/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Execute arbitrary authenticate commands
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"
#include "asterisk/utils.h"
#include "ucm_crypt.h"

enum {
	OPT_ACCOUNT = (1 << 0),
	OPT_DATABASE = (1 << 1),
	OPT_MULTIPLE = (1 << 3),
	OPT_REMOVE = (1 << 4),
};

AST_APP_OPTIONS(auth_app_options, {
	AST_APP_OPTION('a', OPT_ACCOUNT),
	AST_APP_OPTION('d', OPT_DATABASE),
	AST_APP_OPTION('m', OPT_MULTIPLE),
	AST_APP_OPTION('r', OPT_REMOVE),
});


static const char app[] = "Authenticate";
/*** DOCUMENTATION
	<application name="Authenticate" language="en_US">
		<synopsis>
			Authenticate a user
		</synopsis>
		<syntax>
			<parameter name="password" required="true">
				<para>Password the user should know</para>
			</parameter>
			<parameter name="options" required="false">
				<optionlist>
					<option name="a">
						<para>Set the channels' account code to the password that is entered</para>
					</option>
					<option name="d">
						<para>Interpret the given path as database key, not a literal file.</para>
						<note>
							<para>The value is not used at all in the authentication when using this option.
							If the family/key is set to <literal>/pin/100</literal> (value does not matter)
							then the password field needs to be set to <literal>/pin</literal> and the pin entered
							by the user would be authenticated against <literal>100</literal>.</para>
						</note>
					</option>
					<option name="m">
						<para>Interpret the given path as a file which contains a list of account
						codes and password hashes delimited with <literal>:</literal>, listed one per line in
						the file. When one of the passwords is matched, the channel will have
						its account code set to the corresponding account code in the file.</para>
					</option>
					<option name="r">
						<para>Remove the database key upon successful entry (valid with <literal>d</literal> only)</para>
					</option>
				</optionlist>
			</parameter>
			<parameter name="maxdigits" required="false">
				<para>maximum acceptable number of digits. Stops reading after
				maxdigits have been entered (without requiring the user to press the <literal>#</literal> key).
				Defaults to 0 - no limit - wait for the user press the <literal>#</literal> key.</para>
			</parameter>
			<parameter name="prompt" required="false">
				<para>Override the agent-pass prompt file.</para>
			</parameter>
		</syntax>
		<description>
			<para>This application asks the caller to enter a given password in order to continue dialplan execution.</para>
			<para>If the password begins with the <literal>/</literal> character,
			it is interpreted as a file which contains a list of valid passwords, listed 1 password per line in the file.</para>
			<para>When using a database key, the value associated with the key can be anything.</para>
			<para>Users have three attempts to authenticate before the channel is hung up.</para>
		</description>
		<see-also>
			<ref type="application">VMAuthenticate</ref>
			<ref type="application">DISA</ref>
		</see-also>
	</application>
 ***/

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("app_authenticate");
AST_MUTEX_DEFINE_STATIC(pl_mutex);
#define SAFE_FREE(buf) if (buf) {ast_free(buf);	buf = NULL;}

struct pinnode{
	char password[128];
	char name[256];
	struct pinnode *next;
};

struct pinsets{
	char ps_id[64];					//pinsets id
	char addto_cdr[8];				//need add to cdr?
	struct pinnode *pin;				//password and name of pinsets node
	struct pinsets *next;
};

struct pinsets *g_ps = NULL;

static struct pinnode* malloc_pinnode(const char *password, const char *name)
{
	if (NULL == password || NULL == name){
		return NULL;
	}
	struct pinnode *pn = NULL;
	pn = ast_calloc(1, sizeof(struct pinnode));
	if (NULL != pn){
		strncpy(pn->password, password, sizeof(pn->password));
		strncpy(pn->name, name, sizeof(pn->name));
		pn->next = NULL;
	}

	return pn;
}

static void pinsets_destroy(void *obj)
{
	struct pinsets *ps = obj;
	struct pinnode *pn = NULL, *tmp_pn = NULL;

	pn = ps->pin;
	while (pn) {
		tmp_pn = pn;
		pn = pn->next;
		ast_free(tmp_pn);
	}
}

static struct pinsets* malloc_pinsets(const char *to_cdr, const char *ps_id)
{
	if (NULL == to_cdr || NULL == ps_id){
		return NULL;
	}
	struct pinsets *ps = NULL;
	ps = ao2_alloc(sizeof(struct pinsets), pinsets_destroy);
	if (ps){
		strncpy(ps->addto_cdr, to_cdr, sizeof(ps->addto_cdr));
		strncpy(ps->ps_id, ps_id, sizeof(ps->ps_id));
		ps->pin= NULL;
		ps->next= NULL;
	}

	return ps;
}

static struct pinsets *create_pinlist(struct pinsets **ps, char *to_cdr, char *ps_id)
{
	if (NULL == ps || '\0' == to_cdr[0] || '\0' == ps_id[0]){
		return NULL;
	}

	struct pinsets *tmp_ps = *ps, *tmp = NULL;
	tmp = malloc_pinsets(to_cdr, ps_id);
	if (tmp_ps == NULL){
		*ps = tmp;
	}
	else{
		while(tmp_ps->next){
			tmp_ps = tmp_ps->next;
		}
		tmp_ps->next = tmp;
	}
	return tmp;
}

static struct pinnode *create_pinnodes(struct pinsets *ps, char *password, char *name)
{
	if (NULL == ps || '\0' == password[0] || '\0' == name[0]){
		return NULL;
	}
	struct pinnode *tmp_pn = ps->pin;
	struct pinnode *pn = malloc_pinnode(password, name);

	if (NULL == ps->pin){
		ps->pin = pn;
	} else{
		while(tmp_pn->next){
			tmp_pn = tmp_pn->next;
		}
		tmp_pn->next = pn;
	}
	return pn;
}

static struct pinsets *find_pinlist_by_id(const char *pinsets_id){
	if ('\0' == pinsets_id[0]){
		return NULL;
	}
	struct pinsets *tmp_pinsets = NULL;
	ast_mutex_lock(&pl_mutex);
	for(tmp_pinsets=g_ps; tmp_pinsets; tmp_pinsets=tmp_pinsets->next){
		if (0 == strcmp(tmp_pinsets->ps_id, pinsets_id)){
			ao2_ref(tmp_pinsets, +1);
			break;
		}
	}
	ast_mutex_unlock(&pl_mutex);
	return tmp_pinsets;
}

static struct pinnode *check_pinsets(struct pinnode *pn, const char *pinsets_password)
{
	if ('\0' == pinsets_password[0] || NULL == pn){
		return NULL;
	}
	struct pinnode *tmp_pinnode = pn;
	for (; tmp_pinnode; tmp_pinnode=tmp_pinnode->next){
		if (0 == strcmp(tmp_pinnode->password, pinsets_password)){
			ast_log(LOG_NOTICE, "The password is correct");
			return tmp_pinnode;
		}else{
			char *decrypt_pinpwd = NULL;
			ucm_secret_decrypt(tmp_pinnode->password, &decrypt_pinpwd);
			if (decrypt_pinpwd && 0 == strcmp(decrypt_pinpwd, pinsets_password)){
				ast_log(LOG_NOTICE, "The encrypt password is correct");
				SAFE_FREE(decrypt_pinpwd);
				return tmp_pinnode;
			}
			SAFE_FREE(decrypt_pinpwd);
		}
	}
	return NULL;
}

static void destroy_pinlist(void){
	struct pinsets *ps = g_ps, *tmp_ps = NULL;

	while(ps){
		tmp_ps = ps;
		ps = ps->next;
		ao2_cleanup(tmp_ps);
	}
	g_ps = NULL;
}
#endif

static int auth_exec(struct ast_channel *chan, const char *data)
{
	int res = 0, retries, maxdigits;
	char passwd[256], *prompt = "agent-pass", *argcopy = NULL;
	struct ast_flags flags = {0};
#ifdef GRANDSTREAM_NETWORKS
	char pin_name[256] = {0};
	char pin_options[64] = {0};
	RAII_VAR(struct pinsets *, pinset, NULL, ao2_cleanup);
	struct pinnode *pin_node = NULL;
#endif


	AST_DECLARE_APP_ARGS(arglist,
		AST_APP_ARG(password);
		AST_APP_ARG(options);
		AST_APP_ARG(maxdigits);
		AST_APP_ARG(prompt);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "Authenticate requires an argument(password)\n");
		return -1;
	}
#ifdef GRANDSTREAM_NETWORKS
	char special_action_type[32] = {0};
	ast_cdr_getvar(ast_channel_name(chan), "special_action_type", special_action_type,sizeof(special_action_type));
	if (!ast_strlen_zero(special_action_type)){
		char special_action_type_plus[64] = {0};
		snprintf(special_action_type_plus, sizeof(special_action_type_plus)-1, "%s/authenticate", special_action_type);
		ast_cdr_setvar(ast_channel_name(chan),"__special_action_type",special_action_type_plus);
	}else{
	ast_cdr_setvar(ast_channel_name(chan),"special_action_type","authenticate");
	}
#endif

	if (ast_channel_state(chan) != AST_STATE_UP) {
		if ((res = ast_answer(chan)))
			return -1;
	}

	argcopy = ast_strdupa(data);

	AST_STANDARD_APP_ARGS(arglist, argcopy);

#ifdef GRANDSTREAM_NETWORKS
	ast_mutex_lock(&pl_mutex);
	if (strstr(arglist.password, "PIN_SET")){
		pinset = find_pinlist_by_id(arglist.password + 9);
		if (pinset){
			if (0 == strcmp(pinset->addto_cdr, "yes")){
				snprintf(pin_options, sizeof(pin_options), "da");
			} else if (0 == strcmp(pinset->addto_cdr, "no")){
				snprintf(pin_options, sizeof(pin_options), "d");
			}
		} else {
			ast_log(LOG_ERROR, "Can not pin by password, pinset: %p\n", pinset);
			ast_mutex_unlock(&pl_mutex);
			return -1;
		}
	}
	if (!ast_strlen_zero(pin_options)){
		ast_app_parse_options(auth_app_options, &flags, NULL, pin_options);
	} else if (!ast_strlen_zero(arglist.options)){
		ast_app_parse_options(auth_app_options, &flags, NULL, arglist.options);
	}
#else
	if (!ast_strlen_zero(arglist.options))
		ast_app_parse_options(auth_app_options, &flags, NULL, arglist.options);
#endif

	if (!ast_strlen_zero(arglist.maxdigits)) {
		maxdigits = atoi(arglist.maxdigits);
		if ((maxdigits<1) || (maxdigits>sizeof(passwd)-2))
			maxdigits = sizeof(passwd) - 2;
	} else {
		maxdigits = sizeof(passwd) - 2;
	}

	if (!ast_strlen_zero(arglist.prompt)) {
		prompt = arglist.prompt;
	} else {
		prompt = "agent-pass";
	}
#ifdef GRANDSTREAM_NETWORKS /*add by weidu ,for dahdi 166180*/
	if (ast_channel_tech(chan)->type && !strcasecmp(ast_channel_tech(chan)->type, "DAHDI")){
		ast_channel_set_auth(chan, 1);
	}
#endif
	/* Start asking for password */
	for (retries = 0; retries < 3; retries++) {
		if ((res = ast_app_getdata(chan, prompt, passwd, maxdigits, 0)) < 0)
			break;

		res = 0;

		if (arglist.password[0] != '/') {
			/* Compare against a fixed password */
			if (!strcmp(passwd, arglist.password))
				break;
#ifdef GRANDSTREAM_NETWORKS
			else {
				char *decrypt_passwd = NULL;
				ucm_secret_decrypt(arglist.password, &decrypt_passwd);
				if (decrypt_passwd && !strcmp(passwd, decrypt_passwd)){
					SAFE_FREE(decrypt_passwd);
					break;
				}
				SAFE_FREE(decrypt_passwd);
			}
#endif
		} else if (ast_test_flag(&flags,OPT_DATABASE)) {
#ifdef GRANDSTREAM_NETWORKS
			pin_node = check_pinsets(pinset->pin, passwd);
			if (pin_node) {
				snprintf(pin_name, sizeof(pin_name), "%s", pin_node->name);
				break;
			}
#else
			char tmp[256];
			/* Compare against a database key */
			if (!ast_db_get(arglist.password + 1, passwd, tmp, sizeof(tmp))) {
				/* It's a good password */
				if (ast_test_flag(&flags,OPT_REMOVE))
					ast_db_del(arglist.password + 1, passwd);
				break;
			}
#endif

		} else {
			/* Compare against a file */
			FILE *f;
			char buf[256] = "", md5passwd[33] = "", *md5secret = NULL;

			if (!(f = fopen(arglist.password, "r"))) {
				ast_log(LOG_WARNING, "Unable to open file '%s' for authentication: %s\n", arglist.password, strerror(errno));
				continue;
			}

			for (;;) {
				size_t len;

				if (feof(f))
					break;

				if (!fgets(buf, sizeof(buf), f)) {
					continue;
				}

				if (ast_strlen_zero(buf))
					continue;

				len = strlen(buf) - 1;
				if (buf[len] == '\n')
					buf[len] = '\0';

				if (ast_test_flag(&flags, OPT_MULTIPLE)) {
					md5secret = buf;
					strsep(&md5secret, ":");
					if (!md5secret)
						continue;
					ast_md5_hash(md5passwd, passwd);
					if (!strcmp(md5passwd, md5secret)) {
						if (ast_test_flag(&flags, OPT_ACCOUNT)) {
							ast_channel_lock(chan);
							ast_channel_accountcode_set(chan, buf);
							ast_channel_unlock(chan);
						}
						break;
					}
				} else {
					if (!strcmp(passwd, buf)) {
						if (ast_test_flag(&flags, OPT_ACCOUNT)) {
							ast_channel_lock(chan);
							ast_channel_accountcode_set(chan, buf);
							ast_channel_unlock(chan);
						}
						break;
					}
				}
			}

			fclose(f);

			if (!ast_strlen_zero(buf)) {
				if (ast_test_flag(&flags, OPT_MULTIPLE)) {
					if (md5secret && !strcmp(md5passwd, md5secret))
						break;
				} else {
					if (!strcmp(passwd, buf))
						break;
				}
			}
		}
		prompt = "auth-incorrect";
	}
	ast_mutex_unlock(&pl_mutex);

	if ((retries < 3) && !res) {
		if (ast_test_flag(&flags,OPT_ACCOUNT) && !ast_test_flag(&flags,OPT_MULTIPLE)) {
			ast_channel_lock(chan);
#ifdef GRANDSTREAM_NETWORKS
			if(!ast_strlen_zero(pin_name)){
				ast_channel_accountcode_set(chan,pin_name);
			} else {
				ast_channel_accountcode_set(chan, passwd);
			}
#else
			ast_channel_accountcode_set(chan, passwd);
#endif
			ast_channel_unlock(chan);
		}
		if (!(res = ast_streamfile(chan, "auth-thankyou", ast_channel_language(chan))))
			res = ast_waitstream(chan, "");
	} else {
		if (!ast_streamfile(chan, "vm-goodbye", ast_channel_language(chan)))
			res = ast_waitstream(chan, "");
		res = -1;
	}

	return res;
}

#ifdef GRANDSTREAM_NETWORKS
static int load_pinset(void)
{
	char *context = NULL;
	struct ast_config *cfg;
	struct ast_variable *var = NULL;
	struct pinsets *tmp_ps = NULL;
	struct ast_flags config_flags = { 0 };
	if (!(cfg = ast_config_load("pinsets.conf", config_flags))) {
		ast_log(LOG_NOTICE, "No pinsets config file (pinsets.conf), so no pinset\n");
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_log(LOG_NOTICE, "file is not changed\n");
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file pinsets.conf is in an invalid format.  Aborting.\n");
	} else {
		ast_mutex_lock(&pl_mutex);
		while ((context = ast_category_browse(cfg, context) )) {
			char to_cdr[8] = {0};
			char ps_id[64] = {0};
			sscanf(context, "%*[^-]-%7[^-]-%63s", to_cdr, ps_id);
			if ('\0' == to_cdr[0] || '\0' == ps_id[0]){
				break;
			}

			tmp_ps = create_pinlist(&g_ps, to_cdr, ps_id);
			ast_log(LOG_NOTICE, "save to cdr:%s, pinsets id:%s.", to_cdr, ps_id);
			char pin_pwd[128] = {0};
			char pin_name[256] = {0};
			for (var = ast_variable_browse(cfg, context); var; var = var->next) {
				sscanf(var->value, "%127[^,],%255s", pin_pwd, pin_name);
				if ('\0' == pin_pwd[0] || '\0' == pin_name[0]){
					break;
				}

				ast_log(LOG_NOTICE, "pinsets name:%s, pinsets password:%s.", pin_name, pin_pwd);
				if (0 == strncmp(var->name, "pn_pwd",6)) {
					create_pinnodes(tmp_ps, pin_pwd, pin_name);
				}
			}
		}
		ast_mutex_unlock(&pl_mutex);
		ast_config_destroy(cfg);
	}
	return 0;
}
#endif


static int unload_module(void)
{
#ifdef GRANDSTREAM_NETWORKS
	ast_mutex_lock(&pl_mutex);
	destroy_pinlist();
	ast_mutex_unlock(&pl_mutex);
#endif
	return ast_unregister_application(app);
}

static int load_module(void)
{
#ifdef GRANDSTREAM_NETWORKS
	load_pinset();
#endif
	if (ast_register_application_xml(app, auth_exec))
		return AST_MODULE_LOAD_DECLINE;
	return AST_MODULE_LOAD_SUCCESS;
}

#ifdef GRANDSTREAM_NETWORKS
static int reload_module(void)
{
	int res = 0;
	ast_mutex_lock(&pl_mutex);
	destroy_pinlist();
	ast_mutex_unlock(&pl_mutex);
	load_pinset();
	return res;
}
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Authenticate Application with pinsets",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
);
#else
AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Authentication Application");
#endif
