/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief choice a trunk from trunk_group
 *
 * \author yqsheng <yqsheng@grandstream.cn>
 * 
 * The APP returns a channel variable TRUNK_OF_GROUP, and if a trunk is selected, TRUNK_OF_GROUP is set to his index, TRUNK_OF_GROUP if the problem is broken or unavailable.
 *  
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/


#include "asterisk.h"

#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/translate.h"
#include "asterisk/say.h"
#include "asterisk/config.h"
#include "asterisk/features.h"
#include "asterisk/musiconhold.h"
#include "asterisk/callerid.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/causes.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/manager.h"
#include "asterisk/privacy.h"
#include "asterisk/stringfields.h"
#include "asterisk/dsp.h"
#include "asterisk/aoc.h"
#include "asterisk/ccss.h"
#include "asterisk/indications.h"
#include "asterisk/framehook.h"
#include "asterisk/dial.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/bridge_after.h"
#include "asterisk/features_config.h"
#include "asterisk/max_forwards.h"
AST_MODULE_LOG("app_choicetrunk");
/*** DOCUMENTATION
<application name="ChoiceTrunks" language="en_US">
	<synopsis>
		choice a trunk from a group
	</synopsis>
	<syntax>
		<parameter name="group_index" required="true">
			<argument name="index" required="true">
				<para>Voicemail username</para>
			</argument>
		</parameter>
	</syntax>
	<description>
		<para>This application is ChoiceTrunks , configure in <filename>select_trunk.conf</filename></para>
		<note><para> test </para></note>
		<variablelist>
			<variable name="TRUNK_OF_GROUP">
				<para>This is selected trunk index</para>
				<value name="1" />
				<value name="2" />
				<value name="3" />
				<value name="4" />
			</variable>
		</variablelist>
	</description>
</application>
 ***/
#define MAX_TRUNK 24

/*! \brief all sip register trunks in all groups */
struct trunk
{
	int index;	/*!< trunk id */
	int max_count;
	int already_used;
};

/*! \brief sip trunk node which trunk group include */
struct trunk_info
{
	struct trunk *real_addr;	/*!< the addr of trunk list */
	struct trunk_info *next;
};

/*! \brief the sip trunk group node */
struct trunk_group
{
	int group_index;	/*!< group id */
	int is_round;	/*!< the line selection: 1 is round, 0 is linear */
	int allow_call_reg_failed;	/*!< allow out going call when register failed: 1 is allow, 0 is disallow */
	int round_index;	/*!< if is the line selection mode is round, the line which index will been next selected */
	struct trunk_info *trunk_info_head;	/*!< the trunk list of this group */
	struct trunk_group *next;
};

AST_MUTEX_DEFINE_STATIC(trunk_lock);

static struct trunk *trunks = NULL;
static struct trunk_group *group_head = NULL;
static const char * const app = "ChoiceTrunks";
static int group_num = 0;
static int trunk_num = 0;
/*
describe: From the trunk group, select one available trunk and return his index.
*/
static  int is_registered(int index)
{
	int ret = 0;
	char id[20] = {0};
	char state[30] = {0};
	sprintf(id,"trunk_%d",index);
	struct ast_custom_function *custom_function = ast_custom_function_find("GET_TRUNKS_STATE");
	if (custom_function && custom_function->read) {
		custom_function->read(NULL, "GET_TRUNKS_STATE", id, state, sizeof(state) - 1);
		if(state[0] == '\0') {
			ast_log(LOG_ERROR, " get state error %s\n",state);
		} else {
			ast_log(LOG_NOTICE, "state  :%s\n", state);
			if(strcmp(state,"Registered") == 0){
				ast_log(LOG_NOTICE, "choice trunk :%d\n", index);
				return 1;
			}
		}
	}
	return ret;
}
static int  alloc_trunk(int group_index)
{
	int ret = 0;
	struct trunk_group *tmp_group = group_head;
	if(trunk_num == 0 || group_num == 0)
	{
		ast_log(LOG_ERROR, " num == 0 \n");
		return ret;
	}
	while(tmp_group != NULL) {
		/* Find the group of group list by group index. */
		if(tmp_group->group_index == group_index && tmp_group->group_index != 0) {
			struct trunk_info *tmp_trunk = tmp_group->trunk_info_head;
			int have_select = 0;    /* mark the trunk have been select, we need to select the next trunk */
			int first_cycle = 1; /* makr is the first cycle */
			/* The line selection is round */
			if(tmp_group->is_round) {
				ast_log(LOG_NOTICE, "This group[%d] is round select line.\n", group_index);
				/* no trunk have been selectd or all trunk have been selected */
				if (tmp_group->round_index == 0) {
					have_select = 1;
				}

				while(tmp_trunk != NULL && tmp_trunk->real_addr){
					/* find the trunk which have not been selectd */
					if (tmp_group->round_index == tmp_trunk->real_addr->index) {
						/* this is first cycle, mark the select trunk id */
						if (first_cycle == 1) {
							have_select = 1;
							ast_log(LOG_NOTICE, "Start to select from  trunk[%d].\n", tmp_group->round_index);
						/* this is second cycle, all trunk index have been found once, no trunk index can't been found */
						} else {
							ast_log(LOG_NOTICE, "All trunk have been selected, no trunk can been found \n");
							return ret;
						}
					}

					/* the trunk have't been select and not rearch the max count */
					if(have_select == 1 && tmp_trunk->real_addr->max_count > tmp_trunk->real_addr->already_used && tmp_trunk->real_addr->index != 0) {
						if(tmp_group->allow_call_reg_failed || is_registered(tmp_trunk->real_addr->index) == 1) {
							/* this trunk can been select, return this trunk index */
							ret = tmp_trunk->real_addr->index;

							/* Mark the trunk index which will been used next time.*/
							if (tmp_trunk->next != NULL && tmp_trunk->next->real_addr != NULL) {
								tmp_group->round_index = tmp_trunk->next->real_addr->index;
							} else {
								/* If the select id is the end id of this group, set the round_index as 0 */
								tmp_group->round_index = 0;
							}

							ast_log(LOG_NOTICE, "The trunk[%d] is selected, the next trunk is :%d\n", ret, tmp_group->round_index);
							return ret;
						}
					}
					tmp_trunk = tmp_trunk->next;
					/* BUG 224766, if after the last select trunk have noused trunk, we shoud select from the head of trunk list again */
					if (tmp_trunk == NULL && first_cycle == 1) {
						ast_log(LOG_NOTICE, "At the end of trunk list, selected the trunk list again.\n");
						tmp_trunk = tmp_group->trunk_info_head;
						first_cycle = 0;
					}
				}
			} else {
				ast_log(LOG_NOTICE, "This group[%d] is linear select line.\n", group_index);
				while(tmp_trunk != NULL && tmp_trunk->real_addr){
					if(tmp_trunk->real_addr->max_count > tmp_trunk->real_addr->already_used  && tmp_trunk->real_addr->index != 0) {
						ret = tmp_trunk->real_addr->index;
						if(tmp_group->allow_call_reg_failed || is_registered(ret) == 1)
							return ret;
					}
					tmp_trunk = tmp_trunk->next;
				}
			}
		}
		tmp_group = tmp_group->next;
	}
	return ret;
}
/*
describe: Every time this APP is called, this function is used to update the SIP array.
*/
static void update_trunks(struct ast_channel *chan)
{
	int count = -1;
	char group[80] = "", category[80] = "";
	int i;
	for(i = 0; i < trunk_num; i++) {
		if(trunks[i].index == 0)
			continue;
		char *data = (char *)ast_malloc(30);
		sprintf(data, "OUT_%d", trunks[i].index);
		ast_app_group_split_group(data, group, sizeof(group), category,
				  sizeof(category));
		if (ast_strlen_zero(group)) {
			struct ast_group_info *gi = NULL;

			ast_app_group_list_rdlock();
			for (gi = ast_app_group_list_head(); gi; gi = AST_LIST_NEXT(gi, group_list)) {
				if (gi->chan != chan)
					continue;
				if (ast_strlen_zero(category) || (!ast_strlen_zero(gi->category) && !strcasecmp(gi->category, category)))
					break;
			}
			if (gi) {
				ast_copy_string(group, gi->group, sizeof(group));
				if (!ast_strlen_zero(gi->category))
					ast_copy_string(category, gi->category, sizeof(category));
			}
			ast_app_group_list_unlock();
		}
		if((count = ast_app_group_get_count(group, category)) == -1) {
			ast_log(LOG_NOTICE, "No group could be found for channel '%s'\n", ast_channel_name(chan));
		}else {
			ast_log(LOG_NOTICE, "[%d] used by trunk[%s]could be found for channel '%s'\n", count, data, ast_channel_name(chan));
			trunks[i].already_used = count;
		}
		if(data != NULL) {
			ast_free(data);
		}
	}
}

static int choice_exec(struct ast_channel *chan, const char *data)
{
	ast_mutex_lock(&trunk_lock);
	update_trunks(chan);
	ast_mutex_unlock(&trunk_lock);

	int res=0;
	char *parse;
	int ret = 0;
	char ret_str[24] = {0};
	char index_str[24] = {0};
	int var = 0;
	AST_DECLARE_APP_ARGS(args,
				 AST_APP_ARG(param1);
	);

	struct ast_module_user *u;
	const char *in_ringgroup = NULL;
	const char *ringboth_checkper = NULL;
	const char *ringboth_trunk = NULL;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "this app requires an argument (param1)\n");
		return -1;
	}

	u = ast_module_user_add(chan);
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.param1)) {
		ast_log(LOG_WARNING, "this app requires an argument (group num)\n");
		res = -1;
		goto out1;
	}

	if((var = atoi(args.param1)) != 0) {
		/* if is ring both */
		in_ringgroup = pbx_builtin_getvar_helper(chan, "IN_RINGGROUP");
		if (!ast_strlen_zero(in_ringgroup) && strcasecmp(in_ringgroup, "yes") ==0) {
			ringboth_checkper = pbx_builtin_getvar_helper(chan, "ringboth_checkper");
			if (!ast_strlen_zero(ringboth_checkper) && strcasecmp(ringboth_checkper, "yes") ==0) {
				snprintf(index_str, sizeof(index_str), "RINGBOTH_TRUNK_%d", var);
			}
		}

		/* if have get the trunk id by this ringgboth ,no need to choice again */
		if (index_str[0] != '\0') {
			ringboth_trunk = pbx_builtin_getvar_helper(chan, index_str);
			if (!ast_strlen_zero(ringboth_trunk)) {
				ret = atoi(ringboth_trunk);
				sprintf(ret_str, "%d", ret);
				ast_log(LOG_DEBUG, "Is ring both. the trunk index is %d\n", ret);
			}
		}

		/* if is't ringboth and have no trunk selected */
		if (ret == 0) {
			ast_mutex_lock(&trunk_lock);
			ret = alloc_trunk(var);
			ast_mutex_unlock(&trunk_lock);
			sprintf(ret_str, "%d", ret);
			if (ret > 0 && index_str[0] != '\0') {
				pbx_builtin_setvar_helper(chan, index_str, ret_str);
				ast_log(LOG_DEBUG, "Update %s use the trunk %s\n", index_str, ret_str);
			}
		}
	}else{
		ast_log(LOG_ERROR, "this app  argument error \n");
	}
out1:
	ast_channel_lock(chan);
	//The channel variable is used to interact with the diaplan, and the following is the assignment statement, which is passed through the parameters, and remember to lock it.
	ast_channel_stage_snapshot(chan);
	pbx_builtin_setvar_helper(chan, "TRUNK_OF_GROUP", ret_str);
	ast_channel_unlock(chan);

	ast_module_user_remove(u);
	return res;
}

static void add_group(struct trunk_group *tmp)
{
	if(group_head == NULL) {
		group_head = tmp;
	}else {
		struct trunk_group *p = group_head;

		while(p->next != NULL) {
			p = p->next;
		}
		p->next = tmp;
	}
}

static void add_trunk_info(struct trunk_group *tmp_group, struct trunk_info *tmp_info)
{
	if(tmp_group->trunk_info_head == NULL) {
		tmp_group->trunk_info_head = tmp_info;
	}else {
		struct trunk_info *p = tmp_group->trunk_info_head;
		while(p->next != NULL) {
			p = p->next;
		}
		p->next = tmp_info;
	}
}
/*
describe: Gets the trunk index in the array.
*/
static int get_coor(int index)
{
	int i;
	for(i = 0;i < trunk_num; i++) {
		if(index == trunks[i].index) {
			return i;
		}
	}
	return -1;
}
static void free_info(struct trunk_info *p)
{
	struct trunk_info * tmp = p;
	while(p != NULL){
		p = p->next;
		ast_free(tmp);
		tmp = p;
	}
}

static void free_list(struct trunk_group *p)
{
	struct trunk_group * tmp = p;
	while(p != NULL){
		p = p->next;
		free_info(tmp->trunk_info_head);
		ast_free(tmp);
		tmp = p;
	}
}
/*
describe: Clears the array and frees the list space.
*/
static void clear_all(void)
{
	//clear array
	ast_log(LOG_NOTICE, "free all\n");

	ast_mutex_lock(&trunk_lock);
	if(trunks != NULL) {
		ast_free(trunks);
		trunks = NULL;
	}

	//clear list
	struct trunk_group *tmp_group = group_head;
	free_list(tmp_group);
	group_head = NULL;

	ast_mutex_unlock(&trunk_lock);

	group_num = 0;
	trunk_num = 0;
}
/*
describe: When loading a module or reloading the module, load the configuration file to generate the corresponding data structure.\
			Call the select_trunk. Conf file.

*/
static int load_config(int reload)
{

	//if conf is not exist , create it
	if((access("/cfg/etc/asterisk_new/select_trunk.conf",F_OK)) == -1) {
		FILE *fp=fopen("/cfg/etc/asterisk_new/select_trunk.conf","w");
		if(fp == NULL) {
			return -1;
		}
		fprintf(fp," ");
		fclose(fp);
	}
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	const char *temp = NULL;
	struct trunk *tmp_trunk = NULL;
	struct trunk_group *tmp = NULL;
	char tmp_string[1024] = {0};
	int ret = -1;

	if (!(cfg = ast_config_load("select_trunk.conf", config_flags))) {
		ast_log(LOG_ERROR, "Configuration file select_trunk.conf missing.\n");
		return -1;
	}else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	}else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file select_trunk.conf is in an invalid format.  Aborting.\n");
		return -1;
	}
	clear_all();

	if (!(temp = ast_variable_retrieve(cfg, "GROUP", "trunk_num"))) {
		trunk_num = 0;
	}else {
		trunk_num = atoi(temp);

		tmp_trunk = (struct trunk *)ast_malloc(trunk_num*sizeof(struct trunk)+1);
		if(tmp_trunk == NULL){
			ast_log(LOG_ERROR, "memerory is full\n");
			goto reorder;
		}
	}

	ast_mutex_lock(&trunk_lock);
	trunks = tmp_trunk;

	if(trunk_num != 0){
		if (!(temp = ast_variable_retrieve(cfg, "GROUP", "trunk"))) {
			trunk_num = 0;
		}else {
			const char* delim=",";
			char *p_save = NULL;
			snprintf(tmp_string, sizeof(tmp_string) - 1, "%s", temp);
			char* p=strtok_r(tmp_string, delim, &p_save);
			int i = 0;
			while(p != NULL){
				trunks[i].index = atoi(p);
				trunks[i].already_used = 0;
				i++;
				p = strtok_r(NULL, delim, &p_save);
			}
			memset(tmp_string, '\0', sizeof(tmp_string));
		}
		if (!(temp = ast_variable_retrieve(cfg, "GROUP", "trunk_max"))) {
			trunk_num = 0;
		}else {
			const char* delim=",";
			char *p_save = NULL;
			snprintf(tmp_string, sizeof(tmp_string) - 1, "%s", temp);
			char* p=strtok_r(tmp_string, delim, &p_save);
			int i = 0;
			while(p!=NULL) {
				trunks[i].max_count = atoi(p);
				if(trunks[i].max_count == 0){
					trunks[i].max_count = 200;
				}
				i++;
				p=strtok_r(NULL, delim, &p_save);
			}
			memset(tmp_string, '\0', sizeof(tmp_string));
		}
	}

	if (!(temp = ast_variable_retrieve(cfg, "GROUP", "group_num"))) {
		group_num = 0;
	}else {
		group_num = atoi(temp);
		if(group_num == 0){
			ast_log(LOG_DEBUG, "Have no group\n");
		}
	}

	if(group_num != 0) {
		if ((temp = ast_variable_retrieve(cfg, "GROUP", "group"))!= NULL) {
			const char* delim=",";
			char *p_save = NULL;
			snprintf(tmp_string, sizeof(tmp_string) - 1, "%s", temp);
			char* p=strtok_r(tmp_string, delim, &p_save);
			while(p!=NULL){
				tmp = ast_malloc(sizeof(struct trunk_group));
				if(tmp == NULL)
				{
					ast_log(LOG_ERROR, "memerory is full\n");
					ast_mutex_unlock(&trunk_lock);
					goto reorder;
				}
				tmp->trunk_info_head = NULL;
				tmp->next = NULL;
				tmp->group_index = atoi(p);
				tmp->round_index = 0;
				add_group(tmp);
				p = strtok_r(NULL, delim, &p_save);
			}
			memset(tmp_string, '\0', sizeof(tmp_string));
		}
	}

	for(tmp = group_head; tmp != NULL && trunk_num != 0; tmp = tmp->next) {
		char group_name[100];
		sprintf(group_name, "%d", tmp->group_index);
		if (!(temp = ast_variable_retrieve(cfg, group_name, "include"))) {
			tmp->trunk_info_head = NULL;
		}else {
			const char* delim=",";
			char *p_save = NULL;
			snprintf(tmp_string, sizeof(tmp_string) - 1, "%s", temp);
			char* p=strtok_r(tmp_string, delim, &p_save);
			while(p != NULL) {
				int tmp_index = 0;
				if((tmp_index = get_coor(atoi(p))) != -1) {
					struct trunk_info *tmp_info = ast_malloc(sizeof(struct trunk_info));
					tmp_info->real_addr = &trunks[tmp_index];
					tmp_info->next = NULL;
					add_trunk_info(tmp, tmp_info);
				}
				p=strtok_r(NULL, delim, &p_save);
			}
			memset(tmp_string, '\0', sizeof(tmp_string));
		}
		if ((temp = ast_variable_retrieve(cfg, group_name, "line_selection"))!= NULL && !strcmp(temp, "round")) {
			tmp->is_round = 1;
		}else {
			tmp->is_round = 0;
		}

		if ((temp = ast_variable_retrieve(cfg, group_name, "allow_outgoing_calls_if_reg_failed"))!= NULL && !strcmp(temp, "yes")) {
			tmp->allow_call_reg_failed = 1;
		}else {
			tmp->allow_call_reg_failed = 0;
		}
	}
	ast_mutex_unlock(&trunk_lock);
	ret = 0;

reorder:
	ast_config_destroy(cfg);
	ast_log(LOG_NOTICE, "load_config\n");
	return ret;
}


static int unload_module(void)
{
	int res;
	//free the space
	clear_all();

	res = ast_unregister_application(app);
	return res;
}

static int load_module(void)
{
	int res;
	res = load_config(0);
	res |= ast_register_application_xml(app, choice_exec);
	return res;
}

static int reload(void)
{
	int res = 0;
	res = load_config(1);
	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "This Application will select trunk from a trunk group",
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
);
