/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (c) 2004 - 2006 Digium, Inc.  All rights reserved.
 *
 * Mark Spencer <markster@digium.com>
 *
 * This code is released under the GNU General Public License
 * version 2.0.  See LICENSE for more information.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 */

/*! \file
 *
 * \brief page() - Paging application
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<depend>app_confbridge</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#ifdef GRANDSTREAM_NETWORKS
#include <sys/stat.h>
#endif

#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/file.h"
#include "asterisk/app.h"
#include "asterisk/chanvars.h"
#include "asterisk/utils.h"
#include "asterisk/devicestate.h"
#include "asterisk/dial.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/paths.h"	/* use ast_config_AST_DATA_DIR */
#endif

/*** DOCUMENTATION
	<application name="Page" language="en_US">
		<synopsis>
			Page series of phones
		</synopsis>
		<syntax>
			<parameter name="Technology/Resource" required="true" argsep="&amp;">
				<argument name="Technology/Resource" required="true">
					<para>Specification of the device(s) to dial. These must be in the format of
					<literal>Technology/Resource</literal>, where <replaceable>Technology</replaceable>
					represents a particular channel driver, and <replaceable>Resource</replaceable> represents a resource
					available to that particular channel driver.</para>
				</argument>
				<argument name="Technology2/Resource2" multiple="true">
					<para>Optional extra devices to dial in parallel</para>
					<para>If you need more than one, enter them as Technology2/Resource2&amp;
					Technology3/Resource3&amp;.....</para>
				</argument>
			</parameter>
			<parameter name="options">
				<optionlist>
				<option name="b" argsep="^">
					<para>Before initiating an outgoing call, Gosub to the specified
					location using the newly created channel.  The Gosub will be
					executed for each destination channel.</para>
					<argument name="context" required="false" />
					<argument name="exten" required="false" />
					<argument name="priority" required="true" hasparams="optional" argsep="^">
						<argument name="arg1" multiple="true" required="true" />
						<argument name="argN" />
					</argument>
				</option>
				<option name="B" argsep="^">
					<para>Before initiating the outgoing call(s), Gosub to the specified
					location using the current channel.</para>
					<argument name="context" required="false" />
					<argument name="exten" required="false" />
					<argument name="priority" required="true" hasparams="optional" argsep="^">
						<argument name="arg1" multiple="true" required="true" />
						<argument name="argN" />
					</argument>
				</option>
					<option name="d">
						<para>Full duplex audio</para>
					</option>
					<option name="i">
						<para>Ignore attempts to forward the call</para>
					</option>
					<option name="q">
						<para>Quiet, do not play beep to caller</para>
					</option>
					<option name="r">
						<para>Record the page into a file (<literal>CONFBRIDGE(bridge,record_conference)</literal>)</para>
					</option>
					<option name="s">
						<para>Only dial a channel if its device state says that it is <literal>NOT_INUSE</literal></para>
					</option>
					<option name="A">
						<argument name="x" required="true">
							<para>The announcement to playback to all devices</para>
						</argument>
						<para>Play an announcement to all paged participants</para>
					</option>
					<option name="n">
						<para>Do not play announcement to caller (alters <literal>A(x)</literal> behavior)</para>
					</option>
				</optionlist>
			</parameter>
			<parameter name="timeout">
				<para>Specify the length of time that the system will attempt to connect a call.
				After this duration, any page calls that have not been answered will be hung up by the
				system.</para>
			</parameter>
		</syntax>
		<description>
			<para>Places outbound calls to the given <replaceable>technology</replaceable>/<replaceable>resource</replaceable>
			and dumps them into a conference bridge as muted participants. The original
			caller is dumped into the conference as a speaker and the room is
			destroyed when the original caller leaves.</para>
		</description>
		<see-also>
			<ref type="application">ConfBridge</ref>
		</see-also>
	</application>
 ***/
static const char * const app_page= "Page";
#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("page");
#endif

enum page_opt_flags {
	PAGE_DUPLEX = (1 << 0),
	PAGE_QUIET = (1 << 1),
	PAGE_RECORD = (1 << 2),
	PAGE_SKIP = (1 << 3),
	PAGE_IGNORE_FORWARDS = (1 << 4),
	PAGE_ANNOUNCE = (1 << 5),
	PAGE_NOCALLERANNOUNCE = (1 << 6),
	PAGE_PREDIAL_CALLEE = (1 << 7),
	PAGE_PREDIAL_CALLER = (1 << 8),
#ifdef GRANDSTREAM_NETWORKS
	PAGE_PROMPT_RECORD = (1 << 9),
	PAGE_CALLER_INTERCOM = (1 << 10),
#endif
};

enum {
	OPT_ARG_ANNOUNCE = 0,
	OPT_ARG_PREDIAL_CALLEE = 1,
	OPT_ARG_PREDIAL_CALLER = 2,
	OPT_ARG_ARRAY_SIZE = 3,
};

AST_APP_OPTIONS(page_opts, {
	AST_APP_OPTION_ARG('b', PAGE_PREDIAL_CALLEE, OPT_ARG_PREDIAL_CALLEE),
	AST_APP_OPTION_ARG('B', PAGE_PREDIAL_CALLER, OPT_ARG_PREDIAL_CALLER),
	AST_APP_OPTION('d', PAGE_DUPLEX),
	AST_APP_OPTION('q', PAGE_QUIET),
	AST_APP_OPTION('r', PAGE_RECORD),
	AST_APP_OPTION('s', PAGE_SKIP),
	AST_APP_OPTION('i', PAGE_IGNORE_FORWARDS),
	AST_APP_OPTION_ARG('A', PAGE_ANNOUNCE, OPT_ARG_ANNOUNCE),
	AST_APP_OPTION('n', PAGE_NOCALLERANNOUNCE),
#ifdef GRANDSTREAM_NETWORKS
	AST_APP_OPTION('c', PAGE_CALLER_INTERCOM),
#endif
});

/* We use this structure as a way to pass this to all dialed channels */
struct page_options {
	char *opts[OPT_ARG_ARRAY_SIZE];
	struct ast_flags flags;
};

#ifdef GRANDSTREAM_NETWORKS
AST_MUTEX_DEFINE_STATIC(page_mutex);

#define PAGE_MAX_MEMBERS 1024
#define CONTACTS_MAX_LENTH 256

/*nodes of members number*/
struct numbers
{
	char members[PAGE_MAX_MEMBERS];
	struct numbers *next;
};

/*list of head numbers*/
struct headnum_list
{
	char headnum[AST_MAX_EXTENSION];/*this can be defined as ivr-id or paging group number or so*/
	struct numbers *extension;
	struct headnum_list *next;
};

/*a global param used for data transfer*/
struct headnum_list *g_pg_list = NULL;

/*init caller id numbers list, insert first number*/
static struct numbers *malloc_numbers(const char *members)
{
	if (members == NULL) {
		return NULL;
	}
	struct numbers *ext = NULL;
	ext = ast_malloc(sizeof(struct numbers));
	if (ext != NULL) {
		strncpy(ext->members, members, sizeof(ext->members));
		ext->next = NULL;
	}
	return ext;
}

/*create head number list*/
static struct numbers *create_numbers_list(struct headnum_list *list, const char *extension)
{
	if (list == NULL || extension== NULL) {
		return NULL;
	}

	struct numbers *node_ext = malloc_numbers(extension);
	struct numbers *tmp_ext = list->extension;
	if (list->extension == NULL) {
		list->extension = node_ext;
	}
	else {
		while (tmp_ext->next) {
			tmp_ext = tmp_ext->next;
		}
		tmp_ext->next = node_ext;
	}

	return node_ext;
}

/*init ivr list*/
static struct headnum_list* malloc_list(const char *head_num)
{
	if (head_num == NULL) {
		return NULL;
	}

	struct headnum_list *list = NULL;
	list = ast_malloc(sizeof(struct headnum_list));
	if (list) {
		strncpy(list->headnum, head_num, sizeof(list->headnum));
		list->extension = NULL;
		list->next = NULL;
	}

	return list;
}

/*create head number list*/
static struct headnum_list *create_list(struct headnum_list **list, const char *headnum)
{
	if (list == NULL || headnum == NULL) {
		return NULL;
	}

	struct headnum_list *tmp_list = *list, *tmp = NULL;
	tmp = malloc_list(headnum);
	if (tmp_list == NULL) {
		*list = tmp;
	}
	else {
		while( tmp_list->next ) {
			tmp_list = tmp_list->next;
		}
		tmp_list->next = tmp;
	}

	return tmp;
}

/*to see wether global saved right*/
static void print_global_list(void)
{
	struct headnum_list *tmp_list = g_pg_list;
	struct numbers *tmp_ext = NULL;
	ast_log(LOG_DEBUG, "start to show paging group members");
	while (tmp_list) {
		ast_debug(1, "[%s]", tmp_list->headnum);
		if(tmp_list->extension != NULL){
			tmp_ext = tmp_list->extension;
			while (tmp_ext) {
				ast_log(LOG_DEBUG,"members = %s", tmp_ext->members);
				tmp_ext = tmp_ext->next;
			}
		}
		tmp_list = tmp_list->next;
	}
	ast_log(LOG_DEBUG, "end to show paging group members");
}

static void unload_list(void)
{
	struct headnum_list *tmp_list = g_pg_list, *prev_list = NULL;
	struct numbers *tmp_ext = NULL, *prev_ext = NULL;
	while (tmp_list) {
		if(tmp_list->headnum != NULL){
			tmp_ext = tmp_list->extension;
		}

		while (tmp_ext) {
			prev_ext = tmp_ext;
			tmp_ext = tmp_ext->next;
			ast_free(prev_ext);
		}
		prev_list = tmp_list;
		tmp_list = tmp_list->next;
		ast_free(prev_list);
	}
	g_pg_list = NULL;
}

static char *get_members(struct numbers *ext)
{
	if (ext == NULL) {
		ast_log(LOG_DEBUG,"ext error");
		return NULL;
	}
	struct numbers *tmp_ext = ext;
	while (tmp_ext != NULL) {
		if (NULL != tmp_ext->members) {
			ast_log(LOG_DEBUG, "here find members: %s", tmp_ext->members);
			break;
		}
		tmp_ext = tmp_ext->next;
	}

	return tmp_ext ? tmp_ext->members : NULL;
}


/*find list by head number*/
static char *find_members_by_headnum(struct headnum_list *list, const char *headnum)
{
	if (list == NULL || headnum == NULL) {
		return NULL;
	}
	struct headnum_list *tmp_list = list;
	while (tmp_list != NULL) {
		if (!strncmp(tmp_list->headnum, headnum, sizeof(tmp_list->headnum))) {
			ast_log(LOG_DEBUG, "here find paging group: %s", tmp_list->headnum);
			break;
		}
		tmp_list = tmp_list->next;
	}

	return tmp_list ? get_members(tmp_list->extension) : NULL;
}
#endif

/*!
 * \internal
 * \brief Setup the page bridge profile.
 *
 * \param chan Setup bridge profile on this channel.
 * \param options Options to setup bridge profile.
 *
 * \return Nothing
 */
static void setup_profile_bridge(struct ast_channel *chan, struct page_options *options)
{
	/* Use default_bridge as a starting point */
	ast_func_write(chan, "CONFBRIDGE(bridge,template)", "");
	if (ast_test_flag(&options->flags, PAGE_RECORD)) {
		ast_func_write(chan, "CONFBRIDGE(bridge,record_conference)", "yes");
	}
}

/*!
 * \internal
 * \brief Setup the paged user profile.
 *
 * \param chan Setup user profile on this channel.
 * \param options Options to setup paged user profile.
 *
 * \return Nothing
 */
static void setup_profile_paged(struct ast_channel *chan, struct page_options *options)
{
	/* Use default_user as a starting point */
	ast_func_write(chan, "CONFBRIDGE(user,template)", "");
	ast_func_write(chan, "CONFBRIDGE(user,quiet)", "yes");
	ast_func_write(chan, "CONFBRIDGE(user,end_marked)", "yes");
	if (!ast_test_flag(&options->flags, PAGE_DUPLEX)) {
		ast_func_write(chan, "CONFBRIDGE(user,startmuted)", "yes");
	}
	if (ast_test_flag(&options->flags, PAGE_ANNOUNCE)
		&& !ast_strlen_zero(options->opts[OPT_ARG_ANNOUNCE])) {
		ast_func_write(chan, "CONFBRIDGE(user,announcement)", options->opts[OPT_ARG_ANNOUNCE]);
	}
}

/*!
 * \internal
 * \brief Setup the caller user profile.
 *
 * \param chan Setup user profile on this channel.
 * \param options Options to setup caller user profile.
 *
 * \return Nothing
 */
static void setup_profile_caller(struct ast_channel *chan, struct page_options *options)
{
	/* Use default_user as a starting point if not already setup. */
	ast_func_write(chan, "CONFBRIDGE(user,template)", "");
	ast_func_write(chan, "CONFBRIDGE(user,quiet)", "yes");
	ast_func_write(chan, "CONFBRIDGE(user,marked)", "yes");
	if (!ast_test_flag(&options->flags, PAGE_NOCALLERANNOUNCE)
		&& ast_test_flag(&options->flags, PAGE_ANNOUNCE)
		&& !ast_strlen_zero(options->opts[OPT_ARG_ANNOUNCE])) {
		ast_func_write(chan, "CONFBRIDGE(user,announcement)", options->opts[OPT_ARG_ANNOUNCE]);
	}
}

static void page_state_callback(struct ast_dial *dial)
{
	struct ast_channel *chan;
	struct page_options *options;

	if (ast_dial_state(dial) != AST_DIAL_RESULT_ANSWERED ||
	    !(chan = ast_dial_answered(dial)) ||
	    !(options = ast_dial_get_user_data(dial))) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&options->flags, PAGE_PROMPT_RECORD)) {
		ast_channel_check_need_prompt_start_stop_record(chan, NULL, 1);
	}
	if (ast_test_flag(&options->flags, PAGE_CALLER_INTERCOM)) {
		ast_channel_callerintercom_set(chan, 1);
	}
#endif

	setup_profile_bridge(chan, options);
	setup_profile_paged(chan, options);
}

static void set_chan_mp3_flag(struct ast_channel *chan, struct page_options options)
{
	struct stat st;
	char *fn = NULL;
	char *is_multipage = NULL;

	if (!chan) {
		return;
	}

	is_multipage = (char *)pbx_builtin_getvar_helper(chan, "IS_MULTIPAGE");
	if (ast_strlen_zero(is_multipage)) {
		return;
	}

	if (strcasecmp(is_multipage, "yes")) {
		return;
	}

	if (!ast_test_flag(&options.flags, PAGE_ANNOUNCE)) {
		return;
	}

	if (ast_strlen_zero(options.opts[OPT_ARG_ANNOUNCE])) {
		return;
	}

	if (ast_asprintf(&fn, "%s/sounds/%s", ast_config_AST_DATA_DIR, options.opts[OPT_ARG_ANNOUNCE]) < 0) {
		ast_log(LOG_ERROR, "Can't set file exist path\n");
		return;
	}

	ast_log(LOG_DEBUG, "Judge if use mp3 prompt, file:%s\n", fn);

	if (fn == NULL) {
		ast_log(LOG_ERROR, "Can't get file exist path\n");
		return;
	}

	if (stat(fn, &st)) { /* file not existent */
		ast_free(fn);
		fn = NULL;
		return;
	}

	ast_log(LOG_DEBUG, "The prompt is mp3 format, chan:%s set use mp3\n", ast_channel_name(chan));

	ast_channel_set_mp3(chan, 1);

	ast_free(fn);
	fn = NULL;

	return;
}

static int page_exec(struct ast_channel *chan, const char *data)
{
	char *tech;
	char *resource;
	char *tmp;
	char *predial_callee = NULL;
	char confbridgeopts[128];
	char originator[AST_CHANNEL_NAME];
	struct page_options options = { { 0, }, { 0, } };
	unsigned int confid = ast_random();
	struct ast_app *app;
	int res = 0;
	int pos = 0;
	int i = 0;
	struct ast_dial **dial_list;
#ifdef GRANDSTREAM_NETWORKS
	unsigned int num_dials = 0;
	unsigned int record_enable = 0;
#else
	unsigned int num_dials;
#endif
	int timeout = 0;
	char *parse;
#ifdef GRANDSTREAM_NETWORKS
	/*BEGIN: Add by weidu , for 108273*/
	struct timeval calldurationlimit = { 0, };
	char  *dur_limit      = NULL;
	int   duration        = 0;
	/*END: Add by weidu */
	char *page_num = NULL;
	char *extension = NULL;
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(devices);
		AST_APP_ARG(options);
		AST_APP_ARG(timeout);
	);

/*begin: add by weidu ,for 108273*/
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "page_exec Enter.............!\n");
#endif
/*end: add by weidu ,for 108273*/
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires at least one argument (destination(s) to page)\n");
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	/*add by weidu ,for 110820, is paging*/
	ast_channel_paging_set(chan, 1);
#endif

	if (!(app = pbx_findapp("ConfBridge"))) {
		ast_log(LOG_WARNING, "There is no ConfBridge application available!\n");
		return -1;
	};

	parse = ast_strdupa(data);

	AST_STANDARD_APP_ARGS(args, parse);

	ast_copy_string(originator, ast_channel_name(chan), sizeof(originator));
	if ((tmp = strchr(originator, '-'))) {
		*tmp = '\0';
	}

	if (!ast_strlen_zero(args.options)) {
		ast_app_parse_options(page_opts, &options.flags, options.opts, args.options);
	}

	if (!ast_strlen_zero(args.timeout)) {
		timeout = atoi(args.timeout);
	}

	/*begin: add by weidu , for 108273*/
#ifdef GRANDSTREAM_NETWORKS
	dur_limit = (char *)pbx_builtin_getvar_helper(chan, "PAGE_DUR_LIMIT");
	if ( ast_strlen_zero(dur_limit) || atoi(dur_limit) <= 0 )
		dur_limit = (char *)pbx_builtin_getvar_helper(chan, "CALLER_DUR_LIMIT");
	if (!ast_strlen_zero(dur_limit)) {
		duration = atoi(dur_limit);
		calldurationlimit.tv_sec = duration / 1000;
		calldurationlimit.tv_usec = (duration % 1000) * 1000;
		ast_log(LOG_DEBUG, "page_exec Setting call duration limit to %.3lf seconds.\n",
				calldurationlimit.tv_sec + calldurationlimit.tv_usec / 1000000.0);
	}

	record_enable = ast_strlen_zero(pbx_builtin_getvar_helper(chan, "PAGE_RECORD_FILE")) ? 0 : 1;
	const char *frompage = pbx_builtin_getvar_helper(chan, "PAGING_ALERT_INFO");
	if (frompage != NULL) {
		ast_channel_set_pagecall_flag(chan, 1);
		const char* tmp_chan = ast_channel_name(chan);
		if (NULL != strcasestr(tmp_chan, "Local") && NULL != strcasestr(tmp_chan, "call-forward") && NULL != strstr(tmp_chan, ";2")){
			char another_local_chan[256] = {0};
			struct ast_channel *Local_chan = NULL;
			snprintf(another_local_chan, sizeof(another_local_chan), "%s", tmp_chan);
			another_local_chan[strlen(another_local_chan)-1] = '1';
			Local_chan = ast_channel_get_by_name(another_local_chan);
			if(Local_chan){
				ast_channel_set_pagecall_flag(Local_chan, 1);
				ast_channel_unref(Local_chan);
			}
		}
	}
#endif
	/*end: add by weidu , for 108273*/

	snprintf(confbridgeopts, sizeof(confbridgeopts), "ConfBridge,%u", confid);

#ifndef GRANDSTREAM_NETWORKS
	/* Count number of extensions in list by number of ampersands + 1 */
	num_dials = 1;
	tmp = args.devices;
	while (*tmp) {
		if (*tmp == '&') {
			num_dials++;
		}
		tmp++;
	}
#endif

	/* PREDIAL: Preprocess any callee gosub arguments. */
	if (ast_test_flag(&options.flags, PAGE_PREDIAL_CALLEE)
		&& !ast_strlen_zero(options.opts[OPT_ARG_PREDIAL_CALLEE])) {
		ast_replace_subargument_delimiter(options.opts[OPT_ARG_PREDIAL_CALLEE]);
		predial_callee =
			(char *) ast_app_expand_sub_args(chan, options.opts[OPT_ARG_PREDIAL_CALLEE]);
	}

	/* PREDIAL: Run gosub on the caller's channel */
	if (ast_test_flag(&options.flags, PAGE_PREDIAL_CALLER)
		&& !ast_strlen_zero(options.opts[OPT_ARG_PREDIAL_CALLER])) {
		ast_replace_subargument_delimiter(options.opts[OPT_ARG_PREDIAL_CALLER]);
		ast_app_exec_sub(NULL, chan, options.opts[OPT_ARG_PREDIAL_CALLER], 0);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_test_flag(&options.flags, PAGE_DUPLEX)) {
		ast_log(LOG_DEBUG, "Options has announce flag, set channel:%s page mode for announce\n", ast_channel_name(chan));
		ast_channel_page_mode_set(chan, AST_PAGE_MODE_ANNOUNCE);
	} else {
		ast_log(LOG_DEBUG, "Options has duplex flag, set channel:%s page mode for duplex\n", ast_channel_name(chan));
		ast_channel_page_mode_set(chan, AST_PAGE_MODE_DUPLEX);
	}

	/*Start added by zyxia for video paging support,2022.8.9*/
	const char* page_video_support = pbx_builtin_getvar_helper(chan,"PAGE_VIDEO_SUPPORT");
	if (page_video_support && !strcasecmp(page_video_support,"yes")) {
		ast_log(LOG_DEBUG, "set channel:%s page video\n", ast_channel_name(chan));
		ast_channel_set_is_page_video_support(chan, 1);
	}
	else
		ast_channel_set_is_page_video_support(chan, 0);
	/*End added by zyxia for video paging support,2022.8.9*/

	set_chan_mp3_flag(chan, options);

	page_num = (char *)pbx_builtin_getvar_helper(chan, "PAGE_GROUP_NUM");

	ast_mutex_lock(&page_mutex);
	char *tmp_list = find_members_by_headnum(g_pg_list, page_num);
	char *member_list = NULL;
	char *original_member_list = NULL;

	if (tmp_list) {
		member_list = ast_strndup(tmp_list, strlen(tmp_list));
		original_member_list = member_list;
	}

	ast_mutex_unlock(&page_mutex);

	char buf[1024] = {0};
	char function[128] = {0};
	char members_buf[CONTACTS_MAX_LENTH*120] = {0};
	char *members = members_buf;
	while (member_list != NULL && (extension = strsep(&member_list, "&"))) {
		memset(buf, 0, sizeof(buf));
		if (strncmp(extension, "IAX2/", 5) == 0) {
			if (strlen(members) > 0) {
				strncat(members, "&", 1);
			}
			strncat(members, extension, strlen(extension));
		} else {
			snprintf(function, sizeof(function), "PJSIP_DIAL_CONTACTS(%s)", extension);
			if ((0 == ast_func_read(chan, function, buf, sizeof(buf))) && (strlen(buf) > 0)) {
				if (strlen(members) > 0) {
					strncat(members, "&", 1);
				}

				strncat(members, buf, strlen(buf));
			}
			else if(strstr(extension,":") != NULL) {
				//For Multicast Paging.
				if (strlen(members) > 0) {
					strncat(members, "&", 1);
				}
				ast_log(LOG_DEBUG,"Extension is Multicast Paging");
				snprintf(buf, sizeof(buf), "MulticastRTP/basic/%s", extension);
				strncat(members, buf, strlen(buf));
			}

		}
	}
	if (original_member_list) {
		ast_free(original_member_list);
	}

	/*Start added by zyxia for Bug 151660  call dahdi when the page extension exist dahdi and sip at the same */
	char *members_dahdi = ast_strndup(args.devices, strlen(args.devices));
	while ((tech = strsep(&members_dahdi, "&"))){
		if (strncmp(tech, "DAHDI/",6) == 0) {

			if (strlen(members) > 0) {
				strncat(members, "&", 1);
			}
			strncat(members, tech, strlen(tech));
		}
	}
	/*End added by zyxia for Bug 151660  call dahdi when the page extension exist dahdi and sip at the same */

	if (members_dahdi) {
		ast_free(members_dahdi);
	}

	if (members && members[0] == '\0') {
		members = args.devices;
	}

	if (members && members[0] != '\0'){
		char *tmp_members = members;
		num_dials = 1;
		while (*tmp_members) {
			if (*tmp_members == '&') {
				num_dials++;
			}
			tmp_members++;
		}
	}

	if (members && !strlen(members)) {
		ast_log(LOG_WARNING, "No contact to call\n");
		return -1;
	}

	ast_log(LOG_DEBUG, "page_num [%s] members [%s] num_dials [%d]\n", page_num, members ? members : "NULL", num_dials);
#endif

	if (!(dial_list = ast_calloc(num_dials, sizeof(struct ast_dial *)))) {
		ast_log(LOG_ERROR, "Can't allocate %ld bytes for dial list\n", (long)(sizeof(struct ast_dial *) * num_dials));
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&options.flags, PAGE_CALLER_INTERCOM)) {
		ast_channel_callerintercom_set(chan, 1);
	}
#endif

	/* Go through parsing/calling each device */
#ifdef GRANDSTREAM_NETWORKS
	while (members && (tech = strsep(&members, "&")))
#else
	while ((tech = strsep(&args.devices, "&")))
#endif
	{
		int state = 0;
		struct ast_dial *dial = NULL;

		/* don't call the originating device */
		if (!strcasecmp(tech, originator))
			continue;

		/* If no resource is available, continue on */
		if (!(resource = strchr(tech, '/'))) {
			ast_log(LOG_WARNING, "Incomplete destination '%s' supplied.\n", tech);
			continue;
		}

		/* Ensure device is not in use if skip option is enabled */
		if (ast_test_flag(&options.flags, PAGE_SKIP)) {
			state = ast_device_state(tech);
			if (state == AST_DEVICE_UNKNOWN) {
				ast_log(LOG_WARNING, "Destination '%s' has device state '%s'. Paging anyway.\n", tech, ast_devstate2str(state));
			} else if (state != AST_DEVICE_NOT_INUSE) {
				ast_log(LOG_WARNING, "Destination '%s' has device state '%s'.\n", tech, ast_devstate2str(state));
				continue;
			}
		}

		*resource++ = '\0';

		/* Create a dialing structure */
		if (!(dial = ast_dial_create())) {
			ast_log(LOG_WARNING, "Failed to create dialing structure.\n");
			continue;
		}

		/* Append technology and resource */
		if (ast_dial_append(dial, tech, resource, NULL) == -1) {
			ast_log(LOG_ERROR, "Failed to add %s to outbound dial\n", tech);
			ast_dial_destroy(dial);
			continue;
		}

		/* Set ANSWER_EXEC as global option */
		ast_dial_option_global_enable(dial, AST_DIAL_OPTION_ANSWER_EXEC, confbridgeopts);

		if (predial_callee) {
			ast_dial_option_global_enable(dial, AST_DIAL_OPTION_PREDIAL, predial_callee);
		}

		if (timeout) {
			ast_dial_set_global_timeout(dial, timeout * 1000);
		}

		if (ast_test_flag(&options.flags, PAGE_IGNORE_FORWARDS)) {
			ast_dial_option_global_enable(dial, AST_DIAL_OPTION_DISABLE_CALL_FORWARDING, NULL);
		}

#ifdef GRANDSTREAM_NETWORKS
		if (record_enable) {
			ast_set_flag(&options.flags, PAGE_PROMPT_RECORD);
		}
#endif

		ast_dial_set_state_callback(dial, &page_state_callback);
		ast_dial_set_user_data(dial, &options);

		/* Run this dial in async mode */
		ast_dial_run(dial, chan, 1);

		/* Put in our dialing array */
		dial_list[pos++] = dial;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!pos) {
		ast_log(LOG_WARNING, "No dial is running\n");
		return -1;
	}

	int check_num = 30;
	char *ring_timeout = (char *)pbx_builtin_getvar_helper(chan, "RINGTIME");
	if ( !ast_strlen_zero(ring_timeout) && atoi(ring_timeout) > 0 )
		check_num = atoi(ring_timeout);
	enum ast_dial_result result;
	if (pos < 5) {
		usleep(500000); /* sleep 500 ms */
	}

	while (1) {
		struct ast_frame *f = ast_read(chan);
		int failed_count = 0;

		if (!f || ((f->frametype == AST_FRAME_CONTROL) && (f->subclass.integer == AST_CONTROL_HANGUP))) {
			ast_log(LOG_WARNING, "Page channel hangup\n");
			if (f) {
				ast_frfree(f);
			}
			goto clear_dial;
		}
		ast_frfree(f);

		for (i = 0; i < pos; i++) {
			result = ast_dial_state(dial_list[i]);
			if (result == AST_DIAL_RESULT_ANSWERED) {
				goto dial_answered;
			}

			if (result == AST_DIAL_RESULT_FAILED || result == AST_DIAL_RESULT_TIMEOUT || result == AST_DIAL_RESULT_PEER_BUSY || 
				result == AST_DIAL_RESULT_HANGUP || result == AST_DIAL_RESULT_UNANSWERED) {
				failed_count++;
			}
		}

		if (failed_count == pos) {
			ast_log(LOG_WARNING, "All page user dial failed\n");
			goto clear_dial;
		}

		check_num--;
		if (check_num <= 0) {
			ast_log(LOG_WARNING, "No User answered\n");
			goto clear_dial;
		}
		sleep(1);
	}

dial_answered:
#endif
	ast_free(predial_callee);

	if (!ast_test_flag(&options.flags, PAGE_QUIET)) {
		res = ast_streamfile(chan, "beep", ast_channel_language(chan));
		if (!res)
			res = ast_waitstream(chan, "");
	}

#ifdef GRANDSTREAM_NETWORKS
	/*begin: add by weidu , for 108273*/
	if (!ast_tvzero(calldurationlimit)) {
		struct timeval whentohangup = ast_tvadd(ast_tvnow(), calldurationlimit);
		ast_channel_lock(chan);
		ast_channel_whentohangup_set(chan, &whentohangup);
		ast_channel_unlock(chan);
		ast_log(LOG_DEBUG, "page_exec set huntup timer!\n");
	}
#endif
		/*end: add by weidu , for 108273*/

	if (!res) {
		setup_profile_bridge(chan, &options);
		setup_profile_caller(chan, &options);
		snprintf(confbridgeopts, sizeof(confbridgeopts), "%u", confid);
		pbx_exec(chan, app, confbridgeopts);
	}

#ifdef GRANDSTREAM_NETWORKS
	clear_dial:
#endif
	/* Go through each dial attempt cancelling, joining, and destroying */
	for (i = 0; i < pos; i++) {
		struct ast_dial *dial = dial_list[i];

		/* We have to wait for the async thread to exit as it's possible ConfBridge won't throw them out immediately */
		ast_dial_join(dial);

		/* Hangup all channels */
		ast_dial_hangup(dial);

		/* Destroy dialing structure */
		ast_dial_destroy(dial);
	}

	ast_free(dial_list);

	return -1;
}

static int unload_module(void)
{
#ifdef GRANDSTREAM_NETWORKS
	unload_list();
#endif
	return ast_unregister_application(app_page);
}

#ifdef GRANDSTREAM_NETWORKS
static int load_config(int reload)
{
	int res = 0;
	char *tmp = NULL;
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	struct headnum_list *tmp_list = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	if (!(cfg = ast_config_load("pagging_members.conf", config_flags))) {
 		ast_log(LOG_NOTICE, "No config file pagging_members.conf\n");
		return 0;
 	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file pagging_members.conf is in an invalid format.  Aborting.\n");
		return 1;
	}

	if (reload){
		unload_list();
	}

	while ((tmp = ast_category_browse(cfg, tmp)) ) {
		tmp_list = create_list(&g_pg_list, tmp);
		for (var = ast_variable_browse(cfg, tmp); var; var = var->next) {
			if (!strcasecmp(var->name, "members")) {
				ast_log(LOG_DEBUG, "PAGING GROUP: %s | NUMBER ALLOW NUMBER : %s", tmp, var->value);
				create_numbers_list(tmp_list, var->value);
			}
		}
	}
	print_global_list();

	ast_config_destroy(cfg);

	return res;
}

static int load_module(void)
{
	if (load_config(0)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return ast_register_application_xml(app_page, page_exec);
}

static int reload(void)
{
	int res = 0;

	ast_mutex_lock(&page_mutex);
	res = load_config(1);
	ast_mutex_unlock(&page_mutex);

	return res;
}

#else
static int load_module(void)
{
	return ast_register_application_xml(app_page, page_exec);
}
#endif

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Page Multiple Phones",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
#ifdef GRANDSTREAM_NETWORKS
	.reload = reload,
#endif
	.requires = "app_confbridge",
);
