/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Trivial application to playback a sound file
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/file.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/app.h"
/* This file provides config-file based 'say' functions, and implenents
 * some CLI commands.
 */
#include "asterisk/say.h"	/*!< provides config-file based 'say' functions */
#include "asterisk/cli.h"

/*** DOCUMENTATION
	<application name="Playback" language="en_US">
		<synopsis>
			Play a file.
		</synopsis>
		<syntax>
			<parameter name="filenames" required="true" argsep="&amp;">
				<argument name="filename" required="true" />
				<argument name="filename2" multiple="true" />
			</parameter>
			<parameter name="options">
				<para>Comma separated list of options</para>
				<optionlist>
					<option name="skip">
						<para>Do not play if not answered</para>
					</option>
					<option name="noanswer">
						<para>Playback without answering, otherwise the channel will
						be answered before the sound is played.</para>
						<note><para>Not all channel types support playing messages while still on hook.</para></note>
					</option>
				</optionlist>
			</parameter>
		</syntax>
		<description>
			<para>Plays back given filenames (do not put extension of wav/alaw etc).
			The playback command answer the channel if no options are specified.
			If the file is non-existant it will fail</para>
			<para>This application sets the following channel variable upon completion:</para>
			<variablelist>
				<variable name="PLAYBACKSTATUS">
					<para>The status of the playback attempt as a text string.</para>
					<value name="SUCCESS"/>
					<value name="FAILED"/>
				</variable>
			</variablelist>
			<para>See Also: Background (application) -- for playing sound files that are interruptible</para>
			<para>WaitExten (application) -- wait for digits from caller, optionally play music on hold</para>
		</description>
		<see-also>
			<ref type="application">Background</ref>
			<ref type="application">WaitExten</ref>
			<ref type="application">ControlPlayback</ref>
			<ref type="agi">stream file</ref>
			<ref type="agi">control stream file</ref>
			<ref type="manager">ControlPlayback</ref>
		</see-also>
	</application>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/musiconhold.h"
#include "asterisk/bridge_channel.h"
#include "asterisk/core_local.h"
#include "asterisk/indications.h"

AST_MODULE_LOG("app_playback");
static char *app_playbackvideo = "PlaybackVideo";
static char *app_playbackunsync = "PlaybackUnSync";
#endif

static char *app = "Playback";

static struct ast_config *say_cfg = NULL;

/*! \brief save the say' api calls.
 * The first entry is NULL if we have the standard source,
 * otherwise we are sourcing from here.
 * 'say load [new|old]' will enable the new or old method, or report status
 */
static const void *say_api_buf[40];
static const char * const say_old = "old";
static const char * const say_new = "new";

static void save_say_mode(const void *arg)
{
	int i = 0;
	say_api_buf[i++] = arg;

	say_api_buf[i++] = ast_say_number_full;
	say_api_buf[i++] = ast_say_enumeration_full;
	say_api_buf[i++] = ast_say_digit_str_full;
	say_api_buf[i++] = ast_say_character_str_full;
	say_api_buf[i++] = ast_say_phonetic_str_full;
	say_api_buf[i++] = ast_say_datetime;
	say_api_buf[i++] = ast_say_time;
	say_api_buf[i++] = ast_say_date;
	say_api_buf[i++] = ast_say_datetime_from_now;
	say_api_buf[i++] = ast_say_date_with_format;
}

static void restore_say_mode(void *arg)
{
	int i = 0;
	say_api_buf[i++] = arg;

	ast_say_number_full = say_api_buf[i++];
	ast_say_enumeration_full = say_api_buf[i++];
	ast_say_digit_str_full = say_api_buf[i++];
	ast_say_character_str_full = say_api_buf[i++];
	ast_say_phonetic_str_full = say_api_buf[i++];
	ast_say_datetime = say_api_buf[i++];
	ast_say_time = say_api_buf[i++];
	ast_say_date = say_api_buf[i++];
	ast_say_datetime_from_now = say_api_buf[i++];
	ast_say_date_with_format = say_api_buf[i++];
}

/*! \brief
 * Typical 'say' arguments in addition to the date or number or string
 * to say. We do not include 'options' because they may be different
 * in recursive calls, and so they are better left as an external
 * parameter.
 */
typedef struct {
	struct ast_channel *chan;
	const char *ints;
	const char *language;
	int audiofd;
	int ctrlfd;
} say_args_t;

static int s_streamwait3(const say_args_t *a, const char *fn)
{
	int res = ast_streamfile(a->chan, fn, a->language);
	if (res) {
		ast_log(LOG_WARNING, "Unable to play message %s\n", fn);
		return res;
	}
	res = (a->audiofd  > -1 && a->ctrlfd > -1) ?
	ast_waitstream_full(a->chan, a->ints, a->audiofd, a->ctrlfd) :
	ast_waitstream(a->chan, a->ints);
	ast_stopstream(a->chan);
	return res;
}

/*! \brief
 * the string is 'prefix:data' or prefix:fmt:data'
 * with ':' being invalid in strings.
 */
static int do_say(say_args_t *a, const char *s, const char *options, int depth)
{
	struct ast_variable *v;
	char *lang, *x, *rule = NULL;
	int ret = 0;
	struct varshead head = { .first = NULL, .last = NULL };
	struct ast_var_t *n;

	ast_debug(2, "string <%s> depth <%d>\n", s, depth);
	if (depth++ > 10) {
		ast_log(LOG_WARNING, "recursion too deep, exiting\n");
		return -1;
	} else if (!say_cfg) {
		ast_log(LOG_WARNING, "no say.conf, cannot spell '%s'\n", s);
		return -1;
	}

	/* scan languages same as in file.c */
	if (a->language == NULL)
		a->language = "en";     /* default */
	ast_debug(2, "try <%s> in <%s>\n", s, a->language);
	lang = ast_strdupa(a->language);
	for (;;) {
		for (v = ast_variable_browse(say_cfg, lang); v ; v = v->next) {
			if (ast_extension_match(v->name, s)) {
				rule = ast_strdupa(v->value);
				break;
			}
		}
		if (rule)
			break;
		if ( (x = strchr(lang, '_')) )
			*x = '\0';      /* try without suffix */
		else if (strcmp(lang, "en"))
			lang = "en";    /* last resort, try 'en' if not done yet */
		else
			break;
	}
	if (!rule)
		return 0;

	/* skip up to two prefixes to get the value */
	if ( (x = strchr(s, ':')) )
		s = x + 1;
	if ( (x = strchr(s, ':')) )
		s = x + 1;
	ast_debug(2, "value is <%s>\n", s);
	n = ast_var_assign("SAY", s);
	if (!n) {
		ast_log(LOG_ERROR, "Memory allocation error in do_say\n");
		return -1;
	}
	AST_LIST_INSERT_HEAD(&head, n, entries);

	/* scan the body, one piece at a time */
	while ( !ret && (x = strsep(&rule, ",")) ) { /* exit on key */
		char fn[128];
		const char *p, *fmt, *data; /* format and data pointers */

		/* prepare a decent file name */
		x = ast_skip_blanks(x);
		ast_trim_blanks(x);

		/* replace variables */
		pbx_substitute_variables_varshead(&head, x, fn, sizeof(fn));
		ast_debug(2, "doing [%s]\n", fn);

		/* locate prefix and data, if any */
		fmt = strchr(fn, ':');
		if (!fmt || fmt == fn)	{	/* regular filename */
			ret = s_streamwait3(a, fn);
			continue;
		}
		fmt++;
		data = strchr(fmt, ':');	/* colon before data */
		if (!data || data == fmt) {	/* simple prefix-fmt */
			ret = do_say(a, fn, options, depth);
			continue;
		}
		/* prefix:fmt:data */
		for (p = fmt; p < data && ret <= 0; p++) {
			char fn2[sizeof(fn)];
			if (*p == ' ' || *p == '\t')	/* skip blanks */
				continue;
			if (*p == '\'') {/* file name - we trim them */
				char *y;
				strcpy(fn2, ast_skip_blanks(p+1));	/* make a full copy */
				y = strchr(fn2, '\'');
				if (!y) {
					p = data;	/* invalid. prepare to end */
					break;
				}
				*y = '\0';
				ast_trim_blanks(fn2);
				p = strchr(p+1, '\'');
				ret = s_streamwait3(a, fn2);
			} else {
				int l = fmt-fn;
				strcpy(fn2, fn); /* copy everything */
				/* after prefix, append the format */
				fn2[l++] = *p;
				strcpy(fn2 + l, data);
				ret = do_say(a, fn2, options, depth);
			}

			if (ret) {
				break;
			}
		}
	}
	ast_var_delete(n);
	return ret;
}

static int say_full(struct ast_channel *chan, const char *string,
	const char *ints, const char *lang, const char *options,
	int audiofd, int ctrlfd)
{
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	return do_say(&a, string, options, 0);
}

static int say_number_full(struct ast_channel *chan, int num,
	const char *ints, const char *lang, const char *options,
	int audiofd, int ctrlfd)
{
	char buf[64];
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	snprintf(buf, sizeof(buf), "num:%d", num);
	return do_say(&a, buf, options, 0);
}

static int say_enumeration_full(struct ast_channel *chan, int num,
	const char *ints, const char *lang, const char *options,
	int audiofd, int ctrlfd)
{
	char buf[64];
	say_args_t a = { chan, ints, lang, audiofd, ctrlfd };
	snprintf(buf, sizeof(buf), "enum:%d", num);
	return do_say(&a, buf, options, 0);
}

static int say_date_generic(struct ast_channel *chan, time_t t,
	const char *ints, const char *lang, const char *format, const char *timezonename, const char *prefix)
{
	char buf[128];
	struct ast_tm tm;
	struct timeval when = { t, 0 };
	say_args_t a = { chan, ints, lang, -1, -1 };
	if (format == NULL)
		format = "";

	ast_localtime(&when, &tm, timezonename);
	snprintf(buf, sizeof(buf), "%s:%s:%04d%02d%02d%02d%02d.%02d-%d-%3d",
		prefix,
		format,
		tm.tm_year+1900,
		tm.tm_mon+1,
		tm.tm_mday,
		tm.tm_hour,
		tm.tm_min,
		tm.tm_sec,
		tm.tm_wday,
		tm.tm_yday);
	return do_say(&a, buf, NULL, 0);
}

static int say_date_with_format(struct ast_channel *chan, time_t t,
	const char *ints, const char *lang, const char *format, const char *timezonename)
{
	return say_date_generic(chan, t, ints, lang, format, timezonename, "datetime");
}

static int say_date(struct ast_channel *chan, time_t t, const char *ints, const char *lang)
{
	return say_date_generic(chan, t, ints, lang, "", NULL, "date");
}

static int say_time(struct ast_channel *chan, time_t t, const char *ints, const char *lang)
{
	return say_date_generic(chan, t, ints, lang, "", NULL, "time");
}

static int say_datetime(struct ast_channel *chan, time_t t, const char *ints, const char *lang)
{
	return say_date_generic(chan, t, ints, lang, "", NULL, "datetime");
}

/*! \brief
 * remap the 'say' functions to use those in this file
 */
static int say_init_mode(const char *mode) {
	if (!strcmp(mode, say_new)) {
		if (say_cfg == NULL) {
			ast_log(LOG_ERROR, "There is no say.conf file to use new mode\n");
			return -1;
		}
		save_say_mode(say_new);
		ast_say_number_full = say_number_full;

		ast_say_enumeration_full = say_enumeration_full;
#if 0
		/*! \todo XXX
		   These functions doesn't exist.
		   say.conf.sample indicates this is working...
		*/
		ast_say_digits_full = say_digits_full;
		ast_say_digit_str_full = say_digit_str_full;
		ast_say_character_str_full = say_character_str_full;
		ast_say_phonetic_str_full = say_phonetic_str_full;
		ast_say_datetime_from_now = say_datetime_from_now;
#endif
		ast_say_datetime = say_datetime;
		ast_say_time = say_time;
		ast_say_date = say_date;
		ast_say_date_with_format = say_date_with_format;
	} else if (!strcmp(mode, say_old) && say_api_buf[0] == say_new) {
		restore_say_mode(NULL);
	} else if (strcmp(mode, say_old)) {
		ast_log(LOG_WARNING, "unrecognized mode %s\n", mode);
		return -1;
	}

	return 0;
}

static char *__say_cli_init(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	const char *old_mode = say_api_buf[0] ? say_new : say_old;
	const char *mode;
	switch (cmd) {
	case CLI_INIT:
		e->command = "say load [new|old]";
		e->usage =
			"Usage: say load [new|old]\n"
			"       say load\n"
			"           Report status of current say mode\n"
			"       say load new\n"
			"           Set say method, configured in say.conf\n"
			"       say load old\n"
			"           Set old say method, coded in asterisk core\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (a->argc == 2) {
		ast_cli(a->fd, "say mode is [%s]\n", old_mode);
		return CLI_SUCCESS;
	} else if (a->argc != e->args)
		return CLI_SHOWUSAGE;
	mode = a->argv[2];
	if (!strcmp(mode, old_mode))
		ast_cli(a->fd, "say mode is %s already\n", mode);
	else
		if (say_init_mode(mode) == 0)
			ast_cli(a->fd, "setting say mode from %s to %s\n", old_mode, mode);

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_playback[] = {
	AST_CLI_DEFINE(__say_cli_init, "Set or show the say mode"),
};

#ifdef GRANDSTREAM_NETWORKS
static int check_need_prompt_record_other(struct ast_channel *chan, char *prompt)
{
	struct ast_channel *chan_prompt = NULL;
	const char *moh = NULL;
	const char *first_caller_channel_name = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");

	if (strcmp(prompt, "dialog-being-recorded")) {
		return 0;
	}

	pbx_builtin_setvar_helper(chan, "HAVE_PROMPTED", "yes");

	if (strstr(ast_channel_name(chan), "Local/")) {
		/* it's a local channel */
		const char *ringgroup_caller_chan_name = pbx_builtin_getvar_helper(chan, "RINGGROUP_CALLER_CHAN");
		const char *a_xfer_state = pbx_builtin_getvar_helper(chan, "A_XFER_STATE");
		if (ringgroup_caller_chan_name 
			&& strcmp(ringgroup_caller_chan_name, ast_channel_name(chan))
			&& (ast_strlen_zero(a_xfer_state) || strcmp(a_xfer_state, "begin"))) {
			ast_log(LOG_DEBUG, "ringgroup_caller_chan:%s", ringgroup_caller_chan_name);
			chan_prompt = ast_channel_get_by_name(ringgroup_caller_chan_name);
			moh = pbx_builtin_getvar_helper(chan_prompt, "RINGGROUP_MOH");

			ast_log(LOG_DEBUG, "Trigger this case");
		} else if (strstr(ast_channel_name(chan), ";2")) {
			const char *from_disa = pbx_builtin_getvar_helper(chan, "FROM_DISA");
			const char *vq_cb_caller = pbx_builtin_getvar_helper(chan, "VQ_CALLBACK_CALLER");

			if (strstr(ast_channel_name(chan), "call-forward-")) {
				/* added by zgmo for Bug150427, only for 63xx */
				if (!ast_strlen_zero(first_caller_channel_name)
					&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
					chan_prompt = ast_channel_get_by_name(first_caller_channel_name);
					if (chan_prompt) {
						const char *have_prompted = pbx_builtin_getvar_helper(chan_prompt, "HAVE_PROMPTED");
						if (have_prompted && !strcmp(have_prompted, "yes")) {
							ast_channel_unref(chan_prompt);
							chan_prompt = NULL;
						} else {
							moh = pbx_builtin_getvar_helper(chan_prompt, "FOLLOWME_MOH_CLASS");  /* Bug167117 */
							ast_log(LOG_DEBUG, "Trigger this case");
						}
					}
				}
			} else if (!ast_strlen_zero(from_disa) && !strcmp(from_disa, "yes")) {
				if (!ast_strlen_zero(first_caller_channel_name)
					&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
					struct ast_channel *first_caller = ast_channel_get_by_name(first_caller_channel_name);
					if (first_caller) {
						pbx_builtin_setvar_helper(first_caller, "HAVE_PROMPTED", "yes");
						ast_log(LOG_DEBUG, "In DISA, Set initial caller:%s, ChanVar 'HAVE_PROMPTED' = 'yes'",
							first_caller_channel_name);

						if (!chan_prompt && strstr(first_caller_channel_name, "Local/") 
							&& strstr(first_caller_channel_name, ";1")) {
							chan_prompt = ast_channel_real_peer(first_caller);
							ast_log(LOG_DEBUG, "Trigger this case");
						}

						ast_channel_unref(first_caller);
					}
				}
			} else if (!ast_strlen_zero(vq_cb_caller) && !strcmp(vq_cb_caller, "yes")) {
				const char *vq_cb_caller_name = pbx_builtin_getvar_helper(chan, "VQ_CALLBACK_CALLER_NAME");
				if (!ast_strlen_zero(vq_cb_caller_name)) {
					chan_prompt = ast_channel_get_by_name(vq_cb_caller_name);
					moh = pbx_builtin_getvar_helper(chan, "VQ_MOH_CLASS");
					ast_log(LOG_DEBUG, "Trigger this case");
				}
			}
		}
	}else {
		/* no-local channel */
		/* added by zgmo for bug153928 */
		const char *ringboth = pbx_builtin_getvar_helper(chan, "RINGBOTH");
		const char *seamless_transferee_channel = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFEREE_CHANNEL");
		const char *in_callpark = pbx_builtin_getvar_helper(chan, "IN_CALLPARK");
		const char *pickup_caller = pbx_builtin_getvar_helper(chan, "PICKUP_CALLER");
		const char *in_followme = pbx_builtin_getvar_helper(chan, "IN_FOLLOWME");
		const char *fm_enable = pbx_builtin_getvar_helper(chan,"FM_ENABLE");
		const char *from_callback = pbx_builtin_getvar_helper(chan,"FROM_CALLBACK");
		const char *record_check_on_answer_caller = pbx_builtin_getvar_helper(chan, "REC_CHECK_ON_ANSWER_CALLER");
		const char *from_disa = pbx_builtin_getvar_helper(chan, "FROM_DISA");

		if ((!ast_strlen_zero(ringboth) && !strcmp(ringboth, "yes"))) {
			if (!ast_strlen_zero(record_check_on_answer_caller)) {
				chan_prompt = ast_channel_get_by_name(record_check_on_answer_caller);

				if (strstr(record_check_on_answer_caller, "Local/")
					&& !ast_strlen_zero(first_caller_channel_name)) {
					struct ast_channel *initial_chan = ast_channel_get_by_name(first_caller_channel_name);
					if (initial_chan) {
						ast_playtones_stop(initial_chan);
						ast_channel_unref(initial_chan);

						ast_log(LOG_DEBUG, "Stop playing tone on %s", first_caller_channel_name);
					}
				}

				ast_log(LOG_DEBUG, "Trigger this case");
			} else if (!ast_strlen_zero(from_disa) && !strcmp(from_disa, "yes")
				&& !ast_strlen_zero(record_check_on_answer_caller)) {
				chan_prompt = ast_channel_get_by_name(record_check_on_answer_caller);
				ast_log(LOG_DEBUG, "Trigger this case");
			} else {
				if (!ast_strlen_zero(first_caller_channel_name)
					&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
					chan_prompt = ast_channel_get_by_name(first_caller_channel_name);
					ast_log(LOG_DEBUG, "Trigger this case");
				}
			}
		} else if (!ast_strlen_zero(pickup_caller) && !strcmp(pickup_caller, "yes")) {
			if (!ast_strlen_zero(first_caller_channel_name)
				&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
				chan_prompt = ast_channel_get_by_name(first_caller_channel_name);
				ast_log(LOG_DEBUG, "Trigger this case");
			}
		} else if (!ast_strlen_zero(seamless_transferee_channel)) {
			chan_prompt = ast_channel_get_by_name(seamless_transferee_channel);
			ast_log(LOG_DEBUG, "Trigger this case");
		} else if (!ast_strlen_zero(in_callpark) && !strcmp(in_callpark, "yes")) {
			const char *callpark_peer = pbx_builtin_getvar_helper(chan, "CALLPARK_PEER");
			if (!ast_strlen_zero(callpark_peer)) {
				chan_prompt = ast_channel_get_by_name(callpark_peer);
				ast_log(LOG_DEBUG, "Trigger this case");
			}
		} else if (!ast_strlen_zero(from_callback) && !strcmp(from_callback, "yes")) {
			if (!ast_strlen_zero(first_caller_channel_name)
				&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
				chan_prompt = ast_channel_get_by_name(first_caller_channel_name);
				ast_log(LOG_DEBUG, "Trigger this case");
			}
		} else if ((!ast_strlen_zero(in_followme) && !strcmp(in_followme, "yes"))
			|| (!ast_strlen_zero(fm_enable) && !strcmp(fm_enable, "yes"))) {
			const char *fm_caller_name = pbx_builtin_getvar_helper(chan, "FOLLOWME_CALLER");

			if (!ast_strlen_zero(record_check_on_answer_caller)
				&& strstr(record_check_on_answer_caller, "Local/")
				&& strstr(record_check_on_answer_caller, ";2")
				&& strstr(record_check_on_answer_caller, "from-internal-xfer")) {
				chan_prompt = ast_channel_get_by_name(record_check_on_answer_caller);
				ast_answer(chan_prompt);

				ast_log(LOG_DEBUG, "Trigger this case");
			} else if(!ast_strlen_zero(record_check_on_answer_caller)
				&& strstr(record_check_on_answer_caller, "Local/")
				&& strstr(record_check_on_answer_caller, ";2")
				&& strstr(record_check_on_answer_caller, "from-internal")) {
				if (first_caller_channel_name && strlen(first_caller_channel_name)
					&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
					chan_prompt = ast_channel_get_by_name(first_caller_channel_name);
					if (!chan_prompt && !ast_strlen_zero(fm_caller_name)) {
						chan_prompt = ast_channel_get_by_name(fm_caller_name);
						ast_log(LOG_DEBUG, "Trigger this case");
					} else {
						ast_log(LOG_DEBUG, "Trigger this case");
					}
				}
			} else {
				if (!ast_strlen_zero(first_caller_channel_name)
					&& strcmp(first_caller_channel_name, ast_channel_name(chan))) {
					chan_prompt = ast_channel_get_by_name(first_caller_channel_name);
					ast_log(LOG_DEBUG, "Trigger this case");
				}

				if (!chan_prompt && !ast_strlen_zero(record_check_on_answer_caller)
					&& strcmp(record_check_on_answer_caller, ast_channel_name(chan))) {
					chan_prompt = ast_channel_get_by_name(record_check_on_answer_caller);
					ast_log(LOG_DEBUG, "Trigger this case");
				}
			}

			if (chan_prompt) {
				const char *have_prompted = pbx_builtin_getvar_helper(chan_prompt, "HAVE_PROMPTED");
				if (have_prompted && !strcmp(have_prompted, "yes")) {
					ast_channel_unref(chan_prompt);
					chan_prompt = NULL;
				} else {
					moh = pbx_builtin_getvar_helper(chan_prompt, "FOLLOWME_MOH_CLASS");  /* Bug167117 */
				}
			}
		}
	}

	if (chan_prompt) {
		const char *have_prompt = pbx_builtin_getvar_helper(chan_prompt, "HAVE_PROMPTED");
		/*BEGIN: Add by weidu ,for 187404, play dialog-being-recorded*/
		const char *seamless_transfer = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER");
		/*END: Add by weidu*/
		ast_log(LOG_DEBUG, "chan(%s) want chan(%s) to prompt record",
			ast_channel_name(chan), ast_channel_name(chan_prompt));

		if (!ast_strlen_zero(have_prompt) && (!strcmp(have_prompt, "yes") || !strcmp(have_prompt, "doing"))) {
			ast_log(LOG_DEBUG, "chan(%s) HAVE_PROMPTED=%s, skip",
				ast_channel_name(chan_prompt), have_prompt ? have_prompt : "null");
			goto skip_prompt;
		}

		if (ast_channel_async_streamfile_protection_enter(chan_prompt)) {
			ast_log(LOG_WARNING, "chan(%s) failed to start streamfile protection, skip",
				ast_channel_name(chan_prompt));
			goto skip_prompt;
		}

		/*BEGIN: Add by weidu ,for 187404, play dialog-being-recorded and not read hangupcall */
		if (AST_STATE_UP == ast_channel_state(chan_prompt) && !strcasecmp((ast_channel_tech(chan)->type), "DAHDI")
			&& !ast_strlen_zero(seamless_transfer) && !strcasecmp(seamless_transfer,"yes")){
			usleep(80000);
		}
		/*END: Add by weidu*/

		pbx_builtin_setvar_helper(chan_prompt, "HAVE_PROMPTED", "doing");

		struct ast_app *progress_app = pbx_findapp("Progress");
		if (progress_app) {
			pbx_exec(chan_prompt, progress_app, NULL);
		}

		if (!ast_streamfile(chan_prompt, prompt, ast_channel_language(chan_prompt))) {
			ast_waitstream(chan_prompt, "");
			ast_stopstream(chan_prompt);

			if (!ast_strlen_zero(moh)) {
				ast_moh_start(chan_prompt, moh, NULL);
			}

			pbx_builtin_setvar_helper(chan_prompt, "HAVE_PROMPTED", "yes");
		} else {
			pbx_builtin_setvar_helper(chan_prompt, "HAVE_PROMPTED", "no");
		}

		ast_channel_async_streamfile_protection_exit(chan_prompt);
skip_prompt:
		ast_channel_unref(chan_prompt);
	}

	return 0;
}
#endif

static int playback_exec(struct ast_channel *chan, const char *data)
{
	int res = 0;
	int mres = 0;
	char *tmp;
	int option_skip=0;
	int option_say=0;
	int option_noanswer = 0;
	const char *cid;
	const char *esn;
#ifdef GRANDSTREAM_NETWORKS
	int play_tone = 1;
	int option_str = 0;
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(filenames);
		AST_APP_ARG(options);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "Playback requires an argument (filename)\n");
		return -1;
	}

	tmp = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, tmp);

	if (args.options) {
		if (strcasestr(args.options, "skip"))
			option_skip = 1;
		if (strcasestr(args.options, "say"))
			option_say = 1;
		if (strcasestr(args.options, "noanswer"))
			option_noanswer = 1;
#ifdef GRANDSTREAM_NETWORKS
		if (strcasestr(args.options, "str")) {
			option_str = 1;
			option_noanswer = 1;
		}
#endif
	}
	if (ast_channel_state(chan) != AST_STATE_UP) {
		if (option_skip) {
			/* At the user's option, skip if the line is not up */
			goto done;
		} else if (!option_noanswer) {
			/* Otherwise answer unless we're supposed to send this while on-hook */
			res = ast_answer(chan);
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!strcasecmp(args.filenames, "dialog-stop-recording")) {
		struct ast_bridge_channel * bridge_channel = NULL;

		ast_channel_lock(chan);
		bridge_channel = ast_channel_get_bridge_channel(chan);
		ast_channel_unlock(chan);
		if (bridge_channel && bridge_channel->in_bridge &&
			!pthread_equal(pthread_self(), bridge_channel->thread)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, args.filenames, NULL);
			ao2_cleanup(bridge_channel);
			goto done;
		}
		ao2_cleanup(bridge_channel);
	}
#endif

	if (!res) {
		char *back = args.filenames;
		char *front;

		ast_stopstream(chan);
		while (!res && (front = strsep(&back, "&"))) {
			if (option_say)
			{
				res = say_full(chan, front, "", ast_channel_language(chan), NULL, -1, -1);
			}
#ifdef GRANDSTREAM_NETWORKS
			else if (option_str)
			{
				ast_say_character_str(chan, front, AST_DIGIT_ANY, ast_channel_language(chan), AST_SAY_CASE_NONE);
				continue;
			}
#endif
			else
			{
				if (0 == strcmp(front,"tone-emergency-caller"))
				{
					cid = pbx_builtin_getvar_helper(chan,"EMERGENCY_CALLER");
					res = ast_say_digit_str(chan, cid, AST_DIGIT_ANY, ast_channel_language(chan));
				}
				else if (0 == strcmp(front,"tone-emergency-service-number"))
				{
					esn = pbx_builtin_getvar_helper(chan,"EMERGENCY_SERVICE_NUMBER");
					res = ast_say_digit_str(chan, esn, AST_DIGIT_ANY, ast_channel_language(chan));
				}
				else
				{
#ifdef GRANDSTREAM_NETWORKS 
					if (!strcmp(front, "dialog-being-recorded")) {
						/*add by weidu, for 155214*/
						const char *have_play_tone = pbx_builtin_getvar_helper(chan, "HAVE_PROMPTED");
						if (have_play_tone && !strcasecmp(have_play_tone, "yes")
							&& ast_channel_state(chan) != AST_STATE_UP) {
							play_tone = 0;
							continue;
						}
						/* end */

						if (strstr(ast_channel_name(chan), "Local/")) {
							if (strstr(ast_channel_name(chan), ";1")) {
								const char *from_callback = pbx_builtin_getvar_helper(chan, "FROM_CALLBACK");
								if (!ast_strlen_zero(from_callback) && !strcmp(from_callback, "yes")) {
									struct ast_channel *local_peer = ast_local_get_peer(chan);
									if (local_peer) {
										const char *have_prompt = pbx_builtin_getvar_helper(local_peer, "HAVE_PROMPTED");
										if (!ast_strlen_zero(have_prompt) && !strcmp(have_prompt, "yes")) {
											play_tone = 0;
											ast_channel_unref(local_peer);
											continue;
										}

										ast_channel_unref(local_peer);
									}
								}
							}
						}
					}

					play_tone =1;
#endif

					res = ast_streamfile(chan, front, ast_channel_language(chan));
#ifdef GRANDSTREAM_NETWORKS
					check_need_prompt_record_other(chan, front);
#endif
				}
			}

			if (!res && play_tone) {
				res = ast_waitstream(chan, "");
				ast_stopstream(chan);
			}
			if (res) {
				if (!ast_check_hangup(chan)) {
					ast_log(LOG_WARNING, "Playback failed on %s for %s\n", ast_channel_name(chan), (char *)data);
				}

#ifdef GRANDSTREAM_NETWORKS
				/*Start added by zyxia for Bug 236122,recv full frame ,need send to the bridge channel wait thread
				2022.9.13*/
				if (res == -2)
					ast_queue_frame(chan, &ast_null_frame);
				/*End*/
				else
					res = 0;
#else
				res = 0;
#endif
				mres = 1;
			}
		}
	}
done:
	pbx_builtin_setvar_helper(chan, "PLAYBACKSTATUS", mres ? "FAILED" : "SUCCESS");

	return res;
}

#ifdef GRANDSTREAM_NETWORKS
/*Start added by zyxia for Bug 257462,2023.4.19*/
struct playfile_data {
	struct ast_channel *chan;
	char filename[AST_FILENAME_MAX];
};

static void* playfile(void *data){
	struct playfile_data *pf_data = data;

	/*play prompt*/
	if (NULL != pf_data->chan) {
		const char* stop_play = pbx_builtin_getvar_helper(pf_data->chan, "STOP_UNSYNC_PLAY");
		if(ast_strlen_zero(stop_play)) {
			pbx_builtin_setvar_helper(pf_data->chan, "FAX_WAIT_PROMPT_PLAY","yes");
			ast_stream_and_wait(pf_data->chan, pf_data->filename, "");
		}
	}

	ast_log(LOG_DEBUG, "playfile for chan %s exit\n",ast_channel_name(pf_data->chan));
	ao2_cleanup(pf_data->chan);
	ast_free(pf_data);
	pthread_exit(NULL);
	return 0;
}

static int playbackUnSync_exec(struct ast_channel *chan, const char *data)
{
	int res = 0;
	char *tmp;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(filename);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "playbackUnSync_exec requires an argument (filename)\n");
		return -1;
	}

	tmp = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, tmp);


	if(ast_strlen_zero(args.filename)){
		ast_log(LOG_WARNING, "playbackUnSync_exec requires an argument (filename)\n");
		return -1;
	}
	char *filename = args.filename;

	if (ast_channel_state(chan) != AST_STATE_UP) {
		res = ast_answer(chan);
	}

	if (!res && filename) {

		ast_stopstream(chan);

		pthread_t threadid;
		struct playfile_data *pf;
		if (!(pf = ast_calloc(1, sizeof(*pf))))
			return -1;

		pf->chan =  ao2_bump((struct ast_channel *)chan);
		ast_copy_string(pf->filename, filename, sizeof(pf->filename));
		if (ast_pthread_create_detached(&threadid, NULL, playfile, pf)) {
			ao2_cleanup(pf->chan);
			ast_free(pf);
		}
	}
	return res;
}
/*End added by zyxia for Bug 257462,2023.4.19*/

static int playbackVideo_exec(struct ast_channel *chan, const char *data)
{
	int res = 0;
	char *tmp;
	int option_skip=0;
	int option_noanswer = 0;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(filenames);
		AST_APP_ARG(options);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "Playback requires an argument (filename)\n");
		return -1;
	}

	tmp = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, tmp);

	if (args.options) {
		if (strcasestr(args.options, "skip"))
			option_skip = 1;
		if (strcasestr(args.options, "noanswer"))
			option_noanswer = 1;
	}
	if (ast_channel_state(chan) != AST_STATE_UP) {
		if (option_skip) {
			/* At the user's option, skip if the line is not up */
			goto done;
		} else if (!option_noanswer) {
			/* Otherwise answer unless we're supposed to send this while on-hook */
			res = ast_answer(chan);
		}
	}

	usleep(500000);//wait the channnel join bridge, 
	if( ast_channel_name(chan) != NULL && !strncmp(ast_channel_name(chan),"Local/",6)
		&& strstr(ast_channel_name(chan),";1") != NULL){
		char local_channel[64] = {0};
		snprintf(local_channel,sizeof(local_channel)-1,"%s",ast_channel_name(chan));
		if(strlen(local_channel) > 0){
			RAII_VAR(struct ast_channel *, local_chan, NULL, ao2_cleanup);
			local_channel[strlen(local_channel)-1] = '2';
			local_chan = ast_channel_get_by_name(local_channel);
			if( local_chan ){
				const char* peer_chan_name = pbx_builtin_getvar_helper(local_chan,"BRIDGEPEER");
				if (!ast_strlen_zero(peer_chan_name)){

					RAII_VAR(struct ast_channel *, peer_chan, NULL, ao2_cleanup);
					peer_chan = ast_channel_get_by_name(peer_chan_name);
					if (peer_chan) {
						ast_log(LOG_DEBUG, "Playback Video File  to channel %s\n",peer_chan_name);
						pbx_builtin_setvar_helper(peer_chan, "FROM_PLAYBACK_VIDEO", "yes");
						res = playback_exec(peer_chan, data);
						goto done;
					}
				}
			}
		}
	}

	res = playback_exec(chan, data);
done:
	pbx_builtin_setvar_helper(chan, "PLAYBACKSTATUS", res ? "FAILED" : "SUCCESS");

	return res;
}
#endif

static int reload(void)
{
	struct ast_variable *v;
	struct ast_flags config_flags = { CONFIG_FLAG_FILEUNCHANGED };
	struct ast_config *newcfg;

	if ((newcfg = ast_config_load("say.conf", config_flags)) == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (newcfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file say.conf is in an invalid format.  Aborting.\n");
		return 0;
	}

	if (say_cfg) {
		ast_config_destroy(say_cfg);
		ast_log(LOG_NOTICE, "Reloading say.conf\n");
	}
	say_cfg = newcfg;

	if (say_cfg) {
		for (v = ast_variable_browse(say_cfg, "general"); v ; v = v->next) {
    			if (ast_extension_match(v->name, "mode")) {
				say_init_mode(v->value);
				break;
			}
		}
	}

	/*! \todo
	 * XXX here we should sort rules according to the same order
	 * we have in pbx.c so we have the same matching behaviour.
	 */
	return 0;
}

static int unload_module(void)
{
	int res;

	res = ast_unregister_application(app);

	ast_cli_unregister_multiple(cli_playback, ARRAY_LEN(cli_playback));
#ifdef GRANDSTREAM_NETWORKS
	ast_unregister_application(app_playbackvideo);
	ast_unregister_application(app_playbackunsync);
#endif
	if (say_cfg)
		ast_config_destroy(say_cfg);

	return res;
}

static int load_module(void)
{
	struct ast_variable *v;
	struct ast_flags config_flags = { 0 };

	say_cfg = ast_config_load("say.conf", config_flags);
	if (say_cfg && say_cfg != CONFIG_STATUS_FILEINVALID) {
		for (v = ast_variable_browse(say_cfg, "general"); v ; v = v->next) {
    			if (ast_extension_match(v->name, "mode")) {
				say_init_mode(v->value);
				break;
			}
		}
	}

	ast_cli_register_multiple(cli_playback, ARRAY_LEN(cli_playback));
#ifdef GRANDSTREAM_NETWORKS
	ast_register_application_xml(app_playbackvideo, playbackVideo_exec);
	ast_register_application_xml(app_playbackunsync, playbackUnSync_exec);
#endif
	return ast_register_application_xml(app, playback_exec);
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Sound File Playback Application",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
);
