#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2016-2017, chaozhen deng <chzhdeng@grandstream.cn>
 *
 * app_fastagi.c adapted from the pms
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>sqlite3</depend>
        <depend>ucmsecret</depend>
	<support_level>core</support_level>
 ***/

/***
	PMS(arg1,arg2,arg3)
		arg1: wake
		arg2: 0--cancel wakeup, 1--set wakeup, 2--executed wakeup
		arg3: 1--answer, 2--no response, 3--busy, 4--error

		arg1: st
		arg2: 1--checkin, 2--cleaning, 3--maintenance
		arg3: 0--Meaningless

***/
#include "asterisk.h"

#include <stdio.h>
#include <string.h>
#include <curl/curl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sqlite3.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <stdbool.h>
#include <sys/un.h>
#include <dirent.h>

#include "asterisk/file.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/app.h"
#include "asterisk/channel.h"
#include "asterisk/dsp.h"
#include "asterisk/ucm_dbi.h"
#include "asterisk/io.h"
#include "ucm_crypt.h"
#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#endif
#include "asterisk/haserver.h"

AST_MODULE_LOG("pms");

typedef sqlite3* db_conn;

static char *app = "PMS";

#define PMS_BUF_MIN_LEN    16
#define PMS_EXTEN_LEN       64
#define PMS_BUF_MAX_LEN    512

#define PATH_GSM "/data/var/lib/asterisk/sounds/"
typedef enum PMS_STATUS 
{
    PMS_STATUS_0 = 0,
    PMS_STATUS_1 = 1,
    PMS_STATUS_2 = 2,
    PMS_STATUS_3 = 3,
    PMS_STATUS_4 = 4,
    PMS_STATUS_5 = 5
}PMS_STATUS;



typedef struct StrMapping
{
    enum PMS_STATUS status;
    char *file_name;
}StrMapping;

static StrMapping pms_status_mapping[] = {
    {PMS_STATUS_0, "untreated"},
    {PMS_STATUS_1, "answer"},
    {PMS_STATUS_2, "noanswer"},
    {PMS_STATUS_3, "busy"},
    {PMS_STATUS_4, "error"},
    {PMS_STATUS_5, "noanswer"}
};

typedef struct ListWeekNode {
    int week;
    struct ListWeekNode* next;
} ListWeekNode;

struct pms_task{
	char type[PMS_BUF_MIN_LEN];
	char status[PMS_EXTEN_LEN];
	char opt[PMS_BUF_MIN_LEN];
	char extension[PMS_EXTEN_LEN];
#ifdef HAVE_UNIXODBC
	struct odbc_obj *obj;
#endif
};

#define PMS_SIZE sizeof(struct pms_task)

#define WAKE_CANCEL_STATUS     "0"
#define WAKE_SET_STATUS          "1"
#define WAKE_EXECUTE_STATUS    "2"
#define PMS_SET_SUCCESS    "set-success"
#define PMS_CAN_NOT_SET    "can-not-set"


#define ast_play_voice_prompt(chan, content) ast_stream_and_wait(chan, content, AST_DIGIT_ANY)

#define clean_pms_task(pms) \
	if (pms){ \
		ast_free(pms); \
	} \

#define AST_GET_PMS_ARGS(a, b, c) \
	if (!ast_strlen_zero(c)) { \
		strncpy(b, c, sizeof(b)-1); \
	}else{ \
		clean_pms_task(a); \
		return -1; \
	} \

typedef struct __dirstream DIR;

static void insertWeekNode(ListWeekNode** head, int val) {
    ListWeekNode* newNode = (ListWeekNode*)malloc(sizeof(ListWeekNode));
    if (!newNode) return;
    newNode->week = val;
    newNode->next = NULL;

    if (*head == NULL) {
        *head = newNode;
    } else {
        ListWeekNode* temp = *head;
        while (temp->next != NULL) {
            temp = temp->next;
        }
        temp->next = newNode;
    }
}

static void freeWeekList(ListWeekNode* head) {
    ListWeekNode* temp;
    while (head != NULL) {
        temp = head;
        head = head->next;
        free(temp);
    }
}

static int vm_lock_path(const char *path)
{
	switch (ast_lock_path(path)) {
	case AST_LOCK_TIMEOUT:
		return -1;
	default:
		return 0;
	}
}

int g_pms_ha_enable = 0;
static int pms_get_haon(void)
{
	FILE *ha_fp = NULL;
	char on_buf[64] = { 0 };

	/* Get the ha_enable */
	ha_fp = popen("nvram get haon", "r");
	if (ha_fp != NULL) {
		while (ha_fp != NULL && (!feof(ha_fp)) && (fgets(on_buf, sizeof(on_buf) - 1, ha_fp) != NULL)) {
			ast_log(LOG_DEBUG, "haon Read: [%s].", on_buf);
		}
		pclose(ha_fp);
	}
	g_pms_ha_enable = atoi(on_buf);

	ast_log(LOG_DEBUG, "ha_enable is %d.", g_pms_ha_enable);
	return g_pms_ha_enable;
}

#ifdef HAVE_UNIXODBC
struct generic_prepare_struct {
	char *sql;
	int argc;
	char **argv;
};
static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}

static int update_pms_db(const char *sql, struct odbc_obj *obj )
{
	SQLHSTMT stmt;
	struct generic_prepare_struct gps = { .sql = (char *)sql};

	stmt = ast_core_odbc_prepare_and_execute(obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", sql);
		return -1;
	} else {
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}

	ast_ha_send_sql((char *)sql, AST_HA_PMS);

	return 0;
}
#endif

static int count_messages(char *dir)
{

	int vmcount = 0;
	DIR *vmdir = NULL;
	struct dirent *vment = NULL;

	if ((vmdir = opendir(dir))) {
		while ((vment = readdir(vmdir))) {
			if (strlen(vment->d_name) > 7 && !strncmp(vment->d_name + 7, ".txt", 4)) {
#ifdef GRANDSTREAM_NETWORKS
				char tmpfile[256] = {0};
				char tmpfile2[256] = {0};
				/* 9 = strlen("msg0000")+ strlen("/") + strlen("\0")*/
				int len = strlen(dir) + 9;
				if (len > sizeof(tmpfile)) {
					len = sizeof(tmpfile);
					
				}
				snprintf(tmpfile, len, "%s/%s", dir, vment->d_name);
				snprintf(tmpfile + strlen(tmpfile), sizeof(tmpfile) - strlen(tmpfile), ".WAV");
				snprintf(tmpfile2, len, "%s/%s", dir, vment->d_name);
				snprintf(tmpfile2 + strlen(tmpfile2), sizeof(tmpfile2) - strlen(tmpfile2), ".wav");

				if (access(tmpfile, 0)) {
					if (access(tmpfile2, 0)) {
						ast_debug(4, "There is no %s, we need to remove the txt file.\n", tmpfile);
						snprintf(tmpfile, sizeof(tmpfile), "%s/%s", dir, vment->d_name);
						remove(tmpfile);
						continue;
					}
				}
#endif
				vmcount++;
			}
		}
		closedir(vmdir);
	}

	return vmcount;
}

static int conf_runsql_get_int_field(char *sql, struct odbc_obj *obj)
{
	int result = -1;
	char tmp[512] = {0};
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	if (ast_strlen_zero(sql)) {
		return result;
	}
	gps.sql = sql;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sql);
		return result;
	} else {

		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", sql);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			return result;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			result = atoi(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", sql);

	return result;
}

/*Mod by weidu, for Bug 309322*/
static char *conf_runsql_get_varchar_field(char *sql, const char *key, struct odbc_obj *conf_obj)
{
	char *result = NULL;
	char tmp[512] = {0};
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;
	if (ast_strlen_zero(sql) || ast_strlen_zero(key)) {
		return NULL;
	}

	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
		return NULL;
	}

	gps.sql = sql;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sql);
		goto end_clean;
	} else {

		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", sql);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			result = ast_strdup(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", sql);

end_clean:
	return result;
}

static int msg_id_incrementor;

static void generate_msg_id(char *dst)
{
	/* msg id is time of msg_id generation plus an incrementing value
	 * called each time a new msg_id is generated. This should achieve uniqueness,
	 * but only in single system solutions.
	 */
	unsigned int unique_counter = ast_atomic_fetchadd_int(&msg_id_incrementor, +1);
	snprintf(dst, 256, "%ld-%08x", (long) time(NULL), unique_counter);
}

static int get_date(char *s, int len)
{
	struct ast_tm tm;
	struct timeval now = ast_tvnow();

	ast_localtime(&now, &tm, NULL);
	return ast_strftime(s, len, "%a %b %e %r %Z %Y", &tm);
}

static int updateNextWakeupTime(struct pms_task *pms)
{
    int ret = 0;
    int tmp_week = 0;
    char cmd[1024] = {0};
    char *next_start_date = NULL;
    char *now_week = NULL;
    char *psave = NULL;
    char *temp = NULL;
    char *address = NULL;
    ListWeekNode* head = NULL;
    ListWeekNode* tmp = NULL;

    snprintf(cmd, sizeof(cmd)-1, "SELECT address FROM pms_room_members WHERE extension = '%s'",pms->extension);
    address = conf_runsql_get_varchar_field(cmd,"address", pms->obj);
    if(!address)
    {
        AST_SAFE_FREE(address);
        return -1;
    }
    snprintf(cmd, sizeof(cmd), "SELECT w_type FROM pms_wakeup WHERE address ='%s'", address);
    char *w_type = conf_runsql_get_varchar_field(cmd, "w_type", pms->obj);
    snprintf(cmd, sizeof(cmd), "SELECT w_date FROM pms_wakeup WHERE address ='%s'", address);
    char *w_date = conf_runsql_get_varchar_field(cmd, "w_date", pms->obj);
    snprintf(cmd, sizeof(cmd), "SELECT w_week FROM pms_wakeup WHERE address ='%s'", address);
    char *w_week = conf_runsql_get_varchar_field(cmd, "w_week", pms->obj);

    if (w_type && w_date)
    {
        if (strcmp(w_type, "2") == 0)
        {
            snprintf(cmd, sizeof(cmd), "SELECT DATE_ADD(STR_TO_DATE('%s','%%Y%%m%%d'), INTERVAL 1 day) AS next_start_date",w_date);
            next_start_date = conf_runsql_get_varchar_field(cmd, "next_start_date", pms->obj);
        }
        else if (strcmp(w_type, "3") == 0)
        {
            if(!w_week)
            {
		AST_SAFE_FREE(address);
                AST_SAFE_FREE(w_type);
                AST_SAFE_FREE(w_date);
                AST_SAFE_FREE(w_week);
                return -1;
            }
            snprintf(cmd, sizeof(cmd), "SELECT WEEKDAY(STR_TO_DATE('%s','%%Y%%m%%d')) AS now_week;",w_date);
            now_week = conf_runsql_get_varchar_field(cmd, "now_week", pms->obj);
            temp = strtok_r(w_week,",",&psave);;
            while(temp != NULL)
            {
                insertWeekNode(&head,atoi(temp));
                temp = strtok_r(NULL,",",&psave);
            }
            if(NULL != now_week && atoi(now_week) != 6)
            {
                tmp_week = atoi(now_week) + 1;
            }
            else
            {
                tmp_week = 0;
            }
            tmp = head;
            while(tmp)
            {
                if(tmp->week == tmp_week)
                {
                    if(tmp->next == NULL)
                    {
                        snprintf(cmd, sizeof(cmd), "SELECT DATE_ADD(STR_TO_DATE('%s','%%Y%%m%%d'), INTERVAL %d day) AS next_start_date",w_date,(7 - tmp_week + head->week));
                        break;
                    }
                    if(tmp->next != NULL)
                    {
                        snprintf(cmd, sizeof(cmd), "SELECT DATE_ADD(STR_TO_DATE('%s','%%Y%%m%%d'), INTERVAL %d day) AS next_start_date",w_date,(tmp->next->week - tmp_week));
                    }
                    break;
                }
                tmp = tmp->next;
            }

            if(NULL == tmp)
            {
                tmp = head;
                while(tmp)
                {
                    if(tmp->week > tmp_week)
                    {
                        snprintf(cmd, sizeof(cmd), "SELECT DATE_ADD(STR_TO_DATE('%s','%%Y%%m%%d'), INTERVAL %d day) AS next_start_date",w_date,(tmp->week - tmp_week));
                        break;
                    }
                    tmp = tmp->next;
                    if(NULL == tmp)
                    {
                        snprintf(cmd, sizeof(cmd), "SELECT DATE_ADD(STR_TO_DATE('%s','%%Y%%m%%d'), INTERVAL %d day) AS next_start_date",w_date,(7 + head->week - tmp_week));
                        break;
                    }
                }
            }
            next_start_date = conf_runsql_get_varchar_field(cmd, "next_start_date", pms->obj);
        }
        else if (strcmp(w_type, "4") == 0)
        {
            snprintf(cmd, sizeof(cmd), "SELECT DATE_ADD(STR_TO_DATE('%s','%%Y%%m%%d'), INTERVAL 1 month) AS next_start_date",w_date);
            next_start_date = conf_runsql_get_varchar_field(cmd, "next_start_date", pms->obj);
        }
        if (next_start_date)
        {
            char *tmp_next_start_date = next_start_date;
            char output[9] = { 0 };
            memcpy(output, tmp_next_start_date, 4);
            tmp_next_start_date += 5;
            memcpy(output + 4, tmp_next_start_date, 2);
            tmp_next_start_date += 3;
            memcpy(output + 6, tmp_next_start_date, 2);
            output[8] = '\0';
            snprintf( cmd, sizeof(cmd), "UPDATE pms_wakeup SET w_date = '%s' WHERE address = '%s'",output,address);
            ret = update_pms_db(cmd, pms->obj);
        }
        else
        {
            ret = -1;
        }
    }
    else
    {
        ret = -1;
    }
    AST_SAFE_FREE(address);
    AST_SAFE_FREE(w_type);
    AST_SAFE_FREE(w_date);
    AST_SAFE_FREE(w_week);
    AST_SAFE_FREE(now_week);
    AST_SAFE_FREE(next_start_date);
    freeWeekList(head);

    return ret;
}

static int sycn_vm_group(struct ast_channel *chan ,struct pms_task *pms ,const char* vm_extension ,const char* vm_name)
{
	int ret = 0;
	int cnt = 0;
	char cmd[512] = { 0 };
	char msg_name[256] = { 0 };
	char vm_group_path[256] = { 0 };
	char* member_extension_string = NULL;
	char* token = NULL;
	char* lasts = NULL;
	char* newmsgs = NULL;
	char* oldmsgs = NULL;
	char* urgentmsgs = NULL;
	char* email_escape = NULL;
	char sqlprefix[64] = {0};
	char msg_id[256] = {0};
	char date[256] = {0};

	ast_safe_nvram_get("sqlprefix", sqlprefix, sizeof(sqlprefix));
	snprintf(cmd, sizeof(cmd), "SELECT GROUP_CONCAT(member_extension SEPARATOR ',') AS member_extension_string  FROM vmgroup_members  WHERE extension = '%s'",vm_extension);
	member_extension_string = conf_runsql_get_varchar_field(cmd,"member_extension_string", pms->obj);
	snprintf(vm_group_path,sizeof(vm_group_path)-1,"/data/var/spool/asterisk/voicemail/default/%s/INBOX",vm_extension);
	if(member_extension_string != NULL && 0 != strlen(member_extension_string))
	{
		token = strtok_r(member_extension_string, ",", &lasts);  
    	while (token != NULL) {  
        	snprintf(cmd,sizeof(cmd)-1,"/data/var/spool/asterisk/voicemail/default/%s/INBOX",token);
        	cnt = count_messages(cmd);
			snprintf(msg_name, sizeof(msg_name)-1, "msg%04d", cnt);
			if (access(cmd, F_OK) == -1) {
				snprintf(cmd,sizeof(cmd)-1,"mkdir -p /data/var/spool/asterisk/voicemail/default/%s/INBOX",token);
				ast_safe_system(cmd);
			}
			snprintf(cmd,sizeof(cmd)-1,"cp %s/%s.WAV /data/var/spool/asterisk/voicemail/default/%s/INBOX/%s.WAV",vm_group_path,vm_name,token,msg_name);
			ast_safe_system(cmd);
			snprintf(cmd,sizeof(cmd)-1,"cp %s/%s.txt /data/var/spool/asterisk/voicemail/default/%s/INBOX/%s.txt",vm_group_path,vm_name,token,msg_name);
			ast_safe_system(cmd);
			generate_msg_id(msg_id);
			get_date(date, sizeof(date));
			snprintf(cmd, sizeof(cmd), "REPLACE INTO %svoicemail.voicemail_msg_info VALUES ('%s', 'INBOX', '\"Wakeup\" <Wakeup>', '%ld', '%s', 6, 'N', '%s', %d, 1000, 'file', 'WAV', 1,NULL,NULL,NULL)", sqlprefix, token , (long)time(NULL), date, msg_id, cnt);
			ret = update_pms_db(cmd, pms->obj);
			snprintf(cmd, sizeof(cmd)-1, "SELECT COUNT(*) AS cnt FROM %svoicemail.voicemail_msg_info WHERE type = 'INBOX' AND user = '%s';",sqlprefix,token);
			newmsgs = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
			snprintf(cmd, sizeof(cmd)-1, "SELECT COUNT(*) AS cnt FROM %svoicemail.voicemail_msg_info WHERE type = 'Old' AND user = '%s';",sqlprefix,token);
			oldmsgs = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
			snprintf(cmd, sizeof(cmd)-1, "SELECT COUNT(*) AS cnt FROM %svoicemail.voicemail_msg_info WHERE type = 'Urgent' AND user = '%s';",sqlprefix,token);
			urgentmsgs = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
            if(newmsgs && oldmsgs && urgentmsgs)
            {
			    ast_manager_event(chan, EVENT_FLAG_CALL, "MessageWaiting", "Mailbox: %s\r\nNew: %d\r\nOld: %d\r\nUrg: %d\r\n", token, atoi(newmsgs), atoi(oldmsgs), atoi(urgentmsgs));
			    __ast_publish_mwi_state_channel(token, "default", atoi(urgentmsgs), atoi(newmsgs) + atoi(urgentmsgs), atoi(oldmsgs), NULL, NULL);
            }
			snprintf(cmd, sizeof(cmd), "SELECT email FROM users WHERE user_name = '%s'",token);
			email_escape = conf_runsql_get_varchar_field(cmd,"email", pms->obj);
        	if(NULL != email_escape)
            {
				snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/sendMail.py voicemail '%s' %d '%s' \"\" '%s' \"\" '/data/var/spool/asterisk/voicemail/default/%s/INBOX/%s' 6 '' \"no\" \"\" \"\" '%s'"
			        , email_escape, cnt, token, pms->extension, token, msg_name, pms->extension);
				ast_safe_system(cmd);
        	}

        	token = strtok_r(NULL, ",", &lasts);  
			AST_SAFE_FREE(newmsgs);
			AST_SAFE_FREE(oldmsgs);
			AST_SAFE_FREE(urgentmsgs);
			AST_SAFE_FREE(email_escape);
    	} 
	}
	AST_SAFE_FREE(newmsgs);
	AST_SAFE_FREE(oldmsgs);
	AST_SAFE_FREE(urgentmsgs);
	AST_SAFE_FREE(member_extension_string);
	return ret;
}

static int update_wakeup_execute_status(struct ast_channel *chan, struct pms_task *pms)
{
	int ret = 0;
	char* address = NULL;
	char* w_type = NULL;
	char* email_notify_enable = NULL;
	char* vm_notify_enable = NULL;
	char* vm_extension = NULL;
	char* language = NULL;
	char* email_send_way = NULL;
	int cnt = 0;
	char* w_times = NULL;
	char* call_times = NULL;
	char* newmsgs = NULL;
	char* oldmsgs = NULL;
	char* urgentmsgs = NULL;
	char* email_escape = NULL;
	char* count = NULL;
	char* member_extension_string = NULL;
	char* vm_email = NULL;
	char msg_name[256] = {0};
	char cmd[PMS_BUF_MAX_LEN] = {0};
	char sox[1024] = {0};
	char msg_id[256] = {0};
	char date[256] = {0};
	char file_path[256] = {0};
	char path_num_gsm[256] = {0};
	char path_wakeup_gsm[256] = {0};
	FILE *txt;

    time_t timer;
    char w_modified[32] = {0};
    char now_data[32] = {0};
    char now_time[32] = {0};
    char sqlprefix[64] = {0};

    time(&timer);
    strftime(w_modified, sizeof(w_modified)-1, "%Y%m%d %H%M%S", localtime(&timer));
	strftime(now_data, sizeof(now_data)-1, "%Y%m%d", localtime(&timer));
	strftime(now_time, sizeof(now_time)-1, "%H%M%S", localtime(&timer));
	ast_safe_nvram_get("sqlprefix", sqlprefix, sizeof(sqlprefix));

	snprintf(cmd, sizeof(cmd)-1, "UPDATE %smaster.pms_wakeup_log SET call_times = call_times + 1 WHERE w_id = '%s'", sqlprefix, pms->status);
	ret = update_pms_db(cmd, pms->obj);
	
	snprintf(cmd, sizeof(cmd)-1, "SELECT address FROM pms_room_members WHERE extension = '%s'",pms->extension);
	address = conf_runsql_get_varchar_field(cmd,"address", pms->obj);  /*Mod by weidu, for Bug 309322*/
	if( atoi(pms->opt) == 1 )
	{
		snprintf(cmd, sizeof(cmd)-1, "UPDATE %smaster.pms_wakeup_log SET w_action = '2',w_status = '%s',run_date = '%s',run_time = '%s' WHERE w_id = '%s'",
						sqlprefix, pms->opt,now_data,now_time,pms->status);
		ret = update_pms_db(cmd, pms->obj);
		snprintf(cmd, sizeof(cmd)-1, "SELECT w_type FROM pms_wakeup WHERE address = '%s'",address);
		w_type = conf_runsql_get_varchar_field(cmd,"w_type", pms->obj);
		if(NULL != w_type && 0 != strlen(w_type))
		{
			if(1 == atoi(w_type))
			{
				snprintf(cmd, sizeof(cmd)-1, "DELETE FROM pms_wakeup WHERE address = '%s'",address);
				ret = update_pms_db(cmd, pms->obj);
			}
			else
			{
				snprintf(cmd, sizeof(cmd)-1, "UPDATE pms_wakeup SET w_action = '1' WHERE address = '%s'",address);
				ret = update_pms_db(cmd, pms->obj);
				ret = updateNextWakeupTime(pms);
			}
		}
	}
	if( 0 <= atoi(pms->opt) && atoi(pms->opt) < 5 && atoi(pms->opt) != 1 )
	{
		snprintf(cmd, sizeof(cmd)-1, "UPDATE %smaster.pms_wakeup_log SET w_action = '2',w_status = '%s' WHERE w_id = '%s'",
						sqlprefix, pms->opt,pms->status);
		ret = update_pms_db(cmd, pms->obj);
	}
	else if( 5 == atoi(pms->opt) )
	{
		snprintf(cmd, sizeof(cmd)-1, "SELECT w_type FROM pms_wakeup WHERE address = '%s'",address);
		w_type = conf_runsql_get_varchar_field(cmd,"w_type", pms->obj);
		snprintf(cmd, sizeof(cmd)-1, "UPDATE %smaster.pms_wakeup_log SET w_action = '2',w_status = '1' WHERE w_id = '%s'",
						sqlprefix,pms->status);
		ret = update_pms_db(cmd, pms->obj);
		if(NULL != w_type && 0 != strlen(w_type))
		{
			if(1 == atoi(w_type))
			{
				snprintf(cmd, sizeof(cmd)-1, "DELETE FROM pms_wakeup WHERE address = '%s'",address);
				ret = update_pms_db(cmd, pms->obj);
			}
			else
			{
				snprintf(cmd, sizeof(cmd)-1, "UPDATE pms_wakeup SET w_action = '1' WHERE address = '%s'",address);
				ret = update_pms_db(cmd, pms->obj);
				ret = updateNextWakeupTime(pms);
			}
		}
	}
	else if( 6 == atoi(pms->opt) )
	{
		snprintf(cmd, sizeof(cmd)-1, "UPDATE %smaster.pms_wakeup_log SET w_action = '2',w_status = '2',run_date = '%s',run_time = '%s' WHERE w_id = '%s'  AND run_time IS NULL",
						sqlprefix, now_data,now_time,pms->status);
		ret = update_pms_db(cmd, pms->obj);
		snprintf(cmd, sizeof(cmd)-1, "SELECT w_type FROM pms_wakeup WHERE address = '%s'",address);
		w_type = conf_runsql_get_varchar_field(cmd,"w_type", pms->obj); /*Mod by weidu, for Bug 309322*/
		if(NULL != w_type && 0 != strlen(w_type))
		{
			if(1 == atoi(w_type))
			{
				snprintf(cmd, sizeof(cmd)-1, "DELETE FROM pms_wakeup WHERE address = '%s'",address);
				ret = update_pms_db(cmd, pms->obj);
			}else{
				snprintf(cmd, sizeof(cmd)-1, "UPDATE pms_wakeup SET w_action = '1' WHERE address = '%s'",address);
				ret = update_pms_db(cmd, pms->obj);
				ret = updateNextWakeupTime(pms);
			}
		}	
	}

	snprintf(cmd, sizeof(cmd)-1, "SELECT email_notify_enable FROM wakeup_notify_settings;");
	email_notify_enable = conf_runsql_get_varchar_field(cmd,"email_notify_enable", pms->obj);
	snprintf(cmd, sizeof(cmd)-1, "SELECT w_times FROM %smaster.pms_wakeup_log WHERE w_id = '%s'", sqlprefix, pms->status);
	w_times = conf_runsql_get_varchar_field(cmd,"w_times", pms->obj);
	snprintf(cmd, sizeof(cmd)-1, "SELECT call_times FROM %smaster.pms_wakeup_log WHERE w_id = '%s'", sqlprefix, pms->status);
	call_times = conf_runsql_get_varchar_field(cmd,"call_times", pms->obj);
	
	if(call_times != NULL && w_times != NULL && 1 != atoi(pms->opt) && 6 != atoi(pms->opt) && atoi(call_times) >= atoi(w_times) + 1 && 0 == strcmp(email_notify_enable,"1"))
	{
		snprintf(cmd, sizeof(cmd)-1, "SELECT email_send_way FROM wakeup_notify_settings;");
		email_send_way = conf_runsql_get_varchar_field(cmd,"email_send_way", pms->obj);
		if(0 == strcmp(email_send_way,"1"))
		{
			snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /app/asterisk/var/lib/asterisk/scripts/sendMail.py wakeup_fail '' %s", pms->status);
			ast_safe_system2(cmd);
		}
	}

	snprintf(cmd, sizeof(cmd)-1, "SELECT vm_notify_enable FROM wakeup_notify_settings;");
	vm_notify_enable = conf_runsql_get_varchar_field(cmd,"vm_notify_enable", pms->obj);
	snprintf(cmd, sizeof(cmd)-1, "SELECT vm_extension FROM wakeup_notify_settings;");
	vm_extension = conf_runsql_get_varchar_field(cmd,"vm_extension", pms->obj);
	if(NULL != call_times && NULL != w_times && 0 != strlen(call_times) && 0 != strlen(w_times) && 1 != atoi(pms->opt) && 6 != atoi(pms->opt) && atoi(call_times) >= atoi(w_times) + 1 && 0 == strcmp(vm_notify_enable,"1"))
	{
		snprintf(cmd, sizeof(cmd)-1, "SELECT language FROM language_settings;");
		language = conf_runsql_get_varchar_field(cmd,"language", pms->obj);
		if(NULL != language)
		{
			snprintf(path_num_gsm,sizeof(path_num_gsm)-1,"%s%s/digits", PATH_GSM, language);
			snprintf(path_wakeup_gsm,sizeof(path_wakeup_gsm)-1,"%s%s", PATH_GSM, language);
		}
		snprintf(cmd,sizeof(cmd)-1,"/data/var/spool/asterisk/voicemail/default/%s/INBOX",vm_extension);
		if (vm_lock_path(cmd))
			return -100;
        cnt = count_messages(cmd);
		snprintf(msg_name, sizeof(msg_name)-1, "msg%04d", cnt);
		snprintf(sox,sizeof(sox)-1,"sox %s/%s",path_wakeup_gsm,"wakeup1.gsm");
		if(address != NULL)
		{
			for(int i = 0;i < strlen(address);i++)
			{
				snprintf(sox + strlen(sox),sizeof(sox) - strlen(sox) - 1," %s/%c.gsm",path_num_gsm,address[i]);
			}
		}
		snprintf(sox + strlen(sox),sizeof(sox) - strlen(sox) - 1," %s/%s %s/%s.gsm /data/var/spool/asterisk/voicemail/default/%s/INBOX/%s.WAV",path_wakeup_gsm,"wakeup2.gsm",path_wakeup_gsm,pms_status_mapping[atoi(pms->opt)].file_name,vm_extension,msg_name);
		ast_log(LOG_DEBUG, "%s\n", sox);
		if (access(cmd, F_OK) != -1) {
			ast_safe_system(sox);
		}else{
			snprintf(cmd,sizeof(cmd)-1,"mkdir -p /data/var/spool/asterisk/voicemail/default/%s/INBOX",vm_extension);
			ast_log(LOG_DEBUG, "%s",cmd);
			ast_safe_system(cmd);
			ast_safe_system(sox);
		}
		snprintf(file_path,sizeof(file_path)-1,"/data/var/spool/asterisk/voicemail/default/%s/INBOX/%s.txt",vm_extension,msg_name);
		txt = fopen(file_path, "w+");
		if (txt) {
			generate_msg_id(msg_id);
			get_date(date, sizeof(date));
			fprintf(txt,
				";\n"
				"; Message Information file\n"
				";\n"
				"[message]\n"
				"origmailbox=%s\n"
				"context=\n"
				"macrocontext=\n"
				"exten=\n"
				"rdnis=Unknown\n"
				"priority=11\n"
				"callerchan=Wakeup\n"
				"callerid=Wakeup\n"
				"origdate=%s\n"
				"origtime=%ld\n"
				"category=\n"
				"msg_id=%s\n"
				"flag=\n" /* flags not supported in copy from file yet */
				"duration=6\n", /* Don't have any reliable way to get duration of file. */
				pms->extension,
				date, (long) time(NULL),
				msg_id);
			fclose(txt);
			snprintf(cmd, sizeof(cmd), "SELECT COUNT(*) AS cnt FROM users WHERE user_name = '%s'",vm_extension);
			count = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
			if(count != NULL && 0 != strlen(count) && 0 == strcmp(count,"0"))
			{
				ast_log(LOG_DEBUG, "voicemail group");
				snprintf(cmd, sizeof(cmd), "SELECT email FROM vmgroups WHERE extension = '%s'",vm_extension);
				vm_email = conf_runsql_get_varchar_field(cmd,"email", pms->obj);
				if(NULL != vm_email && 0 != strlen(vm_email))
				{
					snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/sendMail.py voicemail '%s' %d '%s' \"\" '%s' \"\" '/data/var/spool/asterisk/voicemail/default/%s/INBOX/%s' 6 '' \"no\" \"\" \"\" '%s'"
			        			, vm_email, cnt, vm_extension, pms->extension, vm_extension, msg_name, pms->extension);
					ast_log(LOG_DEBUG, "%s",cmd);
					ast_safe_system(cmd);
				}
				ret = sycn_vm_group(chan,pms,vm_extension,msg_name);
			}
			else
			{
				snprintf(cmd, sizeof(cmd), "SELECT email FROM users WHERE user_name = '%s'",vm_extension);
				email_escape = conf_runsql_get_varchar_field(cmd,"email", pms->obj);
				if(NULL != email_escape)
            	{
					snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/sendMail.py voicemail '%s' %d '%s' \"\" '%s' \"\" '/data/var/spool/asterisk/voicemail/default/%s/INBOX/%s' 6 '' \"no\" \"\" \"\" '%s'"
			        		, email_escape, cnt, vm_extension, pms->extension, vm_extension, msg_name, pms->extension);
					ast_safe_system(cmd);
            	}	
			}
			snprintf(cmd, sizeof(cmd), "REPLACE INTO %svoicemail.voicemail_msg_info VALUES ('%s', 'INBOX', '\"Wakeup\" <Wakeup>', '%ld', '%s', 6, 'N', '%s', %d, 1000, 'file', 'WAV', 1,NULL,NULL,NULL)", sqlprefix, vm_extension , (long)time(NULL), date, msg_id, cnt);
			ret = update_pms_db(cmd, pms->obj);

			snprintf(cmd, sizeof(cmd)-1, "SELECT COUNT(*) AS cnt FROM %svoicemail.voicemail_msg_info WHERE type = 'INBOX' AND user = '%s';",sqlprefix,vm_extension);
			newmsgs = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
			snprintf(cmd, sizeof(cmd)-1, "SELECT COUNT(*) AS cnt FROM %svoicemail.voicemail_msg_info WHERE type = 'Old' AND user = '%s';",sqlprefix,vm_extension);
			oldmsgs = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
			snprintf(cmd, sizeof(cmd)-1, "SELECT COUNT(*) AS cnt FROM %svoicemail.voicemail_msg_info WHERE type = 'Urgent' AND user = '%s';",sqlprefix,vm_extension);
			urgentmsgs = conf_runsql_get_varchar_field(cmd,"cnt", pms->obj);
            if(newmsgs && oldmsgs && urgentmsgs)
            {
			    ast_manager_event(chan, EVENT_FLAG_CALL, "MessageWaiting", "Mailbox: %s\r\nNew: %d\r\nOld: %d\r\nUrg: %d\r\n", vm_extension, atoi(newmsgs), atoi(oldmsgs), atoi(urgentmsgs));
			    __ast_publish_mwi_state_channel(vm_extension, "default", atoi(urgentmsgs), atoi(newmsgs) + atoi(urgentmsgs), atoi(oldmsgs), NULL, NULL);
            }
            /* Since we are recording from a file, we shouldn't need to do anything else with
			/ * this txt file */
			snprintf(cmd,sizeof(cmd)-1,"/data/var/spool/asterisk/voicemail/default/%s/INBOX",vm_extension);
			ast_unlock_path(cmd);
		}
	}

	manager_event(EVENT_FLAG_CALL, "PMSWakeupStatus",
		"Extension: %s\r\n"
		"WakeupActionStatus: %s\r\n"
		"WakeupAnswerStatus: %s\r\n",
		pms->extension?pms->extension:"",
		pms->status?pms->status:"",
		pms->opt?pms->opt:"");

	AST_SAFE_FREE(email_notify_enable);
	AST_SAFE_FREE(email_send_way);
	AST_SAFE_FREE(address);
	AST_SAFE_FREE(vm_notify_enable);
	AST_SAFE_FREE(vm_extension);
	AST_SAFE_FREE(w_type);
	AST_SAFE_FREE(w_times);
	AST_SAFE_FREE(call_times);
	AST_SAFE_FREE(language);
	AST_SAFE_FREE(newmsgs);
	AST_SAFE_FREE(oldmsgs);
	AST_SAFE_FREE(urgentmsgs);
	AST_SAFE_FREE(email_escape);
	AST_SAFE_FREE(count);
	AST_SAFE_FREE(member_extension_string);
	AST_SAFE_FREE(vm_email);
	return ret;
}

static int update_st_status(struct ast_channel *chan, struct pms_task *pms)
{
	int ret = 0;
	char cmd[PMS_BUF_MAX_LEN] = {0};
	char temp[512] = {0};
	char set_time[64] = {0};
	const char *maid = NULL;
	char* pch = NULL;
	time_t timer;
	char s_date[PMS_BUF_MIN_LEN] = {0};
	char s_time[PMS_BUF_MIN_LEN] = {0};
	const char *report_status = NULL;
	int flag_minibar = 0; 
	int tax = 0;
	int cnt = 0;
	time( &timer);
	strftime(s_time, sizeof(s_time)-1, "%H%M", localtime(&timer));
	strftime(s_date, sizeof(s_date)-1, "%Y%m%d", localtime(&timer));

	report_status = pbx_builtin_getvar_helper(chan, "MINIREPROTSTATUS");
	if (report_status != NULL && strcasecmp(report_status, "yes") == 0){
		flag_minibar = 1;
	}
	maid = pbx_builtin_getvar_helper(chan, "MAID");
	if (flag_minibar){
		snprintf(cmd, sizeof(cmd)-1, "UPDATE pms_room SET minibar_code = '%s', s_date = '%s', s_time = '%s', maid = '%s', mb_send = 1 WHERE address = (select address from pms_room_members where extension='%s')", \
									pms->status, s_date, s_time, maid?maid:"0000", pms->extension);
		snprintf(set_time,sizeof(set_time)-1,"%s%s", s_date, s_time);
		pch = strtok(pms->status, "*");
		while (pch != NULL)
		{
			snprintf(temp,sizeof(temp)-1,"SELECT COUNT(*) FROM minibar_billing WHERE goods_name = (SELECT goods_name FROM minibar_goods WHERE extension = '%s') AND time = %s",pch,set_time);
			cnt = conf_runsql_get_int_field(temp,pms->obj);
			snprintf(temp,sizeof(temp)-1,"SELECT goods_tax FROM minibar_goods WHERE extension ='%s'",pch);
			tax = conf_runsql_get_int_field(temp,pms->obj);
			if(0 == cnt)
			{
				if(0 == tax)
				{
					snprintf(temp,sizeof(temp)-1,"INSERT minibar_billing (order_id,goods_name,count,time,amount,maid,tax)VALUE((SELECT order_id FROM pms_room WHERE address = (select address from pms_room_members where extension='%s')),(SELECT goods_name FROM minibar_goods WHERE extension = '%s'),1,%s,(SELECT amount * (SELECT global_tax FROM minibar_settings) FROM minibar_goods WHERE extension = '%s'),'%s',(SELECT global_tax FROM minibar_settings))" ,\
			                                 	pms->extension,pch,set_time,pch,maid?maid:"0000");
					ret = update_pms_db(temp, pms->obj);
				}
				else
				{
					snprintf(temp,sizeof(temp)-1,"INSERT minibar_billing (order_id,goods_name,count,time,amount,maid,tax)VALUE((SELECT order_id FROM pms_room WHERE address = (select address from pms_room_members where extension='%s')),(SELECT goods_name FROM minibar_goods WHERE extension = '%s'),1,%s,(SELECT amount * goods_tax FROM minibar_goods WHERE extension = '%s'),'%s',(SELECT goods_tax FROM minibar_goods WHERE extension ='%s'))" ,\
			                                 	pms->extension,pch,set_time,pch,maid?maid:"0000",pch);
					ret = update_pms_db(temp, pms->obj);
				}
			}
			else
			{
				if(0 == tax)
				{
					snprintf(temp,sizeof(temp)-1,"UPDATE minibar_billing SET count = count + 1,amount = amount + (SELECT amount * (SELECT global_tax FROM minibar_settings) FROM minibar_goods WHERE extension = '%s') WHERE goods_name = (SELECT goods_name FROM minibar_goods WHERE extension = '%s') AND time = '%s'",\
					                            pch,pch,set_time);
					ret = update_pms_db(temp, pms->obj);
				}
				else
				{
					snprintf(temp,sizeof(temp)-1,"UPDATE minibar_billing SET count = count + 1,amount = amount + (SELECT amount * goods_tax FROM minibar_goods WHERE extension = '%s') WHERE goods_name = (SELECT goods_name FROM minibar_goods WHERE extension = '%s') AND time = '%s'",\
					                            pch,pch,set_time);
					ret = update_pms_db(temp, pms->obj);
				}
			}
			printf("%s\n", pch);
			pch = strtok(NULL, "*");
		}
	}else{
		snprintf(cmd, sizeof(cmd)-1, "UPDATE pms_room SET status = '%s', s_date = '%s', s_time = '%s', maid = '%s', send_status = 1 WHERE address = (select address from pms_room_members where extension='%s');", \
									pms->status, s_date, s_time, maid?maid:"0000", pms->extension);
	}	
#ifdef HAVE_UNIXODBC
	ret = update_pms_db(cmd, pms->obj);
#endif
	if (flag_minibar == 0){
		if(ret == 0){
			ast_play_voice_prompt(chan, PMS_SET_SUCCESS);
		}else{
			ast_play_voice_prompt(chan, PMS_CAN_NOT_SET);
		}
		manager_event(EVENT_FLAG_CALL, "PMSRoomStatus",
			"Extension: %s\r\n"
			"RoomActionStatus: %s\r\n",
			pms->extension?pms->extension:"",
			pms->status?pms->status:"");
		int len_ha = strlen(pms->extension?pms->extension:"")+strlen(pms->status?pms->status:"")+strlen("Extension: \r\nRoomActionStatus: \r\n")+1;
		char *ha_data = ast_malloc(len_ha);
		if(NULL == ha_data)
		{
			ast_log(LOG_ERROR, "malloc ha_data error.");
			return -1;
		}
		memset(ha_data, 0, len_ha);
		snprintf(ha_data, len_ha, "Extension: %s\r\nRoomActionStatus: %s\r\n", pms->extension?pms->extension:"", pms->status?pms->status:"");
		ast_ha_send_sql(ha_data, AST_HA_PMS);
		AST_SAFE_FREE(ha_data);
	}else{
		if(ret != 0){
			pbx_builtin_setvar_helper(chan, "MINIREPROTSTATUS", "no");
		}
	}

	return ret;
}

static int update_sr_status(struct ast_channel *chan, struct pms_task *pms)
{

	int ret = 0;
	char temp[512] = {0};
	char set_time[64] = {0};
	const char *maid = NULL;
	char* pch = NULL;
	time_t timer;
	char s_date[PMS_BUF_MIN_LEN] = {0};
	char s_time[PMS_BUF_MIN_LEN] = {0};
	int tax = 0;
	int cnt = 0;
	time( &timer);
	strftime(s_time, sizeof(s_time)-1, "%H%M", localtime(&timer));
	strftime(s_date, sizeof(s_date)-1, "%Y%m%d", localtime(&timer));
	maid = pbx_builtin_getvar_helper(chan, "MAID");

	snprintf(set_time,sizeof(set_time)-1,"%s%s", s_date, s_time);
	pch = strtok(pms->status, "*");
	while (pch != NULL)
	{
		snprintf(temp,sizeof(temp)-1,"SELECT COUNT(*) FROM minibar_billing WHERE goods_name = (SELECT goods_name FROM minibar_goods WHERE extension = '%s') AND time = %s",pch,set_time);
		cnt = conf_runsql_get_int_field(temp,pms->obj);
		snprintf(temp,sizeof(temp)-1,"SELECT goods_tax FROM minibar_goods WHERE extension ='%s'",pch);
		tax = conf_runsql_get_int_field(temp,pms->obj);
		if(0 == cnt)
		{
			if(0 == tax)
			{
				snprintf(temp,sizeof(temp)-1,"INSERT minibar_billing (order_id,goods_name,count,time,amount,maid,tax)VALUE((SELECT order_id FROM pms_room WHERE address = (select address from pms_room_members where extension='%s')),(SELECT goods_name FROM minibar_goods WHERE extension = '%s'),-1,%s,-(SELECT amount * (SELECT global_tax FROM minibar_settings) FROM minibar_goods WHERE extension = '%s'),'%s',(SELECT global_tax FROM minibar_settings))" ,\
			                                pms->extension,pch,set_time,pch,maid?maid:"0000");
				ret = update_pms_db(temp, pms->obj);
			}
			else
			{
				snprintf(temp,sizeof(temp)-1,"INSERT minibar_billing (order_id,goods_name,count,time,amount,maid,tax)VALUE((SELECT order_id FROM pms_room WHERE address = (select address from pms_room_members where extension='%s')),(SELECT goods_name FROM minibar_goods WHERE extension = '%s'),-1,%s,-(SELECT amount * goods_tax FROM minibar_goods WHERE extension = '%s'),'%s',(SELECT goods_tax FROM minibar_goods WHERE extension ='%s'))" ,\
			                                pms->extension,pch,set_time,pch,maid?maid:"0000",pch);
				ret = update_pms_db(temp, pms->obj);
			}
		}
		else
		{
			if(0 == tax)
			{
				snprintf(temp,sizeof(temp)-1,"UPDATE minibar_billing SET count = count - 1,amount = amount - (SELECT amount * (SELECT global_tax FROM minibar_settings) FROM minibar_goods WHERE extension = '%s') WHERE goods_name = (SELECT goods_name FROM minibar_goods WHERE extension = '%s') AND time = '%s'",\
					                        pch,pch,set_time);
				ret = update_pms_db(temp, pms->obj);
			}
			else
			{
				snprintf(temp,sizeof(temp)-1,"UPDATE minibar_billing SET count = count - 1,amount = amount - (SELECT amount * goods_tax FROM minibar_goods WHERE extension = '%s') WHERE goods_name = (SELECT goods_name FROM minibar_goods WHERE extension = '%s') AND time = '%s'",\
					                        pch,pch,set_time);
				ret = update_pms_db(temp, pms->obj);
			}
		}
		printf("%s\n", pch);
		pch = strtok(NULL, "*");
	}
	return ret;
}

/*** update ucm wakeup execute status ***/
static int update_ucm_wakeup_execute_status(struct ast_channel *chan, struct pms_task *wakeup)
{
	int ret = 0;
	char cmd[256] = {0};
	const char *execute_date = NULL;
	const char *wakeup_index = NULL;

	execute_date = pbx_builtin_getvar_helper(chan, "UCM_WAKEUP_DATE");
	wakeup_index = pbx_builtin_getvar_helper(chan, "WAKEUP_INDEX");

	if (execute_date == NULL || wakeup_index == NULL  || wakeup == NULL)
	{
	    return -1;
	}

	snprintf(cmd, sizeof(cmd)-1, "update ucm_wakeup_members set execute_status='%s',answer_status='%s', execute_date='%s' "
								"where member_extension='%s' and wakeup_index = '%s';",
								wakeup->status, wakeup->opt, execute_date, wakeup->extension, wakeup_index);
#ifdef HAVE_UNIXODBC
	ret = update_pms_db(cmd, wakeup->obj);
#endif

	return ret;
}

static struct pms_task *create_pms_task(void)
{
	struct pms_task *tmp_task = NULL;
	tmp_task = (struct pms_task *)ast_malloc(PMS_SIZE);
	if (tmp_task){
		memset(tmp_task, 0, PMS_SIZE);
		return tmp_task;
	}else{
		return NULL;
	}
}

static int pms_exec(struct ast_channel *chan, const char *data)
{
	int ret = 0;
	char *parse = NULL;
	const char *exten = NULL;
	struct pms_task *p_task = NULL;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(pms_type);
		AST_APP_ARG(pms_status);
		AST_APP_ARG(pms_opt);
	);

	ast_debug(1, "enter the apply of pms status settings! \n");
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires at least one argument (destination(s) to pms status settings)\n");
		return -1;
	}

	p_task = create_pms_task();
	if (p_task == NULL){
		ast_log(LOG_WARNING, "create task fail!");
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	AST_GET_PMS_ARGS(p_task, p_task->type, args.pms_type);
	AST_GET_PMS_ARGS(p_task, p_task->status, args.pms_status);
	AST_GET_PMS_ARGS(p_task, p_task->opt, args.pms_opt);
	exten = pbx_builtin_getvar_helper(chan, "TARGET");
	AST_GET_PMS_ARGS(p_task, p_task->extension, exten);
	ast_debug(1, "type: %s, status: %s, opt: %s, extension: %s ",
				p_task->type, p_task->status, p_task->opt, p_task->extension);

#ifdef HAVE_UNIXODBC
	p_task->obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);

	if (!p_task->obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
		return -1;
	}
#endif

	if(0 == strcmp(p_task->type, "wake")){
		ret = update_wakeup_execute_status(chan, p_task);
	}else if(0 == strcmp(p_task->type, "st")){
		ret = update_st_status(chan, p_task);
	}else if(0 == strcmp(p_task->type, "sr")){
		ret = update_sr_status(chan, p_task);
	}else if(0 == strcmp(p_task->type, "ucmWake")){
		ret = update_ucm_wakeup_execute_status(chan, p_task);
	}

	if(0 != ret){
		ast_log(LOG_WARNING, "update extension[%s] information fail!", p_task->extension);
	}

#ifdef HAVE_UNIXODBC
	ast_core_odbc_release_obj(p_task->obj);
#endif

	clean_pms_task(p_task);

	ast_debug(1, "end the apply of pms status settings! \n");
	return 0;
}

static int unload_module(void)
{
	int res;
	res = ast_unregister_application(app);
	return res;
}

static int load_module(void)
{
	int res;

	/*Get HA status*/
	pms_get_haon();
	res = ast_register_application_xml(app, pms_exec);
	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Trivial PMS Status SettingsApplication",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
);

#endif
