/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2024-5-17, yjzhao <chzhdeng@grandstream.cn>
 *
 * app_reblackcheck.c adapted from the remote black_phone query
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>sqlite3</depend>
	<support_level>core</support_level>
 ***/
#include "asterisk.h"
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <curl/curl.h>
#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/options.h"
#include "asterisk/config.h"
#include "asterisk/module.h"
#include "asterisk/enum.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/ucm_dbi.h"

#include "asterisk/core_odbc.h"
#include "asterisk/haserver.h"

 
static char *tdesc = "connect do-not-call-me.vercel.app :)";
static char *app = "reblackcheck";
static char *synopsis = "This module is used to connect to remote database related APIs"
	"If you need it, you can enable this function on the UCM page: Extension/Relay ->Outbound Routing ->Outbound Blacklist"
	"This feature was proposed by a customer in Israel"
	"If you are unsure that the customer you are calling is not the person you want, you can use this feature";

typedef enum{
	BLACK_INIT = 0,
	BLACK_YES,
	BLACK_NO,
}BlackAction;

#define UCM_CONF_DB "/cfg/etc/ucm_config.db"
//#define REMOTE_MYSQL_URL "https://do-not-call-me.vercel.app/api/number?number=%s"
#define REMOTE_MYSQL_URL "https://dontcallme.vercel.app/api/number?number=%s"
#define LOCAL_MYSQL_REQ "select * from blacklist_remote_settings limit 1"
#define BLACKINFO_SIZE sizeof(BlackOutInfo)
#define BLACKOUT_OK 	  0
#define BLACKOUT_ERR	  1
#define SQL_SIZE		  512
#define DESC(a) 		  1
#define BLACK_TOKEN_SIZE  500
#define BLACK_URL_SIZE	  200
#define BLACK_AUTH_SIZE   500

typedef struct _BlackOutInfo{
	char phone_num[100];
	int time_out;
	long res_http_code;
	BlackAction time_out_action;
	char token[BLACK_TOKEN_SIZE];
	struct odbc_obj* obj;
}BlackOutInfo;

struct generic_prepare_struct{
	char *sql;
	int argc;
	char **argv;
};

AST_MODULE_LOG("app_reblackcheck");

#if DESC("function")

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void* data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res))
	{
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}

	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}

	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}


// id, enable, token, timeout, timeout_handle

static int black_get_arg_from_db(BlackOutInfo* black)
{
	char sql[SQL_SIZE] = { 0 };
	SQLHSTMT stmt = NULL;
	char rowdata[BLACK_TOKEN_SIZE] = { 0 };
	struct generic_prepare_struct gps;
	int res = 0;

	gps.argc = 0;
	gps.argv = NULL;
	snprintf(sql, sizeof(sql) - 1, LOCAL_MYSQL_REQ);
	gps.sql = sql;
	stmt = ast_core_odbc_prepare_and_execute(black->obj, generic_prepare, &gps);
	if (!stmt)
	{
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sql);
		return BLACKOUT_ERR;
	}

	res = SQLFetch(stmt);
	if (!SQL_SUCCEEDED(res))
	{
		ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return BLACKOUT_ERR;
	}

	// get token
	res = SQLGetData(stmt, 2, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res))
	{
		ast_log(LOG_ERROR, "SQL Get Data null!\n[%s]\n\n", sql);
	}
	strncpy(black->token, rowdata, BLACK_TOKEN_SIZE - 1);
	ast_log(LOG_ERROR, "reblackcheck get token is %s", rowdata);

	// get timeout
	memset(rowdata, 0, BLACK_TOKEN_SIZE);
	res = SQLGetData(stmt, 3, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res))
	{
		ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return BLACKOUT_ERR;
	}
	if (rowdata[0]!='\0')
	{
		black->time_out = atoi(rowdata);
		ast_log(LOG_ERROR, "sql data is %s", rowdata);
	}
	else
	{
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return BLACKOUT_ERR;
	}	

	// get timeout_handle
	memset(rowdata, 0, BLACK_TOKEN_SIZE);
	res = SQLGetData(stmt, 4, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res))
	{
		ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return BLACKOUT_ERR;
	}

	if (rowdata[0]!='\0')
	{
		ast_log(LOG_ERROR, "sql data is %s", rowdata);
		if(strncasecmp(rowdata, "no", 2) == 0)
		{
			black->time_out_action = BLACK_NO;
		}
		else if((strncasecmp(rowdata, "yes", 3)) == 0)
		{
			black->time_out_action = BLACK_YES;
		}
		else
		{
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			ast_log(LOG_ERROR, "SQL get error data: %s", rowdata);
			return BLACKOUT_ERR;
		}
	}
	else
	{
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		ast_log(LOG_ERROR, "get data is null");
		return BLACKOUT_ERR;
	}

	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	return BLACKOUT_OK;
}


#endif 


#if DESC("the interface of main process")

static int arg_parser(const char* data, char* out_phone, int phone_len)
{
	char* parser = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(param1);
	);

	if (ast_strlen_zero(data))
	{
		ast_log(LOG_ERROR, "MYAPP requires an argument (param1|param2)\n"); 
		return BLACKOUT_ERR;
	}

	parser = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parser);

	if (ast_strlen_zero(args.param1))
	{
		ast_log(LOG_ERROR, "MYAPP requires an argument (callee|filename)\n"); 
		return BLACKOUT_ERR;
	}
	//*out_phone = atoi(args.param1);
	strncpy(out_phone, args.param1, phone_len - 1);
	out_phone[phone_len] = '\0';
	ast_log(LOG_ERROR, "myapp get param is %s, out_phone is %s", args.param1, out_phone);

	return BLACKOUT_OK;
}


static int black_create_info(BlackOutInfo** black, char* out_phone)
{
	*black = (BlackOutInfo *)malloc(BLACKINFO_SIZE);
	if(!(*black))
	{
		ast_log(LOG_ERROR, "malloc err!\n");
		return BLACKOUT_ERR;
	}
	memset((*black)->token, 0, BLACK_TOKEN_SIZE);
	memset((*black)->phone_num, 0, 100);
	strncpy((*black)->phone_num, out_phone, 99);
	((*black)->phone_num)[99] = '\0';
	//(*black)->phone_num = out_phone;
	(*black)->time_out = 0;
	(*black)->time_out_action = BLACK_INIT;
	(*black)->obj = NULL;
	(*black)->res_http_code = -1;

	return BLACKOUT_OK;
}	


static int black_get_args(BlackOutInfo* black)
{
	int res = 0;

	black->obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);
	if (!black->obj)
	{
		ast_log(LOG_ERROR, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
		return BLACKOUT_ERR;
	}

	// get args from mysql
	res = black_get_arg_from_db(black);
	if(res == BLACKOUT_ERR)
	{
		ast_core_odbc_release_obj(black->obj);
		black->obj = NULL;
		return BLACKOUT_ERR;
	}

	ast_core_odbc_release_obj(black->obj);
	black->obj = NULL;

	return BLACKOUT_OK;
}

static int black_phone_match(BlackOutInfo* black)
{
	CURL* curl;
	CURLcode res;
	char url[BLACK_URL_SIZE] = { 0 };
	char authorization[BLACK_AUTH_SIZE] = { 0 };
	struct curl_slist *headers = NULL;

	res = curl_global_init(CURL_GLOBAL_DEFAULT);
	if(res != CURLE_OK)
	{
		ast_log(AST_LOG_ERROR, "curl global init err");
		return BLACKOUT_ERR;
	}
	curl = curl_easy_init();
	if(curl == NULL)
	{
		ast_log(AST_LOG_ERROR, "crul easy init err");
		return BLACKOUT_ERR;
	}

	snprintf(url, sizeof(url) - 1,REMOTE_MYSQL_URL, black->phone_num);
	ast_log(LOG_ERROR, "%s", url);
	snprintf(authorization, sizeof(authorization) - 1, "Authorization: Bearer %s", black->token);
	headers = curl_slist_append(headers, authorization);
	curl_easy_setopt(curl, CURLOPT_URL, url);
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headers);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, (long)black->time_out);
	//curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, (long)black->time_out);
	ast_log(AST_LOG_ERROR, "cancel ssl crt");
	curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0L);
	curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0L);

	res = curl_easy_perform(curl);
	if(res != CURLE_OK)
	{
		if(res == CURLE_OPERATION_TIMEDOUT)
		{
			ast_log(AST_LOG_ERROR, "request remote server err as curl timeout");
			black->res_http_code = 408;
			goto ext;
		}
		else
		{
			ast_log(AST_LOG_ERROR, "request remote server err,curl exit code is %d", res);
			black->res_http_code = 408;
			goto ext;
		}
		curl_easy_cleanup(curl);
		curl_global_cleanup();
		return BLACKOUT_ERR;
	}
	curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &(black->res_http_code));
	ast_log(LOG_ERROR, "http_code is %ld\n", black->res_http_code);

ext:
	curl_easy_cleanup(curl);
	curl_global_cleanup();
	return BLACKOUT_OK;
}


#endif



#if DESC("main process")
static int reblackcheck_exec(struct ast_channel *chan, const char* data)
{
	int res = 0;
	char out_phone[100] = { 0 };
	BlackOutInfo* black_info = NULL;

	res = arg_parser(data, out_phone, sizeof(out_phone));
	if(res == BLACKOUT_ERR)
	{
		goto out;
	}

	res = black_create_info(&black_info, out_phone);
	if(res == BLACKOUT_ERR)
	{
		free(black_info);
		goto out;
	}

	res = black_get_args(black_info);
	if(res == BLACKOUT_ERR)
	{
		free(black_info);
		goto out;
	}

	ast_log(LOG_ERROR, "get data is complete");

	res = black_phone_match(black_info);
	if(res == BLACKOUT_ERR)
	{
		free(black_info);
		goto out;
	}
	//200:ok  401:authorization err 403:forbidden 408：timeout
	switch (black_info->res_http_code)
	{
		case 200:
			ast_log(LOG_ERROR, "calls allow");
			pbx_builtin_setvar_helper(chan, "BLACK_OUT", "yes");
			break;
		case 403:
			ast_log(LOG_ERROR, "no calls allow");
			pbx_builtin_setvar_helper(chan, "BLACK_OUT", "no");
			break;
		case 401:
		case 408:
		default:
			if(black_info->time_out_action == BLACK_YES)
			{
				ast_log(LOG_ERROR, "calls allow");
				pbx_builtin_setvar_helper(chan, "BLACK_OUT", "yes");
			}
			else
			{
				ast_log(LOG_ERROR, "no calls allow");
				pbx_builtin_setvar_helper(chan, "BLACK_OUT", "no");
			}
	}

	free(black_info);
out:
	//ast_module_user_remove(u);

	return BLACKOUT_OK;
} 

#endif

static int load_module(void)
{
	return ast_register_application(app, reblackcheck_exec, synopsis, tdesc);	
}
static int unload_module(void)
{
	int res;
	res = ast_unregister_application(app);
	ast_module_user_hangup_all();
	return res;
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "reblackcheck");


