/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2024-5-27, yjzhao <chzhdeng@grandstream.cn>
 *
 * app_reblackcheck.c adapted from the remote black_phone query
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
    <depend>sqlite3</depend>
    <support_level>core</support_level>
 ***/
#include "asterisk.h" 

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <curl/curl.h>
#include <mysql/mysql.h>
#include <dbi/dbi.h>

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/options.h"
#include "asterisk/config.h"
#include "asterisk/module.h"
#include "asterisk/enum.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/ucm_dbi.h"

#include "asterisk/core_odbc.h"
#include "asterisk/haserver.h"


static char *tdesc = "query name of incaller_phone_number from remote DB :)"; 
static char *app = "reincallerquery"; 
static char *synopsis = "This module is used to connect to remote databases to query relevant information"
    "If you need it, you can enable this function on the UCM page: Extension/Relay ->Inbound Routing"
    "This feature was proposed by some customer"
    "If you're curious about who dials your number, you can use this feature"; 

enum{
    SQL_INCALLER_HOST = 2,
    SQL_INCALLER_DB,
    SQL_INCALLER_TABLE,
    SQL_INCALLER_NUMBER,
    SQL_INCALLER_NAME,
    SQL_INCALLER_USER,
    SQL_INCALLER_PASSWD,
    SQL_INCALLER_CS
};


#define UCM_CONF_DB "/cfg/etc/ucm_config.db"
#define INCALLER_OK          0
#define INCALLER_ERR         1
#define SQL_SIZE             512
#define DESC(a)              1
#define INCALLER_CMD_LEN    500
#define INCALLER_DEFAULT_LEN 100
#define DEFAULT_TIMEOUT      10
#define DEFAULT_SQL_PORT     "3306"

typedef struct _RemoteQueryInfo{
    char remote_host[INCALLER_DEFAULT_LEN];
    char remote_db[INCALLER_DEFAULT_LEN];
    char remote_table[INCALLER_DEFAULT_LEN];
    char remote_name[INCALLER_DEFAULT_LEN];
    char remote_number[INCALLER_DEFAULT_LEN];
    char remote_user[INCALLER_DEFAULT_LEN];
    char remote_passwd[INCALLER_DEFAULT_LEN];
    char remote_cs[INCALLER_DEFAULT_LEN];
    int phone_number;
    int time_out;
    struct odbc_obj* obj;
}RemoteQueryInfo;
#define INCALLERINFO_SIZE sizeof(RemoteQueryInfo)


struct generic_prepare_struct {
    char *sql;
    int argc;
    char **argv;
};

AST_MODULE_LOG("app_incallerquery");

#if DESC("function")
static SQLHSTMT generic_prepare(struct odbc_obj *obj, void* data)
{
    struct generic_prepare_struct *gps = data;
    int res, i;
    SQLHSTMT stmt;

    res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
    if (!SQL_SUCCEEDED(res)) 
    {
        ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
        return NULL;
    }

    res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
    if (!SQL_SUCCEEDED(res)) 
    {
        ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
        SQLFreeHandle(SQL_HANDLE_STMT, stmt);
        return NULL;
    }

    for (i = 0; i < gps->argc; i++)
        SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

    return stmt;
}




static int incaller_get_arg_from_db(RemoteQueryInfo* incaller)
{
    char sql[SQL_SIZE] = { 0 };
    SQLHSTMT stmt = NULL;
    char rowdata[INCALLER_CMD_LEN] = { 0 };
    struct generic_prepare_struct gps;
    int res = 0;
    int i = 0;

    gps.argc = 0;
    gps.argv = NULL;
    sprintf(sql, "select * from incaller_query_settings limit 1");
    gps.sql = sql;
    stmt = ast_core_odbc_prepare_and_execute(incaller->obj, generic_prepare, &gps);
    if (!stmt) 
    {
        ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sql);
        return INCALLER_ERR;
    }

    res = SQLFetch(stmt);
    if (!SQL_SUCCEEDED(res))
    {
        ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", sql);
        SQLFreeHandle(SQL_HANDLE_STMT, stmt);
        return INCALLER_ERR;
    }

    for(i = 2; i <= 9; ++i)
    {
        memset(rowdata, 0, INCALLER_CMD_LEN);
        res = SQLGetData(stmt, i, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
        if (!SQL_SUCCEEDED(res)) 
        {
            ast_log(LOG_ERROR, "i:%d, SQL Get Data null!\n[%s]\n\n", i, sql);
            SQLFreeHandle(SQL_HANDLE_STMT, stmt);
            return INCALLER_ERR;
        }
        if (strlen(rowdata) == 0)
        {
            ast_log(LOG_ERROR, "i:%d data is null", i);
            SQLFreeHandle(SQL_HANDLE_STMT, stmt);
            return INCALLER_ERR;
        }
        switch(i)
        {
            case SQL_INCALLER_HOST:
                strncpy(incaller->remote_host, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_host[INCALLER_DEFAULT_LEN - 1] = '\0';
                ast_log(LOG_ERROR, "host data is %s", rowdata);
                break;
            case SQL_INCALLER_DB:
                strncpy(incaller->remote_db, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_db[INCALLER_DEFAULT_LEN - 1] = '\0';
                ast_log(LOG_ERROR, "db data is %s", rowdata);
                break;
            case SQL_INCALLER_TABLE:
                strncpy(incaller->remote_table, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_table[INCALLER_DEFAULT_LEN - 1] = '\0';
                ast_log(LOG_ERROR, "table data is %s", rowdata);
                break;
            case SQL_INCALLER_NUMBER:
                strncpy(incaller->remote_number, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_number[INCALLER_DEFAULT_LEN - 1] = '\0';
                ast_log(LOG_ERROR, "number data is %s", rowdata);
                break;
            case SQL_INCALLER_NAME:
                strncpy(incaller->remote_name, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_name[INCALLER_DEFAULT_LEN - 1] = '\0';
                ast_log(LOG_ERROR, "name data is %s", rowdata);
                break;
            case SQL_INCALLER_USER:
                strncpy(incaller->remote_user, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_user[INCALLER_DEFAULT_LEN - 1] = '\0';
                //ast_log(LOG_ERROR, "user data is %s", rowdata);
                break;
            case SQL_INCALLER_PASSWD:
                strncpy(incaller->remote_passwd, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_passwd[INCALLER_DEFAULT_LEN - 1] = '\0';
                //ast_log(LOG_ERROR, "passwd data is %s", rowdata);
                break;
            case SQL_INCALLER_CS:
                strncpy(incaller->remote_cs, rowdata, INCALLER_DEFAULT_LEN - 1);
                incaller->remote_cs[INCALLER_DEFAULT_LEN - 1] = '\0';
                ast_log(LOG_ERROR, "cs data is %s", rowdata);
                break;
            default:
                return INCALLER_ERR;
        }
    }

    SQLFreeHandle(SQL_HANDLE_STMT, stmt);
    return INCALLER_OK; 
}


static int sqlContainComment(const char* sql)
{
    char* str = NULL;

    char *comment = strstr(sql, "--");
    if (NULL == comment)
    {
        return 0;
    }

    char* begin_quote = strchr(sql, '\'');
    if (NULL == begin_quote)
    {
        return 1;
    }

    char *end_quote = strchr(begin_quote+1, '\'');
    if (NULL == end_quote)
    {
        return 1;
    }

    while(1)
    {
        if (comment > end_quote)
        {
            str = end_quote+1;
            if (*str != '\0')
            {
                begin_quote = strchr(str, '\'');
                if (NULL == begin_quote)
                {
                    return 1;
                }
                end_quote = strchr(begin_quote+1, '\'');
                if (NULL == end_quote)
                {
                    //cgilog(SYSLOG_DEBUG, "invalid sql:%s!", sql);
                    return 1;
                }
            }
            else
            {
                return 1;
            }
        }

        if (comment > begin_quote && comment < end_quote)
        {
            str = end_quote+1;
            if (*str != '\0')
            {
                comment = strstr(str, "--");
            }
            else
            {
                return 0;
            }
            if (NULL == comment)
            {
                return 0;
            }
        }

        if (comment < begin_quote)
        {
            return 1;
        }
    }
    return 0;
}



static dbi_conn incaller_dbi_connect_db(RemoteQueryInfo* incaller, dbi_inst* pInst)
{
    dbi_conn conn = NULL;
    char dbdPath[256] = { 0 };

    snprintf(dbdPath, sizeof(dbdPath) - 1, "%s/lib/dbd", DBI_DBD_PATH_PREFIX);
    dbi_initialize_r(dbdPath, pInst);

    conn = dbi_conn_new_r("mysql", *pInst);
    if ( conn == NULL )
    {
        ast_log(LOG_ERROR, "DBI: new connection error !" );
        return NULL;
    }

    dbi_conn_set_option(conn, "host", incaller->remote_host);
    dbi_conn_set_option(conn, "username", incaller->remote_user);
    dbi_conn_set_option(conn, "password", incaller->remote_passwd);
    dbi_conn_set_option(conn, "port", DEFAULT_SQL_PORT);
    dbi_conn_set_option(conn, "dbname", incaller->remote_db);
    dbi_conn_set_option(conn, "encoding", incaller->remote_cs);
    dbi_conn_set_option_numeric(conn, "timeout", 20);

    if(dbi_conn_connect(conn) < 0)
    {
        const char *errorMsg = NULL;
        int errorNum = dbi_conn_error(conn, &errorMsg);
        ast_log(LOG_ERROR, "Could not connect. Please check the option settings, Error code is %d, Msg:{%s}",
               errorNum, errorMsg == NULL ? "" : errorMsg);
        dbi_shutdown_r(*pInst);
        return NULL;
    }

    ast_log(LOG_DEBUG, "ast_dbi_connect_db connect to %s", incaller->remote_db);
    return conn;
}


static char* incaller_dbi_getOneVarcharField(dbi_conn connection, const char *sqlCmd, const char* fieldName )
{
    char* ret = NULL;
    dbi_result result = NULL;

    if ( connection == NULL || sqlCmd == NULL )
    {
        ast_log(LOG_ERROR, "ast_dbi_getOneVarcharField : param error!");
        return ret;
    }
    ast_log(LOG_DEBUG, "Enter ast_dbi_getOneVarcharField : cmd is {%s}", sqlCmd );

    // sql inject
    if(sqlContainComment(sqlCmd))
    {
        ast_log(LOG_ERROR, "Detected sql injection attack!");
        return ret;
    }
    result = dbi_conn_query(connection, sqlCmd);
    if(result)
    {
        while(dbi_result_next_row(result))
        {
            ret = dbi_result_get_as_string_copy(result, fieldName);
            break;
        }

        dbi_result_free(result);
    }
    else
    {
        const char *errorMsg = NULL;
        int errorNum = dbi_conn_error(connection, &errorMsg);
        ast_log(LOG_ERROR, "runSQLGetOneVarcharField : error message is %s, ID is %d,sql:%s", errorMsg ? errorMsg : "", errorNum,sqlCmd);
    }
    return ret;
}



static int incaller_dbi_disconnect_db(dbi_conn connection, dbi_inst* pInst)
{
    if (connection != NULL)
    {
        dbi_conn_close(connection);
        connection = NULL;
    }

    dbi_shutdown_r(*pInst);
    return 0;
}

#endif


#if DESC("main interface function")

static int arg_parser(const char* data, int* out_phone)
{
    char* parser = NULL;
    AST_DECLARE_APP_ARGS(args,
        AST_APP_ARG(param1);
    );

    if (ast_strlen_zero(data)) 
    {
        ast_log(LOG_ERROR, "MYAPP requires an argument (param1|param2)\n"); 
        return INCALLER_ERR; 
    }

    parser = ast_strdupa(data); 
    AST_STANDARD_APP_ARGS(args, parser); 

    if (ast_strlen_zero(args.param1) ) 
    {
        ast_log(LOG_ERROR, "MYAPP requires an argument (callee|filename)\n"); 
        return INCALLER_ERR;
    }
    *out_phone = atoi(args.param1); 
    ast_log(LOG_ERROR, "myapp get param is %s, out_phone is %d", args.param1, *out_phone);

    return INCALLER_OK;
}


static int incaller_create_info(RemoteQueryInfo** incaller, int in_phone)
{
    *incaller = (RemoteQueryInfo *)malloc(INCALLERINFO_SIZE);
    if(!(*incaller))
    {
        ast_log(LOG_ERROR, "malloc err!\n");
        return INCALLER_ERR;
    }
    memset((*incaller)->remote_host, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_db, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_table, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_number, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_name, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_user, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_passwd, 0, INCALLER_DEFAULT_LEN);
    memset((*incaller)->remote_cs, 0, INCALLER_DEFAULT_LEN);
    (*incaller)->phone_number = in_phone;
    (*incaller)->time_out = DEFAULT_TIMEOUT;
    (*incaller)->obj = NULL;

    return INCALLER_OK;
}   


static int incaller_get_args(RemoteQueryInfo* incaller)
{
    int res = 0;

    incaller->obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);
    if (!incaller->obj) 
    {
        ast_log(LOG_ERROR, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
        return INCALLER_ERR;
    }

    res = incaller_get_arg_from_db(incaller);
    if(res == INCALLER_ERR)
    {
        ast_core_odbc_release_obj(incaller->obj);
        incaller->obj = NULL;
        return INCALLER_ERR;
    }

    ast_core_odbc_release_obj(incaller->obj);
    incaller->obj = NULL;
    return INCALLER_OK;
}


static int incaller_remote_query(RemoteQueryInfo* incaller, char* name)
{
    dbi_inst pInst= NULL;
    dbi_conn dbi_connection = NULL;
    char* remote_name= NULL;
    char sql[SQL_SIZE] = { 0 };

    dbi_connection = incaller_dbi_connect_db(incaller, &pInst);
    if (dbi_connection == NULL)
    {
        ast_log(LOG_ERROR, "Connect ucm_config DB failed\n");
        return INCALLER_ERR;
    }

    snprintf(sql, sizeof(sql)-1, "select %s from %s where %s like '%d';", incaller->remote_name, incaller->remote_table, incaller->remote_number, incaller->phone_number);
    ast_log(LOG_ERROR, "cmd is %s\n", sql);

    remote_name = incaller_dbi_getOneVarcharField(dbi_connection,sql,incaller->remote_name);
    if(remote_name == NULL)
    {
        ast_log(LOG_ERROR, "sql get data is null");
        incaller_dbi_disconnect_db(dbi_connection, &pInst);
        return INCALLER_ERR;
    }

    ast_log(LOG_ERROR, "sql get name: %s", remote_name);
    strncpy(name, remote_name, INCALLER_DEFAULT_LEN - 1);
    name[INCALLER_DEFAULT_LEN - 1] = '\0';
    free(remote_name);
    remote_name = NULL;

    incaller_dbi_disconnect_db(dbi_connection, &pInst);
    return INCALLER_OK;
}

#endif



#if DESC("mian process")
static int reincallerquery_exec(struct ast_channel *chan, const char* data) 
{
    int res = 0;
    int in_phone = 0;
    RemoteQueryInfo* incaller_info = NULL;
    struct ast_module_user *u = NULL;
    char name[INCALLER_DEFAULT_LEN] = { 0 };

    u = ast_module_user_add(chan);
    res = arg_parser(data, &in_phone);
    if(res == INCALLER_ERR)
    {
        goto out;
    }

    res = incaller_create_info(&incaller_info, in_phone);
    if(res == INCALLER_ERR)
    {
        free(incaller_info);
        goto out;
    }

    res = incaller_get_args(incaller_info);
    if(res == INCALLER_ERR)
    {
        free(incaller_info);
        goto out;
    }

    res = incaller_remote_query(incaller_info, name);
    if(res == INCALLER_ERR)
    {
        free(incaller_info);
        goto out;
    }

    pbx_builtin_setvar_helper(chan, "CALLERID(name)", name);
    pbx_builtin_setvar_helper(chan, "IS_REPLACE", "yes");
    free(incaller_info);

out:
    ast_module_user_remove(u);
    return INCALLER_OK;
} 

#endif

static int load_module(void)
{
    return ast_register_application(app, reincallerquery_exec, synopsis, tdesc);
}
static int unload_module(void)
{
    int res;
    res = ast_unregister_application(app);
    ast_module_user_hangup_all();
    return res;
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "MY OWN APPS");



