/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, Grandstream Networks.
 *
 * Yang Yanfei <yfyang@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
  	<depend>ucmsecret</depend>
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/


#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjsip_ua.h>

#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/app.h"
#include "asterisk/dsp.h"
#include "asterisk/file.h"
#include "asterisk/astdb.h"
#include "asterisk/channel.h"
#include "asterisk/manager.h"
#include "asterisk/cli.h"
#include "asterisk/utils.h"
#include "asterisk/translate.h"
#include "asterisk/paths.h"
#include "asterisk/dial.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/musiconhold.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/res_pjsip_sca.h"
#include "asterisk/callerid.h"
#include "asterisk/channel.h"
#include "asterisk/format_cap.h"
#include "asterisk/strings.h"
#include "asterisk/format_compatibility.h"
//#include "asterisk/haserver.h"
#include "asterisk/say.h"
#include "asterisk/bridge.h"
#include "asterisk/bridge_features.h"
#include "asterisk/avs_controller.h"
#include "asterisk/indications.h"
#include "asterisk/core_local.h"
#include "ucm_crypt.h"

AST_MODULE_LOG("sca");

/*** DOCUMENTATION
	<application name="Sca" language="en_US">
		<synopsis>
			Call to Shared Call Appearance.
		</synopsis>
		<syntax>
			<parameter name="shardline" required="true">
				<para>the name of a shared line</para>
			</parameter>
		</syntax>
		<description>
			<para>This application should be executed by an SCA group. </para>
			<para>On exit, this application will set the variable <variable>SCA_CALL_STATUS</variable> to
			one of the following values:</para>
			<variablelist>
				<variable name="SCA_CALL_STATUS">
					<value name="FAILURE" />
					<value name="CONGESTION" />
					<value name="SUCCESS" />
				</variable>
			</variablelist>
		</description>
	</application>

 ***/
#define AUTO_RECORD_FILE 	"auto_record.conf"
#define SCA_CONFIG_FILE		"sca.conf"
#define SCA_LINE_MAX_COUNT  20
#define SCA_FRAME_BITS      32
#define SCA_SIZE  320
#define SCA_DELAYDETECTTALK     300
#define SCA_DELAYDETECTENDTALK  1000
#define SCA_BARGE_WARN_FREQ     30

/*! each buffer is 20ms, so this is 640ms total */
#define SCA_DEFAULT_AUDIO_BUFFERS  32

#define SCA_AUTO_RECORD_OFF             0
#define SCA_AUTO_RECORD_INTERNAL        1
#define SCA_AUTO_RECORD_EXTERNAL        2
#define SCA_AUTO_RECORD_ALL             3

#define SCA_BRIDGE_USE_CONFBRIDGE       1

//static const pj_str_t callinfo_name = { "Call-Info", 9 };

static const char * const sca_app = "Sca";

static int sca_record_prompt = 0;  

struct sca_shared;
struct sca_user;
struct sca_line;
struct sca_bridge_user;

static struct ao2_container *con_sca_users;  /* the sca_user objects store in the container*/
static struct ao2_container *con_sca_shareds;   /* the sca_shared objects store in the container*/

static struct ast_sched_context *sca_sched;

enum sca_barge_warn {
	Barge_Warn_None = 0,
	Barge_Warn_Once,
	Barge_Warn_Repeat,
};

enum {
	/*! If set, the bridge will be recorded */
	FLAG_RECORD = (1 << 0),
	/*! If set, the channel will leave the bridge if all marked users leave */
	FLAG_MARKEDEXIT = (1 << 1),
	/*! If set, the user will be marked */
	FLAG_MARKEDUSER = (1 << 2),
	/*! If set, play the rbt */
	FLAG_MOH_RBT = (1 << 3),
	/*! If set, unhold the sca bridge */
	FLAG_UNHOLD = (1 << 4),
	/*! If set, barge in the sca bridge */
	FLAG_BARGE = (1 << 5),
	/*! If set, the sca bridge has a duration limit */
	FLAG_DUR_LIMIT = (1 << 6),
};

struct sca_user_ref {
    struct sca_user        *user;
    struct sca_bridge_user *bridge_user;
	AST_RWLIST_ENTRY(sca_user_ref) ref_entry;
};

struct sca_user {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
	);

	struct sca_shared  *shared;
	struct sca_line    *line;
	struct ast_channel *peer_chan;    /* the peer channle */
	struct ast_channel *local_chan;   /* the local channle */
	struct ast_dial    *dial;         /* dial to the sca user */

	unsigned int is_private:1;
	unsigned int as_caller:1;
	unsigned int as_callee:1;
	unsigned int marked:1;

	AST_RWLIST_ENTRY(sca_user) entry;
};

struct sca_line {
	AST_DECLARE_STRING_FIELDS (
		AST_STRING_FIELD(local_info);   /* the local sca user*/
		AST_STRING_FIELD(remote_info);  /* the remote user of the call */
	);

	AST_LIST_HEAD(,sca_user_ref) user_refs;
	unsigned int index;                 /* the index of the line */
	unsigned int related_count;         /* Number of users currently related to the line */
	unsigned int active_count;          /* Number of active users currently on the line */
	struct sca_shared  *shared;         /* the related sharedline */
	struct ast_channel *remote_chan;    /* the remote channel when the call come to the shared line */
	struct sca_bridge_user *heldby_user;
	struct sca_bridge_user *heldby_normal_user;
	enum   sca_status   status;         /* the status of the line */
	int    record_flag;                 /* to indicate if the line need record */
	int    custom_record;               /* to indicate if the line in custom record */
	unsigned int   ringtime;            /* to indicate the ringing time, record as seconds */
	AST_RWLIST_ENTRY(sca_line) entry;
};

struct sca_shared {
	AST_DECLARE_STRING_FIELDS (
		AST_STRING_FIELD(name);
	);

	AST_LIST_HEAD(,sca_user) users;
	AST_LIST_HEAD(,sca_line) lines;

	unsigned int num_users;     /* Number of users that used the shared line */
	unsigned int active_users;  /* Number of users currently on a call that used the shared line */
	unsigned int hold_lines;    /* Number of lines that currently on hold. */
	unsigned int retrv_remote_call:1;  /* whether allow retrieve a call from another location */
	unsigned int paging_alert_all:1;   /* whether alert all locations for paging calls */
	unsigned int mca:1;    /* multiple call arrangement */
	unsigned int barge:1;  /* whether allow bridging between locations */
	unsigned int marked:1;

	/*
	 *  auto_record:
	 *  0 - off
	 *  1 - internal
	 *  2 - external
	 *  3 - all
	 */
	unsigned int auto_record:2;
	enum sca_barge_warn  warn;
};


/*! \brief A sca user that is ringing */
struct sca_ringing_user {
	struct sca_user   *user;
	AST_LIST_ENTRY(sca_ringing_user) entry;
};


/*! \brief Event types that can be queued up for the SCA thread */
enum sca_event_type {
	SCA_EVENT_HOLD,         /*! A user has put the call on hold */
	SCA_EVENT_DIAL_STATE,   /*! The state of a dial has changed */
	SCA_EVENT_RINGING_USER, /*! The state of a ringing user has changed */
    SCA_EVENT_HANGUP_RINGING_USERS,  /*! Indicate to hangup ringing users */
};

enum sca_record_type {
	SCA_RECORD_EXTENSION,
	SCA_RECORD_TRUNK,
	SCA_RECORD_ROUTE,
	SCA_RECORD_IVR,
};

struct sca_event {
	enum sca_event_type type;
	struct sca_shared  *shared;
	struct sca_line    *line;

	AST_LIST_ENTRY(sca_event) entry;
};


struct run_user_args {
	struct sca_user *user;
	ast_mutex_t *cond_lock;
	ast_cond_t *cond;
};

/*!
 * \brief A structure for data used by the sca thread
 */
static struct {
	pthread_t thread;   /*! The SCA thread ID */
	ast_cond_t  cond;
	ast_mutex_t lock;

	AST_LIST_HEAD(, sca_ringing_user) ringing_users;
	AST_LIST_HEAD_NOLOCK(, sca_event) event_q;
	unsigned int ringing_users_num;
	unsigned int event_q_num;
    unsigned int stop:1;
} sca = {
	.thread = AST_PTHREADT_NULL,
};


/*! \brief The bridge User object */
struct sca_bridge_user {
	int user_no;                            /*!< User Number */
	int sched_id;                           /*!< sched id for play tone */
	struct ast_channel *chan;               /*!< Connected channel */
	struct sca_bridge  *sbridge;
	struct sca_line    *line; 
	unsigned int  marked:1;                 /*!< to indicate whether it is a marked user or not */
	unsigned int  playing_moh:1;            /*!< MOH is currently being played to the user */
	unsigned int  destroy:1;                /*!< to indicate whether the user need to be destroy */
	AST_LIST_ENTRY(sca_bridge_user) list;
};


/*! \brief The SCA bridge object */
struct sca_bridge {
	char name[80]; 
	char uniqueid[128];
    struct ast_bridge  *bridge;
	struct ast_channel *playback_chan;      /*!< Announcements channel */
	struct ast_channel *record_chan;        /*!< Listen/Record channel */
	int user_count;                         /*!< Number of active users */
	int markedusers;                        /*!< Number of marked users */
	int refcount;                           /*!< reference count of usage */
	int duration;                           /*!< the duration limit time (seconds) */
	int limit_warning;                      /*!< Duration into the call when warnings should begin (seconds) */
	int limit_freq;                         /*!< Interval between the warnings (seconds) */
	int exit_flag;                          /*!< to indicate if the bridge need to exit */
	long endtime;                           /*!< When to end the bridge if scheduled */
	char *recordingfilename;                /*!< Filename to record the Conference into */
	char *recordingfilename_cdr;			/*!< Filename to be saved in cdr's recording files name */
	char *recordingfileuuid;
	char *args;
	ast_mutex_t lock;
	struct ao2_container *usercontainer;
	AST_LIST_ENTRY(sca_bridge) list;
};

struct auto_record_node {
	int type:4;
	char extensions[AST_MAX_EXTENSION];
	char enable[20];

	AST_LIST_ENTRY(auto_record_node) entry;
};

static AST_LIST_HEAD_STATIC(record_lists, auto_record_node);
static AST_LIST_HEAD_STATIC(sca_bridges, sca_bridge);

AST_RWLIST_HEAD_STATIC(line_seize_handlers, ast_sip_line_seize_handler);
AST_RWLIST_HEAD_STATIC(call_info_handlers, ast_sip_call_info_handler);

static const char *sca_state2str(enum sca_status state);
static void  sca_dial_state_callback(struct ast_dial *dial);

static void sca_queue_event_full(enum sca_event_type type, struct sca_shared *shared,  struct sca_line *line, int lock);
static int sca_dispose_bridge(struct sca_bridge *bridge);
static int sca_bridge_run(struct ast_channel *chan, struct sca_bridge *bridge, struct sca_line *line, struct sca_user *sca_user, struct ast_flags64 *flags);
static struct sca_bridge *sca_build_bridge(const char *name, int refcount, const struct ast_channel *chan);
static void sca_hangup_users(struct sca_shared *shared, struct sca_line *line, int flag);
static int sca_update_local_user(struct sca_shared *shared, struct sca_line *line);
static int sca_ring_users(struct sca_shared *shared, struct sca_line *line);
static int sca_call_info_terminated(char* sharedline, char* exten);
static void sca_remove_ringing_user(struct sca_user *user);
static int sca_bridge_stop_record(struct sca_bridge *bridge);
static void sca_moh_stop(struct sca_bridge_user *user);
static int sca_play_barge_warn_timeout(const void *data);

#define ____SCA_OBJECT_MANAGEMENT_____

static int sca_shared_hash(const void *obj, const int flags)
{
	const struct sca_shared *shared = obj;

	return ast_str_case_hash(shared->name);
}

static int sca_shared_cmp(void *obj, void *arg, int flags)
{
	struct sca_shared *shared = obj, *shared2 = arg;

	return !strcasecmp(shared->name, shared2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static int sca_user_hash(const void *obj, const int flags)
{
	const struct sca_user *user = obj;

	return ast_str_case_hash(user->name);
}

static int sca_user_cmp(void *obj, void *arg, int flags)
{
	struct sca_user *user = obj, *user2 = arg;

	return !strcasecmp(user->name, user2->name) ? CMP_MATCH | CMP_STOP : 0;
}

/*!
 * \internal
 * \brief Find an SCA shared by name
 */
static struct sca_shared *sca_find_shared(const char *name)
{
	struct sca_shared tmp_shared = {
		.name = name,
	};

	if (!name) {
		return NULL;
	}

	return ao2_find(con_sca_shareds, &tmp_shared, OBJ_POINTER); 
}

/*!
 * \internal
 * \brief Find an SCA user by name
 */
static struct sca_user *sca_find_user(const char *name)
{
	struct sca_user tmp_user = {
		.name = name,
	};

	return ao2_find(con_sca_users, &tmp_user, OBJ_POINTER);
}

/*! release the related list in the sca shared object */
static int sca_shared_release_list(void *obj, void *arg, int flags)
{
	struct sca_shared  *shared = obj;
	struct sca_user    *user;
	struct sca_line    *line;

	while ((user = AST_LIST_REMOVE_HEAD(&shared->users, entry))) {
		ao2_ref(user, -1);
	}

	while ((line = AST_LIST_REMOVE_HEAD(&shared->lines, entry))) {
		ao2_ref(line, -1);
	}

	return 0;
}

static void sca_shared_destructor(void *obj)
{
	struct sca_shared *shared = obj;

	ast_log(LOG_DEBUG, "sca_shared destructor for [%s][%p]\n", shared->name, shared);

	sca_shared_release_list(shared, NULL, 0);

	AST_LIST_HEAD_DESTROY(&shared->lines);
	ast_string_field_free_memory(shared);
}


static void sca_user_destructor(void *obj)
{
	struct sca_user *user = obj;

	ast_log(LOG_NOTICE, "sca_user destructor for [%s][%p]\n", user->name, user);

	if (user->is_private) {
		//to terminated call-info subscription
		sca_call_info_terminated((char *)user->shared->name, (char *)user->name);
	}

    ao2_ref(user->shared, -1);
	ast_string_field_free_memory(user);
}

static void sca_line_destructor(void *obj)
{
	struct sca_line *line = obj;
	struct sca_user_ref  *user = NULL;

	ast_log(LOG_NOTICE, "sca_line_destructor for sharedline[%s], line[%p], index[%d]\n", line->shared->name, line, line->index);

	while ((user = AST_LIST_REMOVE_HEAD(&line->user_refs, ref_entry))) {
		ao2_ref(user, -1);
	}

	ao2_ref(line->shared, -1);

	if (line->remote_chan) {
		ao2_ref(line->remote_chan, -1);
	}

	line->shared = NULL;
	line->remote_chan = NULL;
	ast_string_field_free_memory(line);
}


static void sca_user_ref_destructor(void *obj)
{
	struct sca_user_ref *user = obj;

	ast_log(LOG_NOTICE, "sca_user_ref destructor for [%p]\n", user);

	ao2_ref(user->user, -1);

	if (user->bridge_user) {
		ao2_ref(user->bridge_user, -1);
	}
}


static void sca_bridge_user_destructor(void *obj)
{
	struct sca_bridge_user *user = obj;

	if (user->sched_id != -1) {
		AST_SCHED_DEL(sca_sched, user->sched_id);
		user->sched_id = -1;
	}

	user->destroy = 1;
	ast_log(LOG_NOTICE, "sca_bridge_user_destructor, user[%d]", user->user_no); 

	return;
}

static struct sca_line *sca_line_get(struct sca_shared *shared, unsigned int index)
{
	struct sca_line *line;

	if (!shared || index <= 0) {
		return NULL;
	}

	/* get the line from the list */
	AST_LIST_LOCK(&shared->lines);
	AST_LIST_TRAVERSE(&shared->lines, line, entry) {
		if (line->index == index) {
			ao2_ref(line, 1);
			AST_LIST_UNLOCK(&shared->lines);
			return line;
		}
	}
	AST_LIST_UNLOCK(&shared->lines);

    return NULL;
}


/*!
 * \internal
 * \brief Mark a sca shared object before reload, clean the mark flag if object data exist in the configure file;
 *        otherwise, destroy the object
 */
static int sca_shared_mark(void *obj, void *arg, int flags)
{
	struct sca_shared  *shared = obj;
	struct sca_user *user;

	ao2_lock(shared);

	shared->marked = 1;

	AST_LIST_TRAVERSE(&shared->users, user, entry) {
		user->marked = 1;
	}

	ao2_unlock(shared);

	return 0;
}

/*!
 * \internal
 * \brief Mark a sca user object before reload, clean the mark flag if object data exist in the configure file;
 *        otherwise, destroy the object
 */
static int sca_user_mark(void *obj, void *arg, int flags)
{
	struct sca_user *user = obj;

	ao2_lock(user);
	user->marked = 1;
	ao2_unlock(user);

	return 0;
}

/*! to check the mark flag, marked indicates the object need to be destroyed */
static int sca_shared_is_marked(void *obj, void *arg, int flags)
{
	struct sca_shared  *shared = obj;
	struct sca_user    *user   = NULL;

	ao2_lock(shared);

	if (shared->marked) {
		/* Only remove all of the user/line references if the shared itself is going away */
		sca_shared_release_list(shared, NULL, 0);
	} else {
		/* Otherwise only remove references to users no longer in the config */
		AST_LIST_TRAVERSE_SAFE_BEGIN(&shared->users, user, entry) {
			if (!user->marked) {
				continue;
			}
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_ref(user, -1);
		}
		AST_LIST_TRAVERSE_SAFE_END
	}

	ao2_unlock(shared);

	return shared->marked ? CMP_MATCH : 0;
}

static int sca_user_is_marked(void *obj, void *arg, int flags)
{
	struct sca_user *user = obj;

	return user->marked ? CMP_MATCH : 0;
}


static int sca_bridge_user_set_exit(void *obj, void *check_admin_arg, int flags)
{
	struct sca_bridge_user *user = obj;

	if (user->playing_moh) {
		sca_moh_stop(user);
	}

    ast_log(LOG_NOTICE, "sca_bridge_user_set_exit, user->chan[%s]", ast_channel_name(user->chan));
    ast_bridge_remove(user->sbridge->bridge, user->chan);

    return 0;
}

static int sca_bridge_user_cmp(void *obj, void *arg, int flags)
{
	struct sca_bridge_user *user = obj;
	int *user_no = arg;

	if (user->user_no == *user_no) {
		return (CMP_MATCH | CMP_STOP);
	}

	return 0;
}

static void sca_related_shared_to_user(struct sca_user *user, char *sharedline)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
    struct sca_user *user0 = NULL;
	int existing_user = 0;

	if (!user || !sharedline) {
		return;
	}

	shared = sca_find_shared(sharedline); 
	if (!shared) {
		ast_log(LOG_ERROR, "Sharedline '%s' not found!\n", sharedline);
		return;
	}

	AST_LIST_TRAVERSE(&shared->users, user0, entry) {
		if (user == user0) {
			user0->marked = 0;
			existing_user = 1;
			break;
		}
	}

	if (user->shared) {
		if (user->shared != shared) {
			ao2_ref(user->shared, -1);
			user->shared = ao2_bump(shared);
		}
	} else {
        user->shared = ao2_bump(shared);
    }

	if (!existing_user) {
		ao2_lock(shared);
		AST_LIST_INSERT_TAIL(&shared->users, user, entry);
		ao2_bump(user);
		ao2_unlock(shared);

        ast_log(LOG_DEBUG, "Sharedline[%s][%p] added priv[%s][%p]!\n", shared->name, shared, user->name, user);
	}
}

static int sca_build_shared_user(struct sca_shared *shared)
{
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);

    if (!shared) {
        return -1;
    }

	user = ao2_alloc(sizeof(*user), sca_user_destructor);
    if (!user) {
        return -1;
    }

    if (ast_string_field_init(user, 32)) {
		return -1;
	}

	ast_string_field_set(user, name, shared->name);
	user->as_callee = 1;
	user->as_caller = 1;
	user->is_private = 0;
    user->line = NULL;
    user->local_chan = NULL;
    user->peer_chan  = NULL;
    user->dial = NULL;
	user->shared = ao2_bump(shared);

	AST_LIST_INSERT_TAIL(&shared->users, user, entry);
	ao2_bump(user);
	ao2_link(con_sca_users, user);

	ast_log(LOG_DEBUG, "build shared user[%s][%p]!\n", user->name, user);

    return 0;
}

static int sca_build_shared(struct ast_config *cfg, const char *cat)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	struct ast_variable *var;
	int    existing_shared = 0;
	struct sca_user *user;

	if ((shared = sca_find_shared(cat))) {
		shared->marked = 0;
		existing_shared = 1;

		user = sca_find_user(shared->name);
		if (user) {
			user->marked = 0;
			ao2_ref(user, -1);
		}
	} else if ((shared = ao2_alloc(sizeof(*shared), sca_shared_destructor))) {
		if (ast_string_field_init(shared, 32)) {
			return -1;
		}
		ast_string_field_set(shared, name, cat);
	} else {
		return -1;
	}

	ao2_lock(shared);

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "retrv_remote_call")) {
			shared->retrv_remote_call = ast_true(var->value);
		} else if (!strcasecmp(var->name, "paging_alert_all")) {
			shared->paging_alert_all = ast_true(var->value);
		} else if (!strcasecmp(var->name, "mca")){
			shared->mca = ast_true(var->value);
		} else if (!strcasecmp(var->name, "auto_record")){
			if(!strcasecmp(var->value, "off") || !strcasecmp(var->value, "no")) {
				shared->auto_record = SCA_AUTO_RECORD_OFF;
			} else if (!strcasecmp(var->value, "internal")) {
				shared->auto_record = SCA_AUTO_RECORD_INTERNAL;
			} else if (!strcasecmp(var->value, "external")) {
				shared->auto_record = SCA_AUTO_RECORD_EXTERNAL;
			} else if (!strcasecmp(var->value, "all")) {
				shared->auto_record = SCA_AUTO_RECORD_ALL;
			} else {
				ast_log(LOG_WARNING, "Invalid config: auto_record : %s .\n", var->value);
				shared->auto_record = SCA_AUTO_RECORD_OFF;
			}
		} else if (!strcasecmp(var->name, "barge")){
			shared->barge = ast_true(var->value);
		} else if (!strcasecmp(var->name, "barge_warn")) {
			if (!strcasecmp(var->value, "once")) {
				shared->warn = Barge_Warn_Once;
			} else if (!strcasecmp(var->value, "repeat")) {
				shared->warn = Barge_Warn_Repeat;
			} else {
				shared->warn = Barge_Warn_None;
			}
		}
	}

	if (!existing_shared) {
		AST_LIST_HEAD_INIT(&shared->lines);
		AST_LIST_HEAD_INIT(&shared->users);

		ao2_link(con_sca_shareds, shared);
		ast_log(LOG_NOTICE, "build sharedline[%s][%p]!\n", shared->name, shared);

		sca_build_shared_user(shared);
	} 

	ao2_unlock(shared);
	return 0;
}

static int sca_build_user(struct ast_config *cfg, const char *cat)
{
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);
	struct ast_variable *var;
	int existing_user = 0;

	if ((user = sca_find_user(cat))) {
		user->marked = 0;
		existing_user = 1;
	} else if ((user = ao2_alloc(sizeof(*user), sca_user_destructor))) {
		if (ast_string_field_init(user, 32)) {
			return -1;
		}
		ast_string_field_set(user, name, cat);
	} else {
		return -1;
	}

	ao2_lock(user);

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "shared")) {
			ao2_unlock(user);
			sca_related_shared_to_user(user, (char *)var->value);
			ao2_lock(user);
		} else if (!strcasecmp(var->name, "as_caller")) {
			user->as_caller = ast_true(var->value);
		} else if (!strcasecmp(var->name, "as_callee")) {
			user->as_callee = ast_true(var->value);
		}
	}

	user->is_private = 1;
	ao2_unlock(user);

	if (!existing_user) {
		ao2_link(con_sca_users, user);
        ast_log(LOG_DEBUG, "build user[%s][%p]!\n", user->name, user);
	}

	return 0;
}

static int sca_seize_terminated(char* sharedline, int line_index)
{
	struct ast_sip_line_seize_handler *iter;

	SCOPED_LOCK(lock, &line_seize_handlers, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&line_seize_handlers, iter, next) {
		if (0 == iter->seize_terminated(sharedline, line_index)) {
			return 0;
		}
	}
	return -1;
}

static int sca_change_line_state(struct sca_shared *shared, struct sca_line *line, enum sca_status status)
{
	enum ast_device_state  dev_state = AST_DEVICE_NOT_INUSE;
	struct sca_line  *line0 = NULL;
	enum   sca_status old_status;

	if (!shared || !line || status >= SCA_BUTT) {
		return -1;
	}

	ao2_lock(line);
	old_status = line->status;

	ast_log(LOG_NOTICE, "Sharedline[%s], line[%d], state [%s] ==> [%s]!\n", 
	        shared->name, line->index, sca_state2str(old_status), sca_state2str(status));

	if (old_status == status) {
		ao2_unlock(line);
		return 0;
	}

	line->status = status;
	ao2_unlock(line);

	if ((old_status == SCA_Seized) && (status == SCA_Progressing)) {
		sca_seize_terminated((char *)shared->name, line->index);
	}

	if (status != SCA_Idle) {
		dev_state = AST_DEVICE_INUSE;
	}

	AST_LIST_LOCK(&shared->lines);

	/* if idle, release the line */
	if (status == SCA_Idle) {
        char   bridge_name[80]    = {0};
        struct sca_bridge *bridge = NULL;

		snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", shared->name, line->index);
		AST_LIST_UNLOCK(&shared->lines);
		AST_LIST_LOCK(&sca_bridges);
		AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
			if (!strcmp(bridge_name, bridge->name)) {
				if (status == SCA_Idle) {
					ast_log(LOG_NOTICE, "line[%d] is IDLE, set exit to bridge[%s]", line->index, bridge_name);
					bridge->exit_flag = 1;

					ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
				}
                break; 
			}
		}
		AST_LIST_UNLOCK(&sca_bridges);
		AST_LIST_LOCK(&shared->lines);

		AST_LIST_TRAVERSE_SAFE_BEGIN(&shared->lines, line0, entry) {
			if (line0 != line) {
				continue;
			}
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_ref(line0, -1);
		}

		AST_LIST_TRAVERSE_SAFE_END
    }

	/* manage event report*/
	/* first, clean the line status in the sharedline */
	manager_event(EVENT_FLAG_CALL, "ScaLineStatusClean", "sharedline: %s\r\n", shared->name);

    AST_LIST_TRAVERSE(&shared->lines, line0, entry) {
        if (line0) {
            manager_event(EVENT_FLAG_CALL, "ScaLineStatus", "sharedline: %s\r\n"
                          "line_index: %d\r\n"
                          "sca_status: %s\r\n"
                          "local_info: %s\r\n"
                          "remote_info: %s\r\n", 
                          shared->name, line0->index, sca_state2str(line0->status),
                          (ast_strlen_zero(line0->local_info) ? "-" : line0->local_info),
                          (ast_strlen_zero(line0->remote_info) ? "-" : line0->remote_info));
        }
    }

	AST_LIST_UNLOCK(&shared->lines);

	/* report dev state changed, simple indicate the change, call-info module will retrieve the detailed information later */
	ast_devstate_changed(dev_state, AST_DEVSTATE_NOT_CACHABLE, "SCA:%s", shared->name);
	ast_devstate_changed(dev_state, AST_DEVSTATE_NOT_CACHABLE, "PJSIP/%s", shared->name);

    return 0;
}

static int sca_get_channel_user_name(struct ast_channel *chan, char *name, size_t buf_len)
{
	char *dash = NULL;
	char *slash = NULL;
	char  chan_name[AST_CHANNEL_NAME] = {0};

	if (!chan || !name || !buf_len) {
		return -1;
	}

	ast_copy_string(chan_name, ast_channel_name(chan), AST_CHANNEL_NAME); 
	if ((dash = strrchr(chan_name, '-'))) {
		*dash = '\0';
	}

	if ((slash = strchr(chan_name, '/'))) {
		ast_copy_string(name, slash + 1, buf_len); 
	}

	return 0;
}


static enum ast_device_state sca_state(const char *data)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	enum ast_device_state res = AST_DEVICE_INVALID;
	char *sharedline;

	sharedline = ast_strdupa(data);

	shared = sca_find_shared(sharedline);
	if (shared) {
		if (AST_LIST_EMPTY(&shared->lines)) {
			res = AST_DEVICE_NOT_INUSE;
		} else {
			res = AST_DEVICE_INUSE;
		}
	}

	if (res == AST_DEVICE_INVALID) {
		ast_log(LOG_ERROR, "Could not determine state for sharedline %s!\n", sharedline);
	}

	return res;
}

#define ____SCA_SESSION_HANDLER____

static enum sca_status sca_line_status_get(char *sharedline, int line_index, char *local_info)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
    struct sca_line   *line   = NULL;
    enum sca_status    status = SCA_Idle;

    if (!sharedline || line_index <= 0) {
        return SCA_BUTT; 
    }

    shared = sca_find_shared(sharedline);
    if (!shared) {
        ast_log(LOG_WARNING, "sca_line_status_get, can't find sharedline by %s", sharedline);
        return SCA_BUTT; 
    }

	AST_LIST_LOCK(&shared->lines);
    AST_LIST_TRAVERSE(&shared->lines, line, entry) {
        if (line->index == line_index) {
            status = line->status;
            break;
        }
    }
	AST_LIST_UNLOCK(&shared->lines);

    return status; 
}

/*set the status of the line in the sharedline */
static int sca_line_status_set(char *sharedline, int line_index, enum sca_status status)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);
	enum   sca_status  old_status = SCA_BUTT;
	struct sca_bridge *bridge = NULL;
	char   bridge_name[80] = {0};
	int    bridge_exist = 0;
	int    matched = 0;

	if (!sharedline || (status >= SCA_BUTT) || (line_index <= 0)) {
		return -1; 
	}

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_status_set, can't find sharedline by %s", sharedline);
		return -1; 
	}

	line = sca_line_get(shared, line_index);
	if (!line) {
		ast_log(LOG_NOTICE, "sca_line_status_set, can't find line by index[%d]", line_index);
		return -1; 
	}

	ao2_lock(line);
	old_status = line->status;
	if (old_status == status) {
		ao2_unlock(line);
		return 0;
	}
	ao2_unlock(line);

	/* check if the bridge exists */
	if ((SCA_Held == status) || (SCA_Held_Private == status) || (SCA_Bridge_Active == status) || (SCA_Bridge_Held== status)) {
		snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", sharedline, line_index);
		AST_LIST_LOCK(&sca_bridges);
		AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
			if (!strcmp(bridge_name, bridge->name)) {
				bridge_exist = 1;
				break;
			}
		}
		AST_LIST_UNLOCK(&sca_bridges);
	}

	switch (status) {
	case SCA_Idle:
		matched = 1;  /* all state can change to idle */
		break;

	case SCA_Seized:
		if (old_status == SCA_Idle) {
			matched = 1; 
		}
		break;

	case SCA_Progressing:
		if (old_status == SCA_Idle || old_status == SCA_Seized) {
			matched = 1; 
		}
		break;

	case SCA_Alerting:
		if (old_status == SCA_Idle) {
			matched = 1; 
		}
		break;

	case SCA_Active:
		if (old_status == SCA_Progressing || 
		    old_status == SCA_Alerting || 
		    old_status == SCA_Held ||
		    old_status == SCA_Held_Private ||
		    old_status == SCA_Bridge_Active ) {
			matched = 1; 
		}
		break;

	case SCA_Held:
		if ((old_status == SCA_Active) && bridge_exist) {
			matched = 1; 
		}
		break;

	case SCA_Held_Private:
		if ((old_status == SCA_Active) && bridge_exist) {
			matched = 1; 
		}
		break;

	case SCA_Bridge_Active:
		if ((old_status == SCA_Active || old_status == SCA_Bridge_Held) && bridge_exist) {
			matched = 1; 
		}
		break;

	case SCA_Bridge_Held:
		if ((old_status == SCA_Bridge_Active) && bridge_exist) {
			matched = 1; 
		}
		break;

	default:
		break;
	}

	if (!matched) {
		ast_log(LOG_NOTICE, "sca_line_status_set, index[%d], old state[%s] can't change to [%s]", 
		        line_index, sca_state2str(old_status), sca_state2str(status));
		return -1;
	}

	/* do the change */
	return sca_change_line_state(shared, line, status);
}


static int sca_line_alloc(char *sharedline, char *local_info, struct ast_channel *remote_chan)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
    struct sca_line *line = NULL;
    char remote[80] = {0};
    int line_arr[SCA_LINE_MAX_COUNT+1] = {0};
	int line_index = 0;
    int line_count = 0;
    int i;

	if (!sharedline || !local_info) {
		return -1;
	}

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_alloc, can't find sharedline by %s", sharedline);
		return -1;
	}

    /* check the line array */
	AST_LIST_LOCK(&shared->lines);
	AST_LIST_TRAVERSE(&shared->lines, line, entry) {
		if (line->index <= SCA_LINE_MAX_COUNT) {
			line_arr[line->index] = 1;
		}
		line_count++;
	}
	AST_LIST_UNLOCK(&shared->lines);

	if (line_count > 0) {
		if (!shared->mca) {
			ast_log(LOG_WARNING, "sca_line_alloc, sharedline[%s], don't support MCA", sharedline);
			return -1;
		}
	}

	/* get an available index */
	for (i = 1; i < SCA_LINE_MAX_COUNT+1; i++) {
		if (line_arr[i] == 0) {
			line_index = i;
			break;
		}
	}

	if (!line_index) {
		ast_log(LOG_WARNING, "sca_line_alloc, sharedline[%s], all lines have been used", sharedline);
		return -1; 
	}

	line = ao2_alloc(sizeof(*line), sca_line_destructor);
	if (!line) {
		return -1;
	}

	AST_LIST_HEAD_INIT(&line->user_refs);
	if (ast_string_field_init(line, 256)) {
		ao2_ref(line, -1);
		return -1;
	}

	if (remote_chan) {
		const char *real_cid_num = pbx_builtin_getvar_helper(remote_chan, "REALCALLERIDNUM");
		if (strstr(ast_channel_name(remote_chan), "/trunk")
			&& !ast_strlen_zero(real_cid_num)) {
			strncpy(remote, real_cid_num, sizeof(remote)-1);
		}

		if (ast_strlen_zero(remote)) {
			if (ast_channel_caller(remote_chan)->ani.number.valid 
				&& ast_channel_caller(remote_chan)->ani.number.str
				&& strcmp(ast_channel_caller(remote_chan)->ani.number.str, "s")
				&& strcmp(ast_channel_caller(remote_chan)->ani.number.str, local_info)) {
				strncpy(remote, ast_channel_caller(remote_chan)->ani.number.str, sizeof(remote)-1);
			} else {
				if (ast_channel_caller(remote_chan)->id.number.valid
					&& ast_channel_caller(remote_chan)->id.number.str) {
					strncpy(remote, ast_channel_caller(remote_chan)->id.number.str, sizeof(remote)-1);
				}
			}
		}

		if (strlen(remote)) {
			ast_string_field_set(line, remote_info, remote);
			ast_log(LOG_NOTICE, "sca_line_alloc, sharedline[%s], local_info[%s], remote[%s]", sharedline, local_info, remote);
		}
	}

	ast_string_field_set(line, local_info, local_info);

	line->index = line_index;
	line->shared = ao2_bump(shared);
	line->remote_chan = ao2_bump(remote_chan);
	line->record_flag = 0;
	line->custom_record = 0;
	line->ringtime = 0;

	AST_LIST_LOCK(&shared->lines);
	AST_LIST_INSERT_TAIL(&shared->lines, line, entry);
	AST_LIST_UNLOCK(&shared->lines);

    /* Set the line status to seized */
	sca_change_line_state(shared, line, SCA_Alerting);

	ast_log(LOG_NOTICE, "sharedline[%s][%p], alloc line[%d][%p] to local[%s], remote[%s], remote_chan[%s]", 
	       sharedline, shared, line_index, line, local_info, remote, (remote_chan? ast_channel_name(remote_chan) : "NULL"));

	return line_index;
}

static int sca_line_service_get(char *sharedline, int *retrv_remote, int *barge)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);

    if (!sharedline || !retrv_remote || !barge) {
        return -1;
    }

    shared = sca_find_shared(sharedline);
    if (!shared) {
        ast_log(LOG_WARNING, "sca_line_service_get, can't find sharedline by %s", sharedline);
        return -1; 
    }

    *retrv_remote = shared->retrv_remote_call;
    *barge = shared->barge;

    return 0;
}

static unsigned int sca_line_index_get_by_status(char *sharedline, enum sca_status status, char *local_info)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	struct sca_line *line = NULL;

	if (!sharedline || status >= SCA_BUTT || status == SCA_Idle) {
		return 0;
	}

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_index_get_by_status, can't find sharedline by %s", sharedline);
		return 0; 
	}

	/* check the line array */
	AST_LIST_LOCK(&shared->lines);
	AST_LIST_TRAVERSE(&shared->lines, line, entry) {
		if (line->status == status) {
			if (local_info && strcmp(local_info, line->local_info)) {
				continue;
			}
			AST_LIST_UNLOCK(&shared->lines);
			return line->index; 
		}
	}
	AST_LIST_UNLOCK(&shared->lines);

	return 0;
}


static int sca_line_remove_user(struct ast_channel *chan, char *sharedline, int line_index, char *user_name)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);
	struct sca_user_ref  *user_ref  = NULL;
	struct sca_line   *line0  = NULL;
	int    user_exist = 1;
	int    sip_code = 0;

	if (!sharedline || !user_name || (line_index <= 0)) {
		return -1; 
	}

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_remove_user, can't find sharedline[%s]", sharedline);
		return -1; 
	}

	line = sca_line_get(shared, line_index);
	if (!line) {
		ast_log(LOG_NOTICE, "sca_line_remove_user, sharedline[%s], line index[%d] not exist", sharedline, line_index);
		return -1; 
	}

	user = sca_find_user(user_name);
	if (!user) {
		user_exist = 0;
		ast_log(LOG_WARNING, "sca_line_remove_user, can't find user by [%s]", user_name);
	} 

	if (chan) {
		sip_code = ast_channel_hangupsipcode(chan); 
	}

	if ((line->status == SCA_Active) &&
		 user_exist && user->dial && (ast_dial_state(user->dial) == AST_DIAL_RESULT_ANSWERED)) {
		char   bridge_name[80]    = {0};
		struct sca_bridge *bridge = NULL;

		snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", sharedline, line_index);
		AST_LIST_LOCK(&sca_bridges);
		AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
			if (!strcmp(bridge_name, bridge->name)) {
				if (bridge->user_count == 0) {
					ast_log(LOG_NOTICE, "no user in the bridge, line[%d] set exit!", line_index);

					if (user->dial) {
						ast_dial_join(user->dial); 
						ast_dial_destroy(user->dial);
					}

					user->local_chan = NULL;
					user->dial = NULL;

					if (line->remote_chan) {
						pbx_builtin_setvar_helper(line->remote_chan, "SCA_CALL_STATUS", "SUCCESS");
					}

					bridge->exit_flag = 1;
				}

				break; 
			}
		}
		AST_LIST_UNLOCK(&sca_bridges);
	}

	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
		if (!user_ref || strcmp(user_ref->user->name, user_name)) {
			continue;
		}

		if (!user_exist ) {
			if ((line0 = user_ref->user->line) && (ast_atomic_fetchadd_int((int *)&line0->related_count, 0) > 0)) {
				ast_atomic_fetchadd_int((int *)&line0->related_count, -1);
				ao2_ref(line0, -1); 
				user_ref->user->line = NULL;
			}
		}

        AST_LIST_REMOVE_CURRENT(ref_entry); 
		ao2_ref(user_ref, -1);

		break;
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&line->user_refs);

	if (user_exist) {
		ao2_lock(user);
		if (user->line) {
			ast_atomic_fetchadd_int((int *)&user->line->related_count, -1);
			ao2_ref(user->line, -1);
			user->line = NULL;
		}
		ao2_unlock(user);

		sca_remove_ringing_user(user);
	}

	sca_update_local_user(line->shared, line);

	AST_LIST_LOCK(&line->user_refs);
	if (AST_LIST_EMPTY(&line->user_refs)) {
		AST_LIST_UNLOCK(&line->user_refs);

		/* if no user in the line, line changed to IDLE */
		ast_log(LOG_NOTICE, "sca_line_remove_user, sharedline[%s], user[%s], sip_code[%d], set line[%d] to IDLE", sharedline, user_name, sip_code, line_index);
		if (line->remote_chan && (line->status == SCA_Alerting)) {
			if (ast_strlen_zero(pbx_builtin_getvar_helper(line->remote_chan, "SCA_CALL_STATUS"))) {
				if (sip_code == 408) {
					pbx_builtin_setvar_helper(line->remote_chan, "SCA_CALL_STATUS", "NOANSWER"); 
				} else {
					pbx_builtin_setvar_helper(line->remote_chan, "SCA_CALL_STATUS", "BUSY"); 
				}
			}

			pbx_builtin_setvar_helper(line->remote_chan, "SCA_DIAL_STATUS", "UNANSWERED");
		}

		return sca_change_line_state(shared, line, SCA_Idle);
	}
	AST_LIST_UNLOCK(&line->user_refs);

	return 0;
}


static int sca_line_set_user_info(char *sharedline, int line_index, char *local_user, char *remote_user)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);
	int    ischange = 0;

	if (!sharedline || (line_index <= 0) || (!local_user && !remote_user)) {
		return -1; 
	}

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_set_user_info, can't find sharedline[%s]", sharedline);
		return -1; 
	}

	line = sca_line_get(shared, line_index);
	if (!line) {
		ast_log(LOG_NOTICE, "sca_line_set_user_info, sharedline[%s], line index[%d] not exist", sharedline, line_index);
		return -1; 
	}

	if (local_user && strcmp(local_user, line->local_info)) {
		ast_string_field_set(line, local_info, local_user);
		ischange = 1;
	}

	if (remote_user && strcmp(remote_user, line->remote_info)) {
		ast_string_field_set(line, remote_info, remote_user);
		ischange = 1;
	}

	if (ischange) {
		manager_event(EVENT_FLAG_CALL, "ScaLineStatus", "sharedline: %s\r\n"
                      "line_index: %d\r\n"
                      "sca_status: %s\r\n"
                      "local_info: %s\r\n"
                      "remote_info: %s\r\n", 
                      shared->name, line->index, sca_state2str(line->status),
                      (ast_strlen_zero(line->local_info) ? "-" : line->local_info),
                      (ast_strlen_zero(line->remote_info) ? "-" : line->remote_info));
	}

	return 0; 
}


static int sca_line_get_private_users(char *sharedline, char *user_buf, int len, int paging_flag)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	struct sca_user   *user = NULL;
	int    str_len = 0;

	if (!sharedline || !user_buf || !len) {
		return -1; 
	}
    
	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_get_private_users, can't find sharedline[%s]", sharedline);
		return -1; 
	}

	if ((1 == paging_flag) && !shared->paging_alert_all) {
		return 0;
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&shared->users, user, entry) {
		if (user->is_private && user->as_callee ) {
			if ((2 == paging_flag) && user->line) {
				continue;
			}

			str_len = strlen(user_buf);
			if (!str_len) {
				snprintf(user_buf, len, "%s", user->name); 
			} else {
				snprintf(user_buf+str_len, len-str_len, ",%s", user->name); 
			}
		}
	}
	AST_LIST_TRAVERSE_SAFE_END

	return 0;
}


/* ret value means:
   -1: something wrong
   0 : idle  
   other value: inuse */
static int sca_user_check_status(char *sharedline)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	struct sca_line *line = NULL;
	int    alerting_flag  = 0;

	if (!sharedline) {
		return -1; 
	}
    
	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_NOTICE, "sca_user_check_status, can't find sharedline[%s]", sharedline);
		return -1; 
	}

	if (AST_LIST_EMPTY(&shared->lines)) {
		ast_log(LOG_NOTICE, "sca_user_check_status, sharedline[%s] is IDLE", sharedline);
		return 0;
	}

	AST_LIST_LOCK(&shared->lines);
	AST_LIST_TRAVERSE(&shared->lines, line, entry) {
		if (line->status == SCA_Alerting) {
			alerting_flag = 1;
			break;
		}
	}
	AST_LIST_UNLOCK(&shared->lines);

	if (alerting_flag) {
		return SCA_Alerting;
	}

	return 1;
}


static int sca_line_insert_user(char *sharedline, int line_index, char *local_user)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);
    struct sca_user_ref *user_ref = NULL;
	int    existed = 0;

	if (!sharedline || (line_index <= 0) || !local_user) {
		return -1; 
	}

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sca_line_insert_user, can't find sharedline[%s]", sharedline);
		return -1; 
	}

	user = sca_find_user(local_user);
	if (!user) {
		ast_log(LOG_WARNING, "sca_line_insert_user, can't find user by [%s]", local_user);
		return -1; 
	}

	line = sca_line_get(shared, line_index);
	if (!line) {
		ast_log(LOG_NOTICE, "sca_line_insert_user, sharedline[%s], line index[%d] not exist", sharedline, line_index);
		return -1; 
	}

	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
		if (!user_ref || strcmp(user_ref->user->name, local_user)) {
			continue;
		}

		existed = 1;
		break;
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&line->user_refs);

	if (existed) {
		return 0;
	}

    /* add user_ref */
	ao2_lock(user);
	if (!user->line) {
		user->line = ao2_bump(line);
		ast_atomic_fetchadd_int((int *)&line->related_count, 1);
		ao2_unlock(user);

		user_ref = ao2_alloc(sizeof(*user_ref), sca_user_ref_destructor);
		if (user_ref) {
			user_ref->user = ao2_bump(user);
			user_ref->bridge_user = NULL;

			AST_LIST_LOCK(&line->user_refs);
			AST_LIST_INSERT_TAIL(&line->user_refs, user_ref, ref_entry);
			AST_LIST_UNLOCK(&line->user_refs);

			sca_update_local_user(user->shared, line);

			ast_log(LOG_NOTICE, "sca_line_insert_user, insert user[%s] into line[%d]", local_user, line_index);
		}

		ao2_lock(user);
	}
	ao2_unlock(user);

	return 0;
}

static int sca_line_check_in_bridge(char *sharedline, int line_index)
{
	int    found = 0;
	char   bridge_name[80] = {0};
	struct sca_bridge  *bridge = NULL;

	if (!sharedline || (line_index <= 0)) {
		return 0; 
	}

	snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", sharedline, line_index);

	AST_LIST_LOCK(&sca_bridges);
	AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
		if (!strcmp(bridge_name, bridge->name)) {
			found = 1;
			break;
		}
	}
	AST_LIST_UNLOCK(&sca_bridges);

	return found;
}

struct ast_sip_session_sca_handler  sca_handler = {
	.sca_line_status_get = sca_line_status_get,
	.sca_line_status_set = sca_line_status_set,
	.sca_line_index_get  = sca_line_index_get_by_status,
	.sca_line_alloc = sca_line_alloc,
	.sca_line_service_get = sca_line_service_get,
	.sca_line_remove_user = sca_line_remove_user,
	.sca_line_set_user_info = sca_line_set_user_info,
	.sca_line_get_private_users = sca_line_get_private_users,
    .sca_user_check_status = sca_user_check_status,
    .sca_line_insert_user = sca_line_insert_user,
    .sca_line_check_in_bridge = sca_line_check_in_bridge,
};

#define ____SCA_API____

int ast_sip_line_check_and_seize(char* sharedline, char* local_info, int line_index)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);
	struct sca_user_ref *user_ref = NULL;
	struct sca_line     *line = NULL;
    int    has_seized = 0;
    int    line_count = 0;

    if (!sharedline || !local_info || line_index <= 0 ) {
        return -1; 
    }

    if (line_index > SCA_LINE_MAX_COUNT) {
        ast_log(LOG_WARNING, "ast_sip_line_check_and_seize, shareline[%s], line index[%d] mustn't bigger than 10", sharedline, line_index);
        return -1; 
    }

    /* get the sca shared */
    shared = sca_find_shared(sharedline);
    if (!shared) {
        ast_log(LOG_WARNING, "ast_sip_line_check_and_seize, can't find sharedline by [%s]", sharedline);
        return -1; 
    }

    /* check the right of the local info */
    user = sca_find_user(local_info);
    if (!user) {
        ast_log(LOG_WARNING, "ast_sip_line_check_and_seize, can't find user by [%s]", local_info);
        return -1; 
    }

    if (!user->as_caller) {
        ast_log(LOG_WARNING, "ast_sip_line_check_and_seize, the user [%s] can't act as caller", local_info);
        return -1; 
    }

    /* check whether the line has been used */
	AST_LIST_LOCK(&shared->lines);
    AST_LIST_TRAVERSE(&shared->lines, line, entry) {
        if (line->index == line_index) {
            has_seized = 1;
        }
        line_count++;
    }
	AST_LIST_UNLOCK(&shared->lines);

    if (has_seized) {
        ast_log(LOG_WARNING, "ast_sip_line_check_and_seize, sharedline[%s], index[%d] has been seized", sharedline, line_index);
        return -1; 
    }

    if (line_count > 0) {
        if (!shared->mca) {
            ast_log(LOG_WARNING, "ast_sip_line_check_and_seize, sharedline[%s], don't support MCA", sharedline);
            return -1; 
        }
    }

	line = ao2_alloc(sizeof(*line), sca_line_destructor); 
    if (!line) {
        return -1;
    }

	AST_LIST_HEAD_INIT(&line->user_refs);
	if (ast_string_field_init(line, 256)) {
		ao2_ref(line, -1);
		return -1;
	}

	user_ref = ao2_alloc(sizeof(*user_ref), sca_user_ref_destructor); 
	if (!user_ref) {
        ao2_ref(line, -1);
		return -1;
	}

	ast_string_field_set(line, local_info, local_info);

	line->index = line_index;
	line->shared = ao2_bump(shared);

	AST_LIST_LOCK(&shared->lines);
	AST_LIST_INSERT_TAIL(&shared->lines, line, entry);
	AST_LIST_UNLOCK(&shared->lines);

	user_ref->user = ao2_bump(user);
	user_ref->bridge_user = NULL;

	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_INSERT_TAIL(&line->user_refs, user_ref, ref_entry);
	AST_LIST_UNLOCK(&line->user_refs);

	/* Set the line status to seized */
	sca_change_line_state(shared, line, SCA_Seized);

	ast_log(LOG_NOTICE, "sharedline[%s][%p], line[%p][%d], seized by [%s], user_ref[%p]", sharedline, shared, line, line_index, local_info, user_ref);

    return 0;
}


/*!
 * to get the call info string, the format is as follow:
   <sip:as.foo.com>;appearance-uri="<sip:5000@as.foo.com;user=phone>";appearance-index=1;appearance-state=progressing,
   <sip:as.foo.com>;appearance-index=*;appearance-state=idle
 * 
 */
int ast_sip_get_call_info(char* sharedline, char* local_host, char* buf, int buf_size)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	struct sca_line *line = NULL;
    char *next_pos = buf;
	char  tmp[2048] = {0};
	char *tmp_pos  = NULL;
	char *tmp_at   = NULL;
	int   len = sizeof(tmp);

	if (!sharedline || !local_host || !buf || buf_size <= 0) {
		return -1; 
	}

	/* get the sca shared */
	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "ast_sip_get_call_info, can't find sharedline by [%s]", sharedline);
		return -1; 
	}

    ast_log(LOG_NOTICE, "ast_sip_get_call_info, sharedline[%s], local_host[%s]", sharedline, local_host);

	tmp_at = strchr(local_host, '@');
	/* get each line's info */
	AST_LIST_LOCK(&shared->lines);
	AST_LIST_TRAVERSE(&shared->lines, line, entry) {
		memset(tmp, 0, len);
		tmp_pos = tmp;

        snprintf(tmp_pos, len, "<%s>;appearance-index=%d;", local_host, line->index);
		tmp_pos = tmp + strlen(tmp);

		if (!ast_strlen_zero(line->remote_info)) {
            snprintf(tmp_pos, len-strlen(tmp), "appearance-uri=\"<sip:%s%s>\";", line->remote_info, tmp_at);
            tmp_pos = tmp + strlen(tmp);
		}
        snprintf(tmp_pos, len-strlen(tmp), "appearance-state=%s", sca_state2str(line->status));

		/* copy to buffer */
        if (strlen(buf) + strlen(tmp) >= buf_size) {
            ast_log(LOG_WARNING, "ast_sip_get_call_info, the buf size[%d] is too small to contain all the callinfo", buf_size);
            AST_LIST_UNLOCK(&shared->lines);
            return -1; 
        }

        snprintf(next_pos, buf_size - strlen(buf), "%s,", tmp); 
		next_pos = buf + strlen(buf);
	}

	AST_LIST_UNLOCK(&shared->lines);

    /* to added the idle lines */
    memset(tmp, 0, len);
    snprintf(tmp, len, "<%s>;appearance-state=idle;appearance-index=*", local_host);

    if (strlen(buf) + strlen(tmp) >= buf_size) {
        ast_log(LOG_WARNING, "ast_sip_get_call_info, the buf size[%d] is too small to contain all the callinfo", buf_size);
        return -1; 
    }

    snprintf(next_pos, buf_size - strlen(buf), "%s", tmp); 

    ast_log(LOG_NOTICE, "The call info is [%s]", buf);

    return 0;
}


int ast_sip_line_unseize(char* sharedline, char* local_info, int line_index)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);

	if (!sharedline || !local_info|| line_index <= 0) {
		return -1; 
	}

	/* get the sca shared */
	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "ast_sip_line_unseize, can't find sharedline by [%s]", sharedline);
		return -1; 
	}

    ast_log(LOG_NOTICE, "ast_sip_line_unseize, sharedline[%s], line[%d], local_info[%s]", sharedline, line_index, local_info);

	/* check whether the line has been used */
	line = sca_line_get(shared, line_index);
	if (!line) {
		ast_log(LOG_NOTICE, "ast_sip_line_unseize, not found line[%d] in sharedline[%s]", line_index, sharedline);
		return 0; 
    }

	/* check the line status */
	if (line->status != SCA_Seized) {
		ast_log(LOG_NOTICE, "ast_sip_line_unseize, sharedline[%s], line[%d], status is [%s]", 
		        sharedline, line_index, sca_state2str(line->status));
		return 0; 
	}

	/* check the local info */
	if (strcmp(line->local_info, local_info)) {
		ast_log(LOG_WARNING, "ast_sip_line_unseize, sharedline[%s], the line[%d], seized by [%s], not [%s]", 
		        sharedline, line_index, line->local_info, local_info);
		return -1; 
	}

    ast_log(LOG_NOTICE, "ast_sip_line_unseize, sharedline[%s], line[%d], unseized by [%s]", sharedline, line_index, local_info);
	sca_change_line_state(shared, line, SCA_Idle);

	return 0;
}



int ast_sip_register_line_seize_handler(struct ast_sip_line_seize_handler *handler)
{
	if (!handler || !handler->seize_terminated) {
		return -1;
	}

	SCOPED_LOCK(lock, &line_seize_handlers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&line_seize_handlers, handler, next); 
	ast_module_ref(ast_module_info->self);

    return 0; 
}


void ast_sip_unregister_line_seize_handler(struct ast_sip_line_seize_handler *handler)
{
	struct ast_sip_line_seize_handler *iter;

	SCOPED_LOCK(lock, &line_seize_handlers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&line_seize_handlers, iter, next) {
		if (handler == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

    return;
}

static int sca_call_info_terminated(char* sharedline, char* exten)
{
	struct ast_sip_call_info_handler *iter;

	SCOPED_LOCK(lock, &call_info_handlers, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&call_info_handlers, iter, next) {
		if (0 == iter->call_info_terminated(sharedline, exten)) {
			return 0;
		}
	}
	return -1;
}

int ast_sip_register_call_info_handler(struct ast_sip_call_info_handler *handler)
{
	if (!handler || !handler->call_info_terminated) {
		return -1;
	}

	SCOPED_LOCK(lock, &call_info_handlers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&call_info_handlers, handler, next); 
	ast_module_ref(ast_module_info->self);

    return 0; 
}


void ast_sip_unregister_call_info_handler(struct ast_sip_call_info_handler *handler)
{
	struct ast_sip_call_info_handler *iter;

	SCOPED_LOCK(lock, &call_info_handlers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&call_info_handlers, iter, next) {
		if (handler == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

    return;
}

#define ____SCA_APP_CALL____

static char* get_number_from_chan(const char *chan)
{
	int len = 0;
	char *ret = NULL;
	char *start = NULL, *end = NULL;

	if (ast_strlen_zero(chan)) {
		ast_log(LOG_ERROR, "Invalid param");
		goto END;
	}

	start = strchr(chan, '/');
	end = strchr(chan, '-');

	start = start ? start+1:NULL;

	if (!start || !end) {
		ast_log(LOG_WARNING, "Invalid param, chan:%s", chan);
		goto END;
	}

	/*
	 *  PJSIP/40005-00004563
	 *        |    |
	 *        s    e
	 */
	len = end-start+1;
	ret = (char *)ast_malloc(len);
	if (!ret) {
		ast_log(LOG_ERROR, "Insufficient memory");
		goto END;
	}

	snprintf(ret, len, "%s", start);
	ast_log(LOG_DEBUG, "%s ==> %s", chan, ret);
END:
	return ret;
}

static int sca_get_record_setting(struct ast_channel *chan, struct sca_shared *shared, struct sca_line *line)
{
	int   auto_record_enable = 0;
	char  *conf_inv       = NULL;
	const char *global_record_forbidden = NULL;

	if (!chan || !shared || !line) {
		return 0;
	}

	if (!strncmp(ast_channel_name(chan), "Local/REC", 9)) {
		line->record_flag = 0;
		line->custom_record = 1;
		ast_log(LOG_NOTICE, "sca_get_record_setting, it's a custom record");
		return 0;
	}

	conf_inv = (char *)pbx_builtin_getvar_helper(chan, "CONF_INVITE"); 
	if (!ast_strlen_zero(conf_inv) && !strcmp(conf_inv, "yes")) {
		line->record_flag = 0;
		ast_log(LOG_NOTICE, "sca_get_record_setting, it's a conf invitation.");
		return 0;
	}

	if(ast_test_flag(ast_channel_flags(chan), AST_FLAG_RECOVERYING))
	{
		line->record_flag = 0;
		return 0;
	}

	{
		char trunk[64];
		char route[64];
		const char *from_tk = pbx_builtin_getvar_helper(chan, "FROM_TRUNK");
		const char *leave_tk = pbx_builtin_getvar_helper(chan, "LEAVE_TRUNK");
		const char *ivr_context = pbx_builtin_getvar_helper(chan, "IVRCONTEXT");
		const char *from_disa = pbx_builtin_getvar_helper(chan, "FROM_DISA");
		const char *from_cb = pbx_builtin_getvar_helper(chan, "FROM_CALLBACK");
		const char *init_caller_name = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");

		const char *caller = ast_channel_name(chan);
		const char *callee = shared->name;
		struct auto_record_node* node = NULL;

		trunk[0] = '\0';
		route[0] = '\0';

		if (!ast_strlen_zero(from_disa) && !strcmp(from_disa, "yes")
				&& !ast_strlen_zero(from_cb) && !strcmp(from_cb, "yes")
				&& !ast_strlen_zero(init_caller_name)) {
			struct ast_channel *init_caller = ast_channel_get_by_name(init_caller_name);
			if (init_caller) {
				struct ast_channel *chan_real_caller = ast_channel_real_peer(init_caller);
				if (chan_real_caller) {
					leave_tk = "yes";
					caller = ast_channel_name(chan_real_caller);
					pbx_builtin_setvar_helper(chan, "OUTBOUND_RT_PRE_INDEX",
						pbx_builtin_getvar_helper(chan_real_caller, "OUTBOUND_RT_PRE_INDEX"));
					ast_log(LOG_DEBUG, "Catch special case, caller correct ==> %s", caller);

					ast_channel_unref(chan_real_caller);
				}

				ast_channel_unref(init_caller);
			}
		}

		if ((!ast_strlen_zero(from_tk) && !strcmp(from_tk, "yes"))
			|| (!ast_strlen_zero(leave_tk) && !strcmp(leave_tk, "yes"))) {
			char *trunk_num_begin;
			char *trunk_num_end;
			const char *route_num = NULL;

			if (caller && strstr(caller, "Local/")) {
				caller = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");
				if (ast_strlen_zero(caller)) {
					ast_log(LOG_WARNING, "Channel Var : 'FIRST_CALLER_CHANNEL_NAME' not found!");
					return -1;
				}
				ast_log(LOG_DEBUG, "caller updated to '%s'", caller);
			}

			if (caller && strstr(caller, "DAHDI/o")){
				trunk_num_begin = strchr(caller, 'o');
			}else{
				trunk_num_begin = caller ? strchr(caller, '_') : NULL;
			}

			trunk_num_end = caller ? strchr(caller, '-') : NULL;
			if (!trunk_num_begin || !trunk_num_end || trunk_num_end < trunk_num_begin) {
				ast_log(LOG_WARNING, "Can't retrieve trunk index!, caller:%s", caller);
				return -1;
			}

			if (caller && strstr(caller, "PJSIP")) {
				strcat(trunk,"sip_");
			} else if (caller && strstr(caller, "DAHDI")) {
				strcat(trunk,"dahdi_");
			} else if (caller && strstr(caller, "IAX")) {
				strcat(trunk,"iax_");
			} else {
				ast_log(LOG_WARNING, "Unkown caller:%s", caller);
				return -1;
			}

			strncat(trunk, trunk_num_begin+1, trunk_num_end - trunk_num_begin - 1);
			ast_log(LOG_NOTICE, "trunk : %s", trunk);

			if (!ast_strlen_zero(from_tk) && !strcmp(from_tk, "yes")) {
				strcat(route, "inbound_index_");
				route_num = pbx_builtin_getvar_helper(chan, "INBOUND_RT_INDEX");
			} else {
				strcat(route, "outbound_index_");
				route_num = pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_PRE_INDEX");
				if (ast_strlen_zero(route_num)) {
					route_num = pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_INDEX");
				}
			}

			if (route_num) {
				strcat(route, route_num);
			} else {
				ast_log(LOG_ERROR, "Fail to retrieve route_index");
				return -1;
			}
			ast_log(LOG_NOTICE, "route : %s", route);
		}

		if (caller && strstr(caller, "DAHDI/")) {
			caller = pbx_builtin_getvar_helper(chan, "REALCALLERIDNUM");
			if (ast_strlen_zero(caller)) {
				caller = pbx_builtin_getvar_helper(chan, "ORG_CALLER_ID_NUM");
			}

			if (ast_strlen_zero(caller)) {
				ast_log(LOG_WARNING, "Failed to retrive var 'REALCALLERIDNUM' for %s",
					ast_channel_name(chan));
				return -1;
			}
		} else if (strstr(caller, "Local/")) {
			const char *a_xfer_state = pbx_builtin_getvar_helper(chan, "A_XFER_STATE");
			if (!ast_strlen_zero(a_xfer_state) && !strcmp(a_xfer_state, "begin")) {
				caller = pbx_builtin_getvar_helper(chan, "CALLER_NUM");
				if (ast_strlen_zero(caller)) {
					struct ast_party_caller *party_caller = ast_channel_caller(chan);

					if (party_caller->ani.number.valid && !ast_strlen_zero(party_caller->ani.number.str)) {
						caller = party_caller->ani.number.str;
					} else {
						ast_log(LOG_WARNING, "Failed to retrive var 'CALLER_NUM' for %s",
							ast_channel_name(chan));
						return -1;
					}
				}
			} else {
				caller = pbx_builtin_getvar_helper(chan, "CALLER_NUM");
				if (ast_strlen_zero(caller)) {
					caller = pbx_builtin_getvar_helper(chan, "AAUSER");
				}
			}
		}

		AST_LIST_LOCK(&record_lists);

		AST_LIST_TRAVERSE(&record_lists, node, entry) {
			if (!node) {
				continue;
			}

			if (node->type == SCA_RECORD_EXTENSION) {
				ast_log(LOG_DEBUG, "{type:%d|%s:%s} -- {caller:%s,callee:%s}",
							node->type, node->extensions, node->enable, caller, callee);
				if (!strcmp(node->enable, "all")
					|| (ast_strlen_zero(trunk) && !strcmp(node->enable, "internal"))
					|| (!ast_strlen_zero(trunk) && !strcmp(node->enable, "external"))) {
					if (callee && !strcmp(callee, node->extensions)) {
						auto_record_enable = 1;
						break;
					}

					if (caller && strstr(caller, node->extensions)) {
						char *caller_num = get_number_from_chan(caller);
						if (caller_num) {
							if (!strcmp(caller_num, node->extensions)) {
								auto_record_enable = 1;
							}
							ast_free(caller_num);
						} else {
							auto_record_enable = 1;
						}

						if (auto_record_enable) {
							break;
						}
					}
				}
			} else if (node->type == SCA_RECORD_TRUNK && !ast_strlen_zero(trunk)) {
				ast_log(LOG_DEBUG, "{type:%d|%s:%s} -- {trunk:%s}",
					node->type, node->extensions, node->enable, trunk);
				if (!strcmp(node->extensions, trunk) && !strcmp(node->enable, "yes")) {
					auto_record_enable = 1;
					break;
				}
			} else if (node->type == SCA_RECORD_ROUTE && !ast_strlen_zero(route)) {
				ast_log(LOG_DEBUG, "{type:%d|%s:%s} -- {route:%s}",
					node->type, node->extensions, node->enable, route);
				if (!strcmp(node->extensions, route) && !strcmp(node->enable, "yes")) {
					auto_record_enable = 1;
					break;
				}
			} else if (node->type == SCA_RECORD_IVR && !ast_strlen_zero(ivr_context)) {
				ast_log(LOG_DEBUG, "{type:%d|%s:%s} -- {ivr_context:%s}",
					node->type, node->extensions, node->enable, ivr_context);
				if (!strcmp(node->extensions, ivr_context) && !strcmp(node->enable, "yes")) {
					auto_record_enable = 1;
					break;
				}
			}
		}
		
		AST_LIST_UNLOCK(&record_lists);

		if (!auto_record_enable) {
			char res_buffer[8] = {0};
			char *func_param;

			const char *in_ringgroup = pbx_builtin_getvar_helper(chan, "IN_RINGGROUP");
			const char *group_num = pbx_builtin_getvar_helper(chan, "GROUP_NUMBER");

			struct ast_custom_function *custom_func = ast_custom_function_find("RECORD_CHECK");
			if (!custom_func || !custom_func->read) {
				goto END;
			}
			
			if (!ast_strlen_zero(in_ringgroup) && !strcmp(in_ringgroup, "yes") && !ast_strlen_zero(group_num)) {
				if (-1 == asprintf(&func_param, "RINGGROUP,%s-%s,%s", caller, callee, group_num)) {
					goto END;
				}

				if (!func_param) {
					goto END;
				}

				custom_func->read(chan, NULL, func_param, res_buffer, sizeof(res_buffer));
				if (!ast_strlen_zero(res_buffer) && !strcmp(res_buffer, "yes")) {
					auto_record_enable = 1;
				}

				free(func_param);
			}
		}

		if (!auto_record_enable) {
			int clear = 0;
			const char *cf_reason = pbx_builtin_getvar_helper(chan, "CF_REASON");

			if (!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "13")) {
				clear = 1;
			}

			if (clear) {
				ast_cdr_newest_setvar(ast_channel_name(chan), "RECORD_FILES", "");
			}
		}
	}

	 
END:
	global_record_forbidden = pbx_builtin_getvar_helper(NULL, "RECORD_FORBIDDEN");
	if(global_record_forbidden && !ast_strlen_zero(global_record_forbidden) && !strcmp(global_record_forbidden, "yes")) {
		ast_log(LOG_DEBUG, "reset auto_record_enable from:%d since RECORD_FORBIDDEN enabled", auto_record_enable);
		auto_record_enable = 0;
	}

	ast_log(LOG_DEBUG, "auto_record_enable:%d", auto_record_enable);
	line->record_flag = auto_record_enable;

	if (!auto_record_enable) {
		const char *have_prompted = pbx_builtin_getvar_helper(chan, "HAVE_PROMPTED");
		if (!ast_strlen_zero(have_prompted) && !strcmp(have_prompted, "yes")) {
			ast_log(LOG_DEBUG, "CHAN:%s need to prompt STOP_RECORDING", ast_channel_name(chan));
			ast_stream_and_wait(chan, "dialog-stop-recording", "");
			pbx_builtin_setvar_helper(chan, "HAVE_PROMPTED", "");
		}
	} else {
		pbx_builtin_setvar_helper(chan, "REC_ENABLE", "yes");
	}

	return 0;
}

static int sca_get_limit_setting(struct ast_channel *chan, struct sca_bridge *bridge)
{
	char *from_trunk      = NULL;
	char  *dur_limit      = NULL;
	int   duration        = 0;

    if (!chan || !bridge) {
        return 0;
    }

    from_trunk  = (char *)pbx_builtin_getvar_helper(chan, "FROM_TRUNK"); 

	/* Modified by yfyang, for Bug 118496, 103806*/
	if (ast_strlen_zero(from_trunk)) {
		const char *limit_time  = (char *)pbx_builtin_getvar_helper(chan, "LIMIT_TIME"); 
		if (!ast_strlen_zero(limit_time)) {
			bridge->duration = atoi(limit_time);

			const char *warning_time  = (char *)pbx_builtin_getvar_helper(chan, "WARNING_TIME"); 
			if (!ast_strlen_zero(warning_time)) {
				bridge->limit_warning = atoi(warning_time);

				const char *repeat_time  = (char *)pbx_builtin_getvar_helper(chan, "REPEAT_TIME"); 
				if (!ast_strlen_zero(repeat_time)) {
					bridge->limit_freq = atoi(repeat_time);
				}
			}

			ast_log(LOG_NOTICE, "bridge[%s], call limit[%d], warning[%d], freq[%d]!\n", 
			        bridge->name, bridge->duration, bridge->limit_warning, bridge->limit_freq);
		}
	}

	/* get channel val */
	dur_limit = (char *)pbx_builtin_getvar_helper(chan, "CALLER_DUR_LIMIT");
	if (ast_strlen_zero(dur_limit)) {
		dur_limit = (char *)pbx_builtin_getvar_helper(chan, "LIMITIME_MAX");
	}

	if (!ast_strlen_zero(dur_limit)) {
		duration = atoi(dur_limit);
		if ((duration < bridge->duration) || (bridge->duration == 0)) {
			bridge->duration = duration;
		}
		ast_log(LOG_NOTICE, "chan[%s], duration[%d], set bridge->duration=[%d]!\n", ast_channel_name(chan), duration, bridge->duration);
	}

    return 0;
}

static void sca_set_local_flag(struct ast_channel *chan, char *sca_id)
{
	struct ast_channel *chan1 = NULL;
	char *chan_name1 = NULL;
	char *pos = NULL;

	if (!chan || !sca_id) {
		return;
	}

	/* only process local channel, such as Local/3007@call-forward-00000000;2 */
	if (strncmp(ast_channel_name(chan), "Local/", 6)) {
		return;
	}

	chan_name1 = ast_strdupa(ast_channel_name(chan));
	if ((pos = strstr(chan_name1, ";2"))) {
		pos++;
		*pos = '1';

		ast_log(LOG_NOTICE, "sca_set_local_flag, chan_name1[%s]", chan_name1);
		chan1 = ast_channel_get_by_name(chan_name1);
		if (chan1) {
			pbx_builtin_setvar_helper(chan1, "SCA_UNIQUEID", sca_id);
			ast_channel_unref(chan1);   /* Added by yfyang, for Bug 164956 */
		}
	}
}

static int sca_process_inbridge_call(struct ast_channel *chan, struct sca_shared *shared, struct sca_user *user, struct sca_line *line)
{
	struct sca_bridge   *bridge   = NULL;
	struct sca_user_ref *user_ref = NULL;
	struct ast_flags64 flags = { 0 };
	char  bridge_name[80] = {0};
	int   bridge_exist = 0;

	if (!chan || !shared || !user || !line) {
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
		return -1;
	}

	snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", shared->name, line->index);
	//bridge = sca_build_bridge(bridge_name, 1, chan);

	AST_LIST_LOCK(&sca_bridges);
	AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
		if (!strcmp(bridge_name, bridge->name)) {
			bridge_exist = 1;
			break;
		}
	}
	AST_LIST_UNLOCK(&sca_bridges);

	if (!bridge_exist) {
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
		return -1;
	}

	if (user->is_private && ast_channel_caller(chan)->id.number.valid) {
		/* update the caller number */
		ast_free(ast_channel_caller(chan)->id.number.str);
		ast_channel_caller(chan)->id.number.str = ast_strdup(user->name);
	}

    ast_atomic_fetchadd_int(&bridge->refcount, 1); 

	if (line->status == SCA_Held) {
		ast_set_flag64(&flags, FLAG_MARKEDEXIT | FLAG_UNHOLD); 
	} else {
		ast_set_flag64(&flags, FLAG_MARKEDEXIT | FLAG_BARGE); 
	}

	ast_answer(chan);
	ast_atomic_fetchadd_int((int *) &line->active_count, 1);

	user->line = ao2_bump(line); 
	user->local_chan = chan;
	user->peer_chan = line->remote_chan;

	/* to record the user refs*/
	user_ref = ao2_alloc(sizeof(*user_ref), sca_user_ref_destructor); 
	if (user_ref) {
		user_ref->user = ao2_bump(user);
		user_ref->bridge_user = NULL;

		AST_LIST_LOCK(&line->user_refs);
		AST_LIST_INSERT_TAIL(&line->user_refs, user_ref, ref_entry);
		AST_LIST_UNLOCK(&line->user_refs);

		sca_update_local_user(user->shared, line);
	}

	/* to join the bridge */
	sca_bridge_run(chan, bridge, line, user, &flags);

	if (ast_atomic_dec_and_test((int *) &line->active_count)) {
		ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
		sca_change_line_state(shared, line, SCA_Idle);
	}

	/* user exit the bridge, clear resource */
	sca_dispose_bridge(bridge);
	bridge = NULL;

	ao2_ref(user->line, -1); 
	user->line = NULL;
	user->local_chan = NULL;

	pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "SUCCESS");

	return 0;
}

static int sca_exec(struct ast_channel *chan, const char *data)
{
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);

	struct sca_bridge *bridge = NULL;
	char *sharedline;
	char  bridge_name[80] = {0};
	char  user_name[80]   = {0};
	char  family[80]      = {0};
	char  db_data[80]     = {0};
	char  *codeblue_rt    = NULL;
	char  *rem_chan_name  = NULL;
	char  *from_ivr       = NULL;
	int   ringtime        = 60;
	struct ast_flags64 flags = { 0 };
	int    line_index;
	int    moh_start = 0;
	int    remote_is_dahdi = 0;
	struct ast_frame  *f;
	int alerting_flag = 0; /*Add by weidu for 185654*/

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "Invalid Arguments to Sca!\n");
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
		return 0;
	}

	sharedline = ast_strdupa(data);

	shared = sca_find_shared(sharedline);
	if (!shared) {
		ast_log(LOG_WARNING, "sharedline[%s] not found!\n", sharedline);
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
		return 0;
	}

	line_index = ast_channel_get_line_index(chan);
	if ((line_index > 0) && 
	    (strcmp(ast_channel_tech(chan)->type, "PJSIP") == 0) && 
	    (strcmp(ast_channel_get_sharedline(chan), sharedline) == 0)) {
		ast_log(LOG_NOTICE, "sca_exec, chan[%s] in line index[%d]", ast_channel_name(chan), line_index);

		line = sca_line_get(shared, line_index);
		if (!line) {
			ast_log(LOG_WARNING, "sharedline[%s] get line[%d] failed!\n", sharedline, line_index);
			pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
			return 0;
		}

		sca_get_channel_user_name(chan, user_name, sizeof(user_name));
		user = sca_find_user(user_name);
		if (!user) {
			ast_log(LOG_WARNING, "chan[%s], can't fine sca user by user_name[%s]!\n", ast_channel_name(chan), user_name);
			pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
			return 0;
		}

		if (((line->status == SCA_Held) && shared->retrv_remote_call) ||
		    ((line->status == SCA_Active) && shared->barge)) {
			ast_log(LOG_NOTICE, "sca_exec, shared[%s], line[%d] state[%s], process inbridge call", 
			     sharedline, line_index, sca_state2str(line->status));

			if (sca_process_inbridge_call(chan, shared, user, line)) {
				return 0;
			}
			return 0; 
		}

		ast_log(LOG_WARNING, "sca_exec, line[%d] status is [%s], only held/active could be processed !\n", line_index, sca_state2str(line->status));
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
		return 0;
	} 

	ast_channel_set_caller_extension_flag(chan, 1);

	/* alloc a line */
	line_index = sca_line_alloc(sharedline, sharedline, chan);
	if (line_index <= 0) {
		ast_log(LOG_WARNING, "sharedline[%s] alloc line failed!\n", sharedline);
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "BUSY");
		pbx_builtin_setvar_helper(chan, "SCA_DIAL_STATUS", "UNANSWERED");
		return 0;
	}

	line = sca_line_get(shared, line_index);
	if (!line) {
		ast_log(LOG_WARNING, "sharedline[%s] get line[%d] failed!\n", sharedline, line_index);
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "BUSY");
		pbx_builtin_setvar_helper(chan, "SCA_DIAL_STATUS", "UNANSWERED");
		return 0;
	}

	/* get channel val */
	codeblue_rt = (char *)pbx_builtin_getvar_helper(chan, "CODEBLUE_RT");
	if (!ast_strlen_zero(codeblue_rt)) {
		ringtime = atoi(codeblue_rt);
		line->ringtime = ringtime;
	}

	if (0 == line->ringtime) {
		/* to check the ringtime, get AAUSER RT data from ucmdb */
		snprintf(family, sizeof(family), "AAUSER/%s", sharedline);
		ast_ucm_db_get(family, "RT", db_data, sizeof(db_data));

		if (0 == strlen(db_data)) {
			char *call_rt = (char *)pbx_builtin_getvar_helper(chan, "CALL_RT");
			if (!ast_strlen_zero(call_rt)) {
				ringtime = atoi(call_rt);
				ast_log(LOG_NOTICE, "sharedline[%s], general ringtime[%d]!\n", sharedline, ringtime);
			}
		} else {
			ringtime = atoi(db_data);
		}

		line->ringtime = ringtime;
	}

	pbx_builtin_setvar_helper(chan, "__SCA_APP_CALL", "yes");
	pbx_builtin_setvar_helper(chan, "EXTTOCALL", sharedline);

	/* Added by yfyang, for Bug 188746 */
	if (strstr(ast_channel_name(chan), "Local/") && strstr(ast_channel_name(chan), ";2")) {
		char tmp[256] = {0};
		struct ast_channel *local1 = NULL;

		snprintf(tmp, sizeof(tmp)-1, "%s", ast_channel_name(chan));
		tmp[strlen(tmp)-1] = '1';
		local1 = ast_channel_get_by_name(tmp);
		if (local1) {
			pbx_builtin_setvar_helper(local1, "__SCA_APP_CALL", "yes");
			ast_channel_unref(local1);
		}
	}

	if (sca_ring_users(shared, line)) {
		ast_log(LOG_WARNING, "sharedline[%s] ring all user failed!\n", sharedline);
		pbx_builtin_setvar_helper(chan, "SCA_DIAL_STATUS", "UNANSWERED");
		return sca_change_line_state(shared, line, SCA_Idle);
	}

	snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", shared->name, line->index);
	bridge = sca_build_bridge(bridge_name, 1, chan);
	if (!bridge) {
		pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "FAILURE");
		return sca_change_line_state(shared, line, SCA_Idle);
	}

	/* get record setting */
	sca_get_record_setting(chan, shared, line);
	if (line->record_flag) {
		ast_set_flag64(&flags, FLAG_MARKEDEXIT | FLAG_MARKEDUSER | FLAG_MOH_RBT | FLAG_RECORD); 
	} else {
		ast_set_flag64(&flags, FLAG_MARKEDEXIT | FLAG_MARKEDUSER | FLAG_MOH_RBT); 
	}

	/* get call limit setting */
	sca_get_limit_setting(chan, bridge);

	rem_chan_name = (char *)ast_channel_name(chan);

	/* Added by yfyang, for Bug160209 */
	from_ivr = (char *)pbx_builtin_getvar_helper(chan, "FROM_IVR");
	if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes") && (AST_STATE_UP == ast_channel_state(chan))) {
		/*BEGIN: Mod by weidu for 182683, incoming from ivr to sca extension*/
		/* /AAUSER/3007/MOHFORIVR_ENABLE : no/yes*/
		snprintf(family, sizeof(family), "AAUSER/%s", sharedline);
		ast_ucm_db_get(family, "MOHFORIVR_ENABLE", db_data, sizeof(db_data));
		if (strlen(db_data) && !strncasecmp(db_data, "no",2)){
			ast_indicate(chan, AST_CONTROL_RINGING); 
			/*BEGIN: Add by weidu ,for Bug 185654*/
			alerting_flag =1;
			/*END*/
		}else{
			/* /AAUSER/3007/IVR_MOH : default   */
			snprintf(family, sizeof(family), "AAUSER/%s", sharedline);
			ast_ucm_db_get(family, "IVR_MOH", db_data, sizeof(db_data));

			if (strlen(db_data) > 0) {
				ast_moh_start(chan, db_data, NULL);
			} else {
				ast_moh_start(chan, "rbt", NULL);
			}

			ast_channel_fdno_set(chan, AST_TIMING_FD);
			moh_start = 1;
		}
		/*END: Add by weidu*/
		if (!strncmp(rem_chan_name, "DAHDI/", 6)) {
			remote_is_dahdi = 1;	
			ast_log(LOG_NOTICE, "chan[%s], remote_is_dahdi!\n", rem_chan_name);
		}
	} else {
		if (AST_STATE_UP == ast_channel_state(chan) && !strncmp(ast_channel_name(chan), "PJSIP/", 6)) {
			ast_moh_start(chan, "rbt", NULL);
			ast_channel_fdno_set(chan, AST_TIMING_FD);
			moh_start = 1;
		} else {
			ast_indicate(chan, AST_CONTROL_RINGING); 
			if (!strncmp(rem_chan_name, "DAHDI/", 6)) {
				remote_is_dahdi = 1;	
				ast_log(LOG_NOTICE, "chan[%s], remote_is_dahdi!\n", rem_chan_name);
			}
		}
	}

	/* BEGIN: Added by yfyang, for Bug164130 */
	if (!moh_start) {
		char  *special_tone = NULL;
		special_tone = (char *)pbx_builtin_getvar_helper(chan, "SPECIAL_TONE_NAME");
		if (!ast_strlen_zero(special_tone) && strcmp(special_tone, "nil")) {
			if (AST_STATE_UP != ast_channel_state(chan)) {
				ast_indicate(chan, AST_CONTROL_PROGRESS);
			}
			ast_moh_start(chan, special_tone, NULL);
			moh_start = 1;
			ast_log(LOG_NOTICE, "chan[%s], play special tone[%s]!\n", rem_chan_name, special_tone);
		}
	}
	/* END: Added by yfyang, for Bug164130 */

	/* Added by yfyang, for Bug307993 */
	char *transfer_flag = (char *)pbx_builtin_getvar_helper(chan, "SIPTRANSFER");
	char *blind_transfer = (char *)pbx_builtin_getvar_helper(chan, "BLINDTRANSFER");
	char *sip_transfer = (char *)pbx_builtin_getvar_helper(chan, "SIP_TRANSFER");  /* Added by yfyang, for Bug311270 */
	if ((!ast_strlen_zero(transfer_flag) && !strcmp(transfer_flag, "yes")) || (!ast_strlen_zero(blind_transfer)) || (!ast_strlen_zero(sip_transfer) && !strcmp(sip_transfer, "ft_attended"))) {
		struct ast_party_connected_line connected;

		ast_party_connected_line_init(&connected);
		connected.id.number.str = ast_strdup(sharedline);
		connected.id.number.valid = 1;
		ast_party_connected_line_copy(ast_channel_connected(chan), &connected);

		ast_log(LOG_NOTICE, "indicate CONNECTED_LINE[%s] to chan[%s]", sharedline, ast_channel_name(chan));
		ast_indicate(chan, AST_CONTROL_CONNECTED_LINE);
	}

	/* wait until some user enter the bridge */
	int check_num = ringtime*50;
	while (1) {
		if (ast_check_hangup(chan)) {
			ast_log(LOG_WARNING, "caller channel[%s] has hangup\n", ast_channel_name(chan));
			ast_cdr_setvar(ast_channel_name(chan),"STATUS","NOANSWER");
			goto clear_dial;
		}

		if (bridge->user_count > 0) {
			ast_log(LOG_NOTICE, "user_count:%d\n", bridge->user_count);
			ast_log(LOG_NOTICE, "answer the caller chan[%s]\n", ast_channel_name(chan));
			/*BEGIN: Add by weidu ,for Bug 185654*/
			if (ast_opt_avs && alerting_flag){
				ast_playtones_stop(chan);
			}
			/*END: Add by weidu, for Bug 185654*/
			ast_answer(chan);
			break;
		}

		check_num--;
		if (check_num <= 0) {
			/* if not answered in 60s, clear dial */
			ast_log(LOG_NOTICE, "check timeout, clear dial, chan[%s]\n", ast_channel_name(chan));
			pbx_builtin_setvar_helper(chan, "SCA_DIAL_STATUS", "UNANSWERED");
			pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "NOANSWER");
			goto clear_dial;
		}

		if (bridge->exit_flag) {
			ast_log(LOG_NOTICE, "bridge need to exit, clear dial, chan[%s]\n", ast_channel_name(chan));
			if (ast_strlen_zero(pbx_builtin_getvar_helper(chan, "SCA_CALL_STATUS"))) {
				pbx_builtin_setvar_helper(chan, "SCA_DIAL_STATUS", "UNANSWERED");
			}

			goto clear_dial;
		}

		if (ast_atomic_fetchadd_int((int *)&line->related_count, 0) <= 0) {
			/* no user in line, clear dial */
			ast_log(LOG_NOTICE, "no user in line now, clear dial, chan[%s]\n", ast_channel_name(chan));
			pbx_builtin_setvar_helper(chan, "SCA_DIAL_STATUS", "UNANSWERED");
			pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "BUSY");

			/* Added by yfyang, for Bug 307590 */
			sca_change_line_state(shared, line, SCA_Idle);
			goto clear_dial;
		}

		if (moh_start) {
			ast_channel_fdno_set(chan, AST_TIMING_FD);
			f = ast_read(chan);
			if (!f) {
				ast_log(LOG_WARNING, "caller channel has hangup\n");
				ast_cdr_setvar(ast_channel_name(chan),"STATUS","NOANSWER");
				goto clear_dial;
			}
			ast_frfree(f);
		}

		if (remote_is_dahdi) {
			f = ast_read(chan);
			if (!f || ((f->frametype == AST_FRAME_CONTROL) && (f->subclass.integer == AST_CONTROL_HANGUP))) {
				if (f) {
					ast_frfree(f);
				}

				ast_log(LOG_WARNING, "caller channel has hangup\n");
				ast_cdr_setvar(ast_channel_name(chan),"STATUS","NOANSWER");
				goto clear_dial;
			}
			ast_frfree(f);
		}

		usleep(20000); /* sleep 20 ms */
	}

	if (moh_start) {
		if (ast_channel_fdno(chan) == AST_TIMING_FD) {
			ast_channel_fdno_set(chan, -1);
		}

		ast_moh_stop(chan); 
		moh_start = 0;
	}

	sca_bridge_run(chan, bridge, line, NULL, &flags);
	pbx_builtin_setvar_helper(chan, "SCA_CALL_STATUS", "SUCCESS");

	if (bridge->record_chan) {
		sca_bridge_stop_record(bridge);
		//ast_ha_rsync_file(AST_SCA_RECORDING_FILE_ENUM);
	}

	ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
	ast_log(LOG_NOTICE, "sca_exec, bridge_name[%s], caller exit sca_bridge_run() !\n", bridge_name);

clear_dial:
	sca_dispose_bridge(bridge);
	bridge = NULL;

	/* Caller is out of the call now, cancel the dial to the users */
	sca_hangup_users(shared, line, 0);

	return 0;
}

#define ____SCA_CLI_ENTRY____

static const char *sca_state2str(enum sca_status state)
{
	switch (state) {
	case SCA_Idle:
		return "idle";
    case SCA_Seized:
		return "seized";
    case SCA_Progressing:
		return "progressing";
    case SCA_Alerting:
		return "alerting";
    case SCA_Active:
		return "active";
    case SCA_Held:
		return "held";
    case SCA_Held_Private:
		return "held-private";
    case SCA_Bridge_Active:
		return "bridge-active";
    case SCA_Bridge_Held:
		return "bridge-held";
	default:
		return "unknown";
	}
}


static const char *sca_bargewarn2str(enum sca_barge_warn warn)
{
	switch (warn) {
	case Barge_Warn_None:
		return "none";
    case Barge_Warn_Once:
		return "once";
    case Barge_Warn_Repeat:
		return "repeat";
	default:
		return "unknown";
	}
}

static char *sca_show_shareds(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct sca_shared  *shared;
	struct sca_user    *user;
	struct sca_line    *line;

	switch (cmd) {
	case CLI_INIT:
		e->command = "sca show shareds";
		e->usage =
			"Usage: sca show shareds\n"
			"       This will list all shareds defined in sca.conf\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n" 
	            "=============================================================\n"
	            "----- Configured SCA Shareds --------------------------------\n"
	            "=============================================================\n"
	            "===\n");

	i = ao2_iterator_init(con_sca_shareds, 0);
	for (; (shared = ao2_iterator_next(&i)); ao2_ref(shared, -1)) {
		ao2_lock(shared);

		ast_cli(a->fd, "=== ---------------------------------------------------------\n"
		            "=== Shared Name:           %s\n"
		            "=== ==> Retrv_remote_call: %s\n"
		            "=== ==> Paging_alert_all:  %s\n"
		            "=== ==> Mca:               %s\n"
		            "=== ==> Barge:             %s\n"
		            "=== ==> Barge_warn:        %s\n"
		            "=== ==> Auto_record:       %s\n"
		            "=== ==> Users ..........\n",
		            shared->name, 
		            (shared->retrv_remote_call ? "Yes" : "No"),
		            (shared->paging_alert_all ? "Yes" : "No"), 
		            (shared->mca ? "Yes" : "No"),
		            (shared->barge ? "Yes" : "No"),
		            sca_bargewarn2str(shared->warn),
		            (shared->auto_record ? "Yes" : "No")
			);

		AST_LIST_TRAVERSE(&shared->users, user, entry) {
			ast_cli(a->fd, "===    ==> User name: %s\n", user->name); 
		}

		AST_LIST_TRAVERSE(&shared->lines, line, entry) {
			ast_cli(a->fd, "===    ==> line:   %d\n"
			            "===       ==> state:  %s\n"
			            "===       ==> local:  %s\n"
			            "===       ==> remote: %s\n",
			            line->index,
			            sca_state2str(line->status),
			            (ast_strlen_zero(line->local_info) ? "null" : line->local_info),
			            (ast_strlen_zero(line->remote_info) ? "null" : line->remote_info)
			       );
		}

		ast_cli(a->fd, "=== ---------------------------------------------------------\n===\n");

		ao2_unlock(shared);
	}
	ao2_iterator_destroy(&i);
	ast_cli(a->fd, "=============================================================\n\n");

	return CLI_SUCCESS;
}


static char *sca_show_users(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct sca_user  *user;

	switch (cmd) {
	case CLI_INIT:
		e->command = "sca show users";
		e->usage =
			"Usage: sca show users\n"
			"       This will list all users defined in sca.conf\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n" 
	            "=============================================================\n"
	            "--- Configured SCA Users ---------------------------------\n"
	            "=============================================================\n"
	            "===\n");

	i = ao2_iterator_init(con_sca_users, 0);
	for (; (user = ao2_iterator_next(&i)); ao2_ref(user, -1)) {
		ao2_lock(user);

		ast_cli(a->fd, "=== ---------------------------------------------------------\n"
		            "=== User Name:      %s\n"
		            "=== ==> as_caller:  %s\n"
		            "=== ==> as_callee:  %s\n"
		            "=== ==> Shared:     %s\n",
		            user->name, 
		            (user->as_caller ? "Yes" : "No"),
		            (user->as_callee ? "Yes" : "No"), 
		            user->shared->name);

		if (user->line) {
			ast_cli(a->fd, "=== ==> line:       %d\n", user->line->index); 
		}
		ast_cli(a->fd, "=== ---------------------------------------------------------\n===\n"); 

		ao2_unlock(user);
	}
	ao2_iterator_destroy(&i);
	ast_cli(a->fd, "=============================================================\n\n");

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_sca[] = {
	AST_CLI_DEFINE(sca_show_shareds, "Show SCA Shareds"),
	AST_CLI_DEFINE(sca_show_users, "Show SCA Users"),
};


static int sca_status_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *ptr = NULL;
	int   ret = 0;

	if (!data || !buf || len == 0) {
		return -1;
	}

	if (!strncmp(data, "PJSIP/", 6)) {
		ptr = data + strlen("PJSIP/");
	} else {
		ptr = data;
	}

	ret = sca_user_check_status(ptr);

	if (ret == SCA_Alerting) {
		strncpy(buf, "RINGING", len);
	} else if (ret > 0) {
		strncpy(buf, "BUSY", len);
	}

	return 0;
}

static struct ast_custom_function sca_status_func = {
	.name = "CHECK_SCA_STATUS",
	.read = sca_status_read,
};

static int sca_shared_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);

	if (!data || !buf || len == 0) {
		return -1;
	}

	if (strncmp(ast_channel_name(chan), "PJSIP/", 6)) {
		strncpy(buf, data, len);
		return 0;
	} 

	user = sca_find_user(data);
	if (!user) {
		strncpy(buf, data, len);
		return 0;
	} 

	strncpy(buf, user->shared->name, len);
	return 0;
}

static struct ast_custom_function sca_shared_func = {
	.name = "SCA_SHARED_USER",
	.read = sca_shared_read,
};
#define ____SCA_BRIDGE_EVENT_REPORT____

static void sca_bridge_create_report(struct sca_bridge *bridge)
{
	if (!bridge) {
		return;
	}

	manager_event(EVENT_FLAG_CALL, "BridgeCreate", "BridgeUniqueid: %s\r\n"
                                   "BridgeType: sca\r\n"
                                   "BridgeTechnology: sca_bridge\r\n"
                                   "BridgeNumChannels: %d\r\n",
                                   bridge->uniqueid, bridge->user_count);
}

static void sca_bridge_destroy_report(struct sca_bridge *bridge)
{
	if (!bridge) {
		return;
	}

	manager_event(EVENT_FLAG_CALL, "BridgeDestroy", "BridgeUniqueid: %s\r\n", bridge->uniqueid);
}

static void sca_bridge_enter_report(struct sca_bridge *bridge, struct ast_channel *chan)
{
	if (!bridge || !chan) {
		return;
	}
	char *caller_name;
	caller_name = ast_escape_c_alloc(S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, ""));

	ast_log(LOG_NOTICE, "[%s] users ,valied:%d conne:%s in bridge now\n", ast_channel_name(chan),ast_channel_connected(chan)->id.number.valid,ast_channel_connected(chan)->id.number.str);

	/*Start added by zyxia for Bug 219166,2022.3.28*/
	char *connected_number = NULL;
	const char *org_callee = NULL;
	connected_number = ast_escape_c_alloc(S_COR(ast_channel_connected(chan)->id.number.valid, ast_channel_connected(chan)->id.number.str, ""));

	if (ast_strlen_zero(connected_number)) {
		org_callee = pbx_builtin_getvar_helper(chan, "ORIG_CALLEE");
	}
	/*End added by zyxia for Bug 219166,2022.3.28*/

	manager_event(EVENT_FLAG_CALL, "BridgeEnter", "BridgeUniqueid: %s\r\n"
				"BridgeTechnology: sca_bridge\r\n"
				"BridgeCreator: sca\r\n"
				"BridgeNumChannels: %d\r\n"
				"Channel: %s\r\n"
				"ChannelStateDesc: Up\r\n"
				"CallerIDNum: %s\r\n"
				"CallerIDName: %s\r\n"
				"ConnectedLineNum: %s\r\n"
				"ConnectedLineName: \r\n"
				"Context: from-internal\r\n"
				"Uniqueid: %s\r\n",
				bridge->uniqueid, bridge->user_count,
				ast_channel_name(chan), 
				S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, ""),
				caller_name,
				org_callee ? org_callee : connected_number,
				ast_channel_uniqueid(chan)
				);
	ast_free(caller_name);
	ast_free(connected_number);

	ast_channel_set_sca_bridge_uniqueid(chan,bridge->uniqueid);//added by zyxia for bug 202354,2021.8.30


}

static void sca_bridge_leave_report(struct sca_bridge *bridge, struct ast_channel *chan)
{
	if (!bridge || !chan) {
		return;
	}

	manager_event(EVENT_FLAG_CALL, "BridgeLeave", "BridgeUniqueid: %s\r\n"
	                               "BridgeNumChannels: %d\r\n"
	                               "Channel: %s\r\n"
	                               "Uniqueid: %s\r\n",
	              bridge->uniqueid, bridge->user_count,
                  ast_channel_name(chan), 
                  ast_channel_uniqueid(chan)
                  );

}

#define ____SCA_THREAD____


static void sca_ringing_user_destroy(struct sca_ringing_user *ringing_user)
{
	if (!ringing_user) {
		return;
	}

	if (ringing_user->user) {
		ao2_ref(ringing_user->user, -1);
		ringing_user->user = NULL;
	}

	ast_free(ringing_user);
}

static void sca_stop_ringing_user(struct sca_ringing_user *ringing_user)
{
	struct sca_user    *user = NULL;
	struct sca_line    *line = NULL;
	struct sca_bridge  *bridge = NULL;
	char bridge_name[80] = {0};
	int related_count = 0;

	if (!ringing_user || !ringing_user->user) {
		return;
	}

	user = ringing_user->user;

	ast_dial_join(user->dial); 
	ast_dial_destroy(user->dial);
	user->dial = NULL;

	ao2_lock(user);
	line = user->line;
	if (line) {
		related_count = ast_atomic_fetchadd_int((int *)&line->related_count, -1) - 1;
		if (!related_count) {
			snprintf(bridge_name, sizeof(bridge_name), "SCA_%s_%d", user->shared->name, line->index);

			ao2_unlock(user);
			AST_LIST_LOCK(&sca_bridges);
			AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
				if (!strcmp(bridge_name, bridge->name)) {
					ast_log(LOG_NOTICE, "No user in line[%d], set exit to bridge[%s]", line->index, bridge_name);
					bridge->exit_flag = 1;
					ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
					break;
				}
			}
			AST_LIST_UNLOCK(&sca_bridges);
			ao2_lock(user);
		}

		ao2_ref(line, -1);
		user->line = NULL;
	}
	ao2_unlock(user);

	sca_ringing_user_destroy(ringing_user);
}

static void sca_answer_chan(struct ast_channel *chan)
{
    if (!chan) {
        return;
    }

    ast_answer(chan); 
	ast_indicate(chan, -1);
}

static int sca_update_local_user(struct sca_shared *shared, struct sca_line *line)
{
	struct sca_user_ref *user_ref = NULL;
	struct sca_user     *user = NULL;
	char   user_buf[128] = {0};
	int    size = 0;
	int    str_len = 0;

	if (!shared || !line) {
		return -1;
	}

	size = sizeof(user_buf);
	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
		user = user_ref->user;
		if (!user) {
			continue;
		}

		str_len = strlen(user_buf);
		if (!str_len) {
			snprintf(user_buf, size, "%s", user->name); 
		} else {
			snprintf(user_buf+str_len, size-str_len, ",%s", user->name); 
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&line->user_refs);

	if (strlen(user_buf)) {
		sca_line_set_user_info((char *)shared->name, line->index, user_buf, NULL); 
	}

	return 0; 
}

static int sca_play_barge_warn(struct sca_bridge_user *bridge_user)
{
	struct sca_user_ref *user_ref = NULL;
	struct sca_user     *user     = NULL;
	struct ast_channel  *chan;
	struct sca_line     *line;

	if (!bridge_user || !bridge_user->chan || !bridge_user->line) {
		return 0;
	}

	chan = bridge_user->chan;
	line = bridge_user->line;
	if (line->status != SCA_Bridge_Active || line->heldby_normal_user) {
		if (bridge_user->sbridge && bridge_user->sbridge->user_count > 2) {
			ao2_ref(bridge_user, +1);
			if ((bridge_user->sched_id = ast_sched_add(sca_sched, SCA_BARGE_WARN_FREQ * 1000, sca_play_barge_warn_timeout, bridge_user)) < 0) {
				ast_log(LOG_ERROR, "sca_play_barge_warn, add schedule failed\n");
				ao2_ref(bridge_user, -1);
				return 0;
			}
		}
		return 0;
	}

	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
		user = user_ref->user;
		if (!user) {
			continue;
		}

		if (user->local_chan == chan) {
			if (user->peer_chan && !ast_check_hangup(user->peer_chan)) {
				ast_log(LOG_NOTICE, "sca_play_barge_warn, play tone to chan[%s]", ast_channel_name(user->peer_chan));
				ast_channel_lock(user->peer_chan);
                ast_stream_and_wait(user->peer_chan, "beep", "");
				ast_channel_unlock(user->peer_chan);
			}
			continue;
		}

		if (user->local_chan && !ast_check_hangup(user->local_chan)) {
			ast_log(LOG_NOTICE, "sca_play_barge_warn, play tone to chan[%s]", ast_channel_name(user->local_chan));
			ast_channel_lock(user->local_chan);
            ast_stream_and_wait(user->local_chan, "beep", "");
			ast_channel_unlock(user->local_chan);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&line->user_refs);

	if ((line->status == SCA_Bridge_Active) && (line->shared->warn == Barge_Warn_Repeat)) {
		ao2_ref(bridge_user, +1);
		if ((bridge_user->sched_id = ast_sched_add(sca_sched, SCA_BARGE_WARN_FREQ * 1000, sca_play_barge_warn_timeout, bridge_user)) < 0) {
			ast_log(LOG_ERROR, "sca_play_barge_warn, add schedule failed\n");
			ao2_ref(bridge_user, -1);
			return 0;
		}

		ast_log(LOG_NOTICE, "sca_play_barge_warn, add schedule succeed\n");
	}

	return 0; 
}

static int sca_play_barge_warn_timeout(const void *data)
{
	RAII_VAR(struct sca_bridge_user *, user, NULL, ao2_cleanup);
	user = (struct sca_bridge_user *)data;

	if (!user || user->destroy) {
		return 0;
	}

	return sca_play_barge_warn(user);
}

/*! \brief Remove the bridge from the list and free it.
   We assume that this was called while holding conflock. */
static int sca_bridge_free(struct sca_bridge *bridge)
{
	if (!bridge) {
		return 0;
	}

	ast_log(LOG_NOTICE, "sca_bridge_free, [%s]", bridge->name); 

	/* report AMI Event: BridgeDestroy */
	sca_bridge_destroy_report(bridge);

	ast_hangup(bridge->record_chan);
	ast_hangup(bridge->playback_chan);

	if (bridge->bridge) {
		ast_bridge_destroy(bridge->bridge, 0);
		bridge->bridge = NULL;
	}

	if (bridge->recordingfilename) {
		ast_free(bridge->recordingfilename);
	}

	if (bridge->recordingfilename_cdr) {
		ast_free(bridge->recordingfilename_cdr);
	}

	if (bridge->recordingfileuuid) {
		ast_free(bridge->recordingfileuuid);
	}

	if (bridge->args) {
		ast_free(bridge->args);
	}

	if (bridge->usercontainer) {
		ao2_ref(bridge->usercontainer, -1);
	}

	ast_mutex_destroy(&bridge->lock);
	ast_free(bridge);

	return 0;
}

/*! \brief Decrement reference counts, as incremented by running */
static int sca_dispose_bridge(struct sca_bridge *bridge)
{
	int res = 0;

	ast_log(LOG_NOTICE, "sca_dispose_bridge, bridge_name[%s], refcount[%d] !\n", bridge->name, bridge->refcount);
	AST_LIST_LOCK(&sca_bridges);
	if (ast_atomic_dec_and_test(&bridge->refcount)) {
		AST_LIST_REMOVE(&sca_bridges, bridge, list);
		res = 1;
	}
	AST_LIST_UNLOCK(&sca_bridges);

	if (res) {
		sca_bridge_free(bridge);
	}

	return res;
}


static struct sca_bridge *sca_build_bridge(const char *name, int refcount, const struct ast_channel *chan)
{
	struct sca_bridge     *bridge   = NULL;
	//struct ast_format_cap *cap_slin = NULL;
	time_t creation_time;	
	int    existed   = 0;

	AST_LIST_LOCK(&sca_bridges);
	AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
		if (!strcmp(name, bridge->name))  {
			existed = 1;
			break;
		}
	}

	if (existed) {
		ast_atomic_fetchadd_int(&bridge->refcount, refcount);
		AST_LIST_UNLOCK(&sca_bridges);
		ast_log(LOG_NOTICE, "sca_build_bridge, found existed bridge[%s]", name);
		return bridge;
	}

    AST_LIST_UNLOCK(&sca_bridges);

	/* Make a new one */
	if (!(bridge = ast_calloc(1, sizeof(*bridge)))) {
		ast_log(LOG_NOTICE, "sca_build_bridge, ast_calloc[%s] failed", name);
		return NULL;
	}
        
	if (!(bridge->usercontainer = ao2_container_alloc(1, NULL, sca_bridge_user_cmp))) {
		ast_free(bridge);
		ast_log(LOG_NOTICE, "sca_build_bridge, usercontainer alloc failed, name[%s]", name);
		return NULL;
	}

	creation_time = time(NULL);
	snprintf(bridge->uniqueid, sizeof(bridge->uniqueid), "%s-%li", name, (long)creation_time);
    ast_copy_string(bridge->name, name, sizeof(bridge->name));

	/* set the duration limit */
	bridge->limit_warning = 0;
	bridge->duration   = 0;
	bridge->limit_freq = 0;
	bridge->endtime    = 0;
	bridge->exit_flag  = 0;
	bridge->user_count = 0;
	bridge->markedusers= 0;
	bridge->refcount   = 0;
	//edit by zyxia for Bug 218497, add native and 1to1, if not,it dont reconfigure at fun bridge_reconfigured_connected_line_update,2022.322
	bridge->bridge = ast_bridge_base_new(AST_BRIDGE_CAPABILITY_NATIVE |  AST_BRIDGE_CAPABILITY_1TO1MIX | AST_BRIDGE_CAPABILITY_MULTIMIX, AST_BRIDGE_FLAG_MASQUERADE_ONLY | AST_BRIDGE_FLAG_TRANSFER_BRIDGE_ONLY, sca_app, bridge->name, NULL);
	if (!bridge->bridge) {
		ast_log(LOG_WARNING, "sca_build_bridge, ast_bridge_base_new() failed for [%s]", bridge->name);
		ao2_ref(bridge->usercontainer, -1);
		ast_free(bridge);
		return NULL;
	}

	bridge->bridge->avs_bridge_id = -1;

	ast_mutex_init(&bridge->lock);
	AST_LIST_LOCK(&sca_bridges);
	AST_LIST_INSERT_HEAD(&sca_bridges, bridge, list);
	AST_LIST_UNLOCK(&sca_bridges);

	/* report AMI Event: BridgeCreate */
	sca_bridge_create_report(bridge);
	ast_log(LOG_NOTICE, "Created SCA bridge %p for bridge '%s' by channel '%s'\n", bridge, bridge->name, ast_channel_name(chan));
#if 0
	cap_slin = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!cap_slin) {
		ast_log(LOG_NOTICE, "sca_build_bridge, alloc format cap failed, name[%s]", name);
		ast_atomic_fetchadd_int(&bridge->refcount, refcount);
		return bridge;
	}

    ast_format_cap_append(cap_slin, ast_format_slin, 0); 
	bridge->playback_chan = ast_request("CBAnn", cap_slin, NULL, chan, bridge->name, NULL);
    if (!bridge->playback_chan) {
		ast_log(LOG_NOTICE, "sca_build_bridge, request CBAnn failed for [%s]", bridge->name);
    }

    ast_mutex_init(&bridge->playback_lock);
	ao2_cleanup(cap_slin); 
#endif

	ast_atomic_fetchadd_int(&bridge->refcount, refcount);
	return bridge;
}


static int sca_bridge_set_record_filename(struct sca_bridge *bridge, struct ast_channel *chan, struct sca_line *line)
{
	char recordingtmp[512] = "";
	char recordinguuidtemp[512] = "";
	char recordargs[1024] = "";
	char func_data[512] = {0};

	char destdir[PATH_MAX] = {0};
	//const char *var;
	struct ast_custom_function *func;

	if (!bridge || !chan || !line) {
		return -1;
	}

	func = ast_custom_function_find("RECORD_CHECK");
	if (!func) {
		return -2;
	}

	if (!bridge->recordingfilename) {
		time_t timep;
		struct tm *p;
		const char *caller = line->remote_info;
		const char *from_callback = pbx_builtin_getvar_helper(chan, "FROM_CALLBACK");

		if (!ast_strlen_zero(from_callback) && !strcmp(from_callback,"yes")) {
			const char *cb_src = pbx_builtin_getvar_helper(chan, "CALLBACK_SRC");
			caller = !ast_strlen_zero(cb_src) ? cb_src : caller;
		}

		snprintf(destdir, sizeof(destdir), "%s/sca", ast_config_AST_SPOOL_DIR);
		if (ast_mkdir(destdir, 0777) != 0) {
		    ast_log(LOG_WARNING, "mkdir '%s' failed: %s\n", destdir, strerror(errno));
		    return -1;
		}

		time(&timep);
		p = localtime(&timep);

		snprintf(func_data, sizeof(func_data), "file,%s/%d-%02d/rec-%s-%s-%u.wav", destdir, 1900 + p->tm_year, 1 + p->tm_mon, caller, line->shared->name, (unsigned int)timep);
		func->read(chan, "", func_data, recordingtmp, sizeof(recordingtmp));
		bridge->recordingfilename = ast_strdup(recordingtmp);

		snprintf(func_data, sizeof(func_data), "file,%d-%02d/rec-%s-%s-%u.wav", 1900 + p->tm_year, 1 + p->tm_mon, caller, line->shared->name, (unsigned int)timep);
		func->read(chan, "", func_data, recordingtmp, sizeof(recordingtmp));
		bridge->recordingfilename_cdr = ast_strdup(recordingtmp);

		file_crypt_new_uuid(recordinguuidtemp, sizeof(recordinguuidtemp));
		bridge->recordingfileuuid = ast_strdup(recordinguuidtemp);

		snprintf(recordargs, sizeof(recordargs), "%s,%s", recordingtmp, recordinguuidtemp);
		bridge->args = ast_strdup(recordargs);
	}

	ast_log(LOG_NOTICE, "SCA bridge[%s], set record filename[%s].\n", bridge->name, bridge->recordingfilename);
	ast_log(LOG_NOTICE, "record filename cdr[%s].\n", bridge->recordingfilename_cdr);
	ast_log(LOG_DEBUG, "record file uuid[%s].\n", bridge->recordingfileuuid);

	return 0;
}

static int sca_bridge_start_record(struct sca_bridge *bridge, struct ast_channel *chan, struct sca_line *line)
{
	struct ast_format_cap *cap;
	struct ast_bridge_features *features;
	struct ast_app *mixmonapp;

	if (!bridge || !chan || !line) {
		return -1;
	}

	if (bridge->record_chan) {
		ast_log(LOG_NOTICE, "sca_bridge_start_record, bridge[%s] has started record\n", bridge->name);
		return 0;
	}

	mixmonapp = pbx_findapp("MixMonitor");
	if (!mixmonapp) {
		ast_log(LOG_WARNING, "Cannot record sca bridge, MixMonitor app is not installed\n");
		return -1;
	}

	features = ast_bridge_features_new();
	if (!features) {
		return -1;
	}

	ast_set_flag(&features->feature_flags, AST_BRIDGE_CHANNEL_FLAG_IMMOVABLE);

	cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!cap) {
		ast_bridge_features_destroy(features);
		return -1;
	}

	ast_format_cap_append(cap, ast_format_slin, 0);

	/* Create the recording channel. */
	bridge->record_chan = ast_request("CBRec", cap, NULL, NULL, bridge->name, NULL);
	ao2_ref(cap, -1);
	if (!bridge->record_chan) {
		ast_bridge_features_destroy(features);
		return -1;
	}

	/*added by zyxia for Bug 274086,2023.11.30*/
	if(bridge->record_chan && bridge->bridge && !ast_strlen_zero(bridge->bridge->avs_address))
		ast_channel_set_avs_address(bridge->record_chan,bridge->bridge->avs_address);

	pbx_builtin_setvar_helper(bridge->record_chan, "PARENT_CHAN", pbx_builtin_getvar_helper(chan, "ORI_CALLER_CHANNEL_NAME"));

	/* create avs record channel. */
	if (bridge->bridge->avs_bridge_id > 0) {
		if (ast_avs_create_channel(bridge->record_chan, avs_channel_type_rtp, 0, ast_channel_get_port_group_id(bridge->record_chan)) != 0) {
			ast_log(LOG_ERROR, "Couldn't create avs channel for channel '%s'.", ast_channel_name(bridge->record_chan));
		}
	}

	/* generate recording filename */
	if (!bridge->recordingfilename) {
		sca_bridge_set_record_filename(bridge, chan, line);
	}

	ast_answer(bridge->record_chan);

	ast_channel_ref(bridge->record_chan);
	if (ast_bridge_impart(bridge->bridge, bridge->record_chan, NULL, features, AST_BRIDGE_IMPART_CHAN_INDEPENDENT)) {
		ast_log(LOG_WARNING, "sca_bridge_start_record, record_chan[%s] impart bridge failed", ast_channel_name(bridge->record_chan));
		ast_hangup(bridge->record_chan);
		ast_channel_unref(bridge->record_chan);
		bridge->record_chan = NULL;
		return -1;
	}

	pbx_exec(bridge->record_chan, mixmonapp, (const char*)bridge->args);

	if (line != NULL && line->remote_chan != NULL) {
		char recordFile[1024] = {0};
		const char *oldFile = pbx_builtin_getvar_helper(line->remote_chan, "CDR(RECORD_FILES)");

		if (oldFile != NULL) {
			snprintf(recordFile, sizeof(recordFile), "%s%s", oldFile, bridge->recordingfilename_cdr);
		} else {
			snprintf(recordFile, sizeof(recordFile), "%s@", bridge->recordingfilename_cdr);
		}
		pbx_builtin_setvar_helper(line->remote_chan, "CDR(RECORD_FILES)", recordFile);

		// file encrypt key
		const char *channel_name = ast_channel_name(line->remote_chan);
		if (!ast_strlen_zero(channel_name)) {
			char cdr_record_uuid[1024] = {0};
			ast_cdr_getvar(channel_name, "record_files_uuid", cdr_record_uuid, sizeof(cdr_record_uuid));
    		snprintf(cdr_record_uuid + strlen(cdr_record_uuid), sizeof(cdr_record_uuid) - strlen(cdr_record_uuid), "%s@", bridge->recordingfileuuid);
    		ast_cdr_newest_setvar(channel_name, "record_files_uuid", cdr_record_uuid);
		}
	}

	const char* rec_enable = pbx_builtin_getvar_helper(chan, "REC_ENABLE");
	if (ast_strlen_zero(rec_enable) || strcmp(rec_enable, "yes")) {
		ast_log(LOG_NOTICE, "sca_bridge_start_record, chan[%s] set REC_ENABLE", ast_channel_name(chan));
		pbx_builtin_setvar_helper(chan, "REC_ENABLE", "yes");
		ast_indicate(chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
	}

	return 0;
}


static int sca_bridge_stop_record(struct sca_bridge *bridge)
{
	struct ast_channel *chan;
	struct ast_frame f = { AST_FRAME_CONTROL, .subclass.integer = AST_CONTROL_HANGUP };

	if (!bridge->record_chan) {
		return -1;
	}

    ast_log(LOG_NOTICE, "sca_bridge_stop_record, record_chan[%s]", ast_channel_name(bridge->record_chan));

	/* Remove the recording channel from the sca bridge. */
	chan = bridge->record_chan;
	bridge->record_chan = NULL;

	if (ast_channel_is_avs_engine(chan)) {
		ast_avs_control(chan, AVS_CONTROL_STOP_RECORD, NULL, 0);
	}

	ast_queue_frame(chan, &f);
	ast_channel_unref(chan);

	return 0;
}

static struct sca_bridge_user *sca_find_marked_user(const struct sca_bridge *bridge)
{
	struct sca_bridge_user *user;
	struct ao2_iterator user_iter;
    int    found = 0;

	if (!bridge) {
		return NULL;
	}

	user_iter = ao2_iterator_init(bridge->usercontainer, 0);
	while ((user = ao2_iterator_next(&user_iter))) {
		if (user->marked) {
			found = 1;
			break;
		}

		ao2_ref(user, -1);
	}
	ao2_iterator_destroy(&user_iter);

    if (found) {
        return user;
    }

    return NULL;
}

#if 0
static int sca_bridge_play_sound(struct sca_bridge *bridge, const char *filename)
{
	if (!bridge || ast_strlen_zero(filename)) {
		return 0;
	}

	ast_log(LOG_NOTICE, "sca_bridge_play_sound, filename[%s]", filename);

	ast_mutex_lock(&bridge->playback_lock);
	if (!bridge->playback_chan) {
		ast_mutex_unlock(&bridge->playback_lock);
		return 0;
	}

	if (conf_announce_channel_push(bridge->playback_chan)) {
		ast_mutex_unlock(&bridge->playback_lock);
        ast_log(LOG_NOTICE, "sca_bridge_play_sound, push announce channel failed.");
		return 0;
	}

	ast_stream_and_wait(bridge->playback_chan, filename, "");

	ast_log(LOG_NOTICE, "Departing announcer channel '%s' from SCA bridge '%s'\n", ast_channel_name(bridge->playback_chan), bridge->name);

	conf_announce_channel_depart(bridge->playback_chan);
	ast_mutex_unlock(&bridge->playback_lock);

	return 0;
}
#endif

static void sca_moh_stop(struct sca_bridge_user *user)
{
	int in_bridge;

	if (!user || !user->playing_moh || !user->chan) {
		return;
	}

	user->playing_moh = 0; 

	ast_bridge_lock(user->sbridge->bridge);

	ast_log(LOG_NOTICE, "sca_moh_stop, userno[%d], chan[%s]", user->user_no, ast_channel_name(user->chan));
	ast_channel_hold_state_set(user->chan, AST_CONTROL_UNHOLD);//added by zyxia for Bug 221788,2023.3.17

	in_bridge = !ast_bridge_suspend(user->sbridge->bridge, user->chan);
	ast_moh_stop(user->chan);

	if (in_bridge) {
		if (ast_channel_is_avs_engine(user->chan)) {
 			ast_bridge_set_ignore_unsuspend(user->sbridge->bridge, user->chan, 0);
		}
		ast_bridge_unsuspend(user->sbridge->bridge, user->chan);
	}

    ast_bridge_unlock(user->sbridge->bridge);
}

static void sca_moh_start(struct sca_bridge_user *user, void* mclass)
{
	int in_bridge;

	if (!user || user->playing_moh || !user->chan) {
		return;
	}

	user->playing_moh = 1;

	/* Locking the ast_bridge here is the only way to hold off the
	 * call to ast_bridge_join() in confbridge_exec() from
	 * interfering with the bridge and MOH operations here. */
	ast_bridge_lock(user->sbridge->bridge);

	ast_log(LOG_NOTICE, "sca_moh_start, userno[%d], chan[%s], mclass[%s]", 
			user->user_no, ast_channel_name(user->chan), mclass?(char *)mclass:"default");
	ast_channel_hold_state_set(user->chan, AST_CONTROL_HOLD);//added by zyxia for Bug 221788,2023.3.17

	in_bridge = !ast_bridge_suspend(user->sbridge->bridge, user->chan);
	ast_moh_start(user->chan, (const char *)mclass, NULL);

	if (in_bridge) {
		if (ast_channel_is_avs_engine(user->chan)) {
 			ast_bridge_set_ignore_unsuspend(user->sbridge->bridge, user->chan, 1);
		}
		ast_bridge_unsuspend(user->sbridge->bridge, user->chan);
	}

	ast_bridge_unlock(user->sbridge->bridge);
}


static int sca_bridge_unhold(struct sca_bridge_user *bridge_user)
{
	struct ast_channel  *chan;
    struct sca_bridge   *bridge;
	struct sca_line     *line;
	struct sca_user     *user = NULL;
	struct sca_user_ref *user_ref = NULL;
    struct sca_bridge_user *peer_user = NULL;
	int    is_normal_unhold = 0;
	int    peer_moh = 0;

	if (!bridge_user || !bridge_user->chan || !bridge_user->sbridge || !bridge_user->line) {
		return 0;
	}

	chan = bridge_user->chan;
	bridge = bridge_user->sbridge;
	line = bridge_user->line;

	if (!line->heldby_user && !line->heldby_normal_user) {
		if (line->status == SCA_Bridge_Active || line->status == SCA_Active) {
			ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] is in state[%d]", bridge->name, line->index, line->status);
			return 0;
		}

		ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] hasn't been held", bridge->name, line->index);
		return 0; 
	}

	/* Added by yfyang, for Bug 126320 */
	if ((line->status == SCA_Bridge_Active) && (line->heldby_normal_user != bridge_user)) {
		ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] in bridge_active", bridge->name, line->index);
		return 0;
	}
    
	/* Added by yfyang, for Bug 111341 */
	if (line->status == SCA_Bridge_Held) {
		if (line->heldby_user != bridge_user) {
			if ((line->heldby_normal_user == bridge_user) && line->heldby_user) {
                sca_moh_stop(line->heldby_user);
				line->heldby_normal_user = NULL;
			}

			ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] in bridge_held, chan[%s] try to unhold", bridge->name, line->index, ast_channel_name(chan));
			return 0;
		} 

		if ((line->heldby_user == bridge_user) && (NULL != line->heldby_normal_user)) {
			sca_moh_stop(line->heldby_normal_user);
			line->heldby_user = NULL;

			sca_change_line_state(line->shared, line, SCA_Bridge_Active); 
			ast_log(LOG_NOTICE, "sca_bridge_unhold, stop moh to chan[%s]", ast_channel_name(line->heldby_normal_user->chan));
			return 0;
		}
	}

	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
		user = user_ref->user;
		if (!user || !user->local_chan) {
			continue;
		}

		if (chan == user->peer_chan && user->local_chan) {
			ast_log(LOG_NOTICE, "sca_bridge_unhold, user[%s], stop moh", user->name);
			sca_moh_stop(user_ref->bridge_user);
		} else if ((chan != user->peer_chan) && (chan != user->local_chan)) {
			ast_log(LOG_NOTICE, "sca_bridge_unhold, user[%s], stop moh", user->name);
			sca_moh_stop(user_ref->bridge_user);
		} else if (user->peer_chan) {
			peer_moh = 1;
		}
		continue;
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&line->user_refs);

	if (peer_moh) {
		if ((peer_user = sca_find_marked_user(bridge))) {
			ast_log(LOG_NOTICE, "sca_bridge_unhold, stop moh to peer userno[%d]", peer_user->user_no);
			sca_moh_stop(peer_user);
			ao2_ref(peer_user, -1);
		}
	}

    if (line->heldby_user == bridge_user) {
		ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] retrieved by self, chan[%s] ", 
		        bridge->name, line->index, (chan ? ast_channel_name(chan):"null"));

		if (line->status == SCA_Held_Private) {
			sca_change_line_state(line->shared, line, SCA_Active); 
		}

		line->heldby_user = NULL;

	} else if (line->heldby_normal_user == bridge_user) {
		ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] retrieved by normal channel[%s] ", 
		        bridge->name, line->index, (chan ? ast_channel_name(chan) : "null"));

		line->heldby_normal_user = NULL;
		is_normal_unhold = 1;
	} else if (line->heldby_user != bridge_user) {
		ast_log(LOG_NOTICE, "sca_bridge_unhold, bridge[%s], line[%d] held by[%s], retrieved by chan[%s] ", 
		        bridge->name, line->index, (line->heldby_user ? ast_channel_name(line->heldby_user->chan):"null"), (chan ? ast_channel_name(chan):"null"));

		if ((line->status != SCA_Bridge_Held) && (line->status != SCA_Bridge_Active) && line->heldby_user) {
			ast_bridge_remove(bridge_user->sbridge->bridge, line->heldby_user->chan);
		}
	}

	if (!is_normal_unhold) {
		if (line->status == SCA_Held) {
			sca_change_line_state(line->shared, line, SCA_Active); 
		} else if (line->status == SCA_Bridge_Held) {
			sca_change_line_state(line->shared, line, SCA_Bridge_Active); 
		}

		line->heldby_user = NULL;
	}

	return 0;
}

static int sca_bridge_hold(struct sca_bridge_user *bridge_user, void *mclass)
{
	struct ast_channel *chan;
    struct sca_bridge *bridge;
	struct sca_line *line;
	struct sca_bridge_user *peer_user;
	struct sca_user_ref *user_ref = NULL;
	struct sca_user     *user     = NULL;
	int   is_normal_hold          = 0;
	int   peer_moh = 0;

	RAII_VAR(struct sca_user *, hold_user, NULL, ao2_cleanup);

	if (!bridge_user || !bridge_user->chan || !bridge_user->sbridge || !bridge_user->line) {
		return 0;
	}

	chan = bridge_user->chan;
	bridge = bridge_user->sbridge;
	line = bridge_user->line;

	if (bridge_user->marked) {
		is_normal_hold = 1;
	}

	if (line->heldby_user && !is_normal_hold) {
		ast_log(LOG_NOTICE, "sca_bridge_hold, bridge[%s], line[%d] has been held by chan[%s]", 
		        bridge->name, line->index, ast_channel_name(line->heldby_user->chan));
		return 0;
	}

	if (line->heldby_normal_user && is_normal_hold) {
		ast_log(LOG_NOTICE, "sca_bridge_hold, bridge[%s], line[%d] has been held by chan[%s]", 
		        bridge->name, line->index, ast_channel_name(line->heldby_normal_user->chan));
		return 0;
	}

	AST_LIST_LOCK(&line->user_refs);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
		user = user_ref->user;
		if (!user || !user->local_chan) {
			continue;
		}

		ast_log(LOG_NOTICE, "sca_bridge_hold, line[%d], user[%s], chan[%s], local_chan[%s], peer_chan[%s]", line->index, user->name, 
                chan ? ast_channel_name(chan):"null", 
                user->local_chan ? ast_channel_name(user->local_chan) :"null", 
                user->peer_chan ? ast_channel_name(user->peer_chan) : "null");

		if (user->local_chan == chan) {
			ast_log(LOG_NOTICE, "sca_bridge_hold, found user[%s], line[%d]", user->name, line->index);
			if (user->peer_chan) {
				peer_moh = 1;
			}
			continue;
		}

		if (user->local_chan) {
			ast_log(LOG_NOTICE, "sca_bridge_hold, user[%s] start moh", user->name);
            sca_moh_start(user_ref->bridge_user, mclass);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&line->user_refs);

	if (peer_moh) {
		if ((peer_user = sca_find_marked_user(bridge))) {
			ast_log(LOG_NOTICE, "sca_bridge_hold, start moh to peer userno[%d]", peer_user->user_no);
			sca_moh_start(peer_user, mclass);
			ao2_ref(peer_user, -1);
		}
	}

	if (!is_normal_hold) {
		line->heldby_user = bridge_user; 

		if (line->status == SCA_Active) {
			sca_change_line_state(line->shared, line, SCA_Held); 
		} else if (line->status == SCA_Bridge_Active) {
			sca_change_line_state(line->shared, line, SCA_Bridge_Held); 
		}

		ast_log(LOG_NOTICE, "sca_bridge_hold, line[%d] held by chan[%s]", line->index, ast_channel_name(bridge_user->chan));
  	} else {
		line->heldby_normal_user = bridge_user; 
		ast_log(LOG_NOTICE, "sca_bridge_hold, line[%d] normal held by chan[%s]", line->index, ast_channel_name(bridge_user->chan));
	}

	return 0; 
}

static void sca_hold_hook_destructor(void *hook_pvt)
{
	struct sca_bridge_user *user = hook_pvt;

    ast_log(LOG_NOTICE, "sca_hold_hook_destructor, user[%p], user_no[%d]\n", user, user?user->user_no:0xff);
    return;
}

static int sca_handle_hold_cb(struct ast_bridge_channel *bridge_channel, void *hook_pvt, void *mclass, int hold)
{
	struct sca_bridge_user *user = hook_pvt;
	struct sca_bridge *bridge;
	int  found = 0;

    ast_log(LOG_NOTICE, "sca_handle_hold_cb, bridge_channel[%p], user[%p], hold[%d]\n", bridge_channel, user, hold);
	if (!bridge_channel || !hook_pvt) {
		return 0;
	}

	AST_LIST_LOCK(&sca_bridges); 
	AST_LIST_TRAVERSE(&sca_bridges, bridge, list) {
		if (user->sbridge == bridge) {
			found = 1;
			break;
		}
	}
	AST_LIST_UNLOCK(&sca_bridges);

	if (!found) {
		ast_log(LOG_NOTICE, "sca_handle_hold_cb, can't find sca_bridge[%p] for user[%p]", user->sbridge, user);
		return 0;
	}

    ast_log(LOG_NOTICE, "sca_handle_hold_cb, userno[%d], is hold[%d]", user->user_no, hold);
    if (hold) {
        return sca_bridge_hold(user, mclass);
    }

    return sca_bridge_unhold(user);
}

static int sca_prompt_record(struct ast_channel *chan)
{
	//int res = 0;
	struct ast_channel *local_peer = NULL;
	struct ast_channel *chan_prompt = NULL;

	if (strstr(ast_channel_name(chan), "Local/")) {
		const char *followme_caller_name = NULL;

		/* DO NOT CHANGE the POSITION of this code, it's dangerous!!! */
		if (strstr(ast_channel_name(chan), ";2")) {
			const char *seamless_transfer = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER");
			const char *seamless_transferee_channel = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFEREE_CHANNEL");
			
			if (!ast_strlen_zero(seamless_transfer) && !strcmp(seamless_transfer, "yes")
				&& !ast_strlen_zero(seamless_transferee_channel)) {
				int aboard = 0;
				chan_prompt = ast_channel_get_by_name(seamless_transferee_channel);
				if (chan_prompt) {
					const char *have_prompt = pbx_builtin_getvar_helper(chan_prompt, "HAVE_PROMPTED");
					if (!ast_strlen_zero(have_prompt) && !strcmp(have_prompt, "yes")) {
						aboard = 1;
						ast_log(LOG_DEBUG, "In seamless transfer, transferee(%s) has prompted, aboard prompt record!",
							seamless_transferee_channel);
					}

					ast_channel_unref(chan_prompt);
					chan_prompt = NULL;

					if (aboard) {
						return 0;
					}
				}
			}
		}

		local_peer = ast_local_get_peer(chan);
		followme_caller_name = pbx_builtin_getvar_helper(chan, "FOLLOWME_CALLER");
		if (!ast_strlen_zero(followme_caller_name)) {
			chan_prompt = ast_channel_get_by_name(followme_caller_name);
			if (chan_prompt) {
				ast_log(LOG_DEBUG, "In this case, chang prompt chan to %s",
					ast_channel_name(chan_prompt));
			}
		}
	}

	pbx_builtin_setvar_helper(chan, "RECORD_PROMPTING", "yes");
	if (local_peer) {
		pbx_builtin_setvar_helper(local_peer, "RECORD_PROMPTING", "yes");
	}

	if (chan_prompt) {
		ast_channel_async_streamfile_protection_enter(chan_prompt);
		//res = ast_streamfile(chan_prompt, "dialog-being-recorded", ast_channel_language(chan));
		ast_streamfile(chan_prompt, "dialog-being-recorded", ast_channel_language(chan));
	}

	ast_stream_and_wait(chan, "dialog-being-recorded", "");

	pbx_builtin_setvar_helper(chan, "RECORD_PROMPTING", "");
	if (local_peer) {
		pbx_builtin_setvar_helper(local_peer, "RECORD_PROMPTING", "");
		ast_channel_unref(local_peer);
	}

	pbx_builtin_setvar_helper(chan, "HAVE_PROMPTED", "yes");

	if (chan_prompt) {
#if 0
		const char *moh = pbx_builtin_getvar_helper(chan_prompt, "FOLLOWME_MOH_CLASS");
		if (!res) {
			ast_waitstream(chan_prompt, "");
			if (!ast_strlen_zero(moh)) {
				ast_moh_start(chan_prompt, moh, NULL);
			}
		}
#endif
		ast_channel_async_streamfile_protection_exit(chan_prompt);
		ast_channel_unref(chan_prompt);
	}

	return 0;
}

static int sca_bridge_run(struct ast_channel *chan, struct sca_bridge *bridge, struct sca_line *line, struct sca_user *sca_user, struct ast_flags64 *flags)
{
	struct sca_bridge_user *user     = NULL;
	struct sca_user_ref    *user_ref = NULL;
	struct timeval now = { 0, };
	struct ast_bridge_features chan_features;

	if (!chan || !bridge || !line || !flags) {
		return -1;
	}

	if (!(user = ao2_alloc(sizeof(*user), sca_bridge_user_destructor))) {
		ast_log(LOG_ERROR, "sca_bridge_run, alloc bridge_user for [%s] failed!", ast_channel_name(chan));
		return -1;
	}

	ao2_lock(bridge->usercontainer);
	user->user_no   = bridge->user_count;
	user->chan      = chan;
	user->sbridge   = bridge;
	user->line      = line;
	user->sched_id  = -1;
	user->destroy   = 0;
	user->playing_moh = 0;

	ao2_link(bridge->usercontainer, user);

	bridge->user_count++;
	ao2_unlock(bridge->usercontainer);

	if (ast_test_flag64(flags, FLAG_MARKEDUSER)) {
		bridge->markedusers++;
		user->marked = 1;
	} else {
		user->marked = 0;

		AST_LIST_LOCK(&line->user_refs);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&line->user_refs, user_ref, ref_entry) {
			if (!user_ref || (user_ref->user != sca_user)) {
				continue;
			}

			user_ref->bridge_user = ao2_bump(user);
			break;
		}
		AST_LIST_TRAVERSE_SAFE_END
		AST_LIST_UNLOCK(&line->user_refs);
	}

	/* report AMI Event: BridgeEnter */
	sca_bridge_enter_report(bridge, chan);

	ast_log(LOG_NOTICE, "user[%p][%d], chan[%s] added to bridge, [%d] users in bridge now\n", user, user->user_no, ast_channel_name(chan), bridge->user_count);

	/* return the unique ID of the bridge */
	pbx_builtin_setvar_helper(chan, "SCA_UNIQUEID", bridge->uniqueid);

	/* to do unhold, retrieve by remote location */
	if (ast_test_flag64(flags, FLAG_UNHOLD)) {
		sca_bridge_unhold(user);
	}

	if (ast_test_flag64(flags, FLAG_BARGE) && (bridge->user_count > 2)) {
		sca_change_line_state(line->shared, line, SCA_Bridge_Active); 
		if (line->shared->warn != Barge_Warn_None) {
			ast_log(LOG_NOTICE, "user[%d], sca_play_barge_warn\n", user->user_no);
			sca_play_barge_warn(user);
		}
	}

	/* Initialize bridge features for the channel. */
	if (ast_bridge_features_init(&chan_features)) {
		ast_log(LOG_WARNING, "Unable to init bridge feature\n");
		goto outrun;
	}

	/* set call duration limit */
	if ((bridge->user_count == 2) && (bridge->duration > 0) && (bridge->endtime == 0)) {
		struct ast_bridge_features_limits call_limits;

		now = ast_tvnow();
		bridge->endtime = now.tv_sec + bridge->duration;

		ast_bridge_features_limits_construct(&call_limits);

		call_limits.duration = bridge->duration;
		if (user->marked && (bridge->limit_warning > 0)) {
			call_limits.frequency = bridge->limit_freq;
			call_limits.warning = bridge->limit_warning;
			ast_string_field_set(&call_limits, warning_sound, "timeleft");
		}

		ast_bridge_features_set_limits(&chan_features, &call_limits, 0); 
		ast_bridge_features_limits_destroy(&call_limits);
	}

	ast_bridge_hold_hook(&chan_features, sca_handle_hold_cb, user, sca_hold_hook_destructor, AST_BRIDGE_HOOK_REMOVE_ON_PULL);

	ast_mutex_lock(&bridge->lock);
	if (bridge->exit_flag) {
		ast_log(LOG_NOTICE, "bridge has set exit flag, user[%d] run out\n", user->user_no);
		goto outrun;
	}

	if (ast_channel_is_avs_engine(chan) && (bridge->bridge->avs_bridge_id <= 0)) {
		int bridge_id = -1;

		if (line->custom_record) {
			bridge_id = ast_avs_create_bridge(avs_bridge_p2p, bridge->bridge, -1, 0, ast_channel_get_avs_address(chan));
		} else {
			bridge_id = ast_avs_create_bridge(avs_bridge_conf, bridge->bridge, -1, 0, ast_channel_get_avs_address(chan));
		}

		if (bridge_id > 0) {
			bridge->bridge->avs_bridge_id = bridge_id;
			bridge->bridge->avs_bridge_type = avs_bridge_conf;
			if (bridge->bridge->technology && bridge->bridge->technology->stop) {
				ast_bridge_lock(bridge->bridge);
				bridge->bridge->technology->stop(bridge->bridge);
				ast_bridge_unlock(bridge->bridge);
			}
		}
	}
	ast_mutex_unlock(&bridge->lock);

	if (line->record_flag) {
		const char *have_prompt = pbx_builtin_getvar_helper(chan, "HAVE_PROMPTED");
		ast_mutex_lock(&bridge->lock);
		sca_bridge_start_record(bridge, chan, line);
		ast_mutex_unlock(&bridge->lock);

		/* to play record prompt tone */
		if (sca_record_prompt && bridge->recordingfilename
			&& (ast_strlen_zero(have_prompt) || strcmp(have_prompt, "yes"))) {
			sca_prompt_record(chan);
		}
	}

	if (bridge->exit_flag) {
		ast_log(LOG_NOTICE, "bridge has set exit flag, user[%d] run out\n", user->user_no);
		goto outrun;
	}

	if (user->marked) {
		sca_set_local_flag(chan, bridge->uniqueid);
	}

	ast_log(LOG_NOTICE, "sca_bridge_run, chan[%s], user[%d][%p], join bridge.\n", ast_channel_name(chan), user->user_no, user);
	ast_bridge_join(bridge->bridge, chan, NULL, &chan_features, NULL, 0);

outrun:
	AST_LIST_LOCK(&sca_bridges);

	ao2_lock(bridge->usercontainer);
	bridge->user_count--;
	if (user->marked) {
		bridge->markedusers--;

		if (bridge->markedusers == 0) {
			bridge->exit_flag = 1;
			ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
		}
		ast_log(LOG_NOTICE, "Marked user leave, [%d] markedusers in bridge now\n", bridge->markedusers);
	}
	ao2_unlock(bridge->usercontainer);

	if (bridge->user_count == 2) {
		if (line->status == SCA_Bridge_Active) {
			sca_change_line_state(line->shared, line, SCA_Active); 
		} else if (line->status == SCA_Bridge_Held) {
			if (line->heldby_user == user) {
				ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
			} else {
				sca_change_line_state(line->shared, line, SCA_Held); 
			}
		} 
	}

	/* report AMI Event: BridgeLeave */
	sca_bridge_leave_report(bridge, chan);
	ast_log(LOG_NOTICE, "user_no[%d] leave the bridge, [%d] users in bridge now\n", user->user_no, bridge->user_count);

	user->destroy = 1;
	if (-1 != user->sched_id) {
		AST_SCHED_DEL(sca_sched, user->sched_id); 
		user->sched_id = -1;
		ao2_ref(user, -1);
	}

	/* Remove ourselves from the container */
	ao2_unlink(bridge->usercontainer, user); 
	ao2_ref(user, -1);

	AST_LIST_UNLOCK(&sca_bridges);
	ast_bridge_features_cleanup(&chan_features);

	return 0;
}


static void *sca_run_user(void *data)
{
	RAII_VAR(struct sca_user *, user, NULL, ao2_cleanup);
	RAII_VAR(struct sca_shared *, shared, NULL, ao2_cleanup);//edit Bug 278689 
	RAII_VAR(struct sca_line *, line, NULL, ao2_cleanup);//edit Bug 278689 

	struct ast_str    *bridge_name = ast_str_create(80);
	struct ast_flags64 flags = { 0 };
	struct sca_bridge *bridge;
	struct run_user_args *args = data;

	user = args->user;

	ast_mutex_lock(args->cond_lock);
	ast_cond_signal(args->cond);
	ast_mutex_unlock(args->cond_lock);

	if (!user || !user->shared || !user->line || !bridge_name) {
		ast_log(LOG_NOTICE, "sca_run_user, invalid para, user[%s]", user?user->name:"null");
		if (bridge_name){
			ast_free(bridge_name);
			bridge_name = NULL;
		}
		return NULL;
	}

	shared = ao2_bump(user->shared);//edit Bug 278689 
	line = ao2_bump(user->line);//edit Bug 278689 
	if (line->status == SCA_Idle) {
		ast_log(LOG_NOTICE, "sca_run_user, line is Idle, ignore");
		ast_free(bridge_name);
		bridge_name = NULL;
		return NULL;
	}

	ast_atomic_fetchadd_int((int *) &line->active_count, 1);
	ast_str_set(&bridge_name, 0, "SCA_%s_%d", shared->name, line->index);

	if (line->record_flag) {
		ast_set_flag64(&flags, FLAG_MARKEDEXIT | FLAG_RECORD); 
	} else {
		ast_set_flag64(&flags, FLAG_MARKEDEXIT); 
	}

	ast_log(LOG_NOTICE, "sca_run_user, user[%s], answer local_chan[%s], peer_chan[%s]",
				user->name, ast_channel_name(user->local_chan), user->peer_chan ? ast_channel_name(user->peer_chan):"");
	/*BEGIN: Add by weidu ,for Bug 159519, fxo hangup*/
	if (!user->peer_chan || ast_check_hangup(user->peer_chan)){
		ast_log(LOG_DEBUG, "sca_run_user caller have hangup \n");
		sca_change_line_state(shared, line, SCA_Idle);
		goto EXIT;
	}
	/*END: Add by weidu ,for Bug 159519*/
	sca_answer_chan(user->local_chan);

	if (user->peer_chan) {
		char bridge_peer[256] = {0};

		snprintf(bridge_peer, sizeof(bridge_peer)-1, "%s", ast_channel_name(user->local_chan));
		pbx_builtin_setvar_helper(user->peer_chan, "BRIDGEPEER", bridge_peer);
		pbx_builtin_setvar_helper(user->local_chan, "BRIDGEPEER", ast_channel_name(user->peer_chan));
	}
	ast_str_set(&bridge_name, 0, "SCA_%s_%d", shared->name, line->index);
	bridge = sca_build_bridge(ast_str_buffer(bridge_name), 1, user->local_chan);
	if (bridge) {
		if (!bridge->exit_flag) {
			sca_bridge_run(user->local_chan, bridge, line, user, &flags);
		}
		if (ast_atomic_dec_and_test((int *) &line->active_count)) {
			ao2_callback(bridge->usercontainer, OBJ_NODATA, sca_bridge_user_set_exit, NULL);
			sca_change_line_state(shared, line, SCA_Idle);
		}

		sca_dispose_bridge(bridge);
		bridge = NULL;
	} else if (ast_atomic_dec_and_test((int *) &line->active_count)) {
		ast_atomic_dec_and_test((int *) &line->active_count);
		sca_change_line_state(shared, line, SCA_Idle);
	}

EXIT:
	if (user->dial) {
		ast_dial_join(user->dial); 
		ast_dial_destroy(user->dial);
	}

	user->local_chan = NULL;
	user->dial = NULL;

	if (user->line) {
		ast_atomic_fetchadd_int((int *)&user->line->related_count, -1);
		ao2_ref(user->line, -1); 
		user->line = NULL;
	}

	ast_free(bridge_name);

	return NULL;
}

/*!
 * handle the dial state event of the dialed user
 */
static void sca_handle_dial_state_event(void)
{
	struct sca_ringing_user  *ringing_user;
    struct sca_user *user;

	AST_LIST_LOCK(&sca.ringing_users);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&sca.ringing_users, ringing_user, entry) {
		struct run_user_args args;
		enum ast_dial_result dial_res;
		pthread_t dont_care;
		ast_mutex_t cond_lock;
		ast_cond_t cond;

		user = ringing_user->user;
		if (!user || !user->dial) {
			AST_LIST_REMOVE_CURRENT(entry);
			sca.ringing_users_num--; 

            sca_ringing_user_destroy(ringing_user);
            continue;
        }

        dial_res = ast_dial_state(user->dial); 

		switch (dial_res) {
		case AST_DIAL_RESULT_HANGUP:
		case AST_DIAL_RESULT_INVALID:
		case AST_DIAL_RESULT_FAILED:
		case AST_DIAL_RESULT_TIMEOUT:
		case AST_DIAL_RESULT_PEER_BUSY:
		case AST_DIAL_RESULT_UNANSWERED:
		{
			/* Add by bfdu for bug 102298, 2018.02.01 */
			if (user->peer_chan && (dial_res != AST_DIAL_RESULT_HANGUP)) {
				ast_log(LOG_NOTICE, "sca_handle_dial_state_event, set SCA_DIAL_STATUS to UNANSWERED, sca user[%s], user->peer_chan[%p], dial_res[%d]", user->name, user->peer_chan, dial_res);
				pbx_builtin_setvar_helper(user->peer_chan, "SCA_DIAL_STATUS", "UNANSWERED");

				if ((dial_res == AST_DIAL_RESULT_TIMEOUT) || (dial_res == AST_DIAL_RESULT_UNANSWERED)) {
					pbx_builtin_setvar_helper(user->peer_chan, "SCA_CALL_STATUS", "NOANSWER");
				}
			}
			/* End */

			user->peer_chan = NULL;
			AST_LIST_REMOVE_CURRENT(entry);
			sca_stop_ringing_user(ringing_user);
			sca.ringing_users_num--; 
			break;
		}

		case AST_DIAL_RESULT_ANSWERED:
		{
			AST_LIST_REMOVE_CURRENT(entry);
			sca.ringing_users_num--; 

			/* Track the channel that answered the peer */
			user->local_chan = ast_dial_answered(user->dial);

			/* Add by bfdu for bug 102298, 2018.02.01 */
			if (user->peer_chan) {
				ast_log(LOG_NOTICE, "sca_handle_dial_state_event, set SCA_DIAL_STATUS to ANSWERED, sca user[%s], user->peer_chan[%p]", user->name, user->peer_chan);
				pbx_builtin_setvar_helper(user->peer_chan, "SCA_DIAL_STATUS", "ANSWERED");
				ast_channel_set_peer_is_sca(user->peer_chan, 1);
				ast_channel_set_peer_sca_name(user->peer_chan, ast_channel_name(user->local_chan));

				/* added by zgmo for Bug179232 */
				if (ast_channel_get_turn_session_count(user->local_chan)) {
					ast_indicate(user->peer_chan, AST_CONTROL_PEER_IS_NAT_CALL);
				}
				/* end */
			}
			/* End */

			/* Actually answer the peer */
			sca_answer_chan(user->peer_chan);
#ifdef GRANDSTREAM_NETWORKS
			if (user->peer_chan && ast_channel_tech(user->peer_chan)->type && !strcasecmp(ast_channel_tech(user->peer_chan)->type, "DAHDI")){
				pbx_builtin_setvar_helper(user->peer_chan, "CALLED_REALHOOK", "yes");
			}
#endif
			/* Line state ==> Active */
			if (user->line && user->shared) {
				sca_line_status_set((char *)user->shared->name, user->line->index, SCA_Active);
			}

			/* Now, start a thread that will connect this sca user to the peer.  The rest of
			 * the code here sets up the thread and ensures that it is able to save the arguments
			 * before they are no longer valid since they are allocated on the stack. */
			args.user = ao2_bump(user);
			args.cond = &cond;
			args.cond_lock = &cond_lock;
			sca_ringing_user_destroy(ringing_user);

			ast_mutex_init(&cond_lock);
			ast_cond_init(&cond, NULL);
			ast_mutex_lock(&cond_lock);
			ast_pthread_create_detached_background(&dont_care, NULL, sca_run_user, &args);
			ast_cond_wait(&cond, &cond_lock);
			ast_mutex_unlock(&cond_lock);
			ast_mutex_destroy(&cond_lock);
			ast_cond_destroy(&cond);
			break;
        }
		case AST_DIAL_RESULT_TRYING:
		case AST_DIAL_RESULT_RINGING:
		case AST_DIAL_RESULT_PROGRESS:
		case AST_DIAL_RESULT_PROCEEDING:
			break;
		}

		if (dial_res == AST_DIAL_RESULT_ANSWERED) {
			/* Queue up reprocessing ringing users */
			ast_log(LOG_NOTICE, "sca_handle_dial_state_event, sca user[%s] answered, hangup other ringing user", user->name);

            //sca_hangup_users(user->shared, user->line);
            sca_queue_event_full(SCA_EVENT_HANGUP_RINGING_USERS, user->shared, user->line, 1); 
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&sca.ringing_users);
}


/*! \brief Check to see if this user is already ringing 
 * \note Assumes sca.lock is locked 
 */
static int sca_check_ringing_user(const struct sca_user *user)
{
	struct sca_ringing_user  *ringing_user;

	AST_LIST_LOCK(&sca.ringing_users);
	AST_LIST_TRAVERSE(&sca.ringing_users, ringing_user, entry) {
		if (user == ringing_user->user) {
			AST_LIST_UNLOCK(&sca.ringing_users);
			return 1;
		}
	}
	AST_LIST_UNLOCK(&sca.ringing_users);

	return 0;
}


/*! \brief Check to see if a user is in use
 */
static int sca_check_inuse_user(const struct sca_user *user)
{
    if (!user) {
        return 0;
    }

    if (user->dial || user->line) {
        return 1;
    }

    //any other condition ?

	return 0;
}


static struct sca_ringing_user *sca_create_ringing_user(struct sca_user *user)
{
	struct sca_ringing_user *ringing_user;

	if (!(ringing_user = ast_calloc(1, sizeof(*ringing_user)))) {
		return NULL;
	}

	ao2_ref(user, 1);
	ringing_user->user = user;

	return ringing_user;
}

#if 0
static void sca_set_ring_timeout_value(struct ast_dial *dial, struct sca_user *user)
{
	char family[80]= {0};
	char data[16] = {0};
	int  timeout = 0;

	if (!dial || !user || !user->shared || !user->shared->name) {
		ast_log(LOG_WARNING, "sca_set_timeout_value param is NULL\n");
		return;
	}

	snprintf(family, sizeof(family), "AAUSER/%s", user->shared->name);
	ast_ucm_db_get(family, "RT", data, sizeof(data));
	if (strlen(data) != 0) {
		timeout = atoi(data);
		ast_log(LOG_DEBUG, "Get sca shared account:%s configure, set user:%s timeout value:%d\n", user->shared->name, user->name, timeout);
		ast_dial_set_global_timeout(dial, timeout*1000);
	}

	return;
}
#endif

/*! \brief Ring a user
 * \note Assumes sca.lock is locked
 */
static int sca_ring_user(struct sca_user *user, struct sca_line *line)
{
	struct ast_dial *dial;
	struct sca_ringing_user *ringing_user;
	enum ast_dial_result res;

	if (!user) {
		return -1;
	}

	if (!(dial = ast_dial_create())) {
		return -1;
	}

	pbx_builtin_setvar_helper(user->peer_chan, "FROM_SCA", "yes");

#if 0   /*for bug 107574 */
	/* Add by bfdu for Bug 93221, 2018.2.11 */
	if (line->ringtime > 0) {
		ast_dial_set_global_timeout(dial, line->ringtime*1000);
	}
#endif
	ast_dial_set_state_callback(dial, sca_dial_state_callback);
	ast_dial_set_user_data(dial, user);

	if (ast_dial_append(dial, "PJSIP", user->name, NULL) == -1) {
		ast_dial_destroy(dial);
		return -1;
	}

	/* Do we need to save off the caller ID data? */
	//caller = *ast_channel_caller(user->peer_chan);
	//ast_party_caller_init(ast_channel_caller(user->peer_chan));

	/*Set jitter buffer*/
	ast_dial_option_global_enable(dial, AST_DIAL_OPTION_PREDIAL, "callee-handler^s^1");
	res = ast_dial_run(dial, user->peer_chan, 1);
	
	/* Restore saved caller ID */
	//ast_party_caller_free(ast_channel_caller(user->peer_chan));
	//ast_channel_caller_set(user->peer_chan, &caller);
	
	if (res != AST_DIAL_RESULT_TRYING) {
        ast_log(LOG_WARNING, "sca_ring_user, ast_dial_run() failed, user[%s], res[%d]",user->name, res);
		ast_dial_destroy(dial);
		return -1;
	}

	if (!(ringing_user = sca_create_ringing_user(user))) {
		ast_dial_join(dial);
		ast_dial_destroy(dial);
		return -1;
	}

	user->dial = dial;
	ao2_lock(user);
	if (!user->line && line) {
		struct sca_user_ref *user_ref = NULL;

		user->line = ao2_bump(line);
		ast_atomic_fetchadd_int((int *)&line->related_count, 1);
		ao2_unlock(user);

		user_ref = ao2_alloc(sizeof(*user_ref), sca_user_ref_destructor); 
		if (user_ref) {
			user_ref->user = ao2_bump(user);
			user_ref->bridge_user = NULL;

			AST_LIST_LOCK(&line->user_refs);
			AST_LIST_INSERT_TAIL(&line->user_refs, user_ref, ref_entry);
			AST_LIST_UNLOCK(&line->user_refs);

			sca_update_local_user(user->shared, line);
		}

		ao2_lock(user);
	}
	ao2_unlock(user);

	AST_LIST_LOCK(&sca.ringing_users);
	AST_LIST_INSERT_HEAD(&sca.ringing_users, ringing_user, entry);
	sca.ringing_users_num++; 
	AST_LIST_UNLOCK(&sca.ringing_users);

	return 0;
}

/*! \brief Ring users based on current set of ringing users
 * \note Assumes that sca.lock is locked
 */
static int sca_ring_users(struct sca_shared *shared, struct sca_line *line)
{
	struct sca_user  *user;
	int    ret  = -1;
	int    busy_count = 0;
	int    unavail_count = 0;
	const char *answered_user = NULL;

	if (!shared || !line) {
		return ret;
	}

	ao2_ref(line, 1);
	AST_LIST_LOCK(&shared->users);
	if(line->remote_chan && ast_test_flag(ast_channel_flags(line->remote_chan), AST_FLAG_RECOVERYING))
	{
		answered_user = pbx_builtin_getvar_helper(line->remote_chan, "SCA_ANSWERED_USER");
		ast_cdr_setvar(ast_channel_name(line->remote_chan),"action_type","SCA");
	}

	/* Make sure that every user related to the shared line, is ringing. */
	AST_LIST_TRAVERSE_SAFE_BEGIN(&shared->users, user, entry) {
		if (!user->as_callee) {
			ast_log(LOG_NOTICE, "sca_ring_users, user[%s] can't be a callee", user->name);
			continue;
		}

		/* Is this user already ringing? */
		if (sca_check_ringing_user(user)) {
			ast_log(LOG_NOTICE, "sca_ring_users, user[%s] is ringing", user->name);
			busy_count++;
			continue;
        }

		/* Is this user already in a call? */
		if (sca_check_inuse_user(user)) {
			ast_log(LOG_NOTICE, "sca_ring_users, user[%s] is inuse", user->name);
			busy_count++;
			continue;
		}

		if(answered_user) {
			if(strcmp(user->name,answered_user)) {
				continue;
			}
		}

		ast_log(LOG_NOTICE, "sca_ring_users, ring user[%s], line[%d]\n", user->name, line->index);
		user->peer_chan = line->remote_chan;

		/* It is time to make this user begin to ring.  Do it! */
		if (!sca_ring_user(user, line)) {
			ret = 0;
		} else {
			unavail_count++;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END

	AST_LIST_UNLOCK(&shared->users);

	if (ret && line->remote_chan) {
		if (busy_count > 0) {
			pbx_builtin_setvar_helper(line->remote_chan, "SCA_CALL_STATUS", "BUSY");
		} else if (unavail_count > 0) {
			pbx_builtin_setvar_helper(line->remote_chan, "SCA_CALL_STATUS", "UNREACHABLE");
		}
	}

	ao2_ref(line, -1);
	return ret;
}

static void sca_hangup_users(struct sca_shared *shared, struct sca_line *line, int flag)
{
	struct sca_user  *user;
	struct sca_ringing_user *ringing_user;

	ast_log(LOG_DEBUG, "sca_hangup_users entering, flag[%d]\n",flag);
	AST_LIST_LOCK(&sca.ringing_users);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&sca.ringing_users, ringing_user, entry) {
        user = ringing_user->user;

        if (!user || !user->line) {
            continue;
        }

        if (user->line == line) {
			ast_atomic_fetchadd_int((int *)&line->related_count, -1);
			ao2_ref(user->line, -1); 
			user->line = NULL;

			AST_LIST_REMOVE_CURRENT(entry);
			ast_set_answered_elsewhere(user->dial, flag);
			ast_dial_join(user->dial);
			ast_dial_destroy(user->dial);
			user->dial = NULL;
			user->peer_chan = NULL;
			sca_ringing_user_destroy(ringing_user);
			sca.ringing_users_num--;  
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&sca.ringing_users);
}

static void sca_remove_ringing_user(struct sca_user *user)
{
	struct sca_ringing_user *ringing_user = NULL;

	if (!user) {
		return;
	}

	AST_LIST_LOCK(&sca.ringing_users);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&sca.ringing_users, ringing_user, entry) {
		if (ringing_user->user != user) {
			continue;
		}

		ast_log(LOG_DEBUG, "Remove ringing user:%s", user->name);
		AST_LIST_REMOVE_CURRENT(entry);
		sca.ringing_users_num--;
		user->peer_chan = NULL;
		if (user->dial) {
			sca_stop_ringing_user(ringing_user);
		} else {
			sca_ringing_user_destroy(ringing_user);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&sca.ringing_users);
}

static void sca_handle_ringing_user_event(struct sca_shared *shared, struct sca_line *line)
{
	ast_mutex_lock(&sca.lock);
    ast_log(LOG_NOTICE, "sca_handle_ringing_user_event, CALL sca_ring_users");
    sca_ring_users(shared, line);
	ast_mutex_unlock(&sca.lock);
}

/* to check if there is any sharedline object */
static int sca_in_use(void)
{
	return ao2_container_count(con_sca_users) || ao2_container_count(con_sca_shareds);
}

static void sca_queue_event_full(enum sca_event_type type, struct sca_shared *shared,  struct sca_line *line, int lock)
{
	struct sca_event *event;

	if (sca.thread == AST_PTHREADT_NULL) {
		return;
	}

	if (!(event = ast_calloc(1, sizeof(*event)))) {
		return;
	}

	event->shared = NULL;
	event->line   = NULL;
	event->type   = type;

    if (shared) {
        event->shared = ao2_bump(shared); 
    }

    if (line) {
        event->line = ao2_bump(line); 
    }

	if (!lock) {
		AST_LIST_INSERT_TAIL(&sca.event_q, event, entry);
		sca.event_q_num++;  
		return;
	}

	ast_mutex_lock(&sca.lock);
	AST_LIST_INSERT_TAIL(&sca.event_q, event, entry);
	sca.event_q_num++;  
	ast_cond_signal(&sca.cond);
	ast_mutex_unlock(&sca.lock);
}

static void sca_event_destroy(struct sca_event *event)
{
	if (event->shared) {
		ao2_ref(event->shared, -1);
		event->shared = NULL;
	}

	if (event->line) {
		ao2_ref(event->line, -1);
		event->line = NULL;
	}

	ast_free(event);
}

static void sca_dial_state_callback(struct ast_dial *dial)
{
    struct sca_user  *user;

	user = (struct sca_user *)ast_dial_get_user_data(dial);
    if (user) {
        sca_queue_event_full(SCA_EVENT_DIAL_STATE, user->shared, user->line, 1); 
    }
}

static void *sca_thread(void *data)
{
    ast_mutex_lock(&sca.lock);

	while (!sca.stop) {
		struct sca_event *event;

		if (AST_LIST_EMPTY(&sca.event_q)) {
			ast_cond_wait(&sca.cond, &sca.lock);

			if (sca.stop) {
				break;
            }
		}

		while ((event = AST_LIST_REMOVE_HEAD(&sca.event_q, entry))) {
			sca.event_q_num--;  

			ast_mutex_unlock(&sca.lock);
			switch (event->type) {
			case SCA_EVENT_DIAL_STATE:
				sca_handle_dial_state_event();
				break;
			case SCA_EVENT_RINGING_USER:
				sca_handle_ringing_user_event(event->shared, event->line);
				break;
			case SCA_EVENT_HANGUP_RINGING_USERS:
				sca_hangup_users(event->shared, event->line, 1);
				break;
			default:
				break;
			}

			sca_event_destroy(event);
			ast_mutex_lock(&sca.lock);
		}
	}

	ast_mutex_unlock(&sca.lock);
	return NULL;
}


static void sca_destroy(void)
{
	if (sca.thread != AST_PTHREADT_NULL) {
		ast_mutex_lock(&sca.lock);
		sca.stop = 1;
		ast_cond_signal(&sca.cond);
		ast_mutex_unlock(&sca.lock);
		pthread_join(sca.thread, NULL);
	}

	ast_mutex_destroy(&sca.lock);
	ast_cond_destroy(&sca.cond);

	ao2_callback(con_sca_shareds, 0, sca_shared_release_list, NULL);
	ao2_callback(con_sca_users, 0, NULL, NULL);

	ao2_ref(con_sca_shareds, -1);
	con_sca_shareds = NULL;

	ao2_ref(con_sca_users, -1);
	con_sca_users = NULL;
}

#define ____SCA_LOAD_UNLOAD____

static int load_config(int reload)
{
	struct ast_config *cfg;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	const char *cat = NULL;
	const char *val = NULL;
	int res = 0;

	if (!reload) {
		ast_mutex_init(&sca.lock);
		ast_cond_init(&sca.cond, NULL);
		con_sca_shareds = ao2_container_alloc(1, sca_shared_hash, sca_shared_cmp);
		con_sca_users = ao2_container_alloc(1, sca_user_hash, sca_user_cmp);
	}

	if (!(cfg = ast_config_load(SCA_CONFIG_FILE, config_flags))) {
		/*! Treat no config as normal */
		goto auto_rec;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		goto auto_rec;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file " SCA_CONFIG_FILE " is in an invalid format.  Aborting.\n");
		goto auto_rec;
	}

	if (reload) {
		ao2_callback(con_sca_shareds,  0, sca_shared_mark, NULL);
		ao2_callback(con_sca_users, 0, sca_user_mark, NULL);
	}

	if ((val = ast_variable_retrieve(cfg, "general", "record_prompt"))){
		sca_record_prompt = ast_true(val);
	}

	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		const char *type;
		if (!strcasecmp(cat, "general"))
			continue;
		if (!(type = ast_variable_retrieve(cfg, cat, "type"))) {
			ast_log(LOG_WARNING, "Invalid entry in %s defined with no type!\n", SCA_CONFIG_FILE);
			continue;
		}

		if (!strcasecmp(type, "shared"))
			res = sca_build_shared(cfg, cat);
		else if (!strcasecmp(type, "private"))
			res = sca_build_user(cfg, cat);
		else {
			ast_log(LOG_WARNING, "Entry in %s defined with invalid type '%s'!\n",
				SCA_CONFIG_FILE, type);
		}
	}

	ast_config_destroy(cfg);

	if (reload) {
		ao2_callback(con_sca_shareds, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, sca_shared_is_marked, NULL);
		ao2_callback(con_sca_users, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, sca_user_is_marked, NULL);
	}

	/* Start SCA event processing thread once SCA has been configured. */
	if (sca.thread == AST_PTHREADT_NULL && sca_in_use()) {
		ast_pthread_create(&sca.thread, NULL, sca_thread, NULL);
	}


auto_rec:
	cfg = NULL;
	cat = NULL;
	struct ast_variable* var = NULL;
	struct auto_record_node* node = NULL;

	if (!(cfg = ast_config_load(AUTO_RECORD_FILE, config_flags))) {
		ast_log(LOG_ERROR, "load config file met error: %s not exist\n", AUTO_RECORD_FILE);
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_log(LOG_ERROR, "load config file %s not change\n", AUTO_RECORD_FILE);
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "load config file met error: %s invalid\n", AUTO_RECORD_FILE);
		return 0;
	}

	AST_LIST_LOCK(&record_lists);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&record_lists, node, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(node);
	}
	AST_LIST_TRAVERSE_SAFE_END

	AST_LIST_HEAD_INIT(&record_lists);

	cat = ast_category_browse(cfg, NULL);

	while (cat) {
		if (!strcasecmp(cat, "extension") || !strcasecmp(cat, "trunk") 
			|| !strcasecmp(cat, "route") || !strcasecmp(cat, "ivr")) {
			var = ast_variable_browse(cfg, cat);
			
			while (var) {
				if (!var->value || !strncmp(var->value, "no", 2) || !strncmp(var->value, "off", 3)) {
					var = var->next;
					continue;
				}

				struct auto_record_node *node = (struct auto_record_node *)ast_calloc(1, sizeof(struct auto_record_node));
				if (node == NULL) {
					continue;
				}

				strncpy(node->extensions, var->name, sizeof(node->extensions) - 1);
				strncpy(node->enable, var->value, sizeof(node->enable) - 1);

				if (!strcasecmp(cat, "extension")) {
					node->type = SCA_RECORD_EXTENSION;
				} else if (!strcasecmp(cat, "trunk")) {
					node->type = SCA_RECORD_TRUNK;
				} else if (!strcasecmp(cat, "ivr")) {
					node->type = SCA_RECORD_IVR;
				} else if (!strcasecmp(cat, "route")) {
					node->type = SCA_RECORD_ROUTE;
				}else {
					ast_free(node);
					continue;
				}

				AST_LIST_INSERT_TAIL(&record_lists, node, entry);
				var = var->next;
			}
		}

		cat = ast_category_browse(cfg, cat);
	}

	AST_LIST_UNLOCK(&record_lists);

	ast_config_destroy(cfg);
	cfg = NULL;

	return res;
}

static int load_module(void)
{
	int res = 0;

	res |= load_config(0);

	ast_cli_register_multiple(cli_sca, ARRAY_LEN(cli_sca));

	res |= ast_register_application_xml(sca_app, sca_exec);

	res |= ast_sip_session_register_sca_handler(&sca_handler);

	res |= ast_devstate_prov_add("SCA", sca_state);

	res |= ast_custom_function_register(&sca_status_func);
	res |= ast_custom_function_register(&sca_shared_func);

	sca_sched = ast_sched_context_create();
	if (sca_sched) {
		if (ast_sched_start_thread(sca_sched)) {
			ast_log(LOG_ERROR, "Could not start sca scheduler thread\n");
			return AST_MODULE_LOAD_FAILURE;
		}
	}

	return res;
}

static int reload(void)
{
	return load_config(1);
}


static int unload_module(void)
{
	int res = 0;

	struct auto_record_node* node = NULL;

	AST_LIST_LOCK(&record_lists);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&record_lists, node, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(node);
	}
	AST_LIST_TRAVERSE_SAFE_END

	AST_LIST_UNLOCK(&record_lists);

	ast_cli_unregister_multiple(cli_sca, ARRAY_LEN(cli_sca));
	res |= ast_unregister_application(sca_app);
	res |= ast_custom_function_unregister(&sca_status_func);
	res |= ast_custom_function_unregister(&sca_shared_func);

	ast_devstate_prov_del("SCA");
	
	if (sca_sched) {
		ast_sched_context_destroy(sca_sched);
		sca_sched = NULL;
	}

	sca_destroy();

	return res;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "PJSIP SCA Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_APP_DEPEND,
		.requires = "res_pjsip,res_pjsip_session",
);
