/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2007, Digium, Inc.
 *
 * SLA Implementation by:
 * Russell Bryant <russell@digium.com>
 *
 * Modified by:
 * Yang Yanfei <yfyang@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Shared Line Appearances
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<depend>res_pjsip</depend>
	<depend>app_confbridge</depend>
	<depend>res_rtp_avs_engine</depend>
	<support_level>core</support_level>
***/

#include "asterisk.h"

#include <dahdi/user.h>

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/app.h"
#include "asterisk/dsp.h"
#include "asterisk/musiconhold.h"
#include "asterisk/manager.h"
#include "asterisk/cli.h"
#include "asterisk/say.h"
#include "asterisk/utils.h"
#include "asterisk/astobj2.h"
#include "asterisk/devicestate.h"
#include "asterisk/dial.h"
#include "asterisk/causes.h"
#include "asterisk/test.h"
#include "asterisk/paths.h"
#include "asterisk/format_compatibility.h"
#include "asterisk/format_cap.h"
#include "asterisk/format_cache.h"
#include "asterisk/bridge.h"
#include "asterisk/bridge_features.h"
#include "asterisk/avs_controller.h"
#include "asterisk/haserver.h"

AST_MODULE_LOG("sla");
/*** DOCUMENTATION
	<application name="SLAStation" language="en_US">
		<synopsis>
			Shared Line Appearance Station.
		</synopsis>
		<syntax>
			<parameter name="station" required="true">
				<para>Station name</para>
			</parameter>
		</syntax>
		<description>
			<para>This application should be executed by an SLA station. The argument depends
			on how the call was initiated. If the phone was just taken off hook, then the argument
			<replaceable>station</replaceable> should be just the station name. If the call was
			initiated by pressing a line key, then the station name should be preceded by an underscore
			and the trunk name associated with that line button.</para>
			<para>For example: <literal>station1_line1</literal></para>
			<para>On exit, this application will set the variable <variable>SLASTATION_STATUS</variable> to
			one of the following values:</para>
			<variablelist>
				<variable name="SLASTATION_STATUS">
					<value name="FAILURE" />
					<value name="CONGESTION" />
					<value name="SUCCESS" />
				</variable>
			</variablelist>
		</description>
	</application>
	<application name="SLATrunk" language="en_US">
		<synopsis>
			Shared Line Appearance Trunk.
		</synopsis>
		<syntax>
			<parameter name="trunk" required="true">
				<para>Trunk name</para>
			</parameter>
			<parameter name="options">
				<optionlist>
					<option name="M" hasparams="optional">
						<para>Play back the specified MOH <replaceable>class</replaceable>
						instead of ringing</para>
						<argument name="class" required="true" />
					</option>
				</optionlist>
			</parameter>
		</syntax>
		<description>
			<para>This application should be executed by an SLA trunk on an inbound call. The channel calling
			this application should correspond to the SLA trunk with the name <replaceable>trunk</replaceable>
			that is being passed as an argument.</para>
			<para>On exit, this application will set the variable <variable>SLATRUNK_STATUS</variable> to
			one of the following values:</para>
			<variablelist>
				<variable name="SLATRUNK_STATUS">
					<value name="FAILURE" />
					<value name="SUCCESS" />
					<value name="UNANSWERED" />
					<value name="RINGTIMEOUT" />
				</variable>
			</variablelist>
		</description>
	</application>
 ***/

#define SLA_CONFIG_FILE		"sla.conf"


enum {
	/*! If set, the channel will leave the conference if all marked users leave */
	SLA_FLAG_MARKEDEXIT = (1 << 1),
	/*! If set, the user will be marked */
	SLA_FLAG_MARKEDUSER = (1 << 2),
	SLA_FLAG_SLA_STATION = (1 << 3),
	SLA_FLAG_SLA_TRUNK = (1 << 4),
	SLA_FLAG_MOH = (1 << 5),
};


/*! \brief Event types that can be queued up for the SLA thread */
enum sla_event_type {
	/*! A station has put the call on hold */
	SLA_EVENT_HOLD,
	/*! The state of a dial has changed */
	SLA_EVENT_DIAL_STATE,
	/*! The state of a ringing trunk has changed */
	SLA_EVENT_RINGING_TRUNK,
};


enum sla_which_trunk_refs {
	ALL_TRUNK_REFS,
	INACTIVE_TRUNK_REFS,
};

enum sla_trunk_state {
	SLA_TRUNK_STATE_IDLE,
	SLA_TRUNK_STATE_RINGING,
	SLA_TRUNK_STATE_UP,
	SLA_TRUNK_STATE_ONHOLD,
	SLA_TRUNK_STATE_ONHOLD_BYME,
	SLA_TRUNK_STATE_UNAVAILABLE,
};

enum sla_hold_access {
	/*! This means that any station can put it on hold, and any station
	 * can retrieve the call from hold. */
	SLA_HOLD_OPEN,
	/*! This means that only the station that put the call on hold may
	 * retrieve it from hold. */
	SLA_HOLD_PRIVATE,
};


enum {
	OPT_ARG_MOH_CLASS = 4,
};


static const char * const slastation_app = "SLAStation";
static const char * const slatrunk_app = "SLATrunk";

#define MAX_CONFNUM 80

/*! \brief The MeetMe Conference object */
struct sla_conf {
	ast_mutex_t lock;                       /*!< SLA Conference specific lock  */
	char conf_name[80];
	char uniqueid[128];
    struct ast_bridge  *bridge;
	struct ast_channel *record_chan;        /*!< Listen/Record channel */
	int user_count;                         /*!< Number of active users */
	int markedusers;                        /*!< Number of marked users */
	int refcount;                           /*!< reference count of usage */
	char *recordingfilename;                /*!< Filename to record the Conference into */
	char *recordingformat;                  /*!< Format to record the Conference in */
	int duration;                           /*!< the duration limit time (seconds) */
	int limit_warning;                      /*!< Duration into the call when warnings should begin (seconds) */
	int limit_freq;                         /*!< Interval between the warnings (seconds) */
	long endtime;                           /*!< When to end the bridge if scheduled */
	struct ao2_container *usercontainer;
	unsigned int record_flag:1;             /*!< to indicate if the bridge need to record */
	unsigned int exit_flag:1;               /*!< to indicate if the bridge need to exit */
	AST_LIST_ENTRY(sla_conf) list;
};

static AST_LIST_HEAD_STATIC(sla_confs, sla_conf);


/*! \brief The SLA User object */
struct sla_conf_user {
	int user_no;                            /*!< User Number */
	struct ast_channel *chan;               /*!< Connected channel */
	struct sla_conf    *conf;
	unsigned int  playing_moh:1;            /*!< MOH is currently being played to the trunk_ref */
	unsigned int  marked:1;                 /*!< to indicate whether it is a marked user or not */
	unsigned int  moh_flag:1;               /*!< to indicate whether to play moh when only one in conf */
	unsigned int  destroy:1;                /*!< to indicate whether the user need to be destroy */

	AST_LIST_ENTRY(sla_conf_user) list;
};
struct sla_trunk_ref;

struct sla_station {
	AST_RWLIST_ENTRY(sla_station) entry;
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
		AST_STRING_FIELD(device);
		AST_STRING_FIELD(autocontext);
	);
	AST_LIST_HEAD_NOLOCK(, sla_trunk_ref) trunks;
	struct ast_dial *dial;
	/*! Ring timeout for this station, for any trunk.  If a ring timeout
	 *  is set for a specific trunk on this station, that will take
	 *  priority over this value. */
	unsigned int ring_timeout;
	/*! Ring delay for this station, for any trunk.  If a ring delay
	 *  is set for a specific trunk on this station, that will take
	 *  priority over this value. */
	unsigned int ring_delay;
	/*! This option uses the values in the sla_hold_access enum and sets the
	 * access control type for hold on this station. */
	unsigned int hold_access:1;
	/*! Mark used during reload processing */
	unsigned int mark:1;
};

/*!
 * \brief A reference to a station
 *
 * This struct looks near useless at first glance.  However, its existence
 * in the list of stations in sla_trunk means that this station references
 * that trunk.  We use the mark to keep track of whether it needs to be
 * removed from the sla_trunk's list of stations during a reload.
 */
struct sla_station_ref {
	AST_LIST_ENTRY(sla_station_ref) entry;
	struct sla_station *station;
	/*! Mark used during reload processing */
	unsigned int mark:1;
};

struct sla_trunk {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
		AST_STRING_FIELD(device);
		AST_STRING_FIELD(autocontext);
	);
	AST_LIST_HEAD_NOLOCK(, sla_station_ref) stations;
	/*! Number of stations that use this trunk */
	unsigned int num_stations;
	/*! Number of stations currently on a call with this trunk */
	unsigned int active_stations;
	/*! Number of stations that have this trunk on hold. */
	unsigned int hold_stations;
	struct ast_channel *chan;
	unsigned int ring_timeout;
	/*! If set to 1, no station will be able to join an active call with
	 *  this trunk. */
	unsigned int barge_disabled:1;
	/*! This option uses the values in the sla_hold_access enum and sets the
	 * access control type for hold on this trunk. */
	unsigned int hold_access:1;
	/*! Whether this trunk is currently on hold, meaning that once a station
	 *  connects to it, the trunk channel needs to have UNHOLD indicated to it. */
	unsigned int on_hold:1;
	unsigned int auto_record:1;  /*! Added by yfyang */
	/*! Mark used during reload processing */
	unsigned int mark:1;
	enum sla_trunk_state trunk_state;  /*!< current trunk state */
};

/*!
 * \brief A station's reference to a trunk
 *
 * An sla_station keeps a list of trunk_refs.  This holds metadata about the
 * stations usage of the trunk.
 */
struct sla_trunk_ref {
	AST_LIST_ENTRY(sla_trunk_ref) entry;
	struct sla_trunk *trunk;
	enum sla_trunk_state state;
	struct ast_channel *chan;
	struct sla_conf    *conf;
	/*! Ring timeout to use when this trunk is ringing on this specific
	 *  station.  This takes higher priority than a ring timeout set at
	 *  the station level. */
	unsigned int ring_timeout;
	/*! Ring delay to use when this trunk is ringing on this specific
	 *  station.  This takes higher priority than a ring delay set at
	 *  the station level. */
	unsigned int ring_delay;
	/*! Mark used during reload processing */
	unsigned int mark:1;
	unsigned int playing_moh:1;            /*!< MOH is currently being played to the trunk_ref */
	unsigned int order:8; 
};

static struct ao2_container *sla_stations;
static struct ao2_container *sla_trunks;

static const char sla_registrar[] = "SLA";

struct sla_event {
	enum sla_event_type type;
	struct sla_station *station;
	struct sla_trunk_ref *trunk_ref;
	AST_LIST_ENTRY(sla_event) entry;
};

/*! \brief A station that failed to be dialed
 * \note Only used by the SLA thread. */
struct sla_failed_station {
	struct sla_station *station;
	struct timeval last_try;
	AST_LIST_ENTRY(sla_failed_station) entry;
};

/*! \brief A trunk that is ringing */
struct sla_ringing_trunk {
	struct sla_trunk *trunk;
	/*! The time that this trunk started ringing */
	struct timeval ring_begin;
	AST_LIST_HEAD_NOLOCK(, sla_station_ref) timed_out_stations;
	AST_LIST_ENTRY(sla_ringing_trunk) entry;
};

enum sla_station_hangup {
	SLA_STATION_HANGUP_NORMAL,
	SLA_STATION_HANGUP_TIMEOUT,
#ifdef GRANDSTREAM_NETWORKS
	SLA_STATION_HANGUP_BUSY,
#endif
};

/*! \brief A station that is ringing */
struct sla_ringing_station {
	struct sla_station *station;
	/*! The time that this station started ringing */
	struct timeval ring_begin;
	AST_LIST_ENTRY(sla_ringing_station) entry;
};

/*!
 * \brief A structure for data used by the sla thread
 */
static struct {
	/*! The SLA thread ID */
	pthread_t thread;
	ast_cond_t cond;
	ast_mutex_t lock;
	AST_LIST_HEAD_NOLOCK(, sla_ringing_trunk) ringing_trunks;
	AST_LIST_HEAD_NOLOCK(, sla_ringing_station) ringing_stations;
	AST_LIST_HEAD_NOLOCK(, sla_failed_station) failed_stations;
	AST_LIST_HEAD_NOLOCK(, sla_event) event_q;
	unsigned int stop:1;
	/*! Attempt to handle CallerID, even though it is known not to work
	 *  properly in some situations. */
	unsigned int attempt_callerid:1;
	unsigned int ringing_trunks_num;
	unsigned int ringing_stations_num;
	unsigned int failed_stations_num;
	unsigned int event_q_num;
} sla = {
	.thread = AST_PTHREADT_NULL,
	.ringing_trunks_num = 0,
	.ringing_stations_num = 0,
	.failed_stations_num = 0,
	.event_q_num = 0,
};

enum {
	SLA_TRUNK_OPT_MOH = (1 << 0),
};

enum {
	SLA_TRUNK_OPT_ARG_MOH_CLASS = 0,
	SLA_TRUNK_OPT_ARG_ARRAY_SIZE = 1,
};

AST_APP_OPTIONS(sla_trunk_opts, BEGIN_OPTIONS
	AST_APP_OPTION_ARG('M', SLA_TRUNK_OPT_MOH, SLA_TRUNK_OPT_ARG_MOH_CLASS),
END_OPTIONS );

static struct stasis_subscription *sla_state_sub;
static int gs_record_prompt = 0;  
static const char *trunkstate2str(enum sla_trunk_state state);
static void sla_change_trunk_state(struct sla_trunk *trunk, enum sla_trunk_state state, 
	enum sla_which_trunk_refs inactive_only, const struct sla_trunk_ref *exclude);

static void sla_user_moh_start(struct sla_conf_user *user);
static void sla_user_moh_stop(struct sla_conf_user *user);
static int sla_conf_stop_record(struct sla_conf *conf);
static int sla_conf_user_no_cmp(void *obj, void *arg, int flags)
{
	struct sla_conf_user *user = obj;
	int *user_no = arg;

	if (user->user_no == *user_no) {
		return (CMP_MATCH | CMP_STOP);
	}

	return 0;
}


/*!
 * \brief Find or create a sla conference
 *
 * \param conf_name The conference name/number
 * \param refcount How many references to mark on the conference
 * \param chan The asterisk channel
 *
 * \return A pointer to the conference struct, or NULL if it wasn't found and
 *         make or dynamic were not set.
 */
static struct sla_conf *sla_conf_build(const char *conf_name, int refcount, const struct ast_channel *chan)
{
	struct sla_conf *cnf;
	time_t creation_time;
	int existed = 0;

	AST_LIST_LOCK(&sla_confs);
	AST_LIST_TRAVERSE(&sla_confs, cnf, list) {
		if (!strcmp(conf_name, cnf->conf_name)) {
			existed = 1;
			break;
		}
	}

	if (existed) {
		ast_atomic_fetchadd_int(&cnf->refcount, refcount);
		AST_LIST_UNLOCK(&sla_confs);
		ast_log(LOG_NOTICE, "sla_conf_build, found existed conf[%s]", conf_name);
#ifdef GRANDSTREAM_NETWORKS
		if (NULL != strcasestr(conf_name, "SLA_")){
			char now_callee[128] = {0};
			char now_num[128] = {0};
			char sla_callee[128] = {0};
			ast_cdr_getvar(ast_channel_name(chan), "app_callee", now_callee, sizeof(now_callee));
			sscanf(now_callee, "%127[^_]%*s", now_num);
			if (!ast_strlen_zero(now_num)){
				snprintf(sla_callee, sizeof(sla_callee), "%s%s", now_num, conf_name+strlen("SLA"));
				ast_cdr_setvar(ast_channel_name(chan), "__app_callee", sla_callee);
			}
		}
#endif
		return cnf;
	}
    AST_LIST_UNLOCK(&sla_confs);

	/* Make a new one */
	cnf = ast_calloc(1, sizeof(*cnf));
	if (!cnf) {
		ast_log(LOG_NOTICE, "sla_conf_build, ast_calloc[%s] failed", conf_name);
		return NULL;
	}
        
	if (!(cnf->usercontainer = ao2_container_alloc(1, NULL, sla_conf_user_no_cmp))) {
		ast_free(cnf);
		ast_log(LOG_NOTICE, "sla_conf_build, usercontainer alloc failed, name[%s]", conf_name);
		return NULL;
	}

	creation_time = time(NULL);
	snprintf(cnf->uniqueid, sizeof(cnf->uniqueid), "%s-%li", conf_name, (long)creation_time);
    ast_copy_string(cnf->conf_name, conf_name, sizeof(cnf->conf_name));

	/* init the conf */
	cnf->user_count    = 0;
	cnf->markedusers   = 0;
	cnf->refcount      = 0;
	cnf->duration      = 0;
	cnf->limit_warning = 0;
	cnf->limit_freq    = 0;
	cnf->exit_flag     = 0;
	cnf->endtime       = 0;

	/* create new bridge */
	cnf->bridge = ast_bridge_base_new(AST_BRIDGE_CAPABILITY_MULTIMIX, AST_BRIDGE_FLAG_MASQUERADE_ONLY | AST_BRIDGE_FLAG_TRANSFER_BRIDGE_ONLY,
									  "SLA", cnf->conf_name, NULL);
	if (!cnf->bridge) {
		ast_log(LOG_WARNING, "sla_conf_build, ast_bridge_base_new() failed for [%s]", cnf->conf_name);
		ao2_ref(cnf->usercontainer, -1);
		ast_free(cnf);
		return NULL;
	}

	cnf->bridge->avs_bridge_id = -1;

	ast_mutex_init(&cnf->lock);
	AST_LIST_LOCK(&sla_confs);
	AST_LIST_INSERT_HEAD(&sla_confs, cnf, list);
	AST_LIST_UNLOCK(&sla_confs);

	ast_log(LOG_NOTICE, "Created SLA conf %p for [%s] by channel [%s]\n", cnf, cnf->conf_name, ast_channel_name(chan));

	ast_atomic_fetchadd_int(&cnf->refcount, refcount);
	return cnf;
}


static void sla_moh_stop(struct sla_trunk_ref *trunk_ref)
{
	int in_bridge;

	if (!trunk_ref || !trunk_ref->trunk || !trunk_ref->trunk->chan || !trunk_ref->playing_moh) {
		return;
	}

	trunk_ref->playing_moh = 0; 

	if (!trunk_ref->conf || !trunk_ref->conf->bridge) {
		ast_log(LOG_NOTICE, "sla_moh_stop, but no conf related!");
		return;
	}

	ast_bridge_lock(trunk_ref->conf->bridge);

	/* Temporarily suspend the user from the bridge so we have control to stop MOH if needed. */
	in_bridge = !ast_bridge_suspend(trunk_ref->conf->bridge, trunk_ref->trunk->chan);

	ast_log(LOG_NOTICE, "sla_moh_stop, chan[%s], in_bridge[%d]", ast_channel_name(trunk_ref->trunk->chan), in_bridge);
	ast_moh_stop(trunk_ref->trunk->chan);

	if (in_bridge) {
		ast_bridge_unsuspend(trunk_ref->conf->bridge, trunk_ref->trunk->chan);
	}

	ast_bridge_unlock(trunk_ref->conf->bridge);
}

/*! \brief Remove the conference from the list and free it.

   We assume that this was called while holding conflock. */
static int sla_conf_free(struct sla_conf *conf, struct sla_trunk_ref *trunk_ref)
{
	if (!conf) {
		return 0;
	}

	ast_log(LOG_NOTICE, "sla_conf_free, conf[%s]", conf->conf_name);

	if (trunk_ref && trunk_ref->playing_moh) {
		sla_moh_stop(trunk_ref);
	}
	
	if (conf->record_chan) {
		sla_conf_stop_record(conf);
		ast_hangup(conf->record_chan);
	}

	if (conf->bridge) {
		ast_bridge_destroy(conf->bridge, 0);
		conf->bridge = NULL;
	}

	if (conf->recordingfilename) {
		ast_free(conf->recordingfilename);
	}

	if (conf->usercontainer) {
		ao2_ref(conf->usercontainer, -1);
	}

	if (conf->recordingformat) {
		ast_free(conf->recordingformat);
	}

	ast_mutex_destroy(&conf->lock);
	ast_free(conf);

	return 0;
}


/*! \brief Decrement reference counts */
static int sla_conf_dispose(struct sla_conf *conf, struct sla_trunk_ref *trunk_ref)
{
	int res = 0;

	ast_log(LOG_NOTICE, "sla_conf_dispose, conf[%s], refcount[%d]", conf->conf_name, conf->refcount);

	AST_LIST_LOCK(&sla_confs);
	if (ast_atomic_dec_and_test(&conf->refcount)) {
		AST_LIST_REMOVE(&sla_confs, conf, list);
		sla_conf_free(conf, trunk_ref);
		res = 1;
	}
	AST_LIST_UNLOCK(&sla_confs);

	return res;
}

static const char *sla_hold_str(unsigned int hold_access)
{
	const char *hold = "Unknown";

	switch (hold_access) {
	case SLA_HOLD_OPEN:
		hold = "Open";
		break;
	case SLA_HOLD_PRIVATE:
		hold = "Private";
	default:
		break;
	}

	return hold;
}

static const char *trunkstate2str(enum sla_trunk_state state)
{
#define S(e) case e: return # e;
	switch (state) {
	S(SLA_TRUNK_STATE_IDLE)
	S(SLA_TRUNK_STATE_RINGING)
	S(SLA_TRUNK_STATE_UP)
	S(SLA_TRUNK_STATE_ONHOLD)
	S(SLA_TRUNK_STATE_ONHOLD_BYME)
	S(SLA_TRUNK_STATE_UNAVAILABLE)
	}
	return "Uknown State";
#undef S
}


static char *sla_show_trunks(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct sla_trunk *trunk;

	switch (cmd) {
	case CLI_INIT:
		e->command = "sla show trunks";
		e->usage =
			"Usage: sla show trunks\n"
			"       This will list all trunks defined in sla.conf\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "==================== SLA Thread Info=====================\n"
	            "=== event queue num:       %d\n"
	            "=== ringing trunks num:    %d\n"
	            "=== ringing stations num:  %d\n"
	            "=== failed stations num:   %d\n"
	            ,sla.event_q_num, sla.ringing_trunks_num, sla.ringing_stations_num,sla.failed_stations_num);

	ast_cli(a->fd, "\n"
	            "=============================================================\n"
	            "=== Configured SLA Trunks ===================================\n"
	            "=============================================================\n"
	            "===\n");
	i = ao2_iterator_init(sla_trunks, 0);
	for (; (trunk = ao2_iterator_next(&i)); ao2_ref(trunk, -1)) {
		struct sla_station_ref *station_ref;
		char ring_timeout[16] = "(none)";

		ao2_lock(trunk);

		if (trunk->ring_timeout) {
			snprintf(ring_timeout, sizeof(ring_timeout), "%u Seconds", trunk->ring_timeout);
		}

		ast_cli(a->fd, "=== ---------------------------------------------------------\n"
		            "=== Trunk Name:       %s\n"
		            "=== ==> Device:       %s\n"
		            "=== ==> AutoContext:  %s\n"
		            "=== ==> RingTimeout:  %s\n"
		            "=== ==> BargeAllowed: %s\n"
		            "=== ==> HoldAccess:   %s\n"
		            "=== ==> AutoRecord:   %s\n"
		            "=== ==> Usage:        %d/%d/%s\n"
		            "=== ==> State:        %s\n"
		            "=== ==> Stations ...\n",
		            trunk->name, trunk->device,
		            S_OR(trunk->autocontext, "(none)"),
		            ring_timeout,
		            trunk->barge_disabled ? "No" : "Yes",
		            sla_hold_str(trunk->hold_access),
		            (trunk->auto_record ? "Yes" : "No"),  
		            trunk->active_stations, trunk->hold_stations, (trunk->chan ? ast_channel_name(trunk->chan):"null"),
		            trunkstate2str(trunk->trunk_state));  
		AST_LIST_TRAVERSE(&trunk->stations, station_ref, entry) {
			ast_cli(a->fd, "===    ==> Station name: %s\n", station_ref->station->name);
		}

		ast_cli(a->fd, "=== ---------------------------------------------------------\n===\n");

		ao2_unlock(trunk);
	}
	ao2_iterator_destroy(&i);
	ast_cli(a->fd, "=============================================================\n\n");

	return CLI_SUCCESS;
}


static char *sla_show_stations(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct sla_station *station;

	switch (cmd) {
	case CLI_INIT:
		e->command = "sla show stations";
		e->usage =
			"Usage: sla show stations\n"
			"       This will list all stations defined in sla.conf\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n"
	            "=============================================================\n"
	            "=== Configured SLA Stations =================================\n"
	            "=============================================================\n"
	            "===\n");
	i = ao2_iterator_init(sla_stations, 0);
	for (; (station = ao2_iterator_next(&i)); ao2_ref(station, -1)) {
		struct sla_trunk_ref *trunk_ref;
		char ring_timeout[16] = "(none)";
		char ring_delay[16] = "(none)";

		ao2_lock(station);

		if (station->ring_timeout) {
			snprintf(ring_timeout, sizeof(ring_timeout),
				"%u", station->ring_timeout);
		}
		if (station->ring_delay) {
			snprintf(ring_delay, sizeof(ring_delay),
				"%u", station->ring_delay);
		}
		ast_cli(a->fd, "=== ---------------------------------------------------------\n"
		            "=== Station Name:    %s\n"
		            "=== ==> Device:      %s\n"
		            "=== ==> AutoContext: %s\n"
		            "=== ==> RingTimeout: %s\n"
		            "=== ==> RingDelay:   %s\n"
		            "=== ==> HoldAccess:  %s\n"
		            "=== ==> Trunks ...\n",
		            station->name, station->device,
		            S_OR(station->autocontext, "(none)"),
		            ring_timeout, ring_delay,
		            sla_hold_str(station->hold_access));
		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			if (trunk_ref->ring_timeout) {
				snprintf(ring_timeout, sizeof(ring_timeout),
					"%u", trunk_ref->ring_timeout);
			} else {
				strcpy(ring_timeout, "(none)");
			}
			if (trunk_ref->ring_delay) {
				snprintf(ring_delay, sizeof(ring_delay),
					"%u", trunk_ref->ring_delay);
			} else {
				strcpy(ring_delay, "(none)");
			}

			ast_cli(a->fd, "===    ==> Trunk Name: %s\n"
	            "===       ==> State:       %s\n"
	            "===       ==> RingTimeout: %s\n"
	            "===       ==> RingDelay:   %s\n",
	            trunk_ref->trunk->name,
	            trunkstate2str(trunk_ref->state),
	            ring_timeout, ring_delay);
		}
		ast_cli(a->fd, "=== ---------------------------------------------------------\n"
		            "===\n");

		ao2_unlock(station);
	}
	ao2_iterator_destroy(&i);
	ast_cli(a->fd, "============================================================\n"
	            "\n");

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_sla[] = {
	AST_CLI_DEFINE(sla_show_trunks, "Show SLA Trunks"),
	AST_CLI_DEFINE(sla_show_stations, "Show SLA Stations"),
};


static void sla_queue_event_full(enum sla_event_type type,
	struct sla_trunk_ref *trunk_ref, struct sla_station *station, int lock)
{
	struct sla_event *event;

	if (sla.thread == AST_PTHREADT_NULL) {
		ao2_ref(station, -1);
		ao2_ref(trunk_ref, -1);
		return;
	}

	if (!(event = ast_calloc(1, sizeof(*event)))) {
		ao2_ref(station, -1);
		ao2_ref(trunk_ref, -1);
		return;
	}

	event->type = type;
	event->trunk_ref = trunk_ref;
	event->station = station;

	if (!lock) {
		AST_LIST_INSERT_TAIL(&sla.event_q, event, entry);
		sla.event_q_num++;  /* Added by yfyang, grandstream */
		return;
	}

	ast_mutex_lock(&sla.lock);
	AST_LIST_INSERT_TAIL(&sla.event_q, event, entry);
	sla.event_q_num++;  /* Added by yfyang, grandstream */
	ast_cond_signal(&sla.cond);
	ast_mutex_unlock(&sla.lock);
}

static void sla_queue_event_nolock(enum sla_event_type type)
{
	sla_queue_event_full(type, NULL, NULL, 0);
}

static void sla_queue_event(enum sla_event_type type)
{
	sla_queue_event_full(type, NULL, NULL, 1);
}

/*! \brief Queue a SLA event from the conference */
static void sla_queue_event_conf(enum sla_event_type type, struct ast_channel *chan,
	struct sla_conf *conf)
{
	struct sla_station *station;
	struct sla_trunk_ref *trunk_ref = NULL;
	char *trunk_name;
	struct ao2_iterator i;

	trunk_name = ast_strdupa(conf->conf_name);
	strsep(&trunk_name, "_");
	if (ast_strlen_zero(trunk_name)) {
		ast_log(LOG_ERROR, "Invalid conf name for SLA - '%s'!\n", conf->conf_name);
		return;
	}

	i = ao2_iterator_init(sla_stations, 0);
	while ((station = ao2_iterator_next(&i))) {
		ao2_lock(station);
		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			if (trunk_ref->chan == chan && !strcmp(trunk_ref->trunk->name, trunk_name)) {
				ao2_ref(trunk_ref, 1);
				break;
			}
		}
		ao2_unlock(station);
		if (trunk_ref) {
			/* station reference given to sla_queue_event_full() */
			break;
		}
		ao2_ref(station, -1);
	}
	ao2_iterator_destroy(&i);

	if (!trunk_ref) {
		ast_debug(1, "Trunk not found for event!\n");
		return;
	}

	sla_queue_event_full(type, trunk_ref, station, 1);
}

static void sla_moh_start_withclass(struct ast_channel *chan, const char *musicclass)
{
	char *original_moh;

	ast_channel_lock(chan);
	original_moh = ast_strdupa(ast_channel_musicclass(chan));
	ast_channel_set_musicclass(chan, musicclass);
	ast_channel_unlock(chan);

	ast_moh_start(chan, original_moh, NULL);

	ast_channel_lock(chan);
	ast_channel_set_musicclass(chan, original_moh);
	ast_channel_unlock(chan);
}

static int sla_conf_user_set_exit(void *obj, void *arg, int flags)
{
	struct sla_conf_user *user = obj;

	ast_log(LOG_NOTICE, "sla_conf_user_set_exit, user->chan[%s]", ast_channel_name(user->chan));

	if (user->playing_moh) {
		sla_user_moh_stop(user);
	}

	ast_bridge_remove(user->conf->bridge, user->chan);

	return 0;
}

static int sla_conf_user_set_playing(void *obj, void *arg, int flags)
{
	struct sla_conf_user *user = obj;

	if (user->chan == arg) {
		ast_log(LOG_NOTICE, "sla_conf_user_set_playing, user->chan[%s]", ast_channel_name(user->chan));
		user->playing_moh = 1;
	}

    return 0;
}

static void sla_user_moh_stop(struct sla_conf_user *user)
{
	int in_bridge;

	if (!user || !user->playing_moh || !user->chan) {
		return;
	}

	user->playing_moh = 0; 

	ast_bridge_lock(user->conf->bridge);

	/* Temporarily suspend the user from the bridge so we have control to stop MOH if needed. */
	in_bridge = !ast_bridge_suspend(user->conf->bridge, user->chan);

	ast_log(LOG_NOTICE, "sla_moh_stop, userno[%d], chan[%s], in_bridge[%d]", user->user_no, ast_channel_name(user->chan), in_bridge);
	ast_moh_stop(user->chan);

	if (in_bridge) {
		ast_bridge_unsuspend(user->conf->bridge, user->chan);
	}

	ast_bridge_unlock(user->conf->bridge);
}

static void sla_user_moh_start(struct sla_conf_user *user)
{
	int in_bridge;

	if (!user || user->playing_moh || !user->chan) {
		return;
	}

	user->playing_moh = 1;

	/* Locking the ast_bridge here is the only way to hold off the
	 * call to ast_bridge_join() in confbridge_exec() from
	 * interfering with the bridge and MOH operations here. */
	ast_bridge_lock(user->conf->bridge);

	/* Temporarily suspend the user from the bridge so we have control to start MOH if needed. */
	in_bridge = !ast_bridge_suspend(user->conf->bridge, user->chan);

	ast_log(LOG_NOTICE, "sla_moh_start, userno[%d], chan[%s], in_bridge[%d]", user->user_no, ast_channel_name(user->chan), in_bridge);
	ast_moh_start(user->chan, NULL, NULL);

	if (in_bridge) {
		ast_bridge_unsuspend(user->conf->bridge, user->chan);
	}

	ast_bridge_unlock(user->conf->bridge);
}

static int sla_conf_user_start_moh(void *obj, void *arg, int flags)
{
	struct sla_conf_user *user = obj;

	if (user->playing_moh || !user->moh_flag || !user->chan) {
		return 0;
	}

	sla_user_moh_start(user);

	return 0;
}

static int sla_conf_user_stop_moh(void *obj, void *arg, int flags)
{
	struct sla_conf_user *user = obj;

	if (!user->playing_moh || !user->chan) {
		return 0;
	}

	sla_user_moh_stop(user);

	return 0;
}

static int sla_conf_set_record_filename(struct sla_conf *conf, struct ast_channel *chan)
{
#ifdef GRANDSTREAM_NETWORKS
	char recordingtmp[256] = "";
#else
	char recordingtmp[AST_MAX_EXTENSION] = "";
#endif
	char destdir[PATH_MAX] = {0};
    //const char *var;

	if (!conf || !chan) {
        return -1;
    }

    if (!conf->recordingfilename) {
        time_t timep;
        struct tm *p;
#ifdef GRANDSTREAM_NETWORKS
        snprintf(destdir, sizeof(destdir), "%s/meetme", ast_config_AST_SPOOL_DIR);
#else
        snprintf(destdir, sizeof(destdir), "%s/sla", ast_config_AST_SPOOL_DIR);
#endif
        if (ast_mkdir(destdir, 0777) != 0) {
            ast_log(LOG_WARNING, "mkdir '%s' failed: %s\n", destdir, strerror(errno));
            return -1;
        }

        time(&timep);
        p = localtime(&timep);

#ifdef GRANDSTREAM_NETWORKS
        snprintf(recordingtmp, sizeof(recordingtmp), "%s/%d-%02d/meetme-conf-rec-%s-%u.wav", 
				 destdir, 1900 + p->tm_year, 1 + p->tm_mon, conf->conf_name,  (unsigned int)timep);
#else
        snprintf(recordingtmp, sizeof(recordingtmp), "%s/%d-%02d/rec-%s-%u.wav", 
				 destdir, 1900 + p->tm_year, 1 + p->tm_mon, conf->conf_name,  (unsigned int)timep);
#endif
        conf->recordingfilename = ast_strdup(recordingtmp);
    }

    ast_log(LOG_NOTICE, "SLA conf[%s], set record filename[%s].\n", conf->conf_name, conf->recordingfilename);
    return 0;
}

static int sla_conf_start_record(struct sla_conf *conf, struct ast_channel *chan)
{
	struct ast_format_cap *cap;
	struct ast_bridge_features *features;
	struct ast_app *mixmonapp;

	if (!conf || !chan) {
		return -1;
	}

	if (conf->record_chan) {
		ast_log(LOG_NOTICE, "sla_conf_start_record, conf[%s] has started record\n", conf->conf_name);
		return 0;
	}

	mixmonapp = pbx_findapp("MixMonitor");
	if (!mixmonapp) {
		ast_log(LOG_WARNING, "Cannot record sla conf, MixMonitor app is not installed\n");
		return -1;
	}

	features = ast_bridge_features_new();
	if (!features) {
		return -1;
	}

	ast_set_flag(&features->feature_flags, AST_BRIDGE_CHANNEL_FLAG_IMMOVABLE);

	cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!cap) {
		ast_bridge_features_destroy(features);
		return -1;
	}

	ast_format_cap_append(cap, ast_format_slin, 0);

	/* Create the recording channel. */
	conf->record_chan = ast_request("CBRec", cap, NULL, NULL, conf->conf_name, NULL);
	ao2_ref(cap, -1);
	if (!conf->record_chan) {
		ast_bridge_features_destroy(features);
		return -1;
	}

	/* create avs record channel. */
	if (conf->bridge->avs_bridge_id > 0) {
		if (ast_avs_create_channel(conf->record_chan, avs_channel_type_rtp, 0, ast_channel_get_port_group_id(conf->record_chan)) != 0) {
			ast_log(LOG_ERROR, "Couldn't create avs channel for channel '%s'.", ast_channel_name(conf->record_chan));
		}
	}

	/* generate recording filename */
	if (!conf->recordingfilename) {
		sla_conf_set_record_filename(conf, chan);
	}

	ast_answer(conf->record_chan);

	ast_channel_ref(conf->record_chan);
	if (ast_bridge_impart(conf->bridge, conf->record_chan, NULL, features, AST_BRIDGE_IMPART_CHAN_INDEPENDENT)) {
        ast_log(LOG_WARNING, "sla_conf_start_record, record_chan[%s] impart bridge failed", ast_channel_name(conf->record_chan));
		ast_hangup(conf->record_chan);
		ast_channel_unref(conf->record_chan);
		conf->record_chan = NULL;
		return -1;
	}

	pbx_exec(chan, mixmonapp, (const char*)conf->recordingfilename);

	return 0;
}

static int sla_conf_stop_record(struct sla_conf *conf)
{
	struct ast_channel *chan;
	struct ast_frame f = { AST_FRAME_CONTROL, .subclass.integer = AST_CONTROL_HANGUP };

	if (!conf->record_chan) {
		return -1;
	}

    ast_log(LOG_NOTICE, "sla_conf_stop_record, record_chan[%s]", ast_channel_name(conf->record_chan));

	/* Remove the recording channel from the sla conf. */
	chan = conf->record_chan;
	conf->record_chan = NULL;

	if (ast_channel_is_avs_engine(chan)) {
		ast_avs_control(chan, AVS_CONTROL_STOP_RECORD, NULL, 0);
	}

	ast_queue_frame(chan, &f);
	ast_channel_unref(chan);
	ast_ha_rsync_file(AST_MEETING_FILE_ENUM);

	return 0;
}

static void sla_moh_start(struct sla_trunk_ref *trunk_ref, const char *musicclass)
{
	int in_bridge;

	if (!trunk_ref || !trunk_ref->trunk || !trunk_ref->trunk->chan || trunk_ref->playing_moh) {
		return;
	}

	if (!trunk_ref->conf || !trunk_ref->conf->bridge) {
		ast_log(LOG_NOTICE, "sla_moh_start, but no conf related!");
		return;
	}

	trunk_ref->playing_moh = 1;

	/* Locking the ast_bridge here is the only way to hold off the
	 * call to ast_bridge_join() in confbridge_exec() from
	 * interfering with the bridge and MOH operations here. */
	ast_bridge_lock(trunk_ref->conf->bridge);

	/* Temporarily suspend the user from the bridge so we have control to start MOH if needed. */
	in_bridge = !ast_bridge_suspend(trunk_ref->conf->bridge, trunk_ref->trunk->chan);

    ast_log(LOG_NOTICE, "sla_moh_start, chan[%s], in_bridge[%d]", ast_channel_name(trunk_ref->trunk->chan), in_bridge);

	ao2_callback(trunk_ref->conf->usercontainer, OBJ_NODATA, sla_conf_user_set_playing, trunk_ref->trunk->chan);

	if (musicclass) {
		sla_moh_start_withclass(trunk_ref->trunk->chan, musicclass);
	} else {
		ast_moh_start(trunk_ref->trunk->chan, NULL, NULL);
	}

	if (in_bridge) {
		ast_bridge_unsuspend(trunk_ref->conf->bridge, trunk_ref->trunk->chan);
	}

	ast_bridge_unlock(trunk_ref->conf->bridge);
}

static void sla_handle_hold_event(struct sla_event *event)
{
	ast_atomic_fetchadd_int((int *) &event->trunk_ref->trunk->hold_stations, 1);
	event->trunk_ref->state = SLA_TRUNK_STATE_ONHOLD_BYME;
	event->trunk_ref->trunk->trunk_state = SLA_TRUNK_STATE_ONHOLD_BYME;  

	ast_devstate_changed(AST_DEVICE_ONHOLD, AST_DEVSTATE_CACHABLE, "SLA:%s_%s",
			     event->station->name, event->trunk_ref->trunk->name);
	sla_change_trunk_state(event->trunk_ref->trunk, SLA_TRUNK_STATE_ONHOLD,
		INACTIVE_TRUNK_REFS, event->trunk_ref);

	if (event->trunk_ref->trunk->active_stations == 1) {
		/* The station putting it on hold is the only one on the call, so start Music on hold to the trunk. */
		event->trunk_ref->trunk->on_hold = 1;
		sla_moh_start(event->trunk_ref, NULL);
	}

	if (event->trunk_ref->chan) {
		if (event->trunk_ref->conf && event->trunk_ref->conf->bridge) {
			ast_bridge_remove(event->trunk_ref->conf->bridge, event->trunk_ref->chan);
		}

		event->trunk_ref->chan = NULL;
	}
}

static int sla_handle_hold_cb(struct ast_bridge_channel *bridge_channel, void *hook_pvt, void *mclass, int hold)
{
	struct sla_conf_user *user = hook_pvt;
	struct sla_conf *conf;
	int  found = 0;

    ast_log(LOG_NOTICE, "sla_handle_hold_cb, bridge_channel[%p], user[%p], hold[%d]\n", bridge_channel, user, hold);
	if (!bridge_channel || !hook_pvt || !hold || !user->chan) {
		return 0;
	}

	AST_LIST_LOCK(&sla_confs); 
	AST_LIST_TRAVERSE(&sla_confs, conf, list) {
		if (user->conf == conf) {
			found = 1;
			break;
		}
	}
	AST_LIST_UNLOCK(&sla_confs);

	if (!found) {
		ast_log(LOG_NOTICE, "sla_handle_hold_cb, can't find conf[%p] for user[%p]", user->conf, user);
		return 0;
	}

	ast_log(LOG_NOTICE, "sla_handle_hold_cb, chan[%s], conf[%s], queue hold event", ast_channel_name(user->chan), conf->conf_name);
	sla_queue_event_conf(SLA_EVENT_HOLD, user->chan, conf);

    return 0;
}

static void sla_hold_hook_destructor(void *hook_pvt)
{
	struct sla_conf_user *user = hook_pvt;

    ast_log(LOG_NOTICE, "sla_hold_hook_destructor, user[%p], user_no[%d]\n", user, user?user->user_no:0xff);
    return;
}

static int sla_conf_run(struct ast_channel *chan, struct sla_conf *conf, struct ast_flags64 *flags, char *optargs[])
{
	struct sla_conf_user *user  = NULL;
	struct timeval now = {0};
	struct ast_bridge_features chan_features;

	if (!chan || !conf || !flags) {
		return -1;
	}

	if (!(user = ao2_alloc(sizeof(*user), NULL))) {
		ast_log(LOG_ERROR, "sla_conf_run, alloc conf user for [%s] failed", ast_channel_name(chan));
		return -1;
	}

   	ast_mutex_lock(&conf->lock);

	ao2_lock(conf->usercontainer);
	user->user_no  = conf->user_count;
	user->chan     = chan;
	user->conf     = conf;
	user->destroy  = 0;
	user->moh_flag = 0;
	user->marked   = 0;
	user->playing_moh = 0;
	ao2_link(conf->usercontainer, user);

	conf->user_count++;
	ao2_unlock(conf->usercontainer);

	if (ast_test_flag64(flags, SLA_FLAG_MARKEDUSER)){
		conf->markedusers++;
		user->marked = 1;
	} 

	if (ast_test_flag64(flags, SLA_FLAG_MOH)) {
		user->moh_flag = 1;
	}

	ast_mutex_unlock(&conf->lock);

	if (conf->record_flag) {
		sla_conf_start_record(conf, chan);

		if (gs_record_prompt && conf->recordingfilename) {
			if (strstr(ast_channel_name(chan), "DAHDI/o")
				&& AST_STATE_RING == ast_channel_state(chan)) {
				ast_log(LOG_WARNING, "Chan:%s is analog line so we need first answer before prompt",
					ast_channel_name(chan));
				ast_answer(chan);
			}

			ast_stream_and_wait(chan, "dialog-being-recorded", "");
		}
	}

	/* return the unique ID of the conference */
	pbx_builtin_setvar_helper(chan, "SLA_UNIQUEID", conf->uniqueid);

	/* Initialize bridge features for the channel. */
	if (ast_bridge_features_init(&chan_features)) {
		ast_log(LOG_WARNING, "Unable to init bridge feature\n");
		goto outrun;
	}

    /* set call duration limit */
	if ((conf->user_count == 2) && (conf->duration > 0) && (conf->endtime == 0)) {
		struct ast_bridge_features_limits call_limits;

		now = ast_tvnow();
		conf->endtime = now.tv_sec + conf->duration;

		ast_bridge_features_limits_construct(&call_limits);

		call_limits.duration = conf->duration;
        if (user->marked && (conf->limit_warning > 0)) {
            call_limits.frequency = conf->limit_freq; 
            call_limits.warning = conf->limit_warning;
        }

		ast_bridge_features_set_limits(&chan_features, &call_limits, 0); 
		ast_bridge_features_limits_destroy(&call_limits);
	}

	ast_bridge_hold_hook(&chan_features, sla_handle_hold_cb, user, sla_hold_hook_destructor, AST_BRIDGE_HOOK_REMOVE_ON_PULL);

	ast_mutex_lock(&conf->lock);
	if (ast_channel_is_avs_engine(chan) && (conf->bridge->avs_bridge_id <= 0)) {
		int bridge_id = ast_avs_create_bridge(avs_bridge_conf, conf->bridge, -1, 0, ast_channel_get_avs_address(chan));
		if (bridge_id > 0) {
			conf->bridge->avs_bridge_id = bridge_id;
			conf->bridge->avs_bridge_type = avs_bridge_conf;				
			if (conf->bridge->technology && conf->bridge->technology->stop) {
				ast_bridge_lock(conf->bridge);
				conf->bridge->technology->stop(conf->bridge);
				ast_bridge_unlock(conf->bridge);
			}
		}
	}
	ast_mutex_unlock(&conf->lock);

	if ((conf->user_count == 1) && (user->moh_flag == 1)) {
		sla_moh_start_withclass(chan, optargs[SLA_TRUNK_OPT_ARG_MOH_CLASS]);
		user->playing_moh = 1;
	}

	if (conf->user_count > 1) {
		ao2_callback(conf->usercontainer, OBJ_NODATA | OBJ_MULTIPLE, sla_conf_user_stop_moh, NULL);
	}

	/* join bridge */
	ast_log(LOG_NOTICE, "sla_conf_run, chan[%s], user[%d][%p], join bridge.\n", ast_channel_name(chan), user->user_no, user);
	ast_bridge_join(conf->bridge, chan, NULL, &chan_features, NULL, 0);

	/* exit bridge */
	if (user->playing_moh) {
		ast_moh_stop(user->chan);
	}

 outrun:
	AST_LIST_LOCK(&sla_confs);

	ao2_lock(conf->usercontainer);
	conf->user_count--;

	/* Remove ourselves from the container */
	ao2_unlink(conf->usercontainer, user);

	if (user->marked) {
		conf->markedusers--;
		ast_log(LOG_NOTICE, "Marked user leave, [%d] markedusers in conf now\n", conf->markedusers);
		if ((conf->markedusers == 0) && (conf->user_count > 0)) {
			ao2_callback(conf->usercontainer, OBJ_NODATA, sla_conf_user_set_exit, NULL);
		}
	}
	ao2_unlock(conf->usercontainer);

	ast_log(LOG_NOTICE, "sla_conf_run, user[%d] exit, [%d] users in conf now!\n", user->user_no, conf->user_count);
	user->destroy = 1;

	ao2_ref(user, -1);
	AST_LIST_UNLOCK(&sla_confs);

	ao2_lock(conf->usercontainer);
	if ((conf->user_count == 1) && (conf->markedusers > 0)) {
		ao2_callback(conf->usercontainer, OBJ_NODATA, sla_conf_user_start_moh, NULL);
	}
	ao2_unlock(conf->usercontainer);

	ast_bridge_features_cleanup(&chan_features);

	return 0;
}


/*!
 * \internal
 * \brief Find an SLA trunk by name
 */
static struct sla_trunk *sla_find_trunk(const char *name)
{
	struct sla_trunk tmp_trunk = {
		.name = name,
	};

	return ao2_find(sla_trunks, &tmp_trunk, OBJ_POINTER);
}

/*!
 * \internal
 * \brief Find an SLA station by name
 */
static struct sla_station *sla_find_station(const char *name)
{
	struct sla_station tmp_station = {
		.name = name,
	};

	return ao2_find(sla_stations, &tmp_station, OBJ_POINTER);
}

static int sla_check_station_hold_access(const struct sla_trunk *trunk, const struct sla_station *station)
{
	struct sla_station_ref *station_ref;
	struct sla_trunk_ref *trunk_ref;
	int    has_private_hold = 0;

	/* For each station that has this call on hold, check for private hold. */
	AST_LIST_TRAVERSE(&trunk->stations, station_ref, entry) {
		AST_LIST_TRAVERSE(&station_ref->station->trunks, trunk_ref, entry) {
			if (trunk_ref->trunk != trunk)
				continue;

			if (station_ref->station == station) {
				if (trunk_ref->state == SLA_TRUNK_STATE_ONHOLD_BYME) {
					return 0;
				} else {
					continue;
				}
			}

			if (trunk_ref->state == SLA_TRUNK_STATE_ONHOLD_BYME &&
				station_ref->station->hold_access == SLA_HOLD_PRIVATE) {
				has_private_hold = 1;
			}
		}
	}

	return has_private_hold;
}

static int sla_check_trunk_calltime_limited(struct ast_channel *chan, const char *device)
{
	char dev[128]  = {0};
	char buf[32]   = {0};
	char func[128] = {0};
	char  *pos = NULL;

	if (!chan || !device) {
		return 0;
	}

	pos = strchr(device, '/');
	if (!pos) {
		return 0;
	}

	pos++;
	snprintf(dev, sizeof(dev)-1, "DAHDI/o%s", pos);
	ast_log(LOG_NOTICE, "Get chan[%s]", dev);

	snprintf(func, sizeof(func), "TRUNK_TIME_LIMITED(%s)", dev);
	if ((0 == ast_func_read(chan, func, buf, sizeof(buf))) && (strlen(buf) > 0)) {
		if (!strcmp(buf, "1")) {
			ast_log(LOG_NOTICE, "dev[%s] calltime limited", dev);
			return 1;
		}
	}

	return 0;
}

/*!
 * \brief Find a trunk reference on a station by name
 * \param station the station
 * \param name the trunk's name
 * \pre sla_station is locked
 * \return a pointer to the station's trunk reference.  If the trunk
 *         is not found, it is not idle and barge is disabled, or if
 *         it is on hold and private hold is set, then NULL will be returned.
 */
static struct sla_trunk_ref *sla_find_trunk_ref_byname(struct ast_channel *chan, const struct sla_station *station,
	const char *name)
{
	struct sla_trunk_ref *trunk_ref = NULL;

	AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
		if (strcasecmp(trunk_ref->trunk->name, name)){
			continue;
		}

		if ((trunk_ref->trunk->barge_disabled && trunk_ref->state == SLA_TRUNK_STATE_UP) ||
			(trunk_ref->trunk->hold_stations && 
			 trunk_ref->trunk->hold_access == SLA_HOLD_PRIVATE && 
			 trunk_ref->state != SLA_TRUNK_STATE_ONHOLD_BYME) ||
			sla_check_station_hold_access(trunk_ref->trunk, station) ||
			sla_check_trunk_calltime_limited(chan, trunk_ref->trunk->device))
		{
			trunk_ref = NULL;
		}

		break;
	}

	if (trunk_ref) {
		ao2_ref(trunk_ref, 1);
	}

	return trunk_ref;
}

static void sla_station_ref_destructor(void *obj)
{
	struct sla_station_ref *station_ref = obj;

	if (station_ref->station) {
		ao2_ref(station_ref->station, -1);
		station_ref->station = NULL;
	}
}

static struct sla_station_ref *sla_create_station_ref(struct sla_station *station)
{
	struct sla_station_ref *station_ref;

	if (!(station_ref = ao2_alloc(sizeof(*station_ref), sla_station_ref_destructor))) {
		return NULL;
	}

	ao2_ref(station, 1);
	station_ref->station = station;

	return station_ref;
}

static struct sla_ringing_station *sla_create_ringing_station(struct sla_station *station)
{
	struct sla_ringing_station *ringing_station;

	if (!(ringing_station = ast_calloc(1, sizeof(*ringing_station))))
		return NULL;

	ao2_ref(station, 1);
	ringing_station->station = station;
	ringing_station->ring_begin = ast_tvnow();

	return ringing_station;
}

static void sla_ringing_station_destroy(struct sla_ringing_station *ringing_station)
{
	if (ringing_station->station) {
		ao2_ref(ringing_station->station, -1);
		ringing_station->station = NULL;
	}

	ast_free(ringing_station);
}

static struct sla_failed_station *sla_create_failed_station(struct sla_station *station)
{
	struct sla_failed_station *failed_station;

	if (!(failed_station = ast_calloc(1, sizeof(*failed_station)))) {
		return NULL;
	}

	ao2_ref(station, 1);
	failed_station->station = station;
	failed_station->last_try = ast_tvnow();

	return failed_station;
}

static void sla_failed_station_destroy(struct sla_failed_station *failed_station)
{
	if (failed_station->station) {
		ao2_ref(failed_station->station, -1);
		failed_station->station = NULL;
	}

	ast_free(failed_station);
}

static enum ast_device_state sla_state_to_devstate(enum sla_trunk_state state)
{
	switch (state) {
	case SLA_TRUNK_STATE_IDLE:
		return AST_DEVICE_NOT_INUSE;
	case SLA_TRUNK_STATE_RINGING:
		return AST_DEVICE_RINGING;
	case SLA_TRUNK_STATE_UP:
		return AST_DEVICE_INUSE;
	case SLA_TRUNK_STATE_ONHOLD:
	case SLA_TRUNK_STATE_ONHOLD_BYME:
		return AST_DEVICE_ONHOLD;
	case SLA_TRUNK_STATE_UNAVAILABLE:      /* Added by yfyang */
		return AST_DEVICE_UNAVAILABLE;
	}

	return AST_DEVICE_UNKNOWN;
}

static void sla_change_trunk_state(struct sla_trunk *trunk, enum sla_trunk_state state, 
	enum sla_which_trunk_refs inactive_only, const struct sla_trunk_ref *exclude)
{
	struct sla_station *station;
	struct sla_trunk_ref *trunk_ref;
	struct ao2_iterator i;
	int    other_ringing = 0;

	int  trunk_matched = 0;

	i = ao2_iterator_init(sla_stations, 0);
	while ((station = ao2_iterator_next(&i))) {
		ao2_lock(station);

		/* Added by yfyang, for Bug 195436 */
		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			if ((trunk_ref->trunk != trunk) && (trunk_ref->state == SLA_TRUNK_STATE_RINGING)) {
				other_ringing = 1;
				break;
			}
			continue;
		}

		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			if (trunk_ref->trunk != trunk || (inactive_only ? trunk_ref->chan : 0)
					|| trunk_ref == exclude) {
				continue;
			}

			trunk_matched = 1;

			if (SLA_TRUNK_STATE_ONHOLD_BYME == trunk_ref->state && SLA_TRUNK_STATE_ONHOLD == state) {
				continue;
			}

			if (!other_ringing && (state == SLA_TRUNK_STATE_IDLE) && (trunk_ref->state == SLA_TRUNK_STATE_RINGING)) {
				ast_log(LOG_NOTICE, "sla_change_trunk_state, trunk[%s], RINGING --> IDLE\n", trunk->name);
				if (station->dial && ast_get_state(station->dial) == AST_DIAL_RESULT_RINGING) {
					ast_set_state(station->dial, AST_DIAL_RESULT_HANGUP);
				}
			}

			trunk_ref->state = state;
			if (state == SLA_TRUNK_STATE_IDLE || state == SLA_TRUNK_STATE_UP) {
				if (trunk_ref->playing_moh) {
					sla_moh_stop(trunk_ref);
				}

				if (state == SLA_TRUNK_STATE_IDLE) {
					trunk_ref->conf = NULL;
				}
			}

			if (((SLA_TRUNK_STATE_ONHOLD == trunk->trunk_state) || (SLA_TRUNK_STATE_ONHOLD_BYME == trunk->trunk_state)) &&
				((SLA_TRUNK_STATE_IDLE == state) || (SLA_TRUNK_STATE_UNAVAILABLE == state)|| (SLA_TRUNK_STATE_UP == state) )) {
				ast_log(LOG_NOTICE, "sla_change_trunk_state, trunk[%s], hold_stations[%d].\n", trunk->name, trunk->hold_stations);
				trunk->hold_stations = 0;
			}
			trunk->trunk_state = state;

			ast_devstate_changed(sla_state_to_devstate(state), AST_DEVSTATE_CACHABLE,
					     "SLA:%s_%s", station->name, trunk->name);
			break;
		}
		ao2_unlock(station);
		ao2_ref(station, -1);
	}
	ao2_iterator_destroy(&i);

	if (!trunk_matched) {
		trunk->trunk_state = state;
	}
}


static enum sla_trunk_state sla_devstate_to_slastate(enum ast_device_state state)
{
	switch (state) {
	case AST_DEVICE_NOT_INUSE:
		return SLA_TRUNK_STATE_IDLE;

	case AST_DEVICE_RINGING:
	case AST_DEVICE_RINGINUSE:
		return SLA_TRUNK_STATE_RINGING;

	case AST_DEVICE_INUSE:
	case AST_DEVICE_BUSY:
		return SLA_TRUNK_STATE_UP;

	case AST_DEVICE_ONHOLD:
		return SLA_TRUNK_STATE_ONHOLD;

	case AST_DEVICE_UNAVAILABLE:
	case AST_DEVICE_INVALID:
		return SLA_TRUNK_STATE_UNAVAILABLE;

	default:
		return SLA_TRUNK_STATE_UNAVAILABLE;
	}
}


static enum sla_trunk_state sla_extenstate_to_slastate(enum ast_extension_states state)
{
	switch (state) {
	case AST_EXTENSION_NOT_INUSE:
		return SLA_TRUNK_STATE_IDLE;

	case AST_EXTENSION_RINGING:
		return SLA_TRUNK_STATE_RINGING;

	case AST_EXTENSION_INUSE:
	case AST_EXTENSION_BUSY:
		return SLA_TRUNK_STATE_UP;

	case AST_EXTENSION_ONHOLD:
		return SLA_TRUNK_STATE_ONHOLD;

	case AST_EXTENSION_UNAVAILABLE:
		return SLA_TRUNK_STATE_UNAVAILABLE;

	default:
		return SLA_TRUNK_STATE_UNAVAILABLE;
	}
}

/*! \brief set a sla trunk's status based on device state */
static void sla_state_cb(void *unused, struct stasis_subscription *sub, struct stasis_message *msg)
{
	struct ast_device_state_message *dev_state;
	enum   sla_trunk_state      sla_state;
	struct sla_trunk  *pstTrunk = NULL;
	int    found_trunk  = 0;
	struct ao2_iterator i;

	if (ast_device_state_message_type() != stasis_message_type(msg)) {
		return;
	}

	dev_state = stasis_message_data(msg);
	if (!dev_state || dev_state->eid || strncmp(dev_state->device, "DAHDI/", 6)) {
		/* ignore non-aggregate states and non-DAHDI status */
		return;
	}

	sla_state = sla_extenstate_to_slastate(ast_devstate_to_extenstate((enum ast_device_state)dev_state->state));
	if (sla_state != SLA_TRUNK_STATE_UNAVAILABLE && sla_state != SLA_TRUNK_STATE_IDLE) {
		ast_log(LOG_DEBUG, "report device[%s] in state[%d][%s], ignore it.\n",  dev_state->device, sla_state, trunkstate2str(sla_state));
		return;
	}

	/* check the trunk device */
	i = ao2_iterator_init(sla_trunks, 0);
	for (; (pstTrunk = ao2_iterator_next(&i)); ao2_ref(pstTrunk, -1)) {
		if (0 == strcmp(pstTrunk->device, dev_state->device)) {
			found_trunk = 1;
			ao2_ref(pstTrunk, -1);
			break;
		}
	}
	ao2_iterator_destroy(&i);

	if (0 == found_trunk) {
		return;
	}

	if (pstTrunk->trunk_state != sla_state) {
		ast_log(LOG_NOTICE, "sla_trunkstate, trunk[%s], device[%s], trunk_state[%s] ==> state[%d][%s].\n",
				pstTrunk->name, pstTrunk->device, trunkstate2str(pstTrunk->trunk_state), sla_state, trunkstate2str(sla_state));
		sla_change_trunk_state(pstTrunk, sla_state, ALL_TRUNK_REFS, NULL);
	}

	return;
}

struct run_station_args {
	struct sla_station *station;
	struct sla_trunk_ref *trunk_ref;
	ast_mutex_t *cond_lock;
	ast_cond_t *cond;
};

static void answer_trunk_chan(struct ast_channel *chan)
{
	ast_answer(chan);
	ast_indicate(chan, -1);
}




static void *run_station(void *data)
{
	RAII_VAR(struct sla_station *, station, NULL, ao2_cleanup);
	RAII_VAR(struct sla_trunk_ref *, trunk_ref, NULL, ao2_cleanup);
	struct ast_str *conf_name = ast_str_create(16);
	struct ast_flags64 conf_flags = { 0 };
	struct sla_conf *conf;

	{
		struct run_station_args *args = data;
		station = args->station;
		trunk_ref = args->trunk_ref;
		ast_mutex_lock(args->cond_lock);
		ast_cond_signal(args->cond);
		ast_mutex_unlock(args->cond_lock);
		/* args is no longer valid here. */
	}

	ast_atomic_fetchadd_int((int *) &trunk_ref->trunk->active_stations, 1);

	answer_trunk_chan(trunk_ref->chan);

	ast_str_set(&conf_name, 0, "SLA_%s", trunk_ref->trunk->name);
	conf = sla_conf_build(ast_str_buffer(conf_name), 1, trunk_ref->chan);
	if (conf) {
		trunk_ref->conf = conf;
		ast_set_flag64(&conf_flags, SLA_FLAG_MARKEDEXIT | SLA_FLAG_SLA_STATION);

		ast_log(LOG_NOTICE, "run_station, chan[%s] join conf.", ast_channel_name(trunk_ref->chan));
		sla_conf_run(trunk_ref->chan, conf, &conf_flags, NULL);
	}

	trunk_ref->chan = NULL;
	if (ast_atomic_dec_and_test((int *) &trunk_ref->trunk->active_stations) &&
		trunk_ref->state != SLA_TRUNK_STATE_ONHOLD_BYME) {
		if (trunk_ref->conf) {
			ao2_callback(trunk_ref->conf->usercontainer, OBJ_NODATA, sla_conf_user_set_exit, NULL);
		}

		trunk_ref->trunk->hold_stations = 0;
		sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);
	}
	/*BEGIN; Mod by weidu, for sla crash*/
	if (conf){
		if ( 1 == sla_conf_dispose(conf, trunk_ref)){  /*Mod by weidu, for conf use after free*/
			trunk_ref->conf = NULL;
		}
		conf = NULL;
	}
	/*END; Mod by weidu, for sla crash*/
	if (station->dial) {
		ast_dial_join(station->dial); 
		ast_dial_destroy(station->dial);
	}

	station->dial = NULL;
	ast_free(conf_name);

	return NULL;
}

static void sla_stop_ringing_trunk(struct sla_ringing_trunk *ringing_trunk)
{
	char buf[80];
	struct sla_station_ref *station_ref;
	struct sla_conf *conf;

	snprintf(buf, sizeof(buf), "SLA_%s", ringing_trunk->trunk->name);

	AST_LIST_LOCK(&sla_confs); 
	AST_LIST_TRAVERSE(&sla_confs, conf, list) {
		if (!strcmp(conf->conf_name, buf)) {
			ao2_callback(conf->usercontainer, OBJ_NODATA, sla_conf_user_set_exit, NULL);
			break;
		}
	}
	AST_LIST_UNLOCK(&sla_confs);


	sla_change_trunk_state(ringing_trunk->trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);

	while ((station_ref = AST_LIST_REMOVE_HEAD(&ringing_trunk->timed_out_stations, entry))) {
		ao2_ref(station_ref, -1);
	}

	ao2_ref(ringing_trunk, -1);
}

static void sla_stop_ringing_station(struct sla_ringing_station *ringing_station,
	enum sla_station_hangup hangup)
{
	struct sla_ringing_trunk *ringing_trunk;
	struct sla_trunk_ref *trunk_ref;
	struct sla_station_ref *station_ref;

	ast_dial_join(ringing_station->station->dial);
	ast_dial_destroy(ringing_station->station->dial);
	ringing_station->station->dial = NULL;

	if (hangup == SLA_STATION_HANGUP_NORMAL)
		goto done;

	/* If the station is being hung up because of a timeout, then add it to the
	 * list of timed out stations on each of the ringing trunks.  This is so
	 * that when doing further processing to figure out which stations should be
	 * ringing, which trunk to answer, determining timeouts, etc., we know which
	 * ringing trunks we should ignore. */
	ast_mutex_lock(&sla.lock);
	AST_LIST_TRAVERSE(&sla.ringing_trunks, ringing_trunk, entry) {
		AST_LIST_TRAVERSE(&ringing_station->station->trunks, trunk_ref, entry) {
			if (ringing_trunk->trunk == trunk_ref->trunk)
				break;
		}
		if (!trunk_ref)
			continue;
		if (!(station_ref = sla_create_station_ref(ringing_station->station)))
			continue;
#ifdef GRANDSTREAM_NETWORKS
		if (hangup == SLA_STATION_HANGUP_BUSY && ringing_trunk->trunk->chan) {
			pbx_builtin_setvar_helper(ringing_trunk->trunk->chan, "SLATRUNK_STATUS", "BUSY");
		} else if (hangup == SLA_STATION_HANGUP_TIMEOUT && ringing_trunk->trunk->chan) {
			pbx_builtin_setvar_helper(ringing_trunk->trunk->chan, "SLATRUNK_STATUS", "UNANSWERED");
		}
#endif
		AST_LIST_INSERT_TAIL(&ringing_trunk->timed_out_stations, station_ref, entry);
	}
	ast_mutex_unlock(&sla.lock);

done:
	sla_ringing_station_destroy(ringing_station);
}

static void sla_dial_state_callback(struct ast_dial *dial)
{
	sla_queue_event(SLA_EVENT_DIAL_STATE);
}

/*! \brief Check to see if dialing this station already timed out for this ringing trunk
 * \note Assumes sla.lock is locked
 */
static int sla_check_timed_out_station(const struct sla_ringing_trunk *ringing_trunk,
	const struct sla_station *station)
{
	struct sla_station_ref *timed_out_station;

	AST_LIST_TRAVERSE(&ringing_trunk->timed_out_stations, timed_out_station, entry) {
		if (station == timed_out_station->station)
			return 1;
	}

	return 0;
}

/*! \brief Choose the highest priority ringing trunk for a station
 * \param station the station
 * \param rm remove the ringing trunk once selected
 * \param trunk_ref a place to store the pointer to this stations reference to
 *        the selected trunk
 * \return a pointer to the selected ringing trunk, or NULL if none found
 * \note Assumes that sla.lock is locked
 */
static struct sla_ringing_trunk *sla_choose_ringing_trunk(struct sla_station *station,
	struct sla_trunk_ref **trunk_ref, int rm)
{
	struct sla_trunk_ref *s_trunk_ref;
	struct sla_ringing_trunk *ringing_trunk = NULL;

	AST_LIST_TRAVERSE(&station->trunks, s_trunk_ref, entry) {
		ast_mutex_lock(&sla.lock);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_trunks, ringing_trunk, entry) {
			/* Make sure this is the trunk we're looking for */
			if (s_trunk_ref->trunk != ringing_trunk->trunk)
				continue;

			/* This trunk on the station is ringing.  But, make sure this station
			 * didn't already time out while this trunk was ringing. */
			if (sla_check_timed_out_station(ringing_trunk, station))
				continue;

			if (rm) {
				AST_LIST_REMOVE_CURRENT(entry);
				sla.ringing_trunks_num--;   /* Added by yfyang */
			}

			if (trunk_ref) {
				ao2_ref(s_trunk_ref, 1);
				*trunk_ref = s_trunk_ref;
			}

			break;
		}
		AST_LIST_TRAVERSE_SAFE_END;
		ast_mutex_unlock(&sla.lock);

		if (ringing_trunk)
			break;
	}

	return ringing_trunk;
}

static void sla_handle_dial_state_event(void)
{
	struct sla_ringing_station *ringing_station;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_stations, ringing_station, entry) {
		RAII_VAR(struct sla_trunk_ref *, s_trunk_ref, NULL, ao2_cleanup);
		struct sla_ringing_trunk *ringing_trunk = NULL;
		struct run_station_args args;
		enum ast_dial_result dial_res;
		pthread_t dont_care;
		ast_mutex_t cond_lock;
		ast_cond_t cond;

		switch ((dial_res = ast_dial_state(ringing_station->station->dial))) {
		case AST_DIAL_RESULT_HANGUP:
		case AST_DIAL_RESULT_INVALID:
		case AST_DIAL_RESULT_FAILED:
		case AST_DIAL_RESULT_TIMEOUT:
		case AST_DIAL_RESULT_UNANSWERED:
#ifdef GRANDSTREAM_NETWORKS
		case AST_DIAL_RESULT_PEER_BUSY:
#endif
		{
			AST_LIST_REMOVE_CURRENT(entry);
#ifdef GRANDSTREAM_NETWORKS
			enum sla_station_hangup hangup = SLA_STATION_HANGUP_NORMAL;

			if (dial_res == AST_DIAL_RESULT_PEER_BUSY) {
				hangup = SLA_STATION_HANGUP_BUSY;
			} else if (dial_res == AST_DIAL_RESULT_UNANSWERED || dial_res == AST_DIAL_RESULT_TIMEOUT) {
				hangup = SLA_STATION_HANGUP_TIMEOUT;
			}

			sla_stop_ringing_station(ringing_station, hangup);
#else
			sla_stop_ringing_station(ringing_station, SLA_STATION_HANGUP_NORMAL);
#endif
			sla.ringing_stations_num--; /* Added by yfyang */
			break;
		}

		case AST_DIAL_RESULT_ANSWERED:
		{
			AST_LIST_REMOVE_CURRENT(entry);
			sla.ringing_stations_num--; 
			/* Find the appropriate trunk to answer. */
			ast_mutex_lock(&sla.lock);
			ringing_trunk = sla_choose_ringing_trunk(ringing_station->station, &s_trunk_ref, 1);
			ast_mutex_unlock(&sla.lock);
			if (!ringing_trunk) {
				/* This case happens in a bit of a race condition.  If two stations answer
				 * the outbound call at the same time, the first one will get connected to
				 * the trunk.  When the second one gets here, it will not see any trunks
				 * ringing so we have no idea what to conect it to.  So, we just hang up
				 * on it. */
				ast_debug(1, "Found no ringing trunk for station '%s' to answer!\n", ringing_station->station->name);
				ast_dial_join(ringing_station->station->dial);
				ast_dial_destroy(ringing_station->station->dial);
				ringing_station->station->dial = NULL;
				sla_ringing_station_destroy(ringing_station);
				break;
			}
			/* Track the channel that answered this trunk */
			s_trunk_ref->chan = ast_dial_answered(ringing_station->station->dial);
			/* Actually answer the trunk */
			if(ast_channel_state(ringing_trunk->trunk->chan) != AST_STATE_UP) {
				ast_channel_lock(ringing_trunk->trunk->chan);
				answer_trunk_chan(ringing_trunk->trunk->chan);
				ast_channel_unlock(ringing_trunk->trunk->chan);
			}
			sla_change_trunk_state(ringing_trunk->trunk, SLA_TRUNK_STATE_UP, ALL_TRUNK_REFS, NULL);
			/* Now, start a thread that will connect this station to the trunk.  The rest of
			 * the code here sets up the thread and ensures that it is able to save the arguments
			 * before they are no longer valid since they are allocated on the stack. */
			ao2_ref(s_trunk_ref, 1);
			args.trunk_ref = s_trunk_ref;
			ao2_ref(ringing_station->station, 1);
			args.station = ringing_station->station;
			args.cond = &cond;
			args.cond_lock = &cond_lock;
			ao2_ref(ringing_trunk, -1);
			sla_ringing_station_destroy(ringing_station);
			ast_mutex_init(&cond_lock);
			ast_cond_init(&cond, NULL);
			ast_mutex_lock(&cond_lock);
			ast_pthread_create_detached_background(&dont_care, NULL, run_station, &args);
			ast_cond_wait(&cond, &cond_lock);
			ast_mutex_unlock(&cond_lock);
			ast_mutex_destroy(&cond_lock);
			ast_cond_destroy(&cond);
			break;
		}

		case AST_DIAL_RESULT_TRYING:
		case AST_DIAL_RESULT_RINGING:
		case AST_DIAL_RESULT_PROGRESS:
		case AST_DIAL_RESULT_PROCEEDING:
			break;
		}
		if (dial_res == AST_DIAL_RESULT_ANSWERED) {
			/* Queue up reprocessing ringing trunks, and then ringing stations again */
			sla_queue_event(SLA_EVENT_RINGING_TRUNK);
			sla_queue_event(SLA_EVENT_DIAL_STATE);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
}

/*! \brief Check to see if this station is already ringing
 * \note Assumes sla.lock is locked
 */
static int sla_check_ringing_station(const struct sla_station *station)
{
	struct sla_ringing_station *ringing_station;

	AST_LIST_TRAVERSE(&sla.ringing_stations, ringing_station, entry) {
		if (station == ringing_station->station)
			return 1;
	}

	return 0;
}

/*! \brief Check to see if this station has failed to be dialed in the past minute
 * \note assumes sla.lock is locked
 */
static int sla_check_failed_station(const struct sla_station *station)
{
	struct sla_failed_station *failed_station;
	int res = 0;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.failed_stations, failed_station, entry) {
		if (station != failed_station->station)
			continue;
		if (ast_tvdiff_ms(ast_tvnow(), failed_station->last_try) > 1000) {
			AST_LIST_REMOVE_CURRENT(entry);
			sla.failed_stations_num--;  /* Added by yfyang */
			sla_failed_station_destroy(failed_station);
			break;
		}
		res = 1;
	}
	AST_LIST_TRAVERSE_SAFE_END

	return res;
}

/*! \brief Ring a station
 * \note Assumes sla.lock is locked
 */
static int sla_ring_station(struct sla_ringing_trunk *ringing_trunk, struct sla_station *station)
{
	char *tech, *tech_data;
	char *resource, *data;
	struct ast_dial *dial;
	struct sla_ringing_station *ringing_station;
	enum ast_dial_result res;
	int caller_is_saved;
	struct ast_party_caller caller;
	char buf[1024] = {0};
	char func[128] = {0};

	if (!(dial = ast_dial_create()))
		return -1;

	ast_dial_set_state_callback(dial, sla_dial_state_callback);
	tech_data = ast_strdupa(station->device);
	tech = strsep(&tech_data, "/");

	ast_log(LOG_NOTICE, "sla_ring_station, tech[%s], tech_data[%s]", tech, tech_data);

	/* For Bug 208707 */
	snprintf(func, sizeof(func), "PJSIP_DIAL_CONTACTS(%s)", tech_data);
	if ((0 == ast_func_read(NULL, func, buf, sizeof(buf))) && (strlen(buf) > 0)) {
		data = buf;
		while ((tech = strsep(&data, "&"))) {
			if (!(resource = strchr(tech, '/'))) {
				continue;
			}

			*resource++ = '\0';
			ast_log(LOG_NOTICE, "dial append: tech[%s], resource[%s]", tech, resource);
			if (ast_dial_append(dial, tech, resource, NULL) == -1) {
				continue;
			}
		}
	} else {
		if (ast_dial_append(dial, tech, tech_data, NULL) == -1) {
			ast_dial_destroy(dial);
			return -1;
		}
	}

	/* Do we need to save off the caller ID data? */
	caller_is_saved = 0;
	if (!sla.attempt_callerid) {
		caller_is_saved = 1;
		caller = *ast_channel_caller(ringing_trunk->trunk->chan);
		ast_party_caller_init(ast_channel_caller(ringing_trunk->trunk->chan));
	}

	res = ast_dial_run(dial, ringing_trunk->trunk->chan, 1);

	/* Restore saved caller ID */
	if (caller_is_saved) {
		ast_party_caller_free(ast_channel_caller(ringing_trunk->trunk->chan));
		ast_channel_caller_set(ringing_trunk->trunk->chan, &caller);
	}

	if (res != AST_DIAL_RESULT_TRYING) {
		struct sla_failed_station *failed_station;
		ast_dial_destroy(dial);
		if ((failed_station = sla_create_failed_station(station))) {
			AST_LIST_INSERT_HEAD(&sla.failed_stations, failed_station, entry);
			sla.failed_stations_num++;   /* Added by yfyang */
		}
		return -1;
	}
	if (!(ringing_station = sla_create_ringing_station(station))) {
		ast_dial_join(dial);
		ast_dial_destroy(dial);
		return -1;
	}

	station->dial = dial;

	AST_LIST_INSERT_HEAD(&sla.ringing_stations, ringing_station, entry);
	sla.ringing_stations_num++; /* Added by yfyang */
	return 0;
}

/*! \brief Check to see if a station is in use
 */
static int sla_check_inuse_station(const struct sla_station *station)
{
	struct sla_trunk_ref *trunk_ref;

	AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
		if (trunk_ref->chan)
			return 1;
	}

	return 0;
}

static struct sla_trunk_ref *sla_find_trunk_ref(const struct sla_station *station,
	const struct sla_trunk *trunk)
{
	struct sla_trunk_ref *trunk_ref = NULL;

	AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
		if (trunk_ref->trunk == trunk)
			break;
	}

	ao2_ref(trunk_ref, 1);

	return trunk_ref;
}

/*! \brief Calculate the ring delay for a given ringing trunk on a station
 * \param station the station
 * \param ringing_trunk the trunk.  If NULL, the highest priority ringing trunk will be used
 * \return the number of ms left before the delay is complete, or INT_MAX if there is no delay
 */
static int sla_check_station_delay(struct sla_station *station,
	struct sla_ringing_trunk *ringing_trunk)
{
	RAII_VAR(struct sla_trunk_ref *, trunk_ref, NULL, ao2_cleanup);
	unsigned int delay = UINT_MAX;
	int time_left, time_elapsed;

	if (!ringing_trunk)
		ringing_trunk = sla_choose_ringing_trunk(station, &trunk_ref, 0);
	else
		trunk_ref = sla_find_trunk_ref(station, ringing_trunk->trunk);

	if (!ringing_trunk || !trunk_ref)
		return delay;

	/* If this station has a ring delay specific to the highest priority
	 * ringing trunk, use that.  Otherwise, use the ring delay specified
	 * globally for the station. */
	delay = trunk_ref->ring_delay;
	if (!delay)
		delay = station->ring_delay;
	if (!delay)
		return INT_MAX;

	time_elapsed = ast_tvdiff_ms(ast_tvnow(), ringing_trunk->ring_begin);
	time_left = (delay * 1000) - time_elapsed;

	return time_left;
}

/*! \brief Ring stations based on current set of ringing trunks
 * \note Assumes that sla.lock is locked
 */
static void sla_ring_stations(void)
{
	struct sla_station_ref *station_ref;
	struct sla_ringing_trunk *ringing_trunk;

	/* Make sure that every station that uses at least one of the ringing
	 * trunks, is ringing. */
	AST_LIST_TRAVERSE(&sla.ringing_trunks, ringing_trunk, entry) {
		AST_LIST_TRAVERSE(&ringing_trunk->trunk->stations, station_ref, entry) {
			int time_left;

			/* Is this station already ringing? */
			if (sla_check_ringing_station(station_ref->station))
				continue;

			/* Is this station already in a call? */
			if (sla_check_inuse_station(station_ref->station))
				continue;

			/* Did we fail to dial this station earlier?  If so, has it been
 			 * a minute since we tried? */
			if (sla_check_failed_station(station_ref->station))
				continue;

			/* If this station already timed out while this trunk was ringing,
			 * do not dial it again for this ringing trunk. */
			if (sla_check_timed_out_station(ringing_trunk, station_ref->station))
				continue;

			/* Check for a ring delay in progress */
			time_left = sla_check_station_delay(station_ref->station, ringing_trunk);
			if (time_left != INT_MAX && time_left > 0)
				continue;

			/* It is time to make this station begin to ring.  Do it! */
			sla_ring_station(ringing_trunk, station_ref->station);
		}
	}
	/* Now, all of the stations that should be ringing, are ringing. */
}

static void sla_hangup_stations(void)
{
	struct sla_trunk_ref *trunk_ref;
	struct sla_ringing_station *ringing_station;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_stations, ringing_station, entry) {
		AST_LIST_TRAVERSE(&ringing_station->station->trunks, trunk_ref, entry) {
			struct sla_ringing_trunk *ringing_trunk;
			ast_mutex_lock(&sla.lock);
			AST_LIST_TRAVERSE(&sla.ringing_trunks, ringing_trunk, entry) {
				if (trunk_ref->trunk == ringing_trunk->trunk)
					break;
			}
			ast_mutex_unlock(&sla.lock);
			if (ringing_trunk)
				break;
		}
		if (!trunk_ref) {
			AST_LIST_REMOVE_CURRENT(entry);
			ast_dial_join(ringing_station->station->dial);
			ast_dial_destroy(ringing_station->station->dial);
			ringing_station->station->dial = NULL;
			sla_ringing_station_destroy(ringing_station);
			sla.ringing_stations_num--;  /* Added by yfyang */
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
}

static void sla_handle_ringing_trunk_event(void)
{
	ast_mutex_lock(&sla.lock);
	sla_ring_stations();
	ast_mutex_unlock(&sla.lock);

	/* Find stations that shouldn't be ringing anymore. */
	sla_hangup_stations();
}

/*! \brief Process trunk ring timeouts
 * \note Called with sla.lock locked
 * \return non-zero if a change to the ringing trunks was made
 */
static int sla_calc_trunk_timeouts(unsigned int *timeout)
{
	struct sla_ringing_trunk *ringing_trunk;
	int res = 0;

	ast_mutex_lock(&sla.lock);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_trunks, ringing_trunk, entry) {
		int time_left, time_elapsed;
		if (!ringing_trunk->trunk->ring_timeout)
			continue;
		time_elapsed = ast_tvdiff_ms(ast_tvnow(), ringing_trunk->ring_begin);
		time_left = (ringing_trunk->trunk->ring_timeout * 1000) - time_elapsed;
		if (time_left <= 0) {
			pbx_builtin_setvar_helper(ringing_trunk->trunk->chan, "SLATRUNK_STATUS", "RINGTIMEOUT");
			AST_LIST_REMOVE_CURRENT(entry);
			sla_stop_ringing_trunk(ringing_trunk);
			res = 1;
			sla.ringing_trunks_num--;  /* Added by yfyang */
			continue;
		}
		if (time_left < *timeout)
			*timeout = time_left;
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ast_mutex_unlock(&sla.lock);

	return res;
}

/*! \brief Process station ring timeouts
 * \note Called with sla.lock locked
 * \return non-zero if a change to the ringing stations was made
 */
static int sla_calc_station_timeouts(unsigned int *timeout)
{
	struct sla_ringing_trunk *ringing_trunk;
	struct sla_ringing_station *ringing_station;
	int res = 0;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_stations, ringing_station, entry) {
		unsigned int ring_timeout = 0;
		int time_elapsed, time_left = INT_MAX, final_trunk_time_left = INT_MIN;
		struct sla_trunk_ref *trunk_ref;

		/* If there are any ring timeouts specified for a specific trunk
		 * on the station, then use the highest per-trunk ring timeout.
		 * Otherwise, use the ring timeout set for the entire station. */
		AST_LIST_TRAVERSE(&ringing_station->station->trunks, trunk_ref, entry) {
			struct sla_station_ref *station_ref;
			int trunk_time_elapsed, trunk_time_left;

			AST_LIST_TRAVERSE(&sla.ringing_trunks, ringing_trunk, entry) {
				if (ringing_trunk->trunk == trunk_ref->trunk)
					break;
			}
			if (!ringing_trunk)
				continue;

			/* If there is a trunk that is ringing without a timeout, then the
			 * only timeout that could matter is a global station ring timeout. */
			if (!trunk_ref->ring_timeout)
				break;

			/* This trunk on this station is ringing and has a timeout.
			 * However, make sure this trunk isn't still ringing from a
			 * previous timeout.  If so, don't consider it. */
			AST_LIST_TRAVERSE(&ringing_trunk->timed_out_stations, station_ref, entry) {
				if (station_ref->station == ringing_station->station)
					break;
			}
			if (station_ref)
				continue;

			trunk_time_elapsed = ast_tvdiff_ms(ast_tvnow(), ringing_trunk->ring_begin);
			trunk_time_left = (trunk_ref->ring_timeout * 1000) - trunk_time_elapsed;
			if (trunk_time_left > final_trunk_time_left)
				final_trunk_time_left = trunk_time_left;
		}

		/* No timeout was found for ringing trunks, and no timeout for the entire station */
		if (final_trunk_time_left == INT_MIN && !ringing_station->station->ring_timeout)
			continue;

		/* Compute how much time is left for a global station timeout */
		if (ringing_station->station->ring_timeout) {
			ring_timeout = ringing_station->station->ring_timeout;
			time_elapsed = ast_tvdiff_ms(ast_tvnow(), ringing_station->ring_begin);
			time_left = (ring_timeout * 1000) - time_elapsed;
		}

		/* If the time left based on the per-trunk timeouts is smaller than the
		 * global station ring timeout, use that. */
		if (final_trunk_time_left > INT_MIN && final_trunk_time_left < time_left)
			time_left = final_trunk_time_left;

		/* If there is no time left, the station needs to stop ringing */
		if (time_left <= 0) {
			AST_LIST_REMOVE_CURRENT(entry);
			sla_stop_ringing_station(ringing_station, SLA_STATION_HANGUP_TIMEOUT);
			res = 1;
			sla.ringing_stations_num--;  /* Added by yfyang */
			continue;
		}

		/* There is still some time left for this station to ring, so save that
		 * timeout if it is the first event scheduled to occur */
		if (time_left < *timeout)
			*timeout = time_left;
	}
	AST_LIST_TRAVERSE_SAFE_END;

	return res;
}

/*! \brief Calculate the ring delay for a station
 * \note Assumes sla.lock is locked
 */
static int sla_calc_station_delays(unsigned int *timeout)
{
	struct sla_station *station;
	int res = 0;
	struct ao2_iterator i;

	i = ao2_iterator_init(sla_stations, 0);
	for (; (station = ao2_iterator_next(&i)); ao2_ref(station, -1)) {
		struct sla_ringing_trunk *ringing_trunk;
		int time_left;

		/* Ignore stations already ringing */
		if (sla_check_ringing_station(station))
			continue;

		/* Ignore stations already on a call */
		if (sla_check_inuse_station(station))
			continue;

		/* Ignore stations that don't have one of their trunks ringing */
		if (!(ringing_trunk = sla_choose_ringing_trunk(station, NULL, 0)))
			continue;

		if ((time_left = sla_check_station_delay(station, ringing_trunk)) == INT_MAX)
			continue;

		/* If there is no time left, then the station needs to start ringing.
		 * Return non-zero so that an event will be queued up an event to
		 * make that happen. */
		if (time_left <= 0) {
			res = 1;
			continue;
		}

		if (time_left < *timeout)
			*timeout = time_left;
	}
	ao2_iterator_destroy(&i);

	return res;
}

/*! \brief Calculate the time until the next known event
 *  \note Called with sla.lock locked */
static int sla_process_timers(struct timespec *ts)
{
	unsigned int timeout = UINT_MAX;
	struct timeval wait;
	unsigned int change_made = 0;

	/* Check for ring timeouts on ringing trunks */
	if (sla_calc_trunk_timeouts(&timeout))
		change_made = 1;

	/* Check for ring timeouts on ringing stations */
	if (sla_calc_station_timeouts(&timeout))
		change_made = 1;

	/* Check for station ring delays */
	if (sla_calc_station_delays(&timeout))
		change_made = 1;

	/* queue reprocessing of ringing trunks */
	if (change_made)
		sla_queue_event_nolock(SLA_EVENT_RINGING_TRUNK);

	/* No timeout */
	if (timeout == UINT_MAX)
		return 0;

	if (ts) {
		wait = ast_tvadd(ast_tvnow(), ast_samp2tv(timeout, 1000));
		ts->tv_sec = wait.tv_sec;
		ts->tv_nsec = wait.tv_usec * 1000;
	}

	return 1;
}

static void sla_event_destroy(struct sla_event *event)
{
	if (event->trunk_ref) {
		ao2_ref(event->trunk_ref, -1);
		event->trunk_ref = NULL;
	}

	if (event->station) {
		ao2_ref(event->station, -1);
		event->station = NULL;
	}

	ast_free(event);
}

static void *sla_thread(void *data)
{
	struct sla_failed_station *failed_station;
	struct sla_ringing_station *ringing_station;

	ast_mutex_lock(&sla.lock);

	while (!sla.stop) {
		struct sla_event *event;
		struct timespec ts = { 0, };
		unsigned int have_timeout = 0;

		if (AST_LIST_EMPTY(&sla.event_q)) {
			if ((have_timeout = sla_process_timers(&ts))) {
				ast_cond_timedwait(&sla.cond, &sla.lock, &ts);
			} else {
				ast_cond_wait(&sla.cond, &sla.lock);
			}

			if (sla.stop) {
				break;
			}
		}

		if (have_timeout) {
			sla_process_timers(NULL);
		}

		while ((event = AST_LIST_REMOVE_HEAD(&sla.event_q, entry))) {
			sla.event_q_num--;  /* Added by yfyang */
			ast_mutex_unlock(&sla.lock);
			switch (event->type) {
			case SLA_EVENT_HOLD:
				sla_handle_hold_event(event);
				break;
			case SLA_EVENT_DIAL_STATE:
				sla_handle_dial_state_event();
				break;
			case SLA_EVENT_RINGING_TRUNK:
				sla_handle_ringing_trunk_event();
				break;
			}
			sla_event_destroy(event);
			ast_mutex_lock(&sla.lock);
		}
	}

	ast_mutex_unlock(&sla.lock);

	while ((ringing_station = AST_LIST_REMOVE_HEAD(&sla.ringing_stations, entry))) {
		sla_ringing_station_destroy(ringing_station);
	}

	while ((failed_station = AST_LIST_REMOVE_HEAD(&sla.failed_stations, entry))) {
		sla_failed_station_destroy(failed_station);
	}

	sla.ringing_stations_num = 0;
	sla.failed_stations_num = 0;
	return NULL;
}

struct dial_trunk_args {
	struct sla_trunk_ref *trunk_ref;
	struct sla_station *station;
	ast_mutex_t *cond_lock;
	ast_cond_t *cond;
};

static void *dial_trunk(void *data)
{
	struct dial_trunk_args *args = data;
	struct ast_dial *dial;
	char *tech, *tech_data;
	enum ast_dial_result dial_res;
	char conf_name[MAX_CONFNUM];
	struct sla_conf *conf;
	struct ast_flags64 conf_flags = { 0 };
	RAII_VAR(struct sla_trunk_ref *, trunk_ref, args->trunk_ref, ao2_cleanup);
	RAII_VAR(struct sla_station *, station, args->station, ao2_cleanup);
	int caller_is_saved;
	struct ast_party_caller caller;
	int last_state = 0;
	int current_state = 0;
	int check_count = 600;

	if (!(dial = ast_dial_create())) {
		ast_mutex_lock(args->cond_lock);
		ast_cond_signal(args->cond);
		ast_mutex_unlock(args->cond_lock);
		return NULL;
	}

	tech_data = ast_strdupa(trunk_ref->trunk->device);
	tech = strsep(&tech_data, "/");
	if (ast_dial_append(dial, tech, tech_data, NULL) == -1) {
		ast_mutex_lock(args->cond_lock);
		ast_cond_signal(args->cond);
		ast_mutex_unlock(args->cond_lock);
		ast_dial_destroy(dial);
		return NULL;
	}

	/* Do we need to save of the caller ID data? */
	caller_is_saved = 0;
	if (!sla.attempt_callerid) {
		caller_is_saved = 1;
		caller = *ast_channel_caller(trunk_ref->chan);
		ast_party_caller_init(ast_channel_caller(trunk_ref->chan));
	}

	dial_res = ast_dial_run(dial, trunk_ref->chan, 1);

	/* Restore saved caller ID */
	if (caller_is_saved) {
		ast_party_caller_free(ast_channel_caller(trunk_ref->chan));
		ast_channel_caller_set(trunk_ref->chan, &caller);
	}

	if (dial_res != AST_DIAL_RESULT_TRYING) {
		ast_mutex_lock(args->cond_lock);
		ast_cond_signal(args->cond);
		ast_mutex_unlock(args->cond_lock);
		ast_dial_destroy(dial);
		return NULL;
	}

	for (;;) {
		unsigned int done = 0;
		switch ((dial_res = ast_dial_state(dial))) {
		case AST_DIAL_RESULT_ANSWERED:
			trunk_ref->trunk->chan = ast_dial_answered(dial);
		case AST_DIAL_RESULT_HANGUP:
		case AST_DIAL_RESULT_INVALID:
		case AST_DIAL_RESULT_FAILED:
		case AST_DIAL_RESULT_TIMEOUT:
		case AST_DIAL_RESULT_UNANSWERED:
#ifdef GRANDSTREAM_NETWORKS
		case AST_DIAL_RESULT_PEER_BUSY:
#endif
			done = 1;
			break;
		case AST_DIAL_RESULT_TRYING:
			current_state = AST_CONTROL_PROGRESS;
			break;
		case AST_DIAL_RESULT_RINGING:
		case AST_DIAL_RESULT_PROGRESS:
		case AST_DIAL_RESULT_PROCEEDING:
			current_state = AST_CONTROL_RINGING;
			break;
		}

		if (done) {
			break;
		}

		/* If trunk line state changed, send indication back to originating SLA Station channel */
		if (current_state != last_state) {
			ast_debug(3, "Indicating State Change %d to channel %s\n", current_state, ast_channel_name(trunk_ref->chan));
			ast_indicate(trunk_ref->chan, current_state);
			last_state = current_state;
		}

		/* avoid tight loop... sleep for 1/10th second */
		ast_safe_sleep(trunk_ref->chan, 100);

		/* to avoid endless loop */
		check_count--;
		if (check_count <= 0) {
			break;
		}
	}

	if (!trunk_ref->trunk->chan) {
		ast_mutex_lock(args->cond_lock);
		ast_cond_signal(args->cond);
		ast_mutex_unlock(args->cond_lock);
		ast_dial_join(dial);
		ast_dial_destroy(dial);
		return NULL;
	}

	snprintf(conf_name, sizeof(conf_name), "SLA_%s", trunk_ref->trunk->name);
	conf = sla_conf_build(conf_name, 1, trunk_ref->trunk->chan);

	if (conf) {
		if (trunk_ref->trunk->auto_record) {
			conf->record_flag = 1;
		}

		ast_mutex_lock(args->cond_lock);
		ast_cond_signal(args->cond);
		ast_mutex_unlock(args->cond_lock);

		trunk_ref->conf = conf;
		ast_set_flag64(&conf_flags, SLA_FLAG_MARKEDEXIT | SLA_FLAG_MARKEDUSER | SLA_FLAG_SLA_TRUNK);
		sla_conf_run(trunk_ref->trunk->chan, conf, &conf_flags, NULL);

#ifdef GRANDSTREAM_NETWORKS
		ao2_lock(station); // Added by pxie@grandstream.cn for bug[153768], 2019-12-12.
#endif
		if (1 == sla_conf_dispose(conf, trunk_ref)){  /*Mod by weidu, for conf use after free*/
			trunk_ref->conf = NULL;
		}
#ifdef GRANDSTREAM_NETWORKS
		ao2_unlock(station); // Added by pxie@grandstream.cn for bug[153768], 2019-12-12.

#endif
		conf = NULL;
	}

	/* If the trunk is going away, it is definitely now IDLE. */
	sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);

#ifndef GRANDSTREAM_NETWORKS
	/*! Begin: Added by pxie@grandstream.cn for bug[153768], 2019-12-12. */
	trunk_ref->trunk->chan = NULL;
	trunk_ref->trunk->on_hold = 0;
	/*! Ended: Added by pxie@grandstream.cn for bug[153768], 2019-12-12. */
#endif

	ast_dial_join(dial);
	ast_dial_destroy(dial);
#ifdef GRANDSTREAM_NETWORKS
	/*! Begin: Added by pxie@grandstream.cn for bug[153768], 2019-12-12. */
	trunk_ref->trunk->chan = NULL;
	trunk_ref->trunk->on_hold = 0;
	/*! Ended: Added by pxie@grandstream.cn for bug[153768], 2019-12-12. */
#endif

	return NULL;
}

/*!
 * \brief For a given station, choose the highest priority idle trunk
 * \pre sla_station is locked
 */
static struct sla_trunk_ref *sla_choose_idle_trunk(struct ast_channel *chan, const struct sla_station *station)
{
	struct sla_trunk_ref *trunk_ref = NULL;

	AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
		if (trunk_ref->state == SLA_TRUNK_STATE_IDLE && !sla_check_trunk_calltime_limited(chan, trunk_ref->trunk->device)) {
			ao2_ref(trunk_ref, 1);
			break;
		}
	}

	return trunk_ref;
}

static int sla_station_exec(struct ast_channel *chan, const char *data)
{
	char *station_name, *trunk_name;
	RAII_VAR(struct sla_station *, station, NULL, ao2_cleanup);
	RAII_VAR(struct sla_trunk_ref *, trunk_ref, NULL, ao2_cleanup);
	char conf_name[MAX_CONFNUM];
	struct ast_flags64 conf_flags = { 0 };
	struct sla_conf *conf;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "Invalid Arguments to SLAStation!\n");
		pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "FAILURE");
		return 0;
	}

	trunk_name = ast_strdupa(data);
	station_name = strsep(&trunk_name, "_");

	if (ast_strlen_zero(station_name)) {
		ast_log(LOG_WARNING, "Invalid Arguments to SLAStation!\n");
		pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "FAILURE");
		return 0;
	}

	if(ast_channel_caller(chan)->id.number.valid && ast_channel_caller(chan)->id.number.str)
	{
    	if(strcmp(ast_channel_caller(chan)->id.number.str,station_name) || (strlen(ast_channel_caller(chan)->id.number.str) != strlen(station_name)))
    	{
    	    ast_log(LOG_NOTICE, "No trunks available for call.\n");
    		pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "FAILURE");
    		return 0;
	    }
	}

	station = sla_find_station(station_name);

	if (!station) {
		ast_log(LOG_WARNING, "Station '%s' not found!\n", station_name);
		pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "FAILURE");
		return 0;
	}

	ao2_lock(station);
	if (!ast_strlen_zero(trunk_name)) {
		trunk_ref = sla_find_trunk_ref_byname(chan, station, trunk_name);
	} else {
		trunk_ref = sla_choose_idle_trunk(chan, station);
	}
	ao2_unlock(station);

	if (!trunk_ref || (trunk_ref->state == SLA_TRUNK_STATE_UNAVAILABLE)) {
		if (ast_strlen_zero(trunk_name)) {
			ast_log(LOG_NOTICE, "No trunks available for call.\n");
		} else {
			ast_log(LOG_NOTICE, "Can't join existing call on trunk "
				"'%s' due to access controls.\n", trunk_name);
		}
		pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "CONGESTION");
		return 0;
	}

	if (trunk_ref->state == SLA_TRUNK_STATE_ONHOLD_BYME) {
		if (ast_atomic_dec_and_test((int *) &trunk_ref->trunk->hold_stations) == 1) {
			sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_UP, ALL_TRUNK_REFS, NULL);
		} else {
			trunk_ref->state = SLA_TRUNK_STATE_UP;
			ast_devstate_changed(AST_DEVICE_INUSE, AST_DEVSTATE_CACHABLE,
					     "SLA:%s_%s", station->name, trunk_ref->trunk->name);
		}
	} else if (trunk_ref->state == SLA_TRUNK_STATE_RINGING) {
		struct sla_ringing_trunk *ringing_trunk;

		ast_mutex_lock(&sla.lock);
		AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_trunks, ringing_trunk, entry) {
			if (ringing_trunk->trunk == trunk_ref->trunk) {
				AST_LIST_REMOVE_CURRENT(entry);
				sla.ringing_trunks_num--;  /* Added by yfyang */
				break;
			}
		}
		AST_LIST_TRAVERSE_SAFE_END;
		ast_mutex_unlock(&sla.lock);

		if (ringing_trunk) {
			answer_trunk_chan(ringing_trunk->trunk->chan);
			sla_change_trunk_state(ringing_trunk->trunk, SLA_TRUNK_STATE_UP, ALL_TRUNK_REFS, NULL);

			ao2_ref(ringing_trunk, -1);

			/* Queue up reprocessing ringing trunks, and then ringing stations again */
			sla_queue_event(SLA_EVENT_RINGING_TRUNK);
			sla_queue_event(SLA_EVENT_DIAL_STATE);
		}
	}

	trunk_ref->chan = chan;

	if (!trunk_ref->trunk->chan) {
		ast_mutex_t cond_lock;
		ast_cond_t cond;
		pthread_t dont_care;
		struct dial_trunk_args args = {
			.trunk_ref = trunk_ref,
			.station = station,
			.cond_lock = &cond_lock,
			.cond = &cond,
		};
		ao2_ref(trunk_ref, 1);
		ao2_ref(station, 1);
		sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_UP, ALL_TRUNK_REFS, NULL);
		/* Create a thread to dial the trunk and dump it into the conference.
		 * However, we want to wait until the trunk has been dialed and the
		 * conference is created before continuing on here. */
		ast_autoservice_start(chan);
		ast_mutex_init(&cond_lock);
		ast_cond_init(&cond, NULL);
		ast_mutex_lock(&cond_lock);
		ast_pthread_create_detached_background(&dont_care, NULL, dial_trunk, &args);
		ast_cond_wait(&cond, &cond_lock);
		ast_mutex_unlock(&cond_lock);
		ast_mutex_destroy(&cond_lock);
		ast_cond_destroy(&cond);
		ast_autoservice_stop(chan);
		if (!trunk_ref->trunk->chan) {
			ast_debug(1, "Trunk didn't get created. chan: %lx\n", (unsigned long) trunk_ref->trunk->chan);
			pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "CONGESTION");

			if ((trunk_ref->trunk->active_stations == 0) && (trunk_ref->trunk->hold_stations == 0)) {
				sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);
			}
			trunk_ref->chan = NULL;
			return 0;
		}
	}

	snprintf(conf_name, sizeof(conf_name), "SLA_%s", trunk_ref->trunk->name);
	conf = sla_conf_build(conf_name, 1, chan);
	if (!conf) {
		trunk_ref->chan = NULL;
		sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);
		pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "FAILURE");
		return 0;
	}

	if (ast_atomic_fetchadd_int((int *) &trunk_ref->trunk->active_stations, 1) == 0 &&
		trunk_ref->trunk->on_hold) {
		trunk_ref->trunk->on_hold = 0;
		ast_indicate(trunk_ref->trunk->chan, AST_CONTROL_UNHOLD);
		sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_UP, ALL_TRUNK_REFS, NULL);
	}
	if (trunk_ref->trunk->auto_record) {
		conf->record_flag = 1;
 	}
	 
	ast_answer(chan);
	trunk_ref->conf = conf;

	ast_set_flag64(&conf_flags, SLA_FLAG_MARKEDEXIT | SLA_FLAG_SLA_STATION);
	sla_conf_run(chan, conf, &conf_flags, NULL);

	trunk_ref->chan = NULL;
	if (ast_atomic_dec_and_test((int *) &trunk_ref->trunk->active_stations) &&
		trunk_ref->state != SLA_TRUNK_STATE_ONHOLD_BYME) {
		if (trunk_ref->conf) {
			ao2_callback(trunk_ref->conf->usercontainer, OBJ_NODATA, sla_conf_user_set_exit, NULL);
		}

		trunk_ref->trunk->hold_stations = 0;
		sla_change_trunk_state(trunk_ref->trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);
	}

#ifdef GRANDSTREAM_NETWORKS
	ao2_lock(station); // Added by pxie@grandstream.cn for bug[153768], 2019-12-12.
#endif
	if (1 == sla_conf_dispose(conf, trunk_ref)){  /*Mod by weidu, for conf use after free*/
		trunk_ref->conf = NULL;
	}
#ifdef GRANDSTREAM_NETWORKS
	ao2_unlock(station);
#endif
	conf = NULL;


	pbx_builtin_setvar_helper(chan, "SLASTATION_STATUS", "SUCCESS");

	return 0;
}

static void sla_ringing_trunk_destroy(void *obj)
{
	struct sla_ringing_trunk *ringing_trunk = obj;

	if (ringing_trunk->trunk) {
		ao2_ref(ringing_trunk->trunk, -1);
		ringing_trunk->trunk = NULL;
	}

	ast_log(LOG_NOTICE, "ringing_trunk[%p]", ringing_trunk);
}

static struct sla_ringing_trunk *queue_ringing_trunk(struct sla_trunk *trunk)
{
	struct sla_ringing_trunk *ringing_trunk;

	if (!(ringing_trunk = ao2_alloc(sizeof(*ringing_trunk), sla_ringing_trunk_destroy))) {
		return NULL;
	}

	ao2_ref(trunk, 1);
	ringing_trunk->trunk = trunk;
	ringing_trunk->ring_begin = ast_tvnow();

	sla_change_trunk_state(trunk, SLA_TRUNK_STATE_RINGING, ALL_TRUNK_REFS, NULL);

	ast_mutex_lock(&sla.lock);
	AST_LIST_INSERT_HEAD(&sla.ringing_trunks, ringing_trunk, entry);
	sla.ringing_trunks_num++;     /* Added by yfyang */
	ast_mutex_unlock(&sla.lock);

	sla_queue_event(SLA_EVENT_RINGING_TRUNK);

	return ringing_trunk;
}

static int sla_check_ring_stations(struct sla_ringing_trunk *ringing_trunk)
{
	struct sla_station_ref *station_ref;
	int num = 0;

	if (!ringing_trunk || !ringing_trunk->trunk) {
		return num;
	}

	AST_LIST_TRAVERSE(&ringing_trunk->trunk->stations, station_ref, entry) {
		int time_left;

		/* Is this station already ringing? */
		if (sla_check_ringing_station(station_ref->station)) {
			num++;
			continue;
		}

		/* Is this station already in a call? */
		if (sla_check_inuse_station(station_ref->station)) {
			num++;
			continue;
		}

		/* Check for a ring delay in progress */
		time_left = sla_check_station_delay(station_ref->station, ringing_trunk);
		if (time_left != INT_MAX && time_left > 0) {
			num++;
			continue;
		}
	}

	return num;
}

static int sla_trunk_exec(struct ast_channel *chan, const char *data)
{
	char conf_name[MAX_CONFNUM];
	struct sla_conf *conf;
	struct ast_flags64 conf_flags = { 0 };
	struct ast_frame *f;
	RAII_VAR(struct sla_trunk *, trunk, NULL, ao2_cleanup);
	RAII_VAR(struct sla_ringing_trunk *, ringing_trunk, NULL, ao2_cleanup);
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(trunk_name);
		AST_APP_ARG(options);
	);
	char *opts[SLA_TRUNK_OPT_ARG_ARRAY_SIZE] = { NULL, };
	struct ast_flags opt_flags = { 0 };
	char *parse;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_ERROR, "The SLATrunk application requires an argument, the trunk name\n");
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (args.argc == 2) {
		if (ast_app_parse_options(sla_trunk_opts, &opt_flags, opts, args.options)) {
			ast_log(LOG_ERROR, "Error parsing options for SLATrunk\n");
			return -1;
		}
	}

	trunk = sla_find_trunk(args.trunk_name);

	if (!trunk) {
		ast_log(LOG_ERROR, "SLA Trunk '%s' not found!\n", args.trunk_name);
		pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "FAILURE");
		return 0;
	}

	if (trunk->chan) {
		ast_log(LOG_ERROR, "Call came in on %s, but the trunk is already in use!\n",
			args.trunk_name);
		pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "FAILURE");
		return 0;
	}

	trunk->chan = chan;

	if (!(ringing_trunk = queue_ringing_trunk(trunk))) {
		pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "FAILURE");
		return 0;
	}
	ao2_ref(ringing_trunk, +1);

	snprintf(conf_name, sizeof(conf_name), "SLA_%s", args.trunk_name);
	conf = sla_conf_build(conf_name, 1, chan);
	if (!conf) {
		pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "FAILURE");
		return 0;
	}

	if (trunk->auto_record) {
		conf->record_flag = 1;
	} 

	ast_set_flag64(&conf_flags, SLA_FLAG_MARKEDEXIT | SLA_FLAG_MARKEDUSER | SLA_FLAG_SLA_TRUNK);

	if (ast_test_flag(&opt_flags, SLA_TRUNK_OPT_MOH)) {
		ast_indicate(chan, -1);
		ast_set_flag64(&conf_flags, SLA_FLAG_MOH);
	} else {
		ast_indicate(chan, AST_CONTROL_RINGING);
	}

	/* wait until some user enter the conf */
	int check_num = 300*100;
	while (1) {
		usleep(20000); /*Mod by weidu, for 212818*/

		if (ast_check_hangup(chan)) {
			ast_log(LOG_WARNING, "caller channel[%s] has hangup\n", ast_channel_name(chan));
			goto clear_dial;
		}
		
		if (conf->user_count > 0) {
			ast_log(LOG_NOTICE, "answer the caller chan[%s]\n", ast_channel_name(chan));
			ast_answer(chan);
			break;
		}

		check_num--;
		if (check_num <= 0) {
			/* if not answered in 60s, clear dial */
			ast_log(LOG_NOTICE, "check timeout, clear dial, chan[%s]\n", ast_channel_name(chan));
			pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "UNANSWERED");
			goto clear_dial;
		}

		f = ast_read(chan);
		if (!f || ((f->frametype == AST_FRAME_CONTROL) && (f->subclass.integer == AST_CONTROL_HANGUP))) {
			if (f) {
				ast_frfree(f);
			}

			ast_log(LOG_WARNING, "caller channel has hangup\n");
			goto clear_dial;
		}
		ast_frfree(f);

		ast_mutex_lock(&sla.lock);
		if (sla_check_ring_stations(ringing_trunk) <= 0) {
			ast_mutex_unlock(&sla.lock);
			ast_log(LOG_NOTICE, "No station ringing now.\n");

			if (ast_strlen_zero(pbx_builtin_getvar_helper(chan, "SLATRUNK_STATUS"))) {
				pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "UNANSWERED");
			}

			goto clear_dial;
		} 
		ast_mutex_unlock(&sla.lock);
	}

	//trunk_ref->conf = conf;
	sla_conf_run(chan, conf, &conf_flags, opts);
	pbx_builtin_setvar_helper(chan, "SLATRUNK_STATUS", "SUCCESS");

clear_dial:
	sla_conf_dispose(conf, NULL);
	conf = NULL;
	trunk->chan = NULL;
	trunk->on_hold = 0;

	sla_change_trunk_state(trunk, SLA_TRUNK_STATE_IDLE, ALL_TRUNK_REFS, NULL);

	/* Remove the entry from the list of ringing trunks if it is still there. */
	ast_mutex_lock(&sla.lock);
	struct sla_ringing_trunk *tmp_ringing_trunk;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&sla.ringing_trunks, tmp_ringing_trunk, entry) {
		if (tmp_ringing_trunk->trunk == trunk) {
			AST_LIST_REMOVE_CURRENT(entry);
			sla.ringing_trunks_num--;  
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ast_mutex_unlock(&sla.lock);
	if (tmp_ringing_trunk) {
		ao2_ref(tmp_ringing_trunk, -1);
		/* Queue reprocessing of ringing trunks to make stations stop ringing
		 * that shouldn't be ringing after this trunk stopped. */
		sla_queue_event(SLA_EVENT_RINGING_TRUNK);
	}

	return 0;
}

static enum ast_device_state sla_state(const char *data)
{
	char *buf, *station_name, *trunk_name;
	RAII_VAR(struct sla_station *, station, NULL, ao2_cleanup);
	struct sla_trunk_ref *trunk_ref;
	enum ast_device_state res = AST_DEVICE_INVALID;

	trunk_name = buf = ast_strdupa(data);
	station_name = strsep(&trunk_name, "_");

	station = sla_find_station(station_name);
	if (station) {
		ao2_lock(station);
		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			if (!strcasecmp(trunk_name, trunk_ref->trunk->name)) {
				res = sla_state_to_devstate(trunk_ref->state);
				break;
			}
		}
		ao2_unlock(station);
	}

	if (res == AST_DEVICE_INVALID) {
		ast_log(LOG_ERROR, "Could not determine state for trunk %s on station %s!\n",
			trunk_name, station_name);
	}

	return res;
}

#define ____SLA_OBJECT_MANAGEMENT_START____

static void sla_trunk_ref_destructor(void *obj)
{
	struct sla_trunk_ref *trunk_ref = obj;

	if (trunk_ref->trunk) {
		ao2_ref(trunk_ref->trunk, -1);
		trunk_ref->trunk = NULL;
	}
}

static struct sla_trunk_ref *create_trunk_ref(struct sla_trunk *trunk)
{
	struct sla_trunk_ref *trunk_ref;

	if (!(trunk_ref = ao2_alloc(sizeof(*trunk_ref), sla_trunk_ref_destructor))) {
		return NULL;
	}

	ao2_ref(trunk, 1);
	trunk_ref->trunk = trunk;

	return trunk_ref;
}

static int sla_trunk_release_refs(void *obj, void *arg, int flags)
{
	struct sla_trunk *trunk = obj;
	struct sla_station_ref *station_ref;

	while ((station_ref = AST_LIST_REMOVE_HEAD(&trunk->stations, entry))) {
		ao2_ref(station_ref, -1);
	}
	return 0;
}

static int sla_station_release_refs(void *obj, void *arg, int flags)
{
	struct sla_station *station = obj;
	struct sla_trunk_ref *trunk_ref;

	while ((trunk_ref = AST_LIST_REMOVE_HEAD(&station->trunks, entry))) {
		ao2_ref(trunk_ref, -1);
	}

	return 0;
}

static void sla_station_destructor(void *obj)
{
	struct sla_station *station = obj;

	ast_debug(1, "sla_station destructor for '%s'\n", station->name);

	if (!ast_strlen_zero(station->autocontext)) {
		struct sla_trunk_ref *trunk_ref;

		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			char exten[AST_MAX_EXTENSION];
			char hint[AST_MAX_APP];
			snprintf(exten, sizeof(exten), "%s_%s", station->name, trunk_ref->trunk->name);
			snprintf(hint, sizeof(hint), "SLA:%s", exten);
			ast_context_remove_extension(station->autocontext, exten,
				1, sla_registrar);
			ast_context_remove_extension(station->autocontext, hint,
				PRIORITY_HINT, sla_registrar);
		}
	}

	sla_station_release_refs(station, NULL, 0);

	ast_string_field_free_memory(station);
}

static int sla_trunk_cmp(void *obj, void *arg, int flags)
{
	struct sla_trunk *trunk = obj, *trunk2 = arg;

	return !strcasecmp(trunk->name, trunk2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static int sla_station_cmp(void *obj, void *arg, int flags)
{
	struct sla_station *station = obj, *station2 = arg;

	return !strcasecmp(station->name, station2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static int sla_trunk_ref_cmp(unsigned int cur_id, unsigned int new_id)
{
	if (cur_id > new_id) {
		return 1;
	} else if (cur_id < new_id) {
		return -1;
	}

	return 0;
}

static void sla_destroy(void)
{
	if (sla.thread != AST_PTHREADT_NULL) {
		ast_mutex_lock(&sla.lock);
		sla.stop = 1;
		ast_cond_signal(&sla.cond);
		ast_mutex_unlock(&sla.lock);
		pthread_join(sla.thread, NULL);
	}

	/* Drop any created contexts from the dialplan */
	ast_context_destroy(NULL, sla_registrar);

	ast_mutex_destroy(&sla.lock);
	ast_cond_destroy(&sla.cond);

	ao2_callback(sla_trunks, 0, sla_trunk_release_refs, NULL);
	ao2_callback(sla_stations, 0, sla_station_release_refs, NULL);

	ao2_ref(sla_trunks, -1);
	sla_trunks = NULL;

	ao2_ref(sla_stations, -1);
	sla_stations = NULL;
}

static int sla_check_device(const char *device)
{
	char *tech, *tech_data;

	tech_data = ast_strdupa(device);
	tech = strsep(&tech_data, "/");

	if (ast_strlen_zero(tech) || ast_strlen_zero(tech_data))
		return -1;

	return 0;
}

static void sla_trunk_destructor(void *obj)
{
	struct sla_trunk *trunk = obj;

	ast_debug(1, "sla_trunk destructor for '%s'\n", trunk->name);

	if (!ast_strlen_zero(trunk->autocontext)) {
		ast_context_remove_extension(trunk->autocontext, "s", 1, sla_registrar);
	}

	sla_trunk_release_refs(trunk, NULL, 0);

	ast_string_field_free_memory(trunk);
}

static int sla_build_trunk(struct ast_config *cfg, const char *cat)
{
	RAII_VAR(struct sla_trunk *, trunk, NULL, ao2_cleanup);
	struct ast_variable *var;
	const char *dev;
	int existing_trunk = 0;

	if (!(dev = ast_variable_retrieve(cfg, cat, "device"))) {
		ast_log(LOG_ERROR, "SLA Trunk '%s' defined with no device!\n", cat);
		return -1;
	}

	if (sla_check_device(dev)) {
		ast_log(LOG_ERROR, "SLA Trunk '%s' defined with invalid device '%s'!\n",
			cat, dev);
		return -1;
	}

	if ((trunk = sla_find_trunk(cat))) {
		trunk->mark = 0;
		existing_trunk = 1;
	} else if ((trunk = ao2_alloc(sizeof(*trunk), sla_trunk_destructor))) {
		if (ast_string_field_init(trunk, 32)) {
			return -1;
		}
		ast_string_field_set(trunk, name, cat);
	} else {
		return -1;
	}

	ao2_lock(trunk);

	ast_string_field_set(trunk, device, dev);

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "autocontext"))
			ast_string_field_set(trunk, autocontext, var->value);
		else if (!strcasecmp(var->name, "ringtimeout")) {
			if (sscanf(var->value, "%30u", &trunk->ring_timeout) != 1) {
				ast_log(LOG_WARNING, "Invalid ringtimeout '%s' specified for trunk '%s'\n",
					var->value, trunk->name);
				trunk->ring_timeout = 0;
			}
		} else if (!strcasecmp(var->name, "barge"))
			trunk->barge_disabled = ast_false(var->value);
		else if (!strcasecmp(var->name, "hold")) {
			if (!strcasecmp(var->value, "private"))
				trunk->hold_access = SLA_HOLD_PRIVATE;
			else if (!strcasecmp(var->value, "open"))
				trunk->hold_access = SLA_HOLD_OPEN;
			else {
				ast_log(LOG_WARNING, "Invalid value '%s' for hold on trunk %s\n",
					var->value, trunk->name);
			}
		}
		/* Added by yfyang, grandstream, Bug 25633 */
		else if (!strcasecmp(var->name, "autorecord")) {
			trunk->auto_record = ast_true(var->value);
		} else if (strcasecmp(var->name, "type") && strcasecmp(var->name, "device")) {
			ast_log(LOG_ERROR, "Invalid option '%s' specified at line %d of %s!\n",
				var->name, var->lineno, SLA_CONFIG_FILE);
		}
	}

	ao2_unlock(trunk);

	if (!ast_strlen_zero(trunk->autocontext)) {
		if (!ast_context_find_or_create(NULL, NULL, trunk->autocontext, sla_registrar)) {
			ast_log(LOG_ERROR, "Failed to automatically find or create "
				"context '%s' for SLA!\n", trunk->autocontext);
			return -1;
		}

		if (ast_add_extension(trunk->autocontext, 0 /* don't replace */, "s", 1,
			NULL, NULL, slatrunk_app, ast_strdup(trunk->name), ast_free_ptr, sla_registrar)) {
			ast_log(LOG_ERROR, "Failed to automatically create extension "
				"for trunk '%s'!\n", trunk->name);
			return -1;
		}
	}

	if (!existing_trunk) {
		trunk->trunk_state = sla_devstate_to_slastate(ast_device_state(trunk->device));
		ao2_link(sla_trunks, trunk);
	}

	return 0;
}

/*!
 * \internal
 * \pre station is not locked
 */
static void sla_add_trunk_to_station(struct sla_station *station, struct ast_variable *var, unsigned int order)
{
	RAII_VAR(struct sla_trunk *, trunk, NULL, ao2_cleanup);
	struct sla_trunk_ref *trunk_ref = NULL;
	struct sla_station_ref *station_ref;
	char *trunk_name, *options, *cur;
	int existing_trunk_ref = 0;
	int existing_station_ref = 0;
	int need_insert = 0;

	options = ast_strdupa(var->value);
	trunk_name = strsep(&options, ",");

	trunk = sla_find_trunk(trunk_name);
	if (!trunk) {
		ast_log(LOG_ERROR, "Trunk '%s' not found!\n", var->value);
		return;
	}

	ao2_lock(station);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&station->trunks, trunk_ref, entry) {
		if (trunk_ref->trunk == trunk) {
			trunk_ref->mark = 0;
			existing_trunk_ref = 1;
			
			if (trunk_ref->order != order) {
				AST_LIST_REMOVE_CURRENT(entry);
				need_insert = 1;
			}
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	ao2_unlock(station);

	if (!trunk_ref && !(trunk_ref = create_trunk_ref(trunk))) {
		return;
	}
	
	if (!existing_trunk_ref) {
		trunk_ref->state = trunk->trunk_state;
	}

	trunk_ref->order = order;
	while ((cur = strsep(&options, ","))) {
		char *name, *value = cur;
		name = strsep(&value, "=");
		if (!strcasecmp(name, "ringtimeout")) {
			if (sscanf(value, "%30u", &trunk_ref->ring_timeout) != 1) {
				ast_log(LOG_WARNING, "Invalid ringtimeout value '%s' for "
					"trunk '%s' on station '%s'\n", value, trunk->name, station->name);
				trunk_ref->ring_timeout = 0;
			}
		} else if (!strcasecmp(name, "ringdelay")) {
			if (sscanf(value, "%30u", &trunk_ref->ring_delay) != 1) {
				ast_log(LOG_WARNING, "Invalid ringdelay value '%s' for "
					"trunk '%s' on station '%s'\n", value, trunk->name, station->name);
				trunk_ref->ring_delay = 0;
			}
		} else {
			ast_log(LOG_WARNING, "Invalid option '%s' for "
				"trunk '%s' on station '%s'\n", name, trunk->name, station->name);
		}
	}

	AST_LIST_TRAVERSE(&trunk->stations, station_ref, entry) {
		if (station_ref->station == station) {
			station_ref->mark = 0;
			existing_station_ref = 1;
			break;
		}
	}

	if (!station_ref && !(station_ref = sla_create_station_ref(station))) {
		if (!existing_trunk_ref) {
			ao2_ref(trunk_ref, -1);
		} else {
			trunk_ref->mark = 1;
		}
		return;
	}

	if (!existing_station_ref) {
		ao2_lock(trunk);
		AST_LIST_INSERT_TAIL(&trunk->stations, station_ref, entry);
		ast_atomic_fetchadd_int((int *) &trunk->num_stations, 1);
		ao2_unlock(trunk);
	}

	if (!existing_trunk_ref || need_insert) {
		ao2_lock(station);
		AST_LIST_INSERT_SORTFUNC(&station->trunks, trunk_ref, entry, order, sla_trunk_ref_cmp);
		ao2_unlock(station);
	}
}

static int sla_build_station(struct ast_config *cfg, const char *cat)
{
	RAII_VAR(struct sla_station *, station, NULL, ao2_cleanup);
	struct ast_variable *var;
	const char *dev;
	int existing_station = 0;
	unsigned int trunk_ref_order = 0;

	if (!(dev = ast_variable_retrieve(cfg, cat, "device"))) {
		ast_log(LOG_ERROR, "SLA Station '%s' defined with no device!\n", cat);
		return -1;
	}

	if ((station = sla_find_station(cat))) {
		station->mark = 0;
		existing_station = 1;
	} else if ((station = ao2_alloc(sizeof(*station), sla_station_destructor))) {
		if (ast_string_field_init(station, 32)) {
			return -1;
		}
		ast_string_field_set(station, name, cat);
	} else {
		return -1;
	}

	ao2_lock(station);

	ast_string_field_set(station, device, dev);

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "trunk")) {
			ao2_unlock(station);
			sla_add_trunk_to_station(station, var, trunk_ref_order);
			trunk_ref_order++;
			ao2_lock(station);
		} else if (!strcasecmp(var->name, "autocontext")) {
			ast_string_field_set(station, autocontext, var->value);
		} else if (!strcasecmp(var->name, "ringtimeout")) {
			if (sscanf(var->value, "%30u", &station->ring_timeout) != 1) {
				ast_log(LOG_WARNING, "Invalid ringtimeout '%s' specified for station '%s'\n",
					var->value, station->name);
				station->ring_timeout = 0;
			}
		} else if (!strcasecmp(var->name, "ringdelay")) {
			if (sscanf(var->value, "%30u", &station->ring_delay) != 1) {
				ast_log(LOG_WARNING, "Invalid ringdelay '%s' specified for station '%s'\n",
					var->value, station->name);
				station->ring_delay = 0;
			}
		} else if (!strcasecmp(var->name, "hold")) {
			if (!strcasecmp(var->value, "private"))
				station->hold_access = SLA_HOLD_PRIVATE;
			else if (!strcasecmp(var->value, "open"))
				station->hold_access = SLA_HOLD_OPEN;
			else {
				ast_log(LOG_WARNING, "Invalid value '%s' for hold on station %s\n",
					var->value, station->name);
			}

		} else if (strcasecmp(var->name, "type") && strcasecmp(var->name, "device")) {
			ast_log(LOG_ERROR, "Invalid option '%s' specified at line %d of %s!\n",
				var->name, var->lineno, SLA_CONFIG_FILE);
		}
	}

	ao2_unlock(station);

	if (!ast_strlen_zero(station->autocontext)) {
		struct sla_trunk_ref *trunk_ref;

		if (!ast_context_find_or_create(NULL, NULL, station->autocontext, sla_registrar)) {
			ast_log(LOG_ERROR, "Failed to automatically find or create "
				"context '%s' for SLA!\n", station->autocontext);
			return -1;
		}
		/* The extension for when the handset goes off-hook.
		 * exten => station1,1,SLAStation(station1) */
		if (ast_add_extension(station->autocontext, 0 /* don't replace */, station->name, 1,
			NULL, NULL, slastation_app, ast_strdup(station->name), ast_free_ptr, sla_registrar)) {
			ast_log(LOG_ERROR, "Failed to automatically create extension "
				"for trunk '%s'!\n", station->name);
			return -1;
		}
		AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
			char exten[AST_MAX_EXTENSION];
			char hint[AST_MAX_APP];
			snprintf(exten, sizeof(exten), "%s_%s", station->name, trunk_ref->trunk->name);
			snprintf(hint, sizeof(hint), "SLA:%s", exten);
			/* Extension for this line button
			 * exten => station1_line1,1,SLAStation(station1_line1) */
			if (ast_add_extension(station->autocontext, 0 /* don't replace */, exten, 1,
				NULL, NULL, slastation_app, ast_strdup(exten), ast_free_ptr, sla_registrar)) {
				ast_log(LOG_ERROR, "Failed to automatically create extension "
					"for trunk '%s'!\n", station->name);
				return -1;
			}
			/* Hint for this line button
			 * exten => station1_line1,hint,SLA:station1_line1 */
			if (ast_add_extension(station->autocontext, 0 /* don't replace */, exten, PRIORITY_HINT,
				NULL, NULL, hint, NULL, NULL, sla_registrar)) {
				ast_log(LOG_ERROR, "Failed to automatically create hint "
					"for trunk '%s'!\n", station->name);
				return -1;
			}
		}
	}

	if (!existing_station) {
		ao2_link(sla_stations, station);
	}

	return 0;
}

static int sla_trunk_mark(void *obj, void *arg, int flags)
{
	struct sla_trunk *trunk = obj;
	struct sla_station_ref *station_ref;

	ao2_lock(trunk);

	trunk->mark = 1;

	AST_LIST_TRAVERSE(&trunk->stations, station_ref, entry) {
		station_ref->mark = 1;
	}

	ao2_unlock(trunk);

	return 0;
}

static int sla_station_mark(void *obj, void *arg, int flags)
{
	struct sla_station *station = obj;
	struct sla_trunk_ref *trunk_ref;

	ao2_lock(station);

	station->mark = 1;

	AST_LIST_TRAVERSE(&station->trunks, trunk_ref, entry) {
		trunk_ref->mark = 1;
	}

	ao2_unlock(station);

	return 0;
}

static int sla_trunk_is_marked(void *obj, void *arg, int flags)
{
	struct sla_trunk *trunk = obj;

	ao2_lock(trunk);

	if (trunk->mark) {
		/* Only remove all of the station references if the trunk itself is going away */
		sla_trunk_release_refs(trunk, NULL, 0);
	} else {
		struct sla_station_ref *station_ref;

		/* Otherwise only remove references to stations no longer in the config */
		AST_LIST_TRAVERSE_SAFE_BEGIN(&trunk->stations, station_ref, entry) {
			if (!station_ref->mark) {
				continue;
			}
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_ref(station_ref, -1);
		}
		AST_LIST_TRAVERSE_SAFE_END
	}

	ao2_unlock(trunk);

	return trunk->mark ? CMP_MATCH : 0;
}

static int sla_station_is_marked(void *obj, void *arg, int flags)
{
	struct sla_station *station = obj;

	ao2_lock(station);

	if (station->mark) {
		/* Only remove all of the trunk references if the station itself is going away */
#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug 190851 */
		struct sla_trunk_ref *trunk_ref;
		AST_LIST_TRAVERSE_SAFE_BEGIN(&station->trunks, trunk_ref, entry) {
			ast_devstate_changed(AST_DEVICE_UNAVAILABLE, AST_DEVSTATE_CACHABLE, "SLA:%s_%s", station->name, trunk_ref->trunk->name);
		}
		AST_LIST_TRAVERSE_SAFE_END
#endif
		sla_station_release_refs(station, NULL, 0);
	} else {
		struct sla_trunk_ref *trunk_ref;

		/* Otherwise only remove references to trunks no longer in the config */
		AST_LIST_TRAVERSE_SAFE_BEGIN(&station->trunks, trunk_ref, entry) {
			if (!trunk_ref->mark) {
				ast_devstate_changed(sla_state_to_devstate(trunk_ref->state), AST_DEVSTATE_CACHABLE,
					     "SLA:%s_%s", station->name, trunk_ref->trunk->name);
				continue;
			}
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_ref(trunk_ref, -1);
		}
		AST_LIST_TRAVERSE_SAFE_END
	}

	ao2_unlock(station);

	return station->mark ? CMP_MATCH : 0;
}

#define ____SLA_MODULE_LOAD_UNLOAD____

static int sla_in_use(void)
{
	return ao2_container_count(sla_trunks) || ao2_container_count(sla_stations);
}

static int sla_load_config(int reload)
{
	struct ast_config *cfg;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	const char *cat = NULL;
	int res = 0;
	const char *val;

	if (!reload) {
		ast_mutex_init(&sla.lock);
		ast_cond_init(&sla.cond, NULL);
		sla_trunks = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
			NULL, sla_trunk_cmp);
		sla_stations = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
			NULL, sla_station_cmp);
	}

	if (!(cfg = ast_config_load(SLA_CONFIG_FILE, config_flags))) {
		return 0; /* Treat no config as normal */
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file " SLA_CONFIG_FILE " is in an invalid format.  Aborting.\n");
		return 0;
	}

	if (reload) {
		ao2_callback(sla_trunks, 0, sla_trunk_mark, NULL);
		ao2_callback(sla_stations, 0, sla_station_mark, NULL);
	}

	if ((val = ast_variable_retrieve(cfg, "general", "attemptcallerid")))
		sla.attempt_callerid = ast_true(val);

	if ((val = ast_variable_retrieve(cfg, "general", "recordprompt"))) {
		gs_record_prompt = ast_true(val);
	}

	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		const char *type;
		if (!strcasecmp(cat, "general"))
			continue;
		if (!(type = ast_variable_retrieve(cfg, cat, "type"))) {
			ast_log(LOG_WARNING, "Invalid entry in %s defined with no type!\n",
				SLA_CONFIG_FILE);
			continue;
		}
		if (!strcasecmp(type, "trunk"))
			res = sla_build_trunk(cfg, cat);
		else if (!strcasecmp(type, "station"))
			res = sla_build_station(cfg, cat);
		else {
			ast_log(LOG_WARNING, "Entry in %s defined with invalid type '%s'!\n",
				SLA_CONFIG_FILE, type);
		}
	}

	ast_config_destroy(cfg);

	if (reload) {
		ao2_callback(sla_trunks, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, sla_trunk_is_marked, NULL);
		ao2_callback(sla_stations, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, sla_station_is_marked, NULL);
	}

	/* Start SLA event processing thread once SLA has been configured. */
	if (sla.thread == AST_PTHREADT_NULL && sla_in_use()) {
		ast_pthread_create(&sla.thread, NULL, sla_thread, NULL);
	}

	return res;
}


static int load_config(int reload)
{
	return sla_load_config(reload);
}

static int unload_module(void)
{
	int res = 0;

#ifndef BUILD_ON_CLOUD_PBX
	ast_cli_unregister_multiple(cli_sla, ARRAY_LEN(cli_sla));
	res |= ast_unregister_application(slastation_app);
	res |= ast_unregister_application(slatrunk_app);

	ast_devstate_prov_del("SLA");

	sla_state_sub = stasis_unsubscribe_and_join(sla_state_sub);
	sla_destroy();
#endif
	return res;
}

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	int res = 0;

#ifndef BUILD_ON_CLOUD_PBX
	res |= load_config(0);

	ast_cli_register_multiple(cli_sla, ARRAY_LEN(cli_sla));
	res |= ast_register_application_xml(slastation_app, sla_station_exec);
	res |= ast_register_application_xml(slatrunk_app, sla_trunk_exec);

	res |= ast_devstate_prov_add("SLA", sla_state);

	sla_state_sub = stasis_subscribe(ast_device_state_topic_all(), sla_state_cb, NULL);
	if (!sla_state_sub) {
		return -1;
	}
#endif
	return res;
}

static int reload(void)
{
#ifndef BUILD_ON_CLOUD_PBX
	return load_config(1);
#else
	return 0;
#endif
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "SLA Support",
		.support_level = AST_MODULE_SUPPORT_EXTENDED,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_APP_DEPEND,
);

