#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, Grandstream Networks.
 *
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 * Please follow coding guidelines
 * http://svn.digium.com/view/asterisk/trunk/doc/CODING-GUIDELINES
 */

/*! \file
 *
 * \brief Confbridge resource manager
 *
 * \author\Grandstream Networks
 *
 * This file contains functions that are used for managing the media
 * resource in sfu mode
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/
#include "asterisk.h"

#include "asterisk/logger.h"
#include "asterisk/config.h"
#include "asterisk/config_options.h"
#include "asterisk/astobj2.h"
#include "asterisk/astdb.h"
#include "asterisk/cli.h"
#include "asterisk/stringfields.h"
#include "asterisk/channel.h"
#include "asterisk/bfcp.h"
#include "include/confbridge.h"
#include "include/conf_resource.h"

AST_MODULE_LOG("Bridge");

#define SFU_DEFAULT_MAX_ROOMS     (1)
#define SFU_DEFAULT_MAX_USERS     (15)
#define SFU_DEFAULT_SD_NUM        (3)
#define SFU_DEFAULT_HD_NUM        (1)
#define SFU_DEFAULT_SHARE_NUM     (1)
#define SFU_DEFAULT_HD_BANDWIDTH  (1230)
#define SFU_DEFAULT_SD_BANDWIDTH  (250)
#define SFU_DEFAULT_MAX_VIDEO_RECORD_ROOMS     (3)
#define SFU_DEFAULT_MAX_VIDEO_ENCODE_CAP       (60)
#define SFU_DEFAULT_VIDEO_ENCODE_FRAME         (15)
#define SFU_DEFAULT_SLIDES_ENCODE_FRAME        (3)
#define SFU_DEFAULT_FOUCS_ENCODE_FRAME         (0)

#define SFU_CALL_ID "SFU_CALL_ID"

#define MAX_PVALUE_LENGTH 512
#define NET_WAN   ":wan_device"
#define NET_LAN   ":lan_device"
#define NET_LAN1  ":lan1_device"
#define NET_LAN2  ":lan2_device"
#define DEFAULT_IF  "default_if"

/*! \brief Callbacks for vector deletes */
#define RESOURCE_DELETE_CB(r) ({ \
	reset_resource(r); \
	ast_free(r); \
})

/*! \brief Callbacks for vector initialization */
#define RESOURCE_INIT_CB(r, t) ({ \
	r = ast_calloc(sizeof(*r), 1); \
	r->type = t; \
	r->conference_number = NULL; \
	r->inuse = 0; \
	r->id = NULL; \
})

struct video_attribute {
	int slides_type;
	enum video_resolution resolution;
	char codec[8];
	char slider_name[128];
};

struct audio_attribute {
};

struct resource_owner {
	char *channel_name;
	char *caller_number;
	AST_LIST_ENTRY(resource_owner) entry;
};

struct resource_info {
	enum resource_type type;
	union {
		struct video_attribute video;
		struct audio_attribute audio;
	}attr;
	char *conference_number;
	AST_LIST_HEAD_NOLOCK(, resource_owner) owner_list;
	struct ast_channel *slider;
	int inuse;
	char *id; /*Mark resource assign*/
};

struct global_profile {
	int max_rooms;
	int max_users;
	int max_video_users;
	int video_hd_bt;
	int video_sd_bt;
	int max_video_record_rooms;
	int max_video_encode_cap;
	int video_encode_frame;
	int slides_encode_frame;
	int focus_encode_frame;
};

struct room_profile {
	int max_video_users;
	int max_FHD_video_users;
	int max_share_users;
};

struct conf_resource_room {
	char name[MAX_CONF_NAME];
	struct room_profile profile;
	AST_VECTOR(, struct resource_info *) audio_v;
	AST_VECTOR(, struct resource_info *) video_v;
	int inuse_video_hd_num;
};

struct conf_resources {
	struct global_profile g_profile;
	struct room_profile r_profile;
	struct ao2_container *rooms;
};

/* Globals */
static const char config[] = "license.conf";
static struct conf_resources *g_conf_resources = NULL;

static const char *resource_type_string[RESOURCE_TYPE_END] = {
	"audio",
	"video",
	"slides"
};

static const char *video_resolution_string[VIDEO_RESOLUTION_MAX] = {
	"NONE",
	"360P",
	"720P",
	"1080P"
};

char *handle_cli_confbridge_resource_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);

static void conf_resource_write_operator_log(struct ast_channel *chan, const char *conf_number, enum conf_operation_type op_type, const char *extra_msg)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	char *caller_id = NULL;
	char *caller_name = NULL;
	char call_id[AST_CALLID_BUFFER_LENGTH] = {0};
	char user_agent[AST_USER_AGENT_LENGTH] = {0};
	char oplog_detail[AST_OPLOG_DETAIL_LENGTH] = {0};

	caller_id = S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "Anonymous");
	caller_name = S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, caller_id);
	ast_callid_strnprint(call_id, sizeof(call_id), ast_channel_callid(chan));
	ast_channel_get_device_ua(chan, user_agent, sizeof(user_agent));
	snprintf(oplog_detail, sizeof(oplog_detail), "%s%s", S_OR(extra_msg, ""), user_agent);

	conference = conf_find_bridge(conf_number);
	if (conference) {
		ast_operation_log(caller_id, caller_name, call_id, conference->feature_id, conf_operation_type_to_action(op_type), "%s", S_OR(oplog_detail, ""));
	} else {
		ast_operation_log(caller_id, caller_name, call_id, conf_number, conf_operation_type_to_action(op_type), "%s", S_OR(oplog_detail, ""));
	}
}

static struct resource_owner *create_resource_owner(struct ast_channel *chan)
{
	struct resource_owner *owner = NULL;

	owner = ast_calloc(1, sizeof(*owner));
	if (!owner) {
		return NULL;
	}

	owner->channel_name = ast_strdup(ast_channel_name(chan));
	owner->caller_number = ast_strdup(S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, NULL));
	return owner;
}

static void destroy_resource_owner(struct resource_owner *owner)
{
	if (owner) {
		if (owner->channel_name) {
			ast_free(owner->channel_name);
			owner->channel_name = NULL;
		}

		if (owner->caller_number) {
			ast_free(owner->caller_number);
			owner->caller_number = NULL;
		}

		ast_free(owner);
	}
}

static int match_resource_owner(struct resource_info *resource, const char *channel_name)
{
	struct resource_owner *owner = NULL;
	int match = 0;

	AST_LIST_TRAVERSE(&resource->owner_list, owner, entry) {
		if (!strcmp(owner->channel_name, channel_name)) {
			match = 1;
		}
	}

	return match;
}

static void reset_resource(struct resource_info *resource)
{
	struct resource_owner *cur_owner = NULL;

	resource->inuse = 0;
	if (resource->conference_number) {
		ast_free((void *)resource->conference_number);
		resource->conference_number = NULL;
	}
	if (resource->type == RESOURCE_TYPE_VIDEO) {
		memset(&resource->attr.video, 0, sizeof(struct video_attribute));
	}
	if (resource->id) {
		ast_free((void *)resource->id);
		resource->id = NULL;
	}
	if (resource->type == RESOURCE_TYPE_SLIDES && resource->slider) {
		resource->attr.video.slides_type = 0;
		ao2_ref(resource->slider, -1);
		resource->slider = NULL;
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&resource->owner_list, cur_owner, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		destroy_resource_owner(cur_owner);
	}
	AST_LIST_TRAVERSE_SAFE_END;
}

static void destroy_conf_resource(void *obj)
{
	struct conf_resources *resources = obj;

	ao2_cleanup(resources->rooms);
}

static void destroy_conf_resource_room(void *obj)
{
	struct conf_resource_room *room = obj;
	ast_log(LOG_DEBUG, "Destroying the conference resource room %s.\n", room->name);

	AST_VECTOR_RESET(&room->audio_v, RESOURCE_DELETE_CB);
	AST_VECTOR_FREE(&room->audio_v);
	AST_VECTOR_RESET(&room->video_v, RESOURCE_DELETE_CB);
	AST_VECTOR_FREE(&room->video_v);
}

static int audio_resource_init(struct conf_resource_room *room, int num)
{
	if (num < 0) {
		return -1;
	}

	AST_VECTOR_INIT(&room->audio_v, num);
	AST_VECTOR_CALLBACK_VOID(&room->audio_v, RESOURCE_INIT_CB, RESOURCE_TYPE_AUDIO);

	return 0;
}

static int video_resource_init(struct conf_resource_room *room, int num, enum resource_type type, enum video_resolution resolution)
{
	struct resource_info *resource = NULL;
	int i = 0;

	for (i = 0; i < num; i++) {
		RESOURCE_INIT_CB(resource, type);
		resource->attr.video.resolution = resolution;
		AST_VECTOR_APPEND(&room->video_v, resource);
	}

	return 0;
}

static void unlink_conf_resource_room(struct conf_resource_room *room)
{
	struct resource_info *resource = NULL;
	int i = 0;

	ao2_lock(g_conf_resources);
	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->audio_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->audio_v, i);
		if (!resource) {
			continue;
		}

		if (resource->inuse) {
			ao2_unlock(room);
			ao2_unlock(g_conf_resources);
			return;
		}
	}

	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (!resource) {
			continue;
		}

		if (resource->inuse) {
			ao2_unlock(room);
			ao2_unlock(g_conf_resources);
			return;
		}
	}
	ao2_unlock(room);

	ao2_unlink(g_conf_resources->rooms, room);
	ao2_unlock(g_conf_resources);
}

static struct conf_resource_room *build_conf_resource_room(const char *conf_number)
{
	struct conf_resource_room *room = NULL;

	ao2_lock(g_conf_resources);
	/*check room*/
	if (ao2_container_count(g_conf_resources->rooms) >= g_conf_resources->g_profile.max_rooms) {
		ast_log(LOG_WARNING, "Resource rooms has reached the max number, couldn't allocate a room for %s.\n", conf_number);
		ao2_unlock(g_conf_resources);
		return NULL;
	}

	room = ao2_alloc(sizeof(struct conf_resource_room), destroy_conf_resource_room);
	if (!room) {
		ao2_unlock(g_conf_resources);
		return NULL;
	}
	ast_copy_string(room->name, conf_number, sizeof(room->name));
	room->profile = g_conf_resources->r_profile;
	audio_resource_init(room, 0);
	video_resource_init(room, room->profile.max_video_users, RESOURCE_TYPE_VIDEO, VIDEO_RESOLUTION_360P);
	video_resource_init(room, room->profile.max_share_users, RESOURCE_TYPE_SLIDES, VIDEO_RESOLUTION_1080P);
	ao2_link(g_conf_resources->rooms, room);
	ao2_unlock(g_conf_resources);

	return room;
}

static struct conf_resource_room *find_conf_resource_room(const char *conf_number)
{
	struct conf_resource_room *room = NULL;

	if (ast_strlen_zero(conf_number)) {
		return NULL;
	}

	room = ao2_find(g_conf_resources->rooms, conf_number, OBJ_KEY);
	return room;
}

static struct resource_info *find_resource(struct conf_resource_room *room, const char *owner, enum resource_type type)
{
	int i = 0;
	struct resource_info *resource = NULL;

	ao2_lock(room);
	if (type == RESOURCE_TYPE_AUDIO) {
		for (i = 0; i < AST_VECTOR_SIZE(&room->audio_v); i++, resource = NULL) {
			resource = AST_VECTOR_GET(&room->audio_v, i);
			if (resource->inuse && resource->type == type
				&& match_resource_owner(resource, owner)) {
				break;
			}
		}
	} else {
		for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
			resource = AST_VECTOR_GET(&room->video_v, i);
			if (resource->inuse && resource->type == type
				&& match_resource_owner(resource, owner)) {
				break;
			}
		}
	}
	ao2_unlock(room);

	return resource;
}

/*
static int room_video_active_count(struct conf_resource_room *room)
{
	struct resource_info *resource = NULL;
	int i = 0;
	int count = 0;

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (!resource) {
			continue;
		}

		if (resource->inuse) {
			count++;
		}
	}
	ao2_unlock(room);
	return count;
}

static int all_rooms_video_active_count(void)
{
	struct ao2_iterator iter;
	struct conf_resource_room *room = NULL;
	int count = 0;

	ao2_lock(g_conf_resources);
	iter = ao2_iterator_init(g_conf_resources->rooms, 0);
	while ((room = ao2_iterator_next(&iter))) {
		count += room_video_active_count(room);
		ao2_ref(room, -1);
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(g_conf_resources);

	return count;
}
*/

static struct resource_info *get_audio_resource(struct ast_channel *chan, struct conf_resource_room *room)
{
	int i = 0;
	struct resource_info *resource = NULL;
	struct resource_info *match_resource = NULL;
	struct resource_owner *owner = NULL;
	const char *id = pbx_builtin_getvar_helper(chan, SFU_CALL_ID);

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->audio_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->audio_v, i);
		if (!resource) {
			continue;
		}

		if (!resource->inuse) {
			if (ast_strlen_zero(id)) {
				match_resource = resource;
				break;
			} else if (!match_resource) {
				match_resource = resource;
			}
		} else if (!ast_strlen_zero(resource->id) && !ast_strlen_zero(id) && !strcmp(resource->id, id)) {
			match_resource = resource;
			break;
		}
	}

	if (match_resource) {
		if (!match_resource->inuse) {
			match_resource->inuse = 1;
			match_resource->conference_number = ast_strdup(room->name);
			match_resource->id = ast_strdup(id);
		}

		owner = create_resource_owner(chan);
		if (owner) {
			AST_LIST_INSERT_TAIL(&match_resource->owner_list, owner, entry);
		}
	}

	ao2_unlock(room);

	return match_resource;
}

static int put_audio_resource(struct conf_resource_room *room, const char *channel_name)
{
	int res = -1;
	int i = 0;
	int found = 0;
	struct resource_info *resource = NULL;
	struct resource_owner *owner = NULL;

	if (ast_strlen_zero(channel_name)) {
		return -1;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->audio_v); i++) {
		resource = AST_VECTOR_GET(&room->audio_v, i);
		if (!resource || !resource->inuse) {
			continue;
		}

		AST_LIST_TRAVERSE_SAFE_BEGIN(&resource->owner_list, owner, entry) {
			if (!strcmp(owner->channel_name, channel_name)) {
				AST_LIST_REMOVE_CURRENT(entry);
				destroy_resource_owner(owner);
				found = 1;
				break;
			}
		}
		AST_LIST_TRAVERSE_SAFE_END;

		if (found) {
			if (AST_LIST_EMPTY(&resource->owner_list)) {
				reset_resource(resource);
			}

			res = 0;
			break;
		}
	}
	ao2_unlock(room);

	return res;
}

static struct resource_info *get_video_resource(struct ast_channel *chan, struct conf_resource_room *room, enum resource_type type, enum video_resolution resolution)
{
	int i = 0;
	struct resource_info *resource = NULL;
	struct resource_info *match_resource = NULL;
	struct resource_owner *owner = NULL;
	const char *id = pbx_builtin_getvar_helper(chan, SFU_CALL_ID);

	ao2_lock(room);

	if (chan) {
		resource = find_resource(room, ast_channel_name(chan), type);
		if (resource) {
			ast_log(LOG_DEBUG, "Chan:%s aleady get video resource\n", ast_channel_name(chan));
			if (resolution == VIDEO_RESOLUTION_360P) {
				if (resource->attr.video.resolution == VIDEO_RESOLUTION_1080P || resource->attr.video.resolution == VIDEO_RESOLUTION_720P) {
					room->inuse_video_hd_num--;
				}
				resource->attr.video.resolution = resolution;
			} else if (resolution == VIDEO_RESOLUTION_1080P || resolution == VIDEO_RESOLUTION_720P) {
				if (resource->attr.video.resolution == VIDEO_RESOLUTION_360P) {
					if (room->inuse_video_hd_num >= room->profile.max_FHD_video_users) {
						ast_log(LOG_WARNING, "There is no HD video resource for room %s.\n", room->name);
					} else {
						resource->attr.video.resolution = resolution;
						room->inuse_video_hd_num++;
					}
				} else {
					resource->attr.video.resolution = resolution;
				}
			}
			ao2_unlock(room);
			return resource;
		}
	}

	/*Check the number of HD video*/
	if (type == RESOURCE_TYPE_VIDEO && (resolution == VIDEO_RESOLUTION_1080P || resolution == VIDEO_RESOLUTION_720P)) {
		if (room->inuse_video_hd_num >= room->profile.max_FHD_video_users) {
			ast_log(LOG_WARNING, "There is no HD video resource, but will check SD resource.\n");
			resolution = VIDEO_RESOLUTION_360P;
		}
	}

	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (!resource || resource->type != type) {
			continue;
		}

		if (!resource->inuse) {
			if (ast_strlen_zero(id)) {
				match_resource = resource;
				break;
			} else if (!match_resource) {
				match_resource = resource;
			}
		} else if (!ast_strlen_zero(resource->id) && !ast_strlen_zero(id) && !strcmp(resource->id, id)) {
			match_resource = resource;
			break;
		}
	}

	if (match_resource) {
		if (!match_resource->inuse) {
			match_resource->inuse = 1;
			match_resource->conference_number = ast_strdup(room->name);
			match_resource->id = ast_strdup(id);
			match_resource->attr.video.resolution = resolution;
			if (type == RESOURCE_TYPE_VIDEO && (resolution == VIDEO_RESOLUTION_1080P || resolution == VIDEO_RESOLUTION_720P)) {
				room->inuse_video_hd_num++;
			}

			if (type == RESOURCE_TYPE_SLIDES) {
				match_resource->slider = chan;
				ao2_ref(chan, +1);
			}
		}

		owner = create_resource_owner(chan);
		if (owner) {
			AST_LIST_INSERT_TAIL(&match_resource->owner_list, owner, entry);
		}
	}

	ao2_unlock(room);

	return match_resource;
}

static int put_video_resource(struct conf_resource_room *room, const char *channel_name, enum resource_type type, enum video_resolution resolution)
{
	int res = -1;
	int i = 0;
	int found = 0;
	struct resource_info *resource = NULL;
	struct resource_owner *owner = NULL;

	if (ast_strlen_zero(channel_name)) {
		return -1;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (!resource || !resource->inuse
			|| resource->type != type) {
			continue;
		}

		AST_LIST_TRAVERSE_SAFE_BEGIN(&resource->owner_list, owner, entry) {
			if (!strcmp(owner->channel_name, channel_name)) {
				AST_LIST_REMOVE_CURRENT(entry);
				destroy_resource_owner(owner);
				found = 1;
				break;
			}
		}
		AST_LIST_TRAVERSE_SAFE_END;

		if (found) {
			if (AST_LIST_EMPTY(&resource->owner_list)) {
				if (resource->type == RESOURCE_TYPE_VIDEO && (resource->attr.video.resolution == VIDEO_RESOLUTION_1080P || resource->attr.video.resolution == VIDEO_RESOLUTION_720P)) {
					room->inuse_video_hd_num--;
				}
				reset_resource(resource);
			}

			res = 0;
			break;
		}
	}
	ao2_unlock(room);

	return res;
}

static int conf_alloc_resource(const char *conf_number, struct ast_channel *chan, enum resource_type type, enum video_resolution resolution)
{
	int res = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info  *resource = NULL;

	if (!chan) {
		return -1;
	}

	ao2_lock(g_conf_resources);
	room = find_conf_resource_room(conf_number);
	if (!room) {
		room = build_conf_resource_room(conf_number);
		if (!room) {
			ao2_unlock(g_conf_resources);
			return -3;
		}
	}

	switch (type) {
		case RESOURCE_TYPE_AUDIO:
			if (!(resource = get_audio_resource(chan, room))) {
				ast_log(LOG_DEBUG, "Channel %s couldn't get resource type '%s'",
					ast_channel_name(chan), resource_type_string[type]);
				res = -1;
				break;
			}
			break;

		case RESOURCE_TYPE_VIDEO:
			/*check max video users*/
			/*
			if (all_rooms_video_active_count() >= g_conf_resources->g_profile.max_video_users) {
				ast_log(LOG_DEBUG, "The number of video users has reached the max number, so couldn't allocate video resource to '%s'.\n",
					ast_channel_name(chan));
				res = -1;
				break;
			}
			*/
		case RESOURCE_TYPE_SLIDES:
			if (ast_gloabl_flow_control == 1) {
				ast_log(LOG_WARNING, "The CPU flow control have been enabled, the video resource should been limit !\n");
				res = -2;
				unlink_conf_resource_room(room);
				break;
			}
			else if (!(resource = get_video_resource(chan, room, type, resolution))) {
				ast_log(LOG_DEBUG, "Channel %s couldn't get resource type '%s'",
					ast_channel_name(chan), resource_type_string[type]);
				res = -1;
				break;
			}
			ast_channel_set_is_videocall(chan, 1);

			res = resource->attr.video.resolution;
			break;

		default:
			ast_log(LOG_WARNING, "Unknown resource type '%d'", type);
			res  = -1;
			break;
	}
	ao2_unlock(g_conf_resources);

	ao2_cleanup(room);
	return res;
}

static int conf_release_resource(const char *conf_number, struct ast_channel *chan, enum resource_type type)
{
	int res = 0;
	enum video_resolution resolution = VIDEO_RESOLUTION_360P;
	struct conf_resource_room *room = NULL;

	if (!chan) {
		return -1;
	}

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return -1;
	}

	switch (type) {
		case RESOURCE_TYPE_AUDIO:
			res = put_audio_resource(room, ast_channel_name(chan));
			break;

		case RESOURCE_TYPE_VIDEO:
		case RESOURCE_TYPE_SLIDES:
			res = put_video_resource(room, ast_channel_name(chan), type, resolution);
			break;

		default:
			ast_log(LOG_WARNING, "Unknown resource type '%d'", type);
			res  = -1;
			break;
	}

	if (!res) {
		unlink_conf_resource_room(room);
	}
	ao2_cleanup(room);

	return res;
}

int ast_conf_get_audio_resource(const char *conf_number, struct ast_channel *chan)
{
	return conf_alloc_resource(conf_number, chan, RESOURCE_TYPE_AUDIO, VIDEO_RESOLUTION_NONE);
}

int ast_conf_put_audio_resource(const char *conf_number, struct ast_channel *chan)
{
	return conf_release_resource(conf_number, chan, RESOURCE_TYPE_AUDIO);
}

int ast_conf_get_main_video_resource(const char *conf_number, struct ast_channel *chan, int resolution)
{
	return conf_alloc_resource(conf_number, chan, RESOURCE_TYPE_VIDEO, resolution);
}

int ast_conf_put_main_video_resource(const char *conf_number, struct ast_channel *chan)
{
	return conf_release_resource(conf_number, chan, RESOURCE_TYPE_VIDEO);
}

int ast_conf_get_slides_video_resource(const char *conf_number, struct ast_channel *chan, int share_audio)
{
	int ret = VIDEO_RESOLUTION_NONE;

	ret = conf_alloc_resource(conf_number, chan, RESOURCE_TYPE_SLIDES, VIDEO_RESOLUTION_1080P);
	if (ret == VIDEO_RESOLUTION_1080P) {
		conf_cei_add_sharing_user(ast_channel_name(chan), conf_number, share_audio);
	}

	conf_resource_write_operator_log(chan, conf_number, CONF_OP_OPEN_SHARE, NULL);

	return ret;
}

int ast_conf_put_slides_video_resource(const char *conf_number, struct ast_channel *chan)
{
	int ret = -1;

	ret = conf_release_resource(conf_number, chan, RESOURCE_TYPE_SLIDES);
	if (!ret) {
		conf_cei_delete_sharing_user(ast_channel_name(chan), conf_number);
	}

	conf_resource_write_operator_log(chan, conf_number, CONF_OP_CLOSE_SHARE, NULL);

	return ret;
}

/*Free the object after use*/
char *ast_conf_get_slider(const char *conf_number)
{
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	char *slider = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return NULL;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_SLIDES) {
			slider = ast_strdup(AST_LIST_FIRST(&resource->owner_list)->channel_name);
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);

	return slider;
}

struct ast_channel *ast_conf_get_slider_channel(const char *conf_number)
{
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	//struct resource_owner * owner = NULL;
	struct ast_channel *channel = NULL;
	//char *channel_name = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return NULL;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_SLIDES) {
			/*owner = AST_LIST_FIRST(&resource->owner_list);
			if (owner) {
				channel_name = ast_strdupa(owner->channel_name);
				break;
			}*/
			channel = resource->slider;
			ao2_ref(channel, +1);
		}
	}
	ao2_unlock(room);

	/*if (!ast_strlen_zero(channel_name)) {
		channel = ast_channel_get_by_name_prefix(channel_name, strlen(channel_name));
	}*/

	ao2_cleanup(room);
	return channel;
}

int ast_conf_set_slides_type(const char *conf_number, int type)
{
	int res = -1;
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return -1;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_SLIDES) {
			resource->attr.video.slides_type = type;
			res = 0;
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);
	return res;
}

int ast_conf_get_slides_type(const char *conf_number)
{
	int type = -1;
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return -1;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_SLIDES) {
			type = resource->attr.video.slides_type;
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);
	return type;
}

int ast_conf_set_slider_name(const char *conf_number, char *slider_name)
{
	int res = -1;
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return -1;
	}

	if (ast_strlen_zero(slider_name)) {
		return -1;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_SLIDES) {
			strncpy(resource->attr.video.slider_name, slider_name, sizeof(resource->attr.video.slider_name) - 1);
			res = 0;
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);
	return res;
}

char *ast_conf_get_slider_name(const char *conf_number)
{
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	char *slider_name = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return NULL;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_SLIDES) {
			slider_name = ast_strdup(resource->attr.video.slider_name);
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);
	return slider_name;
}

int ast_conf_switch_main_video_resolution(struct ast_channel *chan, const char *conf_number, int new_resolution)
{
	int res = -1;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return -1;
	}

	ao2_lock(room);
	resource = find_resource(room, ast_channel_name(chan), RESOURCE_TYPE_VIDEO);
	if (!resource) {
		ao2_unlock(room);
		ao2_cleanup(room);
		return -1;
	}

	if (resource->attr.video.resolution == new_resolution) {
		ast_log(LOG_DEBUG, "The resolution is the same with the current resolution, no need to change.\n");
		ao2_unlock(room);
		ao2_cleanup(room);
		res = 0;
		goto SWITCH_SUCCESS;
	}

	if (new_resolution == VIDEO_RESOLUTION_360P) {
		if (resource->attr.video.resolution == VIDEO_RESOLUTION_1080P || resource->attr.video.resolution == VIDEO_RESOLUTION_720P) {
			room->inuse_video_hd_num--;
		}
		resource->attr.video.resolution = new_resolution;
		res = 0;
	} else if (new_resolution == VIDEO_RESOLUTION_1080P || new_resolution == VIDEO_RESOLUTION_720P) {
		if (resource->attr.video.resolution == VIDEO_RESOLUTION_360P) {
			if (room->inuse_video_hd_num >= room->profile.max_FHD_video_users) {
				ast_log(LOG_WARNING, "There is no HD video resource for room %s.\n", room->name);
			} else {
				resource->attr.video.resolution = new_resolution;
				room->inuse_video_hd_num++;
				res = 0;
			}
		} else {
			resource->attr.video.resolution = new_resolution;
			res = 0;
		}
	}

	ao2_unlock(room);
	ao2_cleanup(room);

SWITCH_SUCCESS:

	if (!res) {
		conf_cei_video_resolution_update(ast_channel_name(chan), conf_number, new_resolution);
	}

	conf_resource_write_operator_log(chan, conf_number, CONF_OP_SWITCH_RESOLUTION, video_resolution_type_to_name(new_resolution));

	return res;
}

void ast_conf_cei_video_resolution_update(struct ast_channel *chan, const char *conf_number, int new_resolution)
{
	conf_cei_video_resolution_update(ast_channel_name(chan), conf_number, new_resolution);
}

static void conf_camera_statistics(const char *conf_number, int op_type)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	conference = conf_find_bridge(conf_number);
	if (!conference) {
		ast_log(LOG_DEBUG, "conf_camera_statistics conference not found!");
		return;
	}
	ao2_lock(conference);
	if (op_type == CONF_OP_OPEN_CAMERA){
		conference->num_camera += 1;
		if (conference->num_camera > conference->peak_camera) {
			conference->peak_camera = conference->num_camera;
		}
	}else if (op_type == CONF_OP_CLOSE_CAMERA) {
		if (conference->num_camera > 0) {
			conference->num_camera -= 1;
		}
	}
	ao2_unlock(conference);

}
void ast_conf_indicate(struct ast_channel *chan, const char *conf_number, int condition, void *data)
{
	switch (condition) {
		case CONF_CONTROL_OPEN_CAMERA:
			confbridge_user_media_update(ast_channel_name(chan), conf_number, 1);
			conf_cei_media_update(ast_channel_name(chan), conf_number, 1);
			conf_resource_write_operator_log(chan, conf_number, CONF_OP_OPEN_CAMERA, NULL);
			conf_camera_statistics(conf_number, CONF_OP_OPEN_CAMERA);
			break;
		case CONF_CONTROL_CLOSE_CAMERA:
			confbridge_user_media_update(ast_channel_name(chan), conf_number, 0);
			conf_cei_media_update(ast_channel_name(chan), conf_number, 0);
			conf_resource_write_operator_log(chan, conf_number, CONF_OP_CLOSE_CAMERA, NULL);
			conf_camera_statistics(conf_number, CONF_OP_CLOSE_CAMERA);
			break;
		default:
			ast_log(LOG_WARNING, "Ignore the unknown condition '%d'\n", condition);
			break;
	}
}

int ast_conf_mic_mute_unmute_helper(int mute, const char *conference_name, const char *chan_name)
{
	return conf_mic_mute_unmute_helper(mute, conference_name, chan_name);
}

int ast_conf_user_hand_up_helper(int hand_up, const char *conference_name, const char *chan_name, struct ast_channel *op_chan)
{
	return conf_user_hand_up_helper(hand_up, conference_name, chan_name, op_chan);
}

int ast_conf_set_video_codec(struct ast_channel *chan, const char *conf_number, const char *codec)
{
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	int i = 0;
	int ret = -1;

	if (!chan || ast_strlen_zero(conf_number) || ast_strlen_zero(codec)) {
		return -1;
	}

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return -1;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && resource->type == RESOURCE_TYPE_VIDEO && match_resource_owner(resource, ast_channel_name(chan))) {
			ast_copy_string(resource->attr.video.codec, codec, sizeof(resource->attr.video.codec));
			ret = 0;
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);
	return ret;
}

char *ast_conf_get_video_codec(struct ast_channel *chan, const char *conf_number)
{
	char *codec = NULL;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	int i = 0;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return NULL;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) && resource->type == RESOURCE_TYPE_VIDEO && !ast_strlen_zero(resource->attr.video.codec)) {
			codec = ast_strdup(resource->attr.video.codec);
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);
	return codec;
}

char *ast_conf_get_hd_chan_name(const char *conf_number)
{
	char *name = NULL;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	struct resource_owner *owner = NULL;
	int i = 0;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return NULL;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) &&
			resource->type == RESOURCE_TYPE_VIDEO &&
			(resource->attr.video.resolution == VIDEO_RESOLUTION_720P || resource->attr.video.resolution == VIDEO_RESOLUTION_1080P)) {
			/* For now we don't alloc resource for SFU invitation, the list is only one member */
			AST_LIST_TRAVERSE(&resource->owner_list, owner, entry) {
				name = ast_strdup(owner->channel_name);
				break;
			}
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);

	return name;
}

enum video_resolution ast_conf_get_hd_resolution(const char *conf_number)
{
	enum video_resolution  resolution = VIDEO_RESOLUTION_NONE;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;
	int i = 0;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return VIDEO_RESOLUTION_NONE;
	}

	ao2_lock(room);
	for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
		resource = AST_VECTOR_GET(&room->video_v, i);
		if (resource->inuse && !strcmp(resource->conference_number, conf_number) &&
			resource->type == RESOURCE_TYPE_VIDEO &&
			(resource->attr.video.resolution == VIDEO_RESOLUTION_720P || resource->attr.video.resolution == VIDEO_RESOLUTION_1080P)) {
			resolution = resource->attr.video.resolution;
			break;
		}
	}
	ao2_unlock(room);

	ao2_cleanup(room);

	return resolution;
}


void ast_conf_remove_conference(const char *conf_number)
{
	ast_log(LOG_DEBUG, "ast_conf_remove_conference %s!",conf_number);

	struct cei_conference *host_conf = NULL;
	// Get host conference
	host_conf = get_conf_by_num(conf_number);
	if (host_conf)
		remove_virtual_conference(host_conf);

	return ;
}

enum video_resolution conf_user_get_video_quality(const char *channel_name, const char *conf_number)
{
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;
	struct conf_resource_room *room = NULL;
	struct resource_info * resource = NULL;

	room = find_conf_resource_room(conf_number);
	if (!room) {
		return VIDEO_RESOLUTION_NONE;
	}

	ao2_lock(room);
	resource = find_resource(room, channel_name, RESOURCE_TYPE_VIDEO);
	if (resource) {
		resolution = resource->attr.video.resolution;
	}
	ao2_unlock(room);
	ao2_cleanup(room);

	return resolution;
}

const char *video_resolution_type_to_name(int type)
{
	return video_resolution_string[type];
}

int video_resolution_name_to_type(const char *name)
{
	int i = 0;
	enum video_resolution type = VIDEO_RESOLUTION_NONE;

	if (ast_strlen_zero(name)) {
		return type;
	}

	for (i = 0; i < VIDEO_RESOLUTION_MAX; i++) {
		if (!strcasecmp(video_resolution_string[i], name)) {
			type = i;
			break;
		}
	}

	return type;
}

int sfu_max_capability()
{
	return g_conf_resources ? g_conf_resources->g_profile.max_users : SFU_DEFAULT_MAX_USERS;
}

int sfu_max_video_record_rooms_capability()
{
	if (!g_conf_resources) {
		return SFU_DEFAULT_MAX_VIDEO_RECORD_ROOMS;
	}

	if (g_conf_resources->g_profile.max_video_record_rooms) {
		return g_conf_resources->g_profile.max_video_record_rooms;
	}

	return SFU_DEFAULT_MAX_VIDEO_RECORD_ROOMS;
}

int sfu_max_video_encode_capability()
{
	if (!g_conf_resources) {
		return SFU_DEFAULT_MAX_VIDEO_ENCODE_CAP;
	}

	if (g_conf_resources->g_profile.max_video_encode_cap) {
		return g_conf_resources->g_profile.max_video_encode_cap;
	}

	return SFU_DEFAULT_MAX_VIDEO_ENCODE_CAP;
}

int sfu_video_encode_frame()
{
	if (!g_conf_resources) {
		return SFU_DEFAULT_VIDEO_ENCODE_FRAME;
	}

	if (g_conf_resources->g_profile.video_encode_frame) {
		return g_conf_resources->g_profile.video_encode_frame;
	}

	return SFU_DEFAULT_VIDEO_ENCODE_FRAME;
}

int sfu_slides_encode_frame()
{
	if (!g_conf_resources) {
		return SFU_DEFAULT_SLIDES_ENCODE_FRAME;
	}

	if (g_conf_resources->g_profile.slides_encode_frame) {
		return g_conf_resources->g_profile.slides_encode_frame;
	}

	return SFU_DEFAULT_SLIDES_ENCODE_FRAME;
}

int sfu_focus_encode_frame()
{
	if (!g_conf_resources) {
		return SFU_DEFAULT_FOUCS_ENCODE_FRAME;
	}

	if (g_conf_resources->g_profile.focus_encode_frame) {
		return g_conf_resources->g_profile.focus_encode_frame;
	}

	return SFU_DEFAULT_FOUCS_ENCODE_FRAME;
}

static int cmd_read(const char *cmd, char *dst, int det_len)
{
	FILE *fp = NULL;

	if (!cmd || !dst || !det_len) {
		return -1;
	}

	fp = popen(cmd, "r");
	if (fp == NULL)
	{
		ast_log(LOG_ERROR, "cmd read error");
		return -1;
	}
	while (fp != NULL
		&& (!feof(fp))
		&& (fgets(dst, det_len, fp) != NULL)) {
		ast_log(LOG_DEBUG, "cmd read: [%s].", dst);
	}

	pclose(fp);
	fp = NULL;

	return 0;

}

static void get_ipv6_address(char *name, struct ast_json *json)
{
	FILE *f = NULL;
	char addr6[40] = {0};
	char devname[21] = {0};
	int plen, scope, dad_status, if_idx;
	char addr6p[8][5] = {0};

	if (!name || !json) {
		return;
	}

	f = fopen("/proc/net/if_inet6", "r");
	if (f == NULL) {
		return;
	}

	while (fscanf
		(f, "%4s%4s%4s%4s%4s%4s%4s%4s %08x %02x %02x %02x %20s\n",
			addr6p[0], addr6p[1], addr6p[2], addr6p[3], addr6p[4],
			addr6p[5], addr6p[6], addr6p[7], &if_idx, &plen, &scope,
			&dad_status, devname) != EOF
	) {
		if (!strcmp(devname, name)) {
			sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
					addr6p[0], addr6p[1], addr6p[2], addr6p[3],
					addr6p[4], addr6p[5], addr6p[6], addr6p[7]);

			switch (scope & 0xf0) {
			case 0:
				ast_json_object_set(json,"ipv6",ast_json_string_create(addr6));
				ast_log(LOG_DEBUG, "getIPV6Addr, get ipv6 :%s", addr6);
				break;
			case 0x20:
				ast_json_object_set(json,"ipv6_link",ast_json_string_create(addr6));
				ast_log(LOG_DEBUG, "getIPV6Addr, get ipv6 link :%s", addr6);
				break;
			default:
				break;
			}
		}
	}
	fclose(f);
}

static void get_ip_address(char *ipbuf, char *device, int bufsize)
{
	int fd;
	struct  sockaddr_in *sin;
	struct  ifreq ifr;

	memset(ipbuf, 0, bufsize);
	if ((fd = socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0)) < 0) {
		ast_log(LOG_DEBUG, "Can't talk to kernel! (%d)\n", errno);
		return;
	}
	strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name) - 1);
	if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0)  {
		ast_log(LOG_DEBUG, "Can't get status for %s. (%d)\n", device, errno);
		close(fd);
		return;
	}
	if ((ifr.ifr_flags & IFF_UP) == 0)  {
		ast_log(LOG_DEBUG, "Interface %s not active.\n", device);
		close(fd);
		return;
	}
	if (ioctl(fd, SIOCGIFADDR, &ifr) != 0) {
		ast_log(LOG_DEBUG, "Can't get IP address for %s. (%d)\n", device, errno);
		close(fd);
		return;
	}
	close(fd);
	sin = (struct sockaddr_in *)&ifr.ifr_addr;
	strncpy(ipbuf, ast_inet_ntoa(sin->sin_addr), bufsize);
	ast_log(LOG_DEBUG, "get IP address %s \n", ipbuf);

}

static int get_net_config(char *name, struct ast_json *json)
{
	int res = -1;
	char cmd[512] = {0};
	char tmp_buff[256] = {0};
	struct ast_json *net_json = json;

	if (name == NULL || name[0] == '\0' || net_json == NULL) {
		return res;
	}

	ast_log(LOG_DEBUG, "get net config name %s", name);
	/* get IP */
	get_ip_address(tmp_buff, name, sizeof(tmp_buff));
	if (tmp_buff[0] != '\0') {
		ast_json_object_set(net_json, "ip", ast_json_string_create(tmp_buff));
	} else if (strstr(name, "ppp") != NULL) {
		snprintf(cmd, sizeof(cmd), "ifconfig %s | grep addr | awk -F ':' '{print $2}' | awk -F ' ' '{print $1}'", name);
		cmd_read(cmd, tmp_buff, sizeof(tmp_buff) - 1);
		ast_json_object_set(net_json, "ip", ast_json_string_create(tmp_buff));
	} else {
		ast_log(LOG_DEBUG, "Not handled %s", name);
		return res;
	}

	/* get ipv6 */
	get_ipv6_address(name, net_json);

	res = 0;

	return res;
}

static int get_device_info(conf_net_info *info)
{
	int nat = 0;
	int haon = 0;
	const char *ptr = NULL;
	const char *hafloatip = NULL;
	char pvalue[MAX_PVALUE_LENGTH] = {0};


	if (info == NULL) {
		ast_log(LOG_DEBUG, "NetInfo pointer is NULL");
		return -1;
	}

	if (info->mode == -1) {
		memset(info, 0, sizeof(conf_net_info));
		info->mode = 1;

		ptr = ast_safe_nvram_get("haon", pvalue, sizeof(pvalue));
		if (ptr != NULL) {
			haon = atoi(ptr);
		}
		info->haon = haon;

		if (haon) {
			hafloatip = ast_safe_nvram_get("hafloatip", pvalue, sizeof(pvalue));
			if (hafloatip != NULL) {
				strncpy(info->ip_hafloat, hafloatip, sizeof(info->ip_hafloat));
			}

			ptr = ast_safe_nvram_get(":haRole", pvalue, sizeof(pvalue));
			if (ptr != NULL) {
				info->harole = atoi(ptr);
			}
		}

		ast_log(LOG_DEBUG, "haon:%d", info->haon);
		ast_log(LOG_DEBUG, "harole:%d", info->harole);
		ast_log(LOG_DEBUG, "ip_hafloat:%s", info->ip_hafloat);

		ptr = ast_safe_nvram_get("allow_nat", pvalue, sizeof(pvalue));
		if (ptr != NULL) {
			nat = atoi(ptr);
			ast_log(LOG_DEBUG, "allow nat is (%s)", ptr);
		}

		/* Network mode */
		ptr = ast_safe_nvram_get("231", pvalue, sizeof(pvalue));
		if (ptr != NULL) {
			info->mode = atoi(ptr);
			ast_log(LOG_DEBUG, "231 is (%s)", ptr);
		}

		ptr = ast_safe_nvram_get(DEFAULT_IF, pvalue, sizeof(pvalue));
		if (ptr != NULL) {
			strncpy(info->default_if, ptr, sizeof(info->default_if));
			ast_log(LOG_DEBUG, "default_if is (%s)", ptr);
		}

		if (info->mode == 1) {
			/*
			*  Switch Mode
			*/
			ptr = ast_safe_nvram_get(NET_LAN, pvalue, sizeof(pvalue));
			if (ptr != NULL) {
				strncpy(info->lan, ptr, sizeof(info->lan));
				ast_log(LOG_DEBUG, "lan is (%s)", ptr);
			}
		}
		else if (info->mode == 2) {
			/*
			*  Dual Mode
			*/
			ptr = ast_safe_nvram_get(NET_LAN1, pvalue, sizeof(pvalue));
			if (ptr != NULL) {
				strncpy(info->lan1, ptr, sizeof(info->lan1));
				ast_log(LOG_DEBUG, "lan1 is (%s)", ptr);
			}

			ptr = ast_safe_nvram_get(NET_LAN2, pvalue, sizeof(pvalue));
			if (ptr != NULL) {
				strncpy(info->lan2, ptr, sizeof(info->lan2));
				ast_log(LOG_DEBUG, "lan2 is (%s)", ptr);
			}
		}
		else if (info->mode == 0) {
			/*
			*  Router mode
			*/
			if (nat == 1) {
				ptr = ast_safe_nvram_get(NET_LAN, pvalue, sizeof(pvalue));
				if (ptr != NULL) {
					strncpy(info->lan, ptr, sizeof(info->lan));
					ast_log(LOG_DEBUG, "lan is (%s)", ptr);
				}

				ptr = ast_safe_nvram_get(NET_WAN, pvalue, sizeof(pvalue));
				if (ptr != NULL) {
					strncpy(info->wan, ptr, sizeof(info->wan));
					ast_log(LOG_DEBUG, "wan is (%s)", ptr);
				}
			}
			else {
				ptr = ast_safe_nvram_get(NET_LAN, pvalue, sizeof(pvalue));
				if (ptr != NULL) {
					strncpy(info->lan, ptr, sizeof(info->lan));
					ast_log(LOG_DEBUG, "lan is (%s)", ptr);
				}
			}
		}
		else {
			ast_log(LOG_WARNING, "Unrecorgnized net mode:%d", info->mode);
		}
	}

	return 0;
}

static void get_network_info(conf_net_info *info, struct ast_json *json)
{
	int ret = -1;
	char netmode[16] = {0};
	struct ast_json *values = json;

	if (!values) {
		return;
	}

	if (info != NULL) {
		if (!ast_strlen_zero(info->lan)) {
			strncpy(netmode, "lan", sizeof(netmode) - 1);
			struct ast_json *net_json = ast_json_object_create();
			if (net_json != NULL) {
				ret = get_net_config(info->lan, net_json);/*get lan config*/
				if (ret == 0) {
					ast_json_object_set(values, netmode, net_json);
				} else {
					ast_json_unref(net_json);
				}
			}
		}

		if (!ast_strlen_zero(info->wan)) {
			strncpy(netmode, "wan", sizeof(netmode) - 1);
			struct ast_json *net_json = ast_json_object_create();
			if (net_json != NULL) {
				ret = get_net_config(info->wan, net_json);/*get wan config*/
				if (ret == 0) {
					ast_json_object_set(values, netmode, net_json);
				} else {
					ast_json_unref(net_json);
				}
			}
		}

		if (!ast_strlen_zero(info->lan1)) {
			strncpy(netmode, "lan1", sizeof(netmode) - 1);
			struct ast_json *net_json = ast_json_object_create();
			if (net_json != NULL) {
				ret = get_net_config(info->lan1, net_json);/*get lan1 config*/
				if (ret == 0) {
					ast_json_object_set(values, netmode, net_json);
				} else {
					ast_json_unref(net_json);
				}
			}
		}

		if (!ast_strlen_zero(info->lan2)) {
			strncpy(netmode, "lan2", sizeof(netmode) - 1);
			struct ast_json *net_json = ast_json_object_create();
			if (net_json != NULL) {
				ret = get_net_config(info->lan2, net_json);/*get lan2 config*/
				if (ret == 0) {
					ast_json_object_set(values, netmode, net_json);
				} else {
					ast_json_unref(net_json);
				}
			}
		}
	}
}

void ast_conf_net_init(conf_net_info *info)
{
	if (info) {
		info->haon = -1;
		info->harole = -1;
		info->mode = -1;
	}
}

void ast_conf_get_net_info(struct ast_json *net_json, conf_net_info *info)
{
	if(net_json && info) {
		get_device_info(info);
		get_network_info(info, net_json);
	}
}


char *handle_cli_confbridge_resource_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int i = 0;
	struct conf_resource_room *room = NULL;
	struct resource_info *resource = NULL;
	struct resource_owner *owner = NULL;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge resource list";
		e->usage =
			"Usage: confbridge resource list [<name>]\n"
			"       Lists all currently resources in SFU.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 3) {
		struct ao2_iterator iter;
		ao2_lock(g_conf_resources);
		ast_cli(a->fd, "Conference   Capacity\n");
		ast_cli(a->fd, "============ =========\n");
		iter = ao2_iterator_init(g_conf_resources->rooms, 0);
		while ((room = ao2_iterator_next(&iter))) {
			ast_cli(a->fd, "%-12s %-9d\n",
				room->name, room->profile.max_video_users);
			ao2_ref(room, -1);
		}
		ao2_iterator_destroy(&iter);
		ao2_unlock(g_conf_resources);
		return CLI_SUCCESS;
	}

	if (a->argc == 4) {
		room = find_conf_resource_room(a->argv[3]);
		if (!room) {
			ast_cli(a->fd, "No conference named '%s' found!\n", a->argv[3]);
			return CLI_SUCCESS;
		}

		ao2_lock(room);
		ast_cli(a->fd, "Caller      Type     Owner                Codec        Resolution   Conference   id\n");
		ast_cli(a->fd, "==========  ======== ==================== ============ ============ ============ ============\n");

		for (i = 0; i < AST_VECTOR_SIZE(&room->audio_v); i++, resource = NULL) {
			resource = AST_VECTOR_GET(&room->audio_v, i);
			if (resource->inuse) {
				AST_LIST_TRAVERSE(&resource->owner_list, owner, entry) {
					ast_cli(a->fd, "%-10s %-8s %-20s %-12s %-12s %-12s %-12s\n",
						owner->caller_number, resource_type_string[resource->type], owner->channel_name, "N/A",
						"N/A", resource->conference_number, S_OR(resource->id, ""));
				}
			}
		}

		for (i = 0; i < AST_VECTOR_SIZE(&room->video_v); i++, resource = NULL) {
			resource = AST_VECTOR_GET(&room->video_v, i);
			if (resource->inuse) {
				AST_LIST_TRAVERSE(&resource->owner_list, owner, entry) {
					ast_cli(a->fd, "%-10s %-8s %-20s %-12s %-12s %-12s %-12s\n",
						owner->caller_number, resource_type_string[resource->type], owner->channel_name, resource->attr.video.codec,
						video_resolution_string[resource->attr.video.resolution], resource->conference_number, S_OR(resource->id, ""));
				}
			}
		}
		ao2_unlock(room);
		ao2_cleanup(room);
		return CLI_SUCCESS;
	}

	return CLI_SHOWUSAGE;
}

int unload_conf_resource(void)
{
	ao2_cleanup(g_conf_resources);
	ast_bfcp_unregister_conf_funcs();
	ast_db_del("SFU", "MAX_SIZE");

	return 0;
}

static int conf_resource_room_hash_cb(const void *obj, const int flags)
{
	const struct conf_resource_room *room = obj;
	const char *name = obj;
	int hash;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
		default:
		case OBJ_POINTER:
			name = room->name;
			/* Fall through */
		case OBJ_KEY:
			hash = ast_str_case_hash(name);
			break;
		case OBJ_PARTIAL_KEY:
			/* Should never happen in hash callback. */
			ast_assert(0);
			hash = 0;
			break;
	}
	return hash;
}

static int conf_resource_room_cmp_cb(void *obj, void *arg, int flags)
{
	const struct conf_resource_room *left = obj;
	const struct conf_resource_room *right = arg;
	const char *right_name = arg;
	int cmp;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
		default:
		case OBJ_POINTER:
			right_name = right->name;
			/* Fall through */
		case OBJ_KEY:
			cmp = strcasecmp(left->name, right_name);
			break;
		case OBJ_PARTIAL_KEY:
			cmp = strncasecmp(left->name, right_name, strlen(right_name));
			break;
	}
	return cmp ? 0 : CMP_MATCH;
}

static int conf_resource_set_defaults(void)
{
	/*global*/
	g_conf_resources->g_profile.max_rooms = SFU_DEFAULT_MAX_ROOMS;
	g_conf_resources->g_profile.max_users = SFU_DEFAULT_MAX_USERS;
	g_conf_resources->g_profile.max_video_users = SFU_DEFAULT_SD_NUM + SFU_DEFAULT_HD_NUM;
	g_conf_resources->g_profile.video_hd_bt = SFU_DEFAULT_HD_BANDWIDTH;
	g_conf_resources->g_profile.video_sd_bt = SFU_DEFAULT_SD_BANDWIDTH;
	g_conf_resources->g_profile.max_video_record_rooms = SFU_DEFAULT_MAX_VIDEO_RECORD_ROOMS;
	g_conf_resources->g_profile.max_video_encode_cap = SFU_DEFAULT_MAX_VIDEO_ENCODE_CAP;
	g_conf_resources->g_profile.video_encode_frame = SFU_DEFAULT_VIDEO_ENCODE_FRAME;
	g_conf_resources->g_profile.slides_encode_frame = SFU_DEFAULT_SLIDES_ENCODE_FRAME;
	g_conf_resources->g_profile.focus_encode_frame = SFU_DEFAULT_FOUCS_ENCODE_FRAME;

	/*room*/
	g_conf_resources->r_profile.max_video_users = SFU_DEFAULT_SD_NUM + SFU_DEFAULT_HD_NUM;
	g_conf_resources->r_profile.max_FHD_video_users = SFU_DEFAULT_HD_NUM;
	g_conf_resources->r_profile.max_share_users = SFU_DEFAULT_SHARE_NUM;

	return 0;
}

static int conf_resource_load_config(int reload)
{
	struct ast_config *cfg = NULL;
	const char *category = NULL;
	struct ast_variable *var = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	cfg = ast_config_load(config, config_flags);
	if (cfg == CONFIG_STATUS_FILEMISSING || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "Couldn't find license config file %s.\n", config);
		conf_resource_set_defaults();
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_log(LOG_DEBUG, "The license config file %s isn't changed.\n", config);
		return 0;
	}

	while ((category = ast_category_browse(cfg, category))) {
		if (!strcasecmp(category, "global")) {
			for (var = ast_variable_browse(cfg, category); var; var = var->next) {
				if (!strcasecmp(var->name, "max_rooms")) {
					g_conf_resources->g_profile.max_rooms = atoi(var->value);
				} else if (!strcasecmp(var->name, "max_users")) {
					g_conf_resources->g_profile.max_users = atoi(var->value);
				} else if (!strcasecmp(var->name, "max_video_users")) {
					g_conf_resources->g_profile.max_video_users = atoi(var->value);
				} else if (!strcasecmp(var->name, "fhd_bandwidth")) {
					g_conf_resources->g_profile.video_hd_bt = atoi(var->value);
				} else if (!strcasecmp(var->name, "sd_bandwidth")) {
					g_conf_resources->g_profile.video_sd_bt = atoi(var->value);
				} else if (!strcasecmp(var->name, "max_video_record_rooms")) {
					g_conf_resources->g_profile.max_video_record_rooms = atoi(var->value);
				} else if (!strcasecmp(var->name, "max_video_encode_cap")) {
					g_conf_resources->g_profile.max_video_encode_cap = atoi(var->value);
				} else if (!strcasecmp(var->name, "video_encode_frame")) {
					g_conf_resources->g_profile.video_encode_frame = atoi(var->value);
				} else if (!strcasecmp(var->name, "slides_encode_frame")) {
					g_conf_resources->g_profile.slides_encode_frame = atoi(var->value);
				} else if (!strcasecmp(var->name, "focus_encode_frame")) {
					g_conf_resources->g_profile.focus_encode_frame = atoi(var->value);
				}
			}
		} else if (!strcasecmp(category, "room")) {
			for (var = ast_variable_browse(cfg, category); var; var = var->next) {
				if (!strcasecmp(var->name, "max_video_users")) {
					g_conf_resources->r_profile.max_video_users = atoi(var->value);
				} else if (!strcasecmp(var->name, "max_fhd_video_users")) {
					g_conf_resources->r_profile.max_FHD_video_users = atoi(var->value);
				} else if (!strcasecmp(var->name, "max_share_users")) {
					g_conf_resources->r_profile.max_share_users = atoi(var->value);
				}
			}
		}
	}

	ast_config_destroy(cfg);

	return 0;
}

int load_conf_resource(void)
{
	g_conf_resources = ao2_alloc(sizeof(*g_conf_resources), destroy_conf_resource);
	if (!g_conf_resources) {
		ast_log(LOG_ERROR, "Couldn't create conf resource manager instance.");
		return 0;
	}
	g_conf_resources->rooms = ao2_container_alloc(1, conf_resource_room_hash_cb, conf_resource_room_cmp_cb);
	if (!g_conf_resources->rooms) {
		ast_log(LOG_ERROR, "Couldn't create conf resource rooms.");
		ao2_cleanup(g_conf_resources);
		return 0;
	}

	conf_resource_load_config(0);

	/*register method to bcfp*/
	struct bfcp_conf_resource_manager_funcs bfcp_conf_funcs = {
		.conf_get_slides_video_resource = ast_conf_get_slides_video_resource,
		.conf_put_slides_video_resource = ast_conf_put_slides_video_resource,
		.conf_get_slider_channel = ast_conf_get_slider_channel,
		.conf_get_sfu_moderator = ast_conf_get_sfu_moderator,
		.conf_set_slider_name = ast_conf_set_slider_name
	};
	ast_bfcp_register_conf_funcs(&bfcp_conf_funcs);

	char max_size[8] = {0};
	snprintf(max_size, sizeof(max_size), "%d", g_conf_resources->g_profile.max_users);
	ast_db_put("SFU", "MAX_SIZE", max_size);

	//modify by zyxia for bug212456,2022.1.4
	ast_safe_system2("LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/reload_conf_members_status.py");
	return 0;
}

int reload_conf_resource(void)
{
	ao2_lock(g_conf_resources);
#if 0
	if (ao2_container_count(g_conf_resources->rooms) > 0) {
		ast_log(LOG_ERROR, "Resource rooms is using, couldn't reload.\n");
		ao2_unlock(g_conf_resources);
		return -1;
	}
#endif
	conf_resource_load_config(1);

	ao2_unlock(g_conf_resources);
	return 0;
}
#endif
