/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2012, Terry Wilson
 *
 * Terry Wilson <twilson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 * Please follow coding guidelines
 * http://svn.digium.com/view/asterisk/trunk/doc/CODING-GUIDELINES
 */

/*! \file
 *
 * \brief Confbridge state handling for the INACTIVE state
 *
 * \author\verbatim Terry Wilson <twilson@digium.com> \endverbatim
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "include/confbridge.h"
#include "include/conf_state.h"

static void join_unmarked(struct confbridge_user *user);
static void join_marked(struct confbridge_user *user);
static void leave_waitmarked(struct confbridge_user *user);
#ifdef GRANDSTREAM_NETWORKS
static void state_to_marked(struct confbridge_user *user);
#endif

struct confbridge_state STATE_INACTIVE = {
	.name = "INACTIVE",
	.join_unmarked = join_unmarked,
	.join_waitmarked = conf_default_join_waitmarked,
	.join_marked = join_marked,
	.leave_waitmarked = leave_waitmarked,
#ifdef GRANDSTREAM_NETWORKS
	.state_to_marked = state_to_marked,
#endif
};
struct confbridge_state *CONF_STATE_INACTIVE = &STATE_INACTIVE;

static void join_unmarked(struct confbridge_user *user)
{
	conf_add_user_active(user->conference, user);
#ifndef GRANDSTREAM_NETWORKS
	conf_add_post_join_action(user, conf_handle_only_person);
#endif

	conf_change_state(user, CONF_STATE_SINGLE);
}

static void join_marked(struct confbridge_user *user)
{
	conf_add_user_marked(user->conference, user);
	conf_update_user_mute(user);

	conf_change_state(user, CONF_STATE_MULTI_MARKED);
}

static void leave_waitmarked(struct confbridge_user *user)
{
	conf_default_leave_waitmarked(user);
	if (user->conference->waitingusers == 0) {
		conf_change_state(user, CONF_STATE_EMPTY);
	}
}

#ifdef GRANDSTREAM_NETWORKS
static void state_to_marked(struct confbridge_user *user)
{
	conf_remove_user_waiting(user->conference, user);
	conf_add_user_active(user->conference, user);
	if (user->conference->waitingusers == 0) {
		conf_state_update(user, CONF_STATE_SINGLE_MARKED);
	} else {
		int waitingusers = user->conference->waitingusers;

		if (user->playing_moh) {
			conf_moh_stop(user);
		}
		conf_update_user_mute(user);
		conf_change_state(user, CONF_STATE_MULTI_MARKED);

		/* Play the audio file stating that the conference is beginning */
		if (ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED)
			&& !ast_test_flag(&user->u_profile, USER_OPT_QUIET)
			&& waitingusers) {
			play_sound_file(user->conference,conf_get_sound(CONF_SOUND_BEGIN, user->conference->b_profile.sounds));
		}
	}
}
#endif
