/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2011, Digium, Inc.
 *
 * David Vossel <dvossel@digium.com>
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */


#ifndef _CONFBRIDGE_H
#define _CONFBRIDGE_H

#include "asterisk.h"
#include "asterisk/app.h"
#include "asterisk/logger.h"
#include "asterisk/linkedlists.h"
#include "asterisk/channel.h"
#include "asterisk/bridge.h"
#include "asterisk/bridge_features.h"
#include "conf_state.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/core_local.h"

#define MAX_DEV_NAME 16
#define MAX_RELATED_EXT 130
#define MAX_RECORD_TYPE 8
#define MAX_RECORD_LINK 32
#define MAX_CONF_LAYOUT 16

#define DEFAULT_CONF_LAYOUT "normal"
#endif

/*! Maximum length of a conference bridge name */
#define MAX_CONF_NAME AST_MAX_EXTENSION
/*! Maximum length of a conference pin */
#define MAX_PIN     80
/*! Maximum length of bridge/user/menu profile names */
#define MAX_PROFILE_NAME 128

#define DEFAULT_USER_PROFILE "default_user"
#define DEFAULT_BRIDGE_PROFILE "default_bridge"
#define DEFAULT_MENU_PROFILE "default_menu"

/*! Default minimum average magnitude threshold to determine talking by the DSP. */
#define DEFAULT_TALKING_THRESHOLD 160

/*! Default time in ms of silence necessary to declare talking stopped by the bridge. */
#define DEFAULT_SILENCE_THRESHOLD 2500

enum user_profile_flags {
	USER_OPT_ADMIN =        (1 << 0), /*!< Set if the caller is an administrator */
	USER_OPT_NOONLYPERSON = (1 << 1), /*!< Set if the "you are currently the only person in this conference" sound file should not be played */
	USER_OPT_MARKEDUSER =   (1 << 2), /*!< Set if the caller is a marked user */
	USER_OPT_STARTMUTED =   (1 << 3), /*!< Set if the caller should be initially set muted */
	USER_OPT_MUSICONHOLD =  (1 << 4), /*!< Set if music on hold should be played if nobody else is in the conference bridge */
	USER_OPT_QUIET =        (1 << 5), /*!< Set if no audio prompts should be played */
	USER_OPT_ANNOUNCEUSERCOUNT = (1 << 6), /*!< Set if the number of users should be announced to the caller */
	USER_OPT_WAITMARKED =   (1 << 7), /*!< Set if the user must wait for a marked user before starting */
	USER_OPT_ENDMARKED =    (1 << 8), /*!< Set if the user should be kicked after the last Marked user exits */
	USER_OPT_DENOISE =      (1 << 9), /*!< Sets if denoise filter should be used on audio before mixing. */
	USER_OPT_ANNOUNCE_JOIN_LEAVE = (1 << 10), /*!< Sets if the user's name should be recorded and announced on join and leave. */
	USER_OPT_TALKER_DETECT = (1 << 11), /*!< Sets if start and stop talking events should generated for this user over AMI. */
	USER_OPT_DROP_SILENCE =  (1 << 12), /*!< Sets if silence should be dropped from the mix or not. */
	USER_OPT_DTMF_PASS    =  (1 << 13), /*!< Sets if dtmf should be passed into the conference or not */
	USER_OPT_ANNOUNCEUSERCOUNTALL = (1 << 14), /*!< Sets if the number of users should be announced to everyone. */
	USER_OPT_JITTERBUFFER =  (1 << 15), /*!< Places a jitterbuffer on the user. */
	USER_OPT_ANNOUNCE_JOIN_LEAVE_REVIEW = (1 << 16), /*!< modifies ANNOUNCE_JOIN_LEAVE - user reviews the recording before continuing */
	USER_OPT_SEND_EVENTS = (1 << 17), /*!< Send text message events to users */
	USER_OPT_ECHO_EVENTS = (1 << 18), /*!< Send events only to the admin(s) */
#ifdef GRANDSTREAM_NETWORKS
	USER_OPT_MARKED_MCB = (1 << 19), /*!< Set if the user is a mcb */
	USER_OPT_JOINEDUSER = (1 << 20), /*!< Set if the user already joined */
	USER_OPT_MCB_CREATE_CHAN_IN_USE = (1 << 21), /*!< Set if the user is bridging the conference from mcb-create leg */
	USER_OPT_CAN_INVITE = (1 << 22), /*!< Set if the user has user invite menu */
	USER_OPT_MARKED_MCB_IDLE_KICK = (1 << 23), /*!< Set if then user is a MCB and it's idle to kick when no more user in conference.*/
	USER_OPT_INVITEDUSER = (1 << 24), /*!< Set if the user invited to join */
	USER_OPT_INVITE_ALLOW = (1 << 25), /*!< Set if allowing user invite */
	USER_OPT_MENU_ALLOW = (1 << 26), /*!< Set if allowing conference menu */
#endif
};

enum bridge_profile_flags {
	BRIDGE_OPT_RECORD_CONFERENCE = (1 << 0), /*!< Set if the conference should be recorded */
	BRIDGE_OPT_VIDEO_SRC_LAST_MARKED = (1 << 1), /*!< Set if conference should feed video of last marked user to all participants. */
	BRIDGE_OPT_VIDEO_SRC_FIRST_MARKED = (1 << 2), /*!< Set if conference should feed video of first marked user to all participants. */
	BRIDGE_OPT_VIDEO_SRC_FOLLOW_TALKER = (1 << 3), /*!< Set if conference set the video feed to follow the loudest talker.  */
	BRIDGE_OPT_RECORD_FILE_APPEND = (1 << 4), /*!< Set if the record file should be appended to between start/stops.  */
	BRIDGE_OPT_RECORD_FILE_TIMESTAMP = (1 << 5), /*!< Set if the record file should have a timestamp appended */
	BRIDGE_OPT_BINAURAL_ACTIVE = (1 << 6), /*!< Set if binaural convolution is activated */
	BRIDGE_OPT_VIDEO_SRC_SFU = (1 << 7), /*!< Selective forwarding unit */
	BRIDGE_OPT_REMB_BEHAVIOR_AVERAGE = (1 << 8), /*!< The average of all REMB reports is sent to the sender */
	BRIDGE_OPT_REMB_BEHAVIOR_LOWEST = (1 << 9), /*!< The lowest estimated maximum bitrate is sent to the sender */
	BRIDGE_OPT_REMB_BEHAVIOR_HIGHEST = (1 << 10), /*!< The highest estimated maximum bitrate is sent to the sender */
	BRIDGE_OPT_ENABLE_EVENTS = (1 << 11), /*!< Enable sending events to participants */
#ifdef GRANDSTREAM_NETWORKS
	BRIDGE_OPT_VIDEO_RECORD_CONFERENCE = (1 << 12), /*!< Set if the conference should be video recorded */
	BRIDGE_OPT_ENABLE_MCU = (1 << 13), /*!< Set if the conference will be use mcu */
	BRIDGE_OPT_SLIDES_RECORD_CONFERENCE = (1 << 14), /*!< Set if the conference should be slides video recorded */
	BRIDGE_OPT_FOCUS_RECORD_CONFERENCE = (1 << 15), /*!< Set if the conference should be focus video recorded */
	BRIDGE_OPT_ALLOW_USER_UNMUTE = (1 << 16), /*!< Set if the conference allow user unmute */
	BRIDGE_OPT_RECORD_PROMPT = (1 << 17),/*!< Set if allowing play record prompt*/
#endif
};

enum conf_menu_action_id {
	MENU_ACTION_TOGGLE_MUTE = 1,
	MENU_ACTION_PLAYBACK,
	MENU_ACTION_PLAYBACK_AND_CONTINUE,
	MENU_ACTION_INCREASE_LISTENING,
	MENU_ACTION_DECREASE_LISTENING,
	MENU_ACTION_RESET_LISTENING,
	MENU_ACTION_RESET_TALKING,
	MENU_ACTION_INCREASE_TALKING,
	MENU_ACTION_DECREASE_TALKING,
	MENU_ACTION_DIALPLAN_EXEC,
	MENU_ACTION_ADMIN_TOGGLE_LOCK,
	MENU_ACTION_ADMIN_KICK_LAST,
	MENU_ACTION_LEAVE,
	MENU_ACTION_NOOP,
	MENU_ACTION_SET_SINGLE_VIDEO_SRC,
	MENU_ACTION_RELEASE_SINGLE_VIDEO_SRC,
	MENU_ACTION_PARTICIPANT_COUNT,
	MENU_ACTION_ADMIN_TOGGLE_MUTE_PARTICIPANTS,
	MENU_ACTION_TOGGLE_BINAURAL,
#ifdef GRANDSTREAM_NETWORKS
	MENU_ACTION_ADMIN_KICK_MCBs,		  /* Set to kick MCBs, added by WeilingLiao*/
	MENU_ACTION_ADMIN_KICK_ALL_NON_ADMIN, /* Set to kick all non-administrator users, added by WeilingLiao*/
	MENU_ACTION_ADMIN_TOGGLE_RECORD,	  /* Set to toggle record conference, added by WeilingLiao*/
	MENU_ACTION_CHANGE_ROLE,
	MENU_ACTION_INVITE_CONTACT_GROUP      /* Inivte contact group*/
#endif
};

#ifdef GRANDSTREAM_NETWORKS
enum conf_operation_type {
	CONF_OP_START,
	CONF_OP_END,
	CONF_OP_START_RECORD,
	CONF_OP_END_RECORD,
	CONF_OP_JOIN,
	CONF_OP_LEAVE,
	CONF_OP_INVITE,
	CONF_OP_INVITE_MCB,
	CONF_OP_INVITE_CONTACT_GROUP,
	CONF_OP_OPEN_CAMERA,
	CONF_OP_CLOSE_CAMERA,
	CONF_OP_OPEN_SHARE,
	CONF_OP_CLOSE_SHARE,
	CONF_OP_MUTE,
	CONF_OP_MUTE_PARTICIPANTS,
	CONF_OP_UNMUTE,
	CONF_OP_UNMUTE_PARTICIPANTS,
	CONF_OP_LOCK,
	CONF_OP_UNLOCK,
	CONF_OP_KICK,
	CONF_OP_KICK_MCB,
	CONF_OP_KICK_PARTICIPANTS,
	CONF_OP_KICK_LAST,
	CONF_OP_TRANSFER_HOST,
	CONF_OP_RETRIEVE_HOST,
	CONF_OP_SWITCH_RESOLUTION,
	CONF_OP_CHANGE_ROLE,
	CONF_OP_HANGUP_CONF,
	CONF_OP_INCREASE_LISTEN_VOLUME,
	CONF_OP_DECREASE_LISTEN_VOLUME,
	CONF_OP_INCREASE_TALK_VOLUME,
	CONF_OP_DECREASE_TALK_VOLUME,
	CONF_OP_UPDATE_NAME,
	CONF_OP_MIC_MUTE,
	CONF_OP_MIC_UNMUTE,
	CONF_OP_SET_FOCUS,
	CONF_OP_DEL_FOCUS,
	CONF_OP_SET_LAYOUT,
	CONF_OP_DEL_LAYOUT,
	CONF_OP_SET_USER_PERMISSION,
	CONF_OP_DEL_USER_PERMISSION
};

enum call_type {
	AST_CALL_TYPE_UNKNOWN = 0,
	AST_CALL_TYPE_PSIP_TRUNK,
	AST_CALL_TYPE_DAHDIO_TRUNK,
	AST_CALL_TYPE_DAHDII_TRUNK,
	AST_CALL_TYPE_IAX2_TRUNK,
	AST_CALL_TYPE_ANONYMOUS,
	AST_CALL_TYPE_LOCAL,
};

enum {
	EXTEND_INIT = 0,
	EXTEND_PENDING,
	EXTEND_CONFIRMED,
	EXTEND_REJECTED
};

enum conf_act_type {
	AST_CONF_ACT_TYPE_UNKNOWN = 0,
	AST_CONF_ACT_TYPE_PUT_DOWN_ALL_HANDS,
};

#endif

/*! The conference menu action contains both
 *  the action id that represents the action that
 *  must take place, along with any data associated
 *  with that action. */
struct conf_menu_action {
	enum conf_menu_action_id id;
	union {
		char playback_file[PATH_MAX];
		struct {
			char context[AST_MAX_CONTEXT];
			char exten[AST_MAX_EXTENSION];
			int priority;
		} dialplan_args;
	} data;
	AST_LIST_ENTRY(conf_menu_action) action;
};

/*! Conference menu entries contain the DTMF sequence
 *  and the list of actions that are associated with that
 *  sequence. */
struct conf_menu_entry {
	/*! the DTMF sequence that triggers the actions */
	char dtmf[MAXIMUM_DTMF_FEATURE_STRING];
	/*! The actions associated with this menu entry. */
	AST_LIST_HEAD_NOLOCK(, conf_menu_action) actions;
	AST_LIST_ENTRY(conf_menu_entry) entry;
};

/*! Conference menu structure.  Contains a list
 * of DTMF sequences coupled with the actions those
 * sequences invoke.*/
struct conf_menu {
	char name[MAX_PROFILE_NAME];
	AST_LIST_HEAD_NOLOCK(, conf_menu_entry) entries;
};

struct user_profile {
	char name[MAX_PROFILE_NAME];
	char pin[MAX_PIN];
	char moh_class[128];
	char announcement[PATH_MAX];
#ifdef GRANDSTREAM_NETWORKS
	char mcb_create_chan[128];
	char mcb_type[16];
	char mcb_remote_conf_id[64];
	char admin_pin[MAX_PIN];
	/*! The time in s the user may wait for the schedule confbridge */
	unsigned int wait_timeout;
#endif
	unsigned int flags;
	unsigned int announce_user_count_all_after;
	/*! Minimum average magnitude threshold to determine talking by the DSP. */
	unsigned int talking_threshold;
	/*! Time in ms of silence necessary to declare talking stopped by the bridge. */
	unsigned int silence_threshold;
	/*! The time in ms the user may stay in the confbridge */
	unsigned int timeout;
};

enum conf_sounds {
	CONF_SOUND_HAS_JOINED,
	CONF_SOUND_HAS_LEFT,
	CONF_SOUND_KICKED,
	CONF_SOUND_MUTED,
	CONF_SOUND_UNMUTED,
	CONF_SOUND_ONLY_ONE,
	CONF_SOUND_THERE_ARE,
	CONF_SOUND_OTHER_IN_PARTY,
	CONF_SOUND_PLACE_IN_CONF,
	CONF_SOUND_WAIT_FOR_LEADER,
	CONF_SOUND_LEADER_HAS_LEFT,
	CONF_SOUND_GET_PIN,
	CONF_SOUND_INVALID_PIN,
	CONF_SOUND_ONLY_PERSON,
	CONF_SOUND_LOCKED,
	CONF_SOUND_LOCKED_NOW,
	CONF_SOUND_UNLOCKED_NOW,
	CONF_SOUND_ERROR_MENU,
	CONF_SOUND_JOIN,
	CONF_SOUND_LEAVE,
	CONF_SOUND_PARTICIPANTS_MUTED,
	CONF_SOUND_PARTICIPANTS_UNMUTED,
	CONF_SOUND_BEGIN,
	CONF_SOUND_BINAURAL_ON,
	CONF_SOUND_BINAURAL_OFF,
#ifdef GRANDSTREAM_NETWORKS
	CONF_SOUND_START_RECORD,
	CONF_SOUND_END_RECORD,
	CONF_SOUND_CONTINUE_WAIT,
	CONF_SOUND_GIVEUP_MODERATOR,
	CONF_SOUND_ACQUIRE_MODERATOR,
	CONF_SOUND_YOU_ARE_MODERATOR,
	CONF_SOUND_YOU_ARE_PATICIPANT,
	CONF_SOUND_SORRY_CANNOT_BE_MODERATOR,
	CONF_SOUND_TIMEOUT,
	CONF_SOUND_EXTEND_TIME,
	CONF_SOUND_START_VIDEO,
	CONF_SOUND_HAS_VIDEOED,
	CONF_SOUND_END_FIVE_MINUTES,
	CONF_SOUND_VIDEO_OFF,
	CONF_SOUND_RECORD_OFF,
	CONF_SOUND_START_PAGING_VIDEO,
	CONF_SOUND_START_FOCUS_VIDEO
#endif
};

struct bridge_profile_sounds {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(hasjoin);
		AST_STRING_FIELD(hasleft);
		AST_STRING_FIELD(kicked);
		AST_STRING_FIELD(muted);
		AST_STRING_FIELD(unmuted);
		AST_STRING_FIELD(onlyone);
		AST_STRING_FIELD(thereare);
		AST_STRING_FIELD(otherinparty);
		AST_STRING_FIELD(placeintoconf);
		AST_STRING_FIELD(waitforleader);
		AST_STRING_FIELD(leaderhasleft);
		AST_STRING_FIELD(getpin);
		AST_STRING_FIELD(invalidpin);
		AST_STRING_FIELD(onlyperson);
		AST_STRING_FIELD(locked);
		AST_STRING_FIELD(lockednow);
		AST_STRING_FIELD(unlockednow);
		AST_STRING_FIELD(errormenu);
		AST_STRING_FIELD(leave);
		AST_STRING_FIELD(join);
		AST_STRING_FIELD(participantsmuted);
		AST_STRING_FIELD(participantsunmuted);
		AST_STRING_FIELD(begin);
		AST_STRING_FIELD(binauralon);
		AST_STRING_FIELD(binauraloff);
#ifdef GRANDSTREAM_NETWORKS
		AST_STRING_FIELD(startrecord);
		AST_STRING_FIELD(endrecord);
		AST_STRING_FIELD(continuewait);
		AST_STRING_FIELD(giveupmoderator);
		AST_STRING_FIELD(acquiremoderator);
		AST_STRING_FIELD(youaremoderator);
		AST_STRING_FIELD(youareparticipant);
		AST_STRING_FIELD(sorryyoucantbemoderator);
		AST_STRING_FIELD(conftimeout);
		AST_STRING_FIELD(extendtime);
		AST_STRING_FIELD(pagingrecord);
#endif
	);
};

struct bridge_profile {
	char name[MAX_PROFILE_NAME];
	char language[MAX_LANGUAGE];		  /*!< Language used for playback_chan */
	char rec_file[PATH_MAX];
	char rec_options[128];
	char rec_command[128];
	unsigned int flags;
	unsigned int max_members;          /*!< The maximum number of participants allowed in the conference */
	unsigned int internal_sample_rate; /*!< The internal sample rate of the bridge. 0 when set to auto adjust mode. */
	unsigned int mix_interval;  /*!< The internal mixing interval used by the bridge. When set to 0 the bridgewill use a default interval. */
	struct bridge_profile_sounds *sounds;
	char regcontext[AST_MAX_CONTEXT];
	unsigned int video_update_discard; /*!< Amount of time after sending a video update request that subsequent requests should be discarded */
	unsigned int remb_send_interval; /*!< Interval at which a combined REMB frame is sent to video sources */
#ifdef GRANDSTREAM_NETWORKS
	unsigned int one_person_warning;  /*! Duration into the one person when warnings should begin (specified in milliseconds or 0 to disable) */
	unsigned int one_person_warning_interval;  /*! Interval between the one person warnings (specified in milliseconds or 0 to disable) */
	unsigned int conf_admin_pin;      /*Whether the conference room has a admin pin code*/
	unsigned int n_way_conf;          /*Whether the conference room is N-way-conf*/
	unsigned int is_random_conf;      /*Whether the conference is random conference */
	unsigned int conf_timeout;        /*Maximum duration of the meeting (specified in milliseconds or 0 to disable)*/
#endif
};

// SEND_STATUS is used to mark whether the CDR, recording, and video functions are affected by the lack of storage space
#define SEND_STATUS_CDR       1   //0x01  Takes the first bit of the int data
#define SEND_STATUS_RECORDING 2   //0x10  Takes the second bit of the int data
#define SEND_STATUS_VIDEOING  4   //0x100 Takes the third bit of the int data

#ifdef GRANDSTREAM_NETWORKS
struct conference_one_person_warning {
	struct confbridge_user *user;
	int id;
	int playing;
	int need_warning;
	int extend_time_status;                       /*!< Mark the status of the user timeout notify response */
};

struct conf_contact_member {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);                  /*!< Name of the contact member*/
		AST_STRING_FIELD(number);                /*!< Number of the contact member*/
	);

	AST_LIST_ENTRY(conf_contact_member) entry;   /*!< Linked list information */
};

struct conf_contact_group {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);                                     /*!< Name of the contact group*/
		AST_STRING_FIELD(prompt);                                   /*!< Prompt of the contact group for menu navigator>*/
		AST_STRING_FIELD(pin);                                      /*!< PIN of the contact group*/
		AST_STRING_FIELD(conference);                               /*!< The Conference to use the contact group*/
	);
	int inuse;                                                      /*!< Set it if INUSE*/
	AST_LIST_HEAD_NOLOCK(, conf_contact_member) member_list;        /*!< List of the contact members in the group*/
};

struct conf_report_user_record {
	const char *number;
	const char *name;
	const char *group_name;
	struct timeval start;
	struct timeval end;
};

struct conf_report_conference_record {
	int id;
	struct timeval start;
	struct timeval end;
};

enum conf_user_role {
	CONF_USER_ROLE_UNKNOW,
	CONF_USER_ROLE_ADMIN,
	CONF_USER_ROLE_PARTICIPANT,
};

struct conf_notify_info {
	enum conf_user_role role;
};

struct conf_notify_switch_data {
	char conf_id[MAX_CONF_NAME];
	int is_video_conf;
	int is_admin;
	int enabled_cei;
	int event_type;      /*0 means nway type, 1 means call flip type*/
};

struct schedule_conference_wait_info{
	int conf_type;
	const char *pin;
	const char *admin_pin;
	const char *conf_name;
	const char * schedule_id;
	const char * utc_start_time;
	const char * start_time;
	const char * conf_subject;
	const char * host_exten;
	const char * host_email;
};

struct conf_timeout_notify_data {
	char conf_id[MAX_CONF_NAME];
	int notify_type;     /*0 means user type, 1 means conf type*/
	int status;    /*1 means pending, 2 means confirmed, 3 means rejected*/
	int time;
};

enum conf_record_type {
	CONF_RECORD_INIT,
	CONF_RECORD_SHUTDOWN,
	CONF_RECORD_AUDIO,
	CONF_RECORD_VIDEO,
	CONF_RECORD_SLIDES,
	CONF_RECORD_FOUCS,
};

enum mute_type {
	CONF_MUTE_INIT,
	CONF_MUTE_BY_SELF,
	CONF_MUTE_BY_HOST,
};
#endif

/*! \brief The structure that represents a conference bridge */
struct confbridge_conference {
#ifdef GRANDSTREAM_NETWORKS
	int max_users;                                                    /*!< Maximum number of user number in confernece. Not the total user number */
#endif
	char name[MAX_CONF_NAME];                                         /*!< Name of the conference bridge */
	struct confbridge_state *state;                                   /*!< Conference state information */
	struct ast_bridge *bridge;                                        /*!< Bridge structure doing the mixing */
	struct bridge_profile b_profile;                                  /*!< The Bridge Configuration Profile */
	unsigned int activeusers;                                         /*!< Number of active users present */
	unsigned int markedusers;                                         /*!< Number of marked users present */
	unsigned int waitingusers;                                        /*!< Number of waiting users present */
#ifdef GRANDSTREAM_NETWORKS
	unsigned int mcb_idle_kick_users;                                 /*!< Number of mcb users who is idle to kick */
#endif
	unsigned int locked:1;                                            /*!< Is this conference bridge locked? */
	unsigned int muted:1;                                             /*!< Is this conference bridge muted? */
	struct ast_channel *playback_chan;                                /*!< Channel used for playback into the conference bridge */
	struct ast_channel *record_chan;                                  /*!< Channel used for recording the conference */
	struct ast_str *record_filename;                                  /*!< Recording filename. */
#ifdef GRANDSTREAM_NETWORKS
	struct ast_str *record_files;                                     /*!< all recordfiles in this conference */
	struct ast_str *record_files_desc;
	struct ast_str *record_files_uuid;
	struct ast_str *video_filename;                                  /*!< every Videoing filename. */
	struct ast_str *video_files;                                     /*!< all video files in this conference */
	struct ast_str *video_files_desc;
	struct ast_str *video_files_uuid;
#endif
	struct ast_str *orig_rec_file;                                    /*!< Previous b_profile.rec_file. */
	AST_LIST_HEAD_NOLOCK(, confbridge_user) active_list;              /*!< List of users participating in the conference bridge */
	AST_LIST_HEAD_NOLOCK(, confbridge_user) waiting_list;             /*!< List of users waiting to join the conference bridge */
	struct ast_taskprocessor *playback_queue;                         /*!< Queue for playing back bridge announcements and managing the announcer channel */
#ifdef GRANDSTREAM_NETWORKS
	ast_mutex_t playback_lock;										  /*!< Lock used for playback channel */
	struct conference_one_person_warning only_warning;				  /*!< Warning when only one person in conference*/
	struct conf_report_conference_record report_record;               /*!< Call Record*/
	char start_time[32];                                              /*!< Start time of the conference*/
	int is_conference_service;                                        /*!< Set if the confbridge is for the conference service */
	char *schedule_id_to_start;                                       /*!< The schedule id will start */
	char *schedule_start_time;                                        /*!< The start time of the schedule conference */
	struct timeval schedule_alert_time;                               /*!< The alert time of the schedule conference */
	char *feature_id;                                                 /*!< The feature id of operation log*/
	char *schedule_subject;                                           /*!< The subject of the schedule conference */
	char *schedule_id_processing;									  /*!< The schedule id of the conference in processing */
	time_t conference_start;										  /*!< The start timestamp of the conference*/
	char *schedule_host_exten;										  /*!< The extension of the schedule conference host*/
	char *schedule_host_email;										  /*!< The email of the schedule conference host*/
	char *schedule_utc_start_time;									  /*!< The config utc start time of the schedule conference*/
	int peak_users;                                                   /*!< Peak users in the conference room. Not the total user number */
	int peak_camera;                                                  /*!< Peak of users with cameras on. Not the total cameras number */
	int num_camera;                                                   /*!< Number of users with cameras on now. */
	int conf_end;                                                     /*!< Identifies whether the meeting room is set to end */
	int sch_time;                                                     /*!< How many timed sched tasks are left in the meeting room */
	unsigned int conf_sched_id;                                       /*!< sched id for meeting timeout */
	struct timeval schedule_start_timeval;
	int video_is_recording;
	int slides_is_recording;
	int focus_is_recording;
	time_t record_start_utc_timestamp;
	time_t video_start_utc_timestamp;
	int mcu_enabled;
	int play_videoing_sound;
	char *im_service_prefix;
	int recorded;                                                      /*!< conference bridge record type */
	int conf_will_timeout;                                             /*!< Mark that the meeting room will timeout in 15min */
	int extend_time_status;                                            /*!< Mark the status of meeting timeout notify response */
	struct ast_channel *focus_chan;                                    /*!< Channel used for focusing the conference */
	char record_link[MAX_RECORD_LINK];
	char conf_layout[MAX_CONF_LAYOUT];                                 /*!< conference layout mode */
	char *share_system_audio_chan;                                     /*!< the channel name who share the system audio */
	int allow_user_unmute;
	int user_can_invite;
	struct stasis_message_router *bridge_router;
	int schedule_hangup;
	int is_paging_conf;
	struct ast_channel *paging_caller_chan;
	char *link_auth_id;
	int virtual_trunk_conf;
	char *conf_creator;                                                /*!< The creator of the conference */
	char recordfile_slave_media[256];
	struct timeval last_notify;                                        /*!< The last time for conference timeout notification */
	time_t record_pause_duration;
#endif
};

extern struct ao2_container *conference_bridges;

struct post_join_action {
	int (*func)(struct confbridge_user *user);
	AST_LIST_ENTRY(post_join_action) list;
};

/*! \brief The structure that represents a conference bridge user */
struct confbridge_user {
#ifdef GRANDSTREAM_NETWORKS
	int user_no;                                 /*!< User Number */
	time_t kicktime;                             /*!< Time the user will be kicked from the conference */
	struct conf_report_user_record report_record;/*!< Call Record*/
#endif
	struct confbridge_conference *conference;    /*!< Conference bridge they are participating in */
	struct bridge_profile b_profile;             /*!< The Bridge Configuration Profile */
	struct user_profile u_profile;               /*!< The User Configuration Profile */
	char menu_name[MAX_PROFILE_NAME];            /*!< The name of the DTMF menu assigned to this user */
	char name_rec_location[PATH_MAX];            /*!< Location of the User's name recorded file if it exists */
	struct ast_channel *chan;                    /*!< Asterisk channel participating */
	struct ast_bridge_features features;         /*!< Bridge features structure */
	struct ast_bridge_tech_optimizations tech_args; /*!< Bridge technology optimizations for talk detection */
	unsigned int suspended_moh;                  /*!< Count of active suspended MOH actions. */
	unsigned int muted:1;                        /*!< Has the user requested to be muted? */
	unsigned int kicked:1;                       /*!< User has been kicked from the conference */
	unsigned int playing_moh:1;                  /*!< MOH is currently being played to the user */
	unsigned int talking:1;                      /*!< User is currently talking */
#ifdef GRANDSTREAM_NETWORKS
	unsigned int video:1;                        /*!< User is a vide call */
	unsigned int exe_mute:1;                     /*!< Muted/Umuted will be executed */
	unsigned int play_sound;                      /*!< Force to play the kick sound event or timeout sound event. 1 means kick,2 means timeout*/
	unsigned int mic_muted:1;                    /*!< mark the user's mic is muted? */
	unsigned int talking_before_muted:1;         /*!< User is currently talking before mute*/
	enum ast_terminal_type term_type;
	struct timeval jointime;
	int talking_rms;                              /*!< The Energy value of talking*/
	char related_ext[MAX_RELATED_EXT];            /*!< The related exten for confXXXX exten */
	unsigned int user_sched_id;                   /*!< sched id for user one person warning */
	int menu_operation;
	unsigned int isfocus:1;                       /*!< mark the user is focus? */
	unsigned int hand_up:1;                       /*!< mark the user is hand up? */
	time_t hand_up_time;
	unsigned int had_play_tone;                  //added by zyxia for user is page callee playtone
	unsigned int is_playing_annoncement;                  //mark is playing file
	int mute_type;                                /*!< muted by self or others */
	unsigned int user_mcb_sched_id;               /*!< sched id for mcb user auto release */
	unsigned int suspended_mute;                  /*!< need suspend mute sound */
#endif
	AST_LIST_HEAD_NOLOCK(, post_join_action) post_join_list; /*!< List of sounds to play after joining */;
	AST_LIST_ENTRY(confbridge_user) list;        /*!< Linked list information */
};

/*! \brief load confbridge.conf file */
int conf_load_config(void);

/*! \brief reload confbridge.conf file */
int conf_reload_config(void);

/*! \brief destroy the information loaded from the confbridge.conf file*/
void conf_destroy_config(void);

/*!
 * \brief find a user profile given a user profile's name and store
 * that profile in result structure.
 *
 * \param chan channel the user profile is requested for
 * \param user_profile_name name of the profile requested (optional)
 * \param result data contained by the user profile will be copied to this struct pointer
 *
 * \details If user_profile_name is not provided, this function will
 * check for the presence of a user profile set by the CONFBRIDGE
 * function on a channel datastore. If that doesn't exist, the
 * default_user profile is used.
 *
 * \retval user profile on success
 * \retval NULL on failure
 */
const struct user_profile *conf_find_user_profile(struct ast_channel *chan, const char *user_profile_name, struct user_profile *result);

/*!
 * \brief Find a bridge profile given a bridge profile's name and store
 * that profile in result structure.
 *
 * \param chan channel the bridge profile is requested for
 * \param bridge_profile_name name of the profile requested (optional)
 * \param result data contained by the bridge profile will be copied to this struct pointer
 *
 * \details If bridge_profile_name is not provided, this function will
 * check for the presence of a bridge profile set by the CONFBRIDGE
 * function on a channel datastore. If that doesn't exist, the
 * default_bridge profile is used.
 *
 * \retval bridge profile on success
 * \retval NULL on failure
 */
const struct bridge_profile *conf_find_bridge_profile(struct ast_channel *chan, const char *bridge_profile_name, struct bridge_profile *result);

/*!
 * \brief find a menu profile given a menu profile's name and apply
 * the menu in DTMF hooks.
 *
 * \param chan channel the menu profile is requested for
 * \param user user profile the menu is being applied to
 * \param menu_profile_name name of the profile requested (optional)
 *
 * \details If menu_profile_name is not provided, this function will
 * check for the presence of a menu profile set by the CONFBRIDGE
 * function on a channel datastore. If that doesn't exist, the
 * default_menu profile is used.
 *
 * \retval 0 on success
 * \retval -1 on failure
 */
int conf_set_menu_to_user(struct ast_channel *chan, struct confbridge_user *user, const char *menu_profile_name);

/*!
 * \brief Destroy a bridge profile found by 'conf_find_bridge_profile'
 */
void conf_bridge_profile_destroy(struct bridge_profile *b_profile);

/*!
 * \brief copies a bridge profile
 * \note conf_bridge_profile_destroy must be called on the dst structure
 */
void conf_bridge_profile_copy(struct bridge_profile *dst, struct bridge_profile *src);

/*!
 * \brief Finds a menu_entry in a menu structure matched by DTMF sequence.
 *
 * \note the menu entry found must be destroyed using conf_menu_entry_destroy()
 *
 * \retval 1 success, entry is found and stored in result
 * \retval 0 failure, no entry found for given DTMF sequence
 */
int conf_find_menu_entry_by_sequence(const char *dtmf_sequence, struct conf_menu *menu, struct conf_menu_entry *result);

/*!
 * \brief Destroys and frees all the actions stored in a menu_entry structure
 */
void conf_menu_entry_destroy(struct conf_menu_entry *menu_entry);

/*!
 * \brief Once a DTMF sequence matches a sequence in the user's DTMF menu, this function will get
 * called to perform the menu action.
 *
 * \param bridge_channel Bridged channel this is involving
 * \param user the conference user to perform the action on.
 * \param menu_entry the menu entry that invoked this callback to occur.
 * \param menu an AO2 referenced pointer to the entire menu structure the menu_entry
 *        derived from.
 *
 * \note The menu_entry is a deep copy of the entry found in the menu structure.  This allows
 * for the menu_entry to be accessed without requiring the menu lock.  If the menu must
 * be accessed, the menu lock must be held.  Reference counting of the menu structure is
 * handled outside of the scope of this function.
 *
 * \retval 0 success
 * \retval -1 failure
 */
int conf_handle_dtmf(
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user,
	struct conf_menu_entry *menu_entry,
	struct conf_menu *menu);


/*! \brief Looks to see if sound file is stored in bridge profile sounds, if not
 *  default sound is provided.*/
const char *conf_get_sound(enum conf_sounds sound, struct bridge_profile_sounds *custom_sounds);

int func_confbridge_helper(struct ast_channel *chan, const char *cmd, char *data, const char *value);

/*!
 * \brief Play sound file into conference bridge
 *
 * \param conference The conference bridge to play sound file into
 * \param filename Sound file to play
 *
 * \retval 0 success
 * \retval -1 failure
 */
int play_sound_file(struct confbridge_conference *conference, const char *filename);

/*!
 * \brief Play sound file into conference bridge asynchronously
 *
 * If the initiator parameter is non-NULL, then the playback will wait for
 * that initiator channel to get back in the bridge before playing the sound
 * file. This way, the initiator has no danger of hearing a "clipped" file.
 *
 * \param conference The conference bridge to play sound file into
 * \param filename Sound file to play
 * \param initiator Channel that initiated playback.
 *
 * \retval 0 success
 * \retval -1 failure
 */
int async_play_sound_file(struct confbridge_conference *conference, const char *filename,
	struct ast_channel *initiator);

/*!
 * \brief Indicate the initiator of an async sound file is ready for it to play.
 *
 * When playing an async sound file, the initiator is typically either out of the bridge
 * or not in a position to hear the queued announcement. This function lets the announcement
 * thread know that the initiator is now ready for the sound to play.
 *
 * If an async announcement was queued and no initiator channel was provided, then this is
 * a no-op
 *
 * \param chan The channel that initiated the async announcement
 */
void async_play_sound_ready(struct ast_channel *chan);

#ifdef GRANDSTREAM_NETWORKS
/*!
 * \brief Play sound file into conference bridge by user's playback channel.
 *
 * \param user The user that has the conference bridge to play sound file into
 * \param filename Sound file to play
 *
 * \retval 0 success
 * \retval -1 failure
 */
int user_announcer_play_sound_file(struct confbridge_user *user, const char *filename);
void synchronize_play_sound_file(struct confbridge_conference *conference, const char *filename, int say_number);
#endif

/*! \brief Callback to be called when the conference has become empty
 * \param conference The conference bridge
 */
void conf_ended(struct confbridge_conference *conference);

/*!
 * \brief Update the actual mute status of the user and set it on the bridge.
 *
 * \param user User to update the mute status.
 *
 * \return Nothing
 */
void conf_update_user_mute(struct confbridge_user *user);

/*!
 * \brief Stop MOH for the conference user.
 *
 * \param user Conference user to stop MOH on.
 *
 * \return Nothing
 */
void conf_moh_stop(struct confbridge_user *user);

/*!
 * \brief Start MOH for the conference user.
 *
 * \param user Conference user to start MOH on.
 *
 * \return Nothing
 */
void conf_moh_start(struct confbridge_user *user);

/*! \brief Attempt to mute/play MOH to the only user in the conference if they require it
 * \param conference A conference bridge containing a single user
 */
void conf_mute_only_active(struct confbridge_conference *conference);

/*! \brief Callback to execute any time we transition from zero to one active users
 * \param conference The conference bridge with a single active user joined
 * \retval 0 success
 * \retval -1 failure
 */
void conf_handle_first_join(struct confbridge_conference *conference);

/*! \brief Handle actions every time a waitmarked user joins w/o a marked user present
 * \param user The waitmarked user
 * \retval 0 success
 * \retval -1 failure
 */
int conf_handle_inactive_waitmarked(struct confbridge_user *user);

/*! \brief Handle actions whenever an user joins an empty conference
 *
 * \param user The user
 */
int conf_handle_only_person(struct confbridge_user *user);

/*! \brief Handle when a conference moves to having more than one active participant
 * \param conference The conference bridge with more than one active participant
 */
void conf_handle_second_active(struct confbridge_conference *conference);

/*! \brief Add a conference bridge user as an unmarked active user of the conference
 * \param conference The conference bridge to add the user to
 * \param user The conference bridge user to add to the conference
 */
void conf_add_user_active(struct confbridge_conference *conference, struct confbridge_user *user);

/*! \brief Add a conference bridge user as a marked active user of the conference
 * \param conference The conference bridge to add the user to
 * \param user The conference bridge user to add to the conference
 */
void conf_add_user_marked(struct confbridge_conference *conference, struct confbridge_user *user);

/*! \brief Add a conference bridge user as an waiting user of the conference
 * \param conference The conference bridge to add the user to
 * \param user The conference bridge user to add to the conference
 */
void conf_add_user_waiting(struct confbridge_conference *conference, struct confbridge_user *user);

/*! \brief Remove a conference bridge user from the unmarked active conference users in the conference
 * \param conference The conference bridge to remove the user from
 * \param user The conference bridge user to remove from the conference
 */
void conf_remove_user_active(struct confbridge_conference *conference, struct confbridge_user *user);

/*! \brief Remove a conference bridge user from the marked active conference users in the conference
 * \param conference The conference bridge to remove the user from
 * \param user The conference bridge user to remove from the conference
 */
void conf_remove_user_marked(struct confbridge_conference *conference, struct confbridge_user *user);

/*! \brief Remove a conference bridge user from the waiting conference users in the conference
 * \param conference The conference bridge to remove the user from
 * \param user The conference bridge user to remove from the conference
 */
void conf_remove_user_waiting(struct confbridge_conference *conference, struct confbridge_user *user);

/*! \brief Queue a function to run with the given conference bridge user as an argument once the state transition is complete
 * \param user The conference bridge user to pass to the function
 * \param func The function to queue
 * \retval 0 success
 * \retval non-zero failure
 */
int conf_add_post_join_action(struct confbridge_user *user, int (*func)(struct confbridge_user *user));

/*!
 * \since 12.0
 * \brief get the confbridge start stasis message type
 *
 * \retval stasis message type for confbridge start messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_start_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge end stasis message type
 *
 * \retval stasis message type for confbridge end messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_end_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge join stasis message type
 *
 * \retval stasis message type for confbridge join messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_join_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge leave stasis message type
 *
 * \retval stasis message type for confbridge leave messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_leave_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge start_record stasis message type
 *
 * \retval stasis message type for confbridge start_record messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_start_record_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge stop_record stasis message type
 *
 * \retval stasis message type for confbridge stop_record messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_stop_record_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge mute stasis message type
 *
 * \retval stasis message type for confbridge mute messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_mute_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge unmute stasis message type
 *
 * \retval stasis message type for confbridge unmute messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_unmute_type(void);

/*!
 * \since 12.0
 * \brief get the confbridge talking stasis message type
 *
 * \retval stasis message type for confbridge talking messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_talking_type(void);

/*!
 * \since 15.5
 * \brief get the confbridge welcome stasis message type
 *
 * \retval stasis message type for confbridge welcome messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_welcome_type(void);

#ifdef GRANDSTREAM_NETWORKS
/*!
 * \since 13.0
 * \brief get the confbridge lock status stasis message type
 *
 * \retval stasis message type for confbridge lock status messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_lock_status_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge moderator stasis message type
 *
 * \retval stasis message type for confbridge moderator messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_moderator_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge media update stasis message type
 *
 * \retval stasis message type for confbridge media update messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_media_update_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge role update stasis message type
 *
 * \retval stasis message type for confbridge role update messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_role_update_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge lock status stasis message type
 *
 * \retval stasis message type for confbridge mute all messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_muteall_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge lock status stasis message type
 *
 * \retval stasis message type for confbridge unmute all messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_unmuteall_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge name update stasis message type
 *
 * \retval stasis message type for confbridge name update messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_name_update_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge focus update stasis message type
 *
 * \retval stasis message type for confbridge focus update messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_focus_status_type(void);

/*!
 * \since 13.0
 * \brief get the confbridge layout update stasis message type
 *
 * \retval stasis message type for confbridge layout update messages if it's available
 * \retval NULL if it isn't
 */
struct stasis_message_type *confbridge_layout_status_type(void);

#endif

/*!
 * \since 15.5
 * \brief Get the string representation of a confbridge stasis message type
 *
 * \param event_type The confbridge event type such as 'confbridge_welcome_type()'
 *
 * \retval The string representation of the message type
 * \retval "unknown" if not found
 */
const char *confbridge_event_type_to_string(struct stasis_message_type *event_type);

/*!
 * \since 12.0
 * \brief register stasis message routers to handle manager events for confbridge messages
 *
 * \retval 0 success
 * \retval non-zero failure
 */
int manager_confbridge_init(void);

/*!
 * \since 12.0
 * \brief unregister stasis message routers to handle manager events for confbridge messages
 */
void manager_confbridge_shutdown(void);

/*!
 * \brief Get ConfBridge record channel technology struct.
 * \since 12.0.0
 *
 * \return ConfBridge record channel technology.
 */
struct ast_channel_tech *conf_record_get_tech(void);

/*!
 * \brief Get ConfBridge announce channel technology struct.
 * \since 12.0.0
 *
 * \return ConfBridge announce channel technology.
 */
struct ast_channel_tech *conf_announce_get_tech(void);

/*!
 * \brief Push the announcer channel into the conference.
 * \since 12.0.0
 *
 * \param ast Either channel in the announcer channel pair.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
int conf_announce_channel_push(struct ast_channel *ast);

/*!
 * \brief Find a confbridge by name.
 * \since 13.22.0
 * \since 15.5.0
 *
 * \param confbridge_name The name to search for
 *
 * \return ConfBridge (which must be unreffed) or NULL.
 */
struct confbridge_conference *conf_find_bridge(const char *conference_name);

/*!
 * \brief Send events to bridge participants.
 * \since 15.7
 * \since 16.1
 *
 * \param conference The conference bridge
 * \param chan The channel triggering the action
 * \param msg The stasis message describing the event
 */
void conf_send_event_to_participants(struct confbridge_conference *conference,
	struct ast_channel *chan, struct stasis_message *msg);

#ifdef GRANDSTREAM_NETWORKS
/*!
 * \brief execute h@mcb-create extension in dialplan
 * \since 13.0
 *
 * \param ast Either channel in the announcer channel pair.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
int exec_mcb_authed(struct ast_channel *chan);

/*!
 * \brief check if nedd to play muted prompt to the new user
 * \since 13.0
 *
 * \param conference The conference bridge to mute
 * \param user The conference bridge user to mute
 */
void handle_new_user_muted_status(struct confbridge_conference *conference, struct confbridge_user *user);

/*!
 * \brief check if need to play conference recording prompt to the new user
 * \since 13.0
 *
 * \param conference The conference bridge to mute
 * \param user The conference bridge user to mute
 */
void handle_new_user_recorded_status(struct confbridge_conference *conference, struct confbridge_user *user);

/*!
 * \brief check if need to play conference videoing prompt to the new user
 * \since 13.0
 *
 * \param conference The conference bridge
 * \param user The conference bridge user
 */
void handle_new_user_videoed_status(struct confbridge_conference *conference, struct confbridge_user *user);
#endif

#ifdef GRANDSTREAM_NETWORKS
int manager_cei_init(void);
void manager_cei_shutdown(void);

struct stasis_message_type *cei_join_type(void);
struct stasis_message_type *cei_toggle_mute_type(void);
struct stasis_message_type *cei_toggle_record_type(void);
struct stasis_message_type *cei_lock_status_type(void);
struct stasis_message_type *cei_leave_type(void);
struct stasis_message_type *cei_talk_status_type(void);
struct stasis_message_type *cei_moderator_type(void);
struct stasis_message_type *cei_refresh_all_type(void);
struct stasis_message_type *cei_mute_status_type(void);
struct stasis_message_type *cei_role_update_type(void);
struct stasis_message_type *cei_name_update_type(void);
struct stasis_message_type *cei_focus_status_type(void);
struct stasis_message_type *cei_layout_status_type(void);
struct stasis_message_type *cei_user_permission_type(void);
struct stasis_message_type *cei_gesture_status_type(void);
struct stasis_message_type *cei_act_event_type(void);

int conf_cei_add_sharing_user(const char *channel_name, const char *conf_number, int share_audio);
int conf_cei_delete_sharing_user(const char *channel_name, const char *conf_number);
int conf_cei_video_resolution_update(const char *channel_name, const char *conf_number, int resolution);
int conf_cei_media_update(const char *channel_name, const char *conf_number, int is_video);
void confbridge_user_media_update(const char *channel_name, const char *conf_number, int is_video);
int conf_unset_menu_from_user(struct ast_channel *chan, struct confbridge_user *user);
void conf_uniqueid_strnprint(char *buffer, size_t buffer_size, unsigned int id);
const char *conf_operation_type_to_action(int type);
void conf_get_ucm_dec_info(char *buffer, int buffer_size);
void conf_cei_sche_user_load(const char *conf_name, const char *scheduel_id, const char *extension, const char *related_info, int user_signed);
int conf_record_schedule_users_signed(const char *conference, const char *schedule_id, const char *extension, const char *related_info, const char *schedule_utc_time);
int conf_mic_mute_unmute_helper(int mute, const char *conference_name, const char *chan_name);
int conf_get_exten_by_channel(const char *channel_name, char *buf, int size);
int conf_cei_send_trunk_conf_join_notify(const char *conf_num);
struct cei_conference* get_conf_by_num(const char *conf_num);
void remove_virtual_conference(struct cei_conference *conf);
int conf_user_hand_up_helper(int hand_up, const char *conference_name, const char *chan_name, struct ast_channel *op_chan);
#endif

#endif
