/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2009, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Built in bridging features
 *
 * \author Joshua Colp <jcolp@digium.com>
 *
 * \ingroup bridges
 */

/*** MODULEINFO
  	<depend>ucmsecret</depend>
	<use type="module">res_monitor</use>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/bridge.h"
#include "asterisk/bridge_technology.h"
#include "asterisk/frame.h"
#include "asterisk/file.h"
#include "asterisk/app.h"
#include "asterisk/astobj2.h"
#include "asterisk/pbx.h"
#include "asterisk/parking.h"
#include "asterisk/features_config.h"
#include "asterisk/monitor.h"
#include "asterisk/mixmonitor.h"
#include "asterisk/audiohook.h"
#include "asterisk/causes.h"
#include "ucm_crypt.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/mod_format.h"
#include "asterisk/avs_controller.h"
#include "asterisk/core_local.h"
#include "asterisk/haserver.h"
#include "asterisk/dial.h"
#include "asterisk/multi_media.h"
#endif

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG ("Bridge");
#endif

enum set_touch_variables_res {
	SET_TOUCH_SUCCESS,
	SET_TOUCH_UNSET,
	SET_TOUCH_ALLOC_FAILURE,
};
void replaceTimestamp(char *str);
#ifdef GRANDSTREAM_NETWORKS
struct mixmonitor_ds {
	unsigned int destruction_ok;
	ast_cond_t destruction_condition;
	ast_mutex_t lock;

	/* The filestream is held in the datastore so it can be stopped
	 * immediately during stop_mixmonitor or channel destruction. */
	int fs_quit;

	struct ast_filestream *fs;
	struct ast_filestream *fs_read;
	struct ast_filestream *fs_write;

	struct ast_audiohook *audiohook;

	unsigned int samp_rate;
	char *filename;
	char *beep_id;
};
#endif

static void set_touch_variable(enum set_touch_variables_res *res, struct ast_channel *chan, const char *var_name, char **touch)
{
	const char *c_touch;

	if (*res == SET_TOUCH_ALLOC_FAILURE) {
		return;
	}
	c_touch = pbx_builtin_getvar_helper(chan, var_name);
	if (!ast_strlen_zero(c_touch)) {
		*touch = ast_strdup(c_touch);
		if (!*touch) {
			*res = SET_TOUCH_ALLOC_FAILURE;
		} else {
			*res = SET_TOUCH_SUCCESS;
		}
	}
}

static enum set_touch_variables_res set_touch_variables(struct ast_channel *chan, int is_mixmonitor, char **touch_format, char **touch_monitor, char **touch_monitor_prefix)
{
	enum set_touch_variables_res res = SET_TOUCH_UNSET;
	const char *var_format;
	const char *var_monitor;
	const char *var_prefix;

	SCOPED_CHANNELLOCK(lock, chan);

	if (is_mixmonitor) {
		var_format = "TOUCH_MIXMONITOR_FORMAT";
		var_monitor = "TOUCH_MIXMONITOR";
		var_prefix = "TOUCH_MIXMONITOR_PREFIX";
	} else {
		var_format = "TOUCH_MONITOR_FORMAT";
		var_monitor = "TOUCH_MONITOR";
		var_prefix = "TOUCH_MONITOR_PREFIX";
	}
	set_touch_variable(&res, chan, var_format, touch_format);
	set_touch_variable(&res, chan, var_monitor, touch_monitor);
	set_touch_variable(&res, chan, var_prefix, touch_monitor_prefix);

	return res;
}

static void stop_automonitor(struct ast_bridge_channel *bridge_channel, struct ast_channel *peer_chan, struct ast_features_general_config *features_cfg, const char *stop_message)
{
	ast_verb(4, "AutoMonitor used to stop recording call.\n");

	ast_channel_lock(peer_chan);
	if (ast_channel_monitor(peer_chan)) {
		if (ast_channel_monitor(peer_chan)->stop(peer_chan, 1)) {
			ast_verb(4, "Cannot stop AutoMonitor for %s\n", ast_channel_name(bridge_channel->chan));
			if (features_cfg && !(ast_strlen_zero(features_cfg->recordingfailsound))) {
				ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
			}
			ast_channel_unlock(peer_chan);
			return;
		}
	} else {
		/* Something else removed the Monitor before we got to it. */
		ast_channel_unlock(peer_chan);
		return;
	}

	ast_channel_unlock(peer_chan);

	if (features_cfg && !(ast_strlen_zero(features_cfg->courtesytone))) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}

	if (!ast_strlen_zero(stop_message)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, stop_message, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, stop_message, NULL);
	}
}

static void start_automonitor(struct ast_bridge_channel *bridge_channel, struct ast_channel *peer_chan, struct ast_features_general_config *features_cfg, const char *start_message)
{
	char *touch_filename;
	size_t len;
	int x;
	enum set_touch_variables_res set_touch_res;

	RAII_VAR(char *, touch_format, NULL, ast_free);
	RAII_VAR(char *, touch_monitor, NULL, ast_free);
	RAII_VAR(char *, touch_monitor_prefix, NULL, ast_free);

	set_touch_res = set_touch_variables(bridge_channel->chan, 0, &touch_format,
		&touch_monitor, &touch_monitor_prefix);
	switch (set_touch_res) {
	case SET_TOUCH_SUCCESS:
		break;
	case SET_TOUCH_UNSET:
		set_touch_res = set_touch_variables(peer_chan, 0, &touch_format, &touch_monitor,
			&touch_monitor_prefix);
		if (set_touch_res == SET_TOUCH_ALLOC_FAILURE) {
			return;
		}
		break;
	case SET_TOUCH_ALLOC_FAILURE:
		return;
	}

	if (!ast_strlen_zero(touch_monitor)) {
		len = strlen(touch_monitor) + 50;
		touch_filename = ast_alloca(len);
		snprintf(touch_filename, len, "%s-%ld-%s",
			S_OR(touch_monitor_prefix, "auto"),
			(long) time(NULL),
			touch_monitor);
	} else {
		char *caller_chan_id;
		char *peer_chan_id;

		caller_chan_id = ast_strdupa(S_COR(ast_channel_caller(bridge_channel->chan)->id.number.valid,
			ast_channel_caller(bridge_channel->chan)->id.number.str, ast_channel_name(bridge_channel->chan)));
		peer_chan_id = ast_strdupa(S_COR(ast_channel_caller(peer_chan)->id.number.valid,
			ast_channel_caller(peer_chan)->id.number.str, ast_channel_name(peer_chan)));
		len = strlen(caller_chan_id) + strlen(peer_chan_id) + 50;
		touch_filename = ast_alloca(len);
		snprintf(touch_filename, len, "%s-%ld-%s-%s",
			S_OR(touch_monitor_prefix, "auto"),
			(long) time(NULL),
			caller_chan_id,
			peer_chan_id);
	}

	for (x = 0; x < strlen(touch_filename); x++) {
		if (touch_filename[x] == '/') {
			touch_filename[x] = '-';
		}
	}

	ast_verb(4, "AutoMonitor used to record call. Filename: %s\n", touch_filename);

	if (ast_monitor_start(peer_chan, touch_format, touch_filename, 1, X_REC_IN | X_REC_OUT, NULL)) {
		ast_verb(4, "AutoMonitor feature was tried by '%s' but monitor failed to start.\n",
			ast_channel_name(bridge_channel->chan));
		return;
	}

	ast_monitor_setjoinfiles(peer_chan, 1);

	if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}

	if (!ast_strlen_zero(start_message)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, start_message, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, start_message, NULL);
	}

	pbx_builtin_setvar_helper(bridge_channel->chan, "TOUCH_MONITOR_OUTPUT", touch_filename);
	pbx_builtin_setvar_helper(peer_chan, "TOUCH_MONITOR_OUTPUT", touch_filename);
}

static int feature_automonitor(struct ast_bridge_channel *bridge_channel, void *hook_pvt)
{
	const char *start_message;
	const char *stop_message;
	struct ast_bridge_features_automonitor *options = hook_pvt;
	enum ast_bridge_features_monitor start_stop = options ? options->start_stop : AUTO_MONITOR_TOGGLE;
	int is_monitoring;

	RAII_VAR(struct ast_channel *, peer_chan, NULL, ast_channel_cleanup);
	RAII_VAR(struct ast_features_general_config *, features_cfg, NULL, ao2_cleanup);

	ast_channel_lock(bridge_channel->chan);
	features_cfg = ast_get_chan_features_general_config(bridge_channel->chan);
	ast_channel_unlock(bridge_channel->chan);
	ast_bridge_channel_lock_bridge(bridge_channel);
	peer_chan = ast_bridge_peer_nolock(bridge_channel->bridge, bridge_channel->chan);
	ast_bridge_unlock(bridge_channel->bridge);

	if (!peer_chan) {
		ast_verb(4, "Cannot start AutoMonitor for %s - can not determine peer in bridge.\n",
			ast_channel_name(bridge_channel->chan));
		if (features_cfg && !ast_strlen_zero(features_cfg->recordingfailsound)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
		}
		return 0;
	}

	ast_channel_lock(bridge_channel->chan);
	start_message = pbx_builtin_getvar_helper(bridge_channel->chan,
		"TOUCH_MONITOR_MESSAGE_START");
	start_message = ast_strdupa(S_OR(start_message, ""));
	stop_message = pbx_builtin_getvar_helper(bridge_channel->chan,
		"TOUCH_MONITOR_MESSAGE_STOP");
	stop_message = ast_strdupa(S_OR(stop_message, ""));
	ast_channel_unlock(bridge_channel->chan);

	is_monitoring = ast_channel_monitor(peer_chan) != NULL;
	switch (start_stop) {
	case AUTO_MONITOR_TOGGLE:
		if (is_monitoring) {
			stop_automonitor(bridge_channel, peer_chan, features_cfg, stop_message);
		} else {
			start_automonitor(bridge_channel, peer_chan, features_cfg, start_message);
		}
		return 0;
	case AUTO_MONITOR_START:
		if (!is_monitoring) {
			start_automonitor(bridge_channel, peer_chan, features_cfg, start_message);
			return 0;
		}
		ast_verb(4, "AutoMonitor already recording call.\n");
		break;
	case AUTO_MONITOR_STOP:
		if (is_monitoring) {
			stop_automonitor(bridge_channel, peer_chan, features_cfg, stop_message);
			return 0;
		}
		ast_verb(4, "AutoMonitor already stopped on call.\n");
		break;
	}

	/*
	 * Fake start/stop to invoker so will think it did something but
	 * was already in that mode.
	 */
	if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}
	if (is_monitoring) {
		if (!ast_strlen_zero(start_message)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, start_message, NULL);
		}
	} else {
		if (!ast_strlen_zero(stop_message)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, stop_message, NULL);
		}
	}
	return 0;
}

static void stop_automixmonitor(struct ast_bridge_channel *bridge_channel, struct ast_channel *peer_chan, struct ast_features_general_config *features_cfg, const char *stop_message)
{
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "stop_automixmonitor(), bridge_channel:%s, peer_chan:%s, stop_message:%s, features_cfg:%s",
						(bridge_channel && bridge_channel->chan) ? ast_channel_name(bridge_channel->chan):"",
						peer_chan ? ast_channel_name(peer_chan):"",
						ast_strlen_zero(stop_message) ? "":stop_message,
						(features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) ? features_cfg->courtesytone:"");
#else
	ast_verb(4, "AutoMixMonitor used to stop recording call.\n");
#endif
	if (ast_stop_mixmonitor(peer_chan, NULL)) {
		ast_verb(4, "Failed to stop AutoMixMonitor for %s.\n", ast_channel_name(bridge_channel->chan));
		if (features_cfg && !(ast_strlen_zero(features_cfg->recordingfailsound))) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
		}
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (bridge_channel && bridge_channel->chan) {
		struct ast_channel *peer = ast_channel_real_peer(peer_chan);  /* added by zgmo for Bug176611 */
		struct ast_channel *bridge_chan_real_peer = ast_channel_real_peer(bridge_channel->chan);

		const char *from_ivr = pbx_builtin_getvar_helper(peer_chan, "FROM_IVR");

		if (peer) {
			const char *in_queue = pbx_builtin_getvar_helper(peer, "IN_QUEUE");

			ast_log(LOG_DEBUG, "Chan(%s) get real peer : %s",
				ast_channel_name(peer_chan), ast_channel_name(peer));

			if (in_queue && !strcmp(in_queue, "yes")) {
				const char *is_recording = pbx_builtin_getvar_helper(peer, "IS_RECORDING");

				if (!ast_strlen_zero(is_recording) && !strcmp(is_recording, "yes")) {
					ast_stop_mixmonitor(peer, NULL);

					ast_log(LOG_DEBUG, "Stop chan(%s)'s recording", ast_channel_name(peer));
				}
			}

			pbx_builtin_setvar_helper(peer, "REC_ENABLE", "");
			pbx_builtin_setvar_helper(peer, "LAST_REC_ENABLE", "");

			if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")) {
				pbx_builtin_setvar_helper(peer, "IVR_REC", "");
				pbx_builtin_setvar_helper(peer, "SAVED_IVR_REC", "");
				ast_log(LOG_DEBUG, "%s ivr_rec off", ast_channel_name(peer));
			}

			ast_indicate(peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);

			ast_channel_unref(peer);
		}

		pbx_builtin_setvar_helper(bridge_channel->chan, "REC_ENABLE", "");
		pbx_builtin_setvar_helper(bridge_channel->chan, "LAST_REC_ENABLE", "");

		pbx_builtin_setvar_helper(peer_chan, "REC_ENABLE", "");
		pbx_builtin_setvar_helper(peer_chan, "LAST_REC_ENABLE", "");

		if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")) {
			pbx_builtin_setvar_helper(peer_chan, "IVR_REC", "");
			pbx_builtin_setvar_helper(peer_chan, "SAVED_IVR_REC", "");
			ast_log(LOG_DEBUG, "%s ivr_rec off", ast_channel_name(peer_chan));
		}

		ast_indicate(bridge_channel->chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
		ast_indicate(peer_chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);

		if (bridge_chan_real_peer) {
			ast_log(LOG_DEBUG, "bridge_chan_real_peer:%s", ast_channel_name(bridge_chan_real_peer));

			pbx_builtin_setvar_helper(bridge_chan_real_peer, "REC_ENABLE", "");
			ast_indicate(bridge_chan_real_peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);

			ast_channel_unref(bridge_chan_real_peer);
		}
	}
#endif

	if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}

	if (!ast_strlen_zero(stop_message)) {
#ifdef GRANDSTREAM_NETWORKS
		const char *have_prompted = pbx_builtin_getvar_helper(bridge_channel->chan, "HAVE_PROMPTED");
		if (!ast_strlen_zero(have_prompted) && !strcmp(have_prompted, "yes")) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, stop_message, NULL);
			ast_bridge_channel_write_playfile(bridge_channel, NULL, stop_message, NULL);
			pbx_builtin_setvar_helper(bridge_channel->chan, "HAVE_PROMPTED", "");
			pbx_builtin_setvar_helper(peer_chan, "HAVE_PROMPTED", "");

			pbx_builtin_setvar_helper(bridge_channel->chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
			pbx_builtin_setvar_helper(peer_chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
		}
#else
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, stop_message, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, stop_message, NULL);
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ast_strlen_zero(stop_message)) {
		if (ast_dial_check_outside_can_stop_record(bridge_channel->chan)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, "dialog-stop-recording", NULL);

			pbx_builtin_setvar_helper(bridge_channel->chan, "HAVE_PROMPTED", "");
			pbx_builtin_setvar_helper(bridge_channel->chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
		}

		if (peer_chan && strcmp(ast_channel_name(peer_chan), ast_channel_name(bridge_channel->chan))
				&& ast_dial_check_outside_can_stop_record(peer_chan)) {
			ast_bridge_channel_write_playfile(bridge_channel, NULL, "dialog-stop-recording", NULL);

			pbx_builtin_setvar_helper(peer_chan, "HAVE_PROMPTED", "");
			pbx_builtin_setvar_helper(peer_chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
		}
	}
#endif
}

void replaceTimestamp(char *str) {
	int i = 0;
	char new_timestamp[20];
	char timestamp_str[32];
	char *third_to_last_dash = NULL;
	char *last_dash = strrchr(str, '-');
	
	if (!last_dash) {
		return;
	}
	for (char *p = last_dash; p != str && !third_to_last_dash; --p) {
		if (*(p - 1) == '-') {
			if (i == 1) {
				third_to_last_dash = p - 1;
			}
			i++;
		}
	}

	if (!third_to_last_dash) {
		return;
	}

	time_t now = time(NULL);
	struct tm *tm_info = localtime(&now);

	strftime(new_timestamp, sizeof(new_timestamp), "%Y%m%d-%H%M%S", tm_info);
	snprintf(timestamp_str, sizeof(timestamp_str), "%s-%ld.0", new_timestamp, (long)now);
	memcpy(third_to_last_dash - 8, timestamp_str, 28);
}

static void start_automixmonitor(struct ast_bridge_channel *bridge_channel, struct ast_channel *peer_chan, struct ast_features_general_config *features_cfg, const char *start_message)
{
	char *touch_filename;
	size_t len;
	char *uuid;
	size_t uuid_len;
	int x;
	enum set_touch_variables_res set_touch_res;
	char lastapp[2*AST_MAX_USER_FIELD] = {0};
	char action_type[2*AST_MAX_USER_FIELD] = {0};
	const char *vq_callback_caller = pbx_builtin_getvar_helper(bridge_channel->chan, "VQ_CALLBACK_CALLER");
	const char *vq_callback_callee = pbx_builtin_getvar_helper(bridge_channel->chan, "VQ_CALLBACK_CALLEE");
	const char *queuefilename = NULL;

	RAII_VAR(char *, touch_format, NULL, ast_free);
	RAII_VAR(char *, touch_monitor, NULL, ast_free);
	RAII_VAR(char *, touch_monitor_prefix, NULL, ast_free);

	ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "lastapp", lastapp, sizeof(lastapp));
	ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "action_type", action_type, sizeof(action_type));
	set_touch_res = set_touch_variables(bridge_channel->chan, 1, &touch_format,
		&touch_monitor, &touch_monitor_prefix);
	switch (set_touch_res) {
	case SET_TOUCH_SUCCESS:
		break;
	case SET_TOUCH_UNSET:
		set_touch_res = set_touch_variables(peer_chan, 1, &touch_format, &touch_monitor,
			&touch_monitor_prefix);
		if (set_touch_res == SET_TOUCH_ALLOC_FAILURE) {
			return;
		}
		break;
	case SET_TOUCH_ALLOC_FAILURE:
		return;
	}
	
	uuid_len = 64;
	uuid = ast_alloca(uuid_len);
	
	memset(uuid, 0, uuid_len);
	file_crypt_new_uuid(uuid, uuid_len);
	ast_log(LOG_DEBUG, "[start_automixmonitor] new uuid:%s\n", uuid);

#ifdef GRANDSTREAM_NETWORKS
	/*jklou*/
	if (!ast_strlen_zero(touch_monitor) && (vq_callback_caller == NULL || strcmp(vq_callback_caller, "yes")) && strstr(lastapp, "Queue") == NULL) {
		len = strlen(touch_monitor) + 60;
		touch_filename = ast_alloca(len);
		memset(touch_filename, 0, len);

		time_t timep;
		struct tm *p;
		time(&timep);
		p = localtime(&timep);

		snprintf(touch_filename, len, "%d-%02d/%s-%ld-%s.%s",
			1900 + p->tm_year, 1 + p->tm_mon,
			S_OR(touch_monitor_prefix, "auto"),
			(long) time(NULL),
			touch_monitor,
			S_OR(touch_format, "wav"));
	}
#else
	if (!ast_strlen_zero(touch_monitor)) {
		len = strlen(touch_monitor) + 50;
		touch_filename = ast_alloca(len);
		snprintf(touch_filename, len, "%s-%ld-%s.%s",
			S_OR(touch_monitor_prefix, "auto"),
			(long) time(NULL),
			touch_monitor,
			S_OR(touch_format, "wav"));
	}
#endif
	else {
#ifdef GRANDSTREAM_NETWORKS
		char *caller_chan_id = NULL;
		char *peer_chan_id = NULL;
		char exten[AST_MAX_EXTENSION] = {0};
		char peer_chan_exten[AST_MAX_EXTENSION] = {0};
		const char *is_callee = pbx_builtin_getvar_helper(bridge_channel->chan, "IS_CALLEE");
		int use_peer_exten = 0;
		const char *cf_reason = pbx_builtin_getvar_helper(bridge_channel->chan, "CF_REASON");
		const char *caller_num = pbx_builtin_getvar_helper(bridge_channel->chan, "CALLER_NUM");

		ast_log(LOG_DEBUG, "feature_chan:%s, peer_chan: %s",
			ast_channel_name(bridge_channel->chan), ast_channel_name(peer_chan));

		/*added by wxu@grandstream.cn for bug 43529*/
		if(strlen(ast_channel_caller_num(bridge_channel->chan)) > 0){
			const char *pickup_caller = pbx_builtin_getvar_helper(bridge_channel->chan, "PICKUP_CALLER");

			ast_log(LOG_DEBUG, "%s caller_num:%s", ast_channel_name(bridge_channel->chan), ast_channel_caller_num(bridge_channel->chan));
			if (!ast_strlen_zero(pickup_caller) && !strcmp(pickup_caller, "yes")) {
				if (strstr(ast_channel_name(bridge_channel->chan), "DAHDI/")) {
					struct ast_party_caller *party_caller = ast_channel_caller(bridge_channel->chan);

					if (party_caller->ani.number.valid && !ast_strlen_zero(party_caller->ani.number.str)) {
						caller_chan_id = ast_strdupa(party_caller->ani.number.str);
					} else {
						ast_log(LOG_WARNING,
							"%s is a pickup-caller and should be callee, but cannot retrive ANI NUMBER",
							ast_channel_name(bridge_channel->chan));
						caller_chan_id = "unknown";
					}
				} else {
					if (sscanf(ast_channel_name(bridge_channel->chan), "%*[^/]/%[^-]", exten)) {
						caller_chan_id = exten;
					}
				}
			} else {
				caller_chan_id = ast_strdupa(ast_channel_caller_num(bridge_channel->chan));
			}
		} else {
			if(!strcmp(ast_channel_tech(bridge_channel->chan)->type,"IAX2")){
				if(sscanf(ast_channel_name(bridge_channel->chan),"IAX2/%[^-]-%*s",exten) == 1){
					caller_chan_id = ast_strdupa(exten);
				}
			} else if ((!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "14"))
				&& strstr(ast_channel_name(bridge_channel->chan), "Local/")
				&& strstr(ast_channel_name(bridge_channel->chan), ";2")
				&& !ast_strlen_zero(caller_num)) {
				caller_chan_id = (char *)caller_num;
			} else {
				if(sscanf(ast_channel_name(bridge_channel->chan), "%*[^/]/%[^@]", exten)) {
					caller_chan_id = exten;
				}
			}
		}
		/*end*/

		/* added by zgmo for bug 152711 */
		if (strstr(ast_channel_name(peer_chan), "Local/")) {
			const char *from_cb = pbx_builtin_getvar_helper(bridge_channel->chan, "FROM_CALLBACK");
			const char *do_forward = pbx_builtin_getvar_helper(peer_chan, "DO_FORWARD");
			const char *aa_user = pbx_builtin_getvar_helper(peer_chan, "AAUSER");
			const char *from_park = pbx_builtin_getvar_helper(peer_chan, "FROM_PARK");
			const char *caller_do_park = pbx_builtin_getvar_helper(peer_chan, "CALLER_DO_PARK");
			const char *orig_callee = pbx_builtin_getvar_helper(peer_chan, "ORIG_CALLEE");
			const char *in_followme = pbx_builtin_getvar_helper(peer_chan, "IN_FOLLOWME");
			const char *fm_cidnum = pbx_builtin_getvar_helper(peer_chan, "FMCIDNUM");
			const char *page_intercom = pbx_builtin_getvar_helper(peer_chan, "PAGE_INTERCOM");

			use_peer_exten = 1;

			if ((!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "14"))
				|| (!ast_strlen_zero(from_cb) && !strcmp(from_cb, "yes"))) {
				if (!ast_strlen_zero(caller_num)) {
					strcpy(peer_chan_exten, caller_num);
					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				} else if (strstr(ast_channel_name(peer_chan), ";2")
					&& !ast_strlen_zero(from_park) && !strcmp(from_park, "yes")
					&& !ast_strlen_zero(caller_do_park) && !strcmp(caller_do_park, "yes")
					&& !ast_strlen_zero(orig_callee)) {
					strcpy(peer_chan_exten, orig_callee);
					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				} else {
					sscanf(ast_channel_name(peer_chan), "%*[^/]/%[^@]", peer_chan_exten);
					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				}
			} else if (!ast_strlen_zero(do_forward) && !ast_strlen_zero(aa_user)
				&& strstr(ast_channel_name(peer_chan), ";2")) {
				strcpy(peer_chan_exten, aa_user);
				ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
			} else if (!ast_strlen_zero(in_followme) && !strcmp(in_followme, "yes")
				&& !ast_strlen_zero(fm_cidnum) && strstr(ast_channel_name(peer_chan), ";2")) {
				strcpy(peer_chan_exten, fm_cidnum);
				ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
			} else if (!ast_strlen_zero(page_intercom) && !strcmp(page_intercom, "yes")) {
				use_peer_exten = 0;
				ast_log(LOG_DEBUG, "In this case, clear use_peer_exten");
			} else {
				sscanf(ast_channel_name(peer_chan), "%*[^/]/%[^@]", peer_chan_exten);
				ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
			}
		} else {
			if (strstr(ast_channel_name(peer_chan), "/trunk_")) {
				/* modified by zgmo for bug 145695 */
				const char *ringboth = pbx_builtin_getvar_helper(bridge_channel->chan, "RINGBOTH");
				
				const char *orig_callee = pbx_builtin_getvar_helper(peer_chan, "ORIG_CALLEE");
				const char *is_callee = pbx_builtin_getvar_helper(peer_chan, "IS_CALLEE");
				const char *org_is_callee = pbx_builtin_getvar_helper(peer_chan, "ORG_IS_CALLEE");
				const char *external_number = pbx_builtin_getvar_helper(peer_chan, "EXTERNAL_NUMBER");

				if (!ast_strlen_zero(ringboth) && !strcmp(ringboth, "yes")) {
					if (!ast_strlen_zero(orig_callee)
						&& !ast_strlen_zero(org_is_callee) && !strcmp(org_is_callee, "yes")) {
						strcpy(peer_chan_exten, orig_callee);
						use_peer_exten = 1;

						ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
					}
					else if (!ast_strlen_zero(external_number)
						&& !ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
						strcpy(peer_chan_exten, external_number);
						use_peer_exten = 1;

						ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", external_number);
					}
				} else if (!ast_strlen_zero(orig_callee) 
					&& ((!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes"))
						|| (!ast_strlen_zero(org_is_callee) && !strcmp(org_is_callee, "yes")))) {
					strcpy(peer_chan_exten, orig_callee);
					use_peer_exten = 1;

					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				}
			} else {
				sscanf(ast_channel_name(peer_chan), "%*[^/]/%[^-]", peer_chan_exten);
			}
		}

		/* modified by zgmo for bug 152711 */
		if(strstr(ast_channel_caller(peer_chan)->id.number.str, "Callback")) {
			use_peer_exten = 1;
			ast_log(LOG_DEBUG, "In this case, mark use_peer_exten");
		}

		/* modified by jklou */
		/* modified by zgmo for bug 145695 */
		if (use_peer_exten) {
			peer_chan_id = peer_chan_exten;
		} else {
			peer_chan_id = ast_strdupa(S_COR(ast_channel_caller(peer_chan)->ani.number.valid,
				ast_channel_caller(peer_chan)->ani.number.str,
				ast_channel_caller(peer_chan)->id.number.str));

			ast_log(LOG_DEBUG, "use peer_exten from party-caller :%s", peer_chan_id ? : "none");
		}

		len = strlen(caller_chan_id) + strlen(peer_chan_id) + 128;
		touch_filename = ast_alloca(len);
		memset(touch_filename, 0, len);

		time_t timep;
		struct tm *p;
		time(&timep);
		p = localtime(&timep);
		const char *merge_audio_recording_files = pbx_builtin_getvar_helper(NULL, "MERGE_AUDIO_RECORDING_FILES");
		/* modified by zgmo for bug153684 */
		if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
			if (strstr(lastapp, "Queue") != NULL) {
				char tmp_filename[256] = {0};
				const char *monitor_filename = pbx_builtin_getvar_helper(peer_chan, "MONITOR_FILENAME");

				if (monitor_filename != NULL) {
					char *tmp_file = ast_strdupa(monitor_filename);
					if (ast_strlen_zero(merge_audio_recording_files) || !strcmp(merge_audio_recording_files, "no")) {
						replaceTimestamp(tmp_file);
					}
					snprintf(touch_filename, len, "%s.wav", tmp_file + strlen("/app/asterisk/var/spool/asterisk/queue/"));
					snprintf(tmp_filename, sizeof(tmp_filename), "%s.wav", tmp_file);
					queuefilename = ast_strdupa(tmp_filename);
					ast_log(LOG_DEBUG, "touch_filename:%s, queuefilename:%s", touch_filename, queuefilename);
				}
			}
			else if (strstr(lastapp, "Dial") && ((!ast_strlen_zero(vq_callback_caller) && !strcmp(vq_callback_caller, "yes")) || 
												  (!ast_strlen_zero(vq_callback_callee) && !strcmp(vq_callback_callee, "yes")))) {
				char tmp_filename[256] = {0};
				struct ast_channel *get_chan = bridge_channel->chan;
				struct ast_channel *set_chan = peer_chan;

				if (strstr(ast_channel_name(get_chan), "Local/") == NULL) {
					get_chan = peer_chan;
					set_chan = bridge_channel->chan;
				}
				const char *monitor_filename = pbx_builtin_getvar_helper(get_chan, "MONITOR_FILENAME");
				if (monitor_filename != NULL) {
					char *tmp_file = ast_strdupa(monitor_filename);
					if (ast_strlen_zero(merge_audio_recording_files) || !strcmp(merge_audio_recording_files, "no")) {
						replaceTimestamp(tmp_file);
					}
					snprintf(touch_filename, len, "%s.wav", tmp_file + strlen("/app/asterisk/var/spool/asterisk/queue/"));
					snprintf(tmp_filename, sizeof(tmp_filename), "%s.wav", tmp_file);
					queuefilename = ast_strdupa(tmp_filename);
					pbx_builtin_setvar_helper(ast_channel_get_mychan(set_chan), "MONITOR_FILENAME", monitor_filename);
					pbx_builtin_setvar_helper(ast_channel_get_mychan(set_chan), "VQ_CALLBACK_CALLER", vq_callback_caller);
					ast_log(LOG_DEBUG, "touch_filename:%s, queuefilename:%s", touch_filename, queuefilename);
				}
			}
			else {
				snprintf(touch_filename, len, "%d-%02d/%s-%ld-%s-%s.%s",
					1900 + p->tm_year, 1 + p->tm_mon,
					S_OR(touch_monitor_prefix, "auto"),
					(long) time(NULL),
					peer_chan_id,
					caller_chan_id,
					S_OR(touch_format, "wav"));
			}
		} else {
			if (strstr(lastapp, "Queue") != NULL) {
				char tmp_filename[256] = {0};
				const char *monitor_filename = pbx_builtin_getvar_helper(peer_chan, "MONITOR_FILENAME");

				if (monitor_filename != NULL) {
					char *tmp_file = ast_strdupa(monitor_filename);
					if (ast_strlen_zero(merge_audio_recording_files) || !strcmp(merge_audio_recording_files, "no")) {
						replaceTimestamp(tmp_file);
					}
					snprintf(touch_filename, len, "%s.wav", tmp_file + strlen("/app/asterisk/var/spool/asterisk/queue/"));
					snprintf(tmp_filename, sizeof(tmp_filename), "%s.wav", tmp_file);
					queuefilename = ast_strdupa(tmp_filename);
					ast_log(LOG_DEBUG, "touch_filename:%s, queuefilename:%s", touch_filename, queuefilename);
				}
			}
			else if (strstr(lastapp, "Dial") && ((!ast_strlen_zero(vq_callback_caller) && !strcmp(vq_callback_caller, "yes")) || 
													(!ast_strlen_zero(vq_callback_callee) && !strcmp(vq_callback_callee, "yes")))) {
				char tmp_filename[256] = {0};
				struct ast_channel *get_chan = bridge_channel->chan;
				struct ast_channel *set_chan = peer_chan;

				if (strstr(ast_channel_name(get_chan), "Local/") == NULL) {
					get_chan = peer_chan;
					set_chan = bridge_channel->chan;
				}
				const char *monitor_filename = pbx_builtin_getvar_helper(get_chan, "MONITOR_FILENAME");
				if (monitor_filename != NULL) {
					char *tmp_file = ast_strdupa(monitor_filename);
					if (ast_strlen_zero(merge_audio_recording_files) || !strcmp(merge_audio_recording_files, "no")) {
						replaceTimestamp(tmp_file);
					}
					snprintf(touch_filename, len, "%s.wav", tmp_file + strlen("/app/asterisk/var/spool/asterisk/queue/"));
					snprintf(tmp_filename, sizeof(tmp_filename), "%s.wav", tmp_file);
					queuefilename = ast_strdupa(tmp_filename);
					pbx_builtin_setvar_helper(ast_channel_get_mychan(set_chan), "MONITOR_FILENAME", monitor_filename);
					pbx_builtin_setvar_helper(ast_channel_get_mychan(set_chan), "VQ_CALLBACK_CALLER", vq_callback_caller);
					ast_log(LOG_DEBUG, "touch_filename:%s, queuefilename:%s", touch_filename, queuefilename);
				}
			}  
			else if (!strcmp(action_type, "EMERGENCYCALL")) {
				snprintf(touch_filename, len, "%d-%02d/%s-%s-%s-%ld.%s",
					1900 + p->tm_year, 1 + p->tm_mon,
					S_OR(touch_monitor_prefix, "emergency"),
					peer_chan_id,
					caller_chan_id,
					(long) time(NULL),
					S_OR(touch_format, "wav"));
			}
			else {
				snprintf(touch_filename, len, "%d-%02d/%s-%ld-%s-%s.%s",
					1900 + p->tm_year, 1 + p->tm_mon,
					S_OR(touch_monitor_prefix, "auto"),
					(long) time(NULL),
					caller_chan_id,
					peer_chan_id,
					S_OR(touch_format, "wav"));
			}
		}
		/* end jklou**/
#else
		char *caller_chan_id;
		char *peer_chan_id;

		caller_chan_id = ast_strdupa(S_COR(ast_channel_caller(bridge_channel->chan)->id.number.valid,
			ast_channel_caller(bridge_channel->chan)->id.number.str, ast_channel_name(bridge_channel->chan)));
		peer_chan_id = ast_strdupa(S_COR(ast_channel_caller(peer_chan)->id.number.valid,
			ast_channel_caller(peer_chan)->id.number.str, ast_channel_name(peer_chan)));
		len = strlen(caller_chan_id) + strlen(peer_chan_id) + 50;
		touch_filename = ast_alloca(len);
		snprintf(touch_filename, len, "%s-%ld-%s-%s.%s",
			S_OR(touch_monitor_prefix, "auto"),
			(long) time(NULL),
			caller_chan_id,
			peer_chan_id,
			S_OR(touch_format, "wav"));
#endif
	}
#ifdef GRANDSTREAM_NETWORKS
	 /*added by wxu@grandstream.cn for bug 41694*/
	{
		char tmp[2*AST_MAX_USER_FIELD] = {0};
		char val[2*AST_MAX_USER_FIELD] = {0};
		char record_files[2*AST_MAX_USER_FIELD] = {0};
		int len = sizeof(tmp)-1;

		const char *merge_audio_recording_files = pbx_builtin_getvar_helper(NULL, "MERGE_AUDIO_RECORDING_FILES");
		char record_files_desc[2*AST_MAX_USER_FIELD] = {0};
		char record_files_uuid[2*AST_MAX_USER_FIELD] = {0};

		strncat(tmp, touch_filename, len);
		len -= strlen(touch_filename);
		strncat(tmp, "@", len);

		ast_log(LOG_VERBOSE, "lastapp = %s, new record:%s", lastapp, tmp);

		if (strcmp(lastapp, "ParkedCall") != 0) {
			memset(val, 0, sizeof(val));
			ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "RECORD_FILES", val, sizeof(val));
			ast_log(LOG_VERBOSE, "[start_automixmonitor] Chan:%s, old-cdr-record:%s", ast_channel_name(bridge_channel->chan), val);

			if (!strstr(val, tmp)) {
				snprintf(record_files,sizeof(record_files)-1,"%s%s",val,tmp);
				ast_cdr_setvar(ast_channel_name(bridge_channel->chan), "RECORD_FILES", record_files);
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr RECORD_FILES:%s", ast_channel_name(bridge_channel->chan), record_files);
			}
			
			{
				memset(val, 0, sizeof(val));
				ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "record_files_uuid", val, sizeof(val));
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Chan:%s, old-cdr-record-uuid:%s", ast_channel_name(bridge_channel->chan), val);
				
				snprintf(record_files_uuid, sizeof(record_files_uuid), "%s%s@", val, uuid);
				ast_cdr_setvar(ast_channel_name(bridge_channel->chan), "record_files_uuid", record_files_uuid);
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr record_files_uuid:%s", ast_channel_name(bridge_channel->chan), record_files_uuid);
			}

			if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
				memset(val, 0, sizeof(val));
				
				ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "record_files_desc", val, sizeof(val));
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Chan:%s, old-cdr-record-desc:%s", ast_channel_name(bridge_channel->chan), val);

				snprintf(record_files_desc, sizeof(record_files_desc), "%sduration=0,size=0B,start-time=0@", val);
				ast_cdr_setvar(ast_channel_name(bridge_channel->chan), "record_files_desc", record_files_desc);
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr record_files_desc:%s", ast_channel_name(bridge_channel->chan), record_files_desc);
			}
		}

		if(peer_chan) {
			memset(record_files,0,sizeof(record_files));
			memset(val,0,sizeof(val));

			ast_cdr_getvar(ast_channel_name(peer_chan), "RECORD_FILES", val, sizeof(val));
			ast_log(LOG_VERBOSE, "[start_automixmonitor] Chan:%s, old-cdr-record:%s", ast_channel_name(peer_chan), val);

			/* modified by zgmo for Bug181204 */
			if (!strstr(val, tmp)) {
				snprintf(record_files,sizeof(record_files)-1,"%s%s",val,tmp);
				ast_cdr_setvar(ast_channel_name(peer_chan), "RECORD_FILES", record_files);
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr RECORD_FILES:%s", ast_channel_name(peer_chan), record_files);
			}
			/* modify end */
			
			{
				memset(val, 0, sizeof(val));
				ast_cdr_getvar(ast_channel_name(peer_chan), "record_files_uuid", val, sizeof(val));
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Chan:%s, old-cdr-record-uuid:%s", ast_channel_name(peer_chan), val);
				
				snprintf(record_files_uuid, sizeof(record_files_uuid), "%s%s@", val, uuid);
				ast_cdr_setvar(ast_channel_name(peer_chan), "record_files_uuid", record_files_uuid);
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr record_files_uuid:%s", ast_channel_name(peer_chan), record_files_uuid);
			}

			if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
				memset(val, 0, sizeof(val));
				
				ast_cdr_getvar(ast_channel_name(peer_chan), "record_files_desc", val, sizeof(val));
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Chan:%s, old-cdr-record-desc:%s", ast_channel_name(peer_chan), val);

				snprintf(record_files_desc, sizeof(record_files_desc), "%sduration=0,size=0B,start-time=0@", val);
				ast_cdr_setvar(ast_channel_name(peer_chan), "record_files_desc", record_files_desc);
				ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr record_files_desc:%s", ast_channel_name(peer_chan), record_files_desc);
			}

			/*
			 *  added by zgmo for Bug177441
			 */
			if (strstr(ast_channel_name(peer_chan), "Local/")
				&& strstr(ast_channel_name(peer_chan), ";1")
				&& !ast_strlen_zero(record_files)) {
				struct ast_channel *local_peer = ast_local_get_peer(peer_chan);
				if (local_peer) {
					ast_cdr_setvar(ast_channel_name(local_peer), "RECORD_FILES", record_files);
					ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr RECORD_FILES:%s", ast_channel_name(local_peer), record_files);
					ast_channel_unref(local_peer);
				}
			} else if (strstr(ast_channel_name(peer_chan), "Local/")
						&& strstr(ast_channel_name(peer_chan), ";2")
						&& !ast_strlen_zero(record_files)) {
				const char *do_forward = pbx_builtin_getvar_helper(peer_chan, "DO_FORWARD");
				const char *aa_user = pbx_builtin_getvar_helper(peer_chan, "AAUSER");
				const char *page_caller = pbx_builtin_getvar_helper(peer_chan, "PAGING_INTERCOM_CALLER_CHAN_NAME");

				ast_log(LOG_DEBUG, "aa_user:%s", ast_strlen_zero(aa_user) ? "none":aa_user);
				ast_log(LOG_DEBUG, "do_forward:%s", ast_strlen_zero(do_forward) ? "none":do_forward);
				ast_log(LOG_DEBUG, "page_caller:%s", ast_strlen_zero(page_caller) ? "none":page_caller);

				if (!ast_strlen_zero(aa_user) && !ast_strlen_zero(do_forward)) {
					const char *initial_chan = pbx_builtin_getvar_helper(peer_chan, "FIRST_CALLER_CHANNEL_NAME");

					if (!ast_strlen_zero(initial_chan)) {
						ast_cdr_setvar(initial_chan, "RECORD_FILES", record_files);
						ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr RECORD_FILES:%s", initial_chan, record_files);
					}
				} else if (!ast_strlen_zero(page_caller)) {
					ast_cdr_setvar(page_caller, "RECORD_FILES", record_files);
					ast_log(LOG_VERBOSE, "[start_automixmonitor] Set %s's cdr RECORD_FILES:%s", page_caller, record_files);
				}
			}
		}
	}
	/* end */
#endif

#ifdef GRANDSTREAM_NETWORKS
	for (x = 8; x < strlen(touch_filename); x++) {
		if (touch_filename[x] == '/') {
			touch_filename[x] = '-';
		}
	}

	pbx_builtin_setvar_helper(peer_chan, 
			(!strcmp(action_type, "EMERGENCYCALL")) ? "EMERGENCYFILENAME" : "MIXFILENAME", 
			touch_filename);
#else
	for (x = 0; x < strlen(touch_filename); x++) {
		if (touch_filename[x] == '/') {
			touch_filename[x] = '-';
		}
	}
#endif

	ast_verb(4, "AutoMixMonitor used to record call. Filename: %s\n", touch_filename);

	if ( strstr(lastapp, "Queue") != NULL || (strstr(lastapp, "Dial") && ((!ast_strlen_zero(vq_callback_caller) && !strcmp(vq_callback_caller, "yes")) || 
								 										  (!ast_strlen_zero(vq_callback_callee) && !strcmp(vq_callback_callee, "yes")))) ) {
		if (queuefilename != NULL && strlen(queuefilename) > 0) {
			if (ast_start_mixmonitor(peer_chan, queuefilename, uuid, "b")) {
				ast_verb(4, "AutoMixMonitor feature was tried by '%s' but queue record failed to start.\n",
					ast_channel_name(bridge_channel->chan));

				if (features_cfg && !ast_strlen_zero(features_cfg->recordingfailsound)) {
					ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
				}
			}
		}
	}
	else if (!strcmp(action_type, "EMERGENCYCALL")) {
		pbx_builtin_setvar_helper(bridge_channel->chan, "EMERGENCYCALL", "yes");
		pbx_builtin_setvar_helper(peer_chan, "EMERGENCYCALL", "yes");
		if (ast_start_mixmonitor(peer_chan, touch_filename, uuid, "b")) {
			ast_verb(4, "EMERGENCY-CALL AutoMixMonitor feature was tried by '%s' but MixMonitor failed to start.\n", 
				ast_channel_name(bridge_channel->chan));

			if (features_cfg && !ast_strlen_zero(features_cfg->recordingfailsound)) {
				ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
			}
			return;
		}
	}
	else if (ast_start_mixmonitor(peer_chan, touch_filename, uuid, "b")) {
		ast_verb(4, "AutoMixMonitor feature was tried by '%s' but MixMonitor failed to start.\n",
			ast_channel_name(bridge_channel->chan));

		if (features_cfg && !ast_strlen_zero(features_cfg->recordingfailsound)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
		}
		return;
	}

	if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}

	if (!ast_strlen_zero(start_message)) {

		const char *record_forbidden = pbx_builtin_getvar_helper(NULL, "RECORD_FORBIDDEN");
		if (ast_strlen_zero(record_forbidden)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, start_message, NULL);
			ast_bridge_channel_write_playfile(bridge_channel, NULL, start_message, NULL);
		}

		pbx_builtin_setvar_helper(bridge_channel->chan, "HAVE_PROMPTED", "yes");
		pbx_builtin_setvar_helper(peer_chan, "HAVE_PROMPTED", "yes");

		if (strstr(ast_channel_name(peer_chan), "Local/")
			&& strstr(ast_channel_name(peer_chan), ";1")) {
			struct ast_channel *real_peer_chan = ast_channel_real_peer(peer_chan);
			if (real_peer_chan) {
				ast_log(LOG_DEBUG, "%s's real_peer_chan:%s",
						ast_channel_name(peer_chan), ast_channel_name(real_peer_chan));
				pbx_builtin_setvar_helper(real_peer_chan, "HAVE_PROMPTED", "yes");

				ast_channel_unref(real_peer_chan);
			}
		}
	}

	pbx_builtin_setvar_helper(bridge_channel->chan, "TOUCH_MIXMONITOR_OUTPUT", touch_filename);
	pbx_builtin_setvar_helper(peer_chan, "TOUCH_MIXMONITOR_OUTPUT", touch_filename);

#ifdef GRANDSTREAM_NETWORKS
	ast_bridge_channel_write_playfile(bridge_channel, NULL, "prompt_outside_stop_record", NULL);

	/* added by zgmo for recording indication on sdp */
	pbx_builtin_setvar_helper(bridge_channel->chan, "REC_ENABLE", "yes");
	pbx_builtin_setvar_helper(peer_chan, "REC_ENABLE", "yes");

	pbx_builtin_setvar_helper(bridge_channel->chan, "LAST_REC_ENABLE", "yes");
	pbx_builtin_setvar_helper(peer_chan, "LAST_REC_ENABLE", "yes");

	{
		const char *from_ivr = pbx_builtin_getvar_helper(bridge_channel->chan, "FROM_IVR");
		if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")) {
			pbx_builtin_setvar_helper(bridge_channel->chan, "IVR_REC", "yes");
			pbx_builtin_setvar_helper(bridge_channel->chan, "SAVED_IVR_REC", "yes");
			pbx_builtin_setvar_helper(peer_chan, "IVR_REC", "yes");
			pbx_builtin_setvar_helper(peer_chan, "SAVED_IVR_REC", "yes");

			ast_log(LOG_DEBUG, "%s %s ivr_rec on",
				ast_channel_name(bridge_channel->chan), ast_channel_name(peer_chan));
		}
	}


	const char *record_forbidden = pbx_builtin_getvar_helper(NULL, "RECORD_FORBIDDEN");

	if (peer_chan && strstr(ast_channel_name(peer_chan), "Local/")
		&& strstr(ast_channel_name(peer_chan), ";2")) {
		struct ast_channel *bridge_chan_real_peer = ast_channel_real_peer(bridge_channel->chan);
		if (bridge_chan_real_peer) {
			pbx_builtin_setvar_helper(bridge_chan_real_peer, "REC_ENABLE", "yes");
			pbx_builtin_setvar_helper(bridge_chan_real_peer, "LAST_REC_ENABLE", "yes");
			if (ast_strlen_zero(record_forbidden)) 
				ast_indicate(bridge_chan_real_peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);

			ast_channel_unref(bridge_chan_real_peer);
		}
	}

	if (ast_strlen_zero(record_forbidden))
	{
		ast_indicate(bridge_channel->chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
		ast_indicate(peer_chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
	}
	/* end */
#endif
}

#ifdef GRANDSTREAM_NETWORKS

/// used by record internally
enum rec_t {
	REC_OK,
	REC_CHAN_NULL,
	REC_FILE_UNAVAILABLE,
	REC_FILE_INFO_UNAVAILABLE,
	REC_OPERATION_FAIL
};
typedef enum rec_t rec_t;

static int update_cdr_rec_desc_pause(struct ast_channel *chan)
{
	char old_desc[4096] = {0}, old_desc_dump[4096] = {0}, old_desc_last[1024] = {0};
	char new_desc[4096] = {0};

	char *sep = "@", *token = NULL, *sub = NULL;

	time_t duration_old = 0, duration_new = 0;
	int file_size_old = 0;
	time_t startUtc = 0, resumeUtc = 0;
	
	int ret = 0;
	
	resumeUtc = ast_channel_get_recordresume_utc(chan);
	if (!resumeUtc) {
		resumeUtc = ast_channel_get_recordstart_utc(chan);
		if (!resumeUtc) {
			ast_log(LOG_WARNING ,"[update_record_desc_for_cdr] neither started nor resumed ? ");
			ret = -1; goto END;
		}
	}
	duration_new = ast_now_utc_timestamp() - resumeUtc;

	ast_cdr_newest_getvar(ast_channel_name(chan), "record_files_desc", old_desc, sizeof(old_desc));
	if (strlen(old_desc)) {

		snprintf(old_desc_dump, 
				sizeof(old_desc_dump), 
				"%s", 
				old_desc);
		
    	for (token = strtok_r(old_desc_dump, sep, &sub); token != NULL; token = strtok_r(NULL, sep, &sub)) {
        	snprintf(old_desc_last, sizeof(old_desc_last), "%s", token);
    	}
		old_desc[strlen(old_desc) - strlen(old_desc_last) - 1] = '\0';
		
		sscanf(old_desc_last, "duration=%ld,size=%dB,start-time=%ld", &duration_old, &file_size_old, &startUtc);
		snprintf(new_desc, 
				sizeof(new_desc), 
				"%s", 
				old_desc);
		snprintf(new_desc + strlen(new_desc), 
				sizeof(new_desc) - strlen(new_desc), 
				"duration=%ld,size=%dB,start-time=%ld@", 
				duration_new + duration_old, 
				file_size_old, 
				startUtc);
	} else {

		snprintf(new_desc,  
				sizeof(new_desc), 
				"duration=%ld,size=%dB,start-time=%ld@", 
				duration_new, 
				file_size_old, 
				startUtc);
	}

	ast_log(LOG_DEBUG, "[rec pause/resume] pause curUtc:%ld startUtc:%ld resumeUtc:%ld", 
				ast_now_utc_timestamp(), 
				startUtc, 
				resumeUtc);
	ret = ast_cdr_newest_setvar(ast_channel_name(chan), "record_files_desc", new_desc);
	ast_log(LOG_DEBUG, "[rec pause/resume] channel:%s old_desc:%s -> new_desc:%s %s", 
				ast_channel_name(chan), 
				strlen(old_desc) ? old_desc : "",
				strlen(new_desc) ? new_desc : "",
				!ret ? "succeed" : "failed");

END:

	return ret;
}

static void pause_automixmonitor(struct ast_bridge_channel *bridge_channel,
		struct ast_channel *peer_chan, 
		struct ast_features_general_config *features_cfg, 
		const char *stop_message)
{
	ast_log(LOG_DEBUG, "[pause_automixmonitor] bridge_channel:%s, peer_chan:%s, stop_message:%s, features_cfg:%s",
			(bridge_channel && bridge_channel->chan) ? ast_channel_name(bridge_channel->chan) : "",
			peer_chan ? ast_channel_name(peer_chan) : "",
            ast_strlen_zero(stop_message) ? "" : stop_message,
            (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) ? features_cfg->courtesytone : "");
			
	if (!peer_chan) {
		ast_log(LOG_DEBUG, "[pause_automixmonitor] null peer chan");
        return;
	}

    if (ast_pause_mixmonitor(peer_chan, NULL)) {
        ast_verb(4, "[pause_automixmonitor] failed to pause record for %s.\n", ast_channel_name(bridge_channel->chan));
        
		if (features_cfg && !(ast_strlen_zero(features_cfg->recordingfailsound))) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
        }
        return;
    }

	if (bridge_channel && bridge_channel->chan) {
		
		struct ast_channel *real_peer = ast_channel_real_peer(peer_chan); 
        struct ast_channel *bridge_chan_real_peer = ast_channel_real_peer(bridge_channel->chan);

        if (real_peer) {
			
			ast_log(LOG_DEBUG, "[pause_automixmonitor] chan(%s) get real peer : %s", ast_channel_name(peer_chan), ast_channel_name(real_peer));

            pbx_builtin_setvar_helper(real_peer, "REC_ENABLE", "");
            pbx_builtin_setvar_helper(real_peer, "LAST_REC_ENABLE", "");

        	ast_indicate(real_peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
			
			ast_channel_unref(real_peer);
		}

        if (bridge_chan_real_peer) {

			ast_log(LOG_DEBUG, "[pause_automixmonitor] bridge_chan_real_peer:%s", ast_channel_name(bridge_chan_real_peer));
			
			pbx_builtin_setvar_helper(bridge_chan_real_peer, "REC_ENABLE", "");
            ast_indicate(bridge_chan_real_peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);

            ast_channel_unref(bridge_chan_real_peer);
        }
    }
	
	/*if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
    }*/

    if (!ast_strlen_zero(stop_message)) {
		
		const char *have_prompted = pbx_builtin_getvar_helper(bridge_channel->chan, "HAVE_PROMPTED");
		
		if (!ast_strlen_zero(have_prompted) && 
		    !strcmp(have_prompted, "yes")) {
				
				ast_bridge_channel_queue_playfile(bridge_channel, NULL, stop_message, NULL);
                ast_bridge_channel_write_playfile(bridge_channel, NULL, stop_message, NULL);
                pbx_builtin_setvar_helper(bridge_channel->chan, "HAVE_PROMPTED", "");
                pbx_builtin_setvar_helper(peer_chan, "HAVE_PROMPTED", "");

                pbx_builtin_setvar_helper(bridge_channel->chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
                pbx_builtin_setvar_helper(peer_chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
            }
			
	} else {
		
		if (ast_dial_check_outside_can_stop_record(bridge_channel->chan)) {
				
				ast_bridge_channel_queue_playfile(bridge_channel, NULL, "dialog-stop-recording", NULL);
				
				pbx_builtin_setvar_helper(bridge_channel->chan, "HAVE_PROMPTED", "");
                pbx_builtin_setvar_helper(bridge_channel->chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
        }

        if (peer_chan && 
		    strcmp(ast_channel_name(peer_chan), ast_channel_name(bridge_channel->chan)) && 
			ast_dial_check_outside_can_stop_record(peer_chan)) {
				
				ast_bridge_channel_write_playfile(bridge_channel, NULL, "dialog-stop-recording", NULL);
				
				pbx_builtin_setvar_helper(peer_chan, "HAVE_PROMPTED", "");
				pbx_builtin_setvar_helper(peer_chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED", "");
			}
	}
	
	pbx_builtin_setvar_helper(peer_chan, "REC_ENABLE", "");
	pbx_builtin_setvar_helper(peer_chan, "LAST_REC_ENABLE", "");
	pbx_builtin_setvar_helper(peer_chan, "REC_PAUSED", "yes");
	pbx_builtin_setvar_helper(peer_chan, "MIXAUTORECORDING", NULL); 
	ast_indicate(peer_chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
	
	RAII_VAR(struct ast_channel *, tmp, NULL, ast_channel_cleanup);
	ast_bridge_channel_lock_bridge(bridge_channel);
	tmp = ast_bridge_peer_nolock(bridge_channel->bridge, peer_chan);
	ast_bridge_unlock(bridge_channel->bridge);

	if (tmp) {
		pbx_builtin_setvar_helper(tmp, "REC_ENABLE", "");
		pbx_builtin_setvar_helper(tmp, "LAST_REC_ENABLE", "");
		pbx_builtin_setvar_helper(tmp, "REC_PAUSED", "yes");
		pbx_builtin_setvar_helper(tmp, "MIXAUTORECORDING", NULL); 
		ast_indicate(tmp, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
	}

	ast_log(LOG_DEBUG, "[pause_automixmonitor] update channel:%s 's cdr desc", ast_channel_name(peer_chan));
	update_cdr_rec_desc_pause(peer_chan);
	if (tmp) {
		ast_log(LOG_DEBUG, "[pause_automixmonitor] update channel:%s 's cdr desc", ast_channel_name(tmp));
		update_cdr_rec_desc_pause(tmp);
	}

	return;
}

extern const char *ast_config_AST_SPOOL_DIR;
extern const char *ast_config_AST_MONITOR_DIR;

static int resume_automixmonitor(struct ast_bridge_channel *bridge_channel,
		struct ast_channel *peer_chan, 
		struct ast_features_general_config *features_cfg, 
		const char *start_message)
{
	int ret = REC_OK;

	const char *mix_filename = NULL;
	const char *monitor_filename = NULL;
	const char *record_forbidden = NULL;
	char actual_filename[256]={0};
	char action_type[2*AST_MAX_USER_FIELD] = {0};

	const char *record_file_suffix = ".wav";
	int skip_check = 0;

	RAII_VAR(struct ast_channel *, tmp, NULL, ast_channel_cleanup);
	ast_bridge_channel_lock_bridge(bridge_channel);
	tmp = ast_bridge_peer_nolock(bridge_channel->bridge, peer_chan);
	ast_bridge_unlock(bridge_channel->bridge);

	ast_cdr_getvar(ast_channel_name(peer_chan), "action_type", action_type, sizeof(action_type));
	if (!strlen(action_type)) {
		ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "action_type", action_type, sizeof(action_type));
	}
	
	mix_filename = pbx_builtin_getvar_helper(peer_chan, 
					(!ast_strlen_zero(action_type) && !strcmp(action_type, "EMERGENCYCALL")) ? "EMERGENCYFILENAME" : "MIXFILENAME");
	monitor_filename = pbx_builtin_getvar_helper(peer_chan, "MONITOR_FILENAME");
	if (ast_strlen_zero(mix_filename) && ast_strlen_zero(monitor_filename)) {
		ast_log(LOG_DEBUG, "[resume_automixmonitor] no FILENAME attr from peer_chan:%s", ast_channel_name(peer_chan));
		ret = REC_FILE_INFO_UNAVAILABLE;
		return ret;
	}

	if (monitor_filename) {
		snprintf(actual_filename, sizeof(actual_filename), "%s.wav", monitor_filename);
		ast_log(LOG_DEBUG, "[resume_automixmonitor] actual record file:%s", actual_filename);
	} else {
		if (!strcmp(action_type, "EMERGENCYCALL")) {
			snprintf(actual_filename, sizeof(actual_filename), "%s/emergency/%s", ast_config_AST_SPOOL_DIR, mix_filename);
		} else {
			snprintf(actual_filename, sizeof(actual_filename), "%s/%s", ast_config_AST_MONITOR_DIR, mix_filename);
		}

		if (!ast_strlen_zero(actual_filename) && 
			strlen(actual_filename) > strlen(record_file_suffix) &&
			strncmp(actual_filename+ (strlen(actual_filename) - strlen(record_file_suffix)), record_file_suffix, strlen(record_file_suffix))) {
			snprintf(actual_filename + strlen(actual_filename), sizeof(actual_filename) - strlen(actual_filename), "%s", record_file_suffix);
		}


		/*Start added by zyxia for Bug 303068 ,2024.6.13*/
		if(ast_media_role_get() == MEDIA_ROLE_MASTER
			&& !ast_strlen_zero(ast_channel_get_avs_address(peer_chan))
			&& strcmp("127.0.0.1",ast_channel_get_avs_address(peer_chan)) && !ast_strlen_zero(ast_channel_get_recordfile_slave_media(peer_chan))){
			skip_check = 1;
		}
		/*End added by zyxia for Bug 303068 ,2024.6.13*/
		/*Bug 303633,auto record enable and opertaion pause/resume record in caller channel,need get slave_record in remote channek*/
		else if(ast_media_role_get() == MEDIA_ROLE_MASTER
			&& tmp && !ast_strlen_zero(ast_channel_get_avs_address(tmp))
			&& strcmp("127.0.0.1",ast_channel_get_avs_address(tmp)) && !ast_strlen_zero(ast_channel_get_recordfile_slave_media(tmp))){
			skip_check = 1;
		}


		ast_log(LOG_DEBUG, "[resume_automixmonitor] actual record file:%s", actual_filename);
	}

	if (!skip_check && access(actual_filename, F_OK) != 0) {
		/** File not exists here since storage path
	   	   may change from A to B, like local to USB

	   	   Stop record first and do start again  
		*/
		ret = REC_FILE_UNAVAILABLE;
		return ret;
	}
	
	if (ast_resume_mixmonitor(peer_chan, actual_filename, "b")) {
		ast_verb(4, "AutoMixMonitor feature was tried by '%s' but MixMonitor failed to resume.\n",
		ast_channel_name(bridge_channel->chan));

		if (features_cfg && !ast_strlen_zero(features_cfg->recordingfailsound)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
		}
		ret = REC_OPERATION_FAIL;
		return ret;
	}

	/*if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
		ast_bridge_channel_write_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}*/
	
	if (!ast_strlen_zero(start_message)) {
		
		record_forbidden = pbx_builtin_getvar_helper(NULL, "RECORD_FORBIDDEN");
		if (ast_strlen_zero(record_forbidden)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, start_message, NULL);
			ast_bridge_channel_write_playfile(bridge_channel, NULL, start_message, NULL);
		}

		ast_bridge_channel_write_playfile(bridge_channel, NULL, "prompt_outside_stop_record", NULL);

		pbx_builtin_setvar_helper(peer_chan, "HAVE_PROMPTED", "yes");
		if (tmp) {
			pbx_builtin_setvar_helper(tmp, "HAVE_PROMPTED", "yes");
		}
	}
	
	pbx_builtin_setvar_helper(peer_chan, "REC_ENABLE", "yes");
	pbx_builtin_setvar_helper(peer_chan, "LAST_REC_ENABLE", "yes");
	pbx_builtin_setvar_helper(peer_chan, "REC_PAUSED", NULL);
	pbx_builtin_setvar_helper(peer_chan, "MIXAUTORECORDING", "ON");

	if (tmp) {
		pbx_builtin_setvar_helper(tmp, "REC_ENABLE", "yes");
		pbx_builtin_setvar_helper(tmp, "LAST_REC_ENABLE", "yes");
		pbx_builtin_setvar_helper(tmp, "REC_PAUSED", NULL);
		pbx_builtin_setvar_helper(tmp, "MIXAUTORECORDING", "ON");
	}
	
	if (ast_strlen_zero(record_forbidden)) {
		
		ast_indicate(peer_chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
		if (tmp) {
			ast_indicate(tmp, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
		}       
	}

	return ret;
}

#endif


static int feature_automixmonitor(struct ast_bridge_channel *bridge_channel, void *hook_pvt)
{
	static const char *mixmonitor_spy_type = "MixMonitor";
	const char *stop_message;
	const char *start_message;
	struct ast_bridge_features_automixmonitor *options = hook_pvt;
	enum ast_bridge_features_monitor start_stop = options ? options->start_stop : AUTO_MONITOR_TOGGLE;
	int is_monitoring;
#ifdef GRANDSTREAM_NETWORKS
	int skip_whitelist_check = 0;
	char *cid_num = NULL;
	/*BEGIN: Mod by weidu, for memory use after free*/
	char *recording_chan = NULL;
	const char *recording_chan_tmp = pbx_builtin_getvar_helper(bridge_channel->chan, "RECORDING_CHAN");
	if (recording_chan_tmp){
		recording_chan = ast_strdupa(recording_chan_tmp);
	}
	/*END: Mod by weidu, for memory use after free*/
	const char *local2_is_trunk = pbx_builtin_getvar_helper(bridge_channel->chan, "LOCAL2_IS_TRUNK");;

	RAII_VAR(struct ast_channel *, real_peer_chan, NULL, ast_channel_cleanup);
	RAII_VAR(struct ast_channel *, local_peer_chan, NULL, ast_channel_cleanup);

	char app_callee[2*AST_MAX_USER_FIELD] = {0};
	ast_cdr_getvar(ast_channel_name(bridge_channel->chan), "app_callee", app_callee, sizeof(app_callee));

	const char *merge_audio_recording_files = pbx_builtin_getvar_helper(NULL, "MERGE_AUDIO_RECORDING_FILES");
#endif

	RAII_VAR(struct ast_channel *, peer_chan, NULL, ast_channel_cleanup);
	RAII_VAR(struct ast_features_general_config *, features_cfg, NULL, ao2_cleanup);

	ast_channel_lock(bridge_channel->chan);
	features_cfg = ast_get_chan_features_general_config(bridge_channel->chan);
	ast_channel_unlock(bridge_channel->chan);
	ast_bridge_channel_lock_bridge(bridge_channel);
	peer_chan = ast_bridge_peer_nolock(bridge_channel->bridge, bridge_channel->chan);
	ast_bridge_unlock(bridge_channel->bridge);

#ifdef GRANDSTREAM_NETWORKS
	if (ast_channel_caller(bridge_channel->chan)->ani.number.valid &&
		!ast_strlen_zero(ast_channel_caller(bridge_channel->chan)->ani.number.str)
		&& strcmp(ast_channel_caller(bridge_channel->chan)->ani.number.str, "s")) {
		cid_num = ast_channel_caller(bridge_channel->chan)->ani.number.str;
	} else if (ast_channel_caller(bridge_channel->chan)->id.number.valid
		&&!ast_strlen_zero(ast_channel_caller(bridge_channel->chan)->id.number.str)) {
		cid_num = ast_channel_caller(bridge_channel->chan)->id.number.str;
	}
#endif

	if (!peer_chan) {
		ast_verb(4, "Cannot start AutoMixMonitor for %s - cannot determine peer in bridge.\n",
			ast_channel_name(bridge_channel->chan));

		if (features_cfg && !ast_strlen_zero(features_cfg->recordingfailsound)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->recordingfailsound, NULL);
		}
		return 0;
	}

	ast_channel_lock(bridge_channel->chan);
	start_message = pbx_builtin_getvar_helper(bridge_channel->chan, "TOUCH_MIXMONITOR_MESSAGE_START");
	start_message = ast_strdupa(S_OR(start_message, ""));
	
	stop_message = pbx_builtin_getvar_helper(bridge_channel->chan, "TOUCH_MIXMONITOR_MESSAGE_STOP");
	stop_message = ast_strdupa(S_OR(stop_message, ""));
	ast_channel_unlock(bridge_channel->chan);

	is_monitoring =
		(0 < ast_channel_audiohook_count_by_source(peer_chan, mixmonitor_spy_type, AST_AUDIOHOOK_TYPE_SPY));

#ifdef GRANDSTREAM_NETWORKS
	if(ast_opt_avs) {
		int chan_is_monitoring = 0;
		int peer_is_monitoring = 0;
		const char *mix_auto_recording = pbx_builtin_getvar_helper(bridge_channel->chan, "MIXAUTORECORDING");

		chan_is_monitoring = ast_channel_avs_is_recording(bridge_channel->chan);
		peer_is_monitoring = ast_channel_avs_is_recording(peer_chan);
		is_monitoring = (chan_is_monitoring || peer_is_monitoring);

		ast_verb(LOG_DEBUG, "peer_chan:%s. chan_is_monitoring:%d, peer_is_monitoring:%d",
			ast_channel_name(peer_chan), chan_is_monitoring, peer_is_monitoring);

		if (strstr(ast_channel_name(peer_chan), "Local/")) {
			int is_monitoring_real_peer = 0;
			real_peer_chan = ast_channel_real_peer(peer_chan);
			if (real_peer_chan) {
				is_monitoring_real_peer = ast_channel_avs_is_recording(real_peer_chan);
				ast_log(LOG_WARNING, "peer_chan:%s is LOCAL-CHAN, real_peer_chan: %s, is_monitoring_real_peer:%d",
					ast_channel_name(peer_chan), ast_channel_name(real_peer_chan), is_monitoring_real_peer);
				is_monitoring |= is_monitoring_real_peer;
			}

			local_peer_chan = ast_local_get_peer(peer_chan);
			if (local_peer_chan) {
				const char *in_queue = pbx_builtin_getvar_helper(local_peer_chan, "IN_QUEUE");
				const char *realcallernum = pbx_builtin_getvar_helper(local_peer_chan, "REALCALLERIDNUM");
				const char *forward_context = pbx_builtin_getvar_helper(local_peer_chan, "FORWARD_CONTEXT");
				const char *caller = ast_channel_caller_num(bridge_channel->chan);
				
				if (in_queue != NULL && realcallernum != NULL && forward_context != NULL && caller != NULL) {
					if (!strcmp(in_queue, "yes") && !strcmp(caller, realcallernum) && !strcmp(forward_context, "call-forward")) {
						ast_log(LOG_ERROR, "Callers are not allowed to turn queue recording on and off");
						return 0;
					}
				}
				is_monitoring_real_peer = ast_channel_avs_is_recording(local_peer_chan);
				ast_log(LOG_WARNING, "peer_chan:%s is LOCAL-CHAN, local_peer_chan: %s, is_monitoring_real_peer:%d",
					ast_channel_name(peer_chan), ast_channel_name(local_peer_chan), is_monitoring_real_peer);

				is_monitoring |= is_monitoring_real_peer;
			}
		}

		ast_verb(LOG_DEBUG, "mix_auto_recording:%s", ast_strlen_zero(mix_auto_recording) ? "":mix_auto_recording);
		is_monitoring |= (!ast_strlen_zero(mix_auto_recording) && !strcmp(mix_auto_recording, "ON")
			&& !ast_strlen_zero(recording_chan) && strcmp(recording_chan, ast_channel_name(bridge_channel->chan)));
	} else {
		ast_channel_lock(bridge_channel->chan);
			struct mixmonitor_ds *mixmonitor_ds = NULL;
			struct ast_datastore *datastore = NULL;
			AST_LIST_TRAVERSE(ast_channel_datastores(bridge_channel->chan), datastore, entry) {
				if (strstr(datastore->info->type, "mixmonitor")) {
					mixmonitor_ds = datastore->data;

					char *filename = NULL;
					if (mixmonitor_ds->fs) {
						filename = mixmonitor_ds->fs->filename;
					}

					if (filename) {
						is_monitoring = 1;
						ast_verb(4, "feature_automixmonitor: filename [%s]\n", filename);
					}
				}
			}
		ast_channel_unlock(bridge_channel->chan);
	}

	ast_verb(4, "feature_automixmonitor: start_stop:%d channel:%s is_monitoring:%d\n", 
						start_stop, 
						ast_channel_name(bridge_channel->chan), 
						is_monitoring);

	if (strstr(ast_channel_name(bridge_channel->chan), "PJSIP/trunk")
		|| strstr(ast_channel_name(bridge_channel->chan), "DAHDI/o")
		|| strstr(ast_channel_name(bridge_channel->chan), "IAX2/trunk_")
		|| strstr(ast_channel_name(bridge_channel->chan), "PJSIP/webrtc_trunk_")
		|| (!ast_strlen_zero(local2_is_trunk) && !strcmp(local2_is_trunk, "yes"))) {
		
		const char *outside_ext_stop_record = pbx_builtin_getvar_helper(NULL, "OUTSIDE_EXT_STOP_RECORD");
		if (ast_strlen_zero(outside_ext_stop_record) || strcmp(outside_ext_stop_record, "yes")) {
			ast_log(LOG_DEBUG, "Forbit trigger recording because outside_ext_stop_record not enable");
			return 0;
		}

		if (AUTO_MONITOR_START == start_stop 
			|| (AUTO_MONITOR_TOGGLE == start_stop && 0 == is_monitoring)) {
			ast_log(LOG_WARNING, "forbit outside CHAN to start recording");
			return 0;
		}

		skip_whitelist_check = 1;
	}

	if ( (!ast_strlen_zero(app_callee) && !strncmp(app_callee, "DISA[", strlen("DISA["))) ) {
		skip_whitelist_check = 1;
	}

	if (!skip_whitelist_check && !ast_strlen_zero(cid_num)) {
		if (!ast_channel_check_record_feature_whitelist(bridge_channel->chan, cid_num)) {
			return 0;
		}
	}
#endif

	switch (start_stop) {
	case AUTO_MONITOR_TOGGLE:
		if (is_monitoring) {
#ifdef GRANDSTREAM_NETWORKS
			if(ast_opt_avs) {
				if (ast_channel_avs_is_recording(peer_chan)) {
					if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] pause peer ->%s", ast_channel_name(peer_chan));
						pause_automixmonitor(bridge_channel, peer_chan, features_cfg, stop_message);
					} else {
						stop_automixmonitor(bridge_channel, peer_chan, features_cfg, stop_message);
					}
				}

				if (real_peer_chan && ast_channel_avs_is_recording(real_peer_chan)) {
					if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] pause real peer ->%s", ast_channel_name(real_peer_chan));
						pause_automixmonitor(bridge_channel, real_peer_chan, features_cfg, stop_message);
					} else {
						stop_automixmonitor(bridge_channel, real_peer_chan, features_cfg, stop_message);
					}
				}

				if (local_peer_chan && ast_channel_avs_is_recording(local_peer_chan)) {
					if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] pause local peer ->%s", ast_channel_name(local_peer_chan));
						pause_automixmonitor(bridge_channel, local_peer_chan, features_cfg, stop_message);
					} else {
						stop_automixmonitor(bridge_channel, local_peer_chan, features_cfg, stop_message);
					}
				}
			} else {
#endif
				stop_automixmonitor(bridge_channel, peer_chan, features_cfg, stop_message);
#ifdef GRANDSTREAM_NETWORKS
			}

			if(ast_opt_avs) {
				if (ast_channel_avs_is_recording(bridge_channel->chan)) {
					if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] pause self ->%s", ast_channel_name(bridge_channel->chan));
						pause_automixmonitor(bridge_channel, bridge_channel->chan, features_cfg, stop_message);
					} else {
						stop_automixmonitor(bridge_channel, bridge_channel->chan, features_cfg, stop_message);
					}
				}
			} else {
				stop_automixmonitor(bridge_channel, bridge_channel->chan, features_cfg, stop_message);
			}

			/*if(ast_opt_avs && !ast_strlen_zero(recording_chan)) {
				struct ast_channel *rec_chan = ast_channel_get_by_name(recording_chan);
				if (rec_chan) {
					if (ast_channel_avs_is_recording(rec_chan)) {
						stop_automixmonitor(bridge_channel, rec_chan, features_cfg, stop_message);
					}
					ast_channel_unref(rec_chan);
				}
			}*/
#endif
		} else {
#ifdef GRANDSTREAM_NETWORKS
			if (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) {
				const char *rec_paused_self = NULL, *rec_paused_peer = NULL, 
						   *rec_paused_real_peer = NULL, *rec_paused_local_peer = NULL;
				int r = REC_OK;
				int f = 0;
				time_t record_resume_utc = ast_now_utc_timestamp();

				if (bridge_channel->chan) {
					rec_paused_self = pbx_builtin_getvar_helper(bridge_channel->chan, "REC_PAUSED");
					if (rec_paused_self && !strcmp(rec_paused_self, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] resume self ->%s", ast_channel_name(bridge_channel->chan));
						r = resume_automixmonitor(bridge_channel, bridge_channel->chan, features_cfg, start_message);
						if (r == REC_FILE_UNAVAILABLE) {
							stop_automixmonitor(bridge_channel, bridge_channel->chan, features_cfg, NULL);
							pbx_builtin_setvar_helper(bridge_channel->chan, "REC_PAUSED", NULL);
						} else {
							f = 1;
						}
					}
				}
				
				if (peer_chan) {
					rec_paused_peer = pbx_builtin_getvar_helper(peer_chan, "REC_PAUSED");
					if (rec_paused_peer && !strcmp(rec_paused_peer, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] resume peer ->%s", ast_channel_name(peer_chan));
						r = resume_automixmonitor(bridge_channel, peer_chan, features_cfg, start_message);
						if (r == REC_FILE_UNAVAILABLE) {
							stop_automixmonitor(bridge_channel, peer_chan, features_cfg, NULL);
							pbx_builtin_setvar_helper(peer_chan, "REC_PAUSED", NULL);
						} else {
							f = 1;
						}
					}
				}
				
				if (real_peer_chan) {
					rec_paused_real_peer = pbx_builtin_getvar_helper(real_peer_chan, "REC_PAUSED");
					if (rec_paused_real_peer && !strcmp(rec_paused_real_peer, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] resume real peer ->%s", ast_channel_name(real_peer_chan));
						r = resume_automixmonitor(bridge_channel, real_peer_chan, features_cfg, start_message);
						if (r == REC_FILE_UNAVAILABLE) {
							stop_automixmonitor(bridge_channel, real_peer_chan, features_cfg, NULL);
							pbx_builtin_setvar_helper(real_peer_chan, "REC_PAUSED", NULL);
						} else  {
							f = 1;
						}						
					}
				}
				
				if (local_peer_chan) {
					rec_paused_local_peer = pbx_builtin_getvar_helper(local_peer_chan, "REC_PAUSED");
					if (rec_paused_local_peer && !strcmp(rec_paused_local_peer, "yes")) {
						ast_log(LOG_DEBUG, "[feature_automixmonitor] resume local peer ->%s", ast_channel_name(local_peer_chan));
						r = resume_automixmonitor(bridge_channel, local_peer_chan, features_cfg, start_message);
						if (r == REC_FILE_UNAVAILABLE) {
							stop_automixmonitor(bridge_channel, local_peer_chan, features_cfg, NULL);
							pbx_builtin_setvar_helper(local_peer_chan, "REC_PAUSED", NULL);
						} else {
							f = 1;
						}
					}
				}

				if (f) {
					if (bridge_channel->chan) {
						ast_channel_set_recordresume_utc(bridge_channel->chan, record_resume_utc);
					}
					if (peer_chan) {
						ast_channel_set_recordresume_utc(peer_chan, record_resume_utc);
					}
					if (real_peer_chan) {
						ast_channel_set_recordresume_utc(real_peer_chan, record_resume_utc);
					}
					if (local_peer_chan) {
						ast_channel_set_recordresume_utc(local_peer_chan, record_resume_utc);
					}
				}	
				
				if ((!rec_paused_self  && !rec_paused_peer && !rec_paused_real_peer && !rec_paused_local_peer) ||
				    r == REC_FILE_UNAVAILABLE) {
					if (bridge_channel->chan) {
						ast_channel_set_recordstart_utc(bridge_channel->chan, 0);
						ast_channel_set_recordfile(bridge_channel->chan, "");
					}
					if (peer_chan) {
						ast_channel_set_recordstart_utc(peer_chan, 0);
						ast_channel_set_recordfile(peer_chan, "");
					}
					if (real_peer_chan) {
						ast_channel_set_recordstart_utc(real_peer_chan, 0);
						ast_channel_set_recordfile(real_peer_chan, "");
					}
					if (local_peer_chan) {
						ast_channel_set_recordstart_utc(local_peer_chan, 0);
						ast_channel_set_recordfile(local_peer_chan, "");
					}

					ast_log(LOG_DEBUG, "[feature_automixmonitor] start peer ->%s", ast_channel_name(peer_chan));
					start_automixmonitor(bridge_channel, peer_chan, features_cfg, start_message);
				}
			} else {
				start_automixmonitor(bridge_channel, peer_chan, features_cfg, start_message);
			}
#else
			start_automixmonitor(bridge_channel, peer_chan, features_cfg, start_message);
#endif
		}
		return 0;

	case AUTO_MONITOR_START:
		if (!is_monitoring) {
			start_automixmonitor(bridge_channel, peer_chan, features_cfg, start_message);
			return 0;
		}
		ast_verb(4, "AutoMixMonitor already recording call.\n");
		break;

	case AUTO_MONITOR_STOP:
		if (is_monitoring) {
			stop_automixmonitor(bridge_channel, peer_chan, features_cfg, stop_message);
			return 0;
		}
		ast_verb(4, "AutoMixMonitor already stopped on call.\n");
		break;
	}

	/*
	 * Fake start/stop to invoker so will think it did something but
	 * was already in that mode.
	 */
	if (features_cfg && !ast_strlen_zero(features_cfg->courtesytone)) {
		ast_bridge_channel_queue_playfile(bridge_channel, NULL, features_cfg->courtesytone, NULL);
	}
	if (is_monitoring) {
		if (!ast_strlen_zero(start_message)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, start_message, NULL);
		}
	} else {
		if (!ast_strlen_zero(stop_message)) {
			ast_bridge_channel_queue_playfile(bridge_channel, NULL, stop_message, NULL);
		}
	}
	return 0;
}

/*! \brief Internal built in feature for hangup */
static int feature_hangup(struct ast_bridge_channel *bridge_channel, void *hook_pvt)
{
	/*
	 * This is very simple, we simply change the state on the
	 * bridge_channel to force the channel out of the bridge and the
	 * core takes care of the rest.
	 */
#ifdef GRANDSTREAM_NETWORKS
	if (bridge_channel->chan && (strcasestr(ast_channel_name(bridge_channel->chan),"PJSIP/trunk")
		|| strcasestr(ast_channel_name(bridge_channel->chan),"DAHDI/o"))) {
		const char *from_disa = pbx_builtin_getvar_helper(bridge_channel->chan, "FROM_DISA");
		const char *disa_hangup = pbx_builtin_getvar_helper(bridge_channel->chan, "DISA_HANGUP");

		if (ast_strlen_zero(from_disa) || strcmp(from_disa, "yes")
			|| ast_strlen_zero(disa_hangup) || strcmp(disa_hangup, "Hg")) {
			ast_log(LOG_DEBUG,
				"Not allow Trunk %s to hangup because not in DISA or DISA_HANGUP is disabled\n",
				ast_channel_name(bridge_channel->chan));
			return 0;
		}
	}
#endif
	ast_bridge_channel_leave_bridge(bridge_channel, BRIDGE_CHANNEL_STATE_END,
		AST_CAUSE_NORMAL_CLEARING);
	return 0;
}

static int unload_module(void)
{
	ast_bridge_features_unregister(AST_BRIDGE_BUILTIN_HANGUP);
	ast_bridge_features_unregister(AST_BRIDGE_BUILTIN_AUTOMON);
	ast_bridge_features_unregister(AST_BRIDGE_BUILTIN_AUTOMIXMON);

	return 0;
}

static int load_module(void)
{
	ast_bridge_features_register(AST_BRIDGE_BUILTIN_HANGUP, feature_hangup, NULL);
	ast_bridge_features_register(AST_BRIDGE_BUILTIN_AUTOMON, feature_automonitor, NULL);
	ast_bridge_features_register(AST_BRIDGE_BUILTIN_AUTOMIXMON, feature_automixmonitor, NULL);

	/* This module cannot be unloaded until shutdown */
	ast_module_shutdown_ref(ast_module_info->self);

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Built in bridging features",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.optional_modules = "res_monitor",
);
