/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2007, Tilghman Lesher
 *
 * Tilghman Lesher <cdr_adaptive_odbc__v1@the-tilghman.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 * \brief Adaptive ODBC CDR backend
 *
 * \author Tilghman Lesher <cdr_adaptive_odbc__v1@the-tilghman.com>
 * \ingroup cdr_drivers
 */

/*! \li \ref cdr_adaptive_odbc.c uses the configuration file \ref cdr_adaptive_odbc.conf
 * \addtogroup configuration_file Configuration Files
 */

/*!
 * \page cdr_adaptive_odbc.conf cdr_adaptive_odbc.conf
 * \verbinclude cdr_adaptive_odbc.conf.sample
 */

/*** MODULEINFO
    <depend>ucmsecret</depend>
	<depend>res_odbc</depend>
	<depend>generic_odbc</depend>
	<support_level>core</support_level>
 ***/
#ifdef GRANDSTREAM_NETWORKS
#define FIFO "/tmp/cdr_monitor_fifo"
#define FIFO_CALLSPRO "/tmp/cdr_callspro_fifo"
#endif

#include "asterisk.h"

#include <sys/types.h>
#include <time.h>

#include <sql.h>
#include <sqlext.h>
#include <sqltypes.h>

#include "asterisk/config.h"
#include "asterisk/channel.h"
#include "asterisk/lock.h"
#include "asterisk/linkedlists.h"
#include "asterisk/res_odbc.h"
#include "asterisk/cdr.h"
#include "asterisk/module.h"
#include "asterisk/utils.h"
#include "ucm_crypt.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/astdb.h"
#include "asterisk/logger.h"
#include "asterisk/pbx.h"
#include <sys/stat.h>
#include "asterisk/haserver.h"
#include "asterisk/security_events.h"
#include <sys/types.h>
#include <fcntl.h>
#include "asterisk/cdr_adaptive_odbc.h"
AST_MODULE_LOG("cdr");
#endif

#define	CONFIG	"cdr_adaptive_odbc.conf"

static const char name[] = "Adaptive ODBC";
/* Optimization to reduce number of memory allocations */
static int maxsize = 512, maxsize2 = 512;

#ifdef GRANDSTREAM_NETWORKS
#define FIFO_MISSED_CALL "/tmp/missed_call_fifo"

static int  fifo_fd;
static int  fifo_callspro_fd;
static int  pms_enable;
static int  crm_enable;
static int  cdr_callspro_enable;
static int  ha_enable = 0;
long int acctid = 0;
static int  missed_call_fd;
#endif

struct columns {
	char *name;
	char *cdrname;
	char *filtervalue;
	char *staticvalue;
	SQLSMALLINT type;
	SQLINTEGER size;
	SQLSMALLINT decimals;
	SQLSMALLINT radix;
	SQLSMALLINT nullable;
	SQLINTEGER octetlen;
	AST_LIST_ENTRY(columns) list;
	unsigned int negatefiltervalue:1;
};

struct tables {
	char *connection;
	char *table;
	char *schema;
	char quoted_identifiers;
	unsigned int usegmtime:1;
	AST_LIST_HEAD_NOLOCK(odbc_columns, columns) columns;
	AST_RWLIST_ENTRY(tables) list;
};

static AST_RWLIST_HEAD_STATIC(odbc_tables, tables);
#ifdef GRANDSTREAM_NETWORKS
/*node for exten*/
struct exten_node{
	char extension[128];
	char enabled[4];
	AST_LIST_ENTRY(exten_node) list;
};

/*linklist*/
struct exten_list {
	AST_LIST_HEAD_NOLOCK(missext, exten_node) exten_node;
	AST_RWLIST_ENTRY(exten_list) list;
};

/*global list for missed call*/
static AST_RWLIST_HEAD_STATIC(global_extlists, exten_list);
#endif
static int load_config(void)
{
	struct ast_config *cfg;
	struct ast_variable *var;
	const char *tmp, *catg;
	struct tables *tableptr;
	struct columns *entry;
	struct odbc_obj *obj;
	char columnname[80];
	char connection[40];
	char table[40];
	char schema[40];
	char quoted_identifiers;
	int lenconnection, lentable, lenschema, usegmtime = 0;
	SQLLEN sqlptr;
	int res = 0;
	SQLHSTMT stmt = NULL;
	struct ast_flags config_flags = { 0 }; /* Part of our config comes from the database */

	cfg = ast_config_load(CONFIG, config_flags);
	if (!cfg || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_WARNING, "Unable to load " CONFIG ".  No adaptive ODBC CDRs.\n");
		return -1;
	}

	for (catg = ast_category_browse(cfg, NULL); catg; catg = ast_category_browse(cfg, catg)) {
		var = ast_variable_browse(cfg, catg);
		if (!var)
			continue;

		if (ast_strlen_zero(tmp = ast_variable_retrieve(cfg, catg, "connection"))) {
			ast_log(LOG_WARNING, "No connection parameter found in '%s'.  Skipping.\n", catg);
			continue;
		}
		ast_copy_string(connection, tmp, sizeof(connection));
		lenconnection = strlen(connection);

		if (!ast_strlen_zero(tmp = ast_variable_retrieve(cfg, catg, "usegmtime"))) {
			usegmtime = ast_true(tmp);
		}

		/* When loading, we want to be sure we can connect. */
		obj = ast_odbc_request_obj(connection, 1);
		if (!obj) {
			ast_log(LOG_WARNING, "No such connection '%s' in the '%s' section of " CONFIG ".  Check res_odbc.conf.\n", connection, catg);
			continue;
		}

		if (ast_strlen_zero(tmp = ast_variable_retrieve(cfg, catg, "table"))) {
			ast_log(LOG_NOTICE, "No table name found.  Assuming 'cdr'.\n");
			tmp = "cdr";
		}
		ast_copy_string(table, tmp, sizeof(table));
		lentable = strlen(table);

		if (ast_strlen_zero(tmp = ast_variable_retrieve(cfg, catg, "schema"))) {
			tmp = "";
		}
		ast_copy_string(schema, tmp, sizeof(schema));
		lenschema = strlen(schema);

		if (ast_strlen_zero(tmp = ast_variable_retrieve(cfg, catg, "quoted_identifiers"))) {
			tmp = "";
		}
		quoted_identifiers = tmp[0];
		if (strlen(tmp) > 1) {
			ast_log(LOG_ERROR, "The quoted_identifiers setting only accepts a single character,"
				" while a value of '%s' was provided. This option has been disabled as a result.\n", tmp);
			quoted_identifiers = '\0';
		}

		res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
		if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
			ast_log(LOG_WARNING, "SQL Alloc Handle failed on connection '%s'!\n", connection);
			ast_odbc_release_obj(obj);
			continue;
		}

		res = SQLColumns(stmt, NULL, 0, lenschema == 0 ? NULL : (unsigned char *)schema, SQL_NTS, (unsigned char *)table, SQL_NTS, (unsigned char *)"%", SQL_NTS);
		if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
			ast_log(LOG_ERROR, "Unable to query database columns on connection '%s'.  Skipping.\n", connection);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			ast_odbc_release_obj(obj);
			continue;
		}

		tableptr = ast_calloc(sizeof(char), sizeof(*tableptr) + lenconnection + 1 + lentable + 1 + lenschema + 1 + 1);
		if (!tableptr) {
			ast_log(LOG_ERROR, "Out of memory creating entry for table '%s' on connection '%s'%s%s%s\n", table, connection,
				lenschema ? " (schema '" : "", lenschema ? schema : "", lenschema ? "')" : "");
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			ast_odbc_release_obj(obj);
			res = -1;
			break;
		}

		tableptr->usegmtime = usegmtime;
		tableptr->connection = (char *)tableptr + sizeof(*tableptr);
		tableptr->table = (char *)tableptr + sizeof(*tableptr) + lenconnection + 1;
		tableptr->schema = (char *)tableptr + sizeof(*tableptr) + lenconnection + 1 + lentable + 1;
		ast_copy_string(tableptr->connection, connection, lenconnection + 1);
		ast_copy_string(tableptr->table, table, lentable + 1);
		ast_copy_string(tableptr->schema, schema, lenschema + 1);
		tableptr->quoted_identifiers = quoted_identifiers;

		ast_verb(3, "Found adaptive CDR table %s@%s.\n", tableptr->table, tableptr->connection);

		/* Check for filters first */
		for (var = ast_variable_browse(cfg, catg); var; var = var->next) {
			if (strncmp(var->name, "filter", strlen("filter")) == 0) {
				int negate = 0;
				char *cdrvar = ast_strdupa(var->name + 6);
				cdrvar = ast_strip(cdrvar);
				if (cdrvar[strlen(cdrvar) - 1] == '!') {
					negate = 1;
					cdrvar[strlen(cdrvar) - 1] = '\0';
					ast_trim_blanks(cdrvar);
				}

				ast_verb(3, "Found filter %s'%s' for CDR variable %s in %s@%s\n", negate ? "!" : "", var->value, cdrvar, tableptr->table, tableptr->connection);

				entry = ast_calloc(sizeof(char), sizeof(*entry) + strlen(cdrvar) + 1 + strlen(var->value) + 1);
				if (!entry) {
					ast_log(LOG_ERROR, "Out of memory creating filter entry for CDR variable '%s' in table '%s' on connection '%s'\n", cdrvar, table, connection);
					res = -1;
					break;
				}

				/* NULL column entry means this isn't a column in the database */
				entry->name = NULL;
				entry->cdrname = (char *)entry + sizeof(*entry);
				entry->filtervalue = (char *)entry + sizeof(*entry) + strlen(cdrvar) + 1;
				strcpy(entry->cdrname, cdrvar);
				strcpy(entry->filtervalue, var->value);
				entry->negatefiltervalue = negate;

				AST_LIST_INSERT_TAIL(&(tableptr->columns), entry, list);
			}
		}

		while ((res = SQLFetch(stmt)) != SQL_NO_DATA && res != SQL_ERROR) {
			char *cdrvar = "", *staticvalue = "";

			SQLGetData(stmt,  4, SQL_C_CHAR, columnname, sizeof(columnname), &sqlptr);

			/* Is there an alias for this column? */

			/* NOTE: This seems like a non-optimal parse method, but I'm going
			 * for user configuration readability, rather than fast parsing. We
			 * really don't parse this file all that often, anyway.
			 */
			for (var = ast_variable_browse(cfg, catg); var; var = var->next) {
				if (strncmp(var->name, "alias", strlen("alias")) == 0 && strcasecmp(var->value, columnname) == 0) {
					char *alias = ast_strdupa(var->name + 5);
					cdrvar = ast_strip(alias);
					ast_verb(3, "Found alias %s for column %s in %s@%s\n", cdrvar, columnname, tableptr->table, tableptr->connection);
					break;
				} else if (strncmp(var->name, "static", strlen("static")) == 0 && strcasecmp(var->value, columnname) == 0) {
					char *item = ast_strdupa(var->name + 6);
					item = ast_strip(item);
					if (item[0] == '"' && item[strlen(item) - 1] == '"') {
						/* Remove surrounding quotes */
						item[strlen(item) - 1] = '\0';
						item++;
					}
					staticvalue = item;
				}
			}

			entry = ast_calloc(sizeof(char), sizeof(*entry) + strlen(columnname) + 1 + strlen(cdrvar) + 1 + strlen(staticvalue) + 1);
			if (!entry) {
				ast_log(LOG_ERROR, "Out of memory creating entry for column '%s' in table '%s' on connection '%s'\n", columnname, table, connection);
				res = -1;
				SQLFreeHandle(SQL_HANDLE_STMT, stmt);
				break;
			}
			entry->name = (char *)entry + sizeof(*entry);
			strcpy(entry->name, columnname);

			if (!ast_strlen_zero(cdrvar)) {
				entry->cdrname = entry->name + strlen(columnname) + 1;
				strcpy(entry->cdrname, cdrvar);
			} else { /* Point to same place as the column name */
				entry->cdrname = (char *)entry + sizeof(*entry);
			}

			if (!ast_strlen_zero(staticvalue)) {
				entry->staticvalue = entry->cdrname + strlen(entry->cdrname) + 1;
				strcpy(entry->staticvalue, staticvalue);
			}

			SQLGetData(stmt,  5, SQL_C_SHORT, &entry->type, sizeof(entry->type), NULL);
			SQLGetData(stmt,  7, SQL_C_LONG, &entry->size, sizeof(entry->size), NULL);
			SQLGetData(stmt,  9, SQL_C_SHORT, &entry->decimals, sizeof(entry->decimals), NULL);
			SQLGetData(stmt, 10, SQL_C_SHORT, &entry->radix, sizeof(entry->radix), NULL);
			SQLGetData(stmt, 11, SQL_C_SHORT, &entry->nullable, sizeof(entry->nullable), NULL);
			SQLGetData(stmt, 16, SQL_C_LONG, &entry->octetlen, sizeof(entry->octetlen), NULL);

			/* Specification states that the octenlen should be the maximum number of bytes
			 * returned in a char or binary column, but it seems that some drivers just set
			 * it to NULL. (Bad Postgres! No biscuit!) */
			if (entry->octetlen == 0)
				entry->octetlen = entry->size;

			ast_verb(4, "Found %s column with type %hd with len %ld, octetlen %ld, and numlen (%hd,%hd)\n", entry->name, entry->type, (long) entry->size, (long) entry->octetlen, entry->decimals, entry->radix);
			/* Insert column info into column list */
			AST_LIST_INSERT_TAIL(&(tableptr->columns), entry, list);
			res = 0;
		}

		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		ast_odbc_release_obj(obj);

		if (AST_LIST_FIRST(&(tableptr->columns)))
			AST_RWLIST_INSERT_TAIL(&odbc_tables, tableptr, list);
		else
			ast_free(tableptr);
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Get the pms_enable */
	if (!ast_strlen_zero(tmp = ast_variable_retrieve(cfg, "asterisk_event", "pms_enable"))) {
		pms_enable = atoi(tmp);
		ast_log(LOG_DEBUG, "pms_enable = %d.\n", pms_enable);
	} else {
		ast_log(LOG_DEBUG, "pms_enable not specified.  Assuming to 0.\n");
		pms_enable = 0;
	}
	/* Get the crm_enable */
	if (!ast_strlen_zero(tmp = ast_variable_retrieve(cfg, "asterisk_event", "crm_enable"))) {
		crm_enable = atoi(tmp);
		ast_log(LOG_DEBUG, "crm_enable = %d.\n", crm_enable);
	} else {
		ast_log(LOG_DEBUG, "crm_enable not specified.  Assuming to 0.\n");
		crm_enable = 0;
	}

	/* Get the crm_enable */
	if (!ast_strlen_zero(tmp = ast_variable_retrieve(cfg, "callspro", "enable"))) {
		cdr_callspro_enable = atoi(tmp);
		ast_log(LOG_DEBUG, "cdr_callspro_enable = %d.\n", cdr_callspro_enable);
	} else {
		ast_log(LOG_DEBUG, "cdr_callspro_enable not specified.	Assuming to 0.\n");
		cdr_callspro_enable = 0;
	}

	FILE *ha_fp = NULL;
	char on_buf[64] = { 0 };

	/* Get the ha_enable */
	ha_fp = popen("nvram get haon", "r");
	if (ha_fp != NULL) {
		while (ha_fp != NULL && (!feof(ha_fp)) && (fgets(on_buf, sizeof(on_buf) - 1, ha_fp) != NULL)) {
		ast_log(LOG_DEBUG, "haon Read: %s", on_buf);
	}

	pclose(ha_fp);
	ha_enable = atoi(on_buf);
	}
#endif

	ast_config_destroy(cfg);
	return res;
}

static int free_config(void)
{
	struct tables *table;
	struct columns *entry;
	while ((table = AST_RWLIST_REMOVE_HEAD(&odbc_tables, list))) {
		while ((entry = AST_LIST_REMOVE_HEAD(&(table->columns), list))) {
			ast_free(entry);
		}
		ast_free(table);
	}
	return 0;
}

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	int res, i;
	SQLHSTMT stmt;
	SQLINTEGER nativeerror = 0, numfields = 0;
	SQLSMALLINT diagbytes = 0;
	unsigned char state[10], diagnostic[256];

	res = SQLAllocHandle (SQL_HANDLE_STMT, obj->con, &stmt);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}

	res = SQLPrepare(stmt, (unsigned char *) data, SQL_NTS);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(LOG_WARNING, "SQL Prepare failed![%s]\n", (char *) data);
		SQLGetDiagField(SQL_HANDLE_STMT, stmt, 1, SQL_DIAG_NUMBER, &numfields, SQL_IS_INTEGER, &diagbytes);
		for (i = 0; i < numfields; i++) {
			SQLGetDiagRec(SQL_HANDLE_STMT, stmt, i + 1, state, &nativeerror, diagnostic, sizeof(diagnostic), &diagbytes);
			ast_log(LOG_WARNING, "SQL Execute returned an error %d: %s: %s (%d)\n", res, state, diagnostic, diagbytes);
			if (i > 10) {
				ast_log(LOG_WARNING, "Oh, that was good.  There are really %d diagnostics?\n", (int)numfields);
				break;
			}
		}
		SQLFreeHandle (SQL_HANDLE_STMT, stmt);
		return NULL;
	}

	return stmt;
}

#define LENGTHEN_BUF(size, var_sql)														\
			do {																\
				/* Lengthen buffer, if necessary */								\
				if (ast_str_strlen(var_sql) + size + 1 > ast_str_size(var_sql)) {		\
					if (ast_str_make_space(&var_sql, ((ast_str_size(var_sql) + size + 1) / 512 + 1) * 512) != 0) { \
						ast_log(LOG_ERROR, "Unable to allocate sufficient memory.  Insert CDR '%s:%s' failed.\n", tableptr->connection, tableptr->table); \
						ast_free(sql);											\
						ast_free(sql2);											\
						AST_RWLIST_UNLOCK(&odbc_tables);						\
						return -1;												\
					}															\
				}																\
			} while (0)

#define LENGTHEN_BUF1(size)	\
	LENGTHEN_BUF(size, sql);
#define LENGTHEN_BUF2(size)	\
	LENGTHEN_BUF(size, sql2);
#ifdef GRANDSTREAM_NETWORKS
/*add by wxu@grandstream for CDR check trunk*/

/*flag=0 is inbound, flag=1 is outbound*/
static void set_trunk(struct ast_cdr *cdr, const char *trunk_index, int flag)
{
	char trunkName[128] = {0};
	char family[64] = {0};
	if (ast_strlen_zero(trunk_index)){
		return;
	}
	snprintf(family, sizeof(family) - 1, "TRUNKINDEX/DAHDI_%s", trunk_index);
	ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
	if (ast_strlen_zero(trunkName)) {
		snprintf(family, sizeof(family), "TRUNK/DIGITAL_%s", trunk_index);
		ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
	}
	if (ast_strlen_zero(trunkName)) {
		snprintf(family, sizeof(family) - 1, "TRUNK/IAX2_%s", trunk_index);
		ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
	}
	if (ast_strlen_zero(trunkName)) {
		snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%s", trunk_index);
		ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
	}
	if (!ast_strlen_zero(trunkName)) {
		if (flag){
			snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
		} else{
			snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "%s", trunkName);
		}
	}
	return;
}

static void check_trunk(struct ast_cdr *cdr, const char *call)
{
	char *channel = NULL;
	char family[64] = {0};
	char trunkName[128] = {0};

	if (call != NULL) {
		if (!strcasecmp(call, "caller")) {
			channel = cdr->channel;
		} else if (!strcasecmp(call, "callee")) {
			channel = cdr->dstchannel;
		} else {
			return;
		}
		if (channel != NULL && channel[0] != '\0') {
			if (strstr(channel, "/trunk_") != NULL || strstr(channel, "/webrtc_trunk_") != NULL) {
				if (!strcasecmp(call, "caller") && cdr->src[0] != '\0') {
					char inbound_trunk_index[16] = {0};
					char *result = NULL;
					ast_cdr_format_var(cdr, "inbound_trunk_index", &result, inbound_trunk_index,
									   sizeof(inbound_trunk_index), 0);
					ast_log(LOG_NOTICE, "inbound_trunk_index = %s\n", inbound_trunk_index);
					if (ast_strlen_zero(inbound_trunk_index)
						&& strcasestr(cdr->session, "calleeseamless")
						&& AST_CDR_ANSWERED != cdr->disposition){
						sscanf(channel, "%*[^_]_%[^-]%*s", inbound_trunk_index);
					}
					if (strstr(channel, "IAX2")) {
						snprintf(family, sizeof(family) - 1, "TRUNK/IAX2_%d", atoi(inbound_trunk_index));
					} else {
						snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(inbound_trunk_index));
					}
					ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
					memset(cdr->src_trunk_name, 0, sizeof(cdr->src_trunk_name));
				} else if (!strcasecmp(call, "callee") && cdr->dst[0] != '\0') {
					int trunk = 0;
					if (strstr(channel, "/trunk_") != NULL) {
						trunk = atoi(strstr(channel, "/trunk_") + 7);
					}else if (strstr(channel, "/webrtc_trunk_") != NULL) {
						trunk = atoi(strstr(channel, "/webrtc_trunk_") + 14);
					}
					if (strstr(channel, "IAX2")) {
						snprintf(family, sizeof(family) - 1, "TRUNK/IAX2_%d", trunk);
					} else {
						snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", trunk);
					}
					ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
					memset(cdr->dst_trunk_name, 0, sizeof(cdr->dst_trunk_name));
				}
				if (trunkName[0] != '\0') {
					// ast_log(LOG_DEBUG,"exten:%s,trunkName:%s",exten,trunkName);
					if (!strcasecmp(call, "caller")) {
						if (strcasestr(cdr->session, "calleeseamless") && AST_CDR_ANSWERED != cdr->disposition){
							snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
						}else{
							snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "%s", trunkName);
						}
					} else if (!strcasecmp(call, "callee")) {
						snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
					}
				}
				if (strcasestr(cdr->session, "calleeseamless") && ast_strlen_zero(cdr->dstchannel)
					&& (!ast_strlen_zero(cdr->outbound_trunk_index) || !ast_strlen_zero(cdr->inbound_trunk_index))){
					set_trunk(cdr, cdr->outbound_trunk_index, 1);
					set_trunk(cdr, cdr->inbound_trunk_index, 0);
				}
			} else if (strstr(channel, "DAHDI/i3")) { // digital trunk : SS7 PRI R2 EM_IM EM_EINK
				if (!strcasecmp(call, "callee")) {
					char outbound_trunk_index[16] = {0};
					char *result = NULL;
					ast_cdr_format_var(cdr, "outbound_trunk_index", &result, outbound_trunk_index,
									   sizeof(outbound_trunk_index), 0);
					ast_log(LOG_NOTICE, "outbound_trunk_index = %s\n", outbound_trunk_index);
					snprintf(family, sizeof(family), "TRUNK/DIGITAL_%d", atoi(outbound_trunk_index));
					ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
					snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
				} else if (!strcasecmp(call, "caller")) {
					char inbound_trunk_index[16] = {0};
					char *result = NULL;
					ast_cdr_format_var(cdr, "inbound_trunk_index", &result, inbound_trunk_index,
									   sizeof(inbound_trunk_index), 0);
					ast_log(LOG_NOTICE, "inbound_trunk_index = %s\n", inbound_trunk_index);
					snprintf(family, sizeof(family), "TRUNK/DIGITAL_%d", atoi(inbound_trunk_index));
					ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
					snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "%s", trunkName);
				}
			} else if (strstr(channel, "DAHDI/o")) {
				int Dahdi_chan = atoi(strstr(channel, "DAHDI/o") + strlen("DAHDI/o"));
				snprintf(family, sizeof(family) - 1, "TRUNK/DAHDI_%d", Dahdi_chan);
				ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));

				if (!strcasecmp(call, "caller") && cdr->src[0] != '\0') {
					memset(cdr->src_trunk_name, 0, sizeof(cdr->src_trunk_name));
				} else if (!strcasecmp(call, "callee") && cdr->dst[0] != '\0') {
					memset(cdr->dst_trunk_name, 0, sizeof(cdr->dst_trunk_name));
				}

				if (trunkName[0] != '\0') {
					if (strcasestr(cdr->session, "calleeseamless") && ast_strlen_zero(cdr->dstchannel)){
						snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
						set_trunk(cdr, cdr->inbound_trunk_index, 0);
					} else if (!strcasecmp(call, "caller")) {
						snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "%s", trunkName);
					} else if (!strcasecmp(call, "callee")) {
						snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
					}
				} else if (Dahdi_chan > 0) {
					if (!strcasecmp(call, "caller")) {
						snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "FXO_%d", Dahdi_chan);
					} else if (!strcasecmp(call, "callee")) {
						snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "FXO_%d", Dahdi_chan);
					}
				}
			} else if (!strstr(channel, "DAHDI/i3") && !strstr(channel, "DAHDI/o") && strstr(channel, "DAHDI/")) {
				int numFXO = ast_get_dahdi_fxo_number();
				int numFXS = ast_get_dahdi_fxs_number();
				int Dahdi_chan = atoi(strstr(channel, "DAHDI/") + 6);

				if (Dahdi_chan > numFXO + numFXS) { // EM_WINK OR EM_IM
					if (!strcasecmp(call, "callee")) {
						char outbound_trunk_index[16] = {0};
						char *result = NULL;
						ast_cdr_format_var(cdr, "outbound_trunk_index", &result, outbound_trunk_index,
										   sizeof(outbound_trunk_index), 0);
						ast_log(LOG_NOTICE, "outbound_trunk_index = %s\n", outbound_trunk_index);
						snprintf(family, sizeof(family), "TRUNK/DIGITAL_%d", atoi(outbound_trunk_index));
						ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
						snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
					} else if (!strcasecmp(call, "caller")) {
						char inbound_trunk_index[16] = {0};
						char *result = NULL;
						ast_cdr_format_var(cdr, "inbound_trunk_index", &result, inbound_trunk_index,
										   sizeof(inbound_trunk_index), 0);
						ast_log(LOG_NOTICE, "inbound_trunk_index = %s\n", inbound_trunk_index);
						snprintf(family, sizeof(family), "TRUNK/DIGITAL_%d", atoi(inbound_trunk_index));
						ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
						snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "%s", trunkName);
					}
				} else if ((Dahdi_chan > numFXO) && (Dahdi_chan <= numFXO + numFXS)) { // FXS port
					/*int fxs = Dahdi_chan - numFXO;
					sql = sqlite3_mprintf("SELECT extension FROM fxs_accounts WHERE dahdi='%d';",
					fxs);
					check_ucm_config(sql, exten);
					sqlite3_free(sql);

					if ( exten[0] != '\0' ){
						snprintf(temp, sizeof(temp), "%s", exten);
						ast_cdr_setvar(cdr, "dstanswer", temp, 0);
					}
					else if ( fxs > 0 ){
						snprintf(temp, sizeof(temp), "FXS_%d", fxs);
						ast_cdr_setvar(cdr, "dstanswer", temp, 0);
					}*/
				} else if ((Dahdi_chan >= 1) && (Dahdi_chan <= numFXO)) { // FXO trunk
					snprintf(family, sizeof(family) - 1, "TRUNK/DAHDI_%d", Dahdi_chan);
					ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));

					if (!strcasecmp(call, "caller") && cdr->src[0] != '\0') {
						memset(cdr->src_trunk_name, 0, sizeof(cdr->src_trunk_name));
					} else if (!strcasecmp(call, "callee") && cdr->dst[0] != '\0') {
						memset(cdr->dst_trunk_name, 0, sizeof(cdr->dst_trunk_name));
					}

					if (trunkName[0] != '\0') {
						if (!strcasecmp(call, "caller")) {
							snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "%s", trunkName);
						} else if (!strcasecmp(call, "callee")) {
							snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "%s", trunkName);
						}
					} else if (Dahdi_chan > 0) {
						if (!strcasecmp(call, "caller")) {
							snprintf(cdr->src_trunk_name, sizeof(cdr->src_trunk_name), "FXO_%d", Dahdi_chan);
						} else if (!strcasecmp(call, "callee")) {
							snprintf(cdr->dst_trunk_name, sizeof(cdr->dst_trunk_name), "FXO_%d", Dahdi_chan);
						}
					}
				}
			}
			if(strcasestr(channel, "Local") && strcasestr(channel, ";2")
				&& !ast_strlen_zero(cdr->inbound_trunk_index)
				&& (strcasestr(cdr->special_action_type, "local_disa")
					|| !strncmp(cdr->action_type, "CONFERENCE", strlen("CONFERENCE"))
					|| strstr(channel, "call-forward")
					|| (!strncmp(cdr->action_type, "TRANSFER", strlen("TRANSFER")) && strstr(channel, ";2") && strstr(channel, "internal-xfer")))){
					set_trunk(cdr,cdr->inbound_trunk_index,0);
			}
		}
		if (!strncmp(cdr->action_type, "FOLLOWME", strlen("FOLLOWME")) && strcasestr(cdr->dstchannel, "Local") && !ast_strlen_zero(cdr->outbound_trunk_index)){
			set_trunk(cdr, cdr->outbound_trunk_index, 1);
		}
		if(((strcasestr(cdr->lastdata, "seamless_transfer_user") || !strncmp(cdr->dcontext, "outbound-route-", strlen("outbound-route-")))
					&& !strcasecmp(call, "callee")
					&& (channel == NULL || channel[0] == '\0'))
			|| (strcasestr(cdr->channel, "mcb-create")
					&& !strcasecmp(call, "callee")
					&& strcasestr(channel, "Local"))){
			char outbound_trunk_index[16] = {0};
			char *result = NULL;
			ast_cdr_format_var(cdr, "outbound_trunk_index", &result, outbound_trunk_index,
							   sizeof(outbound_trunk_index), 0);
			ast_log(LOG_NOTICE, "outbound_trunk_index = %s\n", outbound_trunk_index);
			set_trunk(cdr, outbound_trunk_index, 1);
		}

		if (0 == strcmp(cdr->special_action_type, "COUNTRY_CODE_LIMIT")){
			set_trunk(cdr, cdr->outbound_trunk_index, 1);
		}

		if((strcasestr(cdr->clid,"Call Bargin") || strcasestr(cdr->clid,"Call Whisper") || strcasestr(cdr->clid,"Call Listen"))){
			if (AST_CDR_ANSWERED != cdr->disposition && cdr->next && !ast_strlen_zero(cdr->next->inbound_trunk_index)){
				set_trunk(cdr, cdr->next->inbound_trunk_index, 0);
			}
			if (!ast_strlen_zero(cdr->inbound_trunk_index)){
				set_trunk(cdr, cdr->inbound_trunk_index, 0);
			}
		}
		if (0 == strncmp(cdr->special_action_type, "FOLLOWME/external", strlen("FOLLOWME/external"))){
			set_trunk(cdr, cdr->outbound_trunk_index, 1);
		}
		if(!strcasecmp(call, "callee")
			&& (strcasestr(channel, "Local") && strcasestr(channel, "call-forward") && strcasestr(channel, ";1"))
			&& !ast_strlen_zero(cdr->outbound_trunk_index)){
			set_trunk(cdr, cdr->outbound_trunk_index, 1);
		}
		/* For SLA CDR callee display name */
		if (!strncmp(cdr->lastapp, "SLAStation", strlen("SLAStation")) && !strcasecmp(call, "callee") && strstr(channel, "DAHDI/o")) {
			memset(cdr->dst_trunk_name, 0, sizeof(cdr->dst_trunk_name));
			if (ast_strlen_zero(cdr->dst)){
				memset(cdr->dst, 0, sizeof(cdr->dst));
				snprintf(cdr->dst, sizeof(cdr->dst), "%s", cdr->lastdata);
			}
		}
		if (strstr(cdr->action_type, "MULTIMEDIAMEETING") && ast_strlen_zero(cdr->dstchannel) && !ast_strlen_zero(cdr->outbound_trunk_index)
			&& (strstr(channel, "DAHDI/o") || strstr(channel, "trunk_"))){
			memset(cdr->src_trunk_name, 0, sizeof(cdr->src_trunk_name));
			set_trunk(cdr, cdr->outbound_trunk_index, 1);
		}
		if (strcasestr(cdr->session, "calleeseamless") && ast_strlen_zero(cdr->dstchannel) && !ast_strlen_zero(cdr->inbound_trunk_index)){
			set_trunk(cdr, cdr->inbound_trunk_index, 0);
		}
	}

	if (!strcmp(cdr->src, "Callback")) {
		memset(cdr->src_trunk_name, 0, sizeof(cdr->src_trunk_name));
	}

	return;
}

static void check_call_type(struct ast_cdr *cdr)
{
	memset(cdr->userfield, 0, sizeof(cdr->userfield));
	if (strlen(cdr->src_trunk_name) > 0 && strlen(cdr->dst_trunk_name) > 0) {
		snprintf(cdr->userfield, sizeof(cdr->userfield), "%s", "External");
	} else if (strlen(cdr->src_trunk_name) > 0 && strlen(cdr->dst_trunk_name) == 0) {
		snprintf(cdr->userfield, sizeof(cdr->userfield), "%s", "Inbound");
	} else if (strlen(cdr->src_trunk_name) == 0 && strlen(cdr->dst_trunk_name) > 0) {
		snprintf(cdr->userfield, sizeof(cdr->userfield), "%s", "Outbound");
	} else {
		snprintf(cdr->userfield, sizeof(cdr->userfield), "%s", "Internal");
	}
}
/*add end*/

static int load_missed(void)
{
/*load file to see number enable func*/
	int res = 0;
	char *tmp = NULL;
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	struct exten_list * missed_list = NULL;
	struct exten_node *entry = NULL;
	struct ast_flags config_flags = { 0 };
	if (!(cfg = ast_config_load("missedcall.conf", config_flags))) {
		ast_log(LOG_NOTICE, "No config file (missedcall.conf)\n");
		return 1;
	} else if (CONFIG_STATUS_FILEUNCHANGED == cfg ) {
		return 0;
	} else if (CONFIG_STATUS_FILEINVALID == cfg ) {
		ast_log(LOG_ERROR, "Config file missedcall.conf is in an invalid format. Aborting.\n");
		return 1;
	}

	missed_list = ast_calloc(1, sizeof(struct exten_list));
	if (!missed_list) {
		ast_config_destroy(cfg);
		return -1;
	}
	while ((tmp = ast_category_browse(cfg, tmp)) ) {
		for (var = ast_variable_browse(cfg, tmp); var; var = var->next) {
			if (NULL != var->name && NULL != var->value){
				entry = ast_calloc(1, sizeof(struct exten_node));
				if (!entry) {
					res = -1;
					break;
				}
				ast_copy_string(entry->enabled,var->value,sizeof(entry->enabled));
				ast_copy_string(entry->extension,var->name,sizeof(entry->extension));
				ast_log(LOG_DEBUG, "entry->extension:%s entry->enabled:%s", entry->extension, entry->enabled);
				AST_LIST_INSERT_TAIL(&(missed_list->exten_node), entry, list);
			}
		}
	}
	if (AST_LIST_FIRST(&(missed_list->exten_node)))
		AST_RWLIST_INSERT_TAIL(&global_extlists, missed_list, list);
	else
		ast_free(missed_list);
	ast_config_destroy(cfg);

	return res;
}
static int unload_missed(void)
{
	struct exten_list * missed_list = NULL;
	struct exten_node *entry = NULL;
	while ((missed_list = AST_RWLIST_REMOVE_HEAD(&global_extlists, list))) {
		while ((entry = AST_LIST_REMOVE_HEAD(&(missed_list->exten_node), list))) {
			ast_free(entry);
		}
		ast_free(missed_list);
	}
	return 0;
}

static int contacts_number(char *contacts)
{
	int numbers = 1;
	const char *ptr = NULL;
	if (NULL == contacts || '\0' == contacts){
		return 0;
	}
	ptr = contacts;
	while ('\0' != *ptr){
		if ('&' == *ptr){
			numbers++;
		}
		++ptr;
	}
	ast_log(LOG_DEBUG, "dialed numbers:%d", numbers);
	return numbers;
}

static int update_trunk_time(void)
{
	struct ast_custom_function *custom_function = ast_custom_function_find("TRUNK_TIME_UPDATE");

	if (custom_function && custom_function->read) {
		char buffer[8] = {0};
		custom_function->read(NULL, "TRUNK_TIME_UPDATE", NULL, buffer, sizeof(buffer) - 1);
	}

	return 0;
}
#endif

int ast_odbc_insert_cdr(char *sql)
{
	struct tables *tableptr;

	struct odbc_obj *obj;
	SQLHSTMT stmt = NULL;
	SQLLEN rows = 0;
	int ret = -1;

	if(!sql)
	{
		return -1;
	}
	if (AST_RWLIST_RDLOCK(&odbc_tables)) {
		ast_log(LOG_ERROR, "Unable to lock table list.	Insert CDR(s) failed.\n");
		return -1;
	}

	AST_LIST_TRAVERSE(&odbc_tables, tableptr, list) {
		if (strcasecmp(tableptr->table , "cdr"))
		{
			continue;
		}
		if (!(obj = ast_odbc_request_obj(tableptr->connection, 0))) {
			ast_log(LOG_WARNING, "cdr_adaptive_odbc: Unable to retrieve database handle for '%s:%s'.  CDR failed: %s\n", tableptr->connection, tableptr->table, sql);
			
			AST_RWLIST_UNLOCK(&odbc_tables);
			return -1;
		}

		stmt = ast_odbc_prepare_and_execute(obj, generic_prepare, sql);
		if (stmt) {
			SQLRowCount(stmt, &rows);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		}
		if (rows == 0) {
			ast_log(LOG_WARNING, "cdr_adaptive_odbc: Insert failed on '%s:%s'.  CDR failed: %s\n", tableptr->connection, tableptr->table, sql);
			ret = -1;
		}
		else
		{
			ast_log(LOG_DEBUG, "cdr_adaptive_odbc: Insert successed on '%s:%s'.  sql: %s\n", tableptr->connection, tableptr->table, sql);
			ret = 0;
		}
		ast_odbc_release_obj(obj);
		break;
		
	}

	AST_RWLIST_UNLOCK(&odbc_tables);
	return ret;
}

static int odbc_log(struct ast_cdr *cdr)
{
	struct tables *tableptr;
	struct columns *entry;
	struct odbc_obj *obj;
	struct ast_str *sql = ast_str_create(maxsize), *sql2 = ast_str_create(maxsize2);
	char *tmp;
	char colbuf[1024], *colptr;
	SQLHSTMT stmt = NULL;
	SQLLEN rows = 0;
#ifdef GRANDSTREAM_NETWORKS
	static char cdr_value[2048] = "";
	static char cdr_column[2048] = "";
	struct ast_str *sql_rowid = ast_str_create(maxsize);
	int cdr_insert_success =0;
	char rowdata[20];
	int res;
	char start_time[32] = {0};
	char send_buf[512] = {0};
	int need_send = 0;
	struct ast_tm tm;
	char cdr_clid[512] = {0};
	char* tmp_cdr_clid = NULL;
	int bill_sec = 0;
	char main_sql[1024] = {0};
	char tmp_pjsip_chan_head[256] = {0};
	struct odbc_obj *rf_obj = NULL;
#endif
	char *separator;
	int quoted = 0;


#ifdef GRANDSTREAM_NETWORKS
	if (!sql || !sql2 || !sql_rowid) 
#else
	if (!sql || !sql2 ) 
#endif
	{
		if (sql)
			ast_free(sql);
		if (sql2)
			ast_free(sql2);
#ifdef GRANDSTREAM_NETWORKS
		if (sql_rowid)
			ast_free(sql_rowid); 
#endif
		return -1;
	}

	
#ifdef GRANDSTREAM_NETWORKS
		/*add by wxu@grandstream for CDR check trunk and set CDR type*/
		check_trunk(cdr,"caller");
		check_trunk(cdr,"callee");
		check_call_type(cdr);
		/*end*/
#endif

	if (AST_RWLIST_RDLOCK(&odbc_tables)) {
		ast_log(LOG_ERROR, "Unable to lock table list.  Insert CDR(s) failed.\n");
		ast_free(sql);
		ast_free(sql2);
#ifdef GRANDSTREAM_NETWORKS
		ast_free(sql_rowid); 
#endif
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	AST_LIST_TRAVERSE(&odbc_tables, tableptr, list) {
		if (!strcasecmp(tableptr->table, "recording_files")) {
			rf_obj = ast_odbc_request_obj(tableptr->connection, 0);	
			break;
		}
	}

	if (!rf_obj) {
		ast_log(LOG_DEBUG, "[odbc_log] failed to get recording_files database object");
	}
#endif

	AST_LIST_TRAVERSE(&odbc_tables, tableptr, list) {
		separator = "";

		quoted = 0;
		if (tableptr->quoted_identifiers != '\0'){
			quoted = 1;
		}

#ifdef GRANDSTREAM_NETWORKS
		/*if master cdr insert failed ,also neednt insert asterisk_event.cdr_upload*/
		if (!strcasecmp(tableptr->table , "cdr_upload") && ((!pms_enable && !crm_enable )|| (!cdr_insert_success)))
		{
			ast_log(LOG_WARNING, "cdr_adaptive_odbc: PMS Disable and Crm Disable. Don't Insert cdr_upload\n");
			continue;
		}

		if (!strcasecmp(tableptr->table, "recording_files")) {
			ast_log(LOG_WARNING, "[odbc_log] No need handle recording_files");
			continue;
		}
#endif
		if (ast_strlen_zero(tableptr->schema)) {
			if (quoted) {
				ast_str_set(&sql, 0, "INSERT INTO %c%s%c (",
					tableptr->quoted_identifiers, tableptr->table, tableptr->quoted_identifiers );
			}
#ifdef GRANDSTREAM_NETWORKS
			else if (!strcasecmp(tableptr->table , "cdr_upload"))
				ast_str_set(&sql, 0, "INSERT INTO %s (AcctId,", tableptr->table);
#endif
			else{
				ast_str_set(&sql, 0, "INSERT INTO %s (", tableptr->table);
			}
		} else {
			if (quoted) {
				ast_str_set(&sql, 0, "INSERT INTO %c%s%c.%c%s%c (",
						tableptr->quoted_identifiers, tableptr->schema, tableptr->quoted_identifiers,
						tableptr->quoted_identifiers, tableptr->table,  tableptr->quoted_identifiers);
			}
#ifdef GRANDSTREAM_NETWORKS
			else if (!strcasecmp(tableptr->table , "cdr_upload"))
				ast_str_set(&sql, 0, "INSERT INTO %s (AcctId,", tableptr->table);
#endif
			else{
				ast_str_set(&sql, 0, "INSERT INTO %s.%s (", tableptr->schema, tableptr->table);
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		if (!strcasecmp(tableptr->table , "cdr_upload"))
			ast_str_set(&sql2, 0, " VALUES (%ld,",acctid);
		else
#endif
		ast_str_set(&sql2, 0, " VALUES (");

		/* No need to check the connection now; we'll handle any failure in prepare_and_execute */
		if (!(obj = ast_odbc_request_obj(tableptr->connection, 0))) {
			ast_log(LOG_WARNING, "cdr_adaptive_odbc: Unable to retrieve database handle for '%s:%s'.  CDR failed: %s\n", tableptr->connection, tableptr->table, ast_str_buffer(sql));
			continue;
		}

		AST_LIST_TRAVERSE(&(tableptr->columns), entry, list) {
			int datefield = 0;
			if (strcasecmp(entry->cdrname, "start") == 0) {
				datefield = 1;
			} else if (strcasecmp(entry->cdrname, "answer") == 0) {
				datefield = 2;
			} else if (strcasecmp(entry->cdrname, "end") == 0) {
				datefield = 3;
			}

			/* Check if we have a similarly named variable */
			if (entry->staticvalue) {
				colptr = ast_strdupa(entry->staticvalue);
			} else if (datefield && tableptr->usegmtime) {
				struct timeval date_tv = (datefield == 1) ? cdr->start : (datefield == 2) ? cdr->answer : cdr->end;
				struct ast_tm tm = { 0, };
				ast_localtime(&date_tv, &tm, "UTC");
				ast_strftime(colbuf, sizeof(colbuf), "%Y-%m-%d %H:%M:%S", &tm);
				colptr = colbuf;
			}
#ifdef GRANDSTREAM_NETWORKS
			else if (!strcasecmp(entry->cdrname, "pms_flag")){
				ast_str_append(&sql, 0, ",%s", "pms_flag");
				LENGTHEN_BUF2(12);
				ast_str_append(&sql2, 0, ",%d", pms_enable);
				continue;
			} else if (!strcasecmp(entry->cdrname, "crm_flag")){
				ast_str_append(&sql, 0, ",%s", "crm_flag");
				LENGTHEN_BUF2(12);
				ast_str_append(&sql2, 0, ",%d", crm_enable);
				continue;
			}
#endif
			else {
				ast_cdr_format_var(cdr, entry->cdrname, &colptr, colbuf, sizeof(colbuf), datefield ? 0 : 1);
			}

			if (colptr) {
				/* Check first if the column filters this entry.  Note that this
				 * is very specifically NOT ast_strlen_zero(), because the filter
				 * could legitimately specify that the field is blank, which is
				 * different from the field being unspecified (NULL). */
				if ((entry->filtervalue && !entry->negatefiltervalue && strcasecmp(colptr, entry->filtervalue) != 0) ||
					(entry->filtervalue && entry->negatefiltervalue && strcasecmp(colptr, entry->filtervalue) == 0)) {
					ast_verb(4, "CDR column '%s' with value '%s' does not match filter of"
						" %s'%s'.  Cancelling this CDR.\n",
						entry->cdrname, colptr, entry->negatefiltervalue ? "!" : "", entry->filtervalue);
					goto early_release;
				}

				/* Only a filter? */
				if (ast_strlen_zero(entry->name))
					continue;

				LENGTHEN_BUF1(strlen(entry->name));
				switch (entry->type) {
				case SQL_CHAR:
				case SQL_VARCHAR:
				case SQL_LONGVARCHAR:
#ifdef HAVE_ODBC_WCHAR
				case SQL_WCHAR:
				case SQL_WVARCHAR:
				case SQL_WLONGVARCHAR:
#endif
				case SQL_BINARY:
				case SQL_VARBINARY:
				case SQL_LONGVARBINARY:
				case SQL_GUID:
					/* For these two field names, get the rendered form, instead of the raw
					 * form (but only when we're dealing with a character-based field).
					 */
					if (strcasecmp(entry->name, "disposition") == 0) {
						ast_cdr_format_var(cdr, entry->name, &colptr, colbuf, sizeof(colbuf), 0);
					} else if (strcasecmp(entry->name, "amaflags") == 0) {
						ast_cdr_format_var(cdr, entry->name, &colptr, colbuf, sizeof(colbuf), 0);
					}

					/* Truncate too-long fields */
					if (entry->type != SQL_GUID) {
						if (strlen(colptr) > entry->octetlen) {
							colptr[entry->octetlen] = '\0';
						}
					}

					LENGTHEN_BUF2(strlen(colptr));

					/* Encode value, with escaping */
					ast_str_append(&sql2, 0, "%s'", separator);
#ifdef GRANDSTREAM_NETWORKS
					size_t i = 0;
					size_t j = 0;

					if (!ast_utf8_check_string(colptr, strlen(colptr))) {
						ast_verb(4, "The string %s use invalid utf8 char, remov it.\n", colptr);
						for (tmp = colptr; *tmp; tmp++, i++) {
							if (!ast_utf8_check_string(colptr, i+1)) {
								continue;
							} else {
								j = i+1;
							}
						}
						colptr[j] = '\0';
					}
#endif
					for (tmp = colptr; *tmp; tmp++) {
						if (*tmp == '\'') {
							ast_str_append(&sql2, 0, "''");
						} else if (*tmp == '\\' && ast_odbc_backslash_is_escape(obj)) {
							ast_str_append(&sql2, 0, "\\\\");
						} else {
							ast_str_append(&sql2, 0, "%c", *tmp);
						}
					}
					ast_str_append(&sql2, 0, "'");
					break;
				case SQL_TYPE_DATE:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						int year = 0, month = 0, day = 0;
						if (sscanf(colptr, "%4d-%2d-%2d", &year, &month, &day) != 3 || year <= 0 ||
							month <= 0 || month > 12 || day < 0 || day > 31 ||
							((month == 4 || month == 6 || month == 9 || month == 11) && day == 31) ||
							(month == 2 && year % 400 == 0 && day > 29) ||
							(month == 2 && year % 100 == 0 && day > 28) ||
							(month == 2 && year % 4 == 0 && day > 29) ||
							(month == 2 && year % 4 != 0 && day > 28)) {
							ast_log(LOG_WARNING, "CDR variable %s is not a valid date ('%s').\n", entry->name, colptr);
							continue;
						}

						if (year > 0 && year < 100) {
							year += 2000;
						}

						LENGTHEN_BUF2(17);
						ast_str_append(&sql2, 0, "%s{ d '%04d-%02d-%02d' }", separator, year, month, day);
					}
					break;
				case SQL_TYPE_TIME:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						int hour = 0, minute = 0, second = 0;
						int count = sscanf(colptr, "%2d:%2d:%2d", &hour, &minute, &second);

						if ((count != 2 && count != 3) || hour < 0 || hour > 23 || minute < 0 || minute > 59 || second < 0 || second > 59) {
							ast_log(LOG_WARNING, "CDR variable %s is not a valid time ('%s').\n", entry->name, colptr);
							continue;
						}

						LENGTHEN_BUF2(15);
						ast_str_append(&sql2, 0, "%s{ t '%02d:%02d:%02d' }", separator, hour, minute, second);
					}
					break;
				case SQL_TYPE_TIMESTAMP:
				case SQL_TIMESTAMP:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						int year = 0, month = 0, day = 0, hour = 0, minute = 0, second = 0;
						int count = sscanf(colptr, "%4d-%2d-%2d %2d:%2d:%2d", &year, &month, &day, &hour, &minute, &second);

						if ((count != 3 && count != 5 && count != 6) || year <= 0 ||
							month <= 0 || month > 12 || day < 0 || day > 31 ||
							((month == 4 || month == 6 || month == 9 || month == 11) && day == 31) ||
							(month == 2 && year % 400 == 0 && day > 29) ||
							(month == 2 && year % 100 == 0 && day > 28) ||
							(month == 2 && year % 4 == 0 && day > 29) ||
							(month == 2 && year % 4 != 0 && day > 28) ||
							hour > 23 || minute > 59 || second > 59 || hour < 0 || minute < 0 || second < 0) {
							ast_log(LOG_WARNING, "CDR variable %s is not a valid timestamp ('%s').\n", entry->name, colptr);
							continue;
						}

						if (year > 0 && year < 100) {
							year += 2000;
						}

						LENGTHEN_BUF2(26);
						ast_str_append(&sql2, 0, "%s{ ts '%04d-%02d-%02d %02d:%02d:%02d' }", separator, year, month, day, hour, minute, second);
					}
					break;
				case SQL_INTEGER:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						int integer = 0;
						if (sscanf(colptr, "%30d", &integer) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an integer.\n", entry->name);
							continue;
						}

						LENGTHEN_BUF2(12);
						ast_str_append(&sql2, 0, "%s%d", separator, integer);
					}
					break;
				case SQL_BIGINT:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						long long integer = 0;
						if (sscanf(colptr, "%30lld", &integer) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an integer.\n", entry->name);
							continue;
						}

						LENGTHEN_BUF2(24);
						ast_str_append(&sql2, 0, "%s%lld", separator, integer);
					}
					break;
				case SQL_SMALLINT:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						short integer = 0;
						if (sscanf(colptr, "%30hd", &integer) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an integer.\n", entry->name);
							continue;
						}

						LENGTHEN_BUF2(6);
						ast_str_append(&sql2, 0, "%s%d", separator, integer);
					}
					break;
				case SQL_TINYINT:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						signed char integer = 0;
						if (sscanf(colptr, "%30hhd", &integer) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an integer.\n", entry->name);
							continue;
						}

						LENGTHEN_BUF2(4);
						ast_str_append(&sql2, 0, "%s%d", separator, integer);
					}
					break;
				case SQL_BIT:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						signed char integer = 0;
						if (sscanf(colptr, "%30hhd", &integer) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an integer.\n", entry->name);
							continue;
						}
						if (integer != 0)
							integer = 1;

						LENGTHEN_BUF2(2);
						ast_str_append(&sql2, 0, "%s%d", separator, integer);
					}
					break;
				case SQL_NUMERIC:
				case SQL_DECIMAL:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						double number = 0.0;

						if (!strcasecmp(entry->cdrname, "billsec")) {
							if (!ast_tvzero(cdr->answer)) {
								snprintf(colbuf, sizeof(colbuf), "%lf",
											(double) (ast_tvdiff_us(cdr->end, cdr->answer) / 1000000.0));
							} else {
								ast_copy_string(colbuf, "0", sizeof(colbuf));
							}
						} else if (!strcasecmp(entry->cdrname, "duration")) {
							snprintf(colbuf, sizeof(colbuf), "%lf",
										(double) (ast_tvdiff_us(cdr->end, cdr->start) / 1000000.0));

							if (!ast_strlen_zero(colbuf)) {
								colptr = colbuf;
							}
						}

						if (sscanf(colptr, "%30lf", &number) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an numeric type.\n", entry->name);
							continue;
						}

						LENGTHEN_BUF2(entry->decimals);
						ast_str_append(&sql2, 0, "%s%*.*lf", separator, entry->decimals, entry->radix, number);
					}
					break;
				case SQL_FLOAT:
				case SQL_REAL:
				case SQL_DOUBLE:
					if (ast_strlen_zero(colptr)) {
						continue;
					} else {
						double number = 0.0;

						if (!strcasecmp(entry->cdrname, "billsec")) {
							if (!ast_tvzero(cdr->answer)) {
								snprintf(colbuf, sizeof(colbuf), "%lf",
											(double) (ast_tvdiff_us(cdr->end, cdr->answer) / 1000000.0));
							} else {
								ast_copy_string(colbuf, "0", sizeof(colbuf));
							}
						} else if (!strcasecmp(entry->cdrname, "duration")) {
							snprintf(colbuf, sizeof(colbuf), "%lf",
										(double) (ast_tvdiff_us(cdr->end, cdr->start) / 1000000.0));

							if (!ast_strlen_zero(colbuf)) {
								colptr = colbuf;
							}
						}

						if (sscanf(colptr, "%30lf", &number) != 1) {
							ast_log(LOG_WARNING, "CDR variable %s is not an numeric type.\n", entry->name);
							continue;
						}

						LENGTHEN_BUF2(entry->decimals);
						ast_str_append(&sql2, 0, "%s%lf", separator, number);
					}
					break;
				default:
					ast_log(LOG_WARNING, "Column type %d (field '%s:%s:%s') is unsupported at this time.\n", entry->type, tableptr->connection, tableptr->table, entry->name);
					continue;
				}
				if (quoted) {
					ast_str_append(&sql, 0, "%s%c%s%c", separator, tableptr->quoted_identifiers, entry->name, tableptr->quoted_identifiers);
				} else {
					ast_str_append(&sql, 0, "%s%s", separator, entry->name);
				}
				separator = ", ";
			} else if (entry->filtervalue
				&& ((!entry->negatefiltervalue && entry->filtervalue[0] != '\0')
					|| (entry->negatefiltervalue && entry->filtervalue[0] == '\0'))) {
				ast_log(AST_LOG_WARNING, "CDR column '%s' was not set and does not match filter of"
					" %s'%s'.  Cancelling this CDR.\n",
					entry->cdrname, entry->negatefiltervalue ? "!" : "",
					entry->filtervalue);
				goto early_release;
			}
		}

		/* Concatenate the two constructed buffers */
		LENGTHEN_BUF1(ast_str_strlen(sql2));
#ifdef GRANDSTREAM_NETWORKS
		//strcpy(cdr_value,ast_str_buffer(sql2)+9);//get cdr values;
		ast_copy_string(cdr_value, strstr(ast_str_buffer(sql2),"(")+1, sizeof(cdr_value) - 1);//get cdr values;
		ast_copy_string(cdr_column, strstr(ast_str_buffer(sql),"(")+1, sizeof(cdr_column) - 1);//get cdr_coliumn values;
#endif
		ast_str_append(&sql, 0, ")");
		ast_str_append(&sql2, 0, ")");
		ast_str_append(&sql, 0, "%s", ast_str_buffer(sql2));

		ast_debug(3, "Executing [%s]\n", ast_str_buffer(sql));

		if(HA_REDUNDANCE_SECOND_TRANSIT == ast_ha_get_redundance_state()
			|| (1 == ast_get_redundance_flag() && 1 == ast_ha_get_on_flag()
			  && AST_HA_ROLE_BACKUP == ast_ha_get_role() 
			  && HA_REDUNDANCE_SECOND_IDLE == ast_ha_get_redundance_state()) ){ 
			ast_ha_send_cdr_sql(ast_str_buffer(sql));
			ast_debug(3,"i am slave now, send cdr sql to peer for insert db");
			if (-1 != cdr->main_disposition && 0 == strcasecmp(tableptr->table, "cdr")){
				snprintf(main_sql, sizeof(main_sql), "REPLACE INTO main_count(whole_session,whole_duration,whole_billsec,whole_disposition) VALUES('%s','%ld','%ld','%s')", cdr->session, cdr->main_duration, cdr->main_billsec, ast_cdr_disp2str(cdr->main_disposition));
				ast_log(LOG_DEBUG, "%s", main_sql);
				ast_ha_send_cdr_sql(main_sql);
			}
		}
		else{
			stmt = ast_odbc_prepare_and_execute(obj, generic_prepare, ast_str_buffer(sql));
			if (stmt) {
				SQLRowCount(stmt, &rows);
				SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			}
			if (rows == 0) {
				ast_log(LOG_WARNING, "cdr_adaptive_odbc: Insert failed on '%s:%s'.  CDR failed: %s\n", tableptr->connection, tableptr->table, ast_str_buffer(sql));
			}
	/* start added by zyxia */
#ifdef GRANDSTREAM_NETWORKS
			else {
				/* if the table is master.cdr,need do it*/
				if (!strcasecmp(tableptr->table , "cdr"))
				{
					cdr_insert_success = 1;
					if (ast_strlen_zero(tableptr->schema)) {
						ast_str_set(&sql_rowid,0, "SELECT MAX(AcctId) FROM %s ", tableptr->table);
					} else {
						ast_str_set(&sql_rowid,0, "SELECT MAX(AcctId) FROM %s.%s ", tableptr->schema, tableptr->table);
					}

					stmt = ast_odbc_prepare_and_execute(obj, generic_prepare, ast_str_buffer(sql_rowid));
					if (!stmt) {
						ast_log(AST_LOG_WARNING, "SQL Execute error!\n[%s]\n\n", ast_str_buffer(sql_rowid));
						goto early_release;
					}
					res = SQLFetch(stmt);
					if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
						ast_log(AST_LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", ast_str_buffer(sql_rowid));
						SQLFreeHandle (SQL_HANDLE_STMT, stmt);
						goto early_release;
					}
					res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
					if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
						ast_log(AST_LOG_WARNING, "SQL Get Data error!\n[%s]\n\n", ast_str_buffer(sql_rowid));
						SQLFreeHandle (SQL_HANDLE_STMT, stmt);
						goto early_release;
					}
					acctid = atoi(rowdata);
					SQLFreeHandle (SQL_HANDLE_STMT, stmt);

					char tmp[64] = {0};
					/* Added by yfyang, for trunk calltime limit */
					if (!ast_tvzero(cdr->answer)) {
						bill_sec = (ast_tvdiff_us(cdr->end, cdr->answer) / 1000000.0);
						if ((bill_sec > 0) && !ast_strlen_zero(cdr->dstchannel) && (strcasestr(cdr->dstchannel, "trunk_") || strcasestr(cdr->dstchannel, "DAHDI/o"))) {
							update_trunk_time();  /* To set tklimit update */
							if (ha_enable){   /*Add by weidu ,for 278611*/
								ast_ha_send_update_limit_time("update limit time");
							}
						}
					}

					if (ha_enable)
					{
						char ha_sql[PATH_MAX];
						snprintf(ha_sql, sizeof(ha_sql),"INSERT INTO %s (AcctId, %s) VALUES (%ld, %s)", tableptr->table, cdr_column,acctid, cdr_value);
						ast_debug(1, "sql is [%s].", ha_sql);
						//ast_ha_send_cdr_sql(ha_sql);
					}
					if (cdr_callspro_enable) {
						if (access(FIFO_CALLSPRO, F_OK) != -1) { /* wait entail pbxmid create the FIFO */
							if (fifo_callspro_fd <= 0) {
								/* fifo open will fail when write point with O_NONBLOCK and read point not open ahead of write point */
								if ((fifo_callspro_fd = open(FIFO_CALLSPRO, O_WRONLY | O_NONBLOCK)) == -1) {
									ast_log(LOG_DEBUG, "Open FIFO '%s' failed, pass write fifo\n", FIFO_CALLSPRO);
								} else {
									ast_debug(1, "Open FIFO success\n");
								}
								}
							if (fifo_callspro_fd > 0) {
								snprintf(tmp, sizeof(tmp), "%ld", acctid);
								if (write(fifo_callspro_fd, tmp, sizeof(tmp)) > 0) {
									ast_debug(1, "Write fifo success, acctid is %s\n", tmp);
								} else {
									ast_log(LOG_DEBUG, "Write into fifo failed, pass write fifo\n");
								}
							}
						} else {
							ast_log(LOG_DEBUG, "FIFO '%s' not exist, pass write fifo\n", FIFO_CALLSPRO);
						}
					}
					/*here add the missed call send*/
					if (AST_RWLIST_WRLOCK(&global_extlists)) {
						ast_log(LOG_ERROR, "Unable to lock missed ext list.  Load failed.\n");
					}
					ast_localtime(&cdr->start, &tm, NULL);
					ast_strftime(start_time, sizeof(start_time), "%Y-%m-%d %H:%M:%S", &tm);
					char inbound_dst[AST_MAX_EXTENSION] = {0};
					if (strcasestr(cdr->channel, "trunk_") && !ast_strlen_zero(cdr->dstchannel) && !strcasestr(cdr->dstchannel, "trunk_")){
						sscanf(cdr->dstchannel, "%*[^/]/%[^-]", inbound_dst);
					}
					ast_log(LOG_DEBUG, "src:%s dst:%s clid:%s time:%s status:%ld action_type:%s", cdr->src, strlen(inbound_dst)?inbound_dst:cdr->dst, cdr->clid, start_time, cdr->disposition, cdr->action_type);
					//src:4005 dst:4562007 clid:"" <4005> time:2020-01-13 17:28:21 status:0 action_type:DIAL
					int concurrent_num = contacts_number(cdr->lastdata);
					if ((AST_CDR_ANSWERED != cdr->disposition || (AST_CDR_ANSWERED == cdr->disposition && !strcasecmp(cdr->action_type, "VM")))
						&& '\0' != cdr->dst[0]
						&& 0 != strncmp(cdr->action_type, "RINGGROUP", strlen("RINGGROUP"))
						&& 0 != strncmp(cdr->action_type, "QUEUE", strlen("QUEUE"))
						&& 0 != strncmp(cdr->special_action_type, "DO_DND", strlen("DO_DND"))
						&& !(strcasestr(cdr->channel, "trunk_") && !strncmp(cdr->lastapp, "ForkCDR", strlen("ForkCDR")))){
						struct exten_list *missed_extlist = NULL;
						struct exten_node *entry;
						AST_LIST_TRAVERSE(&global_extlists, missed_extlist, list) {
							AST_LIST_TRAVERSE(&(missed_extlist->exten_node), entry, list) {
								snprintf(tmp_pjsip_chan_head, sizeof(tmp_pjsip_chan_head), "PJSIP/%s", entry->extension);
								if ((0 == strcmp(entry->extension, strlen(inbound_dst)?inbound_dst:cdr->dst)
										|| 0 == strncmp(cdr->lastdata, tmp_pjsip_chan_head, strlen(tmp_pjsip_chan_head)))
									&& (cdr->concurrent == concurrent_num
										|| !strcasecmp(cdr->action_type, "VM")
										|| (strcasestr(cdr->special_action_type, "ringboth") && cdr->concurrent == concurrent_num-1 && !strcasestr(cdr->dstchannel, "trunk_"))
										|| (AST_CDR_ANSWERED != cdr->disposition && !strncmp(cdr->action_type, "FOLLOWME", strlen("FOLLOWME"))))){
									if (!strcmp(entry->enabled, "all")
										|| ((!strcmp(cdr->userfield, "External") || !strcmp(cdr->userfield, "Inbound")) && !strcmp(entry->enabled, "out"))
										|| (!strcmp(cdr->userfield, "Internal") && !strcmp(entry->enabled, "in"))){
										need_send=1;
										snprintf(inbound_dst, sizeof(inbound_dst), "%s", entry->extension);
										break;
									}
								}
							}
						}

						if (AST_CDR_ANSWERED != cdr->disposition && cdr->next
							&& (!strcasecmp(cdr->next->action_type, "VM") && strncmp(cdr->action_type, "FOLLOWME", strlen("FOLLOWME")))) {
							need_send = 0;
						}
						/*only for vm after followme failed*/
						if (-1 == cdr->concurrent){
							need_send = 0;
						}
					}
					AST_RWLIST_UNLOCK(&global_extlists);
					if (1 == need_send){
						need_send = 0;
						if (access(FIFO_MISSED_CALL, F_OK) != -1){
							if (missed_call_fd <= 0) {
								/* fifo open will fail when write point with O_NONBLOCK and read point not open ahead of write point */
								if ((missed_call_fd = open(FIFO_MISSED_CALL, O_WRONLY | O_NONBLOCK)) == -1) {
									ast_log(LOG_DEBUG, "Open FIFO '%s' failed, pass write fifo\n", FIFO_MISSED_CALL);
								} else {
									ast_debug(1, "Open FIFO success\n");
								}
							}
							if (missed_call_fd > 0) {
								strcpy(cdr_clid, cdr->clid);
								tmp_cdr_clid = cdr_clid;
								while(tmp_cdr_clid !=NULL)
								{
									if((tmp_cdr_clid = strchr(tmp_cdr_clid, '"')))
									{
										*tmp_cdr_clid = '\'';
										tmp_cdr_clid++;
									}
								}
								snprintf(send_buf, sizeof(send_buf), "%s@%s@%s@%s", cdr->src, strlen(inbound_dst)?inbound_dst:cdr->dst, cdr_clid, start_time);
								ast_log(LOG_DEBUG, "send_buf:%s", send_buf);
								if (write(missed_call_fd, send_buf, sizeof(send_buf)) > 0) {
									ast_debug(1, "Write fifo success\n");
								} else {
									ast_log(LOG_DEBUG, "Write into fifo failed, pass write fifo\n");
								}
							}
						}
						else {
							ast_log(LOG_DEBUG, "FIFO '%s' not exist, pass write fifo\n", FIFO_MISSED_CALL);
						}
					}
				}
				if (!strcasecmp(tableptr->table , "cdr_upload") && (pms_enable || crm_enable))
				{
					if (access(FIFO, F_OK) != -1) { /* wait entail pbxmid create the FIFO */
						if (fifo_fd <= 0) {
							/* fifo open will fail when write point with O_NONBLOCK and read point not open ahead of write point */
							if ((fifo_fd = open(FIFO, O_WRONLY | O_NONBLOCK)) == -1) {
								ast_log(LOG_DEBUG, "Open FIFO '%s' failed, pass write fifo\n", FIFO);
							} else {
								ast_debug(1, "Open FIFO success\n");
							}
						}
						if (fifo_fd > 0) {
							if (write(fifo_fd, "1", 1) > 0) {
								ast_debug(1, "Write fifo success\n");
							} else {
								ast_log(LOG_DEBUG, "Write into fifo failed, pass write fifo\n");
							}
						}
					} 
					else {
						ast_log(LOG_DEBUG, "FIFO '%s' not exist, pass write fifo\n", FIFO);
					}
				}
			}

			if (-1 != cdr->main_disposition && 0 == strcasecmp(tableptr->table, "cdr")){
				snprintf(main_sql, sizeof(main_sql), "REPLACE INTO main_count(whole_session,whole_duration,whole_billsec,whole_disposition) VALUES('%s','%ld','%ld','%s')", cdr->session, cdr->main_duration, cdr->main_billsec, ast_cdr_disp2str(cdr->main_disposition));
				ast_log(LOG_DEBUG, "%s", main_sql);
				stmt = ast_odbc_prepare_and_execute(obj, generic_prepare, main_sql);
				if (stmt) {
					SQLFreeHandle(SQL_HANDLE_STMT, stmt);
				}
			}

			if ( strlen(cdr->record_files) && strlen(cdr->record_files_uuid) && 
				 !strcasecmp(tableptr->table, "cdr") && 
				 rf_obj) {
				int need_record_key = 0;
				char record_files[4096] = "";
				char record_files_uuid[4096] = "";

				char *sep = "@";
    			char *record_files_token = NULL;
				char *record_files_uuid_token = NULL;
    			char *record_files_sub = NULL;
				char *record_files_uuid_sub = NULL;

				char *file_key = NULL;
				char *file_key_check = NULL;
				int has_record = 0;
				int isrecord = 0;

				snprintf(record_files, sizeof(record_files), "%s", cdr->record_files);
				snprintf(record_files_uuid, sizeof(record_files_uuid), "%s", cdr->record_files_uuid);

#ifdef BUILD_ON_CLOUD_PBX
				need_record_key = 1;
#else
				char settingss[128] = { 0 };
				int settingsi = 0x0;
				ast_safe_nvram_get("data_crypt_settings", settingss, sizeof(settingss) - 1);
				settingsi = atoi(settingss);

				if (strstr(record_files, ".wav"))
				{
					isrecord = 1;
				}
				need_record_key = strstr(record_files, ".wav") ? ast_btst(settingsi,AST_RECORD_AUDIO_FILE) : ast_btst(settingsi,AST_RECORD_VIDEO_FILE);
#endif // BUILD_ON_CLOUD_PBX

				record_files_token = strtok_r(record_files, sep, &record_files_sub);
				record_files_uuid_token = strtok_r(record_files_uuid, sep, &record_files_uuid_sub);
				while (record_files_token && record_files_uuid_token) {
					char record_files_without_dir[256] = "";
					char *pos = strstr(record_files_token, "/");
					if (pos) {
						snprintf(record_files_without_dir, sizeof(record_files_without_dir), "%s", ++pos);
					}

					ucm_secret_encrypt(record_files_uuid_token, &file_key);
					if (strlen(record_files_uuid_token) == 32) {
						ucm_secret_encrypt_with_key("grandstream", record_files_uuid_token, &file_key_check);
					}
					
					snprintf(main_sql, 
						sizeof(main_sql), 
						"SELECT count(*) AS cnt FROM recording_files WHERE file_name = '%s';", 
						strlen(record_files_without_dir) ? record_files_without_dir : record_files_token);
					stmt = ast_odbc_prepare_and_execute(rf_obj, generic_prepare, main_sql);
					if (!stmt) {
						ast_log(LOG_DEBUG, "cdr_adaptive_odbc failed to execute %s", main_sql);
						continue;
					}
					res = SQLFetch(stmt);
					if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
						ast_log(LOG_DEBUG, "cdr_adaptive_odbc failed to execute %s", main_sql);
						continue;
					}
					res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
					if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
						ast_log(LOG_DEBUG, "cdr_adaptive_odbc failed to execute %s", main_sql);
						SQLFreeHandle(SQL_HANDLE_STMT, stmt);
						continue;
					}
					int cnt = atoi(rowdata);
					SQLFreeHandle(SQL_HANDLE_STMT, stmt);

					if (!cnt) {
						snprintf(main_sql, 
							sizeof(main_sql), 
							"REPLACE INTO recording_files(file_name, path_type, feature_type, file_date, file_key, file_key_check) "
							"VALUES('%s', '%s', '%s', '%s', '%s', '%s');",
							strlen(record_files_without_dir) ? record_files_without_dir : record_files_token,
							"",
							"",
							"",
							(file_key && strlen(file_key)) ? file_key : "",
							(file_key_check && strlen(file_key_check)) ? file_key_check : "" 
						);
						has_record = 1;
					} else {
						snprintf(main_sql,
							sizeof(main_sql),
							"UPDATE recording_files SET file_key = '%s', file_key_check = '%s' WHERE file_name = '%s';",
							(file_key && strlen(file_key)) ? file_key : "",
							(file_key_check && strlen(file_key_check)) ? file_key_check : "",
							strlen(record_files_without_dir) ? record_files_without_dir : record_files_token
						);
					}
					ast_log(LOG_DEBUG, "%s", main_sql);

					stmt = ast_odbc_prepare_and_execute(rf_obj, generic_prepare, main_sql);
					if (stmt) {
						SQLFreeHandle(SQL_HANDLE_STMT, stmt);
					}

					if (file_key) {
						ast_free(file_key);
						file_key = NULL;
					}
					if (file_key_check) {
						ast_free(file_key_check);
						file_key_check = NULL;
					}

					record_files_token = strtok_r(NULL, sep, &record_files_sub);
					record_files_uuid_token = strtok_r(NULL, sep, &record_files_uuid_sub);
				}
				if (!need_record_key && has_record) {
#ifdef BUILD_ON_CLOUD_PBX
					char sqlprefix[64] = {0};
					ast_safe_nvram_get("sqlprefix", sqlprefix, sizeof(sqlprefix));
					if (!ast_strlen_zero(sqlprefix)) {
						snprintf(main_sql, sizeof(main_sql), "UPDATE %sucm_config.data_encrypt_details SET safety_factor='weak',status='' WHERE data_type='%s' AND safety_factor='strong' AND status='completed' AND device='CloudUCM'", sqlprefix, isrecord ? "record_file" : "video_file");
					}
#else
					snprintf(main_sql, sizeof(main_sql), "UPDATE ucm_config.data_encrypt_details SET safety_factor='weak',status='' WHERE data_type='%s' AND safety_factor='strong' AND status='completed' AND device='UCM'", isrecord ? "record_file" : "video_file");
#endif
					ast_log(LOG_DEBUG, "%s", main_sql);
					stmt = ast_odbc_prepare_and_execute(rf_obj, generic_prepare, main_sql);
					if (stmt) {
						SQLFreeHandle(SQL_HANDLE_STMT, stmt);
					}
				}
			}
		}
	
#endif
early_release:
		ast_odbc_release_obj(obj);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (rf_obj) {
		ast_odbc_release_obj(rf_obj);
	}
#endif

	AST_RWLIST_UNLOCK(&odbc_tables);

	/* Next time, just allocate buffers that are that big to start with. */
	if (ast_str_strlen(sql) > maxsize) {
		maxsize = ast_str_strlen(sql);
	}
	if (ast_str_strlen(sql2) > maxsize2) {
		maxsize2 = ast_str_strlen(sql2);
	}

	ast_free(sql);
	ast_free(sql2);
#ifdef GRANDSTREAM_NETWORKS
	ast_free(sql_rowid);
#endif
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS

static int odbc_tkcall_sum(struct ast_channel *chan, const char *cmd, char *data, char *buffer, size_t buflen)
{
	struct tables *tableptr;
	struct odbc_obj *obj;
	SQLHSTMT stmt = NULL;
	char rowdata[20] = {0};
	int res;

	if (!data || !buffer || !buflen) {
		ast_log(LOG_WARNING, "Invalid para in func(%s).\n", cmd);
		return -1;
	}

	AST_RWLIST_RDLOCK(&odbc_tables);
	AST_RWLIST_TRAVERSE(&odbc_tables, tableptr, list) {
		if (!strcasecmp(tableptr->table, "cdr")) {
			break;
		}
	}

	if (!tableptr) {
		AST_RWLIST_UNLOCK(&odbc_tables);
		ast_log(LOG_DEBUG, "odbc_tkcall_sum, fail to get tableptr\n");
		return -1;
	}

	obj = ast_odbc_request_obj(tableptr->connection, 0);
	if (!obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for [%s]!\n", tableptr->connection);
		AST_RWLIST_UNLOCK(&odbc_tables);
		return -1;
	}

	ast_log(LOG_DEBUG, "odbc_tkcall_sum, data[%s]\n", data);

	stmt = ast_odbc_prepare_and_execute(obj, generic_prepare, data);
	if (!stmt) {
		ast_log(AST_LOG_WARNING, "SQL Execute error!\n\n");
		ast_odbc_release_obj(obj);
		AST_RWLIST_UNLOCK(&odbc_tables);
		return 0;
	}

	res = SQLFetch(stmt);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		ast_log(AST_LOG_WARNING, "SQL Fetch error!\n");
		goto bail;
	}
	res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if ((res != SQL_SUCCESS) && (res != SQL_SUCCESS_WITH_INFO)) {
		goto bail;
	}

	ast_log(AST_LOG_DEBUG, "SQL Get Data [%s]!\n", rowdata);
	ast_copy_string(buffer, rowdata, buflen);   

bail:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	ast_odbc_release_obj(obj);
	AST_RWLIST_UNLOCK(&odbc_tables);
	return 0;
}

static struct ast_custom_function tktime_sum_func = {
	.name = "TRUNK_TIME_SUM",
	.read = odbc_tkcall_sum,
};

#endif

static int unload_module(void)
{
	if (ast_cdr_unregister(name)) {
		return -1;
	}

	if (AST_RWLIST_WRLOCK(&odbc_tables)) {
		ast_cdr_register(name, ast_module_info->description, odbc_log);
		ast_log(LOG_ERROR, "Unable to lock column list.  Unload failed.\n");
		return -1;
	}
#ifdef GRANDSTREAM_NETWORKS
	if (AST_RWLIST_WRLOCK(&global_extlists)) {
		ast_log(LOG_ERROR, "Unable to lock missed ext list.  Load failed.\n");
		return 0;
	}
	unload_missed();
	AST_RWLIST_UNLOCK(&global_extlists);

	ast_custom_function_unregister(&tktime_sum_func);
#endif
	free_config();
	AST_RWLIST_UNLOCK(&odbc_tables);

	ast_cdr_sql_unregister();
	return 0;
}

static int load_module(void)
{
	if (AST_RWLIST_WRLOCK(&odbc_tables)) {
		ast_log(LOG_ERROR, "Unable to lock column list.  Load failed.\n");
		return 0;
	}
#ifdef GRANDSTREAM_NETWORKS
	if (AST_RWLIST_WRLOCK(&global_extlists)) {
		ast_log(LOG_ERROR, "Unable to lock missed ext list.  Load failed.\n");
		return 0;
	}
	load_missed();
	AST_RWLIST_UNLOCK(&global_extlists);
#endif

	load_config();
	AST_RWLIST_UNLOCK(&odbc_tables);
	ast_cdr_register(name, ast_module_info->description, odbc_log);
	ast_cdr_sql_register(ast_odbc_insert_cdr);

#ifdef GRANDSTREAM_NETWORKS
	ast_custom_function_register(&tktime_sum_func);
#endif

	return 0;
}

static int reload(void)
{
	if (AST_RWLIST_WRLOCK(&odbc_tables)) {
		ast_log(LOG_ERROR, "Unable to lock column list.  Reload failed.\n");
		return -1;
	}

	free_config();
	load_config();
#ifdef GRANDSTREAM_NETWORKS
	if (AST_RWLIST_WRLOCK(&global_extlists)) {
		ast_log(LOG_ERROR, "Unable to lock missed ext list.  Load failed.\n");
		return 0;
	}
	unload_missed();
	load_missed();
	AST_RWLIST_UNLOCK(&global_extlists);
#endif
	AST_RWLIST_UNLOCK(&odbc_tables);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Adaptive ODBC CDR backend",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
	.load_pri = AST_MODPRI_CDR_DRIVER,
	.requires = "cdr,res_odbc",
);
