/*!
 * \file
 * \brief Implementation of Real Time Streaming Protocol
 *
 * \author zyxia <zyxia@grandstream.cn>
 *
 */

/*** MODULEINFO
	<depend>ucmsecret</depend>
	<support_level>extended</support_level>
 ***/

#include "asterisk.h"

#include <fcntl.h>
#include <signal.h>
#include <sys/signal.h>
#include <inttypes.h>

#include "asterisk/network.h"
#include "asterisk/lock.h"
#include "asterisk/config.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/sched.h"
#include "asterisk/io.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/acl.h"
#include "asterisk/callerid.h"
#include "asterisk/cli.h"
#include "asterisk/causes.h"
#include "asterisk/utils.h"
#include "asterisk/astobj2.h"
#include "asterisk/translate.h"
#include <asterisk/format_compatibility.h>
#include "asterisk/format.h"
#include "asterisk/ast_version.h"
#include "chan_rtsp.h"
#include "asterisk/avs_controller.h"
#include "../apps/confbridge/include/conf_resource.h"
#ifdef HAVE_UCMSECRET
#include "ucm_crypt.h"
#endif
AST_MODULE_LOG("chan_rtsp");
#define SAFE_FREE(buf) if (buf) {ast_free(buf);	buf = NULL;}

static int usecnt = 0;
AST_MUTEX_DEFINE_STATIC(usecnt_lock);

static unsigned int chan_idx;

static struct ast_sockaddr bindaddr;

struct ast_sched_context *sched;     /*!< The scheduling context */
static struct io_context *io;           /*!< The IO context */

/*! \brief list of users found in the config file */
static AST_RWLIST_HEAD_STATIC(rtsp_users, ast_rtsp_user);
static char *config = "chan_rtsp.conf";

/*Functions*/
static int rtsp_call(struct ast_channel *chan, const char *data, int timeout);
static struct ast_channel *rtsp_request(const char *type, struct ast_format_cap *cap, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *addr, int *cause);
static int rtsp_hangup(struct ast_channel *chan);
static enum ast_terminal_type chan_rtsp_get_terminal_type(struct ast_channel *ast);
static int rtsp_write(struct ast_channel *ast, struct ast_frame *frame);
static int rtsp_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen);
static int rtsp_keepalive_cb(const void *data);
static void* ast_rtsp_keepalive_thread(void *data);


/**
 ** Asterisk structs for ch registration
 **/
/*! \brief Definition of this channel for PBX channel registration */
static struct ast_channel_tech rtsp_tech = {
	.type = "RTSP",
	.description = "Real Time Streaming Protocol (RTSP)",
	.properties = AST_CHAN_TP_WANTSJITTER,
	.requester = rtsp_request,
	.call = rtsp_call,				/* called with chan locked */
	.hangup = rtsp_hangup,			/* called with chan locked */
	.get_terminal_type = chan_rtsp_get_terminal_type,
	.write = rtsp_write,			/* called with chan locked */
	.indicate = rtsp_indicate,		/* called with chan locked */

#if 0 
	.answer = rtsp_answer,			/* called with chan locked */
	.read = rtsp_read,				/* called with chan locked */
	.write = rtsp_write,			/* called with chan locked */
	.write_video = rtsp_write,		/* called with chan locked */

	.setoption = rtsp_setoption,
#endif
};

static const char *ipcstatus_tostr(const int type)
{
	switch (type) {
	case RTSP_UNMONITORNED:
		return "Unmonitored";
	case RTSP_REACHABLE:
		return "Reachable";
	case RTSP_UNREACHABLE:
		return "Unreachable";
	}
	return "Unmonitored";
}

//----------------------------------------------------------------------------------------------------------------
static int CreateMedia(const char *mline, int media, struct ast_rtp_codecs *codecs,unsigned int *first_payload)
{
	int len;
	unsigned int x = 0;
	unsigned int numberofports;
	char protocol[18] = "";	
	const char *payloads;
	unsigned int payload;
	unsigned int find =0;

	if ((sscanf(mline, "m=audio %30u/%30u %17s %n", &x, &numberofports, protocol, &len) == 3 && len > 0) ||
		(sscanf(mline, "m=audio %30u %17s %n", &x, protocol, &len) == 2 && len > 0) ||
		(sscanf(mline, "m=video %30u/%30u %17s %n", &x, &numberofports, protocol, &len) == 3 && len > 0) ||
		(sscanf(mline, "m=video %30u %17s %n", &x, protocol, &len) == 2 && len > 0)) {
		payloads = mline + len;
	} else {
		ast_log(LOG_WARNING, "Rejecting media offer due to invalid or unsupported syntax: %s\n", mline);
		return 0;
	}

	/* For each format */
	/* Scan through the RTP payload types specified in a "m=" line: */
	for (; !ast_strlen_zero(payloads); payloads = ast_skip_blanks(payloads + len)) {
		if (sscanf(payloads, "%30u%n", &payload, &len) != 1) {
			ast_log(LOG_WARNING, "Invalid syntax in RTP media payload list: %s\n", payloads);
			return 0;
		}
		if (!find) {
			*first_payload = payload;
			find = 1;
		}
		ast_verb(1, "[rtsp] Found RTP media id%d payload %u\n", media, payload);
		ast_rtp_codecs_payloads_set_m_type(codecs, NULL, payload);
	}

	/* Return media */
	return 1;
}


static enum video_resolution ast_rtsp_get_main_video_resource(const char* conf_number,struct ast_channel *chan, unsigned int h264_level)
{
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;

	if (h264_level >= AST_1080P_LEVEL_ID) {
		resolution = ast_conf_get_main_video_resource(conf_number, chan, AST_STREAM_VIDEO_RESOLUTION_1080P);
	} else if (h264_level >= AST_720P_LEVEL_ID && h264_level < AST_1080P_LEVEL_ID) {
		resolution = ast_conf_get_main_video_resource(conf_number, chan, AST_STREAM_VIDEO_RESOLUTION_720P);
	} else {
		resolution = ast_conf_get_main_video_resource(conf_number, chan, AST_STREAM_VIDEO_RESOLUTION_360P);
	}

	return resolution;
}

static  int set_h264_resolution_for_rtsp(struct ast_channel * chan,
	struct ast_format *format, struct ast_stream *asterisk_stream)
{
	int ret = R_FAIL;
	enum ast_stream_video_resolution obtain_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;
	const unsigned int *h264_level = NULL;

	h264_level = ast_format_attribute_get(format, "level");
	if (!h264_level) {
		ast_log(LOG_DEBUG, "Can't get H264 level, can't open ipc chan:%s\n",ast_channel_name(chan));
		return ret;
	}

	
	const char* conf_room  = pbx_builtin_getvar_helper((struct ast_channel *)chan, "CONF_ROOM");
	if (conf_room) {
		ret = R_SUCCESS;
		obtain_video_resolution = ast_rtsp_get_main_video_resource(conf_room, chan,*h264_level);
		ast_log(LOG_DEBUG, "[chan:%s] Set H264 level %d, conf_room:%s,resolution:%d\n",ast_channel_name(chan),*h264_level,conf_room,obtain_video_resolution);

		if ((obtain_video_resolution <= AST_STREAM_VIDEO_RESOLUTION_NONE) || (obtain_video_resolution > AST_STREAM_VIDEO_RESOLUTION_1080P)) {
			ret = R_FAIL;
			if (obtain_video_resolution == -2) {
				ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_CPU_LIMIT);
			} else {
				ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_RESOURCE);
			}
		}
	}

	return ret;
}

//static int parseSDP(struct ast_channel *chan, char *buffer, int bufferLen)
static int parseSDP(struct ast_channel *chan, struct avs_play_rtsp_info* rtsp_info)
{
	int res = R_SUCCESS;
	char *i = rtsp_info->sdp_info;
	char *j = NULL;
	char mline[256];
	int media = -1;

	char mslabel[AST_UUID_STR_LEN];
	/*! \brief Track label */
	char label[AST_UUID_STR_LEN];

	// parse fmtp
	unsigned int audio_payload = 0;
	unsigned int video_payload = 0;
	unsigned int payload =0;
	char mimeSubtype[128];
	char fmtp_string[256];
	unsigned int sample_rate;
	long profile_level_id = 0;

	ast_log(LOG_NOTICE, "[rtsp] parseSDP 111\n");

	// parse m=
	struct ast_rtp_codecs newaudiortp = AST_RTP_CODECS_NULL_INIT;
	struct ast_rtp_codecs newvideortp = AST_RTP_CODECS_NULL_INIT;

	struct ast_stream_topology *topology = NULL;
	struct ast_stream *video_stream = NULL;
	struct ast_stream *audio_stream = NULL;

	RAII_VAR(struct ast_format_cap *, audio_caps, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, video_caps, NULL, ao2_cleanup);


	struct rtsp_pvt *p = ast_channel_tech_pvt(chan);
	
	if (!p) {
		ast_log(LOG_NOTICE, "[rtsp] rtsp_call: No channel tech found!\n");
		return 0;
	}

	if (ast_rtp_codecs_payloads_initialize(&newaudiortp) || ast_rtp_codecs_payloads_initialize(&newvideortp)) {
		res = 0;
		goto process_sdp_cleanup;
	}

	topology = ast_stream_topology_alloc();
	if (!topology) {
		res = 0;
		goto process_sdp_cleanup;
	}

	/* Read each line */
	while ((j = strstr(i, "\n")) != NULL) {
		/* if it's not enougth data */
		if (j - i <= 1)
			goto next;

		/* If previous is a \r" */
		if (j[-1] == '\r')
			/* Decrease end */
			j--;

		/* log */
		ast_verbose("[rtsp] -line [%s]\n", strndup(i, j - i));

		/* Check header */
		if (strncmp(i, "m=", 2) == 0) {
			/* media */
			int size = j - i + 1;
			if (size > sizeof(mline)) {
				size = sizeof(mline);
			}
			snprintf(mline, size, "%s", i);
			if (strncmp(i + 2, "video", 5) == 0) {
				/* create video */
				if (!CreateMedia(mline, VIDEO_SDP, &newvideortp,&video_payload)) {
					ast_log(LOG_NOTICE, "[rtsp] Error creating media on Video SDP\n");
					res = 0;
					goto process_sdp_cleanup;
				}
				 media = VIDEO_SDP;
			} else if (strncmp(i + 2, "audio", 5) == 0) {
				/* create audio */
				if (!CreateMedia(mline, AUDIO_SDP, &newaudiortp,&audio_payload)) {
					ast_log(LOG_NOTICE, "[rtsp] Error creating media on Audio SDP\n");
					res = 0;
					goto process_sdp_cleanup;
				}
				p->has_audio = 1; //added by zyxia for Bug 249215,2023.2.6
				media = AUDIO_SDP;
			} else {
				media = -1;
			}
		} else if (sscanf(i, "a=rtpmap: %30u %127[^/]/%30u", &payload, mimeSubtype, &sample_rate) == 3) {
			ast_verb(1, "[rtsp] Got a=rtpmap media id%d, payload %u, mimeSubtype %s, sample_rate %u\n", media, payload, mimeSubtype, sample_rate);
			/* if not in media */
			if (media < 0)
				goto next;
			if (media == VIDEO_SDP) {
				if (!(ast_rtp_codecs_payloads_set_rtpmap_type_rate(&newvideortp, NULL, payload, "video", mimeSubtype, 0, sample_rate))) {
					pbx_builtin_setvar_helper(chan, "RTSP_VIDEO_FORMAT",mimeSubtype);
					ast_verb(1, "[rtsp] Found video description format %s for payload %u\n", mimeSubtype, payload);
				} else {
					ast_rtp_codecs_payloads_unset(&newvideortp, NULL, payload);
					ast_verb(1, "[rtsp] Unknown video description format %s for payload %u\n", mimeSubtype, payload);
				}


			} else if (media == AUDIO_SDP) {
				if (!(ast_rtp_codecs_payloads_set_rtpmap_type_rate(&newaudiortp, NULL, payload, "audio", mimeSubtype, 0, sample_rate))) {
					ast_verb(1, "[rtsp] Found audio description format %s for ID %u\n", mimeSubtype, payload);
					p->has_audio = 1;
					pbx_builtin_setvar_helper(chan, "RTSP_AUDIO_FORMAT",mimeSubtype);
					audio_payload = payload;
				} else {
					ast_rtp_codecs_payloads_unset(&newaudiortp, NULL, payload);
					ast_verb(1, "[rtsp] Unknown audio description format %s for ID %u\n", mimeSubtype, payload);
				}
			}


		} else if (sscanf(i, "a=fmtp: %30u %255[^\t\n]", &payload, fmtp_string) == 2) {
			ast_verb(1, "[rtsp] Got a=fmtp media id%d, codec %u, fmtp_string %s\n", media, payload, fmtp_string);
			if (media != VIDEO_SDP)
				goto next;

			video_stream = ast_stream_alloc("video",AST_MEDIA_TYPE_VIDEO);
			if (!video_stream) {
				goto next;
			}
			
			ast_stream_set_state(video_stream, AST_STREAM_STATE_SENDRECV);
			ast_stream_set_metadata(video_stream, "CONTENT", "main");
			ast_stream_set_sub_type(video_stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
			ast_stream_set_type(video_stream, AST_MEDIA_TYPE_VIDEO);

			snprintf(mslabel, sizeof(mslabel) - 1, "%s_%ld-%s",p->cid_num ,ast_random(),ast_codec_media_sub_type2str(AST_MEDIA_SUB_VIDEO_TYPE_MAIN));
			snprintf(label, sizeof(label) - 1, "%s", ast_codec_media_sub_type2str(AST_MEDIA_SUB_VIDEO_TYPE_MAIN));
			ast_stream_set_metadata(video_stream, "MSID:MSLABEL", mslabel);
			ast_stream_set_metadata(video_stream, "MSID:LABEL", label);

			RAII_VAR(struct ast_format *, video_format, NULL, ao2_cleanup);
			if ((video_format = ast_rtp_codecs_get_payload_format(&newvideortp, payload))) {
				ast_verb(1, "[rtsp] Found video format %s for payload %u\n", ast_format_get_name(video_format), payload);
				if (ast_format_cmp(video_format, ast_format_h264) == AST_FORMAT_CMP_EQUAL) {
					//ast_format_attribute_set_directly(video_format, "profile-level-id", "profile-level-id=4d001f"); //720p

					char *attribs = ast_strdupa(fmtp_string), *attrib;
					while ((attrib = strsep(&attribs, ";"))) {
						ast_verb(1, "[rtsp] H264 %s\n", attrib);
						if (*attrib == ' ') {
							attrib++;
						}
						char val_tmp[8] = "";
						char SPS[512] = "";
						char PPS[512] = "";
						if (sscanf(attrib, "profile-level-id=%s", val_tmp) == 1) {
							long val;
							if (strlen(val_tmp) > 0 && sscanf(val_tmp, "%lx", &val) == 1) {
								pbx_builtin_setvar_helper(chan, "RTSP_PROFILE_LEVEL_ID",attrib);
								ast_format_attribute_set_directly(video_format, "profile-level-id", attrib);
								ast_format_attribute_set_directly(video_format, "packetization-mode", "1");
								profile_level_id = val;
								ast_verb(1, "[rtsp] profile_level_id=%lx\n", profile_level_id);
							}
						}
						else if (sscanf(attrib, "sprop-parameter-sets=%[^,],%s", SPS,PPS)) {
							unsigned char decode_sps[64];
							char profile_level_id[64]="";
							int len = ast_base64decode(decode_sps, SPS, sizeof(decode_sps));
							if (len > 4 ) {
								snprintf(profile_level_id, sizeof(profile_level_id), "profile-level-id=%02x%02x%02x",decode_sps[1],decode_sps[2],decode_sps[3] );
								pbx_builtin_setvar_helper(chan, "RTSP_PROFILE_LEVEL_ID",profile_level_id);
								ast_format_attribute_set_directly(video_format, "profile-level-id", profile_level_id); //720p
								ast_format_attribute_set_directly(video_format, "packetization-mode", "1");
								ast_verb(1, "[rtsp] %s\n", profile_level_id);
							}

						}
					}
				}

				/*Set video stream ,add append*/
				video_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
				if (!video_caps) {
					ast_stream_free(video_stream);
					goto next;
				}

				if (ast_format_cap_append(video_caps, video_format, 0)) {
					ast_stream_free(video_stream);
					goto next;
				}
				ast_stream_set_formats(video_stream, video_caps);

			} else {
				ast_log(LOG_NOTICE, "[rtsp] Unknown video description format for payload %u\n", payload);
			}

			p->has_video = 1;

		} 
next:
		/* if it's a \r */
		if (j[0] == '\r')
			/* skip \r\n to next line */
			i = j + 2;
		else
			/* skip \n to next line */
			i = j + 1;
	}



	if(p->has_audio) {
		char audio_uuid_str[AST_UUID_STR_LEN] = {0};
		char audio_msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

		audio_stream = ast_stream_alloc("audio",AST_MEDIA_TYPE_AUDIO);
		if (!audio_stream) {
			res = R_FAIL;
			goto process_sdp_cleanup;
		}

		ast_stream_set_state(audio_stream, AST_STREAM_STATE_SENDRECV);
		ast_stream_set_sub_type(audio_stream, AST_MEDIA_SUB_AUDIO_TYPE);
		ast_stream_set_type(audio_stream, AST_MEDIA_TYPE_AUDIO);

		memset(mslabel, 0, sizeof(mslabel));
		memset(label, 0, sizeof(label));
		snprintf(mslabel, sizeof(mslabel) - 1, "%s_%ld-%s",p->cid_num ,ast_random(), ast_codec_media_sub_type2str(AST_MEDIA_SUB_AUDIO_TYPE));
		snprintf(label, sizeof(label) - 1, "%s", ast_codec_media_sub_type2str(AST_MEDIA_SUB_AUDIO_TYPE));
		ast_stream_set_metadata(audio_stream, "MSID:MSLABEL", mslabel);
		ast_stream_set_metadata(audio_stream, "MSID:LABEL", label);

		//ast_stream_set_rtp_codecs(audio_stream, &newaudiortp);
		/*Set video stream ,add append*/
		audio_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!audio_caps) {
			ast_stream_free(audio_stream);
			res = R_FAIL;
			goto process_sdp_cleanup;
		}

		RAII_VAR(struct ast_format *, audio_format, NULL, ao2_cleanup);
		audio_format = ast_rtp_codecs_get_payload_format(&newaudiortp, audio_payload);

		if (audio_format && 
			(ast_format_cmp(audio_format, ast_format_g722) == AST_FORMAT_CMP_EQUAL 
			||  ast_format_cmp(audio_format, ast_format_alaw) == AST_FORMAT_CMP_EQUAL 
			||  ast_format_cmp(audio_format, ast_format_ulaw) == AST_FORMAT_CMP_EQUAL ) 
			&& !ast_format_cap_append(audio_caps, audio_format, 0)) {
			ast_log(LOG_NOTICE, "[rtsp] audio   format for payload %u\n", payload);
		}
		else {
			ast_log(LOG_NOTICE, "[rtsp] dont support Unknown audio  format for payload %u \n", payload);
		//edit by zyxia for Bug 278673 ,audio codec dont support
			ast_stream_free(audio_stream);
			res = AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
			goto process_sdp_cleanup;
		}

		ast_stream_set_formats(audio_stream, audio_caps);
		if (ast_stream_topology_append_stream(topology, audio_stream) < 0) {
			ast_stream_free(audio_stream);
			res = R_FAIL;
			goto process_sdp_cleanup;
		}

		struct avs_port_param port_param = {0};
		port_param.port_group_id = p->port_group_id;
		port_param.endpt_ice_support = 0;
		port_param.media_turn_relay = 0;
		port_param.enable_rtsp = 1;
		if(ast_channel_get_avs_address(chan))
			strcpy(port_param.avs_address, ast_channel_get_avs_address(chan));
		snprintf(port_param.rtsp_stream_id,sizeof(port_param.rtsp_stream_id),"%s",rtsp_info->stream_id);

		if (!(p->rtp = ast_rtp_instance_new("avs", sched, &bindaddr, &port_param))) {
			ast_log(LOG_WARNING, "[rtsp] Unable to create audio RTP session: %s\n", strerror(errno));
			res = R_FAIL;
			goto process_sdp_cleanup;
		}
		ast_rtp_codecs_payloads_copy(&newaudiortp, ast_rtp_instance_get_codecs(p->rtp), p->rtp);
		ast_rtp_instance_set_caps(p->rtp, audio_caps);

		ast_rtp_instance_set_remote_ssrc(p->rtp, ast_random());
		ast_rtp_instance_set_remote_cname(p->rtp,ast_uuid_generate_str(audio_uuid_str, sizeof(audio_uuid_str)));

		/* Set  msid */
		snprintf(audio_msid, sizeof(audio_msid), "%s %s", ast_stream_get_metadata(audio_stream, "MSID:MSLABEL"),  ast_stream_get_metadata(audio_stream, "MSID:LABEL"));
		ast_rtp_instance_set_remote_msid(p->rtp,audio_msid);
		ast_rtp_instance_set_local_msid(p->rtp,audio_msid);

		ast_rtp_instance_set_transmode(p->rtp,AST_STREAM_STATE_RECVONLY);
		ast_rtp_instance_set_media_type(p->rtp, AST_MEDIA_SUB_AUDIO_TYPE);
		ast_avs_after_nego_notify(chan, p->rtp);
	}

	if(p->has_video) {

		/*Start added by zyxia for Bug 236254,Bug 247375,2023.1.16*/
		if ( video_stream && ast_stream_topology_append_stream(topology, video_stream) < 0) {
			ast_stream_free(video_stream);
			res = R_FAIL;
			goto process_sdp_cleanup;
		}
		/*End added by zyxia for Bug 236254,Bug 247375,2023.1.16*/

		struct avs_port_param port_param = {0};
		port_param.port_group_id = p->port_group_id;
		port_param.endpt_ice_support = 0;
		port_param.media_turn_relay = 0;
		port_param.enable_rtsp = 1;
		if(ast_channel_get_avs_address(chan))
			strcpy(port_param.avs_address, ast_channel_get_avs_address(chan));

		snprintf(port_param.rtsp_stream_id,sizeof(port_param.rtsp_stream_id),"%s",rtsp_info->stream_id);

		if (!(p->vrtp = ast_rtp_instance_new("avs", sched, &bindaddr, &port_param))) {
			ast_log(LOG_WARNING, "[rtsp] Unable to create audio RTP session: %s\n", strerror(errno));
			res = R_FAIL;
			goto process_sdp_cleanup;
		}
		ast_rtp_codecs_payloads_copy(&newvideortp, ast_rtp_instance_get_codecs(p->vrtp), p->vrtp);

		char video_uuid_str[AST_UUID_STR_LEN] = {0};
		char video_msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

		ast_rtp_instance_set_remote_ssrc(p->vrtp, ast_random());
		ast_rtp_instance_set_remote_cname(p->vrtp,ast_uuid_generate_str(video_uuid_str, sizeof(video_uuid_str)));

		/* Set  msid */
		snprintf(video_msid, sizeof(video_msid), "%s %s", ast_stream_get_metadata(video_stream, "MSID:MSLABEL"),  ast_stream_get_metadata(video_stream, "MSID:LABEL"));
		ast_rtp_instance_set_remote_msid(p->vrtp,video_msid);
		ast_rtp_instance_set_local_msid(p->vrtp,video_msid);

		ast_rtp_instance_set_media_type(p->vrtp, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		ast_rtp_instance_set_transmode(p->vrtp,AST_STREAM_STATE_RECVONLY);
		ast_rtp_instance_set_caps(p->vrtp, video_caps);
		ast_avs_after_nego_notify(chan, p->vrtp);

	}

	ast_channel_set_stream_topology(chan, topology);

process_sdp_cleanup:

	ast_rtp_codecs_payloads_destroy(&newaudiortp);
	ast_rtp_codecs_payloads_destroy(&newvideortp);

	/* Return sdp */
	return res;
}

static enum ast_terminal_type chan_rtsp_get_terminal_type(struct ast_channel *ast)
{
	enum ast_terminal_type type = AST_TERMINAL_TYPE_IPC;
	
	return type;
}

/**
** ast_rtsp_ouraddrfor function
**/
static int ast_rtsp_ouraddrfor(const struct ast_sockaddr *them, struct ast_sockaddr *us, struct rtsp_pvt *p)
{
   if (!ast_sockaddr_isnull(&bindaddr))
      ast_sockaddr_copy(us, &bindaddr);
   else
      return ast_ouraddrfor(them, us);
   
   return 0;
}

static void rtsp_destroy(struct rtsp_pvt *p, int lockowner, int lockdialoglist)
{
	ast_debug(0, "[rtsp] Destroying RTSP pvt [%p]\n", p);

	if (p->owner) {
		if (lockowner) {
			ast_channel_lock(p->owner);
		}
		ast_debug(0, "[rtsp] Detaching from %s\n", ast_channel_name(p->owner));
		ast_channel_tech_pvt_set(p->owner, NULL);
		/* Make sure that the channel knows its backend is going away */
		ast_channel_softhangup_internal_flag_add(p->owner, AST_SOFTHANGUP_DEV);
		if (lockowner) {
			ast_channel_unlock(p->owner);
		}
		/* Give the channel a chance to react before deallocation */
		usleep(1);
	}

	if (p->rtp) {
		ast_rtp_instance_destroy(p->rtp);
		p->rtp = NULL;
	}
	if (p->vrtp) {
		ast_rtp_instance_destroy(p->vrtp);
		p->vrtp = NULL;
	}

	ao2_cleanup(p->caps);
	ao2_cleanup(p->jointcaps);
}

static void rtsp_destroy_fn(void *p)
{
	rtsp_destroy(p, 1, 1);
}

/**
** rtsp_hangup function
 * p ref -1
**/
/* Part of PBX interface, called from ast_hangup */
static int rtsp_hangup(struct ast_channel *chan)
{
	ast_log(LOG_DEBUG, "[rtsp] rtsp_hangup[%s]\n", ast_channel_name(chan));
	struct rtsp_pvt *p = ast_channel_tech_pvt(chan);

	if (!p) {
		ast_log(LOG_NOTICE, "[rtsp] Asked to hangup channel not connected\n");
		return 0;
	}

	ao2_lock(p);

	/* Determine how to disconnect */
	if (p->owner != chan) {
		ast_log(LOG_WARNING, "[rtsp] Huh?  We aren't the owner?\n");
		ao2_unlock(p);
		return 0;
	}

	/* Disconnect */
	if (ast_opt_avs) {
		struct avs_play_rtsp_info rtsp_info;
		memset(&rtsp_info, 0, sizeof(rtsp_info));
		rtsp_info.opt = RTSP_PLAY_STOP;
		//rtsp_info.stream_id =  p->stream_id;
		snprintf(rtsp_info.stream_id, sizeof(rtsp_info.stream_id), "%s",p->stream_id);
		snprintf(rtsp_info.url, sizeof(rtsp_info.url), "%s",p->rtsp_url);
		ast_avs_control(chan,AVS_CONTROL_STOP_RTSP,&rtsp_info,sizeof(rtsp_info));
	}

	p->owner = NULL;
	ast_channel_tech_pvt_set(chan, NULL);

	/* Update usage counter */
	ast_mutex_lock(&usecnt_lock);
	usecnt--;
	if (usecnt < 0) ast_log(LOG_WARNING, "[rtsp] Usecnt < 0\n");
	ast_mutex_unlock(&usecnt_lock);

	ast_update_use_count();
	ast_setstate(chan, AST_STATE_DOWN);

	ao2_ref(p,-1);
	return 0;
}

/**
** rtsp_alloc function
**/
static struct rtsp_pvt *rtsp_alloc(int callid, struct ast_sockaddr *sin, int useglobalnat)
{
	struct rtsp_pvt *p;

	p = ao2_alloc(sizeof(*p), rtsp_destroy_fn);
	if (!p) {
		ast_log(LOG_ERROR, "[rtsp] Couldn't allocate private structure. This is bad\n");
		return NULL;
	}

	if (!sin) {
		p->ourip = bindaddr;
	} else {
		ast_sockaddr_copy(&p->sa, sin);
		ast_rtsp_ouraddrfor(&p->sa, &p->ourip, p);
	}

	/* Assign a default capability */
	p->caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	p->jointcaps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!p->caps|| !p->jointcaps) {
		ao2_cleanup(p->caps);
		ao2_cleanup(p->jointcaps);
		ao2_ref(p, -1);
		return NULL;
	}
	ast_format_cap_append_from_cap(p->caps, rtsp_tech.capabilities, AST_MEDIA_TYPE_UNKNOWN);
	return p;
}

static void ast_rtsp_set_avs_address( struct ast_channel *chan)
{
	char *avs_address = NULL;

	if ( !chan) {
		return;
	}

	if (!ast_get_avs_master_value()) {
		return;
	}

	avs_address = ast_avs_get_avs_address();
	if(avs_address) {
		ast_channel_set_avs_address(chan, avs_address);
		ast_free(avs_address);
	}

	ast_log(LOG_DEBUG, "[chan:%s] set_chan_avs_address:%s",
		ast_channel_name(chan), ast_channel_get_avs_address(chan));

	return;
}

/**
** rtsp_new function
 *
 * \pre p is locked
 *
 * \return New ast_channel locked.
 * p ref +1
**/
static struct ast_channel *rtsp_new(struct rtsp_pvt *p, int state, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *host)
{
	ast_log(LOG_DEBUG, "[rtsp] rtsp_new: host=%s\n", host);
	struct ast_format_cap *caps;
	struct ast_channel *ch;
	struct ast_format *fmt;
	struct ast_format_cap *what = NULL; /* SHALLOW COPY DO NOT DESTROY! */
	struct ast_str *codec_buf = ast_str_alloca(64);

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return NULL;
	}

	/* Don't hold a rtsp pvt lock while we allocate a channel */
	ao2_unlock(p);
	ch = ast_channel_alloc(1, state, p->cid_num, p->cid_name, NULL, 0, 0, assignedids, requestor, p->amaflags, "RTSP/%s-%08x", host, ast_atomic_fetchadd_int((int *)&chan_idx, +1));
	if (!ch) {
		ast_log(LOG_WARNING, "[rtsp] Unable to allocate AST channel structure for RTSP channel\n");
		ao2_ref(caps, -1);
		ao2_lock(p);
		return NULL;
	}


	ao2_lock(p);
	ast_channel_set_port_group_id(ch, p->port_group_id);
	ast_channel_tech_set(ch, &rtsp_tech);

	/* Select our native format based on codec preference until we receive
	something from another device to the contrary. */
	if (ast_format_cap_count(p->jointcaps)) {	/* The joint capabilities of us and peer */
		what = p->jointcaps;
	} else if (ast_format_cap_count(p->caps)) {		/* Our configured capability for this peer */
		what = p->caps;
	} else {
		what = rtsp_tech.capabilities;
	}

	/* Set the native formats */
	ast_format_cap_append_from_cap(caps, what, AST_MEDIA_TYPE_UNKNOWN);
	/* Use only the preferred audio format, which is stored at the '0' index */
	fmt = ast_format_cap_get_best_by_type(what, AST_MEDIA_TYPE_AUDIO); /* get the best audio format */
	if (fmt) {
		ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_AUDIO); /* remove only the other audio formats */
		ast_format_cap_append(caps, fmt, 0); /* add our best choice back */
	} else {
		/* If we don't have an audio format, try to get something */
		fmt = ast_format_cap_get_format(caps, 0);
		if (!fmt) {
			ast_log(LOG_WARNING, "[rtsp] No compatible formats could be found for %s\n", ast_channel_name(ch));
			ao2_ref(caps, -1);
			ch = ast_channel_unref(ch);
			return NULL;
		}
	}
	ast_channel_nativeformats_set(ch, caps);
	ao2_ref(caps, -1);

	ast_debug(3, "[rtsp] *** Our native formats are %s \n", ast_format_cap_get_names(ast_channel_nativeformats(ch), &codec_buf));
	ast_debug(3, "[rtsp] *** Joint capabilities are %s \n", ast_format_cap_get_names(p->jointcaps, &codec_buf));
	ast_debug(3, "[rtsp] *** Our capabilities are %s \n", ast_format_cap_get_names(p->caps, &codec_buf));
	ast_debug(3, "[rtsp] *** AST_CODEC_CHOOSE formats are %s \n", ast_format_get_name(fmt));

	ast_setstate(ch, state);

	if (state == AST_STATE_RING) {
		ast_channel_rings_set(ch, 1);
	}

	ast_channel_adsicpe_set(ch, AST_ADSI_UNAVAILABLE);
	ast_channel_set_writeformat(ch, fmt);
	ast_channel_set_rawwriteformat(ch, fmt);

	ast_channel_set_readformat(ch, fmt);
	ast_channel_set_rawreadformat(ch, fmt);

	ao2_ref(fmt, -1);

	ao2_ref(p, 1);
	ast_channel_tech_pvt_set(ch, p);

	/*
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;
	struct ast_format_cap *caps;


	topology = ast_stream_topology_alloc();
	if (!topology) {
		return -1;
	}

	ast_channel_set_stream_topology(ch, topology);
	*/
	/*  Set the owner of this channel */
	p->owner = ch;

	/* Update usage counter */
	ast_mutex_lock(&usecnt_lock);
	usecnt++;

	ast_mutex_unlock(&usecnt_lock);
	ast_update_use_count();
	ast_channel_priority_set(ch, 1);

	if (!ast_strlen_zero(p->callid)) {
		pbx_builtin_setvar_helper(ch, "RTSPCALLID", p->callid);
	}
	/*
	if (!ast_strlen_zero(p->accountcode)) {
		ast_channel_accountcode_set(ch, p->accountcode);
	}

	if (p->amaflags) {
		ast_channel_amaflags_set(ch, p->amaflags);
	}
	*/
	ast_rtsp_set_avs_address(ch);

	if (ast_opt_avs) {
		ast_avs_create_channel(ch, avs_channel_type_rtp, 0, ast_channel_get_port_group_id(ch));
	}

	ast_log(LOG_DEBUG, "[rtsp] rtsp_new[%s]\n", ast_channel_name(ch));
	return ch;
}

/**
** rtsp_call function
**/
static int rtsp_call(struct ast_channel *chan, const char *data, int timeout)
{
	const char *is_sfu = NULL;
	int ret = R_FAIL;
	ast_log(LOG_DEBUG, "[rtsp] rtsp_call[%s]: data=%s\n", ast_channel_name(chan), data);
	struct rtsp_pvt *p = ast_channel_tech_pvt(chan);
	if (!data || !strlen(data)) {
		ast_log(LOG_WARNING, "[rtsp] rtsp_call: requires an data argument\n");
		return -1;
	}

	if (!p) {
		ast_log(LOG_NOTICE, "[rtsp] rtsp_call: No channel tech found!\n");
		return -1;
	}

	if (ast_opt_avs) {
		struct avs_play_rtsp_info rtsp_info;
		memset(&rtsp_info, 0, sizeof(rtsp_info));
		rtsp_info.opt = RTSP_PLAY_START;
		snprintf(rtsp_info.url, sizeof(rtsp_info.url), "%s",p->rtsp_url);
		snprintf(rtsp_info.username, sizeof(rtsp_info.username), "%s",p->rtsp_username);
		snprintf(rtsp_info.password, sizeof(rtsp_info.password), "%s",p->rtsp_password);
		snprintf(rtsp_info.transport_protocol, sizeof(rtsp_info.transport_protocol), "%s",p->transport_protocol);
		ret = ast_avs_control(chan,AVS_CONTROL_START_RTSP,&rtsp_info,sizeof(rtsp_info));
		ast_log(LOG_NOTICE, "[rtsp] rtsp_call: sdp info:%s\n",rtsp_info.sdp_info);

		if (ret ==R_SUCCESS) {

			snprintf(p->stream_id, sizeof(p->stream_id), "%s",rtsp_info.stream_id);
			ret = parseSDP(chan,&rtsp_info);

			is_sfu = pbx_builtin_getvar_helper((struct ast_channel *)chan, "IS_SFU");
			if (ret == R_SUCCESS && !ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
				/*parsesdpinfo from*/
				struct ast_stream *stream1 =NULL;
				for (int i = 0; i < ast_stream_topology_get_count(ast_channel_get_stream_topology(chan)); ++i) {

					stream1 = ast_stream_topology_get_stream(ast_channel_get_stream_topology(chan), i);
					if (ast_stream_get_type(stream1) == AST_MEDIA_TYPE_VIDEO) {
						for (int index = 0; index < ast_format_cap_count(ast_stream_get_formats(stream1)); ++index) {
							RAII_VAR(struct ast_format *, video_format, NULL, ao2_cleanup);
							video_format = ast_format_cap_get_format(ast_stream_get_formats(stream1), index);
							if (ast_format_cmp(video_format, ast_format_h264) == AST_FORMAT_CMP_EQUAL) {
								ast_log(LOG_NOTICE, "[rtsp] rtsp_call: h264 resouluce:%s\n",ast_channel_name(chan));
								ret = set_h264_resolution_for_rtsp(chan,video_format,stream1);//modify by zyxia for Bug 206905 ,2021.10.29
								break;
							}
						}
					}
				}
			}


			if (ret == R_SUCCESS)
				ast_queue_control(chan, AST_CONTROL_ANSWER);
			else if (ret == AST_CAUSE_BEARERCAPABILITY_NOTAVAIL) {
				ast_channel_hangupcause_set(chan, AST_CAUSE_BEARERCAPABILITY_NOTAVAIL);
				ast_set_hangupsource(chan, ast_channel_name(chan), 0);
				ast_queue_hangup(chan);
			}
			else
				ast_queue_control(chan, AST_CONTROL_CONGESTION);
		}
		else {
			ast_queue_control(chan, AST_CONTROL_CONGESTION);
		}
	}

	return 0;
}

static int rtsp_write(struct ast_channel *ast, struct ast_frame *frame)
{
	return 0;
}

/*Start added by zyxia for Bug 206905 ,2021.10.29*/
static void process_rtsp_join_sfu_control_frame(struct  ast_channel *channel)
{
	const char *is_nway_conf = NULL;
	const char *is_sfu = NULL;
	int resolution=0;
	if (!channel) {
		return;
	}

	is_nway_conf = pbx_builtin_getvar_helper(channel, "NWAY_CONF_NUMBER");
	is_sfu = pbx_builtin_getvar_helper(channel, "IS_SFU");

	const char* conf_room  = pbx_builtin_getvar_helper((struct ast_channel *)channel, "CONF_ROOM");

	if (!ast_strlen_zero(is_nway_conf) && !ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes") && conf_room) {
			resolution = ast_conf_get_main_video_resource(conf_room, channel, AST_STREAM_VIDEO_RESOLUTION_360P);
			if ((resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) || (resolution == -1)) {
				/* Todo, It seems can't failed, if happen need to close camera */
				ast_log(LOG_ERROR, "[channel:%p] process_nway_join_confenerce can't get video resource\n",
					ast_channel_name(channel));
			}
	}
	return;
}

/**
** rtsp_indicate function
**/
static int rtsp_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen)
{
	int res = 0;
	struct rtsp_pvt *p = ast_channel_tech_pvt(ast);
	if (NULL == p) {
		ast_debug(1, "[rtsp] Asked to indicate condition on channel %s with no pvt; ignoring\n", ast_channel_name(ast));
		return res;
	}

	ao2_lock(p);
	switch(condition) {
	case AST_CONTROL_VIDUPDATE:
		break;
	case AST_CONTROL_JOIN_SFU:
		ast_log(LOG_DEBUG, "[channel:%s] Recv AST_CONTROL_JOIN_SFU, \n",
			ast_channel_name(ast));
		process_rtsp_join_sfu_control_frame(ast);
		break;
	/*Start added by zyxia for Bug 291954,2024.3.13 */
	case AST_CONTROL_NOTIFY_OPEN_FOCUS_VIDEO:
		ast_log(LOG_DEBUG, "[channel:%s] Recv AST_CONTROL_NOTIFY_OPEN_FOCUS_VIDEO, \n",
			ast_channel_name(ast));
		ast_avs_mcu_control(ast, p->vrtp, AVS_MCU_MIX_TYPE_FOCUS_VIDEO);
		ast_channel_set_is_focus_record(ast, 1);
		break;
	case AST_CONTROL_NOTIFY_CLOSE_FOCUS_VIDEO:
		ast_log(LOG_DEBUG, "[channel:%s] Recv AST_CONTROL_NOTIFY_CLOSE_FOCUS_VIDEO, \n",
			ast_channel_name(ast));
		ast_avs_mcu_control(ast, NULL, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);
		ast_channel_set_is_focus_record(ast, 0);
		break;
	/*End added by zyxia for Bug 291954,2024.3.13 */
	/*Start added by zyxia for  312473,2024.8.19*/
	case AST_CONTROL_NOTIFY_CHANNEL_LEAVE:
		ast_log(LOG_DEBUG, "[channel:%s] Recv AST_CONTROL_NOTIFY_CHANNEL_LEAVE, \n",
			ast_channel_name(ast));
		if(ast_channel_get_is_focus_record(ast))
			ast_avs_mcu_control(ast, NULL, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);
		break;
	/*End added by zyxia for	312473,2024.8.19*/
	default:
		ast_log(LOG_WARNING, "[rtsp] Don't know how to indicate condition %d\n", condition);
		res = -1;
		break;
	}

	ao2_unlock(p);
	return res;
}
/*End added by zyxia for Bug 206905 ,2021.10.29*/

static void parse_dial_string(char *data, struct ast_rtsp_user *pds)
{
	struct ast_rtsp_user *user= NULL;
	char *type =NULL;
	char *account = NULL;
	if (ast_strlen_zero(data))
		return;

	account = strsep(&data, "/");
	type = strsep(&data, "/");

	ast_log(LOG_DEBUG, "[rtsp] account:%s type:%s\n", account,type);

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&rtsp_users, user, list) {
		ast_log(LOG_DEBUG, "[rtsp] usr_account:%s type:%s\n", user->account,type);
		if (user->account !=NULL && !strcmp(user->account,account)) {
			ast_copy_string(pds->account, user->account, sizeof(pds->account));
			pds->username = ast_strdup(user->username);
			pds->password = ast_strdup(user->password);
			pds->host = ast_strdup(user->host);
			pds->transport_protocol = ast_strdup(user->transport_protocol);
			pds->streaming_protocol = ast_strdup(user->streaming_protocol);
			pds->port = user->port;
			pds->path = ast_strdup(user->path);
			pds->fullname = ast_strdup(user->fullname);
			break;
		}
	}
	ast_log(LOG_NOTICE, "[rtsp] account:%s username:%s,password:%s,host:%s\n", pds->account,pds->username,pds->password,pds->host);

	AST_RWLIST_TRAVERSE_SAFE_END;
}

/**
** rtsp_request function
**/
static struct ast_channel *rtsp_request(const char *type, struct ast_format_cap *cap, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *addr, int *cause)
{
	struct rtsp_pvt *p;
	struct ast_channel *tmpc = NULL;
	struct ast_str *codec_buf = ast_str_alloca(64);
	struct ast_rtsp_user *rtsp_user = NULL;
	char *tmpstr = NULL;

	if (!(rtsp_user = ast_calloc(1, sizeof(*rtsp_user)))) {
		return NULL;
	}

	tmpstr = ast_strdupa(addr);
	parse_dial_string(tmpstr, rtsp_user);

	if (!(ast_format_cap_has_type(cap, AST_MEDIA_TYPE_AUDIO))) {
		ast_log(LOG_NOTICE, "[rtsp] Asked to get a channel of unsupported format %s\n", ast_format_cap_get_names(cap, &codec_buf));
		*cause = AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;	/* Can't find codec to connect to host */
		return NULL;
	}
	ast_debug(1, "[rtsp] Asked to create a RTSP channel with formats: %s\n", ast_format_cap_get_names(cap, &codec_buf));

	if (ast_strlen_zero(addr)) {
		ast_log(LOG_ERROR, "[rtsp] Unable to create channel with empty destination.\n");
		*cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
		return NULL;
	}

	/* allocate a new rtsp_pvt struct for ch */
	if (!(p = rtsp_alloc(0, NULL, 0))) {
		ast_log(LOG_WARNING, "[rtsp] Unable to build pvt data for 'RTSP/%s'\n", (char *)addr);
		return NULL;
	}

	ast_format_cap_get_compatible(cap, p->caps, p->jointcaps);

	p->port_group_id = ast_random();

	/* RTSP timeout */
	//p->session_timeout = global_rtptimeout;
	//int ret = RTSP_CreateURL(p,rtsp_user);

	ast_log(LOG_NOTICE, "[rtsp] fullname:%s username: %s, password:%s\n",rtsp_user->fullname, rtsp_user->username,rtsp_user->password);

	//if (!ast_strlen_zero(rtsp_user->username) && !ast_strlen_zero(rtsp_user->password))
	//	snprintf(p->rtsp_url, sizeof(p->rtsp_url), "%s://%s:%s@%s:%d",rtsp_user->streaming_protocol,rtsp_user->username,rtsp_user->password,rtsp_user->host,rtsp_user->port);
	//else
	snprintf(p->rtsp_url, sizeof(p->rtsp_url) -1, "%s://%s:%d%s",rtsp_user->streaming_protocol,rtsp_user->host,rtsp_user->port,rtsp_user->path);

	if(!ast_strlen_zero(rtsp_user->username))
		snprintf(p->rtsp_username, sizeof(p->rtsp_username) -1, "%s",rtsp_user->username);

	if(!ast_strlen_zero(rtsp_user->password))
		snprintf(p->rtsp_password, sizeof(p->rtsp_password) -1, "%s",rtsp_user->password);

	snprintf(p->transport_protocol, sizeof(p->transport_protocol) -1, "%s",rtsp_user->transport_protocol);
	snprintf(p->cid_num,sizeof(p->cid_num) -1,"%s",rtsp_user->account);
	snprintf(p->cid_name,sizeof(p->cid_name) -1,"%s",rtsp_user->fullname);

	/* create new channel */
	ao2_lock(p);
	tmpc = rtsp_new(p, AST_STATE_DOWN, assignedids, requestor, addr);
	ao2_unlock(p);
	if (NULL != tmpc) {
		ast_channel_unlock(tmpc);
	}
	ao2_ref(p, -1);
	ast_free(rtsp_user);
	return tmpc;
}


static void rtsp_status_report(struct ast_rtsp_user *user,const char* status)
{
	ast_log(LOG_DEBUG, "rtsp_status_report to ipc %s,status:%s\n",user->account,status);
	manager_event(EVENT_FLAG_CALL, "IpcStatus", "Exten: %s\r\n"
									"Status: %s\r\n"
									"ErrorMessage: %s\r\n",
					user->account,
					status ?status :"Unmonitored",
					""
				);
}

static void* ast_rtsp_keepalive_thread(void *data)
{
	int  ret = 0;
	struct ast_rtsp_user *rtsp_user= (struct ast_rtsp_user *)data;
	//ao2_lock(rtsp_user);
	if (ast_opt_avs) {
		struct avs_play_rtsp_info rtsp_info;
		memset(&rtsp_info, 0, sizeof(rtsp_info));
		rtsp_info.opt = RTSP_KEEP_ALIVE;

		snprintf(rtsp_info.url, sizeof(rtsp_info.url), "%s://%s:%d%s",rtsp_user->streaming_protocol,rtsp_user->host,rtsp_user->port,rtsp_user->path);
		snprintf(rtsp_info.username, sizeof(rtsp_info.username), "%s",rtsp_user->username);
		snprintf(rtsp_info.password, sizeof(rtsp_info.password), "%s",rtsp_user->password);
		snprintf(rtsp_info.transport_protocol, sizeof(rtsp_info.transport_protocol), "%s",rtsp_user->transport_protocol);
		ret = ast_avs_control(NULL,AVS_CONTROL_RTSP_KEEPALIVE,&rtsp_info,sizeof(rtsp_info));
		if (ret == R_SUCCESS) {
			/*ipc status changed*/
			if(rtsp_info.code == 0 && rtsp_user->keepalive_status != RTSP_REACHABLE) {
				rtsp_status_report(rtsp_user,ipcstatus_tostr(RTSP_REACHABLE));
			}
			else{
				ast_log(LOG_DEBUG, "Ipc %s status did not changed,status:%d\n",rtsp_user->account,rtsp_user->keepalive_status);
			}
			rtsp_user->keepalive_status = RTSP_REACHABLE;
		}
		else {
			/*send offline*/
			if (rtsp_user->keepalive_status != RTSP_UNREACHABLE) {
				rtsp_status_report(rtsp_user,ipcstatus_tostr(RTSP_UNREACHABLE));
			}
			else{
				ast_log(LOG_DEBUG, "Ipc %s status did not changed,status:%d\n",rtsp_user->account,rtsp_user->keepalive_status);
			}
			rtsp_user->keepalive_status =RTSP_UNREACHABLE;
		}

		rtsp_info.opt = RTSP_KEEP_ALIVE_STOP;
		ast_avs_control(NULL,AVS_CONTROL_CLOSE_RTSP_KEEPALIVE,&rtsp_info,sizeof(rtsp_info));
	}

	/*add sched again*/
	#if 0
	if (rtsp_user->keepalive_sched_id > 0) {
		ret =ast_sched_del(sched,rtsp_user->keepalive_sched_id);
		if (ret != 0)
			ast_log(LOG_DEBUG, "[rtsp] account %s delete sched_id:%d ,ret:%d\n", rtsp_user->account,rtsp_user->keepalive_sched_id,ret);
		else
			rtsp_user->keepalive_sched_id = 0;
	}
	#endif
	rtsp_user->keepalive_sched_id = 0;
	if (rtsp_user->qualify_frequency) {//check if need  add again,for Bug 228062
		rtsp_user->keepalive_sched_id = ast_sched_add(sched, rtsp_user->qualify_frequency*1000, rtsp_keepalive_cb, rtsp_user);
		if(rtsp_user->keepalive_sched_id <= 0 )
			ast_log(LOG_ERROR, "Add keepalive sched failed  to ipc %s\n",rtsp_user->account);
		else
			ast_log(LOG_DEBUG, "Add keepalive sched %d  to ipc %s\n",rtsp_user->keepalive_sched_id,rtsp_user->account);
	}
	else{
		rtsp_status_report(rtsp_user,ipcstatus_tostr(RTSP_UNMONITORNED));//report again,for Bug 228062
	}
	pthread_exit(NULL);
	//ao2_unlock(rtsp_user);

}

static int rtsp_keepalive_cb(const void *data)
{
	pthread_t threadid;
	if (ast_pthread_create_detached(&threadid, NULL, ast_rtsp_keepalive_thread, (void*)data)) {
		ast_log(LOG_ERROR, "Error call to the gswave new address\n");
		return -1;
	}
	return 0;
}




static struct ast_rtsp_user *get_rtsp_by_account_locked(const char *account)
{
	struct ast_rtsp_user *user = NULL;

	AST_RWLIST_TRAVERSE(&rtsp_users, user, list) {
		if (!strcmp(user->account, account)) {
			break;
		}
	}

	return user;
}


static int action_ipcstatus(struct mansession *s, const struct message *m)
{
	const char *exten = astman_get_header(m, "Exten");
	struct ast_rtsp_user *rtsp_user = NULL;

	if (ast_strlen_zero(exten)) {
		astman_send_error(s, m, "Extension not specified");
		return 0;
	}

	rtsp_user = get_rtsp_by_account_locked(exten);

	if (rtsp_user == NULL) {
		astman_send_error(s, m, "Extension not exist");
		return 0;
	}

	rtsp_status_report(rtsp_user,ipcstatus_tostr(rtsp_user->keepalive_status));
	return 0;
}

/* reload_config */
static int reload_config(void)
{
	struct ast_config *cfg;

	struct ast_variable *var;
	struct ast_rtsp_user *rtsp_user = NULL;
	struct ast_flags config_flags = { 0 };
	char user_info[256]={0};
	char info_hash[33]={0};
	char *cat = NULL;
	int value =0;
	int ret = 0;
	int is_exist =0;
	int need_reload=0;
	char ipc_reload[8] = {0};

	cfg = ast_config_load2(config, "rtsp", config_flags);
	if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (!cfg || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_NOTICE, "Unable to open AMI configuration manager.conf, or configuration is invalid.\n");
		return 0;
	}


	/*Start added by zyxia for Bug 228075 ,2022.6.27*/
	ast_safe_nvram_get(":ipc_reload", ipc_reload, sizeof(ipc_reload));
	if ( ipc_reload != NULL && ipc_reload[0] != '\0')
		need_reload = atoi(ipc_reload);
	/*End added by zyxia for Bug 228075 ,2022.6.27*/

	ast_sockaddr_parse(&bindaddr, "0.0.0.0:0", 0);

	AST_RWLIST_WRLOCK(&rtsp_users);

	/*start added by zyxia for Bug 225186,2022.5.31*/
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&rtsp_users, rtsp_user, list) {
		rtsp_user->is_exist = 0;
	}

	AST_RWLIST_TRAVERSE_SAFE_END;
	/*End added by zyxia for Bug 225186,2022.5.31*/

	/* We *must* have a config file otherwise stop immediately */
	if (!cfg) {
		ast_log(LOG_NOTICE, "[rtsp] Unable to load config %s, RTSP disabled\n", config);
		return 0;
	}
	while ((cat = ast_category_browse(cfg, cat))) {

		if (!strcasecmp(cat, "general")) {
			continue;
		}
		/* Look for an existing entry, if none found - create one and add it to the list */
		if (!(rtsp_user = get_rtsp_by_account_locked(cat))) {
			if (!(rtsp_user = ast_calloc(1, sizeof(*rtsp_user)))) {
				break;
			}
			/* Copy name over */
			ast_copy_string(rtsp_user->account, cat, sizeof(rtsp_user->account));
			/* Insert into list */
			AST_RWLIST_INSERT_TAIL(&rtsp_users, rtsp_user, list);
		}
		else {
			is_exist = 1;
			ast_log(LOG_NOTICE, "[rtsp] exist:%s\n", cat);
		}

		var = ast_variable_browse(cfg, cat);
		for (; var; var = var->next) {
			if (!strcasecmp(var->name, "username")) {
				ast_free(rtsp_user->username);
#ifdef HAVE_UCMSECRET
				if (!ast_strlen_zero(var->value)) {
					char *plain_string = NULL;
					ucm_secret_decrypt(var->value, &plain_string);
					if (!ast_strlen_zero(plain_string)) {
						rtsp_user->username = ast_strdup(plain_string);
					} else {
						rtsp_user->username = ast_strdup(var->value);
					}
					if (plain_string) {
						ast_free(plain_string);
					}
				} else {
					rtsp_user->username = ast_strdup(var->value);
				}
#else
				rtsp_user->username = ast_strdup(var->value);
#endif
			} else if (!strcasecmp(var->name, "password")) {
				ast_free(rtsp_user->password);
#ifdef HAVE_UCMSECRET
				if (!ast_strlen_zero(var->value)) {
					char *plain_string = NULL;
					ucm_secret_decrypt(var->value, &plain_string);
					if (!ast_strlen_zero(plain_string)) {
						rtsp_user->password = ast_strdup(plain_string);
					} else {
						rtsp_user->password = ast_strdup(var->value);
					}
					if (plain_string) {
						ast_free(plain_string);
					}
				} else {
					rtsp_user->password = ast_strdup(var->value);
				}
#else
				rtsp_user->password = ast_strdup(var->value);
#endif
			} else if (!strcasecmp(var->name, "host")) {
				ast_free(rtsp_user->host);
				rtsp_user->host = ast_strdup(var->value);
			} else if (!strcasecmp(var->name, "port")) {
				value = atoi(var->value);
				rtsp_user->port = value;
			} else if (!strcasecmp(var->name, "transport_protocol")) {
				ast_free(rtsp_user->transport_protocol);
				rtsp_user->transport_protocol = ast_strdup(var->value);
			} else if (!strcasecmp(var->name, "streaming_protocol")) {
				ast_free(rtsp_user->streaming_protocol);
				rtsp_user->streaming_protocol = ast_strdup(var->value);
			} else if (!strcasecmp(var->name, "path")) {
				ast_free(rtsp_user->path);
				rtsp_user->path = ast_strdup(var->value);
			} else if (!strcasecmp(var->name, "qualify_frequency")) {
				value = atoi(var->value);
				rtsp_user->qualify_frequency = value;
			} else if (!strcasecmp(var->name, "fullname")) {
				ast_free(rtsp_user->fullname);
				rtsp_user->fullname = ast_strdup(var->value);
			} 
			else {
				ast_debug(1, "%s is an unknown option.\n", var->name);
			}
		}

		if (!rtsp_user->qualify_frequency) {
			rtsp_user->keepalive_status= RTSP_UNMONITORNED;//added by zyxia for Bug 225999.2022.6.9
			rtsp_status_report(rtsp_user,ipcstatus_tostr(RTSP_UNMONITORNED));
		}

		memset(user_info,0,sizeof(user_info));
		snprintf(user_info, sizeof(user_info), "%s:%s:%s:%s@%s:%d%s:%d",rtsp_user->fullname,rtsp_user->account,rtsp_user->username,rtsp_user->password,rtsp_user->host,rtsp_user->port,rtsp_user->path,rtsp_user->qualify_frequency);
		ast_md5_hash(info_hash, user_info);
		ast_log(LOG_DEBUG, "[rtsp] info is %s,new_hash:%s old_hash:%s\n", user_info,info_hash,rtsp_user->info_hash);
		//hash is not different,need s
		if (strcmp(rtsp_user->info_hash,info_hash) || need_reload){
			ast_log(LOG_NOTICE, "[rtsp] account %s info is changed, need reload,rtsp_user->keepalive_sched_id:%d \n", rtsp_user->account,rtsp_user->keepalive_sched_id);

			ast_copy_string(rtsp_user->info_hash, info_hash, sizeof(rtsp_user->info_hash));

			if (is_exist && rtsp_user->keepalive_sched_id > 0) {
				ret = ast_sched_del(sched,rtsp_user->keepalive_sched_id );
				if (ret != 0)
					ast_log(LOG_ERROR, "[rtsp] account %s delete sched_id:%d ,ret:%d\n", rtsp_user->account,rtsp_user->keepalive_sched_id,ret);
				else
					rtsp_user->keepalive_sched_id = 0;
			}

			if (rtsp_user->qualify_frequency) {
				rtsp_user->keepalive_status= RTSP_UNREACHABLE;//added by zyxia for Bug 225999.2022.6.9
				rtsp_status_report(rtsp_user,ipcstatus_tostr(RTSP_UNREACHABLE));
				rtsp_user->keepalive_sched_id = ast_sched_add(sched, 1, rtsp_keepalive_cb, rtsp_user);
				ast_log(LOG_DEBUG, "[rtsp] account %s added, rtsp_user->keepalive_sched_id:%d \n", rtsp_user->account,rtsp_user->keepalive_sched_id);
			}
		}
		rtsp_user->is_exist = 1;//mark it is not delete

	}

	/*Start added by zyxia for Bug 225186,2022.5.31*/
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&rtsp_users, rtsp_user, list) {
		ast_log(LOG_DEBUG, "[rtsp] name:%s, account:%s username:%s,password:%s,host:%s,port:%d,protocol:%s,path:%s,keepalive_sched_id:%d\n", rtsp_user->fullname,rtsp_user->account,rtsp_user->username,rtsp_user->password,rtsp_user->host,rtsp_user->port,rtsp_user->transport_protocol,rtsp_user->path,rtsp_user->keepalive_sched_id);
		/*account is not exist ,need delete*/
		if (!rtsp_user->is_exist) {
			if (rtsp_user->keepalive_sched_id > 0)
				ret= ast_sched_del(sched,rtsp_user->keepalive_sched_id );
			AST_RWLIST_REMOVE_CURRENT(list);
			ast_log(LOG_NOTICE, "[rtsp] Delete account:%s \n", rtsp_user->account);
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
	/*End added by zyxia for Bug 225186,2022.5.31*/

	AST_RWLIST_UNLOCK(&rtsp_users);

	ast_format_cap_remove_by_type(rtsp_tech.capabilities, AST_MEDIA_TYPE_UNKNOWN);
	ast_format_cap_append(rtsp_tech.capabilities, ast_format_ulaw, 0);
	ast_format_cap_append(rtsp_tech.capabilities, ast_format_alaw, 0);
	ast_format_cap_append(rtsp_tech.capabilities, ast_format_h264, 0);

	ast_safe_nvram_set(":ipc_reload","0");
	ast_safe_nvram_commit();

	ast_config_destroy(cfg);
	return ret;
}


/* unload_module */
static int unload_module(void)
{  
	ast_sched_dump(sched);

	/* First, take us out of the channel type list */
	ast_channel_unregister(&rtsp_tech);

	/* Unregister CLI commands */
	//ast_cli_unregister_multiple(cli_rtsp, ARRAY_LEN(cli_rtsp));

	io_context_destroy(io);
	ast_sched_context_destroy(sched);

	ao2_cleanup(rtsp_tech.capabilities);
	rtsp_tech.capabilities = NULL;

	return 0;
}

static int reload_module(void)
{
	int res = 0;
	ast_log(LOG_DEBUG, "[rtsp] reload_module rtsp module\n");
	res |= reload_config();

	return res;
}

/* load_module */
static int load_module(void)
{
	ast_log(LOG_DEBUG, "[rtsp] load rtsp module\n");

	if (!(rtsp_tech.capabilities = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}
	ast_format_cap_append_by_type(rtsp_tech.capabilities, AST_MEDIA_TYPE_AUDIO);

	if (!(sched = ast_sched_context_create())) {
		ast_log(LOG_ERROR, "[rtsp] Unable to create scheduler context\n");
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}

	if (ast_sched_start_thread(sched)) {
		ast_log(LOG_ERROR, "[rtsp] Could not start scheduler thread\n");
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}

	if (!(io = io_context_create())) {
		ast_log(LOG_ERROR, "[rtsp] Unable to create I/O context\n");
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}


	/* Load the configuration from rtsp.conf */
	if (reload_config()) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	/* Make sure we can register our rtsp channel type */
	if (ast_channel_register(&rtsp_tech)) {
		ast_log(LOG_ERROR, "[rtsp] Unable to register channel type 'RTSP'\n");
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}

	ast_manager_register_xml_core("IpcStatus", EVENT_FLAG_CALL | EVENT_FLAG_REPORTING, action_ipcstatus);

	/* Register all CLI functions for RTSP */
	//ast_cli_register_multiple(cli_rtsp, ARRAY_LEN(cli_rtsp));


	return AST_MODULE_LOAD_SUCCESS;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Real Time Streaming Protocol (RTSP)",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
	.load_pri = AST_MODPRI_CHANNEL_DRIVER,
);

