#ifdef GRANDSTREAM_NETWORKS

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include <stddef.h>

#include "asterisk.h"
#include "asterisk/lock.h"
#include "asterisk/translate.h"
#include "asterisk/config.h"
#include "asterisk/module.h"
#include "asterisk/cli.h"
#include "asterisk/channel.h"
#include "asterisk/utils.h"
#include "asterisk/linkedlists.h"
#include "asterisk/ulaw.h"
#include "asterisk/channel.h"
#include "asterisk/timing.h"
#include "asterisk/format_compatibility.h"
#include "asterisk/slin.h"
#include "asterisk/avs_controller.h"

AST_MODULE_LOG("transcoding");

struct translator {
	struct ast_translator      t;
	uint64_t                   srcfmt;
	uint64_t                   dstfmt;
	AST_LIST_ENTRY(translator) entry;
};

static AST_LIST_HEAD_STATIC(translators, translator);

#ifndef ast_container_of
#define ast_container_of(ptr, type, member) ((type *)((char *)(ptr) - offsetof(type, member)))
#endif

struct ast_translator_pvt {
	const char *name;
	uint64_t   src_fmt;
	uint64_t   dst_fmt;
} translator_pvt[] = {
	{
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_ALAW | AST_FORMAT_ULAW,
		.dst_fmt = AST_FORMAT_G729,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_G729,
		.dst_fmt = AST_FORMAT_ALAW,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_G729,
		.dst_fmt = AST_FORMAT_ULAW,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_ULAW | AST_FORMAT_ALAW,
		.dst_fmt = AST_FORMAT_OPUS,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_OPUS,
		.dst_fmt = AST_FORMAT_ALAW,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_OPUS,
		.dst_fmt = AST_FORMAT_ULAW,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_ALAW | AST_FORMAT_ULAW,
		.dst_fmt = AST_FORMAT_G723,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_G723,
		.dst_fmt = AST_FORMAT_ALAW,
	}, {
		.name = "Codec Container",
		.src_fmt = AST_FORMAT_G723,
		.dst_fmt = AST_FORMAT_ULAW,
	}
};

static const char *ast_get_codec_name(uint64_t format)
{
	switch (format) {
	case AST_FORMAT_ALAW:
		return "alaw";
	case AST_FORMAT_ULAW:
		return "ulaw";
	case AST_FORMAT_G722:
		return "g722";
	case AST_FORMAT_G723:
		return "g723";
	case AST_FORMAT_G726:
		return "g726rfc3551";
	case AST_FORMAT_G726_AAL2:
		return "g726aal2";
	case AST_FORMAT_ADPCM:
		return "adpcm";
	case AST_FORMAT_G729:
		return "g729";
	case AST_FORMAT_GSM:
		return "gsm";
	case AST_FORMAT_ILBC:
		return "ilbc";
	case AST_FORMAT_OPUS:
		return "opus";
	default:
		break;
	}

	return "";
}

#define AST_TRANSLATOR_PVT_SIZE (sizeof(translator_pvt) / sizeof(translator_pvt[0]))

static const struct ast_codec *ast_get_codec(uint64_t format)
{
	const struct ast_codec *ast_codec_tmp = NULL;

	static const struct ast_codec ast_alaw = {
		.name = "alaw",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
		.default_ms = 20,
	};

	static const struct ast_codec ast_ulaw = {
		.name = "ulaw",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
		.default_ms = 20,
	};

	static const struct ast_codec ast_g729 = {
		.name = "g729",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
		.default_ms = 20,
	};

	static const struct ast_codec ast_opus = {
		.name = "opus",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 48000,
		.default_ms = 20,
	};

	static const struct ast_codec ast_g723 = {
		.name = "g723",
		.type = AST_MEDIA_TYPE_AUDIO,
		.sample_rate = 8000,
		.default_ms = 30,
	};

	switch (format) {
	case AST_FORMAT_ALAW:
		ast_codec_tmp = &ast_alaw;
		break;
	case AST_FORMAT_ULAW:
		ast_codec_tmp = &ast_ulaw;
		break;
	case AST_FORMAT_G729:
		ast_codec_tmp = &ast_g729;
		break;
	case AST_FORMAT_OPUS:
		ast_codec_tmp = &ast_opus;
		break;
	case AST_FORMAT_G723:
		ast_codec_tmp = &ast_g723;
	default:
		break;
	}

	return ast_codec_tmp;
}

static int ast_newpvt(struct ast_trans_pvt *trans_pvt)
{
	struct translator *zt = ast_container_of(trans_pvt->t, struct translator, t);

	if (!zt) {
		return -1;
	}

	return 0;
}

static void ast_destroy(struct ast_trans_pvt *trans_pvt)
{
}

static int ast_framein(struct ast_trans_pvt *trans_pvt, struct ast_frame *f)
{
	struct translator *zt = trans_pvt->pvt;

	if (!zt) {
		return -1;
	}

	trans_pvt->samples += f->samples;
	trans_pvt->datalen = 0;

	return 0;
}

static struct ast_frame *ast_frameout(struct ast_trans_pvt *trans_pvt)
{
	struct translator *zt = trans_pvt->pvt;

	if (!zt) {
		return NULL;
	}

	return ast_trans_frameout(trans_pvt, 160, 160);
}

static int ast_register_translators(uint64_t dstfmt, uint64_t srcfmt)
{
	struct translator *trans = NULL;
	const struct ast_codec *src_codec = NULL, *dst_codec = NULL;
	struct translator *cur_trans = NULL;

	if (!(dst_codec = ast_get_codec(dstfmt)) || !(src_codec = ast_get_codec(srcfmt))) {
		return -1;
	}

	if (NULL == (trans = ast_calloc(1, sizeof(*trans)))) {
		return -1;
	}

	AST_LIST_LOCK(&translators);
	AST_LIST_TRAVERSE(&translators, cur_trans, entry) {
		if ((cur_trans->srcfmt == dstfmt) && (cur_trans->dstfmt == srcfmt)) {
			break;
		}
	}
	AST_LIST_UNLOCK(&translators);

	trans->dstfmt = dstfmt;
	trans->srcfmt = srcfmt;
	trans->t.newpvt = ast_newpvt;
	trans->t.framein = ast_framein;
	trans->t.frameout = ast_frameout;
	trans->t.destroy = ast_destroy;
	trans->t.buf_size = 8000;
	trans->t.buffer_samples = 0;
	trans->t.desc_size = sizeof(trans);
	memcpy(&trans->t.src_codec, src_codec, sizeof(*src_codec));
	memcpy(&trans->t.dst_codec, dst_codec, sizeof(*dst_codec));

	if (ast_register_translator(&trans->t)) {
		ast_log(LOG_ERROR, "ast_register_translator %s---->%s failure!\n", src_codec->name, dst_codec->name);
		ast_free(trans);
		return -1;
	}

	AST_LIST_LOCK(&translators);
	AST_LIST_INSERT_HEAD(&translators, trans, entry);
	AST_LIST_UNLOCK(&translators);

	ast_log(LOG_ERROR, "ast_register_translator %s---->%s success!\n", src_codec->name, dst_codec->name);

	return 0;
}

static void ast_unregister_translators(void)
{
	struct translator *cur = NULL;

	AST_LIST_LOCK(&translators);
	while ((cur = AST_LIST_REMOVE_HEAD(&translators, entry))) {
		ast_unregister_translator(&cur->t);
		ast_free(cur);
	}
	AST_LIST_UNLOCK(&translators);
}

static int ast_translator_already_exist(uint64_t dstfmt, uint64_t srcfmt)
{
	int res = 0;
	const struct translator *oct_trans = NULL;

	AST_LIST_LOCK(&translators);
	AST_LIST_TRAVERSE(&translators, oct_trans, entry) {
		if ((oct_trans->srcfmt == srcfmt) && (oct_trans->dstfmt == dstfmt)) {
			res = 1;
			break;
		}
	}
	AST_LIST_UNLOCK(&translators);

	return res;
}

static void ast_build_translator(uint64_t dstfmts, uint64_t srcfmts)
{
	uint64_t dstfmt, srcfmt;

	AST_LIST_LOCK(&translators);
	for (srcfmt = 1; srcfmt != 0; srcfmt <<= 1) {
		for (dstfmt = 1; dstfmt != 0; dstfmt <<= 1) {
			if (!(dstfmts & dstfmt) || !(srcfmts & srcfmt)) {
				continue;
			}

			if (ast_translator_already_exist(dstfmt, srcfmt)) {
				ast_log(LOG_WARNING, "Translator '%s' ----> '%s' already exist!\n", ast_get_codec_name(srcfmt), ast_get_codec_name(dstfmt));
				continue;
			}

			ast_register_translators(dstfmt, srcfmt);
		}
	}
	AST_LIST_UNLOCK(&translators);
}

static void ast_find_transcoders(void)
{
	int x;
	struct ast_translator_pvt pvt = {0};

	for (x = 0; x < AST_TRANSLATOR_PVT_SIZE; x++) {
		pvt = translator_pvt[x];

		ast_build_translator(pvt.dst_fmt, pvt.src_fmt);
	}

	ast_debug(1, "AST_TRANSLATOR_PVT_SIZE: %ld\n", (long)AST_TRANSLATOR_PVT_SIZE);
}

static int reload(void)
{
	return 0;
}

static int unload_module(void)
{
	int res = 0;

	ast_unregister_translators();

	return res;
}

static int load_module(void)
{
	int res = 0;

	ast_find_transcoders();

	if (res) {
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}	

	return AST_MODULE_LOAD_SUCCESS;

}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Transcoding container",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
	);

#endif
