/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Function to lookup the callerid number, and see if it is blacklisted
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \ingroup functions
 * 
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"



#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/astdb.h"
#include "asterisk/app.h"
#include "asterisk/haserver.h"

/*** DOCUMENTATION
	<function name="BLACKLIST" language="en_US">
		<synopsis>
			Check if the callerid is on the blacklist.
		</synopsis>
		<syntax />
		<description>
			<para>Uses astdb to check if the Caller*ID is in family <literal>blacklist</literal>.
			Returns <literal>1</literal> or <literal>0</literal>.</para>
		</description>
		<see-also>
			<ref type="function">DB</ref>
		</see-also>
	</function>
	<function name="INBOUND_BLACKLIST" language="en_US">
		<synopsis>
			Check if the extension is on the inbound_blacklist.
		</synopsis>
		<syntax />
		<description>
			<para>Uses blacklist.conf to check ifthe extension is in  <literal>blacklist</literal>.
			Returns <literal>1</literal> or <literal>0</literal>.</para>
		</description>
		<see-also>
			<ref type="function">DB</ref>
		</see-also>
	</function>
	<function name="OUTBOUND_BLACKLIST" language="en_US">
		<synopsis>
			Check if the extension is on the outbound_blacklist.
		</synopsis>
		<syntax />
		<description>
			<para>Uses blacklist.conf to check if the extension is in  <literal>blacklist</literal>.
			Returns <literal>1</literal> or <literal>0</literal>.</para>
		</description>
		<see-also>
			<ref type="function">DB</ref>
		</see-also>
	</function>

***/

AST_MODULE_LOG("func_blacklist");

static int blacklist_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char blacklist[1];
	int bl = 0;

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	if (ast_channel_caller(chan)->id.number.valid && ast_channel_caller(chan)->id.number.str) {
		if (!ast_ucm_db_get("blacklist", ast_channel_caller(chan)->id.number.str, blacklist, sizeof (blacklist)))
			bl = 1;
	}
	if (ast_channel_caller(chan)->id.name.valid && ast_channel_caller(chan)->id.name.str) {
		if (!ast_ucm_db_get("blacklist", ast_channel_caller(chan)->id.name.str, blacklist, sizeof (blacklist)))
			bl = 1;
	}
#ifdef GRANDSTREAM_NETWORKS
	if(ast_channel_caller(chan)->id.name.valid && ast_channel_caller(chan)->id.number.valid &&
		ast_strlen_zero(ast_channel_caller(chan)->id.number.str) && ast_strlen_zero(ast_channel_caller(chan)->id.name.str)){
		if (!ast_ucm_db_get("blacklist","Anonymous", blacklist, sizeof (blacklist)))
			bl = 1;
	}
#endif

	snprintf(buf, len, "%d", bl);
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS

#define BLACKLIST_CONF		"blacklist.conf"
#define INBOUND_BLACKLIST_CONF		"inbound_blacklist.conf"
#define OUTBOUND_BLACKLIST_CONF		"outbound_blacklist.conf"

// add for SAFE_FREE
#define SAFE_FREE(buf) if (buf) {ast_free(buf);	buf = NULL;}


struct idx_link {
	char idx[8];
	AST_LIST_ENTRY(idx_link) entry;
};

struct blacklist_link {
	char pattern[80];
	AST_LIST_HEAD(, idx_link) idx_list;
	AST_LIST_ENTRY(blacklist_link) entry;
};


struct blacklist_link *inbound_list = NULL;
struct blacklist_link *outbound_list = NULL;
static AST_LIST_HEAD(, blacklist_link) inbound_list_head;
static AST_LIST_HEAD(, blacklist_link) outbound_list_head;

static int inbound_routes_idx_list_destroy(struct blacklist_link* blacklist)
{
	struct idx_link* tmp = NULL;
	while ((tmp = AST_LIST_REMOVE_HEAD(&blacklist->idx_list, entry))) {
		SAFE_FREE(tmp);
	}
	return 0;
}

static int inbound_routes_idx_list_get(struct blacklist_link* blacklist, const char* idx)
{
	struct idx_link *pr_iter;
	if (AST_LIST_EMPTY(&(blacklist->idx_list))) {
		/*inbound_routes_idx_list == NULL means that it effective on all inbound_routes  */
		return 0;
	}
	AST_LIST_TRAVERSE(&blacklist->idx_list, pr_iter, entry) {
		ast_log(LOG_DEBUG, "[%s][%s]", idx, pr_iter->idx);
		if (NULL != idx && NULL != pr_iter->idx && 0 == strncasecmp(idx, pr_iter->idx, strlen(idx))) {
			return 0;
		}
	}
	return 1;
}

static int inbound_routes_idx_list_add(const char *inbound_routes_list, struct blacklist_link* blacklist)
{
	struct idx_link* tmp_inbound_routes_idx = NULL;
	char *tmp_idx = NULL;
	char *inbound_idx = NULL;
	char *tmp_inbound_routes_list = NULL;
	if (NULL == inbound_routes_list || NULL == blacklist) {
		return 0;
	}
	tmp_inbound_routes_list = ast_strdup(inbound_routes_list);
	inbound_idx = strtok_r(tmp_inbound_routes_list, ",", &tmp_idx);
	while(inbound_idx != NULL){
		tmp_inbound_routes_idx = ast_calloc(1, sizeof(*tmp_inbound_routes_idx));
		ast_copy_string(tmp_inbound_routes_idx->idx, inbound_idx, sizeof(tmp_inbound_routes_idx->idx));
		AST_LIST_INSERT_TAIL(&blacklist->idx_list, tmp_inbound_routes_idx, entry);
		inbound_idx = strtok_r(NULL, ",", &tmp_idx);
	}
	SAFE_FREE(tmp_inbound_routes_list);
	return 0;
}

static int inbound_blacklist_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int bl = 0;
	int effective = 0;
	char* parse = NULL;
	if (!data) {
		return -1;
	}
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(pattern);
		AST_APP_ARG(inbound_index);
	);
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);

	// bl = get_match_in_list(inbound_list_head, data);

	struct blacklist_link *pr_iter;
	AST_LIST_TRAVERSE(&inbound_list_head, pr_iter, entry) {
		effective = 0;
		if (!ast_strlen_zero(args.inbound_index)) {
			effective = inbound_routes_idx_list_get(pr_iter, args.inbound_index);
		}
		if (!ast_strlen_zero(args.pattern) && 0 == effective) {
			int ret = ast_extension_match(pr_iter->pattern, args.pattern);
			if (ret == 1 || ret == 2) {
				bl = 1;
				break;
			}
		}
	}

	if (bl != 0) {
		ast_log(LOG_DEBUG, "find in inbound black list");
	}

	// pbx_builtin_setvar_helper(chan, "BEFORE_CONTEXT", context ? : "");
	snprintf(buf, len, "%d", bl);
	return 0;
}

static int outbound_blacklist_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int bl = 0;

	if (!data) {
		return -1;
	}
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	// bl = get_match_in_list(outbound_list, data);
	struct blacklist_link *pr_iter;
	AST_LIST_TRAVERSE(&outbound_list_head, pr_iter, entry) {
		int ret = ast_extension_match(pr_iter->pattern, data);
		/*when Early match, it will return 2*/
		if (ret == 1 || ret == 2) {
			bl = 1;
			break;
		}
	}

	if (bl != 0) {
		ast_log(LOG_DEBUG, "find in outbound black list");
	}


	// pbx_builtin_setvar_helper(chan, "BEFORE_CONTEXT", context ? : "");

	snprintf(buf, len, "%d", bl);
	return 0;
}

/*!
 * \brief Check if configuration file is valid
 */
static inline int valid_config(const struct ast_config *cfg)
{
	return cfg && cfg != CONFIG_STATUS_FILEINVALID;
}


static int inbound_blacklist_remove_write(struct ast_channel *chan, const char *cmd, char *data, const char *value)
{
	if (!data) {
		return -1;
	}
	char tmp_number[256] = {0};

	struct blacklist_link *pr_iter;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&inbound_list_head, pr_iter, entry) {
		/*pattern is start with _*/
		if (pr_iter->pattern[0] == '_' && strncmp(data, pr_iter->pattern + 1, strlen(data)) == 0) {
			AST_LIST_REMOVE_CURRENT(entry);
			inbound_routes_idx_list_destroy(pr_iter);
			SAFE_FREE(pr_iter);
		}
		else if (strncmp(data, pr_iter->pattern, strlen(data)) == 0) {
			AST_LIST_REMOVE_CURRENT(entry);
			inbound_routes_idx_list_destroy(pr_iter);
			SAFE_FREE(pr_iter);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	if (data[0] != '_'){
		snprintf(tmp_number, sizeof(tmp_number), "_%s", data);
	}else{
		snprintf(tmp_number, sizeof(tmp_number), "%s", data);
	}

	ast_log(LOG_DEBUG, "delete %s to  inbound black list\n", data);

	struct ast_flags config_flags = { CONFIG_FLAG_WITHCOMMENTS };
	struct ast_config   *cfg = NULL;
	struct ast_variable *var = NULL;
	struct ast_category *cat = NULL;
	char *category = NULL;
	if ((cfg = ast_config_load(BLACKLIST_CONF, config_flags)) && valid_config(cfg)) {
		while ((category = ast_category_browse(cfg, category))) {
			if (!strcasecmp(category, "inbound")) {
				if (!ast_variable_retrieve(cfg, category, tmp_number)) {
					ast_debug(3, "looks like we need to make data[%s] = 0!\n", data);
					var = ast_variable_new(tmp_number, "0", "");
				} else {
					var = NULL;
				}
				if (!(cat = ast_category_get(cfg, category, NULL))) {
					ast_debug(4, "failed to get category!\n");
					ast_free(var);
					break;
				}
				if (!var) {
					ast_variable_update(cat, tmp_number, "0", NULL, 0);
				} else {
					ast_variable_append(cat, var);
				}
			}
		}
		ast_config_text_file_save(BLACKLIST_CONF, cfg, "func_blacklist");
	}else{
		ast_log(LOG_ERROR, "there's something wrong with config file:%s", BLACKLIST_CONF);
	}
	if (cfg){
		ast_config_destroy(cfg);
	}

	ast_ha_send_blacklist(data, AST_DB_INBOUND_BLACKLIST_DEL);

	return 0;
}

static int inbound_blacklist_add_write(struct ast_channel *chan, const char *cmd, char *data, const char *value)
{
	if (!data) {
		return -1;
	}

	struct blacklist_link *tmp = NULL;
	tmp = ast_calloc(1, sizeof(*tmp));
	AST_LIST_HEAD_INIT(&(tmp->idx_list));
	ast_copy_string(tmp->pattern, data, sizeof(tmp->pattern));
	AST_LIST_INSERT_TAIL(&inbound_list_head, tmp, entry);
	ast_log(LOG_DEBUG, "add %s to  inbound black list\n", data);
	char tmp_number[256] = {0};
	if (data[0] != '_'){
		snprintf(tmp_number, sizeof(tmp_number), "_%s", data);
	}else{
		snprintf(tmp_number, sizeof(tmp_number), "%s", data);
	}

	struct ast_flags config_flags = { CONFIG_FLAG_WITHCOMMENTS };
	struct ast_config   *cfg = NULL;
	struct ast_variable *var = NULL;
	struct ast_category *cat = NULL;
	char *category = NULL;
	if ((cfg = ast_config_load(BLACKLIST_CONF, config_flags)) && valid_config(cfg)) {
		while ((category = ast_category_browse(cfg, category))) {
			if (!strcasecmp(category, "inbound")) {
				if (!ast_variable_retrieve(cfg, category, tmp_number)) {
					ast_debug(3, "looks like we need to make data[%s] = []!\n", data);
					var = ast_variable_new(tmp_number, "", "");
				} else {
					var = NULL;
				}
				if (!(cat = ast_category_get(cfg, category, NULL))) {
					ast_debug(4, "failed to get category!\n");
					ast_free(var);
					break;
				}
				if (!var) {
					ast_variable_update(cat, tmp_number, "", NULL, 0);
				} else {
					ast_variable_append(cat, var);
				}
			}
		}
		ast_config_text_file_save(BLACKLIST_CONF, cfg, "func_blacklist");
	}else{
		ast_log(LOG_ERROR, "there's something wrong with config file:%s", BLACKLIST_CONF);
	}
	if (cfg){
		ast_config_destroy(cfg);
	}
	/* send inbound blacklist to standby ucm */
	ast_ha_send_blacklist(data, AST_DB_INBOUND_BLACKLIST_ADD);
	return 0;
}


#endif

static int blacklist_read2(struct ast_channel *chan, const char *cmd, char *data, struct ast_str **str, ssize_t len)
{
	/* 2 bytes is a single integer, plus terminating null */
	if (ast_str_size(*str) - ast_str_strlen(*str) < 2) {
		if (len > ast_str_size(*str) || len == 0) {
			ast_str_make_space(str, len ? len : ast_str_strlen(*str) + 2);
		}
	}
	if (ast_str_size(*str) - ast_str_strlen(*str) >= 2) {
		int res = blacklist_read(chan, cmd, data, ast_str_buffer(*str) + ast_str_strlen(*str), 2);
		ast_str_update(*str);
		return res;
	}
	return -1;
}

static struct ast_custom_function blacklist_function = {
	.name = "BLACKLIST",
	.read = blacklist_read,
	.read2 = blacklist_read2,
};

#ifdef GRANDSTREAM_NETWORKS
static struct ast_custom_function inbound_blacklist_function = {
	.name = "INBOUND_BLACKLIST",
	.read = inbound_blacklist_read,
};

static struct ast_custom_function outbound_blacklist_function = {
	.name = "OUTBOUND_BLACKLIST",
	.read = outbound_blacklist_read,
};

static struct ast_custom_function inbound_blacklist_remove_function = {
	.name = "INBOUND_BLACKLIST_DEL",
	.read = NULL,
	.write = inbound_blacklist_remove_write,
};

static struct ast_custom_function inbound_blacklist_add_function = {
	.name = "INBOUND_BLACKLIST_ADD",
	.read = NULL,
	.write = inbound_blacklist_add_write,
};

static void blacklist_destroy_list(void)
{
	struct blacklist_link *tmp = NULL;

	while ((tmp = AST_LIST_REMOVE_HEAD(&inbound_list_head, entry))) {
		// ao2_ref(t_eventlist, -1);
		inbound_routes_idx_list_destroy(tmp);
		ast_free(tmp);
	}
	while ((tmp = AST_LIST_REMOVE_HEAD(&outbound_list_head, entry))) {
		// ao2_ref(t_eventlist, -1);
		ast_free(tmp);
	}
	return;
}

static int load_config(int reload)
{
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	int res = 0;

	struct ast_variable *v;

	if (!(cfg = ast_config_load(BLACKLIST_CONF, config_flags))) {
		ast_log(LOG_ERROR, "Configuration file blacklist.conf missing.\n");
		return -1;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file blacklist.conf is in an invalid format.  Aborting.\n");
		return -1;
	}

	blacklist_destroy_list();

	if (cfg) {
		for (v = ast_variable_browse(cfg, "inbound"); v; v = v->next) {
			//ast_log(LOG_ERROR, "v->name = %s v->value = %s\n", v->name, v->value ? v->value : "");
			if (v->name) {
				if (v->value && strcmp(v->value, "0") == 0) {
					continue;
				}
				struct blacklist_link *tmp = NULL;
				tmp = ast_calloc(1, sizeof(*tmp));
				AST_LIST_HEAD_INIT(&(tmp->idx_list));
				inbound_routes_idx_list_add(v->value, tmp);
				ast_copy_string(tmp->pattern, v->name, sizeof(tmp->pattern));
				AST_LIST_INSERT_TAIL(&inbound_list_head, tmp, entry);
			}
		}

		for (v = ast_variable_browse(cfg, "outbound"); v; v = v->next) {
			// ast_log(LOG_ERROR, "v->name = %s v->value = %s\n", v->name, v->value ? v->value : "");
			if (v->name) {
				struct blacklist_link *tmp = NULL;
				tmp = ast_calloc(1, sizeof(*tmp));
				ast_copy_string(tmp->pattern, v->name, sizeof(tmp->pattern));
				AST_LIST_INSERT_TAIL(&outbound_list_head, tmp, entry);
			}
		}
	}

	if(cfg) {
		ast_config_destroy(cfg);
	}

	return res;
}

static int unload_module(void)
{
	int res = 0;

	blacklist_destroy_list();
	
	res |= ast_custom_function_unregister(&blacklist_function);
	res |= ast_custom_function_unregister(&inbound_blacklist_function);
	res |= ast_custom_function_unregister(&outbound_blacklist_function);
	res |= ast_custom_function_unregister(&inbound_blacklist_remove_function);
	res |= ast_custom_function_unregister(&inbound_blacklist_add_function);

	return res;
}

static int reload(void)
{
	int res = 0;

	ast_log(LOG_DEBUG, "reload module blacklist \n");
	load_config(1);

	return res;
}


static int load_module(void)
{
	int res = 0;
	res |= ast_custom_function_register(&blacklist_function);
	res |= ast_custom_function_register(&inbound_blacklist_function);
	res |= ast_custom_function_register(&outbound_blacklist_function);
	res |= ast_custom_function_register(&inbound_blacklist_remove_function);
	res |= ast_custom_function_register(&inbound_blacklist_add_function);
	load_config(0);

	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "from blacklist conf",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		);

#else

static int unload_module(void)
{
	return ast_custom_function_unregister(&blacklist_function);
}

static int load_module(void)
{
	return ast_custom_function_register(&blacklist_function);
}


AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Look up Caller*ID name/number from blacklist database");
#endif

