#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 208-2019, Zegeng Mo <zgmo@grandstream.cn>
 *
 * func_calltransfer.c source code for CONF_CALLTRANS_READ
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Functions for reading config on call transfer feature
 *
 * \author Zegeng Mo <zgmo@grandstream.cn>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

/* BEGIN:  add */
#undef AST_MODULE
#define AST_MODULE "func_calltransfer"
/*   END:  add */

#include <regex.h>

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"
#include "asterisk/haserver.h"
AST_MODULE_LOG("func_calltransfer");

/*** DOCUMENTATION
	<function name="CFG_CALLTRANS" language="en_US">
		<synopsis>
			Read from or write to the config file call_transfer.conf.
		</synopsis>
		<syntax argsep="/">
			<parameter name="family" required="true" />
			<parameter name="key" required="true" />
		</syntax>
		<description>
			<para>This function will return config about call transfer.</para>
		</description>
		<see-also>
		</see-also>
	</function>
 ***/
#define CALL_TRANSFER_CONF "/cfg/etc/asterisk_new/call_transfer.conf"

AST_MUTEX_DEFINE_STATIC(config_lock);
static struct ast_config *ct_cfg = NULL;
static int read_ct_cfg(struct ast_channel *chan, const char *cmd,
								char *data, char *buf, size_t len)
{
	const char *value;
	struct ast_category *cfg_category;

	SCOPED_MUTEX(lock, &config_lock);

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(family);
		AST_APP_ARG(key);
	);

	AST_STANDARD_APP_ARGS(args, data);
	if(!args.family || !args.key) {
		ast_log(LOG_ERROR, "Invalid param: family:\"%s\", key:\"%s\"", args.family, args.key);
		return -1;
	}

	args.family = ast_str_to_lower(args.family);
	args.key = ast_str_to_lower(args.key);

	ast_log(LOG_DEBUG, "read_ct_cfg(cmd:%s,data:%s, buf:%p, len:%ld),[family:%s, key:%s]", 
		cmd, data, buf, len, args.family, args.key);

	if(!ct_cfg) {
		ast_log(LOG_ERROR, "Config file:%s not loaded", CALL_TRANSFER_CONF);
		return -1;
	}

	cfg_category = ast_category_get(ct_cfg, args.family, NULL);
	if(cfg_category) {
		value = ast_variable_find(cfg_category, args.key);
		if(value) {			
			strncpy(buf, value, len);
			ast_log(LOG_DEBUG, "Read:[%s] %s='%s'", args.family, args.key, value);
		}else {
			ast_log(LOG_DEBUG, "Read failed:[%s] no such key:%s, set the default", args.family, args.key);
			if (args.key != NULL && (strstr(args.key, "time_type") != NULL || strstr(args.key, "dst_type") != NULL)) {
				strncpy(buf, "0", len);
			} else {
				strncpy(buf, "", len);
			}
		}
	}else {
		ast_log(LOG_WARNING, "Read failed:no such category:%s", args.family);
	}

	return 0;
}

static int write_ct_cfg(struct ast_channel *chan, const char *cmd,
								char *data, const char *buf)
{
	int ret = 0;
	struct ast_category *cfg_category;
	struct ast_variable *var = NULL;
	char time_type[126] = {0};
	char dst_type[126] = {0};
	char *ha_data = NULL;
	int len_ha = 0;

	SCOPED_MUTEX(lock, &config_lock);

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(family);
		AST_APP_ARG(key);
	);

	AST_STANDARD_APP_ARGS(args, data);
	if(!args.family || !args.key) {
		ast_log(LOG_ERROR, "Invalid param: family:\"%s\", key:\"%s\"", args.family, args.key);
		return -1;
	}

	args.family = ast_str_to_lower(args.family);
	args.key = ast_str_to_lower(args.key);
	snprintf(time_type, sizeof(time_type), "%s_time_type", args.key);
	snprintf(dst_type, sizeof(dst_type), "%s_dst_type", args.key);

	ast_log(LOG_DEBUG, "write_ct_cfg(cmd:%s,data:%s, buf:%p, len:%ld),[family:%s, key:%s]", 
		cmd, data, buf, strlen(buf), args.family, args.key);

	if(!ct_cfg) {
		ast_log(LOG_ERROR, "Config file:%s not loaded", CALL_TRANSFER_CONF);
		return -1;
	}

	len_ha = strlen(buf)+strlen(args.family)+strlen(args.key)+strlen(",,,")+1;
	ha_data = ast_malloc(len_ha);
	if(NULL == ha_data)
	{
		ast_log(LOG_ERROR, "malloc ha_data error.");
		return -1;
	}
	memset(ha_data, 0, len_ha);
	snprintf(ha_data, len_ha, "%s,%s,%s", args.family, args.key, buf);

	cfg_category = ast_category_get(ct_cfg, args.family, NULL);
	if(!cfg_category) {
		cfg_category = ast_category_new(args.family, "", 1);
		if (cfg_category) {
			ast_category_append(ct_cfg, cfg_category);
		} else {
			ast_log(LOG_ERROR, "Error creating new variable structure\n");
		}
	}

	if(cfg_category) {
		if (!ast_variable_retrieve(ct_cfg, args.family, args.key)) {
			ast_debug(3, "looks like we need to make args.key %s!\n", args.key);
			var = ast_variable_new(args.key, buf, "");
			ast_variable_append(cfg_category, var);
			var = ast_variable_new(time_type, "0", "");
			ast_variable_append(cfg_category, var);
			var = ast_variable_new(dst_type, "0", "");
			ast_variable_append(cfg_category, var);
		} else {
			ast_variable_update(cfg_category, args.key, buf, NULL, 0);
		}

		ret = ast_config_text_file_save(CALL_TRANSFER_CONF, ct_cfg, "func_calltransfer");
		if(!ret) {
			ast_ha_send_sql(ha_data, AST_HA_TF_CONF);
			ast_log(LOG_DEBUG, "Write done:[%s] %s = '%s'", args.family, args.key, buf);
		}else {
			ast_log(LOG_WARNING, "Write failed:[%s] %s = '%s'", args.family, args.key, buf);
		}
	}else {
		ret = -1;
		ast_log(LOG_WARNING, "Write failed:no such category:%s", args.family);
	}

	AST_SAFE_FREE(ha_data);
	return ret;
}

static struct ast_custom_function config_call_transfer_read_function = {
	.name = "CFG_CALLTRANS",
	.read = read_ct_cfg,
	.write = write_ct_cfg,
};

static int pares_config_file(int reload)
{
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	SCOPED_MUTEX(lock, &config_lock);

	cfg = ast_config_load(CALL_TRANSFER_CONF, config_flags);
	if (cfg == CONFIG_STATUS_FILEMISSING || cfg == CONFIG_STATUS_FILEINVALID)
		return -1;

	if(cfg == CONFIG_STATUS_FILEUNCHANGED)
		return 0;

	if(ct_cfg && ct_cfg != cfg)
		ast_config_destroy(ct_cfg);

	ct_cfg = cfg;
	ast_log(LOG_DEBUG, "Config file:%s %s", CALL_TRANSFER_CONF, reload ? "reloaded" : "loaded");

	return 0;
}

static int load_module(void)
{
	int res = 0;

	res |= pares_config_file(0);
	res |= ast_custom_function_register(&config_call_transfer_read_function);

	return res;
}

static int reload_module(void)
{
	int res = 0;
	
	res |= pares_config_file(1);
	
	return res;
}

static int unload_module(void)
{
	int res = 0;

	res |= ast_custom_function_unregister(&config_call_transfer_read_function);

	if(ct_cfg)
		ast_config_destroy(ct_cfg);

	return res;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Call transfer related dialplan functions",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload_module,
);

#endif

