#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2016-2017, jklou  <jklou@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Functions for interaction with the Asterisk config file
 *
 * \author jklou <jklou@grandstream.cn>
 *
 * \codebluegroups check functions
 */

/*** MODULEINFO
     <support_level>core</support_level>
 ***/

#include "asterisk.h"
#ifdef GRANDSTREAM_NETWORKS
#undef AST_MODULE
#define AST_MODULE "database"
#endif

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/lock.h"

/*** DOCUMENTATION
	<function name="CODEBLUE_GROUP" language="en_US">
		<synopsis>
			Check if the extensionis in the codeblue group.
		</synopsis>
		<syntax argsep="/">
			<parameter name="extension" required="2500" >
				<para>number to be compared</para>
			</parameter>
		</syntax>
		<description>
			<para>
			Check if the in the codebluegroups.
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
 ***/

AST_MUTEX_DEFINE_STATIC(list_mutex);
#define NO 0  /*NOT IN*/
#define YES 1 /*IN*/
#define ERROR -1

/*node of number under context*/
struct numbers
{
	char extension[AST_MAX_EXTENSION]; /*the extension of number*/
	char phonenum[AST_MAX_EXTENSION]; /*the phone number of extension*/
	struct numbers *next;
};

/*list of context*/
struct context_list
{
	char context[AST_MAX_EXTENSION];/*this is the [context]*/
	struct numbers *number;         /*numbers under context*/
	struct numbers *out_number;     /*additional number for later adding*/
	struct context_list *next;      /*next context*/
};

/*a global param used for data transfer*/
struct context_list *g_list = NULL;

/*find list by context*/
static struct context_list *find_list_by_context(struct context_list *list, const char *context)
{
	if (NULL == list || NULL == context ) {
		return NULL;
	}
	struct context_list *tmp_list = list;
	while ( NULL != tmp_list ) {
		ast_log(LOG_DEBUG, " context in list : [%s], our context : [%s]", tmp_list->context, context);
		if (!strncmp(tmp_list->context, context, sizeof(tmp_list->context))) {
			ast_log(LOG_DEBUG, " here find context : %s", tmp_list->context);
			break;
		}
		tmp_list = tmp_list->next;
	}
	return tmp_list;
}

/*find wether caller id is allowed*/
static struct numbers *find_numbers_by_ext(struct numbers *ext, const char *caller_id)
{
	if (NULL == ext || NULL == caller_id ) {
		ast_log(LOG_DEBUG,"error");
		return NULL;
	}
	struct numbers *tmp_ext = ext;
	while ( NULL != tmp_ext ) {
		ast_log(LOG_DEBUG, "allowed number : %s | caller id : %s", tmp_ext->extension, caller_id);
		if (!strncmp(tmp_ext->extension, caller_id, sizeof(tmp_ext->extension))) {
			ast_log(LOG_DEBUG, "here find extension number: %s", tmp_ext->extension);
			break;
		}
		tmp_ext = tmp_ext->next;
	}
	return tmp_ext;
}

/*init caller id numbers list, insert first number*/
static struct numbers *malloc_numbers(const char *caller_id)
{
	if (NULL == caller_id) {
		return NULL;
	}
	struct numbers *ext = NULL;
	ext = ast_malloc(sizeof(struct numbers));
	if ( NULL != ext ) {
		strncpy(ext->extension, caller_id, sizeof(ext->extension));
		ext->next = NULL;
	}
	return ext;
}

/*create context number list*/
static struct numbers *create_numbers_list(struct context_list *list, const char *extension)
{
	if (NULL == list || NULL == extension ) {
		return NULL;
	}

	struct numbers *node_ext = malloc_numbers(extension);
	struct numbers *tmp_ext = list->number;
	if (NULL == list->number ) {
		list->number = node_ext;
	}
	else {
		while (tmp_ext->next) {
			tmp_ext = tmp_ext->next;
		}
		tmp_ext->next = node_ext;
	}

	return node_ext;
}

/*init context list*/
static struct context_list* malloc_list(const char *context_num)
{
	if (NULL == context_num ) {
		return NULL;
	}

	struct context_list *list = NULL;
	list = ast_malloc(sizeof(struct context_list));
	if (list) {
		strncpy(list->context, context_num, sizeof(list->context));
		list->number = NULL;
		list->next = NULL;
	}

	return list;
}

/*create context number list*/
static struct context_list *create_list(struct context_list **list, const char *context)
{
	if (NULL == list || NULL == context ) {
		return NULL;
	}

	struct context_list *tmp_list = *list, *tmp = NULL;
	tmp = malloc_list(context);
	if (NULL == tmp_list ) {
		*list = tmp;
	}
	else {
		while( tmp_list->next ) {
			tmp_list = tmp_list->next;
		}
		tmp_list->next = tmp;
	}

	return tmp;
}

/*to see wether global saved right*/
static void print_global_list(void)
{
	struct context_list *tmp_list = g_list;
	struct numbers *tmp_ext = NULL;
	ast_log(LOG_DEBUG, "start to show context and extensions");
	while (tmp_list) {
		ast_debug(1, "[%s]", tmp_list->context);
		if( NULL != tmp_list->number ){
			tmp_ext = tmp_list->number;
			while (tmp_ext) {
				ast_log(LOG_DEBUG,"extension = %s", tmp_ext->extension);
				tmp_ext = tmp_ext->next;
			}
		}
		tmp_list = tmp_list->next;
	}
	ast_log(LOG_DEBUG, "end to show context and extensions");
}

static void unload_list(void)
{
	struct context_list *tmp_list = g_list, *prev_list = NULL;
	struct numbers *tmp_ext = NULL, *prev_ext = NULL;
	while (tmp_list) {
		if( NULL != tmp_list->context ){
			tmp_ext = tmp_list->number;
		}

		while (tmp_ext) {
			prev_ext = tmp_ext;
			tmp_ext = tmp_ext->next;
			ast_free(prev_ext);
		}
		prev_list = tmp_list;
		tmp_list = tmp_list->next;
		ast_free(prev_list);
	}
	g_list = NULL;
}

/*check caller id, to see wether it is allowed*/
static int check_list_privilege(const char *context, const char *ext)
{
	int ret = ERROR;
	if ( NULL == ext || NULL == context || '\0' == ext[0] || '\0' == context[0] ){
		return ret;
	}
	struct context_list *tmp_list = NULL;
	tmp_list = find_list_by_context(g_list, context);
	ast_log(LOG_DEBUG, "context : %s", tmp_list->context);
	if (!tmp_list) {
		return ret;
	}

	ret = find_numbers_by_ext(tmp_list->number, ext)? YES : NO;
	return ret;
}

/*get context and extension, using function*/
static int group_function_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(extension);
	);
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}
	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.extension)) {
		return -1;
	}
	ast_mutex_lock(&list_mutex);
	ast_log(LOG_DEBUG, "check extension = %s\n", args.extension);

	if (check_list_privilege("general", args.extension) == YES) {
		strncpy(buf, "YES", len - 1);
	} else {
		strncpy(buf, "NO", len - 1);
	}
	
	ast_mutex_unlock(&list_mutex);
	return 0;
}


static struct ast_custom_function group_function = {
	.name = "CODEBLUE_GROUP",
	.read = group_function_read,
};

static int unload_module(void)
{
	int res = 0;

	res |= ast_custom_function_unregister(&group_function);

	unload_list();

	return res;
}

/*load conf file and save data into global param*/
static int load_module(void)
{
	int res = 0;
	char *tmp = NULL;
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	struct context_list *tmp_list = NULL;
	struct ast_flags config_flags = { 0 };
	if (!(cfg = ast_config_load("codeblue_group.conf", config_flags))) {
		ast_log(LOG_NOTICE, "No config file (codeblue_group.conf)\n");
        ast_custom_function_register(&group_function);
		return 0;
	} else if (CONFIG_STATUS_FILEUNCHANGED == cfg ) {
		return 0;
	} else if (CONFIG_STATUS_FILEINVALID == cfg ) {
		ast_log(LOG_ERROR, "Config file codeblue_group.conf is in an invalid format. Aborting.\n");
		return 1;
	}

	while ((tmp = ast_category_browse(cfg, tmp)) ) {
		tmp_list = create_list(&g_list, tmp);
		for (var = ast_variable_browse(cfg, tmp); var; var = var->next) {
			//if (!strcasecmp(var->name, "extension")) {
				ast_log(LOG_DEBUG, "context : %s | extension : %s", tmp, var->value);
				create_numbers_list(tmp_list, var->value);
			//}
		}
	}
	print_global_list();
	res |= ast_custom_function_register(&group_function);

	ast_config_destroy(cfg);

	return res;
}

static int reload(void)
{
	int res = 0;

	ast_mutex_lock(&list_mutex);
	res = unload_module();
	res |= load_module();
	ast_mutex_unlock(&list_mutex);

	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "List related dialplan functions",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
);

#endif

