#ifdef GRANDSTREAM_NETWORKS

/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2016-2017, Yuancheng Pang <ychpang@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Functions for interaction relate on dialplan
 *
 * \author Yuancheng Pang <ychpang@grandstream.cn>
 *
 * \inbound call functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/



#include "asterisk.h"

#include <string.h>
#include <curl/curl.h>
#include <sqlite3.h>
#include <stdlib.h>
#include <unistd.h>
#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
//#include "asterisk/haserver.h"
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#endif
#endif
AST_MODULE_LOG("func_dialin");

/*** DOCUMENTATION
	<function name="DIALOUT_CHECK" language="en_US">
		<synopsis>
			Insert into database when extension call outbound number.
		</synopsis>
		<syntax argsep="/">
			<parameter name="src" required="4005" />
			<parameter name="dst" required="159XXXX2158" />
		</syntax>
		<description>
			<para>Insert into database when extension call outbound number.
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="DIALIN_CHECK" language="en_US">
		<synopsis>
			Check if the inbound number had been called by local number.
		</synopsis>
		<syntax argsep="/">
			<parameter name="dst" required="159XXXX2158" />
		</syntax>
		<description>
			<para>Check if the inbound number had been called by local number.
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="CKECK_PLAYBACK" language="en_US">
		<synopsis>
			For peer and pri trunk, when there's number been called, return number.
		</synopsis>
		<syntax argsep="/">
			<parameter name="dst" required="159XXXX2158" />
		</syntax>
		<description>
			<para>For peer and pri trunk, when there's number been called, return number.
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="DIAL_PLAYPROMPT" language="en_US">
		<synopsis>
			For peer and pri trunk, when there's number been called, return "yes".
		</synopsis>
		<syntax argsep="/">
			<parameter name="dst" required="159XXXX2158" />
		</syntax>
		<description>
			<para>For peer and pri trunk, when there's number been called, return "yes".
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="DIALIN_ELIN" language="en_US">
		<synopsis>
			For emergency trunk, when there's number been called, return "yes".
		</synopsis>
		<syntax argsep="/">
			<parameter name="dst" required="XXXXXXXXXX" />
		</syntax>
		<description>
			<para>Foremergency trunk, when there's number been called, return "yes".
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="DIALOUT_ELIN" language="en_US">
		<synopsis>
			For emergency trunk, when there's number been called, return "yes".
		</synopsis>
		<syntax argsep="/">
			<parameter name="dst" required="XXXXXXXXXX" />
		</syntax>
		<description>
			<para>Foremergency trunk, when there's number been called, return "yes".
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="DIALIN_CALLER_WRITE" language="en_US">
		<synopsis>
			Add the caller and callee into db for blacklist.
		</synopsis>
		<syntax argsep="/">
			<parameter name="last_caller" required="xxxx" />
			<parameter name="extension" required="xxxx" />
		</syntax>
		<description>
			<para>Save the last caller dial extension into DB
		</para>
		</description>
		<see-also>
		</see-also>
	</function>
	<function name="DIALIN_CALLER_READ" language="en_US">
		<synopsis>
			Get the last number dial to caller.
		</synopsis>
		<syntax argsep="/">
			<parameter name="caller" required="xxxx" />
		</syntax>
		<description>
			<para>Get the last number dial from DB
		</para>
		</description>
		<see-also>
		</see-also>
	</function>
 ***/

#define CMD_SIZE 1024
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
#define DB_DIALIN "dialin_direct"
#endif
#else
#define DB_DIALIN "/data/var/log/asterisk/dialin_direct.db"
#define TABLE_DIALIN "dialin_direct"
sqlite3 *dialin_db = NULL;
#endif
#define TABLE_ELIN "dialin_elin"
#define TABLE_LAST_CALLER "blacklist_late_caller"

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
struct generic_prepare_struct {
	char *sql;
	int argc;
	char **argv;
};
static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}

static int update_dialin_db(const char *cmd, struct odbc_obj *obj_db)
{
	SQLHSTMT stmt;

	if (NULL == cmd || '\0'==cmd[0]){
		return -1;
	}
	char *my_cmd = ast_strdup(cmd);
	ast_log(LOG_DEBUG, "cmd:%s", cmd);

	if (obj_db) {
		struct generic_prepare_struct gps = {.sql = my_cmd};
		stmt = ast_core_odbc_prepare_and_execute(obj_db, generic_prepare, &gps);
		if (!stmt) {
			ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", my_cmd);
			ast_free(my_cmd);
			return -1;
		} else {
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		}
		ast_free(my_cmd);
	} else {
		struct odbc_obj *obj_di = NULL;

		obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
		if (!obj_di) {
			ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
			return -1;
		}

		struct generic_prepare_struct gps = {.sql = my_cmd};
		stmt = ast_core_odbc_prepare_and_execute(obj_di, generic_prepare, &gps);
		if (!stmt) {
			ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", my_cmd);
			ast_core_odbc_release_obj(obj_di);
			obj_di = NULL;
			ast_free(my_cmd);
			return -1;
		} else {
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		}
		ast_free(my_cmd);
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
	}

	return 0;
}

/*return finded src for user, need free*/
static char *get_dial_src(const char *cmd)
{
	SQLHSTMT stmt = NULL;
	char rowdata[128];
	char *find_src= NULL;
	int res = -1;
	struct odbc_obj *obj_di = NULL;

	if (NULL == cmd || '\0'==cmd[0]){
		return NULL;
	}
	obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
	if (!obj_di) {
		ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
		return NULL;
	}


	char *my_cmd = ast_strdup(cmd);
	struct generic_prepare_struct gps = { .sql = my_cmd };

	stmt = ast_core_odbc_prepare_and_execute(obj_di, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", my_cmd);
		goto bail;
	}
	res = SQLFetch(stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", my_cmd);
		goto bail_with_handle;
	}
	res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", my_cmd);
		goto bail_with_handle;
	}
	if(rowdata != NULL)
	{
		find_src = ast_strdup(rowdata);
	}else{
		ast_log(LOG_DEBUG, "don't find src");
		goto bail_with_handle;
	}
bail_with_handle:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
bail:
	ast_core_odbc_release_obj(obj_di);
	obj_di = NULL;
	if (my_cmd){
		ast_free(my_cmd);
	}
	return find_src;
}

/*compare and return the finded number.
  *dial inbound, compare dst in db with passed src;
  *cmd is like "select dst,src from %s where trunk_index=2;"
  *return finded src, need free.
*/
static char *compare_with_src(const char *src, const char *trunk_index, int flag)
{
	SQLHSTMT stmt = NULL;
	char rowsrc[128];
	char rowdst[128];
	char tmp_cmd[512] = {0};
	char *find_src= NULL;
	char *tmp_rowdst = NULL;
	const char *tmp_src = NULL;
	int len_d = 0;
	int res = -1;
	struct odbc_obj *obj_di = NULL;

	if(NULL == src || '\0' == src[0]){
		return NULL;
	}
	obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
	if (!obj_di) {
		ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
		return NULL;
	}

	snprintf(tmp_cmd, sizeof(tmp_cmd)-1, "select dst,src from %s where trunk_index=%s", DB_DIALIN, trunk_index);
	ast_log(LOG_DEBUG, "my cmd:%s", tmp_cmd);
	char *my_cmd = strdup(tmp_cmd);
	struct generic_prepare_struct gps = { .sql = my_cmd };

	stmt = ast_core_odbc_prepare_and_execute(obj_di, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", my_cmd);
		goto bail;
	}
	res = SQLFetch(stmt);
	while(SQL_SUCCEEDED(res)){
		res = SQLGetData(stmt, 1, SQL_CHAR, rowdst, sizeof(rowdst), NULL);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", my_cmd);
			goto bail_with_handle;
		}
		if (NULL == rowdst){
			ast_log(LOG_DEBUG, "don't find rowdst");
			continue;
		}
		res = SQLGetData(stmt, 2, SQL_CHAR, rowsrc, sizeof(rowsrc), NULL);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", my_cmd);
			goto bail_with_handle;
		}
		if (NULL == rowsrc){
			ast_log(LOG_DEBUG, "don't find rowsrc");
			continue;
		}
		ast_log(LOG_DEBUG, "rowsrc:%s rowdst:%s", rowsrc, rowdst);
		/*cut off the "+" and "0" on the extension's header, both dst and src*/
		tmp_rowdst = rowdst;
		while (tmp_rowdst && (*tmp_rowdst=='+' || *tmp_rowdst=='0'))
		{
			tmp_rowdst++;
		}
		tmp_src = src;
		while (tmp_src && (*tmp_src=='+' || *tmp_src=='0'))
		{
			tmp_src++;
		}
		if (tmp_rowdst && tmp_src)
		{
			len_d = strlen(tmp_rowdst) - strlen(tmp_src);
		}
		if (len_d){
			int cmp_same = -1;
			if (len_d > 0){
				cmp_same = strncmp(tmp_rowdst+len_d, tmp_src, strlen(tmp_src));
			}else if (len_d < 0){
				cmp_same = strncmp(tmp_src-len_d, tmp_rowdst, strlen(tmp_rowdst));
			}
			if (!cmp_same){
				find_src = strdup(rowsrc);
				if (flag){
					memset(tmp_cmd, 0, sizeof(tmp_cmd));
					snprintf(tmp_cmd, sizeof(tmp_cmd)-1, "delete from %s where dst='%s' and trunk_index=%s", DB_DIALIN, rowdst, trunk_index);
					res = update_dialin_db(tmp_cmd, obj_di);
				}
				goto bail_with_handle;
			}
		}
		res = SQLFetch(stmt);
	}

bail_with_handle:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
bail:
	ast_core_odbc_release_obj(obj_di);
	obj_di = NULL;
	if (my_cmd){
		ast_free(my_cmd);
	}
	return find_src;
}
#endif
#else
/*used to reduce Database fragmentation*/
static int clearDBSpace(sqlite3 *db)
{
	int res = 0;
	char sqlcmd[512] = {0};
	char *szErrMsg = 0;
	char **dbresult = NULL;
	int nRow = 0;
	int nCol = 0;
	int id = 0;
	snprintf(sqlcmd, sizeof(sqlcmd), "select max(id) from %s;", TABLE_DIALIN);
	ast_log(LOG_DEBUG, "cmd:%s", sqlcmd);
	sqlite3_busy_timeout(db, 15000);
	res = sqlite3_get_table(db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
	if ( SQLITE_OK != res)
	{
#ifdef GRANDSTREAM_NETWORKS
		sqlite3_free(szErrMsg);
#endif
		res = -1;
	}
	if(0 != nRow && 0 != nCol && NULL != dbresult[0] && NULL != dbresult[1]){
		id = atoi(dbresult[1]);
		ast_log(LOG_DEBUG, "nRow:%d nCol:%d %s : %s", nRow, nCol, dbresult[0], dbresult[1]);
		if(0 == (id%1000))
		{
			memset(sqlcmd, 0, sizeof(sqlcmd));
			snprintf(sqlcmd, sizeof(sqlcmd), "vacuum;");
			ast_log(LOG_DEBUG, "%d : %s", id, sqlcmd);
			res = sqlite3_exec(db, sqlcmd, 0, 0, &szErrMsg);
			if ( SQLITE_OK != res){
			   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
			   sqlite3_free(szErrMsg);
			}
		}
	}
	return res;
}

/*SQL EXEC*/
static int execDBSql(sqlite3 *db, char *sqlcmd)
{
	int res = 0;
	char *zErrMsg = NULL;

	ast_log(LOG_VERBOSE, "cmd:%s", sqlcmd);
	sqlite3_busy_timeout(db, 15000);
	res = sqlite3_exec(db, sqlcmd, 0, 0, &zErrMsg);
	if ( SQLITE_OK != res){
	   ast_log(LOG_ERROR, "SQL error: %s\n", zErrMsg);
	   sqlite3_free(zErrMsg);
	   return -1;
	}
	/*sync sql to peer if ha is enable*/
	//ast_ha_send_sql(sqlcmd, AST_HA_DIALIN);
	return res;
}
#endif

/*read from db for pri and analog*/
static int dialin_function_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	char *finded_src = NULL;
#endif
#else
	char sqlcmd[512] = {0};
	char *szErrMsg = NULL;
	char **dbresult = NULL;
	int nRow = 0;
	int nCol = 0;
	int countR;
	int countC;
	int nIndex;
#endif
	int i = 0;
	int first = 0;
	int presscmd = 0;
	int firstdigittimeout = (ast_channel_pbx(chan) ? ast_channel_pbx(chan)->rtimeoutms : 20000);
	char press[AST_MAX_EXTENSION] = "";
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(src);
		AST_APP_ARG(trunk_index);
	);
#ifdef USE_DB_MYSQL
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.src) || ast_strlen_zero(args.trunk_index)) {
		return -1;
	}

	ast_set_flag(ast_channel_flags(chan), AST_FLAG_END_DTMF_ONLY);
	for(;AST_MAX_EXTENSION-1 > i;){
		if (!presscmd){
			presscmd = ast_play_and_wait(chan, "welcome");
			first = presscmd;
		}
		if (!presscmd){
			usleep(600000);/* for bug 231086 */
		}

		if (!presscmd){
			presscmd = ast_play_and_wait(chan, "press-1-or-2-to-destination");
			first = presscmd;
		}

		if (first)
		{
			press[i++] = first;
			press[i] = 0;
			if ('#' == first){
				break;
			}
			first = -1;
		}

		presscmd = ast_waitfordigit(chan, firstdigittimeout);

		ast_stopstream(chan);
		if (!presscmd)
		{
			if(0 != strlen(press)){
				snprintf(press+strlen(press), sizeof(press), "#");
			}
			else{
				ast_debug(1,"TIME OUT");
				strncpy(press, "TIMEOUT", AST_MAX_EXTENSION);
			}
			break;

		}

		press[i++] = presscmd;
		press[i] = 0;
		if ('#' == presscmd){
			break;
		}

	}

	ast_log(LOG_DEBUG, "press:%s", press);

	if(0 == strcmp(press, "TIMEOUT")) {
		strncpy(buf, press, len-1);
		return 0;
	}
	else if(0 == strcmp(press, "1#") || 0 == strcmp(press, "2#")) {
		/*skip exten and goto default || goto select*/
	}
	else{
		strncpy(buf, "INVALID", len-1);
		return 0;
	}

	ast_log(LOG_DEBUG, "inbound route number : %s trunk index : %s", args.src, args.trunk_index);
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "select src from %s where dst = '%s' and trunk_index = %s", DB_DIALIN, args.src, args.trunk_index);
	ast_log(LOG_DEBUG, "cmd:%s", my_cmd);
	finded_src = get_dial_src(my_cmd);
	if (NULL != finded_src){
		if(0 == strcmp(press, "2#")){
			strncpy(buf, finded_src, len-1);
		}
		memset(my_cmd, 0, sizeof(my_cmd));
		snprintf(my_cmd, sizeof(my_cmd), "delete from %s where dst = '%s' and trunk_index = %s", DB_DIALIN, args.src, args.trunk_index);
		update_dialin_db(my_cmd, NULL);
		ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
		ast_free(finded_src);
	} else{
		finded_src = compare_with_src(args.src, args.trunk_index, 1);
		if (NULL != finded_src){
			if(0 == strcmp(press, "2#")){
				strncpy(buf, finded_src, len-1);
			}
			ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
			ast_free(finded_src);
		}
	}
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "select src from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, args.src, args.trunk_index);
	ast_log(LOG_DEBUG, "cmd:%s", sqlcmd);
	res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
	if ( SQLITE_OK != res){
	   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
	   sqlite3_free(szErrMsg);
	}
	if ( 0 == nRow || 0 == nCol || NULL == dbresult[0] || NULL == dbresult[1]){
		/*start select 2*/
		memset(sqlcmd, 0, sizeof(sqlcmd));
		snprintf(sqlcmd, sizeof(sqlcmd), "select dst,src from %s where trunk_index=%s", TABLE_DIALIN, args.trunk_index);
		ast_log(LOG_DEBUG, "cmd:%s", sqlcmd);
		dbresult = NULL;
		sqlite3_busy_timeout(dialin_db, 15000);
		res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
		if (SQLITE_OK != res){
		   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
		   sqlite3_free(szErrMsg);
		}
		nIndex = nCol;
		ast_log(LOG_DEBUG, "the data has %d row, %d column!", nRow, nCol);
		for(countR = 0 ; countR < nRow ; countR++)
		{
			for(countC = 0 ; countC < nCol ; countC++)
			{
				if (1 == nIndex % 2 || 0 == strcmp(dbresult[nIndex], "dst"))
				{
					++nIndex;
					continue;
				}
				ast_log(LOG_DEBUG, "%s=%s", dbresult[countC],dbresult[nIndex]);
				if( strstr(args.src, dbresult[nIndex]))
				{
					if(0 == strcmp(press, "2#")){
						strncpy(buf, dbresult[nIndex+1], len-1);
					}
					memset(sqlcmd, 0, sizeof(sqlcmd));
					snprintf(sqlcmd, sizeof(sqlcmd), "delete from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, dbresult[nIndex], args.trunk_index);
					res = execDBSql(dialin_db, sqlcmd);
					ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
					break;
				}
				++nIndex;
 			}
		}

	}
	else{
		if(0 == strcmp(press, "2#")){
			strncpy(buf, dbresult[1], len-1);
		}
		memset(sqlcmd, 0, sizeof(sqlcmd));
		snprintf(sqlcmd, sizeof(sqlcmd), "delete from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, args.src, args.trunk_index);
		res = execDBSql(dialin_db, sqlcmd);
		ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
	}
#endif
	ast_log(LOG_DEBUG, "find src:%s", buf);
	return res;
}

static struct ast_custom_function dialin_function = {
	.name = "DIALIN_CHECK",
	.read = dialin_function_read,
};

/*insert into db*/
static int dialout_function_write(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	int sipcode = 0;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
#endif
#else
	char *sqlcmd;
#endif
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(src);
		AST_APP_ARG(dialstatus);
		AST_APP_ARG(saveit);
		AST_APP_ARG(trunkindex);
	);
#ifdef USE_DB_MYSQL
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif
	ast_log(LOG_DEBUG, "to save dial out bound caller and callee numbers.\n");
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}
	char *callerid_num = ast_channel_caller(chan)->ani.number.str;

	if ( !callerid_num)
	{
		ast_log(LOG_WARNING, "This is not a regular call, do not save data.\n");
		return -1;
	}
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.src) || ast_strlen_zero(args.dialstatus) || ast_strlen_zero(args.saveit) || ast_strlen_zero(args.trunkindex)) {
		return -1;
	}
	char *protocol = (char *)pbx_builtin_getvar_helper(chan, "PROTOCOL");
	if (NULL != protocol && 0 == strcasecmp(protocol, "SIP") && 0 == strcasecmp(args.dialstatus, "ANSWER")){
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		snprintf(my_cmd, sizeof(my_cmd), "delete from %s where dst='%s' and src='%s' and trunk_index='%s'", DB_DIALIN,args.src,callerid_num,args.trunkindex);
		ast_log(LOG_VERBOSE, "delete : %s", my_cmd);
		res = update_dialin_db(my_cmd, NULL);
		memset(my_cmd, 0, sizeof(my_cmd)-1);
#endif
#endif
	}
	if ( protocol && strcasecmp(protocol, "PRI") && strcasecmp(protocol, "MFC/R2") && strcasecmp(protocol, "SS7")
		&& (0 == strcasecmp(args.dialstatus, "ANSWER") ||  0 != strcasecmp(args.saveit, "yes"))){
		ast_log(LOG_VERBOSE, "would not save this data");
		return -1;
	}
	/*skip some sip code, do not skip other trunk*/
#if 0
	if (0 == strcasecmp(args.dialstatus, "404CHANUNAVAIL") ||  0 == strcasecmp(args.dialstatus, "480NOANSWER")){
		return -1;
	}
#endif
	sscanf(args.dialstatus, "%d", &sipcode);
	if (0 == sipcode || 486 == sipcode || 408 == sipcode || 487 == sipcode || 603 == sipcode || 480 == sipcode){
		/*only save these*/
	}
	else {
		if (NULL != protocol && 0 == strcasecmp(protocol, "SIP")){
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
			snprintf(my_cmd, sizeof(my_cmd), "delete from %s where dst='%s' and src='%s' and trunk_index='%s'", DB_DIALIN,args.src,callerid_num,args.trunkindex);
			ast_log(LOG_VERBOSE, "delete : %s", my_cmd);
			res = update_dialin_db(my_cmd, NULL);
			memset(my_cmd, 0, sizeof(my_cmd)-1);
#endif
#endif
		}
		ast_log(LOG_VERBOSE, "would not save this data, sipcode:%d", sipcode);
		return -1;
	}
	ast_log(LOG_DEBUG, "caller : %s callee : %s trunk index : %s", callerid_num, args.src, args.trunkindex);
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "replace into %s (dst, src, trunk_index) values('%s','%s','%s')", DB_DIALIN,args.src,callerid_num,args.trunkindex);
	res = update_dialin_db(my_cmd, NULL);
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	//insert into db
	sqlcmd = sqlite3_mprintf("replace into %s (dst, src, trunk_index) values('%s','%s','%s')", TABLE_DIALIN,args.src,callerid_num,args.trunkindex);
	res = execDBSql(dialin_db, sqlcmd);

	sqlite3_free(sqlcmd);
	//clear space in here;
	res = clearDBSpace(dialin_db);
#endif
	return res;
}

static struct ast_custom_function dialout_function = {
	.name = "DIALOUT_CHECK",
	.read = dialout_function_write,
};

/*add for peer*/
static int dialin_function_check(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	char *finded_src = NULL;
#endif
#else
	char sqlcmd[512] = {0};
	char *szErrMsg = NULL;
	char **dbresult = NULL;
	int nRow = 0;
	int nCol = 0;
	int countR;
	int countC;
	int nIndex;
#endif
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(src);
		AST_APP_ARG(trunk_index);
	);
#ifdef USE_DB_MYSQL
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.src) || ast_strlen_zero(args.trunk_index)) {
		return -1;
	}

	ast_log(LOG_VERBOSE, "inbound route number : %s trunk index : %s", args.src, args.trunk_index);

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "select src from %s where dst = '%s' and trunk_index = %s", DB_DIALIN, args.src, args.trunk_index);
	ast_log(LOG_DEBUG, "cmd:%s", my_cmd);
	finded_src = get_dial_src(my_cmd);
	if (NULL != finded_src){
		strncpy(buf, finded_src, len-1);
		memset(my_cmd, 0, sizeof(my_cmd));
		snprintf(my_cmd, sizeof(my_cmd), "delete from %s where dst = '%s' and trunk_index = %s", DB_DIALIN, args.src, args.trunk_index);
		update_dialin_db(my_cmd, NULL);
		ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
		ast_free(finded_src);
	} else{
		finded_src = compare_with_src(args.src, args.trunk_index, 1);
		if (NULL != finded_src){
			strncpy(buf, finded_src, len-1);
			ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
			ast_free(finded_src);
		}
	}
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "select src from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, args.src, args.trunk_index);
	/*sqlcmd = sqlite3_mprintf("select src from %s where dst = '%s'", TABLE_DIALIN, args.src);*/
	ast_log(LOG_VERBOSE, "cmd:%s", sqlcmd);
	res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
	if ( SQLITE_OK != res){
	   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
	   sqlite3_free(szErrMsg);
	}
	if ( 0 == nRow || 0 == nCol || NULL == dbresult[0] || NULL == dbresult[1]){
		//start select 2
		memset(sqlcmd, 0, sizeof(sqlcmd));
		snprintf(sqlcmd, sizeof(sqlcmd), "select dst,src from %s where trunk_index=%s", TABLE_DIALIN, args.trunk_index);
		ast_log(LOG_VERBOSE, "cmd:%s", sqlcmd);
		dbresult = NULL;
		res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
		if (SQLITE_OK != res){
		   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
		   sqlite3_free(szErrMsg);
		}
		nIndex = nCol;
		ast_log(LOG_VERBOSE, "the data has %d row, %d column!", nRow, nCol);
		for(countR = 0 ; countR < nRow ; countR++)
		{
			for(countC = 0 ; countC < nCol ; countC++)
			{
				if (1 == nIndex % 2 || 0 == strcmp(dbresult[nIndex], "dst"))
				{
					++nIndex;
					continue;
				}
				ast_log(LOG_VERBOSE, "%s=%s", dbresult[countC],dbresult[nIndex]);
				if( strstr(args.src, dbresult[nIndex]))
				{
					strncpy(buf, dbresult[nIndex+1], len-1);
					memset(sqlcmd, 0, sizeof(sqlcmd));
					snprintf(sqlcmd, sizeof(sqlcmd), "delete from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, dbresult[nIndex], args.trunk_index);
					res = execDBSql(dialin_db, sqlcmd);
					ast_log(LOG_DEBUG ,"get the number who has called %s : %s", args.src, buf);
					break;
				}
				++nIndex;
			}
		}

	}
	else{
		strncpy(buf, dbresult[1], len-1);
		memset(sqlcmd, 0, sizeof(sqlcmd));
		snprintf(sqlcmd, sizeof(sqlcmd), "delete from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, args.src, args.trunk_index);
		res = execDBSql(dialin_db, sqlcmd);
		ast_log(LOG_VERBOSE ,"get the number who has called %s : %s", args.src, buf);
	}
#endif
	return res;
}

static struct ast_custom_function dialcheck_function = {
	.name = "CKECK_PLAYBACK",
	.read = dialin_function_check,
};

/*add check for playback*/
static int dialin_function_play(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	char *finded_src = NULL;
#endif
#else
	char sqlcmd[512] = {0};
	char *szErrMsg = NULL;
	char **dbresult = NULL;
	int nRow = 0;
	int nCol = 0;
	int countR;
	int countC;
	int nIndex;
#endif
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(src);
		AST_APP_ARG(trunk_index);
	);

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.src) || ast_strlen_zero(args.trunk_index)) {
		return -1;
	}

	ast_log(LOG_VERBOSE, "inbound route number : %s trunk index : %s", args.src, args.trunk_index);

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "select src from %s where dst = '%s' and trunk_index = %s", DB_DIALIN, args.src, args.trunk_index);
	ast_log(LOG_DEBUG, "cmd:%s", my_cmd);
	finded_src = get_dial_src(my_cmd);
	if (NULL != finded_src){
		strncpy(buf, "yes", len-1);
		ast_log(LOG_VERBOSE,"has data");
		ast_free(finded_src);
	} else{
		finded_src = compare_with_src(args.src, args.trunk_index, 0);
		if (NULL != finded_src){
			strncpy(buf, "yes", len-1);
			ast_log(LOG_VERBOSE,"has data");
			ast_free(finded_src);
		}
	}
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "select src from %s where dst = '%s' and trunk_index = %s", TABLE_DIALIN, args.src, args.trunk_index);
	/*sqlcmd = sqlite3_mprintf("select src from %s where dst = '%s'", TABLE_DIALIN, args.src);*/
	ast_log(LOG_VERBOSE, "cmd:%s", sqlcmd);
	res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
	if ( SQLITE_OK != res){
	   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
	   sqlite3_free(szErrMsg);
	}
	if ( 0 == nRow || 0 == nCol || NULL == dbresult[0] || NULL == dbresult[1]){
		//start select 2
		memset(sqlcmd, 0, sizeof(sqlcmd));
		snprintf(sqlcmd, sizeof(sqlcmd), "select dst,src from %s where trunk_index=%s", TABLE_DIALIN, args.trunk_index);
		ast_log(LOG_VERBOSE, "cmd:%s", sqlcmd);
		dbresult = NULL;
		sqlite3_busy_timeout(dialin_db, 15000);
		res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
		if (SQLITE_OK != res){
		   ast_log(LOG_ERROR, "SQL error: %s\n", szErrMsg);
		   sqlite3_free(szErrMsg);
		}
		nIndex = nCol;
		ast_log(LOG_VERBOSE, "the data has %d row, %d column!", nRow, nCol);
		for(countR = 0 ; countR < nRow ; countR++)
		{
			for(countC = 0 ; countC < nCol ; countC++)
			{
				if (1 == nIndex % 2 || 0 == strcmp(dbresult[nIndex], "dst"))
				{
					++nIndex;
					continue;
				}
				ast_log(LOG_VERBOSE, "%s=%s", dbresult[countC],dbresult[nIndex]);
				if( strstr(args.src, dbresult[nIndex]))
				{
					strncpy(buf, "yes", len-1);
					ast_log(LOG_VERBOSE,"has data");
					break;
				}
				++nIndex;
			}
		}

	}
	else{
		strncpy(buf, "yes", len-1);
		ast_log(LOG_VERBOSE,"has data");
	}
#endif
	return res;
}

static struct ast_custom_function dialplay_function = {
	.name = "DIAL_PLAYPROMPT",
	.read = dialin_function_play,
};

static int dial_elin_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	char *finded_src = NULL;
	struct odbc_obj *obj_di = NULL;
#endif
#else	
	char sqlcmd[512] = {0};
	char *szErrMsg = NULL;
	char **dbresult = NULL;
	int nRow = 0;
	int nCol = 0;
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(elin);
	);

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
	if (!obj_di) {
		ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
		return 0;
	}
#endif
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.elin)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "select extension from %s where elin = '%s'", TABLE_ELIN, args.elin);
	finded_src = get_dial_src(my_cmd);
	if (NULL != finded_src){
		strncpy(buf, finded_src, len-1);
		memset(my_cmd, 0, sizeof(my_cmd));
		ast_free(finded_src);
	} else {
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
		return -1;
	}
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "select extension from %s where elin = '%s'", TABLE_ELIN, args.elin);
	ast_log(LOG_DEBUG, "cmd:%s", sqlcmd);
    
	res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
 	if (res != SQLITE_OK || *dbresult == NULL || *(dbresult + 1) == NULL || nRow == 0 || nCol == 0){
		sqlite3_free(szErrMsg);
		sqlite3_free_table(dbresult);
		return -1;
	}
    
	strncpy(buf, *(dbresult + 1), len);
	sqlite3_free_table(dbresult);
	dbresult = NULL;
#endif
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	ast_core_odbc_release_obj(obj_di);
	obj_di = NULL;
#endif
#endif

	return res;
}

static struct ast_custom_function dialin_elin = {
	.name = "DIALIN_ELIN",
	.read = dial_elin_read,
};

static int dial_elin_write(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	struct odbc_obj *obj_di = NULL;
#endif
#else
	char sqlcmd[512] = {0};
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(elin);
		AST_APP_ARG(extension);
	);

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
	if (!obj_di) {
		ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
		return 0;
	}
#endif
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.elin) || ast_strlen_zero(args.extension)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "replace into %s (elin, extension) values('%s','%s')", TABLE_ELIN, args.elin, args.extension);
	res = update_dialin_db(my_cmd, obj_di);
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "replace into %s (elin, extension) values('%s','%s')", TABLE_ELIN, args.elin, args.extension);
    ast_log(LOG_WARNING, "excute cmd is [%s]", sqlcmd);
    
	res = execDBSql(dialin_db, sqlcmd);
	res = clearDBSpace(dialin_db);
#endif
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	ast_core_odbc_release_obj(obj_di);
	obj_di = NULL;
#endif
#endif
	return res;
}

static struct ast_custom_function dialout_elin = {
	.name = "DIALOUT_ELIN",
	.read = dial_elin_write,
};

static int dialin_caller_write_func(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	struct odbc_obj *obj_di = NULL;
#endif
#else
	char sqlcmd[512] = {0};
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(extension);
		AST_APP_ARG(last_caller);
	);

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
	if (!obj_di) {
		ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
		return 0;
	}
#endif
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.extension) || ast_strlen_zero(args.last_caller)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "replace into %s (extension, last_caller) values('%s','%s')", TABLE_LAST_CALLER, args.extension, args.last_caller);
	res = update_dialin_db(my_cmd, obj_di);
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "replace into %s (extension, last_caller) values('%s','%s')", TABLE_LAST_CALLER, args.extension, args.last_caller);
    ast_log(LOG_WARNING, "excute cmd is [%s]", sqlcmd);

	res = execDBSql(dialin_db, sqlcmd);
	res = clearDBSpace(dialin_db);
#endif
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	ast_core_odbc_release_obj(obj_di);
	obj_di = NULL;
#endif
#endif

	return res;
}

static struct ast_custom_function dialin_caller_write = {
	.name = "DIALIN_CALLER_WRITE",
	.read = dialin_caller_write_func,
};

static int dialin_caller_read_func(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	char my_cmd[512] = {0};
	char *finded_src = NULL;
	struct odbc_obj *obj_di = NULL;
#endif
#else
	char sqlcmd[512] = {0};
	char *szErrMsg = NULL;
	char **dbresult = NULL;
	int nRow = 0;
	int nCol = 0;
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(caller);
	);

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	obj_di = ast_core_odbc_request_obj(DB_DIALIN, 0);
	if (!obj_di) {
		ast_log(LOG_WARNING, "Failed to obtain database object for '%s'!\n", DB_DIALIN);
		return 0;
	}
#endif
#else
	if (dialin_db == NULL) {
		/* Should not have loaded, but be failsafe. */
		return 0;
	}
#endif

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.caller)) {
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
#endif
#endif
		return -1;
	}
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	snprintf(my_cmd, sizeof(my_cmd), "select last_caller from %s where extension = '%s'", TABLE_LAST_CALLER, args.caller);
	finded_src = get_dial_src(my_cmd);
	if (NULL != finded_src){
		strncpy(buf, finded_src, len-1);
		memset(my_cmd, 0, sizeof(my_cmd));
		ast_free(finded_src);
	} else {
		ast_core_odbc_release_obj(obj_di);
		obj_di = NULL;
		return -1;
	}
#endif
#else
	sqlite3_busy_timeout(dialin_db, 15000);
	snprintf(sqlcmd, sizeof(sqlcmd), "select last_caller from %s where extension = '%s'", TABLE_LAST_CALLER, args.caller);
	ast_log(LOG_DEBUG, "cmd:%s", sqlcmd);

	res = sqlite3_get_table(dialin_db, sqlcmd, &dbresult, &nRow, &nCol, &szErrMsg);
	if (res != SQLITE_OK || *dbresult == NULL || *(dbresult + 1) == NULL || nRow == 0 || nCol == 0){
		sqlite3_free(szErrMsg);
		sqlite3_free_table(dbresult);
		return -1;
	}

	strncpy(buf, *(dbresult + 1), len);
	sqlite3_free_table(dbresult);
	dbresult = NULL;
#endif
#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	ast_core_odbc_release_obj(obj_di);
	obj_di = NULL;
#endif
#endif

	return res;
}

static struct ast_custom_function dialin_caller_read = {
	.name = "DIALIN_CALLER_READ",
	.read = dialin_caller_read_func,
};

static int unload_module(void)
{
	int res = 0;
	res = ast_custom_function_unregister(&dialin_function);
	res |= ast_custom_function_unregister(&dialout_function);
	res |= ast_custom_function_unregister(&dialcheck_function);
	res |= ast_custom_function_unregister(&dialplay_function);
    res |= ast_custom_function_unregister(&dialin_elin);
    res |= ast_custom_function_unregister(&dialout_elin);
    res |= ast_custom_function_unregister(&dialin_caller_write);
    res |= ast_custom_function_unregister(&dialin_caller_read);
#ifdef USE_DB_MYSQL
#else
	if (dialin_db){
		res |= sqlite3_close(dialin_db);
		dialin_db = NULL;
	}
#endif
	return res;
}

static int load_module(void)
{
	int res = 0;
	ast_log(LOG_DEBUG, "load dialin direct func");
	/*if no trunk enable dialin_direct, do not conn db*/
#ifndef USE_DB_MYSQL
	int connect = 0;
	char *tmp = NULL;
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	struct ast_flags config_flags = { 0 };
	if (!(cfg = ast_config_load("extensions_additional.conf", config_flags))) {
		ast_log(LOG_NOTICE, "No config file (extensions_additional.conf)\n");
		return 1;
	} else if (CONFIG_STATUS_FILEUNCHANGED == cfg ) {
		return 0;
	} else if (CONFIG_STATUS_FILEINVALID == cfg ) {
		ast_log(LOG_ERROR, "Config file extensions_additional.conf is in an invalid format. Aborting.\n");
		return 1;
	}
	while ((tmp = ast_category_browse(cfg, tmp)) ) {
		if (NULL != strcasestr(tmp, "outbound-route-") || NULL != strcasestr(tmp, "emergency-outbounds")){
			for (var = ast_variable_browse(cfg, tmp); var; var = var->next) {
				if (!strcasecmp(var->name, "exten")) {
					if (NULL != strcasestr(var->value, "__SAVETODIALIN=yes") || NULL != strcasestr(var->value, "HASELIN=yes")){
						ast_log(LOG_DEBUG, "context : %s | extension : %s", tmp, var->value);
#endif
#ifdef USE_DB_MYSQL
#else
						res = sqlite3_open(DB_DIALIN, &dialin_db);
						if (SQLITE_OK != res) {
							ast_log(LOG_ERROR, "Could not open database %s.\n", DB_DIALIN);
							return -1;
						}
						connect = 1;
						ast_debug(1, "open dialin_direct db success\n");
						break;
					}
				}
			}
		}

		if (1 == connect){
			break;
		}
	}
	ast_config_destroy(cfg);
#endif
	res |= ast_custom_function_register(&dialin_function);
	res |= ast_custom_function_register(&dialout_function);
	res |= ast_custom_function_register(&dialcheck_function);
	res |= ast_custom_function_register(&dialplay_function);
	res |= ast_custom_function_register(&dialin_elin);
	res |= ast_custom_function_register(&dialout_elin);
	res |= ast_custom_function_register(&dialin_caller_write);
	res |= ast_custom_function_register(&dialin_caller_read);
	return res;
}

static int reload_module(void)
{
	int res = 0;
#ifndef USE_DB_MYSQL
	res = unload_module();
	res |= load_module();
#endif
	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Check dial number had been dialed",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload_module,
);

#endif
