#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2014-2015, Junkai Lou <jklou@grandstream.cn>
 *
 * func_matchcallerid.c adapted from the matchcallerid for MATCH_CALLERID
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Functions for interaction with the Asterisk database
 *
 * \author Junkai Lou <jklou@grandstream.cn>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

/* BEGIN:  add */
#undef AST_MODULE
#define AST_MODULE "func_matchcallerid"
/*   END:  add */

#include <regex.h>

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"

/*** DOCUMENTATION
	<function name="MATCH_CALLERID" language="en_US">
		<synopsis>
			Read from or write to the Asterisk database.
		</synopsis>
		<syntax argsep="/">
			<parameter name="callee" required="true" />
			<parameter name="caller" required="true" />
		</syntax>
		<description>
			<para>This function will check match callerid
			function, return the context.</para>
		</description>
		<see-also>
		</see-also>
	</function>
 ***/

AST_MODULE_LOG("func_matchcallerid");

#define MATCH_CALLERID_CONF "match_callerid.conf"

//static struct ast_config *match_callerid_cfg = NULL;
#define SAFE_FREE(buf) if (buf) {ast_free(buf);	buf = NULL;}


ast_rwlock_t tablelock;

struct cidcallers {
	char caller[64];
	AST_LIST_ENTRY(cidcallers) entry;
};

struct match_callee {
	char exten[64];
	int matchcid;
	char context[64];
	char goto_context[128];
	// AST_LIST_HEAD(, cidcallers)caller_head;
	AST_LIST_ENTRY(match_callee) entry;
};

struct context_callers {
	char context[128];
	AST_LIST_HEAD(, cidcallers)caller_head;
	AST_LIST_ENTRY(context_callers) entry;
};

struct caller_context {
	char caller[64];
	char context[64];
	AST_LIST_ENTRY(caller_context) entry;
};


static AST_LIST_HEAD(, match_callee) match_callee_head;
static AST_LIST_HEAD(, context_callers) context_callers_head;

static int ext_cmp_pattern_pos(const char **p, unsigned char *bitwise)
{
#define BITS_PER	8	/* Number of bits per unit (byte). */
	unsigned char c;
	unsigned char cmin;
	int count;
	const char *end;

	do {
		/* Get character and advance. (Ignore '-' chars as eye candy fluff.) */
		do {
			c = *(*p)++;
		} while (c == '-');

		/* always return unless we have a set of chars */
		switch (c) {
		default:
			/* ordinary character */
			bitwise[c / BITS_PER] = 1 << ((BITS_PER - 1) - (c % BITS_PER));
			return 0x0100 | c;

		case 'n':
		case 'N':
			/* 2..9 */
			bitwise[6] = 0x3f;
			bitwise[7] = 0xc0;
			return 0x0800 | '2';

		case 'x':
		case 'X':
			/* 0..9 */
			bitwise[6] = 0xff;
			bitwise[7] = 0xc0;
			return 0x0A00 | '0';

		case 'z':
		case 'Z':
			/* 1..9 */
			bitwise[6] = 0x7f;
			bitwise[7] = 0xc0;
			return 0x0900 | '1';

		case '.':
			/* wildcard */
			return 0x18000;

		case '!':
			/* earlymatch */
			return 0x28000;	/* less specific than '.' */

		case '\0':
			/* empty string */
			*p = NULL;
			return 0x30000;

		case '[':
			/* char set */
			break;
		}
		/* locate end of set */
		end = strchr(*p, ']');

		if (!end) {
			ast_log(LOG_WARNING, "Wrong usage of [] in the extension\n");
			return 0x40000;	/* XXX make this entry go last... */
		}

		count = 0;
		cmin = 0xFF;
		for (; *p < end; ++*p) {
			unsigned char c1;	/* first char in range */
			unsigned char c2;	/* last char in range */

			c1 = (*p)[0];
			if (*p + 2 < end && (*p)[1] == '-') { /* this is a range */
				c2 = (*p)[2];
				*p += 2;    /* skip a total of 3 chars */
			} else {        /* individual character */
				c2 = c1;
			}
			if (c1 < cmin) {
				cmin = c1;
			}
			for (; c1 <= c2; ++c1) {
				unsigned char mask = 1 << ((BITS_PER - 1) - (c1 % BITS_PER));

				/*
				 * Note: If two character sets score the same, the one with the
				 * lowest ASCII values will compare as coming first.  Must fill
				 * in most significant bits for lower ASCII values to accomplish
				 * the desired sort order.
				 */
				if (!(bitwise[c1 / BITS_PER] & mask)) {
					/* Add the character to the set. */
					bitwise[c1 / BITS_PER] |= mask;
					count += 0x100;
				}
			}
		}
		++*p;
	} while (!count);/* While the char set was empty. */
	return count | cmin;
}

/*!
 * \internal
 * \brief Comparison of exten patterns.
 *
 * \param left Pattern to compare.
 * \param right Pattern to compare.
 *
 * \retval <0 if left < right
 * \retval =0 if left == right
 * \retval >0 if left > right
 * if ext_cmp_pattern < 0, left priority is high
 */
static int ext_cmp_pattern(const char *left, const char *right)
{
	int cmp;
	int left_pos;
	int right_pos;

	for (;;) {
		unsigned char left_bitwise[32] = { 0, };
		unsigned char right_bitwise[32] = { 0, };

		left_pos = ext_cmp_pattern_pos(&left, left_bitwise);
		right_pos = ext_cmp_pattern_pos(&right, right_bitwise);
		cmp = left_pos - right_pos;
		if (!cmp) {
			/*
			 * Are the character sets different, even though they score the same?
			 *
			 * Note: Must swap left and right to get the sense of the
			 * comparison correct.  Otherwise, we would need to multiply by
			 * -1 instead.
			 */
			cmp = memcmp(right_bitwise, left_bitwise, ARRAY_LEN(left_bitwise));
		}
		if (cmp) {
			break;
		}
		if (!left) {
			/*
			 * Get here only if both patterns ended at the same time.  cmp
			 * would be non-zero if only one pattern ended.
			 */
			break;
		}
	}
	return cmp;
}

/* need to free */
static char* find_callee_context(struct ast_channel *chan, char *callee, char *caller)
{
	if (!callee) {
		return NULL;
	}
	char *context = NULL;
	// struct match_callee* tmp = NULL;

	AST_LIST_HEAD(, caller_context) caller_context_head;
	AST_LIST_HEAD_INIT(&caller_context_head);

	char callee_match_first[64] = {0};
	struct match_callee *pr_iter;
	AST_LIST_TRAVERSE(&match_callee_head, pr_iter, entry) {
		int res = ast_extension_match(pr_iter->exten, callee);
		char matched_caller[64] = {0};
		int not_insert = 0;
		int callee_match = 0;
		int caller_match = 0;
		if (res == 1 || res == 2) {
			callee_match = 1;
			if (callee_match_first[0] != '\0' && ext_cmp_pattern(pr_iter->exten, callee_match_first) != 0)
			{
				not_insert = 1;
			}
			ast_copy_string(callee_match_first, pr_iter->exten, sizeof(callee_match_first));

			if (!caller || caller[0] == '\0') {
				caller_match = 1;
				ast_log(LOG_DEBUG, "no caller, match\n");
			} else {
				int find_context_in_list = 0;
				struct context_callers *cc;
				AST_LIST_TRAVERSE(&context_callers_head, cc, entry) {
					if (strcmp(cc->context, pr_iter->context) == 0) {
						if (AST_LIST_EMPTY(&cc->caller_head)) {
							ast_copy_string(matched_caller, "empty", sizeof(matched_caller));
							caller_match = 1;
							not_insert = 0;
						} else {
							struct cidcallers *tmp_caller = NULL;
							AST_LIST_TRAVERSE(&cc->caller_head,tmp_caller, entry) {
								ast_log(LOG_DEBUG, "[%s]tmp_caller->caller = %s, caller = %s\n", cc->context, tmp_caller->caller, caller);
								int ret = ast_extension_match(tmp_caller->caller, caller);
								/*modify for Bug 202478*/
								/*const char *conf_room = pbx_builtin_getvar_helper(chan, "CONF_ROOM");
								const char *conf_invite = pbx_builtin_getvar_helper(chan, "CONF_INVITE");
								if (!ast_strlen_zero(conf_room) && !ast_strlen_zero(conf_invite)) {
									ret = 1;
									ast_log(LOG_DEBUG, "conference members no need do match caller");
								}*/
								ast_copy_string(matched_caller, tmp_caller->caller, sizeof(matched_caller));
								if (ret == 1 || ret == 2) {
									caller_match = 1;
									/* if it has match caller id , we need insert it's context */
									not_insert = 0;
									break;
								}
							}
						}

						find_context_in_list = 1;
					}
					if (find_context_in_list == 1) {
						break;
					}
				}


			}
		}
		if (callee_match == 1 && caller_match == 1) {
			if (not_insert == 1) {
				break;
			}
			struct caller_context *tmp2 = NULL;
			tmp2 = ast_calloc(1, sizeof(*tmp2));

			if (matched_caller[0] == '\0') {
				ast_copy_string(tmp2->caller, "empty", sizeof(tmp2->caller));
			} else {
				ast_copy_string(tmp2->caller, matched_caller, sizeof(tmp2->caller));
			}
			ast_copy_string(tmp2->context, pr_iter->goto_context, sizeof(tmp2->context));
			ast_log(LOG_DEBUG, "caller = [%s]\t context = [%s]\n", tmp2->caller, tmp2->context);
			// Caller context head list needn't be sort
			//AST_LIST_INSERT_SORTFUNC(&caller_context_head, tmp2, entry, caller, ext_cmp_pattern);
			AST_LIST_INSERT_TAIL(&caller_context_head, tmp2, entry);
			// break;
		}
		else if (callee_match  == 1 && caller_match == 0) {
			pbx_builtin_setvar_helper(chan, "MTCAH_CALLERID_STATUS", "CALLER_NOT_ALLOWED");
		}
	}

	struct caller_context *ptr_iter;
	AST_LIST_TRAVERSE(&caller_context_head, ptr_iter, entry) {
		ast_log(LOG_DEBUG, "caller = [%s], context = [%s]\n", ptr_iter->caller, ptr_iter->context);
		context = ast_strdup(ptr_iter->context);
		break;
	}

	struct caller_context *cur = NULL;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&caller_context_head, cur, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(cur);
	}
	AST_LIST_TRAVERSE_SAFE_END;

	return context;
}


static int matchcallerid_function_read(struct ast_channel *chan, const char *cmd,
			char *data, char *buf, size_t len)
{
	int res = 0;
	char *saveptr = NULL;

	if (!data) {
		return -1;
	}
	char *callee = strtok_r(data, ",", &saveptr);
	char *caller = strtok_r(NULL, ",", &saveptr);

	char * context = find_callee_context(chan, callee, caller ? caller : "");

	ast_log(LOG_DEBUG, "read matchcallerid function caller = %s, callee = %s, context = [%s]\n!",
		caller, callee, context);

	// pbx_builtin_setvar_helper(chan, "BEFORE_CONTEXT", context ? : "");
	/*  outbound-route-4, _XXXX, trunkindex_strip_prepend ,trunkgroup index */

	strncpy(buf, context ? context : "", len);

	context ? ast_free(context) : -1;

	return res;
}


static void free_match_list(void)
{
	struct match_callee *tmp = NULL;
	while ((tmp = AST_LIST_REMOVE_HEAD(&match_callee_head, entry))) {
		ast_free(tmp);
	}
}

static void free_context_list(void)
{
	struct context_callers *tmp = NULL;
	while ((tmp = AST_LIST_REMOVE_HEAD(&context_callers_head, entry))) {
		struct cidcallers *tmp2 = NULL;
		while ((tmp2 = AST_LIST_REMOVE_HEAD(&tmp->caller_head, entry))) {
			ast_free(tmp2);
		}
		ast_free(tmp);
	}
}


/*
 * 1: has, 0: not
 */
static int is_has_context(char *context)
{
	int res = 0;
	struct context_callers *ptr_iter;
	AST_LIST_TRAVERSE(&context_callers_head, ptr_iter, entry) {
		if (strcmp(ptr_iter->context, context) == 0) {
			res = 1;
			break;
		}
	}
	return res;
}

static int create_or_find_callee(struct ast_variable *v, struct ast_config *cfg)
{
	if (!v->name || !v->value) {
		return -1;
	}
	struct ast_variable *vv;
	struct match_callee *callee = NULL;
	// struct ast_config *cfg = NULL;
	callee = ast_calloc(1, sizeof(*callee));
	ast_copy_string(callee->exten, v->name, sizeof(callee->exten));
	ast_copy_string(callee->goto_context, v->value, sizeof(callee->goto_context));/*  outbound-route-4, _XXXX, trunkindex_strip_prepend ,trunkgroup index */

	char tmp[128] =  {0};
	char *saveptr = NULL;
	snprintf(tmp, sizeof(tmp), "%s", v->value);
	char *tmp_context = strtok_r(tmp, ",", &saveptr);
	// ast_log(LOG_ERROR,"tmp = %s, tmp_context =%s\n", tmp, tmp_context);
	ast_copy_string(callee->context, tmp_context, sizeof(callee->context));

	int has_context = is_has_context(tmp_context);
	if (!has_context) {
		struct context_callers *context_callers = NULL;
		context_callers = ast_calloc(1, sizeof(*context_callers));
		ast_copy_string(context_callers->context, tmp_context, sizeof(context_callers->context));
		for (vv = ast_variable_browse(cfg, tmp_context); vv; vv = vv->next) {
			struct cidcallers *tmp2 = NULL;
			tmp2 = ast_calloc(1, sizeof(*tmp2));
			ast_copy_string(tmp2->caller, vv->name, sizeof(tmp2->caller));
			//AST_LIST_INSERT_SORTFUNC(&callee->caller_head, tmp2, entry, caller, ext_cmp_pattern);
			//AST_LIST_INSERT_TAIL(&callee->caller_head, tmp2, entry);
			AST_LIST_INSERT_TAIL(&context_callers->caller_head, tmp2, entry);
		}
		AST_LIST_INSERT_TAIL(&context_callers_head, context_callers, entry);
	}

	//AST_LIST_INSERT_TAIL(&match_callee_head, callee, entry);
	AST_LIST_INSERT_SORTFUNC(&match_callee_head, callee, entry, exten, ext_cmp_pattern);

	return 0;
}

static struct ast_custom_function matchcallerid_function = {
	.name = "MATCH_CALLERID",
	.read = matchcallerid_function_read,
};



static int load_config(int reload)
{
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	int res = 0;

	struct ast_variable *v;

	if (!(cfg = ast_config_load(MATCH_CALLERID_CONF, config_flags))) {
		ast_log(LOG_ERROR, "Configuration file match_callerid.conf missing.\n");
		return -1;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_log(LOG_DEBUG, "Config file unchanged.\n");
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file match_callerid.conf is in an invalid format.  Aborting.\n");
		return -1;
	}

	free_match_list();
	free_context_list();

	if (cfg) {
		for (v = ast_variable_browse(cfg, "general"); v; v = v->next) {
			ast_log(LOG_DEBUG, "v->name = %s v->value = %s\n", v->name, v->value ? v->value : "");

			if ((res = create_or_find_callee(v, cfg))) {
				break;
			}
		}
	} else {
		ast_log(LOG_ERROR, "Config file hdl is NULL.\n");
	}

	if(cfg) {
		ast_config_destroy(cfg);
	}

	return res;
}


static int unload_module(void)
{
	int res = 0;

	ast_log(LOG_DEBUG, "Unload module matchcallerid\n");
	free_match_list();
	free_context_list();

	res |= ast_custom_function_unregister(&matchcallerid_function);

	// ast_hashtab_destroy(callees_table, match_callee_free);

	return res;
}

static int reload(void)
{
	int res = 0;

	ast_log(LOG_DEBUG, "reload module matchcallerid\n");
	load_config(1);

	return res;
}

static int load_module(void)
{
	int res = 0;

	res |= ast_custom_function_register(&matchcallerid_function);
	load_config(0);

	return res;
}

//AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Record check related dialplan functions");
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "func matchcallerid",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		);



#endif

