#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2014-2015, Junkai Lou <jklou@grandstream.cn>
 *
 * func_recordcheck.c adapted from the recordcheck for RECORD_CHECK
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Functions for interaction with the Asterisk database
 *
 * \author Junkai Lou <jklou@grandstream.cn>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

/* BEGIN:  add */
#undef AST_MODULE
#define AST_MODULE "func_recordcheck"
/*   END:  add */

#include <regex.h>
#include <stdio.h>

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/astdb.h"

AST_MODULE_LOG("func_recordcheck");

/*** DOCUMENTATION
	<function name="RECORD_CHECK" language="en_US">
		<synopsis>
			Read from or write to the config file: auto_record.conf.
		</synopsis>
		<syntax argsep="/">
			<parameter name="type" required="true" />
			<parameter name="query" required="true" />
		</syntax>
		<description>
			<para>This function will check if the dialog will be recorded
			function.</para>
		</description>
		<see-also>
		</see-also>
	</function>
 ***/
#define RECORD_CHECK_CONF "/cfg/etc/asterisk_new/auto_record.conf"

enum rc_call_region_stat
{
	RC_REGION_INTERNAL = 0,
	RC_REGION_INBOUND,
	RC_REGION_OUTBOUND,
};

AST_MUTEX_DEFINE_STATIC(config_lock);
static struct ast_config *record_check_cfg = NULL;

static char* read_conf_val(const char* category, const char* key)
{
	const char *value = NULL;
	char *res = NULL;
	struct ast_category *cfg_category = NULL;

	SCOPED_MUTEX(lock, &config_lock);

	if(!record_check_cfg)
		return NULL;

	cfg_category = ast_category_get(record_check_cfg, category, NULL);
	if (cfg_category) {
		value = ast_variable_find(cfg_category, key);
		if (value)
			res = ast_strdup(value);
	}

	return res;
}

static char* read_conf_val2(const char* category, const char* key)
{
	const char *value = NULL;
	char *res = NULL;
	struct ast_category *cfg_category = NULL;

	SCOPED_MUTEX(lock, &config_lock);

	if(!record_check_cfg)
		return NULL;

	cfg_category = ast_category_get(record_check_cfg, category, NULL);
	if (cfg_category) {
		value = ast_variable_find(cfg_category, key);
		if (value)
			res = (char *)value;
	}

	return res;
}

static int judge_conf_val(struct ast_channel *chan, char *val)
{
	int ret = 0;
	enum rc_call_region_stat record_check_call_region;
	const char *cv_from_trunk;
	const char *cv_leave_trunk;

	if (!chan || !val) {
		ast_log(LOG_ERROR, "Invalid param! chan:%p, val:%p", chan, val);
		return 0;
	}

	if (!strcasecmp("all", val))
		return 1;

	if (!strcasecmp("off", val) || !strcasecmp("no", val))
		return 0;

	cv_from_trunk = pbx_builtin_getvar_helper(chan, "FROM_TRUNK");
	cv_leave_trunk = pbx_builtin_getvar_helper(chan, "LEAVE_TRUNK");

	if (cv_from_trunk && !cv_leave_trunk) {
		record_check_call_region = RC_REGION_INBOUND;
	} else if (cv_leave_trunk && !cv_from_trunk) {
		record_check_call_region = RC_REGION_OUTBOUND;
	} else if (!cv_from_trunk && !cv_leave_trunk) {
		record_check_call_region = RC_REGION_INTERNAL;
	} else {
		ast_log(LOG_ERROR, "Impossible : FROM_TRUNK=:%s, LEAVE_TRUNK=%s", cv_from_trunk, cv_leave_trunk);
		return 0;
	}

	switch(record_check_call_region) {
	case RC_REGION_INBOUND:
	case RC_REGION_OUTBOUND:
		if (!strcasecmp("external", val))
			ret = 1;
		break;

	case RC_REGION_INTERNAL:
		if (!strcasecmp("internal", val))
			ret = 1;
		break;
	default:
		break;
	}

	return ret;
}

static int judge_conf_val2(char *val, int outbound)
{
	int ret = 0;

	if (!val) {
		ast_log(LOG_ERROR, "Invalid param! val:%p", val);
		return 0;
	}

	if (!strcasecmp("all", val))
		return 1;

	if (!strcasecmp("off", val) || !strcasecmp("no", val))
		return 0;

	if(outbound) {
		if (!strcasecmp("external", val)) {
			ret = 1;
		}
	} else {
		if (!strcasecmp("internal", val)) {
			ret = 1;
		}
	}

	return ret;
}

static int judge_extension(struct ast_channel *chan, char *extension)
{
	int ret = 0;
	char *val;

	if(!record_check_cfg || !chan || !extension) {
		ast_log(LOG_ERROR, "invalid param or no auto_record_config! record_check_cfg:%p, extension:%s", 
			record_check_cfg, extension);
		return 0;
	}

	val = read_conf_val("extension", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension:%s val:%s", 
			extension, val);
		ret = judge_conf_val(chan, val);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}

	val = read_conf_val("ringgroup", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension(ringgroup):%s val:%s", 
			extension, val);
		ret = judge_conf_val(chan, val);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}

	val = read_conf_val("queue", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension(queue):%s val:%s", 
			extension, val);
		ret = judge_conf_val(chan, val);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}

	val = read_conf_val("paginggroup", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension(paginggroup):%s val:%s", 
			extension, val);
		ret = judge_conf_val(chan, val);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}

END:
	if (val) {
		ast_free(val);
	}

	return ret;
}

static int judge_extension2(struct ast_channel *chan, char *extension, int outbound)
{
	int ret = 0;
	char *val;

	if(!record_check_cfg || !chan || !extension)
		return 0;

	val = read_conf_val("extension", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension:%s val:%s", extension, val);
		ret = judge_conf_val2(val, outbound);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}

	val = read_conf_val("ringgroup", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension(ringgroup):%s val:%s", extension, val);
		ret = judge_conf_val2(val, outbound);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}
	
	/* add by wylu */
	val = read_conf_val("paginggroup", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension(paginggroup):%s val:%s", extension, val);
		ret = judge_conf_val2(val, outbound);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}
	/* end */

	val = read_conf_val("queue", extension);
	if (val) {
		ast_log(LOG_DEBUG, "extension(queue):%s val:%s", extension, val);
		ret = judge_conf_val2(val, outbound);
		if (ret) {
			goto END;
		}

		ast_free(val);
		val = NULL;
	}

END:
	if (val) {
		ast_free(val);
	}

	return ret;
}

static int judge_trunk(char *trunk)
{
	int ret = 0;
	char *val;

	if (!record_check_cfg || !trunk)
		return 0;

	val = read_conf_val("trunk", trunk);
	if (val) {
		if (!strcasecmp(val, "yes")) {
			ret = 1;
		}

		ast_free(val);
	}

	return ret;
}

static int judge_is_outbound_call(char *caller, char *callee)
{
	int ret;
	int caller_exist = 0;
	int callee_exist = 0;

	if (!caller || !callee) {
		return -1;
	}

	if(read_conf_val2("extension", caller) 
		|| read_conf_val2("ringgroup", caller)
		|| read_conf_val2("queue", caller)
		|| read_conf_val2("paginggroup", caller)) {
		caller_exist = 1;
	}

	if(read_conf_val2("extension", callee)
		|| read_conf_val2("ringgroup", callee)
		|| read_conf_val2("queue", callee)
		|| read_conf_val2("paginggroup", callee)) {
		callee_exist = 1;
	}

	if (caller_exist && callee_exist) {
		ret = 0;
	} else {
		ret = 1;
	}

	ast_log(LOG_DEBUG, "[judge_is_outbound_call]outbound:%d(caller_exist:%d, callee_exist:%d)",
		ret, caller_exist, callee_exist);

	return ret;
}

static int judge_dir_media(char *caller, char *callee)
{
	char family[32] = {0};
	char key[32] = {0};
	char value[32] = {0};
	int caller_dm_enable = 0;
	int callee_dm_enable = 0;

	snprintf(key, sizeof(key), "DIRECTMEDIA");

	snprintf(family, sizeof(family), "AAUSER/%s", caller);
	if (ast_ucm_db_get(family, key, value, 32-1)) {
		ast_debug(1, "DB: %s/%s not found in database.\n", family, key);
	} else {
		if (strcasecmp(value, "yes") == 0) {
			caller_dm_enable = 1;
		}
	}

	memset(family, 0, sizeof(family));

	snprintf(family, sizeof(family), "AAUSER/%s", callee);
	if (ast_ucm_db_get(family, key, value, 32-1)) {
		ast_debug(1, "DB: %s/%s not found in database.\n", family, key);
	} else {
		if (strcasecmp(value, "yes") == 0) {
			callee_dm_enable = 1;
		}
	}

	return (caller_dm_enable && callee_dm_enable) ? 0 : 1;
}

static int judge_call(struct ast_channel *chan, char *user_list)
{
	const char *chan_var_transparentsdp;
	const char *chan_var_from_ivr;
	const char *chan_var_from_directory;

	int ret = 0;

	char *caller;
	char *callee;
	char *user_save;
	char *user;

	if(!chan || !user_list)
		return 0;

	user_save = ast_strdup(user_list);
	if (user_save) {
		user = user_save;
	} else {
		ast_log(LOG_ERROR, "Fail to alloc memory!");
		return 0;
	}

	/*  xxxx-xxxx or xxxx-xxxxx-xxxx */
	caller = user;
	callee = strrchr(user, '-');
	if (callee) {
		*callee = '\0';
		callee++;
	}

	ast_log(LOG_DEBUG, "[judge_call]caller:%s, callee:%s", caller, callee);

	if (!caller || !callee)
		goto END;

	if (0 == strcasecmp(caller, "schedulepaging/intercom") || 0 == strcasecmp(callee, "schedulepaging/intercom"))
	{
		ast_log(LOG_DEBUG, "Don't need recording!");
		goto END;
	}
	
	chan_var_transparentsdp = pbx_builtin_getvar_helper(chan, "TRANSPARENTSDP");
	chan_var_from_ivr = pbx_builtin_getvar_helper(chan, "FROM_IVR");
	chan_var_from_directory = pbx_builtin_getvar_helper(chan, "FROM_DIRECTORY");

	/*
	 *  if caller and callee both enable "dir_media", they may be
	 *  transparent stream directory to each other(bypass UCM), so auto-recording may
	 *  not work, so disable "autoRecord".
	 */
	if (!((chan_var_transparentsdp && !strcasecmp(chan_var_transparentsdp, "no"))
		||(chan_var_from_ivr && !strcasecmp(chan_var_from_ivr, "yes"))
		||(chan_var_from_directory && !strcasecmp(chan_var_from_directory, "yes")))) {
		if (!(ret = judge_dir_media(caller, callee))) {
			goto END;  /* disable autoRecord */
		}
	}

#if 1
	int is_outbound = judge_is_outbound_call(caller, callee);
	ret = judge_extension2(chan, caller, is_outbound);
	ret |= judge_extension2(chan, callee, is_outbound);
	ret = ret ? 1 : 0;
#else
	ret = (judge_extension(chan, caller) || judge_extension(chan, callee)) ? 1 : 0;
#endif

END:
	if (user_save)
		ast_free(user_save);

	return ret;
}

static int check_global_enable(void)
{
	int ret = 0;
	char *global_enable;

	global_enable = read_conf_val("general", "enable");

	if (global_enable) {
		ret = (!strcasecmp(global_enable, "yes")) ? 1 : 0;
		ast_free(global_enable);
	}

	return ret;
}

static int judge_route(char *route)
{
	int ret = 0;
	char *val;

	if (!record_check_cfg || !route)
		return 0;

	val = read_conf_val("route", route);
	ast_log(LOG_DEBUG, "judge_route(%s) = %s", route, val);

	if (val) {
		if (!strcasecmp(val, "yes")) {
			ret = 1;
		}

		ast_free(val);
	}

	return ret;
}

static int judge_ivr(char *ivr)
{
	int ret = 0;
	char *val;

	if (!record_check_cfg || !ivr)
		return 0;

	val = read_conf_val("ivr", ivr);
	ast_log(LOG_DEBUG, "judge_ivr(%s) = %s", ivr, val);

	if (val) {
		if (!strcasecmp(val, "yes")) {
			ret = 1;
		}

		ast_free(val);
	}

	return ret;
}

static int judge_ringgroup(struct ast_channel *chan, char *ringgroup_exten, char *user_list)
{
	int ret = 0;

	char *caller;
	char *callee;
	char *user_save;
	char *user;

	user_save = ast_strdup(user_list);
	if (user_save) {
		user = user_save;
	} else {
		ast_log(LOG_ERROR, "Fail to alloc memory!");
		return 0;
	}

	caller = strsep(&user, "-");
	callee = strsep(&user, "-");

	ast_log(LOG_DEBUG, "[judge_ringgroup]caller:%s, callee:%s", caller, callee);

	if (!caller || !callee)
		goto END;

	int is_outbound = judge_is_outbound_call(caller, callee);
	ret = judge_extension2(chan, ringgroup_exten, is_outbound);
	ret = ret ? 1 : 0;

END:
	if (user_save)
		ast_free(user_save);

	return ret;
}

static int check_recordfile_name(char *file_name, char *buf, size_t len)
{
	int ret;
	char real_path[256] = {0};
	char *file = ast_strdup(file_name);
	const char *mixmon_dir = pbx_builtin_getvar_helper(NULL, "MIXMON_DIR");

	if (!file) {
		goto END;
	}

	if (ast_strlen_zero(mixmon_dir)) {
		goto END;
	}

	ret = readlink(mixmon_dir, real_path, sizeof(real_path) - 1);
	if (ret < 0 || ret >= sizeof(real_path) - 1) {
		goto END;
	}

	real_path[ret] = '\0';

	ast_log(LOG_DEBUG, "actual record storage path: %s", real_path);

	if (strstr(real_path, "/media/sd") || strstr(real_path, "/media/mmcblk")
		|| strstr(real_path,"/monitor_nas")) {
		char *p = strchr(file, '*');

		if (p) {
			while (p) {
				*p = '_';
				p = strchr(++p, '*');
			}

			ast_log(LOG_DEBUG, "Change %s --> %s", file_name, file);
		}
	}

	strncpy(buf, file, len-1);
	buf[len-1] = '\0';

END:
	if(file) {
		ast_free(file);
	}

	return 0;
}

static int recordcheck_function_read(struct ast_channel *chan, const char *cmd,
			char *data, char *buf, size_t len)
{
	int ret = 0;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(type);
		AST_APP_ARG(query);
		AST_APP_ARG(exten);
	);

	ast_log(LOG_DEBUG, "[recordcheck_function_read]chan:%s, cmd:%s, data:%s", 
		ast_channel_name(chan), cmd, data);

	AST_STANDARD_APP_ARGS(args, data);
	if (!args.type || !args.query) {
		ast_log(LOG_ERROR, "Invalid param: type:\"%s\", query:\"%s\"", args.type, args.query);
		goto END;
	}

	args.type = ast_str_to_lower(args.type);
	//args.query = ast_str_to_lower(args.query);
	if (args.exten) {
		args.exten = ast_str_to_lower(args.exten);
		ast_log(LOG_DEBUG, "exten:%s", args.exten);
	}

	if (!record_check_cfg) {
		ast_log(LOG_ERROR, "Config file:%s not loaded", RECORD_CHECK_CONF);
		return -1;
	}

	ret = check_global_enable();
	if (!ret)
		goto END;

	if (!strcasecmp(args.type, "call")) {
		ret = judge_call(chan, args.query);
	} else if (!strcasecmp(args.type, "extension")) {
		ret = judge_extension(chan, args.query);
	} else if (!strcasecmp(args.type, "trunk")) {
		ret = judge_trunk(args.query);
	} else if (!strcasecmp(args.type, "route")) {
		ret = judge_route(args.query);
	} else if (!strcasecmp(args.type, "ivr")) {
		ret = judge_ivr(args.query);
	} else if (!strcasecmp(args.type, "ringgroup")) {
		ret = judge_ringgroup(chan, args.exten, args.query);
	} else if (!strcasecmp(args.type, "file")) {
		check_recordfile_name(args.query, buf, len);
	} else {
		ret = 0;
		ast_log(LOG_WARNING, "Unknown type:%s", args.type);
	}

END:
	if (strcasecmp(args.type, "file")) {
		strncpy(buf, ret ? "yes" : "no", len);
	}

	return 0;
}

static struct ast_custom_function recordcheck_function = {
	.name = "RECORD_CHECK",
	.read = recordcheck_function_read,
};

static void check_record_disk_protect(void)
{
	char result[16] = {0};
	int ami_send_status = 0;

	ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		ami_send_status = atoi(result);
	}

	if (ami_send_status & 0x02) {
		ast_log(LOG_DEBUG, "Forbit voice recording globally");
		pbx_builtin_setvar_helper(NULL, "RECORD_FORBIDDEN", "yes");
	} else {
		ast_log(LOG_DEBUG, "Allow voice recording globally");
		pbx_builtin_setvar_helper(NULL, "RECORD_FORBIDDEN", "");
	}
}

#ifdef BUILD_ON_CLOUD_PBX
static void check_record_gdms_support(void)
{
	char result[16] = {0};
	int ami_send_status = 0;
	int gdms_record_support = 0;

	ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
	if (strlen(result)) {
		ami_send_status = atoi(result);
	}

	// only record function is not forbidden by ucm_warning (ami_send_status not equals to 0) can we
	// do allow or forbiden record function
	if (!ami_send_status) {
		memset(result, 0, sizeof(result));
		ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
		if (strlen(result) > 0) {
			gdms_record_support = atoi(result);
		}

		if (!gdms_record_support) {
			ast_log(LOG_DEBUG, "[check_record_gdms_support] Forbit voice recording globally");
			pbx_builtin_setvar_helper(NULL, "RECORD_FORBIDDEN", "yes");
		} else {
			ast_log(LOG_DEBUG, "[check_record_gdms_support] Allow voice recording globally");
			pbx_builtin_setvar_helper(NULL, "RECORD_FORBIDDEN", "");
		}
	}
}
#endif // BUILD_ON_CLOUD_PBX

static int pares_config_file(int reload)
{
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	SCOPED_MUTEX(lock, &config_lock);

	check_record_disk_protect();
#ifdef BUILD_ON_CLOUD_PBX
	check_record_gdms_support();
#endif // BUILD_ON_CLOUD_PBX

	cfg = ast_config_load(RECORD_CHECK_CONF, config_flags);
	if (cfg == CONFIG_STATUS_FILEMISSING || cfg == CONFIG_STATUS_FILEINVALID)
		return -1;

	if(cfg == CONFIG_STATUS_FILEUNCHANGED)
		return 0;

	if(record_check_cfg && record_check_cfg != cfg)
		ast_config_destroy(record_check_cfg);

	record_check_cfg = cfg;
	ast_log(LOG_DEBUG, "Config file:%s %s", RECORD_CHECK_CONF, reload ? "reloaded" : "loaded");

	return 0;
}


static int load_module(void)
{
	int res = 0;

	res |= pares_config_file(0);
	res |= ast_custom_function_register(&recordcheck_function);

	return res;
}

static int reload_module(void)
{
	int res = 0;

	res |= pares_config_file(1);

	return res;
}

static int unload_module(void)
{
	int res = 0;

	res |= ast_custom_function_unregister(&recordcheck_function);

	if (record_check_cfg)
		ast_config_destroy(record_check_cfg);

	return res;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Record check related dialplan functions",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload_module,
);

#endif

