#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2016-2017, Kaiming Ye <kmye@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Functions for interaction with the Asterisk database
 *
 * \author Kaiming Ye <kmye@grandstream.cn>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/


#include "asterisk.h"

//ASTERISK_FILE_VERSION(__FILE__, "$Revision$")

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/lock.h"
#include "ucm_crypt.h"
#include "asterisk/ucm_dbi.h"

#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#endif

AST_MODULE_LOG("func_seamless_transfer");

/*** DOCUMENTATION
	<function name="IN_TRANSFER_DIDWHILTELIST" language="en_US">
		<synopsis>
			Check if the Extensionis on the seamless transfer did white list .
		</synopsis>
		<syntax argsep="/">
			<parameter name="did_x" required="did-1" />
			<parameter name="extension" required="1000" />
		</syntax>
		<description>
			<para>Check if the Extensionis on the seamless transfer did white list .
			</para>
		</description>
		<see-also>
		</see-also>
	</function>
 ***/

#define AST_SAFE_FREE(ptr) if (ptr) {\
    ast_free(ptr);\
    ptr = NULL; }

AST_MUTEX_DEFINE_STATIC(cb_mutex);

struct seamless_transfer
{
	char extension[AST_MAX_EXTENSION];
	char phonenumber[AST_MAX_EXTENSION];
	char phonenumber_prefix[AST_MAX_EXTENSION];
	struct seamless_transfer *next;
};


struct transfer
{
	char extension[AST_MAX_EXTENSION];
	struct seamless_transfer *data;
	struct transfer *next;
};
struct transfer *g_cb = NULL;

#ifdef HAVE_UNIXODBC
struct generic_prepare_struct {
	char *sql;
	int argc;
	char **argv;
};

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	if (!obj) {
		ast_log(AST_LOG_WARNING, "The obj is error !\n");
		return NULL;
	}

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}
#endif


static struct transfer *find_transfer_by_name(struct transfer *cb, const char * extension)
{
	if (cb == NULL || extension == NULL) {
		return NULL;
	}
	struct transfer *tmp_cb = cb;
	ast_log(LOG_DEBUG, "cb : %s, extension %s", cb->extension, extension);
	while (tmp_cb != NULL) {
		if (!strncmp(tmp_cb->extension, extension, sizeof(tmp_cb->extension))) {
			break;
		}
		tmp_cb = tmp_cb->next;
	}
	return tmp_cb;
}

static char *find_seamless_transfer_by_name(struct seamless_transfer *cm, const char *extension)
{
	if (cm == NULL || extension == NULL) {
		return NULL;
	}
	struct seamless_transfer *tmp_cm = cm;
	while (tmp_cm != NULL) {
		ast_log(LOG_DEBUG, "tmp_cm->phonenumber : %s  prefix %s  --- extension : %s", tmp_cm->phonenumber, tmp_cm->phonenumber_prefix, extension);
		if (ast_extension_match(tmp_cm->phonenumber, extension) || ast_extension_match(tmp_cm->phonenumber_prefix, extension))
		{
			ast_log(LOG_DEBUG, "find : [%s]", tmp_cm->extension);
			return tmp_cm->extension;
		}
		tmp_cm = tmp_cm->next;
	}
	return NULL;
}

static char* str_strim(const char* src)
{
	if (src == NULL) {
		return NULL;
	}
	int length = strlen(src) + 1;
	char *dst = (char*)ast_malloc(length);
	char *tmp = dst;
	memset(dst, 0, length);
	while (*src != '\0') {
		if (*src != ' ') {
			*tmp++= *src;
		}
		src++;
	}
	return dst;
}

static char* get_ucm_name_decrypt(const char* plain_string)
{
	if (plain_string == NULL)
	{
		return NULL;
	}
	struct odbc_obj *obj = NULL;
	char cmd[1024] = {0};
	char fullname[128] = {0};
	SQLHSTMT stmt = NULL;
	struct generic_prepare_struct gps = { .sql = cmd };
	obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);

	char *master_key = NULL;
	master_key = get_encrypted_key();
	if (master_key) {
		snprintf(cmd ,sizeof(cmd) ,"SELECT CAST(AES_DECRYPT(UNHEX(substring('%s',7)), '%s') AS CHAR CHARSET utf8) AS fullname", plain_string, master_key);
	}
	else
	{
		snprintf(fullname, sizeof(fullname), "%s", plain_string);
	}

	AST_SAFE_FREE(master_key);

	stmt = ast_core_odbc_prepare_and_execute(obj, generic_prepare, &gps);
	if ( stmt && SQL_SUCCEEDED(SQLFetch(stmt)) )
	{
		SQLGetData(stmt, 1, SQL_CHAR, fullname, sizeof(fullname), NULL);
	}
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	if (obj) {
		ast_core_odbc_release_obj(obj);
	}

	return ast_strdup(fullname);
}

static struct seamless_transfer *malloc_seamless_transfer(const char * extension)
{
	if (extension == NULL) {
		return NULL;
	}
	char* p_left = NULL;
	char* p_right = NULL;
	char* p_comma = NULL;
	char *plain_string = NULL;
	char *space_addr = NULL;
	char tmp_phonenumber[AST_MAX_EXTENSION] = {0};
	char *tmp_str = NULL;
	struct seamless_transfer *cm = NULL;
	cm = ast_malloc(sizeof(struct seamless_transfer));
	if (cm != NULL) {
		p_left = strstr(extension, "(");
		p_right = strstr(extension, ")");
		p_comma = strstr(extension, ",");
		if (p_left != NULL && p_right != NULL && p_comma != NULL)
		{
			*p_comma = '\0';
			*p_right = '\0';
			memset(cm, 0, sizeof(struct seamless_transfer));
			snprintf(cm->extension, strlen(p_left+1) + 1, "%s", p_left+1);

			if (*(p_comma+1) != '\0')
			{
				snprintf(tmp_phonenumber,  strlen(p_comma+1) + 1, "%s", (p_comma+1));
				plain_string = get_ucm_name_decrypt(tmp_phonenumber);

				if (plain_string != NULL) {
					space_addr = strchr(plain_string, ' ');
					if (space_addr != NULL)
					{
						snprintf(cm->phonenumber, sizeof(cm->phonenumber), "%s", space_addr + 1);
						tmp_str = str_strim(plain_string);
						snprintf(cm->phonenumber_prefix, sizeof(cm->phonenumber_prefix), "%s", tmp_str);
						AST_SAFE_FREE(tmp_str);
					}
					else
					{
						//we think it has no prefix
						snprintf(cm->phonenumber, sizeof(cm->phonenumber), "%s", plain_string);
						snprintf(cm->phonenumber_prefix, sizeof(cm->phonenumber_prefix), "~None~");
					}
				}
				ast_log(LOG_DEBUG, "plain_string : %s", plain_string);
				AST_SAFE_FREE(plain_string);
			}
			else
			{
				snprintf(cm->phonenumber, sizeof(cm->phonenumber), "~None~");
				snprintf(cm->phonenumber_prefix, sizeof(cm->phonenumber_prefix), "~None~");
			}
		}
		cm->next = NULL;
	}
	return cm;
}


static struct seamless_transfer *create_seamless_transfer(struct transfer *cb, const char *extension)
{
	if (cb == NULL || extension == NULL) {
		return NULL;
	}

	struct seamless_transfer *cm = malloc_seamless_transfer(extension);
	struct seamless_transfer *tmp_cm = cb->data;
	if (cb->data == NULL) {
		cb->data = cm;
	}
	else {
		while (tmp_cm->next) {
			tmp_cm = tmp_cm->next;
		}
		tmp_cm->next = cm;
	}

	return cm;
}

static struct transfer* malloc_transfer(const char *extension)
{
	if (extension == NULL) {
		return NULL;
	}
	struct transfer *cb = NULL;
	cb = ast_malloc(sizeof(struct transfer));
	if (cb) {
		memset(cb->extension, 0, sizeof(cb->extension));
		strncpy(cb->extension, extension, sizeof(cb->extension));
		cb->data = NULL;
		cb->next = NULL;
	}

	return cb;
}

static struct transfer *create_transfer(struct transfer **cb, const char *extension)
{
	if (cb == NULL || extension == NULL) {
		return NULL;
	}

	struct transfer *tmp_cb = *cb, *tmp = NULL;
	tmp = malloc_transfer(extension);
	if (tmp_cb == NULL) {
		*cb = tmp;
	}
	else {
		while( tmp_cb->next ) {
			tmp_cb = tmp_cb->next;
		}
		tmp_cb->next = tmp;
	}

	return tmp;
}

static void print_transfer(void)
{
	struct transfer *tmp_cb = g_cb;
	struct seamless_transfer *tmp_cm = NULL;
	ast_log(LOG_DEBUG, "--------------------seamless transfer list start---------");
	while (tmp_cb) {
		ast_debug(1, "context [%s]", tmp_cb->extension);
		tmp_cm = tmp_cb->data;
		while (tmp_cm) {
			ast_log(LOG_DEBUG,"		extension = %s, phonenumber = %s", tmp_cm->extension, tmp_cm->phonenumber);
			ast_log(LOG_DEBUG,"		extension = %s, phonenumber_prefix = %s", tmp_cm->extension, tmp_cm->phonenumber_prefix);
			tmp_cm = tmp_cm->next;
		}
		tmp_cb = tmp_cb->next;
	}
	ast_log(LOG_DEBUG, "---------seamless transfer list end ---------");
}

static void unload_transfer(void)
{
	struct transfer *tmp_cb = g_cb, *prev_cb = NULL;
	struct seamless_transfer *tmp_cm = NULL, *prev_cm = NULL;
	while (tmp_cb) {
		tmp_cm = tmp_cb->data;
		while (tmp_cm) {
			prev_cm = tmp_cm;
			tmp_cm = tmp_cm->next;
			ast_free(prev_cm);
		}
		prev_cb = tmp_cb;
		tmp_cb = tmp_cb->next;
		ast_free(prev_cb);
	}
	g_cb = NULL;
}

static char *check_transfer_privilege(const char *from, const char *to)
{
	if (from == NULL || to == NULL || from[0] == '\0' || to[0] == '\0'){
		return NULL;
	}
	struct transfer *tmp_cb = NULL;
	ast_log(LOG_DEBUG, "cb : %s, extension %s", g_cb->extension, to);
	tmp_cb = find_transfer_by_name(g_cb, to);
	if (!tmp_cb) {
		return NULL;
	}
	return find_seamless_transfer_by_name(tmp_cb->data, from);
}

static int transfer_function_read(struct ast_channel *chan, const char *cmd,
			char *data, char *buf, size_t len)
{
	char *parse = NULL;
	char *extension = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(did_x);
		AST_APP_ARG(did_phonenumber);
	);
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}
	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.did_x) || ast_strlen_zero(args.did_phonenumber)) {
		return -1;
	}
	ast_log(LOG_DEBUG, "-----from : %s-----", args.did_x);
	ast_mutex_lock(&cb_mutex);
	extension = check_transfer_privilege(args.did_phonenumber, args.did_x);
	if (extension != NULL) {
		strncpy(buf, extension, len-1);
	}
	else {
		strncpy(buf, "None", len-1);
	}
	ast_mutex_unlock(&cb_mutex);
	return 0;
}


static struct ast_custom_function transfer_function = {
	.name = "IN_TRANSFER_DIDWHILTELIST",
	.read = transfer_function_read,
};

static int unload_module(void)
{
	int res = 0;

	res |= ast_custom_function_unregister(&transfer_function);

	unload_transfer();

	return res;
}

static int load_module(void)
{
	int res = 0;
	char *tmp = NULL;
	struct ast_config *cfg = NULL;
	struct ast_variable *var = NULL;
	struct transfer *tmp_cb = NULL;
	struct ast_flags config_flags = { 0 };
	if (!(cfg = ast_config_load("seamless_transfer_did_whitelist.conf", config_flags))) {
                ast_log(LOG_NOTICE, "No  seamless transfer config file (seamless transfer.conf), so no seamless transfer\n");
                ast_custom_function_register(&transfer_function);
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file seamless_transfer.conf is in an invalid format.  Aborting.\n");
		return 1;
	}

	while ((tmp = ast_category_browse(cfg, tmp)) ) {
		tmp_cb = create_transfer(&g_cb, tmp);
		for (var = ast_variable_browse(cfg, tmp); var; var = var->next) {
			if (!strcasecmp(var->name, "didwhite")) {
				create_seamless_transfer(tmp_cb, var->value);
			}
		}
	}
	print_transfer();
	res |= ast_custom_function_register(&transfer_function);

	ast_config_destroy(cfg);

	return res;
}

static int reload(void)
{
	int res = 0;

	ast_mutex_lock(&cb_mutex);
	res = unload_module();
	res |= load_module();
	ast_mutex_unlock(&cb_mutex);

	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Seamless Transfer DID WhiteList related dialplan functions",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
);

#endif

