
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2022, Grandstream, Inc.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Trunk call time statistics and check
 *
 * \author <yfyang@grandstream.com>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/module.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/app.h"
#include "asterisk/haserver.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/json.h"
#include "asterisk/cli.h"
#include "asterisk/astdb.h"

AST_MODULE_LOG("dial");

/*** DOCUMENTATION
	<function name="TRUNK_TIME_LIMITED" language="en_US">
		<synopsis>
			Check whether the trunk has reached the call time limit or not.
		</synopsis>
		<syntax>
			<parameter name="channel name" />
		</syntax>
		<description>
			<para>This function returns <literal>1</literal> if the trunk has reached the limit. Otherwise, it returns <literal>0</literal>.</para>
		</description>
	</function>
	<function name="TRUNK_TIME_UPDATE" language="en_US">
		<synopsis>
			Update the trunk call time record.
		</synopsis>
		<syntax />
		<description>
			<para>Update the trunk call time record.</para>
		</description>
	</function>
 ***/

AST_MUTEX_DEFINE_STATIC(tkl_mutex);
#define CALL_TIME_BUFFER_LEN   2048
#define TKLIMIT_UPDATE_INTVAL  30

#define  TRUNK_CALLLIMIT_CONFIG_FILE  "trunk_calllimit.conf"
#define  TRUNK_CALLDATA_CURRENT       "trunk_calldata_current.json"

#define  GEN_TKDATA_LASTMONTH         "/cfg/var/lib/asterisk/scripts/gen_tkdata_lastmonth.py"
#define  GEN_TKDATA_CURRENT           "/cfg/var/lib/asterisk/scripts/gen_tkdata_current.py"

#define  TRUNK_CALLDATA_FILE_CUR   	  "/app/asterisk/var/run/trunk_calldata_current.json"
#define  TRUNK_CALLDATA_PATH          "/app/asterisk/var/run"
#define  TRUNK_TIMESUM_FUNC           "TRUNK_TIME_SUM"

static unsigned int g_tklimit_enable_num = 0;

enum tk_call_cycle_type {
	TKL_CYCLE_MONTH,
	TKL_CYCLE_QUARTER,
	TKL_CYCLE_BUTT,
};

enum tk_technology_type {
	TK_TECH_PJSIP,       /* PJSIP */
	TK_TECH_IAX2,        /* IAX2 */
	TK_TECH_DAHDI,       /* DAHDI */
	TK_TECH_TRUNK_GROUP, /* TRUNK_GROUP */
	TK_TECH_BUTT,
};

struct tk_limit_chan {
	AST_DECLARE_STRING_FIELDS (
		AST_STRING_FIELD(name);            /* the tkl chan name */
	);

	struct tk_limit_config  *tkl_config;   /* the related tkl_config */
	unsigned int enable_time_limit:1;
	unsigned int cur_consume;
	unsigned int history_consume;
	AST_RWLIST_ENTRY(tk_limit_chan) entry;
};

struct tk_limit_config {
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);
		AST_STRING_FIELD(trunk_name);
		AST_STRING_FIELD(include_chans);
		AST_STRING_FIELD(start_month);
	);

	AST_LIST_HEAD(,tk_limit_chan) tkl_chans;
	enum tk_technology_type  en_tech;
	enum tk_call_cycle_type  en_cycle;
	unsigned int month;
	unsigned int trunk_index;
	unsigned int total_time;
	unsigned int cur_consume;
	unsigned int history_consume;
	unsigned int old_consume;
	unsigned int enable_time_limit:1;
	unsigned int mark:1;
	unsigned int update:1;
};

static struct ao2_container *con_tklimit_configs;
static struct ao2_container *con_tklimit_chans;
static ast_rwlock_t  tklimit_update_lock;
static int  tklimit_update_sched_id = -1;
static unsigned int g_need_update_tklimit = 0;  /* The flag to indicate whether need to update or not */
static struct ast_sched_context *tklimit_update_sched = NULL;
char  g_cur_month_start[32] = {0};
char  g_cur_month[8] = {0};
char  g_cur_month2[8] = {0};
static struct stasis_message_router *tklimit_ha_data_router = NULL;
ast_mutex_t limit_lock;

static int reload(void);

static const char *tklimit_tech2str(enum tk_technology_type tech)
{
	switch (tech) {
	case TK_TECH_PJSIP:
		return "PJSIP";
	case TK_TECH_IAX2:
		return "IAX2";
	case TK_TECH_DAHDI:
		return "DAHDI";
	case TK_TECH_TRUNK_GROUP:
		return "TRUNK_GROUP";
	default:
		return "unknown";
	}
}

static const char *tklimit_cycle2str(enum tk_call_cycle_type en_cycle)
{
	switch (en_cycle) {
	case TKL_CYCLE_MONTH:
		return "month";
	case TKL_CYCLE_QUARTER:
		return "quarter";
	default:
		return "unknown";
	}
}

static int tklimit_get_current_month(char *buffer, int len)
{
	struct ast_tm timeresult;
	struct timeval now = ast_tvnow();

	if (!buffer || !len) {
		return -1;
	}

	ast_localtime(&now, &timeresult, NULL);
	ast_strftime(buffer, len, "%Y-%m", &timeresult);
	ast_strftime(g_cur_month2, sizeof(g_cur_month2), "%Y_%m", &timeresult);
	snprintf(g_cur_month_start, sizeof(g_cur_month_start)-1, "%s-01 00:00:00", buffer);

	return 0;
}

static int tklimit_check_current_month(char *month)
{
	if (!month) {
		return -1;
	}

	if (strncmp(g_cur_month, month, sizeof(g_cur_month))) {
		ast_log(LOG_NOTICE, "g_cur_month[%s] changed to [%s]", g_cur_month, month);
		strncpy(g_cur_month, month, sizeof(g_cur_month));
		return 1; /* changed */
	}

	return 0;  /* not changed */
}

static int tklimit_get_history_num(int start_month)
{
	struct ast_tm timeresult;
	char time_buffer[32] = {0};
	struct timeval now = ast_tvnow();
	int year   = 0;
	int month  = 0;
	int num    = 0;

	ast_localtime(&now, &timeresult, NULL);
	ast_strftime(time_buffer, sizeof(time_buffer), "%Y-%m", &timeresult);
	sscanf(time_buffer, "%4d-%2d", &year, &month);

	if (month >= start_month) {
		num = (month - start_month)%3;
	} else {
		num = (month - start_month + 12)%3;
	}

	ast_log(LOG_DEBUG, "tklimit_get_history_num, num[%d]", num);
	return num;
}

static int tklimit_get_history_month(char *history1, int len1, char *history2, int len2)
{
	struct ast_tm timeresult;
	char time_buffer[32] = {0};
	struct timeval now = ast_tvnow();
	int year   = 0;
	int month  = 0;
	int year1  = 0;
	int month1 = 0;
	int year2  = 0;
	int month2 = 0;

	if (!history1 || !history2) {
		return -1;
	}

	ast_localtime(&now, &timeresult, NULL);
	ast_strftime(time_buffer, sizeof(time_buffer), "%Y-%m", &timeresult);
	sscanf(time_buffer, "%4d-%2d", &year, &month);

	month1 = month-1;
	year1  = year;
	if (month1 <= 0) {
		month1 += 12;
		year1 = year - 1;
	}
	snprintf(history1, len1, "%4d_%02d", year1, month1);

	month2 = month1 - 1;
	year2  = year1;
	if (month2 <= 0) {
		month2 += 12;
		year2 = year1 - 1;
	}
	snprintf(history2, len2, "%4d_%02d", year2, month2);

	ast_log(LOG_NOTICE, "tklimit_get_history_month, history1[%s], history2[%s]", history1, history2);
	return 0;
}

static int tklimit_get_chan_name(struct tk_limit_config *tkl_config, int chan_id, char *buffer, int buf_len)
{
	if (!tkl_config || !buffer || !buf_len) {
		return -1;
	}

	switch (tkl_config->en_tech) {
	case TK_TECH_PJSIP:
		snprintf(buffer, buf_len-1, "PJSIP/%s", tkl_config->name);
		break;
	case TK_TECH_IAX2:
		snprintf(buffer, buf_len-1, "IAX2/%s", tkl_config->name);
		break;
	case TK_TECH_DAHDI:
		snprintf(buffer, buf_len-1, "DAHDI/o%d", chan_id);
		break;
	case TK_TECH_TRUNK_GROUP:
		snprintf(buffer, buf_len-1, "PJSIP/trunk_%d", chan_id);
		break;
	default:
		return -1;
	}

	ast_log(LOG_NOTICE, "tklimit_get_chan_name, trunk[%s], chan_id[%d], tech[%d] ==> [%s] ", tkl_config->name, chan_id, tkl_config->en_tech, buffer);
	return 0;
}

static int tklimit_chan_hash(const void *obj, const int flags)
{
	const struct tk_limit_chan *tkl_chan = obj;

	return ast_str_case_hash(tkl_chan->name);
}

static int tklimit_chan_cmp(void *obj, void *arg, int flags)
{
	struct tk_limit_chan *tkl_chan = obj, *tkl_chan2 = arg;

	return !strcasecmp(tkl_chan->name, tkl_chan2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static int tklimit_config_hash(const void *obj, const int flags)
{
	const struct tk_limit_config *tkl_config = obj;

	return ast_str_case_hash(tkl_config->name);
}

static int tklimit_config_cmp(void *obj, void *arg, int flags)
{
	struct tk_limit_config *tkl_config = obj, *tkl_config2 = arg;

	return !strcasecmp(tkl_config->name, tkl_config2->name) ? CMP_MATCH | CMP_STOP : 0;
}

static struct tk_limit_config *tklimit_find_conf(const char *name)
{
	struct tk_limit_config tmp_config = {
		.name = name,
	};

	if (!name) {
		return NULL;
	}

	return ao2_find(con_tklimit_configs, &tmp_config, OBJ_POINTER);
}


static struct tk_limit_chan *tklimit_find_chan(const char *name)
{
	struct tk_limit_chan tmp_chan = {
		.name = name,
	};

	return ao2_find(con_tklimit_chans, &tmp_chan, OBJ_POINTER);
}


/*! release the related list in the sca shared object */
static int tklimit_config_release_list(void *obj, void *arg, int flags)
{
	struct tk_limit_config *tkl_config = obj;
	struct tk_limit_chan   *tkl_chan;

	while ((tkl_chan = AST_LIST_REMOVE_HEAD(&tkl_config->tkl_chans, entry))) {
		ao2_ref(tkl_chan, -1);
	}

	return 0;
}

static void tklimit_config_destructor(void *obj)
{
	struct tk_limit_config *tkl_config = obj;

	ast_log(LOG_NOTICE, "tklimit_config_destructor for [%s][%p]\n", tkl_config->name, tkl_config);

	tklimit_config_release_list(tkl_config, NULL, 0);

	AST_LIST_HEAD_DESTROY(&tkl_config->tkl_chans);
	ast_string_field_free_memory(tkl_config);
}


static void tklimit_chan_destructor(void *obj)
{
	struct tk_limit_chan *tkl_chan = obj;

	ast_log(LOG_NOTICE, "tklimit_chan_destructor for [%s][%p]\n", tkl_chan->name, tkl_chan);

	tkl_chan->tkl_config = NULL;
	ast_string_field_free_memory(tkl_chan);
}

static int tklimit_config_is_marked(void *obj, void *arg, int flags)
{
	struct tk_limit_config *tkl_config = obj;

	ao2_lock(tkl_config);

	if (tkl_config->mark) {
		tklimit_config_release_list(tkl_config, NULL, 0);
	}

	ao2_unlock(tkl_config);
	return tkl_config->mark ? CMP_MATCH:0;
}

static int tklimit_config_mark(void *obj, void *arg, int flags)
{
	struct tk_limit_config *tkl_config = obj;

	ao2_lock(tkl_config);
	tkl_config->mark = 1;
	ao2_unlock(tkl_config);

	return 0;
}

static int tklimit_config_reset_current(void *obj, void *arg, int flags)
{
	struct tk_limit_config *tkl_config = obj;

	ao2_lock(tkl_config);
	tkl_config->old_consume = tkl_config->cur_consume + tkl_config->history_consume;
	tkl_config->update = 1;
	tkl_config->cur_consume = 0;
	ao2_unlock(tkl_config);

	return 0;
}

static int tklimit_config_reset_consume(void *obj, void *arg, int flags)
{
	struct tk_limit_config *tkl_config = obj;

	ao2_lock(tkl_config);
	tkl_config->old_consume = tkl_config->cur_consume + tkl_config->history_consume;
	tkl_config->update = 1;
	tkl_config->cur_consume = 0;
	tkl_config->history_consume = 0;
	ao2_unlock(tkl_config);

	return 0;
}

static int tklimit_update_chan(struct tk_limit_config *tkl_config, struct tk_limit_chan *tkl_chan)
{
	if (!tkl_config || !tkl_chan) {
		return -1;
	}

	if (tkl_chan->tkl_config) {
		if (tkl_chan->tkl_config != tkl_config) {
			tkl_chan->tkl_config = tkl_config;
		}
	} else {
		tkl_chan->tkl_config = tkl_config;
	}

	ao2_bump(tkl_chan);
	AST_LIST_INSERT_TAIL(&tkl_config->tkl_chans, tkl_chan, entry);
	tkl_chan->enable_time_limit = tkl_config->enable_time_limit;

	ast_log(LOG_NOTICE, "tklimit_update_chan, [%s]", tkl_chan->name);
	return 0;
}


static int tklimit_new_chan(struct tk_limit_config *tkl_config, char *name)
{
	RAII_VAR(struct tk_limit_chan *, tkl_chan, NULL, ao2_cleanup);

	if (!tkl_config || !name) {
		return -1;
	}

	tkl_chan = ao2_alloc(sizeof(*tkl_chan), tklimit_chan_destructor);
	if (!tkl_chan) {
		return -1;
	}

	if (ast_string_field_init(tkl_chan, 32)) {
		return -1;
	}

	ast_string_field_set(tkl_chan, name, name);
	tkl_chan->tkl_config = tkl_config;

	ao2_bump(tkl_chan);
	AST_LIST_INSERT_TAIL(&tkl_config->tkl_chans, tkl_chan, entry);
	tkl_chan->enable_time_limit = tkl_config->enable_time_limit;

	ao2_link(con_tklimit_chans, tkl_chan);
	ast_log(LOG_NOTICE, "tklimit_new_chan, [%s][%p]", name, tkl_chan);
	return 0;
}

static int tklimit_related_chan(struct tk_limit_config *tkl_config, int chan_id)
{
	RAII_VAR(struct tk_limit_chan *, tkl_chan, NULL, ao2_cleanup);
	char  chan_name[20] = {0};

	if (!tkl_config) {
		return -1;
	}

	tklimit_get_chan_name(tkl_config, chan_id, chan_name, sizeof(chan_name)-1);
	tkl_chan = tklimit_find_chan(chan_name);

	if (tkl_chan) {
		tklimit_update_chan(tkl_config, tkl_chan);
		ao2_ref(tkl_chan, -1);
	} else {
		tklimit_new_chan(tkl_config, chan_name);
	}

	ast_log(LOG_NOTICE, "tklimit_related_chan, chan[%s]->config[%s][%p]", chan_name, tkl_config->name, tkl_config);
	return 0;
}

static int tklimit_build_chan(struct tk_limit_config *tkl_config)
{
	RAII_VAR(struct tk_limit_chan *, tkl_chan, NULL, ao2_cleanup);
	char *include_chans = NULL;
	char *cur;
	int   chanid;

	if (!tkl_config) {
		return -1;
	}

	if (!ast_strlen_zero(tkl_config->include_chans)) {
		include_chans = ast_strdupa(tkl_config->include_chans);
		if (!include_chans) {
			return -1;
		}

		while ((cur = strsep(&include_chans, ","))) {
			chanid = atoi(cur);
			tklimit_related_chan(tkl_config, chanid);
		}
	} else {
		tklimit_related_chan(tkl_config, 0);
	}

	return 0;
}

static int tklimit_build_config(struct ast_config *cfg, const char *cat)
{
	RAII_VAR(struct tk_limit_config *, tkl_config, NULL, ao2_cleanup);
	struct ast_variable   *var;
	int    config_existed = 0;
	char   *trunk_id     = NULL;

	if ((tkl_config = tklimit_find_conf(cat))) {
		config_existed = 1;
		tkl_config->mark = 0;
	} else if ((tkl_config = ao2_alloc(sizeof(*tkl_config), tklimit_config_destructor))) {
		if (ast_string_field_init(tkl_config, 128)) {
			return -1;
		}
		ast_string_field_set(tkl_config, name, cat);

		trunk_id = strstr(tkl_config->name, "trunk_");
		if (!trunk_id) {
			return -1;
		}

		trunk_id = trunk_id + strlen("trunk_");
		tkl_config->trunk_index = atoi(trunk_id);

		ast_log(LOG_NOTICE, "build tkl_config, alloc [%s][%p]!\n", tkl_config->name, tkl_config);
	} else {
		return -1;
	}

	ao2_lock(tkl_config);

	tkl_config->total_time = 0;
	tkl_config->cur_consume = 0;
	tkl_config->history_consume = 0;
	tkl_config->old_consume = 0;
	tklimit_config_release_list(tkl_config, NULL, 0);

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "trunk_name")) {
			ast_string_field_set(tkl_config, trunk_name, var->value);
		} else if (!strcasecmp(var->name, "technology")) {
			if (!strcasecmp(var->value, "PJSIP")) {
				tkl_config->en_tech = TK_TECH_PJSIP;
			} else if (!strcasecmp(var->value, "IAX2")) {
				tkl_config->en_tech = TK_TECH_IAX2;
			} else if (!strcasecmp(var->value, "DAHDI")) {
				tkl_config->en_tech = TK_TECH_DAHDI;
			} else if (!strcasecmp(var->value, "TRUNK_GROUP")) {
				tkl_config->en_tech = TK_TECH_TRUNK_GROUP;
			} else {
				tkl_config->en_tech = TK_TECH_BUTT;
			}
		} else if (!strcasecmp(var->name, "include_chans")){
			ast_string_field_set(tkl_config, include_chans, var->value);
		} else if (!strcasecmp(var->name, "start_month")) {
			int year = 0;

			ast_string_field_set(tkl_config, start_month, var->value);
			sscanf(tkl_config->start_month, "%4d-%2d", &year, &tkl_config->month);
		} else if (!strcasecmp(var->name, "enable_time_limit")) {
			tkl_config->enable_time_limit = ast_true(var->value);

			if (tkl_config->enable_time_limit) {
				g_tklimit_enable_num++;
			}
		} else if (!strcasecmp(var->name, "total_time")) {
			tkl_config->total_time = atoi(var->value);
		} else if (!strcasecmp(var->name, "cum_period")) {
			if (!strncmp(var->value, "quarter", 7)) {
				tkl_config->en_cycle = TKL_CYCLE_QUARTER;
			} else {
				tkl_config->en_cycle = TKL_CYCLE_MONTH;
			}
		}
	}

	if (!config_existed) {
		AST_LIST_HEAD_INIT(&tkl_config->tkl_chans);

		ao2_link(con_tklimit_configs, tkl_config);
		ast_log(LOG_NOTICE, "build tkl_config[%s][%p]!\n", tkl_config->name, tkl_config);
	}

	if (tkl_config->enable_time_limit) {
		tklimit_build_chan(tkl_config);
	}

	ao2_unlock(tkl_config);
	return 0;
}

static int tklimit_check_call_time(struct ast_channel *chan, const char *cmd, char *data, char *buffer, size_t buflen)
{
	int   consumed = 0;
	struct tk_limit_config *tkl_config;
	char *tmp = NULL;
	char  tk_name[16] = {0};
	struct tk_limit_chan *tkl_chan = NULL;

	if (!chan || !data) {
		ast_log(LOG_WARNING, "No channel was provided to %s function.\n", cmd);
		return -1;
	}

	ast_log(LOG_NOTICE, "tklimit_check_call_time, chan[%s], data[%s].\n", ast_channel_name(chan), data);

	/* data[DAHDI/g6] */
	tmp = strstr(data, "DAHDI/g");
	if (tmp) {
		tmp = tmp+strlen("DAHDI/g");
		snprintf(tk_name, sizeof(tk_name), "trunk_%s", tmp);
		tkl_config = tklimit_find_conf(tk_name);
		if (!tkl_config || !tkl_config->enable_time_limit || !tkl_config->total_time) {
			ast_copy_string(buffer, "0", buflen);   /* not limited */
			ao2_cleanup(tkl_config);

			ast_log(LOG_NOTICE, "trunk[%s] not limited.\n", tk_name);
			return 0;
		}

		ast_mutex_lock(&limit_lock);
		consumed = tkl_config->cur_consume + tkl_config->history_consume;

		if (consumed >= tkl_config->total_time) {
			ast_log(LOG_NOTICE, "consume[%d] reached configed[%d], limited!\n", consumed, tkl_config->total_time);
			ast_copy_string(buffer, "1", buflen);   /* limited */
		} else {
			ast_log(LOG_NOTICE, "trunk[%s] not reach the limited time.\n", tk_name);
			ast_copy_string(buffer, "0", buflen);   /* not reach the limit */
		}

		ao2_cleanup(tkl_config);
		ast_mutex_unlock(&limit_lock);
		return 0;
	}

	tkl_chan = tklimit_find_chan(data);
	if (!tkl_chan || !tkl_chan->enable_time_limit || !tkl_chan->tkl_config || !tkl_chan->tkl_config->total_time) {
		ast_copy_string(buffer, "0", buflen);   /* not limited */
		ao2_cleanup(tkl_chan);
		return 0;
	}

	tkl_config = tkl_chan->tkl_config;

	ast_mutex_lock(&limit_lock);
	consumed = tkl_config->cur_consume + tkl_config->history_consume;

	if (consumed >= tkl_config->total_time) {
		ast_log(LOG_NOTICE, "consume[%d] reached total_time[%d], limited!\n", consumed, tkl_config->total_time);
	    ast_copy_string(buffer, "1", buflen);   /* limited */
	} else {
		ast_copy_string(buffer, "0", buflen);   /* not reach the limit */
	}

	ao2_cleanup(tkl_chan);
	ast_mutex_unlock(&limit_lock);
	return 0;
}

static int tklimit_set_update(struct ast_channel *chan, const char *cmd, char *data, char *buffer, size_t buflen)
{
	if (!g_need_update_tklimit) {
		ast_rwlock_wrlock(&tklimit_update_lock);
		g_need_update_tklimit = 1;
		ast_rwlock_unlock(&tklimit_update_lock);
	}

	return 0;
}

static struct ast_custom_function tklimit_func = {
	.name = "TRUNK_TIME_LIMITED",
	.read = tklimit_check_call_time,
};

static struct ast_custom_function tklimit_update_func = {
	.name = "TRUNK_TIME_UPDATE",
	.read = tklimit_set_update,
};

static char *tklimit_show_configs(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct tk_limit_config *tkl_config;
	struct tk_limit_chan   *tkl_chan;

	switch (cmd) {
	case CLI_INIT:
		e->command = "tklimit show configs";
		e->usage =
			"Usage: tklimit show configs\n"
			"       This will list all trunks defined in trunk_calllimit.conf \n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n"
	            "==========================================================\n"
	            "----- Configured Trunks Limited --------------------------\n"
	            "==========================================================\n"
	            "===\n");

	ast_mutex_lock(&limit_lock);
	i = ao2_iterator_init(con_tklimit_configs, 0);
	for (; (tkl_config = ao2_iterator_next(&i)); ao2_ref(tkl_config, -1)) {
		if (!tkl_config->enable_time_limit) {
			continue;
		}

		ao2_lock(tkl_config);

		ast_cli(a->fd, "=== -----------------------------------------------------\n"
		            "=== Name:                 %s\n"
		            "=== ==> trunk_name:       %s\n"
		            "=== ==> technology:       %s\n"
		            "=== ==> cum_period:       %s\n"
		            "=== ==> start_month:      %s\n"
		            "=== ==> total_time:       %d(min)\n"
				    "=== ==> cur_consume:      %d(min)\n"
				    "=== ==> history_consume:  %d(min)\n",
		            tkl_config->name,
		            tkl_config->trunk_name,
		            tklimit_tech2str(tkl_config->en_tech),
		            tklimit_cycle2str(tkl_config->en_cycle),
		            ast_strlen_zero(tkl_config->start_month) ? " ":tkl_config->start_month,
		            tkl_config->total_time,
		            tkl_config->cur_consume,
		            tkl_config->history_consume
			);

		AST_LIST_TRAVERSE(&tkl_config->tkl_chans, tkl_chan, entry) {
			ast_cli(a->fd, "===    ==> chan name: %s\n", tkl_chan->name);
		}

		ast_cli(a->fd, "=== -----------------------------------------------------\n===\n");

		ao2_unlock(tkl_config);
	}

	ao2_iterator_destroy(&i);
	ast_mutex_unlock(&limit_lock);
	ast_cli(a->fd, "=========================================================\n\n");

	return CLI_SUCCESS;
}


static char *tklimit_show_chans(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator   i;
	struct tk_limit_chan *tkl_chan;

	switch (cmd) {
	case CLI_INIT:
		e->command = "tklimit show chans";
		e->usage =
			"Usage: tklimit show chans\n"
			"       This will list all chans defined in trunk_calllimit.conf\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n"
	            "==========================================================\n"
	            "----- Configured Trunks chans ----------------------------\n"
	            "==========================================================\n"
	            "===\n");

	ast_mutex_lock(&limit_lock);
	i = ao2_iterator_init(con_tklimit_chans, 0);
	for (; (tkl_chan = ao2_iterator_next(&i)); ao2_ref(tkl_chan, -1)) {
		if (!tkl_chan->enable_time_limit || !tkl_chan->tkl_config) {
			continue;
		}

		ao2_lock(tkl_chan);
		ast_cli(a->fd, "=== ------------------------------------------------------\n"
		            "=== Chan Name:            %s\n"
		            "=== ==> trunk name:       %s\n"
		            "=== ==> cur_consume:      %d(min)\n"
				    "=== ==> history_consume:  %d(min)\n",
		            tkl_chan->name,
		            tkl_chan->tkl_config->name,
		            tkl_chan->cur_consume,
				    tkl_chan->history_consume
				);

		ast_cli(a->fd, "=== ------------------------------------------------------\n===\n");
		ao2_unlock(tkl_chan);
	}
	ao2_iterator_destroy(&i);
	ast_mutex_unlock(&limit_lock);
	ast_cli(a->fd, "==========================================================\n\n");

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_tklimit[] = {
	AST_CLI_DEFINE(tklimit_show_configs, "Show trunk limit configs"),
	AST_CLI_DEFINE(tklimit_show_chans,   "Show trunk limit chans"),
};

static int action_tklimitlists(struct mansession *s, const struct message *m)
{
	const char *actionid = astman_get_header(m, "ActionID");
	struct tk_limit_config *tkl_config;
	struct ao2_iterator     stIterator;
    char   action_id[80] = {0};
	int    remain_time   = 0;
	int    obj_count     = ao2_container_count(con_tklimit_configs);

	astman_send_listack(s, m, "Following are Events for each tklimit_config", "start");

    if (!ast_strlen_zero(actionid)) {
        snprintf(action_id, sizeof(action_id), "ActionID: %s\r\n", actionid);
    }

	stIterator = ao2_iterator_init(con_tklimit_configs, 0);
	while ((tkl_config = ao2_t_iterator_next(&stIterator, "iterate thru tkl_config table"))) {
		ao2_lock(tkl_config);

		remain_time = tkl_config->total_time - tkl_config->history_consume - tkl_config->cur_consume;
		if (remain_time < 0) {
			remain_time = 0;
		}

		astman_append(s, "Event: TrunkTimeUsage\r\n"
                         "%s"
                         "TrunkID: %d\r\n"
                         "TrunkName: %s\r\n"
						 "TotalTime: %d\r\n"
                         "RemainTime: %d\r\n\r\n",
					      action_id, tkl_config->trunk_index, tkl_config->trunk_name, tkl_config->total_time, remain_time);

		ao2_unlock(tkl_config);
		ao2_ref(tkl_config, -1);
	}
	ao2_iterator_destroy(&stIterator);

	/* Send final confirmation */
	astman_send_list_complete_start(s, m, "TrunkTimelistsComplete", obj_count);
	astman_send_list_complete_end(s);

    return 0;
}

static void tklimit_write_json_file(char *filename, char *buffer, int len)
{
	FILE *output;

	if (!filename || !buffer || !len) {
		return;
	}

	if (!(output = fopen(filename, "w+"))) {
		ast_log(LOG_ERROR, "Unable to open/create file %s: %s\n", filename, strerror(errno));
		return;
	}

	fwrite(buffer, len, 1, output);
	fclose(output);
	return;
}

static void tklimit_gen_cur_call_record(void)
{
	char cmd[256] = {0};
	char buffer[32] = {0};
	char file[256]  = {0};
	RAII_VAR(char *, data, NULL, ast_json_free);
	struct ao2_iterator   i;
	struct tk_limit_chan *tkl_chan;
	int  billsec_sum = 0;
	RAII_VAR(struct ast_json *, json_tk_calltime, ast_json_object_create(), ast_json_unref);
	RAII_VAR(struct ast_json *, json_tk_all, NULL, ast_json_unref);

	struct ast_custom_function *custom_func = ast_custom_function_find(TRUNK_TIMESUM_FUNC);

	if (!custom_func || !custom_func->read) {
		snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python %s", GEN_TKDATA_CURRENT);
		ast_safe_system2(cmd);
		return;
	}

	ast_mutex_lock(&limit_lock);
	i = ao2_iterator_init(con_tklimit_chans, 0);
	for (; (tkl_chan = ao2_iterator_next(&i)); ao2_ref(tkl_chan, -1)) {
		if (!tkl_chan->enable_time_limit || !tkl_chan->tkl_config) {
			continue;
		}

		ao2_lock(tkl_chan);

		snprintf(cmd, sizeof(cmd)-1, "SELECT sum(billsec) as calltime FROM cdr WHERE dstchannel LIKE '%s-%%' AND start>='%s';", tkl_chan->name, g_cur_month_start);
		memset(buffer, 0, sizeof(buffer));
		custom_func->read(NULL, TRUNK_TIMESUM_FUNC, cmd, buffer, sizeof(buffer) - 1);

		if (strlen(buffer) > 0) {
			billsec_sum = atoi(buffer);
			if (billsec_sum%60 > 0) {
				billsec_sum = billsec_sum/60 + 1;
			} else {
				billsec_sum = billsec_sum/60;
			}
		} else {
			billsec_sum = 0;
		}

		ast_log(LOG_DEBUG, "%s: %d", tkl_chan->name, billsec_sum);
		ast_json_object_set(json_tk_calltime, tkl_chan->name, ast_json_integer_create(billsec_sum));
		ao2_unlock(tkl_chan);
	}
	ao2_iterator_destroy(&i);
	ast_mutex_unlock(&limit_lock);

	json_tk_all = ast_json_pack("{s: O}", "trunk_calltime", json_tk_calltime);

	data = ast_json_dump_string_format(json_tk_all, AST_JSON_PRETTY);
	tklimit_write_json_file(TRUNK_CALLDATA_FILE_CUR, data, strlen(data));

	/* set current month record */
	snprintf(file, sizeof(file) - 1, "%s/trunk_calldata_%s.json", TRUNK_CALLDATA_PATH, g_cur_month2);

	tklimit_write_json_file(file, data, strlen(data));

	return;
}

static void tklimit_gen_history_call_record(int refresh)
{
	char cmd[256]   = {0};

	snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python %s", GEN_TKDATA_LASTMONTH);
	ast_safe_system2(cmd);
	return;
}

static void tklimit_gen_call_record(int month_refresh)
{
	tklimit_gen_cur_call_record();
	tklimit_gen_history_call_record(month_refresh);
	return;
}

static void tklimit_read_history_call_record(void)
{
	char filename1[256] = {0};
	char filename2[256] = {0};
	char month1[16] = {0};
	char month2[16] = {0};
	RAII_VAR(struct ast_json *, json_s1, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_s2, NULL, ast_json_unref);
	FILE  *file = NULL;
	struct ast_json_iter *iter;
	struct ast_json *json_tk;
	struct ast_json *json_time;
	struct tk_limit_chan   *tkl_chan   = NULL;
	struct tk_limit_config *tkl_config = NULL;
	const char *chan_name;
	int need_month2 = 0;
	int tk_time = 0;
	int num = 0;

	/* Get history call data */
	tklimit_get_history_month(month1, sizeof(month1), month2, sizeof(month2));
	snprintf(filename1, sizeof(filename1)-1, "%s/trunk_calldata_%s.json", TRUNK_CALLDATA_PATH, month1);

	file = fopen(filename1, "r");
	if (!file) {
		return;
	}

	json_s1 = ast_json_load_file(file, NULL);
	if (!json_s1) {
		ast_log(LOG_NOTICE, "Load json object error.");
		return;
	}

	fclose(file);
	json_tk = ast_json_object_get(json_s1, "trunk_calltime");
	if (!json_tk) {
		ast_log(LOG_NOTICE, "Load trunk_calltime object error.");
		return;
	}

	for (iter = ast_json_object_iter(json_tk); iter; iter = ast_json_object_iter_next(json_tk, iter)) {
		chan_name = ast_json_object_iter_key(iter);
		tkl_chan = tklimit_find_chan(chan_name);

		if (!tkl_chan) {
			continue;
		}

		tkl_config = tkl_chan->tkl_config;
		if (!tkl_config) {
			ao2_ref(tkl_chan, -1);
			continue;
		}

		if (tkl_config->en_cycle == TKL_CYCLE_QUARTER) {
			num = tklimit_get_history_num(tkl_config->month);
			if (num > 0) {
				json_time = ast_json_object_iter_value(iter);
				tk_time = (int)ast_json_integer_get(json_time);

				tkl_chan->history_consume += tk_time;
				tkl_config->history_consume += tk_time;
			}

			if (num > 1) {
				ast_log(LOG_NOTICE, "chan[%s], month[%d], num[%d]", chan_name, tkl_config->month, num);
				need_month2 = 1;
			}
		}

		ao2_ref(tkl_chan, -1);
	}

	if (need_month2) {
		snprintf(filename2, sizeof(filename2)-1, "%s/trunk_calldata_%s.json", TRUNK_CALLDATA_PATH, month2);
		file = fopen(filename2, "r");
		if (!file) {
			return;
		}

		json_s2 = ast_json_load_file(file, NULL);
		if (!json_s2) {
			ast_log(LOG_NOTICE, "Load json object error.");
			return;
		}

		fclose(file);
		json_tk = ast_json_object_get(json_s2, "trunk_calltime");
		if (!json_tk) {
			ast_log(LOG_NOTICE, "Load trunk_calltime object error.");
			return;
		}

		for (iter = ast_json_object_iter(json_tk); iter; iter = ast_json_object_iter_next(json_tk, iter)) {
			chan_name = ast_json_object_iter_key(iter);
			tkl_chan = tklimit_find_chan(chan_name);

			if (!tkl_chan) {
				continue;
			}

			tkl_config = tkl_chan->tkl_config;
			if (!tkl_config) {
				ao2_ref(tkl_chan, -1);
				continue;
			}

			if (tkl_config->en_cycle == TKL_CYCLE_QUARTER) {
				num = tklimit_get_history_num(tkl_config->month);
				if (num > 1) {
					json_time = ast_json_object_iter_value(iter);
					tk_time = (int)ast_json_integer_get(json_time);

					tkl_chan->history_consume += tk_time;
					tkl_config->history_consume += tk_time;
				}
			}

			ao2_ref(tkl_chan, -1);
		}
	}

	return;
}

static void tklimit_read_cur_call_record(void)
{
	char filename[256] = {0};
	RAII_VAR(struct ast_json *, json_s, NULL, ast_json_unref);
	FILE  *file = NULL;
	struct ast_json_iter *iter;
	struct ast_json *json_tk;
	struct ast_json *json_time;
	struct tk_limit_chan   *tkl_chan = NULL;
	struct tk_limit_config *tkl_config = NULL;
	struct ao2_iterator i;
	const char *chan_name;
	int    remain_time = 0;
	int    tk_time     = 0;

	/* Get current call data */
	snprintf(filename, sizeof(filename)-1, "%s/%s", TRUNK_CALLDATA_PATH, TRUNK_CALLDATA_CURRENT);
	file = fopen(filename, "r");
	if (!file) {
		return;
	}

	json_s = ast_json_load_file(file, NULL);
	if (!json_s) {
		fclose(file);
		ast_log(LOG_NOTICE, "Load json object error.");
		return;
	}
	fclose(file);

	json_tk = ast_json_object_get(json_s, "trunk_calltime");
	if (!json_tk) {
		ast_log(LOG_NOTICE, "Load trunk_calltime object error.");
		return;
	}

	for (iter = ast_json_object_iter(json_tk); iter; iter = ast_json_object_iter_next(json_tk, iter)) {
		chan_name = ast_json_object_iter_key(iter);
		json_time = ast_json_object_iter_value(iter);
		tk_time = (int)ast_json_integer_get(json_time);
		tkl_chan = tklimit_find_chan(chan_name);

		if (!tkl_chan || !tkl_chan->tkl_config) {
			ao2_cleanup(tkl_chan);
			continue;
		}

		tkl_chan->cur_consume = tk_time;
		tkl_chan->tkl_config->cur_consume += tk_time;
		ao2_cleanup(tkl_chan);
	}

	ast_mutex_lock(&limit_lock);
	i = ao2_iterator_init(con_tklimit_configs, 0);
	for (; (tkl_config = ao2_iterator_next(&i)); ao2_ref(tkl_config, -1)) {
		if (!tkl_config->enable_time_limit) {
			continue;
		}

		//if (tkl_config->update) {
			tkl_config->update = 0;
			if (tkl_config->old_consume != (tkl_config->cur_consume + tkl_config->history_consume)) {
				remain_time = tkl_config->total_time - tkl_config->history_consume - tkl_config->cur_consume;
				if (remain_time < 0) {
					remain_time = 0;
				}

				manager_event(EVENT_FLAG_SYSTEM, "TrunkTimeUsage", "TrunkID: %d\r\nTrunkName: %s\r\nTotalTime: %d\r\nRemainTime: %d\r\n\r\n",
								  tkl_config->trunk_index, tkl_config->trunk_name, tkl_config->total_time, remain_time);

				tkl_config->old_consume = (tkl_config->cur_consume + tkl_config->history_consume);
			}
		//}
	}
	ao2_iterator_destroy(&i);
	ast_mutex_unlock(&limit_lock);

	return;
}

static void tklimit_read_call_record(void)
{
	ast_mutex_lock(&limit_lock);

	tklimit_read_history_call_record();
	tklimit_read_cur_call_record();

	ast_mutex_unlock(&limit_lock);
	return;
}

static void tklimit_update_call_record(int month_update)
{
	ast_mutex_lock(&limit_lock);

	if (month_update) {
		ao2_callback(con_tklimit_configs, 0, tklimit_config_reset_consume, NULL);
		tklimit_gen_call_record(1);
		tklimit_read_call_record();
	} else {
		ao2_callback(con_tklimit_configs, 0, tklimit_config_reset_current, NULL);
		tklimit_gen_cur_call_record();
		tklimit_read_cur_call_record();
	}

	ast_mutex_unlock(&limit_lock);
	return;
}

static int tklimit_update_time(const void *data)
{
	int  month_update = 0;
	char month[16] = {0};

	tklimit_get_current_month(month, sizeof(month)-1);
	month_update = tklimit_check_current_month(month);

	if (g_need_update_tklimit || month_update) {
		ast_log(LOG_DEBUG, "[%s] is new? [%d], update call time: [%d]", g_cur_month, month_update, g_need_update_tklimit);

		ast_rwlock_wrlock(&tklimit_update_lock);
		g_need_update_tklimit = 0;
		ast_rwlock_unlock(&tklimit_update_lock);

		tklimit_update_call_record(month_update);
	}

	return (TKLIMIT_UPDATE_INTVAL * 1000);
}

/* load eventlist configure data, refresh eventlist cb */
static int tklimit_load_proc(int reload)
{
	struct ast_flags config_flags = {0};
	struct ast_config    *cfg = NULL;
	const char *cat = NULL;
	int res = 0;

	if (!reload) {
		con_tklimit_chans = ao2_container_alloc(1, tklimit_chan_hash, tklimit_chan_cmp);
		con_tklimit_configs = ao2_container_alloc(1, tklimit_config_hash, tklimit_config_cmp);
	} else {
		ao2_callback(con_tklimit_configs, 0, tklimit_config_mark, NULL);
	}

    /* load conf file*/
	cfg = ast_config_load(TRUNK_CALLLIMIT_CONFIG_FILE, config_flags);

	if (cfg == CONFIG_STATUS_FILEUNCHANGED || NULL == cfg) {
		return 0;
	}

	if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "trunk_calllimit.conf is invalid.\n");
		return -1;
    }

	g_tklimit_enable_num = 0;

	ast_mutex_lock(&limit_lock);
	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		if (strncasecmp(cat, "trunk_", 6)) {
			continue;
		}
		res = tklimit_build_config(cfg, cat);
	}

	ast_config_destroy(cfg);

	if (reload) {
		ao2_callback(con_tklimit_configs, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, tklimit_config_is_marked, NULL);
	}

	ast_mutex_unlock(&limit_lock);

	ast_log(LOG_NOTICE, "tklimit_load_proc finished, g_tklimit_enable_num[%d].\n", g_tklimit_enable_num);
	return 0;
}

static int tklimit_con_release(void)
{

	ao2_callback(con_tklimit_configs, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, tklimit_config_release_list, NULL);
	ao2_callback(con_tklimit_chans, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	ao2_ref(con_tklimit_configs, -1);
	con_tklimit_configs = NULL;

	ao2_ref(con_tklimit_chans, -1);
	con_tklimit_chans = NULL;

	return 0;
}


static void tklimit_check_update_start(void)
{
	tklimit_update_sched = ast_sched_context_create();
	if (!tklimit_update_sched) {
		return;
	}

	if (ast_sched_start_thread(tklimit_update_sched)) {
		ast_log(LOG_ERROR, "Could not start tklimit_update_sched\n");
		ast_sched_context_destroy(tklimit_update_sched);
		tklimit_update_sched = NULL;
		return;
	}

	tklimit_update_sched_id = ast_sched_add(tklimit_update_sched, TKLIMIT_UPDATE_INTVAL * 1000, tklimit_update_time, NULL);
	if (tklimit_update_sched_id < 0) {
		ast_log(LOG_ERROR, "Could not add update_sched_id\n");
		ast_sched_context_destroy(tklimit_update_sched);
		tklimit_update_sched = NULL;
		return;
	}

	return;
}


static void tklimit_ha_role_switch_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	/* Only do sync when switch to active */
	if (HA_ROLE_ACTIVE != ast_ha_get_role()) {
		return;
	}

	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	reload();
	ast_log(LOG_NOTICE, "tklimit swtiched to Active.\n");

	return;
}

static int tklimit_ha_stasis_route_init(void)
{
	struct stasis_topic *tklimit_ha_data_topic = ast_ha_sync_data_topic_all();
	if (!tklimit_ha_data_topic) {
		ast_log(LOG_WARNING, "Failed to get tklimit_ha_data_topic\n");
		return -1;
	}

	tklimit_ha_data_router = stasis_message_router_create(tklimit_ha_data_topic);
	if (tklimit_ha_data_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create tklimit_ha_data_router\n");
		return -1;
	}

	if (stasis_message_router_add(tklimit_ha_data_router, ast_ha_role_switch_type(), tklimit_ha_role_switch_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_role_switch_type\n");
		return -1;
	}

	return 0;
}

static int unload_module(void)
{
	int res = 0;

	if (tklimit_update_sched) {
		AST_SCHED_DEL(tklimit_update_sched, tklimit_update_sched_id);
		ast_sched_context_destroy(tklimit_update_sched);
		tklimit_update_sched = NULL;
	}

	ast_manager_unregister("TrunkTimelists");
	res |= tklimit_con_release();
	res |= ast_custom_function_unregister(&tklimit_func);
	res |= ast_custom_function_unregister(&tklimit_update_func);
	res |= ast_cli_unregister_multiple(cli_tklimit, ARRAY_LEN(cli_tklimit));

	if (tklimit_ha_data_router) {
		stasis_message_router_unsubscribe(tklimit_ha_data_router);
		tklimit_ha_data_router = NULL;
	}

	ast_mutex_destroy(&limit_lock);
	return res;
}

static int load_module(void)
{
	ast_mutex_init(&limit_lock);

	/*! 1. load trunk_calllimit.conf */
	tklimit_load_proc(0);

	/*! 2. generate trunk call time */
	tklimit_get_current_month(g_cur_month, sizeof(g_cur_month));
	tklimit_gen_call_record(0);

	/*! 3. read json record */
	tklimit_read_call_record();

	ast_cli_register_multiple(cli_tklimit, ARRAY_LEN(cli_tklimit));

	/*! 4. register function */
	ast_custom_function_register(&tklimit_func);
	ast_custom_function_register(&tklimit_update_func);

	ast_manager_register_xml("TrunkTimelists", EVENT_FLAG_SYSTEM | EVENT_FLAG_CALL, action_tklimitlists);

	if (g_tklimit_enable_num > 0) {
		tklimit_check_update_start();
	}

	if (HA_ROLE_STANDBY == ast_ha_get_role()) {
		tklimit_ha_stasis_route_init();
	}

	return 0;
}


static int reload(void)
{
	int res = 0;

	ast_mutex_lock(&tkl_mutex);
	res = unload_module();
	res |= load_module();
	ast_mutex_unlock(&tkl_mutex);

	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Trunk call time statistics and check",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
	.load_pri = AST_MODPRI_APP_DEPEND,
);
