/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

#ifndef _RES_PJSIP_SESSION_H
#define _RES_PJSIP_SESSION_H

/* Needed for pj_timer_entry definition */
#include <pjlib.h>
#include "asterisk/linkedlists.h"
/* Needed for AST_MAX_EXTENSION constant */
#include "asterisk/channel.h"
/* Needed for ast_sockaddr struct */
#include "asterisk/netsock2.h"
/* Needed for ast_sdp_srtp struct */
#include "asterisk/sdp_srtp.h"
/* Needed for ast_media_type */
#include "asterisk/codec.h"
#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP, 2015.07.09 */
#include "asterisk/bfcp.h"
#include "../../apps/confbridge/include/conf_resource.h"
#endif

/* Forward declarations */
struct ast_sip_endpoint;
struct ast_sip_transport;
struct pjsip_inv_session;
struct ast_channel;
struct ast_datastore;
struct ast_datastore_info;
struct ao2_container;
struct pjsip_tx_data;
struct pjsip_rx_data;
struct ast_party_id;
struct pjmedia_sdp_media;
struct pjmedia_sdp_session;
struct ast_dsp;
struct ast_udptl;

/*! \brief T.38 states for a session */
enum ast_sip_session_t38state {
	T38_DISABLED = 0,   /*!< Not enabled */
	T38_LOCAL_REINVITE, /*!< Offered from local - REINVITE */
	T38_PEER_REINVITE,  /*!< Offered from peer - REINVITE */
	T38_ENABLED,        /*!< Negotiated (enabled) */
	T38_REJECTED,       /*!< Refused */
	T38_MAX_ENUM,       /*!< Not an actual state; used as max value in the enum */
};

struct ast_sip_session_sdp_handler;
struct ast_sip_session;
struct ast_sip_session_media;

typedef struct ast_frame *(*ast_sip_session_media_read_cb)(struct ast_sip_session *session, struct ast_sip_session_media *session_media);
typedef int (*ast_sip_session_media_write_cb)(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_frame *frame);

#ifdef GRANDSTREAM_NETWORKS
#define MAX_MSID_LEN 128
#define MAX_CNAME_LEN 128

struct ast_sip_video_rtx_pt {
	int h264_rtx_pt;
	int h265_rtx_pt;
	int vp8_rtx_pt;
	int vp9_rtx_pt;
};
#endif

/*!
 * \brief A structure containing SIP session media information
 */
struct ast_sip_session_media {
	/*! \brief RTP instance itself */
	struct ast_rtp_instance *rtp;
	/*! \brief UDPTL instance itself */
	struct ast_udptl *udptl;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	struct ast_bfcp_udp *bfcp_udp;
	struct ast_bfcp_tcp *bfcp_tcp;
#endif
	/*! \brief Direct media address */
	struct ast_sockaddr direct_media_addr;
	/*! \brief SDP handler that setup the RTP */
	struct ast_sip_session_sdp_handler *handler;
	/*! \brief Holds SRTP information */
	struct ast_sdp_srtp *srtp;
	/*! \brief What type of encryption is in use on this stream */
	enum ast_sip_session_media_encryption encryption;
	/*! \brief The media transport in use for this stream */
	pj_str_t transport;
	/*! \brief Scheduler ID for RTP keepalive */
	int keepalive_sched_id;
	/*! \brief Scheduler ID for RTP timeout */
	int timeout_sched_id;
	/*! \brief Stream is on hold by remote side */
	unsigned int remotely_held:1;
	/*! \brief Stream is on hold by local side */
	unsigned int locally_held:1;
	/*! \brief Does remote support rtcp_mux */
	unsigned int remote_rtcp_mux:1;
	/*! \brief Does remote support ice */
	unsigned int remote_ice:1;
	/*! \brief Media type of this session media */
	enum ast_media_type type;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	enum ast_media_sub_type sub_type;
#endif
	/*! \brief The write callback when writing frames */
	ast_sip_session_media_write_cb write_callback;
	/*! \brief The stream number to place into any resulting frames */
	int stream_num;
	/*! \brief Media identifier for this stream (may be shared across multiple streams) */
	char *mid;
	/*! \brief The bundle group the stream belongs to */
	int bundle_group;
	/*! \brief Whether this stream is currently bundled or not */
	unsigned int bundled;
	/*! \brief Media stream label */
	char mslabel[AST_UUID_STR_LEN];
	/*! \brief Track label */
	char label[AST_UUID_STR_LEN];
	/*! \brief The underlying session has been changed in some fashion */
	unsigned int changed;
	/*! \brief Remote media stream label */
	char *remote_mslabel;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for restore remote msid */
	char remote_msid[MAX_MSID_LEN];
	char remote_cname[MAX_CNAME_LEN];
	/* Add by bfdu for gs-fec */
	unsigned int alone_fec_version:1;
	unsigned int has_rtcp_address:1;
	unsigned int has_nack:1;
	unsigned int has_nack_pli:1;
	unsigned int has_ccm_fir:1;
	unsigned int trickle_ice:1;
	unsigned int rtcp_rsize:1;
	unsigned int ice_restart:1;
	unsigned int setted_turn_target:1;
	unsigned int end_of_candidates:1;
	unsigned int dtls:1;
	unsigned int mpa:1;
	unsigned int zrtp:1;
	char gs_fec_version[MAX_GS_FEC_VERSION_LEN];
	unsigned int transport_cc_id;
	unsigned int remote_port;
	unsigned int remote_rtcp_port;
	unsigned int open_camera;
	unsigned int close_camera;
	unsigned int open_slides;
	int bfcp_sched_id;
	int bfcp_keepalive_id;
	struct ast_sip_video_rtx_pt video_rtx_pt;
	char remote_address[PJ_INET6_ADDRSTRLEN + 2];
	int cn_8k_pt;
	int cn_16k_pt;
	int cn_32k_pt;
	int parent_avs_fd;
	enum ast_media_net_type net_type;
#endif
};

/*!
 * \brief Structure which contains read callback information
 */
struct ast_sip_session_media_read_callback_state {
	/*! \brief The file descriptor itself */
	int fd;
	/*! \brief The callback to invoke */
	ast_sip_session_media_read_cb read_callback;
	/*! \brief The media session */
	struct ast_sip_session_media *session;
};

/*!
 * \brief Structure which contains media state information (streams, sessions)
 */
struct ast_sip_session_media_state {
	/*! \brief Mapping of stream to media sessions */
	AST_VECTOR(, struct ast_sip_session_media *) sessions;
	/*! \brief Added read callbacks - these are whole structs and not pointers */
	AST_VECTOR(, struct ast_sip_session_media_read_callback_state) read_callbacks;
	/*! \brief Default media sessions for each type */
	struct ast_sip_session_media *default_session[AST_MEDIA_TYPE_END];
	/*! \brief The media stream topology */
	struct ast_stream_topology *topology;
};

/*!
 * \brief Opaque structure representing a request that could not be sent
 * due to an outstanding INVITE transaction
 */
struct ast_sip_session_delayed_request;

/*! \brief Opaque struct controlling the suspension of the session's serializer. */
struct ast_sip_session_suspender;

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
#define MAX_FRAMERATE_LEN 8
#define MAX_CONTENT_LEN 8
#define MAX_LABEL_LEN 8
#define MAX_IMAGERTTR_LEN 64
#define MAX_SSRC_ATTR_LEN 128
#define ECAN_PARAM_MAX_LEN 16
#define MAX_MID_LEN 16
#define MAX_CONF_ROOM_LEN 64
#define MAX_USER_AGENT_LEN 128
#define MAX_PROFILE_LEVEID_ID_LEN 64
#define DEFALUT_SLIDES_VIDEO_LABEL 12
#define TERMINAL_TYPE_GVC "SIP Call V=2.0"
#define USER_EXTMAP_TRANSPORT_CC_WHOLE    "http://www.ietf.org/id/draft-holmer-rmcat-transport-wide-cc-extensions-01"
#define AVS_SUPPORT_FEC_VERSION "2,1"

#define H264_PROFILE_LEVELID_1080P  "profile-level-id=42e028"
#define H264_PROFILE_LEVELID_720P   "profile-level-id=42e01f"
#define H264_PROFILE_LEVELID_360P   "profile-level-id=42e016"
#define H264_PROFILE_LEVELID_SLIDE  "profile-level-id=42e028"

#define H264_GVC_PROFILE_LEVELID_1080P  "profile-level-id=428028"   //gvc use:640028 4D0028 428028, for sfu,choose 428028
#define H264_GVC_PROFILE_LEVELID_720P   "profile-level-id=42801f"
#define H264_GVC_PROFILE_LEVELID_360P   "profile-level-id=428016"
#define H264_GVC_PROFILE_LEVELID_SLIDE  "profile-level-id=428028"

#define H264_IMAGEATTR_1080P        "send [x=1920,y=1080] recv [x=1920,y=1080]"
#define H264_IMAGEATTR_720P         "send [x=1280,y=720] recv [x=1280,y=720]"
#define H264_IMAGEATTR_360P         "send [x=640,y=360] recv [x=640,y=360]"
#define H264_WAVE_IMAGEATTR_1080P        "recv [x=1920,y=1080]"
#define H264_WAVE_IMAGEATTR_720P         "recv [x=1280,y=720]"
#define H264_WAVE_IMAGEATTR_360P         "recv [x=640,y=360]"
#define H264_IMAGEATTR_SLIDE        "send [x=1920,y=1080] recv [x=1920,y=1080]"

#define AST_BW_HEADER_SIZE (192)                          /* header : 192(IP+UDP+RTP) */
#define AST_1080P_AS       (1024 + AST_BW_HEADER_SIZE)    /* AS:1216 = 1024 + 192 (kbits/s) */
#define AST_1080P_TIAS     (1024 * 1000)              	  /* 1024 *1000 */
#define AST_720P_AS        (512 + AST_BW_HEADER_SIZE)     /* AS:704 = 512 + 192 (kbits/s) */
#define AST_720P_TIAS      (512 * 1000)                   /* 512 Kbps *1000 */
#define AST_360P_AS        (172 + AST_BW_HEADER_SIZE)     /* AS:364 = 172 + 192 (kbits/s) */
#define AST_360P_TIAS      (172 * 1000)                   /* 250 Kbps *1000 */
#define AST_SLIDES_VIDEO_AS      (512 + 192)              /* AS:704 = 512 + 192 (kbits/s) */
#define AST_SLIDES_VIDEO_TIAS    (512 * 1000)             /* 512 Kbps *1000 */
#define AST_SLIDES_VIDEO_FRAMERATE    (5)                 /* framerate:5 for slides video */

#define AST_1080P_LEVEL_ID 40                             /* 28 hex according Dec is 40 */
#define AST_720P_LEVEL_ID 31                              /* 1f hex according Dec is 31 */

#define MAX_SDP_BODY_LEN 4096

struct ast_crm_info {
	char caller_name[128];
	char caller_number[128];
	char company[256];
};

struct bfcp_info_data {
	struct ast_sip_session *session;
	struct ast_bfcp_info *bfcp_info;
};

enum bfcp_sdp_floorctl
{
	BFCP_FLOORCTL_C_ONLY   = 1,
	BFCP_FLOORCTL_S_ONLY   = 2,
	BFCP_FLOORCTL_C_S      = 3,
	BFCP_FLOORCTL_MAX
};

enum bfcp_sdp_setup
{
	BFCP_SETUP_ACTIVE    = 1,
	BFCP_SETUP_PASSIVE   = 2,
	BFCP_SETUP_ACTPASS   = 3,
	BFCP_SETUP_MAX
};

enum ast_sfu_video_codec {
	AST_SFU_VIDEO_UNKNOW = 0,
	AST_SFU_VIDEO_H264,
	AST_SFU_VIDEO_H265,
	AST_SFU_VIDEO_VP9,
};

enum ast_neg_pool_type {
	AST_NEG_POOL_INV_PROV = 0,
	AST_NEG_POOL_NOTIFY_PROV,
};

enum ast_ssrc_type {
	AST_SSRC_TYPE_UNKNOWN = 0,
	AST_SSRC_TYPE_MEDIA,
	AST_SSRC_TYPE_RTX,
	AST_SSRC_TYPE_FEC,
};

enum ast_gs_state_item_type {
	AST_GS_STATE_ITEM_TYPE_UNKNOW          =  0,
	AST_GS_STATE_ITEM_TYPE_TERMINAL        =  1,
	AST_GS_STATE_ITEM_TYPE_GESTURE         =  2,
};

enum ast_sip_camera_control_action {
	AST_SIP_CAMERA_CONTROL_ACTION_UNKNOW       =  0,
	AST_SIP_CAMERA_CONTROL_ACTION_OPEN         =  1,
	AST_SIP_CAMERA_CONTROL_ACTION_CLOSE        =  2,
};

struct ast_sip_gs_state_data {
	enum ast_gs_state_item_type type;
	unsigned int muted;
	unsigned int hand_up;
	char channel_name[AST_CHANNEL_NAME];
};

struct gs_state_info_data {
	struct ast_sip_session *session;
	struct ast_sip_gs_state_data *gs_state_data;
};

enum ast_sip_session_gs_state_type {
	AST_GS_STATE_TYPE_UNKNOW,
	AST_GS_STATE_TYPE_CONF_TIMEOUT,
	AST_GS_STATE_TYPE_TRUNK_INFO,
	AST_GS_STATE_TYPE_ANSWERED_STATUS,
};

#define MAX_CONFTIMEOUT_CONF 64
#define MAX_CONFTIMEOUT_TYPE 16
#define MAX_CONFTIMEOUT_TIME 16
#define MAX_CONFTIMEOUT_STATUS 16
#define MAX_CONF_LINK_LEN 512
#define MAX_CHANNEL_ID_LEN 152

struct ast_conf_timeout_state {
	char conf[MAX_CONFTIMEOUT_CONF];
	char type[MAX_CONFTIMEOUT_TYPE];
	char time[MAX_CONFTIMEOUT_TIME];
	char status[MAX_CONFTIMEOUT_STATUS];
};

struct ast_trunk_info {
	char conf_link[MAX_CONF_LINK_LEN];
	char channel_id[MAX_CHANNEL_ID_LEN];
};

struct ast_answered_status{
	char answered[6];
};

struct ast_gs_state_data {
	struct ast_sip_session *session;
	enum ast_sip_session_gs_state_type type;
	void *data;
};

struct ast_gs_state_clone_data {
	enum ast_sip_session_gs_state_type type;
	void *data;
};
#endif

/*!
 * \brief A structure describing a SIP session
 *
 * For the sake of brevity, a "SIP session" in Asterisk is referring to
 * a dialog initiated by an INVITE. While "session" is typically interpreted
 * to refer to the negotiated media within a SIP dialog, we have opted
 * to use the term "SIP session" to refer to the INVITE dialog itself.
 */
struct ast_sip_session {
	/*! Dialplan extension where incoming call is destined */
	char exten[AST_MAX_EXTENSION];
	/*! The endpoint with which Asterisk is communicating */
	struct ast_sip_endpoint *endpoint;
	/*! The contact associated with this session */
	struct ast_sip_contact *contact;
	/*! The PJSIP details of the session, which includes the dialog */
	struct pjsip_inv_session *inv_session;
	/*! The Asterisk channel associated with the session */
	struct ast_channel *channel;
	/*! Registered session supplements */
	AST_LIST_HEAD(, ast_sip_session_supplement) supplements;
	/*! Datastores added to the session by supplements to the session */
	struct ao2_container *datastores;
	/*! Serializer for tasks relating to this SIP session */
	struct ast_taskprocessor *serializer;
	/*! Non-null if the session serializer is suspended or being suspended. */
	struct ast_sip_session_suspender *suspended;
	/*! Requests that could not be sent due to current inv_session state */
	AST_LIST_HEAD_NOLOCK(, ast_sip_session_delayed_request) delayed_requests;
	/*! When we need to reschedule a reinvite, we use this structure to do it */
	pj_timer_entry rescheduled_reinvite;
	/*! Format capabilities pertaining to direct media */
	struct ast_format_cap *direct_media_cap;
	/*! When we need to forcefully end the session */
	pj_timer_entry scheduled_termination;
	/*! Identity of endpoint this session deals with */
	struct ast_party_id id;
	/*! Active media state (sessions + streams) - contents are guaranteed not to change */
	struct ast_sip_session_media_state *active_media_state;
	/*! Pending media state (sessions + streams) */
	struct ast_sip_session_media_state *pending_media_state;
	/*! Optional DSP, used only for inband DTMF/Fax-CNG detection if configured */
	struct ast_dsp *dsp;
	/*! Whether the termination of the session should be deferred */
	unsigned int defer_terminate:1;
	/*! Termination requested while termination deferred */
	unsigned int terminate_while_deferred:1;
#ifdef GRANDSTREAM_NETWORKS
	int is_video_page_call;
	int is_gds_call;
	int is_gsc_call;
	int is_hikvision_call;
	int blocking_collect_calls; /*add by pengwang, if 1 blocking backward collect calls*/
	char user_agent[MAX_USER_AGENT_LEN];
	unsigned int cc_offer:1;
	int core_id;
	enum ast_cc_service_type cc_service;
	pj_str_t    cc_subs_user;
	pjsip_uri  *cc_subs_uri; 
	char       *contact_user;   /*! Added by yfyang, for Bug 50520 */
	int is_outgoing;
	unsigned int recv_held_private:1;   /*! to indicate if received held private, for for sca feature */
	unsigned int has_insert_sca_user:1; /*! Added by yfyang, for bug 107738 */
	unsigned int line_index;   /*! the sca line index, for for sca feature */
	unsigned int is_sfu:1;
	unsigned int joint_ipvt:1;
	/* Begin add by bfdu for video reinivite, 2015.05.27 */
	unsigned int is_media_change:1;
	unsigned int is_media_count_change:1;
	unsigned int is_direct_media:1;
	unsigned int is_sdp_ipv6:1;
	unsigned int is_yealink_msg:1;
	unsigned int recv_reinvite_no_sdp:1;
	unsigned int open_slides_video:1;
	unsigned int need_notify_avs_early_media:1;
	struct ast_bfcp_para bfcp_para;    					/*!< bfcp parameters */
	int reinvite_failed_code;
	/* End add by bfdu for video reinivite, 2015.05.27 */
	unsigned int wImageReInvite:1;
	unsigned int wLocalReInvite:1;
	unsigned int wSwitchImageInvite488ToAudioInvite:1;
	/*! End: Added by pxie@grandstream.cn for bug[42261], 2015-04-25*/
	unsigned int fax_alreay_work:1;
	unsigned int is_intelligent_route:1;
	unsigned int is_fax_sending_request:1; //Added by pxie@grandstream.cn for bug[45683 and 42261], 2015-07-16.
	unsigned int is_image_audio_two_media:1;
	unsigned int image_port_forbidden:1;
	unsigned int two_media_is_support_image:1;
	unsigned int wFaxDetect;
	unsigned int intranet_ip_filter:1;
	unsigned int is_webrtc_call;
	unsigned int session_uas:1;
	unsigned int audio_has_modem:1;
	unsigned int need_orig_callee:1;
	unsigned int end_of_candidates:1;
	unsigned int sent_bfcp_hello:1;
	unsigned int sfu_invite_by_local_chan:1;
	unsigned int support_ice:1;
	unsigned int support_trickle_ice:1;
	unsigned int ssh_proxy_flag:1;           /* Added by yfyang, for sip ssh proxy feature */	
	unsigned int camera_control_operater:1;
	unsigned int media_turn_relay:1;
	unsigned int auto_record:1;              /* Added by zgmo for sdp media record attribute */
	unsigned int dump:1;
	unsigned int notify_full_participant_sdp:1;
	unsigned int remote_support_bundle:1;
	unsigned int close_camera_forcely:1;
	unsigned int pending_bfcp_info:1;
	unsigned int pending_bfcp_message:1;
	unsigned int join_sfu:1;
	unsigned int support_whiteboard:1;
	unsigned int conf_video_recording:1;
	unsigned int ignore_avs_response:1; /*Ignore the response for del/stop avs request*/
	unsigned int allow_talk_event:1;
	unsigned int allow_hold_event:1;
	unsigned int cand_host_only:1;
	unsigned int sending_participant_notify:1;
	unsigned int focus_video:1;
	unsigned int sfu_recv_slides_video:1;
	unsigned int notify_share_audio:1;
	unsigned int ucm_trunk:1;
	unsigned int get_peer_session:1;
	unsigned int emergemcy_call:1;
	unsigned int support_gs_fec:1;
	unsigned int send_bfcp_request_by_info:1;
	unsigned int change_resolution_reinvite:1;
	unsigned int release_hd_by_wave:1;
	unsigned int send_final_rsp:1;
	unsigned int disabled_video:1;
	unsigned int trickle_info_delayed:1;
	unsigned int rc_limit_already_processed:1;
	unsigned int ref_added:1;
	unsigned int update_avs_slides_video:1;
	unsigned int insert_video_channel:1;	/* 1:insert, 0:remove */
	unsigned int send_peer_reinvite_for_callflip:1;
	unsigned int delayed_send_session:1;
	unsigned int support_gs_state;
	unsigned int had_send_answered;
	unsigned int is_callflip_call;
	unsigned int diff_multiavs_ip;//mark  wheter if the ip of session is differnt from avs network ip
	unsigned int is_fastprov:1;
	unsigned int initial_authed:1;

	struct ast_bfcp_info last_bfcp_info;
	char audio_ecan_param[ECAN_PARAM_MAX_LEN];
	char conf_room[MAX_CONF_ROOM_LEN];
	char turn_target_address[PJ_INET6_ADDRSTRLEN + 2];
	char avs_address[PJ_INET6_ADDRSTRLEN + 2];
	/*! when hold requst send to peer, both caller and callee this value will inc one, bug 77064 */
	/* Add by bfdu for bug 78568, 2017.02.08 */
	struct ast_crm_info crm_info;
	enum ast_terminal_type terminal_type;
	int msid_count;
	unsigned int has_dec_gs_call_count;/*! added by zyxia for ios gswave,2018.8.31 */
	unsigned int has_add_gs_call_count;/*! added by zyxia for ios gswave,2018.8.31 */
	unsigned int gswave_app_session_active;/*! added by zyxia for ios gswave wake up when change network in app,2021.2.22 */
	unsigned int is_gswave_app_call;
	unsigned int delayed_reinvite_count;
	unsigned int sending_notify_count;
	int conf_hangupcause;
	unsigned int port_group_id;
	unsigned int src_addr_ver;
	struct in_addr src_addr;
	struct in6_addr src_addr6;
	enum ast_sip_dtmf_mode negotiated_dtmf;
	enum ast_sfu_video_codec sfu_video_codec;
	struct ast_sip_session_media_state *active_participant_media_state;
	struct ast_sip_session_media_state *pending_participant_media_state;
	pjmedia_sdp_session *active_notify_sdp;
	pjmedia_sdp_session *neg_notify_sdp;
	struct ast_sip_session_media *participant_bundled_media;
	int configured_video_resolution;
	int reject_call;
	unsigned int is_rtsp_call;//added by zyxia for rtsp call,2021.11.16
	unsigned int is_trunk_conf_invite;
	char rtsp_profile_level_id[MAX_PROFILE_LEVEID_ID_LEN];
	enum ast_sip_camera_control_action camera_control_action;
	unsigned int nat_call_left_time;
	enum ast_record_stop_type record_stop_type;
	struct ast_sip_sched_task *play_sound_task;
	struct ast_sip_sched_task *reject_nat_call_task;
	struct ast_sip_sched_task *send_real_answered;
	char *conf_link;
	char *trunk_channel_id;
	char *replace_channel_id;
	int main_video_first_multi_h264_pt;
	int main_video_first_multi_rtx_pt;
	int comfort_noise;
	char *trunk_conf_num;
	int had_subscribe_trunk_cei;
	unsigned int conf_invite_call;
	char *remote_trunk_channel_id;//for outbound to ucm  conf through trunk
	char caller_gs_client_info[MAX_USER_AGENT_LEN];
	char callee_gs_client_info[MAX_USER_AGENT_LEN];
	int  if_type;
	struct ast_gs_state_clone_data *gs_state_clone_data;
	char sbc_endpoint_addr[80];  /* Added by yfyang, for clouducm support udp */
	char sbc_addr[80];  /* Added by yfyang, for Bug292540 */
#endif
	/*! Deferred incoming re-invite */
	pjsip_rx_data *deferred_reinvite;
	/*! Current T.38 state */
	enum ast_sip_session_t38state t38state;
	/*! The AOR associated with this session */
	struct ast_sip_aor *aor;
	/*! From header saved at invite creation */
	pjsip_fromto_hdr *saved_from_hdr;
	/*! Whether the end of the session should be deferred */
	unsigned int defer_end:1;
	/*! Session end (remote hangup) requested while termination deferred */
	unsigned int ended_while_deferred:1;
	/*! DTMF mode to use with this session, from endpoint but can change */
	enum ast_sip_dtmf_mode dtmf;
	/*! Initial incoming INVITE Request-URI.  NULL otherwise. */
	pjsip_uri *request_uri;
#ifdef GRANDSTREAM_NETWORKS
	AST_LIST_ENTRY(ast_sip_session) next;
#endif
	char *active_contact_addr; /* when multiple phones register on, record the phone host of the call incoming */
	char *trans_type; /* UDP/TCP/TLS */
};

typedef int (*ast_sip_session_request_creation_cb)(struct ast_sip_session *session, pjsip_tx_data *tdata);
typedef int (*ast_sip_session_response_cb)(struct ast_sip_session *session, pjsip_rx_data *rdata);
typedef int (*ast_sip_session_sdp_creation_cb)(struct ast_sip_session *session, pjmedia_sdp_session *sdp);

/*!
 * \brief Describes when a supplement should be called into on incoming responses.
 *
 * In most cases, session supplements will not need to worry about this because in most cases,
 * the correct value will be automatically applied. However, there are rare circumstances
 * when a supplement will want to specify when it should be called.
 *
 * The values below are listed in chronological order.
 */
enum ast_sip_session_response_priority {
	/*!
	 * When processing 3XX responses, the supplement is called into before
	 * the redirecting information is processed.
	 */
	AST_SIP_SESSION_BEFORE_REDIRECTING = (1 << 0),
	/*!
	 * For responses to INVITE transactions, the supplement is called into
	 * before media is negotiated.
	 *
	 * This priority is applied by default to any session supplement that
	 * does not specify a response priority.
	 */
	AST_SIP_SESSION_BEFORE_MEDIA = (1 << 1),
	/*!
	 * For INVITE transactions, the supplement is called into after media
	 * is negotiated.
	 */
	AST_SIP_SESSION_AFTER_MEDIA = (1 << 2),
};

/*!
 * \brief A supplement to SIP message processing
 *
 * These can be registered by any module in order to add
 * processing to incoming and outgoing SIP requests and responses
 */
struct ast_sip_session_supplement {
	/*! Method on which to call the callbacks. If NULL, call on all methods */
	const char *method;
	/*! Priority for this supplement. Lower numbers are visited before higher numbers */
	enum ast_sip_supplement_priority priority;
	/*!
	 * \brief Notification that the session has begun
	 * This method will always be called from a SIP servant thread.
	 */
	void (*session_begin)(struct ast_sip_session *session);
	/*!
	 * \brief Notification that the session has ended
	 *
	 * This method may or may not be called from a SIP servant thread. Do
	 * not make assumptions about being able to call PJSIP methods from within
	 * this method.
	 */
	void (*session_end)(struct ast_sip_session *session);
	/*!
	 * \brief Notification that the session is being destroyed
	 */
	void (*session_destroy)(struct ast_sip_session *session);
	/*!
	 * \brief Called on incoming SIP request
	 * This method can indicate a failure in processing in its return. If there
	 * is a failure, it is required that this method sends a response to the request.
	 * This method is always called from a SIP servant thread.
	 *
	 * \note
	 * The following PJSIP methods will not work properly:
	 * pjsip_rdata_get_dlg()
	 * pjsip_rdata_get_tsx()
	 * The reason is that the rdata passed into this function is a cloned rdata structure,
	 * and its module data is not copied during the cloning operation.
	 * If you need to get the dialog, you can get it via session->inv_session->dlg.
	 *
	 * \note
	 * There is no guarantee that a channel will be present on the session when this is called.
	 */
	int (*incoming_request)(struct ast_sip_session *session, struct pjsip_rx_data *rdata);
	/*!
	 * \brief Called on an incoming SIP response
	 * This method is always called from a SIP servant thread.
	 *
	 * \note
	 * The following PJSIP methods will not work properly:
	 * pjsip_rdata_get_dlg()
	 * pjsip_rdata_get_tsx()
	 * The reason is that the rdata passed into this function is a cloned rdata structure,
	 * and its module data is not copied during the cloning operation.
	 * If you need to get the dialog, you can get it via session->inv_session->dlg.
	 *
	 * \note
	 * There is no guarantee that a channel will be present on the session when this is called.
	 */
	void (*incoming_response)(struct ast_sip_session *session, struct pjsip_rx_data *rdata);
	/*!
	 * \brief Called on an outgoing SIP request
	 * This method is always called from a SIP servant thread.
	 */
	void (*outgoing_request)(struct ast_sip_session *session, struct pjsip_tx_data *tdata);
	/*!
	 * \brief Called on an outgoing SIP response
	 * This method is always called from a SIP servant thread.
	 */
	void (*outgoing_response)(struct ast_sip_session *session, struct pjsip_tx_data *tdata);
	/*! Next item in the list */
	AST_LIST_ENTRY(ast_sip_session_supplement) next;
	/*!
	 * Determines when the supplement is processed when handling a response.
	 * Defaults to AST_SIP_SESSION_BEFORE_MEDIA
	 */
	enum ast_sip_session_response_priority response_priority;
};

enum ast_sip_session_sdp_stream_defer {
	/*! The stream was not handled by this handler. If there are other registered handlers for this stream type, they will be called. */
	AST_SIP_SESSION_SDP_DEFER_NOT_HANDLED,
	/*! There was an error encountered. No further operations will take place and the current negotiation will be abandoned. */
	AST_SIP_SESSION_SDP_DEFER_ERROR,
	/*! Re-invite is not needed */
	AST_SIP_SESSION_SDP_DEFER_NOT_NEEDED,
	/*! Re-invite should be deferred and will be resumed later. No further operations will take place. */
	AST_SIP_SESSION_SDP_DEFER_NEEDED,
};

/*!
 * \brief A handler for SDPs in SIP sessions
 *
 * An SDP handler is registered by a module that is interested in being the
 * responsible party for specific types of SDP streams.
 */
struct ast_sip_session_sdp_handler {
	/*! An identifier for this handler */
	const char *id;
	/*!
	 * \brief Determine whether a stream requires that the re-invite be deferred.
	 * If a stream can not be immediately negotiated the re-invite can be deferred and
	 * resumed at a later time. It is up to the handler which caused deferral to occur
	 * to resume it.
	 *
	 * \param session The session for which the media is being re-invited
	 * \param session_media The media being reinvited
	 * \param sdp The entire SDP. Useful for getting "global" information, such as connections or attributes
	 * \param stream PJSIP incoming SDP media lines to parse by handler.
	 *
	 * \return enum ast_sip_session_defer_stream
	 *
	 * \note This is optional, if not implemented the stream is assumed to not be deferred.
	 */
	enum ast_sip_session_sdp_stream_defer (*defer_incoming_sdp_stream)(struct ast_sip_session *session, struct ast_sip_session_media *session_media, const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream);
	/*!
	 * \brief Set session details based on a stream in an incoming SDP offer or answer
	 * \param session The session for which the media is being negotiated
	 * \param session_media The media session
	 * \param sdp The entire SDP. Useful for getting "global" information, such as connections or attributes
	 * \param index The index for the session media, Asterisk stream, and PJMEDIA stream being negotiated
	 * \param asterisk_stream The Asterisk stream representation
	 * \retval 0 The stream was not handled by this handler. If there are other registered handlers for this stream type, they will be called.
	 * \retval <0 There was an error encountered. No further operation will take place and the current negotiation will be abandoned.
	 * \retval >0 The stream was handled by this handler. No further handler of this stream type will be called.
	 */
	int (*negotiate_incoming_sdp_stream)(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
		const struct pjmedia_sdp_session *sdp, int index, struct ast_stream *asterisk_stream);
	/*!
	 * \brief Create an SDP media stream and add it to the outgoing SDP offer or answer
	 * \param session The session for which media is being added
	 * \param session_media The media to be setup for this session
	 * \param sdp The entire SDP as currently built
	 * \param remote Optional remote SDP if this is an answer
	 * \param stream The stream that is to be added to the outgoing SDP
	 * \retval 0 This handler has no stream to add. If there are other registered handlers for this stream type, they will be called.
	 * \retval <0 There was an error encountered. No further operation will take place and the current SDP negotiation will be abandoned.
	 * \retval >0 The handler has a stream to be added to the SDP. No further handler of this stream type will be called.
	 */
	int (*create_outgoing_sdp_stream)(struct ast_sip_session *session, struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp,
		const struct pjmedia_sdp_session *remote, struct ast_stream *stream);
	/*!
	 * \brief Update media stream with external address if applicable
	 * \param tdata The outgoing message itself
	 * \param stream The stream on which to operate
	 * \param transport The transport the SDP is going out on
	 */
	void (*change_outgoing_sdp_stream_media_address)(struct pjsip_tx_data *tdata, struct pjmedia_sdp_media *stream, struct ast_sip_transport *transport);
	/*!
	 * \brief Apply a negotiated SDP media stream
	 * \param session The session for which media is being applied
	 * \param session_media The media session
	 * \param local The entire local negotiated SDP
	 * \param remote The entire remote negotiated SDP
	 * \param index The index of the session media, SDP streams, and Asterisk streams
	 * \param asterisk_stream The Asterisk stream representation
	 * \retval 0 The stream was not applied by this handler. If there are other registered handlers for this stream type, they will be called.
	 * \retval <0 There was an error encountered. No further operation will take place and the current application will be abandoned.
	 * \retval >0 The stream was handled by this handler. No further handler of this stream type will be called.
	 */
	int (*apply_negotiated_sdp_stream)(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
		const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_session *remote, int index,
		struct ast_stream *asterisk_stream);
	/*!
	 * \brief Stop a session_media created by this handler but do not destroy resources
	 * \param session The session for which media is being stopped
	 * \param session_media The media to destroy
	 */
	void (*stream_stop)(struct ast_sip_session_media *session_media);
	/*!
	 * \brief Destroy a session_media created by this handler
	 * \param session The session for which media is being destroyed
	 * \param session_media The media to destroy
	 */
	void (*stream_destroy)(struct ast_sip_session_media *session_media);
	/*! Next item in the list. */
	AST_LIST_ENTRY(ast_sip_session_sdp_handler) next;
};

/*!
 * \brief A structure which contains a channel implementation and session
 */
struct ast_sip_channel_pvt {
	/*! \brief Pointer to channel specific implementation information, must be ao2 object */
	void *pvt;
	/*! \brief Pointer to session */
	struct ast_sip_session *session;
};

/*!
 * \brief Allocate a new SIP channel pvt structure
 *
 * \param pvt Pointer to channel specific information
 * \param session Pointer to SIP session
 *
 * \retval non-NULL success
 * \retval NULL failure
 */
struct ast_sip_channel_pvt *ast_sip_channel_pvt_alloc(void *pvt, struct ast_sip_session *session);

/*!
 * \brief Allocate a new SIP session
 *
 * This will take care of allocating the datastores container on the session as well
 * as placing all registered supplements onto the session.
 *
 * The endpoint that is passed in will have its reference count increased by one since
 * the session will be keeping a reference to the endpoint. The session will relinquish
 * this reference when the session is destroyed.
 *
 * \param endpoint The endpoint that this session communicates with
 * \param contact The contact associated with this session
 * \param inv_session The PJSIP INVITE session data
 * \param rdata INVITE request received (NULL if for outgoing allocation)
 */
struct ast_sip_session *ast_sip_session_alloc(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, pjsip_inv_session *inv, pjsip_rx_data *rdata);

/*!
 * \brief Request and wait for the session serializer to be suspended.
 * \since 12.7.0
 *
 * \param session Which session to suspend the serializer.
 *
 * \note No channel locks can be held while calling without risk of deadlock.
 *
 * \return Nothing
 */
void ast_sip_session_suspend(struct ast_sip_session *session);

/*!
 * \brief Request the session serializer be unsuspended.
 * \since 12.7.0
 *
 * \param session Which session to unsuspend the serializer.
 *
 * \return Nothing
 */
void ast_sip_session_unsuspend(struct ast_sip_session *session);

/*!
 * \brief Create a new outgoing SIP session
 *
 * The endpoint that is passed in will have its reference count increased by one since
 * the session will be keeping a reference to the endpoint. The session will relinquish
 * this reference when the session is destroyed.
 *
 * \param endpoint The endpoint that this session uses for settings
 * \param contact The contact that this session will communicate with
 * \param location Name of the location to call, be it named location or explicit URI. Overrides contact if present.
 * \param request_user Optional request user to place in the request URI if permitted
 * \param req_topology The requested capabilities
 */
#ifdef GRANDSTREAM_NETWORKS
struct ast_sip_session *ast_sip_session_create_outgoing(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, const char *location, const char *request_user,
	struct ast_stream_topology *req_topology, struct ast_channel *requestor,pj_str_t *local, pj_str_t *remote, pj_str_t *callid);
#else
struct ast_sip_session *ast_sip_session_create_outgoing(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, const char *location, const char *request_user,
	struct ast_stream_topology *req_topology);
#endif

/*!
 * \brief Terminate a session and, if possible, send the provided response code
 *
 * \param session The session to terminate
 * \param response The response code to use for termination if possible
 *
 * \warning Calling this function MAY cause the last session reference to be
 * released and the session destructor to be called.  If you need to do something
 * with session after this call, be sure to bump the ref count before calling terminate.
 */
void ast_sip_session_terminate(struct ast_sip_session *session, int response);

/*!
 * \brief Defer local termination of a session until remote side terminates, or an amount of time passes
 *
 * \param session The session to defer termination on
 *
 * \retval 0 Success
 * \retval -1 Failure
 */
int ast_sip_session_defer_termination(struct ast_sip_session *session);

/*!
 * \brief Cancel a pending deferred termination.
 *
 * \param session The session to cancel a deferred termination on.
 */
void ast_sip_session_defer_termination_cancel(struct ast_sip_session *session);

/*!
 * \brief End the session if it had been previously deferred
 *
 * \param session The session to end if it had been deferred
 */
void ast_sip_session_end_if_deferred(struct ast_sip_session *session);

/*!
 * \brief Register an SDP handler
 *
 * An SDP handler is responsible for parsing incoming SDP streams and ensuring that
 * Asterisk can cope with the contents. Similarly, the SDP handler will be
 * responsible for constructing outgoing SDP streams.
 *
 * Multiple handlers for the same stream type may be registered. They will be
 * visited in the order they were registered. Handlers will be visited for each
 * stream type until one claims to have handled the stream.
 *
 * \param handler The SDP handler to register
 * \param stream_type The type of media stream for which to call the handler
 * \retval 0 Success
 * \retval -1 Failure
 */
int ast_sip_session_register_sdp_handler(struct ast_sip_session_sdp_handler *handler, const char *stream_type);

/*!
 * \brief Unregister an SDP handler
 *
 * \param handler The SDP handler to unregister
 * \param stream_type Stream type for which the SDP handler was registered
 */
void ast_sip_session_unregister_sdp_handler(struct ast_sip_session_sdp_handler *handler, const char *stream_type);

/*!
 * \brief Register a supplement to SIP session processing
 *
 * This allows for someone to insert themselves in the processing of SIP
 * requests and responses. This, for example could allow for a module to
 * set channel data based on headers in an incoming message. Similarly,
 * a module could reject an incoming request if desired.
 *
 * \param module Referenced module(NULL safe)
 * \param supplement The supplement to register
 */
void ast_sip_session_register_supplement_with_module(struct ast_module *module, struct ast_sip_session_supplement *supplement);

#define ast_sip_session_register_supplement(supplement) \
		ast_sip_session_register_supplement_with_module(AST_MODULE_SELF, supplement)

/*!
 * \brief Unregister a an supplement to SIP session processing
 *
 * \param supplement The supplement to unregister
 */
void ast_sip_session_unregister_supplement(struct ast_sip_session_supplement *supplement);

/*!
 * \brief Add supplements to a SIP session
 *
 * \param session The session to initialize
 */
int ast_sip_session_add_supplements(struct ast_sip_session *session);

/*!
 * \brief Remove supplements from a SIP session
 *
 * \param session The session to remove
 */
void ast_sip_session_remove_supplements(struct ast_sip_session *session);

/*!
 * \brief Alternative for ast_datastore_alloc()
 *
 * There are two major differences between this and ast_datastore_alloc()
 * 1) This allocates a refcounted object
 * 2) This will fill in a uid if one is not provided
 *
 * DO NOT call ast_datastore_free() on a datastore allocated in this
 * way since that function will attempt to free the datastore rather
 * than play nicely with its refcount.
 *
 * \param info Callbacks for datastore
 * \param uid Identifier for datastore
 * \retval NULL Failed to allocate datastore
 * \retval non-NULL Newly allocated datastore
 */
struct ast_datastore *ast_sip_session_alloc_datastore(const struct ast_datastore_info *info, const char *uid);

/*!
 * \brief Add a datastore to a SIP session
 *
 * Note that SIP uses reference counted datastores. The datastore passed into this function
 * must have been allocated using ao2_alloc() or there will be serious problems.
 *
 * \param session The session to add the datastore to
 * \param datastore The datastore to be added to the session
 * \retval 0 Success
 * \retval -1 Failure
 */
int ast_sip_session_add_datastore(struct ast_sip_session *session, struct ast_datastore *datastore);

/*!
 * \brief Retrieve a session datastore
 *
 * The datastore retrieved will have its reference count incremented. When the caller is done
 * with the datastore, the reference counted needs to be decremented using ao2_ref().
 *
 * \param session The session from which to retrieve the datastore
 * \param name The name of the datastore to retrieve
 * \retval NULL Failed to find the specified datastore
 * \retval non-NULL The specified datastore
 */
struct ast_datastore *ast_sip_session_get_datastore(struct ast_sip_session *session, const char *name);

/*!
 * \brief Remove a session datastore from the session
 *
 * This operation may cause the datastore's free() callback to be called if the reference
 * count reaches zero.
 *
 * \param session The session to remove the datastore from
 * \param name The name of the datastore to remove
 */
void ast_sip_session_remove_datastore(struct ast_sip_session *session, const char *name);

/*!
 * \brief Send a reinvite or UPDATE on a session
 *
 * This method will inspect the session in order to construct an appropriate
 * session refresh request. As with any outgoing request in res_pjsip_session,
 * this will call into registered supplements in case they wish to add anything.
 *
 * Note: The on_request_creation callback may or may not be called in the same
 * thread where this function is called. Request creation may need to be delayed
 * due to the current INVITE transaction state.
 *
 * \param session The session on which the reinvite will be sent
 * \param on_request_creation Callback called when request is created
 * \param on_sdp_creation Callback called when SDP is created
 * \param on_response Callback called when response for request is received
 * \param method The method that should be used when constructing the session refresh
 * \param generate_new_sdp Boolean to indicate if a new SDP should be created
 * \param media_state Optional requested media state for the SDP
 *
 * \retval 0 Successfully sent refresh
 * \retval -1 Failure to send refresh
 *
 * \note If a media_state is passed in ownership will be taken in all cases
 */
int ast_sip_session_refresh(struct ast_sip_session *session,
		ast_sip_session_request_creation_cb on_request_creation,
		ast_sip_session_sdp_creation_cb on_sdp_creation,
		ast_sip_session_response_cb on_response,
		enum ast_sip_session_refresh_method method,
		int generate_new_sdp,
		struct ast_sip_session_media_state *media_state);

/*!
 * \brief Regenerate SDP Answer
 *
 * This method is used when an SDP offer has been received but an SDP answer
 * has not been sent yet. It requests that a new local SDP be created and
 * set as the SDP answer. As with any outgoing request in res_pjsip_session,
 * this will call into registered supplements in case they wish to add anything.
 *
 * \param session The session on which the answer will be updated
 * \param on_sdp_creation Callback called when SDP is created
 * \param generate_new_sdp Boolean to indicate if a new SDP should be created
 * \retval 0 Successfully updated the SDP answer
 * \retval -1 Failure to updated the SDP answer
 */
int ast_sip_session_regenerate_answer(struct ast_sip_session *session,
		ast_sip_session_sdp_creation_cb on_sdp_creation);

/*!
 * \brief Send a SIP response
 *
 * This will send the SIP response specified in tdata and
 * call into any registered supplements' outgoing_response callback.
 *
 * \param session The session on which to send the response.
 * \param tdata The response to send
 */
void ast_sip_session_send_response(struct ast_sip_session *session, pjsip_tx_data *tdata);

/*!
 * \brief Send a SIP request
 *
 * This will send the SIP request specified in tdata and
 * call into any registered supplements' outgoing_request callback.
 *
 * \param session The session to which to send the request
 * \param tdata The request to send
 */
void ast_sip_session_send_request(struct ast_sip_session *session, pjsip_tx_data *tdata);

/*!
 * \brief Creates an INVITE request.
 *
 * \param session Starting session for the INVITE
 * \param tdata The created request.
 */
int ast_sip_session_create_invite(struct ast_sip_session *session, pjsip_tx_data **tdata);

/*!
 * \brief Send a SIP request and get called back when a response is received
 *
 * This will send the request out exactly the same as ast_sip_send_request() does.
 * The difference is that when a response arrives, the specified callback will be
 * called into
 *
 * \param session The session on which to send the request
 * \param tdata The request to send
 * \param on_response Callback to be called when a response is received
 */
void ast_sip_session_send_request_with_cb(struct ast_sip_session *session, pjsip_tx_data *tdata,
		ast_sip_session_response_cb on_response);

/*!
 * \brief Retrieves a session from a dialog
 *
 * \param dlg The dialog to retrieve the session from
 *
 * \retval non-NULL if session exists
 * \retval NULL if no session
 *
 * \note The reference count of the session is increased when returned
 *
 * \note This function *must* be called with the dialog locked
 */
struct ast_sip_session *ast_sip_dialog_get_session(pjsip_dialog *dlg);

/*!
 * \brief Resumes processing of a deferred incoming re-invite
 *
 * \param session The session which has a pending incoming re-invite
 *
 * \note When resuming a re-invite it is given to the pjsip stack as if it
 *       had just been received from a transport, this means that the deferral
 *       callback will be called again.
 */
void ast_sip_session_resume_reinvite(struct ast_sip_session *session);

/*!
 * \brief Determines if a provided pending stream will be the default stream or not
 * \since 15.0.0
 *
 * \param session The session to check against
 * \param stream The pending stream
 *
 * \retval 1 if stream will be default
 * \retval 0 if stream will NOT be the default
 */
int ast_sip_session_is_pending_stream_default(const struct ast_sip_session *session, const struct ast_stream *stream);

/*!
 * \brief Allocate a session media state structure
 * \since 15.0.0
 *
 * \retval non-NULL success
 * \retval NULL failure
 */
struct ast_sip_session_media_state *ast_sip_session_media_state_alloc(void);

/*!
 * \brief Allocate an ast_session_media and add it to the media state's vector.
 * \since 15.0.0
 *
 * This allocates a session media of the specified type. The position argument
 * determines where in the vector that the new session media will be inserted.
 *
 * \note The returned ast_session_media is the reference held by the vector. Callers
 * of this function must NOT decrement the refcount of the session media.
 *
 * \param session Session on which to query active media state for
 * \param media_state Media state to place the session media into
 * \param type The type of the session media
 * \param position Position at which to insert the new session media.
 *
 * \note The active media state will be queried and if a media session already
 * exists at the given position for the same type it will be reused instead of
 * allocating a new one.
 *
 * \retval non-NULL success
 * \retval NULL failure
 */
struct ast_sip_session_media *ast_sip_session_media_state_add(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state, enum ast_media_type type, int position);

/*!
 * \brief Reset a media state to a clean state
 * \since 15.0.0
 *
 * \param media_state The media state to reset
 */
void ast_sip_session_media_state_reset(struct ast_sip_session_media_state *media_state);

/*!
 * \brief Clone a media state
 * \since 15.0.0
 *
 * \param media_state The media state to clone
 *
 * \retval non-NULL success
 * \retval NULL failure
 */
struct ast_sip_session_media_state *ast_sip_session_media_state_clone(const struct ast_sip_session_media_state *media_state);

/*!
 * \brief Free a session media state structure
 * \since 15.0.0
 */
void ast_sip_session_media_state_free(struct ast_sip_session_media_state *media_state);

/*!
 * \brief Set a read callback for a media session with a specific file descriptor
 * \since 15.0.0
 *
 * \param session The session
 * \param session_media The media session
 * \param fd The file descriptor
 * \param callback The read callback
 *
 * \retval 0 the read callback was successfully added
 * \retval -1 the read callback could not be added
 *
 * \note This operations on the pending media state
 */
int ast_sip_session_media_add_read_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	int fd, ast_sip_session_media_read_cb callback);

/*!
 * \brief Set a write callback for a media session
 * \since 15.0.0
 *
 * \param session The session
 * \param session_media The media session
 * \param callback The write callback
 *
 * \retval 0 the write callback was successfully add
 * \retval -1 the write callback is already set to something different
 *
 * \note This operates on the pending media state
 */
int ast_sip_session_media_set_write_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	ast_sip_session_media_write_cb callback);

/*!
 * \brief Retrieve the underlying media session that is acting as transport for a media session
 * \since 15.0.0
 *
 * \param session The session
 * \param session_media The media session to retrieve the transport for
 *
 * \note This operates on the pending media state
 *
 * \note This function is guaranteed to return non-NULL
 */
struct ast_sip_session_media *ast_sip_session_media_get_transport(struct ast_sip_session *session, struct ast_sip_session_media *session_media);


#ifdef GRANDSTREAM_NETWORKS

enum ast_sip_supplement_type {
	AST_SIP_EXTEN_SUPPLEMENT_NONE = -1,
	AST_SIP_CCSS_ORIGINAL_EXTEN = 0,
	AST_SIP_CCSS_NOTIFY_USER    = 1,
};

struct ast_sip_cc_original_exten_data {
	int  core_id;
	char original_exten[AST_MAX_EXTENSION]; 
	char device_name[AST_MAX_EXTENSION];	
};

struct ast_sip_cc_notify_user_data {
	int    core_id;
	struct ast_cc_monitor *monitor; 
	char   notify_user[AST_MAX_EXTENSION];
};

struct ast_sip_exten_supplement {
	enum ast_sip_supplement_type  exten_type; 
	int (*find_exten)(char *exten, void *data);
	AST_LIST_ENTRY(ast_sip_exten_supplement) next;
};

/* Added by yfyang, for sca feature */
enum sca_status {
	SCA_Idle = 0,    /* Available*/
	SCA_Seized,      /* The line has been seized by a sca user */
	SCA_Progressing, /* There is a call in progressing on the line */
	SCA_Alerting,    /* There is an incoming call on the line */
	SCA_Active,      /* There is a call established on the line */
	SCA_Held,        /* The call has been held by a sca user on the line */
	SCA_Held_Private,/* The call has been private held by a sca user on the line */
	SCA_Bridge_Active, /* A sca user has barged in an active line */
	SCA_Bridge_Held,   /* The bridge call has been held by a sca user */
	SCA_BUTT
};

struct ast_sip_session_sca_handler {
	/* To get the status of the sca line */
	enum sca_status (*sca_line_status_get)(char *sharedline, int line_index, char *local_info);

	/* To set the status of the sca line */
	int (*sca_line_status_set)(char *sharedline, int line_index, enum sca_status status);

	/* To get the line index according to the status */
	unsigned int (*sca_line_index_get)(char *sharedline, enum sca_status status, char *local_info);

	/* To alloc a sca line */
	int (*sca_line_alloc)(char *sharedline, char *local_info, struct ast_channel *remote_chan);

	/* To get the service of the shared line */
	int (*sca_line_service_get)(char *sharedline, int *retrv_remote, int *barge);

	/* To remove a user from the shared line */
	int (*sca_line_remove_user)(struct ast_channel *chan, char *sharedline, int line_index, char *user_name);

	/* To set the user info of sca call */
	int (*sca_line_set_user_info)(char *sharedline, int line_index, char *local_user, char *remote_user);

	/* To get the private users of the shared line */
	int (*sca_line_get_private_users)(char *sharedline, char *user_buf, int len, int paging_flag);

	/* To get the shard line status: inuse or not */
	int (*sca_user_check_status)(char *sharedline);

	/* To check and insert user to line */
	int (*sca_line_insert_user)(char *sharedline, int line_index, char *local_user);

	/* To check if the line is in a sca bridge, return 0: not in bridge, 1: in bridge */
	int (*sca_line_check_in_bridge)(char *sharedline, int line_index);

	AST_LIST_ENTRY(ast_sip_session_sca_handler) next;
};

const char *ast_sip_sca_state2str(enum sca_status state);

enum sca_status ast_sip_get_sca_line_status(char *sharedline, int line_index);
int ast_sip_set_sca_line_status(char *sharedline, int line_index, enum sca_status status);
int ast_sip_get_sca_call_info_from_rdata(pjsip_rx_data *rdata, unsigned int *index, enum sca_status *status);
unsigned int ast_sip_get_sca_line_index(char *sharedline, enum sca_status status, char *local_info);
int ast_sip_sca_line_remove_user(struct ast_channel *chan, char *sharedline, int line_index, char *user_name);
int ast_sip_set_sca_user_info(char *sharedline, int line_index, char *local_user, char *remote_user);
int ast_sip_get_sca_private_users(char *sharedline, char *user_buf, int len, int paging_flag);
int ast_sip_check_sca_user_status(char *sharedline);
int ast_sip_sca_line_insert_user(char *sharedline, int line_index, char *local_user);
unsigned int ast_sip_sca_line_alloc(char *sharedline, char *local_info, struct ast_channel *remote_chan);
int ast_sip_sca_line_check_in_bridge(char *sharedline, int line_index);

int ast_sip_session_register_sca_handler(struct ast_sip_session_sca_handler *sca_handler);
void ast_sip_session_unregister_sca_handler(struct ast_sip_session_sca_handler *sca_handler);

int ast_sip_find_sub_handler_for_supplement_exten(enum ast_sip_supplement_type type, char *exten, void *data);
int ast_sip_exten_register_supplement(struct ast_sip_exten_supplement *supplement);
void ast_sip_exten_unregister_supplement(struct ast_sip_exten_supplement *supplement);

/* Add by bfdu for reinivite */
struct ast_sip_reinvite_data {
	struct ast_sip_session *session;
};

enum ast_sip_application_type {
	AST_SIP_APP_UDP_BFCP = 1,
	AST_SIP_APP_TCP_BFCP,
	AST_SIP_APP_RTP,
	AST_SIP_APP_INVALID,
};

struct ast_sip_recovery_entry{
	char exten[AST_MAX_EXTENSION]; /* caller number */
	char caller_name[128];/* caller_id_name */
	char linkid[AST_MAX_UNIQUEID];
	char uuid[AST_MAX_UNIQUEID];
	char connected_line_id[AST_MAX_EXTENSION];
	char conf_room[32];
	char orig_options[32];/* for conference */
	char peertrunk_aor[64];
	int dahdi_channel; /* for restore call on dahdi same channel */
	int create_sec;  /* for cdr recover */
	int create_usec;
	int answer_sec;
	int answer_usec;
	int cseq;
	char *from_tag;
	char *to_tag;
	char *call_id;
	char *remote_ip;
	char direction; /* inbound or outbound */
	char *data;/* reserved */
	char inbound_trunk_index[AST_MAX_EXTENSION];
	struct ast_channel_media_encryption encryption;
	char *sca_answer_user;
	char delay_foripv6;
	char *session_conf_link;
	char *session_trunk_channel_id;
};

unsigned char ast_sip_session_judge_if_type_is_sip(const struct ast_channel* chan);
unsigned char ast_sip_session_judge_if_type_is_local(const struct ast_channel* chan);
unsigned char ast_sip_session_judge_if_type_is_rtsp(const struct ast_channel* chan);
unsigned char judge_if_support_video_media(struct ast_sip_session *session, const pjmedia_sdp_session *offer);
int ast_sip_session_judge_if_main_video(const pjmedia_sdp_media *offer_media);
int ast_sip_session_judge_if_slides_video(const pjmedia_sdp_media *offer_media);
int ast_sip_session_judge_if_bfcp_stream(const pjmedia_sdp_media *stream);
int ast_sip_session_judge_if_fec_stream(const pjmedia_sdp_media *stream);
int ast_sip_session_judge_if_fecc_stream(const pjmedia_sdp_media *stream);
int ast_sip_session_judge_if_passthrough_attr(char *attr_value);
int ast_sip_session_judge_if_passthrough_bandw(char *bandw_name);
int ast_sip_session_judge_if_need_passthrough(void);
void ast_sip_session_change_gs_fec_version(struct ast_format_cap *caps);
enum ast_sip_application_type ast_sip_session_get_application_type(pj_str_t transport);
enum ast_sip_session_media_encryption ast_sip_session_get_media_encryption_type(pj_str_t transport,
	const struct pjmedia_sdp_media *stream, unsigned int *optimistic);
enum ast_sip_session_media_encryption ast_sip_session_check_endpoint_media_transport(struct ast_sip_endpoint *endpoint,
	const struct pjmedia_sdp_media *stream, unsigned int is_webrtc_call);
void ast_sip_session_set_media_sub_type(struct ast_sip_session *session, const pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream);
void ast_sip_session_modify_outgoing_yealink_sdp(const struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media);
void ast_sip_session_generate_main_video_ssrc_attrs(pj_pool_t *pool, pjmedia_sdp_media *media,
	const struct ast_sip_session *session, unsigned int ssrc, unsigned int rtx_ssrc);
void ast_sip_session_generate_slides_video_ssrc_attrs(pj_pool_t *pool, pjmedia_sdp_media *media,
	const struct ast_sip_session *session, unsigned int ssrc, unsigned int rtx_ssrc);
void ast_sip_session_change_ssrc_attrs(pj_pool_t *pool, const struct ast_sip_session *session,
	pjmedia_sdp_media **p_result, int is_main_video);
void ast_sip_session_generate_rtx_ssrc_attrs(pj_pool_t *pool, pjmedia_sdp_media *media,
	struct ast_rtp_instance *rtp, const struct ast_sip_session *session);
unsigned int ast_sip_session_get_allow_forward_hold_request(void);
enum ast_media_sub_type ast_sip_session_get_media_sub_type(struct ast_sip_session *session, const pjmedia_sdp_media *stream);
struct ast_sip_session_media *ast_sip_session_media_state_add_as_sub_type_and_msid(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state, struct ast_stream *stream, int position);
void ast_sip_session_init_stream_media_direction(struct ast_sip_session *session);
void ast_sip_set_used_gs_call_count(struct ast_sip_session *session,pjsip_rx_data *rdata,int flag);//added by zyxia for ios gswave wakeup
void ast_sip_session_bfcp_data_destroy(void *obj);
struct bfcp_info_data *ast_sip_session_bfcp_data_alloc(struct ast_sip_session *session, const struct ast_bfcp_info *bfcp_info);
const char *ast_sip_session_bfcp_info_reqstatus_2str(enum bfcp_reqstatus status);
const char *ast_sip_session_bfcp_info_prim_2str(enum bfcp_prim  prim);
int ast_sip_session_bfcp_info_generate_body_content(const struct ast_bfcp_info *bfcp_info, char *content, int len);
int ast_sip_session_bfcp_send_conference_sip_info(void *data);
int ast_sip_session_bfcp_send_sip_info(void *data);
char *ast_sip_session_get_video_resolution_according_type(struct ast_sip_session *session, enum video_resolution resolution);
char *ast_sip_session_get_imageattr_according_type(enum video_resolution resolution);
char *ast_sip_session_get_wave_imageattr_according_type(enum video_resolution resolution);
int ast_sip_session_get_as_according_type(struct ast_sip_session *session, enum video_resolution resolution);
int ast_sip_session_get_tias_according_type(struct ast_sip_session *session, enum video_resolution resolution);
void ast_sip_session_set_as_tias_for_sfu(struct ast_sip_session *session, struct ast_stream *stream);
void ast_sip_session_set_imageattr_for_stream(struct ast_sip_session *session,
	struct ast_stream *stream, enum video_resolution resolution);
struct ast_format *ast_sip_session_get_video_format(struct ast_stream *stream, const char *codec_name);
void ast_sip_session_set_all_stream_change_ssrc(struct ast_sip_session *session);
struct ast_stream *ast_sip_session_get_pending_main_video_stream(struct ast_sip_session *session);
struct ast_stream *ast_sip_session_get_active_main_video_stream(struct ast_sip_session *session);
struct ast_stream *ast_sip_session_get_pending_slides_video_stream(struct ast_sip_session *session);
struct ast_stream *ast_sip_session_get_active_slides_video_stream(struct ast_sip_session *session);
enum video_resolution ast_sip_session_get_main_video_resource(struct ast_sip_session *session, unsigned int h264_level);
unsigned int ast_sip_session_get_h264_level(struct ast_sip_session *session, unsigned int is_nway);
void ast_sip_session_set_conf_video_codec(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution);
int ast_sip_session_modify_video_codec_order_for_sfu(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution);
unsigned int ast_sip_session_judge_if_has_valid_slides_video(struct ast_sip_session *session);
struct ast_sip_session_media *ast_sip_session_get_bfcp_active_media(struct ast_sip_session *session);
struct ast_sip_session_media *ast_sip_session_get_bfcp_pending_media(struct ast_sip_session *session);
const char *ast_sip_session_media_encryption_type2str(enum ast_sip_session_media_encryption encryption);
int ast_sip_get_extension_from_channel_name(const char *channel_name, char *buf, int size);
void ast_sip_session_notify_sfu_bfcp_for_join_conference(struct ast_sip_session *session);
int ast_sip_session_get_format_useinbandfec(struct ast_sip_session *session, struct ast_stream *stream, struct ast_format_cap *caps);
void ast_sip_session_set_format_useinbandfec(struct ast_sip_session *session, struct ast_format_cap *caps, struct ast_stream *stream, int useinbandfec);
unsigned int ast_sip_session_get_local_negotiate_prefer(void);
void ast_sip_session_change_video_independent_codec(struct ast_sip_session *session,
	struct ast_format_cap *joint_cap, struct ast_stream *clone_stream);
unsigned int ast_sip_session_need_create_bfcp_media(struct ast_sip_session *session);
void ast_sip_session_append_clone_stream(struct ast_sip_session *session, struct ast_stream *req_stream);
unsigned int ast_sip_session_process_special_stream(struct ast_sip_session *session, struct ast_stream *req_stream);
int ast_sip_session_get_joint_caps(struct ast_sip_session *session, struct ast_stream *req_stream,
	struct ast_format_cap *joint_cap, struct ast_format_cap *req_cap);
void ast_sip_session_process_special_audio_codec(struct ast_sip_session *session,
	struct ast_stream *req_stream, struct ast_format_cap *joint_cap);
int ast_sip_session_process_special_video_codec(struct ast_sip_session *session, struct ast_format_cap *joint_cap);
unsigned char ast_sip_session_support_multiple_video(const struct ast_sip_session *session);
unsigned char ast_sip_session_support_bfcp(const struct ast_sip_session *session);
const char *ast_sip_session_sfu_codec2str(enum ast_sfu_video_codec codec);
struct ast_sip_session_media *ast_sip_session_get_host_main_video_media(struct ast_channel *channel, const char *conf_room);

int ast_sip_session_set_device_ua(struct ast_sip_session* session, struct pjsip_rx_data* rdata, pjsip_method_e method);
int ast_sip_session_get_device_ua(struct ast_sip_session* session, char* buf, uint64_t size);
struct ast_sip_session_media *ast_sip_session_get_active_session_media(struct ast_sip_session *session,
	enum ast_media_sub_type sub_type);
struct ast_sip_session_media *ast_sip_session_get_pending_session_media(struct ast_sip_session *session,
	enum ast_media_sub_type sub_type);
struct ast_sip_session_media *ast_sip_session_get_session_media(struct ast_sip_session *session,
	enum ast_media_sub_type sub_type);
int ast_sip_session_process_special_stream_for_reinvite(struct ast_sip_session* session, struct ast_stream* req_stream);
int ast_sip_session_refresh_participant_for_sfu(struct ast_sip_session *session,
	int generate_new_sdp, ast_sip_session_response_cb on_response, struct ast_sip_session_media_state *refresh_media_state);
int ast_sip_session_participant_topology_change_response(struct ast_sip_session *session, pjsip_rx_data *rdata);
void ast_sip_session_add_missed_line_for_gswave(struct ast_sip_session *session, struct ast_channel *requestor);
int ast_sip_session_print_sdp(const pjmedia_sdp_session *ses, char *buf, pj_ssize_t len);
struct ast_sip_session_media *ast_sip_session_get_media_for_participant(struct ast_sip_session *session, struct ast_sip_session_media *session_media);
unsigned int ast_sip_session_media_use_bundle(const struct ast_sip_session *session);
int ast_sip_session_translate_gswave_to_sip_behavior(struct ast_sip_session *session,
	struct ast_stream *req_stream, struct ast_channel *requestor);
void ast_sip_session_change_rtx_params(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, char *rtx_video_codec_name, pj_str_t video_pt_pj_str);
void ast_sip_session_set_video_format(struct ast_sip_session *session, struct ast_stream *stream, int resolution);
unsigned int ast_sip_session_get_enable_p2p_video_control(void);
int ast_sip_session_get_p2p_video_resolution(void);
void ast_sip_update_dlg_and_invite_tsx_transport_by_endponit(struct ast_sip_endpoint *endpoint,pj_sockaddr *rem_addr, pjsip_transport *tp);
void ast_sip_notify_avs_to_make_early_bridge(struct ast_sip_session *session);
unsigned char ast_sip_session_judge_if_disable_media(pjmedia_sdp_media *media);
void ast_sip_session_remove_attrs_for_disable_media(pjmedia_sdp_media *media);
void ast_sip_session_set_format_dtx(struct ast_sip_session *session,
	struct ast_format_cap *caps, struct ast_stream *stream);
void ast_sip_session_remove_media_from_bundle_group(struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_session *sdp, pjmedia_sdp_media *media);
void ast_sip_session_set_bfcp_info_msid(struct ast_sip_session *session,
	struct ast_bfcp_info *bfcp_info);
unsigned int ast_sip_session_get_only_support_audio(void);
unsigned int ast_sip_session_jugde_if_media_support(struct ast_sip_session *session, enum ast_media_sub_type sub_type);
int ast_sip_record_hdr_from_msg(struct ast_channel *chan, void *msg, void *xml_cdr);
void ast_sip_session_disable_stream(struct ast_sip_session *session, struct ast_stream *stream);
void ast_sip_session_cancel_pending_call(struct ast_sip_session *session);
void ast_sip_session_cancel_pending_call_according_id(const char *conf_id, const char *id);
void ast_sip_session_set_main_video_stream_uptime(struct ast_sip_session *session);
void ast_sip_session_send_session_to_peer(struct ast_sip_session *session);
int ast_sip_set_contact_user(pjsip_tx_data *packet, const  char *user);
unsigned int ast_sip_session_judge_if_need_add_rpid(const struct ast_sip_session *session);
struct ast_stream *ast_sip_session_get_pending_stream_by_subtype(struct ast_sip_session * session, enum ast_media_sub_type sub_type);
int ast_sip_session_add_support_header_for_whiteboard(struct ast_sip_session *session, pjsip_tx_data *tdata);
void ast_sip_session_append_stream_to_topology(struct ast_sip_session *session,
	struct ast_stream_topology *topology, struct ast_stream *stream, enum ast_media_sub_type type);
int ast_sip_session_if_need_send_bfcp_info(struct ast_sip_session *session, char *share_type);
int ast_sip_session_send_notify_to_update_gs_state(struct ast_sip_session *session,
	enum ast_sip_session_gs_state_type type, void *data);
void ast_sip_session_send_avs_mcu_control(struct ast_sip_session *session, int type);
void ast_sip_session_set_ucm_trunk(struct ast_sip_session *session, pjsip_rx_data *rdata);
int ast_sip_session_is_nway_conference_number(const char *exten);
void ast_sip_session_set_trunk_info(struct ast_sip_session *session, pjsip_rx_data *rdata);

int ast_sip_get_diversion_user(pjsip_rx_data *rdata, char *user, int len);
int ast_sip_session_set_conf_link_from_db(struct ast_sip_session* session);
void ast_sip_session_notfiy_trunk_info(struct ast_sip_session* session);
int ast_sip_session_cancel_invite_call(struct ast_sip_session *session, const struct ast_channel *chan,const char* dest);
void ast_sip_session_set_media_transport(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media);
int ast_sip_session_send_info_for_add_candidates(void *data);
void ast_sip_session_add_decline_video_reason(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, struct ast_stream *stream);
int ast_sip_session_trunk_conf_need_subscribe(struct ast_sip_session *session,struct pjsip_rx_data *rdata);
void ast_sip_session_cancel_conf_invitee(struct ast_sip_session *session);
void ast_sip_session_notfiy_answered_status(struct ast_sip_session* session);
int ast_sip_session_is_have_video_codec(struct ast_sip_session *session);
int ast_sip_session_add_support_header_for_gs_state(struct ast_sip_session *session, pjsip_tx_data *tdata);
void ast_sip_session_add_support_video_call_for_webrtc_trunk(struct ast_sip_session *session, pjsip_tx_data *tdata);
void ast_sip_session_send_reinvite_response(struct ast_sip_session *session);
void ast_sip_session_call_to_emergency_call(struct ast_sip_session *session, pjsip_rx_data *rdata);
void ast_sip_session_set_call_flip(struct ast_sip_session *session, pjsip_rx_data *rdata);
int ast_sip_session_trunk_record_need_forward(struct ast_sip_session *session,struct pjsip_rx_data *rdata);
void ast_sip_session_set_avs_address(struct ast_sip_session *session);
void ast_sip_session_set_gs_client_info(pjsip_rx_data *rdata, struct ast_sip_session *session, unsigned int is_caller);
int ast_sip_session_delay_request(struct ast_sip_session *session);

int ast_sip_get_vlan_flag(void);
const char *ast_sip_get_vlan_addr(void);
const char *ast_sip_get_lan_addr(void);
int ast_sip_session_media_stream_has_zrtphash(const struct pjmedia_sdp_media *stream);
int  ast_sip_session_media_stream_has_srtp(pj_str_t transport,const struct pjmedia_sdp_media *stream);


#endif   /* GRANDSTREAM_NETWORKS */

#endif /* _RES_PJSIP_SESSION_H */
