/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2017, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 * \brief Media Stream API
 *
 * \author Joshua Colp <jcolp@digium.com>
 */

#ifndef _AST_STREAM_H_
#define _AST_STREAM_H_

#include "asterisk/codec.h"
#include "asterisk/vector.h"

/*!
 * \brief Forward declaration for a stream, as it is opaque
 */
struct ast_stream;

/*!
 * \brief Forward declaration for a format capability
 */
struct ast_format_cap;

/*!
 * \brief The topology of a set of streams
 */
struct ast_stream_topology;

/*!
 * \brief A mapping of two topologies.
 */
struct ast_stream_topology_map;

typedef void (*ast_stream_data_free_fn)(void *);

/*!
 * \brief States that a stream may be in
 */
enum ast_stream_state {
	/*!
	 * \brief Set when the stream has been removed/declined
	 */
	AST_STREAM_STATE_REMOVED = 0,
	/*!
	 * \brief Set when the stream is sending and receiving media
	 */
	AST_STREAM_STATE_SENDRECV,
	/*!
	 * \brief Set when the stream is sending media only
	 */
	AST_STREAM_STATE_SENDONLY,
	/*!
	 * \brief Set when the stream is receiving media only
	 */
	AST_STREAM_STATE_RECVONLY,
	/*!
	 * \brief Set when the stream is not sending OR receiving media
	 */
	AST_STREAM_STATE_INACTIVE,
};

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for passthrough, 2019.05.10 */
enum ast_passthrough_type {
	AST_PASSTHROUGH_TYPE_ATTR,
	AST_PASSTHROUGH_TYPE_BANDW,
};

enum ast_stream_media_direction {
	UNKNOW,
	SENDRECV,   /*!< The offer with a=sendrecv */
	SENDONLY,   /*!< The offer with a=sendonly */
	INACTIVE,   /*!< The offer with a=inactive */
	RECVONLY,   /*!< The answer with a=recvonly */
};

enum ast_sfu_invite_type {
	AST_SFU_INVITE_TYPE_AUDIO,
	AST_SFU_INVITE_TYPE_VIDEO,
};

enum ast_stream_video_resolution {
	AST_STREAM_VIDEO_RESOLUTION_NONE,
	AST_STREAM_VIDEO_RESOLUTION_360P,
	AST_STREAM_VIDEO_RESOLUTION_720P,
	AST_STREAM_VIDEO_RESOLUTION_1080P,
};

enum ast_video_decline_reason
{
	AST_VIDEO_DECLINE_REASON_NO_VIDEO_CODEC       = 1,
	AST_VIDEO_DECLINE_REASON_NO_VIDEO_RESOURCE    = 2,
	AST_VIDEO_DECLINE_REASON_CPU_LIMIT            = 3,
	AST_VIDEO_DECLINE_REASON_VERSION_LIMIT        = 4,
	AST_VIDEO_DECLINE_REASON_NO_VIDEO_ROOM        = 5,
	AST_VIDEO_DECLINE_REASON_PEER_REJECT          = 6,
};

/* This struct must be according ast_sip_media_encryption */
enum ast_stream_media_encryption {
	AST_STREAM_MEDIA_TRANSPORT_INVALID = 0,
	AST_STREAM_MEDIA_ENCRYPT_NONE,
	AST_STREAM_MEDIA_ENCRYPT_SDES,
	AST_STREAM_MEDIA_ENCRYPT_DTLS,
	AST_STREAM_MEDIA_AUTO_NO_ENCRYPT,
	AST_STREAM_MEDIA_AUTO_ENCRYPT_SRTP,
	AST_STREAM_MEDIA_AUTO_ENCRYPT_DTLS,
};

enum ast_stream_same_codec_type {
	AST_STREAM_SAME_CODEC_TYPE_INVALID = 0,
	AST_STREAM_SAME_CODEC_TYPE_H264,
	AST_STREAM_SAME_CODEC_TYPE_RTX,
};

#define MAX_STREAM_FMTP_PARAM 256

struct ast_same_codec_info {
	int payload;
	char fmtp_param[MAX_STREAM_FMTP_PARAM];
};
#endif

/*!
 * \brief Create a new media stream representation
 *
 * \param name A name for the stream
 * \param type The media type the stream is handling
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \note This is NOT an AO2 object and has no locking. It is expected that a higher level object provides protection.
 *
 * \note The stream will default to an inactive state until changed.
 *
 * \since 15
 */
struct ast_stream *ast_stream_alloc(const char *name, enum ast_media_type type);

/*!
 * \brief Destroy a media stream representation
 *
 * \param stream The media stream
 *
 * \since 15
 */
void ast_stream_free(struct ast_stream *stream);

/*!
 * \brief Create a deep clone of an existing stream
 *
 * \param stream The existing stream
 * \param Optional name for cloned stream. If NULL, then existing stream's name is copied.
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \note Opaque data pointers set with ast_stream_set_data() are not part
 * of the deep clone.  We have no way to clone the data.
 *
 * \since 15
 */
struct ast_stream *ast_stream_clone(const struct ast_stream *stream, const char *name);

/*!
 * \brief Get the name of a stream
 *
 * \param stream The media stream
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \since 15
 */
const char *ast_stream_get_name(const struct ast_stream *stream);

/*!
 * \brief Get the media type of a stream
 *
 * \param stream The media stream
 *
 * \return The media type of the stream (AST_MEDIA_TYPE_UNKNOWN on error)
 *
 * \since 15
 */
enum ast_media_type ast_stream_get_type(const struct ast_stream *stream);

/*!
 * \brief Change the media type of a stream
 *
 * \param stream The media stream
 * \param type The new media type
 *
 * \since 15
 */
void ast_stream_set_type(struct ast_stream *stream, enum ast_media_type type);

/*!
 * \brief Get the current negotiated formats of a stream
 *
 * \param stream The media stream
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \note The reference count is not increased
 *
 * \since 15
 */
struct ast_format_cap *ast_stream_get_formats(const struct ast_stream *stream);

/*!
 * \brief Set the current negotiated formats of a stream
 *
 * \param stream The media stream
 * \param caps The current negotiated formats
 *
 * \note The new format capabilities structure has its refcount bumped and
 * any existing format capabilities structure has its refcount decremented.
 *
 * \since 15
 */
void ast_stream_set_formats(struct ast_stream *stream, struct ast_format_cap *caps);

/*!
 * \brief Get the current state of a stream
 *
 * \param stream The media stream
 *
 * \return The state of the stream (AST_STREAM_STATE_UNKNOWN on error)
 *
 * \since 15
 */
enum ast_stream_state ast_stream_get_state(const struct ast_stream *stream);

/*!
 * \brief Set the state of a stream
 *
 * \param stream The media stream
 * \param state The new state that the stream is in
 *
 * \note Used by stream creator to update internal state
 *
 * \since 15
 */
void ast_stream_set_state(struct ast_stream *stream, enum ast_stream_state state);

/*!
 * \brief Convert the state of a stream into a string
 *
 * \param state The stream state
 *
 * \return The state of the stream in string format
 *
 * \since 15
 */
const char *ast_stream_state2str(enum ast_stream_state state);

/*!
 * \brief Convert a string to a stream state
 *
 * \param str The string to convert
 *
 * \return The stream state
 *
 * \since 15.0.0
 */
enum ast_stream_state ast_stream_str2state(const char *str);

/*!
 * \brief Get a stream metadata value
 *
 * \param stream The media stream
 * \param m_key  An arbitrary metadata key
 *
 * \retval non-NULL metadata value
 * \retval NULL failure or not found
 *
 * \since 15.5
 */
const char *ast_stream_get_metadata(const struct ast_stream *stream,
	const char *m_key);

/*!
 * \brief Get all stream metadata keys
 *
 * \param stream The media stream
 *
 * \retval An ast_variable list of the metadata key/value pairs.
 * \retval NULL if error or no variables are set.
 *
 * When you're finished with the list, you must call
 *  ast_variables_destroy(list);
 *
 * \since 15.5
 */
struct ast_variable *ast_stream_get_metadata_list(const struct ast_stream *stream);

/*!
 * \brief Set a stream metadata value
 *
 * \param stream The media stream
 * \param m_key  An arbitrary metadata key
 * \param value  String metadata value or NULL to remove existing value
 *
 * \retval -1 failure
 * \retval 0  success
 *
 * \since 15.5
 */
int ast_stream_set_metadata(struct ast_stream *stream, const char *m_key, const char *value);

/*!
 * \brief Get the position of the stream in the topology
 *
 * \param stream The media stream
 *
 * \return The position of the stream (-1 on error)
 *
 * \since 15
 */
int ast_stream_get_position(const struct ast_stream *stream);

/*!
 * \brief Get rtp_codecs associated with the stream
 *
 * \param stream The media stream
 *
 * \return The rtp_codecs
 *
 * \since 15.5
 */
struct ast_rtp_codecs *ast_stream_get_rtp_codecs(const struct ast_stream *stream);

/*!
 * \brief Set rtp_codecs associated with the stream
 *
 * \param stream The media stream
 * \param rtp_codecs The rtp_codecs
 *
 * \since 15.5
 */
void ast_stream_set_rtp_codecs(struct ast_stream *stream, struct ast_rtp_codecs *rtp_codecs);

/*!
 * \brief Create a stream topology
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \since 15
 *
 * \note This returns an ao2 refcounted object
 */
struct ast_stream_topology *ast_stream_topology_alloc(void);

/*!
 * \brief Create a deep clone of an existing stream topology
 *
 * \param topology The existing topology of streams
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \since 15
 *
 * \note This returns an ao2 refcounted object
 */
struct ast_stream_topology *ast_stream_topology_clone(
	const struct ast_stream_topology *topology);

/*!
 * \brief Compare two stream topologies to see if they are equal
 *
 * \param left The left topology
 * \param right The right topology
 *
 * \retval 1 topologies are equivalent
 * \retval 0 topologies differ
 *
 * \since 15
 */
int ast_stream_topology_equal(const struct ast_stream_topology *left,
	const struct ast_stream_topology *right);

/*!
 * \brief Unreference and destroy a stream topology
 *
 * \param topology The topology of streams
 *
 * \note All streams contained within the topology will be destroyed
 *
 * \since 15
 */
void ast_stream_topology_free(struct ast_stream_topology *topology);

/*!
 * \brief Append a stream to the topology
 *
 * \param topology The topology of streams
 * \param stream The stream to append
 *
 * \returns the position of the stream in the topology (-1 on error)
 *
 * \since 15
 */
int ast_stream_topology_append_stream(struct ast_stream_topology *topology,
	struct ast_stream *stream);

/*!
 * \brief Get the number of streams in a topology
 *
 * \param topology The topology of streams
 *
 * \return the number of streams (-1 on error)
 *
 * \since 15
 */
int ast_stream_topology_get_count(const struct ast_stream_topology *topology);

/*!
 * \brief Get a specific stream from the topology
 *
 * \param topology The topology of streams
 * \param position The topology position to get
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \since 15
 */
struct ast_stream *ast_stream_topology_get_stream(
	const struct ast_stream_topology *topology, unsigned int position);

/*!
 * \brief Set a specific position in a topology
 *
 * \param topology The topology of streams
 * \param position The topology position to set
 * \param stream The stream to put in its place
 *
 * \retval 0 success
 * \retval -1 failure
 *
 * \note If an existing stream exists it will be destroyed
 *
 * \note You can overwrite an existing position in the topology or set
 * the first unused position.  You can't set positions beyond that.
 *
 * \since 15
 */
int ast_stream_topology_set_stream(struct ast_stream_topology *topology,
	unsigned int position, struct ast_stream *stream);

/*!
 * \brief Delete a specified stream from the given topology.
 * \since 15.0.0
 *
 * \param topology The topology of streams.
 * \param position The topology position to delete.
 *
 * \note Deleting a stream will completely remove it from the topology
 * as if it never existed in it.  i.e., Any following stream positions
 * will shift down so there is no gap.
 *
 * \retval 0 on success.
 * \retval -1 on failure.
 *
 * \return Nothing
 */
int ast_stream_topology_del_stream(struct ast_stream_topology *topology,
	unsigned int position);

/*!
 * \brief A helper function that, given a format capabilities structure,
 * creates a topology and separates the media types in format_cap into
 * separate streams.
 *
 * \param caps The format capabilities structure (NULL creates an empty topology)
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \note The format capabilities reference is NOT altered by this function
 * since a new format capabilities structure is created for each media type.
 *
 * \note Each stream will have its name set to the corresponding media type.
 * For example: "audio".
 *
 * \note Each stream will be set to the sendrecv state.
 *
 * \since 15
 */
struct ast_stream_topology *ast_stream_topology_create_from_format_cap(
	struct ast_format_cap *cap);

/*!
 * \brief Create a format capabilities structure representing the topology.
 *
 * \details
 * A helper function that, given a stream topology, creates a format
 * capabilities structure containing all formats from all active streams.
 *
 * \param topology The topology of streams
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \note The stream topology is NOT altered by this function.
 *
 * \since 15
 */
struct ast_format_cap *ast_format_cap_from_stream_topology(
	struct ast_stream_topology *topology);

/*!
 * \brief Gets the first active stream of a specific type from the topology
 *
 * \param topology The topology of streams
 * \param type The media type
 *
 * \retval non-NULL success
 * \retval NULL failure
 *
 * \since 15
 */
struct ast_stream *ast_stream_topology_get_first_stream_by_type(
	const struct ast_stream_topology *topology,
	enum ast_media_type type);

/*!
 * \brief Map a given topology's streams to the given types.
 *
 * \note The given vectors in which mapping values are placed are reset by
 *       this function. This means if those vectors already contain mapping
 *       values they will be lost.
 *
 * \param topology The topology to map
 * \param types The media types to be mapped
 * \param v0 Index mapping of topology to types
 * \param v1 Index mapping of types to topology
 *
 * \since 15
 */
void ast_stream_topology_map(const struct ast_stream_topology *topology,
	struct ast_vector_int *types, struct ast_vector_int *v0, struct ast_vector_int *v1);

/*!
 * \brief Get the stream group that a stream is part of
 *
 * \param stream The stream
 *
 * \return the numerical stream group (-1 if not in a group)
 *
 * \since 15.2.0
 */
int ast_stream_get_group(const struct ast_stream *stream);

/*!
 * \brief Set the stream group for a stream
 *
 * \param stream The stream
 * \param group The group the stream is part of
 *
 * \since 15.2.0
 */
void ast_stream_set_group(struct ast_stream *stream, int group);

#ifdef GRANDSTREAM_NETWORKS
enum ast_media_sub_type ast_stream_get_sub_type(const struct ast_stream *stream);
void ast_stream_set_sub_type(struct ast_stream *stream, enum ast_media_sub_type sub_type);
const char *ast_stream_get_gs_fec_version(const struct ast_stream *stream);
void ast_stream_set_gs_fec_version(struct ast_stream *stream, const char *gs_fec_version);
unsigned int ast_stream_get_alone_fec_version(const struct ast_stream *stream);
void ast_stream_set_alone_fec_version(struct ast_stream *stream, unsigned int value);
unsigned int ast_stream_get_has_rtcp_address(const struct ast_stream *stream);
void ast_stream_set_has_rtcp_address(struct ast_stream *stream, unsigned int has_rtcp_address);
unsigned int ast_stream_get_has_nack(const struct ast_stream *stream);
void ast_stream_set_has_nack(struct ast_stream *stream, unsigned int has_nack);
unsigned int ast_stream_get_has_nack_pli(const struct ast_stream *stream);
void ast_stream_set_has_nack_pli(struct ast_stream *stream, unsigned int has_nack_pli);
unsigned int ast_stream_get_has_ccm_fir(const struct ast_stream *stream);
void ast_stream_set_has_ccm_fir(struct ast_stream *stream, unsigned int has_ccm_fir);
struct ast_variable *ast_stream_get_passthrough_list(const struct ast_stream *stream, enum ast_passthrough_type passthrough_type);
int ast_stream_set_passthrough(struct ast_stream *stream, const char *m_key, const char *value, enum ast_passthrough_type passthrough_type);
void ast_stream_clear_passthrough(struct ast_stream *stream);
void ast_stream_set_local_media_direction(struct ast_stream *stream, enum ast_stream_media_direction media_direction);
enum ast_stream_media_direction ast_stream_get_local_media_direction(struct ast_stream *stream);
void ast_stream_set_remote_media_direction(struct ast_stream *stream, enum ast_stream_media_direction media_direction);
enum ast_stream_media_direction ast_stream_get_remote_media_direction(struct ast_stream *stream);
struct ast_stream_topology *ast_stream_topology_webrtc_create_from_format_cap_for_sfu(
	struct ast_format_cap *cap, enum ast_sfu_invite_type invite_type);
struct ast_stream_topology *ast_stream_topology_sip_create_from_format_cap_for_sfu(
	struct ast_format_cap *cap, enum ast_sfu_invite_type invite_type, unsigned int need_create_bfcp);
void ast_stream_set_video_resolution(struct ast_stream *stream, enum ast_stream_video_resolution video_resolution);
enum ast_stream_video_resolution ast_stream_get_video_resolution(struct ast_stream *stream);
void ast_stream_clear_bandw_passthrough(struct ast_stream *stream);
unsigned char ast_stream_judge_if_main_video(struct ast_stream *stream);
void ast_stream_set_change_ssrc(struct ast_stream *stream, unsigned int value);
unsigned int ast_stream_get_change_ssrc(const struct ast_stream *stream);
void ast_stream_set_held(struct ast_stream *stream, unsigned int value);
unsigned int ast_stream_get_held(struct ast_stream *stream);
void ast_stream_set_video_decline_reason(struct ast_stream *stream, enum ast_video_decline_reason reason);
enum ast_video_decline_reason ast_stream_get_video_decline_reason(struct ast_stream *stream);
unsigned int ast_stream_has_video_resource(struct ast_stream *stream);
struct ast_stream_topology *ast_stream_topology_clone_participant(const struct ast_stream_topology *topology);
unsigned int ast_stream_judge_if_stream_is_participant(struct ast_stream *stream);
int ast_stream_judge_if_audio(struct ast_stream *stream);
void ast_stream_set_video_rtx_pt(struct ast_stream *stream, const char *name, int pt);
int ast_stream_get_video_rtx_pt(struct ast_stream *stream, const char *name);
void ast_stream_init_video_rtx_pt(struct ast_stream *stream);

struct ast_format_cap *ast_format_cap_from_stream_topology_for_iax(struct ast_stream_topology *topology);
struct ast_format_cap* ast_create_format_by_topology(struct ast_stream_topology* topology, int ignore_video);
void ast_stream_set_uptime(struct ast_stream *stream, char *uptime);
char *ast_stream_get_uptime(struct ast_stream *stream);
void ast_stream_set_media_encryption(struct ast_stream *stream, enum ast_stream_media_encryption encryption);
enum ast_stream_media_encryption ast_stream_get_media_encryption(struct ast_stream *stream);
void ast_stream_set_same_codec_info(struct ast_stream *stream,
	enum ast_stream_same_codec_type type, struct ast_same_codec_info codec_info);
struct ast_same_codec_info *ast_stream_get_same_codec_info(const struct ast_stream *stream,
	enum ast_stream_same_codec_type type, unsigned int number);
int ast_stream_same_codec_info_get_count(const struct ast_stream *stream,
	enum ast_stream_same_codec_type type);
void ast_stream_init_same_codec_info(struct ast_stream *stream);
unsigned int ast_stream_judge_if_same_type(struct ast_stream *stream, enum ast_media_sub_type type);
unsigned int ast_stream_is_enabled_main_video(struct ast_stream *stream);
void ast_stream_set_previous_video_resolution(struct ast_stream *stream, enum ast_stream_video_resolution video_resolution);
enum ast_stream_video_resolution ast_stream_get_previous_video_resolution(struct ast_stream *stream);
#endif

#endif /* _AST_STREAM_H */
