/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthm@freeswitch.org>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Anthony Minessale II <anthm@freeswitch.org>
 *
 *
 * switch_xml.h -- XML PARSER
 *
 * Derived from EZXML http://ezxml.sourceforge.net
 * Original Copyright
 *
 * Copyright 2004, 2005 Aaron Voisine <aaron@voisine.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef AST_XML_H
#define AST_XML_H

#include "asterisk/utils.h"

#ifdef __cplusplus
#define AST_BEGIN_EXTERN_C       extern "C" {
#define AST_END_EXTERN_C         }
#else
#define AST_BEGIN_EXTERN_C
#define AST_END_EXTERN_C
#endif

typedef enum {
	SWITCH_FALSE = 0,
	SWITCH_TRUE = 1
} switch_bool_t;

/* Use UTF-8 as the general encoding */
#define USE_UTF_8_ENCODING SWITCH_TRUE

struct switch_xml_binding;

///\defgroup xml1 XML Library Functions
///\ingroup core1
///\{
#define SWITCH_XML_BUFSIZE 1024	// size of internal memory buffers
	typedef enum {
	SWITCH_XML_ROOT = (1 << 0),	// root
	SWITCH_XML_NAMEM = (1 << 1),	// name is malloced
	SWITCH_XML_TXTM = (1 << 2),	// txt is malloced
	SWITCH_XML_DUP = (1 << 3),	// attribute name and value are strduped
	SWITCH_XML_CDATA = (1 << 4) // body is in CDATA
} switch_xml_flag_t;

#if 1
typedef size_t switch_size_t;
#define switch_test_flag(obj, flag) ((obj)->flags & flag)
#define switch_str_nil(s) (s ? s : "")

#ifdef WIN32
#define SWITCH_DECLARE(type)			__declspec(dllimport) type __stdcall
#else
#define SWITCH_DECLARE(type) type
#endif

static inline char *switch_must_strdup(const char *_s)
{
	char *s = __ast_strdup(_s, __FILE__, __LINE__, __FUNCTION__);
	ast_assert(s);
	return s;
}

static inline void *switch_must_realloc(void *_b, size_t _z)
{
	void *m = __ast_realloc(_b, _z, __FILE__, __LINE__, __FUNCTION__);
	ast_assert(m);
	return m;
}

static inline void *switch_must_malloc(size_t _b)
{
	void *m = ast_malloc(_b);
	ast_assert(m);
	return m;
}

#ifndef __ATTR_SAL
	/* used for msvc code analysis */
	/* http://msdn2.microsoft.com/en-us/library/ms235402.aspx */
#define _In_
#define _In_z_
#define _In_opt_z_
#define _In_opt_
#define _Printf_format_string_
#define _Ret_opt_z_
#define _Ret_z_
#define _Out_opt_
#define _Out_
#define _Check_return_
#define _Inout_
#define _Inout_opt_
#define _In_bytecount_(x)
#define _Out_opt_bytecapcount_(x)
#define _Out_bytecapcount_(x)
#define _Ret_
#define _Post_z_
#define _Out_cap_(x)
#define _Out_z_cap_(x)
#define _Out_ptrdiff_cap_(x)
#define _Out_opt_ptrdiff_cap_(x)
#define _Post_count_(x)
#endif

typedef struct switch_xml *switch_xml_t;
#endif //if 1

/*! \brief A representation of an XML tree */
struct switch_xml {
	/*! tag name */
	char *name;
	/*! tag attributes { name, value, name, value, ... NULL } */
	char **attr;
	/*! tag character content, empty string if none */
	char *txt;
	/*! path to free on destroy */
	char *free_path;
	/*! tag offset from start of parent tag character content */
	switch_size_t off;
	/*! next tag with same name in this section at this depth */
	switch_xml_t next;
	/*! next tag with different name in same section and depth */
	switch_xml_t sibling;
	/*! next tag, same section and depth, in original order */
	switch_xml_t ordered;
	/*! head of sub tag list, NULL if none */
	switch_xml_t child;
	/*! parent tag, NULL if current tag is root tag */
	switch_xml_t parent;
	/*! flags */
	uint32_t flags;
	/*! is_switch_xml_root bool */
	switch_bool_t is_switch_xml_root_t;
	uint32_t refs;
};

/*!
 * \brief Parses a string into a switch_xml_t, ensuring the memory will be freed with switch_xml_free
 * \param s The string to parse
 * \param dup true if you want the string to be strdup()'d automatically
 * \return the switch_xml_t or NULL if an error occured
 */
SWITCH_DECLARE(switch_xml_t) switch_xml_parse_str_dynamic(_In_z_ char *s, _In_ switch_bool_t dup);

/*!
 * \brief Parses a string into a switch_xml_t
 * \param s The string to parse
 * \return the switch_xml_t or NULL if an error occured
 */
#define switch_xml_parse_str_dup(x)  switch_xml_parse_str_dynamic(x, SWITCH_TRUE)

///\brief Given a string of xml data and its length, parses it and creates an switch_xml
///\ structure. For efficiency, modifies the data by adding null terminators
///\ and decoding ampersand sequences. If you don't want this, copy the data and
///\ pass in the copy. Returns NULL on failure.
///\param s a string
///\param len the length of the string
///\return a formated xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_parse_str(_In_z_ char *s, _In_ switch_size_t len);

///\brief returns the first child tag (one level deeper) with the given name or NULL
///\ if not found
///\param xml an xml node
///\param name the name of the child tag
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_child(_In_ switch_xml_t xml, _In_z_ const char *name);

///\brief find a child tag in a node called 'childname' with an attribute 'attrname' which equals 'value'
///\param node the xml node
///\param childname the child tag name
///\param attrname the attribute name
///\param value the value
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_find_child(_In_ switch_xml_t node, _In_z_ const char *childname, _In_opt_z_ const char *attrname,
												   _In_opt_z_ const char *value);
SWITCH_DECLARE(switch_xml_t) switch_xml_find_child_multi(_In_ switch_xml_t node, _In_z_ const char *childname, ...);

///\brief Returns the Nth tag with the same name in the same section at the same depth
///\ or NULL if not found. An index of 0 returns the tag given.
///\param xml the xml node
///\param idx the index
///\return an xml node or NULL
switch_xml_t switch_xml_idx(_In_ switch_xml_t xml, _In_ int idx);

///\brief returns the name of the given tag
///\param xml the xml node
///\return the name
#define switch_xml_name(xml) ((xml) ? xml->name : NULL)

///\brief returns the given tag's character content or empty string if none
///\param xml the xml node
///\return the content
#define switch_xml_txt(xml) ((xml) ? xml->txt : "")

///\brief returns the value of the requested tag attribute, or NULL if not found
///\param xml the xml node
///\param attr the attribute
///\return the value
SWITCH_DECLARE(const char *) switch_xml_attr(_In_opt_ switch_xml_t xml, _In_opt_z_ const char *attr);

///\brief returns the value of the requested tag attribute, or "" if not found
///\param xml the xml node
///\param attr the attribute
///\return the value
SWITCH_DECLARE(const char *) switch_xml_attr_soft(_In_ switch_xml_t xml, _In_z_ const char *attr);

///\brief Traverses the switch_xml structure to retrieve a specific subtag. Takes a
///\ variable length list of tag names and indexes. The argument list must be
///\ terminated by either an index of -1 or an empty string tag name. Example:
///\ title = switch_xml_get(library, "shelf", 0, "book", 2, "title", -1);
///\ This retrieves the title of the 3rd book on the 1st shelf of library.
///\ Returns NULL if not found.
///\param xml the xml node
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_get(_In_ switch_xml_t xml,...);

///\brief Converts an switch_xml structure back to xml in html format. Returns a string of html data that
///\ must be freed.
///\param xml the xml node
///\param prn_header add <?xml version..> header too
///\param use_utf8_encoding encoding into ampersand entities for UTF-8 chars
///\return the ampersanded html text string to display xml
#define switch_xml_toxml(xml, prn_header) switch_xml_toxml_ex(xml, prn_header, USE_UTF_8_ENCODING)
#define switch_xml_toxml_nolock(xml, prn_header) switch_xml_toxml_nolock_ex(xml, prn_header, USE_UTF_8_ENCODING)

SWITCH_DECLARE(char *) switch_xml_toxml_ex(_In_ switch_xml_t xml, _In_ switch_bool_t prn_header, switch_bool_t use_utf8_encoding);
SWITCH_DECLARE(char *) switch_xml_toxml_nolock_ex(switch_xml_t xml, _In_ switch_bool_t prn_header, switch_bool_t use_utf8_encoding);

///\brief Converts an switch_xml structure back to xml using the buffer passed in the parameters.
///\param xml the xml node
///\param buf buffer to use
///\param buflen size of buffer
///\param offset offset to start at
///\param prn_header add <?xml version..> header too
///\param use_utf8_encoding encoding into ampersand entities for UTF-8 chars
///\return the xml text string
#define switch_xml_toxml_buf(xml, buf, buflen, offset, prn_header) switch_xml_toxml_buf_ex(xml, buf, buflen, offset, prn_header, USE_UTF_8_ENCODING);
SWITCH_DECLARE(char *) switch_xml_toxml_buf_ex(_In_ switch_xml_t xml, _In_z_ char *buf, _In_ switch_size_t buflen, _In_ switch_size_t offset,
											_In_ switch_bool_t prn_header, switch_bool_t use_utf8_encoding);

///\brief frees the memory allocated for an switch_xml structure
///\param xml the xml node
///\note in the case of the root node the readlock will be lifted
SWITCH_DECLARE(void) switch_xml_free(_In_opt_ switch_xml_t xml);

///\brief returns parser error message or empty string if none
///\param xml the xml node
///\return the error string or nothing
SWITCH_DECLARE(const char *) switch_xml_error(_In_ switch_xml_t xml);

///\brief returns a new empty switch_xml structure with the given root tag name
///\param name the name of the new root tag
SWITCH_DECLARE(switch_xml_t) switch_xml_new(_In_opt_z_ const char *name);

///\brief wrapper for switch_xml_new() that strdup()s name
///\param name the name of the root
///\return an xml node or NULL
#define switch_xml_new_d(name) switch_xml_set_flag(switch_xml_new(strdup(name)), SWITCH_XML_NAMEM)

///\brief Adds a child tag. off is the offset of the child tag relative to the start
///\ of the parent tag's character content. Returns the child tag.
///\param xml the xml node
///\param name the name of the tag
///\param off the offset
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_add_child(_In_ switch_xml_t xml, _In_z_ const char *name, _In_ switch_size_t off);

///\brief wrapper for switch_xml_add_child() that strdup()s name
///\param xml the xml node
///\param name the name of the child
///\param off the offset
SWITCH_DECLARE(switch_xml_t) switch_xml_add_child_d(_In_ switch_xml_t xml, _In_z_ const char *name, _In_ switch_size_t off);

///\brief sets the character content for the given tag and returns the tag
///\param xml the xml node
///\param txt the text
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_set_txt(switch_xml_t xml, const char *txt);

///\brief wrapper for switch_xml_set_txt() that strdup()s txt
///\ sets the character content for the given tag and returns the tag
///\param xml the xml node
///\param txt the text
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_set_txt_d(switch_xml_t xml, const char *txt);

///\brief Sets the given tag attribute or adds a new attribute if not found. A value
///\ of NULL will remove the specified attribute.
///\param xml the xml node
///\param name the attribute name
///\param value the attribute value
///\return the tag given
SWITCH_DECLARE(switch_xml_t) switch_xml_set_attr(switch_xml_t xml, const char *name, const char *value);

///\brief Wrapper for switch_xml_set_attr() that strdup()s name/value. Value cannot be NULL
///\param xml the xml node
///\param name the attribute name
///\param value the attribute value
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_set_attr_d(switch_xml_t xml, const char *name, const char *value);

///\brief Wrapper for switch_xml_set_attr() that strdup()s name/value. Value cannot be NULL
///\param xml the xml node
///\param name the attribute name
///\param value the attribute value
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_set_attr_d_buf(switch_xml_t xml, const char *name, const char *value);

///\brief sets a flag for the given tag and returns the tag
///\param xml the xml node
///\param flag the flag to set
///\return an xml node or NULL
SWITCH_DECLARE(switch_xml_t) switch_xml_set_flag(switch_xml_t xml, switch_xml_flag_t flag);

///\brief removes a tag along with its subtags without freeing its memory
///\param xml the xml node
SWITCH_DECLARE(switch_xml_t) switch_xml_cut(_In_ switch_xml_t xml);

///\brief inserts an existing tag into an ezxml structure
SWITCH_DECLARE(switch_xml_t) switch_xml_insert(_In_ switch_xml_t xml, _In_ switch_xml_t dest, _In_ switch_size_t off);

///\brief Moves an existing tag to become a subtag of dest at the given offset from
///\ the start of dest's character content. Returns the moved tag.
#define switch_xml_move(xml, dest, off) switch_xml_insert(switch_xml_cut(xml), dest, off)

///\brief removes a tag along with all its subtags
#define switch_xml_remove(xml) switch_xml_free(switch_xml_cut(xml))
///\}

/// * This routine is a variant of the "sprintf()" from the
/// * standard C library.  The resulting string is written into memory
/// * obtained from malloc() so that there is never a possiblity of buffer
// * overflow.  This routine also implement some additional formatting
// * options that are useful for constructing SQL statements.
///\ must be freed.
///\param sql commands
SWITCH_DECLARE(char *) switch_xml_mprintf(const char *zFormat, ...);
#endif // _AST_XML_H
/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 noet:
 */
