#ifdef GRANDSTREAM_NETWORKS
/*********************************************
 * File Name : bfcp.c
 * Author: bfdu
 * Date:   2014.02.09
 * Description: for bfcp
 * Changes: Initial
********************************************/
/*** MODULEINFO
	<depend>bfcp_external</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"
//ASTERISK_FILE_VERSION(__FILE__, "$Revision: 123344 $")
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>

#include "asterisk/bfcp.h"
#include "asterisk/frame.h"
#include "asterisk/channel.h"
#include "asterisk/config.h"
#include "asterisk/lock.h"
#include "asterisk/utils.h"
//#include "asterisk/netsock.h"
#include "asterisk/cli.h"
#include "asterisk/astobj2.h"
#include "asterisk/rtp_engine.h"
#include "../apps/confbridge/include/conf_resource.h"

AST_MODULE_LOG("core");

//static int bfcpstart = 50000;
//static int bfcpend = 52999;
#ifdef SO_NO_CHECK
static int nochecksums = 0;
#endif

//static int bfcp_tcp_start = 53000;
//static int bfcp_tcp_end = 55999;

static struct bfcp_supprim ast_bfcp_supported_primitives;
static struct bfcp_supattr ast_bfcp_supported_attributes;

/*! List of RTP engines that are currently registered */
static AST_RWLIST_HEAD_STATIC(engines, ast_bfcp_engine);

#define MAX_RETRANSMIT_COUNT 5

/* Define to 1 if you have the `fopencookie' function. */
#define HAVE_FOPENCOOKIE 1

#define BFCP_TCP_TIMEOUT 60 * 1000;

static struct bfcp_conf_resource_manager_funcs bfcp_conf_funcs;

void ast_bfcp_setnat(struct ast_bfcp_udp *bfcp, int nat)
{
	if (NULL == bfcp) {
		ast_log(LOG_WARNING, "bfcp is NULL, set nat failed");
		return;
	}
	bfcp->nat = nat;
}

int ast_bfcp_fd(const struct ast_bfcp_udp *bfcp)
{
	if (NULL == bfcp) {
		ast_log(LOG_WARNING, "bfcp is NULL, get fd failed");
		return -1;
	}
	return bfcp->fd;
}

void ast_bfcp_set_peer(struct ast_bfcp_udp *bfcp, const struct ast_sockaddr *them)
{
	if (NULL == bfcp || NULL == them) {
		ast_log(LOG_WARNING, "bfcp is NULL, set peer failed");
		return;
	}
	ast_sockaddr_copy(&bfcp->them, them);
}

void ast_bfcp_get_us(const struct ast_bfcp_udp *bfcp, struct ast_sockaddr *us)
{
	if (NULL == bfcp || NULL == us) {
		ast_log(LOG_WARNING, "bfcp is NULL, get us failed");
		return;
	}
	ast_sockaddr_copy(us, &bfcp->us);
}

void ast_bfcp_stop(struct ast_bfcp_udp *bfcp)
{
	if (NULL == bfcp) {
		ast_log(LOG_WARNING, "bfcp is NULL, stop failed");
		return;
	}
	ast_sockaddr_setnull(&bfcp->them);
}

void ast_bfcp_destroy(struct ast_bfcp_udp *bfcp)
{
	if (NULL == bfcp) {
		ast_log(LOG_WARNING, "bfcp is NULL, destory failed");
		return;
	}

	ast_log(LOG_DEBUG, "ast_bfcp_destroy for bfcp:%p\n", bfcp);

	if (&bfcp->timeout_buffer) {
		bfcp_free_buffer(&bfcp->timeout_buffer);
	}

	if (&bfcp->bfcp_recv_buffer) {
		bfcp_free_buffer(&bfcp->bfcp_recv_buffer);
	}

	if (&bfcp->bfcp_send_buffer) {
		bfcp_free_buffer(&bfcp->bfcp_send_buffer);
	}

	if (bfcp->sched) {
		AST_SCHED_DEL(bfcp->sched, bfcp->sched_id);
	}

	if (bfcp->ioid) {
		ast_io_remove(bfcp->io, bfcp->ioid);
	}
	if (bfcp->fd > -1) {
		close(bfcp->fd);
	}
	if (bfcp->tag) {
		ast_free(bfcp->tag);
	}

	if (bfcp->media_turn_relay && bfcp->engine) {
		bfcp->engine->destory(bfcp);
	}

	ast_free(bfcp);

	return;
}

static int ast_bfcp_init_supprim(struct bfcp_supprim *bfcp_supported_primitives)
{
	if (!bfcp_supported_primitives)
	{
		return -1;
	}

	bfcp_supported_primitives->len = BFCP_SUPPORT_PRIM_SIZE;
	bfcp_supported_primitives->prims = (enum bfcp_prim *)ast_malloc(BFCP_SUPPORT_PRIM_SIZE * sizeof(enum bfcp_prim));
	if (!bfcp_supported_primitives->prims) {
		return -1;
	}
	bfcp_supported_primitives->prims[0] = BFCP_FLOOR_REQUEST;
	bfcp_supported_primitives->prims[1] = BFCP_FLOOR_RELEASE;
	bfcp_supported_primitives->prims[2] = BFCP_FLOOR_REQUEST_QUERY;
	bfcp_supported_primitives->prims[3] = BFCP_FLOOR_REQUEST_STATUS;
	bfcp_supported_primitives->prims[4] = BFCP_FLOOR_QUERY;
	bfcp_supported_primitives->prims[5] = BFCP_FLOOR_STATUS;
	bfcp_supported_primitives->prims[6] = BFCP_HELLO;
	bfcp_supported_primitives->prims[7] = BFCP_HELLO_ACK;
	bfcp_supported_primitives->prims[8] = BFCP_FLOOR_REQ_STATUS_ACK;
	bfcp_supported_primitives->prims[9] = BFCP_FLOOR_STATUS_ACK;

	return 0;
}

static int ast_bfcp_init_supattr(struct bfcp_supattr *bfcp_supported_attributes)
{
	if (!bfcp_supported_attributes)
	{
		return -1;
	}

	bfcp_supported_attributes->len = BFCP_SUPPORT_ATTR_SIZE;
	bfcp_supported_attributes->attrs = (enum bfcp_attribute *)ast_malloc(BFCP_SUPPORT_ATTR_SIZE * sizeof(enum bfcp_attribute));
	if (!bfcp_supported_attributes->attrs) {
		return -1;
	}
	bfcp_supported_attributes->attrs[0] = BFCP_FLOOR_ID;
	bfcp_supported_attributes->attrs[1] = BFCP_FLOOR_REQUEST_ID;
	bfcp_supported_attributes->attrs[2] = BFCP_REQUEST_STATUS;
	bfcp_supported_attributes->attrs[3] = BFCP_FLOOR_REQ_INFO;
	bfcp_supported_attributes->attrs[4] = BFCP_REQUESTED_BY_INFO;
	bfcp_supported_attributes->attrs[5] = BFCP_PRIORITY;
	bfcp_supported_attributes->attrs[6] = BFCP_OVERALL_REQ_STATUS;

	return 0;
}

int ast_bfcp_init()
{
	int res = 0;
	res = ast_bfcp_init_supprim(&ast_bfcp_supported_primitives);
	if (res != 0) {
		return res;
	}
	res = ast_bfcp_init_supattr(&ast_bfcp_supported_attributes);

	return res;
}

void ast_bfcp_set_data(struct ast_bfcp_udp *bfcp, void *data)
{
	bfcp->data = data;
}

void *ast_bfcp_get_data(struct ast_bfcp_udp *bfcp_udp)
{
	if (!bfcp_udp) {
		return NULL;
	}

	return bfcp_udp->data;
}

static void bfcp_new_for_turn_relay(struct ast_bfcp_udp *bfcp,
	const char *engine_name, const char *remote_address, unsigned int media_turn_relay)
{
	struct ast_bfcp_engine *engine = NULL;

	if (!media_turn_relay) {
		return;
	}

	if (!ast_strlen_zero(engine_name)) {
		AST_RWLIST_TRAVERSE(&engines, engine, entry) {
			if (!strcmp(engine->name, engine_name)) {
				break;
			}
		}
	} else {
		engine = AST_RWLIST_FIRST(&engines);
	}

	if (!engine) {
		ast_log(LOG_WARNING, "[BFCP:%p] Can't find BFCP engine\n", bfcp);
		return;
	}

	bfcp->engine = engine;

	if (bfcp->engine->new(bfcp)) {
		ast_log(LOG_WARNING, "[BFCP:%p] Engine '%s' failed to new bfcp\n", bfcp, engine->name);
	}

	if (bfcp->engine->get_turn_relay_address(bfcp)) {
		ast_log(LOG_WARNING, "[BFCP:%p] Engine '%s' failed to get turn relay address\n", bfcp, engine->name);
	}

	if (bfcp->engine->create_turn_permission(bfcp, remote_address)) {
		ast_log(LOG_WARNING, "[BFCP:%p] Engine '%s' failed to create turn permission\n", bfcp, engine->name);
	}

	bfcp->media_turn_relay = 1;

	return;
}

struct ast_bfcp_udp *ast_bfcp_udp_new_with_bindaddr(const char *bfcp_protocol, struct ast_sched_context *sched,
	struct io_context *io, struct ast_sockaddr *addr, const char *engine_name, const char *remote_address, unsigned int media_turn_relay)
{
	struct ast_bfcp_udp *bfcp;
	int x;
	int startplace;
	long int flags;
	int bprotocol = SOCK_DGRAM;

	ast_log(LOG_DEBUG, "ast_bfcp_udp_new_with_bindaddr begin\n");

	int bfcpstart = ast_get_bfcp_start();
	int bfcpend = ast_get_bfcp_end();
	if (NULL == bfcp_protocol) {
		ast_log(LOG_WARNING, "bfcp_protocol is NULL");
		return NULL;
	}

	if (!(bfcp = ast_calloc(1, sizeof(*bfcp)))) {
		ast_log(LOG_WARNING, "Unable to alloc bfcp\n");
		return NULL;
	}

	if (!strcmp(bfcp_protocol, "TCP") || !strcmp(bfcp_protocol, "TCP/TLS")) {
		bprotocol = SOCK_STREAM;
	} else {
		bprotocol = SOCK_DGRAM;
	}
	bfcp->channel = NULL;
	ast_log(LOG_DEBUG, "ast_bfcp_udp_new_with_bindaddr protocol is %d\n", bprotocol);

	/* Record sched information we may need */
	bfcp->sched = sched;

	bfcp_create_buffer(&bfcp->timeout_buffer);
	bfcp_init_buffer(&bfcp->timeout_buffer);

	bfcp_create_buffer(&bfcp->bfcp_recv_buffer);
	bfcp_init_buffer(&bfcp->bfcp_recv_buffer);

	bfcp_create_buffer(&bfcp->bfcp_send_buffer);
	bfcp_init_buffer(&bfcp->bfcp_send_buffer);

	if ((bfcp->fd = socket(ast_sockaddr_is_ipv6(addr) ?
					AF_INET6 : AF_INET, SOCK_DGRAM, 0)) < 0) {
		ast_free(bfcp);
		ast_log(LOG_WARNING, "Unable to allocate socket: %s\n", strerror(errno));
		return NULL;
	}
	flags = fcntl(bfcp->fd, F_GETFL);
	fcntl(bfcp->fd, F_SETFL, flags | O_NONBLOCK);
#ifdef SO_NO_CHECK
	if (nochecksums)
		setsockopt(bfcp->fd, SOL_SOCKET, SO_NO_CHECK, &nochecksums, sizeof(nochecksums));
#endif
	/* Find us a place */
	x = (bfcpstart == bfcpend) ? bfcpstart : (ast_random() % (bfcpend - bfcpstart)) + bfcpstart;
	x = x & ~1;
	startplace = x;
	for (;;) {
		ast_sockaddr_copy(&bfcp->us, addr);
		ast_sockaddr_set_port(&bfcp->us, x);
		if (ast_bind(bfcp->fd, &bfcp->us) == 0) {
			break;
		}

		x++;

		if (x > bfcpend) {
			x = (bfcpstart + 1) & ~1;
		}

		if (x == startplace || (errno != EADDRINUSE && errno != EACCES)) {
			ast_log(LOG_ERROR, "Oh dear... we couldn't allocate a port for UDP BFCP '%p'\n", bfcp);
			close(bfcp->fd);
			ast_free(bfcp);
			return NULL;
		}
	}

	bfcp_new_for_turn_relay(bfcp, engine_name, remote_address, media_turn_relay);

	return bfcp;
}

static int send_bfcp_msg_timeout(const void *data)
{
	struct ast_bfcp_udp *bfcp = NULL;
	int res = 0;

	if (NULL == data) {
		return 0;
	}

	bfcp = (struct ast_bfcp_udp *)data;

	if (!bfcp->sending_msg) {
		ast_log(LOG_DEBUG, "BFCP has receive ACK, don't send again\n");
		return 0;
	}

	if (bfcp->retransmit_count > MAX_RETRANSMIT_COUNT) {
		ast_log(LOG_DEBUG, "BFCP retransmit has reached max counts, don't send again\n");
		return 0;
	}

	if (bfcp->timeout_buffer.pos > 0 && !ast_sockaddr_isnull(&bfcp->them)) {
		if (bfcp->media_turn_relay) {
			if (bfcp->engine->turn_packet_send(bfcp, bfcp->timeout_buffer.buf, bfcp->timeout_buffer.pos)) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			} else {
				bfcp->retransmit_count++;
				return BFCP_T1 * (2 << (bfcp->retransmit_count - 1));
			}
		} else {
			if ((res = ast_sendto(bfcp->fd, bfcp->timeout_buffer.buf, bfcp->timeout_buffer.pos, 0, &bfcp->them)) < 0) {
				ast_log(LOG_NOTICE, "BFCP : retransmission error to %s: %s\n",
				ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			} else {
				bfcp->retransmit_count++;
				return BFCP_T1 * (2 << (bfcp->retransmit_count - 1));
			}
		}
	}

	return 0;
}

static int bfcp_process_hello(BFCP_MSG *bfcp_msg, struct bfcp_buffer *send_buffer)
{
	int res = 0;

	if (!bfcp_msg || !send_buffer)
	{
		ast_log(LOG_WARNING, "bfcp_process_hello, param is NULL\n");
		return -1;
	}

	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_HELLO_ACK,
	                   bfcp_msg->header.conf_id, bfcp_msg->header.transaction_id, bfcp_msg->header.user_id,
	                   2, BFCP_SUPPORTED_PRIMS | BFCP_ATTR_MANDATORY, 0, &ast_bfcp_supported_primitives,
	                   BFCP_SUPPORTED_ATTRS | BFCP_ATTR_MANDATORY, 0, &ast_bfcp_supported_attributes);

	return res;
}

/* Todo: to process hello ack process, now just ignore */
static int bfcp_process_hello_ack(BFCP_MSG *bfcpMsg)
{
	return 0;
}

static int bfcp_process_floor_req_status(BFCP_MSG *bfcp_msg, 
	struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info)
{
	int res = 0;
	BFCP_ATTR *req_status_attr = NULL;

	if (!bfcp_msg || !send_buffer || !bfcp_info)
	{
		ast_log(LOG_WARNING, "bfcp_process_floor_req_status, param is NULL\n");
		return -1;
	}

	req_status_attr = bfcp_attrs_find(&bfcp_msg->attr_list, BFCP_REQUEST_STATUS);
	if (req_status_attr) {
		bfcp_info->status = req_status_attr->v.reqstatus.status;
	}

	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQ_STATUS_ACK,
	bfcp_msg->header.conf_id, bfcp_msg->header.transaction_id, bfcp_msg->header.user_id, 0);

	return res;
}

static void bfcp_release_floor_use_conference_interface(struct ast_bfcp_info *bfcp_info, struct ast_channel *channel)
{
	RAII_VAR(struct ast_channel *, host_channel, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, slider_channel, NULL, ao2_cleanup);

	if (!bfcp_info || !channel) {
		return;
	}

	bfcp_info->primitive = BFCP_FLOOR_REQUEST_STATUS;

	slider_channel = bfcp_conf_funcs.conf_get_slider_channel(bfcp_info->info_conf_id);
	if (slider_channel) {
		if (slider_channel == channel) {
			if (bfcp_conf_funcs.conf_put_slides_video_resource(bfcp_info->info_conf_id, channel)) {
				bfcp_info->status = BFCP_DENIED;
				ast_log(LOG_DEBUG, "Release slides video resource failed for conf:%s, channel:%s\n", bfcp_info->info_conf_id, ast_channel_name(channel));
			} else {
				ast_log(LOG_DEBUG, "Release slides video successful\n");
				bfcp_info->status = BFCP_RELEASED;
			}
		} else {
			host_channel = bfcp_conf_funcs.conf_get_sfu_moderator(channel, bfcp_info->info_conf_id);
			if (channel == host_channel) {
				if (bfcp_conf_funcs.conf_put_slides_video_resource(bfcp_info->info_conf_id, slider_channel)) {
					bfcp_info->status = BFCP_DENIED;
					ast_log(LOG_DEBUG, "Release slides video resource failed for conf:%s, channel:%s\n", bfcp_info->info_conf_id, ast_channel_name(channel));
				} else {
					ast_log(LOG_DEBUG, "Release slides video forcely by host\n");
					bfcp_info->status = BFCP_RELEASED;
				}
			} else {
				ast_log(LOG_DEBUG, "Release slides video forcely failed, it's not host\n");
				bfcp_info->status = BFCP_DENIED;
			}
		}
	} else {
		bfcp_info->status = BFCP_DENIED;
	}

	return;
}

static void set_bfcp_info_msid(struct ast_channel *channel, struct ast_bfcp_info *bfcp_info)
{
	char *msid = NULL;

	if (!channel || !bfcp_info) {
		return;
	}

	msid = ast_channel_get_slides_video_local_msid(channel);
	if (!msid) {
		return;
	}

	strncpy(bfcp_info->msid, msid, sizeof(bfcp_info->msid) - 1);

	ast_free(msid);
	msid = NULL;

	return;
}

static void set_bfcp_slider_name(struct ast_channel *chan, struct ast_bfcp_info *bfcp_info)
{
	char *slider_name = NULL;

	if (!chan || !bfcp_info) {
		return;
	}

	if (!bfcp_conf_funcs.conf_set_slider_name) {
		return;
	}

	slider_name = S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, "");
	if (ast_strlen_zero(slider_name)) {
		ast_log(LOG_DEBUG, "channel:%s has no name, don't need set\n", ast_channel_name(chan));
		return;
	}

	bfcp_conf_funcs.conf_set_slider_name(bfcp_info->info_conf_id, slider_name);

	ast_log(LOG_DEBUG, "set_bfcp_slider_name:%s, channel name:%s\n", slider_name, ast_channel_name(chan));

	return;
}

static void bfcp_process_floor_status_for_conference(BFCP_MSG *bfcp_msg,
	struct ast_bfcp_info *bfcp_info, struct ast_channel *channel)
{
	BFCP_ATTR *req_status_attr = NULL;
	enum bfcp_reqstatus req_status = 0;

	req_status_attr = bfcp_attrs_find(&bfcp_msg->attr_list, BFCP_REQUEST_STATUS);
	if (req_status_attr) {
		req_status = req_status_attr->v.reqstatus.status;
		ast_log(LOG_DEBUG, "bfcp_process_floor_status status:%d\n", req_status);
	}

	switch (req_status) {
		case BFCP_GRANTED:
			if (bfcp_conf_funcs.conf_get_slides_video_resource &&
				VIDEO_RESOLUTION_1080P == bfcp_conf_funcs.conf_get_slides_video_resource(bfcp_info->info_conf_id, channel, ast_true(bfcp_info->share_audio))) {
				ast_log(LOG_DEBUG, "Get slides video resource sucess for conf:%s, channel:%s\n", bfcp_info->info_conf_id, ast_channel_name(channel));
				bfcp_info->status = BFCP_GRANTED;
				set_bfcp_info_msid(channel, bfcp_info);
				set_bfcp_slider_name(channel, bfcp_info);
			} else {
				bfcp_info->status = BFCP_DENIED;
			}
			break;
		case BFCP_CANCELLED:
		case BFCP_RELEASED:
			if (bfcp_conf_funcs.conf_put_slides_video_resource &&
				bfcp_conf_funcs.conf_get_slider_channel &&
				bfcp_conf_funcs.conf_get_sfu_moderator) {
				bfcp_release_floor_use_conference_interface(bfcp_info, channel);
				bfcp_info->primitive = BFCP_FLOOR_STATUS;
			} else {
				bfcp_info->status = BFCP_RELEASED;
			}
			break;
		default:
			ast_log(LOG_DEBUG, "bfcp_process_floor_status can't process status:%d\n", req_status);
			break;
	}

	return;
}

static int bfcp_process_floor_status(BFCP_MSG *bfcp_msg,
	struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info, struct ast_channel *channel)
{
	int res = 0;
	BFCP_ATTR *req_status_attr = NULL;

	if (!bfcp_msg || !send_buffer || !bfcp_info || !channel)
	{
		ast_log(LOG_WARNING, "bfcp_process_floor_status, param is NULL\n");
		return -1;
	}

	ast_log(LOG_DEBUG, "bfcp_process_floor_status send the BFCP_FLOOR_REQUEST_STATUS_ACK\n");

	if (strlen(bfcp_info->info_conf_id)) {
		bfcp_process_floor_status_for_conference(bfcp_msg, bfcp_info, channel);
	} else {
		req_status_attr = bfcp_attrs_find(&bfcp_msg->attr_list, BFCP_REQUEST_STATUS);
		if (req_status_attr) {
			bfcp_info->status = req_status_attr->v.reqstatus.status;
			ast_log(LOG_DEBUG, "bfcp_process_floor_status status:%d\n", bfcp_info->status);
		}
	}

	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_STATUS_ACK,
	                    bfcp_msg->header.conf_id, bfcp_msg->header.transaction_id, bfcp_msg->header.user_id, 0);

	return res;
}

static int bfcp_hello_msg_encode(struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info)
{
	int res = 0;

	if (!send_buffer || !bfcp_info)
	{
		ast_log(LOG_WARNING, "bfcp_hello_msg_encode param is NULL\n");
		return -1;
	}

	bfcp_info->transaction_id++;
	if (bfcp_info->transaction_id >= 65535)
	{
		bfcp_info->transaction_id = 1;
	}
	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_HELLO,
	                   bfcp_info->conf_id, bfcp_info->transaction_id, bfcp_info->user_id, 0);

	ast_log(LOG_DEBUG, "bfcp_hello_msg_encode res=%d\n", res);

	return res;
}

static int bfcp_floor_request_msg_encode(struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info)
{
	int res = 0;

	if (!send_buffer || !bfcp_info)
	{
		ast_log(LOG_WARNING, "bfcp_floor_request_msg_encode param is NULL\n");
		return -1;
	}

	bfcp_info->transaction_id++;
	if(bfcp_info->transaction_id >= 65535)
	{
		bfcp_info->transaction_id = 1;
	}
	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQUEST,
	               bfcp_info->conf_id, bfcp_info->transaction_id, bfcp_info->user_id,
	               1, BFCP_FLOOR_ID | BFCP_ATTR_MANDATORY, 0, &bfcp_info->floor_id );

	ast_log(LOG_DEBUG, "bfcp_floor_request_msg_encode res=%d\n", res);

	return res;
}

static int bfcp_floor_release_msg_encode(struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info)
{
	int res = 0;

	if (!send_buffer || !bfcp_info)
	{
		ast_log(LOG_WARNING, "bfcp_floor_release_msg_encode param is NULL\n");
		return -1;
	}

	bfcp_info->transaction_id++;
	if(bfcp_info->transaction_id >= 65535)
	{
		bfcp_info->transaction_id = 1;
	}
	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_RELEASE,
	               bfcp_info->conf_id, bfcp_info->transaction_id, bfcp_info->user_id,
	               1, BFCP_FLOOR_REQUEST_ID | BFCP_ATTR_MANDATORY, 0, &bfcp_info->floor_requst_id);

	ast_log(LOG_DEBUG, "bfcp_floor_release_msg_encodes res=%d\n", res);

	return res;
}

static int bfcp_floor_request_status_msg_encode(struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info)
{
	int res = 0;
	struct bfcp_reqstatus_struct req_status;

	if (!send_buffer || !bfcp_info)
	{
		ast_log(LOG_WARNING, "bfcp_floor_request_status_msg_encode param is NULL\n");
		return -1;
	}

	bfcp_info->transaction_id++;
	if(bfcp_info->transaction_id >= 65535)
	{
		bfcp_info->transaction_id = 1;
	}
	req_status.status = bfcp_info->status;
	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQUEST_STATUS,
	                   bfcp_info->conf_id, bfcp_info->transaction_id, bfcp_info->user_id,
	                   1, BFCP_FLOOR_REQ_INFO | BFCP_ATTR_MANDATORY, 2, &bfcp_info->floor_requst_id,
	                   BFCP_OVERALL_REQ_STATUS | BFCP_ATTR_MANDATORY, 1, &bfcp_info->floor_requst_id,
	                   BFCP_REQUEST_STATUS | BFCP_ATTR_MANDATORY, 0, &req_status,
	                   BFCP_FLOOR_REQ_STATUS | BFCP_ATTR_MANDATORY, 0, &bfcp_info->floor_id);

	ast_log(LOG_DEBUG, "bfcp_floor_request_status_msg_encode res=%d\n", res);

	return res;
}

static int bfcp_floor_status_msg_encode(struct bfcp_buffer *send_buffer, struct ast_bfcp_info *bfcp_info)
{
	int res = 0;
	struct bfcp_reqstatus_struct req_status;

	if (!send_buffer || !bfcp_info)
	{
		ast_log(LOG_WARNING, "bfcp_floor_status_msg_encode param is NULL\n");
		return -1;
	}

	bfcp_info->transaction_id++;
	if(bfcp_info->transaction_id >= 65535)
	{
		bfcp_info->transaction_id = 1;
	}

	req_status.status = bfcp_info->status;

	res = bfcp_msg_encode(send_buffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_STATUS,
	               bfcp_info->conf_id, bfcp_info->transaction_id, bfcp_info->user_id,
	               2, BFCP_FLOOR_ID | BFCP_ATTR_MANDATORY, 0, &bfcp_info->floor_id,
	               BFCP_FLOOR_REQ_INFO | BFCP_ATTR_MANDATORY, 2, &bfcp_info->floor_requst_id,
	               BFCP_OVERALL_REQ_STATUS | BFCP_ATTR_MANDATORY, 1, &bfcp_info->floor_requst_id,
	               BFCP_REQUEST_STATUS | BFCP_ATTR_MANDATORY, 0, &req_status,
	               BFCP_FLOOR_REQ_STATUS | BFCP_ATTR_MANDATORY, 0, &bfcp_info->floor_id);

	ast_log(LOG_DEBUG, "bfcp_floor_status_msg_encode res=%d\n", res);

	return res;
}

static enum bfcp_info_action bfcp_primitive_to_action(const struct ast_bfcp_info *bfcp_info)
{
	enum bfcp_info_action action = 0;

	switch (bfcp_info->primitive) {
		case BFCP_FLOOR_REQUEST:
			action = BFCP_INFO_ACTION_FLOOR_REQUEST;
			break;
		case BFCP_FLOOR_RELEASE:
			action = BFCP_INFO_ACTION_FLOOR_RELEASE;
			break;
		case BFCP_FLOOR_REQUEST_STATUS:
			action = BFCP_INFO_ACTION_FLOOR_REQUEST_STATUS;
			break;
		case BFCP_FLOOR_STATUS:
			action = BFCP_INFO_ACTION_FLOOR_STATUS;
			break;
		default:
			ast_log(LOG_DEBUG, "Can't get BFCP primitive value\n");
			break;
	}

	return action;
}

void ast_bfcp_set_channel_slides_status(struct ast_channel *channel,
	const struct ast_bfcp_info *bfcp_info, unsigned int is_recv_bfcp_msg)
{
	enum bfcp_info_action action = 0;

	if (!channel || !bfcp_info) {
		return;
	}

	ast_channel_lock(channel);

	ast_log(LOG_DEBUG, "ast_bfcp_set_channel_slides_status action:%d, channel:%s\n",
		bfcp_info->action, ast_channel_name(channel));

	action = bfcp_info->action;
	if (!action) {
		action = bfcp_primitive_to_action(bfcp_info);
	}

	switch (action) {
		case BFCP_INFO_ACTION_FLOOR_REQUEST:
			if (is_recv_bfcp_msg) {
				ast_channel_set_send_slides_video(channel, 1);
			} else {
				ast_channel_set_recv_slides_video(channel, 1);
			}
			break;
		case BFCP_INFO_ACTION_FLOOR_RELEASE:
		case BFCP_INFO_ACTION_FLOOR_CANCEL:
			if (is_recv_bfcp_msg) {
				ast_channel_set_send_slides_video(channel, 0);
			} else {
				ast_channel_set_recv_slides_video(channel, 0);
			}
			break;
		case BFCP_INFO_ACTION_FLOOR_REQUEST_STATUS:
			if (bfcp_info->status == BFCP_DENIED) {
				ast_channel_set_send_slides_video(channel, 0);
				ast_channel_set_recv_slides_video(channel, 0);
			}
			break;
		case BFCP_INFO_ACTION_FLOOR_STATUS:
			if (bfcp_info->status == BFCP_GRANTED) {
				if (is_recv_bfcp_msg) {
					ast_channel_set_send_slides_video(channel, 1);
				} else {
					ast_channel_set_recv_slides_video(channel, 1);
				}
			} else if ((bfcp_info->status == BFCP_DENIED) ||
				(bfcp_info->status == BFCP_RELEASED)) {
				ast_channel_set_send_slides_video(channel, 0);
				ast_channel_set_recv_slides_video(channel, 0);
			}
			break;
		default:
			ast_log(LOG_DEBUG, "bfcp receive unknow notify, can't process now\n");
			break;
	}

	ast_channel_unlock(channel);

	return;
}

static int judge_if_channel_is_trunk(struct ast_channel *channel)
{
	if (!channel) {
		return 0;
	}

	if (!strncmp(ast_channel_name(channel), "PJSIP/trunk", 11)) {
		return 1;
	}

	return 0;
}

static int get_extension_from_channel_name(const char *channel_name, char *buf, int size)
{
	const char *start = NULL;
	const char *end =  NULL;

	memset(buf, 0, size);

	if (strncasecmp(channel_name, "PJSIP/", 6)) {
		return -1;
	}
	start = channel_name + 6;
	end = strrchr(start, '-');
	if (!end) {
		return -1;
	}

	strncpy(buf, start, size > (end - start) ? (end - start) : size);
	return 0;
}

static void ast_bfcp_set_info(struct ast_channel *chan, struct ast_bfcp_info *bfcp_info)
{
	char *slider_name = NULL;
	char extension[128] = { 0 };

	if (!chan || !bfcp_info) {
		ast_log(LOG_ERROR, "ast_bfcp_set_info param is NULL\n");
		return;
	}

	get_extension_from_channel_name(ast_channel_name(chan), extension, sizeof(extension));
	strncpy(bfcp_info->account_id, extension, sizeof(bfcp_info->account_id) - 1);

	ast_log(LOG_DEBUG, "ast_bfcp_set_info account_id:%s, chan:%s\n", bfcp_info->account_id, ast_channel_name(chan));

	if (!judge_if_channel_is_trunk(chan)) {
		slider_name = S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, "");
		if (!ast_strlen_zero(slider_name)) {
			strncpy(bfcp_info->slider_name, slider_name, sizeof(bfcp_info->slider_name) - 1);
			ast_log(LOG_DEBUG, "ast_bfcp_set_info slider_name:%s, chan:%s\n", bfcp_info->slider_name, ast_channel_name(chan));
		}
	}

	return;
}

static int ast_process_bfcp_msg(struct ast_bfcp_udp *bfcp, int data_len)
{
	BFCP_MSG *bfcp_msg = NULL;
	int res = 0;
	enum bfcp_prim prim;
	unsigned int need_send_bfcp_msg = 0;
	struct ast_bfcp_info bfcp_info;

	struct bfcp_buffer bfcp_recv_buffer = bfcp->bfcp_recv_buffer;
	struct bfcp_buffer bfcp_send_buffer = bfcp->bfcp_send_buffer;

	bfcp_init_buffer(&bfcp_recv_buffer);
	bfcp_init_buffer(&bfcp_send_buffer);

	memcpy(bfcp_recv_buffer.buf, bfcp->rawdata, data_len);
	bfcp_recv_buffer.size = data_len;

	res = bfcp_msg_decode(&bfcp_msg, &bfcp_recv_buffer);

	if (res < 0)
	{
		ast_log(LOG_WARNING, "ast_process_bfcp_msg error:%d\n", res);
		goto  PROCESS_BFCP_MSG_CLEANUP;
	}

	prim = bfcp_msg->header.prim;
	if (prim != BFCP_HELLO && prim != BFCP_HELLO_ACK)
	{
		ast_log(LOG_DEBUG, "Received bfcp message header prim:%d for user_id:%d\n", prim, bfcp_msg->header.user_id);
	}

	bfcp->bfcp_info.primitive = bfcp_msg->header.prim;

	memset(&bfcp_info, 0, sizeof(bfcp_info));
	bfcp_info.primitive = prim;

	switch (prim)
	{
		case BFCP_HELLO:
			res = bfcp_process_hello(bfcp_msg, &bfcp_send_buffer);
			need_send_bfcp_msg = 1;
			break;

		case BFCP_HELLO_ACK:
			res = bfcp_process_hello_ack(bfcp_msg);
			AST_SCHED_DEL(bfcp->sched, bfcp->sched_id);
			bfcp->sending_msg = 0;
			need_send_bfcp_msg = 0;
			break;

		case BFCP_FLOOR_REQUEST:
			ast_bfcp_set_info(bfcp->channel, &bfcp_info);
			bfcp->channel ? ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info)) : 1;
			need_send_bfcp_msg = 0;
			break;
		case BFCP_FLOOR_RELEASE:
			ast_bfcp_set_info(bfcp->channel, &bfcp_info);
			if (ast_channel_get_send_slides_video(bfcp->channel)) {
				ast_channel_set_opening_slides_video(bfcp->channel, 0);
				ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 0;
			break;

		case BFCP_FLOOR_REQUEST_STATUS:
			res = bfcp_process_floor_req_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info);
			bfcp->channel ? ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info)) : 1;
			AST_SCHED_DEL(bfcp->sched, bfcp->sched_id);
			bfcp->sending_msg = 0;
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_STATUS:
			ast_bfcp_set_info(bfcp->channel, &bfcp_info);
			res = bfcp_process_floor_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info, bfcp->channel);
			if (bfcp_info.status != BFCP_RELEASED ||
				judge_if_channel_is_trunk(bfcp->channel) ||
				(bfcp_info.status == BFCP_RELEASED &&
				ast_channel_get_send_slides_video(bfcp->channel))) {
				ast_channel_set_opening_slides_video(bfcp->channel, 1);
				ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_REQ_STATUS_ACK:
		case BFCP_FLOOR_STATUS_ACK:
			AST_SCHED_DEL(bfcp->sched, bfcp->sched_id);
			bfcp->sending_msg = 0;
			need_send_bfcp_msg = 0;
			break;

		default:
			need_send_bfcp_msg = 0;
			ast_log(LOG_DEBUG, "Can't process this bfcp message\n");
			break;
	}

	if (res < 0)
	{
		ast_log(LOG_WARNING, "bfcp process error\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	ast_bfcp_set_channel_slides_status(bfcp->channel, &bfcp_info, 1);

	if (!need_send_bfcp_msg)
	{
		ast_log(LOG_DEBUG, "bfcp don't need to send message\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	if (bfcp_send_buffer.pos > 0 && !ast_sockaddr_isnull(&bfcp->them)) {
		if (bfcp->media_turn_relay) {
			if (bfcp->engine->turn_packet_send(bfcp, bfcp_send_buffer.buf, bfcp_send_buffer.pos)) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			}
		} else {
			if ((res = ast_sendto(bfcp->fd, bfcp_send_buffer.buf, bfcp_send_buffer.pos, 0, &bfcp->them)) < 0) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
				ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			}
		}
	}

PROCESS_BFCP_MSG_CLEANUP:
	bfcp_free_msg(bfcp_msg);

	return 1;
}

static const char * const bfcp_reqstatus2string[][1] = {
    { /* 0 BFCP_INIT */             "Init"               },
    { /* 1 BFCP_PENDING */          "Pending"            },
    { /* 2 BFCP_ACCEPTED */         "Accepted"           },
    { /* 3 BFCP_GRANTED */          "Granted"            },
    { /* 4 BFCP_DENIED */           "Denied"             },
    { /* 5 BFCP_CANCELLED */        "Cancelled"          },
    { /* 6 BFCP_RELEASED */         "Released"           },
    { /* 7 BFCP_REVOKED */          "Revoked"            },

    { /* 8 BFCP_UNKNOWN */          "bfcp status unknown"},
};

const static char *bfcp_req_status_2str(enum bfcp_reqstatus status)
{
    if (status > BFCP_REVOKED)
    {
        return "bfcp reqstatus unknown";
    }

    return bfcp_reqstatus2string[status][0];
}

static int bfcp_process_floor_request(struct ast_bfcp_udp *bfcp, BFCP_MSG *bfcpMsg,
	struct bfcp_buffer *sendBuffer)
{
	int res = 0;
	struct bfcp_reqstatus_struct req_status;

	if (!bfcp || !bfcpMsg || !sendBuffer)
	{
		ast_log(LOG_WARNING, "bfcp_process_floor_request, param is NULL\n");
		return -1;
	}

	if (bfcpMsg->header.transaction_id == bfcp->bfcp_info.received_transaction_id)
	{
		ast_log(LOG_DEBUG, "bfcp_process_floor_request, recv a re-transmit message, conf:%d, user:%d\n",
		bfcp->bfcp_info.conf_id, bfcp->bfcp_info.user_id);
		memcpy(sendBuffer->buf, bfcp->bfcp_send_buffer.buf, bfcp->bfcp_send_buffer.size);
		sendBuffer->size = bfcp->bfcp_send_buffer.size;
		sendBuffer->pos = bfcp->bfcp_send_buffer.pos;
		return 0;
	}
    bfcp->bfcp_info.received_transaction_id = bfcpMsg->header.transaction_id;

    if (bfcp_conf_funcs.conf_get_slides_video_resource) {
		if (VIDEO_RESOLUTION_1080P == bfcp_conf_funcs.conf_get_slides_video_resource(bfcp->bfcp_info.info_conf_id, bfcp->channel, ast_true(bfcp->bfcp_info.share_audio))) {
			ast_log(LOG_DEBUG, "Get slides video resource sucess for conf:%s, channel:%s\n", bfcp->bfcp_info.info_conf_id, ast_channel_name(bfcp->channel));
			req_status.status = BFCP_GRANTED;
			bfcp->bfcp_info.status = BFCP_GRANTED;
			set_bfcp_info_msid(bfcp->channel, &bfcp->bfcp_info);
			set_bfcp_slider_name(bfcp->channel, &bfcp->bfcp_info);
		} else {
			req_status.status = BFCP_DENIED;
			bfcp->bfcp_info.status = BFCP_DENIED;
			res = -1;
		}
	} else {
		ast_log(LOG_DEBUG, "Has no registed func conf_get_slides_video_resource\n");
		req_status.status = BFCP_GRANTED;
		bfcp->bfcp_info.status = BFCP_GRANTED;
	}

	ast_log(LOG_DEBUG, "bfcp_process_floor_request send the BFCP_FLOOR_REQUEST_STATUS, reqStatus.status:%s\n",
		bfcp_req_status_2str(req_status.status));
	req_status.qpos = 0;

	res |= bfcp_msg_encode(sendBuffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQUEST_STATUS,
							bfcp->bfcp_info.conf_id, bfcpMsg->header.transaction_id, bfcp->bfcp_info.user_id,
							1, BFCP_FLOOR_REQ_INFO | BFCP_ATTR_MANDATORY, 2, &bfcp->bfcp_info.floor_requst_id,
							BFCP_OVERALL_REQ_STATUS | BFCP_ATTR_MANDATORY, 1, &bfcp->bfcp_info.floor_requst_id,
							BFCP_REQUEST_STATUS | BFCP_ATTR_MANDATORY, 0, &req_status,
							BFCP_FLOOR_REQ_STATUS | BFCP_ATTR_MANDATORY, 0, &bfcp->bfcp_info.floor_id);

	ast_log(LOG_DEBUG, "bfcp_process_floor_request res=%d\n", res);

	return res;
}

static int bfcp_process_floor_release(struct ast_bfcp_udp *bfcp, BFCP_MSG *bfcpMsg,
	struct bfcp_buffer *sendBuffer)
{
	int res = 0;
	struct bfcp_reqstatus_struct req_status;

	if (!bfcp || !bfcpMsg || !sendBuffer)
	{
		ast_log(LOG_WARNING, "bfcp_process_floor_release, param is NULL\n");
		return -1;
	}

	if (bfcpMsg->header.transaction_id == bfcp->bfcp_info.received_transaction_id)
	{
		ast_log(LOG_DEBUG, "bfcp_process_floor_release, recv a re-transmit message, conf:%d, user:%d\n",
		bfcp->bfcp_info.conf_id, bfcp->bfcp_info.user_id);
		memcpy(sendBuffer->buf, bfcp->bfcp_send_buffer.buf, bfcp->bfcp_send_buffer.size);
		sendBuffer->size = bfcp->bfcp_send_buffer.size;
		sendBuffer->pos = bfcp->bfcp_send_buffer.pos;
		return 0;
	}
	bfcp->bfcp_info.received_transaction_id = bfcpMsg->header.transaction_id;

	if (bfcp_conf_funcs.conf_put_slides_video_resource && bfcp_conf_funcs.conf_get_slider_channel && bfcp_conf_funcs.conf_get_sfu_moderator) {
		bfcp_release_floor_use_conference_interface(&bfcp->bfcp_info, bfcp->channel);
		req_status.status = bfcp->bfcp_info.status;
	} else {
		req_status.status = BFCP_RELEASED;
		bfcp->bfcp_info.status = BFCP_RELEASED;
	}

	ast_log(LOG_DEBUG, "bfcp_process_floor_release send the BFCP_FLOOR_REQUEST_STATUS, reqStatus.status:%s\n",
	bfcp_req_status_2str(req_status.status));
	req_status.qpos = 0;

	res = bfcp_msg_encode(sendBuffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQUEST_STATUS,
							bfcp->bfcp_info.conf_id, bfcpMsg->header.transaction_id, bfcp->bfcp_info.user_id,
							1, BFCP_FLOOR_REQ_INFO | BFCP_ATTR_MANDATORY, 2, &bfcp->bfcp_info.floor_requst_id,
							BFCP_OVERALL_REQ_STATUS | BFCP_ATTR_MANDATORY, 1, &bfcp->bfcp_info.floor_requst_id,
							BFCP_REQUEST_STATUS | BFCP_ATTR_MANDATORY, 0, &req_status,
							BFCP_FLOOR_REQ_STATUS | BFCP_ATTR_MANDATORY, 0, &bfcp->bfcp_info.floor_id);

	ast_log(LOG_DEBUG, "bfcp_process_floor_release res=%d\n", res);

	return res;
}

static int ast_process_bfcp_msg_for_conference(struct ast_bfcp_udp *bfcp, int data_len)
{
	BFCP_MSG *bfcp_msg = NULL;
	int res = 0;
	enum bfcp_prim prim;
	unsigned int need_send_bfcp_msg = 0;
	struct ast_bfcp_info bfcp_info;

	struct bfcp_buffer bfcp_recv_buffer = bfcp->bfcp_recv_buffer;
	struct bfcp_buffer bfcp_send_buffer = bfcp->bfcp_send_buffer;

	bfcp_init_buffer(&bfcp_recv_buffer);
	bfcp_init_buffer(&bfcp_send_buffer);

	memset(&bfcp_info, 0, sizeof(bfcp_info));

	memcpy(bfcp_recv_buffer.buf, bfcp->rawdata, data_len);
	bfcp_recv_buffer.size = data_len;

	res = bfcp_msg_decode(&bfcp_msg, &bfcp_recv_buffer);

	if (res < 0)
	{
		ast_log(LOG_WARNING, "ast_process_bfcp_msg error:%d\n", res);
		goto  PROCESS_BFCP_MSG_CLEANUP;
	}

	prim = bfcp_msg->header.prim;
	if (prim != BFCP_HELLO && prim != BFCP_HELLO_ACK)
	{
		ast_log(LOG_DEBUG, "Received bfcp message header prim:%d for user_id:%d\n", prim, bfcp_msg->header.user_id);
	}

	switch (prim)
	{
		case BFCP_HELLO:
			res = bfcp_process_hello(bfcp_msg, &bfcp_send_buffer);
			need_send_bfcp_msg = 1;
			break;

		case BFCP_HELLO_ACK:
			res = bfcp_process_hello_ack(bfcp_msg);
			need_send_bfcp_msg = 0;
			break;

		case BFCP_FLOOR_REQUEST:
			ast_bfcp_set_info(bfcp->channel, &bfcp_info);
			res = bfcp_process_floor_request(bfcp, bfcp_msg, &bfcp_send_buffer);
			if (!res) {
				bfcp_info.status = BFCP_GRANTED;
				bfcp_info.primitive = BFCP_FLOOR_STATUS;
				strcpy(bfcp_info.info_conf_id, bfcp->bfcp_info.info_conf_id);
				strcpy(bfcp_info.floor_owner_id, bfcp->bfcp_info.account_id);
				strcpy(bfcp_info.msid, bfcp->bfcp_info.msid);
				ast_channel_set_opening_slides_video(bfcp->channel, 1);
				ast_indicate(bfcp->channel, AST_CONTROL_NOTIFY_OPEN_SLIDES_VIDEO);
				ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;
		case BFCP_FLOOR_RELEASE:
			ast_bfcp_set_info(bfcp->channel, &bfcp_info);
			res = bfcp_process_floor_release(bfcp, bfcp_msg, &bfcp_send_buffer);
			if (!res && (bfcp->bfcp_info.status == BFCP_RELEASED)) {
				bfcp_info.status = BFCP_RELEASED;
				bfcp_info.primitive = BFCP_FLOOR_STATUS;
				strcpy(bfcp_info.info_conf_id, bfcp->bfcp_info.info_conf_id);
				strcpy(bfcp_info.floor_owner_id, bfcp->bfcp_info.account_id);
				ast_indicate(bfcp->channel, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO);
				ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_REQUEST_STATUS:
			res = bfcp_process_floor_req_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info);
			bfcp_info.primitive = BFCP_FLOOR_STATUS;
			strcpy(bfcp_info.info_conf_id, bfcp->bfcp_info.info_conf_id);
			strcpy(bfcp_info.floor_owner_id, bfcp->bfcp_info.account_id);
			ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			AST_SCHED_DEL(bfcp->sched, bfcp->sched_id);
			bfcp->sending_msg = 0;
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_STATUS:
			ast_bfcp_set_info(bfcp->channel, &bfcp_info);
			strcpy(bfcp_info.info_conf_id, bfcp->bfcp_info.info_conf_id);
			res = bfcp_process_floor_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info, bfcp->channel);
			BFCP_ATTR *req_status = bfcp_attrs_find(&bfcp_msg->attr_list, BFCP_REQUEST_STATUS);
			if (req_status) {
				bfcp_info.status = req_status->v.reqstatus.status;
				bfcp_info.primitive = BFCP_FLOOR_STATUS;
				strcpy(bfcp_info.floor_owner_id, bfcp->bfcp_info.account_id);
				if (bfcp_info.status == BFCP_GRANTED) {
					ast_indicate(bfcp->channel, AST_CONTROL_NOTIFY_OPEN_SLIDES_VIDEO);
				} else if (bfcp_info.status == BFCP_RELEASED) {
					ast_indicate(bfcp->channel, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO);
				}
				ast_queue_control_data(bfcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_REQ_STATUS_ACK:
		case BFCP_FLOOR_STATUS_ACK:
			AST_SCHED_DEL(bfcp->sched, bfcp->sched_id);
			bfcp->sending_msg = 0;
			need_send_bfcp_msg = 0;
			break;

		default:
			need_send_bfcp_msg = 0;
			ast_log(LOG_DEBUG, "Can't process this bfcp message\n");
			break;
	}

	if (res < 0) {
		ast_log(LOG_WARNING, "bfcp process error\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	if (!need_send_bfcp_msg) {
		ast_log(LOG_DEBUG, "bfcp don't need to send message\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	if (bfcp_send_buffer.pos > 0 && !ast_sockaddr_isnull(&bfcp->them)) {
		if (bfcp->media_turn_relay) {
			if (bfcp->engine->turn_packet_send(bfcp, bfcp_send_buffer.buf, bfcp_send_buffer.pos)) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			}
		} else {
			if ((res = ast_sendto(bfcp->fd, bfcp_send_buffer.buf, bfcp_send_buffer.pos, 0, &bfcp->them)) < 0) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
				ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			}
		}
	}

PROCESS_BFCP_MSG_CLEANUP:
	bfcp_free_msg(bfcp_msg);

	return 1;
}


int ast_send_bfcp_msg(struct ast_bfcp_udp *bfcp)
{
	struct bfcp_buffer send_buffer;
	enum bfcp_prim  prim_type;
	int res = 0;

	if (!bfcp)
	{
		ast_log(LOG_DEBUG, "ast_send_bfcp_msg param is NULL\n");
		return -1;
	}

	bfcp_create_buffer(&send_buffer);
	bfcp_init_buffer(&send_buffer);

	prim_type = bfcp->bfcp_info.primitive;
	ast_log(LOG_DEBUG, "ast_send_bfcp_msg prim_type is %d\n", prim_type);

	switch (prim_type)
	{
		case BFCP_HELLO:
			res = bfcp_hello_msg_encode(&send_buffer, &bfcp->bfcp_info);
			break;

		case BFCP_FLOOR_REQUEST:
			res = bfcp_floor_request_msg_encode(&send_buffer, &bfcp->bfcp_info);
			break;

		case BFCP_FLOOR_RELEASE:
			res = bfcp_floor_release_msg_encode(&send_buffer, &bfcp->bfcp_info);
			break;

		case BFCP_FLOOR_REQUEST_STATUS:
			res = bfcp_floor_request_status_msg_encode(&send_buffer, &bfcp->bfcp_info);
			break;

		case BFCP_FLOOR_STATUS:
			res = bfcp_floor_status_msg_encode(&send_buffer, &bfcp->bfcp_info);
			break;

		default:
			break;
	}

	if (res < 0)
	{
		ast_log(LOG_DEBUG, "ast_send_bfcp_msg encode message error\n");
		bfcp_free_buffer(&send_buffer);
		return res;
	}

	if (send_buffer.pos > 0 && !ast_sockaddr_isnull(&bfcp->them)) {
		if (bfcp->media_turn_relay) {
			if (bfcp->engine->turn_packet_send(bfcp, send_buffer.buf, send_buffer.pos)) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			} else {
				bfcp_init_buffer(&bfcp->timeout_buffer);
				memcpy(bfcp->timeout_buffer.buf, send_buffer.buf, send_buffer.size);
				bfcp->timeout_buffer.size = send_buffer.size;
				bfcp->timeout_buffer.pos = send_buffer.pos;
				AST_SCHED_REPLACE_VARIABLE(bfcp->sched_id, bfcp->sched, BFCP_T1, send_bfcp_msg_timeout, bfcp, 1);
				bfcp->sending_msg = 1;
			}
		} else {
			if ((res = ast_sendto(bfcp->fd, send_buffer.buf, send_buffer.pos, 0, &bfcp->them)) < 0) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
				ast_sockaddr_stringify(&bfcp->them), strerror(errno));
			} else {
				bfcp_init_buffer(&bfcp->timeout_buffer);
				memcpy(bfcp->timeout_buffer.buf, send_buffer.buf, send_buffer.size);
				bfcp->timeout_buffer.size = send_buffer.size;
				bfcp->timeout_buffer.pos = send_buffer.pos;
				AST_SCHED_REPLACE_VARIABLE(bfcp->sched_id, bfcp->sched, BFCP_T1, send_bfcp_msg_timeout, bfcp, 1);
				bfcp->sending_msg = 1;
			}
		}
	}

	bfcp_free_buffer(&send_buffer);

	ast_log(LOG_DEBUG, "ast_send_bfcp_msg end\n");

	return res;
}


struct ast_frame *ast_bfcp_read(struct ast_bfcp_udp *bfcp)
{
	int res;
	struct ast_sockaddr addr;
	uint8_t *buf;

	if (NULL == bfcp) {
		ast_log(LOG_WARNING, "ast_bfcp_read bfcp is NULL");
		return &ast_null_frame;
	}

	buf = bfcp->rawdata;

	/* Cache where the header will go */
	res = ast_recvfrom(bfcp->fd,
			buf,
			sizeof(bfcp->rawdata),
			0,
			&addr);
	if (res < 0) {
		if (errno != EAGAIN) {
			ast_log(LOG_WARNING, "bfcp : read error: %s\n", strerror(errno));
		}
		ast_assert(errno != EBADF);
		return &ast_null_frame;
	}

	if (res > BFCP_MAX_BUFFER_LENTH) {
		ast_log(LOG_WARNING, "Receive error bfcp packet!");
		return &ast_null_frame;
	}

	ast_log(LOG_DEBUG, "ast_bfcp_read from %s, len:%d", ast_sockaddr_stringify(&bfcp->them), res);

	/* Ignore if the other side hasn't been given an address yet. */
	if (ast_sockaddr_isnull(&bfcp->them)) {
		return &ast_null_frame;
	}

	if (bfcp->nat) {
		/* Send to whoever sent to us */
		if (ast_sockaddr_cmp(&bfcp->them, &addr)) {
			ast_sockaddr_copy(&bfcp->them, &addr);
			ast_debug(1, "bfcp : NAT, Using address %s\n", ast_sockaddr_stringify(&bfcp->them));
		}
	}

	if (strlen(bfcp->bfcp_info.info_conf_id)) {
		ast_process_bfcp_msg_for_conference(bfcp, res);
	} else {
		ast_process_bfcp_msg(bfcp, res);
	}

	return &ast_null_frame;
}

int ast_bfcp_write(struct ast_bfcp_udp *s, struct ast_frame *f)
{
	unsigned int len = f->datalen;
	int res;
	/* if no max datagram size is provided, use default value */
	const int bufsize = BFCP_MAX_DATAGRAM_LIMIT;
	uint8_t buf[bufsize];

	memset(buf, 0, sizeof(buf));

	ast_log(LOG_DEBUG, "ast_bfcp_write begin send to %s", ast_sockaddr_stringify(&s->them));

	/* If we have no peer, return immediately */
	if (ast_sockaddr_isnull(&s->them)) {
		return 0;
	}

	/* If there is no data length, return immediately */
	if (f->datalen == 0)
		return 0;

	if (f->frametype != AST_FRAME_BFCP) {
		ast_log(LOG_WARNING, "BFCP can only send BFCP data.\n");
		return -1;
	}

	if (len > bufsize) {
		ast_log(LOG_WARNING, "bufsize is less than data len");
		return 0;
	}

	memcpy(buf, f->data.ptr, len);

	if ((signed int) len > 0 && !ast_sockaddr_isnull(&s->them)) {
		if (s->media_turn_relay) {
			if (s->engine->turn_packet_send(s, buf, len)) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&s->them), strerror(errno));
			}
		} else {
			if ((res = ast_sendto(s->fd, buf, len, 0, &s->them)) < 0) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&s->them), strerror(errno));
			}
		}
	}

	return 0;
}

int ast_bfcp_setqos(struct ast_bfcp_udp *bfcp, unsigned int tos, unsigned int cos)
{
	ast_log(LOG_DEBUG, "ast_bfcp_setqos tos:%d, cos:%d\n", tos, cos);
	return ast_set_qos(bfcp->fd, tos, cos, "BFCP");
}

int ast_bfcp_get_and_cmp_requested_target_address(struct ast_bfcp_udp *bfcp,
		struct ast_sockaddr *address)
{
	if (!bfcp || !address) {
		ast_log(LOG_WARNING, "bfcp cmp requested target address, param is null");
		return 0;
	}
	if (ast_sockaddr_cmp(address, &bfcp->them) != 0) {
		ast_sockaddr_copy(address, &bfcp->them);
		return 1;
	}

	return 0;
}

int ast_bfcp_tcp_get_and_cmp_requested_target_address(struct ast_bfcp_tcp *bfcp_tcp,
		struct ast_sockaddr *address)
{
	if (!bfcp_tcp || !address) {
		ast_log(LOG_WARNING, "bfcp cmp requested target address, param is null");
		return 0;
	}
	if (ast_sockaddr_cmp(address, &bfcp_tcp->bfcp_tcp_desc.remote_address) != 0) {
		ast_sockaddr_copy(address, &bfcp_tcp->bfcp_tcp_desc.remote_address);
		return 1;
	}

	return 0;
}

void ast_bfcp_tcp_get_us(const struct ast_bfcp_tcp *bfcp_tcp, struct ast_sockaddr *us)
{
	if (NULL == bfcp_tcp || NULL == us) {
		ast_log(LOG_WARNING, "bfcp is NULL, get us failed");
		return;
	}
	ast_sockaddr_copy(us, &bfcp_tcp->bfcp_tcp_desc.local_address);
}

void ast_bfcp_tcp_set_peer(struct ast_bfcp_tcp *bfcp_tcp, const struct ast_sockaddr *them)
{
	if (NULL == bfcp_tcp || NULL == them) {
		ast_log(LOG_WARNING, "bfcp is NULL, set peer failed");
		return;
	}
	ast_sockaddr_copy(&bfcp_tcp->bfcp_tcp_desc.remote_address, them);
	ast_sockaddr_copy(&bfcp_tcp->remote_address, them);
}

static void ast_bfcp_tcptls_close_session_file(struct ast_bfcp_tcptls_session_instance *tcptls_session)
{
	if (tcptls_session->fd != -1) {
		/*
		 * Issuing shutdown() is necessary here to avoid a race
		 * condition where the last data written may not appear
		 * in the TCP stream.  See ASTERISK-23548
		 */
		shutdown(tcptls_session->fd, SHUT_RDWR);
		if (close(tcptls_session->fd)) {
			ast_log(LOG_ERROR, "close() failed: %s\n", strerror(errno));
		}
		tcptls_session->fd = -1;
	} else {
		ast_log(LOG_DEBUG, "ast_bfcp_tcptls_close_session_file invoked on session instance without file or file descriptor\n");
	}
}

static void bfcp_session_instance_destructor(void *obj)
{
	struct ast_bfcp_tcptls_session_instance *tcptls_session = obj;

	ast_log(LOG_DEBUG, "bfcp_session_instance_destructor begin\n");

	if (!tcptls_session) {
		return;
	}

	if (tcptls_session->bfcp_tcp) {
		tcptls_session->bfcp_tcp->tcptls_session_destoryed = 1;
		ao2_ref(tcptls_session->bfcp_tcp, -1);
	}
	
	if (tcptls_session->stream_cookie) {
		ast_log(LOG_DEBUG, "bfcp_session_instance_destructor Destroying stream\n");
		ao2_t_ref(tcptls_session->stream_cookie, -1, "Destroying tcptls session instance");
		tcptls_session->stream_cookie = NULL;
	}

	ast_free(tcptls_session->overflow_buf);
	tcptls_session->overflow_buf = NULL;

	ao2_cleanup(tcptls_session->private_data);
	tcptls_session->private_data = NULL;

	return;
}

static HOOK_T bfcp_tcptls_stream_read(void *cookie, char *buf, LEN_T size)
{
	struct ast_bfcp_tcptls_stream *stream = cookie;
	struct timeval start;
	int ms;
	int res;

	if (!size) {
		/* You asked for no data you got no data. */
		return 0;
	}

	if (!stream || stream->fd == -1) {
		errno = EBADF;
		return -1;
	}

	if (stream->start.tv_sec) {
		start = stream->start;
	} else {
		start = ast_tvnow();
	}

	for (;;) {
		res = read(stream->fd, buf, size);
		if (0 <= res || !stream->exclusive_input) {
			/* Got data or we cannot wait for it. */
			return res;
		}
		if (errno != EINTR && errno != EAGAIN) {
			/* Not a retryable error. */
			ast_debug(1, "TCP socket error reading data: %s\n",
				strerror(errno));
			return -1;
		}
		ms = ast_remaining_ms(start, stream->timeout);
		if (!ms) {
			/* Report EOF for a timeout */
			ast_debug(1, "TCP timeout reading data\n");
			return 0;
		}
		ast_wait_for_input(stream->fd, ms);
	}
}

static HOOK_T bfcp_tcptls_stream_write(void *cookie, const char *buf, LEN_T size)
{
	struct ast_bfcp_tcptls_stream *stream = cookie;
	struct timeval start;
	int ms;
	int res;
	int written;
	int remaining;

	ast_log(LOG_DEBUG, "bfcp_tcptls_stream_write date len:%ld\n", size);

	if (!size) {
		/* You asked to write no data you wrote no data. */
		return 0;
	}

	if (!stream || stream->fd == -1) {
		errno = EBADF;
		return -1;
	}

	if (stream->start.tv_sec) {
		start = stream->start;
	} else {
		start = ast_tvnow();
	}

	written = 0;
	remaining = size;
	for (;;) {
		res = write(stream->fd, buf + written, remaining);
		if (res == remaining) {
			/* Yay everything was written. */
			return size;
		}
		if (0 < res) {
			/* Successfully wrote part of the buffer.  Try to write the rest. */
			written += res;
			remaining -= res;
			continue;
		}
		if (errno != EINTR && errno != EAGAIN) {
			/* Not a retryable error. */
			ast_debug(1, "TCP socket error writing: %s\n", strerror(errno));
			if (written) {
				return written;
			}
			return -1;
		}
		ms = ast_remaining_ms(start, stream->timeout);
		if (!ms) {
			/* Report partial write. */
			ast_debug(1, "TCP timeout writing data\n");
			return written;
		}
		ast_wait_for_output(stream->fd, ms);
	}
}

static void bfcp_tcptls_stream_dtor(void *cookie)
{
#ifdef AST_DEVMODE
	/* Since the ast_assert below is the only one using stream,
	 * and ast_assert is only available with AST_DEVMODE, we
	 * put this in a conditional to avoid compiler warnings. */
	struct ast_bfcp_tcptls_stream *stream = cookie;
#endif

	ast_assert(stream->fd == -1);
}

static struct ast_bfcp_tcptls_stream *bfcp_tcptls_stream_alloc(void)
{
	struct ast_bfcp_tcptls_stream *stream;

	stream = ao2_alloc_options(sizeof(*stream), bfcp_tcptls_stream_dtor,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (stream) {
		stream->fd = -1;
		stream->timeout = -1;
	}
	return stream;
}

void ast_bfcp_tcptls_server_start(struct ast_bfcp_tcp *bfcp_tcp)
{
	int flags;
	int x = 1;
	struct ast_tcptls_session_args *desc = &bfcp_tcp->bfcp_tcp_desc;

	int bfcp_tcp_start = ast_get_bfcp_tcp_start();
	int bfcp_tcp_end = ast_get_bfcp_tcp_end();
	/* Do nothing if nothing has changed */
	if (!ast_sockaddr_cmp(&desc->old_address, &desc->local_address)) {
		ast_debug(1, "Nothing changed bfcp tcp in %s\n", desc->name);
		return;
	}

	/* If we return early, there is no one listening */
	ast_sockaddr_setnull(&desc->old_address);

	/* Shutdown a running server if there is one */
	if (desc->master != AST_PTHREADT_NULL) {
		pthread_cancel(desc->master);
		pthread_kill(desc->master, SIGURG);
		pthread_join(desc->master, NULL);
	}

	if (desc->accept_fd != -1) {
		close(desc->accept_fd);
	}

	/* If there's no new server, stop here */
	if (ast_sockaddr_isnull(&desc->local_address)) {
		ast_debug(2, "Server disabled:  %s\n", desc->name);
		return;
	}

	desc->accept_fd = socket(ast_sockaddr_is_ipv6(&desc->local_address) ?
				 AF_INET6 : AF_INET, SOCK_STREAM, 0);
	if (desc->accept_fd < 0) {
		ast_log(LOG_ERROR, "Unable to allocate socket for %s: %s\n", desc->name, strerror(errno));
		return;
	}

	ast_log(LOG_DEBUG, "Bfcp tcptls server start socket fd:%d\n", desc->accept_fd);

	setsockopt(desc->accept_fd, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));

	int bfcp_tcp_port = (bfcp_tcp_start == bfcp_tcp_end) ? bfcp_tcp_start : (ast_random() % (bfcp_tcp_end - bfcp_tcp_start)) + bfcp_tcp_start;
	bfcp_tcp_port = bfcp_tcp_port & ~1;
	int startplace = bfcp_tcp_port;
	for (;;) {
		ast_sockaddr_set_port(&desc->local_address, bfcp_tcp_port);
		if (ast_bind(desc->accept_fd, &desc->local_address) == 0) {
			break;
		}

		bfcp_tcp_port++;

		if (bfcp_tcp_port > bfcp_tcp_end) {
			bfcp_tcp_port = (bfcp_tcp_start + 1) & ~1;
		}

		if (bfcp_tcp_port == startplace || (errno != EADDRINUSE && errno != EACCES)) {
			ast_log(LOG_ERROR, "Oh dear... we couldn't allocate a port for TCP BFCP '%p'\n", bfcp_tcp);
			goto error;
		}
	}

	if (listen(desc->accept_fd, 10)) {
		ast_log(LOG_ERROR, "Unable to listen for %s!\n", desc->name);
		goto error;
	}
	flags = fcntl(desc->accept_fd, F_GETFL);
	fcntl(desc->accept_fd, F_SETFL, flags | O_NONBLOCK);
	ao2_ref(bfcp_tcp, +1);
	if (ast_pthread_create_background(&desc->master, NULL, desc->accept_fn, bfcp_tcp)) {
		ast_log(LOG_ERROR, "Unable to launch thread for %s on %s: %s\n",
			desc->name,
			ast_sockaddr_stringify(&desc->local_address),
			strerror(errno));
		ao2_ref(bfcp_tcp, -1);
		goto error;
	}

	/* Set current info */
	ast_sockaddr_copy(&desc->old_address, &desc->local_address);

	return;

error:
	close(desc->accept_fd);
	desc->accept_fd = -1;
}

void *handle_bfcp_tcptls_connection(void *data)
{
	struct ast_bfcp_tcp *bfcp_tcp = data;
	struct ast_bfcp_tcptls_session_instance *tcptls_session = bfcp_tcp->tcptls_session;

	if (!tcptls_session->stream_cookie) {
		tcptls_session->stream_cookie = bfcp_tcptls_stream_alloc();
		if (!tcptls_session->stream_cookie) {
			ast_bfcp_tcptls_close_session_file(tcptls_session);
			ao2_ref(tcptls_session, -1);
			return NULL;
		}
		tcptls_session->stream_cookie->fd = tcptls_session->fd;
		tcptls_session->stream_cookie->timeout = -1;
	}

	if (tcptls_session->parent->worker_fn) {
		return tcptls_session->parent->worker_fn(bfcp_tcp);
	} else {
		return tcptls_session;
	}
}

void *ast_bfcp_tcptls_server_root(void *data)
{
	struct ast_bfcp_tcp *bfcp_tcp = data;
	struct ast_tcptls_session_args *desc = &bfcp_tcp->bfcp_tcp_desc;
	int fd;
	struct ast_sockaddr addr;
	struct ast_bfcp_tcptls_session_instance *tcptls_session;
	pthread_t launched;

	for (;;) {
		int i, flags;

		if (desc->periodic_fn) {
			desc->periodic_fn(desc);
		}
		i = ast_wait_for_input(desc->accept_fd, desc->poll_timeout);
		if (i <= 0) {
			continue;
		}
		fd = ast_accept(desc->accept_fd, &addr);
		if (fd < 0) {
			if ((errno != EAGAIN) && (errno != EWOULDBLOCK) && (errno != EINTR) && (errno != ECONNABORTED)) {
				ast_log(LOG_ERROR, "BFCP TCP accept failed: %s\n", strerror(errno));
				break;
			}
			/* Prevent tight loop from hogging CPU */
			usleep(1);
			continue;
		}

		ast_log(LOG_DEBUG, "Bfcp tcptls server root accept fd:%d\n", fd);

		tcptls_session = ao2_alloc(sizeof(*tcptls_session), bfcp_session_instance_destructor);
		if (!tcptls_session) {
			ast_log(LOG_WARNING, "No memory for new bfcp tcp session: %s\n", strerror(errno));
			if (close(fd)) {
				ast_log(LOG_ERROR, "close() bfcp tcp failed: %s\n", strerror(errno));
			}
			continue;
		}

		tcptls_session->overflow_buf = ast_str_create(128);
		flags = fcntl(fd, F_GETFL);
		fcntl(fd, F_SETFL, flags & ~O_NONBLOCK);
		tcptls_session->fd = fd;
		tcptls_session->parent = desc;
		ast_sockaddr_copy(&tcptls_session->remote_address, &addr);
		tcptls_session->client = 0;
		bfcp_tcp->tcptls_session = tcptls_session;
		bfcp_tcp->tcptls_session_destoryed = 0;
		ao2_ref(bfcp_tcp, +1);
		tcptls_session->bfcp_tcp = bfcp_tcp;

		ast_log(LOG_DEBUG,"ast_bfcp_tcptls_server_root tcptls_session:%p\n", tcptls_session);

		/* This thread is now the only place that controls the single ref to tcptls_session */
		if (ast_pthread_create_detached_background(&launched, NULL, handle_bfcp_tcptls_connection, bfcp_tcp)) {
			ast_log(LOG_ERROR, "Unable to launch helper thread: %s\n", strerror(errno));
			ast_bfcp_tcptls_close_session_file(tcptls_session);
			ao2_ref(tcptls_session, -1);
		}
	}

	ast_log(LOG_ERROR, "BFCP TCP listener thread ended abnormally\n");

	/* Close the listener socket so Asterisk doesn't appear dead. */
	fd = desc->accept_fd;
	desc->accept_fd = -1;
	if (0 <= fd) {
		close(fd);
	}

	return NULL;
}

struct ast_bfcp_tcptls_session_instance *ast_bfcp_tcptls_client_start(struct ast_bfcp_tcp *bfcp_tcp)
{
	struct ast_bfcp_tcptls_session_instance *tcptls_session = bfcp_tcp->tcptls_session;
	struct ast_tcptls_session_args *desc;
	int flags;

	if (!(desc = tcptls_session->parent)) {
		goto client_start_error;
	}

	ast_log(LOG_DEBUG, "ast_bfcp_tcptls_client_start connnect\n");

	if (ast_connect(desc->accept_fd, &desc->remote_address)) {
		ast_log(LOG_ERROR, "Unable to connect %s to %s: %s\n",
			desc->name,
			ast_sockaddr_stringify(&desc->remote_address),
			strerror(errno));
		goto client_start_error;
	}

	flags = fcntl(desc->accept_fd, F_GETFL);
	fcntl(desc->accept_fd, F_SETFL, flags & ~O_NONBLOCK);

	return handle_bfcp_tcptls_connection(bfcp_tcp);

client_start_error:
	if (desc) {
		close(desc->accept_fd);
		desc->accept_fd = -1;
	}
	ao2_ref(tcptls_session, -1);
	tcptls_session = NULL;

	return NULL;
}

struct ast_bfcp_tcptls_session_instance *ast_bfcp_tcptls_client_create(struct ast_tcptls_session_args *desc)
{
	int x = 1;
	struct ast_bfcp_tcptls_session_instance *tcptls_session = NULL;

	ast_log(LOG_DEBUG, "ast_bfcp_tcptls_client_create begin\n");

	int bfcp_tcp_start = ast_get_bfcp_tcp_start();
	int bfcp_tcp_end = ast_get_bfcp_tcp_end();
	/* Do nothing if nothing has changed */
	if (!ast_sockaddr_cmp(&desc->old_address, &desc->remote_address)) {
		ast_log(LOG_DEBUG, "Nothing changed in %s\n", desc->name);
		return NULL;
	}

	/* If we return early, there is no connection */
	ast_sockaddr_setnull(&desc->old_address);

	if (desc->accept_fd != -1) {
		close(desc->accept_fd);
	}

	desc->accept_fd = socket(ast_sockaddr_is_ipv6(&desc->remote_address) ?
				 AF_INET6 : AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (desc->accept_fd < 0) {
		ast_log(LOG_ERROR, "Unable to allocate socket for %s: %s\n",
			desc->name, strerror(errno));
		return NULL;
	}

	ast_log(LOG_DEBUG, "ast_bfcp_tcptls_client_create fd:%d\n", desc->accept_fd);

	/* if a local address was specified, bind to it so the connection will
	   originate from the desired address */
	if (!ast_sockaddr_isnull(&desc->local_address)) {
		setsockopt(desc->accept_fd, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));

		int bfcp_tcp_port = (bfcp_tcp_start == bfcp_tcp_end) ? bfcp_tcp_start : (ast_random() % (bfcp_tcp_end - bfcp_tcp_start)) + bfcp_tcp_start;
		int startplace = bfcp_tcp_port;
		int cnt = 0;
		for (;;) {
			ast_sockaddr_set_port(&desc->local_address, bfcp_tcp_port);
			if (ast_bind(desc->accept_fd, &desc->local_address) == 0) {
				break;
			}
			if (errno != EADDRINUSE) {
				ast_log(LOG_WARNING, "Unexpected bfcp bind error: %s\n", strerror(errno));
				goto error;
			}

			++bfcp_tcp_port;
			cnt++;

			if (bfcp_tcp_port > bfcp_tcp_end)
				bfcp_tcp_port = bfcp_tcp_start;
			if (bfcp_tcp_port == startplace) {
				ast_log(LOG_WARNING, "No bfcp ports remaining\n");
				goto error;
			}
			if (cnt > 10) {
				ast_log(LOG_ERROR, "Could't allocate a port for bfcp\n");
				goto error;
			}
		}

		ast_log(LOG_DEBUG, "ast_bfcp_tcptls_client_create port:%d\n", bfcp_tcp_port);
	}

	if (!(tcptls_session = ao2_alloc(sizeof(*tcptls_session), bfcp_session_instance_destructor))) {
		ast_log(LOG_ERROR, "Could't allocate tcptls_session\n");
		goto error;
	}

	tcptls_session->overflow_buf = ast_str_create(128);
	tcptls_session->client = 1;
	tcptls_session->fd = desc->accept_fd;
	tcptls_session->parent = desc;
	tcptls_session->parent->worker_fn = NULL;
	ast_sockaddr_copy(&tcptls_session->remote_address, &desc->remote_address);

	/* Set current info */
	ast_sockaddr_copy(&desc->old_address, &desc->remote_address);

	ast_log(LOG_DEBUG,"ast_bfcp_tcptls_client_create tcptls_session:%p\n", tcptls_session);

	return tcptls_session;
error:
	close(desc->accept_fd);
	desc->accept_fd = -1;
	if (tcptls_session) {
		ao2_ref(tcptls_session, -1);
	}
	return NULL;
}

int ast_bfcp_tcptls_server_read(struct ast_bfcp_tcptls_session_instance *tcptls_session, void *buf, int count)
{
	if (!tcptls_session) {
		ast_log(LOG_ERROR, "TCP/TLS read called on invalid session.\n");
		errno = EIO;
		return -1;
	}
	if (tcptls_session->fd == -1) {
		ast_log(LOG_ERROR, "TCP/TLS read called on invalid fd.\n");
		errno = EIO;
		return -1;
	}
	if (!tcptls_session->stream_cookie) {
		ast_log(LOG_ERROR, "TCP/TLS read called on invalid stream.\n");
		errno = EIO;
		return -1;
	}
	if (tcptls_session->stream_cookie->fd == -1) {
		ast_log(LOG_ERROR, "TCP/TLS read called on invalid stream.\n");
		errno = EIO;
		return -1;
	}

	return bfcp_tcptls_stream_read(tcptls_session->stream_cookie, buf, count);
}

int ast_bfcp_tcptls_server_write(struct ast_bfcp_tcptls_session_instance *tcptls_session, const void *buf, int count)
{
	if (!tcptls_session) {
		ast_log(LOG_ERROR, "TCP/TLS write called on invalid session.\n");
		errno = EIO;
		return -1;
	}
	if (tcptls_session->fd == -1) {
		ast_log(LOG_ERROR, "TCP/TLS write called on invalid fd.\n");
		errno = EIO;
		return -1;
	}
	if (!tcptls_session->stream_cookie || tcptls_session->stream_cookie->fd == -1) {
		ast_log(LOG_ERROR, "TCP/TLS write called on invalid stream.\n");
		errno = EIO;
		return -1;
	}

	return bfcp_tcptls_stream_write(tcptls_session->stream_cookie, buf, count);
}

void bfcp_init_tcp_desc(struct ast_tcptls_session_args *bfcp_tcp_desc)
{
	bfcp_tcp_desc->accept_fd = -1;
	bfcp_tcp_desc->master = AST_PTHREADT_NULL;
	bfcp_tcp_desc->tls_cfg = NULL;
	bfcp_tcp_desc->poll_timeout = -1;
	bfcp_tcp_desc->name = "BFCP TCP server";
	bfcp_tcp_desc->accept_fn = ast_bfcp_tcptls_server_root;
	bfcp_tcp_desc->worker_fn = bfcp_tcp_worker_fn;

	return;
}

static void bfcp_tcp_destructor(void *obj)
{
	struct ast_bfcp_tcp *bfcp_tcp = obj;

	ast_log(LOG_DEBUG, "bfcp_tcp_destructor begin\n");

	if (!bfcp_tcp) {
		ast_log(LOG_DEBUG, "bfcp_tcp_destructor is already destory\n");
		return;
	}

	if (bfcp_tcp->channel) {
		ast_channel_unref(bfcp_tcp->channel);
	}

	return;
}

struct ast_bfcp_tcp *ast_bfcp_tcp_new_with_bindaddr(struct ast_sockaddr *addr)
{
	struct ast_bfcp_tcp *bfcp_tcp = NULL;

	ast_log(LOG_DEBUG, "ast_bfcp_tcp_new_with_bindaddr begin\n");

	if (!(bfcp_tcp = ao2_alloc(sizeof(*bfcp_tcp), bfcp_tcp_destructor))) {
		ast_log(LOG_WARNING, "Unable to alloc bfcp tcp\n");
		return NULL;
	}

	/* Start TCP server */
	bfcp_init_tcp_desc(&bfcp_tcp->bfcp_tcp_desc);
	if (ast_sockaddr_isnull(&bfcp_tcp->bfcp_tcp_desc.local_address)) {
		ast_sockaddr_copy(&bfcp_tcp->bfcp_tcp_desc.local_address, addr);
	}

	ast_bfcp_tcptls_server_start(bfcp_tcp);
	if (bfcp_tcp->bfcp_tcp_desc.accept_fd == -1) {
		/* TCP server start failed. Tell the admin */
		ast_log(LOG_ERROR, "BFCP TCP Server start failed. Not listening on TCP socket.\n");
	} else {
		ast_log(LOG_DEBUG, "BFCP TCP server started\n");
	}

	return bfcp_tcp;
}

static int bfcp_prepare_local_socket(struct ast_bfcp_tcp *bfcp_tcp)
{
	static const char name[] = "BFCP TCP client";
	struct ast_tcptls_session_args *client_args = NULL;
	pthread_t launched;

	if (!bfcp_tcp->tcptls_session_destoryed &&
		(bfcp_tcp->tcptls_session) &&
		(bfcp_tcp->tcptls_session->fd != -1)) {
		ast_log(LOG_DEBUG, "Bfcp tcp client already have fd now:%p %d\n", bfcp_tcp->tcptls_session, bfcp_tcp->tcptls_session->fd);
		return bfcp_tcp->tcptls_session->fd;
	}

	ast_log(LOG_DEBUG, "bfcp_prepare_local_socket begin\n");

	/* create new session arguments for the client connection */
	if (!(client_args = ao2_alloc(sizeof(*client_args), bfcp_tcptls_client_args_destructor)) ||
		!(client_args->name = ast_strdup(name))) {
		goto create_tcptls_session_fail;
	}

	client_args->accept_fd = -1;
	ast_sockaddr_copy(&client_args->remote_address, &bfcp_tcp->remote_address);

	ast_log(LOG_DEBUG, "BFCP old remote at port %s\n", ast_sockaddr_stringify(&client_args->old_address));
	ast_log(LOG_DEBUG, "BFCP peer remote at port %s\n", ast_sockaddr_stringify(&bfcp_tcp->remote_address));

	/* Create a client connection for address, this does not start the connection, just sets it up. */
	if (!(bfcp_tcp->tcptls_session = ast_bfcp_tcptls_client_create(client_args))) {
		goto create_tcptls_session_fail;
	}

	if (!(bfcp_tcp->tcptls_session = ast_bfcp_tcptls_client_start(bfcp_tcp))) {
		ast_log(LOG_DEBUG, "bfcp tcp client start failed\n");
		goto create_tcptls_session_fail;
	}

	if (bfcp_tcp->tcptls_session) {
		ao2_ref(bfcp_tcp, +1);
		bfcp_tcp->tcptls_session->bfcp_tcp = bfcp_tcp;
		bfcp_tcp->tcptls_session_destoryed = 0;
	}

	if (ast_pthread_create_detached_background(&launched, NULL, bfcp_tcp_worker_fn, bfcp_tcp)) {
		ast_debug(1, "Unable to launch bfcp tcp '%s'.", client_args->name);
		goto create_tcptls_session_fail;
	}

	return bfcp_tcp->tcptls_session->fd;

create_tcptls_session_fail:
	ast_log(LOG_DEBUG, "bfcp_prepare_local_socket cleanup tcptls_session begin");
	if (client_args) {
		ao2_t_ref(client_args, -1, "failed to create client, getting rid of client tcptls_session arguments");
	}
	if (bfcp_tcp->tcptls_session) {
		ast_log(LOG_DEBUG,"bfcp_prepare_local_socket cleanup:%p", bfcp_tcp->tcptls_session);
		ao2_lock(bfcp_tcp->tcptls_session);
		ast_bfcp_tcptls_close_session_file(bfcp_tcp->tcptls_session);
		bfcp_tcp->tcptls_session->parent = NULL;
		ao2_unlock(bfcp_tcp->tcptls_session);

		ao2_ref(bfcp_tcp->tcptls_session, -1);
		bfcp_tcp->tcptls_session = NULL;
	}

	return -1;
}

void bfcp_tcptls_client_args_destructor(void *obj)
{
	struct ast_tcptls_session_args *args = obj;
	ast_log(LOG_DEBUG, "bfcp_tcptls_client_args_destructor begin:%p\n", args);
	if (args->tls_cfg) {
		ast_free(args->tls_cfg->certfile);
		ast_free(args->tls_cfg->pvtfile);
		ast_free(args->tls_cfg->cipher);
		ast_free(args->tls_cfg->cafile);
		ast_free(args->tls_cfg->capath);

		ast_ssl_teardown(args->tls_cfg);
	}
	ast_free(args->tls_cfg);
	ast_free((char *) args->name);
}

static int ast_process_bfcp_tcp_msg(struct ast_bfcp_tcp *bfcp_tcp, char *buf, int data_len)
{
	BFCP_MSG *bfcp_msg = NULL;
	int res = 0;
	enum bfcp_prim prim;
	unsigned int need_send_bfcp_msg = 0;
	struct ast_bfcp_info bfcp_info;

	struct bfcp_buffer bfcp_recv_buffer;
	struct bfcp_buffer bfcp_send_buffer;

	if (!bfcp_tcp || !buf || !bfcp_tcp->channel) {
		ast_log(LOG_ERROR, "ast_process_bfcp_tcp_msg param is NULL\n");
		return -1;
	}

	bfcp_create_buffer(&bfcp_recv_buffer);
	bfcp_create_buffer(&bfcp_send_buffer);
	bfcp_init_buffer(&bfcp_recv_buffer);
	bfcp_init_buffer(&bfcp_send_buffer);

	memcpy(bfcp_recv_buffer.buf, buf, data_len);
	bfcp_recv_buffer.size = data_len;

	res = bfcp_msg_decode(&bfcp_msg, &bfcp_recv_buffer);

	if (res < 0)
	{
		ast_log(LOG_WARNING, "ast_process_bfcp_msg error:%d\n", res);
		goto  PROCESS_BFCP_MSG_CLEANUP;
	}

	prim = bfcp_msg->header.prim;
	if (prim != BFCP_HELLO && prim != BFCP_HELLO_ACK)
	{
		ast_log(LOG_DEBUG, "Received bfcp message header prim:%d for user_id:%d\n", prim, bfcp_msg->header.user_id);
	}

	bfcp_tcp->bfcp_info.primitive = bfcp_msg->header.prim;

	memset(&bfcp_info, 0, sizeof(bfcp_info));
	bfcp_info.primitive = prim;

	switch (prim)
	{
		case BFCP_HELLO:
			res = bfcp_process_hello(bfcp_msg, &bfcp_send_buffer);
			need_send_bfcp_msg = 1;
			break;

		case BFCP_HELLO_ACK:
			res = bfcp_process_hello_ack(bfcp_msg);
			need_send_bfcp_msg = 0;
			break;

		case BFCP_FLOOR_REQUEST:
		case BFCP_FLOOR_RELEASE:
			ast_bfcp_set_info(bfcp_tcp->channel, &bfcp_info);
			ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			need_send_bfcp_msg = 0;
			break;

		case BFCP_FLOOR_REQUEST_STATUS:
			res = bfcp_process_floor_req_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info);
			ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_STATUS:
			ast_bfcp_set_info(bfcp_tcp->channel, &bfcp_info);
			res = bfcp_process_floor_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info, bfcp_tcp->channel);
			ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_REQ_STATUS_ACK:
		case BFCP_FLOOR_STATUS_ACK:
			need_send_bfcp_msg = 0;
			break;

		default:
			need_send_bfcp_msg = 0;
			ast_log(LOG_DEBUG, "Can't process this bfcp message\n");
			break;
	}

	if (res < 0)
	{
		ast_log(LOG_WARNING, "bfcp process error\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	ast_bfcp_set_channel_slides_status(bfcp_tcp->channel, &bfcp_info, 1);

	if (!need_send_bfcp_msg)
	{
		ast_log(LOG_DEBUG, "bfcp don't need to send message\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	if (bfcp_send_buffer.pos > 0) {
		ao2_lock(bfcp_tcp->tcptls_session);
		ast_bfcp_tcptls_server_write(bfcp_tcp->tcptls_session, bfcp_send_buffer.buf, bfcp_send_buffer.pos);
		ao2_unlock(bfcp_tcp->tcptls_session);
	}

PROCESS_BFCP_MSG_CLEANUP:
	bfcp_free_msg(bfcp_msg);

	bfcp_free_buffer(&bfcp_recv_buffer);
	bfcp_free_buffer(&bfcp_send_buffer);

	return 1;
}

static int bfcp_tcp_process_floor_request(struct ast_bfcp_tcp *bfcp_tcp, BFCP_MSG *bfcpMsg,
	struct bfcp_buffer *sendBuffer)
{
    int res = 0;
    struct bfcp_reqstatus_struct req_status;

    if (!bfcp_tcp || !bfcpMsg || !sendBuffer)
    {
        ast_log(LOG_WARNING, "bfcp_process_floor_request, param is NULL\n");
        return -1;
    }

    if (bfcp_conf_funcs.conf_get_slides_video_resource) {
		if (VIDEO_RESOLUTION_1080P == bfcp_conf_funcs.conf_get_slides_video_resource(bfcp_tcp->bfcp_info.info_conf_id, bfcp_tcp->channel, ast_true(bfcp_tcp->bfcp_info.share_audio))) {
			ast_log(LOG_DEBUG, "Get slides video resource sucess for conf:%s, channel:%s\n", bfcp_tcp->bfcp_info.info_conf_id, ast_channel_name(bfcp_tcp->channel));
			req_status.status = BFCP_GRANTED;
			bfcp_tcp->bfcp_info.status = BFCP_GRANTED;
			set_bfcp_info_msid(bfcp_tcp->channel, &bfcp_tcp->bfcp_info);
			set_bfcp_slider_name(bfcp_tcp->channel, &bfcp_tcp->bfcp_info);
		} else {
			req_status.status = BFCP_DENIED;
			bfcp_tcp->bfcp_info.status = BFCP_DENIED;
			res = -1;
		}
	} else {
		ast_log(LOG_DEBUG, "Has no registed func conf_get_slides_video_resource\n");
		req_status.status = BFCP_GRANTED;
		bfcp_tcp->bfcp_info.status = BFCP_GRANTED;
	}

    ast_log(LOG_DEBUG, "bfcp_process_floor_request send the BFCP_FLOOR_REQUEST_STATUS, reqStatus.status:%s\n",
        bfcp_req_status_2str(req_status.status));
    req_status.qpos = 0;

    res = bfcp_msg_encode(sendBuffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQUEST_STATUS,
                           bfcp_tcp->bfcp_info.conf_id, bfcpMsg->header.transaction_id, bfcp_tcp->bfcp_info.user_id,
                           1, BFCP_FLOOR_REQ_INFO | BFCP_ATTR_MANDATORY, 2, &bfcp_tcp->bfcp_info.floor_requst_id,
                           BFCP_OVERALL_REQ_STATUS | BFCP_ATTR_MANDATORY, 1, &bfcp_tcp->bfcp_info.floor_requst_id,
                           BFCP_REQUEST_STATUS | BFCP_ATTR_MANDATORY, 0, &req_status,
                           BFCP_FLOOR_REQ_STATUS | BFCP_ATTR_MANDATORY, 0, &bfcp_tcp->bfcp_info.floor_id);

    ast_log(LOG_DEBUG, "bfcp_process_floor_request res=%d\n", res);

    return res;
}

static int bfcp_tcp_process_floor_release(struct ast_bfcp_tcp *bfcp_tcp, BFCP_MSG *bfcpMsg,
	struct bfcp_buffer *sendBuffer)
{
    int res = 0;
    struct bfcp_reqstatus_struct req_status;

    if (!bfcp_tcp || !bfcpMsg || !sendBuffer)
    {
        ast_log(LOG_WARNING, "bfcp_process_floor_release, param is NULL\n");
        return -1;
    }

	if (bfcp_conf_funcs.conf_put_slides_video_resource && bfcp_conf_funcs.conf_get_slider_channel && bfcp_conf_funcs.conf_get_sfu_moderator) {
		bfcp_release_floor_use_conference_interface(&bfcp_tcp->bfcp_info, bfcp_tcp->channel);
		req_status.status = bfcp_tcp->bfcp_info.status;
	} else {
		req_status.status = BFCP_RELEASED;
		bfcp_tcp->bfcp_info.status = BFCP_RELEASED;
	}

    ast_log(LOG_DEBUG, "bfcp_process_floor_release send the BFCP_FLOOR_REQUEST_STATUS, reqStatus.status:%s\n",
        bfcp_req_status_2str(req_status.status));
    req_status.qpos = 0;

    res = bfcp_msg_encode(sendBuffer, BFCP_VER1, BFCP_RESERVED, BFCP_FLOOR_REQUEST_STATUS,
                           bfcp_tcp->bfcp_info.conf_id, bfcpMsg->header.transaction_id, bfcp_tcp->bfcp_info.user_id,
                           1, BFCP_FLOOR_REQ_INFO | BFCP_ATTR_MANDATORY, 2, &bfcp_tcp->bfcp_info.floor_requst_id,
                           BFCP_OVERALL_REQ_STATUS | BFCP_ATTR_MANDATORY, 1, &bfcp_tcp->bfcp_info.floor_requst_id,
                           BFCP_REQUEST_STATUS | BFCP_ATTR_MANDATORY, 0, &req_status,
                           BFCP_FLOOR_REQ_STATUS | BFCP_ATTR_MANDATORY, 0, &bfcp_tcp->bfcp_info.floor_id);

    ast_log(LOG_DEBUG, "bfcp_process_floor_release res=%d\n", res);

    return res;
}

static int ast_process_bfcp_tcp_msg_for_conference(struct ast_bfcp_tcp *bfcp_tcp, char *buf, int data_len)
{
	BFCP_MSG *bfcp_msg = NULL;
	int res = 0;
	enum bfcp_prim prim;
	unsigned int need_send_bfcp_msg = 0;
	struct ast_bfcp_info bfcp_info;

	struct bfcp_buffer bfcp_recv_buffer;
	struct bfcp_buffer bfcp_send_buffer;

	bfcp_create_buffer(&bfcp_recv_buffer);
	bfcp_create_buffer(&bfcp_send_buffer);
	bfcp_init_buffer(&bfcp_recv_buffer);
	bfcp_init_buffer(&bfcp_send_buffer);

	memcpy(bfcp_recv_buffer.buf, buf, data_len);
	bfcp_recv_buffer.size = data_len;

	res = bfcp_msg_decode(&bfcp_msg, &bfcp_recv_buffer);

	if (res < 0)
	{
		ast_log(LOG_WARNING, "ast_process_bfcp_msg error:%d\n", res);
		goto  PROCESS_BFCP_MSG_CLEANUP;
	}

	prim = bfcp_msg->header.prim;
	if (prim != BFCP_HELLO && prim != BFCP_HELLO_ACK)
	{
		ast_log(LOG_DEBUG, "Received bfcp message header prim:%d for user_id:%d\n", prim, bfcp_msg->header.user_id);
	}

	bfcp_tcp->bfcp_info.primitive = bfcp_msg->header.prim;

	memset(&bfcp_info, 0, sizeof(bfcp_info));
	bfcp_info.primitive = prim;

	switch (prim)
	{
		case BFCP_HELLO:
			res = bfcp_process_hello(bfcp_msg, &bfcp_send_buffer);
			need_send_bfcp_msg = 1;
			break;

		case BFCP_HELLO_ACK:
			res = bfcp_process_hello_ack(bfcp_msg);
			need_send_bfcp_msg = 0;
			break;

		case BFCP_FLOOR_REQUEST:
			ast_bfcp_set_info(bfcp_tcp->channel, &bfcp_info);
			res = bfcp_tcp_process_floor_request(bfcp_tcp, bfcp_msg, &bfcp_send_buffer);
			if (!res) {
				bfcp_info.status = BFCP_GRANTED;
				bfcp_info.primitive = BFCP_FLOOR_STATUS;
				strcpy(bfcp_info.info_conf_id, bfcp_tcp->bfcp_info.info_conf_id);
				strcpy(bfcp_info.floor_owner_id, bfcp_tcp->bfcp_info.account_id);
				strcpy(bfcp_info.msid, bfcp_tcp->bfcp_info.msid);
				ast_channel_set_opening_slides_video(bfcp_tcp->channel, 1);
				ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;
		case BFCP_FLOOR_RELEASE:
			ast_bfcp_set_info(bfcp_tcp->channel, &bfcp_info);
			res = bfcp_tcp_process_floor_release(bfcp_tcp, bfcp_msg, &bfcp_send_buffer);
			if (!res && (bfcp_tcp->bfcp_info.status == BFCP_RELEASED)) {
				bfcp_info.status = BFCP_RELEASED;
				bfcp_info.primitive = BFCP_FLOOR_STATUS;
    			strcpy(bfcp_info.info_conf_id, bfcp_tcp->bfcp_info.info_conf_id);
    			strcpy(bfcp_info.floor_owner_id, bfcp_tcp->bfcp_info.account_id);
				ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_REQUEST_STATUS:
			res = bfcp_process_floor_req_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info);
			bfcp_info.primitive = BFCP_FLOOR_STATUS;
			strcpy(bfcp_info.info_conf_id, bfcp_tcp->bfcp_info.info_conf_id);
			strcpy(bfcp_info.floor_owner_id, bfcp_tcp->bfcp_info.account_id);
			ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_STATUS:
			ast_bfcp_set_info(bfcp_tcp->channel, &bfcp_info);
			strcpy(bfcp_info.info_conf_id, bfcp_tcp->bfcp_info.info_conf_id);
			res = bfcp_process_floor_status(bfcp_msg, &bfcp_send_buffer, &bfcp_info, bfcp_tcp->channel);
			BFCP_ATTR *req_status = bfcp_attrs_find(&bfcp_msg->attr_list, BFCP_REQUEST_STATUS);
			if (req_status) {
				bfcp_info.status = req_status->v.reqstatus.status;
				bfcp_info.primitive = BFCP_FLOOR_STATUS;
				strcpy(bfcp_info.floor_owner_id, bfcp_tcp->bfcp_info.account_id);
				ast_queue_control_data(bfcp_tcp->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
			}
			need_send_bfcp_msg = 1;
			break;

		case BFCP_FLOOR_REQ_STATUS_ACK:
		case BFCP_FLOOR_STATUS_ACK:
			need_send_bfcp_msg = 0;
			break;

		default:
			need_send_bfcp_msg = 0;
			ast_log(LOG_DEBUG, "Can't process this bfcp message\n");
			break;
	}

	if (res < 0)
	{
		ast_log(LOG_WARNING, "bfcp process error\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	if (!need_send_bfcp_msg)
	{
		ast_log(LOG_DEBUG, "bfcp don't need to send message\n");
		goto PROCESS_BFCP_MSG_CLEANUP;
	}

	if (bfcp_send_buffer.pos > 0) {
		ao2_lock(bfcp_tcp->tcptls_session);
		ast_bfcp_tcptls_server_write(bfcp_tcp->tcptls_session, bfcp_send_buffer.buf, bfcp_send_buffer.pos);
		ao2_unlock(bfcp_tcp->tcptls_session);
	}

PROCESS_BFCP_MSG_CLEANUP:
	bfcp_free_msg(bfcp_msg);

	bfcp_free_buffer(&bfcp_recv_buffer);
	bfcp_free_buffer(&bfcp_send_buffer);

	return 1;
}

int ast_send_bfcp_tcp_msg(struct ast_bfcp_tcp *bfcp_tcp)
{
	struct bfcp_buffer send_buffer;
	enum bfcp_prim  prim_type;
	int res = 0;

	if (!bfcp_tcp)
	{
		ast_log(LOG_DEBUG, "ast_send_bfcp_tcp_msg param is NULL\n");
		return -1;
	}

	bfcp_create_buffer(&send_buffer);
	bfcp_init_buffer(&send_buffer);

	prim_type = bfcp_tcp->bfcp_info.primitive;
	ast_log(LOG_DEBUG, "ast_send_bfcp_tcp_msg prim_type is %d\n", prim_type);

	switch (prim_type)
	{
		case BFCP_HELLO:
			res = bfcp_hello_msg_encode(&send_buffer, &bfcp_tcp->bfcp_info);
			break;

		case BFCP_FLOOR_REQUEST:
			res = bfcp_floor_request_msg_encode(&send_buffer, &bfcp_tcp->bfcp_info);
			break;

		case BFCP_FLOOR_RELEASE:
			res = bfcp_floor_release_msg_encode(&send_buffer, &bfcp_tcp->bfcp_info);
			break;

		case BFCP_FLOOR_REQUEST_STATUS:
			res = bfcp_floor_request_status_msg_encode(&send_buffer, &bfcp_tcp->bfcp_info);
			break;

		case BFCP_FLOOR_STATUS:
			res = bfcp_floor_status_msg_encode(&send_buffer, &bfcp_tcp->bfcp_info);
			break;

		default:
			break;
	}

	if (res < 0)
	{
		ast_log(LOG_DEBUG, "ast_send_bfcp_tcp_msg encode message error\n");
		bfcp_free_buffer(&send_buffer);
		return res;
	}

	if (send_buffer.pos > 0) {
		ao2_lock(bfcp_tcp);
		if (bfcp_prepare_local_socket(bfcp_tcp) < 0) {
			ast_log(LOG_ERROR, "Bfcp prepare local socket error");
			bfcp_free_buffer(&send_buffer);
			ao2_unlock(bfcp_tcp);
			return res;
		}
		ast_bfcp_tcptls_server_write(bfcp_tcp->tcptls_session, send_buffer.buf, send_buffer.pos);
		ao2_unlock(bfcp_tcp);
	}

	bfcp_free_buffer(&send_buffer);

	ast_log(LOG_DEBUG, "ast_send_bfcp_tcp_msg end\n");

	return res;
}

void ast_bfcp_tcp_write(struct ast_bfcp_tcp *bfcp_tcp, const void *data, int datalen)
{
	if (!bfcp_tcp || !data) {
		return;
	}

	ao2_lock(bfcp_tcp);

	if (bfcp_prepare_local_socket(bfcp_tcp) < 0) {
		ast_log(LOG_ERROR, "ast_bfcp_tcp_write, prepare local socket error");
		ao2_unlock(bfcp_tcp);
		return;
	}

	ast_bfcp_tcptls_server_write(bfcp_tcp->tcptls_session, data, datalen);

	ao2_unlock(bfcp_tcp);

	return;
}

static void ast_bfcp_tcptls_stream_set_exclusive_input(struct ast_bfcp_tcptls_stream *stream, int exclusive_input)
{
	ast_assert(stream != NULL);

	stream->exclusive_input = exclusive_input;
}

void *_bfcp_tcp_helper_thread(struct ast_bfcp_tcp *bfcp_tcp)
{
	int res = 0;
	int timeout = BFCP_TCP_TIMEOUT;
	int flags = 0;
	struct pollfd fds[1];
	memset(fds, 0, sizeof(fds));
	struct ast_tcptls_session_args *client_args = NULL;
	char read_buf[BFCP_MAX_BUFFER_LENTH] = {0};

	ast_log(LOG_DEBUG, "Bfcp tcp helper thread begin\n");

	struct ast_bfcp_tcptls_session_instance *tcptls_session = bfcp_tcp->tcptls_session;

	if (!tcptls_session) {
		ast_log(LOG_DEBUG, "Bfcp tcp helper thread tcptls session is NULL\n");
		return NULL;
	}

	if (!tcptls_session->client) {
		if ((flags = fcntl(tcptls_session->fd, F_GETFL)) == -1) {
			ast_log(LOG_ERROR, "error setting socket to non blocking mode, fcntl() failed: %s\n", strerror(errno));
			goto cleanup;
		}

		flags |= O_NONBLOCK;
		if (fcntl(tcptls_session->fd, F_SETFL, flags) == -1) {
			ast_log(LOG_ERROR, "error setting socket to non blocking mode, fcntl() failed: %s\n", strerror(errno));
			goto cleanup;
		}
	} else {
		if (!(client_args = tcptls_session->parent)) {
			goto cleanup;
		}
	}

	flags = 1;
	if (setsockopt(tcptls_session->fd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags))) {
		ast_log(LOG_ERROR, "Error enabling TCP keep-alives on bfcp socket: %s\n", strerror(errno));
		goto cleanup;
	}

	/* set up pollfd to watch for reads on the socket */
	fds[0].fd = tcptls_session->fd;
	fds[0].events = POLLIN | POLLPRI;

	/*
	 * We cannot let the stream exclusively wait for data to arrive.
	 * We have to wake up the task to send outgoing messages.
	 */
	ast_bfcp_tcptls_stream_set_exclusive_input(tcptls_session->stream_cookie, 0);

	for (;;) {
		res = ast_poll(fds, 1, timeout); /* poll for bfcp tcp socket */
		if (res < 0) {
			ast_log(LOG_DEBUG, "BFCP TCP %s :: ast_wait_for_input returned %d\n", tcptls_session->client ? "client" : "server", res);
			goto cleanup;
		} else if (res == 0) {
			/* timeout */
			ast_log(LOG_DEBUG, "BFCP TCP %s timed out\n", tcptls_session->client ? "client" : "server");
			goto cleanup;
		}

		if (fds[0].revents) { /* there is data on the socket to be read */
			fds[0].revents = 0;

			memset(read_buf, 0, sizeof(read_buf));
			res = ast_bfcp_tcptls_server_read(tcptls_session, read_buf, sizeof(read_buf) - 1);
			if (res <= 0) {
				ast_log(LOG_DEBUG, "Bfcp TCP %s read no data\n", tcptls_session->client ? "client" : "server");
				goto cleanup;
			}

			if (bfcp_tcp->channel) {
				if (strlen(bfcp_tcp->bfcp_info.info_conf_id)) {
					ast_process_bfcp_tcp_msg_for_conference(bfcp_tcp, read_buf, res);
				} else {
					ast_process_bfcp_tcp_msg(bfcp_tcp, read_buf, res);
				}
			}
		}
	}

	ast_debug(2, "Shutting down thread for BFCP TCP %s\n", tcptls_session->client ? "client" : "server");

cleanup:
	ast_log(LOG_DEBUG, "Bfcp TCP %s helper thread cleanup\n", tcptls_session->client ? "client" : "server");

	/* if client, we own the parent session arguments and must decrement ref */
	if (client_args) {
		ao2_t_ref(client_args, -1, "closing tcptls thread, getting rid of client tcptls_session arguments");
	}

	ao2_ref(bfcp_tcp, +1);
	ao2_lock(bfcp_tcp);

	if (tcptls_session) {
		ast_log(LOG_DEBUG, "tcptls_session cleanup:%p\n", tcptls_session);
		ast_bfcp_tcptls_close_session_file(tcptls_session);
		tcptls_session->parent = NULL;

		ao2_ref(tcptls_session, -1);
		tcptls_session = NULL;
	}

	ao2_unlock(bfcp_tcp);
	ao2_ref(bfcp_tcp, -1);

	return NULL;
}

void *bfcp_tcp_worker_fn(void *data)
{
	struct ast_bfcp_tcp *bfcp_tcp = data;

	return _bfcp_tcp_helper_thread(bfcp_tcp);
}

void ast_bfcp_tcptls_server_stop(struct ast_bfcp_tcp *bfcp_tcp)
{
	if (NULL == bfcp_tcp) {
		ast_log(LOG_WARNING, "bfcp tcp stop, has no bfcp\n");
		return;
	}

	struct ast_tcptls_session_args *desc = &bfcp_tcp->bfcp_tcp_desc;
	if (NULL == desc) {
		ast_log(LOG_WARNING, "bfcp tcp stop, has no desc\n");
		return;
	}

	if (desc->master != AST_PTHREADT_NULL) {
		pthread_cancel(desc->master);
		pthread_kill(desc->master, SIGURG);
		pthread_join(desc->master, NULL);
		desc->master = AST_PTHREADT_NULL;
	}
	if (desc->accept_fd != -1) {
		close(desc->accept_fd);
	}
	desc->accept_fd = -1;

	ao2_ref(bfcp_tcp, -1);

	ast_log(LOG_DEBUG, "Stopped server :: %s\n", desc->name);

	return;
}

void ast_bfcp_tcp_destroy(struct ast_bfcp_tcp *bfcp_tcp)
{
	ast_log(LOG_DEBUG, "ast_bfcp_tcp_destroy begin\n");

	if (NULL == bfcp_tcp) {
		ast_log(LOG_WARNING, "bfcp is NULL, destory failed\n");
		return;
	}
	ast_bfcp_tcptls_server_stop(bfcp_tcp);

	ao2_ref(bfcp_tcp, -1);
	bfcp_tcp = NULL;

	return;
}

void ast_bfcp_udp_reachable_set(int value, const char *engine_name)
{
	struct ast_bfcp_engine *engine = NULL;

	if (!ast_strlen_zero(engine_name)) {
		AST_RWLIST_TRAVERSE(&engines, engine, entry) {
			if (!strcmp(engine->name, engine_name)) {
				break;
			}
		}
	} else {
		engine = AST_RWLIST_FIRST(&engines);
	}

	if (!engine) {
		ast_log(LOG_WARNING, "ast_bfcp_set_udp_reachable can't find BFCP engine\n");
		return;
	}

	engine->set_udp_reachable(value);

	return;
}

int ast_bfcp_register_conf_funcs(struct bfcp_conf_resource_manager_funcs *funcs)
{
	if (NULL == funcs)
	{
		ast_log(LOG_WARNING, "BFCP register conf resource manager funcs failure!\n");
		return -1;
	}

	memcpy(&bfcp_conf_funcs, funcs, sizeof(bfcp_conf_funcs));

	return 0;
}

int ast_bfcp_unregister_conf_funcs(void)
{

	memset(&bfcp_conf_funcs, 0, sizeof(bfcp_conf_funcs));

	return 0;
}

int ast_bfcp_engine_register(struct ast_bfcp_engine *engine, struct ast_module *module)
{
	struct ast_bfcp_engine *current_engine;

	/* Perform a sanity check on the engine structure to make sure it has the basics */
	if (ast_strlen_zero(engine->name)) {
		ast_log(LOG_WARNING, "BFCP Engine '%s' failed sanity check so it was not registered.\n", !ast_strlen_zero(engine->name) ? engine->name : "Unknown");
		return -1;
	}

	/* Link owner module to the RTP engine for reference counting purposes */
	engine->mod = module;

	AST_RWLIST_WRLOCK(&engines);

	/* Ensure that no two modules with the same name are registered at the same time */
	AST_RWLIST_TRAVERSE(&engines, current_engine, entry) {
		if (!strcmp(current_engine->name, engine->name)) {
			ast_log(LOG_WARNING, "An BFCP engine with the name '%s' has already been registered.\n", engine->name);
			AST_RWLIST_UNLOCK(&engines);
			return -1;
		}
	}

	/* The engine survived our critique. Off to the list it goes to be used */
	AST_RWLIST_INSERT_TAIL(&engines, engine, entry);

	AST_RWLIST_UNLOCK(&engines);

	ast_verb(2, "Registered BFCP engine '%s'\n", engine->name);

	return 0;
}

int ast_bfcp_engine_unregister(struct ast_bfcp_engine *engine)
{
	struct ast_bfcp_engine *current_engine = NULL;

	AST_RWLIST_WRLOCK(&engines);

	if ((current_engine = AST_RWLIST_REMOVE(&engines, engine, entry))) {
		ast_verb(2, "Unregistered BFCP engine '%s'\n", engine->name);
	}

	AST_RWLIST_UNLOCK(&engines);

	return current_engine ? 0 : -1;
}
#endif

