/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2012, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Channel Accessor API
 *
 * This file is intended to be the only file that ever accesses the
 * internals of an ast_channel. All other files should use the
 * accessor functions defined here.
 *
 * \author Terry Wilson
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include "asterisk/alertpipe.h"
#include "asterisk/paths.h"
#include "asterisk/channel.h"
#include "asterisk/channel_internal.h"
#include "asterisk/endpoints.h"
#include "asterisk/indications.h"
#include "asterisk/stasis_cache_pattern.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/stasis_endpoints.h"
#include "asterisk/stringfields.h"
#include "asterisk/stream.h"
#include "asterisk/test.h"
#include "asterisk/vector.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/avs_controller.h"
#include "asterisk/pbx.h"
#include "asterisk/core_local.h"
#endif
#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG ("Channel");
#endif

/*!
 * \brief Channel UniqueId structure
 * \note channel creation time used for determining LinkedId Propagation
 */
struct ast_channel_id {
	time_t creation_time;				/*!< Creation time */
	int creation_unique;				/*!< sub-second unique value */
	char unique_id[AST_MAX_UNIQUEID];	/*!< Unique Identifier */
};

/*!
 * \brief Main Channel structure associated with a channel.
 *
 * \note When adding fields to this structure, it is important to add the field
 *       'in position' with like-aligned fields, so as to keep the compiler from
 *       having to add padding to align fields. The structure's fields are sorted
 *       in this order: pointers, structures, long, int/enum, short, char. This
 *       is especially important on 64-bit architectures, where mixing 4-byte
 *       and 8-byte fields causes 4 bytes of padding to be added before many
 *       8-byte fields.
 */
struct ast_channel {
	const struct ast_channel_tech *tech;		/*!< Technology (point to channel driver) */
	void *tech_pvt;					/*!< Private data used by the technology driver */
	void *music_state;				/*!< Music State*/
	void *generatordata;				/*!< Current generator data if there is any */
	struct ast_generator *generator;		/*!< Current active data generator */
	struct ast_channel *masq;			/*!< Channel that will masquerade as us */
	struct ast_channel *masqr;			/*!< Who we are masquerading as */
	const char *blockproc;				/*!< Procedure causing blocking */
	const char *appl;				/*!< Current application */
	const char *data;				/*!< Data passed to current application */
	struct ast_sched_context *sched;                /*!< Schedule context */
	struct ast_filestream *stream;			/*!< Stream itself. */
	struct ast_filestream *vstream;			/*!< Video Stream itself. */
	ast_timing_func_t timingfunc;
	void *timingdata;
	struct ast_pbx *pbx;				/*!< PBX private structure for this channel */
	struct ast_trans_pvt *writetrans;		/*!< Write translation path */
	struct ast_trans_pvt *readtrans;		/*!< Read translation path */
	struct ast_audiohook_list *audiohooks;
	struct ast_framehook_list *framehooks;
	struct ast_cdr *cdr;				/*!< Call Detail Record */
	struct ast_tone_zone *zone;			/*!< Tone zone as set in indications.conf or
							 *   in the CHANNEL dialplan function */
	struct ast_channel_monitor *monitor;		/*!< Channel monitoring */
	ast_callid callid;			/*!< Bound call identifier pointer */
	struct ao2_container *dialed_causes;		/*!< Contains tech-specific and Asterisk cause data from dialed channels */

	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(name);         /*!< ASCII unique channel name */
		AST_STRING_FIELD(language);     /*!< Language requested for voice prompts */
		AST_STRING_FIELD(musicclass);   /*!< Default music class */
		AST_STRING_FIELD(latest_musicclass);   /*!< Latest active music class */
		AST_STRING_FIELD(accountcode);  /*!< Account code for billing */
		AST_STRING_FIELD(peeraccount);  /*!< Peer account code for billing */
		AST_STRING_FIELD(userfield);    /*!< Userfield for CEL billing */
		AST_STRING_FIELD(call_forward); /*!< Where to forward to if asked to dial on this interface */
		AST_STRING_FIELD(parkinglot);   /*! Default parking lot, if empty, default parking lot  */
		AST_STRING_FIELD(hangupsource); /*! Who is responsible for hanging up this channel */
		AST_STRING_FIELD(dialcontext);  /*!< Dial: Extension context that we were called from */
#ifdef GRANDSTREAM_NETWORKS
		AST_STRING_FIELD(sharedline);	/*!< To record the sharedline, Added by yfyang, for sca feature */
		AST_STRING_FIELD(reg_calleenum);	/*!< Add by ychp for call through register trunk*/
		AST_STRING_FIELD(reg_callernum);
		AST_STRING_FIELD(reg_callername);
		AST_STRING_FIELD(device_info);      /*!<for caller device info*/
		AST_STRING_FIELD(device_info_peer); /*!<for callee device info*/
		AST_STRING_FIELD(reject_reason);    /*!<for terminal warning info*/
#endif
	);

	struct ast_channel_id uniqueid;		/*!< Unique Channel Identifier - can be specified on creation */
	struct ast_channel_id linkedid;		/*!< Linked Channel Identifier - oldest propagated when bridged */

	struct timeval whentohangup; /*!< Non-zero, set to actual time when channel is to be hung up */
	pthread_t blocker;           /*!< If anyone is blocking, this is them */

	/*!
	 * \brief Dialed/Called information.
	 * \note Set on incoming channels to indicate the originally dialed party.
	 * \note Dialed Number Identifier (DNID)
	 */
	struct ast_party_dialed dialed;

	/*!
	 * \brief Channel Caller ID information.
	 * \note The caller id information is the caller id of this
	 * channel when it is used to initiate a call.
	 */
	struct ast_party_caller caller;

	/*!
	 * \brief Channel Connected Line ID information.
	 * \note The connected line information identifies the channel
	 * connected/bridged to this channel.
	 */
	struct ast_party_connected_line connected;

	/*!
	 * \brief Channel Connected Line ID information that was last indicated.
	 */
	struct ast_party_connected_line connected_indicated;

	/*! \brief Redirecting/Diversion information */
	struct ast_party_redirecting redirecting;

	struct ast_frame dtmff;				/*!< DTMF frame */
	struct varshead varshead;			/*!< A linked list for channel variables. See \ref AstChanVar */
	ast_group_t callgroup;				/*!< Call group for call pickups */
	ast_group_t pickupgroup;			/*!< Pickup group - which calls groups can be picked up? */
#ifdef GRANDSTREAM_NETWORKS
	struct ast_sip_endpoint_pickup_configuration * pickup_conf; /*The structure contains a list of pickup groups.*/
#endif
	struct ast_namedgroups *named_callgroups;	/*!< Named call group for call pickups */
	struct ast_namedgroups *named_pickupgroups;	/*!< Named pickup group - which call groups can be picked up? */
	struct timeval creationtime;			/*!< The time of channel creation */
	struct timeval answertime;				/*!< The time the channel was answered */
	struct ast_readq_list readq;
	struct ast_jb jb;				/*!< The jitterbuffer state */
	struct timeval dtmf_tv;				/*!< The time that an in process digit began, or the last digit ended */
	struct ast_hangup_handler_list hangup_handlers;/*!< Hangup handlers on the channel. */
	struct ast_datastore_list datastores; /*!< Data stores on the channel */
	struct ast_autochan_list autochans; /*!< Autochans on the channel */
	unsigned long insmpl;				/*!< Track the read/written samples for monitor use */
	unsigned long outsmpl;				/*!< Track the read/written samples for monitor use */

	int blocker_tid;					/*!< If anyone is blocking, this is their thread id */
	AST_VECTOR(, int) fds;				/*!< File descriptors for channel -- Drivers will poll on
							 *   these file descriptors, so at least one must be non -1.
							 *   See \arg \ref AstFileDesc */
	int softhangup;				/*!< Whether or not we have been hung up...  Do not set this value
							 *   directly, use ast_softhangup() */
	int fdno;					/*!< Which fd had an event detected on */
	int streamid;					/*!< For streaming playback, the schedule ID */
	int vstreamid;					/*!< For streaming video playback, the schedule ID */
	struct ast_format *oldwriteformat;  /*!< Original writer format */
	int timingfd;					/*!< Timing fd */
	enum ast_channel_state state;			/*!< State of line -- Don't write directly, use ast_setstate() */
	int rings;					/*!< Number of rings so far */
	int priority;					/*!< Dialplan: Current extension priority */
	int macropriority;				/*!< Macro: Current non-macro priority. See app_macro.c */
	int amaflags;					/*!< Set BEFORE PBX is started to determine AMA flags */
	enum ast_channel_adsicpe adsicpe;		/*!< Whether or not ADSI is detected on CPE */
	unsigned int fin;				/*!< Frames in counters. The high bit is a debug mask, so
							 *   the counter is only in the remaining bits */
	unsigned int fout;				/*!< Frames out counters. The high bit is a debug mask, so
							 *   the counter is only in the remaining bits */
	int hangupcause;				/*!< Why is the channel hanged up. See causes.h */
	unsigned int finalized:1;       /*!< Whether or not the channel has been successfully allocated */
	struct ast_flags flags;				/*!< channel flags of AST_FLAG_ type */
	int alertpipe[2];
	struct ast_format_cap *nativeformats;         /*!< Kinds of data this channel can natively handle */
	struct ast_format *readformat;            /*!< Requested read format (after translation) */
	struct ast_format *writeformat;           /*!< Requested write format (before translation) */
	struct ast_format *rawreadformat;         /*!< Raw read format (before translation) */
	struct ast_format *rawwriteformat;        /*!< Raw write format (after translation) */
	unsigned int emulate_dtmf_duration;		/*!< Number of ms left to emulate DTMF for */
	int visible_indication;                         /*!< Indication currently playing on the channel */
	int hold_state;							/*!< Current Hold/Unhold state */

	unsigned short transfercapability;		/*!< ISDN Transfer Capability - AST_FLAG_DIGITAL is not enough */

	struct ast_bridge *bridge;                      /*!< Bridge this channel is participating in */
	struct ast_bridge_channel *bridge_channel;/*!< The bridge_channel this channel is linked with. */
	struct ast_timer *timer;			/*!< timer object that provided timingfd */

	char context[AST_MAX_CONTEXT];			/*!< Dialplan: Current extension context */
	char exten[AST_MAX_EXTENSION];			/*!< Dialplan: Current extension number */
	char macrocontext[AST_MAX_CONTEXT];		/*!< Macro: Current non-macro context. See app_macro.c */
	char macroexten[AST_MAX_EXTENSION];		/*!< Macro: Current non-macro extension. See app_macro.c */
	char unbridged;							/*!< non-zero if the bridge core needs to re-evaluate the current
											 bridging technology which is in use by this channel's bridge. */
	char is_t38_active;						/*!< non-zero if T.38 is active on this channel. */
	char dtmf_digit_to_emulate;			/*!< Digit being emulated */
	char sending_dtmf_digit;			/*!< Digit this channel is currently sending out. (zero if not sending) */
	struct timeval sending_dtmf_tv;		/*!< The time this channel started sending the current digit. (Invalid if sending_dtmf_digit is zero.) */
#ifdef GRANDSTREAM_NETWORKS
	int ignore_direct_media;
	unsigned int is_parking_ringback:1; /*!< Parking lot timeout ring back call mark */
	int reject_call;
	char isDISAorIVR:1;
	char session[AST_MAX_EXTENSION];		//add by wxu@grandstream.cn
	char connect_num[AST_MAX_EXTENSION];	//add by wxu@grandstream.cn for active call
	char caller_num[AST_MAX_EXTENSION]; 	//add by wxu@grandstream.cn for callerID
	char connect_name[AST_MAX_CID_NAME];	//for active call name
	unsigned int attend_transfered;
	unsigned int phone_attend_transfered;
	struct ast_channel *my_chan;
	char *call_id;	/*!< sip call id */
	int hangupsipcode;				/*!< which sip msg cause this channel to hangup */
	unsigned int line_index;   /* the sca line index, for for sca feature */
	unsigned int peer_is_sac;
	char peer_channel_name[80];
	unsigned int wOriginateCall:1;
	unsigned char call_from_page;
	unsigned char is_vm;   /*!< Add by bfdu for disable video when enter voicemail, 2017.04.26 */
	char trunk_hunt_level; /* add by jlhu for trunk hunt */
	int avs_chanid;
	long avs_tick;
	int avs_chanfd;
	int avs_ctrlfd;    /*only used for avs control cmd*/
	int avs_chan_notifyfd;

	int avs_tonedetectfd;
	unsigned int avs_bridgeid;
	long avs_bridgetick;
	unsigned int avs_early_media;
	long avs_early_bridge_tick;
	long avs_local_bridge_tick;
	int support_2833;
	struct ast_channel_avs_records records;
	unsigned int port_group_id;
	unsigned char no_increase_call;
	/*begin: add by weidu ,for 110820, 20180621, paging send beep tone to caller*/
	unsigned int havesendbeepinpaging:1;
	unsigned int is_paging:1;
	/*end: add by weidu ,for 110820, 20180621, paging send beep tone to caller*/
	unsigned int is_avs_playtone:1;
	unsigned int syncflag:1;
	unsigned int ignore_avs_response:1; /*Ignore the response for del/stop avs request*/
	const char *exedappl; /*the executed application  by jklou*/
	unsigned char has_elin;
	char elin_dialer[AST_MAX_EXTENSION];
	char recordfile[256];
	char recordfile_backup[256];
	char recordfile_slave_media[256];
	char recordfile_master_media[256];

	time_t record_srtat_utc;
	time_t record_srtat_utc_backup;
	time_t record_resume_utc;	/*!< Used to calculate record file's duration after pause */
	time_t record_pause_duration;
	unsigned int elin_addr_ver;
	unsigned int elin_addr;
	struct in6_addr elin_addr6;
	char vfax_filepath[256];
	char vfax_localstationid[MAX_LOCAL_STATION_ID];
	char vfax_headerinfo[MAX_HEADER_INFO];
	char sca_bridge_uniqueid[128];
#endif
	struct stasis_cp_single *topics;		/*!< Topic for all channel's events */
	struct stasis_forward *endpoint_forward;	/*!< Subscription for event forwarding to endpoint's topic */
	struct stasis_forward *endpoint_cache_forward; /*!< Subscription for cache updates to endpoint's topic */
	struct ast_stream_topology *stream_topology; /*!< Stream topology */
	void *stream_topology_change_source; /*!< Source that initiated a stream topology change */
	struct ast_stream *default_streams[AST_MEDIA_TYPE_END]; /*!< Default streams indexed by media type */
#ifdef GRANDSTREAM_NETWORKS
	unsigned int hangup_rtp_timeout:1;
	unsigned int hangup_by_ami_action:1;
	unsigned int respond_18x_flag:1;    /* Added by yfyang, for Bug287522 */

	/* Add by bfdu, 2019.03.28 */
	unsigned char use_final_sdp;
	unsigned char allow_blind_transfer_callback;
	int blind_transfer_timeout;
	unsigned int feature_transfer;
	unsigned char cHasSend183;
	struct timeval first_183_tv;  /*!< Add by bfdu for crash when send two 183 to caller at the same time, 2017.03.16 */
	unsigned char recv_early_media_rsp;
	unsigned char callwaiting;
	unsigned char opening_slides_video;
	int law;
	unsigned char sfu_early_media;
	int conf_hangupcause;
	int need_notify_avs_early_media;
	int not_join_avs_bridge;
	unsigned char send_slides_video;
	unsigned char recv_slides_video;
	unsigned char transferer_hanguped;
	unsigned int turn_session_count;
	unsigned char is_auth;
	unsigned int is_nat_call;
	unsigned int peer_is_nat_call;  /* to be deleted, do not use this! */
	unsigned int local_outbound_index;
	int is_videocall;
	int is_callflip;
	int is_callflip_connected;
	int is_callflip_have_sendupdate;
	enum ast_page_mode page_mode;
	unsigned int talk_detect;
	struct timeval dahdi_bridge_tv;
	int local_is_gds_call;
	int local_is_gsc_call;
	int is_ipc_call;
	int local_is_hikvision_call;
	int is_send_fax;
	int is_vfax;
	int is_t30_vfax;
	int is_enable_fax_detect;
	int is_sdp_image_reinvite;
	unsigned char cCallerExtension;
	/*! Begin: Added by pxie@grandstream.cn for bug[42270], 2015-05-18.*/
	unsigned short wFrameHookAttached;	/*! True: framehook attach success.*/
	/*! End: Added by pxie@grandstream.cn for bug[42270], 2015-05-18.*/
	struct ast_stream_topology *stream_participant_topology;
	char stir_sk_orig_tn[AST_MAX_EXTENSION]; /*Add by weidu ,for outgoing stir and shaken*/
	unsigned char recv_bye;
	unsigned char dahdi_conf;

	/* added by zgmo for async channel streamfile protection */
	int is_streamfile;
	ast_cond_t streamfile_cond;
	ast_mutex_t streamfile_lock;
	/* end */

	// Bug 189664
	int is_spy_ringgroup;
	char ringgroup_chan_name[80];
	//

	struct ast_channel_media_encryption encryption;
	int dynamic_conf_rxgain; /*Add by weidu, for dahdi conf dynamic gain*/
	int dynamic_conf_txgain;
	int already_hangup;
	unsigned int nat_call_play_sound_timeout;
	int is_webrtc_trunk;
	struct ast_chan_move_hooks *move_hooks;
	void *move_hook_data;
	int is_wave_app;
	int is_in_calling_wave_app;
	unsigned int queue_reinvite_rsp_failed;
	int page_video_support;
	unsigned int is_pickup;
	unsigned int is_page_caller;
	unsigned int is_page_focus_record;
	unsigned int is_join_trunk_conf;
	char virtual_conf_num[256];
	unsigned int is_callerintercom;
	int open_file;
	unsigned int is_queue_call;
	char avs_address[MAX_IPADDR_V6_LEN];
	int join_ghost;
	unsigned int is_admin;
	int terminal_type;
	unsigned int mp3;
	unsigned int record_in_media;
	char file_encrypt_key[256];
	char file_encrypt_version[16];
	unsigned int is_focus_record;
#endif
	int already_recovery;
	char *recovery_uniqueid;
	char *recovery_linkedid;
	struct timeval recovery_createtime;  /* originally channel create time */
	struct timeval recovery_answertime;  /* originally channel answer time */
	char *recovery_app_data; /* last dialplan application data when recovery used*/
	char avoid_delete_track; /* need't delete recovery db if true */
};

/*! \brief The monotonically increasing integer counter for channel uniqueids */
static int uniqueint;

/* ACCESSORS */

#define DEFINE_STRINGFIELD_SETTERS_FOR(field, publish, assert_on_null) \
void ast_channel_##field##_set(struct ast_channel *chan, const char *value) \
{ \
	if ((assert_on_null)) ast_assert(!ast_strlen_zero(value)); \
	if (!strcmp(value, chan->field)) return; \
	ast_string_field_set(chan, field, value); \
	if (publish && ast_channel_internal_is_finalized(chan)) ast_channel_publish_snapshot(chan); \
} \
  \
void ast_channel_##field##_build_va(struct ast_channel *chan, const char *fmt, va_list ap) \
{ \
	ast_string_field_build_va(chan, field, fmt, ap); \
	if (publish && ast_channel_internal_is_finalized(chan)) ast_channel_publish_snapshot(chan); \
} \
void ast_channel_##field##_build(struct ast_channel *chan, const char *fmt, ...) \
{ \
	va_list ap; \
	va_start(ap, fmt); \
	ast_channel_##field##_build_va(chan, fmt, ap); \
	va_end(ap); \
}

DEFINE_STRINGFIELD_SETTERS_FOR(name, 0, 1);
DEFINE_STRINGFIELD_SETTERS_FOR(language, 1, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(musicclass, 0, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(latest_musicclass, 0, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(accountcode, 1, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(peeraccount, 1, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(userfield, 0, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(call_forward, 0, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(parkinglot, 0, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(hangupsource, 0, 0);
DEFINE_STRINGFIELD_SETTERS_FOR(dialcontext, 0, 0);

#define DEFINE_STRINGFIELD_GETTER_FOR(field) const char *ast_channel_##field(const struct ast_channel *chan) \
{ \
	return chan->field; \
}

DEFINE_STRINGFIELD_GETTER_FOR(name);
DEFINE_STRINGFIELD_GETTER_FOR(language);
DEFINE_STRINGFIELD_GETTER_FOR(musicclass);
DEFINE_STRINGFIELD_GETTER_FOR(latest_musicclass);
DEFINE_STRINGFIELD_GETTER_FOR(accountcode);
DEFINE_STRINGFIELD_GETTER_FOR(peeraccount);
DEFINE_STRINGFIELD_GETTER_FOR(userfield);
DEFINE_STRINGFIELD_GETTER_FOR(call_forward);
DEFINE_STRINGFIELD_GETTER_FOR(parkinglot);
DEFINE_STRINGFIELD_GETTER_FOR(hangupsource);
DEFINE_STRINGFIELD_GETTER_FOR(dialcontext);

const char *ast_channel_uniqueid(const struct ast_channel *chan)
{
	ast_assert(chan->uniqueid.unique_id[0] != '\0');
	return chan->uniqueid.unique_id;
}

const char *ast_channel_linkedid(const struct ast_channel *chan)
{
	ast_assert(chan->linkedid.unique_id[0] != '\0');
	return chan->linkedid.unique_id;
}
#ifdef GRANDSTREAM_NETWORKS
void ast_channel_set_parking(struct ast_channel* chan, unsigned int parking)
{
	if (!chan) {
		return;
	}

	chan->is_parking_ringback = parking;
}

unsigned int ast_channel_get_parking(const struct ast_channel* chan)
{
	if (!chan) {
		return 0;    
	}

	return chan->is_parking_ringback;
}

void ast_channel_set_musicclass(struct ast_channel *chan, const char *value)
{
	ast_string_field_set(chan, musicclass, value);
}
int ast_channel_hangupsipcode(const struct ast_channel *chan)
{
	return chan->hangupsipcode;
}
void ast_channel_hangupsipcode_set(struct ast_channel *chan, int value)
{
	chan->hangupsipcode= value;
}
void ast_channel_isDISAorIVR_set(struct ast_channel *chan, int value)
{
	chan->isDISAorIVR = value;
}

unsigned char ast_channel_no_increase_call_get(const struct ast_channel *chan)
{
	return chan->no_increase_call;
}

void ast_channel_no_increase_call_set(struct ast_channel *chan, unsigned char value)
{
	chan->no_increase_call = value;
}

/*! end: Added by pxie@grandstream.cn for bug[42270], 2015-05-18.*/
void ast_channel_set_mychan(struct ast_channel *owerchan, struct ast_channel * mychan)
{
	owerchan->my_chan = mychan;
}
struct ast_channel *ast_channel_get_mychan(struct ast_channel *owerchan)
{
	return owerchan->my_chan;
}

/*Begin: Uniqueid for qe, Added by shfguan*/
void ast_channel_vq_uniqueid_set(char *unique_id)
{
	ast_assert(uniqueid != NULL)

	struct ast_channel_id uniqueid;
	uniqueid.creation_time = time(NULL);
	uniqueid.creation_unique = ast_atomic_fetchadd_int(&uniqueint, 1);
	snprintf(unique_id, sizeof(uniqueid.unique_id), "%li.%d",(long)(uniqueid.creation_time),uniqueid.creation_unique);
}
/*End: Uniqueid for qe, Added by shfguan*/


/* Added by yfyang, for sca feature */
void ast_channel_set_line_index(struct ast_channel *chan, int value)
{
	chan->line_index = value;
}

unsigned int ast_channel_get_line_index(const struct ast_channel *chan)
{
	return chan->line_index;
}

void ast_channel_set_sharedline(struct ast_channel *chan, const char *value)
{
	ast_string_field_set(chan, sharedline, value);
}

const char* ast_channel_get_sharedline(struct ast_channel *chan)
{
	return chan->sharedline;
}

void ast_channel_set_peer_is_sca(struct ast_channel* chan, int value)
{
    chan->peer_is_sac = value;
}

int ast_channel_get_peer_is_sca(struct ast_channel* chan)
{
    return chan->peer_is_sac;
}

int ast_channel_set_peer_sca_name(struct ast_channel* chan, const char* chan_name)
{
    if (!chan || !chan_name || strlen(chan_name) <= 0)
	{
        return -1;
    }

    strncpy(chan->peer_channel_name, chan_name, strlen(chan_name));

    return 0;
}

int ast_channel_get_peer_sca_name(struct ast_channel* chan, char* result)
{
    if (!chan || !result)
	{
        return -1;
    }

    strncpy(result, chan->peer_channel_name, strlen(chan->peer_channel_name));

    return 0;
}

int ast_channel_set_reject_call(struct ast_channel* chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->reject_call = value;

	return 0;
}

int ast_channel_get_reject_call(struct ast_channel* chan)
{
	if (!chan) {
		return 0;
	}

	return chan->reject_call;
}

int ast_channel_spy_with_ringgroup(struct ast_channel* chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_spy_ringgroup;
}

int ast_channel_set_ringgroup_name(struct ast_channel* chan, const char* data)
{
	if (!chan || !data || strlen(data) <= 0) {
		return -1;
	}

	strncpy(chan->ringgroup_chan_name, data, strlen(data));
	chan->is_spy_ringgroup = 1;

	return 0;
}

int ast_channel_get_ringgroup_name(struct ast_channel* chan, char* data, int size)
{
	if (!chan || !data || size <= 0) {
		return -1;
	}

	strncpy(data, chan->ringgroup_chan_name, size);

	return 0;
}

void ast_channel_value_to_originate_call(struct ast_channel *chan, int value)
{
	chan->wOriginateCall = value;
}
int ast_channel_get_originate_call_value(struct ast_channel *chan)
{
	return chan->wOriginateCall;
}
void ast_channel_vm_set(struct ast_channel *chan, unsigned char is_vm)
{
	chan->is_vm = is_vm;
}

unsigned char ast_channel_vm_get(const struct ast_channel *chan)
{
	return chan->is_vm;
}

void ast_channel_set_streamfile_flg(struct ast_channel *chan, int flg)
{
	chan->is_streamfile = flg ? 1 : 0;
}

int ast_channel_get_streamfile_flg(struct ast_channel *chan)
{
	return chan->is_streamfile;
}

ast_cond_t *ast_channel_get_streamfile_cond(const struct ast_channel *chan)
{
	return (ast_cond_t *)&chan->streamfile_cond;
}
ast_mutex_t *ast_channel_get_streamfile_lock(const struct ast_channel *chan)
{
	return (ast_mutex_t *)&chan->streamfile_lock;
}
#endif

const char *ast_channel_appl(const struct ast_channel *chan)
{
	return chan->appl;
}
void ast_channel_appl_set(struct ast_channel *chan, const char *value)
{
	chan->appl = value;
}
const char *ast_channel_blockproc(const struct ast_channel *chan)
{
	return chan->blockproc;
}
void ast_channel_blockproc_set(struct ast_channel *chan, const char *value)
{
	chan->blockproc = value;
}
const char *ast_channel_data(const struct ast_channel *chan)
{
	return chan->data;
}
void ast_channel_data_set(struct ast_channel *chan, const char *value)
{
	chan->data = value;
}

const char *ast_channel_context(const struct ast_channel *chan)
{
	return chan->context;
}
void ast_channel_context_set(struct ast_channel *chan, const char *value)
{
	ast_copy_string(chan->context, value, sizeof(chan->context));
}
const char *ast_channel_exten(const struct ast_channel *chan)
{
	return chan->exten;
}
void ast_channel_exten_set(struct ast_channel *chan, const char *value)
{
	ast_copy_string(chan->exten, value, sizeof(chan->exten));
}
const char *ast_channel_macrocontext(const struct ast_channel *chan)
{
	return chan->macrocontext;
}
void ast_channel_macrocontext_set(struct ast_channel *chan, const char *value)
{
	ast_copy_string(chan->macrocontext, value, sizeof(chan->macrocontext));
}
const char *ast_channel_macroexten(const struct ast_channel *chan)
{
	return chan->macroexten;
}
void ast_channel_macroexten_set(struct ast_channel *chan, const char *value)
{
	ast_copy_string(chan->macroexten, value, sizeof(chan->macroexten));
}

char ast_channel_dtmf_digit_to_emulate(const struct ast_channel *chan)
{
	return chan->dtmf_digit_to_emulate;
}
void ast_channel_dtmf_digit_to_emulate_set(struct ast_channel *chan, char value)
{
	chan->dtmf_digit_to_emulate = value;
}

char ast_channel_sending_dtmf_digit(const struct ast_channel *chan)
{
	return chan->sending_dtmf_digit;
}
void ast_channel_sending_dtmf_digit_set(struct ast_channel *chan, char value)
{
	chan->sending_dtmf_digit = value;
}

struct timeval ast_channel_sending_dtmf_tv(const struct ast_channel *chan)
{
	return chan->sending_dtmf_tv;
}
void ast_channel_sending_dtmf_tv_set(struct ast_channel *chan, struct timeval value)
{
	chan->sending_dtmf_tv = value;
}

enum ama_flags ast_channel_amaflags(const struct ast_channel *chan)
{
	return chan->amaflags;
}

void ast_channel_amaflags_set(struct ast_channel *chan, enum ama_flags value)
{
	if (chan->amaflags == value) {
		return;
	}
	chan->amaflags = value;
	ast_channel_publish_snapshot(chan);
}
int ast_channel_fdno(const struct ast_channel *chan)
{
	return chan->fdno;
}
void ast_channel_fdno_set(struct ast_channel *chan, int value)
{
	chan->fdno = value;
}
int ast_channel_hangupcause(const struct ast_channel *chan)
{
	return chan->hangupcause;
}
void ast_channel_hangupcause_set(struct ast_channel *chan, int value)
{
	chan->hangupcause = value;
}
int ast_channel_macropriority(const struct ast_channel *chan)
{
	return chan->macropriority;
}
void ast_channel_macropriority_set(struct ast_channel *chan, int value)
{
	chan->macropriority = value;
}
int ast_channel_priority(const struct ast_channel *chan)
{
	return chan->priority;
}
void ast_channel_priority_set(struct ast_channel *chan, int value)
{
	chan->priority = value;
}
int ast_channel_rings(const struct ast_channel *chan)
{
	return chan->rings;
}
void ast_channel_rings_set(struct ast_channel *chan, int value)
{
	chan->rings = value;
}
int ast_channel_streamid(const struct ast_channel *chan)
{
	return chan->streamid;
}
void ast_channel_streamid_set(struct ast_channel *chan, int value)
{
	chan->streamid = value;
}
int ast_channel_timingfd(const struct ast_channel *chan)
{
	return chan->timingfd;
}
void ast_channel_timingfd_set(struct ast_channel *chan, int value)
{
	chan->timingfd = value;
}
int ast_channel_visible_indication(const struct ast_channel *chan)
{
	return chan->visible_indication;
}
void ast_channel_visible_indication_set(struct ast_channel *chan, int value)
{
	chan->visible_indication = value;
}
int ast_channel_hold_state(const struct ast_channel *chan)
{
	return chan->hold_state;
}
void ast_channel_hold_state_set(struct ast_channel *chan, int value)
{
	chan->hold_state = value;
}
int ast_channel_vstreamid(const struct ast_channel *chan)
{
	return chan->vstreamid;
}
void ast_channel_vstreamid_set(struct ast_channel *chan, int value)
{
	chan->vstreamid = value;
}
unsigned short ast_channel_transfercapability(const struct ast_channel *chan)
{
	return chan->transfercapability;
}
void ast_channel_transfercapability_set(struct ast_channel *chan, unsigned short value)
{
	chan->transfercapability = value;
}
unsigned int ast_channel_emulate_dtmf_duration(const struct ast_channel *chan)
{
	return chan->emulate_dtmf_duration;
}
void ast_channel_emulate_dtmf_duration_set(struct ast_channel *chan, unsigned int value)
{
	chan->emulate_dtmf_duration = value;
}
unsigned int ast_channel_fin(const struct ast_channel *chan)
{
	return chan->fin;
}
void ast_channel_fin_set(struct ast_channel *chan, unsigned int value)
{
	chan->fin = value;
}
unsigned int ast_channel_fout(const struct ast_channel *chan)
{
	return chan->fout;
}
void ast_channel_fout_set(struct ast_channel *chan, unsigned int value)
{
	chan->fout = value;
}
unsigned long ast_channel_insmpl(const struct ast_channel *chan)
{
	return chan->insmpl;
}
void ast_channel_insmpl_set(struct ast_channel *chan, unsigned long value)
{
	chan->insmpl = value;
}
unsigned long ast_channel_outsmpl(const struct ast_channel *chan)
{
	return chan->outsmpl;
}
void ast_channel_outsmpl_set(struct ast_channel *chan, unsigned long value)
{
	chan->outsmpl = value;
}
void *ast_channel_generatordata(const struct ast_channel *chan)
{
	return chan->generatordata;
}
void ast_channel_generatordata_set(struct ast_channel *chan, void *value)
{
	chan->generatordata = value;
}
void *ast_channel_music_state(const struct ast_channel *chan)
{
	return chan->music_state;
}
void ast_channel_music_state_set(struct ast_channel *chan, void *value)
{
	chan->music_state = value;
}
void *ast_channel_tech_pvt(const struct ast_channel *chan)
{
	return chan->tech_pvt;
}
void ast_channel_tech_pvt_set(struct ast_channel *chan, void *value)
{
	chan->tech_pvt = value;
}
void *ast_channel_timingdata(const struct ast_channel *chan)
{
	return chan->timingdata;
}
void ast_channel_timingdata_set(struct ast_channel *chan, void *value)
{
	chan->timingdata = value;
}
struct ast_audiohook_list *ast_channel_audiohooks(const struct ast_channel *chan)
{
	return chan->audiohooks;
}
void ast_channel_audiohooks_set(struct ast_channel *chan, struct ast_audiohook_list *value)
{
	chan->audiohooks = value;
}
struct ast_cdr *ast_channel_cdr(const struct ast_channel *chan)
{
	return chan->cdr;
}
void ast_channel_cdr_set(struct ast_channel *chan, struct ast_cdr *value)
{
	chan->cdr = value;
}
struct ast_channel *ast_channel_masq(const struct ast_channel *chan)
{
	return chan->masq;
}
void ast_channel_masq_set(struct ast_channel *chan, struct ast_channel *value)
{
	chan->masq = value;
}
struct ast_channel *ast_channel_masqr(const struct ast_channel *chan)
{
	return chan->masqr;
}
void ast_channel_masqr_set(struct ast_channel *chan, struct ast_channel *value)
{
	chan->masqr = value;
}
struct ast_channel_monitor *ast_channel_monitor(const struct ast_channel *chan)
{
	return chan->monitor;
}
void ast_channel_monitor_set(struct ast_channel *chan, struct ast_channel_monitor *value)
{
	chan->monitor = value;
}
struct ast_filestream *ast_channel_stream(const struct ast_channel *chan)
{
	return chan->stream;
}
void ast_channel_stream_set(struct ast_channel *chan, struct ast_filestream *value)
{
	chan->stream = value;
}
struct ast_filestream *ast_channel_vstream(const struct ast_channel *chan)
{
	return chan->vstream;
}
void ast_channel_vstream_set(struct ast_channel *chan, struct ast_filestream *value)
{
	chan->vstream = value;
}
struct ast_format_cap *ast_channel_nativeformats(const struct ast_channel *chan)
{
	return chan->nativeformats;
}

static void channel_set_default_streams(struct ast_channel *chan)
{
	enum ast_media_type type;

	ast_assert(chan != NULL);

	for (type = AST_MEDIA_TYPE_UNKNOWN; type < AST_MEDIA_TYPE_END; type++) {
		if (chan->stream_topology) {
			chan->default_streams[type] =
				ast_stream_topology_get_first_stream_by_type(chan->stream_topology, type);
		} else {
			chan->default_streams[type] = NULL;
		}
	}
}

void ast_channel_internal_set_stream_topology(struct ast_channel *chan,
	struct ast_stream_topology *topology)
{
	ast_stream_topology_free(chan->stream_topology);
	chan->stream_topology = topology;
	channel_set_default_streams(chan);
}

void ast_channel_internal_set_stream_topology_change_source(
	struct ast_channel *chan, void *change_source)
{
	chan->stream_topology_change_source = change_source;
}

void *ast_channel_get_stream_topology_change_source(struct ast_channel *chan)
{
	return chan->stream_topology_change_source;
}

void ast_channel_nativeformats_set(struct ast_channel *chan,
	struct ast_format_cap *value)
{
	ast_assert(chan != NULL);

	ao2_replace(chan->nativeformats, value);

	/* If chan->stream_topology is NULL, the channel is being destroyed
	 * and topology is destroyed.
	 */
	if (!chan->stream_topology) {
		return;
	}

	if (!ast_channel_is_multistream(chan) || !value) {
		struct ast_stream_topology *new_topology;

		new_topology = ast_stream_topology_create_from_format_cap(value);
		ast_channel_internal_set_stream_topology(chan, new_topology);
	}
}

struct ast_framehook_list *ast_channel_framehooks(const struct ast_channel *chan)
{
	return chan->framehooks;
}
void ast_channel_framehooks_set(struct ast_channel *chan, struct ast_framehook_list *value)
{
	chan->framehooks = value;
}
struct ast_generator *ast_channel_generator(const struct ast_channel *chan)
{
	return chan->generator;
}
void ast_channel_generator_set(struct ast_channel *chan, struct ast_generator *value)
{
	chan->generator = value;
}
struct ast_pbx *ast_channel_pbx(const struct ast_channel *chan)
{
	return chan->pbx;
}
void ast_channel_pbx_set(struct ast_channel *chan, struct ast_pbx *value)
{
	chan->pbx = value;
}
struct ast_sched_context *ast_channel_sched(const struct ast_channel *chan)
{
	return chan->sched;
}
void ast_channel_sched_set(struct ast_channel *chan, struct ast_sched_context *value)
{
	chan->sched = value;
}
struct ast_timer *ast_channel_timer(const struct ast_channel *chan)
{
	return chan->timer;
}
void ast_channel_timer_set(struct ast_channel *chan, struct ast_timer *value)
{
	chan->timer = value;
}
struct ast_tone_zone *ast_channel_zone(const struct ast_channel *chan)
{
	return chan->zone;
}
void ast_channel_zone_set(struct ast_channel *chan, struct ast_tone_zone *value)
{
	chan->zone = value;
}
struct ast_trans_pvt *ast_channel_readtrans(const struct ast_channel *chan)
{
	return chan->readtrans;
}
void ast_channel_readtrans_set(struct ast_channel *chan, struct ast_trans_pvt *value)
{
	chan->readtrans = value;
}
struct ast_trans_pvt *ast_channel_writetrans(const struct ast_channel *chan)
{
	return chan->writetrans;
}
void ast_channel_writetrans_set(struct ast_channel *chan, struct ast_trans_pvt *value)
{
	chan->writetrans = value;
}
const struct ast_channel_tech *ast_channel_tech(const struct ast_channel *chan)
{
	return chan->tech;
}
void ast_channel_tech_set(struct ast_channel *chan, const struct ast_channel_tech *value)
{
	if (value->read_stream || value->write_stream) {
		ast_assert(value->read_stream && value->write_stream);
	}

	chan->tech = value;
}
enum ast_channel_adsicpe ast_channel_adsicpe(const struct ast_channel *chan)
{
	return chan->adsicpe;
}
void ast_channel_adsicpe_set(struct ast_channel *chan, enum ast_channel_adsicpe value)
{
	chan->adsicpe = value;
}
enum ast_channel_state ast_channel_state(const struct ast_channel *chan)
{
	return chan->state;
}
ast_callid ast_channel_callid(const struct ast_channel *chan)
{
	return chan->callid;
}
void ast_channel_callid_set(struct ast_channel *chan, ast_callid callid)
{
	char call_identifier_from[AST_CALLID_BUFFER_LENGTH];
	char call_identifier_to[AST_CALLID_BUFFER_LENGTH];
	call_identifier_from[0] = '\0';
	ast_callid_strnprint(call_identifier_to, sizeof(call_identifier_to), callid);
	if (chan->callid) {
		ast_callid_strnprint(call_identifier_from, sizeof(call_identifier_from), chan->callid);
		ast_debug(3, "Channel Call ID changing from %s to %s\n", call_identifier_from, call_identifier_to);
	}

	chan->callid = callid;

	ast_test_suite_event_notify("CallIDChange",
		"State: CallIDChange\r\n"
		"Channel: %s\r\n"
		"CallID: %s\r\n"
		"PriorCallID: %s",
		ast_channel_name(chan),
		call_identifier_to,
		call_identifier_from);
}

void ast_channel_state_set(struct ast_channel *chan, enum ast_channel_state value)
{
	chan->state = value;
}
void ast_channel_set_oldwriteformat(struct ast_channel *chan, struct ast_format *format)
{
	ao2_replace(chan->oldwriteformat, format);
}
void ast_channel_set_rawreadformat(struct ast_channel *chan, struct ast_format *format)
{
	ao2_replace(chan->rawreadformat, format);
}
void ast_channel_set_rawwriteformat(struct ast_channel *chan, struct ast_format *format)
{
	ao2_replace(chan->rawwriteformat, format);
}
void ast_channel_set_readformat(struct ast_channel *chan, struct ast_format *format)
{
	ao2_replace(chan->readformat, format);
}
void ast_channel_set_writeformat(struct ast_channel *chan, struct ast_format *format)
{
	ao2_replace(chan->writeformat, format);
}
struct ast_format *ast_channel_oldwriteformat(struct ast_channel *chan)
{
	return chan->oldwriteformat;
}
struct ast_format *ast_channel_rawreadformat(struct ast_channel *chan)
{
	return chan->rawreadformat;
}
struct ast_format *ast_channel_rawwriteformat(struct ast_channel *chan)
{
	return chan->rawwriteformat;
}
struct ast_format *ast_channel_readformat(struct ast_channel *chan)
{
	return chan->readformat;
}
struct ast_format *ast_channel_writeformat(struct ast_channel *chan)
{
	return chan->writeformat;
}
struct ast_hangup_handler_list *ast_channel_hangup_handlers(struct ast_channel *chan)
{
	return &chan->hangup_handlers;
}
struct ast_datastore_list *ast_channel_datastores(struct ast_channel *chan)
{
	return &chan->datastores;
}
struct ast_autochan_list *ast_channel_autochans(struct ast_channel *chan)
{
	return &chan->autochans;
}
struct ast_readq_list *ast_channel_readq(struct ast_channel *chan)
{
	return &chan->readq;
}
struct ast_frame *ast_channel_dtmff(struct ast_channel *chan)
{
	return &chan->dtmff;
}
struct ast_jb *ast_channel_jb(struct ast_channel *chan)
{
	return &chan->jb;
}
struct ast_party_caller *ast_channel_caller(struct ast_channel *chan)
{
	return &chan->caller;
}
struct ast_party_connected_line *ast_channel_connected(struct ast_channel *chan)
{
	return &chan->connected;
}
struct ast_party_connected_line *ast_channel_connected_indicated(struct ast_channel *chan)
{
	return &chan->connected_indicated;
}
struct ast_party_id ast_channel_connected_effective_id(struct ast_channel *chan)
{
	return ast_party_id_merge(&chan->connected.id, &chan->connected.priv);
}
struct ast_party_dialed *ast_channel_dialed(struct ast_channel *chan)
{
	return &chan->dialed;
}
struct ast_party_redirecting *ast_channel_redirecting(struct ast_channel *chan)
{
	return &chan->redirecting;
}
struct ast_party_id ast_channel_redirecting_effective_orig(struct ast_channel *chan)
{
	return ast_party_id_merge(&chan->redirecting.orig, &chan->redirecting.priv_orig);
}
struct ast_party_id ast_channel_redirecting_effective_from(struct ast_channel *chan)
{
	return ast_party_id_merge(&chan->redirecting.from, &chan->redirecting.priv_from);
}
struct ast_party_id ast_channel_redirecting_effective_to(struct ast_channel *chan)
{
	return ast_party_id_merge(&chan->redirecting.to, &chan->redirecting.priv_to);
}
struct timeval *ast_channel_dtmf_tv(struct ast_channel *chan)
{
	return &chan->dtmf_tv;
}
struct timeval *ast_channel_whentohangup(struct ast_channel *chan)
{
	return &chan->whentohangup;
}
struct varshead *ast_channel_varshead(struct ast_channel *chan)
{
	return &chan->varshead;
}
void ast_channel_dtmff_set(struct ast_channel *chan, struct ast_frame *value)
{
	chan->dtmff = *value;
}
void ast_channel_jb_set(struct ast_channel *chan, struct ast_jb *value)
{
	chan->jb = *value;
}
void ast_channel_caller_set(struct ast_channel *chan, struct ast_party_caller *value)
{
	chan->caller = *value;
}
void ast_channel_connected_set(struct ast_channel *chan, struct ast_party_connected_line *value)
{
	chan->connected = *value;
}
void ast_channel_dialed_set(struct ast_channel *chan, struct ast_party_dialed *value)
{
	chan->dialed = *value;
}
void ast_channel_redirecting_set(struct ast_channel *chan, struct ast_party_redirecting *value)
{
	chan->redirecting = *value;
}
void ast_channel_dtmf_tv_set(struct ast_channel *chan, struct timeval *value)
{
	chan->dtmf_tv = *value;
}
void ast_channel_whentohangup_set(struct ast_channel *chan, struct timeval *value)
{
	chan->whentohangup = *value;
}
void ast_channel_varshead_set(struct ast_channel *chan, struct varshead *value)
{
	chan->varshead = *value;
}
struct timeval ast_channel_creationtime(struct ast_channel *chan)
{
	return chan->creationtime;
}
void ast_channel_creationtime_set(struct ast_channel *chan, struct timeval *value)
{
	chan->creationtime = *value;
}

struct timeval ast_channel_answertime(struct ast_channel *chan)
{
	return chan->answertime;
}

void ast_channel_answertime_set(struct ast_channel *chan, struct timeval *value)
{
	chan->answertime = *value;
}

/* Evil softhangup accessors */
int ast_channel_softhangup_internal_flag(struct ast_channel *chan)
{
	return chan->softhangup;
}
void ast_channel_softhangup_internal_flag_set(struct ast_channel *chan, int value)
{
	chan->softhangup = value;
}
void ast_channel_softhangup_internal_flag_add(struct ast_channel *chan, int value)
{
	chan->softhangup |= value;
}
void ast_channel_softhangup_internal_flag_clear(struct ast_channel *chan, int value)
{
	chan ->softhangup &= ~value;
}

int ast_channel_unbridged_nolock(struct ast_channel *chan)
{
	return chan->unbridged;
}

int ast_channel_unbridged(struct ast_channel *chan)
{
	int res;
	ast_channel_lock(chan);
	res = ast_channel_unbridged_nolock(chan);
	ast_channel_unlock(chan);
	return res;
}

void ast_channel_set_unbridged_nolock(struct ast_channel *chan, int value)
{
	chan->unbridged = !!value;
	ast_queue_frame(chan, &ast_null_frame);
}

void ast_channel_set_unbridged(struct ast_channel *chan, int value)
{
	ast_channel_lock(chan);
	ast_channel_set_unbridged_nolock(chan, value);
	ast_channel_unlock(chan);
}

int ast_channel_is_t38_active_nolock(struct ast_channel *chan)
{
	return chan->is_t38_active;
}

int ast_channel_is_t38_active(struct ast_channel *chan)
{
	int res;

	ast_channel_lock(chan);
	res = ast_channel_is_t38_active_nolock(chan);
	ast_channel_unlock(chan);
	return res;
}

void ast_channel_set_is_t38_active_nolock(struct ast_channel *chan, int is_t38_active)
{
	chan->is_t38_active = !!is_t38_active;
}

void ast_channel_set_is_t38_active(struct ast_channel *chan, int is_t38_active)
{
	ast_channel_lock(chan);
	ast_channel_set_is_t38_active_nolock(chan, is_t38_active);
	ast_channel_unlock(chan);
}

void ast_channel_callid_cleanup(struct ast_channel *chan)
{
	chan->callid = 0;
}

/* Typedef accessors */
ast_group_t ast_channel_callgroup(const struct ast_channel *chan)
{
	return chan->callgroup;
}
void ast_channel_callgroup_set(struct ast_channel *chan, ast_group_t value)
{
	chan->callgroup = value;
}
ast_group_t ast_channel_pickupgroup(const struct ast_channel *chan)
{
	return chan->pickupgroup;
}
void ast_channel_pickupgroup_set(struct ast_channel *chan, ast_group_t value)
{
	chan->pickupgroup = value;
}
struct ast_namedgroups *ast_channel_named_callgroups(const struct ast_channel *chan)
{
	return chan->named_callgroups;
}
void ast_channel_named_callgroups_set(struct ast_channel *chan, struct ast_namedgroups *value)
{
	ast_unref_namedgroups(chan->named_callgroups);
	chan->named_callgroups = ast_ref_namedgroups(value);
}
struct ast_namedgroups *ast_channel_named_pickupgroups(const struct ast_channel *chan)
{
	return chan->named_pickupgroups;
}
void ast_channel_named_pickupgroups_set(struct ast_channel *chan, struct ast_namedgroups *value)
{
	ast_unref_namedgroups(chan->named_pickupgroups);
	chan->named_pickupgroups = ast_ref_namedgroups(value);
}
#ifdef GRANDSTREAM_NETWORKS
void ast_channel_pickupgroup_list_set(struct ast_channel *chan, struct ast_sip_endpoint_pickup_configuration *value)
{
	chan->pickup_conf = value;
}
struct ast_sip_endpoint_pickup_configuration *ast_channel_pickupgroup_list(const struct ast_channel *chan)
{
	return chan->pickup_conf;
}
#endif

/* Alertpipe functions */
int ast_channel_alert_write(struct ast_channel *chan)
{
	return ast_alertpipe_write(chan->alertpipe);
}

ast_alert_status_t ast_channel_internal_alert_flush(struct ast_channel *chan)
{
	return ast_alertpipe_flush(chan->alertpipe);
}

ast_alert_status_t ast_channel_internal_alert_read(struct ast_channel *chan)
{
	return ast_alertpipe_read(chan->alertpipe);
}

int ast_channel_alert_writable(struct ast_channel *chan)
{
	return ast_alertpipe_writable(chan->alertpipe);
}

int ast_channel_internal_alert_readable(struct ast_channel *chan)
{
	return ast_alertpipe_readable(chan->alertpipe);
}

void ast_channel_internal_alertpipe_clear(struct ast_channel *chan)
{
	ast_alertpipe_clear(chan->alertpipe);
}

void ast_channel_internal_alertpipe_close(struct ast_channel *chan)
{
	ast_alertpipe_close(chan->alertpipe);
}

int ast_channel_internal_alertpipe_init(struct ast_channel *chan)
{
	return ast_alertpipe_init(chan->alertpipe);
}

int ast_channel_internal_alert_readfd(struct ast_channel *chan)
{
	return ast_alertpipe_readfd(chan->alertpipe);
}

void ast_channel_internal_alertpipe_swap(struct ast_channel *chan1, struct ast_channel *chan2)
{
	ast_alertpipe_swap(chan1->alertpipe, chan2->alertpipe);
}

/* file descriptor array accessors */
void ast_channel_internal_fd_set(struct ast_channel *chan, int which, int value)
{
	int pos;

	/* This ensures that if the vector has to grow with unused positions they will be
	 * initialized to -1.
	 */
	for (pos = AST_VECTOR_SIZE(&chan->fds); pos < which; pos++) {
		AST_VECTOR_REPLACE(&chan->fds, pos, -1);
	}

	AST_VECTOR_REPLACE(&chan->fds, which, value);
}
void ast_channel_internal_fd_clear(struct ast_channel *chan, int which)
{
	if (which >= AST_VECTOR_SIZE(&chan->fds)) {
		return;
	}

	AST_VECTOR_REPLACE(&chan->fds, which, -1);
}
void ast_channel_internal_fd_clear_all(struct ast_channel *chan)
{
	AST_VECTOR_RESET(&chan->fds, AST_VECTOR_ELEM_CLEANUP_NOOP);
}
int ast_channel_fd(const struct ast_channel *chan, int which)
{
	return (which >= AST_VECTOR_SIZE(&chan->fds)) ? -1 : AST_VECTOR_GET(&chan->fds, which);
}
int ast_channel_fd_isset(const struct ast_channel *chan, int which)
{
	return ast_channel_fd(chan, which) > -1;
}

int ast_channel_fd_count(const struct ast_channel *chan)
{
	return AST_VECTOR_SIZE(&chan->fds);
}

int ast_channel_fd_add(struct ast_channel *chan, int value)
{
	int pos = AST_EXTENDED_FDS;

	while (ast_channel_fd_isset(chan, pos)) {
		pos += 1;
	}

	AST_VECTOR_REPLACE(&chan->fds, pos, value);

	return pos;
}

pthread_t ast_channel_blocker(const struct ast_channel *chan)
{
	return chan->blocker;
}
void ast_channel_blocker_set(struct ast_channel *chan, pthread_t value)
{
	chan->blocker = value;
}

int ast_channel_blocker_tid(const struct ast_channel *chan)
{
	return chan->blocker_tid;
}
void ast_channel_blocker_tid_set(struct ast_channel *chan, int value)
{
	chan->blocker_tid = value;
}

ast_timing_func_t ast_channel_timingfunc(const struct ast_channel *chan)
{
	return chan->timingfunc;
}
void ast_channel_timingfunc_set(struct ast_channel *chan, ast_timing_func_t value)
{
	chan->timingfunc = value;
}

struct ast_bridge *ast_channel_internal_bridge(const struct ast_channel *chan)
{
	return chan->bridge;
}
void ast_channel_internal_bridge_set(struct ast_channel *chan, struct ast_bridge *value)
{
	chan->bridge = value;
	ast_channel_publish_snapshot(chan);
}

struct ast_bridge_channel *ast_channel_internal_bridge_channel(const struct ast_channel *chan)
{
	return chan->bridge_channel;
}
void ast_channel_internal_bridge_channel_set(struct ast_channel *chan, struct ast_bridge_channel *value)
{
	chan->bridge_channel = value;
}

struct ast_flags *ast_channel_flags(struct ast_channel *chan)
{
	return &chan->flags;
}

static int collect_names_cb(void *obj, void *arg, int flags) {
	struct ast_control_pvt_cause_code *cause_code = obj;
	struct ast_str **str = arg;

	ast_str_append(str, 0, "%s%s", (ast_str_strlen(*str) ? "," : ""), cause_code->chan_name);

	return 0;
}

struct ast_str *ast_channel_dialed_causes_channels(const struct ast_channel *chan)
{
	struct ast_str *chanlist = ast_str_create(128);

	if (!chanlist) {
		return NULL;
	}

	ao2_callback(chan->dialed_causes, 0, collect_names_cb, &chanlist);

	return chanlist;
}

struct ast_control_pvt_cause_code *ast_channel_dialed_causes_find(const struct ast_channel *chan, const char *chan_name)
{
	return ao2_find(chan->dialed_causes, chan_name, OBJ_KEY);
}

int ast_channel_dialed_causes_add(const struct ast_channel *chan, const struct ast_control_pvt_cause_code *cause_code, int datalen)
{
	struct ast_control_pvt_cause_code *ao2_cause_code;
	ao2_find(chan->dialed_causes, cause_code->chan_name, OBJ_KEY | OBJ_UNLINK | OBJ_NODATA);
	ao2_cause_code = ao2_alloc(datalen, NULL);

	if (ao2_cause_code) {
		memcpy(ao2_cause_code, cause_code, datalen);
		ao2_link(chan->dialed_causes, ao2_cause_code);
		ao2_ref(ao2_cause_code, -1);
		return 0;
	} else {
		return -1;
	}
}

void ast_channel_dialed_causes_clear(const struct ast_channel *chan)
{
	ao2_callback(chan->dialed_causes, OBJ_UNLINK | OBJ_NODATA | OBJ_MULTIPLE, NULL, NULL);
}

/* \brief Hash function for pvt cause code frames */
static int pvt_cause_hash_fn(const void *vpc, const int flags)
{
	const struct ast_control_pvt_cause_code *pc = vpc;
	return ast_str_hash(ast_tech_to_upper(ast_strdupa(pc->chan_name)));
}

/* \brief Comparison function for pvt cause code frames */
static int pvt_cause_cmp_fn(void *obj, void *vstr, int flags)
{
	struct ast_control_pvt_cause_code *pc = obj;
	char *str = ast_tech_to_upper(ast_strdupa(vstr));
	char *pc_str = ast_tech_to_upper(ast_strdupa(pc->chan_name));
	return !strcmp(pc_str, str) ? CMP_MATCH | CMP_STOP : 0;
}

#define DIALED_CAUSES_BUCKETS 37

struct ast_channel *__ast_channel_internal_alloc(void (*destructor)(void *obj), const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *file, int line, const char *function)
{
	struct ast_channel *tmp;

	tmp = __ao2_alloc(sizeof(*tmp), destructor,
		AO2_ALLOC_OPT_LOCK_MUTEX, "", file, line, function);

	if (!tmp) {
		return NULL;
	}

	if ((ast_string_field_init(tmp, 128))) {
		return ast_channel_unref(tmp);
	}

	tmp->dialed_causes = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
		DIALED_CAUSES_BUCKETS, pvt_cause_hash_fn, NULL, pvt_cause_cmp_fn);
	if (!tmp->dialed_causes) {
		return ast_channel_unref(tmp);
	}

	/* set the creation time in the uniqueid */
	tmp->uniqueid.creation_time = time(NULL);
	tmp->uniqueid.creation_unique = ast_atomic_fetchadd_int(&uniqueint, 1);

	/* use provided id or default to historical {system-}time.# format */
	if (assignedids && !ast_strlen_zero(assignedids->uniqueid)) {
		ast_copy_string(tmp->uniqueid.unique_id, assignedids->uniqueid, sizeof(tmp->uniqueid.unique_id));
	} else if (ast_strlen_zero(ast_config_AST_SYSTEM_NAME)) {
		snprintf(tmp->uniqueid.unique_id, sizeof(tmp->uniqueid.unique_id), "%li.%d",
			(long)(tmp->uniqueid.creation_time),
			tmp->uniqueid.creation_unique);
	} else {
		snprintf(tmp->uniqueid.unique_id, sizeof(tmp->uniqueid.unique_id), "%s-%li.%d",
			ast_config_AST_SYSTEM_NAME,
			(long)(tmp->uniqueid.creation_time),
			tmp->uniqueid.creation_unique);
	}

	/* copy linked id from parent channel if known */
	if (requestor) {
		tmp->linkedid = requestor->linkedid;
	} else {
		tmp->linkedid = tmp->uniqueid;
	}

	AST_VECTOR_INIT(&tmp->fds, AST_MAX_FDS);

	return tmp;
}

struct ast_channel *ast_channel_internal_oldest_linkedid(struct ast_channel *a, struct ast_channel *b)
{
	ast_assert(a->linkedid.creation_time != 0);
	ast_assert(b->linkedid.creation_time != 0);

	if (a->linkedid.creation_time < b->linkedid.creation_time) {
		return a;
	}
	if (b->linkedid.creation_time < a->linkedid.creation_time) {
		return b;
	}
	if (a->linkedid.creation_unique < b->linkedid.creation_unique) {
		return a;
	}
	return b;
}

void ast_channel_internal_copy_linkedid(struct ast_channel *dest, struct ast_channel *source)
{
	if (dest->linkedid.creation_time == source->linkedid.creation_time
		&& dest->linkedid.creation_unique == source->linkedid.creation_unique
		&& !strcmp(dest->linkedid.unique_id, source->linkedid.unique_id)) {
		return;
	}
	dest->linkedid = source->linkedid;
	ast_channel_publish_snapshot(dest);
}

void ast_channel_internal_swap_uniqueid_and_linkedid(struct ast_channel *a, struct ast_channel *b)
{
	struct ast_channel_id temp;

	temp = a->uniqueid;
	a->uniqueid = b->uniqueid;
	b->uniqueid = temp;

	temp = a->linkedid;
	a->linkedid = b->linkedid;
	b->linkedid = temp;
}

void ast_channel_internal_swap_topics(struct ast_channel *a, struct ast_channel *b)
{
	struct stasis_cp_single *temp;

	temp = a->topics;
	a->topics = b->topics;
	b->topics = temp;
}

void ast_channel_internal_swap_endpoint_forward_and_endpoint_cache_forward(struct ast_channel *a, struct ast_channel *b)
{
	struct stasis_forward *temp;
	temp = a->endpoint_forward;
	a->endpoint_forward = b->endpoint_forward;
	b->endpoint_forward = temp;

	temp = a->endpoint_cache_forward;
	a->endpoint_cache_forward = b->endpoint_cache_forward;
	b->endpoint_cache_forward = temp;
}

void ast_channel_internal_set_fake_ids(struct ast_channel *chan, const char *uniqueid, const char *linkedid)
{
	ast_copy_string(chan->uniqueid.unique_id, uniqueid, sizeof(chan->uniqueid.unique_id));
	ast_copy_string(chan->linkedid.unique_id, linkedid, sizeof(chan->linkedid.unique_id));
}

void ast_channel_internal_cleanup(struct ast_channel *chan)
{
	if (chan->dialed_causes) {
		ao2_t_ref(chan->dialed_causes, -1,
			"done with dialed causes since the channel is going away");
		chan->dialed_causes = NULL;
	}

	ast_string_field_free_memory(chan);

	chan->endpoint_forward = stasis_forward_cancel(chan->endpoint_forward);
	chan->endpoint_cache_forward = stasis_forward_cancel(chan->endpoint_cache_forward);

	stasis_cp_single_unsubscribe(chan->topics);
	chan->topics = NULL;

	ast_channel_internal_set_stream_topology(chan, NULL);

#ifdef GRANDSTREAM_NETWORKS
	ast_channel_internal_set_participant_stream_topology(chan, NULL);
#endif

	AST_VECTOR_FREE(&chan->fds);
}

void ast_channel_internal_finalize(struct ast_channel *chan)
{
	chan->finalized = 1;
}

int ast_channel_internal_is_finalized(struct ast_channel *chan)
{
	return chan->finalized;
}

struct stasis_topic *ast_channel_topic(struct ast_channel *chan)
{
	if (!chan) {
		return ast_channel_topic_all();
	}

	return stasis_cp_single_topic(chan->topics);
}

struct stasis_topic *ast_channel_topic_cached(struct ast_channel *chan)
{
	if (!chan) {
		return ast_channel_topic_all_cached();
	}

	return stasis_cp_single_topic_cached(chan->topics);
}

int ast_channel_forward_endpoint(struct ast_channel *chan,
	struct ast_endpoint *endpoint)
{
	ast_assert(chan != NULL);
	ast_assert(endpoint != NULL);

	chan->endpoint_forward =
		stasis_forward_all(ast_channel_topic(chan),
			ast_endpoint_topic(endpoint));
	if (!chan->endpoint_forward) {
		return -1;
	}

	chan->endpoint_cache_forward = stasis_forward_all(ast_channel_topic_cached(chan),
		ast_endpoint_topic(endpoint));
	if (!chan->endpoint_cache_forward) {
		chan->endpoint_forward = stasis_forward_cancel(chan->endpoint_forward);
		return -1;
	}

	return 0;
}

int ast_channel_internal_setup_topics(struct ast_channel *chan)
{
	const char *topic_name = chan->uniqueid.unique_id;
	ast_assert(chan->topics == NULL);

	if (ast_strlen_zero(topic_name)) {
		topic_name = "<dummy-channel>";
	}

	chan->topics = stasis_cp_single_create(
		ast_channel_cache_all(), topic_name);
	if (!chan->topics) {
		return -1;
	}

	return 0;
}

AST_THREADSTORAGE(channel_errno);

void ast_channel_internal_errno_set(enum ast_channel_error error)
{
	enum ast_channel_error *error_code = ast_threadstorage_get(&channel_errno, sizeof(*error_code));
	if (!error_code) {
		return;
	}

	*error_code = error;
}

enum ast_channel_error ast_channel_internal_errno(void)
{
	enum ast_channel_error *error_code = ast_threadstorage_get(&channel_errno, sizeof(*error_code));
	if (!error_code) {
		return AST_CHANNEL_ERROR_UNKNOWN;
	}

	return *error_code;
}

struct ast_stream_topology *ast_channel_get_stream_topology(
	const struct ast_channel *chan)
{
	ast_assert(chan != NULL);

	return chan->stream_topology;
}

struct ast_stream_topology *ast_channel_set_stream_topology(struct ast_channel *chan,
	struct ast_stream_topology *topology)
{
	struct ast_stream_topology *new_topology;

	ast_assert(chan != NULL);

	/* A non-MULTISTREAM channel can't manipulate topology directly */
	ast_assert(ast_channel_is_multistream(chan));

	/* Unless the channel is being destroyed, we always want a topology on
	 * it even if its empty.
	 */
	if (!topology) {
		new_topology = ast_stream_topology_alloc();
	} else {
		new_topology = topology;
	}

	if (new_topology) {
		ast_channel_internal_set_stream_topology(chan, new_topology);
	}

	return new_topology;
}

struct ast_stream *ast_channel_get_default_stream(struct ast_channel *chan,
	enum ast_media_type type)
{
	ast_assert(chan != NULL);
	ast_assert(type < AST_MEDIA_TYPE_END);

	return chan->default_streams[type];
}

void ast_channel_internal_swap_stream_topology(struct ast_channel *chan1,
	struct ast_channel *chan2)
{
	struct ast_stream_topology *tmp_topology;

	ast_assert(chan1 != NULL && chan2 != NULL);

	tmp_topology = chan1->stream_topology;
	chan1->stream_topology = chan2->stream_topology;
	chan2->stream_topology = tmp_topology;

	channel_set_default_streams(chan1);
	channel_set_default_streams(chan2);
}

int ast_channel_is_multistream(struct ast_channel *chan)
{
	return (chan->tech && chan->tech->read_stream && chan->tech->write_stream);
}

#ifdef GRANDSTREAM_NETWORKS
/*start added by zyxia for fax ,2020.8.13*/
void ast_channel_set_framehook_attach(struct ast_channel *chan, unsigned short wFrameHookAttached)
{
	chan->wFrameHookAttached = wFrameHookAttached;
}
int ast_channel_get_framehook_attach(struct ast_channel *chan)
{
	return chan->wFrameHookAttached;
}

void ast_channel_set_caller_extension_flag(struct ast_channel *chan, unsigned char cValue)
{
	chan->cCallerExtension = cValue;
}
int ast_channel_get_caller_extension_flag(struct ast_channel *chan)
{
	return chan->cCallerExtension;
}

void ast_channel_set_is_enable_fax_detect(struct ast_channel *chan, int value)
{
	chan->is_enable_fax_detect = value;
}
int ast_channel_get_is_enable_fax_detect(const struct ast_channel *chan)
{
	return chan->is_enable_fax_detect;
}

void ast_channel_set_is_sdp_image_reinvite(struct ast_channel *chan, int value)
{
	chan->is_sdp_image_reinvite = value;
}
int ast_channel_get_is_sdp_image_reinvite(const struct ast_channel *chan)
{
	return chan->is_sdp_image_reinvite;
}

void ast_channel_set_vfax_flag(struct ast_channel *chan, int value)
{
	chan->is_vfax = value;
}
int ast_channel_get_vfax_flag(const struct ast_channel *chan)
{
	return chan->is_vfax;
}

void ast_channel_set_is_send_fax(struct ast_channel *chan, int value)
{
	chan->is_send_fax = value;
}

int ast_channel_get_is_send_fax(const struct ast_channel *chan)
{
	return chan->is_send_fax;
}

void ast_channel_set_is_t30_vfax(struct ast_channel *chan, int value)
{
	chan->is_t30_vfax = value;
}

int ast_channel_get_is_t30_vfax(const struct ast_channel *chan)
{
	return chan->is_t30_vfax;
}

/*end added by zyxia for fax ,2020.8.13*/

/* add by jlhu for trunk hunt begin */
void ast_channel_trunk_priority_set(struct ast_channel *chan, char level)
{
	chan->trunk_hunt_level= level;
}

char ast_channel_trunk_priority_get(struct ast_channel *chan)
{
	return chan->trunk_hunt_level;
}
/* add by jlhu for trunk hunt end */

unsigned int ast_channel_rtp_timeout_hangup_get(const struct ast_channel *chan)
{
	return chan->hangup_rtp_timeout;
}

void ast_channel_rtp_timeout_hangup_set(struct ast_channel *chan, unsigned int value)
{
	chan->hangup_rtp_timeout = value;
}

unsigned int ast_channel_ami_action_hangup_get(const struct ast_channel *chan)
{
	return chan->hangup_by_ami_action;
}

void ast_channel_ami_action_hangup_set(struct ast_channel *chan, unsigned int value)
{
	chan->hangup_by_ami_action = value;
}

const char *ast_channel_session(const struct ast_channel *chan)
{
	return chan->session;
}

void ast_channel_session_set(struct ast_channel *chan, const char *value)
{
	struct timeval time_val = ast_channel_creationtime(chan);
	int time_sec = (int)time_val.tv_sec;
	int time_usec = (int)time_val.tv_usec;

	memset(chan->session,0,sizeof(chan->session));
	/*modify by wxu, for IVR or someone else without [macro-user-callerid] in dialplan*/
	if( !strcasecmp(value,"session") )
	{
		char *pstr = strchr(ast_channel_name(chan),'/');
		if(pstr != NULL)
		{
			snprintf(chan->session,sizeof(chan->session)-1,"%10d%6d-%d",time_sec,time_usec,atoi(pstr + 1));
		}
		else
		{
			ast_log(LOG_ERROR,"ast_channel_session_set,chan->name[%s] caller num is null",ast_channel_name(chan));
		}
	}
	else
	{
		snprintf(chan->session,sizeof(chan->session)-1,"%d%d-%s",time_sec,time_usec,value);
	}
	/*end*/
	//ast_channel_publish_snapshot(chan);
}

void ast_channel_session_set_value(struct ast_channel *chan, const char *value)
{
	memset(chan->session,0,sizeof(chan->session));
	snprintf(chan->session,sizeof(chan->session)-1,"%s",value);
}

/*added by wxu@grandstream.cn for active call*/
const char *ast_channel_connect_num(const struct ast_channel *chan)
{
	return chan->connect_num;
}

void ast_channel_connect_num_set(struct ast_channel *chan, const char *value)
{
	memset(chan->connect_num,0,sizeof(chan->connect_num));
	snprintf(chan->connect_num,sizeof(chan->connect_num)-1,"%s",value);
}

const char *ast_channel_connect_name(const struct ast_channel *chan)
{
	return chan->connect_name;
}

void ast_channel_connect_name_set(struct ast_channel *chan, const char *value)
{
	memset(chan->connect_name,0,sizeof(chan->connect_name));
	snprintf(chan->connect_name,sizeof(chan->connect_name)-1,"%s",value);
}

const char *ast_channel_caller_num(const struct ast_channel *chan)
{
	return chan->caller_num;
}

void ast_channel_caller_num_set(struct ast_channel *chan, const char *value)
{
	if(value && strlen(value) > 0){
		memset(chan->caller_num,0,sizeof(chan->caller_num));
		snprintf(chan->caller_num,sizeof(chan->caller_num)-1,"%s",value);
	}
}
int ast_channel_attend_transfered(const struct ast_channel *chan)
{
	return chan->attend_transfered;
}

void ast_channel_attend_transfered_set(struct ast_channel *chan,unsigned int value)
{
	chan->attend_transfered = value;
}
int ast_channel_phone_attend_transfered(const struct ast_channel *chan)
{
	return chan->phone_attend_transfered;
}

void ast_channel_phone_attend_transfered_set(struct ast_channel *chan,unsigned int value)
{
	chan->phone_attend_transfered = value;
}
const char *ast_channel_call_id(const struct ast_channel *chan)
{
	return chan->call_id;
}

void ast_channel_call_id_set(struct ast_channel *chan, const char *value)
{
	if (chan->call_id) {
		ast_free(chan->call_id);
	}
	chan->call_id = (char *)value;
}

void ast_channel_set_elin_flag(struct ast_channel *chan, unsigned char para)
{
	chan->has_elin = para;
}

unsigned char ast_channel_get_elin_flag(struct ast_channel *chan)
{
	return chan->has_elin;
}

void ast_channel_set_elin_dialer(struct ast_channel *chan, const char *elin_dialer)
{
	memset(chan->elin_dialer, 0, sizeof(chan->elin_dialer));
	snprintf(chan->elin_dialer, sizeof(chan->elin_dialer)-1, "%s", elin_dialer);
}

char *ast_channel_get_elin_dialer(struct ast_channel *chan)
{
	return chan->elin_dialer;
}

void ast_channel_set_syncflag(struct ast_channel *chan, unsigned int syncflag)
{
	chan->syncflag = syncflag;
}

unsigned int ast_channel_get_syncflag(struct ast_channel *chan)
{
	return chan->syncflag;
}

void ast_channel_set_recordfile(struct ast_channel *chan, const char *recordfile)
{
	chan->recordfile[0] = '\0';
	snprintf(chan->recordfile, sizeof(chan->recordfile)-1, "%s", recordfile);
}

char *ast_channel_get_recordfile(struct ast_channel *chan)
{
	return chan->recordfile;
}

void ast_channel_set_recordfile_slave_media(struct ast_channel *chan, const char *recordfile)
{
	chan->recordfile_slave_media[0] = '\0';
	snprintf(chan->recordfile_slave_media, sizeof(chan->recordfile_slave_media)-1, "%s", recordfile);
}

char *ast_channel_get_recordfile_slave_media(struct ast_channel *chan)
{
	return chan->recordfile_slave_media;
}

void ast_channel_set_recordfile_master_media(struct ast_channel *chan, const char *recordfile)
{
	chan->recordfile_master_media[0] = '\0';
	snprintf(chan->recordfile_master_media, sizeof(chan->recordfile_master_media)-1, "%s", recordfile);
}

char *ast_channel_get_recordfile_master_media(struct ast_channel *chan)
{
	return chan->recordfile_master_media;
}

void ast_channel_set_elin_ip_ver(struct ast_channel *chan, unsigned int ver)
{
	chan->elin_addr_ver = ver;
	return;
}

unsigned int ast_channel_get_elin_ip_ver(struct ast_channel *chan)
{
	return chan->elin_addr_ver;
}

void ast_channel_set_elin_ip(struct ast_channel *chan,  struct in_addr *src_addr)
{
	chan->elin_addr = src_addr->s_addr;
	return;
}

unsigned int ast_channel_get_elin_ip(struct ast_channel *chan)
{
	return chan->elin_addr;
}

void ast_channel_set_elin_ip6(struct ast_channel *chan,  struct in6_addr *src_addr6)
{
	memcpy(&chan->elin_addr6, src_addr6, sizeof(struct in6_addr));
	return;
}

struct in6_addr *ast_channel_get_elin_ip6(struct ast_channel *chan)
{
	return &chan->elin_addr6;
}

const char *ast_channel_reg_calleenum(const struct ast_channel *chan)
{
	return chan->reg_calleenum;
}

void ast_channel_reg_calleenum_set(struct ast_channel *chan, const char *value)
{
	ast_string_field_set(chan, reg_calleenum, value);
}

const char *ast_channel_reg_callernum(const struct ast_channel *chan)
{
	return chan->reg_callernum;
}

void ast_channel_reg_callernum_set(struct ast_channel *chan, const char *value)
{
	ast_string_field_set(chan, reg_callernum, value);
}

const char *ast_channel_reg_callername(const struct ast_channel *chan)
{
	return chan->reg_callername;
}

void ast_channel_reg_callername_set(struct ast_channel *chan, const char *value)
{
	ast_string_field_set(chan, reg_callername, value);
}
void ast_channel_set_pagecall_flag(struct ast_channel *chan, unsigned char para)
{
     chan->call_from_page = para;
}

unsigned char ast_channel_get_pagecall_flag(struct ast_channel *chan)
{
     return chan->call_from_page;
}

void ast_channel_set_use_final_sdp(struct ast_channel *chan, unsigned char value)
{
	chan->use_final_sdp = value;
}

unsigned char ast_channel_get_use_final_sdp(struct ast_channel *chan)
{
	return chan->use_final_sdp;
}

unsigned int ast_channel_size_get(void)
{
	return sizeof(struct ast_channel);
}

/* Add by bfdu for blind transfer callback, 2017.03.17 */
unsigned char ast_channel_allow_blind_transfer_callback(const struct ast_channel *chan)
{
	return chan->allow_blind_transfer_callback;
}

void ast_channel_allow_blind_transfer_callback_set(struct ast_channel *chan, unsigned char allow_value)
{
	chan->allow_blind_transfer_callback = allow_value;
}

unsigned int ast_channel_blind_transfer_timeout(const struct ast_channel *chan)
{
	return chan->blind_transfer_timeout;
}

void ast_channel_blind_transfer_timeout_set(struct ast_channel *chan, unsigned int timeout)
{
	chan->blind_transfer_timeout = timeout;
}

int ast_channel_is_avs_engine(struct ast_channel *chan)
{
	return chan->avs_chanid;
}

void ast_channel_set_avs_chanid(struct ast_channel *chan, unsigned int chanid)
{
	chan->avs_chanid = chanid;
}

int ast_channel_get_avs_chanid(struct ast_channel *chan)
{
	return chan->avs_chanid;
}

void ast_channel_set_avs_tick(struct ast_channel *chan, long tick)
{
	chan->avs_tick = tick;
}

long ast_channel_get_avs_tick(struct ast_channel *chan)
{
	return chan->avs_tick;
}

void ast_channel_set_avs_chanfd(struct ast_channel *chan, int chanfd)
{
	chan->avs_chanfd = chanfd;
}

int ast_channel_get_avs_chanfd(struct ast_channel *chan)
{
	return chan->avs_chanfd;
}

void ast_channel_set_avs_chan_notifyfd(struct ast_channel *chan, int chanfd)
{
	chan->avs_chan_notifyfd = chanfd;
}

int ast_channel_get_avs_chan_notifyfd(struct ast_channel *chan)
{
	return chan->avs_chan_notifyfd;
}

void ast_channel_set_avs_ctrlfd(struct ast_channel *chan, int ctrlfd)
{
	chan->avs_ctrlfd = ctrlfd;
}

int ast_channel_get_avs_ctrlfd(struct ast_channel *chan)
{
	return chan->avs_ctrlfd;
}

void ast_channel_set_avs_bridgeid(struct ast_channel *chan, unsigned int bridgeid)
{
	chan->avs_bridgeid = bridgeid;
}

int ast_channel_get_avs_bridgeid(struct ast_channel *chan)
{
	return chan->avs_bridgeid;
}

void ast_channel_set_avs_bridgetick(struct ast_channel *chan, long tick)
{
	chan->avs_bridgetick = tick;
}

long ast_channel_get_avs_bridgetick(struct ast_channel *chan)
{
	return chan->avs_bridgetick;
}

void ast_channel_set_avs_early_media(struct ast_channel *chan, unsigned int flag)
{
	chan->avs_early_media = flag;
}

int ast_channel_get_avs_early_media(struct ast_channel *chan)
{
	return chan->avs_early_media;
}

void ast_channel_set_avs_early_media_bridge_tick(struct ast_channel *chan, long tick)
{
	chan->avs_early_bridge_tick = tick;
}

long ast_channel_get_avs_early_media_bridge_tick(const struct ast_channel *chan)
{
	return chan->avs_early_bridge_tick;
}

void ast_channel_set_avs_local_bridge_tick(struct ast_channel *chan, long tick)
{
	chan->avs_local_bridge_tick = tick;
}

long ast_channel_get_avs_local_bridge_tick(const struct ast_channel *chan)
{
	return chan->avs_local_bridge_tick;
}

void ast_channel_set_avs_tonedectfd(struct ast_channel *chan, int tonedectfd)
{
	chan->avs_tonedetectfd = tonedectfd;
}

int ast_channel_get_avs_tonedectfd(struct ast_channel *chan)
{
	return chan->avs_tonedetectfd;
}

void ast_channel_set_avs_tone_flag(struct ast_channel *chan, int flag)
{
	chan->is_avs_playtone = flag;
}

int ast_channel_get_avs_tone_flag(struct ast_channel *chan)
{
	return chan->is_avs_playtone;
}

int ast_channel_get_2833_flag(struct ast_channel *chan)
{
	return chan->support_2833;
}

void ast_channel_set_2833_flag(struct ast_channel *chan, int flag)
{
	chan->support_2833 = flag;
}

int ast_channel_avs_add_record(struct ast_channel *chan, char *uid, char *filename)
{
	struct ast_channel_avs_record_node *record_tmp = NULL;
	int insert_flag = 0;
	
	if (!chan || !filename) {
		return -1;
	}

	ast_channel_lock(chan);
	if (uid && 0 < chan->records.record_count) {
		AST_LIST_TRAVERSE_SAFE_BEGIN(&chan->records.record_list, record_tmp, entry) {
			if ((record_tmp->uid != NULL) && !strcasecmp(uid, record_tmp->uid)) {
				memset(record_tmp->filename, 0, sizeof(record_tmp->filename));
				strncpy(record_tmp->filename, filename, sizeof(record_tmp->filename));
				/*start added by zyxia for Bug 290242*/
				if(!ast_strlen_zero(ast_channel_get_recordfile_slave_media(chan))) {
					memset(record_tmp->slave_filename, 0, sizeof(record_tmp->slave_filename));
					strncpy(record_tmp->slave_filename, ast_channel_get_recordfile_slave_media(chan), sizeof(record_tmp->slave_filename));
				}
				/*end added by zyxia for Bug 290242*/
				insert_flag = 1;
				break;
			}
		}
		AST_LIST_TRAVERSE_SAFE_END;
	}

	if (insert_flag) {
		ast_channel_unlock(chan);
		return 0;
	}
	
	record_tmp = ast_calloc(1, sizeof(struct ast_channel_avs_record_node));
	strncpy(record_tmp->filename, filename, sizeof(record_tmp->filename));
	/*start added by zyxia for Bug 290242*/
	if(!ast_strlen_zero(ast_channel_get_recordfile_slave_media(chan))) {
		memset(record_tmp->slave_filename, 0, sizeof(record_tmp->slave_filename));
		strncpy(record_tmp->slave_filename, ast_channel_get_recordfile_slave_media(chan), sizeof(record_tmp->slave_filename));
	}
	/*End added by zyxia for Bug 290242*/

	if (uid) {
		strncpy(record_tmp->uid, uid, sizeof(record_tmp->uid));
	}
	AST_LIST_INSERT_TAIL(&chan->records.record_list, record_tmp, entry);
	chan->records.record_count++;
	ast_channel_unlock(chan);
	ast_log(LOG_NOTICE,"channel[%s] has [%d] recording", ast_channel_name(chan), chan->records.record_count);
	return 0;
}

int ast_channel_avs_del_record(struct ast_channel *chan, char *uid, char *filename)
{
	struct ast_channel_avs_record_node *record_tmp = NULL;
	if (!chan || !filename) {
		return -1;
	}

	ast_channel_lock(chan);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&chan->records.record_list, record_tmp, entry) {
		if (uid == NULL) {
			/* del all the record */
			AST_LIST_REMOVE_CURRENT(entry);
			ast_free(record_tmp);
			if (0 < chan->records.record_count) {
				chan->records.record_count--;
			}
			continue;
		} else {
			if ((record_tmp->uid != NULL) && !strcasecmp(uid, record_tmp->uid)) {
				/* Matched uid, del it*/
				if (ast_strlen_zero(record_tmp->filename)) {
					strncpy(filename, record_tmp->filename, sizeof(record_tmp->filename));
				}
				AST_LIST_REMOVE_CURRENT(entry);
				ast_free(record_tmp);
				if (0 < chan->records.record_count) {
					chan->records.record_count--;
				}
				break;
			}
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ast_channel_unlock(chan);
	return 0;
}

int ast_channel_avs_is_recording(struct ast_channel *chan)
{
	int ret = 0;
	
	if (!chan) {
		return ret;
	}	
	
	if (0 < chan->records.record_count) {
		ret = 1;
	}
	return ret;
}

/*start added by zyxia for Bug 290242*/
int ast_channel_avs_get_record_count(struct ast_channel *chan)
{
	int ret = 0;
	
	if (!chan) {
		return ret;
	}

	return chan->records.record_count;
}
/*End added by zyxia for Bug 290242*/

struct ast_record_list *ast_channel_avs_get_recordlist(struct ast_channel *chan)
{
	return &chan->records.record_list;
}

void ast_channel_swap_avs_para(struct ast_channel *chan1, struct ast_channel *chan2)
{
	unsigned int avs_chanid;
	unsigned int avs_bridgeid;
	int avs_chanfd;
	int avs_ctrlfd;
	int avs_chan_notifyfd;
	int support_2833;
	unsigned int record_count;
	unsigned int avs_early_media;
	long avs_tick;
	long avs_early_bridge_tick;
	long avs_local_bridge_tick;
	char avs_address[MAX_IPADDR_V6_LEN]={0};
	if (!chan1 || !chan2) {
		return;
	}

	avs_chanid = ast_channel_get_avs_chanid(chan1);
	ast_channel_set_avs_chanid(chan1, ast_channel_get_avs_chanid(chan2));
	ast_channel_set_avs_chanid(chan2, avs_chanid);

	avs_chanfd = ast_channel_get_avs_chanfd(chan1);
	ast_channel_set_avs_chanfd(chan1, ast_channel_get_avs_chanfd(chan2));
	ast_channel_set_avs_chanfd(chan2, avs_chanfd);

	avs_ctrlfd = ast_channel_get_avs_ctrlfd(chan1);
	ast_channel_set_avs_ctrlfd(chan1, ast_channel_get_avs_ctrlfd(chan2));
	ast_channel_set_avs_ctrlfd(chan2, avs_ctrlfd);

	support_2833 = ast_channel_get_2833_flag(chan1);
	ast_channel_set_2833_flag(chan1, ast_channel_get_2833_flag(chan2));
	ast_channel_set_2833_flag(chan2, support_2833);

	/*Start added by zyxia for Bug 299806,2024.5.16*/
	avs_chan_notifyfd = ast_channel_get_avs_chan_notifyfd(chan1);
	ast_channel_set_avs_chan_notifyfd(chan1, ast_channel_get_avs_chan_notifyfd(chan2));
	ast_channel_set_avs_chan_notifyfd(chan2, avs_chan_notifyfd);
	/*End added by zyxia for Bug 299806,2024.5.16*/

	snprintf(avs_address,sizeof(avs_address)-1,ast_channel_get_avs_address(chan1));
	ast_channel_set_avs_address(chan1, ast_channel_get_avs_address(chan2));
	ast_channel_set_avs_address(chan2, avs_address);

	avs_bridgeid = chan1->avs_bridgeid;
	chan1->avs_bridgeid = chan2->avs_bridgeid;
	chan2->avs_bridgeid = avs_bridgeid;

	avs_tick = chan1->avs_tick;
	chan1->avs_tick = chan2->avs_tick;
	chan2->avs_tick = avs_tick;

	avs_early_media = chan1->avs_early_media;
	chan1->avs_early_media = chan2->avs_early_media;
	chan2->avs_early_media = avs_early_media;

	avs_early_bridge_tick = chan1->avs_early_bridge_tick;
	chan1->avs_early_bridge_tick = chan2->avs_early_bridge_tick;
	chan2->avs_early_bridge_tick = avs_early_bridge_tick;
    
	avs_local_bridge_tick = chan1->avs_local_bridge_tick;
	chan1->avs_local_bridge_tick = chan2->avs_local_bridge_tick;
	chan2->avs_local_bridge_tick = avs_local_bridge_tick;

	record_count = chan1->records.record_count;
	chan1->records.record_count = chan2->records.record_count;
	chan2->records.record_count = record_count;
	struct ast_record_list tmp_recordq;
	struct ast_channel_avs_record_node *current = NULL;
	AST_LIST_HEAD_INIT_NOLOCK(&tmp_recordq);
	AST_LIST_APPEND_LIST(&tmp_recordq, &chan1->records.record_list, entry);
	AST_LIST_APPEND_LIST(&chan1->records.record_list, &chan2->records.record_list, entry);
	while ((current = AST_LIST_REMOVE_HEAD(&tmp_recordq, entry))) {
		AST_LIST_INSERT_TAIL(&chan2->records.record_list, current, entry);
	}
}


void ast_channel_set_port_group_id(struct ast_channel *chan, unsigned int port_group_id)
{
	chan->port_group_id = port_group_id;
	return;
}

unsigned int ast_channel_get_port_group_id(struct ast_channel *chan)
{
	return chan->port_group_id;
}

/*begin: add by weidu ,for 110820, 20180621,paging caller Paging/ Intercom Announcement Warning */			
unsigned int ast_channel_reg_sendbeep_get(const struct ast_channel *chan)
{
	return chan->havesendbeepinpaging;
}
void ast_channel_reg_sendbeep_set(struct ast_channel *chan, unsigned int value)
{
	chan->havesendbeepinpaging = value;
}
unsigned int ast_channel_paging_get(const struct ast_channel *chan)
{
	return chan->is_paging;
}

void ast_channel_paging_set(struct ast_channel *chan, unsigned int is_paging)
{
	chan->is_paging = is_paging;
}
/*end: add by weidu ,for 110820*/

enum ast_terminal_type ast_channel_get_terminate_type(struct ast_channel *chan)
{
	const struct ast_channel_tech * chan_tech = NULL;
	enum ast_terminal_type type = AST_TERMINAL_TYPE_UNKNOW;

	chan_tech = ast_channel_tech(chan);
	if (chan_tech->get_terminal_type) {
		type = chan_tech->get_terminal_type(chan);
	}

	return type;
}
const char *ast_channel_exedappl(const struct ast_channel *chan)
{
	return chan->exedappl;
}

void ast_channel_exedappl_set(struct ast_channel *chan, const char *value)
{
	chan->exedappl = value;
}

void ast_channel_set_feature_transfer(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->feature_transfer = value;
}

unsigned int ast_channel_get_feature_transfer(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->feature_transfer;
}

int ast_channel_respond_18x_flag(const struct ast_channel *chan)
{
	return chan->respond_18x_flag;
}

void ast_channel_respond_18x_flag_set(struct ast_channel *chan, int value)
{
	chan->respond_18x_flag = value;
}

int ast_channel_cHasSend183(const struct ast_channel *chan)
{
	return chan->cHasSend183;
}

void ast_channel_cHasSend183_set(struct ast_channel *chan, int value)
{
	chan->cHasSend183 = value;
}

unsigned char ast_channel_recv_early_media_rsp_get(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->recv_early_media_rsp;
}

void ast_channel_recv_early_media_rsp_set(struct ast_channel *chan, unsigned char value)
{
	if (!chan) {
		return;
	}

	chan->recv_early_media_rsp = value;
}

void ast_set_direct_media_ignore(struct ast_channel* channel, int value)
{
	if (!channel|| channel->ignore_direct_media == value) {
		return;
	}

	channel->ignore_direct_media = value;
}

void ast_get_direct_media_ignore(struct ast_channel* channel, int* value) 
{
	if (!channel) {
		*value = 0;
	}

	*value = channel->ignore_direct_media;
}
unsigned char ast_channel_get_callwaiting(const struct ast_channel *chan)
{
	return chan->callwaiting;
}

void ast_channel_callwaiting_set(struct ast_channel *chan, unsigned char value)
{
	chan->callwaiting = value;
}

unsigned char ast_channel_get_opening_slides_video(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->opening_slides_video;
}

void ast_channel_set_opening_slides_video(struct ast_channel *chan, unsigned char value)
{
	if (!chan) {
		return;
	}

	chan->opening_slides_video = value;
}
int ast_channel_get_law(const struct ast_channel *chan)
{
	return chan->law;
}

void ast_channel_law_set(struct ast_channel *chan, int value)
{
	chan->law = value;
}

int ast_channel_get_if_negotiate_state_done(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 1;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 1;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_if_negotiate_state_done) {
		return chan_tech->get_if_negotiate_state_done(chan);
	}

	return 1;
}

unsigned int ast_channel_get_delayed_reinvite_count(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_delayed_reinvite_count) {
		return chan_tech->get_delayed_reinvite_count(chan);
	}

	return 0;
}

void ast_channel_inc_delayed_reinvite_count(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->inc_delayed_reinvite_count) {
		chan_tech->inc_delayed_reinvite_count(chan);
	}

	return;
}

void ast_channel_release_video_resource(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->release_video_resource) {
		chan_tech->release_video_resource(chan);
	}

	return;
}

unsigned char ast_channel_sfu_early_media(const struct ast_channel *chan)
{
	return chan->sfu_early_media;
}
void ast_channel_sfu_early_media_set(struct ast_channel *chan, unsigned char value)
{
	chan->sfu_early_media = value;
}

int ast_channel_conf_hangupcause(const struct ast_channel *chan)
{
	return chan->conf_hangupcause;
}
void ast_channel_conf_hangupcause_set(struct ast_channel *chan, int value)
{
	chan->conf_hangupcause = value;
}

int ast_channel_get_need_notify_avs_early_media(const struct ast_channel *chan)
{
	return chan->need_notify_avs_early_media;
}
void ast_channel_need_notify_avs_early_media_set(struct ast_channel *chan, int value)
{
	chan->need_notify_avs_early_media = value;
}
int ast_channel_get_not_join_avs_bridge(const struct ast_channel *chan)
{
	return chan->not_join_avs_bridge;
}
void ast_channel_not_join_avs_bridge_set(struct ast_channel *chan, int value)
{
	chan->not_join_avs_bridge = value;
}

void ast_channel_set_send_slides_video(struct ast_channel *chan, unsigned char value)
{
	if (!chan) {
		return;
	}

	chan->send_slides_video = value;

	return;
}

unsigned char ast_channel_get_send_slides_video(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->send_slides_video;
}

void ast_channel_set_recv_slides_video(struct ast_channel *chan, unsigned char value)
{
	if (!chan) {
		return;
	}

	chan->recv_slides_video = value;

	return;
}

unsigned char ast_channel_get_recv_slides_video(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->recv_slides_video;
}

void ast_channel_set_transferer_hanguped(struct ast_channel *chan, unsigned char value)
{
	if (!chan) {
		return;
	}

	chan->transferer_hanguped = value;

	return;
}

unsigned char ast_channel_get_transferer_hanguped(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->transferer_hanguped;
}

void ast_channel_set_turn_session_count(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->turn_session_count = value;

	if (chan->turn_session_count) {
		chan->is_nat_call = 1;
		/* added by zgmo for Bug178229 */
		if (strstr(ast_channel_name(chan), "PJSIP/trunk")) {
			const char *conf_mcb_invite = pbx_builtin_getvar_helper(chan, "CONF_MCB_INVITE");
			const char *peer_name = pbx_builtin_getvar_helper(chan, "DIALOUT_CHANNEL_PARENT");
			if (!ast_strlen_zero(conf_mcb_invite) && !strcmp(conf_mcb_invite, "yes")
				&& !ast_strlen_zero(peer_name) && strstr(peer_name, "Local/")) {
				struct ast_channel *peer = ast_channel_get_by_name(peer_name);
				if (peer) {
					struct ast_channel *peer_local_peer = ast_local_get_peer(peer);
					if (peer_local_peer) {
						peer_local_peer->is_nat_call = 1;
						ast_log(LOG_DEBUG, "Set Chan:%s is_nat_call", ast_channel_name(peer_local_peer));
						ast_channel_unref(peer_local_peer);
					}

					peer->is_nat_call = 1;
					ast_log(LOG_DEBUG, "Set Chan:%s is_nat_call", ast_channel_name(peer));
					ast_channel_unref(peer);
				}
			}
		}
		/* end */
	}

	return;
}

unsigned int ast_channel_get_turn_session_count(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->turn_session_count;
}

void ast_channel_set_auth(struct ast_channel *chan, unsigned char value)
{
	if (!chan) {
		return;
	}

	chan->is_auth = value;

	return;
}

unsigned char ast_channel_get_auth(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_auth;
}

void ast_channel_set_local_outbound_index(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->local_outbound_index = value;
	return;
}

unsigned int ast_channel_get_local_outbound_index(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->local_outbound_index;
}

void ast_channel_set_is_nat_call(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->is_nat_call = value;
	return;
}

unsigned int ast_channel_get_is_nat_call(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_nat_call;
}

void ast_channel_set_peer_is_nat_call(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->peer_is_nat_call = value;

	return;
}

/* This API is depricated */
unsigned int ast_channel_get_peer_is_nat_call(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->peer_is_nat_call;
}

void ast_channel_set_is_gds_call(struct ast_channel* chan, int value) {
	if (!chan) {
		return;
	}

	chan->local_is_gds_call = value;
}

int ast_channel_get_is_gds_call(const struct ast_channel* chan) {
	if (!chan) {
		return 0;
	}

	return chan->local_is_gds_call;
}

void ast_channel_set_is_hikvision_call(struct ast_channel* chan, int value) {
	if (!chan) {
		return;
	}

	chan->local_is_hikvision_call = value;
}

int ast_channel_get_is_hikvision_call(const struct ast_channel* chan) {
	if (!chan) {
		return 0;
	}

	return chan->local_is_hikvision_call;
}

void ast_channel_set_is_gsc_call(struct ast_channel* chan, int value) {
	if (!chan) {
		return;
	}

	chan->local_is_gsc_call = value;
}

int ast_channel_get_is_gsc_call(const struct ast_channel* chan) {
	if (!chan) {
		return 0;
	}

	return chan->local_is_gsc_call;
}

void ast_channel_set_is_ipc_call(struct ast_channel* chan, int value) {
	if (!chan) {
		return;
	}

	chan->is_ipc_call = value;
}

int ast_channel_get_is_ipc_call(const struct ast_channel* chan) {
	if (!chan) {
		return 0;
	}

	return chan->is_ipc_call;
}

int ast_channel_set_is_videocall(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_videocall = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_videocall:%d",
		ast_channel_name(chan), value);

	return 0;
}

int ast_channel_get_is_callflip(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_callflip;
}

int ast_channel_set_is_callflip(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_callflip = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_callflip:%d",
		ast_channel_name(chan), value);

	return 0;
}

int ast_channel_get_is_callflip_connected(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_callflip_connected;
}

int ast_channel_set_is_callflip_connected(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_callflip_connected = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is callflip_connected:%d",
		ast_channel_name(chan), value);

	return 0;
}

int ast_channel_get_is_callflip_have_sendupdate(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_callflip_have_sendupdate;
}

int ast_channel_set_is_callflip_have_sendupdate(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_callflip_have_sendupdate = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is is_callflip_have_sendupdate:%d",
		ast_channel_name(chan), value);

	return 0;
}

int ast_channel_get_is_videocall(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_videocall;
}

void ast_channel_page_mode_set(struct ast_channel *chan, enum ast_page_mode page_mode)
{
	if (!chan) {
		return;
	}

	chan->page_mode = page_mode;
}

/*Start added by zyxia for support video paging,2022.8.5*/
void ast_channel_set_is_page_video_support(struct ast_channel *chan, int page_video_support)
{
	if (!chan) {
		return;
	}

	chan->page_video_support = page_video_support;
}

int ast_channel_get_is_page_video_support(struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->page_video_support;
}

void ast_channel_set_is_page_caller(struct ast_channel *chan,unsigned int is_page_caller)
{
	if (!chan) {
		return;
	}

	chan->is_page_caller = is_page_caller;
}

unsigned int ast_channel_get_is_page_caller(struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_page_caller;
}

void ast_channel_set_is_page_focus_record(struct ast_channel *chan, unsigned int is_page_focus_record)
{
	if (!chan) {
		return;
	}

	chan->is_page_focus_record = is_page_focus_record;
}

unsigned int ast_channel_get_is_page_focus_record(struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_page_focus_record;
}

/*End added by zyxia for support video paging,2022.8.5*/

const char *ast_channel_reject_reason(const struct ast_channel *chan)
{
	return chan->reject_reason;
}

void ast_channel_reject_reason_set(struct ast_channel *chan, const char *value)
{
	if (!chan) {
		return;
	}

	ast_string_field_set(chan, reject_reason, value);
}

const char *ast_channel_device_info(const struct ast_channel *chan)
{
	return chan->device_info;
}

void ast_channel_device_info_set(struct ast_channel *chan, const char *value)
{
	char *p = NULL;
	char agent_buf[256] = {0};
	if (value == NULL || value[0] == '\0')
	{
		memset(agent_buf, 0, sizeof(agent_buf));
	}
	/* if is the wave desktop [Grandstream Wave Desktop null (chrome 85.0.4183.121 windows)]*/
	else if (strcasestr(value, "Desktop") != NULL)
	{
		snprintf(agent_buf, sizeof(agent_buf), "Wave Desktop");
	}
	/* if is the wave web [Grandstream Wave 0.10.3.3 (chrome 87.0.4280.66 windows)]*/
	else if (strcasestr(value, "chrome") != NULL || strcasestr(value, "firefox") != NULL )
	{
		if (strcasestr(value, "chrome") != NULL)
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave Web/Chrome");
		}
		else
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave Web/Firefox");
		}
	}
	/* if is the new wave phone [Grandstream Wave Android 1.0.1.6(HONOR YAL-AL10,Android 10)]*/
	else if (strcasestr(value, " Android ") != NULL || strcasestr(value, " IOS ") != NULL)
	{
		if (strcasestr(value, " Android ") != NULL)
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave App/Android");
		}
		else
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave App/iOS");
		}
	}
	/* if is the old wave phone [Grandstream Wave x.x.x.x]*/
	else if (strcasestr(value, "Grandstream Wave ") != NULL)
	{
		snprintf(agent_buf, sizeof(agent_buf), "Wave App");
	}
	/* others remove the "Grandstream ", and buffer before the ' ' */
	else
	{
		p = strstr(value, "Grandstream ");
		if (p)
		{
			p = p + strlen("Grandstream ");
			snprintf(agent_buf, sizeof(agent_buf), "%s", p ? p : "");
		}
		else
		{
			snprintf(agent_buf, sizeof(agent_buf), "%s", value);
		}

		p = strstr(agent_buf, " ");
		if (p)
		{
			*p = '\0';
		}
	}
	ast_log(LOG_DEBUG, "device_info:%s", agent_buf);
	ast_string_field_set(chan, device_info, agent_buf);
}

const char *ast_channel_device_info_peer(const struct ast_channel *chan)
{
	return chan->device_info_peer;
}

void ast_channel_device_info_peer_set(struct ast_channel *chan, const char *value)
{
	char *p = NULL;
	char agent_buf[256] = {0};
	if (value == NULL || value[0] == '\0')
	{
		memset(agent_buf, 0, sizeof(agent_buf));
	}
	/* if is the wave desktop [Grandstream Wave Desktop null (chrome 85.0.4183.121 windows)]*/
	else if (strcasestr(value, "Desktop") != NULL)
	{
		snprintf(agent_buf, sizeof(agent_buf), "Wave Desktop");
	}
	/* if is the wave web [Grandstream Wave 0.10.3.3 (chrome 87.0.4280.66 windows)]*/
	else if (strcasestr(value, "chrome") != NULL || strcasestr(value, "firefox") != NULL )
	{
		if (strcasestr(value, "chrome") != NULL)
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave Web/Chrome");
		}
		else
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave Web/Firefox");
		}
	}
	/* if is the new wave phone [Grandstream Wave Android 1.0.1.6(HONOR YAL-AL10,Android 10)]*/
	else if (strcasestr(value, " Android ") != NULL || strcasestr(value, " IOS ") != NULL)
	{
		if (strcasestr(value, " Android ") != NULL)
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave App/Android");
		}
		else
		{
			snprintf(agent_buf, sizeof(agent_buf), "Wave App/iOS");
		}
	}
	/* if is the old wave phone [Grandstream Wave x.x.x.x]*/
	else if (strcasestr(value, "Grandstream Wave ") != NULL)
	{
		snprintf(agent_buf, sizeof(agent_buf), "Wave App");
	}
	/* others remove the "Grandstream ", and buffer before the ' ' */
	else
	{
		p = strstr(value, "Grandstream ");
		if (p)
		{
			p = p + strlen("Grandstream ");
			snprintf(agent_buf, sizeof(agent_buf), "%s", p ? p : "");
		}
		else
		{
			snprintf(agent_buf, sizeof(agent_buf), "%s", value);
		}

		p = strstr(agent_buf, " ");
		if (p)
		{
			*p = '\0';
		}
	}
	ast_log(LOG_DEBUG, "device_info:%s", agent_buf);
	ast_string_field_set(chan, device_info_peer, agent_buf);
}

enum ast_page_mode ast_channel_page_mode_get(struct ast_channel *chan)
{
	if (!chan) {
		return AST_PAGE_MODE_NONE;
	}

	return chan->page_mode;
}

unsigned int ast_channel_get_media_turn_relay(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_media_turn_relay) {
		return chan_tech->get_media_turn_relay(chan);
	}

	return 0;
}

unsigned int ast_channel_get_gswave_bundle_enabled(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_gswave_bundle_enabled) {
		return chan_tech->get_gswave_bundle_enabled(chan);
	}

	return 0;
}

unsigned int ast_channel_get_mobile_bundle_enabled(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_mobile_bundle_enabled) {
		return chan_tech->get_mobile_bundle_enabled(chan);
	}

	return 0;
}

void ast_channel_dahdi_bridge_tv_set(struct ast_channel *chan, struct timeval *value)
{
	chan->dahdi_bridge_tv = *value;
}

struct timeval *ast_channel_dahdi_bridge_tv(struct ast_channel *chan)
{
	return &chan->dahdi_bridge_tv;
}

int ast_channel_get_device_ua(struct ast_channel* chan, char* buf, size_t size)
{
	int ret = -1;
	const struct ast_channel_tech* tech = NULL;

	if (!chan || !buf || size <= 0) {
		return ret;
	}

	tech = ast_channel_tech(chan);
	if (tech && tech->get_session_ua) {
		ret = tech->get_session_ua(chan, buf, size);
	}

	return ret;
}

struct ast_stream_topology *ast_channel_get_participant_stream_topology(const struct ast_channel *chan)
{
	if (!chan) {
		return NULL;
	}

	return chan->stream_participant_topology;
}

void ast_channel_internal_set_participant_stream_topology(struct ast_channel *chan,
	struct ast_stream_topology *topology)
{
	if (!chan) {
		return;
	}

	ast_stream_topology_free(chan->stream_participant_topology);
	chan->stream_participant_topology = topology;

	return;
}

struct ast_stream_topology *ast_channel_set_participant_stream_topology(struct ast_channel *chan,
	struct ast_stream_topology *topology)
{
	struct ast_stream_topology *new_topology;

	if (!chan) {
		return NULL;
	}

	/* A non-MULTISTREAM channel can't manipulate topology directly */
	ast_assert(ast_channel_is_multistream(chan));

	/* Unless the channel is being destroyed, we always want a topology on
	 * it even if its empty.
	 */
	if (!topology) {
		new_topology = ast_stream_topology_alloc();
	} else {
		new_topology = topology;
	}

	if (new_topology) {
		ast_channel_internal_set_participant_stream_topology(chan, new_topology);
	}

	return new_topology;
}
/*BEGIN: Add by weidu ,for stir/shaken orig tn*/
void ast_channel_set_stir_shaken_origtn(struct ast_channel *chan, char *stir_sk_orig_tn)
{
	if (!chan || !stir_sk_orig_tn){
		return ;
	}
	memset(chan->stir_sk_orig_tn, 0, sizeof(chan->stir_sk_orig_tn));
	snprintf(chan->stir_sk_orig_tn, sizeof(chan->stir_sk_orig_tn)-1, "%s", stir_sk_orig_tn);
}

char *ast_channel_get_stir_shaken_origtn(struct ast_channel *chan)
{
	if (!chan){
		return NULL;
	}
	return chan->stir_sk_orig_tn;
}
/*END: Add by weidu, for stir/shaken*/
void ast_channel_set_vfax_filepath(struct ast_channel *chan, const char *file_path)
{
	if (!chan || !file_path){
		return ;
	}

	memset(chan->vfax_filepath, 0, sizeof(chan->vfax_filepath));
	snprintf(chan->vfax_filepath, sizeof(chan->vfax_filepath)-1, "%s", file_path);
}

char *ast_channel_get_vfax_filepath(struct ast_channel *chan)
{
	return chan->vfax_filepath;
}

void ast_channel_set_vfax_localstationid(struct ast_channel *chan, const char *localstationid)
{
	if (!chan || !localstationid){
		return ;
	}

	memset(chan->vfax_localstationid, 0, sizeof(chan->vfax_localstationid));
	snprintf(chan->vfax_localstationid, sizeof(chan->vfax_localstationid)-1, "%s", localstationid);
}

char *ast_channel_get_vfax_localstationid(struct ast_channel *chan)
{
	return chan->vfax_localstationid;
}

void ast_channel_set_vfax_headerinfo(struct ast_channel *chan, const char *headerinfo)
{
	if (!chan || !headerinfo){
		return ;
	}

	memset(chan->vfax_headerinfo, 0, sizeof(chan->vfax_headerinfo));
	snprintf(chan->vfax_headerinfo, sizeof(chan->vfax_headerinfo)-1, "%s", headerinfo);
}

char *ast_channel_get_vfax_headerinfo(struct ast_channel *chan)
{
	return chan->vfax_headerinfo;
}

/*Start added by zyxia for Bug 202354 ,2021.8.30*/
void ast_channel_set_sca_bridge_uniqueid(struct ast_channel *chan, const char *uniqueid)
{
	if (!chan || !uniqueid){
		return ;
	}

	memset(chan->sca_bridge_uniqueid, 0, sizeof(chan->sca_bridge_uniqueid));
	snprintf(chan->sca_bridge_uniqueid, sizeof(chan->sca_bridge_uniqueid)-1, "%s", uniqueid);
}

char *ast_channel_get_sca_bridge_uniqueid(struct ast_channel *chan)
{
	return chan->sca_bridge_uniqueid;
}
/*End added by zyxia for Bug 202354 ,2021.8.30*/

/*BEGIN: Add by weidu ,for 167810*/
void ast_channel_talk_detect_set(struct ast_channel *chan, unsigned int value)
{
	if (chan) {
		chan->talk_detect = value;
	}
}

unsigned int ast_channel_get_talk_detect(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}
	return chan->talk_detect;
}
/*END: Add by weidu ,for 167810*/

int ast_channel_get_participant_if_negotiate_state_done(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 1;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 1;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_participant_if_negotiate_state_done) {
		return chan_tech->get_participant_if_negotiate_state_done(chan);
	}

	return 1;
}

unsigned int ast_channel_get_delayed_notify_count(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_delayed_notify_count) {
		return chan_tech->get_delayed_notify_count(chan);
	}

	return 0;
}

void ast_channel_inc_delayed_notify_count(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->inc_delayed_notify_count) {
		chan_tech->inc_delayed_notify_count(chan);
	}

	return;
}

/*Start added by zyxia for Bug202204,2021.8.27*/
void ast_channel_decrease_turn_session_count(struct ast_channel *chan,unsigned int decrease_count)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return ;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return ;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->decrease_turn_session_count) {
		chan_tech->decrease_turn_session_count(chan,decrease_count);
	}

	return;
}
/*End added by zyxia for Bug202204,2021.8.27*/

int ast_channel_get_enable_mcu(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_enable_mcu) {
		return chan_tech->get_enable_mcu(chan);
	}

	return 0;
}

void ast_channel_recv_bye_set(struct ast_channel *chan, unsigned int value)
{
	if (chan) {
		chan->recv_bye = value;
	}

	return;
}

unsigned char ast_channel_get_recv_bye(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->recv_bye;
}

void ast_channel_dahdi_conf_set(struct ast_channel *chan, unsigned int value)
{
	if (chan) {
		chan->dahdi_conf = value;
	}

	return;
}

unsigned char ast_channel_get_dahdi_conf(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->dahdi_conf;
}

char *ast_channel_get_slides_video_local_msid(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return NULL;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return NULL;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_slides_video_local_msid) {
		 return chan_tech->get_slides_video_local_msid(chan);
	}

	return NULL;
}


char *ast_channel_recovery_linkedid(const struct ast_channel* chan)
{
	return chan->recovery_linkedid;
}

char *ast_channel_recovery_uniqueid(const struct ast_channel* chan)
{
	return chan->recovery_uniqueid;
}

void ast_channel_set_recovery_linkedid(struct ast_channel* chan ,char *lnkid)
{
	chan->recovery_linkedid = lnkid;
}

void ast_channel_set_recovery_uniqueid(struct ast_channel* chan ,char *uniqid)
{
	chan->recovery_uniqueid = uniqid;

}

void ast_channel_set_recovery_createtime(struct ast_channel* chan ,struct timeval *create)
{
	chan->recovery_createtime = *create;

}

struct timeval ast_channel_recovery_createtime(const struct ast_channel *chan)
{
	return chan->recovery_createtime;
}

struct timeval ast_channel_recovery_answertime(const struct ast_channel *chan)
{
	return chan->recovery_answertime;
}


void ast_channel_set_recovery_answertime(struct ast_channel* chan ,struct timeval *answer)
{
	chan->recovery_answertime = *answer;

}

const char *ast_channel_recovery_data(const struct ast_channel *chan)
{
	return chan->recovery_app_data;
}
void ast_channel_recovery_data_set(struct ast_channel *chan, const char *value)
{
	chan->recovery_app_data = (char *)value;
}

void ast_channel_already_recovery_set(struct ast_channel *chan, int value)
{
	chan->already_recovery = value;
}

int ast_channel_already_recovery_get(const struct ast_channel *chan)
{
	return chan->already_recovery;
}

void ast_channel_avoid_deltrack_flag_set(struct ast_channel *chan, char value)
{
	chan->avoid_delete_track = value;
}

char ast_channel_avoid_deltrack_flag_get(struct ast_channel *chan)
{
	return chan->avoid_delete_track;
}


void ast_channel_media_encryption_set(struct ast_channel *chan, struct ast_channel_media_encryption encryption)
{
	if (!chan) {
		return;
	}

	memcpy(&chan->encryption, &encryption, sizeof(chan->encryption));

	return;
}

void ast_channel_media_encryption_set_by_type(struct ast_channel *chan,
	enum ast_media_sub_type stream_type, int value)
{
	if (!chan) {
		return;
	}

	switch (stream_type) {
		case AST_MEDIA_SUB_AUDIO_TYPE:
			chan->encryption.audio_encryption = value;
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			chan->encryption.main_video_encryption = value;
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			chan->encryption.slides_video_encryption = value;
			break;
		default:
			break;
	}

	return;
}

enum ast_stream_media_encryption ast_channel_media_encryption_get(const struct ast_channel *chan, int stream_type)
{
	enum ast_stream_media_encryption encryption = 0;

	if (!chan) {
		return 0;
	}

	switch (stream_type) {
		case AST_MEDIA_SUB_AUDIO_TYPE:
			encryption = chan->encryption.audio_encryption;
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			encryption = chan->encryption.main_video_encryption;
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			encryption = chan->encryption.slides_video_encryption;
			break;
		default:
			break;
	}

	return encryption;
}
/*BEGIN: Add by weidu ,for181933 , no dahdi volume*/
void ast_channel_set_conference_volume_rxgain(struct ast_channel *chan, int value)
{
	if (!chan){
		return;
	}
	chan->dynamic_conf_rxgain = value;
}

int ast_channel_get_conference_volume_rxgain(struct ast_channel *chan)
{
	if (!chan){
		return 0;
	}
	return chan->dynamic_conf_rxgain;
}

void ast_channel_set_conference_volume_txgain(struct ast_channel *chan, int value)
{
	if (!chan){
		return;
	}
	chan->dynamic_conf_txgain = value;
}

int ast_channel_get_conference_volume_txgain(struct ast_channel *chan)
{
	if (!chan){
		return 0;
	}
	return chan->dynamic_conf_txgain;
}
/*END: Add by weidu*/

void ast_channel_set_recordstart_utc(struct ast_channel *chan, time_t value)
{
	if (!chan) {
		return;
	}

	chan->record_srtat_utc = value;
	ast_log(LOG_DEBUG, "Chan:%s set record_srtat_utc=%ld", ast_channel_name(chan), value);
}

void ast_channel_set_recordresume_utc(struct ast_channel *chan, time_t value)
{
	if (!chan) {
		return;
	}
	
	chan->record_resume_utc = value;
	ast_log(LOG_DEBUG, "Chane:%s set record_resume_utc=%ld", ast_channel_name(chan), value);
}

time_t ast_channel_get_recordstart_utc(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->record_srtat_utc;
}

time_t ast_channel_get_recordresume_utc(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}
	return chan->record_resume_utc;
}

void ast_channel_set_recordpaused_dur(struct ast_channel *chan, time_t value)
{
	if (!chan) {
		return;
	}
	
	chan->record_pause_duration += value;
	ast_log(LOG_DEBUG, "Chane:%s set record_resume_utc=%ld", ast_channel_name(chan), value);
}

time_t ast_channel_get_recordpaused_dur(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}
	return chan->record_pause_duration;
}

unsigned int ast_channel_ignore_avs_response(struct ast_channel *chan)
{
	return chan->ignore_avs_response;
}

void ast_channel_ignore_avs_response_set(struct ast_channel *chan, unsigned int value)
{
	chan->ignore_avs_response = value;
}

int ast_channel_already_hangup(struct ast_channel *chan)
{
	return chan->already_hangup;
}

void ast_channel_set_already_hangup(struct ast_channel *chan, int value)
{
	chan->already_hangup = value;
}

unsigned int ast_channel_nat_call_play_sound_timeout(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->nat_call_play_sound_timeout;
}

void ast_channel_nat_call_play_sound_timeout_set(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->nat_call_play_sound_timeout = value;

	return;
}

int ast_channel_get_limit_time_reached(struct ast_channel *chan)
{
	const struct ast_channel_tech *chan_tech = NULL;

	if (!chan) {
		return 0;
	}

	chan_tech = ast_channel_tech(chan);
	if (!chan_tech) {
		return 0;
	}

	if (0 == strcasecmp(chan_tech->type, "PJSIP") && chan_tech->get_limit_time_reached) {
		return chan_tech->get_limit_time_reached(chan);
	}

	return 0;
}

int ast_channel_set_is_webrtc_trunk(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_webrtc_trunk = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_webrtc_trunk:%d",
		ast_channel_name(chan), value);

	return 0;
}

int ast_channel_get_is_webrtc_trunk(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_webrtc_trunk;
}

struct ast_chan_move_hooks *ast_channel_move_hooks(const struct ast_channel *chan)
{
	return chan->move_hooks;
}

void ast_channel_move_hooks_set(struct ast_channel *chan, struct ast_chan_move_hooks *hooks, void *data)
{
	chan->move_hooks = hooks;
	chan->move_hook_data = data;
}

void *ast_channel_move_hook_data(const struct ast_channel *chan)
{
	return chan->move_hook_data;
}

int ast_channel_set_is_wave_app(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_wave_app = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_wave_app:%d",
		ast_channel_name(chan), value);

	return 0;
}

int ast_channel_get_is_wave_app(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_wave_app;
}

int ast_channel_get_is_in_calling_wave_app(const struct ast_channel *chan)
{
	if (!chan) {
		return -1;
	}

	return chan->is_in_calling_wave_app;
}

int ast_channel_set_is_in_calling_wave_app(struct ast_channel *chan, int value)
{
	if (!chan) {
		return -1;
	}

	chan->is_in_calling_wave_app = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_in_calling_wave_app:%d",
		ast_channel_name(chan), value);

	return 0;
}



void ast_channel_set_queue_reinvite_rsp_failed(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->queue_reinvite_rsp_failed = value;
	return;
}

unsigned int ast_channel_get_queue_reinvite_rsp_failed(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->queue_reinvite_rsp_failed;
}

void ast_channel_set_is_pickup(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->is_pickup = value;
	return;
}

unsigned int ast_channel_get_is_pickup(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_pickup;
}
void  ast_channel_set_is_join_trunk_conf(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return ;
	}

	chan->is_join_trunk_conf = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_join_trunk_conf:%d",
		ast_channel_name(chan), value);

	return ;
}

unsigned int ast_channel_get_is_join_trunk_conf(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_join_trunk_conf;
}

void ast_channel_set_virtual_conf_num(struct ast_channel *chan, const char *virtual_conf_num)
{
	memset(chan->virtual_conf_num, 0, sizeof(chan->virtual_conf_num));
	snprintf(chan->virtual_conf_num, sizeof(chan->virtual_conf_num)-1, "%s", virtual_conf_num);
}

char *ast_channel_get_virtual_conf_num(struct ast_channel *chan)
{
	return chan->virtual_conf_num;
}

unsigned int ast_channel_callerintercom_get(const struct ast_channel *chan)
{
	return chan->is_callerintercom;
}

void ast_channel_callerintercom_set(struct ast_channel *chan, unsigned int is_callerintercom)
{
	chan->is_callerintercom = is_callerintercom;
}

void ast_channel_open_file_set(struct ast_channel *chan, char *file)
{
	if (file != NULL && chan != NULL) {
		// ast_log(LOG_DEBUG, "file open:%s", file);
		chan->open_file = open(file, O_RDONLY, 0);
	}
}

int ast_channel_open_file_get(struct ast_channel *chan)
{
	if (chan != NULL) {
		return chan->open_file;
	}
	return 0;
}

void ast_channel_open_file_close(struct ast_channel *chan)
{
	if (chan != NULL && chan->open_file > 0) {
		// ast_log(LOG_DEBUG, "file close:%d", chan->open_file);
		if (!close(chan->open_file)) {
			chan->open_file = 0;
		}
		// ast_log(LOG_DEBUG, "file close:%d", chan->open_file);
	}
}

/*Start added by zyxia for queue answeed status report,2023.6.2*/
void  ast_channel_set_is_queue_call(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return ;
	}

	chan->is_queue_call = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_queue_call:%d",
		ast_channel_name(chan), value);

	return ;
}

unsigned int ast_channel_get_is_queue_call(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_queue_call;
}
/*End added by zyxia for queue answeed status report,2023.6.2*/

const char *ast_channel_get_avs_address(const struct ast_channel *chan)
{
	if (!chan) {
		return NULL;
	}

	return chan->avs_address;
}

void ast_channel_set_avs_address(struct ast_channel *chan, const char *avs_address)
{
	if (!chan || !avs_address) {
		return;
	}

	strncpy(chan->avs_address, avs_address, sizeof(chan->avs_address) - 1);

	return;
}

void ast_channel_set_avs_join_ghost(struct ast_channel *chan, int value)
{
	chan->join_ghost = value;
}

int ast_channel_get_avs_join_ghost(struct ast_channel *chan)
{
	return chan->join_ghost;
}

/*Start added by zyxia for Bug 268897,2023.8.16*/
unsigned int ast_channel_get_is_admin(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_admin;
}

void ast_channel_set_is_admin(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return ;
	}

	chan->is_admin = value;
	ast_log(LOG_DEBUG, "Chan:%s, set is_admin:%d",
		ast_channel_name(chan), value);

	return ;
}

int ast_channel_terminal_type(const struct ast_channel *chan)
{
       return chan->terminal_type;
}
void ast_channel_terminal_type_set(struct ast_channel *chan, int value)
{
       chan->terminal_type = value;
}
/*End added by zyxia for Bug 268897,2023.8.16*/

void ast_channel_set_mp3(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->mp3 = value;

	return;
}

unsigned int ast_channel_get_mp3(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->mp3;
}

void ast_channel_set_record_in_media(struct ast_channel *chan, unsigned int value)
{
	if (!chan) {
		return;
	}

	chan->record_in_media = value;

	return;
}

unsigned int ast_channel_get_record_in_media(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->record_in_media;
}

/*Start  added by zyxia for Bug 312085 ,2024.8.16*/
void ast_channel_set_file_encrypt_key(struct ast_channel *chan, const char *key)
{
	if (!chan || !key){
		return ;
	}

	memset(chan->file_encrypt_key, 0, sizeof(chan->file_encrypt_key));
	snprintf(chan->file_encrypt_key, sizeof(chan->file_encrypt_key)-1, "%s", key);
}

char *ast_channel_get_file_encrypt_key(struct ast_channel *chan)
{
	return chan->file_encrypt_key;
}

void ast_channel_set_file_encrypt_version(struct ast_channel *chan, const char *version)
{
	if (!chan || !version){
		return ;
	}

	memset(chan->file_encrypt_version, 0, sizeof(chan->file_encrypt_version));
	snprintf(chan->file_encrypt_version, sizeof(chan->file_encrypt_version)-1, "%s", version);
}

char *ast_channel_get_file_encrypt_version(struct ast_channel *chan)
{
	return chan->file_encrypt_version;
}
/*End  added by zyxia for Bug 312085 ,2024.8.16*/

void ast_channel_set_is_focus_record(struct ast_channel *chan, unsigned int is_focus_record)
{
	if (!chan) {
		return;
	}

	chan->is_focus_record = is_focus_record;

	return;
}

unsigned int ast_channel_get_is_focus_record(const struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	return chan->is_focus_record;
}

#endif
