/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief ASTdb Management
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \note DB3 is licensed under Sleepycat Public License and is thus incompatible
 * with GPL.  To avoid having to make another exception (and complicate
 * licensing even further) we elect to use DB1 which is BSD licensed
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/
#include "asterisk.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/haserver.h"
#endif


#include "asterisk/_private.h"
#include "asterisk/paths.h"	/* use ast_config_AST_DB */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <dirent.h>
#include <sqlite3.h>

#include "asterisk/channel.h"
#include "asterisk/file.h"
#include "asterisk/app.h"
#include "asterisk/dsp.h"
#include "asterisk/astdb.h"
#include "asterisk/cli.h"
#include "asterisk/utils.h"
#include "asterisk/manager.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/switch_xml.h"
AST_MODULE_LOG("database");
#endif

/*** DOCUMENTATION
	<manager name="DBGet" language="en_US">
		<synopsis>
			Get DB Entry.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="DBPut" language="en_US">
		<synopsis>
			Put DB entry.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" required="true" />
			<parameter name="Val" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="DBPutTree" language="en_US">
		<synopsis>
			Put DB entry batch.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Keys" required="true" />
			<parameter name="Vals" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="DBDel" language="en_US">
		<synopsis>
			Delete DB entry.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="DBDelTree" language="en_US">
		<synopsis>
			Delete DB Tree.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" />
		</syntax>
		<description>
		</description>
	</manager>
	 <manager name="UCMDBPut" language="en_US">
		<synopsis>
			Put DB entry.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" required="true" />
			<parameter name="Val" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="UCMDBPutTree" language="en_US">
		<synopsis>
			Put DB entry batch.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Keys" required="true" />
			<parameter name="Vals" />
		</syntax>
		<description>
		</description>
	</manager>
 	<manager name="UCMDBDelTree" language="en_US">
		<synopsis>
			Delete DB Tree.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" />
		</syntax>
		<description>
		</description>
	</manager>
 	<manager name="UCMDBDelAll" language="en_US">
		<synopsis>
			Delete DB All.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="RECOVERYDBPut" language="en_US">
		<synopsis>
			Delete DB.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Family" required="true" />
			<parameter name="Key" />
		</syntax>
		<description>
		</description>
	</manager>
 	<manager name="RECOVERYDBDelAll" language="en_US">
		<synopsis>
			Delete DB All.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
		</syntax>
		<description>
		</description>
	</manager>
 ***/

#define MAX_DB_FIELD 256
AST_MUTEX_DEFINE_STATIC(dblock);
static ast_cond_t dbcond;
static sqlite3 *astdb;
static pthread_t syncthread;
static int doexit;
static int dosync;

static void db_sync(void);

#ifdef GRANDSTREAM_NETWORKS

AST_MUTEX_DEFINE_STATIC(ucm_dblock);
static ast_cond_t ucm_dbcond;
static sqlite3 *ucm_astdb;
static pthread_t ucm_syncthread;
static int ucm_doexit;
static int ucm_dosync;

AST_MUTEX_DEFINE_STATIC(recovery_dblock);
static ast_cond_t recovery_dbcond;
static sqlite3 *recovery_astdb;
static pthread_t recovery_syncthread;
static int recovery_doexit;
static int recovery_dosync;

static void ucm_db_sync(void);

/* Add by bfdu for reduce load pjsip time, 2017.03.09  */
static struct ast_db_entry *db_entry_for_pjsip_load = NULL;
static int g_ast_db_ha_enable = 0;
#endif


#define DEFINE_SQL_STATEMENT(stmt,sql) static sqlite3_stmt *stmt; \
	const char stmt##_sql[] = sql;

DEFINE_SQL_STATEMENT(put_stmt, "INSERT OR REPLACE INTO astdb (key, value) VALUES (?, ?)")
DEFINE_SQL_STATEMENT(get_stmt, "SELECT value FROM astdb WHERE key=?")
DEFINE_SQL_STATEMENT(del_stmt, "DELETE FROM astdb WHERE key=?")
DEFINE_SQL_STATEMENT(deltree_stmt, "DELETE FROM astdb WHERE key || '/' LIKE ? || '/' || '%'")
DEFINE_SQL_STATEMENT(deltree_all_stmt, "DELETE FROM astdb")
DEFINE_SQL_STATEMENT(gettree_stmt, "SELECT key, value FROM astdb WHERE key || '/' LIKE ? || '/' || '%' ORDER BY key")
DEFINE_SQL_STATEMENT(gettree_all_stmt, "SELECT key, value FROM astdb ORDER BY key")
DEFINE_SQL_STATEMENT(showkey_stmt, "SELECT key, value FROM astdb WHERE key LIKE '%' || '/' || ? ORDER BY key")
DEFINE_SQL_STATEMENT(create_astdb_stmt, "CREATE TABLE IF NOT EXISTS astdb(key VARCHAR(256), value VARCHAR(256), PRIMARY KEY(key))")

/* This query begs an explanation:
 *
 * First, the parameter binding syntax used here is slightly different than the other
 * queries in that we use a numbered parameter so that we can bind once and get the same
 * value substituted multiple times within the executed query.
 *
 * Second, the key comparison is being used to find all keys that are lexicographically
 * greater than the provided key, but less than the provided key with a high (but
 * invalid) Unicode codepoint appended to it. This will give us all keys in the database
 * that have 'key' as a prefix and performs much better than the equivalent "LIKE key ||
 * '%'" operation.
 */
DEFINE_SQL_STATEMENT(gettree_prefix_stmt, "SELECT key, value FROM astdb WHERE key > ?1 AND key <= ?1 || X'ffff'")

static int init_stmt(sqlite3_stmt **stmt, const char *sql, size_t len)
{
	ast_mutex_lock(&dblock);
	if (sqlite3_prepare(astdb, sql, len, stmt, NULL) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't prepare statement '%s': %s\n", sql, sqlite3_errmsg(astdb));
		ast_mutex_unlock(&dblock);
		return -1;
	}
	ast_mutex_unlock(&dblock);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
DEFINE_SQL_STATEMENT(ucm_put_stmt, "INSERT OR REPLACE INTO astdb (key, value) VALUES (?, ?)")
DEFINE_SQL_STATEMENT(ucm_get_stmt, "SELECT value FROM astdb WHERE key=?")
DEFINE_SQL_STATEMENT(ucm_del_stmt, "DELETE FROM astdb WHERE key=?")
DEFINE_SQL_STATEMENT(ucm_deltree_stmt, "DELETE FROM astdb WHERE key || '/' LIKE ? || '/' || '%'")
DEFINE_SQL_STATEMENT(ucm_deltree_all_stmt, "DELETE FROM astdb")
DEFINE_SQL_STATEMENT(ucm_gettree_stmt, "SELECT key, value FROM astdb WHERE key || '/' LIKE ? || '/' || '%' ORDER BY key")
DEFINE_SQL_STATEMENT(ucm_gettree_all_stmt, "SELECT key, value FROM astdb ORDER BY key")
DEFINE_SQL_STATEMENT(ucm_showkey_stmt, "SELECT key, value FROM astdb WHERE key LIKE '%' || '/' || ? ORDER BY key")
DEFINE_SQL_STATEMENT(ucm_create_astdb_stmt, "CREATE TABLE IF NOT EXISTS astdb(key VARCHAR(256), value VARCHAR(256), PRIMARY KEY(key))")

static int ucm_init_stmt(sqlite3_stmt **stmt, const char *sql, size_t len)
{
	ast_mutex_lock(&ucm_dblock);
	if (sqlite3_prepare(ucm_astdb, sql, len, stmt, NULL) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't prepare statement '%s': %s\n", sql, sqlite3_errmsg(ucm_astdb));
		ast_mutex_unlock(&ucm_dblock);
		return -1;
	}
	ast_mutex_unlock(&ucm_dblock);

	return 0;
}

static int ucm_clean_stmt(sqlite3_stmt **stmt, const char *sql)
{
	if (sqlite3_finalize(*stmt) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't finalize statement '%s': %s\n", sql, sqlite3_errmsg(ucm_astdb));
		*stmt = NULL;
		return -1;
	}
	*stmt = NULL;
	return 0;
}

static void ucm_clean_statements(void)
{
	ucm_clean_stmt(&ucm_get_stmt, ucm_get_stmt_sql);
	ucm_clean_stmt(&ucm_del_stmt, ucm_del_stmt_sql);
	ucm_clean_stmt(&ucm_deltree_stmt, ucm_deltree_stmt_sql);
	ucm_clean_stmt(&ucm_deltree_all_stmt, ucm_deltree_all_stmt_sql);
	ucm_clean_stmt(&ucm_gettree_stmt, ucm_gettree_stmt_sql);
	ucm_clean_stmt(&ucm_gettree_all_stmt, ucm_gettree_all_stmt_sql);
	ucm_clean_stmt(&ucm_showkey_stmt, ucm_showkey_stmt_sql);
	ucm_clean_stmt(&ucm_put_stmt, ucm_put_stmt_sql);
	ucm_clean_stmt(&ucm_create_astdb_stmt, ucm_create_astdb_stmt_sql);
}

static int ucm_init_statements(void)
{
	/* Don't initialize create_astdb_statment here as the astdb table needs to exist
	 * brefore these statments can be initialized */
	return ucm_init_stmt(&ucm_get_stmt, ucm_get_stmt_sql, sizeof(ucm_get_stmt_sql))
	|| ucm_init_stmt(&ucm_del_stmt, ucm_del_stmt_sql, sizeof(ucm_del_stmt_sql))
	|| ucm_init_stmt(&ucm_deltree_stmt, ucm_deltree_stmt_sql, sizeof(ucm_deltree_stmt_sql))
	|| ucm_init_stmt(&ucm_deltree_all_stmt, ucm_deltree_all_stmt_sql, sizeof(ucm_deltree_all_stmt_sql))
	|| ucm_init_stmt(&ucm_gettree_stmt, ucm_gettree_stmt_sql, sizeof(ucm_gettree_stmt_sql))
	|| ucm_init_stmt(&ucm_gettree_all_stmt, ucm_gettree_all_stmt_sql, sizeof(ucm_gettree_all_stmt_sql))
	|| ucm_init_stmt(&ucm_showkey_stmt, ucm_showkey_stmt_sql, sizeof(ucm_showkey_stmt_sql))
	|| ucm_init_stmt(&ucm_put_stmt, ucm_put_stmt_sql, sizeof(ucm_put_stmt_sql));
}

static int ucm_db_create_astdb(void)
{
	int res = 0;

	if (!ucm_create_astdb_stmt) {
		ucm_init_stmt(&ucm_create_astdb_stmt, ucm_create_astdb_stmt_sql, sizeof(ucm_create_astdb_stmt_sql));
	}

	ast_mutex_lock(&ucm_dblock);
	if (sqlite3_step(ucm_create_astdb_stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't create astdb table: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	}
	sqlite3_reset(ucm_create_astdb_stmt);
	ucm_db_sync();
	ast_mutex_unlock(&ucm_dblock);

	return res;
}

static int ucm_db_open(void)
{
	char *dbname;

	if (!(dbname = ast_alloca(strlen(ast_config_AST_DB) + sizeof(".sqlite3")+sizeof("ucm_")))) {
		return -1;
	}
	strcpy(dbname, ast_config_AST_DB);
	dbname[strlen(ast_config_AST_DB)-5]='\0';
	strcat(dbname, "ucm_astdb.sqlite3");

	ast_mutex_lock(&ucm_dblock);
	if (sqlite3_open(dbname, &ucm_astdb) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Unable to open Asterisk database '%s': %s\n", dbname, sqlite3_errmsg(ucm_astdb));
		sqlite3_close(ucm_astdb);
		ast_mutex_unlock(&ucm_dblock);
		return -1;
	}

	ast_mutex_unlock(&ucm_dblock);

	return 0;
}

static int ucm_db_init(void)
{
	if (ucm_astdb) {
		return 0;
	}

	if (ucm_db_open() || ucm_db_create_astdb() || ucm_init_statements()) {
		return -1;
	}

	return 0;
}

/* We purposely don't lock around the sqlite3 call because the transaction
 * calls will be called with the database lock held. For any other use, make
 * sure to take the dblock yourself. */
static int ucm_db_execute_sql(const char *sql, int (*callback)(void *, int, char **, char **), void *arg)
{
	char *errmsg = NULL;
	int res =0;

	if (sqlite3_exec(ucm_astdb, sql, callback, arg, &errmsg) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Error executing SQL (%s): %s\n", sql, errmsg);
		sqlite3_free(errmsg);
		res = -1;
	}

	return res;
}

static int ast_ucm_db_begin_transaction(void)
{
	return ucm_db_execute_sql("BEGIN TRANSACTION", NULL, NULL);
}

static int ast_ucm_db_commit_transaction(void)
{
	return ucm_db_execute_sql("COMMIT", NULL, NULL);
}

static int ast_ucm_db_rollback_transaction(void)
{
	return ucm_db_execute_sql("ROLLBACK", NULL, NULL);
}

static int ast_ucm_db_delall(void)
{
	return ucm_db_execute_sql(ucm_deltree_all_stmt_sql, NULL, NULL);
}

int ast_ucm_db_put(const char *family, const char *key, const char *value ,int iscommit)
{
	char fullkey[MAX_DB_FIELD];
	size_t fullkey_len;
	int res = 0;

	if (strlen(family) + strlen(key) + 2 > sizeof(fullkey) - 1) {
		ast_log(LOG_WARNING, "Family and key length must be less than %zu bytes\n", sizeof(fullkey) - 3);
		return -1;
	}

	fullkey_len = snprintf(fullkey, sizeof(fullkey), "/%s/%s", family, key);

	ast_mutex_lock(&ucm_dblock);
	if (sqlite3_bind_text(ucm_put_stmt, 1, fullkey, fullkey_len, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	} else if (sqlite3_bind_text(ucm_put_stmt, 2, value, -1, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind value to stmt: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	} else if (sqlite3_step(ucm_put_stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't execute statment: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	}

	sqlite3_reset(ucm_put_stmt);
    if(iscommit){
    	ucm_db_sync();
    }
	ast_mutex_unlock(&ucm_dblock);

	if (0 == res)
    {
        if(iscommit)
        {
            ast_ha_send_astdb(UCM_DB_HA_PUT_CMT, family, key, value);
        }
        else
        {
            ast_ha_send_astdb(UCM_DB_HA_PUT_NOCMT, family, key, value);
        }
    }

	return res;
}

/*!
 * \internal
 * \brief Get key value specified by family/key.
 *
 * Gets the value associated with the specified \a family and \a key, and
 * stores it, either into the fixed sized buffer specified by \a buffer
 * and \a bufferlen, or as a heap allocated string if \a bufferlen is -1.
 *
 * \note If \a bufferlen is -1, \a buffer points to heap allocated memory
 *       and must be freed by calling ast_free().
 *
 * \retval -1 An error occurred
 * \retval 0 Success
 */
static int ucm_db_get_common(const char *family, const char *key, char **buffer, int bufferlen)
{
	const unsigned char *result;
	char fullkey[MAX_DB_FIELD];
	size_t fullkey_len;
	int res = 0;

	if (strlen(family) + strlen(key) + 2 > sizeof(fullkey) - 1) {
		ast_log(LOG_WARNING, "Family and key length must be less than %zu bytes\n", sizeof(fullkey) - 3);
		return -1;
	}

	fullkey_len = snprintf(fullkey, sizeof(fullkey), "/%s/%s", family, key);

	ast_mutex_lock(&ucm_dblock);
	if (sqlite3_bind_text(ucm_get_stmt, 1, fullkey, fullkey_len, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	} else if (sqlite3_step(ucm_get_stmt) != SQLITE_ROW) {
		ast_debug(1, "Unable to find key '%s' in family '%s'\n", key, family);
		res = -1;
	} else if (!(result = sqlite3_column_text(ucm_get_stmt, 0))) {
		ast_log(LOG_WARNING, "Couldn't get value\n");
		res = -1;
	} else {
		const char *value = (const char *) result;

		if (bufferlen == -1) {
			*buffer = ast_strdup(value);
		} else {
			ast_copy_string(*buffer, value, bufferlen);
		}
	}
	sqlite3_reset(ucm_get_stmt);
	ast_mutex_unlock(&ucm_dblock);

	return res;
}

int ast_ucm_db_get(const char *family, const char *key, char *value, int valuelen)
{
	ast_assert(value != NULL);

	/* Make sure we initialize */
	value[0] = 0;

	return ucm_db_get_common(family, key, &value, valuelen);
}

int ast_ucm_db_get_allocated(const char *family, const char *key, char **out)
{
	*out = NULL;

	return ucm_db_get_common(family, key, out, -1);
}

int ast_ucm_db_del(const char *family, const char *key)
{
	char fullkey[MAX_DB_FIELD];
	size_t fullkey_len;
	int res = 0;

	if (strlen(family) + strlen(key) + 2 > sizeof(fullkey) - 1) {
		ast_log(LOG_WARNING, "Family and key length must be less than %zu bytes\n", sizeof(fullkey) - 3);
		return -1;
	}

	fullkey_len = snprintf(fullkey, sizeof(fullkey), "/%s/%s", family, key);

	ast_mutex_lock(&ucm_dblock);
	if (sqlite3_bind_text(ucm_del_stmt, 1, fullkey, fullkey_len, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	} else if (sqlite3_step(ucm_del_stmt) != SQLITE_DONE) {
		ast_debug(1, "Unable to find key '%s' in family '%s'\n", key, family);
		res = -1;
	}
	sqlite3_reset(ucm_del_stmt);
	ucm_db_sync();
	ast_mutex_unlock(&ucm_dblock);
	if (0 == res)
    {
        ast_ha_send_astdb(UCM_DB_HA_DEL, family, key, NULL);
    }
	return res;
}

int ast_ucm_db_deltree(const char *family, const char *keytree)
{
	sqlite3_stmt *stmt = ucm_deltree_stmt;
	char prefix[MAX_DB_FIELD];
	int res = 0;

	if (!ast_strlen_zero(family)) {
		if (!ast_strlen_zero(keytree)) {
			/* Family and key tree */
			snprintf(prefix, sizeof(prefix), "/%s/%s", family, keytree);
		} else {
			/* Family only */
			snprintf(prefix, sizeof(prefix), "/%s", family);
		}
	} else {
		prefix[0] = '\0';
		stmt = ucm_deltree_all_stmt;
	}

	ast_mutex_lock(&ucm_dblock);
	if (!ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", prefix, sqlite3_errmsg(ucm_astdb));
		res = -1;
	} else if (sqlite3_step(stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't execute stmt: %s\n", sqlite3_errmsg(ucm_astdb));
		res = -1;
	}
	//res = sqlite3_changes(ucm_astdb);
	sqlite3_reset(stmt);
	ucm_db_sync();
	ast_mutex_unlock(&ucm_dblock);
	if (0 == res)
    {
        ast_ha_send_astdb(UCM_DB_HA_DELTREE, family, keytree, NULL);
    }
	return res;
}

struct ast_db_entry *ast_ucm_db_gettree(const char *family, const char *keytree)
{
	char prefix[MAX_DB_FIELD];
	sqlite3_stmt *stmt = ucm_gettree_stmt;
	struct ast_db_entry *cur, *last = NULL, *ret = NULL;

	if (!ast_strlen_zero(family)) {
		if (!ast_strlen_zero(keytree)) {
			/* Family and key tree */
			snprintf(prefix, sizeof(prefix), "/%s/%s", family, keytree);
		} else {
			/* Family only */
			snprintf(prefix, sizeof(prefix), "/%s", family);
		}
	} else {
		prefix[0] = '\0';
		stmt = ucm_gettree_all_stmt;
	}

	ast_mutex_lock(&ucm_dblock);
	if (!ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", prefix, sqlite3_errmsg(ucm_astdb));
		sqlite3_reset(stmt);
		ast_mutex_unlock(&ucm_dblock);
		return NULL;
	}

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		const char *key_s, *value_s;
		if (!(key_s = (const char *) sqlite3_column_text(stmt, 0))) {
			break;
		}
		if (!(value_s = (const char *) sqlite3_column_text(stmt, 1))) {
			break;
		}
		if (!(cur = ast_malloc(sizeof(*cur) + strlen(key_s) + strlen(value_s) + 2))) {
			break;
		}
		cur->next = NULL;
		cur->key = cur->data + strlen(value_s) + 1;
		strcpy(cur->data, value_s);
		strcpy(cur->key, key_s);
		if (last) {
			last->next = cur;
		} else {
			ret = cur;
		}
		last = cur;
	}
	sqlite3_reset(stmt);
	ast_mutex_unlock(&ucm_dblock);

	return ret;
}

/* Add by bfdu for reduce load pjsip time, 2017.03.09 */
void ast_db_gettree_for_pjsip_load(void)
{
	if (!g_ast_db_ha_enable) {
		ast_log(LOG_DEBUG, "ast_db_gettree_for_pjsip_load HA is not enable\n");
		return;
	}
	db_entry_for_pjsip_load = ast_db_gettree("registrar/contact", NULL);
	return;
}

void ast_db_freetree_for_pjsip_load(void)
{
	if (!g_ast_db_ha_enable) {
		ast_log(LOG_DEBUG, "ast_db_freetree_for_pjsip_load HA is not enable\n");
		return;
	}
	ast_db_freetree(db_entry_for_pjsip_load);
	db_entry_for_pjsip_load = NULL;
	return;
}

void ast_db_set_ha_enable_for_pjsip_load(void)
{
	FILE *ha_fp = NULL;
	char on_buf[64] = { 0 };
	    
	/* Get the ha_enable */
	ha_fp = popen("nvram get haon", "r");
	if (ha_fp == NULL)
	{
		ast_log(LOG_ERROR, "ast_db haon Read error\n");
		return;
	}
	while (ha_fp != NULL && (!feof(ha_fp)) && (fgets(on_buf, sizeof(on_buf) - 1, ha_fp) != NULL)) 
	{
		ast_log(LOG_DEBUG, "ast_db haon Read: [%s]\n", on_buf);
	}
	g_ast_db_ha_enable = atoi(on_buf);

	ast_log(LOG_DEBUG, "ast_db ha_enable is %d\n", g_ast_db_ha_enable);

	pclose(ha_fp);
	ha_fp = NULL;
	return;
}
/* End by bfdu */
#endif

#ifdef GRANDSTREAM_NETWORKS

DEFINE_SQL_STATEMENT(recovery_create_astdb_stmt, "CREATE TABLE IF NOT EXISTS recovery ( runtime_uuid VARCHAR(255), technology VARCHAR(255), profile_name VARCHAR(255), hostname VARCHAR(255), uuid VARCHAR(255), metadata text)")
DEFINE_SQL_STATEMENT(recovery_put_stmt, "INSERT OR REPLACE INTO recovery (runtime_uuid, technology, profile_name, hostname, uuid, metadata) VALUES (?, ?, ?, ?, ?, ?)")
DEFINE_SQL_STATEMENT(recovery_get_all_stmt, "SELECT runtime_uuid, uuid, metadata FROM recovery ORDER BY runtime_uuid")
DEFINE_SQL_STATEMENT(recovery_get_stmt, "SELECT metadata FROM recovery WHERE runtime_uuid=? and uuid=?")

static int recovery_init_stmt(sqlite3_stmt **stmt, const char *sql, size_t len)
{
	ast_mutex_lock(&recovery_dblock);
	if (sqlite3_prepare(recovery_astdb, sql, len, stmt, NULL) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't prepare statement '%s': %s\n", sql, sqlite3_errmsg(recovery_astdb));
		ast_mutex_unlock(&recovery_dblock);
		return -1;
	}
	ast_mutex_unlock(&recovery_dblock);

	return 0;
}

static int recovery_clean_stmt(sqlite3_stmt **stmt, const char *sql)
{
	if (sqlite3_finalize(*stmt) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't finalize statement '%s': %s\n", sql, sqlite3_errmsg(recovery_astdb));
		*stmt = NULL;
		return -1;
	}
	*stmt = NULL;
	return 0;
}

static void recovery_clean_statements(void)
{
	recovery_clean_stmt(&recovery_create_astdb_stmt, recovery_create_astdb_stmt_sql);
	recovery_clean_stmt(&recovery_put_stmt, recovery_put_stmt_sql);
	recovery_clean_stmt(&recovery_get_stmt, recovery_get_stmt_sql);
	recovery_clean_stmt(&recovery_get_all_stmt, recovery_get_all_stmt_sql);
}

static int recovery_init_statements(void)
{
	/* Don't initialize create_astdb_statment here as the astdb table needs to exist
	 * brefore these statments can be initialized */
	return recovery_init_stmt(&recovery_put_stmt, recovery_put_stmt_sql, sizeof(recovery_put_stmt_sql))
	|| recovery_init_stmt(&recovery_get_stmt, recovery_get_stmt_sql, sizeof(recovery_get_stmt_sql))
	|| recovery_init_stmt(&recovery_get_all_stmt, recovery_get_all_stmt_sql, sizeof(recovery_get_all_stmt_sql));
}

static void recovery_db_sync(void)
{
	recovery_dosync = 1;
	ast_cond_signal(&recovery_dbcond);
}

static int recovery_db_execute_sql(const char *sql, int (*callback)(void *, int, char **, char **), void *arg)
{
	char *errmsg = NULL;
	int res =0;

	if (sqlite3_exec(recovery_astdb, sql, callback, arg, &errmsg) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Error executing SQL (%s): %s\n", sql, errmsg);
		sqlite3_free(errmsg);
		res = -1;
	}

	return res;
}

static int ast_recovery_db_begin_transaction(void)
{
	return recovery_db_execute_sql("BEGIN TRANSACTION", NULL, NULL);
}

static int ast_recovery_db_commit_transaction(void)
{
	return recovery_db_execute_sql("COMMIT", NULL, NULL);
}

static int ast_recovery_db_rollback_transaction(void)
{
	return recovery_db_execute_sql("ROLLBACK", NULL, NULL);
}

static void *recovery_db_sync_thread(void *data)
{
	ast_mutex_lock(&recovery_dblock);
	ast_recovery_db_begin_transaction();
	for (;;) {
		/* If dosync is set, db_sync() was called during sleep(1), 
		 * and the pending transaction should be committed. 
		 * Otherwise, block until db_sync() is called.
		 */
		while (!recovery_dosync) {
			ast_cond_wait(&recovery_dbcond, &recovery_dblock);
		}
		recovery_dosync = 0;
		if (ast_recovery_db_commit_transaction()) {
			ast_recovery_db_rollback_transaction();
		}
		if (recovery_doexit) {
			ast_mutex_unlock(&recovery_dblock);
			break;
		}
		ast_recovery_db_begin_transaction();
		ast_mutex_unlock(&recovery_dblock);
		sleep(1);
		ast_mutex_lock(&recovery_dblock);
	}

	return NULL;
}


static int recovery_db_create_astdb(void)
{
	int res = 0;

	if (!recovery_create_astdb_stmt) {
		recovery_init_stmt(&recovery_create_astdb_stmt, recovery_create_astdb_stmt_sql, sizeof(recovery_create_astdb_stmt_sql));
	}

	ast_mutex_lock(&recovery_dblock);
	if (sqlite3_step(recovery_create_astdb_stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't create astdb table: %s\n", sqlite3_errmsg(recovery_astdb));
		res = -1;
	}
	sqlite3_reset(recovery_create_astdb_stmt);
	recovery_db_sync();
	ast_mutex_unlock(&recovery_dblock);

	return res;
}

static int recovery_db_open(void)
{
	char *dbname;

	if (!(dbname = ast_alloca(strlen(ast_config_AST_DB) + sizeof(".sqlite3")+sizeof("recovery_")))) {
		return -1;
	}
	strcpy(dbname, ast_config_AST_DB);
	dbname[strlen(ast_config_AST_DB)-5]='\0';
	strcat(dbname, "recovery_astdb.sqlite3");

	ast_mutex_lock(&recovery_dblock);
	if (sqlite3_open(dbname, &recovery_astdb) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Unable to open Asterisk database '%s': %s\n", dbname, sqlite3_errmsg(recovery_astdb));
		sqlite3_close(recovery_astdb);
		ast_mutex_unlock(&recovery_dblock);
		return -1;
	}

	ast_mutex_unlock(&recovery_dblock);

	return 0;
}

static int recovery_db_init(void)
{
	if (recovery_astdb) {
		return 0;
	}

	if (recovery_db_open() || recovery_db_create_astdb() || recovery_init_statements()) {
		return -1;
	}

	return 0;
}

int ast_recovery_db_put(const char* sql, int iscommit, int issync)
{
	int res = 0;
	char *errmsg = NULL;
	ast_mutex_lock(&recovery_dblock);

	if (sqlite3_exec(recovery_astdb, sql, NULL, NULL, &errmsg) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Error executing SQL (%s): %s\n", sql, errmsg);
		sqlite3_free(errmsg);
		res = -1;
	}

	if(iscommit){
		recovery_db_sync();
	}

#ifdef GRANDSTREAM_NETWORKS
	if (issync){
		ast_ha_send_recoveryastdb_sql((char* )sql); /*recovery by weidu ,for ha sql sync*/
	}
#endif
	
	ast_mutex_unlock(&recovery_dblock);
	return res;
}

static int recovery_db_get_common(const char* runtime_uuid, const char* uuid, char **buffer, int bufferlen)
{
	const unsigned char *result;
	int res = 0;

	if (NULL == runtime_uuid || strlen(runtime_uuid) == 0) {
		ast_log(LOG_WARNING, "runtiome_uuid is NULL\n");
		return -1;
	}

	if (NULL == uuid || strlen(uuid) == 0) {
		ast_log(LOG_WARNING, "uuid is NULL\n");
		return -1;
	}

	ast_mutex_lock(&recovery_dblock);
	if (sqlite3_bind_text(recovery_get_stmt, 1, runtime_uuid, strlen(runtime_uuid), SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(recovery_astdb));
		res = -1;
	} else if (sqlite3_bind_text(recovery_get_stmt, 2, uuid, strlen(uuid), SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(recovery_astdb));
		res = -1;
	} else if (sqlite3_step(recovery_get_stmt) != SQLITE_ROW) {
		ast_debug(1, "Unable to find runtiome_uuid '%s' in uuid '%s' [%s]\n", runtime_uuid, uuid, sqlite3_errmsg(recovery_astdb));
		res = -1;
	} else if (!(result = sqlite3_column_text(recovery_get_stmt, 0))) {
		ast_log(LOG_WARNING, "Couldn't get value\n");
		res = -1;
	} else {
		const char *value = (const char *) result;

		if (bufferlen == -1) {
			*buffer = ast_strdup(value);
		} else {
			ast_copy_string(*buffer, value, bufferlen);
		}
	}
	sqlite3_reset(recovery_get_stmt);
	ast_mutex_unlock(&recovery_dblock);

	return res;
}

int ast_recovery_db_get(const char* runtime_uuid, const char* uuid, char* value, int valuelen)
{
	ast_assert(value != NULL);

	/* Make sure we initialize */
	value[0] = 0;

	return recovery_db_get_common(runtime_uuid, uuid, &value, valuelen);
}

struct recovery_db_entry *ast_recovery_db_get_all(void)
{
	char prefix[MAX_DB_FIELD] = {0};
	sqlite3_stmt *stmt = recovery_get_all_stmt;
	struct recovery_db_entry *cur, *last = NULL, *ret = NULL;

	ast_mutex_lock(&recovery_dblock);
	if (!ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", prefix, sqlite3_errmsg(recovery_astdb));
		sqlite3_reset(stmt);
		ast_mutex_unlock(&recovery_dblock);
		return NULL;
	}

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		const char *runtime_uuid_s, *uuid_s, *data_s;
		if (!(runtime_uuid_s = (const char *) sqlite3_column_text(stmt, 0))) {
			break;
		}
		if (!(uuid_s = (const char *) sqlite3_column_text(stmt, 1))) {
			break;
		}
		if (!(data_s = (const char *) sqlite3_column_text(stmt, 2))) {
			break;
		}
		if (!(cur = ast_malloc(sizeof(*cur) + strlen(runtime_uuid_s) + strlen(uuid_s) + strlen(data_s) + 3))) {
			break;
		}
		cur->next = NULL;
		cur->uuid = cur->data + strlen(data_s) + 1;
		cur->runtime_uuid = cur->data + strlen(data_s) + strlen(uuid_s) + 2;
		strcpy(cur->data, data_s);
		strcpy(cur->uuid, uuid_s);
		strcpy(cur->runtime_uuid, runtime_uuid_s);
		if (last) {
			last->next = cur;
		} else {
			ret = cur;
		}
		last = cur;
	}
	sqlite3_reset(stmt);
	ast_mutex_unlock(&recovery_dblock);

	return ret;
}

void ast_recovery_freetree(struct recovery_db_entry *dbe)
{
	struct recovery_db_entry *last;
	while (dbe) {
		last = dbe;
		dbe = dbe->next;
		ast_free(last);
	}
}
#endif

/*! \internal
 * \brief Clean up the prepared SQLite3 statement
 * \note dblock should already be locked prior to calling this method
 */
static int clean_stmt(sqlite3_stmt **stmt, const char *sql)
{
	if (sqlite3_finalize(*stmt) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't finalize statement '%s': %s\n", sql, sqlite3_errmsg(astdb));
		*stmt = NULL;
		return -1;
	}
	*stmt = NULL;
	return 0;
}

/*! \internal
 * \brief Clean up all prepared SQLite3 statements
 * \note dblock should already be locked prior to calling this method
 */
static void clean_statements(void)
{
	clean_stmt(&get_stmt, get_stmt_sql);
	clean_stmt(&del_stmt, del_stmt_sql);
	clean_stmt(&deltree_stmt, deltree_stmt_sql);
	clean_stmt(&deltree_all_stmt, deltree_all_stmt_sql);
	clean_stmt(&gettree_stmt, gettree_stmt_sql);
	clean_stmt(&gettree_all_stmt, gettree_all_stmt_sql);
	clean_stmt(&gettree_prefix_stmt, gettree_prefix_stmt_sql);
	clean_stmt(&showkey_stmt, showkey_stmt_sql);
	clean_stmt(&put_stmt, put_stmt_sql);
	clean_stmt(&create_astdb_stmt, create_astdb_stmt_sql);
}

static int init_statements(void)
{
	/* Don't initialize create_astdb_statement here as the astdb table needs to exist
	 * brefore these statements can be initialized */
	return init_stmt(&get_stmt, get_stmt_sql, sizeof(get_stmt_sql))
	|| init_stmt(&del_stmt, del_stmt_sql, sizeof(del_stmt_sql))
	|| init_stmt(&deltree_stmt, deltree_stmt_sql, sizeof(deltree_stmt_sql))
	|| init_stmt(&deltree_all_stmt, deltree_all_stmt_sql, sizeof(deltree_all_stmt_sql))
	|| init_stmt(&gettree_stmt, gettree_stmt_sql, sizeof(gettree_stmt_sql))
	|| init_stmt(&gettree_all_stmt, gettree_all_stmt_sql, sizeof(gettree_all_stmt_sql))
	|| init_stmt(&gettree_prefix_stmt, gettree_prefix_stmt_sql, sizeof(gettree_prefix_stmt_sql))
	|| init_stmt(&showkey_stmt, showkey_stmt_sql, sizeof(showkey_stmt_sql))
	|| init_stmt(&put_stmt, put_stmt_sql, sizeof(put_stmt_sql));
}

static int convert_bdb_to_sqlite3(void)
{
	char *cmd;
	int res;

	res = ast_asprintf(&cmd, "%s/astdb2sqlite3 '%s'\n", ast_config_AST_SBIN_DIR, ast_config_AST_DB);
	if (0 <= res) {
		res = ast_safe_system2(cmd);
		ast_free(cmd);
	}

	return res;
}

static int db_create_astdb(void)
{
	int res = 0;

	if (!create_astdb_stmt) {
		init_stmt(&create_astdb_stmt, create_astdb_stmt_sql, sizeof(create_astdb_stmt_sql));
	}

	ast_mutex_lock(&dblock);
	if (sqlite3_step(create_astdb_stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't create astdb table: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	}
	sqlite3_reset(create_astdb_stmt);
	db_sync();
	ast_mutex_unlock(&dblock);

	return res;
}

static int db_open(void)
{
	char *dbname;
	struct stat dont_care;

	if (!(dbname = ast_alloca(strlen(ast_config_AST_DB) + sizeof(".sqlite3")))) {
		return -1;
	}
	strcpy(dbname, ast_config_AST_DB);
	strcat(dbname, ".sqlite3");

	if (stat(dbname, &dont_care) && !stat(ast_config_AST_DB, &dont_care)) {
		if (convert_bdb_to_sqlite3()) {
			ast_log(LOG_ERROR, "*** Database conversion failed!\n");
			ast_log(LOG_ERROR, "*** Asterisk now uses SQLite3 for its internal\n");
			ast_log(LOG_ERROR, "*** database. Conversion from the old astdb\n");
			ast_log(LOG_ERROR, "*** failed. Most likely the astdb2sqlite3 utility\n");
			ast_log(LOG_ERROR, "*** was not selected for build. To convert the\n");
			ast_log(LOG_ERROR, "*** old astdb, please delete '%s'\n", dbname);
			ast_log(LOG_ERROR, "*** and re-run 'make menuselect' and select astdb2sqlite3\n");
			ast_log(LOG_ERROR, "*** in the Utilities section, then 'make && make install'.\n");
			ast_log(LOG_ERROR, "*** It is also imperative that the user under which\n");
			ast_log(LOG_ERROR, "*** Asterisk runs have write permission to the directory\n");
			ast_log(LOG_ERROR, "*** where the database resides.\n");
			sleep(5);
		} else {
			ast_log(LOG_NOTICE, "Database conversion succeeded!\n");
		}
	}

	ast_mutex_lock(&dblock);
	if (sqlite3_open(dbname, &astdb) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Unable to open Asterisk database '%s': %s\n", dbname, sqlite3_errmsg(astdb));
		sqlite3_close(astdb);
		ast_mutex_unlock(&dblock);
		return -1;
	}

	ast_mutex_unlock(&dblock);

	return 0;
}

static int db_init(void)
{
	if (astdb) {
		return 0;
	}

	if (db_open() || db_create_astdb() || init_statements()) {
		return -1;
	}

	return 0;
}

/* We purposely don't lock around the sqlite3 call because the transaction
 * calls will be called with the database lock held. For any other use, make
 * sure to take the dblock yourself. */
static int db_execute_sql(const char *sql, int (*callback)(void *, int, char **, char **), void *arg)
{
	char *errmsg = NULL;
	int res =0;

	if (sqlite3_exec(astdb, sql, callback, arg, &errmsg) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Error executing SQL (%s): %s\n", sql, errmsg);
		sqlite3_free(errmsg);
		res = -1;
	}

	return res;
}

static int ast_db_begin_transaction(void)
{
	return db_execute_sql("BEGIN TRANSACTION", NULL, NULL);
}

static int ast_db_commit_transaction(void)
{
	return db_execute_sql("COMMIT", NULL, NULL);
}

static int ast_db_rollback_transaction(void)
{
	return db_execute_sql("ROLLBACK", NULL, NULL);
}

int ast_db_put(const char *family, const char *key, const char *value)
{
	char fullkey[MAX_DB_FIELD];
	size_t fullkey_len;
	int res = 0;

	if (strlen(family) + strlen(key) + 2 > sizeof(fullkey) - 1) {
		ast_log(LOG_WARNING, "Family and key length must be less than %zu bytes\n", sizeof(fullkey) - 3);
		return -1;
	}

	fullkey_len = snprintf(fullkey, sizeof(fullkey), "/%s/%s", family, key);

	ast_mutex_lock(&dblock);
	if (sqlite3_bind_text(put_stmt, 1, fullkey, fullkey_len, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	} else if (sqlite3_bind_text(put_stmt, 2, value, -1, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind value to stmt: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	} else if (sqlite3_step(put_stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't execute statement: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	}

	sqlite3_reset(put_stmt);
	db_sync();
	ast_mutex_unlock(&dblock);
#ifdef GRANDSTREAM_NETWORKS
	if (0 == res)
	{
		ast_ha_send_astdb(AST_DB_HA_PUT, family, key, value);
	}
#endif
	return res;
}

/*!
 * \internal
 * \brief Get key value specified by family/key.
 *
 * Gets the value associated with the specified \a family and \a key, and
 * stores it, either into the fixed sized buffer specified by \a buffer
 * and \a bufferlen, or as a heap allocated string if \a bufferlen is -1.
 *
 * \note If \a bufferlen is -1, \a buffer points to heap allocated memory
 *       and must be freed by calling ast_free().
 *
 * \retval -1 An error occurred
 * \retval 0 Success
 */
static int db_get_common(const char *family, const char *key, char **buffer, int bufferlen)
{
	const unsigned char *result;
	char fullkey[MAX_DB_FIELD];
	size_t fullkey_len;
	int res = 0;

	if (strlen(family) + strlen(key) + 2 > sizeof(fullkey) - 1) {
		ast_log(LOG_WARNING, "Family and key length must be less than %zu bytes\n", sizeof(fullkey) - 3);
		return -1;
	}

	fullkey_len = snprintf(fullkey, sizeof(fullkey), "/%s/%s", family, key);

	ast_mutex_lock(&dblock);
	if (sqlite3_bind_text(get_stmt, 1, fullkey, fullkey_len, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	} else if (sqlite3_step(get_stmt) != SQLITE_ROW) {
		ast_debug(1, "Unable to find key '%s' in family '%s'\n", key, family);
		res = -1;
	} else if (!(result = sqlite3_column_text(get_stmt, 0))) {
		ast_log(LOG_WARNING, "Couldn't get value\n");
		res = -1;
	} else {
		const char *value = (const char *) result;

		if (bufferlen == -1) {
			*buffer = ast_strdup(value);
		} else {
			ast_copy_string(*buffer, value, bufferlen);
		}
	}
	sqlite3_reset(get_stmt);
	ast_mutex_unlock(&dblock);

	return res;
}

int ast_db_get(const char *family, const char *key, char *value, int valuelen)
{
	ast_assert(value != NULL);

	/* Make sure we initialize */
	value[0] = 0;

	return db_get_common(family, key, &value, valuelen);
}

int ast_db_get_allocated(const char *family, const char *key, char **out)
{
	*out = NULL;

	return db_get_common(family, key, out, -1);
}

int ast_db_del(const char *family, const char *key)
{
	char fullkey[MAX_DB_FIELD];
	size_t fullkey_len;
	int res = 0;

	if (strlen(family) + strlen(key) + 2 > sizeof(fullkey) - 1) {
		ast_log(LOG_WARNING, "Family and key length must be less than %zu bytes\n", sizeof(fullkey) - 3);
		return -1;
	}

	fullkey_len = snprintf(fullkey, sizeof(fullkey), "/%s/%s", family, key);

	ast_mutex_lock(&dblock);
	if (sqlite3_bind_text(del_stmt, 1, fullkey, fullkey_len, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Couldn't bind key to stmt: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	} else if (sqlite3_step(del_stmt) != SQLITE_DONE) {
		ast_debug(1, "Unable to find key '%s' in family '%s'\n", key, family);
		res = -1;
	}
	sqlite3_reset(del_stmt);
	db_sync();
	ast_mutex_unlock(&dblock);
#ifdef GRANDSTREAM_NETWORKS
	if (0 == res)
	{
		ast_ha_send_astdb(AST_DB_HA_DEL, family, key, NULL);
	}
#endif
	return res;
}

int ast_db_deltree(const char *family, const char *keytree)
{
	sqlite3_stmt *stmt = deltree_stmt;
	char prefix[MAX_DB_FIELD];
	int res = 0;

	if (!ast_strlen_zero(family)) {
		if (!ast_strlen_zero(keytree)) {
			/* Family and key tree */
			snprintf(prefix, sizeof(prefix), "/%s/%s", family, keytree);
		} else {
			/* Family only */
			snprintf(prefix, sizeof(prefix), "/%s", family);
		}
	} else {
		prefix[0] = '\0';
		stmt = deltree_all_stmt;
	}

	ast_mutex_lock(&dblock);
	if (!ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", prefix, sqlite3_errmsg(astdb));
		res = -1;
	} else if (sqlite3_step(stmt) != SQLITE_DONE) {
		ast_log(LOG_WARNING, "Couldn't execute stmt: %s\n", sqlite3_errmsg(astdb));
		res = -1;
	}
	//res = sqlite3_changes(astdb);
	sqlite3_reset(stmt);
	db_sync();
	ast_mutex_unlock(&dblock);
#ifdef GRANDSTREAM_NETWORKS
	if (0 == res)
	{
		ast_ha_send_astdb(AST_DB_HA_DELTREE, family, keytree, NULL);
	}
#endif
	return res;
}

static struct ast_db_entry *db_gettree_common(sqlite3_stmt *stmt)
{
	struct ast_db_entry *head = NULL, *prev = NULL, *cur;

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		const char *key, *value;
		size_t key_len, value_len;

		key   = (const char *) sqlite3_column_text(stmt, 0);
		value = (const char *) sqlite3_column_text(stmt, 1);

		if (!key || !value) {
			break;
		}

		key_len = strlen(key);
		value_len = strlen(value);

		cur = ast_malloc(sizeof(*cur) + key_len + value_len + 2);
		if (!cur) {
			break;
		}

		cur->next = NULL;
		cur->key = cur->data + value_len + 1;
		memcpy(cur->data, value, value_len + 1);
		memcpy(cur->key, key, key_len + 1);

		if (prev) {
			prev->next = cur;
		} else {
			head = cur;
		}
		prev = cur;
	}

	return head;
}

struct ast_db_entry *ast_db_gettree(const char *family, const char *keytree)
{
	char prefix[MAX_DB_FIELD];
	sqlite3_stmt *stmt = gettree_stmt;
	size_t res = 0;
	struct ast_db_entry *ret;

	if (!ast_strlen_zero(family)) {
		if (!ast_strlen_zero(keytree)) {
			/* Family and key tree */
			res = snprintf(prefix, sizeof(prefix), "/%s/%s", family, keytree);
		} else {
			/* Family only */
			res = snprintf(prefix, sizeof(prefix), "/%s", family);
		}

		if (res >= sizeof(prefix)) {
			ast_log(LOG_WARNING, "Requested prefix is too long: %s\n", keytree);
			return NULL;
		}
	} else {
		prefix[0] = '\0';
		stmt = gettree_all_stmt;
	}

	ast_mutex_lock(&dblock);
#ifdef GRANDSTREAM_NETWORKS
		int entry_found = 0;
		if (g_ast_db_ha_enable && db_entry_for_pjsip_load) {	
			struct ast_db_entry *cur = db_entry_for_pjsip_load;
			struct ast_db_entry *ret_entry = NULL;
			struct ast_db_entry *insert_entry = NULL;
			struct ast_db_entry *pre = NULL;
			struct ast_db_entry *cur_del = NULL;
			while (cur) {
				const char *value_s = (const char *)(cur->data);
				int len = strlen(value_s);
				if (!strncmp(cur->key, prefix, strlen(prefix) - 2)) {
					if (!(insert_entry = ast_malloc(sizeof(*cur) + strlen(cur->key) + len + 2))) {
						break;
					}
					insert_entry->next = NULL;
					insert_entry->key = insert_entry->data + len + 1;
					strcpy(insert_entry->data, value_s);
					strcpy(insert_entry->key, cur->key);
					if (ret_entry) {
						while (ret_entry->next) {
							ret_entry = ret_entry->next;
						}
						ret_entry->next = insert_entry;
					} else {
						ret_entry = insert_entry;
					}
					entry_found = 1;
	
					/* delete the matched entery to reduce the list count, so reduce the time to match for next query */
					cur_del = cur;
					if (cur_del == db_entry_for_pjsip_load) {
						db_entry_for_pjsip_load = cur->next;
					} else {
						if (pre) {
							pre->next = cur->next;
						}
					}
					cur = cur->next;
					ast_free(cur_del);
					cur_del = NULL;
				} else {
					pre = cur;
					cur = cur->next;
				}
			}
	
			ret = ret_entry;
		}
	
		if (!entry_found) {
	if (res && !ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, res, SQLITE_STATIC) != SQLITE_OK)) {
#else
	if (res && (sqlite3_bind_text(stmt, 1, prefix, res, SQLITE_STATIC) != SQLITE_OK)) {
#endif
		ast_log(LOG_WARNING, "Could not bind %s to stmt: %s\n", prefix, sqlite3_errmsg(astdb));
		sqlite3_reset(stmt);
		ast_mutex_unlock(&dblock);
		return NULL;
	}

	ret = db_gettree_common(stmt);
	sqlite3_reset(stmt);
#ifdef GRANDSTREAM_NETWORKS
	}
#endif
	ast_mutex_unlock(&dblock);
	return ret;
}

struct ast_db_entry *ast_db_gettree_by_prefix(const char *family, const char *key_prefix)
{
	char prefix[MAX_DB_FIELD];
	size_t res;
	struct ast_db_entry *ret;

	res = snprintf(prefix, sizeof(prefix), "/%s/%s", family, key_prefix);
	if (res >= sizeof(prefix)) {
		ast_log(LOG_WARNING, "Requested key prefix is too long: %s\n", key_prefix);
		return NULL;
	}

	ast_mutex_lock(&dblock);
	if (sqlite3_bind_text(gettree_prefix_stmt, 1, prefix, res, SQLITE_STATIC) != SQLITE_OK) {
		ast_log(LOG_WARNING, "Could not bind %s to stmt: %s\n", prefix, sqlite3_errmsg(astdb));
		sqlite3_reset(gettree_prefix_stmt);
		ast_mutex_unlock(&dblock);
		return NULL;
	}

	ret = db_gettree_common(gettree_prefix_stmt);
	sqlite3_reset(gettree_prefix_stmt);
	ast_mutex_unlock(&dblock);

	return ret;
}

void ast_db_freetree(struct ast_db_entry *dbe)
{
	struct ast_db_entry *last;
	while (dbe) {
		last = dbe;
		dbe = dbe->next;
		ast_free(last);
	}
}

static char *handle_cli_database_put(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;

	switch (cmd) {
	case CLI_INIT:
		e->command = "database put";
		e->usage =
			"Usage: database put <family> <key> <value>\n"
			"       Adds or updates an entry in the Asterisk database for\n"
			"       a given family, key, and value.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5)
		return CLI_SHOWUSAGE;
	res = ast_db_put(a->argv[2], a->argv[3], a->argv[4]);
	if (res)  {
		ast_cli(a->fd, "Failed to update entry\n");
	} else {
		ast_cli(a->fd, "Updated database successfully\n");
	}
	return CLI_SUCCESS;
}

static char *handle_cli_database_get(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;
	char tmp[MAX_DB_FIELD];

	switch (cmd) {
	case CLI_INIT:
		e->command = "database get";
		e->usage =
			"Usage: database get <family> <key>\n"
			"       Retrieves an entry in the Asterisk database for a given\n"
			"       family and key.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4)
		return CLI_SHOWUSAGE;
	res = ast_db_get(a->argv[2], a->argv[3], tmp, sizeof(tmp));
	if (res) {
		ast_cli(a->fd, "Database entry not found.\n");
	} else {
		ast_cli(a->fd, "Value: %s\n", tmp);
	}
	return CLI_SUCCESS;
}

static char *handle_cli_database_del(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;

	switch (cmd) {
	case CLI_INIT:
		e->command = "database del";
		e->usage =
			"Usage: database del <family> <key>\n"
			"       Deletes an entry in the Asterisk database for a given\n"
			"       family and key.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4)
		return CLI_SHOWUSAGE;
	res = ast_db_del(a->argv[2], a->argv[3]);
	if (res) {
		ast_cli(a->fd, "Database entry does not exist.\n");
	} else {
		ast_cli(a->fd, "Database entry removed.\n");
	}
	return CLI_SUCCESS;
}

static char *handle_cli_database_deltree(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int num_deleted;

	switch (cmd) {
	case CLI_INIT:
		e->command = "database deltree";
		e->usage =
			"Usage: database deltree <family> [keytree]\n"
			"   OR: database deltree <family>[/keytree]\n"
			"       Deletes a family or specific keytree within a family\n"
			"       in the Asterisk database.  The two arguments may be\n"
			"       separated by either a space or a slash.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if ((a->argc < 3) || (a->argc > 4))
		return CLI_SHOWUSAGE;
	if (a->argc == 4) {
		num_deleted = ast_db_deltree(a->argv[2], a->argv[3]);
	} else {
		num_deleted = ast_db_deltree(a->argv[2], NULL);
	}
	if (num_deleted < 0) {
		ast_cli(a->fd, "Database unavailable.\n");
	} else if (num_deleted == 0) {
		ast_cli(a->fd, "Database entries do not exist.\n");
	} else {
		ast_cli(a->fd, "%d database entries removed.\n",num_deleted);
	}
	return CLI_SUCCESS;
}

static char *handle_cli_database_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char prefix[MAX_DB_FIELD];
	int counter = 0;
	sqlite3_stmt *stmt = gettree_stmt;

	switch (cmd) {
	case CLI_INIT:
		e->command = "database show";
		e->usage =
			"Usage: database show [family [keytree]]\n"
			"   OR: database show [family[/keytree]]\n"
			"       Shows Asterisk database contents, optionally restricted\n"
			"       to a given family, or family and keytree. The two arguments\n"
			"       may be separated either by a space or by a slash.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 4) {
		/* Family and key tree */
		snprintf(prefix, sizeof(prefix), "/%s/%s", a->argv[2], a->argv[3]);
	} else if (a->argc == 3) {
		/* Family only */
		snprintf(prefix, sizeof(prefix), "/%s", a->argv[2]);
	} else if (a->argc == 2) {
		/* Neither */
		prefix[0] = '\0';
		stmt = gettree_all_stmt;

	} else {
		return CLI_SHOWUSAGE;
	}

	ast_mutex_lock(&dblock);
	if (!ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", prefix, sqlite3_errmsg(astdb));
		sqlite3_reset(stmt);
		ast_mutex_unlock(&dblock);
		return NULL;
	}

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		const char *key_s, *value_s;
		if (!(key_s = (const char *) sqlite3_column_text(stmt, 0))) {
			ast_log(LOG_WARNING, "Skipping invalid key!\n");
			continue;
		}
		if (!(value_s = (const char *) sqlite3_column_text(stmt, 1))) {
			ast_log(LOG_WARNING, "Skipping invalid value!\n");
			continue;
		}
		++counter;
		ast_cli(a->fd, "%-50s: %-25s\n", key_s, value_s);
	}

	sqlite3_reset(stmt);
	ast_mutex_unlock(&dblock);

	ast_cli(a->fd, "%d results found.\n", counter);
	return CLI_SUCCESS;
}

static char *handle_cli_database_showkey(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int counter = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "database showkey";
		e->usage =
			"Usage: database showkey <keytree>\n"
			"       Shows Asterisk database contents, restricted to a given key.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_mutex_lock(&dblock);
	if (!ast_strlen_zero(a->argv[2]) && (sqlite3_bind_text(showkey_stmt, 1, a->argv[2], -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", a->argv[2], sqlite3_errmsg(astdb));
		sqlite3_reset(showkey_stmt);
		ast_mutex_unlock(&dblock);
		return NULL;
	}

	while (sqlite3_step(showkey_stmt) == SQLITE_ROW) {
		const char *key_s, *value_s;
		if (!(key_s = (const char *) sqlite3_column_text(showkey_stmt, 0))) {
			break;
		}
		if (!(value_s = (const char *) sqlite3_column_text(showkey_stmt, 1))) {
			break;
		}
		++counter;
		ast_cli(a->fd, "%-50s: %-25s\n", key_s, value_s);
	}
	sqlite3_reset(showkey_stmt);
	ast_mutex_unlock(&dblock);

	ast_cli(a->fd, "%d results found.\n", counter);
	return CLI_SUCCESS;
}

static int display_results(void *arg, int columns, char **values, char **colnames)
{
	struct ast_cli_args *a = arg;
	size_t x;

	for (x = 0; x < columns; x++) {
		ast_cli(a->fd, "%-5s: %-50s\n", colnames[x], values[x]);
	}
	ast_cli(a->fd, "\n");

	return 0;
}

static char *handle_cli_database_query(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{

	switch (cmd) {
	case CLI_INIT:
		e->command = "database query";
		e->usage =
			"Usage: database query \"<SQL Statement>\"\n"
			"       Run a user-specified SQL query on the database. Be careful.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_mutex_lock(&dblock);
	db_execute_sql(a->argv[2], display_results, a);
	db_sync(); /* Go ahead and sync the db in case they write */
	ast_mutex_unlock(&dblock);

	return CLI_SUCCESS;
}

#ifdef GRANDSTREAM_NETWORKS
static char *handle_cli_database_reconnect(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char *dbname;
	int ret;

	switch (cmd) {
	case CLI_INIT:
		e->command = "database reconnect";
		e->usage =
			"Usage: database reconnect\n"
			"       Reconnect database\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 2)
		return CLI_SHOWUSAGE;

	ast_mutex_lock(&dblock);
	if ((ret = sqlite3_close(astdb)) == SQLITE_OK) {
		astdb = NULL;
	} else {
		if (SQLITE_BUSY == ret) {
			clean_statements();
			if ((ret = sqlite3_close(astdb)) == SQLITE_OK) {
				astdb = NULL;
			} else {
				ast_log(LOG_WARNING, "astdb closed fail [%d]\n", ret);
				ast_cli(a->fd, "astdb closed fail\n");
				return CLI_FAILURE;
			}
		} else {
			ast_log(LOG_WARNING, "astdb closed fail [%d]\n", ret);
			ast_cli(a->fd, "astdb closed fail\n");
			return CLI_FAILURE;
		}
	}

	if (!(dbname = ast_alloca(strlen(ast_config_AST_DB) + sizeof(".sqlite3")))) {
		ast_log(LOG_WARNING, "astdb name alloc fail\n");
		ast_cli(a->fd, "astdb name alloc fail\n");
		return CLI_FAILURE;
	}
	strcpy(dbname, ast_config_AST_DB);
	strcat(dbname, ".sqlite3");

	if (sqlite3_open(dbname, &astdb) != SQLITE_OK || init_statements()) {
		ast_log(LOG_WARNING, "Unable to open Asterisk database '%s': %s\n", dbname, sqlite3_errmsg(astdb));
		ast_cli(a->fd, "Unable to open Asterisk database '%s': %s\n", dbname, sqlite3_errmsg(astdb));
		sqlite3_close(astdb);
		ast_mutex_unlock(&dblock);
		return CLI_FAILURE;
	}
	ast_mutex_unlock(&dblock);

	ast_cli(a->fd, "database reconnect successfully\n");

	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_put(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database put";
		e->usage =
			"Usage: ucm_database put <family> <key> <value>\n"
			"       Adds or updates an entry in the Asterisk ucm_database for\n"
			"       a given family, key, and value.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5)
		return CLI_SHOWUSAGE;
	res = ast_ucm_db_put(a->argv[2], a->argv[3], a->argv[4], 1);
	if (res)  {
		ast_cli(a->fd, "Failed to update entry\n");
	} else {
		ast_cli(a->fd, "Updated database successfully\n");
	}
	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_get(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;
	char tmp[MAX_DB_FIELD];

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database get";
		e->usage =
			"Usage: ucm_database get <family> <key>\n"
			"       Retrieves an entry in the Asterisk ucm_database for a given\n"
			"       family and key.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4)
		return CLI_SHOWUSAGE;
	res = ast_ucm_db_get(a->argv[2], a->argv[3], tmp, sizeof(tmp));
	if (res) {
		ast_cli(a->fd, "Database entry not found.\n");
	} else {
		ast_cli(a->fd, "Value: %s\n", tmp);
	}
	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_del(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int res;

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database del";
		e->usage =
			"Usage: ucm_database del <family> <key>\n"
			"       Deletes an entry in the Asterisk ucm_database for a given\n"
			"       family and key.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4)
		return CLI_SHOWUSAGE;
	res = ast_ucm_db_del(a->argv[2], a->argv[3]);
	if (res) {
		ast_cli(a->fd, "Database entry does not exist.\n");
	} else {
		ast_cli(a->fd, "Database entry removed.\n");
	}
	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_deltree(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int num_deleted;

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database deltree";
		e->usage =
			"Usage: ucm_database deltree <family> [keytree]\n"
			"   OR: ucm_database deltree <family>[/keytree]\n"
			"       Deletes a family or specific keytree within a family\n"
			"       in the Asterisk ucm_database.  The two arguments may be\n"
			"       separated by either a space or a slash.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if ((a->argc < 3) || (a->argc > 4))
		return CLI_SHOWUSAGE;
	if (a->argc == 4) {
		num_deleted = ast_ucm_db_deltree(a->argv[2], a->argv[3]);
	} else {
		num_deleted = ast_ucm_db_deltree(a->argv[2], NULL);
	}
	if (num_deleted < 0) {
		ast_cli(a->fd, "Database unavailable.\n");
	} else if (num_deleted == 0) {
		ast_cli(a->fd, "Database entries do not exist.\n");
	} else {
		ast_cli(a->fd, "%d ucm_database entries removed.\n",num_deleted);
	}
	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	char prefix[MAX_DB_FIELD];
	int counter = 0;
	sqlite3_stmt *stmt = ucm_gettree_stmt;

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database show";
		e->usage =
			"Usage: ucm_database show [family [keytree]]\n"
			"   OR: ucm_database show [family[/keytree]]\n"
			"       Shows Asterisk ucm_database contents, optionally restricted\n"
			"       to a given family, or family and keytree. The two arguments\n"
			"       may be separated either by a space or by a slash.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 4) {
		/* Family and key tree */
		snprintf(prefix, sizeof(prefix), "/%s/%s", a->argv[2], a->argv[3]);
	} else if (a->argc == 3) {
		/* Family only */
		snprintf(prefix, sizeof(prefix), "/%s", a->argv[2]);
	} else if (a->argc == 2) {
		/* Neither */
		prefix[0] = '\0';
		stmt = ucm_gettree_all_stmt;

	} else {
		return CLI_SHOWUSAGE;
	}

	ast_mutex_lock(&ucm_dblock);
	if (!ast_strlen_zero(prefix) && (sqlite3_bind_text(stmt, 1, prefix, -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", prefix, sqlite3_errmsg(ucm_astdb));
		sqlite3_reset(stmt);
		ast_mutex_unlock(&ucm_dblock);
		return NULL;
	}

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		const char *key_s, *value_s;
		if (!(key_s = (const char *) sqlite3_column_text(stmt, 0))) {
			ast_log(LOG_WARNING, "Skipping invalid key!\n");
			continue;
		}
		if (!(value_s = (const char *) sqlite3_column_text(stmt, 1))) {
			ast_log(LOG_WARNING, "Skipping invalid value!\n");
			continue;
		}
		++counter;
		ast_cli(a->fd, "%-50s: %-25s\n", key_s, value_s);
	}

	sqlite3_reset(stmt);
	ast_mutex_unlock(&ucm_dblock);

	ast_cli(a->fd, "%d results found.\n", counter);
	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_showkey(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int counter = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database showkey";
		e->usage =
			"Usage: ucm_database showkey <keytree>\n"
			"       Shows Asterisk ucm_database contents, restricted to a given key.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_mutex_lock(&ucm_dblock);
	if (!ast_strlen_zero(a->argv[2]) && (sqlite3_bind_text(ucm_showkey_stmt, 1, a->argv[2], -1, SQLITE_STATIC) != SQLITE_OK)) {
		ast_log(LOG_WARNING, "Could bind %s to stmt: %s\n", a->argv[2], sqlite3_errmsg(ucm_astdb));
		sqlite3_reset(ucm_showkey_stmt);
		ast_mutex_unlock(&ucm_dblock);
		return NULL;
	}

	while (sqlite3_step(ucm_showkey_stmt) == SQLITE_ROW) {
		const char *key_s, *value_s;
		if (!(key_s = (const char *) sqlite3_column_text(ucm_showkey_stmt, 0))) {
			break;
		}
		if (!(value_s = (const char *) sqlite3_column_text(ucm_showkey_stmt, 1))) {
			break;
		}
		++counter;
		ast_cli(a->fd, "%-50s: %-25s\n", key_s, value_s);
	}
	sqlite3_reset(ucm_showkey_stmt);
	ast_mutex_unlock(&ucm_dblock);

	ast_cli(a->fd, "%d results found.\n", counter);
	return CLI_SUCCESS;
}

static char *handle_cli_ucm_database_query(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{

	switch (cmd) {
	case CLI_INIT:
		e->command = "ucm_database query";
		e->usage =
			"Usage: ucm_database query \"<SQL Statement>\"\n"
			"       Run a user-specified SQL query on the ucm_database. Be careful.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}

	ast_mutex_lock(&ucm_dblock);
	ucm_db_execute_sql(a->argv[2], display_results, a);
	ucm_db_sync(); /* Go ahead and sync the db in case they write */
	ast_mutex_unlock(&ucm_dblock);

	return CLI_SUCCESS;
}

#endif


static struct ast_cli_entry cli_database[] = {
	AST_CLI_DEFINE(handle_cli_database_show,    "Shows database contents"),
	AST_CLI_DEFINE(handle_cli_database_showkey, "Shows database contents"),
	AST_CLI_DEFINE(handle_cli_database_get,     "Gets database value"),
	AST_CLI_DEFINE(handle_cli_database_put,     "Adds/updates database value"),
	AST_CLI_DEFINE(handle_cli_database_del,     "Removes database key/value"),
	AST_CLI_DEFINE(handle_cli_database_deltree, "Removes database keytree/values"),
	AST_CLI_DEFINE(handle_cli_database_query,   "Run a user-specified query on the astdb"),
#ifdef GRANDSTREAM_NETWORKS
	AST_CLI_DEFINE(handle_cli_database_reconnect,   "Reconnect to the astdb"),
	AST_CLI_DEFINE(handle_cli_ucm_database_show,	"Shows ucm_database contents"),
	AST_CLI_DEFINE(handle_cli_ucm_database_showkey, "Shows ucm_database contents"),
	AST_CLI_DEFINE(handle_cli_ucm_database_get, 	"Gets ucm_database value"),
	AST_CLI_DEFINE(handle_cli_ucm_database_put, 	"Adds/updates ucm_database value"),
	AST_CLI_DEFINE(handle_cli_ucm_database_del, 	"Removes ucm_database key/value"),
	AST_CLI_DEFINE(handle_cli_ucm_database_deltree, "Removes ucm_database keytree/values"),
	AST_CLI_DEFINE(handle_cli_ucm_database_query,	"Run a user-specified query on the ucm_astdb"),
#endif
};

static int manager_dbput(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	const char *val = astman_get_header(m, "Val");
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified");
		return 0;
	}
	if (ast_strlen_zero(key)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}

	res = ast_db_put(family, key, S_OR(val, ""));
	if (res) {
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int manager_dbputtree(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *keys = astman_get_header(m, "Keys");
	const char *vals = astman_get_header(m, "Vals");
	const char *key = NULL;
	const char *val = NULL;
	char *key_saved = NULL;
	char *val_saved = NULL;
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified");
		return 0;
	}
	if (ast_strlen_zero(keys)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}

        key = (const char *)strtok_r((char *)keys, ";,;", &key_saved);
        val = (const char *)strtok_r((char *)vals, ";,;", &val_saved);
        while ( key != NULL )
        {
            if (val == NULL)
            {
                val = "";
            }
            else if (strcasecmp(val, "~none~") == 0)
            {
                /*the val is maked as empty*/
                val = "";
            }

            ast_log(LOG_DEBUG, "manager_dbputtree while key is [%s] val is [%s]\n", key, val);

            if (!ast_strlen_zero(key))
            {
                 res = ast_db_put(family, key, S_OR(val, ""));
                 if (res)
                 {
                     /*TODO: Need a rollback !*/
                     break;
                 }
            }
            key = strtok_r(NULL, ";,;", &key_saved);
            val = strtok_r(NULL, ";,;", &val_saved);
        }

	if (res) {
                /*TODO: Need a rollback !*/
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	return 0;
}
#endif

#ifdef GRANDSTREAM_NETWORKS  /*Add by weidu ,for ha astdb*/
static int manager_dbhaput(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	const char *val = astman_get_header(m, "Val");
	int res;

	ast_log(LOG_DEBUG, "manager_dbhaput entering");
	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified");
		return 0;
	}
	if (ast_strlen_zero(key)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}
	ast_log(LOG_DEBUG, "manager_dbhaput family %s, key %s \n", family, key);
	
	if (!strncmp(family, "registrar", 9) && !strncmp(key, "contact", 7)) {
		ast_log(LOG_DEBUG, "key+8 = %s \n", key+8);
		ast_ha_publish_astdb_reg_data(1, (char *)(key+8),(char *) val);
	    astman_send_ack(s, m, "Updated database successfully");
		return 0;
	}

	if (!strncmp(family, "Identify", 8) && !strncmp(family, "matches", 7)) {
		ast_ha_publish_astdb_ip_data(1, (char *)(key+8), (char *)val);
		astman_send_ack(s, m, "Updated database successfully");
		return 0;
	}
	res = ast_db_put(family, key, S_OR(val, ""));
	if (res) {
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	return 0;
}

static int manager_dbhadel(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified.");
		return 0;
	}

	if (ast_strlen_zero(key)) {
		astman_send_error(s, m, "No key specified.");
		return 0;
	}

	ast_log(LOG_DEBUG, "family %s, key %s \n",family, key);
	if (!strncmp(family, "registrar", 9) && !strncmp(key, "contact", 7)) {
		ast_ha_publish_astdb_reg_data(2, (char *)(key+8), NULL);
		astman_send_ack(s, m, "Updated database successfully");
		return 0;
	}
	res = ast_db_del(family, key);
	if (res)
		astman_send_error(s, m, "Database entry not found");
	else
		astman_send_ack(s, m, "Key deleted successfully");

	return 0;
}
/*END: Add by weidu*/
#endif

static int manager_dbget(struct mansession *s, const struct message *m)
{
	const char *id = astman_get_header(m,"ActionID");
	char idText[256];
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	char tmp[MAX_DB_FIELD];
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified.");
		return 0;
	}
	if (ast_strlen_zero(key)) {
		astman_send_error(s, m, "No key specified.");
		return 0;
	}

	idText[0] = '\0';
	if (!ast_strlen_zero(id))
		snprintf(idText, sizeof(idText) ,"ActionID: %s\r\n", id);

	res = ast_db_get(family, key, tmp, sizeof(tmp));
	if (res) {
		astman_send_error(s, m, "Database entry not found");
	} else {
		astman_send_listack(s, m, "Result will follow", "start");

		astman_append(s, "Event: DBGetResponse\r\n"
				"Family: %s\r\n"
				"Key: %s\r\n"
				"Val: %s\r\n"
				"%s"
				"\r\n",
				family, key, tmp, idText);

		astman_send_list_complete_start(s, m, "DBGetComplete", 1);
		astman_send_list_complete_end(s);
	}
	return 0;
}

static int manager_dbdel(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified.");
		return 0;
	}

	if (ast_strlen_zero(key)) {
		astman_send_error(s, m, "No key specified.");
		return 0;
	}

	res = ast_db_del(family, key);
	if (res)
		astman_send_error(s, m, "Database entry not found");
	else
		astman_send_ack(s, m, "Key deleted successfully");

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static void ast_db_publish_reg_contact_del(char *data)
{
	struct stasis_topic *ha_db_data_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);
	char *end_flag = NULL;

	if (!data) {
		return;
	}

	end_flag = strchr(data, ';');
	if (end_flag) {
		*end_flag = '\0';
	}

	ast_log(LOG_NOTICE, "ast_db_publish_reg_contact_del, data[%s]", data);

	ha_db_data_topic = ast_ha_sync_data_topic_all();
	if (ha_db_data_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_ha_sync_data_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s}", "aor", data);
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return;
	}

	payload = ast_json_payload_create(json_object);
	if (!payload) {
		ast_log(LOG_WARNING, "Failed to create json payload");
		return;
	}

	msg = stasis_message_create(ast_ha_registrar_contact_del_type(), payload);
	if (!msg) {
		ast_log(LOG_WARNING, "Failed to create stasis message");
		return;
	}

	stasis_publish(ha_db_data_topic, msg);
}
#endif

static int manager_dbdeltree(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	int num_deleted;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified.");
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for 146256 */
	if (!strncmp(family, "registrar/contact/", 18)) {
		char *data = (char *)family + strlen("registrar/contact/");

		ast_db_publish_reg_contact_del(data);
		astman_send_ack(s, m, "Key tree deleted successfully");
		return 0;
	}
#endif

	if (!ast_strlen_zero(key)) {
		num_deleted = ast_db_deltree(family, key);
	} else {
		num_deleted = ast_db_deltree(family, NULL);
	}

	if (num_deleted < 0) {
		astman_send_error(s, m, "Database unavailable");
	} else if (num_deleted == 0) {
		astman_send_error(s, m, "Database entry not found");
	} else {
		astman_send_ack(s, m, "Key tree deleted successfully");
	}

	return 0;
}

/*!
 * \internal
 * \brief Signal the astdb sync thread to do its thing.
 *
 * \note dblock is assumed to be held when calling this function.
 */
static void db_sync(void)
{
	dosync = 1;
	ast_cond_signal(&dbcond);
}

/*!
 * \internal
 * \brief astdb sync thread
 *
 * This thread is in charge of syncing astdb to disk after a change.
 * By pushing it off to this thread to take care of, this I/O bound operation
 * will not block other threads from performing other critical processing.
 * If changes happen rapidly, this thread will also ensure that the sync
 * operations are rate limited.
 */
static void *db_sync_thread(void *data)
{
	ast_mutex_lock(&dblock);
	ast_db_begin_transaction();
	for (;;) {
		/* If dosync is set, db_sync() was called during sleep(1),
		 * and the pending transaction should be committed.
		 * Otherwise, block until db_sync() is called.
		 */
		while (!dosync) {
			ast_cond_wait(&dbcond, &dblock);
		}
		dosync = 0;
		if (ast_db_commit_transaction()) {
			ast_db_rollback_transaction();
		}
		if (doexit) {
			ast_mutex_unlock(&dblock);
			break;
		}
		ast_db_begin_transaction();
		ast_mutex_unlock(&dblock);
		sleep(1);
		ast_mutex_lock(&dblock);
	}

	return NULL;
}

#ifdef GRANDSTREAM_NETWORKS
static int manager_ucm_dbput(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	const char *val = astman_get_header(m, "Val");
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified");
		return 0;
	}
	if (ast_strlen_zero(key)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}

	res = ast_ucm_db_put(family, key, S_OR(val, ""),1);
	if (res) {
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	return 0;
}

static int manager_ucm_dbputtree(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *keys = astman_get_header(m, "Keys");
	const char *vals = astman_get_header(m, "Vals");
	const char *key = NULL;
	const char *val = NULL;
	char *key_saved = NULL;
	char *val_saved = NULL;
	int res;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified");
		return 0;
	}
	if (ast_strlen_zero(keys)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}

        key = (const char *)strtok_r((char *)keys, ";,;", &key_saved);
        val = (const char *)strtok_r((char *)vals, ";,;", &val_saved);
        while ( key != NULL )
        {
            if (val == NULL)
            {
                val = "";
            }
            else if (strcasecmp(val, "~none~") == 0)
            {
                /*the val is maked as empty*/
                val = "";
            }

            ast_log(LOG_DEBUG, "manager_dbputtree while key is [%s] val is [%s]\n", key, val);

            if (!ast_strlen_zero(key))
            {
                 res = ast_ucm_db_put(family, key, S_OR(val, ""),0);
                 if (res)
                 {
                     /*TODO: Need a rollback !*/
                     break;
                 }
            }
            key = strtok_r(NULL, ";,;", &key_saved);
            val = strtok_r(NULL, ";,;", &val_saved);
        }
	    ast_mutex_lock(&ucm_dblock);
    	ucm_db_sync();
	    ast_mutex_unlock(&ucm_dblock);
 

	if (res) {
                /*TODO: Need a rollback !*/
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	return 0;
}


static int manager_ucm_dbdeltree(struct mansession *s, const struct message *m)
{
	const char *family = astman_get_header(m, "Family");
	const char *key = astman_get_header(m, "Key");
	int num_deleted = 1;
	const char *tmp_family = NULL;
	char *family_saved = NULL;

	if (ast_strlen_zero(family)) {
		astman_send_error(s, m, "No family specified.");
		return 0;
	}

	if (NULL != strstr(family, ";,;")){
		tmp_family = (const char *)strtok_r((char *)family, ";,;", &family_saved);
		while ( tmp_family != NULL ){
			if (!ast_strlen_zero(tmp_family)){
				num_deleted = ast_ucm_db_deltree(tmp_family, NULL);
				if (num_deleted < 0){
					break;
				}
			}
			tmp_family = strtok_r(NULL, ";,;", &family_saved);
		}
	}
	else{
		if (!ast_strlen_zero(key)) {
			num_deleted = ast_ucm_db_deltree(family, key);
		} else {
			num_deleted = ast_ucm_db_deltree(family, NULL);
		}
	}

	if (num_deleted < 0) {
		astman_send_error(s, m, "Database unavailable");
	} else if (num_deleted == 0) {
		astman_send_error(s, m, "Database entry not found");
	} else {
		astman_send_ack(s, m, "Key tree deleted successfully");
	}

	return 0;
}
static int manager_ucm_dbdelall(struct mansession *s, const struct message *m)
{

	ast_ucm_db_delall();

	astman_send_ack(s, m, "deleted successfully");

	return 0;
}

static int manager_dbpmsput(struct mansession *s, const struct message *m)
{
	const char *familys = astman_get_header(m, "Familys");
	const char *keys = astman_get_header(m, "Keys");
	const char *vals = astman_get_header(m, "Vals");
	const char *family = NULL;
	const char *key = NULL;
	const char *val = NULL;
	char *family_saved = NULL;
	char *key_saved = NULL;
	char *val_saved = NULL;
	int res=0;
	ast_log(LOG_ERROR, "familys:%s | keys:%s | vals:%s", familys?familys:"familys", keys?keys:"keys", vals?vals:"vals");

	if (ast_strlen_zero(familys)) {
		astman_send_error(s, m, "No family specified");
		return 0;
	}
	if (ast_strlen_zero(keys)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}

	family = (const char *)strtok_r((char *)familys, ";,;", &family_saved);
	key = (const char *)strtok_r((char *)keys, ";,;", &key_saved);
	val = (const char *)strtok_r((char *)vals, ";,;", &val_saved);
	while ( NULL != family && NULL != key )
	{
		if (NULL == val || 0 == strcasecmp(val, "~none~"))
		{
			val = "";
		}

		ast_log(LOG_DEBUG, "manager_dbpmsput while family is [%s] key is [%s] val is [%s]\n", family, key, val);

		if (!ast_strlen_zero(key))
		{
			 res = ast_ucm_db_put(family, key, val,0);
			 if (res)
			 {
				 break;
			 }
		}
		family = strtok_r(NULL, ";,;", &family_saved);
		key = strtok_r(NULL, ";,;", &key_saved);
		val = strtok_r(NULL, ";,;", &val_saved);
	}
	ast_mutex_lock(&ucm_dblock);
	ucm_db_sync();
	ast_mutex_unlock(&ucm_dblock);

	if (res) {
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	return 0;
}

static int manager_recovery_dbput(struct mansession *s, const struct message *m)
{
	char *sql = NULL;
	const char *runtime_uuid = astman_get_header(m, "runtime_uuid");
	const char *technology = astman_get_header(m, "technology");
	const char *profile_name = astman_get_header(m, "profile_name");
	const char *hostname = astman_get_header(m, "hostname");
	const char *uuid = astman_get_header(m, "uuid");
	const char *metadata = astman_get_header(m, "metadata");
	int res;

	if (ast_strlen_zero(runtime_uuid)) {
		astman_send_error(s, m, "No specified");
		return 0;
	}
	if (ast_strlen_zero(technology)) {
		astman_send_error(s, m, "No key specified");
		return 0;
	}
	sql = switch_xml_mprintf("insert into recovery (runtime_uuid, technology, profile_name, hostname, uuid, metadata) values ('%q','%q','%q','%q','%q','%q')",
					 			runtime_uuid, technology, profile_name, hostname, uuid, metadata);

	res = ast_recovery_db_put(sql, 1, 0);
	if (res) {
		astman_send_error(s, m, "Failed to update entry");
	} else {
		astman_send_ack(s, m, "Updated database successfully");
	}
	ast_free(sql);
	return 0;
}

static int manager_recovery_dbdelall(struct mansession *s, const struct message *m)
{
	ast_recovery_db_put("delete from recovery;", 1, 0);
	astman_send_ack(s, m, "deleted successfully");
	return 0;
}

/*!
 * \internal
 * \brief Signal the astdb sync thread to do its thing.
 *
 * \note dblock is assumed to be held when calling this function.
 */
static void ucm_db_sync(void)
{
	ucm_dosync = 1;
	ast_cond_signal(&ucm_dbcond);
}

/*!
 * \internal
 * \brief astdb sync thread
 *
 * This thread is in charge of syncing astdb to disk after a change.
 * By pushing it off to this thread to take care of, this I/O bound operation
 * will not block other threads from performing other critical processing.
 * If changes happen rapidly, this thread will also ensure that the sync
 * operations are rate limited.
 */
static void *ucm_db_sync_thread(void *data)
{
	ast_mutex_lock(&ucm_dblock);
	ast_ucm_db_begin_transaction();
	for (;;) {
		/* If dosync is set, db_sync() was called during sleep(1), 
		 * and the pending transaction should be committed. 
		 * Otherwise, block until db_sync() is called.
		 */
		while (!ucm_dosync) {
			ast_cond_wait(&ucm_dbcond, &ucm_dblock);
		}
		ucm_dosync = 0;
		if (ast_ucm_db_commit_transaction()) {
			ast_ucm_db_rollback_transaction();
		}
		if (ucm_doexit) {
			ast_mutex_unlock(&ucm_dblock);
			break;
		}
		ast_ucm_db_begin_transaction();
		ast_mutex_unlock(&ucm_dblock);
		sleep(1);
		ast_mutex_lock(&ucm_dblock);
	}

	return NULL;
}

#endif


/*!
 * \internal
 * \brief Clean up resources on Asterisk shutdown
 */
static void astdb_atexit(void)
{
	ast_cli_unregister_multiple(cli_database, ARRAY_LEN(cli_database));
	ast_manager_unregister("DBGet");
	ast_manager_unregister("DBPut");
#ifdef GRANDSTREAM_NETWORKS	
	ast_manager_unregister("DBPutTree");
#endif
	ast_manager_unregister("DBDel");
	ast_manager_unregister("DBDelTree");
#ifdef GRANDSTREAM_NETWORKS	
	ast_manager_unregister("UCMDBPut");
	ast_manager_unregister("UCMDBPutTree");
	ast_manager_unregister("UCMDBDelTree");
	ast_manager_unregister("RECOVERYDBPut");
	ast_manager_unregister("RECOVERYDBDelAll");

	ast_manager_unregister("UCMDBDelAll");
	/* Set doexit to 1 to kill thread. db_sync must be called with
	 * mutex held. */
	ast_mutex_lock(&ucm_dblock);
	ucm_doexit = 1;
	ucm_db_sync();
	ast_mutex_unlock(&ucm_dblock);

	pthread_join(ucm_syncthread, NULL);
	ast_mutex_lock(&ucm_dblock);
	ucm_clean_statements();
	if (sqlite3_close(ucm_astdb) == SQLITE_OK) {
		ucm_astdb = NULL;
	}
	ast_mutex_unlock(&ucm_dblock);

	ast_mutex_lock(&recovery_dblock);
	recovery_doexit = 1;
	recovery_db_sync();
	ast_mutex_unlock(&recovery_dblock);

	pthread_join(recovery_syncthread, NULL);
	ast_mutex_lock(&recovery_dblock);
	recovery_clean_statements();
	if (sqlite3_close(recovery_astdb) == SQLITE_OK) {
		recovery_astdb = NULL;
	}
	ast_mutex_unlock(&recovery_dblock);
#endif

	/* Set doexit to 1 to kill thread. db_sync must be called with
	 * mutex held. */
	ast_mutex_lock(&dblock);
	doexit = 1;
	db_sync();
	ast_mutex_unlock(&dblock);

	pthread_join(syncthread, NULL);
	ast_mutex_lock(&dblock);
	clean_statements();
	if (sqlite3_close(astdb) == SQLITE_OK) {
		astdb = NULL;
	}
	ast_mutex_unlock(&dblock);
}

int astdb_init(void)
{
	if (db_init()) {
		return -1;
	}
#ifdef GRANDSTREAM_NETWORKS
		if (ucm_db_init()) {
			return -1;
		}
		ast_cond_init(&ucm_dbcond, NULL);
		if (ast_pthread_create_background(&ucm_syncthread, NULL, ucm_db_sync_thread, NULL)) {
			return -1;
		}

		if (recovery_db_init()) {
			return -1;
		}
		ast_cond_init(&recovery_dbcond, NULL);
		if (ast_pthread_create_background(&recovery_syncthread, NULL, recovery_db_sync_thread, NULL)) {
			return -1;
		}
#endif

	ast_cond_init(&dbcond, NULL);
	if (ast_pthread_create_background(&syncthread, NULL, db_sync_thread, NULL)) {
		return -1;
	}

	ast_register_atexit(astdb_atexit);
	ast_cli_register_multiple(cli_database, ARRAY_LEN(cli_database));
	ast_manager_register_xml_core("DBGet", EVENT_FLAG_SYSTEM | EVENT_FLAG_REPORTING, manager_dbget);
	ast_manager_register_xml_core("DBPut", EVENT_FLAG_SYSTEM, manager_dbput);
#ifdef GRANDSTREAM_NETWORKS
	ast_manager_register_xml_core("DBPutTree", EVENT_FLAG_SYSTEM, manager_dbputtree);
#endif
	ast_manager_register_xml_core("DBDel", EVENT_FLAG_SYSTEM, manager_dbdel);
	ast_manager_register_xml_core("DBDelTree", EVENT_FLAG_SYSTEM, manager_dbdeltree);
#ifdef GRANDSTREAM_NETWORKS
	ast_manager_register_xml_core("UCMDBDelTree", EVENT_FLAG_SYSTEM, manager_ucm_dbdeltree);
	ast_manager_register_xml_core("UCMDBPut", EVENT_FLAG_SYSTEM, manager_ucm_dbput);
	ast_manager_register_xml_core("UCMDBPutTree", EVENT_FLAG_SYSTEM, manager_ucm_dbputtree);
	ast_manager_register_xml_core("UCMDBDelAll", EVENT_FLAG_SYSTEM, manager_ucm_dbdelall);
	ast_manager_register_xml_core("RECOVERYDBPut", EVENT_FLAG_SYSTEM, manager_recovery_dbput);
	ast_manager_register_xml_core("RECOVERYDBDelAll", EVENT_FLAG_SYSTEM, manager_recovery_dbdelall);
	ast_manager_register_xml_core("DBHAPut", EVENT_FLAG_SYSTEM, manager_dbhaput); /*Add by weidu for ha sync*/
	ast_manager_register_xml_core("DBHADel", EVENT_FLAG_SYSTEM, manager_dbhadel);
	ast_manager_register_xml_core("DBPMSPut", EVENT_FLAG_SYSTEM, manager_dbpmsput);
#endif

	return 0;
}
