
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2023, Grandstream, Inc.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief time check for forward timing settings
 *
 * \author <yfyang@grandstream.com>
 *
 * \ingroup functions
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"
#include <syslog.h>
#include <signal.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>

#include "asterisk/astobj2.h"
#include "asterisk/_private.h"
#include "asterisk/module.h"
#include "asterisk/config.h"
#include "asterisk/channel.h"
#include "asterisk/utils.h"

#include "asterisk/app.h"
#include "asterisk/strings.h"
#include "asterisk/linkedlists.h"
#include "asterisk/presencestate.h"
#include "asterisk/threadstorage.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/pbx.h"
//#include "asterisk/cli.h"
#include "asterisk/astdb.h"
#include "asterisk/fwd_timing.h"

AST_MODULE_LOG("core");


struct ext_fwd_setting {
	char   ext[80];
	char   state_context[128];
    enum   ast_presence_state   state;
    struct ast_fwd_setting    cfu_setting;
    struct ast_fwd_setting    cfn_setting;
    struct ast_fwd_setting    cfb_setting;
    struct ast_fwd_setting    rcfu_setting;
    struct ast_fwd_setting    rcfn_setting;
    struct ast_fwd_setting    rcfb_setting;
    int    mark:1;
};

/*! \brief The Timing object */
struct timing_obj {
	struct ast_timing timing; 
	AST_LIST_ENTRY(timing_obj) entry;
};

struct timing_specific {
	char   ext[80];
    AST_LIST_HEAD(,timing_obj) time_list; 
};


static AST_LIST_HEAD_STATIC(list_office_timings, timing_obj);
static AST_LIST_HEAD_STATIC(list_holiday_timings, timing_obj);
static AST_LIST_HEAD_STATIC(list_fwd_ext_status, ast_fwd_ext_status);

struct ao2_container *con_specific_timings = NULL;
struct ao2_container *con_extension_fwd_settings = NULL;

ast_mutex_t time_lock;

static int  reload_flag = 0;

static int  fwd_check_sched_id = -1;
static int  fwd_notify_sched_id = -1;
static struct ast_sched_context *fwd_check_sched = NULL;

#define  FWD_CHECK_INTVAL  30
#define  FWD_NOTIFY_INTVAL (FWD_CHECK_INTVAL+10)
#define  TIME_SETTINGS_CONFIG_FILE  "time_settings.conf"
#define  CALL_TRANSFER_CONFIG_FILE  "call_transfer.conf"

#ifdef LOW_MEMORY
static const int HASH_EXT_SIZE = 17;
#else
static const int HASH_EXT_SIZE = 563;
#endif

static const struct fwd_services {
    enum ast_fwd_service_type type;
    const char* text;
} fwd_services_table[] = {
	{FWD_CFU,  "forwardImmediate"},
	{FWD_CFN,  "forwardNoAns"},
	{FWD_CFB,  "forwardBusy"},
	{FWD_RCFU, "forwardImmediate"},
	{FWD_RCFN, "forwardNoAns"},
	{FWD_RCFB, "forwardBusy"},
    {FWD_BUTT, "unknown"},
};


static const struct fwd_dst_types {
	enum ast_fwd_dst_type type;
	const char* text;
} fwd_dsttype_table[] = {
	{FWD_DST_USER,    "user"},
	{FWD_DST_CUSTOM,  "custom"},
	{FWD_DST_VM,      "voicemail"},
	{FWD_DST_RINGROUP,"ringgroup"},
	{FWD_DST_QUEUE,   "queues"},
	{FWD_DST_VMG,     "vmgroups"},
	{FWD_DST_PROMPT,  "prompt"},
	{FWD_DST_BUTT,    "unknown"},
};

static struct ext_fwd_setting *fwd_ext_setting_find(char *name);

const char *ast_fwd_type_to_str(enum ast_fwd_service_type stype)
{
	int idx;

	for (idx = 0; idx < ARRAY_LEN(fwd_services_table); ++idx) {
		if (stype == fwd_services_table[idx].type) {
			return fwd_services_table[idx].text;
		}
	}

	return "unknown";
}

enum ast_fwd_service_type ast_fwd_type_get(char *text)
{
	int idx;

    if (!text) {
        return FWD_BUTT;
    }

	for (idx = 0; idx < ARRAY_LEN(fwd_services_table); ++idx) {
		if (!strcmp(text, fwd_services_table[idx].text)) {
			return fwd_services_table[idx].type;
		}
	}

    return FWD_BUTT;
}


enum ast_fwd_dst_type ast_fwd_dst_type_get(char *ext)
{
	int  idx;
	char family[128] = {0};
	char data[16] = {0};

	if (!ext) {
		return FWD_DST_CUSTOM;
	}

	snprintf(family, sizeof(family), "AAUSER/%s", ext);
	ast_ucm_db_get(family, "TYPE", data, sizeof(data));
	if (ast_strlen_zero(data)) {
		return FWD_DST_CUSTOM;
	}

	for (idx = 0; idx < ARRAY_LEN(fwd_dsttype_table); ++idx) {
		if (!strcmp(data, fwd_dsttype_table[idx].text)) {
			return fwd_dsttype_table[idx].type;
		}
	}

    return FWD_DST_CUSTOM;
}

enum ast_presence_state ast_fwd_ext_get_presence_status(char *ext)
{
	RAII_VAR(struct ext_fwd_setting *, fwd_setting, NULL, ao2_cleanup);

    if (!ext) {
        return AST_PRESENCE_AVAILABLE;
    }

    fwd_setting = fwd_ext_setting_find(ext);
    if (!fwd_setting) {
        return AST_PRESENCE_AVAILABLE;
    }

    return fwd_setting->state;
}

static struct timing_obj* fwd_timing_build(char *time_info)
{
	struct timing_obj *fwd_timing = NULL;

	/* Make a new one */
	if (!(fwd_timing = ast_calloc(1, sizeof(*fwd_timing)))) {
		ast_log(LOG_NOTICE, "fwd_timing_build, ast_calloc failed");
		return NULL;
	}

	if (!ast_build_timing(&fwd_timing->timing, time_info)) {
		ast_free(fwd_timing);
		ast_log(LOG_NOTICE, "fwd_timing_build, build_timing failed, time[%s]", time_info);
		return NULL;
	}

	return fwd_timing;

}

/*
[holiday_time]
time = 0:0-23:59,*,5,apr,*
        time = 0:0-23:59,*,1&2&3,may,*
        time = 0:0-23:59,*,1&2&3&4&5&6&7,oct,*
*/
static int fwd_timing_build_holiday(struct ast_config *cfg, const char *cat)
{
	struct ast_variable *var;
	struct timing_obj   *fwd_timing = NULL;

    AST_LIST_LOCK(&list_holiday_timings);
	while ((fwd_timing = AST_LIST_REMOVE_HEAD(&list_holiday_timings, entry))) {
		ast_free(fwd_timing);
	};

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "time")) {
			if ((fwd_timing = fwd_timing_build((char *)var->value))) {
                AST_LIST_INSERT_TAIL(&list_holiday_timings, fwd_timing, entry);
                ast_log(LOG_NOTICE, "build holiday timing, %s", var->value);
            }
		}
	}

    AST_LIST_UNLOCK(&list_holiday_timings);
	return 0;
}

/*
[office_time]
time = 8:30-18:0,mon&tue&wed&thu&fri,*,*,*
        time = 8:30-17:40,*,28,jan&jun&sep&dec,*
*/
static int fwd_timing_build_office(struct ast_config *cfg, const char *cat)
{
	struct ast_variable *var;
	struct timing_obj   *fwd_timing = NULL;

    AST_LIST_LOCK(&list_office_timings);
	while ((fwd_timing = AST_LIST_REMOVE_HEAD(&list_office_timings, entry))) {
		ast_free(fwd_timing);
	};

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "time")) {
			if ((fwd_timing = fwd_timing_build((char *)var->value))) {
                AST_LIST_INSERT_TAIL(&list_office_timings, fwd_timing, entry);
                ast_log(LOG_NOTICE, "build office timing, %s", var->value);
            }
		}
	}

    AST_LIST_UNLOCK(&list_office_timings);
	return 0;
}

static int fwd_timing_spec_hash(const void *obj, const int flags)
{
	const struct timing_specific *tmp_spec = obj;

	return ast_str_case_hash(tmp_spec->ext);
}

static int fwd_timing_spec_cmp(void *obj, void *arg, int flags)
{
	struct timing_specific *tmp_spec = obj, *tmp_spec2 = arg;

	return !strcasecmp(tmp_spec->ext, tmp_spec2->ext) ? CMP_MATCH | CMP_STOP : 0;
}

static struct timing_specific *fwd_timing_find_spec(const char *name)
{
	struct timing_specific tmp_spec;

	if (!name) {
		return NULL;
	}

    strncpy(tmp_spec.ext, name, sizeof(tmp_spec.ext));
	return ao2_find(con_specific_timings, &tmp_spec, OBJ_POINTER); 
}

static void fwd_timing_spec_destructor(void *obj)
{
	struct timing_specific *timing_spec = obj;
	struct timing_obj      *fwd_timing  = NULL;

	ast_log(LOG_NOTICE, "fwd_timing_spec_destructor for ext[%s]\n", timing_spec->ext);

    while ((fwd_timing = AST_LIST_REMOVE_HEAD(&timing_spec->time_list, entry))) {
        ast_free(fwd_timing);
    }

	AST_LIST_HEAD_DESTROY(&timing_spec->time_list);
}

/*
[specific_time_9004]
time = 8:50-18:0,mon&tue&wed&thu&fri,*,*,*

struct timing_specific {
	char   ext[80];
    AST_LIST_HEAD(,timing_obj) time_list; 
};

static struct ao2_container *con_specific_timings;
*/
static int fwd_timing_build_specific(struct ast_config *cfg, const char *cat)
{
	struct ast_variable *var;
	struct timing_obj   *fwd_timing = NULL;
	RAII_VAR(struct timing_specific *, timing_spec, NULL, ao2_cleanup);
	char  *ext = NULL;

	/* 1. get ext */
	ext = strstr(cat, "specific_time_");
	if (ext) {
		ext = ext+strlen("specific_time_");
	}

	if (!ext) {
		return -1;
	}

	/* 2. alloc the ao2 obj */
	timing_spec = ao2_alloc(sizeof(*timing_spec), fwd_timing_spec_destructor);
	if (!timing_spec) {
		return -1;
	}

	AST_LIST_HEAD_INIT(&timing_spec->time_list);
	ast_copy_string(timing_spec->ext, ext, sizeof(timing_spec->ext));

	ao2_lock(timing_spec);

	/* 3. set the time_list */
    AST_LIST_LOCK(&timing_spec->time_list);
	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "time")) {
			if ((fwd_timing = fwd_timing_build((char *)var->value))) {
                AST_LIST_INSERT_TAIL(&timing_spec->time_list, fwd_timing, entry);
                ast_log(LOG_NOTICE, "build specific timing, %s", var->value);
            }
		}
	}

    AST_LIST_UNLOCK(&timing_spec->time_list);
	ao2_link(con_specific_timings, timing_spec);

    ast_log(LOG_NOTICE, "fwd_timing_build_specific, ext[%s]", ext);
    ao2_unlock(timing_spec);
	return 0;
}

/* load time_settings.conf, call_transfer.conf */
static int fwd_timing_load_timing(int reload)
{
	struct ast_flags config_flags = {reload ? CONFIG_FLAG_FILEUNCHANGED : 0};
	struct ast_config   *cfg = NULL;
	const char          *cat = NULL;
	int res = 0;

    /* load conf file*/
	cfg = ast_config_load(TIME_SETTINGS_CONFIG_FILE, config_flags);

	if (cfg == CONFIG_STATUS_FILEUNCHANGED || NULL == cfg) {
		return 0;
	} 

	if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "time_settings.conf is invalid.\n");
		return -1;
    }

	if (!con_specific_timings) {
		con_specific_timings = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0, HASH_EXT_SIZE, fwd_timing_spec_hash, NULL, fwd_timing_spec_cmp);
    } else {
        /* release con_specific_timings obj */
        ao2_callback(con_specific_timings, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
    }

	ast_mutex_lock(&time_lock);
	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		if (!strncasecmp(cat, "holiday_time", 12)) {
			res = fwd_timing_build_holiday(cfg, cat);
		} else if (!strncasecmp(cat, "office_time", 11)) {
			res = fwd_timing_build_office(cfg, cat);
		} else {
			res = fwd_timing_build_specific(cfg, cat);
		}
	}

	ast_config_destroy(cfg);
	ast_mutex_unlock(&time_lock);

	ast_log(LOG_NOTICE, "fwd_timing_load_timing finished.\n");
	return 0;
}


static int fwd_ext_setting_hash(const void *obj, const int flags)
{
	const struct ext_fwd_setting *tmp_ext_fwd = obj;

	return ast_str_case_hash(tmp_ext_fwd->ext);
}

static int fwd_ext_setting_cmp(void *obj, void *arg, int flags)
{
	struct ext_fwd_setting *tmp_setting = obj, *tmp_setting2 = arg;

	return !strcasecmp(tmp_setting->ext, tmp_setting2->ext) ? CMP_MATCH | CMP_STOP : 0;
}

static int fwd_ext_setting_mark(void *obj, void *arg, int flags)
{
	struct ext_fwd_setting *tmp_ext_fwd = obj;

	ao2_lock(tmp_ext_fwd);
	tmp_ext_fwd->mark = 1;
	ao2_unlock(tmp_ext_fwd);

	return 0;
}

static int fwd_ext_setting_is_marked(void *obj, void *arg, int flags)
{
	struct ext_fwd_setting *tmp_ext_fwd = obj;

	return tmp_ext_fwd->mark ? CMP_MATCH:0;
}

static struct ext_fwd_setting *fwd_ext_setting_find(char *name)
{
	struct ext_fwd_setting tmp_setting;

	if (!name) {
		return NULL;
	}

    strncpy(tmp_setting.ext, name, sizeof(tmp_setting.ext));
	return ao2_find(con_extension_fwd_settings, &tmp_setting, OBJ_POINTER); 
}

static void fwd_ext_setting_destructor(void *obj)
{
	struct ext_fwd_setting *fwd_setting = obj;

	ast_log(LOG_NOTICE, "fwd_ext_setting_destructor for ext[%s]\n", fwd_setting->ext);
}

#if 0
static int fwd_setting_cmp(struct ast_fwd_setting *setting1, struct ast_fwd_setting *setting2)
{
    if (!setting1 || !setting2) {
        return -1;
    }

    if (setting1->enabled != setting2->enabled) {
        return -1;
    }

    if (setting1->time_type != setting2->time_type) {
        return -1;
    }

    if (strncmp(setting1->dst_number, setting2->dst_number, sizeof(setting1->dst_number))) {
        return -1;
    }

    return 0;
}
#endif

static int fwd_setting_copy(struct ast_fwd_setting *setting_src, struct ast_fwd_setting *setting_dst)
{
    if (!setting_src || !setting_dst) {
        return -1;
    }

    setting_dst->enabled = setting_src->enabled;
    setting_dst->time_type = setting_src->time_type;
    strncpy(setting_dst->dst_number, setting_src->dst_number, sizeof(setting_dst->dst_number));

    return 0;
}

static int fwd_ext_setting_init(struct ext_fwd_setting *fwd_setting)
{
	if (!fwd_setting) {
		return -1;
	}

	struct ast_fwd_setting  default_setting = {0, 0, {0}};

	fwd_setting_copy(&default_setting, &fwd_setting->cfu_setting);
	fwd_setting_copy(&default_setting, &fwd_setting->cfn_setting);
	fwd_setting_copy(&default_setting, &fwd_setting->cfb_setting);
	fwd_setting_copy(&default_setting, &fwd_setting->rcfu_setting);
	fwd_setting_copy(&default_setting, &fwd_setting->rcfn_setting);
	fwd_setting_copy(&default_setting, &fwd_setting->rcfb_setting);

	//ast_log(LOG_DEBUG, "fwd_ext_setting_init, [%s]", fwd_setting->ext);
	return 0;
}

static int fwd_ext_setting_build(struct ast_config *cfg, const char *cat, int reload)
{
	struct ast_variable      *var;
	int  existed = 0;
	RAII_VAR(struct ext_fwd_setting *, fwd_setting, NULL, ao2_cleanup);

	if (reload) {
		fwd_setting = fwd_ext_setting_find((char *)cat);
	}

	if (fwd_setting) {
		existed = 1;
	} else if ((fwd_setting = ao2_alloc(sizeof(*fwd_setting), fwd_ext_setting_destructor))) {
		strncpy(fwd_setting->ext, cat, sizeof(fwd_setting->ext));
	} else {
		return -1;
	}
/*
[1000]  
status = available
*/
    fwd_setting->mark = 0;
	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "status")) {
            fwd_setting->state = ast_presence_state_val(var->value);
            snprintf(fwd_setting->state_context, sizeof(fwd_setting->state_context), "%s-%s", cat, var->value);
            break;
		}
	}

	if (!existed) {
		ao2_link(con_extension_fwd_settings, fwd_setting);
		ast_log(LOG_NOTICE, "fwd_ext_setting_build, ext[%s], state_context[%s]", fwd_setting->ext, fwd_setting->state_context);
	} else {
		if (!ast_category_get(cfg, fwd_setting->state_context, NULL)) {
			fwd_ext_setting_init(fwd_setting);
		}
	}

	return 0;
}


static int fwd_ext_setting_update(struct ast_config *cfg, const char *cat)
{
	struct ast_variable   *var;
    char *ext = NULL;
    char *context = ast_strdupa(cat);
	RAII_VAR(struct ext_fwd_setting *, fwd_setting, NULL, ao2_cleanup);
    struct ast_fwd_setting    cfu_setting  = {0, 0, {0}};
    struct ast_fwd_setting    cfn_setting  = {0, 0, {0}};
    struct ast_fwd_setting    cfb_setting  = {0, 0, {0}};
    struct ast_fwd_setting    rcfu_setting = {0, 0, {0}};
    struct ast_fwd_setting    rcfn_setting = {0, 0, {0}};
    struct ast_fwd_setting    rcfb_setting = {0, 0, {0}};

    /* get ext */
    ext = strsep(&context, "-");

    if (!(fwd_setting = fwd_ext_setting_find(ext))) {
        ast_log(LOG_NOTICE, "fwd_ext_setting_update, can't find ext[%s]", ext);
        return -1;
    }

    if (strcmp(cat, fwd_setting->state_context)) {
        //ast_log(LOG_NOTICE, "fwd_ext_setting_update, state_context[%s] not match cat[%s]", fwd_setting->state_context, cat);
        return -1;
    }

	for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
		if (!strcasecmp(var->name, "cfu")) {
            if (!ast_strlen_zero(var->value)) {
                cfu_setting.enabled = 1;
                ast_copy_string(cfu_setting.dst_number, var->value, sizeof(cfu_setting.dst_number));
            }
        } else if (!strcasecmp(var->name, "cfu_time_type")) {
            cfu_setting.time_type = atoi(var->value);
        } else if (!strcasecmp(var->name, "cfn")) {
            if (!ast_strlen_zero(var->value)) {
                cfn_setting.enabled = 1;
                ast_copy_string(cfn_setting.dst_number, var->value, sizeof(cfn_setting.dst_number));
            } 
        } else if (!strcasecmp(var->name, "cfn_time_type")) {
            cfn_setting.time_type = atoi(var->value);
        } else if (!strcasecmp(var->name, "cfb")) {
            if (!ast_strlen_zero(var->value)) {
                cfb_setting.enabled = 1;
                ast_copy_string(cfb_setting.dst_number, var->value, sizeof(cfb_setting.dst_number));
            } 
        } else if (!strcasecmp(var->name, "cfb_time_type")) {
            cfb_setting.time_type = atoi(var->value);
        } else if (!strcasecmp(var->name, "rcfu")) {
            if (!ast_strlen_zero(var->value)) {
                rcfu_setting.enabled = 1;
                ast_copy_string(rcfu_setting.dst_number, var->value, sizeof(rcfu_setting.dst_number));
            } 
        } else if (!strcasecmp(var->name, "rcfu_time_type")) {
            rcfu_setting.time_type = atoi(var->value);
        } else if (!strcasecmp(var->name, "rcfn")) {
            if (!ast_strlen_zero(var->value)) {
                rcfn_setting.enabled = 1;
                ast_copy_string(rcfn_setting.dst_number, var->value, sizeof(rcfn_setting.dst_number));
            }
        } else if (!strcasecmp(var->name, "rcfn_time_type")) {
            rcfn_setting.time_type = atoi(var->value);
        } else if (!strcasecmp(var->name, "rcfb")) {
            if (!ast_strlen_zero(var->value)) {
                rcfb_setting.enabled = 1;
                ast_copy_string(rcfb_setting.dst_number, var->value, sizeof(rcfb_setting.dst_number));
            }
        } else if (!strcasecmp(var->name, "rcfb_time_type")) {
            rcfb_setting.time_type = atoi(var->value);
        } 
    }

    fwd_setting_copy(&cfu_setting, &fwd_setting->cfu_setting);
    fwd_setting_copy(&cfn_setting, &fwd_setting->cfn_setting);
    fwd_setting_copy(&cfb_setting, &fwd_setting->cfb_setting);
    fwd_setting_copy(&rcfu_setting, &fwd_setting->rcfu_setting);
    fwd_setting_copy(&rcfn_setting, &fwd_setting->rcfn_setting);
    fwd_setting_copy(&rcfb_setting, &fwd_setting->rcfb_setting);

    //ast_log(LOG_DEBUG, "fwd_ext_setting_update, [%s]", cat);
	return 0;
}

static int fwd_timing_load_transfer(int reload)
{
    struct ast_flags config_flags = {reload ? CONFIG_FLAG_FILEUNCHANGED : 0};
	struct ast_config   *cfg = NULL;
	const char          *cat = NULL;
	int    res = 0;

    /* load conf file*/
	cfg = ast_config_load(CALL_TRANSFER_CONFIG_FILE, config_flags);

	if (cfg == CONFIG_STATUS_FILEUNCHANGED || NULL == cfg) {
		return 0;
	} 

	if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "call_transfer.conf is invalid.\n");
		return -1;
    }

	if (!reload) {
		con_extension_fwd_settings = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0, HASH_EXT_SIZE, fwd_ext_setting_hash, NULL, fwd_ext_setting_cmp);
	} else {
		ao2_callback(con_extension_fwd_settings, 0, fwd_ext_setting_mark, NULL);
	}

	ast_mutex_lock(&time_lock);
	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		if (!strstr(cat, "-")) {
            fwd_ext_setting_build(cfg, cat, reload);
			continue;
		}
		fwd_ext_setting_update(cfg, cat);
	}
	ast_config_destroy(cfg);

	if (reload) {
		ao2_callback(con_extension_fwd_settings, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, fwd_ext_setting_is_marked, NULL);
	}

	ast_mutex_unlock(&time_lock);
	ast_log(LOG_NOTICE, "fwd_timing_load_transfer finished.\n");
	return 0;
}

static int fwd_check_office_timing(struct timeval tv)
{
    int matched = 0;
    struct timing_obj  *timing_obj;

    AST_LIST_LOCK(&list_office_timings);
    AST_LIST_TRAVERSE(&list_office_timings, timing_obj, entry) {
        if (timing_obj && ast_check_timing2(&timing_obj->timing, tv)) {
            matched = 1;
            break;
        }
    }
    AST_LIST_UNLOCK(&list_office_timings);

    return matched;
}

static int fwd_check_holiday_timing(struct timeval tv)
{
    int matched = 0;
    struct timing_obj  *timing_obj;

    AST_LIST_LOCK(&list_holiday_timings);
    AST_LIST_TRAVERSE(&list_holiday_timings, timing_obj, entry) {
        if (timing_obj && ast_check_timing2(&timing_obj->timing, tv)) {
            matched = 1;
            break;
        }
    }
    AST_LIST_UNLOCK(&list_holiday_timings);

    return matched;
}


static int fwd_check_specific_timing(char *ext, struct timeval tv)
{
    int matched = 0;
    struct timing_obj  *timing_obj;
	RAII_VAR(struct timing_specific *, timing_spec, NULL, ao2_cleanup);

    timing_spec = fwd_timing_find_spec(ext);
    if (!timing_spec) {
        return 0;
    }

    AST_LIST_LOCK(&timing_spec->time_list);
    AST_LIST_TRAVERSE(&timing_spec->time_list, timing_obj, entry) {
        if (timing_obj && ast_check_timing2(&timing_obj->timing, tv)) {
            matched = 1;
            break;
        }
    }
    AST_LIST_UNLOCK(&timing_spec->time_list);

    return matched;
}

static int fwd_ext_timing_get_status(char *ext, struct ast_fwd_setting *setting, struct ast_fwd_status *status)
{
	struct timeval tv = ast_tvnow();

    if (!ext || !setting || !status) {
        return 0;
    }

    if (!setting->enabled) {
        status->active = 0;
        memset(status->dst_num, 0, sizeof(status->dst_num));
        return 0;
    }

    strncpy(status->dst_num, setting->dst_number, sizeof(status->dst_num));
    status->active = 0;

    switch (setting->time_type) {
    case TIME_ALL:
        status->active = 1;
        break;

    case TIME_OFFICE:
        if (fwd_check_office_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_NON_OFFICE:
        if (!fwd_check_office_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_HOLIDAY:
        if (fwd_check_holiday_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_NON_HOLIDAY:
        if (!fwd_check_holiday_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_NON_OFFICE_OR_HOLIDAY:
        if (!fwd_check_office_timing(tv) || fwd_check_holiday_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_SPECIFIC:
        if (fwd_check_specific_timing(ext, tv)) {
            status->active = 1;
        }
        break;

    case TIME_SPECIFIC_AND_NON_HOLIDAY:
        if (fwd_check_specific_timing(ext, tv) && !fwd_check_holiday_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_OFFICE_AND_NON_HOLIDAY:
        if (fwd_check_office_timing(tv) && !fwd_check_holiday_timing(tv)) {
            status->active = 1;
        }
        break;

    case TIME_NON_SPECIFIC:
        if (!fwd_check_specific_timing(ext, tv)) {
            status->active = 1;
        }
        break;

    case TIME_NON_SPECIFIC_OR_HOLIDAY:
        if (!fwd_check_specific_timing(ext, tv) || fwd_check_holiday_timing(tv)) {
            status->active = 1;
        }
        break;

    default:
        break;
    }

    return 0;
}

int ast_fwd_timetype_check(char *ext, int time_type)
{
	struct timeval tv = ast_tvnow();
    int matched = 0;

    switch (time_type) {
    case TIME_ALL:
        matched = 1;
        break;

    case TIME_OFFICE:
        if (fwd_check_office_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_NON_OFFICE:
        if (!fwd_check_office_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_HOLIDAY:
        if (fwd_check_holiday_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_NON_HOLIDAY:
        if (!fwd_check_holiday_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_NON_OFFICE_OR_HOLIDAY:
        if (!fwd_check_office_timing(tv) || fwd_check_holiday_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_SPECIFIC:
        if (fwd_check_specific_timing(ext, tv)) {
            matched = 1;
        }
        break;

    case TIME_SPECIFIC_AND_NON_HOLIDAY:
        if (fwd_check_specific_timing(ext, tv) && !fwd_check_holiday_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_OFFICE_AND_NON_HOLIDAY:
        if (fwd_check_office_timing(tv) && !fwd_check_holiday_timing(tv)) {
            matched = 1;
        }
        break;

    case TIME_NON_SPECIFIC:
        if (!fwd_check_specific_timing(ext, tv)) {
            matched = 1;
        }
        break;

    case TIME_NON_SPECIFIC_OR_HOLIDAY:
        if (!fwd_check_specific_timing(ext, tv) || fwd_check_holiday_timing(tv)) {
            matched = 1;
        }
        break;

    default:
        break;
    }

    return matched;
}

static int fwd_ext_status_cmp(struct ast_fwd_status *status1, struct ast_fwd_status *status2)
{
    if (!status1 || !status2) {
        return 1;
    }

    if (status1->active != status2->active) {
        return 1;
    }

    if (strncmp(status1->dst_num, status2->dst_num, sizeof(status1->dst_num))) {
        return 1;
    }

    return 0;
}


static int fwd_status_cpy(struct ast_fwd_status *status_src, struct ast_fwd_status *status_dst)
{
    if (!status_dst || !status_src) {
        return 1;
    }

    status_dst->active = status_src->active;
    status_dst->update = status_src->update;
    strncpy(status_dst->dst_num, status_src->dst_num, sizeof(status_dst->dst_num));

    return 0;
}

int ast_fwd_active_status_cpy(struct ast_fwd_active_status *status_src, struct ast_fwd_active_status *status_dst)
{
    if (!status_src || !status_dst) {
        return 1;
    }

    fwd_status_cpy(&status_src->cfu_status,  &status_dst->cfu_status);
    fwd_status_cpy(&status_src->cfn_status,  &status_dst->cfn_status);
    fwd_status_cpy(&status_src->cfb_status,  &status_dst->cfb_status);
    fwd_status_cpy(&status_src->rcfu_status, &status_dst->rcfu_status);
    fwd_status_cpy(&status_src->rcfn_status, &status_dst->rcfn_status);
    fwd_status_cpy(&status_src->rcfb_status, &status_dst->rcfb_status);

    return 0;
}

static int fwd_ext_set_update(struct ast_fwd_ext_status *fwd_status)
{
    if (!fwd_status) {
        return 0;
    }

    fwd_status->cur_status.cfu_status.update = fwd_ext_status_cmp(&fwd_status->cur_status.cfu_status, &fwd_status->last_status.cfu_status);
    fwd_status->cur_status.cfn_status.update = fwd_ext_status_cmp(&fwd_status->cur_status.cfn_status, &fwd_status->last_status.cfn_status);
    fwd_status->cur_status.cfb_status.update = fwd_ext_status_cmp(&fwd_status->cur_status.cfb_status, &fwd_status->last_status.cfb_status);
    fwd_status->cur_status.rcfu_status.update = fwd_ext_status_cmp(&fwd_status->cur_status.rcfu_status, &fwd_status->last_status.rcfu_status);
    fwd_status->cur_status.rcfn_status.update = fwd_ext_status_cmp(&fwd_status->cur_status.rcfn_status, &fwd_status->last_status.rcfn_status);
    fwd_status->cur_status.rcfb_status.update = fwd_ext_status_cmp(&fwd_status->cur_status.rcfb_status, &fwd_status->last_status.rcfb_status);

    //ast_fwd_active_status_cpy(&fwd_status->cur_status, &fwd_status->last_status);

    ast_log(LOG_DEBUG, "ext[%s], update/active cfu:%d/%d, cfn:%d/%d, cfb:%d/%d, rcfu:%d/%d, rcfn:%d/%d, rcfb:%d/%d.", fwd_status->ext, 
            fwd_status->cur_status.cfu_status.update, fwd_status->cur_status.cfu_status.active, 
            fwd_status->cur_status.cfn_status.update, fwd_status->cur_status.cfn_status.active,
            fwd_status->cur_status.cfb_status.update, fwd_status->cur_status.cfb_status.active,
            fwd_status->cur_status.rcfu_status.update, fwd_status->cur_status.rcfu_status.active,
            fwd_status->cur_status.rcfn_status.update, fwd_status->cur_status.rcfn_status.active,
            fwd_status->cur_status.rcfb_status.update, fwd_status->cur_status.rcfb_status.active);
    return 0;
}


int ast_fwd_notify_info_cpy(struct ast_fwd_notify_info *info_src, struct ast_fwd_notify_info *info_dst)
{
    if (!info_src || !info_dst) {
        return 0;
    }

    strncpy(info_dst->ext, info_src->ext, sizeof(info_dst->ext));
    ast_fwd_active_status_cpy(&info_src->notify_status, &info_dst->notify_status);
    return 0;
    
}

static int fwd_ext_check_process(struct ast_fwd_ext_status *fwd_status)
{
	RAII_VAR(struct ext_fwd_setting *, fwd_setting, NULL, ao2_cleanup);

    if (!fwd_status) {
        ast_log(LOG_NOTICE, "input param is NULL");
        return 0;
    }

    /* 1. get ext_fwd_setting object */
    fwd_setting = fwd_ext_setting_find(fwd_status->ext);
    if (!fwd_setting) {
        ast_log(LOG_NOTICE, "can't find fwd_setting, ext[%s]", fwd_status->ext);
        return 0;
    }

    ast_log(LOG_DEBUG, "fwd_ext_check_process, ext[%s]", fwd_status->ext);

    /* 2. update fwd status */
    fwd_ext_timing_get_status(fwd_status->ext, &fwd_setting->cfu_setting, &fwd_status->cur_status.cfu_status);
    fwd_ext_timing_get_status(fwd_status->ext, &fwd_setting->cfn_setting, &fwd_status->cur_status.cfn_status);
    fwd_ext_timing_get_status(fwd_status->ext, &fwd_setting->cfb_setting, &fwd_status->cur_status.cfb_status);
    fwd_ext_timing_get_status(fwd_status->ext, &fwd_setting->rcfu_setting, &fwd_status->cur_status.rcfu_status);
    fwd_ext_timing_get_status(fwd_status->ext, &fwd_setting->rcfn_setting, &fwd_status->cur_status.rcfn_status);
    fwd_ext_timing_get_status(fwd_status->ext, &fwd_setting->rcfb_setting, &fwd_status->cur_status.rcfb_status);

    /* 3. set update flag */
    fwd_ext_set_update(fwd_status);

    return 0;
}

static int fwd_ext_check_timeout(const void *data)
{
    struct ast_fwd_ext_status *fwd_status = NULL;

    fwd_timing_load_timing(reload_flag);
    fwd_timing_load_transfer(reload_flag);

    if (AST_LIST_EMPTY(&list_fwd_ext_status)) {
        return (FWD_CHECK_INTVAL * 1000);
    }

    AST_LIST_LOCK(&list_fwd_ext_status);
    AST_LIST_TRAVERSE(&list_fwd_ext_status, fwd_status, entry) {
        fwd_ext_check_process(fwd_status);
    }

    AST_LIST_UNLOCK(&list_fwd_ext_status);

	return (FWD_CHECK_INTVAL * 1000);
}

int ast_fwd_active_status_update_set(struct ast_fwd_active_status *active_status, int flag)
{
    if (!active_status) {
        return 0;
    }

    active_status->cfb_status.update = flag;
    active_status->cfu_status.update = flag;
    active_status->cfn_status.update = flag;
    active_status->rcfb_status.update = flag;
    active_status->rcfu_status.update = flag;
    active_status->rcfn_status.update = flag;

    return 0;
}

static struct ast_fwd_ext_status *fwd_ext_status_build(char *ext)
{
    struct ast_fwd_ext_status *ext_status = NULL;

    if (!ext) {
        return NULL;
    }

    /* alloc */
	if (!(ext_status = ast_calloc(1, sizeof(*ext_status)))) {
		ast_log(LOG_NOTICE, "fwd_ext_status_build, ast_calloc failed");
		return NULL;
	}

    strncpy(ext_status->ext, ext, sizeof(ext_status->ext));

    /* get data */
    fwd_ext_check_process(ext_status);

    /* insert into list */
    AST_LIST_INSERT_TAIL(&list_fwd_ext_status, ext_status, entry);

    ast_log(LOG_NOTICE, "fwd_ext_status_build, ext[%s]", ext_status->ext);
    return ext_status;
}

/* full:1 get all fwd status
   full:0 get update fwd status
*/
int ast_fwd_ext_get_notify_data(char *ext, struct ast_fwd_active_status *active_status, int full)
{

    struct ast_fwd_ext_status *ext_status = NULL;
    int need_insert = 1;

    if (!ext || !active_status) {
        return -1;
    }

    ast_log(LOG_NOTICE, "fwd_ext_get_notify_data, ext[%s].", ext);

    AST_LIST_LOCK(&list_fwd_ext_status);
    AST_LIST_TRAVERSE(&list_fwd_ext_status, ext_status, entry) {
        if (ext_status && !strcmp(ext, ext_status->ext)) {
            /* get active_status from ext_status */
            ast_fwd_active_status_cpy(&ext_status->cur_status, active_status);

            if (full) {
                ast_fwd_active_status_update_set(active_status, 1);
            }

            need_insert = 0;
            break;
        }
    }

    AST_LIST_UNLOCK(&list_fwd_ext_status);

    if (need_insert) {
        ext_status = fwd_ext_status_build(ext);
        ast_fwd_active_status_cpy(&ext_status->cur_status, active_status);
        ast_fwd_active_status_update_set(active_status, 1);
        return 0;
    }

    return 0;
}

static int fwd_ext_notify_process(struct ast_fwd_ext_status *fwd_status)
{
	RAII_VAR(struct ast_fwd_notify_info *, info, NULL, ao2_cleanup);
	RAII_VAR(struct stasis_message *, message, NULL, ao2_cleanup);
	struct stasis_topic *forward_notify_topic;

    if (!fwd_status) {
        return 0;
    }

    /* check update flag */
    if (!fwd_status->cur_status.cfu_status.update && !fwd_status->cur_status.cfn_status.update && 
        !fwd_status->cur_status.cfb_status.update && !fwd_status->cur_status.rcfu_status.update &&
        !fwd_status->cur_status.rcfn_status.update && !fwd_status->cur_status.rcfb_status.update) {
        return 0;
    }

	if (!ast_feature_notify_type()) {
		return -1;
	}

	info = ao2_alloc(sizeof(*info), NULL);
	if (!info) {
		return -1;
	}

	snprintf(info->ext, sizeof(info->ext), "%s", fwd_status->ext);
    ast_fwd_active_status_cpy(&fwd_status->cur_status, &info->notify_status);
    ast_fwd_active_status_cpy(&fwd_status->cur_status, &fwd_status->last_status);

	message = stasis_message_create(ast_feature_notify_type(), info);
	if (!message) {
		return -1;
	}

	forward_notify_topic = ast_forward_notify_topic();
	if (!forward_notify_topic) {
		return -1;
	}

	ast_log(LOG_NOTICE, "push notify data, ext[%s]\n", fwd_status->ext);
	stasis_publish(forward_notify_topic, message);

	return 0;
}

static int fwd_ext_notify_timeout(const void *data)
{
    struct ast_fwd_ext_status *fwd_status = NULL;

    if (AST_LIST_EMPTY(&list_fwd_ext_status)) {
        return (FWD_NOTIFY_INTVAL * 1000);
    }

    AST_LIST_LOCK(&list_fwd_ext_status);

    AST_LIST_TRAVERSE(&list_fwd_ext_status, fwd_status, entry) {
        fwd_ext_notify_process(fwd_status);
    }

    AST_LIST_UNLOCK(&list_fwd_ext_status);

	return (FWD_NOTIFY_INTVAL * 1000);
}

static void fwd_ext_check_start(void)
{
	fwd_check_sched = ast_sched_context_create();
	if (!fwd_check_sched) {
		return;
	}

	if (ast_sched_start_thread(fwd_check_sched)) {
		ast_log(LOG_ERROR, "Could not start fwd_check_sched\n");
		ast_sched_context_destroy(fwd_check_sched);
		fwd_check_sched = NULL;
		return;
	}

    fwd_timing_load_timing(0);
    fwd_timing_load_transfer(0);

    reload_flag = 1;
	fwd_check_sched_id = ast_sched_add(fwd_check_sched, FWD_CHECK_INTVAL * 1000, fwd_ext_check_timeout, NULL);
	if (fwd_check_sched_id < 0) {
		ast_log(LOG_ERROR, "Could not add fwd_check_sched_id\n");
		ast_sched_context_destroy(fwd_check_sched);
		fwd_check_sched = NULL;
		return;
	}

	fwd_notify_sched_id = ast_sched_add(fwd_check_sched, FWD_NOTIFY_INTVAL * 1000, fwd_ext_notify_timeout, NULL);
	if (fwd_notify_sched_id < 0) {
		ast_log(LOG_ERROR, "Could not add fwd_notify_sched_id\n");
		AST_SCHED_DEL(fwd_check_sched, fwd_check_sched_id);
		ast_sched_context_destroy(fwd_check_sched);
		fwd_check_sched = NULL;
		return;
	}

	return;
}


static void fwd_timing_shutdown(void)
{
    struct ast_fwd_ext_status *ext_status = NULL;
    struct timing_obj *t_obj = NULL;

	if (fwd_check_sched) {
		AST_SCHED_DEL(fwd_check_sched, fwd_check_sched_id);
		AST_SCHED_DEL(fwd_check_sched, fwd_notify_sched_id);
		ast_sched_context_destroy(fwd_check_sched);
		fwd_check_sched = NULL;
	}

    ao2_callback(con_specific_timings, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
    ao2_callback(con_extension_fwd_settings, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	while ((ext_status = AST_LIST_REMOVE_HEAD(&list_fwd_ext_status, entry))) {
		ast_free(ext_status);
	}

	while ((t_obj = AST_LIST_REMOVE_HEAD(&list_office_timings, entry))) {
		ast_free(t_obj);
	}

	while ((t_obj = AST_LIST_REMOVE_HEAD(&list_holiday_timings, entry))) {
		ast_free(t_obj);
	}

	ast_mutex_destroy(&time_lock);
}

int ast_fwd_timing_init(void)
{
	ast_mutex_init(&time_lock);

    AST_LIST_HEAD_INIT(&list_fwd_ext_status);
	fwd_ext_check_start();

    ast_register_cleanup(fwd_timing_shutdown);
	return 0;
}

