/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2013, Digium, Inc.
 * Copyright (C) 2012, Russell Bryant
 *
 * Matt Jordan <mjordan@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Routines implementing call pickup
 *
 * \author Matt Jordan <mjordan@digium.com>
 */

/*!
 * \li Call pickup uses the configuration file \ref features.conf
 * \addtogroup configuration_file Configuration Files
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

/*** DOCUMENTATION
	<managerEvent language="en_US" name="Pickup">
		<managerEventInstance class="EVENT_FLAG_CALL">
			<synopsis>Raised when a call pickup occurs.</synopsis>
			<syntax>
				<channel_snapshot/>
				<channel_snapshot prefix="Target"/>
			</syntax>
		</managerEventInstance>
	</managerEvent>
 ***/

#include "asterisk.h"

#include "asterisk/pickup.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/app.h"
#include "asterisk/callerid.h"
#include "asterisk/causes.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/features_config.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/utils.h"
#include "asterisk/dial.h"
AST_MODULE_LOG("Channel");
#endif

static struct ast_manager_event_blob *call_pickup_to_ami(struct stasis_message *message);

STASIS_MESSAGE_TYPE_DEFN(
	ast_call_pickup_type,
	.to_ami = call_pickup_to_ami);


/*!
 * The presence of this datastore on the channel indicates that
 * someone is attemting to pickup or has picked up the channel.
 * The purpose is to prevent a race between two channels
 * attempting to pickup the same channel.
 */
static const struct ast_datastore_info pickup_active = {
	.type = "pickup-active",
};

int ast_can_pickup(struct ast_channel *chan)
{
	if (!ast_channel_pbx(chan) && !ast_channel_masq(chan) && !ast_test_flag(ast_channel_flags(chan), AST_FLAG_ZOMBIE)
		&& (ast_channel_state(chan) == AST_STATE_RINGING
#ifndef GRANDSTREAM_NETWORKS
			|| ast_channel_state(chan) == AST_STATE_RING
#endif
			/*
			 * Check the down state as well because some SIP devices do not
			 * give 180 ringing when they can just give 183 session progress
			 * instead.  Issue 14005.  (Some ISDN switches as well for that
			 * matter.)
			 */
			|| ast_channel_state(chan) == AST_STATE_DOWN)
		&& !ast_channel_datastore_find(chan, &pickup_active, NULL)) {

#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug 108793 */
		const char* sca_app = pbx_builtin_getvar_helper(chan, "SCA_APP_CALL");
		if (!ast_strlen_zero(sca_app) && !strcmp(sca_app, "yes")) {
			return 0;
		}
#endif

		return 1;
	}
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int match_pickup_group_list(struct ast_channel *chan, struct ast_channel *target)
{
	if (chan == NULL || target == NULL || ast_channel_pickupgroup_list(chan) == NULL || ast_channel_pickupgroup_list(target) == NULL) {
		ast_log(LOG_DEBUG,"chan is NULL or chan->pickup_conf is NULL");
		return 0;
	}

	struct group_id_list *call_list;
	struct group_id_list *pick_list;

	AST_LIST_TRAVERSE(&(ast_channel_pickupgroup_list(target)->call_group_id_list),call_list,entry) {
		AST_LIST_TRAVERSE(&(ast_channel_pickupgroup_list(chan)->pick_group_id_list),pick_list,entry) {
			if ( call_list->group_id == pick_list->group_id) {
				ast_log(LOG_DEBUG,"match group id = %d",call_list->group_id);
				return 1;
			}
		}
	}

	return 0;
}
#endif

static int find_channel_by_group(void *obj, void *arg, void *data, int flags)
{
	struct ast_channel *target = obj; /*!< Potential pickup target */
	struct ast_channel *chan = arg;   /*!< Channel wanting to pickup call */

	if (chan == target) {
		return 0;
	}

	ast_channel_lock(target);
	if (ast_can_pickup(target)) {
		/* Lock both channels. */
		while (ast_channel_trylock(chan)) {
			ast_channel_unlock(target);
			sched_yield();
			ast_channel_lock(target);
		}
#ifdef GRANDSTREAM_NETWORKS
		int match = match_pickup_group_list(chan,target);
#endif

		/*
		 * Both callgroup and namedcallgroup pickup variants are
		 * matched independently.  Checking for named group match is
		 * done last since it's a more expensive operation.
		 */
		 /*Keep two matching methods,update by ffzhao.*/
#ifdef GRANDSTREAM_NETWORKS
		if ((match || (ast_channel_pickupgroup(chan) & ast_channel_callgroup(target)))
#else
		if ((ast_channel_pickupgroup(chan) & ast_channel_callgroup(target))
#endif
			|| (ast_namedgroups_intersect(ast_channel_named_pickupgroups(chan),
				ast_channel_named_callgroups(target)))) {
			struct ao2_container *candidates = data;/*!< Candidate channels found. */

			/* This is a candidate to pickup */
			ao2_link(candidates, target);
		}
		ast_channel_unlock(chan);
	}
	ast_channel_unlock(target);

	return 0;
}

struct ast_channel *ast_pickup_find_by_group(struct ast_channel *chan)
{
	struct ao2_container *candidates;/*!< Candidate channels found to pickup. */
	struct ast_channel *target;/*!< Potential pickup target */

	candidates = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0, NULL, NULL);
	if (!candidates) {
		return NULL;
	}

	/* Find all candidate targets by group. */
	ast_channel_callback(find_channel_by_group, chan, candidates, 0);

	/* Find the oldest pickup target candidate */
	target = NULL;
	for (;;) {
		struct ast_channel *candidate;/*!< Potential new older target */
		struct ao2_iterator iter;

		iter = ao2_iterator_init(candidates, 0);
		while ((candidate = ao2_iterator_next(&iter))) {
			if (!target) {
				/* First target. */
				target = candidate;
				continue;
			}
			if (ast_tvcmp(ast_channel_creationtime(candidate), ast_channel_creationtime(target)) < 0) {
				/* We have a new target. */
				ast_channel_unref(target);
				target = candidate;
				continue;
			}
			ast_channel_unref(candidate);
		}
		ao2_iterator_destroy(&iter);
		if (!target) {
			/* No candidates found. */
			break;
		}

		/* The found channel must be locked and ref'd. */
		ast_channel_lock(target);

		/* Recheck pickup ability */
		if (ast_can_pickup(target)) {
			/* This is the channel to pickup. */
			break;
		}

		/* Someone else picked it up or the call went away. */
		ast_channel_unlock(target);
		ao2_unlink(candidates, target);
		target = ast_channel_unref(target);
	}
	ao2_ref(candidates, -1);

	return target;
}

/*!
 * \brief Pickup a call
 * \param chan channel that initiated pickup.
 *
 * Walk list of channels, checking it is not itself, channel is pbx one,
 * check that the callgroup for both channels are the same and the channel is ringing.
 * Answer calling channel, flag channel as answered on queue, masq channels together.
 */
int ast_pickup_call(struct ast_channel *chan)
{
	struct ast_channel *target;/*!< Potential pickup target */
	int res = -1;
	RAII_VAR(struct ast_features_pickup_config *, pickup_cfg, NULL, ao2_cleanup);
	const char *pickup_sound;
	const char *fail_sound;

	ast_debug(1, "Pickup attempt by %s\n", ast_channel_name(chan));
	ast_channel_lock(chan);
	pickup_cfg = ast_get_chan_features_pickup_config(chan);
	if (!pickup_cfg) {
		ast_log(LOG_ERROR, "Unable to retrieve pickup configuration. Unable to play pickup sounds\n");
	}
	pickup_sound = ast_strdupa(pickup_cfg ? pickup_cfg->pickupsound : "");
	fail_sound = ast_strdupa(pickup_cfg ? pickup_cfg->pickupfailsound : "");
	ast_channel_unlock(chan);

	/* The found channel is already locked. */
	target = ast_pickup_find_by_group(chan);
	if (target) {
		ast_log(LOG_NOTICE, "Pickup %s attempt by %s\n", ast_channel_name(target), ast_channel_name(chan));

		res = ast_do_pickup(chan, target);
		ast_channel_unlock(target);
		if (!res) {
			if (!ast_strlen_zero(pickup_sound)) {
				pbx_builtin_setvar_helper(target, "BRIDGE_PLAY_SOUND", pickup_sound);
			}
		} else {
			ast_log(LOG_WARNING, "Pickup %s failed by %s\n", ast_channel_name(target), ast_channel_name(chan));
		}
		target = ast_channel_unref(target);
	}

	if (res < 0) {
		ast_debug(1, "No call pickup possible... for %s\n", ast_channel_name(chan));
		if (!ast_strlen_zero(fail_sound)) {
			ast_answer(chan);
			ast_stream_and_wait(chan, fail_sound, "");
		}
	}

	return res;
}

static struct ast_manager_event_blob *call_pickup_to_ami(struct stasis_message *message)
{
	struct ast_multi_channel_blob *contents = stasis_message_data(message);
	struct ast_channel_snapshot *chan;
	struct ast_channel_snapshot *target;
	struct ast_manager_event_blob *res;

	RAII_VAR(struct ast_str *, channel_str, NULL, ast_free);
	RAII_VAR(struct ast_str *, target_str, NULL, ast_free);

	chan = ast_multi_channel_blob_get_channel(contents, "channel");
	target = ast_multi_channel_blob_get_channel(contents, "target");

	ast_assert(chan != NULL && target != NULL);

	if (!(channel_str = ast_manager_build_channel_state_string(chan))) {
		return NULL;
	}

	if (!(target_str = ast_manager_build_channel_state_string_prefix(target, "Target"))) {
		return NULL;
	}

	res = ast_manager_event_blob_create(EVENT_FLAG_CALL, "Pickup",
		"%s"
		"%s",
		ast_str_buffer(channel_str),
		ast_str_buffer(target_str));

	return res;
}

static int send_call_pickup_stasis_message(struct ast_channel *picking_up, struct ast_channel_snapshot *chan, struct ast_channel_snapshot *target)
{
	RAII_VAR(struct ast_multi_channel_blob *, pickup_payload, NULL, ao2_cleanup);
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);

	if (!ast_call_pickup_type()) {
		return -1;
	}

	if (!(pickup_payload = ast_multi_channel_blob_create(ast_json_null()))) {
		return -1;
	}

	ast_multi_channel_blob_add_channel(pickup_payload, "channel", chan);
	ast_multi_channel_blob_add_channel(pickup_payload, "target", target);

	if (!(msg = stasis_message_create(ast_call_pickup_type(), pickup_payload))) {
		return -1;
	}

	stasis_publish(ast_channel_topic(picking_up), msg);
	return 0;
}

int ast_do_pickup(struct ast_channel *chan, struct ast_channel *target)
{
	struct ast_party_connected_line connected_caller;
	struct ast_datastore *ds_pickup;
	const char *chan_name;/*!< A masquerade changes channel names. */
	const char *target_name;/*!< A masquerade changes channel names. */
	int res = -1;

	RAII_VAR(struct ast_channel_snapshot *, chan_snapshot, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel_snapshot *, target_snapshot, NULL, ao2_cleanup);

	target_name = ast_strdupa(ast_channel_name(target));
	ast_debug(1, "Call pickup on '%s' by '%s'\n", target_name, ast_channel_name(chan));

	/* Mark the target to block any call pickup race. */
	ds_pickup = ast_datastore_alloc(&pickup_active, NULL);
	if (!ds_pickup) {
		ast_log(LOG_WARNING,
			"Unable to create channel datastore on '%s' for call pickup\n", target_name);
		return -1;
	}
	ast_channel_datastore_add(target, ds_pickup);

	ast_party_connected_line_init(&connected_caller);
	ast_party_connected_line_copy(&connected_caller, ast_channel_connected(target));
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for Bug 233713, 2022.8.18 */
	ast_channel_set_is_pickup(target, 1);
#endif
	ast_channel_unlock(target);/* The pickup race is avoided so we do not need the lock anymore. */
	/* Reset any earlier private connected id representation */
	ast_party_id_reset(&connected_caller.priv);

	connected_caller.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;
	if (ast_channel_connected_line_sub(NULL, chan, &connected_caller, 0) &&
		ast_channel_connected_line_macro(NULL, chan, &connected_caller, 0, 0)) {
		ast_channel_update_connected_line(chan, &connected_caller, NULL);
	}
	ast_party_connected_line_free(&connected_caller);

	ast_channel_lock(chan);
	chan_name = ast_strdupa(ast_channel_name(chan));
	ast_connected_line_copy_from_caller(&connected_caller, ast_channel_caller(chan));
	ast_channel_unlock(chan);
	connected_caller.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;

#ifdef GRANDSTREAM_NETWORKS
	pbx_builtin_setvar_helper(chan, "__TRANSPARENTSDP", "no");
	pbx_builtin_setvar_helper(chan, "PICKUP_CALLER", "yes");

	/* Added by yfyang, for Bug 118333 */
	const char *conf_user = pbx_builtin_getvar_helper(target, "CONF_USER_INVITE");
	if (!ast_strlen_zero(conf_user) && !strncmp(conf_user, "yes", 3)) {

		struct ast_app *setfocus = pbx_findapp("SIPSetConfFocus");
		if (setfocus) {
			const char *conf_room = pbx_builtin_getvar_helper(target, "CONF_ROOM");
			ast_log(LOG_NOTICE, "About to call SIPSetConfFocus to chan[%s]\n", ast_channel_name(chan));
			pbx_exec(chan, setfocus, conf_room);
		}
	}

	/* Add by bfdu for Bug 121880, 2018.09.29 */
	const char *is_sfu = pbx_builtin_getvar_helper(target, "IS_SFU");
	if (is_sfu && !strcasecmp(is_sfu, "yes")) {
		ast_log(LOG_DEBUG, "target is sfu, can't support pickup\n");
		goto pickup_failed;
	}

	/* Added by yfyang, for Bug 91578 */
	const char *callinfo_msg = pbx_builtin_getvar_helper(target, "CALL_INFO_MSG");
	if (!ast_strlen_zero(callinfo_msg)) {
		const char *callinfo_val = pbx_builtin_getvar_helper(target, "CALL_INFO");
		if (!ast_strlen_zero(callinfo_val)) {
			pbx_builtin_setvar_helper(chan, "CALL_INFO_MSG", callinfo_msg);
			pbx_builtin_setvar_helper(chan, "CALL_INFO", callinfo_val);
		}
	}

	/* Add by bfdu for Bug 130876, 2019.01.28 */
	if (target) {
		struct ast_format_cap *nativeformats = ast_channel_nativeformats(target);
		if (chan && nativeformats && !ast_format_cap_has_type(nativeformats, AST_MEDIA_TYPE_VIDEO)) {
			ast_log(LOG_DEBUG, "Do pickup, target has no video caps. chan:%s, target:%s\n", ast_channel_name(chan), ast_channel_name(target));
			pbx_builtin_setvar_helper(chan, "HAS_NO_VIDEO_CAPS", "yes");
		}
	}
#endif

	if (ast_answer(chan)) {
		ast_log(LOG_WARNING, "Unable to answer '%s'\n", chan_name);
		goto pickup_failed;
	}

#ifdef GRANDSTREAM_NETWORKS  /*Add by weidu, for Bug276338 */
	if (0 == strcasecmp(ast_channel_tech(chan)->type, "DAHDI")){
		pbx_builtin_setvar_helper(chan, "CALLED_REALHOOK", "yes");
	}
#endif

	if (ast_queue_control(chan, AST_CONTROL_ANSWER)) {
		ast_log(LOG_WARNING, "Unable to queue answer on '%s'\n", chan_name);
		goto pickup_failed;
	}

	ast_channel_queue_connected_line_update(chan, &connected_caller, NULL);

	/* setting the HANGUPCAUSE so the ringing channel knows this call was not a missed call */
	ast_channel_hangupcause_set(chan, AST_CAUSE_ANSWERED_ELSEWHERE);

	ast_channel_lock(chan);
	chan_snapshot = ast_channel_snapshot_create(chan);
	ast_channel_unlock(chan);
	if (!chan_snapshot) {
		goto pickup_failed;
	}

	target_snapshot = ast_channel_snapshot_get_latest(ast_channel_uniqueid(target));
	if (!target_snapshot) {
		goto pickup_failed;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* added by zgmo for Bug144395 */
	if (chan && target) {
		const char *from_callback = pbx_builtin_getvar_helper(target, "FROM_CALLBACK");
		const char *callback_src = pbx_builtin_getvar_helper(target, "CALLBACK_SRC");
		const char *first_caller_channel_name = pbx_builtin_getvar_helper(target, "FIRST_CALLER_CHANNEL_NAME");

		ast_log(LOG_DEBUG, "from_callback:%s, callback_src:%s",
			ast_strlen_zero(from_callback) ? "":from_callback,
			ast_strlen_zero(callback_src) ? "":callback_src);

		pbx_builtin_setvar_helper(target, "PICKUPEE", "yes");
		pbx_builtin_setvar_helper(target, "CALLEE_NUM", NULL);
		pbx_builtin_setvar_helper(chan, "IS_CALLEE", "yes");
		pbx_builtin_setvar_helper(chan, "PICKUPEE_CHAN", ast_channel_name(target));
		pbx_builtin_setvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME", first_caller_channel_name);

		if (!ast_strlen_zero(first_caller_channel_name)) {
			struct ast_channel *initial_caller = ast_channel_get_by_name(first_caller_channel_name);
			if (initial_caller) {
				pbx_builtin_setvar_helper(chan, "FROM_TRUNK", pbx_builtin_getvar_helper(initial_caller, "FROM_TRUNK"));
				pbx_builtin_setvar_helper(chan, "TRUNK_INDEX", pbx_builtin_getvar_helper(initial_caller, "TRUNK_INDEX"));
				pbx_builtin_setvar_helper(chan, "INBOUND_RT_INDEX", pbx_builtin_getvar_helper(initial_caller, "INBOUND_RT_INDEX"));

				ast_channel_unref(initial_caller);
			}
		}

		pbx_builtin_setvar_helper(chan, "FROM_IVR", pbx_builtin_getvar_helper(target, "FROM_IVR"));
		pbx_builtin_setvar_helper(chan, "FROM_DISA", pbx_builtin_getvar_helper(target, "FROM_DISA"));
		pbx_builtin_setvar_helper(chan, "IVR_REC", pbx_builtin_getvar_helper(target, "IVR_REC"));
		pbx_builtin_setvar_helper(chan, "IN_RINGGROUP", pbx_builtin_getvar_helper(target, "IN_RINGGROUP"));
		pbx_builtin_setvar_helper(chan, "GROUP_NUMBER", pbx_builtin_getvar_helper(target, "GROUP_NUMBER"));
		pbx_builtin_setvar_helper(chan, "DIAL_EXEC_FULL_CALLER", pbx_builtin_getvar_helper(target, "DIAL_EXEC_FULL_CALLER"));

#if 0
		char *caller_num = NULL;
		if (!ast_strlen_zero(from_callback) && !ast_strlen_zero(callback_src)) {
			caller_num = (char *)callback_src;
		} else {
			struct ast_party_connected_line *party_target_connected = ast_channel_connected(target);
			caller_num = party_target_connected->ani.number.str;
		}

		if (ast_dial_check_record2(caller_num, NULL, chan)) {
			if (!ast_strlen_zero(first_caller_channel_name)) {
				ast_cdr_setvar(first_caller_channel_name, "RECORD_FILES",
					pbx_builtin_getvar_helper(chan, "RECORDING_FILE_CDR"));
			}
			pbx_builtin_setvar_helper(chan, "RECORD_ATTR_SDP_TRIGGER", "yes");
		} else {
			const char *from_ivr = pbx_builtin_getvar_helper(target, "FROM_IVR");
			if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")) {
				const char *ringgroup_caller_chan = pbx_builtin_getvar_helper(target, "RINGGROUP_CALLER_CHAN");
				if (!ast_strlen_zero(ringgroup_caller_chan)) {
					char cdr_record[128];

					cdr_record[0] = '\0';
					ast_cdr_getvar(ringgroup_caller_chan, "record_files", cdr_record, sizeof(cdr_record));
					cdr_record[sizeof(cdr_record) - 1] = '\0';

					if (!ast_strlen_zero(cdr_record)) {
						ast_log(LOG_DEBUG, "Chan %s -> IVR -> RG(rec on) -> %s(pickup by %s)",
							ringgroup_caller_chan, ast_channel_name(target), ast_channel_name(chan));
						ast_log(LOG_DEBUG, "After pickup, no need to record, clear %s'cdr record_files",
							ringgroup_caller_chan);

						ast_cdr_setvar(ringgroup_caller_chan, "record_files", "");
					}
				}
			}
		}

		pbx_builtin_setvar_helper(target, "REC_ENABLE", pbx_builtin_getvar_helper(chan, "REC_ENABLE"));
#endif
		ast_cdr_setvar(ast_channel_name(target), "special_action_type", "masquerade");
		ast_cdr_setvar(ast_channel_name(target), "status", ast_channel_get_is_nat_call(target) ? "1" : "0");
	}
#endif

	if (ast_channel_move(target, chan)) {
		ast_log(LOG_WARNING, "Unable to complete call pickup of '%s' with '%s'\n",
			chan_name, target_name);
		goto pickup_failed;
	}

	/* target points to the channel that did the pickup at this point, so use that channel's topic instead of chan */
	send_call_pickup_stasis_message(target, chan_snapshot, target_snapshot);

	res = 0;

pickup_failed:
	ast_channel_lock(target);
#ifdef GRANDSTREAM_NETWORKS
	ast_channel_set_is_pickup(target, 0);
#endif
	if (!ast_channel_datastore_remove(target, ds_pickup)) {
		ast_datastore_free(ds_pickup);
	}
	ast_party_connected_line_free(&connected_caller);

	return res;
}

/*!
 * \internal
 * \brief Clean up resources on Asterisk shutdown
 */
static void pickup_shutdown(void)
{
	STASIS_MESSAGE_TYPE_CLEANUP(ast_call_pickup_type);
}

int ast_pickup_init(void)
{
	STASIS_MESSAGE_TYPE_INIT(ast_call_pickup_type);
	ast_register_cleanup(pickup_shutdown);

	return 0;
}
