/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2017, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Media Stream API
 *
 * \author Joshua Colp <jcolp@digium.com>
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/logger.h"
#include "asterisk/stream.h"
#include "asterisk/strings.h"
#include "asterisk/format.h"
#include "asterisk/format_cap.h"
#include "asterisk/vector.h"
#include "asterisk/config.h"
#include "asterisk/rtp_engine.h"

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("stream");

#define SOFTBRIDGE_VIDEO_DEST_PREFIX "softbridge_dest"
#define SOFTBRIDGE_VIDEO_DEST_LEN strlen(SOFTBRIDGE_VIDEO_DEST_PREFIX)

#define MAX_UPTIME_LEN 64

struct ast_video_rtx_pt {
	int h264_rtx_pt;
	int h265_rtx_pt;
	int vp8_rtx_pt;
	int vp9_rtx_pt;
};
#endif

struct ast_stream_metadata_entry {
	size_t length;
	int value_start;
	char name_value[0];
};

struct ast_stream {
	/*!
	 * \brief The type of media the stream is handling
	 */
	enum ast_media_type type;

#ifdef GRANDSTREAM_NETWORKS
	enum ast_media_sub_type sub_type;
#endif

	/*!
	 * \brief The position of the stream in the topology
	 */
	unsigned int position;

	/*!
	 * \brief Current formats negotiated on the stream
	 */
	struct ast_format_cap *formats;

	/*!
	 * \brief The current state of the stream
	 */
	enum ast_stream_state state;

	/*!
	 * \brief Stream metadata vector
	 */
	struct ast_variable *metadata;

	/*!
	 * \brief The group that the stream is part of
	 */
	int group;

	/*!
	 * \brief The rtp_codecs used by the stream
	 */
	struct ast_rtp_codecs *rtp_codecs;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for gs-fec */
	unsigned int alone_fec_version:1;
	unsigned int has_rtcp_address:1;
	unsigned int has_nack:1;
	unsigned int has_nack_pli:1;
	unsigned int has_ccm_fir:1;
	unsigned int change_ssrc:1;
	unsigned int held:1;
	char gs_fec_version[MAX_GS_FEC_VERSION_LEN];
	struct ast_variable *passthrough_attr;
	struct ast_variable *passthrough_bandw;
	enum ast_stream_media_direction local_direction;
	enum ast_stream_media_direction remote_direction;
	enum ast_stream_video_resolution video_resolution;
	enum ast_stream_video_resolution previous_video_resolution;
	enum ast_video_decline_reason video_decline_reason;
	struct ast_video_rtx_pt video_rtx_pt;
	char uptime[MAX_UPTIME_LEN];
	enum ast_stream_media_encryption encryption;
	AST_VECTOR(, struct ast_same_codec_info *) h264_info;
	AST_VECTOR(, struct ast_same_codec_info *) rtx_info;
#endif

	/*!
	 * \brief Name for the stream within the context of the channel it is on
	 */
	char name[0];
};

struct ast_stream_topology {
	/*!
	 * \brief A vector of all the streams in this topology
	 */
	AST_VECTOR(, struct ast_stream *) streams;
};

#ifdef GRANDSTREAM_NETWORKS
static void ast_stream_set_sub_type_according_media_type(struct ast_stream *stream, enum ast_media_type type);
#endif

struct ast_stream *ast_stream_alloc(const char *name, enum ast_media_type type)
{
	struct ast_stream *stream;

	stream = ast_calloc(1, sizeof(*stream) + strlen(S_OR(name, "")) + 1);
	if (!stream) {
		return NULL;
	}

	stream->type = type;
	stream->state = AST_STREAM_STATE_INACTIVE;
	stream->group = -1;
	strcpy(stream->name, S_OR(name, "")); /* Safe */

#ifdef GRANDSTREAM_NETWORKS
	AST_VECTOR_INIT(&stream->h264_info, 0);
	AST_VECTOR_INIT(&stream->rtx_info, 0);
#endif

	return stream;
}

#ifdef GRANDSTREAM_NETWORKS
static void ast_stream_clone_same_codec_info(struct ast_stream *new_stream, const struct ast_stream *stream)
{
	unsigned int i = 0;
	struct ast_same_codec_info *original_info = NULL;
	struct ast_same_codec_info *info = NULL;

	AST_VECTOR_INIT(&new_stream->h264_info, 0);
	AST_VECTOR_INIT(&new_stream->rtx_info, 0);

	ast_log(LOG_DEBUG, "Clone from stream:%p to new_stream:%p\n", stream, new_stream);

	for (i = 0; i < AST_VECTOR_SIZE(&stream->h264_info); ++i) {
		original_info = ast_stream_get_same_codec_info(stream, AST_STREAM_SAME_CODEC_TYPE_H264, i);
		if (!original_info) {
			ast_log(LOG_ERROR, "[stream:%p] Can't get H264 codec info\n", stream);
			continue;
		}

		info = ast_calloc(1, sizeof(struct ast_same_codec_info));
		if (!info) {
			ast_log(LOG_ERROR, "[stream:%p] Append H264 codec info, alloc failed\n", stream);
			break;
		}
		info->payload = original_info->payload;
		strncpy(info->fmtp_param, original_info->fmtp_param, sizeof(info->fmtp_param) - 1);

		if (AST_VECTOR_APPEND(&new_stream->h264_info, info)) {
			ast_log(LOG_ERROR, "[stream:%p] Append H264 codec info failed\n", stream);
		}
	}

	for (i = 0; i < AST_VECTOR_SIZE(&stream->rtx_info); i++) {
		original_info = ast_stream_get_same_codec_info(stream, AST_STREAM_SAME_CODEC_TYPE_RTX, i);
		if (!original_info) {
			ast_log(LOG_ERROR, "[stream:%p] Can't get RTX codec info\n", stream);
			continue;
		}

		info = ast_calloc(1, sizeof(struct ast_same_codec_info));
		if (!info) {
			ast_log(LOG_ERROR, "[stream:%p] Append rtx codec info, alloc failed\n", stream);
			break;
		}
		info->payload = original_info->payload;
		strncpy(info->fmtp_param, original_info->fmtp_param, sizeof(info->fmtp_param) - 1);

		if (AST_VECTOR_APPEND(&new_stream->rtx_info, info)) {
			ast_log(LOG_ERROR, "[stream:%p] Append rtx codec info failed\n", stream);
		}
	}

	return;
}

static void ast_stream_same_codec_info_free(struct ast_same_codec_info *info)
{
	if (!info) {
		return;
	}

	ast_free(info);
	info = NULL;

	return;
}

static void ast_stream_topology_init_same_codec_info(struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	unsigned int count = 0;
	struct ast_stream *stream = NULL;

#ifdef GRANDSTREAM_NETWORKS
	if (!topology){
		return;
	}
#endif
	count = ast_stream_topology_get_count(topology);
	for (i = 0; i < count; i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		AST_VECTOR_INIT(&stream->h264_info, 0);
		AST_VECTOR_INIT(&stream->rtx_info, 0);
	}

	return;
}
#endif

struct ast_stream *ast_stream_clone(const struct ast_stream *stream, const char *name)
{
	struct ast_stream *new_stream;
	size_t stream_size;
	const char *stream_name;

	if (!stream) {
		return NULL;
	}

	stream_name = name ?: stream->name;
	stream_size = sizeof(*stream) + strlen(stream_name) + 1;
	new_stream = ast_calloc(1, stream_size);
	if (!new_stream) {
		return NULL;
	}

	memcpy(new_stream, stream, sizeof(*new_stream));
	strcpy(new_stream->name, stream_name); /* Safe */
	new_stream->group = -1;
	if (new_stream->formats) {
		ao2_ref(new_stream->formats, +1);
	}

	new_stream->metadata = ast_stream_get_metadata_list(stream);
#ifdef GRANDSTREAM_NETWORKS
	new_stream->passthrough_attr = ast_stream_get_passthrough_list(stream, AST_PASSTHROUGH_TYPE_ATTR);
	new_stream->passthrough_bandw = ast_stream_get_passthrough_list(stream, AST_PASSTHROUGH_TYPE_BANDW);

	ast_stream_clone_same_codec_info(new_stream, stream);
#endif
	/* rtp_codecs aren't cloned */

	return new_stream;
}

void ast_stream_free(struct ast_stream *stream)
{
	if (!stream) {
		return;
	}

	ast_variables_destroy(stream->metadata);
#ifdef GRANDSTREAM_NETWORKS
	ast_variables_destroy(stream->passthrough_attr);
	stream->passthrough_attr = NULL;
	ast_variables_destroy(stream->passthrough_bandw);
	stream->passthrough_bandw = NULL;

	AST_VECTOR_CALLBACK_VOID(&stream->h264_info, ast_stream_same_codec_info_free);
	AST_VECTOR_FREE(&stream->h264_info);
	AST_VECTOR_CALLBACK_VOID(&stream->rtx_info, ast_stream_same_codec_info_free);
	AST_VECTOR_FREE(&stream->rtx_info);
#endif

	if (stream->rtp_codecs) {
		ast_rtp_codecs_payloads_destroy(stream->rtp_codecs);
	}

	ao2_cleanup(stream->formats);
	ast_free(stream);
}

const char *ast_stream_get_name(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->name;
}

enum ast_media_type ast_stream_get_type(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->type;
}

void ast_stream_set_type(struct ast_stream *stream, enum ast_media_type type)
{
	ast_assert(stream != NULL);

	stream->type = type;
}

struct ast_format_cap *ast_stream_get_formats(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->formats;
}

void ast_stream_set_formats(struct ast_stream *stream, struct ast_format_cap *caps)
{
	ast_assert(stream != NULL);

	ao2_cleanup(stream->formats);
	stream->formats = ao2_bump(caps);
}

enum ast_stream_state ast_stream_get_state(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->state;
}

void ast_stream_set_state(struct ast_stream *stream, enum ast_stream_state state)
{
	ast_assert(stream != NULL);

	stream->state = state;
}

const char *ast_stream_state2str(enum ast_stream_state state)
{
	switch (state) {
	case AST_STREAM_STATE_REMOVED:
		return "removed";
	case AST_STREAM_STATE_SENDRECV:
		return "sendrecv";
	case AST_STREAM_STATE_SENDONLY:
		return "sendonly";
	case AST_STREAM_STATE_RECVONLY:
		return "recvonly";
	case AST_STREAM_STATE_INACTIVE:
		return "inactive";
	default:
		return "<unknown>";
	}
}

enum ast_stream_state ast_stream_str2state(const char *str)
{
	if (!strcmp("sendrecv", str)) {
		return AST_STREAM_STATE_SENDRECV;
	}
	if (!strcmp("sendonly", str)) {
		return AST_STREAM_STATE_SENDONLY;
	}
	if (!strcmp("recvonly", str)) {
		return AST_STREAM_STATE_RECVONLY;
	}
	if (!strcmp("inactive", str)) {
		return AST_STREAM_STATE_INACTIVE;
	}
	return AST_STREAM_STATE_REMOVED;
}

const char *ast_stream_get_metadata(const struct ast_stream *stream, const char *m_key)
{
	struct ast_variable *v;

	ast_assert_return(stream != NULL, NULL);
	ast_assert_return(m_key != NULL, NULL);

	for (v = stream->metadata; v; v = v->next) {
		if (strcmp(v->name, m_key) == 0) {
			return v->value;
		}
	}

	return NULL;
}

struct ast_variable *ast_stream_get_metadata_list(const struct ast_stream *stream)
{
	struct ast_variable *v;
	struct ast_variable *vout = NULL;

	ast_assert_return(stream != NULL, NULL);

	for (v = stream->metadata; v; v = v->next) {
		struct ast_variable *vt = ast_variable_new(v->name, v->value, "");

		if (!vt) {
			ast_variables_destroy(vout);
			return NULL;
		}

		ast_variable_list_append(&vout, vt);
	}

	return vout;
}

int ast_stream_set_metadata(struct ast_stream *stream, const char *m_key, const char *value)
{
	struct ast_variable *v;
	struct ast_variable *prev;

	ast_assert_return(stream != NULL, -1);
	ast_assert_return(m_key != NULL, -1);

	prev = NULL;
	v = stream->metadata;
	while(v) {
		struct ast_variable *next = v->next;
		if (strcmp(v->name, m_key) == 0) {
			if (prev) {
				prev->next = next;
			} else {
				stream->metadata = next;
			}
			ast_free(v);
			break;
		} else {
			prev = v;
		}
		v = next;
	}

	if (!value) {
		return 0;
	}

	v = ast_variable_new(m_key, value, "");
	if (!v) {
		return -1;
	}

	ast_variable_list_append(&stream->metadata, v);

	return 0;
}

int ast_stream_get_position(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->position;
}

struct ast_rtp_codecs *ast_stream_get_rtp_codecs(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->rtp_codecs;
}

void ast_stream_set_rtp_codecs(struct ast_stream *stream, struct ast_rtp_codecs *rtp_codecs)
{
	ast_assert(stream != NULL);

	if (stream->rtp_codecs) {
		ast_rtp_codecs_payloads_destroy(rtp_codecs);
	}

	stream->rtp_codecs = rtp_codecs;
}

static void stream_topology_destroy(void *data)
{
	struct ast_stream_topology *topology = data;

	AST_VECTOR_CALLBACK_VOID(&topology->streams, ast_stream_free);
	AST_VECTOR_FREE(&topology->streams);
}

#define TOPOLOGY_INITIAL_STREAM_COUNT 2
struct ast_stream_topology *ast_stream_topology_alloc(void)
{
	struct ast_stream_topology *topology;

	topology = ao2_alloc_options(sizeof(*topology), stream_topology_destroy, AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!topology) {
		return NULL;
	}

	if (AST_VECTOR_INIT(&topology->streams, TOPOLOGY_INITIAL_STREAM_COUNT)) {
		ao2_ref(topology, -1);
		topology = NULL;
#ifdef GRANDSTREAM_NETWORKS
		return NULL;	/*Add by weidu, for asterisk coredump*/
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_stream_topology_init_same_codec_info(topology);
#endif

	return topology;
}

struct ast_stream_topology *ast_stream_topology_clone(
	const struct ast_stream_topology *topology)
{
	struct ast_stream_topology *new_topology;
	int i;

	ast_assert(topology != NULL);

	new_topology = ast_stream_topology_alloc();
	if (!new_topology) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&topology->streams); i++) {
		struct ast_stream *existing = AST_VECTOR_GET(&topology->streams, i);
		struct ast_stream *stream = ast_stream_clone(existing, NULL);

		if (!stream || AST_VECTOR_APPEND(&new_topology->streams, stream)) {
			ast_stream_free(stream);
			ast_stream_topology_free(new_topology);
			return NULL;
		}

		ast_stream_set_group(stream, ast_stream_get_group(existing));
	}

	return new_topology;
}

int ast_stream_topology_equal(const struct ast_stream_topology *left,
	const struct ast_stream_topology *right)
{
	int index;

	ast_assert(left != NULL);
	ast_assert(right != NULL);

	if (ast_stream_topology_get_count(left) != ast_stream_topology_get_count(right)) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "ast_stream_topology_equal count is not equal, left:%d, right:%d\n",
			ast_stream_topology_get_count(left), ast_stream_topology_get_count(right));
#endif
		return 0;
	}

	for (index = 0; index < ast_stream_topology_get_count(left); ++index) {
		const struct ast_stream *left_stream = ast_stream_topology_get_stream(left, index);
		const struct ast_stream *right_stream = ast_stream_topology_get_stream(right, index);

		if (ast_stream_get_type(left_stream) != ast_stream_get_type(right_stream)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "ast_stream_topology_equal type is not equal, left:%d, right:%d, index:%d\n",
				ast_stream_get_type(left_stream), ast_stream_get_type(right_stream), index);
#endif
			return 0;
		}

		if (ast_stream_get_state(left_stream) != ast_stream_get_state(right_stream)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "ast_stream_topology_equal state is not equal, left:%d, right:%d, index:%d, left_stream:%p, right_stream:%p\n",
				ast_stream_get_state(left_stream), ast_stream_get_state(right_stream), index, left_stream, right_stream);
#endif
			return 0;
		}

		if (!ast_stream_get_formats(left_stream) && ast_stream_get_formats(right_stream) &&
			ast_format_cap_count(ast_stream_get_formats(right_stream))) {
			/* A NULL format capabilities and an empty format capabilities are the same, as they have
			 * no formats inside. If one does though... they are not equal.
			 */
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "ast_stream_topology_equal left_stream is NULL\n");
#endif
			return 0;
		} else if (!ast_stream_get_formats(right_stream) && ast_stream_get_formats(left_stream) &&
			ast_format_cap_count(ast_stream_get_formats(left_stream))) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "ast_stream_topology_equal right_stream is NULL\n");
#endif
			return 0;
		}
#ifdef GRANDSTREAM_NETWORKS
		else if (ast_stream_get_formats(left_stream) && ast_stream_get_formats(right_stream) &&
			!ast_format_cap_identical(ast_stream_get_formats(left_stream), ast_stream_get_formats(right_stream)) &&
			ast_stream_get_type(left_stream) != AST_MEDIA_TYPE_AUDIO) {
			/* But if both are actually present we need to do an actual identical check. */
			struct ast_str *left_buf = ast_str_alloca(AST_FORMAT_CAP_NAMES_LEN);
			struct ast_str *right_buf = ast_str_alloca(AST_FORMAT_CAP_NAMES_LEN);
			ast_log(LOG_DEBUG, "ast_stream_topology_equal formats is not equal, left_stream:%s, right_stream:%s\n",
				ast_format_cap_get_names(ast_stream_get_formats(left_stream), &left_buf),
				ast_format_cap_get_names(ast_stream_get_formats(right_stream), &right_buf));
			return 0;
		}
#else
		else if (ast_stream_get_formats(left_stream) && ast_stream_get_formats(right_stream) &&
			!ast_format_cap_identical(ast_stream_get_formats(left_stream), ast_stream_get_formats(right_stream))) {
			/* But if both are actually present we need to do an actual identical check. */
			return 0;
		}
#endif

		if (strcmp(ast_stream_get_name(left_stream), ast_stream_get_name(right_stream))) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_DEBUG, "ast_stream_topology_equal name is not equal, left:%s, right:%s, index:%d\n",
				ast_stream_get_name(left_stream), ast_stream_get_name(right_stream), index);
#endif
			return 0;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (ast_stream_get_change_ssrc(left_stream)) {
			ast_log(LOG_DEBUG, "ast_stream_topology_equal left_stream:%s is changed ssrc\n", ast_stream_get_name(left_stream));
			return 0;
		}

		if (ast_stream_get_change_ssrc(right_stream)) {
			ast_log(LOG_DEBUG, "ast_stream_topology_equal right_stream:%s is changed ssrc\n", ast_stream_get_name(right_stream));
			return 0;
		}
#endif
	}

	return 1;
}

void ast_stream_topology_free(struct ast_stream_topology *topology)
{
	ao2_cleanup(topology);
}

int ast_stream_topology_append_stream(struct ast_stream_topology *topology, struct ast_stream *stream)
{
#ifdef GRANDSTREAM_NETWORKS
	if (!topology || !stream){    /*Mod by weidu ,for Bug 266919*/
		return -1;
	}
#else
	ast_assert(topology && stream);
#endif
	if (AST_VECTOR_APPEND(&topology->streams, stream)) {
		return -1;
	}

	stream->position = AST_VECTOR_SIZE(&topology->streams) - 1;

	return AST_VECTOR_SIZE(&topology->streams) - 1;
}

int ast_stream_topology_get_count(const struct ast_stream_topology *topology)
{
#ifdef GRANDSTREAM_NETWORKS   /*Mod by weidu ,for Bug 266919*/
	if (!topology){
		return 0;
	}
#else
	ast_assert(topology != NULL);
#endif
	return AST_VECTOR_SIZE(&topology->streams);
}

struct ast_stream *ast_stream_topology_get_stream(
	const struct ast_stream_topology *topology, unsigned int stream_num)
{
#ifdef GRANDSTREAM_NETWORKS    /*Mod by weidu ,for Bug 266919*/
	if (!topology){
		return NULL;
	}
#else
	ast_assert(topology != NULL);
#endif

	return AST_VECTOR_GET(&topology->streams, stream_num);
}

int ast_stream_topology_set_stream(struct ast_stream_topology *topology,
	unsigned int position, struct ast_stream *stream)
{
	struct ast_stream *existing_stream;

	ast_assert(topology && stream);

	if (position > AST_VECTOR_SIZE(&topology->streams)) {
		return -1;
	}

	if (position < AST_VECTOR_SIZE(&topology->streams)) {
		existing_stream = AST_VECTOR_GET(&topology->streams, position);
		ast_stream_free(existing_stream);
	}

	stream->position = position;

	if (position == AST_VECTOR_SIZE(&topology->streams)) {
		return AST_VECTOR_APPEND(&topology->streams, stream);
	}

	return AST_VECTOR_REPLACE(&topology->streams, position, stream);
}

int ast_stream_topology_del_stream(struct ast_stream_topology *topology,
	unsigned int position)
{
	struct ast_stream *stream;

	ast_assert(topology != NULL);

	if (AST_VECTOR_SIZE(&topology->streams) <= position) {
		return -1;
	}

	stream = AST_VECTOR_REMOVE_ORDERED(&topology->streams, position);
	ast_stream_free(stream);

	/* Fix up higher stream position indices */
	for (; position < AST_VECTOR_SIZE(&topology->streams); ++position) {
		stream = AST_VECTOR_GET(&topology->streams, position);
		stream->position = position;
	}

	return 0;
}

struct ast_stream_topology *ast_stream_topology_create_from_format_cap(
	struct ast_format_cap *cap)
{
	struct ast_stream_topology *topology;
	enum ast_media_type type;

	topology = ast_stream_topology_alloc();
	if (!topology || !cap || !ast_format_cap_count(cap)) {
		return topology;
	}

	for (type = AST_MEDIA_TYPE_UNKNOWN + 1; type < AST_MEDIA_TYPE_END; type++) {
		struct ast_format_cap *new_cap;
		struct ast_stream *stream;

		if (!ast_format_cap_has_type(cap, type)) {
			continue;
		}

		new_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!new_cap) {
			ast_stream_topology_free(topology);
			return NULL;
		}

		ast_format_cap_set_framing(new_cap, ast_format_cap_get_framing(cap));
		if (ast_format_cap_append_from_cap(new_cap, cap, type)) {
			ao2_cleanup(new_cap);
			ast_stream_topology_free(topology);
			return NULL;
		}

		stream = ast_stream_alloc(ast_codec_media_type2str(type), type);
		if (!stream) {
			ao2_cleanup(new_cap);
			ast_stream_topology_free(topology);
			return NULL;
		}
		/* We're transferring the initial ref so no bump needed */
		stream->formats = new_cap;
		stream->state = AST_STREAM_STATE_SENDRECV;

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for set sub_type, 2019.12.26 */
		ast_stream_set_sub_type_according_media_type(stream, type);
		ast_stream_set_local_media_direction(stream, SENDRECV);
#endif

		if (ast_stream_topology_append_stream(topology, stream) == -1) {
			ast_stream_free(stream);
			ast_stream_topology_free(topology);
			return NULL;
		}
	}

	return topology;
}

struct ast_format_cap *ast_format_cap_from_stream_topology(
    struct ast_stream_topology *topology)
{
	struct ast_format_cap *caps;
	int i;

	ast_assert(topology != NULL);

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&topology->streams); i++) {
		struct ast_stream *stream;

		stream = AST_VECTOR_GET(&topology->streams, i);
		if (!stream->formats
			|| stream->state == AST_STREAM_STATE_REMOVED) {
			continue;
		}

		ast_format_cap_append_from_cap(caps, stream->formats, AST_MEDIA_TYPE_UNKNOWN);
	}

	return caps;
}

#ifdef GRANDSTREAM_NETWORKS
struct ast_format_cap *ast_format_cap_from_stream_topology_for_iax(
    struct ast_stream_topology *topology)
{
	struct ast_format_cap *caps;
	int i;

	ast_assert(topology != NULL);

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&topology->streams); i++) {
		struct ast_stream *stream;

		stream = AST_VECTOR_GET(&topology->streams, i);
		ast_log(LOG_NOTICE, "stream[%s], state[%d], type[%d], subtype[%d], local_dir[%d], rem_dir[%d]", 
				stream->name, stream->state, stream->type, stream->sub_type, stream->local_direction, stream->remote_direction);
		if (!stream->formats
			|| stream->state == AST_STREAM_STATE_REMOVED
			|| stream->local_direction == INACTIVE
			|| stream->sub_type >= AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			continue;
		}

		ast_format_cap_append_from_cap(caps, stream->formats, AST_MEDIA_TYPE_UNKNOWN);
	}

	return caps;
}
#endif

struct ast_stream *ast_stream_topology_get_first_stream_by_type(
	const struct ast_stream_topology *topology,
	enum ast_media_type type)
{
	int i;

	ast_assert(topology != NULL);

	for (i = 0; i < AST_VECTOR_SIZE(&topology->streams); i++) {
		struct ast_stream *stream;

		stream = AST_VECTOR_GET(&topology->streams, i);
		if (stream->type == type
			&& stream->state != AST_STREAM_STATE_REMOVED) {
			return stream;
		}
	}

	return NULL;
}

void ast_stream_topology_map(const struct ast_stream_topology *topology,
	struct ast_vector_int *types, struct ast_vector_int *v0, struct ast_vector_int *v1)
{
	int i;
	int nths[AST_MEDIA_TYPE_END] = {0};
	int size = ast_stream_topology_get_count(topology);

	/*
	 * Clear out any old mappings and initialize the new ones
	 */
	AST_VECTOR_FREE(v0);
	AST_VECTOR_FREE(v1);

	/*
	 * Both vectors are sized to the topology. The media types vector is always
	 * guaranteed to be the size of the given topology or greater.
	 */
	AST_VECTOR_INIT(v0, size);
	AST_VECTOR_INIT(v1, size);

	for (i = 0; i < size; ++i) {
		struct ast_stream *stream = ast_stream_topology_get_stream(topology, i);
		enum ast_media_type type = ast_stream_get_type(stream);
		int index = AST_VECTOR_GET_INDEX_NTH(types, ++nths[type],
			type, AST_VECTOR_ELEM_DEFAULT_CMP);

		if (index == -1) {
			/*
			 * If a given type is not found for an index level then update the
			 * media types vector with that type. This keeps the media types
			 * vector always at the max topology size.
			 */
			AST_VECTOR_APPEND(types, type);
			index = AST_VECTOR_SIZE(types) - 1;
		}

		/*
		 * The mapping is reflexive in the sense that if it maps in one direction
		 * then the reverse direction maps back to the other's index.
		 */
		AST_VECTOR_REPLACE(v0, i, index);
		AST_VECTOR_REPLACE(v1, index, i);
	}
}

int ast_stream_get_group(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->group;
}

void ast_stream_set_group(struct ast_stream *stream, int group)
{
	ast_assert(stream != NULL);

	stream->group = group;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC */
enum ast_media_sub_type ast_stream_get_sub_type(const struct ast_stream *stream)
{
	ast_assert(stream != NULL);

	return stream->sub_type;
}

void ast_stream_set_sub_type(struct ast_stream *stream, enum ast_media_sub_type sub_type)
{
	ast_assert(stream != NULL);

	stream->sub_type = sub_type;
}

const char *ast_stream_get_gs_fec_version(const struct ast_stream *stream)
{
	if (!stream) {
		return NULL;
	}

	return stream->gs_fec_version;
}

void ast_stream_set_gs_fec_version(struct ast_stream *stream, const char *gs_fec_version)
{
	if (!stream) {
		return;
	}

	if (!gs_fec_version || ast_strlen_zero(gs_fec_version)) {
		memset(stream->gs_fec_version, 0, sizeof(stream->gs_fec_version));
		return;
	}

	strncpy(stream->gs_fec_version, gs_fec_version, sizeof(stream->gs_fec_version) - 1);

	return;
}

unsigned int ast_stream_get_alone_fec_version(const struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	return stream->alone_fec_version;
}

void ast_stream_set_alone_fec_version(struct ast_stream *stream, unsigned int value)
{
	if (!stream) {
		return;
	}

	stream->alone_fec_version = value;

	return;
}

unsigned int ast_stream_get_has_rtcp_address(const struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	return stream->has_rtcp_address;
}

void ast_stream_set_has_rtcp_address(struct ast_stream *stream, unsigned int has_rtcp_address)
{
	if (!stream) {
		return;
	}

	stream->has_rtcp_address = has_rtcp_address;

	return;
}

unsigned int ast_stream_get_has_nack(const struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	return stream->has_nack;
}

void ast_stream_set_has_nack(struct ast_stream *stream, unsigned int has_nack)
{
	if (!stream) {
		return;
	}

	stream->has_nack = has_nack;

	return;
}

unsigned int ast_stream_get_has_nack_pli(const struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	return stream->has_nack_pli;
}

void ast_stream_set_has_nack_pli(struct ast_stream *stream, unsigned int has_nack_pli)
{
	if (!stream) {
		return;
	}

	stream->has_nack_pli = has_nack_pli;

	return;
}

unsigned int ast_stream_get_has_ccm_fir(const struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	return stream->has_ccm_fir;
}

void ast_stream_set_has_ccm_fir(struct ast_stream *stream, unsigned int has_ccm_fir)
{
	if (!stream) {
		return;
	}

	stream->has_ccm_fir = has_ccm_fir;

	return;
}

struct ast_variable *ast_stream_get_passthrough_list(const struct ast_stream *stream, enum ast_passthrough_type passthrough_type)
{
	struct ast_variable *v = NULL;
	struct ast_variable *vout = NULL;
	struct ast_variable *passthrough_list = NULL;

	ast_assert_return(stream != NULL, NULL);

	switch (passthrough_type) {
		case AST_PASSTHROUGH_TYPE_ATTR:
			passthrough_list = stream->passthrough_attr;
			break;
		case AST_PASSTHROUGH_TYPE_BANDW:
			passthrough_list = stream->passthrough_bandw;
			break;
		default:
			ast_log(LOG_WARNING, "We can't support other passthrough type\n");
			return NULL;
	}

	for (v = passthrough_list; v; v = v->next) {
		struct ast_variable *vt = ast_variable_new(v->name, v->value, "");

		if (!vt) {
			ast_variables_destroy(vout);
			return NULL;
		}

		ast_variable_list_append(&vout, vt);
	}

	return vout;
}

int ast_stream_set_passthrough(struct ast_stream *stream, const char *m_key, const char *value, enum ast_passthrough_type passthrough_type)
{
	struct ast_variable *v = NULL;

	ast_assert_return(stream != NULL, -1);
	ast_assert_return(m_key != NULL, -1);

	if (!value) {
		return 0;
	}

	v = ast_variable_new(m_key, value, "");
	if (!v) {
		return -1;
	}

	switch (passthrough_type) {
		case AST_PASSTHROUGH_TYPE_ATTR:
			ast_variable_list_append(&stream->passthrough_attr, v);
			break;
		case AST_PASSTHROUGH_TYPE_BANDW:
			ast_variable_list_append(&stream->passthrough_bandw, v);
			break;
		default:
			ast_log(LOG_WARNING, "We can't support other passthrough type\n");
	}

	return 0;
}

void ast_stream_clear_passthrough(struct ast_stream *stream)
{
	if (!stream) {
		return;
	}

	if (stream->passthrough_attr) {
		ast_variables_destroy(stream->passthrough_attr);
		stream->passthrough_attr = NULL;
	}
	if (stream->passthrough_bandw) {
		ast_variables_destroy(stream->passthrough_bandw);
		stream->passthrough_bandw = NULL;
	}

	return;
}

void ast_stream_clear_bandw_passthrough(struct ast_stream *stream)
{
	if (!stream) {
		return;
	}

	if (stream->passthrough_bandw) {
		ast_variables_destroy(stream->passthrough_bandw);
		stream->passthrough_bandw = NULL;
	}

	return;
}

void ast_stream_set_local_media_direction(struct ast_stream *stream, enum ast_stream_media_direction media_direction)
{
	if (!stream) {
		return;
	}

	ast_log(LOG_DEBUG, "ast_stream_set_local_media_direction:%d, stream:%p\n", media_direction, stream);

	stream->local_direction = media_direction;
}

enum ast_stream_media_direction ast_stream_get_local_media_direction(struct ast_stream *stream)
{
	if (!stream) {
		return INACTIVE;
	}

	return stream->local_direction;
}

void ast_stream_set_remote_media_direction(struct ast_stream *stream, enum ast_stream_media_direction media_direction)
{
	if (!stream) {
		return;
	}

	ast_log(LOG_DEBUG, "ast_stream_set_remote_media_direction:%d, stream:%p\n", media_direction, stream);

	stream->remote_direction = media_direction;
}

enum ast_stream_media_direction ast_stream_get_remote_media_direction(struct ast_stream *stream)
{
	if (!stream) {
		return INACTIVE;
	}

	return stream->remote_direction;
}

static enum ast_media_type ast_stream_sub_type_to_type(enum ast_media_sub_type sub_type)
{
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;

	switch (sub_type) {
		case AST_MEDIA_SUB_AUDIO_TYPE:
			type = AST_MEDIA_TYPE_AUDIO;
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			type = AST_MEDIA_TYPE_VIDEO;
			break;
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP:
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP:
		case AST_MEDIA_SUB_APPLICATION_TYPE_FEC:
		case AST_MEDIA_SUB_APPLICATION_TYPE_FECC:
			type = AST_MEDIA_TYPE_APPLICATION;
			break;
		default:
			break;
	}

	return type;
}

struct ast_stream_topology *ast_stream_topology_webrtc_create_from_format_cap_for_sfu(
	struct ast_format_cap *cap, enum ast_sfu_invite_type invite_type)
{
	struct ast_stream_topology *topology = NULL;
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;
	struct ast_format_cap *new_cap = NULL;
	struct ast_stream *stream = NULL;

	topology = ast_stream_topology_alloc();
	if (!topology || !cap || !ast_format_cap_count(cap)) {
		return topology;
	}

	for (sub_type = AST_MEDIA_SUB_AUDIO_TYPE; sub_type <= AST_MEDIA_SUB_VIDEO_TYPE_SLIDES; sub_type++) {
		type = ast_stream_sub_type_to_type(sub_type);

		if (!ast_format_cap_has_type(cap, type)) {
			continue;
		}

		if ((invite_type == AST_SFU_INVITE_TYPE_AUDIO) && (type != AST_MEDIA_TYPE_AUDIO)) {
			continue;
		}

		new_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!new_cap) {
			ast_stream_topology_free(topology);
			return NULL;
		}

		ast_format_cap_set_framing(new_cap, ast_format_cap_get_framing(cap));
		if (ast_format_cap_append_from_cap(new_cap, cap, type)) {
			ao2_cleanup(new_cap);
			ast_stream_topology_free(topology);
			return NULL;
		}

		if (sub_type == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			ast_format_cap_remove_codecs_except_conf_video_codec(new_cap, "H264");
			if (!ast_format_cap_count(new_cap)) {
				ast_log(LOG_DEBUG, "Has no H264 codec in enpoint\n");
				ao2_cleanup(new_cap);
				continue;
			} 
		}

		if (type == AST_MEDIA_TYPE_AUDIO) {
			if (ast_format_cap_has_red(new_cap)) {
				ast_format_cap_remove_red(new_cap);
			}
			if (ast_format_cap_has_audio_fec(new_cap)) {
				ast_format_cap_remove_audio_fec(new_cap);
			}
		}

		stream = ast_stream_alloc(ast_codec_media_type2str(type), type);
		if (!stream) {
			ao2_cleanup(new_cap);
			ast_stream_topology_free(topology);
			return NULL;
		}
		/* We're transferring the initial ref so no bump needed */
		stream->formats = new_cap;
		stream->state = AST_STREAM_STATE_SENDRECV;
		switch (sub_type) {
			case AST_MEDIA_SUB_AUDIO_TYPE:
				ast_stream_set_sub_type(stream, AST_MEDIA_SUB_AUDIO_TYPE);
				break;
			case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
				ast_stream_set_metadata(stream, "CONTENT", "main");
				ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
				ast_stream_set_local_media_direction(stream, RECVONLY);
				break;
			case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
				ast_stream_set_metadata(stream, "CONTENT", "slides");
				ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
				break;
			default:
				break;
		}

		if (sub_type != AST_MEDIA_SUB_AUDIO_TYPE) {
			ast_stream_set_has_nack(stream, 1);
			ast_stream_set_has_nack_pli(stream, 1);
			ast_stream_set_has_ccm_fir(stream, 1);
		}

		if (ast_stream_topology_append_stream(topology, stream) == -1) {
			ast_stream_free(stream);
			ast_stream_topology_free(topology);
			return NULL;
		}
	}

	return topology;
}

static int ast_stream_topology_create_and_append_bfcp_stream(struct ast_stream_topology *topology,
	enum ast_media_sub_type sub_type)
{
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	struct ast_stream *stream = NULL;

	type = ast_stream_sub_type_to_type(sub_type);
	stream = ast_stream_alloc(ast_codec_media_type2str(type), type);
	if (!stream) {
		return -1;
	}

	ast_stream_set_sub_type(stream, sub_type);

	if (ast_stream_topology_append_stream(topology, stream) < 0) {
		ast_stream_free(stream);
		return -1;
	}

	return 0;
}

static int judge_if_has_valid_video_codec(struct ast_format_cap *caps)
{

	RAII_VAR(struct ast_format_cap *, video_caps, NULL, ao2_cleanup);

	if (!caps) {
		return 0;
	}

	if (!(video_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate capabilities for handle no pending media\n");
		return -1;
	}

	ast_format_cap_append_from_cap(video_caps, caps, AST_MEDIA_TYPE_VIDEO);

	if (ast_format_cap_has_rtx(video_caps) && ast_format_cap_count(video_caps) == 1) {
		ast_log(LOG_DEBUG, "Only has RTX codec, don't need add\n");
		return 0;
	}

	if (ast_format_cap_has_video_fec(video_caps) && ast_format_cap_count(video_caps) == 1) {
		ast_log(LOG_DEBUG, "Only has video FEC codec, don't need add\n");
		return 0;
	}

	if (ast_format_cap_has_rtx(video_caps) && ast_format_cap_has_video_fec(video_caps) && ast_format_cap_count(video_caps) == 2) {
		ast_log(LOG_DEBUG, "Only has RTX and video FEC codec, don't need add\n");
		return 0;
	}

	return 1;
}

struct ast_stream_topology *ast_stream_topology_sip_create_from_format_cap_for_sfu(
	struct ast_format_cap *cap, enum ast_sfu_invite_type invite_type, unsigned int need_create_bfcp)
{
	struct ast_stream_topology *topology = NULL;
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;
	struct ast_format_cap *new_cap = NULL;
	struct ast_stream *stream = NULL;

	if (!cap || !ast_format_cap_count(cap)) {
		return NULL;
	}

	topology = ast_stream_topology_alloc();
	if (!topology) {
		return NULL;
	}

	for (sub_type = AST_MEDIA_SUB_AUDIO_TYPE; sub_type <= AST_MEDIA_SUB_VIDEO_TYPE_SLIDES; sub_type++) {
		type = ast_stream_sub_type_to_type(sub_type);

		if (!ast_format_cap_has_type(cap, type)) {
			continue;
		}

		if ((invite_type == AST_SFU_INVITE_TYPE_AUDIO) && (type != AST_MEDIA_TYPE_AUDIO)) {
			ast_log(LOG_DEBUG, "We don't add other media except audio for audio invite\n");
			continue;
		}

		if (type == AST_MEDIA_TYPE_VIDEO && !judge_if_has_valid_video_codec(cap)) {
			ast_log(LOG_DEBUG, "Has no valid video codec, don't need add\n");
			continue;
		}

		new_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!new_cap) {
			ast_stream_topology_free(topology);
			return NULL;
		}

		ast_format_cap_set_framing(new_cap, ast_format_cap_get_framing(cap));
		if (ast_format_cap_append_from_cap(new_cap, cap, type)) {
			ao2_cleanup(new_cap);
			ast_stream_topology_free(topology);
			return NULL;
		}

		/* For SFU slides video, we just support H264 codec */
		if (sub_type == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			ast_format_cap_remove_codecs_except_conf_video_codec(new_cap, "H264");
			if (!ast_format_cap_count(new_cap)) {
				ast_log(LOG_DEBUG, "Has no H264 codec in enpoint\n");
				ao2_cleanup(new_cap);
				continue;
			} 
		}

		stream = ast_stream_alloc(ast_codec_media_type2str(type), type);
		if (!stream) {
			ao2_cleanup(new_cap);
			ast_stream_topology_free(topology);
			return NULL;
		}
		/* We're transferring the initial ref so no bump needed */
		stream->formats = new_cap;
		stream->state = AST_STREAM_STATE_SENDRECV;
		ast_stream_set_local_media_direction(stream, SENDRECV);
		if (ast_format_cap_has_video_fec(cap)) {
			ast_stream_set_alone_fec_version(stream, 1);
		}
		switch (sub_type) {
			case AST_MEDIA_SUB_AUDIO_TYPE:
				ast_stream_set_sub_type(stream, AST_MEDIA_SUB_AUDIO_TYPE);
				break;
			case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
				ast_stream_set_metadata(stream, "CONTENT", "main");
				ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
				break;
			case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
				ast_stream_set_metadata(stream, "CONTENT", "slides");
				ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
				break;
			default:
				break;
		}

		if (sub_type != AST_MEDIA_SUB_AUDIO_TYPE) {
			ast_stream_set_has_nack(stream, 1);
			ast_stream_set_has_nack_pli(stream, 1);
			ast_stream_set_has_ccm_fir(stream, 1);
		}

		if (sub_type == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES && need_create_bfcp) {
			/* We append bfcp udp for default */
			if (ast_stream_topology_create_and_append_bfcp_stream(topology, AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP)) {
				ast_stream_free(stream);
				ast_stream_topology_free(topology);
				return NULL;
			}
		}

		if (ast_stream_topology_append_stream(topology, stream) == -1) {
			ast_stream_free(stream);
			ast_stream_topology_free(topology);
			return NULL;
		}
	}

	return topology;
}

void ast_stream_set_video_resolution(struct ast_stream *stream, enum ast_stream_video_resolution video_resolution)
{
	if (!stream) {
		return;
	}

	stream->video_resolution = video_resolution;

	ast_log(LOG_DEBUG, "ast_stream_set_video_resolution video_resolution:%d\n", video_resolution);

	return;
}

enum ast_stream_video_resolution ast_stream_get_video_resolution(struct ast_stream *stream)
{
	if (!stream) {
		return AST_STREAM_VIDEO_RESOLUTION_NONE;
	}

	return stream->video_resolution;
}

void ast_stream_set_previous_video_resolution(struct ast_stream *stream, enum ast_stream_video_resolution video_resolution)
{
	if (!stream) {
		return;
	}

	stream->previous_video_resolution = video_resolution;

	ast_log(LOG_DEBUG, "ast_stream_set_video_resolution video_resolution:%d\n", video_resolution);

	return;
}

enum ast_stream_video_resolution ast_stream_get_previous_video_resolution(struct ast_stream *stream)
{
	if (!stream) {
		return AST_STREAM_VIDEO_RESOLUTION_NONE;
	}

	return stream->previous_video_resolution;
}

unsigned char ast_stream_judge_if_main_video(struct ast_stream *stream)
{
	if (!stream) {
		ast_log(LOG_DEBUG, "judge_if_main_video_stream stream is NULL\n");
		return 0;
	}

	if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
		ast_log(LOG_DEBUG, "judge_if_main_video_stream stream is not main type\n");
		return 0;
	}

	if (strncmp(ast_stream_get_name(stream), SOFTBRIDGE_VIDEO_DEST_PREFIX, SOFTBRIDGE_VIDEO_DEST_LEN)) {
		ast_log(LOG_DEBUG, "judge_if_main_video_stream stream stream is not append video\n");
		return 1;
	}

	ast_log(LOG_DEBUG, "judge_if_main_video_stream stream is append video");

	return 0;
}

int ast_stream_judge_if_audio(struct ast_stream *stream)
{
	if (!stream) {
		ast_log(LOG_DEBUG, "ast_stream_judge_if_audio stream is NULL\n");
		return 0;
	}

	if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_AUDIO_TYPE) {
		ast_log(LOG_DEBUG, "ast_stream_judge_if_audio stream is not audio type\n");
		return 0;
	}

	ast_log(LOG_DEBUG, "ast_stream_judge_if_audio stream is audo");

	return 1;
}

unsigned int ast_stream_judge_if_same_type(struct ast_stream *stream, enum ast_media_sub_type type)
{
	unsigned int res = 0;
	enum ast_media_sub_type stream_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!stream) {
		ast_log(LOG_DEBUG, "ast_stream_judge_if_same_type stream is NULL\n");
		return 0;
	}

	stream_type = ast_stream_get_sub_type(stream);

	switch (type) {
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			if (ast_stream_judge_if_main_video(stream)) {
				res = 1;
			}
			break;
		default:
			if (stream_type == type) {
				res = 1;
			}
			break;
	}

	return res;
}


void ast_stream_set_change_ssrc(struct ast_stream *stream, unsigned int value)
{
	if (!stream) {
		return;
	}

	stream->change_ssrc = value;

	ast_log(LOG_DEBUG, "ast_stream_set_change_ssrc value:%d\n", value);

	return;
}

unsigned int ast_stream_get_change_ssrc(const struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	ast_log(LOG_DEBUG, "ast_stream_get_change_ssrc value:%d\n", stream->change_ssrc);

	return stream->change_ssrc;
}

void ast_stream_set_held(struct ast_stream *stream, unsigned int value)
{
	if (!stream) {
		return;
	}

	stream->held = value;

	ast_log(LOG_DEBUG, "ast_stream_set_held value:%d\n", value);

	return;
}

unsigned int ast_stream_get_held(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	ast_log(LOG_DEBUG, "ast_stream_get_held value:%d\n", stream->held);

	return stream->held;
}

static void ast_stream_set_sub_type_according_media_type(struct ast_stream *stream, enum ast_media_type type)
{
	if (!stream) {
		ast_log(LOG_WARNING, "Can't get stream, don't set sub_type\n");
		return;
	}

	switch (type) {
		case AST_MEDIA_TYPE_AUDIO:
			ast_stream_set_sub_type(stream, AST_MEDIA_SUB_AUDIO_TYPE);
			break;
		case AST_MEDIA_TYPE_VIDEO:
			ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
			break;
		default:
			break;
	}

	return;
}

void ast_stream_set_video_decline_reason(struct ast_stream *stream, enum ast_video_decline_reason reason)
{
	if (!stream) {
		return;
	}

	ast_log(LOG_DEBUG, "ast_stream_set_video_decline_reason value:%d\n", reason);

	stream->video_decline_reason = reason;

	return;
}

enum ast_video_decline_reason ast_stream_get_video_decline_reason(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	ast_log(LOG_DEBUG, "ast_stream_get_video_decline_reason value:%d\n", stream->video_decline_reason);

	return stream->video_decline_reason;
}

unsigned int ast_stream_has_video_resource(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_SENDONLY ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_INACTIVE) {
		return 0;
	}

	return 1;
}

unsigned int ast_stream_judge_if_stream_is_participant(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_AUDIO_TYPE ||
		ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES ||
		ast_stream_judge_if_main_video(stream)) {
		return 0;
	}

	return 1;
}

struct ast_stream_topology *ast_stream_topology_clone_participant(const struct ast_stream_topology *topology)
{
	int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream *existing = NULL;
	struct ast_stream_topology *new_topology = NULL;

	if (!topology) {
		return NULL;
	}

	new_topology = ast_stream_topology_alloc();
	if (!new_topology) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&topology->streams); i++) {
		existing = AST_VECTOR_GET(&topology->streams, i);

		if (!ast_stream_judge_if_stream_is_participant(existing)) {
			continue;
		}

		stream = ast_stream_clone(existing, NULL);

		if (!stream || AST_VECTOR_APPEND(&new_topology->streams, stream)) {
			ast_stream_free(stream);
			ast_stream_topology_free(new_topology);
			return NULL;
		}

		ast_stream_set_group(stream, ast_stream_get_group(existing));
	}

	return new_topology;
}

void ast_stream_set_video_rtx_pt(struct ast_stream *stream, const char *name, int pt)
{
	if (!stream || !name) {
		return;
	}

	if (!strcasecmp(name, "H264") && !stream->video_rtx_pt.h264_rtx_pt) {
		stream->video_rtx_pt.h264_rtx_pt = pt;
	} else if (!strcasecmp(name, "H265") && !stream->video_rtx_pt.h265_rtx_pt) {
		stream->video_rtx_pt.h265_rtx_pt = pt;
	} else if (!strcasecmp(name, "VP8") && !stream->video_rtx_pt.vp8_rtx_pt) {
		stream->video_rtx_pt.vp8_rtx_pt = pt;
	} else if (!strcasecmp(name, "VP9") && !stream->video_rtx_pt.vp9_rtx_pt) {
		stream->video_rtx_pt.vp9_rtx_pt = pt;
	} else {
		ast_log(LOG_DEBUG, "Can't support video codec:%s or stream has already setted for RTX\n", name);
	}

	return;
}

int ast_stream_get_video_rtx_pt(struct ast_stream *stream, const char *name)
{
	if (!stream || !name) {
		return 0;
	}

	if (!strcasecmp(name, "H264")) {
		return stream->video_rtx_pt.h264_rtx_pt;
	} else if (!strcasecmp(name, "H265")) {
		return stream->video_rtx_pt.h265_rtx_pt;
	} else if (!strcasecmp(name, "VP8")) {
		return stream->video_rtx_pt.vp8_rtx_pt;
	} else if (!strcasecmp(name, "VP9")) {
		return stream->video_rtx_pt.vp9_rtx_pt;
	} else {
		ast_log(LOG_DEBUG, "Can't support video codec:%s for RTX\n", name);
	}

	return 0;
}


void ast_stream_init_video_rtx_pt(struct ast_stream *stream)
{
	if (!stream) {
		return;
	}

	stream->video_rtx_pt.h264_rtx_pt = 0;
	stream->video_rtx_pt.h265_rtx_pt = 0;
	stream->video_rtx_pt.vp8_rtx_pt = 0;
	stream->video_rtx_pt.vp9_rtx_pt = 0;

	return;
}

// Not support slides video.
struct ast_format_cap* ast_create_format_by_topology(struct ast_stream_topology* topology, int ignore_video)
{
	if (!topology) {
		return NULL;
	}

	int i = 0;
	struct ast_format_cap* caps = NULL;
	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&topology->streams); i++) {
		struct ast_stream* stream = NULL;
		stream = AST_VECTOR_GET(&topology->streams, i);
		if (!stream->formats || stream->state == AST_STREAM_STATE_REMOVED) {
			continue;
		}

		if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO) {
			if (ignore_video) {
				continue;
			}

			if (ast_stream_get_remote_media_direction(stream) == INACTIVE ||
				ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
				continue;
			}
		}

		ast_format_cap_append_from_cap(caps, stream->formats, AST_MEDIA_TYPE_UNKNOWN);
	}

	return caps;
}

void ast_stream_set_uptime(struct ast_stream *stream, char *uptime)
{
	if (!stream) {
		return;
	}

	strncpy(stream->uptime, uptime, sizeof(stream->uptime) - 1);

	return;
}

char *ast_stream_get_uptime(struct ast_stream *stream)
{
	if (!stream) {
		return NULL;
	}

	return stream->uptime;
}

void ast_stream_set_media_encryption(struct ast_stream *stream, enum ast_stream_media_encryption encryption)
{
	if (!stream) {
		return;
	}

	stream->encryption = encryption;

	return;
}

enum ast_stream_media_encryption ast_stream_get_media_encryption(struct ast_stream *stream)
{
	if (!stream) {
		return AST_STREAM_MEDIA_TRANSPORT_INVALID;
	}

	return stream->encryption;
}

static void ast_stream_set_same_h264_codec_info(struct ast_stream *stream, struct ast_same_codec_info codec_info)
{
	struct ast_same_codec_info *info = NULL;

	if (!stream) {
		return;
	}

	info = ast_calloc(1, sizeof(struct ast_same_codec_info));
	if (!info) {
		ast_log(LOG_ERROR, "[stream:%p] Append H264 codec info, alloc failed\n", stream);
		return;
	}

	ast_log(LOG_DEBUG, "[stream:%p] Set H264 codec info, pt:%d, fmtp_param:%s\n", stream, codec_info.payload, codec_info.fmtp_param);

	info->payload = codec_info.payload;
	strncpy(info->fmtp_param, codec_info.fmtp_param, sizeof(info->fmtp_param) - 1);

	if (AST_VECTOR_APPEND(&stream->h264_info, info)) {
		ast_log(LOG_ERROR, "[stream:%p] Append H264 codec info failed\n", stream);
	}

	return;
}

static void ast_stream_set_same_rtx_codec_info(struct ast_stream *stream, struct ast_same_codec_info codec_info)
{
	struct ast_same_codec_info *info = NULL;

	if (!stream) {
		return;
	}

	info = ast_calloc(1, sizeof(struct ast_same_codec_info));
	if (!info) {
		ast_log(LOG_ERROR, "[stream:%p] Append rtx codec info, alloc failed\n", stream);
		return;
	}

	ast_log(LOG_DEBUG, "[stream:%p] Set RTX codec info, pt:%d, fmtp_param:%s\n", stream, codec_info.payload, codec_info.fmtp_param);

	info->payload = codec_info.payload;
	strncpy(info->fmtp_param, codec_info.fmtp_param, sizeof(info->fmtp_param) - 1);

	if (AST_VECTOR_APPEND(&stream->rtx_info, info)) {
		ast_log(LOG_ERROR, "[stream:%p] Append rtx codec info failed\n", stream);
	}

	return;
}

void ast_stream_set_same_codec_info(struct ast_stream *stream,
	enum ast_stream_same_codec_type type, struct ast_same_codec_info codec_info)
{
	if (!stream) {
		return;
	}

	switch (type) {
		case AST_STREAM_SAME_CODEC_TYPE_H264:
			ast_stream_set_same_h264_codec_info(stream, codec_info);
			break;
		case AST_STREAM_SAME_CODEC_TYPE_RTX:
			ast_stream_set_same_rtx_codec_info(stream, codec_info);
			break;
		default:
			break;
	}

	return;
}

struct ast_same_codec_info *ast_stream_get_same_codec_info(const struct ast_stream *stream,
	enum ast_stream_same_codec_type type, unsigned int number)
{
	if (!stream) {
		return NULL;
	}

	switch (type) {
		case AST_STREAM_SAME_CODEC_TYPE_H264:
			return AST_VECTOR_GET(&stream->h264_info, number);
		case AST_STREAM_SAME_CODEC_TYPE_RTX:
			return AST_VECTOR_GET(&stream->rtx_info, number);
		default:
			break;
	}

	return NULL;
}

int ast_stream_same_codec_info_get_count(const struct ast_stream *stream,
	enum ast_stream_same_codec_type type)
{
	if (!stream) {
		return 0;
	}

	switch (type) {
		case AST_STREAM_SAME_CODEC_TYPE_H264:
			return AST_VECTOR_SIZE(&stream->h264_info);
		case AST_STREAM_SAME_CODEC_TYPE_RTX:
			return AST_VECTOR_SIZE(&stream->rtx_info);
		default:
			break;
	}

	return 0;
}

void ast_stream_init_same_codec_info(struct ast_stream *stream)
{
	if (!stream) {
		return;
	}

	if (ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_H264)) {
		AST_VECTOR_CALLBACK_VOID(&stream->h264_info, ast_stream_same_codec_info_free);
		AST_VECTOR_FREE(&stream->h264_info);
	}
	if (ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_RTX)) {
		AST_VECTOR_CALLBACK_VOID(&stream->rtx_info, ast_stream_same_codec_info_free);
		AST_VECTOR_FREE(&stream->rtx_info);
	}

	AST_VECTOR_INIT(&stream->h264_info, 0);
	AST_VECTOR_INIT(&stream->rtx_info, 0);

	return;
}

unsigned int ast_stream_is_enabled_main_video(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (!ast_stream_judge_if_main_video(stream)) {
		return 0;
	}

	if (ast_stream_get_local_media_direction(stream) == SENDRECV ||
		ast_stream_get_local_media_direction(stream) == RECVONLY) {
		return 1;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_INACTIVE ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_SENDONLY) {
		return 0;
	}

	return 1;
}
#endif
