/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthm@freeswitch.org>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Anthony Minessale II <anthm@freeswitch.org>
 * Simon Capper <skyjunky@sbcglobal.net>
 * Marc Olivier Chouinard <mochouinard@moctel.com>
 * Raymond Chandler <intralanman@freeswitch.org>
 *
 * switch_xml.c -- XML PARSER
 *
 * Derived from ezxml http://ezxml.sourceforge.net
 * Original Copyright
 *
 * Copyright 2004, 2006 Aaron Voisine <aaron@voisine.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include "asterisk.h"
#include "asterisk/switch_xml.h"
#include "asterisk/logger.h"
#include "asterisk/utils.h"
#include "asterisk/autoconfig.h"
#include "asterisk/utils.h"
#include "asterisk/astdb.h"

#define SWITCH_XML_WS   "\t\r\n "	/* whitespace */
#define SWITCH_XML_ERRL 128		/* maximum error string length */

typedef struct switch_xml_root *switch_xml_root_t;
struct switch_xml_root {		/* additional data for the root tag */
	struct switch_xml xml;		/* is a super-struct built on top of switch_xml struct */
	switch_xml_t cur;			/* current xml tree insertion point */
	char *m;					/* original xml string */
	switch_size_t len;			/* length of allocated memory */
	uint8_t dynamic;			/* Free the original string when calling switch_xml_free */
	char *u;					/* UTF-8 conversion of string if original was UTF-16 */
	char *s;					/* start of work area */
	char *e;					/* end of work area */
	char **ent;					/* general entities (ampersand sequences) */
	char ***attr;				/* default attributes */
	char ***pi;					/* processing instructions */
	short standalone;			/* non-zero if <?xml standalone="yes"?> */
	char err[SWITCH_XML_ERRL];	/* error string */
};

char *SWITCH_XML_NIL[] = { NULL };	/* empty, null terminated array of strings */

static ast_mutex_t *REFLOCK = NULL;

switch_xml_t switch_xml_find_child(switch_xml_t node, const char *childname, const char *attrname, const char *value)
{
	switch_xml_t p = NULL;

	if (!(childname && attrname && value)) {
		return node;
	}

	for (p = switch_xml_child(node, childname); p; p = p->next) {
		const char *aname = switch_xml_attr(p, attrname);
		if (aname && !strcasecmp(aname, value)) {
			break;
		}
	}

	return p;
}

switch_xml_t switch_xml_find_child_multi(switch_xml_t node, const char *childname,...)
{
	switch_xml_t p = NULL;
	const char *names[256] = { 0 };
	const char *vals[256] = { 0 };
	int x, i = 0;
	va_list ap;
	const char *attrname, *value = NULL;

	va_start(ap, childname);

	while (i < 255) {
		if ((attrname = va_arg(ap, const char *))) {
			value = va_arg(ap, const char *);
		}
		if (attrname && value) {
			names[i] = attrname;
			vals[i] = value;
		} else {
			break;
		}
		i++;
	}

	va_end(ap);

	if (!(childname && i)) {
		return node;
	}

	for (p = switch_xml_child(node, childname); p; p = p->next) {
		for (x = 0; x < i; x++) {
			if (names[x] && vals[x]) {
				const char *aname = switch_xml_attr(p, names[x]);

				if (aname) {
					if (*vals[x] == '!') {
						const char *sval = vals[x] + 1;
						if (strcasecmp(aname, sval)) {
							goto done;
						}
					} else {
						if (!strcasecmp(aname, vals[x])) {
							goto done;
						}
					}
				}
			}
		}
	}

  done:

	return p;
}

/* returns the first child tag with the given name or NULL if not found */
switch_xml_t switch_xml_child(switch_xml_t xml, const char *name)
{
	xml = (xml) ? xml->child : NULL;
	while (xml && strcmp(name, xml->name))
		xml = xml->sibling;
	return xml;
}

/* returns the Nth tag with the same name in the same subsection or NULL if not found */
switch_xml_t switch_xml_idx(switch_xml_t xml, int idx)
{
	for (; xml && idx; idx--)
		xml = xml->next;
	return xml;
}

/* returns the value of the requested tag attribute or "" if not found */
const char *switch_xml_attr_soft(switch_xml_t xml, const char *attr)
{
	const char *ret = switch_xml_attr(xml, attr);

	return ret ? ret : "";
}

/* returns the value of the requested tag attribute or NULL if not found */
const char *switch_xml_attr(switch_xml_t xml, const char *attr)
{
	int i = 0, j = 1;
	switch_xml_root_t root = (switch_xml_root_t) xml;

	if (!xml || !xml->attr)
		return NULL;
	while (xml->attr[i] && attr && strcmp(attr, xml->attr[i]))
		i += 2;
	if (xml->attr[i])
		return xml->attr[i + 1];	/* found attribute */

	while (root->xml.parent)
		root = (switch_xml_root_t) root->xml.parent;	/* root tag */

	if (!root->attr) {
		return NULL;
	}

	for (i = 0; root->attr[i] && xml->name && strcmp(xml->name, root->attr[i][0]); i++);
	if (!root->attr[i])
		return NULL;			/* no matching default attributes */
	while (root->attr[i][j] && attr && strcmp(attr, root->attr[i][j]))
		j += 3;
	return (root->attr[i][j]) ? root->attr[i][j + 1] : NULL;	/* found default */
}

/* same as switch_xml_get but takes an already initialized va_list */
static switch_xml_t switch_xml_vget(switch_xml_t xml, va_list ap)
{
	char *name = va_arg(ap, char *);
	int idx = -1;

	if (name && *name) {
		idx = va_arg(ap, int);
		xml = switch_xml_child(xml, name);
	}
	return (idx < 0) ? xml : switch_xml_vget(switch_xml_idx(xml, idx), ap);
}

/* Traverses the xml tree to retrieve a specific subtag. Takes a variable
   length list of tag names and indexes. The argument list must be terminated
   by either an index of -1 or an empty string tag name. Example:
   title = switch_xml_get(library, "shelf", 0, "book", 2, "title", -1);
   This retrieves the title of the 3rd book on the 1st shelf of library.
   Returns NULL if not found. */
switch_xml_t switch_xml_get(switch_xml_t xml,...)
{
	va_list ap;
	switch_xml_t r;

	va_start(ap, xml);
	r = switch_xml_vget(xml, ap);
	va_end(ap);
	return r;
}

/* set an error string and return root */
static switch_xml_t switch_xml_err(switch_xml_root_t root, char *s, const char *err, ...)
{
	va_list ap;
	int line = 1;
	char *t, fmt[SWITCH_XML_ERRL];

	if (!root || !root->s) {
		return NULL;
	}

	for (t = root->s; t && t < s; t++)
		if (*t == '\n')
			line++;
	snprintf(fmt, SWITCH_XML_ERRL, "[error near line %d]: %s", line, err);
	ast_log(LOG_ERROR, "[error near line %d]: %s", line, err);

	va_start(ap, err);
	vsnprintf(root->err, SWITCH_XML_ERRL, fmt, ap);
	va_end(ap);

	return &root->xml;
}

/* Recursively decodes entity and character references and normalizes new lines
   ent is a null terminated array of alternating entity names and values. set t
   to '&' for general entity decoding, '%' for parameter entity decoding, 'c'
   for cdata sections, ' ' for attribute normalization, or '*' for non-cdata
   attribute normalization. Returns s, or if the decoded string is longer than
   s, returns a malloced string that must be freed. */
static char *switch_xml_decode(char *s, char **ent, char t)
{
	char *e, *r = s, *m = s;
	unsigned long b, c, d, l;

	for (; *s; s++) {			/* normalize line endings */
		while (*s == '\r') {
			*(s++) = '\n';
			if (*s == '\n')
				memmove(s, (s + 1), strlen(s));
		}
	}

	for (s = r;;) {
		while (*s && *s != '&' && (*s != '%' || t != '%') && !isspace((unsigned char) (*s)))
			s++;

		if (!*s)
			break;
		else if (t != 'c' && !strncmp(s, "&#", 2)) {	/* character reference */
			char *code = s + 2;
			int base = 10;
			if (*code == 'x') {
				code++;
				base = 16;
			}
			if (!isxdigit((int)*code)) { /* "&# 1;" and "&#-1;" are invalid */
				s++;
				continue;
			}
			c = strtoul(code, &e, base);
			if (!c || *e != ';') {
				s++;
				continue;
			}
			/* not a character ref */
			if (c < 0x80)
				*(s++) = (char) c;	/* US-ASCII subset */
			else if (c > 0x7FFFFFFF) { /* out of UTF-8 range */
				s++;
				continue;
			} else {				/* multi-byte UTF-8 sequence */
				for (b = 0, d = c; d; d /= 2)
					b++;		/* number of bits in c */
				b = (b - 2) / 5;	/* number of bytes in payload */
				ast_assert(b < 7);		/* because c <= 0x7FFFFFFF */
				*(s++) = (char) ((0xFF << (7 - b)) | (c >> (6 * b)));	/* head */
				while (b)
					*(s++) = (char) (0x80 | ((c >> (6 * --b)) & 0x3F));	/* payload */
			}

			memmove(s, strchr(s, ';') + 1, strlen(strchr(s, ';')));
		} else if ((*s == '&' && (t == '&' || t == ' ' || t == '*')) || (*s == '%' && t == '%')) {	/* entity reference */
			for (b = 0; ent[b] && strncmp(s + 1, ent[b], strlen(ent[b])); b += 2);	/* find entity in entity list */

			if (ent[b++]) {		/* found a match */
				if ((c = (unsigned long) strlen(ent[b])) - 1 > (e = strchr(s, ';')) - s) {
					l = (d = (unsigned long) (s - r)) + c + (unsigned long) strlen(e);	/* new length */
					if (l) {
						if (r == m) {
							char *tmp = (char *) switch_must_malloc(l);
							r = strcpy(tmp, r);
						} else {
							r = (char *) switch_must_realloc(r, l);
						}
					}
					e = strchr((s = r + d), ';');	/* fix up pointers */
				}

				memmove(s + c, e + 1, strlen(e));	/* shift rest of string */
				strncpy(s, ent[b], c);	/* copy in replacement text */
			} else
				s++;			/* not a known entity */
		} else if ((t == ' ' || t == '*') && isspace((int) (*s)))
			*(s++) = ' ';
		else
			s++;				/* no decoding needed */
	}

	if (t == '*') {				/* normalize spaces for non-cdata attributes */
		for (s = r; *s; s++) {
			if ((l = (unsigned long) strspn(s, " ")))
				memmove(s, s + l, strlen(s + l) + 1);
			while (*s && *s != ' ')
				s++;
		}
		if (--s >= r && *s == ' ')
			*s = '\0';			/* trim any trailing space */
	}
	return r;
}

/* called when parser finds start of new tag */
static void switch_xml_open_tag(switch_xml_root_t root, char *name, char **attr)
{
	switch_xml_t xml;

	if (!root || !root->cur) {
		return;
	}

	xml = root->cur;

	if (xml->name)
		xml = switch_xml_add_child(xml, name, strlen(xml->txt));
	else
		xml->name = name;		/* first open tag */

	xml->attr = attr;
	root->cur = xml;			/* update tag insertion point */
}

/* called when parser finds character content between open and closing tag */
static void switch_xml_char_content(switch_xml_root_t root, char *s, switch_size_t len, char t)
{
	switch_xml_t xml;
	char *m = s;
	switch_size_t l;

	if (!root || !root->cur) {
		return;
	}

	xml = root->cur;

	if (!xml || !xml->name || !len)
		return;					/* sanity check */

	s[len] = '\0';				/* null terminate text (calling functions anticipate this) */
	len = strlen(s = switch_xml_decode(s, root->ent, t)) + 1;

	if (!*(xml->txt))
		xml->txt = s;			/* initial character content */
	else {						/* allocate our own memory and make a copy */
		if ((xml->flags & SWITCH_XML_TXTM)) {	/* allocate some space */
			xml->txt = (char *) switch_must_realloc(xml->txt, (l = strlen(xml->txt)) + len);
		} else {
			char *tmp = (char *) switch_must_malloc((l = strlen(xml->txt)) + len);

			xml->txt = strcpy(tmp, xml->txt);
		}
		strcpy(xml->txt + l, s);	/* add new char content */
		if (s != m)
			free(s);			/* free s if it was malloced by switch_xml_decode() */
	}

	if (xml->txt != m)
		switch_xml_set_flag(xml, SWITCH_XML_TXTM);
}

/* called when parser finds closing tag */
static switch_xml_t switch_xml_close_tag(switch_xml_root_t root, char *name, char *s)
{
	if (!root || !root->cur || !root->cur->name || strcmp(name, root->cur->name))
		return switch_xml_err(root, s, "unexpected closing tag </%s>", name);

	root->cur = root->cur->parent;
	return NULL;
}

/* checks for circular entity references, returns non-zero if no circular
   references are found, zero otherwise */
static int switch_xml_ent_ok(char *name, char *s, char **ent)
{
	int i;

	for (;; s++) {
		while (*s && *s != '&')
			s++;				/* find next entity reference */
		if (!*s)
			return 1;
		if (!strncmp(s + 1, name, strlen(name)))
			return 0;			/* circular ref. */
		for (i = 0; ent[i] && strncmp(ent[i], s + 1, strlen(ent[i])); i += 2);
		if (ent[i] && !switch_xml_ent_ok(name, ent[i + 1], ent))
			return 0;
	}
}

/* called when the parser finds a processing instruction */
static void switch_xml_proc_inst(switch_xml_root_t root, char *s, switch_size_t len)
{
	int i = 0, j = 1;
	char *target = s;
	char **sstmp;
	char *stmp;

	s[len] = '\0';				/* null terminate instruction */
	if (*(s += strcspn(s, SWITCH_XML_WS))) {
		*s = '\0';				/* null terminate target */
		s += strspn(s + 1, SWITCH_XML_WS) + 1;	/* skip whitespace after target */
	}

	if (!root)
		return;

	if (!strcmp(target, "xml")) {	/* <?xml ... ?> */
		if ((s = strstr(s, "standalone")) && !strncmp(s + strspn(s + 10, SWITCH_XML_WS "='\"") + 10, "yes", 3))
			root->standalone = 1;
		return;
	}

	if (!root->pi || !root->pi[0]) {
		root->pi = (char ***) switch_must_malloc(sizeof(char **));
		*(root->pi) = NULL;		/* first pi */
	}

	while (root->pi[i] && strcmp(target, root->pi[i][0]))
		i++;					/* find target */
	if (!root->pi[i]) {			/* new target */
		char ***ssstmp = (char ***) switch_must_realloc(root->pi, sizeof(char **) * (i + 2));

		root->pi = ssstmp;
		if (!root->pi)
			return;
		root->pi[i] = (char **) switch_must_malloc(sizeof(char *) * 3);
		root->pi[i][0] = target;
		root->pi[i][1] = (char *) (root->pi[i + 1] = NULL);	/* terminate pi list */
		root->pi[i][2] = switch_must_strdup("");	/* empty document position list */
	}

	while (root->pi[i][j])
		j++;					/* find end of instruction list for this target */
	sstmp = (char **) switch_must_realloc(root->pi[i], sizeof(char *) * (j + 3));
	root->pi[i] = sstmp;
	stmp = (char *) switch_must_realloc(root->pi[i][j + 1], j + 1);
	root->pi[i][j + 2] = stmp;
	strcpy(root->pi[i][j + 2] + j - 1, (root->xml.name) ? ">" : "<");
	root->pi[i][j + 1] = NULL;	/* null terminate pi list for this target */
	root->pi[i][j] = s;			/* set instruction */
}

/* called when the parser finds an internal doctype subset */
static short switch_xml_internal_dtd(switch_xml_root_t root, char *s, switch_size_t len)
{
	char q, *c, *t, *n = NULL, *v, **ent, **pe;
	int i, j;
	char **sstmp;

	pe = (char **) memcpy(switch_must_malloc(sizeof(SWITCH_XML_NIL)), SWITCH_XML_NIL, sizeof(SWITCH_XML_NIL));

	for (s[len] = '\0'; s;) {
		while (*s && *s != '<' && *s != '%')
			s++;				/* find next declaration */

		if (!*s)
			break;
		else if (!strncmp(s, "<!ENTITY", 8)) {	/* parse entity definitions */
			c = s += strspn(s + 8, SWITCH_XML_WS) + 8;	/* skip white space separator */
			n = s + strspn(s, SWITCH_XML_WS "%");	/* find name */
			*(s = n + strcspn(n, SWITCH_XML_WS)) = ';';	/* append ; to name */

			v = s + strspn(s + 1, SWITCH_XML_WS) + 1;	/* find value */
			if ((q = *(v++)) != '"' && q != '\'') {	/* skip externals */
				s = strchr(s, '>');
				continue;
			}

			for (i = 0, ent = (*c == '%') ? pe : root->ent; ent[i]; i++);
			sstmp = (char **) switch_must_realloc(ent, (i + 3) * sizeof(char *));	/* space for next ent */
			ent = sstmp;
			if (*c == '%')
				pe = ent;
			else
				root->ent = ent;

			*(++s) = '\0';		/* null terminate name */
			if ((s = strchr(v, q)))
				*(s++) = '\0';	/* null terminate value */
			ent[i + 1] = switch_xml_decode(v, pe, '%');	/* set value */
			ent[i + 2] = NULL;	/* null terminate entity list */
			if (!switch_xml_ent_ok(n, ent[i + 1], ent)) {	/* circular reference */
				if (ent[i + 1] != v)
					free(ent[i + 1]);
				switch_xml_err(root, v, "circular entity declaration &%s", n);
				break;
			} else
				ent[i] = n;		/* set entity name */
		} else if (!strncmp(s, "<!ATTLIST", 9)) {	/* parse default attributes */
			t = s + strspn(s + 9, SWITCH_XML_WS) + 9;	/* skip whitespace separator */
			if (!*t) {
				switch_xml_err(root, t, "unclosed <!ATTLIST");
				break;
			}
			if (*(s = t + strcspn(t, SWITCH_XML_WS ">")) == '>')
				continue;
			else
				*s = '\0';		/* null terminate tag name */
			for (i = 0; root->attr[i] && strcmp(n, root->attr[i][0]); i++);

			//while (*(n = ++s + strspn(s, SWITCH_XML_WS)) && *n != '>') {
			// gcc 4.4 you are a creep
			for (;;) {
				s++;
				if (!(*(n = s + strspn(s, SWITCH_XML_WS)) && *n != '>')) {
					break;
				}
				if (*(s = n + strcspn(n, SWITCH_XML_WS)))
					*s = '\0';	/* attr name */
				else {
					switch_xml_err(root, t, "malformed <!ATTLIST");
					break;
				}

				s += strspn(s + 1, SWITCH_XML_WS) + 1;	/* find next token */
				c = (strncmp(s, "CDATA", 5)) ? (char *) "*" : (char *) " ";	/* is it cdata? */
				if (!strncmp(s, "NOTATION", 8))
					s += strspn(s + 8, SWITCH_XML_WS) + 8;
				s = (*s == '(') ? strchr(s, ')') : s + strcspn(s, SWITCH_XML_WS);
				if (!s) {
					switch_xml_err(root, t, "malformed <!ATTLIST");
					break;
				}

				s += strspn(s, SWITCH_XML_WS ")");	/* skip white space separator */
				if (!strncmp(s, "#FIXED", 6))
					s += strspn(s + 6, SWITCH_XML_WS) + 6;
				if (*s == '#') {	/* no default value */
					s += strcspn(s, SWITCH_XML_WS ">") - 1;
					if (*c == ' ')
						continue;	/* cdata is default, nothing to do */
					v = NULL;
				} else if ((*s == '"' || *s == '\'') &&	/* default value */
						   (s = strchr(v = s + 1, *s)))
					*s = '\0';
				else {
					switch_xml_err(root, t, "malformed <!ATTLIST");
					break;
				}

				if (!root->attr[i]) {	/* new tag name */
					root->attr = (!i) ? (char ***) switch_must_malloc(2 * sizeof(char **))
						: (char ***) switch_must_realloc(root->attr, (i + 2) * sizeof(char **));
					root->attr[i] = (char **) switch_must_malloc(2 * sizeof(char *));
					root->attr[i][0] = t;	/* set tag name */
					root->attr[i][1] = (char *) (root->attr[i + 1] = NULL);
				}

				for (j = 1; root->attr[i][j]; j += 3);	/* find end of list */
				sstmp = (char **) switch_must_realloc(root->attr[i], (j + 4) * sizeof(char *));

				root->attr[i] = sstmp;
				root->attr[i][j + 3] = NULL;	/* null terminate list */
				root->attr[i][j + 2] = c;	/* is it cdata? */
				root->attr[i][j + 1] = (v) ? switch_xml_decode(v, root->ent, *c) : NULL;
				root->attr[i][j] = n;	/* attribute name  */
			}
		} else if (!strncmp(s, "<!--", 4))
			s = strstr(s + 4, "-->");	/* comments */
		else if (!strncmp(s, "<?", 2)) {	/* processing instructions */
			if ((s = strstr(c = s + 2, "?>")))
				switch_xml_proc_inst(root, c, s++ - c);
		} else if (*s == '<')
			s = strchr(s, '>');	/* skip other declarations */
		else if (*(s++) == '%' && !root->standalone)
			break;
	}

	free(pe);
	return !*root->err;
}

/* Converts a UTF-16 string to UTF-8. Returns a new string that must be freed
   or NULL if no conversion was needed. */
static char *switch_xml_str2utf8(char **s, switch_size_t *len)
{
	char *u;
	switch_size_t l = 0, sl, max = *len;
	long c, d;
	int b, be = (**s == '\xFE') ? 1 : (**s == '\xFF') ? 0 : -1;

	if (be == -1)
		return NULL;			/* not UTF-16 */

	u = (char *) switch_must_malloc(max);
	for (sl = 2; sl < *len - 1; sl += 2) {
		c = (be) ? (((*s)[sl] & 0xFF) << 8) | ((*s)[sl + 1] & 0xFF)	/* UTF-16BE */
			: (((*s)[sl + 1] & 0xFF) << 8) | ((*s)[sl] & 0xFF);	/* UTF-16LE */
		if (c >= 0xD800 && c <= 0xDFFF && (sl += 2) < *len - 1) {	/* high-half */
			d = (be) ? (((*s)[sl] & 0xFF) << 8) | ((*s)[sl + 1] & 0xFF)
				: (((*s)[sl + 1] & 0xFF) << 8) | ((*s)[sl] & 0xFF);
			c = (((c & 0x3FF) << 10) | (d & 0x3FF)) + 0x10000;
		}

		while (l + 6 > max) {
			char *tmp;
			tmp = (char *) switch_must_realloc(u, max += SWITCH_XML_BUFSIZE);
			u = tmp;
		}
		if (c < 0x80)
			u[l++] = (char) c;	/* US-ASCII subset */
		else {					/* multi-byte UTF-8 sequence */
			for (b = 0, d = c; d; d /= 2)
				b++;			/* bits in c */
			b = (b - 2) / 5;	/* bytes in payload */
			u[l++] = (char) ((0xFF << (7 - b)) | (c >> (6 * b)));	/* head */
			while (b)
				u[l++] = (char) (0x80 | ((c >> (6 * --b)) & 0x3F));	/* payload */
		}
	}
	return *s = (char *) switch_must_realloc(u, *len = l);
}

/* frees a tag attribute list */
static void switch_xml_free_attr(char **attr)
{
	int i = 0;
	char *m;

	if (!attr || attr == SWITCH_XML_NIL)
		return;					/* nothing to free */
	while (attr[i])
		i += 2;					/* find end of attribute list */
	m = attr[i + 1];			/* list of which names and values are malloced */
	for (i = 0; m[i]; i++) {
		if (m[i] & SWITCH_XML_NAMEM)
			free(attr[i * 2]);
		if (m[i] & SWITCH_XML_TXTM)
			free(attr[(i * 2) + 1]);
	}
	free(m);
	free(attr);
}

switch_xml_t switch_xml_parse_str_dynamic(char *s, switch_bool_t dup)
{
	switch_xml_root_t root;
	char *data;

	ast_assert(s);
	data = dup ? switch_must_strdup(s) : s;

	if ((root = (switch_xml_root_t) switch_xml_parse_str(data, strlen(data)))) {
		root->dynamic = 1;		/* Make sure we free the memory is switch_xml_free() */
		return &root->xml;
	} else {
		if (dup) {
			free(data);
		}
		return NULL;
	}
}

/* parse the given xml string and return a switch_xml structure */
switch_xml_t switch_xml_parse_str(char *s, switch_size_t len)
{
	switch_xml_root_t root = (switch_xml_root_t) switch_xml_new(NULL);
	char q, e, *d, **attr, **a = NULL;	/* initialize a to avoid compile warning */
	int l, i, j;

	root->m = s;
	if (!len)
		return switch_xml_err(root, s, "root tag missing");
	root->u = switch_xml_str2utf8(&s, &len);	/* convert utf-16 to utf-8 */
	root->e = (root->s = s) + len;	/* record start and end of work area */

	e = s[len - 1];				/* save end char */
	s[len - 1] = '\0';			/* turn end char into null terminator */

	while (*s && *s != '<')
		s++;					/* find first tag */
	if (!*s)
		return switch_xml_err(root, s, "root tag missing");

	for (;;) {
		attr = (char **) SWITCH_XML_NIL;
		d = ++s;

		if (isalpha((int) (*s)) || *s == '_' || *s == ':' || (int8_t) * s < '\0') {	/* new tag */
			if (!root->cur)
				return switch_xml_err(root, d, "markup outside of root element");

			s += strcspn(s, SWITCH_XML_WS "/>");
			while (isspace((int) (*s)))
				*(s++) = '\0';	/* null terminate tag name */

			if (*s && *s != '/' && *s != '>')	/* find tag in default attr list */
				for (i = 0; (a = root->attr[i]) && strcmp(a[0], d); i++);

			for (l = 0; *s && *s != '/' && *s != '>'; l += 2) {	/* new attrib */
				attr = (l) ? (char **) switch_must_realloc(attr, (l + 4) * sizeof(char *))
					: (char **) switch_must_malloc(4 * sizeof(char *));	/* allocate space */
				attr[l + 3] = (l) ? (char *) switch_must_realloc(attr[l + 1], (l / 2) + 2)
					: (char *) switch_must_malloc(2);	/* mem for list of maloced vals */
				strcpy(attr[l + 3] + (l / 2), " ");	/* value is not malloced */
				attr[l + 2] = NULL;	/* null terminate list */
				attr[l + 1] = (char *) "";	/* temporary attribute value */
				attr[l] = s;	/* set attribute name */

				s += strcspn(s, SWITCH_XML_WS "=/>");
				if (*s == '=' || isspace((int) (*s))) {
					*(s++) = '\0';	/* null terminate tag attribute name */
					q = *(s += strspn(s, SWITCH_XML_WS "="));
					if (q == '"' || q == '\'') {	/* attribute value */
						attr[l + 1] = ++s;
						while (*s && *s != q)
							s++;
						if (*s)
							*(s++) = '\0';	/* null terminate attribute val */
						else {
							switch_xml_free_attr(attr);
							return switch_xml_err(root, d, "missing %c", q);
						}

						for (j = 1; a && a[j] && strcmp(a[j], attr[l]); j += 3);
						attr[l + 1] = switch_xml_decode(attr[l + 1], root->ent, (a && a[j]) ? *a[j + 2] : ' ');
						if (attr[l + 1] < d || attr[l + 1] > s)
							attr[l + 3][l / 2] = SWITCH_XML_TXTM;	/* value malloced */
					}
				}
				while (isspace((int) (*s)))
					s++;
			}

			if (*s == '/') {	/* self closing tag */
				*(s++) = '\0';
				if ((*s && *s != '>') || (!*s && e != '>')) {
					if (l)
						switch_xml_free_attr(attr);
					return switch_xml_err(root, d, "missing >");
				}
				switch_xml_open_tag(root, d, attr);
				switch_xml_close_tag(root, d, s);
			} else if ((q = *s) == '>' || (!*s && e == '>')) {	/* open tag */
				*s = '\0';		/* temporarily null terminate tag name */
				switch_xml_open_tag(root, d, attr);
				*s = q;
			} else {
				if (l)
					switch_xml_free_attr(attr);
				return switch_xml_err(root, d, "missing >");
			}
		} else if (*s == '/') {	/* close tag */
			s += strcspn(d = s + 1, SWITCH_XML_WS ">") + 1;
			if (!(q = *s) && e != '>')
				return switch_xml_err(root, d, "missing >");
			*s = '\0';			/* temporarily null terminate tag name */
			if (switch_xml_close_tag(root, d, s))
				return &root->xml;
			if (isspace((int) (*s = q)))
				s += strspn(s, SWITCH_XML_WS);
		} else if (!strncmp(s, "!--", 3)) {	/* xml comment */
			if (!(s = strstr(s + 3, "--")) || (*(s += 2) != '>' && *s) || (!*s && e != '>'))
				return switch_xml_err(root, d, "unclosed <!--");
		} else if (!strncmp(s, "![CDATA[", 8)) {	/* cdata */
			if ((s = strstr(s, "]]>"))) {
				if (root && root->cur) {
					root->cur->flags |= SWITCH_XML_CDATA;
				}
				switch_xml_char_content(root, d + 8, (s += 2) - d - 10, 'c');
			} else {
				return switch_xml_err(root, d, "unclosed <![CDATA[");
			}
		} else if (!strncmp(s, "!DOCTYPE", 8)) {	/* dtd */
			for (l = 0; *s && ((!l && *s != '>') || (l && (*s != ']' || *(s + strspn(s + 1, SWITCH_XML_WS) + 1) != '>'))); l = (*s == '[') ? 1 : l)
				s += strcspn(s + 1, "[]>") + 1;
			if (!*s && e != '>')
				return switch_xml_err(root, d, "unclosed <!DOCTYPE");
			d = (l) ? strchr(d, '[') + 1 : d;
			if (l && !switch_xml_internal_dtd(root, d, s++ - d))
				return &root->xml;
		} else if (*s == '?') {	/* <?...?> processing instructions */
			do {
				s = strchr(s, '?');
			} while (s && *(++s) && *s != '>');
			if (!s || (!*s && e != '>'))
				return switch_xml_err(root, d, "unclosed <?");
			else
				switch_xml_proc_inst(root, d + 1, s - d - 2);
		} else
			return switch_xml_err(root, d, "unexpected <");

		if (!s || !*s)
			break;
		*s = '\0';
		d = ++s;
		if (*s && *s != '<') {	/* tag character content */
			while (*s && *s != '<')
				s++;
			if (*s)
				switch_xml_char_content(root, d, s - d, '&');
			else
				break;
		} else if (!*s)
			break;
	}

	if (!root->cur)
		return &root->xml;
	else if (!root->cur->name)
		return switch_xml_err(root, d, "root tag missing");
	else
		return switch_xml_err(root, d, "unclosed tag <%s>", root->cur->name);
}

/* Encodes ampersand sequences appending the results to *dst, reallocating *dst
   if length exceeds max. a is non-zero for attribute encoding. Returns *dst */
static char *switch_xml_ampencode(const char *s, switch_size_t len, char **dst, switch_size_t *dlen, switch_size_t *max, short a, switch_bool_t use_utf8_encoding)
{
	const char *e = NULL;
	int immune = 0;
	int expecting_x_utf_8_char = 0;
	int unicode_char = 0x000000;

	if (!(s && *s))
		return *dst;

	if (len) {
		e = s + len;
	}

	while (s != e) {
		while (*dlen + 10 > *max) {
			*dst = (char *) switch_must_realloc(*dst, *max += SWITCH_XML_BUFSIZE);
		}

		if (immune) {
			if (*s == '\0') {
				return *dst;
			}
			(*dst)[(*dlen)++] = *s;
		} else
			switch (*s) {
			case '\0':
				return *dst;
			case '&':
				*dlen += sprintf(*dst + *dlen, "&amp;");
				break;
			case '<':
				if (*(s + 1) == '!') {
					(*dst)[(*dlen)++] = *s;
					immune++;
					break;
				}
				*dlen += sprintf(*dst + *dlen, "&lt;");
				break;
			case '>':
				*dlen += sprintf(*dst + *dlen, "&gt;");
				break;
			case '"':
				*dlen += sprintf(*dst + *dlen, (a) ? "&quot;" : "\"");
				break;
			case '\n':
				*dlen += sprintf(*dst + *dlen, (a) ? "&#xA;" : "\n");
				break;
			case '\t':
				*dlen += sprintf(*dst + *dlen, (a) ? "&#x9;" : "\t");
				break;
			case '\r':
				*dlen += sprintf(*dst + *dlen, "&#xD;");
				break;
			default:
				if (use_utf8_encoding && expecting_x_utf_8_char == 0 && ((*s >> 8) & 0x01)) {
					int num = 1;
					for (;num<4;num++) {
						if (! ((*s >> (7-num)) & 0x01)) {
							break;
						}
					}
					switch (num) {
						case 2:
							unicode_char = *s & 0x1f;
							break;
						case 3:
							unicode_char = *s & 0x0f;
							break;
						case 4:
							unicode_char = *s & 0x07;
							break;
						default:
							ast_log(LOG_WARNING, "Invalid UTF-8 Initial charactere, skip it\n");
							/* ERROR HERE */
							break;
					}
					expecting_x_utf_8_char = num - 1;

				} else if (use_utf8_encoding && expecting_x_utf_8_char > 0) {
					if (((*s >> 6) & 0x03) == 0x2) {

						unicode_char = unicode_char << 6;
						unicode_char = unicode_char | (*s & 0x3f);
					} else {
						ast_log(LOG_WARNING, "Invalid UTF-8 character to ampersand, skip it\n");
						expecting_x_utf_8_char = 0;
						break;
					}
					expecting_x_utf_8_char--;
					if (expecting_x_utf_8_char == 0) {
						*dlen += sprintf(*dst + *dlen, "&#x%X;", unicode_char);
					}
				} else {
					(*dst)[(*dlen)++] = *s;
				}
			}
		s++;
	}
	return *dst;
}

#define XML_INDENT "  "
/* Recursively converts each tag to xml appending it to *s. Reallocates *s if
   its length exceeds max. start is the location of the previous tag in the
   parent tag's character content. Returns *s. */
static char *switch_xml_toxml_r(switch_xml_t xml, char **s, switch_size_t *len, switch_size_t *max, switch_size_t start, char ***attr, uint32_t *count, int isroot, switch_bool_t use_utf8_encoding)
{
	int i, j;
	char *txt;
	switch_size_t off;
	uint32_t lcount;
	uint32_t loops = 0;

  tailrecurse:
	off = 0;
	txt = "";

	if (loops++) {
		isroot = 0;
	}

	if (!isroot && xml->parent) {
		txt = (char *) xml->parent->txt;
	}

	/* parent character content up to this tag */
	*s = switch_xml_ampencode(txt + start, xml->off - start, s, len, max, 0, use_utf8_encoding);

	while (*len + strlen(xml->name) + 5 + (strlen(XML_INDENT) * (*count)) + 1 > *max) {	/* reallocate s */
		*s = (char *) switch_must_realloc(*s, *max += SWITCH_XML_BUFSIZE);
	}

	if (*len && *(*s + (*len) - 1) == '>') {
		*len += sprintf(*s + *len, "\n");	/* indent */
	}
	for (lcount = 0; lcount < *count; lcount++) {
		*len += sprintf(*s + *len, "%s", XML_INDENT);	/* indent */
	}

	*len += sprintf(*s + *len, "<%s", xml->name);	/* open tag */
	for (i = 0; xml->attr[i]; i += 2) {	/* tag attributes */
		if (switch_xml_attr(xml, xml->attr[i]) != xml->attr[i + 1])
			continue;
		while (*len + strlen(xml->attr[i]) + 7 + (strlen(XML_INDENT) * (*count)) > *max) {	/* reallocate s */
			*s = (char *) switch_must_realloc(*s, *max += SWITCH_XML_BUFSIZE);
		}

		*len += sprintf(*s + *len, " %s=\"", xml->attr[i]);
		switch_xml_ampencode(xml->attr[i + 1], 0, s, len, max, 1, use_utf8_encoding);
		*len += sprintf(*s + *len, "\"");
	}

	for (i = 0; attr[i] && strcmp(attr[i][0], xml->name); i++);
	for (j = 1; attr[i] && attr[i][j]; j += 3) {	/* default attributes */
		if (!attr[i][j + 1] || switch_xml_attr(xml, attr[i][j]) != attr[i][j + 1])
			continue;			/* skip duplicates and non-values */
		while (*len + strlen(attr[i][j]) + 8 + (strlen(XML_INDENT) * (*count)) > *max) {	/* reallocate s */
			*s = (char *) switch_must_realloc(*s, *max += SWITCH_XML_BUFSIZE);
		}

		*len += sprintf(*s + *len, " %s=\"", attr[i][j]);
		switch_xml_ampencode(attr[i][j + 1], 0, s, len, max, 1, use_utf8_encoding);
		*len += sprintf(*s + *len, "\"");
	}

	*len += sprintf(*s + *len, (xml->child || xml->txt) ? ">" : "/>\n");

	if (xml->child) {
		(*count)++;
		*s = switch_xml_toxml_r(xml->child, s, len, max, 0, attr, count, 0, use_utf8_encoding);

	} else {
		*s = switch_xml_ampencode(xml->txt, 0, s, len, max, 0, use_utf8_encoding);	/* data */
	}

	while (*len + strlen(xml->name) + 5 + (strlen(XML_INDENT) * (*count)) > *max) {	/* reallocate s */
		*s = (char *) switch_must_realloc(*s, *max += SWITCH_XML_BUFSIZE);
	}

	if (xml->child || xml->txt) {
		if (*(*s + (*len) - 1) == '\n') {
			for (lcount = 0; lcount < *count; lcount++) {
				*len += sprintf(*s + *len, "%s", XML_INDENT);	/* indent */
			}
		}
		*len += sprintf(*s + (*len), "</%s>\n", xml->name);	/* close tag */
	}

	while (txt[off] && off < xml->off)
		off++;					/* make sure off is within bounds */

	if (!isroot && xml->ordered) {
		xml = xml->ordered;
		start = off;
		goto tailrecurse;
/*
		return switch_xml_toxml_r(xml->ordered, s, len, max, off, attr, count, use_utf8_encoding);
*/
	} else {
		if (*count > 0)
			(*count)--;
		return switch_xml_ampencode(txt + off, 0, s, len, max, 0, use_utf8_encoding);
	}
}

char *switch_xml_toxml_nolock_ex(switch_xml_t xml, switch_bool_t prn_header, switch_bool_t use_utf8_encoding)
{
	char *s = (char *) switch_must_malloc(SWITCH_XML_BUFSIZE);

	return switch_xml_toxml_buf_ex(xml, s, SWITCH_XML_BUFSIZE, 0, prn_header, use_utf8_encoding);
}

char *switch_xml_toxml_ex(switch_xml_t xml, switch_bool_t prn_header, switch_bool_t use_utf8_encoding)
{
	char *r, *s;

	s = (char *) switch_must_malloc(SWITCH_XML_BUFSIZE);

	r = switch_xml_toxml_buf_ex(xml, s, SWITCH_XML_BUFSIZE, 0, prn_header, use_utf8_encoding);

	return r;
}

/* converts a switch_xml structure back to xml, returning a string of xml data that
   must be freed */
char *switch_xml_toxml_buf_ex(switch_xml_t xml, char *buf, switch_size_t buflen, switch_size_t offset, switch_bool_t prn_header, switch_bool_t use_utf8_encoding)
{
	switch_xml_t p = (xml) ? xml->parent : NULL;
	switch_xml_root_t root = (switch_xml_root_t) xml;
	switch_size_t len = 0, max = buflen;
	char *s, *t, *n;
	int i, j, k;
	uint32_t count = 0;

	s = buf;
	ast_assert(s != NULL);
	memset(s, 0, max);
	len += offset;
	if (prn_header) {
		len += sprintf(s + len, "<?xml version=\"1.0\"?>\n");
	}

	if (!xml || !xml->name) {
		return (char *) switch_must_realloc(s, len + 1);
	}

	while (root->xml.parent) {
		root = (switch_xml_root_t) root->xml.parent;	/* root tag */
	}

	for (i = 0; !p && root->pi[i]; i++) {	/* pre-root processing instructions */
		for (k = 2; root->pi[i][k - 1]; k++);
		for (j = 1; (n = root->pi[i][j]); j++) {
			if (root->pi[i][k][j - 1] == '>') {
				continue;		/* not pre-root */
			}
			while (len + strlen(t = root->pi[i][0]) + strlen(n) + 7 > max) {
				s = (char *) switch_must_realloc(s, max += SWITCH_XML_BUFSIZE);
			}
			len += sprintf(s + len, "<?%s%s%s?>", t, *n ? " " : "", n);
		}
	}

	s = switch_xml_toxml_r(xml, &s, &len, &max, 0, root->attr, &count, 1, use_utf8_encoding);

	for (i = 0; !p && root->pi[i]; i++) {	/* post-root processing instructions */
		for (k = 2; root->pi[i][k - 1]; k++);
		for (j = 1; (n = root->pi[i][j]); j++) {
			if (root->pi[i][k][j - 1] == '<') {
				continue;		/* not post-root */
			}
			while (len + strlen(t = root->pi[i][0]) + strlen(n) + 7 > max) {
				s = (char *) switch_must_realloc(s, max += SWITCH_XML_BUFSIZE);
			}
			len += sprintf(s + len, "\n<?%s%s%s?>", t, *n ? " " : "", n);
		}
	}

	return (char *) switch_must_realloc(s, len + 1);
}

/* free the memory allocated for the switch_xml structure */
void switch_xml_free(switch_xml_t xml)
{
	switch_xml_root_t root;
	int i, j;
	char **a, *s;
	switch_xml_t orig_xml;
	int refs = 0;

tailrecurse:
	root = (switch_xml_root_t) xml;
	if (!xml) {
		return;
	}

	if (switch_test_flag(xml, SWITCH_XML_ROOT)) {
		ast_mutex_lock(REFLOCK);

		if (xml->refs) {
			xml->refs--;
			refs = xml->refs;
		}
		ast_mutex_unlock(REFLOCK);
	}

	if (refs) {
		return;
	}

	if (xml->free_path) {
		if (unlink(xml->free_path) != 0) {
			ast_log(LOG_WARNING, "Failed to delete file [%s]\n", xml->free_path);
		}
		ast_free(xml->free_path);
	}

	switch_xml_free(xml->child);
	/*switch_xml_free(xml->ordered); */

	if (!xml->parent) {			/* free root tag allocations */
#if (_MSC_VER >= 1400)			// VC8+
		__analysis_assume(sizeof(root->ent) > 44);	/* tail recursion confuses code analysis */
#endif
		for (i = 10; root->ent[i]; i += 2)	/* 0 - 9 are default entities (<>&"') */
			if ((s = root->ent[i + 1]) < root->s || s > root->e)
				free(s);
		free(root->ent);		/* free list of general entities */

		for (i = 0; (a = root->attr[i]); i++) {
			for (j = 1; a[j++]; j += 2)	/* free malloced attribute values */
				if (a[j] && (a[j] < root->s || a[j] > root->e))
					free(a[j]);
			free(a);
		}
		if (root->attr[0])
			free(root->attr);	/* free default attribute list */

		for (i = 0; root->pi[i]; i++) {
			for (j = 1; root->pi[i][j]; j++);
			free(root->pi[i][j + 1]);
			free(root->pi[i]);
		}
		if (root->pi[0])
			free(root->pi);		/* free processing instructions */

		if (root->dynamic == 1)
			free(root->m);		/* malloced xml data */
		if (root->u)
			free(root->u);		/* utf8 conversion */
	}

	switch_xml_free_attr(xml->attr);	/* tag attributes */
	if ((xml->flags & SWITCH_XML_TXTM))
		free(xml->txt);			/* character content */
	if ((xml->flags & SWITCH_XML_NAMEM))
		free(xml->name);		/* tag name */
	if (xml->ordered) {
		orig_xml = xml;
		xml = xml->ordered;
		free(orig_xml);
		goto tailrecurse;
	}
	free(xml);
}

/* return parser error message or empty string if none */
const char *switch_xml_error(switch_xml_t xml)
{
	while (xml && xml->parent)
		xml = xml->parent;		/* find root tag */
	return (xml) ? ((switch_xml_root_t) xml)->err : "";
}

/* returns a new empty switch_xml structure with the given root tag name */
switch_xml_t switch_xml_new(const char *name)
{
	static const char *ent[] = { "lt;", "&#60;", "gt;", "&#62;", "quot;", "&#34;",
		"apos;", "&#39;", "amp;", "&#38;", NULL
	};
	switch_xml_root_t root = (switch_xml_root_t) switch_must_malloc(sizeof(struct switch_xml_root));

	memset(root, '\0', sizeof(struct switch_xml_root));
	root->xml.name = (char *) name;
	root->cur = &root->xml;
	strcpy(root->err, root->xml.txt = (char *) "");
	root->ent = (char **) memcpy(switch_must_malloc(sizeof(ent)), ent, sizeof(ent));
	root->attr = root->pi = (char ***) (root->xml.attr = SWITCH_XML_NIL);
	return &root->xml;
}

/* inserts an existing tag into a switch_xml structure */
switch_xml_t switch_xml_insert(switch_xml_t xml, switch_xml_t dest, switch_size_t off)
{
	switch_xml_t cur, prev, head;

	xml->next = xml->sibling = xml->ordered = NULL;
	xml->off = off;
	xml->parent = dest;

	if ((head = dest->child)) {	/* already have sub tags */
		if (head->off <= off) {	/* not first subtag */
			for (cur = head; cur->ordered && cur->ordered->off <= off; cur = cur->ordered);
			xml->ordered = cur->ordered;
			cur->ordered = xml;
		} else {				/* first subtag */
			xml->ordered = head;
			dest->child = xml;
		}

		for (cur = head, prev = NULL; cur && strcmp(cur->name, xml->name); prev = cur, cur = cur->sibling);	/* find tag type */
		if (cur && cur->off <= off) {	/* not first of type */
			while (cur->next && cur->next->off <= off)
				cur = cur->next;
			xml->next = cur->next;
			cur->next = xml;
		} else {				/* first tag of this type */
			if (prev && cur)
				prev->sibling = cur->sibling;	/* remove old first */
			xml->next = cur;	/* old first tag is now next */
			for (cur = head, prev = NULL; cur && cur->off <= off; prev = cur, cur = cur->sibling);	/* new sibling insert point */
			xml->sibling = cur;
			if (prev)
				prev->sibling = xml;
		}
	} else
		dest->child = xml;		/* only sub tag */

	return xml;
}

/* Adds a child tag. off is the offset of the child tag relative to the start
   of the parent tag's character content. Returns the child tag */
switch_xml_t switch_xml_add_child(switch_xml_t xml, const char *name, switch_size_t off)
{
	switch_xml_t child;

	if (!xml)
		return NULL;
	child = (switch_xml_t) switch_must_malloc(sizeof(struct switch_xml));

	memset(child, '\0', sizeof(struct switch_xml));
	child->name = (char *) name;
	child->attr = SWITCH_XML_NIL;
	child->off = off;
	child->parent = xml;
	child->txt = (char *) "";

	return switch_xml_insert(child, xml, off);
}

/* Adds a child tag. off is the offset of the child tag relative to the start
   of the parent tag's character content. Returns the child tag */
switch_xml_t switch_xml_add_child_d(switch_xml_t xml, const char *name, switch_size_t off)
{
	if (!xml) return NULL;
	return switch_xml_set_flag(switch_xml_add_child(xml, strdup(switch_str_nil(name)), off), SWITCH_XML_NAMEM);
}

/* sets the character content for the given tag and returns the tag */
switch_xml_t switch_xml_set_txt(switch_xml_t xml, const char *txt)
{
	if (!xml)
		return NULL;
	if (xml->flags & SWITCH_XML_TXTM)
		free(xml->txt);			/* existing txt was malloced */
	xml->flags &= ~SWITCH_XML_TXTM;
	xml->txt = (char *) txt;
	return xml;
}

/* sets the character content for the given tag and returns the tag */
switch_xml_t switch_xml_set_txt_d(switch_xml_t xml, const char *txt)
{
	if (!xml) return NULL;
	return switch_xml_set_flag(switch_xml_set_txt(xml, strdup(switch_str_nil(txt))), SWITCH_XML_TXTM);
}

/* Sets the given tag attribute or adds a new attribute if not found. A value
   of NULL will remove the specified attribute.  Returns the tag given */
switch_xml_t switch_xml_set_attr(switch_xml_t xml, const char *name, const char *value)
{
	int l = 0, c;

	if (!xml)
		return NULL;
	while (xml->attr[l] && strcmp(xml->attr[l], name))
		l += 2;
	if (!xml->attr[l]) {		/* not found, add as new attribute */
		if (!value)
			return xml;			/* nothing to do */
		if (xml->attr == SWITCH_XML_NIL) {	/* first attribute */
			xml->attr = (char **) switch_must_malloc(4 * sizeof(char *));
			xml->attr[l + 1] = switch_must_strdup("");	/* empty list of malloced names/vals */
		} else {
			xml->attr = (char **) switch_must_realloc(xml->attr, (l + 4) * sizeof(char *));
		}

		xml->attr[l] = (char *) name;	/* set attribute name */
		xml->attr[l + 2] = NULL;	/* null terminate attribute list */
		xml->attr[l + 3] = (char *) switch_must_realloc(xml->attr[l + 1], (c = (int) strlen(xml->attr[l + 1])) + 2);
		strcpy(xml->attr[l + 3] + c, " ");	/* set name/value as not malloced */
		if (xml->flags & SWITCH_XML_DUP)
			xml->attr[l + 3][c] = SWITCH_XML_NAMEM;
	} else if (xml->flags & SWITCH_XML_DUP)
		free((char *) name);	/* name was strduped */

	for (c = l; xml->attr[c]; c += 2);	/* find end of attribute list */
	if (xml->attr[c + 1][l / 2] & SWITCH_XML_TXTM)
		free(xml->attr[l + 1]);	/* old val */
	if (xml->flags & SWITCH_XML_DUP)
		xml->attr[c + 1][l / 2] |= SWITCH_XML_TXTM;
	else
		xml->attr[c + 1][l / 2] &= ~SWITCH_XML_TXTM;

	if (value)
		xml->attr[l + 1] = (char *) value;	/* set attribute value */
	else {						/* remove attribute */
		if (xml->attr[c + 1][l / 2] & SWITCH_XML_NAMEM)
			free(xml->attr[l]);
		memmove(xml->attr + l, xml->attr + l + 2, (c - l + 2) * sizeof(char *));
		xml->attr = (char **) switch_must_realloc(xml->attr, (c + 2) * sizeof(char *));
		memmove(xml->attr[c + 1] + (l / 2), xml->attr[c + 1] + (l / 2) + 1, (c / 2) - (l / 2));	/* fix list of which name/vals are malloced */
	}
	xml->flags &= ~SWITCH_XML_DUP;	/* clear strdup() flag */

	return xml;
}

/* Sets the given tag attribute or adds a new attribute if not found. A value
   of NULL will remove the specified attribute.  Returns the tag given */
switch_xml_t switch_xml_set_attr_d(switch_xml_t xml, const char *name, const char *value)
{
	if (!xml) return NULL;
	return switch_xml_set_attr(switch_xml_set_flag(xml, SWITCH_XML_DUP), strdup(switch_str_nil(name)), strdup(switch_str_nil(value)));
}

/* Sets the given tag attribute or adds a new attribute if not found. A value
   of NULL will remove the specified attribute.  Returns the tag given */
switch_xml_t switch_xml_set_attr_d_buf(switch_xml_t xml, const char *name, const char *value)
{
	if (!xml) return NULL;
	return switch_xml_set_attr(switch_xml_set_flag(xml, SWITCH_XML_DUP), strdup(switch_str_nil(name)), strdup(switch_str_nil(value)));
}

/* sets a flag for the given tag and returns the tag */
switch_xml_t switch_xml_set_flag(switch_xml_t xml, switch_xml_flag_t flag)
{
	if (xml)
		xml->flags |= flag;
	return xml;
}

/* removes a tag along with its subtags without freeing its memory */
switch_xml_t switch_xml_cut(switch_xml_t xml)
{
	switch_xml_t cur;

	if (!xml)
		return NULL;			/* nothing to do */
	if (xml->next)
		xml->next->sibling = xml->sibling;	/* patch sibling list */

	if (xml->parent) {			/* not root tag */
		cur = xml->parent->child;	/* find head of subtag list */
		if (cur == xml)
			xml->parent->child = xml->ordered;	/* first subtag */
		else {					/* not first subtag */
			while (cur->ordered != xml)
				cur = cur->ordered;
			cur->ordered = cur->ordered->ordered;	/* patch ordered list */

			cur = xml->parent->child;	/* go back to head of subtag list */
			if (strcmp(cur->name, xml->name)) {	/* not in first sibling list */
				while (strcmp(cur->sibling->name, xml->name))
					cur = cur->sibling;
				if (cur->sibling == xml) {	/* first of a sibling list */
					cur->sibling = (xml->next) ? xml->next : cur->sibling->sibling;
				} else
					cur = cur->sibling;	/* not first of a sibling list */
			}

			while (cur->next && cur->next != xml)
				cur = cur->next;
			if (cur->next)
				cur->next = cur->next->next;	/* patch next list */
		}
	}
	xml->ordered = xml->sibling = xml->next = NULL;	/* prevent switch_xml_free() from clobbering ordered list */
	return xml;
}

#define LONGDOUBLE_TYPE	long double

/*
** Conversion types fall into various categories as defined by the
** following enumeration.
*/
#define etRADIX       1			/* Integer types.  %d, %x, %o, and so forth */
#define etFLOAT       2			/* Floating point.  %f */
#define etEXP         3			/* Exponentional notation. %e and %E */
#define etGENERIC     4			/* Floating or exponential, depending on exponent. %g */
#define etSIZE        5			/* Return number of characters processed so far. %n */
#define etSTRING      6			/* Strings. %s */
#define etDYNSTRING   7			/* Dynamically allocated strings. %z */
#define etPERCENT     8			/* Percent symbol. %% */
#define etCHARX       9			/* Characters. %c */
/* The rest are extensions, not normally found in printf() */
#define etCHARLIT    10			/* Literal characters.  %' */
#define etSQLESCAPE  11			/* Strings with '\'' doubled.  %q */
#define etSQLESCAPE2 12			/* Strings with '\'' doubled and enclosed in '',
								   NULL pointers replaced by SQL NULL.  %Q */
#ifdef __UNSUPPORTED__
#define etTOKEN      13			/* a pointer to a Token structure */
#define etSRCLIST    14			/* a pointer to a SrcList */
#endif
#define etPOINTER    15			/* The %p conversion */
#define etSQLESCAPE3 16
#define etSQLESCAPE4 17

/*
** An "etByte" is an 8-bit unsigned value.
*/
typedef unsigned char etByte;

/*
** Each builtin conversion character (ex: the 'd' in "%d") is described
** by an instance of the following structure
*/
typedef struct et_info {		/* Information about each format field */
	char fmttype;				/* The format field code letter */
	etByte base;				/* The base for radix conversion */
	etByte flags;				/* One or more of FLAG_ constants below */
	etByte type;				/* Conversion paradigm */
	etByte charset;				/* Offset into aDigits[] of the digits string */
	etByte prefix;				/* Offset into aPrefix[] of the prefix string */
} et_info;

/*
** Allowed values for et_info.flags
*/
#define FLAG_SIGNED  1			/* True if the value to convert is signed */
#define FLAG_INTERN  2			/* True if for internal use only */
#define FLAG_STRING  4			/* Allow infinity precision */


/*
** The following table is searched linearly, so it is good to put the
** most frequently used conversion types first.
*/
static const char aDigits[] = "0123456789ABCDEF0123456789abcdef";
static const char aPrefix[] = "-x0\000X0";
static const et_info fmtinfo[] = {
	{'d', 10, 1, etRADIX, 0, 0},
	{'s', 0, 4, etSTRING, 0, 0},
	{'g', 0, 1, etGENERIC, 30, 0},
	{'z', 0, 6, etDYNSTRING, 0, 0},
	{'q', 0, 4, etSQLESCAPE, 0, 0},
	{'Q', 0, 4, etSQLESCAPE2, 0, 0},
	{'w', 0, 4, etSQLESCAPE3, 0, 0},
	{'y', 0, 4, etSQLESCAPE4, 0, 0},
	{'c', 0, 0, etCHARX, 0, 0},
	{'o', 8, 0, etRADIX, 0, 2},
	{'u', 10, 0, etRADIX, 0, 0},
	{'x', 16, 0, etRADIX, 16, 1},
	{'X', 16, 0, etRADIX, 0, 4},
#ifndef SWITCH_OMIT_FLOATING_POINT
	{'f', 0, 1, etFLOAT, 0, 0},
	{'e', 0, 1, etEXP, 30, 0},
	{'E', 0, 1, etEXP, 14, 0},
	{'G', 0, 1, etGENERIC, 14, 0},
#endif
	{'i', 10, 1, etRADIX, 0, 0},
	{'n', 0, 0, etSIZE, 0, 0},
	{'%', 0, 0, etPERCENT, 0, 0},
	{'p', 16, 0, etPOINTER, 0, 1},
#ifdef __UNSUPPORTED__
	{'T', 0, 2, etTOKEN, 0, 0},
	{'S', 0, 2, etSRCLIST, 0, 0},
#endif
};

#define etNINFO  (sizeof(fmtinfo)/sizeof(fmtinfo[0]))

/*
** If SWITCH_OMIT_FLOATING_POINT is defined, then none of the floating point
** conversions will work.
*/
#ifndef SWITCH_OMIT_FLOATING_POINT
/*
** "*val" is a double such that 0.1 <= *val < 10.0
** Return the ascii code for the leading digit of *val, then
** multiply "*val" by 10.0 to renormalize.
**
** Example:
**     input:     *val = 3.14159
**     output:    *val = 1.4159    function return = '3'
**
** The counter *cnt is incremented each time.  After counter exceeds
** 16 (the number of significant digits in a 64-bit float) '0' is
** always returned.
*/
static int et_getdigit(LONGDOUBLE_TYPE * val, int *cnt)
{
	int digit;
	LONGDOUBLE_TYPE d;
	if ((*cnt)++ >= 16)
		return '0';
	digit = (int) *val;
	d = digit;
	digit += '0';
	*val = (*val - d) * 10.0;
	return digit;
}
#endif /* SWITCH_OMIT_FLOATING_POINT */

/*
** On machines with a small stack size, you can redefine the
** SWITCH_PRINT_BUF_SIZE to be less than 350.  But beware - for
** smaller values some %f conversions may go into an infinite loop.
*/
#ifndef SWITCH_PRINT_BUF_SIZE
# define SWITCH_PRINT_BUF_SIZE 350
#endif
#define etBUFSIZE SWITCH_PRINT_BUF_SIZE	/* Size of the output buffer */

/*
** The root program.  All variations call this core.
**
** INPUTS:
**   func   This is a pointer to a function taking three arguments
**            1. A pointer to anything.  Same as the "arg" parameter.
**            2. A pointer to the list of characters to be output
**               (Note, this list is NOT null terminated.)
**            3. An integer number of characters to be output.
**               (Note: This number might be zero.)
**
**   arg    This is the pointer to anything which will be passed as the
**          first argument to "func".  Use it for whatever you like.
**
**   fmt    This is the format string, as in the usual print.
**
**   ap     This is a pointer to a list of arguments.  Same as in
**          vfprint.
**
** OUTPUTS:
**          The return value is the total number of characters sent to
**          the function "func".  Returns -1 on a error.
**
** Note that the order in which automatic variables are declared below
** seems to make a big difference in determining how fast this beast
** will run.
*/
static int vxprintf(void (*func) (void *, const char *, int),	/* Consumer of text */
					void *arg,	/* First argument to the consumer */
					int useExtended,	/* Allow extended %-conversions */
					const char *fmt,	/* Format string */
					va_list ap	/* arguments */
	)
{
	int c;						/* Next character in the format string */
	char *bufpt;				/* Pointer to the conversion buffer */
	int precision;				/* Precision of the current field */
	int length;					/* Length of the field */
	int idx;					/* A general purpose loop counter */
	int count;					/* Total number of characters output */
	int width;					/* Width of the current field */
	etByte flag_leftjustify;	/* True if "-" flag is present */
	etByte flag_plussign;		/* True if "+" flag is present */
	etByte flag_blanksign;		/* True if " " flag is present */
	etByte flag_alternateform;	/* True if "#" flag is present */
	etByte flag_altform2;		/* True if "!" flag is present */
	etByte flag_zeropad;		/* True if field width constant starts with zero */
	etByte flag_long;			/* True if "l" flag is present */
	etByte flag_longlong;		/* True if the "ll" flag is present */
	etByte done;				/* Loop termination flag */
	uint64_t longvalue;			/* Value for integer types */
	LONGDOUBLE_TYPE realvalue;	/* Value for real types */
	const et_info *infop;		/* Pointer to the appropriate info structure */
	char buf[etBUFSIZE];		/* Conversion buffer */
	char prefix;				/* Prefix character.  "+" or "-" or " " or '\0'. */
	etByte errorflag = 0;		/* True if an error is encountered */
	etByte xtype = 0;			/* Conversion paradigm */
	char *zExtra;				/* Extra memory used for etTCLESCAPE conversions */
	static const char spaces[] = "                                                                         ";
#define etSPACESIZE (sizeof(spaces)-1)
#ifndef SWITCH_OMIT_FLOATING_POINT
	int exp, e2;				/* exponent of real numbers */
	double rounder;				/* Used for rounding floating point values */
	etByte flag_dp;				/* True if decimal point should be shown */
	etByte flag_rtz;			/* True if trailing zeros should be removed */
	etByte flag_exp;			/* True to force display of the exponent */
	int nsd;					/* Number of significant digits returned */
#endif

	func(arg, "", 0);
	count = length = 0;
	bufpt = 0;
	for (; (c = (*fmt)) != 0; ++fmt) {
		if (c != '%') {
			int amt;
			bufpt = (char *) fmt;
			amt = 1;
			while ((c = (*++fmt)) != '%' && c != 0)
				amt++;
			(*func) (arg, bufpt, amt);
			count += amt;
			if (c == 0)
				break;
		}
		if ((c = (*++fmt)) == 0) {
			errorflag = 1;
			(*func) (arg, "%", 1);
			count++;
			break;
		}
		/* Find out what flags are present */
		flag_leftjustify = flag_plussign = flag_blanksign = flag_alternateform = flag_altform2 = flag_zeropad = 0;
		done = 0;
		do {
			switch (c) {
			case '-':
				flag_leftjustify = 1;
				break;
			case '+':
				flag_plussign = 1;
				break;
			case ' ':
				flag_blanksign = 1;
				break;
			case '#':
				flag_alternateform = 1;
				break;
			case '!':
				flag_altform2 = 1;
				break;
			case '0':
				flag_zeropad = 1;
				break;
			default:
				done = 1;
				break;
			}
		} while (!done && (c = (*++fmt)) != 0);
		/* Get the field width */
		width = 0;
		if (c == '*') {
			width = va_arg(ap, int);
			if (width < 0) {
				flag_leftjustify = 1;
				width = -width;
			}
			c = *++fmt;
		} else {
			while (c >= '0' && c <= '9') {
				width = width * 10 + c - '0';
				c = *++fmt;
			}
		}
		if (width > etBUFSIZE - 10) {
			width = etBUFSIZE - 10;
		}
		/* Get the precision */
		if (c == '.') {
			precision = 0;
			c = *++fmt;
			if (c == '*') {
				precision = va_arg(ap, int);
				if (precision < 0)
					precision = -precision;
				c = *++fmt;
			} else {
				while (c >= '0' && c <= '9') {
					precision = precision * 10 + c - '0';
					c = *++fmt;
				}
			}
		} else {
			precision = -1;
		}
		/* Get the conversion type modifier */
		if (c == 'l') {
			flag_long = 1;
			c = *++fmt;
			if (c == 'l') {
				flag_longlong = 1;
				c = *++fmt;
			} else {
				flag_longlong = 0;
			}
		} else {
			flag_long = flag_longlong = 0;
		}
		/* Fetch the info entry for the field */
		infop = 0;
		for (idx = 0; idx < etNINFO; idx++) {
			if (c == fmtinfo[idx].fmttype) {
				infop = &fmtinfo[idx];
				if (useExtended || (infop->flags & FLAG_INTERN) == 0) {
					xtype = infop->type;
				} else {
					return -1;
				}
				break;
			}
		}
		zExtra = 0;
		if (infop == 0) {
			return -1;
		}


		/* Limit the precision to prevent overflowing buf[] during conversion */
		if (precision > etBUFSIZE - 40 && (infop->flags & FLAG_STRING) == 0) {
			precision = etBUFSIZE - 40;
		}

		/*
		 ** At this point, variables are initialized as follows:
		 **
		 **   flag_alternateform          TRUE if a '#' is present.
		 **   flag_altform2               TRUE if a '!' is present.
		 **   flag_plussign               TRUE if a '+' is present.
		 **   flag_leftjustify            TRUE if a '-' is present or if the
		 **                               field width was negative.
		 **   flag_zeropad                TRUE if the width began with 0.
		 **   flag_long                   TRUE if the letter 'l' (ell) prefixed
		 **                               the conversion character.
		 **   flag_longlong               TRUE if the letter 'll' (ell ell) prefixed
		 **                               the conversion character.
		 **   flag_blanksign              TRUE if a ' ' is present.
		 **   width                       The specified field width.  This is
		 **                               always non-negative.  Zero is the default.
		 **   precision                   The specified precision.  The default
		 **                               is -1.
		 **   xtype                       The class of the conversion.
		 **   infop                       Pointer to the appropriate info struct.
		 */
		switch (xtype) {
		case etPOINTER:
			flag_longlong = sizeof(char *) == sizeof(int64_t);
			flag_long = sizeof(char *) == sizeof(long int);
			/* Fall through into the next case */
		case etRADIX:
			if (infop->flags & FLAG_SIGNED) {
				int64_t v;
				if (flag_longlong)
					v = va_arg(ap, int64_t);
				else if (flag_long)
					v = va_arg(ap, long int);
				else
					v = va_arg(ap, int);
				if (v < 0) {
					longvalue = -v;
					prefix = '-';
				} else {
					longvalue = v;
					if (flag_plussign)
						prefix = '+';
					else if (flag_blanksign)
						prefix = ' ';
					else
						prefix = 0;
				}
			} else {
				if (flag_longlong)
					longvalue = va_arg(ap, uint64_t);
				else if (flag_long)
					longvalue = va_arg(ap, unsigned long int);
				else
					longvalue = va_arg(ap, unsigned int);
				prefix = 0;
			}
			if (longvalue == 0)
				flag_alternateform = 0;
			if (flag_zeropad && precision < width - (prefix != 0)) {
				precision = width - (prefix != 0);
			}
			bufpt = &buf[etBUFSIZE - 1];
			{
				register const char *cset;	/* Use registers for speed */
				register int base;
				cset = &aDigits[infop->charset];
				base = infop->base;
				do {			/* Convert to ascii */
					*(--bufpt) = cset[longvalue % base];
					longvalue = longvalue / base;
				} while (longvalue > 0);
			}
			length = (int)(&buf[etBUFSIZE - 1] - bufpt);
			for (idx = precision - length; idx > 0; idx--) {
				*(--bufpt) = '0';	/* Zero pad */
			}
			if (prefix)
				*(--bufpt) = prefix;	/* Add sign */
			if (flag_alternateform && infop->prefix) {	/* Add "0" or "0x" */
				const char *pre;
				char x;
				pre = &aPrefix[infop->prefix];
				if (*bufpt != pre[0]) {
					for (; (x = (*pre)) != 0; pre++)
						*(--bufpt) = x;
				}
			}
			length = (int)(&buf[etBUFSIZE - 1] - bufpt);
			break;
		case etFLOAT:
		case etEXP:
		case etGENERIC:
			realvalue = va_arg(ap, double);
#ifndef SWITCH_OMIT_FLOATING_POINT
			if (precision < 0)
				precision = 6;	/* Set default precision */
			if (precision > etBUFSIZE / 2 - 10)
				precision = etBUFSIZE / 2 - 10;
			if (realvalue < 0.0) {
				realvalue = -realvalue;
				prefix = '-';
			} else {
				if (flag_plussign)
					prefix = '+';
				else if (flag_blanksign)
					prefix = ' ';
				else
					prefix = 0;
			}
			if (xtype == etGENERIC && precision > 0)
				precision--;
#if 0
			/* Rounding works like BSD when the constant 0.4999 is used.  Wierd! */
			for (idx = precision, rounder = 0.4999; idx > 0; idx--, rounder *= 0.1);
#else
			/* It makes more sense to use 0.5 */
			for (idx = precision, rounder = 0.5; idx > 0; idx--, rounder *= 0.1) {
			}
#endif
			if (xtype == etFLOAT)
				realvalue += rounder;
			/* Normalize realvalue to within 10.0 > realvalue >= 1.0 */
			exp = 0;
			if (realvalue > 0.0) {
				while (realvalue >= 1e32 && exp <= 350) {
					realvalue *= 1e-32;
					exp += 32;
				}
				while (realvalue >= 1e8 && exp <= 350) {
					realvalue *= 1e-8;
					exp += 8;
				}
				while (realvalue >= 10.0 && exp <= 350) {
					realvalue *= 0.1;
					exp++;
				}
				while (realvalue < 1e-8 && exp >= -350) {
					realvalue *= 1e8;
					exp -= 8;
				}
				while (realvalue < 1.0 && exp >= -350) {
					realvalue *= 10.0;
					exp--;
				}
				if (exp > 350 || exp < -350) {
					bufpt = "NaN";
					length = 3;
					break;
				}
			}
			bufpt = buf;
			/*
			 ** If the field type is etGENERIC, then convert to either etEXP
			 ** or etFLOAT, as appropriate.
			 */
			flag_exp = xtype == etEXP;
			if (xtype != etFLOAT) {
				realvalue += rounder;
				if (realvalue >= 10.0) {
					realvalue *= 0.1;
					exp++;
				}
			}
			if (xtype == etGENERIC) {
				flag_rtz = !flag_alternateform;
				if (exp < -4 || exp > precision) {
					xtype = etEXP;
				} else {
					precision = precision - exp;
					xtype = etFLOAT;
				}
			} else {
				flag_rtz = 0;
			}
			if (xtype == etEXP) {
				e2 = 0;
			} else {
				e2 = exp;
			}
			nsd = 0;
			flag_dp = (precision > 0) | flag_alternateform | flag_altform2;
			/* The sign in front of the number */
			if (prefix) {
				*(bufpt++) = prefix;
			}
			/* Digits prior to the decimal point */
			if (e2 < 0) {
				*(bufpt++) = '0';
			} else {
				for (; e2 >= 0; e2--) {
					*(bufpt++) = (char) et_getdigit(&realvalue, &nsd);
				}
			}
			/* The decimal point */
			if (flag_dp) {
				*(bufpt++) = '.';
			}
			/* "0" digits after the decimal point but before the first
			 ** significant digit of the number */
			for (e2++; e2 < 0 && precision > 0; precision--, e2++) {
				*(bufpt++) = '0';
			}
			/* Significant digits after the decimal point */
			while ((precision--) > 0) {
				*(bufpt++) = (char) et_getdigit(&realvalue, &nsd);
			}
			/* Remove trailing zeros and the "." if no digits follow the "." */
			if (flag_rtz && flag_dp) {
				while (bufpt[-1] == '0')
					*(--bufpt) = 0;
				ast_assert(bufpt > buf);
				if (bufpt[-1] == '.') {
					if (flag_altform2) {
						*(bufpt++) = '0';
					} else {
						*(--bufpt) = 0;
					}
				}
			}
			/* Add the "eNNN" suffix */
			if (flag_exp || (xtype == etEXP && exp)) {
				*(bufpt++) = aDigits[infop->charset];
				if (exp < 0) {
					*(bufpt++) = '-';
					exp = -exp;
				} else {
					*(bufpt++) = '+';
				}
				if (exp >= 100) {
					*(bufpt++) = (char) (exp / 100) + '0';	/* 100's digit */
					exp %= 100;
				}
				*(bufpt++) = (char) exp / 10 + '0';	/* 10's digit */
				*(bufpt++) = exp % 10 + '0';	/* 1's digit */
			}
			*bufpt = 0;

			/* The converted number is in buf[] and zero terminated. Output it.
			 ** Note that the number is in the usual order, not reversed as with
			 ** integer conversions. */
			length = (int)(bufpt - buf);
			bufpt = buf;

			/* Special case:  Add leading zeros if the flag_zeropad flag is
			 ** set and we are not left justified */
			if (flag_zeropad && !flag_leftjustify && length < width) {
				int i;
				int nPad = width - length;
				for (i = width; i >= nPad; i--) {
					bufpt[i] = bufpt[i - nPad];
				}
				i = prefix != 0;
				while (nPad--)
					bufpt[i++] = '0';
				length = width;
			}
#endif
			break;
		case etSIZE:
			*(va_arg(ap, int *)) = count;
			length = width = 0;
			break;
		case etPERCENT:
			buf[0] = '%';
			bufpt = buf;
			length = 1;
			break;
		case etCHARLIT:
		case etCHARX:
			c = buf[0] = (char) (xtype == etCHARX ? va_arg(ap, int) : *++fmt);
			if (precision >= 0) {
				for (idx = 1; idx < precision; idx++)
					buf[idx] = (char) c;
				length = precision;
			} else {
				length = 1;
			}
			bufpt = buf;
			break;
		case etSTRING:
		case etDYNSTRING:
			bufpt = va_arg(ap, char *);
			if (bufpt == 0) {
				bufpt = "";
			} else if (xtype == etDYNSTRING) {
				zExtra = bufpt;
			}
			length = (int)strlen(bufpt);
			if (precision >= 0 && precision < length)
				length = precision;
			break;
		case etSQLESCAPE:
		case etSQLESCAPE2:
		case etSQLESCAPE4:
		case etSQLESCAPE3:{
				int i, j, n, ch, isnull;
				int needQuote;
				char *escarg = va_arg(ap, char *);
				isnull = escarg == 0;
				if (isnull)
					escarg = (xtype == etSQLESCAPE2 ? "NULL" : "(NULL)");
				for (i = n = 0; (ch = escarg[i]) != 0; i++) {
					if (ch == '\'' || (xtype == etSQLESCAPE3 && ch == '\\'))
						n++;
				}
				needQuote = !isnull && xtype == etSQLESCAPE2;
				n += i + 1 + needQuote * 2;
				if (n > etBUFSIZE) {
					bufpt = zExtra = malloc(n);
					if (bufpt == 0)
						return -1;
				} else {
					bufpt = buf;
				}
				j = 0;
				if (needQuote)
					bufpt[j++] = '\'';
				for (i = 0; (ch = escarg[i]) != 0; i++) {
					bufpt[j++] = (char) ch;
					if (xtype == etSQLESCAPE4) {
						if (ch == '\'' || (xtype == etSQLESCAPE3 && ch == '\\')) {
							bufpt[j] = (char) ch;
							bufpt[j-1] = (char) '\\';
							j++;
						}
					} else {
						if (ch == '\'' || (xtype == etSQLESCAPE3 && ch == '\\'))
							bufpt[j++] = (char) ch;
					}
				}
				if (needQuote)
					bufpt[j++] = '\'';
				bufpt[j] = 0;
				length = j;
				/* The precision is ignored on %q and %Q */
				/* if ( precision>=0 && precision<length ) length = precision; */
				break;
			}
#ifdef __UNSUPPORTED__
		case etTOKEN:{
				Token *pToken = va_arg(ap, Token *);
				if (pToken && pToken->z) {
					(*func) (arg, (char *) pToken->z, pToken->n);
				}
				length = width = 0;
				break;
			}
		case etSRCLIST:{
				SrcList *pSrc = va_arg(ap, SrcList *);
				int k = va_arg(ap, int);
				struct SrcList_item *pItem = &pSrc->a[k];
				ast_assert(k >= 0 && k < pSrc->nSrc);
				if (pItem->zDatabase && pItem->zDatabase[0]) {
					(*func) (arg, pItem->zDatabase, strlen(pItem->zDatabase));
					(*func) (arg, ".", 1);
				}
				(*func) (arg, pItem->zName, strlen(pItem->zName));
				length = width = 0;
				break;
			}
#endif
		}						/* End switch over the format type */
		/*
		 ** The text of the conversion is pointed to by "bufpt" and is
		 ** "length" characters long.  The field width is "width".  Do
		 ** the output.
		 */
		if (!flag_leftjustify) {
			register int nspace;
			nspace = width - length;
			if (nspace > 0) {
				count += nspace;
				while (nspace >= etSPACESIZE) {
					(*func) (arg, spaces, etSPACESIZE);
					nspace -= etSPACESIZE;
				}
				if (nspace > 0)
					(*func) (arg, spaces, nspace);
			}
		}
		if (length > 0) {
			(*func) (arg, bufpt, length);
			count += length;
		}
		if (flag_leftjustify) {
			register int nspace;
			nspace = width - length;
			if (nspace > 0) {
				count += nspace;
				while (nspace >= etSPACESIZE) {
					(*func) (arg, spaces, etSPACESIZE);
					nspace -= etSPACESIZE;
				}
				if (nspace > 0)
					(*func) (arg, spaces, nspace);
			}
		}
		if (zExtra) {
			free(zExtra);
		}
	}							/* End for loop over the format string */
	return errorflag ? -1 : count;
}								/* End of function */


/* This structure is used to store state information about the
** write to memory that is currently in progress.
*/
struct sgMprintf {
	char *zBase;				/* A base allocation */
	char *zText;				/* The string collected so far */
	int nChar;					/* Length of the string so far */
	int nTotal;					/* Output size if unconstrained */
	int nAlloc;					/* Amount of space allocated in zText */
	void *(*xRealloc) (void *, int);	/* Function used to realloc memory */
};

/*
** This function implements the callback from vxprintf.
**
** This routine add nNewChar characters of text in zNewText to
** the sgMprintf structure pointed to by "arg".
*/
static void mout(void *arg, const char *zNewText, int nNewChar)
{
	struct sgMprintf *pM = (struct sgMprintf *) arg;
	pM->nTotal += nNewChar;
	if (pM->nChar + nNewChar + 1 > pM->nAlloc) {
		if (pM->xRealloc == 0) {
			nNewChar = pM->nAlloc - pM->nChar - 1;
		} else {
			pM->nAlloc = pM->nChar + nNewChar * 2 + 1;
			if (pM->zText == pM->zBase) {
				pM->zText = pM->xRealloc(0, pM->nAlloc);
				if (pM->zText && pM->nChar) {
					memcpy(pM->zText, pM->zBase, pM->nChar);
				}
			} else {
				char *zNew;
				zNew = pM->xRealloc(pM->zText, pM->nAlloc);
				if (zNew) {
					pM->zText = zNew;
				}
			}
		}
	}
	if (pM->zText) {
		if (nNewChar > 0) {
			memcpy(&pM->zText[pM->nChar], zNewText, nNewChar);
			pM->nChar += nNewChar;
		}
		pM->zText[pM->nChar] = 0;
	}
}

/*
** This routine is a wrapper around xprintf() that invokes mout() as
** the consumer.
*/
static char *base_vprintf(void *(*xRealloc) (void *, int),	/* Routine to realloc memory. May be NULL */
						  int useInternal,	/* Use internal %-conversions if true */
						  char *zInitBuf,	/* Initially write here, before mallocing */
						  int nInitBuf,	/* Size of zInitBuf[] */
						  const char *zFormat,	/* format string */
						  va_list ap	/* arguments */
	)
{
	struct sgMprintf sM;
	sM.zBase = sM.zText = zInitBuf;
	sM.nChar = sM.nTotal = 0;
	sM.nAlloc = nInitBuf;
	sM.xRealloc = xRealloc;
	vxprintf(mout, &sM, useInternal, zFormat, ap);
	if (xRealloc) {
		if (sM.zText == sM.zBase) {
			sM.zText = xRealloc(0, sM.nChar + 1);
			if (sM.zText) {
				memcpy(sM.zText, sM.zBase, sM.nChar + 1);
			}
		} else if (sM.nAlloc > sM.nChar + 10) {
			char *zNew = xRealloc(sM.zText, sM.nChar + 1);
			if (zNew) {
				sM.zText = zNew;
			}
		}
	}
	return sM.zText;
}
/*
** Realloc that is a real function, not a macro.
*/
static void *printf_realloc(void *old, int size)
{
	return __ast_realloc(old, size, __FILE__, __LINE__, __FUNCTION__);
}

/*
** Print into memory.  Omit the internal %-conversion extensions.
** Need Free.
*/
char *switch_xml_mprintf(const char *zFormat, ...)
{
	va_list ap;
	char *z;
	char zBase[SWITCH_PRINT_BUF_SIZE];
	va_start(ap, zFormat);
	z = base_vprintf(printf_realloc, 0, zBase, sizeof(zBase), zFormat, ap);
	va_end(ap);
	return z;
}


#ifdef FOR_DEMO
static void recovery_test(void)
{
	ast_log(LOG_WARNING, "====enter recovery_test======");
	switch_xml_t variables, cdr, x_main_cp, x_caller_profile, x_caller_extension, x_times, time_tag,
		x_application, x_callflow, x_inner_extension, x_apps, x_o, x_channel_data, x_field, xhr, x_hold;
	int cdr_off = 0, cd_off = 0;
	char *xml_cdr_text = NULL;

	cdr = switch_xml_new("cdr");
	switch_xml_set_attr_d(cdr, "core-uuid", "core-uuid_attr");
	switch_xml_set_attr_d(cdr, "switchname", "switchname_attr");

	if (!(x_channel_data = switch_xml_add_child_d(cdr, "channel_data", cdr_off++))) {
		goto error;
	}

	x_field = switch_xml_add_child_d(x_channel_data, "state", cd_off++);
	switch_xml_set_txt_d(x_field, "state_txt");

	x_field = switch_xml_add_child_d(x_channel_data, "direction", cd_off++);
	switch_xml_set_txt_d(x_field, "direction_txt");

	x_field = switch_xml_add_child_d(x_channel_data, "textlog", cd_off++);
	switch_xml_set_txt_d(x_field, "textlog_txt");

	x_field = switch_xml_add_child_d(x_channel_data, "state_number", cd_off++);
	switch_xml_set_txt_d(x_field, "state_number_txt");

	x_field = switch_xml_add_child_d(x_channel_data, "flags", cd_off++);
	switch_xml_set_txt_d(x_field, "flags_txt");

	x_field = switch_xml_add_child_d(x_channel_data, "caps", cd_off++);
	switch_xml_set_txt_d(x_field, "caps_txt");

	if (!(variables = switch_xml_add_child_d(cdr, "call-stats", cdr_off++))) {
		goto error;
	}

	if (!(variables = switch_xml_add_child_d(cdr, "variables", cdr_off++))) {
		goto error;
	}

	if (1) {
		int app_off = 0;

		if (!(x_apps = switch_xml_add_child_d(cdr, "app_log", cdr_off++))) {
			goto error;
		}

		if (!(x_application = switch_xml_add_child_d(x_apps, "application", app_off++))) {
			goto error;
		}

		switch_xml_set_attr_d(x_application, "app_name", "app_name_attr");
		switch_xml_set_attr_d(x_application, "app_data", "app_data_attr");
		switch_xml_set_attr_d_buf(x_application, "app_stamp", "app_stamp_attr");
	}

	if (1) {
		int cf_off = 0;

		if (!(xhr = switch_xml_add_child_d(cdr, "hold-record", cdr_off++))) {
			goto error;
		}

		if (!(x_hold = switch_xml_add_child_d(xhr, "hold", cf_off++))) {
			goto error;
		}

		switch_xml_set_attr_d(x_hold, "on", "on_attr");
		switch_xml_set_attr_d(x_hold, "off", "off_attr");
		switch_xml_set_attr_d(x_hold, "bridged-to", "bridged-to_attr");
	}

	if (1) {
		int cf_off = 0;
		int cp_off = 0;

		if (!(x_callflow = switch_xml_add_child_d(cdr, "callflow", cdr_off++))) {
			goto error;
		}

		switch_xml_set_attr_d(x_callflow, "dialplan", "dialplan_attr");
		switch_xml_set_attr_d(x_callflow, "unique-id", "unique-id_attr");
		switch_xml_set_attr_d(x_callflow, "clone-of", "clone-of_attr");
		switch_xml_set_attr_d(x_callflow, "profile_index", "profile_index_attr");

		int app_off = 0;

		if (!(x_caller_extension = switch_xml_add_child_d(x_callflow, "extension", cf_off++))) {
			goto error;
		}

		switch_xml_set_attr_d(x_caller_extension, "name", "name_attr");
		switch_xml_set_attr_d(x_caller_extension, "number", "number_attr");
		switch_xml_set_attr_d(x_caller_extension, "current_app", "current_app_attr");

		if (!(x_application = switch_xml_add_child_d(x_caller_extension, "application", app_off++))) {
			goto error;
		}
		switch_xml_set_attr_d(x_application, "last_executed", "true");

		switch_xml_set_attr_d(x_application, "app_name", "app_name_attr");
		switch_xml_set_attr_d(x_application, "app_data", "app_data_attr");

		int i_off = 0, i_app_off = 0;

		if (!(x_inner_extension = switch_xml_add_child_d(x_caller_extension, "sub_extensions", app_off++))) {
			goto error;
		}

		if (!(x_caller_extension = switch_xml_add_child_d(x_inner_extension, "extension", i_off++))) {
			goto error;
		}
		switch_xml_set_attr_d(x_caller_extension, "name", "name_attr");
		switch_xml_set_attr_d(x_caller_extension, "number", "number_attr");
		switch_xml_set_attr_d(x_caller_extension, "dialplan", "dialplan_attr");
		switch_xml_set_attr_d(x_caller_extension, "current_app", "current_app_attr");

		if (!(x_application = switch_xml_add_child_d(x_caller_extension, "application", i_app_off++))) {
			goto error;
		}
		switch_xml_set_attr_d(x_application, "last_executed", "true");
		switch_xml_set_attr_d(x_application, "app_name", "app_name_attr");
		switch_xml_set_attr_d(x_application, "app_data", "app_data_attr");

		if (!(x_main_cp = switch_xml_add_child_d(x_callflow, "caller_profile", cf_off++))) {
			goto error;
		}


		int off = 0;
		if (!(x_o = switch_xml_add_child_d(x_main_cp, "origination", cp_off++))) {
			goto error;
		}

		if (!(x_caller_profile = switch_xml_add_child_d(x_o, "origination_caller_profile", off++))) {
			goto error;
		}

		if (!(x_o = switch_xml_add_child_d(x_main_cp, "originator", cp_off++))) {
			goto error;
		}

		if (!(x_caller_profile = switch_xml_add_child_d(x_o, "originator_caller_profile", off++))) {
			goto error;
		}

		if (!(x_o = switch_xml_add_child_d(x_main_cp, "originatee", cp_off++))) {
			goto error;
		}
		if (!(x_caller_profile = switch_xml_add_child_d(x_o, "originatee_caller_profile", off++))) {
			goto error;
		}

		int t_off = 0;
		if (!(x_times = switch_xml_add_child_d(x_callflow, "times", cf_off++))) {
			goto error;
		}
		if (!(time_tag = switch_xml_add_child_d(x_times, "created_time", t_off++))) {
			goto error;
		}

		if (!(time_tag = switch_xml_add_child_d(x_times, "profile_created_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "progress_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");


		if (!(time_tag = switch_xml_add_child_d(x_times, "progress_media_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "answered_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "bridged_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "last_hold_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "hold_accum_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "hangup_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "resurrect_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

		if (!(time_tag = switch_xml_add_child_d(x_times, "transfer_time", t_off++))) {
			goto error;
		}
		switch_xml_set_txt_d(time_tag, "time_tag_txt");

	}

	if (cdr) {
		xml_cdr_text = switch_xml_toxml_nolock(cdr, SWITCH_FALSE);
		ast_log(LOG_WARNING, "%s", xml_cdr_text);
		char *sql = NULL;
		sql = switch_xml_mprintf("insert into recovery (runtime_uuid, technology, profile_name, hostname, uuid, metadata) "
								 "values ('test','test','test','test','test','%q')", xml_cdr_text);
		if (sql) {
			ast_log(LOG_WARNING, "%s", sql);
			ast_recovery_db_put(sql, 1, 1);
			ast_free(sql);
		}
	}

	// Get one data from database
	char buf[4096] = {0};
	ast_recovery_db_get("test1", "test5", buf, sizeof(buf)-1);
	ast_log(LOG_WARNING, "%s", buf);

	// restore database to xml struct, need free
	switch_xml_t restore_xml = switch_xml_parse_str_dynamic(buf, SWITCH_TRUE);
	// Get child txt
	switch_xml_t channel_data = switch_xml_child(restore_xml, "channel_data");
	switch_xml_t state = switch_xml_child(channel_data, "state");
	ast_log(LOG_WARNING, "====%s====", state->txt);

	if (restore_xml) {
		switch_xml_free(restore_xml);
	}

	// Get all data from database
	struct recovery_db_entry* entry = NULL;
	struct recovery_db_entry* inbound_mode_sub_entry = NULL;
	inbound_mode_sub_entry = ast_recovery_db_get_all();
	int i = 0;
	for (entry = inbound_mode_sub_entry; entry; entry = entry->next) {
		ast_log(LOG_WARNING, "!!!!!!!!!!!!!=============== [%d] ===============!!!!!!!!!!!!!\n data [%s]", i, entry->data);
		i++;
	}
	if (inbound_mode_sub_entry) {
		ast_recovery_freetree(inbound_mode_sub_entry);
	}

error:
	if (cdr) {
		switch_xml_free(cdr);
	}

	if (xml_cdr_text) {
		ast_free(xml_cdr_text);
	}

	return;
}
#endif
/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 noet:
 */
