#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2023, Grandstream.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>

#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/utils.h"
#include "asterisk/astobj2.h"
#include "asterisk/avs_controller.h"

AST_MODULE_LOG("RTP");

static const char config[] = "avs_resource.conf";

#define MINIMUM_MULTICAST_PORT 1024 	/*!< Minimum port number to accept */
#define MAXIMUM_MULTICAST_PORT 65535	/*!< Maximum port number to accept */

static int avs_resource_reload(int reload)
{
	struct ast_config *cfg = NULL;
	const char *s;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	int avs_master = 0;
	char avs_multicast_address[MAX_IPADDR_V6_LEN] = {0};
	int avs_multicast_port_start = 10000;
	int avs_multicast_port_end = 10300;
	char avs_multicast_network_ip[MAX_IPADDR_V6_LEN] = {0};

	cfg = ast_config_load(config, config_flags);
	if (!cfg) {
		ast_log(LOG_NOTICE, "Unable to load config %s\n", config);
		return -1;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file %s is in an invalid format.  Aborting.\n", config);
		return -1;
	}

	if ((s = ast_variable_retrieve(cfg, "general", "avs_master"))) {
		avs_master = ast_true(s);
		ast_set_avs_master_value(avs_master);
	}

	if ((s = ast_variable_retrieve(cfg, "general", "avs_multicast_address"))) {
		ast_copy_string(avs_multicast_address, s, sizeof(avs_multicast_address));
		ast_set_avs_multicast_address(avs_multicast_address);
	}

	if ((s = ast_variable_retrieve(cfg, "general", "avs_multicast_port_start"))) {
		avs_multicast_port_start = atoi(s);
		if (avs_multicast_port_start < MINIMUM_MULTICAST_PORT) {
			avs_multicast_port_start = MINIMUM_MULTICAST_PORT;
		}
		if (avs_multicast_port_start > MAXIMUM_MULTICAST_PORT) {
			avs_multicast_port_start = MAXIMUM_MULTICAST_PORT;
		}
		ast_set_avs_multicast_start_port(avs_multicast_port_start);
	}

	if ((s = ast_variable_retrieve(cfg, "general", "avs_multicast_port_end"))) {
		avs_multicast_port_end= atoi(s);
		if (avs_multicast_port_end < MINIMUM_MULTICAST_PORT) {
			avs_multicast_port_end = MINIMUM_MULTICAST_PORT;
		}
		if (avs_multicast_port_end > MAXIMUM_MULTICAST_PORT) {
			avs_multicast_port_end = MAXIMUM_MULTICAST_PORT;
		}
		ast_set_avs_multicast_end_port(avs_multicast_port_end );
	}

	if (avs_master) {
		ast_init_avs_address();
	}

	if ((s = ast_variable_retrieve(cfg, "general", "avs_multicast_network_ip"))) {
		ast_copy_string(avs_multicast_network_ip, s, sizeof(avs_multicast_address));
		ast_set_avs_multicast_network_ip(avs_multicast_network_ip);
	}

	ast_config_destroy(cfg);
	return 0;
}

static int reload_module(void)
{
	avs_resource_reload(1);

	return 0;
}

static int load_module(void)
{
	avs_resource_reload(0);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "AVS resource",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND,
#ifdef HAVE_PJPROJECT
	.requires = "res_pjproject",
#endif
);

#endif

