#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2020, Grandstream.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>

#include "asterisk/module.h"
#include "asterisk/bfcp.h"
#include "asterisk/lock.h"
#include "asterisk/utils.h"
#include "asterisk/astobj2.h"
#include "asterisk/res_pjsip.h"

#include <pjlib.h>
#include <pjlib-util.h>
#include <pjnath.h>
#include <ifaddrs.h>
#include "asterisk/res_pjproject.h"

AST_MODULE_LOG("res_bfcp");

#define BFCP_DEFAULT_TURN_PORT 				3478
#define BFCP_TURN_STATE_WAIT_TIME 			2000
#define BFCP_PJ_IOQUEUE_MAX_HANDLES  		64

#if 0
/* Retain for query DNS in sched to speed up call setup time */
#define BFCP_DNS_FIRST_REFRESH_INTERVAL   	1
#define BFCP_DNS_REFRESH_INTERVAL   			300
static struct ast_sched_context *bfcp_sched;
static int bfcp_turn_sched_id = -1; 
static pj_str_t bfcp_turn_address[2];
static int bfcp_turn_port = BFCP_DEFAULT_TURN_PORT;
static pj_pool_t *pool;
#endif

/*! \brief Pool factory used by pjlib to allocate memory. */
static pj_caching_pool cachingpool;

/*! \brief List of ioqueue threads */
static AST_LIST_HEAD_STATIC(ioqueues, ast_bfcp_ioqueue_thread);

static int udp_reachable = 1;

struct ast_bfcp_ioqueue_thread {
	/*! \brief Pool used by the thread */
	pj_pool_t *pool;
	/*! \brief The thread handling the queue and timer heap */
	pj_thread_t *thread;
	/*! \brief Ioqueue which polls on sockets */
	pj_ioqueue_t *ioqueue;
	/*! \brief Timer heap for scheduled items */
	pj_timer_heap_t *timerheap;
	/*! \brief Termination request */
	int terminate;
	/*! \brief Current number of descriptors being waited on */
	unsigned int count;
	/*! \brief Linked list information */
	AST_LIST_ENTRY(ast_bfcp_ioqueue_thread) next;
};

static int ast_bfcp_new(struct ast_bfcp_udp *bfcp_udp);
static int ast_bfcp_destory(struct ast_bfcp_udp *bfcp_udp);
static int ast_bfcp_get_turn_relay_address(struct ast_bfcp_udp *bfcp_udp);
static int ast_bfcp_turn_packet_send(struct ast_bfcp_udp *bfcp_udp, void *pkt, int size);
static int ast_bfcp_create_turn_permission(struct ast_bfcp_udp *bfcp_udp, const char *peer_address);
static int ast_bfcp_set_peer_address(struct ast_bfcp_udp *bfcp_udp, struct ast_sockaddr *remote_addr);
static void ast_bfcp_set_udp_reachable(int value);

struct ast_bfcp {
	ast_cond_t cond;               				/*!< Condition for signaling */
	struct ast_sockaddr bfcp_loop;	 			/*!< Loopback address for forwarding BFCP message from TURN */
	struct ast_bfcp_ioqueue_thread *ioqueue;		/*!< The ioqueue thread handling us */
	pj_turn_sock *turn_socket;
	pj_turn_state_t turn_state;
	pj_sockaddr peer_addr;
};

/* BFCP Engine Declaration */
static struct ast_bfcp_engine asterisk_bfcp_engine = {
	.name = "asterisk",
	.new = ast_bfcp_new,
	.destory = ast_bfcp_destory,
	.get_turn_relay_address = ast_bfcp_get_turn_relay_address,
	.turn_packet_send = ast_bfcp_turn_packet_send,
	.create_turn_permission = ast_bfcp_create_turn_permission,
	.set_peer_address = ast_bfcp_set_peer_address,
	.set_udp_reachable = ast_bfcp_set_udp_reachable,
};

#if 0
/* Retain for query DNS in sched to speed up call setup time */
static int bfcp_turn_addr_refresh_timeout(const void *data)
{
	struct ast_sockaddr addr;
	char *turn_addr_str = NULL;

	turn_addr_str = ast_sip_get_turn_server_address();
	if (!turn_addr_str) {
		ast_log(LOG_DEBUG, "Can't get turn server address for BFCP module\n");
		return 0;
	}

	memset(&addr, 0, sizeof(addr));
	if (!ast_sockaddr_parse(&addr, turn_addr_str, PARSE_PORT_FORBID)) {
		ast_log(LOG_WARNING, "Invalid IP address: %s\n", turn_addr_str);
		ast_free(turn_addr_str);
		turn_addr_str = NULL;
		return 0;
	}

	ast_sockaddr_set_port(&addr, BFCP_DEFAULT_TURN_PORT);
	pj_strdup2_with_null(pool, &bfcp_turn_address[0], ast_sockaddr_stringify_addr(&addr));
	bfcp_turn_port = ast_sockaddr_port(&addr);

	ast_log(LOG_NOTICE, "BFCP get turn address:%.*s:%d\n", (int)bfcp_turn_address[0].slen, bfcp_turn_address[0].ptr, bfcp_turn_port);

	ast_free(turn_addr_str);
	turn_addr_str = NULL;

	return BFCP_DNS_REFRESH_INTERVAL*1000;
}

static int sched_get_turn_address(void)
{
	char *turn_addr_str = NULL;

	turn_addr_str = ast_sip_get_turn_server_address();
	if (!turn_addr_str) {
		ast_log(LOG_DEBUG, "Can't get turn server address for BFCP module\n");
		return -1;
	}

	if (!ast_sched_find_data(bfcp_sched, bfcp_turn_sched_id)) {
		ast_log(LOG_NOTICE, "Add query turn address dns sched, turn_sched_id:%d, turn_addr_str:%s\n", bfcp_turn_sched_id, turn_addr_str);
		bfcp_turn_sched_id = ast_sched_add(bfcp_sched, BFCP_DNS_FIRST_REFRESH_INTERVAL*1000, bfcp_turn_addr_refresh_timeout, turn_addr_str);
	} else {
		ast_log(LOG_NOTICE, "Don't add query turn address dns sched, already scheded, turn_sched_id:%d\n", bfcp_turn_sched_id);
	}

	ast_free(turn_addr_str);
	turn_addr_str = NULL;

	return 0;
}
#endif

AST_THREADSTORAGE(pj_thread_storage);

/*! \brief Function used to check if the calling thread is registered with pjlib. If it is not it will be registered. */
static void pj_thread_register_check(void)
{
	pj_thread_desc *desc;
	pj_thread_t *thread;

	if (pj_thread_is_registered() == PJ_TRUE) {
		return;
	}

	desc = ast_threadstorage_get(&pj_thread_storage, sizeof(pj_thread_desc));
	if (!desc) {
		ast_log(LOG_ERROR, "Could not get thread desc from thread-local storage. Expect awful things to occur\n");
		return;
	}
	pj_bzero(*desc, sizeof(*desc));

	if (pj_thread_register("Asterisk Thread", *desc, &thread) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Coudln't register thread with PJLIB.\n");
	}

	return;
}

static void bfcp_terminate_pjproject(void)
{
	pj_thread_register_check();

	ast_pjproject_caching_pool_destroy(&cachingpool);
	pj_shutdown();

	return;
}

/* PJPROJECT TURN callback */
static void ast_bfcp_on_turn_rx_data(pj_turn_sock *turn_sock, void *pkt,
	unsigned pkt_len, const pj_sockaddr_t *peer_addr, unsigned addr_len)
{
	struct ast_bfcp_udp *bfcp_udp = NULL;
	struct ast_bfcp *bfcp = NULL;
	char new_peer_addr[PJ_INET6_ADDRSTRLEN + 2] = { 0 };
	char old_peer_addr[PJ_INET6_ADDRSTRLEN + 2] = { 0 };

	bfcp_udp = pj_turn_sock_get_user_data(turn_sock);
	if (!bfcp_udp) {
		return;
	}

	bfcp = ast_bfcp_get_data(bfcp_udp);
	if (!bfcp) {
		return;
	}

	pj_sockaddr_print(peer_addr, new_peer_addr, sizeof(new_peer_addr), 0);
	ast_log(LOG_DEBUG, "[BFCP:%p] Receive data len:%d from turn, peer address:%s port:%d\n",
		bfcp_udp, pkt_len, new_peer_addr, pj_sockaddr_get_port(peer_addr));

	if (pj_sockaddr_cmp(peer_addr, &bfcp->peer_addr)) {
		pj_sockaddr_print(&bfcp->peer_addr, old_peer_addr, sizeof(old_peer_addr), 0);
		ast_log(LOG_NOTICE, "[BFCP:%p] New peer address:%s port:%d is not the same as old address:%s port:%d, create new permission\n",
			bfcp_udp, new_peer_addr, pj_sockaddr_get_port(peer_addr), old_peer_addr, pj_sockaddr_get_port(&bfcp->peer_addr));

		pj_turn_sock_set_perm(bfcp->turn_socket, 1, peer_addr, 1);
		pj_sockaddr_cp(&bfcp->peer_addr, peer_addr);
	}

	ast_sendto(bfcp_udp->fd, pkt, pkt_len, 0, &bfcp->bfcp_loop);

	return;
}

/* PJPROJECT TURN callback */
static void ast_bfcp_on_turn_state(pj_turn_sock *turn_sock, pj_turn_state_t old_state, pj_turn_state_t new_state)
{
	struct ast_bfcp_udp *bfcp_udp = pj_turn_sock_get_user_data(turn_sock);
	struct ast_bfcp *bfcp = NULL;

	if (!bfcp_udp) {
		return;
	}

	bfcp = ast_bfcp_get_data(bfcp_udp);
	if (!bfcp) {
		return;
	}

	ast_log(LOG_DEBUG, "[BFCP:%p] ast_bfcp_on_turn_state new_state:%d old_state:%d\n", bfcp_udp, new_state, old_state);

	ao2_lock(bfcp);

	/* We store the new state so the other thread can actually handle it */
	bfcp->turn_state = new_state;
	ast_cond_signal(&bfcp->cond);

	if (new_state == PJ_TURN_STATE_DESTROYING) {
		pj_turn_sock_set_user_data(bfcp->turn_socket, NULL);
		bfcp->turn_socket = NULL;
	}

	ao2_unlock(bfcp);

	return;
}

/* BFCP TURN Socket interface declaration */
static pj_turn_sock_cb ast_bfcp_turn_sock_cb = {
	.on_rx_data = ast_bfcp_on_turn_rx_data,
	.on_state = ast_bfcp_on_turn_state,
};

/*! \brief Worker thread for ioqueue and timerheap */
static int bfcp_ioqueue_worker_thread(void *data)
{
	struct ast_bfcp_ioqueue_thread *ioqueue = data;

	while (!ioqueue->terminate) {
		const pj_time_val delay = {0, 10};

		pj_ioqueue_poll(ioqueue->ioqueue, &delay);

		pj_timer_heap_poll(ioqueue->timerheap, NULL);
	}

	return 0;
}

/*! \brief Destroyer for ioqueue thread */
static void bfcp_ioqueue_thread_destroy(struct ast_bfcp_ioqueue_thread *ioqueue)
{
	if (ioqueue->thread) {
		ioqueue->terminate = 1;
		pj_thread_join(ioqueue->thread);
		pj_thread_destroy(ioqueue->thread);
	}

	if (ioqueue->pool) {
		/* This mimics the behavior of pj_pool_safe_release
		 * which was introduced in pjproject 2.6.
		 */
		pj_pool_t *temp_pool = ioqueue->pool;

		ioqueue->pool = NULL;
		pj_pool_release(temp_pool);
	}

	ast_free(ioqueue);

	return;
}

/*! \brief Removal function for ioqueue thread, determines if it should be terminated and destroyed */
static void bfcp_ioqueue_thread_remove(struct ast_bfcp_ioqueue_thread *ioqueue)
{
	int destroy = 0;

	/* If nothing is using this ioqueue thread destroy it */
	AST_LIST_LOCK(&ioqueues);
	if ((ioqueue->count - 1) == 0) {
		ast_log(LOG_DEBUG, "BFCP ioqueues decreased to zero, remove it\n");
		destroy = 1;
		AST_LIST_REMOVE(&ioqueues, ioqueue, next);
	}
	AST_LIST_UNLOCK(&ioqueues);

	if (!destroy) {
		return;
	}

	bfcp_ioqueue_thread_destroy(ioqueue);

	return;
}

/*! \brief Finder and allocator for an ioqueue thread */
static struct ast_bfcp_ioqueue_thread *bfcp_ioqueue_thread_get_or_create(void)
{
	struct ast_bfcp_ioqueue_thread *ioqueue;
	pj_lock_t *lock;

	AST_LIST_LOCK(&ioqueues);

	/* See if an ioqueue thread exists that can handle more */
	AST_LIST_TRAVERSE(&ioqueues, ioqueue, next) {
		if ((ioqueue->count + 1) < BFCP_PJ_IOQUEUE_MAX_HANDLES) {
			break;
		}
	}

	/* If we found one bump it up and return it */
	if (ioqueue) {
		ioqueue->count += 1;
		goto end;
	}

	ioqueue = ast_calloc(1, sizeof(*ioqueue));
	if (!ioqueue) {
		goto end;
	}

	ioqueue->pool = pj_pool_create(&cachingpool.factory, "bfcp", 512, 512, NULL);

	/* We use a timer on the ioqueue thread for TURN so that two threads aren't operating
	 * on a session at the same time
	 */
	if (pj_timer_heap_create(ioqueue->pool, 4, &ioqueue->timerheap) != PJ_SUCCESS) {
		goto fatal;
	}

	if (pj_lock_create_recursive_mutex(ioqueue->pool, "bfcp%p", &lock) != PJ_SUCCESS) {
		goto fatal;
	}

	pj_timer_heap_set_lock(ioqueue->timerheap, lock, PJ_TRUE);

	if (pj_ioqueue_create(ioqueue->pool, BFCP_PJ_IOQUEUE_MAX_HANDLES, &ioqueue->ioqueue) != PJ_SUCCESS) {
		goto fatal;
	}

	if (pj_thread_create(ioqueue->pool, "bfcp", &bfcp_ioqueue_worker_thread, ioqueue, 0, 0, &ioqueue->thread) != PJ_SUCCESS) {
		goto fatal;
	}

	AST_LIST_INSERT_HEAD(&ioqueues, ioqueue, next);

	ioqueue->count = 1;

	goto end;

fatal:
	bfcp_ioqueue_thread_destroy(ioqueue);
	ioqueue = NULL;

end:
	AST_LIST_UNLOCK(&ioqueues);
	return ioqueue;
}

static int ast_bfcp_turn_packet_send(struct ast_bfcp_udp *bfcp_udp, void *pkt, int size)
{
	struct ast_bfcp *bfcp = ast_bfcp_get_data(bfcp_udp);

	if (!bfcp) {
		return -1;
	}

	ast_log(LOG_DEBUG, "[BFCP:%p] ast_bfcp_turn_packet_send pkt size:%d\n", bfcp_udp, size);

	ao2_lock(bfcp);

	if (!bfcp->turn_socket) {
		ao2_unlock(bfcp);
		return -1;
	}

	if (bfcp->turn_state != PJ_TURN_STATE_READY) {
		ast_log(LOG_ERROR, "[BFCP:%p] Turn state is not ready, can't send packet\n", bfcp_udp);
		ao2_unlock(bfcp);
		return -1;
	}

	pj_turn_sock_sendto(bfcp->turn_socket, pkt, size, &bfcp->peer_addr, pj_sockaddr_get_len(&bfcp->peer_addr));

	ao2_unlock(bfcp);

	return 0;
}

static int ast_bfcp_set_peer_address(struct ast_bfcp_udp *bfcp_udp, struct ast_sockaddr *remote_addr)
{
	pj_str_t address;
	pj_sockaddr peer_addr;
	struct ast_bfcp *bfcp = ast_bfcp_get_data(bfcp_udp);

	if (!bfcp || !remote_addr) {
		return -1;
	}

	if (pj_sockaddr_has_addr(&bfcp->peer_addr)) {
		ast_log(LOG_NOTICE, "[BFCP:%p] BFCP already has peer address\n", bfcp_udp);
		return 0;
	}

	ao2_lock(bfcp);

	if (!bfcp->turn_socket) {
		ao2_unlock(bfcp);
		return -1;
	}

	if (bfcp->turn_state != PJ_TURN_STATE_READY) {
		ast_log(LOG_ERROR, "[BFCP:%p] Turn state is not ready, can't send packet\n", bfcp_udp);
		ao2_unlock(bfcp);
		return -1;
	}

	ast_log(LOG_VERBOSE, "[BFCP:%p] ast_bfcp_set_peer_address remote_addr:%s\n", bfcp_udp, ast_sockaddr_stringify(remote_addr));

	pj_sockaddr_parse(pj_AF_UNSPEC(), 0, pj_cstr(&address, ast_sockaddr_stringify(remote_addr)), &peer_addr);
	pj_sockaddr_set_port(&peer_addr, ast_sockaddr_port(remote_addr));
	pj_sockaddr_cp(&bfcp->peer_addr, &peer_addr);
	pj_turn_sock_set_perm(bfcp->turn_socket, 1, &peer_addr, 1);

	ao2_unlock(bfcp);

	return 0;
}

static int ast_bfcp_create_turn_permission(struct ast_bfcp_udp *bfcp_udp, const char *peer_address)
{
	pj_sockaddr permission_addr;
	pj_str_t		permission_addr_str;
	struct ast_bfcp *bfcp = ast_bfcp_get_data(bfcp_udp);

	if (!bfcp) {
		return -1;
	}

	ao2_lock(bfcp);

	if (!bfcp->turn_socket) {
		ao2_unlock(bfcp);
		return -1;
	}

	if (bfcp->turn_state != PJ_TURN_STATE_READY) {
		ao2_unlock(bfcp);
		ast_log(LOG_ERROR, "[BFCP:%p] Turn state is not ready, can't create turn permission\n", bfcp_udp);
		return -1;
	}

	ast_log(LOG_VERBOSE, "[BFCP:%p] Create turn permission for peer address:%s\n", bfcp_udp, peer_address);

	pj_strset2(&permission_addr_str, (char *)peer_address);
	pj_sockaddr_parse(pj_AF_UNSPEC(), 0, &permission_addr_str, &permission_addr);
	pj_sockaddr_set_port(&permission_addr, 6666);
	pj_turn_sock_set_perm(bfcp->turn_socket, 1, &permission_addr, 1);

	ao2_unlock(bfcp);

	return 0;
}

static void bfcp_set_relay_address(struct ast_bfcp_udp *bfcp_udp, struct ast_bfcp *bfcp, pj_turn_session_info info)
{
	char address[PJ_INET6_ADDRSTRLEN + 2] = { 0 };

	if (!bfcp_udp || !bfcp) {
		return;
	}

	strncpy(bfcp_udp->turn_relay_address, pj_sockaddr_print(&info.relay_addr, address, sizeof(address), 0), sizeof(bfcp_udp->turn_relay_address) - 1);
	bfcp_udp->turn_relay_port = pj_sockaddr_get_port(&info.relay_addr);

	ast_log(LOG_VERBOSE, "[BFCP:%p] Get turn relay address:%s, relay port:%d\n",
		bfcp_udp, bfcp_udp->turn_relay_address, bfcp_udp->turn_relay_port);

	memset(address, 0, sizeof(address));
	ast_log(LOG_VERBOSE, "[BFCP:%p] Get mapped address:%s, port:%d\n",
		bfcp_udp, pj_sockaddr_print(&info.mapped_addr, address, sizeof(address), 0), pj_sockaddr_get_port(&info.mapped_addr));

	return;
}

static void bfcp_set_loop_address(struct ast_bfcp_udp *bfcp_udp, struct ast_bfcp *bfcp)
{
	struct ast_sockaddr local;
	struct ast_sockaddr loop;

	if (!bfcp_udp || !bfcp) {
		return;
	}

	ast_bfcp_get_us(bfcp_udp, &local);
	if (ast_sockaddr_is_ipv4(&local)) {
		ast_sockaddr_parse(&loop, "127.0.0.1", PARSE_PORT_FORBID);
	} else {
		ast_sockaddr_parse(&loop, "::1", PARSE_PORT_FORBID);
	}
	ast_sockaddr_set_port(&loop, ast_sockaddr_port(&local));

	ast_sockaddr_copy(&bfcp->bfcp_loop, &loop);

	return;
}

static void bfcp_turn_request(struct ast_bfcp_udp *bfcp_udp,
		enum ast_transport transport, const char *server, unsigned int port, const char *username, const char *password)
{
	struct ast_bfcp *bfcp = NULL;
	pj_turn_sock **turn_sock;
	const pj_turn_sock_cb *turn_cb;
	pj_turn_tp_type conn_type;
	pj_stun_auth_cred cred = { 0, };
	pj_str_t turn_addr;
	RAII_VAR(struct ast_sockaddr *, addr, NULL, ast_free);
	pj_stun_config stun_config;
	struct timeval wait = ast_tvadd(ast_tvnow(), ast_samp2tv(BFCP_TURN_STATE_WAIT_TIME, 1000));
	struct timespec ts = { .tv_sec = wait.tv_sec, .tv_nsec = wait.tv_usec * 1000, };
	pj_turn_session_info info;
	pj_status_t status;
	pj_turn_sock_cfg turn_sock_cfg;
	char *server_ip = NULL;
	int num_addrs = 0;

	ast_log(LOG_DEBUG, "[BFCP:%p] Start to turn request for BFCP\n", bfcp_udp);

	bfcp = ast_bfcp_get_data(bfcp_udp);
	if (!bfcp) {
		ast_log(LOG_WARNING, "[BFCP:%p] Can't get bfcp data\n", bfcp_udp);
		return;
	}

	turn_sock = &bfcp->turn_socket;
	turn_cb = &ast_bfcp_turn_sock_cb;

	if (transport == AST_TRANSPORT_UDP) {
		conn_type = PJ_TURN_TP_UDP;
	} else if (transport == AST_TRANSPORT_TCP) {
		conn_type = PJ_TURN_TP_TCP;
	} else {
		ast_assert(0);
		return;
	}

	num_addrs = ast_sockaddr_resolve(&addr, server, PARSE_PORT_FORBID, AST_AF_UNSPEC);
	if (num_addrs) {
		server_ip = ast_sockaddr_stringify_fmt(addr, AST_SOCKADDR_STR_ADDR);
		if (ast_strlen_zero(server_ip)) {
			ast_log(LOG_WARNING, "[BFCP:%p] Turn server IP is invalid\n", bfcp_udp);
			return;
		}
		ast_log(LOG_DEBUG, "[BFCP:%p] Get turn server IP:%s\n", bfcp_udp, server_ip);
	} else {
		ast_log(LOG_WARNING, "[BFCP:%p] Can't get turn server IP\n", bfcp_udp);
		return;
	}

	if (*turn_sock) {
		ast_log(LOG_VERBOSE, "[BFCP:%p] Already has turn sock, don't process again\n", bfcp_udp);
		return;
	}

	if (!bfcp->ioqueue) {
		bfcp->ioqueue = bfcp_ioqueue_thread_get_or_create();
		if (!bfcp->ioqueue) {
			ast_log(LOG_WARNING, "[BFCP:%p] Can't get or create ioqueue\n", bfcp_udp);
			return;
		}
	}

	pj_stun_config_init(&stun_config, &cachingpool.factory, 0, bfcp->ioqueue->ioqueue, bfcp->ioqueue->timerheap);

	/* Use ICE session group lock for TURN session to avoid deadlock */
	pj_turn_sock_cfg_default(&turn_sock_cfg);

	status = pj_turn_sock_create(&stun_config, ast_sockaddr_is_ipv4(addr) ? pj_AF_INET() : pj_AF_INET6(),
		conn_type, turn_cb, &turn_sock_cfg, bfcp_udp, turn_sock);
	if (status != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "[BFCP:%p] Could not create a TURN client socket\n", bfcp_udp);
		return;
	}

	cred.type = PJ_STUN_AUTH_CRED_STATIC;
	pj_strset2(&cred.data.static_cred.username, (char*)username);
	cred.data.static_cred.data_type = PJ_STUN_PASSWD_PLAIN;
	pj_strset2(&cred.data.static_cred.data, (char*)password);

	pj_turn_sock_alloc(*turn_sock, pj_cstr(&turn_addr, server_ip), port, NULL, &cred, NULL);

	/*
	 * Because the TURN socket is asynchronous and we are synchronous we need to
	 * wait until it is done
	 */
	ao2_lock(bfcp);
	while (bfcp->turn_state < PJ_TURN_STATE_READY) {
		ast_cond_timedwait(&bfcp->cond, ao2_object_get_lockaddr(bfcp), &ts);
	}
	ao2_unlock(bfcp);

	if (bfcp->turn_state != PJ_TURN_STATE_READY) {
		ast_log(LOG_WARNING, "[BFCP:%p] TURN state is not ready\n", bfcp_udp);
		return;
	}

	pj_turn_sock_get_info(*turn_sock, &info);

	bfcp_set_relay_address(bfcp_udp, bfcp, info);

	bfcp_set_loop_address(bfcp_udp, bfcp);

	return;
}

static void bfcp_destructor(void *obj)
{
	struct ast_bfcp *bfcp = obj;
	struct timeval wait = ast_tvadd(ast_tvnow(), ast_samp2tv(BFCP_TURN_STATE_WAIT_TIME, 1000));
	struct timespec ts = { .tv_sec = wait.tv_sec, .tv_nsec = wait.tv_usec * 1000, };

	if (!bfcp) {
		return;
	}

	pj_thread_register_check();

	if (bfcp->turn_socket) {
		bfcp->turn_state = PJ_TURN_STATE_NULL;

		pj_turn_sock_destroy(bfcp->turn_socket);

		ao2_lock(bfcp);
		while (bfcp->turn_state != PJ_TURN_STATE_DESTROYING) {
			ast_cond_timedwait(&bfcp->cond, ao2_object_get_lockaddr(bfcp), &ts);
		}
		ao2_unlock(bfcp);

		bfcp->turn_socket = NULL;
	}

	if (bfcp->ioqueue) {
		bfcp_ioqueue_thread_remove(bfcp->ioqueue);
		bfcp->ioqueue = NULL;
	}

	ast_cond_destroy(&bfcp->cond);

	return;
}

static int ast_bfcp_new(struct ast_bfcp_udp *bfcp_udp)
{
	struct ast_bfcp *bfcp = NULL;

	if (!(bfcp = ao2_alloc(sizeof(*bfcp), bfcp_destructor))) {
		return -1;
	}

	ast_bfcp_set_data(bfcp_udp, bfcp);

	ast_cond_init(&bfcp->cond, NULL);

	return 0;
}

static int ast_bfcp_destory(struct ast_bfcp_udp *bfcp_udp)
{
	struct ast_bfcp *bfcp = ast_bfcp_get_data(bfcp_udp);

	if (!bfcp) {
		ast_log(LOG_WARNING, "BFCP already destoryed\n");
		return -1;
	}

	ao2_ref(bfcp, -1);
	
	return 0;
}

static int ast_bfcp_get_turn_relay_address(struct ast_bfcp_udp *bfcp_udp)
{
	unsigned int port = BFCP_DEFAULT_TURN_PORT;
	char *host_str = NULL;
	char *port_str = NULL;
	RAII_VAR(char *, user_name, NULL, ast_free);
	RAII_VAR(char *, password, NULL, ast_free);
	RAII_VAR(char *, sever, NULL, ast_free);

	user_name = ast_sip_get_turn_user_name();
	password = ast_sip_get_turn_password();
	sever = ast_sip_get_turn_server_address();
	
	if (!bfcp_udp) {
		return -1;
	}

	if (ast_strlen_zero(user_name) || ast_strlen_zero(password) || ast_strlen_zero(sever)) {
		ast_log(LOG_WARNING, "[BFCP:%p] Can't get user name or password or sever\n", bfcp_udp);
		return -1;
	}

	ast_sockaddr_split_hostport(sever, &host_str, &port_str, 0);
	if (!ast_strlen_zero(port_str)) {
		port = atoi(port_str);
		ast_log(LOG_DEBUG, "[BFCP:%p] Get port:%d from server:%s\n", bfcp_udp, port, sever);
	}

	if (udp_reachable) {
		bfcp_turn_request(bfcp_udp, AST_TRANSPORT_UDP, sever, port, user_name, password);
	} else {
		ast_log(LOG_DEBUG, "[BFCP:%p] UDP is unreachable, use TCP to allocate\n", bfcp_udp);
		bfcp_turn_request(bfcp_udp, AST_TRANSPORT_TCP, sever, port, user_name, password);
	}

	return 0;
}

static void ast_bfcp_set_udp_reachable(int value)
{
	if (udp_reachable != value) {
		ast_log(LOG_WARNING, "udp_reachable change to value:%d\n", value);
	}

	udp_reachable = value;

	return;
}

static int unload_module(void)
{
	ast_bfcp_engine_unregister(&asterisk_bfcp_engine);

	pj_thread_register_check();
	bfcp_terminate_pjproject();

	return 0;
}

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	if (pj_init() != PJ_SUCCESS) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (pjlib_util_init() != PJ_SUCCESS) {
		bfcp_terminate_pjproject();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (pjnath_init() != PJ_SUCCESS) {
		bfcp_terminate_pjproject();
		return AST_MODULE_LOAD_DECLINE;
	}

	pj_caching_pool_init(&cachingpool, &pj_pool_factory_default_policy, 0);

	if (ast_bfcp_engine_register(&asterisk_bfcp_engine, AST_MODULE_SELF)) {
		bfcp_terminate_pjproject();
		return AST_MODULE_LOAD_DECLINE;
	}

#if 0
/* Retain for query DNS in sched to speed up call setup time */
	if (!(bfcp_sched = ast_sched_context_create())) {
		ast_log(LOG_ERROR, "Unable to create scheduler context.\n");
		bfcp_terminate_pjproject();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sched_start_thread(bfcp_sched)) {
		ast_log(LOG_ERROR, "Unable to create scheduler context thread.\n");
		bfcp_terminate_pjproject();
		return AST_MODULE_LOAD_DECLINE;
	}

	sched_get_turn_address();
#endif

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "BFCP message handler",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjproject",
);

#endif

