/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2022, Grandstream, Inc.
 *
 * Binfeng Du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/module.h"
#include "asterisk/format.h"

AST_MODULE_LOG("format");

#define MAX_G723_ANNEXA_LEN 8

#define G723_ATTR_ANNEXA "annexa"

struct g723_attr {
	char annexa[MAX_G723_ANNEXA_LEN];
};

/* Destroy is a required callback and must exist */
static void g723_destroy(struct ast_format *format)
{
	/*BEGIN; Add by weidu, for g723 memory free*/
	if (format){
		struct g723_attr *attr = ast_format_get_attribute_data(format);
		ast_free(attr);
		ast_format_set_attribute_data(format, NULL);
	}
	/*END; Add by weidu, for g723 memory free*/
}

/* Clone is a required callback and must exist */
static int g723_clone(const struct ast_format *src, struct ast_format *dst)
{
	struct g723_attr *original = ast_format_get_attribute_data(src);
	struct g723_attr *attr = ast_calloc(1, sizeof(*attr));

	if (!attr) {
		return -1;
	}

	if (original) {
		*attr = *original;
	}

	ast_format_set_attribute_data(dst, attr);

	return 0;
}

static void g723_generate_sdp_fmtp(const struct ast_format *format, unsigned int payload, struct ast_str **str)
{
	struct g723_attr *attr = NULL;

	attr = ast_format_get_attribute_data(format);
	if (!attr) {
		ast_str_append(str, 0, "a=fmtp:%u annexa=no\r\n", payload);
		return;
	}
	
	if (ast_strlen_zero(attr->annexa)) {
		ast_str_append(str, 0, "a=fmtp:%u annexa=no\r\n", payload);
	} else {
		ast_str_append(str, 0, "a=fmtp:%u annexa=%s\r\n", payload, attr->annexa);
	}
	
	return;
}

static struct ast_format *g723_parse_sdp_fmtp(const struct ast_format *format, const char *attributes)
{
	char *attribs = ast_strdupa(attributes);
	char *attrib = NULL;
	struct ast_format *cloned = NULL;
	struct g723_attr *attr = NULL;
	char val[MAX_G723_ANNEXA_LEN] = { 0 };

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);
	if (!attr) {
		ast_log(LOG_ERROR, "Can't get g723 attr\n");
		return NULL;
	}

	/* Use yes as default */
	memset(attr->annexa, 0, sizeof(attr->annexa));
	strcpy(attr->annexa, "yes");

	while ((attrib = strsep(&attribs, ";"))) {
		attrib = ast_strip(attrib);

		if (sscanf(attrib, "annexa=%s", val) == 1) {
			memset(attr->annexa, 0, sizeof(attr->annexa));
			strncpy(attr->annexa, val, sizeof(attr->annexa) - 1);
		}
	}

	return cloned;
}

static struct ast_format *g723_getjoint(const struct ast_format *format1, const struct ast_format *format2)
{
	struct ast_format *cloned = NULL;

	cloned = ast_format_clone(format1);
	if (!cloned) {
		return NULL;
	}

	return cloned;
}

static struct ast_format *g723_set(const struct ast_format *format, const char *name, const char *value)
{
	struct ast_format *cloned = NULL;
	struct g723_attr *attr = NULL;

	if (!format || !name || !value) {
		return NULL;
	}

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	if (!strcasecmp(name, G723_ATTR_ANNEXA)) {
		memset(attr->annexa, 0, sizeof(attr->annexa));
		strncpy(attr->annexa, value, sizeof(attr->annexa) - 1);
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return cloned;
}

static struct ast_format *g723_set_directly(struct ast_format *format, const char *name, const char *value)
{
	struct g723_attr *attr = NULL;

	if (!format || !name || !value) {
		return NULL;
	}

	attr = ast_format_get_attribute_data(format);
	if (!attr) {
		return NULL;
	}

	if (!strcasecmp(name, G723_ATTR_ANNEXA)) {
		memset(attr->annexa, 0, sizeof(attr->annexa));
		strncpy(attr->annexa, value, sizeof(attr->annexa) - 1);
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return format;
}

static const void *g723_get(const struct ast_format *format, const char *name)
{
	struct g723_attr *attr = ast_format_get_attribute_data(format);
	char *val = NULL;

	if (!attr) {
		return NULL;
	}

	if (!strcasecmp(name, G723_ATTR_ANNEXA)) {
		val = attr->annexa;
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return val;
}

static struct ast_format_interface g723_interface = {
	.format_destroy = g723_destroy,
	.format_clone = g723_clone,
	.format_generate_sdp_fmtp = g723_generate_sdp_fmtp,
	.format_parse_sdp_fmtp = g723_parse_sdp_fmtp,
	.format_get_joint = g723_getjoint,
	.format_attribute_set = g723_set,
	.format_attribute_set_directly = g723_set_directly,
	.format_attribute_get = g723_get,
};

static int load_module(void)
{
	if (ast_format_interface_register("g723", &g723_interface)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "G.723 Format Attribute Module",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND,
);
