/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2016, Digium, Inc.
 *
 * Jason Parker <jparker@sangoma.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/module.h"
#include "asterisk/format.h"

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("format");

#define MAX_G729_ANNEXB_LEN 8

#define G729_ATTR_ANNEXB "annexb"

struct g729_attr {
	char annexb[MAX_G729_ANNEXB_LEN];
};
#endif

/* Destroy is a required callback and must exist */
static void g729_destroy(struct ast_format *format)
{
	/*BEGIN; Add by weidu, for g729 memory free*/
	if (format){
		struct g729_attr *attr = ast_format_get_attribute_data(format);
		ast_free(attr);
		ast_format_set_attribute_data(format, NULL);
	}
	/*END; Add by weidu, for g729 memory free*/
}

/* Clone is a required callback and must exist */
static int g729_clone(const struct ast_format *src, struct ast_format *dst)
{
#ifdef GRANDSTREAM_NETWORKS
	struct g729_attr *original = ast_format_get_attribute_data(src);
	struct g729_attr *attr = ast_calloc(1, sizeof(*attr));

	if (!attr) {
		return -1;
	}

	if (original) {
		*attr = *original;
	}

	ast_format_set_attribute_data(dst, attr);
#endif
	return 0;
}

static void g729_generate_sdp_fmtp(const struct ast_format *format, unsigned int payload, struct ast_str **str)
{
#ifdef GRANDSTREAM_NETWORKS
	struct g729_attr *attr = NULL;

	attr = ast_format_get_attribute_data(format);
	if (!attr) {
		ast_str_append(str, 0, "a=fmtp:%u annexb=no\r\n", payload);
		return;
	}

	if (ast_strlen_zero(attr->annexb)) {
		ast_str_append(str, 0, "a=fmtp:%u annexb=no\r\n", payload);
	} else {
		ast_str_append(str, 0, "a=fmtp:%u annexb=%s\r\n", payload, attr->annexb);
	}
#else
	/*
	 * According to the rfc the joint annexb format parameter should be set to 'yes'
	 * or 'no' based on the answerer (rfc7261 - 3.3). However, Asterisk being a B2BUA
	 * makes things tricky. So for now Asterisk will set annexb=no.
	 */
	ast_str_append(str, 0, "a=fmtp:%u annexb=no\r\n", payload);
#endif
}

#ifdef GRANDSTREAM_NETWORKS
static struct ast_format *g729_parse_sdp_fmtp(const struct ast_format *format, const char *attributes)
{
	char *attribs = ast_strdupa(attributes);
	char *attrib = NULL;
	struct ast_format *cloned = NULL;
	struct g729_attr *attr = NULL;
	char val[MAX_G729_ANNEXB_LEN] = { 0 };

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);
	if (!attr) {
		ast_log(LOG_ERROR, "Can't get g723 attr\n");
		return NULL;
	}

	/* Use yes as default */
	memset(attr->annexb, 0, sizeof(attr->annexb));
	strcpy(attr->annexb, "yes");

	while ((attrib = strsep(&attribs, ";"))) {
		attrib = ast_strip(attrib);

		if (sscanf(attrib, "annexb=%s", val) == 1) {
			memset(attr->annexb, 0, sizeof(attr->annexb));
			strncpy(attr->annexb, val, sizeof(attr->annexb) - 1);
		}
	}

	return cloned;
}

static struct ast_format *g729_getjoint(const struct ast_format *format1, const struct ast_format *format2)
{
	struct ast_format *cloned = NULL;

	cloned = ast_format_clone(format1);
	if (!cloned) {
		return NULL;
	}

	return cloned;
}

static struct ast_format *g729_set(const struct ast_format *format, const char *name, const char *value)
{
	struct ast_format *cloned = NULL;
	struct g729_attr *attr = NULL;

	if (!format || !name || !value) {
		return NULL;
	}

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	if (!strcasecmp(name, G729_ATTR_ANNEXB)) {
		memset(attr->annexb, 0, sizeof(attr->annexb));
		strncpy(attr->annexb, value, sizeof(attr->annexb) - 1);
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return cloned;
}

static struct ast_format *g729_set_directly(struct ast_format *format, const char *name, const char *value)
{
	struct g729_attr *attr = NULL;

	if (!format || !name || !value) {
		return NULL;
	}

	attr = ast_format_get_attribute_data(format);
	if (!attr) {
		return NULL;
	}

	if (!strcasecmp(name, G729_ATTR_ANNEXB)) {
		memset(attr->annexb, 0, sizeof(attr->annexb));
		strncpy(attr->annexb, value, sizeof(attr->annexb) - 1);
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return format;
}

static const void *g729_get(const struct ast_format *format, const char *name)
{
	struct g729_attr *attr = ast_format_get_attribute_data(format);
	char *val = NULL;

	if (!attr) {
		return NULL;
	}

	if (!strcasecmp(name, G729_ATTR_ANNEXB)) {
		val = attr->annexb;
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return val;
}
#endif

static struct ast_format_interface g729_interface = {
	.format_destroy = g729_destroy,
	.format_clone = g729_clone,
	.format_generate_sdp_fmtp = g729_generate_sdp_fmtp,
#ifdef GRANDSTREAM_NETWORKS
	.format_parse_sdp_fmtp = g729_parse_sdp_fmtp,
	.format_get_joint = g729_getjoint,
	.format_attribute_set = g729_set,
	.format_attribute_set_directly = g729_set_directly,
	.format_attribute_get = g729_get,
#endif
};

static int load_module(void)
{
	if (ast_format_interface_register("g729", &g729_interface)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "G.729 Format Attribute Module",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND,
);
