/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, grandstream, Inc.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 */

/*! \file
 *
 * \brief RTX Format Attribute Module
 *
 * This is a format attribute module for the GS-FEC codec.
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/module.h"
#include "asterisk/format.h"

AST_MODULE_LOG("pjsip");

#define MAX_VERSION_LEN 8

struct gs_fec_attr {
	char version[MAX_VERSION_LEN];
};


static void gs_fec_destroy(struct ast_format *format)
{
	struct gs_fec_attr *attr = ast_format_get_attribute_data(format);

	ast_free(attr);
}

static int gs_fec_clone(const struct ast_format *src, struct ast_format *dst)
{
	struct gs_fec_attr *original = ast_format_get_attribute_data(src);
	struct gs_fec_attr *attr = ast_calloc(1, sizeof(*attr));

	if (!attr) {
		return -1;
	}

	if (original) {
		*attr = *original;
	}

	ast_format_set_attribute_data(dst, attr);

	return 0;
}

static enum ast_format_cmp_res gs_fec_cmp(const struct ast_format *format1, const struct ast_format *format2)
{
	struct gs_fec_attr *attr1 = ast_format_get_attribute_data(format1);
	struct gs_fec_attr *attr2 = ast_format_get_attribute_data(format2);

	if (!attr1 || !attr2) {
		return AST_FORMAT_CMP_EQUAL;
	}

	if (!strcasecmp(attr1->version, attr2->version)) {
		return AST_FORMAT_CMP_EQUAL;
	}

	return AST_FORMAT_CMP_NOT_EQUAL;
}

static struct ast_format *gs_fec_getjoint(const struct ast_format *format1, const struct ast_format *format2)
{
	struct ast_format *cloned = NULL;

	cloned = ast_format_clone(format1);
	if (!cloned) {
		return NULL;
	}

	return cloned;
}

static struct ast_format *gs_fec_set(const struct ast_format *format, const char *name, const char *value)
{
	struct ast_format *cloned = NULL;
	struct gs_fec_attr *attr = NULL;

	if (!format || !name || !value) {
		return 0;
	}

	if (strlen(value) == 0) {
		return NULL;
	}

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	ast_log(LOG_DEBUG, "gs_fec_set value:%s\n", value);

	if (!strcasecmp(name, "version")) {
		strncpy(attr->version, value, sizeof(attr->version) - 1);
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return cloned;
}

static void set_max_gs_fec_version(struct gs_fec_attr *attr, char *gs_fec_version)
{
	unsigned int version = 0;
	unsigned int max_version = 0;
	char *version_str = NULL;
	char max_version_str[MAX_GS_FEC_VERSION_LEN] = {0};

	if (ast_strlen_zero(gs_fec_version)) {
		return;
	}

	ast_log(LOG_DEBUG, "determine_gs_fec_version init version:%s\n", gs_fec_version);

	if (strchr(gs_fec_version, ',') == NULL) {
		strncpy(attr->version, gs_fec_version, sizeof(attr->version) - 1);
		return;
	}

	while ((version_str = strsep(&gs_fec_version, ","))) {
		version_str = ast_strip(version_str);
		if (sscanf(version_str, "%d", &version) == 1) {
			if (max_version < version) {
				max_version = version;
			}
		}
	}

	snprintf(max_version_str, sizeof(max_version_str) - 1, "%d", max_version);
	strncpy(attr->version, max_version_str, sizeof(attr->version) - 1);

	ast_log(LOG_DEBUG, "get_max_gs_fec_version final version:%s\n", max_version_str);

	return;
}

static struct ast_format *gs_fec_parse_sdp_fmtp(const struct ast_format *format, const char *attributes)
{
	char *attribs = ast_strdupa(attributes);
	char *attrib = NULL;
	struct ast_format *cloned = NULL;
	struct gs_fec_attr *attr = NULL;
	char version[MAX_VERSION_LEN] = {0};

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	while ((attrib = strsep(&attribs, ";"))) {
		attrib = ast_strip(attrib);

		if (sscanf(attrib, "version=%s", version) == 1) {
			ast_log(LOG_DEBUG, "gs_fec_parse_sdp_fmtp version:%s\n", version);
			set_max_gs_fec_version(attr, version);
		}
	}

	return cloned;
}

#define APPEND_IF_NONZERO(field, str, name) do {		\
	if (strlen(field)) {	\
		if (added) {	\
			ast_str_append(str, 0, "; ");	\
		} else if (0 < ast_str_append(str, 0, "a=fmtp:%u ", payload)) {	\
			added = 1;	\
		}	\
		ast_str_append(str, 0, "%s=%s", name, field);	\
	}	\
} while (0)

static void gs_fec_generate_sdp_fmtp(const struct ast_format *format, unsigned int payload, struct ast_str **str)
{
	struct gs_fec_attr *attr = ast_format_get_attribute_data(format);
	int added = 0;

	if (!attr) {
		return;
	}

	/* Now we just need generate version attr, if don't has, just return */
	if (strlen(attr->version) == 0) {
		return;
	}

	APPEND_IF_NONZERO(attr->version, str, "version");

	if (added) {
		ast_str_append(str, 0, "\r\n");
	}

	return;
}

static const void *gs_fec_get(const struct ast_format *format, const char *name)
{
	struct gs_fec_attr *attr = ast_format_get_attribute_data(format);
	char *val = NULL;

	if (!strcasecmp(name, "version")) {
		val = attr->version;
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
		return NULL;
	}

	return val;
}

static struct ast_format_interface gs_fec_interface = {
	.format_destroy = gs_fec_destroy,
	.format_clone = gs_fec_clone,
	.format_cmp = gs_fec_cmp,
	.format_get_joint = gs_fec_getjoint,
	.format_attribute_set = gs_fec_set,
	.format_attribute_get = gs_fec_get,
	.format_parse_sdp_fmtp = gs_fec_parse_sdp_fmtp,
	.format_generate_sdp_fmtp = gs_fec_generate_sdp_fmtp,
};

static int unload_module(void)
{
	return 0;
}

static int load_module(void)
{
	if (ast_format_interface_register("fec", &gs_fec_interface)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "GS-FEC Format Attribute Module",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
);

