/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, grandstream, Inc.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 */

/*! \file
 *
 * \brief H.265 Format Attribute Module
 *
 * This is a format attribute module for the H.265 codec.
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/module.h"
#include "asterisk/format.h"

#define H265_MAX_FMTP_PARAM 256

struct h265_attr {
	char fmtp_param[H265_MAX_FMTP_PARAM];
};


static void h265_destroy(struct ast_format *format)
{
	struct h265_attr *attr = ast_format_get_attribute_data(format);

	ast_free(attr);
}

static int h265_clone(const struct ast_format *src, struct ast_format *dst)
{
	struct h265_attr *original = ast_format_get_attribute_data(src);
	struct h265_attr *attr = ast_calloc(1, sizeof(*attr));

	if (!attr) {
		return -1;
	}

	if (original) {
		*attr = *original;
	}

	ast_format_set_attribute_data(dst, attr);

	return 0;
}

static enum ast_format_cmp_res h265_cmp(const struct ast_format *format1, const struct ast_format *format2)
{
	struct h265_attr *attr1 = ast_format_get_attribute_data(format1);
	struct h265_attr *attr2 = ast_format_get_attribute_data(format2);

	if (!attr1 || !attr2) {
		return AST_FORMAT_CMP_EQUAL;
	}

	if (!strlen(attr1->fmtp_param) || !strlen(attr2->fmtp_param)) {
		return AST_FORMAT_CMP_EQUAL;
	}

	if (!strcmp(attr1->fmtp_param, attr2->fmtp_param)) {
		return AST_FORMAT_CMP_EQUAL;
	}

	return AST_FORMAT_CMP_NOT_EQUAL;
}

static struct ast_format *h265_getjoint(const struct ast_format *format1, const struct ast_format *format2)
{
	struct ast_format *cloned = NULL;

	cloned = ast_format_clone(format1);
	if (!cloned) {
		return NULL;
	}

	return cloned;
}


static struct ast_format *h265_parse_sdp_fmtp(const struct ast_format *format, const char *attributes)
{
	struct ast_format *cloned = NULL;
	struct h265_attr *attr = NULL;

	if (!format || !attributes) {
		return NULL;
	}

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	if (attr) {
		strcpy(attr->fmtp_param, attributes);
	}

	return cloned;
}

static void h265_generate_sdp_fmtp(const struct ast_format *format, unsigned int payload, struct ast_str **str)
{
	struct h265_attr *attr = ast_format_get_attribute_data(format);

	if (!attr) {
		return;
	}

	ast_str_append(str, 0, "a=fmtp:%u ", payload);
	ast_str_append(str, 0, attr->fmtp_param);
	ast_str_append(str, 0, "\r\n");

	return;
}

static struct ast_format_interface h265_interface = {
	.format_destroy = h265_destroy,
	.format_clone = h265_clone,
	.format_cmp = h265_cmp,
	.format_get_joint = h265_getjoint,
	.format_parse_sdp_fmtp = h265_parse_sdp_fmtp,
	.format_generate_sdp_fmtp = h265_generate_sdp_fmtp,
};

static int unload_module(void)
{
	return 0;
}

static int load_module(void)
{
	if (ast_format_interface_register("h265", &h265_interface)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "H.265 Format Attribute Module",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
);

