/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, grandstream, Inc.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 */

/*! \file
 *
 * \brief RTX Format Attribute Module
 *
 * This is a format attribute module for the RTX codec.
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include "asterisk/module.h"
#include "asterisk/format.h"

#define CODEC_RTX_ATTR_APT "apt"
#define CODEC_RTX_ATTR_RTX_TIME "rtx-time"

struct rtx_attr {
	unsigned int apt;
	unsigned int rtx_time;
};


static void rtx_destroy(struct ast_format *format)
{
	struct rtx_attr *attr = ast_format_get_attribute_data(format);

	ast_free(attr);
}

static int rtx_clone(const struct ast_format *src, struct ast_format *dst)
{
	struct rtx_attr *original = ast_format_get_attribute_data(src);
	struct rtx_attr *attr = ast_calloc(1, sizeof(*attr));

	if (!attr) {
		return -1;
	}

	if (original) {
		*attr = *original;
	}

	ast_format_set_attribute_data(dst, attr);

	return 0;
}

static enum ast_format_cmp_res rtx_cmp(const struct ast_format *format1, const struct ast_format *format2)
{
	struct rtx_attr *attr1 = ast_format_get_attribute_data(format1);
	struct rtx_attr *attr2 = ast_format_get_attribute_data(format2);

	if (!attr1 || !attr2) {
		return AST_FORMAT_CMP_EQUAL;
	}

	if ((attr1->apt == attr2->apt) && (attr1->rtx_time == attr2->rtx_time)) {
		return AST_FORMAT_CMP_EQUAL;
	}

	return AST_FORMAT_CMP_NOT_EQUAL;
}

static struct ast_format *rtx_getjoint(const struct ast_format *format1, const struct ast_format *format2)
{
	struct ast_format *cloned = NULL;

	cloned = ast_format_clone(format1);
	if (!cloned) {
		return NULL;
	}

	return cloned;
}

static struct ast_format *rtx_set(const struct ast_format *format, const char *name, const char *value)
{
	struct ast_format *cloned;
	struct rtx_attr *attr;
	unsigned int val;

	if (sscanf(value, "%30u", &val) != 1) {
		ast_log(LOG_WARNING, "Unknown value '%s' for attribute type '%s'\n",
			value, name);
		return NULL;
	}

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	if (!strcasecmp(name, CODEC_RTX_ATTR_APT)) {
		attr->apt = val;
	} else if (!strcasecmp(name, CODEC_RTX_ATTR_RTX_TIME)) {
		attr->rtx_time = val;
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return cloned;
}

static struct ast_format *rtx_parse_sdp_fmtp(const struct ast_format *format, const char *attributes)
{
	char *attribs = ast_strdupa(attributes);
	char *attrib = NULL;
	struct ast_format *cloned = NULL;
	struct rtx_attr *attr = NULL;
	unsigned int val = 0;

	cloned = ast_format_clone(format);
	if (!cloned) {
		return NULL;
	}
	attr = ast_format_get_attribute_data(cloned);

	while ((attrib = strsep(&attribs, ";"))) {
		attrib = ast_strip(attrib);

		if (sscanf(attrib, "apt=%30u", &val) == 1) {
			attr->apt = val;
		} else if (sscanf(attrib, "rtx-time=%30u", &val) == 1) {
			attr->rtx_time = val;
		}
	}

	return cloned;
}

#define APPEND_IF_NONZERO(field, str, name) do {		\
	if (field) {	\
		if (added) {	\
			ast_str_append(str, 0, "; ");	\
		} else if (0 < ast_str_append(str, 0, "a=fmtp:%u ", payload)) {	\
			added = 1;	\
		}	\
		ast_str_append(str, 0, "%s=%u", name, field);	\
	}	\
} while (0)

static void rtx_generate_sdp_fmtp(const struct ast_format *format, unsigned int payload, struct ast_str **str)
{
	struct rtx_attr *attr = ast_format_get_attribute_data(format);
	int added = 0;

	if (!attr) {
		return;
	}

	APPEND_IF_NONZERO(attr->apt, str, "apt");
	APPEND_IF_NONZERO(attr->rtx_time, str, "rtx-time");

	if (added) {
		ast_str_append(str, 0, "\r\n");
	}

	return;
}

static const void *rtx_get(const struct ast_format *format, const char *name)
{
	struct rtx_attr *attr = ast_format_get_attribute_data(format);
	unsigned int *val = NULL;

	if (!attr) {
		return NULL;
	}

	if (!strcasecmp(name, CODEC_RTX_ATTR_APT)) {
		val = &attr->apt;
	} else if (!strcasecmp(name, CODEC_RTX_ATTR_RTX_TIME)) {
		val = &attr->rtx_time;
	} else {
		ast_log(LOG_WARNING, "unknown attribute type %s\n", name);
	}

	return val;
}

static struct ast_format_interface rtx_interface = {
	.format_destroy = rtx_destroy,
	.format_clone = rtx_clone,
	.format_cmp = rtx_cmp,
	.format_get_joint = rtx_getjoint,
	.format_attribute_set = rtx_set,
	.format_parse_sdp_fmtp = rtx_parse_sdp_fmtp,
	.format_generate_sdp_fmtp = rtx_generate_sdp_fmtp,
	.format_attribute_get = rtx_get,
};

static int unload_module(void)
{
	return 0;
}

static int load_module(void)
{
	if (ast_format_interface_register("rtx", &rtx_interface)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "RTX Format Attribute Module",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
);

