/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

#include "asterisk.h"
#include <pjsip.h>
#include <pjlib.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/logger.h"
#include "asterisk/astobj2.h"
#include "asterisk/paths.h"
#include "asterisk/sorcery.h"
#include "asterisk/taskprocessor.h"
#include "include/res_pjsip_private.h"
#include "asterisk/res_pjsip_cli.h"
#include "asterisk/statsd.h"
#include "asterisk/named_locks.h"

#include "asterisk/res_pjproject.h"

#ifdef GRANDSTREAM_NETWORKS
#include <pjsip_ua.h>
#include "asterisk/app.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/astdb.h"

AST_MODULE_LOG("pjsip");
#endif

static int pj_max_hostname = PJ_MAX_HOSTNAME;
static int pjsip_max_url_size = PJSIP_MAX_URL_SIZE;

#ifdef GRANDSTREAM_NETWORKS
static unsigned int g_proxy_user_num = 0;
static unsigned int g_local_user_num = 0;
static ast_rwlock_t  proxy_user_count_lock;
static ast_rwlock_t  local_user_count_lock;
static ast_rwlock_t  reg_user_update_lock;

static unsigned int g_max_rc_users = 0;    /* The maximum of the RC registered users */
static unsigned int g_max_local_users = 0;    /* The maximum of the RC registered users */
static unsigned int g_need_update_local = 0;  /* The flag to indicate whether need to update or not */
static unsigned int g_need_update_rc = 0;     /* The flag to indicate whether need to update or not */
static unsigned int g_need_update_max = 0;    
static int  update_sched_id = -1;

static unsigned int g_max_reg_users = 0;    /* The maximum of the registered users */
static struct ao2_container *user_reginfos;
static ast_rwlock_t  reg_user_max_lock;

static struct ao2_container *rc_reginfos;
static struct ao2_container *local_reginfos;
static struct ao2_container *account_tk_infos;
static struct ao2_container *account_tk_confs;
static ast_rwlock_t  rc_user_max_lock;
static ast_rwlock_t  local_user_max_lock;
static struct ast_sched_context *local_update_sched = NULL;

#define REGISTER_FAMILY     "REGISTER"
#define REGISTER_LOCAL_MAX  "MaxLocalUsers"
#define REGISTER_LOCAL_UPDATE_INT   30

static struct ao2_container *cli_contact_get_container(const char *regex);
static struct ao2_container *cli_get_aors(void);

static int ast_sip_get_useragent(char *str, char *user, int len);

static int ast_sip_account_tk_cmp(void *obj, void *arg, int flags)
{
	struct ast_sip_account_tk_info *info1 = obj, *info2 = arg;

	return !strcasecmp(info1->endpoint, info2->endpoint) ? CMP_MATCH | CMP_STOP : 0;
}

struct ast_sip_endpoint *ast_sip_get_accounttk_endpoint_by_address(char *addr)
{
	struct ao2_iterator iter;
	struct ast_sip_account_tk_info *info;
	struct ast_sip_endpoint *endpoint = NULL;
	int found = 0;

	if (!addr) {
		return NULL;
	}

	if (!ao2_container_count(account_tk_infos)) {
		return NULL;
	}

	iter = ao2_iterator_init(account_tk_infos, AO2_ITERATOR_DONTLOCK);
	while ((info = ao2_iterator_next(&iter))) {
		if (!strcasecmp(info->address, addr)) {
            ast_log(LOG_DEBUG, "Found account trunk info: %s, username[%s], address[%s]", info->endpoint, info->username, info->address);
			found = 1;
			break;
		}
		ao2_ref(info, -1);
	}
	ao2_iterator_destroy(&iter);

	if (found) {
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", info->endpoint);
		ao2_ref(info, -1);
		return endpoint;
	}

	return NULL;
}

struct ast_sip_account_tk_info *ast_sip_find_account_tk_info(const char *name)
{
	struct ast_sip_account_tk_info tk_info;

	if (!name) {
		return NULL;
	}

	strncpy(tk_info.endpoint, name, sizeof(tk_info.endpoint) - 1);

	return ao2_find(account_tk_infos, &tk_info, OBJ_POINTER);
}

static unsigned int ast_sip_delete_account_tk_info(struct ast_sip_contact *contact)
{
	RAII_VAR(struct ast_sip_account_tk_info *, tk_info, NULL, ao2_cleanup);

	if (!contact || ast_strlen_zero(contact->endpoint_name)) {
		return -1;
	}

	if (!contact->endpoint) {
		struct ast_sip_endpoint *endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", contact->endpoint_name);
		if (!endpoint) {
			return -1;
		}

		if (ast_strlen_zero(endpoint->trunk_type) || strncmp(endpoint->trunk_type, "account", 7)) {
			return -1;
		}
    } else {
		if (ast_strlen_zero(contact->endpoint->trunk_type) || strncmp(contact->endpoint->trunk_type, "account", 7)) {
			return -1;
		}
	}

	tk_info = ast_sip_find_account_tk_info(contact->endpoint_name);
	if (tk_info) {
		/* check address */
		char address[80] = {0};
		char user[80]    = {0};

		if (!ast_strlen_zero(contact->original_address)) {
			strncpy(address, contact->original_address, sizeof(address)-1);
		} else if (!ast_strlen_zero(contact->uri)) {
			sscanf(contact->uri, "%79[^@]@%79[^;]", user, address);
		} else {
			return -1;
		}

		if (!strcmp(tk_info->address, address)) {
			ast_log(LOG_DEBUG, "Delete account trunk info: %s, username[%s], address[%s]", tk_info->endpoint, tk_info->username, tk_info->address);
			ao2_unlink(account_tk_infos, tk_info);
		}
    }

	return 0;
}

static unsigned int ast_sip_update_account_tk_info(struct ast_sip_contact *contact)
{
    char  endpt_id[16] = {0};
	char  address[80]  = {0};
	RAII_VAR(struct ast_sip_account_tk_info *, tk_info, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (!contact || ast_strlen_zero(contact->endpoint_name)) {
		return -1;
	}

    endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", contact->endpoint_name);
	if (!endpoint) {
		return -1;
	}

	if (ast_strlen_zero(endpoint->trunk_type) || strncmp(endpoint->trunk_type, "account", 7)) {
		return -1;
	}

	/* get ip from original_address or via */
	if (!ast_strlen_zero(contact->original_address)) {
		strncpy(address, contact->original_address, sizeof(address)-1);
	} else if (!ast_strlen_zero(contact->uri)) {
		char user[80] = {0};
        sscanf(contact->uri, "%79[^@]@%79[^;]", user, address);
	} else {
        return -1;
    }

	strncpy(endpt_id, contact->endpoint_name, sizeof(endpt_id)-1);
	tk_info = ast_sip_find_account_tk_info(endpt_id);

	if (tk_info) {
		snprintf(tk_info->address, sizeof(tk_info->address)-1, "%s", address);
		snprintf(tk_info->username, sizeof(tk_info->username)-1, "%s", endpoint->username);
		ast_log(LOG_DEBUG, "Update account trunk info: %s, username[%s], address[%s]", endpt_id, tk_info->username, tk_info->address);
		return 0;
	}

	tk_info = ao2_alloc(sizeof(*tk_info), NULL);
	if (!tk_info) {
		return -1;
	}

    snprintf(tk_info->endpoint, sizeof(tk_info->endpoint)-1, "%s", endpt_id);
    snprintf(tk_info->address, sizeof(tk_info->address)-1, "%s", address);
    snprintf(tk_info->username, sizeof(tk_info->username)-1, "%s", endpoint->username);

	ast_log(LOG_DEBUG, "Add account trunk info: %s, username[%s], address[%s]", endpt_id, tk_info->username, tk_info->address);
	ao2_link(account_tk_infos, tk_info);
	return 0;
}


static int ast_sip_acctk_conf_cmp(void *obj, void *arg, int flags)
{
	struct ast_sip_account_tk_conf *info1 = obj, *info2 = arg;

	return !strcasecmp(info1->endpoint, info2->endpoint) ? CMP_MATCH | CMP_STOP : 0;
}

static int ast_sip_acctk_conf_mark(void *obj, void *arg, int flags)
{
	struct ast_sip_account_tk_conf *tk_conf = obj;

	if (tk_conf) {
		tk_conf->mark = 1;
	}
	return 0;
}

static int ast_sip_acctk_conf_marked_match(void *obj, void *arg, int flag)
{
	struct ast_sip_account_tk_conf *tk_conf = obj;

	if (tk_conf && tk_conf->mark) {
		return CMP_MATCH;
	}
	return 0;
}

void ast_sip_acctk_markall(void)
{
    ao2_callback(account_tk_confs, OBJ_NODATA, ast_sip_acctk_conf_mark, NULL);
}

void ast_sip_acctk_unlink_marked(void)
{
	int flag = 0;
	ao2_callback(account_tk_confs, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, ast_sip_acctk_conf_marked_match, &flag);
}

static struct ast_sip_account_tk_conf *ast_sip_find_account_tk(const char *name)
{
	struct ast_sip_account_tk_conf tk_info;

	if (!name) {
		return NULL;
	}

	strncpy(tk_info.endpoint, name, sizeof(tk_info.endpoint) - 1);

	return ao2_find(account_tk_confs, &tk_info, OBJ_POINTER);
}

struct ast_sip_endpoint *ast_sip_get_account_tk_endpoint(char *username, char *transport)
{
	struct ast_sip_endpoint        *endpoint = NULL;
	struct ast_sip_account_tk_conf *tk_conf  = NULL;
	struct ao2_iterator iter;

	if (!username || !transport || !account_tk_confs) {
		return NULL;
	}

	if (!ao2_container_count(account_tk_confs)) {
		return NULL;
	}

	iter = ao2_iterator_init(account_tk_confs, AO2_ITERATOR_DONTLOCK);
	while ((tk_conf = ao2_iterator_next(&iter))) {
		if (!strcasecmp(tk_conf->username, username) && !strcasecmp(tk_conf->transport, transport)) {
			ast_log(LOG_DEBUG, "Found account trunk: %s", tk_conf->endpoint);
			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", tk_conf->endpoint);
			ao2_ref(tk_conf, -1);
			break;
		}

		ao2_ref(tk_conf, -1);
	}
	ao2_iterator_destroy(&iter);

	return endpoint;
}

static void acctk_destroy(void *obj)
{
	struct ast_sip_account_tk_conf *tk_conf = obj;
	struct ast_sip_account_tk_info *tk_info = NULL;

	/* Remove from account_tk_infos */
	tk_info = ast_sip_find_account_tk_info(tk_conf->endpoint);
	if (tk_info) {
		ao2_unlink(account_tk_infos, tk_info);
	}

	ast_log(LOG_DEBUG, "acctk_destroy, [%s]", tk_conf->endpoint);
	return;
}

unsigned int ast_sip_set_account_tk_info(struct ast_sip_endpoint *endpoint)
{
	char  endpt_id[16] = {0};
	RAII_VAR(struct ast_sip_account_tk_conf *, tk_info, NULL, ao2_cleanup);

	if (!endpoint) {
		return -1;
	}

	if (ast_strlen_zero(endpoint->trunk_type) || strncmp(endpoint->trunk_type, "account", 7)) {
		return -1;
	}

	strncpy(endpt_id, ast_sorcery_object_get_id(endpoint), sizeof(endpt_id)-1);
	tk_info = ast_sip_find_account_tk(endpt_id);

	if (tk_info) {
		snprintf(tk_info->transport, sizeof(tk_info->transport)-1, "%s", endpoint->transport);
		snprintf(tk_info->username, sizeof(tk_info->username)-1, "%s", endpoint->username);
		ast_log(LOG_DEBUG, "Update account trunk info: %s, username[%s], transport[%s]", endpt_id, tk_info->username, tk_info->transport);
		tk_info->mark = 0;
		return 0;
	}

	tk_info = ao2_alloc(sizeof(*tk_info), acctk_destroy);
	if (!tk_info) {
		return -1;
	}

	snprintf(tk_info->endpoint, sizeof(tk_info->endpoint)-1, "%s", endpt_id);
	snprintf(tk_info->transport, sizeof(tk_info->transport)-1, "%s", endpoint->transport);
	snprintf(tk_info->username, sizeof(tk_info->username)-1, "%s", endpoint->username);
	tk_info->mark = 0;

	ast_log(LOG_DEBUG, "Add account trunk info: %s, username[%s], transport[%s]", endpt_id, tk_info->username, tk_info->transport);
	ao2_link(account_tk_confs, tk_info);
	return 0;
}

static unsigned int ast_sip_get_max_reg_users(void)
{
	unsigned int reg_user_num = 0;

	ast_rwlock_wrlock(&reg_user_max_lock);
	reg_user_num = g_max_reg_users;
	ast_rwlock_unlock(&reg_user_max_lock);

	return reg_user_num;
}

static unsigned int ast_sip_set_max_reg_users(unsigned int num)
{
	ast_rwlock_wrlock(&reg_user_max_lock);
	g_max_reg_users = num;
	ast_rwlock_unlock(&reg_user_max_lock);

	ast_log(LOG_DEBUG, "g_max_reg_users is [%d] now.\n", g_max_reg_users);
	return g_max_reg_users;
}

static struct ast_sip_reginfo *ast_sip_find_user_reginfo(const char *name, int rc_flag)
{
	struct ast_sip_reginfo reg_info;
	strncpy(reg_info.user_agent, name, sizeof(reg_info.user_agent)-1);
	reg_info.rc_flag = rc_flag;

	return ao2_find(user_reginfos, &reg_info, OBJ_POINTER);
}

static unsigned int ast_sip_add_reg_users(struct ast_sip_contact *contact)
{
	char *user_agent = NULL;
	char prod[256] = {0};
	int  rc_flag   = 0;

	RAII_VAR(struct ast_sip_reginfo *, reginfo, NULL, ao2_cleanup);

	if (!contact) {
		return -1;
	}

	if (ast_strlen_zero(contact->user_agent)) {
		strncpy(prod, "Others", sizeof(prod)-1);
	} else {
		user_agent = ast_strdupa(contact->user_agent);
		ast_sip_get_useragent(user_agent, prod, sizeof(prod)-1);
	}

	if (contact->ssh_proxy_flag) {
		rc_flag = 1;
	}

	reginfo = ast_sip_find_user_reginfo(prod, rc_flag);
	if (reginfo) {
		reginfo->number++;
		ast_log(LOG_DEBUG, "Add reg users, user_agent[%s], num[%d]", prod, reginfo->number);
		return 0;
	}

	reginfo = ao2_alloc(sizeof(*reginfo), NULL);

	strncpy(reginfo->user_agent, prod, sizeof(reginfo->user_agent)-1);
	reginfo->number = 1;
	reginfo->rc_flag = rc_flag;

	ast_log(LOG_DEBUG, "Add reg users, user_agent[%s], rc_flag[%d]", prod, rc_flag);
	ao2_link(user_reginfos, reginfo);
	return 0;
}


static void ast_sip_update_accounttk_reginfos(void)
{
	struct ao2_container   *con_contacts;
	struct ast_sip_contact_wrapper *wrapper;
	struct ao2_iterator i;

	ao2_callback(account_tk_infos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	con_contacts = cli_contact_get_container("");
	if (!con_contacts) {
		return;
	}

	i = ao2_iterator_init(con_contacts, 0);
	for (; (wrapper = ao2_iterator_next(&i)); ao2_ref(wrapper, -1)) {
		if (!strncmp(wrapper->aor_id, "trunk_", 6)) {
            ast_sip_update_account_tk_info(wrapper->contact);
		}
	}

	ao2_iterator_destroy(&i);
	ao2_ref(con_contacts, -1);
}

static void ast_sip_update_user_reginfos(void)
{
	struct ao2_container   *con_contacts;
	struct ast_sip_contact_wrapper *wrapper;
	struct ast_sip_contact *contact;
	struct ao2_iterator i;
	unsigned int num = 0;

	ao2_callback(user_reginfos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	con_contacts = cli_contact_get_container("");
	if (!con_contacts) {
		return;
	}

	i = ao2_iterator_init(con_contacts, 0);
	for (; (wrapper = ao2_iterator_next(&i)); ao2_ref(wrapper, -1)) {
		if (!strncmp(wrapper->aor_id, "trunk_", 6)) {
			continue;
		}

		contact = wrapper->contact;
		ast_sip_add_reg_users(contact);
		num++;
	}

	ao2_iterator_destroy(&i);
	ao2_ref(con_contacts, -1);
	ast_sip_set_max_reg_users(num);
}

static unsigned int ast_sip_get_max_rc_users(void)
{
	unsigned int rc_user_num = 0;

	ast_rwlock_wrlock(&rc_user_max_lock);
	rc_user_num = g_max_rc_users;
	ast_rwlock_unlock(&rc_user_max_lock);

	return rc_user_num;
}

static unsigned int ast_sip_set_max_rc_users(unsigned int num)
{
	ast_rwlock_wrlock(&rc_user_max_lock);
	g_max_rc_users = num;
	ast_rwlock_unlock(&rc_user_max_lock);

	return g_max_rc_users;
}

static void ast_sip_set_current_proxy_user_num(unsigned int num)
{
	ast_rwlock_wrlock(&proxy_user_count_lock);
	g_proxy_user_num = num;
	ast_rwlock_unlock(&proxy_user_count_lock);

	ast_log(LOG_DEBUG, "g_proxy_user_num is [%d].\n", num);
	return;
}

static struct ast_sip_reginfo *ast_sip_find_rc_reginfo(const char *name)
{
	struct ast_sip_reginfo reg_info;
	strncpy(reg_info.user_agent, name, sizeof(reg_info.user_agent)-1);
	reg_info.rc_flag = 1;

	return ao2_find(rc_reginfos, &reg_info, OBJ_POINTER);
}

static unsigned int ast_sip_get_max_local_users(void)
{
	unsigned int local_user_num = 0;

	ast_rwlock_wrlock(&local_user_max_lock);
	local_user_num = g_max_local_users;
	ast_rwlock_unlock(&local_user_max_lock);

	return local_user_num;
}

static unsigned int ast_sip_set_max_local_users(unsigned int num)
{
	char num_str[8] = {0};

	ast_rwlock_wrlock(&local_user_max_lock);
	g_max_local_users = num;
	ast_rwlock_unlock(&local_user_max_lock);

	snprintf(num_str, sizeof(num_str), "%d", num);
	ast_db_put(REGISTER_FAMILY, REGISTER_LOCAL_MAX, num_str);

	return g_max_local_users;
}

static struct ast_sip_reginfo *ast_sip_find_local_reginfo(const char *name)
{
	struct ast_sip_reginfo reg_info;
	strncpy(reg_info.user_agent, name, sizeof(reg_info.user_agent)-1);
	reg_info.rc_flag = 0;

	return ao2_find(local_reginfos, &reg_info, OBJ_POINTER);
}

static int ast_sip_get_useragent(char *str, char *user, int len)
{
	char prod0[80+1] = {0};
	char prod1[80+1] = {0};
	char prod2[80+1] = {0};
	char prod3[80+1] = {0};

	if (!str || !user) {
		return -1;
	}

	if (sscanf(str, "%80s %80s %80s %80s", prod0, prod1, prod2, prod3) >= 3) {
		if (!strncmp(prod0, "Grandstream", 11)) {
			if (!strncmp(prod1, "Wave", 4)) {
				snprintf(user, len, "%s %s", prod1, prod2);
			} else {
				strncpy(user, prod1, len);
			}

			return 0;
		}
	} 

	strncpy(user, str, len);
	return 0;
}

static unsigned int ast_sip_add_local_users(struct ast_sip_contact *contact)
{
	char *user_agent = NULL;
	char prod[256] = {0};

	RAII_VAR(struct ast_sip_reginfo *, reginfo, NULL, ao2_cleanup);

	if (!contact || ast_strlen_zero(contact->user_agent)) {
		return -1;
	}

	user_agent = ast_strdupa(contact->user_agent);
	ast_sip_get_useragent(user_agent, prod, sizeof(prod)-1);

	reginfo = ast_sip_find_local_reginfo(prod);
	if (reginfo) {
		reginfo->number++;
		ast_log(LOG_DEBUG, "Add local users, user_agent[%s], num[%d]", prod, reginfo->number);
		return 0;
	}

	reginfo = ao2_alloc(sizeof(*reginfo), NULL);

	strncpy(reginfo->user_agent, prod, sizeof(reginfo->user_agent)-1);
	reginfo->number = 1;
	reginfo->rc_flag = 0;

	ast_log(LOG_DEBUG, "Add local users, user_agent[%s]", prod);
	ao2_link(local_reginfos, reginfo);
	return 0;
}

static unsigned int ast_sip_add_rc_users(struct ast_sip_contact *contact)
{
	char *user_agent = NULL;
	char prod[256] = {0};
	RAII_VAR(struct ast_sip_reginfo *, rc_reginfo, NULL, ao2_cleanup);

	if (!contact || ast_strlen_zero(contact->user_agent)) {
		return -1;
	}

	user_agent = ast_strdupa(contact->user_agent);
	ast_sip_get_useragent(user_agent, prod, sizeof(prod)-1);

	rc_reginfo = ast_sip_find_rc_reginfo(prod);
	if (rc_reginfo) {
		rc_reginfo->number++;
		ast_log(LOG_NOTICE, "Add rc users, user_agent[%s], num[%d]", prod, rc_reginfo->number);
		return 0;
	}

	rc_reginfo = ao2_alloc(sizeof(*rc_reginfo), NULL);

	strncpy(rc_reginfo->user_agent, prod, sizeof(rc_reginfo->user_agent)-1);
	rc_reginfo->number = 1;
	rc_reginfo->rc_flag = 1;

	ast_log(LOG_NOTICE, "Add rc users, user_agent[%s]", prod);
	ao2_link(rc_reginfos, rc_reginfo);
	return 0;
}


static void ast_sip_update_rc_reginfos(void)
{
	struct ao2_container   *con_contacts;
	struct ast_sip_contact_wrapper *wrapper;
	struct ast_sip_contact *contact;
	struct ao2_iterator i;
	unsigned int num = 0;
	char  host[256] = {0};

	/* To check nvram :WsProxyHost */
	ast_sip_get_nvram_value(":WsProxyHost", host, sizeof(host));
	if (!strlen(host)) {
		return;
	}

	ao2_callback(rc_reginfos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	con_contacts = cli_contact_get_container("");
	if (!con_contacts) {
		return;
	}

	i = ao2_iterator_init(con_contacts, 0);
	for (; (wrapper = ao2_iterator_next(&i)); ao2_ref(wrapper, -1)) {
		if (!strncmp(wrapper->aor_id, "trunk_", 6)) {
			continue;
		}

		contact = wrapper->contact;
		if (contact->ssh_proxy_flag) {
			ast_sip_add_rc_users(contact);
			num++;
		}
	}

	ao2_iterator_destroy(&i);
	ao2_ref(con_contacts, -1);
	ast_sip_set_max_rc_users(num);

	ast_sip_set_current_proxy_user_num(num);   /* Added by yfyang, for Bug 293443*/
}

static void ast_sip_update_local_reginfos(void)
{
	struct ao2_container   *con_contacts;
	struct ast_sip_contact_wrapper *wrapper;
	struct ast_sip_contact *contact;
	struct ao2_iterator i;
	unsigned int num = 0;
	char  host[256] = {0};

	/* To check nvram :WsProxyHost */
	ast_sip_get_nvram_value(":WsProxyHost", host, sizeof(host));
	if (!strlen(host)) {
		return;
	}

	ao2_callback(local_reginfos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	con_contacts = cli_contact_get_container("");
	if (!con_contacts) {
		return;
	}

	i = ao2_iterator_init(con_contacts, 0);
	for (; (wrapper = ao2_iterator_next(&i)); ao2_ref(wrapper, -1)) {
		if (!strncmp(wrapper->aor_id, "trunk_", 6)) {
			continue;
		}

		contact = wrapper->contact;
		if (contact && !contact->ssh_proxy_flag) {
			ast_sip_add_local_users(contact);
			num++;
		}
	}

	ao2_iterator_destroy(&i);
	ao2_ref(con_contacts, -1);
	ast_sip_set_max_local_users(num);
	ast_sip_set_current_local_user_num(num);
}

static int ast_sip_update_reg_users(const void *data)
{
	if (g_need_update_local) {
		ast_rwlock_wrlock(&reg_user_update_lock);
		g_need_update_local = 0;
		ast_rwlock_unlock(&reg_user_update_lock);

		ast_sip_update_local_reginfos();
	}

	if (g_need_update_rc) {
		ast_rwlock_wrlock(&reg_user_update_lock);
		g_need_update_rc = 0;
		ast_rwlock_unlock(&reg_user_update_lock);

		ast_sip_update_rc_reginfos();
	}

	if (g_need_update_max) {
		ast_rwlock_wrlock(&reg_user_update_lock);
		g_need_update_max = 0;
		ast_rwlock_unlock(&reg_user_update_lock);

		ast_sip_update_user_reginfos();
	}

	update_sched_id = ast_sched_add(local_update_sched, REGISTER_LOCAL_UPDATE_INT * 1000, ast_sip_update_reg_users, NULL);
	return 0;
}

static void ast_sip_check_update_start(void)
{
	local_update_sched = ast_sched_context_create();
	if (!local_update_sched) {
		return;
	}

	if (ast_sched_start_thread(local_update_sched)) {
		ast_log(LOG_ERROR, "Could not start local_update_sched\n");
		ast_sched_context_destroy(local_update_sched);
		local_update_sched = NULL;
		return;
	}

	update_sched_id = ast_sched_add(local_update_sched, REGISTER_LOCAL_UPDATE_INT * 1000, ast_sip_update_reg_users, NULL);
	if (update_sched_id < 0) {
		ast_log(LOG_ERROR, "Could not add update_sched_id\n");
		ast_sched_context_destroy(local_update_sched);
		local_update_sched = NULL;
		return;
	}

	return;
}

void ast_sip_check_and_update_local_reginfo(void)
{
	char db_value[8] = {0};
	unsigned int local_num = ast_sip_get_max_local_users();
	unsigned int db_num = 0;

	ast_db_get(REGISTER_FAMILY, REGISTER_LOCAL_MAX, db_value, sizeof(db_value));
	if (strlen(db_value)) {
		db_num = atoi(db_value);
	}

	if (!db_num || db_num != local_num) {
		ast_sip_update_local_reginfos();
	}

	ast_sip_update_user_reginfos();
	ast_sip_check_update_start();
    ast_sip_update_accounttk_reginfos();
}

static char *ast_sip_show_account_tks_reginfo(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct ast_sip_account_tk_info *info;
	int  tk_num = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show account trunk reginfo";
		e->usage =
			"Usage: pjsip show account trunk reginfo\n"
			"       This will list registered account trunk info \n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	i = ao2_iterator_init(account_tk_infos, 0);
	for (; (info = ao2_iterator_next(&i)); ao2_ref(info, -1)) {
		ast_cli(a->fd, "\nAccount Trunk: %s\t username: %s\t Address: %s\n", info->endpoint, info->username, info->address);
		tk_num++;
	}
	ao2_iterator_destroy(&i);

	ast_cli(a->fd, "\n=== %d Account trunk registered ===\n\n", tk_num);

	return CLI_SUCCESS;
}

static char *ast_sip_show_account_tks(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct ast_sip_account_tk_conf *info;
	int  tk_num = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show account trunks";
		e->usage =
			"Usage: pjsip show account trunks\n"
			"       This will list account trunks\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

    i = ao2_iterator_init(account_tk_confs, 0);
	for (; (info = ao2_iterator_next(&i)); ao2_ref(info, -1)) {
		ast_cli(a->fd, "\nAccount Trunk: %s\t username: %s\t transport: %s\n", info->endpoint, info->username, info->transport);
		tk_num++;
	}
	ao2_iterator_destroy(&i);

	ast_cli(a->fd, "\n=== Total %d account trunks ===\n\n", tk_num);

	return CLI_SUCCESS;
}

static char *ast_sip_show_reginfos(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct ast_sip_reginfo *info;
	unsigned int max_rc_users  = 0;
	unsigned int max_lcl_users = 0;
	unsigned int rc_count    = 0;
	unsigned int local_count = 0;

	RAII_VAR(struct ast_json *, json_localtype, ast_json_object_create(), ast_json_unref);
	RAII_VAR(struct ast_json *, json_rctype, ast_json_object_create(), ast_json_unref);
	RAII_VAR(struct ast_json *, json_local, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_remote, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_all, NULL, ast_json_unref);

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show reginfos";
		e->usage =
			"Usage: pjsip show reginfos\n"
			"       This will list the maximum register info \n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	max_rc_users = ast_sip_get_max_rc_users();
	max_lcl_users = ast_sip_get_max_local_users();

	if (max_lcl_users) {
		i = ao2_iterator_init(local_reginfos, 0);
		for (; (info = ao2_iterator_next(&i)); ao2_ref(info, -1)) {
			ao2_lock(info);
			ast_json_object_set(json_localtype, info->user_agent, ast_json_integer_create(info->number));
			local_count = local_count + info->number;
			ao2_unlock(info);
		}
		ao2_iterator_destroy(&i);

		if (max_lcl_users > local_count) {
			ast_json_object_set(json_localtype, "Others", ast_json_integer_create(max_lcl_users-local_count));
		}
	}

	if (max_rc_users) {
		i = ao2_iterator_init(rc_reginfos, 0);
		for (; (info = ao2_iterator_next(&i)); ao2_ref(info, -1)) {
			ao2_lock(info);
			ast_json_object_set(json_rctype, info->user_agent, ast_json_integer_create(info->number));
			rc_count = rc_count + info->number;
			ao2_unlock(info);
		}
		ao2_iterator_destroy(&i);

		if (max_rc_users > rc_count) {
			ast_json_object_set(json_rctype, "Others", ast_json_integer_create(max_rc_users-rc_count));
		}
	}

	json_local = ast_json_pack("{s: i, s: O}",
							  "maxLocalUsers", max_lcl_users,
							  "type",          json_localtype);

	json_remote = ast_json_pack("{s: i, s: O}",
							  "maxRcUsers", max_rc_users,
							  "type",       json_rctype);

	json_all = ast_json_pack("{s: O, s: O}",
							  "local",  json_local,
							  "remote", json_remote);

	ast_cli(a->fd, "%s\n", ast_json_dump_string_format(json_all, AST_JSON_PRETTY));

	return CLI_SUCCESS;
}


static char *ast_sip_reset_reginfos(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip reset reginfos";
		e->usage =
			"Usage: pjsip reset reginfos\n"
			"       This will reset the max record of the registration.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_sip_update_rc_reginfos();
	ast_sip_update_local_reginfos();
	ast_cli(a->fd, "=== %d MaxRCUsers, %d MaxLocalUsers ===\n\n", ast_sip_get_max_rc_users(), ast_sip_get_max_local_users());

	return CLI_SUCCESS;
}

static char *ast_sip_show_max_reginfos(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct ast_sip_reginfo *info;
	unsigned int max_users  = 0;
	unsigned int rc_count    = 0;
	unsigned int local_count = 0;

	RAII_VAR(struct ast_json *, json_localtype, ast_json_object_create(), ast_json_unref);
	RAII_VAR(struct ast_json *, json_rctype, ast_json_object_create(), ast_json_unref);
	RAII_VAR(struct ast_json *, json_local, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_remote, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_all, NULL, ast_json_unref);

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show max reginfos";
		e->usage =
			"Usage: pjsip show max reginfos\n"
			"       This will list the maximum register info \n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	max_users = ast_sip_get_max_reg_users();

	if (max_users) {
		i = ao2_iterator_init(user_reginfos, 0);
		for (; (info = ao2_iterator_next(&i)); ao2_ref(info, -1)) {
			ao2_lock(info);
			if (info->rc_flag) {
				ast_json_object_set(json_rctype, info->user_agent, ast_json_integer_create(info->number));
				rc_count = rc_count + info->number;
			} else {
				ast_json_object_set(json_localtype, info->user_agent, ast_json_integer_create(info->number));
				local_count = local_count + info->number;
			}
			
			ao2_unlock(info);
		}
		ao2_iterator_destroy(&i);
	}

	json_local = ast_json_pack("{s: i, s: O}",
							  "localUsers", local_count,
							  "type",       json_localtype);

	json_remote = ast_json_pack("{s: i, s: O}",
							  "rcUsers", rc_count,
							  "type",    json_rctype);

	json_all = ast_json_pack("{s: O, s: O}",
							  "local",  json_local,
							  "remote", json_remote);

	ast_cli(a->fd, "%s\n", ast_json_dump_string_format(json_all, AST_JSON_PRETTY));

	return CLI_SUCCESS;
}

static char *ast_sip_reset_max_reginfos(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip reset max reginfos";
		e->usage =
			"Usage: pjsip reset max reginfos\n"
			"       This will reset the max record of the total registration.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_sip_update_user_reginfos();
	ast_cli(a->fd, "=== %d Max Registered Users ===\n\n", ast_sip_get_max_reg_users());

	return CLI_SUCCESS;
}

static char *ast_sip_show_overview(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator iter_aor;
	struct ao2_iterator iter_contact;
	struct ao2_container *contact_cons = NULL;
	struct ast_sip_aor *aor = NULL;
	struct ast_sip_contact *contact = NULL;
	unsigned int reg_num   = 0;
	unsigned int unreg_num = 0;
	unsigned int proxy_user_num = 0;
	unsigned int local_udp_num = 0;
	unsigned int local_tcp_num = 0;
	unsigned int local_tls_num = 0;
	unsigned int local_ws_num = 0;
	char proxy_host[128] = {0};
	char udp_host[128] = {0};
	char tcp_host[128] = {0};
	char tls_host[128] = {0};
	char ws_host[128]  = {0};
	char mac[16]  = {0};
	char *proto   = NULL;
	RAII_VAR(struct ast_json *, json_final, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_info, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_extension, NULL, ast_json_unref);
	RAII_VAR(struct ast_json *, json_server, NULL, ast_json_unref);
	RAII_VAR(struct ao2_container *, aor_cons, NULL, ao2_cleanup);

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show device overview";
		e->usage =
			"Usage: pjsip show device overview\n"
			"       This will display the sip devices overview information. \n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_sip_get_nvram_value(":WsProxyHost", proxy_host, sizeof(proxy_host));
	ast_sip_get_nvram_value("4511", mac, sizeof(mac));
	if (!strlen(mac) || !strlen(proxy_host)) {
		return NULL;
	}

	ast_sip_check_host_ip();
	snprintf(udp_host, sizeof(udp_host), "%s (UDP)", ast_sip_get_host_ip_string(pj_AF_INET()));
	snprintf(tcp_host, sizeof(tcp_host), "%s (TCP)", ast_sip_get_host_ip_string(pj_AF_INET()));
	snprintf(tls_host, sizeof(tls_host), "%s (TLS)", ast_sip_get_host_ip_string(pj_AF_INET()));
	snprintf(ws_host, sizeof(ws_host), "%s (WS)", ast_sip_get_host_ip_string(pj_AF_INET()));

	aor_cons = cli_get_aors();
	if (ao2_container_count(aor_cons)) {
		iter_aor = ao2_iterator_init(aor_cons, 0);
		for (; (aor = ao2_iterator_next(&iter_aor)); ao2_ref(aor, -1)) {
			if (!strncmp("conf", ast_sorcery_object_get_id(aor), 4) || !strncmp("trunk_", ast_sorcery_object_get_id(aor), 6)) {
				continue;
			}

			contact_cons = ast_sip_location_retrieve_aor_contacts(aor);
			if (contact_cons && ao2_container_count(contact_cons)) {
				reg_num++;

				iter_contact = ao2_iterator_init(contact_cons, 0);
				for (; (contact = ao2_iterator_next(&iter_contact)); ao2_ref(contact, -1)) {
					if (contact->ssh_proxy_flag) {
						proxy_user_num++;
					} else {
						proto = strstr(contact->uri, "transport=");
						if (!proto) {
							local_udp_num++;
						} else {
							proto = proto+strlen("transport=");
							if (!strcasecmp(proto, "ws")) {
								local_ws_num++;
							} else if (!strcasecmp(proto, "tcp")) {
								local_tcp_num++;
							} else {
								local_tls_num++;
							}
						}
					}
				}

				ao2_iterator_cleanup(&iter_contact);
				ao2_cleanup(contact_cons);
			} else {
				unreg_num++;
			}
		}

		ao2_iterator_cleanup(&iter_aor);
	}

	json_extension = ast_json_pack("{s: i, s: i}",
								   "registered",   reg_num,
								   "unregistered", unreg_num);

	json_server = ast_json_pack("{s: i, s: i, s: i, s: i, s: i}",
								proxy_host, proxy_user_num,
								udp_host, local_udp_num,
								tcp_host, local_tcp_num,
								tls_host, local_tls_num,
								ws_host,  local_ws_num);

	json_info = ast_json_pack("{s: O, s: O}",
							  "extension", json_extension,
							  "sipServer", json_server);


	json_final = ast_json_pack("{s: s, s: O}",
							  "mac", mac,
							  "info", json_info);

	ast_cli(a->fd, "%s\n", ast_json_dump_string_format(json_final, AST_JSON_PRETTY));

	return CLI_SUCCESS;
}

unsigned int ast_sip_get_current_proxy_user_num(void)
{
	unsigned int proxy_user_num = 0;

	ast_rwlock_wrlock(&proxy_user_count_lock);
	proxy_user_num = g_proxy_user_num;
	ast_rwlock_unlock(&proxy_user_count_lock);

	return proxy_user_num;
}

unsigned int ast_sip_inc_current_proxy_user_num(void)
{
	ast_rwlock_wrlock(&proxy_user_count_lock);
	g_proxy_user_num++;
	ast_rwlock_unlock(&proxy_user_count_lock);

	ast_log(LOG_DEBUG, "g_proxy_user_num is [%d] now.\n", g_proxy_user_num);
	return g_proxy_user_num;
}

unsigned int ast_sip_dec_current_proxy_user_num(void)
{
	ast_rwlock_wrlock(&proxy_user_count_lock);
	if (g_proxy_user_num > 0) {
		g_proxy_user_num--;
	}
	ast_rwlock_unlock(&proxy_user_count_lock);

	ast_log(LOG_DEBUG, "g_proxy_user_num is [%d] now.\n", g_proxy_user_num);
	return g_proxy_user_num;
}

unsigned int ast_sip_get_current_local_user_num(void)
{
	unsigned int user_num = 0;

	ast_rwlock_wrlock(&local_user_count_lock);
	user_num = g_local_user_num;
	ast_rwlock_unlock(&local_user_count_lock);

	return user_num;
}

unsigned int ast_sip_set_current_local_user_num(unsigned int num)
{
	ast_rwlock_wrlock(&local_user_count_lock);
	g_local_user_num = num;
	ast_rwlock_unlock(&local_user_count_lock);

	return num;
}

unsigned int ast_sip_inc_current_local_user_num(void)
{
	ast_rwlock_wrlock(&local_user_count_lock);
	g_local_user_num++;
	ast_rwlock_unlock(&local_user_count_lock);

	ast_log(LOG_DEBUG, "g_local_user_num is [%d] now.\n", g_local_user_num);
	return g_local_user_num;
}

unsigned int ast_sip_dec_current_local_user_num(void)
{
	ast_rwlock_wrlock(&local_user_count_lock);
	if (g_local_user_num > 0) {
		g_local_user_num--;
	}
	ast_rwlock_unlock(&local_user_count_lock);

	ast_log(LOG_DEBUG, "g_local_user_num is [%d] now.\n", g_local_user_num);
	return g_local_user_num;
}


static int ast_sip_reginfo_cmp(void *obj, void *arg, int flags)
{
	struct ast_sip_reginfo *info1 = obj, *info2 = arg;
	if (info1->rc_flag != info2->rc_flag) {
		return 0;
	}

	return !strcasecmp(info1->user_agent, info2->user_agent) ? CMP_MATCH | CMP_STOP : 0;
}

static void ast_sip_reginfo_destroy(void)
{
	if (local_update_sched) {
		AST_SCHED_DEL(local_update_sched, update_sched_id);
		ast_sched_context_destroy(local_update_sched);
		local_update_sched = NULL;
	}

	ao2_callback(rc_reginfos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
	ao2_callback(local_reginfos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
	ao2_callback(user_reginfos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
    ao2_callback(account_tk_infos, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
    ao2_callback(account_tk_confs, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);

	ao2_ref(rc_reginfos, -1);
	ao2_ref(local_reginfos, -1);
	ao2_ref(user_reginfos, -1);
	ao2_ref(account_tk_infos, -1);
	ao2_ref(account_tk_confs, -1);
}

unsigned int ast_sip_check_proxy_user_num_limit(void)
{
	int ret = 0;

	unsigned int max_registrar = ast_sip_get_turn_max_registrar();
	if (!max_registrar) {
		return 0;
	}

	ast_rwlock_wrlock(&proxy_user_count_lock);
	if (max_registrar <= g_proxy_user_num) {
		ast_log(LOG_WARNING, "The g_proxy_user_num[%d] has reached the turn_max_registrar[%d] now.\n", g_proxy_user_num, max_registrar);
		ret = 1;
	}

	ast_rwlock_unlock(&proxy_user_count_lock);
	return ret;
}

#endif

/*! \brief Destructor for AOR */
static void aor_destroy(void *obj)
{
	struct ast_sip_aor *aor = obj;

	ao2_cleanup(aor->permanent_contacts);
	ast_string_field_free_memory(aor);
	ast_free(aor->voicemail_extension);
}

/*! \brief Allocator for AOR */
static void *aor_alloc(const char *name)
{
	void *lock;
	struct ast_sip_aor *aor;

	lock = ast_named_lock_get(AST_NAMED_LOCK_TYPE_MUTEX, "aor", name);
	if (!lock) {
		return NULL;
	}

	aor = ast_sorcery_lockable_alloc(sizeof(struct ast_sip_aor), aor_destroy, lock);
	ao2_ref(lock, -1);

	if (!aor) {
		return NULL;
	}
	ast_string_field_init(aor, 128);

	return aor;
}

/*! \brief Internal callback function which destroys the specified contact */
static int destroy_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;

	ast_sip_location_delete_contact(contact);

	return CMP_MATCH;
}

static void aor_deleted_observer(const void *object)
{
	const struct ast_sip_aor *aor = object;
	const char *aor_id = ast_sorcery_object_get_id(object);
	/* Give enough space for ;@ at the end, since that is our object naming scheme */
#ifdef GRANDSTREAM_NETWORKS
	size_t prefix_len = strlen(aor_id) * 2 + sizeof(";@") - 1;
	char prefix[prefix_len + 1];
	char tmp_prefix[prefix_len + 1];
	struct ao2_container *contacts;
#else
	size_t prefix_len = strlen(aor_id) + sizeof(";@") - 1;
	char prefix[prefix_len + 1];
	struct ao2_container *contacts;
#endif

	if (aor->permanent_contacts) {
		ao2_callback(aor->permanent_contacts, OBJ_NODATA | OBJ_MULTIPLE | OBJ_UNLINK, destroy_contact, NULL);
	}


#ifdef GRANDSTREAM_NETWORKS
	sprintf(tmp_prefix, "%s;@", aor_id); /* Safe */
	int i = 0, j = 0;
	for (i =0; i< strlen (tmp_prefix); i++, j++) {
		if (tmp_prefix[i] == '+') {
			prefix[j++] = '\\';
		}
		prefix[j] = tmp_prefix[i];
	}
	prefix[j]='\0';
#else
	sprintf(prefix, "%s;@", aor_id); /* Safe */
#endif
	if (!(contacts = ast_sorcery_retrieve_by_prefix(ast_sip_get_sorcery(), "contact", prefix, prefix_len))) {
		return;
	}
	/* Destroy any contacts that may still exist that were made for this AoR */
	ao2_callback(contacts, OBJ_NODATA | OBJ_MULTIPLE | OBJ_UNLINK, destroy_contact, NULL);

	ao2_ref(contacts, -1);
}

/*! \brief Observer for contacts so state can be updated on respective endpoints */
static const struct ast_sorcery_observer aor_observer = {
	.deleted = aor_deleted_observer,
};


/*! \brief Destructor for contact */
static void contact_destroy(void *obj)
{
	struct ast_sip_contact *contact = obj;

	ast_string_field_free_memory(contact);
	ao2_cleanup(contact->endpoint);
}

/*! \brief Allocator for contact */
static void *contact_alloc(const char *name)
{
	struct ast_sip_contact *contact = ast_sorcery_generic_alloc(sizeof(*contact), contact_destroy);
	char *id = ast_strdupa(name);
	char *aor = id;
	char *aor_separator = NULL;

	if (!contact) {
		return NULL;
	}

	if (ast_string_field_init(contact, 256)) {
		ao2_cleanup(contact);
		return NULL;
	}

	/* Dynamic contacts are delimited with ";@" and static ones with "@@" */
	if ((aor_separator = strstr(id, ";@")) || (aor_separator = strstr(id, "@@"))) {
		*aor_separator = '\0';
	}
	ast_assert(aor_separator != NULL);

	ast_string_field_set(contact, aor, aor);

	return contact;
}

struct ast_sip_aor *ast_sip_location_retrieve_aor(const char *aor_name)
{
	return ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "aor", aor_name);
}

/*! \brief Internal callback function which deletes and unlinks any expired contacts */
static int contact_expire(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;

	/* If the contact has not yet expired it is valid */
	if (ast_tvdiff_ms(contact->expiration_time, ast_tvnow()) > 0) {
		return 0;
	}

	ast_sip_location_delete_contact(contact);

	return CMP_MATCH;
}

/*! \brief Internal callback function which links static contacts into another container */
static int contact_link_static(void *obj, void *arg, int flags)
{
	struct ao2_container *dest = arg;

	ao2_link(dest, obj);
	return 0;
}

/*! \brief Internal callback function which removes any contact which is unreachable */
static int contact_remove_unreachable(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct ast_sip_contact_status *status;
	int unreachable;

	status = ast_sip_get_contact_status(contact);
	if (!status) {
		return 0;
	}

	unreachable = (status->status == UNAVAILABLE);
	ao2_ref(status, -1);

	return unreachable ? CMP_MATCH : 0;
}

struct ast_sip_contact *ast_sip_location_retrieve_first_aor_contact(const struct ast_sip_aor *aor)
{
	return ast_sip_location_retrieve_first_aor_contact_filtered(aor, AST_SIP_CONTACT_FILTER_DEFAULT);
}

struct ast_sip_contact *ast_sip_location_retrieve_first_aor_contact_filtered(const struct ast_sip_aor *aor,
	unsigned int flags)
{
	struct ao2_container *contacts;
	struct ast_sip_contact *contact = NULL;

	contacts = ast_sip_location_retrieve_aor_contacts_filtered(aor, flags);
	if (contacts && ao2_container_count(contacts)) {
		/* Get the first AOR contact in the container. */
		contact = ao2_callback(contacts, 0, NULL, NULL);
	}
	ao2_cleanup(contacts);
	return contact;
}

struct ao2_container *ast_sip_location_retrieve_aor_contacts_nolock(const struct ast_sip_aor *aor)
{
	return ast_sip_location_retrieve_aor_contacts_nolock_filtered(aor, AST_SIP_CONTACT_FILTER_DEFAULT);
}

struct ao2_container *ast_sip_location_retrieve_aor_contacts_nolock_filtered(const struct ast_sip_aor *aor,
	unsigned int flags)
{
	/* Give enough space for ;@ at the end, since that is our object naming scheme */
#ifdef GRANDSTREAM_NETWORKS
	size_t prefix_len = strlen(ast_sorcery_object_get_id(aor)) * 2 + sizeof(";@") - 1;
	char prefix[prefix_len + 1];
	char tmp_prefix[prefix_len + 1];
	struct ao2_container *contacts;

	sprintf(tmp_prefix, "%s;@", ast_sorcery_object_get_id(aor)); /* Safe */
	int i = 0, j = 0;
	for (i =0; i< strlen (tmp_prefix); i++, j++) {
		if (tmp_prefix[i] == '+') {
			prefix[j++] = '\\';
		}
		prefix[j] = tmp_prefix[i];
	}
	prefix[j]='\0';
#else
	size_t prefix_len = strlen(ast_sorcery_object_get_id(aor)) + sizeof(";@") - 1;
	char prefix[prefix_len + 1];
	struct ao2_container *contacts;

	sprintf(prefix, "%s;@", ast_sorcery_object_get_id(aor)); /* Safe */
#endif
	if (!(contacts = ast_sorcery_retrieve_by_prefix(ast_sip_get_sorcery(), "contact", prefix, prefix_len))) {
		return NULL;
	}

	/* Prune any expired contacts and delete them, we do this first because static contacts can never expire */
	ao2_callback(contacts, OBJ_NODATA | OBJ_MULTIPLE | OBJ_UNLINK, contact_expire, NULL);

	/* Add any permanent contacts from the AOR */
	if (aor->permanent_contacts) {
		ao2_callback(aor->permanent_contacts, OBJ_NODATA, contact_link_static, contacts);
	}

	if (flags & AST_SIP_CONTACT_FILTER_REACHABLE) {
		ao2_callback(contacts, OBJ_NODATA | OBJ_MULTIPLE | OBJ_UNLINK, contact_remove_unreachable, NULL);
	}

	return contacts;
}

struct ao2_container *ast_sip_location_retrieve_aor_contacts(const struct ast_sip_aor *aor)
{
	return ast_sip_location_retrieve_aor_contacts_filtered(aor, AST_SIP_CONTACT_FILTER_DEFAULT);
}

struct ao2_container *ast_sip_location_retrieve_aor_contacts_filtered(const struct ast_sip_aor *aor,
	unsigned int flags)
{
	struct ao2_container *contacts;

	/* ao2_lock / ao2_unlock do not actually write aor since it has an ao2 lockobj. */
	ao2_lock((void*)aor);
	contacts = ast_sip_location_retrieve_aor_contacts_nolock_filtered(aor, flags);
	ao2_unlock((void*)aor);

	return contacts;
}


void ast_sip_location_retrieve_contact_and_aor_from_list(const char *aor_list, struct ast_sip_aor **aor,
	struct ast_sip_contact **contact)
{
	ast_sip_location_retrieve_contact_and_aor_from_list_filtered(aor_list, AST_SIP_CONTACT_FILTER_DEFAULT, aor, contact);
}

void ast_sip_location_retrieve_contact_and_aor_from_list_filtered(const char *aor_list, unsigned int flags,
	struct ast_sip_aor **aor, struct ast_sip_contact **contact)
{
	char *aor_name;
	char *rest;

	/* If the location is still empty we have nowhere to go */
	if (ast_strlen_zero(aor_list) || !(rest = ast_strdupa(aor_list))) {
		ast_log(LOG_WARNING, "Unable to determine contacts from empty aor list\n");
		return;
	}

	*aor = NULL;
	*contact = NULL;

	while ((aor_name = ast_strip(strsep(&rest, ",")))) {
		*aor = ast_sip_location_retrieve_aor(aor_name);

		if (!(*aor)) {
			continue;
		}
		*contact = ast_sip_location_retrieve_first_aor_contact_filtered(*aor, flags);
		/* If a valid contact is available use its URI for dialing */
		if (*contact) {
			break;
		}

		ao2_ref(*aor, -1);
		*aor = NULL;
	}
}

struct ast_sip_contact *ast_sip_location_retrieve_contact_from_aor_list(const char *aor_list)
{
	struct ast_sip_aor *aor;
	struct ast_sip_contact *contact;

	ast_sip_location_retrieve_contact_and_aor_from_list(aor_list, &aor, &contact);

	ao2_cleanup(aor);

	return contact;
}

#ifdef GRANDSTREAM_NETWORKS
struct ast_sip_contact *ast_sip_location_retrieve_contact_from_aor_uri(const char *aor_name, const char *uri)
{
	RAII_VAR(struct ast_sip_aor *, aor, ast_sip_location_retrieve_aor(aor_name), ao2_cleanup);
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	struct ao2_iterator it_contacts;
	struct ast_sip_contact *contact;

	if (!aor_name || !uri || !aor) {
		return NULL;
	}

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", aor_name);
	if (!endpoint) {
		return NULL;
	}

	/* Added by yfyang, for Bug 297978*/
	do {
		if (endpoint->sca_enable && endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_SHARED) {
			char aor_priv[80] = {0};
			char *pos1 = strstr(uri, ":");
			if (!pos1) {
				break;
			}

			pos1++;
			sscanf(pos1, "%79[^@]", aor_priv);
			ast_log(LOG_DEBUG, "retrieve_contact_from_aor_uri, aor_priv[%s]", aor_priv);

			if (strlen(aor_priv) && strcmp(aor_name, aor_priv)) {
				ao2_cleanup(aor);
				aor = ast_sip_location_retrieve_aor(aor_priv);
				if (!aor) {
					return NULL;
				}
			}
		}
	} while (0);

	if (!(contacts = ast_sip_location_retrieve_aor_contacts_filtered(aor, AST_SIP_CONTACT_FILTER_REACHABLE))) {
		return NULL;
	}

	if (!ao2_container_count(contacts)) {
		return NULL;
	}

	it_contacts = ao2_iterator_init(contacts, 0);
	for (; (contact = ao2_iterator_next(&it_contacts)); ao2_ref(contact, -1)) {
		if (!strcmp(uri, contact->uri)) {
			//ast_log(LOG_NOTICE, "Found contact[%p], aor[%s], uri[%s]", contact, aor_name, uri);
			ao2_iterator_destroy(&it_contacts);
			return contact;
		}
	}

	ao2_iterator_destroy(&it_contacts);
	return NULL;
}

#endif

static int permanent_uri_sort_fn(const void *obj_left, const void *obj_right, int flags);
static int cli_contact_populate_container(void *obj, void *arg, int flags);

static int gather_contacts_for_aor(void *obj, void *arg, int flags)
{
	struct ao2_container *aor_contacts;
	struct ast_sip_aor *aor = obj;
	struct ao2_container *container = arg;

	aor_contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!aor_contacts) {
		return 0;
	}
	ao2_callback(aor_contacts, OBJ_MULTIPLE | OBJ_NODATA, cli_contact_populate_container,
		container);
	ao2_ref(aor_contacts, -1);
	return CMP_MATCH;
}

struct ao2_container *ast_sip_location_retrieve_contacts_from_aor_list(const char *aor_list)
{
	struct ao2_container *contacts;

	contacts = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK,
		AO2_CONTAINER_ALLOC_OPT_DUPS_REJECT, permanent_uri_sort_fn, NULL);
	if (!contacts) {
		return NULL;
	}

	ast_sip_for_each_aor(aor_list, gather_contacts_for_aor, contacts);

	return contacts;
}

struct ast_sip_contact *ast_sip_location_retrieve_contact(const char *contact_name)
{
	return ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "contact", contact_name);
}

#ifdef GRANDSTREAM_NETWORKS


struct ast_sip_contact *ast_sip_location_create_contact(struct ast_sip_aor *aor,
	const char *uri, struct timeval expiration_time, const char *path_info,
	const char *user_agent, const char *via_addr, int via_port, const char *call_id,
	int prune_on_boot, struct ast_sip_endpoint *endpoint,
	const char *gs_appid,  const char *original_address, const char *sbc_endpoint_addr, pjsip_rx_data *rdata,
	struct ast_sip_contact_extra_info extra_info)
#else
struct ast_sip_contact *ast_sip_location_create_contact(struct ast_sip_aor *aor,
	const char *uri, struct timeval expiration_time, const char *path_info,
	const char *user_agent, const char *via_addr, int via_port, const char *call_id,
	int prune_on_boot, struct ast_sip_endpoint *endpoint)
#endif
{
	struct ast_sip_contact *contact;
#ifdef GRANDSTREAM_NETWORKS
	char name[MAX_OBJECT_FIELD * 4 + 6];
	char tmp_name[MAX_OBJECT_FIELD * 2 + 3];
	char hash[33];

	if (extra_info.ssh_proxy_flag) {
		if (ast_sip_check_proxy_user_num_limit()) {
			ast_log(LOG_WARNING, "MaxRegistrar reached, can't create contact[%s] now.\n", uri);
			return NULL;
		}
	}
#ifdef BUILD_ON_CLOUD_PBX
	if (!ast_strlen_zero(sbc_endpoint_addr)) {
		ast_md5_hash(hash, sbc_endpoint_addr);
	} else {
		ast_md5_hash(hash, uri);
	}
#else
	ast_md5_hash(hash, uri);
#endif

	snprintf(tmp_name, sizeof(tmp_name),"%s;@%s", ast_sorcery_object_get_id(aor), hash);
	int i = 0, j = 0;
	for (i =0; i< strlen (tmp_name); i++, j++) {
		if (tmp_name[i] == '+') {
			name[j++] = '\\';
		}
		name[j] = tmp_name[i];
	}
	name[j]='\0';
#else
	char name[MAX_OBJECT_FIELD * 2 + 3];
	char hash[33];

	ast_md5_hash(hash, uri);
	snprintf(name, sizeof(name), "%s;@%s", ast_sorcery_object_get_id(aor), hash);
#endif
	contact = ast_sorcery_alloc(ast_sip_get_sorcery(), "contact", name);
	if (!contact) {
		return NULL;
	}

	ast_string_field_set(contact, uri, uri);
	contact->expiration_time = expiration_time;
	contact->qualify_frequency = aor->qualify_frequency;
	contact->qualify_timeout = aor->qualify_timeout;
	contact->authenticate_qualify = aor->authenticate_qualify;
	if (path_info && aor->support_path) {
		ast_string_field_set(contact, path, path_info);
	}

	if (!ast_strlen_zero(aor->outbound_proxy)) {
		ast_string_field_set(contact, outbound_proxy, aor->outbound_proxy);
#ifdef GRANDSTREAM_NETWORKS
		contact->rmv_obp_from_route = aor->rmv_obp_from_route;
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_strlen_zero(original_address)) {
		ast_string_field_set(contact, original_address, original_address);
	}

	if (!ast_strlen_zero(extra_info.to_host)) {
		ast_string_field_set(contact, to_host, extra_info.to_host);
	}

	if (!ast_strlen_zero(extra_info.instance)) {
		ast_string_field_set(contact, instance, extra_info.instance);
	}

#ifdef BUILD_ON_CLOUD_PBX
	if (!ast_strlen_zero(sbc_endpoint_addr)) {
		ast_string_field_set(contact, sbc_endpoint_addr, sbc_endpoint_addr);
	}
#endif
	contact->support_missed_call = extra_info.support_missed_call;  /* Added by yfyang, for missed call feature */
	contact->support_ldap_settings = extra_info.support_ldap_settings;                  /* Added by yfyang, for ldap settings sync feature */
	contact->ssh_proxy_flag = extra_info.ssh_proxy_flag;            /* Added by yfyang, for sip ssh proxy feature */
	contact->support_ice = extra_info.support_ice;  /* Added by bfdu, for ice support */
	contact->support_trickle_ice = extra_info.support_trickle_ice;	/* Added by bfdu, for ice support */

	contact->gswave_mobile = extra_info.gswave_mobile; /* Added by bfdu, for record gswave mobile */

	contact->support_gs_config = extra_info.support_gs_config;
	contact->support_gs_state = extra_info.support_gs_state;
	contact->support_gs_fec = extra_info.support_gs_fec;
	contact->support_gs_message = extra_info.support_gs_message;
	contact->support_gs_conf_invite = extra_info.support_gs_conf_invite;
	contact->support_gs_control = extra_info.support_gs_control;
	contact->require_feature_code_settings = extra_info.require_feature_code_settings;
	contact->require_turn_info_settings = extra_info.require_turn_info_settings;
	contact->require_basic_settings = extra_info.require_basic_settings;
	contact->require_plugin = extra_info.require_plugin;
	contact->require_phonebook = extra_info.require_phonebook;
    contact->reg_id = extra_info.reg_id;

	/*Start added by zyxia for callflip  when new register */
	if (contact->support_gs_state && ast_sip_get_callflip_enable() )
	{
		//RAII_VAR(struct callflip_user *, callflip_state, NULL, ao2_cleanup);
		struct ast_sip_callflip_state callflip_info;
		memset(&callflip_info, 0, sizeof(struct ast_sip_callflip_state));
		struct callflip_user *callflip_state;
		if ((callflip_state = ast_sip_callflip_find_user(endpoint->aors))) {
			snprintf(callflip_info.enable, sizeof(callflip_info.enable), "%s", callflip_state->enable);
			snprintf(callflip_info.channel_id, sizeof(callflip_info.channel_id), "%s", callflip_state->channel_id);
			if ( ast_sip_check_if_need_sendcallflip(contact,&callflip_info)) { //edit by zyxia for Bug 261256,2023.5.30
				snprintf(callflip_info.remote_id, sizeof(callflip_info.remote_id), "%s", callflip_state->remote_id);
				snprintf(callflip_info.remote_name, sizeof(callflip_info.remote_name), "%s",callflip_state->remote_name);
				snprintf(callflip_info.call_type, sizeof(callflip_info.call_type), "%s", callflip_state->call_type);
				snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", callflip_state->media_type);
				snprintf(callflip_info.conf_link, sizeof(callflip_info.conf_link), "%s", callflip_state->conf_link);
				snprintf(callflip_info.replace_channel_id, sizeof(callflip_info.replace_channel_id), "%s", callflip_state->replace_channel_id);
				ast_sip_send_notify_to_update_gs_state(contact,GS_STATE_TYPE_CALLFLIP,&callflip_info);
				ast_log(LOG_DEBUG, "Contact [%s] now reached, need send CallFlip Notify.\n", uri);
			}
		}
	}
	/*End added by zyxia for callflip  when new register */
#endif

	
#ifdef GRANDSTREAM_NETWORKS
	/*Start added by zyxia for WakeUp wave app*/
	if (contact->gswave_mobile && !ast_strlen_zero(contact->instance)) {
		char key[128]={0},key_value[1024]={0};
		unsigned int update_new_uri = 0;
		struct timeval register_time_val;
		int old_app=0;

		//old app instance have <urn:uuid:00000000-0000,if not have ,is new app
		if(!ast_strlen_zero(contact->instance) && strstr(contact->instance,"<urn:uuid:00000000-0000") && !ast_strlen_zero(gs_appid)) {
			memset(key_value, 0, sizeof(key_value));
			snprintf(key,sizeof(key),"%s/%s",endpoint->aors,gs_appid);
			ast_db_get(GSWAVE_DB_FAMILY, key, key_value, sizeof(key_value));
			old_app = 1;
		}
		//new app
		else if(!ast_strlen_zero(contact->instance) && !strstr(contact->instance,"<urn:uuid:00000000-0000")) {
			memset(key_value, 0, sizeof(key_value));
			snprintf(key,sizeof(key),"%s/%s",endpoint->aors,contact->instance);
			ast_db_get(GSWAVE_DB_FAMILY, key, key_value, sizeof(key_value));

			//need delete old app info when the app update from old app
			if(!ast_strlen_zero(gs_appid)){
				struct sip_waveapp_info* app_info = NULL;
				app_info = ast_sip_find_waveapp_info(endpoint,gs_appid);
				if(app_info != NULL){
					ast_sip_remove_waveinfo_from_database(endpoint->aors, app_info);
					ast_sip_clear_waveapp_info(endpoint,gs_appid);
				}
			}
		}


		if (ast_strlen_zero(key_value)) { 
			ast_debug(3, "It's new register message contact '%s' on AOR '%s' \n",uri, endpoint->aors);
			update_new_uri = 1;
		}
		//need check the register message is refresh register or push register,if the contact isn't changed,dont refresh
		else if(!ast_strlen_zero(key_value) && !ast_strlen_zero(uri) && strcmp(key_value,uri)) {
			update_new_uri = 1;
			ast_debug(3, "It's not refreshed contact '%s' on AOR '%s',old_uri_value:%s\n",
				uri, endpoint->aors,key_value);
		}
		register_time_val = ast_tvnow();

		if(old_app)
			ast_sip_add_waveapp_info(endpoint, gs_appid, uri, contact->support_gs_conf_invite,register_time_val,update_new_uri,1);
		else
			ast_sip_add_waveapp_info(endpoint, contact->instance, uri, contact->support_gs_conf_invite,register_time_val,update_new_uri,1);

		//added by zyxia for 188323
		ast_sip_update_dlg_and_invite_tsx_transport_by_endponit(endpoint,&rdata->tp_info.transport->key.rem_addr,rdata->tp_info.transport);

		memset(key_value,0,sizeof(key_value));
		snprintf(key_value, sizeof(key_value), "%s,%d,%d",uri, (int)register_time_val.tv_sec,contact->support_gs_conf_invite);
		ast_db_put(GSWAVE_DB_FAMILY, key, key_value);
	}
	/*End added by zyxia for WakeUp ios gswave ,2018.7.25 */
#endif

	if (!ast_strlen_zero(user_agent)) {
		ast_string_field_set(contact, user_agent, user_agent);
	}

	if (!ast_strlen_zero(ast_config_AST_SYSTEM_NAME)) {
		ast_string_field_set(contact, reg_server, ast_config_AST_SYSTEM_NAME);
	}

	if (!ast_strlen_zero(via_addr)) {
		ast_string_field_set(contact, via_addr, via_addr);
	}
	contact->via_port = via_port;

	if (!ast_strlen_zero(call_id)) {
		ast_string_field_set(contact, call_id, call_id);
	}

	contact->endpoint = ao2_bump(endpoint);
	if (endpoint) {
		ast_string_field_set(contact, endpoint_name, ast_sorcery_object_get_id(endpoint));
	}

	contact->prune_on_boot = prune_on_boot;

	if (ast_sorcery_create(ast_sip_get_sorcery(), contact)) {
		ao2_ref(contact, -1);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	unsigned int user_num = 0;
	unsigned int rc_num   = 0;
	unsigned int local_num = 0;

	if (!aor->account_trunk) {
		if (contact->ssh_proxy_flag) {
			rc_num = ast_sip_inc_current_proxy_user_num();
			if (rc_num > ast_sip_get_max_rc_users()) {
				if (!g_need_update_rc) {
					ast_rwlock_wrlock(&reg_user_update_lock);
					g_need_update_rc = 1;
					ast_rwlock_unlock(&reg_user_update_lock);
				}
			}

			local_num = ast_sip_get_current_local_user_num();

		} else {
			local_num = ast_sip_inc_current_local_user_num();
			if (local_num > ast_sip_get_max_local_users()) {
				if (!g_need_update_local) {
					ast_rwlock_wrlock(&reg_user_update_lock);
					g_need_update_local = 1;
					ast_rwlock_unlock(&reg_user_update_lock);
				}
			}

			rc_num = ast_sip_get_current_proxy_user_num();
		}
	}

	user_num = rc_num + local_num;
	if (user_num > ast_sip_get_max_reg_users()) {
		if (!g_need_update_max) {
			ast_rwlock_wrlock(&reg_user_update_lock);
			g_need_update_max = 1;
			ast_rwlock_unlock(&reg_user_update_lock);
		}
	}

	ast_sip_update_account_tk_info(contact);
#endif

	return contact;
}

#ifdef GRANDSTREAM_NETWORKS
/* Added input parameter: uuid, gs_appid, support_missed_call */
int ast_sip_location_add_contact_nolock(struct ast_sip_aor *aor, const char *uri,
		struct timeval expiration_time, const char *path_info, const char *user_agent,
		const char *via_addr, int via_port, const char *call_id,
		struct ast_sip_endpoint *endpoint, const char *gs_appid,
		const char *original_address, const char *sbc_endpoint_addr, pjsip_rx_data *rdata, struct ast_sip_contact_extra_info extra_info)
{
	struct ast_sip_contact *contact;

	contact = ast_sip_location_create_contact(aor, uri, expiration_time, path_info,
		user_agent, via_addr, via_port, call_id, 0, endpoint, gs_appid,
		original_address, sbc_endpoint_addr, rdata, extra_info);

	ao2_cleanup(contact);
	return contact ? 0 : -1;
}
#if 0
int ast_sip_location_add_contact(struct ast_sip_aor *aor, const char *uri,
		struct timeval expiration_time, const char *path_info, const char *user_agent,
		const char *via_addr, int via_port, const char *call_id,
		struct ast_sip_endpoint *endpoint, 
		const char *uuid, const char *gs_appid, int support_missed_call, unsigned int support_ice,
		unsigned int gswave_mobile, int ssh_proxy_flag, const char *original_address,pjsip_rx_data *rdata)
{
	int res;

	ao2_lock(aor);
	res = ast_sip_location_add_contact_nolock(aor, uri, expiration_time, path_info, user_agent,
		via_addr, via_port, call_id,
		endpoint, uuid, gs_appid, support_missed_call,support_ice,
		gswave_mobile, ssh_proxy_flag, original_address,rdata);
	ao2_unlock(aor);

	return res;
}
#endif



int ast_sip_location_add_contact(struct ast_sip_endpoint *endpoint,struct ast_sip_aor *aor, const char *uri,
		struct timeval expiration_time, const char *path_info, const char *user_agent, const char *instance, const char *gs_appid, 
		int support_missed_call, int support_ldap_settings, int support_gs_message,int support_gs_conf_invite, int reg_id)
{
	char name[MAX_OBJECT_FIELD * 2 + 3];
	char hash[128] = {0};
	
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
	/*BEGIN: Add by weidu, for ha sync register contact*/
	ast_md5_hash(hash, uri);
	/*END: Add by weidu*/
	snprintf(name, sizeof(name), "%s;@%s", ast_sorcery_object_get_id(aor), hash);
	ast_log(LOG_DEBUG, "ast_sip_location_add_contact contact %s \n", name);
	if (!(contact = ast_sorcery_alloc(ast_sip_get_sorcery(), "contact", name))) {
		return -1;
	}

	ast_string_field_set(contact, uri, uri);
	contact->expiration_time = expiration_time;
	contact->qualify_frequency = aor->qualify_frequency;
	contact->qualify_timeout = aor->qualify_timeout;
	contact->authenticate_qualify = aor->authenticate_qualify;
	if (path_info && aor->support_path) {
		ast_string_field_set(contact, path, path_info);
	}

	if (!ast_strlen_zero(aor->outbound_proxy)) {
		ast_string_field_set(contact, outbound_proxy, aor->outbound_proxy);
		contact->rmv_obp_from_route = aor->rmv_obp_from_route;

	}

	if (!ast_strlen_zero(user_agent)) {
		ast_string_field_set(contact, user_agent, user_agent);
	}

	if (!ast_strlen_zero(instance)) {
		ast_string_field_set(contact, instance, instance);
	}

	contact->reg_id = reg_id;  /* Added by yfyang, for Bug313577 */
	contact->support_missed_call = support_missed_call;  /* Added by yfyang, for missed call feature */
	contact->support_gs_message = support_gs_message ;   /* Added by yfyang, for missed call feature */
	contact->support_ldap_settings = support_ldap_settings;
	contact->support_gs_conf_invite = support_gs_conf_invite;

	if (ast_sorcery_create(ast_sip_get_sorcery(), contact)) {
		ast_sorcery_delete(ast_sip_get_sorcery(), contact); 
		return -1;
	}

	/*Start added by zyxia for WakeUp ios gswave ,2018.7.25 */
	if (!ast_strlen_zero(gs_appid)) {
		char register_time[256];
		char key[64];
		char *ip_addr = NULL;

		ip_addr = ast_sip_get_ip_addr_from_contact_uri(uri);

		ast_string_field_set(endpoint,gs_uri,ip_addr);
		ast_string_field_set(endpoint,gs_new_contact_uri,uri);
		ast_string_field_set(endpoint, gs_appid, gs_appid);
		endpoint->register_time = ast_tvnow();//set it after update contact

		snprintf(key,sizeof(key),"%s/%s","gs_uri",endpoint->aors);
		ast_db_put(GSWAVE_DB_FAMILY, key, ip_addr);

		snprintf(key,sizeof(key),"%s/%s","gs_appid",endpoint->aors);
		ast_db_put(GSWAVE_DB_FAMILY, key, gs_appid);

		snprintf(key,sizeof(key),"%s/%s","register_time",endpoint->aors);
		snprintf(register_time, sizeof(register_time), "%d", (int)endpoint->register_time.tv_sec);
		ast_db_put(GSWAVE_DB_FAMILY, key, register_time);

		memset(key,0,sizeof(key));
		snprintf(key,sizeof(key),"%s/%s","conf_invite",endpoint->aors);
		if(contact->support_gs_conf_invite){
			ast_db_put(GSWAVE_DB_FAMILY, key, "yes");
		}
		else {
			ast_db_put(GSWAVE_DB_FAMILY, key, "no");
		}

	}
	/*End added by zyxia for WakeUp ios gswave ,2018.7.25 */

	return 0;
}



#else
int ast_sip_location_add_contact_nolock(struct ast_sip_aor *aor, const char *uri,
		struct timeval expiration_time, const char *path_info, const char *user_agent,
		const char *via_addr, int via_port, const char *call_id,
		struct ast_sip_endpoint *endpoint)
{
	struct ast_sip_contact *contact;

	contact = ast_sip_location_create_contact(aor, uri, expiration_time, path_info,
		user_agent, via_addr, via_port, call_id, 0, endpoint);
	ao2_cleanup(contact);
	return contact ? 0 : -1;
}

int ast_sip_location_add_contact(struct ast_sip_aor *aor, const char *uri,
		struct timeval expiration_time, const char *path_info, const char *user_agent,
		const char *via_addr, int via_port, const char *call_id,
		struct ast_sip_endpoint *endpoint)
{
	int res;

	ao2_lock(aor);
	res = ast_sip_location_add_contact_nolock(aor, uri, expiration_time, path_info, user_agent,
		via_addr, via_port, call_id,
		endpoint);
	ao2_unlock(aor);

	return res;
}
#endif

int ast_sip_location_update_contact(struct ast_sip_contact *contact)
{
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 179692 */
	if (contact && contact->ssh_proxy_flag) {
		unsigned int max_registrar = ast_sip_get_turn_max_registrar();

		if (max_registrar) {
			unsigned int current_num = ast_sip_get_current_proxy_user_num();

			if (current_num > max_registrar) {
				ast_log(LOG_NOTICE, "The current_num[%d] has over the turn_max_registrar[%d], can't update contact[%s] now.",
						current_num, max_registrar, contact->uri);
				return -1;
			}
		}
	}

	ast_sip_update_account_tk_info(contact);
#endif

	return ast_sorcery_update(ast_sip_get_sorcery(), contact);
}

int ast_sip_location_delete_contact(struct ast_sip_contact *contact)
{
#ifdef GRANDSTREAM_NETWORKS
	int is_account_tk = 0;

	if (contact && !ast_strlen_zero(contact->endpoint_name) && !strncmp(contact->endpoint_name, "trunk_", 6)) {
		is_account_tk = 1;
	}

	if (!is_account_tk) {
		if (contact && contact->ssh_proxy_flag) {
			ast_sip_dec_current_proxy_user_num();
		} else {
			ast_sip_dec_current_local_user_num();
		}
	} else {
    	ast_sip_delete_account_tk_info(contact);
	}
#endif
	return ast_sorcery_delete(ast_sip_get_sorcery(), contact);
}

static int prune_boot_contacts_cb(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;

	if (contact->prune_on_boot
		&& !strcmp(contact->reg_server, ast_config_AST_SYSTEM_NAME ?: "")) {
		ast_verb(3, "Removed contact '%s' from AOR '%s' due to system boot\n",
			contact->uri, contact->aor);
		ast_sip_location_delete_contact(contact);
	}

	return 0;
}

void ast_sip_location_prune_boot_contacts(void)
{
	struct ao2_container *contacts;

	contacts = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "contact",
		AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL);
	if (contacts) {
		ao2_callback(contacts, 0, prune_boot_contacts_cb, NULL);
		ao2_ref(contacts, -1);
	}
}

/*! \brief Custom handler for translating from a string timeval to actual structure */
static int expiration_str2struct(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ast_sip_contact *contact = obj;
	return ast_get_timeval(var->value, &contact->expiration_time, ast_tv(0, 0), NULL);
}

/*! \brief Custom handler for translating from an actual structure timeval to string */
static int expiration_struct2str(const void *obj, const intptr_t *args, char **buf)
{
	const struct ast_sip_contact *contact = obj;
	return (ast_asprintf(buf, "%ld", contact->expiration_time.tv_sec) < 0) ? -1 : 0;
}

static int permanent_uri_sort_fn(const void *obj_left, const void *obj_right, int flags)
{
	const struct ast_sip_contact *object_left = obj_left;
	const struct ast_sip_contact *object_right = obj_right;
	const char *right_key = obj_right;
	int cmp;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_OBJECT:
		right_key = ast_sorcery_object_get_id(object_right);
		/* Fall through */
	case OBJ_SEARCH_KEY:
		cmp = strcmp(ast_sorcery_object_get_id(object_left), right_key);
		break;
	case OBJ_SEARCH_PARTIAL_KEY:
		/*
		 * We could also use a partial key struct containing a length
		 * so strlen() does not get called for every comparison instead.
		 */
		cmp = strncmp(ast_sorcery_object_get_id(object_left), right_key, strlen(right_key));
		break;
	default:
		/* Sort can only work on something with a full or partial key. */
		ast_assert(0);
		cmp = 0;
		break;
	}
	return cmp;
}

int ast_sip_validate_uri_length(const char *contact_uri)
{
	int max_length = pj_max_hostname - 1;
	char *contact = ast_strdupa(contact_uri);
	char *host;
	char *at;
	int theres_a_port = 0;

	if (strlen(contact_uri) > pjsip_max_url_size - 1) {
		return -1;
	}

	contact = ast_strip_quoted(contact, "<", ">");

	if (!strncasecmp(contact, "sip:", 4)) {
		host = contact + 4;
	} else if (!strncasecmp(contact, "sips:", 5)) {
		host = contact + 5;
	} else {
		/* Not a SIP URI */
		return -1;
	}

	at = strchr(contact, '@');
	if (at) {
		/* sip[s]:user@host */
		host = at + 1;
	}

	if (host[0] == '[') {
		/* Host is an IPv6 address. Just get up to the matching bracket */
		char *close_bracket;

		close_bracket = strchr(host, ']');
		if (!close_bracket) {
			return -1;
		}
		close_bracket++;
		if (*close_bracket == ':') {
			theres_a_port = 1;
		}
		*close_bracket = '\0';
	} else {
		/* uri parameters could contain ';' so trim them off first */
		host = strsep(&host, ";?");
		/* Host is FQDN or IPv4 address. Need to find closing delimiter */
		if (strchr(host, ':')) {
			theres_a_port = 1;
			host = strsep(&host, ":");
		}
	}

	if (!theres_a_port) {
		max_length -= strlen("_sips.tcp.");
	}

	if (strlen(host) > max_length) {
		return -1;
	}

	return 0;
}

/*! \brief Custom handler for permanent URIs */
static int permanent_uri_handler(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ast_sip_aor *aor = obj;
	const char *aor_id = ast_sorcery_object_get_id(aor);
	char *contacts;
	char *contact_uri;

	if (ast_strlen_zero(var->value)) {
		return 0;
	}

	contacts = ast_strdupa(var->value);
	while ((contact_uri = ast_strip(strsep(&contacts, ",")))) {
		struct ast_sip_contact *contact;
		struct ast_sip_contact_status *status;
		char hash[33];
		char contact_id[strlen(aor_id) + sizeof(hash) + 2];

		if (ast_strlen_zero(contact_uri)) {
			continue;
		}

		if (ast_sip_validate_uri_length(contact_uri)) {
			ast_log(LOG_ERROR, "Contact uri or hostname length exceeds pjproject limit or is not a sip(s) uri: %s\n", contact_uri);
			return -1;
		}

		if (!aor->permanent_contacts) {
			aor->permanent_contacts = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK,
				AO2_CONTAINER_ALLOC_OPT_DUPS_REJECT, permanent_uri_sort_fn, NULL);
			if (!aor->permanent_contacts) {
				return -1;
			}
		}

		ast_md5_hash(hash, contact_uri);
		snprintf(contact_id, sizeof(contact_id), "%s@@%s", aor_id, hash);
		contact = ast_sorcery_alloc(ast_sip_get_sorcery(), "contact", contact_id);
		if (!contact) {
			return -1;
		}

		ast_string_field_set(contact, uri, contact_uri);

#ifdef GRANDSTREAM_NETWORKS
		ast_string_field_set(contact, aor, aor_id);
#endif
		status = ast_res_pjsip_find_or_create_contact_status(contact);
		if (!status) {
			ao2_ref(contact, -1);
			return -1;
		}
		ao2_ref(status, -1);

		ao2_link(aor->permanent_contacts, contact);
		ao2_ref(contact, -1);
	}

	return 0;
}

static int contact_to_var_list(void *object, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = object;
	struct ast_variable **var = arg;

	ast_variable_list_append(&*var, ast_variable_new("contact", wrapper->contact->uri, ""));

	return 0;
}

static int contacts_to_var_list(const void *obj, struct ast_variable **fields)
{
	const struct ast_sip_aor *aor = obj;

	ast_sip_for_each_contact(aor, contact_to_var_list, fields);

	return 0;
}

static int voicemail_extension_handler(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ast_sip_aor *aor = obj;

	aor->voicemail_extension = ast_strdup(var->value);

	return aor->voicemail_extension ? 0 : -1;
}

static int voicemail_extension_to_str(const void *obj, const intptr_t *args, char **buf)
{
	const struct ast_sip_aor *aor = obj;

	*buf = ast_strdup(aor->voicemail_extension);

	return 0;
}

int ast_sip_for_each_aor(const char *aors, ao2_callback_fn on_aor, void *arg)
{
	char *copy;
	char *name;
	int res;

	if (!on_aor || ast_strlen_zero(aors)) {
		return 0;
	}

	copy = ast_strdupa(aors);
	while ((name = ast_strip(strsep(&copy, ",")))) {
		struct ast_sip_aor *aor;

		aor = ast_sip_location_retrieve_aor(name);
		if (aor) {
			res = on_aor(aor, arg, 0);
			ao2_ref(aor, -1);
			if (res) {
				return -1;
			}
		}
	}
	return 0;
}

static void contact_wrapper_destroy(void *obj)
{
	struct ast_sip_contact_wrapper *wrapper = obj;

	ast_free(wrapper->aor_id);
	ast_free(wrapper->contact_id);
	ao2_cleanup(wrapper->contact);
}

int ast_sip_for_each_contact(const struct ast_sip_aor *aor,
		ao2_callback_fn on_contact, void *arg)
{
	struct ao2_container *contacts;
	struct ao2_iterator i;
	int res = 0;
	void *object = NULL;

	if (!on_contact ||
	    !(contacts = ast_sip_location_retrieve_aor_contacts(aor))) {
		return 0;
	}

	i = ao2_iterator_init(contacts, 0);
	while ((object = ao2_iterator_next(&i))) {
		RAII_VAR(struct ast_sip_contact *, contact, object, ao2_cleanup);
		RAII_VAR(struct ast_sip_contact_wrapper *, wrapper, NULL, ao2_cleanup);
		const char *aor_id = ast_sorcery_object_get_id(aor);

		wrapper = ao2_alloc_options(sizeof(struct ast_sip_contact_wrapper),
			contact_wrapper_destroy, AO2_ALLOC_OPT_LOCK_NOLOCK);
		if (!wrapper) {
			res = -1;
			break;
		}
		wrapper->contact_id = ast_malloc(strlen(aor_id) + strlen(contact->uri) + 2);
		if (!wrapper->contact_id) {
			res = -1;
			break;
		}
		sprintf(wrapper->contact_id, "%s/%s", aor_id, contact->uri);
		wrapper->aor_id = ast_strdup(aor_id);
		if (!wrapper->aor_id) {
			res = -1;
			break;
		}
		wrapper->contact = contact;
		ao2_bump(wrapper->contact);

		if ((res = on_contact(wrapper, arg, 0))) {
			break;
		}
	}
	ao2_iterator_destroy(&i);
	ao2_ref(contacts, -1);
	return res;
}

int ast_sip_contact_to_str(void *object, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = object;
	struct ast_str **buf = arg;

	ast_str_append(buf, 0, "%s,", wrapper->contact_id);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
int ast_sip_contact_origaddr_to_str(void *object, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = object;
	struct ast_str **buf = arg;

#ifdef BUILD_ON_CLOUD_PBX
	if (wrapper->contact && !ast_strlen_zero(wrapper->contact->original_address) && strncmp(wrapper->contact->original_address, "127.0.0.1", 9)) {
		char *addr_str = ast_strdupa(wrapper->contact_id);
		char *orig_addr_str = ast_strdupa(wrapper->contact->original_address);
		char  addr_buf[256] = {0};
		char *paddr_start = NULL;
		char *paddr_end   = NULL;
		char *paddr_pos   = NULL;
		/*BEGIN: Add by weidu, for Bug 307709 */
		/*normal: 5000/sip:5000@10.2.254.237:5060
		* sometimes: 5000/sip:10.2.254.253:5060
		*/
		int  offset = 0;
		
		paddr_start = strstr(addr_str, "@");
		if (paddr_start){
			offset = 1;
		}else {
			paddr_start = strstr(addr_str, "sip:");
			if (paddr_start){
				offset = 4;
			}else {
				paddr_start = strstr(addr_str, "sips:");
				if (paddr_start){
					offset = 5; 
				}
			}
		}

		if (!paddr_start){
			ast_log(LOG_ERROR, "ast_sip_contact_origaddr_to_str invaild contact uri: %s\n", wrapper->contact_id);
			return -1;
		}
		paddr_start = paddr_start + offset;
		/*END: Add by weidu, for Bug 307709 */
		
		paddr_end = strstr(paddr_start, ";");

		strncat(addr_buf, addr_str, paddr_start-addr_str);
		paddr_pos = addr_buf+strlen(addr_buf);
		snprintf(paddr_pos, sizeof(addr_buf)-1, "%s%s",orig_addr_str, paddr_end?paddr_end:"");

		ast_log(LOG_NOTICE, "contact_id[%s], addr_buf[%s]", wrapper->contact_id, addr_buf);
		ast_str_append(buf, 0, "%s,", addr_buf);
		return 0;
	}
#else
	char *p_addr = NULL;

	if ((p_addr = strstr(wrapper->contact_id, "@127.0.0.1"))) {
		if (wrapper->contact && !ast_strlen_zero(wrapper->contact->original_address) && strncmp(wrapper->contact->original_address, "127.0.0.1", 9)) {
			char *addr_str = ast_strdupa(wrapper->contact_id);
			char *orig_addr_str = ast_strdupa(wrapper->contact->original_address);
			char  addr_buf[256] = {0};
			char *paddr_start = NULL;
			char *paddr_end   = NULL;
			char *paddr_pos   = NULL;

			paddr_start = strstr(addr_str, "@127.0.0.1");
			paddr_start++;
			paddr_end = strstr(paddr_start, ";");

			strncat(addr_buf, addr_str, paddr_start-addr_str);
			paddr_pos = addr_buf+strlen(addr_buf);
			snprintf(paddr_pos, sizeof(addr_buf)-1, "%s%s",orig_addr_str, paddr_end?paddr_end:"");

			//ast_log(LOG_NOTICE, "contact_id[%s], addr_buf[%s]", wrapper->contact_id, addr_buf);
			ast_str_append(buf, 0, "%s,", addr_buf);
			return 0;
		}
	}
#endif

	ast_str_append(buf, 0, "%s,", wrapper->contact_id);

	return 0;
}
#endif

static int sip_aor_to_ami(const struct ast_sip_aor *aor, struct ast_str **buf)
{
	struct ast_variable *objset;
	struct ast_variable *i;

	objset = ast_sorcery_objectset_create2(ast_sip_get_sorcery(), aor,
		AST_HANDLER_ONLY_STRING);
	if (!objset) {
		return -1;
	}

	ast_str_append(buf, 0, "ObjectType: %s\r\n",
		       ast_sorcery_object_get_type(aor));
	ast_str_append(buf, 0, "ObjectName: %s\r\n",
		       ast_sorcery_object_get_id(aor));

	for (i = objset; i; i = i->next) {
		char *camel = ast_to_camel_case(i->name);

		if (strcmp(camel, "Contact") == 0) {
			ast_free(camel);
			camel = NULL;
		}
		ast_str_append(buf, 0, "%s: %s\r\n", S_OR(camel, "Contacts"), i->value);
		ast_free(camel);
	}

	ast_variables_destroy(objset);
	return 0;
}

static int contacts_to_str(const void *obj, const intptr_t *args, char **buf)
{
	const struct ast_sip_aor *aor = obj;
	struct ast_str *str;

	str = ast_str_create(MAX_OBJECT_FIELD);
	if (!str) {
		*buf = NULL;
		return -1;
	}

	ast_sip_for_each_contact(aor, ast_sip_contact_to_str, &str);
	ast_str_truncate(str, -1);

	*buf = ast_strdup(ast_str_buffer(str));
	ast_free(str);

	return *buf ? 0 : -1;
}

static int format_ami_aor_handler(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;
	struct ast_sip_ami *ami = arg;
	const struct ast_sip_endpoint *endpoint = ami->arg;
	struct ast_str *buf;
	struct ao2_container *contacts;
	int total_contacts;
	int num_permanent;

	buf = ast_sip_create_ami_event("AorDetail", ami);
	if (!buf) {
		return -1;
	}
	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts) {
		ast_free(buf);
		return -1;
	}

	sip_aor_to_ami(aor, &buf);
	total_contacts = ao2_container_count(contacts);
	num_permanent = aor->permanent_contacts ?
		ao2_container_count(aor->permanent_contacts) : 0;

	ast_str_append(&buf, 0, "TotalContacts: %d\r\n", total_contacts);
	ast_str_append(&buf, 0, "ContactsRegistered: %d\r\n",
		       total_contacts - num_permanent);
	ast_str_append(&buf, 0, "EndpointName: %s\r\n",
		       ast_sorcery_object_get_id(endpoint));

	astman_append(ami->s, "%s\r\n", ast_str_buffer(buf));
	ami->count++;

	ast_free(buf);
	ao2_ref(contacts, -1);
	return 0;
}

static int format_ami_endpoint_aor(const struct ast_sip_endpoint *endpoint,
				   struct ast_sip_ami *ami)
{
	ami->arg = (void *)endpoint;
	return ast_sip_for_each_aor(endpoint->aors,
				    format_ami_aor_handler, ami);
}

struct ast_sip_endpoint_formatter endpoint_aor_formatter = {
	.format_ami = format_ami_endpoint_aor
};

static struct ao2_container *cli_aor_get_container(const char *regex)
{
	struct ao2_container *container;
	struct ao2_container *s_container;

	container = ast_sorcery_retrieve_by_regex(ast_sip_get_sorcery(), "aor", regex);
	if (!container) {
		return NULL;
	}

	/* Create a sorted container of aors. */
	s_container = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
		ast_sorcery_object_id_sort, ast_sorcery_object_id_compare);
	if (s_container
		&& ao2_container_dup(s_container, container, 0)) {
		ao2_ref(s_container, -1);
		s_container = NULL;
	}
	ao2_ref(container, -1);

	return s_container;
}

static int cli_contact_populate_container(void *obj, void *arg, int flags)
{
	ao2_link(arg, obj);

	return 0;
}

static int cli_aor_gather_contacts(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;

	return ast_sip_for_each_contact(aor, cli_contact_populate_container, arg);
}

static const char *cli_contact_get_id(const void *obj)
{
	const struct ast_sip_contact_wrapper *wrapper = obj;
	return wrapper->contact_id;
}

static int cli_contact_sort(const void *obj, const void *arg, int flags)
{
	const struct ast_sip_contact_wrapper *left_wrapper = obj;
	const struct ast_sip_contact_wrapper *right_wrapper = arg;
	const char *right_key = arg;
	int cmp = 0;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_OBJECT:
		right_key = right_wrapper->contact_id;
		/* Fall through */
	case OBJ_SEARCH_KEY:
		cmp = strcmp(left_wrapper->contact_id, right_key);
		break;
	case OBJ_SEARCH_PARTIAL_KEY:
		cmp = strncmp(left_wrapper->contact_id, right_key, strlen(right_key));
		break;
	default:
		cmp = 0;
		break;
	}

	return cmp;
}

static int cli_contact_compare(void *obj, void *arg, int flags)
{
	const struct ast_sip_contact_wrapper *left_wrapper = obj;
	const struct ast_sip_contact_wrapper *right_wrapper = arg;
	const char *right_key = arg;
	int cmp = 0;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_OBJECT:
		right_key = right_wrapper->contact_id;
		/* Fall through */
	case OBJ_SEARCH_KEY:
		if (strcmp(left_wrapper->contact_id, right_key) == 0) {;
			cmp = CMP_MATCH | CMP_STOP;
		}
		break;
	case OBJ_SEARCH_PARTIAL_KEY:
		if (strncmp(left_wrapper->contact_id, right_key, strlen(right_key)) == 0) {
			cmp = CMP_MATCH;
		}
		break;
	default:
		cmp = 0;
		break;
	}

	return cmp;
}

static int cli_contact_iterate(void *container, ao2_callback_fn callback, void *args)
{
	return ast_sip_for_each_contact(container, callback, args);
}

static int cli_filter_contacts(void *obj, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = obj;
	regex_t *regexbuf = arg;

	if (!regexec(regexbuf, wrapper->contact_id, 0, NULL, 0)) {
		return 0;
	}

	return CMP_MATCH;
}

#ifndef GRANDSTREAM_NETWORKS
static int cli_gather_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	RAII_VAR(struct ast_sip_contact_wrapper *, wrapper, NULL, ao2_cleanup);

	if (strcmp(contact->reg_server, ast_config_AST_SYSTEM_NAME ?: "")) {
		return 0;
	}

	wrapper = ao2_alloc_options(sizeof(struct ast_sip_contact_wrapper),
		contact_wrapper_destroy, AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!wrapper) {
		return -1;
	}

	wrapper->contact_id = ast_malloc(strlen(contact->aor) + strlen(contact->uri) + 2);
	if (!wrapper->contact_id) {
		return -1;
	}
	sprintf(wrapper->contact_id, "%s/%s", contact->aor, contact->uri);

	wrapper->aor_id = ast_strdup(contact->aor);
	if (!wrapper->aor_id) {
		return -1;
	}

	wrapper->contact = ao2_bump(contact);

	ao2_link(arg, wrapper);

	return 0;
}
#endif

static struct ao2_container *cli_contact_get_container(const char *regex)
{
	RAII_VAR(struct ao2_container *, aors, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	RAII_VAR(struct ast_variable *, var_aor, NULL, ast_variables_destroy);
	struct ao2_container *contacts_container;
	regex_t regexbuf;

	if (!(var_aor = ast_variable_new("contact !=", "", ""))) {
		return NULL;
	}

	contacts_container = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
		cli_contact_sort, cli_contact_compare);
	if (!contacts_container) {
		return NULL;
	}

	contacts = ast_sorcery_retrieve_by_regex(ast_sip_get_sorcery(), "contact", regex);
	if (!contacts) {
		ao2_ref(contacts_container, -1);
		return NULL;
	}

#ifndef GRANDSTREAM_NETWORKS
	/* Modified by yfyang, for void redundancy contact record */
	ao2_callback(contacts, OBJ_NODATA, cli_gather_contact, contacts_container);
#endif

	aors = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(),
		"aor", AST_RETRIEVE_FLAG_MULTIPLE, var_aor);
	if (!aors) {
		ao2_ref(contacts_container, -1);
		return NULL;
	}

	ao2_callback(aors, OBJ_NODATA, cli_aor_gather_contacts, contacts_container);

	if (!ast_strlen_zero(regex)) {
		if (regcomp(&regexbuf, regex, REG_EXTENDED | REG_NOSUB)) {
			ao2_ref(contacts_container, -1);
			return NULL;
		}
		ao2_callback(contacts_container, OBJ_UNLINK | OBJ_MULTIPLE | OBJ_NODATA, cli_filter_contacts, &regexbuf);
		regfree(&regexbuf);
	}

	return contacts_container;
}

static void *cli_contact_retrieve_by_id(const char *id)
{
	struct ao2_container *container;
	void *obj;

	container = cli_contact_get_container("");
	if (!container) {
		return NULL;
	}

	obj = ao2_find(container, id, OBJ_SEARCH_KEY);
	ao2_ref(container, -1);
	return obj;
}

static int cli_contact_print_header(void *obj, void *arg, int flags)
{
	struct ast_sip_cli_context *context = arg;
	int indent = CLI_INDENT_TO_SPACES(context->indent_level);
	int filler = CLI_LAST_TABSTOP - indent - 23;

	ast_assert(context->output_buffer != NULL);

	ast_str_append(&context->output_buffer, 0,
		"%*s:  <Aor/ContactUri%*.*s> <Hash....> <Status> <RTT(ms)..>\n",
		indent, "Contact", filler, filler, CLI_HEADER_FILLER);

	return 0;
}

static int cli_contact_print_body(void *obj, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = obj;
	struct ast_sip_contact *contact = wrapper->contact;
	struct ast_sip_cli_context *context = arg;
	int indent;
	int flexwidth;
	const char *contact_id = ast_sorcery_object_get_id(contact);
	const char *hash_start = contact_id + strlen(contact->aor) + 2;
	struct ast_sip_contact_status *status;

	ast_assert(contact->uri != NULL);
	ast_assert(context->output_buffer != NULL);

	status = ast_sip_get_contact_status(contact);

	indent = CLI_INDENT_TO_SPACES(context->indent_level);
	flexwidth = CLI_LAST_TABSTOP - indent - 9 - strlen(contact->aor) + 1;

	ast_str_append(&context->output_buffer, 0, "%*s:  %s/%-*.*s %-10.10s %-7.7s %11.3f\n",
		indent,
		"Contact",
		contact->aor,
		flexwidth, flexwidth,
		contact->uri,
		hash_start,
		ast_sip_get_contact_short_status_label(status ? status->status : UNKNOWN),
		(status && (status->status == AVAILABLE)) ? ((long long) status->rtt) / 1000.0 : NAN);

	ao2_cleanup(status);
	return 0;
}

static int cli_aor_iterate(void *container, ao2_callback_fn callback, void *args)
{
	const char *aor_list = container;

	return ast_sip_for_each_aor(aor_list, callback, args);
}

static void *cli_aor_retrieve_by_id(const char *id)
{
	return ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "aor", id);
}

static const char *cli_aor_get_id(const void *obj)
{
	return ast_sorcery_object_get_id(obj);
}

static int cli_aor_print_header(void *obj, void *arg, int flags)
{
	struct ast_sip_cli_context *context = arg;
	int indent = CLI_INDENT_TO_SPACES(context->indent_level);
	int filler = CLI_LAST_TABSTOP - indent - 7;

	ast_assert(context->output_buffer != NULL);

	ast_str_append(&context->output_buffer, 0,
		"%*s:  <Aor%*.*s>  <MaxContact>\n",
		indent, "Aor", filler, filler, CLI_HEADER_FILLER);

	if (context->recurse) {
		struct ast_sip_cli_formatter_entry *formatter_entry;

		context->indent_level++;
		formatter_entry = ast_sip_lookup_cli_formatter("contact");
		if (formatter_entry) {
			formatter_entry->print_header(NULL, context, 0);
			ao2_ref(formatter_entry, -1);
		}
		context->indent_level--;
	}

	return 0;
}

static int cli_aor_print_body(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;
	struct ast_sip_cli_context *context = arg;
	int indent;
	int flexwidth;

	ast_assert(context->output_buffer != NULL);

//	context->current_aor = aor;

	indent = CLI_INDENT_TO_SPACES(context->indent_level);
	flexwidth = CLI_LAST_TABSTOP - indent - 12;

	ast_str_append(&context->output_buffer, 0, "%*s:  %-*.*s %12u\n",
		indent,
		"Aor",
		flexwidth, flexwidth,
		ast_sorcery_object_get_id(aor), aor->max_contacts);

	if (context->recurse) {
		struct ast_sip_cli_formatter_entry *formatter_entry;

		context->indent_level++;

		formatter_entry = ast_sip_lookup_cli_formatter("contact");
		if (formatter_entry) {
			formatter_entry->iterate(aor, formatter_entry->print_body, context);
			ao2_ref(formatter_entry, -1);
		}

		context->indent_level--;

		if (context->indent_level == 0) {
			ast_str_append(&context->output_buffer, 0, "\n");
		}
	}

	if (context->show_details || (context->show_details_only_level_0 && context->indent_level == 0)) {
		ast_str_append(&context->output_buffer, 0, "\n");
		ast_sip_cli_print_sorcery_objectset(aor, context, 0);
	}

	return 0;
}

static struct ao2_container *cli_get_aors(void)
{
	struct ao2_container *aors;

	aors = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "aor",
			AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL);

	return aors;
}

static int format_ami_aorlist_handler(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;
	struct ast_sip_ami *ami = arg;
	struct ast_str *buf;

	buf = ast_sip_create_ami_event("AorList", ami);
	if (!buf) {
		return -1;
	}

	sip_aor_to_ami(aor, &buf);

	astman_append(ami->s, "%s\r\n", ast_str_buffer(buf));
	ami->count++;

	ast_free(buf);

	return 0;
}

static int ami_show_aors(struct mansession *s, const struct message *m)
{
	struct ast_sip_ami ami = { .s = s, .m = m, .action_id = astman_get_header(m, "ActionID"), };
	struct ao2_container *aors;

	aors = cli_get_aors();
	if (!aors) {
		astman_send_error(s, m, "Could not get AORs\n");
		return 0;
	}

	if (!ao2_container_count(aors)) {
		astman_send_error(s, m, "No AORs found\n");
		ao2_ref(aors, -1);
		return 0;
	}

	astman_send_listack(s, m, "A listing of AORs follows, presented as AorList events",
			"start");

	ao2_callback(aors, OBJ_NODATA, format_ami_aorlist_handler, &ami);

	astman_send_list_complete_start(s, m, "AorListComplete", ami.count);
	astman_send_list_complete_end(s);

	ao2_ref(aors, -1);

	return 0;
}

static struct ast_cli_entry cli_commands[] = {
	AST_CLI_DEFINE(ast_sip_cli_traverse_objects, "List PJSIP Aors",
		.command = "pjsip list aors",
		.usage = "Usage: pjsip list aors [ like <pattern> ]\n"
				"       List the configured PJSIP Aors\n"
				"       Optional regular expression pattern is used to filter the list.\n"),
	AST_CLI_DEFINE(ast_sip_cli_traverse_objects, "Show PJSIP Aors",
		.command = "pjsip show aors",
		.usage = "Usage: pjsip show aors [ like <pattern> ]\n"
				"       Show the configured PJSIP Aors\n"
				"       Optional regular expression pattern is used to filter the list.\n"),
	AST_CLI_DEFINE(ast_sip_cli_traverse_objects, "Show PJSIP Aor",
		.command = "pjsip show aor",
		.usage = "Usage: pjsip show aor <id>\n"
				 "       Show the configured PJSIP Aor\n"),

	AST_CLI_DEFINE(ast_sip_cli_traverse_objects, "List PJSIP Contacts",
		.command = "pjsip list contacts",
		.usage = "Usage: pjsip list contacts [ like <pattern> ]\n"
				"       List the configured PJSIP contacts\n"
				"       Optional regular expression pattern is used to filter the list.\n"),
	AST_CLI_DEFINE(ast_sip_cli_traverse_objects, "Show PJSIP Contacts",
		.command = "pjsip show contacts",
		.usage = "Usage: pjsip show contacts [ like <pattern> ]\n"
				"       Show the configured PJSIP contacts\n"
				"       Optional regular expression pattern is used to filter the list.\n"),
	AST_CLI_DEFINE(ast_sip_cli_traverse_objects, "Show PJSIP Contact",
		.command = "pjsip show contact",
		.usage = "Usage: pjsip show contact\n"
				 "       Show the configured PJSIP contact\n"),
#ifdef GRANDSTREAM_NETWORKS
	AST_CLI_DEFINE(ast_sip_show_reginfos, "Show REGISTER Information"),
	AST_CLI_DEFINE(ast_sip_reset_reginfos,"Reset REGISTER Information"),
	AST_CLI_DEFINE(ast_sip_show_overview, "Show SIP Device Overview Information"),
	AST_CLI_DEFINE(ast_sip_show_max_reginfos,  "Show Maximum REGISTER USERS Information"),
	AST_CLI_DEFINE(ast_sip_reset_max_reginfos,  "Reset Maximum REGISTER USERS Information"),
	AST_CLI_DEFINE(ast_sip_show_account_tks_reginfo,  "Show account trunk address"),
	AST_CLI_DEFINE(ast_sip_show_account_tks,  "Show account trunks"),
#endif
};

struct ast_sip_cli_formatter_entry *contact_formatter;
struct ast_sip_cli_formatter_entry *aor_formatter;

/*! \brief Always create a contact_status for each contact */
static int contact_apply_handler(const struct ast_sorcery *sorcery, void *object)
{
	struct ast_sip_contact_status *status;
	struct ast_sip_contact *contact = object;

	if (ast_strlen_zero(contact->uri)) {
		ast_log(LOG_ERROR, "A URI on dynamic contact '%s' is empty\n",
			ast_sorcery_object_get_id(contact));
		return -1;
	}
	status = ast_res_pjsip_find_or_create_contact_status(contact);
	ao2_cleanup(status);

	return status ? 0 : -1;
}

/*! \brief Initialize sorcery with location support */
int ast_sip_initialize_sorcery_location(void)
{
	struct ast_sorcery *sorcery = ast_sip_get_sorcery();
	int i;

	ast_pjproject_get_buildopt("PJ_MAX_HOSTNAME", "%d", &pj_max_hostname);
	/* As of pjproject 2.4.5, PJSIP_MAX_URL_SIZE isn't exposed yet but we try anyway. */
	ast_pjproject_get_buildopt("PJSIP_MAX_URL_SIZE", "%d", &pjsip_max_url_size);

	ast_sorcery_apply_default(sorcery, "contact", "astdb", "registrar");
	ast_sorcery_object_set_congestion_levels(sorcery, "contact", -1,
		3 * AST_TASKPROCESSOR_HIGH_WATER_LEVEL);

#ifndef GRANDSTREAM_NETWORKS    /* modified by yfyang, for performance optimization */
	ast_sorcery_apply_default(sorcery, "aor", "config", "pjsip.conf,criteria=type=aor");
#else
	ast_sorcery_apply_default(sorcery, "aor", "config", "pjsip_aors.conf,criteria=type=aor");
#endif

	if (ast_sorcery_object_register(sorcery, "contact", contact_alloc, NULL, contact_apply_handler) ||
		ast_sorcery_object_register(sorcery, "aor", aor_alloc, NULL, NULL)) {
		return -1;
	}

	ast_sorcery_observer_add(sorcery, "aor", &aor_observer);

	ast_sorcery_object_field_register(sorcery, "contact", "type", "", OPT_NOOP_T, 0, 0);
	ast_sorcery_object_field_register(sorcery, "contact", "uri", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, uri));
	ast_sorcery_object_field_register(sorcery, "contact", "path", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, path));
	ast_sorcery_object_field_register_custom(sorcery, "contact", "expiration_time", "", expiration_str2struct, expiration_struct2str, NULL, 0, 0);
	ast_sorcery_object_field_register(sorcery, "contact", "qualify_frequency", 0, OPT_UINT_T,
		PARSE_IN_RANGE, FLDSET(struct ast_sip_contact, qualify_frequency), 0, 86400);
	ast_sorcery_object_field_register(sorcery, "contact", "qualify_timeout", "3.0", OPT_DOUBLE_T, 0, FLDSET(struct ast_sip_contact, qualify_timeout));
	ast_sorcery_object_field_register(sorcery, "contact", "authenticate_qualify", "no", OPT_YESNO_T, 1, FLDSET(struct ast_sip_contact, authenticate_qualify));
	ast_sorcery_object_field_register(sorcery, "contact", "outbound_proxy", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, outbound_proxy));
	ast_sorcery_object_field_register(sorcery, "contact", "user_agent", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, user_agent));
	ast_sorcery_object_field_register(sorcery, "contact", "endpoint", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, endpoint_name));
	ast_sorcery_object_field_register(sorcery, "contact", "reg_server", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, reg_server));
	ast_sorcery_object_field_register(sorcery, "contact", "via_addr", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, via_addr));
	ast_sorcery_object_field_register(sorcery, "contact", "via_port", "0", OPT_UINT_T, 0, FLDSET(struct ast_sip_contact, via_port));
	ast_sorcery_object_field_register(sorcery, "contact", "call_id", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, call_id));
	ast_sorcery_object_field_register(sorcery, "contact", "prune_on_boot", "no", OPT_YESNO_T, 1, FLDSET(struct ast_sip_contact, prune_on_boot));
#ifdef GRANDSTREAM_NETWORKS
	ast_sorcery_object_field_register(sorcery, "contact", "uuid", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, uuid));
	ast_sorcery_object_field_register(sorcery, "contact", "rmv_obp_from_route", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, rmv_obp_from_route));
	ast_sorcery_object_field_register(sorcery, "contact", "support_missed_call", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_missed_call));
	ast_sorcery_object_field_register(sorcery, "contact", "support_gs_message", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_gs_message));
	ast_sorcery_object_field_register(sorcery, "contact", "ssh_proxy_flag", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, ssh_proxy_flag));
	ast_sorcery_object_field_register(sorcery, "contact", "support_ice", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_ice));
	ast_sorcery_object_field_register(sorcery, "contact", "support_trickle_ice", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_trickle_ice));
	ast_sorcery_object_field_register(sorcery, "contact", "gswave_mobile", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, gswave_mobile));
	ast_sorcery_object_field_register(sorcery, "contact", "original_address", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, original_address));
	ast_sorcery_object_field_register(sorcery, "contact", "support_ldap_settings", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_ldap_settings));
	ast_sorcery_object_field_register(sorcery, "contact", "support_gs_config", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_gs_config));
	ast_sorcery_object_field_register(sorcery, "contact", "support_gs_state", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_gs_state));
	ast_sorcery_object_field_register(sorcery, "contact", "support_gs_fec", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_gs_fec));
	ast_sorcery_object_field_register(sorcery, "contact", "require_feature_code_settings", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, require_feature_code_settings));
	ast_sorcery_object_field_register(sorcery, "contact", "require_turn_info_settings", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, require_turn_info_settings));
	ast_sorcery_object_field_register(sorcery, "contact", "require_basic_settings", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, require_basic_settings));
	ast_sorcery_object_field_register(sorcery, "contact", "to_host", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, to_host));
	ast_sorcery_object_field_register(sorcery, "contact", "instance", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, instance));
	ast_sorcery_object_field_register(sorcery, "contact", "require_plugin", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, require_plugin));
	ast_sorcery_object_field_register(sorcery, "contact", "require_phonebook", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, require_phonebook));
	ast_sorcery_object_field_register(sorcery, "contact", "support_gs_conf_invite", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_gs_conf_invite));
	ast_sorcery_object_field_register(sorcery, "contact", "support_gs_control", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_contact, support_gs_control));
	ast_sorcery_object_field_register(sorcery, "contact", "sbc_endpoint_addr", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_contact, sbc_endpoint_addr));

	ast_sorcery_object_field_register(sorcery, "contact", "reg_id", "0", OPT_UINT_T, 0, FLDSET(struct ast_sip_contact, reg_id));
#endif

	ast_sorcery_object_field_register(sorcery, "aor", "type", "", OPT_NOOP_T, 0, 0);
	ast_sorcery_object_field_register(sorcery, "aor", "minimum_expiration", "60", OPT_UINT_T, 0, FLDSET(struct ast_sip_aor, minimum_expiration));
	ast_sorcery_object_field_register(sorcery, "aor", "maximum_expiration", "7200", OPT_UINT_T, 0, FLDSET(struct ast_sip_aor, maximum_expiration));
	ast_sorcery_object_field_register(sorcery, "aor", "default_expiration", "3600", OPT_UINT_T, 0, FLDSET(struct ast_sip_aor, default_expiration));
	ast_sorcery_object_field_register(sorcery, "aor", "qualify_frequency", 0, OPT_UINT_T, PARSE_IN_RANGE, FLDSET(struct ast_sip_aor, qualify_frequency), 0, 86400);
	ast_sorcery_object_field_register(sorcery, "aor", "qualify_timeout", "3.0", OPT_DOUBLE_T, 0, FLDSET(struct ast_sip_aor, qualify_timeout));
	ast_sorcery_object_field_register(sorcery, "aor", "authenticate_qualify", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_aor, authenticate_qualify));
	ast_sorcery_object_field_register(sorcery, "aor", "max_contacts", "0", OPT_UINT_T, 0, FLDSET(struct ast_sip_aor, max_contacts));

#ifndef GRANDSTREAM_NETWORKS
	ast_sorcery_object_field_register(sorcery, "aor", "remove_existing", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_aor, remove_existing));
#else
	ast_sorcery_object_field_register(sorcery, "aor", "remove_existing", "yes", OPT_BOOL_T, 1, FLDSET(struct ast_sip_aor, remove_existing));
#endif
	ast_sorcery_object_field_register_custom(sorcery, "aor", "contact", "", permanent_uri_handler, contacts_to_str, contacts_to_var_list, 0, 0);
	ast_sorcery_object_field_register(sorcery, "aor", "mailboxes", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_aor, mailboxes));
	ast_sorcery_object_field_register_custom(sorcery, "aor", "voicemail_extension", "", voicemail_extension_handler, voicemail_extension_to_str, NULL, 0, 0);
	ast_sorcery_object_field_register(sorcery, "aor", "outbound_proxy", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ast_sip_aor, outbound_proxy));
	ast_sorcery_object_field_register(sorcery, "aor", "support_path", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_aor, support_path));
#ifdef GRANDSTREAM_NETWORKS
	ast_sorcery_object_field_register(sorcery, "aor", "rmv_obp_from_route", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_aor, rmv_obp_from_route));
	ast_sorcery_object_field_register(sorcery, "aor", "account_trunk", "no", OPT_BOOL_T, 1, FLDSET(struct ast_sip_aor, account_trunk));
#endif

	ast_sip_register_endpoint_formatter(&endpoint_aor_formatter);

	contact_formatter = ao2_alloc(sizeof(struct ast_sip_cli_formatter_entry), NULL);
	if (!contact_formatter) {
		ast_log(LOG_ERROR, "Unable to allocate memory for contact_formatter\n");
		return -1;
	}
	contact_formatter->name = "contact";
	contact_formatter->print_header = cli_contact_print_header;
	contact_formatter->print_body = cli_contact_print_body;
	contact_formatter->get_container = cli_contact_get_container;
	contact_formatter->iterate = cli_contact_iterate;
	contact_formatter->get_id = cli_contact_get_id;
	contact_formatter->retrieve_by_id = cli_contact_retrieve_by_id;

	aor_formatter = ao2_alloc(sizeof(struct ast_sip_cli_formatter_entry), NULL);
	if (!aor_formatter) {
		ast_log(LOG_ERROR, "Unable to allocate memory for aor_formatter\n");
		return -1;
	}
	aor_formatter->name = "aor";
	aor_formatter->print_header = cli_aor_print_header;
	aor_formatter->print_body = cli_aor_print_body;
	aor_formatter->get_container = cli_aor_get_container;
	aor_formatter->iterate = cli_aor_iterate;
	aor_formatter->get_id = cli_aor_get_id;
	aor_formatter->retrieve_by_id = cli_aor_retrieve_by_id;

	ast_sip_register_cli_formatter(contact_formatter);
	ast_sip_register_cli_formatter(aor_formatter);
	ast_cli_register_multiple(cli_commands, ARRAY_LEN(cli_commands));

	if (ast_manager_register_xml("PJSIPShowAors", EVENT_FLAG_SYSTEM, ami_show_aors)) {
		return -1;
	}

	/*
	 * Reset StatsD gauges in case we didn't shut down cleanly.
	 * Note that this must done here, as contacts will create the contact_status
	 * object before PJSIP options handling is initialized.
	 */
	for (i = 0; i <= REMOVED; i++) {
		ast_statsd_log_full_va("PJSIP.contacts.states.%s", AST_STATSD_GAUGE, 0, 1.0, ast_sip_get_contact_status_label(i));
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_rwlock_init(&proxy_user_count_lock);
	ast_rwlock_init(&local_user_count_lock);
	ast_rwlock_init(&reg_user_update_lock);
	ast_rwlock_init(&rc_user_max_lock);
	ast_rwlock_init(&local_user_max_lock);
	ast_rwlock_init(&reg_user_max_lock);

	rc_reginfos = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0, NULL, ast_sip_reginfo_cmp);
	local_reginfos = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0, NULL, ast_sip_reginfo_cmp);
	user_reginfos = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0, NULL, ast_sip_reginfo_cmp);
    account_tk_infos = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0, NULL, ast_sip_account_tk_cmp);
    account_tk_confs = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0, NULL, ast_sip_acctk_conf_cmp);
#endif

	return 0;
}

int ast_sip_destroy_sorcery_location(void)
{
	ast_sorcery_observer_remove(ast_sip_get_sorcery(), "aor", &aor_observer);
	ast_cli_unregister_multiple(cli_commands, ARRAY_LEN(cli_commands));
	ast_sip_unregister_cli_formatter(contact_formatter);
	ast_sip_unregister_cli_formatter(aor_formatter);
	ast_manager_unregister("PJSIPShowAors");

	ast_sip_unregister_endpoint_formatter(&endpoint_aor_formatter);

#ifdef GRANDSTREAM_NETWORKS
	ast_rwlock_destroy(&proxy_user_count_lock);
	ast_rwlock_destroy(&local_user_count_lock);
	ast_rwlock_destroy(&reg_user_update_lock);
	ast_rwlock_destroy(&rc_user_max_lock);
	ast_rwlock_destroy(&local_user_max_lock);
	ast_rwlock_destroy(&reg_user_max_lock);

	ast_sip_reginfo_destroy();
#endif

	return 0;
}

