/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*!
 * \file
 *
 * \brief Generate security events in the PJSIP channel
 *
 * \author Joshua Colp <jcolp@digium.com>
 */

#include "asterisk.h"

#include <pjsip.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/security_events.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/http_websocket.h"
#include "asterisk/astdb.h"

AST_MODULE_LOG("security_events");
#endif

static enum ast_transport security_event_get_transport(pjsip_rx_data *rdata)
{
	if (rdata->tp_info.transport->key.type == PJSIP_TRANSPORT_UDP ||
		rdata->tp_info.transport->key.type == PJSIP_TRANSPORT_UDP6) {
		return AST_TRANSPORT_UDP;
	} else if (rdata->tp_info.transport->key.type == PJSIP_TRANSPORT_TCP ||
		rdata->tp_info.transport->key.type == PJSIP_TRANSPORT_TCP6) {
		return AST_TRANSPORT_TCP;
	} else if (rdata->tp_info.transport->key.type == PJSIP_TRANSPORT_TLS ||
		rdata->tp_info.transport->key.type == PJSIP_TRANSPORT_TLS6) {
		return AST_TRANSPORT_TLS;
	} else if (!strcasecmp(rdata->tp_info.transport->type_name, "WS")) {
		return AST_TRANSPORT_WS;
	} else if (!strcasecmp(rdata->tp_info.transport->type_name, "WSS")) {
		return AST_TRANSPORT_WSS;
	} else {
		return 0;
	}
}

#ifdef GRANDSTREAM_NETWORKS
// GS-GDMS-Endpoint-Proto:udp/tcp/tls/ws/wss
// GS-GDMS-Endpoint-Address:x.x.x.x
static int get_gdms_protocol_and_address(struct pjsip_rx_data* rdata,
								char* protocol_buf, int protocol_size,
								char* address_buf, int address_size)
{
	if (!rdata || (!protocol_buf && !address_buf)) {
		ast_log(LOG_ERROR, "get_gdms_protocol_and_address: Invalid args.");
		return -1;
	}

	pj_str_t GDMS_PROTO_HDR_STRING = { "GS-GDMS-Endpoint-Proto", 22 };
	pj_str_t GDMS_ADDRESS_HDR_STRING = { "GS-GDMS-Endpoint-Address", 24 };
	pjsip_generic_string_hdr* protocol_hdr = NULL;
	pjsip_generic_string_hdr* address_hdr = NULL;

	if (rdata->msg_info.msg->type != PJSIP_REQUEST_MSG) {
		return -1;
	}

	if (0 == ast_sip_check_ssh_proxy_flag(rdata)) {
		return -1;
	}

	if (protocol_buf && protocol_size > 0) {
		if (!(protocol_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &GDMS_PROTO_HDR_STRING, NULL))) {
			ast_log(LOG_DEBUG, "Can't get GS-GDMS-Endpoint-Proto header.");
			return -1;
		}

		if (protocol_hdr->hvalue.slen == 0) {
			ast_log(LOG_DEBUG, "Can't get GS-GDMS-Endpoint-Proto header.");
			return -1;
		}

		memset(protocol_buf, 0, protocol_size);
		ast_copy_pj_str(protocol_buf, &protocol_hdr->hvalue, protocol_size);

		//ast_log(LOG_NOTICE, "The gdms protocol:%s", protocol_buf);
	}

	if (address_buf && address_size > 0) {
		if (!(address_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &GDMS_ADDRESS_HDR_STRING, NULL))) {
			ast_log(LOG_DEBUG, "Can't get GS-GDMS-Endpoint-Address header.");
			return -1;
		}

		if (address_hdr->hvalue.slen == 0) {
			ast_log(LOG_DEBUG, "Can't get GS-GDMS-Endpoint-Address header.");
			return -1;
		}

		memset(address_buf, 0, address_size);
		ast_copy_pj_str(address_buf, &address_hdr->hvalue, address_size);

		//ast_log(LOG_NOTICE, "The gdms address:%s", address_buf);
	}

	return 0;
}

static enum ast_transport update_transport_type(pjsip_rx_data *rdata)
{
	char protocol[32] = {0};
	enum ast_transport transport = AST_TRANSPORT_UDP;

	if (!rdata) {
		ast_log(LOG_ERROR, "[update_transport_type] The rdata is nil.");
		return transport;
	}

	if (get_gdms_protocol_and_address(rdata, protocol, sizeof(protocol), NULL, 0)) {
		transport = security_event_get_transport(rdata);
	} else {
		if (!strncasecmp(protocol, "udp", 3)) {
			transport = AST_TRANSPORT_UDP;
		} else if (!strncasecmp(protocol, "tcp", 3)) {
			transport = AST_TRANSPORT_TCP;
		} else if (!strncasecmp(protocol, "tls", 3)) {
			transport = AST_TRANSPORT_TLS;
		} else if (!strncasecmp(protocol, "wss", 3)) {
			transport = AST_TRANSPORT_WSS;
		} else if (!strncasecmp(protocol, "ws", 2)) {
			transport = AST_TRANSPORT_WS;
		} else {
			transport = security_event_get_transport(rdata);
			//ast_log(LOG_DEBUG, "[update_transport_type] Used rdata transport type.");
		}
	}

	//ast_log(LOG_NOTICE, "[update_transport_type] current used transport type:%d.", transport);

	return transport;
}

/* Added by yfyang, for websocket remote address */
static int get_websocket_address(struct pjsip_rx_data* rdata, char* buf, int len)
{
	if (!rdata || !buf || !len) {
		return -1;
	}

	if (strncmp(rdata->tp_info.transport->type_name, "WS", 2)) {
		return -1;
	}

	char  port[8] = {0};
	snprintf(port, sizeof(port), "%d", rdata->pkt_info.src_port);
	ast_db_get(WEBSOCKET_REMOTE, port, buf, len-1);

	if (strlen(buf)) {
		return 0;
	}

	return -1;
}

#endif
static void security_event_populate(pjsip_rx_data *rdata, char *call_id, size_t call_id_size, struct ast_sockaddr *local, struct ast_sockaddr *remote)
{
	char host[NI_MAXHOST];

	ast_copy_pj_str(call_id, &rdata->msg_info.cid->id, call_id_size);

	ast_copy_pj_str(host, &rdata->tp_info.transport->local_name.host, sizeof(host));
	ast_sockaddr_parse(local, host, PARSE_PORT_FORBID);
	ast_sockaddr_set_port(local, rdata->tp_info.transport->local_name.port);

#ifdef GRANDSTREAM_NETWORKS
	char address[256] = {0};

	if (get_gdms_protocol_and_address(rdata, NULL, 0, address, sizeof(address))) {
		/* Added by yfyang, for websocket remote address */
		if (!get_websocket_address(rdata, address, sizeof(address))) {
			ast_sockaddr_parse(remote, address, PARSE_PORT_FORBID);
		} else {
			ast_sockaddr_parse(remote, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
		}

		ast_sockaddr_set_port(remote, rdata->pkt_info.src_port);
	} else {
		ast_sockaddr_parse(remote, address, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(remote, rdata->pkt_info.src_port);
	}
#else
	ast_sockaddr_parse(remote, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
	ast_sockaddr_set_port(remote, rdata->pkt_info.src_port);
#endif
}

static const char *get_account_id(struct ast_sip_endpoint *endpoint)
{
	RAII_VAR(struct ast_sip_endpoint *, artificial, ast_sip_get_artificial_endpoint(), ao2_cleanup);

	return endpoint == artificial ? "<unknown>" : ast_sorcery_object_get_id(endpoint);
}

#ifdef GRANDSTREAM_NETWORKS
/*added by wxu@grandstream for failed_challenge_response account_id is null*/
static void get_account_id_from_data( char *account_id, pjsip_rx_data *rdata, int buf_size )
{
	pjsip_sip_uri *uri_to = NULL;

	if( account_id == NULL || rdata == NULL){
		return;
	}

	uri_to = pjsip_uri_get_uri(rdata->msg_info.to->uri);
	if(uri_to != NULL){
		ast_copy_pj_str(account_id, &uri_to->user, buf_size);
	}
}
/*end*/
#endif

void ast_sip_report_invalid_endpoint(const char *name, pjsip_rx_data *rdata)
{
#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif
	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;
#ifdef GRANDSTREAM_NETWORKS
	char method[64] = {0};
	struct pjsip_request_line *req;

	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) {
		req = &rdata->msg_info.msg->line.req;
		ast_copy_pj_str(method, &req->method.name, sizeof(method));
	}
#endif

	struct ast_security_event_inval_acct_id inval_acct_id = {
		.common.event_type = AST_SECURITY_EVENT_INVAL_ACCT_ID,
		.common.version    = AST_SECURITY_EVENT_INVAL_ACCT_ID_VERSION,
		.common.service    = "PJSIP",
		.common.account_id = name,
		.common.local_addr = {
			.addr      = &local,
			.transport = transport,
		},
		.common.remote_addr = {
			.addr       = &remote,
			.transport = transport,
		},
		.common.session_id = call_id,
#ifdef GRANDSTREAM_NETWORKS
		.request_type	   = method,
#endif

	};

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&inval_acct_id));
}

void ast_sip_report_failed_acl(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata, const char *name)
{
#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif

	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;
#ifdef GRANDSTREAM_NETWORKS
	/*added by wxu@grandstream.cn set request_type for event warning*/
	struct pjsip_request_line req = rdata->msg_info.msg->line.req;
	char method_name[64] = {0};
	char account_id[64] = {0};
	ast_copy_pj_str(method_name, &req.method.name, sizeof(method_name));
	/*end*/
#endif

	struct ast_security_event_failed_acl failed_acl_event = {
			.common.event_type  = AST_SECURITY_EVENT_FAILED_ACL,
			.common.version     = AST_SECURITY_EVENT_FAILED_ACL_VERSION,
			.common.service     = "PJSIP",
#ifdef GRANDSTREAM_NETWORKS
			.common.account_id	= account_id,/* add by wxu@grandstream.cn for bug 42834*/
#else
			.common.account_id  = get_account_id(endpoint),
#endif
			.common.local_addr  = {
					.addr       = &local,
					.transport  = transport,
			},
			.common.remote_addr = {
					.addr       = &remote,
					.transport  = transport,
			},
			.common.session_id  = call_id,
			.acl_name           = name,
#ifdef GRANDSTREAM_NETWORKS
			/*added by wxu@grandstream.cn set request_type for event warning*/
			.request_type		= method_name,
			/*end*/
#endif

	};
#ifdef GRANDSTREAM_NETWORKS
	get_account_id_from_data(account_id,rdata,sizeof(account_id));/* add by wxu@grandstream.cn for bug 42834*/
#endif

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&failed_acl_event));
}

void ast_sip_report_auth_failed_challenge_response(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata)
{
	pjsip_authorization_hdr *auth = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_AUTHORIZATION, NULL);
#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif

	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	char nonce[64] = "", response[256] = "";
	struct ast_sockaddr local, remote;
#ifdef GRANDSTREAM_NETWORKS
	/*added by wxu@grandstream.cn set request_type for event warning*/
	struct pjsip_request_line req = rdata->msg_info.msg->line.req;
	char method_name[64] = {0};
	char account_id[64] = {0};
	ast_copy_pj_str(method_name, &req.method.name, sizeof(method_name));
	/*end*/
#endif

	struct ast_security_event_chal_resp_failed chal_resp_failed = {
				.common.event_type = AST_SECURITY_EVENT_CHAL_RESP_FAILED,
				.common.version    = AST_SECURITY_EVENT_CHAL_RESP_FAILED_VERSION,
				.common.service    = "PJSIP",
#ifdef GRANDSTREAM_NETWORKS
				.common.account_id	= account_id,
#else
				.common.account_id	= get_account_id(endpoint),
#endif

				.common.local_addr = {
						.addr      = &local,
						.transport = transport,
				},
				.common.remote_addr = {
						.addr      = &remote,
						.transport = transport,
				},
				.common.session_id = call_id,

				.challenge         = nonce,
				.response          = response,
				.expected_response = "",
#ifdef GRANDSTREAM_NETWORKS
				/*added by wxu@grandstream.cn set request_type for event warning*/
				.request_type		= method_name,
				/*end*/
#endif

	};

	if (auth && !pj_strcmp2(&auth->scheme, "Digest")) {
		ast_copy_pj_str(nonce, &auth->credential.digest.nonce, sizeof(nonce));
		ast_copy_pj_str(response, &auth->credential.digest.response, sizeof(response));
	}
#ifdef GRANDSTREAM_NETWORKS
	get_account_id_from_data(account_id,rdata,sizeof(account_id));
#endif
	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&chal_resp_failed));
}

void ast_sip_report_auth_success(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata)
{
	pjsip_authorization_hdr *auth = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_AUTHORIZATION, NULL);
#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif

	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;
#ifdef GRANDSTREAM_NETWORKS
	/*added by wxu@grandstream.cn set request_type for event warning*/
	struct pjsip_request_line req = rdata->msg_info.msg->line.req;
	char method_name[64] = {0};
	ast_copy_pj_str(method_name, &req.method.name, sizeof(method_name));
	/*end*/
#endif

	struct ast_security_event_successful_auth successful_auth = {
			.common.event_type  = AST_SECURITY_EVENT_SUCCESSFUL_AUTH,
			.common.version     = AST_SECURITY_EVENT_SUCCESSFUL_AUTH_VERSION,
			.common.service     = "PJSIP",
			.common.account_id  = get_account_id(endpoint),
			.common.local_addr  = {
					.addr       = &local,
					.transport  = transport,
			},
			.common.remote_addr = {
					.addr       = &remote,
					.transport  = transport,
			},
			.common.session_id  = call_id,
			.using_password     = auth ? 1 : 0,
#ifdef GRANDSTREAM_NETWORKS
			/*added by wxu@grandstream.cn set request_type for event warning*/
			.request_type		= method_name,
			/*end*/
#endif

	};

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&successful_auth));
}

void ast_sip_report_auth_challenge_sent(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata, pjsip_tx_data *tdata)
{
	pjsip_www_authenticate_hdr *auth = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_WWW_AUTHENTICATE, NULL);

#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif

	char nonce[64] = "", call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;

	struct ast_security_event_chal_sent chal_sent = {
			.common.event_type = AST_SECURITY_EVENT_CHAL_SENT,
			.common.version    = AST_SECURITY_EVENT_CHAL_SENT_VERSION,
			.common.service    = "PJSIP",
			.common.account_id = get_account_id(endpoint),
			.common.local_addr = {
					.addr      = &local,
					.transport = transport,
			},
			.common.remote_addr = {
					.addr      = &remote,
					.transport = transport,
			},
			.common.session_id = call_id,
			.challenge         = nonce,
	};

	if (auth && !pj_strcmp2(&auth->scheme, "digest")) {
		ast_copy_pj_str(nonce, &auth->challenge.digest.nonce, sizeof(nonce));
	}

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&chal_sent));
}

void ast_sip_report_req_no_support(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata,
				   const char* req_type)
{
#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif

	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;

	struct ast_security_event_req_no_support req_no_support_event = {
		.common.event_type  = AST_SECURITY_EVENT_REQ_NO_SUPPORT,
		.common.version     = AST_SECURITY_EVENT_REQ_NO_SUPPORT_VERSION,
		.common.service     = "PJSIP",
		.common.account_id  = get_account_id(endpoint),
		.common.local_addr  = {
			.addr       = &local,
			.transport  = transport,
		},
		.common.remote_addr = {
			.addr       = &remote,
			.transport  = transport,
		},
		.common.session_id  = call_id,
		.request_type       = req_type
	};

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&req_no_support_event));
}

void ast_sip_report_mem_limit(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata)
{
#ifdef GRANDSTREAM_NETWORKS
	enum ast_transport transport = update_transport_type(rdata);
#else
	enum ast_transport transport = security_event_get_transport(rdata);
#endif

	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;

	struct ast_security_event_mem_limit mem_limit_event = {
		.common.event_type  = AST_SECURITY_EVENT_MEM_LIMIT,
		.common.version     = AST_SECURITY_EVENT_MEM_LIMIT_VERSION,
		.common.service     = "PJSIP",
		.common.account_id  = get_account_id(endpoint),
		.common.local_addr  = {
			.addr       = &local,
			.transport  = transport,
		},
		.common.remote_addr = {
			.addr       = &remote,
			.transport  = transport,
		},
		.common.session_id  = call_id
	};

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&mem_limit_event));
}
#ifdef GRANDSTREAM_NETWORKS

/* Added by yfyang, for click2call */
void ast_sip_report_click2call(const char *name, pjsip_rx_data *rdata)
{
	enum ast_transport transport = update_transport_type(rdata);
	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;

	struct ast_security_event_inval_acct_id click2call_id = {
		.common.event_type = AST_SECURITY_EVENT_CLICK2CALL,
		.common.version    = AST_SECURITY_EVENT_CLICK2CALL_VERSION,
		.common.service    = "PJSIP",
		.common.account_id = name,
		.common.local_addr = {
			.addr      = &local,
			.transport = transport,
		},
		.common.remote_addr = {
			.addr       = &remote,
			.transport = transport,
		},
		.common.session_id = call_id,
	};

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&click2call_id));
}

void ast_sip_report_invalid_sipmsg(const char *name, pjsip_rx_data *rdata)
{
	enum ast_transport transport = update_transport_type(rdata);
	char call_id[pj_strlen(&rdata->msg_info.cid->id) + 1];
	struct ast_sockaddr local, remote;
	char method_name[64] = {0};
	int  is_request = 0;
	int  resp_code  = 0;

	if (rdata->msg_info.msg->type == PJSIP_RESPONSE_MSG) {
		resp_code = rdata->msg_info.msg->line.status.code;
	} else {
		struct pjsip_request_line req = rdata->msg_info.msg->line.req;
		ast_copy_pj_str(method_name, &req.method.name, sizeof(method_name));
		is_request = 1;
	}

	struct ast_security_event_invalid_sipmsg invalid_sipmsg_id = {
		.common.event_type = AST_SECURITY_EVENT_INVALID_SIP_MSG,
		.common.version    = AST_SECURITY_EVENT_INVALID_SIP_MSG_VERSION,
		.common.service    = "PJSIP",
		.common.account_id = name,
		.common.local_addr = {
			.addr      = &local,
			.transport = transport,
		},
		.common.remote_addr = {
			.addr       = &remote,
			.transport = transport,
		},
		.common.session_id = call_id,
	};

	if (is_request) {
		invalid_sipmsg_id.request_type = method_name;
	} else {
		invalid_sipmsg_id.resp_code = resp_code;
	}

	security_event_populate(rdata, call_id, sizeof(call_id), &local, &remote);

	ast_security_event_report(AST_SEC_EVT(&invalid_sipmsg_id));
}

/*add by wxu@grandstream.cn for system event*/
void sip_report_trunk_reachable(const char *peer_endpoint, int is_reachable)
{
	/* If is_reachable is '1', means 'UNREACHABLE' status, '2' means 'REACHABLE' status. */
	char peer_name[80];
	struct ast_security_event_sip_trunk_reachable sip_trunk_reachable = {
		.common.event_type = AST_SECURITY_EVENT_SIP_TRUNK_REACHABLE,
		.common.version = AST_SECURITY_EVENT_SIP_TRUNK_REACHABLE_VERSION,
		.common.service = "PJSIP",

		.peer_name = peer_name,
		.state = is_reachable,
	};
	ast_copy_string(peer_name, peer_endpoint, sizeof(peer_name));
	ast_security_event_report(AST_SEC_EVT(&sip_trunk_reachable));
}

void sip_report_lost_registration(const char *endpoint_id, const char *endpoint_addr, int transport, int state)
{
	char peer_name[80];
	char user_name[256];
	char host[256];
	struct ast_sockaddr remote;
	struct ast_security_event_sip_lost_registration sip_lost_registration = {
		.common.event_type = AST_SECURITY_EVENT_SIP_LOST_REGISTRATION,
		.common.version = AST_SECURITY_EVENT_SIP_LOST_REGISTRATION_VERSION,
		.common.service = "PJSIP",
		.common.remote_addr = {
			.addr       = &remote,
			.transport  = transport,
		},
		.peer_name = peer_name,
		.user_name = user_name,
		.host = host,
		.state = state,
	};

	ast_copy_string(peer_name, endpoint_id, sizeof(peer_name));
	ast_copy_string(user_name, endpoint_id, sizeof(user_name));
	ast_copy_string(host, "<unkown>", sizeof(host));

	ast_sockaddr_parse(&remote, endpoint_addr, 0);

	ast_log(LOG_DEBUG, "A peer lost registration! name:%s, addr:%s, state[%d]", endpoint_id, endpoint_addr, state);
	ast_security_event_report(AST_SEC_EVT(&sip_lost_registration));
}
static int report_warning_status(void *obj, void *arg, int flags)
{
    struct ast_sip_endpoint *endpoint = obj;
    if (endpoint == NULL)
        return CMP_STOP;
    if (endpoint->aors != NULL && strstr(endpoint->aors, "trunk_") != NULL) {
        sip_report_trunk_reachable(endpoint->aors, endpoint->peer_status);
    }
    ast_log(LOG_DEBUG, "report_warning_status---[%s][%d]", endpoint->aors, endpoint->peer_status);
    return 0;
}

void ast_sip_resend_warning_status(void)
{
    RAII_VAR(struct ao2_container *, endpoints, NULL, ao2_cleanup);
    endpoints =  ast_sip_get_endpoints();
    ao2_callback(endpoints, OBJ_NODATA, report_warning_status, NULL);
}

/*end*/
#endif

