/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2019, Digium, Inc.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/module.h"
#include "asterisk/xml.h"
#include "asterisk/bfcp.h"
#include "asterisk/res_pjsip_bfcp_info.h"
#include "../apps/confbridge/include/conf_resource.h"

AST_MODULE_LOG("pjsip");

struct bfcp_info_pvt {
	struct ast_sip_session *session;
	struct ast_bfcp_info *bfcp_info;
};

static int is_bfcp_info_type(pjsip_rx_data *rdata, char *subtype)
{
	return rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, subtype);
}

static void send_response(struct ast_sip_session *session,
			  struct pjsip_rx_data *rdata, int code)
{
	pjsip_tx_data *tdata;
	pjsip_dialog *dlg = session->inv_session->dlg;

	if (pjsip_dlg_create_response(dlg, rdata, code, NULL, &tdata) == PJ_SUCCESS) {
		struct pjsip_transaction *tsx = pjsip_rdata_get_tsx(rdata);
		pjsip_dlg_send_response(dlg, tsx, tdata);
	}
}

static enum bfcp_prim bfcp_info_str_2_prim(const char *prim)
{
	int i = 0;

	for (i = 0; i <= BFCP_GOODBYE_ACK; i++) {
		if (!strcmp(prim, bfcp_info_prim_strmap[i][0])) {
			return (enum bfcp_prim)i;
		}
	}

	return BFCP_INFO_INVALID;
}

static enum bfcp_reqstatus bfcp_info_str_2_reqstatus(const char *status)
{
	int i;

	for (i = 1; i <= BFCP_REVOKED; i++) {
		if (!strcmp(status, bfcp_info_reqstatus_strmap[i][0])) {
			return (enum bfcp_reqstatus)i;
		}
	}

	return BFCP_INFO_INVALID;
}

static enum bfcp_info_action bfcp_info_str_2_action(const char *prim)
{
	int i = 0;

	for (i = 0; i < BFCP_INFO_ACTION_MAX; i++) {
		if (!strcmp(prim, bfcp_info_action_strmap[i][0])) {
			return (enum bfcp_info_action)i;
		}
	}

	return BFCP_INFO_INVALID;
}

static enum bfcp_info_share_type bfcp_info_str_2_share_type(const char *share_type)
{
	int i = 0;

	for (i = 0; i <= BFCP_INFO_SHARE_TYPE_DOCUMENT; i++) {
		if (!strcmp(share_type, bfcp_info_share_type_strmap[i][0])) {
			return (enum bfcp_info_share_type)i;
		}
	}

	return BFCP_INFO_SHARE_TYPE_UNKNOW;
}

static void bfcp_info_process_floor_request(struct ast_sip_session *session, struct ast_channel *channel,
	struct ast_bfcp_info *bfcp_info, struct bfcp_info_data *bfcp_data)
{
	int resolution = 0;

	if (!session || !channel || !bfcp_info || !bfcp_data) {
		return;
	}

	if (!ast_sip_session_judge_if_has_valid_slides_video(session)) {
		bfcp_data->bfcp_info->status = BFCP_DENIED;
	} else {
		resolution = ast_conf_get_slides_video_resource(bfcp_info->info_conf_id, channel, ast_true(bfcp_info->share_audio));
		if (VIDEO_RESOLUTION_1080P == resolution) {
			ast_log(LOG_DEBUG, "[session:%p] Get slides video resource sucess for conf:%s, channel:%s\n",
				session, bfcp_info->info_conf_id, ast_channel_name(channel));
			bfcp_data->bfcp_info->status = BFCP_GRANTED;
			ast_channel_set_opening_slides_video(channel, 1);
			ast_sip_session_set_bfcp_info_msid(session, bfcp_info);
		} else {
			bfcp_data->bfcp_info->status = BFCP_DENIED;
			if (resolution == -2) {
				bfcp_data->bfcp_info->denied_reason = BFCP_DENIED_REASON_CPU_LIMIT;
			} else if (resolution == -3) {
				bfcp_data->bfcp_info->denied_reason = BFCP_DENIED_REASON_NO_VIDEO_ROOM;
			} else {
				bfcp_data->bfcp_info->denied_reason = BFCP_DENIED_REASON_NO_RESOURCE;
			}
		}
	}

	bfcp_data->bfcp_info->primitive = BFCP_FLOOR_REQUEST_STATUS;

	return;
}

static void bfcp_info_process_floor_release(struct ast_channel *channel,
	struct ast_bfcp_info *bfcp_info, struct bfcp_info_data *bfcp_data)
{
	RAII_VAR(struct ast_channel *, host_channel, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, slider_channel, NULL, ao2_cleanup);

	bfcp_data->bfcp_info->primitive = BFCP_FLOOR_REQUEST_STATUS;

	slider_channel = ast_conf_get_slider_channel(bfcp_info->info_conf_id);
	if (!slider_channel) {
		bfcp_data->bfcp_info->status = BFCP_DENIED;
		return;
	}
	if (slider_channel == channel) {
		if (ast_conf_put_slides_video_resource(bfcp_info->info_conf_id, channel)) {
			bfcp_data->bfcp_info->status = BFCP_DENIED;
			ast_log(LOG_DEBUG, "Release slides video resource failed for conf:%s, channel:%s\n", bfcp_info->info_conf_id, ast_channel_name(channel));
		} else {
			ast_log(LOG_DEBUG, "Release slides video successful\n");
			bfcp_data->bfcp_info->status = BFCP_RELEASED;
		}
	} else {
		host_channel = ast_conf_get_sfu_moderator(channel, bfcp_info->info_conf_id);
		if (channel == host_channel) {
			if (ast_conf_put_slides_video_resource(bfcp_info->info_conf_id, slider_channel)) {
				bfcp_data->bfcp_info->status = BFCP_DENIED;
				ast_log(LOG_DEBUG, "Release slides video resource failed for conf:%s, channel:%s\n", bfcp_info->info_conf_id, ast_channel_name(channel));
			} else {
				ast_log(LOG_DEBUG, "Release slides video forcely by host\n");
				bfcp_data->bfcp_info->status = BFCP_RELEASED;
				bfcp_data->bfcp_info->closed_by_host = 1;
				ast_indicate(slider_channel, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO);
			}
		} else {
			ast_log(LOG_DEBUG, "Release slides video forcely failed, it's not host\n");
			bfcp_data->bfcp_info->status = BFCP_DENIED;
		}
	}

	return;
}

static void bfcp_info_process_floor_cancel(struct bfcp_info_data *bfcp_data)
{
	if (!bfcp_data || !bfcp_data->bfcp_info) {
		return;
	}

	bfcp_data->bfcp_info->status = BFCP_RELEASED;
	bfcp_data->bfcp_info->primitive = BFCP_FLOOR_REQUEST_STATUS;

	return;
}

static void bfcp_info_set_chan_floor_owner_id(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	if (!session || !bfcp_info) {
		return;
	}

	if (strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return;
	}

	if (!ast_strlen_zero(bfcp_info->slider_name)) {
		pbx_builtin_setvar_helper(session->channel, "FLOOR_OWNER_ID", bfcp_info->slider_name);
		ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_chan_floor_owner_id floor_owner_id:%s, endpoint:%s\n",
			session, bfcp_info->slider_name, ast_sorcery_object_get_id(session->endpoint));
	} else if (!ast_strlen_zero(bfcp_info->floor_owner_id)) {
		pbx_builtin_setvar_helper(session->channel, "FLOOR_OWNER_ID", bfcp_info->floor_owner_id);
		ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_chan_floor_owner_id floor_owner_id:%s, endpoint:%s\n",
			session, bfcp_info->floor_owner_id, ast_sorcery_object_get_id(session->endpoint));
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_chan_floor_owner_id not set, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void bfcp_info_process_floor_granted_status(struct ast_sip_session *session, struct ast_channel *channel,
	struct ast_bfcp_info *bfcp_info)
{
	int resolution = 0;

	ast_log(LOG_DEBUG, "[session:%p] Receive floor status granted for conf:%s, channel:%s\n",
		session, bfcp_info->info_conf_id, ast_channel_name(channel));

	ast_channel_set_opening_slides_video(channel, 1);
	bfcp_info_set_chan_floor_owner_id(session, bfcp_info);

	resolution = ast_conf_get_slides_video_resource(bfcp_info->info_conf_id, channel, ast_true(bfcp_info->share_audio));
	if (VIDEO_RESOLUTION_1080P == resolution) {
		ast_log(LOG_DEBUG, "[session:%p] Get slides video resource sucess for conf:%s, channel:%s\n",
			session, bfcp_info->info_conf_id, ast_channel_name(channel));
		ast_sip_session_set_bfcp_info_msid(session, bfcp_info);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Can't get slides video resource sucess for conf:%s, channel:%s\n",
			session, bfcp_info->info_conf_id, ast_channel_name(channel));
		bfcp_info->status = BFCP_DENIED;
	}

	return;
}

static void bfcp_info_process_floor_released_status(struct ast_sip_session *session, struct ast_channel *channel,
	struct ast_bfcp_info *bfcp_info)
{
	RAII_VAR(struct ast_channel *, slider_channel, NULL, ao2_cleanup);

	ast_log(LOG_DEBUG, "[session:%p] Receive floor status released for conf:%s, channel:%s\n",
		session, bfcp_info->info_conf_id, ast_channel_name(channel));
	ast_channel_set_opening_slides_video(channel, 0);

	slider_channel = ast_conf_get_slider_channel(bfcp_info->info_conf_id);
	if (!slider_channel) {
		return;
	}

	if (ast_conf_put_slides_video_resource(bfcp_info->info_conf_id, slider_channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Release slides video resource failed for conf:%s, channel:%s\n",
			session, bfcp_info->info_conf_id, ast_channel_name(channel));
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Release slides video resource forcely for conf:%s, channel:%s\n",
			session, bfcp_info->info_conf_id, ast_channel_name(channel));
	}

	return;
}


static void bfcp_info_process_floor_status(struct ast_sip_session *session, struct ast_channel *channel,
	struct ast_bfcp_info *bfcp_info)
{
	if (!session || !channel || !bfcp_info) {
		return;
	}

	switch (bfcp_info->status) {
		case BFCP_GRANTED:
			bfcp_info_process_floor_granted_status(session, channel, bfcp_info);
			break;
		case BFCP_RELEASED:
			bfcp_info_process_floor_released_status(session, channel, bfcp_info);
			break;
		default:
			break;
	}

	return;
}

static void bfcp_info_notify_open_share_audio(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	struct ast_app *app = NULL;
	char appdata[64] = {0};

	if (!session || !session->channel || !bfcp_info) {
		return;
	}

	if (strcasecmp(bfcp_info->share_audio, "true")) {
		return;
	}

	if (ast_strlen_zero(session->conf_room)) {
		ast_log(LOG_ERROR, "[session:%p] bfcp_info_notify_share_audio_for_open, can't get conf_room endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(appdata, sizeof(appdata), "%s,%s", session->conf_room, "true");
	app = pbx_findapp("ConfBridgeFloorContrl");
	if (app) {
		ast_log(LOG_VERBOSE, "Launching %s(%s) on %s\n", "ConfBridgeFloorContrl", S_OR(appdata, ""),
			ast_channel_name(session->channel));
		pbx_exec(session->channel, app, S_OR(appdata, ""));
		session->notify_share_audio = 1;
	} else {
		ast_log(LOG_ERROR, "[session:%p] bfcp_info_notify_share_audio_for_open, can't ConfBridgeFloorContrl APP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void bfcp_info_set_conf_slider_name(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	char *slider_name = NULL;

	if (!session || !session->channel || !bfcp_info) {
		return;
	}

	if (strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		slider_name = S_COR(ast_channel_caller(session->channel)->id.name.valid, ast_channel_caller(session->channel)->id.name.str, "");
		if (!ast_strlen_zero(slider_name)) {
			ast_conf_set_slider_name(bfcp_info->info_conf_id, slider_name);
			ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_conf_slider_name:%s, endpoint:%s\n",
				session, slider_name, ast_sorcery_object_get_id(session->endpoint));
		}
	} else {
		ast_conf_set_slider_name(bfcp_info->info_conf_id, bfcp_info->account_id);
		ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_conf_slider_name:%s, endpoint:%s\n",
			session, bfcp_info->account_id, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void bfcp_info_set_data_for_floor_granted(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info)
{
	ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_data_for_floor_granted, set conf:%s share type:%s endpoint:%s\n",
		session, bfcp_info->info_conf_id, bfcp_info->share_type, ast_sorcery_object_get_id(session->endpoint));

	bfcp_info->primitive = BFCP_FLOOR_STATUS;
	bfcp_info->status = BFCP_GRANTED;
	strcpy(bfcp_info->floor_owner_id, bfcp_info->account_id);

	ast_indicate(channel, AST_CONTROL_NOTIFY_OPEN_SLIDES_VIDEO);

	bfcp_info_notify_open_share_audio(session, bfcp_info);

	ast_queue_control_data(channel, AST_CONTROL_BFCP_CONTROL, bfcp_info, sizeof(struct ast_bfcp_info));

	ast_conf_set_slides_type(bfcp_info->info_conf_id, bfcp_info_str_2_share_type(bfcp_info->share_type));

	bfcp_info_set_conf_slider_name(session, bfcp_info);

	return;
}

static void bfcp_info_after_process_floor_request(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info, struct bfcp_info_data *bfcp_data)
{
	if (bfcp_data->bfcp_info->status != BFCP_GRANTED) {
		return;
	}

	bfcp_info_set_data_for_floor_granted(session, channel, bfcp_info);

	return;
}

static void bfcp_info_after_process_floor_release(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info, struct bfcp_info_data *bfcp_data)
{
	if (bfcp_data->bfcp_info->status != BFCP_RELEASED) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] bfcp_info_after_process_floor_release for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	bfcp_info->primitive = BFCP_FLOOR_STATUS;
	bfcp_info->status = BFCP_RELEASED;
	if (!bfcp_data->bfcp_info->closed_by_host) {
		ast_indicate(channel, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO);
	} else {
		ast_indicate_data(channel, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO_BY_HOST, bfcp_info, sizeof(struct ast_bfcp_info));
	}

	ast_queue_control_data(channel, AST_CONTROL_BFCP_CONTROL, bfcp_info, sizeof(struct ast_bfcp_info));

	return;
}

static void bfcp_info_after_process_floor_granted_status(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info)
{
	if (bfcp_info->status != BFCP_GRANTED) {
		return;
	}

	bfcp_info_set_data_for_floor_granted(session, channel, bfcp_info);

	return;
}

static void bfcp_info_after_process_floor_released_status(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info)
{
	ast_log(LOG_DEBUG, "[session:%p] bfcp_info_after_process_floor_released_status for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	bfcp_info->primitive = BFCP_FLOOR_STATUS;
	bfcp_info->status = BFCP_RELEASED;

	ast_indicate(channel, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO);

	ast_queue_control_data(channel, AST_CONTROL_BFCP_CONTROL, bfcp_info, sizeof(struct ast_bfcp_info));

	return;
}

static void bfcp_info_after_process_floor_status(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info)
{
	switch (bfcp_info->status) {
		case BFCP_GRANTED:
			bfcp_info_after_process_floor_granted_status(session, channel, bfcp_info);
			break;
		case BFCP_RELEASED:
			bfcp_info_after_process_floor_released_status(session, channel, bfcp_info);
			break;
		default:
			ast_log(LOG_DEBUG, "[session:%p] BFCP receive unknow info, can't process now for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
	}

	return;
}

static void bfcp_info_after_process_conf_action(struct ast_sip_session *session,
	struct ast_channel *channel, struct ast_bfcp_info *bfcp_info, struct bfcp_info_data *bfcp_data)
{
	switch (bfcp_info->action) {
		case BFCP_INFO_ACTION_FLOOR_REQUEST:
			bfcp_info_after_process_floor_request(session, channel, bfcp_info, bfcp_data);
			break;
		case BFCP_INFO_ACTION_FLOOR_RELEASE:
			bfcp_info_after_process_floor_release(session, channel, bfcp_info, bfcp_data);
			break;
		case BFCP_INFO_ACTION_FLOOR_STATUS:
			bfcp_info_after_process_floor_status(session, channel, bfcp_info);
			break;
		default:
			ast_log(LOG_DEBUG, "[session:%p] BFCP receive unknow info, can't process now for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
	}

	return;
}

static int bfcp_info_process_conf_action(struct ast_channel *channel, struct ast_bfcp_info *bfcp_info)
{
	struct ast_sip_channel_pvt *channel_pvt = NULL;
	struct bfcp_info_data *bfcp_data = NULL;

	if (!channel || !bfcp_info) {
		return -1;
	}

	channel_pvt = ast_channel_tech_pvt(channel);
	if (!channel_pvt || !channel_pvt->session) {
		ast_log(LOG_ERROR, "Can't get channel pvt\n");
		return -1;
	}

	strcpy(bfcp_info->floor_owner_id, bfcp_info->account_id);

	bfcp_data = ast_sip_session_bfcp_data_alloc(channel_pvt->session, bfcp_info);
	if (!bfcp_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc bfcp data for endpoint:%s\n",
			channel_pvt->session, ast_sorcery_object_get_id(channel_pvt->session->endpoint));
		return -1;
	}

	switch (bfcp_info->action) {
		case BFCP_INFO_ACTION_FLOOR_REQUEST:
			bfcp_info_process_floor_request(channel_pvt->session, channel, bfcp_info, bfcp_data);
			break;
		case BFCP_INFO_ACTION_FLOOR_RELEASE:
			bfcp_info_process_floor_release(channel, bfcp_info, bfcp_data);
			break;
		case BFCP_INFO_ACTION_FLOOR_CANCEL:
			bfcp_info_process_floor_cancel(bfcp_data);
			break;
		case BFCP_INFO_ACTION_FLOOR_STATUS:
			bfcp_info_process_floor_status(channel_pvt->session, channel, bfcp_info);
			break;
		default:
			ast_log(LOG_DEBUG, "[session:%p] BFCP receive unknow info, can't process now for endpoint:%s\n",
				channel_pvt->session, ast_sorcery_object_get_id(channel_pvt->session->endpoint));
			break;
	}

	if (bfcp_data && (bfcp_info->action != BFCP_INFO_ACTION_FLOOR_STATUS)) {
		if (ast_sip_push_task(channel_pvt->session->serializer, ast_sip_session_bfcp_send_sip_info, bfcp_data)) {
			ao2_cleanup(bfcp_data);
		}
	}

	bfcp_info_after_process_conf_action(channel_pvt->session, channel, bfcp_info, bfcp_data);

	return 0;
}

static int bfcp_info_process_xml(char *xml, struct ast_bfcp_info *bfcp_info)
{
	struct ast_xml_doc  *xml_doc          = NULL;
	struct ast_xml_node *bfcp_info_node   = NULL;
	struct ast_xml_node *child_nodes      = NULL;
	struct ast_xml_node *node_iterator    = NULL;
	struct ast_xml_node *node_conf_id     = NULL;
	struct ast_xml_node *node_user_id     = NULL;
	struct ast_xml_node *node_share_type  = NULL;
	struct ast_xml_node *node_action      = NULL;
	struct ast_xml_node *node_share_audio = NULL;
	struct ast_xml_node *node_floor_owned_user_id = NULL;
	const char *node_name;
	const char *content;
	const char *action_name;
	char status[32]     = {0};
	//char user_id[256]   = {0};

	if (!xml) {
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		return -1;
	}

	bfcp_info_node = ast_xml_get_root(xml_doc);
	if (!bfcp_info_node) {
		ast_xml_close(xml_doc);
		return -1;
	}

	child_nodes = ast_xml_node_get_children(bfcp_info_node);
	if (child_nodes) {
		for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "Action") && !node_action) {
				node_action = node_iterator;
			} else if (!strcmp(node_name, "ConfID") && !node_conf_id) {
				node_conf_id = node_iterator;
			} else if (!strcmp(node_name, "UserID") && !node_user_id) {
				node_user_id = node_iterator;
			} else if (!strcmp(node_name, "ShareType") && !node_share_type) {
				node_share_type = node_iterator;
			} else if (!strcmp(node_name, "ShareAudio") && !node_share_audio) {
				node_share_audio = node_iterator;
			} else if (!strcmp(node_name, "FloorOwnedUserID") && !node_floor_owned_user_id) {
				node_floor_owned_user_id = node_iterator;
			}
		}
	}

	if (node_action) {
		action_name = ast_xml_get_attribute(node_action, "name");
		bfcp_info->primitive = bfcp_info_str_2_prim(action_name);
		bfcp_info->action = bfcp_info_str_2_action(action_name);
		ast_log(LOG_NOTICE, "bfcp_info_process_xml, action[%s]\n", action_name);
		ast_xml_free_attr(action_name);

		for (node_iterator = ast_xml_node_get_children(node_action); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			content = ast_xml_get_text(node_iterator);
			if (!content) {
				continue;
			}

			if (!strcmp(node_name, "Status")) {
				snprintf(status, sizeof(status), "%s", content);
				bfcp_info->status = bfcp_info_str_2_reqstatus(status);
				ast_log(LOG_NOTICE, "bfcp_info_process_xml, Status[%s]\n", status);
			}
			ast_xml_free_text(content);
		}
	}

	if (node_conf_id) {
		content = ast_xml_get_text(node_conf_id);
		if (content) {
			strncpy(bfcp_info->info_conf_id, content, sizeof(bfcp_info->info_conf_id) - 1);
			ast_log(LOG_NOTICE, "bfcp_info_process_xml, info_conf_id[%s]\n", content);
		}
		ast_xml_free_text(content);
	}

	if (node_user_id) {
		content = ast_xml_get_text(node_user_id);
		if (content) {
			strncpy(bfcp_info->account_id, content, sizeof(bfcp_info->account_id) - 1);
			ast_log(LOG_NOTICE, "bfcp_info_process_xml, account_id[%s]\n", content);
		}
		ast_xml_free_text(content);
	}

	if (node_share_type) {
		content = ast_xml_get_text(node_share_type);
		if (content) {
			strncpy(bfcp_info->share_type, content, sizeof(bfcp_info->share_type) - 1);
			ast_log(LOG_NOTICE, "bfcp_info_process_xml, share_type[%s]\n", content);
		}
		ast_xml_free_text(content);
	}

	if (node_share_audio) {
		content = ast_xml_get_text(node_share_audio);
		if (content) {
			strncpy(bfcp_info->share_audio, content, sizeof(bfcp_info->share_audio) - 1);
			ast_log(LOG_NOTICE, "bfcp_info_process_xml, share_audio[%s]\n", content);
		}
		ast_xml_free_text(content);
	}

	if (node_floor_owned_user_id) {
		content = ast_xml_get_text(node_floor_owned_user_id);
		if (content) {
			strncpy(bfcp_info->floor_owner_id, content, sizeof(bfcp_info->floor_owner_id) - 1);
			ast_log(LOG_NOTICE, "bfcp_info_process_xml, floor_owned_user_id[%s]\n", content);
		}
		ast_xml_free_text(content);
	}

	ast_xml_close(xml_doc);

	return 0;
}

static void bfcp_info_deny_request(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	struct bfcp_info_data *bfcp_data = NULL;
	struct ast_bfcp_info new_bfcp_info;

	if (!session || !bfcp_info) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] UCM type is not support BFCP, deny bfcp request for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	memcpy(&new_bfcp_info, bfcp_info, sizeof(struct ast_bfcp_info));

	switch (bfcp_info->action) {
		case BFCP_INFO_ACTION_FLOOR_REQUEST:
		case BFCP_INFO_ACTION_FLOOR_RELEASE:
		case BFCP_INFO_ACTION_FLOOR_CANCEL:
			new_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
			new_bfcp_info.status = BFCP_DENIED;
			break;
		default:
			new_bfcp_info.primitive = BFCP_FLOOR_STATUS;
			new_bfcp_info.status = BFCP_RELEASED;
			ast_log(LOG_DEBUG, "[session:%p] BFCP receive unknow info, send release as default for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
	}

	bfcp_data = ast_sip_session_bfcp_data_alloc(session, &new_bfcp_info);
	if (!bfcp_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc bfcp data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_push_task(session->serializer, ast_sip_session_bfcp_send_sip_info, bfcp_data)) {
		ast_log(LOG_ERROR, "[session:%p] Push task failed for deny bfcp request, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(bfcp_data);
	}

	return;
}

static void bfcp_info_set_session_share_audio(struct ast_sip_session *session, struct ast_bfcp_info bfcp_info)
{
	if (!session) {
		return;
	}

	if (!strcasecmp(bfcp_info.share_audio, "true")) {
		session->notify_share_audio = 1;
	} else {
		session->notify_share_audio = 0;
	}

	ast_log(LOG_ERROR, "[session:%p] bfcp_info_set_session_share_audio notify_share_audio:%d, endpoint:%s\n",
		session, session->notify_share_audio, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void bfcp_info_set_slider_name(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	char *slider_name = NULL;

	if (!session || !bfcp_info) {
		ast_log(LOG_ERROR, "bfcp_info_set_slider_name param is NULL\n");
		return;
	}

	if (strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		slider_name = S_COR(ast_channel_caller(session->channel)->id.name.valid, ast_channel_caller(session->channel)->id.name.str, "");
		if (!ast_strlen_zero(slider_name)) {
			strncpy(bfcp_info->slider_name, slider_name, sizeof(bfcp_info->slider_name) - 1);
			ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_slider_name:%s, endpoint:%s\n",
				session, slider_name, ast_sorcery_object_get_id(session->endpoint));
		}
	} else {
		strncpy(bfcp_info->slider_name, bfcp_info->account_id, sizeof(bfcp_info->slider_name) - 1);
		ast_log(LOG_DEBUG, "[session:%p] bfcp_info_set_slider_name:%s, endpoint:%s\n",
			session, bfcp_info->account_id, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static int bfcp_info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	int res = 0;
	pjsip_msg_body *body = NULL;
	char *content = NULL;
	struct ast_bfcp_info bfcp_info;
	int is_bfcp_info = 0;

	if (!session || !rdata || !rdata->msg_info.msg) {
		return 0;
	}

	is_bfcp_info = is_bfcp_info_type(rdata, BFCP_INFO_SUB_TYPE);
	if (!is_bfcp_info) {
		is_bfcp_info = is_bfcp_info_type(rdata, BFCP_INFO_CONF_SUB_TYPE);
		if (!is_bfcp_info) {
			return 0;
		}
	}

	body = rdata->msg_info.msg->body;

	if (!body || !body->len) {
		return 0;
	}

	char buf[body ? (body->len + 1) : 0];
	memset(buf, 0, sizeof(buf));
	body->print_body(body, buf, body->len);

	if (!session->channel) {
		ast_log(LOG_DEBUG, "Session channel is NULL\n");
		return 0;
	}

	send_response(session, rdata, 200);

	ast_channel_lock(session->channel);

	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ZOMBIE) || ast_check_hangup(session->channel)) {
		ast_channel_unlock(session->channel);
		ast_log(LOG_DEBUG, "Channel has been hung up\n");
		return 0;
	}

	content = (char*) ast_malloc(body->len + 1);
	if (!content) {
		ast_channel_unlock(session->channel);
		return 0;
	}
	memset(content, 0, body->len + 1);
	strncpy(content, (char*)body->data, body->len);

	memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
	res = bfcp_info_process_xml(content, &bfcp_info);

	if (ast_sip_session_get_only_support_audio()) {
		bfcp_info_deny_request(session, &bfcp_info);
		ast_free(content);
		content = NULL;
		ast_channel_unlock(session->channel);
		return 0;
	}

	if (!res) {
		bfcp_info_set_slider_name(session, &bfcp_info);
		if (session->is_sfu) {
			if (strlen(bfcp_info.info_conf_id) == 0) {
				strncpy(bfcp_info.info_conf_id, session->conf_room, sizeof(bfcp_info.info_conf_id) - 1);
			}
			bfcp_info_process_conf_action(session->channel, &bfcp_info);
		} else {
			memset(bfcp_info.info_conf_id, 0, sizeof(bfcp_info.info_conf_id) - 1);
			ast_bfcp_set_channel_slides_status(session->channel, &bfcp_info, 1);
			bfcp_info_set_session_share_audio(session, bfcp_info);
			bfcp_info_set_chan_floor_owner_id(session, &bfcp_info);
			ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(bfcp_info));
		}
	}

	ast_log(LOG_NOTICE, "bfcp_info_incoming_request, process this info\n");

    if (content) {
        ast_free(content);
    }

    ast_channel_unlock(session->channel);

	return 1;
}

static void bfcp_info_incoming_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	if (!session || !rdata) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] bfcp_info_incoming_response clear pending state for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session->pending_bfcp_info = 0;

	return;
}

static struct ast_sip_session_supplement bfcp_info_supplement = {
	.method = "INFO",
	.incoming_request  = bfcp_info_incoming_request,
	.incoming_response = bfcp_info_incoming_response,
	.response_priority = AST_SIP_SESSION_BEFORE_MEDIA | AST_SIP_SESSION_AFTER_MEDIA,
};

static int load_module(void)
{
	ast_sip_session_register_supplement(&bfcp_info_supplement);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&bfcp_info_supplement);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP BFCP INFO Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
		.requires = "res_pjsip,res_pjsip_pubsub",
);

