/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Cui Xiaoxia <xxcui@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_pubsub</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#include "asterisk/module.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/res_pjsip_sca.h"
#include "asterisk/logger.h"
#include "asterisk/astobj2.h"
#include "asterisk/taskprocessor.h"

AST_MODULE_LOG("pjsip");

#define SCA_CONTEXT "sca-lines"

struct call_info_subscription {
	/*! The SIP subscription */
	struct ast_sip_subscription *sip_sub;
	/*! The serializer to use for notifications */
	struct ast_taskprocessor *serializer;
	/*! Context in which subscription looks for updates */
	char context[AST_MAX_CONTEXT];
	/*! The subscriber */
	char exten[AST_MAX_EXTENSION];
	/*! The shared line */
	char sharedline[AST_MAX_EXTENSION];
	/*! The exten role, shared or private */
	char  is_shared;
	/*! to indicate if the sca user has been removed */
	char  sca_removed; 
	/*! The buf to store call-info from sca module */
	char call_info[2048];

	/*! Next item in the list */
	AST_LIST_ENTRY(call_info_subscription) next;
};

static struct stasis_subscription *call_info_state_sub;

static void subscription_shutdown(struct ast_sip_subscription *sub);
static int new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource);
static int subscription_established(struct ast_sip_subscription *sub);
static void *get_notify_data(struct ast_sip_subscription *sub);
static void *check_and_get_notify_data(struct ast_sip_subscription *sub);

AST_RWLIST_HEAD_STATIC(callinfo_subs, call_info_subscription);

struct ast_sip_notifier call_info_notifier = {
	.new_subscribe = new_subscribe,
	.subscription_established = subscription_established,
	.get_notify_data = get_notify_data,
    .check_and_get_notify_data = check_and_get_notify_data,
};

struct ast_sip_subscription_handler call_info_handler = {
	.event_name = "call-info",
	.subscription_shutdown = subscription_shutdown,
	.notifier = &call_info_notifier,
};

static struct call_info_subscription *find_callinfo_subscription(char* sharedline, char* exten)
{
	struct call_info_subscription *i;
	SCOPED_LOCK(lock, &callinfo_subs, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	if (AST_LIST_EMPTY(&callinfo_subs)) {
		return NULL;
	}

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&callinfo_subs, i, next) {
		if (!strcmp(i->sharedline, sharedline) && !strcmp(i->exten, exten)) {
			return i;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

	return NULL;
}

static void call_info_subscription_destructor(void *obj)
{
	struct call_info_subscription *sub = obj;

	ast_sip_subscription_destroy(sub->sip_sub);
	sub->sip_sub = NULL;
	ast_taskprocessor_unreference(sub->serializer);
}

/*!
 * \internal
 * \brief Allocates an call_info_subscription object.
 *
 * Creates the underlying SIP subscription for the given request. First makes
 * sure that there are registered handler and provider objects available.
 */
static struct call_info_subscription *call_info_subscription_alloc(
		struct ast_sip_subscription *sip_sub, struct ast_sip_endpoint *endpoint)
{
	struct call_info_subscription *call_info_sub;

	call_info_sub = ao2_alloc(sizeof(*call_info_sub), call_info_subscription_destructor);
	if (!call_info_sub) {
		return NULL;
	}

	call_info_sub->sip_sub = sip_sub;

	/* We keep our own reference to the serializer as there is no guarantee in state_changed
	 * that the subscription tree is still valid when it is called. This can occur when
	 * the subscription is terminated at around the same time as the state_changed
	 * callback is invoked.
	 */
	call_info_sub->serializer = ao2_bump(ast_sip_subscription_get_serializer(sip_sub));
	return call_info_sub;
}

static int send_call_info_notify(void *obj)
{
	RAII_VAR(struct call_info_subscription *, task_data, obj, ao2_cleanup);
	struct call_info_subscription *call_info_sub = task_data;
	struct ast_sip_body_data data;
	char uri[PJSIP_MAX_URL_SIZE] = {0};

	if(!call_info_sub || !call_info_sub->sip_sub) {
		return -1;
	}

	ast_sip_subscription_get_local_uri(call_info_sub->sip_sub, uri, sizeof(uri));
	memset(call_info_sub->call_info, 0, sizeof(call_info_sub->call_info));

	if (call_info_sub->sca_removed) {
        ast_log(LOG_NOTICE, "send_call_info_notify, sca user[%s] removed, set default call-info", call_info_sub->sharedline);
		snprintf(call_info_sub->call_info, sizeof(call_info_sub->call_info), "<%s>;appearance-index=*;appearance-state=idle", uri);
	} else {
		if (ast_sip_get_call_info(call_info_sub->sharedline, uri, call_info_sub->call_info, sizeof(call_info_sub->call_info))) {
			return -1;
		}
	}

	data.body_data = call_info_sub->call_info;
    ast_log(LOG_NOTICE, "send_call_info_notify, body_data[%s]", (char *)data.body_data);
	ast_sip_subscription_notify(call_info_sub->sip_sub, &data, call_info_sub->sca_removed);

	return 0;
}

static void call_info_state_cb(void *unused, struct stasis_subscription *sub, struct stasis_message *msg)
{
	struct ast_device_state_message *dev_state;
    char   sharedline[80] = {0};
	struct call_info_subscription *call_info_sub;
	SCOPED_LOCK(lock, &callinfo_subs, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	if (AST_LIST_EMPTY(&callinfo_subs)) {
		return;
	}

	if (ast_device_state_message_type() != stasis_message_type(msg)) {
		return;
	}

	dev_state = stasis_message_data(msg);
	if (dev_state->eid) {
		/* ignore non-aggregate states */
		return;
	}

    if (strncmp(dev_state->device, "SCA:", 4)) {
        return;
    }

    strncpy(sharedline, dev_state->device + 4, sizeof(sharedline) -1);

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&callinfo_subs, call_info_sub, next) {
		if (!strcmp(call_info_sub->sharedline, sharedline))  {

			ao2_ref(call_info_sub, +1);

			if (AST_DEVICE_REMOVED == dev_state->state) {
				call_info_sub->sca_removed = 1;
			}

			if (ast_sip_push_task(call_info_sub->serializer, send_call_info_notify, call_info_sub)) {
				ao2_cleanup(call_info_sub);
				continue;
			}
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

	return;
}


static struct ast_datastore_info ds_info = { };
static const char ds_name[] = "call-info datastore";

/*!
 * \internal
 * \brief Add a datastore for call_info_subscription.
 *
 * Adds the call_info_subscription wrapper object to a datastore so it can be retrieved
 * later based upon its association with the ast_sip_subscription.
 */
static int call_info_add_datastore(struct call_info_subscription  *call_info_sub)
{
	RAII_VAR(struct ast_datastore *, datastore,
		 ast_sip_subscription_alloc_datastore(&ds_info, ds_name), ao2_cleanup);

	if (!datastore) {
		return -1;
	}

	datastore->data = call_info_sub;
	ast_sip_subscription_add_datastore(call_info_sub->sip_sub, datastore);
	ao2_ref(call_info_sub, +1);
	return 0;
}

/*!
 * \internal
 * \brief Get the call_info_subscription object associated with the given
 * ast_sip_subscription in the datastore.
 */
static struct call_info_subscription *get_call_info_sub(
	struct ast_sip_subscription *sub)
{
	RAII_VAR(struct ast_datastore *, datastore,
		 ast_sip_subscription_get_datastore(sub, ds_name), ao2_cleanup);

	return datastore ? datastore->data : NULL;
}

static void subscription_shutdown(struct ast_sip_subscription *sub)
{
	struct call_info_subscription *call_info_sub = get_call_info_sub(sub);
	struct call_info_subscription *call_info_sub0 = NULL;
	int    need_terminated = 1;

	if (!call_info_sub) {
		return;
	}

	ast_sip_subscription_remove_datastore(call_info_sub->sip_sub, ds_name);
	ao2_cleanup(call_info_sub);  /* remove data store reference */

	AST_RWLIST_WRLOCK(&callinfo_subs);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&callinfo_subs, call_info_sub0, next) {
		if (!call_info_sub0) {
			break;
		}

		if (call_info_sub0 == call_info_sub) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ao2_ref(call_info_sub, -1);
			continue;
		}

		if (!strcmp(call_info_sub->exten, call_info_sub0->exten) && !strcmp(call_info_sub->sharedline, call_info_sub0->sharedline)) {
			need_terminated = 0;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
	AST_RWLIST_UNLOCK(&callinfo_subs);

	if (need_terminated) {
		manager_event(EVENT_FLAG_SYSTEM, "SubscriptionStatus", "Type: call-info\r\nRole: %s\r\nExten: %s\r\nResource: %s\r\nStatus: terminated\r\n",
			                          (call_info_sub->is_shared ? "shared":"private"), call_info_sub->exten, call_info_sub->sharedline);
	} else {
		ast_log(LOG_NOTICE, "subscription_shutdown, don't report terminated to exten[%s]", call_info_sub->exten);
	}

	ao2_cleanup(call_info_sub);
}

static int new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource)
{
	if (strcmp(endpoint->aors, resource)) {
		ast_log(LOG_NOTICE, "Call-info subscription failed: Extension %s does not subscribe itself but %s\n",
			endpoint->aors, resource);
		return 403;
	}

	if (!endpoint->sca_enable || ast_strlen_zero(endpoint->sca_sharedline)) {
		ast_log(LOG_NOTICE, "Call-info subscription failed: Extension %s does not enable SCA\n",
			resource);
		return 403;
	}

	return 200;
}

static int subscription_established(struct ast_sip_subscription *sip_sub)
{
	struct ast_sip_endpoint *endpoint = ast_sip_subscription_get_endpoint(sip_sub);
	const char *sharedline = endpoint->sca_sharedline;
	struct call_info_subscription *call_info_sub;

	if (!(call_info_sub = call_info_subscription_alloc(sip_sub, endpoint))) {
		ao2_cleanup(endpoint);
		return -1;
	}

	if (endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_NONE) {
		ao2_cleanup(endpoint);
		ao2_cleanup(call_info_sub);
		return -1;
	} else if (endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_SHARED) {
		call_info_sub->is_shared = 1;
	} else {
		call_info_sub->is_shared = 0;
	}

	call_info_sub->sca_removed = 0;

	ast_copy_string(call_info_sub->context, SCA_CONTEXT, sizeof(call_info_sub->context));
	ast_copy_string(call_info_sub->sharedline, sharedline, sizeof(call_info_sub->sharedline));
	ast_copy_string(call_info_sub->exten, ast_sorcery_object_get_id(endpoint), sizeof(call_info_sub->exten));

    AST_RWLIST_WRLOCK(&callinfo_subs);
	ao2_ref(call_info_sub, +1);
	AST_RWLIST_INSERT_TAIL(&callinfo_subs, call_info_sub, next);
	AST_RWLIST_UNLOCK(&callinfo_subs);

	manager_event(EVENT_FLAG_SYSTEM, "SubscriptionStatus", "Type: call-info\r\nRole: %s\r\nExten: %s\r\nResource: %s\r\nStatus: active\r\n",
			                         (call_info_sub->is_shared ? "shared":"private"), call_info_sub->exten, call_info_sub->sharedline);

	/* Go ahead and cleanup the endpoint since we don't need it anymore */
	ao2_cleanup(endpoint);

	/* bump the ref since ast_extension_state_add holds a reference */
	ao2_ref(call_info_sub, +1);

	if (call_info_add_datastore(call_info_sub)) {
		ast_log(LOG_WARNING, "Unable to add to call info subscription datastore.\n");
		ao2_cleanup(call_info_sub);
		return -1;
	}

	ao2_cleanup(call_info_sub);
	return 0;
}

static void *get_notify_data(struct ast_sip_subscription *sub)
{
	struct call_info_subscription *call_info_sub;
	char uri[PJSIP_MAX_URL_SIZE] = {0};

	call_info_sub = get_call_info_sub(sub);
	if (!call_info_sub) {
		return NULL;
	}

    ast_log(LOG_NOTICE, "get_notify_data, sharedline[%s]", call_info_sub->sharedline);
	ast_sip_subscription_get_local_uri(sub, uri, sizeof(uri));
	memset(call_info_sub->call_info, 0, sizeof(call_info_sub->call_info));
	if (ast_sip_get_call_info(call_info_sub->sharedline, uri, call_info_sub->call_info, sizeof(call_info_sub->call_info))) {
		return NULL;
	}
	return call_info_sub->call_info;
}


static void *check_and_get_notify_data(struct ast_sip_subscription *sub)
{
	struct call_info_subscription *call_info_sub;
	char uri[PJSIP_MAX_URL_SIZE] = {0};

	call_info_sub = get_call_info_sub(sub);
	if (!call_info_sub) {
		return NULL;
	}

    ast_log(LOG_NOTICE, "check_and_get_notify_data, sharedline[%s]", call_info_sub->sharedline);

    if (!call_info_sub->sca_removed) {
		return NULL;
    }

	call_info_sub->sca_removed = 0;

	ast_sip_subscription_get_local_uri(sub, uri, sizeof(uri));
	memset(call_info_sub->call_info, 0, sizeof(call_info_sub->call_info));

	if (0 == ast_sip_get_call_info(call_info_sub->sharedline, uri, call_info_sub->call_info, sizeof(call_info_sub->call_info))) {
		ast_log(LOG_NOTICE, "check_and_get_notify_data, get call_info[%s]", call_info_sub->call_info);
		return call_info_sub->call_info;
	}

	return NULL;
}


static int notify_callinfo_terminated(void *obj)
{
	RAII_VAR(struct call_info_subscription *, task_data, obj, ao2_cleanup);
	struct call_info_subscription *callinfo_sub = task_data;

	if (ast_sip_notify_terminated(callinfo_sub->sip_sub)) {
		return -1;
	}

	return 0;
}

static int call_info_terminated(char* sharedline, char* exten)
{
	struct call_info_subscription *callinfo_sub;

    if (!sharedline || !exten) {
        return -1;
    }

	callinfo_sub = find_callinfo_subscription(sharedline, exten);
	if (!callinfo_sub || !callinfo_sub->sip_sub) {
		return -1;
	}

	ao2_ref(callinfo_sub, +1);
	if (ast_sip_push_task(callinfo_sub->serializer, notify_callinfo_terminated, callinfo_sub)) {
		ao2_cleanup(callinfo_sub);
		return -1;
	}

	return 0;
}

struct ast_sip_call_info_handler sca_callinfo_handler = {
	.call_info_terminated = call_info_terminated,
};

static int load_module(void)
{
	if (ast_sip_register_subscription_handler(&call_info_handler)) {
		ast_log(LOG_WARNING, "Unable to register subscription handler %s\n",
			call_info_handler.event_name);
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_sip_register_call_info_handler(&sca_callinfo_handler);

	call_info_state_sub = stasis_subscribe(ast_device_state_topic_all(), call_info_state_cb, NULL);
	if (!call_info_state_sub) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_subscription_handler(&call_info_handler);
    ast_sip_unregister_call_info_handler(&sca_callinfo_handler);

	call_info_state_sub = stasis_unsubscribe_and_join(call_info_state_sub);

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Call Info Notifications",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
		.requires = "res_pjsip,res_pjsip_pubsub",
);

