/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/channel.h"
#include "asterisk/module.h"
#include "asterisk/callerid.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/pbx.h"
#include "asterisk/astdb.h"
AST_MODULE_LOG("pjsip");
#endif

/* emergency elin node. */
struct emergency_elin {
	char *elin;
	int net_ver;
	struct in_addr addr;
	struct in6_addr addr6;
	unsigned int mask;
	char *location;
	char *loc_route;
	AST_LIST_ENTRY(emergency_elin) entry;
};
static AST_LIST_HEAD_NOLOCK(, emergency_elin) emergency_elin_list;

/*!
 * \internal
 * \brief Set an ast_party_id name and number based on an identity header.
 * \param hdr From, P-Asserted-Identity, or Remote-Party-ID header on incoming message
 * \param[out] id The ID to set data on
 */
static void set_id_from_hdr(pjsip_fromto_hdr *hdr, struct ast_party_id *id)
{
#ifdef GRANDSTREAM_NETWORKS
	char cid_name[AST_MAX_CID_NAME] = {0};
#else
	char cid_name[AST_CHANNEL_NAME];
#endif
	char cid_num[AST_CHANNEL_NAME];
	pjsip_sip_uri *uri;
	pjsip_name_addr *id_name_addr = (pjsip_name_addr *) hdr->uri;
	char *semi;

	uri = pjsip_uri_get_uri(id_name_addr);
	ast_copy_pj_str(cid_name, &id_name_addr->display, sizeof(cid_name));
	ast_copy_pj_str(cid_num, &uri->user, sizeof(cid_num));

	/* Always truncate caller-id number at a semicolon. */
	semi = strchr(cid_num, ';');
	if (semi) {
		/*
		 * We need to be able to handle URI's looking like
		 * "sip:1235557890;phone-context=national@x.x.x.x;user=phone"
		 *
		 * Where the uri->user field will result in:
		 * "1235557890;phone-context=national"
		 *
		 * People don't care about anything after the semicolon
		 * showing up on their displays even though the RFC
		 * allows the semicolon.
		 */
		*semi = '\0';
	}

	ast_free(id->name.str);
#ifdef GRANDSTREAM_NETWORKS
	ast_party_name_init(&id->name);
#endif

	id->name.str = ast_strdup(cid_name);
	if (!ast_strlen_zero(cid_name)) {
		id->name.valid = 1;
	}
	ast_free(id->number.str);
	id->number.str = ast_strdup(cid_num);
	if (!ast_strlen_zero(cid_num)) {
		id->number.valid = 1;
	}
}

/*!
 * \internal
 * \brief Get a P-Asserted-Identity or Remote-Party-ID header from an incoming message
 *
 * This function will parse the header as if it were a From header. This allows for us
 * to easily manipulate the URI, as well as add, modify, or remove parameters from the
 * header
 *
 * \param rdata The incoming message
 * \param header_name The name of the ID header to find
 * \retval NULL No ID header present or unable to parse ID header
 * \retval non-NULL The parsed ID header
 */
static pjsip_fromto_hdr *get_id_header(pjsip_rx_data *rdata, const pj_str_t *header_name)
{
	static const pj_str_t from = { "From", 4 };
	pj_str_t header_content;
	pjsip_fromto_hdr *parsed_hdr;
	pjsip_generic_string_hdr *ident = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg,
			header_name, NULL);
	int parsed_len;

	if (!ident) {
		return NULL;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &header_content, &ident->hvalue);

	parsed_hdr = pjsip_parse_hdr(rdata->tp_info.pool, &from, header_content.ptr,
			pj_strlen(&header_content), &parsed_len);

	if (!parsed_hdr) {
		return NULL;
	}

	return parsed_hdr;
}

/*!
 * \internal
 * \brief Set an ast_party_id structure based on data in a P-Asserted-Identity header
 *
 * This makes use of \ref set_id_from_hdr for setting name and number. It uses
 * the contents of a Privacy header in order to set presentation information.
 *
 * \param rdata The incoming message
 * \param[out] id The ID to set
 * \retval 0 Successfully set the party ID
 * \retval non-zero Could not set the party ID
 */
static int set_id_from_pai(pjsip_rx_data *rdata, struct ast_party_id *id)
{
	static const pj_str_t pai_str = { "P-Asserted-Identity", 19 };
	static const pj_str_t privacy_str = { "Privacy", 7 };
	pjsip_fromto_hdr *pai_hdr = get_id_header(rdata, &pai_str);
	pjsip_generic_string_hdr *privacy;

	if (!pai_hdr) {
		return -1;
	}

	set_id_from_hdr(pai_hdr, id);

	if (!id->number.valid) {
		return -1;
	}

	privacy = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &privacy_str, NULL);
	if (!privacy || !pj_stricmp2(&privacy->hvalue, "none")) {
		id->number.presentation = AST_PRES_ALLOWED_USER_NUMBER_NOT_SCREENED;
		id->name.presentation = AST_PRES_ALLOWED_USER_NUMBER_NOT_SCREENED;
	} else {
		id->number.presentation = AST_PRES_PROHIB_USER_NUMBER_NOT_SCREENED;
		id->name.presentation = AST_PRES_PROHIB_USER_NUMBER_NOT_SCREENED;
	}

	return 0;
}

/*!
 * \internal
 * \brief Set an ast_party_id structure based on data in a Remote-Party-ID header
 *
 * This makes use of \ref set_id_from_hdr for setting name and number. It uses
 * the privacy and screen parameters in order to set presentation information.
 *
 * \param rdata The incoming message
 * \param[out] id The ID to set
 * \retval 0 Succesfully set the party ID
 * \retval non-zero Could not set the party ID
 */
static int set_id_from_rpid(pjsip_rx_data *rdata, struct ast_party_id *id)
{
	static const pj_str_t rpid_str = { "Remote-Party-ID", 15 };
	static const pj_str_t privacy_str = { "privacy", 7 };
	static const pj_str_t screen_str = { "screen", 6 };
	pjsip_fromto_hdr *rpid_hdr = get_id_header(rdata, &rpid_str);
	pjsip_param *screen;
	pjsip_param *privacy;

	if (!rpid_hdr) {
		return -1;
	}

	set_id_from_hdr(rpid_hdr, id);

	if (!id->number.valid) {
		return -1;
	}

	privacy = pjsip_param_find(&rpid_hdr->other_param, &privacy_str);
	screen = pjsip_param_find(&rpid_hdr->other_param, &screen_str);
	if (privacy && !pj_stricmp2(&privacy->value, "full")) {
		id->number.presentation = AST_PRES_RESTRICTED;
		id->name.presentation = AST_PRES_RESTRICTED;
	} else {
		id->number.presentation = AST_PRES_ALLOWED;
		id->name.presentation = AST_PRES_ALLOWED;
	}
	if (screen && !pj_stricmp2(&screen->value, "yes")) {
		id->number.presentation |= AST_PRES_USER_NUMBER_PASSED_SCREEN;
		id->name.presentation |= AST_PRES_USER_NUMBER_PASSED_SCREEN;
	} else {
		id->number.presentation |= AST_PRES_USER_NUMBER_UNSCREENED;
		id->name.presentation |= AST_PRES_USER_NUMBER_UNSCREENED;
	}

	return 0;
}

/*!
 * \internal
 * \brief Set an ast_party_id structure based on data in a From
 *
 * This makes use of \ref set_id_from_hdr for setting name and number. It uses
 * no information from the message in order to set privacy. It relies on endpoint
 * configuration for privacy information.
 *
 * \param rdata The incoming message
 * \param[out] id The ID to set
 * \retval 0 Succesfully set the party ID
 * \retval non-zero Could not set the party ID
 */
static int set_id_from_from(struct pjsip_rx_data *rdata, struct ast_party_id *id)
{
	pjsip_fromto_hdr *from = pjsip_msg_find_hdr(rdata->msg_info.msg,
			PJSIP_H_FROM, rdata->msg_info.msg->hdr.next);

	if (!from) {
		/* This had better not happen */
		return -1;
	}

	set_id_from_hdr(from, id);

	if (!id->number.valid) {
		return -1;
	}

	return 0;
}

/*!
 * \internal
 * \brief Determine if a connected line update should be queued
 *
 * This uses information about the session and the ID that would be queued
 * in the connected line update in order to determine if we should queue
 * a connected line update.
 *
 * \param session The session whose channel we wish to queue the connected line update on
 * \param id The identification information that would be queued on the connected line update
 * \retval 0 We should not queue a connected line update
 * \retval non-zero We should queue a connected line update
 */
static int should_queue_connected_line_update(const struct ast_sip_session *session, const struct ast_party_id *id)
{
	/* Invalid number means no update */
	if (!id->number.valid) {
		return 0;
	}

	/* If the session has never communicated an update or if the
	 * new ID has a different number than the session, then we
	 * should queue an update
	 */
	if (ast_strlen_zero(session->id.number.str) ||
			strcmp(session->id.number.str, id->number.str)) {
		return 1;
	}

	/* By making it to this point, it means the number is not enough
	 * to determine if an update should be sent. Now we look at
	 * the name
	 */

	/* If the number couldn't warrant an update and the name is
	 * invalid, then no update
	 */
	if (!id->name.valid) {
		return 0;
	}

	/* If the name has changed or we don't have a name set for the
	 * session, then we should send an update
	 */
	if (ast_strlen_zero(session->id.name.str) ||
			strcmp(session->id.name.str, id->name.str)) {
		return 1;
	}

	/* Neither the name nor the number have changed. No update */
	return 0;
}

/*!
 * \internal
 * \brief Queue a connected line update on a session's channel.
 * \param session The session whose channel should have the connected line update queued upon.
 * \param id The identification information to place in the connected line update
 */
static void queue_connected_line_update(struct ast_sip_session *session, const struct ast_party_id *id)
{
	struct ast_party_connected_line connected;
	struct ast_party_caller caller;

	/* Fill connected line information */
	ast_party_connected_line_init(&connected);
	connected.id = *id;
	connected.id.tag = session->endpoint->id.self.tag;
#ifndef GRANDSTREAM_NETWORKS
	connected.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;
#else
	connected.source = AST_CONNECTED_LINE_UPDATE_SOURCE_RPID;
#endif

	/* Save to channel driver copy */
	ast_party_id_copy(&session->id, &connected.id);

	/* Update our channel CALLERID() */
	ast_party_caller_init(&caller);
	caller.id = connected.id;
	caller.ani = connected.id;
	caller.ani2 = ast_channel_caller(session->channel)->ani2;
	ast_channel_set_caller_event(session->channel, &caller, NULL);

	/* Tell peer about the new connected line information. */
	ast_channel_queue_connected_line_update(session->channel, &connected, NULL);
}

#ifdef GRANDSTREAM_NETWORKS
static int set_id_from_to(struct pjsip_rx_data *rdata, struct ast_party_id *id)
{
	pjsip_fromto_hdr *to = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_TO, rdata->msg_info.msg->hdr.next);

	if (!to) {
		/* This had better not happen */
		return -1;
	}

	set_id_from_hdr(to, id);

	if (!id->number.valid) {
		return -1;
	}

	return 0;
}

static void update_ringboth_connected_line(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session || !session->channel || !rdata) {
		return;
	}

	if ((rdata->msg_info.msg->type != PJSIP_RESPONSE_MSG) || (rdata->msg_info.msg->line.status.code != 200)) {
		return;
	}

	if (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		char *en_both = (char *)pbx_builtin_getvar_helper(session->channel, "EN_BOTH");
		if (!ast_strlen_zero(en_both) && !strncmp(en_both, "yes", 3)) {
			struct ast_party_id id;

			ast_party_id_init(&id);
			if (!set_id_from_pai(rdata, &id) || !set_id_from_rpid(rdata, &id) || !set_id_from_to(rdata, &id)) {
				if (should_queue_connected_line_update(session, &id)) {
					queue_connected_line_update(session, &id);
				}
			}
			ast_party_id_free(&id);
		}
	}
}
#endif

/*!
 * \internal
 * \brief Make updates to connected line information based on an incoming request.
 *
 * This will get identity information from an incoming request. Once the identification is
 * retrieved, we will check if the new information warrants a connected line update and queue
 * a connected line update if so.
 *
 * \param session The session on which we received an incoming request
 * \param rdata The incoming request
 */
static void update_incoming_connected_line(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	struct ast_party_id id;
#ifdef GRANDSTREAM_NETWORKS
	int updated = 0;
#endif

	if (!session->endpoint->trust_connected_line
		|| !session->endpoint->id.trust_inbound) {

#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug294748 */
		update_ringboth_connected_line(session, rdata);
#endif
		return;
	}

	ast_party_id_init(&id);
	if (!set_id_from_pai(rdata, &id) || !set_id_from_rpid(rdata, &id)) {
		if (should_queue_connected_line_update(session, &id)) {
			queue_connected_line_update(session, &id);
		}

#ifdef GRANDSTREAM_NETWORKS
		updated = 1;
#endif
	}
	ast_party_id_free(&id);

#ifdef GRANDSTREAM_NETWORKS
	if (!updated) {
		/* Added by yfyang, for Bug294748 */
		update_ringboth_connected_line(session, rdata);
	}
#endif
}

/*!
 * \internal
 * \brief Session supplement callback on an incoming INVITE request
 *
 * If we are receiving an initial INVITE, then we will set the session's identity
 * based on the INVITE or configured endpoint values. If we are receiving a reinvite,
 * then we will potentially queue a connected line update via the \ref update_incoming_connected_line
 * function
 *
 * \param session The session that has received an INVITE
 * \param rdata The incoming INVITE
 */
static int caller_id_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session->channel) {
#ifdef GRANDSTREAM_NETWORKS     /* Added by yfyang, for sca feature */
		if (session->endpoint->sca_enable && 
		    (session->endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_PRIVATE) &&
		    !ast_strlen_zero(session->endpoint->sca_sharedline)) {
			set_id_from_from(rdata, &session->id);

			/* private user use the shared line number */
			ast_free(session->id.number.str);
			session->id.number.str = ast_strdup(session->endpoint->sca_sharedline);
			return 0;
		}
#endif

		/*
		 * Since we have no channel this must be the initial inbound
		 * INVITE.  Set the session ID directly because the channel
		 * has not been created yet.
		 */
#ifndef GRANDSTREAM_NETWORKS     /* Modified by yfyang, for livechat PAI process */
		if (session->endpoint->id.trust_inbound
#else
		if ((session->endpoint->id.trust_inbound || !strncmp(ast_sorcery_object_get_id(session->endpoint), "webrtc_trunk_", 13))
#endif
			&& (!set_id_from_pai(rdata, &session->id)
				|| !set_id_from_rpid(rdata, &session->id))) {
			ast_free(session->id.tag);
			session->id.tag = ast_strdup(session->endpoint->id.self.tag);
			return 0;
		}
#ifndef GRANDSTREAM_NETWORKS     /* Modified by yfyang, for bug59445 */
		ast_party_id_copy(&session->id, &session->endpoint->id.self);
		if (!session->endpoint->id.self.number.valid) {
			set_id_from_from(rdata, &session->id);
		}
#else
		if (strncmp(ast_sorcery_object_get_id(session->endpoint), "trunk_", 6) &&
			strncmp(ast_sorcery_object_get_id(session->endpoint), "conf", 4)) {
			ast_party_id_copy(&session->id, &session->endpoint->id.self);
			if (!session->endpoint->id.self.number.valid) {
				set_id_from_from(rdata, &session->id);
			}
		} else {
			set_id_from_from(rdata, &session->id);
		}
#endif
	} else {
		/*
		 * ReINVITE or UPDATE.  Check for changes to the ID and queue
		 * a connected line update if necessary.
		 */
		update_incoming_connected_line(session, rdata);
	}
	return 0;
}

/*!
 * \internal
 * \brief Session supplement callback on INVITE response
 *
 * INVITE responses could result in queuing connected line updates.
 *
 * \param session The session on which communication is happening
 * \param rdata The incoming INVITE response
 */
static void caller_id_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session->channel) {
		return;
	}

	update_incoming_connected_line(session, rdata);
}

#ifdef GRANDSTREAM_NETWORKS
/* add by chdruan@grandsteam.cn for P-Preferred-Identity header Bug 30238*/
static void modify_contact_header(const struct ast_sip_session *session, pj_pool_t *pool, pjsip_contact_hdr *id_hdr, const struct ast_party_id *id)
{
	pjsip_name_addr *id_name_addr;
	pjsip_sip_uri *id_uri;

	id_name_addr = (pjsip_name_addr *) id_hdr->uri;
	id_uri = pjsip_uri_get_uri(id_name_addr->uri);

	if (id->name.valid && id->name.str) {
		int name_buf_len = strlen(id->name.str) * 2 + 1;
		char *name_buf = ast_alloca(name_buf_len);

		ast_escape_quoted(id->name.str, name_buf, name_buf_len);
		pj_strdup2(pool, &id_name_addr->display, name_buf);
	}

	if (id->number.valid) {
		pj_strdup2(pool, &id_uri->user, id->number.str);
	}

	/* Add by bfdu for Bug 298222, 2024.5.23 */
	if (session && session->inv_session && session->inv_session->dlg) {
		session->inv_session->dlg->local.contact = (pjsip_contact_hdr*)pjsip_hdr_clone(session->inv_session->dlg->pool, id_hdr);
		ast_log(LOG_DEBUG, "modify_contact_header update dlg local contact\n");
	}
}


/*!
 * \internal
 * \brief Set name and number information on an identity header.
 * \param pool Memory pool to use for string duplication
 * \param id_hdr A From, P-Asserted-Identity, or Remote-Party-ID header to modify
 * \param id The identity information to apply to the header
 */
static void modify_id_header(pj_pool_t *pool, pjsip_fromto_hdr *id_hdr, const struct ast_party_id *id)
{
	pjsip_name_addr *id_name_addr;
	pjsip_sip_uri *id_uri;

	id_name_addr = (pjsip_name_addr *) id_hdr->uri;
#ifdef GRANDSTREAM_NETWORKS
	/*! Begin: Added by pxie@grandstream.cn for bug[164443], 2020-05-07. */
	if (!id_name_addr) {
		return;
	}
	/*! Ended: Added by pxie@grandstream.cn for bug[164443], 2020-05-07. */
#endif

	id_uri = pjsip_uri_get_uri(id_name_addr->uri);

	if (id->name.valid && id->name.str) {
		int name_buf_len = strlen(id->name.str) * 2 + 1;
		char *name_buf = ast_alloca(name_buf_len);

		ast_escape_quoted(id->name.str, name_buf, name_buf_len);
		pj_strdup2(pool, &id_name_addr->display, name_buf);
	}
	else {
		pj_strdup2(pool, &id_name_addr->display, "");
	}

	if (id->number.valid) {
		pj_strdup2(pool, &id_uri->user, id->number.str);
	}
}

static int ipv6_prefix_is_equal(const struct in6_addr *a1, const struct in6_addr *a2, unsigned int prefixlen)
{
	unsigned pdw, pbi;
	pdw = prefixlen >> 5;
	if (pdw && memcmp(a1->s6_addr32, a2->s6_addr32, pdw << 2))
		return 0;

	/* check incomplete u32 in prefix */
	pbi = prefixlen & 0x1f;
	if (pbi && ((a1->s6_addr32[pdw] ^ a2->s6_addr32[pdw]) & htonl((0xffffffff) << (32 - pbi))))
		return 0;

	return 1;
}

/* Added by yfyang, for IPVT Server application */
static void add_ipvt_client_info_header(const struct ast_sip_session *session, pjsip_tx_data *tdata, int is_response)
{
	const char *client_info_msg = NULL;
	const char *client_info     = NULL;
	pjsip_generic_string_hdr  *hdr_old;
	pj_str_t clientinfo_name = { "X-IPVT-Client-Info", 18};
	char buf[512] = {0};
	int  add = 0;

	if (!session || !tdata) {
		return;
	}

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &clientinfo_name, NULL))) {
		return;
	}

	client_info_msg = pbx_builtin_getvar_helper(session->channel, "IPVT_CLIENT_INFO_MSG");
	if (ast_strlen_zero(client_info_msg)) {
		return;
	}

	if (!strncmp(client_info_msg, "response", 8)) {
		if (is_response){
			add = 1;
		} 
	} else if (!strncmp(client_info_msg, "request", 7)) {
		if (!is_response){
			add = 1;
		} 
	}

	if (!add) {
		return;
	}

	client_info = pbx_builtin_getvar_helper(session->channel, "IPVT_CLIENT_INFO");
	if (!ast_strlen_zero(client_info)) {
		snprintf(buf, sizeof(buf)-1, "%s", client_info);
	}

	ast_sip_add_header(tdata, "X-IPVT-Client-Info", buf);
	ast_log(LOG_NOTICE, "add_ipvt_client_info_header, buf[%s]", buf);
}

#ifdef BUILD_ON_CLOUD_PBX
static void add_trunk_host_header(const struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_generic_string_hdr  *hdr_old;
	pj_str_t tkhost_name = { "Trunk-Host", 10};
	char  addrs[128] = {0};

	if (!session || !tdata || !session->endpoint) {
		return;
	}

	if (ast_strlen_zero(session->endpoint->trunk_host)) {
		return;
	}

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &tkhost_name, NULL))) {
		return;
	}

	ast_sip_add_header(tdata, "Trunk-Host", session->endpoint->trunk_host);

	ast_sip_resolve_identifier_get_destaddrs((char *)session->endpoint->trunk_host, addrs, sizeof(addrs));

	if (strlen(addrs)) {
		ast_sip_add_header(tdata, "Trunk-IP", addrs);
	}
    return;
}
#endif
#endif

/*!
 * \internal
 * \brief Create an identity header for an outgoing message
 * \param hdr_name The name of the header to create
 * \param tdata The message to place the header on
 * \param id The identification information for the new header
 * \return newly-created header
 */
static pjsip_fromto_hdr *create_new_id_hdr(const pj_str_t *hdr_name, pjsip_fromto_hdr *base, pjsip_tx_data *tdata, const struct ast_party_id *id)
{
	pjsip_fromto_hdr *id_hdr;
	pjsip_name_addr *id_name_addr;
	pjsip_sip_uri *id_uri;

	id_hdr = pjsip_from_hdr_create(tdata->pool);
	id_hdr->type = PJSIP_H_OTHER;
	id_hdr->sname = id_hdr->name = *hdr_name;

	id_name_addr = pjsip_uri_clone(tdata->pool, base->uri);
	id_uri = pjsip_uri_get_uri(id_name_addr->uri);

	if (id->name.valid && !ast_strlen_zero(id->name.str)) {
		int name_buf_len = strlen(id->name.str) * 2 + 1;
		char *name_buf = ast_alloca(name_buf_len);

		ast_escape_quoted(id->name.str, name_buf, name_buf_len);
		pj_strdup2(tdata->pool, &id_name_addr->display, name_buf);
	} else {
		/*
		 * We need to clear the remnants of the clone or it'll be left set.
		 * pj_strdup2 is safe to call with a NULL src and it resets both slen and ptr.
		 */
		pj_strdup2(tdata->pool, &id_name_addr->display, NULL);
	}

	if (id->number.valid) {
		pj_strdup2(tdata->pool, &id_uri->user, id->number.str);
	} else {
		/* Similar to name, make sure the number is also cleared when invalid */
		pj_strdup2(tdata->pool, &id_uri->user, NULL);
	}

	id_hdr->uri = (pjsip_uri *) id_name_addr;
	return id_hdr;
}

/*!
 * \internal
 * \brief Add a Privacy header to an outbound message
 *
 * When sending a P-Asserted-Identity header, if privacy is requested, then we
 * will need to indicate such by adding a Privacy header. Similarly, if no
 * privacy is requested, and a Privacy header already exists on the message,
 * then the old Privacy header should be removed.
 *
 * \param tdata The outbound message to add the Privacy header to
 * \param id The id information used to determine privacy
 */
static void add_privacy_header(pjsip_tx_data *tdata, const struct ast_party_id *id)
{
	static const pj_str_t pj_privacy_name = { "Privacy", 7 };
#ifndef GRANDSTREAM_NETWORKS
	static const pj_str_t pj_privacy_value = { "id", 2 };
#else
	static const pj_str_t pj_privacy_allow = { "none", 4 };
	static const pj_str_t pj_privacy_block = { "user;id;critical", 16 };
	pjsip_generic_string_hdr *privacy_hdr = NULL;
#endif	
	pjsip_hdr *old_privacy;

	old_privacy = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_privacy_name, NULL);

#ifdef GRANDSTREAM_NETWORKS
	if (old_privacy) {
		pj_list_erase(old_privacy);
	}

	if ((ast_party_id_presentation(id) & AST_PRES_RESTRICTION) == AST_PRES_ALLOWED) {
		privacy_hdr = pjsip_generic_string_hdr_create(tdata->pool, &pj_privacy_name, &pj_privacy_allow);
	} else {
		privacy_hdr = pjsip_generic_string_hdr_create(tdata->pool, &pj_privacy_name, &pj_privacy_block);
	}

	if (privacy_hdr) {
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)privacy_hdr);
	}
#else

	if ((ast_party_id_presentation(id) & AST_PRES_RESTRICTION) == AST_PRES_ALLOWED) {
		if (old_privacy) {
			pj_list_erase(old_privacy);
		}

	} else if (!old_privacy) {
		pjsip_generic_string_hdr *privacy_hdr = pjsip_generic_string_hdr_create(
				tdata->pool, &pj_privacy_name, &pj_privacy_value);
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)privacy_hdr);
	}
#endif
}


#ifdef GRANDSTREAM_NETWORKS	
static void add_privacy_header_block(pjsip_tx_data *tdata)
{
	static const pj_str_t pj_privacy_name = { "Privacy", 7 };
	static const pj_str_t pj_privacy_block = { "user;id;critical", 16 };
	pjsip_hdr *old_privacy;

	old_privacy = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_privacy_name, NULL);
	if (old_privacy) {
		pj_list_erase(old_privacy);
	}

	pjsip_generic_string_hdr *privacy_hdr = pjsip_generic_string_hdr_create(
				tdata->pool, &pj_privacy_name, &pj_privacy_block);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)privacy_hdr);
}

static int check_inner_user(char *user)
{
    char family[80] = {0};
    char data[16]   = {0};

    if (!user) {
        return 0;
    }

    // AAUSER/9000/TYPE : user
    snprintf(family, sizeof(family), "AAUSER/%s", user);
    ast_ucm_db_get(family, "TYPE", data, sizeof(data));

    if (!strcmp(data, "user")) {
        return 1;
    }

    return 0;
}
#endif

/*!
 * \internal
 * \brief Add a P-Asserted-Identity header to an outbound message
 * \param tdata The message to add the header to
 * \param id The identification information used to populate the header
 */
static void add_pai_header(const struct ast_sip_session *session, pjsip_tx_data *tdata, const struct ast_party_id *id)
{
	static const pj_str_t pj_pai_name = { "P-Asserted-Identity", 19 };
	pjsip_fromto_hdr *base;
	pjsip_fromto_hdr *pai_hdr;
	pjsip_fromto_hdr *old_pai;

#ifdef GRANDSTREAM_NETWORKS
	if (0 == session->endpoint->id.send_pai) {
		goto add_header;
	}

	const char *from_user = session->endpoint ? session->endpoint->fromuser : NULL;
	const char *dod = pbx_builtin_getvar_helper(session->channel, "DOD");
	const char *pai_number = session->endpoint ? session->endpoint->pai_number : NULL;

	if (!pai_number && (!from_user || (dod && !strcmp(from_user, dod)))) {
		/* add by chdruan@grandsteam.cn for P-Asserted-Identity header Bug 25547*/
		pjsip_contact_hdr *contact;
		contact = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CONTACT, tdata->msg->hdr.next);
		if (contact) {
			modify_contact_header(session, tdata->pool, contact, id);
		}
	}

add_header:
#endif

	/* Since inv_session reuses responses, we have to make sure there's not already
	 * a P-Asserted-Identity present. If there is, we just modify the old one.
	 */
	old_pai = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_pai_name, NULL);
	if (old_pai) {
		/* If type is OTHER, then the existing header was most likely
		 * added by the PJSIP_HEADER dial plan function as a simple
		 * name/value pair.  We can't pass this to modify_id_header because
		 * there are no virtual functions to get the uri.  We could parse
		 * it into a pjsip_fromto_hdr but it isn't worth it since
		 * modify_id_header is just going to overwrite the name and number
		 * anyway.  We'll just remove it from the header list instead
		 * and create a new one.
		 */
		if (old_pai->type == PJSIP_H_OTHER) {
			pj_list_erase(old_pai);
		} else {
			ast_sip_modify_id_header(tdata->pool, old_pai, id);
			add_privacy_header(tdata, id);
			return;
		}
	}

	if (tdata->msg->type == PJSIP_REQUEST_MSG) {
		base = session->saved_from_hdr ? session->saved_from_hdr : PJSIP_MSG_FROM_HDR(tdata->msg);
	} else {
		base = PJSIP_MSG_TO_HDR(tdata->msg);
	}

	pai_hdr = create_new_id_hdr(&pj_pai_name, base, tdata, id);
	if (!pai_hdr) {
		return;
	}
	add_privacy_header(tdata, id);

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)pai_hdr);
}

#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, for missed call feature */
static int add_calltype_params_to_from(const struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_from_hdr *from_hdr = NULL;
	static const pj_str_t calltype_str = { "x-gs-call-type", 14};
	static const pj_str_t name_str     = { "x-gs-group-name", 15};
	static const pj_str_t cq_str = { "CQ", 2 };
	static const pj_str_t rg_str = { "RG", 2 };
	pj_str_t     name_value_str;
	pjsip_param *call_type = NULL;
	pjsip_param *call_name = NULL;
	pjsip_param *call_type1 = NULL;
	pjsip_param *call_name1 = NULL;
	int is_group_call = 0;
	int is_queue_call = 0;
	pjsip_dialog *dlg = NULL;

	if (!session || !session->channel || !tdata || !session->inv_session || !session->inv_session->dlg) {
		ast_log(LOG_WARNING, "TX packet is NULL\n");
		return -1;
	}

	from_hdr = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, NULL);
	if (!from_hdr) {
		return -1;
	}

	/*to check if From header already has the parameter */
	call_type = pjsip_param_find(&from_hdr->other_param, &calltype_str);
	if (call_type) {
		ast_log(LOG_NOTICE, "From header already has param: x-gs-call-type \n");
		return 0;
	}

	/* to get channel value: IN_RINGGROUP or IN_QUEUE */
	const char *in_group = pbx_builtin_getvar_helper(session->channel, "IN_RINGGROUP");
	if (!ast_strlen_zero(in_group) && !strncmp(in_group, "yes", 3)) {
		is_group_call = 1;
	} else {
		const char *in_queue = pbx_builtin_getvar_helper(session->channel, "IN_QUEUE");
		if (!ast_strlen_zero(in_queue) && !strncmp(in_queue, "yes", 3)) {
			is_queue_call = 1;
		}
	}

	if (!is_group_call && !is_queue_call) {
		ast_log(LOG_NOTICE, "This is not a CQ or RG call\n");
		return 0;
	}

	call_type = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	if (!call_type) {
		ast_log(LOG_WARNING, "PJ_POOL_ALLOC_T failed\n");
		return -1;
	}

	/* append parameter */
	call_type->name = calltype_str;
	if (is_group_call) {
		call_type->value = rg_str;

		const char *group_name = pbx_builtin_getvar_helper(session->channel, "GROUP_NAME");
		if (!ast_strlen_zero(group_name)) {
			call_name = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
			if (call_name) {
				pj_strdup2_with_null(tdata->pool, &name_value_str, group_name);
				call_name->value = name_value_str; 
			}
		}
	} else if (is_queue_call) {
		call_type->value = cq_str;
		const char *queue_name = pbx_builtin_getvar_helper(session->channel, "QUEUE_NAME");
		if (!ast_strlen_zero(queue_name)) {
			call_name = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
			if (call_name) {
				pj_strdup2_with_null(tdata->pool, &name_value_str, queue_name);
				call_name->value = name_value_str; 
			}
		}
	}

	dlg = session->inv_session->dlg;
	call_type1 = PJ_POOL_ALLOC_T(dlg->pool, pjsip_param);
	if (call_type1) {
		pj_strdup_with_null(dlg->pool, &call_type1->name, &call_type->name);
		pj_strdup_with_null(dlg->pool, &call_type1->value, &call_type->value);
		pj_list_push_back(&dlg->local.info->other_param, call_type1);
	}

	pj_list_push_back(&from_hdr->other_param, call_type);

	if (call_name) {
		call_name->name = name_str;
		pj_list_push_back(&from_hdr->other_param, call_name);

		call_name1 = PJ_POOL_ALLOC_T(dlg->pool, pjsip_param);
		if (call_name1) {
			pj_strdup_with_null(dlg->pool, &call_name1->name, &call_name->name);
			pj_strdup_with_null(dlg->pool, &call_name1->value, &call_name->value);
			pj_list_push_back(&dlg->local.info->other_param, call_name1);
		}
	}

	return 0; 
}

/* add by chdruan@grandsteam.cn for P-Preferred-Identity header Bug 30238*/
/*!
 * \internal
 * \brief Add a P-Preferred-Identity header to an outbound message
 * \param tdata The message to add the header to
 * \param id The identification information used to populate the header
 */
static void add_ppi_header(const struct ast_sip_session *session, pjsip_tx_data *tdata, const struct ast_party_id *id)
{
	static const pj_str_t pj_ppi_name = { "P-Preferred-Identity", 20 };
	pjsip_fromto_hdr  *ppi_hdr;
	pjsip_fromto_hdr  *old_ppi;
	pjsip_fromto_hdr  *base;
	pjsip_contact_hdr *contact;
	struct ast_party_id connected_id;
#ifdef GRANDSTREAM_NETWORKS
	struct ast_party_id connected_id1;
	char cid_name[AST_MAX_CID_NAME] = {0};
#else
	char cid_name[80] = {0};
#endif
	char cid_num[80]  = {0};
	int  set_name = 1;
	const char *cid_str = NULL;

	if (!id->number.valid) {
		return;
	}

	ast_party_id_init(&connected_id);
	ast_party_id_copy(&connected_id, id);

	/* added for bug51673, bug 210894 */
	char *dod = (char *)pbx_builtin_getvar_helper(session->channel, "OUTDOD_CID");
	if (ast_strlen_zero(dod)) {
		dod = (char *)pbx_builtin_getvar_helper(session->channel, "DOD");
	}

	const char *orig_cid = pbx_builtin_getvar_helper(session->channel, "tmp_peer_fullcallid");  /* added for bug81134 */
	const char *from_tk = pbx_builtin_getvar_helper(session->channel, "FROM_TRUNK");            /* added for bug182507 */

	if (ast_strlen_zero(orig_cid)) {
		orig_cid = pbx_builtin_getvar_helper(session->channel, "FIRST_CALLERID");  
	}

	if (session->endpoint->id.use_dod_in_ppi && dod && strlen(dod)) {
		cid_str = dod;
	} else if (session->endpoint->id.use_origcid_in_ppi && !ast_strlen_zero(orig_cid)) {
		if (!ast_strlen_zero(from_tk) && !strncmp(from_tk, "yes", 3)) {
			cid_str = orig_cid;
		}
	}

	if (cid_str) {
		ast_free(connected_id.number.str);

		/* Modified by yfyang, Bug 117298 */
		ast_free(connected_id.name.str);
		connected_id.name.str = NULL;
		connected_id.name.valid = 0;

		set_name = 0;

		/* dod without name */
		if (!strchr(cid_str, '<')) {
			connected_id.number.str = ast_strdup(cid_str); 
		} else { /* with name, split name and num */
			ast_callerid_split(cid_str, cid_name, sizeof(cid_name), cid_num, sizeof(cid_num)); 
			if (!ast_strlen_zero(cid_name)) {
				connected_id.name.str = ast_strdup(cid_name);
				connected_id.name.valid = 1;
			}

			if (!ast_strlen_zero(cid_num)) {
				connected_id.number.str = ast_strdup(cid_num); 
				connected_id.number.valid = 1;
			}
		}
	} else {
		if (!ast_strlen_zero(session->endpoint->username) && !strcmp(session->endpoint->trunk_type, "register")) {
			ast_free(connected_id.number.str);
			connected_id.number.str = ast_strdup(session->endpoint->username);
		}
	}

	if (set_name) {
		ast_free(connected_id.name.str); 
		connected_id.name.str = NULL;
		connected_id.name.valid = 0;

		if (session->endpoint->id.self.name.valid && session->endpoint->id.self.name.str) {
			connected_id.name.str = ast_strdup(session->endpoint->id.self.name.str);
			connected_id.name.valid = 1;
		}
	}

	contact = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CONTACT, tdata->msg->hdr.next);
	if (contact) {
        if (dod && (!ast_strlen_zero(session->endpoint->username) && session->endpoint->id.use_dod_as_contact_user)) {
			ast_party_id_init(&connected_id1);

			ast_callerid_split(dod, cid_name, sizeof(cid_name), cid_num, sizeof(cid_num)); 
			if (!ast_strlen_zero(cid_name)) {
				connected_id1.name.str = ast_strdup(cid_name);
				connected_id1.name.valid = 1;
			}

			if (!ast_strlen_zero(cid_num)) {
				connected_id1.number.str = ast_strdup(cid_num); 
				connected_id1.number.valid = 1;
			}
			modify_contact_header(session, tdata->pool, contact, &connected_id1);
			ast_party_id_free(&connected_id1);
		} else {
			modify_contact_header(session, tdata->pool, contact, &connected_id);
		}
	}

	/* Since inv_session reuses responses, we have to make sure there's not already
	 * a P-Preferred-Identity present. If there is, we just modify the old one.
	 */
	old_ppi = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_ppi_name, NULL);
	if (old_ppi) {
        pj_list_erase(old_ppi);
	}

	if (tdata->msg->type == PJSIP_REQUEST_MSG) {
		base = session->saved_from_hdr ? session->saved_from_hdr : PJSIP_MSG_FROM_HDR(tdata->msg);
	} else {
		base = PJSIP_MSG_TO_HDR(tdata->msg);
	}

	ppi_hdr = create_new_id_hdr(&pj_ppi_name, base, tdata, &connected_id);
	if (!ppi_hdr) {
		ast_party_id_free(&connected_id);
		return;
	}

	add_privacy_header(tdata, &connected_id);

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)ppi_hdr);
	ast_party_id_free(&connected_id);

}

/* Added by yfyang, for Broadsoft Test */
static void add_peer_pai_info_header(const struct ast_sip_session *session, pjsip_tx_data *tdata, const char* peer_info)
{
	struct ast_sip_pai_payload  payload;
	struct ast_party_id         pai_id;
	pjsip_generic_string_hdr   *hdr_old = NULL;
	static const pj_str_t pj_pai_name = { "P-Asserted-Identity", 19 };
	char* tmp   = NULL;
	int   pos   = 0;
	char  num[256]  = {0};
	char  name[256] = {0};
	char* info  = NULL;

	if (!session || !tdata || !peer_info) {
		return;
	}

	info = ast_strdupa(peer_info);
	ast_log(LOG_NOTICE, "peer_info[%s]", info);
	memset(&payload, 0, sizeof(payload));

	while ((tmp = strsep(&info, ","))) {
		switch (pos) {
		case 0:
			payload.is_blocked = atoi(tmp);
			break;
		case 1:
			payload.is_response = atoi(tmp);
			break;
		case 2:
			payload.code = atoi(tmp);
			break;
		case 3:
			strncpy(payload.method, tmp, sizeof(payload.method)-1);
			break;
		case 4:
			strncpy(payload.pai_string, tmp, sizeof(payload.pai_string) - 1);
			break;
		}

		pos++;
	}

	if (!strlen(payload.pai_string)) {
		ast_log(LOG_NOTICE, "add_peer_pai_info_header, NO pai_string, do nothing!");
		return;
	}

	/* check the payload value */
	if (tdata->msg->type == PJSIP_REQUEST_MSG) {
		if (payload.is_response) {
			return;
		}

		if (tdata->msg->line.req.method.id == PJSIP_INVITE_METHOD) {
			if (strcmp(payload.method, "INVITE")) {
				return;
			}
		} else if (tdata->msg->line.req.method.id == PJSIP_OTHER_METHOD) {
			if (strncmp(tdata->msg->line.req.method.name.ptr, "UPDATE", 6)) {
				return;
			}
		}
	} else {
		if (!payload.code) {
			return;
		}

		if (tdata->msg->line.status.code != payload.code) {
			return;
		}
	}

	ast_party_id_init(&pai_id);
	ast_callerid_split(payload.pai_string, name, sizeof(name), num, sizeof(num));
	if (strlen(name)) {
		pai_id.name.str = ast_strdup(name); 
		pai_id.name.valid = 1;
		if (payload.is_blocked) {
			pai_id.name.presentation = AST_PRES_RESTRICTED;
		}
	}

	if (strlen(num)) {
		pai_id.number.str = ast_strdup(num); 
		pai_id.number.valid = 1;
		if (payload.is_blocked) {
			pai_id.number.presentation = AST_PRES_RESTRICTED;
		}
	}

	pj_lock_acquire(tdata->lock);
	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_pai_name, NULL))) {
		pj_list_erase(hdr_old);
	}

	add_pai_header(session, tdata, &pai_id);
	pj_lock_release(tdata->lock);

	ast_party_id_free(&pai_id);

	return;
}

#endif

/*!
 * \internal
 * \brief Add party parameter to a Remote-Party-ID header.
 *
 * \param tdata The message where the Remote-Party-ID header is
 * \param hdr The header on which the parameters are being added
 * \param session The session involved
 */
static void add_party_param(pjsip_tx_data *tdata, pjsip_fromto_hdr *hdr, const struct ast_sip_session *session)
{
	static const pj_str_t party_str = { "party", 5 };
	static const pj_str_t calling_str = { "calling", 7 };
	static const pj_str_t called_str = { "called", 6 };
	pjsip_param *party;

	/* The party value can't change throughout the lifetime, so it is set only once */
	party = pjsip_param_find(&hdr->other_param, &party_str);
	if (party) {
		return;
	}

	party = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	party->name = party_str;
	party->value = (session->inv_session->role == PJSIP_ROLE_UAC) ? calling_str : called_str;
	pj_list_insert_before(&hdr->other_param, party);
}

/*!
 * \internal
 * \brief Add privacy and screen parameters to a Remote-Party-ID header.
 *
 * If privacy is requested, then the privacy and screen parameters need to
 * reflect this. Similarly, if no privacy or screening is to be communicated,
 * we need to make sure that any previously set values are updated.
 *
 * \param tdata The message where the Remote-Party-ID header is
 * \param hdr The header on which the parameters are being added
 * \param id The identification information used to determine privacy
 */
static void add_privacy_params(pjsip_tx_data *tdata, pjsip_fromto_hdr *hdr, const struct ast_party_id *id)
{
	static const pj_str_t privacy_str = { "privacy", 7 };
	static const pj_str_t screen_str = { "screen", 6 };
	static const pj_str_t privacy_full_str = { "full", 4 };
	static const pj_str_t privacy_off_str = { "off", 3 };
	static const pj_str_t screen_yes_str = { "yes", 3 };
	static const pj_str_t screen_no_str = { "no", 2 };
	pjsip_param *old_privacy;
	pjsip_param *old_screen;
	pjsip_param *privacy;
	pjsip_param *screen;
	int presentation;

	old_privacy = pjsip_param_find(&hdr->other_param, &privacy_str);
	old_screen = pjsip_param_find(&hdr->other_param, &screen_str);

	if (!old_privacy) {
		privacy = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
		privacy->name = privacy_str;
		pj_list_insert_before(&hdr->other_param, privacy);
	} else {
		privacy = old_privacy;
	}

	if (!old_screen) {
		screen = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
		screen->name = screen_str;
		pj_list_insert_before(&hdr->other_param, screen);
	} else {
		screen = old_screen;
	}

	presentation = ast_party_id_presentation(id);
	if ((presentation & AST_PRES_RESTRICTION) == AST_PRES_ALLOWED) {
		privacy->value = privacy_off_str;
	} else {
		privacy->value = privacy_full_str;
	}
	if ((presentation & AST_PRES_NUMBER_TYPE) == AST_PRES_USER_NUMBER_PASSED_SCREEN) {
		screen->value = screen_yes_str;
	} else {
		screen->value = screen_no_str;
	}
}

/*!
 * \internal
 * \brief Add a Remote-Party-ID header to an outbound message
 * \param tdata The message to add the header to
 * \param id The identification information used to populate the header
 */
static void add_rpid_header(const struct ast_sip_session *session, pjsip_tx_data *tdata, const struct ast_party_id *id)
{
	static const pj_str_t pj_rpid_name = { "Remote-Party-ID", 15 };
	pjsip_fromto_hdr *base;
	pjsip_fromto_hdr *rpid_hdr;
	pjsip_fromto_hdr *old_rpid;

#ifdef GRANDSTREAM_NETWORKS
	struct ast_party_id tmp_id;
	ast_party_id_init(&tmp_id);
	ast_party_id_copy(&tmp_id, id);

	const char *anonymous = pbx_builtin_getvar_helper(session->channel, "ANONYMOUS");
	if (session->endpoint->id.send_anonymous || (anonymous && !strcasecmp(anonymous, "anonymous"))) {
		if (tmp_id.name.valid) {
			tmp_id.name.presentation = AST_PRES_RESTRICTED;
		}

		tmp_id.number.presentation = AST_PRES_RESTRICTED | AST_PRES_USER_NUMBER_PASSED_SCREEN;
	}
#endif

	/* Since inv_session reuses responses, we have to make sure there's not already
	 * a P-Asserted-Identity present. If there is, we just modify the old one.
	 */
	old_rpid = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_rpid_name, NULL);
	if (old_rpid) {
		/* If type is OTHER, then the existing header was most likely
		 * added by the PJSIP_HEADER dial plan function as a simple
		 * name/value pair.  We can't pass this to modify_id_header because
		 * there are no virtual functions to get the uri.  We could parse
		 * it into a pjsip_fromto_hdr but it isn't worth it since
		 * modify_id_header is just going to overwrite the name and number
		 * anyway.  We'll just remove it from the header list instead
		 * and create a new one.
		 */
		if (old_rpid->type == PJSIP_H_OTHER) {
			pj_list_erase(old_rpid);
		} else {
#ifdef GRANDSTREAM_NETWORKS
			ast_sip_modify_id_header(tdata->pool, old_rpid, &tmp_id);
			add_party_param(tdata, old_rpid, session);
			add_privacy_params(tdata, old_rpid, &tmp_id);
			ast_party_id_free(&tmp_id);
#else
			ast_sip_modify_id_header(tdata->pool, old_rpid, id);
			add_party_param(tdata, old_rpid, session);
			add_privacy_params(tdata, old_rpid, id);
#endif
			return;
		}
	}

	if (tdata->msg->type == PJSIP_REQUEST_MSG) {
		base = session->saved_from_hdr ? session->saved_from_hdr : PJSIP_MSG_FROM_HDR(tdata->msg);
	} else {
		base = PJSIP_MSG_TO_HDR(tdata->msg);
	}

#ifdef GRANDSTREAM_NETWORKS
	rpid_hdr = create_new_id_hdr(&pj_rpid_name, base, tdata, &tmp_id);
	if (!rpid_hdr) {
		ast_party_id_free(&tmp_id);
		return;
	}

	/*start added by for Bug 218934,2022.3.24*/
	if(ast_channel_get_is_callflip(session->channel) && anonymous && !strcasecmp(anonymous, "anonymous")){
		pjsip_name_addr *id_name_addr;
		pjsip_sip_uri *id_uri;
		id_name_addr = (pjsip_name_addr *) rpid_hdr->uri;
		id_uri = pjsip_uri_get_uri(id_name_addr->uri);
		pj_strdup2(tdata->pool, &id_name_addr->display, "Anonymous");
		pj_strdup2(tdata->pool, &id_uri->user, "anonymous");
	}
	/*end added by for Bug 218934,2022.3.24*/

	add_party_param(tdata, rpid_hdr, session);
	add_privacy_params(tdata, rpid_hdr, &tmp_id);
	ast_party_id_free(&tmp_id);
#else
	rpid_hdr = create_new_id_hdr(&pj_rpid_name, base, tdata, id);
	if (!rpid_hdr) {
		return;
	}
	add_party_param(tdata, rpid_hdr, session);
	add_privacy_params(tdata, rpid_hdr, id);
#endif
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)rpid_hdr);
}

#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, Bug 103782 */
static void add_alertinfo_header(struct ast_sip_session *session, pjsip_tx_data *tdata, const char *alertinfo)
{
	static const pj_str_t alertinfo_name = { "Alert-Info", 10 };
	pjsip_generic_string_hdr  *hdr_old = NULL;
	pjsip_generic_string_hdr  *hdr_new = NULL;
	pj_str_t  alertinfo_value;
	char alert_str[256] = {0};

	if (!tdata || !alertinfo) {
		return;
	}

	hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &alertinfo_name, NULL); 
	if (hdr_old) {
		if (strstr(hdr_old->hvalue.ptr, "info=")) {
			return;
		}
	}

	const char *alert_schema = pbx_builtin_getvar_helper(session->channel, "DID_ALERT_SCHEMA"); 
	if (ast_strlen_zero(alert_schema)) {
        return;
	} 

	if (!strncmp(alert_schema, "uri", 3)) {
        snprintf(alert_str, sizeof(alert_str) - 1, "<https://127.0.0.1/%s>", alertinfo);
	} else if (!strncmp(alert_schema, "info", 4)) {
		snprintf(alert_str, sizeof(alert_str) - 1, "<https://127.0.0.1>;info=%s", alertinfo);
	} else {
		const char *alert_rt = pbx_builtin_getvar_helper(session->channel, "DID_ALERT_RINGTONE"); 
		if (ast_strlen_zero(alert_rt)) {
			snprintf(alert_str, sizeof(alert_str) - 1, "<https://127.0.0.1>;info=%s", alertinfo);
		} else {
			snprintf(alert_str, sizeof(alert_str) - 1, "<https://127.0.0.1/ringtones/%s>;info=%s", alert_rt, alertinfo);
		}
    }

	pj_strdup2_with_null(tdata->pool, &alertinfo_value, alert_str);
	hdr_new = pjsip_generic_string_hdr_create(tdata->pool, &alertinfo_name, &alertinfo_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_new);
}

static void replace_alertinfo_header(pjsip_tx_data* tdata, const char* alert_info)
{
	if (!tdata || !alert_info) {
		return;
	}

	char alert_str[256] = {0};
	pj_str_t alertinfo_value;
	const pj_str_t ALERT_INFO_STR = {"Alert-Info", 10};
	pjsip_generic_string_hdr* hdr_new = NULL;
	pjsip_generic_string_hdr* hdr_old = NULL;

	pj_lock_acquire(tdata->lock);
	while ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &ALERT_INFO_STR, NULL))) {
		pj_list_erase(hdr_old);
	}
	pj_lock_release(tdata->lock);

	snprintf(alert_str, sizeof(alert_str) - 1, "<http://127.0.0.1>;info=%s", alert_info);

	pj_strdup2_with_null(tdata->pool, &alertinfo_value, alert_str);

	hdr_new = pjsip_generic_string_hdr_create(tdata->pool, &ALERT_INFO_STR, &alertinfo_value);
	if (hdr_new) {
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_new);
	}
}

static void add_callinfo_header(struct ast_sip_session *session, pjsip_tx_data *tdata, int auto_answer, int chan_val, int sca)
{
	pjsip_generic_string_hdr  *hdr_old;
	pjsip_generic_string_hdr  *hdr_new;
	pjsip_fromto_hdr *from_hdr;
	pjsip_name_addr  *id_name_addr = NULL;
	pjsip_fromto_hdr *hdr;
	pjsip_fromto_hdr *hdr2;
	pjsip_param      *param_p;
	pj_str_t callinfo_name = { "Call-Info", 9 };
	pj_str_t callinfo_value;
	char  *call_info_str    = NULL;
	char   aa_str[256] = {0};
	char   ss_str[256] = {0};
	char   buf[512] = {0};
	int    len = 0;

	if (!session || !tdata) {
		return;
	}

	if (!auto_answer && !chan_val && !sca) {
		return;
	}

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &callinfo_name, NULL))) {
		return;
	}

	if (chan_val) {
		call_info_str = (char *)pbx_builtin_getvar_helper(session->channel, "CALL_INFO");
	}

	if (auto_answer) {
		from_hdr = PJSIP_MSG_FROM_HDR(tdata->msg); 
		id_name_addr = pjsip_uri_clone(tdata->pool, from_hdr->uri);

		/* create hdr */
		hdr = pjsip_from_hdr_create(tdata->pool);
		hdr->type = PJSIP_H_OTHER;
		pj_strdup(tdata->pool, &hdr->name, &callinfo_name);
		hdr->sname = hdr->name;
		hdr->uri = (pjsip_uri *)id_name_addr;

		param_p = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
		param_p->name = pj_str("answer-after");
		param_p->value = pj_str("0");
		pj_list_insert_before(&hdr->other_param, param_p);

		if (pjsip_hdr_print_on((pjsip_hdr*)hdr, aa_str, sizeof(aa_str)-1) < 0) {
			ast_log(LOG_NOTICE, "add_callinfo_header, print failed\n");
		}
	}

	/* Call-Info: <sip:as.iop1.broadworks.net>;appearance-index=1 */
	/* For bug 128208, when auto answer, don't add line index */
	if (sca && session->line_index && !auto_answer) {
		char sca_id[4] = {0};
        
		from_hdr = PJSIP_MSG_FROM_HDR(tdata->msg); 
		id_name_addr = pjsip_uri_clone(tdata->pool, from_hdr->uri);

		/* create hdr */
		hdr2 = pjsip_from_hdr_create(tdata->pool);
		hdr2->type = PJSIP_H_OTHER;
		pj_strdup(tdata->pool, &hdr2->name, &callinfo_name);
		hdr2->sname = hdr2->name; 
		if (id_name_addr) {
			hdr2->uri = (pjsip_uri *)id_name_addr; 
		} else {
			from_hdr = PJSIP_MSG_FROM_HDR(tdata->msg); 
			id_name_addr = pjsip_uri_clone(tdata->pool, from_hdr->uri);
			hdr2->uri = (pjsip_uri *)id_name_addr; 
		}

		param_p = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
		param_p->name = pj_str("appearance-index");
		snprintf(sca_id, sizeof(sca_id)-1, "%d", session->line_index);
		param_p->value = pj_str(sca_id);
		pj_list_insert_before(&hdr2->other_param, param_p);

		if (pjsip_hdr_print_on((pjsip_hdr*)hdr2, ss_str, sizeof(ss_str)-1) < 0) {
			ast_log(LOG_NOTICE, "add_callinfo_header, sca print failed\n");
		}
	}

	/* generate Call-Info value */
	if (!ast_strlen_zero(call_info_str)) {
		snprintf(buf, sizeof(buf)-1, "%s", call_info_str); 
	}

	if (strlen(aa_str)) {
		len = strlen(buf);
		if (len) {
			snprintf(buf + len, sizeof(buf) - len - 1, ",%s", aa_str + 11);   /* add 11 to removed the header name: Call-Info: */
		} else {
			snprintf(buf, sizeof(buf) - 1, "%s", aa_str + 11); 
		}
	} 

	if (strlen(ss_str)) {
		len = strlen(buf);
		if (len) {
			snprintf(buf + len, sizeof(buf) - len - 1, ",%s", ss_str + 11);   /* add 11 to removed the header name: Call-Info: */
		} else {
			snprintf(buf, sizeof(buf) - 1, "%s", ss_str + 11); 
		}
	} 

	if (!strlen(buf)) {
		return;
	}

	ast_log(LOG_NOTICE, "add_callinfo_header, [%s]", buf);

	pj_lock_acquire(tdata->lock);
	/* create hdr and insert to tdata */
	pj_strdup2_with_null(tdata->pool, &callinfo_value, buf);
	hdr_new = pjsip_generic_string_hdr_create(tdata->pool, &callinfo_name, &callinfo_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_new);
	pj_lock_release(tdata->lock);

	return;
}

static void add_pai_headers_for_anonymous_call(const struct ast_sip_session *session, pjsip_tx_data *tdata, 
											   const struct ast_party_id *id, int pai_added)
{
	if (!session || !session->endpoint || (id && !id->number.valid)
		|| (!session->endpoint->id.trust_outbound && (ast_party_id_presentation(id) & AST_PRES_RESTRICTION) != AST_PRES_ALLOWED))
	{
		return;
	}

	if ( session && session->endpoint && session->endpoint->aors ) {
		if (session->channel) {
			const char *ppi_info = pbx_builtin_getvar_helper(session->channel, "PPI_INFO");
			const char *anonymous = pbx_builtin_getvar_helper(session->channel, "ANONYMOUS");
			char caller_name[128] = {0};
			char caller_number[128] = {0};
			char format[128] = {0};
			pjsip_fromto_hdr *from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, tdata->msg->hdr.next);
			if (session->endpoint->id.send_anonymous || (anonymous && !strcasecmp(anonymous, "anonymous"))) {
				if (!pai_added) {
					/* add pai header */
					struct ast_party_id tmp_id;

					ast_party_id_init(&tmp_id);
					ast_party_id_copy(&tmp_id, id);
					tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_RESTRICTED;

					if (!ast_strlen_zero(ppi_info)) {
						/* get caller number and caller name from PPI header */
						snprintf(format, sizeof(format) - 1, "\"%%%ld[^\"]\" <sip:%%%ld[^@]@%%*s>", sizeof(caller_name) - 1, sizeof(caller_number) - 1);
						sscanf(ppi_info, format, caller_name, caller_number);
						if (ast_strlen_zero(caller_name)) {
							memset(format, 0, sizeof(format));
							memset(caller_number, 0, sizeof(caller_number));
							snprintf(format, sizeof(format) - 1, "<sip:%%%ld[^@]@%%*s>", sizeof(caller_number) - 1);
							sscanf(ppi_info, format, caller_number);
						}

						if (!ast_strlen_zero(caller_number)) {
							ast_free(tmp_id.name.str);
							if (!ast_strlen_zero(caller_name)) {
								tmp_id.name.str = ast_strdup(caller_name);
								tmp_id.name.valid = 1;
							} else {
								tmp_id.name.valid = 0;
							}

							ast_free(tmp_id.number.str);
							tmp_id.number.str = ast_strdup(caller_number);
						}
					}
					add_pai_header(session, tdata, &tmp_id);
					ast_party_id_free(&tmp_id);
				} else {
					add_privacy_header_block(tdata);
				}

				/* modify from header */
				if (from) {
					pjsip_name_addr *id_name_addr;
					pjsip_sip_uri *id_uri;
					id_name_addr = (pjsip_name_addr *) from->uri;
					if (id_name_addr) {
						id_uri = pjsip_uri_get_uri(id_name_addr->uri);
						pj_strdup2(tdata->pool, &id_name_addr->display, "Anonymous");
						if (id_uri) {
							pj_strdup2(tdata->pool, &id_uri->host, "anonymous.invalid");
							pj_strdup2(tdata->pool, &id_uri->user, "anonymous");
						}
					}
				}
			}
		}
	}
	return;
}

static void modify_from_user_for_conference(const struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_fromto_hdr *from = NULL;
	pjsip_sip_uri *uri = NULL;
	const char *conf_room = NULL;

	if (!session || !session->channel || !tdata || !tdata->msg) {
		ast_log(LOG_WARNING, "modify_from_user_for_conference param is NULL\n");
		return;
	}

	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_NOTICE, "[session:%p] modify_from_user_for_conference endpoint:%s is trunk, don't modify from user\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if ((tdata->msg->type == PJSIP_REQUEST_MSG) && (tdata->msg->line.req.method.id == PJSIP_INVITE_METHOD)) {
		from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, tdata->msg->hdr.next);
		if (!from) {
			ast_log(LOG_WARNING, "modify_from_user_for_conference from header is NULL\n");
			return;
		}
		if (PJSIP_URI_SCHEME_IS_SIP(from->uri) || PJSIP_URI_SCHEME_IS_SIPS(from->uri)) {
			uri = pjsip_uri_get_uri(from->uri);
			if (!uri) {
				ast_log(LOG_WARNING, "modify_from_user_for_conference from uri is NULL\n");
				return;
			}
			conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
			if (!ast_strlen_zero(conf_room)) {
				ast_log(LOG_DEBUG, "It's conf invite, set from user to %s\n", conf_room);
				pj_strdup2(tdata->pool, &uri->user, conf_room);
			}
		}
	}

	return;
}

/* Added by yfyang, for bug 155444 */
static void set_rpid_headers(const struct ast_sip_session *session, pjsip_tx_data *tdata, const struct ast_party_id *id)
{
	struct ast_party_id tmp_id;
	char *anonymous = NULL;
	char *reg_calleenum = NULL;
	char *cf_info = NULL;
	char *sip_tsf = NULL;
	char *from_park = NULL;
	char *in_park = NULL;

	if (!session || !tdata || !id) {
		return;
	}

	if (!session->channel) {
		add_rpid_header(session, tdata, id);
		return;
	}

	ast_party_id_init(&tmp_id);
	anonymous = (char *)pbx_builtin_getvar_helper(session->channel, "ANONYMOUS");
	if (anonymous && !strcasecmp(anonymous, "anonymous")) {
		tmp_id.number.str = ast_strdup(anonymous);
		tmp_id.number.valid = 1;
		tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_RESTRICTED;
		add_rpid_header(session, tdata, &tmp_id);

		ast_party_id_free(&tmp_id);
		return;
	}

	reg_calleenum = (char *)ast_channel_reg_calleenum(session->channel);
	cf_info = (char *)pbx_builtin_getvar_helper(session->channel, "CF_ALL");
	sip_tsf = (char *)pbx_builtin_getvar_helper(session->channel, "SIP_TRANSFER");
	from_park = (char *)pbx_builtin_getvar_helper(session->channel, "FROM_PARK");
	in_park = (char *)pbx_builtin_getvar_helper(session->channel, "IN_CALLPARK");

	ast_log(LOG_NOTICE, "set_rpid_headers, chan[%s], calleenum[%s], cf_info[%s], sip_tsf[%s], from_park[%s]", ast_channel_name(session->channel),
						reg_calleenum?reg_calleenum:"nil", cf_info?cf_info:"nil", sip_tsf?sip_tsf:"nil", ast_strlen_zero(from_park)? "no":from_park);

	/* Added by yfyang, for bug 180666 */
	if (ast_strlen_zero(cf_info) && ast_strlen_zero(sip_tsf) && ast_strlen_zero(from_park) && ast_strlen_zero(in_park)) {
		/* For Bug 226167 */
		if (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
			if (!check_inner_user(id->number.str)) {
				add_rpid_header(session, tdata, id);
				return;
			}
		}
	}

	if (ast_strlen_zero(reg_calleenum) && !ast_strlen_zero(sip_tsf) && !strncmp(sip_tsf, "ft_attended", 11)) {
		reg_calleenum = (char *)pbx_builtin_getvar_helper(session->channel, "REG_CALLEENUM");
		if (!ast_strlen_zero(reg_calleenum)) {
			ast_channel_reg_calleenum_set(session->channel, (const char *)reg_calleenum);
		}
		if ((session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) && strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
			char *aauser = (char *)pbx_builtin_getvar_helper(session->channel, "AAUSER");
			if (!ast_strlen_zero(aauser)) {
				tmp_id.number.str = ast_strdup(aauser);
				tmp_id.number.valid = 1;
				tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;
				add_rpid_header(session, tdata, &tmp_id);
				ast_party_id_free(&tmp_id);

				ast_log(LOG_NOTICE, "set rpid number[%s]", aauser);
				return;
			}
		}
	}

	if (!ast_strlen_zero(reg_calleenum) && (!ast_strlen_zero(cf_info) || !ast_strlen_zero(sip_tsf)) &&
		strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		tmp_id.number.str = ast_strdup(reg_calleenum);
		tmp_id.number.valid = 1;
		tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;
		add_rpid_header(session, tdata, &tmp_id);
		ast_party_id_free(&tmp_id);
		ast_log(LOG_NOTICE, "set rpid number[%s]", reg_calleenum);
		return;
	} 

	if (!ast_strlen_zero(reg_calleenum) && !strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		char *endpoint_id = NULL;
		char *trunk_id = NULL;
		char *rt_id = NULL;
		char *keep_ocid = NULL;
		char num[256]  = {0};
		char name[256] = {0};
		int  skip_name = 1;

		keep_ocid = (char *)pbx_builtin_getvar_helper(session->channel, "KEEP_ORIGCID");
		if (!ast_strlen_zero(keep_ocid) && !strncmp(keep_ocid, "yes", 3)) {
			ast_log(LOG_NOTICE, "chan[%s], RPID use original cid", ast_channel_name(session->channel));
			add_rpid_header(session, tdata, id);
			return;
		}

		endpoint_id = ast_strdupa(ast_sorcery_object_get_id(session->endpoint));
		trunk_id = endpoint_id + strlen("trunk_");

		rt_id = (char *)pbx_builtin_getvar_helper(session->channel, "OUTBOUND_RT_INDEX");
		ast_sip_inner_cid_to_outer_cid((char *)reg_calleenum, trunk_id, rt_id, num, sizeof(num)-1, name, sizeof(name)-1, &skip_name);

		if (strlen(num)) {
			tmp_id.number.str = ast_strdup(num);
			if (strlen(name)) {
				tmp_id.name.str = ast_strdup(name);
				tmp_id.name.valid = 1;
			}

			tmp_id.number.valid = 1;
			tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;
			add_rpid_header(session, tdata, &tmp_id);
			ast_log(LOG_NOTICE, "set rpid number[%s]", strlen(num)? num:reg_calleenum);

			ast_party_id_free(&tmp_id);
			return;
		}
	}

	/* Added by yfyang, for Bug 222347 */
	if ((!ast_strlen_zero(from_park) && !strncmp(from_park, "yes", 3)) ||(ast_channel_get_is_callflip(session->channel))) {//added by zyxia for bug226169.2022.6.13
		char *out_callee = (char *)pbx_builtin_getvar_helper(session->channel, "THE_REAL_OUT_CALLEE");
		if (!ast_strlen_zero(out_callee)) {
			tmp_id.number.str = ast_strdup(out_callee);
			tmp_id.number.valid = 1;
			tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;
			add_rpid_header(session, tdata, &tmp_id);
			ast_party_id_free(&tmp_id);
			ast_log(LOG_NOTICE, "set rpid number[%s]", out_callee);
			return;
		}
	}

	add_rpid_header(session, tdata, id);
	return;
}
#endif

/*!
 * \internal
 * \brief Add any appropriate identification headers to an outbound SIP message
 *
 * This will determine if an outbound message should have identification headers and
 * will add the appropriately configured headers
 *
 * \param session The session on which we will be sending the message
 * \param tdata The outbound message
 * \param The identity information to place on the message
 */
static void add_id_headers(const struct ast_sip_session *session, pjsip_tx_data *tdata, const struct ast_party_id *id)
{
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug259603 */
	char *speeddial_cid = (char *)pbx_builtin_getvar_helper(session->channel, "SpeedDial_CID");
	if (!id->number.valid
		|| (!session->endpoint->id.trust_outbound
			&& (ast_party_id_presentation(id) & AST_PRES_RESTRICTION) != AST_PRES_ALLOWED)) {
		if ((session->inv_session->role == PJSIP_ROLE_UAS) && !ast_strlen_zero(speeddial_cid)) {
			ast_log(LOG_NOTICE, "the SpeedDial_CID is %s", speeddial_cid);
		} else {
			return;
		}
	}
#else
	if (!id->number.valid
		|| (!session->endpoint->id.trust_outbound
			&& (ast_party_id_presentation(id) & AST_PRES_RESTRICTION) != AST_PRES_ALLOWED)) {
		return;
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	int pai_added = 0;
	int is_regtk  = 0;
	unsigned int need_add_rpid = ast_sip_session_judge_if_need_add_rpid(session);
	pjsip_contact_hdr *contact;

	pj_lock_acquire(tdata->lock);
	/* Added by yfyang, for Bug60850 */
	if (session->inv_session->role == PJSIP_ROLE_UAS) {
		/* Added by yfyang, for Bug69902 */
		if (session->endpoint->id.send_rpid || session->endpoint->id.send_pai || need_add_rpid) {
			/* Added by yfyang, for Bug 131843, Bug 157479 */
			const char *calleenum = pbx_builtin_getvar_helper(session->channel, "REG_CALLEENUM");
			const char *calleename = pbx_builtin_getvar_helper(session->channel, "REG_CALLEENAME");
			/*BEGIN: Mod by weidu, for use after free*/
			ast_channel_lock(session->channel);
			const char *peername = pbx_builtin_getvar_helper(session->channel, "BRIDGEPEER");
			char *pickup_caller = NULL;

			if (ast_strlen_zero(peername)) {
				peername = pbx_builtin_getvar_helper(session->channel, "custom");

				/* Added by yfyang, for Bug 221759 */
				if (ast_strlen_zero(peername)) {
					peername = pbx_builtin_getvar_helper(session->channel, "DIALEDPEERNAME");
				}
			}

			char *tmp_peername = NULL;
			if (peername){
				tmp_peername = ast_strdupa(peername);
				peername = tmp_peername;
			}
			ast_channel_unlock(session->channel);
			/*END: Mod by weidu*/

			if (!ast_strlen_zero(peername) && !strncmp(peername, "Local/", 6) && strstr(peername, ";1")) {
				struct ast_channel *local_chan = NULL;
				char *chan_name = strdupa(peername);

				if (chan_name) {
					chan_name[strlen(chan_name) - 1] = '2';
					ast_log(LOG_NOTICE, "local chan[%s]", chan_name);
					local_chan = ast_channel_get_by_name(chan_name);
					if (local_chan) {
						calleenum = pbx_builtin_getvar_helper(local_chan, "REG_CALLEENUM");
						calleename = pbx_builtin_getvar_helper(local_chan, "REG_CALLEENAME");
						peername = pbx_builtin_getvar_helper(local_chan, "DIALEDPEERNAME");
						if (ast_strlen_zero(peername)) {
							peername = pbx_builtin_getvar_helper(local_chan, "custom");
						}

						ast_channel_unref(local_chan);
					}
				}
			}

			if (!ast_strlen_zero(peername)) {
				struct ast_channel *peer_chan = ast_channel_get_by_name(peername);
				if (peer_chan) {
					pickup_caller = (char *)pbx_builtin_getvar_helper(peer_chan, "PICKUP_CALLER");
					ast_channel_unref(peer_chan);
				}
			}

			/*Start added by zyxia for Bug 247276 ,2022.12.29*/
			if(session->is_sfu && ast_channel_name(session->channel) && !strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)){
				contact = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CONTACT, tdata->msg->hdr.next);
				if (contact) {
					modify_contact_header(session, tdata->pool, contact, id);
				}
			}
			/*End added by zyxia for Bug 247276 ,2022.12.29*/

			/* Added by yfyang, for Bug259603 */
			if (!ast_strlen_zero(speeddial_cid)) {
				char spd_name[AST_MAX_CID_NAME] = {0};
				char spd_num[80] = {0};
				struct ast_party_id  spd_id;

				ast_party_id_init(&spd_id);
				ast_callerid_split(speeddial_cid, spd_name, sizeof(spd_name), spd_num, sizeof(spd_num));

				if (strlen(spd_num)) {
					spd_id.number.str = ast_strdup(spd_num);
					spd_id.number.valid = 1;
				}
				if (strlen(spd_name)) {
					spd_id.name.str = ast_strdup(spd_name);
					spd_id.name.valid = 1;
				}

				if (session->endpoint->id.send_rpid || need_add_rpid) {
					ast_log(LOG_NOTICE, "add_rpid_header, use SpeedDial_CID[%s]", speeddial_cid);
					add_rpid_header(session, tdata, &spd_id);
				}

				if (session->endpoint->id.send_pai) {
					add_pai_header(session, tdata, &spd_id); 
				}

				ast_party_id_free(&spd_id);
				pj_lock_release(tdata->lock);
				return;
			}

			ast_log(LOG_NOTICE, "calleenum[%s], peername[%s], trust_inbound[%d]", 
					ast_strlen_zero(calleenum)?" ":calleenum, ast_strlen_zero(peername)?" ":peername, session->endpoint->id.trust_inbound);
			if (!ast_strlen_zero(calleenum) && (!ast_strlen_zero(peername) && !strncmp(peername, "PJSIP/trunk_", 12))) {
				struct ast_party_id  rpid_id;

				ast_party_id_init(&rpid_id);
				ast_party_id_copy(&rpid_id, id);
				ast_free(rpid_id.number.str);
				rpid_id.number.str = ast_strdup(calleenum);

				if (rpid_id.name.valid) {
					ast_free(rpid_id.name.str);
					rpid_id.name.str = NULL;
					rpid_id.name.valid = 0;

					if (!ast_strlen_zero(calleename)) {
						rpid_id.name.str = ast_strdup(calleename);
						rpid_id.name.valid = 1;
					}
				}

				/*Start  added by zyxia for Bug 227429,callflip ringboth extension,2022.6.23*/
				if (ast_channel_get_is_callflip(session->channel) ) {
					char *out_callee = (char *)pbx_builtin_getvar_helper(session->channel, "THE_REAL_OUT_CALLEE");
					if (!ast_strlen_zero(out_callee)) {
						ast_log(LOG_NOTICE, "add_id_headers, THE_REAL_OUT_CALLEE[%s]", out_callee);

						if (rpid_id.number.valid) {
							ast_free(rpid_id.number.str);
						}
						rpid_id.number.str = ast_strdup(out_callee);
					}
				}
				/*End  added by zyxia for Bug 227429,callflip ringboth extension,2022.6.23*/

				/*Start added by zyxia for Bug 263901,update rpi when caller do n-way,2023.7.4*/
				if(session->is_sfu){
					const char* conf_room  = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
					if (!ast_strlen_zero(conf_room)) {

						if (rpid_id.number.valid) {
							ast_free(rpid_id.number.str);
						}
						rpid_id.number.str = ast_strdup(conf_room);
					}
				}
				/*End added by zyxia for Bug 263901,update rpi when caller do n-way,2023.7.4*/

				if (session->endpoint->id.send_rpid || need_add_rpid) {
					ast_log(LOG_NOTICE, "add_rpid_header, THE_REAL_OUT_CALLEE %s",rpid_id.name.str );
					add_rpid_header(session, tdata, &rpid_id);
				}

				if (session->endpoint->id.send_pai) {
					add_pai_header(session, tdata, &rpid_id); 
				}

				ast_party_id_free(&rpid_id);
			} else {
				/* Added by yfyang, for Bug 217269 */
				if (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
					char *aa_user = NULL;
					const char *park_by_exten = pbx_builtin_getvar_helper(session->channel, "CALLPARK_BY_EXTEN");
					if (!ast_strlen_zero(park_by_exten) && !strncmp(park_by_exten, "yes", 3)) {
						aa_user = (char *)pbx_builtin_getvar_helper(session->channel, "AAUSER");
					} else {
						/* Added by yfyang, for Bug 223915 */
						const char *from_disa = pbx_builtin_getvar_helper(session->channel, "FROM_DISA");
						if (!ast_strlen_zero(from_disa) && !strncmp(from_disa, "yes", 3)) {
							aa_user = (char *)pbx_builtin_getvar_helper(session->channel, "Callee_NUM");
						}
					}

					if (!ast_strlen_zero(aa_user)) {
						struct ast_party_id  rpid_id;

						ast_party_id_init(&rpid_id);
						ast_party_id_copy(&rpid_id, id);
						ast_free(rpid_id.number.str);
						rpid_id.number.str = ast_strdup(aa_user);


						ast_log(LOG_NOTICE, "add_id_headers, aa_user[%s]", aa_user);
						if (session->endpoint->id.send_rpid || need_add_rpid) {
							add_rpid_header(session, tdata, &rpid_id);
						}

						if (session->endpoint->id.send_pai) {
							add_pai_header(session, tdata, &rpid_id); 
						}

						ast_party_id_free(&rpid_id);
						pj_lock_release(tdata->lock);
						return; 
					}
				}

				if (ast_channel_get_is_callflip(session->channel) && strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
					struct ast_party_id  rpid_id;
					char *out_callee = (char *)pbx_builtin_getvar_helper(session->channel, "THE_REAL_OUT_CALLEE");
					/*Start added by zyxia for Bug 251671,2023.2.23*/
					if(ast_channel_get_is_page_caller(session->channel)) {
						const char *page_num = pbx_builtin_getvar_helper(session->channel, "PAGE_GROUP_NUM");
						if(page_num){
							ast_party_id_init(&rpid_id);
							ast_party_id_copy(&rpid_id, id);
							ast_free(rpid_id.number.str);
							rpid_id.number.str = ast_strdup(page_num);
							rpid_id.number.valid = 1;
							rpid_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;
							ast_log(LOG_NOTICE, "add_id_headers, real_user[%s]", page_num);
							if (session->endpoint->id.send_rpid || need_add_rpid) {
								add_rpid_header(session, tdata, &rpid_id);
							}

							if (session->endpoint->id.send_pai) {
								add_pai_header(session, tdata, &rpid_id); 
							}
							ast_party_id_free(&rpid_id);
							pj_lock_release(tdata->lock);
							return; 
						}
					}
					/*End added by zyxia for Bug 251671,2023.2.23*/
					/*Start added by zyxia for Bug 226169 - Call Flip 2022.6.13*/
					else if (!ast_strlen_zero(out_callee)) {
						ast_party_id_init(&rpid_id);
						ast_party_id_copy(&rpid_id, id);
						ast_free(rpid_id.number.str);
						rpid_id.number.str = ast_strdup(out_callee);
						rpid_id.number.valid = 1;
						rpid_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;

						ast_log(LOG_NOTICE, "add_id_headers, real_user[%s]", out_callee);
						if (session->endpoint->id.send_rpid || need_add_rpid) {
							add_rpid_header(session, tdata, &rpid_id);
						}

						if (session->endpoint->id.send_pai) {
							add_pai_header(session, tdata, &rpid_id); 
						}

						ast_party_id_free(&rpid_id);
						pj_lock_release(tdata->lock);
						return; 
					}
				}
				/*End added by zyxia for Bug 226169 - Call Flip 2022.6.13*/

				if (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
					if (session->endpoint && !strncmp(session->endpoint->trunk_type, "register", 8)) {
						is_regtk = 1;
						ast_log(LOG_NOTICE, "chan[%s] is register trunk.", ast_channel_name(session->channel));
					}
				}

				/* Added by yfyang, for Bug 226195 */
				if (is_regtk || (!ast_strlen_zero(pickup_caller) && !strncmp(pickup_caller, "yes", 3))) {
					if (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12) && (session->endpoint->id.send_rpid || need_add_rpid)) {
						char *trunk_id = (char *)pbx_builtin_getvar_helper(session->channel, "TRUNK_NUM");
						char *rpid_num = (char *)id->number.str;
						char num[256]  = {0};
						char name[256] = {0};
						int  skip_name = 1;

						if (ast_strlen_zero(trunk_id)) {
							trunk_id = (char *)pbx_builtin_getvar_helper(session->channel, "TRUNK_INDEX");
						}

						ast_log(LOG_NOTICE, "chan[%s], id_num[%s], trunk_id[%s]", ast_channel_name(session->channel), rpid_num, trunk_id?trunk_id:"");

						ast_sip_inner_cid_to_outer_cid(rpid_num, trunk_id, NULL, num, sizeof(num)-1, name, sizeof(name)-1, &skip_name);
						if (strlen(num)) {
							struct ast_party_id  tmp_id;
							ast_party_id_init(&tmp_id);

							tmp_id.number.str = ast_strdup(num);
							if (strlen(name)) {
								tmp_id.name.str = ast_strdup(name);
								tmp_id.name.valid = 1;
							}

							tmp_id.number.valid = 1;
							tmp_id.number.presentation = ast_party_id_presentation(id) | AST_PRES_ALLOWED;
							add_rpid_header(session, tdata, &tmp_id);
							ast_log(LOG_NOTICE, "set rpid number[%s]", num);

							ast_party_id_free(&tmp_id);
							pj_lock_release(tdata->lock);
							return;
						}
					}
				}

				if (session->endpoint->id.send_rpid || need_add_rpid) {
					ast_log(LOG_NOTICE, "add_rpid_header number[%s]", id->number.str);
					add_rpid_header(session, tdata, id);
				}

				if (session->endpoint->id.send_pai) {
					add_pai_header(session, tdata, id); 
				}
			}
		}

		pj_lock_release(tdata->lock);
		return; 
	}

	/* Added by yfyang, for Bug 95215 */
	if (!session->endpoint->id.send_ppi) {
		struct ast_party_id connected_id;

		contact = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CONTACT, tdata->msg->hdr.next);
		if (contact) {
			/* Added by yfyang, for Bug 240954 */
			char *from_tk = (char *)pbx_builtin_getvar_helper(session->channel, "FROM_TRUNK");
			char *cf_user = (char *)pbx_builtin_getvar_helper(session->channel, "CF_USER");

			if (!ast_strlen_zero(cf_user) && !ast_strlen_zero(from_tk) && !strcmp(from_tk, "yes") &&
			    !strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
				ast_party_id_init(&connected_id);
				connected_id.number.str = ast_strdup(cf_user);
				connected_id.number.valid = 1;
				modify_contact_header(session, tdata->pool, contact, &connected_id);
				ast_party_id_free(&connected_id);
			} else {
				if(ast_channel_get_is_join_trunk_conf(session->channel)) {
					//&& (tdata->msg->line.req.method.id == PJSIP_INVITE_METHOD || tdata->msg->line.req.method.id == PJSIP_UPDATE_METHOD)) {
					ast_log(LOG_NOTICE,"channel %s is join trunk conf,dont modify the invite contact\n",ast_channel_name(session->channel));
				}
				else if (!ast_strlen_zero(session->endpoint->username) && !strncmp(session->endpoint->trunk_type, "register", 8)) {
					int  dod_set = 0;
					ast_party_id_init(&connected_id);

					/* Added by yfyang, for bug 266343 */
					if (session->endpoint->id.use_dod_as_contact_user) {
						char *dod = (char *)pbx_builtin_getvar_helper(session->channel, "OUTDOD_CID");
						if (ast_strlen_zero(dod)) {
							dod = (char *)pbx_builtin_getvar_helper(session->channel, "DOD");
						}

						if (!ast_strlen_zero(dod)) {
							char dod_name[AST_MAX_CID_NAME] = {0};
							char dod_num[80] = {0};

							ast_callerid_split(dod, dod_name, sizeof(dod_name), dod_num, sizeof(dod_num));
							if (strlen(dod_num)) {
								connected_id.number.str = ast_strdup(dod_num);
								dod_set = 1;
							}
							if (strlen(dod_name)) {
								connected_id.name.str = ast_strdup(dod_name);
								connected_id.name.valid = 1;;
							}
						}
					}

					if (!dod_set) {
						connected_id.number.str = ast_strdup(session->endpoint->username);
					}

					connected_id.number.valid = 1;
					modify_contact_header(session, tdata->pool, contact, &connected_id);
					ast_party_id_free(&connected_id);
				} else {
					modify_contact_header(session, tdata->pool, contact, id);
				}
			}
		}
	}

	if (session->endpoint->id.send_pai) {
        /* for Bug 69810 */
		if (!ast_strlen_zero(session->endpoint->pai_number)) {
			struct ast_party_id  pai_id;
			char cid_name[AST_MAX_CID_NAME] = {0};
			char cid_num[256]  = {0};
			char *p_num = cid_num;

			if (strchr(session->endpoint->pai_number, '<') && strchr(session->endpoint->pai_number, '>')) {
				if (session->endpoint->pai_number[0] == '<') {
					sscanf(session->endpoint->pai_number, "<%255[^>]", cid_num); 
				} else {
					sscanf(session->endpoint->pai_number, "%255[^<]<%255[^>]", cid_name, cid_num); 
				}
			} else {
				p_num = (char *)session->endpoint->pai_number;
			}

			if (strlen(p_num)) {
				ast_party_id_init(&pai_id);
				if (strlen(cid_name)) {
					pai_id.name.str = ast_strdup(cid_name);
					pai_id.name.valid = 1;
				}

				pai_id.number.str = ast_strdup(p_num);
				pai_id.number.valid = 1;
				add_pai_header(session, tdata, &pai_id); 
				ast_party_id_free(&pai_id);
				pai_added = 1;
			}
		} else {
			add_pai_header(session, tdata, id); 
			pai_added = 1;
		}
	} else if (session->endpoint->id.passthrough_pai) {
		/* to send the caller pai to the callee*/
		const char *caller_pai = pbx_builtin_getvar_helper(session->channel, "CALLER_PAI");
		if (!ast_strlen_zero(caller_pai)) {
			char cid_name[AST_MAX_CID_NAME] = {0};
			char cid_num[80]  = {0};
			struct ast_party_id  pai_id2;

			ast_callerid_split(caller_pai, cid_name, sizeof(cid_name), cid_num, sizeof(cid_num));
			if (strlen(cid_num)) {
				ast_party_id_init(&pai_id2); 

				if (strlen(cid_name)) {
					pai_id2.name.str = ast_strdup(cid_name);
					pai_id2.name.valid = 1;
				}

				pai_id2.number.str = ast_strdup(cid_num);
				pai_id2.number.valid = 1;

				add_pai_header(session, tdata, &pai_id2); 
				ast_party_id_free(&pai_id2);
				pai_added = 1;
			}
		}
	}

	/* add by chdruan@grandsteam.cn for P-Preferred-Identity header Bug 30238*/
	if (session->endpoint->id.send_ppi) {
		add_ppi_header(session, tdata, id);
		pai_added = 1;
	}

	if (session->endpoint->id.send_rpid || need_add_rpid) {
		/* Modified by yfyang, for Bug 155444 */
		set_rpid_headers(session, tdata, id);
	}

	add_pai_headers_for_anonymous_call(session, tdata, id, pai_added);

	pj_lock_release(tdata->lock);

#else
	if (session->endpoint->id.send_pai) {
		add_pai_header(session, tdata, id);
	}
	if (session->endpoint->id.send_rpid) {
		add_rpid_header(session, tdata, id);
	}
#endif
}

/*!
 * \internal
 * \brief Session supplement callback for outgoing INVITE requests
 *
 * On all INVITEs (initial and reinvite) we may add other identity headers
 * such as P-Asserted-Identity and Remote-Party-ID based on configuration
 * and privacy settings
 *
 * \param session The session on which the INVITE will be sent
 * \param tdata The outbound INVITE request
 */
static void caller_id_outgoing_request(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_party_id effective_id;
	struct ast_party_id connected_id;
#ifdef  GRANDSTREAM_NETWORKS
	int auto_answer = 0;
	int call_info_req = 0;
#endif

	if (!session->channel) {
		return;
	}

	ast_party_id_init(&connected_id);
	ast_channel_lock(session->channel);
	effective_id = ast_channel_connected_effective_id(session->channel);
	ast_party_id_copy(&connected_id, &effective_id);
	ast_channel_unlock(session->channel);

#ifdef  GRANDSTREAM_NETWORKS
	if (session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) {
		pjsip_fromto_hdr *from;
		pjsip_dialog *dlg;

		char name[256] = {0},num[256] = {0};
		const char *xfer_callerid = pbx_builtin_getvar_helper(session->channel, "XFER_CALLERID");
		if (!ast_strlen_zero(xfer_callerid)) {
			ast_callerid_split(xfer_callerid, name, sizeof(name), num, sizeof(num));
			ast_log(LOG_NOTICE,"xfer_num = %s,xfer_name = %s\n",num,name);

			ast_free(connected_id.name.str);
			connected_id.name.str = NULL;
			connected_id.name.valid = 0;

			if (strlen(name)) {
				connected_id.name.str = ast_strdup(name);
				connected_id.name.valid = 1;
			}

			ast_free(connected_id.number.str);
			connected_id.number.str = ast_strdup(num);
		}

		from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, tdata->msg->hdr.next);
		dlg = session->inv_session->dlg;

		if (!ast_strlen_zero(session->endpoint->fromuser)) {
			/* Modify by bfdu for local_id only use in this case, 2016.05.30 */
			struct ast_party_id  local_id;
			ast_party_id_init(&local_id);
			ast_party_id_copy(&local_id, &connected_id);
			/* End */

			ast_free(local_id.number.str);
			local_id.number.str = ast_strdup(session->endpoint->fromuser);

			ast_free(local_id.name.str);
			local_id.name.str = NULL;
			local_id.name.valid = 0;

			/* Added by yfyang, for bug90890 */
			char *dod = (char *)pbx_builtin_getvar_helper(session->channel, "OUTDOD_CID");
			if (ast_strlen_zero(dod)) {
				dod = (char *)pbx_builtin_getvar_helper(session->channel, "DOD");
			}

			if (!ast_strlen_zero(dod) && session->endpoint->id.use_dod_as_from_name) {
				/* added by kmye. for bug 103282 */
				char *start_ptr = NULL;
				char *end_ptr = NULL;
				if (NULL != (start_ptr = strchr(dod, '<')) && NULL != (end_ptr = strrchr(dod, '>'))) {
					char dod_tmp[128] = {0};
					snprintf(dod_tmp, sizeof(dod_tmp), "%.*s", (int)(end_ptr-start_ptr-1), start_ptr+1);
					if (!ast_strlen_zero(dod_tmp)) {
						local_id.name.str = ast_strdup(dod_tmp);
						local_id.name.valid = 1;
					}
				}
				else {
					local_id.name.str = ast_strdup(dod);
					local_id.name.valid = 1;
				}
			} else if (session->endpoint->id.self.name.valid && session->endpoint->id.self.name.str) {
				local_id.name.str = ast_strdup(session->endpoint->id.self.name.str);
				local_id.name.valid = 1;
			}
			modify_id_header(tdata->pool, from, &local_id);
			modify_id_header(dlg->pool, dlg->local.info, &local_id);
			ast_channel_set_stir_shaken_origtn(session->channel, local_id.number.str); /*Add by weidu ,for Bug 186634 stir/shaken*/
			/* Add by bfdu for memory leak, 2016.05.30 */
			ast_party_id_free(&local_id);
		} else {
			/* Added by yfyang, for Bug193120 */
			char *reg_calleenum = (char *)ast_channel_reg_calleenum(session->channel);
			char *cf_info = (char *)pbx_builtin_getvar_helper(session->channel, "CF_ALL");
			char *sip_tsf = (char *)pbx_builtin_getvar_helper(session->channel, "SIP_TRANSFER");

			if (!ast_strlen_zero(reg_calleenum) && (!ast_strlen_zero(cf_info) || !ast_strlen_zero(sip_tsf)) && strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
				struct ast_party_id  local_id;
				ast_party_id_init(&local_id);
				ast_party_id_copy(&local_id, &connected_id);

				ast_free(local_id.number.str);
				local_id.number.str = ast_strdup(reg_calleenum);

				ast_free(local_id.name.str);
				local_id.name.str = NULL;
				local_id.name.valid = 0;

				modify_id_header(tdata->pool, from, &local_id);
				modify_id_header(dlg->pool, dlg->local.info, &local_id);
				ast_party_id_free(&local_id);
				ast_log(LOG_NOTICE, "set from number[%s]", reg_calleenum);
			} else {
				if (!ast_strlen_zero(sip_tsf) && !strncmp(sip_tsf, "ft_attended", 11) && !strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
					struct ast_party_id  local_id;
					char *endpoint_id = NULL;
					char *trunk_id = NULL;
					char *rt_id = NULL;
					char num[256]  = {0};
					char name[256] = {0};
					int  skip_name = 1;

					ast_party_id_init(&local_id);
					endpoint_id = ast_strdupa(ast_sorcery_object_get_id(session->endpoint));
					trunk_id = endpoint_id + strlen("trunk_");

					rt_id = (char *)pbx_builtin_getvar_helper(session->channel, "OUTBOUND_RT_INDEX");
					ast_sip_inner_cid_to_outer_cid(connected_id.number.str, trunk_id, rt_id, num, sizeof(num)-1, name, sizeof(name)-1, &skip_name);

					if (strlen(num)) {
						local_id.number.str = ast_strdup(num);
						if (strlen(name)) {
							local_id.name.str = ast_strdup(name);
							local_id.name.valid = 1;
						}

						local_id.number.valid = 1;
						local_id.number.presentation = ast_party_id_presentation(&connected_id) | AST_PRES_ALLOWED;
						ast_log(LOG_NOTICE, "set local_id number[%s]", num);

						ast_party_id_copy(&connected_id, &local_id);
						modify_id_header(tdata->pool, from, &local_id);
						modify_id_header(dlg->pool, dlg->local.info, &local_id);
					}

					ast_party_id_free(&local_id);
				} else if (!ast_strlen_zero(sip_tsf) && !strncmp(sip_tsf, "ft_attended", 11) && strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
					char *fullcid = (char *)pbx_builtin_getvar_helper(session->channel, "tmp_peer_fullcallid");

					if (!ast_strlen_zero(fullcid)) {
						char num[256]  = {0};
						char name[256] = {0};
						struct ast_party_id  local_id;

						ast_party_id_init(&local_id);
						ast_callerid_split(fullcid, name, sizeof(name)-1, num, sizeof(num)-1);
						if (strlen(num)) {
							local_id.number.str = ast_strdup(num);
							if (strlen(name)) {
								local_id.name.str = ast_strdup(name);
								local_id.name.valid = 1;
							}

							local_id.number.valid = 1;
							local_id.number.presentation = ast_party_id_presentation(&connected_id) | AST_PRES_ALLOWED;
							ast_log(LOG_NOTICE, "set local_id number[%s]", num);

							ast_party_id_copy(&connected_id, &local_id);
							modify_id_header(tdata->pool, from, &local_id);
							modify_id_header(dlg->pool, dlg->local.info, &local_id);
						}

						ast_party_id_free(&local_id);
					}
				}
			}
		}

		if (session->endpoint->use_for_sfu_conf && !ast_strlen_zero(ast_channel_caller_num(session->channel))) {
			struct ast_party_id  local_id;
			ast_party_id_init(&local_id);
			ast_party_id_copy(&local_id, &connected_id);

			ast_free(local_id.number.str);
			local_id.number.str = ast_strdup(ast_channel_caller_num(session->channel));
			modify_id_header(tdata->pool, from, &local_id);
			modify_id_header(dlg->pool, dlg->local.info, &local_id);
#ifdef  GRANDSTREAM_NETWORKS /*Add by weidu ,for stir/shaken*/
			ast_channel_set_stir_shaken_origtn(session->channel, local_id.number.str); /*Add by weidu ,for Bug 186634 stir/shaken*/
#endif
			ast_party_id_free(&local_id);
		}
 
		if (ast_channel_get_elin_flag(session->channel)) {
			int ip_ver;
			char elin[16] = "invalid";
			struct emergency_elin *tmp_elin = NULL;
			int mask = 0;
			unsigned int elin_ip;
			unsigned int src;   
			int mask_flag = 0;
			char loc_route[8] = {0};
			char location[512] = {0};
			ip_ver = ast_channel_get_elin_ip_ver((session->channel));
			AST_LIST_TRAVERSE_SAFE_BEGIN(&emergency_elin_list, tmp_elin, entry) {
				if (tmp_elin->net_ver != ip_ver) {
					continue;
				}
				if (tmp_elin->net_ver == AF_INET) {
					src = ntohl(ast_channel_get_elin_ip(session->channel));
					elin_ip = ntohl(tmp_elin->addr.s_addr);
					mask = 0xFFFFFFFF << (32 - tmp_elin->mask);
					ast_log(LOG_DEBUG, "find src[0X%x] elin_ip[0X%x], mask[%d].", src, elin_ip, mask);
					if (((elin_ip ^ src) & mask) == 0 && tmp_elin->elin && tmp_elin->mask >= mask_flag) {
						strncpy(elin, tmp_elin->elin, sizeof(elin) - 1);
						mask_flag = tmp_elin->mask;
						strncpy(location, tmp_elin->location, sizeof(location) - 1);
						strncpy(loc_route, tmp_elin->loc_route, sizeof(loc_route) - 1);
						ast_log(LOG_DEBUG, "find elin[%s] match the src.", tmp_elin->elin);
						//break;
					}
				} else if (tmp_elin->net_ver == AF_INET6) {
					if (ipv6_prefix_is_equal(ast_channel_get_elin_ip6(session->channel), &tmp_elin->addr6, tmp_elin->mask) &&
						tmp_elin->mask >= mask_flag) {
						strncpy(elin, tmp_elin->elin, sizeof(elin) - 1);
						mask_flag = tmp_elin->mask;
						strncpy(loc_route, tmp_elin->loc_route, sizeof(loc_route) - 1);
						ast_log(LOG_DEBUG, "find elin[%s] match the src.", tmp_elin->elin);
						//break;                    
					}
				}
			}
			AST_LIST_TRAVERSE_SAFE_END;
			
			if(strncmp(elin, "invalid", sizeof(elin))){
				struct ast_party_id  local_id;
				char *extension = NULL;
				char buf[1024] = {0};
				char function[128] = {0};
				char loc_uri[512] = {0};
				const char *isregtrunk = NULL;

				extension = ast_calloc(1, AST_MAX_EXTENSION);
				strncpy(extension, ast_channel_get_elin_dialer(session->channel), AST_MAX_EXTENSION);
				ast_party_id_init(&local_id);
				ast_party_id_copy(&local_id, &connected_id);
				ast_free(local_id.number.str);
				local_id.number.str = ast_strdup(elin);
				modify_id_header(tdata->pool, from, &local_id);
				modify_id_header(dlg->pool, dlg->local.info, &local_id);
#ifdef  GRANDSTREAM_NETWORKS /*Add by weidu ,for stir/shaken*/
				ast_channel_set_stir_shaken_origtn(session->channel, local_id.number.str); /*Add by weidu ,for Bug 186634 stir/shaken*/
#endif
				ast_party_id_free(&local_id);
				snprintf(loc_uri, sizeof(loc_uri)-1, "<https://%s>", location);
				ast_sip_add_header(tdata,"Geolocation",loc_uri);
				ast_sip_add_header(tdata,"Geolocation-Routing",loc_route);


				/*store to the db*/
				isregtrunk = pbx_builtin_getvar_helper((struct ast_channel *)session->channel, "ELIN_ON_REGTRUNK");
				if(!isregtrunk)
				{
					snprintf(function, sizeof(function), "DIALOUT_ELIN(%s,%s)", elin, extension);
					ast_log(LOG_DEBUG, "change callerid [%s] to[%s] excu func[%s].", extension, elin, function);
					ast_func_read(session->channel, function, buf, sizeof(buf));                 
				}
				ast_free(extension);
			}
		}
        
		if (session->endpoint && session->endpoint->custom_autoanswer) {
			auto_answer = 1;
		}

		/* Added by yfyang, for Broadsoft Test, support P-Access-Network-Info */
		if (session->endpoint && session->endpoint->id.send_pani && !ast_strlen_zero(session->endpoint->access_info)) {			
			static const pj_str_t pani_name = { "P-Access-Network-Info", 21 };	
			
			pjsip_generic_string_hdr *pani_hdr = pjsip_msg_find_hdr_by_name(tdata->msg, &pani_name, NULL);
			if (!pani_hdr) {
				pj_str_t pani_value;
				 
				pj_strdup2(tdata->pool, &pani_value, (const char*)session->endpoint->access_info);
				pjsip_generic_string_hdr *pani_hdr = pjsip_generic_string_hdr_create(tdata->pool, &pani_name, &pani_value);
				pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)pani_hdr); 
	
				ast_log(LOG_NOTICE, "Outgoing msg with PANI!");
			}
		}
	}		
#endif
	add_id_headers(session, tdata, &connected_id);
#ifdef  GRANDSTREAM_NETWORKS /*Add by weidu ,for Bug 186634 stir/shaken*/
	char* origintn = ast_channel_get_stir_shaken_origtn(session->channel);
	if (ast_strlen_zero(origintn)){
		ast_channel_set_stir_shaken_origtn(session->channel, connected_id.number.str);
		ast_log(LOG_DEBUG, "sip stir/shaken origin tn %s \n", ast_channel_get_stir_shaken_origtn(session->channel));
	}else{
		ast_log(LOG_DEBUG, "sip stir/shaken origin tn %s \n", origintn);
	}
#endif
	ast_party_id_free(&connected_id);

#ifdef  GRANDSTREAM_NETWORKS
	const char *callinfo_msg = pbx_builtin_getvar_helper(session->channel, "CALL_INFO_MSG");
	if (!ast_strlen_zero(callinfo_msg)) {
		if (!strncmp(callinfo_msg, "request", 7)) {
			call_info_req = 1;
		}
	}

    /* Bug 102993 */
	const char *codeblue_auto = pbx_builtin_getvar_helper(session->channel, "CODEBLUE_AUTO");
	if (!ast_strlen_zero(codeblue_auto)) {
		if (!strncmp(codeblue_auto, "yes", 3)) {
			auto_answer = 1;
		}
	}

	if (auto_answer || call_info_req || session->endpoint->sca_enable) {
		add_callinfo_header(session, tdata, auto_answer, call_info_req, session->endpoint->sca_enable); 
	}

	/* Bug 103782 */
	if (session->endpoint->sca_enable) {
		const char *alert_info = pbx_builtin_getvar_helper(session->channel, "DID_ALERTINFO"); 
		if (!ast_strlen_zero(alert_info)) {
			add_alertinfo_header(session, tdata, alert_info);
		}
	}

	/* Add alert info header for parking lot timeout ring back call */
	const char* parking_alert_info = pbx_builtin_getvar_helper(session->channel, "PARKING_LOT_ALERT_INFO");
	const unsigned int is_parking_ringback = ast_channel_get_parking(session->channel);
	if (is_parking_ringback && !ast_strlen_zero(parking_alert_info)) {
		replace_alertinfo_header(tdata, parking_alert_info);
	}

	/* Added by yfyang, for missed call feature */
	add_calltype_params_to_from(session, tdata);

	modify_from_user_for_conference(session, tdata);

	/* Added by yfyang, for IPVT Server application */
	add_ipvt_client_info_header(session, tdata, 0); 

	/* Added by yfyang, for Broadsoft Test */
	if (session->endpoint->id.passthrough_pai && !session->endpoint->id.send_pai && !session->endpoint->id.send_anonymous) {
		const char *peer_pai_info = pbx_builtin_getvar_helper(session->channel, "PEER_PAI_INFO");
		if (!ast_strlen_zero(peer_pai_info)) {
			add_peer_pai_info_header(session, tdata, peer_pai_info); 
		}
	}

	/* Added by yfyang, for sip ssh proxy feature */
	const char *proxy_flag = pbx_builtin_getvar_helper(session->channel, "SSH_PROXY_FLAG");
	if (!ast_strlen_zero(proxy_flag) && !strncmp(proxy_flag, "yes", 3)) {
		ast_log(LOG_NOTICE, "chan[%s] has set SSH_PROXY_FLAG, added to Contact.", ast_channel_name(session->channel));
		ast_sip_set_contact_param_for_sshproxy(tdata);
	}

	/* Added by yfyang, for Bug 232651, support P-Called-Party-ID */
	if (session->endpoint->id.send_pcpid && (tdata->msg->line.req.method.id == PJSIP_INVITE_METHOD)) {
		const char *peer_pcpid_info = pbx_builtin_getvar_helper(session->channel, "PEER_PCPID_INFO");
		if (!ast_strlen_zero(peer_pcpid_info)) {
			ast_sip_add_header(tdata, "P-Called-Party-ID", peer_pcpid_info);
			ast_log(LOG_NOTICE, "Add P-Called-Party-ID: [%s]", peer_pcpid_info);
		}
	}

#ifdef BUILD_ON_CLOUD_PBX
	add_trunk_host_header(session, tdata);
#endif
#endif
}

/*!
 * \internal
 * \brief Session supplement for outgoing INVITE response
 *
 * This will add P-Asserted-Identity and Remote-Party-ID headers if necessary
 *
 * \param session The session on which the INVITE response is to be sent
 * \param tdata The outbound INVITE response
 */
static void caller_id_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_party_id effective_id;
	struct ast_party_id connected_id;

	if (!session->channel
		|| (!session->endpoint->send_connected_line
			&& session->inv_session
			&& session->inv_session->state >= PJSIP_INV_STATE_EARLY)) {
		return;
	}

	/* Must do a deep copy unless we hold the channel lock the entire time. */
	ast_party_id_init(&connected_id);
	ast_channel_lock(session->channel);
	effective_id = ast_channel_connected_effective_id(session->channel);
	ast_party_id_copy(&connected_id, &effective_id);
	ast_channel_unlock(session->channel);

#ifdef  GRANDSTREAM_NETWORKS
	if (!connected_id.number.valid && strlen(ast_channel_connect_num(session->channel))) {
		connected_id.number.str = ast_strdup(ast_channel_connect_num(session->channel));
		connected_id.number.valid = 1;
	}
#endif

	add_id_headers(session, tdata, &connected_id);
	ast_party_id_free(&connected_id);

#ifdef  GRANDSTREAM_NETWORKS
	int  call_info_resp = 0;
	const char *callinfo_msg = pbx_builtin_getvar_helper(session->channel, "CALL_INFO_MSG");
	if (!ast_strlen_zero(callinfo_msg)) {
		if (!strncmp(callinfo_msg, "response", 8)) {
			call_info_resp = 1;
		} else if (!strncmp(callinfo_msg, "request", 7)) {

			/* Added by yfyang, for Bug 91578 */
			const char *pickup_caller = pbx_builtin_getvar_helper(session->channel, "PICKUP_CALLER");
			if (!ast_strlen_zero(pickup_caller) && !strncmp(pickup_caller, "yes", 3)) {
				call_info_resp = 1;
			}
		}
    }

	if (call_info_resp || session->endpoint->sca_enable) {
		add_callinfo_header(session, tdata, 0, 1, session->endpoint->sca_enable); 
	}

	/* Added by yfyang, for IPVT Server application */
	add_ipvt_client_info_header(session, tdata, 1); 
	
	/* Added by yfyang, for Broadsoft Test */
	if (session->endpoint->id.passthrough_pai) {
		const char *peer_pai_info = pbx_builtin_getvar_helper(session->channel, "PEER_PAI_INFO");
		if (!ast_strlen_zero(peer_pai_info)) {
			add_peer_pai_info_header(session, tdata, peer_pai_info); 
		}
	}
#endif
}

static struct ast_sip_session_supplement caller_id_supplement = {
	.method = "INVITE,UPDATE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL - 1000,
	.incoming_request = caller_id_incoming_request,
	.incoming_response = caller_id_incoming_response,
	.outgoing_request = caller_id_outgoing_request,
	.outgoing_response = caller_id_outgoing_response,
};

#ifdef  GRANDSTREAM_NETWORKS
static int load_config(int reload)
{
	struct ast_config *cfg = NULL;
	const char *category = NULL;
	const char *location = "";
	const char *route = "no";
	const char *network_segment = "";
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct emergency_elin *elin_t = NULL;
	char delim[] = "/";
	char *token  = NULL;
	char *save_ptr = NULL;

	if (!(cfg = ast_config_load("emergency_elin.conf", config_flags))) {
		return 0; 
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		return 0;
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&emergency_elin_list, elin_t, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(elin_t->elin);
		ast_free(elin_t->location);
		ast_free(elin_t->loc_route);
		ast_free(elin_t);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_HEAD_INIT_NOLOCK(&emergency_elin_list);

	ast_log(LOG_WARNING, "load_config in emergency_elin");
	while ((category = ast_category_browse(cfg, category))) {
		struct emergency_elin *tmp_elin = NULL;
		tmp_elin = ast_calloc(1, sizeof(struct emergency_elin));
		tmp_elin->elin = strdup(category);
		ast_log(LOG_WARNING, "tmp_elin->elin, [%s]", tmp_elin->elin);

		if ((network_segment = ast_variable_retrieve(cfg, category, "network_segment")) && !ast_strlen_zero(network_segment)) {
			token = strtok_r((char *)network_segment, delim, &save_ptr);
			if (token) {
				if(inet_pton(AF_INET, token, (void *)(&tmp_elin->addr))){
					tmp_elin->net_ver = AF_INET;
				} else if (inet_pton(AF_INET6, token, (void *)(&tmp_elin->addr6))) {
					tmp_elin->net_ver = AF_INET6;
				}
			}

			if (save_ptr) {
				tmp_elin->mask = atoi(save_ptr);
			}
		}

		if ((location = ast_variable_retrieve(cfg, category, "location")) && !ast_strlen_zero(location)) {
			tmp_elin->location = strdup(location);
		}

		if ((route = ast_variable_retrieve(cfg, category, "loc_route")) && !ast_strlen_zero(route)) {
			tmp_elin->loc_route = strdup(route);
		}

		AST_LIST_INSERT_TAIL(&emergency_elin_list, tmp_elin, entry);		
	}
	ast_config_destroy(cfg);
	return 0;
}

static int reload_module(void)
{
	load_config(1);
	return 0;
}
#endif	

static int load_module(void)
{
	ast_module_shutdown_ref(AST_MODULE_SELF);
#ifdef  GRANDSTREAM_NETWORKS
	load_config(0);
#endif	
	ast_sip_session_register_supplement(&caller_id_supplement);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
#ifdef  GRANDSTREAM_NETWORKS
	struct emergency_elin *elin_t = NULL;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&emergency_elin_list, elin_t, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(elin_t->elin);
		ast_free(elin_t->location);
		ast_free(elin_t);
	}
	AST_LIST_TRAVERSE_SAFE_END;
#endif
	ast_sip_session_unregister_supplement(&caller_id_supplement);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Caller ID Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
#ifdef  GRANDSTREAM_NETWORKS
	.reload = reload_module,
#endif
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip",
);

