/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2020, Digium, Inc.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/module.h"
#include "asterisk/xml.h"
#include "asterisk/res_pjsip_camera_control.h"

AST_MODULE_LOG("pjsip");

static int is_camera_control_type(pjsip_rx_data *rdata, char *subtype)
{
	return rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, subtype);
}

static void send_response(struct ast_sip_session *session,
			  struct pjsip_rx_data *rdata, int code)
{
	pjsip_tx_data *tdata;
	pjsip_dialog *dlg = session->inv_session->dlg;

	if (pjsip_dlg_create_response(dlg, rdata, code, NULL, &tdata) == PJ_SUCCESS) {
		struct pjsip_transaction *tsx = pjsip_rdata_get_tsx(rdata);
		pjsip_dlg_send_response(dlg, tsx, tdata);
	}
}

static enum ast_camera_control_action camera_control_info_str_2_action(const char *action)
{
	int i = 0;

	if (!action) {
		return CAMERA_CONTROL_INFO_INVALID;
	}

	for (i = 0; i < CAMERA_CONTROL_ACTION_MAX; i++) {
		if (!strcmp(action, camera_control_action_strmap[i][0])) {
			return (enum ast_camera_control_action)i;
		}
	}

	ast_log(LOG_WARNING, "Action:%s is invalid\n", action);

	return CAMERA_CONTROL_INFO_INVALID;
}

static enum ast_camera_control_status camera_control_info_str_2_reqstatus(const char *status)
{
	int i = 0;

	if (!status) {
		return CAMERA_CONTROL_INFO_INVALID;
	}

	for (i = 0; i < CAMERA_CONTROL_STATUS_MAX; i++) {
		if (!strcmp(status, camera_control_status_strmap[i][0])) {
			return (enum ast_camera_control_status)i;
		}
	}

	ast_log(LOG_WARNING, "Status:%s is invalid\n", status);

	return CAMERA_CONTROL_INFO_INVALID;
}

static unsigned int camera_control_info_str_2_capability(const char *capability)
{
	if (!capability) {
		return 0;
	}

	if (!strcasecmp(capability, "True") ||
		!strcasecmp(capability, "Yes")) {
		return 1;
	}

	if (!strcasecmp(capability, "False") ||
		!strcasecmp(capability, "No")) {
		return 0;
	}

	return 0;
}


static int camera_control_info_process_xml(char *xml, struct ast_camera_control_info *camera_control_info)
{
	struct ast_xml_doc  *xml_doc                     = NULL;
	struct ast_xml_node *camera_control_info_node    = NULL;
	struct ast_xml_node *child_nodes                 = NULL;
	struct ast_xml_node *node_iterator               = NULL;
	struct ast_xml_node *node_participant_channel_id = NULL;
	struct ast_xml_node *node_action                 = NULL;
	const char *node_name;
	const char *content;
	const char *action_name;
	char status[32] = {0};
	char failed_reason[128] = {0};
	char capability[16] = {0};

	if (!xml) {
		ast_log(LOG_DEBUG, "camera_control_info_process_xml xml is NULL\n");
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		ast_log(LOG_DEBUG, "camera_control_info_process_xml xml_doc is NULL\n");
		return -1;
	}

	camera_control_info_node = ast_xml_get_root(xml_doc);
	if (!camera_control_info_node) {
		ast_log(LOG_DEBUG, "camera_control_info_process_xml camera_control_info_node is NULL\n");
		ast_xml_close(xml_doc);
		return -1;
	}

	child_nodes = ast_xml_node_get_children(camera_control_info_node);
	if (child_nodes) {
		for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "Action") && !node_action) {
				node_action = node_iterator;
			} else if (!strcmp(node_name, "ParticipantChannelID") && !node_participant_channel_id) {
				node_participant_channel_id = node_iterator;
			}
		}
	}

	if (node_action) {
		action_name = ast_xml_get_attribute(node_action, "name");
		camera_control_info->action = camera_control_info_str_2_action(action_name);
		ast_log(LOG_NOTICE, "camera_control_info_process_xml, action[%s]\n", action_name);
		ast_xml_free_attr(action_name);

		for (node_iterator = ast_xml_node_get_children(node_action); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			content = ast_xml_get_text(node_iterator);
			if (!content) {
				continue;
			}

			if (!strcmp(node_name, "Status")) {
				snprintf(status, sizeof(status), "%s", content);
				camera_control_info->status = camera_control_info_str_2_reqstatus(status);
				ast_log(LOG_NOTICE, "camera_control_info_process_xml, Status[%s]\n", status);
			}

			if (!strcmp(node_name, "FailedReason")) {
				snprintf(failed_reason, sizeof(failed_reason), "%s", content);
				strncpy(camera_control_info->failed_reason, failed_reason, sizeof(camera_control_info->failed_reason) - 1);
				ast_log(LOG_NOTICE, "camera_control_info_process_xml, failed_reason[%s]\n", failed_reason);
			}

			if (!strcmp(node_name, "Capability")) {
				snprintf(capability, sizeof(capability), "%s", content);
				camera_control_info->camera_capability = camera_control_info_str_2_capability(capability);
				ast_log(LOG_NOTICE, "camera_control_info_process_xml, camera_capability[%s]\n", capability);
			}
			ast_xml_free_text(content);
		}
	}

	if (node_participant_channel_id) {
		content = ast_xml_get_text(node_participant_channel_id);
		if (content) {
			strncpy(camera_control_info->participant_channel_id, content, sizeof(camera_control_info->participant_channel_id) - 1);
			ast_log(LOG_NOTICE, "camera_control_info_process_xml, participant_channel_id[%s]\n", content);
		}
		ast_xml_free_text(content);
	}

	ast_xml_close(xml_doc);

	return 0;
}

static int camera_control_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	int res = 0;
	pjsip_msg_body *body = NULL;
	char *content = NULL;
	int is_camera_control_info = 0;
	struct ast_camera_control_info camera_control_info;

	ast_log(LOG_DEBUG, "camera_control_incoming_request begin\n");

	if (!session || !rdata || !rdata->msg_info.msg) {
		return 0;
	}

	is_camera_control_info = is_camera_control_type(rdata, CAMERA_CONTROL_INFO_SUB_TYPE);
	if (!is_camera_control_info) {
		return 0;
	}

	body = rdata->msg_info.msg->body;

	if (!body || !body->len) {
		return 0;
	}

	char buf[body ? (body->len + 1) : 0];
	memset(buf, 0, sizeof(buf));
	body->print_body(body, buf, body->len);

	if (!session->channel) {
		ast_log(LOG_DEBUG, "Session channel is NULL\n");
		return 0;
	}

	send_response(session, rdata, 200);

	ast_channel_lock(session->channel);

	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ZOMBIE) || ast_check_hangup(session->channel)) {
		ast_channel_unlock(session->channel);
		ast_log(LOG_DEBUG, "Channel has been hung up\n");
		return 0;
	}

	content = (char*) ast_malloc(body->len + 1);
	if (!content) {
		ast_channel_unlock(session->channel);
		return 0;
	}
	memset(content, 0, body->len + 1);
	strncpy(content, (char*)body->data, body->len);

	memset(&camera_control_info, 0, sizeof(struct ast_camera_control_info));
	res = camera_control_info_process_xml(content, &camera_control_info);
	if (!res) {
		switch (camera_control_info.action) {
			case CAMERA_CONTROL_ACTION_REPORT_CAMERA_CAPABILITY:
				//Todo
				break;
			case CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA:
			case CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA:
			case CAMERA_CONTROL_ACTION_OPEN_OWN_CAMERA:
			case CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA:
				session->camera_control_operater = 1;
				ast_queue_control_data(session->channel, AST_CONTROL_CAMERA_CONTROL, &camera_control_info, sizeof(camera_control_info));
				break;
			default:
				ast_queue_control_data(session->channel, AST_CONTROL_CAMERA_CONTROL, &camera_control_info, sizeof(camera_control_info));
				break;
		}
		
	}

	ast_log(LOG_NOTICE, "[session:%p] camera_control_incoming_request, process this info for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (content) {
		ast_free(content);
		content = NULL;
	}

	ast_channel_unlock(session->channel);

	return 1;
}

static struct ast_sip_session_supplement camera_control_supplement = {
	.method = "INFO",
	.incoming_request = camera_control_incoming_request,
};

static int load_module(void)
{
	ast_sip_session_register_supplement(&camera_control_supplement);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&camera_control_supplement);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP CAMERA CONTROL INFO Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
		.requires = "res_pjsip,res_pjsip_pubsub",
);

