/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2014, Digium, Inc.
 *
 * Thammanoon Kawinfruangfukul <thammanoon@grandstream.com>
 * Weiling Liao <weiling.liao@grandstream.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_pubsub</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>
#include <zlib.h>

#include "asterisk/astdb.h"
#include "asterisk/pjsip_cei.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/module.h"
#include "asterisk/strings.h"

AST_MODULE_LOG("pjsip");

static char* to_str_user_status(int status) {
	if (status == CEI_STATUS_CONNECTED) {
		return "connected";
	}
	else if (status == CEI_STATUS_DISCONNECTED) {
		return "disconnected";
	}
	else if (status == CEI_STATUS_ONHOLD) {
		return "on-hold";
	}
	else if (status == CEI_STATUS_MUTEVIAFOCUS) {
		return "muted-via-focus";
	}
	else if (status == CEI_STATUS_PENDING) {
		return "pending";
	}
	else if (status == CEI_STATUS_ALERTING) {
		return "alerting";
	}
	else if (status == CEI_STATUS_DIALINGIN) {
		return "dialing-in";
	}
	else if (status == CEI_STATUS_DIALINGOUT) {
		return "dialing-out";
	}
	else if (status == CEI_STATUS_DISCONNECTING) {
		return "disconnecting";
	}
	else if (status == CEI_STATUS_PAUSEVIAFOCUS) {
		return "connected";
	}

	return "unknown";
}

static void *cei_allocate_body(void *data)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	struct ast_str **cei_str;

	cei_str = ast_malloc(sizeof(*cei_str));
	if (!cei_str) {
		return NULL;
	}
	*cei_str = ast_str_create(64);
	if (!*cei_str) {
		ast_free(cei_str);
		return NULL;
	}
	return cei_str;
}

/* Hexadecimal string to binary string */
static void cei_hex_to_bin(char *source, char *dest, int len)
{
    int i = 0;
    char Dict[17][5] = {
        "0000", "0001", "0010", "0011",
        "0100", "0101", "0110", "0111",
        "1000", "1001", "1010", "1011",
        "1100", "1101", "1110", "1111",
    };
    for (i = 0; i < len; i++) {
        int n = 16;
        if (source[i] >= 'a' && source[i] <= 'f') {
            n = source[i] - 'a' + 10;
        }
        if (source[i] >= 'A' && source[i] <= 'F') {
            n = source[i] - 'A' + 10;
        }
        if (source[i] >= '0' && source[i] <= '9') {
            n = source[i] - '0';
        }
        memcpy(&dest[i*4],Dict[n],4);
    }
    return;
}

/*! Return need to be free  */
static char *escape_xml(char *data) {
	if (data == NULL || ast_strlen_zero(data)) {
		return NULL;
	}

	int num_escapes = 0;
	char *escape_pointer = strpbrk(data, "<>&\"\'");
	while (escape_pointer != NULL) {
		num_escapes++;
		escape_pointer = strpbrk(escape_pointer + 1, "<>&\"\'");
	}
	char escaped_data[strlen(data) + (num_escapes * 5) + 1];
	if (num_escapes == 0) {
		strcpy(escaped_data, data);
	}
	else {
		int num_copied = 0;
		char *head_pointer = data;
		escape_pointer = strpbrk(data, "<>&\"\'");
		while (escape_pointer != NULL) {
			strncpy(escaped_data + num_copied, head_pointer, (escape_pointer - head_pointer));
			num_copied += escape_pointer - head_pointer;
			if (strncmp(escape_pointer, "<", 1) == 0) {
				strcpy(escaped_data + num_copied, "&lt;");
				num_copied += 4;
			}
			else if (strncmp(escape_pointer, ">", 1) == 0) {
				strcpy(escaped_data + num_copied, "&gt;");
				num_copied += 4;
			}
			else if (strncmp(escape_pointer, "\"", 1) == 0) {
				strcpy(escaped_data + num_copied, "&quot;");
				num_copied += 6;
			}
			else if (strncmp(escape_pointer, "\'", 1) == 0) {
				strcpy(escaped_data + num_copied, "&apos;");
				num_copied += 6;
			}
			else if (strncmp(escape_pointer, "&", 1) == 0) {
				strcpy(escaped_data + num_copied, "&amp;");
				num_copied += 5;
			}
			head_pointer = escape_pointer + 1;
			escape_pointer = strpbrk(escape_pointer + 1, "<>&\"\'");
		}
		strcpy(escaped_data + num_copied, head_pointer);
	}
	//ast_log(LOG_DEBUG, "[WL] Escape %s to %s\n", data, escaped_data);

	return ast_strdup(escaped_data);
}

static int cei_generate_body_content(void *body, void *data)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	struct ast_str **cei = body;
	struct pjsip_cei_notify_data *cei_data = data;
	enum cei_notify_type notifyType = CEI_NOTIFY_UNKNOWN;
	int total_user_count = 0;
	int index = 0;
	int num_local_update_user = 0;
	int num_remote_conf = 0;
	char confid[41];
	char *escaped_content = NULL;
	char *saveptr = NULL;
	char *notifyPointer = NULL;
	char *local_count = NULL;
	char *remote_count = NULL;
	char *local_user_count = NULL;
	char *remote_user_count = NULL;
	char *total_update_count = NULL;
	char *conf_locked = NULL;
	char *subject = NULL;
	char *free_text = NULL;
	char *conf_start = NULL;
	char *conf_max_user = NULL;
	char *conference_type = NULL;
	char *conf_user_count = NULL;
	char *user_number = NULL;
	char *cid_number = NULL;
	char *cid_name = NULL;
	char *channel_name = NULL;
	char *conference_id = NULL;
	char *self_conference_id = NULL;
	char *user_role = NULL;
	char *entity_status = NULL;
	struct ast_str *database_tree = NULL;

	char *pres_user = NULL;
	char *muted     = NULL;
	char *v_muted   = NULL;
	char *mic_muted = NULL;
	char *v_capable = NULL;
	char *is_video  = NULL;
	char *avail_res = NULL;
	char *v_codec   = NULL;
	char *v_reso    = NULL;
	char *v_fr      = NULL;
	char *transfer_host = NULL;
	char *talking   = NULL;
	char *term_type = NULL;
	char *web_page  = NULL;
	char *extension = NULL;
	char *msid      = NULL;
	char *conf_muted = NULL;
	char *feature_id = NULL;
	char *sharing    = NULL;
	char *conf_config = NULL;
	char *record_type = NULL;
	char *record_start_timestamp_utc = NULL;
	char conf_config_str[16] = {0};
	char *related_ext  = NULL;
	char *isfocus = NULL;
	char *conf_layout = NULL;
	char *allow_user_unmute = NULL;
	char *hand_up_time = NULL;
	char *act_event_type = NULL;
	char *conf_creator = NULL;

	// Sanatity check
	if (cei_data == NULL) {
		return 0;
	}

	// Terminate notify
	if (cei_data->isTerminated == 1) {
		// Construct NOTIFY content
		ast_str_append(cei, 0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n");
		ast_str_append(cei, 0, "<conference-info xmlns=\"urn:ietf:params:xml:ns:conference-info\" entity=\"%s\" state=\"deleted\" version=\"%d\">\r\n",
															cei_data->ceiURI, cei_data->notifyVer);
		ast_str_append(cei, 0, "</conference-info>");
		return 0;
	}

	if (cei_data->channelName == NULL ||
		cei_data->conference == NULL || cei_data->content == NULL) {
		ast_log(LOG_WARNING, "[TK] Invalid CEI data");
		return 0;
	}

	//ast_log(LOG_DEBUG, "[TK] content: %s\n", cei_data->content);

	// Escaped the content
	escaped_content = escape_xml(cei_data->content);
	if (escaped_content == NULL) {
		return 0;
	}

	// Notify type
	notifyPointer = strtok_r(escaped_content, "|", &saveptr);

	if (notifyPointer == NULL) {
		ast_log(LOG_DEBUG, "[TK] Invalid notify type\n");
		ast_free(escaped_content);
		return 0;
	}
	notifyType = atoi(notifyPointer);
	ast_log(LOG_DEBUG, "[TK] notifyType: %d\n", notifyType);
	//ast_log(LOG_DEBUG, "[TK] notifyContent: %s\n", cei_data->content);

	local_count = strtok_r(NULL, "|", &saveptr);           // Local users (role:user/admin + role:conference)
	remote_count = strtok_r(NULL, "|", &saveptr);          // Remote users (role:user/admin + role:conference)
	local_user_count = strtok_r(NULL, "|", &saveptr);      // Local users (role:user/admin)
	remote_user_count = strtok_r(NULL, "|", &saveptr);     // Remote users (role:user/admin)
	total_update_count = strtok_r(NULL, "|", &saveptr);    // user/conference entity need to be updated/motified
	conf_locked = strtok_r(NULL, "|", &saveptr);
	record_type = strtok_r(NULL, "|", &saveptr);
	record_start_timestamp_utc = strtok_r(NULL, "|", &saveptr);
	subject = strtok_r(NULL, "|", &saveptr);
	free_text = strtok_r(NULL, "|", &saveptr);
	conf_start = strtok_r(NULL, "|", &saveptr);
	conf_max_user = strtok_r(NULL, "|", &saveptr);
	feature_id = strtok_r(NULL, "|", &saveptr);
	conf_config = strtok_r(NULL, "|", &saveptr);
	conf_layout = strtok_r(NULL, "|", &saveptr);
	allow_user_unmute = strtok_r(NULL, "|", &saveptr);
	act_event_type = strtok_r(NULL, "|", &saveptr);
	conf_creator = strtok_r(NULL, "|", &saveptr);

	/*We need to convert the hexadecimal information of the configuration to binary*/
	if (conf_config && strcasecmp(conf_config, "NULL")) {
		cei_hex_to_bin(conf_config, conf_config_str, strlen(conf_config));
	}

	if (notifyType == CEI_NOTIFY_FULL_EXTRA || notifyType == CEI_NOTIFY_PARTIAl_EXTRA) {
		pres_user = strtok_r(NULL, "|", &saveptr);
		web_page = strtok_r(NULL, "|", &saveptr);
		conf_muted = strtok_r(NULL, "|", &saveptr);
	} else {
		conf_muted = strtok_r(NULL, "|", &saveptr);
	}

	// Sanity check
	if (local_count == NULL || remote_count == NULL ||
		local_user_count == NULL || remote_user_count == NULL ||
		total_update_count == NULL || conf_locked == NULL ||
		subject == NULL || free_text == NULL ||
		conf_start == NULL || conf_max_user == NULL) {
		ast_log(LOG_WARNING, "[TK] Invalid CEI conference header data");
		ast_free(escaped_content);
		return 0;
	}

	// Get conference id
	memset(confid, 0, sizeof(confid));
	database_tree = ast_str_create(37);
	ast_str_append(&database_tree, 0, "MCB/%s", cei_data->conference);
	ast_db_get(ast_str_buffer(database_tree), "LOCAL_IDENTITY", confid, sizeof(confid));
	ast_free(database_tree);

	// Calculate user count
	total_user_count = atoi(local_user_count) + atoi(remote_user_count);

	// Construct NOTIFY content
	ast_str_append(cei, 0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n");
	ast_str_append(cei, 0, "<conference-info xmlns=\"urn:ietf:params:xml:ns:conference-info\" entity=\"%s\" state=\"%s\" version=\"%d\">\r\n",
							cei_data->ceiURI, (((notifyType == CEI_NOTIFY_FULL) || (notifyType == CEI_NOTIFY_FULL_EXTRA)) ? "full" : "partial"), cei_data->notifyVer);

	if (notifyType == CEI_NOTIFY_FULL || notifyType == CEI_NOTIFY_FULL_EXTRA) {
		ast_str_append(cei, 0, "<conference-description>\r\n");
		if (notifyType == CEI_NOTIFY_FULL_EXTRA) {
			ast_str_append(cei, 0, "<type>video conference</type>\r\n");
		} else {
			ast_str_append(cei, 0, "<type>conference</type>\r\n");
		}
		if (!ast_strlen_zero(conf_creator) && strcasecmp(conf_creator, "--")) {
			ast_str_append(cei, 0, "<creator>%s</creator>\r\n", conf_creator);
		}
		ast_str_append(cei, 0, "<subject>%s</subject>\r\n", subject);
		ast_str_append(cei, 0, "<free-text>%s</free-text>\r\n",free_text);
		ast_str_append(cei, 0, "<when>%s</when>\r\n", conf_start);
		ast_str_append(cei, 0, "<maximum-user-count>%s</maximum-user-count>\r\n",conf_max_user);
		ast_str_append(cei, 0, "<feature-id>%s</feature-id>\r\n", feature_id);
		ast_str_append(cei, 0, "</conference-description>\r\n");
	}

	ast_str_append(cei, 0, "<host-info>\r\n");
	ast_str_append(cei, 0, "<display-text>%s</display-text>\r\n", confid);
	if (notifyType == CEI_NOTIFY_FULL_EXTRA) {
		if (!ast_strlen_zero(web_page)) {
			if (!ast_strlen_zero(cei_data->webHost) && !strstr(web_page, cei_data->webHost)) {
				char scheme[8] = {0};
				int port = 0;

				sscanf(web_page, "%[^:]://%*[^:]:%d", scheme, &port);
				ast_str_append(cei, 0, "<web-page>%s://%s:%d</web-page>\r\n", scheme, cei_data->webHost, port);
			} else {
				ast_str_append(cei, 0, "<web-page>%s</web-page>\r\n", web_page);
			}
		}
	}
	ast_str_append(cei, 0, "</host-info>\r\n");

	ast_str_append(cei, 0, "<conference-state>\r\n");
	ast_str_append(cei, 0, "<user-count>%d</user-count>\r\n", total_user_count);
	ast_str_append(cei, 0, "<locked>%s</locked>\r\n", atoi(conf_locked) ? "true" : "false");

	ast_str_append(cei, 0, "<record-state>\r\n");
	if (!ast_strlen_zero(record_type) && strcasecmp(record_type, "none")) {
		time_t current_utc = ast_now_utc_timestamp();

		ast_str_append(cei, 0, "<type>%s</type>\r\n", record_type);
		ast_str_append(cei, 0, "<recording>true</recording>\r\n");
		ast_str_append(cei, 0, "<start-time>%s</start-time>\r\n", record_start_timestamp_utc);
		ast_str_append(cei, 0, "<current-time>%ld</current-time>\r\n", current_utc);
	} else {
		ast_str_append(cei, 0, "<recording>false</recording>\r\n");
	}
	ast_str_append(cei, 0, "</record-state>\r\n");

	if (notifyType == CEI_NOTIFY_FULL_EXTRA || notifyType == CEI_NOTIFY_PARTIAl_EXTRA) {
		ast_str_append(cei, 0, "<muted>%s</muted>\r\n", conf_muted ? (atoi(conf_muted) ? "true" : "false") : "false");
		ast_str_append(cei, 0, "<presentation-user>%s</presentation-user>\r\n", pres_user);
	} else {
		ast_str_append(cei, 0, "<muted>%s</muted>\r\n", conf_muted ? (atoi(conf_muted) ? "true" : "false") : "false");
	}
	ast_str_append(cei, 0, "<layout>%s</layout>\r\n", conf_layout ? conf_layout : "normal");

	ast_str_append(cei, 0, "</conference-state>\r\n");

	ast_str_append(cei, 0, "<conference-config>\r\n");
	ast_str_append(cei, 0, "<conf-config-bin>%s</conf-config-bin>\r\n", conf_config_str[0] != '\0' ? conf_config_str : "NULL");
	ast_str_append(cei, 0, "</conference-config>\r\n");
	ast_str_append(cei, 0, "<permissions>\r\n");
	ast_str_append(cei, 0, "<allow-unmute>%s</allow-unmute>\r\n", allow_user_unmute ? allow_user_unmute : "no");
	ast_str_append(cei, 0, "</permissions>\r\n");

	if (!ast_strlen_zero(act_event_type) && atoi(act_event_type) == CEI_ACT_TYPE_PUT_DOWN_ALL_HANDS) {
		ast_str_append(cei, 0, "<conference-event>\r\n");
		ast_str_append(cei, 0, "<hand-up>false</hand-up>\r\n");
		ast_str_append(cei, 0, "</conference-event>\r\n");
	}

	for (index = 0; index < atoi(total_update_count); index++) {
		conference_type = strtok_r(NULL, "\t", &saveptr);
		conf_user_count = strtok_r(NULL, "\t", &saveptr);
		user_number = strtok_r(NULL, "\t", &saveptr);
		cid_number = strtok_r(NULL, "\t", &saveptr);
		/*!begin Added by pyzhong@grandstream.cn for bug[68670], 2017-9-23*/
		if (saveptr[0] != '\t') {
			cid_name = strtok_r(NULL, "\t", &saveptr);
		} else {
			cid_name = "";
		}
		/*!end Added by pyzhong@grandstream.cn for bug[68670], 2017-9-23*/
		channel_name = strtok_r(NULL, "\t", &saveptr);
		conference_id = strtok_r(NULL, "\t", &saveptr);
		self_conference_id = strtok_r(NULL, "\t", &saveptr);
		user_role = strtok_r(NULL, "\t", &saveptr);

		if (!ast_strlen_zero(conference_type) && (atoi(conference_type) == CEI_CONFERENCE_REMOTE || atoi(conference_type) == CEI_CONFERENCE_VIRTUAL_LOCAL)) {
			entity_status = strtok_r(NULL, "|", &saveptr); 
		} else {
			if (notifyType != CEI_NOTIFY_FULL_EXTRA && notifyType != CEI_NOTIFY_PARTIAl_EXTRA) {
				entity_status = strtok_r(NULL, "\t", &saveptr); 
				muted = strtok_r(NULL, "\t", &saveptr);
				mic_muted = strtok_r(NULL, "\t", &saveptr);
				is_video = strtok_r(NULL, "\t", &saveptr);
				transfer_host = strtok_r(NULL, "\t", &saveptr);
				talking = strtok_r(NULL, "\t", &saveptr);
				term_type = strtok_r(NULL, "\t", &saveptr);
				extension = strtok_r(NULL, "|", &saveptr);
			} else {
				entity_status = strtok_r(NULL, "\t", &saveptr); 
				muted = strtok_r(NULL, "\t", &saveptr); 
				mic_muted = strtok_r(NULL, "\t", &saveptr);
				isfocus = strtok_r(NULL, "\t", &saveptr);
				hand_up_time = strtok_r(NULL, "\t", &saveptr);
				v_muted = strtok_r(NULL, "\t", &saveptr); 
				v_capable = strtok_r(NULL, "\t", &saveptr); 
				is_video = strtok_r(NULL, "\t", &saveptr); 
				avail_res = strtok_r(NULL, "\t", &saveptr); 
				v_codec = strtok_r(NULL, "\t", &saveptr); 
				v_reso = strtok_r(NULL, "\t", &saveptr); 
				v_fr = strtok_r(NULL, "\t", &saveptr);
				transfer_host = strtok_r(NULL, "\t", &saveptr);
				talking = strtok_r(NULL, "\t", &saveptr);
				term_type = strtok_r(NULL, "\t", &saveptr);
				extension = strtok_r(NULL, "\t", &saveptr);
				msid = strtok_r(NULL, "\t", &saveptr);
				sharing = strtok_r(NULL, "\t", &saveptr);
				related_ext = strtok_r(NULL, "|", &saveptr);
			}
		}


		if (conference_type == NULL || conf_user_count == NULL || user_number == NULL ||
			cid_number == NULL || cid_name == NULL || channel_name == NULL ||
			conference_id == NULL || user_role == NULL ||
			entity_status == NULL  || (self_conference_id == NULL && atoi(conference_type) != CEI_CONFERENCE_VIRTUAL_LOCAL )){ 
			ast_log(LOG_WARNING, "[TK] Invalid CEI conference body data");
			continue;
		}

		// Add user number to CID number to make it unique
		char xml_cid_number[strlen(cid_number) + strlen(user_number) + 2];
		strcpy(xml_cid_number, cid_number);
		strcpy(xml_cid_number + strlen(cid_number), "-");
		strcpy(xml_cid_number + strlen(cid_number) + 1, user_number);


		// Append remote conference
		if (atoi(conference_type) == CEI_CONFERENCE_REMOTE) {
			if (num_remote_conf == 0) {
				if (num_local_update_user > 0) {
					ast_str_append(cei, 0, "</users>\r\n");
				}
				ast_str_append(cei, 0, "<sidebars-by-val state=\"%s\">\r\n", (notifyType == CEI_NOTIFY_FULL || notifyType == CEI_NOTIFY_FULL_EXTRA) ? "full" : "partial");
			}
			else {
				ast_str_append(cei, 0, "</users>\r\n");
				ast_str_append(cei, 0, "</entry>\r\n");
			}

			if (notifyType == CEI_NOTIFY_FULL) {
				ast_str_append(cei, 0, "<entry entity=\"%s@%s.invalid\" state=\"full\">\r\n",
										xml_cid_number,
										conference_id);
			} else {
				ast_str_append(cei, 0, "<entry entity=\"%s@%s.invalid\" state=\"%s\">\r\n",
										xml_cid_number,
										conference_id,
										(atoi(entity_status) == CEI_STATUS_DISCONNECTED) ? "deleted" : "partial");
			}

			ast_str_append(cei, 0, "<conference-state>\r\n");
			ast_str_append(cei, 0, "<user-count>%s</user-count>\r\n",conf_user_count);
			ast_str_append(cei, 0, "</conference-state>\r\n");
			ast_str_append(cei, 0, "<users>\r\n");

			num_remote_conf++;

			continue;
		}

		// <!-- USERS -->
		if (atoi(conference_type) == CEI_CONFERENCE_LOCAL || atoi(conference_type) == CEI_CONFERENCE_VIRTUAL_LOCAL ) {
			if (num_remote_conf == 0 && num_local_update_user == 0) {
				ast_str_append(cei, 0, "<users>\r\n");
			}
			num_local_update_user++;
		}

		// Append user
		ast_str_append(cei, 0, "<user entity=\"%s@%s.invalid\" state=\"full\" userid=\"%s\">\r\n", xml_cid_number, conference_id, user_number);
		ast_str_append(cei, 0, "<display-text>%s</display-text>\r\n", cid_name);
		ast_str_append(cei, 0, "<extension>%s</extension>\r\n", (extension && strcmp(extension, "NULL")) ? extension : "");
		ast_str_append(cei, 0, "<real-ext>%s</real-ext>\r\n", (related_ext && strcmp(related_ext, "NULL")) ? related_ext : cid_number);

		if (strcmp(self_conference_id, "NULL")) {
			ast_str_append(cei, 0, "<associated-aors>%s</associated-aors>\r\n", self_conference_id);
		}
		ast_str_append(cei, 0, "<roles>\r\n");

		if (ast_sip_cei_same_channel(channel_name, cei_data->channelName, cid_number) == 0) {
			ast_str_append(cei, 0, "<entry>you</entry>\r\n");
		}
		switch (atoi(user_role)) {
			case CEI_ROLE_CONFERENCE:
			{
				ast_str_append(cei, 0, "<entry>conference</entry>\r\n");
			}
			break;
			case CEI_ROLE_ADMIN:
			{
				ast_str_append(cei, 0, "<entry>admin</entry>\r\n");
				ast_str_append(cei, 0, "<entry>invite</entry>\r\n");
			}
			break;
			case CEI_ROLE_USER_INVITE:
			{
				ast_str_append(cei, 0, "<entry>user</entry>\r\n");
				ast_str_append(cei, 0, "<entry>invite</entry>\r\n");
			}
			break;
			case CEI_ROLE_USER_NORMAL:
			{
				ast_str_append(cei, 0, "<entry>user</entry>\r\n");
			}
			break;
		}
		ast_str_append(cei, 0, "</roles>\r\n");
		ast_str_append(cei, 0, "<endpoint entity=\"%s\" state=\"full\">\r\n", channel_name);
		if (atoi(user_role) != CEI_ROLE_CONFERENCE) {
			ast_str_append(cei, 0, "<type>%s</type>\r\n", term_type ? term_type : "unknown");
		}

		ast_str_append(cei, 0, "<status>%s</status>\r\n", to_str_user_status(atoi(entity_status)));

		if (notifyType == CEI_NOTIFY_FULL_EXTRA || notifyType == CEI_NOTIFY_PARTIAl_EXTRA) {
			ast_str_append(cei, 0, "<muted>%s</muted>\r\n",  (muted && atoi(muted)) ? "true" : "false");
			ast_str_append(cei, 0, "<mic-muted>%s</mic-muted>\r\n",  (mic_muted && atoi(mic_muted)) ? "true" : "false");
			ast_str_append(cei, 0, "<isfocus>%s</isfocus>\r\n",  (isfocus && atoi(isfocus)) ? "true" : "false");
			ast_str_append(cei, 0, "<hand-up-time>%s</hand-up-time>\r\n",  hand_up_time ? hand_up_time : "0");
			ast_str_append(cei, 0, "<video-muted>%s</video-muted>\r\n", (v_muted && atoi(v_muted)) ? "true" : "false");
			ast_str_append(cei, 0, "<media-state>\r\n");
			ast_str_append(cei, 0, "  <video-capable>%s</video-capable>\r\n", (v_capable && atoi(v_capable)) ? "true" : "false");
			ast_str_append(cei, 0, "  <media-type>%s</media-type>\r\n", (is_video && atoi(is_video)) ? "video" : "audio");

			if (avail_res && strcmp(avail_res, "NULL")) {
				char *token = NULL;
				char *saveptr = NULL;

				for (token = strtok_r(avail_res, ",", &saveptr); token != NULL; token = strtok_r(NULL, ",", &saveptr)) {
					ast_str_append(cei, 0, "  <avail-resolution>%s</avail-resolution>\r\n", token);
				} 
			}

			if ((is_video && atoi(is_video)) && v_codec && strcmp(v_codec, "NULL")) {
				ast_str_append(cei, 0, "  <video>\r\n");
				ast_str_append(cei, 0, "    <codec>%s</codec>\r\n", v_codec);

				if (v_reso && strcmp(v_reso, "NULL")) {
					ast_str_append(cei, 0, "    <resolution>%s</resolution>\r\n", v_reso);
				}

				if (v_fr && strcmp(v_fr, "NULL")) {
					ast_str_append(cei, 0, "    <frame-rate>%s</frame-rate>\r\n", v_fr);
				}

				if (msid && strcmp(msid, "NULL")) {
					ast_str_append(cei, 0, "    <msid>%s</msid>\r\n", msid);
				}

				ast_str_append(cei, 0, "  </video>\r\n");
			}

			/*audio*/
			if (!is_video || !atoi(is_video)) {
				ast_str_append(cei, 0, "  <audio>\r\n");
				if (msid && strcmp(msid, "NULL")) {
					ast_str_append(cei, 0, "    <msid>%s</msid>\r\n", msid);
				}
				ast_str_append(cei, 0, "  </audio>\r\n");
			}

			ast_str_append(cei, 0, "</media-state>\r\n");
			ast_str_append(cei, 0, "<transfer-host>%s</transfer-host>\r\n", (transfer_host && atoi(transfer_host)) ? "true" : "false");
			ast_str_append(cei, 0, "<talk-status>%s</talk-status>\r\n", (talking && atoi(talking)) ? "talking" : "silence");
			ast_str_append(cei, 0, "<sharing-status>%s</sharing-status>\r\n", (sharing && atoi(sharing)) ? "true" : "false");
		}else {
			ast_str_append(cei, 0, "<muted>%s</muted>\r\n",  (muted && atoi(muted)) ? "true" : "false");
			ast_str_append(cei, 0, "<mic-muted>%s</mic-muted>\r\n",  (mic_muted && atoi(mic_muted)) ? "true" : "false");
			ast_str_append(cei, 0, "<media-state>\r\n");
			ast_str_append(cei, 0, "  <media-type>%s</media-type>\r\n", (is_video && atoi(is_video)) ? "video" : "audio");
			ast_str_append(cei, 0, "</media-state>\r\n");
			ast_str_append(cei, 0, "<transfer-host>%s</transfer-host>\r\n", (transfer_host && atoi(transfer_host)) ? "true" : "false");
			ast_str_append(cei, 0, "<talk-status>%s</talk-status>\r\n", (talking && atoi(talking)) ? "talking" : "silence");
		}

		ast_str_append(cei, 0, "</endpoint>\r\n");
		ast_str_append(cei, 0, "</user>\r\n");
	}

	if (num_remote_conf > 0) {
		ast_str_append(cei, 0, "</users>\r\n");
		ast_str_append(cei, 0, "</entry>\r\n");
		ast_str_append(cei, 0, "</sidebars-by-val>\r\n");
	}
	else if(num_local_update_user > 0){
		ast_str_append(cei, 0, "</users>\r\n");
	}
	ast_str_append(cei, 0, "</conference-info>\r\n");
	
	ast_free(escaped_content);
	ast_log(LOG_DEBUG, "[TK] Exit...%s\n", __FUNCTION__);

	return 0;
}

static void cei_to_string(void *body, struct ast_str **str)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
 	struct ast_str **cei = body;

	ast_str_set(str, 0, "%s", ast_str_buffer(*cei));
}

static void cei_to_encode_string(void *body, struct ast_str **str)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_str **cei = body;
	int body_lenth = 0;
	int encode_lenth = 0;
	char *str_in = NULL;
	char *str_out = NULL;
	unsigned long compress_len = 0;

	if (!*cei) {
		return;
	}
	body_lenth = ast_str_strlen(*cei) + 1;
	if (body_lenth <= 1) {
		return;
	}
	compress_len = compressBound(body_lenth) * 2;
	encode_lenth = (body_lenth/3 + 1)*4;

	str_in = (char*) ast_malloc(encode_lenth);
	str_out = (char*) ast_malloc(compress_len);
	if (str_in == NULL || str_out == NULL) {
		if (str_in) {
			ast_free(str_in);
		}
		if (str_out) {
			ast_free(str_out);
		}
		return;
	}

	//Zlib Compression
	if (ast_sip_compress(str_out, &compress_len, ast_str_buffer(*cei), body_lenth) != Z_OK) {
		ast_free(str_in);
		ast_free(str_out);
		return;
	}

	//Base64 Encoding
	memset(str_in, 0, encode_lenth);  /*Mod by weidu, for Bug 260211*/
	ast_base64encode(str_in, (const unsigned char *)str_out, compress_len, encode_lenth);
	ast_log(LOG_DEBUG, "[WL] Got encoded notify data: %s \n", str_in);

	ast_str_set(str, 0, "%s", str_in);

	// Clean up
	ast_free(str_in);
	ast_free(str_out);
}

static void cei_destroy_body(void *body)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_str **cei = body;

	ast_free(*cei);
	ast_free(cei);
}

static struct ast_sip_pubsub_body_generator cei_generator = {
	.type = CEI_TYPE,
	.subtype = CEI_SUBTYPE,
	.content_encoding = NOTIFY_CONTENT_ENCODING,
	.body_type = AST_SIP_CEI_NOTIFY_DATA,
	.allocate_body = cei_allocate_body,
	.generate_body_content = cei_generate_body_content,
	.to_string = cei_to_string,
	.to_encode_string = cei_to_encode_string,
	.destroy_body = cei_destroy_body,
};

static int load_module(void)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);

	if (ast_sip_pubsub_register_body_generator(&cei_generator)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	ast_sip_pubsub_unregister_body_generator(&cei_generator);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP CEI resource",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND,
	.requires = "res_pjsip,res_pjsip_pubsub",
);

#endif
