/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Kevin Harwell <kharwell@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/callerid.h"
#include "asterisk/channel.h"
#include "asterisk/module.h"
#include "asterisk/strings.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/pbx.h"
#include "asterisk/astdb.h"

AST_MODULE_LOG("pjsip");
#endif

static const pj_str_t diversion_name = { "Diversion", 9 };

/*!
 * \internal
 * \brief Determine if the given string is a SIP token.
 * \since 13.8.0
 *
 * \param str String to determine if is a SIP token.
 *
 * \note A token is defined by RFC3261 Section 25.1
 *
 * \return Non-zero if the string is a SIP token.
 */
static int sip_is_token(const char *str)
{
	int is_token;

	if (ast_strlen_zero(str)) {
		/* An empty string is not a token. */
		return 0;
	}

	is_token = 1;
	do {
		if (!isalnum(*str)
			&& !strchr("-.!%*_+`'~", *str)) {
			/* The character is not allowed in a token. */
			is_token = 0;
			break;
		}
	} while (*++str);

	return is_token;
}

/*! \brief Diversion header reasons
 *
 * The core defines a bunch of constants used to define
 * redirecting reasons. This provides a translation table
 * between those and the strings which may be present in
 * a SIP Diversion header
 */
static const struct reasons {
	enum AST_REDIRECTING_REASON code;
	const char *text;
} reason_table[] = {
	{ AST_REDIRECTING_REASON_UNKNOWN, "unknown" },
	{ AST_REDIRECTING_REASON_USER_BUSY, "user-busy" },
	{ AST_REDIRECTING_REASON_NO_ANSWER, "no-answer" },
	{ AST_REDIRECTING_REASON_UNAVAILABLE, "unavailable" },
	{ AST_REDIRECTING_REASON_UNCONDITIONAL, "unconditional" },
	{ AST_REDIRECTING_REASON_TIME_OF_DAY, "time-of-day" },
	{ AST_REDIRECTING_REASON_DO_NOT_DISTURB, "do-not-disturb" },
	{ AST_REDIRECTING_REASON_DEFLECTION, "deflection" },
	{ AST_REDIRECTING_REASON_FOLLOW_ME, "follow-me" },
	{ AST_REDIRECTING_REASON_OUT_OF_ORDER, "out-of-service" },
	{ AST_REDIRECTING_REASON_AWAY, "away" },
	{ AST_REDIRECTING_REASON_CALL_FWD_DTE, "cf_dte" },		/* Non-standard */
	{ AST_REDIRECTING_REASON_SEND_TO_VM, "send_to_vm" },	/* Non-standard */
#ifdef GRANDSTREAM_NETWORKS
	{ AST_REDIRECTING_REASON_BLIND_TRANSFER, "blind-transfer"},   /* Added by yfyang, for  Bug 127821 */
#endif
};

static const char *reason_code_to_str(const struct ast_party_redirecting_reason *reason)
{
	int idx;
	int code;

	/* use specific string if given */
	if (!ast_strlen_zero(reason->str)) {
		return reason->str;
	}

	code = reason->code;
	for (idx = 0; idx < ARRAY_LEN(reason_table); ++idx) {
		if (code == reason_table[idx].code) {
			return reason_table[idx].text;
		}
	}

	return "unknown";
}

static pjsip_fromto_hdr *get_diversion_header(pjsip_rx_data *rdata)
{
	static const pj_str_t from_name = { "From", 4 };

	pjsip_generic_string_hdr *hdr;
	pj_str_t value;
	int size;

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &diversion_name, NULL))) {
		return NULL;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);

	/* parse as a fromto header */
	return pjsip_parse_hdr(rdata->tp_info.pool, &from_name, value.ptr,
			       pj_strlen(&value), &size);
}

static void set_redirecting_value(char **dst, const pj_str_t *src)
{
	ast_free(*dst);
	*dst = ast_malloc(pj_strlen(src) + 1);
	if (*dst) {
		ast_copy_pj_str(*dst, src, pj_strlen(src) + 1);
	}
}

static void set_redirecting_id(pjsip_name_addr *name_addr, struct ast_party_id *data,
			       struct ast_set_party_id *update)
{
	pjsip_sip_uri *uri = pjsip_uri_get_uri(name_addr->uri);
	char *semi;
	pj_str_t uri_user;

	uri_user = uri->user;

	/* Always truncate redirecting number at a semicolon. */
	semi = pj_strchr(&uri_user, ';');
	if (semi) {
		/*
		 * We need to be able to handle URI's looking like
		 * "sip:1235557890;phone-context=national@x.x.x.x;user=phone"
		 *
		 * Where the uri->user field will result in:
		 * "1235557890;phone-context=national"
		 *
		 * People don't care about anything after the semicolon
		 * showing up on their displays even though the RFC
		 * allows the semicolon.
		 */
		pj_strset(&uri_user, (char *) pj_strbuf(&uri_user), semi - pj_strbuf(&uri_user));
	}

	if (pj_strlen(&uri_user)) {
		update->number = 1;
		data->number.valid = 1;
		set_redirecting_value(&data->number.str, &uri_user);
	}

	if (pj_strlen(&name_addr->display)) {
		update->name = 1;
		data->name.valid = 1;
		set_redirecting_value(&data->name.str, &name_addr->display);
	}
}

static void copy_redirecting_id(struct ast_party_id *dst, const struct ast_party_id *src,
				struct ast_set_party_id *update)
{
	ast_party_id_copy(dst, src);

	if (dst->number.valid) {
		update->number = 1;
	}

	if (dst->name.valid) {
		update->name = 1;
	}
}

static void set_redirecting_reason(pjsip_fromto_hdr *hdr,
				   struct ast_party_redirecting_reason *data)
{
	static const pj_str_t reason_name = { "reason", 6 };
	pjsip_param *reason = pjsip_param_find(&hdr->other_param, &reason_name);
	char *reason_str;

	if (!reason) {
		return;
	}

	set_redirecting_value(&data->str, &reason->value);
	if (!data->str) {
		/* Oops, allocation failure */
		return;
	}
	reason_str = ast_strdupa(data->str);

	/* Remove any enclosing double-quotes */
	if (*reason_str == '"') {
		reason_str = ast_strip_quoted(reason_str, "\"", "\"");
	}

	data->code = ast_redirecting_reason_parse(reason_str);
	if (data->code < 0) {
		data->code = AST_REDIRECTING_REASON_UNKNOWN;
	} else {
		ast_free(data->str);
		data->str = ast_strdup("");
	}
}

static void set_redirecting(struct ast_sip_session *session,
			    pjsip_fromto_hdr *from_info,
			    pjsip_name_addr *to_info)
{
	struct ast_party_redirecting data;
	struct ast_set_party_redirecting update;

	if (!session->channel) {
		return;
	}

	ast_party_redirecting_init(&data);
	memset(&update, 0, sizeof(update));

	if (from_info) {
		set_redirecting_id((pjsip_name_addr*)from_info->uri,
			&data.from, &update.from);
		set_redirecting_reason(from_info, &data.reason);
	} else {
		copy_redirecting_id(&data.from, &session->id, &update.from);
	}

#ifdef GRANDSTREAM_NETWORKS    /*for bug 72906*/
	if (to_info) {
		set_redirecting_id(to_info, &data.to, &update.to);
	}
#else
	set_redirecting_id(to_info, &data.to, &update.to);
#endif
	ast_set_party_id_all(&update.priv_orig);
	ast_set_party_id_all(&update.priv_from);
	ast_set_party_id_all(&update.priv_to);
	++data.count;

	ast_channel_set_redirecting(session->channel, &data, &update);
	ast_party_redirecting_free(&data);
}

static int diversion_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_fromto_hdr *hdr = get_diversion_header(rdata);

	if (hdr) {
#ifdef GRANDSTREAM_NETWORKS    /*for bug 72906*/
		set_redirecting(session, hdr, NULL);
#else
		set_redirecting(session, hdr, (pjsip_name_addr*)
				PJSIP_MSG_TO_HDR(rdata->msg_info.msg)->uri);
#endif
	}

	return 0;
}

static void diversion_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	static const pj_str_t contact_name = { "Contact", 7 };
	static const pj_str_t contact_name_s = { "m", 1 };

	pjsip_status_line status = rdata->msg_info.msg->line.status;
	pjsip_fromto_hdr *div_hdr;
	pjsip_contact_hdr *contact_hdr;

	if ((status.code != 302) && (status.code != 181)) {
		return;
	}
#ifdef GRANDSTREAM_NETWORKS
	const char* en_diversion = pbx_builtin_getvar_helper(session->channel, "EN_DIVERSION");
	if (en_diversion && 0 == strcasecmp(en_diversion, "no")) {
		return;
	}
#endif

	/* use the diversion header info if there is one. if not one then use the
           session caller id info. if that doesn't exist use info from the To hdr*/
	if (!(div_hdr = get_diversion_header(rdata)) && !session->id.number.valid) {
		div_hdr = PJSIP_MSG_TO_HDR(rdata->msg_info.msg);
	}

	contact_hdr = pjsip_msg_find_hdr_by_names(rdata->msg_info.msg, &contact_name, &contact_name_s, NULL);

	set_redirecting(session, div_hdr, contact_hdr ?	(pjsip_name_addr*)contact_hdr->uri :
			(pjsip_name_addr*)PJSIP_MSG_FROM_HDR(rdata->msg_info.msg)->uri);
}

#ifdef GRANDSTREAM_NETWORKS
static void remove_same_header_by_name(pjsip_msg *msg, const pj_str_t *name, const struct ast_party_id *id)
{
	pjsip_fromto_hdr *old_hdr = NULL;
	const pjsip_hdr *hdr=msg->hdr.next, *end=&msg->hdr;

	for ( ; hdr!=end; ) {
		if (pj_stricmp(&hdr->name, name) == 0) {
			old_hdr = (pjsip_fromto_hdr *)hdr;
		}

		hdr = hdr->next;

		if(old_hdr) {
			pjsip_name_addr *id_name_addr;
			pjsip_sip_uri *id_uri;

			id_name_addr = (pjsip_name_addr *) old_hdr->uri;
			id_uri = pjsip_uri_get_uri(id_name_addr->uri);

			if (!pj_strcmp2(&id_uri->user, id->number.str)) {
					pj_list_erase(old_hdr);
			}
			old_hdr = NULL;
		}
	}
}
#endif

/*!
 * \internal
 * \brief Adds diversion header information to an outbound SIP message
 *
 * \param tdata The outbound message
 * \param data The redirecting data used to fill parts of the diversion header
 */
static void add_diversion_header(pjsip_tx_data *tdata, struct ast_party_redirecting *data)
{
	pjsip_fromto_hdr *hdr;
	pjsip_name_addr *name_addr;
	pjsip_sip_uri *uri;
	pjsip_param *param;
#ifndef GRANDSTREAM_NETWORKS
	pjsip_fromto_hdr *old_hdr;
#endif
	
	const char *reason_str;
	const char *quote_str;
	char *reason_buf;

	struct ast_party_id *id = &data->from;
	pjsip_uri *base = PJSIP_MSG_FROM_HDR(tdata->msg)->uri;

	if (!id->number.valid || ast_strlen_zero(id->number.str)) {
		return;
	}

#if 0
	if (!strcmp(reason_code_to_str(&data->reason), "unknown")) {
		return;
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug	253588 */
	if (data->reason.code == AST_REDIRECTING_REASON_DEFLECTION) {
		if (!ast_sip_get_send_deflection_div()) {
			return;
		}
	}
#endif

	hdr = pjsip_from_hdr_create(tdata->pool);
	hdr->type = PJSIP_H_OTHER;
	pj_strdup(tdata->pool, &hdr->name, &diversion_name);
	hdr->sname = hdr->name;

	name_addr = pjsip_uri_clone(tdata->pool, base);
	uri = pjsip_uri_get_uri(name_addr->uri);

	pj_strdup2(tdata->pool, &name_addr->display, id->name.str);
	pj_strdup2(tdata->pool, &uri->user, id->number.str);

	param = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param->name = pj_str("reason");

	reason_str = reason_code_to_str(&data->reason);

	/* Reason is either already quoted or it is a token to not need quotes added. */
	quote_str = *reason_str == '\"' || sip_is_token(reason_str) ? "" : "\"";

	reason_buf = pj_pool_alloc(tdata->pool, strlen(reason_str) + 3);
	sprintf(reason_buf, "%s%s%s", quote_str, reason_str, quote_str);/* Safe */

	param->value = pj_str(reason_buf);

	pj_list_insert_before(&hdr->other_param, param);

	hdr->uri = (pjsip_uri *) name_addr;

#ifdef GRANDSTREAM_NETWORKS
	pj_lock_acquire(tdata->lock);
	remove_same_header_by_name(tdata->msg, &diversion_name, id);
	pj_lock_release(tdata->lock);
#else
	old_hdr = pjsip_msg_find_hdr_by_name(tdata->msg, &diversion_name, NULL);
	if (old_hdr) {
		pj_list_erase(old_hdr);
	}
#endif

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
}

static void get_redirecting_add_diversion(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_party_redirecting *data;
#ifdef GRANDSTREAM_NETWORKS
	int  reason_type = -1;
	char *endpoint_id = NULL;
	char *trunk_id  = NULL;
	char num[256] = {0};
	char name[256] = {0};
	int  skip_name = 1;
	struct ast_party_redirecting redirecting;

	if (!session->endpoint->id.send_diversion) {
		ast_log(LOG_NOTICE, "send_diversion is FALSE, do nothing");
		return;
	}

	/*add by chdruan, Bug 35682 - UCM6100 [Feature request] Include Diversion when call forward is enabled*/
	const char* en_diversion = pbx_builtin_getvar_helper(session->channel, "EN_DIVERSION");
	const char* en_reason = pbx_builtin_getvar_helper(session->channel, "CF_REASON");
	if (!ast_strlen_zero(en_reason)) {
		reason_type = atoi(en_reason);
		if (reason_type == AST_REDIRECTING_REASON_DEFLECTION) {
			if ((session->channel != NULL) && (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12))) {
				reason_type = -1;
			}
		}
	}

	if (session->channel != NULL && (reason_type != AST_REDIRECTING_REASON_ATTENDED_TRANSFER) && (reason_type != -1)
			&& ((en_diversion == NULL) || (en_diversion && strcasecmp(en_diversion, "no") != 0 ))) {

		char* callforward  = NULL;
		char* sep          = NULL;
		char* tmp          = NULL;
		int   cf_count       = 0;
		char* cf_all = NULL;

		/* for bug69641 */
		char *cf_user  = NULL;
		char *rt_id    = NULL;
		int  div_count = 0;

		if (!strncmp(ast_sorcery_object_get_id(session->endpoint), "trunk_", 6)) {
			cf_user = (char *)pbx_builtin_getvar_helper(session->channel, "CF_USER"); 
			if (cf_user) {
				endpoint_id = ast_strdupa(ast_sorcery_object_get_id(session->endpoint));
				trunk_id = endpoint_id + strlen("trunk_");

				rt_id = (char *)pbx_builtin_getvar_helper(session->channel, "OUTBOUND_RT_INDEX");
				ast_sip_inner_cid_to_outer_cid(cf_user, trunk_id, rt_id, num, sizeof(num)-1, name, sizeof(name)-1, &skip_name);
			}
		}

		ast_party_redirecting_init(&redirecting);
		data = &redirecting;

		cf_all = (char *)pbx_builtin_getvar_helper(session->channel, "CF_ALL");

		if (cf_all) {
			if (cf_all[0] == '@') {
				cf_all++;
			}
			callforward = ast_strdupa(cf_all);
		}

		ast_log(LOG_NOTICE, "get_redirecting_add_diversion, callforward[%s], divertor[%s]", callforward, num);

		/*Data formats of callforward is like "1000,lucy,4@1001,1001,3@1002,,4"*/
		while ((sep = strsep(&callforward, "@"))) {
			data->from.number.valid = 1;
			cf_count = 0;
			while ((tmp = strsep(&sep, ","))) {
				switch (cf_count) {
				case 0:
					ast_free(data->from.number.str);
					data->from.number.str = NULL;
					if ((div_count == 0) && strlen(num)) {
						data->from.number.str = ast_strdup(num);
					} else {
						data->from.number.str = ast_strdup(tmp);
					}
					break;
				case 1:
					ast_free(data->from.name.str);
					data->from.name.str = NULL;
					if ((div_count == 0) && strlen(num)) {
						if (strlen(name)) {
							data->from.name.str = ast_strdup(name); 
						} else if (skip_name) {
							data->from.name.str = ast_strdup(tmp); 
						}
					} else {
						data->from.name.str = ast_strdup(tmp); 
					}
					break;
				case 2:
					data->reason.code = atoi(tmp);
					break;
				default:
					break;
				}
				cf_count++;
			}
			add_diversion_header(tdata, data);
			div_count++;
		}
		data->count--;

		ast_party_redirecting_free(&redirecting);
	} 

	if (session->channel && (data = ast_channel_redirecting(session->channel))->count) {
		ast_log(LOG_NOTICE, "redirecting count[%d]", data->count);

		/* Added by yfyang, for Bug 129944 */
		if (!strncmp(ast_sorcery_object_get_id(session->endpoint), "trunk_", 6)) {
			char  hint[80] = {0};

			struct ast_party_id *id = &data->from; 
			if (id->number.valid && id->number.str) {
				endpoint_id = ast_strdupa(ast_sorcery_object_get_id(session->endpoint));
				trunk_id = endpoint_id + strlen("trunk_");

				/*! Modified by yfyang, for Bug 178709, check if the id is an extension and get divert id */
				if (0 != ast_get_hint(hint, sizeof(hint), NULL, 0, NULL, "ext-local", id->number.str)) {
					char *rt_id = NULL;

					rt_id = (char *)pbx_builtin_getvar_helper(session->channel, "OUTBOUND_RT_INDEX");
					ast_sip_inner_cid_to_outer_cid(id->number.str, trunk_id, rt_id, num, sizeof(num)-1, name, sizeof(name)-1, &skip_name);
				}

				if (strlen(name)) {
					ast_free(id->name.str);
					id->name.str = ast_strdup(name); 
					id->name.valid = 1;
				}

				if (strlen(num)) {
					ast_free(id->number.str);
					id->number.str = ast_strdup(num); 
					id->number.valid = 1;
				}
			}
		/* Added by yfyang, for Bug 254969 */
		} else {
			struct ast_party_id *id = &data->from; 
			if (id->number.valid && id->number.str) {
				char  family[80]= {0};
				char  data[256] = {0};

				/*! get CIDNAME from astdb /AAUSER/9004/CIDNAME : Dallas */
				snprintf(family, sizeof(family), "AAUSER/%s", id->number.str);
				ast_ucm_db_get(family, "CIDNAME", data, sizeof(data));

				if (strlen(data) != 0) {
					ast_free(id->name.str);
					id->name.str = ast_strdup(data); 
					id->name.valid = 1;
				}
			}
		}

		add_diversion_header(tdata, data);
	}
#else
	if (session->channel && session->endpoint->id.send_diversion &&
	    (data = ast_channel_redirecting(session->channel))->count) {
		add_diversion_header(tdata, data);
	}
#endif
}

/*!
 * \internal
 * \brief Adds a diversion header to an outgoing INVITE request if
 *  redirecting information is available.
 *
 * \param session The session on which the INVITE request is to be sent
 * \param tdata The outbound INVITE request
 */
static void diversion_outgoing_request(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	get_redirecting_add_diversion(session, tdata);
}

/*!
 * \internal
 * \brief Adds a diversion header to an outgoing 3XX response
 *
 * \param session The session on which the INVITE response is to be sent
 * \param tdata The outbound INVITE response
 */
static void diversion_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct pjsip_status_line status = tdata->msg->line.status;

	/* add to 302 and 181 */
	if (PJSIP_IS_STATUS_IN_CLASS(status.code, 300) || (status.code == 181)) {
#ifdef GRANDSTREAM_NETWORKS
		session->is_direct_media = 0;
		ast_log(LOG_DEBUG, "diversion_outgoing_response disable direct media\n");
#endif
		get_redirecting_add_diversion(session, tdata);
	}
#ifdef GRANDSTREAM_NETWORKS
	else {
		pjsip_generic_string_hdr  *hdr_old;

		pj_lock_acquire(tdata->lock);
		while ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &diversion_name, NULL))) {
			pj_list_erase(hdr_old); 
		}
		pj_lock_release(tdata->lock);
	}
#endif
}

static struct ast_sip_session_supplement diversion_supplement = {
	.method = "INVITE",
	/* this supplement needs to be called after caller id
           and after the channel has been created */
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL + 100,
	.incoming_request = diversion_incoming_request,
	.incoming_response = diversion_incoming_response,
	.outgoing_request = diversion_outgoing_request,
	.outgoing_response = diversion_outgoing_response,
	.response_priority = AST_SIP_SESSION_BEFORE_REDIRECTING,
};

static int load_module(void)
{
	ast_sip_session_register_supplement(&diversion_supplement);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&diversion_supplement);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Add Diversion Header Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip,res_pjsip_session",
);
