/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_cli.h"
#include "asterisk/module.h"
#include "asterisk/acl.h"
#include "asterisk/manager.h"
#include "res_pjsip/include/res_pjsip_private.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/linkedlists.h"
#include "asterisk/pbx.h"
#include "asterisk/res_pjsip_session.h"

#include "asterisk/astdb.h"
#include "asterisk/app.h"
#include "asterisk/haserver.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"

AST_MODULE_LOG("pjsip");
#endif

/*** DOCUMENTATION
	<configInfo name="res_pjsip_endpoint_identifier_ip" language="en_US">
		<synopsis>Module that identifies endpoints</synopsis>
		<configFile name="pjsip.conf">
			<configObject name="identify">
				<synopsis>Identifies endpoints via some criteria.</synopsis>
				<description>
					<para>This module provides alternatives to matching inbound requests to
					a configured endpoint. At least one of the matching mechanisms
					must be provided, or the object configuration is invalid.</para>
					<para>The matching mechanisms are provided by the following
					configuration options:</para>
					<enumlist>
						<enum name="match"><para>Match by source IP address.</para></enum>
						<enum name="match_header"><para>Match by SIP header.</para></enum>
					</enumlist>
					<note><para>If multiple matching criteria are provided then an inbound
					request will be matched to the endpoint if it matches
					<emphasis>any</emphasis> of the criteria.</para></note>
				</description>
				<configOption name="endpoint">
					<synopsis>Name of endpoint identified</synopsis>
				</configOption>
				<configOption name="match">
					<synopsis>IP addresses or networks to match against.</synopsis>
					<description>
						<para>The value is a comma-delimited list of IP addresses or
						hostnames.  IP addresses may have a subnet mask appended.  The
						subnet mask may be written in either CIDR or dotted-decimal
						notation.  Separate the IP address and subnet mask with a slash
						('/').
						</para>
					</description>
				</configOption>
				<configOption name="peer_host">
					<synopsis>The original peer host.</synopsis>
					<description>
						<para>The value is a IP addresse or a hostname to indicate a trunk host.  
						</para>
					</description>
                </configOption>
				<configOption name="destination">
					<synopsis>The remote proxy or host if SBC existed.</synopsis>
					<description>
						<para>The value is a IP addresse or a hostname to indicate the destination that SBC will route to .  
						</para>
					</description>
				</configOption>
				<configOption name="srv_lookups" default="yes">
					<synopsis>Perform SRV lookups for provided hostnames.</synopsis>
					<description>
						<para>When enabled, <replaceable>srv_lookups</replaceable> will
						perform SRV lookups for _sip._udp, _sip._tcp, and _sips._tcp of
						the given hostnames to determine additional addresses that traffic
						may originate from.
						</para>
					</description>
				</configOption>
				<configOption name="match_header">
					<synopsis>Header/value pair to match against.</synopsis>
					<description>
						<para>A SIP header whose value is used to match against.  SIP
						requests containing the header, along with the specified value,
						will be mapped to the specified endpoint.  The header must be
						specified with a <literal>:</literal>, as in
						<literal>match_header = SIPHeader: value</literal>.
						</para>
						<para>The specified SIP header value can be a regular
						expression if the value is of the form
						/<replaceable>regex</replaceable>/.
						</para>
						<note><para>Use of a regex is expensive so be sure you need
						to use a regex to match your endpoint.
						</para></note>
					</description>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'identify'.</synopsis>
				</configOption>
			</configObject>
		</configFile>
	</configInfo>
 ***/

/*! \brief The number of buckets for storing hosts for resolution */
#define HOSTS_BUCKETS 53

/*! \brief Structure for an IP identification matching object */
struct ip_identify_match {
	/*! \brief Sorcery object details */
	SORCERY_OBJECT(details);
	/*! \brief Stringfields */
	AST_DECLARE_STRING_FIELDS(
		/*! The name of the endpoint */
		AST_STRING_FIELD(endpoint_name);
		/*! If matching by header, the header/value to match against */
		AST_STRING_FIELD(match_header);
		/*! SIP header name of the match_header string */
		AST_STRING_FIELD(match_header_name);
		/*! SIP header value of the match_header string */
		AST_STRING_FIELD(match_header_value);
#ifdef GRANDSTREAM_NETWORKS
		AST_STRING_FIELD(service);    /*! the service of dnssrv */
		AST_STRING_FIELD(peer_host);  /*! the original peer host */
#endif

#ifdef BUILD_ON_CLOUD_PBX
		AST_STRING_FIELD(destination);  /*! the message target */
		AST_STRING_FIELD(dst_service);    /*! the service of dnssrv */
		AST_STRING_FIELD(dst_to_resolve); /*! the service of dnssrv */
#endif
	);
	/*! Compiled match_header regular expression when is_regex is non-zero */
	regex_t regex_buf;
	/*! \brief Networks or addresses that should match this */
	struct ast_ha *matches;
	/*! \brief Hosts to be resolved when applying configuration */
	struct ao2_container *hosts;
	/*! \brief Perform SRV resolution of hostnames */
	unsigned int srv_lookups;
	/*! Non-zero if match_header has a regular expression (i.e., regex_buf is valid) */
	unsigned int is_regex:1;
#ifdef GRANDSTREAM_NETWORKS
	int sched_id;    /*! the sched to refresh dns resolve*/
	unsigned short a_port;   /*! to record the A/AAAA record port */
#endif

#ifdef BUILD_ON_CLOUD_PBX
	/*! \brief Networks or addresses that should match this for clouducm */
	struct ast_ha *dst_matches;
	unsigned short dst_port;   /*! to record the A/AAAA record port */
#endif
};

#ifdef GRANDSTREAM_NETWORKS

struct srv_entry {
	unsigned short priority;
	unsigned short weight;
	unsigned short port;
	unsigned int weight_sum;
	AST_LIST_ENTRY(srv_entry) list;
	char host[1];
};

struct srv_context {
	unsigned int have_weights:1;
	struct srv_entry *prev;
	unsigned int num_records;
	AST_LIST_HEAD_NOLOCK(srv_entries, srv_entry) entries;
};


static struct ast_sched_context *dns_sched = NULL;

#define IP_IDENTIFY_DNS_REFRESH_INTERVAL   300
#define IP_IDENTIFY_DNS_FIRST_LOOKUP       1

#ifndef BUILD_ON_CLOUD_PBX
static int ip_identify_match_srv_lookup(struct ip_identify_match *identify, const char *prefix, const char *host, int results);
static int ip_identify_match_host_lookup(struct ip_identify_match *identify, const char *host, unsigned short srvport);

#else

struct ip_sbc_matches {
	char domain[128];
	char service[16];
	unsigned short port;
	unsigned short need_resolve;
	int  sched_id;    /*! the sched to refresh dns resolve*/
	struct ast_ha *matches;
};

struct ip_sbc_matches *g_sbc_matches = NULL;

static int ip_identify_dst_match_check(void *obj, void *arg, int flags);
static int ip_identify_dst_match_srv_lookup(struct ip_identify_match *identify, const char *prefix, const char *host, int results);
static int ip_identify_dst_match_host_lookup(struct ip_identify_match *identify, const char *host, unsigned short srvport);
static int ip_identify_check_match_domains(void *obj, void *arg, int flags);
static int ip_identify_sbc_host_lookup(struct ip_sbc_matches *sbc_matches, const char *host, unsigned short srvport);
static int ip_identify_sbc_srv_lookup(struct ip_sbc_matches *sbc_matches, const char *prefix, const char *host, int results);
static int ip_identify_sbc_refresh_timeout(const void *data);
#endif

/* returns   1: network address was successfully converted
             0: is returned if src does not contain a character string representing a valid network address in the specified address family.
            -1: af does not contain a valid address family
*/
static int is_a_valid_IP_address(const char *str)
{
	struct in_addr addr;
	struct in6_addr addr6;
	char host[MAXHOSTNAMELEN] = {0};
	int i = 0;

	if (*str == '[') {
		++str;
		for (; *str && *str != ']';) {
			host[i++] = *str++;
		}
	}

	/* for ipv6 */
	if (inet_pton(AF_INET, str, &addr) < 1) {
		return inet_pton(AF_INET6, host, &addr6);
	}

	return 1;
}

static int get_from_header(pjsip_rx_data *rdata, char *username, size_t username_size, char *domain, size_t domain_size)
{
	pjsip_uri *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_from;

	if (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from)) {
		return -1;
	}
	sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
	ast_copy_pj_str(username, &sip_from->user, username_size);
	ast_copy_pj_str(domain, &sip_from->host, domain_size);
	return 0;
}

static int get_host_info_from_string(const char *content, char *host, int host_len, char *service, int serv_len, unsigned short *port)
{
	char transport[MAXHOSTNAMELEN] = {0};
	char tmp_current[MAXHOSTNAMELEN] = {0};
	char *tmp_ts = NULL;
	char *tmp_port = NULL;
	char *tmp_transport = NULL;

    if (!content) {
        return 0;
    }

    snprintf(tmp_current, MAXHOSTNAMELEN, "%s", content); 

	/* for ipv6 */
	if (tmp_current[0] != '[') {
		tmp_port = strrchr(tmp_current, ':');
	} else {
		tmp_ts = strrchr(tmp_current, ']');
		if (tmp_ts) {
			tmp_port = strrchr(tmp_ts, ':');
		}
	}

	tmp_transport = strstr(tmp_current, ";transport=");

	if (tmp_port == NULL) {
		tmp_ts = tmp_transport ? tmp_transport + strlen(";transport=") : NULL;
		if (tmp_ts && !strcmp(tmp_ts, "TLS")) {
			strncpy(transport, "_sips._tcp", MAXHOSTNAMELEN); 
		} else {
			snprintf(transport, MAXHOSTNAMELEN, "_sip._%s", tmp_ts ? tmp_ts : "udp"); 
		}

		if (tmp_transport != NULL) {
			*tmp_transport = '\0';
		}
	} else {
		*tmp_port = '\0';
		tmp_port++;
		*port = (unsigned short)atoi(tmp_port);
	}

	snprintf(host, host_len, "%s", tmp_current);
	if (strlen(transport)) {
		snprintf(service, serv_len, "%s", transport);
	}

	ast_log(LOG_NOTICE, "value[%s] => host[%s], service[%s], port[%d]", content, host, strlen(service) ? service:" ", *port);
	return 0;
}


static struct ast_sip_endpoint *ip_identify_match_contact(const char* contact_str)
{
	struct ast_sip_endpoint *pstEndpoint = NULL;
	struct ao2_iterator      j;
	struct ast_sip_contact  *contact;
	int    found_endpoint = 0;

	char *contact_id     = NULL;
	char *uri_start      = NULL;
	char endpoint_id[80] = {0};
	char contact_uri[80] = {0};
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);

	if (!(contacts = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "contact", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(contacts)) {
		ast_log(LOG_NOTICE, "No contact sections to match against\n");
		return NULL;
	}

	j = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_t_iterator_next(&j, "iterate contacts table"))) {

		if (ast_strlen_zero(contact->uri)) {
			ao2_ref(contact, -1);
			continue;
		}

        contact_id = ast_strdupa(contact->uri);
		uri_start = strstr(contact_id, "sip:");
		if (!uri_start) {
			uri_start = strstr(contact_id, "sips:");
			if (uri_start) {
				uri_start = uri_start + strlen("sips:");
			}
		} else {
			uri_start = uri_start + strlen("sip:");
		}

		if (uri_start) {
			sscanf(uri_start, "%79[^@]@%79[^;]", endpoint_id, contact_uri);
		}

		if (!strlen(contact_uri)) {
			ao2_ref(contact, -1);
			continue;
		}

		if (!strcmp(contact_str, contact_uri)) {
			sscanf(endpoint_id, "%79[^;]", endpoint_id);
			found_endpoint = 1;
			ao2_ref(contact, -1);
			break;
		}

		ao2_ref(contact, -1);
	}
	ao2_iterator_destroy(&j);

	if (found_endpoint) {
		pstEndpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpoint_id);
		return pstEndpoint;
	}
	return NULL;
}

/*! \brief do dns resolve every 5 mins */
static int ip_identify_refresh_timeout(const void *data)
{
	struct ip_identify_match *identify = (struct ip_identify_match *)data;
	struct ip_identify_match *match;

	int    found      = 0;
	struct ao2_iterator j;
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);

#ifndef BUILD_ON_CLOUD_PBX
	int  str_len = 0;
	char addr_buf[512] = {0};
	struct ast_ha *ha;
	char *current_string;
	struct ao2_iterator i;
#endif

	if (NULL == data) {
		return 0;
	}

	if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections to refresh\n");
		return 0;
	}

	j = ao2_iterator_init(candidates, 0);
	while ((match = ao2_t_iterator_next(&j, "iterate identify table"))) {
		if (match == identify) {
			found = 1;
			ao2_ref(match, -1);
			break;
		}
		ao2_ref(match, -1);
	}
	ao2_iterator_destroy(&j);

	if (!found) {
		ast_log(LOG_WARNING, "No identify matched to refresh\n");
		return 0;
	}

	if (!identify->hosts) {
#ifdef BUILD_ON_CLOUD_PBX
		if (ast_strlen_zero(identify->dst_to_resolve)) {
			return 0;
		}
#else
		return 0;
#endif
	}


	ao2_ref(identify, +1);

#ifndef BUILD_ON_CLOUD_PBX
	/* Modified by yfyang, not do dns query when standby, for HA service data sync */
	if (HA_ROLE_STANDBY != ast_ha_get_role()) {
		/* Resolve the match addresses now */
		i = ao2_iterator_init(identify->hosts, 0);
		while ((current_string = ao2_iterator_next(&i))) {
			struct ast_sockaddr address;
			int results = 0;

			/* If the provided string is not an IP address perform SRV resolution on it */
			if (identify->srv_lookups && !ast_sockaddr_parse(&address, current_string, 0)) {
				if (!ast_strlen_zero(identify->service)) {
					results = ip_identify_match_srv_lookup(identify, identify->service, current_string, results);
				}
			}

			/* If SRV fails fall back to a normal lookup on the host itself */
			if (!results) {
				results = ip_identify_match_host_lookup(identify, current_string, identify->a_port);
			}

			if (results == 0) {
				ast_log(LOG_WARNING, "Identify '%s' provided address '%s' did not resolve to any address\n", ast_sorcery_object_get_id(identify), current_string);
			} else if (results == -1) {
				ast_log(LOG_WARNING, "Identify '%s' failed when adding resolution results of '%s'\n", ast_sorcery_object_get_id(identify), current_string);
			}

			ao2_ref(current_string, -1);
		}
		ao2_iterator_destroy(&i);

		/* BEGIN: Added by yfyang, for HA service data sync */
		if (HA_ROLE_ACTIVE == ast_ha_get_role()) {
			ao2_lock(identify);
			for (ha = identify->matches; ha; ha = ha->next) {
				if (!str_len) {
					snprintf(addr_buf, sizeof(addr_buf), "%s:%s:%d", ha->hostname, ast_sockaddr_stringify_addr(&ha->addr), ha->port);
				} else {
					snprintf(addr_buf+str_len, sizeof(addr_buf)-str_len, ",%s:%s:%d", ha->hostname, ast_sockaddr_stringify_addr(&ha->addr), ha->port);
				}

				str_len = strlen(addr_buf);
			}
			ao2_unlock(identify);
			if (str_len) {
				ast_db_put("Identify/matches", ast_sorcery_object_get_id(identify), addr_buf);
			}
		}
		/* END: Added by yfyang, for HA service data sync */
	}
#endif

#ifdef BUILD_ON_CLOUD_PBX
    if (!ast_strlen_zero(identify->dst_to_resolve)) {
			struct ast_sockaddr addr;
			int dst_results = 0;

			/* If the provided string is not an IP address perform SRV resolution on it */
			if (!identify->dst_port && !ast_strlen_zero(identify->dst_service) && !ast_sockaddr_parse(&addr, identify->dst_to_resolve, 0)) {
				dst_results = ip_identify_dst_match_srv_lookup(identify, identify->dst_service, identify->dst_to_resolve, dst_results);
			}

			/* If SRV fails fall back to a normal lookup on the host itself */
			if (!dst_results) {
				dst_results = ip_identify_dst_match_host_lookup(identify, identify->dst_to_resolve, identify->dst_port);
			}

			if (dst_results == 0) {
				ast_log(LOG_WARNING, "Identify '%s' resovle for '%s' did not resolve to any address\n", ast_sorcery_object_get_id(identify), identify->dst_to_resolve);
			} else if (dst_results == -1) {
				ast_log(LOG_WARNING, "Identify '%s' resovle failed for '%s'\n", ast_sorcery_object_get_id(identify), identify->dst_to_resolve);
			}
		}
#endif

	identify->sched_id = ast_sched_add(dns_sched, IP_IDENTIFY_DNS_REFRESH_INTERVAL*1000, ip_identify_refresh_timeout, identify);
	ao2_ref(identify, -1);
	return 0;
}

static struct ast_ha *ip_identify_get_dns_addr(pj_str_t* host)
{
#ifndef BUILD_ON_CLOUD_PBX
	struct ao2_iterator j;
	struct ao2_iterator i;
	struct ip_identify_match *match;
	char  *current_str = NULL;
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);
#endif
	struct ast_ha *result = NULL;
	int    found          = 0;

	if (!host || !host->slen || !host->ptr) {
		return NULL;
	}

#ifdef BUILD_ON_CLOUD_PBX
	if (!g_sbc_matches) {
		return NULL;
	}

	if (!strncmp(g_sbc_matches->domain, host->ptr, host->slen)) {
		result = ast_duplicate_ha_list(g_sbc_matches->matches);
		found = 1;
	}
#else
	if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections\n");
		return NULL;
	}

	j = ao2_iterator_init(candidates, 0);
	while ((match = ao2_t_iterator_next(&j, "iterate identify table"))) {
		if (!match->hosts || !ao2_container_count(match->hosts)) {
			continue;
		}

		i = ao2_iterator_init(match->hosts, 0); 
		while ((current_str = ao2_t_iterator_next(&i, "iterate host list"))) {
			ast_log(LOG_NOTICE, "current_str[%s]\n", current_str);
			if (!strncmp(current_str, host->ptr, host->slen)) {
				result = ast_duplicate_ha_list(match->matches);
				ao2_ref(current_str, -1);
				found = 1;
				break;
			}

			ao2_ref(current_str, -1);
		}
		ao2_iterator_destroy(&i);

		ao2_ref(match, -1);
		if (found) {
			break;
		}
	}
	ao2_iterator_destroy(&j);
#endif

	if (!found) {
		ast_log(LOG_NOTICE, "No identify matched to refresh\n");
		return NULL;
	}

	return result;
}

#ifdef BUILD_ON_CLOUD_PBX
static void ip_identify_get_addrs_by_host(char *host, char *addrs, int len)
{
	struct ao2_iterator j;
	struct ip_identify_match *ident;
	struct ast_ha *match;
	char  *pos = addrs;
	char  *end = NULL;
	int    l = 0;
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);

	if (!host || !addrs || !len) {
		return;
	}
	
	if (1 == is_a_valid_IP_address(host)) {
		return;
	}

    if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
		!ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections\n");
		return;
	}

	end = strchr(host, ':');
	if (end) {
		*end = '\0';  /* remove port part */
		if (1 == is_a_valid_IP_address(host)) {
			return;
		}
	}

	j = ao2_iterator_init(candidates, 0);
	while ((ident = ao2_t_iterator_next(&j, "iterate identify table"))) {
		if (strcmp(ident->destination, host)) {
			ao2_ref(ident, -1);
			continue;
		}

		for (match = ident->dst_matches; match; match = match->next) {
			const char *addr = ast_sockaddr_stringify_addr(&match->addr);
			if (match->port > 0) {
				snprintf(pos, len - l, "%s:%d,", addr, match->port);
			} else {
				snprintf(pos, len - l, "%s,", addr);
			}

			l = strlen(addrs);
			pos = addrs+l;
		}

		ao2_ref(ident, -1);

        if (l) {
            addrs[l-1] = '\0';
            ast_log(LOG_NOTICE, "get_addrs_by_host, [%s] => [%s]\n", host, addrs);
        }
        break;
	}
	ao2_iterator_destroy(&j);

	return;
}

static void ip_identify_set_used_by_addr(char *endpoint_id, char *addr)
{
	if (!endpoint_id || !addr) {
		return;
	}

	if (!g_sbc_matches) {
		return;
	}

	ast_log(LOG_NOTICE, "ip_identify_set_used_by_addr, endpoint[%s], addr[%s]\n", endpoint_id, addr);
	ast_ha_set_used_by_addr(g_sbc_matches->matches, addr);
	return;
}
#else
static void ip_identify_get_addrs_by_host(char *host, char *addrs, int len)
{
    return;  //do nothing.
}

static void ip_identify_set_used_by_addr(char *endpoint_id, char *addr)
{
	RAII_VAR(struct ip_identify_match *, identify, NULL, ao2_cleanup);
	char identify_id[80] = {0};

	if (!endpoint_id || !addr) {
		return;
	}

	snprintf(identify_id, sizeof(identify_id), "id_%s", endpoint_id);
	identify = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "identify", identify_id);
	if (!identify) {
		ast_log(LOG_NOTICE, "No identify[%s] object.\n", identify_id);
		return;
	}

	ast_log(LOG_NOTICE, "ip_identify_set_used_by_addr, endpoint[%s], addr[%s]\n", endpoint_id, addr);
	ast_ha_set_used_by_addr(identify->matches, addr);
	return;
}
#endif

static struct ast_sip_resolve_identifier ip_resolve_identify = {
	.get_dns_resolve_addr = ip_identify_get_dns_addr,
	.set_used_by_addr = ip_identify_set_used_by_addr,
    .get_addrs_by_host = ip_identify_get_addrs_by_host,
};

#endif

/*! \brief Destructor function for a matching object */
static void ip_identify_destroy(void *obj)
{
	struct ip_identify_match *identify = obj;

	ast_string_field_free_memory(identify);
	ast_free_ha(identify->matches);
	ao2_cleanup(identify->hosts);
	if (identify->is_regex) {
		regfree(&identify->regex_buf);
	}

#ifdef GRANDSTREAM_NETWORKS
	AST_SCHED_DEL(dns_sched, identify->sched_id);
#endif
#ifdef BUILD_ON_CLOUD_PBX
	ast_free_ha(identify->dst_matches);
#endif
}

/*! \brief Allocator function for a matching object */
static void *ip_identify_alloc(const char *name)
{
	struct ip_identify_match *identify = ast_sorcery_generic_alloc(sizeof(*identify), ip_identify_destroy);

	if (!identify || ast_string_field_init(identify, 256)) {
		ao2_cleanup(identify);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	identify->a_port = 0;
#endif

	return identify;
}

/*! \brief Comparator function for matching an object by header */
static int header_identify_match_check(void *obj, void *arg, int flags)
{
	struct ip_identify_match *identify = obj;
	struct pjsip_rx_data *rdata = arg;
	pjsip_hdr *header;
	pj_str_t pj_header_name;
	int header_present;

	if (ast_strlen_zero(identify->match_header)) {
		return 0;
	}

	pj_header_name = pj_str((void *) identify->match_header_name);

	/* Check all headers of the given name for a match. */
	header_present = 0;
	for (header = NULL;
		(header = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &pj_header_name, header));
		header = header->next) {
		char *pos;
		int len;
		char buf[PATH_MAX];

		header_present = 1;

		/* Print header line to buf */
		len = pjsip_hdr_print_on(header, buf, sizeof(buf) - 1);
		if (len < 0) {
			/* Buffer not large enough or no header vptr! */
			ast_assert(0);
			continue;
		}
		buf[len] = '\0';

		/* Remove header name from pj_buf and trim blanks. */
		pos = strchr(buf, ':');
		if (!pos) {
			/* No header name?  Bug in PJPROJECT if so. */
			ast_assert(0);
			continue;
		}
		pos = ast_strip(pos + 1);

		/* Does header value match what we are looking for? */
		if (identify->is_regex) {
			if (!regexec(&identify->regex_buf, pos, 0, NULL, 0)) {
				return CMP_MATCH;
			}
		} else if (!strcmp(identify->match_header_value, pos)) {
			return CMP_MATCH;
		}

		ast_debug(3, "Identify '%s': SIP message has '%s' header but value '%s' does not match '%s'.\n",
			ast_sorcery_object_get_id(identify),
			identify->match_header_name,
			pos,
			identify->match_header_value);
	}
	if (!header_present) {
		ast_debug(3, "Identify '%s': SIP message does not have '%s' header.\n",
			ast_sorcery_object_get_id(identify),
			identify->match_header_name);
	}
	return 0;
}

#ifndef BUILD_ON_CLOUD_PBX
/*! \brief Comparator function for matching an object by IP address */
static int ip_identify_match_check(void *obj, void *arg, int flags)
{
	struct ip_identify_match *identify = obj;
	struct ast_sockaddr *addr = arg;
	int sense;

	sense = ast_apply_ha(identify->matches, addr);
#ifndef GRANDSTREAM_NETWORKS
	if (sense != AST_SENSE_ALLOW) 
#else
	if (sense == AST_SENSE_ALLOW) 
#endif
     {
		ast_log(LOG_DEBUG, "Source address %s matches identify '%s'\n",
				ast_sockaddr_stringify(addr),
				ast_sorcery_object_get_id(identify));
		return CMP_MATCH;
	} else {
		ast_log(LOG_DEBUG, "Source address %s does not match identify '%s'\n",
				ast_sockaddr_stringify(addr),
				ast_sorcery_object_get_id(identify));
		return 0;
	}
}

static struct ast_sip_endpoint *ip_identify_match_fromdomain(pjsip_rx_data *rdata, struct ao2_container *candidates)
{
	if (!rdata || !candidates) {
		return NULL;
	}

	struct ip_identify_match *match = NULL;
	struct ast_sip_endpoint  *endpoint = NULL;
	int found_domain = 0;
	pj_str_t host;
	struct ao2_iterator it_matches_domain = ao2_iterator_init(candidates, 0);

	/* Get from-domain */
	pjsip_uri  *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_uri = pjsip_uri_get_uri(from);
	if (!from || !sip_uri || (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from))) {
		return NULL;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &host, &sip_uri->host);
	while ((match = ao2_iterator_next(&it_matches_domain))) {
		if (ast_strlen_zero(match->peer_host)) {
			ao2_cleanup(match);
			continue;
		}

		if (!pj_strcmp2(&host, match->peer_host)) {
			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);
			if (!endpoint) {
				ao2_cleanup(match);
				continue;
			}

			ast_log(LOG_NOTICE, "Found endpoint[%s] by from_domain[%s]", match->endpoint_name, match->peer_host);
			found_domain = 1;
			ao2_cleanup(match);
			break;
		}

		ao2_cleanup(match);
	}

	ao2_iterator_destroy(&it_matches_domain);
	if (found_domain) {
		return endpoint;
	}

	return NULL;
}
#endif

#ifdef BUILD_ON_CLOUD_PBX
static struct ast_sip_endpoint *ip_identify_match_fromdomains(pjsip_rx_data *rdata, struct ao2_container *candidates, char *domain, int *num)
{
	if (!rdata || !candidates || !domain || !num) {
		return NULL;
	}

	struct ip_identify_match *match = NULL;
	struct ast_sip_endpoint  *endpoint = NULL;
	struct ast_sip_endpoint  *found_endpoint = NULL;
	int found_domain = 0;
	int id_count = 0;
	struct ao2_iterator it_matches_domain = ao2_iterator_init(candidates, 0);

	while ((match = ao2_iterator_next(&it_matches_domain))) {
		if (ast_strlen_zero(match->peer_host)) {
			ao2_cleanup(match);
			continue;
		}

		if (!strcmp(domain, match->peer_host)) {
			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);
			if (!endpoint) {
				ao2_cleanup(match);
				continue;
			}

			ast_log(LOG_DEBUG, "Found endpoint[%s] by from_domain[%s]", match->endpoint_name, match->peer_host);
			found_domain = 1;
			ao2_cleanup(match);

			if (strncmp(endpoint->trunk_type, "register", 8)) {
				if (found_endpoint) {
					ao2_cleanup(found_endpoint);
				}
				found_endpoint = endpoint;
				break; /*peer trunk, matched*/
			}

			id_count++;
			if ((id_count > 1) && found_endpoint) {
				ast_log(LOG_DEBUG, "Found %d endpoints, ignore", id_count);
				ao2_cleanup(endpoint);
				ao2_cleanup(found_endpoint);

				found_domain = 0;
				break; /* more than one register trunk, ignore */
			}

			found_endpoint = endpoint;
		} else {
			ao2_cleanup(match);
		}
	}

	ao2_iterator_destroy(&it_matches_domain);
	if (found_domain) {
		return found_endpoint;
	}

    *num = id_count;
	return NULL;
}
#endif

static struct ast_sip_endpoint *common_identify(ao2_callback_fn *identify_match_cb, void *arg)
{
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);
	struct ip_identify_match *match;
	struct ast_sip_endpoint *endpoint;

	/* If no possibilities exist return early to save some time */
	candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify",
		AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL);
	if (!candidates || !ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections to match against\n");
		return NULL;
	}

	match = ao2_callback(candidates, 0, identify_match_cb, arg);
	if (!match) {
		return NULL;
	}

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
		match->endpoint_name);
	if (endpoint) {
		ast_debug(3, "Identify '%s' SIP message matched to endpoint %s\n",
			ast_sorcery_object_get_id(match), match->endpoint_name);
	} else {
		ast_log(LOG_WARNING, "Identify '%s' points to endpoint '%s' but endpoint could not be found\n",
			ast_sorcery_object_get_id(match), match->endpoint_name);
	}

	ao2_ref(match, -1);
	return endpoint;
}

#ifdef BUILD_ON_CLOUD_PBX
static struct ast_sip_endpoint *ip_identify(pjsip_rx_data *rdata)
{
	struct ast_sockaddr addr = { { 0, } };
	struct ast_sip_endpoint  *endpoint = NULL;

	/*1. get GS-GDMS-Endpoint-Address header */
	pj_str_t  addr_hdr_str = { "GS-GDMS-Endpoint-Address", 24 };
	pj_str_t  sbcaddr_hdr_str = { "SBC-Endpoint-Address", 20 };
    pj_str_t  domain;
	pjsip_generic_string_hdr* gdms_addr_hdr = NULL;
	char host[PJ_INET6_ADDRSTRLEN] = { 0 };
	int  domain_count = 0;

	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG &&
		rdata->msg_info.msg->line.req.method.id == PJSIP_REGISTER_METHOD) {
		ast_log(LOG_DEBUG, "It's a REGISTER method, don't use ip identify\n");
		return NULL;
	}

	/* Get from-domain */
	pjsip_uri  *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_uri = pjsip_uri_get_uri(from);
	if (!from || !sip_uri || (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from))) {
		return NULL;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &domain, &sip_uri->host);

	if (!(gdms_addr_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &sbcaddr_hdr_str, NULL))) {
		gdms_addr_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &addr_hdr_str, NULL);
	}

	if (gdms_addr_hdr && gdms_addr_hdr->hvalue.slen) {
		ast_copy_pj_str(host, &gdms_addr_hdr->hvalue, PJ_INET6_ADDRSTRLEN);

		if ((endpoint = ast_sip_get_accounttk_endpoint_by_address(host))) {
			return endpoint;
		}

		ast_sockaddr_parse(&addr, host, PARSE_ADDR);
		ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);
	} else {
		ast_sockaddr_parse(&addr, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);
    }

	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, account_tks, NULL, ao2_cleanup);
	struct ip_identify_match *match = NULL;
	struct ao2_iterator      *it_matches = NULL;
	char  *did_num     = NULL;
	int    is_inv_req  = 0;
	struct ast_sip_cc_original_exten_data exten_data;
	char endpoint_name[80] = {0};
	char domain_name[80]   = {0};
	char to_num[80]        = {0};
	char req_num[80]       = {0};

    get_from_header(rdata, endpoint_name, sizeof(endpoint_name), domain_name, sizeof(domain_name));

	/* Add by bfdu for Bug 113319, 2018.06.22 */
	if (rdata && rdata->tp_info.transport && rdata->msg_info.cseq && 
		(!strcasecmp(rdata->tp_info.transport->type_name, "WS") || !strcasecmp(rdata->tp_info.transport->type_name, "WSS")) &&
		(0 == pj_stricmp2(&rdata->msg_info.cseq->method.name, "INVITE"))) {
		ast_log(LOG_DEBUG, "It's a webrtc call, don't use ip identify\n");
		return NULL;
	}

	/* If no possibilities exist return early to save some time */
	if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
	    !ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections to match against\n");

		return ast_sip_get_accounttk_endpoint_by_address(ast_sockaddr_stringify(&addr));
	}

	/*2. check dst_matches */
	if (!(it_matches = ao2_callback(candidates, OBJ_MULTIPLE, ip_identify_dst_match_check, &addr)) ||
		 (it_matches && ao2_iterator_count(it_matches) == 0)) 
    {
		ast_debug(3, "'%s' did not match any identify section rules\n", ast_sockaddr_stringify_addr(&addr));
		if (it_matches) {
			ao2_iterator_destroy(it_matches);
			it_matches = NULL;
		}

		/* check from domain */
		if ((endpoint = ip_identify_match_fromdomains(rdata, candidates, domain_name, &domain_count))) {
			ast_debug(3, "Retrieved endpoint %s.\n", ast_sorcery_object_get_id(endpoint));
			return endpoint;
		}

		if (domain_count > 1) {
			goto check_did;
		}

		if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG && 
			((rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) || !pj_stricmp2(&rdata->msg_info.cseq->method.name, "SUBSCRIBE"))) {

			endpoint = ip_identify_match_contact(ast_sockaddr_stringify(&addr));
			if (endpoint) {
				if (strcasecmp(endpoint_name, "anonymous") == 0) {
					ast_string_field_set(endpoint, anonymous, "anonymous");
					ast_log(LOG_NOTICE, "Set anonymous to endpoint[%s].", ast_sorcery_object_get_id(endpoint));
				}
				ast_debug(3, "Retrieved endpoint %s from contact\n", ast_sorcery_object_get_id(endpoint));
				return endpoint;
			}
		}

		return ast_sip_get_accounttk_endpoint_by_address(ast_sockaddr_stringify(&addr));
	}

	if (ao2_iterator_count(it_matches) > 1) {
		if ((endpoint = ip_identify_match_fromdomains(rdata, candidates, domain_name, &domain_count))) {
			ao2_iterator_destroy(it_matches);
			return endpoint;
		}
	}

check_did:
	if (domain_count > 1) {
		if (!(it_matches = ao2_callback(candidates, OBJ_MULTIPLE, ip_identify_check_match_domains, domain_name))) {
			ast_debug(3, "Can't match identify by domain[%s].\n", domain_name);
			return NULL;
		}
	}

	/*! get to-user and request-user from the INVITE request */
    if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG &&
		rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) {

		pjsip_uri      *to   = NULL;
		pjsip_uri      *ruri = NULL;
		pjsip_sip_uri  *uri  = NULL;
		pjsip_sip_uri  *sip_ruri = NULL;
		pjsip_tel_uri  *tel_ruri = NULL;

		/* Get to-user */
		to = rdata->msg_info.to->uri;
		uri = pjsip_uri_get_uri(to);

		if (PJSIP_URI_SCHEME_IS_SIP(to) || PJSIP_URI_SCHEME_IS_SIPS(to)) {
			ast_copy_pj_str(to_num, &uri->user, sizeof(to_num));
		} else if (PJSIP_URI_SCHEME_IS_TEL(to)) {
			tel_ruri = (pjsip_tel_uri *)uri;
			ast_copy_pj_str(to_num, &tel_ruri->number, sizeof(to_num));
		}

		/* Get request-user */
		ruri = rdata->msg_info.msg->line.req.uri;
		sip_ruri = pjsip_uri_get_uri(ruri);

		if (PJSIP_URI_SCHEME_IS_SIP(ruri) || PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
			ast_copy_pj_str(req_num, &sip_ruri->user, sizeof(req_num));
		} else if (PJSIP_URI_SCHEME_IS_TEL(ruri)) {
			tel_ruri = (pjsip_tel_uri *)sip_ruri;
			ast_copy_pj_str(req_num, &tel_ruri->number, sizeof(req_num));
		}

		is_inv_req = 1;
	}

	ast_log(LOG_NOTICE, "ip_identify, match count[%d], is_inv_req[%d]\n",  ao2_iterator_count(it_matches), is_inv_req);
	for (; (match = ao2_iterator_next(it_matches)); ao2_cleanup(match)) {
		ast_log(LOG_NOTICE, "ip_identify, check match->endpoint_name[%s], check To user\n", match->endpoint_name);

		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);
		if (!endpoint) {
			continue;
		}

		if (!is_inv_req) {
			ao2_cleanup(match);
			break;
		}

		/*! peer trunk, don't need to check the to header */
		if (ast_strlen_zero(endpoint->username) || strncmp(endpoint->trunk_type, "register", 8)) {
			ao2_cleanup(endpoint);
			endpoint = NULL;
			continue;
		}

		if (strcasecmp(endpoint->didmode, "to-header") == 0) {
			did_num = to_num;
		} else {
			did_num = req_num;
		}

		if (!strncmp(endpoint->username, did_num, sizeof(to_num))) {
			ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], To user[%s] matched username\n", ast_sorcery_object_get_id(endpoint), did_num);
			ao2_cleanup(match);
			break;  /*! to matched */
		}

		ao2_cleanup(endpoint);
		endpoint = NULL;
	}
	ao2_iterator_destroy(it_matches);

	if (endpoint) {
		ast_log(LOG_NOTICE, "Retrieved endpoint %s by To user\n", ast_sorcery_object_get_id(endpoint));
		return endpoint; 
	}

	if (!(it_matches = ao2_callback(candidates, OBJ_MULTIPLE, ip_identify_dst_match_check, &addr))) {
		ast_debug(3, "Can't match identify by addr[%s].\n", ast_sockaddr_stringify_addr(&addr));
		return NULL;
	}

	for (; (match = ao2_iterator_next(it_matches)); ao2_cleanup(match)) {
		char diversion_user[80] = {0};

		ast_log(LOG_NOTICE, "ip_identify, check match->endpoint_name[%s], check DID\n", match->endpoint_name);
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);

		if (!endpoint) {
			continue;
		} 

		if (strcasecmp(endpoint->didmode, "to-header") == 0) {
			did_num = to_num;
		} else {
			did_num = req_num;
		}

		if (ast_exists_extension(NULL, endpoint->context, did_num, 1, "NO_NEED_MATCH_CALLERID")) {
			ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], To user[%s] matched DID mode\n", ast_sorcery_object_get_id(endpoint), did_num);
			ao2_cleanup(match);
			break;  /*! did matched */
		}

		/* Added by yfyang, for Bug 234702 */
		if (!ast_sip_get_diversion_user(rdata, diversion_user, sizeof(diversion_user))) {
			if (ast_exists_extension2(NULL, endpoint->context, did_num, 1, diversion_user, 1)) {
				ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], DID user[%s], DiversionID[%s] matched\n", ast_sorcery_object_get_id(endpoint), did_num, diversion_user);
				ao2_cleanup(match);
				break;  /*! did matched */
			}
		}

		if (0 == ast_sip_find_sub_handler_for_supplement_exten(AST_SIP_CCSS_ORIGINAL_EXTEN, did_num, &exten_data)) {
			char *device = exten_data.device_name + strlen("PJSIP/");
			/* Added by yfyang, for Bug292944 */
			char *ptr = strchr(device, '/');
			if (ptr) {
				*ptr = '\0';
			}

			if (!strncmp(device,  ast_sorcery_object_get_id(endpoint), AST_MAX_EXTENSION)) {
				ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], DID user[%s] matched cc-user\n", ast_sorcery_object_get_id(endpoint), did_num);
				ao2_cleanup(match);
				break;  /*! did matched */
			}
		}

		ao2_cleanup(endpoint);
		endpoint = NULL;
	}
	ao2_iterator_destroy(it_matches);

	if (endpoint) {
		ast_log(LOG_NOTICE, "Retrieved endpoint %s by DID mode \n", ast_sorcery_object_get_id(endpoint));
	}

	return endpoint; 
}
#else
static struct ast_sip_endpoint *ip_identify(pjsip_rx_data *rdata)
{
	struct ast_sockaddr addr = { { 0, } };

	ast_sockaddr_parse(&addr, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
	ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);

#ifndef GRANDSTREAM_NETWORKS
	return common_identify(ip_identify_match_check, &addr);
#else
	RAII_VAR(struct ao2_container *, candidates, NULL, ao2_cleanup);
	struct ip_identify_match *match = NULL;
	struct ast_sip_endpoint  *endpoint = NULL;
	struct ao2_iterator      *it_matches;
	struct ao2_iterator      *it_matches_did;
	char  *did_num     = NULL;
	int    is_inv_req  = 0;
	struct ast_sip_cc_original_exten_data exten_data;
	char endpoint_name[80] = {0};
	char domain_name[80]   = {0};
	char to_num[80]        = {0};
	char req_num[80]       = {0};

	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG &&
		rdata->msg_info.msg->line.req.method.id == PJSIP_REGISTER_METHOD) {
		ast_log(LOG_DEBUG, "It's a REGISTER method, don't use ip identify\n");
		return NULL;
	}

	pj_str_t  addr_hdr_str = { "contact-original-address", 24 };
	pjsip_generic_string_hdr* gdms_addr_hdr = NULL;
	char host[PJ_INET6_ADDRSTRLEN] = { 0 };

	if ((gdms_addr_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &addr_hdr_str, NULL))) {
		if (gdms_addr_hdr->hvalue.slen) {
			ast_copy_pj_str(host, &gdms_addr_hdr->hvalue, PJ_INET6_ADDRSTRLEN);
			ast_sockaddr_parse(&addr, host, PARSE_PORT_FORBID);
			ast_sockaddr_set_port(&addr, rdata->pkt_info.src_port);
		}
	}

    get_from_header(rdata, endpoint_name, sizeof(endpoint_name), domain_name, sizeof(domain_name));

	/* Add by bfdu for Bug 113319, 2018.06.22 */
	if (rdata && rdata->tp_info.transport && rdata->msg_info.cseq && 
		(!strcasecmp(rdata->tp_info.transport->type_name, "WS") || !strcasecmp(rdata->tp_info.transport->type_name, "WSS")) &&
		(0 == pj_stricmp2(&rdata->msg_info.cseq->method.name, "INVITE"))) {
		ast_log(LOG_DEBUG, "It's a webrtc call, don't use ip identify\n");
		return NULL;
	}

	/* If no possibilities exist return early to save some time */
	if (!(candidates = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL)) ||
	    !ao2_container_count(candidates)) {
		ast_debug(3, "No identify sections to match against\n");

		if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG && 
		    ((rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) || !pj_stricmp2(&rdata->msg_info.cseq->method.name, "SUBSCRIBE"))) {

			endpoint = ip_identify_match_contact(ast_sockaddr_stringify(&addr));
			if (endpoint) {
				if (strcasecmp(endpoint_name, "anonymous") == 0) {
					ast_string_field_set(endpoint, anonymous, "anonymous");
					ast_log(LOG_NOTICE, "Set anonymous to endpoint[%s].", ast_sorcery_object_get_id(endpoint));
				}
				ast_debug(3, "Retrieved endpoint %s from contact\n", ast_sorcery_object_get_id(endpoint));
				return endpoint;
			}
		}

		return ast_sip_get_accounttk_endpoint_by_address(ast_sockaddr_stringify(&addr));
	}

	if (!(it_matches = ao2_callback(candidates, OBJ_MULTIPLE, ip_identify_match_check, &addr)) ||
		(it_matches && ao2_iterator_count(it_matches) == 0)) {
		ast_debug(3, "'%s' did not match any identify section rules\n", ast_sockaddr_stringify(&addr));
		if (it_matches) {
			ao2_iterator_destroy(it_matches);
		}

		/* check from domain, to support trunk through GDMS */
		if (strcasecmp(rdata->tp_info.transport->type_name, "UDP")) {
			if ((endpoint = ip_identify_match_fromdomain(rdata, candidates))) {
				return endpoint;
			}
		}

		if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG && 
			((rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) || !pj_stricmp2(&rdata->msg_info.cseq->method.name, "SUBSCRIBE"))) {

			endpoint = ip_identify_match_contact(ast_sockaddr_stringify(&addr));
			if (endpoint) {
				if (strcasecmp(endpoint_name, "anonymous") == 0) {
					ast_string_field_set(endpoint, anonymous, "anonymous");
					ast_log(LOG_NOTICE, "Set anonymous to endpoint[%s].", ast_sorcery_object_get_id(endpoint));
				}
				ast_debug(3, "Retrieved endpoint %s from contact\n", ast_sorcery_object_get_id(endpoint));
				return endpoint;
			}
		}

		return ast_sip_get_accounttk_endpoint_by_address(ast_sockaddr_stringify(&addr));
	}

	/*! get to-user and request-user from the INVITE request */
    if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG &&
		rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) {

		pjsip_uri      *to   = NULL;
		pjsip_uri      *ruri = NULL;
		pjsip_sip_uri  *uri  = NULL;
		pjsip_sip_uri  *sip_ruri = NULL;
		pjsip_tel_uri  *tel_ruri = NULL;

		/* Get to-user */
		to = rdata->msg_info.to->uri;
		uri = pjsip_uri_get_uri(to);

		if (PJSIP_URI_SCHEME_IS_SIP(to) || PJSIP_URI_SCHEME_IS_SIPS(to)) {
			ast_copy_pj_str(to_num, &uri->user, sizeof(to_num));
		} else if (PJSIP_URI_SCHEME_IS_TEL(to)) {
			tel_ruri = (pjsip_tel_uri *)uri;
			ast_copy_pj_str(to_num, &tel_ruri->number, sizeof(to_num));
		}

		/* Get request-user */
		ruri = rdata->msg_info.msg->line.req.uri;
		sip_ruri = pjsip_uri_get_uri(ruri);

		if (PJSIP_URI_SCHEME_IS_SIP(ruri) || PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
			ast_copy_pj_str(req_num, &sip_ruri->user, sizeof(req_num));
		} else if (PJSIP_URI_SCHEME_IS_TEL(ruri)) {
			tel_ruri = (pjsip_tel_uri *)sip_ruri;
			ast_copy_pj_str(req_num, &tel_ruri->number, sizeof(req_num));
		}

		is_inv_req = 1;
	}

	ast_log(LOG_NOTICE, "ip_identify, match count[%d], is_inv_req[%d]\n",  ao2_iterator_count(it_matches), is_inv_req);
	for (; (match = ao2_iterator_next(it_matches)); ao2_cleanup(match)) {
		ast_log(LOG_NOTICE, "ip_identify, check match->endpoint_name[%s], check To user\n", match->endpoint_name);

		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);
		if (!endpoint) {
			continue;
		}

		if (!is_inv_req) {
			ao2_cleanup(match);
			break;
		}

		/*! peer trunk, don't need to check the to header */
		if (ast_strlen_zero(endpoint->username) || strncmp(endpoint->trunk_type, "register", 8)) {
			ao2_cleanup(endpoint);
			endpoint = NULL;
			continue;
		}

		if (strcasecmp(endpoint->didmode, "to-header") == 0) {
			did_num = to_num;
		} else {
			did_num = req_num;
		}

		if (!strncmp(endpoint->username, did_num, sizeof(to_num))) {
			ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], To user[%s] matched username\n", ast_sorcery_object_get_id(endpoint), did_num);
			ao2_cleanup(match);
			break;  /*! to matched */
		}

		ao2_cleanup(endpoint);
		endpoint = NULL;
	}
	ao2_iterator_destroy(it_matches);

	if (endpoint) {
		ast_log(LOG_NOTICE, "Retrieved endpoint %s by To user\n", ast_sorcery_object_get_id(endpoint));
		return endpoint; 
	}

	/* to match did mode */
	if (!(it_matches_did = ao2_callback(candidates, OBJ_MULTIPLE, ip_identify_match_check, &addr))) {
		return NULL;
	}

	for (; (match = ao2_iterator_next(it_matches_did)); ao2_cleanup(match)) {
		char diversion_user[80] = {0};

		ast_log(LOG_NOTICE, "ip_identify, check match->endpoint_name[%s], check DID\n", match->endpoint_name);
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", match->endpoint_name);

		if (!endpoint) {
			continue;
		} 

		if (strcasecmp(endpoint->didmode, "to-header") == 0) {
			did_num = to_num;
		} else {
			did_num = req_num;
		}

		if (ast_exists_extension(NULL, endpoint->context, did_num, 1, "NO_NEED_MATCH_CALLERID")) {
			ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], To user[%s] matched DID mode\n", ast_sorcery_object_get_id(endpoint), did_num);
			ao2_cleanup(match);
			break;  /*! did matched */
		}

		/* Added by yfyang, for Bug 234702 */
		if (!ast_sip_get_diversion_user(rdata, diversion_user, sizeof(diversion_user))) {
			if (ast_exists_extension2(NULL, endpoint->context, did_num, 1, diversion_user, 1)) {
				ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], DID user[%s], DiversionID[%s] matched\n", ast_sorcery_object_get_id(endpoint), did_num, diversion_user);
				ao2_cleanup(match);
				break;  /*! did matched */
			}
		}

		if (0 == ast_sip_find_sub_handler_for_supplement_exten(AST_SIP_CCSS_ORIGINAL_EXTEN, did_num, &exten_data)) {
			char *device = exten_data.device_name + strlen("PJSIP/");

			/* Added by yfyang, for Bug303207 */
			char *ptr = strchr(device, '/');
			if (ptr) {
				*ptr = '\0';
			}

			if (!strncmp(device,  ast_sorcery_object_get_id(endpoint), AST_MAX_EXTENSION)) {
				ast_log(LOG_NOTICE, "ip_identify, endpoint[%s], DID user[%s] matched cc-user\n", ast_sorcery_object_get_id(endpoint), did_num);
				ao2_cleanup(match);
				break;  /*! did matched */
			}
		}

		ao2_cleanup(endpoint);
		endpoint = NULL;
	}
	ao2_iterator_destroy(it_matches_did);

	if (endpoint) {
		ast_log(LOG_NOTICE, "Retrieved endpoint %s by DID mode \n", ast_sorcery_object_get_id(endpoint));
	}

	return endpoint; 
#endif
}
#endif

static struct ast_sip_endpoint_identifier ip_identifier = {
	.identify_endpoint = ip_identify,
};

static struct ast_sip_endpoint *header_identify(pjsip_rx_data *rdata)
{
	return common_identify(header_identify_match_check, rdata);
}

static struct ast_sip_endpoint_identifier header_identifier = {
	.identify_endpoint = header_identify,
};

#ifndef BUILD_ON_CLOUD_PBX
/*! \brief Helper function which performs a host lookup and adds result to identify match */
#ifndef GRANDSTREAM_NETWORKS
static int ip_identify_match_host_lookup(struct ip_identify_match *identify, const char *host)
#else
static int ip_identify_match_host_lookup(struct ip_identify_match *identify, const char *host, unsigned short srvport)
#endif
{
	struct ast_sockaddr *addrs;
	int num_addrs = 0, error = 0, i;
	int results = 0;

	num_addrs = ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC);
	if (!num_addrs) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	struct ast_ha  *ha_inuse = NULL;
	char   addr[48] = {0};

	ha_inuse = ast_get_ha_inuse(identify->matches, host);
	if (ha_inuse) {
		snprintf(addr, sizeof(addr), "%s", ast_sockaddr_stringify_addr(&ha_inuse->addr));
	}

    ast_log(LOG_NOTICE, "ip_identify_match_host_lookup, host[%s], num_addrs[%d]", host, num_addrs);
    identify->matches = ast_free_ha_by_host(identify->matches, host);
#endif

	for (i = 0; i < num_addrs; ++i) {
#ifndef GRANDSTREAM_NETWORKS
		/* Check if the address is already in the list, if so don't add it again */
		if (identify->matches && (ast_apply_ha(identify->matches, &addrs[i]) != AST_SENSE_ALLOW)) {
			continue;
		}

		/* We deny what we actually want to match because there is an implicit permit all rule for ACLs */
		identify->matches = ast_append_ha("d", ast_sockaddr_stringify_addr(&addrs[i]), identify->matches, &error);
#else
		if (!ast_sockaddr_peer_is_valid(&addrs[i])) {
			continue;
		}

		identify->matches = ast_append_ha_1("p", ast_sockaddr_stringify_addr(&addrs[i]), host, srvport, identify->matches, &error);
#endif
		if (!identify->matches || error) {
			results = -1;
			break;
		}

		results += 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ha_inuse) {
		ast_log(LOG_NOTICE, "set host inuse addr[%s]", addr);
		ast_ha_set_used_by_addr(identify->matches, addr);
		ast_free(ha_inuse);
	}
#endif

	ast_free(addrs);

	return results;
}

/*! \brief Helper function which performs an SRV lookup and then resolves the hostname */
static int ip_identify_match_srv_lookup(struct ip_identify_match *identify, const char *prefix, const char *host, int results)
{
	char service[NI_MAXHOST];
	struct srv_context *context = NULL;
	int srv_ret;
	const char *srvhost;
	unsigned short srvport;

	snprintf(service, sizeof(service), "%s.%s", prefix, host);

	while (!(srv_ret = ast_srv_lookup(&context, service, &srvhost, &srvport))) {
		int hosts;

		/* In the case of the SRV lookup we don't care if it fails, we will output a log message
		 * when we fallback to a normal lookup.
		 */
#ifndef GRANDSTREAM_NETWORKS
		hosts = ip_identify_match_host_lookup(identify, srvhost);
#else
		hosts = ip_identify_match_host_lookup(identify, srvhost, srvport);
#endif
		if (hosts == -1) {
#ifndef GRANDSTREAM_NETWORKS
			results = -1;
			break;
#else
			continue;
#endif
		} else {
			results += hosts;
		}
	}

	ast_srv_cleanup(&context);

	return results;
}
#endif

#ifdef BUILD_ON_CLOUD_PBX

/*! \brief Comparator function for matching an object by IP address */
static int ip_identify_dst_match_check(void *obj, void *arg, int flags)
{
	struct ip_identify_match *identify = obj;
	struct ast_sockaddr *addr = arg;
	int sense;

	sense = ast_apply_ha(identify->dst_matches, addr);
	if (sense == AST_SENSE_ALLOW) {
		ast_log(LOG_DEBUG, "Source address %s matches identify '%s'\n",
				ast_sockaddr_stringify_addr(addr), ast_sorcery_object_get_id(identify));
		return CMP_MATCH;
	} 

	ast_log(LOG_DEBUG, "Source address %s does not match identify '%s'\n",
			ast_sockaddr_stringify_addr(addr), ast_sorcery_object_get_id(identify));
	return 0;
}

static int ip_identify_check_match_domains(void *obj, void *arg, int flags)
{
	struct ip_identify_match *identify = obj;
	char  *domain = arg;

	if (!domain || ast_strlen_zero(identify->peer_host)) {
		return 0;
	}

	if (!strcmp(domain, identify->peer_host)) {
		ast_log(LOG_DEBUG, "From domain %s matches identify '%s'\n", domain, ast_sorcery_object_get_id(identify));
		return CMP_MATCH;
	}

	return 0;
}

static int ip_identify_dst_match_host_lookup(struct ip_identify_match *identify, const char *host, unsigned short srvport)
{
	struct ast_sockaddr *addrs;
	int num_addrs = 0, error = 0, i;
	int results = 0;

	num_addrs = ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC);
	if (!num_addrs) {
		return -1;
	}

    ast_log(LOG_NOTICE, "ip_identify_dst_match_host_lookup, host[%s], srvport[%d], num_addrs[%d]", host, srvport, num_addrs);
    identify->dst_matches = ast_free_ha_by_host(identify->dst_matches, host);

	for (i = 0; i < num_addrs; ++i) {
		if (!ast_sockaddr_peer_is_valid(&addrs[i])) {
			continue;
		}

		identify->dst_matches = ast_append_ha_1("p", ast_sockaddr_stringify_addr(&addrs[i]), host, srvport, identify->dst_matches, &error);
		if (!identify->dst_matches || error) {
			results = -1;
			break;
		}

		results += 1;
	}

	ast_free(addrs);

	return results;
}

static int ip_identify_dst_match_srv_lookup(struct ip_identify_match *identify, const char *prefix, const char *host, int results)
{
	char service[NI_MAXHOST];
	struct srv_context *context = NULL;
	int srv_ret;
	const char *srvhost;
	unsigned short srvport;

	snprintf(service, sizeof(service), "%s.%s", prefix, host);

	while (!(srv_ret = ast_srv_lookup(&context, service, &srvhost, &srvport))) {
		int hosts;

		hosts = ip_identify_dst_match_host_lookup(identify, srvhost, srvport);
		if (hosts == -1) {
			continue;
		} else {
			results += hosts;
		}
	}

	ast_srv_cleanup(&context);

	return results;
}

/*! \brief Custom handler for match field */
static int ip_identify_dst_match_handler(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ip_identify_match *identify = obj;
	char *input_string = ast_strdupa(var->value);
	char host[80] = {0};
	char service[80] = {0};
    unsigned short srvport = 0;
 	int error = 0;

	if (ast_strlen_zero(input_string)) {
		return 0;
	}

	get_host_info_from_string((const char *)input_string, host, sizeof(host), service, sizeof(service), &srvport);
	ast_string_field_set(identify, destination, host);

	if (is_a_valid_IP_address(host) == 1) {
		identify->dst_matches = ast_append_ha_1("p", host, host, srvport, identify->dst_matches, &error);  
        return 0;
	}

	identify->dst_port = srvport;
	ast_string_field_set(identify, dst_service, service);
	ast_string_field_set(identify, dst_to_resolve, host);
	return 0;
}

static int dst_match_to_str(const void *obj, const intptr_t *args, char **buf)
{
	RAII_VAR(struct ast_str *, str, ast_str_create(MAX_OBJECT_FIELD), ast_free);
	const struct ip_identify_match *identify = obj;

	ast_ha_join(identify->dst_matches, &str);
	*buf = ast_strdup(ast_str_buffer(str));
	return 0;
}

static int dst_match_to_var_list(const void *obj, struct ast_variable **fields)
{
	char str[MAX_OBJECT_FIELD];
	const struct ip_identify_match *identify = obj;
	struct ast_variable *head = NULL;
	struct ast_ha *ha = identify->dst_matches;

	for (; ha; ha = ha->next) {
		const char *addr = ast_strdupa(ast_sockaddr_stringify_addr(&ha->addr));
		snprintf(str, MAX_OBJECT_FIELD, "%s%s/%s", ha->sense == AST_SENSE_ALLOW ? "!" : "",
			addr, ast_sockaddr_stringify_addr(&ha->netmask));

		ast_variable_list_append(&head, ast_variable_new("destination", str, ""));
	}

	if (head) {
		*fields = head;
	}

	return 0;
}

static int ip_identify_sbc_host_lookup(struct ip_sbc_matches *sbc_matches, const char *host, unsigned short srvport)
{
	struct ast_sockaddr *addrs;
	struct ast_ha   *ha_inuse = NULL;
	char  addr[48] = {0};
	int num_addrs = 0, error = 0, i;
	int results = 0;

	num_addrs = ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC);
	if (!num_addrs) {
		return -1;
	}

	ha_inuse = ast_get_ha_inuse(sbc_matches->matches, host);
	if (ha_inuse) {
		snprintf(addr, sizeof(addr), "%s", ast_sockaddr_stringify_addr(&ha_inuse->addr));
	}

	ast_log(LOG_NOTICE, "ip_identify_sbc_host_lookup, host[%s], num_addrs[%d], inuse addr[%s]", host, num_addrs, addr);
	sbc_matches->matches = ast_free_ha_by_host(sbc_matches->matches, host);

	for (i = 0; i < num_addrs; ++i) {
		if (!ast_sockaddr_peer_is_valid(&addrs[i])) {
			continue;
		}

		sbc_matches->matches = ast_append_ha_1("p", ast_sockaddr_stringify_addr(&addrs[i]), host, srvport, sbc_matches->matches, &error);
		if (!sbc_matches->matches || error) {
			results = -1;
			break;
		}

		results += 1;
	}

	if (ha_inuse) {
		ast_log(LOG_NOTICE, "set SBC inuse addr[%s]", addr);
		ast_ha_set_used_by_addr(sbc_matches->matches, addr);
		ast_free(ha_inuse);
	}

	ast_free(addrs);
	return results;
}

static int ip_identify_sbc_srv_lookup(struct ip_sbc_matches *sbc_matches, const char *prefix, const char *host, int results)
{
	char service[NI_MAXHOST];
	struct srv_context *context = NULL;
	int srv_ret;
	const char *srvhost;
	unsigned short srvport;

	snprintf(service, sizeof(service), "%s.%s", prefix, host);

	while (!(srv_ret = ast_srv_lookup(&context, service, &srvhost, &srvport))) {
		int hosts;

		hosts = ip_identify_sbc_host_lookup(sbc_matches, srvhost, srvport);
		if (hosts == -1) {
			continue;
		} else {
			results += hosts;
		}
	}

	ast_srv_cleanup(&context);

	return results;
}


static int ip_identify_sbc_refresh_timeout(const void *data)
{
	struct ip_sbc_matches *sbc_matches = (struct ip_sbc_matches *)data;

	if (NULL == data) {
		return 0;
	}

	if (!sbc_matches->need_resolve) {
		return 0;
	}

	if (!ast_strlen_zero(sbc_matches->domain)) {
		struct ast_sockaddr addr;
		int dst_results = 0;

		/* If the provided string is not an IP address perform SRV resolution on it */
		if (!sbc_matches->port && !ast_strlen_zero(sbc_matches->service) && !ast_sockaddr_parse(&addr, sbc_matches->domain, 0)) {
			dst_results = ip_identify_sbc_srv_lookup(sbc_matches, sbc_matches->service, sbc_matches->domain, dst_results);
		}

		/* If SRV fails fall back to a normal lookup on the host itself */
		if (!dst_results) {
			dst_results = ip_identify_sbc_host_lookup(sbc_matches, sbc_matches->domain, sbc_matches->port);
		}

		if (dst_results == 0) {
			ast_log(LOG_WARNING, "sbc[%s] did not resolve to any address\n", sbc_matches->domain);
		} else if (dst_results == -1) {
			ast_log(LOG_WARNING, "sbc[%s] resovle failed\n", sbc_matches->domain);
		}
	}

	sbc_matches->sched_id = ast_sched_add(dns_sched, IP_IDENTIFY_DNS_REFRESH_INTERVAL*1000, ip_identify_sbc_refresh_timeout, sbc_matches);
	return 0;
}

static int ip_identify_sbc_matches_alloc(char *domain, char *service, unsigned short port)
{
	int error = 0;

	if (!domain) {
		return -1;
	}

	if (g_sbc_matches) {
		return 0;
	}

	g_sbc_matches = ast_calloc(1, sizeof(*g_sbc_matches)+1);
	if (!g_sbc_matches) {
		return -1;
	}

	g_sbc_matches->need_resolve = 1;
	g_sbc_matches->port = port;

	strncpy(g_sbc_matches->domain, domain, sizeof(g_sbc_matches->domain));
	if (strlen(service)) {
		strncpy(g_sbc_matches->service, service, sizeof(g_sbc_matches->service));
	}

	if (is_a_valid_IP_address(domain) == 1) {
		g_sbc_matches->matches = ast_append_ha_1("p", domain, domain, port, g_sbc_matches->matches, &error);
		g_sbc_matches->need_resolve = 0;
	} else {
		g_sbc_matches->sched_id = ast_sched_add(dns_sched, IP_IDENTIFY_DNS_FIRST_LOOKUP*1000, ip_identify_sbc_refresh_timeout, g_sbc_matches);
	}

	return 0;
}


static void ip_identify_sbc_matches_destroy(void)
{
	if (!g_sbc_matches) {
		return;
	}

	AST_SCHED_DEL(dns_sched, g_sbc_matches->sched_id);
	ast_free_ha(g_sbc_matches->matches);
	ast_free(g_sbc_matches);
	g_sbc_matches = NULL;
}
#endif

/*! \brief Custom handler for match field */
static int ip_identify_match_handler(const struct aco_option *opt, struct ast_variable *var, void *obj)
{
	struct ip_identify_match *identify = obj;
	char *input_string = ast_strdupa(var->value);
	char *current_string;

#ifdef GRANDSTREAM_NETWORKS
	char host[80] = {0};
	char service[80] = {0};
	unsigned short srvport = 0;
#endif

	if (ast_strlen_zero(var->value)) {
		return 0;
	}

	while ((current_string = ast_strip(strsep(&input_string, ",")))) {
		char *mask = strrchr(current_string, '/');
		int error = 0;

		if (ast_strlen_zero(current_string)) {
			continue;
		}

#ifdef GRANDSTREAM_NETWORKS
		get_host_info_from_string((const char *)current_string, host, sizeof(host), service, sizeof(service), &srvport);
        current_string = host;
#endif

		if (mask) {
#ifndef GRANDSTREAM_NETWORKS
			identify->matches = ast_append_ha("d", current_string, identify->matches, &error);
#else
			identify->matches = ast_append_ha_1("p", current_string, current_string, srvport, identify->matches, &error);   //modified by yfyang, d-->p
#endif
			if (!identify->matches || error) {
				ast_log(LOG_ERROR, "Failed to add address '%s' to ip endpoint identifier '%s'\n",
					current_string, ast_sorcery_object_get_id(obj));
				return -1;
			}

			continue;
		}

#ifdef BUILD_ON_CLOUD_PBX
		ip_identify_sbc_matches_alloc(host, service, srvport);
#endif
#ifdef GRANDSTREAM_NETWORKS
		if (is_a_valid_IP_address(host) == 1) {
			identify->matches = ast_append_ha_1("p", host, host, srvport, identify->matches, &error);   
			continue;
		}

		if (strlen(service)) {
			ast_string_field_set(identify, service, service);
		} else {
			identify->a_port = srvport;
		}
#endif
		if (!identify->hosts) {
			identify->hosts = ast_str_container_alloc_options(AO2_ALLOC_OPT_LOCK_NOLOCK, HOSTS_BUCKETS);
			if (!identify->hosts) {
				ast_log(LOG_ERROR, "Failed to create container to store hosts on ip endpoint identifier '%s'\n",
					ast_sorcery_object_get_id(obj));
				return -1;
			}
		}

		error = ast_str_container_add(identify->hosts, current_string);
		if (error) {
			ast_log(LOG_ERROR, "Failed to store host '%s' for resolution on ip endpoint identifier '%s'\n",
				current_string, ast_sorcery_object_get_id(obj));
			return -1;
		}
	}

	return 0;
}

/*! \brief Apply handler for identify type */
static int ip_identify_apply(const struct ast_sorcery *sorcery, void *obj)
{
	struct ip_identify_match *identify = obj;
#ifndef GRANDSTREAM_NETWORKS
	char *current_string;
	struct ao2_iterator i;
#endif

	/* Validate the identify object configuration */
	if (ast_strlen_zero(identify->endpoint_name)) {
		ast_log(LOG_ERROR, "Identify '%s' missing required endpoint name.\n",
			ast_sorcery_object_get_id(identify));
		return -1;
	}
	if (ast_strlen_zero(identify->match_header) /* No header to match */
		/* and no static IP addresses with a mask */
		&& !identify->matches
		/* and no addresses to resolve */
		&& (!identify->hosts || !ao2_container_count(identify->hosts))) {
		ast_log(LOG_ERROR, "Identify '%s' is not configured to match anything.\n",
			ast_sorcery_object_get_id(identify));
		return -1;
	}

	if (!ast_strlen_zero(identify->match_header)) {
		char *c_header;
		char *c_value;
		int len;

		/* Split the header name and value */
		c_header = ast_strdupa(identify->match_header);
		c_value = strchr(c_header, ':');
		if (!c_value) {
			ast_log(LOG_ERROR, "Identify '%s' missing ':' separator in match_header '%s'.\n",
				ast_sorcery_object_get_id(identify), identify->match_header);
			return -1;
		}
		*c_value = '\0';
		c_value = ast_strip(c_value + 1);
		c_header = ast_strip(c_header);

		if (ast_strlen_zero(c_header)) {
			ast_log(LOG_ERROR, "Identify '%s' has no SIP header to match in match_header '%s'.\n",
				ast_sorcery_object_get_id(identify), identify->match_header);
			return -1;
		}

		if (!strcmp(c_value, "//")) {
			/* An empty regex is the same as an empty literal string. */
			c_value = "";
		}

		if (ast_string_field_set(identify, match_header_name, c_header)
			|| ast_string_field_set(identify, match_header_value, c_value)) {
			return -1;
		}

		len = strlen(c_value);
		if (2 < len && c_value[0] == '/' && c_value[len - 1] == '/') {
			/* Make "/regex/" into "regex" */
			c_value[len - 1] = '\0';
			++c_value;

			if (regcomp(&identify->regex_buf, c_value, REG_EXTENDED | REG_NOSUB)) {
				ast_log(LOG_ERROR, "Identify '%s' failed to compile match_header regex '%s'.\n",
					ast_sorcery_object_get_id(identify), c_value);
				return -1;
			}
			identify->is_regex = 1;
		}
	}

	if (!identify->hosts) {
#ifdef BUILD_ON_CLOUD_PBX
		if (ast_strlen_zero(identify->dst_to_resolve)) {
			return 0;
		}
#else
		/* No match addresses to resolve */
		return 0;
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	identify->sched_id = ast_sched_add(dns_sched, IP_IDENTIFY_DNS_FIRST_LOOKUP*1000, ip_identify_refresh_timeout, identify);
#else
	/* Resolve the match addresses now */
	i = ao2_iterator_init(identify->hosts, 0);
	while ((current_string = ao2_iterator_next(&i))) {
		struct ast_sockaddr address;
		int results = 0;

		/* If the provided string is not an IP address perform SRV resolution on it */
		if (identify->srv_lookups && !ast_sockaddr_parse(&address, current_string, 0)) {
			results = ip_identify_match_srv_lookup(identify, "_sip._udp", current_string,
				results);
			if (results != -1) {
				results = ip_identify_match_srv_lookup(identify, "_sip._tcp",
					current_string, results);
			}
			if (results != -1) {
				results = ip_identify_match_srv_lookup(identify, "_sips._tcp",
					current_string, results);
			}
		}

		/* If SRV fails fall back to a normal lookup on the host itself */
		if (!results) {
			results = ip_identify_match_host_lookup(identify, current_string);
		}

		if (results == 0) {
			ast_log(LOG_WARNING, "Identify '%s' provided address '%s' did not resolve to any address\n",
				ast_sorcery_object_get_id(identify), current_string);
		} else if (results == -1) {
			ast_log(LOG_ERROR, "Identify '%s' failed when adding resolution results of '%s'\n",
				ast_sorcery_object_get_id(identify), current_string);
			ao2_ref(current_string, -1);
			ao2_iterator_destroy(&i);
			return -1;
		}

		ao2_ref(current_string, -1);
	}
	ao2_iterator_destroy(&i);

	ao2_ref(identify->hosts, -1);
	identify->hosts = NULL;
#endif

	return 0;
}

static int match_to_str(const void *obj, const intptr_t *args, char **buf)
{
	RAII_VAR(struct ast_str *, str, ast_str_create(MAX_OBJECT_FIELD), ast_free);
	const struct ip_identify_match *identify = obj;

	ast_ha_join(identify->matches, &str);
	*buf = ast_strdup(ast_str_buffer(str));
	return 0;
}

static int match_to_var_list(const void *obj, struct ast_variable **fields)
{
	char str[MAX_OBJECT_FIELD];
	const struct ip_identify_match *identify = obj;
	struct ast_variable *head = NULL;
	struct ast_ha *ha = identify->matches;

	for (; ha; ha = ha->next) {
		const char *addr = ast_strdupa(ast_sockaddr_stringify_addr(&ha->addr));
		snprintf(str, MAX_OBJECT_FIELD, "%s%s/%s", ha->sense == AST_SENSE_ALLOW ? "!" : "",
			addr, ast_sockaddr_stringify_addr(&ha->netmask));

		ast_variable_list_append(&head, ast_variable_new("match", str, ""));

	}

	if (head) {
		*fields = head;
	}

	return 0;
}

static int sip_identify_to_ami(const struct ip_identify_match *identify,
			       struct ast_str **buf)
{
	return ast_sip_sorcery_object_to_ami(identify, buf);
}

static int send_identify_ami_event(void *obj, void *arg, void *data, int flags)
{
	struct ip_identify_match *identify = obj;
	const char *endpoint_name = arg;
	struct ast_sip_ami *ami = data;
	struct ast_str *buf;

	/* Build AMI event */
	buf = ast_sip_create_ami_event("IdentifyDetail", ami);
	if (!buf) {
		return CMP_STOP;
	}
	if (sip_identify_to_ami(identify, &buf)) {
		ast_free(buf);
		return CMP_STOP;
	}
	ast_str_append(&buf, 0, "EndpointName: %s\r\n", endpoint_name);

	/* Send AMI event */
	astman_append(ami->s, "%s\r\n", ast_str_buffer(buf));
	++ami->count;

	ast_free(buf);
	return 0;
}

static int format_ami_endpoint_identify(const struct ast_sip_endpoint *endpoint,
					struct ast_sip_ami *ami)
{
	struct ao2_container *identifies;
	struct ast_variable fields = {
		.name = "endpoint",
		.value = ast_sorcery_object_get_id(endpoint),
	};

	identifies = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify",
		AST_RETRIEVE_FLAG_MULTIPLE, &fields);
	if (!identifies) {
		return -1;
	}

	/* Build and send any found identify object's AMI IdentifyDetail event. */
	ao2_callback_data(identifies, OBJ_MULTIPLE | OBJ_NODATA,
		send_identify_ami_event,
		(void *) ast_sorcery_object_get_id(endpoint),
		ami);

	ao2_ref(identifies, -1);
	return 0;
}

struct ast_sip_endpoint_formatter endpoint_identify_formatter = {
	.format_ami = format_ami_endpoint_identify
};

static int cli_iterator(void *container, ao2_callback_fn callback, void *args)
{
	const struct ast_sip_endpoint *endpoint = container;
	struct ao2_container *identifies;

	struct ast_variable fields = {
		.name = "endpoint",
		.value = ast_sorcery_object_get_id(endpoint),
		.next = NULL,
	};

	identifies = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "identify",
		AST_RETRIEVE_FLAG_MULTIPLE, &fields);
	if (!identifies) {
		return -1;
	}

	ao2_callback(identifies, OBJ_NODATA, callback, args);
	ao2_cleanup(identifies);

	return 0;
}

static struct ao2_container *cli_get_container(const char *regex)
{
	RAII_VAR(struct ao2_container *, container, NULL, ao2_cleanup);
	struct ao2_container *s_container;

	container =  ast_sorcery_retrieve_by_regex(ast_sip_get_sorcery(), "identify", regex);
	if (!container) {
		return NULL;
	}

	s_container = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
		ast_sorcery_object_id_sort, ast_sorcery_object_id_compare);
	if (!s_container) {
		return NULL;
	}

	if (ao2_container_dup(s_container, container, 0)) {
		ao2_ref(s_container, -1);
		return NULL;
	}

	return s_container;
}

static void *cli_retrieve_by_id(const char *id)
{
	return ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "identify", id);
}

static int cli_print_header(void *obj, void *arg, int flags)
{
	struct ast_sip_cli_context *context = arg;
	int indent = CLI_INDENT_TO_SPACES(context->indent_level);
	int filler = CLI_MAX_WIDTH - indent - 22;

	ast_assert(context->output_buffer != NULL);

	ast_str_append(&context->output_buffer, 0,
		"%*s:  <Identify/Endpoint%*.*s>\n",
		indent, "Identify", filler, filler, CLI_HEADER_FILLER);

	if (context->recurse) {
		context->indent_level++;
		indent = CLI_INDENT_TO_SPACES(context->indent_level);
		filler = CLI_LAST_TABSTOP - indent - 24;

		ast_str_append(&context->output_buffer, 0,
			"%*s:  <criteria%*.*s>\n",
			indent, "Match", filler, filler, CLI_HEADER_FILLER);

		context->indent_level--;
	}

	return 0;
}

static int cli_print_body(void *obj, void *arg, int flags)
{
	RAII_VAR(struct ast_str *, str, ast_str_create(MAX_OBJECT_FIELD), ast_free);
	struct ip_identify_match *ident = obj;
	struct ast_sip_cli_context *context = arg;
	struct ast_ha *match;
	struct ast_ha *matches;
	int indent;

	ast_assert(context->output_buffer != NULL);

	ast_str_append(&context->output_buffer, 0, "%*s:  %s/%s\n",
		CLI_INDENT_TO_SPACES(context->indent_level), "Identify",
		ast_sorcery_object_get_id(ident), ident->endpoint_name);

	if (context->recurse) {
		context->indent_level++;
		indent = CLI_INDENT_TO_SPACES(context->indent_level);

		matches = ident->matches;
#ifdef BUILD_ON_CLOUD_PBX
		if (g_sbc_matches) {
			matches = g_sbc_matches->matches;
		}
#endif

		for (match = matches; match; match = match->next) {
			const char *addr = ast_sockaddr_stringify_addr(&match->addr);

			ast_str_append(&context->output_buffer, 0, "%*s: %s -> %s%s/%d/%d\n",
				indent,
				"Match",
#ifdef GRANDSTREAM_NETWORKS
				match->hostname,
				match->sense != AST_SENSE_ALLOW ? "!" : "",
				addr, ast_sockaddr_cidr_bits(&match->netmask), match->inuse);
#else
				match->sense == AST_SENSE_ALLOW ? "!" : "",
				addr, ast_sockaddr_cidr_bits(&match->netmask));
#endif
		}

#ifdef BUILD_ON_CLOUD_PBX
		for (match = ident->dst_matches; match; match = match->next) {
			const char *addr = ast_sockaddr_stringify_addr(&match->addr);

			ast_str_append(&context->output_buffer, 0, "%*s: %s -> %s%s:%d\n",
				indent,
				"Destination",
				ident->destination,
				match->sense != AST_SENSE_ALLOW ? "!" : "",
				addr, match->port);
		}
#endif

		if (!ast_strlen_zero(ident->match_header)) {
			ast_str_append(&context->output_buffer, 0, "%*s: %s\n",
				indent,
				"Match",
				ident->match_header);
		}

		context->indent_level--;

		if (context->indent_level == 0) {
			ast_str_append(&context->output_buffer, 0, "\n");
		}
	}

	if (context->show_details
		|| (context->show_details_only_level_0 && context->indent_level == 0)) {
		ast_str_append(&context->output_buffer, 0, "\n");
		ast_sip_cli_print_sorcery_objectset(ident, context, 0);
	}

	return 0;
}

/*
 * A function pointer to callback needs to be within the
 * module in order to avoid problems with an undefined
 * symbol when the module is loaded.
 */
static char *my_cli_traverse_objects(struct ast_cli_entry *e, int cmd,
	struct ast_cli_args *a)
{
	return ast_sip_cli_traverse_objects(e, cmd, a);
}

static struct ast_cli_entry cli_identify[] = {
AST_CLI_DEFINE(my_cli_traverse_objects, "List PJSIP Identifies",
	.command = "pjsip list identifies",
	.usage = "Usage: pjsip list identifies [ like <pattern> ]\n"
	"       List the configured PJSIP Identifies\n"
	"       Optional regular expression pattern is used to filter the list.\n"),
AST_CLI_DEFINE(my_cli_traverse_objects, "Show PJSIP Identifies",
	.command = "pjsip show identifies",
	.usage = "Usage: pjsip show identifies [ like <pattern> ]\n"
	"       Show the configured PJSIP Identifies\n"
	"       Optional regular expression pattern is used to filter the list.\n"),
AST_CLI_DEFINE(my_cli_traverse_objects, "Show PJSIP Identify",
	.command = "pjsip show identify",
	.usage = "Usage: pjsip show identify <id>\n"
	"       Show the configured PJSIP Identify\n"),
};

static struct ast_sip_cli_formatter_entry *cli_formatter;

#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, for HA service data sync */

static struct stasis_message_router *ip_ha_data_router = NULL;

/*! data format, for example: 
 {"id": "id_trunk_1"
  "matches" : "sip2.jzhang.com:192.168.129.228,sip1.jzhang.com:192.168.129.73"
 }
*/ 
static void ip_ha_identify_sync_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	const char			*id_str      = NULL;
	const char			*matches_str = NULL;
	RAII_VAR(struct ip_identify_match *, identify, NULL, ao2_cleanup);
	char   delim[]  = ","; 
	char  *token    = NULL;
	char  *save_ptr = NULL;
	char   buffer[512] = {0};

	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (!payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	id_str = ast_json_string_get(ast_json_object_get(json_object, "id"));
	matches_str = ast_json_string_get(ast_json_object_get(json_object, "matches"));
	if (!id_str || !matches_str) {
		ast_log(LOG_WARNING, "No id/matches in JSON object data");
		return;
	}

	identify = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "identify", id_str);
	if (!identify) {
		ast_log(LOG_WARNING, "No identify object[%s]", id_str);
		return;
	}

	ast_log(LOG_NOTICE, "ip_ha_identify_sync_cb, id[%s], matches[%s]", id_str, matches_str);
	/* append ha */
	ao2_lock(identify);

	strncpy(buffer, matches_str, sizeof(buffer)-1);
	token = buffer;

	/* free previous info first */
	struct ast_ha *p = identify->matches;
	struct ast_ha *cur;
	while(p)
	{
		cur = p;
		p = p->next;
		ast_free(cur);
	}
	identify->matches = NULL; 
	while (NULL != (token = (strtok_r(token, delim, &save_ptr)))) {
		unsigned short port = 0;
		char host[128] = {0};
		char addr[128] = {0};
		char port_str[8] = {0};
		int sep_num;
		char *tmp = NULL,*start, *last;
		char tmp_buf[512] = {0};
		/* host:addr:port */

		tmp = token;
		sep_num = 0;
		start = NULL;
		last = NULL;
		memset(tmp_buf, 0, sizeof(tmp_buf));
		/* matches[ucm242.a.gdms.work:52.83.137.81:0,ucm242.a.gdms.work:2404:c2c0:86e7:5d10:9eb4:cdf3:5c97:27f2:0] */
		while(*tmp != '\0')
		{
			if(*tmp == ':')
			{
				sep_num++;
			}
			tmp++;
		}
		/* ipv4 */
		if(sep_num <= 2)
		{
			sscanf(token, "%127[^:]:%127[^:]:%7s", host, addr, port_str);
		}
		else /* ipv6 */
		{
			strcpy(tmp_buf, token);
			tmp  = tmp_buf;

			while(*tmp != '\0')
			{
				if(*tmp == ':' )
				{
					if(start == NULL)
					{
						*tmp = '\0';
						strncpy(host, tmp_buf, strlen(tmp_buf));
						start = tmp+1;
					}
					last = tmp;
				}
				tmp++;
			}

			if(start != NULL && last != NULL)
			{
				*last = '\0';
				strncpy(addr, start, strlen(start));
				strncpy(port_str, last+1, strlen(last+1));
			}
		}
		
		token = NULL;
		if (strlen(port_str)) {
			port = atoi(port_str);
		}

		if (strlen(host) && strlen(addr)) {
			//identify->matches = ast_free_ha_by_host(identify->matches, host); /* bug231202 */
			identify->matches = ast_append_ha_1("p", addr, host, port, identify->matches, NULL);
			ast_log(LOG_NOTICE, "ip_ha_identify_sync_cb, host[%s], addr[%s], port[%d]", host, addr, port);
		}
	}
	ao2_unlock(identify);

	return;
}

static void ip_ha_role_switch_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	int    role  = 0;
	int    retry = 10;

	/* Only do sync when switch to active */
	if (HA_ROLE_ACTIVE != ast_ha_get_role()) {
		return;
    }

	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (!payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	role = ast_json_integer_get(ast_json_object_get(json_object, "ha_role"));
	if (!role) {
		ast_log(LOG_WARNING, "Incomplete JSON object data"); 
		return;
	}

	/* Delete the matches entry */
	ast_db_deltree("Identify/matches", NULL);

	ast_safe_system2("nvram set :sip_load=1");
	while (retry > 0 && 0 != ast_ha_send_sipload(1)) {
		sleep(1);
		retry--;
	}

	ast_log(LOG_NOTICE, "identifier_ip swtiched to Active.\n");

	return;
}

static int ip_ha_stasis_route_init(void)
{
	struct stasis_topic * ip_ha_data_topic = ast_ha_sync_data_topic_all();
	if (!ip_ha_data_topic) {
		ast_log(LOG_WARNING, "Failed to get ip_ha_data_topic\n");
		return -1;
	}

	ip_ha_data_router = stasis_message_router_create(ip_ha_data_topic);
	if (ip_ha_data_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create ip_ha_data_router\n");
		return -1;
	}

	if (stasis_message_router_add(ip_ha_data_router, ast_ha_role_switch_type(), ip_ha_role_switch_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_role_switch_type\n");
		return -1;
	}

	if (stasis_message_router_add(ip_ha_data_router, ast_ha_identify_ip_type(), ip_ha_identify_sync_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_identify_ip_type \n");
		return -1;
	}

	return 0;
}
#endif


static int load_module(void)
{
	ast_sorcery_apply_config(ast_sip_get_sorcery(), "res_pjsip_endpoint_identifier_ip");

#ifndef GRANDSTREAM_NETWORKS    /* modified by yfyang, for performance optimization */
	ast_sorcery_apply_default(ast_sip_get_sorcery(), "identify", "config", "pjsip.conf,criteria=type=identify");
#else
	ast_sorcery_apply_default(ast_sip_get_sorcery(), "identify", "config", "pjsip_identify.conf,criteria=type=identify");
#endif

	if (ast_sorcery_object_register(ast_sip_get_sorcery(), "identify", ip_identify_alloc, NULL, ip_identify_apply)) {
		return AST_MODULE_LOAD_DECLINE;
	}

#ifdef GRANDSTREAM_NETWORKS
	dns_sched = ast_sched_context_create();
	if (!dns_sched) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sched_start_thread(dns_sched)) {
		ast_log(LOG_ERROR, "Could not start scheduler thread for identify dns refresh\n");
		ast_sched_context_destroy(dns_sched);
		return AST_MODULE_LOAD_FAILURE;
	}
#endif

	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "type", "", OPT_NOOP_T, 0, 0);
	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "endpoint", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ip_identify_match, endpoint_name));
	ast_sorcery_object_field_register_custom(ast_sip_get_sorcery(), "identify", "match", "", ip_identify_match_handler, match_to_str, match_to_var_list, 0, 0);
	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "match_header", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ip_identify_match, match_header));
	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "srv_lookups", "yes", OPT_BOOL_T, 1, FLDSET(struct ip_identify_match, srv_lookups));
#ifdef GRANDSTREAM_NETWORKS
	ast_sorcery_object_field_register(ast_sip_get_sorcery(), "identify", "peer_host", "", OPT_STRINGFIELD_T, 0, STRFLDSET(struct ip_identify_match, peer_host));
#endif
#ifdef BUILD_ON_CLOUD_PBX
	ast_sorcery_object_field_register_custom(ast_sip_get_sorcery(), "identify", "destination", "", ip_identify_dst_match_handler, dst_match_to_str, dst_match_to_var_list, 0, 0);
#endif
	ast_sorcery_load_object(ast_sip_get_sorcery(), "identify");

	ast_sip_register_endpoint_identifier_with_name(&ip_identifier, "ip");
	ast_sip_register_endpoint_identifier_with_name(&header_identifier, "header");
	ast_sip_register_endpoint_formatter(&endpoint_identify_formatter);

	cli_formatter = ao2_alloc(sizeof(struct ast_sip_cli_formatter_entry), NULL);
	if (!cli_formatter) {
		ast_log(LOG_ERROR, "Unable to allocate memory for cli formatter\n");
		return AST_MODULE_LOAD_DECLINE;
	}
	cli_formatter->name = "identify";
	cli_formatter->print_header = cli_print_header;
	cli_formatter->print_body = cli_print_body;
	cli_formatter->get_container = cli_get_container;
	cli_formatter->iterate = cli_iterator;
	cli_formatter->get_id = ast_sorcery_object_get_id;
	cli_formatter->retrieve_by_id = cli_retrieve_by_id;

	ast_sip_register_cli_formatter(cli_formatter);
	ast_cli_register_multiple(cli_identify, ARRAY_LEN(cli_identify));

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_register_resolve_identifier(&ip_resolve_identify);
	/* Added by yfyang, for HA service data sync */
	/*if ((HA_ROLE_STANDBY == ast_ha_get_role()) && ip_ha_stasis_route_init()) {*/
	if ( ip_ha_stasis_route_init()) {
		ast_sched_context_destroy(dns_sched);
		return AST_MODULE_LOAD_FAILURE;
	}

#endif

	return AST_MODULE_LOAD_SUCCESS;
}

static int reload_module(void)
{
	ast_sorcery_reload_object(ast_sip_get_sorcery(), "identify");

	return 0;
}

static int unload_module(void)
{
	ast_cli_unregister_multiple(cli_identify, ARRAY_LEN(cli_identify));
	ast_sip_unregister_cli_formatter(cli_formatter);
	ast_sip_unregister_endpoint_formatter(&endpoint_identify_formatter);
	ast_sip_unregister_endpoint_identifier(&header_identifier);
	ast_sip_unregister_endpoint_identifier(&ip_identifier);

#ifdef BUILD_ON_CLOUD_PBX
	ip_identify_sbc_matches_destroy();
#endif
#ifdef GRANDSTREAM_NETWORKS
	ast_sip_unregister_resolve_identifier(&ip_resolve_identify);

	if (dns_sched) {
		ast_sched_context_destroy(dns_sched);
		dns_sched = NULL;
	}
#endif

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP IP endpoint identifier",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.reload = reload_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND - 4,
	.requires = "res_pjsip",
);
