/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/module.h"

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("pjsip");
#endif

static int get_from_header(pjsip_rx_data *rdata, char *username, size_t username_size, char *domain, size_t domain_size)
{
	pjsip_uri *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_from;

	if (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from)) {
		return -1;
	}
	sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
	ast_copy_pj_str(username, &sip_from->user, username_size);
	ast_copy_pj_str(domain, &sip_from->host, domain_size);
	return 0;
}

static pjsip_authorization_hdr *get_auth_header(pjsip_rx_data *rdata, char *username,
	size_t username_size, char *realm, size_t realm_size, pjsip_authorization_hdr *start)
{
	pjsip_authorization_hdr *header;

	header = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_AUTHORIZATION, start);

	if (!header || pj_stricmp2(&header->scheme, "digest")) {
		return NULL;
	}

	ast_copy_pj_str(username, &header->credential.digest.username, username_size);
	ast_copy_pj_str(realm, &header->credential.digest.realm, realm_size);

	return header;
}

static int find_transport_state_in_use(void *obj, void *arg, int flags)
{
	struct ast_sip_transport_state *transport_state = obj;
	pjsip_rx_data *rdata = arg;

	if (transport_state->transport == rdata->tp_info.transport
		|| (transport_state->factory
			&& !pj_strcmp(&transport_state->factory->addr_name.host, &rdata->tp_info.transport->local_name.host)
			&& transport_state->factory->addr_name.port == rdata->tp_info.transport->local_name.port)) {
		return CMP_MATCH;
	}

	return 0;
}

#define DOMAIN_NAME_LEN 255
#define USERNAME_LEN    255

static struct ast_sip_endpoint *find_endpoint(pjsip_rx_data *rdata, char *endpoint_name,
	char *domain_name)
{
	struct ast_sip_endpoint *endpoint;

	if (!ast_sip_get_disable_multi_domain()) {
		struct ast_sip_domain_alias *alias;
		struct ao2_container *transport_states;
		struct ast_sip_transport_state *transport_state = NULL;
		struct ast_sip_transport *transport = NULL;
		char id[DOMAIN_NAME_LEN + USERNAME_LEN + sizeof("@")];

		/* Attempt to find the endpoint given the name and domain provided */
#ifdef GRANDSTREAM_NETWORKS
        if (domain_name) {
            snprintf(id, sizeof(id), "%s@%s", endpoint_name, domain_name);
        } else {
            snprintf(id, sizeof(id), "%s", endpoint_name);
        }
#else
		snprintf(id, sizeof(id), "%s@%s", endpoint_name, domain_name);
#endif
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", id);
		if (endpoint) {
			return endpoint;
		}

		/* See if an alias exists for the domain provided */
		alias = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "domain_alias",
			domain_name);
		if (alias) {
			snprintf(id, sizeof(id), "%s@%s", endpoint_name, alias->domain);
			ao2_ref(alias, -1);
			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", id);
			if (endpoint) {
				return endpoint;
			}
		}

		/* See if the transport this came in on has a provided domain */
		if ((transport_states = ast_sip_get_transport_states())
			&& (transport_state = ao2_callback(transport_states, 0, find_transport_state_in_use, rdata))
			&& (transport = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "transport", transport_state->id))
			&& !ast_strlen_zero(transport->domain)) {
			snprintf(id, sizeof(id), "%s@%s", endpoint_name, transport->domain);
			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", id);
		}
		ao2_cleanup(transport);
		ao2_cleanup(transport_state);
		ao2_cleanup(transport_states);
		if (endpoint) {
			return endpoint;
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug 295956 */
		endpoint = ast_sip_get_account_tk_endpoint(endpoint_name, rdata->tp_info.transport->type_name);
		if (endpoint) {
			return endpoint;
		}
#endif
	}

	/* Fall back to no domain */
	return ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpoint_name);
}

#ifdef GRANDSTREAM_NETWORKS
/* identify user by username+ip, after username matched, we check the user'contact */
static int username_check_contact(pjsip_rx_data *rdata, struct ast_sip_endpoint *endpoint, int auth)
{
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	struct ao2_iterator i;
	struct ast_sip_contact *contact;
	char aor_id[80]      = {0};
	char contact_uri[80] = {0};
	char src_addr[80]    = {0};
	int  contact_matched = 0;

	if (!rdata || !endpoint) {
		return -1;
	}

	aor = ast_sip_location_retrieve_aor(endpoint->aors); 
	if (!aor) {
		return -1;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts || (aor->max_contacts > ao2_container_count(contacts))) {
        if (auth) {
            ast_log(LOG_NOTICE, "The endpoint[%s] not reach its max contacts, pass!", ast_sorcery_object_get_id(endpoint));
            return 0;
        }
	}

	if (contacts) {
		if (strchr(rdata->pkt_info.src_name, ':')) {
			snprintf(src_addr, sizeof(src_addr) - 1, "[%s]:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port); 
		} else {
			snprintf(src_addr, sizeof(src_addr) - 1, "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port); 
		}

		i = ao2_iterator_init(contacts, 0);
		while ((contact = ao2_iterator_next(&i))) {
			sscanf(contact->uri, "%79[^@]@%79[^;]", aor_id, contact_uri);
			ao2_ref(contact, -1); 

			if (!strcmp(contact_uri, src_addr)) {
				contact_matched = 1;
				break;
			}
		}
		ao2_iterator_destroy(&i);

		if (contact_matched) {
			ast_log(LOG_NOTICE, "contact_matched!");
			return 0;
		}
	}

	/* check from domain */
	char proxy_host[256] = {0};
	ast_sip_get_nvram_value(":WsProxyHost", proxy_host, sizeof(proxy_host));
	if (!strlen(proxy_host)) {
		char *public_host = getenv("PUBLIC_HOST");
		if (!ast_strlen_zero(public_host)) {
			strncpy(proxy_host, public_host, sizeof(proxy_host)-1);
		}
	}

	if (strlen(proxy_host)) {
		pj_str_t host;
		pjsip_uri  *from = rdata->msg_info.from->uri;
		pjsip_sip_uri *sip_uri = pjsip_uri_get_uri(from);

		if (!from || !sip_uri || (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from))) {
			return -1;
		}

		pj_strdup_with_null(rdata->tp_info.pool, &host, &sip_uri->host);
		if (!pj_strcmp2(&host, proxy_host)) {
			ast_log(LOG_NOTICE, "From domain matched WsProxyHost[%s]!", proxy_host);
			return 0;
		}
    }

	return -1;
}

static unsigned int webrtc_call_through_peer_trunk(pjsip_rx_data *rdata, char *username)
{
	if (!rdata || !username) {
		return 0;
	}

	if (strncasecmp(username, "webrtc_trunk", 12)) {
		return 0;
	}

	if (rdata->msg_info.cseq && pj_stricmp2(&rdata->msg_info.cseq->method.name, "INVITE")) {
		return 0;
	}

	if (ast_sip_judge_if_wave_terminal_type(rdata)) {
		return 0;
	}

	/* If it's not wave type, and tranport is not ws, we think it's from ucm peer trunk */
	if (rdata->tp_info.transport &&
		strcasecmp(rdata->tp_info.transport->type_name, "WS") && strcasecmp(rdata->tp_info.transport->type_name, "WSS")) {
		return 1;
	}

	return 0;
}

static int check_from_anonymous(pjsip_rx_data *rdata)
{
	if (!rdata) {
		return 0;
	}

	pjsip_uri *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_from;
	char username[80] = {0};

	if (!PJSIP_URI_SCHEME_IS_SIP(from) && !PJSIP_URI_SCHEME_IS_SIPS(from)) {
		return 0;
	}

	sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
	if (!sip_from) {
		return 0;
	}

    ast_copy_pj_str(username, &sip_from->user, sizeof(username));
	if (strcasecmp(username, "anonymous") == 0) {
		return 1;
	}

	return 0;
}
#endif

static struct ast_sip_endpoint *username_identify(pjsip_rx_data *rdata)
{
	char username[USERNAME_LEN + 1];
	char domain[DOMAIN_NAME_LEN + 1];
	struct ast_sip_endpoint *endpoint;

	if (get_from_header(rdata, username, sizeof(username), domain, sizeof(domain))) {
		return NULL;
	}

	/*
	 * We may want to be matched without any user options getting
	 * in the way.
	 */
	AST_SIP_USER_OPTIONS_TRUNCATE_CHECK(username);

	ast_debug(3, "Attempting identify by From username '%s' domain '%s'\n", username, domain);

	endpoint = find_endpoint(rdata, username, domain);
	if (!endpoint) {
		ast_debug(3, "Endpoint not found for From username '%s' domain '%s'\n", username, domain);
		return NULL;
	}
	if (!(endpoint->ident_method & AST_SIP_ENDPOINT_IDENTIFY_BY_USERNAME)) {
		ast_debug(3, "Endpoint found for '%s' but 'username' method not supported'\n", username);
		ao2_cleanup(endpoint);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* check the contact, to non-register request*/
	if (!(((rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) && (rdata->msg_info.msg->line.req.method.id == PJSIP_REGISTER_METHOD)) || endpoint->click2call_enable)
		   && username_check_contact(rdata, endpoint, 0)) {
		ast_debug(3, "Checkout contact failed, endpoint[%s]\n", ast_sorcery_object_get_id(endpoint));
		ao2_ref(endpoint, -1);
		return NULL;
	}

	if (endpoint->click2call_enable) {
		char  host_value[256] = {0};

		ast_sip_get_nvram_value(":WsProxyHost", host_value, sizeof(host_value));
		if (!strlen(host_value)) {
			ast_log(LOG_NOTICE, "No ProxyHost, not support click2call.");
			ao2_ref(endpoint, -1);
			return NULL;
		}
	}

	if (webrtc_call_through_peer_trunk(rdata, username)) {
		ast_log(LOG_NOTICE, "It's a webrtc call through peer trunk, don't use user name identify\n");
		ao2_ref(endpoint, -1);
		return NULL;
	}
#endif
	ast_debug(3, "Identified [%s] by From username '%s' domain '%s'\n", ast_sorcery_object_get_id(endpoint), username, domain);

	return endpoint;
}

static struct ast_sip_endpoint *auth_username_identify(pjsip_rx_data *rdata)
{
	char username[USERNAME_LEN + 1], realm[DOMAIN_NAME_LEN + 1];
	struct ast_sip_endpoint *endpoint;
	pjsip_authorization_hdr *auth_header = NULL;

	while ((auth_header = get_auth_header(rdata, username, sizeof(username), realm, sizeof(realm),
		auth_header ? auth_header->next : NULL))) {
		ast_debug(3, "Attempting identify by Authorization username '%s' realm '%s'\n", username,
			realm);
#ifdef GRANDSTREAM_NETWORKS
		endpoint = find_endpoint(rdata, username, NULL);
#else
		endpoint = find_endpoint(rdata, username, realm);
#endif
		if (!endpoint) {
			ast_debug(3, "Endpoint not found for Authentication username '%s' realm '%s'\n",
				username, realm);
			ao2_cleanup(endpoint);
			continue;
		}
		if (!(endpoint->ident_method & AST_SIP_ENDPOINT_IDENTIFY_BY_AUTH_USERNAME)) {
			ast_debug(3, "Endpoint found for '%s' but 'auth_username' method not supported'\n",
				username);
			ao2_cleanup(endpoint);
			continue;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (username_check_contact(rdata, endpoint, 1)) {
			ast_debug(3, "Endpoint found for '%s' but contact check failed.\n", username);
			ao2_cleanup(endpoint);
			continue;
		}

		/* Added by yfyang, for Bug 287915 */
		if (check_from_anonymous(rdata)) {
			ast_string_field_set(endpoint, anonymous, "anonymous");
			ast_log(LOG_NOTICE, "Set anonymous to endpoint[%s].", ast_sorcery_object_get_id(endpoint));
		}
#endif
		ast_debug(3, "Identified by Authorization username '%s' realm '%s'\n", username, realm);

		return endpoint;
	}

	return NULL;
}


static struct ast_sip_endpoint_identifier username_identifier = {
	.identify_endpoint = username_identify,
};

static struct ast_sip_endpoint_identifier auth_username_identifier = {
	.identify_endpoint = auth_username_identify,
};


static int load_module(void)
{
	ast_sip_register_endpoint_identifier_with_name(&username_identifier, "username");
	ast_sip_register_endpoint_identifier_with_name(&auth_username_identifier, "auth_username");
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_endpoint_identifier(&auth_username_identifier);
	ast_sip_unregister_endpoint_identifier(&username_identifier);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP username endpoint identifier",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND - 4,
	.requires = "res_pjsip",
);
