/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2023, Grandstream, Inc.
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_pubsub</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <regex.h>

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/module.h"
#include "asterisk/logger.h"
#include "asterisk/astobj2.h"
#include "asterisk/sorcery.h"
#include "asterisk/app.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/astdb.h"
#include "asterisk/callerid.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/fwd_timing.h"

AST_MODULE_LOG("pjsip");


/*!
 * \brief A subscription for as feature event
 *
 * This structure acts as the owner for the underlying SIP subscription. It
 * also keeps a pointer to an associated "provider" so when a state changes
 * a notify data creator is quickly accessible.
 */
struct feature_event_subscription {
	struct ast_sip_subscription *sip_sub;   /*! The SIP subscription */
	struct ast_taskprocessor *serializer;   /*! The serializer to use for notifications */
	char exten[AST_MAX_EXTENSION];          /*! Extension within the context to receive updates from */
};

AST_VECTOR(fe_body_part_vec, pjsip_multipart_part *);

#define DEFAULT_FEATURE_EVENT_BODY "application/x-as-feature-event+xml"
#define FWD_EXECUTE_SQL_FILE "/app/cfg_default/var/lib/asterisk/scripts/fwd_execute_sql.py"
#define FWD_EXECUTE_SQL_ENV  "source /app/asterisk/Python-3.8.3/setup_python_env.sh"

int g_python_env_set = 0;

static void subscription_shutdown(struct ast_sip_subscription *sub);
static int new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource);
static int subscription_established(struct ast_sip_subscription *sub);
static void *get_notify_data(struct ast_sip_subscription *sub);
static int update_notify_data(void *notify_data, int stype, void *fwd_status);
static int subscription_send_fe_notify(struct ast_sip_subscription *sub, void *info);
//static void feature_event_to_string(void *body, struct ast_str **str);
static void feature_event_state_change(struct ast_sip_subscription *sub, pjsip_msg_body *body, enum pjsip_evsub_state state);
struct ast_fwd_status *feature_event_body_part_first(struct ast_sip_feature_event_data *data, enum ast_fwd_service_type *stype);

struct ast_sip_notifier feature_event_notifier = {
	.default_accept = DEFAULT_FEATURE_EVENT_BODY,
	.new_subscribe = new_subscribe,
	.subscription_established = subscription_established,
	.get_notify_data = get_notify_data,
	.update_notify_data = update_notify_data,
	.send_feature_event_notify = subscription_send_fe_notify,
};

static struct ast_sip_subscriber feature_event_subscriber = {
	.state_change = feature_event_state_change,
};

struct ast_sip_subscription_handler feature_event_handler = {
	.event_name = "as-feature-event",
	.body_type = AST_SIP_FEATURE_EVENT_DATA,
	.accept = { DEFAULT_FEATURE_EVENT_BODY, },
	.subscription_shutdown = subscription_shutdown,
	.notifier = &feature_event_notifier,
	.subscriber = &feature_event_subscriber,
};

static void feature_event_subscription_destructor(void *obj)
{
	struct feature_event_subscription *sub = obj;

	ast_sip_subscription_destroy(sub->sip_sub);
	ast_taskprocessor_unreference(sub->serializer);
}

/*!
 * \internal
 * \brief Allocates an feature_event_subscription object.
 *
 * Creates the underlying SIP subscription for the given request. First makes
 * sure that there are registered handler and provider objects available.
 */
static struct feature_event_subscription *feature_event_subscription_alloc(struct ast_sip_subscription *sip_sub, struct ast_sip_endpoint *endpoint)
{
	struct feature_event_subscription * feature_event_sub;

	feature_event_sub = ao2_alloc(sizeof(*feature_event_sub), feature_event_subscription_destructor);
	if (!feature_event_sub) {
		return NULL;
	}

	feature_event_sub->sip_sub = sip_sub;
	feature_event_sub->serializer = ao2_bump(ast_sip_subscription_get_serializer(sip_sub));

	return feature_event_sub;
}

struct fe_notify_task_data {
	struct ast_sip_feature_event_data  feature_event_data;
	struct feature_event_subscription *feature_event_sub;
	int terminate;
};

struct fe_sync_task_data {
	char cmd[1024];
	char context[128];
    char service[8];
    char fwd_num[80];
	int fwd_num_type;
};

struct dnd_sync_task_data {
	char cmd[1024];
	char custom[128];
    char aauser[128];
	int  active;
};


static struct dnd_sync_task_data *alloc_dnd_task_data(char *cmd, char *custom, char *aauser, int active)
{
	if (!cmd || !custom || !aauser) {
		return NULL;
	}

	struct dnd_sync_task_data *task_data = ao2_alloc(sizeof(*task_data), NULL);

	if (!task_data) {
		ast_log(LOG_WARNING, "Unable to create sync task data\n");
		return NULL;
	}

	strncpy(task_data->cmd, cmd, sizeof(task_data->cmd));
	strncpy(task_data->custom, custom, sizeof(task_data->custom));
	strncpy(task_data->aauser, aauser, sizeof(task_data->aauser));
	task_data->active = active;

	return task_data;
}

static int dnd_sync_task(void *obj)
{
	RAII_VAR(struct dnd_sync_task_data *, task_data, obj, ao2_cleanup);
    int res = 0;

	if (!g_python_env_set) {
        ast_safe_system(FWD_EXECUTE_SQL_ENV);
		g_python_env_set = 1;
	}

    if (!task_data) {
        return 0;
    }

    res = ast_safe_system(task_data->cmd);
    ast_log(LOG_NOTICE, "cmd[%s]\n res [%d]", task_data->cmd, res);

    ast_db_put("CustomPresence", task_data->custom, (task_data->active ? "dnd,,":"available,,"));
	ast_ucm_db_put(task_data->aauser, "DND", (task_data->active ? "yes":"no"), 1);               
	ast_ucm_db_put(task_data->aauser, "DNDTIMETYPE", "0", 1); 
	return 0;
}

static struct fe_sync_task_data *alloc_sync_task_data(char *cmd, char *context, char *service, char *fwd_num, int num_type)
{
    if (!cmd || !context || !service) {
        return NULL;
    }

	struct fe_sync_task_data *task_data = ao2_alloc(sizeof(*task_data), NULL);

	if (!task_data) {
		ast_log(LOG_WARNING, "Unable to create sync task data\n");
		return NULL;
	}

	strncpy(task_data->cmd, cmd, sizeof(task_data->cmd));
	strncpy(task_data->context, context, sizeof(task_data->context));
	strncpy(task_data->service, service, sizeof(task_data->service));

    if (ast_strlen_zero(fwd_num)) {
        memset(task_data->fwd_num, 0, sizeof(task_data->fwd_num));
        task_data->fwd_num_type = 0;
    } else {
        strncpy(task_data->fwd_num, fwd_num, sizeof(task_data->fwd_num));
        task_data->fwd_num_type = num_type;
    }

	return task_data;
}

static void notify_task_data_destructor(void *obj)
{
	struct fe_notify_task_data *task_data = obj;

	ao2_ref(task_data->feature_event_sub, -1);
}

static struct fe_notify_task_data *alloc_notify_task_data(const char *exten,
	struct feature_event_subscription *feature_event_sub,
	struct ast_fwd_active_status *fwd_state)
{
	struct fe_notify_task_data *task_data = ao2_alloc(sizeof(*task_data), notify_task_data_destructor);

	if (!task_data) {
		ast_log(LOG_WARNING, "Unable to create notify task data\n");
		return NULL;
	}

    task_data->feature_event_sub = feature_event_sub;
	ao2_ref(task_data->feature_event_sub, +1);

	strncpy(task_data->feature_event_data.exten, feature_event_sub->exten, sizeof(task_data->feature_event_data.exten));

    ast_fwd_active_status_cpy(fwd_state, &task_data->feature_event_data.fwd_state);
	return task_data;
}

static int notify_task(void *obj)
{
	RAII_VAR(struct fe_notify_task_data *, task_data, obj, ao2_cleanup);
	struct ast_sip_body_data data = {
		.body_type = AST_SIP_FEATURE_EVENT_DATA,
		.body_data = &task_data->feature_event_data,
	};

	/* Terminated subscriptions are no longer associated with a valid tree, and sending
	 * NOTIFY messages on a subscription which has already been terminated won't work.
	 */
	if (ast_sip_subscription_is_terminated(task_data->feature_event_sub->sip_sub)) {
		return 0;
	}

	if (ao2_ref(task_data->feature_event_sub, 0) <= 0) {
		ast_log(LOG_ERROR, "notify_task, the subscription destroyed");
		return 0;
	}

	ast_sip_subscription_notify(task_data->feature_event_sub->sip_sub, &data, task_data->terminate);

	return 0;
}


static struct ast_datastore_info ds_info = { };
static const char ds_name[] = "feature event datastore";

/*!
 * \internal
 * \brief Add a datastore for exten feature_event_subscription.
 *
 * Adds the feature_event_subscription wrapper object to a datastore so it can be retrieved
 * later based upon its association with the ast_sip_subscription.
 */
static int add_datastore(struct feature_event_subscription *feature_event_sub)
{
	RAII_VAR(struct ast_datastore *, datastore, ast_sip_subscription_alloc_datastore(&ds_info, ds_name), ao2_cleanup);

	if (!datastore) {
		return -1;
	}

	datastore->data = feature_event_sub;
	ast_sip_subscription_add_datastore(feature_event_sub->sip_sub, datastore);
	ao2_ref(feature_event_sub, +1);
	return 0;
}

/*!
 * \internal
 * \brief Get the feature_event_subscription object associated with the given
 * ast_sip_subscription in the datastore.
 */
static struct feature_event_subscription *get_feature_event_sub(struct ast_sip_subscription *sub)
{
	RAII_VAR(struct ast_datastore *, datastore, ast_sip_subscription_get_datastore(sub, ds_name), ao2_cleanup);

	return datastore ? datastore->data : NULL;
}

static int sync_task(void *obj)
{
	RAII_VAR(struct fe_sync_task_data *, task_data, obj, ao2_cleanup);
    int res = 0;

	if (!g_python_env_set) {
        ast_safe_system(FWD_EXECUTE_SQL_ENV);
		g_python_env_set = 1;
	}

    if (!task_data) {
        return 0;
    }

    res = ast_safe_system(task_data->cmd);
    ast_log(LOG_NOTICE, "cmd[%s]\n res [%d]", task_data->cmd, res);

    /* to modify the config file */
    struct ast_custom_function *custom_function = ast_custom_function_find("CFG_CALLTRANS");
    if (custom_function && custom_function->write) {
        char data[256] = {0};
        char buf[128]  = {0};

        snprintf(data, sizeof(data), "%s,%s", task_data->context, task_data->service);
        snprintf(buf, sizeof(buf), "%s", task_data->fwd_num);
        custom_function->write(NULL, "CFG_CALLTRANS", data, buf);

        snprintf(data, sizeof(data), "%s,%s_dst_type", task_data->context, task_data->service);
        snprintf(buf, sizeof(buf), "%d", task_data->fwd_num_type);
        custom_function->write(NULL, "CFG_CALLTRANS", data, buf);

        snprintf(data, sizeof(data), "%s,r%s", task_data->context, task_data->service);
        snprintf(buf, sizeof(buf), "%s", task_data->fwd_num);
        custom_function->write(NULL, "CFG_CALLTRANS", data, buf);

        snprintf(data, sizeof(data), "%s,r%s_dst_type", task_data->context, task_data->service);
        snprintf(buf, sizeof(buf), "%d", task_data->fwd_num_type);
        custom_function->write(NULL, "CFG_CALLTRANS", data, buf);
    }
	return 0;
}

static void feature_event_dnd_change(struct ast_sip_subscription *sub, pjsip_msg_body *body, enum pjsip_evsub_state state)
{
	char *xml_start  = NULL;
	char *xml_end    = NULL;
	char *type_start = NULL;
	char *type_end   = NULL;
	char *pos        = NULL;
	char  type_str[20] = {0};
    char  appdata[1024] = {0};
	char  device[128] = {0};
	char  aauser[128] = {0};
    struct dnd_sync_task_data *task_data;
	int   active_flag = 0;
    char *resource;

	if (!sub || !body) {
		ast_log(LOG_WARNING, "feature_event_state_change, invalid params\n");
		return;
	}

	ast_log(LOG_NOTICE, "feature_event_dnd_change start process\n");
/*
<SetDoNotDisturb xmlns="http://www.ecmainternational.org/standards/ecma-323/csta/ed3">
<device>9000</device>
<doNotDisturbOn>true</doNotDisturbOn>
</SetDoNotDisturb>
*/

    xml_start = strstr((const char*)body->data, "<SetDoNotDisturb");
    xml_end = strstr((const char*)body->data, "</SetDoNotDisturb");
	if (!xml_start || !xml_end) {
        ast_log(LOG_NOTICE, "No SetDoNotDisturb node, exit.");
		return;
	}

	/* get body data */
	type_start = strstr((const char*)body->data, "<doNotDisturbOn>");
	type_end   = strstr((const char*)body->data, "</doNotDisturbOn>");
	if (!type_start || !type_end) {
		ast_log(LOG_WARNING, "no doNotDisturbOn in the body[%s]\n", (char *)body->data);
		return;
	}

    pos = type_start + strlen("<doNotDisturbOn>");
	if (pos >= type_end) {
		return;
	}

	strncpy(type_str, pos, ((type_end-pos)>sizeof(type_str))? sizeof(type_str):(type_end-pos));
	active_flag = ast_true(type_str);

	resource = (char *)ast_sip_subscription_get_resource_name(sub);
	snprintf(device, sizeof(device), "PJSIP/%s", resource);
	snprintf(aauser, sizeof(aauser), "AAUSER/%s", resource);

	ast_log(LOG_NOTICE, "Extension[%s] set DND to [%d]", resource, active_flag);

	if (active_flag) {
		snprintf(appdata, sizeof(appdata), "LD_PRELOAD=/usr/lib/libjemalloc.so python %s \"update sip_accounts set dnd='yes',presence_status='dnd' where extension='%s'\" "
		                                   "\"%s\"  \"dnd\"  \"yes\" ", FWD_EXECUTE_SQL_FILE, resource, resource);
	} else {
		snprintf(appdata, sizeof(appdata), "LD_PRELOAD=/usr/lib/libjemalloc.so python %s \"update sip_accounts set dnd='no',presence_status='available' where extension='%s'\" "
                      "\"%s\"  \"dnd\"  \"no\" ", FWD_EXECUTE_SQL_FILE, resource, resource);
	}

    task_data = alloc_dnd_task_data(appdata, device, aauser, active_flag);
    if (!task_data) {
        ast_log(LOG_NOTICE, "alloc_dnd_task_data failed.");
        return;
    }

    if (ast_sip_push_task(ast_sip_subscription_get_serializer(sub), dnd_sync_task, task_data)) {
        ao2_cleanup(task_data);
        ast_log(LOG_NOTICE, "ast_sip_push_task failed.");
        return;
    }

	return;
}

/*
Accept: application/as-feature-event
                                                                                  >
<?xml version="1.0" encoding="ISO-8859-1"?>
<SetForwarding xmlns="http://www.ecmainternational.org/standards/ecma-323/csta/ed3">
<device>9725552011</device>
<forwardingType>forwardImmediate</forwardingType>
<activateForward>true</activateForward>
<forwardDN>2013</forwardDN>
</SetForwarding>
*/
static void feature_event_state_change(struct ast_sip_subscription *sub, pjsip_msg_body *body, enum pjsip_evsub_state state)
{
	char *xml_start  = NULL;
	char *xml_end    = NULL;
	char *type_start = NULL;
	char *type_end   = NULL;
	char *active_start  = NULL;
	char *active_end    = NULL;
	char *num_start  = NULL;
	char *num_end    = NULL;
	char *pos        = NULL;
	char  type_str[20] = {0};
	char  active_str[20] = {0};
	char  fwd_num[80]   = {0};
    char  appdata[1024] = {0};
	char  context[128]  = {0};
	enum  ast_fwd_service_type fwd_type = FWD_BUTT;
    enum  ast_fwd_dst_type fwd_dst_type = FWD_DST_CUSTOM;
    struct ast_fwd_status  status;
    struct fe_sync_task_data *task_data;
	int   active_flag = 0;
    const char *status_str;
    char *resource;
    char *fwd_type_str = NULL;

	if (!sub || !body) {
		ast_log(LOG_WARNING, "feature_event_state_change, invalid params\n");
		return;
	}

	/* check content type*/
	if (0 != pj_stricmp2(&body->content_type.type, "application") || 0 != pj_stricmp2(&body->content_type.subtype, "x-as-feature-event+xml")) {
		ast_log(LOG_WARNING, "the content type is [%s/%s], ignore\n", body->content_type.type.ptr, body->content_type.subtype.ptr);
		return;
	}

	ast_log(LOG_NOTICE, "feature_event_state_change start process\n");

    xml_start = strstr((const char*)body->data, "<SetForwarding");
    xml_end = strstr((const char*)body->data, "</SetForwarding");
	if (!xml_start || !xml_end) {
		ast_log(LOG_NOTICE, "No SetForwarding node.");
		feature_event_dnd_change(sub, body, state);
		return;
	}

	/* get body data */
	type_start = strstr((const char*)body->data, "<forwardingType>");
	type_end   = strstr((const char*)body->data, "</forwardingType>");
	if (!type_start || !type_end) {
		ast_log(LOG_WARNING, "no forwardingType in the body[%s]\n", (char *)body->data);
		return;
	}

    pos = type_start + strlen("<forwardingType>");
	if (pos >= type_end) {
		return;
	}

	strncpy(type_str, pos, ((type_end-pos)>sizeof(type_str))? sizeof(type_str):(type_end-pos));
    fwd_type = ast_fwd_type_get(type_str);
    if (fwd_type >= FWD_BUTT) {
        ast_log(LOG_NOTICE, "forwardingType[%s] invalid, exit.", type_str);
		return;
    }

	active_start = strstr((const char*)body->data, "<activateForward>");
	active_end   = strstr((const char*)body->data, "</activateForward>");
	if (!active_start || !active_end) {
		ast_log(LOG_WARNING, "no activateForward in the body[%s]\n", (char *)body->data);
		return;
	}

    pos = active_start + strlen("<activateForward>");
	if (pos >= active_end) {
		return;
	}

	strncpy(active_str, pos, ((active_end-pos)>sizeof(active_str))?sizeof(active_str):(active_end-pos));
    active_flag = ast_true(active_str);

	if (active_flag) {
        /* if active, get fwd dst number */
        num_start = strstr((const char*)body->data, "<forwardDN>");
        num_end   = strstr((const char*)body->data, "</forwardDN>");
        if (!num_start || !num_end) {
            ast_log(LOG_WARNING, "no forwardDN in the body[%s]\n", (char *)body->data);
            return;
        }

        pos = num_start + strlen("<forwardDN>");
        if (pos >= num_end) {
            return;
        }

        strncpy(fwd_num, pos, ((num_end-pos)>sizeof(fwd_num))?sizeof(fwd_num):(num_end-pos));
	}

	fwd_dst_type = ast_fwd_dst_type_get(fwd_num);
    resource = (char *)ast_sip_subscription_get_resource_name(sub);
    status_str = ast_presence_state2str(ast_fwd_ext_get_presence_status(resource));

    if (fwd_type == FWD_CFU || fwd_type == FWD_RCFU) {
        fwd_type_str = "cfu";
    } else if (fwd_type == FWD_CFN || fwd_type == FWD_RCFN) {
        fwd_type_str = "cfn";
    } else {
        fwd_type_str = "cfb";
    }

	ast_log(LOG_NOTICE, "forward type[%s], active[%d], dst_num[%s], dst_type[%d], status_str[%s]", fwd_type_str, active_flag, fwd_num, fwd_dst_type, status_str);

    /* to update notify data */
    status.active = (active_flag? 1:0);
    memcpy(status.dst_num, fwd_num, sizeof(status.dst_num));
    ast_sip_update_fe_notify_data(sub, fwd_type, (void *)&status);

    if (active_flag) {
        snprintf(appdata, sizeof(appdata), "LD_PRELOAD=/usr/lib/libjemalloc.so python %s \"update sip_presence_settings set %s_destination_type='%d',%s='%s', "
                                           "r%s_destination_type='%d',r%s='%s' where extension='%s' and presence_status='%s'\" "
                                           "\"%s\"  \"%s\"  \"%s\" ", FWD_EXECUTE_SQL_FILE, fwd_type_str, fwd_dst_type, fwd_type_str, fwd_num,
                                           fwd_type_str,fwd_dst_type,fwd_type_str,fwd_num, resource, status_str,
                                           resource, fwd_type_str, fwd_num);
    } else {
            snprintf(appdata, sizeof(appdata), "LD_PRELOAD=/usr/lib/libjemalloc.so python %s \"update sip_presence_settings set %s_destination_type='0',%s='',r%s_destination_type='0',r%s='' "
                     " where extension='%s' and presence_status='%s'\"  \"%s\"  \"%s\"  \"\" ",
                     FWD_EXECUTE_SQL_FILE, fwd_type_str, fwd_type_str, fwd_type_str, fwd_type_str, resource, status_str, resource, fwd_type_str);
    }

    snprintf(context, sizeof(context), "%s-%s", resource, status_str);

    task_data = alloc_sync_task_data(appdata, context, fwd_type_str, fwd_num, fwd_dst_type);
    if (!task_data) {
        ast_log(LOG_NOTICE, "alloc_sync_task_data failed.");
        return;
    }

    if (ast_sip_push_task(ast_sip_subscription_get_serializer(sub), sync_task, task_data)) {
        ao2_cleanup(task_data);
        ast_log(LOG_NOTICE, "ast_sip_push_task failed.");
        return;
    }

	return;
}

static void subscription_shutdown(struct ast_sip_subscription *sub)
{
	struct feature_event_subscription *feature_event_sub = get_feature_event_sub(sub);

	if (!feature_event_sub) {
		return;
	}

    //to do

	ast_sip_subscription_remove_datastore(feature_event_sub->sip_sub, ds_name);
	/* remove data store reference */
	ao2_cleanup(feature_event_sub);
}

static int subscription_send_fe_notify(struct ast_sip_subscription *sub, void *info)
{
	struct fe_notify_task_data *task_data;
	struct feature_event_subscription *feature_event_sub;
	struct ast_sip_feature_event_data *feature_event_data;

	if (!sub || !info) {
		return -1;
	}

	feature_event_sub = get_feature_event_sub(sub);
	if (NULL == feature_event_sub) {
		return -1;
	}

    feature_event_data = (struct ast_sip_feature_event_data *)info;
	if (!(task_data = alloc_notify_task_data((const char *)feature_event_data->exten, feature_event_sub, &feature_event_data->fwd_state))) {
		return -1;
	}

	/* safe to push this async since we copy the data from info and
	   add a ref for the device state info */
	if (ast_sip_push_task(ast_sip_subscription_get_serializer(sub), notify_task, task_data)) {
		ao2_cleanup(task_data);
		return -1;
	}
	return 0;
}

static int new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource)
{
    if (!endpoint || !resource) {
        return 403;
    }

    if (strcmp(ast_sorcery_object_get_id(endpoint), resource)) {
        ast_log(LOG_NOTICE, "Endpoint[%s] can't subscribe [%s], you can only subscribe as-feature-event event of yourself.", ast_sorcery_object_get_id(endpoint), resource);
        return 403;
    }

	return 200;
}

static int subscription_established(struct ast_sip_subscription *sip_sub)
{
	struct ast_sip_endpoint *endpoint = ast_sip_subscription_get_endpoint(sip_sub);
	const char *resource = ast_sip_subscription_get_resource_name(sip_sub);
	struct feature_event_subscription *feature_event_sub;

	if (!(feature_event_sub = feature_event_subscription_alloc(sip_sub, endpoint))) {
		ao2_cleanup(endpoint);
		return -1;
	}

	ast_copy_string(feature_event_sub->exten, resource, sizeof(feature_event_sub->exten));

	/* Go ahead and cleanup the endpoint since we don't need it anymore */
	ao2_cleanup(endpoint);

	if (add_datastore(feature_event_sub)) {
		ast_log(LOG_WARNING, "Unable to add to subscription datastore.\n");
		ao2_cleanup(feature_event_sub);
		return -1;
	}

	ao2_cleanup(feature_event_sub);
	ast_log(LOG_NOTICE, "subscription_established, resource[%s]", resource);
	return 0;
}

static void *get_notify_data(struct ast_sip_subscription *sub)
{
	struct feature_event_subscription *feature_event_sub = NULL;
	struct ast_sip_feature_event_data *feature_event_data= NULL;

	feature_event_sub = get_feature_event_sub(sub);
	if (!feature_event_sub) {
		return NULL;
	}

	feature_event_data = ast_calloc(1, sizeof(*feature_event_data)+1);
	if (!feature_event_data) {
		return NULL;
	}

	strncpy(feature_event_data->exten, feature_event_sub->exten, sizeof(feature_event_data->exten));

    ast_log(LOG_NOTICE, "get_notify_data, exten[%s]", feature_event_sub->exten);

    /* to get fwd data */
    ast_fwd_ext_get_notify_data(feature_event_sub->exten, &feature_event_data->fwd_state, 1);

#if 0
    {
        struct ast_fwd_active_status *status = &feature_event_data->fwd_state;

        ast_log(LOG_NOTICE, "CFB u/a/d=%d/%d/%s", status->cfb_status.update, status->cfb_status.active, status->cfb_status.dst_num);
        ast_log(LOG_NOTICE, "CFN u/a/d=%d/%d/%s", status->cfn_status.update, status->cfn_status.active, status->cfn_status.dst_num);
        ast_log(LOG_NOTICE, "CFU u/a/d=%d/%d/%s", status->cfu_status.update, status->cfu_status.active, status->cfu_status.dst_num);
        ast_log(LOG_NOTICE, "RCFB u/a/d=%d/%d/%s", status->rcfb_status.update, status->rcfb_status.active, status->rcfb_status.dst_num);
        ast_log(LOG_NOTICE, "RCFN u/a/d=%d/%d/%s", status->rcfn_status.update, status->rcfn_status.active, status->rcfn_status.dst_num);
        ast_log(LOG_NOTICE, "RCFU u/a/d=%d/%d/%s", status->rcfu_status.update, status->rcfu_status.active, status->rcfu_status.dst_num);
    }
#endif
	return feature_event_data;
}

static int update_notify_data(void *notify_data, int stype, void *fwd_status)
{
	struct ast_sip_feature_event_data *fe_data = notify_data;
    struct ast_fwd_status *status = fwd_status;

    if (!fe_data || !fwd_status || stype >= FWD_BUTT) {
        return 0;
    }

    ast_fwd_active_status_update_set(&fe_data->fwd_state, 0);

    if (stype == FWD_CFU) {
        memcpy(&fe_data->fwd_state.cfu_status, status, sizeof(fe_data->fwd_state.cfu_status));
        memcpy(&fe_data->fwd_state.rcfu_status, status, sizeof(fe_data->fwd_state.rcfu_status));
        fe_data->fwd_state.cfu_status.update = 1;
        fe_data->fwd_state.rcfu_status.update = 1;
    } else if (stype == FWD_CFB) {
        memcpy(&fe_data->fwd_state.cfb_status, status, sizeof(fe_data->fwd_state.cfb_status));
        memcpy(&fe_data->fwd_state.rcfb_status, status, sizeof(fe_data->fwd_state.rcfb_status));
        fe_data->fwd_state.cfb_status.update = 1;
        fe_data->fwd_state.rcfb_status.update = 1;
    } else if (stype == FWD_CFN){
        memcpy(&fe_data->fwd_state.cfn_status, status, sizeof(fe_data->fwd_state.cfn_status));
        memcpy(&fe_data->fwd_state.rcfn_status, status, sizeof(fe_data->fwd_state.rcfn_status));
        fe_data->fwd_state.cfn_status.update = 1;
        fe_data->fwd_state.rcfn_status.update = 1;
    }

    return 0;
}


static void feature_event_destroy_body(void *body)
{
	struct ast_str **feture_body = body;

	ast_free(*feture_body);
	ast_free(feture_body);
}

static void *feature_event_allocate_body(void)
{
	struct ast_str **feature_event_str;

	feature_event_str = ast_malloc(sizeof(*feature_event_str));
	if (!feature_event_str) {
		return NULL;
	}

	*feature_event_str = ast_str_create(128);
	if (!*feature_event_str) {
		ast_free(feature_event_str);
		return NULL;
	}
	return feature_event_str;
}


static int feature_event_body_parts_count(struct ast_sip_feature_event_data *event_data)
{
    int num = 0;
    struct ast_fwd_active_status *fwd_state = NULL;

    if (!event_data) {
        return num;
    }

    fwd_state = &event_data->fwd_state;

    if (fwd_state->cfb_status.update) {
        num++;
    }

    if (fwd_state->cfu_status.update) {
        num++;
    }

    if (fwd_state->cfn_status.update) {
        num++;
    }

    if (fwd_state->rcfb_status.update) {
        num++;
    }

    if (fwd_state->rcfu_status.update) {
        num++;
    }

    if (fwd_state->rcfn_status.update) {
        num++;
    }

    return num;
}

struct ast_fwd_status *feature_event_body_part_first(struct ast_sip_feature_event_data *data, enum ast_fwd_service_type *stype)
{
    struct ast_fwd_active_status *fwd_state = NULL;

    if (!data) {
        return NULL;
    }

    fwd_state = &data->fwd_state;

    if (fwd_state->cfb_status.update) {
        *stype = FWD_CFB;
        return &fwd_state->cfb_status;
    }

    if (fwd_state->cfu_status.update) {
        *stype = FWD_CFU;
        return &fwd_state->cfu_status;
    }

    if (fwd_state->cfn_status.update) {
        *stype = FWD_CFN;
        return &fwd_state->cfn_status;
    }

    if (fwd_state->rcfb_status.update) {
        *stype = FWD_RCFB;
        return &fwd_state->rcfb_status;
    }

    if (fwd_state->rcfu_status.update) {
        *stype = FWD_RCFU;
        return &fwd_state->rcfu_status;
    }

    if (fwd_state->rcfn_status.update) {
        *stype = FWD_RCFN;
        return &fwd_state->rcfn_status;
    }

    return NULL;
}

static int feature_event_generate_body_part(void *body, char *ext, enum ast_fwd_service_type stype, struct ast_fwd_status *status_data)
{
    struct ast_str **fwd_event = body;

    if (stype >= FWD_BUTT || !status_data || !ext) {
        return -1;
    }

    ast_str_append(fwd_event, 0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n");
    ast_str_append(fwd_event, 0, "<ForwardingEvent xmlns=\"http://www.ecma-international.org/standards/ecma-323/csta/ed3\">\r\n");
    ast_str_append(fwd_event, 0, "<device>%s</device>\r\n", ext);
    ast_str_append(fwd_event, 0, "<forwardingType>%s</forwardingType>\r\n", ast_fwd_type_to_str(stype));
    ast_str_append(fwd_event, 0, "<forwardStatus>%s</forwardStatus>\r\n", status_data->active? "true":"false");
    ast_str_append(fwd_event, 0, "<forwardTo>%s</forwardTo>\r\n", status_data->dst_num);

    if (stype == FWD_CFN || stype == FWD_RCFN) {
        if (status_data->active) {
            char family[128] = { 0 };
            char data[16] = {0};

            snprintf(family, sizeof(family), "AAUSER/%s", ext);
            ast_ucm_db_get(family, "RT", data, sizeof(data));
            if (!ast_strlen_zero(data)) {
                ast_str_append(fwd_event, 0, "<ringCount>%s</ringCount>\r\n", data);
            } else {
                char *ringtime = (char *)pbx_builtin_getvar_helper(NULL, "RINGTIME");
                if (!ast_strlen_zero(ringtime)) {
                    ast_str_append(fwd_event, 0, "<ringCount>%s</ringCount>\r\n", ringtime);
                }
            }
        } else {
            ast_str_append(fwd_event, 0, "<ringCount>60</ringCount>\r\n");
        }
    }

    ast_str_append(fwd_event, 0, "<callType>%s</callType>\r\n", (stype > FWD_CFB) ? "external" : "internal");
    ast_str_append(fwd_event, 0, "</ForwardingEvent>\r\n");

    return 0;
}

static void feature_event_build_body_part(pj_pool_t *pool, struct fe_body_part_vec *parts, char *ext, enum ast_fwd_service_type stype, struct ast_fwd_status *status)
{
	pjsip_multipart_part *part;
	pjsip_msg_body  *body;
    struct ast_str **body_str = NULL;
	pj_str_t type;
	pj_str_t subtype;
	pj_str_t text;

    /* build single body */
    body_str = feature_event_allocate_body();
    feature_event_generate_body_part(body_str, ext, stype, status);

    pj_cstr(&type, "application");
    pj_cstr(&subtype, "x-as-feature-event+xml");
    pj_cstr(&text, ast_str_buffer(*body_str));

    body = pjsip_msg_body_create(pool, &type, &subtype, &text);
    part = pjsip_multipart_create_part(pool);
	part->body = body;

    //ast_log(LOG_NOTICE, "body[%s]", body->data);
	AST_VECTOR_APPEND(parts, part);
    feature_event_destroy_body(body_str);
}



static pjsip_msg_body *feature_event_create_multipart_body(pj_pool_t *pool)
{
	pjsip_media_type media_type;
	pj_str_t pj_boundary;
	char boundary[24] = {0};

	pjsip_media_type_init2(&media_type, "multipart", "mixed");

	pj_cstr(&pj_boundary, ast_generate_random_string(boundary, sizeof(boundary)));
	return pjsip_multipart_create(pool, &media_type, &pj_boundary);
}

static pjsip_msg_body *feature_event_generate_list_body(pj_pool_t *pool, struct ast_sip_feature_event_data *event_data, int num)
{
	pjsip_msg_body       *multipart;
	struct fe_body_part_vec      body_parts;
    struct ast_fwd_active_status    *fwd_state = NULL;
	int    i;

    if (!pool || !event_data || !num) {
        return NULL;
    }

    ast_log(LOG_NOTICE, "feature_event_generate_list_body, num[%d].", num);
    if (AST_VECTOR_INIT(&body_parts, num)) {
        ast_log(LOG_NOTICE, "AST_VECTOR_INIT failed.");
		return NULL;
	}

    ast_log(LOG_NOTICE, "feature_event_generate_list_body, start.");
    fwd_state = &event_data->fwd_state;

    if (fwd_state->cfu_status.update) {
        feature_event_build_body_part(pool, &body_parts, (char *)event_data->exten, FWD_CFU, &fwd_state->cfu_status);
    }

    if (fwd_state->cfb_status.update) {
        feature_event_build_body_part(pool, &body_parts, (char *)event_data->exten, FWD_CFB, &fwd_state->cfb_status);
    }

    if (fwd_state->cfn_status.update) {
        feature_event_build_body_part(pool, &body_parts, (char *)event_data->exten, FWD_CFN, &fwd_state->cfn_status);
    }

    if (fwd_state->rcfu_status.update) {
        feature_event_build_body_part(pool, &body_parts, (char *)event_data->exten, FWD_RCFU, &fwd_state->rcfu_status);
    }

    if (fwd_state->rcfb_status.update) {
        feature_event_build_body_part(pool, &body_parts, (char *)event_data->exten, FWD_RCFB, &fwd_state->rcfb_status);
    }

    if (fwd_state->rcfn_status.update) {
        feature_event_build_body_part(pool, &body_parts, (char *)event_data->exten, FWD_RCFN, &fwd_state->rcfn_status);
    }

	/* This can happen if issuing partial state and no children of the list have changed state */
	if (AST_VECTOR_SIZE(&body_parts) == 0) {
		AST_VECTOR_FREE(&body_parts);

        ast_log(LOG_NOTICE, "body_parts size is 0.");
		return NULL;
	}

	multipart = feature_event_create_multipart_body(pool);

	for (i = 0; i < AST_VECTOR_SIZE(&body_parts); ++i) {
		pjsip_multipart_add_part(pool, multipart, AST_VECTOR_GET(&body_parts, i));
	}

	AST_VECTOR_FREE(&body_parts);
	return multipart;
}

static pjsip_msg_body *feature_event_generate_sip_body(pj_pool_t *pool, void *data, int full)
{
	struct ast_sip_feature_event_data *event_data = data;
    struct ast_fwd_status *fwd_state = NULL;
    enum   ast_fwd_service_type stype;
    pjsip_msg_body *sip_body  = NULL;
    struct ast_str **body_str = NULL;
	int    num = 0;
	pj_str_t type;
	pj_str_t subtype;
	pj_str_t text;

    if (!pool || !data) {
        return NULL;
    }

    if (full) {
        ast_fwd_active_status_update_set(&event_data->fwd_state, 1);
    }

    num = feature_event_body_parts_count(event_data);
    if (!num) {
        ast_log(LOG_NOTICE, "No fwd status need to update.");
        return NULL;
    }

    ast_log(LOG_NOTICE, "body parts count[%d].", num);
    if (num == 1) {
        /* build single body */
		fwd_state = feature_event_body_part_first(event_data, &stype);
		body_str = feature_event_allocate_body();
        feature_event_generate_body_part(body_str, (char *)event_data->exten, stype, fwd_state);

        pj_cstr(&type, "application");
        pj_cstr(&subtype, "x-as-feature-event+xml");
        pj_cstr(&text, ast_str_buffer(*body_str));
        sip_body = pjsip_msg_body_create(pool, &type, &subtype, &text);

        feature_event_destroy_body(body_str);
        return sip_body;
    }

    /* multipart body */
    sip_body = feature_event_generate_list_body(pool, event_data, num);
    return sip_body;
}

static struct ast_sip_pubsub_body_generator feature_event_body_generator = {
	.type = "application",
	.subtype = "x-as-feature-event+xml",
	.body_type = AST_SIP_FEATURE_EVENT_DATA,
	 //.allocate_body = feature_event_allocate_body,
	 //.generate_body_content = feature_event_generate_body_content,
    .generate_sip_body = feature_event_generate_sip_body,
	 //.to_string = feature_event_to_string,
     //.destroy_body = feature_event_destroy_body,
};

static int unload_module(void)
{
	ast_sip_unregister_subscription_handler(&feature_event_handler);
	ast_sip_pubsub_unregister_body_generator(&feature_event_body_generator);

	return 0;
}

static int load_module(void)
{
	if (ast_sip_register_subscription_handler(&feature_event_handler)) {
		ast_log(LOG_WARNING, "Unable to register subscription handler %s\n", feature_event_handler.event_name);
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sip_pubsub_register_body_generator(&feature_event_body_generator)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Feature Event State Notifications",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND + 5,
	.requires = "res_pjsip,res_pjsip_pubsub",
);

