/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2021, grandstream.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
 ***/


#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/module.h"

AST_MODULE_LOG("pjsip");

const char * const info_state_item_strmap[][1] = {
	{ /* 0 AST_GS_STATE_ITEM_TYPE_UNKNOW           */      "UNKNOW"                },
	{ /* 1 AST_GS_STATE_ITEM_TYPE_TERMINAL         */      "terminal-state"        },
	{ /* 1 AST_GS_STATE_ITEM_TYPE_GESTURE          */      "gesture-state"         },
};

static int is_general_control_info(pjsip_rx_data *rdata)
{
	return rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, "gs-general-control+xml");
}

static int is_gs_state_info(pjsip_rx_data *rdata)
{
	return rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, "x-gs-state+xml");
}


static void process_cancel_call_control_info(struct ast_sip_session *session, struct ast_xml_node *root_node)
{
	struct ast_xml_node *child_nodes      = NULL;
	struct ast_xml_node *node_iterator    = NULL;
	struct ast_xml_node *node_conf_id     = NULL;
	struct ast_xml_node *node_user_id     = NULL;
	struct ast_xml_node *node_prefix_id   = NULL;
	const char *node_name = NULL;
	const char *content   = NULL;
	char conf_id[128] = { 0 };
	char prefix_id[128] = { 0 };
	char matched_id[256] = { 0 };

	child_nodes = ast_xml_node_get_children(root_node);
	if (child_nodes) {
		for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "ConfID") && !node_conf_id) {
				node_conf_id = node_iterator;
			} else if (!strcmp(node_name, "UserID") && !node_user_id) {
				node_user_id = node_iterator;
			} else if (!strcmp(node_name, "PrefixID") && !node_prefix_id) {
				node_prefix_id = node_iterator;
			}
		}
	}

	if (node_conf_id) {
		content = ast_xml_get_text(node_conf_id);
		if (content) {
			ast_log(LOG_NOTICE, "[session:%p] process_cancel_call_control_info, ConfID[%s], endpoint:%s\n",
				session, content, ast_sorcery_object_get_id(session->endpoint));
			strncpy(conf_id, content, sizeof(conf_id) - 1);
		} else {
			ast_log(LOG_NOTICE, "[session:%p] process_cancel_call_control_info, can't get ConfID, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
		ast_xml_free_text(content);
	}

	if (node_prefix_id) {
		content = ast_xml_get_text(node_prefix_id);
		if (content) {
			ast_log(LOG_NOTICE, "[session:%p] process_cancel_call_control_info, PrefixID[%s], endpoint:%s\n",
				session, content, ast_sorcery_object_get_id(session->endpoint));
			strncpy(prefix_id, content, sizeof(prefix_id) - 1);
		} else {
			ast_log(LOG_WARNING, "[session:%p] process_cancel_call_control_info, can't get PrefixID, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
		ast_xml_free_text(content);
	}

	if (node_user_id) {
		content = ast_xml_get_text(node_user_id);
		if (content) {
			ast_log(LOG_NOTICE, "[session:%p] process_cancel_call_control_info, UserID[%s], endpoint:%s\n",
				session, content, ast_sorcery_object_get_id(session->endpoint));
			snprintf(matched_id, sizeof(matched_id) -1, "%s%s", prefix_id, content);
			ast_sip_session_cancel_pending_call_according_id(conf_id, matched_id);
		} else {
			ast_log(LOG_NOTICE, "[session:%p] process_cancel_call_control_info, can't get UserID, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
		ast_xml_free_text(content);
	}

	return;
}

static int general_info_process_xml(struct ast_sip_session *session, char *xml)
{
	struct ast_xml_doc  *xml_doc = NULL;
	struct ast_xml_node *root_node = NULL;
	const char *root_node_name = NULL;

	if (!session || !xml) {
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		return -1;
	}

	root_node = ast_xml_get_root(xml_doc);
	if (!root_node) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get root node, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_xml_close(xml_doc);
		return -1;
	}

	root_node_name = ast_xml_node_get_name(root_node);
	if (!root_node_name) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get root node name, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_xml_close(xml_doc);
		return -1;
	}
	
	if (!strcmp(root_node_name, "CancelCallControl")) {
		process_cancel_call_control_info(session, root_node);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Can't process action:%s now, endpoint:%s\n",
			session, root_node_name, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_xml_close(xml_doc);

	return 0;
}

static void send_response(struct ast_sip_session *session,
			  struct pjsip_rx_data *rdata, int code)
{
	pjsip_tx_data *tdata;
	pjsip_dialog *dlg = session->inv_session->dlg;

	if (pjsip_dlg_create_response(dlg, rdata, code, NULL, &tdata) == PJ_SUCCESS) {
		struct pjsip_transaction *tsx = pjsip_rdata_get_tsx(rdata);
		pjsip_dlg_send_response(dlg, tsx, tdata);
	}

	return;
}

static enum ast_gs_state_item_type item_name_str_2_prim(const char *item_name)
{
	int i = 0;

	for (i = 0; i <= AST_GS_STATE_ITEM_TYPE_GESTURE; i++) {
		if (!strcmp(item_name, info_state_item_strmap[i][0])) {
			return (enum ast_gs_state_item_type)i;
		}
	}

	return AST_GS_STATE_ITEM_TYPE_UNKNOW;
}

static void gs_state_info_process_terminal_state_xml(struct ast_sip_session *session,
	struct ast_xml_node *node_item, struct ast_sip_gs_state_data *gs_state)
{
	struct ast_xml_node *node_iterator = NULL;
	const char *node_name;
	const char *content;
	char muted[8] = {0};

	for (node_iterator = ast_xml_node_get_children(node_item); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
		node_name = ast_xml_node_get_name(node_iterator);
		content = ast_xml_get_text(node_iterator);
		if (!content) {
			continue;
		}

		if (!strcmp(node_name, "muted")) {
			snprintf(muted, sizeof(muted), "%s", content);
			if (ast_true(muted)) {
				gs_state->muted = 1;
				ast_log(LOG_DEBUG, "[session:%p] Terminal report state muted is true, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}
		ast_xml_free_text(content);
	}

	return;
}

static void gs_state_info_process_gesture_state_xml(struct ast_sip_session *session,
	struct ast_xml_node *node_item, struct ast_sip_gs_state_data *gs_state)
{
	struct ast_xml_node *node_iterator = NULL;
	const char *node_name;
	const char *content;
	char hand_up[8] = {0};
	char channel_name[AST_CHANNEL_NAME] = {0};

	for (node_iterator = ast_xml_node_get_children(node_item); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
		node_name = ast_xml_node_get_name(node_iterator);
		content = ast_xml_get_text(node_iterator);
		if (!content) {
			continue;
		}

		if (!strcmp(node_name, "hand-up")) {
			snprintf(hand_up, sizeof(hand_up), "%s", content);
			if (ast_true(hand_up)) {
				gs_state->hand_up = 1;
				ast_log(LOG_DEBUG, "[session:%p] Terminal report state hand_up is true, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}
		else if (!strcmp(node_name, "chan")) {
			snprintf(channel_name, sizeof(channel_name), "%s", content);
			if (!ast_strlen_zero(channel_name)) {
				ast_copy_string(gs_state->channel_name, channel_name, sizeof(gs_state->channel_name));
				ast_log(LOG_DEBUG, "[session:%p] Terminal report state channel_name is %s, endpoint:%s\n",
					session, channel_name, ast_sorcery_object_get_id(session->endpoint));
			}
		}
		ast_xml_free_text(content);
	}

	return;
}
static int gs_state_info_process_xml(struct ast_sip_session *session, char *xml, struct ast_sip_gs_state_data *gs_state)
{
	struct ast_xml_doc  *xml_doc        = NULL;
	struct ast_xml_node *root_node      = NULL;
	struct ast_xml_node *child_nodes    = NULL;
	struct ast_xml_node *node_item      = NULL;
	const char *node_name;
	const char *item_name;

	if (!xml) {
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		return -1;
	}

	root_node = ast_xml_get_root(xml_doc);
	if (!root_node) {
		ast_xml_close(xml_doc);
		return -1;
	}

	child_nodes = ast_xml_node_get_children(root_node);
	if (!child_nodes) {
		ast_xml_close(xml_doc);
		return -1;
	}

	for (node_item = child_nodes; node_item; node_item = ast_xml_node_get_next(node_item)) {
		node_name = ast_xml_node_get_name(node_item);
		if (strcmp(node_name, "item")) {
			continue;
		}

		item_name = ast_xml_get_attribute(node_item, "name");
		gs_state->type = item_name_str_2_prim(item_name);
		ast_xml_free_attr(item_name);

		switch (gs_state->type) {
			case AST_GS_STATE_ITEM_TYPE_TERMINAL:
				gs_state_info_process_terminal_state_xml(session, node_item, gs_state);
				break;
			case AST_GS_STATE_ITEM_TYPE_GESTURE:
				gs_state_info_process_gesture_state_xml(session, node_item, gs_state);
				break;
			default:
				break;
		}
	}

	ast_xml_close(xml_doc);

	return 0;
}

static int general_info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_msg_body *body = NULL;
	char *content = NULL;

	body = rdata->msg_info.msg->body;

	if (!body || !body->len) {
		ast_log(LOG_ERROR, "general_info_incoming_request, has no body info\n");
		return 0;
	}

	send_response(session, rdata, 200);

	char buf[body ? (body->len + 1) : 0];
	memset(buf, 0, sizeof(buf));
	body->print_body(body, buf, body->len);

	content = (char*) ast_malloc(body->len + 1);
	if (!content) {
		return 0;
	}

	memset(content, 0, body->len + 1);
	strncpy(content, (char*)body->data, body->len);

	general_info_process_xml(session, content);

	ast_log(LOG_NOTICE, "[session:%p] general_info_incoming_request, process this info, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

    if (content) {
        ast_free(content);
		content = NULL;
    }

	return 1;
}

static int gs_state_info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_msg_body *body = NULL;
	char *content = NULL;
	struct ast_sip_gs_state_data gs_state;
	int res = 0;

	body = rdata->msg_info.msg->body;

	if (!body || !body->len) {
		ast_log(LOG_ERROR, "gs_state_info_incoming_request, has no body info\n");
		return 0;
	}
	
	send_response(session, rdata, 200);
	
	char buf[body ? (body->len + 1) : 0];
	memset(buf, 0, sizeof(buf));
	body->print_body(body, buf, body->len);

	content = (char*) ast_malloc(body->len + 1);
	if (!content) {
		return 0;
	}

	memset(content, 0, body->len + 1);
	strncpy(content, (char*)body->data, body->len);

	ast_log(LOG_NOTICE, "[session:%p] gs_state_info_incoming_request, content:%s\n",
		session, content);

	memset(&gs_state, 0, sizeof(struct ast_sip_gs_state_data));

	res = gs_state_info_process_xml(session, content, &gs_state);
	if (!res) {
		ast_log(LOG_NOTICE, "[session:%p] gs_state_info_incoming_request, process this info success, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_indicate_data(session->channel, AST_CONTROL_NOTIFY_GS_STATE, &gs_state, sizeof(struct ast_sip_gs_state_data));
	} else {
		ast_log(LOG_WARNING, "[session:%p] gs_state_info_incoming_request, process this info failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

    if (content) {
        ast_free(content);
		content = NULL;
    }

	return 1;
}

static int update_register_trunk_info(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t trunk = { "X-GS-Trunk", 10 };

	if (!session || !rdata) {
		return -1;
	}

	if (!session->ucm_trunk) {
		hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &trunk, NULL);
		if (!hdr) {
			return -1;
		}
		ast_log(LOG_DEBUG, "[session:%p] update_register_trunk_info, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->ucm_trunk = 1;
	} else{
		ast_log(LOG_DEBUG, "[session:%p] update_register_trunk_info already set ucm_trunk, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (!session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] update_register_trunk_info is not sfu endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ast_sip_session_notfiy_trunk_info(session);

	return 0;
}

static int info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	if (!session || !session->channel || !rdata) {
		return 0;
	}

	if (!update_register_trunk_info(session, rdata)) {
		send_response(session, rdata, 200);
		return 1;
	}

	if (!rdata->msg_info.msg) {
		return 0;
	}

	if (is_general_control_info(rdata)) {
		return general_info_incoming_request(session, rdata);
	}

	if (is_gs_state_info(rdata)) {
		return gs_state_info_incoming_request(session, rdata);
	}

	return 0;
}

static struct ast_sip_session_supplement info_supplement = {
	.method = "INFO",
	.incoming_request  = info_incoming_request,
	.response_priority = AST_SIP_SESSION_BEFORE_MEDIA | AST_SIP_SESSION_AFTER_MEDIA,
};

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&info_supplement);
	
	return 0;
}

static int load_module(void)
{
	ast_sip_session_register_supplement(&info_supplement);

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP INFO Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip,res_pjsip_session",
);

