/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Cui Xiaoxia <xxcui@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_pubsub</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"


#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#include "asterisk/module.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/res_pjsip_sca.h"
#include "asterisk/logger.h"
#include "asterisk/astobj2.h"
#include "asterisk/taskprocessor.h"

AST_MODULE_LOG("pjsip");

struct line_seize_subscription {
	/*! The SIP subscription */
	struct ast_sip_subscription *sip_sub;
	/*! The serializer to use for notifications */
	struct ast_taskprocessor *serializer;
	/*! The shared line */
	char sharedline[AST_MAX_EXTENSION];
	/*! The line-index */
	int line_index;
	/*! The buf to store call-info header */
	char call_info[256];
	AST_LIST_ENTRY(line_seize_subscription) next;
};

AST_RWLIST_HEAD_STATIC(line_seize_subscriptions, line_seize_subscription);

static void subscription_shutdown(struct ast_sip_subscription *sub);
static int new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource);
static int subscription_established(struct ast_sip_subscription *sub);
static void *get_notify_data(struct ast_sip_subscription *sub);
static int line_seize_terminated(char* sharedline, int line_index);


struct ast_sip_notifier line_seize_notifier = {
	.new_subscribe = new_subscribe,
	.subscription_established = subscription_established,
	.get_notify_data = get_notify_data,
};

struct ast_sip_subscription_handler line_seize_handler = {
	.event_name = "line-seize",
	.subscription_shutdown = subscription_shutdown,
	.notifier = &line_seize_notifier,
};

struct ast_sip_line_seize_handler sca_seize_handler = {
	.seize_terminated = line_seize_terminated,
};

static void line_seize_subscription_destructor(void *obj)
{
	struct line_seize_subscription *sub = obj;

	ast_sip_subscription_destroy(sub->sip_sub);
	sub->sip_sub = NULL;
	ast_taskprocessor_unreference(sub->serializer);
}

/*!
 * \internal
 * \brief Allocates an line_seize_subscription object.
 *
 * Creates the underlying SIP subscription for the given request. First makes
 * sure that there are registered handler and provider objects available.
 */
static struct line_seize_subscription *line_seize_subscription_alloc(
		struct ast_sip_subscription *sip_sub, struct ast_sip_endpoint *endpoint)
{
	struct line_seize_subscription *line_seize_sub;

	line_seize_sub = ao2_alloc(sizeof(*line_seize_sub), line_seize_subscription_destructor);
	if (!line_seize_sub) {
		return NULL;
	}

	line_seize_sub->sip_sub = sip_sub;

	/* We keep our own reference to the serializer as there is no guarantee in state_changed
	 * that the subscription tree is still valid when it is called. This can occur when
	 * the subscription is terminated at around the same time as the state_changed
	 * callback is invoked.
	 */
	line_seize_sub->serializer = ao2_bump(ast_sip_subscription_get_serializer(sip_sub));
	return line_seize_sub;
}

static struct ast_datastore_info ds_info = { };
static const char ds_name[] = "line-seize datastore";

/*!
 * \internal
 * \brief Add a datastore for line_seize_subscription.
 *
 * Adds the line_seize_subscription wrapper object to a datastore so it can be retrieved
 * later based upon its association with the ast_sip_subscription.
 */
static int line_seize_add_datastore(struct line_seize_subscription  *line_seize_sub)
{
	RAII_VAR(struct ast_datastore *, datastore,
		 ast_sip_subscription_alloc_datastore(&ds_info, ds_name), ao2_cleanup);

	if (!datastore) {
		return -1;
	}

	datastore->data = line_seize_sub;
	ast_sip_subscription_add_datastore(line_seize_sub->sip_sub, datastore);
	ao2_ref(line_seize_sub, +1);
	return 0;
}

/*!
 * \internal
 * \brief Get the line_seize_subscription object associated with the given
 * ast_sip_subscription in the datastore.
 */
static struct line_seize_subscription *get_line_seize_sub(
	struct ast_sip_subscription *sub)
{
	RAII_VAR(struct ast_datastore *, datastore,
		 ast_sip_subscription_get_datastore(sub, ds_name), ao2_cleanup);

	return datastore ? datastore->data : NULL;
}

static void add_line_seize_subscription(struct line_seize_subscription *obj)
{
	SCOPED_LOCK(lock, &line_seize_subscriptions, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&line_seize_subscriptions, obj, next);
}

static void remove_line_seize_subscription(struct line_seize_subscription *obj)
{
	struct line_seize_subscription *i;
	SCOPED_LOCK(lock, &line_seize_subscriptions, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&line_seize_subscriptions, i, next) {
		if (i == obj) {
			AST_RWLIST_REMOVE_CURRENT(next);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

static struct line_seize_subscription *find_line_seize_subscription(char* sharedline, int line_index)
{
	struct line_seize_subscription *i;
	SCOPED_LOCK(lock, &line_seize_subscriptions, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	if (AST_LIST_EMPTY(&line_seize_subscriptions)) {
		return NULL;
	}

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&line_seize_subscriptions, i, next) {
		if (!strcmp(i->sharedline, sharedline) && (i->line_index == line_index)) {
			return i;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

	return NULL;
}

static int notify_line_seize_terminated(void *obj)
{
	RAII_VAR(struct line_seize_subscription *, task_data, obj, ao2_cleanup);
	struct line_seize_subscription *line_seize_sub = task_data;

	if (NULL == line_seize_sub->sip_sub) {
		return -1;
	}

	if (ast_sip_notify_terminated(line_seize_sub->sip_sub)) {
		ast_log(LOG_WARNING, "ast_sip_notify_terminated failed, line_seize_sub[%p]", line_seize_sub);
		return -1;
	}

	return 0;
}

static void subscription_shutdown(struct ast_sip_subscription *sub)
{
	struct ast_sip_endpoint *endpoint = ast_sip_subscription_get_endpoint(sub);
	struct line_seize_subscription *line_seize_sub = get_line_seize_sub(sub);

	if (!line_seize_sub) {
		return;
	}

	ast_sip_line_unseize(line_seize_sub->sharedline, (char *)endpoint->aors, line_seize_sub->line_index);

	ast_sip_subscription_remove_datastore(line_seize_sub->sip_sub, ds_name);
	remove_line_seize_subscription(line_seize_sub);

	/* remove data store reference */
	ao2_cleanup(line_seize_sub);
}

static int new_subscribe(struct ast_sip_endpoint *endpoint, const char *resource)
{
	if (strcmp(endpoint->aors, resource)) {
		ast_log(LOG_NOTICE, "Call-info subscription failed: Extension %s does not subscribe itself but %s\n",
			endpoint->aors, resource);
		return 403;
	}

	if (!endpoint->sca_enable || ast_strlen_zero(endpoint->sca_sharedline)) {
		ast_log(LOG_NOTICE, "Call-info subscription failed: Extension %s does not enable SCA\n", resource);
		return 403;
	}

	if (ast_sip_line_check_and_seize((char *)endpoint->sca_sharedline, (char *)resource, endpoint->line_index)) {
		ast_log(LOG_NOTICE, "Call-info subscription failed: Extension %s line check failed\n", resource);
		return 480;
	}

	return 200;
}

static int subscription_established(struct ast_sip_subscription *sip_sub)
{
	struct ast_sip_endpoint *endpoint = ast_sip_subscription_get_endpoint(sip_sub);
	const char *sharedline = endpoint->sca_sharedline;
	struct line_seize_subscription *line_seize_sub;
	char uri[PJSIP_MAX_URL_SIZE] = {0};
	char *port;

	if (!(line_seize_sub = line_seize_subscription_alloc(sip_sub, endpoint))) {
		ao2_cleanup(endpoint);
		return -1;
	}

	ast_copy_string(line_seize_sub->sharedline, sharedline, sizeof(line_seize_sub->sharedline));
	line_seize_sub->line_index = endpoint->line_index;

	ast_sip_subscription_get_local_uri(sip_sub, uri, sizeof(uri));
	port = strrchr(uri, ':');
	if (port)
		*port = '\0';

	sprintf(line_seize_sub->call_info, "<%s>;appearance-index=%d", uri, line_seize_sub->line_index);

	/* Go ahead and cleanup the endpoint since we don't need it anymore */
	ao2_cleanup(endpoint);

	if (line_seize_add_datastore(line_seize_sub)) {
		ast_log(LOG_WARNING, "Unable to add to call info subscription datastore.\n");
		ao2_cleanup(line_seize_sub);
		return -1;
	}

	add_line_seize_subscription(line_seize_sub);

	ao2_cleanup(line_seize_sub);
	return 0;
}

static void *get_notify_data(struct ast_sip_subscription *sub)
{
	struct line_seize_subscription *line_seize_sub;

	line_seize_sub = get_line_seize_sub(sub);
	if (!line_seize_sub) {
		return NULL;
	}

	return line_seize_sub->call_info;
}

static int line_seize_terminated(char* sharedline, int line_index)
{
	struct line_seize_subscription *line_seize_sub;

	if (!sharedline || line_index <= 0) {
		return -1;
	}

	line_seize_sub = find_line_seize_subscription(sharedline, line_index);
	if (!line_seize_sub) {
		return -1;
	}

	if (!line_seize_sub->sip_sub) {
		return -1;
	}

	ao2_ref(line_seize_sub, +1);
	if (ast_sip_push_task_synchronous(line_seize_sub->serializer, notify_line_seize_terminated,
		line_seize_sub)) {
		remove_line_seize_subscription(line_seize_sub);
		ast_sip_subscription_remove_datastore(line_seize_sub->sip_sub, ds_name);
		ao2_cleanup(line_seize_sub);   /* remove data store reference */
		ao2_cleanup(line_seize_sub);   /* remove current task reference */
	}

	return 0;
}

static int load_module(void)
{
	if (ast_sip_register_subscription_handler(&line_seize_handler)) {
		ast_log(LOG_WARNING, "Unable to register subscription handler %s\n",
			line_seize_handler.event_name);
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_sip_register_line_seize_handler(&sca_seize_handler);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_subscription_handler(&line_seize_handler);
	ast_sip_unregister_line_seize_handler(&sca_seize_handler);

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Line Seize Notifications",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
		.requires = "res_pjsip,res_pjsip_pubsub",
);

