
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2018, Grandstream Networks.
 *
 * Yang Yanfei <yfyang@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
    <depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>

#include "asterisk/app.h"
#include "asterisk/callerid.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/module.h"
#include "asterisk/strings.h"

AST_MODULE_LOG("pjsip");

#define MESSAGE_GS_TYPE        "application"
#define MESSAGE_GS_SUBTYPE     "x-gs-message+xml"
#define MESSAGE_PLAIN_TYPE     "text"
#define MESSAGE_PLAIN_SUBTYPE  "plain"

ast_mutex_t  mc_lock;

enum mc_msg_status {
    MC_INITIAL      = 0,
    MC_SENDING      = 1,
    MC_SEND_SUCCESS = 2,
    MC_SEND_FAILED  = 3,
    MC_OFFLINE      = 4,
};

struct mc_msg_data {
	struct ast_sip_endpoint  *endpoint;
	struct ast_sip_body      *body;
	char   message_id[32];    /* the uuid of the message */
	char   source[80];        /* the source/sender of the message */
	char   contact_uri[128];
#ifdef BUILD_ON_CLOUD_PBX
	char   sbc_endpoint_addr[80];
#endif
};

/* the member to receive message */
struct message_member {
    char extension[80];
    enum mc_msg_status status;
	AST_LIST_ENTRY(message_member) entry;
};


/*! \brief A message record */
#if 0
Action: SendMessage
ActionID: <value>
MessageId: <value>      //每次触发的消息有一个唯一的标识id
MessageType: <value>    //消息类型：消息中心页面下发的是 "notification", 会议中的聊天消息为“chat”
Timestamp: <value>      //消息发送的时间点
Name: <value>           //本次广播名称
Source: <value>         //消息来源（发送者）
Text: <value>           //文本消息内容
File: <value>           //文件url（可以直接打开该url下载文件），如消息不带文件，则不包含该字段
Destination: <value>    //消息发送目的号码，多个号码用','分隔，如果消息中心一次指定的目的成员太多，建议分成多次下发
#endif

struct message_entry {
	char   message_id[32];      /* the uuid of the message */
	char   message_name[256];    /* the topic of the message */
	char   source[256];          /* the source/sender of the message */
	char   text[2048];          /* the plain text */
	char   file[256];           /* the file url */
	long   timestamp;           /* The time that the message send */
	int    is_notification;     /* 1: notification, 0: chat */
    int    member_num;          /* the number of the members */
    AST_LIST_HEAD(, message_member) members; /* the message receivers list */
	AST_LIST_ENTRY(message_entry) entry;
};


/*!
 * \brief A structure for data used by the message_task thread
 */
static struct {
	pthread_t thread;   /*! The message dispatch thread ID */
	ast_cond_t  cond;
	ast_mutex_t lock;

	AST_LIST_HEAD(, message_entry) messages;
	unsigned int message_num;

} message_task = {
	.thread = AST_PTHREADT_NULL,
};


static const struct {
    const char *suffix;
    const char *mime_type;
} suffix2mime[] = {
    {"aac", "audio/aac"},
    {"abw", "application/x-abiword"},
    {"arc", "application/x-freearc"},
    {"avi", "video/x-msvideo"},
    {"azw", "application/vnd.amazon.ebook"},
    {"bin", "application/octet-stream"},
    {"bmp", "image/bmp"},
    {"bz",  "application/x-bzip"},
    {"bz2", "application/x-bzip2"},
    {"csh", "application/x-csh"},
    {"css", "text/css"},
    {"csv", "text/csv"},
    {"doc", "application/msword"},
    {"docx", "application/vnd.openxmlformats-officedocument.wordprocessingml.document"},
    {"eot", "application/vnd.ms-fontobject"},
    {"epub", "application/epub+zip"},
    {"gif", "image/gif"},
    {"htm", "text/html"},
    {"html", "text/html"},
    {"ico", "image/vnd.microsoft.icon"},
    {"ics", "text/calendar"},
    {"jar", "application/java-archive"},
    {"jpeg", "image/jpeg"},
    {"jpg", "image/jpeg"},
    {"js",  "text/javascript"},
    {"json", "application/json"},
    {"jsonld", "application/ld+json"},
    {"mid", "audio/midi"},
    {"midi", "audio/midi"},
    {"mjs", "text/javascript"},
    {"mp3", "audio/mpeg"},
    {"mpeg", "video/mpeg"},
    {"mpkg", "application/vnd.apple.installer+xml"},
    {"odp", "application/vnd.oasis.opendocument.presentation"},
    {"ods", "application/vnd.oasis.opendocument.spreadsheet"},
    {"odt", "application/vnd.oasis.opendocument.text"},
    {"oga", "audio/ogg"},
    {"ogv", "video/ogg"},
    {"ogx", "application/ogg"},
    {"otf", "font/otf"},
    {"png", "image/png"},
    {"pdf", "application/pdf"},
    {"ppt", "application/vnd.ms-powerpoint"},
    {"pptx", "application/vnd.openxmlformats-officedocument.presentationml.presentation"},
    {"rar", "application/x-rar-compressed"},
    {"rtf", "application/rtf"},
    {"sh",  "application/x-sh"},
    {"svg", "image/svg+xml"},
    {"swf", "application/ogg"},
    {"tar", "application/x-tar"},
    {"tif", "image/tiff"},
    {"tiff", "image/tiff"},
    {"ttf", "font/ttf"},
    {"txt", "text/plain"},
    {"vsd", "application/vnd.visio"},
    {"wav", "audio/wav"},
    {"weba", "audio/webm"},
    {"webm", "video/webm"},
    {"webp", "image/webp"},
    {"woff", "font/woff"},
    {"woff2", "font/woff2"},
    {"xhtml", "application/xhtml+xml"},
    {"xls", "application/vnd.ms-excel"},
    {"xlsx", "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"},
    {"xml", "application/xml"},
    {"xul", "application/vnd.mozilla.xul+xml"},
    {"zip", "application/zip"},
    {"3gp", "video/3gpp"},
    {"3g2", "video/3gpp2"},
    {"7z", "application/x-7z-compressed"},
};

static const char *mc_file_url2mimetype(char *file)
{
    int   i = 0;
    char *suffix = NULL;

    suffix = strrchr(file, '.');
    if (!suffix) {
        return "application/octet-stream";
    }

    suffix++;

    for (i=0; i<ARRAY_LEN(suffix2mime); i++) {
        if (!strcmp(suffix, suffix2mime[i].suffix)) {
            return suffix2mime[i].mime_type;
        }
    }

    return "application/octet-stream";
}

static int ReplaceEscapeChars(char *intext, char *outtext, int len, char *oldstr, char newchar)
{
	char *ptr = NULL;
	char *data = ast_strdupa(intext);
	char *start = data;
	char *pos = outtext;

	if (!intext || !outtext || !len || !oldstr) {
		return -1;
	}

	while ((ptr = strstr(start, oldstr))) {
		*ptr = newchar;
		strncpy(pos, start, (ptr-start+1));
		start = ptr+strlen(oldstr);
		pos = outtext + strlen(outtext);
	}

	if (start && pos && (start != data)) {
		strcpy(pos, start);
	}

	if (!strlen(outtext)) {
		snprintf(outtext, len, "%s", data);
	}

	return 0;
}

static struct ast_str *create_gs_message_body(struct message_entry *data)
{
	struct ast_str *body_text = NULL;

	if (!data) {
		return NULL;
	}

	body_text = ast_str_create(2048);
	if (!body_text) {
		ast_log(LOG_ERROR, "Alloc buffer for x-gs-message body failed.\n");
		return NULL;
	}

	ast_str_append(&body_text, 0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	ast_str_append(&body_text, 0, "<message-info>\n");
    ast_str_append(&body_text, 0, "  <message id=\"%s\" type=\"%s\" >\n", data->message_id, data->is_notification ? "notification" : "chat");
    ast_str_append(&body_text, 0, "    <name>%s</name>\n", data->message_name);
    ast_str_append(&body_text, 0, "    <source>%s</source>\n", data->source);
    ast_str_append(&body_text, 0, "    <timestamp>%li</timestamp>\n", data->timestamp);
    ast_str_append(&body_text, 0, "    <msg>\n");

    if (!ast_strlen_zero(data->text)) {
        ast_str_append(&body_text, 0, "      <text>%s</text>\n", data->text);
    }

    if (!ast_strlen_zero(data->file)) {
        ast_str_append(&body_text, 0, "      <file url=\"%s\" type=\"%s\" />\n", data->file, mc_file_url2mimetype(data->file));
    }

    ast_str_append(&body_text, 0, "    </msg>\n");
	ast_str_append(&body_text, 0, "  </message>\n");
	ast_str_append(&body_text, 0, "</message-info>");

	return body_text;
}

static struct ast_str *create_plain_message_body(struct message_entry *data)
{
	struct ast_str *body_text = NULL;
	char buffer[2048] = {0};
	char buffer1[2048] = {0};

	if (!data || ast_strlen_zero(data->text)) {
		return NULL;
	}

	body_text = ast_str_create(2048);
	if (!body_text) {
		ast_log(LOG_ERROR, "Alloc buffer for plain body failed.\n");
		return NULL;
	}

	if (data->is_notification) {
		ast_str_append(&body_text, 0, "Notification: %s\n", strlen(data->message_name) ? data->message_name : "");
	}

	ReplaceEscapeChars((char *)data->text, buffer, sizeof(buffer)-1, "&amp;", '&');
	ReplaceEscapeChars(buffer, buffer1, sizeof(buffer1)-1, "&lt;", '<');

	memset(buffer, 0, sizeof(buffer));
	ReplaceEscapeChars(buffer1, buffer, sizeof(buffer)-1, "&gt;", '>');

	ast_log(LOG_NOTICE, "plain text[%s]", buffer);
	ast_str_append(&body_text, 0, "%s", buffer);
	return body_text;
}

#if 0  /* to process cb on stage 2 */
/* to set message send status */
static void mc_send_cb(void *token, pjsip_event *e)
{
	struct mc_msg_data *msg_data = token;
    int succeed = 0;

    if (!token || !e) {
        return;
    }

	if (e->type == PJSIP_EVENT_TSX_STATE) {
		switch(e->body.tsx_state.type) {
        case PJSIP_EVENT_RX_MSG:
            ast_log(LOG_DEBUG, "MESSAGE ID[%s] to contact[%s] recv response.", msg_data->message_id, msg_data->contact_uri);
            succeed = 1;
            break;
		case PJSIP_EVENT_TRANSPORT_ERROR:
        case PJSIP_EVENT_TIMER:
            ast_log(LOG_DEBUG, "MESSAGE ID[%s] to contact[%s] failed, event[%d].", msg_data->message_id, msg_data->contact_uri, e->body.tsx_state.type);
            break;
        default:
            break;
        }
    }

    /*to set message send status */

    ao2_ref(msg_data, -1);
    return;
}
#endif

static int mc_send_message_to_contact(void *data)
{
	pjsip_tx_data *tdata;
	struct mc_msg_data *msg_data = data;

	if (!msg_data || !msg_data->endpoint || !msg_data->contact_uri || !msg_data->body) {
        ast_log(LOG_NOTICE, "invalid param.");
		return 0;
	}

    if (ast_sip_create_request("MESSAGE", NULL, msg_data->endpoint, msg_data->contact_uri, NULL, &tdata)) {
		ast_log(LOG_WARNING, "Unable to create MESSAGE request to endpoint %s, URI %s\n", ast_sorcery_object_get_id(msg_data->endpoint), msg_data->contact_uri);
		return 0;
	}

    /* set From user */
    pjsip_fromto_hdr *from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, NULL);
    if (from) {
        pjsip_name_addr *id_name_addr = (pjsip_name_addr *)from->uri;
        if (id_name_addr) {
            pjsip_sip_uri *id_uri = pjsip_uri_get_uri(id_name_addr->uri);
            if (id_uri) {
                pj_strdup2(tdata->pool, &id_uri->user, msg_data->source);
            }
        }
    }

#ifdef BUILD_ON_CLOUD_PBX
	if (strlen(msg_data->sbc_endpoint_addr)) {
		ast_sip_add_sbc_endpoint_address_header(msg_data->sbc_endpoint_addr, tdata);
	}
#endif

    /* Add body */
	ast_sip_add_body(tdata, msg_data->body);

    //ao2_ref(msg_data, 1);
	//ast_sip_send_request(tdata, NULL, msg_data->endpoint, msg_data, mc_send_cb);
	ast_sip_send_request(tdata, NULL, msg_data->endpoint, NULL, NULL);

	return 0;
}

static int mc_send_message_to_contacts(struct message_entry *msg_entry, char *exten, struct ast_sip_body *body, char *contacts)
{
	char  delim_t[]= "&";
	char  *token_t = NULL;
	char  *saveptr = NULL;
	struct mc_msg_data  msg_data;

    if (!msg_entry || !exten || !body || !contacts) {
        return -1;
    }

    /* get endpoint object */
    RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", exten), ao2_cleanup);
    if (!endpoint) {
        ast_log(LOG_WARNING, "mc_send_message_to_contacts, can't get endpoint[%s]", exten);
        return -1;
    }

	/* parse the contacts */
	for (token_t = strtok_r(contacts, delim_t, &saveptr); token_t != NULL; token_t = strtok_r(NULL, delim_t, &saveptr)) {
		struct ast_sip_contact *contact = (struct ast_sip_contact *)ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "contact", token_t);

		if (!contact) {
			ast_log(LOG_WARNING, "can't get contact object by id[%s]", token_t);
			continue;
		}

		snprintf(msg_data.contact_uri, sizeof(msg_data.contact_uri), "%s", contact->uri);
		snprintf(msg_data.message_id, sizeof(msg_data.message_id), "%s", msg_entry->message_id);
		snprintf(msg_data.source, sizeof(msg_data.source), "%s", msg_entry->source);
		msg_data.endpoint = endpoint;
		msg_data.body = body;
#ifdef BUILD_ON_CLOUD_PBX
		if (!ast_strlen_zero(contact->sbc_endpoint_addr)) {
			snprintf(msg_data.sbc_endpoint_addr, sizeof(msg_data.sbc_endpoint_addr), "%s", contact->sbc_endpoint_addr);
		}
#endif

        ast_log(LOG_NOTICE, "mc_send_message_to_contacts, contact_uri[%s]", msg_data.contact_uri);
        ast_sip_push_task_synchronous(NULL, mc_send_message_to_contact, &msg_data);
        ao2_cleanup(contact);
    }

    return 0;
}

static int mc_send_process(void *userdata)
{
	struct message_entry   *msg_entry = userdata;
	struct message_member  *member_entry = NULL;
	struct ast_sip_body     gs_body;
	struct ast_sip_body     plain_body;
	struct ast_str *gs_body_text = NULL;
	struct ast_str *plain_body_text = NULL;
    char   gs_buffer[512] = {0};
    char   plain_buffer[512] = {0};

	if (!msg_entry) {
        ast_log(LOG_NOTICE, "msg_entry is null\n");
		return -1;
	}

    if (!msg_entry->member_num) {
        ast_log(LOG_NOTICE, "No member in msg_entry[%s]\n", msg_entry->message_id);
		return -1;
    }

	struct ast_custom_function *gs_contacts_func = ast_custom_function_find("PJSIP_GS_MESSAGE_CONTACTS");
	if (!gs_contacts_func || !gs_contacts_func->read) {
        ast_log(LOG_NOTICE, "No PJSIP_GS_MESSAGE_CONTACTS func\n");
		return -1;
	}

	struct ast_custom_function *plain_contacts_func = ast_custom_function_find("PJSIP_PLAIN_MESSAGE_CONTACTS");
	if (!plain_contacts_func || !plain_contacts_func->read) {
        ast_log(LOG_NOTICE, "No PJSIP_PLAIN_MESSAGE_CONTACTS func\n");
		return -1;
	}

	/* generate application/x-gs-message body text */
	if (!(gs_body_text = create_gs_message_body(msg_entry))) {
		ast_log(LOG_ERROR, "Can't create x-gs-message body.\n");
		return -1;
	}

	/* generate text/plain text */
	if (!(plain_body_text = create_plain_message_body(msg_entry))) {
		ast_log(LOG_ERROR, "Can't create text/plain body.\n");
        ast_free(gs_body_text);
		return -1;
	}

    gs_body.type = MESSAGE_GS_TYPE;
    gs_body.subtype = MESSAGE_GS_SUBTYPE;
	gs_body.body_text = ast_str_buffer(gs_body_text);

    plain_body.type = MESSAGE_PLAIN_TYPE;
    plain_body.subtype = MESSAGE_PLAIN_SUBTYPE;
	plain_body.body_text = ast_str_buffer(plain_body_text);

    AST_LIST_LOCK(&msg_entry->members);
    while ((member_entry = AST_LIST_REMOVE_HEAD(&msg_entry->members, entry))) {
        msg_entry->member_num--;

        memset(gs_buffer, 0, sizeof(gs_buffer));
        memset(plain_buffer, 0, sizeof(plain_buffer));

		if (0 == gs_contacts_func->read(NULL, "PJSIP_GS_MESSAGE_CONTACTS", (char *)member_entry->extension, gs_buffer, sizeof(gs_buffer) - 1)) {
			if (strlen(gs_buffer)) {
				//ast_log(LOG_NOTICE, "get gs_contact[%s]", gs_buffer);
                mc_send_message_to_contacts(msg_entry, member_entry->extension, &gs_body, gs_buffer);
            }
        }

		if (0 == plain_contacts_func->read(NULL, "PJSIP_PLAIN_MESSAGE_CONTACTS", (char *)member_entry->extension, plain_buffer, sizeof(plain_buffer) - 1)) {
			if (strlen(plain_buffer)) {
				//ast_log(LOG_NOTICE, "get plain_buffer[%s]", plain_buffer);
                mc_send_message_to_contacts(msg_entry, member_entry->extension, &plain_body, plain_buffer);
            }
        }

        if (!strlen(gs_buffer) && !strlen(plain_buffer)) {
            member_entry->status = MC_OFFLINE;
        } else {
            member_entry->status = MC_SENDING;
        }

        /* member_entry MOVE TO another list */
        //todo on stage 2
        //just free it now
        ast_free(member_entry);
    }

    AST_LIST_UNLOCK(&msg_entry->members);

	ast_free(gs_body_text);
    ast_free(plain_body_text);

	return 0;
}

static void *mc_thread(void *data)
{
	ast_mutex_lock(&message_task.lock);

	for (;;) {
		struct message_entry  *msg_entry;

		if (message_task.message_num == 0) {
			ast_cond_wait(&message_task.cond, &message_task.lock);
		}

		AST_LIST_LOCK(&message_task.messages);
		while ((msg_entry = AST_LIST_REMOVE_HEAD(&message_task.messages, entry))) {
			message_task.message_num--;

			ast_mutex_unlock(&message_task.lock);
			mc_send_process(msg_entry);
			ast_free(msg_entry);
			ast_mutex_lock(&message_task.lock);
		}

		AST_LIST_UNLOCK(&message_task.messages);
	}

	ast_mutex_unlock(&message_task.lock);
	return NULL;
}

static int UnescapeLFchars(char *intext, char *outtext, int len)
{
    int l = 0;
    char *ptr = NULL;
    char *data = ast_strdupa(intext);

    if (!intext || !outtext || !len) {
        return -1;
    }

    if ((ptr = strstr(data, "%0A"))) {
        *ptr = '\0';
        ptr+=3;
        snprintf(outtext, len, "%s\n", data);

        if (!ast_strlen_zero(ptr)) {
            l = strlen(outtext);
            UnescapeLFchars(ptr, outtext+l, len-l);
        }
    } else {
        snprintf(outtext, len, "%s", data);
    }

    return 0;
}

static int action_send_message(struct mansession *s, const struct message *m)
{
	struct message_entry  *msg_entry = NULL;
	const char *message_id = astman_get_header(m, "MessageId");
	const char *message_type = astman_get_header(m, "MessageType");
	//const char *timestamp = astman_get_header(m, "Timestamp");
	const char *name = astman_get_header(m, "Name");
	const char *source = astman_get_header(m, "Source");
	const char *text = astman_get_header(m, "Text");
	const char *file = astman_get_header(m, "File");
	const char *destination = astman_get_header(m, "Destination");
	char  buffer[2048] = {0};
	char  *members = NULL;
	char  delim_t[]= ",";
	char  *token_t = NULL;
	char  *saveptr = NULL;

	if (ast_strlen_zero(message_id)) {
		astman_send_error(s, m, "No MessageId specified");
		return 0;
	}

	if (ast_strlen_zero(source)) {
		astman_send_error(s, m, "No Source specified");
		return 0;
	}

	if (ast_strlen_zero(destination)) {
		astman_send_error(s, m, "No Destination specified");
		return 0;
	}

    if (ast_strlen_zero(text) && ast_strlen_zero(file)) {
		astman_send_error(s, m, "No Text or File specified");
		return 0;
    }

    /* alloc message_entry */
    msg_entry = ast_calloc(1, sizeof(*msg_entry));
    if (!msg_entry) {
		astman_send_error(s, m, "Alloc object failed.");
		return 0;
    }

    ast_log(LOG_NOTICE, "action_send_message, dest[%s], text[%s]", destination, text);
    msg_entry->timestamp = (long)time(NULL);

    strncpy(msg_entry->message_id, message_id, sizeof(msg_entry->message_id)-1);
    strncpy(msg_entry->source, source, sizeof(msg_entry->source)-1);

    if (!ast_strlen_zero(name)) {
        strncpy(msg_entry->message_name, name, sizeof(msg_entry->message_name) - 1);
    } else {
        memset(msg_entry->message_name, 0, sizeof(msg_entry->message_name));
    }

    if (!ast_strlen_zero(text)) {
        UnescapeLFchars((char *)text, buffer, sizeof(buffer)-1);
        strncpy(msg_entry->text, buffer, sizeof(msg_entry->text) - 1);
    } else {
        memset(msg_entry->text, 0, sizeof(msg_entry->text));
    }

    if (!ast_strlen_zero(file)) {
        strncpy(msg_entry->file, file, sizeof(msg_entry->file) - 1);
    } else {
        memset(msg_entry->file, 0, sizeof(msg_entry->file));
    }

    msg_entry->is_notification = 1;
    if (!ast_strlen_zero(message_type) && !strcmp(message_type, "chat")) {
        msg_entry->is_notification = 0;
    }

    /* split the destination members */
    members = ast_strdupa(destination);
    msg_entry->member_num= 0;
    AST_LIST_HEAD_INIT(&msg_entry->members);

    //ast_log(LOG_NOTICE, "action_send_message, members[%s]", members);
    for (token_t = strtok_r(members, delim_t, &saveptr); token_t != NULL; token_t = strtok_r(NULL, delim_t, &saveptr)) {
        struct message_member *member_entry = NULL;

		if (!(member_entry = ast_calloc(1, sizeof(*member_entry)))) {
			ast_log(LOG_WARNING, "ast_calloc member_entry failed!");
			break;
		}

		member_entry->status = 0;
		strncpy(member_entry->extension, token_t, sizeof(member_entry->extension)-1);

        AST_LIST_INSERT_TAIL(&msg_entry->members, member_entry, entry);
        msg_entry->member_num++;
    }

    AST_LIST_LOCK(&message_task.messages);
    AST_LIST_INSERT_TAIL(&message_task.messages, msg_entry, entry);
    message_task.message_num++;

    AST_LIST_UNLOCK(&message_task.messages);

	astman_send_ack(s, m, "Success");

	ast_mutex_lock(&message_task.lock);
	ast_cond_signal(&message_task.cond);
	ast_mutex_unlock(&message_task.lock);

	return 0;
}


static int unload_module(void)
{
	struct message_entry  *msg_entry = NULL;

	ast_manager_unregister("SendMessage");

	AST_LIST_LOCK(&message_task.messages);

	while ((msg_entry = AST_LIST_REMOVE_HEAD(&message_task.messages, entry))) {
		ast_free(msg_entry);
	}

	AST_LIST_UNLOCK(&message_task.messages);

	if (message_task.thread != AST_PTHREADT_NULL) {
		ast_mutex_lock(&message_task.lock);
		ast_cond_signal(&message_task.cond);
		ast_mutex_unlock(&message_task.lock);
		pthread_join(message_task.thread, NULL);
	}

	ast_mutex_destroy(&message_task.lock);
	ast_cond_destroy(&message_task.cond);

	return 0;
}

static int load_module(void)
{
	ast_manager_register_xml("SendMessage", EVENT_FLAG_SYSTEM | EVENT_FLAG_CALL, action_send_message);

	ast_mutex_init(&message_task.lock);
	ast_cond_init(&message_task.cond, NULL);
	ast_pthread_create(&message_task.thread, NULL, mc_thread, NULL);

	message_task.message_num = 0;
	AST_LIST_HEAD_INIT(&message_task.messages);

	return AST_MODULE_LOAD_SUCCESS;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Message Center Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_CHANNEL_DEPEND,
);

