
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2018, Grandstream Networks.
 *
 * Yang Yanfei <yfyang@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
    <depend>res_pjsip_session</depend> 
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>

#include "asterisk/app.h"
#include "asterisk/callerid.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/module.h"
#include "asterisk/strings.h"

AST_MODULE_LOG("pjsip");
/*** DOCUMENTATION
	<application name="MissedCall" language="en_US">
		<synopsis>
			To record the missed call info, then the UCM can send NOTIFY to the callee .
		</synopsis>
		<syntax>
			<parameter name="call_type" required="true">
				<para>The call type: CQ or RG.</para>
            </parameter>
			<parameter name="group_name" required="true">
				<para>The name of the queue or ringgroup.</para>
			</parameter>
			<parameter name="group_exten" required="true">
				<para>The extension of the queue or ringgroup.</para>
            </parameter>
			<parameter name="callerid" required="true">
				<para>The ID of the caller party.</para>
            </parameter>
			<parameter name="contacts" required="true">
				<para>The callee contacts.</para>
            </parameter>
		</syntax>
		<description>
			<para>This application will set the missed call info:</para>
		</description>
		<see-also>
			<ref type="function">PJSIP_MISSED_CALL_CONTACTS</ref>
		</see-also>
	</application>
 ***/
 
 
#define MISSED_INFO_TYPE    "application"
#define MISSED_INFO_SUBTYPE "gs-missed-info+xml"
#define MISSED_INFO_EVENT   "gs-missed-call"

static char *missed_app = "MissedCall";


struct missed_call_msg_data {
	struct ast_sip_endpoint   *endpoint;
	struct ast_sip_contact    *contact;
	const struct ast_sip_body *body;
};

/*! \brief A missed call record */
struct missed_call_entry {
	char   userid[80];      /* The endpoint id of the callee user */
	char   call_type[4];    /* The type of the missed call: RG or CQ */ 
	char   group_exten[80]; /* The extension of the CQ or RG */
	char   group_name[80];  /* The name of the CQ or RG */
	char   caller_id[80];   /* The number of the caller side */
	char   caller_name[80]; /* The name of the caller side */
	long   call_time;       /* The time that the missed call occur */
	struct ast_sip_contact  *contact;  /* The contact to send the NOTIFY */

	AST_LIST_ENTRY(missed_call_entry) entry;
};


/*!
 * \brief A structure for data used by the missed call thread
 */
static struct {
	pthread_t thread;   /*! The missed call thread ID */
	ast_cond_t  cond;
	ast_mutex_t lock;

	AST_LIST_HEAD(, missed_call_entry)  missed_calls;
	unsigned int missed_calls_num;

} missed_call = {
	.thread = AST_PTHREADT_NULL,
};


/*
<?xml version="1.0" encoding="UTF-8"?>
<missed-info xmlns="urn:ietf:params:xml:ns:missed-info" entity="sip:5000@192.168.130.212:5060" state="partial">
  <missed-call type="RG" name="RG-GS" extension="6400">
    <caller id="5555" name="User-5555" time=1534347600/>
  </missed-call>  
</missed-info>
*/
static struct ast_str *create_missed_call_body(struct missed_call_entry  *data)
{
	struct ast_str *body_text = NULL;

	if (!data || !data->contact) {
		return NULL;
	}

	body_text = ast_str_create(1000); 
	if (!body_text) {
		ast_log(LOG_ERROR, "Alloc buffer for missed-call Notify failed.\n");
		return NULL;
	}

	ast_str_append(&body_text, 0, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	ast_str_append(&body_text, 0, "<missed-info xmlns=\"urn:ietf:params:xml:ns:missed-info\" entity=\"%s\" state=\"partial\">\n", data->contact->uri);
	ast_str_append(&body_text, 0, "  <missed-call type=\"%s\" name=\"%s\" extension=\"%s\">\n", data->call_type, data->group_name, data->group_exten);
	ast_str_append(&body_text, 0, "    <caller id=\"%s\" name=\"%s\" time=\"%li\"/>\n", data->caller_id, data->caller_name, data->call_time);
	ast_str_append(&body_text, 0, "  </missed-call>\n");
	ast_str_append(&body_text, 0, "</missed-info>");

	return body_text;
}


static int missed_call_send_notify_to_contact(void *data)
{
	pjsip_tx_data *tdata;
	pjsip_sub_state_hdr *sub_state;
	pjsip_event_hdr *hdr_event;
	struct missed_call_msg_data *msg_data = data;
	//const pjsip_hdr *allow_events = pjsip_evsub_get_allow_events_hdr(NULL);

	if (!msg_data || !msg_data->endpoint || !msg_data->contact || !msg_data->body) {
		return 0;
	}

	if (ast_sip_create_request("NOTIFY", NULL, msg_data->endpoint, NULL, msg_data->contact, &tdata)) {
		ast_log(LOG_WARNING, "Unable to create missed_call NOTIFY request to endpoint %s, URI %s\n", ast_sorcery_object_get_id(msg_data->endpoint), msg_data->contact->uri);
		return 0;
	}

	/* Add header,  Subscription-State: terminated */
	sub_state = pjsip_sub_state_hdr_create(tdata->pool);
	pj_cstr(&sub_state->sub_state, "terminated");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *) sub_state);

	/* Add header,  Event: gs-missed-call */
	hdr_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&hdr_event->event_type, MISSED_INFO_EVENT);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *) hdr_event);

	//pjsip_msg_add_hdr(tdata->msg, pjsip_hdr_shallow_clone(tdata->pool, allow_events));

	/* Add body */
	ast_sip_add_body(tdata, msg_data->body);

	ast_sip_send_request(tdata, NULL, msg_data->endpoint, NULL, NULL);

	return 0;
}

static int missed_call_send_process(void *userdata)
{
	struct missed_call_entry     *call_entry = userdata;
	struct missed_call_msg_data   msg_data;
	struct ast_sip_body body;
	struct ast_str *body_text = NULL;

	if (!call_entry) {
		return -1;
	}

	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", call_entry->userid), ao2_cleanup); 

	if (!endpoint) {
		ast_log(LOG_WARNING, "Can't get endpoint by id[%s]\n", call_entry->userid);
		return -1;
	}

	/* generate body text */
	if (!(body_text = create_missed_call_body(call_entry))) {
		ast_log(LOG_ERROR, "Can't create missed call body text.\n");
		return -1;
	}

	body.type = MISSED_INFO_TYPE;
	body.subtype = MISSED_INFO_SUBTYPE;
	body.body_text = ast_str_buffer(body_text);

	msg_data.body = &body;
	msg_data.endpoint = endpoint;
	msg_data.contact  = call_entry->contact;

	ast_sip_push_task_synchronous(NULL, missed_call_send_notify_to_contact, &msg_data);
	ast_free(body_text);

	return 0;
}


static int missed_call_exec(struct ast_channel *chan, const char *data)
{
	struct missed_call_entry  *call_entry = NULL;
	struct ast_sip_contact    *contact = NULL;
	char  *parse = NULL;
#ifdef GRANDSTREAM_NETWORKS
	char  cid_name[AST_MAX_CID_NAME] = {0};
#else
	char  cid_name[80] = {0};
#endif
	char  cid_num[80]  = {0};
	char  delim_t[] = "&";
	char  *token_t = NULL;
	char  *user_id = NULL;
	char  *ptr     = NULL;
	char  *saveptr = NULL;
	time_t  call_time;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(call_type);
		AST_APP_ARG(group_name);
		AST_APP_ARG(group_exten);
		AST_APP_ARG(callerid);
		AST_APP_ARG(contacts);
	);

	if (ast_strlen_zero(data)) {
		return 0; 
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);

	if (!args.call_type || !args.group_name || !args.group_exten || !args.callerid || !args.contacts) {
		ast_log(LOG_WARNING, "missed_call_exec, some para not in the data[%s]", data);
		return 0;
	}

	ast_log(LOG_NOTICE, "missed_call_exec, call_type[%s], group[%s <%s>], callerid[%s], contacts[%s]", 
		args.call_type, args.group_name, args.group_exten, args.callerid, args.contacts);

	/* get caller id and caller name */
	if (!strchr(args.callerid, '<')) {
		strncpy(cid_num, args.callerid, sizeof(cid_num)-1);
		strncpy(cid_name, args.callerid, sizeof(cid_name)-1);
	} else { 
		/* with name, split name and num */
		ast_callerid_split(args.callerid, cid_name, sizeof(cid_name), cid_num, sizeof(cid_num)); 
	}

	call_time = time(NULL);

	/* parse the contacts */
	for (token_t = strtok_r(args.contacts, delim_t, &saveptr); token_t != NULL; token_t = strtok_r(NULL, delim_t, &saveptr)) {
		user_id = ast_strdupa(token_t);
		if ((ptr = strchr(user_id, ';'))) {
			*ptr = '\0';
		}

		contact = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "contact", token_t);
		if (!contact) {
			ast_log(LOG_WARNING, "can't get contact object by id[%s]", token_t);
			continue;
		}

		if (!(call_entry = ast_calloc(1, sizeof(*call_entry)))) {
			ast_log(LOG_WARNING, "ast_calloc failed!!!");
			break;
		}

		call_entry->contact = contact;
		call_entry->call_time = (long)call_time;
		strncpy(call_entry->userid, user_id, sizeof(call_entry->userid)-1);
		strncpy(call_entry->call_type, args.call_type, sizeof(call_entry->call_type)-1);
		strncpy(call_entry->group_exten, args.group_exten, sizeof(call_entry->group_exten)-1);
		strncpy(call_entry->group_name, args.group_name, sizeof(call_entry->group_name)-1);
		strncpy(call_entry->caller_id, cid_num, sizeof(call_entry->caller_id)-1);
		strncpy(call_entry->caller_name, cid_name, sizeof(call_entry->caller_name)-1);

		AST_LIST_LOCK(&missed_call.missed_calls);
		AST_LIST_INSERT_HEAD(&missed_call.missed_calls, call_entry, entry);
		missed_call.missed_calls_num++; 
		AST_LIST_UNLOCK(&missed_call.missed_calls);
	}

	pbx_builtin_setvar_helper(chan, "MISSED_CALL_SETTED", "yes");

	ast_mutex_lock(&missed_call.lock);
	ast_cond_signal(&missed_call.cond);
	ast_mutex_unlock(&missed_call.lock);

	return 0; 
}

static void *missed_call_thread(void *data)
{
	ast_mutex_lock(&missed_call.lock);

	for (;;) {
		struct missed_call_entry  *call_entry;

		if (missed_call.missed_calls_num == 0) {
			ast_cond_wait(&missed_call.cond, &missed_call.lock);
		}

		AST_LIST_LOCK(&missed_call.missed_calls);
		while ((call_entry = AST_LIST_REMOVE_HEAD(&missed_call.missed_calls, entry))) {
			missed_call.missed_calls_num--;  

			ast_mutex_unlock(&missed_call.lock);

			missed_call_send_process(call_entry);

			/* free call entry */
			ao2_ref(call_entry->contact, -1);
			ast_free(call_entry);

			ast_mutex_lock(&missed_call.lock);
		}

		AST_LIST_UNLOCK(&missed_call.missed_calls);
	}

	ast_mutex_unlock(&missed_call.lock);
	return NULL;
}

static int unload_module(void)
{
	struct missed_call_entry  *call_entry;

	ast_unregister_application(missed_app);

	AST_LIST_LOCK(&missed_call.missed_calls);
	while ((call_entry = AST_LIST_REMOVE_HEAD(&missed_call.missed_calls, entry))) {
		/* free call entry */
		ao2_ref(call_entry->contact, -1);
		ast_free(call_entry);
	}
	AST_LIST_UNLOCK(&missed_call.missed_calls);

	/* cancel missed_info_notify_thread */
	if (missed_call.thread != AST_PTHREADT_NULL) {
		ast_mutex_lock(&missed_call.lock);
		ast_cond_signal(&missed_call.cond);
		ast_mutex_unlock(&missed_call.lock);
		pthread_join(missed_call.thread, NULL);
	}

	ast_mutex_destroy(&missed_call.lock);
	ast_cond_destroy(&missed_call.cond);

	return 0;
}

static int load_module(void)
{
	if (ast_register_application_xml(missed_app, missed_call_exec)) {
		ast_log(LOG_WARNING, "register application: MissedCall failed");
		unload_module();
		return -1;
	}

	ast_mutex_init(&missed_call.lock);
	ast_cond_init(&missed_call.cond, NULL);
	ast_pthread_create(&missed_call.thread, NULL, missed_call_thread, NULL);
	missed_call.missed_calls_num = 0;
	AST_LIST_HEAD_INIT(&missed_call.missed_calls);

	return AST_MODULE_LOAD_SUCCESS;
}


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Missed Call Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_CHANNEL_DEPEND,
);

