/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/module.h"
#include "asterisk/acl.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/haserver.h"
#endif

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("pjsip");

#ifdef HAVE_GCC601X
#define LOCAL_HOST "192.0.2.100"
#else
#define LOCAL_HOST "127.0.0.1"
#endif

#endif

static void rewrite_uri(pjsip_rx_data *rdata, pjsip_sip_uri *uri)
{
	pj_cstr(&uri->host, rdata->pkt_info.src_name);
	uri->port = rdata->pkt_info.src_port;
	if (!strcasecmp("WSS", rdata->tp_info.transport->type_name)) {
		/* WSS is special, we don't want to overwrite the URI at all as it needs to be ws */
	} else if (strcasecmp("udp", rdata->tp_info.transport->type_name)) {
		uri->transport_param = pj_str(rdata->tp_info.transport->type_name);
	} else {
		uri->transport_param.slen = 0;
	}
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for bug 149435, 2019.11.11 */
static unsigned char cmp_uri_host_and_src_addr(pjsip_rx_data *rdata, pjsip_sip_uri *uri)
{
	char host[PJ_INET6_ADDRSTRLEN] = { 0 };

	ast_copy_pj_str(host, &uri->host, PJ_INET6_ADDRSTRLEN);

	if (strcmp(host, rdata->pkt_info.src_name)) {
		ast_log(LOG_DEBUG, "rdata src address:%s is no equal to uri host:%s, rewrite uri host\n", rdata->pkt_info.src_name, host);
		return FALSE;
	}

	return TRUE;
}

static void rewrite_route_uri(pjsip_rx_data *rdata, pjsip_sip_uri *uri, pjsip_dialog *dlg)
{
	if (dlg && dlg->pool && !cmp_uri_host_and_src_addr(rdata, uri)) {
		pj_strdup2_with_null(dlg->pool, &uri->host, rdata->pkt_info.src_name);
	}
	uri->port = rdata->pkt_info.src_port;
	if (!strcasecmp("WSS", rdata->tp_info.transport->type_name)) {
		/* WSS is special, we don't want to overwrite the URI at all as it needs to be ws */
	} else if (strcasecmp("udp", rdata->tp_info.transport->type_name)) {
		uri->transport_param = pj_str(rdata->tp_info.transport->type_name);
	} else {
		uri->transport_param.slen = 0;
	}

	return;
}
#endif

/*
 * Update the Record-Route headers in the request or response and in the dialog
 * object if exists.
 *
 * When NAT is in use, the address of the next hop in the SIP may be incorrect.
 * To address this  asterisk uses two strategies in parallel:
 *  1. intercept the messages at the transaction level and rewrite the
 *     messages before arriving at the dialog layer
 *  2. after the application processing, update the dialog object with the
 *     correct information
 *
 * The first strategy has a limitation that the SIP message may not have all
 * the information required to determine if the next hop is in the route set
 * or in the contact. Causing risk that asterisk will update the Contact on
 * receipt of an in-dialog message despite there being a route set saved in
 * the dialog.
 *
 * The second strategy has a limitation that not all UAC layers have interfaces
 * available to invoke this module after dialog creation.  (pjsip_sesion does
 * but pjsip_pubsub does not), thus this strategy can't update the dialog in
 * all cases needed.
 *
 * The ideal solution would be to implement an "incomming_request" event
 * in pubsub module that can then pass the dialog object to this module
 * on SUBSCRIBE, this module then should add itself as a listener to the dialog
 * for the subsequent requests and responses & then be able to properly update
 * the dialog object for all required events.
 */

static int rewrite_route_set(pjsip_rx_data *rdata, pjsip_dialog *dlg)
{
	pjsip_rr_hdr *rr = NULL;
	pjsip_sip_uri *uri;
	int res = -1;
	int ignore_rr = 0;
	int pubsub = 0;

	if (rdata->msg_info.msg->type == PJSIP_RESPONSE_MSG) {
		pjsip_hdr *iter;
		for (iter = rdata->msg_info.msg->hdr.prev; iter != &rdata->msg_info.msg->hdr; iter = iter->prev) {
			if (iter->type == PJSIP_H_RECORD_ROUTE) {
				rr = (pjsip_rr_hdr *)iter;
				break;
			}
		}
	} else if (pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_register_method)) {
		rr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_RECORD_ROUTE, NULL);
	} else {
		/**
		 * Record-Route header has no meaning in REGISTER requests
		 * and should be ignored
		 */
		ignore_rr = 1;
	}

	if (!pjsip_method_cmp(&rdata->msg_info.cseq->method, &pjsip_subscribe_method) ||
		!pjsip_method_cmp(&rdata->msg_info.cseq->method, &pjsip_notify_method)) {
		/**
		 * There is currently no good way to get the dlg object for a pubsub dialog
		 * so we will just look at the rr & contact of the current message and
		 * hope for the best
		 */
		pubsub = 1;
	}

	if (rr) {
		uri = pjsip_uri_get_uri(&rr->name_addr);
		rewrite_uri(rdata, uri);
		res = 0;
	}

	if (dlg && !pj_list_empty(&dlg->route_set) && !dlg->route_set_frozen) {
		pjsip_routing_hdr *route = dlg->route_set.next;
		uri = pjsip_uri_get_uri(&route->name_addr);
#ifdef GRANDSTREAM_NETWORKS
		rewrite_route_uri(rdata, uri, dlg);
#else
		rewrite_uri(rdata, uri);
#endif
		res = 0;
	}

	if (!dlg && !rr && !ignore_rr  && !pubsub && rdata->msg_info.to->tag.slen){
		/**
		 * Even if this message doesn't have any route headers
		 * the dialog may, so wait until a later invocation that
		 * has a dialog reference to make sure there isn't a
		 * previously saved routset in the dialog before deciding
		 * the contact needs to be modified
		 */
		res = 0;
	}

	return res;
}

static int rewrite_contact(pjsip_rx_data *rdata, pjsip_dialog *dlg)
{
	pjsip_contact_hdr *contact;

	contact = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
	if (contact && !contact->star && (PJSIP_URI_SCHEME_IS_SIP(contact->uri) || PJSIP_URI_SCHEME_IS_SIPS(contact->uri))) {
		pjsip_sip_uri *uri = pjsip_uri_get_uri(contact->uri);

		rewrite_uri(rdata, uri);

		if (dlg && pj_list_empty(&dlg->route_set) && (!dlg->remote.contact
			|| pjsip_uri_cmp(PJSIP_URI_IN_REQ_URI, dlg->remote.contact->uri, contact->uri))) {
			dlg->remote.contact = (pjsip_contact_hdr*)pjsip_hdr_clone(dlg->pool, contact);
			dlg->target = dlg->remote.contact->uri;
		}
		return 0;
	}

	return -1;
}


#ifdef GRANDSTREAM_NETWORKS
static int ast_sip_rdata_add_header(pjsip_rx_data *rdata, const char *name, const char *value)
{
	pj_str_t hdr_name;
	pj_str_t hdr_value;
	pjsip_generic_string_hdr *hdr;

	pj_cstr(&hdr_name, name);
	pj_cstr(&hdr_value, value);

	hdr = pjsip_generic_string_hdr_create(rdata->tp_info.pool, &hdr_name, &hdr_value);

	pjsip_msg_add_hdr(rdata->msg_info.msg, (pjsip_hdr *) hdr);

	return 0;
}

static void record_contact_original_address(pjsip_rx_data *rdata)
{
	char host[PJ_INET6_ADDRSTRLEN] = { 0 };
	pjsip_contact_hdr *contact = NULL;
#if 0
	if (pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_register_method)) {
		return;
	}
#endif

#ifdef BUILD_ON_CLOUD_PBX
	pj_str_t  sbc_addr_hdr_str = { "SBC-Endpoint-Address", 20 };
	pjsip_generic_string_hdr* sbc_addr_hdr = NULL;

	if ((sbc_addr_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &sbc_addr_hdr_str, NULL))) {
		if (sbc_addr_hdr->hvalue.slen) {
			if (ast_sip_check_ssh_proxy_flag(rdata)) {
				ast_copy_pj_str(host, &sbc_addr_hdr->hvalue, PJ_INET6_ADDRSTRLEN);
				ast_sip_rdata_add_header(rdata, "contact-original-address", host);

				ast_log(LOG_DEBUG, "Record sbc original address:%s\n", host);
				return;
			}
		}
	}
#endif

	pj_str_t  addr_hdr_str = { "GS-GDMS-Endpoint-Address", 24 };
	pjsip_generic_string_hdr* gdms_addr_hdr = NULL;

	if ((gdms_addr_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &addr_hdr_str, NULL))) {
		if (gdms_addr_hdr->hvalue.slen) {
			if (ast_sip_check_ssh_proxy_flag(rdata)) {
				ast_copy_pj_str(host, &gdms_addr_hdr->hvalue, PJ_INET6_ADDRSTRLEN);
				ast_sip_rdata_add_header(rdata, "contact-original-address", host);

				ast_log(LOG_DEBUG, "Record contact original address:%s\n", host);
				return;
			}
		}
	}

	if (strcasecmp(rdata->tp_info.transport->type_name, "udp")) {
		contact = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
		if (contact && !contact->star && (PJSIP_URI_SCHEME_IS_SIP(contact->uri) || PJSIP_URI_SCHEME_IS_SIPS(contact->uri))) {
			pjsip_sip_uri *uri = pjsip_uri_get_uri(contact->uri);

			ast_copy_pj_str(host, &uri->host, PJ_INET6_ADDRSTRLEN);
			ast_sip_rdata_add_header(rdata, "contact-original-address", host);

			ast_log(LOG_DEBUG, "Record contact original address:%s\n", host);
		}
	}

	return;
}
#endif

static pj_bool_t handle_rx_message(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata)
{
	pjsip_dialog *dlg = pjsip_rdata_get_dlg(rdata);

	if (!endpoint) {
		return PJ_FALSE;
	}

	if (endpoint->nat.rewrite_contact) {
		/* rewrite_contact is intended to ensure we send requests/responses to
		 * a routeable address when NAT is involved. The URI that dictates where
		 * we send requests/responses can be determined either by Record-Route
		 * headers or by the Contact header if no Record-Route headers are present.
		 * We therefore will attempt to rewrite a Record-Route header first, and if
		 * none are present, we fall back to rewriting the Contact header instead.
		 */
#ifdef GRANDSTREAM_NETWORKS
		record_contact_original_address(rdata);
#endif
		if (rewrite_route_set(rdata, dlg)) {
			rewrite_contact(rdata, dlg);
		}
	}

	if (endpoint->nat.force_rport) {
		rdata->msg_info.via->rport_param = rdata->pkt_info.src_port;
	}

	return PJ_FALSE;
}

static pj_bool_t nat_on_rx_message(pjsip_rx_data *rdata)
{
	pj_bool_t res;
	struct ast_sip_endpoint *endpoint;

	endpoint = ast_pjsip_rdata_get_endpoint(rdata);
	res = handle_rx_message(endpoint, rdata);
	ao2_cleanup(endpoint);
	return res;
}

/*! \brief Structure which contains information about a transport */
struct request_transport_details {
	/*! \brief Type of transport */
	enum ast_transport type;
	/*! \brief Potential pointer to the transport itself, if UDP */
	pjsip_transport *transport;
	/*! \brief Potential pointer to the transport factory itself, if TCP/TLS */
	pjsip_tpfactory *factory;
	/*! \brief Local address for transport */
	pj_str_t local_address;
	/*! \brief Local port for transport */
	int local_port;
};

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for Bug 199072, 2021.7.7 */
static int judge_if_transport_host_match_local_address(struct ast_sip_transport_state *transport_state, struct request_transport_details *details)
{
	if (!transport_state || !transport_state->factory || !details) {
		ast_log(LOG_DEBUG, "judge_if_transport_host_match_local_address params error\n");
		return 0;
	}

	if (!pj_strcmp(&transport_state->factory->addr_name.host, &details->local_address)) {
		return 1;
	}

#ifdef BUILD_ON_CLOUD_PBX
	ast_log(LOG_DEBUG, "judge_if_transport_host_match_local_address is cloudUCM, just return true\n");
	return 1;
#endif

	/* Added by yfyang, for Bug 275136 */
	if (details->local_address.slen && details->local_address.ptr) {
		char factory_addr[64] = { 0 };
		char localhost[64] = {0};
		char local_address[256] = {0};

		ast_copy_string(local_address, details->local_address.ptr, details->local_address.slen);
		if (strchr(local_address, ':')) {
			strcpy(localhost, "fe80::");
		} else {
			strcpy(localhost, LOCAL_HOST);
		}

		snprintf(factory_addr, sizeof(factory_addr), "%.*s", (int)pj_strlen(&transport_state->factory->addr_name.host), pj_strbuf(&transport_state->factory->addr_name.host));
		if (!strncmp(factory_addr, localhost, strlen(localhost))) {
			return 1;
		}
	}

	if (HA_ROLE_ACTIVE == ast_ha_get_role()) {
		/* Todo HA just support IPV4, need to modify when support IPV6 */
		if (transport_state->host.addr.sa_family != pj_AF_INET()) {
			ast_log(LOG_DEBUG, "It HA mode, transport host address is not IPV4, can't matched\n");
			return 0;
		}

		pj_str_t host_ip_str = pj_str((char *)ast_sip_get_host_ip_string(pj_AF_INET()));
		if (!pj_strcmp(&host_ip_str, &details->local_address)) {
			ast_log(LOG_DEBUG, "It HA mode, local address matched float IP\n");
			return 1;
		}
	}

	if (pj_strlen(&transport_state->factory->addr_name.host) > 0) {
		ast_log(LOG_DEBUG, "factory host[%.*s], not match\n", (int)pj_strlen(&transport_state->factory->addr_name.host), pj_strbuf(&transport_state->factory->addr_name.host));
	}
	return 0;
}
#endif

/*! \brief Callback function for finding the transport the request is going out on */
static int find_transport_state_in_use(void *obj, void *arg, int flags)
{
	struct ast_sip_transport_state *transport_state = obj;
	struct request_transport_details *details = arg;
#ifdef GRANDSTREAM_NETWORKS
	int res = 0;
#endif

	/* If an explicit transport or factory matches then this is what is in use, if we are unavailable
	 * to compare based on that we make sure that the type is the same and the source IP address/port are the same
	 */
	if (transport_state && ((details->transport && details->transport == transport_state->transport) ||
		(details->factory && details->factory == transport_state->factory) ||
		((details->type == transport_state->type) && (transport_state->factory) &&
#ifdef GRANDSTREAM_NETWORKS
			judge_if_transport_host_match_local_address(transport_state, details)
#else
			!pj_strcmp(&transport_state->factory->addr_name.host, &details->local_address)
			&& transport_state->factory->addr_name.port == details->local_port
#endif
		))) {
#ifdef GRANDSTREAM_NETWORKS
		res = CMP_MATCH;
#else
		return CMP_MATCH;
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
#if 0
	if (transport_state) {
		ast_log(LOG_DEBUG, "[%s] state[%p]: transp[%p], factory[%p], type[%d]", (res == CMP_MATCH) ? "Matched":"Not-Match",
				transport_state, transport_state->transport, transport_state->factory, transport_state->type);
		if (transport_state->factory) {
			ast_log(LOG_DEBUG, "factory host[%.*s]\n", (int)pj_strlen(&transport_state->factory->addr_name.host), pj_strbuf(&transport_state->factory->addr_name.host));
		}
	}
#endif
	return res;
#else
	return 0;
#endif
}

/*! \brief Helper function which returns the SIP URI of a Contact header */
static pjsip_sip_uri *nat_get_contact_sip_uri(pjsip_tx_data *tdata)
{
	pjsip_contact_hdr *contact = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CONTACT, NULL);

	if (!contact || (!PJSIP_URI_SCHEME_IS_SIP(contact->uri) && !PJSIP_URI_SCHEME_IS_SIPS(contact->uri))) {
		return NULL;
	}

	return pjsip_uri_get_uri(contact->uri);
}

/*! \brief Structure which contains hook details */
struct nat_hook_details {
	/*! \brief Outgoing message itself */
	pjsip_tx_data *tdata;
	/*! \brief Chosen transport */
	struct ast_sip_transport *transport;
};

/*! \brief Callback function for invoking hooks */
static int nat_invoke_hook(void *obj, void *arg, int flags)
{
	struct ast_sip_nat_hook *hook = obj;
	struct nat_hook_details *details = arg;

	if (hook->outgoing_external_message) {
		hook->outgoing_external_message(details->tdata, details->transport);
	}

	return 0;
}

static pj_status_t nat_on_tx_message(pjsip_tx_data *tdata)
{
	RAII_VAR(struct ao2_container *, transport_states, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_transport *, transport, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_transport_state *, transport_state, NULL, ao2_cleanup);
	struct request_transport_details details = { 0, };
	pjsip_via_hdr *via = NULL;
#ifndef BUILD_ON_CLOUD_PBX
	struct ast_sockaddr addr = { { 0, } };
#endif
	pjsip_sip_uri *uri = NULL;
	RAII_VAR(struct ao2_container *, hooks, NULL, ao2_cleanup);

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "nat_on_tx_message. tp_sel.type[%d], enter...", tdata->tp_sel.type);
#endif
	/* If a transport selector is in use we know the transport or factory, so explicitly find it */
	if (tdata->tp_sel.type == PJSIP_TPSELECTOR_TRANSPORT) {
		details.transport = tdata->tp_sel.u.transport;
	} else if (tdata->tp_sel.type == PJSIP_TPSELECTOR_LISTENER) {
		details.factory = tdata->tp_sel.u.listener;
	} else if (tdata->tp_info.transport->key.type == PJSIP_TRANSPORT_UDP || tdata->tp_info.transport->key.type == PJSIP_TRANSPORT_UDP6) {
		/* Connectionless uses the same transport for all requests */
		details.type = AST_TRANSPORT_UDP;
		details.transport = tdata->tp_info.transport;
	} else {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "transport type[%ld]", tdata->tp_info.transport->key.type);
#endif
		if (tdata->tp_info.transport->key.type == PJSIP_TRANSPORT_TCP) {
			details.type = AST_TRANSPORT_TCP;
		} else if (tdata->tp_info.transport->key.type == PJSIP_TRANSPORT_TLS) {
			details.type = AST_TRANSPORT_TLS;
		} else {
			/* Unknown transport type, we can't map and thus can't apply NAT changes */
			return PJ_SUCCESS;
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug 275136 */
		pjsip_cseq_hdr *cseq = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CSEQ, NULL);
		if (cseq && (tdata->msg->type == PJSIP_RESPONSE_MSG) && !pj_strcmp2(&cseq->method.name, "REGISTER")) {
			return PJ_SUCCESS;
		}
#endif
		if ((uri = nat_get_contact_sip_uri(tdata))) {
			details.local_address = uri->host;
			details.local_port = uri->port;
		} else if ((tdata->msg->type == PJSIP_REQUEST_MSG) &&
			(via = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_VIA, NULL))) {
			details.local_address = via->sent_by.host;
			details.local_port = via->sent_by.port;
		} else {
			return PJ_SUCCESS;
		}

		if (!details.local_port) {
			details.local_port = (details.type == AST_TRANSPORT_TLS) ? 5061 : 5060;
		}
	}

	if (!(transport_states = ast_sip_get_transport_states())) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "can't get transport_states.");
#endif
		return PJ_SUCCESS;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "details, local_addr[%.*s:%d], transp[%p], factory[%p], type[%d]", 
			(int)pj_strlen(&details.local_address), pj_strbuf(&details.local_address), details.local_port, details.transport, details.factory, details.type);
#endif

	if (!(transport_state = ao2_callback(transport_states, 0, find_transport_state_in_use, &details))) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "can't get transport_state.");
#endif
		return PJ_SUCCESS;
	}

	if (!(transport = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "transport", transport_state->id))) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "can't get transport.");
#endif
		return PJ_SUCCESS;
	}

#ifndef BUILD_ON_CLOUD_PBX
	if (transport_state->localnet) {
		ast_sockaddr_parse(&addr, tdata->tp_info.dst_name, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(&addr, tdata->tp_info.dst_port);

		/* See if where we are sending this request is local or not, and if not that we can get a Contact URI to modify */
		if (ast_sip_transport_is_local(transport_state, &addr)) {
			ast_debug(5, "Request is being sent to local address, skipping NAT manipulation\n");
			return PJ_SUCCESS;
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 102383, 2018.02.02 */
		RAII_VAR(struct ast_sockaddr *, local_addrs, NULL, ast_free);
		if (ast_sockaddr_resolve(&local_addrs, "127.0.0.1", PARSE_PORT_FORBID, AST_AF_UNSPEC) > 0) {
			if (!ast_sockaddr_cmp_addr(local_addrs, &addr)) {
				ast_log(LOG_DEBUG, "Dest address is 127.0.0.1, don't change, just return\n");
				return PJ_SUCCESS;
			}
		}
#endif
	}
#endif

#ifndef GRANDSTREAM_NETWORKS
	if (!ast_sockaddr_isnull(&transport_state->external_signaling_address)) {
		/* Update the contact header with the external address */
		if (uri || (uri = nat_get_contact_sip_uri(tdata))) {
			pj_strdup2(tdata->pool, &uri->host, ast_sockaddr_stringify_host(&transport_state->external_signaling_address));
			if (transport->external_signaling_port) {
				uri->port = transport->external_signaling_port;
				ast_debug(4, "Re-wrote Contact URI port to %d\n", uri->port);
			}
		}

		/* Update the via header if relevant */
		if ((tdata->msg->type == PJSIP_REQUEST_MSG) && (via || (via = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_VIA, NULL)))) {
			pj_strdup2(tdata->pool, &via->sent_by.host, ast_sockaddr_stringify_host(&transport_state->external_signaling_address));
			if (transport->external_signaling_port) {
				via->sent_by.port = transport->external_signaling_port;
			}
		}
	}
#else
	/* Added by yfyang, get external ip from STUN */
	char ext_addr[256] = {0};

#ifndef BUILD_ON_CLOUD_PBX
	if (transport->get_extaddr_from_stun) {
		ast_sip_nat_get_external_ip(ext_addr, sizeof(ext_addr)-1);
	}

	if (!strlen(ext_addr)) {
		ast_log(LOG_DEBUG, "external_signaling_address[%s]", ast_sockaddr_stringify_host(&transport_state->external_signaling_address));
		if (!ast_sockaddr_isnull(&transport_state->external_signaling_address)) {
			strncpy(ext_addr, ast_sockaddr_stringify_host(&transport_state->external_signaling_address), sizeof(ext_addr)-1);
		} 
	} else if (ast_strlen_zero(transport->external_signaling_address)) {
		ast_string_field_set(transport, external_signaling_address, ext_addr);
		ast_string_field_set(transport, external_media_address, ext_addr);
	}
#endif

#ifdef BUILD_ON_CLOUD_PBX

	ast_sip_get_nvram_value(":WsProxyHost", ext_addr, sizeof(ext_addr));
	if (!strlen(ext_addr)) {
		char *clouducm_public_ip = getenv("PUBLIC_HOST");
		if (!ast_strlen_zero(clouducm_public_ip)) {
			strncpy(ext_addr, clouducm_public_ip, sizeof(ext_addr)-1);
		}
	}

	ast_log(LOG_DEBUG, "Get HOST_IP:%s for cloud UCM\n", ext_addr);
#endif

	if (!strlen(ext_addr)) {
		ast_log(LOG_DEBUG, "No external addr.\n");
		return PJ_SUCCESS;
	}

	ast_log(LOG_DEBUG, "Get external addr[%s], re-write.\n", ext_addr);

	/* Update the contact header with the external address */
	if (uri || (uri = nat_get_contact_sip_uri(tdata))) {
		pj_strdup2(tdata->pool, &uri->host, ext_addr);
		if (transport->external_signaling_port) {
			uri->port = transport->external_signaling_port;
			ast_debug(4, "Re-wrote Contact URI port to %d\n", uri->port);
		}
	}

	/* Update the via header if relevant */
	if ((tdata->msg->type == PJSIP_REQUEST_MSG) && (via || (via = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_VIA, NULL)))) {
		pj_strdup2(tdata->pool, &via->sent_by.host, ext_addr);
		if (transport->external_signaling_port) {
			via->sent_by.port = transport->external_signaling_port;
		}
	}
#endif

	/* Invoke any additional hooks that may be registered */
	if ((hooks = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "nat_hook", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL))) {
		struct nat_hook_details hook_details = {
			.tdata = tdata,
			.transport = transport,
		};
		ao2_callback(hooks, 0, nat_invoke_hook, &hook_details);
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "nat_on_tx_message. exit...");
#endif
	return PJ_SUCCESS;
}

static pjsip_module nat_module = {
	.name = { "NAT", 3 },
	.id = -1,
	.priority = PJSIP_MOD_PRIORITY_TSX_LAYER - 2,
	.on_rx_request = nat_on_rx_message,
	.on_rx_response = nat_on_rx_message,
	.on_tx_request = nat_on_tx_message,
	.on_tx_response = nat_on_tx_message,
};

/*! \brief Function called when an INVITE goes out */
static int nat_incoming_invite_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	if (session->inv_session->state == PJSIP_INV_STATE_INCOMING) {
		pjsip_dlg_add_usage(session->inv_session->dlg, &nat_module, NULL);
	}

	return 0;
}

/*! \brief Function called when an INVITE response comes in */
static void nat_incoming_invite_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	handle_rx_message(session->endpoint, rdata);
}

/*! \brief Function called when an INVITE comes in */
static void nat_outgoing_invite_request(struct ast_sip_session *session, struct pjsip_tx_data *tdata)
{
	if (session->inv_session->state == PJSIP_INV_STATE_NULL) {
		pjsip_dlg_add_usage(session->inv_session->dlg, &nat_module, NULL);
	}
}

/*! \brief Supplement for adding NAT functionality to dialog */
static struct ast_sip_session_supplement nat_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_FIRST + 1,
	.incoming_request = nat_incoming_invite_request,
	.outgoing_request = nat_outgoing_invite_request,
	.incoming_response = nat_incoming_invite_response,
};


static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&nat_supplement);
	ast_sip_unregister_service(&nat_module);
	return 0;
}

static int load_module(void)
{
	if (ast_sip_register_service(&nat_module)) {
		ast_log(LOG_ERROR, "Could not register NAT module for incoming and outgoing requests\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_sip_session_register_supplement(&nat_supplement);

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP NAT Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip,res_pjsip_session",
);

