/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Kevin Harwell <kharwell@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/cli.h"
#include "asterisk/config.h"
#include "asterisk/manager.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/sorcery.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/taskprocessor.h"
#include "asterisk/astdb.h"
#include "asterisk/xml.h"
#include "asterisk/app.h"
#include "asterisk/pjsip_cei.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"
AST_MODULE_LOG("pjsip");

#endif

/*** DOCUMENTATION
	<manager name="PJSIPNotify" language="en_US">
		<synopsis>
			Send a NOTIFY to either an endpoint, an arbitrary URI, or inside a SIP dialog.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Endpoint" required="false">
				<para>The endpoint to which to send the NOTIFY.</para>
			</parameter>
			<parameter name="URI" required="false">
				<para>Abritrary URI to which to send the NOTIFY.</para>
			</parameter>
			<parameter name="channel" required="false">
				<para>Channel name to send the NOTIFY. Must be a PJSIP channel.</para>
			</parameter>
			<parameter name="Variable" required="true">
				<para>Appends variables as headers/content to the NOTIFY. If the variable is
				named <literal>Content</literal>, then the value will compose the body
				of the message if another variable sets <literal>Content-Type</literal>.
				<replaceable>name</replaceable>=<replaceable>value</replaceable></para>
			</parameter>
		</syntax>
		<description>
			<para>Sends a NOTIFY to an endpoint, an arbitrary URI, or inside a SIP dialog.</para>
			<para>All parameters for this event must be specified in the body of this
			request	via multiple <literal>Variable: name=value</literal> sequences.</para>
			<note><para>One (and only one) of <literal>Endpoint</literal>,
			<literal>URI</literal>, or <literal>Channel</literal> must be specified.
			If <literal>URI</literal> is used, the default outbound endpoint will be used
			to send the message. If the default outbound endpoint isn't configured, this command
			can not send to an arbitrary URI.</para></note>
		</description>
	</manager>
	<configInfo name="res_pjsip_notify" language="en_US">
		<synopsis>Module that supports sending NOTIFY requests to endpoints from external sources</synopsis>
		<configFile name="pjsip_notify.conf">
			<configObject name="general">
				<synopsis>Unused, but reserved.</synopsis>
			</configObject>
			<configObject name="notify">
				<synopsis>Configuration of a NOTIFY request.</synopsis>
				<description>
					<para>Each key-value pair in a <literal>notify</literal>
					configuration section defines either a SIP header to send
					in the request or a line of content in the request message
					body. A key of <literal>Content</literal> is treated
					as part of the message body and is appended in sequential
					order; any other header is treated as part of the SIP
					request.</para>
				</description>
				<configOption name="">
					<synopsis>A key/value pair to add to a NOTIFY request.</synopsis>
					<description>
						<para>If the key is <literal>Content</literal>,
						it will be treated as part of the message body. Otherwise,
						it will be added as a header in the NOTIFY request.</para>
						<para>The following headers are reserved and cannot be
						specified:</para>
						<enumlist>
							<enum name="Call-ID" />
							<enum name="Contact" />
							<enum name="CSeq" />
							<enum name="To" />
							<enum name="From" />
							<enum name="Record-Route" />
							<enum name="Route" />
							<enum name="Via" />
						</enumlist>
					</description>
				</configOption>
			</configObject>
		</configFile>
	</configInfo>
 ***/

#define CONTENT_TYPE_SIZE 64
#define CONTENT_SIZE 512

/*!
 * \internal
 * \brief The configuration file containing NOTIFY payload types to send.
 */
static const char notify_config[] = "pjsip_notify.conf";
#ifdef GRANDSTREAM_NETWORKS
/* add by kmye@grandstream.cn. it is for notify forwarding*/
static struct ast_taskprocessor *notify_serializer;
/*! \brief Message router for stasis messages regarding channel state */
static struct stasis_message_router *stasis_router = NULL;

struct msg_data {
	char* user_name;
	char* uri;
	pjsip_msg_body* body;
	pjsip_hdr* hdr;
	struct pjsip_rx_data *rdata;
};
#endif

struct notify_option_item {
	const char *name;
	const char *value;
	char buf[0];
};

struct notify_option {
	/*! Contains header and/or content information */
	struct ao2_container *items;
	/*! The name of the notify option */
	char name[0];
};

static int notify_option_hash(const void *obj, int flags)
{
	const struct notify_option *option = obj;
	return ast_str_case_hash(flags & OBJ_KEY ? obj : option->name);
}

static int notify_option_cmp(void *obj, void *arg, int flags)
{
	struct notify_option *option1 = obj;
	struct notify_option *option2 = arg;
	const char *key = flags & OBJ_KEY ? arg : option2->name;

	return strcasecmp(option1->name, key) ? 0 : CMP_MATCH;
}

static void notify_option_destroy(void *obj)
{
	struct notify_option *option = obj;
	ao2_cleanup(option->items);
}

static void *notify_option_alloc(const char *category)
{
	int category_size = strlen(category) + 1;

	struct notify_option *option = ao2_alloc(
		sizeof(*option) + category_size, notify_option_destroy);

	if (!option) {
		return NULL;
	}

	ast_copy_string(option->name, category, category_size);

	if (!(option->items = ao2_container_alloc_list(
		      AO2_ALLOC_OPT_LOCK_NOLOCK,
		      AO2_CONTAINER_ALLOC_OPT_DUPS_ALLOW, NULL, NULL))) {
		ao2_cleanup(option);
		return NULL;
	}

	return option;
}

static void *notify_option_find(struct ao2_container *container, const char *category)
{
	return ao2_find(container, category, OBJ_KEY);
}

static int notify_option_handler(const struct aco_option *opt,
				 struct ast_variable *var, void *obj)
{
	struct notify_option *option = obj;

	int name_size = strlen(var->name) + 1;
	int value_size = strlen(var->value) + 1;

	RAII_VAR(struct notify_option_item *, item,
		 ao2_alloc(sizeof(*item) + name_size + value_size,
			   NULL), ao2_cleanup);

	item->name = item->buf;
	item->value = item->buf + name_size;

	ast_copy_string(item->buf, var->name, name_size);
	ast_copy_string(item->buf + name_size, var->value, value_size);

	if (!ao2_link(option->items, item)) {
		return -1;
	}

	return 0;
}

struct notify_cfg {
	struct ao2_container *notify_options;
};

static void notify_cfg_destroy(void *obj)
{
	struct notify_cfg *cfg = obj;
	ao2_cleanup(cfg->notify_options);
}

static void *notify_cfg_alloc(void)
{
	struct notify_cfg *cfg;

	if (!(cfg = ao2_alloc(sizeof(*cfg), notify_cfg_destroy))) {
		return NULL;
	}

	cfg->notify_options = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
		20, notify_option_hash, NULL, notify_option_cmp);
	if (!cfg->notify_options) {
		ao2_cleanup(cfg);
		return NULL;
	}

	return cfg;
}

static struct aco_type notify_option = {
	.type = ACO_ITEM,
	.name = "notify",
	.category_match = ACO_BLACKLIST_EXACT,
	.category = "general",
	.item_offset = offsetof(struct notify_cfg, notify_options),
	.item_alloc = notify_option_alloc,
	.item_find = notify_option_find
};

static struct aco_type *notify_options[] = ACO_TYPES(&notify_option);

static struct aco_file module_conf = {
	.filename = notify_config,
	.types = ACO_TYPES(&notify_option),
};

static AO2_GLOBAL_OBJ_STATIC(globals);

CONFIG_INFO_STANDARD(notify_cfg, globals, notify_cfg_alloc,
	.files = ACO_FILES(&module_conf)
);

#ifdef GRANDSTREAM_NETWORKS
enum notify_info_type {
	NOTIFY_INFO_TYPE_FORWARD = 1,
	NOTIFY_INFO_TYPE_AMI,
	NOTIFY_INFO_TYPE_CLI,
};
#endif

/*!
 * \internal
 * \brief Structure to hold task data for notifications.
 */
struct notify_data {
	/*! The endpoint being notified */
	struct ast_sip_endpoint *endpoint;
	/*! The info of headers, types and content */
	void *info;
	/*! Function to help build notify request */
	void (*build_notify)(pjsip_tx_data *, void *);

#ifdef GRANDSTREAM_NETWORKS
	enum notify_info_type info_type;
#endif
};

/*!
 * \internal
 * \brief Destroy the notify CLI data releasing any resources.
 */
static void notify_cli_data_destroy(void *obj)
{
	struct notify_data *data = obj;

	ao2_cleanup(data->endpoint);
	ao2_cleanup(data->info);
}

/*!
 * \internal
 * \brief Structure to hold task data for notifications (URI variant)
 */
struct notify_uri_data {
	char *uri;
	void *info;
	void (*build_notify)(pjsip_tx_data *, void *);
};

/*!
 * \internal
 * \brief Structure to hold task data for notifications (channel variant)
 */
struct notify_channel_data {
	struct ast_sip_session *session;
	void *info;
	void (*build_notify)(pjsip_tx_data *, void *);
};

static void notify_cli_uri_data_destroy(void *obj)
{
	struct notify_uri_data *data = obj;

	ast_free(data->uri);
	ao2_cleanup(data->info);
}

/*!
 * \internal
 * \brief Destroy the notify CLI data releasing any resources (URI variant)
 */
static void build_cli_notify(pjsip_tx_data *tdata, void *info);

/*!
 * \internal
 * \brief Construct a notify data object for CLI.
 */
static struct notify_data* notify_cli_data_create(
	struct ast_sip_endpoint *endpoint, void *info)
{
	struct notify_data *data = ao2_alloc(sizeof(*data),
					     notify_cli_data_destroy);
	if (!data) {
		return NULL;
	}

	data->endpoint = endpoint;
	ao2_ref(data->endpoint, +1);

	data->info = info;
	ao2_ref(data->info, +1);

	data->build_notify = build_cli_notify;

#ifdef GRANDSTREAM_NETWORKS
	data->info_type = NOTIFY_INFO_TYPE_CLI;
#endif

	return data;
}

/*!
 * \internal
 * \brief Construct a notify URI data object for CLI.
 */
static struct notify_uri_data* notify_cli_uri_data_create(
	const char *uri, void *info)
{
	struct notify_uri_data *data = ao2_alloc(sizeof(*data),
		notify_cli_uri_data_destroy);

	if (!data) {
		return NULL;
	}

	data->uri = ast_strdup(uri);
	if (!data->uri) {
		ao2_ref(data, -1);
		return NULL;
	}

	data->info = info;
	ao2_ref(data->info, +1);

	data->build_notify = build_cli_notify;

	return data;
}

/*!
 * \internal
 * \brief Destroy the notify AMI data releasing any resources.
 */
static void notify_ami_data_destroy(void *obj)
{
	struct notify_data *data = obj;
	struct ast_variable *info = data->info;

	ao2_cleanup(data->endpoint);
	ast_variables_destroy(info);
}

/*!
 * \internal
 * \brief Destroy the notify AMI URI data releasing any resources.
 */
static void notify_ami_uri_data_destroy(void *obj)
{
	struct notify_uri_data *data = obj;
	struct ast_variable *info = data->info;

	ast_free(data->uri);
	ast_variables_destroy(info);
}

/*!
 * \internal
 * \brief Destroy the notify AMI channel data releasing any resources.
 */
static void notify_ami_channel_data_destroy(void *obj)
{
	struct notify_channel_data *data = obj;
	struct ast_variable *info = data->info;

	ao2_cleanup(data->session);
	ast_variables_destroy(info);
}

static void build_ami_notify(pjsip_tx_data *tdata, void *info);

/*!
 * \internal
 * \brief Construct a notify data object for AMI.
 */
static struct notify_data* notify_ami_data_create(
	struct ast_sip_endpoint *endpoint, void *info)
{
	struct notify_data *data = ao2_alloc(sizeof(*data),
					     notify_ami_data_destroy);
	if (!data) {
		return NULL;
	}

	data->endpoint = endpoint;
	ao2_ref(data->endpoint, +1);

	data->info = info;
	data->build_notify = build_ami_notify;

#ifdef GRANDSTREAM_NETWORKS
	data->info_type = NOTIFY_INFO_TYPE_AMI;
#endif

	return data;
}

/*!
 * \internal
 * \brief Construct a notify URI data object for AMI.
 */
static struct notify_uri_data* notify_ami_uri_data_create(
	const char *uri, void *info)
{
	struct notify_uri_data *data = ao2_alloc(sizeof(*data),
							notify_ami_uri_data_destroy);
	if (!data) {
		return NULL;
	}

	data->uri = ast_strdup(uri);
	if (!data->uri) {
		ao2_ref(data, -1);
		return NULL;
	}

	data->info = info;
	data->build_notify = build_ami_notify;

	return data;
}

/*!
 * \internal
 * \brief Construct a notify channel data object for AMI.
 */
static struct notify_channel_data *notify_ami_channel_data_create(
	struct ast_sip_session *session, void *info)
{
	struct notify_channel_data *data;

	data = ao2_alloc_options(sizeof(*data), notify_ami_channel_data_destroy,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!data) {
		return NULL;
	}

	data->session = session;
	data->info = info;
	data->build_notify = build_ami_notify;

	return data;
}

/*!
 * \internal
 * \brief Checks if the given header name is not allowed.
 *
 * \details Some headers are not allowed to be set by the user within the
 *          scope of a NOTIFY request.  If the given var header name is
 *          found in the "not allowed" list then return true.
 */
static int not_allowed(const char *name)
{
	int i;
	static const char *names[] = {
		"Call-ID",
		"Contact",
		"CSeq",
		"To",
		"From",
		"Record-Route",
		"Route",
		"Request-URI",
		"Via",
	};

	for (i = 0; i < ARRAY_LEN(names); ++i) {
		if (!strcasecmp(name, names[i])) {
			return 1;
		}
	}
	return 0;
}

/*!
 * \internal
 * \brief If a content type was specified add it and the content body to the
 *        NOTIFY request.
 */
static void build_notify_body(pjsip_tx_data *tdata, struct ast_str *content_type,
			      struct ast_str *content)
{
	if (content_type) {
		char *p;
		struct ast_sip_body body;

		if (content) {
			body.body_text = ast_str_buffer(content);
		}

		body.type = ast_str_buffer(content_type);
		if ((p = strchr(body.type, '/'))) {
			*p++ = '\0';
			body.subtype = p;
		}
		ast_sip_add_body(tdata, &body);
	}
}

/*!
 * \internal
 * \brief Build the NOTIFY request adding content or header info.
 */
static void build_notify(pjsip_tx_data *tdata, const char *name, const char *value,
			 struct ast_str **content_type, struct ast_str **content)
{
	if (not_allowed(name)) {
		ast_log(LOG_WARNING, "Cannot specify %s header, "
			"ignoring\n", name);
		return;
	}

	if (!strcasecmp(name, "Content-type")) {
		if (!(*content_type)) {
			*content_type = ast_str_create(CONTENT_TYPE_SIZE);
		}
		ast_str_set(content_type, 0,"%s", value);
	} else if (!strcasecmp(name, "Content")) {
		if (!(*content)) {
			*content = ast_str_create(CONTENT_SIZE);
		}

		if (ast_str_strlen(*content)) {
			ast_str_append(content, 0, "\r\n");
		}
		ast_str_append(content, 0, "%s", value);
	} else {
		ast_sip_add_header(tdata, name, value);
	}
}

/*!
 * \internal
 * \brief Build the NOTIFY request from CLI info adding header and content
 *        when specified.
 */
static void build_cli_notify(pjsip_tx_data *tdata, void *info)
{
	struct notify_option *option = info;
	RAII_VAR(struct ast_str *, content_type, NULL, ast_free);
	RAII_VAR(struct ast_str *, content, NULL, ast_free);

	struct notify_option_item *item;
	struct ao2_iterator i = ao2_iterator_init(option->items, 0);

	while ((item = ao2_iterator_next(&i))) {
		build_notify(tdata, item->name, item->value,
			     &content_type, &content);
		ao2_cleanup(item);
	}
	ao2_iterator_destroy(&i);

	build_notify_body(tdata, content_type, content);
}

/*!
 * \internal
 * \brief Build the NOTIFY request from AMI info adding header and content
 *        when specified.
 */
static void build_ami_notify(pjsip_tx_data *tdata, void *info)
{
	struct ast_variable *vars = info;
	RAII_VAR(struct ast_str *, content_type, NULL, ast_free);
	RAII_VAR(struct ast_str *, content, NULL, ast_free);
	struct ast_variable *i;

	for (i = vars; i; i = i->next) {
		if (!strcasecmp(i->name, "Content-Length")) {
			ast_log(LOG_NOTICE, "It is not necessary to specify Content-Length, ignoring.\n");
			continue;
		}
#ifdef GRANDSTREAM_NETWORKS
		if (!strcasecmp(i->name, "PJSIPNotify-ip")
			|| !strcasecmp(i->name, "PJSIPNotify-port")) {
			ast_log(LOG_NOTICE, "It is not necessary to specify %s, ignoring.\n", i->name);
			continue;
		}
#endif
		build_notify(tdata, i->name, i->value,
			     &content_type, &content);
	}

	build_notify_body(tdata, content_type, content);
}

/*!
 * \internal
 * \brief Build and send a NOTIFY request to a contact.
 */
static int notify_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct notify_data *data = arg;
	pjsip_tx_data *tdata;

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "In notify_contact(),data-info-type: %d", data->info_type);

	if (NOTIFY_INFO_TYPE_AMI == data->info_type) {
		const char *ip = NULL;
		const char *port = NULL;
		struct ast_variable *vars = data->info;

		if (vars)
		{
			ip = ast_variable_find_in_list(vars, "PJSIPNotify-ip");
			port = ast_variable_find_in_list(vars, "PJSIPNotify-port");

			if (!ast_strlen_zero(ip) && !ast_strlen_zero(port))
			{
				ast_log(LOG_DEBUG, "contact-uri: %s", contact->uri);
				ast_log(LOG_DEBUG, "PJSIPNotify: %s:%s", ip, port);
				ast_log(LOG_DEBUG, "ssh_proxy_flag: %d, original_address:%s\n", contact->ssh_proxy_flag, ast_strlen_zero(contact->original_address)?" ":contact->original_address);

#ifdef BUILD_ON_CLOUD_PBX
				if (!strstr(contact->uri, port)) //CloudUCM send ip and contact ip not match. only match port.
#else
				if (!(strstr(contact->uri, ip) && strstr(contact->uri, port)))
#endif
				{
					if (!(contact->ssh_proxy_flag && strstr(contact->original_address, ip))) {
						ast_log(LOG_WARNING, "contact-uri not match, skip sending notify");
						return 0;
					}
				}
			}
		}
	}
#endif

	if (ast_sip_create_request("NOTIFY", NULL, data->endpoint,
				   NULL, contact, &tdata)) {
		ast_log(LOG_WARNING, "SIP NOTIFY - Unable to create request for "
			"contact %s\n",	contact->uri);
		return -1;
	}

	ast_sip_add_header(tdata, "Subscription-State", "terminated");
#ifdef GRANDSTREAM_NETWORKS
	/* add the support attributes to peer ucm if has more use "," to split . add by wylu*/
	ast_sip_add_header(tdata, "X-GS-LdapAttrSupport", "Type,Position");
#endif
	data->build_notify(tdata, data->info);

	if (ast_sip_send_request(tdata, NULL, data->endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "SIP NOTIFY - Unable to send request for "
			"contact %s\n",	contact->uri);
		return -1;
	}

	return 0;
}

/*!
 * \internal
 * \brief Send a NOTIFY request to the endpoint.
 *
 * \detail Iterates over an endpoint's AORs sending a NOTIFY request
 *         with the appropriate payload information to each contact.
 */
static int notify_endpoint(void *obj)
{
	RAII_VAR(struct notify_data *, data, obj, ao2_cleanup);
	char *aor_name, *aors;

	if (ast_strlen_zero(data->endpoint->aors)) {
		ast_log(LOG_WARNING, "Unable to NOTIFY - "
			"endpoint has no configured AORs\n");
		return -1;
	}

	aors = ast_strdupa(data->endpoint->aors);

	while ((aor_name = ast_strip(strsep(&aors, ",")))) {
		RAII_VAR(struct ast_sip_aor *, aor,
			 ast_sip_location_retrieve_aor(aor_name), ao2_cleanup);
		RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);

		if (!aor || !(contacts = ast_sip_location_retrieve_aor_contacts(aor))) {
			continue;
		}

		ao2_callback(contacts, OBJ_NODATA, notify_contact, data);
	}

	return 0;
}

/*!
 * \internal
 * \brief Send a notify request to the URI.
 */
static int notify_uri(void *obj)
{
	RAII_VAR(struct notify_uri_data *, data, obj, ao2_cleanup);
	RAII_VAR(struct ast_sip_endpoint *, endpoint,
		ast_sip_default_outbound_endpoint(), ao2_cleanup);
	pjsip_tx_data *tdata;

	if (!endpoint) {
		ast_log(LOG_WARNING, "No default outbound endpoint set, can not send "
			"NOTIFY requests to arbitrary URIs.\n");
		return -1;
	}

	if (ast_strlen_zero(data->uri)) {
		ast_log(LOG_WARNING, "Unable to NOTIFY - URI is blank.\n");
		return -1;
	}

	if (ast_sip_create_request("NOTIFY", NULL, endpoint,
				   data->uri, NULL, &tdata)) {
		ast_log(LOG_WARNING, "SIP NOTIFY - Unable to create request for "
			"uri %s\n",	data->uri);
		return -1;
	}

	ast_sip_add_header(tdata, "Subscription-State", "terminated");

	data->build_notify(tdata, data->info);

	if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "SIP NOTIFY - Unable to send request for "
			"uri %s\n",	data->uri);
		return -1;
	}

	return 0;
}

/*!
 * \internal
 * \brief Send a notify request to a channel.
 */
static int notify_channel(void *obj)
{
	RAII_VAR(struct notify_channel_data *, data, obj, ao2_cleanup);
	pjsip_tx_data *tdata;
	struct pjsip_dialog *dlg;

	if (!data->session->channel
		|| !data->session->inv_session
		|| data->session->inv_session->state < PJSIP_INV_STATE_EARLY
		|| data->session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		return -1;
	}

	ast_debug(1, "Sending notify on channel %s\n", ast_channel_name(data->session->channel));

	dlg = data->session->inv_session->dlg;

	if (ast_sip_create_request("NOTIFY", dlg, NULL, NULL, NULL, &tdata)) {
		return -1;
	}

	ast_sip_add_header(tdata, "Subscription-State", "terminated");
	data->build_notify(tdata, data->info);

	if (ast_sip_send_request(tdata, dlg, NULL, NULL, NULL)) {
		return -1;
	}

	return 0;
}

enum notify_result {
	SUCCESS,
	INVALID_ENDPOINT,
	INVALID_CHANNEL,
	ALLOC_ERROR,
	TASK_PUSH_ERROR
};

typedef struct notify_data *(*task_data_create)(
	struct ast_sip_endpoint *, void *info);

typedef struct notify_uri_data *(*task_uri_data_create)(
	const char *uri, void *info);

typedef struct notify_channel_data *(*task_channel_data_create)(
	struct ast_sip_session *session, void *info);

/*!
 * \internal
 * \brief Send a NOTIFY request to the endpoint within a threaded task.
 */
static enum notify_result push_notify(const char *endpoint_name, void *info,
				      task_data_create data_create)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	struct notify_data *data;

	if (!(endpoint = ast_sorcery_retrieve_by_id(
		ast_sip_get_sorcery(), "endpoint", endpoint_name))) {
		return INVALID_ENDPOINT;
	}

	if (!(data = data_create(endpoint, info))) {
		return ALLOC_ERROR;
	}

	if (ast_sip_push_task(NULL, notify_endpoint, data)) {
		ao2_cleanup(data);
		return TASK_PUSH_ERROR;
	}

	return SUCCESS;
}

/*!
 * \internal
 * \brief Send a NOTIFY request to the URI within an threaded task.
 */
static enum notify_result push_notify_uri(const char *uri, void *info,
	task_uri_data_create data_create)
{
	struct notify_uri_data *data;

	if (!(data = data_create(uri, info))) {
		return ALLOC_ERROR;
	}

	if (ast_sip_push_task(NULL, notify_uri, data)) {
		ao2_cleanup(data);
		return TASK_PUSH_ERROR;
	}

	return SUCCESS;
}

/*!
 * \internal
 * \brief Send a NOTIFY request in a channel within an threaded task.
 */
static enum notify_result push_notify_channel(const char *channel_name, void *info,
	task_channel_data_create data_create)
{
	struct notify_channel_data *data;
	struct ast_channel *ch;
	struct ast_sip_session *session;
	struct ast_sip_channel_pvt *ch_pvt;

	/* note: this increases the refcount of the channel */
	ch = ast_channel_get_by_name(channel_name);
	if (!ch) {
		ast_debug(1, "No channel found with name %s", channel_name);
		return INVALID_CHANNEL;
	}

	if (strcmp(ast_channel_tech(ch)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "Channel was a non-PJSIP channel: %s\n", channel_name);
		ast_channel_unref(ch);
		return INVALID_CHANNEL;
	}

	ast_channel_lock(ch);
	ch_pvt = ast_channel_tech_pvt(ch);
	session = ch_pvt->session;

	if (!session || !session->inv_session
			|| session->inv_session->state < PJSIP_INV_STATE_EARLY
			|| session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_debug(1, "No active session for channel %s\n", channel_name);
		ast_channel_unlock(ch);
		ast_channel_unref(ch);
		return INVALID_CHANNEL;
	}

	ao2_ref(session, +1);
	ast_channel_unlock(ch);

	/* don't keep a reference to the channel, we've got a reference to the session */
	ast_channel_unref(ch);

	/*
	 * data_create will take ownership of the session,
	 * and take care of releasing the ref.
	 */
	data = data_create(session, info);
	if (!data) {
		ao2_ref(session, -1);
		return ALLOC_ERROR;
	}

	if (ast_sip_push_task(session->serializer, notify_channel, data)) {
		ao2_ref(data, -1);
		return TASK_PUSH_ERROR;
	}

	return SUCCESS;
}

/*!
 * \internal
 * \brief Do completion on the endpoint.
 */
static char *cli_complete_endpoint(const char *word)
{
	int wordlen = strlen(word);
	struct ao2_container * endpoints;
	struct ast_sip_endpoint *endpoint;
	struct ao2_iterator i;

	endpoints = ast_sorcery_retrieve_by_prefix(ast_sip_get_sorcery(),
		"endpoint", word, wordlen);
	if (endpoints == NULL) {
		return NULL;
	}

	i = ao2_iterator_init(endpoints, 0);
	while ((endpoint = ao2_iterator_next(&i))) {
		ast_cli_completion_add(
			ast_strdup(ast_sorcery_object_get_id(endpoint)));
		ao2_cleanup(endpoint);
	}
	ao2_iterator_destroy(&i);

	ao2_ref(endpoints, -1);

	return NULL;
}

/*!
 * \internal
 * \brief Do completion on the notify CLI command.
 */
static char *cli_complete_notify(const char *line, const char *word,
				 int pos, int state, int using_uri)
{
	char *c = NULL;

	if (pos == 3) {
		int which = 0;
		int wordlen = strlen(word);

		RAII_VAR(struct notify_cfg *, cfg,
			 ao2_global_obj_ref(globals), ao2_cleanup);
		struct notify_option *option;

		/* do completion for notify type */
		struct ao2_iterator i = ao2_iterator_init(cfg->notify_options, 0);
		while ((option = ao2_iterator_next(&i))) {
			if (!strncasecmp(word, option->name, wordlen) && ++which > state) {
				c = ast_strdup(option->name);
			}

			ao2_cleanup(option);
			if (c) {
				break;
			}
		}
		ao2_iterator_destroy(&i);
		return c;
	}

	if (pos == 4) {
		int wordlen = strlen(word);

		if (ast_strlen_zero(word)) {
		    if (state == 0) {
		        c = ast_strdup("endpoint");
		    } else if (state == 1) {
		        c = ast_strdup("uri");
		    }
		} else if (state == 0) {
		    if (!strncasecmp(word, "endpoint", wordlen)) {
		        c = ast_strdup("endpoint");
		    } else if (!strncasecmp(word, "uri", wordlen)) {
		        c = ast_strdup("uri");
		    }
		}

		return c;
	}

	return pos > 4 && !using_uri ? cli_complete_endpoint(word) : NULL;
}

/*!
 * \internal
 * \brief CLI command to send a SIP notify to an endpoint.
 *
 * \details Attempts to match the "type" given in the CLI command to a
 *          configured one.  If found, sends a NOTIFY to the endpoint
 *          with the associated payload.
 */
static char *cli_notify(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	RAII_VAR(struct notify_cfg *, cfg, NULL, ao2_cleanup);
	RAII_VAR(struct notify_option *, option, NULL, ao2_cleanup);

	int i;
	int using_uri = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip send notify";
		e->usage =
			"Usage: pjsip send notify <type> {endpoint|uri} <peer> [<peer>...]\n"
			"       Send a NOTIFY request to an endpoint\n"
			"       Message types are defined in pjsip_notify.conf\n";
		return NULL;
	case CLI_GENERATE:
		if (a->argc > 4 && (!strcasecmp(a->argv[4], "uri"))) {
			using_uri = 1;
		}

		return cli_complete_notify(a->line, a->word, a->pos, a->n, using_uri);
	}

	if (a->argc < 6) {
		return CLI_SHOWUSAGE;
	}

	if (!strcasecmp(a->argv[4], "uri")) {
		using_uri = 1;
	} else if (strcasecmp(a->argv[4], "endpoint")) {
		return CLI_SHOWUSAGE;
	}

	cfg = ao2_global_obj_ref(globals);

	if (!(option = notify_option_find(cfg->notify_options, a->argv[3])))
	{
		ast_cli(a->fd, "Unable to find notify type '%s'\n",
			a->argv[3]);
		return CLI_FAILURE;
	}

	for (i = 5; i < a->argc; ++i) {
		ast_cli(a->fd, "Sending NOTIFY of type '%s' to '%s'\n",
			a->argv[3], a->argv[i]);

		switch (using_uri ? push_notify_uri(a->argv[i], option, notify_cli_uri_data_create) :
		                    push_notify(a->argv[i], option, notify_cli_data_create)) {
		case INVALID_ENDPOINT:
			ast_cli(a->fd, "Unable to retrieve endpoint %s\n",
				a->argv[i]);
			break;
		case ALLOC_ERROR:
			ast_cli(a->fd, "Unable to allocate NOTIFY task data\n");
			return CLI_FAILURE;
		case TASK_PUSH_ERROR:
			ast_cli(a->fd, "Unable to push NOTIFY task\n");
			return CLI_FAILURE;
		default:
			break;
		}
	}

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_options[] = {
	AST_CLI_DEFINE(cli_notify, "Send a NOTIFY request to a SIP endpoint")
};

/*!
 * \interanl
 * \brief Completes SIPNotify AMI command in Endpoint mode.
 */
static void manager_notify_endpoint(struct mansession *s,
	const struct message *m, const char *endpoint_name)
{
	struct ast_variable *vars = astman_get_variables_order(m, ORDER_NATURAL);

	if (!strncasecmp(endpoint_name, "sip/", 4)) {
		endpoint_name += 4;
	}

	if (!strncasecmp(endpoint_name, "pjsip/", 6)) {
		endpoint_name += 6;
	}

	switch (push_notify(endpoint_name, vars, notify_ami_data_create)) {
	case INVALID_CHANNEL:
		/* Shouldn't be possible. */
		ast_assert(0);
		break;
	case INVALID_ENDPOINT:
		ast_variables_destroy(vars);
		astman_send_error_va(s, m, "Unable to retrieve endpoint %s",
			endpoint_name);
		break;
	case ALLOC_ERROR:
		ast_variables_destroy(vars);
		astman_send_error(s, m, "Unable to allocate NOTIFY task data");
		break;
	case TASK_PUSH_ERROR:
		/* Don't need to destroy vars since it is handled by cleanup in push_notify */
		astman_send_error(s, m, "Unable to push NOTIFY task");
		break;
	case SUCCESS:
		astman_send_ack(s, m, "NOTIFY sent");
		break;
	}
}

/*!
 * \internal
 * \brief Completes SIPNotify AMI command in URI mode.
 */
static void manager_notify_uri(struct mansession *s,
	const struct message *m, const char *uri)
{
	struct ast_variable *vars = astman_get_variables_order(m, ORDER_NATURAL);

	switch (push_notify_uri(uri, vars, notify_ami_uri_data_create)) {
	case INVALID_CHANNEL:
		/* Shouldn't be possible. */
		ast_assert(0);
		break;
	case INVALID_ENDPOINT:
		/* Shouldn't be possible. */
		ast_assert(0);
		break;
	case ALLOC_ERROR:
		ast_variables_destroy(vars);
		astman_send_error(s, m, "Unable to allocate NOTIFY task data");
		break;
	case TASK_PUSH_ERROR:
		/* Don't need to destroy vars since it is handled by cleanup in push_notify_uri */
		astman_send_error(s, m, "Unable to push Notify task");
		break;
	case SUCCESS:
		astman_send_ack(s, m, "NOTIFY sent");
		break;
	}
}

/*!
 * \internal
 * \brief Completes SIPNotify AMI command in channel mode.
 */
static void manager_notify_channel(struct mansession *s,
	const struct message *m, const char *channel)
{
	struct ast_variable *vars = astman_get_variables_order(m, ORDER_NATURAL);

	switch (push_notify_channel(channel, vars, notify_ami_channel_data_create)) {
	case INVALID_CHANNEL:
		ast_variables_destroy(vars);
		astman_send_error(s, m, "Channel not found");
		break;
	case INVALID_ENDPOINT:
		/* Shouldn't be possible. */
		ast_assert(0);
		break;
	case ALLOC_ERROR:
		ast_variables_destroy(vars);
		astman_send_error(s, m, "Unable to allocate NOTIFY task data");
		break;
	case TASK_PUSH_ERROR:
		/* Don't need to destroy vars since it is handled by cleanup in push_notify_channel */
		astman_send_error(s, m, "Unable to push Notify task");
		break;
	case SUCCESS:
		astman_send_ack(s, m, "NOTIFY sent");
		break;
	}
}

/*!
 * \internal
 * \brief AMI entry point to send a SIP notify to an endpoint.
 */
static int manager_notify(struct mansession *s, const struct message *m)
{
	const char *endpoint_name = astman_get_header(m, "Endpoint");
	const char *uri = astman_get_header(m, "URI");
	const char *channel = astman_get_header(m, "Channel");
	int count = 0;

	if (!ast_strlen_zero(endpoint_name)) {
		++count;
	}
	if (!ast_strlen_zero(uri)) {
		++count;
	}
	if (!ast_strlen_zero(channel)) {
		++count;
	}

	if (1 < count) {
		astman_send_error(s, m,
			"PJSIPNotify requires either an endpoint name, a SIP URI, or a channel.  "
			"You must use only one of them.");
	} else if (!ast_strlen_zero(endpoint_name)) {
		manager_notify_endpoint(s, m, endpoint_name);
	} else if (!ast_strlen_zero(uri)) {
		manager_notify_uri(s, m, uri);
	} else if (!ast_strlen_zero(channel)) {
		manager_notify_channel(s, m, channel);
	} else {
		astman_send_error(s, m,
			"PJSIPNotify requires either an endpoint name, a SIP URI, or a channel.");
	}

	return 0;
}


#ifdef GRANDSTREAM_NETWORKS

int ast_sip_get_nvram_value(char *pvalue, char* buf, int len )
{
#if 1
	if (ast_safe_nvram_get(pvalue, buf, len) == NULL) {
		memset(buf, 0, len);
	}
#else
	char cmd[32] = {0};

	if (pvalue == NULL) {
		return -1;
	}

	FILE *fp = NULL;
	memset(buf, 0, len);
	snprintf(cmd, sizeof(cmd), "nvram get %s", pvalue);

	fp = popen(cmd, "r");
	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
			   && (fgets(buf, len - 1, fp) != NULL)) {
			int length = strlen(buf);
			if (buf[length-1] == '\n') {
				buf[length-1] = '\0';
			}
			ast_log(LOG_DEBUG, "cmd Read: %s", buf);
		}
		pclose(fp);
		fp = NULL;
	}
#endif

	return 0;
}

/*add by kmye@grandstream.com. it's for notify forwarding */

/* free msg data */
static void msg_data_destroy(void *obj)
{
	struct msg_data *mdata = obj;
	if (!mdata)
		return;
	ast_free(mdata->user_name);
	ast_free(mdata->uri);
	if (mdata->rdata) {
		pjsip_rx_data_free_cloned(mdata->rdata);
	}
	ast_log(LOG_DEBUG, "-------msg_data_destroy %p---------", mdata->rdata);
	return;
}

static int send_response(struct pjsip_rx_data *rdata, int st_code)
{
	pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, st_code, NULL, NULL, NULL);
	return PJ_TRUE;
}

static struct ast_sip_endpoint* get_endpoint_by_exten(char* exten)
{
	struct ast_sip_endpoint* endpoint;
	if (NULL == exten) {
		return NULL;
	}
	if (!(endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", exten))) {
		ast_log(LOG_DEBUG, "can't find endpoint[%s]", exten);
		return NULL;
	}
	return endpoint;
}
static struct msg_data* create_notify_msg(struct ast_sip_endpoint* endpoint, struct pjsip_rx_data *rdata, struct pjsip_generic_string_hdr *event_hdr)
{
	RAII_VAR(struct ast_sip_aor *, found_aor, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_contact *, found_contact, NULL, ao2_cleanup);
	struct msg_data* mdata;

	if (!endpoint) {
		return NULL;
	}

	/* find extension's contact. eg. sip:1002@172.16.0.101:5060  */
	ast_sip_location_retrieve_contact_and_aor_from_list(endpoint->aors, &found_aor, &found_contact);
	if (!found_contact) {
		return NULL;
	}

	mdata = ao2_alloc(sizeof(*mdata), msg_data_destroy);
	if (pjsip_rx_data_clone(rdata, 0, &mdata->rdata) != PJ_SUCCESS) {
		return NULL;
	}

	mdata->uri = ast_strdup(found_contact->uri);
	mdata->user_name = ast_strdup(endpoint->aors);
	ast_log(LOG_DEBUG, "username: %s, uri: %s", endpoint->aors, found_contact->uri);

	if (event_hdr) {
		mdata->hdr = pjsip_hdr_clone(mdata->rdata->tp_info.pool, (pjsip_hdr *)event_hdr);
	}

	if (mdata->rdata->msg_info.msg->body) {
		mdata->body = pjsip_msg_body_clone(mdata->rdata->tp_info.pool, mdata->rdata->msg_info.msg->body);
	}

	return mdata;
}

static int getItem(const char *source, const char *item, char *buf, int buf_len)
{
	const char *ptr = NULL;
	const char *p = NULL;
	int i = 0;

	if ( source == NULL || item  == NULL || buf == NULL )
	{
		return -1;
	}
	memset(buf, 0, buf_len);

	ptr = strcasestr(source, item);
	if ( ptr == NULL )
	{
		return -1;
	}

	ptr += strlen(item);
	p = ptr;

	while ( *p != '\0' && *p != '\r' && i < (buf_len - 1) )
	{
		if (*p == '\'')
		{
			buf[i++] = *p;
			buf[i++] = *p;
		}
		else
		{
			buf[i++] = *p;
		}
		p++;
	}
	return 0;
}

static int send_notify(void* data)
{
	struct pjsip_tx_data *tdata = NULL;
	RAII_VAR(struct msg_data *, mdata, data, ao2_cleanup);
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	int  is_ipvt_notify = 0;

	if (!mdata) {
		return 0;
	}

	endpoint = get_endpoint_by_exten(mdata->user_name);
	if (!endpoint) {
		return 0;
	}

	ast_log(LOG_DEBUG, "send notify");
	if (ast_sip_create_request("NOTIFY", NULL, endpoint, mdata->uri, NULL, &tdata)) {
		ast_log(LOG_ERROR, "PJSIP MESSAGE - Could not create request\n");
		return -1;
	}

	if (mdata->hdr && mdata->hdr->vptr) {
		pjsip_generic_string_hdr *event_hdr = (pjsip_generic_string_hdr *)mdata->hdr;
		const char* ipvt_event = "X-GS-CONFERENCE";
		const char* ipvt_event1 = "X-GS-Meeting-Info";

		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)pjsip_hdr_clone(tdata->pool, mdata->hdr));

		if (!strncmp(pj_strbuf(&event_hdr->hvalue), ipvt_event, strlen(ipvt_event)) ||
			!strncmp(pj_strbuf(&event_hdr->hvalue), ipvt_event1, strlen(ipvt_event1))) {
			is_ipvt_notify = 1;
		}
	}

	if (mdata->body) {
		tdata->msg->body = pjsip_msg_body_clone(tdata->pool, mdata->body);
	}

	if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "send_notify - Could not send request\n");
		return -1;
	}

	if (is_ipvt_notify) {
		ast_log(LOG_NOTICE, "send_notify, is ipvt notify, done.\n");
		return 0;
	}

	/* send  AMI NotifyForwarding */
	manager_event(EVENT_FLAG_CALL, "NotifyForwarding",
			"Extension: %s\r\n"
			"%s\r\n\r\n",
			mdata->user_name?mdata->user_name:"",
			(char*)tdata->msg->body->data?(char*)tdata->msg->body->data:""
	);

	return 0;
}
static void saved_notify_message(struct pjsip_rx_data *rdata, const char* username)
{
	char voice_msg[256] = {0};
	char message_waiting[8] = {0};
	char* tmp = NULL;
	char family[128]={0};
	/* saved voice message into astdb */
	if (rdata->msg_info.msg->body && rdata->msg_info.msg->body->data) {
		getItem(rdata->msg_info.msg->body->data, "Voice-Message:", voice_msg, sizeof(voice_msg));
		getItem(rdata->msg_info.msg->body->data, "Messages-Waiting:", message_waiting, sizeof(message_waiting));
		snprintf(family, sizeof(family), "NotifyForwarding/PJSIP");
		if (strlen(voice_msg) > 0) {
			ast_db_put(family, username, voice_msg);
		}
		snprintf(family, sizeof(family), "NotifyForwardingStatus/PJSIP");
		if (!ast_strlen_zero(message_waiting)) {
			tmp = ast_strip(message_waiting);
			ast_db_put(family, username, tmp);
		}
	}
	return;
}
static int notify_forwarding(struct pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *event_hdr;
	struct msg_data *mdata;
	pj_str_t str_event = { "Event", 5 };
	const char* event = "message-summary";
	const char* ipvt_event  = "X-GS-CONFERENCE";
	const char* ipvt_event1 = "X-GS-Meeting-Info";
	pjsip_sip_uri *uri_to = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	char user_name[256] = {0};

	ast_log(LOG_DEBUG, "enter notify_forwarding...");
	if (0 != pj_stricmp2(&rdata->msg_info.cseq->method.name, "NOTIFY")) {
		return PJ_FALSE;
	}

	uri_to = pjsip_uri_get_uri(rdata->msg_info.to->uri);
	if (uri_to != NULL) {
		ast_copy_pj_str(user_name, &uri_to->user, sizeof(user_name));
	}

	endpoint = get_endpoint_by_exten(user_name);
	if (!endpoint) {
		return PJ_FALSE;
	}

	event_hdr= pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &str_event, NULL);
	if (!event_hdr) {
		ast_log(LOG_WARNING, "Incoming NOTIFY request without Event header\n");
		return send_response(rdata, PJSIP_SC_BAD_EVENT);
	}

	if (!strncmp(pj_strbuf(&event_hdr->hvalue), event, strlen(event))) {
		if (0 != strncmp(endpoint->remote_mailbox, "yes", 3)) {
			ast_log(LOG_NOTICE, "Incoming mwi NOTIFY, but endpoint[%s] hasn't remote mailbox\n", user_name);
			return PJ_FALSE;
		}

		saved_notify_message(rdata, user_name);

	} else if (strncmp(pj_strbuf(&event_hdr->hvalue), ipvt_event, strlen(ipvt_event)) &&
			   strncmp(pj_strbuf(&event_hdr->hvalue), ipvt_event1, strlen(ipvt_event1))) {
		ast_log(LOG_NOTICE, "Incoming NOTIFY event[%s], ignore\n", pj_strbuf(&event_hdr->hvalue));
		return PJ_FALSE;
	}

	if (!(mdata = create_notify_msg(endpoint, rdata, event_hdr)) ||
		ast_sip_push_task(notify_serializer, send_notify, mdata)) {
		ao2_ref(mdata, -1);
	}

	send_response(rdata, PJSIP_SC_OK);
	return PJ_TRUE;
}

/* add by kmye@grandstream.cn */
static pjsip_module notify_forward_module = {
	.name = {"Notify Forwarding Module", 16},
	.id = -1,
	.priority = PJSIP_MOD_PRIORITY_APPLICATION - 1,
	.on_rx_request = notify_forwarding,
};


#define _______IN_DIALOG_NOTIFY____________

#if 0
static const char * const cei_user_status_2str[][1] = {
	{ /* 0 CEI_STATUS_UNKNOWN   */     "unknown"         },
	{ /* 1 CEI_STATUS_CONNECTED */     "connected"       },
	{ /* 2 CEI_STATUS_DISCONNECTED */  "disconnected"    },
	{ /* 3 CEI_STATUS_ONHOLD */        "on-hold"         },
	{ /* 4 CEI_STATUS_MUTEVIAFOCUS */  "muted-via-focus" },
	{ /* 5 CEI_STATUS_PENDING */       "pending"         },
	{ /* 6 CEI_STATUS_ALERTING */      "alerting"        },
	{ /* 7 CEI_STATUS_DIALINGIN */     "dialing-in"      },
	{ /* 8 CEI_STATUS_DIALINGOUT */    "dialing-out"     },
	{ /* 9 CEI_STATUS_DISCONNECTING */ "disconnecting"   },
	{ /* 10 CEI_STATUS_PAUSEVIAFOCUS */"paused-via-focus"},
};
#endif

static void send_in_dialog_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata, int code)
{
	pjsip_tx_data *tdata = NULL;
	pjsip_dialog *dlg = NULL;

	if (!session || !session->inv_session || !session->inv_session->dlg || !rdata) {
		return;
	}

	dlg = session->inv_session->dlg;

	if (pjsip_dlg_create_response(dlg, rdata, code, NULL, &tdata) == PJ_SUCCESS) {
		struct pjsip_transaction *tsx = pjsip_rdata_get_tsx(rdata);
		pjsip_dlg_send_response(dlg, tsx, tdata);
	}

	return;
}

static int prase_gs_state_notify_xml(struct ast_sip_session *session, char *xml, struct ast_trunk_info *trunk_info,struct ast_answered_status *call_status)
{
	struct ast_xml_doc  *xml_doc          = NULL;
	struct ast_xml_node *root_node  		  = NULL;
	struct ast_xml_node *child_nodes      = NULL;
	struct ast_xml_node *node_iterator    = NULL;
	struct ast_xml_node *node_item        = NULL;
	struct ast_xml_node *node_trunk_info = NULL;
	struct ast_xml_node *node_call_status = NULL;
	const char *node_name;
	const char *content;
	const char *item_name;
	int ret = -1;

	if (!session || !xml || !trunk_info || !call_status) {
		ast_log(LOG_ERROR, "prase_gs_state_notify_xml param is NULL\n");
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		ast_log(LOG_ERROR, "[session:%p] ast_xml_read_memory error, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	root_node = ast_xml_get_root(xml_doc);
	if (!root_node) {
		ast_log(LOG_ERROR, "[session:%p] ast_xml_get_root error, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_xml_close(xml_doc);
		return -1;
	}

	child_nodes = ast_xml_node_get_children(root_node);
	if (child_nodes) {
		for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "item") && !node_item) {
				node_item = node_iterator;
				ast_log(LOG_DEBUG, "[session:%p] prase_gs_state_notify_xml get item node, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}
	}

	if (node_item) {
		item_name = ast_xml_get_attribute(node_item, "name");

		if (!strcasecmp(item_name, "call-state")) {
			for (node_iterator = ast_xml_node_get_children(node_item); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
				node_name = ast_xml_node_get_name(node_iterator);
				if (!strcasecmp(node_name, "trunkInfo")) {
					node_trunk_info = node_iterator;
				}
				/* Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
				else if(!strcasecmp(node_name, "callStatus")) {
					node_call_status = node_iterator;
				}
				/* End by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
				else {
					ast_log(LOG_DEBUG, "[session:%p] prase_gs_state_notify_xml node name:%s is not trunkInfo or callStatus, can't prase now endpoint:%s\n",
						session, node_name, ast_sorcery_object_get_id(session->endpoint));
				}
			}
		} else {
			ast_log(LOG_DEBUG, "[session:%p] prase_gs_state_notify_xml item name:%s is not call-state, can't prase now endpoint:%s\n",
				session, item_name, ast_sorcery_object_get_id(session->endpoint));
		}

		ast_xml_free_attr(item_name);
	}

	if (node_trunk_info) {
		for (node_iterator = ast_xml_node_get_children(node_trunk_info); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "confLink") ) {
				content = ast_xml_get_text(node_iterator);
				if (!content) {
					continue;
				}
				strncpy(trunk_info->conf_link, content, sizeof(trunk_info->conf_link) - 1);

				ast_log(LOG_DEBUG, "[session:%p] prase_gs_state_notify_xml get conf_link:%s, endpoint:%s\n",
					session, trunk_info->conf_link, ast_sorcery_object_get_id(session->endpoint));

				ast_xml_free_text(content);
				ret = 0;
			}

			if (!strcmp(node_name, "replaceChannelID")) {
				content = ast_xml_get_text(node_iterator);
				if (!content) {
					continue;
				}
				strncpy(trunk_info->channel_id, content, sizeof(trunk_info->channel_id) - 1);

				ast_log(LOG_DEBUG, "[session:%p] prase_gs_state_notify_xml get channel_id:%s, endpoint:%s\n",
					session, trunk_info->channel_id, ast_sorcery_object_get_id(session->endpoint));

				ast_xml_free_text(content);
				ret = 0;
			}
		}
	}

	/* Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
	if (node_call_status) {
		for (node_iterator = ast_xml_node_get_children(node_call_status); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "answered") ) {
				content = ast_xml_get_text(node_iterator);
				if (!content) {
					continue;
				}
				strncpy(call_status->answered, content, sizeof(call_status->answered) - 1);

				ast_log(LOG_DEBUG, "[session:%p] prase_gs_state_notify_xml get answered:%s, endpoint:%s\n",
					session,call_status->answered, ast_sorcery_object_get_id(session->endpoint));

				ast_xml_free_text(content);
				ret = 0;
			}

		}
	}
	/* End by zyxia for x-gs-state Bug 262507 , 2023.6.14 */

	ast_xml_close(xml_doc);

	return ret;
}

static int process_gs_state_notify(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	int res = 0;
	struct ast_trunk_info trunk_info;
	struct ast_answered_status answerd_status;
	pjsip_msg_body *body = NULL;
	char *content = NULL;

	if (!session || !session->channel || !rdata || !rdata->msg_info.msg) {
		ast_log(LOG_ERROR, "process_gs_state_notify param is NULL\n");
		return 0;
	}

	body = rdata->msg_info.msg->body;

	if (!body || !body->len) {
		ast_log(LOG_ERROR, "[session:%p] process_gs_state_notify msg has no body, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	char buf[body ? (body->len + 1) : 0];
	memset(buf, 0, sizeof(buf));
	body->print_body(body, buf, body->len);

	ast_channel_lock(session->channel);

	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ZOMBIE) || ast_check_hangup(session->channel)) {
		ast_channel_unlock(session->channel);
		ast_log(LOG_ERROR, "[session:%p] Channel has been hung up, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	content = (char*) ast_malloc(body->len + 1);
	if (!content) {
		ast_log(LOG_ERROR, "[session:%p] process_gs_state_notify malloc memory failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_unlock(session->channel);
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_gs_state_notify begin, chan:%s, endpoint:%s\n",
		session, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));

	memset(content, 0, body->len + 1);
	strncpy(content, (char*)body->data, body->len);

	memset(&trunk_info, 0, sizeof(struct ast_trunk_info));
	memset(&answerd_status, 0, sizeof(struct ast_answered_status));
	res = prase_gs_state_notify_xml(session, content, &trunk_info,&answerd_status);
	if (!res) {
		if(strlen(trunk_info.conf_link) || strlen(trunk_info.channel_id)) {
			if (ast_queue_control_data(session->channel, AST_CONTROL_NOTIFY_TRUNK_INFO, &trunk_info, sizeof(trunk_info))) {
				ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_NOTIFY_TRUNK_INFO frame failed for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}

			if (!session->conf_link && strlen(trunk_info.conf_link)) {
				session->conf_link = ast_calloc(1, strlen(trunk_info.conf_link) + 1);
				if (session->conf_link) {
					strcpy(session->conf_link, trunk_info.conf_link);
					ast_log(LOG_DEBUG, "[session:%p] process_gs_state_notify conf link:%s, chan:%s, endpoint:%s\n",
						session, session->conf_link, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
				}
			}
			if (!session->trunk_channel_id && strlen(trunk_info.channel_id)) {
				session->trunk_channel_id = ast_calloc(1, strlen(trunk_info.channel_id) + 1);
				if (session->trunk_channel_id) {
					strcpy(session->trunk_channel_id, trunk_info.channel_id);
					ast_log(LOG_DEBUG, "[session:%p] process_gs_state_notify channel_id:%s, chan:%s, endpoint:%s\n",
						session, session->trunk_channel_id, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
				}
			}
		}

		/* Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
		if(strlen(answerd_status.answered)) {
			if (ast_queue_control(session->channel, AST_CONTROL_REAL_ANSWERED)) {
				ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_REAL_ANSWERED frame failed for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}
		/* End by zyxia for x-gs-state Bug 262507 , 2023.6.14 */

	} else {
		ast_log(LOG_ERROR, "[session:%p] prase_gs_state_notify_xml error, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	if (content) {
		ast_free(content);
		content = NULL;
	}

	ast_channel_unlock(session->channel);

	return 1;
}

static int incoming_in_dialog_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	struct ast_frame f;
	char buf[1024] = {0};
	pjsip_msg_body *body = rdata->msg_info.msg->body;

	if (!body || !body->len) {
		/* need to return 200 OK on empty body */
		goto end_process;
	}

	ast_log(LOG_DEBUG, "[session:%p] Process incoming_in_dialog_request: NOTIFY, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, "conference-information")) {

		ast_log(LOG_DEBUG, "[session:%p] Process incoming_in_dialog_request: it's a mobile APP NOTIFY, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));

		snprintf(buf, sizeof(buf), "NOTIFY mobile-app-conf-info:%.*s", body->len, (char *)body->data);

		memset(&f, 0, sizeof(f));
		f.frametype = AST_FRAME_TEXT;
		f.subclass.integer = 0;
		f.offset = 0;
		f.data.ptr = buf;
		f.datalen = strlen(buf) + 1;
		ast_queue_frame(session->channel, &f);
	}

	if (rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, "x-gs-state+xml")) {

		ast_log(LOG_DEBUG, "[session:%p] Process incoming_in_dialog_request: it's a gs-state NOTIFY, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));

		process_gs_state_notify(session, rdata);
	}

end_process:
	send_in_dialog_response(session, rdata, 200);
	return 0;
}


static struct ast_sip_session_supplement notify_supplement = {
	.method = "NOTIFY",
	.incoming_request = incoming_in_dialog_request,
};


struct vmforward_body_info {
	struct notify_option *option;
	char local_ext[128];            /* The number of the local mailbox */
	char remote_ext[128];           /* the number of the remote mailbox */
	char addr[128];                 /* the address of trunk:192.168.124.77 */
	char folder[128];               /* the folder of voicemail, INBOX or Urgent*/
	int port;                       /* The tcp port to forward this voicemail */
};

static int is_private_network(const char *remote_addr)
{
	char buf1[4] = {0};
	char buf2[4] = {0};
	int ip1 = 0;
	int ip2 = 0;
	int res = -1;

	if ( remote_addr == NULL ) {
		return -1;
	}

	sscanf(remote_addr, "%3[^.].%3[^.]", buf1, buf2);

	ip1 = atoi(buf1);
	ip2 = atoi(buf2);

	if ( (ip1 == 10) || (ip1 == 192 && ip2 == 168) || (ip1 == 172 && ip2 >= 16 && ip2 <= 31 ) ) {
		res = 0;
	}
	ast_log(LOG_DEBUG, "is_private_network: host [%s] result [%d]", remote_addr, res);
	return res;
}

/* local_ip : "0" means to return the local ip, others means to return the NAT IP*/
static int getOutgoingIP(char *ipbuf, int local_ip)
{
	struct ifreq ifr_ip;
	struct sockaddr_in  *sin;
	int  fd;
	char buf[64] = { 0 };
	char buffer[128] = { 0 };
	char device[16] = { 0 };
	int  device_mode = 1;

	if (NULL == ipbuf) {
		return -1;
	}

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (-1 == fd) {
		return -1;
	}
	FILE *fp = NULL;
	fp = popen("nvram get :ext_ip && nvram get 231 && nvram get wan_device && nvram get lan_device && nvram get default_if && nvram get lan1_device && nvram get lan2_device", "r");
	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
			   && (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
			ast_log(LOG_DEBUG, "getOutgoingIP Read: %s", buf);
			strcat(buffer, buf);
		}
		pclose(fp);
		fp = NULL;

		if (buffer[0] != '\0') {
			/*parsing result*/
			int i = 0;
			char *ptr = buffer;
			char *sub = NULL;
			char *saveptr = NULL;
			char value[5][64] = {
				{ 0 }, { 0 }, { 0 }, {0}, {0}
			};
			ast_log(LOG_DEBUG, "getOutgoingIP Reslut: {%s}", buffer);

			while ((sub = strtok_r(ptr, "\n", &saveptr)) != NULL && i < 5) {
				ast_copy_string(value[i], sub, sizeof(value[i]) - 1);
				i++;
				ptr = NULL;
			}

			if ( local_ip != 0 && value[0] != NULL && (strchr(value[0], '.') != NULL || strchr(value[0], ':') != NULL)) {
				/*Try to support IPv4 and IPv6*/
				ast_safe_nvram_get("WsProxyHost", buf, sizeof(buf) - 1);
				if(strlen(buf) > 0)
				{
					strcpy(ipbuf, buf);
				}
				else{
					ast_log(LOG_DEBUG, "Have the external IP via STUN, use it : {%s}", value[0]);
					strcpy(ipbuf, value[0]);
				}
			}
			else {
				int i = 0;
				ast_log(LOG_DEBUG, "Use the local IP");

				if (NULL != strchr(value[0], '.')) {
					device_mode = atoi(value[1]);
					i = 1;
				}
				else {
					device_mode = atoi(value[0]);
				}

				if (device_mode == 0) {
					/*Router mode, using wan_device */
					/*0 ethx ethxx LANx*/
					ast_copy_string(device, value[1+i], sizeof(device) - 1);
				}
				else if (device_mode == 1) {
					/*Switch mode, using lan_device */
					/*1 ethx ethx LANX*/
					ast_copy_string(device, value[2+i], sizeof(device) - 1);
				}
				else {
					/*Dual mode, using the default interface*/
					/*2 LANX ethx etxx*/
					ast_copy_string(device, value[3+i], sizeof(device) - 1);
				}

				memset(&ifr_ip, 0, sizeof(ifr_ip));
				strncpy(ifr_ip.ifr_ifrn.ifrn_name, device[0] == '\0' ? "eth0" : device, sizeof(ifr_ip.ifr_ifrn.ifrn_name) - 1);

				if (ioctl(fd, SIOCGIFADDR, &ifr_ip) < 0) {
					ast_log(LOG_DEBUG, "getOutgoingIP: failed\n");
					close(fd);
					return -1;
				}

				sin = (struct sockaddr_in *)&ifr_ip.ifr_ifru.ifru_addr;
				strcpy(ipbuf, ast_inet_ntoa(sin->sin_addr));

			}
		}

	}
	ast_log(LOG_DEBUG, "getOutgoingIP: [%s]\n", ipbuf);
	close(fd);
	return 0;
}

/*!
 * \internal
 * \brief Build the NOTIFY request from CLI info adding header and content
 *        when specified.
 */
static void build_vmforward_notify(pjsip_tx_data *tdata, void *info)
{
	struct vmforward_body_info *data = info;
	struct notify_option *option = NULL;
	RAII_VAR(struct ast_str *, content_type, NULL, ast_free);
	RAII_VAR(struct ast_str *, content, NULL, ast_free);
	char buf[512] = {0};
	char address[256] = {0};
	char dev_base[32] = {0};

	if (!data) {
		return;
	}

	option = data->option;

	if (!option) {
		return;
	}

	struct notify_option_item *item;
	struct ao2_iterator i = ao2_iterator_init(option->items, 0);


	while ((item = ao2_iterator_next(&i))) {
		build_notify(tdata, item->name, item->value,
			     &content_type, &content);
		ao2_cleanup(item);
	}
	ao2_iterator_destroy(&i);

	ast_sip_get_nvram_value("dev_base", dev_base, sizeof(dev_base)); 
	if (strlen(dev_base) != 0 && strcmp(dev_base, "CloudUCM") == 0)
		ast_sip_get_nvram_value(":WsProxyHost", address, sizeof(address)); 
	else
		getOutgoingIP(address, is_private_network( data->addr ));

	snprintf(buf, sizeof(buf), "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
				   "<vmForward>\n"
				   "  <addr>%s</addr>\n"
				   "  <port>%d</port>\n"
				   "  <folder>%s</folder>\n"
				   "  <mailbox>%s</mailbox>\n"
				   "  <localmailbox>%s</localmailbox>\n"
				   "  <providerMode>security</providerMode>\n"
				   "  <devBase>%s</devBase>\n"
				   "</vmForward>", address, data->port, data->folder, data->remote_ext, data->local_ext, dev_base);
	build_notify(tdata, "Content", buf,
			     &content_type, &content);

	build_notify_body(tdata, content_type, content);
}

/*!
 * \internal
 * \brief Construct a notify data object for CLI.
 */
static struct notify_data* notify_vmforward_data_create(
	struct ast_sip_endpoint *endpoint, void *info)
{
	struct notify_data *data = ao2_alloc(sizeof(*data),
					     notify_cli_data_destroy);
	if (!data) {
		return NULL;
	}

	data->endpoint = endpoint;
	ao2_ref(data->endpoint, +1);

	data->info = info;
	ao2_ref(data->info, +1);

	data->build_notify = build_vmforward_notify;

#ifdef GRANDSTREAM_NETWORKS
	data->info_type = NOTIFY_INFO_TYPE_FORWARD;
#endif

	return data;
}

/*!
 * \brief Handler for when a channel is parked
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message The message about who got parked
 * */
static void handle_forward_notify_message(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	struct ast_vmforward_info *obj = stasis_message_data(message);

	RAII_VAR(struct notify_cfg *, cfg, NULL, ao2_cleanup);
	RAII_VAR(struct notify_option *, option, NULL, ao2_cleanup);
	RAII_VAR(struct vmforward_body_info *, info, NULL, ao2_cleanup);

	if (!obj) {
		return;
	}

	cfg = ao2_global_obj_ref(globals);

	if (!(option = notify_option_find(cfg->notify_options, "grandstream-vmforward")))
	{
		ast_log(LOG_VERBOSE, "Unable to find notify type 'grandstream-vmforward'\n");
		return;
	}

	info = ao2_alloc(sizeof(*info), NULL);
	if (!info) {
		return;
	}

	info->option = option;
	snprintf(info->remote_ext, sizeof(info->remote_ext), "%s", obj->remote_ext);
	snprintf(info->local_ext, sizeof(info->local_ext), "%s", obj->local_ext);
	snprintf(info->addr, sizeof(info->addr), "%s", obj->host);
	snprintf(info->folder, sizeof(info->folder), "%s", obj->folder);
	info->port = obj->port;

	push_notify(obj->trunk, info, notify_vmforward_data_create);

	return;
}
#endif


static int load_module(void)
{
	if (aco_info_init(&notify_cfg)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	aco_option_register_custom(&notify_cfg, "", ACO_PREFIX, notify_options,
				   "", notify_option_handler, 0);
#ifdef GRANDSTREAM_NETWORKS
	if (ast_sip_register_service(&notify_forward_module) != PJ_SUCCESS) {
		return AST_MODULE_LOAD_DECLINE;
	}
	char name[AST_UUID_STR_LEN];
	ast_uuid_generate_str(name, sizeof(name));
	notify_serializer = ast_sip_create_serializer(name);
	if (!notify_serializer) {
		ast_sip_unregister_service(&notify_forward_module);
		return AST_MODULE_LOAD_DECLINE;
	}
#endif

	if (aco_process_config(&notify_cfg, 0)) {
		aco_info_destroy(&notify_cfg);
#ifdef GRANDSTREAM_NETWORKS
		ast_sip_unregister_service(&notify_forward_module);
#endif
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_cli_register_multiple(cli_options, ARRAY_LEN(cli_options));
	ast_manager_register_xml("PJSIPNotify", EVENT_FLAG_SYSTEM, manager_notify);
#ifdef GRANDSTREAM_NETWORKS
	ast_sip_session_register_supplement(&notify_supplement);

	struct stasis_topic *notify_topic = ast_forward_notify_topic();
	if (!notify_topic) {
		return -1;
	}

	stasis_router = stasis_message_router_create(notify_topic);

	if (!stasis_router) {
		return -1;
	}

	stasis_message_router_add(stasis_router, ast_forward_notify_type(), handle_forward_notify_message, NULL);
#endif

	return AST_MODULE_LOAD_SUCCESS;
}

static int reload_module(void)
{
	if (aco_process_config(&notify_cfg, 1) == ACO_PROCESS_ERROR) {
		return AST_MODULE_LOAD_DECLINE;
	}

	return 0;
}

static int unload_module(void)
{
	ast_manager_unregister("PJSIPNotify");
	ast_cli_unregister_multiple(cli_options, ARRAY_LEN(cli_options));
#ifdef GRANDSTREAM_NETWORKS
	ast_sip_unregister_service(&notify_forward_module);
	ast_taskprocessor_unreference(notify_serializer);
	ast_sip_session_unregister_supplement(&notify_supplement);
	stasis_message_router_unsubscribe_and_join(stasis_router);
	stasis_router = NULL;
#endif

	aco_info_destroy(&notify_cfg);
	ao2_global_obj_release(globals);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "CLI/AMI PJSIP NOTIFY Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.reload = reload_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip",
);
