/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/logger.h"
#include "asterisk/module.h"
#include "asterisk/xml.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"

AST_MODULE_LOG("general_info");

static int general_info_incoming_request(struct ast_sip_session* session, struct pjsip_rx_data* rdata);

static struct ast_sip_session_supplement general_info_supplement = {
	.method = "INFO",
	.incoming_request = general_info_incoming_request,
};

static int process_general_sip_info(char* xml, char* buf, int size) {
	struct ast_xml_doc* xml_doc = NULL;
	struct ast_xml_node* info_node = NULL;
	struct ast_xml_node* child_nodes = NULL;
	struct ast_xml_node* node_iterator = NULL;
	struct ast_xml_node* node_channel = NULL;  
	const char* node_name = NULL;
	const char* content = NULL;

	if (!xml || !buf || size <= 0) {
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		return -1;
	}

	info_node = ast_xml_get_root(xml_doc);
	if (!info_node) {
		ast_xml_close(xml_doc);
		return -1;
	}

	child_nodes = ast_xml_node_get_children(info_node);

	if (child_nodes) {
		for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "channelID") && !node_channel) {
				node_channel = node_iterator;
			}
		}
	}

	if (node_channel) {
		content = ast_xml_get_text(node_channel);
		if (content) {
			ast_log(LOG_NOTICE, "[process_general_sip_info] The peer channel name: %s.\n", content);
			memset(buf, 0, size);
			strncpy(buf, content, size);
			ast_xml_free_text(content);
		}
	} else {
		ast_log(LOG_DEBUG, "[process_general_sip_info] The xml is invalid.\n");
	}

	ast_xml_close(xml_doc);

	return 0;
}

static int general_info_incoming_request(struct ast_sip_session* session, struct pjsip_rx_data* rdata)
{
	ast_log(LOG_DEBUG, "[general_info_incoming_request] Incoming");

	char channame[128] = {0};
	char* content = NULL;
	pjsip_msg_body* body = NULL;
	struct pjsip_tx_data *tdata = NULL;
	struct pjsip_transaction* tsx = NULL;
	RAII_VAR(struct ast_channel*, peer_chan, NULL, ast_channel_cleanup);

	if (!session || !rdata || !rdata->msg_info.msg) {
		ast_log(LOG_ERROR, "[general_info_incoming_request] The session or rdata is nil.");
		return 0;
	}

	if (!session->channel
		|| !ast_sip_is_content_type(&rdata->msg_info.msg->body->content_type,
			"application",
			"general_sip_control+xml")) {
		return 0;
	}

	body = rdata->msg_info.msg->body;
	if (!body || !body->len) {
		return 0;
	}

	content = (char*)ast_malloc(body->len + 1);
	if (!content) {
		return 0;
	}

	memset(content, 0, body->len + 1);
	strncpy(content, (char*)body->data, body->len);

	tsx = pjsip_rdata_get_tsx(rdata);

	if (process_general_sip_info(content, channame, sizeof(channame))) {
		ast_free(content);
		return 0;
	}

	if (pjsip_dlg_create_response(session->inv_session->dlg, rdata, 200, NULL, &tdata) == PJ_SUCCESS) {
		pjsip_dlg_send_response(session->inv_session->dlg, tsx, tdata);
	}

	peer_chan = ast_channel_get_by_name(channame);
	if (peer_chan) {
		ast_indicate_data2(peer_chan, AST_CONTROL_GERERAL_SIP_INFO, content, body->len + 1);
	}

	ast_free(content);

	ast_log(LOG_DEBUG, "[general_info_incoming_request] Outgoing");

	return 1;
}

static int load_module(void)
{
	ast_sip_session_register_supplement(&general_info_supplement);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&general_info_supplement);

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP General INFO Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DRIVER,
	.requires = "res_pjsip,res_pjsip_session",
);
