#ifdef GRANDSTREAM_NETWORKS
/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2017, GrandStream, Inc.
 *
 * Mark pxie <pxie@grandstream.cn>
 *
 * UCM docking with the ip phone for agent login or logoff
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

//ASTERISK_FILE_VERSION(__FILE__, "$Revision: 13.0 $")

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>

#include "asterisk/module.h"
#include "asterisk/linkedlists.h"
#include "asterisk/astobj2.h"
#include "asterisk/datastore.h"
#include "asterisk/uuid.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/sched.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/res_pjsip_body_generator_types.h"
#include "asterisk/callerid.h"
#include "asterisk/manager.h"
#include "asterisk/app.h"
#include "asterisk/test.h"
#include "res_pjsip/include/res_pjsip_private.h"
#include "asterisk/res_pjsip_presence_xml.h"
#include "asterisk/lock.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/cli.h"
#include "asterisk/sched.h"

AST_MODULE_LOG("pjsip")

struct stasis_topic *queue_agent_topic = NULL;
struct stasis_message_router *queue_agent_router = NULL;

static struct ast_sched_context *queue_agent_sched;
unsigned int g_queue_agent_sched_id = 0;
#define QUEUE_AGENT_REFRESH_TIMEOUT (5 * 60 * 1000)

struct queue_agent_subscription {
	char *aors;
	char id[1];
};

AO2_GLOBAL_OBJ_STATIC(queue_agent_unsolicited);

static int queue_agent_send_notify_contact(void *userdata);

#define QUEUE_AGENT_TYPE "application"
#define QUEUE_AGENT_SUBTYPE "ucm-cc+xml"

struct ast_pjsip_queue_agent {
	char         queue_extension[40];
	char         member_extension[40];
	char         queue_agent[40];
	char         need_login[40];
	char         cur_status[40];
	char         queue_name[40];
	unsigned int refresh_flag;

	AST_LIST_ENTRY(ast_pjsip_queue_agent) entry;
};

static AST_LIST_HEAD_STATIC(g_queue_agent, ast_pjsip_queue_agent);


#define CALLQUEUEAGENT	"\
  <CallQueueAgent>\n\
    <QueueExtension>%s</QueueExtension>\n\
    <Agent>%s</Agent>\n\
    <NeedLogin>%s</NeedLogin>\n\
    <CurStatus>%s</CurStatus>\n\
    <QueueName>%s</QueueName>\n\
  </CallQueueAgent>\n"

static struct ast_str *create_queue_agent_body(const char *agent_exten, unsigned int *find_endpt_flag)
{
	struct ast_str *body_text = NULL;

	body_text = ast_str_create(1000);
	if (!body_text) {
		ast_log(LOG_ERROR, "Could not allocate buffer for queue agent notify.\n");
		return NULL;
	}

	ast_str_append(&body_text, 0,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	ast_str_append(&body_text, 0, "<UCMCallCenterStatus>\n");

	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		ast_log(LOG_NOTICE, "queue extension: %s, member: %s, agent: %s, need login: %s, status: %s, queue name: %s\n",
			cur->queue_extension, cur->member_extension, cur->queue_agent, cur->need_login, cur->cur_status, cur->queue_name);

		if (agent_exten && !strcasecmp(cur->member_extension, agent_exten)) {
			ast_str_append(&body_text, 0, CALLQUEUEAGENT, cur->queue_extension, cur->queue_agent,
				cur->need_login, cur->cur_status, cur->queue_name);
			cur->refresh_flag = 1;
			*find_endpt_flag = 1;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);

	ast_str_append(&body_text, 0, "</UCMCallCenterStatus>\n");

	return body_text;
}

struct queue_agent_unsolicited_data {
	struct queue_agent_subscription *sub;
	struct ast_sip_endpoint *endpoint;
	pjsip_evsub_state state;
	const struct ast_sip_body *body;
};

static int queue_agent_send_unsolicited_notify_to_contact(void *obj, void *arg, int flags)
{
	struct queue_agent_unsolicited_data *agent_data = arg;
	struct queue_agent_subscription *sub = agent_data->sub;
	struct ast_sip_endpoint *endpoint = agent_data->endpoint;
	pjsip_evsub_state state = agent_data->state;
	const struct ast_sip_body *body = agent_data->body;
	struct ast_sip_contact *contact = obj;
	const char *state_name;
	pjsip_tx_data *tdata;
	pjsip_sub_state_hdr *sub_state;
	pjsip_event_hdr *event;
	const pjsip_hdr *allow_events = pjsip_evsub_get_allow_events_hdr(NULL);

	if (ast_sip_create_request("NOTIFY", NULL, endpoint, NULL, contact, &tdata)) {
		ast_log(LOG_WARNING, "Unable to create unsolicited NOTIFY request to endpoint %s URI %s\n",
			sub ? sub->id : "null", contact->uri);
		return 0;
	}

	if (!ast_strlen_zero(endpoint->subscription.mwi.fromuser)) {
		pjsip_fromto_hdr *from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, NULL);
		pjsip_name_addr *from_name_addr = (pjsip_name_addr *) from->uri;
		pjsip_sip_uri *from_uri = pjsip_uri_get_uri(from_name_addr->uri);

		pj_strdup2(tdata->pool, &from_uri->user, endpoint->subscription.mwi.fromuser);
	}

	switch (state) {
	case PJSIP_EVSUB_STATE_ACTIVE:
		state_name = "active";
		break;
	case PJSIP_EVSUB_STATE_TERMINATED:
	default:
		state_name = "terminated";
		break;
	}

	sub_state = pjsip_sub_state_hdr_create(tdata->pool);
	pj_cstr(&sub_state->sub_state, state_name);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *) sub_state);

	event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&event->event_type, "UCM-CC");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *) event);

	pjsip_msg_add_hdr(tdata->msg, pjsip_hdr_shallow_clone(tdata->pool, allow_events));
	ast_sip_add_body(tdata, body);
	ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL);

	return 0;
}

static int queue_agent_send_notify_contact(void *userdata)
{
	const char *endp_aor = userdata;

	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(),
		"endpoint", endp_aor), ao2_cleanup);
	char *endpoint_aors;
	char *aor_name;
	struct ast_sip_body body;
	struct ast_str *body_text = NULL;
	unsigned int find_endpt_flag = 0;

	if (!endpoint) {
		ast_log(LOG_WARNING, "Unable to send unsolicited NOTIFY to %s because endpoint does not exist\n", endp_aor);
		return -1;
	}

	if (ast_strlen_zero(endpoint->aors)) {
		ast_log(LOG_WARNING, "Unable to send unsolicited NOTIFY to %s because the endpoint has no"
			" configured AORs\n", endp_aor);
		return -1;
	}

	body.type = QUEUE_AGENT_TYPE;
	body.subtype = QUEUE_AGENT_SUBTYPE;

	if (!(body_text = create_queue_agent_body(endp_aor, &find_endpt_flag))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for queue agent notify.\n");
		return -1;
	}

	/*! Begin: Added by pxie@grandstream.cn for bug[87803], 2017-06-21. */
	if (0 == find_endpt_flag) {
		ast_log(LOG_VERBOSE, "Endpoint(%s) was not queue agents.\n", endp_aor);
		ast_free(body_text);
		return -1;
	}
	/*! Ended: Added by pxie@grandstream.cn for bug[87803], 2017-06-21. */

	body.body_text = ast_str_buffer(body_text);
	endpoint_aors = ast_strdupa(endpoint->aors);
	ast_debug(5, "Sending unsolicited NOTIFY to endpoint %s, endpoint_aors: %s\n",
		endp_aor, endpoint_aors);

	while ((aor_name = strsep(&endpoint_aors, ","))) {
		RAII_VAR(struct ast_sip_aor *, aor, ast_sip_location_retrieve_aor(aor_name), ao2_cleanup);
		RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
		struct queue_agent_unsolicited_data queue_agent_data = {
			.sub = NULL,
			.endpoint = endpoint,
			.body = &body,
		};

		if (!aor) {
			ast_log(LOG_WARNING, "Unable to locate AOR %s for unsolicited NOTIFY\n", aor_name);
			continue;
		}

		contacts = ast_sip_location_retrieve_aor_contacts(aor);
		if (!contacts || (ao2_container_count(contacts) == 0)) {
			ast_log(LOG_NOTICE, "No contacts bound to AOR %s. Cannot send unsolicited NOTIFY until a contact registers.\n", aor_name);
			continue;
		}

		ao2_callback(contacts, OBJ_NODATA, queue_agent_send_unsolicited_notify_to_contact, &queue_agent_data);
	}
	ast_free(body_text);

	return 0;
}

static void queue_agent_event_insert_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	struct ast_json			*json_object	= NULL;
	struct ast_json_payload *payload		= NULL;
	const char              *queue_exten    = NULL;
	const char              *member_exten   = NULL;
	const char              *queue_agent    = NULL;
	const char              *need_login     = NULL;
	const char              *cur_status     = NULL;
	const char              *queue_name     = NULL;

	if (!msg) {
		ast_log(LOG_WARNING, "Stasis message is NULL!\n");
		return;
	}

	payload = stasis_message_data(msg);
	if (!payload) {
		ast_debug(LOG_WARNING, "Got NULL payload from stasis message!\n");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL!\n");
		return;
	}

	queue_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueExtension"));
	member_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueMember"));
	queue_agent = ast_json_string_get(ast_json_object_get(json_object, "QueueAgent"));
	need_login = ast_json_string_get(ast_json_object_get(json_object, "NeedLogin"));
	cur_status = ast_json_string_get(ast_json_object_get(json_object, "CurStatus"));
	queue_name = ast_json_string_get(ast_json_object_get(json_object, "QueueName"));

	if (!queue_exten || !member_exten || !queue_agent || !need_login || !cur_status || !queue_name) {
		ast_log(LOG_WARNING, "Incomplete JSON object data\n");
		return;
	}

	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *qagent = ast_calloc(1, sizeof(*qagent));
	if (qagent) {
		ast_copy_string(qagent->queue_extension, queue_exten, sizeof(qagent->queue_extension));
		char *tmp = strstr(member_exten, "/");
		ast_copy_string(qagent->member_extension, ++tmp, sizeof(qagent->member_extension));
		ast_copy_string(qagent->queue_agent, queue_agent, sizeof(qagent->queue_agent));
		ast_copy_string(qagent->need_login, need_login, sizeof(qagent->need_login));
		ast_copy_string(qagent->cur_status, cur_status, sizeof(qagent->cur_status));
		ast_copy_string(qagent->queue_name, queue_name, sizeof(qagent->queue_name));
		AST_LIST_INSERT_TAIL(&g_queue_agent, qagent, entry);
		ast_log(LOG_DEBUG, "Queue agent receive 'insert' event!\n");
		queue_agent_send_notify_contact(tmp);
	}
	AST_LIST_UNLOCK(&g_queue_agent);
}

static void queue_agent_event_update_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	struct ast_json			*json_object	= NULL;
	struct ast_json_payload *payload		= NULL;
	const char              *queue_exten    = NULL;
	const char              *member_exten   = NULL;
	const char              *queue_agent    = NULL;
	const char              *need_login     = NULL;
	const char              *cur_status     = NULL;
	const char              *queue_name     = NULL;
	unsigned int            update_flag     = 0;

	if (!msg) {
		ast_log(LOG_WARNING, "Stasis message is NULL!\n");
		return;
	}

	payload = stasis_message_data(msg);
	if (!payload) {
		ast_debug(LOG_WARNING, "Got NULL payload from stasis message!\n");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL!\n");
		return;
	}

	queue_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueExtension"));
	member_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueMember"));
	queue_agent = ast_json_string_get(ast_json_object_get(json_object, "QueueAgent"));
	need_login = ast_json_string_get(ast_json_object_get(json_object, "NeedLogin"));
	cur_status = ast_json_string_get(ast_json_object_get(json_object, "CurStatus"));
	queue_name = ast_json_string_get(ast_json_object_get(json_object, "QueueName"));
	
	ast_log(LOG_NOTICE, "queue_exten: %s, member_exten: %s, queue_agent: %s, need_login: %s, cur_status: %s, queue_name: %s\n",
		queue_exten, member_exten, queue_agent, need_login, cur_status, queue_name);
	if (!queue_exten || !member_exten || !queue_agent || !need_login || !cur_status || !queue_name) {
		ast_log(LOG_WARNING, "Incomplete JSON object data\n");
		return;
	}

	char *member = strstr(member_exten, "/");
	if (member) {
		member++;
	}

	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		if (!strcmp(cur->queue_extension, queue_exten) && !strcmp(cur->member_extension, member)) {
			update_flag = 1;
			ast_copy_string(cur->queue_agent, queue_agent, sizeof(cur->queue_agent));
			ast_copy_string(cur->need_login, need_login, sizeof(cur->need_login));
			ast_copy_string(cur->cur_status, cur_status, sizeof(cur->cur_status));
			ast_copy_string(cur->queue_name, queue_name, sizeof(cur->queue_name));
			ast_log(LOG_NOTICE, "queue extension: %s, member: %s, agent: %s, need login: %s, status: %s, queue name:%s\n",
				cur->queue_extension, cur->member_extension, cur->queue_agent, cur->need_login, cur->cur_status, cur->queue_name);
			ast_log(LOG_DEBUG, "Queue agent receive 'update' event!\n");
			queue_agent_send_notify_contact(member);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	/*! Begin: Added by pxie@grandstream.cn for bug[87586], 2017-06-21. */
	if (0 == update_flag) {
		struct ast_pjsip_queue_agent *qagent = ast_calloc(1, sizeof(*qagent));
		if (qagent) {
			ast_copy_string(qagent->queue_extension, queue_exten, sizeof(qagent->queue_extension));
			ast_copy_string(qagent->member_extension, member, sizeof(qagent->member_extension));
			ast_copy_string(qagent->queue_agent, queue_agent, sizeof(qagent->queue_agent));
			ast_copy_string(qagent->need_login, need_login, sizeof(qagent->need_login));
			ast_copy_string(qagent->cur_status, cur_status, sizeof(qagent->cur_status));
			ast_copy_string(qagent->queue_name, queue_name, sizeof(qagent->queue_name));
			AST_LIST_INSERT_TAIL(&g_queue_agent, qagent, entry);
			ast_log(LOG_DEBUG, "Queue agent receive 'update' event!\n");
			queue_agent_send_notify_contact(member);
		}
	}
	/*! Ended: Added by pxie@grandstream.cn for bug[87586], 2017-06-21. */
	AST_LIST_UNLOCK(&g_queue_agent);
}

static void queue_agent_event_remove_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	struct ast_json			*json_object	= NULL;
	struct ast_json_payload *payload		= NULL;
	const char              *queue_exten    = NULL;
	const char              *member_exten   = NULL;
	const char              *queue_agent    = NULL;
	const char              *need_login     = NULL;
	const char              *cur_status     = NULL;
	const char              *queue_name     = NULL;

	if (!msg) {
		ast_log(LOG_WARNING, "Stasis message is NULL!\n");
		return;
	}

	payload = stasis_message_data(msg);
	if (!payload) {
		ast_debug(LOG_WARNING, "Got NULL payload from stasis message!\n");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL!\n");
		return;
	}

	queue_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueExtension"));
	member_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueMember"));
	queue_agent = ast_json_string_get(ast_json_object_get(json_object, "QueueAgent"));
	need_login = ast_json_string_get(ast_json_object_get(json_object, "NeedLogin"));
	cur_status = ast_json_string_get(ast_json_object_get(json_object, "CurStatus"));
	queue_name = ast_json_string_get(ast_json_object_get(json_object, "QueueName"));
	
	if (!queue_exten || !member_exten || !queue_agent || !need_login || !cur_status || !queue_name) {
		ast_log(LOG_WARNING, "Incomplete JSON object data\n");
		return;
	}

	char *member = strstr(member_exten, "/");
	if (member) {
		member++;
	}

	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		if (!strcmp(cur->queue_extension, queue_exten) && !strcmp(cur->member_extension, member)) {
			ast_copy_string(cur->queue_agent, queue_agent, sizeof(cur->queue_agent));
			ast_copy_string(cur->need_login, need_login, sizeof(cur->need_login));
			ast_copy_string(cur->cur_status, cur_status, sizeof(cur->cur_status));
			ast_copy_string(cur->queue_name, queue_name, sizeof(cur->queue_name));
			ast_log(LOG_DEBUG, "Queue agent receive 'update' event!\n");
			queue_agent_send_notify_contact(member);
			AST_LIST_REMOVE_CURRENT(entry);
			ast_free(cur);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);
}

static void queue_agent_event_destroy_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(cur);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);
}

static void queue_agent_publish_cb(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	struct ast_json			*json_object	= NULL;
	struct ast_json_payload *payload		= NULL;
	const char              *queue_exten    = NULL;
	const char              *member_exten   = NULL;
	const char              *cur_status     = NULL;

	if (!msg) {
		ast_log(LOG_WARNING, "Stasis message is NULL!\n");
		return;
	}

	payload = stasis_message_data(msg);
	if (!payload) {
		ast_debug(LOG_WARNING, "Got NULL payload from stasis message!\n");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL!\n");
		return;
	}

	queue_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueExtension"));
	member_exten = ast_json_string_get(ast_json_object_get(json_object, "QueueMember"));
	cur_status = ast_json_string_get(ast_json_object_get(json_object, "CurStatus"));
	if (!queue_exten || !member_exten || !cur_status) {
		ast_log(LOG_WARNING, "Incomplete JSON object data\n");
		return;
	}

	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		if (!strcmp(cur->queue_extension, queue_exten) && !strcmp(cur->member_extension, member_exten)) {
			ast_copy_string(cur->cur_status, cur_status, sizeof(cur->cur_status));

			RAII_VAR(struct stasis_message *, tmp_msg, NULL, ao2_cleanup);
			tmp_msg = stasis_message_create(ast_queue_agent_login_logoff_type(), payload);
			if (!tmp_msg) {
				ast_log(LOG_WARNING, "Create queue agent message failed!\n");
				return;
			}
			stasis_publish(ast_queue_topic_all(), tmp_msg);
			
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);
}

static int queue_agent_send_to_one_contact(struct ast_sip_contact *contact)
{
	struct ast_sip_body body;
	struct ast_str *body_text = NULL;
	unsigned int find_endpt_flag = 0;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (!contact) {
		ast_log(LOG_WARNING, "The contact is invalid.\n");
		return -1;
	}

	body.type = QUEUE_AGENT_TYPE;
	body.subtype = QUEUE_AGENT_SUBTYPE;

	if (!(body_text = create_queue_agent_body(contact->aor, &find_endpt_flag))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for queue agent notify.\n");
		return -1;
	}

	if (0 == find_endpt_flag) {
		ast_log(LOG_VERBOSE, "Endpoint(%s) was not queue agents.\n", contact->aor);
		ast_free(body_text);
		return -1;
	}

	if (contact->endpoint) {
		endpoint = ao2_bump(contact->endpoint);
	} else {
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", contact->aor);
	}

	body.body_text = ast_str_buffer(body_text);
	ast_log(LOG_NOTICE, "Sending ucm-cc NOTIFY to endpoint %s\n", contact->aor);

	struct queue_agent_unsolicited_data queue_agent_data = {
		.sub = NULL,
		.endpoint = endpoint,
		.body = &body,
	};

	queue_agent_send_unsolicited_notify_to_contact((void *)contact, (void *)&queue_agent_data, 0);
	ast_free(body_text);

	return 0;
}

static void queue_agent_contact_added(const void *object)
{
	struct ast_sip_contact *contact = (struct ast_sip_contact *)object;
	struct ast_pjsip_queue_agent *cur;
	int update_flag = 0;

	if (!contact) {
		return;
	}

	ast_log(LOG_NOTICE, "queue_agent_contact_added, contact[%s]", ast_sorcery_object_get_id(contact));

	AST_LIST_LOCK(&g_queue_agent);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		if (!strcmp(cur->member_extension, contact->aor)) {
			update_flag = 1;
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);

	if (update_flag) {
		queue_agent_send_to_one_contact(contact);
	}

	return;
}

/*! \brief Observer for contacts so unsolicited MWI is sent when a contact changes */
static const struct ast_sorcery_observer queue_agent_contact_observer = {
	.created = queue_agent_contact_added,
};

static int reload(void)
{
#if 0
	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		ast_log(LOG_DEBUG, "Queue agent start reload!\n");
		queue_agent_send_notify_contact(cur->queue_agent);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);
#endif
	return 0;
}

#define HSMCONN_FORMAT " %-10s  %-10s  %-25.30s  %-10.30s  %-15s\n"
static char *queue_agent_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
		case CLI_INIT:
			e->command = "queue show agents";
			e->usage =
					"Usage: queue show\n"
					"       Provides summary information on a specified queue.\n";
			return NULL;
		case CLI_GENERATE:
			break;
	}

	ast_cli(a->fd, HSMCONN_FORMAT, "QueueName", "QueueAgent", "IsAgent", "NeedLogin", "CurrentStatus");
	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		ast_cli(a->fd, HSMCONN_FORMAT, cur->queue_extension, cur->member_extension,
			cur->queue_agent, cur->need_login, cur->cur_status);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_queue_agent[] = {
	AST_CLI_DEFINE(queue_agent_show, "Show status of queue agent"),
};

static int queue_agent_refresh_notify(const void *data)
{
	g_queue_agent_sched_id = 0;

	AST_LIST_LOCK(&g_queue_agent);
	struct ast_pjsip_queue_agent *cur;
	/*! Begin: Added by pxie@grandstream.cn for bug[87714], 2017-06-21. */
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		cur->refresh_flag = 0;
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);
	/*! Ended: Added by pxie@grandstream.cn for bug[87714], 2017-06-21. */

	AST_LIST_LOCK(&g_queue_agent);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&g_queue_agent, cur, entry) {
		if (0 == cur->refresh_flag && strcmp(cur->cur_status, "logoff")) {
			ast_log(LOG_NOTICE, "Queue agent start 'refresh' event!\n");
			queue_agent_send_notify_contact(cur->member_extension);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&g_queue_agent);

	if ((g_queue_agent_sched_id = ast_sched_add(queue_agent_sched, QUEUE_AGENT_REFRESH_TIMEOUT,
		queue_agent_refresh_notify, NULL)) < 0) {
		ast_log(LOG_ERROR, "Add scheduler timer failure!\n");
		return -1;
	}

	return 0;
}

static int unload_module(void)
{
	RAII_VAR(struct ao2_container *, queue_agent_subscriptions, ao2_global_obj_ref(queue_agent_unsolicited), ao2_cleanup);
	int res = 0;

	if (queue_agent_subscriptions) {
		ao2_global_obj_release(queue_agent_unsolicited);
	}

	queue_agent_event_destroy_cb(NULL, NULL, NULL);

	stasis_message_router_unsubscribe_and_join(queue_agent_router);
	queue_agent_router = NULL;

	ast_sorcery_observer_remove(ast_sip_get_sorcery(), "contact", &queue_agent_contact_observer);
	ast_cli_unregister_multiple(cli_queue_agent, ARRAY_LEN(cli_queue_agent));

	if (queue_agent_sched) {
		if (g_queue_agent_sched_id) {
			res = ast_sched_del(queue_agent_sched, g_queue_agent_sched_id);
		}

		ast_sched_context_destroy(queue_agent_sched);
	}

	return res;
}

static int load_module(void)
{
	CHECK_PJSIP_MODULE_LOADED();

	int res = 0;

	queue_agent_topic = ast_queue_topic_all();
	if (!queue_agent_topic) {
		ast_log(LOG_WARNING, "Cann't get queue topic!\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	queue_agent_router = stasis_message_router_create(queue_agent_topic);
	if (!queue_agent_router) {
		ast_log(LOG_WARNING, "Cann't create queue message router!\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	res |= stasis_message_router_add(queue_agent_router,
		ast_queue_agent_event_insert_type(),
		queue_agent_event_insert_cb,
		NULL);

	res |= stasis_message_router_add(queue_agent_router,
		ast_queue_agent_event_remove_type(),
		queue_agent_event_remove_cb,
		NULL);

	res |= stasis_message_router_add(queue_agent_router,
		ast_queue_agent_event_update_type(),
		queue_agent_event_update_cb,
		NULL);

	res |= stasis_message_router_add(queue_agent_router,
		ast_queue_agent_event_destroy_type(),
		queue_agent_event_destroy_cb,
		NULL);

	res |= stasis_message_router_add(queue_agent_router,
		ast_queue_agent_publish_type(),
		queue_agent_publish_cb,
		NULL);

	res |= ast_cli_register_multiple(cli_queue_agent, ARRAY_LEN(cli_queue_agent));

	ast_sorcery_observer_add(ast_sip_get_sorcery(), "contact", &queue_agent_contact_observer);

	if (!(queue_agent_sched = ast_sched_context_create())) {
		ast_log(LOG_ERROR, "Unable to create scheduler context.\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (ast_sched_start_thread(queue_agent_sched)) {
		ast_log(LOG_ERROR, "Unable to create scheduler context thread.\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if ((g_queue_agent_sched_id = ast_sched_add(queue_agent_sched, QUEUE_AGENT_REFRESH_TIMEOUT,
		queue_agent_refresh_notify, NULL)) < 0) {
		ast_log(LOG_ERROR, "Unable to add scheduler timer.\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP queue agent resource",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_CHANNEL_DEPEND,
);
#endif
