/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/bridge.h"
#include "asterisk/framehook.h"
#include "asterisk/stasis_bridges.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/causes.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/channel.h"
#include "asterisk/dial.h"
#include "asterisk/mixmonitor.h"

AST_MODULE_LOG("pjsip");
#endif

/*! \brief REFER Progress structure */
struct refer_progress {
	/*! \brief Subscription to provide updates on */
	pjsip_evsub *sub;
	/*! \brief Dialog for subscription */
	pjsip_dialog *dlg;
	/*! \brief Received packet, used to construct final response in case no subscription exists */
	pjsip_rx_data *rdata;
	/*! \brief Frame hook for monitoring REFER progress */
	int framehook;
	/*! \brief Last received subclass in frame hook */
	int subclass;
	/*! \brief Serializer for notifications */
	struct ast_taskprocessor *serializer;
	/*! \brief Stasis subscription for bridge events */
	struct stasis_subscription *bridge_sub;
	/*! \brief Reference to transfer_channel_data related to the refer */
	struct transfer_channel_data *transfer_data;
	/*! \brief Uniqueid of transferee channel */
	char *transferee;
	/*! \brief Non-zero if the 100 notify has been sent */
	int sent_100;
	/*! \brief Whether to notifies all the progress details on blind transfer */
	unsigned int refer_blind_progress;
};

/*! \brief REFER Progress notification structure */
struct refer_progress_notification {
	/*! \brief Refer progress structure to send notification on */
	struct refer_progress *progress;
	/*! \brief SIP response code to send */
	int response;
	/*! \brief Subscription state */
	pjsip_evsub_state state;
};

/*! \brief REFER Progress module, used to attach REFER progress structure to subscriptions */
static pjsip_module refer_progress_module = {
	.name = { "REFER Progress", 14 },
	.id = -1,
};

/*! \brief Destructor for REFER Progress notification structure */
static void refer_progress_notification_destroy(void *obj)
{
	struct refer_progress_notification *notification = obj;

	ao2_cleanup(notification->progress);
}

/*! \brief Allocator for REFER Progress notification structure */
static struct refer_progress_notification *refer_progress_notification_alloc(struct refer_progress *progress, int response,
	pjsip_evsub_state state)
{
	struct refer_progress_notification *notification = ao2_alloc(sizeof(*notification), refer_progress_notification_destroy);

	if (!notification) {
		return NULL;
	}

	ao2_ref(progress, +1);
	notification->progress = progress;
	notification->response = response;
	notification->state = state;

	return notification;
}

/*! \brief Serialized callback for subscription notification */
static int refer_progress_notify(void *data)
{
	RAII_VAR(struct refer_progress_notification *, notification, data, ao2_cleanup);
	pjsip_evsub *sub;
	pjsip_tx_data *tdata;

	/* If the subscription has already been terminated we can't send a notification */
	if (!(sub = notification->progress->sub)) {
		ast_debug(3, "Not sending NOTIFY of response '%d' and state '%u' on progress monitor '%p' as subscription has been terminated\n",
			notification->response, notification->state, notification->progress);
		return 0;
	}

	/* If the subscription is being terminated we want to actually remove the progress structure here to
	 * stop a deadlock from occurring - basically terminated changes the state which queues a synchronous task
	 * but we are already running a task... thus it would deadlock */
	if (notification->state == PJSIP_EVSUB_STATE_TERMINATED) {
		ast_debug(3, "Subscription '%p' is being terminated as a result of a NOTIFY, removing REFER progress structure early on progress monitor '%p'\n",
			notification->progress->sub, notification->progress);
		pjsip_dlg_inc_lock(notification->progress->dlg);
#ifdef GRANDSTREAM_NETWORKS
		pjsip_evsub_set_mod_data(sub, refer_progress_module.id, NULL);
#else
		pjsip_evsub_set_mod_data(notification->progress->sub, refer_progress_module.id, NULL);
#endif
		pjsip_dlg_dec_lock(notification->progress->dlg);

		/* This is for dropping the reference on the subscription */
		ao2_cleanup(notification->progress);

		notification->progress->sub = NULL;
	}

	/* Send a deferred initial 100 Trying SIP frag NOTIFY if we haven't already. */
	if (!notification->progress->sent_100) {
		notification->progress->sent_100 = 1;
		if (notification->response != 100) {
			ast_debug(3, "Sending initial 100 Trying NOTIFY for progress monitor '%p'\n",
				notification->progress);
			if (pjsip_xfer_notify(sub, PJSIP_EVSUB_STATE_ACTIVE, 100, NULL, &tdata) == PJ_SUCCESS) {
				pjsip_xfer_send_request(sub, tdata);
			}
		}
	}

	ast_debug(3, "Sending NOTIFY with response '%d' and state '%u' on subscription '%p' and progress monitor '%p'\n",
		notification->response, notification->state, sub, notification->progress);

	/* Actually send the notification */
	if (pjsip_xfer_notify(sub, notification->state, notification->response, NULL, &tdata) == PJ_SUCCESS) {
		pjsip_xfer_send_request(sub, tdata);
	}

	return 0;
}

static void refer_progress_bridge(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	struct refer_progress *progress = data;
	struct ast_bridge_blob *enter_blob;
	struct refer_progress_notification *notification;
	struct ast_channel *chan;

	if (stasis_subscription_final_message(sub, message)) {
		ao2_ref(progress, -1);
		return;
	}

	if (ast_channel_entered_bridge_type() != stasis_message_type(message)) {
		/* Don't care */
		return;
	}

	enter_blob = stasis_message_data(message);
	if (strcmp(enter_blob->channel->uniqueid, progress->transferee)) {
		/* Don't care */
		return;
	}

	if (!progress->transfer_data->completed) {
		/* We can't act on this message because the transfer_channel_data doesn't show that
		 * the transfer is ready to progress */
		return;
	}

	/* OMG the transferee is joining a bridge. His call got answered! */
	notification = refer_progress_notification_alloc(progress, 200, PJSIP_EVSUB_STATE_TERMINATED);
	if (notification) {
		if (ast_sip_push_task(progress->serializer, refer_progress_notify, notification)) {
			ao2_cleanup(notification);
		}
#ifdef GRANDSTREAM_NETWORKS
		/*Add for Bug223248*/
		struct stasis_subscription *bridge_sub = NULL;
		ao2_lock(progress);
		if (progress->bridge_sub) {
			bridge_sub = progress->bridge_sub;
			progress->bridge_sub = NULL;
		}
		ao2_unlock(progress);

		if (bridge_sub) {
			stasis_unsubscribe(bridge_sub);
		}
#else
		progress->bridge_sub = stasis_unsubscribe(progress->bridge_sub);
#endif
	}

	chan = ast_channel_get_by_name(progress->transferee);
	if (!chan) {
		/* The channel is already gone */
		return;
	}

	ast_channel_lock(chan);
	ast_debug(3, "Detaching REFER progress monitoring hook from '%s' as it has joined a bridge\n",
		ast_channel_name(chan));
	ast_framehook_detach(chan, progress->framehook);
	ast_channel_unlock(chan);

	ast_channel_unref(chan);
}

/*! \brief Progress monitoring frame hook - examines frames to determine state of transfer */
static struct ast_frame *refer_progress_framehook(struct ast_channel *chan, struct ast_frame *f, enum ast_framehook_event event, void *data)
{
	struct refer_progress *progress = data;
	struct refer_progress_notification *notification = NULL;

	/* We only care about frames *to* the channel */
	if (!f || (event != AST_FRAMEHOOK_EVENT_WRITE)) {
		return f;
	}

	/* If the completed flag hasn't been raised, skip this pass. */
	if (!progress->transfer_data->completed) {
		return f;
	}

	/* Determine the state of the REFER based on the control frames (or voice frames) passing */
	if (f->frametype == AST_FRAME_VOICE && !progress->subclass) {
		/* Media is passing without progress, this means the call has been answered */
		notification = refer_progress_notification_alloc(progress, 200, PJSIP_EVSUB_STATE_TERMINATED);
	} else if (f->frametype == AST_FRAME_CONTROL) {
		/* Based on the control frame being written we can send a NOTIFY advising of the progress */
		if ((f->subclass.integer == AST_CONTROL_RING) || (f->subclass.integer == AST_CONTROL_RINGING)) {
			progress->subclass = f->subclass.integer;
			notification = refer_progress_notification_alloc(progress, 180, PJSIP_EVSUB_STATE_ACTIVE);
		} else if (f->subclass.integer == AST_CONTROL_BUSY) {
			progress->subclass = f->subclass.integer;
			notification = refer_progress_notification_alloc(progress, 486, PJSIP_EVSUB_STATE_TERMINATED);
		} else if (f->subclass.integer == AST_CONTROL_CONGESTION) {
			progress->subclass = f->subclass.integer;
			notification = refer_progress_notification_alloc(progress, 503, PJSIP_EVSUB_STATE_TERMINATED);
		} else if (f->subclass.integer == AST_CONTROL_PROGRESS) {
			progress->subclass = f->subclass.integer;
			notification = refer_progress_notification_alloc(progress, 183, PJSIP_EVSUB_STATE_ACTIVE);
		} else if (f->subclass.integer == AST_CONTROL_PROCEEDING) {
			progress->subclass = f->subclass.integer;
			notification = refer_progress_notification_alloc(progress, 100, PJSIP_EVSUB_STATE_ACTIVE);
		} else if (f->subclass.integer == AST_CONTROL_ANSWER) {
			progress->subclass = f->subclass.integer;
			notification = refer_progress_notification_alloc(progress, 200, PJSIP_EVSUB_STATE_TERMINATED);
		}
	}

	/*Start added by zyxia for  Bug 143733  telephone blind transfer,ucm didnot send notify 200 ok to telelphone 2019.8.20  */
#ifdef GRANDSTREAM_NETWORKS
	if (!notification) {
		notification = refer_progress_notification_alloc(progress, 200, PJSIP_EVSUB_STATE_TERMINATED);
	}
#endif
	/*End added by zyxia for  Bug 143733  telephone blind transfer,ucm didnot send notify 200 ok to telelphone 2019.8.20  */

	/* If a notification is due to be sent push it to the thread pool */
	if (notification) {
#ifdef GRANDSTREAM_NETWORKS
		struct refer_progress_notification *oknotification = NULL;
		if ((f->subclass.integer == AST_CONTROL_RING) || (f->subclass.integer == AST_CONTROL_RINGING) || (f->subclass.integer == AST_CONTROL_PROGRESS)) {
			progress->subclass = f->subclass.integer;
			oknotification = refer_progress_notification_alloc(progress, 200, PJSIP_EVSUB_STATE_TERMINATED);
		}
		if (oknotification) {
			if (ast_sip_push_task(progress->serializer, refer_progress_notify, oknotification)) {
				ao2_cleanup(oknotification);
			}
		}
#endif
		/* If the subscription is being terminated we don't need the frame hook any longer */
		if (notification->state == PJSIP_EVSUB_STATE_TERMINATED) {
			ast_debug(3, "Detaching REFER progress monitoring hook from '%s' as subscription is being terminated\n",
				ast_channel_name(chan));
			ast_framehook_detach(chan, progress->framehook);
		}

		if (ast_sip_push_task(progress->serializer, refer_progress_notify, notification)) {
			ao2_cleanup(notification);
		}
	}

	return f;
}

/*! \brief Destroy callback for monitoring framehook */
static void refer_progress_framehook_destroy(void *data)
{
	struct refer_progress *progress = data;
	struct refer_progress_notification *notification = refer_progress_notification_alloc(progress, 503, PJSIP_EVSUB_STATE_TERMINATED);

	if (notification && ast_sip_push_task(progress->serializer, refer_progress_notify, notification)) {
		ao2_cleanup(notification);
	}

#ifdef GRANDSTREAM_NETWORKS
	/*Add for Bug223248*/
	struct stasis_subscription *bridge_sub = NULL;
	ao2_lock(progress);
	if (progress->bridge_sub) {
		bridge_sub = progress->bridge_sub;
		progress->bridge_sub = NULL;
	}
	ao2_unlock(progress);

	if (bridge_sub) {
		stasis_unsubscribe(bridge_sub);
	}
#else
	if (progress->bridge_sub) {
		progress->bridge_sub = stasis_unsubscribe(progress->bridge_sub);
	}
#endif

	ao2_cleanup(progress);
}

/*! \brief Serialized callback for subscription termination */
static int refer_progress_terminate(void *data)
{
	struct refer_progress *progress = data;

	/* The subscription is no longer valid */
	progress->sub = NULL;

	return 0;
}

/*! \brief Callback for REFER subscription state changes */
static void refer_progress_on_evsub_state(pjsip_evsub *sub, pjsip_event *event)
{
	struct refer_progress *progress = pjsip_evsub_get_mod_data(sub, refer_progress_module.id);

	/* If being destroyed queue it up to the serializer */
	if (progress && (pjsip_evsub_get_state(sub) == PJSIP_EVSUB_STATE_TERMINATED)) {
		/* To prevent a deadlock race condition we unlock the dialog so other serialized tasks can execute */
		ast_debug(3, "Subscription '%p' has been remotely terminated, waiting for other tasks to complete on progress monitor '%p'\n",
			sub, progress);

		/* It's possible that a task is waiting to remove us already, so bump the refcount of progress so it doesn't get destroyed */
		ao2_ref(progress, +1);
		pjsip_dlg_dec_lock(progress->dlg);
		/*
		 * XXX We are always going to execute this inline rather than
		 * in the serializer because this function is a PJPROJECT
		 * callback and thus has to be a SIP servant thread.
		 *
		 * The likely remedy is to push most of this function into
		 * refer_progress_terminate() with ast_sip_push_task().
		 */
		ast_sip_push_task_wait_servant(progress->serializer, refer_progress_terminate, progress);
		pjsip_dlg_inc_lock(progress->dlg);
		ao2_ref(progress, -1);

		ast_debug(3, "Subscription '%p' removed from progress monitor '%p'\n", sub, progress);

		/* Since it was unlocked it is possible for this to have been removed already, so check again */
		if (pjsip_evsub_get_mod_data(sub, refer_progress_module.id)) {
			pjsip_evsub_set_mod_data(sub, refer_progress_module.id, NULL);
			ao2_cleanup(progress);
		}
	}
}

/*! \brief Callback structure for subscription */
static pjsip_evsub_user refer_progress_evsub_cb = {
	.on_evsub_state = refer_progress_on_evsub_state,
};

/*! \brief Destructor for REFER progress sutrcture */
static void refer_progress_destroy(void *obj)
{
	struct refer_progress *progress = obj;

	if (progress->bridge_sub) {
		progress->bridge_sub = stasis_unsubscribe(progress->bridge_sub);
	}

	ao2_cleanup(progress->transfer_data);

	ast_free(progress->transferee);
	ast_taskprocessor_unreference(progress->serializer);
}

/*! \brief Internal helper function which sets up a refer progress structure if needed */
static int refer_progress_alloc(struct ast_sip_session *session, pjsip_rx_data *rdata, struct refer_progress **progress)
{
	const pj_str_t str_refer_sub = { "Refer-Sub", 9 };
	pjsip_generic_string_hdr *refer_sub = NULL;
	const pj_str_t str_true = { "true", 4 };
	pjsip_hdr hdr_list;
	char tps_name[AST_TASKPROCESSOR_MAX_NAME + 1];

	*progress = NULL;

	/* Grab the optional Refer-Sub header, it can be used to suppress the implicit subscription */
	refer_sub = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &str_refer_sub, NULL);
	if ((refer_sub && pj_strnicmp(&refer_sub->hvalue, &str_true, 4))) {
		return 0;
	}

	if (!(*progress = ao2_alloc(sizeof(struct refer_progress), refer_progress_destroy))) {
		return -1;
	}

	ast_debug(3, "Created progress monitor '%p' for transfer occurring from channel '%s' and endpoint '%s'\n",
		progress, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));

	(*progress)->refer_blind_progress = session->endpoint->refer_blind_progress;

	(*progress)->framehook = -1;

	/* To prevent a potential deadlock we need the dialog so we can lock/unlock */
	(*progress)->dlg = session->inv_session->dlg;

	/* Create name with seq number appended. */
	ast_taskprocessor_build_name(tps_name, sizeof(tps_name), "pjsip/refer/%s",
		ast_sorcery_object_get_id(session->endpoint));

	if (!((*progress)->serializer = ast_sip_create_serializer(tps_name))) {
		goto error;
	}

	/* Create the implicit subscription for monitoring of this transfer */
	if (pjsip_xfer_create_uas(session->inv_session->dlg, &refer_progress_evsub_cb, rdata, &(*progress)->sub) != PJ_SUCCESS) {
		goto error;
	}

	/* Associate the REFER progress structure with the subscription */
	ao2_ref(*progress, +1);
	pjsip_evsub_set_mod_data((*progress)->sub, refer_progress_module.id, *progress);

	pj_list_init(&hdr_list);
	if (refer_sub) {
		pjsip_hdr *hdr = (pjsip_hdr*)pjsip_generic_string_hdr_create(session->inv_session->dlg->pool, &str_refer_sub, &str_true);

		pj_list_push_back(&hdr_list, hdr);
	}

	/* Accept the REFER request */
	ast_debug(3, "Accepting REFER request for progress monitor '%p'\n", *progress);
	pjsip_xfer_accept((*progress)->sub, rdata, 202, &hdr_list);

	return 0;

error:
	ao2_cleanup(*progress);
	*progress = NULL;
	return -1;
}

/*! \brief Structure for attended transfer task */
struct refer_attended {
	/*! \brief Transferer session */
	struct ast_sip_session *transferer;
	/*! \brief Transferer channel */
	struct ast_channel *transferer_chan;
	/*! \brief Second transferer session */
	struct ast_sip_session *transferer_second;
	/*! \brief Optional refer progress structure */
	struct refer_progress *progress;
};

/*! \brief Destructor for attended transfer task */
static void refer_attended_destroy(void *obj)
{
	struct refer_attended *attended = obj;

	ao2_cleanup(attended->transferer);
	ast_channel_cleanup(attended->transferer_chan);
	ao2_cleanup(attended->transferer_second);
	ao2_cleanup(attended->progress);
}

/*! \brief Allocator for attended transfer task */
static struct refer_attended *refer_attended_alloc(struct ast_sip_session *transferer,
	struct ast_sip_session *transferer_second,
	struct refer_progress *progress)
{
	struct refer_attended *attended;

	attended = ao2_alloc_options(sizeof(*attended), refer_attended_destroy,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!attended) {
		return NULL;
	}

	ao2_ref(transferer, +1);
	attended->transferer = transferer;
	ast_channel_ref(transferer->channel);
	attended->transferer_chan = transferer->channel;
	ao2_ref(transferer_second, +1);
	attended->transferer_second = transferer_second;

	if (progress) {
		ao2_ref(progress, +1);
		attended->progress = progress;
	}

	return attended;
}

static int session_end_if_deferred_task(void *data)
{
	struct ast_sip_session *session = data;

	ast_sip_session_end_if_deferred(session);
	ao2_ref(session, -1);
	return 0;
}

static int defer_termination_cancel_task(void *data)
{
	struct ast_sip_session *session = data;

	ast_sip_session_end_if_deferred(session);
	ast_sip_session_defer_termination_cancel(session);
	ao2_ref(session, -1);
	return 0;
}

/*!
 * \internal
 * \brief Convert transfer enum to SIP response code.
 * \since 13.3.0
 *
 * \param xfer_code Core transfer function enum result.
 *
 * \return SIP response code
 */
static int xfer_response_code2sip(enum ast_transfer_result xfer_code)
{
	int response;

	response = 503;
	switch (xfer_code) {
	case AST_BRIDGE_TRANSFER_INVALID:
		response = 400;
		break;
	case AST_BRIDGE_TRANSFER_NOT_PERMITTED:
		response = 403;
		break;
	case AST_BRIDGE_TRANSFER_FAIL:
		response = 500;
		break;
	case AST_BRIDGE_TRANSFER_SUCCESS:
		response = 200;
		break;
	}
	return response;
}

#ifdef GRANDSTREAM_NETWORKS
struct pjmedia_sdp_neg
{
    pjmedia_sdp_neg_state state;	    /**< Negotiator state.	     */
    pj_bool_t		  prefer_remote_codec_order;
    pj_bool_t             answer_with_multiple_codecs;
    pj_bool_t		  has_remote_answer;
    pj_bool_t		  answer_was_remote;

    pjmedia_sdp_session	*initial_sdp,	    /**< Initial local SDP	     */
			*initial_sdp_tmp,   /**< Temporary initial local SDP */
			*active_local_sdp,  /**< Currently active local SDP. */
			*active_remote_sdp, /**< Currently active remote's.  */
			*neg_local_sdp,	    /**< Temporary local SDP.	     */
			*neg_remote_sdp;    /**< Temporary remote SDP.	     */
};

typedef struct pjmedia_sdp_neg pjmedia_sdp_neg;

static int judge_if_need_send_reinvite(struct ast_sip_session *session)
{
	pjmedia_sdp_neg_state chan_neg_state = session->inv_session->neg->state;
	pjmedia_sdp_session *chan_local_sdp = NULL;
	pj_pool_t *pool = session->inv_session->pool_prov;

	if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		chan_local_sdp = pjmedia_sdp_session_clone(pool, session->inv_session->neg->neg_local_sdp);
	} else if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_DONE) {
		chan_local_sdp = pjmedia_sdp_session_clone(pool, session->inv_session->neg->active_local_sdp);
	}

	if (chan_local_sdp) {
		unsigned i;
		pjmedia_sdp_media *im;
		for (i = 0; i < chan_local_sdp->media_count; ++i) {
			im = chan_local_sdp->media[i];
			if (!pjmedia_sdp_media_find_attr2(im, "sendrecv", NULL)) {
				return 1;
			} else {
				continue;
			}
		}
		return 0;
	}

	return 0;
}

static void exchange_peer_remote_sdp(struct ast_channel *chan, const struct ast_channel *peer)
{
	if (NULL == chan || NULL == peer) {
		ast_log(LOG_ERROR, "exchange peer remote sdp error");
		return;
	}
	ast_log(LOG_DEBUG, "exchange peer remote sdp begin");

	if (ast_channel_tech(chan) && (0 == strcmp(ast_channel_tech(chan)->type, "PJSIP")) &&
		ast_channel_tech(peer) && (0 == strcmp(ast_channel_tech(peer)->type, "PJSIP"))) {
		struct ast_sip_channel_pvt *chan_pvt = ast_channel_tech_pvt(chan);
		struct ast_sip_channel_pvt *peer_pvt = ast_channel_tech_pvt(peer);
		if (!chan_pvt || !peer_pvt) {
			ast_log(LOG_WARNING, "exchange peer remote sdp, channel has no pvt\n");
			return;
		}
		
		if (!chan_pvt->session || !peer_pvt->session) {
			ast_log(LOG_WARNING, "exchange peer remote sdp, channel has no session");
			return;
		}

		if (!chan_pvt->session->inv_session || !peer_pvt->session->inv_session) {
			ast_log(LOG_WARNING, "exchange peer remote sdp, channel has no invite session");
			return;
		}

		if (!chan_pvt->session->is_direct_media || !peer_pvt->session->is_direct_media) {
			ast_log(LOG_DEBUG, "exchange peer remote sdp, direct media is not enabled");
			chan_pvt->session->is_direct_media = 0;
			peer_pvt->session->is_direct_media = 0;
			ast_sip_session_init_stream_media_direction(chan_pvt->session);
			ast_sip_session_init_stream_media_direction(peer_pvt->session);
			return;
		}

		pjmedia_sdp_neg_state chan_neg_state = chan_pvt->session->inv_session->neg->state;
		pjmedia_sdp_neg_state peer_neg_state = peer_pvt->session->inv_session->neg->state;
		pjmedia_sdp_session *chan_local_sdp = NULL;
		pjmedia_sdp_session *peer_local_sdp = NULL;
		if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_DONE && peer_neg_state == PJMEDIA_SDP_NEG_STATE_DONE) {
			chan_local_sdp = pjmedia_sdp_session_clone(chan_pvt->session->inv_session->pool, peer_pvt->session->inv_session->neg->active_remote_sdp);
			peer_local_sdp = pjmedia_sdp_session_clone(peer_pvt->session->inv_session->pool, chan_pvt->session->inv_session->neg->active_remote_sdp);
			unsigned i;
			pjmedia_sdp_media *im;
			for (i = 0; i < chan_local_sdp->media_count; ++i) {
				im = chan_local_sdp->media[i];
				if (pjmedia_sdp_media_find_attr2(im, "recvonly", NULL)) {
					pjmedia_sdp_attr *attr;
					pjmedia_sdp_media_remove_all_attr(im, "recvonly");
					attr = pjmedia_sdp_attr_create(chan_pvt->session->inv_session->pool, "sendrecv", NULL);
					pjmedia_sdp_media_add_attr(im, attr);
				}
			}
			for (i = 0; i < peer_local_sdp->media_count; ++i) {
				im = peer_local_sdp->media[i];
				if (pjmedia_sdp_media_find_attr2(im, "recvonly", NULL)) {
					pjmedia_sdp_attr *attr;
					pjmedia_sdp_media_remove_all_attr(im, "recvonly");
					attr = pjmedia_sdp_attr_create(peer_pvt->session->inv_session->pool, "sendrecv", NULL);
					pjmedia_sdp_media_add_attr(im, attr);
				}
			}
		}

		char buf[1024] = {0};
		if (chan_local_sdp) {
			pjsip_inv_set_local_sdp(chan_pvt->session->inv_session, chan_local_sdp);
			{
				pjmedia_sdp_print(chan_pvt->session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
				ast_log(LOG_DEBUG, "exchange peer remote sdp, chan local sdp is:\n%s", buf);
			}
		}
		if (peer_local_sdp) {
			pjsip_inv_set_local_sdp(peer_pvt->session->inv_session, peer_local_sdp);
			{
				memset(buf, 0, sizeof(buf));
				pjmedia_sdp_print(peer_pvt->session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
				ast_log(LOG_DEBUG, "exchange peer remote sdp, peer local sdp is:\n%s", buf);
			}
		}
	}
	return;
}

static void update_info_for_refer_attend(struct refer_attended *attended, int response, struct ast_channel *transfer_target)
{
	if (response != 200) {
		ast_log(LOG_DEBUG, "Transfer is failed, response:%d, don't need process\n", response);
		return;
	}

	if (!attended || !attended->transferer_chan || !attended->transferer_second || !attended->transferer_second->channel) {
		return;
	}

	if (transfer_target) {
		ast_log(LOG_DEBUG, "update_info_for_refer_attend, transfer_target:%s transferer_chan:%s, send_slides_video:%d, recv_slides_video:%d\n",
			ast_channel_name(transfer_target), ast_channel_name(attended->transferer_chan),
			ast_channel_get_send_slides_video(attended->transferer_chan),
			ast_channel_get_recv_slides_video(attended->transferer_chan));
		ast_channel_set_send_slides_video(transfer_target, ast_channel_get_send_slides_video(attended->transferer_chan));
		ast_channel_set_recv_slides_video(transfer_target, ast_channel_get_recv_slides_video(attended->transferer_chan));
		ast_indicate(transfer_target, AST_CONTROL_REFER_COMPLETE);
	} else {
		ast_log(LOG_DEBUG, "update_info_for_refer_attend attended->transferer_chan:%s, transferer_second->channel:%s, send_slides_video:%d, recv_slides_video:%d\n",
			ast_channel_name(attended->transferer_chan), ast_channel_name(attended->transferer_second->channel),
			ast_channel_get_send_slides_video(attended->transferer_chan),
			ast_channel_get_recv_slides_video(attended->transferer_chan));
		ast_channel_set_send_slides_video(attended->transferer_second->channel, ast_channel_get_send_slides_video(attended->transferer_chan));
		ast_channel_set_recv_slides_video(attended->transferer_second->channel, ast_channel_get_recv_slides_video(attended->transferer_chan));
		ast_queue_control(attended->transferer_second->channel, AST_CONTROL_REFER_TRANSFERER_HANGUP);
	}

	return;
}

static void notify_trunk_info_for_refer(struct ast_channel *transferer_peer_chan, struct ast_channel *transfer_target)
{
	struct ast_sip_channel_pvt *channel_pvt = NULL;
	struct ast_sip_session *session = NULL;
	struct ast_trunk_info trunk_info;

	if (!transferer_peer_chan) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_refer has no transferer_peer_chan\n");
		return;
	}

	if (!transfer_target) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_refer has no transfer_target\n");
		return;
	}

	if (strncmp(ast_channel_name(transferer_peer_chan), "PJSIP/trunk_", 12)) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_refer it's not trunk\n");
		return;
	}

	channel_pvt = ast_channel_tech_pvt(transferer_peer_chan);
	if (!channel_pvt) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_refer can't get channel_pvt\n");
		return;
	}

	session = channel_pvt->session;
	if (!channel_pvt->session) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_refer can't get session\n");
		return;
	}

	if (ast_strlen_zero(session->conf_link)) {
		ast_log(LOG_DEBUG, "[session:%p] notify_trunk_info_for_refer can't get conf_link, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_strlen_zero(session->trunk_channel_id)) {
		ast_log(LOG_DEBUG, "[session:%p] notify_trunk_info_for_refer can't get trunk_channel_id, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	memset(&trunk_info, 0, sizeof(struct ast_trunk_info));
	strncpy(trunk_info.conf_link, session->conf_link, sizeof(trunk_info.conf_link) - 1);
	strncpy(trunk_info.channel_id, session->trunk_channel_id, sizeof(trunk_info.channel_id) - 1);

	ast_log(LOG_DEBUG, "notify_trunk_info_for_refer transferer_peer_chan:%s, transfer_target:%s, conf_link:%s, trunk_channel_id:%s\n",
		ast_channel_name(transferer_peer_chan), ast_channel_name(transfer_target), trunk_info.conf_link, trunk_info.channel_id);

	ast_indicate_data(transfer_target, AST_CONTROL_NOTIFY_TRUNK_INFO, &trunk_info, sizeof(trunk_info));

	return;
}

static int refer_refresh(void *data)
{
	RAII_VAR(struct ast_sip_session *, refer_session, data, ao2_cleanup);

	ast_sip_session_refresh(refer_session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);

	return 0;
}

static int refer_refresh_without_new_sdp(void *data)
{
	RAII_VAR(struct ast_sip_session *, refer_session, data, ao2_cleanup);

	ast_sip_session_refresh(refer_session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 0, NULL);

	return 0;
}
#endif

/*! \brief Task for attended transfer executed by attended->transferer_second serializer */
static int refer_attended_task(void *data)
{
	struct refer_attended *attended = data;
	int response;
	int (*task_cb)(void *data);

	if (attended->transferer_second->channel) {
		ast_debug(3, "Performing a REFER attended transfer - Transferer #1: %s Transferer #2: %s\n",
			ast_channel_name(attended->transferer_chan),
			ast_channel_name(attended->transferer_second->channel));

#ifdef GRANDSTREAM_NETWORKS

		/* added by jklou for bug 55705 */
		RAII_VAR(struct ast_channel *, bridged_1, ast_channel_bridge_peer(attended->transferer_chan), ast_channel_cleanup);
		RAII_VAR(struct ast_channel *, bridged_2, ast_channel_bridge_peer(attended->transferer_second->channel), ast_channel_cleanup);

		unsigned int allow_forward_hold_request = ast_sip_session_get_allow_forward_hold_request();

		/*Start added by zyxia for Bug 241705,2023.4.11*/
		if(bridged_1 && bridged_2 ) {
			if(strstr(ast_channel_name(bridged_1),"RTSP")){
				pbx_builtin_setvar_helper(bridged_2, "CALL_TO_IPC", "yes");
			}
			/*Start added by zyxia for Bug 261649 ,2023.6.9*/
			else if (ast_channel_get_is_gds_call(bridged_1)){
				pbx_builtin_setvar_helper(bridged_2, "CALL_TO_GDS", "yes");
			}
			else if (ast_channel_get_is_hikvision_call(bridged_1)){
				pbx_builtin_setvar_helper(bridged_2, "CALL_TO_HIK", "yes");
			}
			/*End added by zyxia for Bug 261649 ,2023.6.9*/
		}
		/*End added by zyxia for Bug 241705,2023.4.11*/

#endif

		response = xfer_response_code2sip(ast_bridge_transfer_attended(
			attended->transferer_chan,
			attended->transferer_second->channel));

		ast_debug(3, "Final response for REFER attended transfer - Transferer #1: %s Transferer #2: %s is '%d'\n",
			ast_channel_name(attended->transferer_chan),
			ast_channel_name(attended->transferer_second->channel),
			response);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "bridge_1(transferer_chan's peer):%s, bridge_2(transferer_second's peer):%s",
					bridged_1 ? ast_channel_name(bridged_1) : "null",
					bridged_2 ? ast_channel_name(bridged_2) : "null");

		/*
		 *              ATTENDED TRANSFER BY ENDPOINT(!!!NOT FEATURE CODE!!!)
		 *        --------------------desired calling-----------------------
		 *       |                                                          |
		 *  (bridged_1) transferee -----> transferer                        |
		 *                             transferer_second --------> transfer_target (bridged_2)
		 */
		if (bridged_1) {
			/* added by zgmo for bug133681 */
			if (bridged_2) {
				int i;
				const char *callfilename;
				const char *mixmon_format;
				const char *rec_enable;
				const char *recording_chan;

				ast_log(LOG_DEBUG, "Refer Attended Transfer Normally");

				/* 2. prepare environments*/
				pbx_builtin_setvar_helper(bridged_2, "CF_REASON", "15");  /* attended transfer trigger by phone */
				pbx_builtin_setvar_helper(bridged_2, "TRUNK_INDEX", pbx_builtin_getvar_helper(bridged_1, "TRUNK_INDEX"));
				pbx_builtin_setvar_helper(bridged_2, "INBOUND_RT_INDEX", pbx_builtin_getvar_helper(bridged_1, "INBOUND_RT_INDEX"));
				pbx_builtin_setvar_helper(bridged_2, "OUTBOUND_TRUNK_INDEX", pbx_builtin_getvar_helper(bridged_1, "OUTBOUND_TRUNK_INDEX"));
				pbx_builtin_setvar_helper(bridged_2, "OUTBOUND_RT_INDEX", pbx_builtin_getvar_helper(bridged_1, "OUTBOUND_RT_INDEX"));
				pbx_builtin_setvar_helper(bridged_2, "__IN_RINGGROUP", pbx_builtin_getvar_helper(bridged_1, "IN_RINGGROUP"));
				pbx_builtin_setvar_helper(bridged_2, "__GROUP_NUMBER", pbx_builtin_getvar_helper(bridged_1, "GROUP_NUMBER"));
				pbx_builtin_setvar_helper(bridged_2, "AAUSER", pbx_builtin_getvar_helper(bridged_1, "AAUSER"));

				pbx_builtin_setvar_helper(bridged_2, "A_XFER_STATE", "refer success");

				/* 3. Stop recording if need */
				struct ast_channel *chan_array[3] = {bridged_1, bridged_2, attended->transferer_chan};

				for (i=0; i<3; i++) {
					if (!chan_array[i]) {
						continue;
					}

					recording_chan = pbx_builtin_getvar_helper(chan_array[i], "RECORDING_CHAN");
					ast_log(LOG_DEBUG, "%s's RECORDING_CHAN:%s",
						ast_channel_name(chan_array[i]), recording_chan ? : "none");

					if (!ast_strlen_zero(recording_chan) && !strcmp(recording_chan, ast_channel_name(chan_array[i]))) {
						ast_stop_mixmonitor(chan_array[i], NULL);
						ast_log(LOG_DEBUG, "Chan:%s is the recording channel.Stop it!",
							ast_channel_name(chan_array[i]));
						
					}
				}

				/* 4. Call Macro */
				if (ast_dial_check_record(bridged_1, bridged_2)) {
					callfilename = pbx_builtin_getvar_helper(bridged_2, "CALLFILENAME");
					mixmon_format = pbx_builtin_getvar_helper(bridged_2, "MIXMON_FORMAT");
					if (!ast_strlen_zero(callfilename) && !ast_strlen_zero(mixmon_format)) {
						char *record_file = NULL;
						ast_asprintf(&record_file, "%s.%s@", callfilename, mixmon_format);
						if (!ast_strlen_zero(record_file)) {
							ast_cdr_newest_setvar(ast_channel_name(bridged_1), "RECORD_FILES", record_file);
							ast_free(record_file);
						}
					}
				}

				rec_enable = pbx_builtin_getvar_helper(bridged_2, "REC_ENABLE");

				/* added by zgmo for sdp media record attribute */
				pbx_builtin_setvar_helper(bridged_1, "REC_ENABLE", rec_enable);
				if (!allow_forward_hold_request) {
					ast_indicate(bridged_1, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
					ast_indicate(bridged_2, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
				}
				/* end */
			} else {
				/*
				 *              ATTENDED TRANSFER BY ENDPOINT(!!!NOT FEATURE CODE!!!)
				 *        --------------------desired calling-----------------------
				 *       |                                                          |
				 *  (bridged_1) transferee -----> transferer                        |
				 *                             transferer_second --------> transfer_target (bridged_2)
				 */
				/*
				 *  This means transferer hangup before target answer
				 */
				const char *recording_chan = pbx_builtin_getvar_helper(attended->transferer_chan, "RECORDING_CHAN");
				const char *rec_enable = pbx_builtin_getvar_helper(bridged_1, "REC_ENABLE");
				const char *trunk_put_number = pbx_builtin_getvar_helper(attended->transferer_second->channel, "TRUNKPUTNUMBER");

				ast_log(LOG_DEBUG, "Refer Attended Transferer Hangup before target answer");

				ast_log(LOG_DEBUG, "transferer_chan:%s | RECORDING_CHAN:%s, trunk_put_number:%s",
					ast_channel_name(attended->transferer_chan),
					ast_strlen_zero(recording_chan) ? "":recording_chan,
					ast_strlen_zero(trunk_put_number) ? "":trunk_put_number);

				ast_log(LOG_DEBUG, "bridged_1:%s | RECORDING_CHAN:%s",
					ast_channel_name(bridged_1),
					ast_strlen_zero(rec_enable) ? "":rec_enable);

				if (!ast_strlen_zero(recording_chan)) {
					if (!strcmp(recording_chan, ast_channel_name(attended->transferer_chan))) {
						ast_stop_mixmonitor(attended->transferer_chan, NULL);
						ast_log(LOG_DEBUG, "Chan:%s is the recording channel.Stop it!",
								ast_channel_name(attended->transferer_chan));
					} else if(!strcmp(recording_chan, ast_channel_name(bridged_1))) {
						ast_stop_mixmonitor(bridged_1, NULL);
						ast_log(LOG_DEBUG, "Chan:%s is the recording channel.Stop it!",
							ast_channel_name(bridged_1));
					}

					pbx_builtin_setvar_helper(bridged_1, "REC_ENABLE", NULL);
					pbx_builtin_setvar_helper(bridged_1, "AUTO_RECORDING", NULL);
					pbx_builtin_setvar_helper(bridged_1, "MIXAUTORECORDING", NULL);
					pbx_builtin_setvar_helper(bridged_1, "MIXFILENAME", NULL);
				}

				pbx_builtin_setvar_helper(bridged_1, "CALL_OPTIONS", NULL);
				pbx_builtin_setvar_helper(bridged_1, "CF_REASON", "15");  /* attended transfer trigger by phone */
				pbx_builtin_setvar_helper(bridged_1, "A_XFER_STATE", "transferer hangup");
				pbx_builtin_setvar_helper(bridged_1, "A_XFEREE_NAME", ast_channel_name(bridged_1));

				if (!ast_strlen_zero(trunk_put_number)) {
					pbx_builtin_setvar_helper(bridged_1, "CALLEE_NUM", trunk_put_number);
				}

				ast_cdr_setvar(ast_channel_name(attended->transferer_second->channel),
					"app_caller", ast_channel_caller(attended->transferer_second->channel)->ani.number.str);

				/*Added by yfyang, for Bug 335592 */
				if (!strncmp(ast_channel_name(bridged_1), "PJSIP/trunk_", 12)) {
					char tmp_dstchan[128] = {0};
					ast_cdr_getvar(ast_channel_name(attended->transferer_second->channel), "dstchannel", tmp_dstchan, sizeof(tmp_dstchan));
					if (strlen(tmp_dstchan)) {
						ast_log(LOG_NOTICE, "the second dstchannel is %s", tmp_dstchan);
						struct ast_channel *dst_channel = ast_channel_get_by_name(tmp_dstchan);
						if (dst_channel) {
							const char *real_out_callee = pbx_builtin_getvar_helper(bridged_1, "THE_REAL_OUT_CALLEE");
							if (!ast_strlen_zero(real_out_callee)) {
								ast_channel_reg_calleenum_set(dst_channel, real_out_callee);
								pbx_builtin_setvar_helper(dst_channel, "SIP_TRANSFER", "yes");
							}

							ast_channel_cleanup(dst_channel);
						}
					}
				}
			}
		}
		/* end add by zgmo*/

		if (bridged_1 && bridged_2) {
			exchange_peer_remote_sdp(bridged_1, bridged_2);
		}

		if (bridged_1) {
			ast_log(LOG_NOTICE, "bridge_1 = %s\n", ast_channel_name(bridged_1));
			if (ast_channel_tech(bridged_1) && (0 == strcmp(ast_channel_tech(bridged_1)->type, "PJSIP"))) {
				struct ast_sip_channel_pvt *channel_pvt = ast_channel_tech_pvt(bridged_1);
				if (channel_pvt && channel_pvt->session) {
					if (channel_pvt->session->is_direct_media) {
						/* modified by xxcui for Bug 76639, attended transfer under direct media, one channel is bridged and the other isn't.
						Refresh the bridged transferee first */
						if (bridged_2) {
							if (channel_pvt->session
								&& ast_sip_push_task(channel_pvt->session->serializer, refer_refresh_without_new_sdp, ao2_bump(channel_pvt->session))) {
								ao2_cleanup(channel_pvt->session);
							}
						} else {
							channel_pvt->session->is_direct_media = 0;
							ast_sip_session_init_stream_media_direction(channel_pvt->session);
							if (channel_pvt->session
								&& ast_sip_push_task(channel_pvt->session->serializer, refer_refresh, ao2_bump(channel_pvt->session))) {
								ao2_cleanup(channel_pvt->session);
							}
						}
					} else {
						/* Modify by bfdu, just send re-invite(unhold) when send hold before, 2018.06.07 */
						if (allow_forward_hold_request) {
							ast_sip_session_set_all_stream_change_ssrc(channel_pvt->session);
							ast_sip_session_init_stream_media_direction(channel_pvt->session);
							if (bridged_1) {
								ast_channel_set_feature_transfer(bridged_1, 1);
							}
							if (channel_pvt->session
								&& ast_sip_push_task(channel_pvt->session->serializer, refer_refresh, ao2_bump(channel_pvt->session))) {
								ao2_cleanup(channel_pvt->session);
							}
						}
					}
				}
			}
		}
		if (bridged_2){
			ast_log(LOG_NOTICE, "bridge_2 = %s\n", ast_channel_name(bridged_2));
			if (ast_channel_tech(bridged_2) && (0 == strcmp(ast_channel_tech(bridged_2)->type, "PJSIP"))) {
				struct ast_sip_channel_pvt *channel_pvt_peer = ast_channel_tech_pvt(bridged_2);
				if (channel_pvt_peer && channel_pvt_peer->session) {
					if (channel_pvt_peer->session->is_direct_media) {
						if (bridged_1) {
							if (channel_pvt_peer->session
								&& ast_sip_push_task(channel_pvt_peer->session->serializer, refer_refresh_without_new_sdp, ao2_bump(channel_pvt_peer->session))) {
								ao2_cleanup(channel_pvt_peer->session);
							}
						} else {
							channel_pvt_peer->session->is_direct_media = 0;
							ast_sip_session_init_stream_media_direction(channel_pvt_peer->session);
							if (channel_pvt_peer->session
								&& ast_sip_push_task(channel_pvt_peer->session->serializer, refer_refresh, ao2_bump(channel_pvt_peer->session))) {
								ao2_cleanup(channel_pvt_peer->session);
							}
						}
					} else {
						/* Modify by bfdu, just send re-invite(unhold) when send hold before, 2018.06.07 */
						if (allow_forward_hold_request) {
							ast_sip_session_set_all_stream_change_ssrc(channel_pvt_peer->session);
							ast_sip_session_init_stream_media_direction(channel_pvt_peer->session);
							if (channel_pvt_peer->session
								&& ast_sip_push_task(channel_pvt_peer->session->serializer, refer_refresh, ao2_bump(channel_pvt_peer->session))) {
								ao2_cleanup(channel_pvt_peer->session);
							}
						}
					}
				}
			}
		}

		update_info_for_refer_attend(attended, response, bridged_2);
		notify_trunk_info_for_refer(bridged_1, bridged_2);
#endif
	} else {
		ast_debug(3, "Received REFER request on channel '%s' but other channel has gone.\n",
			ast_channel_name(attended->transferer_chan));
		response = 603;
	}

	if (attended->progress) {
		struct refer_progress_notification *notification;

		notification = refer_progress_notification_alloc(attended->progress, response,
			PJSIP_EVSUB_STATE_TERMINATED);
		if (notification) {
			refer_progress_notify(notification);
		}
	}

	if (response == 200) {
		task_cb = session_end_if_deferred_task;
	} else {
		task_cb = defer_termination_cancel_task;
	}
	if (!ast_sip_push_task(attended->transferer->serializer,
		task_cb, attended->transferer)) {
		/* Gave the ref to the pushed task. */
		attended->transferer = NULL;
	} else {
		/* Do this anyway even though it is the wrong serializer. */
		ast_sip_session_end_if_deferred(attended->transferer);
	}

	ao2_ref(attended, -1);
	return 0;
}

/*! \brief Structure for blind transfer callback details */
struct refer_blind {
	/*! \brief Context being used for transfer */
	const char *context;
	/*! \brief Optional progress structure */
	struct refer_progress *progress;
	/*! \brief REFER message */
	pjsip_rx_data *rdata;
	/*! \brief Optional Replaces header */
	pjsip_replaces_hdr *replaces;
	/*! \brief Optional Refer-To header */
	pjsip_sip_uri *refer_to;
	/*! \brief Attended transfer flag */
	unsigned int attended:1;
};

/*! \brief Blind transfer callback function */
static void refer_blind_callback(struct ast_channel *chan, struct transfer_channel_data *user_data_wrapper,
	enum ast_transfer_type transfer_type)
{
	struct refer_blind *refer = user_data_wrapper->data;
	pjsip_generic_string_hdr *referred_by;

	static const pj_str_t str_referred_by = { "Referred-By", 11 };
	static const pj_str_t str_referred_by_s = { "b", 1 };

	pbx_builtin_setvar_helper(chan, "SIPTRANSFER", "yes");

	if (refer->progress && !refer->attended && !refer->progress->refer_blind_progress) {
		/* If blind transfer and endpoint doesn't want to receive all the progress details */
		struct refer_progress_notification *notification = refer_progress_notification_alloc(refer->progress, 200,
			PJSIP_EVSUB_STATE_TERMINATED);

		if (notification) {
			refer_progress_notify(notification);
		}
	} else if (refer->progress) {
		/* If attended transfer and progress monitoring is being done attach a frame hook so we can monitor it */
		struct ast_framehook_interface hook = {
			.version = AST_FRAMEHOOK_INTERFACE_VERSION,
			.event_cb = refer_progress_framehook,
			.destroy_cb = refer_progress_framehook_destroy,
			.data = refer->progress,
			.disable_inheritance = 1,
		};

		refer->progress->transferee = ast_strdup(ast_channel_uniqueid(chan));
#ifdef GRANDSTREAM_NETWORKS
		user_data_wrapper->transferee = ast_strdup(ast_channel_uniqueid(chan));
#endif
		if (!refer->progress->transferee) {
			struct refer_progress_notification *notification = refer_progress_notification_alloc(refer->progress, 200,
				PJSIP_EVSUB_STATE_TERMINATED);

			ast_log(LOG_WARNING, "Could not copy channel name '%s' during transfer - assuming success\n",
				ast_channel_name(chan));

			if (notification) {
				refer_progress_notify(notification);
			}
		}

		/* Progress needs a reference to the transfer_channel_data so that it can track the completed status of the transfer */
		ao2_ref(user_data_wrapper, +1);
		refer->progress->transfer_data = user_data_wrapper;

		/* We need to bump the reference count up on the progress structure since it is in the frame hook now */
		ao2_ref(refer->progress, +1);

		/* If we can't attach a frame hook for whatever reason send a notification of success immediately */
		ast_channel_lock(chan);
		refer->progress->framehook = ast_framehook_attach(chan, &hook);
		ast_channel_unlock(chan);
		if (refer->progress->framehook < 0) {
			struct refer_progress_notification *notification = refer_progress_notification_alloc(refer->progress, 200,
				PJSIP_EVSUB_STATE_TERMINATED);

			ast_log(LOG_WARNING, "Could not attach REFER transfer progress monitoring hook to channel '%s' - assuming success\n",
				ast_channel_name(chan));

			if (notification) {
				refer_progress_notify(notification);
			}

			ao2_cleanup(refer->progress);
		}

		/* We need to bump the reference count for the stasis subscription */
		ao2_ref(refer->progress, +1);
		/* We also will need to detect if the transferee enters a bridge. This is currently the only reliable way to
		 * detect if the transfer target has answered the call
		 */
#ifdef GRANDSTREAM_NETWORKS
		/** added by jklou for bug 54856,55705*/
		if (chan && ast_channel_tech(chan) && (0 == strcmp(ast_channel_tech(chan)->type, "PJSIP"))) {
			struct ast_sip_channel_pvt *channel_pvt = ast_channel_tech_pvt(chan);
			if (channel_pvt && channel_pvt->session) {
				channel_pvt->session->is_direct_media = 0;
				/* add judgement for Bug 68242 */
				ast_sip_session_init_stream_media_direction(channel_pvt->session);
				if (judge_if_need_send_reinvite(channel_pvt->session) || channel_pvt->session->is_webrtc_call) {
					ast_sip_session_set_all_stream_change_ssrc(channel_pvt->session);
					if (channel_pvt->session
						&& ast_sip_push_task(channel_pvt->session->serializer, refer_refresh, ao2_bump(channel_pvt->session))) {
						ao2_cleanup(channel_pvt->session);
					}
				} else {
					ast_log(LOG_DEBUG, "Don't need send reinvite for refer\n");
				}
			}
		} else if (chan && ast_channel_tech(chan) && (0 == strcmp(ast_channel_tech(chan)->type, "Local"))) {
			struct ast_channel* peer_local_chan = ast_channel_get_mychan(chan);
			if (peer_local_chan) {
				ast_queue_control_data(peer_local_chan, AST_CONTROL_BLIND_REFER_UNHOLD, NULL, 0);
			} else {
				ast_log(LOG_WARNING, "Can't get peer local channel: %s\n", ast_channel_name(chan));
			}
		}

		pbx_builtin_setvar_helper(chan, "__TRANSPARENTSDP", "no");
#endif
		refer->progress->bridge_sub = stasis_subscribe_pool(ast_bridge_topic_all(), refer_progress_bridge, refer->progress);
		if (!refer->progress->bridge_sub) {
			struct refer_progress_notification *notification = refer_progress_notification_alloc(refer->progress, 200,
				PJSIP_EVSUB_STATE_TERMINATED);

			ast_log(LOG_WARNING, "Could not create bridge stasis subscription for monitoring progress on transfer of channel '%s' - assuming success\n",
					ast_channel_name(chan));

			if (notification) {
				refer_progress_notify(notification);
			}

			ast_channel_lock(chan);
			ast_framehook_detach(chan, refer->progress->framehook);
			ast_channel_unlock(chan);

			ao2_cleanup(refer->progress);
		} else {
#ifdef GRANDSTREAM_NETWORKS
			/*Add for Bug223248*/
			ao2_lock(refer->progress);
			RAII_VAR(struct stasis_subscription *, bridge_sub, ao2_bump(refer->progress->bridge_sub), ao2_cleanup);
			ao2_unlock(refer->progress);
			if (bridge_sub) {
				stasis_subscription_accept_message_type(bridge_sub, ast_channel_entered_bridge_type());
				stasis_subscription_accept_message_type(bridge_sub, stasis_subscription_change_type());
				stasis_subscription_set_filter(bridge_sub, STASIS_SUBSCRIPTION_FILTER_SELECTIVE);
			}
#else
			stasis_subscription_accept_message_type(refer->progress->bridge_sub, ast_channel_entered_bridge_type());
			stasis_subscription_accept_message_type(refer->progress->bridge_sub, stasis_subscription_change_type());
			stasis_subscription_set_filter(refer->progress->bridge_sub, STASIS_SUBSCRIPTION_FILTER_SELECTIVE);
#endif
		}
	}

	pbx_builtin_setvar_helper(chan, "SIPREFERRINGCONTEXT", S_OR(refer->context, NULL));

	referred_by = pjsip_msg_find_hdr_by_names(refer->rdata->msg_info.msg,
		&str_referred_by, &str_referred_by_s, NULL);
	if (referred_by) {
		size_t uri_size = pj_strlen(&referred_by->hvalue) + 1;
		char *uri = ast_alloca(uri_size);

		ast_copy_pj_str(uri, &referred_by->hvalue, uri_size);
		pbx_builtin_setvar_helper(chan, "__SIPREFERREDBYHDR", S_OR(uri, NULL));
	} else {
		pbx_builtin_setvar_helper(chan, "SIPREFERREDBYHDR", NULL);
	}

	if (refer->replaces) {
		char replaces[512];
		char *replaces_val = NULL;
		int len;

		len = pjsip_hdr_print_on(refer->replaces, replaces, sizeof(replaces) - 1);
		if (len != -1) {
			/* pjsip_hdr_print_on does not NULL terminate the buffer */
			replaces[len] = '\0';
			replaces_val = replaces + sizeof("Replaces:");
		}
		pbx_builtin_setvar_helper(chan, "__SIPREPLACESHDR", replaces_val);
	} else {
		pbx_builtin_setvar_helper(chan, "SIPREPLACESHDR", NULL);
	}

	if (refer->refer_to) {
		char refer_to[PJSIP_MAX_URL_SIZE];

		pjsip_uri_print(PJSIP_URI_IN_REQ_URI, refer->refer_to, refer_to, sizeof(refer_to));
		pbx_builtin_setvar_helper(chan, "SIPREFERTOHDR", S_OR(refer_to, NULL));
	} else {
		pbx_builtin_setvar_helper(chan, "SIPREFERTOHDR", NULL);
	}
}

/*!
 * \internal
 * \brief Set the passed in context variable to the determined transfer context.
 * \since 13.3.0
 *
 * \param context Set to the determined transfer context.
 * \param session INVITE dialog SIP session.
 */
#define DETERMINE_TRANSFER_CONTEXT(context, session)									\
	do {																				\
		ast_channel_lock((session)->channel);											\
		context = pbx_builtin_getvar_helper((session)->channel, "TRANSFER_CONTEXT");	\
		if (ast_strlen_zero(context)) {													\
			context = (session)->endpoint->context;										\
		} else {																		\
			context = ast_strdupa(context);												\
		}																				\
		ast_channel_unlock((session)->channel);											\
	} while (0)																			\

static int refer_incoming_attended_request(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_sip_uri *target_uri,
	pjsip_param *replaces_param, struct refer_progress *progress)
{
	const pj_str_t str_replaces = { "Replaces", 8 };
	pj_str_t replaces_content;
	pjsip_replaces_hdr *replaces;
	int parsed_len;
	pjsip_dialog *dlg;

	pj_strdup_with_null(rdata->tp_info.pool, &replaces_content, &replaces_param->value);

	/* Parsing the parameter as a Replaces header easily grabs the needed information */
	if (!(replaces = pjsip_parse_hdr(rdata->tp_info.pool, &str_replaces, replaces_content.ptr,
		pj_strlen(&replaces_content), &parsed_len))) {
		ast_log(LOG_ERROR, "Received REFER request on channel '%s' from endpoint '%s' with invalid Replaces header, rejecting\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
		return 400;
	}

	/* See if the dialog is local, or remote */
	if ((dlg = pjsip_ua_find_dialog(&replaces->call_id, &replaces->to_tag, &replaces->from_tag, PJ_TRUE))) {
		RAII_VAR(struct ast_sip_session *, other_session, ast_sip_dialog_get_session(dlg), ao2_cleanup);
		struct refer_attended *attended;

		pjsip_dlg_dec_lock(dlg);

		if (!other_session) {
			ast_debug(3, "Received REFER request on channel '%s' from endpoint '%s' for local dialog but no session exists on it\n",
				ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
			return 603;
		}

		/* We defer actually doing the attended transfer to the other session so no deadlock can occur */
		if (!(attended = refer_attended_alloc(session, other_session, progress))) {
			ast_log(LOG_ERROR, "Received REFER request on channel '%s' from endpoint '%s' for local dialog but could not allocate structure to complete, rejecting\n",
				ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
			return 500;
		}

		if (ast_sip_session_defer_termination(session)) {
			ast_log(LOG_ERROR, "Received REFER request on channel '%s' from endpoint '%s' for local dialog but could not defer termination, rejecting\n",
				ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
			ao2_cleanup(attended);
			return 500;
		}

#ifdef GRANDSTREAM_NETWORKS 
		RAII_VAR(struct ast_channel*, caller_channel, ast_channel_bridge_peer(attended->transferer_chan), ast_channel_cleanup);
		RAII_VAR(struct ast_channel*, callee_channel, ast_channel_bridge_peer(attended->transferer_second->channel), ast_channel_cleanup);

		if (caller_channel && callee_channel
			&& (0 == strcmp(ast_channel_tech(caller_channel)->type, "PJSIP"))
			&& (0 == strcmp(ast_channel_tech(callee_channel)->type, "PJSIP"))) {

			struct ast_sip_channel_pvt* caller_channel_pvt = NULL;
			struct ast_sip_channel_pvt* callee_channel_pvt = NULL;
			struct ast_sip_session* caller_session = NULL;
			struct ast_sip_session* callee_session = NULL;
			struct ast_sip_session* middle_session = other_session;

			if (caller_channel) {
				caller_channel_pvt = ast_channel_tech_pvt(caller_channel);
				if (caller_channel_pvt) {
					caller_session = caller_channel_pvt->session;
				}
				ast_cdr_setvar(ast_channel_name(caller_channel),"special_action_type", "atx_refer");
			}

			if (callee_channel) {
				callee_channel_pvt = ast_channel_tech_pvt(callee_channel);
				if (callee_channel_pvt) {
					callee_session = callee_channel_pvt->session;
				}
			}

			if (caller_session && callee_session 
				&& caller_session->endpoint->media.direct_media.enabled
				&& middle_session->endpoint->media.direct_media.enabled
				&& !callee_session->endpoint->media.direct_media.enabled) {
				struct ast_bridge* second_bridge = ast_channel_get_bridge(middle_session->channel);
				if (second_bridge) {
					ast_bridge_destroy(second_bridge, 0);
					second_bridge = NULL;
				}
			
				ast_sip_session_defer_termination_cancel(session);
				ao2_cleanup(attended);
				ast_log(LOG_NOTICE, "The refer is not support this case, please change extension direct media option\n");
				return 603;
			}

			/* BEGIN: Added by yfyang, for Bug 155444 */
			if (!strncmp(ast_channel_name(caller_channel), "PJSIP/trunk_", 12)) {
				const char *reg_calleenum = ast_channel_reg_calleenum(caller_channel);
				if (!ast_strlen_zero(reg_calleenum)) {
					ast_log(LOG_NOTICE, "set reg_calleenum[%s] to chan[%s]", reg_calleenum, ast_channel_name(callee_channel));
					ast_channel_reg_calleenum_set(callee_channel, reg_calleenum);
				}
				pbx_builtin_setvar_helper(callee_channel, "__SIP_TRANSFER", "rf_attended");
			}
			/* END: Added by yfyang, for Bug 155444 */
		}
#endif

		/* Push it to the other session, which will have both channels with minimal locking */
		if (ast_sip_push_task(other_session->serializer, refer_attended_task, attended)) {
			ast_sip_session_end_if_deferred(session);
			ast_sip_session_defer_termination_cancel(session);
			ao2_cleanup(attended);
			return 500;
		}

		ast_debug(3, "Attended transfer from '%s' pushed to second channel serializer\n",
			ast_channel_name(session->channel));

		return 200;
	} else {
		const char *context;
		struct refer_blind refer = { 0, };
		int response;

		DETERMINE_TRANSFER_CONTEXT(context, session);

		if (!ast_exists_extension(NULL, context, "external_replaces", 1, NULL)) {
			ast_log(LOG_ERROR, "Received REFER for remote session on channel '%s' from endpoint '%s' but 'external_replaces' extension not found in context %s\n",
				ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint), context);
			return 404;
		}

		refer.context = context;
		refer.progress = progress;
		refer.rdata = rdata;
		refer.replaces = replaces;
		refer.refer_to = target_uri;
		refer.attended = 1;

		if (ast_sip_session_defer_termination(session)) {
			ast_log(LOG_ERROR, "Received REFER for remote session on channel '%s' from endpoint '%s' but could not defer termination, rejecting\n",
				ast_channel_name(session->channel),
				ast_sorcery_object_get_id(session->endpoint));
			return 500;
		}

		response = xfer_response_code2sip(ast_bridge_transfer_blind(1, session->channel,
			"external_replaces", context, refer_blind_callback, &refer));

		ast_sip_session_end_if_deferred(session);
		if (response != 200) {
			ast_sip_session_defer_termination_cancel(session);
		}

		return response;
	}
}

static int refer_incoming_blind_request(struct ast_sip_session *session, pjsip_rx_data *rdata, pjsip_sip_uri *target,
	struct refer_progress *progress)
{
	const char *context;
	char exten[AST_MAX_EXTENSION];
	struct refer_blind refer = { 0, };
	int response;

	/* If no explicit transfer context has been provided use their configured context */
	DETERMINE_TRANSFER_CONTEXT(context, session);

	/* Using the user portion of the target URI see if it exists as a valid extension in their context */
	ast_copy_pj_str(exten, &target->user, sizeof(exten));

	/*
	 * We may want to match in the dialplan without any user
	 * options getting in the way.
	 */
	AST_SIP_USER_OPTIONS_TRUNCATE_CHECK(exten);

	/* Uri without exten */
	if (ast_strlen_zero(exten)) {
		ast_copy_string(exten, "s", sizeof(exten));
		ast_debug(3, "Channel '%s' from endpoint '%s' attempted blind transfer to a target without extension. Target was set to 's@%s'\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint), context);
	}

	if (!ast_exists_extension(NULL, context, exten, 1, NULL)) {
		ast_log(LOG_ERROR, "Channel '%s' from endpoint '%s' attempted blind transfer to '%s@%s' but target does not exist\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint), exten, context);
		return 404;
	}

#ifdef GRANDSTREAM_NETWORKS
	else if (ast_findapp_extension(NULL, NULL, context, exten, 1, NULL, NULL, E_MATCH, "ParkedCall"))
	{
		ast_log(LOG_ERROR, "ParkedCall");
		return 486;
	}

#endif

	refer.context = context;
	refer.progress = progress;
	refer.rdata = rdata;
	refer.refer_to = target;
	refer.attended = 0;

	if (ast_sip_session_defer_termination(session)) {
		ast_log(LOG_ERROR, "Channel '%s' from endpoint '%s' attempted blind transfer but could not defer termination, rejecting\n",
			ast_channel_name(session->channel),
			ast_sorcery_object_get_id(session->endpoint));
		return 500;
	}

#ifdef GRANDSTREAM_NETWORKS
	pbx_builtin_setvar_helper(session->channel, "REFER_BLINDTRANSFER", "yes");
#endif
	response = xfer_response_code2sip(ast_bridge_transfer_blind(1, session->channel,
		exten, context, refer_blind_callback, &refer));

	ast_sip_session_end_if_deferred(session);
	if (response != 200) {
		ast_sip_session_defer_termination_cancel(session);
	}

	return response;
}

/*! \brief Structure used to retrieve channel from another session */
struct invite_replaces {
	/*! \brief Session we want the channel from */
	struct ast_sip_session *session;
	/*! \brief Channel from the session (with reference) */
	struct ast_channel *channel;
	/*! \brief Bridge the channel is in */
	struct ast_bridge *bridge;
};

/*! \brief Task for invite replaces */
static int invite_replaces(void *data)
{
	struct invite_replaces *invite = data;

	if (!invite->session->channel) {
		return -1;
	}

	ast_channel_ref(invite->session->channel);
	invite->channel = invite->session->channel;

	invite->bridge = ast_bridge_transfer_acquire_bridge(invite->channel);
	return 0;
}

static int refer_incoming_invite_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_dialog *other_dlg = NULL;
	pjsip_tx_data *packet;
	int response = 0;
	RAII_VAR(struct ast_sip_session *, other_session, NULL, ao2_cleanup);
	struct invite_replaces invite;

	/* If a Replaces header is present make sure it is valid */
	if (pjsip_replaces_verify_request(rdata, &other_dlg, PJ_TRUE, &packet) != PJ_SUCCESS) {
		response = packet->msg->line.status.code;
		ast_assert(response != 0);
		pjsip_tx_data_dec_ref(packet);
		goto inv_replace_failed;
	}

	/* If no other dialog exists then this INVITE request does not have a Replaces header */
	if (!other_dlg) {
		return 0;
	}

	other_session = ast_sip_dialog_get_session(other_dlg);
	pjsip_dlg_dec_lock(other_dlg);

	/* Don't accept an in-dialog INVITE with Replaces as it does not make much sense */
	if (session->inv_session->dlg->state == PJSIP_DIALOG_STATE_ESTABLISHED) {
		response = 488;
		goto inv_replace_failed;
	}

	if (!other_session) {
		ast_debug(3, "INVITE with Replaces received on channel '%s' from endpoint '%s', but requested session does not exist\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
		response = 481;
		goto inv_replace_failed;
	}

	invite.session = other_session;

	if (ast_sip_push_task_wait_serializer(other_session->serializer, invite_replaces,
		&invite)) {
		response = 481;
		goto inv_replace_failed;
	}

	ast_channel_lock(session->channel);
	ast_setstate(session->channel, AST_STATE_RING);
	ast_channel_unlock(session->channel);
#ifdef GRANDSTREAM_NETWORKS
	ast_raw_answer(session->channel, NULL);
#else
	ast_raw_answer(session->channel);
#endif

#ifdef GRANDSTREAM_NETWORKS

	struct ast_channel *peer_chan = ast_channel_bridge_peer(invite.channel);
#endif

	ast_debug(3, "INVITE with Replaces being attempted.  '%s' --> '%s'\n",
		ast_channel_name(session->channel), ast_channel_name(invite.channel));

	if (!invite.bridge) {
		struct ast_channel *chan = session->channel;

		/*
		 * This will use a synchronous task but we aren't operating in
		 * the serializer at this point in time, so it won't deadlock.
		 */
		if (!ast_channel_move(invite.channel, chan)) {
			/*
			 * We can't directly use session->channel because ast_channel_move()
			 * does a masquerade which changes session->channel to a different
			 * channel.  To ensure we work on the right channel we store a
			 * pointer locally before we begin so it remains valid.
			 */
			ast_hangup(chan);
		} else {
			response = AST_CAUSE_FAILURE;
		}
	} else {
		if (ast_bridge_impart(invite.bridge, session->channel, invite.channel, NULL,
			AST_BRIDGE_IMPART_CHAN_INDEPENDENT)) {
			response = AST_CAUSE_FAILURE;
		}
	}

	ast_channel_unref(invite.channel);
	ao2_cleanup(invite.bridge);

	if (!response) {
		/*
		 * On success we cannot use session->channel in the debug message.
		 * This thread either no longer has a ref to session->channel or
		 * session->channel is no longer the original channel.
		 */

#ifdef GRANDSTREAM_NETWORKS

		ast_debug(3, "INVITE with Replaces successfully completed.\n");
		if (peer_chan) {
			unsigned int allow_forward_hold_request = ast_sip_session_get_allow_forward_hold_request();
			ast_log(LOG_NOTICE, "peer_chan = %s\n", ast_channel_name(peer_chan));
			if (ast_channel_tech(peer_chan) && (0 == strcmp(ast_channel_tech(peer_chan)->type, "PJSIP"))) {
				struct ast_sip_channel_pvt *channel_pvt = ast_channel_tech_pvt(peer_chan);
				if (channel_pvt && channel_pvt->session) {
					if (channel_pvt->session->is_direct_media) {
						channel_pvt->session->is_direct_media = 0;
						ast_sip_session_init_stream_media_direction(channel_pvt->session);
						if (channel_pvt->session
							&& ast_sip_push_task(channel_pvt->session->serializer, refer_refresh, ao2_bump(channel_pvt->session))) {
							ao2_cleanup(channel_pvt->session);
						}
					} else {
						if (allow_forward_hold_request) {
							ast_sip_session_set_all_stream_change_ssrc(channel_pvt->session);
							ast_sip_session_init_stream_media_direction(channel_pvt->session);
							if (channel_pvt->session
								&& ast_sip_push_task(channel_pvt->session->serializer, refer_refresh, ao2_bump(channel_pvt->session))) {
								ao2_cleanup(channel_pvt->session);
							}
						}
					}
				}
			}
			ast_channel_unref(peer_chan);
		}
#endif
	} else {
		ast_debug(3, "INVITE with Replaces failed on channel '%s', hanging up with cause '%d'\n",
			ast_channel_name(session->channel), response);
		ast_channel_lock(session->channel);
		ast_channel_hangupcause_set(session->channel, response);
		ast_channel_unlock(session->channel);
		ast_hangup(session->channel);
	}

	return 1;

inv_replace_failed:
	if (session->inv_session->dlg->state != PJSIP_DIALOG_STATE_ESTABLISHED) {
		ast_debug(3, "INVITE with Replaces failed on channel '%s', sending response of '%d'\n",
			ast_channel_name(session->channel), response);
		session->defer_terminate = 1;
		ast_hangup(session->channel);

		if (pjsip_inv_end_session(session->inv_session, response, NULL, &packet) == PJ_SUCCESS
			&& packet) {
			ast_sip_session_send_response(session, packet);
		}
	} else {
		ast_debug(3, "INVITE with Replaces in-dialog on channel '%s', hanging up\n",
			ast_channel_name(session->channel));
		ast_queue_hangup(session->channel);
	}

	return 1;
}

#ifdef GRANDSTREAM_NETWORKS
static void tel_to_sip_uri(pj_pool_t *pool, pjsip_tel_uri *tel_uri, pjsip_sip_uri *sip_uri)
{
    static pj_str_t tel_to_sip_param_str;
    pjsip_param *sip_param;

    pj_strdup( pool, &sip_uri->user, &tel_uri->number);
    pjsip_param_clone(pool, &sip_uri->other_param, &tel_uri->other_param);

    pj_strdup2(pool, &tel_to_sip_param_str, "user");
    sip_param = pjsip_param_find(&tel_uri->other_param, &tel_to_sip_param_str);
    if(sip_param){
       pj_strdup( pool, &sip_uri->user_param, &sip_param->value);
    }

    pj_strdup2(pool, &tel_to_sip_param_str, "method");
    sip_param = pjsip_param_find(&tel_uri->other_param, &tel_to_sip_param_str);
    if(sip_param){
       pj_strdup( pool, &sip_uri->method_param, &sip_param->value);
    }

    pj_strdup2(pool, &tel_to_sip_param_str, "transport");
    sip_param = pjsip_param_find(&tel_uri->other_param, &tel_to_sip_param_str);
    if(sip_param){
       pj_strdup( pool, &sip_uri->transport_param, &sip_param->value);
    }

    pj_strdup2(pool, &tel_to_sip_param_str, "lr");
    sip_param = pjsip_param_find(&tel_uri->other_param, &tel_to_sip_param_str);
    if(sip_param){
       sip_uri->lr_param = pj_strtoul(&sip_param->value);
    }

    pj_strdup2(pool, &tel_to_sip_param_str, "ttl");
    sip_param = pjsip_param_find(&tel_uri->other_param, &tel_to_sip_param_str);
    if(sip_param){
       sip_uri->ttl_param = pj_strtoul(&sip_param->value);
    }

    pj_strdup2(pool, &tel_to_sip_param_str, "maddr");
    sip_param = pjsip_param_find(&tel_uri->other_param, &tel_to_sip_param_str);
    if(sip_param){
       pj_strdup( pool, &sip_uri->maddr_param, &sip_param->value);
    }
}

static unsigned int target_is_conference(pjsip_sip_uri *target_uri)
{
	char target_user[128] = { 0 };

	if (!target_uri) {
		return 0;
	}

	ast_copy_pj_str(target_user, &target_uri->user, sizeof(target_user));

	ast_log(LOG_DEBUG, "target_is_conference target_user:%s\n", target_user);

	if (ast_sip_check_is_mcm_conf_number(target_user) ||
		ast_sip_session_is_nway_conference_number(target_user)) {
		ast_log(LOG_DEBUG, "target_is_conference target:%s is a conference\n", target_user);
		return 1;
	}

	return 0;
}
#endif

static int refer_incoming_refer_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *refer_to;
	char *uri;
	size_t uri_size;
	pjsip_uri *target;
	pjsip_sip_uri *target_uri;
#ifdef GRANDSTREAM_NETWORKS
    pjsip_tel_uri *tel_target_uri;
    pjsip_sip_uri *tel_uri_to_sip_uri;
    pjsip_host_info host_info;
#endif
	RAII_VAR(struct refer_progress *, progress, NULL, ao2_cleanup);
	pjsip_param *replaces;
	int response;

	static const pj_str_t str_refer_to = { "Refer-To", 8 };
	static const pj_str_t str_refer_to_s = { "r", 1 };
	static const pj_str_t str_replaces = { "Replaces", 8 };

	if (!session->channel) {
		/* No channel to refer.  Likely because the call was just hung up. */
		pjsip_dlg_respond(session->inv_session->dlg, rdata, 404, NULL, NULL, NULL);
		ast_debug(3, "Received a REFER on a session with no channel from endpoint '%s'.\n",
			ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!session->endpoint->allowtransfer) {
		pjsip_dlg_respond(session->inv_session->dlg, rdata, 603, NULL, NULL, NULL);
		ast_log(LOG_WARNING, "Endpoint %s transfer attempt blocked due to configuration\n",
				ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 133005*/
	const char *conf = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
	const char *paging = pbx_builtin_getvar_helper(session->channel, "PAGING_ALERT_INFO");
	if (!ast_strlen_zero(conf) || (!ast_strlen_zero(paging) && !strcmp(paging, "yes"))) {
		pjsip_dlg_respond(session->inv_session->dlg, rdata, 603, NULL, NULL, NULL);
		ast_log(LOG_WARNING, "This transfer from conference or paging, forbidden!");
		return 0;
	}
#endif

	/* A Refer-To header is required */
	refer_to = pjsip_msg_find_hdr_by_names(rdata->msg_info.msg, &str_refer_to, &str_refer_to_s, NULL);
	if (!refer_to) {
		pjsip_dlg_respond(session->inv_session->dlg, rdata, 400, NULL, NULL, NULL);
		ast_debug(3, "Received a REFER without Refer-To on channel '%s' from endpoint '%s'\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/* The ast_copy_pj_str to uri is needed because it puts the NULL terminator to the uri
	 * as pjsip_parse_uri require a NULL terminated uri
	 */

	uri_size = pj_strlen(&refer_to->hvalue) + 1;
	uri = ast_alloca(uri_size);
	ast_copy_pj_str(uri, &refer_to->hvalue, uri_size);

	target = pjsip_parse_uri(rdata->tp_info.pool, uri, uri_size - 1, 0);
#ifndef GRANDSTREAM_NETWORKS
	if (!target
		|| (!PJSIP_URI_SCHEME_IS_SIP(target)
			&& !PJSIP_URI_SCHEME_IS_SIPS(target))) 
#else
    	if (!target
		|| (!PJSIP_URI_SCHEME_IS_SIP(target)
			&& !PJSIP_URI_SCHEME_IS_SIPS(target)
			&& !PJSIP_URI_SCHEME_IS_TEL(target))) 
#endif
	{

		pjsip_dlg_respond(session->inv_session->dlg, rdata, 400, NULL, NULL, NULL);
		ast_debug(3, "Received a REFER without a parseable Refer-To ('%s') on channel '%s' from endpoint '%s'\n",
			uri, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}
	target_uri = pjsip_uri_get_uri(target);
#ifdef GRANDSTREAM_NETWORKS
    if(PJSIP_URI_SCHEME_IS_TEL(target))
    {
       tel_target_uri = (pjsip_tel_uri *)pjsip_uri_get_uri(target);
       tel_uri_to_sip_uri = pjsip_sip_uri_create(rdata->tp_info.pool, PJ_FALSE);

       if(pjsip_get_dest_info(session->inv_session->dlg->target, NULL, rdata->tp_info.pool,&host_info)== PJ_SUCCESS){
          pj_strdup( rdata->tp_info.pool, &tel_uri_to_sip_uri->host, &host_info.addr.host);
       }

       if(tel_uri_to_sip_uri){
          tel_to_sip_uri(rdata->tp_info.pool, tel_target_uri, tel_uri_to_sip_uri);
          target_uri = tel_uri_to_sip_uri;
    	}  
    }

	/* Check SCA transfer, Added by yfyang, for Bug 153263 */
	if ((replaces = pjsip_param_find(&target_uri->header_param, &str_replaces)) ||
		(replaces = pjsip_param_find(&target_uri->other_param, &str_replaces))) {
		/* Add by bfdu for forbid attened transfer to conference */
		if (target_is_conference(target_uri)) {
			pjsip_dlg_respond(session->inv_session->dlg, rdata, 603, NULL, NULL, NULL);
			ast_log(LOG_WARNING, "This transfer is to conference, forbidden!");
			return 0;
		}

		const char *sca_id = NULL;

		sca_id = pbx_builtin_getvar_helper(session->channel, "SCA_UNIQUEID");
		if (ast_strlen_zero(sca_id)) {
			const pj_str_t str_replaces = { "Replaces", 8 };
			pj_str_t replaces_content;
			pjsip_replaces_hdr *replaces_hdr;
			int parsed_len;
			pjsip_dialog *dlg;

			pj_strdup_with_null(rdata->tp_info.pool, &replaces_content, &replaces->value);

			/* Parsing the parameter as a Replaces header easily grabs the needed information */
			if (!(replaces_hdr = pjsip_parse_hdr(rdata->tp_info.pool, &str_replaces, replaces_content.ptr, pj_strlen(&replaces_content), &parsed_len))) {
				pjsip_dlg_respond(session->inv_session->dlg, rdata, 400, NULL, NULL, NULL);
				ast_log(LOG_WARNING, "This transfer with invalid replaces!");
				return 0;
			}

			/* See if the dialog is local, or remote */
			if ((dlg = pjsip_ua_find_dialog(&replaces_hdr->call_id, &replaces_hdr->to_tag, &replaces_hdr->from_tag, PJ_TRUE))) {
				RAII_VAR(struct ast_sip_session *, other_session, ast_sip_dialog_get_session(dlg), ao2_cleanup);

				pjsip_dlg_dec_lock(dlg);

				if (!other_session) {
					pjsip_dlg_respond(session->inv_session->dlg, rdata, 603, NULL, NULL, NULL);
					ast_log(LOG_WARNING, "This replaces can't find dlg!");
					return 0;
				}

				sca_id = pbx_builtin_getvar_helper(other_session->channel, "SCA_UNIQUEID");
			}
		}

		/* if SCA call, refuse the refer */
		if (!ast_strlen_zero(sca_id)) {
			pjsip_dlg_respond(session->inv_session->dlg, rdata, 603, NULL, NULL, NULL);
			ast_log(LOG_WARNING, "This transfer from sca call, forbidden!");
			return 0;
		}
	}
#endif

	/* Set up REFER progress subscription if requested/possible */
	if (refer_progress_alloc(session, rdata, &progress)) {
		pjsip_dlg_respond(session->inv_session->dlg, rdata, 500, NULL, NULL, NULL);
		ast_debug(3, "Could not set up subscription for REFER on channel '%s' from endpoint '%s'\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/* Determine if this is an attended or blind transfer */
	if ((replaces = pjsip_param_find(&target_uri->header_param, &str_replaces)) ||
		(replaces = pjsip_param_find(&target_uri->other_param, &str_replaces))) {
		response = refer_incoming_attended_request(session, rdata, target_uri, replaces, progress);
	} else {
		response = refer_incoming_blind_request(session, rdata, target_uri, progress);
	}

	if (!progress) {
		/* The transferer has requested no subscription, so send a final response immediately */
		pjsip_tx_data *tdata;
		const pj_str_t str_refer_sub = { "Refer-Sub", 9 };
		const pj_str_t str_false = { "false", 5 };
		pjsip_hdr *hdr;

		ast_debug(3, "Progress monitoring not requested for REFER on channel '%s' from endpoint '%s', sending immediate response of '%d'\n",
			ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint), response);

		if (pjsip_dlg_create_response(session->inv_session->dlg, rdata, response, NULL, &tdata) != PJ_SUCCESS) {
			pjsip_dlg_respond(session->inv_session->dlg, rdata, response, NULL, NULL, NULL);
			return 0;
		}

		hdr = (pjsip_hdr*)pjsip_generic_string_hdr_create(tdata->pool, &str_refer_sub, &str_false);
		pjsip_msg_add_hdr(tdata->msg, hdr);

		pjsip_dlg_send_response(session->inv_session->dlg, pjsip_rdata_get_tsx(rdata), tdata);
	} else if (response != 200) {
		/* Since this failed we can send a final NOTIFY now and terminate the subscription */
		struct refer_progress_notification *notification = refer_progress_notification_alloc(progress, response, PJSIP_EVSUB_STATE_TERMINATED);

		if (notification) {
			/* The refer_progress_notify function will call ao2_cleanup on this for us */
			refer_progress_notify(notification);
		}
	}

	return 0;
}

static int refer_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, pjsip_get_refer_method())) {
		return refer_incoming_refer_request(session, rdata);
	} else if (!pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_invite_method)) {
		return refer_incoming_invite_request(session, rdata);
	} else {
		return 0;
	}
}

#ifdef GRANDSTREAM_NETWORKS


/* Added by yfyang, for Bug 156466 */
static int get_new_referredby_val(pj_pool_t *pool, struct ast_channel *chan, const char *old_referredby, char *new_val, int len)
{
	static const pj_str_t from_name = { "From", 4 };
	static pjsip_fromto_hdr *pFrom;
	pjsip_name_addr *id_name_addr;
	pjsip_sip_uri   *sip_uri;
	char  transfer_user[80] = {0};
	char *old_val = NULL;
	char *rt_id   = NULL;
	const char *chan_name = NULL;
	char num[256]   = {0};
	char name[256]  = {0};
	char trunk_id[8]= {0};
	int  size  = 0;
	int  tk_id = 0;
	int  skip_name = 1;

	if (!pool || !chan || !old_referredby || !new_val || !len) {
		return -1;
	}

	chan_name = ast_channel_name(chan);
	if (strncmp(chan_name, "PJSIP/trunk_", 12)) {
		return -1;
	}

	tk_id = atoi(chan_name + strlen("PJSIP/trunk_"));
	if (tk_id == 0) {
		return -1;
	}

	old_val = ast_strdupa(old_referredby);
	if (ast_strlen_zero(old_val)) {
		return -1;
	}

	pFrom = pjsip_parse_hdr(pool, &from_name, old_val, strlen(old_val), &size);
	if (!pFrom) {
		return -1;
	}

	id_name_addr = pjsip_uri_clone(pool, pFrom->uri);
	sip_uri = pjsip_uri_get_uri(id_name_addr->uri);
	ast_copy_pj_str(transfer_user, &sip_uri->user, AST_MAX_EXTENSION);
	snprintf(trunk_id, sizeof(trunk_id), "%d", tk_id);

	ast_log(LOG_NOTICE, "get_new_referredby_val, transfer_user[%s], tk_id[%s]\n", transfer_user, trunk_id);

	rt_id = (char *)pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_INDEX");
	ast_sip_inner_cid_to_outer_cid(transfer_user, trunk_id, rt_id, num, sizeof(num)-1, name, sizeof(name)-1, &skip_name);
	if (!strlen(num)) {
		return -1;
	}
	pj_strdup2(pool, &sip_uri->user, num);

	if (strlen(name)) {
		pj_strdup2(pool, &id_name_addr->display, name);
	}

	pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, id_name_addr, new_val, len);
	ast_log(LOG_NOTICE, "get_new_referredby_val, old_referredby[%s], new_val[%s]\n", old_referredby, new_val);
	return 0;
}
#endif

/*!
 * \brief Use the value of a channel variable as the value of a SIP header
 *
 * This looks up a variable name on a channel, then takes that value and adds
 * it to an outgoing SIP request. If the header already exists on the message,
 * then no action is taken.
 *
 * \pre chan is locked.
 *
 * \param chan The channel on which to find the variable.
 * \param var_name The name of the channel variable to use.
 * \param header_name The name of the SIP header to add to the outgoing message.
 * \param tdata The outgoing SIP message on which to add the header
 */
static void add_header_from_channel_var(struct ast_channel *chan, const char *var_name, const char *header_name, pjsip_tx_data *tdata)
{
	const char *var_value;
	pj_str_t pj_header_name;
	pjsip_hdr *header;

	var_value = pbx_builtin_getvar_helper(chan, var_name);
	if (ast_strlen_zero(var_value)) {
		return;
	}

	pj_cstr(&pj_header_name, header_name);
	header = pjsip_msg_find_hdr_by_name(tdata->msg, &pj_header_name, NULL);
	if (header) {
		return;
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 156466 */
	if (!strcmp(header_name, "Referred-By") && !strncmp(ast_channel_name(chan), "PJSIP/trunk_", 12)) {
		char buffer[256] = {0};

		if (!get_new_referredby_val(tdata->pool, chan, var_value, buffer, sizeof(buffer) - 1)) {
			ast_sip_add_header(tdata, header_name, buffer);
			return;
		}
	}
#endif

	ast_sip_add_header(tdata, header_name, var_value);
}

static void refer_outgoing_request(struct ast_sip_session *session, struct pjsip_tx_data *tdata)
{
	if (pjsip_method_cmp(&tdata->msg->line.req.method, &pjsip_invite_method)
		|| !session->channel
		|| session->inv_session->state != PJSIP_INV_STATE_NULL) {
		return;
	}

	ast_channel_lock(session->channel);
	add_header_from_channel_var(session->channel, "SIPREPLACESHDR", "Replaces", tdata);
	add_header_from_channel_var(session->channel, "SIPREFERREDBYHDR", "Referred-By", tdata);
	ast_channel_unlock(session->channel);
}

static struct ast_sip_session_supplement refer_supplement = {
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL + 1,
	.incoming_request = refer_incoming_request,
	.outgoing_request = refer_outgoing_request,
};

static int load_module(void)
{
	const pj_str_t str_norefersub = { "norefersub", 10 };

	pjsip_replaces_init_module(ast_sip_get_pjsip_endpoint());
	pjsip_xfer_init_module(ast_sip_get_pjsip_endpoint());
	pjsip_endpt_add_capability(ast_sip_get_pjsip_endpoint(), NULL, PJSIP_H_SUPPORTED, NULL, 1, &str_norefersub);

	ast_sip_register_service(&refer_progress_module);
	ast_sip_session_register_supplement(&refer_supplement);

	ast_module_shutdown_ref(ast_module_info->self);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&refer_supplement);
	ast_sip_unregister_service(&refer_progress_module);

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Blind and Attended Transfer Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip,res_pjsip_session,res_pjsip_pubsub",
);
