/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <signal.h>
#include <pjsip.h>
#include <pjsip_ua.h>

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/astdb.h"
#include "asterisk/app.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/causes.h"
#endif

#include "asterisk/res_pjsip.h"
#include "asterisk/module.h"
#include "asterisk/paths.h"
#include "asterisk/test.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/manager.h"
#include "asterisk/named_locks.h"
#include "asterisk/res_pjproject.h"
#include "res_pjsip/include/res_pjsip_private.h"
#include "asterisk/haserver.h"

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("pjsip");
#endif
/*** DOCUMENTATION
	<manager name="PJSIPShowRegistrationsInbound" language="en_US">
		<synopsis>
			Lists PJSIP inbound registrations.
		</synopsis>
		<syntax />
		<description>
			<para>
			In response, <literal>InboundRegistrationDetail</literal> events showing configuration
			and status information are raised for all contacts, static or dynamic.  Once all events
			are completed an <literal>InboundRegistrationDetailComplete</literal> is issued.
			</para>
			<warning><para>
				This command just dumps all coonfigured AORs with contacts, even if the contact
				is a permanent one.  To really get just inbound registrations, use
				<literal>PJSIPShowRegistrationInboundContactStatuses</literal>.
			</para>
			</warning>
		</description>
		<see-also>
			<ref type="manager" module="res_pjsip_registrar">PJSIPShowRegistrationInboundContactStatuses</ref>
		</see-also>
	</manager>
	<manager name="PJSIPShowRegistrationInboundContactStatuses" language="en_US">
		<synopsis>
			Lists ContactStatuses for PJSIP inbound registrations.
		</synopsis>
		<syntax />
		<description>
			<para>
			In response, <literal>ContactStatusDetail</literal> events showing status information
			are raised for each inbound registration (dynamic contact) object.  Once all events
			are completed a <literal>ContactStatusDetailComplete</literal> event is issued.
			</para>
		</description>
	</manager>
 ***/

static int pj_max_hostname = PJ_MAX_HOSTNAME;
static int pjsip_max_url_size = PJSIP_MAX_URL_SIZE;
struct pjsip_transport *mobile_tp=  NULL;

typedef struct ast_wave_mobile_watch{
	char aor_name[32];
	struct pjsip_transport *tp;

	struct ast_wave_mobile_watch *prev;
	struct ast_wave_mobile_watch *next;
}ast_wave_mobile_watch_node;

typedef struct {
    ast_wave_mobile_watch_node *head;
    ast_wave_mobile_watch_node *tail;
    unsigned int count_node;
    ast_mutex_t lock;
}ast_wave_mobile_watch_queue;

ast_wave_mobile_watch_queue g_wave_mobile_watch_queue;

void ast_wave_mobile_tp_queue_init(ast_wave_mobile_watch_queue *queue);
void ast_wave_mobile_tp_queue_clean(ast_wave_mobile_watch_queue *queue);
int ast_wave_mobile_tp_enqueue(ast_wave_mobile_watch_queue *q, char *aor, struct pjsip_transport *tp );
int ast_wave_mobile_tp_quitqueue(ast_wave_mobile_watch_queue *q, char *aor);
void ast_reset_wave_mobile_connect(ast_wave_mobile_watch_queue *q);


/*! \brief Internal function which returns the expiration time for a contact */
static int registrar_get_expiration(const struct ast_sip_aor *aor, const pjsip_contact_hdr *contact, const pjsip_rx_data *rdata)
{
	pjsip_expires_hdr *expires;
	int expiration = aor->default_expiration;

	if (contact && contact->expires != -1) {
		/* Expiration was provided with the contact itself */
		expiration = contact->expires;
	} else if ((expires = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_EXPIRES, NULL))) {
		/* Expiration was provided using the Expires header */
		expiration = expires->ivalue;
	}

	/* If the value has explicitly been set to 0, do not enforce */
	if (!expiration) {
		return expiration;
	}

	/* Enforce the range that we will allow for expiration */
	if (expiration < aor->minimum_expiration) {
		expiration = aor->minimum_expiration;
	} else if (expiration > aor->maximum_expiration) {
		expiration = aor->maximum_expiration;
	}

	return expiration;
}

/*! \brief Structure used for finding contact */
struct registrar_contact_details {
	/*! \brief Pool used for parsing URI */
	pj_pool_t *pool;
	/*! \brief URI being looked for */
	pjsip_sip_uri *uri;
#ifdef GRANDSTREAM_NETWORKS
	char sbc_endpoint_addr[80];
#endif
};

#ifdef BUILD_ON_CLOUD_PBX
static const pj_str_t sbc_endpoint_addr_name = {"SBC-Endpoint-Address", 20};
#endif

/*! \brief Callback function for finding a contact */
static int registrar_find_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	const struct registrar_contact_details *details = arg;
	pjsip_uri *contact_uri;

	if (ast_tvzero(contact->expiration_time)) {
		return 0;
	}

#ifdef BUILD_ON_CLOUD_PBX
	if (strlen(details->sbc_endpoint_addr) && !ast_strlen_zero(contact->sbc_endpoint_addr)) {
		if (strcmp(details->sbc_endpoint_addr, contact->sbc_endpoint_addr)) {
			return 0;
		}

		return CMP_MATCH;
	}
#endif

	contact_uri = pjsip_parse_uri(details->pool, (char*)contact->uri, strlen(contact->uri), 0);

	return (pjsip_uri_cmp(PJSIP_URI_IN_CONTACT_HDR, details->uri, contact_uri) == PJ_SUCCESS) ? CMP_MATCH : 0;
}

/*! \brief Internal function which validates provided Contact headers to confirm that they are acceptable, and returns number of contacts */
static int registrar_validate_contacts(const pjsip_rx_data *rdata, pj_pool_t *pool, struct ao2_container *contacts,
	struct ast_sip_aor *aor, int permanent, int *added, int *updated, int *deleted)
{
	pjsip_contact_hdr *previous = NULL;
	pjsip_contact_hdr *contact = (pjsip_contact_hdr *)&rdata->msg_info.msg->hdr;
	struct registrar_contact_details details = {
		.pool = pool,
	};

	memset(details.sbc_endpoint_addr, 0, sizeof(details.sbc_endpoint_addr));
#ifdef BUILD_ON_CLOUD_PBX
	/* get SBC-Endpoint-Address hdr value */
	pjsip_generic_string_hdr *hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &sbc_endpoint_addr_name, NULL);
	if (hdr) {
		snprintf(details.sbc_endpoint_addr, sizeof(details.sbc_endpoint_addr), "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);
	} 
#endif

	for (; (contact = (pjsip_contact_hdr *) pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, contact->next)); pj_pool_reset(pool)) {
		int expiration = registrar_get_expiration(aor, contact, rdata);
		struct ast_sip_contact *existing;
		char contact_uri[pjsip_max_url_size];

		if (contact->star) {
			/* The expiration MUST be 0 when a '*' contact is used and there must be no other contact */
			if (expiration != 0 || previous) {
				return -1;
			}
			/* Count all contacts to delete */
			*deleted = ao2_container_count(contacts) - permanent;
			previous = contact;
			continue;
		} else if (previous && previous->star) {
			/* If there is a previous contact and it is a '*' this is a deal breaker */
			return -1;
		}
		previous = contact;

		if (!PJSIP_URI_SCHEME_IS_SIP(contact->uri) && !PJSIP_URI_SCHEME_IS_SIPS(contact->uri)) {
			continue;
		}

		details.uri = pjsip_uri_get_uri(contact->uri);

		/* pjsip_uri_print returns -1 if there's not enough room in the buffer */
		if (pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, details.uri, contact_uri, sizeof(contact_uri)) < 0) {
			/* If the total length of the uri is greater than pjproject can handle, go no further */
			return -1;
		}

		if (details.uri->host.slen >= pj_max_hostname) {
			/* If the length of the hostname is greater than pjproject can handle, go no further */
			return -1;
		}

		/* Determine if this is an add, update, or delete for policy enforcement purposes */
		existing = ao2_callback(contacts, 0, registrar_find_contact, &details);
		ao2_cleanup(existing);
		if (!existing) {
			if (expiration) {
				++*added;
			}
		} else if (expiration) {
			++*updated;
		} else {
			++*deleted;
		}
	}

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS

/*! \brief Internal function used to delete the first contact from an AOR */
static int registrar_delete_first_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	const char *aor_name = arg;

	ast_sip_location_delete_contact(contact);
	if (!ast_strlen_zero(aor_name)) {
		ast_verb(3, "Removed contact '%s' from AOR '%s' due to replace\n", contact->uri, aor_name);
		ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
				"Contact: %s\r\n"
				"AOR: %s\r\n"
				"UserAgent: %s",
				contact->uri,
				aor_name,
				contact->user_agent);
	}

	return CMP_MATCH;
}


/* Added by yfyang, for GsWave Login */
static int regitrar_get_gs_access_info(int *server_port, int *https_flag)
{
	char value[64] = {0};

	ast_sip_get_nvram_value("900", value, sizeof(value));
	if (strlen(value) > 0) {
		*https_flag = atoi(value);
	}

	ast_sip_get_nvram_value("gswave_port", value, sizeof(value));
	if (strlen(value) > 0) {
		*server_port = atoi(value);
	}

	return 0;
}

/*! Adds a X-GS-Access-Info header to a response 
    eg: X-GS-Access-Info: <https://192.168.129.171:8090>
*/
static void registrar_add_access_info_header(pjsip_tx_data *tdata)
{
	pjsip_fromto_hdr *to = NULL;
	pjsip_name_addr  *id_name_addr = NULL;
	pjsip_sip_uri    *uri = NULL;
	char access[256]     = {0};
	char host_value[256] = {0};
	int  port = 0;
	int  flag = 0;

	regitrar_get_gs_access_info(&port, &flag);
	if (!flag) {
		/* not https, do nothing */
		return;
	}

	to = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_TO, NULL);
	if (!to) {
		return;
	}

	id_name_addr = (pjsip_name_addr *)to->uri;
	uri = pjsip_uri_get_uri(id_name_addr);
	if (!uri) {
		return;
	}

	/* cmp uri and WsProxyHost */
	ast_sip_get_nvram_value(":WsProxyHost", host_value, sizeof(host_value));
	if (!strlen(host_value) || strncmp(host_value, uri->host.ptr, uri->host.slen)) {
		snprintf(access, sizeof(access) - 1, "<https://%.*s:%d>", (int)uri->host.slen, uri->host.ptr, port);
	} else {
		snprintf(access, sizeof(access) - 1, "<https://%s>", host_value);
	}

	ast_sip_add_header(tdata, "X-GS-Access-Info", access);
}

static pj_str_t GS_CONFIG_SUPPORTED_NAME = { "x-gs-config", 11 };
static pj_str_t GS_STATE_SUPPORTED_NAME = { "x-gs-state", 10 };
static pj_str_t GS_RESOLUTION_SUPPORTED_NAME = { "x-gs-resolution", 15 };
static pj_str_t GS_CONTROL_NAME = { "x-gs-control", 12 };

static int registrar_add_support_header(pjsip_tx_data *tdata, struct ast_sip_contact_extra_info extra_info)
{
	pjsip_supported_hdr *hdr;

	if (!extra_info.support_gs_config && !extra_info.support_gs_state) {
		return 0;
	}

	hdr = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_SUPPORTED, NULL);
	if (!hdr) {
		/* insert a new Supported header */
		hdr = pjsip_supported_hdr_create(tdata->pool);
		if (!hdr) {
			return -1;
		}

		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
	}

	if (extra_info.support_gs_config) {
		ast_log(LOG_DEBUG, "Add x-gs-config support header\n");
		pj_strassign(&hdr->values[hdr->count++], &GS_CONFIG_SUPPORTED_NAME);
	}

	if (extra_info.support_gs_state) {
		ast_log(LOG_DEBUG, "Add x-gs-state support header\n");
		pj_strassign(&hdr->values[hdr->count++], &GS_STATE_SUPPORTED_NAME);
	}

	if (extra_info.support_gs_resolution) {
		ast_log(LOG_DEBUG, "Add x-gs-resolution support header\n");
		pj_strassign(&hdr->values[hdr->count++], &GS_RESOLUTION_SUPPORTED_NAME);
	}

	if (extra_info.support_gs_control) {
		ast_log(LOG_DEBUG, "Add x-gs-control support header\n");
		pj_strassign(&hdr->values[hdr->count++], &GS_CONTROL_NAME);
	}

	return 0;
}
#endif

/*! \brief Internal function used to delete a contact from an AOR */
static int registrar_delete_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	const char *aor_name = arg;

	/* Permanent contacts can't be deleted */
	if (ast_tvzero(contact->expiration_time)) {
		return 0;
	}

	ast_sip_location_delete_contact(contact);
	if (!ast_strlen_zero(aor_name)) {
		ast_verb(3, "Removed contact '%s' from AOR '%s' due to request\n", contact->uri, aor_name);
		ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
				"Contact: %s\r\n"
				"AOR: %s\r\n"
				"UserAgent: %s",
				contact->uri,
				aor_name,
				contact->user_agent);
	}

	return CMP_MATCH;
}

/*! \brief Internal function which adds a contact to a response */
static int registrar_add_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	pjsip_tx_data *tdata = arg;
	pjsip_contact_hdr *hdr = pjsip_contact_hdr_create(tdata->pool);
	pj_str_t uri;

	pj_strdup2_with_null(tdata->pool, &uri, contact->uri);
	hdr->uri = pjsip_parse_uri(tdata->pool, uri.ptr, uri.slen, PJSIP_PARSE_URI_AS_NAMEADDR);
	hdr->expires = ast_tvdiff_ms(contact->expiration_time, ast_tvnow()) / 1000;

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr*)hdr);

	return 0;
}

static const pj_str_t path_hdr_name = { "Path", 4 };

static int build_path_data(pjsip_rx_data *rdata, struct ast_str **path_str)
{
	pjsip_generic_string_hdr *path_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &path_hdr_name, NULL);

	if (!path_hdr) {
		return 0;
	}

	*path_str = ast_str_create(64);
	if (!*path_str) {
		return -1;
	}

	ast_str_set(path_str, 0, "%.*s", (int)path_hdr->hvalue.slen, path_hdr->hvalue.ptr);

	while ((path_hdr = (pjsip_generic_string_hdr *) pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &path_hdr_name, path_hdr->next))) {
		ast_str_append(path_str, 0, ",%.*s", (int)path_hdr->hvalue.slen, path_hdr->hvalue.ptr);
	}

	return 0;
}

static int registrar_validate_path(pjsip_rx_data *rdata, struct ast_sip_aor *aor, struct ast_str **path_str)
{
	const pj_str_t path_supported_name = { "path", 4 };
	pjsip_supported_hdr *supported_hdr;
	int i;

	if (!aor->support_path) {
		return 0;
	}

	if (build_path_data(rdata, path_str)) {
		return -1;
	}

	if (!*path_str) {
		return 0;
	}

	supported_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_SUPPORTED, NULL);
	if (!supported_hdr) {
		return -1;
	}

	/* Find advertised path support */
	for (i = 0; i < supported_hdr->count; i++) {
		if (!pj_stricmp(&supported_hdr->values[i], &path_supported_name)) {
			return 0;
		}
	}

	/* Path header present, but support not advertised */
	return -1;
}

#ifdef GRANDSTREAM_NETWORKS

/*! \brief Structure used for finding contact by uuid */
struct registrar_contact_uuid {
	/*! \brief URI being looked for */
	pjsip_uri *uri;
	/*! \brief uuid being looked for */
	char *uuid;
	int  reg_id;
};

/*! \brief Callback function for finding a contact */
static int registrar_find_uuid(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	const struct registrar_contact_uuid *uuid_obj = arg;
	int res = 0;
	char contact_uri[PJSIP_MAX_URL_SIZE] = {0};

	if (uuid_obj == NULL || uuid_obj->uri == NULL || contact == NULL || contact->uri == NULL) {
		return res;
	}

	/* if the uuid is not null */
	if (!ast_strlen_zero(uuid_obj->uuid)) {
		if (!ast_strlen_zero(contact->instance) && strcasecmp(uuid_obj->uuid, contact->instance) == PJ_SUCCESS) {
			//same device
            if (contact->reg_id == uuid_obj->reg_id) {
                res = CMP_MATCH | CMP_STOP;
            }
        }
        //res = (contact->uuid && strcasecmp(uuid_obj->uuid, contact->uuid) == PJ_SUCCESS) ? CMP_MATCH | CMP_STOP : 0;
	}
	else {
		pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, uuid_obj->uri, contact_uri, sizeof(contact_uri));

		char *p = strchr( contact_uri,  '@');
		if ( p != NULL && (p = strchr( p, ':')) != NULL ) {
			p++;
			*p = '\0';
		}

		if (ast_strlen_zero(contact->instance) && contact_uri[0] != '\0' && strstr(contact->uri, contact_uri) != NULL) {
			res = CMP_MATCH | CMP_STOP;
		}
		//res = (ast_strlen_zero(contact->uuid) && contact_uri[0] != '\0' && strstr(contact->uri, contact_uri) != NULL) ? CMP_MATCH | CMP_STOP : 0;
	}

	return res;
}

#if 0
static char *get_uuid_from_contact (pjsip_contact_hdr *contact_hdr)
{
	static const pj_str_t uuid_str = { "+sip.instance", 13 };
	pjsip_param *uuid;
	char *ptr = NULL;
	char uuid_buf[64] = {0};

	if (!contact_hdr) {
		return NULL;
	}

	uuid = pjsip_param_find(&contact_hdr->other_param, &uuid_str);
	if (uuid && uuid->value.ptr) {
		sscanf(uuid->value.ptr,"%*[^:]:uuid:%63[^>*]",uuid_buf);
		ptr = ast_strdup(uuid_buf);
	}

	return ptr;
}
#endif

static int get_instance_from_contact (pjsip_contact_hdr *contact_hdr, struct ast_sip_contact_extra_info *extra_info)
{
	static const pj_str_t tmp_str = { "+sip.instance", 13 };
	static const pj_str_t id_str = { "reg-id", 6 };
	pjsip_param *instance = NULL;
	pjsip_param *reg_id_param = NULL;
	char buf[64] = {0};

	if (!contact_hdr) {
		return 0;
	}

	instance = pjsip_param_find(&contact_hdr->other_param, &tmp_str);
	if (instance && instance->value.ptr) {
		sscanf(instance->value.ptr,"\"%[^\"]\"",buf);

		ast_copy_string(extra_info->instance, buf, sizeof(extra_info->instance));
	}

	extra_info->reg_id = 0;
	reg_id_param = pjsip_param_find(&contact_hdr->other_param, &id_str);
	if (reg_id_param && reg_id_param->value.ptr) {
		extra_info->reg_id = atoi(reg_id_param->value.ptr);
	}

	ast_log(LOG_NOTICE, "get instance[%s], reg_id[%d]", extra_info->instance, extra_info->reg_id);

	return 0;
}


static int check_support_header_value(pjsip_rx_data *rdata, struct ast_sip_contact_extra_info *extra_info)
{
	const pj_str_t missed_call_tag = { "gs-missed-call", 14 };
	const pj_str_t gs_config_tag = { "x-gs-config", 11 };
	const pj_str_t gs_state_tag = { "x-gs-state", 10 };
	const pj_str_t gs_message_tag = { "x-gs-message", 12 };
	const pj_str_t gs_resolution_tag = { "x-gs-resolution", 15 };
	const pj_str_t gs_fec_tag = { "GS-FEC", 6 };
	const pj_str_t gs_conf_invite_tag = { "x-gs-conf-invite", 16};
	const pj_str_t gs_control = { "x-gs-control", 12};

	pjsip_supported_hdr *supported_hdr;
	int i;

	if (!rdata) {
		return 0;
	}

	supported_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_SUPPORTED, NULL);
	if (!supported_hdr) {
		return 0;
	}

	/* Find advertised path support */
	for (i = 0; i < supported_hdr->count; i++) {
		if (!pj_stricmp(&supported_hdr->values[i], &missed_call_tag)) {
			ast_log(LOG_NOTICE, "Support gs-missed-call!\n");
			extra_info->support_missed_call = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_config_tag)) {
			ast_log(LOG_NOTICE, "Support x-gs-config!\n");
			extra_info->support_gs_config = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_state_tag)) {
			ast_log(LOG_NOTICE, "Support x-gs-state!\n");
			extra_info->support_gs_state = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_fec_tag)) {
			ast_log(LOG_NOTICE, "Support GS-FEC!\n");
			extra_info->support_gs_fec = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_message_tag)) {
			ast_log(LOG_NOTICE, "x-gs-message!\n");
            extra_info->support_gs_message = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_resolution_tag)) {
			ast_log(LOG_NOTICE, "x-gs-resolution!\n");
			extra_info->support_gs_resolution = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_conf_invite_tag)) {
			ast_log(LOG_NOTICE, "Support x-gs-conf-invite!\n");
			extra_info->support_gs_conf_invite = 1;
			continue;
		}

		if (!pj_stricmp(&supported_hdr->values[i], &gs_control)) {
			ast_log(LOG_NOTICE, "Support x-gs-control!\n");
			extra_info->support_gs_control = 1;
			continue;
		}
	}

	return 0;
}

static unsigned int check_support_ice (pjsip_contact_hdr *contact_hdr)
{
	static const pj_str_t ice_str = { "+sip.ice", 8 };
	pjsip_param *ice;

	if (!contact_hdr) {
		return 0;
	}

	ice = pjsip_param_find(&contact_hdr->other_param, &ice_str);
	if (ice) {
		ast_log(LOG_DEBUG, "Contact contain sip.ice, support ice\n");
		return 1;
	}

	return 0;
}

static unsigned int check_support_trickle_ice (pjsip_contact_hdr *contact_hdr)
{
	static const pj_str_t trickle_ice_str = { "trickle-ice", 11 };
	pjsip_param *trickle_ice = NULL;

	if (!contact_hdr) {
		return 0;
	}

	trickle_ice = pjsip_param_find(&contact_hdr->other_param, &trickle_ice_str);
	if (trickle_ice) {
		ast_log(LOG_DEBUG, "Contact contain trickle-ice, support trickle ice\n");
		return 1;
	}

	return 0;
}

static unsigned int check_if_register_by_gswave_mobile(pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t hdr_type = {"Grandstream-type", 16};

	if (!rdata) {
		return FALSE;
	}

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &hdr_type, NULL))) {
		return FALSE;
	}

	if (pj_stricmp2(&hdr->hvalue, "GSWave mobile")) {
		return FALSE;
	}

	return TRUE;
}
#endif

/*! Transport monitor for incoming REGISTER contacts */
struct contact_transport_monitor {
	/*!
	 * \brief Sorcery contact name to remove on transport shutdown
	 * \note Stored after aor_name in space reserved when struct allocated.
	 */
	char *contact_name;
	/*! Indicates that the monitor is in the process of removing a contact */
	int removing;
	/*! AOR name the contact is associated */
	char aor_name[0];
};

static int contact_transport_monitor_matcher(void *a, void *b)
{
	struct contact_transport_monitor *ma = a;
	struct contact_transport_monitor *mb = b;

	return strcmp(ma->aor_name, mb->aor_name) == 0
		&& strcmp(ma->contact_name, mb->contact_name) == 0;
}

static int register_contact_transport_remove_cb(void *data)
{
	struct contact_transport_monitor *monitor = data;
	struct ast_sip_contact *contact;
	struct ast_sip_aor *aor;

	aor = ast_sip_location_retrieve_aor(monitor->aor_name);
	if (!aor) {
		ao2_lock(monitor);
		monitor->removing = 0;
		ao2_unlock(monitor);
		ao2_ref(monitor, -1);
		return 0;
	}

	ao2_lock(aor);

	contact = ast_sip_location_retrieve_contact(monitor->contact_name);
	if (contact) {
		ast_sip_location_delete_contact(contact);
		ast_verb(3, "Removed contact '%s' from AOR '%s' due to transport shutdown\n",
			contact->uri, monitor->aor_name);
		ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
			"Contact: %s\r\n"
			"AOR: %s\r\n"
			"UserAgent: %s",
			contact->uri,
			monitor->aor_name,
			contact->user_agent);
		ao2_ref(contact, -1);
	}
	ao2_unlock(aor);
	ao2_ref(aor, -1);

	ao2_ref(monitor, -1);
	return 0;
}

/*!
 * \internal
 * \brief The reliable transport we registered as a contact has shutdown.
 *
 * \param data What contact needs to be removed.
 *
 * \note Normally executed by the pjsip monitor thread.
 *
 * \return Nothing
 */
static void register_contact_transport_shutdown_cb(void *data)
{
	struct contact_transport_monitor *monitor = data;

	/*
	 * It's possible for this shutdown handler to get called multiple times for the
	 * same monitor from different threads. Only one of the calls needs to do the
	 * actual removing of the contact, so if one is currently removing then any
	 * subsequent calls can skip.
	 */
	ao2_lock(monitor);
	if (monitor->removing) {
		ao2_unlock(monitor);
		return;
	}

	monitor->removing = 1;

	/*
	 * Push off to a default serializer.  This is in case sorcery
	 * does database accesses for contacts.  Database accesses may
	 * not be on this machine.  We don't want to tie up the pjsip
	 * monitor thread with potentially long access times.
	 */
	ao2_ref(monitor, +1);
	if (ast_sip_push_task(NULL, register_contact_transport_remove_cb, monitor)) {
		monitor->removing = 0;
		ao2_ref(monitor, -1);
	}

	ao2_unlock(monitor);
}

AST_VECTOR(excess_contact_vector, struct ast_sip_contact *);

static int vec_contact_cmp(struct ast_sip_contact *left, struct ast_sip_contact *right)
{
	struct ast_sip_contact *left_contact = left;
	struct ast_sip_contact *right_contact = right;

	/* Sort from soonest to expire to last to expire */
	return ast_tvcmp(left_contact->expiration_time, right_contact->expiration_time);
}

static int vec_contact_add(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct excess_contact_vector *contact_vec = arg;

	/*
	 * Performance wise, an insertion sort is fine because we
	 * shouldn't need to remove more than a handful of contacts.
	 * I expect we'll typically be removing only one contact.
	 */
	AST_VECTOR_ADD_SORTED(contact_vec, contact, vec_contact_cmp);
	if (AST_VECTOR_SIZE(contact_vec) == AST_VECTOR_MAX_SIZE(contact_vec)) {
		/*
		 * We added a contact over the number we need to remove.
		 * Remove the longest to expire contact from the vector
		 * which is the last element in the vector.  It may be
		 * the one we just added or the one we just added pushed
		 * out an earlier contact from removal consideration.
		 */
		--AST_VECTOR_SIZE(contact_vec);
	}
	return 0;
}

/*!
 * \internal
 * \brief Remove excess existing contacts that expire the soonest.
 * \since 13.18.0
 *
 * \param contacts Container of unmodified contacts that could remove.
 * \param to_remove Maximum number of contacts to remove.
 *
 * \return Nothing
 */
static void remove_excess_contacts(struct ao2_container *contacts, struct ao2_container *response_contacts,
	unsigned int to_remove)
{
	struct excess_contact_vector contact_vec;

	/*
	 * Create a sorted vector to hold the to_remove soonest to
	 * expire contacts.  The vector has an extra space to
	 * temporarily hold the longest to expire contact that we
	 * won't remove.
	 */
	if (AST_VECTOR_INIT(&contact_vec, to_remove + 1)) {
		return;
	}
	ao2_callback(contacts, OBJ_NODATA | OBJ_MULTIPLE, vec_contact_add, &contact_vec);

	/*
	 * The vector should always be populated with the number
	 * of contacts we need to remove.  Just in case, we will
	 * remove all contacts in the vector even if the contacts
	 * container had fewer contacts than there should be.
	 */
	ast_assert(AST_VECTOR_SIZE(&contact_vec) == to_remove);
	to_remove = AST_VECTOR_SIZE(&contact_vec);

	/* Remove the excess contacts that expire the soonest */
	while (to_remove--) {
		struct ast_sip_contact *contact;

		contact = AST_VECTOR_GET(&contact_vec, to_remove);

		ast_sip_location_delete_contact(contact);
		ast_verb(3, "Removed contact '%s' from AOR '%s' due to remove_existing\n",
			contact->uri, contact->aor);
		ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
			"Contact: %s\r\n"
			"AOR: %s\r\n"
			"UserAgent: %s",
			contact->uri,
			contact->aor,
			contact->user_agent);
		ao2_unlink(response_contacts, contact);
	}

	AST_VECTOR_FREE(&contact_vec);
}

#ifndef GRANDSTREAM_NETWORKS
/*! \brief Callback function which adds non-permanent contacts to a container */
static int registrar_add_non_permanent(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct ao2_container *container = arg;

	if (ast_tvzero(contact->expiration_time)) {
		return 0;
	}

	ao2_link(container, contact);

	return 0;
}
#endif

struct aor_core_response {
	/*! Tx data to use for statefull response.  NULL for stateless response. */
	pjsip_tx_data *tdata;
	/*! SIP response code to send in stateless response */
	int code;
#ifdef GRANDSTREAM_NETWORKS
	int reason;
#endif
};

#ifdef GRANDSTREAM_NETWORKS
struct wave_register_data {
	char aor_name[80];
	char user_agent[128];
	char instance[128];
	int enable_duplicate_logins;
};

/* Add by yfyang for replace gswave mobile contact */
static int registrar_find_gsmobile_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct wave_register_data *data = arg;
	int enable_duplicate_logins = 0;

	enable_duplicate_logins = data ? data->enable_duplicate_logins : 0;
	ast_log(LOG_DEBUG, "The duplicate login is %s \n", enable_duplicate_logins ? "enable" : "disabled");

	ast_log(LOG_DEBUG, "The duplicate login is %s \n", enable_duplicate_logins ? "enable" : "disabled");

	if (enable_duplicate_logins) {
		if ( contact && data && data->instance && !strcasecmp(contact->instance, data->instance)) {
			ast_log(LOG_DEBUG, "Get the same instance %s \n", data->instance);
			return CMP_MATCH | CMP_STOP;
		}
	}
	else {
		if (contact && contact->gswave_mobile) {
			return CMP_MATCH | CMP_STOP;
		}
	}

	return  0;
}

/* Add by bfdu for replace gswave contact, 2019.12.27 */
static int registrar_find_ws_contact_by_aor_name(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct wave_register_data *data = arg;
	char* contact_uri = (char*)contact->uri;
	char data_user_agent_param1[16] = { 0 };
	char data_user_agent_param2[16] = { 0 };
	char data_user_agent_param3[64] = { 0 };
	char data_user_agent_param4[32] = { 0 };
	char contact_user_agent_param1[16] = { 0 };
	char contact_user_agent_param2[16] = { 0 };
	char contact_user_agent_param3[64] = { 0 };
	char contact_user_agent_param4[32] = { 0 };
	unsigned int is_wave = 0;
	int enable_duplicate_logins = 0;

	enable_duplicate_logins = data ? data->enable_duplicate_logins : 0;
	ast_log(LOG_DEBUG, "The duplicate login is %s \n", enable_duplicate_logins ? "enable" : "disabled");

	if (contact->user_agent && (strcasestr(contact->user_agent, "Grandstream Wave Web") || strcasestr(contact->user_agent, "Grandstream Wave Desktop"))) {
		is_wave = 1;
	}

	if (contact_uri != NULL && (strstr(contact_uri, data->aor_name) == NULL || (strcasestr(contact_uri, "WS") == NULL && !is_wave))) {
		return 0;
	}

	if (enable_duplicate_logins) {
		if (contact->instance != NULL && data->instance && data->instance[0] != '\0' && !strcasecmp(contact->instance, data->instance)) {
			ast_log(LOG_DEBUG, "Get the same instance\n"); 
			return CMP_MATCH | CMP_STOP;
		}
	} else {
		if (sscanf(data->user_agent, "%15s %15s %63s %31s", data_user_agent_param1, data_user_agent_param2,
			data_user_agent_param3, data_user_agent_param4) < 4) {
			ast_log(LOG_WARNING, "Can't get user agent params from register message\n");
		}

		if (contact->user_agent && sscanf(contact->user_agent, "%15s %15s %63s %31s", contact_user_agent_param1, contact_user_agent_param2,
			contact_user_agent_param3, contact_user_agent_param4) < 4) {
			ast_log(LOG_WARNING, "Can't get user agent params from contact info\n");
		}

			ast_log(LOG_WARNING, "%s %s %s %s\n", data_user_agent_param1, data_user_agent_param2,
			data_user_agent_param3, data_user_agent_param4);
			ast_log(LOG_WARNING, "%s %s %s %s\n", contact_user_agent_param1, contact_user_agent_param2,
			contact_user_agent_param3, contact_user_agent_param4);

		if (!strlen(data_user_agent_param3) && !strlen(contact_user_agent_param3)) {
			ast_log(LOG_DEBUG, "Both can't get user-agent, consider not match\n");
			return 0;
		}

		if (!strcasecmp(data_user_agent_param3, contact_user_agent_param3)) {
			ast_log(LOG_DEBUG, "Get the same user agent terminal type\n");
			return CMP_MATCH | CMP_STOP;
		}

		if ((!strcasecmp(data_user_agent_param3, "Desktop") ||
			!strcasecmp(data_user_agent_param3, "Web") ||
			!strcasecmp(data_user_agent_param3, "Web_Extension")) &&
			(!strcasecmp(contact_user_agent_param3, "Desktop") ||
			!strcasecmp(contact_user_agent_param3, "Web") ||
			!strcasecmp(contact_user_agent_param3, "Web_Extension"))) {
			ast_log(LOG_DEBUG, "Web and desktop need replace\n");
			return CMP_MATCH | CMP_STOP;
		}
	}

	return 0;
}

static struct ast_sip_contact *get_exist_contact_for_ws(pjsip_sip_uri *uri,
	struct wave_register_data *data, struct ao2_container *contacts, pjsip_rx_data *rdata)
{
	pj_str_t type_name;
	unsigned int is_wave = 0;

	if (!uri || !data || !contacts) {
		ast_log(LOG_WARNING, "get_exist_contact_for_ws param is NULL\n");
		return NULL;
	}

	is_wave = ast_sip_judge_if_wave_terminal_type(rdata);

	if (pj_stricmp(&uri->transport_param, pj_cstr(&type_name, "ws")) && !is_wave) {
		ast_log(LOG_DEBUG, "It's not websocket connect\n");
		return NULL;
	}

	struct ast_sip_contact *contact = ao2_callback(contacts, OBJ_UNLINK, registrar_find_ws_contact_by_aor_name, data);
	if (contact) {
		ast_log(LOG_DEBUG, "Find ws contact for aor:%s\n", data->aor_name);
		return contact;
	}

	return NULL;
}

static void send_notify_to_update_turn_info(struct ast_sip_contact *contact)
{
	RAII_VAR(char *, turn_sever_addr, NULL, ast_free);
	RAII_VAR(char *, stun_sever_addr, NULL, ast_free);
	RAII_VAR(char *, user_name, NULL, ast_free);
	RAII_VAR(char *, password, NULL, ast_free);
	RAII_VAR(char *, protocol_type, NULL, ast_free);

	turn_sever_addr = ast_sip_get_turn_server_address();
	stun_sever_addr = ast_sip_get_stun_address();
	user_name = ast_sip_get_turn_user_name();
	password = ast_sip_get_turn_password();
	protocol_type = ast_sip_get_turn_protocol_type();

	ast_sip_send_notify_to_update_turn_info(contact, turn_sever_addr, user_name, password, protocol_type);

	return;
}


#ifdef BUILD_ON_CLOUD_PBX
static int ast_sip_check_if_clouducm(pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	const pj_str_t name = { "User-Agent", 10 };
	char user_agent[128] = { 0 };

	if (!rdata || !rdata->msg_info.msg) {
		return 0;
	}

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &name, NULL);
	if (!hdr) {
		return 0;
	}

	snprintf(user_agent, sizeof(user_agent) - 1, "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);

	if (strcasestr(user_agent, "Grandstream CloudUCM")) {
		return 1;
	}

	return 0;
}
#endif

static char *get_contact_original_address(struct pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t original_address = { "contact-original-address", 24 };
	char original_address_str[128] = {0};
	struct ast_sockaddr remote;
	char addr[256] = {0};

	if (!rdata) {
		return NULL;
	}

	if (!strncasecmp(rdata->tp_info.transport->type_name, "WS", 2)) {
		char port[8] = {0};

		snprintf(port, sizeof(port), "%d", rdata->pkt_info.src_port);
		ast_db_get("WSRemote", port, addr, sizeof(addr)-1);
		if (strlen(addr)) {
			ast_sockaddr_parse(&remote, addr, PARSE_PORT_FORBID);
			ast_sockaddr_set_port(&remote, rdata->pkt_info.src_port);
			snprintf(original_address_str, sizeof(original_address_str) - 1, "%s", ast_sockaddr_stringify(&remote));

			ast_log(LOG_NOTICE, "get ws original address: %s\n", original_address_str);
			return ast_strdup(original_address_str);
		}
	}

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &original_address, NULL))) {
		ast_log(LOG_NOTICE, "can't get_contact_original_address\n");
		return NULL;
	}

	if (0 == hdr->hvalue.slen) {
		ast_log(LOG_NOTICE, "can't get_contact_original_address\n");
		return NULL;
	}

	snprintf(addr, sizeof(addr) - 1, "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);

#ifdef BUILD_ON_CLOUD_PBX
	pj_str_t  sbc_addr_hdr_str = { "SBC-Endpoint-Address", 20 };

	if (pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &sbc_addr_hdr_str, NULL) && ast_sip_check_if_clouducm(rdata)) {
		ast_sockaddr_parse(&remote, addr, PARSE_ADDR);
	} else {
		ast_sockaddr_parse(&remote, addr, PARSE_PORT_IGNORE);
		ast_sockaddr_set_port(&remote, rdata->pkt_info.src_port);
	}
#else
	ast_sockaddr_parse(&remote, addr, PARSE_PORT_IGNORE);
	ast_sockaddr_set_port(&remote, rdata->pkt_info.src_port);
#endif

	snprintf(original_address_str, sizeof(original_address_str) - 1, "%s", ast_sockaddr_stringify(&remote));

	/* Added by yfyang, for Bug 292028 */
	if (!strlen(original_address_str) || !strcmp(original_address_str, "(null)")) {
		if (strchr(addr, ':' )) {
			snprintf(original_address_str, sizeof(original_address_str) - 1, "%s", addr);
		} else {
			snprintf(original_address_str, sizeof(original_address_str) - 1, "%s:%d", addr, rdata->pkt_info.src_port);
		}
	}

	ast_log(LOG_NOTICE, "get_contact_original_address: %s\n", original_address_str);

	return ast_strdup(original_address_str);
}

static void set_extra_info_for_remove_exsit_contact(struct ast_sip_contact_extra_info *extra_info,
	struct ast_sip_contact *contact)
{
	if (!extra_info || !contact) {
		return;
	}

	extra_info->support_ldap_settings = contact->support_ldap_settings;
	extra_info->require_feature_code_settings = contact->require_feature_code_settings;
	extra_info->require_turn_info_settings = contact->require_turn_info_settings;
	extra_info->require_basic_settings = contact->require_basic_settings;
	extra_info->require_plugin = contact->require_plugin;
	extra_info->require_phonebook = contact->require_phonebook;

	return;
}

static void set_to_host_addr_extra_info(struct ast_sip_contact_extra_info *extra_info, struct pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_from_hdr *to_hdr = NULL;
	pjsip_sip_uri *uri_to = NULL;

	if (!extra_info || !rdata) {
		return;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		ast_log(LOG_WARNING, "Has no pjsip msg, can't set to host\n");
		return;
	}

	to_hdr = pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
	if (!to_hdr) {
		ast_log(LOG_WARNING, "Can't find To header, can't set to host\n");
		return;
	}

	uri_to = pjsip_uri_get_uri(to_hdr->uri);
	if (!uri_to) {
		ast_log(LOG_WARNING, "Can't get To header uri, can't set to host\n");
		return;
	}

	ast_copy_pj_str(extra_info->to_host, &uri_to->host, sizeof(extra_info->to_host));

	return;
}
#endif

static void register_aor_core(pjsip_rx_data *rdata,
	struct ast_sip_endpoint *endpoint,
	struct ast_sip_aor *aor,
	const char *aor_name,
	struct ao2_container *contacts,
	struct aor_core_response *response)
{
	static const pj_str_t USER_AGENT = { "User-Agent", 10 };

	int added = 0;
	int updated = 0;
	int deleted = 0;
	int permanent = 0;
	int contact_count;
	struct ao2_container *existing_contacts = NULL;
	pjsip_contact_hdr *contact_hdr = (pjsip_contact_hdr *)&rdata->msg_info.msg->hdr;
	struct registrar_contact_details details = { 0, };
	pjsip_tx_data *tdata;
	RAII_VAR(struct ast_str *, path_str, NULL, ast_free);
	struct ast_sip_contact *response_contact;
	char *user_agent = NULL;
	pjsip_user_agent_hdr *user_agent_hdr;
	pjsip_expires_hdr *expires_hdr;
	pjsip_via_hdr *via_hdr;
	pjsip_via_hdr *via_hdr_last;
	char *via_addr = NULL;
	int via_port = 0;
	pjsip_cid_hdr *call_id_hdr;
	char *call_id = NULL;
	size_t alloc_size;
#ifdef GRANDSTREAM_NETWORKS
	pjsip_expires_hdr     *expires;

	/*Start added by zyxia for wakeup gswave ,get ios unique appid from regisger,2018.7.25*/
	pjsip_generic_string_hdr *gs_appid_hdr;
	char *gs_appid = NULL;
	static const pj_str_t GS_APPID = { "GS-APPID", 8 };
	int replace_first = 0;
	/*End added by zyxia for wakeup gswave ,get ios unique appid from regisger,2018.7.25*/
	char *original_address = NULL;

	struct ast_sip_contact_extra_info extra_info;
	memset(&extra_info, 0, sizeof(struct ast_sip_contact_extra_info));
#endif

	/* We create a single pool and use it throughout this function where we need one */
	details.pool = pjsip_endpt_create_pool(ast_sip_get_pjsip_endpoint(),
		"Contact Comparison", 1024, 256);
	if (!details.pool) {
		response->code = 500;
		return;
	}

	/* If there are any permanent contacts configured on the AOR we need to take them
	 * into account when counting contacts.
	 */
	if (aor->permanent_contacts) {
		permanent = ao2_container_count(aor->permanent_contacts);
	}

	if (registrar_validate_contacts(rdata, details.pool, contacts, aor, permanent, &added, &updated, &deleted)) {
		/* The provided Contact headers do not conform to the specification */
		ast_sip_report_failed_acl(endpoint, rdata, "registrar_invalid_contacts_provided");
		ast_log(LOG_WARNING, "Failed to validate contacts in REGISTER request from '%s'\n",
				ast_sorcery_object_get_id(endpoint));
		response->code = 400;
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
		return;
	}

	if (registrar_validate_path(rdata, aor, &path_str)) {
		/* Ensure that intervening proxies did not make invalid modifications to the request */
		ast_log(LOG_WARNING, "Invalid modifications made to REGISTER request from '%s' by intervening proxy\n",
				ast_sorcery_object_get_id(endpoint));
		response->code = 420;
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
		return;
	}
	
#ifdef GRANDSTREAM_NETWORKS
	if ((expires = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_EXPIRES, NULL))) {
		if ((0 != expires->ivalue) && (expires->ivalue < aor->minimum_expiration)) {
			ast_log(LOG_WARNING, "Expires[%d] too small, respond 423 with Min-Expires[%d].\n", expires->ivalue, aor->minimum_expiration);
			response->code = 423;
			pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
			return;
		}
	}
#endif
#ifndef GRANDSTREAM_NETWORKS
	if (aor->remove_existing) {
		/* Cumulative number of contacts affected by this registration */
		contact_count = MAX(updated + added - deleted,  0);

		/* We need to keep track of only existing contacts so we can later
		 * remove them if need be.
		 */
		existing_contacts = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_NOLOCK, 0,
			NULL, ast_sorcery_object_id_compare);
		if (!existing_contacts) {
			response->code = 500;
			pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
			return;
		}

		ao2_callback(contacts, OBJ_NODATA, registrar_add_non_permanent, existing_contacts);
	} else {
		/* Total contacts after this registration */
		contact_count = ao2_container_count(contacts) - permanent + added - deleted;
	}

	if (contact_count > aor->max_contacts ) {
		/* Enforce the maximum number of contacts */
		ast_sip_report_failed_acl(endpoint, rdata, "registrar_attempt_exceeds_maximum_configured_contacts");
		ast_log(LOG_WARNING, "Registration attempt from endpoint '%s' to AOR '%s' will exceed max contacts of %u\n",
				ast_sorcery_object_get_id(endpoint), aor_name, aor->max_contacts);
		response->code = 403;
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
		ao2_cleanup(existing_contacts);
		return;
	}
#else
	/* Total contacts after this registration */
	contact_count = ao2_container_count(contacts) - permanent + added - deleted;
	ast_log(LOG_DEBUG, "register_aor_core: added[%d], updated[%d], deleted[%d], contact_count[%d]", added, updated, deleted, contact_count);
#endif

	user_agent_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (user_agent_hdr) {
		alloc_size = pj_strlen(&user_agent_hdr->hvalue) + 1;
		user_agent = ast_alloca(alloc_size);
		ast_copy_pj_str(user_agent, &user_agent_hdr->hvalue, alloc_size);
	}

	/* Find the first Via header */
	via_hdr = via_hdr_last = (pjsip_via_hdr*) pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_VIA, NULL);
	if (via_hdr) {
		/* Find the last Via header */
		while ( (via_hdr = (pjsip_via_hdr*) pjsip_msg_find_hdr(rdata->msg_info.msg,
				PJSIP_H_VIA, via_hdr->next)) != NULL) {
			via_hdr_last = via_hdr;
		}
		alloc_size = pj_strlen(&via_hdr_last->sent_by.host) + 1;
		via_addr = ast_alloca(alloc_size);
		ast_copy_pj_str(via_addr, &via_hdr_last->sent_by.host, alloc_size);
		via_port=via_hdr_last->sent_by.port;
	}

	call_id_hdr = (pjsip_cid_hdr*) pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CALL_ID, NULL);
	if (call_id_hdr) {
		alloc_size = pj_strlen(&call_id_hdr->id) + 1;
		call_id = ast_alloca(alloc_size);
		ast_copy_pj_str(call_id, &call_id_hdr->id, alloc_size);
	}

#ifdef GRANDSTREAM_NETWORKS
	/*Start added by zyxia for WakeUp Gswave,get the ios unique appid from register message,2018.7.25*/
	gs_appid_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &GS_APPID, NULL);
	if (gs_appid_hdr && aor_name && strncmp("conf", aor_name, 4)) {
		size_t alloc_size = pj_strlen(&gs_appid_hdr->hvalue) + 1;
		gs_appid = ast_alloca(alloc_size);
		ast_copy_pj_str(gs_appid, &gs_appid_hdr->hvalue, alloc_size);
		ast_log(LOG_NOTICE, "Endpoint[%s] Get GS_APPID '%s',user_agent:%s.\n",ast_sorcery_object_get_id(endpoint),gs_appid,user_agent);
	}
	/*End added by zyxia for WakeUp Gswave,get the ios unique appid from register message,2018.7.25*/

	check_support_header_value(rdata, &extra_info);
	memset(details.sbc_endpoint_addr, 0, sizeof(details.sbc_endpoint_addr));
#ifdef BUILD_ON_CLOUD_PBX
	/* get SBC-Endpoint-Address hdr value */
	pjsip_generic_string_hdr *hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &sbc_endpoint_addr_name, NULL);
	if (hdr) {
		snprintf(details.sbc_endpoint_addr, sizeof(details.sbc_endpoint_addr), "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);
	}
#endif
#endif

	/* Iterate each provided Contact header and add, update, or delete */
	for (; (contact_hdr = (pjsip_contact_hdr *) pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, contact_hdr->next)); pj_pool_reset(details.pool)) {
		int expiration;
		char contact_uri[pjsip_max_url_size];
		RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

		if (contact_hdr->star) {
			/* A star means to unregister everything, so do so for the possible contacts */
			ao2_callback(contacts, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE,
				registrar_delete_contact, (void *)aor_name);
			/* If we are keeping track of existing contacts for removal then, well, there is
			 * absolutely nothing left so no need to try to remove any.
			 */
			if (existing_contacts) {
				ao2_ref(existing_contacts, -1);
				existing_contacts = NULL;
			}
			break;
		}

		if (!PJSIP_URI_SCHEME_IS_SIP(contact_hdr->uri) && !PJSIP_URI_SCHEME_IS_SIPS(contact_hdr->uri)) {
			/* This registrar only currently supports sip: and sips: URI schemes */
			continue;
		}

		expiration = registrar_get_expiration(aor, contact_hdr, rdata);
		details.uri = pjsip_uri_get_uri(contact_hdr->uri);
		pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, details.uri, contact_uri, sizeof(contact_uri));

#ifdef GRANDSTREAM_NETWORKS
		extra_info.ssh_proxy_flag = ast_sip_check_ssh_proxy_flag(rdata);
		ast_log(LOG_DEBUG, "ssh_proxy_flag = [%d]", extra_info.ssh_proxy_flag);

		if (extra_info.ssh_proxy_flag && !expiration && !strstr(contact_uri, aor_name)) {
			pj_strdup2(details.pool, &details.uri->user, aor_name);
			memset(contact_uri, 0, pjsip_max_url_size);
			pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, details.uri, contact_uri, sizeof(contact_uri));
			ast_log(LOG_DEBUG, "The new contact uri with aor_name: %s", contact_uri);
		}

		/* Added by yfyang, for Broadsoft Test, support GIN */
		if (strstr(contact_uri, ";bnc")) {
			/* bnc register, remove bnc parameter and get To user */
			static const pj_str_t bnc_name = { "bnc", 3 };
			pjsip_param  *x_bnc = NULL;

			pj_strdup2(details.pool, &details.uri->user, aor_name);  /* clear the user part */
			while ((x_bnc = pjsip_param_find(&details.uri->other_param, &bnc_name))) {
				pj_list_erase(x_bnc);
			}
			pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, details.uri, contact_uri, sizeof(contact_uri));
		}

		char *pTok = NULL;
		pTok = strstr(contact_uri, ";pn-tok");
		if (pTok) {
			*pTok = '\0';
		}

		/*Start added by zyxia for Wake Up ios gswave, the gswave register timeout is 7 days */
		if (expiration && !ast_strlen_zero(gs_appid)) {
			expiration = REGISGTER_VALID_TIME;
		}
		/*End added by zyxia for Wake Up ios gswave, the gswave register timeout is 7 days */
#endif
		contact = ao2_callback(contacts, OBJ_UNLINK, registrar_find_contact, &details);

		/* If a contact was returned and we need to keep track of existing contacts then it
		 * should be removed.
		 */
		if (contact && existing_contacts) {
			ao2_unlink(existing_contacts, contact);
		}

		if (!contact) {
			int prune_on_boot;

			/* If they are actually trying to delete a contact that does not exist... be forgiving */
			if (!expiration) {
				ast_verb(3, "Attempted to remove non-existent contact '%s' from AOR '%s' by request\n",
					contact_uri, aor_name);
				continue;
			}

#ifdef GRANDSTREAM_NETWORKS
			/*Modified by yfyang, for Bug 313577 */
			//RAII_VAR(char *, uuid, get_uuid_from_contact (contact_hdr), ast_free);
			get_instance_from_contact(contact_hdr, &extra_info);

			struct registrar_contact_uuid uuid_obj;
			uuid_obj.uri = pjsip_uri_get_uri(contact_hdr->uri);
            uuid_obj.uuid = extra_info.instance;
            uuid_obj.reg_id = extra_info.reg_id;

			extra_info.gswave_mobile = check_if_register_by_gswave_mobile(rdata);

			ast_verb(3, "Get the uuid '%s' reg_id %d from contact, gswave_mobile[%d].\n", uuid_obj.uuid ? uuid_obj.uuid : "null", uuid_obj.reg_id, extra_info.gswave_mobile);
			/* if there is the uuid of this contact ,delete it and add angin*/
			RAII_VAR(struct ast_sip_contact *, ws_contact, NULL, ao2_cleanup);
			struct wave_register_data data;
			memset(&data, 0, sizeof(struct wave_register_data));
			if (!ast_strlen_zero(aor_name)) {
				strncpy(data.aor_name , aor_name, sizeof(data.aor_name) - 1);
			}
			if (!ast_strlen_zero(user_agent)) {
				strncpy(data.user_agent, user_agent, sizeof(data.user_agent) - 1);
			}
			if (!ast_strlen_zero(extra_info.instance)) {
				strncpy(data.instance, extra_info.instance, sizeof(data.instance) - 1);
			}
			data.enable_duplicate_logins = endpoint ? endpoint->enable_duplicate_logins: 0;

			if (((ws_contact = get_exist_contact_for_ws(details.uri, &data, contacts, rdata)) != NULL)) {
				ast_verb(3, "Have the ws contact '%s', removed it.\n", ws_contact->uri);
				set_extra_info_for_remove_exsit_contact(&extra_info, ws_contact);
				ast_sip_location_delete_contact(ws_contact);
				ast_verb(3, "Removed ws contact '%s' from AOR '%s' due to request\n", ws_contact->uri, aor_name);
                ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
						"Contact: %s\r\n"
						"AOR: %s\r\n"
						"UserAgent: %s",
						ws_contact->uri,
						aor_name,
						user_agent);
#if 1 /*Modified by yfyang, for Bug 313577*/
			} else if ((strlen(extra_info.instance) && (contact = ao2_callback(contacts, OBJ_UNLINK, registrar_find_uuid, &uuid_obj)) != NULL)) {
				ast_verb(3, "Have the contact '%s', removed it.\n", contact->uri);
				set_extra_info_for_remove_exsit_contact(&extra_info, contact);
				ast_sip_location_delete_contact(contact);
				ast_verb(3, "Removed contact '%s' from AOR '%s' due to request\n", contact->uri, aor_name);
                ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
						"Contact: %s\r\n"
						"AOR: %s\r\n"
						"UserAgent: %s",
						contact->uri,
						aor_name,
						user_agent);
#endif
			} else if ((extra_info.gswave_mobile && (contact = ao2_callback(contacts, OBJ_UNLINK, registrar_find_gsmobile_contact, &data)) != NULL)) {
				ast_verb(3, "Have the GsWave mobile contact '%s', removed it.\n", contact->uri);
				set_extra_info_for_remove_exsit_contact(&extra_info, contact);
				ast_sip_location_delete_contact(contact);
                ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
						"Contact: %s\r\n"
						"AOR: %s\r\n"
						"UserAgent: %s",
						contact->uri,
						aor_name,
						user_agent);
			}
			/* if this is the new contact, we need to check the max_contacts */
			else if ((MAX(added - deleted, 0) + ao2_container_count(contacts)) > aor->max_contacts) {
				if (aor->max_contacts > 1) {
					ast_sip_report_failed_acl(endpoint, rdata, "registrar_attempt_exceeds_maximum_configured_contacts");
					ast_log(LOG_WARNING, "Registration attempt from endpoint '%s' to AOR '%s' will exceed max contacts of %u\n",
						ast_sorcery_object_get_id(endpoint), ast_sorcery_object_get_id(aor), aor->max_contacts);

					if (existing_contacts) {
						ao2_ref(existing_contacts, -1);
						existing_contacts = NULL;
					}

					response->code = 403;
					response->reason = AST_CAUSE_CONFIGURED_CONTACTS_REACHED;
					pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
					return;
				} else {
					if (aor->remove_existing) {
						ast_log(LOG_NOTICE, "Max-Contact is 1, endpoint '%s' register to AOR '%s', replace the first contact\n",
						ast_sorcery_object_get_id(endpoint), ast_sorcery_object_get_id(aor));
						replace_first = 1; 
					} else {
						ast_sip_report_failed_acl(endpoint, rdata, "registrar_attempt_exceeds_maximum_configured_contacts");
						ast_log(LOG_NOTICE, "Registration attempt from endpoint '%s' to AOR '%s' has exceed max contacts %u, not remove existing.\n",
							ast_sorcery_object_get_id(endpoint), ast_sorcery_object_get_id(aor), aor->max_contacts);

						if (existing_contacts) {
							ao2_ref(existing_contacts, -1);
							existing_contacts = NULL;
						}

						response->code = 403;
						response->reason = AST_CAUSE_CONFIGURED_CONTACTS_REACHED;
						pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
						return;
					}
				}
			}

			extra_info.support_ice = check_support_ice(contact_hdr);
			extra_info.support_trickle_ice = check_support_trickle_ice(contact_hdr);

			if (extra_info.ssh_proxy_flag) {
				/* check the max registrar num */
				if (ast_sip_check_proxy_user_num_limit()) {
					if (existing_contacts) {
						ao2_ref(existing_contacts, -1);
						existing_contacts = NULL;
					}

					response->code = 403;
					response->reason = AST_CAUSE_TURN_REGISTRAR_LIMITED;
					pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
					return;
				}
			}
#endif

			prune_on_boot = !ast_sip_will_uri_survive_restart(details.uri, endpoint, rdata);

#ifdef GRANDSTREAM_NETWORKS
			original_address = get_contact_original_address(rdata);
			set_to_host_addr_extra_info(&extra_info, rdata);
#endif
			contact = ast_sip_location_create_contact(aor, contact_uri,
				ast_tvadd(ast_tvnow(), ast_samp2tv(expiration, 1)),
				path_str ? ast_str_buffer(path_str) : NULL,
#ifdef GRANDSTREAM_NETWORKS
				user_agent, via_addr, via_port, call_id, prune_on_boot, endpoint, gs_appid,
				original_address, details.sbc_endpoint_addr,rdata,extra_info);
			if (original_address) {
				ast_free(original_address);
				original_address = NULL;
			}
#else
				user_agent, via_addr, via_port, call_id, prune_on_boot, endpoint);
#endif
			if (!contact) {
				ast_log(LOG_ERROR, "Unable to bind contact '%s' to AOR '%s'\n",
					contact_uri, aor_name);
#ifdef GRANDSTREAM_NETWORKS
				if (extra_info.ssh_proxy_flag) {
					if (existing_contacts) {
						ao2_ref(existing_contacts, -1);
						existing_contacts = NULL;
					}

					response->code = 403;
					pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
					return;
				}
#endif
				continue;
			}

			if (prune_on_boot) {
				const char *contact_name;
				struct contact_transport_monitor *monitor;

				/*
				 * Monitor the transport in case it gets disconnected because
				 * the contact won't be valid anymore if that happens.
				 */
				contact_name = ast_sorcery_object_get_id(contact);
				monitor = ao2_alloc(sizeof(*monitor) + 2 + strlen(aor_name)
					+ strlen(contact_name), NULL);
				if (monitor) {
					strcpy(monitor->aor_name, aor_name);/* Safe */
					monitor->contact_name = monitor->aor_name + strlen(aor_name) + 1;
					strcpy(monitor->contact_name, contact_name);/* Safe */

					ast_sip_transport_monitor_register(rdata->tp_info.transport,
						register_contact_transport_shutdown_cb, monitor);
					ao2_ref(monitor, -1);
				}
			}
			ast_verb(3, "Added contact '%s' to AOR '%s' with expiration of %d seconds\n",
				contact_uri, aor_name, expiration);
			memset(rdata->tp_info.transport->aor_name, 0, sizeof(rdata->tp_info.transport->aor_name));
			strcpy(rdata->tp_info.transport->aor_name,aor_name);
			if(contact->gswave_mobile && ast_ha_and_redundance())
			{
				ast_verb(3, "aor:%s, tp:%p add to tp queue!",aor_name, rdata->tp_info.transport);
				ast_wave_mobile_tp_enqueue(&g_wave_mobile_watch_queue, (char *)aor_name,  rdata->tp_info.transport);
			}
			ast_test_suite_event_notify("AOR_CONTACT_ADDED",
					"Contact: %s\r\n"
					"AOR: %s\r\n"
					"Expiration: %d\r\n"
					"UserAgent: %s",
					contact_uri,
					aor_name,
					expiration,
					user_agent);

			ao2_link(contacts, contact);
		} else if (expiration) {
#ifdef GRANDSTREAM_NETWORKS
			if (contact_count > aor->max_contacts) {
				/* Enforce the maximum number of contacts */
				ast_sip_report_failed_acl(endpoint, rdata, "registrar_attempt_exceeds_maximum_configured_contacts.");

				if (existing_contacts) {
					ao2_ref(existing_contacts, -1);
				}

				ast_sip_location_delete_contact(contact);
				pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);
				response->code = 403;
				response->reason = AST_CAUSE_CONFIGURED_CONTACTS_REACHED;

				return;
			}
#endif

			struct ast_sip_contact *contact_update;

			contact_update = ast_sorcery_copy(ast_sip_get_sorcery(), contact);
			if (!contact_update) {
				ast_log(LOG_ERROR, "Failed to update contact '%s' expiration time to %d seconds.\n",
					contact->uri, expiration);
				continue;
			}

			contact_update->expiration_time = ast_tvadd(ast_tvnow(), ast_samp2tv(expiration, 1));
			contact_update->qualify_frequency = aor->qualify_frequency;
			contact_update->authenticate_qualify = aor->authenticate_qualify;

			if (path_str) {
				ast_string_field_set(contact_update, path, ast_str_buffer(path_str));
			}
			if (user_agent) {
				ast_string_field_set(contact_update, user_agent, user_agent);
			}
			if (!ast_strlen_zero(ast_config_AST_SYSTEM_NAME)) {
				ast_string_field_set(contact_update, reg_server, ast_config_AST_SYSTEM_NAME);
			}

#ifdef GRANDSTREAM_NETWORKS
			contact_update->support_missed_call = extra_info.support_missed_call;
			contact_update->support_gs_message = extra_info.support_gs_message;
			contact_update->support_gs_conf_invite = extra_info.support_gs_conf_invite;
			if (!ast_strlen_zero(extra_info.instance)) {
				ast_string_field_set(contact_update, instance, extra_info.instance);
			}

#ifdef BUILD_ON_CLOUD_PBX
			ast_string_field_set(contact_update, uri, contact_uri);
#endif
#endif

			if (ast_sip_location_update_contact(contact_update)) {
				ast_log(LOG_ERROR, "Failed to update contact '%s' expiration time to %d seconds.\n",
					contact->uri, expiration);
				ast_sip_location_delete_contact(contact);
				continue;
			}

			/*Start added by zyxia for WakeUp phone gswave，2018.7.23 */
#ifdef GRANDSTREAM_NETWORKS
			//need check the register message is refresh register or push register,if the contact isn't changed,dont refresh
			if(contact->gswave_mobile && !ast_strlen_zero(contact_update->instance)){
				char key[128]={0},key_value[1024]={0};
				struct sip_waveapp_info* app_info = NULL;
				struct timeval register_time;
				register_time = ast_tvnow();

				//old app instance have <urn:uuid:00000000-0000,if not have ,is new app
				if(!ast_strlen_zero(contact_update->instance) && strstr(contact_update->instance,"<urn:uuid:00000000-0000") && !ast_strlen_zero(gs_appid)) {
					memset(key, 0, sizeof(key));
					snprintf(key,sizeof(key),"%s/%s",endpoint->aors,gs_appid);
					app_info = ast_sip_find_waveapp_info(endpoint, gs_appid);
				}
				//new app
				else if(!ast_strlen_zero(contact->instance) && !strstr(contact_update->instance,"<urn:uuid:00000000-0000")) {
					memset(key, 0, sizeof(key));
					snprintf(key,sizeof(key),"%s/%s",endpoint->aors,contact_update->instance);
					app_info = ast_sip_find_waveapp_info(endpoint, contact_update->instance);
				}

				if (app_info != NULL) {
					/*update it*/
					ast_string_field_set(app_info, gs_uri, contact->uri);
					ast_string_field_set(app_info, gs_new_contact_uri, "");
					app_info->conf_invite = contact_update->support_gs_conf_invite;
					app_info->register_time = register_time;
					ast_log(LOG_DEBUG, "app clientId %s  already add to endpoint %s",
					contact_update->instance, ast_sorcery_object_get_id(endpoint));
				}

				memset(key_value,0,sizeof(key_value));
				snprintf(key_value, sizeof(key_value), "%s,%d,%d",contact->uri, (int)register_time.tv_sec,contact_update->support_gs_conf_invite);
				ast_db_put(GSWAVE_DB_FAMILY, key, key_value);
			}
#endif
			/*End added by zyxia for WakeUp phone gswave，2018.7.23 */

			ast_debug(3, "Refreshed contact '%s' on AOR '%s' with new expiration of %d seconds\n",
				contact_uri, aor_name, expiration);
			ast_test_suite_event_notify("AOR_CONTACT_REFRESHED",
					"Contact: %s\r\n"
					"AOR: %s\r\n"
					"Expiration: %d\r\n"
					"UserAgent: %s",
					contact_uri,
					aor_name,
					expiration,
					contact_update->user_agent);
			ao2_link(contacts, contact_update);
			ao2_cleanup(contact_update);
		} else {
			if (contact->prune_on_boot) {
				struct contact_transport_monitor *monitor;
				const char *contact_name =
					ast_sorcery_object_get_id(contact);

				monitor = ast_alloca(sizeof(*monitor) + 2 + strlen(aor_name)
					+ strlen(contact_name));
				strcpy(monitor->aor_name, aor_name);/* Safe */
				monitor->contact_name = monitor->aor_name + strlen(aor_name) + 1;
				strcpy(monitor->contact_name, contact_name);/* Safe */

				ast_sip_transport_monitor_unregister(rdata->tp_info.transport,
					register_contact_transport_shutdown_cb, monitor, contact_transport_monitor_matcher);
			}
			/*Start added by zyxia for WakeUp phone gswave，2018.7.23 */
			/* if unregister ,remove data in endpoint and database*/
#ifdef GRANDSTREAM_NETWORKS
			if(contact->gswave_mobile ){
				struct sip_waveapp_info* app_info = NULL;
				app_info = ast_sip_find_waveapp_info_by_uri(endpoint,contact_uri);

				if(app_info != NULL){
					ast_sip_remove_waveinfo_from_database(aor_name, app_info);
					ast_sip_clear_waveapp_info(endpoint,app_info->client_uniqueId);
				}

				ast_verb(3, "Removed Gswave data  from	endpoint '%s' due to request\n",  aor_name);

				if(ast_ha_and_redundance())
				{
					ast_debug(3,"aor:%s quit tp queue",aor_name);
					ast_wave_mobile_tp_quitqueue(&g_wave_mobile_watch_queue, (char *)aor_name);
				}
				
			}
			
			memset(rdata->tp_info.transport->aor_name,0, sizeof(rdata->tp_info.transport->aor_name));
#endif
			/*End added by zyxia for WakeUp phone gswave，2018.7.23 */

			/* We want to report the user agent that was actually in the removed contact */
			ast_sip_location_delete_contact(contact);

			ast_verb(3, "Removed contact '%s' from AOR '%s' due to request\n", contact_uri, aor_name);
			ast_test_suite_event_notify("AOR_CONTACT_REMOVED",
					"Contact: %s\r\n"
					"AOR: %s\r\n"
					"UserAgent: %s",
					contact_uri,
					aor_name,
					contact->user_agent);
		}
#ifdef GRANDSTREAM_NETWORKS
		if (!extra_info.support_gs_config && extra_info.gswave_mobile && contact && expiration) {
			ast_log(LOG_DEBUG, "It's gswave mobile, send turn info\n");
			send_notify_to_update_turn_info(contact);
		}
#endif
	}

	pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), details.pool);

	/*
	 * If the AOR is configured to remove any contacts over max_contacts
	 * that have not been updated/added/deleted as a result of this
	 * REGISTER do so.
	 *
	 * The existing contacts container holds all contacts that were not
	 * involved in this REGISTER.
	 * The contacts container holds the current contacts of the AOR.
	 */
	if (aor->remove_existing && existing_contacts) {
		/* Total contacts after this registration */
		contact_count = ao2_container_count(existing_contacts) + updated + added;
		if (contact_count > aor->max_contacts) {
			/* Remove excess existing contacts that expire the soonest */
			remove_excess_contacts(existing_contacts, contacts, contact_count - aor->max_contacts);
		}
		ao2_ref(existing_contacts, -1);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (replace_first) {
		/*Start added by zyxia for Bug 122155 ios gswave，2018.9.30 */
		char gs_appid_key[64],gs_appid_db[256];
		/*get gs_appid from astdb*/
		snprintf(gs_appid_key,sizeof(gs_appid_key),"%s/%s","gs_appid",aor_name);
		if(ast_strlen_zero(gs_appid) && !ast_db_get(GSWAVE_DB_FAMILY, gs_appid_key, gs_appid_db, sizeof(gs_appid_db)) && !ast_strlen_zero(gs_appid_db)){
			//ast_sip_delete_gswave_data(task_data->endpoint,aor_name);
			char register_key[64]={0},gs_uri_key[64]={0},gs_conf_invite_key[64]={0};

			snprintf(gs_appid_key,sizeof(gs_appid_key),"%s/%s","gs_appid",aor_name);
			snprintf(gs_uri_key,sizeof(gs_uri_key),"%s/%s","gs_uri",aor_name);
			snprintf(register_key,sizeof(register_key),"%s/%s","register_time",aor_name);
			snprintf(gs_conf_invite_key,sizeof(gs_conf_invite_key),"%s/%s","conf_invite",aor_name);

			//gswave didn't use long time,del it.
			ast_db_del(GSWAVE_DB_FAMILY, gs_appid_key);
			ast_db_del(GSWAVE_DB_FAMILY, register_key);
			ast_db_del(GSWAVE_DB_FAMILY, gs_uri_key);
			ast_db_del(GSWAVE_DB_FAMILY, gs_conf_invite_key);
				
			ast_get_timeval(0, &endpoint->register_time, ast_tv(0, 0), NULL);
			ast_string_field_set(endpoint, gs_appid,"");
			ast_string_field_set(endpoint, gs_uri,"");
			ast_string_field_set(endpoint, gs_new_contact_uri,"");
			ast_verb(3, "Removed Gswave data  from	endpoint '%s' due to max contacts\n",  aor_name);
		}
		/*End added by zyxia for Bug 122155 ios gswave，2018.9.30	*/
		
		ao2_callback(contacts, OBJ_NODATA|OBJ_UNLINK, registrar_delete_first_contact, (void *)aor_name);		
	}
#endif

	response_contact = ao2_callback(contacts, 0, NULL, NULL);

	/* Send a response containing all of the contacts (including static) that are present on this AOR */
	if (ast_sip_create_response(rdata, 200, response_contact, &tdata) != PJ_SUCCESS) {
		ao2_cleanup(response_contact);
		ao2_cleanup(contacts);
		response->code = 500;
		return;
	}
	ao2_cleanup(response_contact);

	/* Add the date header to the response, some UAs use this to set their date and time */
	ast_sip_add_date_header(tdata);

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_add_header(tdata, "X-GS-Config", "LDAP-Settings, featureCode, turnInfo, basic, plugin, phonebook");

	if ((aor->max_contacts > 1) && expires && (0 != expires->ivalue)) {
		registrar_add_access_info_header(tdata);
	}

	registrar_add_support_header(tdata, extra_info);
#endif

	ao2_callback(contacts, 0, registrar_add_contact, tdata);

	if ((expires_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_EXPIRES, NULL))) {
		expires_hdr = pjsip_expires_hdr_create(tdata->pool, registrar_get_expiration(aor, NULL, rdata));
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr*)expires_hdr);
	}

	response->tdata = tdata;
}

static int register_aor(pjsip_rx_data *rdata,
	struct ast_sip_endpoint *endpoint,
	struct ast_sip_aor *aor,
	const char *aor_name)
{
	struct aor_core_response response = {
		.code = 500,
#ifdef GRANDSTREAM_NETWORKS
		.reason = 0,
#endif
	};
	struct ao2_container *contacts = NULL;

	ao2_lock(aor);
	contacts = ast_sip_location_retrieve_aor_contacts_nolock(aor);
	if (!contacts) {
		ao2_unlock(aor);
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(),
			rdata, response.code, NULL, NULL, NULL);
		return PJ_TRUE;
	}

	register_aor_core(rdata, endpoint, aor, aor_name, contacts, &response);

	ao2_cleanup(contacts);
	ao2_unlock(aor);

	/* Now send the REGISTER response to the peer */
	if (response.tdata) {
		ast_sip_send_stateful_response(rdata, response.tdata, endpoint);
	} else {
#ifdef GRANDSTREAM_NETWORKS
		if (response.code == 423) {
			pjsip_hdr              hdr_list;
			pjsip_min_expires_hdr *hdr_minexpires;

			if (!(hdr_minexpires = pjsip_min_expires_hdr_create(rdata->tp_info.pool, aor->minimum_expiration))) {
				ast_log(LOG_WARNING, "create Min-Expires header failed\n");
			} else {
				pj_list_init(&hdr_list);
				pj_list_push_back(&hdr_list, hdr_minexpires);
				pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 423, NULL, &hdr_list, NULL);
				return PJ_TRUE;
			}
		}

		if ((response.code == 403) && (response.reason == AST_CAUSE_TURN_REGISTRAR_LIMITED)) {
			static const pj_str_t pj_reason = { "Reason", 6 };
			static const pj_str_t pj_value  = { "SIP;cause=403;text=\"Max remote registrar reached\"", 49 };

			pjsip_generic_string_hdr *reason_hdr = pjsip_generic_string_hdr_create(rdata->tp_info.pool, &pj_reason, &pj_value);
			if (reason_hdr) {
				pjsip_hdr  hdr_list;
				pj_list_init(&hdr_list);
				pj_list_push_back(&hdr_list, reason_hdr);
				pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 403, NULL, &hdr_list, NULL);
				return PJ_TRUE;
			}
		}

		if ((response.code == 403) && (response.reason == AST_CAUSE_CONFIGURED_CONTACTS_REACHED)) {
			static const pj_str_t pj_reason = { "Reason", 6 };
			static const pj_str_t pj_value  = { "SIP;cause=403;text=\"Max configured contacts reached\"", 52 };

			pjsip_generic_string_hdr *reason_hdr = pjsip_generic_string_hdr_create(rdata->tp_info.pool, &pj_reason, &pj_value);
			if (reason_hdr) {
				pjsip_hdr  hdr_list;
				pj_list_init(&hdr_list);
				pj_list_push_back(&hdr_list, reason_hdr);
				pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 403, NULL, &hdr_list, NULL);
				return PJ_TRUE;
			}
		}
#endif
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(),
			rdata, response.code, NULL, NULL, NULL);
	}
	return PJ_TRUE;
}

static int match_aor(const char *aor_name, const char *id)
{
	if (ast_strlen_zero(aor_name)) {
		return 0;
	}

	if (!strcmp(aor_name, id)) {
		ast_debug(3, "Matched id '%s' to aor '%s'\n", id, aor_name);
		return 1;
	}

	return 0;
}

static char *find_aor_name(const char *username, const char *domain, const char *aors)
{
	char *configured_aors;
	char *aors_buf;
	char *aor_name;
	char *id_domain;
	struct ast_sip_domain_alias *alias;

	id_domain = ast_alloca(strlen(username) + strlen(domain) + 2);
	sprintf(id_domain, "%s@%s", username, domain);

	aors_buf = ast_strdupa(aors);

	/* Look for exact match on username@domain */
	configured_aors = aors_buf;
	while ((aor_name = ast_strip(strsep(&configured_aors, ",")))) {
		if (match_aor(aor_name, id_domain)) {
			return ast_strdup(aor_name);
		}
	}

	/* If there's a domain alias, look for exact match on username@domain_alias */
	alias = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "domain_alias", domain);
	if (alias) {
		char *id_domain_alias = ast_alloca(strlen(username) + strlen(alias->domain) + 2);

		sprintf(id_domain, "%s@%s", username, alias->domain);
		ao2_cleanup(alias);

		configured_aors = strcpy(aors_buf, aors);/* Safe */
		while ((aor_name = ast_strip(strsep(&configured_aors, ",")))) {
			if (match_aor(aor_name, id_domain_alias)) {
				return ast_strdup(aor_name);
			}
		}
	}

	if (ast_strlen_zero(username)) {
		/* No username, no match */
		return NULL;
	}

	/* Look for exact match on username only */
	configured_aors = strcpy(aors_buf, aors);/* Safe */
	while ((aor_name = ast_strip(strsep(&configured_aors, ",")))) {
		if (match_aor(aor_name, username)) {
			return ast_strdup(aor_name);
		}
	}

	return NULL;
}

static struct ast_sip_aor *find_registrar_aor(struct pjsip_rx_data *rdata, struct ast_sip_endpoint *endpoint)
{
	struct ast_sip_aor *aor = NULL;
	char *aor_name = NULL;
	char *domain_name;
	char *username = NULL;
	int i;

#ifdef GRANDSTREAM_NETWORKS
	if (endpoint && !ast_strlen_zero(endpoint->trunk_type) && !strncmp(endpoint->trunk_type, "account", 7)) {
		aor = ast_sip_location_retrieve_aor(ast_sorcery_object_get_id(endpoint));
		return aor;
	}
#endif

	for (i = 0; i < AST_VECTOR_SIZE(&endpoint->ident_method_order); ++i) {
		pjsip_sip_uri *uri;
		pjsip_authorization_hdr *header = NULL;

		switch (AST_VECTOR_GET(&endpoint->ident_method_order, i)) {
		case AST_SIP_ENDPOINT_IDENTIFY_BY_USERNAME:
			uri = pjsip_uri_get_uri(rdata->msg_info.to->uri);

			domain_name = ast_alloca(uri->host.slen + 1);
			ast_copy_pj_str(domain_name, &uri->host, uri->host.slen + 1);
			username = ast_alloca(uri->user.slen + 1);
			ast_copy_pj_str(username, &uri->user, uri->user.slen + 1);

			/*
			 * We may want to match without any user options getting
			 * in the way.
			 */
			AST_SIP_USER_OPTIONS_TRUNCATE_CHECK(username);

			aor_name = find_aor_name(username, domain_name, endpoint->aors);
			if (aor_name) {
				ast_debug(3, "Matched aor '%s' by To username\n", aor_name);
			}
			break;
		case AST_SIP_ENDPOINT_IDENTIFY_BY_AUTH_USERNAME:
			while ((header = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_AUTHORIZATION,
				header ? header->next : NULL))) {
				if (header && !pj_stricmp2(&header->scheme, "digest")) {
					username = ast_alloca(header->credential.digest.username.slen + 1);
					ast_copy_pj_str(username, &header->credential.digest.username, header->credential.digest.username.slen + 1);
					domain_name = ast_alloca(header->credential.digest.realm.slen + 1);
					ast_copy_pj_str(domain_name, &header->credential.digest.realm, header->credential.digest.realm.slen + 1);

					aor_name = find_aor_name(username, domain_name, endpoint->aors);
					if (aor_name) {
						ast_debug(3, "Matched aor '%s' by Authentication username\n", aor_name);
						break;
					}
				}
			}
			break;
		default:
			continue;
		}

		if (aor_name) {
			break;
		}
	}

	if (ast_strlen_zero(aor_name) || !(aor = ast_sip_location_retrieve_aor(aor_name))) {
		/* The provided AOR name was not found (be it within the configuration or sorcery itself) */
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 404, NULL, NULL, NULL);
		ast_sip_report_req_no_support(endpoint, rdata, "registrar_requested_aor_not_found");
		ast_log(LOG_WARNING, "AOR '%s' not found for endpoint '%s'\n",
			username ?: "", ast_sorcery_object_get_id(endpoint));
	}
	ast_free(aor_name);
	return aor;
}

static pj_bool_t registrar_on_rx_request(struct pjsip_rx_data *rdata)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint,
		 ast_pjsip_rdata_get_endpoint(rdata), ao2_cleanup);
	struct ast_sip_aor *aor;
	const char *aor_name;

	if (pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_register_method) || !endpoint) {
		return PJ_FALSE;
	}

	if (ast_strlen_zero(endpoint->aors)) {
		/* Short circuit early if the endpoint has no AORs configured on it, which means no registration possible */
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 403, NULL, NULL, NULL);
		ast_sip_report_failed_acl(endpoint, rdata, "registrar_attempt_without_configured_aors");
		ast_log(LOG_WARNING, "Endpoint '%s' has no configured AORs\n", ast_sorcery_object_get_id(endpoint));
		return PJ_TRUE;
	}

	if (!PJSIP_URI_SCHEME_IS_SIP(rdata->msg_info.to->uri) && !PJSIP_URI_SCHEME_IS_SIPS(rdata->msg_info.to->uri)) {
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 416, NULL, NULL, NULL);
		ast_sip_report_failed_acl(endpoint, rdata, "registrar_invalid_uri_in_to_received");
		ast_log(LOG_WARNING, "Endpoint '%s' attempted to register to an AOR with a non-SIP URI\n", ast_sorcery_object_get_id(endpoint));
		return PJ_TRUE;
	}

	aor = find_registrar_aor(rdata, endpoint);
	if (!aor) {
		/* We've already responded about not finding an AOR. */
		return PJ_TRUE;
	}

	aor_name = ast_sorcery_object_get_id(aor);

	if (!aor->max_contacts) {
		/* Registration is not permitted for this AOR */
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 403, NULL, NULL, NULL);
		ast_sip_report_req_no_support(endpoint, rdata, "registrar_attempt_without_registration_permitted");
		ast_log(LOG_WARNING, "AOR '%s' has no configured max_contacts. Endpoint '%s' unable to register\n",
			aor_name, ast_sorcery_object_get_id(endpoint));
	} else {
		register_aor(rdata, endpoint, aor, aor_name);
	}
	ao2_ref(aor, -1);
	return PJ_TRUE;
}

/* function pointer to callback needs to be within the module
   in order to avoid problems with an undefined symbol */
static int sip_contact_to_str(void *acp, void *arg, int flags)
{
	return ast_sip_contact_to_str(acp, arg, flags);
}

static int ami_registrations_aor(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;
	struct ast_sip_ami *ami = arg;
	int *count = ami->arg;
	RAII_VAR(struct ast_str *, buf,
		 ast_sip_create_ami_event("InboundRegistrationDetail", ami), ast_free);

	if (!buf) {
		return -1;
	}

	ast_sip_sorcery_object_to_ami(aor, &buf);
	ast_str_append(&buf, 0, "Contacts: ");
	ast_sip_for_each_contact(aor, sip_contact_to_str, &buf);
	ast_str_append(&buf, 0, "\r\n");

	astman_append(ami->s, "%s\r\n", ast_str_buffer(buf));
	(*count)++;
	return 0;
}

static int ami_registrations_endpoint(void *obj, void *arg, int flags)
{
	struct ast_sip_endpoint *endpoint = obj;
	return ast_sip_for_each_aor(
		endpoint->aors, ami_registrations_aor, arg);
}

static int ami_registrations_endpoints(void *arg)
{
	RAII_VAR(struct ao2_container *, endpoints,
		 ast_sip_get_endpoints(), ao2_cleanup);

	if (!endpoints) {
		return 0;
	}

	ao2_callback(endpoints, OBJ_NODATA, ami_registrations_endpoint, arg);
	return 0;
}

static int ami_show_registrations(struct mansession *s, const struct message *m)
{
	int count = 0;
	struct ast_sip_ami ami = { .s = s, .m = m, .arg = &count, .action_id = astman_get_header(m, "ActionID"), };

	astman_send_listack(s, m, "Following are Events for each Inbound registration",
		"start");

	ami_registrations_endpoints(&ami);

	astman_send_list_complete_start(s, m, "InboundRegistrationDetailComplete", count);
	astman_send_list_complete_end(s);
	return 0;
}

static int ami_show_registration_contact_statuses(struct mansession *s, const struct message *m)
{
	int count = 0;
	struct ast_sip_ami ami = { .s = s, .m = m, .arg = NULL, .action_id = astman_get_header(m, "ActionID"), };
	struct ao2_container *contacts = ast_sorcery_retrieve_by_fields(
		ast_sip_get_sorcery(), "contact", AST_RETRIEVE_FLAG_MULTIPLE | AST_RETRIEVE_FLAG_ALL, NULL);
	struct ao2_iterator i;
	struct ast_sip_contact *contact;

	astman_send_listack(s, m, "Following are ContactStatusEvents for each Inbound "
			    "registration", "start");

	if (contacts) {
		i = ao2_iterator_init(contacts, 0);
		while ((contact = ao2_iterator_next(&i))) {
			struct ast_sip_contact_wrapper wrapper;

			wrapper.aor_id = (char *)contact->aor;
			wrapper.contact = contact;
			wrapper.contact_id = (char *)ast_sorcery_object_get_id(contact);

			ast_sip_format_contact_ami(&wrapper, &ami, 0);
			count++;

			ao2_ref(contact, -1);
		}
		ao2_iterator_destroy(&i);
		ao2_ref(contacts, -1);
	}

	astman_send_list_complete_start(s, m, "ContactStatusDetailComplete", count);
	astman_send_list_complete_end(s);
	return 0;
}

#define AMI_SHOW_REGISTRATION_CONTACT_STATUSES "PJSIPShowRegistrationInboundContactStatuses"
#define AMI_SHOW_REGISTRATIONS "PJSIPShowRegistrationsInbound"

static pjsip_module registrar_module = {
	.name = { "Registrar", 9 },
	.id = -1,
	.priority = PJSIP_MOD_PRIORITY_APPLICATION,
	.on_rx_request = registrar_on_rx_request,
};


#ifdef GRANDSTREAM_NETWORKS

/* Added by yfyang, for HA service data sync */
static struct stasis_message_router *ha_reg_data_router = NULL;

/*! data format, for example: 
{"contact": "1605;@sip:1605@192.168.129.166:5066" 
 "value": {"qualify_timeout":"0.000000",
    	   "expiration_time":"1543459416",
    	   "path":"",
    	   "qualify_frequency":"0",
    	   "rmv_obp_from_route":"false",
    	   "support_missed_call":"false",
    	   "outbound_proxy":"",
    	   "uri":"sip:1605@192.168.129.103:5070",
    	   "uuid":"00000000-0000-1000-8000-000B8248EC41",
		   "user_agent":"Grandstream GXP2200 1.0.3.27" }
} 
*/ 
static void ha_reg_data_put_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	//struct ast_json			*val_object  = NULL;
	const char				*contact_id  = NULL;
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	const char *val_str    = NULL;
	const char *uri        = NULL;
	const char *path_info  = NULL;
	const char *user_agent = NULL;
	const char *uuid       = NULL;
	char  *user_id  = NULL;
	char  *ptr      = NULL;
	char  *val_buf  = NULL;
	char   delim[]  = ","; 
	char   delim1[] = ":"; 
	char  *token    = NULL;
	char  *save_ptr = NULL;
	char  *key      = NULL;
	char  *value0   = NULL;
	char  *instance = NULL;
    int   reg_id = 0;
	int support_missed_call = 0;
	int support_ldap_settings = 0;
	unsigned int support_gs_config = 0;
	unsigned int support_gs_state = 0;
	unsigned int support_gs_fec = 0;
	unsigned int support_gs_message = 0;
	unsigned int require_feature_code_settings = 0;
	unsigned int require_turn_info_settings = 0;
	unsigned int require_basic_settings = 0;
	unsigned int require_plugin = 0;
	unsigned int require_phonebook = 0;
	unsigned int support_ice = 0;
	unsigned int support_trickle_ice = 0;
	unsigned int support_gs_conf_invite = 0;
	unsigned int support_gs_control = 0;

	struct timeval expiration_time;

	ast_log(LOG_NOTICE, "ha_reg_data_put_cb, enter.....");
	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (!payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	contact_id = ast_json_string_get(ast_json_object_get(json_object, "contact"));
	val_str = ast_json_string_get(ast_json_object_get(json_object, "value"));
	//val_object = ast_json_object_get(json_object, "value");
	if (ast_strlen_zero(contact_id) || ast_strlen_zero(val_str)) {
		ast_log(LOG_WARNING, "Incomplete JSON object data"); 
		return;
	}

	ast_log(LOG_NOTICE, "ha_reg_data_put_cb, contact[%s], value[%s]", contact_id, val_str);

	/* get endpoint, aor */
	user_id = ast_strdupa(contact_id);
	if ((ptr = strchr(user_id, ';'))) {
		*ptr = '\0';
	}

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", user_id);
	aor = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "aor", user_id);
	if (!endpoint || !aor) {
		ast_log(LOG_WARNING, "Can't get endpoint or aor by id[%s]", user_id);
		return;
	}

	/* parse the value */
	val_buf = ast_strdupa(val_str);

	/* "qualify_timeout":"0.000000","expiration_time":"1543485628","rmv_obp_from_route":"false","uuid":"00000000-0000-1000-8000-000B823E1704","path":"","outbound_proxy":"",
	    "qualify_frequency":"0","user_agent":"Grandstream GXP2200 20.13.9.13","uri":"sip:5003@192.168.129.122:5064","support_missed_call":"false"}*/
	val_buf = ast_strip_quoted(val_buf, "{", "}");
	for (token = strtok_r(val_buf, delim, &save_ptr); token != NULL; token = strtok_r(NULL, delim, &save_ptr)) {
		key = strtok_r(token, delim1, &value0);
		if (key && value0) {
			key = ast_strip_quoted(key, "\"", "\"");
			value0 = ast_strip_quoted(value0, "\"", "\"");

			if (!strcmp(key, "expiration_time") && value0) {
				expiration_time = ast_samp2tv(atoi(value0), 1);
			} else if (!strcmp(key, "uuid")) {
				uuid = value0;
			} else if (!strcmp(key, "path")) {
				path_info = value0;
			} else if (!strcmp(key, "user_agent")) {
				user_agent = value0;
			} else if (!strcmp(key, "uri")) {
				uri = value0;
			} else if (!strcmp(key, "support_missed_call")) {
				support_missed_call = ast_true(value0);
			} else if (!strcmp(key, "support_ldap_settings")) {
				support_ldap_settings = ast_true(value0);
			} else if (!strcmp(key, "support_gs_config")) {
				support_gs_config = ast_true(value0);
			} else if (!strcmp(key, "support_gs_state")) {
				support_gs_state = ast_true(value0);
			} else if (!strcmp(key, "support_gs_fec")) {
				support_gs_fec = ast_true(value0);
			} else if (!strcmp(key, "require_feature_code_settings")) {
				require_feature_code_settings = ast_true(value0);
			} else if (!strcmp(key, "require_turn_info_settings")) {
				require_turn_info_settings = ast_true(value0);
			} else if (!strcmp(key, "require_basic_settings")) {
				require_basic_settings = ast_true(value0);
			} else if (!strcmp(key, "require_plugin")) {
				require_plugin = ast_true(value0);
			} else if (!strcmp(key, "require_phonebook")) {
				require_phonebook = ast_true(value0);
			} else if (!strcmp(key, "support_ice")) {
				support_ice = ast_true(value0);
			} else if (!strcmp(key, "support_trickle_ice")) {
				support_trickle_ice = ast_true(value0);
			} else if (!strcmp(key, "support_gs_message")) {
				support_gs_message = ast_true(value0);
			} else if (!strcmp(key, "support_gs_conf_invite")) {
				support_gs_conf_invite = ast_true(value0);
			} else if (!strcmp(key, "support_gs_control")) {
				support_gs_control = ast_true(value0);
			} else if (!strcmp(key, "instance")) {
				instance = value0;
			} else if (!strcmp(key, "reg_id")) {
				reg_id = atoi(value0);
			}
		}
	}

	contact = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "contact", contact_id);
	if (!contact) {
		/* Not add TCP/TLS contact */
		char *transport = NULL;
		if (!ast_strlen_zero(uri) && (transport = strstr(uri, "transport="))) {
			transport = transport + strlen("transport=");
			if (!strncmp(transport, "TCP", 3) || !strncmp(transport, "TLS", 3) || !strncmp(transport, "WS", 2) || !strncmp(transport, "ws", 2)) {
				ast_log(LOG_NOTICE, "Not sync register conatct with transport[%s]", transport);
				return;
			}
		}

		ast_log(LOG_NOTICE, "ha_reg_data_put_cb, to add contact[%s], uri[%s], instance[%s], reg_id[%d]", contact_id, uri, instance?instance:" ", reg_id);
		ast_sip_location_add_contact(endpoint, aor, uri, expiration_time, path_info, user_agent, instance, NULL, support_missed_call, support_ldap_settings, support_gs_message,support_gs_conf_invite, reg_id);
		return;
	}

	/* update contact */
	ast_string_field_set(contact, uri, uri);
	contact->expiration_time = expiration_time;
	contact->qualify_frequency = aor->qualify_frequency;
	contact->qualify_timeout = aor->qualify_timeout;
	contact->authenticate_qualify = aor->authenticate_qualify;
	contact->support_missed_call = support_missed_call;
	contact->support_ldap_settings = support_ldap_settings;
	contact->support_gs_config = support_gs_config;
	contact->support_gs_state = support_gs_state;
	contact->support_gs_fec = support_gs_fec;
	contact->require_feature_code_settings = require_feature_code_settings;
	contact->require_turn_info_settings = require_turn_info_settings;
	contact->require_basic_settings = require_basic_settings;
	contact->require_plugin = require_plugin;
	contact->require_phonebook = require_phonebook;
	contact->support_ice = support_ice;
	contact->support_trickle_ice = support_trickle_ice;
	contact->support_gs_message = support_gs_message;
	contact->support_gs_conf_invite = support_gs_conf_invite;
	contact->support_gs_control = support_gs_control;
	contact->reg_id = reg_id;

	if (path_info && aor->support_path) {
		ast_string_field_set(contact, path, path_info);
	}

	if (!ast_strlen_zero(aor->outbound_proxy)) {
		ast_string_field_set(contact, outbound_proxy, aor->outbound_proxy);
		contact->rmv_obp_from_route = aor->rmv_obp_from_route;
	}

	if (!ast_strlen_zero(user_agent)) {
		ast_string_field_set(contact, user_agent, user_agent);
	}

	if (!ast_strlen_zero(uuid)) {
		ast_string_field_set(contact, uuid, uuid);
	}

	ast_log(LOG_NOTICE, "ha_reg_data_put_cb, update registrar contact[%s]", contact_id);
	ast_sip_location_update_contact(contact);

	return;
}

static void remove_aor_contacts(struct ast_sip_aor *aor)
{
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	struct ast_sip_contact *contact;
	struct ao2_iterator it_contacts;

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts || (ao2_container_count(contacts) == 0)) {
		return;
	}

	it_contacts = ao2_iterator_init(contacts, 0);
	for (; (contact = ao2_iterator_next(&it_contacts)); ao2_ref(contact, -1)) {
		ast_sip_location_delete_contact(contact);
	}
	ao2_iterator_destroy(&it_contacts);

	return;
}

/*! data format, for example: 
 {"contact": "1605;@sip:1605@192.168.129.166:5066"  or
 {"aor": "1605"}
*/ 
static void ha_reg_data_del_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	const char				*contact_id  = NULL;
	const char				*aor_id      = NULL;
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);

	ast_log(LOG_NOTICE, "ha_reg_data_del_cb, enter.....");
	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (!payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	contact_id = ast_json_string_get(ast_json_object_get(json_object, "contact"));
	if (!contact_id) {
		aor_id = ast_json_string_get(ast_json_object_get(json_object, "aor"));
		if (aor_id) {
			aor = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "aor", aor_id);
			if (!aor) {
				ast_log(LOG_WARNING, "No aor object[%s]", aor_id);
				return;
			}

			ast_log(LOG_NOTICE, "CALL remove_aor_contacts(%s)", aor_id);
			remove_aor_contacts(aor);
			return;
		}

		ast_log(LOG_WARNING, "No contact in JSON object data");
		return;
	}

	contact = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "contact", contact_id);
	if (!contact) {
		ast_log(LOG_WARNING, "No contact object[%s]", contact_id);
		return;
	}

	/* delete contact object */
	ast_sip_location_delete_contact(contact);
	ast_log(LOG_NOTICE, "ha_reg_data_del_cb, registrar contact[%s] removed", contact_id);

	return;
}

static int ha_registrar_stasis_route_init(void)
{
	struct stasis_topic * ha_reg_data_topic = ast_ha_sync_data_topic_all();
	if (!ha_reg_data_topic) {
		ast_log(LOG_WARNING, "Failed to get ast_ha_sync_data_topic_all\n");
		return -1;
	}

	ha_reg_data_router = stasis_message_router_create(ha_reg_data_topic);
	if (ha_reg_data_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create ha_reg_data_router\n");
		return -1;
	}

	if (stasis_message_router_add(ha_reg_data_router, ast_ha_registrar_contact_put_type(), ha_reg_data_put_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_registrar_contact_put_type\n");
		return -1;
	}

	if (stasis_message_router_add(ha_reg_data_router, ast_ha_registrar_contact_del_type(), ha_reg_data_del_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_registrar_contact_del_type\n");
		return -1;
	}

	return 0;
}

#endif

/*! \brief Thread keeping things alive */
static pthread_t check_thread = AST_PTHREADT_NULL;

/*! \brief The global interval at which to check for contact expiration */
static unsigned int check_interval;

/*! \brief Callback function which deletes a contact */
static int expire_contact(void *obj, void *arg, int flags)
{
	struct ast_sip_contact *contact = obj;
	struct ast_named_lock *lock;

	lock = ast_named_lock_get(AST_NAMED_LOCK_TYPE_MUTEX, "aor", contact->aor);
	if (!lock) {
		return 0;
	}

	/*
	 * We need to check the expiration again with the aor lock held
	 * in case another thread is attempting to renew the contact.
	 */
	ao2_lock(lock);
	if (ast_tvdiff_ms(ast_tvnow(), contact->expiration_time) > 0) {
		if (contact->prune_on_boot) {
			struct contact_transport_monitor *monitor;
			const char *contact_name = ast_sorcery_object_get_id(contact);

			monitor = ast_alloca(sizeof(*monitor) + 2 + strlen(contact->aor)
				+ strlen(contact_name));
			strcpy(monitor->aor_name, contact->aor);/* Safe */
			monitor->contact_name = monitor->aor_name + strlen(contact->aor) + 1;
			strcpy(monitor->contact_name, contact_name);/* Safe */

			ast_sip_transport_monitor_unregister_all(register_contact_transport_shutdown_cb,
				monitor, contact_transport_monitor_matcher);
		}
		ast_sip_location_delete_contact(contact);
	}
	ao2_unlock(lock);
	ast_named_lock_put(lock);

	return 0;
}

static void *check_expiration_thread(void *data)
{
	struct ao2_container *contacts;
	struct ast_variable *var;
	char *time = alloca(64);

	while (check_interval) {
		sleep(check_interval);

		sprintf(time, "%ld", ast_tvnow().tv_sec);
		var = ast_variable_new("expiration_time <=", time, "");

		ast_debug(4, "Woke up at %s  Interval: %d\n", time, check_interval);

		contacts = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "contact",
			AST_RETRIEVE_FLAG_MULTIPLE, var);

		ast_variables_destroy(var);
		if (contacts) {
			ast_debug(3, "Expiring %d contacts\n", ao2_container_count(contacts));
			ao2_callback(contacts, OBJ_NODATA, expire_contact, NULL);
			ao2_ref(contacts, -1);
		}
		if(1 == ast_get_reset_connect_state())
		{
			ast_debug(3, "start shutdown wave mobile transport \n");
			ast_reset_wave_mobile_connect(&g_wave_mobile_watch_queue);
		
        }
	}

	return NULL;
}

static void expiration_global_loaded(const char *object_type)
{
	check_interval = ast_sip_get_contact_expiration_check_interval();

	/* Observer calls are serialized so this is safe without it's own lock */
	if (check_interval) {
		if (check_thread == AST_PTHREADT_NULL) {
			if (ast_pthread_create_background(&check_thread, NULL, check_expiration_thread, NULL)) {
				ast_log(LOG_ERROR, "Could not create thread for checking contact expiration.\n");
				return;
			}
			ast_debug(3, "Interval = %d, starting thread\n", check_interval);
		}
	} else {
		if (check_thread != AST_PTHREADT_NULL) {
			pthread_kill(check_thread, SIGURG);
			pthread_join(check_thread, NULL);
			check_thread = AST_PTHREADT_NULL;
			ast_debug(3, "Interval = 0, shutting thread down\n");
		}
	}
}

/*! \brief Observer which is used to update our interval when the global setting changes */
static struct ast_sorcery_observer expiration_global_observer = {
	.loaded = expiration_global_loaded,
};

static int load_module(void)
{
	const pj_str_t STR_REGISTER = { "REGISTER", 8 };

	ast_pjproject_get_buildopt("PJ_MAX_HOSTNAME", "%d", &pj_max_hostname);
	/* As of pjproject 2.4.5, PJSIP_MAX_URL_SIZE isn't exposed yet but we try anyway. */
	ast_pjproject_get_buildopt("PJSIP_MAX_URL_SIZE", "%d", &pjsip_max_url_size);

	if (ast_sip_register_service(&registrar_module)) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (pjsip_endpt_add_capability(ast_sip_get_pjsip_endpoint(), NULL, PJSIP_H_ALLOW, NULL, 1, &STR_REGISTER) != PJ_SUCCESS) {
		ast_sip_unregister_service(&registrar_module);
		return AST_MODULE_LOAD_DECLINE;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ha_registrar_stasis_route_init()) {
		ast_sip_unregister_service(&registrar_module);
		return AST_MODULE_LOAD_DECLINE;
	}
	ast_wave_mobile_tp_queue_init(&g_wave_mobile_watch_queue);
#endif

	ast_manager_register_xml(AMI_SHOW_REGISTRATIONS, EVENT_FLAG_SYSTEM,
				 ami_show_registrations);
	ast_manager_register_xml(AMI_SHOW_REGISTRATION_CONTACT_STATUSES, EVENT_FLAG_SYSTEM,
				 ami_show_registration_contact_statuses);

	ast_sorcery_observer_add(ast_sip_get_sorcery(), "global", &expiration_global_observer);
	ast_sorcery_reload_object(ast_sip_get_sorcery(), "global");

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	if (check_thread != AST_PTHREADT_NULL) {
		check_interval = 0;
		pthread_kill(check_thread, SIGURG);
		pthread_join(check_thread, NULL);

		check_thread = AST_PTHREADT_NULL;
	}

	ast_sorcery_observer_remove(ast_sip_get_sorcery(), "global", &expiration_global_observer);

	ast_manager_unregister(AMI_SHOW_REGISTRATIONS);
	ast_manager_unregister(AMI_SHOW_REGISTRATION_CONTACT_STATUSES);
	ast_sip_unregister_service(&registrar_module);
	ast_sip_transport_monitor_unregister_all(register_contact_transport_shutdown_cb, NULL, NULL);

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for HA service data sync */
	if (ha_reg_data_router) {
		stasis_message_router_unsubscribe(ha_reg_data_router);
		ha_reg_data_router = NULL;
	}
	ast_wave_mobile_tp_queue_clean(&g_wave_mobile_watch_queue);
	ast_mutex_destroy(&(g_wave_mobile_watch_queue.lock));
#endif

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Registrar Support",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND - 3,
	.requires = "res_pjproject,res_pjsip",
);
#ifdef GRANDSTREAM_NETWORKS

void ast_wave_mobile_tp_queue_init(ast_wave_mobile_watch_queue *queue)
{
	if(queue == NULL)
	{
		return;
	}
    queue->head = queue->tail= NULL;
    queue->count_node = 0;

    ast_mutex_init(&queue->lock);
    return;
}

void ast_wave_mobile_tp_queue_clean(ast_wave_mobile_watch_queue *queue)
{
	ast_wave_mobile_watch_node *free_q_head_node = NULL;

	if (!queue)
	{
		return;
	}

	ast_mutex_lock(&queue->lock);
	while(queue->head != NULL)
	{
		free_q_head_node = queue->head->next;
		ast_free(queue->head);
		queue->head = free_q_head_node;
	}
	queue->head = queue->tail = NULL;
	queue->count_node = 0;
	ast_mutex_unlock(&queue->lock);
}

int ast_wave_mobile_tp_enqueue(ast_wave_mobile_watch_queue *q, char *aor, struct pjsip_transport *tp )
{
    ast_wave_mobile_watch_node *node = NULL;
	ast_wave_mobile_watch_node *loop = NULL;

	if(!q || !aor || !tp)
	{
		return 1;
	}

	ast_mutex_lock(&(q->lock));

	loop = q->head;
	while(loop)
	{
		if(!strcmp(loop->aor_name, aor))
		{
			ast_log(LOG_DEBUG," replace new mobile tp:%p for old tp:%p",tp, loop->tp);
			loop->tp = tp;
			ast_mutex_unlock(&(q->lock));
			return 0;
		}
		loop = loop->next;
	}
	
	/* insert new node */
    node = (ast_wave_mobile_watch_node *)ast_malloc(sizeof(ast_wave_mobile_watch_node));
    if (NULL == node)
    {
        ast_log(LOG_ERROR, "ast_wave_mobile_enqueue malloc error!");
        return 1;
    }
	memset(node, 0, sizeof(ast_wave_mobile_watch_node));
	strcpy(node->aor_name, aor);
	node->tp = tp;
	node->next =  NULL;
	
    if (0 == q->count_node)
    {
        q->head = q->tail = node;
        q->count_node = 1;
		node->prev = NULL;
    }
    else
    {
        node->prev = q->tail;
        q->tail->next = node;
        q->tail = node;
        q->count_node = q->count_node + 1;
    }
    ast_mutex_unlock(&(q->lock));
    return 0;
}



int ast_wave_mobile_tp_quitqueue(ast_wave_mobile_watch_queue *q, char *aor)
{
	ast_wave_mobile_watch_node* loop;

    if (!q || !aor)
    {
        return 1;
    }

    ast_mutex_lock(&(q->lock));

	loop = q->head;
	while(loop)
	{
		if(!strcmp(loop->aor_name, aor))
		{
			if(loop == q->head)
			{
				if(loop->next){
					loop->next->prev = NULL;
				}
				q->head = loop->next;
				if(loop == q->tail)
				{
					q->tail = NULL;
				}
				ast_free(loop);
				q->count_node = q->count_node - 1;
			}
			else
			{
				if(loop == q->tail)
				{
					q->tail = loop->prev;
				}
				if(loop->next){
					loop->next->prev = loop->prev;
				}
				loop->prev->next = loop->next;
				ast_free(loop);
				q->count_node = q->count_node - 1;
				
			}
			break;
		}
		loop = loop->next;
	}

    ast_mutex_unlock(&(q->lock));

    return 0;
}

void ast_reset_wave_mobile_connect(ast_wave_mobile_watch_queue *q)
{
	ast_wave_mobile_watch_node* loop;

	if(NULL == q)
	{
		return;
	}
	ast_mutex_lock(&(q->lock));
	loop = q->head;
	while(loop)
	{
		if(!ast_search_relogin_queue(loop->aor_name))
		{
			if(!loop->tp->tpmgr || !strstr(loop->tp->obj_name,"tlss"))
			{
				ast_debug(3,"aor:%s  tp:%p has bean released else where!",loop->aor_name,loop->tp );
				ast_wave_mobile_tp_quitqueue(&g_wave_mobile_watch_queue, loop->aor_name);
				loop = loop->next;
				continue;
			}
			ast_relogin_enqueue(loop->aor_name);
			ast_append_reset_connect_tofile(loop->aor_name);
			ast_log(LOG_DEBUG, " aor:%s, tp:%p begin shutdown", loop->aor_name,loop->tp);
			if(PJ_SUCCESS == pjsip_transport_shutdown(loop->tp))
			{
				ast_wave_mobile_tp_quitqueue(&g_wave_mobile_watch_queue, loop->aor_name);
			}
			usleep(10000);
		}
		loop = loop->next;
	}
	
	ast_mutex_unlock(&(q->lock));
}

#endif
