/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Chuandong Ruan <chdruan@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>ucmsecret</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>

#include "asterisk/module.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/app.h"
#include "asterisk/acl.h"
#include "asterisk/haserver.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/stasis_channels.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"
#include <sqlite3.h>
#include "asterisk/astdb.h"
#include "asterisk/haserver.h"
#include "asterisk/ucm_dbi.h"
#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#endif
#ifdef HAVE_UCMSECRET
#include "ucm_crypt.h"
#endif
#endif


#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("pjsip");
#endif

#define  g_ldap_sync_key   "wohi90237402ns"
/*LDAP settings on sip_users.conf trunk_xx context*/
#define LDAP_ENABLE        "ldap_sync_enable"
#define LDAP_KEY           "ldap_sync_passwd"
#define LDAP_LOCAL_PORT    "ldap_sync_port"
#define LDAP_OUTGOING_RULE "ldap_default_outrt"
#define LDAP_EXT_PERFIX    "ldap_default_outrt_prefix"
#define LDAP_TRUNK_NAME    "trunk_name"
#define LDAP_TRUNK_HOST    "host"
#define HERE_DOC_MARKE     "TTsf7G0"
#ifdef GRANDSTREAM_NETWORKS /*Add by weidu ,for ldap sync*/
#define LDAP_SYNC_TYPE    "ldap_sync_type"
#endif
#define SAFE_FREE(buf) if (buf) {ast_free(buf);	buf = NULL;}

static int    g_ldap_changed   = 0;

#ifdef GRANDSTREAM_NETWORKS /*Add by weidu ,for ldap sync*/
enum sync_type
{
	SYNC_RSYNC_TYPE  = 1,
	SYNC_WGET_TYPE   = 2,
	
	SYNC_INVAILD_TYPE = 24
};

/* add by wylu */
struct generic_prepare_struct {
	char *sql;
	int argc;
	char **argv;
};

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}


#endif

struct st_ldap_server
{
	struct ast_sip_endpoint *endpoint;
	int enabled;
	int local_port;
	int remote_port;
	char *remote_host;
	char *decrypt_key;
	char *extension_prefix;
	char *trunk_name;
	char *trunk_name_markid;
	char *trunk_context;
	char *check_sum;/*The MD5 checksum of encrypted file*/
#ifdef GRANDSTREAM_NETWORKS /*Add by weidu ,for ldap sync*/
	int from_gdms;
	int sync_type;
#endif
	AST_LIST_ENTRY(st_ldap_server) entry;
};

struct st_ldap_sync_list
{
	int sip_tk_num;
	//struct ast_sip_endpoint  *sip_tk_sync_pos;
	AST_LIST_HEAD(, st_ldap_server) ast_ldap_server_list;
};

#ifdef GRANDSTREAM_NETWORKS
struct st_device_information
{
	char *exp_ip;
	int device_mode;
	char *wan_device;
	char *lan_device;
	char *default_if;
	char *lan1_device;
	char *lan2_device;
};

struct st_device_information g_ldap_device_mode = {NULL, 0, NULL, NULL, NULL, NULL, NULL};
AST_MUTEX_DEFINE_STATIC(g_device_info_mutex);
static char last_options_prd_ip[64] = {0};
static int last_options_cseq = 0;
#endif

struct st_ldap_sync_list  g_ldap_sync_list = {0, };

/*CONTENT TYPE BODY XML*/
static const pj_str_t type         = { "application", 11 };
static const pj_str_t subtype      = { "x-gs-ldap-sync+xml", 18 };
static const pj_str_t ldapSync     = { "ldapSync", 8};
static const pj_str_t providerName = { "providerName", 12};
static const pj_str_t providerAddr = { "providerAddr", 12};
static const pj_str_t providerExtIPAddr = { "providerExtIPAddr", 17};
static const pj_str_t fileCheckSum = { "fileCheckSum", 12};
static const pj_str_t providerMode = { "providerMode", 12};
#ifdef GRANDSTREAM_NETWORKS
static const pj_str_t cloudPrefix = { "cloudPrefix", 11};
static const pj_str_t cloudDevice = { "cloudDevice", 11};
#endif

/*EVENT HEADER*/
static const pj_str_t str_event_name = { "Event", 5 };


static pj_bool_t optnot_on_rx_request(pjsip_rx_data *rdata);
static int url_to_IP(const char *src, char *ip, int length);
static int check_dnssrv_equal_addr(const char* prefix, const char* host_address, const char* prdip);

static struct pjsip_module notify_module = {
	.name = { "NOTIFY Module", 13 },
	.priority = PJSIP_MOD_PRIORITY_APPLICATION - 1,
	.on_rx_request = optnot_on_rx_request,
};

/* need free */
static char* get_random_string(void)
{
	char buff[32] = {0};
	char metachar[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
	int i = 0;
	for (i = 0; i < sizeof(buff) - 1; i++)
	{
		buff[i] = metachar[rand() % 62]; 
	}
	buff[sizeof(buff) - 1] = '\0';
	return ast_strdup(buff);
}

static int is_private_network(const char *remote_addr, struct ast_sip_transport * transport)
{
	struct ast_sockaddr addr;

	if ( remote_addr == NULL || transport == NULL) {
		return -1;
	}
	if (-1 == ast_sockaddr_parse(&addr, remote_addr, 0))
		return -1;
	return (ast_apply_ha(transport->localnet, &addr) == AST_SENSE_ALLOW) ? 0 : 1;
}
#if 1
static int getLocalAddress(pj_pool_t *pool, char* remote_host, char* ipbuf, int buf_len, enum ast_transport transport_type)
{
	pjsip_tpmgr_fla2_param prm;
	if (remote_host == NULL)
		return -1;
	pjsip_tpmgr_fla2_param_default(&prm);
	enum pjsip_transport_type_e tp_type = (pjsip_transport_type_e)transport_type;
	if (strchr(remote_host, ':'))
		tp_type = (pjsip_transport_type_e)(((int)tp_type) + PJSIP_TRANSPORT_IPV6);
	prm.tp_type = tp_type;
	pj_strset2(&prm.dst_host, remote_host);
	prm.local_if = PJ_TRUE;
	if (pjsip_tpmgr_find_local_addr2(pjsip_endpt_get_tpmgr(ast_sip_get_pjsip_endpoint()), pool, &prm) == PJ_SUCCESS) {
		ast_copy_string(ipbuf, (&prm.ret_addr)->ptr, (prm.ret_addr.slen + 1) > buf_len ? buf_len : (prm.ret_addr.slen + 1));
	}
	return 0;
}

/* local_ip : "0" means to return the local ip, others means to return the NAT IP*/
static int getOutgoingIP(char *ipbuf, int local_ip, struct ast_sip_transport * transport, int buf_len)
{
	if (NULL == ipbuf) {
		return -1;
	}

	if (local_ip != 0 && transport && !ast_strlen_zero(transport->external_signaling_address)) {
		ast_log(LOG_DEBUG, "Have the external IP via NAT, use it : {%s}", transport->external_signaling_address);
		ast_copy_string(ipbuf, transport->external_signaling_address, buf_len);
	}
	else if (!ast_strlen_zero(ipbuf)) {
		ast_log(LOG_DEBUG, "Have the local_address IP, use it : {%s}", ipbuf);
	}

	return 0;
}

static int getExtIPAddr(char *ipbuf, size_t size)
{
	char buffer[128] = { 0 };
	char cmd[32] = {0};
	char buf[64] = {0};
	FILE *fp = NULL;
	snprintf(cmd, sizeof(cmd), "nvram get ext_ip");
	fp = popen(cmd, "r");
	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
			   && (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
			ast_log(LOG_DEBUG, "ext_ip Read: %s", buf);
			strcat(buffer, buf);
		}
		pclose(fp);
		fp = NULL;
		char *ptr = strchr(buffer, '\n');
		if (ptr != NULL) {
			*ptr = '\0';
		}
	}
	if (buffer[0] != '\0' && 0 != strcasecmp(ipbuf, buffer)) {
		ast_log(LOG_DEBUG, "get Ext ip Read: %s", buffer);
		snprintf(ipbuf , (strlen(buffer) + 1) > size ? size: strlen(buffer) + 1, "%s", buffer);
	}
	return 0;
}

#else
/* local_ip : "0" means to return the local ip, others means to return the NAT IP*/
static int getOutgoingIP(char *ipbuf, int local_ip)
{
	struct ifreq ifr_ip;
	struct sockaddr_in  *sin;
	int  fd;
	char device[16] = { 0 };

	if (NULL == ipbuf) {
		return -1;
	}

	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (-1 == fd) {
		return -1;
	}
	ast_mutex_lock(&g_device_info_mutex);
	if ( local_ip != 0 && g_ldap_device_mode.exp_ip != NULL && (strchr(g_ldap_device_mode.exp_ip, '.') != NULL || strchr(g_ldap_device_mode.exp_ip, ':') != NULL)) {
		/*Try to support IPv4 and IPv6*/
		ast_log(LOG_DEBUG, "Have the external IP via STUN, use it : {%s}", g_ldap_device_mode.exp_ip);
		strcpy(ipbuf, g_ldap_device_mode.exp_ip);
	}
	else {
		if (g_ldap_device_mode.device_mode== 0 && g_ldap_device_mode.wan_device) {
			/*Router mode, using wan_device */
			/*0 ethx ethxx LANx*/
			ast_copy_string(device, g_ldap_device_mode.wan_device, sizeof(device) - 1);
		}
		else if (g_ldap_device_mode.device_mode == 1 && g_ldap_device_mode.lan_device) {
			/*Switch mode, using lan_device */
			/*1 ethx ethx LANX*/
			ast_copy_string(device, g_ldap_device_mode.lan_device, sizeof(device) - 1);
		}
		else if (g_ldap_device_mode.default_if){
			/*Dual mode, using the default interface*/
			/*2 LANX ethx etxx*/
			ast_copy_string(device, g_ldap_device_mode.default_if, sizeof(device) - 1);
		}
		memset(&ifr_ip, 0, sizeof(ifr_ip));
		strncpy(ifr_ip.ifr_ifrn.ifrn_name, device[0] == '\0' ? "eth0" : device, sizeof(ifr_ip.ifr_ifrn.ifrn_name) - 1);
		if (ioctl(fd, SIOCGIFADDR, &ifr_ip) < 0) {
			ast_log(LOG_DEBUG, "getOutgoingIP: failed\n");
			ast_mutex_unlock(&g_device_info_mutex);
			close(fd);
			return -1;
		}
			sin = (struct sockaddr_in *)&ifr_ip.ifr_ifru.ifru_addr;
			strcpy(ipbuf, ast_inet_ntoa(sin->sin_addr));

	}
	ast_mutex_unlock(&g_device_info_mutex);

	ast_log(LOG_DEBUG, "getOutgoingIP: [%s]\n", ipbuf);
	close(fd);
	return 0;
}
#endif
static char* strip_brackets_ip(const char* addr)
{
	char actual_addr[256] = {0};
	if (strstr(addr, ":") != NULL && addr[0] == '[' && addr[strlen(addr) - 1] == ']') {
		snprintf(actual_addr, strlen(addr) - 1, "%s", addr + 1);
	}
	else {
		snprintf(actual_addr, strlen(addr) + 1, "%s", addr);
	}
	return ast_strdup(actual_addr);
}

static int check_equal_addr(const char* addr1, const char* addr2)
{
	int addr1_af = AF_INET;
	int addr2_af = AF_INET;
	int ret = -1;
	struct in6_addr addr1_in6;
	struct in6_addr addr2_in6;
	char* addr1_ipv6 = NULL;
	char* addr2_ipv6 = NULL;
	if (addr1 == NULL || addr2 == NULL)
		return ret;

	ast_log(LOG_DEBUG, "check_equal_addr [%s][%s]", addr1, addr2);
#ifdef GRANDSTREAM_NETWORKS /*Add by weidu , for ldap sync*/
	if (ast_strlen_zero(addr1) || ast_strlen_zero(addr2)){
		return -1;
	}
#endif
	if (strstr(addr1, ":") != NULL) {
		addr1_af = AF_INET6;
	}

	if (strstr(addr2, ":") != NULL) {
		addr2_af = AF_INET6;
	}

	if (addr1_af != addr2_af) {
		return ret;
	}
	else if (addr1_af == AF_INET6) {
		addr1_ipv6 = strip_brackets_ip(addr1);
		addr2_ipv6 = strip_brackets_ip(addr2);
		if (!ast_strlen_zero(addr1_ipv6) && inet_pton(addr1_af, addr1_ipv6, &addr1_in6) <= 0) {
			ast_log(LOG_DEBUG, "inet_pton failed [%s]", addr1_ipv6);
			ret = -1;
		}

		if (!ast_strlen_zero(addr2_ipv6) && inet_pton(addr2_af, addr1_ipv6, &addr2_in6) <= 0) {
			ast_log(LOG_DEBUG, "inet_pton failed [%s]", addr2_ipv6);
			ret = -1;
		}
		if (memcmp(&addr1_in6, &addr2_in6, sizeof(struct in6_addr)) == 0) {
			ret = 0;
		}
		SAFE_FREE(addr1_ipv6);
		SAFE_FREE(addr2_ipv6);
	}
	else {
		char host_address1[256] = {0};
		char host_address2[256] = {0};
		url_to_IP(addr1, host_address1, sizeof(host_address1));
		url_to_IP(addr2, host_address2, sizeof(host_address2));
#ifdef GRANDSTREAM_NETWORKS /*Mod by weidu , for ldap sync*/
		if (!strlen(host_address1) || !strlen(host_address2)){
			if (check_dnssrv_equal_addr("_ldap._tcp", addr1, addr2) == PJ_TRUE){
				return 0;
			}
		}else{
			return strcasecmp(host_address1, host_address2);
		}
#endif
	}
	return ret;
}

static int url_to_IP(const char *src, char *ip, int length)
{
	struct addrinfo hints, *res;
	struct in_addr addr;
	int err;
	unsigned char buf[sizeof(struct in6_addr)] = {0};
	char url[256] = {0};
	if (src == NULL || ip == NULL) {
		ast_log(LOG_DEBUG, "Enter url_to_IP : Input error");
		return -1;
	}
	/*remove [] for ipv6 */
	if (strstr(src, ":") != NULL && src[0] == '[' && src[strlen(src) - 1] == ']')
	{
		snprintf(url, strlen(src) - 1, "%s", src + 1);
	}
	else {
		snprintf(url, strlen(src) + 1, "%s", src);
	}
	ast_log(LOG_DEBUG, "url_to_IP : URL: %s", url);

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC;	/* Allow IPv4 or IPv6 */
	/*it is a vaild IPV4/IPV6 addr*/
	if (inet_pton(AF_INET, url, buf) > 0 || inet_pton(AF_INET6, url, buf) > 0 )
	{
		snprintf(ip, length, "%s", src); /*In case the url is an IP address*/
		return 1;
	}
	if ( (err = getaddrinfo(url, NULL, &hints, &res)) != 0) {
		ast_log(LOG_DEBUG, "url_to_IP : error %d", err);
#ifdef GRANDSTREAM_NETWORKS  /*Mody by weidu , for ldap sync*/
		return -1;
#else
		snprintf(ip, length, "%s", src); /*In case the url is an IP address*/
		return 1;
#endif
	}

	addr.s_addr = ((struct sockaddr_in *)(res->ai_addr))->sin_addr.s_addr;
	const char *ptr = ast_inet_ntoa(addr);
	ast_log(LOG_DEBUG, "url_to_IP : IP address: %s", ptr==NULL?"UNKNOW":ptr);

	snprintf(ip, length, "%s", ptr==NULL?src:ptr);

	freeaddrinfo(res);

	return 0;
}

static int check_dnssrv_equal_addr(const char* prefix, const char* host_address, const char* prdip)
{
	char service[NI_MAXHOST] = {0};
	char host[256] = {0};
	struct srv_context *context = NULL;
	int srv_ret;
	int ret = PJ_FALSE;
	const char *srvhost;
	unsigned short srvport;

	snprintf(service, sizeof(service), "%s.%s", prefix, host_address);

	while (!(srv_ret = ast_srv_lookup(&context, service, &srvhost, &srvport))) {
		url_to_IP(srvhost, host, sizeof(host));
#ifdef GRANDSTREAM_NETWORKS
		if (!strlen(host)){
			continue;
		}
#endif
		if (check_equal_addr(host, prdip) == 0) {
			ret = PJ_TRUE;
			break;
		}
	}

	ast_srv_cleanup(&context);
	return ret;
}

static int get_ldap_checksum(char* trunk_name_markid, char *checksum, int length)
{
	char cmd[64] = {0};
	char buf[64] = {0};
	char buffer[128] = {0};

	if (NULL == checksum) {
		return -1;
	}

	FILE *fp = NULL;
	snprintf(cmd, sizeof(cmd), "nvram get :ldapsync_%s", trunk_name_markid);
	fp = popen(cmd, "r");
	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
			   && (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
			ast_log(LOG_DEBUG, "get_ldap_checksum Read: %s", buf);
			strcat(buffer, buf);
		}
		pclose(fp);
		fp = NULL;

		char *ptr = strchr(buffer, '\n');
		if (ptr != NULL) {
			*ptr = '\0';
		}
	}
	ast_log(LOG_DEBUG, "get_ldap_checksum: [%s]\n", buffer);
	strncpy(checksum, buffer, length);

#ifdef GRANDSTREAM_NETWORKS
	snprintf(cmd, sizeof(cmd), "nvram unset :ldapsync_%s", trunk_name_markid);
	ast_safe_system2(cmd);
#endif

	return 0;
}

/* "1" means yes, "0 means no"*/
static int getNeedToSyncLDAP(void)
{
	ast_log(LOG_VERBOSE, "Enter getNeedToSyncLDAP!!!\n");

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for HA service data sync, don't do ldap sync when standby */
	if (HA_ROLE_STANDBY == ast_ha_get_role()) {
		g_ldap_changed = 0;
		return 0;
	}
#endif

	if ( g_ldap_changed == 0 ) {
		char buf[64] = { 0 };
		char buffer[256] = { 0 };

		FILE *fp = NULL;
		fp = popen("nvram get :needLDAPSync", "r");

		if (fp != NULL) {
			while (fp != NULL && (!feof(fp))
				   && (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
				//ast_log(LOG_DEBUG, "getNeedToSyncLDAP Read: %s", buf);
				strcat(buffer, buf);
			}
			pclose(fp);
			fp = NULL;

			if (buffer[0] != '\0') {
				/*parsing result*/
				g_ldap_changed = atoi( buffer );
			}
		}
	}
	ast_log(LOG_DEBUG, "getNeedToSyncLDAP result is : %d", g_ldap_changed);

	return g_ldap_changed;
}

/* Function to print XML message body.*/
static int xml_print_body( struct pjsip_msg_body *msg_body, char *buf, pj_size_t size)
{
	return pj_xml_print((const pj_xml_node*)msg_body->data, buf, size, PJ_TRUE);
}

/*Function to clone XML document.*/
static void* xml_clone_data(pj_pool_t *pool, const void *data, unsigned len)
{
	PJ_UNUSED_ARG(len);
	return pj_xml_clone( pool, (const pj_xml_node*)data);
}

#ifdef GRANDSTREAM_NETWORKS
static char *get_cloud_im_device_id_and_prefix(char *type)
{
	char cmd[1024] = {0};
	struct odbc_obj *obj = NULL;
	SQLHSTMT stmt = NULL;
	char rowdata[256] = {0};
	char *result = NULL;
	int res = -1;

	if (type == NULL)
	{
		ast_log(AST_LOG_ERROR, "The type is NULL!");
		return NULL;
	}

	struct generic_prepare_struct gps = { .sql = cmd };
	obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);
	if (!obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
		return NULL;
	}
	memset(cmd,0,sizeof(cmd));
	snprintf(cmd, sizeof(cmd), "SELECT %s FROM cloud_im_settings WHERE cloud_im_enabled = 'on'", type);
	stmt = ast_core_odbc_prepare_and_execute(obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(AST_LOG_WARNING, "SQL Execute error!\n[%s]\n\n", cmd);
		goto fail;
	}
	res = SQLFetch(stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", cmd);
		goto fail_with_handle;
	}
	res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Get Data error!\n[%s]\n\n", cmd);
		goto fail_with_handle;
	}
	result = ast_strdup(rowdata);


fail_with_handle:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
fail:
	ast_core_odbc_release_obj(obj);

	return result;

}
#endif
static pjsip_msg_body *gs_ldap_sync_build_xml(pj_pool_t *pool, struct ast_sip_endpoint *endpoint)
{
	pj_str_t providerNameVal;
	pj_str_t providerAddrVal;
	pj_str_t providerAddrExtIPVal;
	pj_str_t providerModeVal;
	pj_str_t fileCheckSumVal;
#ifdef GRANDSTREAM_NETWORKS
	pj_str_t cloudPrefixVal;
	pj_str_t cloudDeviceVal;
#endif
	pj_xml_node *doc, *node;
	char address[256]       = {0};
	char checksum[128]      = {0};
	char trunk_context[128] = {0};
#ifdef GRANDSTREAM_NETWORKS
	char ha_on[15] = {0};
#endif
	pjsip_msg_body *body    = NULL;
	int flag = 0; /*Add by weidu , for gdms ldap*/
	struct st_ldap_server  *pstLdapServer    = NULL;
	const char *name = ast_sorcery_object_get_id(endpoint);
	snprintf(trunk_context, sizeof(trunk_context), "ldap_%s", name);

	RAII_VAR(struct ast_sip_transport *, transport, NULL, ao2_cleanup);
	if (!(transport = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "transport", "trans-udp"))) {
		ast_log(LOG_DEBUG, "trans-udp transport is null");
		return NULL;
	}

	AST_LIST_TRAVERSE(&(g_ldap_sync_list.ast_ldap_server_list), pstLdapServer, entry) {
		if ( NULL != pstLdapServer->endpoint&& pstLdapServer->enabled == 1
		  && NULL != pstLdapServer->remote_host && strcasecmp(pstLdapServer->trunk_context, trunk_context ) == 0 ) {
		  
			char host[256] = {0};
			doc = pj_xml_node_new(pool, &ldapSync);
			flag = 0;
			node = pj_xml_node_new(pool, &providerName);
			providerNameVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
			providerNameVal.slen = pj_ansi_snprintf(providerNameVal.ptr, PJSIP_MAX_URL_SIZE,"%s",pstLdapServer->trunk_name_markid);
			node->content = providerNameVal;
			pj_xml_add_node(doc, node);

			node = pj_xml_node_new(pool, &providerAddr);
			memset(address, 0, sizeof(address));
#ifdef GRANDSTREAM_NETWORKS
			ast_sip_get_nvram_value("haon", ha_on, sizeof(ha_on));
			if (strlen(ha_on))
			{
				if (strcmp(ha_on, "1") == 0){
					ast_sip_get_nvram_value("hafloatip", address, sizeof(address));
				}else{
					getLocalAddress(pool, pstLdapServer->remote_host, address, sizeof(address), transport->type);
				}
			}else{
				getLocalAddress(pool, pstLdapServer->remote_host, address, sizeof(address), transport->type);
			}
#else
			getLocalAddress(pool, pstLdapServer->remote_host, address, sizeof(address), transport->type);
#endif
			getOutgoingIP(address, is_private_network( pstLdapServer->remote_host, transport), transport, sizeof(address));
			/*BEGIN: Mod by weidu , for gdms ldap*/
#ifdef GRANDSTREAM_NETWORKS
			providerAddrVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
			ast_sip_get_nvram_value(":WsProxyHost", host, sizeof(host));

			if (pstLdapServer->sync_type == SYNC_RSYNC_TYPE){
				if (strchr(address, ':')){
					providerAddrVal.slen = pj_ansi_snprintf(providerAddrVal.ptr, PJSIP_MAX_URL_SIZE,"[%s]",address);
				}
				else{
					providerAddrVal.slen = pj_ansi_snprintf(providerAddrVal.ptr, PJSIP_MAX_URL_SIZE,"%s",address);
				}	
			}
			else if (pstLdapServer->from_gdms && !ast_strlen_zero(host)){
				providerAddrVal.slen = pj_ansi_snprintf(providerAddrVal.ptr, PJSIP_MAX_URL_SIZE,"%s", host);
				flag = 2;  /*gdms*/
			}else{
				if (strchr(address, ':')){
					providerAddrVal.slen = pj_ansi_snprintf(providerAddrVal.ptr, PJSIP_MAX_URL_SIZE,"[%s]:%d",address, pstLdapServer->local_port);
				}
				else{
					providerAddrVal.slen = pj_ansi_snprintf(providerAddrVal.ptr, PJSIP_MAX_URL_SIZE,"%s:%d",address, pstLdapServer->local_port);
				}	
				flag = 1;
			}
			pstLdapServer->from_gdms = 0;
#endif
			/*END: Add by weidu , for gdms ldap*/
			node->content = providerAddrVal;
			pj_xml_add_node(doc, node);

			/* add by kmye.
			   append ext_ip info xml info
			*/
			memset(address, 0, sizeof(address));
			getExtIPAddr(address, sizeof(address));
			if (!ast_strlen_zero(address)) {
				node = pj_xml_node_new(pool, &providerExtIPAddr);
				providerAddrExtIPVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
				if (strchr(address, ':'))
					providerAddrExtIPVal.slen = pj_ansi_snprintf(providerAddrExtIPVal.ptr, PJSIP_MAX_URL_SIZE,"[%s]",
						address);
				else
					providerAddrExtIPVal.slen = pj_ansi_snprintf(providerAddrExtIPVal.ptr, PJSIP_MAX_URL_SIZE,"%s",
						address);

				node->content = providerAddrExtIPVal;
				pj_xml_add_node(doc, node);
			}
			/* add providerMode node. marks to use rsync */
			node = pj_xml_node_new(pool, &providerMode);
			providerModeVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
#ifdef GRANDSTREAM_NETWORKS
			/*BEGIN: Mod by weidu , for ldap*/
			if (flag == 1){
				providerModeVal.slen = pj_ansi_snprintf(providerModeVal.ptr, PJSIP_MAX_URL_SIZE,"normal");
			}else if (flag == 2){
				providerModeVal.slen = pj_ansi_snprintf(providerModeVal.ptr, PJSIP_MAX_URL_SIZE,"gdms");
			}else{
				providerModeVal.slen = pj_ansi_snprintf(providerModeVal.ptr, PJSIP_MAX_URL_SIZE,"security");
			}
			/*END: Mod by weidu , for ldap*/
#endif
			node->content = providerModeVal;
			pj_xml_add_node(doc, node);

			node = pj_xml_node_new(pool, &fileCheckSum);
			get_ldap_checksum(pstLdapServer->trunk_name_markid, checksum, sizeof(checksum));
			fileCheckSumVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
			fileCheckSumVal.slen = pj_ansi_snprintf(fileCheckSumVal.ptr, PJSIP_MAX_URL_SIZE,"%s",checksum);
			node->content = fileCheckSumVal;
			pj_xml_add_node(doc, node);
#ifdef GRANDSTREAM_NETWORKS
			char *cloudPrefix_str = get_cloud_im_device_id_and_prefix("cloud_im_prefix");
			if (cloudPrefix_str != NULL)
			{
				ast_log(AST_LOG_DEBUG, "The cloudPrefix_str is %s", cloudPrefix_str);
				node = pj_xml_node_new(pool, &cloudPrefix);
				cloudPrefixVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
				cloudPrefixVal.slen = pj_ansi_snprintf(cloudPrefixVal.ptr, PJSIP_MAX_URL_SIZE,"%s",cloudPrefix_str);
				node->content = cloudPrefixVal;
				pj_xml_add_node(doc, node);
				ast_free(cloudPrefix_str);
			}
			char *cloudDevice_str = get_cloud_im_device_id_and_prefix("cloud_im_device_id");
			if (cloudDevice_str != NULL)
			{
				ast_log(AST_LOG_DEBUG, "The cloudDevice_str is %s", cloudDevice_str);
				node = pj_xml_node_new(pool, &cloudDevice);
				cloudDeviceVal.ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
				cloudDeviceVal.slen = pj_ansi_snprintf(cloudDeviceVal.ptr, PJSIP_MAX_URL_SIZE,"%s",cloudDevice_str);
				node->content = cloudDeviceVal;
				pj_xml_add_node(doc, node);
				ast_free(cloudDevice_str);
			}
			
#endif
			body = PJ_POOL_ZALLOC_T(pool, pjsip_msg_body);
			body->content_type.type = type;
			body->content_type.subtype = subtype;

			body->data = doc;
			body->len = 0;

			body->print_body = &xml_print_body;
			body->clone_data = &xml_clone_data;
		}
	}

	return body;

}

static int send_options(struct ast_sip_endpoint *endpoint, struct ast_sip_contact *contact)
{
	pjsip_tx_data *tdata;

	if (ast_sip_create_request("OPTIONS", NULL, NULL, NULL, contact, &tdata)) {
		ast_log(LOG_ERROR, "Unable to create OPTIONS request\n");
		return -1;
	}

	/* If an outbound proxy is specified set it on this request */
	if (!ast_strlen_zero(endpoint->outbound_proxy) &&
		ast_sip_set_outbound_proxy(tdata, endpoint->outbound_proxy, endpoint->rmv_obp_from_route))
	{
		pjsip_tx_data_dec_ref(tdata);
		ast_log(LOG_ERROR, "Unable to apply outbound proxy on OPTIONS request\n");
		return -1;
	}

	tdata->msg->body = gs_ldap_sync_build_xml(tdata->pool, endpoint);
	if (NULL == tdata->msg->body) {
		pjsip_tx_data_dec_ref(tdata);
		ast_log(LOG_ERROR, "Create LDAP XML body failure.\n");
		return -1;
	}

	ao2_ref(contact, +1);
	if (ast_sip_send_request(tdata, NULL, endpoint, contact, NULL)
		!= PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Unable to send OPTIONS request\n");
		ao2_ref(contact, -1);
		return -1;
	}

	return 0;
}

static int sip_poke_endpoint(void *data)
{
	struct ast_sip_endpoint *endpoint = (struct ast_sip_endpoint *)data;
	char *aors;
	char *aor_name;

	if (!endpoint) {
		ast_log(LOG_ERROR, "Endpoint is NULL\n");
		return -1;
	}
	if (ast_strlen_zero(endpoint->aors)) {
		ast_log(LOG_ERROR, "Endpoint %s has no AoR's configured\n", ast_sorcery_object_get_id(endpoint));
		return -1;
	}

	aors = ast_strdupa(endpoint->aors);
	while ((aor_name = strsep(&aors, ","))) {
		struct ast_sip_aor *aor;
		struct ast_sip_contact *contact;

		aor = ast_sip_location_retrieve_aor(aor_name);
		if (!aor) {
			continue;
		}

		contact = ast_sip_location_retrieve_first_aor_contact(aor);
		if (contact) {
			send_options(endpoint,contact);
		}

		ao2_ref(aor, -1);
	}
	return PJ_TRUE;
}

/*! \brief Send a ldap sync optons to all sip trunk endpoints */
#ifdef GRANDSTREAM_NETWORKS
static void sip_ldap_sync_manual(const char *context, char *to_domain, char *ldap_support_attr)
#else
static void sip_ldap_sync_manual(const char *context)
#endif
{
	ast_log(LOG_VERBOSE, "Enter sip_ldap_sync_manual !!!\n");

	struct st_ldap_server  *pstLdapServer = NULL;
	char   command[512] = {0};
	char   trunk_context[128] = {0};
	char host[256] = {0}; /*Add by weidu , for ldap*/
	char gs_port[32] = {0};

	if ( context == NULL) {
		ast_log(LOG_ERROR, "sip_ldap_sync_manual, input error!\n");
		return;
	}

	snprintf(trunk_context, sizeof(trunk_context), "ldap_%s", context);
	ast_log(LOG_DEBUG, "sip_ldap_sync_manual host: [%s]!!!", trunk_context);

	if (0 == g_ldap_sync_list.sip_tk_num) {
		ast_log(LOG_ERROR, "sip_ldap_sync_msnual, no sip trunk\n");
		return;
	}

	AST_LIST_LOCK(&(g_ldap_sync_list.ast_ldap_server_list));
	AST_LIST_TRAVERSE(&(g_ldap_sync_list.ast_ldap_server_list), pstLdapServer, entry) {
		ast_log(LOG_DEBUG, "trunk_context is (%s), context is (%s)", pstLdapServer->trunk_context, trunk_context);

		if ( NULL != pstLdapServer->endpoint && pstLdapServer->enabled == 1
			&& strcasecmp(pstLdapServer->trunk_context, trunk_context) == 0 ) {
			ast_log(LOG_DEBUG, "prepareLDAPSync for trunk [%s] ~", pstLdapServer->trunk_name);
			ast_log(LOG_DEBUG, "prepareLDAPSync for trunk_mark_id [%s] ~", pstLdapServer->trunk_name_markid);
#ifdef GRANDSTREAM_NETWORKS
			/*BEGIN: Add by weidu ,for ldap*/
			ast_sip_get_nvram_value(":WsProxyHost", host, sizeof(host));
			ast_sip_get_nvram_value("gswave_port", gs_port, sizeof(gs_port));
			if (!ast_strlen_zero(gs_port)){ /*use gswave*/
				pstLdapServer->local_port = atoi(gs_port);
			}else{
				pstLdapServer->local_port = 8090;
			}
			ast_log(LOG_DEBUG, "sip_ldap_sync_manual to_domain %s \n" , to_domain);
			if (!ast_strlen_zero(host) && !ast_strlen_zero(to_domain) && !strcasecmp(host, to_domain)){
				pstLdapServer->from_gdms = 1;
			}else{
				pstLdapServer->from_gdms = 0;
			}
			/*END: Add by weidu ,for ldap*/
#endif
			ast_log(LOG_DEBUG, "prepareLDAPSync for <%s> (%d)", pstLdapServer->decrypt_key==NULL?"NULL":pstLdapServer->decrypt_key, pstLdapServer->local_port);
			
			if (pstLdapServer->sync_type == SYNC_RSYNC_TYPE){
				snprintf(command, sizeof(command), "prepareLDAPFile.sh `cat << '%s'\n%s %s rsnc %s \n%s\n`", HERE_DOC_MARKE, pstLdapServer->decrypt_key==NULL?g_ldap_sync_key:pstLdapServer->decrypt_key
				  , pstLdapServer->trunk_name_markid, ast_strlen_zero(ldap_support_attr) ? "NULL" : ldap_support_attr, HERE_DOC_MARKE);
			}else{
				snprintf(command, sizeof(command), "prepareLDAPFile.sh `cat << '%s'\n%s %s wget %s \n%s\n`", HERE_DOC_MARKE, pstLdapServer->decrypt_key==NULL?g_ldap_sync_key:pstLdapServer->decrypt_key
				  , pstLdapServer->trunk_name_markid, ast_strlen_zero(ldap_support_attr) ? "NULL" : ldap_support_attr, HERE_DOC_MARKE);
			}
			ast_log(LOG_DEBUG, "command = (%s)", command);
			ast_safe_system2(command);

			sip_poke_endpoint(pstLdapServer->endpoint);
			pstLdapServer->from_gdms = 0;
		}
	}
	AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));

	ast_log(LOG_VERBOSE, "Exit sip_ldap_sync_manual !!!\n");

	return;
}

static pj_bool_t notify_on_rx_request(pjsip_rx_data *rdata)
{
	ast_log(LOG_VERBOSE, "Enter notify_on_rx_request !!!\n");

	pjsip_event_hdr *event_header;
	char event[32];
#ifdef GRANDSTREAM_NETWORKS
	pjsip_sip_uri *uri_to = NULL;  /*Add by weidu ,for ldap gdms*/
	char notify_ip[128] = {0};
	uri_to = pjsip_uri_get_uri(rdata->msg_info.to->uri);
	/* Add by wylu for get ldap support attr */
	pjsip_generic_string_hdr *attr_header;
	char ldap_support_attr[64] = { 0 };
	const pj_str_t ldap_attr_header = { "X-GS-LdapAttrSupport", 20 };
#endif
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	endpoint = ast_pjsip_rdata_get_endpoint(rdata);
	ast_assert(endpoint != NULL);

	const char *name = ast_sorcery_object_get_id(endpoint);

	event_header = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &str_event_name, rdata->msg_info.msg->hdr.next);
	if (!event_header) {
		ast_log(LOG_ERROR, "Incoming NOTIFY request with no Event header\n");
		return PJ_TRUE;
	}
	ast_copy_pj_str(event, &event_header->event_type, sizeof(event));

	if (strcmp("gs-request-ldap-sync", event)) {
		return PJ_FALSE;
	}

#ifdef GRANDSTREAM_NETWORKS  /*Add by weidu ,for ldap gdms*/
	if (uri_to) {
		ast_copy_pj_str(notify_ip, &(uri_to->host), sizeof(notify_ip));
	}

	/* Add by wylu for get ldap support attr */
	attr_header = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &ldap_attr_header, NULL);
	if (!attr_header) {
		ast_log(LOG_DEBUG, "Incoming NOTIFY request with no LdapAttrSupport header\n");
	}else {
		ast_copy_pj_str(ldap_support_attr, &attr_header->hvalue, sizeof(ldap_support_attr));
	}
#endif

	pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 200, NULL, NULL, NULL);

	sip_ldap_sync_manual(name, notify_ip, ldap_support_attr);

	return PJ_TRUE;
}

static pj_status_t sip_get_ldap_sync_info_from_msg( pj_pool_t *pool, char *msg, pj_size_t len,
	pj_str_t **pName, pj_str_t **pIPAddr, pj_str_t **pChecksum, pj_str_t **pExtIPAddr, pj_str_t **pMode, pj_str_t **pCloudDevice, pj_str_t **pCloudPrefix)
{
	pj_xml_node *doc, *node;

	doc = pj_xml_parse( pool, msg, len);
	if (!doc) {
		return -1;
	}

	if (pj_stricmp(&doc->name, &ldapSync) != 0) {
		return -1;
	}

	node = pj_xml_find_node(doc, &providerName);
	if (node == NULL) {
		return -1;
	}
	*pName = &node->content;

	node = pj_xml_find_node(doc, &providerAddr);
	if (node == NULL) {
		return -1;
	}
	*pIPAddr = &node->content;

	node = pj_xml_find_node(doc, &fileCheckSum);
	if (node == NULL) {
		return -1;
	}
	*pChecksum = &node->content;
	/* add by kmye. providerExtIPAddr is't be include in xml at 20.X or lower version */
	node = pj_xml_find_node(doc, &providerExtIPAddr);
	if (node != NULL) {
		*pExtIPAddr = &node->content;
	}
	node = pj_xml_find_node(doc, &providerMode);
	if (node != NULL) {
		*pMode = &node->content;
	}

	/* add by wylu for cloudim device_id and prefix */
	node = pj_xml_find_node(doc, &cloudDevice);
	if (node != NULL) {
		*pCloudDevice = &node->content;
	}
	node = pj_xml_find_node(doc, &cloudPrefix);
	if (node != NULL) {
		*pCloudPrefix = &node->content;
	}

	return PJ_SUCCESS;
}

static pj_bool_t options_on_rx_request(pjsip_rx_data *rdata)
{
	ast_log(LOG_VERBOSE, "Enter options_on_rx_request !!!\n");

	pj_int32_t cseq = rdata->msg_info.cseq->cseq;
	pjsip_ctype_hdr *ctype_header;
	pjsip_msg_body *body = rdata->msg_info.msg->body;
	pj_str_t *pj_name = NULL;
	pj_str_t *pj_ip = NULL;
	pj_str_t *pj_ext_ip = NULL;
	pj_str_t *pj_checksum = NULL;
	pj_str_t *pj_mode = NULL;
	pj_str_t *pj_device_id = NULL;
	pj_str_t *pj_device_prefix = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	endpoint = ast_pjsip_rdata_get_endpoint(rdata);

	ast_assert(endpoint != NULL);

	ctype_header = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTENT_TYPE, NULL);
	if (!ctype_header) {
		ast_log(LOG_DEBUG, "Incoming OPTIONS request with no CONTENT TYPE header\n");
		return PJ_FALSE;
	}

	if (pj_stricmp(&ctype_header->media.type, &type)) {
		ast_log(LOG_DEBUG, "Type is not application\n");
		return PJ_FALSE;
	}
	if (pj_stricmp(&ctype_header->media.subtype, &subtype)) {
		ast_log(LOG_DEBUG, "Content_type is not gs-request-ldap-sync\n");
		return PJ_FALSE;
	}

	if (!body || !body->data) {
		ast_log(LOG_DEBUG, "There is no message body in received packet\n");
		return PJ_FALSE;
	}

	if (PJ_SUCCESS == sip_get_ldap_sync_info_from_msg(rdata->tp_info.pool, (char*)body->data, body->len,
			&pj_name, &pj_ip, &pj_checksum, &pj_ext_ip, &pj_mode, &pj_device_id, &pj_device_prefix)) {
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 200, NULL, NULL, NULL);

		char prd_name[80] = {0};
		char prd_ip[64] = {0};
		char prd_ext_ip[64] = {0};
		char prd_checksum[128] = {0};
		char prd_mode[16] = {0};
		char prd_device_id[16] = {0};
		char prd_device_prefix[16] = {0};
		int  prd_port   = 0;
		int  prd_ext_ip_port = 0;
		char command[512] = {0};
		char host_address[256] = {0};
		struct st_ldap_server *pstLdapServer = NULL;
#ifdef GRANDSTREAM_NETWORKS
		char prd_gdms_header_ip[64] = {0};
		pjsip_sip_uri *uri_from = NULL;
		uri_from = pjsip_uri_get_uri(rdata->msg_info.from->uri);
#endif
		snprintf(prd_name, sizeof(prd_name), "%.*s", (int)pj_name->slen, pj_name->ptr);
		snprintf(prd_ip, sizeof(prd_ip), "%.*s", (int)pj_ip->slen, pj_ip->ptr);
		char *pExtPort = NULL;
		char *pPort = NULL;

		if (NULL != pj_mode) {
			snprintf(prd_mode, sizeof(prd_mode), "%.*s", (int)pj_mode->slen, pj_mode->ptr);
		} else {
			snprintf(prd_mode, sizeof(prd_mode), "empty");
		}
#ifdef GRANDSTREAM_NETWORKS
		if (NULL != pj_device_id) {
			snprintf(prd_device_id, sizeof(prd_device_id), "%.*s", (int)pj_device_id->slen, pj_device_id->ptr);
		} else {
			snprintf(prd_device_id, sizeof(prd_device_id), "empty");
		}

		if (NULL != pj_device_prefix && !ast_strlen_zero(pj_device_prefix->ptr)) {
			snprintf(prd_device_prefix, sizeof(prd_device_prefix), "%.*s", (int)pj_device_prefix->slen, pj_device_prefix->ptr);
		} else {
			snprintf(prd_device_prefix, sizeof(prd_device_prefix), "empty");
		}
#endif
		/*it need't support PORT when we sync by rsync*/

		if (ast_strlen_zero(prd_mode) || 0 == strncmp(prd_mode, "normal", strlen(prd_mode))
			|| 0 == strncmp(prd_mode, "empty", strlen(prd_mode))) {
			pPort = strrchr(prd_ip, ':');
			if (pPort != NULL) {
				*pPort = '\0';
				++pPort;
				prd_port = atoi(pPort);
			}

			if (NULL != pj_ext_ip) {
				snprintf(prd_ext_ip, sizeof(prd_ext_ip), "%.*s", (int)pj_ext_ip->slen, pj_ext_ip->ptr);
				pExtPort = strrchr(prd_ext_ip, ':');
				if (pExtPort != NULL) {
					*pExtPort = '\0';
					++pExtPort;
					prd_ext_ip_port = atoi(pExtPort);
				}
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		if (ast_strlen_zero(prd_ip) && uri_from) {
			ast_copy_pj_str(prd_ip, &(uri_from->host), strlen(prd_ip));
		}
		if (uri_from) {
			/* If the gdms url can not compare, use the header ip */
			ast_copy_pj_str(prd_gdms_header_ip, &(uri_from->host), strlen(prd_gdms_header_ip));
		}
#endif
		snprintf(prd_checksum, sizeof(prd_checksum), "%.*s", (int)pj_checksum->slen, pj_checksum->ptr);
		ast_log(LOG_DEBUG, "prd_name = %s, prd_addr = %s:%d, prd_ext_ip_add = [%s:%d] prd_checksum = [%s] prd_mode = [%s] prd_device_id = [%s] prd_device_prefix = [%s]", prd_name, prd_ip, prd_port, prd_ext_ip, prd_ext_ip_port, prd_checksum, prd_mode, prd_device_id, prd_device_prefix);
		/* added by xxcui for Bug 78011 */
		if ((last_options_cseq == cseq) && !strcasecmp(last_options_prd_ip, prd_ip)) {
			return PJ_TRUE;
		} else {
			last_options_cseq = cseq;
			memset(last_options_prd_ip, 0, sizeof(last_options_prd_ip));
			memcpy(last_options_prd_ip, prd_ip, strlen(last_options_prd_ip) > strlen(prd_ip) ? strlen(prd_ip) : strlen(last_options_prd_ip));
		}
		char* argv[16] = {NULL};
		int idx= 0;
		argv[idx++] = "/usr/bin/syncRemoteLDAP.sh";
		/*UPdate list*/
		AST_LIST_LOCK(&(g_ldap_sync_list.ast_ldap_server_list));
		AST_LIST_TRAVERSE(&(g_ldap_sync_list.ast_ldap_server_list), pstLdapServer, entry)
		{
			if ( pstLdapServer->remote_host != NULL ) {
				if (strcasecmp(pstLdapServer->remote_host, prd_ip)){
					if (strcasecmp(prd_mode, "gdms") == 0)/* add by wylu */
					{
						ast_log(LOG_DEBUG, "The prd mode is gdms, don't need to exchange url to ip,just compare web public url!");
					}else{
						url_to_IP(pstLdapServer->remote_host, host_address, sizeof(host_address));
						ast_log(LOG_DEBUG, "Remote host is:[%s], toIP[%s], PRD_IP is [%s]", pstLdapServer->remote_host, host_address, prd_ip);
					}
				}else{
					ast_log(LOG_DEBUG, "Remote host is:[%s], PRD_IP is [%s]", pstLdapServer->remote_host,  prd_ip);
				}
				if ( 1 == pstLdapServer->enabled && !ast_strlen_zero(prd_ip) && !ast_strlen_zero(pstLdapServer->remote_host) && (strcasecmp(pstLdapServer->remote_host, prd_ip) == 0 || check_equal_addr(host_address, prd_ip) == 0 || (strcasecmp(prd_mode, "gdms") == 0 && strcasecmp(pstLdapServer->remote_host, prd_gdms_header_ip) == 0))) {
					pstLdapServer->remote_port = prd_port;
					argv[idx++] = prd_ip;
					argv[idx++] = ast_strlen_zero(pPort) ? "0" : pPort;
					argv[idx++] = pstLdapServer->trunk_name;
					argv[idx++] = pstLdapServer->decrypt_key;
					argv[idx++] = prd_checksum;
					argv[idx++] = prd_mode;
					argv[idx++] = prd_name;
					argv[idx++] = pstLdapServer->extension_prefix==NULL?"NULL":pstLdapServer->extension_prefix;
					argv[idx++] = prd_device_id;
					argv[idx++] = prd_device_prefix;
					argv[idx++] = NULL;
					snprintf(command, sizeof(command), "syncRemoteLDAP.sh `cat << '%s'\n%s %d %s %s %s %s %s %s %s %s\n%s\n` &",
							HERE_DOC_MARKE, prd_ip, prd_port, pstLdapServer->trunk_name, pstLdapServer->decrypt_key, prd_checksum,
							prd_mode, prd_name, pstLdapServer->extension_prefix==NULL?"NULL":pstLdapServer->extension_prefix, prd_device_id, prd_device_prefix, HERE_DOC_MARKE);
					break;
				}
				else if ( 1 == pstLdapServer->enabled && !ast_strlen_zero(prd_ext_ip) && (strcasecmp(pstLdapServer->remote_host, prd_ext_ip) == 0 || check_equal_addr(host_address, prd_ext_ip) == 0)) {
					pstLdapServer->remote_port = prd_ext_ip_port;
					argv[idx++] = prd_ext_ip;
					argv[idx++] = ast_strlen_zero(pExtPort) ? "0" : pExtPort;
					argv[idx++] = pstLdapServer->trunk_name;
					argv[idx++] = pstLdapServer->decrypt_key;
					argv[idx++] = prd_checksum;
					argv[idx++] = prd_mode;
					argv[idx++] = prd_name;
					argv[idx++] = pstLdapServer->extension_prefix==NULL?"NULL":pstLdapServer->extension_prefix;
					argv[idx++] = prd_device_id;
					argv[idx++] = prd_device_prefix;
					argv[idx++] = NULL;
					snprintf(command, sizeof(command), "syncRemoteLDAP.sh `cat << '%s'\n%s %d %s %s %s %s %s %s %s %s\n%s\n` &",
							HERE_DOC_MARKE, prd_ext_ip, prd_ext_ip_port, pstLdapServer->trunk_name, pstLdapServer->decrypt_key, prd_checksum,
							prd_mode, prd_name, pstLdapServer->extension_prefix==NULL?"NULL":pstLdapServer->extension_prefix, prd_device_id, prd_device_prefix, HERE_DOC_MARKE);
					break;
				}

			}
		}
		AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));
		ast_safe_execvp(1,argv[0], argv);
		ast_log(LOG_DEBUG, "command is : %s\n", command);

		return PJ_TRUE;
	} else {
		return PJ_FALSE;
	}
}

static pj_bool_t optnot_on_rx_request(pjsip_rx_data *rdata)
{
	if (!pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_notify_method)) {
		return notify_on_rx_request(rdata);
	} else if (!pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_options_method)) {
		return options_on_rx_request(rdata);
	}

	return PJ_FALSE;
}

#if 0
static void gs_ldap_get_settings_by_trunk_name(struct st_ldap_server  *node, const char *name)
{
	char buffer[256] = {0};
	char trunk_context[128] = {0};
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = {0};

	if ( node == NULL || name == NULL ) {
		ast_log(LOG_WARNING, "LDAP get settings: Invalid parameter !");
		return;
	}

	snprintf(trunk_context, sizeof(trunk_context), "ldap_%s", name);
	ast_log(LOG_VERBOSE, "Enter gs_ldap_get_settings_by_trunk_name [%s] !!!\n", trunk_context);

	/*Parsing context from pjsip_ldap.conf*/
	if ((cfg = ast_config_load("pjsip_ldap.conf", config_flags)) && cfg != CONFIG_STATUS_FILEINVALID
			&& cfg != CONFIG_STATUS_FILEUNCHANGED) {
		ast_log(LOG_VERBOSE, "pjsip_ldap.conf is OK!!!");

		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_ENABLE), "no"), sizeof(buffer));
		node->trunk_context= ast_strdup(trunk_context);
		if ( strcasecmp(buffer, "yes") == 0 ) {
			node->enabled = 1;
		}
		memset(buffer, 0, sizeof(buffer));

		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_KEY), ""), sizeof(buffer));
		if ( buffer[0] != '\0' ) {
			if ( node->decrypt_key != NULL ) {
				ast_free(node->decrypt_key);
				node->decrypt_key = NULL;
			}
			node->decrypt_key = ast_strdup(buffer);
		}
		memset(buffer, 0, sizeof(buffer));

		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_LOCAL_PORT), ""), sizeof(buffer));
		if ( buffer[0] != '\0' ) {
			node->local_port = atoi(buffer);
		}
		memset(buffer, 0, sizeof(buffer));

		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_EXT_PERFIX), ""), sizeof(buffer));
		if ( buffer[0] != '\0' ) {
			if ( node->extension_prefix != NULL ) {
				ast_free(node->extension_prefix);
				node->extension_prefix = NULL;
			}
			node->extension_prefix = ast_strdup(buffer);
		}
		memset(buffer, 0, sizeof(buffer));

		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_TRUNK_NAME), ""), sizeof(buffer));
		if ( buffer[0] != '\0' ) {
			if ( node->trunk_name != NULL ) {
				ast_free(node->trunk_name);
				node->trunk_name = NULL;
			}
			node->trunk_name = ast_strdup(buffer);

			if ( node->trunk_name_markid != NULL) {
				ast_free(node->trunk_name_markid);
				node->trunk_name_markid = NULL;
			}
			node->trunk_name_markid = get_random_string();
		}
		memset(buffer, 0, sizeof(buffer));

		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_TRUNK_HOST), ""), sizeof(buffer));
		if ( buffer[0] != '\0' ) {
			if ( node->remote_host != NULL ) {
				ast_free(node->remote_host);
				node->remote_host = NULL;
			}
			node->remote_host = ast_strdup(buffer);
		}
		memset(buffer, 0, sizeof(buffer));
#ifdef GRANDSTREAM_NETWORKS
		/*BEGIN: Add by weidu ,for ldap*/
		ast_copy_string(buffer, S_OR(ast_variable_retrieve(cfg, trunk_context, LDAP_SYNC_TYPE), ""), sizeof(buffer));
		if ( buffer[0] != '\0' && strcasecmp(buffer, "rsync") == 0) {
			node->sync_type = SYNC_RSYNC_TYPE;

		}else{
			node->sync_type = SYNC_WGET_TYPE;
		}
		memset(buffer, 0, sizeof(buffer));
#endif
		/*END: Add by weidu ,for ldap*/
		ast_config_destroy(cfg);
	}

	ast_log(LOG_DEBUG, "Trunk Name : %s", node->trunk_name==NULL?"":node->trunk_name);
	ast_log(LOG_DEBUG, "Trunk Name mark id: %s", node->trunk_name_markid==NULL?"":node->trunk_name_markid);
	ast_log(LOG_DEBUG, "Enabled    : %d", node->enabled);
	ast_log(LOG_DEBUG, "Key        : %s", node->decrypt_key==NULL?"":node->decrypt_key);
	ast_log(LOG_DEBUG, "Prefix     : %s", node->extension_prefix==NULL?"":node->extension_prefix);
	ast_log(LOG_DEBUG, "Local Port : %d", node->local_port);
	ast_log(LOG_DEBUG, "Remote Host: %s", node->remote_host==NULL?"":node->remote_host);
	ast_log(LOG_DEBUG, "sysnc type: %d", node->sync_type);
	ast_log(LOG_VERBOSE, "Exit gs_ldap_get_settings_by_trunk_name [%s] !!!\n", trunk_context);

}

static void gs_ldap_get_trunk_endpoint(void)
{
	ast_log(LOG_VERBOSE, "Enter gs_ldap_get_trunk_endpoint !!!\n");

	struct st_ldap_server  *pstLdapServer    = NULL;
	struct st_ldap_server  *pstLdapServerNew = NULL;
	struct ast_sip_endpoint *pstEndpoint     = NULL;
	struct ao2_iterator     i;
	/* Set random number seed for trunk name mark id*/
	srand(time(NULL));
	AST_LIST_LOCK(&g_ldap_sync_list.ast_ldap_server_list);
	/* remove old list member */
	while ((pstLdapServer = AST_LIST_REMOVE_HEAD(&g_ldap_sync_list.ast_ldap_server_list, entry))) {
		if (pstLdapServer->decrypt_key != NULL) {
			ast_free(pstLdapServer->decrypt_key);
			pstLdapServer->decrypt_key = NULL;
		}

		if (pstLdapServer->extension_prefix != NULL) {
			ast_free(pstLdapServer->extension_prefix);
			pstLdapServer->extension_prefix = NULL;
		}

		if (pstLdapServer->remote_host != NULL) {
			ast_free(pstLdapServer->remote_host);
			pstLdapServer->remote_host = NULL;
		}

		if (pstLdapServer->trunk_name != NULL) {
			ast_free(pstLdapServer->trunk_name);
			pstLdapServer->trunk_name = NULL;
		}

		if (pstLdapServer->trunk_context != NULL) {
			ast_free(pstLdapServer->trunk_context);
			pstLdapServer->trunk_context = NULL;
		}

		if (pstLdapServer->endpoint != NULL) {
			ao2_ref(pstLdapServer->endpoint, -1);
		}
		if ( pstLdapServer->trunk_name_markid != NULL) {
			ast_free(pstLdapServer->trunk_name_markid);
			pstLdapServer->trunk_name_markid = NULL;
		}

		ast_free(pstLdapServer);
	}

	AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));

	g_ldap_sync_list.sip_tk_num = 0;

	/* inset new list member */
	RAII_VAR(struct ao2_container *, endpoints, ast_sip_get_endpoints(), ao2_cleanup);
	i = ao2_iterator_init(endpoints, 0);
	while ((pstEndpoint = ao2_t_iterator_next(&i, "iterate thruk endpoints table"))) {
		ao2_lock(pstEndpoint);
		const char *name = ast_sorcery_object_get_id(pstEndpoint);

		if (0 == strstr(name, "trunk_")) {
			ao2_unlock(pstEndpoint);
			ao2_cleanup(pstEndpoint);
			continue;
		}

		pstLdapServerNew = ast_calloc(1, sizeof(struct st_ldap_server));
		if (NULL == pstLdapServerNew) {
			ao2_unlock(pstEndpoint);
			ao2_cleanup(pstEndpoint);
			continue;
		}

		pstLdapServerNew->decrypt_key      = NULL;
		pstLdapServerNew->extension_prefix = NULL;
		pstLdapServerNew->remote_host      = NULL;
		pstLdapServerNew->trunk_name       = NULL;
		pstLdapServerNew->trunk_context    = NULL;
		pstLdapServerNew->enabled          = 0;
		pstLdapServerNew->local_port       = 0;
		pstLdapServerNew->remote_port      = 0;
		pstLdapServerNew->endpoint         = pstEndpoint;
#ifdef GRANDSTREAM_NETWORKS
		pstLdapServerNew->from_gdms = 0;
		pstLdapServerNew->sync_type = SYNC_WGET_TYPE;
#endif
		ao2_ref(pstLdapServerNew->endpoint, +1);
		gs_ldap_get_settings_by_trunk_name(pstLdapServerNew, name);

		AST_LIST_LOCK(&(g_ldap_sync_list.ast_ldap_server_list));
		AST_LIST_INSERT_TAIL(&g_ldap_sync_list.ast_ldap_server_list, pstLdapServerNew, entry);
		AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));
		g_ldap_sync_list.sip_tk_num++;

		ao2_unlock(pstEndpoint);
		ao2_cleanup(pstEndpoint);
	}
	ao2_iterator_destroy(&i);

	ast_log(LOG_DEBUG, "gs_ldap_get_trunk_endpoint, has [%d] sip trunk", g_ldap_sync_list.sip_tk_num);
	ast_log(LOG_VERBOSE, "Exit gs_ldap_get_trunk_endpoint !!!\n");

}
#endif

static void sip_ldap_trunk_list_free(void)
{
	struct st_ldap_server  *pstLdapServer = NULL;

	AST_LIST_LOCK(&g_ldap_sync_list.ast_ldap_server_list);

	/* remove old list member */
	while ((pstLdapServer = AST_LIST_REMOVE_HEAD(&g_ldap_sync_list.ast_ldap_server_list, entry))) {
		if (pstLdapServer->decrypt_key != NULL) {
			ast_free(pstLdapServer->decrypt_key);
			pstLdapServer->decrypt_key = NULL;
		}

		if (pstLdapServer->extension_prefix != NULL) {
			ast_free(pstLdapServer->extension_prefix);
			pstLdapServer->extension_prefix = NULL;
		}

		if (pstLdapServer->remote_host != NULL) {
			ast_free(pstLdapServer->remote_host);
			pstLdapServer->remote_host = NULL;
		}

		if (pstLdapServer->trunk_name != NULL) {
			ast_free(pstLdapServer->trunk_name);
			pstLdapServer->trunk_name = NULL;
		}

		if (pstLdapServer->trunk_context != NULL) {
			ast_free(pstLdapServer->trunk_context);
			pstLdapServer->trunk_context = NULL;
		}

		if (pstLdapServer->endpoint != NULL) {
			ao2_ref(pstLdapServer->endpoint, -1);
		}
		if ( pstLdapServer->trunk_name_markid != NULL) {
			ast_free(pstLdapServer->trunk_name_markid);
			pstLdapServer->trunk_name_markid = NULL;
		}

		ast_free(pstLdapServer);
	}

	AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));

	g_ldap_sync_list.sip_tk_num = 0;
}

static void gs_ldap_get_trunk_endpoint(int reload)
{
	struct ast_config *cfg = NULL;
	struct ast_variable *var;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	const char *cat = NULL;
	const char *type = NULL;
	const char *tk_id = NULL;
	struct ast_sip_endpoint *endpoint = NULL;
	struct st_ldap_server  *pstLdapServerNew = NULL;
	int res = 0;
	char *ldap_sync_password_decrypt = NULL;
	cfg = ast_config_load("pjsip_ldap.conf", config_flags);
	if (!cfg || (cfg == CONFIG_STATUS_FILEINVALID) || (cfg == CONFIG_STATUS_FILEUNCHANGED) || (cfg == CONFIG_STATUS_FILEMISSING)) {
		return;
	}

	if (reload) {
		//free ldap trunk list
		sip_ldap_trunk_list_free();
	}

	/* Set random number seed for trunk name mark id*/
	srand(time(NULL));

	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		if (!(type = ast_variable_retrieve(cfg, cat, "type"))) {
			continue;
		}

		if (strcasecmp(type, "ldapserver")) {
			continue;
		}

		if (strncmp(cat, "ldap_", 5)) {
			continue;
		}

		tk_id = cat + strlen("ldap_");
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", tk_id);
		if (!endpoint) {
			continue;
		}

		pstLdapServerNew = ast_calloc(1, sizeof(struct st_ldap_server));
		if (NULL == pstLdapServerNew) {
			ao2_cleanup(endpoint);
			continue;
		}

		pstLdapServerNew->decrypt_key      = NULL;
		pstLdapServerNew->extension_prefix = NULL;
		pstLdapServerNew->remote_host      = NULL;
		pstLdapServerNew->trunk_name       = NULL;
		pstLdapServerNew->trunk_context    = ast_strdup(cat);
		pstLdapServerNew->enabled          = 0;
		pstLdapServerNew->local_port       = 0;
		pstLdapServerNew->remote_port      = 0;
		pstLdapServerNew->endpoint         = endpoint;
		pstLdapServerNew->from_gdms = 0;
		pstLdapServerNew->sync_type = SYNC_WGET_TYPE;

		for (var = ast_variable_browse(cfg, cat); var; var = var->next) {
			if (!strcasecmp(var->name, "ldap_sync_enable") && ast_true(var->value)) {
				pstLdapServerNew->enabled = 1;
			} else if (!strcasecmp(var->name, "ldap_sync_passwd")) {
				if (!ast_strlen_zero(var->value)) {
					ucm_secret_decrypt(var->value, &ldap_sync_password_decrypt);
					if (ldap_sync_password_decrypt == NULL || strlen(ldap_sync_password_decrypt) == 0)
						pstLdapServerNew->decrypt_key = ast_strdup(var->value);
					else
						pstLdapServerNew->decrypt_key = ast_strdup(ldap_sync_password_decrypt);
					AST_SAFE_FREE(ldap_sync_password_decrypt);
				}
			} else if (!strcasecmp(var->name, "ldap_sync_port")) {
				if (!ast_strlen_zero(var->value)) {
					pstLdapServerNew->local_port = atoi(var->value);
				}
			} else if (!strcasecmp(var->name, "ldap_default_outrt_prefix")){
				if (!ast_strlen_zero(var->value)) {
					pstLdapServerNew->extension_prefix = ast_strdup(var->value);
				}
			} else if (!strcasecmp(var->name, "trunk_name")){
				if (!ast_strlen_zero(var->value)) {
					pstLdapServerNew->trunk_name = ast_strdup(var->value);
					pstLdapServerNew->trunk_name_markid = get_random_string();
				}
			} else if (!strcasecmp(var->name, "host")) {
				if (!ast_strlen_zero(var->value)) {
					pstLdapServerNew->remote_host = ast_strdup(var->value);
				}
			} else if (!strcasecmp(var->name, "ldap_sync_type")) {
				if (!ast_strlen_zero(var->value) && !strcasecmp(var->value, "rsync")) {
					pstLdapServerNew->sync_type = SYNC_RSYNC_TYPE;
				}
			}
		}

		ast_log(LOG_DEBUG, "Trunk Name : %s", pstLdapServerNew->trunk_name==NULL?"":pstLdapServerNew->trunk_name);
		ast_log(LOG_DEBUG, "Trunk Name Mark id: %s", pstLdapServerNew->trunk_name_markid==NULL?"":pstLdapServerNew->trunk_name_markid);
		ast_log(LOG_DEBUG, "Enabled    : %d", pstLdapServerNew->enabled);
		ast_log(LOG_DEBUG, "Key        : %s", pstLdapServerNew->decrypt_key==NULL?"":pstLdapServerNew->decrypt_key);
		ast_log(LOG_DEBUG, "Prefix     : %s", pstLdapServerNew->extension_prefix==NULL?"":pstLdapServerNew->extension_prefix);
		ast_log(LOG_DEBUG, "Local Port : %d", pstLdapServerNew->local_port);
		ast_log(LOG_DEBUG, "Remote Host: %s", pstLdapServerNew->remote_host==NULL?"":pstLdapServerNew->remote_host);
		ast_log(LOG_DEBUG, "sysnc type : %d", pstLdapServerNew->sync_type);

		AST_LIST_LOCK(&(g_ldap_sync_list.ast_ldap_server_list));
		AST_LIST_INSERT_TAIL(&g_ldap_sync_list.ast_ldap_server_list, pstLdapServerNew, entry);
		AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));
		g_ldap_sync_list.sip_tk_num++;
	}

	ast_log(LOG_DEBUG, "gs_ldap_get_trunk_endpoint, [%d] sip trunks in total", g_ldap_sync_list.sip_tk_num);
	ast_config_destroy(cfg);

}

static void sip_ldap_sync_all(void)
{
	ast_log(LOG_VERBOSE, "Enter sip_ldap_sync_all !!!\n");

	struct st_ldap_server  *pstLdapServer = NULL;
	char   command[512] = {0};
	char host[256] = {0};  /*Add by weidu , for ldap*/
	char gs_port[32] = {0};

	if ( 0 == g_ldap_changed)
	{
		ast_log(LOG_DEBUG, "sip_ldap_sync_all, no change\n");
		return;
	}

	if (0 == g_ldap_sync_list.sip_tk_num)
	{
		ast_log(LOG_DEBUG, "sip_ldap_sync_all, no sip trunk\n");
		return;
	}

	AST_LIST_LOCK(&(g_ldap_sync_list.ast_ldap_server_list));
	AST_LIST_TRAVERSE(&(g_ldap_sync_list.ast_ldap_server_list), pstLdapServer, entry) {
		if ( NULL != pstLdapServer->endpoint && pstLdapServer->enabled == 1 ) {
			ast_log(LOG_DEBUG, "prepareLDAPSync for trunk [%s] ~", pstLdapServer->trunk_name);
			ast_log(LOG_DEBUG, "prepareLDAPSync for <%s>", pstLdapServer->decrypt_key==NULL?"NULL":pstLdapServer->decrypt_key);
#ifdef GRANDSTREAM_NETWORKS
			/*BEGIN: Add by weidu ,for ldap*/
			ast_sip_get_nvram_value(":WsProxyHost", host, sizeof(host));  
			ast_sip_get_nvram_value("gswave_port", gs_port, sizeof(gs_port));

			if (!ast_strlen_zero(gs_port)){ /*use gswave port*/
				pstLdapServer->local_port = atoi(gs_port);
			}else{
				pstLdapServer->local_port = 8090;
			}
			if (!ast_strlen_zero(host)){
				pstLdapServer->from_gdms = 1;
			}else{
				pstLdapServer->from_gdms = 0;
			}
			ast_log(LOG_DEBUG, "sip_ldap_sync_manual pstLdapServer->from_gdms %d \n" , pstLdapServer->from_gdms);
#endif
			/*END: Add by weidu, for ldap*/
#ifdef GRANDSTREAM_NETWORKS
			/*BEGIN: Mod by weidu ,for 188505, reload/load module ldap sync*/
			if (pstLdapServer->sync_type == SYNC_RSYNC_TYPE){
				snprintf(command, sizeof(command), "prepareLDAPFile.sh `cat << '%s'\n%s %s rsync NULL \n%s\n`", HERE_DOC_MARKE, pstLdapServer->decrypt_key==NULL?g_ldap_sync_key:pstLdapServer->decrypt_key, pstLdapServer->trunk_name_markid, HERE_DOC_MARKE);
			}else{
				snprintf(command, sizeof(command), "prepareLDAPFile.sh `cat << '%s'\n%s %s wget NULL \n%s\n`", HERE_DOC_MARKE, pstLdapServer->decrypt_key==NULL?g_ldap_sync_key:pstLdapServer->decrypt_key, pstLdapServer->trunk_name_markid, HERE_DOC_MARKE);
			}
			ast_log(LOG_DEBUG, "command = (%s)", command);
			/*END: Mod by weidu ,for 188505, reload/load module ldap sync*/
#endif
			ast_safe_system2(command);

			if (ast_sip_push_task(NULL, sip_poke_endpoint, pstLdapServer->endpoint)) {
				ast_log(LOG_ERROR, "Can't send options to sync ldap!\n");
				return;
			}
		}
	}
	AST_LIST_UNLOCK(&(g_ldap_sync_list.ast_ldap_server_list));

	g_ldap_changed = 0;

	ast_log(LOG_VERBOSE, "Exit sip_ldap_sync_all !!!\n");


	return;
}

static void log_device_infomation(void)
{
	ast_log(LOG_DEBUG, "exp_ip      : %s", g_ldap_device_mode.exp_ip?g_ldap_device_mode.exp_ip:"");
	ast_log(LOG_DEBUG, "device_mode : %d", g_ldap_device_mode.device_mode);
	ast_log(LOG_DEBUG, "wan_device  : %s", g_ldap_device_mode.wan_device?g_ldap_device_mode.wan_device:"");
	ast_log(LOG_DEBUG, "lan_device  : %s", g_ldap_device_mode.lan_device?g_ldap_device_mode.lan_device:"");
	ast_log(LOG_DEBUG, "default_if  : %s", g_ldap_device_mode.default_if?g_ldap_device_mode.default_if:"");
	ast_log(LOG_DEBUG, "lan1_device : %s", g_ldap_device_mode.lan1_device?g_ldap_device_mode.lan1_device:"");
	ast_log(LOG_DEBUG, "lan2_device : %s\n", g_ldap_device_mode.lan2_device?g_ldap_device_mode.lan2_device:"");
}

static void unload_device_information(void)
{
	ast_free(g_ldap_device_mode.exp_ip);
	ast_free(g_ldap_device_mode.lan_device);
	ast_free(g_ldap_device_mode.lan1_device);
	ast_free(g_ldap_device_mode.lan2_device);
	ast_free(g_ldap_device_mode.wan_device);
	ast_free(g_ldap_device_mode.default_if);
	/*BEGIN: Add by weidu , for 290147*/
	g_ldap_device_mode.exp_ip = NULL;
	g_ldap_device_mode.lan_device = NULL;
	g_ldap_device_mode.lan1_device = NULL;
	g_ldap_device_mode.lan2_device = NULL;
	g_ldap_device_mode.wan_device = NULL;
	g_ldap_device_mode.default_if = NULL;
	/*END: Add by weidu , for 290147*/
}

static void load_device_information(int reload)
{
	char *tmp = NULL;
	struct ast_config *cfg;
	struct ast_variable *var = NULL;
	struct ast_flags config_flags = {reload ? CONFIG_FLAG_FILEUNCHANGED : 0};

	if (!(cfg = ast_config_load("device_mode.conf", config_flags))) {
		ast_log(LOG_NOTICE, "No device mode config file (device_mode.conf)\n");
		return ;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return ;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "device_mode.conf is in an invalid format.  Aborting.\n");
		return ;
	}

	unload_device_information();

	while ((tmp = ast_category_browse(cfg, tmp)) ) {
		if (strstr(tmp, "global"))
		for (var = ast_variable_browse(cfg, tmp); var; var = var->next) {
			if (!strcasecmp(var->name, "exp_ip")) {
				g_ldap_device_mode.exp_ip = ast_strdup(var->value);
			}else if (!strcasecmp(var->name, "device_mode")){
				g_ldap_device_mode.device_mode = atoi(var->value);
			}else if (!strcasecmp(var->name, "wan_device")){
				g_ldap_device_mode.wan_device = ast_strdup(var->value);
			}else if (!strcasecmp(var->name, "lan_device")){
				g_ldap_device_mode.lan_device = ast_strdup(var->value);
			}else if (!strcasecmp(var->name, "default_if")){
				g_ldap_device_mode.default_if = ast_strdup(var->value);
			}else if (!strcasecmp(var->name, "lan1_device")){
				g_ldap_device_mode.lan1_device = ast_strdup(var->value);
			}else if (!strcasecmp(var->name, "lan2_device")){
				g_ldap_device_mode.lan2_device = ast_strdup(var->value);
			}
		}
	}
	log_device_infomation();

#ifdef GRANDSTREAM_NETWORKS
	ast_config_destroy(cfg);
#endif
}

#if 0
static int check_need_reload_ldap_server(void)
{
	struct ast_config *cfg = NULL;
	struct ast_flags config_flags = {CONFIG_FLAG_FILEUNCHANGED};

	/*Parsing context from pjsip_endpoints.conf*/
	if ((cfg = ast_config_load("pjsip_ldap.conf", config_flags)) && 
		(cfg != CONFIG_STATUS_FILEINVALID) && 
		(cfg != CONFIG_STATUS_FILEUNCHANGED) &&
		(cfg != CONFIG_STATUS_FILEMISSING)) {

		ast_config_destroy(cfg);
		return 1;
	}

	ast_log(LOG_DEBUG, "ldap sync not changed.\n");
	return 0;
}
#endif

static int load_module(void)
{
	CHECK_PJSIP_MODULE_LOADED();

	if (ast_sip_register_service(&notify_module)) {
		ast_log(LOG_ERROR, "Could not register notify service\n");
		return AST_MODULE_LOAD_FAILURE;
	}

	ast_log(LOG_DEBUG, "ldap sync feature PROCESS START\n");
	gs_ldap_get_trunk_endpoint(0);
	getNeedToSyncLDAP();
	sip_ldap_sync_all();
	load_device_information(0);
	ast_log(LOG_DEBUG, "ldap sync feature PROCESS END\n");

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_service(&notify_module);
	unload_device_information();
	return 0;
}

static int reload(void)
{
	ast_log(LOG_DEBUG, "Sync ldap module reload\n");
	//if (check_need_reload_ldap_server()) {
		gs_ldap_get_trunk_endpoint(1);
	//}

	getNeedToSyncLDAP();
	sip_ldap_sync_all();
	ast_mutex_lock(&g_device_info_mutex);
	//unload_device_information();
	load_device_information(1);
	ast_mutex_unlock(&g_device_info_mutex);
#ifndef GRANDSTREAM_NETWORKS
	return 1;
#else
	return AST_MODULE_LOAD_SUCCESS;
#endif
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP sync LDAP",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_APP_DEPEND,
);
