/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2020, Digium, Inc.
 *
 * binfeng du <bfdu@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjmedia.h>
#include <pjlib.h>

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/module.h"
#include "asterisk/res_pjsip_trickle_ice_info.h"
#include "asterisk/avs_controller.h"

AST_MODULE_LOG("pjsip");

static int is_trickle_ice_info_type(pjsip_rx_data *rdata, char *subtype)
{
	return rdata->msg_info.ctype
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.type, "application")
		&& !pj_strcmp2(&rdata->msg_info.ctype->media.subtype, subtype);
}

static void send_response(struct ast_sip_session *session,
			  struct pjsip_rx_data *rdata, int code)
{
	pjsip_tx_data *tdata;
	pjsip_dialog *dlg = session->inv_session->dlg;

	if (pjsip_dlg_create_response(dlg, rdata, code, NULL, &tdata) == PJ_SUCCESS) {
		struct pjsip_transaction *tsx = pjsip_rdata_get_tsx(rdata);
		pjsip_dlg_send_response(dlg, tsx, tdata);
	}
}

static unsigned int compare_ufrag_pwd_attr(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, char *remote_ufrag, char *remote_pwd)
{
	const char *local_ufrag = NULL;
	const char *local_pwd = NULL;

	if (!session || !session_media || !session_media->rtp || !remote_ufrag || !remote_pwd) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get session_media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!(local_ufrag = ast_rtp_instance_get_remote_ice_ufrag(session_media->rtp))) {
		ast_log(LOG_DEBUG, "[session:%p] Has no remote ufrag, don't need process trickle ICE for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (strcmp(local_ufrag, remote_ufrag)) {
		return 0;
	}

	if (!(local_pwd = ast_rtp_instance_get_remote_ice_pwd(session_media->rtp))) {
		ast_log(LOG_DEBUG, "[session:%p] Has no remote password, don't need process trickle ICE for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (strcmp(local_pwd, remote_pwd)) {
		return 0;
	}

	return 1;
}

static struct ast_sip_session_media *get_session_media_as_match_ice_attr(struct ast_sip_session *session,
	pjmedia_sdp_media *remote_stream)
{
	unsigned int position = 0;
	char remote_ufrag[256] = { 0 };
	char remote_pwd[256] = { 0 };
	pjmedia_sdp_attr *attr = NULL;
	struct ast_sip_session_media *session_media = NULL;
	struct ast_sip_session_media_state *media_state = NULL;

	if (!session || !session->channel || !remote_stream) {
		ast_log(LOG_DEBUG, "[session:%p] Param is NULL, don't get session_media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	if (!session->pending_media_state || !session->pending_media_state->topology) {
		media_state = session->active_media_state;
	} else {
		media_state = session->pending_media_state;
	}
	if (!media_state) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no media_state\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-ufrag", NULL);
	if (!attr) {
		ast_log(LOG_DEBUG, "[session:%p] Remote media has no ice-ufrag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}
	ast_copy_pj_str(remote_ufrag, (pj_str_t*)&attr->value, sizeof(remote_ufrag));

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-pwd", NULL);
	if (!attr) {
		ast_log(LOG_DEBUG, "[session:%p] Remote media has no ice-pwd for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}
	ast_copy_pj_str(remote_pwd, (pj_str_t*)&attr->value, sizeof(remote_pwd));

	for (position = 0; position < AST_VECTOR_SIZE(&media_state->sessions); position++) {
		session_media = AST_VECTOR_GET(&media_state->sessions, position);
		if (!compare_ufrag_pwd_attr(session, session_media, remote_ufrag, remote_pwd)) {
			continue;
		}

		ast_log(LOG_DEBUG, "[session:%p] Find session_media:%s at postion:%d, remote ufrag and pwd as the same for endpoint:%s\n",
				session, ast_codec_media_sub_type2str(session_media->sub_type), position, ast_sorcery_object_get_id(session->endpoint));

		return session_media;
	}

	return NULL;
}

static int is_candidate_already_exsit(struct ast_sip_session_media *session_media, struct ast_rtp_engine_ice_candidate *new_candidate)
{
	RAII_VAR(struct ao2_container *, remote_candidates, NULL, ao2_cleanup);
	RAII_VAR(struct ast_rtp_engine_ice_candidate *, candidate, NULL, ao2_cleanup);

	if (!session_media || !session_media->rtp || !new_candidate) {
		return 0;
	}

	remote_candidates = ast_rtp_instance_get_remote_candidates(session_media->rtp);

	candidate = ao2_find(remote_candidates, new_candidate, OBJ_POINTER);

	if (!candidate) {
		return 0;
	}

	return 1;
}

static void process_ice_candidate(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *remote_stream)
{
	unsigned int added = 0;
	unsigned int attr_i = 0;
	pjmedia_sdp_attr *attr = NULL;
	char attr_value[256] = { 0 };
	struct ast_rtp_engine_ice *ice = NULL;

	if (!session || !session_media || !session_media->rtp || !remote_stream) {
		return;
	}

	/* If ICE support is not enabled or available exit early */
	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		return;
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		char foundation[33], transport[32], address[PJ_INET6_ADDRSTRLEN + 1], cand_type[6], relay_address[PJ_INET6_ADDRSTRLEN + 1] = "";
		unsigned int port, relay_port = 0;
		struct ast_rtp_engine_ice_candidate candidate = { 0, };

		attr = remote_stream->attr[attr_i];

		/* If this is not a candidate line skip it */
		if (pj_strcmp2(&attr->name, "candidate")) {
			continue;
		}

		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));

		if (sscanf(attr_value, "%32s %30u %31s %30u %46s %30u typ %5s %*s %23s %*s %30u", foundation, &candidate.id, transport,
			(unsigned *)&candidate.priority, address, &port, cand_type, relay_address, &relay_port) < 7) {
			/* Candidate did not parse properly */
			ast_log(LOG_DEBUG, "[session:%p] Can't parse Candidate:%s for endpoint:%s\n",
				session, attr_value, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		if (session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux && candidate.id > 1) {
			/* Remote side may have offered RTP and RTCP candidates. However, if we're using RTCP MUX,
			 * then we should ignore RTCP candidates.
			 */
			continue;
		}

		candidate.foundation = foundation;
		candidate.transport = transport;

		ast_sockaddr_parse(&candidate.address, address, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(&candidate.address, port);

		if (!strcasecmp(cand_type, "host")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_HOST;
		} else if (!strcasecmp(cand_type, "srflx")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_SRFLX;
		} else if (!strcasecmp(cand_type, "relay")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_RELAYED;
		} else {
			continue;
		}

		if (!ast_strlen_zero(relay_address)) {
			ast_sockaddr_parse(&candidate.relay_address, relay_address, PARSE_PORT_FORBID);
		}

		if (relay_port) {
			ast_sockaddr_set_port(&candidate.relay_address, relay_port);
		}

		if (is_candidate_already_exsit(session_media, &candidate)) {
			continue;
		}

		ast_log(LOG_DEBUG, "[session:%p] Add new Candidate:%s for endpoint:%s\n",
				session, attr_value, ast_sorcery_object_get_id(session->endpoint));

		ice->add_new_remote_candidate(session_media->rtp, &candidate);

		added = 1;
	}

	if (added) {
		ast_avs_update_cand(session->channel, session_media->rtp);
		ast_rtp_instance_clear_new_remote_candidates(session_media->rtp);
	}

	return;
}

static int trickle_ice_info_process_sdp(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	unsigned int i = 0;
	pjsip_rdata_sdp_info *sdp_info = NULL;
	struct ast_sip_session_media *session_media = NULL;
	pjmedia_sdp_media *remote_stream = NULL;

	if (!session || !rdata) {
		return -1;
	}

	sdp_info = pjsip_rdata_get_special_sdp_info(rdata, "trickle-ice-sdpfrag");
	if (sdp_info && (sdp_info->sdp_err == PJ_SUCCESS) && sdp_info->sdp) {
		for (i = 0; i < sdp_info->sdp->media_count; ++i) {
			remote_stream = sdp_info->sdp->media[i];
			if (!remote_stream) {
				ast_log(LOG_WARNING, "[session:%p] Can't get remote_stream for endpoint:%s, postion:%d\n",
					session, ast_sorcery_object_get_id(session->endpoint), i);
				continue;
			}

			session_media = get_session_media_as_match_ice_attr(session, remote_stream);
			if (!session_media) {
				ast_log(LOG_WARNING, "[session:%p] Can't get session_media for endpoint:%s, postion:%d\n",
					session, ast_sorcery_object_get_id(session->endpoint), i);
				continue;
			}

			process_ice_candidate(session, session_media, remote_stream);
		}
	} else {
		ast_log(LOG_WARNING, "[session:%p] Can't get trickle ice sdp for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return 0;
}

static int trickle_ice_info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	int is_trickle_ice_info = 0;

	if (!session || !rdata || !rdata->msg_info.msg) {
		return 0;
	}

	is_trickle_ice_info = is_trickle_ice_info_type(rdata, TRICKLE_ICE_INFO_SUB_TYPE);
	if (!is_trickle_ice_info) {
		return 0;
	}

	if (!session->channel) {
		ast_log(LOG_DEBUG, "[session:%p] Session channel is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	send_response(session, rdata, 200);

	trickle_ice_info_process_sdp(session, rdata);

	ast_log(LOG_NOTICE, "[session:%p] trickle_ice_info_incoming_request, process this info for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	return 1;
}

static struct ast_sip_session_supplement trickle_ice_info_supplement = {
	.method = "INFO",
	.incoming_request = trickle_ice_info_incoming_request,
};

static int load_module(void)
{
	ast_sip_session_register_supplement(&trickle_ice_info_supplement);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&trickle_ice_info_supplement);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Trickle ICE INFO Support",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.load_pri = AST_MODPRI_APP_DEPEND,
		.requires = "res_pjsip",
);

