/****************************************************************************
 * Asterisk AVS Engine Module.     
 * FILENAME:        $RCSfile: res_rtp_avs_engine.c,v $
 * LAST REVISION:   $Revision: 1.0$
 * LAST MODIFIED:   $Date: 2019/3/3 17:00:00 $
 *
 * Copyright (c) 2019 by Grandstream Networks, Inc.
 * All rights reserved.
 *
 * This material is proprietary to Grandstream Networks, Inc. and,
 * in addition to the above mentioned Copyright, may be
 * subject to protection under other intellectual property
 * regimes, including patents, trade secrets, designs and/or
 * trademarks.
 *
 * Any use of this material for any purpose, except with an
 * express license from Grandstream Networks, Inc. is strictly
 * prohibited.
 *
 *
 * \brief AVS Engine for RTP.
 *
 * \author 
 *
 ***************************************************************************/

/*** MODULEINFO
    <depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_pubsub</depend>
	<depend>res_pjsip_outbound_publish</depend>
	<support_level>core</support_level>
***/

/*** DOCUMENTATION
	<managerEvent language="en_US" name="RTCPSent">
		<managerEventInstance class="EVENT_FLAG_REPORTING">
			<synopsis>Raised when an RTCP packet is sent.</synopsis>
			<syntax>
				<channel_snapshot/>
				<parameter name="SSRC">
					<para>The SSRC identifier for our stream</para>
				</parameter>
				<parameter name="PT">
					<para>The type of packet for this RTCP report.</para>
					<enumlist>
						<enum name="200(SR)"/>
						<enum name="201(RR)"/>
					</enumlist>
				</parameter>
				<parameter name="To">
					<para>The address the report is sent to.</para>
				</parameter>
				<parameter name="ReportCount">
					<para>The number of reports that were sent.</para>
					<para>The report count determines the number of ReportX headers in
					the message. The X for each set of report headers will range from 0 to
					<literal>ReportCount - 1</literal>.</para>
				</parameter>
				<parameter name="SentNTP" required="false">
					<para>The time the sender generated the report. Only valid when
					PT is <literal>200(SR)</literal>.</para>
				</parameter>
				<parameter name="SentRTP" required="false">
					<para>The sender's last RTP timestamp. Only valid when PT is
					<literal>200(SR)</literal>.</para>
				</parameter>
				<parameter name="SentPackets" required="false">
					<para>The number of packets the sender has sent. Only valid when PT
					is <literal>200(SR)</literal>.</para>
				</parameter>
				<parameter name="SentOctets" required="false">
					<para>The number of bytes the sender has sent. Only valid when PT is
					<literal>200(SR)</literal>.</para>
				</parameter>
				<parameter name="ReportXSourceSSRC">
					<para>The SSRC for the source of this report block.</para>
				</parameter>
				<parameter name="ReportXFractionLost">
					<para>The fraction of RTP data packets from <literal>ReportXSourceSSRC</literal>
					lost since the previous SR or RR report was sent.</para>
				</parameter>
				<parameter name="ReportXCumulativeLost">
					<para>The total number of RTP data packets from <literal>ReportXSourceSSRC</literal>
					lost since the beginning of reception.</para>
				</parameter>
				<parameter name="ReportXHighestSequence">
					<para>The highest sequence number received in an RTP data packet from
					<literal>ReportXSourceSSRC</literal>.</para>
				</parameter>
				<parameter name="ReportXSequenceNumberCycles">
					<para>The number of sequence number cycles seen for the RTP data
					received from <literal>ReportXSourceSSRC</literal>.</para>
				</parameter>
				<parameter name="ReportXIAJitter">
					<para>An estimate of the statistical variance of the RTP data packet
					interarrival time, measured in timestamp units.</para>
				</parameter>
				<parameter name="ReportXLSR">
					<para>The last SR timestamp received from <literal>ReportXSourceSSRC</literal>.
					If no SR has been received from <literal>ReportXSourceSSRC</literal>,
					then 0.</para>
				</parameter>
				<parameter name="ReportXDLSR">
					<para>The delay, expressed in units of 1/65536 seconds, between
					receiving the last SR packet from <literal>ReportXSourceSSRC</literal>
					and sending this report.</para>
				</parameter>
			</syntax>
			<see-also>
				<ref type="managerEvent">RTCPReceived</ref>
			</see-also>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="RTCPReceived">
		<managerEventInstance class="EVENT_FLAG_REPORTING">
			<synopsis>Raised when an RTCP packet is received.</synopsis>
			<syntax>
				<channel_snapshot/>
				<parameter name="SSRC">
					<para>The SSRC identifier for the remote system</para>
				</parameter>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='RTCPSent']/managerEventInstance/syntax/parameter[@name='PT'])" />
				<parameter name="From">
					<para>The address the report was received from.</para>
				</parameter>
				<parameter name="RTT">
					<para>Calculated Round-Trip Time in seconds</para>
				</parameter>
				<parameter name="ReportCount">
					<para>The number of reports that were received.</para>
					<para>The report count determines the number of ReportX headers in
					the message. The X for each set of report headers will range from 0 to
					<literal>ReportCount - 1</literal>.</para>
				</parameter>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='RTCPSent']/managerEventInstance/syntax/parameter[@name='SentNTP'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='RTCPSent']/managerEventInstance/syntax/parameter[@name='SentRTP'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='RTCPSent']/managerEventInstance/syntax/parameter[@name='SentPackets'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='RTCPSent']/managerEventInstance/syntax/parameter[@name='SentOctets'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='RTCPSent']/managerEventInstance/syntax/parameter[contains(@name, 'ReportX')])" />
			</syntax>
			<see-also>
				<ref type="managerEvent">RTCPSent</ref>
			</see-also>
		</managerEventInstance>
	</managerEvent>
 ***/
 
#include "asterisk.h"
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>

#ifdef HAVE_OPENSSL
#include <openssl/opensslconf.h>
#include <openssl/opensslv.h>
#if !defined(OPENSSL_NO_SRTP) && (OPENSSL_VERSION_NUMBER >= 0x10001000L)
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#if !defined(OPENSSL_NO_ECDH) && (OPENSSL_VERSION_NUMBER >= 0x10000000L)
#include <openssl/bn.h>
#endif
#ifndef OPENSSL_NO_DH
#include <openssl/dh.h>
#endif
#endif
#endif

#include <pjlib.h>
#include <pjlib-util.h>
#include <pjnath.h>
#include <ifaddrs.h>

#include "asterisk/options.h"
#include "asterisk/stun.h"
#include "asterisk/pbx.h"
#include "asterisk/frame.h"
#include "asterisk/format_cache.h"
#include "asterisk/channel.h"
#include "asterisk/acl.h"
#include "asterisk/config.h"
#include "asterisk/lock.h"
#include "asterisk/utils.h"
#include "asterisk/cli.h"
#include "asterisk/manager.h"
#include "asterisk/unaligned.h"
#include "asterisk/module.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/smoother.h"
#include "asterisk/uuid.h"
#include "asterisk/test.h"
#include "asterisk/data_buffer.h"
#include "asterisk/res_pjproject.h"
#include "asterisk/avs_controller.h"

#ifdef GRANDSTREAM_NETWORKS
AST_MODULE_LOG("RTP");
#endif

#define __AVS_DTLS_WRAPPER__ 1
#define __AVS_ICE_WRAPPER__ 1
#define __AVS_ENGINE_WRAPPER__ 1


/*! \brief RTP session description */
struct avs_rtp {
	int s;
	unsigned int avs_sock_id;
	char avs_client_path[AVS_UNIX_PATH_MAX];
	struct ast_frame f;
	
	int nego_flag;
	unsigned char rawdata[8192 + AST_FRIENDLY_OFFSET];
	unsigned int avs_bridge_id;
	unsigned int avs_chan_id;
	unsigned long avs_port_id;
	unsigned long avs_parent_port_id;
	unsigned long avs_track_id;
	int child_flag;
	
	unsigned int net_version;
	char local_ipaddr[MAX_IPADDR_LEN];	
	char local_ipaddr_v6[MAX_IPADDR_V6_LEN];
	unsigned int local_rtp_port;
	unsigned int local_rtcp_port;
	unsigned int remote_rtcp_port;
	char avs_remote_conn[MAX_IPADDR_LEN];
	unsigned int avs_remote_port;
	unsigned int avs_multicast_port;
	char avs_multicast_conn[MAX_IPADDR_LEN];

	/*for rtsp*/
	unsigned long rtsp_streamid; //added by zyxia for rtsp.2021.9.10
	/*for ICE*/
	int enable_ice;
	enum ast_rtp_ice_role role; /*!< Our role in ICE negotiation */
	char local_cname[AST_UUID_STR_LEN]; /*!< Our local CNAME */
	char remote_cname[AST_UUID_STR_LEN]; /*!< remote CNAME */
	unsigned int ice_port;      /*!< Port that ICE was started with if it was previously started */

	unsigned int local_ssrc;
	unsigned int local_rtx_ssrc;
	unsigned int local_fec_ssrc;
	char local_ufrag[MAX_ICE_UFRAG_LEN];  /*!< The local ICE username */
	char local_passwd[MAX_ICE_PASSWROD_LEN]; /*!< The local ICE password */
	char local_msid[MAX_AVS_MSID_LEN];
	struct ao2_container *local_candidates;           /*!< The local ICE candidates */

	unsigned int remote_ssrc;
	unsigned int remote_rtx_ssrc;
	unsigned int remote_fec_ssrc;
	char remote_ufrag[MAX_ICE_UFRAG_LEN];  /*!< The remote ICE username */
	char remote_passwd[MAX_ICE_PASSWROD_LEN]; /*!< The remote ICE password */
	char remote_msid[MAX_AVS_MSID_LEN];
	struct ao2_container *remote_candidates;           /*!< The remote ICE candidates */
	struct ao2_container *proposed_remote_candidates;  /*!< The proposed remote ICE candidates */
	struct ao2_container *new_remote_candidates;  	   /*!< The new remote ICE candidates */

	/*for dtls*/
	enum ast_rtp_dtls_setup dtls_setup;
	char local_fingerprint[MAX_FINGERPRINT_LEN];
	char remote_fingerprint[MAX_FINGERPRINT_LEN]; 
	
	/*for srtp*/
	enum ast_srtp_suite suite;   /*!< SRTP crypto suite */
	char local_crypto[MAX_LOCALCRYPTO_LEN]; /*normal set crypto*/
	char remote_crypto[MAX_LOCALCRYPTO_LEN]; /*normal set crypto*/
	enum ast_rtp_dtls_hash local_hash; /*!< Local hash used for the fingerprint */
	enum ast_rtp_dtls_hash remote_hash; /*!< Remote hash used for the fingerprint */

	/*for zrtp*/
	char local_zrtphash[MAX_ZRTP_HASH_LEN]; /*normal set local zrtp hash*/
	char remote_zrtphash[MAX_ZRTP_HASH_LEN]; /*normal set remote zrtp hash*/

	struct ast_rtp_jitterbuffer jitter_buffer;
	struct ast_format_cap *caps;
	int transmode;
	int pt_2833;
	int pt_gsfec;
	int pt_red;
	int pt_rtx;
	int pt_8k_cn;
	int pt_16k_cn;
	int pt_32k_cn;
	int ptmie;
	int media_type;
	int rtcp_mux;
	int nack_flag;
	int pli_flag;
	int fir_flag;
	int rsize_flag;
	int inbandfec;
	int tranport_cc_id;
	int teminal_type;
	int gsfec_ver;
	int fecc_pt;
	unsigned int ilbc_frame_mode;
	unsigned int opus_dtx_flag;
	unsigned int crypto_update_flag;
	unsigned int port_update_flag;
	unsigned int symrtp;
	int ssrc_reset_flag;
	char fax_filename[256];
	char fax_localstationid[MAX_LOCAL_STATION_ID];
	char fax_headerinfo[MAX_HEADER_INFO];
	unsigned int fax_mode;
	unsigned int mpa;

	//unsigned char rawdata[8192 + AST_FRIENDLY_OFFSET];
	//unsigned int ssrc;		/*!< Synchronization source, RFC 3550, page 10. */
	unsigned int themssrc;		/*!< Their SSRC */
	unsigned int themssrc_valid;	/*!< True if their SSRC is available. */
	unsigned int lastts;
	unsigned int lastividtimestamp;
	unsigned int lastovidtimestamp;
	unsigned int lastitexttimestamp;
	unsigned int lastotexttimestamp;
	int lastrxseqno;                /*!< Last received sequence number, from the network */
	int expectedrxseqno;		/*!< Next expected sequence number, from the network */
	AST_VECTOR(, int) missing_seqno; /*!< A vector of sequence numbers we never received */
	int expectedseqno;		/*!< Next expected sequence number, from the core */
	unsigned short seedrxseqno;     /*!< What sequence number did they start with?*/
	unsigned int seedrxts;          /*!< What RTP timestamp did they start with? */
	unsigned int rxcount;           /*!< How many packets have we received? */
	unsigned int rxoctetcount;      /*!< How many octets have we received? should be rxcount *160*/
	unsigned int txcount;           /*!< How many packets have we sent? */
	unsigned int txoctetcount;      /*!< How many octets have we sent? (txcount*160)*/
	unsigned int cycles;            /*!< Shifted count of sequence number cycles */
	double rxjitter;                /*!< Interarrival jitter at the moment in seconds */
	double rxtransit;               /*!< Relative transit time for previous packet */
	struct ast_format *lasttxformat;
	struct ast_format *lastrxformat;

	/* DTMF Reception Variables */
	char resp;                        /*!< The current digit being processed */
	unsigned int last_seqno;          /*!< The last known sequence number for any DTMF packet */
	unsigned int last_end_timestamp;  /*!< The last known timestamp received from an END packet */
	unsigned int dtmf_duration;       /*!< Total duration in samples since the digit start event */
	unsigned int dtmf_timeout;        /*!< When this timestamp is reached we consider END frame lost and forcibly abort digit */
	unsigned int dtmfsamples;
	enum ast_rtp_dtmf_mode dtmfmode;  /*!< The current DTMF mode of the RTP stream */
	/* DTMF Transmission Variables */
	unsigned int lastdigitts;
	char sending_digit;	/*!< boolean - are we sending digits */
	char send_digit;	/*!< digit we are sending */
	int send_payload;
	int send_duration;
	unsigned int flags;
	struct timeval rxcore;
	struct timeval txcore;
	double drxcore;                 /*!< The double representation of the first received packet */
	struct timeval dtmfmute;
	struct ast_smoother *smoother;
	unsigned short seqno;		/*!< Sequence number, RFC 3550, page 13. */
	struct ast_sched_context *sched;
	struct ast_rtcp *rtcp;
	unsigned int asymmetric_codec;  /*!< Indicate if asymmetric send/receive codecs are allowed */

	struct ast_rtp_instance *bundled; /*!< The RTP instance we are bundled to */
	int stream_num; /*!< Stream num for this RTP instance */
	AST_VECTOR(, struct rtp_ssrc_mapping) ssrc_mapping; /*!< Mappings of SSRC to RTP instances */
	struct ast_sockaddr bind_address; /*!< Requested bind address for the sockets */

	//enum strict_rtp_state strict_rtp_state; /*!< Current state that strict RTP protection is in */
	struct ast_sockaddr strict_rtp_address;  /*!< Remote address information for strict RTP purposes */
	struct rtp_red *red;
	struct ast_data_buffer *send_buffer;		/*!< Buffer for storing sent packets for retransmission */
	struct ast_data_buffer *recv_buffer;		/*!< Buffer for storing received packets for retransmission */

#ifdef HAVE_PJPROJECT
	ast_cond_t cond;            /*!< ICE/TURN condition for signaling */
	struct ice_wrap *ice;       /*!< ao2 wrapped ICE session */
	pj_turn_sock *turn_rtp;     /*!< RTP TURN relay */
	pj_turn_sock *turn_rtcp;    /*!< RTCP TURN relay */
	pj_turn_state_t turn_state; /*!< Current state of the TURN relay session */
	unsigned int passthrough:1; /*!< Bit to indicate that the received packet should be passed through */
	unsigned int rtp_passthrough:1; /*!< Bit to indicate that TURN RTP should be passed through */
	unsigned int rtcp_passthrough:1; /*!< Bit to indicate that TURN RTCP should be passed through */
	struct ast_sockaddr rtp_loop; /*!< Loopback address for forwarding RTP from TURN */
	struct ast_sockaddr rtcp_loop; /*!< Loopback address for forwarding RTCP from TURN */
	struct ast_rtp_ioqueue_thread *ioqueue; /*!< The ioqueue thread handling us */
	//struct ao2_container *ice_local_candidates;           /*!< The local ICE candidates */
	//struct ao2_container *ice_active_remote_candidates;   /*!< The remote ICE candidates */
	//struct ao2_container *ice_proposed_remote_candidates; /*!< Incoming remote ICE candidates for new session */
	struct ast_sockaddr ice_original_rtp_addr;            /*!< rtp address that ICE started on first session */
	unsigned int ice_num_components; /*!< The number of ICE components */
#endif

	SSL_CTX *ssl_ctx; /*!< SSL context */
	enum ast_rtp_dtls_verify dtls_verify; /*!< What to verify */
	unsigned int rekey; /*!< Interval at which to renegotiate and rekey */
	int rekeyid; /*!< Scheduled item id for rekeying */
	int use_parent_fd;
};

static const char config[] = "avs.conf";
static int ice_support = 1;

/*!
 * \brief SSRC mapping comparator for AST_VECTOR_REMOVE_CMP_UNORDERED()
 *
 * \param elem Element to compare against
 * \param value Value to compare with the vector element.
 *
 * \return 0 if element does not match.
 * \return Non-zero if element matches.
 */
#define AVS_SSRC_MAPPING_ELEM_CMP(elem, value) ((elem).instance == (value))
//static char *generate_random_string(char *buf, size_t size);
static int avs_rtp_set_avs_fd(struct ast_rtp_instance *instance, int fd);
static int avs_rtp_set_avs_nego_flag(struct ast_rtp_instance *instance, int flag);
static int avs_rtp_get_avs_nego_flag(struct ast_rtp_instance *instance);
static int avs_rtp_multicast_address_set(struct ast_rtp_instance *instance,	const struct ast_sockaddr *address);
static int avs_rtp_set_avs_client_path(struct ast_rtp_instance *instance, char *path);
static char *avs_rtp_get_avs_client_path(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_caps(struct ast_rtp_instance *instance, struct ast_format_cap *caps);
struct ast_format_cap *avs_rtp_get_avs_caps(struct ast_rtp_instance *instance);
static const char *avs_rtp_get_remote_cname(struct ast_rtp_instance *instance);
static int avs_rtp_set_remote_cname(struct ast_rtp_instance *instance, char *remote_cname);
static int avs_rtp_set_avs_2833pt(struct ast_rtp_instance *instance, int pt);
static int avs_rtp_get_avs_2833pt(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_gsfec_pt(struct ast_rtp_instance *instance, int pt);
static int avs_rtp_get_avs_gsfec_pt(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_red_pt(struct ast_rtp_instance *instance, int pt);
static int avs_rtp_get_avs_red_pt(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_transmode(struct ast_rtp_instance *instance, int transmode);
static int avs_rtp_get_avs_transmode(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_ptime(struct ast_rtp_instance *instance, int ptime);
static int avs_rtp_get_avs_ptime(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_media_type(struct ast_rtp_instance *instance, int media_type);
static int avs_rtp_get_avs_media_type(struct ast_rtp_instance *instance);
static int avs_rtp_set_rtcp_mux(struct ast_rtp_instance *instance, int rtcp_mux);
static int avs_rtp_get_rtcp_mux(struct ast_rtp_instance *instance);
static int avs_rtp_get_symrtp(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_nack(struct ast_rtp_instance *instance, int nack_flag);
static int avs_rtp_get_avs_nack(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_pli(struct ast_rtp_instance *instance, int pli_flag);
static int avs_rtp_get_avs_pli(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_fir(struct ast_rtp_instance *instance, int fir_flag);
static int avs_rtp_get_avs_fir(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_rsize(struct ast_rtp_instance *instance, int rsize_flag);
static int avs_rtp_get_avs_rsize(struct ast_rtp_instance *instance);
static int avs_rtp_set_inbandfec(struct ast_rtp_instance *instance, int inbandfec);
static int avs_rtp_get_inbandfec(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_transport(struct ast_rtp_instance *instance, int tranport_cc_id);
static int avs_rtp_get_avs_transport(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_teminal_type(struct ast_rtp_instance *instance, int teminal_type);
static int avs_rtp_get_avs_teminal_type(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_gsfec_ver(struct ast_rtp_instance *instance, int gsfec_ver);
static int avs_rtp_get_avs_gsfec_ver(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_rx_ssrc(struct ast_rtp_instance *instance, unsigned int remote_ssrc);
static unsigned int avs_rtp_get_avs_rx_ssrc(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_rx_rtx_ssrc(struct ast_rtp_instance *instance, unsigned int remote_rtx_ssrc);
static unsigned int avs_rtp_get_avs_rx_rtx_ssrc(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_tx_ssrc(struct ast_rtp_instance *instance, unsigned int local_ssrc);
static unsigned int avs_rtp_get_avs_tx_ssrc(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_tx_rtx_ssrc(struct ast_rtp_instance *instance, unsigned int local_rtx_ssrc);
static unsigned int avs_rtp_get_avs_tx_rtx_ssrc(struct ast_rtp_instance *instance);
static int avs_rtp_change_local_ssrc(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_bridgeid(struct ast_rtp_instance *instance, int bridge_id);
static int avs_rtp_set_avs_chanid(struct ast_rtp_instance *instance, int chanid);
static int avs_rtp_get_avs_chanid(struct ast_rtp_instance *instance);
static unsigned long avs_rtp_get_avs_portid(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_portid(struct ast_rtp_instance *instance, unsigned long portid);
static int avs_rtp_set_avs_child_flag(struct ast_rtp_instance *instance, int child_flag);
static int avs_rtp_get_avs_child_flag(struct ast_rtp_instance *instance);
static unsigned long avs_rtp_get_avs_parent_portid(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_parent_portid(struct ast_rtp_instance *instance, unsigned long portid);
static unsigned long avs_rtp_get_avs_trackid(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_trackid(struct ast_rtp_instance *instance, unsigned long trackid);
static int avs_rtp_set_avs_ice_flag(struct ast_rtp_instance *instance, int ice_flag);
static int avs_rtp_get_avs_ice_flag(struct ast_rtp_instance *instance);
static int avs_rtp_get_avs_ice_role(struct ast_rtp_instance *instance);
static int avs_rtp_set_avs_net_version(struct ast_rtp_instance *instance, int net_version);
static int avs_rtp_set_local_rtp_port(struct ast_rtp_instance *instance, int rtp_port);
static int avs_rtp_set_local_rtcp_port(struct ast_rtp_instance *instance, int rtcp_port);
static int avs_rtp_set_local_ipaddr(struct ast_rtp_instance *instance, char *ipaddr);
static int avs_rtp_set_local_ipaddr_v6(struct ast_rtp_instance *instance, char *ipaddr_v6);
static const char *avs_rtp_get_remote_ice_ufrag(struct ast_rtp_instance *instance);
static const char *avs_rtp_get_remote_ice_pwd(struct ast_rtp_instance *instance);
static int avs_rtp_set_local_ice_ufrag(struct ast_rtp_instance *instance, char *ice_ufrag);
static int avs_rtp_set_local_ice_pwd(struct ast_rtp_instance *instance,  char *ice_pwd);
static int avs_rtp_set_local_fingerprint(struct ast_rtp_instance *instance, char *fingerprint);
static int avs_rtp_set_local_fingerprint_hash(struct ast_rtp_instance *instance, enum ast_rtp_dtls_hash hash);
static enum ast_rtp_dtls_hash avs_rtp_get_remote_fingerprint_hash(struct ast_rtp_instance *instance);
static const char *avs_rtp_get_remote_fingerprint(struct ast_rtp_instance *instance);
static int avs_rtp_set_local_candidates(struct ast_rtp_instance *instance, void *cand_list);
static int avs_rtp_set_remote_candidates(struct ast_rtp_instance *instance, const struct ast_rtp_engine_ice_candidate *candidate);
static int avs_rtp_ice_active_candidates(struct ast_rtp_instance *instance);
static struct ao2_container *avs_rtp_ice_get_remote_candidates(struct ast_rtp_instance *instance);
static int avs_rtp_set_local_msid(struct ast_rtp_instance *instance, const char *local_msid);
static const char *avs_rtp_get_local_msid(struct ast_rtp_instance *instance);
static int avs_rtp_set_remote_msid(struct ast_rtp_instance *instance, char *remote_msid);
static const char *avs_rtp_get_remote_msid(struct ast_rtp_instance *instance);
static int avs_rtp_set_local_crypto(struct ast_rtp_instance *instance, char *local_crypto);
static const char *avs_rtp_get_local_crypto(struct ast_rtp_instance *instance);
static int avs_rtp_set_remote_crypto(struct ast_rtp_instance *instance, char *remote_crypto);
static const char *avs_rtp_get_remote_crypto(struct ast_rtp_instance *instance);
static int avs_rtp_set_crypto_suite(struct ast_rtp_instance *instance, int ice_flag);
static int avs_rtp_get_crypto_suite(struct ast_rtp_instance *instance);
static int avs_rtp_set_jitter_buffer(struct ast_rtp_instance *instance, struct ast_rtp_jitterbuffer *jitter_buffer);
static struct ast_rtp_jitterbuffer *avs_rtp_get_jitter_buffer(struct ast_rtp_instance *instance);
static int avs_rtp_set_ilbc_frame_mode(struct ast_rtp_instance *instance, unsigned int mode);
static unsigned int avs_rtp_get_ilbc_frame_mode(struct ast_rtp_instance *instance);
static int avs_rtp_set_opus_dtx_flag(struct ast_rtp_instance *instance, unsigned int flag);
static unsigned int avs_rtp_get_opus_dtx_flag(struct ast_rtp_instance *instance);
static int avs_rtp_set_crypto_update_flag(struct ast_rtp_instance *instance, unsigned int flag);
static unsigned int avs_rtp_get_crypto_update_flag(struct ast_rtp_instance *instance);
static int avs_rtp_set_port_update_flag(struct ast_rtp_instance *instance, unsigned int flag);
static unsigned int avs_rtp_get_port_update_flag(struct ast_rtp_instance *instance);
static int avs_rtp_set_ssrc_reset_flag(struct ast_rtp_instance *instance, int flag);
static int avs_rtp_get_ssrc_reset_flag(struct ast_rtp_instance *instance);
static int avs_rtp_set_fax_filename(struct ast_rtp_instance *instance,const char *fax_filename);
static const char *avs_rtp_get_fax_filename(struct ast_rtp_instance *instance);
static int avs_rtp_set_fax_mode(struct ast_rtp_instance *instance, unsigned int mode);
static unsigned int avs_rtp_get_fax_mode(struct ast_rtp_instance *instance);
static int avs_rtp_set_fax_localstationid(struct ast_rtp_instance *instance,const char *fax_localstationi);
static const char *avs_rtp_get_fax_localstationid(struct ast_rtp_instance *instance);
static int avs_rtp_set_fax_headerinfo(struct ast_rtp_instance *instance,const char *fax_headerinfo);
static const char *avs_rtp_get_fax_headerinfo(struct ast_rtp_instance *instance);
static int avs_rtp_set_fecc_pt(struct ast_rtp_instance *instance, int fecc_pt);
static int avs_rtp_get_fecc_pt(struct ast_rtp_instance *instance);
static int avs_rtp_set_rtsp_streamid(struct ast_rtp_instance *instance, unsigned long streamid);//added by zyxia for rtsp.2021.9.10
static unsigned long avs_rtp_get_rtsp_streamid(struct ast_rtp_instance *instance);//added by zyxia for rtsp.2021.9.10
static int avs_rtp_set_mpa(struct ast_rtp_instance *instance, unsigned int mpa);
static unsigned int avs_rtp_get_mpa(struct ast_rtp_instance *instance);

#ifdef __AVS_ICE_WRAPPER__
static void avs_rtp_ice_set_authentication(struct ast_rtp_instance *instance, const char *ufrag, const char *password)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (!ast_strlen_zero(ufrag)) {
		ast_copy_string(rtp->remote_ufrag, ufrag, sizeof(rtp->remote_ufrag));
	}

	if (!ast_strlen_zero(password)) {
		ast_copy_string(rtp->remote_passwd, password, sizeof(rtp->remote_passwd));
	}
	return;
}

/*! \brief Destructor for locally created ICE candidates */
static void avs_rtp_ice_candidate_destroy(void *obj)
{
	struct ast_rtp_engine_ice_candidate *candidate = obj;

	if (candidate->foundation) {
		ast_free(candidate->foundation);
	}

	if (candidate->transport) {
		ast_free(candidate->transport);
	}
}

static int avs_candidate_cmp(void *obj, void *arg, int flags)
{
	struct ast_rtp_engine_ice_candidate *candidate1 = obj, *candidate2 = arg;

	if (strcmp(candidate1->foundation, candidate2->foundation) ||
			candidate1->id != candidate2->id ||
			ast_sockaddr_cmp(&candidate1->address, &candidate2->address) ||
			candidate1->type != candidate2->type) {
		return 0;
	}
	return CMP_MATCH | CMP_STOP;
}

static void avs_rtp_ice_add_remote_candidate(struct ast_rtp_instance *instance, const struct ast_rtp_engine_ice_candidate *candidate)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	struct ast_rtp_engine_ice_candidate *remote_candidate;

	/*if (strcasecmp(candidate->transport, "udp")) {
		ast_log(LOG_ERROR, "unsupported transport %s!", candidate->transport);
		return;
	}*/

	if (!rtp->proposed_remote_candidates && !(rtp->proposed_remote_candidates = ao2_container_alloc(1, NULL, avs_candidate_cmp))) {
		ast_log(LOG_ERROR, "alloc proposed remote candidataes error!");
		return;
	}

 	if (ao2_container_count(rtp->proposed_remote_candidates) == PJ_ICE_MAX_CAND) {
		ast_log(LOG_ERROR, "too many candidates, the upper limit is %u!", PJ_ICE_MAX_CAND);
		return;
	}

	if (!(remote_candidate = ao2_alloc(sizeof(*remote_candidate), avs_rtp_ice_candidate_destroy))) {
		ast_log(LOG_ERROR, "alloc remote candidate error!");
		return;
	}

	remote_candidate->foundation = ast_strdup(candidate->foundation);
	remote_candidate->id = candidate->id;
	remote_candidate->transport = ast_strdup(candidate->transport);
	remote_candidate->priority = candidate->priority;
	ast_sockaddr_copy(&remote_candidate->address, &candidate->address);
	ast_sockaddr_copy(&remote_candidate->relay_address, &candidate->relay_address);
	remote_candidate->type = candidate->type;
	ast_log(LOG_DEBUG, "avs_rtp_ice_add_remote_candidate candidataes error[%s], [%s]!", remote_candidate->foundation, remote_candidate->transport);

	ao2_link(rtp->proposed_remote_candidates, remote_candidate);
	ao2_ref(remote_candidate, -1);
	return;
}

static void avs_rtp_ice_add_new_remote_candidate(struct ast_rtp_instance *instance, const struct ast_rtp_engine_ice_candidate *candidate)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	struct ast_rtp_engine_ice_candidate *remote_candidate;

	/*if (strcasecmp(candidate->transport, "udp")) {
		ast_log(LOG_ERROR, "unsupported transport %s!", candidate->transport);
		return;
	}*/

	if (!rtp->remote_candidates && !(rtp->remote_candidates = ao2_container_alloc(1, NULL, avs_candidate_cmp))) {
		ast_log(LOG_ERROR, "alloc proposed remote candidataes error!");
		return;
	}

	if (!rtp->new_remote_candidates && !(rtp->new_remote_candidates = ao2_container_alloc(1, NULL, avs_candidate_cmp))) {
		ast_log(LOG_ERROR, "alloc proposed remote candidataes error!");
		return;
	}

	if (ao2_container_count(rtp->new_remote_candidates) == PJ_ICE_MAX_CAND) {
		ast_log(LOG_ERROR, "too many new remote candidates, the upper limit is %u!", PJ_ICE_MAX_CAND);
		return;
	}

	if (!(remote_candidate = ao2_alloc(sizeof(*remote_candidate), avs_rtp_ice_candidate_destroy))) {
		ast_log(LOG_ERROR, "alloc remote candidate error!");
		return;
	}

	remote_candidate->foundation = ast_strdup(candidate->foundation);
	remote_candidate->id = candidate->id;
	remote_candidate->transport = ast_strdup(candidate->transport);
	remote_candidate->priority = candidate->priority;
	ast_sockaddr_copy(&remote_candidate->address, &candidate->address);
	ast_sockaddr_copy(&remote_candidate->relay_address, &candidate->relay_address);
	remote_candidate->type = candidate->type;

	ast_log(LOG_DEBUG, "avs_rtp_ice_add_new_remote_candidate candidataes foundation[%s], transport[%s]!", remote_candidate->foundation, remote_candidate->transport);

	ao2_link(rtp->new_remote_candidates, remote_candidate);
	ao2_link(rtp->remote_candidates, remote_candidate);

	ao2_ref(remote_candidate, -1);

	return;
}

static void avs_rtp_ice_stop(struct ast_rtp_instance *instance)
{
	return;
}

static int avs_candidates_compare(struct ao2_container *left, struct ao2_container *right)
{
	struct ao2_iterator i;
	struct ast_rtp_engine_ice_candidate *right_candidate;

	if (ao2_container_count(left) != ao2_container_count(right)) {
		return -1;
	}

	i = ao2_iterator_init(right, 0);
	while ((right_candidate = ao2_iterator_next(&i))) {
		struct ast_rtp_engine_ice_candidate *left_candidate = ao2_find(left, right_candidate, OBJ_POINTER);

		if (!left_candidate) {
			ao2_ref(right_candidate, -1);
			ao2_iterator_destroy(&i);
			return -1;
		}

		ao2_ref(left_candidate, -1);
		ao2_ref(right_candidate, -1);
	}
	ao2_iterator_destroy(&i);

	return 0;
}

static void avs_rtp_ice_start(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (!rtp->proposed_remote_candidates) {
		ast_log(LOG_ERROR, "%s ICE instance is invalid or remote candidates are empty!", __FUNCTION__);
		return;
	}

	/* Check for equivalence in the lists */
	if (rtp->remote_candidates && !avs_candidates_compare(rtp->proposed_remote_candidates, rtp->remote_candidates)) {
		ao2_cleanup(rtp->proposed_remote_candidates);
		rtp->proposed_remote_candidates = NULL;
		ast_log(LOG_WARNING, "%s proposed remote candidates and active remote candidates are the same!", __FUNCTION__);
		return;
	}

	/* Out with the old, in with the new */
	ao2_cleanup(rtp->remote_candidates);
	rtp->remote_candidates = rtp->proposed_remote_candidates;
	rtp->proposed_remote_candidates = NULL;
	return;
}

static const char *avs_rtp_ice_get_ufrag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_ufrag;
}

static const char *avs_rtp_ice_get_password(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_passwd;
}

static struct ao2_container *avs_rtp_ice_get_local_candidates(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->local_candidates) {
		ao2_ref(rtp->local_candidates, +1);
	}

	return rtp->local_candidates;
}

static void avs_rtp_ice_lite(struct ast_rtp_instance *instance)
{
	return;
}

static void avs_rtp_ice_set_role(struct ast_rtp_instance *instance, enum ast_rtp_ice_role role)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	ast_debug(3, "Set role to %s (%p)\n",
		role == AST_RTP_ICE_ROLE_CONTROLLED ? "CONTROLLED" : "CONTROLLING", instance);

	rtp->role = role;
	return;
}

static void avs_rtp_ice_turn_request(struct ast_rtp_instance *instance, enum ast_rtp_ice_component_type component,
		enum ast_transport transport, const char *server, unsigned int port, const char *username, const char *password)
{
	return;
}

static void avs_rtp_ice_change_components(struct ast_rtp_instance *instance, int num_components)
{
	return;
}

/* ICE RTP Engine interface declaration */
static struct ast_rtp_engine_ice avs_rtp_ice = {
	.set_authentication = avs_rtp_ice_set_authentication,
	.add_remote_candidate = avs_rtp_ice_add_remote_candidate,
	.add_new_remote_candidate = avs_rtp_ice_add_new_remote_candidate,
	.start = avs_rtp_ice_start,
	.stop = avs_rtp_ice_stop,
	.get_ufrag = avs_rtp_ice_get_ufrag,
	.get_password = avs_rtp_ice_get_password,
	.get_local_candidates = avs_rtp_ice_get_local_candidates,
	.ice_lite = avs_rtp_ice_lite,
	.set_role = avs_rtp_ice_set_role,
	.turn_request = avs_rtp_ice_turn_request,
	.change_components = avs_rtp_ice_change_components,
};
#endif

#ifdef __AVS_DTLS_WRAPPER__
static int avs_rtp_dtls_set_configuration(struct ast_rtp_instance *instance, const struct ast_rtp_dtls_cfg *dtls_cfg)
{
	int res = 0;
	//struct avs_rtp *rtp = (struct avs_rtp *)ast_rtp_instance_get_data(instance);

	//rtp->local_hash = dtls_cfg->hash;
	return res;
}

static int avs_rtp_dtls_active(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = (struct avs_rtp *)ast_rtp_instance_get_data(instance);

	return !rtp->ssl_ctx ? 0 : 1;
}

static void avs_rtp_dtls_stop(struct ast_rtp_instance *instance)
{
	return;
}

static void avs_rtp_dtls_reset(struct ast_rtp_instance *instance)
{
	return;
}

static enum ast_rtp_dtls_connection avs_rtp_dtls_get_connection(struct ast_rtp_instance *instance)
{
	return AST_RTP_DTLS_CONNECTION_NEW;
}

static enum ast_rtp_dtls_setup avs_rtp_dtls_get_setup(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->dtls_setup;
}

static void avs_rtp_dtls_set_setup(struct ast_rtp_instance *instance, enum ast_rtp_dtls_setup setup)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	
	switch (setup) {
		case AST_RTP_DTLS_SETUP_ACTIVE:
			rtp->dtls_setup = AST_RTP_DTLS_SETUP_PASSIVE;
			break;
		case AST_RTP_DTLS_SETUP_PASSIVE:
			rtp->dtls_setup = AST_RTP_DTLS_SETUP_ACTIVE;
			break;
		case AST_RTP_DTLS_SETUP_ACTPASS:
			if (rtp->dtls_setup == AST_RTP_DTLS_SETUP_ACTPASS) {
				rtp->dtls_setup = AST_RTP_DTLS_SETUP_ACTIVE;
			} else {
				rtp->dtls_setup = AST_RTP_DTLS_SETUP_PASSIVE;
			}
			break;
		case AST_RTP_DTLS_SETUP_HOLDCONN:
			rtp->dtls_setup = AST_RTP_DTLS_SETUP_HOLDCONN;
			break;
		default:
			return;
	}

	return;
}

static void avs_rtp_dtls_set_fingerprint(struct ast_rtp_instance *instance, enum ast_rtp_dtls_hash hash, const char *fingerprint)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	rtp->remote_hash = hash;
	snprintf(rtp->remote_fingerprint, sizeof(rtp->remote_fingerprint), "%s %s", 
						(AST_RTP_DTLS_HASH_SHA256 == hash) ? "sha-256" : "sha-1", fingerprint);	
	return;
}

static enum ast_rtp_dtls_hash avs_rtp_dtls_get_fingerprint_hash(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_hash;
}

static const char *avs_rtp_dtls_get_fingerprint(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_fingerprint;
}

static struct ast_rtp_engine_dtls avs_rtp_dtls = {
	.set_configuration = avs_rtp_dtls_set_configuration,
	.active = avs_rtp_dtls_active,
	.stop = avs_rtp_dtls_stop,
	.reset = avs_rtp_dtls_reset,
	.get_connection = avs_rtp_dtls_get_connection,
	.get_setup = avs_rtp_dtls_get_setup,
	.set_setup = avs_rtp_dtls_set_setup,
	.set_fingerprint = avs_rtp_dtls_set_fingerprint,
	.get_fingerprint_hash = avs_rtp_dtls_get_fingerprint_hash,
	.get_fingerprint = avs_rtp_dtls_get_fingerprint,
};
#endif

#ifdef __AVS_ENGINE_WRAPPER__
static int avs_rtp_new(struct ast_rtp_instance *instance,
						struct ast_sched_context *sched, struct ast_sockaddr *addr,
						void *data)
{
	struct avs_rtp *rtp = NULL;
	struct avs_port_param *port_param;
	struct avs_port_param add_port_param;

	ast_log(LOG_DEBUG, "avs_rtp_new enter!");

	if (!(rtp = ast_calloc(1, sizeof(*rtp)))) {
		return -1;
	}

	rtp->local_ssrc = ast_random();
	rtp->local_rtx_ssrc = ast_random();
	rtp->local_fec_ssrc = ast_random();
	ast_uuid_generate_str(rtp->local_cname, sizeof(rtp->local_cname));
	rtp->seqno = ast_random() & 0x7fff;
	rtp->expectedrxseqno = -1;
	rtp->expectedseqno = -1;
	rtp->sched = sched;
	ast_rtp_instance_set_data(instance, rtp);

	memset(&add_port_param, 0, sizeof(add_port_param));

	if (data) {
		port_param = (struct avs_port_param *)data;
		add_port_param.endpt_ice_support = port_param->endpt_ice_support;
		add_port_param.port_group_id = port_param->port_group_id;
		add_port_param.media_turn_relay = port_param->media_turn_relay;
		add_port_param.enable_rtsp = port_param->enable_rtsp;
		add_port_param.cand_host_only = port_param->cand_host_only;
		add_port_param.trickle_ice = port_param->trickle_ice;
		add_port_param.parent_avs_fd = port_param->parent_avs_fd;
		snprintf(add_port_param.rtsp_stream_id,sizeof(add_port_param.rtsp_stream_id),"%s",port_param->rtsp_stream_id);
		add_port_param.net_type = port_param->net_type;
		add_port_param.if_type = port_param->if_type;
		add_port_param.enable_dtls = port_param->enable_dtls;
		add_port_param.enable_zrtp = port_param->enable_zrtp ;

		strcpy(add_port_param.avs_address, port_param->avs_address);
		ast_rtp_instance_set_avs_address(instance, port_param->avs_address);
	}

	if (ice_support && add_port_param.endpt_ice_support) {
		add_port_param.enable_dtls = 1;
		add_port_param.enable_ice = 1;
		rtp->enable_ice = 1;
	}

	if (add_port_param.parent_avs_fd > 0) {
		ast_log(LOG_DEBUG, "It's bundled, don't need add port, use parent_avs_fd:%d\n", add_port_param.parent_avs_fd);
		ast_rtp_instance_set_fd(instance, add_port_param.parent_avs_fd);
		ast_rtp_instance_set_use_parent_fd(instance, 1);
		return 0;
	}

	if (ast_avs_add_port(instance, add_port_param)) {
		return -1;
	}

	/**address from avs?????**/
	ast_sockaddr_copy(&rtp->bind_address, addr);
	ast_sockaddr_set_port(&rtp->bind_address, rtp->local_rtp_port);
	ast_rtp_instance_set_local_address(instance, &rtp->bind_address);
			
	if (add_port_param.enable_ice) {
		rtp->ice_num_components = 1;
		rtp->ice_port = rtp->local_rtp_port;
		ast_sockaddr_copy(&rtp->ice_original_rtp_addr, &rtp->bind_address);
		//generate_random_string(rtp->local_ufrag, sizeof(rtp->local_ufrag));
		//generate_random_string(rtp->local_passwd, sizeof(rtp->local_passwd));
	}

	return 0;
}

static int avs_rtp_destroy(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	
	ast_avs_rtp_destroy_notify(instance);

	if (rtp->caps) {
		ao2_ref(rtp->caps, -1);
		rtp->caps = NULL;
	}

	if (rtp->local_candidates) {
		ao2_ref(rtp->local_candidates, -1);
		rtp->local_candidates = NULL;
	}

	if (rtp->remote_candidates) {
		ao2_ref(rtp->remote_candidates, -1);
		rtp->remote_candidates = NULL;
	}

	if (rtp->proposed_remote_candidates) {
		ao2_ref(rtp->proposed_remote_candidates, -1);
		rtp->proposed_remote_candidates = NULL;
	}

	if (rtp->new_remote_candidates) {
		ao2_ref(rtp->new_remote_candidates, -1);
		rtp->new_remote_candidates = NULL;
	}

	ast_free(rtp);
	return 0;
}

static int avs_rtp_dtmf_begin(struct ast_rtp_instance *instance, char digit)
{
	struct ast_sockaddr remote_address = { {0,} };

	ast_rtp_instance_get_remote_address(instance, &remote_address);
	return 0;
}

static int avs_rtp_dtmf_end(struct ast_rtp_instance *instance, char digit)
{
	return 0;
}

static int avs_rtp_dtmf_end_with_duration(struct ast_rtp_instance *instance, char digit, unsigned int duration)
{
	struct ast_sockaddr remote_address = { {0,} };
	int res = -1;

	ast_rtp_instance_get_remote_address(instance, &remote_address);
	return res;
}

/*! \pre instance is locked */
static int avs_rtp_dtmf_mode_set(struct ast_rtp_instance *instance, enum ast_rtp_dtmf_mode dtmf_mode)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->dtmfmode = dtmf_mode;
	return 0;
}

static enum ast_rtp_dtmf_mode avs_rtp_dtmf_mode_get(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->dtmfmode;
}


/*! \pre instance is locked */
static void avs_rtp_update_source(struct ast_rtp_instance *instance)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return;
}

/*! \pre instance is locked */
static void avs_rtp_change_source(struct ast_rtp_instance *instance)
{


	return;
}

/*! \pre instance is locked */
static int avs_rtp_write(struct ast_rtp_instance *instance, struct ast_frame *frame)
{




	return 0;

}

static struct ast_frame *avs_rtp_read(struct ast_rtp_instance *instance, int rtcp)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	
	if (!rtp){
		return &ast_null_frame;
	}
	
	int res = 0;
	struct ast_sockaddr addr;
	unsigned char *read_area = rtp->rawdata + AST_FRIENDLY_OFFSET;
	size_t read_area_size = sizeof(rtp->rawdata) - AST_FRIENDLY_OFFSET;

	if ((res = ast_recvfrom(rtp->s, read_area, read_area_size, 0, &addr)) < 0) {
		if (errno != EAGAIN) {
			ast_log(LOG_WARNING, "AVS RTP Read error: %s.  Hanging up.\n",
				(errno) ? strerror(errno) : "Unspecified");
			return NULL;
		}
		return &ast_null_frame;
	}

	ast_log(LOG_DEBUG, "avs_rtp_read %s.", read_area);
	/*only for playsound test*/
	rtp->f.frametype = AST_FRAME_CONTROL;
	rtp->f.subclass.integer = AST_CONTROL_STREAM_STOP;
	return ast_frdup(&rtp->f);
}

/*! \pre instance is locked */
static void avs_rtp_prop_set(struct ast_rtp_instance *instance, enum ast_rtp_property property, int value)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (AST_RTP_PROPERTY_NAT == property) {
		rtp->symrtp = value ? 1 : 0;
		ast_log(LOG_DEBUG, "set symrtp %d.", value);
	}

	return;
}

/*! \pre instance is locked */
static int avs_rtp_fd(struct ast_rtp_instance *instance, int rtcp)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->s;
}

/*! \pre instance is locked */
static void avs_rtp_remote_address_set(struct ast_rtp_instance *instance, struct ast_sockaddr *addr)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (!rtp){
		return;
	}

	if (rtp && rtp->enable_ice) {

	} else {
		rtp->avs_remote_port = ast_sockaddr_port(addr);
		strncpy(rtp->avs_remote_conn, ast_sockaddr_stringify_addr_remote(addr), sizeof(rtp->avs_remote_conn) - 1);
		ast_log(LOG_DEBUG, "avs_rtp_remote_address_set port is %d, addr is [%s]!", rtp->avs_remote_port, rtp->avs_remote_conn);
	}
	return;
}

/*! \pre instance is locked */
static int avs_rtp_red_init(struct ast_rtp_instance *instance, int buffer_time, int *payloads, int generations)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	//int x;

	return 0;
}

/*! \pre instance is locked */
static int avs_rtp_red_buffer(struct ast_rtp_instance *instance, struct ast_frame *frame)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	//struct rtp_red *red = rtp->red;


	return 0;
}

/*! \pre Neither instance0 nor instance1 are locked */
static int avs_rtp_local_bridge(struct ast_rtp_instance *instance0, struct ast_rtp_instance *instance1)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance0);


	return 0;
}

/*! \pre instance is locked */
static int avs_rtp_get_stat(struct ast_rtp_instance *instance, struct ast_rtp_instance_stats *stats, enum ast_rtp_instance_stat stat)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return 0;
}

/*! \pre Neither instance0 nor instance1 are locked */
static int avs_rtp_dtmf_compatible(struct ast_channel *chan0, struct ast_rtp_instance *instance0, struct ast_channel *chan1, struct ast_rtp_instance *instance1)
{
	return 0;
}

/*! \pre instance is NOT locked */
static void avs_rtp_stun_request(struct ast_rtp_instance *instance, struct ast_sockaddr *suggestion, const char *username)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	//struct sockaddr_in suggestion_tmp;

	return;
}

/*! \pre instance is locked */
static void avs_rtp_stop(struct ast_rtp_instance *instance)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	//struct ast_sockaddr addr = { {0,} };

	return;
}

/*! \pre instance is locked */
static int avs_rtp_qos_set(struct ast_rtp_instance *instance, int tos, int cos, const char *desc)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return 0;
}

/*!
 * \brief generate comfort noice (CNG)
 *
 * \pre instance is locked
 */
static int avs_rtp_sendcng(struct ast_rtp_instance *instance, int level)
{
	int res = 0;
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	struct ast_sockaddr remote_address = { {0,} };

	ast_rtp_instance_get_remote_address(instance, &remote_address);

	return res;
}

/*! \pre instance is locked */
static int avs_rtp_activate(struct ast_rtp_instance *instance)
{
	//struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);


	return 0;
}

/*! \pre instance is locked */
static unsigned int avs_rtp_get_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->local_ssrc;
}

/*! \pre instance is locked */
static const char *avs_rtp_get_cname(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->local_cname;
}

static const char *avs_rtp_get_remote_cname(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->remote_cname;
}

static int avs_rtp_set_remote_cname(struct ast_rtp_instance *instance, char *remote_cname)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !remote_cname) {
		return -1;
	}

	strncpy(rtp->remote_cname, remote_cname, sizeof(rtp->remote_cname));
	return 0;
}

static void avs_rtp_set_remote_ssrc(struct ast_rtp_instance *instance, unsigned int ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	rtp->remote_ssrc = ssrc;
	return;
}

static void avs_rtp_set_stream_num(struct ast_rtp_instance *instance, int stream_num)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	rtp->stream_num = stream_num;
	return;
}

static int avs_rtp_extension_enable(struct ast_rtp_instance *instance, enum ast_rtp_extension extension)
{
	switch (extension) {
		case AST_RTP_EXTENSION_ABS_SEND_TIME:
			return 1;
		default:
			return 0;
	}
}

/*! \pre child is locked */
static int avs_rtp_bundle(struct ast_rtp_instance *child, struct ast_rtp_instance *parent)
{

	ast_rtp_instance_set_child_flag(child, 1);
	ast_rtp_instance_set_avs_parent_portid(child, ast_rtp_instance_get_avs_portid(parent));
	//delete child avs port
	ast_avs_bundle_delete_child_port(child);
	return 0;
}

#define __AVS_ENGINE__
static int avs_rtp_set_avs_fd(struct ast_rtp_instance *instance, int fd)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->s = fd;
	return 0;
}

static int avs_rtp_set_use_parent_fd(struct ast_rtp_instance *instance, int value)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->use_parent_fd = value;
	return 0;
}

static int avs_rtp_get_use_parent_fd(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->use_parent_fd;
}

static int avs_rtp_set_avs_nego_flag(struct ast_rtp_instance *instance, int flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->nego_flag = flag;
	return 0;
}

static int avs_rtp_get_avs_nego_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->nego_flag;
}

static int avs_rtp_multicast_address_set(struct ast_rtp_instance *instance,	const struct ast_sockaddr *address)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	rtp->avs_multicast_port = ast_sockaddr_port(address);
	strncpy(rtp->avs_multicast_conn, ast_sockaddr_stringify_addr_remote(address), sizeof(rtp->avs_remote_conn) - 1);
	ast_log(LOG_DEBUG, "avs_rtp_multicast_address_set port is %d, addr is [%s]!", rtp->avs_multicast_port, rtp->avs_multicast_conn);
	return 0;
}

static unsigned int avs_rtp_get_multicast_port(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->avs_multicast_port;
}

static char *avs_rtp_get_multicast_ip(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->avs_multicast_conn;
}

static int avs_rtp_set_avs_client_path(struct ast_rtp_instance *instance, char *path)
{
	int res = 0;
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	strncpy(rtp->avs_client_path, path, sizeof(rtp->avs_client_path));
	return res;
}

static char *avs_rtp_get_avs_client_path(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->avs_client_path;
}

static int avs_rtp_set_avs_caps(struct ast_rtp_instance *instance, struct ast_format_cap *caps)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->caps) {
		ao2_ref(rtp->caps, -1);
		rtp->caps = NULL;
	}

	if (!(rtp->caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate caps.");
		return -1;
	}

	ast_format_cap_append_from_cap(rtp->caps, caps, AST_MEDIA_TYPE_UNKNOWN);
	return 0;
}

struct ast_format_cap *avs_rtp_get_avs_caps(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->caps;
}

static int avs_rtp_set_avs_2833pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_2833 = pt;
	return 0;
}

static int avs_rtp_get_avs_2833pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_2833;
}

static int avs_rtp_set_avs_gsfec_pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_gsfec = pt;
	return 0;
}

static int avs_rtp_get_avs_gsfec_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_gsfec;
}

static int avs_rtp_set_avs_red_pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_red = pt;
	return 0;
}

static int avs_rtp_get_avs_red_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_red;
}

static int avs_rtp_set_avs_rtx_pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_rtx = pt;
	return 0;
}

static int avs_rtp_get_avs_rtx_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_rtx;
}

static int avs_rtp_set_avs_cn_8k_pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_8k_cn = pt;
	return 0;
}

static int avs_rtp_get_avs_cn_8k_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_8k_cn;
}

static int avs_rtp_set_avs_cn_16k_pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_16k_cn = pt;
	return 0;
}

static int avs_rtp_get_avs_cn_16k_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_16k_cn;
}

static int avs_rtp_set_avs_cn_32k_pt(struct ast_rtp_instance *instance, int pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pt_32k_cn = pt;
	return 0;
}

static int avs_rtp_get_avs_cn_32k_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pt_32k_cn;
}

static int avs_rtp_set_avs_transmode(struct ast_rtp_instance *instance, int transmode)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->transmode = transmode;
	return 0;
}

static int avs_rtp_get_avs_transmode(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->transmode;
}

static int avs_rtp_set_avs_ptime(struct ast_rtp_instance *instance, int ptime)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->ptmie = ptime;
	return 0;
}

static int avs_rtp_get_avs_ptime(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->ptmie;
}

static int avs_rtp_set_avs_media_type(struct ast_rtp_instance *instance, int media_type)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->media_type = media_type;
	return 0;
}

static int avs_rtp_get_avs_media_type(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->media_type;
}

static int avs_rtp_set_rtcp_mux(struct ast_rtp_instance *instance, int rtcp_mux)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->rtcp_mux = rtcp_mux;
	return 0;
}

static int avs_rtp_get_rtcp_mux(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->rtcp_mux;
}

static int avs_rtp_get_symrtp(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->symrtp;
}

static int avs_rtp_set_avs_nack(struct ast_rtp_instance *instance, int nack_flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->nack_flag = nack_flag;
	return 0;
}

static int avs_rtp_get_avs_nack(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->nack_flag;
}

static int avs_rtp_set_avs_pli(struct ast_rtp_instance *instance, int pli_flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->pli_flag = pli_flag;
	return 0;
}

static int avs_rtp_get_avs_pli(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->pli_flag;
}

static int avs_rtp_set_avs_fir(struct ast_rtp_instance *instance, int fir_flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->fir_flag = fir_flag;
	return 0;
}

static int avs_rtp_get_avs_fir(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->fir_flag;
}

static int avs_rtp_set_avs_rsize(struct ast_rtp_instance *instance, int rsize_flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->rsize_flag = rsize_flag;
	return 0;
}

static int avs_rtp_get_avs_rsize(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->rsize_flag;
}

static int avs_rtp_set_inbandfec(struct ast_rtp_instance *instance, int inbandfec)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->inbandfec = inbandfec;
	return 0;
}

static int avs_rtp_get_inbandfec(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->inbandfec;
}

static int avs_rtp_set_avs_transport(struct ast_rtp_instance *instance, int tranport_cc_id)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->tranport_cc_id = tranport_cc_id;
	return 0;
}

static int avs_rtp_get_avs_transport(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->tranport_cc_id;
}

static int avs_rtp_set_avs_teminal_type(struct ast_rtp_instance *instance, int teminal_type)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->teminal_type = teminal_type;
	return 0;
}
static int avs_rtp_get_avs_teminal_type(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->teminal_type;
}

static int avs_rtp_set_avs_gsfec_ver(struct ast_rtp_instance *instance, int gsfec_ver)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->gsfec_ver = gsfec_ver;
	return 0;
}

static int avs_rtp_get_avs_gsfec_ver(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->gsfec_ver;
}

static int avs_rtp_set_avs_rx_ssrc(struct ast_rtp_instance *instance, unsigned int remote_ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->remote_ssrc = remote_ssrc;
	return 0;
}

static unsigned int avs_rtp_get_avs_rx_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->remote_ssrc;
}

static int avs_rtp_set_avs_rx_rtx_ssrc(struct ast_rtp_instance *instance, unsigned int remote_rtx_ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->remote_rtx_ssrc = remote_rtx_ssrc;
	return 0;
}

static unsigned int avs_rtp_get_avs_rx_rtx_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->remote_rtx_ssrc;
}

static int avs_rtp_set_avs_rx_fec_ssrc(struct ast_rtp_instance *instance, unsigned int remote_fec_ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->remote_fec_ssrc = remote_fec_ssrc;
	return 0;
}

static unsigned int avs_rtp_get_avs_rx_fec_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->remote_fec_ssrc;
}

static int avs_rtp_set_avs_tx_ssrc(struct ast_rtp_instance *instance, unsigned int local_ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->local_ssrc = local_ssrc;
	return 0;
}

static unsigned int avs_rtp_get_avs_tx_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_ssrc;
}

static int avs_rtp_set_avs_tx_rtx_ssrc(struct ast_rtp_instance *instance, unsigned int local_rtx_ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->local_rtx_ssrc = local_rtx_ssrc;
	return 0;
}

static unsigned int avs_rtp_get_avs_tx_rtx_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_rtx_ssrc;
}

static int avs_rtp_set_avs_tx_fec_ssrc(struct ast_rtp_instance *instance, unsigned int local_fec_ssrc)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->local_fec_ssrc = local_fec_ssrc;
	return 0;
}

static unsigned int avs_rtp_get_avs_tx_fec_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->local_fec_ssrc;
}

static int avs_rtp_change_local_ssrc(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->local_ssrc = ast_random();
	rtp->local_rtx_ssrc = ast_random();
	return 0;
}

static unsigned long avs_rtp_get_avs_trackid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return 	rtp->avs_track_id;
}
static int avs_rtp_set_avs_chanid(struct ast_rtp_instance *instance, int chanid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->avs_chan_id = chanid;
	return 0;
}

static int avs_rtp_get_avs_chanid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return 	rtp->avs_chan_id;
}

static unsigned long avs_rtp_get_avs_portid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return 	rtp->avs_port_id;
}


static int avs_rtp_set_avs_portid(struct ast_rtp_instance *instance, unsigned long portid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->avs_port_id = portid;
	return 0;
}

static int avs_rtp_set_avs_child_flag(struct ast_rtp_instance *instance, int child_flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->child_flag = child_flag;
	return 0;
}

/*Start added by zyxia for rtsp,2021.9.10*/
static int avs_rtp_set_rtsp_streamid(struct ast_rtp_instance *instance, unsigned long streamid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->rtsp_streamid = streamid;
	return 0;
}

static unsigned long avs_rtp_get_rtsp_streamid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return 	rtp->rtsp_streamid;
}
/*End added by zyxia for rtsp,2021.9.10*/

static int avs_rtp_get_avs_child_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->child_flag;
}

static unsigned long avs_rtp_get_avs_parent_portid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return 	rtp->avs_parent_port_id;
}

static int avs_rtp_set_avs_parent_portid(struct ast_rtp_instance *instance, unsigned long portid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->avs_parent_port_id = portid;
	return 0;
}

static int avs_rtp_set_avs_bridgeid(struct ast_rtp_instance *instance, int bridge_id)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->avs_bridge_id = bridge_id;
	return 0;
}

static int avs_rtp_set_avs_trackid(struct ast_rtp_instance *instance, unsigned long trackid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->avs_track_id = trackid;
	return 0;
}

static int avs_rtp_set_avs_ice_flag(struct ast_rtp_instance *instance, int ice_flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->enable_ice = ice_flag;
	return 0;
}

static int avs_rtp_get_avs_ice_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->enable_ice;
}

static int avs_rtp_get_avs_ice_role(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->role;
}

static int avs_rtp_set_avs_net_version(struct ast_rtp_instance *instance, int net_version)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->net_version = net_version;
	return 0;
}

static int avs_rtp_set_local_rtp_port(struct ast_rtp_instance *instance, int rtp_port)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	rtp->local_rtp_port = rtp_port;
	return 0;
}

static int avs_rtp_get_local_rtp_port(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!rtp) {
		return 0;
	}

	return rtp->local_rtp_port;
}

static int avs_rtp_set_local_rtcp_port(struct ast_rtp_instance *instance, int rtcp_port)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->local_rtcp_port = rtcp_port;
	return 0;
}

static int avs_rtp_get_local_rtcp_port(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!rtp) {
		return 0;
	}

	return rtp->local_rtcp_port;
}
	
static int avs_rtp_set_local_ipaddr(struct ast_rtp_instance *instance, char *ipaddr)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !ipaddr) {
		return -1;
	}

	strncpy(rtp->local_ipaddr, ipaddr, sizeof(rtp->local_ipaddr));
	return 0;
}

static int avs_rtp_get_remote_rtcp_port(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!rtp) {
		return 0;
	}

	return rtp->remote_rtcp_port;
}

static int avs_rtp_set_remote_rtcp_port(struct ast_rtp_instance *instance, int rtcp_port)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!rtp) {
		return 0;
	}

	rtp->remote_rtcp_port = rtcp_port;
	return 0;
}

static const char *avs_rtp_get_local_ipaddr(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!rtp) {
		return NULL;
	}

	return rtp->local_ipaddr;
}

static int avs_rtp_set_local_ipaddr_v6(struct ast_rtp_instance *instance, char *ipaddr_v6)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !ipaddr_v6) {
		return -1;
	}
	memset(rtp->local_ipaddr_v6, 0, sizeof(rtp->local_ipaddr_v6));
	strncpy(rtp->local_ipaddr_v6, ipaddr_v6, sizeof(rtp->local_ipaddr_v6));
	return 0;
}

static const char *avs_rtp_get_local_ipaddr_v6(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!rtp) {
		return NULL;
	}

	return rtp->local_ipaddr_v6;
}

static const char *avs_rtp_get_remote_ice_ufrag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->remote_ufrag;
}

static const char *avs_rtp_get_remote_ice_pwd(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->remote_passwd;
}

static int avs_rtp_set_local_ice_ufrag(struct ast_rtp_instance *instance, char *ice_ufrag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !ice_ufrag) {
		return -1;
	}
	memset(rtp->local_ufrag, 0, sizeof(rtp->local_ufrag));
	strncpy(rtp->local_ufrag, ice_ufrag, sizeof(rtp->local_ufrag));
	return 0;
}

static int avs_rtp_set_local_ice_pwd(struct ast_rtp_instance *instance,  char *ice_pwd)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	
	if(!instance || !ice_pwd) {
		return -1;
	}
	memset(rtp->local_passwd, 0, sizeof(rtp->local_passwd));
	strncpy(rtp->local_passwd, ice_pwd, sizeof(rtp->local_passwd));
	return 0;
}

static int avs_rtp_set_local_fingerprint(struct ast_rtp_instance *instance, char *fingerprint)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	char delim[] = " ";
	char *token  = NULL;
	char tmp_fin[256] = {0};
	char *save_ptr = NULL;

	strncpy(tmp_fin, fingerprint, 256);
	if(!instance || !fingerprint) {
		return -1;
	}

	if (strcasestr(fingerprint, "SHA-1")) {
		rtp->local_hash = AST_RTP_DTLS_HASH_SHA1;
	} else {
		rtp->local_hash = AST_RTP_DTLS_HASH_SHA256;
	}

	if ((token = strtok_r(tmp_fin, delim, &save_ptr)) && save_ptr) {
		strncpy(rtp->local_fingerprint, save_ptr, sizeof(rtp->local_fingerprint));
	}
		
	return 0;
}

static int avs_rtp_set_local_fingerprint_hash(struct ast_rtp_instance *instance, enum ast_rtp_dtls_hash hash)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (hash != AST_RTP_DTLS_HASH_SHA1 && hash != AST_RTP_DTLS_HASH_SHA256) {
		return -1;
	}

	rtp->local_hash = hash;	
	return 0;
}

static enum ast_rtp_dtls_hash avs_rtp_get_remote_fingerprint_hash(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->remote_hash;
}

static const char *avs_rtp_get_remote_fingerprint(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->remote_fingerprint;
}

static int avs_rtp_set_local_candidates(struct ast_rtp_instance *instance, void *cand_list)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
    struct candidate * cand_tmp = NULL;
	struct ast_rtp_engine_ice_candidate *candidates, *existing;
		
	if(!instance || !cand_list) {
		return -1;
	}

    AST_LIST_HEAD_NOLOCK(, candidate) *candidate_list = cand_list;

	/*delete old candidate*/
 	if (rtp->local_candidates) {
		ao2_ref(rtp->local_candidates, -1);
		rtp->local_candidates = NULL;
	}
	
    AST_LIST_TRAVERSE_SAFE_BEGIN(candidate_list, cand_tmp, entry)
    {
    	char foundation[32] = {0};
		char transport[32] = {0};
		char address[MAX_IPADDR_V6_LEN] = {0};
		char cand_type[6] = {0};
		char relay_address[MAX_IPADDR_V6_LEN] = {0};
		unsigned int port = 0;
		unsigned int relay_port = 0;
		
		if (!(candidates = ao2_alloc(sizeof(*candidates), avs_rtp_ice_candidate_destroy))) {
			ao2_ref(candidates, -1);
			continue;
		}
		
		if (!rtp->local_candidates && !(rtp->local_candidates = ao2_container_alloc(1, NULL, avs_candidate_cmp))) {
			ast_log(LOG_ERROR, "%s alloc proposed remote candidataes error!", __FUNCTION__);
			ao2_ref(candidates, -1);
			continue;
		}

		if (sscanf(cand_tmp->candidate_s, "%31s %30u %31s %30u %46s %30u typ %5s raddr %46s rport %30u", 
							foundation, &candidates->id, transport, (unsigned *)&candidates->priority, 
							address, &port, cand_type, relay_address, &relay_port) < 9) {
			if (sscanf(cand_tmp->candidate_s, "%31s %30u %31s %30u %46s %30u typ %5s", foundation, &candidates->id, transport,
				(unsigned *)&candidates->priority, address, &port, cand_type) < 7) {
				/* Candidate did not parse properly */
				ao2_ref(candidates, -1);
				continue;
			}
		}

		if (strcasecmp(transport, "udp")) {
			ast_log(LOG_ERROR, "unsupported transport %s!", transport);
			ao2_ref(candidates, -1);
			continue;
		}

	 	if (ao2_container_count(rtp->local_candidates) == PJ_ICE_MAX_CAND) {
			ast_log(LOG_ERROR, "too many candidates, the upper limit is %u!", PJ_ICE_MAX_CAND);
			ao2_ref(candidates, -1);
			continue;
		}

		candidates->foundation = ast_strdup(foundation);
		candidates->transport = ast_strdup(transport);
		ast_sockaddr_parse(&candidates->address, address, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(&candidates->address, port);
		if (!strcasecmp(cand_type, "host")) {
			candidates->type = AST_RTP_ICE_CANDIDATE_TYPE_HOST;
		} else if (!strcasecmp(cand_type, "srflx")) {
			candidates->type = AST_RTP_ICE_CANDIDATE_TYPE_SRFLX;
		} else if (!strcasecmp(cand_type, "relay")) {
			candidates->type = AST_RTP_ICE_CANDIDATE_TYPE_RELAYED;
		} else {
			ao2_ref(candidates, -1);
			continue;
		}

		if (!ast_strlen_zero(relay_address)) {
			ast_sockaddr_parse(&candidates->relay_address, relay_address, PARSE_PORT_FORBID);
		}

		if (relay_port) {
			ast_sockaddr_set_port(&candidates->relay_address, relay_port);
		}

		if ((existing = ao2_find(rtp->local_candidates, candidates, OBJ_POINTER))) {
			ao2_ref(existing, -1);
			ao2_ref(candidates, -1);
			continue;
		}
		ao2_link(rtp->local_candidates, candidates);
		ao2_ref(candidates, -1);
    }
    AST_LIST_TRAVERSE_SAFE_END;

	return 0;
}

static int avs_rtp_set_remote_candidates(struct ast_rtp_instance *instance, const struct ast_rtp_engine_ice_candidate *candidate)
{
	avs_rtp_ice_add_remote_candidate(instance, candidate);
	return 0;
}

static struct ao2_container *avs_rtp_ice_get_remote_candidates(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->remote_candidates) {
		ao2_ref(rtp->remote_candidates, +1);
	}

	return rtp->remote_candidates;
}

static void avs_rtp_ice_clear_remote_candidates(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->remote_candidates) {
		ao2_ref(rtp->remote_candidates, -1);
		rtp->remote_candidates = NULL;
	}

	return;
}

static void avs_rtp_ice_clear_local_candidates(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->local_candidates) {
		ao2_ref(rtp->local_candidates, -1);
		rtp->local_candidates = NULL;
	}

	return;
}

static struct ao2_container *avs_rtp_ice_get_new_remote_candidates(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->new_remote_candidates) {
		ao2_ref(rtp->new_remote_candidates, +1);
	}

	return rtp->new_remote_candidates;
}

static void avs_rtp_ice_clear_new_remote_candidates(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (rtp->new_remote_candidates) {
		ao2_ref(rtp->new_remote_candidates, -1);
		rtp->new_remote_candidates = NULL;
	}

	return;
}

static int avs_rtp_ice_active_candidates(struct ast_rtp_instance *instance)
{
	avs_rtp_ice_start(instance);
	return 0;
}

static int avs_rtp_set_local_msid(struct ast_rtp_instance *instance, const char *local_msid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !local_msid) {
		return -1;
	}

	strncpy(rtp->local_msid, local_msid, sizeof(rtp->local_msid));
 	return 0;
}

static const char *avs_rtp_get_local_msid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->local_msid;
}

static int avs_rtp_set_remote_msid(struct ast_rtp_instance *instance, char *remote_msid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !remote_msid) {
		return -1;
	}

	strncpy(rtp->remote_msid, remote_msid, sizeof(rtp->remote_msid));
	return 0;
}

static const char *avs_rtp_get_remote_msid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
 	return rtp->remote_msid;
}

static int avs_rtp_set_local_crypto(struct ast_rtp_instance *instance, char *local_crypto)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !local_crypto) {
		return -1;
	}

	strncpy(rtp->local_crypto, local_crypto, sizeof(rtp->local_crypto));
 	return 0;
}

static const char *avs_rtp_get_local_crypto(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->local_crypto;
}

static int avs_rtp_set_remote_crypto(struct ast_rtp_instance *instance, char *remote_crypto)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !remote_crypto) {
		return -1;
	}

	strncpy(rtp->remote_crypto, remote_crypto, sizeof(rtp->remote_crypto));
	return 0;
}

static const char *avs_rtp_get_remote_crypto(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
 	return rtp->remote_crypto;
}

static int avs_rtp_set_crypto_suite(struct ast_rtp_instance *instance, int crypto_suite)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->suite = crypto_suite;
	return 0;
}

static int avs_rtp_get_crypto_suite(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->suite;
}

static int avs_rtp_set_jitter_buffer(struct ast_rtp_instance *instance, struct ast_rtp_jitterbuffer *jitter_buffer)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	if (!jitter_buffer) {
		return -1;
	}
	
	rtp->jitter_buffer.mode = jitter_buffer->mode;
	rtp->jitter_buffer.normal_mode = jitter_buffer->normal_mode;
	rtp->jitter_buffer.normal_size = jitter_buffer->normal_size;
	rtp->jitter_buffer.max_normal_size = jitter_buffer->max_normal_size;

	return 0;
}

static struct ast_rtp_jitterbuffer *avs_rtp_get_jitter_buffer(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return &rtp->jitter_buffer;
}

static int avs_rtp_set_ilbc_frame_mode(struct ast_rtp_instance *instance, unsigned int mode)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->ilbc_frame_mode = mode;
	return 0;
}

static unsigned int avs_rtp_get_ilbc_frame_mode(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->ilbc_frame_mode;
}

static int avs_rtp_set_opus_dtx_flag(struct ast_rtp_instance *instance, unsigned int flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->opus_dtx_flag = flag;
	return 0;
}

static unsigned int avs_rtp_get_opus_dtx_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->opus_dtx_flag;
}

static int avs_rtp_set_crypto_update_flag(struct ast_rtp_instance *instance, unsigned int flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->crypto_update_flag = flag;
	return 0;
}

static unsigned int avs_rtp_get_crypto_update_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->crypto_update_flag;
}

static int avs_rtp_set_port_update_flag(struct ast_rtp_instance *instance, unsigned int flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->port_update_flag = flag;
	return 0;
}

static unsigned int avs_rtp_get_port_update_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->port_update_flag;
}

static int avs_rtp_set_ssrc_reset_flag(struct ast_rtp_instance *instance, int flag)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->ssrc_reset_flag = flag;
	return 0;
}

static int avs_rtp_get_ssrc_reset_flag(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->ssrc_reset_flag;
}

static int avs_rtp_set_fax_filename(struct ast_rtp_instance *instance,const char *fax_filename)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !fax_filename) {
		return -1;
	}

	strncpy(rtp->fax_filename, fax_filename, sizeof(rtp->fax_filename));
 	return 0;
}

static const char *avs_rtp_get_fax_filename(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->fax_filename;
}

static int avs_rtp_set_fax_mode(struct ast_rtp_instance *instance, unsigned int mode)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	rtp->fax_mode = mode;
	return 0;
}

static unsigned int avs_rtp_get_fax_mode(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	return rtp->fax_mode;
}

static int avs_rtp_set_fax_localstationid(struct ast_rtp_instance *instance,const char *fax_localstationid)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !fax_localstationid) {
		return -1;
	}

	strncpy(rtp->fax_localstationid, fax_localstationid, sizeof(rtp->fax_localstationid));
 	return 0;
}

static const char *avs_rtp_get_fax_localstationid(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->fax_localstationid;
}

static int avs_rtp_set_fax_headerinfo(struct ast_rtp_instance *instance,const char *fax_headerinfo)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !fax_headerinfo) {
		return -1;
	}

	strncpy(rtp->fax_headerinfo, fax_headerinfo, sizeof(rtp->fax_headerinfo));
 	return 0;
}

static const char *avs_rtp_get_fax_headerinfo(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->fax_headerinfo;
}

static int avs_rtp_set_fecc_pt(struct ast_rtp_instance *instance, int fecc_pt)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance) {
		return -1;
	}

	rtp->fecc_pt = fecc_pt;
 	return 0;
}

static int avs_rtp_get_fecc_pt(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->fecc_pt;
}

static int avs_rtp_set_mpa(struct ast_rtp_instance *instance, unsigned int mpa)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if (!instance) {
		return -1;
	}

	rtp->mpa = mpa;
	return 0;
}

static unsigned int avs_rtp_get_mpa(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->mpa;
}
/*Start added by zyxia for support zrtp,2024.4.12*/
static int avs_rtp_set_local_zrtphash(struct ast_rtp_instance *instance, char *local_zrtphash)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !local_zrtphash) {
		return -1;
	}

	strncpy(rtp->local_zrtphash, local_zrtphash, sizeof(rtp->local_zrtphash));
 	return 0;
}

static const char *avs_rtp_get_local_zrtphash(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);

	return rtp->local_zrtphash;
}

static int avs_rtp_set_remote_zrtphash(struct ast_rtp_instance *instance, char *remote_zrtphash)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
	if(!instance || !remote_zrtphash) {
		return -1;
	}

	strncpy(rtp->remote_zrtphash, remote_zrtphash, sizeof(rtp->remote_zrtphash));
	return 0;
}

static const char *avs_rtp_get_remote_zrtphash(struct ast_rtp_instance *instance)
{
	struct avs_rtp *rtp = ast_rtp_instance_get_data(instance);
 	return rtp->remote_zrtphash;
}
/*End added by zyxia for support zrtp,2024.4.12*/


/* RTP Engine Declaration */
static struct ast_rtp_engine avs_rtp_engine = {
	.name = "avs",
	.new = avs_rtp_new,
	.destroy = avs_rtp_destroy,
	.dtmf_begin = avs_rtp_dtmf_begin,
	.dtmf_end = avs_rtp_dtmf_end,
	.dtmf_end_with_duration = avs_rtp_dtmf_end_with_duration,
	.dtmf_mode_set = avs_rtp_dtmf_mode_set,
	.dtmf_mode_get = avs_rtp_dtmf_mode_get,
	.update_source = avs_rtp_update_source,
	.change_source = avs_rtp_change_source,
	.write = avs_rtp_write,
	.read = avs_rtp_read,
	.prop_set = avs_rtp_prop_set,
	.fd = avs_rtp_fd,
	.remote_address_set = avs_rtp_remote_address_set,
	.red_init = avs_rtp_red_init,
	.red_buffer = avs_rtp_red_buffer,
	.local_bridge = avs_rtp_local_bridge,
	.get_stat = avs_rtp_get_stat,
	.dtmf_compatible = avs_rtp_dtmf_compatible,
	.stun_request = avs_rtp_stun_request,
	.stop = avs_rtp_stop,
	.qos = avs_rtp_qos_set,
	.sendcng = avs_rtp_sendcng,
	.ice = &avs_rtp_ice,
	.dtls = &avs_rtp_dtls,
	.activate = avs_rtp_activate,
	.set_stream_num = avs_rtp_set_stream_num,
	.extension_enable = avs_rtp_extension_enable,
	.bundle = avs_rtp_bundle,

	.ssrc_get = avs_rtp_get_ssrc,
	.cname_get = avs_rtp_get_cname,
	.remote_cname_get = avs_rtp_get_remote_cname,
	.remote_cname_set = avs_rtp_set_remote_cname,
	.set_remote_ssrc = avs_rtp_set_remote_ssrc,
	.set_avs_fd = avs_rtp_set_avs_fd,
	.set_avs_nego_flag = avs_rtp_set_avs_nego_flag,
	.get_avs_nego_flag = avs_rtp_get_avs_nego_flag,
	.multicast_address_set = avs_rtp_multicast_address_set,
	.get_multicast_port = avs_rtp_get_multicast_port,
	.get_multicast_ip = avs_rtp_get_multicast_ip,
	.set_avs_client_path = avs_rtp_set_avs_client_path,
	.get_avs_client_path = avs_rtp_get_avs_client_path,
	.set_avs_caps = avs_rtp_set_avs_caps,
	.get_avs_caps = avs_rtp_get_avs_caps,
	.set_avs_2833pt = avs_rtp_set_avs_2833pt,
	.get_avs_2833pt = avs_rtp_get_avs_2833pt,
	.set_avs_gsfec_pt = avs_rtp_set_avs_gsfec_pt,
	.get_avs_gsfec_pt = avs_rtp_get_avs_gsfec_pt,
	.set_avs_red_pt = avs_rtp_set_avs_red_pt,
	.get_avs_red_pt = avs_rtp_get_avs_red_pt,
	.set_avs_rtx_pt = avs_rtp_set_avs_rtx_pt,
	.get_avs_rtx_pt = avs_rtp_get_avs_rtx_pt,
	.set_avs_cn_8k_pt = avs_rtp_set_avs_cn_8k_pt,
	.get_avs_cn_8k_pt = avs_rtp_get_avs_cn_8k_pt,
	.set_avs_cn_16k_pt = avs_rtp_set_avs_cn_16k_pt,
	.get_avs_cn_16k_pt = avs_rtp_get_avs_cn_16k_pt,
	.set_avs_cn_32k_pt = avs_rtp_set_avs_cn_32k_pt,
	.get_avs_cn_32k_pt = avs_rtp_get_avs_cn_32k_pt,
	.set_avs_transmode = avs_rtp_set_avs_transmode,
	.get_avs_transmode = avs_rtp_get_avs_transmode,
	.set_avs_ptime = avs_rtp_set_avs_ptime,
	.get_avs_ptime = avs_rtp_get_avs_ptime,
	.set_avs_media_type = avs_rtp_set_avs_media_type,
	.get_avs_media_type = avs_rtp_get_avs_media_type,
	.set_rtcp_mux = avs_rtp_set_rtcp_mux,
	.get_rtcp_mux = avs_rtp_get_rtcp_mux,
	.get_symrtp = avs_rtp_get_symrtp,
	.set_avs_nack = avs_rtp_set_avs_nack,
	.get_avs_nack = avs_rtp_get_avs_nack,
	.set_avs_pli = avs_rtp_set_avs_pli,
	.get_avs_pli = avs_rtp_get_avs_pli,
	.set_avs_fir = avs_rtp_set_avs_fir,
	.get_avs_fir = avs_rtp_get_avs_fir,
	.set_avs_rsize = avs_rtp_set_avs_rsize,
	.get_avs_rsize = avs_rtp_get_avs_rsize,
	.set_avs_inbandfec = avs_rtp_set_inbandfec,
	.get_avs_inbandfec = avs_rtp_get_inbandfec,
	.set_avs_transport = avs_rtp_set_avs_transport,
	.get_avs_transport = avs_rtp_get_avs_transport,
	.set_avs_teminal_type = avs_rtp_set_avs_teminal_type,
	.get_avs_teminal_type = avs_rtp_get_avs_teminal_type,
	.set_avs_gsfec_ver = avs_rtp_set_avs_gsfec_ver,
	.get_avs_gsfec_ver = avs_rtp_get_avs_gsfec_ver,
	.set_avs_rx_ssrc = avs_rtp_set_avs_rx_ssrc,
	.get_avs_rx_ssrc = avs_rtp_get_avs_rx_ssrc,
	.set_avs_rx_rtx_ssrc = avs_rtp_set_avs_rx_rtx_ssrc,
	.get_avs_rx_rtx_ssrc = avs_rtp_get_avs_rx_rtx_ssrc,
	.set_avs_rx_fec_ssrc = avs_rtp_set_avs_rx_fec_ssrc,
	.get_avs_rx_fec_ssrc = avs_rtp_get_avs_rx_fec_ssrc,
	.set_avs_tx_ssrc = avs_rtp_set_avs_tx_ssrc,
	.get_avs_tx_ssrc = avs_rtp_get_avs_tx_ssrc,
	.set_avs_tx_rtx_ssrc = avs_rtp_set_avs_tx_rtx_ssrc,
	.get_avs_tx_rtx_ssrc = avs_rtp_get_avs_tx_rtx_ssrc,
	.set_avs_tx_fec_ssrc = avs_rtp_set_avs_tx_fec_ssrc,
	.get_avs_tx_fec_ssrc = avs_rtp_get_avs_tx_fec_ssrc,
	.change_local_ssrc = avs_rtp_change_local_ssrc,
	.set_avs_bridgeid = avs_rtp_set_avs_bridgeid,
	.set_avs_chanid = avs_rtp_set_avs_chanid,
	.get_avs_chanid = avs_rtp_get_avs_chanid,
	.set_avs_portid = avs_rtp_set_avs_portid,
	.get_avs_portid = avs_rtp_get_avs_portid,
	.set_avs_child_flag = avs_rtp_set_avs_child_flag,
	.get_avs_child_flag = avs_rtp_get_avs_child_flag,
	.set_avs_parent_portid = avs_rtp_set_avs_parent_portid,
	.get_avs_parent_portid = avs_rtp_get_avs_parent_portid,
	.set_avs_trackid = avs_rtp_set_avs_trackid,
	.get_avs_trackid = avs_rtp_get_avs_trackid,	
	.set_avs_ice_flag = avs_rtp_set_avs_ice_flag,
	.get_avs_ice_flag = avs_rtp_get_avs_ice_flag,
	.get_avs_ice_role = avs_rtp_get_avs_ice_role,
	.set_avs_net_version = avs_rtp_set_avs_net_version,
	.set_local_rtp_port = avs_rtp_set_local_rtp_port,
	.get_local_rtp_port = avs_rtp_get_local_rtp_port,
	.set_local_rtcp_port = avs_rtp_set_local_rtcp_port,
	.get_local_rtcp_port = avs_rtp_get_local_rtcp_port,
	.set_remote_rtcp_port = avs_rtp_set_remote_rtcp_port,
	.get_remote_rtcp_port = avs_rtp_get_remote_rtcp_port,
	.set_local_ipaddr = avs_rtp_set_local_ipaddr,
	.get_local_ipaddr = avs_rtp_get_local_ipaddr,
	.set_local_ipaddr_v6 = avs_rtp_set_local_ipaddr_v6,
	.get_local_ipaddr_v6 = avs_rtp_get_local_ipaddr_v6,
	.get_remote_ice_ufrag = avs_rtp_get_remote_ice_ufrag,
	.get_remote_ice_pwd = avs_rtp_get_remote_ice_pwd,
	.set_local_ice_ufrag = avs_rtp_set_local_ice_ufrag,
	.set_local_ice_pwd = avs_rtp_set_local_ice_pwd,
	.set_local_fingerprint = avs_rtp_set_local_fingerprint,
	.set_local_fingerprint_hash = avs_rtp_set_local_fingerprint_hash,
	.get_remote_fingerprint = avs_rtp_get_remote_fingerprint,
	.get_remote_fingerprint_hash = avs_rtp_get_remote_fingerprint_hash,
	.set_local_candidates = avs_rtp_set_local_candidates,
	.set_remote_candidates = avs_rtp_set_remote_candidates,
	.get_remote_candidates = avs_rtp_ice_get_remote_candidates,
	.clear_remote_candidates = avs_rtp_ice_clear_remote_candidates,
	.clear_local_candidates = avs_rtp_ice_clear_local_candidates,
	.get_new_remote_candidates = avs_rtp_ice_get_new_remote_candidates,
	.clear_new_remote_candidates = avs_rtp_ice_clear_new_remote_candidates,
	.active_candidates = avs_rtp_ice_active_candidates,
	.set_local_msid = avs_rtp_set_local_msid,
	.get_local_msid = avs_rtp_get_local_msid,
	.set_remote_msid = avs_rtp_set_remote_msid,
	.get_remote_msid = avs_rtp_get_remote_msid,
	.set_local_crypto = avs_rtp_set_local_crypto,
	.get_local_crypto = avs_rtp_get_local_crypto,
	.set_remote_crypto = avs_rtp_set_remote_crypto,
	.get_remote_crypto = avs_rtp_get_remote_crypto,
	.set_crypto_suite = avs_rtp_set_crypto_suite,
	.get_crypto_suite = avs_rtp_get_crypto_suite,
	.set_jitterbuffer = avs_rtp_set_jitter_buffer,
	.get_jitterbuffer = avs_rtp_get_jitter_buffer,
	.set_ilbc_frame_mode = avs_rtp_set_ilbc_frame_mode,
	.get_ilbc_frame_mode = avs_rtp_get_ilbc_frame_mode,
	.set_opus_dtx_flag = avs_rtp_set_opus_dtx_flag,
	.get_opus_dtx_flag = avs_rtp_get_opus_dtx_flag,
	.set_crypto_update_flag = avs_rtp_set_crypto_update_flag,
	.get_crypto_update_flag = avs_rtp_get_crypto_update_flag,	
	.set_port_update_flag = avs_rtp_set_port_update_flag,
	.get_port_update_flag = avs_rtp_get_port_update_flag,	
	.set_ssrc_reset_flag = avs_rtp_set_ssrc_reset_flag,
	.get_ssrc_reset_flag = avs_rtp_get_ssrc_reset_flag,
	.set_fax_filename = avs_rtp_set_fax_filename,
	.get_fax_filename = avs_rtp_get_fax_filename,
	.set_fax_mode = avs_rtp_set_fax_mode,
	.get_fax_mode = avs_rtp_get_fax_mode,
	.set_fax_localstationid = avs_rtp_set_fax_localstationid,
	.get_fax_localstationid= avs_rtp_get_fax_localstationid,
	.set_fax_headerinfo = avs_rtp_set_fax_headerinfo,
	.get_fax_headerinfo = avs_rtp_get_fax_headerinfo,
	.set_fecc_pt = avs_rtp_set_fecc_pt,
	.get_fecc_pt = avs_rtp_get_fecc_pt,
	.set_rtsp_streamid = avs_rtp_set_rtsp_streamid,
	.get_rtsp_streamid = avs_rtp_get_rtsp_streamid,
	.set_use_parent_fd = avs_rtp_set_use_parent_fd,
	.get_use_parent_fd = avs_rtp_get_use_parent_fd,
	.set_mpa = avs_rtp_set_mpa,
	.get_mpa = avs_rtp_get_mpa,
	.set_local_zrtphash = avs_rtp_set_local_zrtphash,
	.get_local_zrtphash = avs_rtp_get_local_zrtphash,
	.set_remote_zrtphash = avs_rtp_set_remote_zrtphash,
	.get_remote_zrtphash = avs_rtp_get_remote_zrtphash,
};
#endif
#if 0
static char *generate_random_string(char *buf, size_t size)
{
	unsigned long val[4];
	int x;

	for (x = 0; x < 4; x++) {
		val[x] = ast_random();
	}
	snprintf(buf, size, "%08lx%08lx%08lx%08lx", val[0], val[1], val[2], val[3]);

	return buf;
}
#endif
static int avs_reload(int reload)
{
	struct ast_config *cfg = NULL;
	const char *s;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	cfg = ast_config_load(config, config_flags);
	if (!cfg) {
		ast_log(LOG_NOTICE, "Unable to load config %s\n", config);
		return -1;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file %s is in an invalid format.  Aborting.\n", config);
		return -1;
	}

	if ((s = ast_variable_retrieve(cfg, "general", "icesupport"))) {
		ice_support = ast_true(s);
	}
	
	ast_avs_control(NULL, AVS_CONTROL_CONFIG, NULL, 0);
	ast_config_destroy(cfg);
	return 0;
}

static int reload_module(void)
{
	avs_reload(1);


    return 0;
}

static int load_module(void)
{
	if (ast_rtp_engine_register(&avs_rtp_engine)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	
	avs_reload(0);
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_rtp_engine_unregister(&avs_rtp_engine);
	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "AVS RTP Engine",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND,
#ifdef HAVE_PJPROJECT
    .requires = "res_pjproject",
#endif
);


