/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Chuandong Ruan <chdruan@grandstream.cn>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk.h"
#include <pjsip.h>
#include <pjsip_simple.h>
#include <pjlib.h>

#include "asterisk/module.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/app.h"
#include <regex.h>

AST_MODULE_LOG("pjsip");
static char last_options_prd_ip[32] = {0};
static int last_options_cseq = 0;

#define HERE_DOC_MARKE        "TTsf7G0"

/*CONTENT TYPE BODY XML*/
static const pj_str_t type         = { "application", 11 };
static const pj_str_t subtype      = { "x-gs-vmforward+xml", 18 };
static const pj_str_t vmForward     = { "vmForward", 9};
static const pj_str_t addr = { "addr", 4};
static const pj_str_t port = { "port", 4};
static const pj_str_t folder = { "folder", 6};
static const pj_str_t mailbox = { "mailbox", 7};
static const pj_str_t localmailbox = { "localmailbox", 12};
static const pj_str_t providerMode = { "providerMode", 12};
static const pj_str_t devBase = { "devBase", 7};

static pj_bool_t optnot_on_rx_request(pjsip_rx_data *rdata);

static struct pjsip_module notify_module = {
	.name = { "VMFORWARD Module", 13 },
	.priority = PJSIP_MOD_PRIORITY_APPLICATION - 1,
	.on_rx_request = optnot_on_rx_request,
};

static pj_status_t sip_get_vmforward_info_from_msg( pj_pool_t *pool, char *msg, pj_size_t len,
	pj_str_t **pAddr, pj_str_t **pLocal_ext, pj_str_t **pPort, pj_str_t **pMailbox, pj_str_t **pFolder, pj_str_t** pMode, pj_str_t** pDev)
{
	pj_xml_node *doc, *node;

	doc = pj_xml_parse( pool, msg, len);
	if (!doc) {
		return -1;
	}

	if (pj_stricmp(&doc->name, &vmForward) != 0) {
		return -1;
	}

	node = pj_xml_find_node(doc, &addr);
	if (node == NULL) {
		return -1;
	}
	*pAddr = &node->content;

	node = pj_xml_find_node(doc, &port);
	if (node == NULL) {
		return -1;
	}
	*pPort = &node->content;

	node = pj_xml_find_node(doc, &mailbox);
	if (node == NULL) {
		return -1;
	}
	*pMailbox = &node->content;

	node = pj_xml_find_node(doc, &folder);
	if (node == NULL) {
		return -1;
	}
	*pFolder = &node->content;

	node = pj_xml_find_node(doc, &providerMode);
	if (node != NULL) {
		*pMode = &node->content;
	}

	node = pj_xml_find_node(doc, &localmailbox);
	if (node != NULL) {
		*pLocal_ext = &node->content;
	}

	node = pj_xml_find_node(doc, &devBase);
	if (node != NULL) {
		*pDev = &node->content;
	}

	return PJ_SUCCESS;
}

/**
 * Function to check if the str is match the regexp pattern
 *
 * @param cmd         the string to check
 * @param pattern     regular expression
 * @return            0 is mathc, 1 is not match
 */
static int isMatchRegex(const char *cmd, const char *pattern)
{
	int  ret = 0;

	if (NULL == cmd || NULL == pattern) {
		return 0;
	}

	regex_t reg;

	ret = regcomp(&reg, pattern, REG_EXTENDED | REG_NOSUB);
	if (ret != 0) {
		return -1;
	}

	ret = regexec(&reg, cmd, 0, NULL, 0);

	regfree(&reg);

	return ret;
}

static pj_bool_t notify_on_rx_request(pjsip_rx_data *rdata)
{
	ast_log(LOG_VERBOSE, "Enter notify_on_rx_request !!!\n");

	pj_int32_t cseq = rdata->msg_info.cseq->cseq;
	pjsip_ctype_hdr *ctype_header;
	pjsip_msg_body *body = rdata->msg_info.msg->body;
	pj_str_t *pj_addr = NULL;
	pj_str_t *pj_port = NULL;
	pj_str_t *pj_mailbox = NULL;
	pj_str_t *pj_localmailbox = NULL;
	pj_str_t *pj_folder = NULL;
	pj_str_t *pj_mode = NULL;
	pj_str_t *pj_dev_base = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	endpoint = ast_pjsip_rdata_get_endpoint(rdata);

	ast_assert(endpoint != NULL);

	ctype_header = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTENT_TYPE, NULL);
	if (!ctype_header) {
		ast_log(LOG_DEBUG, "Incoming OPTIONS request with no CONTENT TYPE header\n");
		return PJ_FALSE;
	}

	if (pj_stricmp(&ctype_header->media.type, &type)) {
		ast_log(LOG_DEBUG, "Type is not application\n");
		return PJ_FALSE;
	}
	if (pj_stricmp(&ctype_header->media.subtype, &subtype)) {
		ast_log(LOG_DEBUG, "Content_type is not gs-request-vmforward\n");
		return PJ_FALSE;
	}

	if (!body || !body->data) {
		ast_log(LOG_DEBUG, "There is no message body in received packet\n");
		return PJ_FALSE;
	}

	if (PJ_SUCCESS == sip_get_vmforward_info_from_msg(rdata->tp_info.pool, (char*)body->data, body->len,
			&pj_addr,&pj_localmailbox, &pj_port, &pj_mailbox, &pj_folder, &pj_mode, &pj_dev_base)) {
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 200, NULL, NULL, NULL);

		char prd_ip[80] = {0};
		char prd_mailbox[128] = {0};
		char prd_folder[128] = {0};
		char prd_localmailbox[128] = {0};
		char prd_mode[16] = {0};
		char prd_dev_base[16] = {0};
		int  prd_port   = 0;
		char command[512] = {0};
		const char *prd_regex = "[&`;|$>\n]+";

		snprintf(prd_ip, sizeof(prd_ip), "%.*s", (int)pj_addr->slen, pj_addr->ptr);
		if (pj_port->ptr) {
			prd_port = atoi(pj_port->ptr);
		} 

		snprintf(prd_mailbox, sizeof(prd_mailbox), "%.*s", (int)pj_mailbox->slen, pj_mailbox->ptr);
		snprintf(prd_localmailbox, sizeof(prd_localmailbox), "%.*s", (int)pj_localmailbox->slen, pj_localmailbox->ptr);
		snprintf(prd_folder, sizeof(prd_folder), "%.*s", (int)pj_folder->slen, pj_folder->ptr);
		snprintf(prd_dev_base, sizeof(prd_dev_base), "%.*s", (int)pj_dev_base->slen, pj_dev_base->ptr);
		if (NULL != pj_mode) {
				snprintf(prd_mode, sizeof(prd_mode), "%.*s", (int)pj_mode->slen, pj_mode->ptr);
		} else {
				snprintf(prd_mode, sizeof(prd_mode), "normal");
		}

		ast_log(LOG_DEBUG, "prd_ip = %s, prd_port = %d,  prd_mailbox = %s,  prd_folder = %s pro_mode = [%s] pro_dev_base = [%s]", prd_ip, prd_port, prd_mailbox, prd_folder, prd_mode, prd_dev_base);

		if (0 == isMatchRegex(prd_ip, prd_regex) || 0 == isMatchRegex(prd_mailbox, prd_regex) || 0 == isMatchRegex(prd_folder, prd_regex)) {
			return PJ_FALSE;
		}

		if ((last_options_cseq == cseq) && !strcasecmp(last_options_prd_ip, prd_ip)) {
			return PJ_TRUE;
		} else {
			last_options_cseq = cseq;
			snprintf(last_options_prd_ip, sizeof(last_options_prd_ip), "%s", prd_ip);
		}

		snprintf(command, sizeof(command), "/cfg/var/lib/asterisk/scripts/forward_voicemail.sh `cat << '%s'\nrecv %d %s %s %s %s %s\n%s\n`&",
					HERE_DOC_MARKE, prd_port, prd_mailbox, prd_ip, prd_folder, prd_mode, prd_dev_base, HERE_DOC_MARKE);

		manager_event(EVENT_FLAG_CALL, "PushVoicemailMessage", "Send: %s\r\nRecv: %s\r\nUrgent: %s\r\nIsLocal: %s\r\n", prd_localmailbox, prd_mailbox, strcasecmp(prd_folder, "Urgent") == 0 ? "1" : "0" , "1");

		ast_log(LOG_DEBUG, "command is : %s\n", command);
		ast_safe_system2(command);

		return PJ_TRUE;
	} else {
		return PJ_FALSE;
	}
}

static pj_bool_t optnot_on_rx_request(pjsip_rx_data *rdata)
{
	if (!pjsip_method_cmp(&rdata->msg_info.msg->line.req.method, &pjsip_notify_method)) {
		return notify_on_rx_request(rdata);
	}

	return PJ_FALSE;
}

static int load_module(void)
{
	CHECK_PJSIP_MODULE_LOADED();

	if (ast_sip_register_service(&notify_module)) {
		ast_log(LOG_ERROR, "Could not register notify service\n");
		return AST_MODULE_LOAD_FAILURE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_service(&notify_module);
	return 0;
}

static int reload(void)
{
	ast_log(LOG_DEBUG, "Forward Voicemail module reload\n");
	return AST_MODULE_LOAD_SUCCESS;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "Forward Voicemail to Remote",
		.support_level = AST_MODULE_SUPPORT_CORE,
		.load = load_module,
		.unload = unload_module,
		.reload = reload,
		.load_pri = AST_MODPRI_APP_DEPEND,
);
#endif
