/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2007-2008, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 * David Vossel <dvossel@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Conference Bridge application
 *
 * \author\verbatim Joshua Colp <jcolp@digium.com> \endverbatim
 * \author\verbatim David Vossel <dvossel@digium.com> \endverbatim
 *
 * This is a conference bridge application utilizing the bridging core.
 * \ingroup applications
 */

/*! \li \ref app_confbridge.c uses the configuration file \ref confbridge.conf
 * \addtogroup configuration_file Configuration Files
 */

/*!
 * \page confbridge.conf confbridge.conf
 * \verbinclude confbridge.conf.sample
 */

/*** MODULEINFO
  	<depend>ucmsecret</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#ifdef GRANDSTREAM_NETWORKS
#include <dirent.h>                 /* for closedir, opendir, readdir, DIR */
#include <sys/stat.h>
#endif

#include "asterisk/cli.h"
#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/lock.h"
#include "asterisk/bridge.h"
#include "asterisk/musiconhold.h"
#include "asterisk/say.h"
#include "asterisk/audiohook.h"
#include "asterisk/astobj2.h"
#include "confbridge/include/confbridge.h"
#include "asterisk/paths.h"
#include "asterisk/manager.h"
#include "asterisk/test.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_bridges.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/json.h"
#include "asterisk/format_cache.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/stream.h"
#include "asterisk/message.h"
#include "asterisk/utils.h"
#include "ucm_crypt.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/stasis_message_router.h"
#include "asterisk/bridge_channel_internal.h"//added by zhywang
#include "asterisk/avs_controller.h"
#include "confbridge/include/conf_resource.h"
#include "asterisk/causes.h"
#include "asterisk/haserver.h"
#include "asterisk/astdb.h"
#include "asterisk/bridge_internal.h"
#include "asterisk/multi_media.h"

#ifdef USE_DB_MYSQL
#include "asterisk/ucm_dbi.h"
#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#endif
#include "asterisk/features.h"
#endif

AST_MODULE_LOG("Bridge");
#endif

/*** DOCUMENTATION
	<application name="ConfBridge" language="en_US">
		<synopsis>
			Conference bridge application.
		</synopsis>
		<syntax>
			<parameter name="conference" required="true">
				<para>Name of the conference bridge.  You are not limited to just
				numbers.</para>
			</parameter>
			<parameter name="bridge_profile">
				<para>The bridge profile name from confbridge.conf.  When left blank,
				a dynamically built bridge profile created by the CONFBRIDGE dialplan
				function is searched for on the channel and used.  If no dynamic
				profile is present, the 'default_bridge' profile found in
				confbridge.conf is used. </para>
				<para>It is important to note that while user profiles may be unique
				for each participant, mixing bridge profiles on a single conference
				is _NOT_ recommended and will produce undefined results.</para>
			</parameter>
			<parameter name="user_profile">
				<para>The user profile name from confbridge.conf.  When left blank,
				a dynamically built user profile created by the CONFBRIDGE dialplan
				function is searched for on the channel and used.  If no dynamic
				profile is present, the 'default_user' profile found in
				confbridge.conf is used.</para>
			</parameter>
			<parameter name="menu">
				<para>The name of the DTMF menu in confbridge.conf to be applied to
				this channel.  When left blank, a dynamically built menu profile
				created by the CONFBRIDGE dialplan function is searched for on
				the channel and used. If no dynamic profile is present, the
				'default_menu' profile found in confbridge.conf is used.</para>
			</parameter>
		</syntax>
		<description>
			<para>Enters the user into a specified conference bridge.  The user can
			exit the conference by hangup or DTMF menu option.</para>
			<para>This application sets the following channel variable upon completion:</para>
			<variablelist>
				<variable name="CONFBRIDGE_RESULT">
					<value name="FAILED">The channel encountered an error and could not enter the conference.</value>
					<value name="HANGUP">The channel exited the conference by hanging up.</value>
					<value name="KICKED">The channel was kicked from the conference.</value>
					<value name="ENDMARKED">The channel left the conference as a result of the last marked user leaving.</value>
					<value name="DTMF">The channel pressed a DTMF sequence to exit the conference.</value>
					<value name="TIMEOUT">The channel reached its configured timeout.</value>
				</variable>
			</variablelist>
		</description>
		<see-also>
			<ref type="application">ConfBridge</ref>
			<ref type="function">CONFBRIDGE</ref>
			<ref type="function">CONFBRIDGE_INFO</ref>
		</see-also>
	</application>
	<application name="ConfBridgeSFUHelper" language="en_US">
		<synopsis>
			Conference bridge sfu helper application.
		</synopsis>
		<syntax>
			<parameter name="conference" required="true">
				<para>Name of the conference bridge.  You are not limited to just
				numbers.</para>
			</parameter>
			<parameter name="command">
				<para>The command to be executed</para>
			</parameter>
		</syntax>
		<description>
			<para>It will help you to handle the command in sfu mode</para>
			<para>This application sets the following channel variable upon completion:</para>
			<variablelist>
				<variable name="CONFBRIDGE_SFU_HELPER_RESULT">
					<value name="FAILED">The command couldn't be executed.</value>
					<value name="SUCCESS">The command could be executed.</value>
				</variable>
			</variablelist>
		</description>
		<see-also>
			<ref type="application">ConfBridge</ref>
			<ref type="function">CONFBRIDGE</ref>
			<ref type="function">CONFBRIDGE_INFO</ref>
		</see-also>
	</application>
	<application name="ConfBridgeLog" language="en_US">
			<synopsis>
				Writes to the conference report database.
			</synopsis>
			<syntax>
				<parameter name="confname" required="true" />
				<parameter name="name" required="true" />
				<parameter name="number" required="true" />
				<parameter name="way" required="true" />
				<parameter name="status" required="true" />
				<parameter name="group" />
			</syntax>
			<description>
				<para>Allows you to write your own events into the report database.</para>
				<para>Example: ConfBridgeLog(6300,'Alice','1000','CALLIN','NOANSWER','support')</para>
			</description>
	</application>
	<application name="ConfBridgeFloorContrl" language="en_US">
			<synopsis>
				Handle share audio for floor contrl.
			</synopsis>
			<syntax>
				<parameter name="conf" required="true" />
				<parameter name="audioshare" required="true" />
			</syntax>
			<description>
				<para>Handle share audio for floor contrl</para>
				<para>Example: ConfBridgeFloorContrl(6300, 'true')</para>
			</description>
	</application>
	<application name="ConfBridgeEarlyMedia" language="en_US">
			<synopsis>
				Handle early media for SFU.
			</synopsis>
			<syntax>
				<parameter name="conf" required="true" />
				<parameter name="action" required="true" />
			</syntax>
			<description>
				<para>Handle the early media for SFU</para>
				<para>Example: ConfBridgeEarlyMedia(6300,'join')</para>
			</description>
	</application>
	<application name="ConfBridgeCommand" language="en_US">
			<synopsis>
				Handle a command for SFU.
			</synopsis>
			<syntax>
				<parameter name="command" required="true" />
				<parameter name="conf" required="true" />
				<parameter name="args" />
			</syntax>
			<description>
				<para>Handle a command for SFU</para>
				<para>Example: ConfBridgeCommand(6300,'DeleteAuthUser','1000','webrtc')</para>
			</description>
	</application>
	<function name="CONFBRIDGE" language="en_US">
		<synopsis>
			Set a custom dynamic bridge, user, or menu profile on a channel for the
			ConfBridge application using the same options available in confbridge.conf.
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>To what type of conference profile the option applies.</para>
				<enumlist>
					<enum name="bridge"></enum>
					<enum name="menu"></enum>
					<enum name="user"></enum>
				</enumlist>
			</parameter>
			<parameter name="option" required="true">
				<para>Option refers to a <filename>confbridge.conf</filename> option
				that is being set dynamically on this channel, or <literal>clear</literal>
				to remove already applied profile options from the channel.</para>
			</parameter>
		</syntax>
		<description>
			<para>A custom profile uses the default profile type settings defined in
			<filename>confbridge.conf</filename> as defaults if the profile template
			is not explicitly specified first.</para>
			<para>For <literal>bridge</literal> profiles the default template is <literal>default_bridge</literal>.</para>
			<para>For <literal>menu</literal> profiles the default template is <literal>default_menu</literal>.</para>
			<para>For <literal>user</literal> profiles the default template is <literal>default_user</literal>.</para>
			<para>---- Example 1 ----</para>
			<para>In this example the custom user profile set on the channel will
			automatically be used by the ConfBridge application.</para>
			<para>exten => 1,1,Answer()</para>
			<para>; In this example the effect of the following line is</para>
			<para>; implied:</para>
			<para>; same => n,Set(CONFBRIDGE(user,template)=default_user)</para>
			<para>same => n,Set(CONFBRIDGE(user,announce_join_leave)=yes)</para>
			<para>same => n,Set(CONFBRIDGE(user,startmuted)=yes)</para>
			<para>same => n,ConfBridge(1) </para>
			<para>---- Example 2 ----</para>
			<para>This example shows how to use a predefined user profile in
			<filename>confbridge.conf</filename> as a template for a dynamic profile.
			Here we make an admin/marked user out of the <literal>my_user</literal>
			profile that you define in <filename>confbridge.conf</filename>.</para>
			<para>exten => 1,1,Answer()</para>
			<para>same => n,Set(CONFBRIDGE(user,template)=my_user)</para>
			<para>same => n,Set(CONFBRIDGE(user,admin)=yes)</para>
			<para>same => n,Set(CONFBRIDGE(user,marked)=yes)</para>
			<para>same => n,ConfBridge(1)</para>
		</description>
	</function>
	<function name="CONFBRIDGE_INFO" language="en_US">
		<synopsis>
			Get information about a ConfBridge conference.
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>What conference information is requested.</para>
				<enumlist>
					<enum name="admins">
						<para>Get the number of admin users in the conference.</para>
					</enum>
					<enum name="locked">
						<para>Determine if the conference is locked. (0 or 1)</para>
					</enum>
					<enum name="marked">
						<para>Get the number of marked users in the conference.</para>
					</enum>
					<enum name="muted">
						<para>Determine if the conference is muted. (0 or 1)</para>
					</enum>
					<enum name="parties">
						<para>Get the number of users in the conference.</para>
					</enum>
				</enumlist>
			</parameter>
			<parameter name="conf" required="true">
				<para>The name of the conference being referenced.</para>
			</parameter>
		</syntax>
		<description>
			<para>This function returns a non-negative integer for valid conference
			names and an empty string for invalid conference names.</para>
		</description>
	</function>
	<function name="CONFBRIDGE_SFU_INFO" language="en_US">
		<synopsis>
			Get information about a ConfBridge conference.
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>What conference information is requested.</para>
				<enumlist>
					<enum name="role">
						<para>Get the number of admin users in the conference.</para>
					</enum>
				</enumlist>
			</parameter>
			<parameter name="conf" required="true">
				<para>The name of the conference being referenced.</para>
			</parameter>
		</syntax>
		<description>
			<para>This function returns a non-negative integer for valid conference
			names and an empty string for invalid conference names.</para>
		</description>
	</function>
	<function name="CONFBRIDGE_IM_INFO" language="en_US">
		<synopsis>
			Get im information about a ConfBridge conference.
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>What im conference information is requested.</para>
				<enumlist>
					<enum name="groupid">
						<para>Get the groupid of im in the conference.</para>
					</enum>
				</enumlist>
			</parameter>
			<parameter name="conf" required="true">
				<para>The name of the conference being referenced.</para>
			</parameter>
		</syntax>
		<description>
			<para>This function returns a non-negative integer for valid conference
			names and an empty string for invalid conference names.</para>
		</description>
	</function>
	<function name="CONFBRIDGE_USER_TERMINAL_TYPE" language="en_US">
		<synopsis>
			Get user terminal type about a confbridge user.
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>What information is requested.</para>
				<enumlist>
					<enum name="terminal">
						<para>Get the terminal type of the confbridge user.</para>
					</enum>
				</enumlist>
			</parameter>
		</syntax>
		<description>
			<para>This function returns a terminal type of the confbridge user.</para>
		</description>
	</function>
	<function name="CONFBRIDGE_EARLY_JOIN" language="en_US">
		<synopsis>
			Get schedule conference information that you joined early.
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>What conference information is requested.</para>
				<enumlist>
					<enum name="password">
						<para>Verify that the password of the schedule conference is correct.</para>
					</enum>
					<enum name="bookid">
						<para>Get the bookid of the schedule conference that you joined early.</para>
					</enum>
					<enum name="earlyinfo">
						<para>Determine whether there is schedule conference info in the waiting list.</para>
					</enum>
				</enumlist>
			</parameter>
			<parameter name="datastr" required="false">
				<para>If verify the password, datastr Needed .</para>
			</parameter>
			<parameter name="conf" required="true">
				<para>The name of the conference that you joined early.</para>
			</parameter>
		</syntax>
		<description>
			<para>This function returns a non-negative integer for valid conference
			names and an empty string for invalid conference names.</para>
		</description>
	</function>
	<function name="CONFBRIDGE_SCHEDULE_CHECK_ROLE" language="en_US">
		<synopsis>
			Get the configured role about a confbridge schedule user in the database .
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>What information is requested.</para>
				<enumlist>
					<enum name="role">
						<para>Get the configured role of the confbridge schedule user in the database .</para>
					</enum>
				</enumlist>
			</parameter>
			<parameter name="dataval" required="false">
				<para>The reservation id of the schedule meeting.</para>
			</parameter>
			<parameter name="conf" required="true">
				<para>The name of the conference.</para>
			</parameter>
		</syntax>
		<description>
			<para>This function returns a configured role of the confbridge schedule user in the database .</para>
		</description>
	</function>
	<manager name="ConfbridgeList" language="en_US">
		<synopsis>
			List participants in a conference.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true">
				<para>Conference number.</para>
			</parameter>
		</syntax>
		<description>
			<para>Lists all users in a particular ConfBridge conference.
			ConfbridgeList will follow as separate events, followed by a final event called
			ConfbridgeListComplete.</para>
		</description>
	</manager>
	<managerEvent language="en_US" name="ConfbridgeList">
		<managerEventInstance class="EVENT_FLAG_REPORTING">
			<synopsis>Raised as part of the ConfbridgeList action response list.</synopsis>
			<syntax>
				<parameter name="Conference">
					<para>The name of the Confbridge conference.</para>
				</parameter>
				<parameter name="Admin">
					<para>Identifies this user as an admin user.</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="MarkedUser">
					<para>Identifies this user as a marked user.</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="WaitMarked">
					<para>Must this user wait for a marked user to join?</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="EndMarked">
					<para>Does this user get kicked after the last marked user leaves?</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="Waiting">
					<para>Is this user waiting for a marked user to join?</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="Muted">
					<para>The current mute status.</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="Talking">
					<para>Is this user talking?</para>
					<enumlist>
						<enum name="Yes"/>
						<enum name="No"/>
					</enumlist>
				</parameter>
				<parameter name="AnsweredTime">
					<para>The number of seconds the channel has been up.</para>
				</parameter>
				<channel_snapshot/>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<manager name="ConfbridgeListRooms" language="en_US">
		<synopsis>
			List active conferences.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
		</syntax>
		<description>
			<para>Lists data about all active conferences.
				ConfbridgeListRooms will follow as separate events, followed by a final event called
				ConfbridgeListRoomsComplete.</para>
		</description>
	</manager>
	<manager name="ConfbridgeMute" language="en_US">
		<synopsis>
			Mute a Confbridge user.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true">
				<para>If this parameter is not a complete channel name, the first channel with this prefix will be used.</para>
				<para>If this parameter is "all", all channels will be muted.</para>
				<para>If this parameter is "participants", all non-admin channels will be muted.</para>
			</parameter>
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeUnmute" language="en_US">
		<synopsis>
			Unmute a Confbridge user.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true">
				<para>If this parameter is not a complete channel name, the first channel with this prefix will be used.</para>
				<para>If this parameter is "all", all channels will be unmuted.</para>
				<para>If this parameter is "participants", all non-admin channels will be unmuted.</para>
			</parameter>
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeKick" language="en_US">
		<synopsis>
			Kick a Confbridge user.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true" >
				<para>If this parameter is "all", all channels will be kicked from the conference.</para>
				<para>If this parameter is "participants", all non-admin channels will be kicked from the conference.</para>
			</parameter>
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeLock" language="en_US">
		<synopsis>
			Lock a Confbridge conference.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeUnlock" language="en_US">
		<synopsis>
			Unlock a Confbridge conference.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeStartRecord" language="en_US">
		<synopsis>
			Start recording a Confbridge conference.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="RecordFile" required="false" />
		</syntax>
		<description>
			<para>Start recording a conference. If recording is already present an error will be returned. If RecordFile is not provided, the default record file specified in the conference's bridge profile will be used, if that is not present either a file will automatically be generated in the monitor directory.</para>
		</description>
	</manager>
	<manager name="ConfbridgeStopRecord" language="en_US">
		<synopsis>
			Stop recording a Confbridge conference.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeSetSingleVideoSrc" language="en_US">
		<synopsis>
			Set a conference user as the single video source distributed to all other participants.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true">
				<para>If this parameter is not a complete channel name, the first channel with this prefix will be used.</para>
			</parameter>
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeSfuAuthorizeUser" language="en_US">
		<synopsis>
			Authorize a user and cache it for be authenticated subsequently.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Extension" required="true" />
			<parameter name="Role" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeSfuTransferModerator" language="en_US">
		<synopsis>
			Transfer the moderator of sfu to the sepecial participant.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true" />
			<parameter name="Mode" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeSfuSwitchVideoQuality" language="en_US">
		<synopsis>
			Switch the video quality between 360P and 1080P
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true" />
			<parameter name="Quality" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeAuthorizeUser" language="en_US">
		<synopsis>
			Authorize a user and cache it for be authenticated subsequently.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Extension" required="true" />
			<parameter name="Role" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeTransferModerator" language="en_US">
		<synopsis>
			Transfer the moderator to the special participant.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true" />
			<parameter name="Mode" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeRenameUser" language="en_US">
		<synopsis>
			Change the name to the special participant.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Channel" required="true" />
			<parameter name="Name" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeUpdateAdminPin" language="en_US">
		<synopsis>
			Change the admin pin for a conference.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Admin_pincode" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeReportStatisAndRemoveSchedule" language="en_US">
		<synopsis>
			Clean schedule conference list.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Reservation_id" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeScheduleStart" language="en_US">
		<synopsis>
			Inform the schedule conference information.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Reservation_id" required="true" />
			<parameter name="Conf_type" required="true" />
			<parameter name="Utc_start_time" required="true" />
			<parameter name="Conf_subject" required="true" />
			<parameter name="Host_exten" required="false" />
			<parameter name="Host_email" required="false" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeScheduleEnd" language="en_US">
		<synopsis>
			Inform the schedule conference information that the schedule conference is end.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeScheduleJoinEarly" language="en_US">
		<synopsis>
			Inform the schedule conference information that joined in advance .
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Reservation_id" required="true" />
			<parameter name="Conf_type" required="true" />
			<parameter name="Pin_code" required="true" />
			<parameter name="Adminpin_code" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeScheduleWaitInfoModify" language="en_US">
		<synopsis>
			Modify the schedule conference information that joined in advance .
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Reservation_id" required="true" />
			<parameter name="Request_name" required="true" />
			<parameter name="Schedule_data" required="false" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeDumpPCMData" language="en_US">
		<synopsis>
			Start/Stop dump PCM data.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Conference" required="true" />
			<parameter name="Opt" required="true" />
		</syntax>
		<description>
		</description>
	</manager>
	<manager name="ConfbridgeHaSyncRecordCount" language="en_US">
		<synopsis>
			Sync conference record count for HA switch master
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
		</syntax>
		<description>
		</description>
	</manager>

***/

/*!
 * \par Playing back a file to a channel in a conference
 * You might notice in this application that while playing a sound file
 * to a channel the actual conference bridge lock is not held. This is done so
 * that other channels are not blocked from interacting with the conference bridge.
 * Unfortunately because of this it is possible for things to change after the sound file
 * is done being played. Data must therefore be checked after reacquiring the conference
 * bridge lock if it is important.
 */

static const char app[] = "ConfBridge";
#ifdef GRANDSTREAM_NETWORKS
static const char app_sfu[] = "ConfBridgeSFUHelper";
static const char app_confearlymedia[] = "ConfBridgeEarlyMedia";
static const char app_confbridge_command[] = "ConfBridgeCommand";
static const char app_conffloorcontrl[] = "ConfBridgeFloorContrl";
static const char app_pageconf_change_caller[] = "PageConfChangeCaller";
static const char app_privateintercom_callflip[] = "PrivateintercomCallflip";
static const char app_confrecordcontrl[] = "ConfBridgeRecordContrl";
#endif

/*! Number of buckets our conference bridges container can have */
#define CONFERENCE_BRIDGE_BUCKETS 53

/*! Initial recording filename space. */
#define RECORD_FILENAME_INITIAL_SPACE	128

#ifdef GRANDSTREAM_NETWORKS
#define RECORD_FILES_INITIAL_SPACE			1024
#define RECORD_FILES_DESC_INITIAL_SPACE		4096
#define RECORD_FILES_UUID_INITIAL_SPACE		4096


#define MCB_RELEASE_COUNT 1 //Every number of MCB RELEASE COUNT
#define MCB_RELEASE_FREQUENCY 1 //Per second, so 1 would be every second, 1000 would be milliseconds
#define MCB_IDLE_KICK_SECONDS 120 //auto release mcb in 120 sec when no other user in conference

#define SOFTBRIDGE_VIDEO_DEST_PREFIX "softbridge_dest"
#define SOFTBRIDGE_VIDEO_DEST_LEN strlen(SOFTBRIDGE_VIDEO_DEST_PREFIX)
#define SOFTBRIDGE_VIDEO_DEST_SEPARATOR '_'

#define DATE_FORMAT "%Y-%m-%d %H:%M:%S"
#define MENU_SFU_USER "sfu_user_menu"
#define MENU_SFU_ADMIN "sfu_admin_menu"
#define MENU_ADMIN_CONF_MENU "ucm_AdminCallerMenu_UserInviteMenu"
#define MENU_USER_CONF_MENU_INVITE "ucm_UserCallerMenu_UserInviteMenu"
#define MENU_USER_CONF_MENU "ucm_UserCallerMenu"
#define MENU_INVITE "ucm_UserInviteMenu"
#define MENU_DEFAULT DEFAULT_MENU_PROFILE
static char *app_cbl = "ConfBridgeLog";
#define CONF_CONFIG_MAX_LEN 16
#define MAX_USER_NUM_NO_IGNORE_AVS_RESP  (100)
#define AST_MAX_SCHEDULE_ID      128 /*!< Max length of the schedule conference id */
#define AST_MAX_START_TIME       64  /*!< The utc start time of schedule conference */
#define AST_MAX_CONF_SUBJECT     512 /*!< The conf subject of schedule conference */
#define AST_MAX_HOST_EXTEN       64  /*!< The host extension of schedule conference */
#define AST_MAX_HOST_EMAIL       128 /*!< The host email of schedule conference */
#define AST_MAX_IM_GROUP_ID      128 /*!< The group id of im group conference */
#define AST_MAX_RELATED_INFO   256
#define AST_MAX_PATH_LENGTH    512
#endif

/*! \brief Container to hold all conference bridges in progress */
struct ao2_container *conference_bridges;

#ifdef GRANDSTREAM_NETWORKS
static struct ast_sched_context *conf_sched;
static struct ast_sched_context *user_sched;
static struct ast_sched_context *recycle_sched;
struct conf_sched_task_data{
	char *user_channel;
	char *conf_name;
};
#define CONF_SCH_RETRY 3
#define CONF_PEER_TIMEOUT (5*60*1000) //5min
#define CONF_EARLY_JOIN_MAX_TIME (10*60*1000) //10min
#define CONF_SCH_RECYCLE_TIMEOUT (15*60*1000) //15min

static int globalEnabledCEI = 0;
static struct stasis_message_router *ccm_command_router = NULL;
static struct stasis_message_router *cei_toggle_enable_router = NULL;
static struct stasis_message_router *operation_log_router = NULL;
/*! \brief Container to hold the contact groups*/
struct ao2_container *conference_contact_groups = NULL;
/*! \brief Conference Report Database Handler*/

static int alarm_on = 0;

#define CONF_LOG_DB      "conference_report"
#define RECORDING_FILES_DB	"recording_files"

#define safe_ast_free(buf) if (buf) {ast_free(buf);	buf = NULL;}

#define AST_DB_NAME_MAX_LEN 64
char s_dbAmiName[AST_DB_NAME_MAX_LEN] = {0};
char s_dbConfigName[AST_DB_NAME_MAX_LEN] = {0};
char s_dbOperationLog[AST_DB_NAME_MAX_LEN] = {0};

enum {
	CONF_TYPE_NONE,
	CONF_TYPE_AUDIO,
	CONF_TYPE_VIDEO
};

enum {
	CONF_HANGUP_NONE,
	CONF_HANGUP_PALYKICKED,   /*!< hangup by schedule task*/
	CONF_HANGUP_CONFTIMEOUT,  /*!< hangup by conf max time*/
	CONF_HANGUP_WEBHOST,      /*!< hangup by web host*/
	CONF_HANGUP_ACTIVECALL,   /*!< hangup by active call*/
	CONF_HANGUP_MEETINGLIST    /*!< hangup by meeting list*/
};

#ifdef HAVE_UNIXODBC

int count;				/*!< Count of conference call*/
AST_MUTEX_DEFINE_STATIC(record_count_lock);

struct generic_prepare_struct {
	char *sql;
	int argc;
	char **argv;
};

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}

static const char *conf_report_get_user_role(struct confbridge_user *user);
static const char *conf_report_get_user_join_way(struct confbridge_user *user);
static int conf_report_update_call_count(void);
static int conf_report_get_call_count(void);
static int conf_report_db_write_conference(struct confbridge_conference *conference);
static int conference_write_statistics(struct confbridge_conference *conference, int id, time_t start_timep, char *utcStartTime, char *utcEndTime);
static int conf_report_db_write_user(struct confbridge_user *user);
static int conf_report_get_user_record(const char *number, int conf_id, const char*type);
static int conf_record_get_file_record(const char *filename);
static int conf_report_get_user_by_number_and_device(const char *number, const char *device_type, int conf_id);
static int conf_report_db_write(char *cmd);
static int conf_record_db_write(char *cmd);
static struct confbridge_user *find_conf_user_by_number(struct confbridge_conference *conference, const char *number);
static void destroy_conf_user_report_record(struct conf_report_user_record *report_record);
static int conf_get_useragent(struct ast_channel* chan, char* buf, size_t size);
static char* str_replace(const char *str, const char *match, const char *replace);

static struct conf_sched_task_data *conf_sched_task_data_alloc(const char *conf_name, const char *user_channel);
static int only_user_timeout_notify(const void *data);
static int conf_timeout_notify_build_send(struct ast_channel *chan, char *conf_name, int type, int status, int time);
static int cache_conf_update_moderator(const char *conf_number, const char *extension, struct ast_channel *chan);
static int reload_menu_to_user(struct confbridge_user *user, const char *menu_profile_name);
static void playPageVideoRecord_Callee(struct confbridge_conference *conference);
static int conf_get_sca_shareline(struct ast_channel* chan, char* buf, size_t size);
static void clean_schedule_meeting_status(const char *conf_name, int conf_type);
static void conf_user_bridge_suspend(struct confbridge_user *user);
static void conf_user_bridge_unsuspend(struct confbridge_user *user);
#endif

#if 0
//MCB Auto Release
AST_MUTEX_DEFINE_STATIC(mcb_release_lock);
static ast_cond_t mcb_release_cond = PTHREAD_COND_INITIALIZER;
static pthread_t mcb_release_thread = AST_PTHREADT_NULL;
static unsigned char mcb_release_thread_run;
#endif

struct cache_authorized_user{
	struct timeval start_time;
	char extension[AST_MAX_EXTENSION];
	char channel_name[AST_CHANNEL_NAME];
	char terminal_type[AST_MAX_TERMINAL_NAME];
	char user_schedule_id[AST_MAX_SCHEDULE_ID];
	char related_info[AST_MAX_RELATED_INFO];
	int is_admin;
	AST_LIST_ENTRY(cache_authorized_user) entry;
};

struct cache_conference{
	char name[MAX_CONF_NAME];
	char schedule_id[AST_MAX_SCHEDULE_ID];
	int type;
	struct ast_channel *moderator;
	AST_LIST_HEAD_NOLOCK(, cache_authorized_user) auth_user_list;  /*pre-authorized users from gswave*/
	AST_LIST_ENTRY(cache_conference) entry;
};
AST_LIST_HEAD_STATIC(cache_conference_list, cache_conference);

static volatile int next_conf_uniqueid = 1; /* Used to assign unique id to conference */

static const char *conf_operation_str[] = {
	"START",
	"END",
	"START RECORD",
	"STOP RECORD",
	"JOIN",
	"LEAVE",
	"INVITE",
	"MCB",
	"INVITE CONTACT GROUP",
	"OPEN CAMERA",
	"CLOSE CAMERA",
	"OPEN SHARE",
	"CLOSE SHARE",
	"MUTE",
	"MUTE PARTICIPANTS",
	"UNMUTE",
	"UNMUTE PARTICIPANTS",
	"LOCK",
	"UNLOCK",
	"KICK",
	"KICK MCB",
	"KICK PARTICIPANTS",
	"KICK LAST",
	"TRANSFER HOST",
	"RETRIEVE HOST",
	"SWITCH RESOLUTION",
	"CHANGE ROLE",
	"HANGUP",
	"INCREASE LISTENING VOLUME",
	"DECREASE LISTENING VOLUME",
	"INCREASE TALKING VOLUME",
	"DECREASE TALKING VOLUME",
	"UPDATE NAME",
	"MIC MUTE",
	"MIC UNMUTE",
	"SET FOCUS",
	"DEL FOCUS",
	"SET LAYOUT",
	"DEL LAYOUT",
	"SET USER PERMISSION",
	"DEL USER PERMISSION"
};

enum {
	CONF_TYPE_WAVE_WEB,
	CONF_TYPE_WAVE_DESKTOP,
	CONF_TYPE_WAVE_ANDROID,
	CONF_TYPE_WAVE_IOS,
	CONF_TYPE_GVC,
	CONF_TYPE_OTHER
};

static const char *confuser_device_full_type[] = {
	"Grandstream Wave Web",
	"Grandstream Wave Desktop",
	"Grandstream Wave Android",
	"Grandstream Wave iOS",
	"GVC",
	"Other"
};

struct schedule_attendees{
	char extension[AST_MAX_EXTENSION];
	char channel_name[AST_CHANNEL_NAME];
	char terminal_type[AST_MAX_TERMINAL_NAME];
	AST_LIST_ENTRY(schedule_attendees) entry;
};

struct schedule_conference{
	int type;
	int max_user;
	int total_user;
	int is_random_conf;
	time_t conference_start;
	char name[MAX_CONF_NAME];
	char schedule_id[AST_MAX_SCHEDULE_ID];
	char utc_start_time[AST_MAX_START_TIME];
	char start_time[AST_MAX_START_TIME];
	char conf_subject[AST_MAX_CONF_SUBJECT];
	char host_exten[AST_MAX_HOST_EXTEN];
	char host_email[AST_MAX_HOST_EMAIL];
	unsigned int recycle_sched_id;			/*!< sched id for 15min auto release */
	AST_LIST_HEAD_NOLOCK(, schedule_attendees) schedule_attendee_list;
	AST_LIST_ENTRY(schedule_conference) entry;
};
AST_LIST_HEAD_STATIC(schedule_conference_list, schedule_conference);

struct im_conference{
	int type;
	int status;
	int strong_related;
	time_t im_conference_start;
	char name[MAX_CONF_NAME];
	char im_group_id[AST_MAX_IM_GROUP_ID];
	char conf_subject[AST_MAX_CONF_SUBJECT];
	AST_LIST_ENTRY(im_conference) entry;
};
AST_LIST_HEAD_STATIC(im_conference_list, im_conference);

enum {
	CONF_IM_NONE,
	CONF_IM_INIT,
	CONF_IM_INUSE,
	CONF_IM_END
};

struct schedule_attendees_wait{
	int is_admin;
	char channel_name[AST_CHANNEL_NAME];
	char extension[AST_MAX_EXTENSION];
	struct ast_channel *chan;
	AST_LIST_ENTRY(schedule_attendees_wait) entry;
};

struct schedule_conference_wait{
	int type;
	int cancelled;
	int has_notify;
	int wait_users;
	char pin[MAX_PIN];
	char admin_pin[MAX_PIN];
	char name[MAX_CONF_NAME];
	char schedule_id[AST_MAX_SCHEDULE_ID];
	char utc_start_time[AST_MAX_START_TIME];
	char start_time[AST_MAX_START_TIME];
	char conf_subject[AST_MAX_CONF_SUBJECT];
	char host_exten[AST_MAX_HOST_EXTEN];
	char host_email[AST_MAX_HOST_EMAIL];
	AST_LIST_HEAD_NOLOCK(, schedule_attendees_wait) schedule_attendees_wait_list;
	AST_LIST_ENTRY(schedule_conference_wait) entry;
};
AST_LIST_HEAD_STATIC(schedule_conference_wait_list, schedule_conference_wait);

static void publish_cei_unsub(char *conference, const char *channel, const char *cid_number);
static void generic_mute_unmute_user(struct confbridge_conference *conference,
									 struct confbridge_user *user, int mute, int is_mic, int play_sound);
static void conf_set_contact_group_inuse(struct confbridge_user *user);
static int conf_contact_group_update_caller_name(struct confbridge_user *user);
#endif
static void leave_conference(struct confbridge_user *user);
static int play_sound_number(struct confbridge_conference *conference, int say_number);
static int execute_menu_entry(struct confbridge_conference *conference,
	struct confbridge_user *user,
	struct ast_bridge_channel *bridge_channel,
	struct conf_menu_entry *menu_entry,
	struct conf_menu *menu);
#ifdef GRANDSTREAM_NETWORKS
static int kick_conference_participant(struct confbridge_conference *conference, const char *channel, const char *originator);
static int conf_user_join_handle_one_person_warning(struct confbridge_user *user);
static int conf_user_leave_handle_one_person_warning(struct confbridge_user *user);
int conf_user_is_video_call(struct confbridge_user *user);
static int cache_conf_remove_moderator(const char *conf_number, const struct ast_channel *chan);
static void cache_conf_remove(const char *conf_number);
static struct cache_conference *cache_conf_fetch_add(const char *conf_number, int type, int add, const char *schedule_id);
//static void cache_conf_remove_user(const char *conf_number, const char *extension);
static void cache_conf_remove_user_by_channel_name(const char *conf_number, const char *channel_name);
static int cache_conf_authorized_user_set_channel(const char *conf_number, const char *extension, struct ast_channel *chan);
static int action_change_role(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user);
static char *handle_cli_confbridge_cache_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
char *handle_cli_confbridge_resource_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static int sfu_topologies_on_moderator_change(struct confbridge_conference *conference, struct confbridge_user *moderator_user);
static int get_schedule_conference(struct confbridge_conference *conference);
static unsigned int get_schedule_notify_countdown(struct confbridge_conference *conference);
static char *build_schedule_notify_alert_body(const char *conf_number, const char *schedule_id, const char *start_time, unsigned int countdown);
static struct confbridge_user* find_active_user_by_channel(struct confbridge_conference *conference, const char *channel);
/*Generate pdf report for scheduled meeting*/
static char *handle_cli_schedule_conference_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static int conference_write_user_account(struct confbridge_user *user);
static int conference_user_write_reservation_log(struct confbridge_user *user, int op_type, char *user_agent);
static int conference_write_statistics_by_confschedule(struct schedule_conference *conference);
static void schedule_conference_remove(const char *conf_number, const char *reservation_id);
static void schedule_conference_destroy(void *obj);
static struct schedule_conference *schedule_conf_fetch_add(struct confbridge_conference *conference);
static struct schedule_conference *schedule_conf_cli_fetch(const char *conference);
static void schedule_attendees_destroy(void *obj);
static int schedule_attendee_add(struct schedule_conference *conference, const char *extension, const char *terminal_type);
static struct schedule_attendees *schedule_attendee_find(struct schedule_conference *conference, const char *extension, const char *terminal_type);
static void schedule_conference_error_recovery_detect(struct confbridge_conference *conference);
static void schedule_conference_attendee_join_list(struct confbridge_conference *conference, struct confbridge_user *user);
static void schedule_attendee_list(struct schedule_conference *conference);
const char *get_confuser_device_type(const char *user_agent);
static void schedule_conference_end_delay(struct confbridge_conference *conference);
/*for im group mconference*/
static void im_conf_list_remove_error(void);
static void im_conf_list_remove(const char *conf_number);
static int im_conf_list_update_status(const char *conf_number, int conf_status);
static int conf_status_astdb_del(const char *family_name, const char *value, const char *key_name);
static struct im_conference *im_conf_list_fetch_add(const char *conf_number, const char *im_group_id, int strong_related, int type, int add, const char *conf_subject);
static void im_conference_list_destroy(void *obj);
static char *handle_cli_im_conference_info_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);

static void schedule_conference_wait_destroy(void *obj);
static struct schedule_attendees_wait *schedule_attendee_wait_find(struct schedule_conference_wait *conference, const char *chan_name);
//static struct schedule_conference_wait *schedule_conf_wait_fetch_add(const char *conference, int add, const char *pincode, const char *admin_pincode, const char *schedule_id, int conf_type, const char *subject, const char *host_exten, const char *host_email, const char *utc_start_time);
static struct schedule_conference_wait *schedule_conf_wait_info_fetch_add(int add, struct schedule_conference_wait_info* conf_sch_w);
static void schedule_attendee_wait_list_destroy(void *obj);
static int schedule_attendee_wait_add(struct schedule_conference_wait *conference, struct ast_channel *chan);
static int schedule_conference_attendee_join_wait_list(const char *conference, struct ast_channel *chan);
static void schedule_conference_wait_list_remove(const char *conf_number);
static int schedule_conference_wait_user_add_cache(struct ast_channel *chan, const char *conference_name, int type, const char *schedule_id);
static int schedule_conference_wait_user_remove_cache(const char *conference, struct ast_channel *chan);
static void schedule_conference_wait_list_remove_by_scheduleid(const char *conf_number, const char *schedule_id);
static void schedule_conference_wait_user_remove_by_channel(const char *conference, const char *chan_name);
static void schedule_conference_wait_list_update_by_scheduleid(const char *conf_number, const char *schedule_id, const char *cofn_info);
static int schedule_conference_need_wait(const char *conference_name, const char *schedule_id, int *inuse_notify);
//static void conference_attendee_auth(const char *conference_name, const char *schedule_id);
static int schedule_conference_wait_user_join(const char *conference_name);
static char *handle_cli_schedule_conference_early_start_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_schedule_conference_early_start_modify(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_schedule_conference_need_wait(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_conference_current_users(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static char *handle_cli_conference_current_host(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static void schedule_conference_inuse_notify(struct confbridge_conference *conference);
static void schedule_conference_wait_inuse_notify(struct confbridge_conference *conference, const char *schedule_id);
static int schedule_conference_wait_add_info(struct confbridge_conference *conference, struct schedule_conference_wait* conference_w);
static int hangup_conference(struct confbridge_conference *conference, int play_sound);
static void set_conference_record_link(struct confbridge_conference *conference, enum conf_record_type type);
const char *getLinkResult(const char *link_path);
static void check_timeout_before_schedule_notify(struct confbridge_conference *conference);
static void cancel_remote_calling(struct confbridge_user*user);
static void schedule_conference_wait_list_cancel(const char *conference);
static void handle_mcb_user_join(struct confbridge_user * user);
static void handle_mcb_user_remove(struct confbridge_user * user);
static struct schedule_conference_wait *schedule_conf_wait_info_update(struct schedule_conference_wait_info* conf_sch_w);
#endif


/*! \brief Hashing function used for conference bridges container */
static int conference_bridge_hash_cb(const void *obj, const int flags)
{
	const struct confbridge_conference *conference = obj;
	const char *name = obj;
	int hash;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
	default:
	case OBJ_POINTER:
		name = conference->name;
		/* Fall through */
	case OBJ_KEY:
		hash = ast_str_case_hash(name);
		break;
	case OBJ_PARTIAL_KEY:
		/* Should never happen in hash callback. */
		ast_assert(0);
		hash = 0;
		break;
	}
	return hash;
}

/*! \brief Comparison function used for conference bridges container */
static int conference_bridge_cmp_cb(void *obj, void *arg, int flags)
{
	const struct confbridge_conference *left = obj;
	const struct confbridge_conference *right = arg;
	const char *right_name = arg;
	int cmp;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
	default:
	case OBJ_POINTER:
		right_name = right->name;
		/* Fall through */
	case OBJ_KEY:
		cmp = strcasecmp(left->name, right_name);
		break;
	case OBJ_PARTIAL_KEY:
		cmp = strncasecmp(left->name, right_name, strlen(right_name));
		break;
	}
	return cmp ? 0 : CMP_MATCH;
}

#ifdef GRANDSTREAM_NETWORKS
static int mark_chan_has_joined(struct ast_channel *chan)
{
	struct ast_channel *remote_peer;
	struct ast_channel *local_peer = NULL;
	struct ast_channel *ccm_local_peer;

	/*
	 *  "|--|:bridge"
	 *  "<-->" : local bridge
	 *  case 1(without ccm):
	 *  conference |--| local/chan;1 <--> local/chan;2 |--| xxx/chan
	 *
	 *  case 2(with ccm):
	 *  conference |--| local/ccm-chan;1 <--> local/ccm-chan;2 |--| local/chan;1 <--> local/chan;2 |--| xxx/chan
	 */
	if(strstr(ast_channel_name(chan), "ccm-invite")) {
		ccm_local_peer = ast_local_get_peer(chan);
		if(ccm_local_peer) {
			struct ast_channel *local;
			pbx_builtin_setvar(ccm_local_peer, "HAS_JOIN=yes");
			ast_log(LOG_DEBUG, "Chan(%s) Set HAS_JOIN=yes.", ast_channel_name(ccm_local_peer));

			local = ast_channel_bridge_peer(ccm_local_peer);
			if(local) {
				pbx_builtin_setvar(local, "HAS_JOIN=yes");
				ast_log(LOG_DEBUG, "Chan(%s) Set HAS_JOIN=yes.", ast_channel_name(local));
				local_peer = ast_local_get_peer(local);
				ast_channel_unref(local);
			}else {
				ast_log(LOG_ERROR, "Chan(%s) fail to get peer.", ast_channel_name(chan));
			}

			ast_channel_unref(ccm_local_peer);
		}else {
			ast_log(LOG_ERROR, "Chan(%s) fail to get local peer.", ast_channel_name(chan));
		}
	}else {
		local_peer = ast_local_get_peer(chan);
	}

	if(local_peer) {
		pbx_builtin_setvar(local_peer, "HAS_JOIN=yes");
		ast_log(LOG_DEBUG, "Chan(%s) Set HAS_JOIN=yes.", ast_channel_name(local_peer));

		const char *from_paging = pbx_builtin_getvar_helper(chan, "PAGE_INTERCOM");
		if (!ast_strlen_zero(from_paging)) {
			ast_log(LOG_DEBUG, "Chan(%s) PAGE_INTERCOM=%s.", ast_channel_name(chan), from_paging);
			pbx_builtin_setvar(local_peer, "FROM_PAGING=yes");
			ast_log(LOG_DEBUG, "Local Peer Chan(%s) Set FROM_PAGING=yes.", ast_channel_name(local_peer));
		}

		char outbound_trunk_index[16] = {0};
		remote_peer = ast_channel_bridge_peer(local_peer);
		if(remote_peer) {
			pbx_builtin_setvar(remote_peer, "HAS_JOIN=yes");
			if (!ast_strlen_zero(from_paging)) {
				pbx_builtin_setvar(remote_peer, "FROM_PAGING=yes");
				ast_log(LOG_DEBUG, "Local Peer Bridge Chan(%s) Set FROM_PAGING=yes.", ast_channel_name(remote_peer));
				struct ast_channel *local_remote_peer = NULL;
				local_remote_peer = ast_local_get_peer(remote_peer);
				if (local_remote_peer) {
					pbx_builtin_setvar(local_remote_peer, "FROM_PAGING=yes");
					ast_log(LOG_DEBUG, "Local Peer Bridge Chan Local Peer Chan(%s) Set FROM_PAGING=yes.", ast_channel_name(local_remote_peer));
					const char *conf_user_invite = pbx_builtin_getvar_helper(remote_peer, "CONF_USER_INVITE");
					if (!ast_strlen_zero(conf_user_invite)) {
						pbx_builtin_setvar(local_remote_peer, "CONF_USER_INVITE=yes");
						ast_log(LOG_DEBUG, "Local Peer Bridge Chan Local Peer Chan(%s) Set CONF_USER_INVITE=yes.", ast_channel_name(local_remote_peer));
					}
					ast_channel_unref(local_remote_peer);
				}
			}
			ast_log(LOG_DEBUG, "Chan(%s) Set HAS_JOIN=yes.", ast_channel_name(remote_peer));
			ast_cdr_getvar(ast_channel_name(remote_peer),"outbound_trunk_index",outbound_trunk_index,sizeof(outbound_trunk_index)-1);
			ast_channel_unref(remote_peer);
		}else {
			ast_log(LOG_WARNING, "Chan(%s) fail to get remote peer.", ast_channel_name(chan));
		}
		if ('\0'!=outbound_trunk_index[0]){
			ast_cdr_setvar(ast_channel_name(local_peer),"outbound_trunk_index",outbound_trunk_index);
		}

		ast_channel_unref(local_peer);
	}else {
		ast_log(LOG_WARNING, "Chan(%s) fail to get local peer.", ast_channel_name(chan));
	}

	return 0;
}

static int get_pjsip_extension(const char *channel_name, char *buf, int size)
{
	const char *start = NULL;
	const char *end =  NULL;

	memset(buf, 0, size);

	if (strncasecmp(channel_name, "PJSIP/", 6)) {
		return -1;
	}
	start = channel_name + 6;
	end = strrchr(start, '-');
	if (!end) {
		return -1;
	}

	strncpy(buf, start, size > (end - start) ? (end - start) : size);
	return 0;
}

static char *get_caller_name_from_phonebooks(const char *number)
{
	char *sqlcmd = NULL;
	char *name = NULL;
	int res = -1;

	if (ast_strlen_zero(number)) {
		return NULL;
	}

#ifdef USE_DB_MYSQL
#ifdef HAVE_UNIXODBC
	struct odbc_obj *obj = NULL;
	SQLHSTMT stmt = NULL;
	char rowdata[256] = {0};

	obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);
	if (!obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
		return NULL;
	}

	ast_asprintf(&sqlcmd, "SELECT CASE WHEN firstname IS NOT NULL AND firstname <> '' AND lastname <> '' AND  lastname iS NOT NULL THEN CONCAT(CONCAT(firstname,' '),lastname) "
			"WHEN firstname IS NOT NULL AND firstname <> '' THEN firstname "
			"WHEN lastname IS NOT NULL AND lastname <> '' THEN lastname "
			"ELSE '' END AS fullname FROM phonebooks WHERE accountnumber = %s "
			"UNION ALL "
			"SELECT CASE WHEN first_name IS NOT NULL AND first_name <> '' AND last_name <> '' AND  last_name iS NOT NULL THEN CONCAT(CONCAT(first_name,' '),last_name) "
			"WHEN first_name IS NOT NULL AND first_name <> '' THEN first_name "
			"WHEN last_name IS NOT NULL AND last_name <> '' THEN last_name "
			"ELSE '' END AS fullname FROM cloud_users WHERE user_name = %s "
			"UNION ALL "
			"SELECT CASE WHEN first_name IS NOT NULL AND first_name <> '' AND last_name <> '' AND  last_name iS NOT NULL THEN CONCAT(CONCAT(first_name,' '),last_name) "
			"WHEN first_name IS NOT NULL AND first_name <> '' THEN first_name "
			"WHEN last_name IS NOT NULL AND last_name <> '' THEN last_name "
			"ELSE '' END AS fullname FROM external_contact_users WHERE phone_number = %s "
			"LIMIT 1;", number, number, number);

	struct generic_prepare_struct gps = { .sql = sqlcmd };

	stmt = ast_core_odbc_prepare_and_execute(obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(AST_LOG_WARNING, "SQL Execute error!\n[%s]\n\n", sqlcmd);
		goto fail;
	}
	res = SQLFetch(stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", sqlcmd);
		goto fail_with_handle;
	}
	res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Get Data error!\n[%s]\n\n", sqlcmd);
		goto fail_with_handle;
	}
	name = ast_strdup(rowdata);

fail_with_handle:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
fail:
	ast_core_odbc_release_obj(obj);
	obj = NULL;
	safe_ast_free(sqlcmd)
#endif
#endif

    return name;
}

static void update_caller_name_for_trunk(struct ast_channel *chan)
{
	char *name = NULL;

	if (!strstr(ast_channel_name(chan), "PJSIP/trunk") && !strstr(ast_channel_name(chan), "DAHDI/o")) {
		return;
	}

	if (ast_channel_caller(chan)->id.number.valid) {
		name = get_caller_name_from_phonebooks(ast_channel_caller(chan)->id.number.str);
		if (name) {
			if (!ast_strlen_zero(name)) {
				ast_func_write(chan, "CALLERID(name)", name);
			}
			ast_free(name);
		}
	}
}

static int conf_operation_action_to_type(const char *name)
{
	int i = 0;

	for (i = 0; i < ARRAY_LEN(conf_operation_str); i++) {
		if (!strcasecmp(conf_operation_str[i], name)) {
			return i;
		}
	}

	return -1;
}

const char *conf_operation_type_to_action(int type)
{
	return conf_operation_str[type];
}

static int conf_create_operation_feature_id(struct confbridge_conference *conference)
{
	char unique_id[AST_UUID_STR_LEN*2] = {0};
	char feature_id[AST_FEATUREID_BUFFER_LENGTH] = {0};

	if (!conference->is_conference_service) {
		return -1;
	}
	if (!ast_strlen_zero(conference->schedule_id_processing) && !ast_strlen_zero(conference->schedule_utc_start_time)) {
		ast_log(LOG_DEBUG, "conference->schedule_utc_start_time %s!", conference->schedule_utc_start_time);
		struct tm start_tm = {0};
		struct timeval start_tv = {0};
		strptime(conference->schedule_utc_start_time, "%Y-%m-%d %H:%M", &start_tm);
		start_tv.tv_sec = mktime(&start_tm);

		snprintf(unique_id, sizeof(unique_id), "%s_%u", conference->schedule_id_processing, (unsigned int)start_tv.tv_sec);
	} else if (!ast_strlen_zero(conference->schedule_id_processing)){
		ast_log(LOG_DEBUG, "conference->schedule_id_processing %s!", conference->schedule_id_processing);
		snprintf(unique_id, sizeof(unique_id), "%s_%u", conference->schedule_id_processing, (unsigned int)ast_tvnow().tv_sec);
	} else {
		char pbx_uuid[AST_UUID_STR_LEN] = {0};
		ast_uuid_init();
		ast_uuid_generate_str(pbx_uuid, sizeof(pbx_uuid));
		snprintf(unique_id, sizeof(unique_id), "%s_%u", pbx_uuid, (unsigned int)ast_tvnow().tv_sec);
	}
	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		ast_create_featureid(FEATURE_VIDEO_CONFERENCE, conference->name, unique_id, feature_id, sizeof(feature_id));
	} else {
		ast_create_featureid(FEATURE_CONFERENCE, conference->name, unique_id, feature_id, sizeof(feature_id));
	}

	conference->feature_id = ast_strdup(feature_id);

	return 0;
}

static int conference_write_statistics_by_confschedule(struct schedule_conference *conference)
{
	char *cmd = NULL;
	int nobody_join = 0;
	char *subject_safe = NULL;

	if (!conference) {
		ast_log(LOG_ERROR, "conference null!");
		return -1;
	}
	if (ast_strlen_zero(conference->schedule_id)) {
		ast_log(LOG_ERROR, "conference schedule_id null!");
		return -1;
	}

	if (!conference->total_user) {
		ast_log(LOG_DEBUG, "conference schedule may be nobody join!");
		nobody_join = 1;
	}
	if (!ast_strlen_zero(conference->conf_subject)) {
		subject_safe = str_replace(conference->conf_subject, "'", "''");
	}
	ast_asprintf(&cmd, "INSERT INTO conference_reservation_statistics (reservation_id,conf_num,conf_type,conf_subject,start_time,conf_duration,utc_start_time,host_exten,host_email,max_users,total_users, is_random_conf) VALUES ('%s','%s','%d','%s','%s','%ld','%s','%s','%s','%d','%d', '%s')",
			conference->schedule_id, conference->name, conference->type, subject_safe ? subject_safe : "", conference->start_time, nobody_join ? 0 : (time(NULL) - conference->conference_start), conference->utc_start_time,conference->host_exten,conference->host_email, conference->max_user, conference->total_user, conference->is_random_conf ? "yes":"no");
	conf_report_db_write(cmd);
	safe_ast_free(cmd);
	safe_ast_free(subject_safe);

	return 0;
}

static int conference_write_user_account(struct confbridge_user *user)
{
	char *cmd = NULL;
	char strNowTime[80] = {0};
	struct ast_tm timeresult;
	struct timeval now_time = ast_tvnow();

	if (!user || !user->conference) {
		ast_log(LOG_ERROR, "user or conference null!");
		return -1;
	}
	ast_log(LOG_DEBUG, "conference->activeusers + conference->waitingusers '%d'", user->conference->activeusers + user->conference->waitingusers);

	if (ast_strlen_zero(user->conference->schedule_id_processing) || ast_strlen_zero(user->conference->schedule_subject)
		|| ast_strlen_zero(user->conference->schedule_utc_start_time) || ast_strlen_zero(user->conference->name)
		|| ast_strlen_zero(user->conference->start_time)) {
		ast_log(LOG_DEBUG, "some schedule value null!");
		return -1;
	}

	ast_localtime(&now_time, &timeresult, NULL);
	ast_strftime(strNowTime, sizeof(strNowTime), DATE_FORMAT, &timeresult);
	ast_asprintf(&cmd, "INSERT INTO conference_reservation_user_count_realtime (reservation_id,conf_num,user_count,utc_start_time,real_time) VALUES ('%s','%s','%d','%s','%s')",
			user->conference->schedule_id_processing, user->conference->name, user->conference->activeusers + user->conference->waitingusers, user->conference->schedule_utc_start_time, strNowTime);
	conf_report_db_write(cmd);
	safe_ast_free(cmd);

	return 0;
}

const char *get_confuser_device_type(const char *user_agent)
{
	int i = 0;

	if (NULL == user_agent || user_agent[0] == '\0') {
		ast_log(LOG_DEBUG, "user agent is null, use 'Other'!");
		return "Other";
	}

	for (i = 0; i < (sizeof(confuser_device_full_type) / sizeof(const char *)); i++) {
		if (strcasestr(user_agent, confuser_device_full_type[i])) {
			switch(i) {
				case CONF_TYPE_WAVE_WEB:
					if (strcasestr(user_agent, "Chrome")) {
						return "Chrome";
					} else if (strcasestr(user_agent, "Firefox")) {
						return "Firefox";
					} else {
						return "Other";
					}
				case CONF_TYPE_WAVE_DESKTOP:
					if (strcasestr(user_agent, "windows")) {
						return "Wave Desktop(Windows)";
					} else {
						return "Wave Desktop(Mac)";
					}
				case CONF_TYPE_WAVE_ANDROID:
					return "Android";
				case CONF_TYPE_WAVE_IOS:
					return "IOS";
				case CONF_TYPE_GVC:
					return "GVC";
				case CONF_TYPE_OTHER:
					return "Other";
			}

		}
	}
	if (strcasestr(user_agent, "Grandstream")) {
		return "Phone";
	}
	return "Other";
}

static int conference_user_write_reservation_log(struct confbridge_user *user, int op_type, char *user_agent)
{
	char *cmd = NULL;
	char *caller_id = NULL;
	char *caller_name = NULL;
	char * name_tmp = NULL;
	char strStartTime[80] = {0};
	char strEndTime[80] = {0};
	struct ast_tm timeresult;
	struct timeval now_time = ast_tvnow();
	struct timeval t_start,t_end;
	int is_mcb = 0;

	if (!user || !user->conference) {
		ast_log(LOG_ERROR, "user or conference null!");
		return -1;
	}
	if (ast_strlen_zero(user->conference->schedule_id_processing) || ast_strlen_zero(user->conference->schedule_subject)
		|| ast_strlen_zero(user->conference->schedule_utc_start_time) || ast_strlen_zero(user->conference->name)
		|| ast_strlen_zero(user->conference->start_time)) {
		ast_log(LOG_DEBUG, "some schedule value null!");
		return -1;
	}
	if (op_type != CONF_OP_LEAVE)
	{
		return -1;
	}
	const char *device_type = get_confuser_device_type(user_agent);
	t_start = user->jointime;

	ast_log(LOG_DEBUG, "user->conference->schedule_id_processing '%s'", user->conference->schedule_id_processing);
	caller_id = S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous");
	caller_name = S_COR(ast_channel_caller(user->chan)->id.name.valid, ast_channel_caller(user->chan)->id.name.str, caller_id);
	ast_localtime(&user->jointime, &timeresult, NULL);
	ast_strftime(strStartTime, sizeof(strStartTime), DATE_FORMAT, &timeresult);
	ast_localtime(&now_time, &timeresult, NULL);
	ast_strftime(strEndTime, sizeof(strEndTime), DATE_FORMAT, &timeresult);
	gettimeofday(&t_end, NULL);
	if (caller_name && strstr(caller_name, "'")) {
		name_tmp = str_replace(caller_name, "'", "''");
		ast_log(LOG_DEBUG, "name is %s after replace is %s\n", caller_name, name_tmp);
	}
	/*get real exten*/
	char real_exten[AST_MAX_EXTENSION] = {0};
	if (user->chan) {
		conf_get_exten_by_channel(ast_channel_name(user->chan), real_exten, AST_MAX_EXTENSION - 1);
		if (ast_strlen_zero(real_exten) || !strncasecmp(real_exten, "trunk", 5)) {
			ast_copy_string(real_exten, caller_id, sizeof(real_exten));
		}
	}
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB)) {
		is_mcb = 1;
	}
	ast_asprintf(&cmd, "INSERT INTO conference_reservation_log (reservation_id,user_id,user_name,conf_num,id_admin,join_time,leave_time,utc_start_time,action_name,meet_time,device_type,is_random_conf,real_exten,related_exten) "
			"VALUES ('%s','%s','%s','%s','%s','%s','%s','%s','%s','%ld','%s','%s','%s','%s')", user->conference->schedule_id_processing, caller_id, name_tmp?name_tmp:caller_name,
			user->conference->name, ast_test_flag(&user->u_profile, USER_OPT_ADMIN) ? "yes" : "no", strStartTime, strEndTime, user->conference->schedule_utc_start_time, is_mcb ? "BRIDGE" : "JOIN", (long)t_end.tv_sec - (long)t_start.tv_sec, device_type ? device_type : "Other", user->b_profile.is_random_conf ? "yes":"no", real_exten, !ast_strlen_zero(user->related_ext) ? user->related_ext : real_exten);
	conf_report_db_write(cmd);
	safe_ast_free(name_tmp);
	safe_ast_free(cmd);
	return 0;
}

static void conf_user_write_operator_log(struct confbridge_user *user, int op_type, const char *extra_msg)
{
	char *caller_id = NULL;
	char *caller_name = NULL;
	char call_id[AST_CALLID_BUFFER_LENGTH] = {0};
	char user_agent[AST_USER_AGENT_LENGTH] = {0};
	char oplog_detail[AST_OPLOG_DETAIL_LENGTH] = {0};

	if (!user || op_type < 0) {
		return;
	}

	if (!user->conference->is_conference_service) {
		return;
	}

	caller_id = S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous");
	caller_name = S_COR(ast_channel_caller(user->chan)->id.name.valid, ast_channel_caller(user->chan)->id.name.str, caller_id);
	ast_callid_strnprint(call_id, sizeof(call_id), ast_channel_callid(user->chan));
	conf_get_useragent(user->chan, user_agent, sizeof(user_agent));
	snprintf(oplog_detail, sizeof(oplog_detail), "%s&%s", S_OR(extra_msg, ""), user_agent);
	conference_user_write_reservation_log(user, op_type, user_agent);
	ast_operation_log(caller_id, caller_name, call_id, user->conference->feature_id, conf_operation_str[op_type], "%s", S_OR(oplog_detail, ""));
}

static int ha_need_sync_record_count(void)
{
	char conf_sync_count[8] = { 0 };
	int need_sync = 0;

	/*get ConfSyncRecordCount*/
	if (ast_safe_nvram_get(":ConfSyncRecordCount", conf_sync_count, sizeof(conf_sync_count)) == NULL) {
		memset(conf_sync_count, 0, sizeof(conf_sync_count));
	}

	if (conf_sync_count[0] != '\0') {
		/*parsing result*/
		need_sync = atoi(conf_sync_count);
	} else {
		ast_log(LOG_DEBUG, "No need to sync!");
		goto end;
	}
	/*unset ConfSyncRecordCount*/
	ast_safe_nvram_unset(":ConfSyncRecordCount");

	/*commit unset*/
	ast_safe_nvram_commit();

	ast_log(LOG_DEBUG, "need to sync and unset ConfSyncRecordCount!");
end:

	return need_sync;
}

static void conf_operation_log_stasis_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;
	struct ast_json_payload *payload = NULL;
	struct ast_json	*json_object = NULL;
	const char *conference_name = NULL;
	const char *action = NULL;
	const char *extension = NULL;
	const char *originator = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));
	action = ast_json_string_get(ast_json_object_get(json_object, "action"));
	originator = ast_json_string_get(ast_json_object_get(json_object, "originator"));
	extension = ast_json_string_get(ast_json_object_get(json_object, "extension"));
	if (!conference_name || !action || !originator || !extension) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		ast_log(LOG_WARNING, "Couldn't find the conference '%s'", conference_name);
		return;
	}

	ao2_lock(conference);
	user = find_active_user_by_channel(conference, originator);
	if (user) {
		conf_user_write_operator_log(user, conf_operation_action_to_type(action), extension);
	}
	ao2_unlock(conference);
}

static int conf_operation_log_interface(struct ast_channel *chan, const char *data)
{
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf_name);
		AST_APP_ARG(action);
		AST_APP_ARG(act_args);
	);
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;

	if (ast_strlen_zero(data)) {
		return 0;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf_name) || ast_strlen_zero(args.action)) {
		return 0;
	}

	conference = ao2_find(conference_bridges, args.conf_name, OBJ_KEY);
	if (!conference) {
		return 0;
	}

	ao2_lock(conference);
	user = find_active_user_by_channel(conference, ast_channel_name(chan));
	if (!user) {
		ao2_unlock(conference);
		return 0;
	}

	if (!strcasecmp(args.action, "invite")) {
		conf_user_write_operator_log(user, CONF_OP_INVITE, args.act_args);
	} else if (!strcasecmp(args.action, "mcb")) {
		conf_user_write_operator_log(user, CONF_OP_INVITE_MCB, args.act_args);
	}

	ao2_unlock(conference);

	return 0;
}
#endif

const char *conf_get_sound(enum conf_sounds sound, struct bridge_profile_sounds *custom_sounds)
{
	switch (sound) {
	case CONF_SOUND_HAS_JOINED:
		return S_OR(custom_sounds->hasjoin, "conf-hasjoin");
	case CONF_SOUND_HAS_LEFT:
		return S_OR(custom_sounds->hasleft, "conf-hasleft");
	case CONF_SOUND_KICKED:
		return S_OR(custom_sounds->kicked, "conf-kicked");
	case CONF_SOUND_MUTED:
		return S_OR(custom_sounds->muted, "conf-muted");
	case CONF_SOUND_UNMUTED:
		return S_OR(custom_sounds->unmuted, "conf-unmuted");
	case CONF_SOUND_BINAURAL_ON:
		return S_OR(custom_sounds->binauralon, "confbridge-binaural-on");
	case CONF_SOUND_BINAURAL_OFF:
		return S_OR(custom_sounds->binauraloff, "confbridge-binaural-off");
	case CONF_SOUND_ONLY_ONE:
		return S_OR(custom_sounds->onlyone, "conf-onlyone");
	case CONF_SOUND_THERE_ARE:
		return S_OR(custom_sounds->thereare, "conf-thereare");
	case CONF_SOUND_OTHER_IN_PARTY:
		return S_OR(custom_sounds->otherinparty, "conf-otherinparty");
	case CONF_SOUND_PLACE_IN_CONF:
		return S_OR(custom_sounds->placeintoconf, "conf-placeintoconf");
	case CONF_SOUND_WAIT_FOR_LEADER:
		return S_OR(custom_sounds->waitforleader, "conf-waitforleader");
	case CONF_SOUND_LEADER_HAS_LEFT:
		return S_OR(custom_sounds->leaderhasleft, "conf-leaderhasleft");
	case CONF_SOUND_GET_PIN:
		return S_OR(custom_sounds->getpin, "conf-getpin");
	case CONF_SOUND_INVALID_PIN:
		return S_OR(custom_sounds->invalidpin, "conf-invalidpin");
	case CONF_SOUND_ONLY_PERSON:
		return S_OR(custom_sounds->onlyperson, "conf-onlyperson");
	case CONF_SOUND_LOCKED:
		return S_OR(custom_sounds->locked, "conf-locked");
	case CONF_SOUND_LOCKED_NOW:
		return S_OR(custom_sounds->lockednow, "conf-lockednow");
	case CONF_SOUND_UNLOCKED_NOW:
		return S_OR(custom_sounds->unlockednow, "conf-unlockednow");
	case CONF_SOUND_ERROR_MENU:
		return S_OR(custom_sounds->errormenu, "conf-errormenu");
	case CONF_SOUND_JOIN:
		return S_OR(custom_sounds->join, "confbridge-join");
	case CONF_SOUND_LEAVE:
		return S_OR(custom_sounds->leave, "confbridge-leave");
	case CONF_SOUND_PARTICIPANTS_MUTED:
		return S_OR(custom_sounds->participantsmuted, "conf-now-muted");
	case CONF_SOUND_PARTICIPANTS_UNMUTED:
		return S_OR(custom_sounds->participantsunmuted, "conf-now-unmuted");
	case CONF_SOUND_BEGIN:
		return S_OR(custom_sounds->begin, "confbridge-conf-begin");
#ifdef GRANDSTREAM_NETWORKS
	case CONF_SOUND_START_RECORD:
		return S_OR(custom_sounds->startrecord, "conf-now-recording");
	case CONF_SOUND_END_RECORD:
		return S_OR(custom_sounds->endrecord, "dialog-stop-recording");
	case CONF_SOUND_CONTINUE_WAIT:
		return S_OR(custom_sounds->continuewait, "conf-continue-wait");
	case CONF_SOUND_GIVEUP_MODERATOR:
		return S_OR(custom_sounds->giveupmoderator, "conf-giveup-moderator");
	case CONF_SOUND_ACQUIRE_MODERATOR:
		return S_OR(custom_sounds->acquiremoderator, "conf-acquire-moderator");
	case CONF_SOUND_YOU_ARE_MODERATOR:
		return S_OR(custom_sounds->youaremoderator, "conf-your-are-moderator");
	case CONF_SOUND_YOU_ARE_PATICIPANT:
		return S_OR(custom_sounds->youareparticipant, "conf-your-are-participant");
	case CONF_SOUND_SORRY_CANNOT_BE_MODERATOR:
		return S_OR(custom_sounds->sorryyoucantbemoderator, "conf-sorry-you-cant-be-the-moderator");
	case CONF_SOUND_TIMEOUT:
		return S_OR(custom_sounds->conftimeout, "conf-timeout");
	case CONF_SOUND_EXTEND_TIME:
		return S_OR(custom_sounds->extendtime, "conf-extend-time");
	case CONF_SOUND_START_VIDEO:
		return S_OR(custom_sounds->extendtime, "conf-now-videoing");
	case CONF_SOUND_HAS_VIDEOED:
		return S_OR(custom_sounds->extendtime, "conf-has-videoed");
	case CONF_SOUND_END_FIVE_MINUTES:
		return S_OR(custom_sounds->extendtime, "conf-end-five-minutes");
	case CONF_SOUND_VIDEO_OFF:
		return S_OR(custom_sounds->extendtime, "conf-video-off");
	case CONF_SOUND_RECORD_OFF:
		return S_OR(custom_sounds->extendtime, "conf-record-off");
	case CONF_SOUND_START_PAGING_VIDEO:
		return S_OR(custom_sounds->pagingrecord, "paging-video-recording");
	case CONF_SOUND_START_FOCUS_VIDEO:
		return S_OR(custom_sounds->extendtime, "conf-now-focus-videoing");

#endif
	}

	return "";
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu, 2022.5.12.
 * If record enable by UCM config, host can't enable/disable record
 */
static unsigned int get_host_record_authority(struct confbridge_conference *conference)
{
	if (!conference) {
		return 0;
	}

	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE) ||
		ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE) ||
		ast_test_flag(&conference->b_profile, BRIDGE_OPT_SLIDES_RECORD_CONFERENCE) ||
		ast_test_flag(&conference->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE)) {
		return 0;
	}

	return 1;
}

static unsigned int get_conf_config(struct confbridge_conference *conference)
{
	unsigned int res = 0;

	if (!conference) {
		return 0;
	}

	res = conference->b_profile.conf_admin_pin + (get_host_record_authority(conference) << 1);

	return res;
}

static int is_record_merge_enable(void)
{
	const char *merge_record = pbx_builtin_getvar_helper(NULL, "MERGE_AUDIO_RECORDING_FILES");
	return (merge_record && !strcmp(merge_record, "yes"));
}

static int is_conf_record_paused(struct confbridge_conference *conference)
{
	if (!conference->record_chan) { return 0; }
	
	const char *rec_paused = pbx_builtin_getvar_helper(conference->record_chan, "REC_PAUSED");
	return (rec_paused && !strcmp(rec_paused, "yes"));
}
#endif

static void send_conf_stasis(struct confbridge_conference *conference, struct ast_channel *chan,
	struct stasis_message_type *type, struct ast_json *extras, int channel_topic)
{
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
#ifdef GRANDSTREAM_NETWORKS
	char conf_config_buf[CONF_CONFIG_MAX_LEN] = {0};

	sprintf(conf_config_buf,"%X", get_conf_config(conference));
#endif
#ifndef GRANDSTREAM_NETWORKS
	json_object = ast_json_pack("{s: s}",
		"conference", conference->name);
#else
	json_object = ast_json_pack("{s: s, s: s, s: s}",
		"conference", conference->name,
		"conference_type", ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU) ? "video" : "audio",
		"conference_config", conf_config_buf);
#endif
	if (!json_object) {
		return;
	}

	if (extras) {
		ast_json_object_update(json_object, extras);
	}

	ast_bridge_lock(conference->bridge);
	msg = ast_bridge_blob_create(type,
		conference->bridge,
		chan,
		json_object);
	ast_bridge_unlock(conference->bridge);
	if (!msg) {
		return;
	}

	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_ENABLE_EVENTS)) {
		conf_send_event_to_participants(conference, chan, msg);
	}

	if (channel_topic) {
		stasis_publish(ast_channel_topic(chan), msg);
	} else {
		stasis_publish(ast_bridge_topic(conference->bridge), msg);
	}
}

static void send_conf_start_event(struct confbridge_conference *conference)
{
#ifdef GRANDSTREAM_NETWORKS
	struct ast_tm timeresult;
	char family[64] = {0};
	char conf_start_time[AST_MAX_START_TIME] = {0};
	struct timeval now_time = ast_tvnow();
	ast_localtime(&now_time, &timeresult, NULL);
	ast_strftime(conf_start_time, sizeof(conf_start_time), "%Y-%m-%d %H:%M:%S", &timeresult);
	if (conference && !ast_strlen_zero(conf_start_time)) {
		snprintf(family, sizeof(family), "CONF_STATUS/%s", conference->name);
		ast_db_put(family, "starttime", conf_start_time);
	}
#endif
	send_conf_stasis(conference, NULL, confbridge_start_type(), NULL, 0);
#ifdef GRANDSTREAM_NETWORKS
#ifdef HAVE_UNIXODBC
	if (conference->is_conference_service /*&& !ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)*/) {
		ast_mutex_lock(&record_count_lock);
		conference->report_record.id = ++count;
		conf_report_update_call_count();
		ast_mutex_unlock(&record_count_lock);
		conference->report_record.start = ast_tvnow();
		conf_report_db_write_conference(conference);
	}
#endif
#endif
}

static void send_conf_end_event(struct confbridge_conference *conference)
{
#ifdef GRANDSTREAM_NETWORKS
	char family[64] = {0};
	if (conference) {
		snprintf(family, sizeof(family), "CONF_STATUS/%s", conference->name);
		ast_db_deltree(family, "starttime");
	}
#ifdef HAVE_UNIXODBC
	if (conference->is_conference_service /*&& !ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)*/) {
		conference->report_record.end = ast_tvnow();
		conf_report_db_write_conference(conference);
	}
#endif // HAVE_UNIXODBC
#endif // GRANDSTREAM_NETWORKS

#ifdef GRANDSTREAM_NETWORKS
	struct ast_json *json_object = NULL;
	json_object = ast_json_pack("{s: s, s: s}",
		"FeatureId", S_OR(conference->feature_id, ""),
		"ScheduleEnd", conference->schedule_hangup ? "yes" : "no");
	send_conf_stasis(conference, NULL, confbridge_end_type(), json_object, 0);
#else 
	send_conf_stasis(conference, NULL, confbridge_end_type(), NULL, 0);
#endif // GRANDSTREAM_NETWORKS

#ifdef GRANDSTREAM_NETWORKS
	if (json_object) {
		ast_json_unref(json_object);
	}
#endif // GRANDSTREAM_NETWORKS
}

static void send_join_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object;

#ifndef GRANDSTREAM_NETWORKS
	json_object = ast_json_pack("{s: b, s: b}",
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"muted", user->muted);
#else
	struct timeval now_time = ast_tvnow();
	struct conf_notify_info info;
	const char *from_call_flip = NULL;
	char record_type[MAX_RECORD_TYPE] = { 0 };
	char sca_shareline[AST_MAX_EXTENSION] = {0};

	mark_chan_has_joined(user->chan);

	if (ast_strlen_zero(conference->start_time)) {
		struct timeval now = ast_tvnow();
		struct ast_tm tm;
		if (ast_localtime(&now, &tm, NULL)) {
			ast_strftime(conference->start_time, sizeof(conference->start_time), "%Y-%m-%dT%H:%M:%S%z", &tm);
		}
	}
	if (!conference->conference_start) {
		conference->conference_start = time(NULL);
	}
	/*Check and update the caller name for contact group*/
	conf_contact_group_update_caller_name(user);

	/*Get sca share line number*/
	conf_get_sca_shareline(user->chan, sca_shareline, sizeof(sca_shareline));

	/*Check and update the caller name for trunk call*/
	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		update_caller_name_for_trunk(user->chan);
	}

	from_call_flip = pbx_builtin_getvar_helper(user->chan, "FROM_CALL_FLIP");

	if (conference->record_chan) {
		if (is_conf_record_paused(conference)) {
			strncpy(record_type, "none", sizeof(record_type) - 1);
		} else {
			strncpy(record_type, "audio", sizeof(record_type) - 1);
		}
	} else if (conference->video_is_recording) {
		strncpy(record_type, "video", sizeof(record_type) - 1);
	} else if (conference->focus_is_recording) {
		strncpy(record_type, "focus", sizeof(record_type) - 1);
	} else {
		strncpy(record_type, "none", sizeof(record_type) - 1);
	}

	json_object = ast_json_pack("{s: i, s: i, s: s, s: s, s: s, s: b, s: i, s: i, s: i, s: i, s: s, s: s, s: s, s: s, s: s, s: s, s: s, s: s, s: s, s: i, s: s, s: s, s: i, s: s, s: s, s: s, s: s}",
		"user_count", conference->activeusers,
		"user_number", user->user_no ,
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		/*!begin Added by pyzhong@grandstream.cn for bug[68670], 2017-9-23*/
		"caller_id_name", S_COR(ast_channel_caller(user->chan)->id.name.valid, ast_channel_caller(user->chan)->id.name.str, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous")),
		/*!end Added by pyzhong@grandstream.cn for bug[68670], 2017-9-23*/
		"channel_name", ast_channel_name(user->chan),
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"muted", user->muted,
		"mic_muted", user->mic_muted,
		"can_invite", ast_test_flag(&user->u_profile, USER_OPT_CAN_INVITE) ? 1 : 0,
		"invited", ast_test_flag(&user->u_profile, USER_OPT_INVITEDUSER) ? 1 : 0,
		"mcb_type", user->u_profile.mcb_type,
		"mcb_remote_conf_id", user->u_profile.mcb_remote_conf_id,
		"conf_starttime", conference->start_time,
		"media", user->video ? "video" : "audio",
		"feature_id", S_OR(conference->feature_id, ""),
		/*Begin Added by gangwang@grandstream.cn for Wave App*/
		"subject", S_OR(conference->schedule_subject, ""),
		/*End Added by gangwang@grandstream.cn for Wave App*/
		"record_type", record_type,
		"related_info", S_OR(user->related_ext, ""),
		"schedule_id", S_OR(conference->schedule_id_processing, ""),
		"record_start_timestamp_utc", conference->record_start_utc_timestamp,
		"im_service_prefix", S_OR(conference->im_service_prefix, ""),
		"call_flip", from_call_flip ? "yes" : "no",
		"isfocus", user->isfocus,
		"allow_user_unmute", conference->allow_user_unmute ? "yes" : "no",
		"schedule_time", S_OR(conference->schedule_utc_start_time, ""),
		"sca_shareline", S_OR(sca_shareline, ""),
		"creator", S_OR(conference->conf_creator, "")
	);
#endif

	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_join_type(), json_object, 0);
#ifdef GRANDSTREAM_NETWORKS
	if (!ast_strlen_zero(conference->schedule_id_processing)){
		pbx_builtin_setvar_helper(user->chan, "RESOURCE_SCHEDULE_ID", conference->schedule_id_processing);
	}
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_join_type(), obj, 0);
	}
	if (!globalEnabledCEI && conference->is_conference_service && !ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
			info.role = CONF_USER_ROLE_ADMIN;
		} else {
			info.role = CONF_USER_ROLE_PARTICIPANT;
		}
		ast_indicate_data(user->chan, AST_CONTROL_NOTIFY_CEI_STATUS, &info, sizeof(struct conf_notify_info));
	}

	if (conference->is_conference_service) {
		if (ast_tvzero(conference->report_record.start)) {
			conference->report_record.start = now_time;
		}
		user->report_record.start = now_time;
		user->report_record.number = ast_strdup(ast_json_string_get(ast_json_object_get(json_object, "caller_id_number")));
		user->report_record.name = ast_strdup(ast_json_string_get(ast_json_object_get(json_object, "caller_id_name")));
		user->report_record.group_name = ast_strdup(pbx_builtin_getvar_helper(user->chan, "CONTACT_GROUP"));
	}
	user->jointime = now_time;

	conference_write_user_account(user);
	conf_user_write_operator_log(user, CONF_OP_JOIN, NULL);
#endif
	ast_json_unref(json_object);
}

static void send_leave_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object;

#ifndef GRANDSTREAM_NETWORKS
	json_object = ast_json_pack("{s: b}",
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN)
	);
#else
	json_object = ast_json_pack("{s: b, s: i, s: i, s: s, s: s}",
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"user_count", conference->activeusers,
		"user_number", user->user_no ,
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"channel_name", ast_channel_name(user->chan)
	);
#endif
	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_leave_type(), json_object, 0);
#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_leave_type(), obj, 0);
	}
#ifdef HAVE_UNIXODBC
	if (conference->is_conference_service /*&& !ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)*/) {
		user->report_record.end = ast_tvnow();
		conf_report_db_write_user(user);
	}
#endif

	conf_user_write_operator_log(user, CONF_OP_LEAVE, NULL);
#endif
	ast_json_unref(json_object);
}

static void send_start_record_event(struct confbridge_conference *conference)
{
	send_conf_stasis(conference, NULL, confbridge_start_record_type(), NULL, 0);
}

static void send_stop_record_event(struct confbridge_conference *conference)
{
	send_conf_stasis(conference, NULL, confbridge_stop_record_type(), NULL, 0);
}

#ifdef GRANDSTREAM_NETWORKS
static void send_mute_event(struct confbridge_user *user, struct confbridge_conference *conference, int wait, int is_mic)
{
	struct ast_json *json_object;

	json_object = ast_json_pack("{s: i, s: b, s: i, s: i, s: s, s: s, s: i, s: i}",
		"user_number", user->user_no ,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"muted", user->muted,
		"mic_muted", user->mic_muted,
		"talking_status", user->talking ? "on" : "off",
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"is_mic", is_mic,
		"wait", wait
	);

	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_mute_type(), json_object, 1);

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_toggle_mute_type(), obj, 1);
	}

	ast_json_unref(json_object);
}

static void send_unmute_event(struct confbridge_user *user, struct confbridge_conference *conference, int wait, int is_mic)
{
	struct ast_json *json_object;

	json_object = ast_json_pack("{s: i, s: b, s: i, s: i, s: s, s: s, s: i, s: i}",
		"user_number", user->user_no ,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"muted", user->muted,
		"mic_muted", user->mic_muted,
		"talking_status", user->talking ? "on" : "off",
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"is_mic", is_mic,
		"wait", wait
	);

	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_unmute_type(), json_object, 1);

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_toggle_mute_type(), obj, 1);
	}

	ast_json_unref(json_object);
}
#else
static void send_mute_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object;

	json_object = ast_json_pack("{s: b}",
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN)
	);

	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_mute_type(), json_object, 1);
	ast_json_unref(json_object);
}

static void send_unmute_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object;

	json_object = ast_json_pack("{s: b}",
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN)
	);

	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_unmute_type(), json_object, 1);
	ast_json_unref(json_object);
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static void send_muteall_event(struct confbridge_user *user, struct confbridge_conference *conference, int wait, int count_num)
{
	struct ast_json *json_object;

	// Sanity check
	if (conference == NULL) {
		return;
	}

	json_object = ast_json_pack("{s: i, s: b, s: i, s: i, s: s, s: s, s: i, s: s}",
		"user_number", user->user_no ,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"muted", user->muted,
		"mic_muted", user->mic_muted,
		"talking_status", user->talking ? "on" : "off",
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"wait", wait,
		"object", "all"
	);

	if (!json_object) {
		return;
	}

	if (!count_num) {
		send_conf_stasis(conference, NULL, confbridge_muteall_type(), json_object, 0);
	}
	/*message for cei*/
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_toggle_mute_type(), obj, 1);
	}

	// Clean up
	ast_json_unref(json_object);
}

static void send_unmuteall_event(struct confbridge_user *user, struct confbridge_conference *conference, int wait, int count_num)
{
	struct ast_json *json_object;

	// Sanity check
	if (conference == NULL) {
		return;
	}

	json_object = ast_json_pack("{s: i, s: b, s: i, s: i, s: s, s: s, s: i, s: s}",
		"user_number", user->user_no ,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"muted", user->muted,
		"mic_muted", user->mic_muted,
		"talking_status", user->talking ? "on" : "off",
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"wait", wait,
		"object", "all"
	);

	if (!json_object) {
		return;
	}

	if (!count_num) {
		send_conf_stasis(conference, NULL, confbridge_unmuteall_type(), json_object, 0);
	}
	/*message for cei*/
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_toggle_mute_type(), obj, 1);
	}

	// Clean up
	ast_json_unref(json_object);
}

static void send_lock_status_event(struct confbridge_conference *conference, int lock)
{
	struct ast_json *json_object;

	// Sanity check
	if (conference == NULL) {
		return;
	}

	json_object = ast_json_pack("{s: s}",
		"status", lock ? "locked" : "unlocked"
	);
	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, NULL, confbridge_lock_status_type(), json_object, 0);
#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, NULL, cei_lock_status_type(), obj, 0);
	}
#endif
	// Clean up
	ast_json_unref(json_object);
}

/*
static void send_moderator_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object = NULL;

	json_object = ast_json_pack("{s: i, s: b, s: s, s: s}",
		"user_number", user->user_no ,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"channel_name", ast_channel_name(user->chan)
	);

	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, NULL, confbridge_moderator_type(), json_object, 0);

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, NULL, cei_moderator_type(), obj, 0);
	}

	ast_json_unref(json_object);
}
*/

static void send_media_update_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object = NULL;

	json_object = ast_json_pack("{s: i, s: b, s: s}",
		"user_number", user->user_no,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"media", user->video ? "video" : "audio"
	);

	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, user->chan, confbridge_media_update_type(), json_object, 0);

	ast_json_unref(json_object);
}

static void send_role_update_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object = NULL;

	json_object = ast_json_pack("{s: i, s: b, s: s, s: s, s: i, s: i, s: i, s: i}",
		"user_number", user->user_no ,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"),
		"channel_name", ast_channel_name(user->chan),
		"wait", 0,
		"muted", user->muted,
		"mic_muted", user->mic_muted,
		"can_invite", ast_test_flag(&user->u_profile, USER_OPT_CAN_INVITE) ? 1 : 0
	);

	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, user->chan, confbridge_role_update_type(), json_object, 0);

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_role_update_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

void confbridge_user_media_update(const char *channel_name, const char *conf_number, int is_video)
{
	struct confbridge_conference *conference = NULL;
	struct confbridge_user *user = NULL;

	conference = conf_find_bridge(conf_number);
	if (!conference) {
		return;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (!strcasecmp(ast_channel_name(user->chan), channel_name)) {
			user->video = is_video ? 1 : 0;
			send_media_update_event(user, conference);
			break;
		}
	}
	ao2_unlock(conference);

	ao2_cleanup(conference);
}

static void send_conf_mute_status_event(struct confbridge_conference *conference)
{
	struct ast_json *json_object = NULL;

	// Sanity check
	if (conference == NULL) {
		return;
	}

	json_object = ast_json_pack("{s: i}",
		"muted", conference->muted
	);

	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, NULL, cei_mute_status_type(), json_object, 0);

	// Clean up
	ast_json_unref(json_object);
}

static void send_name_update_event(struct confbridge_user *user, struct confbridge_conference *conference, const char *name)
{
	struct ast_json *json_object = NULL;

	json_object = ast_json_pack("{s: i, s: b, s: s, s: s}",
		"user_number", user->user_no,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"channel_name", ast_channel_name(user->chan),
		"name", name
	);

	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, user->chan, confbridge_name_update_type(), json_object, 0);

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);

	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_name_update_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

static void send_record_status_event(struct confbridge_conference *conference, const char *type, int onoff)
{
	struct ast_json *json_object;

	if (!conference || !type) {
		return;
	}

	json_object = ast_json_pack("{s: s, s: s, s: s, s: i}",
		"type", type,
		"status", onoff ? "on":"off",
		"conference", conference->name,
		"start_timestamp_utc", conference->record_start_utc_timestamp
	);
	if (!json_object) {
		return;
	}

	send_conf_stasis(conference, NULL, cei_toggle_record_type(), json_object, 0);

	// Clean up
	ast_json_unref(json_object);
}

static void send_focus_status_event(struct confbridge_user *user, struct confbridge_conference *conference, int isfocus)
{
	struct ast_json *json_object = NULL;
	json_object = ast_json_pack("{s: i, s: b, s: s, s: i, s: s}",
		"user_number", user->user_no,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"channel_name", ast_channel_name(user->chan),
		"isfocus", isfocus,
		"layout", conference->conf_layout
	);
	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, user->chan, confbridge_focus_status_type(), json_object, 0);
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);
	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_focus_status_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

static void send_layout_status_event(struct confbridge_conference *conference)
{
	struct ast_json *json_object = NULL;
	json_object = ast_json_pack("{s: s}",
		"layout", conference->conf_layout
	);
	if (!json_object) {
		return;
	}
	send_conf_stasis(conference, NULL, confbridge_layout_status_type(), json_object, 0);
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);
	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, NULL, cei_layout_status_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

static void send_hand_up_status_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *json_object = NULL;
	json_object = ast_json_pack("{s: i, s: b, s: s, s: i}",
		"user_number", user->user_no,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"channel_name", ast_channel_name(user->chan),
		"hand_up_time", user->hand_up_time
	);
	if (!json_object) {
		return;
	}

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);
	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_gesture_status_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

static void send_conference_act_event(struct confbridge_conference *conference, int event_type)
{
	struct ast_json *json_object = NULL;
	json_object = ast_json_pack("{s: i}",
		"act_event_type", event_type
	);
	if (!json_object) {
		return;
	}

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);
	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, NULL, cei_act_event_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

static unsigned int chan_need_update(struct ast_channel *chan)
{
	if (!chan) {
		return 0;
	}

	if (ast_channel_get_terminate_type(chan) == AST_TERMINAL_TYPE_SIP
			|| ((ast_channel_get_terminate_type(chan) == AST_TERMINAL_TYPE_GSWAVE_MOBILE) &&
				!ast_channel_get_mobile_bundle_enabled(chan))
			|| ast_channel_get_terminate_type(chan) == AST_TERMINAL_TYPE_ICE_TRUNK) {
		return 1;
	}

	return 0;
}

static int chan_update_msid_for_focus(struct confbridge_conference *conference,
	struct ast_channel *msid_chan, struct ast_channel *update_chan)
{
	RAII_VAR(char *, msid, NULL, ast_free);

	if (!conference) {
		ast_log(LOG_ERROR, "chan_update_msid_for_focus conference is NULL\n");
		return -1;
	}

	if (!chan_need_update(update_chan)) {
		return 0;
	}

	ao2_lock(conference);

	if (msid_chan) {
		msid = ast_channel_get_msid(msid_chan, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		if (!msid) {
			msid = ast_strdup("");
		}
	} else {
		msid = ast_strdup("");
	}

	ao2_unlock(conference);

	ast_indicate_data(update_chan, AST_CONTROL_CONF_HOST_CHANGED, msid, strlen(msid) + 1);

	return 0;
}

static unsigned int has_valid_main_video_stream(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) != AST_STREAM_STATE_REMOVED
		&& ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO
		&& ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN
		&& strncmp(ast_stream_get_name(stream), SOFTBRIDGE_VIDEO_DEST_PREFIX,
			SOFTBRIDGE_VIDEO_DEST_LEN)) {
		return 1;
	}

	return 0;
}

static int notify_all_chan_update_msid_for_focus(struct confbridge_conference *conference,
	struct confbridge_user *user)
{
	unsigned int i = 0;
	RAII_VAR(char *, msid, NULL, ast_free);
	struct confbridge_user *cur_user = NULL;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;

	if (!conference) {
		ast_log(LOG_ERROR, "notify_all_chan_update_msid_for_focus conference is NULL\n");
		return -1;
	}

	if (user) {
		msid = ast_channel_get_msid(user->chan, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		if (!msid) {
			msid = ast_strdup("");
		}
	} else {
		msid = ast_strdup("");
	}

	ast_log(LOG_DEBUG, "Update chan msid for focus, conference:%s, msid:%s\n", conference->name, msid);

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if (!chan_need_update(cur_user->chan)) {
			continue;
		}

		topology = ast_channel_get_stream_topology(cur_user->chan);
		if (!topology) {
			continue;
		}

		for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
			stream = ast_stream_topology_get_stream(topology, i);
			if (has_valid_main_video_stream(stream)){
				ast_channel_lock(cur_user->chan);
				if (ast_channel_tech(cur_user->chan)->indicate) {
					ast_channel_tech(cur_user->chan)->indicate(cur_user->chan, AST_CONTROL_CONF_HOST_CHANGED, msid, strlen(msid) + 1);
				}
				ast_channel_unlock(cur_user->chan);
				break;
			}
		}
	}
	ao2_unlock(conference);

	return 0;
}

static void conf_update_msid_for_focus(struct confbridge_conference *conference,
	struct confbridge_user *user, int isfocus)
{
	unsigned int updated = 0;
	struct confbridge_user *cur_user = NULL;

	if (!conference) {
		ast_log(LOG_ERROR, "conf_update_msid_for_focus conference is NULL\n");
		return;
	}

	ast_log(LOG_DEBUG, "Update msid for focus, conference:%s, isfocus:%d\n", conference->name, isfocus);

	if (isfocus) {
		notify_all_chan_update_msid_for_focus(conference, user);
	} else {
		AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
			if (ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN)) {
				notify_all_chan_update_msid_for_focus(conference, cur_user);
				updated = 1;
				break;
			}
		}

		if (!updated) {
			ast_log(LOG_DEBUG, "Disable focus and has no host, set msid NULL for conference:%s\n", conference->name);
			notify_all_chan_update_msid_for_focus(conference, NULL);
		}
	}

	return;
}

static void send_user_permission_event(struct confbridge_conference *conference, const char *permission_type)
{
	if (!permission_type || !conference) {
		return;
	}
	struct ast_json *json_object = NULL;
	json_object = ast_json_pack("{s: s, s: s}",
		"permission_type", permission_type,
		"allow", conference->allow_user_unmute ? "yes" : "no"
	);
	if (!json_object) {
		return;
	}
	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(json_object), ast_json_unref);
	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, NULL, cei_user_permission_type(), obj, 0);
	}
	ast_json_unref(json_object);
}

static void handle_focus_user_leave(struct confbridge_user *user, struct confbridge_conference *conference)
{
	if (!conference || !user) {
		return;
	}
	ao2_lock(conference);
	if (user->isfocus) {
		memset(conference->conf_layout, 0, sizeof(conference->conf_layout));
		ast_copy_string(conference->conf_layout, DEFAULT_CONF_LAYOUT, sizeof(conference->conf_layout));
		ast_log(LOG_DEBUG, "reset conf layout!");
		send_layout_status_event(conference);
		conf_update_msid_for_focus(user->conference, user, 0);

		if (conference->focus_chan) {
			ast_channel_unref(conference->focus_chan);
			conference->focus_chan = NULL;
		}
	}
	ao2_unlock(conference);
}

static void check_reserve_host(struct confbridge_user *user, struct confbridge_conference *conference)
{
	int host_exist = 0;
	struct confbridge_user *user_list = NULL;

	if (!conference || !user || !user->chan) {
		return;
	}
	if (conference->conf_end) {
		ast_log(LOG_DEBUG, "conf end!");
		return;
	}
	const char *want_be_host = NULL;
	want_be_host = pbx_builtin_getvar_helper(user->chan, "WANT_BE_HOST");

	if (!ast_strlen_zero(want_be_host) && !strcasecmp(want_be_host, "yes")) {
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user_list, list) {
			if (ast_test_flag(&user_list->u_profile, USER_OPT_ADMIN)) {
				host_exist = 1;
				break;
			}
		}
		if (!host_exist) {
			if (user->term_type != AST_TERMINAL_TYPE_WEBRTC
				&& user->term_type != AST_TERMINAL_TYPE_GVC
				&& user->term_type != AST_TERMINAL_TYPE_SIP
				&& user->term_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE
				&& user->term_type != AST_TERMINAL_TYPE_ICE_TRUNK) {
				ast_log(LOG_DEBUG, "The terminal type can't be transferred and become the host!");
				ao2_unlock(conference);
				return;
			}
			if (user->muted) {
				user->muted = 0;
				conf_update_user_mute(user);
				send_unmute_event(user, conference, 1, 0);
			}
			ast_set_flag(&user->u_profile, USER_OPT_ADMIN);
			ast_set_flag(&user->u_profile, USER_OPT_MARKEDUSER);
			ast_set_flag(&user->u_profile, USER_OPT_CAN_INVITE);
			conference->markedusers++;
			reload_menu_to_user(user, MENU_SFU_ADMIN);
			send_role_update_event(user, conference);

			if (conference->state->state_to_marked) {
				conference->state->state_to_marked(user);
			}
			if (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
				cache_conf_update_moderator(conference->name,
											S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, ""),
											user->chan);
				/*send Notify to new admin*/
				if (conference->conf_will_timeout && conference->extend_time_status == EXTEND_PENDING) {
					int conf_timeleft = ast_sched_when(conf_sched, conference->conf_sched_id);
					ast_log(LOG_DEBUG, "send Notify to new admin!");
					conf_timeout_notify_build_send(user->chan, conference->name, 1, conference->extend_time_status, conf_timeleft+(conference->sch_time * (CONF_PEER_TIMEOUT/1000)));
				}
			}
			if (!conference->focus_chan) {
				sfu_topologies_on_moderator_change(conference, user);
			} else {
				ast_log(LOG_DEBUG, "already has focus chan, don't upate host msid\n");
			}
			ast_log(LOG_DEBUG, "%s become the host!", ast_channel_name(user->chan));
		} else {
			if (user_list && user_list->chan) {
				ast_log(LOG_DEBUG, "host exist, chan name is %s!", ast_channel_name(user_list->chan));
			}
		}
		ao2_unlock(conference);
	}
}

static void handle_auto_transfer_moderator(struct confbridge_user *left_user, struct confbridge_conference *conference)
{
	int host_from_trunk = 0;
	struct confbridge_user *user = NULL;
	struct confbridge_user *admin_user = NULL;
	char host_exten[AST_MAX_EXTENSION] = {0};
	char host_trunk_exten[AST_MAX_EXTENSION] = {0};
	char host_exten_find[AST_MAX_EXTENSION] = {0};
	ast_log(LOG_DEBUG, "handle auto transfer moderator!");

	if (!conference || !left_user || !left_user->chan) {
		return;
	}
	if (!ast_test_flag(&left_user->u_profile, USER_OPT_ADMIN)){
		ast_log(LOG_DEBUG, "left_user no admin!");
		return;
	}

	if (conference->conf_end) {
		ast_log(LOG_DEBUG, "conf end!");
		return;
	}

	/*Modify for auto transfer moderator when switch to conf mode*/
	if (ast_channel_hangupcause(left_user->chan) == AST_CAUSE_SWITCH_TO_CONF_MODE) {
		const char *switch_host = NULL;
		switch_host = pbx_builtin_getvar_helper(left_user->chan, "SWITCH_HOST");
		snprintf(host_exten_find, sizeof(host_exten_find) -1, "%s", switch_host ? switch_host : "UNKNOW");
		ast_log(LOG_DEBUG, "get switch to conf mode new host %s!", host_exten_find);
	} else {
		conf_get_exten_by_channel(ast_channel_name(left_user->chan), host_exten, AST_MAX_EXTENSION - 1);
		if (!strncasecmp(host_exten, "trunk", 5)) {
			host_from_trunk = 1;
			ast_copy_string(host_trunk_exten, host_exten, sizeof(host_trunk_exten));
			ast_log(LOG_DEBUG, "host_trunk_exten %s!", host_trunk_exten);
		}
		if (ast_strlen_zero(host_exten) || host_from_trunk) {
			ast_copy_string(host_exten, S_COR(ast_channel_caller(left_user->chan)->id.number.valid, ast_channel_caller(left_user->chan)->id.number.str, ""), sizeof(host_exten));
		}

		snprintf(host_exten_find, sizeof(host_exten_find) - 1, "/%s-", host_exten);
	}
	ast_log(LOG_DEBUG, "host exten find %s!", host_exten_find);

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user->chan && user->chan != left_user->chan && ast_channel_name(user->chan) && strcasestr(ast_channel_name(user->chan), host_exten_find)) {
			ast_log(LOG_DEBUG, "find the same exten, chan is %s!", ast_channel_name(user->chan));
			admin_user = user;
			break;
		} else if (user->chan && user->chan != left_user->chan && ast_channel_name(user->chan) && host_from_trunk && strcasestr(ast_channel_name(user->chan), host_trunk_exten)) {
			char user_exten[AST_MAX_EXTENSION] = {0};
			ast_copy_string(user_exten, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, ""), sizeof(user_exten));
			if (strcasestr(host_exten_find, user_exten)) {
				ast_log(LOG_DEBUG, "find the same exten from the same trunk, chan is %s!", ast_channel_name(user->chan));
				admin_user = user;
				break;
			}
		}
	}

	if (admin_user) {
		if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
			ast_log(LOG_DEBUG, "already the host!");
			ao2_unlock(conference);
			return;
		}
		if (admin_user->term_type != AST_TERMINAL_TYPE_WEBRTC
			&& admin_user->term_type != AST_TERMINAL_TYPE_GVC
			&& admin_user->term_type != AST_TERMINAL_TYPE_SIP
			&& admin_user->term_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE
			&& admin_user->term_type != AST_TERMINAL_TYPE_ICE_TRUNK) {
			ast_log(LOG_DEBUG, "The terminal type can't be transferred and become the host!");
			ao2_unlock(conference);
			return;
		}
		if (left_user->chan) {
			pbx_builtin_setvar_helper(left_user->chan, "AUTO_TRANSFER_HOST", "yes");
		}
		if (admin_user->muted) {
			admin_user->muted = 0;
			conf_update_user_mute(admin_user);
			send_unmute_event(admin_user, conference, 1, 0);
		}
		ast_set_flag(&admin_user->u_profile, USER_OPT_ADMIN);
		ast_set_flag(&admin_user->u_profile, USER_OPT_MARKEDUSER);
		ast_set_flag(&admin_user->u_profile, USER_OPT_CAN_INVITE);
		conference->markedusers++;
		reload_menu_to_user(admin_user, MENU_SFU_ADMIN);
		send_role_update_event(admin_user, conference);

		if (conference->state->state_to_marked) {
			conference->state->state_to_marked(admin_user);
		}
		if (admin_user->term_type == AST_TERMINAL_TYPE_WEBRTC || admin_user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			cache_conf_update_moderator(conference->name,
										S_COR(ast_channel_caller(admin_user->chan)->id.number.valid, ast_channel_caller(admin_user->chan)->id.number.str, ""),
										admin_user->chan);
			/*send Notify to new admin*/
			if (conference->conf_will_timeout && conference->extend_time_status == EXTEND_PENDING) {
				int conf_timeleft = ast_sched_when(conf_sched, conference->conf_sched_id);
				ast_log(LOG_DEBUG, "send Notify to new admin!");
				conf_timeout_notify_build_send(admin_user->chan, conference->name, 1, conference->extend_time_status, conf_timeleft+(conference->sch_time * (CONF_PEER_TIMEOUT/1000)));
			}
		}
		if (!conference->focus_chan) {
			sfu_topologies_on_moderator_change(conference, admin_user);
		} else {
			ast_log(LOG_DEBUG, "__transfer_moderator already has focus chan, don't upate host msid\n");
		}
		ast_log(LOG_DEBUG, "auto transfer moderator success!now admin user is %s", ast_channel_name(admin_user->chan));
	}
	ao2_unlock(conference);

}

static void conf_get_rec_time(char *buf, int buflen, char *format)
{

	struct timeval when;
	struct ast_tm tm;
	char *epoch = NULL;

	if (!format || buflen <= 0) {
		ast_log(LOG_ERROR, "Parameter error!");
	}
	ast_asprintf(&epoch, "%d", (int) time(NULL));
	if (epoch) {
		ast_get_timeval(epoch, &when, ast_tvnow(), NULL);
		ast_localtime(&when, &tm, NULL);
		if (ast_strftime(buf, buflen, format, &tm) <= 0) {
			ast_log(LOG_ERROR, "function strftime() error!\n");
		} else {
			ast_log(LOG_DEBUG, "get rec time %s\n", buf);
		}
	} else {
		ast_log(LOG_ERROR, "function ast_asprintf() error!\n");
	}
	AST_SAFE_FREE(epoch);
}

#endif

#ifdef GRANDSTREAM_NETWORKS
static void set_rec_filename(struct confbridge_conference *conference, struct ast_str **filename, struct ast_str **uuid, int is_new)
#else
static void set_rec_filename(struct confbridge_conference *conference, struct ast_str **filename, int is_new)
#endif
{
	char *rec_file = conference->b_profile.rec_file;
#ifdef GRANDSTREAM_NETWORKS
	char destdir[PATH_MAX];
	char tmp[64];
#endif
	char *ext;
	time_t now;

	if (ast_str_strlen(*filename)
		&& ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_APPEND)
		&& !is_new) {
		return;
	}

	time(&now);
#ifdef GRANDSTREAM_NETWORKS
	snprintf(destdir, sizeof(destdir), "%s/meetme", ast_config_AST_SPOOL_DIR);
	if (ast_mkdir(destdir, 0777) != 0) {
		ast_log(LOG_WARNING, "mkdir '%s' failed: %s\n", destdir, strerror(errno));
		return;
	}
#endif

	ast_str_reset(*filename);
	if (ast_strlen_zero(rec_file)) {
#ifdef GRANDSTREAM_NETWORKS
		char rec_time_yy_mm[64] = {0};
		char rec_time_timestamp[128] = {0};
		conf_get_rec_time(rec_time_yy_mm, sizeof(rec_time_yy_mm), "%Y-%m");
		conf_get_rec_time(rec_time_timestamp, sizeof(rec_time_timestamp), "%s");

		ast_str_set(filename, 0, "%s/%s/meeting-%s-%s.wav", destdir, rec_time_yy_mm,
			conference->name, rec_time_timestamp);
		ast_str_append(&conference->record_files, 0, "%s/meeting-%s-%s.wav@",
			rec_time_yy_mm, conference->name, rec_time_timestamp);
#else
		ast_str_set(filename, 0, "confbridge-%s-%u.wav", conference->name,
			(unsigned int) now);
#endif
	} else if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP)) {
		/* insert time before file extension */
		ext = strrchr(rec_file, '.');
		if (ext) {
			ast_str_set_substr(filename, 0, rec_file, ext - rec_file);
			ast_str_append(filename, 0, "-%u%s", (unsigned int) now, ext);
		} else {
			ast_str_set(filename, 0, "%s-%u", rec_file, (unsigned int) now);
		}
	} else {
		ast_str_set(filename, 0, "%s", rec_file);
	}

#ifdef GRANDSTREAM_NETWORKS
	file_crypt_new_uuid(tmp, sizeof(tmp));
	ast_log(LOG_DEBUG, "[set_rec_filename] uuid previous:%s new:%s", ast_str_buffer(*uuid), tmp);
	ast_str_append(uuid, 0, "%s@", tmp);
#endif
	
#ifdef GRANDSTREAM_NETWORKS
	ast_str_append(filename, 0, ",%s,%s%s,%s",
	tmp,
	ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_APPEND) ? "a" : "",
	conference->b_profile.rec_options,
	conference->b_profile.rec_command);
#else
	ast_str_append(filename, 0, ",%s%s,%s",
	ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_APPEND) ? "a" : "",
	conference->b_profile.rec_options,
	conference->b_profile.rec_command);
#endif
}

static int is_new_rec_file(const char *rec_file, struct ast_str **orig_rec_file)
{
	if (!ast_strlen_zero(rec_file)) {
		if (!*orig_rec_file) {
			*orig_rec_file = ast_str_create(RECORD_FILENAME_INITIAL_SPACE);
		}

		if (*orig_rec_file
			&& strcmp(ast_str_buffer(*orig_rec_file), rec_file)) {
			ast_str_set(orig_rec_file, 0, "%s", rec_file);
			return 1;
		}
	}
	return 0;
}

struct confbridge_conference *conf_find_bridge(const char *conference_name)
{
	return ao2_find(conference_bridges, conference_name, OBJ_KEY);
}

/*!
 * \internal
 * \brief Returns whether or not conference is being recorded.
 *
 * \param conference The bridge to check for recording
 *
 * \note Must be called with the conference locked
 *
 * \retval 1, conference is recording.
 * \retval 0, conference is NOT recording.
 */
static int conf_is_recording(struct confbridge_conference *conference)
{
	return conference->record_chan != NULL;
}

/*!
 * \internal
 * \brief Stop recording a conference bridge
 *
 * \param conference The conference bridge on which to stop the recording
 *
 * \note Must be called with the conference locked
 *
 * \retval -1 Failure
 * \retval 0 Success
 */
static int conf_stop_record(struct confbridge_conference *conference)
{
	struct ast_channel *chan;
	struct ast_frame f = { AST_FRAME_CONTROL, .subclass.integer = AST_CONTROL_HANGUP };
#ifdef GRANDSTREAM_NETWORKS
	struct ast_str *str_tmp;
#endif

	if (!conf_is_recording(conference)) {
		return -1;
	}

	/* Remove the recording channel from the conference bridge. */
	chan = conference->record_chan;
	conference->record_chan = NULL;

#ifdef GRANDSTREAM_NETWORKS
	if (ast_channel_is_avs_engine(chan)) {
		ast_avs_control(chan, AVS_CONTROL_STOP_RECORD, NULL, 0);
	}

	str_tmp = ast_str_create(256);
	if (str_tmp) {
		int recording_file_size = 0;
		char *tmp = NULL;
		time_t now, duration;

		ast_str_reset(str_tmp);
		ast_str_copy_string(&str_tmp, conference->record_filename);

		tmp =ast_str_buffer(str_tmp);
		ast_strsep(&tmp, '.', AST_STRSEP_TRIM);
		ast_filesize(ast_str_buffer(str_tmp), &recording_file_size, "wav");
		if (conference->record_pause_duration > 0) {
			duration = conference->record_pause_duration;
		} else {
			now = ast_now_utc_timestamp();
			duration = now - conference->record_start_utc_timestamp;
		}
		ast_str_append(&conference->record_files_desc, -1,
			"duration=%ld,size=%dB,start-time=%ld@", duration, recording_file_size, conference->record_start_utc_timestamp);
		conference->record_start_utc_timestamp = -1;
		/*Start added by zyxia for paging video recording*/
		if (conference->is_paging_conf && conference->paging_caller_chan) {
			ast_cdr_setvar(ast_channel_name(conference->paging_caller_chan), "record_files_desc", ast_str_buffer(conference->record_files_desc));
		}
		/*End added by zyxia for paging video recording*/

		ast_log(LOG_DEBUG, "conf_stop_record | recordfiles:%s, recordfiles_desc:%s",
			ast_str_buffer(conference->record_files), ast_str_buffer(conference->record_files_desc));
		ast_log(LOG_DEBUG, "recording file:%s, size:%d", ast_str_buffer(str_tmp), recording_file_size);

		ast_free(str_tmp);
	}
	conference->recorded = CONF_RECORD_SHUTDOWN;
	memset(conference->record_link, 0, sizeof(conference->record_link));
#endif

	ast_queue_frame(chan, &f);
	ast_channel_unref(chan);

	ast_test_suite_event_notify("CONF_STOP_RECORD", "Message: stopped conference recording channel\r\nConference: %s", conference->b_profile.name);
	send_stop_record_event(conference);
	ast_ha_rsync_file(AST_MEETING_FILE_ENUM);

	return 0;
}


#ifdef GRANDSTREAM_NETWORKS

static int conf_pause_record(struct confbridge_conference *conference)
{
	struct ast_channel *chan;
	time_t pause_duration;

	if (!conf_is_recording(conference)) {
		return -1;
	}

	chan = conference->record_chan;
	if (ast_channel_is_avs_engine(chan)) {
		ast_avs_control(chan, AVS_CONTROL_PAUSE_RECORD, NULL, 0);
		pause_duration = ast_now_utc_timestamp() - conference->record_start_utc_timestamp;
		if (pause_duration > 0) {
			ast_channel_set_recordpaused_dur(chan, pause_duration);
			conference->record_pause_duration += pause_duration;
		}
		pbx_builtin_setvar_helper(chan, "REC_PAUSED", "yes");
	}
	ast_log(LOG_DEBUG, "record_pause_duration %ld", conference->record_pause_duration);
	ast_test_suite_event_notify("CONF_STOP_RECORD", "Message: stopped conference recording channel\r\nConference: %s", conference->b_profile.name);
	send_stop_record_event(conference);

	return 0;
}

#endif 

#ifdef GRANDSTREAM_NETWORKS
/*!
 * \internal
 * \brief Returns whether or not conference is being videoed.
 *
 * \param conference The bridge to check for videoing
 *
 * \note Must be called with the conference locked
 *
 * \retval 1, conference is videoing.
 * \retval 0, conference is NOT videoing.
 */
static int conf_is_videoing(struct confbridge_conference *conference)
{
	return conference->video_is_recording | conference->slides_is_recording | conference->focus_is_recording;
}
#endif


#if 0//#ifdef GRANDSTREAM_NETWORKS
static int conf_start_seamless_transfer_reocrd(const char* record_filename, struct confbridge_conference *conference)
{
		struct ast_app *mixmonapp;
		struct ast_channel *chan;
		struct ast_format_cap *cap;
		struct ast_bridge_features *features;

		if (conf_is_recording(conference) || ast_strlen_zero(record_filename)) {
			return -1;
		}

		mixmonapp = pbx_findapp("MixMonitor");
		if (!mixmonapp) {
			ast_log(LOG_WARNING, "Cannot record ConfBridge, MixMonitor app is not installed\n");
			return -1;
		}

		features = ast_bridge_features_new();
		if (!features) {
			return -1;
		}
		ast_set_flag(&features->feature_flags, AST_BRIDGE_CHANNEL_FLAG_IMMOVABLE);

		cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!cap) {
			ast_bridge_features_destroy(features);
			return -1;
		}
		ast_format_cap_append(cap, ast_format_slin, 0);

		/* Create the recording channel. */
		chan = ast_request("CBRec", cap, NULL, NULL, conference->name, NULL);
		ao2_ref(cap, -1);
		if (!chan) {
			ast_bridge_features_destroy(features);
			return -1;
		}

		/* Start recording. */
		if (chan && conference->bridge->avs_bridge_id > 0) {
			if (ast_avs_create_channel(chan, avs_channel_type_rtp, 0, ast_channel_get_port_group_id(chan)) != 0) {
				ast_log(LOG_ERROR, "Couldn't create avs channel for channel '%s'.", ast_channel_name(chan));
			}
		}

		ast_answer(chan);
		/* Put the channel into the conference bridge. */
		ast_channel_ref(chan);
		conference->record_chan = chan;

		if (ast_bridge_impart(conference->bridge, chan, NULL, features,
			AST_BRIDGE_IMPART_CHAN_INDEPENDENT)) {
			ast_hangup(chan);
			ast_channel_unref(chan);
			conference->record_chan = NULL;
			return -1;
		}
		pbx_exec(chan, mixmonapp, record_filename);

		ast_test_suite_event_notify("CONF_START_RECORD", "Message: started conference recording channel\r\nConference: %s", conference->b_profile.name);
		send_start_record_event(conference);
		return 0;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static int conf_start_record(struct confbridge_conference *conference, struct ast_channel *caller_chan) {
	int res = 0;
	struct ast_app *mixmonapp;
	struct ast_channel *chan;
	struct ast_format_cap *cap;
	struct ast_bridge_features *features;

	if (conf_is_recording(conference)) {
		return -1;
	}

	mixmonapp = pbx_findapp("MixMonitor");
	if (!mixmonapp) {
		ast_log(LOG_WARNING, "Cannot record ConfBridge, MixMonitor app is not installed\n");
		return -1;
	}

	features = ast_bridge_features_new();
	if (!features) {
		return -1;
	}
	ast_set_flag(&features->feature_flags, AST_BRIDGE_CHANNEL_FLAG_IMMOVABLE);

	cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!cap) {
		ast_bridge_features_destroy(features);
		return -1;
	}
	ast_format_cap_append(cap, ast_format_slin, 0);

	/* Create the recording channel. */
	chan = ast_request("CBRec", cap, NULL, NULL, conference->name, NULL);
	ao2_ref(cap, -1);
	if (!chan) {
		ast_bridge_features_destroy(features);
		return -1;
	}

	if (caller_chan) {
		pbx_builtin_setvar_helper(chan, "CONF_RECORD", conference->name);
		pbx_builtin_setvar_helper(chan, "PARENT_CHAN", ast_channel_name(caller_chan));
		pbx_builtin_setvar_helper(caller_chan, "CONF_RECORDING_CHAN", ast_channel_name(chan));
	}

	ast_channel_set_avs_address(chan, (char*)ast_channel_get_avs_address(caller_chan));

	/* Start recording. */
	if (chan && conference->bridge->avs_bridge_id > 0) {
		if (ast_avs_create_channel(chan, avs_channel_type_rtp, 0, ast_channel_get_port_group_id(chan)) != 0) {
			ast_log(LOG_ERROR, "Couldn't create avs channel for channel '%s'.", ast_channel_name(chan));
		}
	}

	set_rec_filename(conference, &conference->record_filename, &conference->record_files_uuid,
		is_new_rec_file(conference->b_profile.rec_file, &conference->orig_rec_file));
	ast_answer(chan);

	/* Put the channel into the conference bridge. */
	ast_channel_ref(chan);
	conference->record_chan = chan;
	conference->record_start_utc_timestamp = ast_now_utc_timestamp();

	if (ast_bridge_impart(conference->bridge, chan, NULL, features,
		AST_BRIDGE_IMPART_CHAN_INDEPENDENT)) {
		ast_hangup(chan);
		ast_channel_unref(chan);
		conference->record_chan = NULL;
		return -1;
	}

	// args here passed are a little tricky, [conference->record_filename] contains all record params
	// separated by comma.
	res = pbx_exec(chan, mixmonapp, ast_str_buffer(conference->record_filename));
	if (!res) {
		conference->recorded = CONF_RECORD_AUDIO;
		set_conference_record_link(conference, CONF_RECORD_AUDIO);

		ast_test_suite_event_notify("CONF_START_RECORD", "Message: started conference recording channel\r\nConference: %s", conference->b_profile.name);
		send_start_record_event(conference);
	} else {
		ast_channel_unref(chan);
		conference->record_chan = NULL;
		conference->recorded = CONF_RECORD_SHUTDOWN;
		ast_log(LOG_DEBUG, "conf record failed!");
	}

	return res;
}

static int conf_resume_record(struct confbridge_conference *conference, struct ast_channel *caller_chan) {
	int res = 0;
	struct ast_channel *chan = NULL;
	struct avs_record_file record_file = {0};

	chan = conference->record_chan;
	if (!chan) {
		ast_log(LOG_DEBUG, "[conf_resume_record] null record channel");
		return res = -1;
	}

	if (ast_channel_is_avs_engine(chan)) {
		snprintf(record_file.file_path, sizeof(record_file.file_path), "%s", ast_str_buffer(conference->record_filename));
		ast_avs_control(chan, AVS_CONTROL_RESUME_RECORD, &record_file, 0);
		pbx_builtin_setvar_helper(chan, "REC_PAUSED", NULL);

		if (caller_chan) {
			pbx_builtin_setvar_helper(chan, "CONF_RECORD", conference->name);
			pbx_builtin_setvar_helper(chan, "PARENT_CHAN", ast_channel_name(caller_chan));
			pbx_builtin_setvar_helper(caller_chan, "CONF_RECORDING_CHAN", ast_channel_name(chan));
		}

		conference->record_start_utc_timestamp = ast_now_utc_timestamp();
		conference->recorded = CONF_RECORD_AUDIO;
		set_conference_record_link(conference, CONF_RECORD_AUDIO);

		ast_test_suite_event_notify("CONF_START_RECORD", "Message: started conference recording channel\r\nConference: %s", conference->b_profile.name);
		send_start_record_event(conference);
	}
	
	return res;
}

#else
/*!
 * \internal
 * \brief Start recording the conference
 *
 * \param conference The conference bridge to start recording
 *
 * \note Must be called with the conference locked
 *
 * \retval 0 success
 * \retval non-zero failure
 */
static int conf_start_record(struct confbridge_conference *conference)
{
	struct ast_app *mixmonapp;
	struct ast_channel *chan;
	struct ast_format_cap *cap;
	struct ast_bridge_features *features;

	if (conf_is_recording(conference)) {
		return -1;
	}

	mixmonapp = pbx_findapp("MixMonitor");
	if (!mixmonapp) {
		ast_log(LOG_WARNING, "Cannot record ConfBridge, MixMonitor app is not installed\n");
		return -1;
	}

	features = ast_bridge_features_new();
	if (!features) {
		return -1;
	}
	ast_set_flag(&features->feature_flags, AST_BRIDGE_CHANNEL_FLAG_IMMOVABLE);

	cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!cap) {
		ast_bridge_features_destroy(features);
		return -1;
	}
	ast_format_cap_append(cap, ast_format_slin, 0);

	/* Create the recording channel. */
	chan = ast_request("CBRec", cap, NULL, NULL, conference->name, NULL);
	ao2_ref(cap, -1);
	if (!chan) {
		ast_bridge_features_destroy(features);
		return -1;
	}

	/* Start recording. */
	set_rec_filename(conference, &conference->record_filename,
		is_new_rec_file(conference->b_profile.rec_file, &conference->orig_rec_file));
	ast_answer(chan);

	/* Put the channel into the conference bridge. */
	ast_channel_ref(chan);
	conference->record_chan = chan;
	conference->record_start_utc_timestamp = ast_now_utc_timestamp();

	if (ast_bridge_impart(conference->bridge, chan, NULL, features,
		AST_BRIDGE_IMPART_CHAN_INDEPENDENT)) {
		ast_hangup(chan);
		ast_channel_unref(chan);
		conference->record_chan = NULL;
		return -1;
	}
	pbx_exec(chan, mixmonapp, ast_str_buffer(conference->record_filename));
	ast_test_suite_event_notify("CONF_START_RECORD", "Message: started conference recording channel\r\nConference: %s", conference->b_profile.name);
	send_start_record_event(conference);

	return 0;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static void notify_video_record_control(struct confbridge_conference *conference, enum avs_control_type type)
{
	struct confbridge_user *cur_user = NULL;

	if (!conference) {
		return;
	}

	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if (cur_user->chan) {
			switch (type) {
				case AVS_CONTROL_START_CONF_VIDEO_RECORD:
				case AVS_CONTROL_START_CONF_SLIDES_RECORD:
				case AVS_CONTROL_START_CONF_FOCUS_RECORD:
					ast_indicate(cur_user->chan, AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_START);
					break;
				case AVS_CONTROL_STOP_CONF_VIDEO_RECORD:
				case AVS_CONTROL_STOP_CONF_SLIDES_RECORD:
				case AVS_CONTROL_STOP_CONF_FOCUS_RECORD:
					ast_indicate(cur_user->chan, AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_STOP);
					break;
				default:
					break;
			}
		}
	}

	return;
}

static char *parse_filename(char *filename, char *buffer, size_t len)
{
	char *slash = NULL;

	if (ast_strlen_zero(filename)) {
		ast_log(LOG_WARNING, "No file name was provided\n");
		return NULL;
	}

	ast_copy_string(buffer, filename, len);

	if ((slash = strrchr(filename, '/'))) {
		*slash = '\0';
	}
	ast_mkdir(filename, 0777);

	return buffer;
}

static unsigned int dir_is_exist(const char *dir_name)
{
	DIR *dir = NULL;

	if (ast_strlen_zero(dir_name)) {
		return 0;
	}

	if ((dir = opendir(dir_name))) {
		closedir(dir);
	} else {
		return 0;
	}

	return 1;
}

const char *getLinkResult(const char *link_path)
{
	char full_path[256] = {0};
	char usb1_path[512] = {0};
	char usb2_path[512] = {0};

	if (link_path == NULL)
	{
		ast_log(LOG_ERROR, "param error");
		return "No_Device";
	}

	ast_safe_nvram_get("Device_USB1_path", usb1_path, sizeof(usb1_path));
	ast_safe_nvram_get("Device_USB2_path", usb2_path, sizeof(usb2_path));

	memset(full_path, 0, sizeof(full_path));
	if (readlink(link_path, full_path, sizeof(full_path) - 1) != -1) {
		ast_log(LOG_DEBUG, "get_link_result, read link success\n");
	}

	if (access(full_path, F_OK) == -1)
	{
		ast_log(LOG_ERROR, "path does not exist");
		return "No_Device";
	}

	if ( strstr(full_path, "_local") != NULL )
	{
		return "Local";
	}
	else if ( strncmp( full_path, "/media/mmcblk", strlen("/media/mmcblk") ) == 0 )
	{
		return "SD";
	}
	else if ( strncmp( full_path, "/media/sd", strlen("/media/sd") ) == 0 )
	{
		if ( usb1_path != NULL && usb1_path[0] != '\0' && strstr(full_path,usb1_path) != NULL )
		{
			return "USB1";
		}
		else if ( usb2_path != NULL && usb2_path[0] != '\0' && strstr(full_path,usb2_path) != NULL )
		{
			return "USB2";
		}
		else
		{
			ast_log(LOG_ERROR, "path does not exist");
		}
	}
	else if ( strstr(full_path, "_nas") != NULL )
	{
		return "Netdisk";
	}
	else
	{
		ast_log(LOG_ERROR, "path does not exist");
	}

	return "No_Device";
}

static int set_video_rec_filename(struct confbridge_conference *conference,
	struct avs_record_file *record_file, const char *record_file_path)
{
	char file_name[MAX_SOUNDFILE_LEN] = { 0 };
	char dest_dir[PATH_MAX];
	time_t now;
	struct tm *p;

	if (!ast_strlen_zero(record_file_path)) {
		snprintf(record_file->file_path, sizeof(record_file->file_path), "%s", record_file_path);
		return 0;
	}

	if (!ast_strlen_zero(conference->b_profile.rec_file)) {
		snprintf(record_file->file_path, sizeof(record_file->file_path), "%s", conference->b_profile.rec_file);
		return 0;
	}

	if (conference->is_paging_conf) {
		snprintf(dest_dir, sizeof(dest_dir), "%s/paging_video", ast_config_AST_SPOOL_DIR);
		if (!dir_is_exist(dest_dir)) {
			ast_log(LOG_WARNING, "set_video_rec_filename dir:%s is not exist\n", dest_dir);
			return -1;
		}

		strcpy(conference->b_profile.rec_file, ast_str_buffer(conference->video_filename));
		strcpy(record_file->file_path, conference->b_profile.rec_file);
		ast_str_append(&conference->video_files, 0, "%s", conference->b_profile.rec_file);
	}
	else {
		snprintf(dest_dir, sizeof(dest_dir), "%s/video", ast_config_AST_SPOOL_DIR);
		if (!dir_is_exist(dest_dir)) {
			ast_log(LOG_WARNING, "set_video_rec_filename dir:%s is not exist\n", dest_dir);
			return -1;
		}

		time(&now);
		p = localtime(&now);
		snprintf(file_name, sizeof(file_name), "%s/%d-%02d/meeting-%s-%u.mkv", dest_dir, 1900 + p->tm_year, 1 + p->tm_mon,
			conference->name,(unsigned int) now);
		ast_log(LOG_DEBUG, "set_video_rec_filename dir:%s file_name %s\n", dest_dir, file_name);

		ast_str_reset(conference->video_filename);
		ast_str_set(&conference->video_filename, 0, "%s",file_name);
		ast_str_append(&conference->video_files, 0, "%d-%02d/meeting-%s-%u.mkv@", 1900 + p->tm_year, 1 + p->tm_mon, conference->name, (unsigned int) now);

		strcpy(record_file->file_path, file_name);
	}
	return 0;
}

static int get_current_video_record_rooms_number(void)
{
	int count = 0;
	struct ao2_iterator iter;
	struct confbridge_conference *conference = NULL;

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (conference->video_is_recording) {
			count++;
		}
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	return count;
}

static int get_current_slides_record_rooms_number(void)
{
	int count = 0;
	struct ao2_iterator iter;
	struct confbridge_conference *conference = NULL;

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (conference->slides_is_recording) {
			count++;
		}
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	return count;
}

static int get_current_focus_record_rooms_number(void)
{
	int count = 0;
	struct ao2_iterator iter;
	struct confbridge_conference *conference = NULL;

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (conference->focus_is_recording) {
			count++;
		}
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	return count;
}

static int reach_video_record_limit(enum avs_control_type type)
{
	int current_video_record_count = 0;
	int current_slides_record_count = 0;
	int current_focus_record_count = 0;
	int max_video_record_count = 0;
	int max_video_encode_cap = 0;
	int current_used_encode_cap = 0;
	int video_encode_frame = 0;
	int slides_encode_frame = 0;
	int focus_encode_frame = 0;

	current_video_record_count = get_current_video_record_rooms_number();
	current_slides_record_count = get_current_slides_record_rooms_number();
	current_focus_record_count = get_current_focus_record_rooms_number();

	max_video_record_count = sfu_max_video_record_rooms_capability();
	max_video_encode_cap = sfu_max_video_encode_capability();

	video_encode_frame = sfu_video_encode_frame();
	slides_encode_frame = sfu_slides_encode_frame();
	focus_encode_frame = sfu_focus_encode_frame();

	current_used_encode_cap = current_video_record_count*video_encode_frame + current_slides_record_count*slides_encode_frame + current_focus_record_count*focus_encode_frame;

	if ((type == AVS_CONTROL_START_CONF_FOCUS_RECORD) && (focus_encode_frame == 0)) {
		ast_log(LOG_DEBUG, "reach_video_record_limit it's focus record don't need check\n");
		return 0;
	}

	if (current_video_record_count >= max_video_record_count && (type == AVS_CONTROL_START_CONF_VIDEO_RECORD)) {
		ast_log(LOG_DEBUG, "reach_video_record_limit current_count:%d, max_count:%d\n", current_video_record_count, max_video_record_count);
		return 1;
	}

	if (current_used_encode_cap >= max_video_encode_cap) {
		ast_log(LOG_DEBUG, "reach_video_record_limit current_used_encode_cap:%d, max_video_encode_cap:%d\n", current_used_encode_cap, max_video_encode_cap);
		return 1;
	}

	ast_log(LOG_DEBUG, "reach_video_record_limit type:%d current_video_record_count:%d, current_slides_record_count:%d, current_used_encode_cap:%d, max_video_record_count:%d, max_video_encode_cap:%d\n",
		type, current_video_record_count, current_slides_record_count, current_used_encode_cap, max_video_record_count, max_video_encode_cap);

	switch (type) {
		case AVS_CONTROL_START_CONF_VIDEO_RECORD:
			if ((current_used_encode_cap + video_encode_frame) >= max_video_encode_cap) {
				ast_log(LOG_DEBUG, "alloc video reach max video encode capability\n");
				return 1;
			}
			break;
		case AVS_CONTROL_START_CONF_SLIDES_RECORD:
			if ((current_used_encode_cap + slides_encode_frame) >= max_video_encode_cap) {
				ast_log(LOG_DEBUG, "alloc slides reach max video encode capability\n");
				return 1;
			}
			break;
		case AVS_CONTROL_START_CONF_FOCUS_RECORD:
			if ((current_used_encode_cap + focus_encode_frame) >= max_video_encode_cap) {
				ast_log(LOG_DEBUG, "alloc slides reach max video encode capability\n");
				return 1;
			}
			break;
		default:
			break;
	}

	return 0;
}

static int type_is_start_record(enum avs_control_type type)
{
	if (type == AVS_CONTROL_START_CONF_VIDEO_RECORD ||
		type == AVS_CONTROL_START_CONF_SLIDES_RECORD ||
		type == AVS_CONTROL_START_CONF_FOCUS_RECORD) {
		return 1;
	}

	return 0;
}

static int type_is_stop_record(enum avs_control_type type)
{
	if (type == AVS_CONTROL_STOP_CONF_VIDEO_RECORD ||
		type == AVS_CONTROL_STOP_CONF_SLIDES_RECORD ||
		type == AVS_CONTROL_STOP_CONF_FOCUS_RECORD) {
		return 1;
	}

	return 0;
}

/*Determine if video recording is currently available*/
static int video_record_control_resource_judgment(struct confbridge_conference *conference, enum avs_control_type type)
{
	char dest_dir[PATH_MAX] = {0};
	char result[16] = {0};
	int ami_send_status = 0;

	ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		ami_send_status = atoi(result);
	}

#ifdef BUILD_ON_CLOUD_PBX
	int gdms_record_support = 0;
	memset(result, 0, sizeof(result));
	ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		gdms_record_support = atoi(result);
	}

	if ( ( alarm_on || (ami_send_status & SEND_STATUS_VIDEOING) || !gdms_record_support ) && type_is_start_record(type) ) {
#else 
	if ( ( alarm_on || (ami_send_status & SEND_STATUS_VIDEOING) ) && type_is_start_record(type) ) {
#endif // BUILD_ON_CLOUD_PBX
		ast_log(LOG_DEBUG, "Alarm is on, can start video record for conference:%s\n", conference->name);
		return -1;
	}

	if (reach_video_record_limit(type) && type_is_start_record(type)) {
		ast_log(LOG_DEBUG, "Already reach video reocord rooms limit, can't record for conference:%s\n", conference->name);
		return -1;
	}

	if (type_is_start_record(type)) {
		snprintf(dest_dir, sizeof(dest_dir), "%s/video", ast_config_AST_SPOOL_DIR);
		if (!dir_is_exist(dest_dir)) {
			ast_log(LOG_WARNING, "set_video_rec_filename dir:%s is not exist\n", dest_dir);
			return -1;
		}
	}

	return 0;
}

static int send_avs_video_record_control(struct confbridge_conference *conference,
	const char *record_file_path, enum avs_control_type type)
{
	struct avs_record_file record_file;
	char buffer[1024] = { 0 };
	char result[16] = {0};
	int ami_send_status = 0;
	int ret = -1;
	char *file_name = NULL;
	char filename[AST_FILENAME_MAX]={0};
	char encrypt_key[64];
	char encrypt_version[64];

	if (!conference || !conference->bridge) {
		ast_log(LOG_ERROR, "Can't get conference bridge\n");
		return -1;
	}

	conference->play_videoing_sound = 0;

	ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		ami_send_status = atoi(result);
	}

#ifdef BUILD_ON_CLOUD_PBX
	int gdms_record_support = 0;
	memset(result, 0, sizeof(result));
	ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		gdms_record_support = atoi(result);
	}

	if ( ( alarm_on || (ami_send_status & SEND_STATUS_VIDEOING) || !gdms_record_support ) && type_is_start_record(type) ) {
#else 
	if ( ( alarm_on || (ami_send_status & SEND_STATUS_VIDEOING) ) && type_is_start_record(type) ) {
#endif // BUILD_ON_CLOUD_PBX
		ast_log(LOG_DEBUG, "Alarm is on, can start video record for conference:%s\n", conference->name);
		return -1;
	}

	if (reach_video_record_limit(type) && type_is_start_record(type)) {
		ast_log(LOG_DEBUG, "Already reach video reocord rooms limit, can't record for conference:%s\n", conference->name);
		return -1;
	}

	memset(&record_file, 0, sizeof(record_file));

	if (type_is_start_record(type)) {
		if (set_video_rec_filename(conference, &record_file, record_file_path)) {
			ast_log(LOG_ERROR, "Can't set conference video record filename\n");
			return -1;
		}

		ast_copy_string(record_file.file_path, ast_strdupa(parse_filename(record_file.file_path, buffer, sizeof(buffer))), sizeof(record_file.file_path));
		if (ast_strlen_zero(record_file.file_path)) {
			ast_log(LOG_ERROR, "Can't parse file path\n");
			return -1;
		}
		
		file_crypt_algorithm_version(encrypt_version, sizeof(encrypt_version));
		snprintf(record_file.encrypt_version, sizeof(record_file.encrypt_version), "%s", encrypt_version);
		
		if (!strlen(ast_str_buffer(conference->video_files_uuid))) {
			file_crypt_new_uuid(encrypt_key, sizeof(encrypt_key));
			snprintf(record_file.encrypt_key, sizeof(record_file.encrypt_key), "%s", encrypt_key);
			ast_str_append(&(conference->video_files_uuid), 0, "%s", encrypt_key);
			ast_log(LOG_DEBUG, "[send_avs_video_record_control] new encrypt_key:%s encrypt_version:%s", encrypt_key, encrypt_version);
		} else {
			snprintf(record_file.encrypt_key, sizeof(record_file.encrypt_key), "%s", ast_str_buffer(conference->video_files_uuid));
			ast_log(LOG_DEBUG, "[send_avs_video_record_control] existing encrypt key:%s encrypt_version:%s", ast_str_buffer(conference->video_files_uuid), encrypt_version);
		}
	}

	if(!ast_strlen_zero(record_file.file_path) && ast_media_role_get() == MEDIA_ROLE_MASTER
		&& conference->bridge && !ast_strlen_zero(conference->bridge->avs_address)
		&& strcmp("127.0.0.1",conference->bridge->avs_address)){

		file_name = strrchr(record_file.file_path, '/');
		if (file_name) {
			ast_create_record_file(record_file.file_path);
			snprintf(filename, sizeof(filename), "/data/slave_media_recording%s", file_name);
			ast_copy_string(record_file.file_path,filename, sizeof(record_file.file_path));
			memset(conference->recordfile_slave_media,0,sizeof(conference->recordfile_slave_media));
			strncpy(conference->recordfile_slave_media, filename,  sizeof(conference->recordfile_slave_media) - 1);
			ast_increase_media_records();
		}
	}

	ret = ast_avs_bridge_control(conference->bridge, type, &record_file, sizeof(record_file));
	if (ret != R_SUCCESS) {
		ast_log(LOG_ERROR, "Send avs bridge control failed\n");
		ast_decrease_media_records(1);
		return ret;
	}

	conference->play_videoing_sound = 1;
	switch (type) {
		case AVS_CONTROL_START_CONF_VIDEO_RECORD:
			conference->recorded = CONF_RECORD_VIDEO;
			conference->video_is_recording = 1;
			conference->video_start_utc_timestamp = ast_now_utc_timestamp();
			conference->record_start_utc_timestamp = ast_now_utc_timestamp();
			set_conference_record_link(conference, CONF_RECORD_VIDEO);
			break;
		case AVS_CONTROL_STOP_CONF_VIDEO_RECORD:
			conference->recorded = CONF_RECORD_SHUTDOWN;
			conference->video_is_recording = 0;
			memset(conference->record_link, 0, sizeof(conference->record_link));
			break;
		case AVS_CONTROL_START_CONF_SLIDES_RECORD:
			conference->recorded = CONF_RECORD_SLIDES;
			conference->slides_is_recording = 1;
			conference->video_start_utc_timestamp = ast_now_utc_timestamp();
			conference->record_start_utc_timestamp = ast_now_utc_timestamp();
			set_conference_record_link(conference, CONF_RECORD_SLIDES);
			break;
		case AVS_CONTROL_STOP_CONF_SLIDES_RECORD:
			conference->recorded = CONF_RECORD_SHUTDOWN;
			conference->slides_is_recording = 0;
			memset(conference->record_link, 0, sizeof(conference->record_link));
			break;
		case AVS_CONTROL_START_CONF_FOCUS_RECORD:
			conference->recorded = CONF_RECORD_FOUCS;
			conference->focus_is_recording = 1;
			conference->video_start_utc_timestamp = ast_now_utc_timestamp();
			conference->record_start_utc_timestamp = ast_now_utc_timestamp();
			set_conference_record_link(conference, CONF_RECORD_FOUCS);
			break;
		case AVS_CONTROL_STOP_CONF_FOCUS_RECORD:
			conference->recorded = CONF_RECORD_SHUTDOWN;
			conference->focus_is_recording = 0;
			memset(conference->record_link, 0, sizeof(conference->record_link));
			break;
		default:
			break;
	}

	if (type_is_stop_record(type)) {
		int videoing_file_size = 0;
		struct ast_str *str_tmp = NULL;
		str_tmp = ast_str_create(1024);
		if (str_tmp) {
			char *tmp = NULL;
			time_t now, duration;
			struct stat st;

			if(!ast_strlen_zero(conference->recordfile_slave_media) && ast_media_role_get() == MEDIA_ROLE_MASTER
				&& conference->bridge && !ast_strlen_zero(conference->bridge->avs_address)
				&& strcmp("127.0.0.1",conference->bridge->avs_address)){
				ast_decrease_media_records(1);
				ast_sync_record_from_slave_and_move(AST_RSYNC_RECORD_CMD_NO_LINK,conference->bridge->avs_address,conference->recordfile_slave_media,ast_str_buffer(conference->video_filename),AST_AVS_SYNC_RECORD_MODULE);
				memset(conference->recordfile_slave_media, 0, sizeof(conference->recordfile_slave_media));
				usleep(500000);
			}

			ast_str_reset(str_tmp);
			ast_str_copy_string(&str_tmp, conference->video_filename);
			tmp =ast_str_buffer(str_tmp);
			ast_log(LOG_DEBUG, "video_filename %s", tmp);

			if (!stat(tmp, &st)) {
				videoing_file_size = st.st_size;
			}
			now = ast_now_utc_timestamp();
			duration = now - conference->video_start_utc_timestamp;
			ast_str_append(&conference->video_files_desc, -1,
				"duration=%ld,size=%dB,start-time=%ld@", duration, videoing_file_size, conference->video_start_utc_timestamp);
			conference->video_start_utc_timestamp = -1;
			/*Start added by zyxia for paging video recording*/
			if (conference->is_paging_conf && conference->paging_caller_chan) {
				ast_str_append(&conference->record_files_desc, -1,
					"%s", ast_str_buffer(conference->video_files_desc));
				ast_cdr_setvar(ast_channel_name(conference->paging_caller_chan), "record_files_desc", ast_str_buffer(conference->record_files_desc));
			}
			/*End added by zyxia for paging video recording*/

			ast_log(LOG_DEBUG, "conf_stop_videoing | videofiles:%s, videofiles_desc:%s",
				ast_str_buffer(conference->video_files), ast_str_buffer(conference->video_files_desc));
			ast_log(LOG_DEBUG, "videoing file:%s, size:%d", ast_str_buffer(str_tmp), videoing_file_size);

			ast_free(str_tmp);
		}
	}

	notify_video_record_control(conference, type);

	return ret;
}

static int conf_start_video_record(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference) {
		return -1;
	}

	if (conference->video_is_recording) {
		return -1;
	}

	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_START_CONF_VIDEO_RECORD);

	//ast_test_suite_event_notify("CONF_START_RECORD", "Message: started conference recording channel\r\nConference: %s", conference->b_profile.name);
	//send_start_record_event(conference);

	return ret;
}

static int conf_stop_video_record(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference->video_is_recording) {
		return -1;
	}

	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_VIDEO_RECORD);

	//ast_test_suite_event_notify("CONF_STOP_RECORD", "Message: stopped conference recording channel\r\nConference: %s", conference->b_profile.name);
	//send_stop_record_event(conference);
	if(!ret)
	{
		ast_ha_rsync_file(AST_CONFERENCE_VIDEO_FILE_ENUM);
		ast_ha_rsync_file(AST_PAGE_VIDEO_FILE_ENUM);
	}
	return ret;
}

static int conf_start_slides_record(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference) {
		return -1;
	}

	if (conference->slides_is_recording) {
		ast_log(LOG_WARNING, "Conference slides video is already recording\n");
		return -1;
	}

	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_START_CONF_SLIDES_RECORD);

	return ret;
}

static int conf_stop_slides_record(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference->slides_is_recording) {
		ast_log(LOG_WARNING, "Conference slides video is already stop record\n");
		return -1;
	}

	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_SLIDES_RECORD);

	if(!ret)
	{
		ast_ha_rsync_file(AST_CONFERENCE_VIDEO_FILE_ENUM);
		ast_ha_rsync_file(AST_PAGE_VIDEO_FILE_ENUM);
	}

	return ret;
}

static int conf_start_focus_record(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference) {
		return -1;
	}

	if (conference->focus_is_recording) {
		ast_log(LOG_WARNING, "Conference focus video is already recording\n");
		return -1;
	}

	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_START_CONF_FOCUS_RECORD);

	return ret;
}

static int conf_stop_focus_record(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference->focus_is_recording) {
		ast_log(LOG_WARNING, "Conference foucs video is already stop record\n");
		return -1;
	}

	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_FOCUS_RECORD);
	if(!ret)
	{
		ast_ha_rsync_file(AST_CONFERENCE_VIDEO_FILE_ENUM);
		ast_ha_rsync_file(AST_PAGE_VIDEO_FILE_ENUM);
	}

	return ret;
}

static int stop_record(struct confbridge_conference *conference)
{
	if (!conference) {
		return -1;
	}

	if (conf_is_recording(conference)) {
		conf_stop_record(conference);
	} else if (conference->video_is_recording) {
		conf_stop_video_record(conference);
	} else if (conference->focus_is_recording) {
		conf_stop_focus_record(conference);
	} else if (conference->slides_is_recording) {
		conf_stop_slides_record(conference);
	} else {
		ast_log(LOG_DEBUG, "conference is not enable recording\n");
	}

	return 0;
}

static int conf_send_avs_mcu_control(struct confbridge_conference *conference,
	enum avs_control_type type)
{
	int ret = -1;

	if (!conference || !conference->bridge) {
		ast_log(LOG_ERROR, "Can't get conference bridge\n");
		return -1;
	}

	ret = ast_avs_bridge_control(conference->bridge, type, NULL, 0);
	if (ret != R_SUCCESS) {
		ast_log(LOG_ERROR, "Send avs bridge control failed, conference:%s\n", conference->name);
		return ret;
	}

	switch (type) {
		case AVS_CONTROL_ENABLE_MCU:
			conference->mcu_enabled = 1;
			break;
		case AVS_CONTROL_DISABLE_MCU:
			conference->mcu_enabled  = 0;
			break;
		default:
			break;
	}

	return ret;
}

static int conf_enable_mcu(struct confbridge_conference *conference)
{
	int ret = 0;

	if (!conference) {
		return -1;
	}

	if (conference->mcu_enabled) {
		ast_log(LOG_DEBUG, "conference:%s already enabled MCU\n", conference->name);
		return 0;
	}

	ret = conf_send_avs_mcu_control(conference, AVS_CONTROL_ENABLE_MCU);

	return ret;
}
#endif

/* \brief Playback the given filename and monitor for any dtmf interrupts.
 *
 * This function is used to playback sound files on a given channel and optionally
 * allow dtmf interrupts to occur.
 *
 * If the optional bridge_channel parameter is given then sound file playback
 * is played on that channel and dtmf interruptions are allowed. However, if
 * bridge_channel is not set then the channel parameter is expected to be set
 * instead and non interruptible playback is played on that channel.
 *
 * \param bridge_channel Bridge channel to play file on
 * \param channel Optional channel to play file on if bridge_channel not given
 * \param filename The file name to playback
 *
 * \retval -1 failure during playback, 0 on file was fully played, 1 on dtmf interrupt.
 */
static int play_file(struct ast_bridge_channel *bridge_channel, struct ast_channel *channel,
		     const char *filename)
{
	struct ast_channel *chan;
	const char *stop_digits;
	int digit;

	if (bridge_channel) {
		chan = bridge_channel->chan;
		stop_digits = AST_DIGIT_ANY;
	} else {
		chan = channel;
		stop_digits = AST_DIGIT_NONE;
	}

	digit = ast_stream_and_wait(chan, filename, stop_digits);
	if (digit < 0) {
		ast_log(LOG_WARNING, "Failed to playback file '%s' to channel\n", filename);
		return -1;
	}

	if (digit > 0) {
		ast_stopstream(chan);
		if (bridge_channel){
			ast_bridge_channel_feature_digit_add(bridge_channel, digit);
		}
		return 1;
	}

	return 0;
}

/*!
 * \internal
 * \brief Complain if the given sound file does not exist.
 *
 * \param filename Sound file to check if exists.
 *
 * \retval non-zero if the file exists.
 */
static int sound_file_exists(const char *filename)
{
	if (ast_fileexists(filename, NULL, NULL)) {
		return -1;
	}
	ast_log(LOG_WARNING, "File %s does not exist in any format\n", filename);
	return 0;
}

/*!
 * \brief Announce number of users in the conference bridge to the caller
 *
 * \param conference Conference bridge to peek at
 * \param user Optional Caller
 * \param bridge_channel The bridged channel involved
 *
 * \note if caller is NULL, the announcment will be sent to all participants in the conference.
 * \return Returns 0 on success, -1 if the user hung up
 */
static int announce_user_count(struct confbridge_conference *conference, struct confbridge_user *user,
			       struct ast_bridge_channel *bridge_channel)
{
#ifdef GRANDSTREAM_NETWORKS
	const char *only_you = conf_get_sound(CONF_SOUND_ONLY_PERSON, conference->b_profile.sounds);
#endif
	const char *other_in_party = conf_get_sound(CONF_SOUND_OTHER_IN_PARTY, conference->b_profile.sounds);
	const char *only_one = conf_get_sound(CONF_SOUND_ONLY_ONE, conference->b_profile.sounds);
	const char *there_are = conf_get_sound(CONF_SOUND_THERE_ARE, conference->b_profile.sounds);

	if (conference->activeusers <= 1) {
		/* Awww we are the only person in the conference bridge OR we only have waitmarked users */
#ifdef GRANDSTREAM_NETWORKS
		ast_stream_and_wait(user->chan, only_you, "");
#endif
		return 0;
	} else if (conference->activeusers == 2) {
		if (user) {
			/* Eep, there is one other person */
			if (play_file(bridge_channel, user->chan, only_one) < 0) {
				return -1;
			}
		} else {
			play_sound_file(conference, only_one);
		}
	} else {
		/* Alas multiple others in here */
		if (user) {
			if (ast_stream_and_wait(user->chan,
				there_are,
				"")) {
				return -1;
			}
			if (ast_say_number(user->chan, conference->activeusers - 1, "", ast_channel_language(user->chan), NULL)) {
				return -1;
			}
			if (play_file(bridge_channel, user->chan, other_in_party) < 0) {
				return -1;
			}
		} else if (sound_file_exists(there_are) && sound_file_exists(other_in_party)) {
			play_sound_file(conference, there_are);
			play_sound_number(conference, conference->activeusers - 1);
			play_sound_file(conference, other_in_party);
		}
	}
	return 0;
}

/*!
 * \brief Play back an audio file to a channel
 *
 * \param user User to play audio prompt to
 * \param filename Prompt to play
 *
 * \return Returns 0 on success, -1 if the user hung up
 * \note Generally this should be called when the conference is unlocked to avoid blocking
 * the entire conference while the sound is played. But don't unlock the conference bridge
 * in the middle of a state transition.
 */
static int play_prompt_to_user(struct confbridge_user *user, const char *filename)
{
	return ast_stream_and_wait(user->chan, filename, "");
}

static void handle_video_on_join(struct confbridge_conference *conference, struct ast_channel *chan, int marked)
{
	/* Right now, only marked users are automatically set as the single src of video.*/
	if (!marked) {
		return;
	}

	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_FIRST_MARKED)) {
		int set = 1;
		struct confbridge_user *user = NULL;

		ao2_lock(conference);
		/* see if anyone is already the video src */
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (user->chan == chan) {
				continue;
			}
			if (ast_bridge_is_video_src(conference->bridge, user->chan)) {
				set = 0;
				break;
			}
		}
		ao2_unlock(conference);
		if (set) {
			ast_bridge_set_single_src_video_mode(conference->bridge, chan);
		}
	} else if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_LAST_MARKED)) {
		/* we joined and are video capable, we override anyone else that may have already been the video feed */
		ast_bridge_set_single_src_video_mode(conference->bridge, chan);
	}
}

static void handle_video_on_exit(struct confbridge_conference *conference, struct ast_channel *chan)
{
	struct confbridge_user *user = NULL;

	/* if this isn't a video source, nothing to update */
	if (!ast_bridge_is_video_src(conference->bridge, chan)) {
		return;
	}

	ast_bridge_remove_video_src(conference->bridge, chan);

	/* If in follow talker mode, make sure to restore this mode on the
	 * bridge when a source is removed.  It is possible this channel was
	 * only set temporarily as a video source by an AMI or DTMF action. */
	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_FOLLOW_TALKER)) {
		ast_bridge_set_talker_src_video_mode(conference->bridge);
	}

	/* if the video_mode isn't set to automatically pick the video source, do nothing on exit. */
	if (!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_FIRST_MARKED) &&
		!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_LAST_MARKED)) {
		return;
	}

	/* Make the next available marked user the video src.  */
	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user->chan == chan) {
			continue;
		}
		if (ast_test_flag(&user->u_profile, USER_OPT_MARKEDUSER)) {
			ast_bridge_set_single_src_video_mode(conference->bridge, user->chan);
			break;
		}
	}
	ao2_unlock(conference);
}

struct hangup_data
{
	struct confbridge_conference *conference;
	ast_mutex_t lock;
	ast_cond_t cond;
	int hungup;
};

/*!
 * \brief Hang up the announcer channel
 *
 * This hangs up the announcer channel in the conference. This
 * runs in the playback queue taskprocessor since we do not want
 * to hang up the channel while it's trying to play an announcement.
 *
 * This task is performed synchronously, so there is no need to
 * perform any cleanup on the passed-in data.
 *
 * \param data A hangup_data structure
 * \return 0
 */
static int hangup_playback(void *data)
{
	struct hangup_data *hangup = data;

	ast_autoservice_stop(hangup->conference->playback_chan);

	ast_hangup(hangup->conference->playback_chan);
	hangup->conference->playback_chan = NULL;

	ast_mutex_lock(&hangup->lock);
	hangup->hungup = 1;
	ast_cond_signal(&hangup->cond);
	ast_mutex_unlock(&hangup->lock);

	return 0;
}

static void hangup_data_init(struct hangup_data *hangup, struct confbridge_conference *conference)
{
	ast_mutex_init(&hangup->lock);
	ast_cond_init(&hangup->cond, NULL);

	hangup->conference = conference;
	hangup->hungup = 0;
}

static void hangup_data_destroy(struct hangup_data *hangup)
{
	ast_mutex_destroy(&hangup->lock);
	ast_cond_destroy(&hangup->cond);
}

/*!
 * \brief Destroy a conference bridge
 *
 * \param obj The conference bridge object
 *
 * \return Returns nothing
 */
static void destroy_conference_bridge(void *obj)
{
	struct confbridge_conference *conference = obj;

	ast_debug(1, "Destroying conference bridge '%s'\n", conference->name);

	if (conference->playback_chan) {
		if (conference->playback_queue) {
			struct hangup_data hangup;
			hangup_data_init(&hangup, conference);

			if (!ast_taskprocessor_push(conference->playback_queue, hangup_playback, &hangup)) {
				ast_mutex_lock(&hangup.lock);
				while (!hangup.hungup) {
					ast_cond_wait(&hangup.cond, &hangup.lock);
				}
				ast_mutex_unlock(&hangup.lock);
			}

			hangup_data_destroy(&hangup);
		} else {
			/* Playback queue is not yet allocated. Just hang up the channel straight */
			ast_hangup(conference->playback_chan);
			conference->playback_chan = NULL;
		}
	}

	/* Destroying a conference bridge is simple, all we have to do is destroy the bridging object */
	if (conference->bridge) {
		ast_bridge_destroy(conference->bridge, 0);
		conference->bridge = NULL;
	}

	ast_channel_cleanup(conference->record_chan);
	ast_free(conference->orig_rec_file);
	ast_free(conference->record_filename);
#ifdef GRANDSTREAM_NETWORKS
	ast_free(conference->record_files);
	ast_free(conference->record_files_desc);
	ast_free(conference->video_filename);
	ast_free(conference->video_files);
	ast_free(conference->video_files_desc);
	ast_free(conference->record_files_uuid);
	ast_free(conference->video_files_uuid);
	if (conference->focus_chan) {
		ast_channel_unref(conference->focus_chan);
		conference->focus_chan = NULL;
	}
#endif

	conf_bridge_profile_destroy(&conference->b_profile);
	ast_taskprocessor_unreference(conference->playback_queue);
	ast_mutex_destroy(&conference->playback_lock);

#ifdef GRANDSTREAM_NETWORKS
	if (conference->schedule_id_to_start) {
		ast_free(conference->schedule_id_to_start);
		conference->schedule_id_to_start = NULL;
	}

	if (conference->schedule_start_time) {
		ast_free(conference->schedule_start_time);
		conference->schedule_start_time = NULL;
	}

	if (conference->feature_id) {
		ast_free(conference->feature_id);
		conference->feature_id = NULL;
	}
	if (conference->schedule_subject) {
		ast_free(conference->schedule_subject);
		conference->schedule_subject = NULL;
	}
	if (conference->schedule_id_processing) {
		ast_free(conference->schedule_id_processing);
		conference->schedule_id_processing = NULL;
	}
	if (conference->schedule_host_exten) {
		ast_free(conference->schedule_host_exten);
		conference->schedule_host_exten = NULL;
	}
	if (conference->schedule_host_email) {
		ast_free(conference->schedule_host_email);
		conference->schedule_host_email = NULL;
	}
	if (conference->schedule_utc_start_time) {
		ast_free(conference->schedule_utc_start_time);
		conference->schedule_utc_start_time = NULL;
	}
	if (conference->im_service_prefix) {
		ast_free(conference->im_service_prefix);
		conference->im_service_prefix = NULL;
	}
	if (conference->share_system_audio_chan) {
		ast_free(conference->share_system_audio_chan);
		conference->share_system_audio_chan = NULL;
	}

	if (conference->bridge_router) {
		stasis_message_router_unsubscribe_and_join(conference->bridge_router);
		conference->bridge_router = NULL;
	}
	if (conference->link_auth_id) {
		ast_free(conference->link_auth_id);
		conference->link_auth_id = NULL;
	}
	if (conference->conf_creator) {
		ast_free(conference->conf_creator);
		conference->conf_creator = NULL;
	}
#endif
}

/*! \brief Call the proper join event handler for the user for the conference bridge's current state
 * \internal
 * \param user The conference bridge user that is joining
 * \retval 0 success
 * \retval -1 failure
 */
static int handle_conf_user_join(struct confbridge_user *user)
{
	conference_event_fn handler;
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKEDUSER)) {
		handler = user->conference->state->join_marked;
	} else if (ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED)) {
		handler = user->conference->state->join_waitmarked;
	} else {
		handler = user->conference->state->join_unmarked;
	}

	ast_assert(handler != NULL);

	if (!handler) {
		conf_invalid_event_fn(user);
		return -1;
	}

	handler(user);

	return 0;
}

/*! \brief Call the proper leave event handler for the user for the conference bridge's current state
 * \internal
 * \param user The conference bridge user that is leaving
 * \retval 0 success
 * \retval -1 failure
 */
static int handle_conf_user_leave(struct confbridge_user *user)
{
	conference_event_fn handler;
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKEDUSER)) {
		handler = user->conference->state->leave_marked;
	} else if (ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED)) {
		handler = user->conference->state->leave_waitmarked;
	} else {
		handler = user->conference->state->leave_unmarked;
	}

	ast_assert(handler != NULL);

	if (!handler) {
		/* This should never happen. If it does, though, it is bad. The user will not have been removed
		 * from the appropriate list, so counts will be off and stuff. The conference won't be torn down, etc.
		 * Shouldn't happen, though. */
		conf_invalid_event_fn(user);
		return -1;
	}

	handler(user);

	return 0;
}

void conf_update_user_mute(struct confbridge_user *user)
{
	int mute_user;
	int mute_system;
	int mute_effective;

#ifdef GRANDSTREAM_NETWORKS
	/*Modify for Bug 233423*/
	if (user->conference && !ast_strlen_zero(user->conference->share_system_audio_chan)) {
		mute_user = user->muted;
		ast_debug(1, "wave share system audio, muteCtrl MuteVoip 0!");
	} else {
		/* User level mute request. */
		mute_user = user->muted || user->mic_muted;
	}
#else
	/* User level mute request. */
	mute_user = user->muted;
#endif
	/* System level mute request. */
	mute_system = user->playing_moh
		/*
		 * Do not allow waitmarked users to talk to anyone unless there
		 * is a marked user present.
		 */
		|| (!user->conference->markedusers
			&& ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED));

	mute_effective = mute_user || mute_system;

	ast_debug(1, "User %s is %s: user:%d system:%d.\n",
		ast_channel_name(user->chan), mute_effective ? "muted" : "unmuted",
		mute_user, mute_system);
	user->features.mute = mute_effective;
	ast_test_suite_event_notify("CONF_MUTE_UPDATE",
		"Mode: %s\r\n"
		"Conference: %s\r\n"
		"Channel: %s",
		mute_effective ? "muted" : "unmuted",
		user->conference->b_profile.name,
		ast_channel_name(user->chan));
#ifdef GRANDSTREAM_NETWORKS

	ast_avs_control(user->chan, (mute_effective ? AVS_CONTROL_MUTE : AVS_CONTROL_UNMUTE), NULL, 0);
#endif
}

#ifdef GRANDSTREAM_NETWORKS
static void conf_update_user_mic_mute(struct confbridge_user *user)
{
	int mute_user = 0;
	int mute_system = 0;
	int mute_effective = 0;

	if (!user || !user->conference || !user->chan) {
		return;
	}
	if (user->mic_muted && user->conference->share_system_audio_chan) {
		if (ast_channel_name(user->chan)
			&& !strcasecmp(ast_channel_name(user->chan), user->conference->share_system_audio_chan)) {
			ast_debug(1, "user %s share the system audio,can't mute to AVS.", ast_channel_name(user->chan));
			return;
		}
	}
	/* User level mic mute request. */
	mute_user = user->mic_muted;

	/* System level mute request. */
	mute_system = user->playing_moh
		/*
		 * Do not allow waitmarked users to talk to anyone unless there
		 * is a marked user present.
		 */
		|| (!user->conference->markedusers
			&& ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED));

	mute_effective = mute_user || mute_system;

	ast_debug(1, "User %s is %s: user:%d mic muted:%d.\n",
			ast_channel_name(user->chan), mute_effective ? "muted" : "unmuted",
			mute_user, mute_user);

	ast_avs_control(user->chan, (mute_effective ? AVS_CONTROL_MUTE : AVS_CONTROL_UNMUTE), NULL, 0);
}
#endif

/*
 * \internal
 * \brief Mute/unmute a single user.
 */
#ifdef GRANDSTREAM_NETWORKS
static void generic_mute_unmute_user(struct confbridge_conference *conference, struct confbridge_user *user, int mute, int is_mic, int play_sound)
{

	/*!begin Added by pyzhong@grandstream.cn for bug[83133], 2017-4-12*/
	if (user->kicked) {
		ast_log(LOG_WARNING, "Channel %s has been kicked.", ast_channel_name(user->chan));
		return;
	}
	/*!end Added by pyzhong@grandstream.cn for bug[83133], 2017-4-12*/


	/* Set user level mute request. */
	if (is_mic) {
		user->mic_muted = mute ? 1 : 0;
	} else {
		user->muted = mute ? 1 : 0;
	}

	if (is_mic) {
		if (user->muted) {
			ast_log(LOG_DEBUG, "user is muted!,mic mute control no need send to AVS.");
		} else {
			conf_update_user_mic_mute(user);
		}
	} else {
		conf_update_user_mute(user);
	}

	ast_test_suite_event_notify("CONF_MUTE",
		"Message: participant %s %s %s\r\n"
		"Conference: %s\r\n"
		"Channel: %s",
		ast_channel_name(user->chan),
		user->muted ? "muted" : "unmuted",
		user->mic_muted ? "mic_muted" : "mic_unmuted",,
		conference->b_profile.name,
		ast_channel_name(user->chan));
	if (mute) {
		send_mute_event(user, conference, 0, is_mic);
	} else {
		send_unmute_event(user, conference, 0, is_mic);
	}

	//added by zhywang. Add prompt for ami. Refer to function "bridge_channel_handle_action"
	ast_channel_lock(user->chan);
	struct ast_bridge_channel *bridge_channel = ast_channel_get_bridge_channel(user->chan);
	ast_channel_unlock(user->chan);
	if (!bridge_channel) {
		ast_log(LOG_WARNING, "Channel %s has no bridge channel.", ast_channel_name(user->chan));
		return;
	}

	if (play_sound) {
		bridge_channel_suspend(bridge_channel);
		ast_indicate(user->chan, AST_CONTROL_SRCUPDATE);

		if(play_prompt_to_user(user,conf_get_sound(mute ? CONF_SOUND_MUTED : CONF_SOUND_UNMUTED, user->b_profile.sounds))){
			ast_log(LOG_ERROR,"conf mute/unmute file failed to playback");
		}
		ast_indicate(user->chan, AST_CONTROL_SRCUPDATE);
		bridge_channel_unsuspend(bridge_channel);
	}

	ao2_ref(bridge_channel, -1);
}

#else
static void generic_mute_unmute_user(struct confbridge_conference *conference, struct confbridge_user *user, int mute)
{
	/* Set user level mute request. */
	if (is_mic) {
		user->mic_muted = mute ? 1 : 0;
	} else {
		user->muted = mute ? 1 : 0;
	}

	conf_update_user_mute(user);

	ast_test_suite_event_notify("CONF_MUTE",
		"Message: participant %s %s %s\r\n"
		"Conference: %s\r\n"
		"Channel: %s",
		ast_channel_name(user->chan),
		user->muted ? "muted" : "unmuted",
		user->mic_muted ? "mic_muted" : "mic_unmuted",,
		conference->b_profile.name,
		ast_channel_name(user->chan));
	if (mute) {
		send_mute_event(user, conference);
	} else {
		send_unmute_event(user, conference);
	}
}
#endif

void conf_moh_stop(struct confbridge_user *user)
{
	user->playing_moh = 0;
	if (!user->suspended_moh) {
		int in_bridge;

		/*
		 * Locking the ast_bridge here is the only way to hold off the
		 * call to ast_bridge_join() in confbridge_exec() from
		 * interfering with the bridge and MOH operations here.
		 */
		ast_bridge_lock(user->conference->bridge);

		/*
		 * Temporarily suspend the user from the bridge so we have
		 * control to stop MOH if needed.
		 */
		in_bridge = !ast_bridge_suspend(user->conference->bridge, user->chan);
		ast_moh_stop(user->chan);
		if (in_bridge) {
#ifdef GRANDSTREAM_NETWORKS
			if (ast_channel_is_avs_engine(user->chan)) {
				ast_bridge_set_ignore_unsuspend(user->conference->bridge, user->chan, 0);
			}
#endif
			ast_bridge_unsuspend(user->conference->bridge, user->chan);
		}

		ast_bridge_unlock(user->conference->bridge);
	}
}

void conf_moh_start(struct confbridge_user *user)
{
	user->playing_moh = 1;
	if (!user->suspended_moh) {
		int in_bridge;

		/*
		 * Locking the ast_bridge here is the only way to hold off the
		 * call to ast_bridge_join() in confbridge_exec() from
		 * interfering with the bridge and MOH operations here.
		 */
		ast_bridge_lock(user->conference->bridge);

		/*
		 * Temporarily suspend the user from the bridge so we have
		 * control to start MOH if needed.
		 */
		in_bridge = !ast_bridge_suspend(user->conference->bridge, user->chan);
		ast_moh_start(user->chan, user->u_profile.moh_class, NULL);
		if (in_bridge) {
#ifdef GRANDSTREAM_NETWORKS
			if (ast_channel_is_avs_engine(user->chan)) {
				ast_bridge_set_ignore_unsuspend(user->conference->bridge, user->chan, 1);
			}
#endif
			ast_bridge_unsuspend(user->conference->bridge, user->chan);
		}

		ast_bridge_unlock(user->conference->bridge);
	}
}

/*!
 * \internal
 * \brief Unsuspend MOH for the conference user.
 *
 * \param user Conference user to unsuspend MOH on.
 *
 * \return Nothing
 */
static void conf_moh_unsuspend(struct confbridge_user *user)
{
	ao2_lock(user->conference);
	if (--user->suspended_moh == 0 && user->playing_moh) {
#ifdef GRANDSTREAM_NETWORKS
		/* Be sure no generators are left on it, otherwise MOH may be stopped after being started.
		ie. playing tone before here*/
		ast_deactivate_generator(user->chan);

		if (ast_channel_is_avs_engine(user->chan)) {
			if (!ast_bridge_suspend(user->conference->bridge, user->chan)) {
				ast_bridge_set_ignore_unsuspend(user->conference->bridge, user->chan, 1);
				ast_bridge_unsuspend(user->conference->bridge, user->chan);
			}
		}
#endif
		ast_moh_start(user->chan, user->u_profile.moh_class, NULL);
	}
	ao2_unlock(user->conference);
}

/*!
 * \internal
 * \brief Suspend MOH for the conference user.
 *
 * \param user Conference user to suspend MOH on.
 *
 * \return Nothing
 */
static void conf_moh_suspend(struct confbridge_user *user)
{
	ao2_lock(user->conference);
	if (user->suspended_moh++ == 0 && user->playing_moh) {
		ast_moh_stop(user->chan);
#ifdef GRANDSTREAM_NETWORKS
		if (ast_channel_is_avs_engine(user->chan)) {
			if (!ast_bridge_suspend(user->conference->bridge, user->chan)) {
				ast_bridge_set_ignore_unsuspend(user->conference->bridge, user->chan, 0);
				ast_bridge_unsuspend(user->conference->bridge, user->chan);
			}
		}
#endif
	}
	ao2_unlock(user->conference);
}

int conf_handle_inactive_waitmarked(struct confbridge_user *user)
{
	/* If we have not been quieted play back that they are waiting for the leader */
	if (!ast_test_flag(&user->u_profile, USER_OPT_QUIET) && play_prompt_to_user(user,
			conf_get_sound(CONF_SOUND_WAIT_FOR_LEADER, user->conference->b_profile.sounds))) {
		/* user hungup while the sound was playing */
		return -1;
	}
	return 0;
}

int conf_handle_only_person(struct confbridge_user *user)
{
#ifdef GRANDSTREAM_NETWORKS
	if (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		ast_set_flag(&user->u_profile, USER_OPT_NOONLYPERSON);
		ast_log(LOG_DEBUG, "Wave web and wave app sound file should not be played!");
	}
#endif
	/* If audio prompts have not been quieted or this prompt quieted play it on out */
	if (!ast_test_flag(&user->u_profile, USER_OPT_QUIET | USER_OPT_NOONLYPERSON)
		&& !ast_test_flag(ast_channel_flags(user->chan), AST_FLAG_RECOVERYING)) {
		if (play_prompt_to_user(user,
			conf_get_sound(CONF_SOUND_ONLY_PERSON, user->conference->b_profile.sounds))) {
			/* user hungup while the sound was playing */
			return -1;
		}
	}
	return 0;
}

int conf_add_post_join_action(struct confbridge_user *user, int (*func)(struct confbridge_user *user))
{
	struct post_join_action *action;
	if (!(action = ast_calloc(1, sizeof(*action)))) {
		return -1;
	}
	action->func = func;
	AST_LIST_INSERT_TAIL(&user->post_join_list, action, list);
	return 0;
}


void conf_handle_first_join(struct confbridge_conference *conference)
{
	ast_devstate_changed(AST_DEVICE_INUSE, AST_DEVSTATE_CACHABLE, "confbridge:%s", conference->name);
}

void conf_handle_second_active(struct confbridge_conference *conference)
{
	/* If we are the second participant we may need to stop music on hold on the first */
	struct confbridge_user *first_user = AST_LIST_FIRST(&conference->active_list);

	if (ast_test_flag(&first_user->u_profile, USER_OPT_MUSICONHOLD)) {
		conf_moh_stop(first_user);
	}
	conf_update_user_mute(first_user);
}

#ifdef GRANDSTREAM_NETWORKS
static void conf_clear_contact_group(struct confbridge_conference *conference)
{
	struct ao2_iterator iter;
	struct conf_contact_group *contact_group = NULL;

	if (!conference) {
		return;
	}

	ao2_lock(conference_contact_groups);
	iter = ao2_iterator_init(conference_contact_groups, 0);
	while ((contact_group = ao2_iterator_next(&iter))) {
		ao2_lock(contact_group);
		if (contact_group->inuse && !strcmp(contact_group->conference, conference->name)) {
			contact_group->inuse = 0;
			ast_string_field_set(contact_group, conference, "");
		}
		ao2_unlock(contact_group);
		ao2_ref(contact_group, -1);
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(conference_contact_groups);
}

static int get_conf_schedule_status(struct confbridge_conference *conference)
{
	int ret = 1;
	int inuse_flag = 0;
	char cmd[512]={0};
	struct ast_tm tm;
	char now_time_buf[AST_MAX_START_TIME] = { 0 };
	struct odbc_obj *conf_obj = NULL;

	if (!conference || ast_strlen_zero(conference->schedule_id_processing) || conference->schedule_hangup) {
		return ret;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return ret;
	}

	struct timeval now = ast_tvnow();
	if (ast_localtime(&now, &tm, NULL)) {
		ast_strftime(now_time_buf, sizeof(now_time_buf), "%Y-%m-%d %H:%M", &tm);
	}
	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		snprintf(cmd, sizeof(cmd), "SELECT COUNT(conf_number) AS count FROM %s.multimedia_conference_reservation WHERE conf_reservation_status='inuse' AND '%s'< end_time AND reservation_id='%s' AND conf_number='%s'", ast_strlen_zero(s_dbConfigName) ? "ucm_config" : s_dbConfigName, now_time_buf, conference->schedule_id_processing, conference->name);
	} else {
		snprintf(cmd, sizeof(cmd), "SELECT COUNT(confno) AS count FROM %s.meetme WHERE conf_reservation_status='inuse' AND '%s'<endtime AND bookid='%s' AND confno='%s'", ast_strlen_zero(s_dbConfigName) ? "ucm_config" : s_dbConfigName, now_time_buf, conference->schedule_id_processing, conference->name);
	}

	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256] = {0};
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			inuse_flag = atoi(tmp);
			if (inuse_flag) {
				ret = 0;
			}
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return ret;
}

static int conf_timeout_broadcast(struct confbridge_user *user)
{
	int res = 0;
	int ret  = 0;

	if (!user || !user->conference) {
		ast_log(LOG_WARNING, "user null!\n");
	}
	/* See if music on hold is playing */
	conf_moh_suspend(user);
	ast_bridge_suspend(user->conference->bridge, user->chan);

	if (!user->kicked) {
		res = ast_play_and_wait(user->chan, conf_get_sound(CONF_SOUND_EXTEND_TIME, user->b_profile.sounds));
		if (res < 0) {
			ast_log(LOG_WARNING, "Failed to playback file %s to channel %s\n", conf_get_sound(CONF_SOUND_EXTEND_TIME, user->b_profile.sounds), ast_channel_name(user->chan));
			goto out;
		}
	}

	ast_stopstream(user->chan);

	if (!res && !user->kicked) {
		res = ast_waitfordigit(user->chan, 5000);//5s
	}

	if (user->kicked) {
		goto out;
	}
	ast_log(LOG_DEBUG, "conf_timeout_broadcast play sound  end res: %d", res);

	if (res == '1') {
		ast_log(LOG_DEBUG, "press 1");
		ret = 1;
	} else if (res == 0) {
		ast_log(LOG_DEBUG, "nothing input");
		ret = 2;
	}else {
		ast_log(LOG_DEBUG, "other input");
		res = 3;
	}

out:

	ast_bridge_unsuspend(user->conference->bridge, user->chan);
	/* See if music on hold needs to be started back up again */
	conf_moh_unsuspend(user);
	return ret;
}

static void conf_timeout_detect_and_resend_notify(struct confbridge_conference *conference, struct confbridge_user *user)
{
	int conf_time_left = 0;
	if (!conference || !user) {
		ast_log(LOG_DEBUG, "conference or user not exist!");
		return;
	}
	/*Avoid the conf_sched lock and the conference lock cross each other, resulting in deadlock*/
	conf_time_left = ast_sched_when(conf_sched,  conference->conf_sched_id);

	ao2_lock(conference);
	if (conference->conf_will_timeout && ast_test_flag(&user->u_profile, USER_OPT_ADMIN) && conference->extend_time_status == EXTEND_PENDING) {
		if (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			ast_log(LOG_DEBUG, "Conf timeout detect,WAVE and APP resend Notify!");
			conf_timeout_notify_build_send(user->chan, conference->name, 1, EXTEND_PENDING, conf_time_left + (conference->sch_time * (CONF_PEER_TIMEOUT/1000)));
		} else {
			ast_log(LOG_DEBUG, "Terminal type is Phone,do nothing!");
		}
	}
	ao2_unlock(conference);
}

static int conference_timeout_notify(const void *data)
{
	int ret = 2;  /*Don't have admin user,do as nothing input*/
	int delay_time = 0;
	int extra_time = 0;
	ast_log(LOG_DEBUG, "conference_timeout_notify entry!\n");
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;

	char *conference_name = (char *)data;
	if (ast_strlen_zero(conference_name)) {
		ast_log(LOG_ERROR,"conference_name null!");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		ast_log(LOG_ERROR,"can not find conference!");
		return 0;
	}
	ao2_lock(conference);
	conference->conf_will_timeout = 1;
	conference->extend_time_status = EXTEND_PENDING;
	if (conference->sch_time <= 0) {
		ast_log(LOG_DEBUG,"sch_time reach, hangup_conference!");
		hangup_conference(conference, CONF_HANGUP_CONFTIMEOUT);
		ao2_unlock(conference);
		return 0;
	}
	struct timeval now_tv = {0,0};
	struct timeval delay_tv = {0,0};
	now_tv = ast_tvnow();
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
			if (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
				ast_log(LOG_DEBUG, "WAVE and APP only send Notify!");
				if (!ast_tvzero(conference->schedule_alert_time) &&
					!ast_strlen_zero(conference->schedule_id_to_start) &&
					(ast_strlen_zero(conference->schedule_id_processing) ||
					(!ast_strlen_zero(conference->schedule_id_processing) &&
					strcasecmp(conference->schedule_id_to_start, conference->schedule_id_processing)))) {
					ast_log(LOG_DEBUG, "Schedule alert happend, don't send timeout Notify!");
				} else {
					conf_timeout_notify_build_send(user->chan, conference->name, 1, conference->extend_time_status, (conference->sch_time*CONF_PEER_TIMEOUT)/1000);
				}
				ret = 2;
				conference->last_notify = ast_tvnow();
			} else {
				ret = conf_timeout_broadcast(user);
			}
		}
	}
	delay_tv = ast_tvnow();
	/*extra time for play file and get digital.*/
	extra_time = ast_tvdiff_ms(delay_tv, now_tv);
	ast_log(LOG_DEBUG, "extra_time %d", extra_time);
	if (ret == 1) {
		/*extend the meeting*/
		delay_time = conference->b_profile.conf_timeout - (CONF_SCH_RETRY - conference->sch_time)*CONF_PEER_TIMEOUT;
		conference->sch_time = CONF_SCH_RETRY;
		conference->conf_will_timeout = 0;
		conference->extend_time_status = EXTEND_CONFIRMED;
	} else if (ret == 2) {
		/*nothing input*/
		delay_time = CONF_PEER_TIMEOUT;
		conference->sch_time--;
	} else {
		/*other input,stop next sched task*/
		delay_time = CONF_PEER_TIMEOUT *conference->sch_time;
		conference->sch_time = 1;
		conference->sch_time--;
		conference->extend_time_status = EXTEND_REJECTED;
	}
	ao2_unlock(conference);
	if (delay_time - extra_time > 0) {
		delay_time = delay_time - extra_time;
	}
	if (delay_time > 0) {
		ast_log(LOG_DEBUG, "conference_timeout_notify next delay_time %d conference->sch_time %d\n", delay_time, conference->sch_time);
		if ((conference->conf_sched_id = ast_sched_add(conf_sched, delay_time, conference_timeout_notify, conference->name)) < 0) {
			ast_log(LOG_ERROR, "Add scheduler timer failure!\n");
			return 0;
		}
	}

	return 0;
}

static int get_schedule_timeout(struct confbridge_conference *conference)
{
	int extra_time = 0;

	if (!conference) {
		ast_log(LOG_ERROR, "get_schedule_timeout conference null!");
		return 0;
	}
	if (ast_tvzero(conference->schedule_start_timeval)) {
		ast_log(LOG_DEBUG, "schedule start time null, may be instant|nway|chat !");
		return 0;
	}
	struct timeval now_tv = {0,0};
	now_tv = ast_tvnow();

	extra_time = ast_tvdiff_ms(now_tv, conference->schedule_start_timeval);
	/*early join max time is 10min*/
	if (extra_time < 0 && -extra_time > CONF_EARLY_JOIN_MAX_TIME) {
		extra_time = -CONF_EARLY_JOIN_MAX_TIME;
	}

	return extra_time;
}

static int wait_for_conference_end(struct ast_channel *chan, const char *conference_name)
{
	ast_log(LOG_DEBUG, "wait for conference end!");

	int res = 0;
	struct confbridge_conference *conference;
	time_t expire = time(NULL) + 60;
	if (!chan || ast_strlen_zero(conference_name)) {
		ast_log(LOG_DEBUG, "wait_for_conference_end chan or conference_name null!");
		return res;
	}

	/* This is the holding pen for callers*/
	for (;;) {
		if (chan && ast_check_hangup(chan)) {
			ast_log(LOG_DEBUG, "chan hangup,exist!");
			res = 1;
			break;
		}
		if (time(NULL) >= expire) {
			ast_log(LOG_DEBUG, "exit waiting,wait timeout!");
			res = 1;
			break;
		}
		ao2_lock(conference_bridges);
		if (!(conference = ao2_find(conference_bridges, conference_name, OBJ_KEY))) {
			ast_log(LOG_DEBUG, "last confbridge release, let's join conference!");
			ao2_unlock(conference_bridges);
			break;
		} else {
			ao2_ref(conference, -1);
		}
		ao2_unlock(conference_bridges);
		sleep(1);
	}

	return res;
}

static int check_random_schedule_conf_end(const char *conf_name, struct ast_channel *chan)
{
	int ret = 0;
	char family[64] = {0};
	char random_conf_start[8] = {0};
	const char *conf_type = NULL;
	const char *schedule_id = NULL;
	if (!conf_name || !chan) {
		goto end_reslut;
	}
	conf_type = pbx_builtin_getvar_helper(chan, "CONF_TYPE");
	schedule_id = pbx_builtin_getvar_helper(chan, "schedule_id");
	if (!ast_strlen_zero(conf_type) && !ast_strlen_zero(schedule_id) && !strcasecmp(conf_type, "random_schedule")) {
		snprintf(family, sizeof(family)-1, "SCHEDULE/%s", conf_name);
		ast_db_get(family, "CONF_START", random_conf_start, sizeof(random_conf_start));
		if (ast_strlen_zero(random_conf_start)) {
			ret = 1;
			goto end_reslut;
		}
	}
end_reslut:
	ast_log(LOG_DEBUG, "check_random_schedule_conf_end %d", ret);
	return ret;
}

/*Now confbridge doesn't need to handle an attended transfer event*/
#if 0
/*!
 * \brief Handle an attended transfer event
 *
 * This event is important in order to be able to update the cei information.
 *
 * \param userdata Data pertaining to the particular confbridge data.
 * \param sub The stasis subscription on which the message occurred.
 * \param topic The topic for this event.
 * \param msg The stasis message for the attended transfer event.
 */
static void confbridge_handle_attended_transfer(void *userdata, struct stasis_subscription *sub,
		struct stasis_message *msg)
{
	struct confbridge_conference *conference = userdata;
	struct ast_attended_transfer_message *atxfer_msg = stasis_message_data(msg);
	struct confbridge_user *user = NULL;

	if (atxfer_msg->result != AST_BRIDGE_TRANSFER_SUCCESS ||
			atxfer_msg->dest_type == AST_ATTENDED_TRANSFER_DEST_THREEWAY) {
		return;
	}

	ast_log(LOG_DEBUG, "atxfer_msg->transferee->name:%s, atxfer_msg->target->name:%s, "
					   "atxfer_msg->to_transferee.channel_snapshot->name:%s, atxfer_msg->to_transfer_target.channel_snapshot->name:%s",
			atxfer_msg->transferee ? atxfer_msg->transferee->name : "null",
			atxfer_msg->target ? atxfer_msg->target->name : "null",
			atxfer_msg->to_transferee.channel_snapshot ? atxfer_msg->to_transferee.channel_snapshot->name : "null",
			atxfer_msg->to_transfer_target.channel_snapshot ? atxfer_msg->to_transfer_target.channel_snapshot->name : "null");

	ast_log(LOG_DEBUG, "atxfer_msg->to_transfer_target.channel_snapshot->bridgeid:%s, conference->bridge->uniqueid:%s",
		atxfer_msg->to_transfer_target.channel_snapshot ? atxfer_msg->to_transfer_target.channel_snapshot->bridgeid : "null",
		conference->bridge->uniqueid);

	if ((!atxfer_msg->to_transfer_target.channel_snapshot) || (!atxfer_msg->transferee) ||
		strcmp(conference->bridge->uniqueid, atxfer_msg->to_transfer_target.channel_snapshot->bridgeid)) {
		return;
	}
}

static int confbridge_setup_bridge_stasis_subs(struct confbridge_conference *conference)
{
	if (!conference) {
		return -1;
	}

	conference->bridge_router = stasis_message_router_create(ast_bridge_topic_all_cached());
	if (!conference->bridge_router) {
		return -1;
	}

	if (stasis_message_router_add(conference->bridge_router, ast_attended_transfer_type(),
		confbridge_handle_attended_transfer, conference)) {
		stasis_message_router_unsubscribe_and_join(conference->bridge_router);
		conference->bridge_router = NULL;
		return -1;
	}

	return 0;
}
#endif
static int release_schedule_meeting(const void *data)
{
	if (!data) {
		return -1;
	}
	ast_log(LOG_DEBUG, "release schedule meeting!");
	char *conference_name = (char *)data;
	clean_schedule_meeting_status(conference_name, -1);
	return 0;
}

static void schedule_meeting_release_detect(struct confbridge_conference *conference)
{
	struct schedule_conference *schedule_conf = NULL;
	if(!conference) {
		return;
	}
	if (conference->schedule_hangup) {
		ast_log(LOG_ERROR, "schedule hangup should not add scheduler timer.\n");
	}
	schedule_conf = schedule_conf_cli_fetch(conference->name);
	if (schedule_conf) {
		if ((schedule_conf->recycle_sched_id = ast_sched_add(recycle_sched, CONF_SCH_RECYCLE_TIMEOUT, release_schedule_meeting, schedule_conf->name)) < 0) {
			ast_log(LOG_ERROR, "Unable to add scheduler timer.\n");
		} else {
			ast_log(LOG_DEBUG, "create schedule conf sched success! user_sch_timeout %d, recycle_sched_id %d!", CONF_SCH_RECYCLE_TIMEOUT, schedule_conf->recycle_sched_id);
		}
		ao2_ref(schedule_conf, -1);
	}

}

/* The schedule meeting astdb information has not been set yet, and the actual time has arrived */
static void detect_meeting_start(struct confbridge_conference *conference)
{
	struct tm start_tm = {0};
	struct timeval start_tv = {0};
	struct timeval now_tv = {0};
	int64_t countdown = 0;

	if (!conference) {
		return;
	}
	if (ast_strlen_zero(conference->schedule_id_processing) && !ast_strlen_zero(conference->schedule_id_to_start) && !ast_strlen_zero(conference->schedule_start_time)) {
		if (!conference->b_profile.is_random_conf && !conference->b_profile.n_way_conf) {
			strptime(conference->schedule_start_time, "%Y-%m-%dT%H:%M:%S%z", &start_tm);
			start_tv.tv_sec = mktime(&start_tm);
			now_tv = ast_tvnow();
			countdown = ast_tvdiff_sec(start_tv, now_tv);
			if (countdown <= 0) {
				ast_log(LOG_DEBUG, "conference schedule %s at %s start!", conference->schedule_id_to_start, conference->schedule_start_time);
				schedule_conference_wait_inuse_notify(conference, conference->schedule_id_to_start);
			}
		}
	}
}
#endif

void conf_ended(struct confbridge_conference *conference)
{
#ifdef GRANDSTREAM_NETWORKS
	schedule_meeting_release_detect(conference);

	/*Clear device state stasis added in conf_handle_first_join*/
	char device[64] = {0};
	snprintf(device, sizeof(device), "confbridge:%s", conference->name);
	ast_device_state_clear_cache(device);
	stasis_topic_pool_delete_topic(ast_device_state_topic_pool(), device);

	ast_log(LOG_DEBUG, "conf_ended entry [%s]!", conference->name);
	if (!conference->conf_end) {
		conference->conf_end = 1;
	}
#endif
	struct pbx_find_info q = { .stacklen = 0 };

	/* Called with a reference to conference */
	ao2_unlink(conference_bridges, conference);
#ifdef GRANDSTREAM_NETWORKS
	ao2_lock(conference);
	conf_stop_record(conference);
	conf_stop_video_record(conference);
	conf_stop_slides_record(conference);
	conf_stop_focus_record(conference);
	ao2_unlock(conference);
#endif
	send_conf_end_event(conference);
	if (!ast_strlen_zero(conference->b_profile.regcontext) &&
			pbx_find_extension(NULL, NULL, &q, conference->b_profile.regcontext,
				conference->name, 1, NULL, "", E_MATCH)) {
		ast_context_remove_extension(conference->b_profile.regcontext,
				conference->name, 1, NULL);
	}
#ifndef GRANDSTREAM_NETWORKS
	ao2_lock(conference);
	conf_stop_record(conference);
	ao2_unlock(conference);
#endif

#ifdef GRANDSTREAM_NETWORKS
	conf_clear_contact_group(conference);
	if (conference->is_conference_service) {
		ao2_unlock(conference);
		cache_conf_remove(conference->name);
		ao2_lock(conference);
		/*End the schedule conference which has been delayed*/
		schedule_conference_end_delay(conference);
		im_conf_list_remove(conference->name);
		/*user in waiting list can join now*/
		if (get_conf_schedule_status(conference)) {
			schedule_conference_wait_user_join(conference->name);
		}
	}
	if (conference->conf_sched_id) {
		int ret = ast_sched_del(conf_sched, conference->conf_sched_id);
		ast_log(LOG_DEBUG, "conf ended,ast sched del ret %d", ret);
	}
#endif
}

/*!
 * \internal
 * \brief Allocate playback channel for a conference.
 * \pre expects conference to be locked before calling this function
 */
static int alloc_playback_chan(struct confbridge_conference *conference)
{
	struct ast_format_cap *cap;
	char taskprocessor_name[AST_TASKPROCESSOR_MAX_NAME + 1];

	cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!cap) {
		return -1;
	}
	ast_format_cap_append(cap, ast_format_slin, 0);
	conference->playback_chan = ast_request("CBAnn", cap, NULL, NULL,
		conference->name, NULL);
	ao2_ref(cap, -1);
	if (!conference->playback_chan) {
		return -1;
	}

	/* To make sure playback_chan has the same language as the bridge */
	ast_channel_lock(conference->playback_chan);
	ast_channel_language_set(conference->playback_chan, conference->b_profile.language);
	ast_channel_unlock(conference->playback_chan);

	ast_debug(1, "Created announcer channel '%s' to conference bridge '%s'\n",
		ast_channel_name(conference->playback_chan), conference->name);

	ast_taskprocessor_build_name(taskprocessor_name, sizeof(taskprocessor_name),
		"Confbridge/%s", conference->name);
	conference->playback_queue = ast_taskprocessor_get(taskprocessor_name, TPS_REF_DEFAULT);
	if (!conference->playback_queue) {
		ast_hangup(conference->playback_chan);
		conference->playback_chan = NULL;
		return -1;
	}
	return 0;
}

/*!
 * \brief Push the announcer channel into the bridge
 *
 * \param conference Conference bridge to push the announcer to
 * \retval 0 Success
 * \retval -1 Failed to push the channel to the bridge
 */
static int push_announcer(struct confbridge_conference *conference)
{
	if (conf_announce_channel_push(conference->playback_chan)) {
		ast_hangup(conference->playback_chan);
		conference->playback_chan = NULL;
		return -1;
	}

	ast_autoservice_start(conference->playback_chan);
	return 0;
}

/*!
 * \brief Join a conference bridge
 *
 * \param conference_name The conference name
 * \param user Conference bridge user structure
 *
 * \return A pointer to the conference bridge struct, or NULL if the conference room wasn't found.
 */
static struct confbridge_conference *join_conference_bridge(const char *conference_name, struct confbridge_user *user)
{
	struct confbridge_conference *conference;
	struct post_join_action *action;
	int max_members_reached = 0;
#ifdef GRANDSTREAM_NETWORKS
	struct confbridge_user *conf_tone_user	= NULL;
	struct ast_channel   *chan_user = NULL ;
	int in_bridge = 0;
	int is_call_flip = 0;
	char result[32] = {0};
	int ami_send_status = 0;
	int ignore_paging_callee_record=0;
	int is_paging_conf =0;
	const char *page_intercom =NULL;
	const char *is_callee =NULL;
	const char *record_prompt =NULL;
#endif

	/* We explictly lock the conference bridges container ourselves so that other callers can not create duplicate conferences at the same */
	ao2_lock(conference_bridges);

	ast_debug(1, "Trying to find conference bridge '%s'\n", conference_name);

	/* Attempt to find an existing conference bridge */
	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (conference && conference->b_profile.max_members) {
		max_members_reached = conference->b_profile.max_members > conference->activeusers ? 0 : 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (check_random_schedule_conf_end(conference_name, user->chan)) {
		ast_log(LOG_DEBUG, "random schedule conf end %s!", conference_name);
		ao2_unlock(conference_bridges);
		if (conference) {
			ao2_ref(conference, -1);
		}
		return NULL;
	}
	/*Start added by zyxia for bug 202118,when lock the conf,call flip failed,2021.8.26*/
	const char* from_call_flip = pbx_builtin_getvar_helper(user->chan, "FROM_CALL_FLIP");//added by zyxia for Bug 202171,call flip in conference need send notify
	if (!ast_strlen_zero(from_call_flip) && strcasecmp(from_call_flip, "yes") == 0) {
		is_call_flip = 1;
	}
	/* When finding a conference bridge that already exists make sure that it is not locked, and if so that we are not an admin */
	if (conference && (max_members_reached || (conference->locked && !is_call_flip)) && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		ao2_unlock(conference_bridges);

		if ((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
		}

		ast_debug(1, "Conference '%s' is locked and caller is not an admin\n", conference_name);
		ast_stream_and_wait(user->chan,
			conf_get_sound(CONF_SOUND_LOCKED, conference->b_profile.sounds),
			"");
		ao2_ref(conference, -1);
		return NULL;
	}
#else
	if (conference && (max_members_reached || conference->locked) && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		ao2_unlock(conference_bridges);
		ast_debug(1, "Conference '%s' is locked and caller is not an admin\n", conference_name);
		ast_stream_and_wait(user->chan,
			conf_get_sound(CONF_SOUND_LOCKED, conference->b_profile.sounds),
			"");
		ao2_ref(conference, -1);
		return NULL;
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	/*conference hangup, let's wait for conference resource release!*/
	if (conference && conference->conf_end) {
		ao2_unlock(conference_bridges);
		if (wait_for_conference_end(user->chan, conference->name)) {
			ast_debug(LOG_WARNING, "conference %s resource busy can't join now!", conference->name);
			ao2_ref(conference, -1);
			return NULL;
		} else {
			ao2_ref(conference, -1);
			conference = NULL;
		}
		ao2_lock(conference_bridges);
		/*must to find it again. There may be multiple users waiting.
		The first one creates a new bridge, and the subsequent ones are not needed */
		conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	}

	/*Start added by zyxia for Bug 235220 ,2022.8.31,callee channel dont record*/
	page_intercom = pbx_builtin_getvar_helper(user->chan, "PAGE_INTERCOM");
	if (!ast_strlen_zero(page_intercom) && !strcmp(page_intercom, "yes")) {
		is_callee =pbx_builtin_getvar_helper(user->chan, "IS_CALLEE");
		ast_log(LOG_DEBUG, " channel_name %s,is_callee:%s", ast_channel_name(user->chan),is_callee ? is_callee:"");
		if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
			ignore_paging_callee_record =1;
		}
		is_paging_conf =1;
		record_prompt = pbx_builtin_getvar_helper(NULL, "RECORD_PROMPT");

	}
	/*End added by zyxia for Bug 235220 ,2022.8.31*/

#endif


	/* If no conference bridge was found see if we can create one */
	if (!conference) {
		/* Try to allocate memory for a new conference bridge, if we fail... this won't end well. */
		if (!(conference = ao2_alloc(sizeof(*conference), destroy_conference_bridge))) {
			ao2_unlock(conference_bridges);
			ast_log(LOG_ERROR, "Conference '%s' could not be created.\n", conference_name);
			return NULL;
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Setup lock for playback channel */
		ast_mutex_init(&conference->playback_lock);

		conference->recorded = CONF_RECORD_INIT;
		ast_copy_string(conference->conf_layout, DEFAULT_CONF_LAYOUT, sizeof(conference->conf_layout));

		conference->record_files = ast_str_create(RECORD_FILES_INITIAL_SPACE);
		if (!conference->record_files) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}
		conference->record_files_desc = ast_str_create(RECORD_FILES_DESC_INITIAL_SPACE);
		if (!conference->record_files_desc) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}
		conference->record_files_uuid = ast_str_create(RECORD_FILES_UUID_INITIAL_SPACE);
		if (!conference->record_files_uuid) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}

		conference->video_filename = ast_str_create(RECORD_FILES_INITIAL_SPACE);
		if (!conference->video_filename) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}
		conference->video_files = ast_str_create(RECORD_FILES_INITIAL_SPACE);
		if (!conference->video_files) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}
		conference->video_files_desc = ast_str_create(RECORD_FILES_DESC_INITIAL_SPACE);
		if (!conference->video_files_desc) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}
		conference->video_files_uuid = ast_str_create(RECORD_FILES_DESC_INITIAL_SPACE);
		if (!conference->video_files_uuid) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}
#endif

		/* Setup for the record channel */
		conference->record_filename = ast_str_create(RECORD_FILENAME_INITIAL_SPACE);
		if (!conference->record_filename) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			return NULL;
		}

		/* Setup conference bridge parameters */
		ast_copy_string(conference->name, conference_name, sizeof(conference->name));
		conf_bridge_profile_copy(&conference->b_profile, &user->b_profile);
		/* Create an actual bridge that will do the audio mixing */
#ifdef GRANDSTREAM_NETWORKS
		conference->bridge = ast_bridge_base_new_gs(AST_BRIDGE_CAPABILITY_MULTIMIX,
			AST_BRIDGE_FLAG_MASQUERADE_ONLY | AST_BRIDGE_FLAG_TRANSFER_BRIDGE_ONLY,
			app, conference_name, NULL, ast_channel_linkedid(user->chan));
#else
		conference->bridge = ast_bridge_base_new(AST_BRIDGE_CAPABILITY_MULTIMIX,
			AST_BRIDGE_FLAG_MASQUERADE_ONLY | AST_BRIDGE_FLAG_TRANSFER_BRIDGE_ONLY,
			app, conference_name, NULL);
#endif
		if (!conference->bridge) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			ast_log(LOG_ERROR, "Conference '%s' mixing bridge could not be created.\n", conference_name);
			return NULL;
		}

		/* Set the internal sample rate on the bridge from the bridge profile */
		ast_bridge_set_internal_sample_rate(conference->bridge, conference->b_profile.internal_sample_rate);
		/* Set the internal mixing interval on the bridge from the bridge profile */
		ast_bridge_set_mixing_interval(conference->bridge, conference->b_profile.mix_interval);
		ast_bridge_set_binaural_active(conference->bridge, ast_test_flag(&conference->b_profile, BRIDGE_OPT_BINAURAL_ACTIVE));

		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_FOLLOW_TALKER)) {
			ast_bridge_set_talker_src_video_mode(conference->bridge);
		} else if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			ast_bridge_set_sfu_video_mode(conference->bridge);
			ast_bridge_set_video_update_discard(conference->bridge, conference->b_profile.video_update_discard);
			ast_bridge_set_remb_send_interval(conference->bridge, conference->b_profile.remb_send_interval);
			if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_REMB_BEHAVIOR_AVERAGE)) {
				ast_brige_set_remb_behavior(conference->bridge, AST_BRIDGE_VIDEO_SFU_REMB_AVERAGE);
			} else if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_REMB_BEHAVIOR_LOWEST)) {
				ast_brige_set_remb_behavior(conference->bridge, AST_BRIDGE_VIDEO_SFU_REMB_LOWEST);
			} else if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_REMB_BEHAVIOR_HIGHEST)) {
				ast_brige_set_remb_behavior(conference->bridge, AST_BRIDGE_VIDEO_SFU_REMB_HIGHEST);
			}
		}

		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_ENABLE_EVENTS)) {
			ast_bridge_set_send_sdp_label(conference->bridge, 1);
		}

#ifdef GRANDSTREAM_NETWORKS
		if (ast_channel_is_avs_engine(user->chan)) {
			if (ast_channel_callerintercom_get(user->chan)) {
				ast_log(LOG_DEBUG, "is private intercom paging");
				int bridge_id = ast_avs_create_bridge(avs_bridge_private, conference->bridge, -1, 0, ast_channel_get_avs_address(user->chan));
				if (bridge_id > 0) {
					conference->bridge->avs_bridge_id = bridge_id;
					conference->bridge->avs_bridge_type = avs_bridge_private;
					if (conference->bridge->technology && conference->bridge->technology->stop) {
						ast_bridge_lock(conference->bridge);
						conference->bridge->technology->stop(conference->bridge);
						ast_bridge_unlock(conference->bridge);
					}
				}
			} else {
				int bridge_id = ast_avs_create_bridge(avs_bridge_conf, conference->bridge, -1, 0, ast_channel_get_avs_address(user->chan));
				if (bridge_id > 0) {
					conference->bridge->avs_bridge_id = bridge_id;
					conference->bridge->avs_bridge_type = avs_bridge_conf;
					if (conference->bridge->technology && conference->bridge->technology->stop) {
						ast_bridge_lock(conference->bridge);
						conference->bridge->technology->stop(conference->bridge);
						ast_bridge_unlock(conference->bridge);
					}
				}
			}
		}
#endif

		/* Link it into the conference bridges container */
		if (!ao2_link(conference_bridges, conference)) {
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			ast_log(LOG_ERROR,
				"Conference '%s' could not be added to the conferences list.\n", conference_name);
			return NULL;
		}

		/* Set the initial state to EMPTY */
		conference->state = CONF_STATE_EMPTY;

		if (alloc_playback_chan(conference)) {
			ao2_unlink(conference_bridges, conference);
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			ast_log(LOG_ERROR, "Could not allocate announcer channel for conference '%s'\n", conference_name);
			return NULL;
		}

		if (push_announcer(conference)) {
			ao2_unlink(conference_bridges, conference);
			ao2_ref(conference, -1);
			ao2_unlock(conference_bridges);
			ast_log(LOG_ERROR, "Could not add announcer channel for conference '%s' bridge\n", conference_name);
			return NULL;
		}

#ifdef GRANDSTREAM_NETWORKS
		ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
		if (strlen(result) > 0) {
			ami_send_status = atoi(result);
		}

#ifdef BUILD_ON_CLOUD_PBX
		int gdms_record_support = 0;
		memset(result, 0, sizeof(result));
		ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
		if (strlen(result) > 0) {
			gdms_record_support = atoi(result);
		}
#endif // BUILD_ON_CLOUD_PBX

		if (is_paging_conf) {
			conference->is_paging_conf = 1;
		}

		/*if the channel is paging callee,dont record*/
		if(ignore_paging_callee_record){
			ast_clear_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE);
			ast_clear_flag(&user->b_profile, BRIDGE_OPT_RECORD_CONFERENCE);
			ast_clear_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE);
			ast_clear_flag(&user->b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE);
			ast_clear_flag(&conference->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE);
			ast_clear_flag(&user->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE);
		}

#ifdef BUILD_ON_CLOUD_PBX
		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE) && !ignore_paging_callee_record && !(ami_send_status & SEND_STATUS_RECORDING) && gdms_record_support) {
#else
		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE) && !ignore_paging_callee_record && !(ami_send_status & SEND_STATUS_RECORDING) ) {
#endif // BUILD_ON_CLOUD_PBX
			/*prompt to  page caller at here ,callee prompt record play in function page_state_callback at app_page.c */
			if (conference->is_paging_conf) {
				if (!ast_strlen_zero(record_prompt) && !strcmp(record_prompt, "yes") ){
					ast_channel_check_need_prompt_start_stop_record(user->chan, NULL, 1);
					ast_log(LOG_DEBUG, "[debug]play prompt to  chan:%s",ast_channel_name(user->chan));
				}
			}

			ao2_lock(conference);
			conf_start_record(conference, user->chan);
			ao2_unlock(conference);
		}
#else
		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE)) {
			ao2_lock(conference);
			conf_start_record(conference);
			ao2_unlock(conference);
		}
#endif
#ifdef GRANDSTREAM_NETWORKS
		if (conference->is_paging_conf && !ignore_paging_callee_record && ast_test_flag(&conference->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE)) {
			ast_log(LOG_DEBUG, "[debug]start video confrecord,%s",user->b_profile.rec_file);
			ast_clear_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);
			strcpy(conference->b_profile.rec_file, user->b_profile.rec_file);
			const char *mix_filename = pbx_builtin_getvar_helper(user->chan, "MIXFILENAME");
			ast_str_reset(conference->video_filename);
			ast_str_set(&conference->video_filename, 0, "%s",user->b_profile.rec_file);

			if (mix_filename) {
				char record_files[1024];
				snprintf(record_files,sizeof(record_files),"%s.mkv@",mix_filename);
				ast_cdr_setvar(ast_channel_name(user->chan), "RECORD_FILES", record_files);
				if (!ast_strlen_zero(record_files)) {
					ast_log(LOG_DEBUG, "caller:%s cdr-record:%s",
						ast_channel_name(user->chan), record_files);
				}
			}
		}

		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE)) {
			ao2_lock(conference);
			conf_start_video_record(conference);
			ao2_unlock(conference);
		}

		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_ENABLE_MCU)) {
			ao2_lock(conference);
			conf_enable_mcu(conference);
			ao2_unlock(conference);
		}

		if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_SLIDES_RECORD_CONFERENCE)) {
			ao2_lock(conference);
			conf_start_slides_record(conference);
			ao2_unlock(conference);
		}

		if (!ignore_paging_callee_record && ast_test_flag(&conference->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE)) {
			ao2_lock(conference);
			conf_start_focus_record(conference);

			if(conference->is_paging_conf)
				ast_channel_set_is_page_focus_record(user->chan, 1);
			ao2_unlock(conference);

			if (strlen(ast_str_buffer(conference->video_files_uuid))) {
				char record_files_uuid[1024] = "";
				snprintf(record_files_uuid, sizeof(record_files_uuid), "%s", ast_str_buffer(conference->video_files_uuid));
				ast_cdr_setvar(ast_channel_name(user->chan), "record_files_uuid", record_files_uuid);
				ast_log(LOG_DEBUG, "[join_conference_bridge] Set channel:%s's cdr record_files_uuid:%s", ast_channel_name(user->chan), record_files_uuid);
			}
		}

		if (!ast_strlen_zero(pbx_builtin_getvar_helper(user->chan, "CONF_ROOM")) ||
			!ast_strlen_zero(pbx_builtin_getvar_helper(user->chan, "MCB_HASH"))) {
			conference->is_conference_service = 1;
		} else {
			conference->is_conference_service = 0;
		}
		conference->allow_user_unmute = ast_test_flag(&conference->b_profile, BRIDGE_OPT_ALLOW_USER_UNMUTE) ? 1 : 0;
		conference->user_can_invite = ast_true(pbx_builtin_getvar_helper(user->chan, "USER_CAN_INVITE")) ? 1 : 0;

		if (conference->is_conference_service || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			get_schedule_conference(conference);
			/*set public conference name as subject*/
			if (ast_strlen_zero(conference->schedule_id_processing)) {
				const char *public_conf_name = pbx_builtin_getvar_helper(user->chan, "CONF_NAME");
				ast_log(LOG_DEBUG, "public_conf_name %s", public_conf_name);
				if (!ast_strlen_zero(public_conf_name)) {
					if (conference->schedule_subject) {
						ast_free(conference->schedule_subject);
						conference->schedule_subject = NULL;
					}
					conference->schedule_subject = ast_strdup(public_conf_name);
					ast_log(LOG_DEBUG, "set schedule_subject for public conference %s.\n", public_conf_name);
				}
			}
			const char *conf_creator = pbx_builtin_getvar_helper(user->chan, "CONF_CREATOR");
			if (!ast_strlen_zero(conf_creator)) {
				if (conference->conf_creator) {
					ast_free(conference->conf_creator);
					conference->conf_creator = NULL;
				}
				conference->conf_creator = ast_strdup(conf_creator);
				ast_log(LOG_DEBUG, "set conf_creator for conference %s.\n", conf_creator);
			}
			/*schedule conference start early.notify update status*/
			/*need notify first,then get schedule info*/
			const char *inuse_notify = pbx_builtin_getvar_helper(user->chan, "INUSE_NOTIFY");
			if (!ast_strlen_zero(inuse_notify) && !strcasecmp(inuse_notify, "yes")) {
				const char *wait_reservation_id = pbx_builtin_getvar_helper(user->chan, "WAIT_RESERVATION_ID");
				schedule_conference_wait_inuse_notify(conference, wait_reservation_id);
			} else {
				detect_meeting_start(conference);
				schedule_conference_inuse_notify(conference);
			}
			conf_create_operation_feature_id(conference);
		}
		im_conf_list_update_status(conference->name, CONF_IM_INUSE);

		/*conference timeout ast sched notify for SFU.Added by gangwang*/
		int conf_sch_timeout = -1;
		if (!ast_strlen_zero(conference->schedule_id_processing) && ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			int extra_time = get_schedule_timeout(conference);
			if (conference->b_profile.conf_timeout && (conference->b_profile.conf_timeout - extra_time) > 0) {
				conf_sch_timeout = conference->b_profile.conf_timeout - extra_time;
				/*Need to broadcast three times*/
				if (conf_sch_timeout > CONF_PEER_TIMEOUT*3) {
					conf_sch_timeout = conf_sch_timeout - CONF_PEER_TIMEOUT*3;
					conference->sch_time = CONF_SCH_RETRY;
				} else {
					/*After joining the meeting, the remain time is not enough to broadcast three times*/
					conference->sch_time = conf_sch_timeout/CONF_PEER_TIMEOUT;
					conf_sch_timeout = conf_sch_timeout%CONF_PEER_TIMEOUT;
				}
			}
			ast_log(LOG_DEBUG, "schedule_id_processing %s extra_time is %d.\n", conference->schedule_id_processing, extra_time);
		} else {
			conf_sch_timeout = conference->b_profile.conf_timeout - CONF_PEER_TIMEOUT*3;
			conference->sch_time = CONF_SCH_RETRY;
		}
		if (conf_sch_timeout >= 0) {
			conference->extend_time_status = EXTEND_INIT;
			if ((conference->conf_sched_id = ast_sched_add(conf_sched, conf_sch_timeout, conference_timeout_notify, conference->name)) < 0) {
				ast_log(LOG_ERROR, "Unable to add scheduler timer.\n");
			} else {
				ast_log(LOG_DEBUG, "conf sch time left %ld!!", ast_sched_when(conf_sched,  conference->conf_sched_id) + (conference->sch_time * (CONF_PEER_TIMEOUT/1000)));
				ast_log(LOG_DEBUG,"create conf sched success! conf_sch_timeout %d, conference->sch_time %d!", conf_sch_timeout, conference->sch_time);
			}
		}

		/*Now confbridge doesn't need to handle an attended transfer event*/
		//confbridge_setup_bridge_stasis_subs(conference);
#endif
		send_conf_start_event(conference);

		if (!ast_strlen_zero(conference->b_profile.regcontext)) {
			if (!ast_exists_extension(NULL, conference->b_profile.regcontext, conference->name, 1, NULL)) {
				ast_add_extension(conference->b_profile.regcontext, 1, conference->name, 1, NULL, NULL, "Noop", NULL, NULL, "ConfBridge");
			}
		}

		ast_debug(1, "Created conference '%s' and linked to container.\n", conference_name);
	}

	ao2_unlock(conference_bridges);

	/* Setup conference bridge user parameters */
	user->conference = conference;

	ao2_lock(conference);
#ifdef GRANDSTREAM_NETWORKS
	/*Avoid creating multiple conference bridges*/
	if (conference->conf_end) {
		ast_debug(1, "conference end,cann't join conference bridge![%s]", conference->name);
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		user->conference = NULL;
		return NULL;
	}
#endif

	/* Determine if the new user should join the conference muted. */
	if (ast_test_flag(&user->u_profile, USER_OPT_STARTMUTED)
		|| (!ast_test_flag(&user->u_profile, USER_OPT_ADMIN) && conference->muted)) {
		/* Set user level mute request. */
		user->muted = 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if(conference->is_paging_conf  && !ignore_paging_callee_record) {
		conference->paging_caller_chan = user->chan; //added by zyxia for Bug 235579,set the caller chan
		ast_channel_set_is_page_caller(user->chan, 1);
	}
	if (!ignore_paging_callee_record && ast_test_flag(&user->b_profile, BRIDGE_OPT_RECORD_CONFERENCE)
		&& !ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE)) {
		/*prompt to  page caller at here ,callee prompt record play in function page_state_callback at app_page.c */
		if (conference->is_paging_conf ) {
			if (!ast_strlen_zero(record_prompt) && !strcmp(record_prompt, "yes") ){
				ast_log(LOG_DEBUG, "[debug]play prompt to  chan:%s",ast_channel_name(user->chan));
				ast_channel_check_need_prompt_start_stop_record(user->chan, NULL, 1);
			}
		}
		ast_log(LOG_DEBUG, "[debug]start confrecord");
		ast_clear_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);
		strcpy(conference->b_profile.rec_file, user->b_profile.rec_file);
		conf_start_record(conference, user->chan);
	}

	if (conference->is_paging_conf && !ignore_paging_callee_record && ast_test_flag(&user->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE)
		&& !ast_test_flag(&conference->b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE)) {
		ast_log(LOG_DEBUG, "[debug]start video confrecord,%s",user->b_profile.rec_file);

		ast_clear_flag(&conference->b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);
		strcpy(conference->b_profile.rec_file, user->b_profile.rec_file);
		const char *mix_filename = pbx_builtin_getvar_helper(user->chan, "MIXFILENAME");
		ast_str_reset(conference->video_filename);
		ast_str_set(&conference->video_filename, 0, "%s",user->b_profile.rec_file);


		if (conf_start_focus_record(conference) ==R_SUCCESS){
			if (mix_filename) {
				char record_files[1024];
				snprintf(record_files,sizeof(record_files),"%s.mkv@",mix_filename);
				ast_cdr_setvar(ast_channel_name(user->chan), "RECORD_FILES", record_files);
				if (!ast_strlen_zero(record_files)) {
					ast_log(LOG_DEBUG, "caller:%s cdr-record:%s",
						ast_channel_name(user->chan), record_files);
				}

				if (strlen(ast_str_buffer(conference->video_files_uuid))) {
					char record_files_uuid[1024] = "";
					snprintf(record_files_uuid, sizeof(record_files_uuid), "%s", ast_str_buffer(conference->video_files_uuid));
					ast_cdr_setvar(ast_channel_name(user->chan), "record_files_uuid", record_files_uuid);
					ast_log(LOG_DEBUG, "[join_conference_bridge] Set channel:%s's cdr record_files_uuid:%s", ast_channel_name(user->chan), record_files_uuid);
				}
			}
			ast_channel_set_is_page_focus_record(user->chan, 1);
		}
	}

	if (conf_is_videoing(conference)) {
		ast_indicate(user->chan, AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_START);
	}

#endif

	/*
	 * Suspend any MOH until the user actually joins the bridge of
	 * the conference.  This way any pre-join file playback does not
	 * need to worry about MOH.
	 */
	user->suspended_moh = 1;

	if (handle_conf_user_join(user)) {
		/* Invalid event, nothing was done, so we don't want to process a leave. */
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		user->conference = NULL;
		return NULL;
	}

	if (ast_check_hangup(user->chan)) {
		ao2_unlock(conference);
		leave_conference(user);
		return NULL;
	}

	ao2_unlock(conference);

	/* If an announcement is to be played play it */
#ifdef GRANDSTREAM_NETWORKS
	check_reserve_host(user, conference);
	if (!ast_strlen_zero(user->u_profile.announcement) && ast_fileexists(user->u_profile.announcement, NULL, NULL)) {
		user->is_playing_annoncement =1; //added by zyxia Bug 244581,2023.3.23

		int play_time = 1,i = 0;
		const char* play_prompt_time = pbx_builtin_getvar_helper(user->chan, "PAGE_PLAY_PROMPT_TIME");
		if (play_prompt_time && play_prompt_time[0] != '\0')
		{
			play_time = atoi(play_prompt_time);
		}
		ast_log(LOG_DEBUG,"announcement=%s, play_time=%d",user->u_profile.announcement, play_time);
		for (i = 0; i < play_time; i++)
		{
			if (play_prompt_to_user(user,
				user->u_profile.announcement)) {
				ast_log(LOG_DEBUG,"Play filed:%s",user->u_profile.announcement);
				//leave_conference(user);
				//return NULL;
			}
		}

		user->is_playing_annoncement =0;
		/*begin: add by weidu ,for 110820, 20180621,paging caller Paging/ Intercom Announcement Warning */
		ao2_lock(conference);
		if (conference->activeusers >= 2) {
			AST_LIST_TRAVERSE(&conference->active_list, conf_tone_user, list) {
				if (ast_strlen_zero(conf_tone_user->u_profile.announcement)
					|| !ast_fileexists(conf_tone_user->u_profile.announcement, NULL, NULL)) {
					chan_user = conf_tone_user->chan;
					if (chan_user && !ast_channel_reg_sendbeep_get(chan_user) && ast_channel_paging_get(chan_user)) {
						ast_bridge_lock(conf_tone_user->conference->bridge);
						in_bridge = !ast_bridge_suspend(conf_tone_user->conference->bridge, chan_user);
						ast_stream_and_wait(chan_user, "beep", "");
						if (in_bridge) {
							ast_bridge_unsuspend(conf_tone_user->conference->bridge, chan_user);
						}
						ast_bridge_unlock(conf_tone_user->conference->bridge);
						ast_channel_reg_sendbeep_set(chan_user, 1);
						ast_log(LOG_DEBUG, "join_conference_bridge have send beep tone to %s !\n", ast_channel_name(chan_user));
					}
				}
			}
		}
		ao2_unlock(conference);
		/*end: add by weidu ,for 110820*/
	}
	/*Start added by zyxia for sigle paging video support */
	if (user->chan && ast_channel_page_mode_get(user->chan)) {
		const char* page_caller_name = pbx_builtin_getvar_helper(user->chan, "PAGING_INTERCOM_CALLER_CHAN_NAME");
		if (!ast_strlen_zero(page_caller_name)) {
			struct ast_channel *page_caller_chan = NULL;
			page_caller_chan = ast_channel_get_by_name(page_caller_name);

			if (page_caller_chan) {
				RAII_VAR(char *, msid, NULL, ast_free);
				msid = ast_channel_get_msid(page_caller_chan, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
				if (!msid) {
					msid = ast_strdup("");
				}

				ast_log(LOG_NOTICE, "channel:%s Msid:%s\n",ast_channel_name(user->chan),msid);
				ast_indicate_data(user->chan, AST_CONTROL_UPDATE_PAGE_MSID, msid, strlen(msid) + 1);
				ast_channel_unref(page_caller_chan);
			}
		}
	}
	/*End added by zyxia for sigle paging video support */
#else
	if (!ast_strlen_zero(user->u_profile.announcement)) {
		if (play_prompt_to_user(user,
			user->u_profile.announcement)) {
			leave_conference(user);
			return NULL;
		}
	}
#endif
	/* Announce number of users if need be */
	if (ast_test_flag(&user->u_profile, USER_OPT_ANNOUNCEUSERCOUNT)) {
		if (announce_user_count(conference, user, NULL)) {
			leave_conference(user);
			return NULL;
		}
	}

	if (ast_test_flag(&user->u_profile, USER_OPT_ANNOUNCEUSERCOUNTALL) &&
		(conference->activeusers > user->u_profile.announce_user_count_all_after)) {
		int user_count_res;

		/*
		 * We have to autoservice the new user because he has not quite
		 * joined the conference yet.
		 */
		ast_autoservice_start(user->chan);
		user_count_res = announce_user_count(conference, NULL, NULL);
		ast_autoservice_stop(user->chan);
		if (user_count_res) {
			leave_conference(user);
			return NULL;
		}
	}

	/* Handle post-join actions */
	while ((action = AST_LIST_REMOVE_HEAD(&user->post_join_list, list))) {
		action->func(user);
		ast_free(action);
	}

	return conference;
}

/*!
 * \brief Leave a conference
 *
 * \param user The conference user
 */
static void leave_conference(struct confbridge_user *user)
{
	struct post_join_action *action;

	ao2_lock(user->conference);
	handle_conf_user_leave(user);
#ifdef GRANDSTREAM_NETWORKS
	conf_user_leave_handle_one_person_warning(user);
	handle_mcb_user_remove(user);
	/* Add by bfdu for Bug 144763, 2020.4.7 */
	if (user->conference->focus_chan) {
		ast_log(LOG_DEBUG, "leave_conference, but conference has focus chan, don't indicate moderator change\n");
	} else {
		if (ast_test_flag(&user->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU) && ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
			sfu_topologies_on_moderator_change(user->conference, NULL);
		}
	}

	conference_write_user_account(user);
	/*conference camera_statistics.*/
	if (user->video && user->conference->num_camera > 0) {
		user->conference->num_camera--;
	}
	if (user->conference->share_system_audio_chan
		&& ast_channel_name(user->chan)
		&& !strcasecmp(ast_channel_name(user->chan), user->conference->share_system_audio_chan)) {
		ast_debug(LOG_DEBUG, "share system audio chan %s leave!", user->conference->share_system_audio_chan);
		ast_free(user->conference->share_system_audio_chan);
		user->conference->share_system_audio_chan = NULL;
	}
#endif
	ao2_unlock(user->conference);

	/* Discard any post-join actions */
	while ((action = AST_LIST_REMOVE_HEAD(&user->post_join_list, list))) {
		ast_free(action);
	}

	/* Done mucking with the conference, huzzah */
	ao2_ref(user->conference, -1);
	user->conference = NULL;
}

static void playback_common(struct confbridge_conference *conference, const char *filename, int say_number)
{
	/* Don't try to play if the playback channel has been hung up */
	if (!conference->playback_chan) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_mutex_lock(&conference->playback_lock);
	int other_chan_num = 0;
	int join_prompt = 0;

	/*modify for Bug 160323*/
	if (filename && strstr(filename, "confbridge-join")) {
		join_prompt = 1;
	}
	other_chan_num = (conference->playback_chan ? 1 : 0) + (conference->record_chan ? 1 : 0);
	/*If only playback channel is in bridge, don't playback.*/
	if (conference->bridge->num_channels <= other_chan_num && !join_prompt) {
		ast_mutex_unlock(&conference->playback_lock);
		return;
	}
#endif
	ast_autoservice_stop(conference->playback_chan);

	/* The channel is all under our control, in goes the prompt */
	if (!ast_strlen_zero(filename)) {
		ast_stream_and_wait(conference->playback_chan, filename, "");
	} else if (say_number >= 0) {
		ast_say_number(conference->playback_chan, say_number, "",
			ast_channel_language(conference->playback_chan), NULL);
	}

	ast_autoservice_start(conference->playback_chan);
#ifdef GRANDSTREAM_NETWORKS
	ast_mutex_unlock(&conference->playback_lock);
#endif
}

struct playback_task_data {
	struct confbridge_conference *conference;
	const char *filename;
	int say_number;
	int playback_finished;
	ast_mutex_t lock;
	ast_cond_t cond;
};

/*!
 * \brief Play an announcement into a confbridge
 *
 * This runs in the playback queue taskprocessor. This ensures that
 * all playbacks are handled in sequence and do not play over top one
 * another.
 *
 * This task runs synchronously so there is no need for performing any
 * sort of cleanup on the input parameter.
 *
 * \param data A playback_task_data
 * \return 0
 */
static int playback_task(void *data)
{
	struct playback_task_data *ptd = data;

	playback_common(ptd->conference, ptd->filename, ptd->say_number);

	ast_mutex_lock(&ptd->lock);
	ptd->playback_finished = 1;
	ast_cond_signal(&ptd->cond);
	ast_mutex_unlock(&ptd->lock);

	return 0;
}

static void playback_task_data_init(struct playback_task_data *ptd, struct confbridge_conference *conference,
		const char *filename, int say_number)
{
	ast_mutex_init(&ptd->lock);
	ast_cond_init(&ptd->cond, NULL);

	ptd->filename = filename;
	ptd->say_number = say_number;
	ptd->conference = conference;
	ptd->playback_finished = 0;
}

static void playback_task_data_destroy(struct playback_task_data *ptd)
{
	ast_mutex_destroy(&ptd->lock);
	ast_cond_destroy(&ptd->cond);
}

static int play_sound_helper(struct confbridge_conference *conference, const char *filename, int say_number)
{
	struct playback_task_data ptd;

	/* Do not waste resources trying to play files that do not exist */
	if (ast_strlen_zero(filename)) {
		if (say_number < 0) {
			return 0;
		}
	} else if (!sound_file_exists(filename)) {
		return 0;
	}

	playback_task_data_init(&ptd, conference, filename, say_number);
	if (ast_taskprocessor_push(conference->playback_queue, playback_task, &ptd)) {
		if (!ast_strlen_zero(filename)) {
			ast_log(LOG_WARNING, "Unable to play file '%s' to conference %s\n",
				filename, conference->name);
		} else {
			ast_log(LOG_WARNING, "Unable to say number '%d' to conference %s\n",
				say_number, conference->name);
		}
		playback_task_data_destroy(&ptd);
		return -1;
	}

	/* Wait for the playback to complete */
	ast_mutex_lock(&ptd.lock);
	while (!ptd.playback_finished) {
		ast_cond_wait(&ptd.cond, &ptd.lock);
	}
	ast_mutex_unlock(&ptd.lock);

	playback_task_data_destroy(&ptd);

	return 0;
}

int play_sound_file(struct confbridge_conference *conference, const char *filename)
{
	return play_sound_helper(conference, filename, -1);
}

struct async_playback_task_data {
	struct confbridge_conference *conference;
	int say_number;
	struct ast_channel *initiator;
	char filename[0];
};

struct async_datastore_data {
	ast_mutex_t lock;
	ast_cond_t cond;
	int wait;
};

static void async_datastore_data_destroy(void *data)
{
	struct async_datastore_data *add = data;

	ast_mutex_destroy(&add->lock);
	ast_cond_destroy(&add->cond);

	ast_free(add);
}

/*!
 * \brief Datastore used for timing of async announcement playback
 *
 * Announcements that are played to the entire conference can be played
 * asynchronously (i.e. The channel that queues the playback does not wait
 * for the playback to complete before continuing)
 *
 * The thing about async announcements is that the channel that queues the
 * announcement is either not in the bridge or is in some other way "occupied"
 * at the time the announcement is queued. Because of that, the initiator of
 * the announcement may enter after the announcement has already started,
 * resulting in the sound being "clipped".
 *
 * This datastore makes it so that the channel that queues the async announcement
 * can say "I'm ready now". This way the announcement does not start until the
 * initiator of the announcement is ready to hear the sound.
 */
static struct ast_datastore_info async_datastore_info = {
	.type = "Confbridge async playback",
	.destroy = async_datastore_data_destroy,
};

static struct async_datastore_data *async_datastore_data_alloc(void)
{
	struct async_datastore_data *add;

	add = ast_malloc(sizeof(*add));
	if (!add) {
		return NULL;
	}

	ast_mutex_init(&add->lock);
	ast_cond_init(&add->cond, NULL);
	add->wait = 1;

	return add;
}

/*!
 * \brief Prepare the async playback datastore
 *
 * This is done prior to queuing an async announcement. If the
 * datastore has not yet been created, it is allocated and initialized.
 * If it already exists, we set it to be in "waiting" mode.
 *
 * \param initiator The channel that is queuing the async playback
 * \retval 0 Success
 * \retval -1 Failure :(
 */
static int setup_async_playback_datastore(struct ast_channel *initiator)
{
	struct ast_datastore *async_datastore;

	async_datastore = ast_channel_datastore_find(initiator, &async_datastore_info, NULL);
	if (async_datastore) {
		struct async_datastore_data *add;

		add = async_datastore->data;
		add->wait = 1;

		return 0;
	}

	async_datastore = ast_datastore_alloc(&async_datastore_info, NULL);
	if (!async_datastore) {
		return -1;
	}

	async_datastore->data = async_datastore_data_alloc();
	if (!async_datastore->data) {
		ast_datastore_free(async_datastore);
		return -1;
	}

	ast_channel_datastore_add(initiator, async_datastore);
	return 0;
}

static struct async_playback_task_data *async_playback_task_data_alloc(
	struct confbridge_conference *conference, const char *filename, int say_number,
	struct ast_channel *initiator)
{
	struct async_playback_task_data *aptd;

	aptd = ast_malloc(sizeof(*aptd) + strlen(filename) + 1);
	if (!aptd) {
		return NULL;
	}

	/* Safe */
	strcpy(aptd->filename, filename);
	aptd->say_number = say_number;

	/* You may think that we need to bump the conference refcount since we are pushing
	 * this task to the taskprocessor.
	 *
	 * In this case, that actually causes a problem. The destructor for the conference
	 * pushes a hangup task into the taskprocessor and waits for it to complete before
	 * continuing. If the destructor gets called from a taskprocessor task, we're
	 * deadlocked.
	 *
	 * So is there a risk of the conference being freed out from under us? No. Since
	 * the destructor pushes a task into the taskprocessor and waits for it to complete,
	 * the destructor cannot free the conference out from under us. No further tasks
	 * can be queued onto the taskprocessor after the hangup since no channels are referencing
	 * the conference at that point any more.
	 */
	aptd->conference = conference;

	aptd->initiator = initiator;
	if (initiator) {
		ast_channel_ref(initiator);
		ast_channel_lock(aptd->initiator);
		/* We don't really care if this fails. If the datastore fails to get set up
		 * we'll still play the announcement. It's possible that the sound will be
		 * clipped for the initiator, but that's not the end of the world.
		 */
		setup_async_playback_datastore(aptd->initiator);
		ast_channel_unlock(aptd->initiator);
	}

	return aptd;
}

static void async_playback_task_data_destroy(struct async_playback_task_data *aptd)
{
	ast_channel_cleanup(aptd->initiator);
	ast_free(aptd);
}

/*!
 * \brief Wait for the initiator of an async playback to be ready
 *
 * See the description on the async_datastore_info structure for more
 * information about what this is about.
 *
 * \param initiator The channel that queued the async announcement
 */
static void wait_for_initiator(struct ast_channel *initiator)
{
	struct ast_datastore *async_datastore;
	struct async_datastore_data *add;

	ast_channel_lock(initiator);
	async_datastore = ast_channel_datastore_find(initiator, &async_datastore_info, NULL);
	ast_channel_unlock(initiator);

	if (!async_datastore) {
		return;
	}

	add = async_datastore->data;

	ast_mutex_lock(&add->lock);
	while (add->wait) {
		ast_cond_wait(&add->cond, &add->lock);
	}
	ast_mutex_unlock(&add->lock);
}

/*!
 * \brief Play an announcement into a confbridge asynchronously
 *
 * This runs in the playback queue taskprocessor. This ensures that
 * all playbacks are handled in sequence and do not play over top one
 * another.
 *
 * \param data An async_playback_task_data
 * \return 0
 */
static int async_playback_task(void *data)
{
	struct async_playback_task_data *aptd = data;

	/* Wait for the initiator to get back in the bridge or be hung up */
	if (aptd->initiator) {
		wait_for_initiator(aptd->initiator);
	}

	playback_common(aptd->conference, aptd->filename, aptd->say_number);

	async_playback_task_data_destroy(aptd);
	return 0;
}

static int async_play_sound_helper(struct confbridge_conference *conference,
	const char *filename, int say_number, struct ast_channel *initiator)
{
	struct async_playback_task_data *aptd;

	/* Do not waste resources trying to play files that do not exist */
	if (ast_strlen_zero(filename)) {
		if (say_number < 0) {
			return 0;
		}
	} else if (!sound_file_exists(filename)) {
		return 0;
	}

	aptd = async_playback_task_data_alloc(conference, filename, say_number, initiator);
	if (!aptd) {
		return -1;
	}

	if (ast_taskprocessor_push(conference->playback_queue, async_playback_task, aptd)) {
		if (!ast_strlen_zero(filename)) {
			ast_log(LOG_WARNING, "Unable to play file '%s' to conference '%s'\n",
				filename, conference->name);
		} else {
			ast_log(LOG_WARNING, "Unable to say number '%d' to conference '%s'\n",
				say_number, conference->name);
		}
		async_playback_task_data_destroy(aptd);
		return -1;
	}

	return 0;
}

int async_play_sound_file(struct confbridge_conference *conference,
	const char *filename, struct ast_channel *initiator)
{
	return async_play_sound_helper(conference, filename, -1, initiator);
}

void async_play_sound_ready(struct ast_channel *chan)
{
	struct ast_datastore *async_datastore;
	struct async_datastore_data *add;

	ast_channel_lock(chan);
	async_datastore = ast_channel_datastore_find(chan, &async_datastore_info, NULL);
	ast_channel_unlock(chan);
	if (!async_datastore) {
		return;
	}

	add = async_datastore->data;

	ast_mutex_lock(&add->lock);
	add->wait = 0;
	ast_cond_signal(&add->cond);
	ast_mutex_unlock(&add->lock);
}
#ifdef GRANDSTREAM_NETWORKS
void synchronize_play_sound_file(struct confbridge_conference *conference, const char *filename, int say_number)
{
	playback_common(conference, filename, say_number);
}

static int wait_for_schedule_conference_start(struct confbridge_user *user)
{
	int res = 0;
	time_t expire = time(NULL) + user->u_profile.wait_timeout + 60;
	/* This is the holding pen for callers*/
	for (;;) {
		if ((res = ast_waitfordigit(user->chan, 1000))) {
			if (res <= 0) {
				ast_log(LOG_DEBUG, "exit waiting, res %d!", res);
				res = 0;
				break;
			}
		}
		if (user->u_profile.wait_timeout && (time(NULL) >= expire)) {
			ast_log(LOG_DEBUG, "exit waiting,wait timeout!");
			res = 1;
			break;
		}

	}
	return res;
}

#endif
/*!
 * \brief Play number into the conference bridge
 *
 * \param conference The conference bridge to say the number into
 * \param say_number number to say
 *
 * \retval 0 success
 * \retval -1 failure
 */
static int play_sound_number(struct confbridge_conference *conference, int say_number)
{
	return play_sound_helper(conference, NULL, say_number);
}

#ifdef GRANDSTREAM_NETWORKS
static void terminate_cei_subscription(struct confbridge_user *user)
{
	struct ast_channel *chan = NULL;
	int terminate = 0;

	if (!user || !user->chan || !user->conference) {
		return;
	}

	if (!((globalEnabledCEI && user->conference->is_conference_service) || ast_test_flag(&user->conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU))) {
		return;
	}

	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB)) {
		return;
	}

	/*Start added by zyxia for Bug 247244, trunk user dond send ternmianl notify*/
	if (ast_channel_name(user->chan) &&
		!strncmp(ast_channel_name(user->chan), "PJSIP/trunk_", 12)) {
		return;
	}
	/*End added by zyxia for Bug 247244, trunk user dond send ternmianl notify*/

	chan = user->chan;
	ast_channel_lock(chan);
	if (!strcasecmp(ast_channel_tech(chan)->type, "PJSIP") || !strcasecmp(ast_channel_tech(chan)->type, "Local")) {
		if (ast_channel_rtp_timeout_hangup_get(chan) ||
			ast_channel_ami_action_hangup_get(chan) ||
			!ast_channel_get_recv_bye(chan)) {
			terminate = 1;
		}
	}
	ast_channel_unlock(chan);

	if (terminate) {
		ast_log(LOG_DEBUG, "publish cei unsub for channel '%s'", ast_channel_name(chan));
		publish_cei_unsub(user->conference->name, ast_channel_name(chan), ast_channel_caller(chan)->id.number.str);
	}
}

static int send_talking_event(struct confbridge_user *user, struct confbridge_conference *conference)
{
	struct ast_json *talking_extras = NULL;

	talking_extras = ast_json_pack("{s: i, s: s, s: i, s: b, s: s}",
		"user_number", user->user_no,
		"talking_status", user->talking ? "on" : "off",
		"talking_rms", user->talking_rms,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous")
	);

	if (!talking_extras) {
		return 0;
	}

	send_conf_stasis(conference, user->chan, confbridge_talking_type(), talking_extras, 0);

	RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(talking_extras), ast_json_unref);
	if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
		send_conf_stasis(conference, user->chan, cei_talk_status_type(), obj, 1);
	}

	ast_json_unref(talking_extras);
	return 0;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static int conf_handle_talker_cb(struct ast_bridge_channel *bridge_channel, void *hook_pvt, int talking, int rms)
#else
static int conf_handle_talker_cb(struct ast_bridge_channel *bridge_channel, void *hook_pvt, int talking)
#endif
{
	struct confbridge_user *user = hook_pvt;
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct ast_json *talking_extras;
#ifdef GRANDSTREAM_NETWORKS
	int last_talking = user->talking ? 1 : 0;
#endif

	conference = ao2_find(conference_bridges, user->conference->name, OBJ_KEY);
	if (!conference) {
		/* Remove the hook since the conference does not exist. */
		return -1;
	}

	ao2_lock(conference);
	user->talking = talking;
#ifdef GRANDSTREAM_NETWORKS
	user->talking_rms = rms;
	if (user->muted) {
		ao2_unlock(conference);
		return 0;
	}
#endif
	ao2_unlock(conference);

#ifndef GRANDSTREAM_NETWORKS
	talking_extras = ast_json_pack("{s: s, s: b}",
		"talking_status", talking ? "on" : "off",
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN));
#else
	talking_extras = ast_json_pack("{s: i, s: s, s: i, s: b, s: s}",
		"user_number", user->user_no,
		"talking_status", talking ? "on" : "off",
		"talking_rms", rms,
		"admin", ast_test_flag(&user->u_profile, USER_OPT_ADMIN),
		"caller_id_number", S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous")
	);
#endif
	if (!talking_extras) {
		return 0;
	}

	send_conf_stasis(conference, bridge_channel->chan, confbridge_talking_type(), talking_extras, 0);

#ifdef GRANDSTREAM_NETWORKS
	if (last_talking != talking) {
		RAII_VAR(struct ast_json *, obj, ast_json_deep_copy(talking_extras), ast_json_unref);

		if (((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) && obj != NULL) {
			send_conf_stasis(conference, user->chan, cei_talk_status_type(), obj, 1);
		}
	}
#endif
	ast_json_unref(talking_extras);
	return 0;
}

static int conf_get_pin(struct ast_channel *chan, struct confbridge_user *user)
{
	char pin_guess[MAX_PIN+1] = { 0, };
	const char *pin = user->u_profile.pin;
	char *tmp = pin_guess;
	int i, res;
	unsigned int len = MAX_PIN;

	/*
	 * NOTE: We have not joined a conference yet so we have to use
	 * the bridge profile requested by the user.
	 */

	/* give them three tries to get the pin right */
	for (i = 0; i < 3; i++) {
		if (ast_app_getdata(chan,
			conf_get_sound(CONF_SOUND_GET_PIN, user->b_profile.sounds),
			tmp, len, 0) >= 0) {
			if (!strcasecmp(pin, pin_guess)) {
				return 0;
			}
		}
		ast_streamfile(chan,
			conf_get_sound(CONF_SOUND_INVALID_PIN, user->b_profile.sounds),
			ast_channel_language(chan));
		res = ast_waitstream(chan, AST_DIGIT_ANY);
		if (res > 0) {
			/* Account for digit already read during ivalid pin playback
			 * resetting pin buf. */
			pin_guess[0] = res;
			pin_guess[1] = '\0';
			tmp = pin_guess + 1;
			len = MAX_PIN - 1;
		} else {
			/* reset pin buf as empty buffer. */
			tmp = pin_guess;
			len = MAX_PIN;
		}
	}
	return -1;
}

static int user_timeout(struct ast_bridge_channel *bridge_channel, void *ignore)
{
	ast_bridge_channel_leave_bridge(bridge_channel, BRIDGE_CHANNEL_STATE_END, 0);
	pbx_builtin_setvar_helper(bridge_channel->chan, "CONFBRIDGE_RESULT", "TIMEOUT");
	return -1;
}

static int conf_rec_name(struct confbridge_user *user, const char *conf_name)
{
	char destdir[PATH_MAX];
	int res;
	int duration = 20;

	snprintf(destdir, sizeof(destdir), "%s/confbridge", ast_config_AST_SPOOL_DIR);

	if (ast_mkdir(destdir, 0777) != 0) {
		ast_log(LOG_WARNING, "mkdir '%s' failed: %s\n", destdir, strerror(errno));
		return -1;
	}
	snprintf(user->name_rec_location, sizeof(user->name_rec_location),
		 "%s/confbridge-name-%s-%s", destdir,
		 conf_name, ast_channel_uniqueid(user->chan));

	if (!(ast_test_flag(&user->u_profile, USER_OPT_ANNOUNCE_JOIN_LEAVE_REVIEW))) {
#ifdef GRANDSTREAM_NETWORKS
		if (ast_channel_is_avs_engine(user->chan)) {
			/*AVS does not support sln format at this time.*/
			res = ast_avs_play_and_record(user->chan,
			"vm-rec-name",
			user->name_rec_location,
			10,
			"wav",
			&duration,
			NULL,
			0,
			ast_dsp_get_threshold_from_settings(THRESHOLD_SILENCE),
			0,
			NULL,
			"#",
			NULL,
			0,
			AST_RECORD_IF_EXISTS_OVERWRITE,
			NULL);
		} else {
			res = ast_play_and_record(user->chan,
				"vm-rec-name",
				user->name_rec_location,
				10,
				"sln",
				&duration,
				NULL,
				ast_dsp_get_threshold_from_settings(THRESHOLD_SILENCE),
				0,
				NULL);
		}
#else
		res = ast_play_and_record(user->chan,
			"vm-rec-name",
			user->name_rec_location,
			10,
			"sln",
			&duration,
			NULL,
			ast_dsp_get_threshold_from_settings(THRESHOLD_SILENCE),
			0,
			NULL);
#endif
	} else {
		res = ast_record_review(user->chan,
			"vm-rec-name",
			user->name_rec_location,
			10,
			"sln",
			&duration,
			NULL);
	}

	if (res == -1) {
		ast_filedelete(user->name_rec_location, NULL);
		user->name_rec_location[0] = '\0';
		return -1;
	}
	return 0;
}

struct async_delete_name_rec_task_data {
	struct confbridge_conference *conference;
	char filename[0];
};

static struct async_delete_name_rec_task_data *async_delete_name_rec_task_data_alloc(
	struct confbridge_conference *conference, const char *filename)
{
	struct async_delete_name_rec_task_data *atd;

	atd = ast_malloc(sizeof(*atd) + strlen(filename) + 1);
	if (!atd) {
		return NULL;
	}

	/* Safe */
	strcpy(atd->filename, filename);
	atd->conference = conference;

	return atd;
}

static void async_delete_name_rec_task_data_destroy(struct async_delete_name_rec_task_data *atd)
{
	ast_free(atd);
}

/*!
 * \brief Delete user's name file asynchronously
 *
 * This runs in the playback queue taskprocessor. This ensures that
 * sound file is removed after playback is finished and not before.
 *
 * \param data An async_delete_name_rec_task_data
 * \return 0
 */
static int async_delete_name_rec_task(void *data)
{
	struct async_delete_name_rec_task_data *atd = data;

	ast_filedelete(atd->filename, NULL);
	ast_log(LOG_DEBUG, "Conference '%s' removed user name file '%s'\n",
		atd->conference->name, atd->filename);

	async_delete_name_rec_task_data_destroy(atd);
	return 0;
}

static int async_delete_name_rec(struct confbridge_conference *conference,
	const char *filename)
{
	struct async_delete_name_rec_task_data *atd;

	if (ast_strlen_zero(filename)) {
		return 0;
	} else if (!sound_file_exists(filename)) {
		return 0;
	}

	atd = async_delete_name_rec_task_data_alloc(conference, filename);
	if (!atd) {
		return -1;
	}

	if (ast_taskprocessor_push(conference->playback_queue, async_delete_name_rec_task, atd)) {
		ast_log(LOG_WARNING, "Conference '%s' was unable to remove user name file '%s'\n",
			conference->name, filename);
		async_delete_name_rec_task_data_destroy(atd);
		return -1;
	}

	return 0;
}

static int join_callback(struct ast_bridge_channel *bridge_channel, void *ignore)
{
	async_play_sound_ready(bridge_channel->chan);
#ifdef GRANDSTREAM_NETWORKS
	if (ast_channel_callerintercom_get(bridge_channel->chan)) {
		ast_log(LOG_DEBUG, "Channel %s join private intercom", ast_channel_name(bridge_channel->chan));
		struct confbridge_user *user = ignore;
		struct confbridge_conference *conference = user->conference;
		struct confbridge_user *cur_user = NULL;
		static int caller_is_in = 0;

		ao2_lock(conference);
		if (bridge_channel->chan == conference->paging_caller_chan) {
			AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
				if (bridge_channel->chan == cur_user->chan) {
					continue;
				}
				ast_log(LOG_DEBUG, "Channel %s subscribe Channel %s", ast_channel_name(bridge_channel->chan), ast_channel_name(cur_user->chan));
				ast_avs_subscribe_bridge(bridge_channel->chan, cur_user->chan, bridge_channel->bridge, bridge_channel->bridge->avs_bridge_id, bridge_channel->bridge->avs_bridge_tick);
				ast_avs_subscribe_bridge(cur_user->chan, bridge_channel->chan, bridge_channel->bridge, bridge_channel->bridge->avs_bridge_id, bridge_channel->bridge->avs_bridge_tick);
			}
			caller_is_in = 1;
		} else {
			if (caller_is_in && conference->paging_caller_chan) {
				ast_log(LOG_DEBUG, "Channel %s subscribe Channel %s", ast_channel_name(bridge_channel->chan), ast_channel_name(conference->paging_caller_chan));
				ast_avs_subscribe_bridge(bridge_channel->chan, conference->paging_caller_chan, bridge_channel->bridge, bridge_channel->bridge->avs_bridge_id, bridge_channel->bridge->avs_bridge_tick);
				ast_avs_subscribe_bridge(conference->paging_caller_chan, bridge_channel->chan, bridge_channel->bridge, bridge_channel->bridge->avs_bridge_id, bridge_channel->bridge->avs_bridge_tick);
			}
		}
		ao2_unlock(conference);
	}
#endif
#if 0
	struct confbridge_user *user = ignore;
	if (user->playing_moh) {
		ast_bridge_suspend(user->conference->bridge, user->chan);
		ast_bridge_set_ignore_unsuspend(user->conference->bridge, user->chan, 1);
	}
#endif
	return 0;
}

struct confbridge_hook_data {
	struct confbridge_conference *conference;
	struct confbridge_user *user;
	enum ast_bridge_hook_type hook_type;
};

#ifndef GRANDSTREAM_NETWORKS
static int send_event_hook_callback(struct ast_bridge_channel *bridge_channel, void *data)
{
	struct confbridge_hook_data *hook_data = data;

	if (hook_data->hook_type == AST_BRIDGE_HOOK_TYPE_JOIN) {
		send_join_event(hook_data->user, hook_data->conference);
	} else {
		send_leave_event(hook_data->user, hook_data->conference);
	}

	return 0;
}
#else
static int conf_user_hangup_hook(struct ast_bridge_channel *bridge_channel, void *hook_pvt)
{
	struct confbridge_user *user = hook_pvt;
	struct confbridge_conference *conference = user->conference;

	if (!conference) {
		return 0;
	}

	//ast_log(LOG_DEBUG, "execute hangup hook for channel %s.\n", ast_channel_name(bridge_channel->chan));
	if (ast_channel_ami_action_hangup_get(user->chan) && (conference->activeusers + conference->waitingusers) > MAX_USER_NUM_NO_IGNORE_AVS_RESP) {
		ast_channel_ignore_avs_response_set(bridge_channel->chan, 1);
	}

	return 0;
}

static int call_duration_init(struct ast_channel *chan, struct ast_bridge_features *features)
{
	const char *dur_limit;
	struct timeval calldurationlimit = { 0, };
	struct ast_bridge_config bridge_config = { { 0, } };
	if (!chan || !features) {
		return -1;
	}

	dur_limit = pbx_builtin_getvar_helper(chan, "DURLIMIT");
	if (!dur_limit || !strlen(dur_limit)) {
		return -1;
	}

	ast_bridge_timelimit(chan, &bridge_config, (char *)dur_limit, &calldurationlimit);
	if (bridge_config.timelimit) {
		int abandon_call = 0; /* TRUE if set limits fails so we can abandon the call. */
		struct ast_bridge_features_limits call_duration_limits_chan;

		if (ast_bridge_features_limits_construct(&call_duration_limits_chan)) {
			ast_log(LOG_ERROR, "Could not construct caller duration limits. Bridge canceled.\n");

			return -1;
		}

		if (ast_test_flag(&bridge_config.features_caller, AST_FEATURE_PLAY_WARNING)) {
			if (bridge_config.end_sound) {
				ast_string_field_set(&call_duration_limits_chan, duration_sound, bridge_config.end_sound);
			}

			if (bridge_config.warning_sound) {
				ast_string_field_set(&call_duration_limits_chan, warning_sound, bridge_config.warning_sound);
			}

			if (bridge_config.start_sound) {
				ast_string_field_set(&call_duration_limits_chan, connect_sound, bridge_config.start_sound);
			}

			call_duration_limits_chan.frequency = bridge_config.warning_freq;
			call_duration_limits_chan.warning = bridge_config.play_warning;
		}
		call_duration_limits_chan.duration = bridge_config.timelimit;

		if(ast_bridge_features_set_limits(features, &call_duration_limits_chan, 0)) {
			abandon_call = 1;
		}
		ast_bridge_features_limits_destroy(&call_duration_limits_chan);

		if (abandon_call) {
			ast_log(LOG_ERROR, "Could not set duration limits on one or more sides of the call. Bridge canceled.\n");
			return -1;
		}
	}

	return 0;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static unsigned int judge_if_nway_sfu_conference(struct ast_channel *chan, struct confbridge_conference *conference)
{
	const char *is_nway_conf = NULL;
	const char *is_sfu = NULL;

	if (!chan || !conference) {
		return 0;
	}

	if (!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		return 0;
	}

	is_nway_conf = pbx_builtin_getvar_helper(chan, "NWAY_CONF_NUMBER");
	is_sfu = pbx_builtin_getvar_helper(chan, "IS_SFU");

	if (ast_strlen_zero(is_nway_conf) || ast_strlen_zero(is_sfu) || strcasecmp(is_sfu, "yes")) {
		return 0;
	}

	return 1;
}

static void notify_avs_slider_video_msid(struct ast_channel *chan, struct confbridge_conference *conference)
{
	RAII_VAR(char *, msid, NULL, ast_free);
	struct confbridge_user *cur_user = NULL;

	if (!chan || !conference) {
		return;
	}

	if (ast_channel_get_opening_slides_video(chan)) {
		msid = ast_channel_get_slides_video_local_msid(chan);
		if (ast_strlen_zero(msid)) {
			ast_log(LOG_DEBUG, "Can't get slides video msid for chan:%s\n", ast_channel_name(chan));
			return;
		}
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
			if (cur_user->chan) {
				ast_indicate_data(cur_user->chan, AST_CONTROL_NOTIFY_SLIDER_MSID, msid, strlen(msid) + 1);
			}
		}
		ao2_unlock(conference);
	} else if (ast_channel_get_recv_slides_video(chan)) {
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
			if (cur_user->chan && ast_channel_get_opening_slides_video(cur_user->chan)) {
				msid = ast_channel_get_slides_video_local_msid(cur_user->chan);
				break;
			}
		}
		ao2_unlock(conference);

		if (!ast_strlen_zero(msid)) {
			ast_indicate_data(chan, AST_CONTROL_NOTIFY_SLIDER_MSID, msid, strlen(msid) + 1);
		}
	}

	return;
}

static void user_exit_delay(struct confbridge_user *user)
{
	AST_MUTEX_DEFINE_STATIC(user_exit_lock);
	static struct timeval start_tv = {0,0};
	struct timeval now_tv = {0,0};
	struct timeval delay_tv = {0,0};
	int delay_max_sec = 20;    /*The max delay time before use exiting, seconds.*/
	int delay_interval = 100;  /*The interval between two user exiting, milliseconds*/
	int delay_ms = 0;
	char ucm_dev[MAX_DEV_NAME] = {0};

	/*Check UCM6301/UCM6302*/
	conf_get_ucm_dec_info(ucm_dev, MAX_DEV_NAME);
	if (!strcasecmp(ucm_dev, "UCM6301") || !strcasecmp(ucm_dev, "UCM6300A") || !strcasecmp(ucm_dev, "UCM6302A")) {
		delay_interval = 100;
	} else if (!strcasecmp(ucm_dev, "UCM6302") || !strcasecmp(ucm_dev, "UCM6304A") || !strcasecmp(ucm_dev, "UCM6308A")) {
		delay_interval = 100;
	} else if (!strcasecmp(ucm_dev, "UCM6308")) {
		delay_interval = 100;
	} else {
		return;
	}

	ast_mutex_lock(&user_exit_lock);
	now_tv = ast_tvnow();
	if (ast_tvzero(start_tv)) {
		start_tv = now_tv;
	} else if (ast_tvcmp(now_tv, start_tv) > 0) {
		if (ast_tvdiff_ms(now_tv, start_tv) < delay_interval) {
			delay_ms = delay_interval;
		} else {
			start_tv = now_tv;
		}
	} else {
		if (ast_tvdiff_sec(start_tv, now_tv) <= delay_max_sec) {
			delay_ms = delay_interval;
		} else {
			start_tv = now_tv;
		}
	}

	if (delay_ms > 0) {
		delay_tv.tv_sec = delay_ms / 1000;
		delay_tv.tv_usec = (delay_ms % 1000) * 1000;
		start_tv = ast_tvadd(start_tv, delay_tv);

		delay_ms = ast_tvdiff_ms(start_tv, now_tv);
		if (delay_ms < 0) {
			delay_ms = 0;
		}
	}
	ast_mutex_unlock(&user_exit_lock);

	if (delay_ms > 0) {
		usleep(delay_ms*1000);
	}
}

#if 0
/*The application hook before moving the channel*/
static void conf_move_before_hook(struct ast_channel *chan)
{
	struct confbridge_user *user = NULL;

	if (!chan || !ast_channel_is_avs_engine(chan)) {
		return;
	}

	user = ast_channel_move_hook_data(chan);
	if (!user) {
		return;
	}

	ast_log(LOG_DEBUG, "Channel %s does the channel move before hook.", ast_channel_name(chan));
	terminate_cei_subscription(user);
	send_leave_event(user, user->conference);
}

/*The application hook after moving the channel*/
static void conf_move_after_hook(struct ast_channel *chan)
{
	struct confbridge_user *user = NULL;

	if (!chan || !ast_channel_is_avs_engine(chan)) {
		return;
	}

	user = ast_channel_move_hook_data(chan);
	if (!user) {
		return;
	}

	ast_log(LOG_DEBUG, "Channel %s does the channel move after hook.", ast_channel_name(chan));
	/* Set a talker indicate call back if talking detection is requested */
	if (ast_test_flag(&user->u_profile, USER_OPT_TALKER_DETECT)) {
		struct ast_talk_detect_param talk_detect_param = {
			.opt = AVS_TALK_DETECTION_CTRL_OPT_START,
			.talk_threshold = user->tech_args.talking_threshold,
			.silence_threshold = user->tech_args.silence_threshold,
		};
		ast_avs_control(chan, AVS_CONTROL_TALK_DETECT, &talk_detect_param, sizeof(talk_detect_param));
	}
	send_join_event(user, user->conference);
}
#endif
static int conf_runsql_get_int_field(char *sql, const char *key)
{
	int result = -1;
	char tmp[512] = {0};
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;
	struct odbc_obj *conf_obj= NULL;

	if (ast_strlen_zero(sql) || ast_strlen_zero(key)) {
		return result;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return result;
	}

	gps.sql = sql;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sql);
		goto end_clean;
	} else {

		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", sql);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			result = atoi(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", sql);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return result;
}

static char *conf_runsql_get_varchar_field(char *sql, const char *key)
{
	char *result = NULL;
	char tmp[512] = {0};
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;
	struct odbc_obj *conf_obj = NULL;
	if (ast_strlen_zero(sql) || ast_strlen_zero(key)) {
		return NULL;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return NULL;
	}

	gps.sql = sql;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sql);
		goto end_clean;
	} else {

		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", sql);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			result = ast_strdup(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", sql);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return result;
}

static int conf_write_db(char *cmd)
{
	int result = 0;
	struct odbc_obj *conf_obj = NULL;

	if (ast_strlen_zero(cmd)) {
		return -1;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	SQLHSTMT stmt;
	struct generic_prepare_struct gps = { .sql = cmd};

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", cmd);
		result = -1;
		goto end_clean;
	} else {
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}

	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return result;
}

static int net_detect_running(const char *interface)
{
	int fd = 0;
	int ret = 0;
	int rval = 0;
	struct ifreq ifr;

	if (interface == NULL || interface[0] == '\0') {
		ast_log(LOG_ERROR, "Interface error");
		return -1;
	}

	fd = socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
	if (fd < 0) {
		ast_log(LOG_ERROR, "socket err, %d %s", errno, strerror(errno));
		return -1;
	}

	memset(&ifr, 0, sizeof(ifr));
	snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s", interface);
	rval = ioctl(fd, SIOCGIFFLAGS, &ifr);

	if (rval < 0) {
		ast_log(LOG_ERROR, "ioctl err, %d %s", errno, strerror(errno));
	}
	else {
		ast_log(LOG_DEBUG, "ifr_flags: 0x%hx, %s\n", ifr.ifr_flags,
			ifr.ifr_flags & IFF_RUNNING ? "running" : "nolink");

		ret = ifr.ifr_flags & IFF_RUNNING ? 1 : 0;
	}
	close(fd);
	return ret;
}

static void general_meeting_link(struct confbridge_conference *conference)
{
	int web_port = 0;
	int web_https = 0;
	char *lan_ip = NULL;
	char *lan1_ip = NULL;
	char *lan2_ip = NULL;
	char *lan_ip_switch = NULL;
	char *link_full = NULL;
	const char *ip_val = NULL;
	char *web_public_url = NULL;
	char array_cmd[256] = {0};
	RAII_VAR(conf_net_info *, info, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, net_json, ast_json_object_create(), ast_json_unref);

	ao2_lock(conference);
	if (!ast_strlen_zero(conference->link_auth_id)) {
		ast_log(LOG_DEBUG, "conference link auth id %s", conference->link_auth_id);
		goto end_clean;;
	}
	info = ao2_alloc(sizeof(conf_net_info), NULL);
	if (!info || !net_json) {
		ast_log(LOG_DEBUG, "conf net info allco error!");
		goto end_clean;
	}
	/*init conf net info*/
	ast_conf_net_init(info);

	ast_conf_get_net_info(net_json, info);

	if (info->mode == 1) {
		if (net_detect_running(info->lan) == 1) {
			struct ast_json* lan_net_switch = NULL;
			lan_net_switch = ast_json_object_get(net_json, "lan");
			if (lan_net_switch) {
				struct ast_json* ip = ast_json_object_get(lan_net_switch, "ip");
				if (ip) {
					ip_val = ast_json_string_get(ip);
					ast_log(LOG_DEBUG, "lan ip = %s", ip_val);
					if (ip_val && strlen(ip_val)) {
						if (info->haon) {
							lan_ip_switch = ast_strdup(info->ip_hafloat);
						} else {
							lan_ip_switch = ast_strdup(ip_val);
						}
					} else {
						struct ast_json* ipv6 = ast_json_object_get(lan_net_switch, "ipv6");
						if (ipv6) {
							ip_val = ast_json_string_get(ipv6);
							if (ip_val) {
								lan_ip_switch = ast_strdup(ip_val);
								ast_log(LOG_DEBUG, "lan ipv6 = %s", lan_ip_switch);
							}
						}
					}
				}
			}
		}
	} else if (info->mode == 0) {
		struct ast_json* lan_net = NULL;
		lan_net = ast_json_object_get(net_json, "lan");
		if (lan_net) {
			struct ast_json* ip = ast_json_object_get(lan_net, "ip");
			if (ip) {
				ip_val = ast_json_string_get(ip);
				ast_log(LOG_DEBUG, "lan ip = %s", ip_val);
				if (ip_val && strlen(ip_val)) {
					if (info->haon) {
						lan_ip = ast_strdup(info->ip_hafloat);
					} else {
						lan_ip = ast_strdup(ip_val);
					}
				} else {
					struct ast_json* ipv6 = ast_json_object_get(lan_net, "ipv6");
					if (ipv6) {
						ip_val = ast_json_string_get(ipv6);
						if (ip_val) {
							lan_ip = ast_strdup(ip_val);
							ast_log(LOG_DEBUG, "lan ipv6 = %s", lan_ip);
						}
					}
				}
			}
		}
	} else if (info->mode == 2) {
		if (net_detect_running(info->lan1) == 1) {
			struct ast_json* lan1_net = NULL;
			lan1_net = ast_json_object_get(net_json, "lan1");
			if (lan1_net) {
				struct ast_json* ip = ast_json_object_get(lan1_net, "ip");
				if (ip) {
					ast_log(LOG_DEBUG, "lan1 ip = %s", ast_json_string_get(ip));
					ip_val = ast_json_string_get(ip);
					if (ip_val && strlen(ip_val)) {
						if (info->haon && !strcmp(info->default_if, "LAN1")) {
							lan1_ip = ast_strdup(info->ip_hafloat);
						} else {
							lan1_ip = ast_strdup(ip_val);
						}
					} else {
						struct ast_json* ipv6 = ast_json_object_get(lan1_net, "ipv6");
						if (ipv6) {
							ip_val = ast_json_string_get(ipv6);
							if (ip_val) {
								lan1_ip = ast_strdup(ip_val);
								ast_log(LOG_DEBUG, "lan1 ipv6 = %s", lan1_ip);
							}
						}
					}
				}
			}
		}
		if (net_detect_running(info->lan2) == 1) {
			struct ast_json* lan2_net = NULL;
			lan2_net = ast_json_object_get(net_json, "lan2");
			if (lan2_net) {
				struct ast_json* ip = ast_json_object_get(lan2_net, "ip");
				if (ip) {
					ip_val = ast_json_string_get(ip);
					ast_log(LOG_DEBUG, "lan2 ip = %s", ip_val);
					if (ip_val && strlen(ip_val)) {
						if (info->haon && !strcmp(info->default_if, "LAN2")) {
							lan2_ip = ast_strdup(info->ip_hafloat);
						} else {
							lan2_ip = ast_strdup(ip_val);
						}
					} else {
						struct ast_json* ipv6 = ast_json_object_get(lan2_net, "ipv6");
						if (ipv6) {
							ip_val = ast_json_string_get(ipv6);
							if (ip_val) {
								lan2_ip = ast_strdup(ip_val);
								ast_log(LOG_DEBUG, "lan2 ipv6 = %s", lan2_ip);
							}
						}
					}
				}
			}
		}
	}

#ifdef BUILD_ON_CLOUD_PBX
	web_https = 1;
#else
	const char *ptr = ast_safe_nvram_get("900", array_cmd, sizeof(array_cmd));
	if (ptr && strcasecmp(ptr, "1") == 0) {
		web_https = 1;
	}
#endif
	snprintf(array_cmd, sizeof(array_cmd), "SELECT gswave_port FROM %s.httpserver", ast_strlen_zero(s_dbConfigName) ? "ucm_config" : s_dbConfigName);
	web_port = conf_runsql_get_int_field(array_cmd, "gswave_port");
	snprintf(array_cmd, sizeof(array_cmd), "SELECT web_public_url FROM %s.httpserver", ast_strlen_zero(s_dbConfigName) ? "ucm_config" : s_dbConfigName);
	web_public_url = conf_runsql_get_varchar_field(array_cmd, "web_public_url");

	if (web_public_url) {
		ast_asprintf(&link_full, "%s://%s/#/invite/anonymous/%s/%s/0/0?join_meeting=no",
			web_https == 1 ? "https" : "http", web_public_url, conference->name, conference->feature_id);
	} else if (lan_ip_switch || lan_ip) {
		ast_asprintf(&link_full, "%s://%s:%d/#/invite/anonymous/%s/%s/0/0?join_meeting=no",
			web_https == 1 ? "https" : "http", lan_ip_switch ? lan_ip_switch : lan_ip, web_port, conference->name, conference->feature_id);
	} else if (lan1_ip) {
		ast_asprintf(&link_full, "%s://%s:%d/#/invite/anonymous/%s/%s/0/0?join_meeting=no",
			web_https == 1 ? "https" : "http", lan1_ip, web_port, conference->name, conference->feature_id);
	} else if (lan2_ip) {
		ast_asprintf(&link_full, "%s://%s:%d/#/invite/anonymous/%s/%s/0/0?join_meeting=no",
			web_https == 1 ? "https" : "http", lan2_ip, web_port, conference->name, conference->feature_id);
	} else {
		ast_log(LOG_ERROR, "A useless network!");
		goto end_clean;
	}
	if (link_full) {
		int enc_len = 0;
		int orig_len = 0;
		char *str_out = NULL;
		char *sql_cmd = NULL;
		char utc_start_time[80] = {0};

		orig_len = strlen(link_full);
		enc_len = (orig_len/3 + 2)*4;
		str_out = (char*) ast_malloc(enc_len + 1);
		if (str_out == NULL) {
			ast_log(LOG_ERROR, "ast_malloc error!");
			goto end_clean;
		}

		memset(str_out, 0, enc_len + 1);
		ast_base64encode(str_out, (const unsigned char *)link_full, orig_len, enc_len);
		snprintf(array_cmd, sizeof(array_cmd),"CONF_STATUS/%s", conference->name);
		ast_db_put(array_cmd, "LINK_FULL_BASE64", str_out);
		ast_log(LOG_DEBUG, "link full put into database!");

		struct ast_tm tm = { 0, };
		struct timeval now = ast_tvnow();
		ast_localtime(&now, &tm, "UTC");
		ast_strftime(utc_start_time, sizeof(utc_start_time), "%Y-%m-%d %H:%M:%S.%6q", &tm);

		ast_asprintf(&sql_cmd, "REPLACE INTO %s.AMIST_conference_link_auth_status(conf_number,feature_id,time_begin) VALUES('%s','%s','%s')",
			ast_strlen_zero(s_dbAmiName) ? "ami_status" : s_dbAmiName, conference->name, conference->feature_id, utc_start_time);
		conf_write_db(sql_cmd);
		conference->link_auth_id = ast_strdup(conference->feature_id);
		AST_SAFE_FREE(sql_cmd);
		AST_SAFE_FREE(str_out);
	}

end_clean:
	ao2_unlock(conference);

	AST_SAFE_FREE(lan_ip);
	AST_SAFE_FREE(lan1_ip);
	AST_SAFE_FREE(lan2_ip);
	AST_SAFE_FREE(link_full);
	AST_SAFE_FREE(lan_ip_switch);
	AST_SAFE_FREE(web_public_url);

}
#endif

/*! \brief The ConfBridge application */
static int confbridge_exec(struct ast_channel *chan, const char *data)
{
	int res = 0, volume_adjustments[2];
	int quiet = 0;
	int async_delete_task_pushed = 0;
	char *parse;
	const char *b_profile_name = NULL;
	const char *u_profile_name = NULL;
	const char *menu_profile_name = NULL;
	/* add by kmye */
#ifdef GRANDSTREAM_NETWORKS
	struct confbridge_user *extra_user  = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	int is_conference_service = 0;
	int seamless_transfer_redirect_flag = 0;
	char result[32] = {0};
	int ami_send_status = 0;
	/*For attended transfer a call to conference*/
	/*
	struct ast_chan_move_hooks move_hooks = {
		.before_hook = conf_move_before_hook,
		.after_hook = conf_move_after_hook,
	};
	*/
#endif
	struct confbridge_conference *conference = NULL;
	struct confbridge_user user = {
		.chan = chan,
		.tech_args.talking_threshold = DEFAULT_TALKING_THRESHOLD,
		.tech_args.silence_threshold = DEFAULT_SILENCE_THRESHOLD,
		.tech_args.drop_silence = 0,
	};
#ifndef GRANDSTREAM_NETWORKS
	struct confbridge_hook_data *join_hook_data;
	struct confbridge_hook_data *leave_hook_data;
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf_name);
		AST_APP_ARG(b_profile_name);
		AST_APP_ARG(u_profile_name);
		AST_APP_ARG(menu_profile_name);
		/* add by kmye */
#ifdef GRANDSTREAM_NETWORKS
		AST_APP_ARG(extra_profile_name);
#endif
	);

	if(!ast_test_flag(ast_channel_flags(chan), AST_FLAG_RECOVERYING))
	{
		/* only update uesr channel, not local  */
		if(ast_test_flag(ast_channel_flags(chan), AST_FLAG_TRACKED))
		{
			ast_channel_recovery_track(chan, NULL, NULL);
		}
		else if(strstr(ast_channel_name(chan), "DAHDI/"))
		{
			ast_channel_recovery_track(chan, NULL, NULL);
		}
	}
	if (ast_ha_permit_send() && ha_need_sync_record_count()) {
		int count_db = 0;
		ast_mutex_lock(&record_count_lock);
		count_db = conf_report_get_call_count();
		if (count_db && count_db != count) {
			ast_log(LOG_DEBUG, "count_db %d and count %d, need sync\n", count_db, count);
			count = count_db;
		} else {
			ast_log(LOG_DEBUG,"db count [%d] VS count [%d]!", count_db, count);
		}
		ast_mutex_unlock(&record_count_lock);

	}

#ifdef GRANDSTREAM_NETWORKS
	terminate_type = ast_channel_get_terminate_type(chan);
	user.term_type = terminate_type;
	if (!ast_strlen_zero(pbx_builtin_getvar_helper(chan, "CONF_ROOM")) ||
		!ast_strlen_zero(pbx_builtin_getvar_helper(chan, "MCB_HASH"))) {
		is_conference_service = 1;
	}
	const char * do_redirect = pbx_builtin_getvar_helper(chan, "DO_REDIRECT");
	if (!ast_strlen_zero(do_redirect) && 0 == strcmp(do_redirect, "yes") && 0 == strcmp(ast_channel_context(chan), "sub-dial-conference")){
		ast_channel_context_set(chan,  "sub-dial-conference-redirect");
	}
#endif

	if (ast_channel_state(chan) != AST_STATE_UP) {
		ast_answer(chan);
	}

	if (ast_bridge_features_init(&user.features)) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		res = -1;
		goto confbridge_cleanup;
	}

#ifdef GRANDSTREAM_NETWORKS
	call_duration_init(chan, &user.features);
#endif

	/* We need to make a copy of the input string if we are going to modify it! */
	parse = ast_strdupa(data);

	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.conf_name)) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		ast_log(LOG_WARNING, "%s requires an argument (conference name[,options])\n", app);
		res = -1;
		goto confbridge_cleanup;
	}

	if (strlen(args.conf_name) >= MAX_CONF_NAME) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		ast_log(LOG_WARNING, "%s does not accept conference names longer than %d\n", app, MAX_CONF_NAME - 1);
		res = -1;
		goto confbridge_cleanup;
	}

	/* bridge profile name */
	if (args.argc > 1 && !ast_strlen_zero(args.b_profile_name)) {
		b_profile_name = args.b_profile_name;
	}
	if (!conf_find_bridge_profile(chan, b_profile_name, &user.b_profile)) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		ast_log(LOG_WARNING, "Conference bridge profile %s does not exist\n", b_profile_name ?
			b_profile_name : DEFAULT_BRIDGE_PROFILE);
		res = -1;
		goto confbridge_cleanup;
	}

	/* user profile name */
	if (args.argc > 2 && !ast_strlen_zero(args.u_profile_name)) {
		u_profile_name = args.u_profile_name;
	}
	if (!conf_find_user_profile(chan, u_profile_name, &user.u_profile)) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		ast_log(LOG_WARNING, "Conference user profile %s does not exist\n", u_profile_name ?
			u_profile_name : DEFAULT_USER_PROFILE);
		res = -1;
		goto confbridge_cleanup;
	}

	quiet = ast_test_flag(&user.u_profile, USER_OPT_QUIET);

	/* ask for a PIN immediately after finding user profile.  This has to be
	 * prompted for requardless of quiet setting. */
	if (!ast_strlen_zero(user.u_profile.pin)) {
		if (conf_get_pin(chan, &user)) {
			pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
			res = -1; /* invalid PIN */
			goto confbridge_cleanup;
		}
	}

	/* See if we need them to record a intro name */
	if (!quiet &&
		(ast_test_flag(&user.u_profile, USER_OPT_ANNOUNCE_JOIN_LEAVE) ||
		(ast_test_flag(&user.u_profile, USER_OPT_ANNOUNCE_JOIN_LEAVE_REVIEW)))) {
		if (conf_rec_name(&user, args.conf_name)) {
			pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
			res = -1; /* Hangup during name recording */
			goto confbridge_cleanup;
		}
	}

	/* menu name */
	if (args.argc > 3 && !ast_strlen_zero(args.menu_profile_name)) {
		menu_profile_name = args.menu_profile_name;
	}

	if (conf_set_menu_to_user(chan, &user, menu_profile_name)) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		ast_log(LOG_WARNING, "Conference menu profile %s does not exist\n", menu_profile_name ?
			menu_profile_name : DEFAULT_MENU_PROFILE);
		res = -1;
		goto confbridge_cleanup;
	}

	/* Set if DTMF should pass through for this user or not */
	if (ast_test_flag(&user.u_profile, USER_OPT_DTMF_PASS)) {
		user.features.dtmf_passthrough = 1;
	} else {
		user.features.dtmf_passthrough = 0;
	}

	/* Set dsp threshold values if present */
	if (user.u_profile.talking_threshold) {
		user.tech_args.talking_threshold = user.u_profile.talking_threshold;
	}
	if (user.u_profile.silence_threshold) {
		user.tech_args.silence_threshold = user.u_profile.silence_threshold;
	}

	/* Set a talker indicate call back if talking detection is requested */
	if (ast_test_flag(&user.u_profile, USER_OPT_TALKER_DETECT)) {
		if (ast_bridge_talk_detector_hook(&user.features, conf_handle_talker_cb,
			&user, NULL, AST_BRIDGE_HOOK_REMOVE_ON_PULL)) {
			pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
			res = -1;
			goto confbridge_cleanup;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (ast_channel_is_avs_engine(chan)) {
			struct ast_talk_detect_param talk_detect_param = {
				.opt = AVS_TALK_DETECTION_CTRL_OPT_START,
				.talk_threshold = user.tech_args.talking_threshold,
				.silence_threshold = user.tech_args.silence_threshold,
			};
			ast_avs_control(chan, AVS_CONTROL_TALK_DETECT, &talk_detect_param, sizeof(talk_detect_param));
		}
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add for schedule join early */
	int inuse_notify = 0;
	const char *wait_reservation_id = pbx_builtin_getvar_helper(chan, "WAIT_RESERVATION_ID");
	ast_log(LOG_DEBUG, "wait_reservation_id %s, user.u_profile.wait_timeout %d!", wait_reservation_id, user.u_profile.wait_timeout);
	if (user.u_profile.wait_timeout && !ast_strlen_zero(wait_reservation_id) && schedule_conference_need_wait(args.conf_name, wait_reservation_id, &inuse_notify)) {
			int ret = schedule_conference_attendee_join_wait_list(args.conf_name, chan);
			if (!ret) {
				/*Cancel video source and notify the terminal*/
				ast_channel_lock(chan);
				if (ast_channel_tech(chan)->indicate) {
					ast_channel_tech(chan)->indicate(chan, AST_CONTROL_WAIT_JOIN_CONFERENCE, NULL, 0);
				}
				ast_channel_unlock(chan);
				ast_log(LOG_DEBUG, "we need wait for schedule conference start!");
				play_file(NULL, chan, "conf-schedule-wait");
				ast_moh_start(chan, user.u_profile.moh_class, NULL);
				int is_timeout = wait_for_schedule_conference_start(&user);
				if (ast_check_hangup(chan) || is_timeout) {
					res = -1;
					if (is_timeout) {
						ast_log(LOG_DEBUG, "wait user time out,some thing may error!");
					} else {
						ast_log(LOG_DEBUG, "wait user hangup the call!");
					}
					schedule_conference_wait_user_remove_cache(args.conf_name, chan);
					schedule_conference_wait_user_remove_by_channel(args.conf_name, ast_channel_name(chan));
					schedule_conference_wait_list_cancel(args.conf_name);
					goto confbridge_cleanup;
				}
				ast_moh_stop(chan);
				play_file(NULL, chan, "conf-schedule-begin");
				inuse_notify = 1;
				ast_log(LOG_DEBUG, "now schedule conference start,join conference!");
				schedule_conference_wait_user_add_cache(chan, args.conf_name, ast_test_flag(&user.b_profile, BRIDGE_OPT_VIDEO_SRC_SFU) ? CONF_TYPE_VIDEO : CONF_TYPE_AUDIO, wait_reservation_id);
				schedule_conference_wait_user_remove_by_channel(args.conf_name, ast_channel_name(chan));
			}
	}
	if (inuse_notify) {
		pbx_builtin_setvar_helper(chan, "INUSE_NOTIFY", "yes");
	}
	if (ast_test_flag(&user.b_profile, BRIDGE_OPT_VIDEO_SRC_SFU) && (terminate_type == AST_TERMINAL_TYPE_WEBRTC || terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		char pjsip_extension[AST_MAX_EXTENSION] = {0};
		get_pjsip_extension(ast_channel_name(chan), pjsip_extension, sizeof(pjsip_extension));
		cache_conf_authorized_user_set_channel(args.conf_name, pjsip_extension, chan);
		const char *related_info = pbx_builtin_getvar_helper(chan, "RELATED_INFO");
		if (!ast_strlen_zero(related_info)) {
			ast_copy_string(user.related_ext, related_info, sizeof(user.related_ext));
		}
	}

	/* added by zgmo for Bug168748 */
	{
		const char *seamless_transfer = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER");
		const char *seamless_transfer_bridge =  pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_BRIDGE");
		const char *seamless_transfer_redirect = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_REDIRECT");

		if ((!ast_strlen_zero(seamless_transfer) && !strcmp(seamless_transfer, "yes"))
			|| (!ast_strlen_zero(seamless_transfer_bridge) && !strcmp(seamless_transfer_bridge, "yes"))) {
			const char *seamless_transfer_record = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_RECORD");
			if (!ast_strlen_zero(seamless_transfer_record)) {
				ast_set_flag(&user.b_profile, BRIDGE_OPT_RECORD_CONFERENCE);
				ast_clear_flag(&user.b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);

				strcpy(user.b_profile.rec_file, seamless_transfer_record);
			}
		} else if (!ast_strlen_zero(seamless_transfer_redirect) && !strcmp(seamless_transfer_redirect, "yes")) {
			int max = 100;
			struct confbridge_conference *finding = NULL;

			ast_log(LOG_DEBUG, "%s is redirected and need to wait confbridge to be created",
				ast_channel_name(chan));
			/*
			 *  In seamless transfer,
			 *  Only channel originated have the right to enable confbridge recording,
			 *  So we need to guarantee that channel originated is first go into confbridge
			 *  and channel redirected must wait.
			 */
			while (1) {
				finding = ao2_find(conference_bridges, args.conf_name, OBJ_KEY);

				if (finding) {
					ast_log(LOG_DEBUG, "confbridge:%s created, join it!", args.conf_name);
					ao2_ref(finding, -1);
					break;
				} else {
					max--;
					if (!max) {
						ast_log(LOG_DEBUG, "confbridge:%s still not created, aboard for max waiting time!",
							args.conf_name);
						break;
					}

					usleep(200*1000);
				}
			}
		}
	}
	/* end */
#endif

#ifdef GRANDSTREAM_NETWORKS
		/* add by wylu for BUG 172985 paging/intercom recording */
		/*Edit by zyxia for paging recording,2022.8.15*/
		const char *page_intercom = pbx_builtin_getvar_helper(chan, "PAGE_INTERCOM");
		if (!ast_strlen_zero(page_intercom) && !strcmp(page_intercom, "yes")) {
			const char *page_record_file = NULL;
			char dest_dir[PATH_MAX]={0};
			pbx_builtin_setvar_helper(chan,"PAGE_CONF_NUM",args.conf_name);//added by zyxia for Bug251489,2023.2.23
			snprintf(dest_dir, sizeof(dest_dir), "%s/paging_video", ast_config_AST_SPOOL_DIR);
			if (ast_channel_get_is_videocall(chan) && dir_is_exist(dest_dir)) {//added by zyxia for Bug 236453,if is not external disk,can't not record
				page_record_file = pbx_builtin_getvar_helper(chan, "PAGE_VIDEO_RECORD_FILE");
				if (!ast_strlen_zero(page_record_file)) {

					ast_set_flag(&user.b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE);
					ast_clear_flag(&user.b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE);
					ast_clear_flag(&user.b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);
					ast_clear_flag(&user.b_profile, BRIDGE_OPT_RECORD_CONFERENCE);
					strcpy(user.b_profile.rec_file, page_record_file);
				}
				else {
					page_record_file = pbx_builtin_getvar_helper(chan, "PAGE_RECORD_FILE");
					if (!ast_strlen_zero(page_record_file)) {
						ast_set_flag(&user.b_profile, BRIDGE_OPT_RECORD_CONFERENCE);
						ast_clear_flag(&user.b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE);
						ast_clear_flag(&user.b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);
						ast_clear_flag(&user.b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE);
						strcpy(user.b_profile.rec_file, page_record_file);
					}
				}
			}
			else {
				page_record_file = pbx_builtin_getvar_helper(chan, "PAGE_RECORD_FILE");
				if (!ast_strlen_zero(page_record_file)) {
					ast_set_flag(&user.b_profile, BRIDGE_OPT_RECORD_CONFERENCE);
					ast_clear_flag(&user.b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE);
					ast_clear_flag(&user.b_profile, BRIDGE_OPT_RECORD_FILE_TIMESTAMP);
					ast_clear_flag(&user.b_profile, BRIDGE_OPT_FOCUS_RECORD_CONFERENCE);
					strcpy(user.b_profile.rec_file, page_record_file);
				}
			}

			ast_indicate(chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
		}
		/* end */
#endif

	if(strstr(ast_channel_name(chan), "Local/")
		&& HA_ROLE_ACTIVE == ast_ha_get_role())
	{

		char local_2[64] = {0};
		struct ast_channel *local2, *peer_user = NULL;
		strcpy(local_2, ast_channel_name(chan));
		local_2[strlen(ast_channel_name(chan)) - 1] = '2';
		struct ast_bridge *local_bridge = NULL;
		char oldcontext[AST_MAX_CONTEXT] = {0};
		char oldexten[AST_MAX_EXTENSION] = {0};
		char *old_data;
		int old_priority;
		const char *ori_options = NULL;
		local2 = ast_channel_get_by_name(local_2);
		ori_options = pbx_builtin_getvar_helper(chan, "ORI_OPTIONS");
		if(local2)
		{

			local_bridge = ast_channel_internal_bridge(local2);
			if(local_bridge){
				peer_user = ast_bridge_peer(local_bridge, local2);
			}
			ast_channel_unref(local2);

			if(peer_user)
			{
				ast_channel_lock(peer_user);
				strcpy(oldcontext, ast_channel_context(peer_user));
				strcpy(oldexten, ast_channel_exten(peer_user));
				old_data = (char *)ast_channel_data(peer_user);
				old_priority = ast_channel_priority(peer_user);

				ast_channel_context_set(peer_user, ast_channel_context(chan));
				ast_channel_exten_set(peer_user, ast_channel_exten(chan));
				ast_channel_priority_set(peer_user, ast_channel_priority(chan));
				ast_channel_data_set(peer_user, ast_channel_data(chan));
				/* copy var ORI_OPTIONS  from local:1 to user channel */
				if(ori_options){
					pbx_builtin_setvar_helper(peer_user, "ORI_OPTIONS", ori_options);
				}
				ast_channel_recovery_track(peer_user, NULL, NULL);

				ast_channel_context_set(peer_user, oldcontext);
				ast_channel_exten_set(peer_user, oldexten);
				ast_channel_priority_set(peer_user, old_priority);
				ast_channel_data_set(peer_user, old_data);

				ast_channel_unlock(peer_user);
				ast_channel_unref(peer_user);
			}

		}
	}

	/* Look for a conference bridge matching the provided name */
	if (!(conference = join_conference_bridge(args.conf_name, &user))) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "FAILED");
		res = -1;
		goto confbridge_cleanup;
	}
#ifdef GRANDSTREAM_NETWORKS
	{
		const char *seamless_transfer_redirect = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_REDIRECT");
		const char *record_prompt = pbx_builtin_getvar_helper(NULL, "RECORD_PROMPT");
		if (!ast_strlen_zero(record_prompt) && !strcmp(record_prompt, "yes")
			&& !ast_strlen_zero(seamless_transfer_redirect) && !strcmp(seamless_transfer_redirect, "yes")) {
			const char *have_prompted = pbx_builtin_getvar_helper(chan, "HAVE_PROMPTED");
			if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE)) {
				if (ast_strlen_zero(have_prompted) || strcmp(have_prompted, "yes")) {
					ast_channel_async_streamfile_protection_enter(chan);
					ast_stream_and_wait(chan, "dialog-being-recorded", "");
					ast_channel_async_streamfile_protection_exit(chan);

					ast_log(LOG_DEBUG, "In seamless-transfer, confbridge is recording, so %s need prompt",
						ast_channel_name(chan));
				}
			} else {
				if (!ast_strlen_zero(have_prompted) && !strcmp(have_prompted, "yes")) {
					ast_channel_async_streamfile_protection_enter(chan);
					ast_stream_and_wait(chan, "dialog-stop-recording", "");
					ast_channel_async_streamfile_protection_exit(chan);

					ast_log(LOG_DEBUG, "In seamless-transfer, confbridge isn't recording, but %s has prompted record, so need prompt stop recording",
						ast_channel_name(chan));
				}
			}
		}
	}

	is_conference_service = conference->is_conference_service;
#if 0
	const char* seamless_transfer_record_enable = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_REOCRD_ENABLE");
	if (!ast_strlen_zero(seamless_transfer_record_enable) && 0 == strcasecmp(seamless_transfer_record_enable, "yes")) {
		const char* seamless_transfer_reocrd_files = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_REOCRD_FILES");
		if (!ast_strlen_zero(seamless_transfer_reocrd_files)) {
			conf_start_seamless_transfer_reocrd(seamless_transfer_reocrd_files, conference);
		}
	}
#endif

	/* we can't count this call if it is called by *45 Pickup call */

	const char* seamless_transfer = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER");
	if	(args.argc > 4 && !ast_strlen_zero(args.extra_profile_name)
		&& !ast_strlen_zero(seamless_transfer) && strcasecmp(seamless_transfer, "yes") == 0) {
		seamless_transfer_redirect_flag = 1;
		ast_decrease_call_count();
	}

#endif
	/* Keep a copy of volume adjustments so we can restore them later if need be */
	volume_adjustments[0] = ast_audiohook_volume_get(chan, AST_AUDIOHOOK_DIRECTION_READ);
	volume_adjustments[1] = ast_audiohook_volume_get(chan, AST_AUDIOHOOK_DIRECTION_WRITE);

	if (ast_test_flag(&user.u_profile, USER_OPT_DROP_SILENCE)) {
		user.tech_args.drop_silence = 1;
	}

	if (ast_test_flag(&user.u_profile, USER_OPT_JITTERBUFFER)) {
		ast_func_write(chan, "JITTERBUFFER(adaptive)", "default");
	}

	if (ast_test_flag(&user.u_profile, USER_OPT_DENOISE)) {
		ast_func_write(chan, "DENOISE(rx)", "on");
	}

	/* if this user has a intro, play it before entering */
	if (!ast_strlen_zero(user.name_rec_location)) {
		ast_autoservice_start(chan);
		play_sound_file(conference, user.name_rec_location);
		play_sound_file(conference,
			conf_get_sound(CONF_SOUND_HAS_JOINED, conference->b_profile.sounds));
		ast_autoservice_stop(chan);
	}

	/* Play the Join sound to both the conference and the user entering. */
	if (!quiet) {
		const char *join_sound = conf_get_sound(CONF_SOUND_JOIN, conference->b_profile.sounds);

		if (strcmp(conference->b_profile.language, ast_channel_language(chan))) {
			ast_stream_and_wait(chan, join_sound, "");
			ast_autoservice_start(chan);
			play_sound_file(conference, join_sound);
			ast_autoservice_stop(chan);
		} else {
#ifdef GRANDSTREAM_NETWORKS
			async_play_sound_file(conference, join_sound, NULL);
#else
			async_play_sound_file(conference, join_sound, chan);
#endif
		}
	}

	if (user.u_profile.timeout) {
		ast_bridge_interval_hook(&user.features,
			0,
			user.u_profile.timeout * 1000,
			user_timeout,
			NULL,
			NULL,
			AST_BRIDGE_HOOK_REMOVE_ON_PULL);
	}

	/* See if we need to automatically set this user as a video source or not */
	handle_video_on_join(conference, user.chan, ast_test_flag(&user.u_profile, USER_OPT_MARKEDUSER));

#ifndef GRANDSTREAM_NETWORKS
	conf_moh_unsuspend(&user);

	join_hook_data = ast_malloc(sizeof(*join_hook_data));
	if (!join_hook_data) {
		res = -1;
		goto confbridge_cleanup;
	}
	join_hook_data->user = &user;
	join_hook_data->conference = conference;
	join_hook_data->hook_type = AST_BRIDGE_HOOK_TYPE_JOIN;
	res = ast_bridge_join_hook(&user.features, send_event_hook_callback,
		join_hook_data, ast_free_ptr, 0);
	if (res) {
		ast_free(join_hook_data);
		ast_log(LOG_ERROR, "Couldn't add bridge join hook for channel '%s'\n", ast_channel_name(chan));
		goto confbridge_cleanup;
	}

	leave_hook_data = ast_malloc(sizeof(*leave_hook_data));
	if (!leave_hook_data) {
		/* join_hook_data is cleaned up by ast_bridge_features_cleanup via the goto */
		res = -1;
		goto confbridge_cleanup;
	}
	leave_hook_data->user = &user;
	leave_hook_data->conference = conference;
	leave_hook_data->hook_type = AST_BRIDGE_HOOK_TYPE_LEAVE;
	res = ast_bridge_leave_hook(&user.features, send_event_hook_callback,
		leave_hook_data, ast_free_ptr, 0);
	if (res) {
		/* join_hook_data is cleaned up by ast_bridge_features_cleanup via the goto */
		ast_free(leave_hook_data);
		ast_log(LOG_ERROR, "Couldn't add bridge leave hook for channel '%s'\n", ast_channel_name(chan));
		goto confbridge_cleanup;
	}
#else
	conf_user_is_video_call(&user);

	const char* from_call_flip = pbx_builtin_getvar_helper(chan, "FROM_CALL_FLIP");//added by zyxia for Bug 202171,call flip in conference need send notify
	if (!ast_strlen_zero(from_call_flip) && strcasecmp(from_call_flip, "yes") == 0) {
		ast_channel_set_is_callflip(chan, 1);
		ast_indicate(chan,AST_CONTROL_CONNECTED_LINE);
	}
	if (user.conference->b_profile.n_way_conf  || (!ast_strlen_zero(from_call_flip) && strcasecmp(from_call_flip, "yes") == 0)) {
		// Send "SIP NOTIFY" to tell the user already join N way conference.
		struct conf_notify_switch_data s_data;
		memset(&s_data, 0, sizeof(s_data));
		snprintf(s_data.conf_id, sizeof(s_data.conf_id), "%s", user.conference->name);

		// The globalEnabledCEI is only available for audio conference rooms,
		// and video conference rooms are enabled by default.
		if (ast_test_flag(&user.b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			s_data.is_video_conf = 1;
			s_data.enabled_cei = 1;
		} else {
			s_data.is_video_conf = 0;
			s_data.enabled_cei = globalEnabledCEI;
		}

		if (ast_test_flag(&user.u_profile, USER_OPT_ADMIN)) {
			s_data.is_admin = 1;
		} else {
			s_data.is_admin = 0;
		}

		if (ast_channel_get_is_callflip(chan)) {
			s_data.event_type = 1;
		}
		ast_channel_set_is_admin(chan,s_data.is_admin);//added by zyxia Bug 268897,2023.8.16
		ast_log(LOG_DEBUG, "N-Way-Conference Or Call Flip will be sent SIP NOTIFY(%s).", ast_channel_name(user.chan));

		ast_channel_lock(chan);
		if (ast_channel_tech(user.chan)->indicate) {
			ast_channel_tech(user.chan)->indicate(user.chan, AST_CONTROL_NOTIFY_N_WAY_CONF, &s_data, sizeof(s_data));
		}
		ast_channel_unlock(chan);
	}

	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		general_meeting_link(conference);
		ast_indicate(chan, AST_CONTROL_JOIN_SFU);
	}

	/* Join our conference bridge for real */
	if (!ast_test_flag(&user.u_profile, USER_OPT_JOINEDUSER)) {
		user.user_no = ++conference->max_users;
		send_join_event(&user, conference);
		schedule_conference_error_recovery_detect(conference);
		schedule_conference_attendee_join_list(conference, &user);
		cancel_remote_calling(&user);
		/*conference camera_statistics. join conference with video.*/
		if (user.video) {
			conference->num_camera++;
		}
		if (conference->num_camera > conference->peak_camera) {
			conference->peak_camera = conference->num_camera;
		}
	}

	if (judge_if_nway_sfu_conference(chan, conference) ) {
		ast_indicate(chan, AST_CONTROL_JOIN_NWAY_CONF);
		notify_avs_slider_video_msid(chan, conference);
	}

	/* Play the corresponding muted sound to new user. */
	handle_new_user_muted_status(conference, &user);
#ifdef GRANDSTREAM_NETWORKS
	ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		ami_send_status = atoi(result);
	}

#ifdef BUILD_ON_CLOUD_PBX
	int gdms_record_support = 0;
	memset(result, 0, sizeof(result));
	ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		gdms_record_support = atoi(result);
	}

	/* Play the corresponding recoreded sound to new user. */
	if (!(ami_send_status & SEND_STATUS_RECORDING) && gdms_record_support) {
#else
	if (!(ami_send_status & SEND_STATUS_RECORDING) ) {
#endif // BUILD_ON_CLOUD_PBX
		handle_new_user_recorded_status(conference, &user);
	}

#ifdef BUILD_ON_CLOUD_PBX
	/* Play the sound that starts the recording to the new user. */
	if (!(ami_send_status & SEND_STATUS_VIDEOING) && gdms_record_support) {
#else
	if (!(ami_send_status & SEND_STATUS_VIDEOING)) {
#endif // BUILD_ON_CLOUD_PBX
		handle_new_user_videoed_status(conference, &user);
	}
#else
	/* Play the corresponding recoreded sound to new user. */
	handle_new_user_recorded_status(conference, &user);
#endif

	/*Set contact group to inuse*/
	conf_set_contact_group_inuse(&user);

		/* Hangup the mcb-create-channel to execute h@mcb-create in dialplan */
	if (ast_test_flag(&user.u_profile, USER_OPT_MCB_CREATE_CHAN_IN_USE)) {
		struct ast_channel *c = NULL;
		if ((c = ast_channel_get_by_name(user.u_profile.mcb_create_chan)) != NULL) {
			exec_mcb_authed(c);
			ast_channel_unref(c);
		}
	}

	/* Remove bridge from channel, if it is joined user, to avoid hangup joind user */
	if (ast_test_flag(&user.u_profile, USER_OPT_JOINEDUSER)) {
		if (ast_channel_internal_bridge(chan)) {
			ast_channel_internal_bridge_set(chan, NULL);
		}
	}

	/*playing moh after all voice prompts*/
	conf_moh_unsuspend(&user);

	conf_user_join_handle_one_person_warning(&user);
	handle_mcb_user_join(&user);

	/*Check if it needs to update host msid*/
	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		if (conference->focus_chan) {
			chan_update_msid_for_focus(conference, conference->focus_chan, chan);
		} else {
			if (ast_test_flag(&user.u_profile, USER_OPT_ADMIN)) {
				sfu_topologies_on_moderator_change(conference, &user);
			} else {
				if (terminate_type == AST_TERMINAL_TYPE_SIP) {
					struct confbridge_user *cur_user = NULL;
					ao2_lock(conference);
					AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
						if (ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN)) {
							sfu_topologies_on_moderator_change(conference, cur_user);
							break;
						}
					}
					ao2_unlock(conference);
				}
			}
		}
	}

	if (!strcasecmp("seamless_transfer_user", S_OR(u_profile_name, ""))) {
		const char *seamless_transferee;
		pbx_builtin_setvar_helper(chan, "CONFERENCE_FEATURE", "seamless_transfer");

		/* added by zgmo for updating record in sdp */
		ast_indicate(chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);

		seamless_transferee = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFEREE_CHANNEL");
		if (!ast_strlen_zero(seamless_transferee) && strcmp(ast_channel_name(chan), seamless_transferee)) {
			struct ast_channel *seamless_transferee_chan = ast_channel_get_by_name(seamless_transferee);
			if (seamless_transferee_chan) {
				pbx_builtin_setvar_helper(seamless_transferee_chan, "REC_ENABLE",
					pbx_builtin_getvar_helper(chan, "REC_ENABLE"));
				ast_channel_unref(seamless_transferee_chan);
			}
		}
		/* end */
	}

	if (!ast_strlen_zero(conference->schedule_id_to_start)) {
		if (!ast_strlen_zero(conference->schedule_id_processing) && !strcmp(conference->schedule_id_to_start, conference->schedule_id_processing)) {
			ast_log(LOG_DEBUG, "schedule_id_to_start %s is already in processing, no need to notify alert!", conference->schedule_id_processing);
		} else {
			char *body = build_schedule_notify_alert_body(conference->name, conference->schedule_id_to_start, conference->schedule_start_time, get_schedule_notify_countdown(conference));
			if (body) {
				check_timeout_before_schedule_notify(conference);
				ast_sendtext(user.chan, body);
				ast_free(body);
			}
		}
	}
	conf_timeout_detect_and_resend_notify(conference, &user);

	ast_bridge_hangup_hook(&user.features, conf_user_hangup_hook, &user, NULL, AST_BRIDGE_HOOK_REMOVE_ON_PULL);
#endif

#ifdef GRANDSTREAM_NETWORKS
	if (ast_bridge_join_hook(&user.features, join_callback, &user, NULL, 0))
#else
	if (ast_bridge_join_hook(&user.features, join_callback, NULL, NULL, 0))
#endif
	{
		async_play_sound_ready(user.chan);
	}

	if(ast_test_flag(ast_channel_flags(chan), AST_FLAG_RECOVERYING))
	{
		char family[64] = {0};
		char value[32] = {0};
		char conf_size[32] = {0};
		char increase_conf_num = 0;
		int num = 0;

		snprintf(family, sizeof(family), "CONF_STATUS/%s",args.conf_name);
		ast_db_get("CONF_STATUS/GLOBAL", "CONF_SIZE", conf_size, sizeof(conf_size));
		if(0 == ast_db_get(family, "SIZE", value, sizeof(value)) )
		{
			if(0 == strcmp(value, "0"))
			{
				increase_conf_num = 1;
			}
			/* /CONF_STATUS/CONF_ROOM/SIZE */
			num = atoi(value)+1;
			sprintf(value, "%d", num);
			ast_db_put(family, "SIZE", value);
			/* /CONF_STATUS/CONF_ROOM/USERS */
			ast_db_get(family, "USERS", value, sizeof(value));
			num = atoi(value)+1;
			sprintf(value, "%d", num);
			ast_db_put(family, "USERS", value);
			/* /CONF_STATUS/GLOBAL/SIZE */
			ast_db_get("CONF_STATUS/GLOBAL", "SIZE", value, sizeof(value));
			num = atoi(value)+1;
			sprintf(value, "%d", num);
			ast_db_put("CONF_STATUS/GLOBAL", "SIZE", value);


		}
		else
		{
			increase_conf_num = 1;

			ast_db_put(family, "SIZE", "1");
			ast_db_put(family, "USERS", "1");
			if(0 == ast_db_get("CONF_STATUS/GLOBAL", "SIZE", value, sizeof(value)))
			{
				num = atoi(value)+1;
				sprintf(value, "%d", num);
				ast_db_put("CONF_STATUS/GLOBAL", "SIZE", value);
			}
			else
			{
				ast_db_put("CONF_STATUS/GLOBAL", "SIZE", "1");
			}
		}

		if(increase_conf_num)
		{
			if(conf_size[0] == '\0')
			{
				ast_db_put("CONF_STATUS/GLOBAL", "CONF_SIZE", "1");
			}
			else
			{
				num = atoi(conf_size) + 1;
				sprintf(conf_size, "%d", num);
				ast_db_put("CONF_STATUS/GLOBAL", "CONF_SIZE", conf_size);
			}
		}
		ast_channel_lock(chan);
		pbx_builtin_setvar_helper(chan, "CONF_SIZE_INCREASED", "1");
		ast_channel_unlock(chan);

	}

#ifdef GRANDSTREAM_NETWORKS
	if (conference->conf_end) {
		ast_log(LOG_DEBUG, "conf end can not join now!");
		res = ast_stream_and_wait(chan,
			conf_get_sound(CONF_SOUND_KICKED, conference->b_profile.sounds),
			"");
		send_leave_event(&user, conference);
		leave_conference(&user);
		conference = NULL;
		goto confbridge_cleanup;
	}

	if (ast_channel_is_avs_engine(chan)) {
		/*Set the application hooks for attended tranfering a call to the conference.
		 * But now disable it because there are many issues to be fixed for attended tranfer.
		 */
		//ast_channel_move_hooks_set(chan, &move_hooks, &user);
	}
#endif

	ast_bridge_join(conference->bridge,
		chan,
		NULL,
		&user.features,
		&user.tech_args,
		0);

	/* This is a catch-all in case joining the bridge failed or for some reason
	 * an async announcement got queued up and hasn't been told to play yet
	 */
	async_play_sound_ready(chan);

	if (!user.kicked && ast_check_hangup(chan)) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_RESULT", "HANGUP");
#ifdef GRANDSTREAM_NETWORKS
		terminate_cei_subscription(&user);
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	handle_focus_user_leave(&user, conference);
	send_leave_event(&user, conference);
	quiet = ast_test_flag(&user.u_profile, USER_OPT_QUIET);
	if (ast_channel_is_avs_engine(chan)) {
		ast_channel_move_hooks_set(chan, NULL, NULL);
	}
#endif

	/* if we're shutting down, don't attempt to do further processing */
	if (ast_shutting_down()) {
		/*
		 * Not taking any new calls at this time.  We cannot create
		 * the announcer channel if this is the first channel into
		 * the conference and we certainly cannot create any
		 * recording channel.
		 */
		leave_conference(&user);
		conference = NULL;
		goto confbridge_cleanup;
	}

	/* If this user was a video source, we need to clean up and possibly pick a new source. */
	handle_video_on_exit(conference, user.chan);

	/* if this user has a intro, play it when leaving */
	if (!quiet && !ast_strlen_zero(user.name_rec_location)) {
#ifdef GRANDSTREAM_NETWORKS
		synchronize_play_sound_file(conference, user.name_rec_location, -1);
		synchronize_play_sound_file(conference,
			conf_get_sound(CONF_SOUND_HAS_LEFT, conference->b_profile.sounds), -1);
#else
		async_play_sound_file(conference, user.name_rec_location, NULL);
		async_play_sound_file(conference,
			conf_get_sound(CONF_SOUND_HAS_LEFT, conference->b_profile.sounds), NULL);
#endif
		async_delete_name_rec(conference, user.name_rec_location);
		async_delete_task_pushed = 1;
	}

	/* play the leave sound */
#ifdef GRANDSTREAM_NETWORKS
	if (!quiet && !(2==conference->bridge->num_channels && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE)))
#else
	if (!quiet)
#endif
	{
		const char *leave_sound = conf_get_sound(CONF_SOUND_LEAVE, conference->b_profile.sounds);
#ifdef GRANDSTREAM_NETWORKS
		synchronize_play_sound_file(conference, leave_sound, -1);
#else
		async_play_sound_file(conference, leave_sound, NULL);
#endif
	}

	/* If the user was kicked from the conference play back the audio prompt for it */
	/* add by kmye */
#ifdef GRANDSTREAM_NETWORKS
	if ((!quiet || user.play_sound) && user.kicked && args.argc < 5 && !ast_test_flag(&user.u_profile, USER_OPT_MARKED_MCB)) {
		if (user.play_sound == CONF_HANGUP_CONFTIMEOUT) {
			res = ast_stream_and_wait(chan,
				conf_get_sound(CONF_SOUND_TIMEOUT, conference->b_profile.sounds),
				"");
		} else {
			res = ast_stream_and_wait(chan,
				conf_get_sound(CONF_SOUND_KICKED, conference->b_profile.sounds),
				"");
		}
	}
#else
	if (!quiet && user.kicked) {
		res = ast_stream_and_wait(chan,
			conf_get_sound(CONF_SOUND_KICKED, conference->b_profile.sounds),
			"");
	}
#endif
#ifdef GRANDSTREAM_NETWORKS
	/* The ref count should be increased by one,
	and the conference object will be released after the last member leave the conference.
	Reuse conference object here will cause a crash (Seamless transfer)*/
	ao2_ref(conference, +1);
#endif
	/* Easy as pie, depart this channel from the conference bridge */
	leave_conference(&user);
	/* add by kmye*/
#ifdef GRANDSTREAM_NETWORKS
	if (!conference->conf_end) {
		handle_auto_transfer_moderator(&user, conference);
		if	(args.argc > 4 && !ast_strlen_zero(args.extra_profile_name)) {
			ao2_lock(conference);
			if (conference->activeusers < 2) {
				AST_LIST_TRAVERSE(&conference->active_list, extra_user, list) {
					if (extra_user) {
						extra_user->kicked = 0;
						if (extra_user->chan) {
							kick_conference_participant(conference, ast_channel_name(extra_user->chan), NULL);
						}
					}
				}
			}
			ao2_unlock(conference);
		}
	}
	ao2_ref(conference, -1);
#endif
	conference = NULL;

	/* Restore volume adjustments to previous values in case they were changed */
	if (volume_adjustments[0]) {
		ast_audiohook_volume_set(chan, AST_AUDIOHOOK_DIRECTION_READ, volume_adjustments[0]);
	}
	if (volume_adjustments[1]) {
		ast_audiohook_volume_set(chan, AST_AUDIOHOOK_DIRECTION_WRITE, volume_adjustments[1]);
	}

confbridge_cleanup:
	if (!async_delete_task_pushed && !ast_strlen_zero(user.name_rec_location)) {
		ast_filedelete(user.name_rec_location, NULL);
	}
#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&user.b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		if(ast_test_flag(&user.u_profile, USER_OPT_ADMIN)) {
			cache_conf_remove_moderator(args.conf_name, chan);
		}
		ast_conf_put_main_video_resource(args.conf_name, chan);
	}
	if (is_conference_service && (terminate_type == AST_TERMINAL_TYPE_WEBRTC || terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		cache_conf_remove_user_by_channel_name(args.conf_name, ast_channel_name(chan));
	}
	destroy_conf_user_report_record(&user.report_record);

	/*Clear CONFERENCE_FEATURE*/
	if (!strcasecmp("seamless_transfer_user", S_OR(u_profile_name, ""))) {
		pbx_builtin_setvar_helper(chan, "CONFERENCE_FEATURE", "");
	}
	if (seamless_transfer_redirect_flag) {
		ast_increase_call_count();
	}

	user_exit_delay(&user);
#endif
	ast_bridge_features_cleanup(&user.features);
	conf_bridge_profile_destroy(&user.b_profile);
	return res;
}

#ifdef GRANDSTREAM_NETWORKS
int exec_mcb_authed(struct ast_channel *chan)
{
	struct ast_pbx_args args;
	struct ast_pbx *pbx;
	char *exten;
	char *context;
	int priority;
	int res;
	memset(&args, 0, sizeof(args));
	args.no_hangup_chan = 1;
	args.no_increase_call = 1;

	ast_channel_lock(chan);

	/*save off*/
	exten = ast_strdupa(ast_channel_exten(chan));
	context = ast_strdupa(ast_channel_context(chan));
	priority = ast_channel_priority(chan);
	pbx = ast_channel_pbx(chan);
	ast_channel_pbx_set(chan, NULL);

	/*set new*/
	ast_channel_exten_set(chan, "mcb-add-record");
	ast_channel_context_set(chan, "mcb-create");
	ast_channel_priority_set(chan, 1);

	ast_channel_unlock(chan);

	/*execute*/
	res = ast_pbx_run_args(chan, &args);

	/*restore*/
	ast_channel_lock(chan);

	ast_channel_exten_set(chan, exten);
	ast_channel_context_set(chan, context);
	ast_channel_priority_set(chan, priority);
	ast_channel_pbx_set(chan, pbx);

	ast_channel_unlock(chan);
	return res;
}

void handle_new_user_muted_status(struct confbridge_conference *conference, struct confbridge_user *user)
{
	int quiet = ast_test_flag(&user->u_profile, USER_OPT_QUIET);

	if (user->muted) {
		conf_update_user_mute(user);
		send_mute_event(user, conference, 0, 0);
		if (!quiet) {
			play_prompt_to_user(user, conf_get_sound(CONF_SOUND_MUTED, user->b_profile.sounds));
		}

	} else if (conference->muted && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		user->muted = conference->muted;
		conf_update_user_mute(user);
		send_mute_event(user, conference, 0, 0);
		if (!quiet) {
			play_prompt_to_user(user, conf_get_sound(CONF_SOUND_MUTED, user->b_profile.sounds));
			play_prompt_to_user(user, conf_get_sound(CONF_SOUND_PARTICIPANTS_MUTED, user->b_profile.sounds));
		}

	} else if (conference->muted && ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		if (!quiet) {
			play_prompt_to_user(user, conf_get_sound(CONF_SOUND_PARTICIPANTS_MUTED, user->b_profile.sounds));
		}

	}
}

/* TODO: May need new conf-is-recording prompt to be displaied here */
void handle_new_user_recorded_status(struct confbridge_conference *conference, struct confbridge_user *user)
{
	int quiet = ast_test_flag(&user->u_profile, USER_OPT_QUIET);

	int recording = conf_is_recording(conference);
	int record_paused = is_conf_record_paused(conference);
	if (recording && !record_paused) {
		if (!quiet && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
			play_prompt_to_user(user, conf_get_sound(CONF_SOUND_START_RECORD, user->b_profile.sounds));
		}

	} else if (!recording && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_CONFERENCE) && conference->recorded != CONF_RECORD_SHUTDOWN) {
		conf_start_record(conference, user->chan);
		if (!quiet && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
			play_prompt_to_user(user, conf_get_sound(CONF_SOUND_START_RECORD, user->b_profile.sounds));
		}
	}
}

void handle_new_user_videoed_status(struct confbridge_conference *conference, struct confbridge_user *user)
{
	int videoing = conf_is_videoing(conference);
	if (videoing && conference->play_videoing_sound) {
		if(conference->is_paging_conf){
			if(!user->had_play_tone ) {
				user->had_play_tone = 1;//set it before play, Bug 258837
				play_prompt_to_user(user, conf_get_sound(CONF_SOUND_START_PAGING_VIDEO, user->b_profile.sounds));
			}
			if(ast_channel_get_is_page_caller(user->chan))
				playPageVideoRecord_Callee(conference);//added by zyxia for bug244581.2022.12.28
		}
		else {
			if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
				if (conference->video_is_recording) {
					play_prompt_to_user(user, conf_get_sound(CONF_SOUND_START_VIDEO, user->b_profile.sounds));
				}
				else if (conference->focus_is_recording) {
					play_prompt_to_user(user, conf_get_sound(CONF_SOUND_START_FOCUS_VIDEO, user->b_profile.sounds));
				}
			}
		}
	} else if (!videoing && conference->play_videoing_sound && ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_RECORD_CONFERENCE) && conference->recorded != CONF_RECORD_SHUTDOWN) {

		if (conf_start_video_record(conference) == 0 && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)){
				play_prompt_to_user(user, conf_get_sound(CONF_SOUND_START_VIDEO, user->b_profile.sounds));
		}
	}


}

#endif

static int action_toggle_mute(struct confbridge_conference *conference,
			      struct confbridge_user *user,
			      struct ast_bridge_channel *bridge_channel)
{
	int mute;

	/* Toggle user level mute request. */
	mute = !user->muted;
#ifdef GRANDSTREAM_NETWORKS
	if (((user->muted && user->mute_type == CONF_MUTE_BY_HOST && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN))
		|| (user->muted && conference->muted && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN)))
		&& !conference->allow_user_unmute) {
		return play_file(bridge_channel, NULL, "you-have-no-permission") < 0;
	}
	conf_user_write_operator_log(user, mute ? CONF_OP_MUTE : CONF_OP_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
	generic_mute_unmute_user(conference, user, mute, 0, 0);
	if (mute) {
		user->mute_type = CONF_MUTE_BY_SELF;
	} else {
		user->mute_type = CONF_MUTE_INIT;
	}
#else
	generic_mute_unmute_user(conference, user, mute);
#endif

	return play_file(bridge_channel, NULL,
		conf_get_sound(mute ? CONF_SOUND_MUTED : CONF_SOUND_UNMUTED,
			conference->b_profile.sounds)) < 0;
}

static int action_toggle_binaural(struct confbridge_conference *conference,
		struct confbridge_user *user,
		struct ast_bridge_channel *bridge_channel)
{
	unsigned int binaural;
	ast_bridge_channel_lock_bridge(bridge_channel);
	binaural = !bridge_channel->binaural_suspended;
	bridge_channel->binaural_suspended = binaural;
	ast_bridge_unlock(bridge_channel->bridge);
	return play_file(bridge_channel, NULL, (binaural ?
				conf_get_sound(CONF_SOUND_BINAURAL_OFF, user->b_profile.sounds) :
				conf_get_sound(CONF_SOUND_BINAURAL_ON, user->b_profile.sounds))) < 0;
}

static int action_toggle_mute_participants(struct confbridge_conference *conference, struct confbridge_user *user)
{
	struct confbridge_user *cur_user = NULL;
	const char *sound_to_play;
	int mute;

	ao2_lock(conference);

	/* Toggle bridge level mute request. */
	mute = !conference->muted;
	conference->muted = mute;
#ifdef GRANDSTREAM_NETWORKS
	send_conf_mute_status_event(conference);
#endif

	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if (!ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN)) {
#ifdef GRANDSTREAM_NETWORKS
			if (cur_user->muted == mute) {
				cur_user->suspended_mute = 1;
				conf_user_bridge_suspend(cur_user);
				continue;
			}
			generic_mute_unmute_user(conference, cur_user, mute, 0, 0);
			if (mute) {
				cur_user->mute_type = CONF_MUTE_BY_HOST;
			} else {
				cur_user->mute_type = CONF_MUTE_INIT;
			}
#else
			/* Set user level to bridge level mute request. */
			cur_user->muted = mute;
			conf_update_user_mute(cur_user);
#endif
		}
#ifdef GRANDSTREAM_NETWORKS
		else {
			conf_user_bridge_suspend(cur_user);
		}
#endif
	}

	ao2_unlock(conference);

#ifdef GRANDSTREAM_NETWORKS
	sound_to_play = conf_get_sound((mute ? CONF_SOUND_MUTED : CONF_SOUND_UNMUTED), user->b_profile.sounds);
	play_sound_helper(conference, sound_to_play, 0);

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if (ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN)) {
			conf_user_bridge_unsuspend(cur_user);
		}
		if (cur_user->suspended_mute) {
			conf_user_bridge_unsuspend(cur_user);
			cur_user->suspended_mute = 0;
		}
	}
	ao2_unlock(conference);

	conf_user_write_operator_log(user, mute ? CONF_OP_MUTE_PARTICIPANTS : CONF_OP_UNMUTE_PARTICIPANTS, NULL);
#else

	sound_to_play = conf_get_sound(
		mute ? CONF_SOUND_PARTICIPANTS_MUTED : CONF_SOUND_PARTICIPANTS_UNMUTED,
		conference->b_profile.sounds);

	if (strcmp(conference->b_profile.language, ast_channel_language(user->chan))) {
		/* The host needs to hear it seperately, as they don't get the audio from play_sound_helper */
		ast_stream_and_wait(user->chan, sound_to_play, "");

		/* Announce to the group that all participants are muted */
		ast_autoservice_start(user->chan);
		play_sound_file(conference, sound_to_play);
		ast_autoservice_stop(user->chan);
	} else {
		/* Playing the sound asynchronously lets the sound be heard by everyone at once */
		async_play_sound_file(conference, sound_to_play, user->chan);
	}
#endif
	return 0;
}

static int action_playback(struct ast_bridge_channel *bridge_channel, const char *playback_file)
{
	char *file_copy = ast_strdupa(playback_file);
	char *file = NULL;

	while ((file = strsep(&file_copy, "&"))) {
		if (ast_stream_and_wait(bridge_channel->chan, file, "")) {
			ast_log(LOG_WARNING, "Failed to playback file %s to channel\n", file);
			return -1;
		}
	}
	return 0;
}

static int action_playback_and_continue(struct confbridge_conference *conference,
	struct confbridge_user *user,
	struct ast_bridge_channel *bridge_channel,
	struct conf_menu *menu,
	const char *playback_file,
	const char *cur_dtmf,
	int *stop_prompts)
{
	int i;
	int digit = 0;
	char dtmf[MAXIMUM_DTMF_FEATURE_STRING];
	struct conf_menu_entry new_menu_entry = { { 0, }, };
	char *file_copy = ast_strdupa(playback_file);
	char *file = NULL;

	while ((file = strsep(&file_copy, "&"))) {
		if (ast_streamfile(bridge_channel->chan, file, ast_channel_language(bridge_channel->chan))) {
			ast_log(LOG_WARNING, "Failed to playback file %s to channel\n", file);
			return -1;
		}

		/* now wait for more digits. */
		if (!(digit = ast_waitstream(bridge_channel->chan, AST_DIGIT_ANY))) {
			/* streaming finished and no DTMF was entered */
			continue;
		} else if (digit == -1) {
			/* error */
			return -1;
		} else {
			break; /* dtmf was entered */
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_stopstream(bridge_channel->chan);
#endif
	if (!digit) {
#ifndef GRANDSTREAM_NETWORKS
		/* streaming finished on all files and no DTMF was entered */
		return -1;
#else
		digit = ast_waitfordigit(bridge_channel->chan, 5*1000);
		if (digit < 0) {
			/*Hangup or something error*/
			ast_debug(1,"Hangup or something error");
			return -1;
		} else if (!digit) {
			/* streaming finished on all files and no DTMF was entered */
			ast_debug(1,"streaming finished on all files and no DTMF was entered");
			return -1;
		} else {
			ast_debug(1,"get digit in 5s");
		}
#endif
	}
#ifndef GRANDSTREAM_NETWORKS
	ast_stopstream(bridge_channel->chan);
#endif

	/* If we get here, then DTMF has been entered, This means no
	 * additional prompts should be played for this menu entry */
	*stop_prompts = 1;

	/* If a digit was pressed during the payback, update
	 * the dtmf string and look for a new menu entry in the
	 * menu structure */
	ast_copy_string(dtmf, cur_dtmf, sizeof(dtmf));
	for (i = 0; i < (MAXIMUM_DTMF_FEATURE_STRING - 1); i++) {
		dtmf[i] = cur_dtmf[i];
		if (!dtmf[i]) {
			dtmf[i] = (char) digit;
			dtmf[i + 1] = '\0';
			i = -1;
			break;
		}
	}
	/* If i is not -1 then the new dtmf digit was _NOT_ added to the string.
	 * If this is the case, no new DTMF sequence should be looked for. */
	if (i != -1) {
		return 0;
	}

	if (conf_find_menu_entry_by_sequence(dtmf, menu, &new_menu_entry)) {
		execute_menu_entry(conference,
			user,
			bridge_channel,
			&new_menu_entry, menu);
		conf_menu_entry_destroy(&new_menu_entry);
	}
	return 0;
}

static int action_kick_last(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	struct confbridge_user *last_user = NULL;
	int isadmin = ast_test_flag(&user->u_profile, USER_OPT_ADMIN);

	if (!isadmin) {
		play_file(bridge_channel, NULL,
			conf_get_sound(CONF_SOUND_ERROR_MENU, conference->b_profile.sounds));
		ast_log(LOG_WARNING, "Only admin users can use the kick_last menu action. Channel %s of conf %s is not an admin.\n",
			ast_channel_name(bridge_channel->chan),
			conference->name);
		return -1;
	}

	ao2_lock(conference);
	last_user = AST_LIST_LAST(&conference->active_list);
	if (!last_user) {
		ao2_unlock(conference);
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	conf_user_write_operator_log(user, CONF_OP_KICK_LAST, NULL);
#endif

	if (last_user == user || ast_test_flag(&last_user->u_profile, USER_OPT_ADMIN)) {
		ao2_unlock(conference);
		play_file(bridge_channel, NULL,
			conf_get_sound(CONF_SOUND_ERROR_MENU, conference->b_profile.sounds));
	} else if (!last_user->kicked) {
#ifdef GRANDSTREAM_NETWORKS
		if(!strstr(ast_channel_name(last_user->chan),"PJSIP/trunk")) //added by zyxia for Bug 247244,2023.1.3
			publish_cei_unsub(conference->name, ast_channel_name(last_user->chan), ast_channel_caller(last_user->chan)->id.number.str);
#endif
		last_user->kicked = 1;
		pbx_builtin_setvar_helper(last_user->chan, "CONFBRIDGE_RESULT", "KICKED");
		ast_bridge_remove(conference->bridge, last_user->chan);
		ao2_unlock(conference);
	}
#ifdef GRANDSTREAM_NETWORKS
	else {
		ao2_unlock(conference);
	}
#endif
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int action_kick_mcb(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	struct confbridge_user *cur_user = NULL;
	RAII_VAR(struct ast_str *, extra_msg, ast_str_create(128), ast_free);

	ast_log(LOG_WARNING, "[TK] flags: %X", user->u_profile.flags);

	// Sanity check
	if (!ast_test_flag(&user->u_profile, USER_OPT_ADMIN) &&
		!ast_test_flag(&user->u_profile, USER_OPT_CAN_INVITE)) {
		if (bridge_channel != NULL) {
			ast_stream_and_wait(bridge_channel->chan,
			conf_get_sound(CONF_SOUND_ERROR_MENU, user->b_profile.sounds), "");
		}

		ast_log(LOG_WARNING, "Only admin/can-invite users can use the kick_mcb menu action. "
							 "Channel %s of conf %s is not an admin.\n",
							 (bridge_channel != NULL) ? ast_channel_name(bridge_channel->chan) : "unknown",
							 conference->name);

		return -1;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		ast_log(LOG_WARNING, "[TK] chan: %s, USER_OPT_MARKED_MCB: 0x%X",
							 ast_channel_name(cur_user->chan),
							 ast_test_flag(&cur_user->u_profile, USER_OPT_MARKED_MCB));

		// Kick the MCB out of conference
		if (ast_test_flag(&cur_user->u_profile, USER_OPT_MARKED_MCB)) {
			cur_user->kicked = 1;
			pbx_builtin_setvar_helper(cur_user->chan, "CONFBRIDGE_RESULT", "KICKED");
			ast_bridge_remove(conference->bridge, cur_user->chan);
			if (!ast_str_strlen(extra_msg)) {
				ast_str_append(&extra_msg, 0, "%s",
					S_COR(ast_channel_caller(cur_user->chan)->id.number.valid, ast_channel_caller(cur_user->chan)->id.number.str, "Anonymous"));
			} else {
				ast_str_append(&extra_msg, 0, ",%s",
					S_COR(ast_channel_caller(cur_user->chan)->id.number.valid, ast_channel_caller(cur_user->chan)->id.number.str, "Anonymous"));
			}
		}
	}
	ao2_unlock(conference);

	conf_user_write_operator_log(user, CONF_OP_KICK_MCB, ast_str_buffer(extra_msg));
	return 0;
}

static int action_kick_all_non_admin(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	struct confbridge_user *cur_user = NULL;
	RAII_VAR(struct ast_str *, extra_msg, ast_str_create(128), ast_free);

	ast_log(LOG_WARNING, "[WL] flags: %X", user->u_profile.flags);

	// Sanity check
	if (!ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		if (bridge_channel != NULL) {
			ast_stream_and_wait(bridge_channel->chan,
			conf_get_sound(CONF_SOUND_ERROR_MENU, user->b_profile.sounds), "");
		}

		ast_log(LOG_WARNING, "Only admin users can use the kick_all_non_admin menu action. "
							 "Channel %s of conf %s is not an admin.\n",
							 (bridge_channel != NULL) ? ast_channel_name(bridge_channel->chan) : "unknown",
							 conference->name);
		return -1;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		ast_log(LOG_WARNING, "[WL] chan: %s, USER_OPT_ADMIN: 0x%X",
							 ast_channel_name(cur_user->chan),
							 ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN));

		// Kick non-administrator users
		if (!ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN)) {
			publish_cei_unsub(conference->name, ast_channel_name(cur_user->chan), ast_channel_caller(cur_user->chan)->id.number.str);
			cur_user->kicked = 1;
			pbx_builtin_setvar_helper(cur_user->chan, "CONFBRIDGE_RESULT", "KICKED");
			ast_bridge_remove(conference->bridge, cur_user->chan);
			if (!ast_str_strlen(extra_msg)) {
				ast_str_append(&extra_msg, 0, "%s",
					S_COR(ast_channel_caller(cur_user->chan)->id.number.valid, ast_channel_caller(cur_user->chan)->id.number.str, "Anonymous"));
			} else {
				ast_str_append(&extra_msg, 0, ",%s",
					S_COR(ast_channel_caller(cur_user->chan)->id.number.valid, ast_channel_caller(cur_user->chan)->id.number.str, "Anonymous"));
			}
		}
	}
	ao2_unlock(conference);

	conf_user_write_operator_log(user, CONF_OP_KICK_PARTICIPANTS, ast_str_buffer(extra_msg));

	return 0;
}

static int action_toggle_record(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	int recording;
	int videoing;
	int res = 0;
	char result[32] = {0};
	int ami_send_status = 0;
	enum avs_control_type opt;

	/* Toggle user level record request. */
	recording = conf_is_recording(conference);
	videoing  = conf_is_videoing(conference);
	ast_log(LOG_WARNING, "[WL] flags: %d", recording);
	ast_log(LOG_WARNING, "[video] flags: %d", videoing);

	// Sanity check
	if (!ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		ast_stream_and_wait(bridge_channel->chan,
		conf_get_sound(CONF_SOUND_ERROR_MENU, user->b_profile.sounds), "");
		ast_log(LOG_WARNING, "Only admin users can use the toggle_record menu action. "
							 "Channel %s of conf %s is not an admin.\n",
							 ast_channel_name(bridge_channel->chan),
							 conference->name);
		return -1;
	}

	if (videoing)
	{
		ast_stream_and_wait(bridge_channel->chan,
		conf_get_sound(CONF_SOUND_HAS_VIDEOED, user->b_profile.sounds), "");
		ast_log(LOG_WARNING, "recording and video cannot be switched on at the same time");
		return -1;
	}

	ast_safe_nvram_get("ami_send_status", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		ami_send_status = atoi(result);
	}

#ifdef BUILD_ON_CLOUD_PBX
	int gdms_record_support = 0;
	memset(result, 0, sizeof(result));
	ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
	if (strlen(result) > 0) {
		gdms_record_support = atoi(result);
	}

	if ( (ami_send_status & SEND_STATUS_RECORDING) || !gdms_record_support ) {
#else
	if ( (ami_send_status & SEND_STATUS_RECORDING) ) {
#endif // BUILD_ON_CLOUD_PBX
		ast_log(LOG_WARNING, "Recording cannot be enabled on when the storage threshold is exceeded");
		return -1;
	}

	if (recording && 
	    (!is_record_merge_enable() || (is_record_merge_enable() && !is_conf_record_paused(conference))) ) {
		ao2_lock(conference);
		conf_user_write_operator_log(user, CONF_OP_END_RECORD, NULL);

		if (is_record_merge_enable()) {
			opt = AVS_CONTROL_PAUSE_RECORD;
			res = conf_pause_record(conference);
		} else {
			opt = AVS_CONTROL_STOP_RECORD;
			res = conf_stop_record(conference);
		}

		if(ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)){//added by zyxia for Bug 260093
			ast_stream_and_wait(bridge_channel->chan,
				conf_get_sound(CONF_SOUND_RECORD_OFF, user->b_profile.sounds), "");
			play_sound_file(conference,
				conf_get_sound(CONF_SOUND_RECORD_OFF, user->b_profile.sounds));
		}
		ao2_unlock(conference);
	} else {
		ao2_lock(conference);
		conf_user_write_operator_log(user, CONF_OP_START_RECORD, NULL);
		if(ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)){//added by zyxia for Bug 260093
			ast_stream_and_wait(bridge_channel->chan,
				conf_get_sound(CONF_SOUND_START_RECORD, user->b_profile.sounds), "");
			play_sound_file(conference,
				conf_get_sound(CONF_SOUND_START_RECORD, user->b_profile.sounds));
		}

		if (is_record_merge_enable() && is_conf_record_paused(conference)) {
			opt = AVS_CONTROL_RESUME_RECORD;
			res = conf_resume_record(conference, user->chan);
		} else {
			opt = AVS_CONTROL_START_RECORD;
			res = conf_start_record(conference, user->chan);
		}
		ao2_unlock(conference);
	}

	if (!res) {

		int status = 0;
		if (opt == AVS_CONTROL_START_RECORD || opt == AVS_CONTROL_RESUME_RECORD) {
			status = 1;
		}

		send_record_status_event(conference, "audio", status);
	}

	return res;
}

static struct conf_contact_group *find_contact_group_by_index(struct ao2_container *c, int index)
{
	struct conf_contact_group *contact_group = NULL;
	struct ao2_iterator iter = {0};
	int idx = 0;

	ao2_lock(c);
	iter = ao2_iterator_init(c, 0);
	while ((contact_group = ao2_iterator_next(&iter))) {
		if ((++idx) == index) {
			break;
		}

		ao2_cleanup(contact_group);
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(c);

	return contact_group;
}

static struct conf_contact_group *find_contact_group_by_name(struct ao2_container *c, const char *name)
{
	struct conf_contact_group *contact_group = NULL;
	struct ao2_iterator iter = {0};

	ao2_lock(c);
	iter = ao2_iterator_init(c, 0);
	while ((contact_group = ao2_iterator_next(&iter))) {
		if (!strcmp(contact_group->name, name)) {
			break;
		}

		ao2_cleanup(contact_group);
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(c);

	return contact_group;
}

static void conf_set_contact_group_inuse(struct confbridge_user *user)
{
	struct conf_contact_group *contact_group = NULL;
	const char *contact_group_name = NULL;

	if (!user || !user->chan || !user->conference) {
		return;
	}

	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB)) {
		return;
	}

	contact_group_name = pbx_builtin_getvar_helper(user->chan, "CONTACT_GROUP");
	if (!contact_group_name) {
		return;
	}

	contact_group = find_contact_group_by_name(conference_contact_groups, contact_group_name);
	if (!contact_group) {
		return;
	}
	ao2_lock(contact_group);
	if (!contact_group->inuse) {
		contact_group->inuse = 1;
		ast_string_field_set(contact_group, conference, user->conference->name);
	}
	ao2_unlock(contact_group);
	ao2_cleanup(contact_group);
}

static int contact_group_playback(struct ast_channel *chan, const char *playback_file)
{
	int digit = 0;
	char *file_copy = ast_strdupa(playback_file);
	char *file = NULL;

	while ((file = strsep(&file_copy, "&"))) {
		/*Play prompts and wait digits*/
		digit = ast_stream_and_wait(chan, file, AST_DIGIT_ANY);
		if (!digit) {
			continue;
		} else if (digit < 0) {
			ast_log(LOG_WARNING, "Failed to playback file %s to channel %s\n", file, ast_channel_name(chan));
			break;
		} else {
			break;
		}
	}

	return digit;
}

static void contact_group_invite_members(struct confbridge_user *user, struct confbridge_conference *conference, struct conf_contact_group *contact_group)
{
	char cmd[256] = {0};

	conf_user_write_operator_log(user, CONF_OP_INVITE_CONTACT_GROUP, contact_group->name);
	//edit by zyxia for Bug 188821 set  CONF_MENU_CHAN var 2021.3.11
	snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/DB2File_PY/conf_contact_group_call.py --originator '%s' --conf %s --group '%s' --channame '%s'",
		S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"), conference->name, contact_group->name,ast_channel_name(user->chan));
	ast_log(LOG_DEBUG, "contact_group_invite_members cmd:%s\n",cmd);
	ast_safe_system2(cmd);
}

static int contact_group_authenticate(struct conf_contact_group *contact_group, struct ast_channel *chan)
{
	char pin_guess[MAX_PIN+1] = { 0, };
	const char *pin = contact_group->pin;
	char *tmp = pin_guess;
	int i, res;
	unsigned int len = MAX_PIN ;

	/* give them three tries to get the pin right */
	for (i = 0; i < 3; i++) {
		if (ast_app_getdata(chan, (!strlen(pin_guess) ? "vm-password" : ""), tmp, len, 0) >= 0) {
			if (!strcasecmp(pin, pin_guess)) {
				return 0;
			}
		}

		if (i == 2) {
			break;
		}

		ast_streamfile(chan, "auth-incorrect", ast_channel_language(chan));
		res = ast_waitstream(chan, AST_DIGIT_ANY);
		if (res > 0) {
			/* Account for digit already read during ivalid pin playback
			 * resetting pin buf. */
			pin_guess[0] = res;
			pin_guess[1] = '\0';
			tmp = pin_guess + 1;
			len = MAX_PIN - 1;
		} else {
			/* reset pin buf as empty buffer. */
			pin_guess[0] = '\0';
			tmp = pin_guess;
			len = MAX_PIN;
		}
	}

	return -1;
}

static int contact_group_menu_play_and_wait(struct ao2_container *c, struct ast_channel *chan)
{
	struct ao2_iterator iter = {0};
	struct conf_contact_group *contact_group = NULL;
	struct ast_str *files = NULL;
	int index = 0;
	int digit = 0;

	digit = ast_stream_and_wait(chan, "conf-invite-contact-group", AST_DIGIT_ANY);
	if (digit > 0) {
		/*stop stream*/
		ast_stopstream(chan);
		return digit;
	}

	if (!(files = ast_str_create(64))) {
		ast_log(LOG_ERROR, "Couldn't malloc a ast string.\n");
		return -1;
	}

	ao2_lock(c);
	iter = ao2_iterator_init(c, 0);
	while ((contact_group = ao2_iterator_next(&iter))) {
		ao2_unlock(c);

		++index;
		ast_str_reset(files);
		/*Get prompts*/
		if (ast_strlen_zero(contact_group->prompt) || !strcmp(contact_group->prompt, "default")) {
			ast_str_set(&files, 0, "conf-contact-group&digits/%d&silence/1", index);
		} else {
			ast_str_set(&files, 0, "record/%s", contact_group->prompt);
		}

		ast_str_append(&files, 0, "&press&digits/%d", index);
		digit = contact_group_playback(chan, ast_str_buffer(files));
		/*wait a digit*/
		if (digit <= 0) {
			digit = ast_waitfordigit(chan, 500);
		}

		ao2_cleanup(contact_group);
		ao2_lock(c);

		if (digit > 0) {
			break;
		}
	}
	ao2_iterator_destroy(&iter);
	ao2_unlock(c);

	ast_free(files);

	/*stop stream*/
	ast_stopstream(chan);

	/*wait a digit*/
	if (digit <= 0) {
		digit = ast_waitfordigit(chan, 5000);
	}

	return digit;
}

static int action_invite_contact_group(struct confbridge_conference *conference,
	struct confbridge_user *user,
	struct ast_bridge_channel *bridge_channel)
{
	int res = 0;
	struct ao2_container *c = NULL;
	struct conf_contact_group *contact_group = NULL;
	int digit = 0;
	int tries = 3;

	/* Added by yfyang, for Bug 143094 */
	if (!conference || conference->locked) {
		ast_log(LOG_NOTICE, "action_invite_contact_group, conf locked, chan[%s]", bridge_channel->chan? ast_channel_name(bridge_channel->chan):"");
		contact_group_playback(bridge_channel->chan, "im-sorry&conf-lockednow");
		return 0;
	}

	c = ao2_bump(conference_contact_groups);
	if (ao2_container_count(c) == 0) {
		ast_stream_and_wait(bridge_channel->chan, "set-valid-conf-contact-group", AST_DIGIT_NONE);
		goto out;
	}

	while (tries--) {
		contact_group = NULL;

		if (digit <= 0) {
			digit = contact_group_menu_play_and_wait(c, bridge_channel->chan);
		}

		if (digit <= 0) {
			continue;
		}

		digit -= '0';
		if ((digit < 1) || (digit > ao2_container_count(c))) {
			digit = ast_stream_and_wait(bridge_channel->chan, "wrong-number", AST_DIGIT_ANY);
			continue;
		} else {
			contact_group = find_contact_group_by_index(c, digit);
			if (!contact_group) {
				ast_log(LOG_WARNING, "Couldn't find the contact group by index %d", digit);
				digit = 0;
				continue;
			}

			ao2_lock(contact_group);
			/*check if the contact group is inuse*/
			if (contact_group->inuse) {
				ast_log(LOG_DEBUG, "The contact group %s is inuse.", contact_group->name);
				ao2_unlock(contact_group);
				ao2_cleanup(contact_group);
				contact_group = NULL;
				digit = contact_group_playback(bridge_channel->chan, "conf-contact-group&is-in-use");
				continue;
			}
			/*set it to inuse*/
			contact_group->inuse = 1;
			ast_string_field_set(contact_group, conference, conference->name);
			ao2_unlock(contact_group);
		}

		if (contact_group) {
			break;
		}

		digit = 0;
	}

	if (contact_group) {
		if ((res = contact_group_authenticate(contact_group, bridge_channel->chan))) {
			ast_log(LOG_WARNING, "Failed to authenticate contact group %s", contact_group->name);
			contact_group_playback(bridge_channel->chan, "sorry-cant-let-you-do-that&auth-thankyou");
			/*clear inuse*/
			ao2_lock(contact_group);
			contact_group->inuse = 0;
			ast_string_field_set(contact_group, conference, "");
			ao2_unlock(contact_group);
		} else {
			contact_group_invite_members(user, conference, contact_group);
		}

		ao2_cleanup(contact_group);
	} else {
		contact_group_playback(bridge_channel->chan, "sorry-cant-let-you-do-that&auth-thankyou");
	}

out:
	/*stop stream*/
	ast_stopstream(bridge_channel->chan);

	ao2_cleanup(c);

	return res;
}
#endif

static int action_dialplan_exec(struct ast_bridge_channel *bridge_channel, struct conf_menu_action *menu_action)
{
	struct ast_pbx_args args;
	struct ast_pbx *pbx;
	char *exten;
	char *context;
	int priority;
	int res;

	memset(&args, 0, sizeof(args));
	args.no_hangup_chan = 1;
#ifdef GRANDSTREAM_NETWORKS
	args.no_increase_call = 1;
#endif

	ast_channel_lock(bridge_channel->chan);

	/*save off*/
	exten = ast_strdupa(ast_channel_exten(bridge_channel->chan));
	context = ast_strdupa(ast_channel_context(bridge_channel->chan));
	priority = ast_channel_priority(bridge_channel->chan);
	pbx = ast_channel_pbx(bridge_channel->chan);
	ast_channel_pbx_set(bridge_channel->chan, NULL);

	/*set new*/
	ast_channel_exten_set(bridge_channel->chan, menu_action->data.dialplan_args.exten);
	ast_channel_context_set(bridge_channel->chan, menu_action->data.dialplan_args.context);
	ast_channel_priority_set(bridge_channel->chan, menu_action->data.dialplan_args.priority);

	ast_channel_unlock(bridge_channel->chan);

	/*execute*/
	res = ast_pbx_run_args(bridge_channel->chan, &args);

	/*restore*/
	ast_channel_lock(bridge_channel->chan);

	ast_channel_exten_set(bridge_channel->chan, exten);
	ast_channel_context_set(bridge_channel->chan, context);
	ast_channel_priority_set(bridge_channel->chan, priority);
	ast_channel_pbx_set(bridge_channel->chan, pbx);

	ast_channel_unlock(bridge_channel->chan);
	return res;
}

static int execute_menu_entry(struct confbridge_conference *conference,
	struct confbridge_user *user,
	struct ast_bridge_channel *bridge_channel,
	struct conf_menu_entry *menu_entry,
	struct conf_menu *menu)
{
	struct conf_menu_action *menu_action;
	int isadmin = ast_test_flag(&user->u_profile, USER_OPT_ADMIN);
	int stop_prompts = 0;
	int res = 0;

	AST_LIST_TRAVERSE(&menu_entry->actions, menu_action, action) {
		switch (menu_action->id) {
		case MENU_ACTION_TOGGLE_MUTE:
			res |= action_toggle_mute(conference, user, bridge_channel);
			break;
		case MENU_ACTION_TOGGLE_BINAURAL:
			action_toggle_binaural(conference, user, bridge_channel);
			break;
		case MENU_ACTION_ADMIN_TOGGLE_MUTE_PARTICIPANTS:
			if (!isadmin) {
				break;
			}
			action_toggle_mute_participants(conference, user);
			break;
		case MENU_ACTION_PARTICIPANT_COUNT:
			announce_user_count(conference, user, bridge_channel);
			break;
		case MENU_ACTION_PLAYBACK:
			if (!stop_prompts) {
				res |= action_playback(bridge_channel, menu_action->data.playback_file);
				ast_test_suite_event_notify("CONF_MENU_PLAYBACK",
					"Message: %s\r\nChannel: %s",
					menu_action->data.playback_file, ast_channel_name(bridge_channel->chan));
			}
			break;
		case MENU_ACTION_RESET_LISTENING:
			ast_audiohook_volume_set(user->chan, AST_AUDIOHOOK_DIRECTION_WRITE, 0);
			break;
		case MENU_ACTION_RESET_TALKING:
			ast_audiohook_volume_set(user->chan, AST_AUDIOHOOK_DIRECTION_READ, 0);
			break;
		case MENU_ACTION_INCREASE_LISTENING:
			ast_audiohook_volume_adjust(user->chan,
				AST_AUDIOHOOK_DIRECTION_WRITE, 1);
#ifdef GRANDSTREAM_NETWORKS
			conf_user_write_operator_log(user, CONF_OP_INCREASE_LISTEN_VOLUME, NULL);
#endif
			break;
		case MENU_ACTION_DECREASE_LISTENING:
			ast_audiohook_volume_adjust(user->chan,
				AST_AUDIOHOOK_DIRECTION_WRITE, -1);
#ifdef GRANDSTREAM_NETWORKS
			conf_user_write_operator_log(user, CONF_OP_DECREASE_LISTEN_VOLUME, NULL);
#endif
			break;
		case MENU_ACTION_INCREASE_TALKING:
			ast_audiohook_volume_adjust(user->chan,
				AST_AUDIOHOOK_DIRECTION_READ, 1);
#ifdef GRANDSTREAM_NETWORKS
			conf_user_write_operator_log(user, CONF_OP_INCREASE_TALK_VOLUME, NULL);
#endif
			break;
		case MENU_ACTION_DECREASE_TALKING:
			ast_audiohook_volume_adjust(user->chan,
				AST_AUDIOHOOK_DIRECTION_READ, -1);
#ifdef GRANDSTREAM_NETWORKS
			conf_user_write_operator_log(user, CONF_OP_DECREASE_TALK_VOLUME, NULL);
#endif
			break;
		case MENU_ACTION_PLAYBACK_AND_CONTINUE:
			user->menu_operation = MENU_ACTION_PLAYBACK_AND_CONTINUE;
			if (!(stop_prompts)) {
				res |= action_playback_and_continue(conference,
					user,
					bridge_channel,
					menu,
					menu_action->data.playback_file,
					menu_entry->dtmf,
					&stop_prompts);
			}
			break;
		case MENU_ACTION_DIALPLAN_EXEC:
#ifdef GRANDSTREAM_NETWORKS
			user->menu_operation = MENU_ACTION_DIALPLAN_EXEC;
			ast_avs_leave_bridge(bridge_channel->chan, conference->bridge, conference->bridge->avs_bridge_id, conference->bridge->avs_bridge_tick);
#endif

			res |= action_dialplan_exec(bridge_channel, menu_action);
#ifdef GRANDSTREAM_NETWORKS

			if (ast_channel_get_avs_bridgeid(bridge_channel->chan) <= 0 ||
				ast_channel_get_avs_bridgeid(bridge_channel->chan) != conference->bridge->avs_bridge_id) {
				ast_avs_join_bridge(bridge_channel->chan, conference->bridge, conference->bridge->avs_bridge_id, conference->bridge->avs_bridge_tick);
			}

			if (!ast_test_flag(&user->u_profile, USER_OPT_QUIET)) {
				/* Play the Join sound to the user. */
				play_file(bridge_channel, NULL, conf_get_sound(CONF_SOUND_JOIN, user->b_profile.sounds));
			} else {
				/* Add by bfdu for bug 153465, need to stop play ringback tone for invite user */
				if (bridge_channel->chan && ast_channel_generator(bridge_channel->chan)) {
					ast_log(LOG_DEBUG, "Stop play tone for chan:%s\n", ast_channel_name(bridge_channel->chan));
					ast_deactivate_generator(bridge_channel->chan);
				}
			}
#endif
			break;
		case MENU_ACTION_ADMIN_TOGGLE_LOCK:
			if (!isadmin) {
				break;
			}
			conference->locked = (!conference->locked ? 1 : 0);
#ifdef GRANDSTREAM_NETWORKS
			conf_user_write_operator_log(user, conference->locked ? CONF_OP_LOCK : CONF_OP_UNLOCK, NULL);
#endif
			res |= play_file(bridge_channel, NULL,
				conf_get_sound(
					conference->locked ? CONF_SOUND_LOCKED_NOW : CONF_SOUND_UNLOCKED_NOW,
					conference->b_profile.sounds)) < 0;
#ifdef GRANDSTREAM_NETWORKS
			res |= play_sound_file(conference, (conference->locked ?
								   conf_get_sound(CONF_SOUND_LOCKED_NOW, conference->b_profile.sounds) :
								   conf_get_sound(CONF_SOUND_UNLOCKED_NOW, conference->b_profile.sounds))) < 0;
			send_lock_status_event(conference, conference->locked);
#endif
			break;
		case MENU_ACTION_ADMIN_KICK_LAST:
			res |= action_kick_last(conference, bridge_channel, user);
			break;
		case MENU_ACTION_LEAVE:
			pbx_builtin_setvar_helper(bridge_channel->chan, "CONFBRIDGE_RESULT", "DTMF");
			ao2_lock(conference);
			ast_bridge_remove(conference->bridge, bridge_channel->chan);
			ast_test_suite_event_notify("CONF_MENU_LEAVE",
				"Channel: %s",
				ast_channel_name(bridge_channel->chan));
			ao2_unlock(conference);
			break;
		case MENU_ACTION_NOOP:
			break;
		case MENU_ACTION_SET_SINGLE_VIDEO_SRC:
			ao2_lock(conference);
			if (!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
				ast_bridge_set_single_src_video_mode(conference->bridge, bridge_channel->chan);
			}
			ao2_unlock(conference);
			break;
		case MENU_ACTION_RELEASE_SINGLE_VIDEO_SRC:
			handle_video_on_exit(conference, bridge_channel->chan);
			break;
#ifdef GRANDSTREAM_NETWORKS
		case MENU_ACTION_ADMIN_KICK_MCBs:      /* kick all MCBs, added by Weiling Liao*/
			res |= action_kick_mcb(conference, bridge_channel, user);
			break;
		case MENU_ACTION_ADMIN_KICK_ALL_NON_ADMIN:
			res |= action_kick_all_non_admin(conference, bridge_channel, user);
			break;
		case MENU_ACTION_ADMIN_TOGGLE_RECORD:
			res |= action_toggle_record(conference, bridge_channel, user);
			break;
		case MENU_ACTION_CHANGE_ROLE:
			res |= action_change_role(conference, bridge_channel, user);
			break;
		case MENU_ACTION_INVITE_CONTACT_GROUP:
			res |= action_invite_contact_group(conference, user, bridge_channel);
			break;
#endif
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	user->menu_operation = 0;
#endif

	return res;
}

#ifdef GRANDSTREAM_NETWORKS
/*!
 * \brief Prepare and publish an \ref ast_cei_unsub_type message to the
 * \ref ast_cei_sub_topic_all.
 */
static void publish_cei_unsub(char *conference, const char *channel, const char *cid_number)
{
    struct stasis_topic *cei_sub_topic = NULL;
    RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
    RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

    // Sanity check
    if (conference == NULL || channel == NULL) {
        return;
    }

    cei_sub_topic = ast_cei_sub_topic_all();
	if (cei_sub_topic == NULL) {
        ast_log(LOG_WARNING, "Failed to get ast_cei_sub_topic_all");
        return;
	}

    json_object = ast_json_pack("{s: s, s: s, s: s}",
                                "conference", conference,
                                "channel", channel,
                                "cid_number", cid_number?cid_number:"Anonymous");

    if (json_object == NULL) {
        ast_log(LOG_WARNING, "Failed to create json_object");
        return;
    }

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

    msg = stasis_message_create(ast_cei_unsub_type(), payload);

    if (msg == NULL) {
        ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_unsub_type");
        return;
    }

    stasis_publish(cei_sub_topic, msg);
}
#endif

int conf_handle_dtmf(struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user,
	struct conf_menu_entry *menu_entry,
	struct conf_menu *menu)
{
	/* See if music on hold is playing */
	conf_moh_suspend(user);
#ifdef GRANDSTREAM_NETWORKS
	ast_bridge_suspend(user->conference->bridge, user->chan);
#endif

	/* execute the list of actions associated with this menu entry */
	execute_menu_entry(user->conference, user, bridge_channel, menu_entry, menu);

#ifdef GRANDSTREAM_NETWORKS
	ast_bridge_unsuspend(user->conference->bridge, user->chan);
#endif

	/* See if music on hold needs to be started back up again */
	conf_moh_unsuspend(user);

	async_play_sound_ready(bridge_channel->chan);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int kick_conference_participant(struct confbridge_conference *conference,
	const char *channel, const char *originator)
#else
static int kick_conference_participant(struct confbridge_conference *conference,
	const char *channel)
#endif
{
	int res = -1;
	int match;
	struct confbridge_user *user = NULL;
	int all = !strcasecmp("all", channel);
	int participants = !strcasecmp("participants", channel);
#ifdef GRANDSTREAM_NETWORKS
	struct ast_frame f = { AST_FRAME_CONTROL, .subclass.integer = AST_CONTROL_STREAM_STOP };
	struct confbridge_user *originator_user = NULL;
#endif

	SCOPED_AO2LOCK(bridge_lock, conference);
#ifdef GRANDSTREAM_NETWORKS
	originator_user = find_active_user_by_channel(conference, originator);
#endif

	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user->kicked) {
			continue;
		}
		match = !strcasecmp(channel, ast_channel_name(user->chan));
		if (match || all
				|| (participants && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN))) {
#ifdef GRANDSTREAM_NETWORKS
			if(!strstr(ast_channel_name(user->chan),"PJSIP/trunk")) //added by zyxia for Bug 247244,2023.1.3
				publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
#endif
			user->kicked = 1;
			pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
			ast_bridge_remove(conference->bridge, user->chan);
#ifdef GRANDSTREAM_NETWORKS
			if (conference->only_warning.playing && conference->only_warning.user == user) {
				ast_queue_frame(user->chan, &f);
			}
			ast_channel_conf_hangupcause_set(user->chan, AST_CAUSE_CONF_KICK);
#endif
			res = 0;
			if (match) {
#ifdef GRANDSTREAM_NETWORKS
				if (originator_user) {
					conf_user_write_operator_log(originator_user, CONF_OP_KICK, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
				}
#endif
				return res;
			}
		}
	}
	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		if (user->kicked) {
			continue;
		}
		match = !strcasecmp(channel, ast_channel_name(user->chan));
		if (match || all
				|| (participants && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN))) {
#ifdef GRANDSTREAM_NETWORKS
			if(!strstr(ast_channel_name(user->chan),"PJSIP/trunk")) //added by zyxia for Bug 247244,2023.1.3
				publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
#endif
			user->kicked = 1;
			pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
			ast_bridge_remove(conference->bridge, user->chan);
#ifdef GRANDSTREAM_NETWORKS
			if (conference->only_warning.playing && conference->only_warning.user == user) {
				ast_queue_frame(user->chan, &f);
			}
			ast_channel_conf_hangupcause_set(user->chan, AST_CAUSE_CONF_KICK);
#endif
			res = 0;
			if (match) {
#ifdef GRANDSTREAM_NETWORKS
				if (originator_user) {
					conf_user_write_operator_log(originator_user, CONF_OP_KICK, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
				}
#endif
				return res;
			}
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	if (originator_user) {
		conf_user_write_operator_log(originator_user, CONF_OP_KICK_PARTICIPANTS, NULL);
	}
#endif

	return res;
}

static char *complete_confbridge_name(const char *line, const char *word, int pos, int state)
{
	int which = 0;
	struct confbridge_conference *conference;
	char *res = NULL;
	int wordlen = strlen(word);
	struct ao2_iterator iter;

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (!strncasecmp(conference->name, word, wordlen) && ++which > state) {
			res = ast_strdup(conference->name);
			ao2_ref(conference, -1);
			break;
		}
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	return res;
}

static char *complete_confbridge_participant(const char *conference_name, const char *line, const char *word, int pos, int state)
{
	int which = 0;
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user;
	char *res = NULL;
	int wordlen = strlen(word);

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		return NULL;
	}

	if (!strncasecmp("all", word, wordlen) && ++which > state) {
		return ast_strdup("all");
	}

	if (!strncasecmp("participants", word, wordlen) && ++which > state) {
		return ast_strdup("participants");
	}

	{
		SCOPED_AO2LOCK(bridge_lock, conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (!strncasecmp(ast_channel_name(user->chan), word, wordlen) && ++which > state) {
				res = ast_strdup(ast_channel_name(user->chan));
				return res;
			}
		}
		AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
			if (!strncasecmp(ast_channel_name(user->chan), word, wordlen) && ++which > state) {
				res = ast_strdup(ast_channel_name(user->chan));
				return res;
			}
		}
	}

	return NULL;
}

static char *handle_cli_confbridge_kick(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;
	int not_found;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge kick";
#ifdef GRANDSTREAM_NETWORKS
		e->usage =
			"Usage: confbridge kick <conference> <channel> [originator]\n"
			"       Kicks a channel out of the conference bridge.\n"
			"             (all to kick everyone, participants to kick non-admins).\n";
#else
		e->usage =
			"Usage: confbridge kick <conference> <channel>\n"
			"		Kicks a channel out of the conference bridge.\n"
			"			  (all to kick everyone, participants to kick non-admins).\n";
#endif
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		if (a->pos == 3) {
			return complete_confbridge_participant(a->argv[2], a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (a->argc != 4 && a->argc != 5) {
		return CLI_SHOWUSAGE;
	}
#else
	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
#endif

	conference = ao2_find(conference_bridges, a->argv[2], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "No conference bridge named '%s' found!\n", a->argv[2]);
		return CLI_SUCCESS;
	}
#ifdef GRANDSTREAM_NETWORKS
	not_found = kick_conference_participant(conference, a->argv[3], a->argc == 5 ? a->argv[4] : NULL);
#else
	not_found = kick_conference_participant(conference, a->argv[3]);
#endif
	ao2_ref(conference, -1);
	if (not_found) {
		if (!strcasecmp("all", a->argv[3]) || !strcasecmp("participants", a->argv[3])) {
			ast_cli(a->fd, "No participants found!\n");
		} else {
			ast_cli(a->fd, "No participant named '%s' found!\n", a->argv[3]);
		}
		return CLI_SUCCESS;
	}
	ast_cli(a->fd, "Kicked '%s' out of conference '%s'\n", a->argv[3], a->argv[2]);
	return CLI_SUCCESS;
}

static void handle_cli_confbridge_list_item(struct ast_cli_args *a, struct confbridge_user *user, int waiting)
{
	char flag_str[6 + 1];/* Max flags + terminator */
	int pos = 0;

	/* Build flags column string. */
	if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		flag_str[pos++] = 'A';
	}
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKEDUSER)) {
		flag_str[pos++] = 'M';
	}
	if (ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED)) {
		flag_str[pos++] = 'W';
	}
	if (ast_test_flag(&user->u_profile, USER_OPT_ENDMARKED)) {
		flag_str[pos++] = 'E';
	}
	if (user->muted) {
		flag_str[pos++] = 'm';
	}
	if (waiting) {
		flag_str[pos++] = 'w';
	}
	flag_str[pos] = '\0';

	ast_cli(a->fd, "%-30s %-6s %-16s %-16s %-16s %s\n",
		ast_channel_name(user->chan),
		flag_str,
		user->u_profile.name,
		user->conference->b_profile.name,
		user->menu_name,
		S_COR(ast_channel_caller(user->chan)->id.number.valid,
			ast_channel_caller(user->chan)->id.number.str, "<unknown>"));
}

static char *handle_cli_confbridge_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge list";
		e->usage =
			"Usage: confbridge list [<name>]\n"
			"       Lists all currently active conference bridges or a specific conference bridge.\n"
			"\n"
			"       When a conference bridge name is provided, flags may be shown for users. Below\n"
			"       are the flags and what they represent.\n"
			"\n"
			"       Flags:\n"
			"         A - The user is an admin\n"
			"         M - The user is a marked user\n"
			"         W - The user must wait for a marked user to join\n"
			"         E - The user will be kicked after the last marked user leaves the conference\n"
			"         m - The user is muted\n"
			"         w - The user is waiting for a marked user to join\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}

	if (a->argc == 2) {
		struct ao2_iterator iter;

		ast_cli(a->fd, "Conference Bridge Name           Users  Marked Locked Muted\n");
		ast_cli(a->fd, "================================ ====== ====== ====== =====\n");
		iter = ao2_iterator_init(conference_bridges, 0);
		while ((conference = ao2_iterator_next(&iter))) {
			ast_cli(a->fd, "%-32s %6u %6u %-6s %s\n",
				conference->name,
				conference->activeusers + conference->waitingusers,
				conference->markedusers,
				AST_CLI_YESNO(conference->locked),
				AST_CLI_YESNO(conference->muted));
			ao2_ref(conference, -1);
		}
		ao2_iterator_destroy(&iter);
		return CLI_SUCCESS;
	}

	if (a->argc == 3) {
		struct confbridge_user *user;

		conference = ao2_find(conference_bridges, a->argv[2], OBJ_KEY);
		if (!conference) {
			ast_cli(a->fd, "No conference bridge named '%s' found!\n", a->argv[2]);
			return CLI_SUCCESS;
		}
		ast_cli(a->fd, "Channel                        Flags  User Profile     Bridge Profile   Menu             CallerID\n");
		ast_cli(a->fd, "============================== ====== ================ ================ ================ ================\n");
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			handle_cli_confbridge_list_item(a, user, 0);
		}
		AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
			handle_cli_confbridge_list_item(a, user, 1);
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_SUCCESS;
	}

	return CLI_SHOWUSAGE;
}

/* \internal
 * \brief finds a conference by name and locks/unlocks.
 *
 * \retval 0 success
 * \retval -1 conference not found
 */
#ifdef GRANDSTREAM_NETWORKS
static int generic_lock_unlock_helper(int lock, const char *conference_name, const char *originator)
#else
static int generic_lock_unlock_helper(int lock, const char *conference_name)
#endif
{
	struct confbridge_conference *conference;
	int res = 0;

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		return -1;
	}
	ao2_lock(conference);
#ifdef GRANDSTREAM_NETWORKS
	if (conference->locked == lock) {
		ast_log(LOG_DEBUG, "conference already %s", lock ? "locked" : "unlocked");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return res;
	}
#endif
	conference->locked = lock;
	ast_test_suite_event_notify("CONF_LOCK", "Message: conference %s\r\nConference: %s", conference->locked ? "locked" : "unlocked", conference->b_profile.name);

#ifdef GRANDSTREAM_NETWORKS
	struct confbridge_user *user = NULL;
	struct confbridge_user *originator_user = find_active_user_by_channel(conference, originator);
	if (originator_user) {
		conf_user_write_operator_log(originator_user, lock ? CONF_OP_LOCK : CONF_OP_UNLOCK, NULL);
	}

	/*active list*/
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		/* See if music on hold is playing */
		conf_moh_suspend(user);
	}
	/*waiting list*/
	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		/* See if music on hold is playing */
		conf_moh_suspend(user);
	}

	play_sound_file(conference, conference->locked ?
                conf_get_sound(CONF_SOUND_LOCKED_NOW, conference->b_profile.sounds) :
                conf_get_sound(CONF_SOUND_UNLOCKED_NOW, conference->b_profile.sounds));
	send_lock_status_event(conference, lock);

	/*active list*/
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		/* See if music on hold needs to be started back up again */
		conf_moh_unsuspend(user);
	}
	/*waiting list*/
	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		/* See if music on hold needs to be started back up again */
		conf_moh_unsuspend(user);
	}
#endif

	ao2_unlock(conference);
	ao2_ref(conference, -1);

	return res;
}

#ifdef GRANDSTREAM_NETWORKS
static void conf_user_bridge_suspend(struct confbridge_user *user)
{
	struct ast_bridge_channel *bridge_channel = NULL;

	ast_channel_lock(user->chan);
	bridge_channel = ast_channel_get_bridge_channel(user->chan);
	ast_channel_unlock(user->chan);
	if (bridge_channel) {
		bridge_channel_suspend(bridge_channel);
		ao2_cleanup(bridge_channel);
	}
}

static void conf_user_bridge_unsuspend(struct confbridge_user *user)
{
	struct ast_bridge_channel *bridge_channel = NULL;

	ast_channel_lock(user->chan);
	bridge_channel = ast_channel_get_bridge_channel(user->chan);
	ast_channel_unlock(user->chan);
	if (bridge_channel) {
		bridge_channel_unsuspend(bridge_channel);
		ao2_cleanup(bridge_channel);
	}
}

int conf_mic_mute_unmute_helper(int mute, const char *conference_name, const char *chan_name)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;
	int res = -2;

	ast_log(LOG_DEBUG, "Enter conf_mic_mute_unmute_helper(mute:%d, chan:%s)", mute, chan_name);

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		return -1;
	}

	SCOPED_AO2LOCK(bridge_lock, conference);

	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		int match = !strncasecmp(chan_name, ast_channel_name(user->chan), strlen(chan_name));
		if (match)
		{
			ast_log(LOG_DEBUG, "User:%d, muted:%d, mic_muted:%d",
				user->user_no, user->muted, user->mic_muted);

			if (user->mic_muted == mute) {
				return res;
			}

			/* See if music on hold is playing */
			conf_moh_suspend(user);
			generic_mute_unmute_user(conference, user, mute, 1, 0);
			/* See if music on hold is playing */
			conf_moh_unsuspend(user);

			res = 0;
			conf_user_write_operator_log(user, mute ? CONF_OP_MUTE : CONF_OP_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
			return res;
		}
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		int match = !strncasecmp(chan_name, ast_channel_name(user->chan), strlen(chan_name));
		if (match)
		{
			if (user->mic_muted == mute) {
				return res;
			}
			/* See if music on hold is playing */
			conf_moh_suspend(user);
			generic_mute_unmute_user(conference, user, mute, 1, 0);
			/* See if music on hold is playing */
			conf_moh_unsuspend(user);

			res = 0;
			conf_user_write_operator_log(user, mute ? CONF_OP_MUTE : CONF_OP_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));

			return res;
		}
	}

	return res;
}

int conf_user_hand_up_helper(int hand_up, const char *conference_name, const char *chan_name, struct ast_channel *op_chan)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;
	int res = 0;

	if (!conference_name  || !op_chan) {
		return -1;
	}

	ast_log(LOG_DEBUG, "Enter conf_user_hand_up_helper(hand_up:%d, chan:%s)", hand_up, chan_name ? chan_name : "");

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		return -1;
	}

	SCOPED_AO2LOCK(bridge_lock, conference);
	if (!ast_strlen_zero(chan_name)) {
		if (strcasecmp(chan_name, "all")) {
			AST_LIST_TRAVERSE(&conference->active_list, user, list) {
				int match = !strncasecmp(chan_name, ast_channel_name(user->chan), strlen(chan_name));
				if (match)
				{
					ast_log(LOG_DEBUG, "User:%d, hand_up:%d", user->user_no, user->hand_up);

					if (user->hand_up == hand_up) {
						return res;
					}
					user->hand_up = hand_up;
					if (user->hand_up) {
						user->hand_up_time = ast_now_utc_timestamp();
					} else {
						user->hand_up_time = 0;
					}
					send_hand_up_status_event(user, conference);
					break;
				}
			}
		} else if (!strcasecmp(chan_name, "all")) {
				AST_LIST_TRAVERSE(&conference->active_list, user, list) {
					ast_log(LOG_DEBUG, "User:%d, hand_up:%d", user->user_no, user->hand_up);

					if (user->hand_up == hand_up) {
						continue;
					}
					user->hand_up = hand_up;
			}
			send_conference_act_event(conference, AST_CONF_ACT_TYPE_PUT_DOWN_ALL_HANDS);
		}
	} else {
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		int match = 0;
		if (ast_channel_name(op_chan) && ast_channel_name(user->chan)) {
			match = !strncasecmp(ast_channel_name(op_chan), ast_channel_name(user->chan), strlen(ast_channel_name(op_chan)));
		}
		if (match)
		{
			ast_log(LOG_DEBUG, "User:%d, hand_up:%d", user->user_no, user->hand_up);

			if (user->hand_up == hand_up) {
				return res;
			}
			user->hand_up = hand_up;
			if (user->hand_up) {
				user->hand_up_time = ast_now_utc_timestamp();
			} else {
				user->hand_up_time = 0;
			}
			send_hand_up_status_event(user, conference);
			break;
		}
	}
	}
	return res;
}

#endif

/* \internal
 * \brief finds a conference user by channel name and mutes/unmutes them.
 *
 * \retval 0 success
 * \retval -1 conference not found
 * \retval -2 user not found
 */
#ifdef GRANDSTREAM_NETWORKS
static int generic_mute_unmute_helper(int mute, const char *conference_name,
	const char *chan_name, const char *originator, int is_mic)
#else
static int generic_mute_unmute_helper(int mute, const char *conference_name,
	const char *chan_name)
#endif
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user;
	int all = !strcasecmp("all", chan_name);
	int participants = !strcasecmp("participants", chan_name);
	int res = -2;
#ifdef GRANDSTREAM_NETWORKS
	int count_num = 0;
	int mute_unmute_self = 0;

	struct confbridge_user *originator_user = NULL;

	ast_log(LOG_DEBUG, "Enter generic_mute_unmute_helper(mute:%d, chan:%s, orig:%s, is_mic:%d)",
		mute, chan_name, originator, is_mic);
#endif

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (participants) {
		conference->muted = mute;
		send_conf_mute_status_event(conference);
	}
#endif

	{
		SCOPED_AO2LOCK(bridge_lock, conference);

#ifdef GRANDSTREAM_NETWORKS
		originator_user = find_active_user_by_channel(conference, originator);
		if (originator_user && originator_user->chan && !ast_test_flag(&originator_user->u_profile, USER_OPT_ADMIN) && !is_mic) {
			if (!conference->allow_user_unmute) {
				ast_log(LOG_DEBUG, "participants no permission (un)mute!");
				return -1;
			} else {
				if (ast_channel_name(originator_user->chan) && strcasecmp(chan_name, ast_channel_name(originator_user->chan))) {
					ast_log(LOG_DEBUG, "participants no permission (un)mute other users[%s] VS [%s]!",
						ast_channel_name(originator_user->chan), chan_name);
					return -1;
				} else if (ast_channel_name(originator_user->chan) && !strcasecmp(chan_name, ast_channel_name(originator_user->chan))) {
					ast_log(LOG_DEBUG, "users[%s] VS [%s]!", ast_channel_name(originator_user->chan), chan_name);
					mute_unmute_self = 1;
				}
			}
		}
#endif

		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			int match = !strncasecmp(chan_name, ast_channel_name(user->chan),
				strlen(chan_name));
#ifdef GRANDSTREAM_NETWORKS
			if (match)
#else
			if (match || all
				|| (participants && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN)))
#endif
			{
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_DEBUG, "User:%d, muted:%d, mic_muted:%d",
					user->user_no, user->muted, user->mic_muted);

				if (is_mic) {
					if (user->mic_muted == mute) {
						return res;
					}
				} else {
					if (user->muted == mute) {
						return res;
					}

					if (mute) {
						user->talking_before_muted = (user->talking ? 1 : 0);
					}
				}
				/* See if music on hold is playing */
				conf_moh_suspend(user);
#endif

#ifdef GRANDSTREAM_NETWORKS
				generic_mute_unmute_user(conference, user, mute, is_mic, is_mic ? 0 : 1);
#else
				generic_mute_unmute_user(conference, user, mute);
#endif

#ifdef GRANDSTREAM_NETWORKS
				/* See if music on hold is playing */
				conf_moh_unsuspend(user);
				if (!is_mic) {
					if (mute && mute_unmute_self) {
						user->mute_type = CONF_MUTE_BY_SELF;
					} else if (mute && !mute_unmute_self) {
						user->mute_type = CONF_MUTE_BY_HOST;
					} else {
						user->mute_type = CONF_MUTE_INIT;
					}
				}
#endif
				res = 0;
				if (match) {
#ifdef GRANDSTREAM_NETWORKS
					if (originator_user) {
						if (is_mic) {
							conf_user_write_operator_log(originator_user, mute ? CONF_OP_MIC_MUTE : CONF_OP_MIC_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
						} else {
							conf_user_write_operator_log(originator_user, mute ? CONF_OP_MUTE : CONF_OP_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
						}
					}
#endif
					return res;
				}
			}
#ifdef GRANDSTREAM_NETWORKS
			else if (all || (participants && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN))) {
				if (user->muted == mute) {
					continue;
				}

				if (mute) {
					user->talking_before_muted = (user->talking ? 1 : 0);
				}
				res = 0;
				user->exe_mute = 1;
				if (!is_mic) {
					if (mute && mute_unmute_self) {
						user->mute_type = CONF_MUTE_BY_SELF;
					} else if (mute && !mute_unmute_self) {
						user->mute_type = CONF_MUTE_BY_HOST;
					} else {
						user->mute_type = CONF_MUTE_INIT;
					}
				}
			}
#endif
		}

		AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
			int match = !strncasecmp(chan_name, ast_channel_name(user->chan),
				strlen(chan_name));
#ifdef GRANDSTREAM_NETWORKS
			if (match)
#else
			if (match || all
				|| (participants && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN)))
#endif
			{
#ifdef GRANDSTREAM_NETWORKS
				if (is_mic) {
					if (user->mic_muted == mute) {
						return res;
					}
				} else {
					if (user->muted == mute) {
						return res;
					}

					if (mute) {
						user->talking_before_muted = (user->talking ? 1 : 0);
					}
				}
				/* See if music on hold is playing */
				conf_moh_suspend(user);
#endif

#ifdef GRANDSTREAM_NETWORKS
				generic_mute_unmute_user(conference, user, mute, is_mic, is_mic ? 0 : 1);
#else
				generic_mute_unmute_user(conference, user, mute);
#endif

#ifdef GRANDSTREAM_NETWORKS
				/* See if music on hold is playing */
				conf_moh_unsuspend(user);
#endif

				res = 0;
				if (match) {
#ifdef GRANDSTREAM_NETWORKS
					if (originator_user) {
						if (is_mic) {
							conf_user_write_operator_log(originator_user, mute ? CONF_OP_MIC_MUTE : CONF_OP_MIC_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
						} else {
							conf_user_write_operator_log(originator_user, mute ? CONF_OP_MUTE : CONF_OP_UNMUTE, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"));
						}
					}
#endif
					return res;
				}
			}
#ifdef GRANDSTREAM_NETWORKS
			else if (all || (participants && !ast_test_flag(&user->u_profile, USER_OPT_ADMIN))) {
				if (user->muted == mute) {
					continue;
				}

				if (mute) {
					user->talking_before_muted = (user->talking ? 1 : 0);
				}
				res = 0;
				user->exe_mute = 1;
			}
#endif
		}

#ifdef GRANDSTREAM_NETWORKS
		if (all || participants) {
			if (originator_user) {
				conf_user_write_operator_log(originator_user, mute ? CONF_OP_MUTE_PARTICIPANTS : CONF_OP_UNMUTE_PARTICIPANTS, NULL);
			}
		}

		if (!res) {
			/*active list*/
			AST_LIST_TRAVERSE(&conference->active_list, user, list) {
				if (user->exe_mute) {
					/* See if music on hold is playing */
					conf_moh_suspend(user);

					/* Set user level mute request. */
					user->muted = mute ? 1 : 0;
					conf_update_user_mute(user);
					if (mute) {
						if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
							send_muteall_event(user, conference, 1, count_num++);
						} else {
							send_mute_event(user, conference, 0, is_mic);
						}
					} else {
						if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
							send_unmuteall_event(user, conference, 1, count_num++);
							if (user->talking || user->talking_before_muted) {
								send_talking_event(user, conference);
							}
						} else {
							send_unmute_event(user, conference, 0, is_mic);
						}
					}
				} else {
					conf_user_bridge_suspend(user);
				}
			}
			/*waiting list*/
			count_num = 0;
			AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
				if (user->exe_mute) {
					/* See if music on hold is playing */
					conf_moh_suspend(user);

					/* Set user level mute request. */
					user->muted = mute ? 1 : 0;
					conf_update_user_mute(user);
					if (mute) {
						if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
							send_muteall_event(user, conference, 1, count_num++);
						} else {
							send_mute_event(user, conference, 0, is_mic);
						}
					} else {
						if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
							send_unmuteall_event(user, conference, 1, count_num++);
							if (user->talking || user->talking_before_muted) {
								send_talking_event(user, conference);
							}
						} else {
							send_unmute_event(user, conference, 0, is_mic);
						}
					}
				} else {
					conf_user_bridge_suspend(user);
				}
			}

			if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
				/*update talking*/
				RAII_VAR(struct ast_json *, talking_extras, NULL, ast_json_unref);
				talking_extras = ast_json_pack("{s: i}", "update_talking_only", 1);
				if (talking_extras) {
					send_conf_stasis(conference, NULL, cei_refresh_all_type(), talking_extras, 1);
				}
			}
			play_sound_file(conference, conf_get_sound(mute ? CONF_SOUND_MUTED : CONF_SOUND_UNMUTED, conference->b_profile.sounds));

			/*active list*/
			AST_LIST_TRAVERSE(&conference->active_list, user, list) {
				if (user->exe_mute) {
					/* See if music on hold needs to be started back up again */
					conf_moh_unsuspend(user);

					user->exe_mute = 0;
				} else {
					conf_user_bridge_unsuspend(user);
				}
			}
			/*waiting list*/
			AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
				if (user->exe_mute) {
					/* See if music on hold needs to be started back up again */
					conf_moh_unsuspend(user);

					user->exe_mute = 0;
				} else {
					conf_user_bridge_unsuspend(user);
				}
			}
		}
#endif
	}

	return res;
}

static int cli_mute_unmute_helper(int mute, struct ast_cli_args *a)
{
#ifdef GRANDSTREAM_NETWORKS
	int res;

	if (a->argc == 5) {
		res = generic_mute_unmute_helper(mute, a->argv[2], a->argv[3], a->argv[4], 0);
	} else if (a->argc == 6) {
		res = generic_mute_unmute_helper(mute, a->argv[2], a->argv[3], a->argv[4], !strcmp(a->argv[5], "mic") ? 1 : 0);
	} else {
		res = generic_mute_unmute_helper(mute, a->argv[2], a->argv[3], NULL, 0);
	}
#else
	int res = generic_mute_unmute_helper(mute, a->argv[2], a->argv[3]);
#endif

	if (res == -1) {
		ast_cli(a->fd, "No conference bridge named '%s' found!\n", a->argv[2]);
		return -1;
	} else if (res == -2) {
		if (!strcasecmp("all", a->argv[3]) || !strcasecmp("participants", a->argv[3])) {
			ast_cli(a->fd, "No participants found in conference %s\n", a->argv[2]);
		} else {
			ast_cli(a->fd, "No channel named '%s' found in conference %s\n", a->argv[3], a->argv[2]);
		}
		return -1;
	}
	ast_cli(a->fd, "%s %s from confbridge %s\n", mute ? "Muting" : "Unmuting", a->argv[3], a->argv[2]);
	return 0;
}

static char *handle_cli_confbridge_mute(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge mute";
#ifdef GRANDSTREAM_NETWORKS
		e->usage =
			"Usage: confbridge mute <conference> <channel> [originator]\n"
			"       Mute a channel in a conference.\n"
			"              (all to mute everyone, participants to mute non-admins)\n"
			"       If the specified channel is a prefix,\n"
			"       the action will be taken on the first\n"
			"       matching channel.\n";
#else
		e->usage =
			"Usage: confbridge mute <conference> <channel>\n"
			"		Mute a channel in a conference.\n"
			"			   (all to mute everyone, participants to mute non-admins)\n"
			"		If the specified channel is a prefix,\n"
			"		the action will be taken on the first\n"
			"		matching channel.\n";
#endif
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		if (a->pos == 3) {
			return complete_confbridge_participant(a->argv[2], a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (a->argc < 4 || a->argc > 6) {
		return CLI_SHOWUSAGE;
	}
#else
	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
#endif

	cli_mute_unmute_helper(1, a);

	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_unmute(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge unmute";
#ifdef GRANDSTREAM_NETWORKS
		e->usage =
			"Usage: confbridge unmute <conference> <channel> [originator]\n"
			"		Unmute a channel in a conference.\n"
			"			   (all to unmute everyone, participants to unmute non-admins)\n"
			"		If the specified channel is a prefix,\n"
			"		the action will be taken on the first\n"
			"		matching channel.\n";
#else
		e->usage =
			"Usage: confbridge unmute <conference> <channel>\n"
			"       Unmute a channel in a conference.\n"
			"              (all to unmute everyone, participants to unmute non-admins)\n"
			"       If the specified channel is a prefix,\n"
			"       the action will be taken on the first\n"
			"       matching channel.\n";
#endif
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		if (a->pos == 3) {
			return complete_confbridge_participant(a->argv[2], a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
#ifdef GRANDSTREAM_NETWORKS
	if (a->argc < 4 || a->argc > 6) {
		return CLI_SHOWUSAGE;
	}
#else
	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
#endif

	cli_mute_unmute_helper(0, a);

	return CLI_SUCCESS;
}

#ifdef GRANDSTREAM_NETWORKS
static void unsuspend_conference_moh(struct confbridge_conference *conference)
{
	struct confbridge_user *user = NULL;

	AST_LIST_TRAVERSE(&conference->active_list, user, list){
		conf_moh_unsuspend(user);
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list){
		conf_moh_unsuspend(user);
	}
}

static void suspend_conference_moh(struct confbridge_conference *conference)
{
	struct confbridge_user *user = NULL;

	AST_LIST_TRAVERSE(&conference->active_list, user, list){
		conf_moh_suspend(user);
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list){
		conf_moh_suspend(user);
	}
}

static char *handle_cli_confbridge_playback(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference = NULL;
	char *front = NULL;
	int res = 0;
	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge playback";
		e->usage =
			"Usage: confbridge playback <conference> <file>\n"
			"       Playback an audio file to a specified conference room.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[2], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_FAILURE;
	}
	ao2_lock(conference);
	/*suspend moh before playing sound file*/
	suspend_conference_moh(conference);
	char *filename = ast_strdupa(a->argv[3]);
	while (!res && (front = strsep(&filename, "&"))) {
		res = play_sound_file(conference,front);
	}
	/*unsuspend moh after playing sound file*/
	unsuspend_conference_moh(conference);
	ao2_unlock(conference);
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static int hangup_conference(struct confbridge_conference *conference, int hangup_type)
{
	char family[64] = {0};
	struct confbridge_user *user = NULL;
	int ignore = 0;
	struct ast_frame frame = {
		.frametype = AST_FRAME_DTMF_END,
		.subclass.integer = '#',
	};
	int has_notify = 0;

	ast_log(LOG_DEBUG, "hangup_conference %s hangup_type %d", conference->name, hangup_type);

	/*stop bridge to decrease the usage of avs cpu*/
	ast_avs_stop_bridge(conference->bridge, conference->bridge->avs_bridge_id, -1);

	SCOPED_AO2LOCK(bridge_lock, conference);

	if ((hangup_type == CONF_HANGUP_WEBHOST || hangup_type == CONF_HANGUP_PALYKICKED || hangup_type == CONF_HANGUP_MEETINGLIST) &&
		conference->b_profile.is_random_conf && !ast_strlen_zero(conference->schedule_id_processing)) {
		snprintf(family, sizeof(family)-1,"SCHEDULE/%s", conference->name);
		ast_db_del(family, "CONF_START");
		ast_log(LOG_DEBUG, "hangup_conference del CONF_START first for %s", conference->name);
	}

	if ((hangup_type == CONF_HANGUP_WEBHOST || hangup_type == CONF_HANGUP_PALYKICKED || hangup_type == CONF_HANGUP_MEETINGLIST) &&
		!ast_strlen_zero(conference->schedule_id_processing) &&
		!ast_strlen_zero(conference->schedule_utc_start_time)) {
		conference->schedule_hangup = 1;
	}

	if ((conference->activeusers + conference->waitingusers) > MAX_USER_NUM_NO_IGNORE_AVS_RESP) {
		ignore = 1;

		if (conference->playback_chan) {
			ast_channel_ignore_avs_response_set(conference->playback_chan, ignore);
		}

		if (conference->record_chan) {
			ast_channel_ignore_avs_response_set(conference->record_chan, ignore);
		}
	}
	conference->conf_end = 1;
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user->kicked) {
			continue;
		}
		/* Notify invitees who have not yet joined the meeting */
		if (!has_notify) {
			ast_indicate(user->chan, AST_CONTROL_END_CONF);
			ast_log(LOG_DEBUG, "Conference has end, Notify invitees who have not yet joined the meeting.");
		}
		has_notify++;
		if (user->menu_operation == MENU_ACTION_PLAYBACK_AND_CONTINUE) {
			ast_queue_frame(user->chan, &frame);
		}

		ast_channel_ignore_avs_response_set(user->chan, ignore);
		publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
		user->kicked = 1;
		pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
		ast_set_flag(&user->u_profile, USER_OPT_QUIET);
		if (hangup_type == CONF_HANGUP_CONFTIMEOUT || hangup_type == CONF_HANGUP_PALYKICKED) {
			user->play_sound = hangup_type;
		}
		ast_bridge_remove(conference->bridge, user->chan);
		if (user->menu_operation == MENU_ACTION_DIALPLAN_EXEC) {
			ast_queue_hangup(user->chan);
		}
		ast_channel_conf_hangupcause_set(user->chan, AST_CAUSE_CONF_END);
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		if (user->kicked) {
			continue;
		}
		if (user->menu_operation == MENU_ACTION_PLAYBACK_AND_CONTINUE) {
			ast_queue_frame(user->chan, &frame);
		}

		ast_channel_ignore_avs_response_set(user->chan, ignore);
		publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
		user->kicked = 1;
		pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
		ast_set_flag(&user->u_profile, USER_OPT_QUIET);
		if (hangup_type == CONF_HANGUP_CONFTIMEOUT || hangup_type == CONF_HANGUP_PALYKICKED) {
			user->play_sound = hangup_type;
		}
		ast_bridge_remove(conference->bridge, user->chan);
		if (user->menu_operation == MENU_ACTION_DIALPLAN_EXEC) {
			ast_queue_hangup(user->chan);
		}
		ast_channel_conf_hangupcause_set(user->chan, AST_CAUSE_CONF_END);
	}

	return 0;
}

static void clean_schedule_meeting_status(const char *conf_name, int conf_type)
{
	char family[256] = {0};
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
	char step[16] = {0};
	char start[16] = {0};
	char type[16] = {0};
	char delay[16] = {0};
	ast_log(LOG_DEBUG, "clean_schedule_meeting_status!");

	if (ast_strlen_zero(conf_name)) {
		ast_log(LOG_DEBUG, "conference name null!");
		return;
	}

	snprintf(family, sizeof(family), "SCHEDULE/%s", conf_name);
	if (conf_type < 0)
	{
		ast_db_get(family, "ConfType", type, sizeof(type));
		if (!strcmp(type, "video")) {
			conf_type = 1;
		} else if (!strcmp(type, "audio")) {
			conf_type = 0;
		} else {
			ast_log(LOG_DEBUG, "unknow conf type!");
			return;
		}
	}
	if (ast_db_get(family, "ScheduleID", schedule_id, sizeof(schedule_id)) == 0 && !ast_strlen_zero(schedule_id)) {
		ast_log(LOG_DEBUG, "clean_schedule_meeting_status %s!", schedule_id);
		snprintf(family, sizeof(family), "SCHEDULE/%s", schedule_id);
		ast_db_get(family, "DELAY", delay, sizeof(delay));
		if (!strcmp(delay, "yes")) {
			ast_log(LOG_DEBUG, "conf delay %s clean_schedule_meeting_status return, schedule_conference_end_delay will handle this case!", schedule_id);
			return;
		}
		ast_db_get(family, "STEP", step, sizeof(step));
		snprintf(family, sizeof(family), "SCHEDULE/%s", conf_name);
		ast_db_get(family, "CONF_START", start, sizeof(start));
		if (!strcmp(step, "call") || !strcmp(step, "earlyjoin") || !strcmp(start, "yes")) {
			char cmd[256] = {0};
			memset(family, 0, sizeof(family));
			snprintf(family, sizeof(family), "SCHEDULE/%s", schedule_id);
			ast_db_put(family, "confhangup", "yes");
			if (conf_type) {
				/* Must be executed in the background, otherwise schedule_conference_remove and release_schedule_meeting will fall into a deadlock*/
				snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/exe_mcm_reservation.py --type clean --id %s --conf %s --before yes &",
					schedule_id, conf_name);
			} else {
				snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/DB2File_PY/ucm_meetme_kick_playback.py --bookid \"%s\" --action \"clear\" --data \"%s\" &",
					schedule_id, conf_name);
			}
			ast_safe_system2(cmd);
			ast_log(LOG_DEBUG, "Clean the schedule conference with cmd:%s.", cmd);
		}
	}

}

static char *handle_cli_confbridge_hangup(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference = NULL;
	struct confbridge_user *user = NULL;
	int hangup_type = CONF_HANGUP_NONE;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge hangup";
		e->usage =
			"Usage: confbridge hangup <conference> <playkick|web|activecall> [originator]\n"
			"       hangup a conference.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 3 && a->argc != 4 && a->argc != 5) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[2], OBJ_KEY);
	if (!conference) {
		clean_schedule_meeting_status(a->argv[2], -1);
		if (a->argc == 5) {
			ast_operation_log(a->argv[4], a->argv[4], "NONE", "NONE", conf_operation_str[CONF_OP_HANGUP_CONF], "hangup from meeting list");
		}
		ast_cli(a->fd, "No conference bridge named '%s' found,Only clean up the schedule meeting status!\n", a->argv[2]);
		return CLI_SUCCESS;
	}

	if (a->argc == 4 || a->argc == 5) {
		/*set hangup type*/
		if (!strcasecmp(a->argv[3], "web")) {
			hangup_type = CONF_HANGUP_WEBHOST;
		} else if (!strcasecmp(a->argv[3], "playkick")) {
			hangup_type = CONF_HANGUP_PALYKICKED;
		} else if (!strcasecmp(a->argv[3], "activecall")) {
			hangup_type = CONF_HANGUP_ACTIVECALL;
		} else if (!strcasecmp(a->argv[3], "meetinglist")) {
			hangup_type = CONF_HANGUP_MEETINGLIST;
		}
		if (a->argc == 5 && hangup_type != CONF_HANGUP_ACTIVECALL) {
			ao2_lock(conference);
			user = find_active_user_by_channel(conference, a->argv[4]);
			if (user) {
				conf_user_write_operator_log(user, CONF_OP_HANGUP_CONF, NULL);
			} else {
				ast_operation_log(a->argv[4], a->argv[4], "NONE", conference->feature_id, conf_operation_str[CONF_OP_HANGUP_CONF], "hangup from meeting list");
			}
			ao2_unlock(conference);
		}
		if (a->argc == 5 && hangup_type == CONF_HANGUP_ACTIVECALL) {
			ast_operation_log(a->argv[4], a->argv[4], "NONE", conference->feature_id, conf_operation_str[CONF_OP_HANGUP_CONF], "hangup from active call");
		}
		hangup_conference(conference, hangup_type);
		/*Request from cgi*/
		if (hangup_type == CONF_HANGUP_WEBHOST || hangup_type == CONF_HANGUP_MEETINGLIST) {
			clean_schedule_meeting_status(conference->name, ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU) ? 1 : 0);
		}
	} else {
		hangup_conference(conference, hangup_type);
	}
	ao2_ref(conference, -1);
	ast_cli(a->fd, "Conference %s is hung up.\n", a->argv[2]);
	return CLI_SUCCESS;
}

static int load_schedule_conference_users(struct confbridge_conference *conference)
{
	int is_schedule = 0;
	int is_random = 0;
	ast_log(LOG_DEBUG, "load_schedule_conference_users entry");
	if (!conference) {
		return -1;
	}
	if (!ast_strlen_zero(conference->name) &&
		!ast_strlen_zero(conference->schedule_id_processing) &&
		!ast_strlen_zero(conference->schedule_utc_start_time)) {
		is_schedule = 1;
	} else if (!ast_strlen_zero(conference->name) &&
			!ast_strlen_zero(conference->schedule_id_processing) &&
			ast_strlen_zero(conference->schedule_utc_start_time) &&
			conference->b_profile.is_random_conf) {
		is_random = 1;
	}

	if (is_schedule || is_random) {
		SQLHSTMT stmt = NULL;
		char row_exten[128] = {0};
		char row_email[256] = {0};
		char row_signed[4] = {0};
		char *sqlcmd = NULL;
		int res = -1;
		int user_signed = 0;
		struct odbc_obj *conf_obj = NULL;

		conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
		if (!conf_obj) {
			ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
			return -1;
		}

		if (is_schedule) {
			ast_asprintf(&sqlcmd, "select member_extension,email,signed from multimedia_conference_user_signed where reservation_id='%s' and conf_number='%s' and utc_start_time='%s'",
						conference->schedule_id_processing, conference->name, conference->schedule_utc_start_time);
		} else {
			ast_asprintf(&sqlcmd, "select member_extension,email,signed from multimedia_conference_user_signed where reservation_id='%s' and conf_number='%s'",
						conference->schedule_id_processing, conference->name);
		}
		ast_log(LOG_DEBUG, "sqlcmd:%s", sqlcmd);
		struct generic_prepare_struct gps = { .sql = sqlcmd };

		stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
		if (!stmt) {
			ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", sqlcmd);
			goto bail;
		}
		while((res = SQLFetch(stmt)) != SQL_NO_DATA && SQL_SUCCEEDED(res)) {
			memset(row_exten, 0, sizeof(row_exten));
			memset(row_email, 0, sizeof(row_email));
			memset(row_signed, 0, sizeof(row_signed));
			res = SQLGetData(stmt, 3, SQL_CHAR, row_signed, sizeof(row_signed), NULL);
			if (!SQL_SUCCEEDED(res)) {
				ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", sqlcmd);
				goto bail_with_handle;
			} else {
				user_signed = atoi(row_signed);
			}
			/*member_extension in database can be NULL!*/
			SQLGetData(stmt, 1, SQL_CHAR, row_exten, sizeof(row_exten), NULL);
			SQLGetData(stmt, 2, SQL_CHAR, row_email, sizeof(row_email), NULL);
			ast_log(LOG_DEBUG, "row_email:%s row_exten:%s, user_signed:%d", row_email, row_exten, user_signed);
			if (row_email[0] == '\0' && row_exten[0] == '\0')
			{
				continue;
			}
			/*load conference schedule users to list*/
			conf_cei_sche_user_load(conference->name, conference->schedule_id_processing, row_exten, row_email, user_signed);
		}

	bail_with_handle:
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	bail:
		ast_core_odbc_release_obj(conf_obj);
		conf_obj = NULL;
		safe_ast_free(sqlcmd);
	} else {
		ast_log(LOG_ERROR, "load_schedule_conference_users no a schedule or random!");
	}
	return 0;
}

int conf_record_schedule_users_signed(const char *conference, const char *schedule_id, const char *extension, const char *related_info, const char *schedule_utc_time)
{
	int res = -1;
	ast_log(LOG_DEBUG, "record_schedule_users_signed entry");

	if (!ast_strlen_zero(conference) && !ast_strlen_zero(schedule_id) && !ast_strlen_zero(extension))
	{
		SQLHSTMT stmt = NULL;
		char *sqlcmd = NULL;
		struct odbc_obj *conf_obj = NULL;

		conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
		if (!conf_obj) {
			ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
			return -1;
		}

		if (!strstr(extension, "conf")) {
			if (!ast_strlen_zero(schedule_utc_time)) {
				ast_asprintf(&sqlcmd, "UPDATE multimedia_conference_user_signed SET signed = 1,real_exten='%s' WHERE conf_number='%s' AND reservation_id='%s' AND member_extension='%s' AND utc_start_time='%s'", extension, conference, schedule_id, extension, schedule_utc_time);
			} else {
				ast_asprintf(&sqlcmd, "UPDATE multimedia_conference_user_signed SET signed = 1,real_exten='%s' WHERE conf_number='%s' AND reservation_id='%s' AND member_extension='%s'", extension, conference, schedule_id, extension);
			}
		} else if (!ast_strlen_zero(related_info)) {
			if (!ast_strlen_zero(schedule_utc_time)) {
				ast_asprintf(&sqlcmd, "UPDATE multimedia_conference_user_signed SET signed = 1,real_exten='%s' WHERE conf_number='%s' AND reservation_id='%s' AND ((member_extension='%s') or (IF(member_extension IS NULL,0,LENGTH(member_extension)) = 0 and email='%s')) AND utc_start_time='%s'", extension, conference, schedule_id, related_info, related_info, schedule_utc_time);
			} else {
				ast_asprintf(&sqlcmd, "UPDATE multimedia_conference_user_signed SET signed = 1,real_exten='%s' WHERE conf_number='%s' AND reservation_id='%s' AND ((member_extension='%s') or (IF(member_extension IS NULL,0,LENGTH(member_extension)) = 0 and email='%s'))", extension, conference, schedule_id, related_info, related_info);
			}
		} else {
			ast_log(LOG_ERROR, "no related info!");
			ast_core_odbc_release_obj(conf_obj);
			conf_obj = NULL;
			return res;
		}

		struct generic_prepare_struct gps = { .sql = sqlcmd };

		stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
		if (!stmt) {
			ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", sqlcmd);
			safe_ast_free(sqlcmd);
			ast_core_odbc_release_obj(conf_obj);
			conf_obj = NULL;
			return res;
		} else {
			res = 0;
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		}

		if (ast_ha_permit_send()){
			ast_ha_send_conf_report_sql(sqlcmd);
		}

		ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", sqlcmd);

		safe_ast_free(sqlcmd);
		ast_core_odbc_release_obj(conf_obj);
		conf_obj = NULL;
	} else {
		ast_log(LOG_ERROR, "conf_record_schedule_users_signed no a schedule!");
	}

	return res;
}

static int get_schedule_conference(struct confbridge_conference *conference)
{
	struct ast_tm tm;
	char family[256] = {0};
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
	char step[8] = {0};
	char start_time[32] = {0};
	char alert_time[32] = {0};
	char cofn_subject[AST_MAX_CONF_SUBJECT] = {0};
	char cofn_info[1024] = {0};
	char im_service_prefix[120] = {0};
	struct ast_json  *utc_time, *host_email, *host_exten, *start_time_val;
	RAII_VAR(struct ast_json *, json_s, NULL, ast_json_unref);
	struct ast_json_error err;

	/*get im_service_prefix*/
	if (ast_safe_nvram_get("im_service_prefix", im_service_prefix, sizeof(im_service_prefix)) == NULL) {
		memset(im_service_prefix, 0, sizeof(im_service_prefix));
	}
	ast_log(LOG_DEBUG, "im_service_prefix %s!", im_service_prefix);
	if (!ast_strlen_zero(im_service_prefix)) {
		if (conference->im_service_prefix) {
			ast_free(conference->im_service_prefix);
			conference->im_service_prefix = NULL;
		}
		conference->im_service_prefix = ast_strdup(im_service_prefix);
		ast_log(LOG_DEBUG, "conference->im_service_prefix %s!", conference->im_service_prefix);
	}
	/*Begin Added by gangwang@grandstream.cn for Wave App*/
	if (conference->schedule_subject) {
		ast_free(conference->schedule_subject);
		conference->schedule_subject = NULL;
	}
	snprintf(family, sizeof(family), "SCHEDULE/%s", conference->name);
	if (!ast_db_get(family, "ScheduleID", schedule_id, sizeof(schedule_id)) && !ast_strlen_zero(schedule_id)) {
		/*Conf schedule id*/
		ast_log(LOG_DEBUG, "ScheduleID %s!", schedule_id);
		if (conference->schedule_id_processing) {
			ast_free(conference->schedule_id_processing);
			conference->schedule_id_processing = NULL;
		}
		conference->schedule_id_processing = ast_strdup(schedule_id);

		/*conf info*/
		snprintf(family, sizeof(family), "SCHEDULE/%s", schedule_id);
		if (ast_db_get(family, "Info", cofn_info, sizeof(cofn_info)) || ast_strlen_zero(cofn_info)) {
			ast_log(LOG_DEBUG, "Did not get the Info!");
		}
		/*Conf Subject*/
		snprintf(family, sizeof(family), "SCHEDULE/%s", schedule_id);
		if (ast_db_get(family, "ConfSubject", cofn_subject, sizeof(cofn_subject)) || ast_strlen_zero(cofn_subject)) {
			ast_log(LOG_DEBUG, "Did not get the ConfSubject!");
		}
		memset(family, 0, sizeof(family));
		memset(schedule_id, 0, sizeof(schedule_id));
	}
	if (!ast_strlen_zero(cofn_subject)) {
		conference->schedule_subject = ast_strdup(cofn_subject);
	} else {
		RAII_VAR(struct im_conference *, conference_im, NULL, ao2_cleanup);
		conference_im = im_conf_list_fetch_add(conference->name, NULL, 0, CONF_TYPE_NONE, 0, NULL);
		if (conference_im) {
			ao2_lock(conference_im);
			conference->schedule_subject = ast_strdup(conference_im->conf_subject);
			ao2_unlock(conference_im);
		}
	}
	if (!ast_strlen_zero(cofn_info)) {
		json_s = ast_json_load_string(cofn_info, &err);
		if(!json_s) {
			ast_log(LOG_WARNING, "load oject error.");
		} else {
			if ((host_exten = ast_json_object_get(json_s, "host_exten"))) {
				if (conference->schedule_host_exten) {
					ast_free(conference->schedule_host_exten);
					conference->schedule_host_exten = NULL;
				}
				conference->schedule_host_exten = ast_strdup(ast_json_string_get(host_exten) ? ast_json_string_get(host_exten) : "");
			}
			if ((host_email = ast_json_object_get(json_s, "host_email"))) {
				if (conference->schedule_host_email) {
					ast_free(conference->schedule_host_email);
					conference->schedule_host_email = NULL;
				}
				conference->schedule_host_email = ast_strdup(ast_json_string_get(host_email) ? ast_json_string_get(host_email) : "");
			}
			if ((utc_time = ast_json_object_get(json_s, "utc_start_time"))) {
				if (conference->schedule_utc_start_time) {
					ast_free(conference->schedule_utc_start_time);
					conference->schedule_utc_start_time = NULL;
				}
				conference->schedule_utc_start_time = ast_strdup(ast_json_string_get(utc_time) ? ast_json_string_get(utc_time) : "");
			}
			if ((start_time_val = ast_json_object_get(json_s, "start_time"))) {
				/*Set start timeval*/
				if (ast_strlen_zero(ast_json_string_get(start_time_val))) {
					conference->schedule_start_timeval = ast_tvnow();
				} else {
					struct tm start_tm = {0};
					struct timeval start_tv = {0};
					strptime(ast_json_string_get(start_time_val), "%Y-%m-%d %H:%M", &start_tm);
					start_tv.tv_sec = mktime(&start_tm);
					conference->schedule_start_timeval = start_tv;
					ast_log(LOG_DEBUG, "get_schedule_conference time diff %ld\n", ast_tvdiff_sec(ast_tvnow(), start_tv));
				}
			}
		}
	}
	load_schedule_conference_users(conference);
	/*End Added by gangwang@grandstream.cn for Wave App*/
	/*schedule id*/
	snprintf(family, sizeof(family), "SCHEDULE/%s", conference->name);
	if (ast_db_get(family, "AlertScheduleID", schedule_id, sizeof(schedule_id)) || ast_strlen_zero(schedule_id)) {
		return -1;
	}
	/*step*/
	snprintf(family, sizeof(family), "SCHEDULE/%s", schedule_id);
	if (ast_db_get(family, "STEP", step, sizeof(step)) || strcmp(step, "alert")) {
		return -1;
	}
	/*start time*/
	if (ast_db_get(family, "StartTime", start_time, sizeof(start_time)) || ast_strlen_zero(start_time)) {
		return -1;
	}
	/*alert time*/
	if (ast_db_get(family, "AlertTime", alert_time, sizeof(alert_time)) || ast_strlen_zero(alert_time)) {
		return -1;
	}

	if (conference->schedule_id_to_start) {
		ast_free(conference->schedule_id_to_start);
		conference->schedule_id_to_start = NULL;
	}
	conference->schedule_id_to_start = ast_strdup(schedule_id);

	if (conference->schedule_start_time) {
		ast_free(conference->schedule_start_time);
		conference->schedule_start_time = NULL;
	}
	conference->schedule_start_time = ast_strdup(start_time);

	ast_strptime(alert_time, "%Y-%m-%d %H:%M:%S%z", &tm);
	conference->schedule_alert_time = ast_mktime(&tm, NULL);

	return 0;
}

static unsigned int get_schedule_notify_countdown(struct confbridge_conference *conference)
{
	struct tm start_tm = {0};
	struct timeval start_tv = {0};
	struct timeval now_tv = {0};
	int64_t countdown = 0;

	strptime(conference->schedule_start_time, "%Y-%m-%dT%H:%M:%S%z", &start_tm);
	start_tv.tv_sec = mktime(&start_tm);
	now_tv = ast_tvnow();
	countdown = start_tv.tv_sec - now_tv.tv_sec + conference->schedule_alert_time.tv_sec % 60;
	if (countdown < 0) {
		countdown = 0;
	}

	return (unsigned int)countdown;
}

static char *build_schedule_notify_alert_body(const char *conf_number, const char *schedule_id, const char *start_time, unsigned int countdown)
{
	char body[256] = {0};
	char day_time[32] = {0};
	char *token = NULL;

	if (!countdown) {
		return NULL;
	}

	strncpy(day_time, start_time, sizeof(day_time)-1);
	if ((token = strchr(day_time, 'T'))) {
		*token = ' ';
	}
	snprintf(body, sizeof(body), "Event: x-gs-schedule-conference\r\n"
		"Schedule-Conference: %s\r\n"
		"Schedule-Id: %s\r\n"
		"Action: alert\r\n"
		"Start-Time: %s\r\n"
		"Countdown: %u\r\n", conf_number, schedule_id, day_time, countdown);

	return ast_strdup(body);
}

static char *build_schedule_notify_cancel_body(const char *conf_number, const char *schedule_id)
{
	char body[256] = {0};

	snprintf(body, sizeof(body), "Event: x-gs-schedule-conference\r\n"
		"Schedule-Conference: %s\r\n"
		"Schedule-Id: %s\r\n"
		"Action: cancel\r\n", conf_number, schedule_id);

	return ast_strdup(body);
}

static void schedule_notify(struct confbridge_conference *conference, const char *body)
{
	struct confbridge_user *user = NULL;

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		ast_sendtext(user->chan, body);
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		ast_sendtext(user->chan, body);
	}
	ao2_unlock(conference);
}

static void check_timeout_before_schedule_notify(struct confbridge_conference *conference)
{
	struct confbridge_user *user = NULL;
	ast_log(LOG_DEBUG, "check timeout before schedule notify!");

	if (!conference) {
		return;
	}

	ao2_lock(conference);
	if (conference->extend_time_status == EXTEND_PENDING) {
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
				if (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
					ast_log(LOG_DEBUG, "Before schedule notify, WAVE and APP only send conf type Notify!");
					conf_timeout_notify_build_send(user->chan, conference->name, 1, EXTEND_CONFIRMED, 0);
				}
			}
		}
	} else if (conference->only_warning.extend_time_status == EXTEND_PENDING && conference->only_warning.user) {
		if (conference->only_warning.user->term_type == AST_TERMINAL_TYPE_WEBRTC || conference->only_warning.user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			ast_log(LOG_DEBUG, "Before schedule notify, WAVE and APP only send user type Notify!");
			conf_timeout_notify_build_send(conference->only_warning.user->chan, conference->name, 0, EXTEND_CONFIRMED, 0);
		}
	}
	ao2_unlock(conference);

}

static char *handle_cli_confbridge_schedule_notify(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference = NULL;
	RAII_VAR(char *, body, NULL, ast_free);

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge schedule notify";
		e->usage =
			"Usage: confbridge schedule notify <conference> <schedule id> <action> <starttime>\n"
			"       Send a schedule conference notify for all users in conference.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 6 && a->argc != 7) {
		return CLI_SHOWUSAGE;
	}

	if (strcasecmp(a->argv[5], "alert") && strcasecmp(a->argv[5], "cancel")) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[3], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "No conference bridge named '%s' found!\n", a->argv[3]);
		return CLI_SUCCESS;
	}

	ao2_lock(conference);
	if (!strcasecmp(a->argv[5], "alert")) {
		if (conference->schedule_id_to_start) {
			ast_free(conference->schedule_id_to_start);
			conference->schedule_id_to_start = NULL;
		}

		if (conference->schedule_start_time) {
			ast_free(conference->schedule_start_time);
			conference->schedule_start_time = NULL;
		}
		conference->schedule_id_to_start = ast_strdup(a->argv[4]);
		conference->schedule_start_time = ast_strdup(a->argv[6]);
		conference->schedule_alert_time = ast_tvnow();
		if (!ast_strlen_zero(conference->schedule_id_processing) && !strcmp(conference->schedule_id_processing, conference->schedule_id_to_start)) {
			ast_log(LOG_DEBUG, "schedule_id_to_start %s is already in processing , no need to notify alert!", conference->schedule_id_processing);
		} else {
			body = build_schedule_notify_alert_body(a->argv[3], a->argv[4], a->argv[6], get_schedule_notify_countdown(conference));
			if (body) {
				check_timeout_before_schedule_notify(conference);
				schedule_notify(conference, body);
			}
		}
	} else if (!strcasecmp(a->argv[5], "cancel")) {
		if (conference->schedule_id_to_start && !strcmp(a->argv[4], conference->schedule_id_to_start)) {
			body = build_schedule_notify_cancel_body(a->argv[3], a->argv[4]);
			schedule_notify(conference, body);
			ast_free(conference->schedule_id_to_start);
			conference->schedule_id_to_start = NULL;
			ast_free(conference->schedule_start_time);
			conference->schedule_start_time = NULL;
			conference->schedule_alert_time = ast_tv(0,0);
		}
	}
	ao2_unlock(conference);

	ao2_ref(conference, -1);
	ast_cli(a->fd, "The notify of the schedule conference %s has been sent.\n", a->argv[3]);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_kick_extension(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference = NULL;
	struct confbridge_user *user = NULL;
	struct ao2_iterator iter;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge kickextension";
		e->usage =
			"Usage: confbridge kickextension <extension> <type>\n"
			"       Kicks a user specified by extension out of the conference bridge.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			ao2_ref(conference, -1);
			continue;
		}

		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_channel_tech(user->chan)->get_terminal_type &&
				ast_channel_tech(user->chan)->get_terminal_type(user->chan) == AST_TERMINAL_TYPE_WEBRTC) {
				char pjsip_extension[AST_MAX_EXTENSION] = {0};
				get_pjsip_extension(ast_channel_name(user->chan), pjsip_extension, sizeof(pjsip_extension));
				if (!strcmp(pjsip_extension, a->argv[2])) {
					publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
					user->kicked = 1;
					pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
					ast_bridge_remove(conference->bridge, user->chan);
					ast_cli(a->fd, "Kicked '%s' out of conference '%s'\n", a->argv[2], conference->name);
				}
			}
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	return CLI_SUCCESS;
}
#endif

static char *handle_cli_confbridge_lock(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge lock";
#ifdef GRANDSTREAM_NETWORKS
		e->usage =
			"Usage: confbridge lock <conference> [originator]\n"
			"       Lock a conference. While locked, no new non-admins\n"
			"       may join the conference.\n";
#else
		e->usage =
			"Usage: confbridge lock <conference>\n"
			"		Lock a conference. While locked, no new non-admins\n"
			"		may join the conference.\n";
#endif
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (a->argc != 3 && a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
#else
	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}
#endif
#ifdef GRANDSTREAM_NETWORKS
	if (generic_lock_unlock_helper(1, a->argv[2], a->argc == 4 ? a->argv[3] : NULL))
#else
	if (generic_lock_unlock_helper(1, a->argv[2]))
#endif
	{
		ast_cli(a->fd, "Conference %s is not found\n", a->argv[2]);
	} else {
		ast_cli(a->fd, "Conference %s is locked.\n", a->argv[2]);
	}
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_unlock(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge unlock";
#ifdef GRANDSTREAM_NETWORKS
		e->usage =
			"Usage: confbridge unlock <conference> [originator]\n"
			"       Unlock a previously locked conference.\n";
#else
		e->usage =
			"Usage: confbridge unlock <conference>\n"
			"       Unlock a previously locked conference.\n";
#endif
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 2) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
#ifdef GRANDSTREAM_NETWORKS
	if (a->argc != 3 && a->argc != 4) {
		return CLI_SHOWUSAGE;
	}
#else
	if (a->argc != 3) {
		return CLI_SHOWUSAGE;
	}
#endif
#ifdef GRANDSTREAM_NETWORKS
	if (generic_lock_unlock_helper(0, a->argv[2], a->argc == 4 ? a->argv[3] : NULL))
#else
	if (generic_lock_unlock_helper(0, a->argv[2]))
#endif
	{
		ast_cli(a->fd, "Conference %s is not found\n", a->argv[2]);
	} else {
		ast_cli(a->fd, "Conference %s is unlocked.\n", a->argv[2]);
	}
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_start_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	const char *rec_file = NULL;
	struct confbridge_conference *conference;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge record start";
		e->usage =
			"Usage: confbridge record start <conference> <file>\n"
			"       <file> is optional, Otherwise the bridge profile\n"
			"       record file will be used.  If the bridge profile\n"
			"       has no record file specified, a file will automatically\n"
			"       be generated in the monitor directory\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}
	if (a->argc == 5) {
		rec_file = a->argv[4];
	}

	conference = ao2_find(conference_bridges, a->argv[3], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_FAILURE;
	}
	ao2_lock(conference);
	if (conf_is_recording(conference)) {
		ast_cli(a->fd, "Conference is already being recorded.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_SUCCESS;
	}
	if (!ast_strlen_zero(rec_file)) {
		ast_copy_string(conference->b_profile.rec_file, rec_file, sizeof(conference->b_profile.rec_file));
	}

#ifdef GRANDSTREAM_NETWORKS
	if (conf_start_record(conference, NULL)) {
		ast_cli(a->fd, "Could not start recording due to internal error.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_FAILURE;
	}
#else
	if (conf_start_record(conference)) {
		ast_cli(a->fd, "Could not start recording due to internal error.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_FAILURE;
	}
#endif
	ao2_unlock(conference);

	ast_cli(a->fd, "Recording started\n");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_stop_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;
	int ret;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge record stop";
		e->usage =
			"Usage: confbridge record stop <conference>\n"
			"       Stop a previously started recording.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[3], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_SUCCESS;
	}
	ao2_lock(conference);
	ret = conf_stop_record(conference);
	ao2_unlock(conference);
	ast_cli(a->fd, "Recording %sstopped.\n", ret ? "could not be " : "");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

#ifdef GRANDSTREAM_NETWORKS
static char *handle_cli_confbridge_start_video_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	const char *rec_file = NULL;
	struct confbridge_conference *conference;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge video record start";
		e->usage =
			"Usage: confbridge video record start <conference> <file>\n"
			"       <file> is optional, Otherwise the bridge profile\n"
			"       record file will be used.  If the bridge profile\n"
			"       has no record file specified, a file will automatically\n"
			"       be generated in the monitor directory\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc < 5) {
		return CLI_SHOWUSAGE;
	}
	if (a->argc == 6) {
		rec_file = a->argv[5];
	}

	conference = ao2_find(conference_bridges, a->argv[4], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_FAILURE;
	}
	ao2_lock(conference);
	if (conference->video_is_recording) {
		ast_cli(a->fd, "Conference is already being recorded.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_SUCCESS;
	}

	if (send_avs_video_record_control(conference, rec_file, AVS_CONTROL_START_CONF_VIDEO_RECORD)) {
		ast_cli(a->fd, "Could not start recording due to internal error.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_FAILURE;
	}
	ao2_unlock(conference);

	ast_cli(a->fd, "Recording started\n");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_stop_video_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;
	int ret;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge video record stop";
		e->usage =
			"Usage: confbridge video record stop <conference>\n"
			"       Stop a previously started video recording.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 5) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[4], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_SUCCESS;
	}
	ao2_lock(conference);
	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_VIDEO_RECORD);
	ao2_unlock(conference);
	ast_cli(a->fd, "Video recording %sstopped.\n", ret ? "could not be " : "");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_start_slides_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	const char *rec_file = NULL;
	struct confbridge_conference *conference;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge slides record start";
		e->usage =
			"Usage: confbridge slides record start <conference> <file>\n"
			"       <file> is optional, Otherwise the bridge profile\n"
			"       record file will be used.  If the bridge profile\n"
			"       has no record file specified, a file will automatically\n"
			"       be generated in the monitor directory\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc < 5) {
		return CLI_SHOWUSAGE;
	}
	if (a->argc == 6) {
		rec_file = a->argv[5];
	}

	conference = ao2_find(conference_bridges, a->argv[4], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_FAILURE;
	}
	ao2_lock(conference);
	if (conference->slides_is_recording) {
		ast_cli(a->fd, "Conference is already being recorded.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_SUCCESS;
	}

	if (send_avs_video_record_control(conference, rec_file, AVS_CONTROL_START_CONF_SLIDES_RECORD)) {
		ast_cli(a->fd, "Could not start recording due to internal error.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_FAILURE;
	}
	ao2_unlock(conference);

	ast_cli(a->fd, "Slides recording started\n");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_stop_slides_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;
	int ret;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge slides record stop";
		e->usage =
			"Usage: confbridge slides record stop <conference>\n"
			"       Stop a previously started slides recording.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 5) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[4], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_SUCCESS;
	}
	ao2_lock(conference);
	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_SLIDES_RECORD);
	ao2_unlock(conference);
	ast_cli(a->fd, "Slides recording %sstopped.\n", ret ? "could not be " : "");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_start_focus_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	const char *rec_file = NULL;
	struct confbridge_conference *conference;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge focus record start";
		e->usage =
			"Usage: confbridge focus record start <conference> <file>\n"
			"       <file> is optional, Otherwise the bridge profile\n"
			"       record file will be used.  If the bridge profile\n"
			"       has no record file specified, a file will automatically\n"
			"       be generated in the monitor directory\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc < 5) {
		return CLI_SHOWUSAGE;
	}
	if (a->argc == 6) {
		rec_file = a->argv[5];
	}

	conference = ao2_find(conference_bridges, a->argv[4], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_FAILURE;
	}
	ao2_lock(conference);
	if (conference->focus_is_recording) {
		ast_cli(a->fd, "Conference is already being recorded.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_SUCCESS;
	}

	if (send_avs_video_record_control(conference, rec_file, AVS_CONTROL_START_CONF_FOCUS_RECORD)) {
		ast_cli(a->fd, "Could not start recording due to internal error.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_FAILURE;
	}
	ao2_unlock(conference);

	ast_cli(a->fd, "Focus recording started\n");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_stop_focus_record(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;
	int ret;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge focus record stop";
		e->usage =
			"Usage: confbridge focus record stop <conference>\n"
			"       Stop a previously started foucs recording.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 4) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 5) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[4], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_SUCCESS;
	}
	ao2_lock(conference);
	ret = send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_FOCUS_RECORD);
	ao2_unlock(conference);
	ast_cli(a->fd, "Focus recording %sstopped.\n", ret ? "could not be " : "");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_enable_mcu(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge mcu enable";
		e->usage =
			"Usage: confbridge mcu enable <conference>\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc < 4) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[3], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_FAILURE;
	}
	ao2_lock(conference);
	if (conference->mcu_enabled) {
		ast_cli(a->fd, "Conference is already enabled MCU.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_SUCCESS;
	}

	if (conf_send_avs_mcu_control(conference, AVS_CONTROL_ENABLE_MCU)) {
		ast_cli(a->fd, "Could not enable MCU due to internal error.\n");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return CLI_FAILURE;
	}
	ao2_unlock(conference);

	ast_cli(a->fd, "Conference MCU enabled\n");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}

static char *handle_cli_confbridge_disable_mcu(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct confbridge_conference *conference;
	int ret;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge mcu disable";
		e->usage =
			"Usage: confbridge mcu disable <conference>\n"
			"       Stop a previously enable mcu.\n";
		return NULL;
	case CLI_GENERATE:
		if (a->pos == 3) {
			return complete_confbridge_name(a->line, a->word, a->pos, a->n);
		}
		return NULL;
	}
	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}

	conference = ao2_find(conference_bridges, a->argv[3], OBJ_KEY);
	if (!conference) {
		ast_cli(a->fd, "Conference not found.\n");
		return CLI_SUCCESS;
	}
	ao2_lock(conference);
	ret = conf_send_avs_mcu_control(conference, AVS_CONTROL_DISABLE_MCU);
	ao2_unlock(conference);
	ast_cli(a->fd, "Conference MCU %sdisabled.\n", ret ? "could not be " : "");
	ao2_ref(conference, -1);
	return CLI_SUCCESS;
}
#endif

static struct ast_cli_entry cli_confbridge[] = {
	AST_CLI_DEFINE(handle_cli_confbridge_list, "List conference bridges and participants."),
	AST_CLI_DEFINE(handle_cli_confbridge_kick, "Kick participants out of conference bridges."),
	AST_CLI_DEFINE(handle_cli_confbridge_mute, "Mute participants."),
	AST_CLI_DEFINE(handle_cli_confbridge_unmute, "Unmute participants."),
	AST_CLI_DEFINE(handle_cli_confbridge_lock, "Lock a conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_unlock, "Unlock a conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_start_record, "Start recording a conference"),
	AST_CLI_DEFINE(handle_cli_confbridge_stop_record, "Stop recording a conference."),
#ifdef GRANDSTREAM_NETWORKS
	AST_CLI_DEFINE(handle_cli_confbridge_playback, "Playback an audio file to a specified conference room."),
	AST_CLI_DEFINE(handle_cli_confbridge_cache_list, "List cache conference and participants."),
	AST_CLI_DEFINE(handle_cli_confbridge_resource_list, "List the resources in SFU."),
	AST_CLI_DEFINE(handle_cli_confbridge_hangup, "Hangup a conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_schedule_notify, "Notify a schedule conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_kick_extension, "Kick the participant specified by extension out of conference bridges."),
	AST_CLI_DEFINE(handle_cli_schedule_conference_list, "List scheduled conference information and member information."),
	AST_CLI_DEFINE(handle_cli_im_conference_info_list, "List im conference information."),
	AST_CLI_DEFINE(handle_cli_schedule_conference_early_start_list, "List scheduled conference wait information and member information."),
	AST_CLI_DEFINE(handle_cli_schedule_conference_early_start_modify, "Modify scheduled conference wait information when no users waiting."),
	AST_CLI_DEFINE(handle_cli_schedule_conference_need_wait, "Determine whether you need to wait for schedule conference start."),
	AST_CLI_DEFINE(handle_cli_conference_current_users, "List the number of current users in the conference."),
	AST_CLI_DEFINE(handle_cli_conference_current_host, "Judge whether the extension is host in the conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_start_video_record, "Start video recording a conference"),
	AST_CLI_DEFINE(handle_cli_confbridge_stop_video_record, "Stop video recording a conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_enable_mcu, "Enable MCU a conference"),
	AST_CLI_DEFINE(handle_cli_confbridge_disable_mcu, "Disable MCU a conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_start_slides_record, "Start slides recording a conference"),
	AST_CLI_DEFINE(handle_cli_confbridge_stop_slides_record, "Stop slides recording a conference."),
	AST_CLI_DEFINE(handle_cli_confbridge_start_focus_record, "Start focus recording a conference"),
	AST_CLI_DEFINE(handle_cli_confbridge_stop_focus_record, "Stop focus recording a conference."),
#endif
};
static struct ast_custom_function confbridge_function = {
	.name = "CONFBRIDGE",
	.write = func_confbridge_helper,
};

static int func_confbridge_info(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len);
static struct ast_custom_function confbridge_info_function = {
	.name = "CONFBRIDGE_INFO",
	.read = func_confbridge_info,
};

static int action_confbridgelist_item(struct mansession *s, const char *id_text, struct confbridge_conference *conference, struct confbridge_user *user, int waiting)
{
	struct ast_channel_snapshot *snapshot;
	struct ast_str *snap_str;

	snapshot = ast_channel_snapshot_get_latest(ast_channel_uniqueid(user->chan));
	if (!snapshot) {
		return 0;
	}

	snap_str = ast_manager_build_channel_state_string(snapshot);
	if (!snap_str) {
		ao2_ref(snapshot, -1);
		return 0;
	}

	astman_append(s,
		"Event: ConfbridgeList\r\n"
		"%s"
		"Conference: %s\r\n"
		"Admin: %s\r\n"
		"MarkedUser: %s\r\n"
		"WaitMarked: %s\r\n"
		"EndMarked: %s\r\n"
		"Waiting: %s\r\n"
		"Muted: %s\r\n"
		"Talking: %s\r\n"
		"AnsweredTime: %d\r\n"
		"%s"
		"\r\n",
		id_text,
		conference->name,
		AST_YESNO(ast_test_flag(&user->u_profile, USER_OPT_ADMIN)),
		AST_YESNO(ast_test_flag(&user->u_profile, USER_OPT_MARKEDUSER)),
		AST_YESNO(ast_test_flag(&user->u_profile, USER_OPT_WAITMARKED)),
		AST_YESNO(ast_test_flag(&user->u_profile, USER_OPT_ENDMARKED)),
		AST_YESNO(waiting),
		AST_YESNO(user->muted),
		AST_YESNO(user->talking),
		ast_channel_get_up_time(user->chan),
		ast_str_buffer(snap_str));

	ast_free(snap_str);
	ao2_ref(snapshot, -1);

	return 1;
}

static int action_confbridgelist(struct mansession *s, const struct message *m)
{
	const char *actionid = astman_get_header(m, "ActionID");
	const char *conference_name = astman_get_header(m, "Conference");
	struct confbridge_user *user;
	struct confbridge_conference *conference;
	char id_text[80];
	int total = 0;

	id_text[0] = '\0';
	if (!ast_strlen_zero(actionid)) {
		snprintf(id_text, sizeof(id_text), "ActionID: %s\r\n", actionid);
	}
	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}
	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	astman_send_listack(s, m, "Confbridge user list will follow", "start");

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		total += action_confbridgelist_item(s, id_text, conference, user, 0);
	}
	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		total += action_confbridgelist_item(s, id_text, conference, user, 1);
	}
	ao2_unlock(conference);
	ao2_ref(conference, -1);

	astman_send_list_complete_start(s, m, "ConfbridgeListComplete", total);
	astman_send_list_complete_end(s);

	return 0;
}

static int action_confbridgelistrooms(struct mansession *s, const struct message *m)
{
	const char *actionid = astman_get_header(m, "ActionID");
	struct confbridge_conference *conference;
	struct ao2_iterator iter;
	char id_text[512] = "";
	int totalitems = 0;

	if (!ast_strlen_zero(actionid)) {
		snprintf(id_text, sizeof(id_text), "ActionID: %s\r\n", actionid);
	}

	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	astman_send_listack(s, m, "Confbridge conferences will follow", "start");

	/* Traverse the conference list */
	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		totalitems++;

		ao2_lock(conference);
		astman_append(s,
		"Event: ConfbridgeListRooms\r\n"
		"%s"
		"Conference: %s\r\n"
		"Parties: %u\r\n"
		"Marked: %u\r\n"
		"Locked: %s\r\n"
		"Muted: %s\r\n"
		"\r\n",
		id_text,
		conference->name,
		conference->activeusers + conference->waitingusers,
		conference->markedusers,
		AST_YESNO(conference->locked),
		AST_YESNO(conference->muted));
		ao2_unlock(conference);

		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	/* Send final confirmation */
	astman_send_list_complete_start(s, m, "ConfbridgeListRoomsComplete", totalitems);
	astman_send_list_complete_end(s);
	return 0;
}

static int action_mute_unmute_helper(struct mansession *s, const struct message *m, int mute)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	int res = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No channel name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	res = generic_mute_unmute_helper(mute, conference_name, channel_name, NULL, 0);
#else
	res = generic_mute_unmute_helper(mute, conference_name, channel_name);
#endif

	if (res == -1) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	} else if (res == -2) {
		astman_send_error(s, m, "No Channel by that name found in Conference.");
		return 0;
	}

	astman_send_ack(s, m, mute ? "User muted" : "User unmuted");
	return 0;
}

static int action_confbridgeunmute(struct mansession *s, const struct message *m)
{
	return action_mute_unmute_helper(s, m, 0);
}
static int action_confbridgemute(struct mansession *s, const struct message *m)
{
	return action_mute_unmute_helper(s, m, 1);
}

static int action_lock_unlock_helper(struct mansession *s, const struct message *m, int lock)
{
	const char *conference_name = astman_get_header(m, "Conference");
	int res = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}
#ifdef GRANDSTREAM_NETWORKS
	if ((res = generic_lock_unlock_helper(lock, conference_name, NULL)))
#else
	if ((res = generic_lock_unlock_helper(lock, conference_name)))
#endif
	{
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}
	astman_send_ack(s, m, lock ? "Conference locked" : "Conference unlocked");
	return 0;
}
static int action_confbridgeunlock(struct mansession *s, const struct message *m)
{
	return action_lock_unlock_helper(s, m, 0);
}
static int action_confbridgelock(struct mansession *s, const struct message *m)
{
	return action_lock_unlock_helper(s, m, 1);
}

static int action_confbridgekick(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel = astman_get_header(m, "Channel");
	struct confbridge_conference *conference;
	int found;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	found = !kick_conference_participant(conference, channel, NULL);
#else
	found = !kick_conference_participant(conference, channel);
#endif
	ao2_ref(conference, -1);

	if (found) {
		astman_send_ack(s, m, !strcmp("all", channel) ? "All participants kicked" : "User kicked");
	} else {
		astman_send_error(s, m, "No Channel by that name found in Conference.");
	}
	return 0;
}

static int action_confbridgestartrecord(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *recordfile = astman_get_header(m, "RecordFile");
	struct confbridge_conference *conference;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);
	if (conf_is_recording(conference)) {
		astman_send_error(s, m, "Conference is already being recorded.");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return 0;
	}

	if (!ast_strlen_zero(recordfile)) {
		ast_copy_string(conference->b_profile.rec_file, recordfile, sizeof(conference->b_profile.rec_file));
	}

#ifdef GRANDSTREAM_NETWORKS
	if (conf_start_record(conference, NULL)) {
		astman_send_error(s, m, "Internal error starting conference recording.");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return 0;
	}
#else
	if (conf_start_record(conference)) {
		astman_send_error(s, m, "Internal error starting conference recording.");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return 0;
	}
#endif
	ao2_unlock(conference);

	ao2_ref(conference, -1);
	astman_send_ack(s, m, "Conference Recording Started.");
	return 0;
}
static int action_confbridgestoprecord(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	struct confbridge_conference *conference;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);
	if (conf_stop_record(conference)) {
		ao2_unlock(conference);
		astman_send_error(s, m, "Internal error while stopping recording.");
		ao2_ref(conference, -1);
		return 0;
	}
	ao2_unlock(conference);

	ao2_ref(conference, -1);
	astman_send_ack(s, m, "Conference Recording Stopped.");
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int action_confbridgestartvideorecord(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *recordfile = astman_get_header(m, "RecordFile");
	struct confbridge_conference *conference = NULL;
	int ret = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);

	if (conference->video_is_recording) {
		astman_send_error(s, m, "Conference is already being video recorded.");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return 0;
	}

	ret = send_avs_video_record_control(conference, recordfile, AVS_CONTROL_START_CONF_VIDEO_RECORD);

	ao2_unlock(conference);

	ao2_ref(conference, -1);

	if (ret == R_SUCCESS)	{
		astman_send_ack(s, m, "Conference Video Recording Started.");
	} else {
		astman_send_error(s, m, "send avs control error");
	}

	return 0;
}

static int action_confbridgestopvideorecord(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *recordfile = astman_get_header(m, "RecordFile");
	struct confbridge_conference *conference = NULL;
	int ret = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);

	ret = send_avs_video_record_control(conference, recordfile, AVS_CONTROL_STOP_CONF_VIDEO_RECORD);

	ao2_unlock(conference);

	ao2_ref(conference, -1);

	if (ret == R_SUCCESS)	{
		astman_send_ack(s, m, "Conference Video Recording Stopped.");
	} else {
		astman_send_error(s, m, "send avs control error");
	}

	return 0;
}

static int action_confbridgestartslidesrecord(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *recordfile = astman_get_header(m, "RecordFile");
	struct confbridge_conference *conference = NULL;
	int ret = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);

	if (conference->slides_is_recording) {
		astman_send_error(s, m, "Conference is already being slides recorded.");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return 0;
	}

	ret = send_avs_video_record_control(conference, recordfile, AVS_CONTROL_START_CONF_SLIDES_RECORD);

	ao2_unlock(conference);

	ao2_ref(conference, -1);

	if (ret == R_SUCCESS)	{
		astman_send_ack(s, m, "Conference Slides Recording Started.");
	} else {
		astman_send_error(s, m, "send avs control error");
	}

	return 0;
}

static int action_confbridgestopslidesrecord(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *recordfile = astman_get_header(m, "RecordFile");
	struct confbridge_conference *conference = NULL;
	int ret = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);

	ret = send_avs_video_record_control(conference, recordfile, AVS_CONTROL_STOP_CONF_SLIDES_RECORD);

	ao2_unlock(conference);

	ao2_ref(conference, -1);

	if (ret == R_SUCCESS)	{
		astman_send_ack(s, m, "Conference Slides Recording Stopped.");
	} else {
		astman_send_error(s, m, "send avs control error");
	}

	return 0;
}

static int action_confbridgeenablemcu(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	struct confbridge_conference *conference = NULL;
	int ret = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);

	if (conference->mcu_enabled) {
		astman_send_error(s, m, "Conference is already enabled MCU.");
		ao2_unlock(conference);
		ao2_ref(conference, -1);
		return 0;
	}

	ret = conf_send_avs_mcu_control(conference, AVS_CONTROL_ENABLE_MCU);

	ao2_unlock(conference);

	ao2_ref(conference, -1);

	if (ret == R_SUCCESS)	{
		astman_send_ack(s, m, "Conference MCU enabled.");
	} else {
		astman_send_error(s, m, "send avs control error");
	}

	return 0;
}

static int action_confbridgedisablemcu(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	struct confbridge_conference *conference = NULL;
	int ret = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	ao2_lock(conference);

	ret = conf_send_avs_mcu_control(conference, AVS_CONTROL_DISABLE_MCU);

	ao2_unlock(conference);

	ao2_ref(conference, -1);

	if (ret == R_SUCCESS)	{
		astman_send_ack(s, m, "Conference MCU diabled.");
	} else {
		astman_send_error(s, m, "send avs control error");
	}

	return 0;
}

static int action_videorecordalarmon(struct mansession *s, const struct message *m)
{
	struct ao2_iterator iter;
	struct confbridge_conference *conference = NULL;

	ast_log(LOG_DEBUG, "Receive action ConfbridgeVideoRecordAlarmOn, set alarm_on true\n");

	alarm_on = 1;

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (conference->video_is_recording) {
			ao2_lock(conference);
			send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_VIDEO_RECORD);
			ao2_unlock(conference);
		}
		if (conference->slides_is_recording) {
			ao2_lock(conference);
			send_avs_video_record_control(conference, NULL, AVS_CONTROL_STOP_CONF_SLIDES_RECORD);
			ao2_unlock(conference);
		}
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	astman_send_ack(s, m, "video record alarm on success");

	return 0;
}

static int action_videorecordalarmoff(struct mansession *s, const struct message *m)
{
	ast_log(LOG_DEBUG, "Receive action ConfbridgeVideoRecordAlarmOff, set alarm_on false\n");

	alarm_on = 0;

	astman_send_ack(s, m, "video record alarm on success");

	return 0;
}
#endif

static int action_confbridgesetsinglevideosrc(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel = astman_get_header(m, "Channel");
	struct confbridge_user *user;
	struct confbridge_conference *conference;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel)) {
		astman_send_error(s, m, "No channel name provided.");
		return 0;
	}
	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	/* find channel and set as video src. */
	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (!strncmp(channel, ast_channel_name(user->chan), strlen(channel))) {
			ast_bridge_set_single_src_video_mode(conference->bridge, user->chan);
			break;
		}
	}
	ao2_unlock(conference);
	ao2_ref(conference, -1);

	/* do not access user after conference unlock.  We are just
	 * using this check to see if it was found or not */
	if (!user) {
		astman_send_error(s, m, "No channel by that name found in conference.");
		return 0;
	}
	astman_send_ack(s, m, "Conference single video source set.");
	return 0;
}

static int action_confbridge_dump_pcm_data(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *opt = astman_get_header(m, "Opt");
	struct confbridge_conference *conference;
	struct confbridge_user *user = NULL;
	struct ast_avs_dump_pcm pcm_param;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided\n");
		return 0;
	}

	if (ast_strlen_zero(opt)) {
		astman_send_error(s, m, "No opt provided\n");
		return 0;
	}

	if (!ao2_container_count(conference_bridges)) {
		astman_send_error(s, m, "No active conferences.");
		return 0;
	}

	ast_log(LOG_DEBUG, "action_confbridge_dump_pcm_data conference:%s, opt:%s\n", conference_name, opt);

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "No Conference by that name found.");
		return 0;
	}

	memset(&pcm_param, 0, sizeof(pcm_param));
	pcm_param.opt = atoi(opt);
	pcm_param.type = AVS_DUMP_PCM;

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user->chan) {
			strncpy(pcm_param.chan_name, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"), sizeof(pcm_param.chan_name) - 1);
			ast_avs_control(user->chan, AVS_CONTROL_DUMP_PCM, &pcm_param, sizeof(pcm_param));

			if( ast_media_role_get() == MEDIA_ROLE_MASTER
				&& !ast_strlen_zero(ast_channel_get_avs_address(user->chan))
				&& strcmp("127.0.0.1",ast_channel_get_avs_address(user->chan))){

				if(pcm_param.opt)
					ast_multi_avs_rsync_file(AST_MULTI_AVS_DUMP_FILE_ENUM,1);//rsync all time,added by zyxia for Bug 274060,2024.2.21
				else {
					ast_multi_avs_rsync_file(AST_MULTI_AVS_DUMP_FILE_ENUM,0);//stop rsync all time
					ast_sync_record_from_slave(AST_RSYNC_RECORD_CMD_NO_LINK,ast_channel_get_avs_address(user->chan),AST_MULTI_AVS_DUMP_FILE_PATH,AST_MULTI_AVS_DUMP_FILE_PATH,AST_AVS_SYNC_RECORD_MODULE);
				}
			}
			if (ast_ha_permit_send() && !pcm_param.opt) {
				ast_ha_rsync_file(AST_AVS_DUMP_ENUM);
			}
		}
	}
	ao2_unlock(conference);
	ao2_ref(conference, -1);

	astman_send_ack(s, m, "Conference dump PCM data action started\n");

	return 0;
}

static int action_confbridge_extend_time_helper(struct mansession *s, const struct message *m, int allow)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	const char *type = astman_get_header(m, "Type");
	struct confbridge_user *user = NULL;
	struct conf_sched_task_data *task_data = NULL;
	int conf_sch_timeout = 0;
	int user_sch_timeout = 0;
	ast_log(LOG_DEBUG, "action_confbridge_extend_time entry");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}
	if (ast_strlen_zero(type)) {
		astman_send_error(s, m, "No Extend provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	if (!strcasecmp(type, "conf")) {
		ast_log(LOG_DEBUG, "action_confbridge_extend_time conf");
		int conf_timeleft = ast_sched_when(conf_sched,  conference->conf_sched_id);
		ao2_lock(conference);
		if (conference->sch_time < 0 || conference->conf_end) {
			ao2_unlock(conference);
			astman_send_error(s, m, "nothing need to do.");
			ast_log(LOG_DEBUG, "conference sch_time reach.\n");
			return 0;
		} else if (0 == conference->sch_time && !allow) {
			AST_LIST_TRAVERSE(&conference->active_list, user, list) {
				if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
					conference->extend_time_status = EXTEND_REJECTED;
					conf_timeout_notify_build_send(user->chan, conference->name, 1, conference->extend_time_status, conf_timeleft);
				}
			}
			ao2_unlock(conference);
			astman_send_ack(s, m, "rejected extend time.");
			ast_log(LOG_DEBUG, "conference rejected extend time,in the last time.\n");
			return 0;
		}

		if (allow) {
			int extra_time = 0;
			/*extend the meeting*/
			if (0 == conference->sch_time && !ast_tvzero(conference->last_notify)) {
				struct timeval delay_tv = {0,0};
				delay_tv = ast_tvnow();
				/*extra time wait for input.*/
				extra_time = ast_tvdiff_ms(delay_tv, conference->last_notify);
				ast_log(LOG_DEBUG, "extra_time %d", extra_time);
			}
			conf_sch_timeout = conference->b_profile.conf_timeout - (CONF_SCH_RETRY - conference->sch_time)*CONF_PEER_TIMEOUT;
			/* Get the remaining time of the last 5 minutes */
			if (CONF_PEER_TIMEOUT - extra_time > 0) {
				conf_sch_timeout = conf_sch_timeout + (CONF_PEER_TIMEOUT - extra_time);
			}
			conference->sch_time = CONF_SCH_RETRY;
			conference->conf_will_timeout = 0;
			conference->extend_time_status = EXTEND_CONFIRMED;
		} else {
			/*other input,stop next sched task*/
			conf_sch_timeout = CONF_PEER_TIMEOUT *conference->sch_time;
			conference->sch_time = 1;
			conference->sch_time--;
			conference->extend_time_status = EXTEND_REJECTED;
		}

		ao2_unlock(conference);
		if ((conference->conf_sched_id = ast_sched_replace(conference->conf_sched_id, conf_sched, conf_sch_timeout, conference_timeout_notify, conference->name)) < 0) {
			ast_log(LOG_ERROR, "Unable to replace scheduler timer.\n");
		} else {
			ao2_lock(conference);
			AST_LIST_TRAVERSE(&conference->active_list, user, list) {
				if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
					conf_timeout_notify_build_send(user->chan, conference->name, 1, conference->extend_time_status, conf_sch_timeout/1000);
				}
			}
			ao2_unlock(conference);
			ast_log(LOG_DEBUG,"replace ast sched success! conf_sch_timeout %d, conference->sch_time %d!", conf_sch_timeout, conference->sch_time);
		}
		astman_send_ack(s, m, "Conf scheduler timer started.");
		ast_log(LOG_DEBUG, "conf sch time left %ld!!", ast_sched_when(conf_sched,  conference->conf_sched_id));
	} else {
		ast_log(LOG_DEBUG, "action_confbridge_extend_time user");
		user = find_active_user_by_channel(conference, channel_name);
		if (user && user->chan) {
			ao2_lock(conference);
			if (-1 == user->user_sched_id) {
				ast_log(LOG_DEBUG, "user sched task exec,return!");
				astman_send_error(s, m, "User sched task exec.");
				ao2_unlock(conference);
				return 0;
			}
			if (!conference->only_warning.user) {
				ast_log(LOG_ERROR, "only warning user null!");
				astman_send_error(s, m, "User null.");
				ao2_unlock(conference);
				return 0;
			}
			if (conference->only_warning.user != user) {
				ast_log(LOG_ERROR, "only warning user not match!");
				astman_send_error(s, m, "User not match.");
				ao2_unlock(conference);
				return 0;
			}
			task_data  = (struct conf_sched_task_data *)ast_sched_find_data(user_sched, user->user_sched_id);
			if (!task_data) {
				ast_log(LOG_ERROR, "Couldn't malloc task data.");
				astman_send_error(s, m, "Couldn't malloc task data.");
				ao2_unlock(conference);
				return 0;
			}
			ast_log(LOG_DEBUG, "user %s has sche!!", task_data ? task_data->user_channel : "");
			ast_log(LOG_DEBUG, "user sch time left %ld!!", ast_sched_when(user_sched,  user->user_sched_id));
			/*extend the meeting*/
			if (allow) {
				/*extend the meeting*/
				user_sch_timeout = user->b_profile.one_person_warning;
				conference->only_warning.need_warning = 1;
				conference->only_warning.extend_time_status = EXTEND_CONFIRMED;
			} else {
				/*nothing input*/
				user_sch_timeout = CONF_PEER_TIMEOUT;
				conference->only_warning.need_warning = 0;
				conference->only_warning.extend_time_status = EXTEND_REJECTED;
			}
			ast_log(LOG_DEBUG, "user->user_sched_id %d.\n", user->user_sched_id);
			if (!user_sch_timeout) {
				ast_log(LOG_ERROR, "user timeout not enable.");
				astman_send_error(s, m, "user timeout not enable.");
				ao2_unlock(conference);
				return 0;
			}
			conf_timeout_notify_build_send(user->chan, conference->name, 0, allow ? EXTEND_CONFIRMED : EXTEND_REJECTED, user_sch_timeout/1000);
			if ((user->user_sched_id = ast_sched_replace(user->user_sched_id, user_sched, user_sch_timeout, only_user_timeout_notify, task_data)) < 0) {
				ast_log(LOG_ERROR, "Unable to replace scheduler timer.\n");
			} else {
				ast_log(LOG_DEBUG,"replace ast sched success! conf_sch_timeout %d, need_warning %d user_sched_id %d!", user_sch_timeout, conference->only_warning.need_warning, user->user_sched_id);
			}
			ao2_unlock(conference);
			ast_log(LOG_DEBUG, "user sch time left %ld!!", ast_sched_when(user_sched, user->user_sched_id));
			astman_send_ack(s, m, "User scheduler timer started.");
		} else {
			astman_send_error(s, m, "Couldn't find user.");
			ast_log(LOG_DEBUG, "Couldn't find user.");
		}
	}

	return 0;
}

static int action_confbridge_focus_helper(struct mansession *s, const struct message *m, int cancel)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	const char *originator = astman_get_header(m, "Originator");
	const char *layout = astman_get_header(m, "Layout");
	struct confbridge_user *user_admin = NULL;
	struct confbridge_user *user = NULL;
	struct confbridge_user *focus_cancel_user = NULL;
	int isfocus = 1;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}
	if (ast_strlen_zero(originator)) {
		astman_send_error(s, m, "No originator provided.");
		return 0;
	}
	if (ast_strlen_zero(layout)) {
		astman_send_error(s, m, "No Conference layout provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	ao2_lock(conference);
	user = find_active_user_by_channel(conference, channel_name);
	if (!user || !user->chan) {
		ao2_unlock(conference);
		astman_send_error(s, m, "Couldn't find the conference user.");
		return 0;
	}

	user_admin = find_active_user_by_channel(conference, originator);
	if (user_admin) {
		if (!ast_test_flag(&user_admin->u_profile, USER_OPT_ADMIN) && strcasecmp(channel_name, originator)) {
			ao2_unlock(conference);
			ast_log(LOG_DEBUG, "originator not admin.");
			astman_send_error(s, m, "originator not admin.");
			return 0;
		}
	} else {
		ao2_unlock(conference);
		ast_log(LOG_DEBUG, "originator not exist.");
		astman_send_error(s, m, "originator not exist.");
		return 0;
	}
	if (cancel) {
		const char *user_chan_name = NULL;
		const char *focus_chan_name = NULL;
		/*cancel focus*/
		if (!conference->focus_chan) {
			ao2_unlock(conference);
			astman_send_error(s, m, "confernce no focus.");
			return 0;
		}
		user_chan_name = ast_channel_name(user->chan);
		focus_chan_name = ast_channel_name(conference->focus_chan);
		if (user_chan_name && focus_chan_name && strcasecmp(user_chan_name, focus_chan_name)) {
			ao2_unlock(conference);
			ast_log(LOG_DEBUG, "focus chan not match.");
			astman_send_error(s, m, "focus chan not match.");
			return 0;
		}
		isfocus = 0;
		user->isfocus = 0;
		ast_channel_unref(conference->focus_chan);
		conference->focus_chan = NULL;
		memset(conference->conf_layout, 0, sizeof(conference->conf_layout));
		ast_copy_string(conference->conf_layout, DEFAULT_CONF_LAYOUT, sizeof(conference->conf_layout));
		ast_log(LOG_DEBUG, "cancel focus done %s.", focus_chan_name);
	} else {
		/*focus configure*/
		if (conference->focus_chan && !strcasecmp(ast_channel_name(conference->focus_chan), channel_name)) {
			ao2_unlock(conference);
			ast_log(LOG_DEBUG, "already focus, now focus chan %s.", ast_channel_name(conference->focus_chan));
			astman_send_error(s, m, "focus channel already focus.");
			return 0;
		} else if (conference->focus_chan && strcasecmp(ast_channel_name(conference->focus_chan), channel_name)) {
			/*replace focus*/
			ast_log(LOG_DEBUG, "old focus channel %s.", ast_channel_name(conference->focus_chan));
			if (ast_channel_has_enabled_main_video_stream(user->chan)) {
				/*Notify pjsip session to disable focus video*/
				ast_indicate(conference->focus_chan, AST_CONTROL_NOTIFY_DISABLE_FOCUS_VIDEO);
			} else {
				/*Notify AVS*/
				ast_indicate(conference->focus_chan, AST_CONTROL_NOTIFY_CLOSE_FOCUS_VIDEO);
			}
			focus_cancel_user = find_active_user_by_channel(conference, ast_channel_name(conference->focus_chan));
			ast_channel_cleanup(conference->focus_chan);
			conference->focus_chan = ao2_bump(user->chan);
		} else if (!conference->focus_chan) {
			/*set focus*/
			conference->focus_chan = ao2_bump(user->chan);
			ast_log(LOG_DEBUG, "set focus done.");
		}
		user->isfocus = 1;
		memset(conference->conf_layout, 0, sizeof(conference->conf_layout));
		ast_copy_string(conference->conf_layout, layout, sizeof(conference->conf_layout));
	}
	if (conference->focus_chan) {
		ast_log(LOG_DEBUG, "now focus channel %s.", ast_channel_name(conference->focus_chan));
	}
	/*Notify AVS*/
	ast_channel_lock(user->chan);
	if (ast_channel_tech(user->chan)->indicate) {
		ast_channel_tech(user->chan)->indicate(user->chan, isfocus ? AST_CONTROL_NOTIFY_OPEN_FOCUS_VIDEO : AST_CONTROL_NOTIFY_CLOSE_FOCUS_VIDEO, NULL, 0);
	}
	ast_channel_unlock(user->chan);

	conf_update_msid_for_focus(conference, user, isfocus);

	ao2_unlock(conference);

	send_focus_status_event(user, conference, isfocus);
	if (focus_cancel_user && focus_cancel_user->chan) {
		/*Modify for Bug 227210*/
		focus_cancel_user->isfocus = 0;
		ast_log(LOG_DEBUG, "Notify cancel status to focus user.");
		send_focus_status_event(focus_cancel_user, conference, 0);
	}
	/*record operation*/
	if (user_admin) {
		conf_user_write_operator_log(user_admin, isfocus ? CONF_OP_SET_FOCUS : CONF_OP_DEL_FOCUS, NULL);
	} else {
		ast_log(LOG_DEBUG, "originator not found!");
	}
	astman_send_ack(s, m, "Conference focus set.");
	return 0;
}

static int action_confbridge_layout_helper(struct mansession *s, const struct message *m, int cancel)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *originator = astman_get_header(m, "Originator");
	const char *layout = astman_get_header(m, "Layout");
	struct confbridge_user *originator_user = NULL;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(originator)) {
		astman_send_error(s, m, "No originator provided.");
		return 0;
	}
	if (ast_strlen_zero(layout)) {
		astman_send_error(s, m, "No Conference layout provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}
	ao2_lock(conference);
	memset(conference->conf_layout, 0, sizeof(conference->conf_layout));
	if (cancel) {
		ast_copy_string(conference->conf_layout, DEFAULT_CONF_LAYOUT, sizeof(conference->conf_layout));
	} else {
		ast_copy_string(conference->conf_layout, layout, sizeof(conference->conf_layout));
	}
	ast_log(LOG_DEBUG, "conf layout %s!", layout);

	send_layout_status_event(conference);

	originator_user = find_active_user_by_channel(conference, originator);
	if (originator_user) {
		conf_user_write_operator_log(originator_user, cancel ? CONF_OP_DEL_LAYOUT : CONF_OP_SET_LAYOUT, conference->conf_layout);
	}
	ao2_unlock(conference);

	astman_send_ack(s, m, "Conference focus set.");
	return 0;
}

static int action_confbridge_allow_extend_time(struct mansession *s, const struct message *m)
{
	return action_confbridge_extend_time_helper(s, m, 1);
}

static int action_confbridge_reject_extend_time(struct mansession *s, const struct message *m)
{
	return action_confbridge_extend_time_helper(s, m, 0);
}

static int action_confbridge_set_layout(struct mansession *s, const struct message *m)
{
	const char *layout = astman_get_header(m, "Layout");
	if (ast_strlen_zero(layout)) {
		astman_send_error(s, m, "No Conference layout provided.");
		return 0;
	}
	if (!strcasecmp(layout, "focus")) {
		return action_confbridge_focus_helper(s, m, 0);
	} else {
		return action_confbridge_layout_helper(s, m, 0);
	}
}

static int action_confbridge_cancel_layout(struct mansession *s, const struct message *m)
{
	const char *layout = astman_get_header(m, "Layout");
	if (ast_strlen_zero(layout)) {
		astman_send_error(s, m, "No Conference layout provided.");
		return 0;
	}
	if (!strcasecmp(layout, "focus")) {
		return action_confbridge_focus_helper(s, m, 1);
	} else {
		return action_confbridge_layout_helper(s, m, 1);
	}
}

static int action_confbridge_user_permission_helper(struct mansession *s, const struct message *m, int cancel)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *originator = astman_get_header(m, "Originator");
	const char *permission_type = astman_get_header(m, "Type");
	struct confbridge_user *admin_user = NULL;
	struct confbridge_user *originator_user = NULL;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(originator)) {
		astman_send_error(s, m, "No originator provided.");
		return 0;
	}
	if (ast_strlen_zero(permission_type)) {
		astman_send_error(s, m, "No Conference Type provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	ao2_lock(conference);
	originator_user = find_active_user_by_channel(conference, originator);
	if (!originator_user || !originator_user->chan) {
		ao2_unlock(conference);
		astman_send_error(s, m, "Couldn't find the conference originator user.");
		return 0;
	}

	AST_LIST_TRAVERSE(&conference->active_list, admin_user, list) {
		if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
			break;
		}
	}

	if (!admin_user || admin_user->chan != originator_user->chan) {
		ao2_unlock(conference);
		astman_send_error(s, m, "No permission");
		return 0;
	}

	if (cancel) {
		if (!strcasecmp(permission_type, "mute")) {
			/*cancel allow_user_unmute*/
			conference->allow_user_unmute = 0;
		}
	} else {
		if (!strcasecmp(permission_type, "mute")) {
			conference->allow_user_unmute = 1;
		}
	}
	send_user_permission_event(conference, permission_type);
	if (originator_user) {
		conf_user_write_operator_log(originator_user, cancel ? CONF_OP_DEL_USER_PERMISSION : CONF_OP_SET_USER_PERMISSION, permission_type);
	}
	ao2_unlock(conference);

	astman_send_ack(s, m, "permission set.");
	return 0;
}

static int action_confbridge_set_user_permission(struct mansession *s, const struct message *m)
{
	return action_confbridge_user_permission_helper(s, m, 0);
}

static int action_confbridge_cancel_user_permission(struct mansession *s, const struct message *m)
{
	return action_confbridge_user_permission_helper(s, m, 1);
}

static int action_confbridge_remove_wait_users(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct schedule_conference_wait *, conference_obj, NULL, ao2_cleanup);
	struct schedule_attendees_wait *schedule_wait_user = NULL;
	struct schedule_conference_wait_info conf_sch_w = { 0 };

	const char *conference_name = astman_get_header(m, "Conference");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}

	conf_sch_w.conf_name = conference_name;
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference_name);
		astman_send_error(s, m, "Couldn't find the conference.");
		return 0;
	}

	ao2_lock(conference_obj);

	AST_LIST_TRAVERSE(&conference_obj->schedule_attendees_wait_list, schedule_wait_user, entry) {
		if(schedule_wait_user->chan) {
			ast_log(LOG_DEBUG, "softhangup chan '%s'\n", ast_channel_name(schedule_wait_user->chan));
			ast_softhangup(schedule_wait_user->chan, AST_SOFTHANGUP_DEV);
		}
	}
	conference_obj->cancelled = 1;
	if (0 == conference_obj->wait_users) {
		schedule_conference_wait_list_remove(conference_name);
		ast_log(LOG_DEBUG, "schedule conference no users, wait list cancel'%s'\n", conference_name);
	}
	ao2_unlock(conference_obj);
	astman_send_ack(s, m, "remove wait users.");
	return 0;
}

static int indicate_notify_stop_recording(struct confbridge_conference *conference, enum ast_record_stop_type type)
{
	struct confbridge_user *cur_user = NULL;
	enum ast_record_stop_type stop_type = type;

	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		ast_indicate_data(cur_user->chan, AST_CONTROL_NOTIFY_STOP_RECORD, &stop_type, sizeof(stop_type));
	}
	ao2_unlock(conference);

	return 0;
}

static void conf_notify_stop_record(enum ast_record_stop_type type, const char *device_type)
{
	struct ao2_iterator iter;
	struct confbridge_conference *conference = NULL;

	if (ast_strlen_zero(device_type)) {
		return;
	}

	ast_log(LOG_DEBUG, "conf_notify_device_pulled_out pulled out device type:%s\n", device_type);

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		ao2_lock(conference);
		ast_log(LOG_DEBUG, "conference record_link:%s\n", conference->record_link);
		if (!strcasecmp(conference->record_link, device_type)) {
			indicate_notify_stop_recording(conference, type);
			stop_record(conference);
			send_record_status_event(conference, "none", 0);
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

	return;
}

static int action_device_monitor(struct mansession *s, const struct message *m)
{
	const char *monitor_type = astman_get_header(m, "MonitorType");
	const char *device_type = astman_get_header(m, "DeviceType");

	if (ast_strlen_zero(monitor_type)) {
		astman_send_error(s, m, "No monitor type provided.");
		return AMI_SUCCESS;
	}

	if (ast_strlen_zero(device_type)) {
		astman_send_error(s, m, "No device type provided.");
		return AMI_SUCCESS;
	}

	ast_log(LOG_DEBUG, "action_device_monitor monitor_type:%s, device_type:%s\n", monitor_type, device_type);

	if (!strcasecmp(monitor_type, "out")) {
		conf_notify_stop_record(AST_RECORD_STOP_TYPE_DEVICE_OUT, device_type);
		ast_stop_all_recording_call(AST_RECORD_STOP_TYPE_DEVICE_OUT, device_type);
	}

	astman_send_ack(s, m, "Device monitor success");

	return AMI_SUCCESS;
}

static int func_confbridge_info(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *parse;
	struct confbridge_conference *conference;
	struct confbridge_user *user;
	int count = 0;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(type);
		AST_APP_ARG(confno);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.confno) || ast_strlen_zero(args.type)) {
		return -1;
	}
	conference = ao2_find(conference_bridges, args.confno, OBJ_KEY);
	if (!conference) {
		snprintf(buf, len, "0");
		return 0;
	}

	/* get the correct count for the type requested */
	ao2_lock(conference);
	if (!strcasecmp(args.type, "parties")) {
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			count++;
		}
		AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
			count++;
		}
	} else if (!strcasecmp(args.type, "admins")) {
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
				count++;
			}
		}
	} else if (!strcasecmp(args.type, "marked")) {
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_MARKEDUSER)) {
				count++;
			}
		}
	} else if (!strcasecmp(args.type, "locked")) {
		count = conference->locked;
	} else if (!strcasecmp(args.type, "muted")) {
		count = conference->muted;
	} else {
		ast_log(LOG_ERROR, "Invalid keyword '%s' passed to CONFBRIDGE_INFO.\n", args.type);
	}
	snprintf(buf, len, "%d", count);
	ao2_unlock(conference);
	ao2_ref(conference, -1);
	return 0;
}

void conf_add_user_active(struct confbridge_conference *conference, struct confbridge_user *user)
{
	AST_LIST_INSERT_TAIL(&conference->active_list, user, list);
	conference->activeusers++;
#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
		conference->mcb_idle_kick_users++;
	}
#endif
}

void conf_add_user_marked(struct confbridge_conference *conference, struct confbridge_user *user)
{
	AST_LIST_INSERT_TAIL(&conference->active_list, user, list);
	conference->activeusers++;
	conference->markedusers++;
#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
		conference->mcb_idle_kick_users++;
	}
#endif
}

void conf_add_user_waiting(struct confbridge_conference *conference, struct confbridge_user *user)
{
	AST_LIST_INSERT_TAIL(&conference->waiting_list, user, list);
	conference->waitingusers++;
}

void conf_remove_user_active(struct confbridge_conference *conference, struct confbridge_user *user)
{
	AST_LIST_REMOVE(&conference->active_list, user, list);
	conference->activeusers--;
#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
		conference->mcb_idle_kick_users--;
	}
#endif
}

void conf_remove_user_marked(struct confbridge_conference *conference, struct confbridge_user *user)
{
	AST_LIST_REMOVE(&conference->active_list, user, list);
	conference->activeusers--;
	conference->markedusers--;
#ifdef GRANDSTREAM_NETWORKS
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
		conference->mcb_idle_kick_users--;
	}
#endif
}

void conf_mute_only_active(struct confbridge_conference *conference)
{
	struct confbridge_user *only_user = AST_LIST_FIRST(&conference->active_list);

	/* Turn on MOH if the single participant is set up for it */
	if (ast_test_flag(&only_user->u_profile, USER_OPT_MUSICONHOLD)) {
		conf_moh_start(only_user);
	}
	conf_update_user_mute(only_user);
}

void conf_remove_user_waiting(struct confbridge_conference *conference, struct confbridge_user *user)
{
	AST_LIST_REMOVE(&conference->waiting_list, user, list);
	conference->waitingusers--;
}

#ifdef GRANDSTREAM_NETWORKS
static int ast_sip_cei_same_channel(const char* channel, const char* mychannel, const char *cid_number)
{
    // Sanity check
    if (channel == NULL || mychannel == NULL) {
        return -1;
    }

    // Exact match
    if (strcmp(channel, mychannel) == 0) {
        return 0;
    }

    // Invited call: Match extension name
    // channel: Local/1002-446300@ccm-invite-00000001;2 (by CCM)
	//          Local/1002@from-internal-00000001;1 (by WebUI)
    // mychannel: PJSIP/1002-00000001
    if (strstr(channel, "Local") != NULL) {
		char *extension = ast_strdupa(channel + 6);
        char *pchannel = NULL;
        if (extension == NULL) {
            return -1;
        }

		// CCM invite
		if (strstr(channel, "ccm-invite") != NULL) {
			pchannel = strchr(extension, '-');
		} else if (strstr(channel, "from-internal") != NULL) {
			pchannel = strchr(extension, '@');
		}

		if (pchannel == NULL) {
			return -1;
		}

		*pchannel = '\0';

		if ((strstr(mychannel, extension) != NULL) || (cid_number && strstr(mychannel, cid_number) != NULL)) {
            return 0;
        }
	}

    return -1;
}

static struct confbridge_user* find_active_user_by_channel(struct confbridge_conference *conference,
														   const char *channel)
{
	struct confbridge_user *user;

	// Sanity check
	if (conference == NULL || channel == NULL) {
		return NULL;
	}

	ao2_lock(conference);

	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (ast_sip_cei_same_channel(ast_channel_name(user->chan), channel, ast_channel_caller(user->chan)->id.number.str) == 0)
		{
			ao2_unlock(conference);
			return user;
		}
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		if (ast_sip_cei_same_channel(ast_channel_name(user->chan), channel, ast_channel_caller(user->chan)->id.number.str) == 0)
		{
			ao2_unlock(conference);
			return user;
		}
	}

	ao2_unlock(conference);

	return NULL;
}

static const char* get_channel_name_by_user_no(struct confbridge_conference *conference,
										       int user_no)
{
	struct confbridge_user *user;

	// Sanity check
	if (conference == NULL) {
		return NULL;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user != NULL && user->user_no == user_no)
		{
			ao2_unlock(conference);
			return ast_channel_name(user->chan);
		}
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		if (user != NULL && user->user_no == user_no)
		{
			ao2_unlock(conference);
			return ast_channel_name(user->chan);
		}
	}
	ao2_unlock(conference);

	return NULL;
}

static int stasis_action_lock_unlock_helper(const char *conference_name, int lock, const char *originator)
{
	int res = 0;

	if (ast_strlen_zero(conference_name)) {
		ast_log(LOG_WARNING, "No Conference name provided.");
		return 0;
	}

	if (!ao2_container_count(conference_bridges)) {
		ast_log(LOG_WARNING, "No active conferences.");
		return 0;
	}

	if ((res = generic_lock_unlock_helper(lock, conference_name, originator))) {
		ast_log(LOG_WARNING, "No Conference by that name found.");
		return 0;
	}

	ast_log(LOG_WARNING, lock ? "Conference locked" : "Conference unlocked");

	return 0;
}

static int stasis_action_kick_user(struct confbridge_conference *conference,
								   struct confbridge_user *user, int user_no)
{
	struct confbridge_user *cur_user = NULL;

	// Sanity check
	if (conference == NULL || user == NULL) {
		return -1;
	}

	// Check user permission
	if (!ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		ast_log(LOG_WARNING, "Only admin users can kick a user. "
							 "Channel %s of conf %s is not an admin.\n",
							 ast_channel_name(user->chan),
							 conference->name);
		return -1;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		ast_log(LOG_WARNING, "[WL] user_no: %d, USER_OPT_MARKED_MCB: 0x%X",
							 cur_user->user_no,
							 ast_test_flag(&cur_user->u_profile, USER_OPT_MARKED_MCB));

		if (cur_user->user_no == user_no &&
			!ast_test_flag(&cur_user->u_profile, USER_OPT_MARKED_MCB)) {
			publish_cei_unsub(conference->name, ast_channel_name(cur_user->chan), ast_channel_caller(cur_user->chan)->id.number.str);
			cur_user->kicked = 1;
			pbx_builtin_setvar_helper(cur_user->chan, "CONFBRIDGE_RESULT", "KICKED");
			ast_bridge_remove(conference->bridge, cur_user->chan);
			conf_user_write_operator_log(user, CONF_OP_KICK, S_COR(ast_channel_caller(cur_user->chan)->id.number.valid, ast_channel_caller(cur_user->chan)->id.number.str, "Anonymous"));
		}
	}
	ao2_unlock(conference);

	return 0;
}

static void ccm_command_kick_cb(void *data, struct stasis_subscription *sub,
							    struct stasis_message *message)
{
	struct ast_json			*json_object	= NULL;
	struct ast_json_payload *payload		= NULL;
	const char				*conference_name = NULL;
	const char				*user_channel	= NULL;
	const char				*entity			= NULL;
	int user_no = -1;
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_DEBUG, "Got NULL payload from stasis message.");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));
	user_channel = ast_json_string_get(ast_json_object_get(json_object, "user_channel"));
	entity = ast_json_string_get(ast_json_object_get(json_object, "entity"));
	if (conference_name == NULL || user_channel == NULL || entity == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (conference == NULL) {
		ast_log(LOG_WARNING, "Failed to find conference %s", conference_name);
		return;
	}

	user = find_active_user_by_channel(conference, user_channel);
	if (user == NULL) {
		ast_log(LOG_WARNING, "Failed to find user with channel %s", user_channel);
		return;
	}

	// Kick all users
	if (strcmp(entity, "ALL") == 0) {
		action_kick_all_non_admin(conference, NULL, user);
	}
	// Kick all MCBs
	else if (strcmp(entity, "ALL_MCB") == 0) {
		action_kick_mcb(conference, NULL, user);
	}
	// Kick a user
	else {
		user_no = atoi(entity);
		stasis_action_kick_user(conference, user, user_no);
	}
}

static void ccm_command_mute_unmute_cb(void *data, struct stasis_subscription *sub,
									    struct stasis_message *message)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	const char *conference_name = NULL;
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	const char *entity = NULL;
	const char *entity_channel = NULL;
	const char *originator = NULL;
	int muted = 0;
	int user_no = -1;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_DEBUG, "Got NULL payload from stasis message.");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));
	entity = ast_json_string_get(ast_json_object_get(json_object, "entity"));
	muted = ast_json_integer_get(ast_json_object_get(json_object, "muted"));
	originator = ast_json_string_get(ast_json_object_get(json_object, "originator"));
	if (conference_name == NULL || entity == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}

	// Mute/Unmute all non-admin users
	if (strcmp(entity, "ALL") == 0) {
		generic_mute_unmute_helper(muted, conference_name, "participants", originator, 0);

		const char *sound_to_play = NULL;

		conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
		if (conference == NULL) {
			ast_log(LOG_WARNING, "Failed to find conference %s", conference_name);
			return;
		}

		sound_to_play = conf_get_sound((muted ? CONF_SOUND_PARTICIPANTS_MUTED : CONF_SOUND_PARTICIPANTS_UNMUTED),
			conference->b_profile.sounds);

		play_sound_helper(conference, sound_to_play, -1);
	}
	// Mute/Unmute a user
	else {
		conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
		if (conference == NULL) {
			ast_log(LOG_WARNING, "Failed to find conference %s", conference_name);
			return;
		}

		user_no = atoi(entity);
		entity_channel = get_channel_name_by_user_no(conference, user_no);
		if (entity_channel == NULL) {
			ast_log(LOG_WARNING, "Failed to get user channel of user# %s in %s",
								 entity, conference_name);
			return;
		}

		generic_mute_unmute_helper(muted, conference_name, entity_channel, originator, 0);
	}
}

static void ccm_command_lock_cb(void *data, struct stasis_subscription *sub,
							    struct stasis_message *message)
{
	struct ast_json			*json_object = NULL;
	struct ast_json_payload *payload	 = NULL;
	const char *conference_name = NULL;
	const char *originator = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_DEBUG, "Got NULL payload from stasis message.");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));
	if (conference_name == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}
	originator = ast_json_string_get(ast_json_object_get(json_object, "originator"));

	stasis_action_lock_unlock_helper(conference_name, 1, originator);
}

static void ccm_command_unlock_cb(void *data, struct stasis_subscription *sub,
								  struct stasis_message *message)
{
	struct ast_json			*json_object = NULL;
	struct ast_json_payload *payload	 = NULL;
	const char *conference_name = NULL;
	const char *originator = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_DEBUG, "Got NULL payload from stasis message.");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	conference_name = ast_json_string_get(ast_json_object_get(json_object, "conference"));
	if (conference_name == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}
	originator = ast_json_string_get(ast_json_object_get(json_object, "originator"));

	stasis_action_lock_unlock_helper(conference_name, 0, originator);
}

static void cei_toggle_enable_cb(void *data, struct stasis_subscription *sub,
								  struct stasis_message *message)
{
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	const char				*enableCEI   = NULL;
	char buffer[128] = {0};
	int res = 0;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	enableCEI = ast_json_string_get(ast_json_object_get(json_object, "enableCEI"));
	if (enableCEI == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}

	if (strcmp(enableCEI, "true") == 0) {
		globalEnabledCEI = 1;
	} else if (strcmp(enableCEI, "false") == 0) {
		globalEnabledCEI = 0;
	}

	snprintf(buffer, sizeof(buffer), "%s%d", "nvram set enableCEI=", globalEnabledCEI);
	/*Add by yuyang for replace the system with popen*/
	FILE  *fp = NULL;
	fp = popen(buffer, "r");
	if (fp != NULL)
	{
	    pclose(fp);
	    res = 0;
	}
	else
	{
	    res = -1;
	}
	fp = NULL;
	fp = popen("nvram commit", "r");
	if (fp != NULL)
	{
	    pclose(fp);
	    res = 0;
	}
	else
	{
	    res = -1;
	}
	//res = system(buffer);
	//res = system("nvram commit");
	/*End by yuyang for replace the system with popen*/

	if (res) {
		return;
	}
}

static int ast_cei_get_enable_configuration(void)
{
	char buf[64] = { 0 };
	char buffer[256] = { 0 };
	int enableCEI = 0;

	FILE *fp = NULL;
	fp = popen("nvram get enableCEI", "r");

	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
		      && (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
			strcat(buffer, buf);
		}
		pclose(fp);
		fp = NULL;

		if (buffer[0] != '\0') {
		   /*parsing result*/
		   enableCEI = atoi( buffer );
		}
	}

	return enableCEI;
}

static int mcb_auto_release_process(const void *data)
{
	const char *conference_name = NULL;
	struct confbridge_user * user = NULL;
	struct confbridge_conference *conference = NULL;

	conference_name = (const char *)data;
	if (!conference_name) {
		return -1;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (conference) {
		if (conference->mcb_idle_kick_users == 0) {
			ao2_ref(conference, -1);
			return 0;
		}
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
				if (conference->activeusers + conference->waitingusers == conference->mcb_idle_kick_users) {
					ast_log(LOG_DEBUG, "[MCB] kick mcb user %s", ast_channel_name(user->chan));
					user->kicked = 1;
					pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
					ast_bridge_remove(conference->bridge, user->chan);
				}
			}
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
	}
	return 0;
}

static void mcb_user_check_remove_sched(const char *conference_name)
{
	struct confbridge_user * user = NULL;
	struct confbridge_conference *conference = NULL;

	if (!conference_name) {
		return;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (conference) {
		if (conference->mcb_idle_kick_users == 0) {
			ao2_ref(conference, -1);
			return;
		}
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (user->user_mcb_sched_id) {
				int ret = ast_sched_del(recycle_sched, user->user_mcb_sched_id);
				/* ast_sched_del Will not clean the value of user->user_mcb_sched_id */
				user->user_mcb_sched_id = 0;
				ast_log(LOG_DEBUG, "[MCB] mcb user %s remove sched,ast sched del ret %d", ast_channel_name(user->chan), ret);
			}
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
	}
}

static void mcb_user_check_add_sched(const char *conference_name)
{
	struct confbridge_user * user = NULL;
	struct confbridge_conference *conference = NULL;

	if (!conference_name) {
		return;
	}

	conference = ao2_find(conference_bridges, conference_name, OBJ_KEY);
	if (conference) {
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (user->user_mcb_sched_id <= 0) {
				if ((user->conference->activeusers + user->conference->waitingusers) == user->conference->mcb_idle_kick_users) {
					if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
						if ((user->user_mcb_sched_id = ast_sched_add(recycle_sched, MCB_IDLE_KICK_SECONDS * 1000, mcb_auto_release_process, user->conference->name)) < 0) {
							ast_log(LOG_ERROR, "[MCB] Unable to add mcb user %s scheduler timer!", ast_channel_name(user->chan));
						} else {
							ast_log(LOG_DEBUG,"[MCB] create mcb user %s sched success! user_sch_timeout %d, user_sched_id %d!", ast_channel_name(user->chan), MCB_IDLE_KICK_SECONDS * 1000, user->user_mcb_sched_id);
						}
					} else {
						ast_log(LOG_DEBUG, "[MCB] normal user %s join", ast_channel_name(user->chan));
					}
				}
			}
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
	}
}

static void handle_mcb_user_join(struct confbridge_user * user)
{
	if (!user || !user->conference) {
		return;
	}

	if (user->conference->mcb_idle_kick_users == 0) {
		return;
	}
	if ((user->conference->activeusers + user->conference->waitingusers) == user->conference->mcb_idle_kick_users) {
		if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
			if ((user->user_mcb_sched_id = ast_sched_add(recycle_sched, MCB_IDLE_KICK_SECONDS * 1000, mcb_auto_release_process, user->conference->name)) < 0) {
				ast_log(LOG_ERROR, "[MCB] Unable to add mcb user %s scheduler timer!", ast_channel_name(user->chan));
			} else {
				ast_log(LOG_DEBUG,"[MCB] create mcb user %s sched success! user_sch_timeout %d, user_sched_id %d!", ast_channel_name(user->chan), MCB_IDLE_KICK_SECONDS * 1000, user->user_mcb_sched_id);
			}
		} else {
			ast_log(LOG_DEBUG, "[MCB] normal user %s join", ast_channel_name(user->chan));
		}
	}
	else {
		mcb_user_check_remove_sched(user->conference->name);
	}
}

static void handle_mcb_user_remove(struct confbridge_user * user)
{
	if (!user || !user->conference) {
		return;
	}
	if (user->user_mcb_sched_id) {
		int ret = ast_sched_del(recycle_sched, user->user_mcb_sched_id);
		/* ast_sched_del Will not clean the value of user->user_mcb_sched_id */
		user->user_mcb_sched_id = 0;
		ast_log(LOG_DEBUG, "[MCB] mcb user %s remove sched,ast sched del ret %d", ast_channel_name(user->chan), ret);
	}
	mcb_user_check_add_sched(user->conference->name);
}

#if 0
static void mcb_auto_release_process(void)
{
	struct confbridge_conference *conference;
	struct confbridge_user * user;
	struct ao2_iterator iter;
	struct timeval now;


	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		if (conference->mcb_idle_kick_users == 0) {
			ao2_ref(conference, -1);
			continue;
		}
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB) && ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB_IDLE_KICK)) {
				now = ast_tvnow();
				if (!user->kicktime) {
					user->kicktime = now.tv_sec + MCB_IDLE_KICK_SECONDS;
				}
				else {
					if (conference->activeusers != conference->mcb_idle_kick_users || conference->waitingusers > 0) {
						user->kicktime = now.tv_sec + MCB_IDLE_KICK_SECONDS;
					}
				}

				if (user->kicktime && (user->kicktime <= now.tv_sec)) {
					user->kicked = 1;
					pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
					ast_bridge_remove(conference->bridge, user->chan);
				}
			}
		}
		ao2_unlock(conference);
		ao2_ref(conference, -1);
	}
	ao2_iterator_destroy(&iter);

}
static void* mcb_release_thread_action(void *data) {
	while (mcb_release_thread_run) {
		struct timespec ts = { 0, };
		struct timeval wait;

		wait = ast_tvadd(ast_tvnow(), ast_samp2tv(MCB_RELEASE_COUNT, MCB_RELEASE_FREQUENCY));
		ts.tv_sec = wait.tv_sec;
		ts.tv_nsec = wait.tv_usec * 1000;

		ast_mutex_lock(&mcb_release_lock);
		ast_cond_timedwait(&mcb_release_cond, &mcb_release_lock, &ts);
		ast_mutex_unlock(&mcb_release_lock);

		if (!mcb_release_thread_run)
			break;

		mcb_auto_release_process();
	}

	return NULL;
}

static void start_mcb_release_thread(void)
{
	int errcode;

	if (mcb_release_thread != AST_PTHREADT_NULL) {
		ast_log(LOG_ERROR, "ZAKU DEBUG ---> MCB RELEASE THREAD IS NOT NULL, NOT CREATING ANOTHER ONE");
		return;
	}

	mcb_release_thread_run = 1;
	if ((errcode = ast_pthread_create(&mcb_release_thread, NULL, mcb_release_thread_action, NULL))) {
		ast_log(LOG_ERROR, "ZAKU DEBUG ---> COULD NOT CREATE THREAD: %s\n", strerror(errcode));
	} else {
		ast_log(LOG_DEBUG, "ZAKU DEBUG ---> CREATED MCB RELEASE THREAD");
	}
}

static void stop_mcb_release_thread(void) {
	mcb_release_thread_run = 0;

	ast_mutex_lock(&mcb_release_lock);
	ast_cond_signal(&mcb_release_cond);
	ast_mutex_unlock(&mcb_release_lock);

	pthread_join(mcb_release_thread, NULL);
	mcb_release_thread = AST_PTHREADT_NULL;

}
#endif
static void print_channel_info(struct ast_channel *c)
{
	if (!c)
	{
		return;
	}

	ast_channel_lock(c);
	ast_log(LOG_DEBUG, "Channel:%s, Stream:%p, Timingfunc:%p, Timingdata:%p\n",
		ast_channel_name(c), ast_channel_stream(c), ast_channel_timingfunc(c), ast_channel_timingdata(c));
	if (ast_channel_stream(c)) {
		ast_log(LOG_DEBUG, "Channel:%s, Stream ref:%d\n", ast_channel_name(c), ao2_ref(ast_channel_stream(c), 0));
	}
	if (ast_channel_timingdata(c) && ast_test_flag(ast_channel_flags(c), AST_FLAG_TIMINGDATA_IS_AO2_OBJ)) {
		ast_log(LOG_DEBUG, "Channel:%s, Timingdata ref:%d\n", ast_channel_name(c), ao2_ref(ast_channel_timingdata(c), 0));
	}
	ast_channel_unlock(c);
}

#if 0
static int conf_one_person_warning_callback(struct ast_bridge_channel *bridge_channel, void *hook_pvt)
{
	int res = 0;
	int file_buf_len = 1024;
	char *files = ast_alloca(file_buf_len);
	char *file = NULL;
	struct confbridge_user *user = hook_pvt;
	struct confbridge_conference *conference = user->conference;
	struct conference_one_person_warning *warning = &conference->only_warning;
	int already_suspended = 0;

	ao2_lock(conference);
	if (!warning->user) {
		ao2_unlock(conference);
		return -1;
	}
	warning->playing = 1;
	ao2_unlock(conference);

	/* See if music on hold is playing */
	conf_moh_suspend(user);

	already_suspended = bridge_channel->suspended;
	if (!already_suspended) {
		bridge_channel_suspend(bridge_channel);
	}
	ast_indicate(bridge_channel->chan, AST_CONTROL_SRCUPDATE);

	print_channel_info(bridge_channel->chan);

	snprintf(files, file_buf_len, "%s&%s",
			conf_get_sound(CONF_SOUND_ONLY_PERSON, user->b_profile.sounds),
			conf_get_sound(CONF_SOUND_CONTINUE_WAIT, user->b_profile.sounds));
	while ((file = strsep(&files, "&")) && !user->kicked) {
		res = ast_stream_and_wait(bridge_channel->chan, file, "1");
		if (!res) {
			continue;
		} else if (res < 0) {
			ast_log(LOG_WARNING, "Failed to playback file %s to channel %s\n", file, ast_channel_name(bridge_channel->chan));
			warning->id = -1;
			warning->playing = 0;
			goto out;
		} else {
			break;
		}
	}

	print_channel_info(bridge_channel->chan);
	ast_stopstream(bridge_channel->chan);

	if (!res && !user->kicked) {
		res = ast_waitfordigit(bridge_channel->chan, 5000);
	}

	warning->playing = 0;

	if (user->kicked) {
		goto out;
	}

	if (res != '1') {
		if ((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			publish_cei_unsub(conference->name, ast_channel_name(bridge_channel->chan), ast_channel_caller(bridge_channel->chan)->id.number.str);
		}
		user->kicked = 1;
		pbx_builtin_setvar_helper(bridge_channel->chan, "CONFBRIDGE_RESULT", "KICKED");
		ast_bridge_remove(bridge_channel->bridge, bridge_channel->chan);
	} else {
		ao2_lock(conference);
		if (warning->user && warning->user == user) {
			if (ast_bridge_interval_hook_seqno(bridge_channel->features, 0, user->b_profile.one_person_warning_interval,
				conf_one_person_warning_callback, hook_pvt, NULL, AST_BRIDGE_HOOK_REMOVE_ON_PULL, (unsigned int *)&warning->id)) {
				ast_log(LOG_WARNING, "Could not add conference warning callback into the bridge channel '%s' interval hook.\n",
					ast_channel_name(bridge_channel->chan));
				warning->id = -1;
			}
		}
		ao2_unlock(conference);
	}

out:
	ast_indicate(bridge_channel->chan, AST_CONTROL_SRCUPDATE);
	if (!already_suspended) {
		bridge_channel_unsuspend(bridge_channel);
	}

	/* See if music on hold needs to be started back up again */
	conf_moh_unsuspend(user);
	return -1;
}
#endif

static int conf_one_person_warning_remove(struct confbridge_conference *conference)
{
	struct confbridge_user *user = NULL;

	if (!conference) {
		return -1;
	}

	user = conference->only_warning.user;
	if (!user) {
		ast_log(LOG_WARNING, "There is no one person warning user.\n");
		return -1;
	}

	conference->only_warning.user = NULL;
#if 0
	if (conference->only_warning.id < 0) {
		ast_log(LOG_WARNING, "The one person warning id for the user '%s' is invalid.\n", ast_channel_name(user->chan));
		return -1;
	}
#endif
	ast_log(LOG_DEBUG, "Remove conference one person warning from the user '%s'. Warning is %s\n",
		ast_channel_name(user->chan), conference->only_warning.playing ? "playing" : "not playing");
	if (conference->only_warning.playing) {
		struct ast_frame frame = {
			.frametype = AST_FRAME_DTMF_END,
			.subclass.integer = '1',
		};
		ast_queue_frame(user->chan, &frame);
	} else {
		//ast_bridge_interval_hook_remove(&user->features, conference->only_warning.id);
	}

	//conference->only_warning.id = -1;
	conference->only_warning.need_warning = 0;
	if (user->user_sched_id) {
		struct conf_sched_task_data *task_data =NULL;
		task_data  = (struct conf_sched_task_data *)ast_sched_find_data(user_sched, user->user_sched_id);
		int ret = ast_sched_del(user_sched, user->user_sched_id);
		if (-1 == user->user_sched_id) {
			ast_log(LOG_DEBUG, "user sched exec happened!");
		} else {
			if (task_data) {
				ast_log(LOG_DEBUG, "task_data %s", task_data->user_channel);
				ao2_ref(task_data, -1);
			}
		}
		ast_log(LOG_DEBUG, "only user remove,ast sched del ret %d", ret);
		/*Send Notify to wave cancel user tiemout*/
		if (conference->only_warning.extend_time_status == EXTEND_PENDING
			&& (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
			ast_log(LOG_DEBUG, "WAVE and APP only send Notify!");
			conf_timeout_notify_build_send(user->chan, conference->name, 0, EXTEND_CONFIRMED, 0);
		}
	}
	return 0;
}

static int conf_timeout_notify_build_send(struct ast_channel *chan, char *conf_name, int type, int status, int time)
{
	if (!chan && ast_strlen_zero(conf_name)) {
		return -1;
	}
	struct conf_timeout_notify_data s_data;
	memset(&s_data, 0, sizeof(s_data));
	snprintf(s_data.conf_id, sizeof(s_data.conf_id), "%s", conf_name);
	/*0 means user type, 1 means conf type*/
	s_data.notify_type = type;
	/*1 means pending, 2 means confirmed, 3 means rejected*/
	s_data.status = status;
	s_data.time = time;
	ast_channel_lock(chan);
	if (ast_channel_tech(chan)->indicate) {
		ast_channel_tech(chan)->indicate(chan, AST_CONTROL_NOTIFY_CONF_TIMEOUT, &s_data, sizeof(s_data));
	}
	ast_channel_unlock(chan);
	return 0;
}

/*If callback function return non-zero, we should schedule them to be
run again.*/
static int only_user_timeout_notify(const void *data)
{
	int delay_time = 0;
	int extra_time = 0;
	int res = 0;
	int file_buf_len = 1024;
	int notify_to_wave = 0;
	char *files = ast_alloca(file_buf_len);
	char *file = NULL;
	ast_log(LOG_DEBUG, "only_user_timeout_notify entry!\n");
	struct confbridge_user *user = NULL;
	RAII_VAR(struct conf_sched_task_data *, task_data, NULL, ao2_cleanup);
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct conference_one_person_warning *warning = NULL;
	int need_kick = 0;

	task_data = (struct conf_sched_task_data *)data;
	if (!task_data || !task_data->conf_name || !task_data->user_channel) {
		ast_log(LOG_ERROR,"user or conference null!");
		return 0;
	}
	conference = ao2_find(conference_bridges, task_data->conf_name, OBJ_KEY);
	if (!conference) {
		ast_log(LOG_ERROR,"can not find conference!");
		return 0;
	}

	ao2_lock(conference);
	warning = &conference->only_warning;
	if (!warning->user) {
		ast_log(LOG_ERROR,"can not find warning user!");
		ao2_unlock(conference);
		return 0;
	}
	/* -1 means user sche task runing */
	warning->user->user_sched_id = -1;
	user = find_active_user_by_channel(conference, task_data->user_channel);
	if (!user) {
		ast_log(LOG_ERROR,"can not find user!");
		ao2_unlock(conference);
		return 0;
	}

	ao2_unlock(conference);

	struct timeval now_tv = {0,0};
	struct timeval delay_tv = {0,0};
	now_tv = ast_tvnow();

	if (warning->need_warning > 0) {
		int default_extend = 0;
		int conf_timeleft = 0;
		if (conference->conf_sched_id) {
			conf_timeleft = ast_sched_when(conf_sched, conference->conf_sched_id);
		}
		ao2_lock(conference);
		warning->extend_time_status = EXTEND_PENDING;
		conf_timeleft = conf_timeleft + conference->sch_time*CONF_PEER_TIMEOUT/1000;
		ast_log(LOG_DEBUG, "conf timeout timeleft %d(s), conference->sch_time %d!", conf_timeleft, conference->sch_time);
		if (conf_timeleft <= (CONF_PEER_TIMEOUT/1000)*6) {
			default_extend = 1;
			ast_log(LOG_DEBUG, "conf timeout timeleft less than 30min, extend user time 30min for default.");
		}
		ao2_unlock(conference);
		if (!default_extend) {
			if (warning->user->term_type == AST_TERMINAL_TYPE_WEBRTC || warning->user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
				ao2_lock(conference);
				ast_log(LOG_DEBUG, "WAVE and APP only send Notify!");
				if (!ast_tvzero(conference->schedule_alert_time) &&
					!ast_strlen_zero(conference->schedule_id_to_start) &&
					(ast_strlen_zero(conference->schedule_id_processing) ||
					(!ast_strlen_zero(conference->schedule_id_processing) &&
					strcasecmp(conference->schedule_id_to_start, conference->schedule_id_processing)))) {
					ast_log(LOG_DEBUG, "Schedule alert happend, don't send user timeout Notify!");
				} else {
					conf_timeout_notify_build_send(user->chan, conference->name, 0, EXTEND_PENDING, 5*60);
				}
				res = 0;
				notify_to_wave = 1;
				ao2_unlock(conference);
			} else {
				ast_log(LOG_DEBUG, "Play sound to Phone!");
				ao2_lock(conference);
				warning->playing = 1;
				ao2_unlock(conference);

				/* See if music on hold is playing */
				conf_moh_suspend(user);

				ast_bridge_suspend(conference->bridge, user->chan);
				ast_indicate(user->chan, AST_CONTROL_SRCUPDATE);

				print_channel_info(user->chan);

				snprintf(files, file_buf_len, "%s&%s",
						conf_get_sound(CONF_SOUND_ONLY_PERSON, user->b_profile.sounds),
						conf_get_sound(CONF_SOUND_END_FIVE_MINUTES, user->b_profile.sounds));
				while ((file = strsep(&files, "&")) && !user->kicked) {
					res = ast_play_and_wait(user->chan, file);
					if (!res) {
						continue;
					} else if (res < 0) {
						ast_log(LOG_WARNING, "Failed to playback file %s to channel %s\n", file, ast_channel_name(user->chan));
						warning->playing = 0;
						need_kick = 1;
						goto out;
					} else {
						break;
					}
				}

				print_channel_info(user->chan);
				ast_stopstream(user->chan);

				if (!res && !user->kicked) {
					res = ast_waitfordigit(user->chan, 5000);
				}
				if (res < 0) {
					need_kick = 1;
					ast_log(LOG_DEBUG, "waitfordigit error!");
					goto out;
				}

				warning->playing = 0;

				if (user->kicked) {
					goto out;
				}
			}
		} else {
			/*extend the meeting for default 30min*/
			res = 49;
		}
		delay_tv = ast_tvnow();
		/*extra time for play file and get digital.*/
		extra_time = ast_tvdiff_ms(delay_tv, now_tv);
		ast_log(LOG_DEBUG, "extra_time %d res %c", extra_time, res);

		ao2_lock(conference);
		if (!warning->user) {
			ast_log(LOG_ERROR,"can not find warning user!");
			ao2_unlock(conference);
			/*Modify for Bug 241160,
			Unsuspend is required,
			otherwise the channel will be blocked in "bridge_channel_wait" and cannot hang up*/
			goto out;
		}
		if (res == '1') {
			/*extend the meeting*/
			delay_time = user->b_profile.one_person_warning;
			warning->need_warning = 1;
			warning->extend_time_status = EXTEND_CONFIRMED;
		} else {
			/*nothing input*/
			/*other input,stop next sched task*/
			delay_time = CONF_PEER_TIMEOUT;
			warning->need_warning = 0;
			/*Notify to wave, status only can be changed by AMI action*/
			if (!notify_to_wave) {
				warning->extend_time_status = EXTEND_REJECTED;
			}
		}

		if (delay_time - extra_time > 0) {
			delay_time = delay_time - extra_time;
		}
		if (delay_time > 0) {
			ao2_ref(task_data, +1);
			if ((user->user_sched_id = ast_sched_add(user_sched, delay_time, only_user_timeout_notify, task_data)) < 0) {
				ast_log(LOG_ERROR, "Add scheduler timer failure!\n");
				ao2_ref(task_data, -1);
			} else {
				ast_log(LOG_DEBUG, "only_user_timeout_notify next delay_time %d need_warning %d user_sched_id %d\n", delay_time, warning->need_warning, user->user_sched_id);
			}
		}
		ao2_unlock(conference);
	} else {
		ast_log(LOG_DEBUG, "need warning is zero!\n");
		need_kick = 1;
	}

out:
	if (need_kick) {
		if ((globalEnabledCEI && conference->is_conference_service) || ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			publish_cei_unsub(conference->name, ast_channel_name(user->chan), ast_channel_caller(user->chan)->id.number.str);
		}
		user->kicked = 1;
		pbx_builtin_setvar_helper(user->chan, "CONFBRIDGE_RESULT", "KICKED");
		ast_bridge_remove(conference->bridge, user->chan);
		ast_log(LOG_DEBUG, "only user leave %s", ast_channel_name(user->chan));
	}
	ast_indicate(user->chan, AST_CONTROL_SRCUPDATE);
	ast_bridge_unsuspend(conference->bridge, user->chan);
	/* See if music on hold needs to be started back up again */
	conf_moh_unsuspend(user);

	return 0;
}

static void conf_sched_task_data_destructor(void *obj)
{
	struct conf_sched_task_data *task_data = obj;
	ast_log(LOG_DEBUG, "conf sched task data destructor!");
	if (task_data) {
		ast_free(task_data->user_channel);
		ast_free(task_data->conf_name);
	}
}

static struct conf_sched_task_data *conf_sched_task_data_alloc(const char *conf_name, const char *user_channel)
{
	if (!conf_name || !user_channel) {
		return NULL;
	}
	struct conf_sched_task_data *task_data = NULL;

	task_data = ao2_alloc_options(sizeof(*task_data), conf_sched_task_data_destructor,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!task_data) {
		return NULL;
	}

	task_data->user_channel = ast_strdup(user_channel);
	task_data->conf_name = ast_strdup(conf_name);

	return task_data;
}

static int conf_one_person_warning_add(struct confbridge_user *user)
{
	int user_sch_timeout = -1;
	struct confbridge_conference *conference = NULL;

	if (!user || !user->chan || !user->conference) {
		return -1;
	}

	ast_log(LOG_DEBUG, "Add conference one person warning to the user '%s'\n", ast_channel_name(user->chan));
	struct conf_sched_task_data *task_data = NULL;
	task_data = conf_sched_task_data_alloc(user->conference->name, ast_channel_name(user->chan));
	if (!task_data) {
		return -1;
	}
	conference = user->conference;
	conference->only_warning.user = user;
	conference->only_warning.id = -1;
	conference->only_warning.playing = 0;
	conference->only_warning.need_warning = 1;
	conference->only_warning.extend_time_status = EXTEND_INIT;
#if 0
	if (ast_bridge_interval_hook_seqno(&user->features, 0, user->b_profile.one_person_warning,
		conf_one_person_warning_callback, user, NULL, AST_BRIDGE_HOOK_REMOVE_ON_PULL,
		(unsigned int *)&conference->only_warning.id)) {
		ast_log(LOG_WARNING, "Could not add conference one person warning callback into the user '%s' interval hook.\n",
			ast_channel_name(user->chan));
		conference->only_warning.id = -1;
		return -1;
	}

	if (ast_channel_is_avs_engine(user->chan)) {
		/* Wake up the bridge channel thread. */
		ast_queue_frame(user->chan, &ast_null_frame);
	}
#endif
	user_sch_timeout = user->b_profile.one_person_warning;
	if (user_sch_timeout >= 0) {
		if ((user->user_sched_id = ast_sched_add(user_sched, user_sch_timeout, only_user_timeout_notify, task_data)) < 0) {
			ast_log(LOG_ERROR, "Unable to add scheduler timer.\n");
			ao2_ref(task_data, -1);
		} else {
			ast_log(LOG_DEBUG,"create user sched success! user_sch_timeout %d, need_warning %d user_sched_id %d!", user_sch_timeout, conference->only_warning.need_warning, user->user_sched_id);
		}
	}

	return 0;
}

static int conf_user_join_handle_one_person_warning(struct confbridge_user *user)
{
	int ret = -1;
	struct confbridge_conference *conference = NULL;

	if (!user) {
		return -1;
    }

	if (!user->b_profile.one_person_warning) {
		return 0;
	}

	conference = user->conference;
	ao2_lock(conference);
	if ((conference->activeusers + conference->waitingusers) == 1) {
		if (!ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB)) {
			if (!conference->only_warning.user) {
				ast_log(LOG_DEBUG, "Add conference one person warning to the user '%s'\n", ast_channel_name(user->chan));
				ret = conf_one_person_warning_add(user);
			}
		}
	} else if (conference->only_warning.user) {
		ast_log(LOG_DEBUG, "Remove conference one person warning from the user '%s'\n", ast_channel_name(conference->only_warning.user->chan));
		ret = conf_one_person_warning_remove(conference);
	}
	ao2_unlock(conference);

	return ret;
}

static int conf_user_leave_handle_one_person_warning(struct confbridge_user *user)
{
	struct confbridge_conference *conference = NULL;

	if (!user) {
		return -1;
    }

	if (!user->b_profile.one_person_warning) {
		return 0;
	}

	conference = user->conference;
	ao2_lock(conference);
	if (conference->only_warning.user) {
		if (conference->only_warning.user == user) {
			conference->only_warning.user = NULL;
			conference->only_warning.id = -1;
			conference->only_warning.need_warning = 0;
			if (user->user_sched_id) {
				ast_log(LOG_DEBUG, "user_sched_id %d!", user->user_sched_id);
				struct conf_sched_task_data *task_data =NULL;
				task_data  = (struct conf_sched_task_data *)ast_sched_find_data(user_sched, user->user_sched_id);
				int ret = ast_sched_del(user_sched, user->user_sched_id);
				if (-1 == user->user_sched_id) {
					ast_log(LOG_DEBUG, "user sched task exec!");
				} else {
					if (task_data) {
						ast_log(LOG_DEBUG, "task_data %s", task_data->user_channel);
						ao2_ref(task_data, -1);
					}
				}
				ast_log(LOG_DEBUG, "only user remove,ast sched del ret %d", ret);
			}
		} else {
			ast_log(LOG_WARNING, "Should not reach here! Warning for the user '%s'\n", ast_channel_name(conference->only_warning.user->chan));
		}
	} else if ((conference->activeusers + conference->waitingusers) == 1) {
		struct confbridge_user *only_user = NULL;

		if (conference->activeusers == 1) {
			only_user = AST_LIST_FIRST(&conference->active_list);
		} else if (conference->waitingusers == 1) {
			only_user = AST_LIST_FIRST(&conference->waiting_list);
		}

		if (only_user && !ast_test_flag(&only_user->u_profile, USER_OPT_MARKED_MCB)) {
			ast_log(LOG_DEBUG, "Add conference one person warning to the user '%s'\n", ast_channel_name(only_user->chan));
			conf_one_person_warning_add(only_user);
		}
	}
	ao2_unlock(conference);

	return 0;
}

int conf_user_is_video_call(struct confbridge_user *user)
{
	struct ast_stream_topology *topology = NULL;
	int i = 0;
	int is_video = 0;

	ast_channel_lock(user->chan);
	if (ast_channel_state(user->chan) != AST_STATE_UP) {
		ast_channel_unlock(user->chan);
		return -1;
	}

	topology =  ast_channel_get_stream_topology(user->chan);
	if (!topology) {
		ast_channel_unlock(user->chan);
		return -1;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		struct ast_stream *stream = ast_stream_topology_get_stream(topology, i);

		if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO
			&& ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
			if (ast_stream_get_state(stream) != AST_STREAM_STATE_REMOVED &&
				ast_stream_get_state(stream) != AST_STREAM_STATE_SENDONLY &&
				ast_stream_get_state(stream) != AST_STREAM_STATE_INACTIVE) {
				is_video = 1;
			}
			break;
		}
	}

	ast_channel_unlock(user->chan);

	user->video = is_video;
	return is_video;
}

struct ast_channel *ast_conf_get_sfu_moderator(struct ast_channel *chan, const char *conf_number)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	RAII_VAR(struct cache_conference *, cache_conf, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;
	struct cache_authorized_user *auth_user = NULL;
	struct ast_channel *moderator = NULL;

	if (ast_strlen_zero(conf_number)) {
		return NULL;
	}

	if ((conference = ao2_find(conference_bridges, conf_number, OBJ_KEY))) {
		if (!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			return NULL;
		}

		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
				moderator = ao2_bump(user->chan);
				break;
			}
		}
		ao2_unlock(conference);
	}

	if (!moderator && (cache_conf = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL))) {
		ao2_lock(cache_conf);
		if (cache_conf->moderator) {
			moderator = ao2_bump(cache_conf->moderator);
		} else {
			AST_LIST_TRAVERSE(&cache_conf->auth_user_list, auth_user, entry) {
				if (auth_user->is_admin) {
					char channel_name[AST_CHANNEL_NAME] = {0};
					snprintf(channel_name, sizeof(channel_name), "PJSIP/%s", auth_user->extension);
					/*Auth by ConfbridgeSfuAuthorizeUser but channel not set.Bug 234362*/
					if (chan && ast_channel_name(chan) && strstr(ast_channel_name(chan), channel_name)) {
							int terminate_type = AST_TERMINAL_TYPE_UNKNOW;
							const char *terminate_str = NULL;
							terminate_type = ast_channel_get_terminate_type(chan);
							if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
								terminate_str = "webrtc";
							} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
								terminate_str = "wave_app";
							} else {
								terminate_str = "other_type";
							}
							if (strcasecmp(terminate_str, "other_type") && !strcasecmp(auth_user->terminal_type, terminate_str)) {
								moderator = ao2_bump(chan);
								ast_log(LOG_DEBUG, "moderator is himself");
							}
							else {
								ast_log(LOG_DEBUG, "The identity of the moderator is not clear");
							}
					}
					break;
				}
			}
		}
		ao2_unlock(cache_conf);
	}

	return moderator;
}

struct ast_channel *ast_conf_get_sfu_focus(struct ast_channel *chan, const char *conf_number)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;
	struct ast_channel *focus = NULL;

	if (ast_strlen_zero(conf_number)) {
		return NULL;
	}

	if ((conference = ao2_find(conference_bridges, conf_number, OBJ_KEY))) {
		if (!ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
			return NULL;
		}

		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (user->chan == conference->focus_chan) {
				focus = ao2_bump(user->chan);
				break;
			}
		}
		ao2_unlock(conference);
	}

	return focus;
}


static void cache_authorized_user_destroy(void *obj)
{
	struct cache_authorized_user *user = obj;

	ast_log(LOG_DEBUG, "cache authorized user '%s|%s' is destroying.\n", user->extension, user->channel_name);
}

static int cache_conf_authorized_user_update(struct cache_authorized_user *user, int is_admin, const char *scehdule_id)
{
	user->start_time = ast_tvnow();
	user->is_admin = is_admin;
	/*Modify for Bug 191227*/
	memset(user->user_schedule_id, 0, sizeof(user->user_schedule_id));
	ast_copy_string(user->user_schedule_id, scehdule_id ? scehdule_id : "", sizeof(user->user_schedule_id));
	return 0;
}

static int cache_conf_authorized_user_add(struct cache_conference *conference, const char *extension, int is_admin, const char *terminal_type, const char *scehdule_id,  const char *related_info)
{
	struct cache_authorized_user *user = NULL;

	if (!conference || ast_strlen_zero(extension)) {
		return -1;
	}

	user = ao2_alloc(sizeof(*user), cache_authorized_user_destroy);
	if (!user) {
		return -1;
	}
	ast_copy_string(user->extension, extension, sizeof(user->extension));
	user->start_time = ast_tvnow();
	user->is_admin = is_admin;
	ast_copy_string(user->terminal_type, terminal_type?terminal_type:"other_type", sizeof(user->terminal_type));
	ast_copy_string(user->user_schedule_id, scehdule_id?scehdule_id:"", sizeof(user->user_schedule_id));
	ast_copy_string(user->related_info, related_info?related_info:"", sizeof(user->related_info));
	ast_log(LOG_DEBUG, "Add cache authorized user '%s' is_admin:%d.\n", user->extension, user->is_admin);
	ao2_lock(conference);
	AST_LIST_INSERT_TAIL(&conference->auth_user_list, user, entry);
	ao2_unlock(conference);

	return 0;
}

static struct cache_authorized_user *cache_conf_authorized_user_find(struct cache_conference *conference, const char *extension, const char *terminal_type)
{
	struct cache_authorized_user *user = NULL;
	if (ast_strlen_zero(terminal_type)) {
		return NULL;
	}
	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
		if (!strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminal_type) && ast_strlen_zero(user->channel_name)) {
			ao2_ref(user, +1);
			break;
		}
	}
	ao2_unlock(conference);

	return user;
}

static void cache_conf_authorized_user_remove(struct cache_conference *conference, const char *extension, const char *terminal_type)
{
	struct cache_authorized_user *user = NULL;

	ao2_lock(conference);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->auth_user_list, user, entry) {
		if (!strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminal_type)) {
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(user);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ao2_unlock(conference);
}

static void cache_conf_remove_user(const char *conf_number, const char *extension, const char *terminal_type)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(extension)) {
		return;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return;
	}

	cache_conf_authorized_user_remove(conference, extension, terminal_type);
}

static void cache_conf_authorized_user_remove_by_channel_name(struct cache_conference *conference, const char *channel_name)
{
	struct cache_authorized_user *user = NULL;

	ao2_lock(conference);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->auth_user_list, user, entry) {
		if (!strcmp(user->channel_name, channel_name)) {
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(user);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ao2_unlock(conference);
}

static void cache_conf_remove_user_by_channel_name(const char *conf_number, const char *channel_name)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(channel_name)) {
		return;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return;
	}

	cache_conf_authorized_user_remove_by_channel_name(conference, channel_name);
}

static int cache_conf_authorized_user_set_channel(const char *conf_number, const char *extension, struct ast_channel *chan)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	RAII_VAR(struct cache_authorized_user *, user, NULL, ao2_cleanup);
	const char * terminate_str = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;

	if (!chan ||ast_strlen_zero(conf_number) || ast_strlen_zero(extension) || ast_strlen_zero(ast_channel_name(chan))) {
		return -1;
	}

	terminate_type = ast_channel_get_terminate_type(chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	user = cache_conf_authorized_user_find(conference, extension, terminate_str);
	if (!user) {
		return -1;
	}
	pbx_builtin_setvar_helper(chan, "RELATED_INFO", user->related_info);
	ast_copy_string(user->channel_name, ast_channel_name(chan), sizeof(user->channel_name));
	return 0;
}

static void cache_conference_destroy(void *obj)
{
	struct cache_conference *conference = obj;
	struct cache_authorized_user *user=  NULL;

	ast_log(LOG_DEBUG, "cache conference '%s' is destroying.\n", conference->name);

	if (conference->moderator) {
		ast_channel_unref(conference->moderator);
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->auth_user_list, user, entry){
		AST_LIST_REMOVE_CURRENT(entry);
		ao2_cleanup(user);
	}
	AST_LIST_TRAVERSE_SAFE_END;
}

static struct cache_conference *cache_conf_fetch_add(const char *conf_number, int type, int add, const char *schedule_id)
{
	struct cache_conference *conference = NULL;

	if (ast_strlen_zero(conf_number)) {
		return NULL;
	}

	AST_LIST_LOCK(&cache_conference_list);

	AST_LIST_TRAVERSE(&cache_conference_list, conference, entry) {
		if (!strcmp(conference->name, conf_number)) {
			ao2_ref(conference, +1);
			break;
		}
	}

	if (add && !conference) {
		if ((conference = ao2_alloc(sizeof(*conference), cache_conference_destroy))) {
			ast_log(LOG_DEBUG, "Add a new cache conference '%s'.\n", conf_number);
			ast_copy_string(conference->name, conf_number, sizeof(conference->name));
			conference->type = type;
			ast_copy_string(conference->schedule_id, schedule_id ? schedule_id : "", sizeof(conference->schedule_id));
			AST_LIST_INSERT_TAIL(&cache_conference_list, conference, entry);
			ao2_ref(conference, +1);
		}
	}
	AST_LIST_UNLOCK(&cache_conference_list);

	return conference;
}

static void cache_conf_remove(const char *conf_number)
{
	struct cache_conference *conference = NULL;

	if (ast_strlen_zero(conf_number)) {
		return;
	}

	ast_log(LOG_DEBUG, "Remove the cache conference '%s'.\n", conf_number);
	AST_LIST_LOCK(&cache_conference_list);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&cache_conference_list, conference, entry)
	{
		if (!strcasecmp(conf_number, "all") || !strcmp(conference->name, conf_number)) {
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(conference);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&cache_conference_list);
}

static int cache_conf_remove_moderator(const char *conf_number, const struct ast_channel *chan)
{
	int res = -1;
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;

	if (!chan) {
		return -1;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	if (conference->moderator == chan) {
		ast_channel_unref(conference->moderator);
		conference->moderator = NULL;
		res = 0;
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->auth_user_list, user, entry) {
		if (!strcmp(user->channel_name, ast_channel_name(chan))) {
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(user);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ao2_unlock(conference);

	return res;
}

static int cache_conf_clear_moderator(const char *conf_number, const struct ast_channel *chan)
{
	int res = -1;
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;

	if (!chan) {
		return -1;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	if (conference->moderator == chan) {
		ast_channel_unref(conference->moderator);
		conference->moderator = NULL;
		res = 0;
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->auth_user_list, user, entry) {
		if (user->is_admin && !strcmp(user->channel_name, ast_channel_name(chan))) {
			user->is_admin = 0;
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ao2_unlock(conference);

	return res;
}

static int cache_conf_check_user_authorization(const char *conf_number, const char *extension, char *schedule_id)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;

	if (ast_strlen_zero(extension)) {
		return -1;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}
	ast_log(LOG_DEBUG, "extension %s!", extension);

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
		if (!strcmp(user->extension, extension)) {
			ast_copy_string(schedule_id, user->user_schedule_id, AST_MAX_SCHEDULE_ID);
			break;
		}
	}
	ao2_unlock(conference);

	return (user ? 0 : -1);
}

static int cache_conf_check_user_role(const char *conf_number, struct ast_channel *chan, const char *extension, enum ast_terminal_type terminate_type)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, moderator_chan, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;
	int is_admin = -1;
	const char * terminate_str = NULL;
	int host_exist = 0;

	if (ast_strlen_zero(extension) || !chan) {
		return -1;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	moderator_chan = ast_conf_get_sfu_moderator(NULL, conf_number);
	if (moderator_chan && ast_channel_name(chan) && ast_channel_name(moderator_chan) && strcmp(ast_channel_name(chan), ast_channel_name(moderator_chan))) {
		is_admin = 0;
		host_exist = 1;
		ast_log(LOG_DEBUG, "The host exist, %s want be host!", ast_channel_name(chan));
		pbx_builtin_setvar_helper(chan, "WANT_BE_HOST", "yes");
		const char *wait_reservation_id = pbx_builtin_getvar_helper(chan, "WAIT_RESERVATION_ID");
		if (wait_reservation_id && strcasecmp(wait_reservation_id, conference->schedule_id)) {
			pbx_builtin_setvar_helper(chan, "WANT_BE_HOST", "");
			ast_log(LOG_DEBUG, "chan waiting for schedule meeting start!");
			ao2_lock(conference);
			AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
				if (user->is_admin) {
					if (!strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminate_str)) {
						is_admin = user->is_admin;
						host_exist = 0;
						ast_log(LOG_DEBUG, "when schedule meeting start %s can be host!", ast_channel_name(chan));
						break;
					}
				}
			}
			ao2_unlock(conference);

		}
	} else if (moderator_chan && ast_channel_name(moderator_chan) && ast_channel_name(chan) && !strcmp(ast_channel_name(chan), ast_channel_name(moderator_chan))) {
		is_admin = 1;
	} else {
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
			if (!strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminate_str)) {
				is_admin = user->is_admin;
				break;
			}
		}
		ao2_unlock(conference);
	}

	if (host_exist) {
		pbx_builtin_setvar_helper(chan, "HOST_EXIST", "yes");
	}

	return is_admin;
}

/*This function is only for wave*/
static int cache_conf_update_moderator(const char *conf_number, const char *extension, struct ast_channel *chan)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(extension) || !chan) {
		return -1;
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
		if (!strcmp(user->channel_name, ast_channel_name(chan))) {
			user->is_admin = 1;
		} else if (user->is_admin) {
			user->is_admin = 0;
		}
	}

	ao2_cleanup(conference->moderator);
	conference->moderator = ao2_bump(chan);
	ao2_unlock(conference);

	return 0;
}

static int confbridge_sfu_helper(struct ast_channel *chan, const char *data)
{
	int res = 0;
	char *parse = NULL;
	char extension[AST_MAX_EXTENSION] = {0};

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(command);
		AST_APP_ARG(dataval);
	);
	RAII_VAR(struct ast_channel *, moderator, NULL, ao2_cleanup);
	RAII_VAR(struct cache_conference *, cache_conf, NULL, ao2_cleanup);
	SCOPED_MUTEX(lock, &cache_conference_list.lock);

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_sfu);
		res = -1;
		goto end;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		res = -1;
		goto end;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.command)) {
		res = -1;
		goto end;
	}

	conf_get_exten_by_channel(ast_channel_name(chan), extension, sizeof(extension)-1);
	if (!strncasecmp(extension, "trunk", 5)) {
		ast_copy_string(extension, ast_channel_caller_num(chan), sizeof(extension));
	}
	if (ast_strlen_zero(extension)) {
		ast_copy_string(extension, S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "Anonymous"), sizeof(extension));
	}

	if (!strcasecmp(args.command, "AddAdmin")) {
		int inuse_notify = 0;
		const char *wait_reservation_id = pbx_builtin_getvar_helper(chan, "WAIT_RESERVATION_ID");
		if (!ast_strlen_zero(wait_reservation_id) && schedule_conference_need_wait(args.conf, wait_reservation_id, &inuse_notify)) {

			RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
			struct cache_authorized_user *user = NULL;
			int has_moderator = 0;
			enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
			const char * terminate_str = NULL;

			terminate_type = ast_channel_get_terminate_type(chan);
			if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
				terminate_str = "webrtc";
			} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
				terminate_str = "wave_app";
			} else {
				terminate_str = "other_type";
			}

			conference = cache_conf_fetch_add(args.conf, CONF_TYPE_VIDEO, 1, wait_reservation_id);
			if (!conference) {
				ast_log(LOG_WARNING, "Failed to fetch the conference '%s'\n", args.conf);
				res = -1;
				goto end;
			}

			ao2_lock(conference);
			AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
				if (user->is_admin && !strcmp(user->user_schedule_id, wait_reservation_id)) {
					has_moderator = 1;
					res = -1;
					ast_log(LOG_DEBUG, "The moderator has exist, user->extension: %s\n", user->extension);
					if (user->extension && !strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminate_str) && !strcmp(user->user_schedule_id, wait_reservation_id)) {
						ast_log(LOG_DEBUG, "The moderator has exist, but extension: %s and user->extension: %s is the same one\n",
							extension, user->extension);
						res = 0;
					}
					break;
				}
			}
			if (!has_moderator) {
				user = cache_conf_authorized_user_find(conference, extension, terminate_str);
				if (!user) {
					if (cache_conf_authorized_user_add(conference, extension, 1, terminate_str, wait_reservation_id, NULL) < 0) {
						ast_log(LOG_WARNING, "Failed to authorize '%s' in conference '%s'\n", extension, args.conf);
						res = -1;
					}
				} else {
					ast_log(LOG_DEBUG, "User '%s' has already been in cache user.\n", extension);
					cache_conf_authorized_user_update(user, 1, wait_reservation_id);
					ao2_cleanup(user);
				}
			}
			ao2_unlock(conference);
		} else {
			moderator = ast_conf_get_sfu_moderator(chan, args.conf);
			if (moderator) {
				if (ast_channel_name(moderator) && ast_channel_name(chan) && !strcasecmp(ast_channel_name(moderator), ast_channel_name(chan))) {
					res = 0;
				} else {
					ast_log(LOG_DEBUG, "Admin(%s) has existed, couldn't add %s as the new admin",
						ast_channel_name(moderator), ast_channel_name(chan));
					res = -1;
				}
			} else {
				if (!(cache_conf = cache_conf_fetch_add(args.conf, CONF_TYPE_VIDEO, 1, args.dataval))) {
					res = -1;
				} else {
					ao2_lock(cache_conf);
					if (!cache_conf->moderator) {
						cache_conf->moderator = ao2_bump(chan);
						res = 0;
					} else {
						ast_log(LOG_DEBUG, "Admin(%s) has existed, couldn't add %s as the new admin",
							ast_channel_name(cache_conf->moderator), ast_channel_name(chan));
						res = -1;
					}
					ao2_unlock(cache_conf);
				}
			}
		}
	} else if (!strcasecmp(args.command, "RemoveAdmin")) {
		cache_conf_remove_moderator(args.conf, chan);
		res = 0;
	} else if (!strcasecmp(args.command, "Authenticate")) {
		char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
		/*check if it is gswave*/
		if (ast_channel_get_terminate_type(chan) == AST_TERMINAL_TYPE_WEBRTC || AST_TERMINAL_TYPE_GSWAVE_MOBILE == ast_channel_get_terminate_type(chan)) {
			res = cache_conf_check_user_authorization(args.conf, extension, schedule_id);
			/*join by the link need set JOIN_EARLY*/
			ast_log(LOG_DEBUG, "get cache schedule_id %s!", schedule_id);
			if (!ast_strlen_zero(schedule_id)) {
				struct schedule_conference_wait_info conf_sch_w = { 0 };
				conf_sch_w.conf_name = args.conf;
				RAII_VAR(struct schedule_conference_wait*, conference_obj, NULL, ao2_cleanup);
				//conference_obj = schedule_conf_wait_fetch_add(args.conf, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
				conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
				if (conference_obj) {
					ao2_lock(conference_obj);
					if (!strcmp(schedule_id, conference_obj->schedule_id)) {
						ast_log(LOG_DEBUG, "Authenticate set JOIN_EARLY!");
						pbx_builtin_setvar_helper(chan, "JOIN_EARLY", "yes");
					} else {
						ast_log(LOG_DEBUG, "Authenticate set JOIN_CURRENT!");
						pbx_builtin_setvar_helper(chan, "JOIN_CURRENT", "yes");
					}
					ao2_unlock(conference_obj);
				}
			}
		} else {
			res = -1;
		}
		if (res == -1) {
			const char *from_call_flip =  NULL;
			from_call_flip = pbx_builtin_getvar_helper(chan, "FROM_CALL_FLIP");
			if (!ast_strlen_zero(from_call_flip) && !strcmp(from_call_flip, "yes")) {
				ast_log(LOG_DEBUG, "the call from call flip!");
				res = 0;
			}
		}
	} else {
		ast_log(LOG_WARNING, "%s(%s,%s) cannot be executed. Invalid type.\n",
			app_sfu, args.conf, args.command);
		res = -1;
	}

end:
	if (!res) {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_SFU_HELPER_RESULT", "SUCCESS");
	} else {
		pbx_builtin_setvar_helper(chan, "CONFBRIDGE_SFU_HELPER_RESULT", "FAILED");
	}
	return 0;
}

static int func_confbridge_sfu_info(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *parse = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(type);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.type)) {
		return -1;
	}
	terminate_type = ast_channel_get_terminate_type(chan);

	if (!strcasecmp(args.type, "role")) {
		int role = 0;
		role = cache_conf_check_user_role(args.conf, chan, S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, ""), terminate_type);
		if (role < 0) {
			ast_log(LOG_WARNING, "Couldn't find the role of user %s.\n", ast_channel_name(chan));
		} else if (role > 0) {
			strncpy(buf, "admin", len);
		} else {
			strncpy(buf, "nonadmin", len);
		}
	} else {
		ast_log(LOG_WARNING, "Invalid keyword '%s' passed to CONFBRIDGE_SFU_INFO.\n", args.type);
	}

	return 0;
}

static struct ast_custom_function confbridge_sfu_info_function = {
	.name = "CONFBRIDGE_SFU_INFO",
	.read = func_confbridge_sfu_info,
};

static int func_confbridge_im_info(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
	int strong_related = 0;
	RAII_VAR(struct im_conference *, conference, NULL, ao2_cleanup);
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(type);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		res = -1;
		goto end;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.type)) {
		res = -1;
		goto end;
	}

	if (!strcasecmp(args.type, "groupid")) {
		conference = im_conf_list_fetch_add(args.conf, NULL, 0, CONF_TYPE_NONE, 0, NULL);
		if (!conference) {
			ast_log(LOG_DEBUG, "CONFBRIDGE_IM_INFO can't find conference!");
			res = -1;
			goto end;
		}
		ao2_lock(conference);
		strncpy(buf, conference->im_group_id, len);
		strong_related = conference->strong_related;
		/*Modify for Bug 190402*/
		if (conference->type == CONF_TYPE_VIDEO)
		{
			pbx_builtin_setvar_helper(chan, "IS_SFU",  "yes");
		}
		ao2_unlock(conference);
	} else {
		res = -1;
		ast_log(LOG_ERROR, "Invalid keyword '%s' passed to CONFBRIDGE_IM_INFO.\n", args.type);
	}

end:
	if (!res) {
		pbx_builtin_setvar_helper(chan, "STRONG_RELATED_CHAT", strong_related ? "yes" : "no");
	}

	return res;
}

static struct ast_custom_function confbridge_im_conf_function = {
	.name = "CONFBRIDGE_IM_INFO",
	.read = func_confbridge_im_info,
};

static int func_confbridge_early_join_info(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = 0;
	char *parse = NULL;
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	RAII_VAR(struct schedule_conference_wait *, conference, NULL, ao2_cleanup);
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(type);
		AST_APP_ARG(datastr);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		res = -1;
		goto end;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.type)) {
		res = -1;
		goto end;
	}
	//ast_log(LOG_DEBUG, "args.type %s, args.datastr %s!", args.type, args.datastr);

	conf_sch_w.conf_name = args.conf;
	//conference = schedule_conf_wait_fetch_add(args.conf, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conference = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference) {
		ast_log(LOG_DEBUG, "CONFBRIDGE_EARLY_JOIN can't find conference!");
		res = -1;
		goto end;
	}
	ao2_lock(conference);
	if (!strcasecmp(args.type, "password") && args.datastr) {
		if (!strcasecmp(conference->admin_pin, args.datastr)) {
			ast_log(LOG_DEBUG, "admin user!");
			strncpy(buf, "admin", len);
			pbx_builtin_setvar_helper(chan, "WAIT_IS_ADMIN", "yes");
		} else if (!strcasecmp(conference->pin, args.datastr)) {
			ast_log(LOG_DEBUG, "participant user!");
			strncpy(buf, "participant", len);
			pbx_builtin_setvar_helper(chan, "WAIT_IS_ADMIN", "no");
		} else {
			ast_log(LOG_DEBUG, "password not match!");
			strncpy(buf, "failed", len);
		}
	} else if (!strcasecmp(args.type, "bookid")) {
		strncpy(buf, conference->schedule_id, len);
		pbx_builtin_setvar_helper(chan, "WAIT_RESERVATION_ID", conference->schedule_id);
	} else if (!strcasecmp(args.type, "earlyinfo")) {
		if (!ast_strlen_zero(conference->schedule_id)) {
			strncpy(buf, "yes", len);
		}
	} else {
		res = -1;
		ast_log(LOG_ERROR, "Invalid keyword '%s' passed to CONFBRIDGE_EARLY_JOIN.\n", args.type);
	}
	ao2_unlock(conference);

end:

	return res;
}

static struct ast_custom_function confbridge_early_join_function = {
	.name = "CONFBRIDGE_EARLY_JOIN",
	.read = func_confbridge_early_join_info,
};

static int func_confbridge_user_terminal_type(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *parse = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(type);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.type)) {
		return -1;
	}

	if (!strcasecmp(args.type, "terminal")) {
		terminate_type = ast_channel_get_terminate_type(chan);

		if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			strncpy(buf, "waveapp", len);
		} else if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
			strncpy(buf, "webrtc", len);
		} else {
			strncpy(buf, "others", len);
		}
	} else {
		ast_log(LOG_ERROR, "Invalid keyword '%s' passed to CONFBRIDGE_USER_TERMINAL_TYPE.\n", args.type);
	}

	return 0;
}

static struct ast_custom_function confbridge_user_terminal_type_function = {
	.name = "CONFBRIDGE_USER_TERMINAL_TYPE",
	.read = func_confbridge_user_terminal_type,
};

static int conf_schedule_get_user_info(const char *reservation_id, const char *if_sfu, const char *extension)
{
	int ret = -1;
	char cmd[512]={0};
	struct odbc_obj *conf_obj = NULL;

	if (ast_strlen_zero(reservation_id) || ast_strlen_zero(if_sfu) || ast_strlen_zero(extension)) {
		return ret;
	}

	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	if (!strcasecmp(if_sfu,"yes")) {
		snprintf(cmd, sizeof(cmd), "SELECT is_admin FROM %s.multimedia_conference_reservation_members WHERE reservation_id='%s' AND member_extension='%s'", ast_strlen_zero(s_dbConfigName) ? "ucm_config" : s_dbConfigName, reservation_id, extension);
	} else {
		snprintf(cmd, sizeof(cmd), "SELECT is_admin FROM %s.meetme_members WHERE bookid='%s' AND member_extension='%s'", ast_strlen_zero(s_dbConfigName) ? "ucm_config" : s_dbConfigName, reservation_id, extension);
	}

	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256] = {0};
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			//ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			if (!strcasecmp(tmp, "yes")) {
				ret = 1;
			} else {
				ret = 0;
			}
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	//ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return ret;
}

static int conf_chan_if_same_endpoint(struct ast_channel *chan1, struct ast_channel *chan2)
{
	int ret = 0;
	int chan1_len = 0;
	int chan2_len = 0;
	const char *start = NULL;
	const char *end = NULL;
	const char *chan1_name = NULL;
	const char *chan2_name = NULL;

	if (!chan1 || !chan2) {
		return ret;
	}

	chan1_name = ast_channel_name(chan1);
	chan2_name = ast_channel_name(chan2);

	if (!chan1_name || !chan2_name) {
		return ret;
	}

	start = chan1_name;
	end = strrchr(start, '-');
	if (!end) {
		return ret;
	}
	chan1_len = end - start;

	start = chan2_name;
	end = strrchr(start, '-');
	if (!end) {
		return ret;
	}
	chan2_len = end - start;

	if (chan1_len && chan2_len && chan1_len == chan2_len) {
		if (strncmp(chan1_name, chan2_name, chan2_len) == 0) {
			ret = 1;
			ast_log(LOG_DEBUG, "conf chan is the same endpoint chan1 %s VS chan2 %s\n", chan1_name, chan2_name);
		}
	}

	return ret;
}

static int confbridge_if_admin_exist(struct ast_channel *chan, const char *reservation_id, const char *conf_name, const char *extension)
{
	int res = 0;

	if (!chan || ast_strlen_zero(reservation_id) || ast_strlen_zero(conf_name) || ast_strlen_zero(extension)) {
		res = -1;
		ast_log(LOG_WARNING, "parament null\n");
		return res;
	}

	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	struct cache_authorized_user *user = NULL;
	const char * terminate_str = NULL;

	terminate_type = ast_channel_get_terminate_type(chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	conference = cache_conf_fetch_add(conf_name, CONF_TYPE_VIDEO, 1, reservation_id);
	if (!conference) {
		ast_log(LOG_WARNING, "Failed to fetch the conference '%s'\n", conf_name);
		res = -1;
		return res;
	}

	ao2_lock(conference);
	if (conference->moderator) {
		if (ast_channel_name(chan) && ast_channel_name(conference->moderator) && !strcmp(ast_channel_name(chan), ast_channel_name(conference->moderator))) {
			ast_log(LOG_DEBUG, "The moderator has exist, but chan is the same one: %s\n", ast_channel_name(conference->moderator));
			res = 0;
		} else {
			ast_log(LOG_DEBUG, "The moderator has exist, moderator: %s\n", ast_channel_name(conference->moderator));
			if (conf_chan_if_same_endpoint(chan, conference->moderator)) {
				pbx_builtin_setvar_helper(chan, "WANT_BE_HOST", "yes");
			}
			res = 1;
		}
		if (res && strcmp(conference->schedule_id, reservation_id)) {
			ast_log(LOG_DEBUG, "The moderator has exist, but not the same reservation: [%s] VS [%s]\n", conference->schedule_id, reservation_id);
			/*The last cache info may not have been cleared*/
			AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
				if (user->is_admin && !strcmp(user->user_schedule_id, reservation_id)) {
					res = 1;
					ast_log(LOG_DEBUG, "The moderator has exist, user->extension: %s\n", user->extension);
					if (user->extension && !strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminate_str) && !strcmp(user->user_schedule_id, reservation_id)) {
						ast_log(LOG_DEBUG, "The moderator has exist, but extension: %s and user->extension: %s is the same one\n",
							extension, user->extension);
						res = 0;
					}
					break;
				}
			}
		}
	} else {
		AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
			if (user->is_admin && !strcmp(user->user_schedule_id, reservation_id)) {
				res = 1;
				ast_log(LOG_DEBUG, "The moderator has exist, user->extension: %s\n", user->extension);
				if (user->extension && !strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminate_str) && !strcmp(user->user_schedule_id, reservation_id)) {
					ast_log(LOG_DEBUG, "The moderator has exist, but extension: %s and user->extension: %s is the same one\n",
						extension, user->extension);
					res = 0;
				}
				break;
			}
		}
	}
	ao2_unlock(conference);
	/*moderator transfor to phone,cache list moderator will be null. Bug 199575*/
	if (!res) {
		RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
		struct confbridge_user *conf_user = NULL;

		if ((conference = ao2_find(conference_bridges, conf_name, OBJ_KEY))) {
			ao2_lock(conference);
			AST_LIST_TRAVERSE(&conference->active_list, conf_user, list) {
				if (ast_test_flag(&conf_user->u_profile, USER_OPT_ADMIN)) {
					if (conf_user->chan && ast_channel_name(chan) && ast_channel_name(conf_user->chan) && strcmp(ast_channel_name(chan), ast_channel_name(conf_user->chan))) {
						if (!ast_strlen_zero(conference->schedule_id_processing) && !strcmp(conference->schedule_id_processing, reservation_id)) {
							ast_log(LOG_DEBUG, "The moderator has exist, host chan %s\n", ast_channel_name(conf_user->chan));
							res = 1;
						}
					}
					break;
				}
			}
			ao2_unlock(conference);
		}
	}
	return res;
}

static int func_confbridge_schedule_check_role(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *parse = NULL;
	char extension[AST_MAX_EXTENSION] = {0};
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(type);
		AST_APP_ARG(conf);
		AST_APP_ARG(dataval);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.type) || ast_strlen_zero(args.conf) || ast_strlen_zero(args.dataval)) {
		return -1;
	}

	conf_get_exten_by_channel(ast_channel_name(chan), extension, sizeof(extension)-1);
	if (!strncasecmp(extension, "trunk", 5)) {
		ast_copy_string(extension, ast_channel_caller_num(chan), sizeof(extension));
	}
	if (ast_strlen_zero(extension)) {
		ast_copy_string(extension, S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "Anonymous"), sizeof(extension));
	}

	if (!strcasecmp(args.type, "role")) {
		const char *is_sfu = pbx_builtin_getvar_helper(chan, "IS_SFU");
		int role_val = conf_schedule_get_user_info(args.dataval, is_sfu ? is_sfu : "no", extension);
		if (1 == role_val) {
			if (is_sfu && !strcmp(is_sfu, "yes") && confbridge_if_admin_exist(chan, args.dataval, args.conf, extension)) {
				strncpy(buf, "participant", len);
				pbx_builtin_setvar_helper(chan, "ADMIN_EXIST", "yes");
			} else {
				strncpy(buf, "admin", len);
			}
			if (ast_channel_name(chan)) {
				ast_log(LOG_DEBUG, "%s want be host!", ast_channel_name(chan));
				pbx_builtin_setvar_helper(chan, "WANT_BE_HOST", "yes");
			}
		} else if (0 == role_val) {
			/*Bug 241850*/
			const char *nway_conf_admin = NULL;
			nway_conf_admin = pbx_builtin_getvar_helper(chan, "NWAY_CONF_ADMIN");
			if (!ast_strlen_zero(nway_conf_admin) && !strcmp(nway_conf_admin, "yes")) {
				strncpy(buf, "admin", len);
			} else {
				strncpy(buf, "participant", len);
			}
		} else {
			strncpy(buf, "unknow", len);
		}
	} else {
		ast_log(LOG_ERROR, "Invalid keyword '%s' passed to CONFBRIDGE_SCHEDULE_CHECK_ROLE.\n", args.type);
	}

	return 0;
}

static struct ast_custom_function confbridge_schedule_check_role_function = {
	.name = "CONFBRIDGE_SCHEDULE_CHECK_ROLE",
	.read = func_confbridge_schedule_check_role,
};

static int func_confbridge_user_related_info(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	char *parse = NULL;
	const char * terminate_str = NULL;
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	RAII_VAR(struct cache_authorized_user *, cache_user, NULL, ao2_cleanup);
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(exten);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.exten)) {
		return -1;
	}

	terminate_type = ast_channel_get_terminate_type(chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	conference = cache_conf_fetch_add(args.conf, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}
	ao2_lock(conference);

	cache_user = cache_conf_authorized_user_find(conference, args.exten, terminate_str);
	if (!cache_user) {
		ao2_unlock(conference);
		return -1;
	}
	if (!ast_strlen_zero(cache_user->related_info)) {
		strncpy(buf, cache_user->related_info, len);
	}

	ao2_unlock(conference);

	return 0;
}

static struct ast_custom_function confbridge_user_related_info_function = {
	.name = "CONFBRIDGE_USER_RELATED_INFO",
	.read = func_confbridge_user_related_info,
};


#if 0
static char *complete_cache_conf_name(const char *line, const char *word, int pos, int state)
{
	int which = 0;
	struct cache_conference *conference = NULL;
	char *res = NULL;
	int wordlen = strlen(word);

	AST_LIST_LOCK(&cache_conference_list);
	AST_LIST_TRAVERSE(&cache_conference_list, conference, entry) {
		if (!strncasecmp(conference->name, word, wordlen) && ++which > state) {
			res = ast_strdup(conference->name);
			break;
		}
	}
	AST_LIST_UNLOCK(&cache_conference_list);

	return res;
}

static char *complete_cache_conf_user(const char *conference_name, const char *line, const char *word, int pos, int state)
{
	int which = 0;
	char *res = NULL;
	int wordlen = strlen(word);
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;

	conference = cache_conf_fetch_add(conference_name, CONF_TYPE_NONE, 0);
	if (!conference) {
		return NULL;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
		if (!strncasecmp(user->extension, word, wordlen) && ++which > state) {
			res = ast_strdup(user->extension);
			break;
		}
	}
	ao2_unlock(conference);

	return res;
}
#endif

static int reload_menu_to_user(struct confbridge_user *user, const char *menu_profile_name)
{
	int ret = 0;

	conf_unset_menu_from_user(NULL, user);
	ret = conf_set_menu_to_user(NULL, user, menu_profile_name);

	return ret;
}

static int __sfu_cache_authorize_user(const char *conf_number, int type, const char *extension, int *is_admin, const char *terminal_type, const char *schedule_id, const char *related_info)
{
	RAII_VAR(struct ast_channel *, moderator_chan, NULL, ao2_cleanup);
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;
	int has_moderator = 0;
	int no_change = 0;

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(extension)) {
		return -1;
	}

	moderator_chan = ast_conf_get_sfu_moderator(NULL, conf_number);
	if (moderator_chan) {
		has_moderator = 1;
	}

	conference = cache_conf_fetch_add(conf_number, type, 1, schedule_id);
	if (!conference) {
		ast_log(LOG_WARNING, "Failed to fetch the conference '%s'\n", conf_number);
		return -1;
	}
	/*schedule wait users schedule id will be differenct*/
	if (has_moderator && !ast_strlen_zero(schedule_id) && strcasecmp(conference->schedule_id, schedule_id)) {
		ast_log(LOG_DEBUG, "may be schedule waiting schedule_id '%s'\n", schedule_id);
		has_moderator = 0;
	}
	ao2_lock(conference);

	if (!has_moderator) {
		AST_LIST_TRAVERSE(&conference->auth_user_list, user, entry) {
			if (user->is_admin && !strcmp(user->user_schedule_id, schedule_id)) {
				has_moderator = 1;
				if (*is_admin && user->extension && !strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminal_type)) {
					ast_log(LOG_DEBUG, "The moderator has exist, but extension: %s and user->extension: %s is the same one,Ignore update\n",
						extension, user->extension);
					no_change = 1;
				}
				break;
			}
		}
	}

	/*update the role*/
	if (*is_admin && has_moderator && !no_change)
	{
		ast_log(LOG_DEBUG, "The moderator has exist, so the role of extension %s will be changed from admin to participant.\n",
			extension);
		*is_admin = 0;
	}

	user = cache_conf_authorized_user_find(conference, extension, terminal_type);
	if (!user) {
		if (cache_conf_authorized_user_add(conference, extension, *is_admin, terminal_type, schedule_id, related_info) < 0) {
			ast_log(LOG_WARNING, "Failed to authorize '%s' in conference '%s'\n", extension, conf_number);
			ao2_unlock(conference);
			return -1;
		}
	} else {
		ast_log(LOG_DEBUG, "User '%s' has already been in cache user.\n", extension);
		cache_conf_authorized_user_update(user, *is_admin, schedule_id);
		ao2_cleanup(user);
	}
	ao2_unlock(conference);

	return 0;
}

static int __audio_conf_cache_authorize_user(const char *conf_number, int type, const char *extension, int *is_admin, const char *terminal_type, const char *schedule_id)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	struct cache_authorized_user *user = NULL;

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(extension)) {
		return -1;
	}

	conference = cache_conf_fetch_add(conf_number, type, 1, schedule_id);
	if (!conference) {
		ast_log(LOG_WARNING, "Failed to fetch the conference '%s'\n", conf_number);
		return -1;
	}

	ao2_lock(conference);

	user = cache_conf_authorized_user_find(conference, extension, terminal_type);
	if (!user) {
		if (cache_conf_authorized_user_add(conference, extension, *is_admin, terminal_type, schedule_id, NULL) < 0) {
			ast_log(LOG_WARNING, "Failed to authorize '%s' in conference '%s'\n", extension, conf_number);
			ao2_unlock(conference);
			return -1;
		}
	} else {
		ast_log(LOG_DEBUG, "User '%s' has already been in cache user.\n", extension);
		cache_conf_authorized_user_update(user, *is_admin, schedule_id);
		ao2_cleanup(user);
	}
	ao2_unlock(conference);

	return 0;
}

static int __cache_authorize_user(const char *conf_number, int type, const char *extension, int *is_admin, const char *terminal_type, const char *schedule_id, const char *related_info)
{
	int ret = 0;
	AST_MUTEX_DEFINE_STATIC(_sync_lock);

	SCOPED_MUTEX(lock, &_sync_lock);
	if (type == CONF_TYPE_VIDEO) {
		ret = __sfu_cache_authorize_user(conf_number, type, extension, is_admin, terminal_type, schedule_id, related_info);
	} else {
		ret = __audio_conf_cache_authorize_user(conf_number, type, extension, is_admin, terminal_type, schedule_id);
	}

	return ret;
}

static int action_confbridge_sfu_authorize_user(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *extension_name = astman_get_header(m, "Extension");
	const char *role = astman_get_header(m, "Role");
	const char *terminal_type = astman_get_header(m, "Terminal");
	const char *schedule_id = astman_get_header(m, "ScheduleId");
	const char *related_info = astman_get_header(m, "RelatedInfo");
	int is_admin = 0;
	int last_is_admin = 0;
	int res = 0;
	char msg[128] = {0};

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(extension_name)) {
		astman_send_error(s, m, "No extension name provided.");
		return 0;
	}
	if (ast_strlen_zero(role)) {
		astman_send_error(s, m, "No role provided.");
		return 0;
	}

	is_admin = !strcasecmp(role, "admin") ? 1 : 0;
	last_is_admin = is_admin;
	res = __cache_authorize_user(conference_name, CONF_TYPE_VIDEO, extension_name, &is_admin, terminal_type, schedule_id, related_info);
	if (res < 0) {
		astman_send_error(s, m, "Fail to authorize user.");
		return 0;
	}

	if (is_admin) {
		ast_copy_string(msg, "Role is admin.", sizeof(msg));
	} else {
		ast_copy_string(msg, "Role is participant.", sizeof(msg));
	}

	if (last_is_admin && !is_admin) {
		ast_copy_string(msg + strlen(msg), "Host exists", sizeof(msg) - strlen(msg));
	}
	astman_send_ack(s, m, msg);
	return 0;
}

static unsigned int stream_has_video_resource(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_SENDONLY ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_INACTIVE) {
		return 0;
	}

	return 1;
}

static char *get_msid(struct ast_channel *chan, enum ast_media_sub_type type)
{
	struct ast_stream_topology *topology = NULL;
	int i = 0;
	char msid[(AST_UUID_STR_LEN * 2) + 24] = {0};
	char *p_msid = NULL;

	if (strcasecmp(ast_channel_tech(chan)->type, "PJSIP")) {
		return NULL;
	}

	topology = ast_channel_get_stream_topology(chan);
	if (!topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		struct ast_stream *stream = ast_stream_topology_get_stream(topology, i);
		if (ast_stream_get_sub_type(stream) == type
			&& stream_has_video_resource(stream)
			&& strncmp(ast_stream_get_name(stream), SOFTBRIDGE_VIDEO_DEST_PREFIX,
				SOFTBRIDGE_VIDEO_DEST_LEN)) {
			snprintf(msid, sizeof(msid), "%s %s", ast_stream_get_metadata(stream, "MSID:MSLABEL"), ast_stream_get_metadata(stream, "MSID:LABEL"));
			p_msid = ast_strdup(msid);
			ast_log(LOG_DEBUG, "The host's msid is '%s'", msid);
			break;
		}
	}

	return p_msid;
}

static int sfu_topologies_on_moderator_change(struct confbridge_conference *conference, struct confbridge_user *moderator_user)
{
	struct confbridge_user *cur_user = NULL;
	RAII_VAR(char *, msid, NULL, ast_free);

	if (!conference) {
		return -1;
	}

	if (moderator_user) {
		msid = get_msid(moderator_user->chan, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		if (!msid) {
			msid = ast_strdup("");
		}
	} else {
		msid = ast_strdup("");
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if (ast_channel_get_terminate_type(cur_user->chan) == AST_TERMINAL_TYPE_SIP
			|| ast_channel_get_terminate_type(cur_user->chan) == AST_TERMINAL_TYPE_GVC
			|| ((ast_channel_get_terminate_type(cur_user->chan) == AST_TERMINAL_TYPE_GSWAVE_MOBILE) &&
				!ast_channel_get_mobile_bundle_enabled(cur_user->chan))
			|| ast_channel_get_terminate_type(cur_user->chan) == AST_TERMINAL_TYPE_ICE_TRUNK) {
			struct ast_stream_topology *topology = ast_channel_get_stream_topology(cur_user->chan);
			if (topology) {
				int i = 0;
				for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
					struct ast_stream *stream = ast_stream_topology_get_stream(topology, i);
					/*Check if it supports video.*/
					if (ast_stream_get_state(stream) != AST_STREAM_STATE_REMOVED
						&& ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO
						&& ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN
						&& strncmp(ast_stream_get_name(stream), SOFTBRIDGE_VIDEO_DEST_PREFIX,
							SOFTBRIDGE_VIDEO_DEST_LEN)) {
						ast_channel_lock(cur_user->chan);
						if (ast_channel_tech(cur_user->chan)->indicate) {
							ast_channel_tech(cur_user->chan)->indicate(cur_user->chan, AST_CONTROL_CONF_HOST_CHANGED, msid, strlen(msid) + 1);
						}
						ast_channel_unlock(cur_user->chan);
						break;
					}
				}
			}
		}
	}
	ao2_unlock(conference);

	return 0;
}

static int __transfer_moderator(struct mansession *s, const struct message *m, struct confbridge_conference *conference, const char *channel)
{
	struct confbridge_user *user = NULL;
	struct confbridge_user *admin_user = NULL;
	struct confbridge_user *old_admin_user = NULL;
	int transferred = 0;
	int can_invite = 0;
	const char *originator = astman_get_header(m, "Originator");

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
			if (!ast_strlen_zero(originator)) {
				if (!strcasecmp(ast_channel_name(user->chan), originator)) {
					old_admin_user = user;
				}
			} else {
				old_admin_user = user;
			}
		}

		if (!transferred && !strcasecmp(ast_channel_name(user->chan), channel)) {
			transferred = 1;
			can_invite = ast_test_flag(&user->u_profile, USER_OPT_CAN_INVITE);
			admin_user = user;
		}
	}

	if (transferred) {
		if (old_admin_user == admin_user) {
			ao2_unlock(conference);
			astman_send_ack(s, m, "");
			return 0;
		}

		if (old_admin_user) {
			ast_clear_flag(&old_admin_user->u_profile, USER_OPT_ADMIN);
			ast_clear_flag(&old_admin_user->u_profile, USER_OPT_MARKEDUSER);
			conference->markedusers--;
			if (!can_invite) {
				ast_clear_flag(&old_admin_user->u_profile, USER_OPT_CAN_INVITE);
			}
			/*if muteall,then we should mute the old host.Bug 176453*/
			if (conference->muted && !old_admin_user->muted) {
				old_admin_user->muted = conference->muted;
				conf_update_user_mute(old_admin_user);
				send_mute_event(old_admin_user, conference, 1, 0);
			}
			reload_menu_to_user(old_admin_user, MENU_SFU_USER);
			send_role_update_event(old_admin_user, conference);
			cache_conf_clear_moderator(old_admin_user->conference->name, old_admin_user->chan);
			/*let old admin cancel Notify*/
			if (conference->conf_will_timeout && conference->extend_time_status == EXTEND_PENDING
				&& (admin_user->term_type == AST_TERMINAL_TYPE_WEBRTC || admin_user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
				ast_log(LOG_DEBUG, "let old admin cancel Notify");
				conf_timeout_notify_build_send(old_admin_user->chan, conference->name, 1, EXTEND_CONFIRMED, 0);
			}
		}

		if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
			ao2_unlock(conference);
			astman_send_ack(s, m, "");
			return 0;
		}

		if (admin_user->muted) {
			admin_user->muted = 0;
			conf_update_user_mute(admin_user);
			send_unmute_event(admin_user, conference, 1, 0);
		}
		ast_set_flag(&admin_user->u_profile, USER_OPT_ADMIN);
		ast_set_flag(&admin_user->u_profile, USER_OPT_MARKEDUSER);
		conference->markedusers++;
		reload_menu_to_user(admin_user, MENU_SFU_ADMIN);
		send_role_update_event(admin_user, conference);

		if (conference->state->state_to_marked) {
			conference->state->state_to_marked(admin_user);
		}
		if (admin_user->term_type == AST_TERMINAL_TYPE_WEBRTC || admin_user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			cache_conf_update_moderator(conference->name,
										S_COR(ast_channel_caller(admin_user->chan)->id.number.valid, ast_channel_caller(admin_user->chan)->id.number.str, ""),
										admin_user->chan);
			/*send Notify to new admin*/
			if (conference->conf_will_timeout && conference->extend_time_status == EXTEND_PENDING) {
				int conf_timeleft = ast_sched_when(conf_sched, conference->conf_sched_id);
				ast_log(LOG_DEBUG, "send Notify to new admin!");
				conf_timeout_notify_build_send(admin_user->chan, conference->name, 1, conference->extend_time_status, conf_timeleft+(conference->sch_time * (CONF_PEER_TIMEOUT/1000)));
			}
		}
		//send_moderator_event(admin_user, conference);

		user = find_active_user_by_channel(conference, astman_get_header(m, "Originator"));
		if (user) {
			conf_user_write_operator_log(user, CONF_OP_TRANSFER_HOST, S_COR(ast_channel_caller(admin_user->chan)->id.number.valid, ast_channel_caller(admin_user->chan)->id.number.str, ""));
		}
	}
	ao2_unlock(conference);

	if (transferred) {
		astman_send_ack(s, m, "");
		if (!conference->focus_chan) {
			sfu_topologies_on_moderator_change(conference, admin_user);
		} else {
			ast_log(LOG_DEBUG, "__transfer_moderator already has focus chan, don't upate host msid\n");
		}
	} else {
		astman_send_error(s, m, "No participant.");
	}
	return 0;
}

static int __retrieve_moderator(struct mansession *s, const struct message *m, struct confbridge_conference *conference, const char *channel)
{
	struct confbridge_user *user = NULL;
	struct confbridge_user *admin_user = NULL;
	char msg[256] = {0};

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
			if (strcasecmp(ast_channel_name(user->chan), channel)) {
				snprintf(msg, sizeof(msg), "Moderator Exists: %s",
					S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, ""));
				astman_send_ack(s, m, msg);
				ao2_unlock(conference);
				return 0;
			} else {
				astman_send_ack(s, m, "");
				conf_user_write_operator_log(user, CONF_OP_RETRIEVE_HOST, NULL);
				ao2_unlock(conference);
				return 0;
			}
		} else if (!strcasecmp(ast_channel_name(user->chan), channel)) {
			admin_user = user;
		}
	}

	if (admin_user) {
		admin_user->muted = 0;
		ast_set_flag(&admin_user->u_profile, USER_OPT_ADMIN);
		ast_set_flag(&admin_user->u_profile, USER_OPT_MARKEDUSER);
		conference->markedusers++;
		/*send unmute event.Bug 194629*/
		send_unmute_event(admin_user, conference, 1, 0);
		reload_menu_to_user(admin_user, MENU_SFU_ADMIN);
		if (conference->state->state_to_marked) {
			conference->state->state_to_marked(admin_user);
		}
		send_role_update_event(admin_user, conference);
		if (admin_user->term_type == AST_TERMINAL_TYPE_WEBRTC || admin_user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			cache_conf_update_moderator(conference->name,
										S_COR(ast_channel_caller(admin_user->chan)->id.number.valid, ast_channel_caller(admin_user->chan)->id.number.str, ""),
										admin_user->chan);
			/*send Notify to new admin*/
			if (conference->conf_will_timeout && conference->extend_time_status == EXTEND_PENDING) {
				int conf_timeleft = ast_sched_when(conf_sched, conference->conf_sched_id);
				ast_log(LOG_DEBUG, "send Notify to new admin!");
				conf_timeout_notify_build_send(admin_user->chan, conference->name, 1, conference->extend_time_status, conf_timeleft+(conference->sch_time * (CONF_PEER_TIMEOUT/1000)));
			}
		}
		/*send event to AVS Bug 179322*/
		conf_update_user_mute(admin_user);

		conf_user_write_operator_log(admin_user, CONF_OP_RETRIEVE_HOST, NULL);
	}
	ao2_unlock(conference);

	if (admin_user) {
		//send_moderator_event(admin_user, conference);
		astman_send_ack(s, m, "");
		if (!conference->focus_chan) {
			sfu_topologies_on_moderator_change(conference, admin_user);
		} else {
			ast_log(LOG_DEBUG, "__retrieve_moderator already has focus chan, don't upate host msid\n");
		}
	} else {
		astman_send_error(s, m, "No participant.");
	}

	return 0;
}

static int action_confbridge_sfu_toggle_record(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");

	struct ast_channel *chan = NULL;
	struct ast_bridge_channel *bridge_chan = NULL;
	struct confbridge_user *admin_user = NULL;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	chan = ast_channel_get_by_name(channel_name);
	if (!chan) {
		astman_send_error(s, m, "No Channel found");
		return 0;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, admin_user, list) {
		if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
			break;
		}
	}
	ao2_unlock(conference);

	ast_log(LOG_DEBUG, "admin_user:%p, admin_chan:%s, chan:%s", admin_user,
		(admin_user && admin_user->chan) ? ast_channel_name(admin_user->chan) : "none",
		ast_channel_name(chan));

	if (!admin_user || admin_user->chan != chan) {
		astman_send_error(s, m, "No permission");
		goto END;
	}

	bridge_chan = bridge_find_channel(conference->bridge, chan);

	ast_bridge_suspend(conference->bridge, admin_user->chan);
	action_toggle_record(conference, bridge_chan, admin_user);
	ast_bridge_unsuspend(conference->bridge, admin_user->chan);

END:
	ast_channel_unref(chan);
	return 0;
}

static int judge_if_conference_is_recording(struct confbridge_conference *conference)
{
	if (!conference) {
		return 0;
	}

	return conf_is_recording(conference) || conf_is_videoing(conference);
}

static int toggle_video_record(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	int res = 0;
	int preivous_recording_state = conference->video_is_recording;

	if (judge_if_conference_is_recording(conference) && !conference->video_is_recording) {
		return -1;
	}

	if (conference->video_is_recording) {
		ao2_lock(conference);
		conf_user_write_operator_log(user, CONF_OP_END_RECORD, NULL);
		res = conf_stop_video_record(conference);
		if (!res && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
			play_sound_file(conference, conf_get_sound(CONF_SOUND_VIDEO_OFF, user->b_profile.sounds));
		}
		ao2_unlock(conference);
	} else {
		ao2_lock(conference);
		res = video_record_control_resource_judgment(conference, AVS_CONTROL_START_CONF_VIDEO_RECORD);
		if (!res && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
			play_sound_file(conference, conf_get_sound(CONF_SOUND_START_VIDEO, user->b_profile.sounds));
		}
		res = conf_start_video_record(conference);
		if (!res) {
			conf_user_write_operator_log(user, CONF_OP_START_RECORD, NULL);
		}
		ao2_unlock(conference);
	}

	if (!res) {
		send_record_status_event(conference, "video", !preivous_recording_state);
	}

	return res;
}

static int toggle_foucs_record(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	int res = 0;
	int preivous_recording_state = conference->focus_is_recording;

	if (judge_if_conference_is_recording(conference) && !conference->focus_is_recording) {
		return -1;
	}

	if (conference->focus_is_recording) {
		ao2_lock(conference);
		conf_user_write_operator_log(user, CONF_OP_END_RECORD, NULL);
		res = conf_stop_focus_record(conference);
		if (!res && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
			play_sound_file(conference, conf_get_sound(CONF_SOUND_VIDEO_OFF, user->b_profile.sounds));
		}
		ao2_unlock(conference);
	} else {
		ao2_lock(conference);
		res = video_record_control_resource_judgment(conference, AVS_CONTROL_START_CONF_FOCUS_RECORD);
		if (!res && ast_test_flag(&conference->b_profile, BRIDGE_OPT_RECORD_PROMPT)) {
			play_sound_file(conference, conf_get_sound(CONF_SOUND_START_FOCUS_VIDEO, user->b_profile.sounds));
		}
		res = conf_start_focus_record(conference);
		if (!res) {
			conf_user_write_operator_log(user, CONF_OP_START_RECORD, NULL);
		}
		ao2_unlock(conference);
	}

	if (!res) {
		send_record_status_event(conference, "focus", !preivous_recording_state);
	}

	return res;
}

static int action_confbridge_toggle_auto_record(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");

	struct ast_channel *chan = NULL;
	struct ast_bridge_channel *bridge_chan = NULL;
	struct confbridge_user *admin_user = NULL;
	int res = -1;
	char response[128] = { 0 };
	char meetme_dir[PATH_MAX] = {0};
	char video_dir[PATH_MAX] = {0};
	char ucm_dev[MAX_DEV_NAME] = {0};

	/*Check UCM630XA*/
	conf_get_ucm_dec_info(ucm_dev, MAX_DEV_NAME);

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}

	ast_log(LOG_DEBUG, "action_confbridge_toggle_auto_record conference_name:%s, channel_name:%s\n", conference_name, channel_name);

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	chan = ast_channel_get_by_name(channel_name);
	if (!chan) {
		astman_send_error(s, m, "No Channel found");
		return 0;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, admin_user, list) {
		if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
			break;
		}
	}
	ao2_unlock(conference);

	ast_log(LOG_DEBUG, "action_confbridge_toggle_auto_record admin_user:%p, admin_chan:%s, chan:%s", admin_user,
		(admin_user && admin_user->chan) ? ast_channel_name(admin_user->chan) : "none",
		ast_channel_name(chan));

	if (!admin_user || admin_user->chan != chan) {
		astman_send_error(s, m, "No permission");
		goto END;
	}

	ao2_lock(conference);
	bridge_chan = bridge_find_channel(conference->bridge, chan);
	snprintf(meetme_dir, sizeof(meetme_dir), "%s/meetme", ast_config_AST_SPOOL_DIR);
	snprintf(video_dir, sizeof(video_dir), "%s/video", ast_config_AST_SPOOL_DIR);

	if (strstr(ucm_dev, "A") == NULL) {
#if defined(BUILD_ON_CLOUD_PBX) || defined(HAVE_GCC601X) || defined(HAVE_GCC603X)
		if (!toggle_foucs_record(conference, bridge_chan, admin_user)) {
			res = 0;
			ast_log(LOG_DEBUG, "toggle focus record success, conference_name:%s, channel_name:%s\n",
				conference_name, channel_name);
			snprintf(response, sizeof(response), "Success\r\nRecordType: focus\r\nDeviceType: %s\r\nControlType: %s", getLinkResult(video_dir), conference->focus_is_recording ? "Open" : "Close");
		}
#else
		if (!toggle_video_record(conference, bridge_chan, admin_user)) {
			res = 0;
			ast_log(LOG_DEBUG, "toggle video record success, conference_name:%s, channel_name:%s\n",
				conference_name, channel_name);
			snprintf(response, sizeof(response), "Success\r\nRecordType: video\r\nDeviceType: %s\r\nControlType: %s", getLinkResult(video_dir), conference->video_is_recording ? "Open" : "Close");
		} else if (!toggle_foucs_record(conference, bridge_chan, admin_user)) {
			res = 0;
			ast_log(LOG_DEBUG, "toggle focus record success, conference_name:%s, channel_name:%s\n",
				conference_name, channel_name);
			snprintf(response, sizeof(response), "Success\r\nRecordType: focus\r\nDeviceType: %s\r\nControlType: %s", getLinkResult(video_dir), conference->focus_is_recording ? "Open" : "Close");
		}
#endif
	}

	if (res) {
		ast_bridge_suspend(conference->bridge, admin_user->chan);
		res = action_toggle_record(conference, bridge_chan, admin_user);
		ast_bridge_unsuspend(conference->bridge, admin_user->chan);
		snprintf(response, sizeof(response), "Success\r\nRecordType: audio\r\nDeviceType: %s\r\nControlType: %s", getLinkResult(meetme_dir), (conference->record_chan != NULL && !is_conf_record_paused(conference)) ? "Open" : "Close");
	}
	ao2_unlock(conference);

	if (res) {
		ast_log(LOG_DEBUG, "action_confbridge_toggle_auto_record failed, conference_name:%s, channel_name:%s\n",
			conference_name, channel_name);
		astman_send_error(s, m, "toggle auto record failed");
	} else {
		astman_send_response(s, m, response, NULL);
	}

END:
	ast_channel_unref(chan);
	return 0;
}

static int action_confbridge_sfu_transfer_moderator(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	const char *Mode = astman_get_header(m, "Mode");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}
	if (ast_strlen_zero(Mode)) {
		astman_send_error(s, m, "No Mode name provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	ast_log(LOG_DEBUG, "%s the host for %s\n", Mode, channel_name);
	if (!strcasecmp(Mode, "transfer")) {
		__transfer_moderator(s, m, conference, channel_name);
	} else if (!strcasecmp(Mode, "retrieve")) {
		__retrieve_moderator(s, m, conference, channel_name);
	}

	return 0;
}

static int action_confbridge_sfu_switch_video_quality(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	const char *quality = astman_get_header(m, "Quality");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}
	if (ast_strlen_zero(quality)) {
		astman_send_error(s, m, "No quality name provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	resolution = video_resolution_name_to_type(quality);
	if (resolution == VIDEO_RESOLUTION_NONE) {
		astman_send_error(s, m, "Quality name is not supported.");
		return 0;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (strcasecmp(ast_channel_tech(user->chan)->type, "PJSIP") || !strcmp(ast_channel_name(user->chan), channel_name)) {
			continue;
		}

		if (ast_conf_switch_main_video_resolution(user->chan, conference->name, resolution)) {
			astman_send_error(s, m, "No Video Resource.");
		} else {
			astman_send_ack(s, m, "");
			/*refresh*/
			ast_channel_lock(user->chan);
			struct ast_stream_topology *topology = ast_stream_topology_clone(ast_channel_get_stream_topology(user->chan));
			if (topology) {
				int i = 0;
				for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
					struct ast_stream *stream = ast_stream_topology_get_stream(topology, i);
					if (ast_stream_get_state(stream) != AST_STREAM_STATE_REMOVED
						&& ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO
						&& ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN
						&& strncmp(ast_stream_get_name(stream), SOFTBRIDGE_VIDEO_DEST_PREFIX,
							SOFTBRIDGE_VIDEO_DEST_LEN)) {
						ast_stream_set_video_resolution(stream, resolution);
						break;
					}
				}

				ast_channel_tech(user->chan)->indicate(user->chan, AST_CONTROL_STREAM_TOPOLOGY_REQUEST_CHANGE, topology, sizeof(topology));
				ast_stream_topology_free(topology);
			}
			ast_channel_unlock(user->chan);

			conf_user_write_operator_log(user, CONF_OP_SWITCH_RESOLUTION, quality);
		}

		break;
	}
	ao2_unlock(conference);

	return 0;
}

static int change_role_get_pin(const char *pin, struct ast_channel *chan)
{
	char pin_guess[MAX_PIN+1] = {0};
	char *tmp = pin_guess;
	int i, res;
	unsigned int len = MAX_PIN ;

	/* give them three tries to get the pin right */
	for (i = 0; i < 3; i++) {
		if (ast_app_getdata(chan, (!strlen(pin_guess) ? "vm-password" : ""), tmp, len, 0) >= 0) {
			if (!strcasecmp(pin, pin_guess)) {
				return 0;
			}
		}

		if (i == 2) {
			break;
		}

		ast_streamfile(chan, "auth-incorrect", ast_channel_language(chan));
		res = ast_waitstream(chan, AST_DIGIT_ANY);
		if (res > 0) {
			/* Account for digit already read during ivalid pin playback
			 * resetting pin buf. */
			pin_guess[0] = res;
			pin_guess[1] = '\0';
			tmp = pin_guess + 1;
			len = MAX_PIN - 1;
		} else {
			/* reset pin buf as empty buffer. */
			pin_guess[0] = '\0';
			tmp = pin_guess;
			len = MAX_PIN;
		}
	}

	return -1;
}

static int action_change_role(struct confbridge_conference *conference,
	struct ast_bridge_channel *bridge_channel,
	struct confbridge_user *user)
{
	int res = 0;
	const char *file = NULL;
	char digits[2] = {0};
	int new_role_admin = 0;
	struct confbridge_user *cur_user = NULL;

	if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		file = conf_get_sound(CONF_SOUND_GIVEUP_MODERATOR, conference->b_profile.sounds);
		new_role_admin = 0;
	} else {
		file = conf_get_sound(CONF_SOUND_ACQUIRE_MODERATOR, conference->b_profile.sounds);
		new_role_admin = 1;
	}

	res = ast_app_getdata(user->chan, file, digits, 1, 5000);
	if (res < 0) {
		return -1;
	} else if (strcmp(digits, "1")) {
		return 0;
	}

	conf_user_write_operator_log(user, CONF_OP_CHANGE_ROLE, new_role_admin ? "moderator" : "participant");

	if (!new_role_admin) {
		ao2_lock(user->conference);
		/*Determine whether normal users can invite*/
		ast_clear_flag(&user->u_profile, USER_OPT_ADMIN);
		ast_clear_flag(&user->u_profile, USER_OPT_MARKEDUSER);
		if (!user->conference->user_can_invite) {
			ast_clear_flag(&user->u_profile, USER_OPT_CAN_INVITE);
		}
		/*Added for Bug 152674*/
		cache_conf_remove_moderator(user->conference->name,user->chan);
		conference->markedusers--;
		/*if muteall,then we should mute the old host.Bug 267891*/
		if (conference->muted && !user->muted) {
			user->muted = conference->muted;
			conf_update_user_mute(user);
			send_mute_event(user, conference, 1, 0);
		}
		reload_menu_to_user(user, MENU_SFU_USER);
		if (conference->state->state_to_unmarked) {
			conference->state->state_to_unmarked(user);
		}
		send_role_update_event(user, conference);
		ao2_unlock(user->conference);
		//send_moderator_event(user, conference);
		if (!conference->focus_chan) {
			sfu_topologies_on_moderator_change(conference, NULL);
		} else {
			ast_log(LOG_DEBUG, "action_change_role, already has focus chan, don't upate host msid\n");
		}
		play_file(bridge_channel, bridge_channel->chan, conf_get_sound(CONF_SOUND_YOU_ARE_PATICIPANT, conference->b_profile.sounds));
	} else {
		if (!ast_strlen_zero(user->u_profile.admin_pin)) {
			res = change_role_get_pin(user->u_profile.admin_pin, bridge_channel->chan);
			if (res < 0) {
				action_playback(bridge_channel, "sorry-cant-let-you-do-that&auth-thankyou");
				return 0;
			}
		}

		ao2_lock(user->conference);
		AST_LIST_TRAVERSE(&user->conference->active_list, cur_user, list) {
			if (ast_test_flag(&cur_user->u_profile, USER_OPT_ADMIN)) {
				break;
			}
		}
		if (cur_user) {
			play_file(bridge_channel, bridge_channel->chan, conf_get_sound(CONF_SOUND_SORRY_CANNOT_BE_MODERATOR, conference->b_profile.sounds));
			ao2_unlock(user->conference);
			return 0;
		} else {
			user->muted = 0;
			ast_set_flag(&user->u_profile, USER_OPT_ADMIN);
			ast_set_flag(&user->u_profile, USER_OPT_MARKEDUSER);
			ast_set_flag(&user->u_profile, USER_OPT_CAN_INVITE);
			conference->markedusers++;
			send_unmute_event(user, conference, 1, 0);
			reload_menu_to_user(user, MENU_SFU_ADMIN);
			if (conference->state->state_to_marked) {
				conference->state->state_to_marked(user);
			}
			send_role_update_event(user, conference);
			//send_moderator_event(user, conference);
			conf_update_user_mute(user);
			send_unmute_event(user, conference, 1, 0);
			if (user->term_type == AST_TERMINAL_TYPE_WEBRTC || user->term_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
				cache_conf_update_moderator(conference->name,
											S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, ""),
											user->chan);
			}
			if (!conference->focus_chan) {
				sfu_topologies_on_moderator_change(conference, user);
			} else {
				ast_log(LOG_DEBUG, "action_change_role, already has focus chan, don't upate host msid\n");
			}
			play_file(bridge_channel, bridge_channel->chan, conf_get_sound(CONF_SOUND_YOU_ARE_MODERATOR, conference->b_profile.sounds));
		}
		ao2_unlock(user->conference);
	}

	return 0;
}

static int action_confbridge_authorize_user(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *extension_name = astman_get_header(m, "Extension");
	const char *role = astman_get_header(m, "Role");
	const char *terminal_type = astman_get_header(m, "Terminal");
	const char *schedule_id = astman_get_header(m, "ScheduleId");
	int is_admin = 0;
	int res = 0;
	char msg[128] = {0};

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(extension_name)) {
		astman_send_error(s, m, "No extension name provided.");
		return 0;
	}
	if (ast_strlen_zero(role)) {
		astman_send_error(s, m, "No role provided.");
		return 0;
	}

	is_admin = !strcasecmp(role, "admin") ? 1 : 0;
	res = __cache_authorize_user(conference_name, CONF_TYPE_AUDIO, extension_name, &is_admin, terminal_type, schedule_id, NULL);
	if (res < 0) {
		astman_send_error(s, m, "Fail to authorize user.");
		return 0;
	}

	if (is_admin) {
		ast_copy_string(msg, "Role is admin.", sizeof(msg));
	} else {
		ast_copy_string(msg, "Role is participant.", sizeof(msg));
	}

	astman_send_ack(s, m, msg);
	return 0;
}

static const char *get_admin_menu(struct confbridge_user *user)
{
	if (ast_test_flag(&user->u_profile, USER_OPT_MENU_ALLOW)) {
		return MENU_ADMIN_CONF_MENU;
	} else {
		return MENU_INVITE;
	}
}

static const char *get_participant_menu(struct confbridge_user *user)
{
	if (ast_test_flag(&user->u_profile, USER_OPT_MENU_ALLOW) && ast_test_flag(&user->u_profile, USER_OPT_INVITE_ALLOW)) {
		return MENU_USER_CONF_MENU_INVITE;
	} else if (ast_test_flag(&user->u_profile, USER_OPT_MENU_ALLOW)) {
		return MENU_USER_CONF_MENU;
	} else if (ast_test_flag(&user->u_profile, USER_OPT_INVITE_ALLOW)) {
		return MENU_INVITE;
	} else {
		return MENU_DEFAULT;
	}
}

static int __conf_transfer_moderator(struct mansession *s, const struct message *m, struct confbridge_conference *conference, const char *channel)
{
	struct confbridge_user *admin_user = NULL;
	struct confbridge_user *old_admin_user = NULL;

	ao2_lock(conference);
	old_admin_user = find_active_user_by_channel(conference, astman_get_header(m, "Originator"));
	if (old_admin_user && !ast_test_flag(&old_admin_user->u_profile, USER_OPT_ADMIN)) {
		ao2_unlock(conference);
		astman_send_error(s, m, "The originator has no admin privilege.");
		return 0;
	}

	admin_user = find_active_user_by_channel(conference, channel);
	if (!admin_user) {
		ao2_unlock(conference);
		astman_send_error(s, m, "No participant.");
		return 0;
	}

	if (old_admin_user == admin_user) {
		ao2_unlock(conference);
		astman_send_ack(s, m, "");
		return 0;
	}

	if (old_admin_user) {
		ast_clear_flag(&old_admin_user->u_profile, USER_OPT_ADMIN);
		ast_clear_flag(&old_admin_user->u_profile, USER_OPT_MARKEDUSER);
		conference->markedusers--;
		/*if muteall,then we should mute the old host.Bug 178615*/
		if (conference->muted && !old_admin_user->muted) {
			old_admin_user->muted = conference->muted;
			conf_update_user_mute(old_admin_user);
			send_unmute_event(old_admin_user, conference, 1, 0);
		}
		reload_menu_to_user(old_admin_user, get_participant_menu(old_admin_user));
		send_role_update_event(old_admin_user, conference);
	}

	if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
		ao2_unlock(conference);
		astman_send_ack(s, m, "");
		return 0;
	}

	/*new host should not be mueted,Bug 176454*/
	if (admin_user->muted) {
		admin_user->muted = 0;
		conf_update_user_mute(admin_user);
		send_unmute_event(admin_user, conference, 1, 0);
	}
	ast_set_flag(&admin_user->u_profile, USER_OPT_ADMIN);
	ast_set_flag(&admin_user->u_profile, USER_OPT_MARKEDUSER);
	conference->markedusers++;
	reload_menu_to_user(admin_user, get_admin_menu(admin_user));
	send_role_update_event(admin_user, conference);

	if (conference->state->state_to_marked) {
		conference->state->state_to_marked(admin_user);
	}
	send_conf_stasis(conference, NULL, cei_refresh_all_type(), NULL, 1);

	conf_user_write_operator_log(admin_user, CONF_OP_TRANSFER_HOST, S_COR(ast_channel_caller(admin_user->chan)->id.number.valid, ast_channel_caller(admin_user->chan)->id.number.str, ""));

	ao2_unlock(conference);

	astman_send_ack(s, m, "");
	return 0;
}

static int __conf_retrieve_moderator(struct mansession *s, const struct message *m, struct confbridge_conference *conference, const char *channel)
{
	struct confbridge_user *admin_user = NULL;

	ao2_lock(conference);
	admin_user = find_active_user_by_channel(conference, astman_get_header(m, "Originator"));
	if (!admin_user) {
		ao2_unlock(conference);
		astman_send_error(s, m, "No participant.");
		return 0;
	} else if (ast_test_flag(&admin_user->u_profile, USER_OPT_ADMIN)) {
		ao2_unlock(conference);
		astman_send_ack(s, m, "");
		return 0;
	}

	admin_user->muted = 0;
	ast_set_flag(&admin_user->u_profile, USER_OPT_ADMIN);
	ast_set_flag(&admin_user->u_profile, USER_OPT_MARKEDUSER);
	conference->markedusers++;
	reload_menu_to_user(admin_user, get_admin_menu(admin_user));
	if (conference->state->state_to_marked) {
		conference->state->state_to_marked(admin_user);
	}
	send_role_update_event(admin_user, conference);
	conf_update_user_mute(admin_user);
	send_conf_stasis(conference, NULL, cei_refresh_all_type(), NULL, 1);

	conf_user_write_operator_log(admin_user, CONF_OP_RETRIEVE_HOST, NULL);

	ao2_unlock(conference);

	astman_send_ack(s, m, "");
	return 0;
}

static int action_confbridge_transfer_moderator(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	const char *Mode = astman_get_header(m, "Mode");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}
	if (ast_strlen_zero(Mode)) {
		astman_send_error(s, m, "No Mode name provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	ast_log(LOG_DEBUG, "%s the host for %s\n", Mode, channel_name);
	if (!strcasecmp(Mode, "transfer")) {
		__conf_transfer_moderator(s, m, conference, channel_name);
	} else if (!strcasecmp(Mode, "retrieve")) {
		__conf_retrieve_moderator(s, m, conference, channel_name);
	}

	return 0;
}

static int update_user_name(struct confbridge_conference *conference, const char *channel, const char *name)
{
	struct confbridge_user *user = NULL;

	ao2_lock(conference);
	user = find_active_user_by_channel(conference, channel);
	if (!user) {
		ao2_unlock(conference);
		return -1;
	}
	send_name_update_event(user, conference, name);
	conf_user_write_operator_log(user, CONF_OP_UPDATE_NAME, name);
	ao2_unlock(conference);

	return 0;
}

static int action_confbridge_rename_user(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");
	const char *name = astman_get_header(m, "Name");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}
	if (ast_strlen_zero(name)) {
		astman_send_error(s, m, "No Name provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	update_user_name(conference, channel_name, name);

	return 0;
}

static int action_confbridge_update_admin_pincode(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *admin_pincode = astman_get_header(m, "Admin_pincode");
	struct confbridge_user *user;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(admin_pincode)) {
		astman_send_error(s, m, "No Admin_pincode provided.");
		return 0;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		astman_send_error(s, m, "Couldn't find the conference instance.");
		return 0;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		if (user && strcmp(user->u_profile.admin_pin, admin_pincode)) {
			memset(user->u_profile.admin_pin, 0, MAX_PIN);
			snprintf(user->u_profile.admin_pin, MAX_PIN, admin_pincode);
		}
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		if (user && strcmp(user->u_profile.admin_pin, admin_pincode)) {
			memset(user->u_profile.admin_pin, 0, MAX_PIN);
			snprintf(user->u_profile.admin_pin, MAX_PIN, admin_pincode);
		}
	}
	ao2_unlock(conference);
	astman_send_ack(s, m, "");

	return 0;

}

static int action_confbridge_schedule_end(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct schedule_conference *, conference_s, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	ast_log(LOG_DEBUG, "action_confbridge_schedule_end entry");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}

	/*user in waiting list can join now*/
	schedule_conference_wait_user_join(conference_name);
	astman_send_ack(s, m, "");

	return 0;
}

/*  @case 1: If has user in wait list we can't remove info,
*   when the schedule conference start,the user will be notified and leave the wait list(Bug 200980).
*
*   @case 2: If no user is waiting in the wait list,
*   we must remove the wait list info.(Bug 204175)
*/
static int schedule_wait_user_join_and_list_remove(const char *conference)
{
	ast_log(LOG_DEBUG, "Check whether the waiting information needs to be removed. \n");

	int need_remove = 1;
	struct schedule_attendees_wait *user = NULL;
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	RAII_VAR(struct schedule_conference_wait *, conf_wait, NULL, ao2_cleanup);

	if (ast_strlen_zero(conference)) {
		ast_log(LOG_DEBUG, "schedule_wait_list_remove_when_no_wait_users conference null. \n");
		return 0;
	}

	conf_sch_w.conf_name = conference;
	conf_wait = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conf_wait) {
		ast_log(LOG_DEBUG, "no wait list info!");
		return 0;
	}

	ao2_lock(conf_wait);
	AST_LIST_TRAVERSE(&conf_wait->schedule_attendees_wait_list, user, entry) {
		if (user->chan) {
			ast_queue_control(user->chan, AST_CONTROL_NOTIFY_SCHEDULE_START);
			ast_log(LOG_DEBUG, "send schedule start notify to wait_user channel_name %s", user->channel_name);
		}
		need_remove = 0;
	}

	if (need_remove) {
		schedule_conference_wait_list_remove(conference);
	} else {
		ast_log(LOG_DEBUG, "Join notify has sended, no need to remove, wait for the user to join and remove automatically \n");
	}
	ao2_unlock(conf_wait);

	return 0;
}

static int action_confbridge_schedule_start(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct schedule_conference *, conference_s, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *conf_type = astman_get_header(m, "Conf_type");
	const char *schedule_id = astman_get_header(m, "Reservation_id");
	const char *utc_start_time = astman_get_header(m, "Utc_start_time");
	const char *conf_subject = astman_get_header(m, "Conf_subject");
	const char *host_exten = astman_get_header(m, "Host_exten");
	const char *host_email = astman_get_header(m, "Host_email");
	const char *start_time = astman_get_header(m, "Start_time");

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(conf_type)) {
		astman_send_error(s, m, "No Conference type provided.");
		return 0;
	}
	if (ast_strlen_zero(schedule_id)) {
		astman_send_error(s, m, "No Reservation id provided.");
		return 0;
	}
	if (ast_strlen_zero(utc_start_time)) {
		astman_send_error(s, m, "No Utc start time provided.");
		return 0;
	}
	if (ast_strlen_zero(conf_subject)) {
		astman_send_error(s, m, "No Conference subject provided.");
		return 0;
	}
	if (ast_strlen_zero(start_time)) {
		astman_send_error(s, m, "No start time provided.");
		return 0;
	}

	/*schedule conference start, check whether need remove wait list*/
	schedule_wait_user_join_and_list_remove(conference_name);

	/*user in waiting list can join now*/
	//schedule_conference_wait_user_join(conference_name);

	AST_LIST_LOCK(&schedule_conference_list);

	AST_LIST_TRAVERSE(&schedule_conference_list, conference_s, entry) {
		if (!strcmp(conference_s->name, conference_name) && !strcmp(conference_s->schedule_id, schedule_id)) {
			ao2_ref(conference_s, +1);
			break;
		}
	}

	if (!conference_s) {
		if ((conference_s = ao2_alloc(sizeof(*conference_s), schedule_conference_destroy))) {
			ast_log(LOG_DEBUG, "Add a new schedule conference '%s'.\n", conference_name);
			ast_copy_string(conference_s->name, conference_name, sizeof(conference_s->name));
			ast_copy_string(conference_s->schedule_id, schedule_id, sizeof(conference_s->schedule_id));
			ast_copy_string(conference_s->utc_start_time, utc_start_time, sizeof(conference_s->utc_start_time));
			ast_copy_string(conference_s->conf_subject, conf_subject, sizeof(conference_s->conf_subject));
			ast_copy_string(conference_s->host_exten, host_exten ? host_exten : "", sizeof(conference_s->host_exten));
			ast_copy_string(conference_s->host_email, host_email ? host_email : "", sizeof(conference_s->host_email));
			ast_copy_string(conference_s->start_time, start_time, sizeof(conference_s->start_time));
			conference_s->type = !strcasecmp(conf_type, "video") ? CONF_TYPE_VIDEO : CONF_TYPE_AUDIO;
			AST_LIST_INSERT_TAIL(&schedule_conference_list, conference_s, entry);
			ao2_ref(conference_s, +1);
		}
	}

	AST_LIST_UNLOCK(&schedule_conference_list);
	astman_send_ack(s, m, "");

	return 0;
}

int conf_get_exten_by_channel(const char *channel_name, char *buf, int size)
{
	const char *start = NULL;
	const char *end =  NULL;

	if (ast_strlen_zero(channel_name)) {
		return -1;
	}

	memset(buf, 0, size);

	if (!strncasecmp(channel_name, "PJSIP/", 6)) {
		start = channel_name + 6;
		end = strrchr(start, '-');
		if (!end) {
			return -1;
		}
		strncpy(buf, start, size > (end - start) ? (end - start) : size);
	} else if (!strncasecmp(channel_name, "Local/", 6)) {
		start = channel_name + 6;
		end = strrchr(start, '@');
		if (!end) {
			return -1;
		}
		strncpy(buf, start, size > (end - start) ? (end - start) : size);
		char * token = strrchr(buf, '-');
		if (token) {
			*token = '\0';
		}
	} else if (!strncasecmp(channel_name, "IAX2/", 5)) {
		start = channel_name + 5;
		end = strrchr(start, '-');
		if (!end) {
			return -1;
		}
		strncpy(buf, start, size > (end - start) ? (end - start) : size);
	} else if (!strncasecmp(channel_name, "DAHDI/", 6)) {
		char family[64] = {0};
		char key[64] = {0};

		end = strrchr(channel_name, '-');
		if (!end) {
			return -1;
		}
		strncpy(family, channel_name, sizeof(family) > (end - channel_name) ? (end - channel_name) : sizeof(family));
		snprintf(key, sizeof(key), "FXS_EXTENSION");
		ast_ucm_db_get(family, key, buf, size);
	} else {
		return -1;
	}

	return 0;
}

static int schedule_conference_wait_user_add_cache(struct ast_channel *chan, const char *conference_name, int type, const char *schedule_id)
{
	int res = 0;
	int is_admin = 0;
	char extension[AST_MAX_EXTENSION] = {0};
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	const char * terminate_str = NULL;
	const char *related_info = NULL;

	if (!chan || ast_strlen_zero(conference_name) || ast_strlen_zero(schedule_id)) {
		res = -1;
	}
	const char *wait_is_admin = pbx_builtin_getvar_helper(chan, "WAIT_IS_ADMIN");
	if (!ast_strlen_zero(wait_is_admin) && !strcasecmp(wait_is_admin, "yes")) {
		is_admin = 1;
	} else {
		is_admin = 0;
	}

	terminate_type = ast_channel_get_terminate_type(chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	conf_get_exten_by_channel(ast_channel_name(chan), extension, sizeof(extension)-1);
	if (!strncasecmp(extension, "trunk", 5)) {
		ast_log(LOG_DEBUG, "extension %s", extension);
		ast_copy_string(extension, ast_channel_caller_num(chan), sizeof(extension));
	}
	if (ast_strlen_zero(extension)) {
		ast_copy_string(extension, S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "Anonymous"), sizeof(extension));
	}
	related_info = pbx_builtin_getvar_helper(chan, "RELATED_INFO");

	res = __cache_authorize_user(conference_name, type, extension, &is_admin, terminate_str, schedule_id, related_info);
	if (res < 0) {
		ast_log(LOG_WARNING, "fail to add wait user into cache!");
	}
	if (!res && is_admin) {
		cache_conf_update_moderator(conference_name, extension, chan);
	}
	return res;
}

static int schedule_conference_wait_user_remove_cache(const char *conference, struct ast_channel *chan)
{
	RAII_VAR(struct schedule_conference_wait *, conference_obj, NULL, ao2_cleanup);
	struct schedule_attendees_wait *schedule_wait_user = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	const char * terminate_str = NULL;
	int res = 0;

	if (ast_strlen_zero(conference) || !chan) {
		ast_log(LOG_WARNING, "conference or chan is null");
		res = -1;
		return res;
	}

	//conference_obj = schedule_conf_wait_fetch_add(conference, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conf_sch_w.conf_name = conference;
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference);
		res = -1;
		return res;
	}

	terminate_type = ast_channel_get_terminate_type(chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	ao2_lock(conference_obj);

	schedule_wait_user = schedule_attendee_wait_find(conference_obj, ast_channel_name(chan));
	if (schedule_wait_user) {
		cache_conf_remove_user(conference, schedule_wait_user->extension, terminate_str);
	}
	ao2_unlock(conference_obj);
	return res;
}

static int schedule_conference_wait_user_join(const char *conference_name)
{
	ast_log(LOG_DEBUG, "action_user_start_join entry");
	RAII_VAR(struct schedule_conference_wait*, conference_obj, NULL, ao2_cleanup);
	struct schedule_attendees_wait *schedule_wait_user = NULL;
	struct schedule_conference_wait_info conf_sch_w = { 0 };

	if (ast_strlen_zero(conference_name)) {
		ast_log(LOG_DEBUG, "No Conference name provided.");
		return 0;
	}

	//conference_obj = schedule_conf_wait_fetch_add(conference_name, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conf_sch_w.conf_name = conference_name;
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference_name);
		return 0;
	}

	ao2_lock(conference_obj);
	AST_LIST_TRAVERSE(&conference_obj->schedule_attendees_wait_list, schedule_wait_user, entry) {
		if (schedule_wait_user->chan) {
			ast_queue_control(schedule_wait_user->chan, AST_CONTROL_NOTIFY_SCHEDULE_START);
			ast_log(LOG_DEBUG, "send schedule start notify to wait_user channel_name %s", schedule_wait_user->channel_name);
		}
	}
	ao2_unlock(conference_obj);

	ast_log(LOG_DEBUG, "action_user_start_join exit");

	return 0;
}

static int schedule_conference_join_early(struct mansession *s, const struct message *m)
{
	ast_log(LOG_DEBUG, "schedule_conference_join_early entry");
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	RAII_VAR(struct schedule_conference_wait*, conference_obj, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *conf_type = astman_get_header(m, "Conf_type");
	const char *schedule_id = astman_get_header(m, "Reservation_id");
	const char *pin_code = astman_get_header(m, "Pin_code");
	const char *adminpin_code = astman_get_header(m, "Adminpin_code");
	const char *utc_start_time = astman_get_header(m, "Utc_start_time");
	const char *conf_subject = astman_get_header(m, "Conf_subject");
	const char *host_exten = astman_get_header(m, "Host_exten");
	const char *host_email = astman_get_header(m, "Host_email");
	const char *start_time = astman_get_header(m, "Start_time");
	int is_sfu = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(conf_type)) {
		astman_send_error(s, m, "No Conference type provided.");
		return 0;
	}
	if (ast_strlen_zero(schedule_id)) {
		astman_send_error(s, m, "No Reservation id provided.");
		return 0;
	}
	if (ast_strlen_zero(utc_start_time)) {
		astman_send_error(s, m, "No Utc start time provided.");
		return 0;
	}
	if (ast_strlen_zero(conf_subject)) {
		astman_send_error(s, m, "No Conference subject provided.");
		return 0;
	}
	if (ast_strlen_zero(start_time)) {
		astman_send_error(s, m, "No start time provided.");
		return 0;
	}

	is_sfu = !strcasecmp(conf_type, "video") ? 1 : 0;
	conf_sch_w.conf_name = conference_name;
	conf_sch_w.conf_type = is_sfu ? CONF_TYPE_VIDEO : CONF_TYPE_AUDIO;
	conf_sch_w.pin = pin_code;
	conf_sch_w.admin_pin = adminpin_code;
	conf_sch_w.schedule_id = schedule_id;
	conf_sch_w.conf_subject = conf_subject;
	conf_sch_w.host_exten = host_exten;
	conf_sch_w.host_email = host_email;
	conf_sch_w.utc_start_time = utc_start_time;
	conf_sch_w.start_time = start_time;

	conference_obj = schedule_conf_wait_info_update(&conf_sch_w);
	if (!conference_obj) {
		conference_obj = schedule_conf_wait_info_fetch_add(1, &conf_sch_w);
		if (!conference_obj) {
			ast_log(LOG_DEBUG, "Failed to add the schedule conference waiting'%s'\n", conference_name);
			astman_send_ack(s, m, "");
			return 0;
		}
	}
	ast_log(LOG_DEBUG, "schedule_conference_join_early exit");
	astman_send_ack(s, m, "");

	return 0;
}

static int schedule_conference_wait_info_modify(struct mansession *s, const struct message *m)
{
	ast_log(LOG_DEBUG, "schedule_conference_join_early entry");
	RAII_VAR(struct schedule_conference_wait*, conference_obj, NULL, ao2_cleanup);
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	const char *conference_name = astman_get_header(m, "Conference");
	const char *action = astman_get_header(m, "Request_name");
	const char *schedule_id = astman_get_header(m, "Reservation_id");
	const char *schedule_data = astman_get_header(m, "Schedule_data");
	int user_count = 0;
	char msg[128] = {0};
	char schedule_id_wait[AST_MAX_SCHEDULE_ID] = {0};

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(action)) {
		astman_send_error(s, m, "No Action name provided.");
		return 0;
	}
	if (ast_strlen_zero(schedule_id)) {
		astman_send_error(s, m, "No Reservation id provided.");
		return 0;
	}
	if (strcasecmp(action, "delete") && strcasecmp(action, "update")) {
		astman_send_error(s, m, "Action name not support.");
		return 0;
	}

	conf_sch_w.conf_name = conference_name;
	//conference_obj = schedule_conf_wait_fetch_add(conference_name, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference_name);
		astman_send_ack(s, m, "Success");
		return 0;
	}
	ao2_lock(conference_obj);
	snprintf(schedule_id_wait, sizeof(schedule_id_wait), conference_obj->schedule_id);
	user_count = conference_obj->wait_users;
	ao2_unlock(conference_obj);
	if (!user_count) {
		if (!strcasecmp(action, "delete")) {
			schedule_conference_wait_list_remove_by_scheduleid(conference_name, schedule_id);
			ast_copy_string(msg, "Success", sizeof(msg));
		} else if (!strcasecmp(action, "update") && !ast_strlen_zero(schedule_data) && !strcmp(schedule_id_wait, schedule_id)) {
			schedule_conference_wait_list_update_by_scheduleid(conference_name, schedule_id, schedule_data);
			ast_copy_string(msg, "Success", sizeof(msg));
		} else {
			ast_copy_string(msg, "Success", sizeof(msg));
		}
	} else {
		if (!strcasecmp(action, "delete") && strcmp(schedule_id_wait, schedule_id)) {
			schedule_conference_wait_list_remove_by_scheduleid(conference_name, schedule_id);
			ast_copy_string(msg, "Success", sizeof(msg));
		} else if (!strcasecmp(action, "update") && !ast_strlen_zero(schedule_data) && strcmp(schedule_id_wait, schedule_id)) {
			schedule_conference_wait_list_update_by_scheduleid(conference_name, schedule_id, schedule_data);
			ast_copy_string(msg, "Success", sizeof(msg));
		} else {
			ast_copy_string(msg, "Fail", sizeof(msg));
		}
	}
	astman_send_ack(s, m, msg);
	return 0;
}

static int action_confbridge_report_statistics_remove_schedule(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *reservation_id = astman_get_header(m, "Reservation_id");

	if (ast_strlen_zero(conference_name) || ast_strlen_zero(reservation_id)) {
		astman_send_error(s, m, "No Conference name or Reservation_id provided.");
		return 0;
	}
	/*if early join schedule conference ended, clean wait list.*/
	schedule_conference_wait_list_remove_by_scheduleid(conference_name, reservation_id);
	schedule_conference_remove(conference_name, reservation_id);
	astman_send_ack(s, m, "");

	return 0;
}

static int action_confbridge_im_authorize(struct mansession *s, const struct message *m)
{
	RAII_VAR(struct im_conference *, conference, NULL, ao2_cleanup);
	const char *conference_name = astman_get_header(m, "Conference");
	const char *im_group_id = astman_get_header(m, "Groupid");
	const char *conf_type = astman_get_header(m, "Type");
	const char *strong_related = astman_get_header(m, "Related");
	const char *conf_subject = astman_get_header(m, "Subject");

	int is_sfu = 0;
	int strong_related_chat = 0;

	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(im_group_id)) {
		astman_send_error(s, m, "No Groupid name provided.");
		return 0;
	}
	if (ast_strlen_zero(conf_type)) {
		astman_send_error(s, m, "No Type provided.");
		return 0;
	}
	if (ast_strlen_zero(strong_related)) {
		astman_send_error(s, m, "No Related provided.");
		return 0;
	}

	is_sfu = !strcasecmp(conf_type, "video") ? 1 : 0;
	strong_related_chat = !strcasecmp(strong_related, "yes") ? 1 : 0;
	/*we need remove error first*/
	im_conf_list_remove_error();

	conference = im_conf_list_fetch_add(conference_name, im_group_id, strong_related_chat, is_sfu ? CONF_TYPE_VIDEO : CONF_TYPE_AUDIO, 1, conf_subject ? conf_subject : "");
	if (!conference) {
		ast_log(LOG_WARNING, "Failed to fetch the conference '%s'\n", conference_name);
		astman_send_error(s, m, "Fail to authorize user.");
		return 0;
	}

	return 0;
}

static int action_confbridge_ha_sync_record_count(struct mansession *s, const struct message *m)
{
	int count_current = 0;
	ast_mutex_lock(&record_count_lock);
	count_current = conf_report_get_call_count();
	if (count_current && count_current != count) {
		ast_log(LOG_DEBUG, "count_current %d and count %d, need sync\n", count_current, count);
		count = count_current;
	} else {
		ast_log(LOG_DEBUG,"db count [%d] VS count [%d]!", count_current, count);
	}
	ast_mutex_unlock(&record_count_lock);

	astman_send_ack(s, m, "Conference record count HA switch master sync down\n");
	return 0;
}
/*Start added by zyxia for CALL FLIP 20218.16*/
static int action_confbridge_remove_moderator_cache(struct mansession *s, const struct message *m)
{
	const char *conference_name = astman_get_header(m, "Conference");
	const char *channel_name = astman_get_header(m, "Channel");

	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;

	ast_log(LOG_DEBUG, "action_confbridge_remove_moderator_cache  111!");
	if (ast_strlen_zero(conference_name)) {
		astman_send_error(s, m, "No Conference name provided.");
		return 0;
	}
	if (ast_strlen_zero(channel_name)) {
		astman_send_error(s, m, "No Channel name provided.");
		return 0;
	}

	if ((conference = ao2_find(conference_bridges, conference_name, OBJ_KEY))) {
		ao2_lock(conference);
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_channel_name(user->chan) && !strcmp(ast_channel_name(user->chan),channel_name)
				&& ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
				ast_clear_flag(&user->u_profile, USER_OPT_ADMIN);
				ast_log(LOG_DEBUG, "action_confbridge_remove_moderator_cache  clear admin flag for channel:%s ",channel_name);
				break;
			}
		}
		ao2_unlock(conference);
	}


	struct ast_channel *admin_chan = NULL;

	admin_chan = ast_channel_get_by_name(channel_name);
	if(admin_chan) {

		if ( cache_conf_remove_moderator(conference_name, admin_chan)) {
			ast_channel_unref(admin_chan);
			astman_send_error(s, m, "Remove Moderator Cache error");
			return 0;
		}
		ast_channel_unref(admin_chan);
	}

	astman_send_ack(s, m, "Remove Moderator Succed");

	return 0;
}

/*End added by zyxia for CALL FLIP 20218.16*/

static void handle_cli_cache_user_list_item(struct ast_cli_args *a, struct cache_authorized_user *user)
{
	char strTime[80] = {0};
	struct ast_tm timeresult;

	ast_localtime(&user->start_time, &timeresult, NULL);
	ast_strftime(strTime, sizeof(strTime), DATE_FORMAT, &timeresult);

	ast_cli(a->fd, "%-14s %-30s %-14s %-21s %-16s %-14s %-18s\n",
		user->extension, user->channel_name, user->is_admin ? "admin" : "non-admin", strTime, user->terminal_type, user->user_schedule_id, user->related_info);
}

static char *handle_cli_confbridge_cache_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct cache_conference *cache_conf = NULL;
	struct cache_authorized_user *cache_user = NULL;
	char moderator[128] = {0};
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge cache list";
		e->usage =
			"Usage: confbridge cache list [<name>]\n"
			"       Lists all currently active cache conference or a specific cache conference.\n"
			"\n"
			"       When a conference name is provided, flags may be shown for users.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 3) {
		ast_cli(a->fd, "Cache Conference Name            Moderator      Schedule Id\n");
		ast_cli(a->fd, "================================ ======         ============\n");

		AST_LIST_LOCK(&cache_conference_list);
		AST_LIST_TRAVERSE(&cache_conference_list, cache_conf, entry) {
			ao2_lock(cache_conf);
			if (cache_conf->moderator) {
				snprintf(moderator, sizeof(moderator), S_COR(ast_channel_caller(cache_conf->moderator)->id.number.valid, ast_channel_caller(cache_conf->moderator)->id.number.str, ""));
			} else {
				AST_LIST_TRAVERSE(&cache_conf->auth_user_list, cache_user, entry) {
					if (cache_user->is_admin) {
						snprintf(moderator, sizeof(moderator), cache_user->extension);
						break;
					}
				}
			}
			ast_copy_string(schedule_id, cache_conf->schedule_id, sizeof(schedule_id));
			ao2_unlock(cache_conf);
			ast_cli(a->fd, "%-32s %-16s %s\n",
				cache_conf->name, moderator, schedule_id);
		}
		AST_LIST_UNLOCK(&cache_conference_list);
		return CLI_SUCCESS;
	}

	if (a->argc == 4) {
		ast_cli(a->fd, "User         Channel                      Role         Start Time          Terminal Type  Schedule Id  Related Info       \n");
		ast_cli(a->fd, "============ ====================         ============ =================== ============   ============ ===================\n");

		cache_conf = cache_conf_fetch_add(a->argv[3], CONF_TYPE_NONE, 0, NULL);
		if (cache_conf) {
			ao2_lock(cache_conf);
			AST_LIST_TRAVERSE(&cache_conf->auth_user_list, cache_user, entry) {
				handle_cli_cache_user_list_item(a, cache_user);
			}
			ao2_unlock(cache_conf);
			ao2_ref(cache_conf, -1);
		}
		return CLI_SUCCESS;
	}

	return CLI_SHOWUSAGE;
}

static char *handle_cli_schedule_conference_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct schedule_conference *schedule_conf = NULL;
	struct schedule_attendees *schedule_user = NULL;
	char conf_name[MAX_CONF_NAME] = {0};
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
	char utc_start_time[AST_MAX_START_TIME] = {0};
	char start_time[AST_MAX_START_TIME] = {0};
	char conf_subject[AST_MAX_CONF_SUBJECT] = {0};
	char host_exten[AST_MAX_HOST_EXTEN] = {0};
	char host_email[AST_MAX_HOST_EMAIL] = {0};
	char user_extension[AST_MAX_EXTENSION] = {0};
	char terminal_type[AST_MAX_TERMINAL_NAME] = {0};
	int max_user = 0;
	int total_user = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge schedule conference list";
		e->usage =
			"Usage: confbridge schedule conference list [<name>]\n"
			"       List scheduled conference information and member information.\n"
			"\n"
			"       When a conference name is provided, flags may be shown for schedule conference.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 4) {
		ast_cli(a->fd, "Conference Name      Reservation Id       Conference Subject                   Start Time          Max User    Total User   Utc Start Time        Host Exten          Host Email          \n");
		ast_cli(a->fd, "===============      ===============      ==================                   ==============      ========    ==========   ===============       ==========          =================\n");

		AST_LIST_LOCK(&schedule_conference_list);
		AST_LIST_TRAVERSE(&schedule_conference_list, schedule_conf, entry) {
			ao2_lock(schedule_conf);
			snprintf(conf_name, sizeof(conf_name), schedule_conf->name);
			snprintf(schedule_id, sizeof(schedule_id), schedule_conf->schedule_id);
			snprintf(utc_start_time, sizeof(utc_start_time), schedule_conf->utc_start_time);
			snprintf(start_time, sizeof(start_time), schedule_conf->start_time);
			snprintf(conf_subject, sizeof(conf_subject), schedule_conf->conf_subject);
			snprintf(host_exten, sizeof(host_exten), schedule_conf->host_exten);
			snprintf(host_email, sizeof(host_email), schedule_conf->host_email);
			max_user = schedule_conf->max_user;
			total_user = schedule_conf->total_user;
			ao2_unlock(schedule_conf);
			ast_cli(a->fd, "%-20s %-20s %-35s %-16s %-5d %-5d %-16s %-8s %-20s\n",
				conf_name, schedule_id, conf_subject, start_time, max_user, total_user, utc_start_time, host_exten, host_email);
		}
		AST_LIST_UNLOCK(&schedule_conference_list);
		return CLI_SUCCESS;
	}

	if (a->argc == 5) {
		ast_cli(a->fd, "User             Terminal Type\n");
		ast_cli(a->fd, "============     ============\n");

		schedule_conf = schedule_conf_cli_fetch(a->argv[4]);
		if (schedule_conf) {
			ao2_lock(schedule_conf);
			AST_LIST_TRAVERSE(&schedule_conf->schedule_attendee_list, schedule_user, entry) {
				snprintf(user_extension, sizeof(user_extension), schedule_user->extension);
				snprintf(terminal_type, sizeof(terminal_type), schedule_user->terminal_type);
				ast_cli(a->fd, "%-18s %-12s\n", user_extension, terminal_type);
			}
			ao2_unlock(schedule_conf);
			ao2_ref(schedule_conf, -1);
		}
		return CLI_SUCCESS;
	}

	return CLI_SHOWUSAGE;
}

static char *handle_cli_im_conference_info_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct im_conference *im_conf = NULL;
	int type = 0;
	int status = 0;
	int strong_related = 0;
	char name[MAX_CONF_NAME] = {0};
	char im_group_id[AST_MAX_IM_GROUP_ID] = {0};

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge im info list";
		e->usage =
			"Usage: confbridge im conference info list [<groupid>]\n"
			"       List im conference information.\n"
			"\n"
			"       When a im group id is provided, flags may be shown for conference name&strong related.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 4) {
		ast_cli(a->fd, "Conference Name      Group Id               Status      Strong Related     Type\n");
		ast_cli(a->fd, "===============      ===============        =======     ==============     ========  \n");

		AST_LIST_LOCK(&im_conference_list);
		AST_LIST_TRAVERSE(&im_conference_list, im_conf, entry) {
			ao2_lock(im_conf);
			snprintf(name, sizeof(name), im_conf->name);
			snprintf(im_group_id, sizeof(im_group_id), im_conf->im_group_id);
			status = im_conf->status;
			strong_related = im_conf->strong_related;
			type = im_conf->type;
			ao2_unlock(im_conf);
			ast_cli(a->fd, "%-20s %-20s %-6d %-6d %-6d\n", name, im_group_id, status, strong_related, type);
		}
		AST_LIST_UNLOCK(&im_conference_list);
		return CLI_SUCCESS;
	}

	if (a->argc == 5) {
		AST_LIST_LOCK(&im_conference_list);
		AST_LIST_TRAVERSE(&im_conference_list, im_conf, entry) {
			ao2_lock(im_conf);
			if(!strcasecmp(im_conf->im_group_id, a->argv[4])) {
				type = im_conf->type;
				strong_related = im_conf->strong_related;
				snprintf(name, sizeof(name), im_conf->name);
				ao2_unlock(im_conf);
				break;
			}
			ao2_unlock(im_conf);
		}
		if (!ast_strlen_zero(name)) {
			ast_cli(a->fd, "Info: %s&%s&%s\n", name, strong_related ? "yes" : "no", type == CONF_TYPE_VIDEO ? "video" : (type == CONF_TYPE_AUDIO ? "audio" : "unknow"));
		} else {
			ast_cli(a->fd, "Info: NoImConf\n");
		}
		AST_LIST_UNLOCK(&im_conference_list);

		return CLI_SUCCESS;
	}

	return CLI_SHOWUSAGE;
}

static char *handle_cli_schedule_conference_early_start_list(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct schedule_conference_wait *wait_conf = NULL;
	struct schedule_attendees_wait *wait_user = NULL;
	int type = 0;
	int user_count = 0;
	char pincode[MAX_PIN] = {0};
	char admin_pincode[MAX_PIN] = {0};
	char name[MAX_CONF_NAME] = {0};
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
	char utc_start_time[AST_MAX_START_TIME] = {0};
	char conf_subject[AST_MAX_CONF_SUBJECT] = {0};
	char host_exten[AST_MAX_HOST_EXTEN] = {0};
	char host_email[AST_MAX_HOST_EMAIL] = {0};

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge schedule conference start early list";
		e->usage =
			"Usage: confbridge schedule conference start early list [<name>]\n"
			"       List schedule conference  start early information.\n"
			"\n"
			"       When a name is provided, flags may be shown for schedule conference wait user info.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc == 6) {
		ast_cli(a->fd, "Conference Name      Rchedule Id           Pin              Admin Pin        Type    Subject              Host        Host Mial                 Utc Start Time      \n");
		ast_cli(a->fd, "===============      ===============       ============     ============     =====   ===================  ==========  ========================  ==================  \n");

		AST_LIST_LOCK(&schedule_conference_wait_list);
		AST_LIST_TRAVERSE(&schedule_conference_wait_list, wait_conf, entry) {
			ao2_lock(wait_conf);
			snprintf(name, sizeof(name), wait_conf->name);
			snprintf(pincode, sizeof(pincode), wait_conf->pin);
			snprintf(admin_pincode, sizeof(admin_pincode), wait_conf->admin_pin);
			snprintf(schedule_id, sizeof(schedule_id), wait_conf->schedule_id);
			snprintf(conf_subject, sizeof(conf_subject), wait_conf->conf_subject);
			snprintf(host_exten, sizeof(host_exten), wait_conf->host_exten);
			snprintf(host_email, sizeof(host_email), wait_conf->host_email);
			snprintf(utc_start_time, sizeof(utc_start_time), wait_conf->utc_start_time);
			type = wait_conf->type;
			ao2_unlock(wait_conf);
			ast_cli(a->fd, "%-20s %-23s %-14s %-16s %-7d %-20s %-10s %-25s %-20s\n", name, schedule_id, pincode, admin_pincode, type, conf_subject, host_exten, host_email, utc_start_time);
		}
		AST_LIST_UNLOCK(&schedule_conference_wait_list);
		return CLI_SUCCESS;
	}

	if (a->argc == 7) {
		ast_cli(a->fd, "Channel Name        Extension\n");
		ast_cli(a->fd, "============        =========\n");
		struct schedule_conference_wait_info conf_sch_w = { 0 };
		conf_sch_w.conf_name = a->argv[6];
		//wait_conf = schedule_conf_wait_fetch_add(a->argv[6], 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
		wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
		if (wait_conf) {
			ao2_lock(wait_conf);
			AST_LIST_TRAVERSE(&wait_conf->schedule_attendees_wait_list, wait_user, entry) {
				ast_cli(a->fd, "%-20s %-12s\n", wait_user->channel_name, wait_user->extension);
			}
			ast_cli(a->fd, "Total wait users\n");
			ast_cli(a->fd, "%-10d\n", wait_conf->wait_users);
			ao2_unlock(wait_conf);
			ao2_ref(wait_conf, -1);
		}

		return CLI_SUCCESS;
	}

	if (a->argc == 8) {
		struct schedule_conference_wait_info conf_sch_w = { 0 };
		conf_sch_w.conf_name = a->argv[6];
		if (a->argv[7] && !strcmp(a->argv[7], "schedule_info")) {
			//wait_conf = schedule_conf_wait_fetch_add(a->argv[6], 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
			wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
			if (wait_conf) {
				ao2_lock(wait_conf);
				snprintf(schedule_id, sizeof(schedule_id), wait_conf->schedule_id);
				ao2_unlock(wait_conf);
				ao2_ref(wait_conf, -1);
			}
			if (!ast_strlen_zero(schedule_id)) {
				ast_cli(a->fd, "schedule_info: %s\n", schedule_id);
			} else {
				ast_cli(a->fd, "schedule_info: NoInfo\n");
			}
		} else if (a->argv[7] && !strcmp(a->argv[7], "wait_users")) {
			wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
			if (wait_conf) {
				ao2_lock(wait_conf);
				user_count = wait_conf->wait_users;
				ao2_unlock(wait_conf);
				ao2_ref(wait_conf, -1);
			}
			ast_cli(a->fd, "wait_users: %d\n", user_count);
		} else if (a->argv[7] && !strcmp(a->argv[7], "users_list")) {
			char *user_list = NULL;
			wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
			if (wait_conf) {
				ao2_lock(wait_conf);
				user_count = wait_conf->wait_users;
				int calloc_len = (user_count + 1) * AST_MAX_EXTENSION;
				user_list = ast_calloc(1, calloc_len);
				if (user_list) {
					AST_LIST_TRAVERSE(&wait_conf->schedule_attendees_wait_list, wait_user, entry) {
						snprintf(user_list + strlen(user_list), (calloc_len - strlen(user_list)), "%s ", wait_user->extension);
					}
				}
				ao2_unlock(wait_conf);
				ao2_ref(wait_conf, -1);
			}
			ast_cli(a->fd, "%s\n", user_list ? user_list : " ");
			if (user_list) {
				ast_free(user_list);
				user_list = NULL;
			}
		}
		return CLI_SUCCESS;
	}

	return CLI_SHOWUSAGE;
}

static char *handle_cli_schedule_conference_early_start_modify(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct schedule_conference_wait *wait_conf = NULL;
	int user_count = 0;
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
	struct schedule_conference_wait_info conf_sch_w = { 0 };

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge schedule conference start early";
		e->usage =
			"Usage: confbridge schedule conference start early [<action>] [<confnum>] [<scheduleid>]\n"
			"       delete or update schedule conference start early information.\n"
			"\n"
			"       When confnum and scheduleid is provided and no wait users in schedule conference,we can delete or update it.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (!a->argv[5] || (strcasecmp(a->argv[5], "delete") && strcasecmp(a->argv[5], "update"))) {
		return CLI_SHOWUSAGE;
	}

	if (a->argc == 7) {
		if (a->argv[5] && a->argv[6]) {
			if (!strcasecmp(a->argv[5], "delete")) {
				conf_sch_w.conf_name = a->argv[6];
				wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
				if (wait_conf) {
					ao2_lock(wait_conf);
					user_count = wait_conf->wait_users;
					ao2_unlock(wait_conf);
					ao2_ref(wait_conf, -1);

					/*If no wait users, delete it*/
					if (!user_count) {
						schedule_conference_wait_list_remove(a->argv[6]);
						ast_cli(a->fd, "Status: success\n");
					} else {
						ast_cli(a->fd, "Status: fail\n");
					}
				} else {
					ast_cli(a->fd, "Status: success\n");
				}
			}
		}
		return CLI_SUCCESS;
	} else if (a->argc == 8) {
		if (a->argv[5] && a->argv[6] && a->argv[7]) {
			if (!strcasecmp(a->argv[5], "delete")) {
				conf_sch_w.conf_name = a->argv[6];
				wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
				if (wait_conf) {
					ao2_lock(wait_conf);
					snprintf(schedule_id, sizeof(schedule_id), wait_conf->schedule_id);
					user_count = wait_conf->wait_users;
					ao2_unlock(wait_conf);
					ao2_ref(wait_conf, -1);

					/*If no wait users, delete it*/
					if (!ast_strlen_zero(schedule_id) && !strcasecmp(schedule_id, a->argv[7]) && !user_count) {
						schedule_conference_wait_list_remove_by_scheduleid(a->argv[6], a->argv[7]);
						ast_cli(a->fd, "Status: success\n");
					} else if (!ast_strlen_zero(schedule_id) && !strcasecmp(schedule_id, a->argv[7]) && user_count) {
						ast_cli(a->fd, "Status: fail\n");
					} else {
						ast_cli(a->fd, "Status: success\n");
					}
				} else {
					ast_cli(a->fd, "Status: success\n");
				}
			}
		}
		return CLI_SUCCESS;
	} else if (a->argc == 9) {
			if (!strcasecmp(a->argv[5], "update")) {
				conf_sch_w.conf_name = a->argv[6];
				//wait_conf = schedule_conf_wait_fetch_add(a->argv[6], 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
				wait_conf = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
				if (wait_conf) {
					ao2_lock(wait_conf);
					snprintf(schedule_id, sizeof(schedule_id), wait_conf->schedule_id);
					user_count = wait_conf->wait_users;
					ao2_unlock(wait_conf);
					ao2_ref(wait_conf, -1);

					/*If no wait users, delete it*/
					if (!ast_strlen_zero(schedule_id) && !strcasecmp(schedule_id, a->argv[7]) && !user_count) {
						schedule_conference_wait_list_update_by_scheduleid(a->argv[6], a->argv[7], a->argv[8]);
						ast_cli(a->fd, "Status: success\n");
					} else if (!user_count) {
						ast_cli(a->fd, "Status: success\n");
					} else {
						ast_cli(a->fd, "Status: fail\n");
					}
				} else {
					ast_cli(a->fd, "Status: success\n");
				}
			}
			return CLI_SUCCESS;
		}
	return CLI_SHOWUSAGE;
}

static char *handle_cli_schedule_conference_need_wait(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int inuse_notify = 0;
	int need_wait = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge schedule conference judge wait";
		e->usage =
			"Usage: confbridge schedule conference judge wait [<confnum>] [<scheduleid>]\n"
			"       Determine whether you need to wait for schedule conference start.\n"
			"\n"
			"       When confnum and scheduleid is provided and wait info is reutrn.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 7) {
		return CLI_SHOWUSAGE;
	}

	if (a->argv[5] && a->argv[6]) {
		need_wait = schedule_conference_need_wait(a->argv[5], a->argv[6], &inuse_notify);
		if (need_wait) {
			ast_cli(a->fd, "Status: yes\n");
		} else {
			ast_cli(a->fd, "Status: no\n");
		}
		return CLI_SUCCESS;
	}
	return CLI_SHOWUSAGE;
}

static char *handle_cli_conference_current_host(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int is_host = 0;
	struct confbridge_user *user = NULL;
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge judge current host";
		e->usage =
			"Usage: confbridge conference current host [<confnum>] [<exten>]\n"
			"       judge whether the extension is host in the conference.\n"
			"\n"
			"       When confnum and extension is provided the result is reutrn.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 6) {
		return CLI_SHOWUSAGE;
	}

	conference = conf_find_bridge(a->argv[4]);
	if (conference) {
		char host_exten[AST_MAX_EXTENSION] = {0};
		ao2_lock(conference);
		/*active list*/
		AST_LIST_TRAVERSE(&conference->active_list, user, list) {
			if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
				if (user->chan) {
					conf_get_exten_by_channel(ast_channel_name(user->chan), host_exten, sizeof(host_exten));
					if (!ast_strlen_zero(host_exten) && !strcmp(host_exten, a->argv[5])) {
						is_host = 1;
						break;
					}
				}
			}
		}
		ao2_unlock(conference);
		ast_cli(a->fd, "Host: %d\n", is_host);
		return CLI_SUCCESS;
	} else {
		ast_cli(a->fd, "Host: -1\n");
		return CLI_SUCCESS;
	}
	return CLI_SHOWUSAGE;
}

static char *handle_cli_conference_current_users(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	int active_users = 0;
	int waiting_users = 0;
	struct confbridge_user *user = NULL;
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);

	switch (cmd) {
	case CLI_INIT:
		e->command = "confbridge conference current users";
		e->usage =
			"Usage: confbridge conference current users [<confnum>] [<bookid>]\n"
			"       list the number of current users in the conference.\n"
			"\n"
			"       When confnum is provided and current users number is reutrn.\n"
			"       When bookid is provided and current users number of which schedule is reutrn.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 5 && a->argc != 6) {
		return CLI_SHOWUSAGE;
	}

	conference = conf_find_bridge(a->argv[4]);
	if (conference) {
		if (a->argc == 5) {
			ao2_lock(conference);
			/*active list*/
			AST_LIST_TRAVERSE(&conference->active_list, user, list) {
				active_users++;
			}
			/*waiting list*/
			AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
				waiting_users++;
			}
			ao2_unlock(conference);
			ast_cli(a->fd, "Users: %d\n", active_users + waiting_users);
			return CLI_SUCCESS;
		} else {
			ao2_lock(conference);
			if (conference->schedule_id_processing && !strcmp(conference->schedule_id_processing, a->argv[5])) {
				/*active list*/
				AST_LIST_TRAVERSE(&conference->active_list, user, list) {
					active_users++;
				}
				/*waiting list*/
				AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
					waiting_users++;
				}
			}
			ao2_unlock(conference);
			ast_cli(a->fd, "Users: %d\n", active_users + waiting_users);
			return CLI_SUCCESS;
		}
	} else {
		ast_cli(a->fd, "Users: 0\n");
		return CLI_SUCCESS;
	}
	return CLI_SHOWUSAGE;
}

#endif

/*!
 * \internal
 * \brief Unregister a ConfBridge channel technology.
 * \since 12.0.0
 *
 * \param tech What to unregister.
 *
 * \return Nothing
 */
static void unregister_channel_tech(struct ast_channel_tech *tech)
{
	ast_channel_unregister(tech);
	ao2_cleanup(tech->capabilities);
}

/*!
 * \internal
 * \brief Register a ConfBridge channel technology.
 * \since 12.0.0
 *
 * \param tech What to register.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int register_channel_tech(struct ast_channel_tech *tech)
{
	tech->capabilities = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!tech->capabilities) {
		return -1;
	}
	ast_format_cap_append_by_type(tech->capabilities, AST_MEDIA_TYPE_UNKNOWN);
	if (ast_channel_register(tech)) {
		ast_log(LOG_ERROR, "Unable to register channel technology %s(%s).\n",
			tech->type, tech->description);
		return -1;
	}
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/*
 * need to free
*/
static char* str_replace(const char *str, const char *match, const char *replace)
{
	if ( str == NULL || match == NULL || replace == NULL )
	{
		return NULL;
	}

	char *ret, *r;
	const char *p, *q;
	size_t oldlen = strlen(match);
	size_t count, retlen, newlen = strlen(replace);

	if (oldlen != newlen)
	{
		for (count = 0, p = str; (q = strstr(p, match)) != NULL; p = q + oldlen)
			count++;
		/* this is undefined if p - str > PTRDIFF_MAX */
		retlen = p - str + strlen(p) + count * (newlen - oldlen);
	}
	else
		retlen = strlen(str);

	if ((ret = ast_malloc(retlen + 1)) == NULL)
		return NULL;

	for (r = ret, p = str; (q = strstr(p, match)) != NULL; p = q + oldlen)
	{
		/* this is undefined if q - p > PTRDIFF_MAX */
		ptrdiff_t l = q - p;
		memcpy(r, p, l);
		r += l;
		memcpy(r, replace, newlen);
		r += newlen;
	}
	strcpy(r, p);

	return ret;
}

static void destroy_contact_member(void *obj)
{
	struct conf_contact_member *member = obj;

	ast_log(LOG_DEBUG, "Destroy contact member '%s'", member->number);
	ast_string_field_free_memory(member);
}

static struct conf_contact_member *alloc_contact_member(const char *name, const char *number)
{
	struct conf_contact_member *member = NULL;

	ast_log(LOG_DEBUG, "Alloc contact member '%s'", number);
	if ((member = ao2_alloc(sizeof(*member), destroy_contact_member))) {
		if (ast_string_field_init(member, 64)) {
			ao2_cleanup(member);
			return NULL;
		}
		ast_string_field_set(member, name, name);
		ast_string_field_set(member, number, number);
	}
	return member;
}

static void destroy_contact_group(void *obj)
{
	struct conf_contact_group *contact_group = obj;
	struct conf_contact_member *member = NULL;

	ast_log(LOG_DEBUG, "Destroy contact group '%s'", contact_group->name);
	ast_string_field_free_memory(contact_group);
	/*free members*/
	AST_LIST_TRAVERSE_SAFE_BEGIN(&contact_group->member_list, member, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ao2_cleanup(member);
	}
	AST_LIST_TRAVERSE_SAFE_END;
}

static struct conf_contact_group *alloc_contact_group(const char *name)
{
	struct conf_contact_group *contact_group = NULL;

	ast_log(LOG_DEBUG, "Alloc contact group '%s'", name);
	if ((contact_group = ao2_alloc(sizeof(*contact_group), destroy_contact_group))) {
		if (ast_string_field_init(contact_group, 64)) {
			ao2_cleanup(contact_group);
			return NULL;
		}
		ast_string_field_set(contact_group, name, name);
	}
	return contact_group;
}

/*! \brief Comparison function used for conference contact groups container */
static int conf_contact_groups_cmp_cb(void *obj, void *arg, int flags)
{
	const struct conf_contact_group *left = obj;
	const struct conf_contact_group *right = arg;
	const char *right_name = arg;
	int cmp;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
		default:
		case OBJ_POINTER:
			right_name = right->name;
			/* Fall through */
		case OBJ_KEY:
			cmp = strcasecmp(left->name, right_name);
			break;
		case OBJ_PARTIAL_KEY:
			cmp = strncasecmp(left->name, right_name, strlen(right_name));
			break;
	}
	return cmp ? 0 : CMP_MATCH;
}

static void destroy_conf_contact_groups(void)
{
	ao2_cleanup(conference_contact_groups);
	conference_contact_groups = NULL;
}

static int load_conf_contact_member(struct conf_contact_group *contact_group, const char *data)
{
	int ret = -1;
	struct conf_contact_member *member = NULL;
	AST_DECLARE_APP_ARGS(member_args,
		AST_APP_ARG(number);
		AST_APP_ARG(name);
	);
	char *parse = NULL;

	if (!contact_group || ast_strlen_zero(data)) {
		return -1;
	}

	parse = ast_strdupa(data);
	if (parse) {
		AST_STANDARD_APP_ARGS(member_args, parse);
		if (member_args.argc == 2) {
			member = alloc_contact_member(member_args.name, member_args.number);
			if (member) {
				AST_LIST_INSERT_TAIL(&contact_group->member_list, member, entry);
				ret = 0;
			}
		}
	}

	return ret;
}

static struct conf_contact_group *load_conf_contact_group(struct ast_config *cfg, const char *name)
{
	struct ast_variable *var = NULL;
	struct conf_contact_group *contact_group = NULL;

	if (!cfg || ast_strlen_zero(name)) {
		return NULL;
	}

	if (!(contact_group = alloc_contact_group(name))) {
		ast_log(LOG_WARNING, "Fail to alloc a contact group for '%s'", name);
		return NULL;
	}

	for (var = ast_variable_browse(cfg, name); var; var = var->next) {
		if (!strcasecmp(var->name, "pin")) {
			ast_string_field_set(contact_group, pin, var->value);
		} else if (!strcasecmp(var->name, "prompt")) {
			ast_string_field_set(contact_group, prompt, var->value);
		} else if (!strcasecmp(var->name, "member")) {
			load_conf_contact_member(contact_group, var->value);
		}
	}

	return contact_group;
}

static int load_conf_contact_groups_conf(int reload)
{
	struct ao2_iterator* it_groups = NULL;
	struct ast_config *cfg = NULL;
	char *cat = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct conf_contact_group *new_contact_group = NULL;
	struct conf_contact_group *contact_group = NULL;

	if (!(cfg = ast_config_load("conference_contact_groups.conf", config_flags))) {
		ast_log(LOG_NOTICE, "No contact groups config file (conference_contact_groups.conf), so no contact groups\n");
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Config file conference_contact_groups.conf is in an invalid format. Aborting.\n");
		return -1;
	}

	ao2_lock(conference_contact_groups);
	/*Retrieve last contact groups*/
	if (!(it_groups = ao2_callback(conference_contact_groups, OBJ_MULTIPLE|OBJ_UNLINK, NULL, NULL))) {
		ast_log(LOG_ERROR, "Unable to create iterator for contact groups container.\n");
		ao2_unlock(conference_contact_groups);
		ast_config_destroy(cfg);
		return -1;
	}

	/* Chug through config file */
	cat = NULL;
	while ((cat = ast_category_browse(cfg, cat)) ) {
		new_contact_group = load_conf_contact_group(cfg, cat);
		if (!new_contact_group) {
			continue;
		}

		contact_group = ao2_callback(it_groups->c, OBJ_UNLINK|OBJ_POINTER, conf_contact_groups_cmp_cb, new_contact_group);
		if (contact_group) {
			/*Update data from last contact group*/
			new_contact_group->inuse = contact_group->inuse;
			ast_string_field_set(new_contact_group, conference, contact_group->conference);
			ao2_cleanup(contact_group);
			contact_group = NULL;
		}
		ao2_link(conference_contact_groups, new_contact_group);
		ao2_cleanup(new_contact_group);
	}
	/*Destroy the last contact groups*/
	ao2_iterator_destroy(it_groups);
	ao2_unlock(conference_contact_groups);

	ast_config_destroy(cfg);
	return 0;
}

static int conf_contact_group_update_caller_name(struct confbridge_user *user)
{
	RAII_VAR(struct conf_contact_group *, contact_group, NULL, ao2_cleanup);
	struct conf_contact_member *contact_member = NULL;
	const char *contact_group_name = NULL;

	contact_group_name = pbx_builtin_getvar_helper(user->chan, "CONTACT_GROUP");
	if (ast_strlen_zero(contact_group_name)) {
		return 0;
	}

	contact_group = find_contact_group_by_name(conference_contact_groups, contact_group_name);
	if (!contact_group) {
		return -1;
	}

	ao2_lock(contact_group);
	AST_LIST_TRAVERSE(&contact_group->member_list, contact_member, entry) {
		if (!strcmp(contact_member->number, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, ""))
			&& !ast_strlen_zero(contact_member->name)) {
			char *func_callerid = NULL;
			if ((func_callerid = ast_module_helper("", "func_callerid", 0, 0, 0, 0))) {
				ast_free(func_callerid);
				ast_func_write(user->chan, "CALLERID(name)", contact_member->name);
			}
			break;
		}
	}
	ao2_unlock(contact_group);

	return 0;
}

#ifdef HAVE_UNIXODBC
static int conf_report_get_call_count(void)
{
	char cmd[512] = {0};
	int count = 0;
	SQLHSTMT stmt;
	struct odbc_obj *conf_obj = NULL;
	struct generic_prepare_struct gps;

	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return count;
	}

	snprintf(cmd, sizeof(cmd), "SELECT count FROM report_stat");
	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256];
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			count = atoi(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return count;
}

static int conf_report_update_call_count(void)
{
	char cmd[512] = {0};
	int ret = -1;
	struct odbc_obj *conf_obj = NULL;

	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	SQLHSTMT stmt;
	struct generic_prepare_struct gps = { .sql = cmd};
	snprintf(cmd, sizeof(cmd), "UPDATE report_stat SET count=%d", count);

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		ret = 0;
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}

	if (ast_ha_permit_send()){
		ast_ha_send_conf_report_sql(cmd);
	}

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return ret;
}

static const char *conf_report_get_user_role(struct confbridge_user *user)
{
	if (ast_test_flag(&user->u_profile, USER_OPT_ADMIN)) {
		return "ADMIN";
	} else if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB)) {
		return "BRIDGE";
	} else {
		return "USER";
	}
}

static const char *conf_report_get_user_join_way(struct confbridge_user *user)
{
	if (ast_test_flag(&user->u_profile, USER_OPT_MARKED_MCB)) {
		return "BRIDGE";
	} else if (!strcasecmp(ast_channel_tech(user->chan)->type, "Local") && (NULL == pbx_builtin_getvar_helper(user->chan,"DISA"))) {
		return "INVITE";
	} else if (user->chan && (ast_true(pbx_builtin_getvar_helper(user->chan, "JOIN_INVITE")) || ast_true(pbx_builtin_getvar_helper(user->chan, "CONF_SCHEDULE_INVITE")))) {
		return "INVITE";
	} else {
		return "CALLIN";
	}
}

static int conf_report_db_write(char *cmd)
{
	int result = 0;
	struct odbc_obj* conf_obj = NULL;

	if (ast_strlen_zero(cmd)) {
		return -1;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	SQLHSTMT stmt;
	struct generic_prepare_struct gps = { .sql = cmd};

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_WARNING, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}

	if (ast_ha_permit_send()){
		ast_ha_send_conf_report_sql(cmd);
	}

	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return result;
}

static int conf_record_db_write(char *cmd)
{
	int result = 0;
	struct odbc_obj* record_obj = NULL;

	if (ast_strlen_zero(cmd)) {
		return -1;
	}

	record_obj = ast_core_odbc_request_obj(RECORDING_FILES_DB, 0);
	if (!record_obj) {
		ast_log(AST_LOG_WARNING, "[conf_record_db_write] Failed to obtain database object for '%s'!\n", RECORDING_FILES_DB);
		return -1;
	}

	SQLHSTMT stmt;
	struct generic_prepare_struct gps = { .sql = cmd};

	stmt = ast_core_odbc_prepare_and_execute(record_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_WARNING, "[conf_record_db_write] SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}

	if (ast_ha_permit_send()){
		ast_ha_send_conf_report_sql(cmd);
	}

	ast_log(LOG_DEBUG, "[conf_record_db_write] SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(record_obj);
	record_obj = NULL;
	return result;
}


static int conf_report_get_user_record(const char *number, int conf_id, const char *type)
{
	char cmd[512]={0};
	int id = -1;
	struct odbc_obj *conf_obj = NULL;

	if (ast_strlen_zero(number) || ast_strlen_zero(type)) {
		return -1;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	if (!strcasecmp(type,"members")) {
		snprintf(cmd, sizeof(cmd), "SELECT conf_id FROM members WHERE conf_id=%d AND extension='%s'", conf_id, number);
	} else if (!strcasecmp(type,"conferences")){
		snprintf(cmd, sizeof(cmd), "SELECT id FROM conferences WHERE id=%d AND extension='%s'", conf_id, number);
	}
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256];
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			id = atoi(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return id;
}

static int conf_record_get_file_record(const char *filename) {
	char cmd[512]={0};
	int count = 0;
	struct odbc_obj *record_obj = NULL;

	if (ast_strlen_zero(filename)) {
		return 0;
	}

	record_obj = ast_core_odbc_request_obj(RECORDING_FILES_DB, 0);
	if (!record_obj) {
		ast_log(AST_LOG_WARNING, "[conf_record_get_file_record] Failed to obtain database object for '%s'!\n", RECORDING_FILES_DB);
		return 0;
	}

	snprintf(cmd, sizeof(cmd), "SELECT count(*) AS cnt FROM recording_files WHERE file_name = '%s';", filename);

	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(record_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "[conf_record_get_file_record] SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256];
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "[conf_record_get_file_record] SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			count = atoi(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "[conf_record_get_file_record] SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(record_obj);
	record_obj = NULL;
	return count;
}

static char* conf_report_get_one_varchar(char *cmd)
{
	char *result = NULL;
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;
	struct odbc_obj *conf_obj = NULL;

	if (ast_strlen_zero(cmd)) {
		return NULL;
	}

	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return NULL;
	}

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256];
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			result = tmp;
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return result;
}

static int conf_report_get_user_by_number_and_device(const char *number, const char *device_type, int conf_id)
{
	char cmd[512]={0};
	int id = -1;
	struct odbc_obj *conf_obj = NULL;

	snprintf(cmd, sizeof(cmd), "SELECT conf_id FROM members WHERE conf_id=%d AND extension='%s' AND device_type='%s' AND start_time!='0000-00-00 00:00:00'", conf_id, number, device_type);

	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		char tmp[256];
		memset(tmp, 0, sizeof(tmp));
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_WARNING, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_CHAR, tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			id = atoi(tmp);
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return id;
}


static int conf_report_get_schedule_logindex(const char *conf_num, const char *reservation_id, const char *utc_start_time)
{
	int log_index = -1;
	char cmd[1024] = {0};
	struct odbc_obj *conf_obj = NULL;

	if (ast_strlen_zero(conf_num) || ast_strlen_zero(reservation_id)) {
		ast_log(LOG_DEBUG, "conf_num or reservation_id null!");
		return -1;
	}
	conf_obj = ast_core_odbc_request_obj(CONF_LOG_DB, 0);
	if (!conf_obj) {
		ast_log(AST_LOG_WARNING, "Failed to obtain database object for '%s'!\n", CONF_LOG_DB);
		return -1;
	}

	if (ast_strlen_zero(utc_start_time)) {
		snprintf(cmd, sizeof(cmd), "SELECT log_index FROM multimedia_conference_reservation_log WHERE reservation_id='%s' AND conf_number='%s' LIMIT 1", reservation_id, conf_num);
	} else {
		snprintf(cmd, sizeof(cmd), "SELECT log_index FROM multimedia_conference_reservation_log WHERE reservation_id='%s' AND conf_number='%s' AND utc_start_time='%s' LIMIT 1", reservation_id, conf_num, utc_start_time);
	}
	SQLHSTMT stmt;
	struct generic_prepare_struct gps;

	gps.sql = cmd;
	gps.argc = 0;
	gps.argv = NULL;

	stmt = ast_core_odbc_prepare_and_execute(conf_obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto end_clean;
	} else {
		int tmp = 0;
		int res = SQLFetch(stmt);
		if (!SQL_SUCCEEDED(res)) {
			ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
			SQLFreeHandle(SQL_HANDLE_STMT, stmt);
			goto end_clean;
		}
		res = SQLGetData(stmt, 1, SQL_INTEGER, &tmp, sizeof(tmp), NULL);
		if (SQL_SUCCEEDED(res)) {
			log_index = tmp;
		}
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
	}
	ast_log(LOG_DEBUG, "SQL Execute success!\n[%s]\n\n", cmd);

end_clean:
	ast_core_odbc_release_obj(conf_obj);
	conf_obj = NULL;
	return log_index;
}

static int conf_report_record_schedule_user(struct confbridge_user *user)
{
	int res = 0;
	int log_index = 0;
	char *sql_cmd = NULL;
	char *cid_number = NULL;
	char member_exten[AST_MAX_EXTENSION] = {0};
	struct confbridge_conference *conference = NULL;

	if (!user || !user->chan) {
		ast_log(LOG_DEBUG, "user or conf_obj null!");
		return -1;
	}
	conference = user->conference;
	if (!conference) {
		ast_log(LOG_DEBUG, "conference null!");
		return -1;
	}
	if (ast_strlen_zero(conference->schedule_id_processing)) {
		ast_log(LOG_DEBUG, "conference may not schedule!");
		return -1;
	}

	log_index = conf_report_get_schedule_logindex(conference->name, conference->schedule_id_processing, conference->schedule_utc_start_time);
	if (log_index <= 0) {
		ast_log(LOG_DEBUG, "can't get log_index,no need record!");
		res = -1;
		goto clean;
	}

	cid_number = S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "");
	conf_get_exten_by_channel(ast_channel_name(user->chan), member_exten, AST_MAX_EXTENSION - 1);
	if (ast_strlen_zero(member_exten) || !strncasecmp(member_exten, "trunk", 5)) {
		ast_copy_string(member_exten, cid_number, AST_MAX_EXTENSION - 1);
	}

	ast_asprintf(&sql_cmd, "INSERT INTO multimedia_conference_reservation_members_log(log_index,reservation_id,member_extension,member_type) SELECT %d, '%s','%s',0 FROM DUAL WHERE NOT EXISTS(SELECT member_extension FROM multimedia_conference_reservation_members_log WHERE log_index=%d AND reservation_id='%s' AND member_extension='%s')",
				log_index, conference->schedule_id_processing, member_exten, log_index, conference->schedule_id_processing, member_exten);

	conf_report_db_write(sql_cmd);

clean:
	safe_ast_free(sql_cmd);
	return res;
}

static int conf_report_db_write_user(struct confbridge_user *user)
{
	struct confbridge_conference *conference = NULL;
	char *cmd = NULL;
	char strStartTime[80] = {0};
	char strEndTime[80] = {0};
	char user_agent[AST_USER_AGENT_LENGTH] = {0};
	int id = -1;
	struct ast_tm timeresult;
	char * name_tmp = NULL;
	char strUtcStartTime[80] = {0};
	char strUtcEndTime[80] = {0};
	time_t timep;
	struct tm tm_time;

	if (!user) {
		return -1;
	}

	conference = user->conference;

	ast_localtime(&user->report_record.start, &timeresult, NULL);
	ast_strftime(strStartTime, sizeof(strStartTime), DATE_FORMAT, &timeresult);
	ast_localtime(&user->report_record.end, &timeresult, NULL);
	ast_strftime(strEndTime, sizeof(strEndTime), DATE_FORMAT, &timeresult);

	strptime(strStartTime, "%Y-%m-%d %H:%M:%S", &tm_time);
	timep = mktime(&tm_time);
	gmtime_r(&timep,&tm_time);
	strftime(strUtcStartTime, sizeof(strUtcStartTime)-1, "%Y-%m-%d %H:%M:%S", &tm_time);
	strptime(strEndTime, "%Y-%m-%d %H:%M:%S", &tm_time);
	timep = mktime(&tm_time);
	gmtime_r(&timep,&tm_time);
	strftime(strUtcEndTime, sizeof(strUtcEndTime)-1, "%Y-%m-%d %H:%M:%S", &tm_time);

	/*Added for Bug 163451*/
	if (user->report_record.name && strstr(user->report_record.name, "'")) {
		name_tmp = str_replace(user->report_record.name, "'", "''");
		ast_log(LOG_DEBUG, "name is %s after replace is %s\n", user->report_record.name, name_tmp);
	}
	conf_get_useragent(user->chan, user_agent, sizeof(user_agent));
	const char *device_type = get_confuser_device_type(user_agent);
	/*get real exten*/
	if (user->report_record.number){    /*Mod by weidu, for 312449*/
		char real_exten[AST_MAX_EXTENSION] = {0};
		if (user->chan) {
			conf_get_exten_by_channel(ast_channel_name(user->chan), real_exten, AST_MAX_EXTENSION - 1);
			if (ast_strlen_zero(real_exten) || !strncasecmp(real_exten, "trunk", 5)) {
				ast_copy_string(real_exten, user->report_record.number, sizeof(real_exten));
			}
		}
		//id = conf_report_get_user_record(user->report_record.number, conference->report_record.id, "members");
		id = conf_report_get_user_by_number_and_device(user->report_record.number, device_type, conference->report_record.id);
		if (id == -1) {
			ast_asprintf(&cmd, "INSERT INTO members (conf_id,extension,name,group_name,role,join_way,invite_time,start_time,end_time,status,device_type,real_exten,utc_start_time,utc_end_time,related_exten) "
				"VALUES (%d,'%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s')", conference->report_record.id, user->report_record.number, name_tmp?name_tmp:S_OR(user->report_record.name, ""),
				S_OR(user->report_record.group_name, ""), conf_report_get_user_role(user), conf_report_get_user_join_way(user), "0000-00-00 00:00:00", strStartTime, strEndTime, "ANSWER", device_type?device_type:"Other", !ast_strlen_zero(real_exten)?real_exten:"", strUtcStartTime, strUtcEndTime, !ast_strlen_zero(user->related_ext)?user->related_ext:real_exten);
		} else {
			ast_asprintf(&cmd, "UPDATE members SET end_time='%s',status='%s',utc_end_time='%s' WHERE conf_id=%d AND device_type='%s'", strEndTime, "ANSWER", strUtcEndTime, id, device_type);
		}
		conf_report_db_write(cmd);
	}
	conf_report_record_schedule_user(user);

	safe_ast_free(cmd);
	safe_ast_free(name_tmp);
	return 0;
}

static int conf_report_db_write_conference(struct confbridge_conference *conference)
{
	char *cmd = NULL;
	char *cmd_record = NULL;
	char strStartTime[80] = {0};
	char strEndTime[80] = {0};
	char utcStartTime[80] = {0};
	char utcEndTime[80] = {0};
	char family[64] = {0};
	char is_random_now[32] = {0};
	int id = -1;
	char type[32] = "audio";
	struct ast_tm timeresult;
	time_t timep;
	time_t start_timep;
	struct tm tm_time;
	char *public_subject = NULL;
	int need_record_key = 0;
	int has_record = 0;

	if (!conference) {
		return -1;
	}

	ast_localtime(&conference->report_record.start, &timeresult, NULL);
	ast_strftime(strStartTime, sizeof(strStartTime), DATE_FORMAT, &timeresult);
	ast_localtime(&conference->report_record.end, &timeresult, NULL);
	ast_strftime(strEndTime, sizeof(strEndTime), DATE_FORMAT, &timeresult);
	id = conf_report_get_user_record(conference->name, conference->report_record.id, "conferences");

	strptime(strStartTime, "%Y-%m-%d %H:%M:%S", &tm_time);
	timep = mktime(&tm_time);
	start_timep = timep;
	gmtime_r(&timep,&tm_time);
	strftime( utcStartTime, sizeof(utcStartTime)-1, "%Y-%m-%d %H:%M:%S", &tm_time);
	strptime(strEndTime, "%Y-%m-%d %H:%M:%S", &tm_time);
	timep = mktime(&tm_time);
	gmtime_r(&timep,&tm_time);
	strftime( utcEndTime, sizeof(utcEndTime)-1, "%Y-%m-%d %H:%M:%S", &tm_time);

	if (conference->b_profile.is_random_conf/* || conference->b_profile.n_way_conf*/) {
		strcpy(type, "random");
		snprintf(family, sizeof(family), "NWAY_CONF_STATUS/%s", conference->name);
		if (!ast_db_get(family, "IS_RANDOM_CONF", is_random_now, sizeof(is_random_now)) && !ast_strlen_zero(is_random_now) && !strcasecmp(is_random_now, "yes")) {
			ast_log(LOG_DEBUG,"The conference is random now meetme!");
		} else {
			ast_log(LOG_DEBUG,"The conference is random schedule meetme!");
			strcat(type, "_sche");
		}
	} else if(conference->b_profile.n_way_conf){
		strcpy(type, "random");
	} else if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		strcpy(type,"video");
		if (conference->schedule_id_processing) {
			strcat(type, "_sche");
		}
	} else if (conference->schedule_id_processing) {
		strcat(type, "_sche");
	}
	/*get public conference room subject*/
	if (ast_strlen_zero(conference->schedule_id_processing) && !ast_strlen_zero(conference->schedule_subject)) {
		public_subject = str_replace(conference->schedule_subject, "'", "''");
	}

#ifdef BUILD_ON_CLOUD_PBX
	need_record_key = 1;
#else
	char settingss[128] = { 0 };
	int settingsi = 0x0;
	ast_safe_nvram_get("data_crypt_settings", settingss, sizeof(settingss) - 1);
	settingsi = atoi(settingss);

	need_record_key = !ast_strlen_zero(ast_str_buffer(conference->record_files)) 
				? ast_btst(settingsi,AST_RECORD_AUDIO_FILE) 
				: ast_btst(settingsi,AST_RECORD_VIDEO_FILE);
#endif // BUILD_ON_CLOUD_PBX


	/** NOTE: video files only recorded in conferences and without 'RecordingFileInsert' ami event
			 and share file key among all video files. so here update into recording_files immediately */
	if (id == -1) {
		ast_asprintf(&cmd, "INSERT INTO conferences (id, extension, type, start_time, utc_start_time, "
									"feature_id, reservation_id, subject) "
									"VALUES (%d,'%s','%s','%s','%s','%s','%s','%s')",
					conference->report_record.id, conference->name, type, strStartTime, utcStartTime,
					conference->feature_id, ast_strlen_zero(conference->schedule_id_processing) ? "" : conference->schedule_id_processing, public_subject ? public_subject : "");
	} else {		
		char video_files_uuid[256];
		char *video_files_key = NULL;
		char *video_files_key_check = NULL;
		
		if (strlen(ast_str_buffer(conference->video_files_uuid))) {
			snprintf(video_files_uuid, sizeof(video_files_uuid), "%s", ast_str_buffer(conference->video_files_uuid));
			ast_log(LOG_DEBUG, "[conf_report_db_write_conference] video_files_uuid:%s", video_files_uuid);
			
			ucm_secret_encrypt(video_files_uuid, &video_files_key);
			if (strlen(video_files_uuid) == 32) {
				ucm_secret_encrypt_with_key("grandstream", video_files_uuid, &video_files_key_check);
			}
			has_record = 1;
		}

		ast_asprintf(&cmd, "UPDATE conferences "
							"SET end_time='%s',utc_end_time='%s',"
							"record_files='%s',record_files_desc='%s',"
							"video_files='%s',video_files_desc='%s',video_files_key='%s',video_files_key_check='%s' "
							"WHERE id=%d",
				strEndTime, 
				utcEndTime , 
				ast_str_buffer(conference->record_files), 
				ast_str_buffer(conference->record_files_desc), 
				ast_str_buffer(conference->video_files), 
				ast_str_buffer(conference->video_files_desc),
				(video_files_key && strlen(video_files_key)) ? video_files_key : "",
				(video_files_key_check && strlen(video_files_key_check)) ? video_files_key_check : "",
				id);
		
		if (video_files_key) {
			free(video_files_key);
		}
		if (video_files_key_check) {
			free(video_files_key_check);
		}

		//
		if (strlen(ast_str_buffer(conference->record_files)) && strlen(ast_str_buffer(conference->record_files_uuid))) {
			char record_files[4096] = "";
			char record_files_uuid[4096] = "";

			char *sep = "@";
    		char *record_files_token = NULL;
			char *record_files_uuid_token = NULL;
    		char *record_files_sub = NULL;
			char *record_files_uuid_sub = NULL;

			char *file_key = NULL;
			char *file_key_check = NULL;

			snprintf(record_files, sizeof(record_files), "%s", ast_str_buffer(conference->record_files));
			snprintf(record_files_uuid, sizeof(record_files_uuid), "%s", ast_str_buffer(conference->record_files_uuid));

			record_files_token = strtok_r(record_files, sep, &record_files_sub);
			record_files_uuid_token = strtok_r(record_files_uuid, sep, &record_files_uuid_sub);

			while (record_files_token && record_files_uuid_token) {
				char record_files_without_dir[256] = "";
				char *pos = strstr(record_files_token, "/");
				if (pos) {
					snprintf(record_files_without_dir, sizeof(record_files_without_dir), "%s", ++pos);
				}
				
				ucm_secret_encrypt(record_files_uuid_token, &file_key);
				if (strlen(record_files_uuid_token) == 32) {
					ucm_secret_encrypt_with_key("grandstream", record_files_uuid_token, &file_key_check);
				}

				// This happen when conferences ended, while RecordingFileInsert ami event is reported every time 
				// stop record.
				// if end meeting during recording, cannot make sure pbxmid will finish earlier than here, so add check
				// first 
				int cnt = conf_record_get_file_record(record_files_without_dir);

				if (!cnt) {
					ast_asprintf(&cmd_record, 
						"REPLACE INTO recording_files(file_name, path_type, feature_type, file_date, file_key, file_key_check) "
					 	"VALUES('%s', '%s', '%s', '%s', '%s', '%s');",
					 	strlen(record_files_without_dir) ? record_files_without_dir : record_files_token,
					 	"",
					 	"",
					 	"",
						(file_key && strlen(file_key)) ? file_key : "",
						(file_key_check && strlen(file_key_check)) ? file_key_check : ""
					);
					has_record = 2;
				} else {
					ast_asprintf(&cmd_record, 
						"UPDATE recording_files SET file_key = '%s', file_key_check = '%s' WHERE file_name = '%s';", 
						(file_key && strlen(file_key)) ? file_key : "",
						(file_key_check && strlen(file_key_check)) ? file_key_check : "",
						strlen(record_files_without_dir) ? record_files_without_dir : record_files_token
					);
				}

				if (cmd_record) {
					ast_log(LOG_DEBUG, "[conf_report_db_write_conference] SQL %s", cmd_record);
					conf_record_db_write(cmd_record);

					ast_free(cmd_record);
					cmd_record = NULL;
				} else {
					ast_log(LOG_ERROR, "[conf_report_db_write_conference] Insufficient memory!");
				}

				if (file_key) {
					ast_free(file_key);
					file_key = NULL;
				}
				if (file_key_check) {
					ast_free(file_key_check);
					file_key_check = NULL;
				}

				record_files_token = strtok_r(NULL, sep, &record_files_sub);
				record_files_uuid_token = strtok_r(NULL, sep, &record_files_uuid_sub);
			}
		}
		if (!need_record_key && has_record) {
#ifdef BUILD_ON_CLOUD_PBX
			char sqlprefix[AST_DB_NAME_MAX_LEN] = {0};
			ast_safe_nvram_get("sqlprefix", sqlprefix, sizeof(sqlprefix));
			if (!ast_strlen_zero(sqlprefix)) {
				ast_asprintf(&cmd_record, "UPDATE %sucm_config.data_encrypt_details SET safety_factor='weak',status='' WHERE data_type='%s' AND safety_factor='strong' AND status='completed' AND device='CloudUCM'", sqlprefix, has_record > 1 ? "record_file" : "video_file");
			}
#else
			ast_asprintf(&cmd_record, "UPDATE ucm_config.data_encrypt_details SET safety_factor='weak',status='' WHERE data_type='%s' AND safety_factor='strong' AND status='completed' AND device='UCM'", has_record > 1 ? "record_file" : "video_file");
#endif
			if (cmd_record) {
				ast_log(LOG_DEBUG, "[conf_report_db_write_conference] SQL %s", cmd_record);
				conf_record_db_write(cmd_record);
				ast_free(cmd_record);
				cmd_record = NULL;
			}
		}
	}

	if (cmd) {
		ast_log(LOG_DEBUG, "SQL Execute return id!\n[%d]\n\ncmd(%ld):\n%s", id, strlen(cmd), cmd);
		conf_report_db_write(cmd);

		ast_free(cmd);
		cmd = NULL;
	} else {
		ast_log(LOG_ERROR, "Insufficient memory!");
	}

	if (cmd_record) {
		ast_free(cmd_record);
		cmd_record = NULL;
	}

	conference_write_statistics(conference,id,start_timep,utcStartTime,utcEndTime);
	safe_ast_free(public_subject);

	return 0;
}

static int conference_write_statistics(struct confbridge_conference *conference, int id, time_t start_timep, char *utcStartTime, char *utcEndTime)
{
	/*Statistics of conference information other than audio conference.*/
	char cmd[512] = {0};
	char is_random_now[32] = {0};
	char family[64] = {0};
	char *share_screen = NULL;
	int is_random = 0;
	int is_schedule = 0;
	if (!conference) {
		return -1;
	}
	if (conference->b_profile.is_random_conf) {
		is_random = 1;
		snprintf(family, sizeof(family), "NWAY_CONF_STATUS/%s", conference->name);
		if (ast_db_get(family, "IS_RANDOM_CONF", is_random_now, sizeof(is_random_now)) && ast_strlen_zero(is_random_now) && strcasecmp(is_random_now, "yes")) {
			is_schedule = 1;
		}
	}else if (conference->b_profile.n_way_conf) {
		is_random = 1;
	}else if (conference->schedule_id_processing) {
		is_schedule = 1;
	}

	if (id == -1) {
		snprintf(cmd, sizeof(cmd), "INSERT INTO conference_statistics(id, conf_num, max_user, utc_start_time, is_random, is_schedule) VALUES (%d,'%s',1,'%s','%s','%s')",
			conference->report_record.id, conference->name, utcStartTime, is_random?"yes":"no", is_schedule?"yes":"no");
	} else {
		snprintf(cmd, sizeof(cmd), "SELECT action_name FROM %s.ast_operation_log WHERE feature_id='%s' AND action_name='OPEN SHARE'", ast_strlen_zero(s_dbOperationLog) ? "operation_log" : s_dbOperationLog, conference->feature_id);
		share_screen = conf_report_get_one_varchar(cmd);
		snprintf(cmd, sizeof(cmd), "UPDATE conference_statistics SET camera_num=%d,share_screen='%s',max_user=%d,duration=%ld,utc_end_time='%s' WHERE id=%d",
			conference->peak_camera?conference->peak_camera:0, share_screen?"yes":"no", conference->peak_users , time(NULL)-start_timep, utcEndTime , id);
	}
	conf_report_db_write(cmd);

	return 0;
}

static void destroy_conf_user_report_record(struct conf_report_user_record *report_record)
{
	if (report_record) {
		if (report_record->number) {
			ast_free((char *)report_record->number);
			report_record->number = NULL;
		}
		if (report_record->name) {
			ast_free((char *)report_record->name);
			report_record->name = NULL;
		}
		if (report_record->group_name) {
			ast_free((char *)report_record->group_name);
			report_record->group_name = NULL;
		}
	}
}

static struct confbridge_user *find_conf_user_by_number(struct confbridge_conference *conference, const char *number)
{
	struct confbridge_user *user = NULL;
	const char *user_number = NULL;

	if (!conference || ast_strlen_zero(number)) {
		return NULL;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, user, list) {
		user_number = S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "");
		if (!strcmp(user_number, number)) {
			ao2_unlock(conference);
			return user;
		}
	}

	AST_LIST_TRAVERSE(&conference->waiting_list, user, list) {
		user_number = S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "");
		if (!strcmp(user_number, number)) {
			ao2_unlock(conference);
			return user;
		}
	}
	ao2_unlock(conference);

	return user;
}

/*! \brief ConfBridgeLog application */
static int confbridge_log_exec(struct ast_channel *chan, const char *data)
{
	ast_log(LOG_DEBUG, "confbridge_log_exec in\n\n");

	char *parse = NULL;
	struct confbridge_conference *conference = NULL;
	struct confbridge_user *user = NULL;
	const char *device_type = NULL;
	char cmd[512] = {0};
	char user_agent[AST_USER_AGENT_LENGTH] = {0};
	int id = -1;
	char *name_tmp = NULL;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf_name);
		AST_APP_ARG(number);
		AST_APP_ARG(name);
		AST_APP_ARG(way);
		AST_APP_ARG(status);
		AST_APP_ARG(group_name);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "ConfBridgeLog requires arguments (confname,name,number,way,status,group_name)\n");
		return 0;
	}

	parse = ast_strdupa(data);

	AST_STANDARD_APP_ARGS(args, parse);

	if (chan) {
		if (ast_strlen_zero(args.conf_name)) {
			args.conf_name = (char *)pbx_builtin_getvar_helper(chan, "CONF_ROOM");
		}
		if (ast_strlen_zero(args.number)) {
			args.number = S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "");
		}
		if (ast_strlen_zero(args.name)) {
			args.name = S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, args.number);
		}
		if (ast_strlen_zero(args.group_name)) {
			args.group_name = (char *)pbx_builtin_getvar_helper(chan, "CONTACT_GROUP");
		}
		conf_get_useragent(chan, user_agent, sizeof(user_agent));
		device_type = get_confuser_device_type(user_agent);
	}

	if (ast_strlen_zero(args.conf_name) || ast_strlen_zero(args.number)
	    || ast_strlen_zero(args.status) || ast_strlen_zero(args.way)) {
		ast_log(LOG_WARNING, "ConfBridgeLog requires arguments (confname,name,number,way,status,group_name)\n");
		return 0;
	}

	/*Convert status into BUSY/NOANSWER/REJECT/FAILURE*/
	if (!strcmp(args.status, "CONGESTION")) {
		args.status = "BUSY";
	} else if (!strcmp(args.status, "CHANUNAVAIL") || !strcmp(args.status, "CANCEL")) {
		args.status = "FAILURE";
	}

	ao2_lock(conference_bridges);

	/* Attempt to find an existing conference bridge */
	conference = ao2_find(conference_bridges, args.conf_name, OBJ_KEY);
	if (!conference) {
		ast_log(LOG_WARNING, "Couldn't find the conference %s\n", args.conf_name);
		ao2_unlock(conference_bridges);
		return 0;
	}

	ao2_unlock(conference_bridges);

	if (!conference->is_conference_service) {
		ao2_cleanup(conference);
		return 0;
	}

	/*Check if the extension is already in this conference.*/
	user = find_conf_user_by_number(conference, args.number);
	if (user) {
		ast_log(LOG_DEBUG, "The user '%s' is alread in the conference %s\n", args.number, args.conf_name);
		ao2_cleanup(conference);
		return 0;
	}
	char family[256] = {0};
	char conf_subject[256] = {0};
	snprintf(family, sizeof(family), "CONF_STATUS/%s", conference->name);

	/*Added for Bug 163451*/
	if (args.name && strstr(args.name, "'")) {
		name_tmp = str_replace(args.name, "'", "''");
		ast_log(LOG_DEBUG, "name is %s after replace is %s\n", args.name, name_tmp);
	}
	if (args.name) {
		safe_ast_free(name_tmp);
		if (!strncmp(args.name, "Conference invitation", strlen("Conference invitation"))) {
			name_tmp = ast_strdup(args.number);
			ast_log(LOG_DEBUG, "name is %s, , so use number\n", args.name);
		} else if (!ast_db_get(family, "CONF_NAME", conf_subject, sizeof(conf_subject)) && !strcmp(conf_subject, args.name)) {
			ast_log(LOG_DEBUG, "caller name is same to conference name %s, so use number", conf_subject);
			name_tmp = ast_strdup(args.number);
		} else if (conference->schedule_subject && !strcmp(args.name, conference->schedule_subject)) {
			ast_log(LOG_DEBUG, "caller name is same to conference subject %s, so use number", conference->schedule_subject);
			name_tmp = ast_strdup(args.number);
		}
	}
	/*get real exten*/
	char real_exten[AST_MAX_EXTENSION] = {0};
	const char *menu_invite = pbx_builtin_getvar_helper(chan, "CONF_MENU_CHAN");
	if (chan && ast_strlen_zero(menu_invite)) {
		conf_get_exten_by_channel(ast_channel_name(chan), real_exten, AST_MAX_EXTENSION - 1);
		if (ast_strlen_zero(real_exten) || !strncasecmp(real_exten, "trunk", 5)) {
			ast_copy_string(real_exten, args.number, sizeof(real_exten));
		} else if (!ast_strlen_zero(real_exten) && strcmp(real_exten, args.number)) {
			/*Bug Bug 218723*/
			safe_ast_free(name_tmp);
			args.number = real_exten;
			name_tmp = ast_strdup(real_exten);
		}
	} else {
		ast_copy_string(real_exten, args.number, sizeof(real_exten));
	}

	id = conf_report_get_user_record(args.number, conference->report_record.id, "members");
	if (id == -1) {
		snprintf(cmd, sizeof(cmd), "INSERT INTO members (conf_id,extension,name,group_name,role,join_way,invite_time,start_time,end_time,status,device_type,real_exten,related_exten)"
				"VALUES (%d,'%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s')", conference->report_record.id, args.number, name_tmp?name_tmp:S_OR(args.name, ""),
				S_OR(args.group_name, ""), "USER", args.way, "0000-00-00 00:00:00", "0000-00-00 00:00:00", "0000-00-00 00:00:00", args.status, device_type?device_type:"Other", !ast_strlen_zero(real_exten)?real_exten:"",!ast_strlen_zero(real_exten)?real_exten:"");
		/*bug 193324*/
		if ( 0 == strcmp(args.status,"BUSY") || 0 == strcmp(args.status,"NOANSWER"))
		{
			conf_report_db_write(cmd);
		}
	}

	ao2_cleanup(conference);
	safe_ast_free(name_tmp);

	return 0;
}

static int conf_get_useragent(struct ast_channel* chan, char* buf, size_t size)
{
	int ret = -1;

	if (!chan || !buf || size <= 0) {
		return ret;
	}

	if (ast_channel_tech(chan) && ast_channel_tech(chan)->get_session_ua) {
		ret = ast_channel_tech(chan)->get_session_ua(chan, buf, size);
	}

	return ret;
}

static int conf_get_sca_shareline(struct ast_channel* chan, char* buf, size_t size)
{
	int ret = -1;

	if (!chan || !buf || size <= 0) {
		return ret;
	}

	if (ast_channel_tech(chan) && ast_channel_tech(chan)->get_sca_sharedline) {
		ret = ast_channel_tech(chan)->get_sca_sharedline(chan, buf, size);
	}

	return ret;
}

#endif

static int confbridge_early_media(struct ast_channel *chan, const char *data)
{
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(action);
	);
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_confearlymedia);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.action)) {
		return -1;
	}

	ast_log(LOG_DEBUG, "'%s' executes conference '%s' early media, action[%s].\n", ast_channel_name(chan), args.conf, args.action);
	conference = conf_find_bridge(args.conf);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	if (conference->bridge->avs_bridge_id > 0) {
		if (!strcasecmp(args.action, "join")) {
			ast_avs_control(chan, AVS_CONTROL_MUTE, NULL, 0);
			ast_avs_join_bridge(chan, conference->bridge, conference->bridge->avs_bridge_id, conference->bridge->avs_bridge_tick);
			ast_channel_sfu_early_media_set(chan, 1);
		} else if (!strcasecmp(args.action, "leave")) {
			ast_avs_leave_bridge(chan, conference->bridge, conference->bridge->avs_bridge_id, conference->bridge->avs_bridge_tick);
			ast_avs_control(chan, AVS_CONTROL_UNMUTE, NULL, 0);
			ast_channel_sfu_early_media_set(chan, 0);
		}
	}
	ao2_unlock(conference);

	return 0;
}

static int confbridge_floor_contrl(struct ast_channel *chan, const char *data)
{
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
		AST_APP_ARG(audioshare);
	);
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *user = NULL;

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_conffloorcontrl);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.audioshare)) {
		return -1;
	}

	ast_log(LOG_DEBUG, "'%s' executes conference '%s' floor contrl, audioshare[%s].\n", ast_channel_name(chan), args.conf, args.audioshare);
	conference = conf_find_bridge(args.conf);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	user = find_active_user_by_channel(conference, ast_channel_name(chan));
	if (user) {
		if (user->muted) {
			ast_log(LOG_DEBUG, "user is muted!,mic mute control no need send to AVS.");
			if (ast_true(args.audioshare)) {
				conference->share_system_audio_chan = ast_strdup(ast_channel_name(chan));
			}
		} else {
			ast_debug(1, "User %s is %s: mic muted:%d.\n",
					ast_channel_name(chan), ast_true(args.audioshare) ? "audioshare" : "not audioshare", user->mic_muted);
			if (ast_true(args.audioshare)) {
				ast_avs_control(user->chan, AVS_CONTROL_UNMUTE, NULL, 0);
				conference->share_system_audio_chan = ast_strdup(ast_channel_name(chan));
			} else if (user->mic_muted && !ast_true(args.audioshare)) {
				ast_avs_control(user->chan, AVS_CONTROL_MUTE, NULL, 0);
			}
			if (!ast_true(args.audioshare)) {
				if (conference->share_system_audio_chan) {
					ast_free(conference->share_system_audio_chan);
					conference->share_system_audio_chan = NULL;
				}
			}
		}
	}
	ao2_unlock(conference);

	return 0;
}

/*Start added by zyxia for Bug 235579, call flip the page caller,need update the video track msid*/
static int  pageconf_change_caller(struct ast_channel *chan, const char *data)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct ao2_iterator iter;
	struct confbridge_user *cur_user = NULL;

	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(msid);
		AST_APP_ARG(new_chan_name);
	);

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_pageconf_change_caller);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "No data was provided to %s application.\n", app_pageconf_change_caller);
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.msid) ) {
		ast_log(LOG_WARNING, "No msid was provided to %s application.\n", app_pageconf_change_caller);
		return -1;
	}

	if (ast_strlen_zero(args.new_chan_name) ) {
		ast_log(LOG_WARNING, "No new_chan_name was provided to %s application.\n", app_pageconf_change_caller);
		return -1;
	}

	iter = ao2_iterator_init(conference_bridges, 0);
	while ((conference = ao2_iterator_next(&iter))) {
		ao2_lock(conference);
		if (conference->is_paging_conf && conference->paging_caller_chan
			&& ast_channel_name(conference->paging_caller_chan)
			&& !strcmp(ast_channel_name(conference->paging_caller_chan), ast_channel_name(chan))) {
				AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
					if(strcmp(ast_channel_name(chan),ast_channel_name(cur_user->chan))){//is not caller chan
						pbx_builtin_setvar_helper(cur_user->chan, "__PAGING_INTERCOM_CALLER_CHAN_NAME", args.new_chan_name);//added by zyxia for Bug 250509
						ast_indicate_data(cur_user->chan, AST_CONTROL_UPDATE_PAGE_MSID, args.msid, strlen(args.msid) + 1);}
				}
			ao2_unlock(conference);
			break;
		}
		ao2_unlock(conference);
	}
	ao2_iterator_destroy(&iter);

	return 0;
}
/*End added by zyxia for Bug 235579, call flip the page caller,need update the video track msid*/


/*Start added by zyxia for Bug 251489 , call flip the private intercom,2023.2.23*/
static int  privateintercom_callflip(struct ast_channel *chan, const char *data)
{
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);
	struct confbridge_user *cur_user = NULL;
	RAII_VAR(struct ast_bridge *, bridge, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, new_chan, NULL, ao2_cleanup);
	const char *page_conf_num = NULL;

	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(new_chan_name);
	);

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "No data was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.new_chan_name) ) {
		ast_log(LOG_WARNING, "No new_chan_name was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	new_chan = ast_channel_get_by_name(args.new_chan_name);
	if(!new_chan){
		ast_log(LOG_WARNING, "No chan was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	bridge = ast_channel_get_bridge(chan);
	if(!bridge){
		ast_log(LOG_WARNING, "No bridge was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	page_conf_num = pbx_builtin_getvar_helper(chan,"PAGE_CONF_NUM");
	if(!page_conf_num) {
		ast_log(LOG_WARNING, "No page_conf_num was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	conference = conf_find_bridge(page_conf_num);
	if(!conference) {
		ast_log(LOG_WARNING, "No conference was provided to %s application.\n", app_privateintercom_callflip);
		return -1;
	}

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if(!strcmp(ast_channel_name(chan),ast_channel_name(cur_user->chan)))
			continue;

		if (conference->paging_caller_chan && chan == conference->paging_caller_chan) {
			ast_log(LOG_DEBUG, "Channel %s subscribe Channel %s,confername:%s", ast_channel_name(cur_user->chan), ast_channel_name(new_chan),conference->name);
			ast_avs_subscribe_bridge(new_chan, cur_user->chan, bridge, bridge->avs_bridge_id, bridge->avs_bridge_tick);
			ast_avs_subscribe_bridge(cur_user->chan, new_chan, bridge, bridge->avs_bridge_id, bridge->avs_bridge_tick);
		} else {
			ast_log(LOG_DEBUG, "Channel %s subscribe Channel %s,confername:%s", ast_channel_name(new_chan), ast_channel_name(conference->paging_caller_chan),conference->name);
			ast_avs_subscribe_bridge(new_chan, conference->paging_caller_chan, bridge, bridge->avs_bridge_id,bridge->avs_bridge_tick);
			ast_avs_subscribe_bridge(conference->paging_caller_chan, new_chan, bridge, bridge->avs_bridge_id,bridge->avs_bridge_tick);
			break;
		}
	}

	ao2_unlock(conference);
	return 0;
}
/*End added by zyxia for Bug 251489 , call flip the private intercom,2023.2.23*/

static int confbridge_record_contrl(struct ast_channel *chan, const char *data)
{
	int res = 0;
	int recording = 0;
	char *parse = NULL;

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
	);
	RAII_VAR(struct confbridge_conference *, conference, NULL, ao2_cleanup);

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_confrecordcontrl);
		return -1;
	}

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf)) {
		return -1;
	}

	ast_log(LOG_DEBUG, "'%s' executes conference '%s' record contrl.\n", ast_channel_name(chan), args.conf);
	conference = conf_find_bridge(args.conf);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	recording = conf_is_recording(conference);
	ast_log(LOG_DEBUG, "recording flags: %d", recording);

	if (recording) {
		res = conf_stop_record(conference);
		if (!res) {
			send_record_status_event(conference, "audio", !recording);
		}
	}
	ao2_unlock(conference);

	return res;
}

#if 1//remove by Bug 252408
/*Start added by zyxia for Bug 242841,244581.2022.12.28*/
static void playPageVideoRecord_Callee(struct confbridge_conference *conference)
{
	struct confbridge_user *cur_user = NULL;
	ao2_lock(conference);

	AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
		if(!ast_channel_get_is_page_caller(cur_user->chan) && !cur_user->had_play_tone && !cur_user->is_playing_annoncement){
			ast_log(LOG_DEBUG, "playPageVideoRecord_Callee %s to channel\n", ast_channel_name(cur_user->chan));
			cur_user->had_play_tone = 1;//set it before play, Bug 258837
			//edit by zyxia for Bug 265467
			ast_indicate_data(cur_user->chan,AST_CONTROL_PLAY_PAGING_VIDEO_RECORD, conf_get_sound(CONF_SOUND_START_PAGING_VIDEO, cur_user->b_profile.sounds), strlen(conf_get_sound(CONF_SOUND_START_PAGING_VIDEO, cur_user->b_profile.sounds)));
			//play_prompt_to_user(cur_user, conf_get_sound(CONF_SOUND_START_PAGING_VIDEO, cur_user->b_profile.sounds));
		}
	}
	ao2_unlock(conference);

	return ;
}
/*End added by zyxia for Bug 242841,244581.2022.12.28*/
#endif
static int command_delete_auth_user(const char *conf_number, struct ast_channel *chan, const char *params)
{
	int res = 0;

	if (chan) {
		cache_conf_remove_user_by_channel_name(conf_number, ast_channel_name(chan));
	} else {
		char *parse = NULL;
		AST_DECLARE_APP_ARGS(args,
			AST_APP_ARG(extension);
			AST_APP_ARG(type);
		);

		parse = ast_strdupa(params);
		AST_STANDARD_APP_ARGS(args, parse);
		cache_conf_remove_user(conf_number, args.extension, args.type);
	}

	return res;
}

static int confbridge_command(struct ast_channel *chan, const char *data)
{
	char *parse = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(command);
		AST_APP_ARG(conf);
		AST_APP_ARG(params);
	);

	/* parse all the required arguments and make sure they exist. */
	if (ast_strlen_zero(data)) {
		return -1;
	}
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf) || ast_strlen_zero(args.command)) {
		return -1;
	}

	if (!strcasecmp(args.command, "DeleteAuthUser")) {
		command_delete_auth_user(args.conf, chan, args.params);
	} else if (!strcasecmp(args.command, "DeleteAuthUserWithoutChan")) {
		char *parse_params = NULL;
		AST_DECLARE_APP_ARGS(args_params,
			AST_APP_ARG(extension);
			AST_APP_ARG(type);
		);

		parse_params = ast_strdupa(args.params);
		AST_STANDARD_APP_ARGS(args_params, parse_params);

		cache_conf_remove_user(args.conf, args_params.extension, args_params.type);
	}

	return 0;
}

static void schedule_conference_remove(const char *conf_number, const char *reservation_id)
{
	struct schedule_conference *conference = NULL;

	if (ast_strlen_zero(conf_number) || (strcasecmp(conf_number, "all") && ast_strlen_zero(reservation_id))) {
		return;
	}

	ast_log(LOG_DEBUG, "Remove the schedule conference '%s' reservation_id '%s'.\n", conf_number, reservation_id);
	AST_LIST_LOCK(&schedule_conference_list);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&schedule_conference_list, conference, entry)
	{
		if (!strcasecmp(conf_number, "all") || (!strcmp(conference->name, conf_number) && !strcmp(conference->schedule_id, reservation_id))) {
			if (strcasecmp(conf_number, "all")) {
				conference_write_statistics_by_confschedule(conference);
			}
			if (conference->recycle_sched_id) {
				int ret = ast_sched_del(recycle_sched, conference->recycle_sched_id);
				conference->recycle_sched_id = 0;
				ast_log(LOG_DEBUG, "schedule conference sched '%s' is removed, ret %d.\n", conference->name, ret);
			}
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(conference);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&schedule_conference_list);
}

static void schedule_conference_destroy(void *obj)
{
	struct schedule_conference *conference = obj;
	struct schedule_attendees *user=  NULL;

	ast_log(LOG_DEBUG, "schedule conference '%s' is destroying.\n", conference->name);
	if (conference->recycle_sched_id) {
		int ret = ast_sched_del(recycle_sched, conference->recycle_sched_id);
		conference->recycle_sched_id = 0;
		ast_log(LOG_DEBUG, "schedule conference sched '%s' is removed, ret %d.\n", conference->name, ret);
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->schedule_attendee_list, user, entry){
		AST_LIST_REMOVE_CURRENT(entry);
		ao2_cleanup(user);
	}
	AST_LIST_TRAVERSE_SAFE_END;
}

/*
For instant meetings or other scheduled meetings, must clear the remaining information first.
*/
static void schedule_conference_error_recovery_detect(struct confbridge_conference *conference)
{
	struct schedule_conference *conference_rm = NULL;

	if (!conference) {
		ast_log(LOG_DEBUG, "conference null!");
		return;
	}

	AST_LIST_LOCK(&schedule_conference_list);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&schedule_conference_list, conference_rm, entry)
	{
		if (conference_rm && !strcmp(conference->name, conference_rm->name)) {
			if (ast_strlen_zero(conference->schedule_id_processing) || (conference->schedule_id_processing && strcmp(conference->schedule_id_processing, conference_rm->schedule_id))) {
				ast_log(LOG_DEBUG, "schedule conference error recovery detect '%s'conference->schedule_id_processing %s conference_rm->schedule_id %s.\n", conference_rm->name, conference->schedule_id_processing, conference_rm->schedule_id);
				conference_write_statistics_by_confschedule(conference_rm);
				if (conference_rm->recycle_sched_id) {
					int ret = ast_sched_del(recycle_sched, conference_rm->recycle_sched_id);
					conference_rm->recycle_sched_id = 0;
					ast_log(LOG_DEBUG, "schedule conference sched '%s' is removed, ret %d.\n", conference_rm->name, ret);
				}
				AST_LIST_REMOVE_CURRENT(entry);
				ao2_cleanup(conference_rm);
				break;
			}
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&schedule_conference_list);
}

static struct schedule_conference *schedule_conf_fetch_add(struct confbridge_conference *conference)
{
	struct schedule_conference *conference_s = NULL;
	char start_time_buf[AST_MAX_START_TIME] = { 0 };
	struct ast_tm tm;

	if (!conference) {
		ast_log(LOG_DEBUG, "conference null!");
		return NULL;
	}

	if (ast_strlen_zero(conference->schedule_id_processing) || ast_strlen_zero(conference->schedule_subject)
		|| ast_strlen_zero(conference->schedule_utc_start_time) || ast_strlen_zero(conference->name)
		|| ast_strlen_zero(conference->start_time)) {
		ast_log(LOG_DEBUG, "some schedule value null!");
		return NULL;
	}

	AST_LIST_LOCK(&schedule_conference_list);

	AST_LIST_TRAVERSE(&schedule_conference_list, conference_s, entry) {
		if (!strcmp(conference_s->name, conference->name)) {
			if (conference->peak_users > conference_s->max_user) {
				conference_s->max_user = conference->peak_users;
			}
			ao2_ref(conference_s, +1);
			break;
		}
	}

	if (!conference_s) {
		if ((conference_s = ao2_alloc(sizeof(*conference_s), schedule_conference_destroy))) {
			ast_log(LOG_DEBUG, "Add a new schedule conference '%s'.\n", conference->name);

			ast_copy_string(conference_s->name, conference->name, sizeof(conference_s->name));
			ast_copy_string(conference_s->schedule_id, conference->schedule_id_processing, sizeof(conference_s->schedule_id));
			ast_copy_string(conference_s->utc_start_time, conference->schedule_utc_start_time, sizeof(conference_s->utc_start_time));
			ast_copy_string(conference_s->conf_subject, conference->schedule_subject, sizeof(conference_s->conf_subject));
			ast_copy_string(conference_s->host_exten, conference->schedule_host_exten ? conference->schedule_host_exten : "", sizeof(conference_s->host_exten));
			ast_copy_string(conference_s->host_email, conference->schedule_host_email ? conference->schedule_host_email : "", sizeof(conference_s->host_email));
			conference_s->type = ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU) ? CONF_TYPE_VIDEO : CONF_TYPE_AUDIO;
			conference_s->max_user = conference->peak_users;
			conference_s->is_random_conf = conference->b_profile.is_random_conf;
			AST_LIST_INSERT_TAIL(&schedule_conference_list, conference_s, entry);
			ao2_ref(conference_s, +1);
		}
	}

	if (ast_strlen_zero(conference_s->start_time)) {
		ast_strptime(conference->start_time, "%Y-%m-%dT%H:%M:%S%z", &tm);
		ast_strftime(start_time_buf, sizeof(start_time_buf), "%Y-%m-%d %H:%M", &tm);
		ast_copy_string(conference_s->start_time, start_time_buf, sizeof(conference_s->start_time));
	}
	if (!conference_s->conference_start) {
		conference_s->conference_start = conference->conference_start;
	}
	AST_LIST_UNLOCK(&schedule_conference_list);

	return conference_s;
}

static struct schedule_conference *schedule_conf_cli_fetch(const char *conference)
{
	struct schedule_conference *conference_s = NULL;

	if (ast_strlen_zero(conference)) {
		ast_log(LOG_DEBUG, "conference name null!");
		return NULL;
	}

	AST_LIST_LOCK(&schedule_conference_list);

	AST_LIST_TRAVERSE(&schedule_conference_list, conference_s, entry) {
		if (!strcmp(conference_s->name, conference)) {
			ao2_ref(conference_s, +1);
			break;
		}
	}
	AST_LIST_UNLOCK(&schedule_conference_list);
	return conference_s;
}

static void schedule_attendees_destroy(void *obj)
{
	struct schedule_attendees *user = obj;

	ast_log(LOG_DEBUG, "schedule attendees user '%s|%s' is destroying.\n", user->extension, user->terminal_type);
}

static int schedule_attendee_add(struct schedule_conference *conference, const char *extension, const char *terminal_type)
{
	struct schedule_attendees *user = NULL;

	if (!conference || ast_strlen_zero(extension)) {
		return -1;
	}

	user = ao2_alloc(sizeof(*user), schedule_attendees_destroy);
	if (!user) {
		return -1;
	}
	ast_copy_string(user->extension, extension, sizeof(user->extension));
	ast_copy_string(user->terminal_type, terminal_type?terminal_type:"other_type", sizeof(user->terminal_type));

	ast_log(LOG_DEBUG, "Add schedule attendees '%s'.\n", user->extension);
	ao2_lock(conference);
	AST_LIST_INSERT_TAIL(&conference->schedule_attendee_list, user, entry);
	ao2_unlock(conference);

	return 0;
}

static struct schedule_attendees *schedule_attendee_find(struct schedule_conference *conference, const char *extension, const char *terminal_type)
{
	struct schedule_attendees *user = NULL;

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->schedule_attendee_list, user, entry) {
		if (!strcmp(user->extension, extension) && !strcmp(user->terminal_type, terminal_type)) {
			ao2_ref(user, +1);
			break;
		}
	}
	ao2_unlock(conference);

	return user;
}

static void schedule_attendee_list(struct schedule_conference *conference)
{
	int total_users = 0;
	struct schedule_attendees *user = NULL;

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->schedule_attendee_list, user, entry) {
		ast_log(LOG_DEBUG, "schedule_attendee_list user '%s' terminal_type '%s'\n", user->extension, user->terminal_type);
		total_users++;
	}
	if (total_users > conference->total_user) {
		conference->total_user = total_users;
	}
	ao2_unlock(conference);
}

static void schedule_conference_attendee_join_list(struct confbridge_conference *conference, struct confbridge_user *user)
{
	RAII_VAR(struct schedule_conference *, conference_obj, NULL, ao2_cleanup);
	struct schedule_attendees *schedule_user = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	const char * terminate_str = NULL;
	char extension[AST_MAX_EXTENSION] = {0};

	if (!conference || !user) {
		ast_log(LOG_WARNING, "conference or user is null");
		return;
	}

	ao2_lock(conference);
	if ((conference->activeusers + conference->waitingusers) > conference->peak_users) {
		conference->peak_users = conference->activeusers + conference->waitingusers;
	}
	ao2_unlock(conference);

	conference_obj = schedule_conf_fetch_add(conference);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference '%s'\n", conference->name);
		return;
	}

	if (conference_obj->recycle_sched_id) {
		int ret = ast_sched_del(recycle_sched, conference_obj->recycle_sched_id);
		conference_obj->recycle_sched_id = 0;
		ast_log(LOG_DEBUG, "schedule conference sched '%s' is removed, ret %d.\n", conference_obj->name, ret);
	}

	terminate_type = ast_channel_get_terminate_type(user->chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}

	conf_get_exten_by_channel(ast_channel_name(user->chan), extension, sizeof(extension)-1);
	if (!strncasecmp(extension, "trunk", 5)) {
		ast_log(LOG_DEBUG, "extension %s", extension);
		ast_copy_string(extension, ast_channel_caller_num(user->chan), sizeof(extension));
	}
	if (ast_strlen_zero(extension)) {
		ast_copy_string(extension, S_COR(ast_channel_caller(user->chan)->id.number.valid, ast_channel_caller(user->chan)->id.number.str, "Anonymous"), sizeof(extension));
	}

	ao2_lock(conference_obj);

	schedule_user = schedule_attendee_find(conference_obj, extension, terminate_str);
	if (!schedule_user) {
		if (schedule_attendee_add(conference_obj, extension, terminate_str) < 0) {
			ast_log(LOG_WARNING, "Failed to add attends '%s' in conference '%s'\n", extension, conference->name);
			ao2_unlock(conference_obj);
			return;
		}
	} else {
		ast_log(LOG_DEBUG, "User '%s' has already been in schedule attendee list(terminate type %s).\n", extension, terminate_str);
		ao2_cleanup(schedule_user);
	}
	schedule_attendee_list(conference_obj);
	ao2_unlock(conference_obj);

}

static void schedule_conference_end_delay(struct confbridge_conference *conference)
{
	char cmd[256] = {0};

	if (!conference) {
		ast_log(LOG_ERROR, "conference null!");
		return;
	}
	if (ast_strlen_zero(conference->schedule_id_processing)) {
		ast_log(LOG_DEBUG, "not schedule conference!");
		return;
	}
	if (ast_strlen_zero(conference->schedule_utc_start_time)) {
		ast_log(LOG_DEBUG, "schedule utc start time null!");
		return;
	}

	if (ast_test_flag(&conference->b_profile, BRIDGE_OPT_VIDEO_SRC_SFU)) {
		snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/exe_mcm_reservation.py --type delayclean --id %s --conf %s --utctime \"%s\" &",
			conference->schedule_id_processing, conference->name, conference->schedule_utc_start_time);
	} else {
		snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/DB2File_PY/ucm_meetme_kick_playback.py --bookid \"%s\" --action \"delayclean\" --data \"%s\" --utctime \"%s\" &",
			conference->schedule_id_processing, conference->name, conference->schedule_utc_start_time);
	}
	ast_log(LOG_DEBUG, "cmd is %s!",cmd);
	ast_safe_system2(cmd);
}

static void im_conference_list_destroy(void *obj)
{
	struct im_conference *conference = obj;
	ast_log(LOG_DEBUG, "im conference '%s' is destroying.\n", conference->name);
}

static struct im_conference *im_conf_list_fetch_add(const char *conf_number, const char *im_group_id, int strong_related, int type, int add, const char *conf_subject)
{
	struct im_conference *conference = NULL;

	if (ast_strlen_zero(conf_number) || (ast_strlen_zero(im_group_id) && add)) {
		return NULL;
	}

	AST_LIST_LOCK(&im_conference_list);

	AST_LIST_TRAVERSE(&im_conference_list, conference, entry) {
		if (!strcmp(conference->name, conf_number)) {
			ao2_ref(conference, +1);
			break;
		}
	}

	if (add && !conference) {
		if ((conference = ao2_alloc(sizeof(*conference), im_conference_list_destroy))) {
			ast_log(LOG_DEBUG, "Add a new im conference '%s' im_group_id '%s'.\n", conf_number, im_group_id);
			ast_copy_string(conference->name, conf_number, sizeof(conference->name));
			conference->type = type;
			conference->im_conference_start = time(NULL);
			conference->status = CONF_IM_INIT;
			conference->strong_related = strong_related;
			ast_copy_string(conference->conf_subject, conf_subject ? conf_subject : "", sizeof(conference->conf_subject));
			ast_copy_string(conference->im_group_id, im_group_id ? im_group_id : "", sizeof(conference->im_group_id));
			AST_LIST_INSERT_TAIL(&im_conference_list, conference, entry);
			ao2_ref(conference, +1);
		}
	}
	AST_LIST_UNLOCK(&im_conference_list);

	if (conference) {
		ast_log(LOG_DEBUG, "im conference '%s' im_group_id '%s'.\n", conference->name, conference->im_group_id);
	}
	return conference;
}

static int im_conf_list_update_status(const char *conf_number, int conf_status)
{
	RAII_VAR(struct im_conference *, conference, NULL, ao2_cleanup);

	if (ast_strlen_zero(conf_number)) {
		return -1;
	}

	conference = im_conf_list_fetch_add(conf_number, NULL, 0, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	ao2_lock(conference);
	if (conference->status != conf_status)
	{
		conference->status = conf_status;
		ast_log(LOG_DEBUG, "conference->status %d", conference->status);
	}
	ao2_unlock(conference);

	return 0;
}

static struct schedule_attendees_wait *schedule_attendee_wait_find(struct schedule_conference_wait *conference, const char *chan_name)
{
	struct schedule_attendees_wait *user = NULL;

	ao2_lock(conference);
	AST_LIST_TRAVERSE(&conference->schedule_attendees_wait_list, user, entry) {
		if (!strcmp(user->channel_name, chan_name)) {
			ao2_ref(user, +1);
			break;
		}
	}
	ao2_unlock(conference);

	return user;
}

static void schedule_conference_wait_destroy(void *obj)
{
	struct schedule_conference_wait *conference = obj;
	struct schedule_attendees_wait *user=  NULL;

	ast_log(LOG_DEBUG, "schedule conference waiting '%s' is destroying.\n", conference->name);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference->schedule_attendees_wait_list, user, entry){
		AST_LIST_REMOVE_CURRENT(entry);
		ao2_cleanup(user);
	}
	AST_LIST_TRAVERSE_SAFE_END;
}

#if 0
static struct schedule_conference_wait *schedule_conf_wait_fetch_add(const char *conference, int add, const char *pincode, const char *admin_pincode, const char *schedule_id, int conf_type, const char *subject, const char *host_exten, const char *host_email, const char *utc_start_time)
{
	struct schedule_conference_wait *conference_w = NULL;

	if (ast_strlen_zero(conference)) {
		ast_log(LOG_DEBUG, "conference null!");
		return NULL;
	}

	if (add && ast_strlen_zero(schedule_id)) {
		ast_log(LOG_DEBUG, "schedule_id null!");
		return NULL;
	}
	AST_LIST_LOCK(&schedule_conference_wait_list);

	AST_LIST_TRAVERSE(&schedule_conference_wait_list, conference_w, entry) {
		if (!strcmp(conference_w->name, conference)) {
			ao2_ref(conference_w, +1);
			break;
		}
	}

	if (add && !conference_w) {
		if ((conference_w = ao2_alloc(sizeof(*conference_w), schedule_conference_wait_destroy))) {
			ast_log(LOG_DEBUG, "Add a new schedule conference waiting %s schedule_id %s.\n", conference, schedule_id);
			if (!ast_strlen_zero(pincode)) {
				ast_copy_string(conference_w->pin, pincode, sizeof(conference_w->pin));
			}
			if (!ast_strlen_zero(admin_pincode)) {
				ast_copy_string(conference_w->admin_pin, admin_pincode, sizeof(conference_w->admin_pin));
			}
			if (!ast_strlen_zero(subject)) {
				ast_copy_string(conference_w->conf_subject, subject, sizeof(conference_w->conf_subject));
			}
			if (!ast_strlen_zero(host_exten)) {
				ast_copy_string(conference_w->host_exten, host_exten, sizeof(conference_w->host_exten));
			}
			if (!ast_strlen_zero(host_email)) {
				ast_copy_string(conference_w->host_email, host_email, sizeof(conference_w->host_email));
			}
			if (!ast_strlen_zero(utc_start_time)) {
				ast_copy_string(conference_w->utc_start_time, utc_start_time, sizeof(conference_w->utc_start_time));
			}
			ast_copy_string(conference_w->name, conference, sizeof(conference_w->name));
			conference_w->type = conf_type;
			ast_copy_string(conference_w->schedule_id, schedule_id, sizeof(conference_w->schedule_id));
			AST_LIST_INSERT_TAIL(&schedule_conference_wait_list, conference_w, entry);
			ao2_ref(conference_w, +1);
		}
	}

	AST_LIST_UNLOCK(&schedule_conference_wait_list);

	return conference_w;
}
#endif

static struct schedule_conference_wait *schedule_conf_wait_info_fetch_add(int add, struct schedule_conference_wait_info* conf_sch_w)
{
	struct schedule_conference_wait *conference_w = NULL;

	if (!conf_sch_w || ast_strlen_zero(conf_sch_w->conf_name)) {
		ast_log(LOG_DEBUG, "conference null!");
		return NULL;
	}

	if (add && ast_strlen_zero(conf_sch_w->schedule_id)) {
		ast_log(LOG_DEBUG, "schedule_id null!");
		return NULL;
	}
	AST_LIST_LOCK(&schedule_conference_wait_list);

	AST_LIST_TRAVERSE(&schedule_conference_wait_list, conference_w, entry) {
		if (!strcmp(conference_w->name, conf_sch_w->conf_name)) {
			ao2_ref(conference_w, +1);
			break;
		}
	}

	if (add && !conference_w) {
		if ((conference_w = ao2_alloc(sizeof(*conference_w), schedule_conference_wait_destroy))) {
			ast_log(LOG_DEBUG, "Add a new schedule conference waiting %s schedule_id %s.\n", conf_sch_w->conf_name, conf_sch_w->schedule_id);
			if (!ast_strlen_zero(conf_sch_w->pin)) {
				ast_copy_string(conference_w->pin, conf_sch_w->pin, sizeof(conference_w->pin));
			}
			if (!ast_strlen_zero(conf_sch_w->admin_pin)) {
				ast_copy_string(conference_w->admin_pin, conf_sch_w->admin_pin, sizeof(conference_w->admin_pin));
			}
			if (!ast_strlen_zero(conf_sch_w->conf_subject)) {
				ast_copy_string(conference_w->conf_subject, conf_sch_w->conf_subject, sizeof(conference_w->conf_subject));
			}
			if (!ast_strlen_zero(conf_sch_w->host_exten)) {
				ast_copy_string(conference_w->host_exten, conf_sch_w->host_exten, sizeof(conference_w->host_exten));
			}
			if (!ast_strlen_zero(conf_sch_w->host_email)) {
				ast_copy_string(conference_w->host_email, conf_sch_w->host_email, sizeof(conference_w->host_email));
			}
			if (!ast_strlen_zero(conf_sch_w->utc_start_time)) {
				ast_copy_string(conference_w->utc_start_time, conf_sch_w->utc_start_time, sizeof(conference_w->utc_start_time));
			}
			if (!ast_strlen_zero(conf_sch_w->start_time)) {
				ast_copy_string(conference_w->start_time, conf_sch_w->start_time, sizeof(conference_w->start_time));
			}
			ast_copy_string(conference_w->name, conf_sch_w->conf_name, sizeof(conference_w->name));
			conference_w->type = conf_sch_w->conf_type;
			ast_copy_string(conference_w->schedule_id, conf_sch_w->schedule_id, sizeof(conference_w->schedule_id));
			AST_LIST_INSERT_TAIL(&schedule_conference_wait_list, conference_w, entry);
			ao2_ref(conference_w, +1);
		}
	}

	AST_LIST_UNLOCK(&schedule_conference_wait_list);

	return conference_w;
}

static struct schedule_conference_wait *schedule_conf_wait_info_update(struct schedule_conference_wait_info* conf_sch_w)
{
	struct schedule_conference_wait *conference_w = NULL;

	if (!conf_sch_w || ast_strlen_zero(conf_sch_w->conf_name)) {
		ast_log(LOG_DEBUG, "conference null!");
		return NULL;
	}

	if (ast_strlen_zero(conf_sch_w->schedule_id)) {
		ast_log(LOG_DEBUG, "schedule_id null!");
		return NULL;
	}
	AST_LIST_LOCK(&schedule_conference_wait_list);

	AST_LIST_TRAVERSE(&schedule_conference_wait_list, conference_w, entry) {
		if (!strcmp(conference_w->name, conf_sch_w->conf_name)) {
			ao2_ref(conference_w, +1);
			break;
		}
	}

	if (conference_w) {
		ast_log(LOG_DEBUG, "update schedule conference %s waiting info schedule_id %s.\n", conf_sch_w->conf_name, conf_sch_w->schedule_id);
		conference_w->type = conf_sch_w->conf_type;
		ast_copy_string(conference_w->name, conf_sch_w->conf_name, sizeof(conference_w->name));
		ast_copy_string(conference_w->pin, conf_sch_w->pin, sizeof(conference_w->pin));
		ast_copy_string(conference_w->admin_pin, conf_sch_w->admin_pin, sizeof(conference_w->admin_pin));
		ast_copy_string(conference_w->conf_subject, conf_sch_w->conf_subject, sizeof(conference_w->conf_subject));
		ast_copy_string(conference_w->host_exten, conf_sch_w->host_exten, sizeof(conference_w->host_exten));
		ast_copy_string(conference_w->host_email, conf_sch_w->host_email, sizeof(conference_w->host_email));
		ast_copy_string(conference_w->utc_start_time, conf_sch_w->utc_start_time, sizeof(conference_w->utc_start_time));
		ast_copy_string(conference_w->start_time, conf_sch_w->start_time, sizeof(conference_w->start_time));
		ast_copy_string(conference_w->schedule_id, conf_sch_w->schedule_id, sizeof(conference_w->schedule_id));
	}

	AST_LIST_UNLOCK(&schedule_conference_wait_list);

	return conference_w;
}

static void schedule_attendee_wait_list_destroy(void *obj)
{
	struct schedule_attendees_wait *user_list = obj;
	ast_log(LOG_DEBUG, "wait join user list '%s' is destroying.\n", user_list->channel_name);
}

static int schedule_attendee_wait_add(struct schedule_conference_wait *conference, struct ast_channel *chan)
{
	struct schedule_attendees_wait *user = NULL;
	char extension[AST_MAX_EXTENSION] = {0};

	if (!conference || !chan) {
		return -1;
	}

	conf_get_exten_by_channel(ast_channel_name(chan), extension, sizeof(extension)-1);
	if (!strncasecmp(extension, "trunk", 5)) {
		ast_copy_string(extension, ast_channel_caller_num(chan), sizeof(extension));
	}
	if (ast_strlen_zero(extension)) {
		ast_copy_string(extension, S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "Anonymous"), sizeof(extension));
	}

	user = ao2_alloc(sizeof(*user), schedule_attendee_wait_list_destroy);
	if (!user) {
		return -1;
	}
	ast_copy_string(user->channel_name, ast_channel_name(chan), sizeof(user->channel_name));
	ast_copy_string(user->extension, extension, sizeof(user->extension));
	user->chan = chan;
	const char *wait_is_admin = pbx_builtin_getvar_helper(chan, "WAIT_IS_ADMIN");
	if (!ast_strlen_zero(wait_is_admin) && !strcasecmp(wait_is_admin, "yes")) {
		user->is_admin = 1;
	} else {
		user->is_admin = 0;
	}
	ast_log(LOG_DEBUG, "Add schedule attendees waiting '%s'.\n", user->channel_name);
	ao2_lock(conference);
	conference->wait_users++;
	AST_LIST_INSERT_TAIL(&conference->schedule_attendees_wait_list, user, entry);
	ao2_unlock(conference);

	return 0;
}

static int schedule_wait_user_set_related_info(const char *conf_number, struct ast_channel *chan)
{
	RAII_VAR(struct cache_conference *, conference, NULL, ao2_cleanup);
	RAII_VAR(struct cache_authorized_user *, user, NULL, ao2_cleanup);
	const char * terminate_str = NULL;
	enum ast_terminal_type terminate_type = AST_TERMINAL_TYPE_UNKNOW;
	char extension[AST_MAX_EXTENSION] = {0};

	if (!chan || ast_strlen_zero(ast_channel_name(chan)) || !strstr(ast_channel_name(chan), "conf")) {
		ast_log(LOG_DEBUG, "Not anonymous, clear identity");
		return -1;
	}
	if (ast_strlen_zero(conf_number)) {
		return -1;
	}

	terminate_type = ast_channel_get_terminate_type(chan);
	if (terminate_type == AST_TERMINAL_TYPE_WEBRTC) {
		terminate_str = "webrtc";
	} else if (terminate_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		terminate_str = "wave_app";
	} else {
		terminate_str = "other_type";
	}
	conf_get_exten_by_channel(ast_channel_name(chan), extension, sizeof(extension)-1);
	if (!strncasecmp(extension, "trunk", 5)) {
		ast_log(LOG_DEBUG, "extension %s", extension);
		ast_copy_string(extension, ast_channel_caller_num(chan), sizeof(extension));
	}
	if (ast_strlen_zero(extension)) {
		ast_copy_string(extension, S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, "Anonymous"), sizeof(extension));
	}

	conference = cache_conf_fetch_add(conf_number, CONF_TYPE_NONE, 0, NULL);
	if (!conference) {
		return -1;
	}

	user = cache_conf_authorized_user_find(conference, extension, terminate_str);
	if (!user) {
		return -1;
	}
	pbx_builtin_setvar_helper(chan, "RELATED_INFO", user->related_info);
	ast_log(LOG_DEBUG, "set %s RELATED_INFO %s", extension, user->related_info);
	return 0;
}

static int schedule_conference_attendee_join_wait_list(const char *conference, struct ast_channel *chan)
{
	RAII_VAR(struct schedule_conference_wait *, conference_obj, NULL, ao2_cleanup);
	struct schedule_attendees_wait *schedule_wait_user = NULL;
	int res = 0;

	if (ast_strlen_zero(conference) || !chan) {
		ast_log(LOG_WARNING, "conference or chan is null");
		res = -1;
		return res;
	}
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	conf_sch_w.conf_name = conference;
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);

	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference);
		res = -1;
		return res;
	}

	ao2_lock(conference_obj);

	if (conference_obj->cancelled) {
		ast_log(LOG_DEBUG, "the schedule conference cancelled '%s'\n", conference);
		res = -1;
		ao2_unlock(conference_obj);
		return res;
	}

	schedule_wait_user = schedule_attendee_wait_find(conference_obj, ast_channel_name(chan));
	if (!schedule_wait_user) {
		if (schedule_attendee_wait_add(conference_obj, chan)) {
			ast_log(LOG_WARNING, "Failed to add attends '%s' in conference waiting'%s'\n", ast_channel_name(chan), conference);
			ao2_unlock(conference_obj);
			res = -1;
			return res;
		}
		if (schedule_wait_user_set_related_info(conference, chan)) {
			ast_log(LOG_WARNING, "set related info failed!");
		}
	} else {
		ast_log(LOG_DEBUG, "User '%s' has already been in schedule attendee wait list.\n", ast_channel_name(chan));
		ao2_cleanup(schedule_wait_user);
	}
	ao2_unlock(conference_obj);
	return res;
}

static void schedule_conference_wait_list_remove(const char *conf_number)
{
	struct schedule_conference_wait *conference = NULL;

	if (ast_strlen_zero(conf_number)) {
		return;
	}

	ast_log(LOG_DEBUG, "Remove the schedule conference waiting '%s'.\n", conf_number);
	AST_LIST_LOCK(&schedule_conference_wait_list);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&schedule_conference_wait_list, conference, entry)
	{
		if (!strcasecmp(conf_number, "all") || !strcmp(conference->name, conf_number)) {
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(conference);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&schedule_conference_wait_list);
}

/*must clean by hangup conference by host or crontab clean time is each*/
static void schedule_conference_wait_list_remove_by_scheduleid(const char *conf_number, const char *schedule_id)
{
	struct schedule_conference_wait *conference = NULL;

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(schedule_id)) {
		return;
	}

	ast_log(LOG_DEBUG, "Remove the schedule conference waiting '%s' schedule id %s.\n", conf_number, schedule_id);
	AST_LIST_LOCK(&schedule_conference_wait_list);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&schedule_conference_wait_list, conference, entry)
	{
		if (!strcmp(conference->name, conf_number) && !strcmp(conference->schedule_id, schedule_id)) {
			AST_LIST_REMOVE_CURRENT(entry);
			conference->wait_users--;
			ao2_cleanup(conference);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&schedule_conference_wait_list);
}

static void schedule_conference_wait_list_update_by_scheduleid(const char *conf_number, const char *schedule_id, const char *cofn_info)
{
	RAII_VAR(struct schedule_conference_wait *, conference, NULL, ao2_cleanup);
	struct ast_json  *utc_time, *host_email, *host_exten, *subject, *pin_code, *admin_pincode;
	RAII_VAR(struct ast_json *, json_s, NULL, ast_json_unref);
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	struct ast_json_error err;

	if (ast_strlen_zero(conf_number) || ast_strlen_zero(schedule_id) || ast_strlen_zero(cofn_info)) {
		ast_log(LOG_DEBUG, "value null!\n");
		return;
	}

	conf_sch_w.conf_name = conf_number;
	//conference = schedule_conf_wait_fetch_add(conf_number, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conference = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conf_number);
		return;
	}

	ast_log(LOG_DEBUG, "update the schedule conference waiting '%s' schedule id %s.\n", conf_number, schedule_id);

	ao2_lock(conference);
	if (!ast_strlen_zero(cofn_info) && !ast_strlen_zero(conference->schedule_id) && !strcmp(schedule_id, conference->schedule_id)) {
		json_s = ast_json_load_string(cofn_info, &err);
		if(!json_s) {
			ast_log(LOG_WARNING, "load oject error.");
		} else {
			if ((host_exten = ast_json_object_get(json_s, "host_exten"))) {
				if (!ast_strlen_zero(conference->host_exten)) {
					memset(conference->host_exten, 0, sizeof(conference->host_exten));
				}
				ast_copy_string(conference->host_exten, ast_json_string_get(host_exten) ? ast_json_string_get(host_exten) : "", sizeof(conference->host_exten));
			}
			if ((host_email = ast_json_object_get(json_s, "host_email"))) {
				if (!ast_strlen_zero(conference->host_email)) {
					memset(conference->host_email, 0, sizeof(conference->host_email));
				}
				ast_copy_string(conference->host_email, ast_json_string_get(host_email) ? ast_json_string_get(host_email) : "", sizeof(conference->host_email));
			}
			if ((utc_time = ast_json_object_get(json_s, "utc_start_time"))) {
					if (!ast_strlen_zero(conference->utc_start_time)) {
						memset(conference->utc_start_time, 0, sizeof(conference->utc_start_time));
					}
					ast_copy_string(conference->utc_start_time, ast_json_string_get(utc_time) ? ast_json_string_get(utc_time) : "", sizeof(conference->utc_start_time));
			}
			if ((pin_code = ast_json_object_get(json_s, "pin_code"))) {
				if (!ast_strlen_zero(conference->pin)) {
					memset(conference->pin, 0, sizeof(conference->pin));
				}
				ast_copy_string(conference->pin, ast_json_string_get(pin_code) ? ast_json_string_get(pin_code) : "", sizeof(conference->pin));
			}
			if ((admin_pincode = ast_json_object_get(json_s, "admin_pincode"))) {
				if (!ast_strlen_zero(conference->admin_pin)) {
					memset(conference->admin_pin, 0, sizeof(conference->admin_pin));
				}
				ast_copy_string(conference->admin_pin, ast_json_string_get(admin_pincode) ? ast_json_string_get(admin_pincode) : "", sizeof(conference->admin_pin));
			}
			if ((subject = ast_json_object_get(json_s, "conf_subject"))) {
				if (!ast_strlen_zero(conference->conf_subject)) {
					memset(conference->conf_subject, 0, sizeof(conference->conf_subject));
				}
				ast_copy_string(conference->conf_subject, ast_json_string_get(subject) ? ast_json_string_get(subject) : "", sizeof(conference->conf_subject));
			}
		}
	}
	ao2_unlock(conference);
}

static void schedule_conference_wait_user_remove_by_channel(const char *conference, const char *chan_name)
{
	RAII_VAR(struct schedule_conference_wait *, conference_obj, NULL, ao2_cleanup);
	struct schedule_attendees_wait *user = NULL;
	struct schedule_conference_wait_info conf_sch_w = { 0 };
	if (ast_strlen_zero(chan_name) || ast_strlen_zero(conference)) {
		return;
	}

	conf_sch_w.conf_name = conference;
	//conference_obj = schedule_conf_wait_fetch_add(conference, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference);
		return;
	}

	ao2_lock(conference_obj);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&conference_obj->schedule_attendees_wait_list, user, entry) {
		if (!strcmp(user->channel_name, chan_name)) {
			conference_obj->wait_users--;
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(user);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	ao2_unlock(conference_obj);
}

static void schedule_conference_wait_list_cancel(const char *conference)
{
	RAII_VAR(struct schedule_conference_wait *, conference_obj, NULL, ao2_cleanup);
	struct schedule_conference_wait_info conf_sch_w = { 0 };

	if (ast_strlen_zero(conference)) {
		return;
	}

	conf_sch_w.conf_name = conference;
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference);
		return;
	}

	ao2_lock(conference_obj);
	if (0 == conference_obj->wait_users && conference_obj->cancelled) {
		schedule_conference_wait_list_remove(conference);
		ast_log(LOG_DEBUG, "schedule conference wait list cancel'%s'\n", conference);
	}
	ao2_unlock(conference_obj);
}

static int schedule_conference_need_wait(const char *conference_name, const char *schedule_id, int *inuse_notify)
{
	int res = 0;
	char family[64] = {0};
	char schedule_id_db[AST_MAX_SCHEDULE_ID] = {0};
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);

	if (ast_strlen_zero(conference_name)) {
		ast_log(LOG_ERROR, "No Conference name provided.");
		res = 0;
		goto auth_end;
	}

	if (ast_strlen_zero(schedule_id)) {
		ast_log(LOG_DEBUG, "No schedule id provided.");
		res = 0;
		goto auth_end;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		snprintf(family, sizeof(family), "SCHEDULE/%s", conference_name);
		if (!ast_db_get(family, "ScheduleID", schedule_id_db, sizeof(schedule_id_db)) && !ast_strlen_zero(schedule_id_db) && strcasecmp(schedule_id_db, schedule_id)) {
			res = 1;
			ast_log(LOG_DEBUG, "schedule conference %s no user,but still inuse!", schedule_id_db);
			goto auth_end;
		} else {
			ast_log(LOG_DEBUG, "No Conference found.we can start schedule conference early!");
			res = 0;
			*inuse_notify = 1;
			goto auth_end;
		}
	}

	ao2_lock(conference);
	if (ast_strlen_zero(conference->schedule_id_processing)) {
		res = 1;
	} else if (conference->schedule_id_processing && strcasecmp(conference->schedule_id_processing, schedule_id)) {
		ast_log(LOG_DEBUG, "conference->schedule_id_processing %s!", conference->schedule_id_processing);
		res = 1;
	} else {
		ast_log(LOG_DEBUG, "No need wait,schedule conference is going we can join it!");
		res = 0;
	}
	ao2_unlock(conference);

auth_end:
	ast_log(LOG_DEBUG, "schedule_conference_need_wait res %d.", res);
	return res;
}

#if 0
static void conference_attendee_auth(const char *conference_name, const char *schedule_id)
{
	RAII_VAR(struct confbridge_conference*, conference, NULL, ao2_cleanup);
	struct confbridge_user *cur_user = NULL;

	if (ast_strlen_zero(conference_name)) {
		ast_log(LOG_ERROR, "No Conference name provided.");
		return;
	}

	if (ast_strlen_zero(schedule_id)) {
		ast_log(LOG_DEBUG, "No schedule id provided.");
		return;
	}

	conference = conf_find_bridge(conference_name);
	if (!conference) {
		ast_log(LOG_DEBUG, "No Conference found, no need auth.");
		return;
	}

	ao2_lock(conference);
	if (ast_strlen_zero(conference->schedule_id_processing) ||
		(conference->schedule_id_processing && strcasecmp(conference->schedule_id_processing, schedule_id))) {
		AST_LIST_TRAVERSE(&conference->active_list, cur_user, list) {
			ast_log(LOG_DEBUG, "[WL] Conference %s chan: %s need auth before next schedule(%s) begin.",
								conference->name,
								ast_channel_name(cur_user->chan),
								schedule_id);
			pbx_builtin_setvar_helper(cur_user->chan, "CONFBRIDGE_ADJOURN", "yes");
			//publish_cei_unsub(conference->name, ast_channel_name(cur_user->chan), ast_channel_caller(cur_user->chan)->id.number.str);
			ast_bridge_remove(conference->bridge, cur_user->chan);
			ast_channel_conf_hangupcause_set(cur_user->chan, AST_CAUSE_CONF_END);
		}
	}
	ao2_unlock(conference);
}
#endif

static void schedule_conference_inuse_notify(struct confbridge_conference *conference)
{
	char cmd[256] = {0};
	char key_step[128]={0};
	char key_status[128]={0};
	char is_random_now[32] = {0};
	int random_schedule = 0;

	if (!conference) {
		ast_log(LOG_ERROR, "conference null!");
		return;
	}
	if (ast_strlen_zero(conference->schedule_id_processing)) {
		ast_log(LOG_ERROR, "schedule id null!");
		return;
	}
	if (conference->b_profile.is_random_conf) {
		snprintf(cmd, sizeof(cmd), "NWAY_CONF_STATUS/%s", conference->name);
		if (ast_db_get(cmd, "IS_RANDOM_CONF", is_random_now, sizeof(is_random_now)) && ast_strlen_zero(is_random_now) && strcasecmp(is_random_now, "yes")) {
			random_schedule = 1;
		}
	}
	if (!random_schedule) {
		ast_log(LOG_ERROR, "not random schedule!");
		return;
	}

	snprintf(cmd, sizeof(cmd),"SCHEDULE/%s", conference->schedule_id_processing);
	ast_db_get(cmd, "STEP", key_step, sizeof(key_step));
	ast_db_get(cmd, "EARLYSTATUS", key_status, sizeof(key_status));
	ast_log(LOG_DEBUG, "key_step %s, key_status %s!", key_step, key_status);
	if (!strcmp(key_step, "call") || !strcmp(key_status, "start")) {
		ast_log(LOG_DEBUG, "schedule meeting already start, no need updatestatus!");
	} else {
		snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/exe_mcm_reservation.py --type updatestatus --id %s --conf %s",
			conference->schedule_id_processing, conference->name);
		ast_log(LOG_DEBUG, "cmd is %s!",cmd);
		ast_safe_system(cmd);
	}
}

static void schedule_conference_wait_inuse_notify(struct confbridge_conference *conference, const char *schedule_id)
{
	char cmd[256] = {0};
	char key_step[128]={0};
	int conf_type = 0;
	RAII_VAR(struct schedule_conference_wait*, conference_obj, NULL, ao2_cleanup);
	struct schedule_conference_wait_info conf_sch_w = { 0 };

	if (!conference) {
		ast_log(LOG_ERROR, "conference null!");
		return;
	}
	if (ast_strlen_zero(schedule_id)) {
		ast_log(LOG_ERROR, "schedule id null!");
		return;
	}

	conf_sch_w.conf_name = conference->name;
	//conference_obj = schedule_conf_wait_fetch_add(conference->name, 0, NULL, NULL, NULL, CONF_TYPE_NONE, NULL, NULL, NULL, NULL);
	conference_obj = schedule_conf_wait_info_fetch_add(0, &conf_sch_w);
	if (!conference_obj) {
		ast_log(LOG_DEBUG, "Failed to fetch the schedule conference waiting'%s'\n", conference->name);
		return;
	}

	ao2_lock(conference_obj);
	if (conference_obj->has_notify) {
		ast_log(LOG_DEBUG, "has notify update status.");
		ao2_unlock(conference_obj);
		return ;
	} else {
		conference_obj->has_notify = 1;
	}
	conf_type = conference_obj->type;
	ao2_unlock(conference_obj);

	/*inuse notify may be later, so set info here.*/
	schedule_conference_wait_add_info(conference, conference_obj);
	/*conference schdeule start, remove the waiting list*/
	schedule_conference_wait_list_remove(conference->name);
	load_schedule_conference_users(conference);

	snprintf(cmd, sizeof(cmd),"SCHEDULE/%s", schedule_id);
	ast_db_get(cmd, "STEP", key_step, sizeof(key_step));
	ast_log(LOG_DEBUG, "key_step %s!", key_step);
	if (!strcmp(key_step, "call")) {
		ast_log(LOG_DEBUG, "schedule meeting already start, no need updatestatus!");
	} else {
		if (conf_type == CONF_TYPE_VIDEO) {
			snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/exe_mcm_reservation.py --type updatestatus --id %s --conf %s",
				schedule_id, conference->name);
		} else if (conf_type == CONF_TYPE_AUDIO) {
			snprintf(cmd, sizeof(cmd), "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/DB2File_PY/ucm_meetme_kick_playback.py --bookid \"%s\" --action \"updatestatus\" --data \"%s\"",
				schedule_id, conference->name);
		} else {
			ast_log(LOG_ERROR, "conf_type error!");
		}
		ast_log(LOG_DEBUG, "cmd is %s!",cmd);
		ast_safe_system(cmd);
	}
}

static int schedule_conference_wait_add_info(struct confbridge_conference *conference, struct schedule_conference_wait* conference_w)
{
	RAII_VAR(struct schedule_conference *, conference_s, NULL, ao2_cleanup);
	char schedule_id[AST_MAX_SCHEDULE_ID] = {0};
	char utc_start_time[AST_MAX_START_TIME] = {0};
	char conf_subject[AST_MAX_CONF_SUBJECT] = {0};
	char host_exten[AST_MAX_HOST_EXTEN] = {0};
	char host_email[AST_MAX_HOST_EMAIL] = {0};
	char start_time[AST_MAX_START_TIME] = {0};
	char family[128]={0};
	char key[512]={0};
	int conf_type = 0;

	if (!conference) {
		ast_log(LOG_DEBUG, "conference null!");
		return 0;
	}
	if (!conference_w) {
		ast_log(LOG_DEBUG, "conference wait null!");
		return 0;
	}

	ao2_lock(conference_w);
	ast_copy_string(schedule_id, conference_w->schedule_id, sizeof(schedule_id));
	ast_copy_string(utc_start_time, conference_w->utc_start_time, sizeof(utc_start_time));
	ast_copy_string(conf_subject, conference_w->conf_subject, sizeof(conf_subject));
	ast_copy_string(host_exten, conference_w->host_exten, sizeof(host_exten));
	ast_copy_string(host_email, conference_w->host_email, sizeof(host_email));
	ast_copy_string(start_time, conference_w->start_time, sizeof(start_time));
	conf_type = conference_w->type;
	/*set astdb status*/
	snprintf(family, sizeof(family)-1,"SCHEDULE/%s", schedule_id);
	ast_db_get(family, "STEP", key, sizeof(key));
	if (!strcmp(key, "call")) {
		ast_log(LOG_DEBUG, "schedule meeting already call!");
	} else {
		memset(key, 0, sizeof(key));
		memset(family, 0, sizeof(family));
		ast_log(LOG_DEBUG, "schedule_conference_wait_add_info put astdb!");
		snprintf(family, sizeof(family)-1,"SCHEDULE/%s", conference_w->name);
		ast_db_del(family, "AlertScheduleID");
		ast_db_put(family, "ConfType", (conf_type == CONF_TYPE_VIDEO) ? "video" : "audio");
		ast_db_put(family, "Pincode", conference_w->pin);
		ast_db_put(family, "AdminPincode", conference_w->admin_pin);
		ast_db_put(family, "ScheduleID", schedule_id);
		memset(family, 0, sizeof(family));
		snprintf(family, sizeof(family)-1,"SCHEDULE/%s", schedule_id);
		ast_db_put(family, "STEP", "earlyjoin");
		ast_db_put(family, "EARLYSTATUS", "start");
		ast_db_put(family, "ConfSubject", conf_subject);
		snprintf(key, sizeof(key)-1, "{\"host_exten\":\"%s\",\"host_email\":\"%s\",\"utc_start_time\":\"%s\",\"start_time\":\"%s\"}", host_exten, host_email, utc_start_time, start_time);
		ast_db_put(family, "Info", key);
	}
	ao2_unlock(conference_w);

	AST_LIST_LOCK(&schedule_conference_list);

	AST_LIST_TRAVERSE(&schedule_conference_list, conference_s, entry) {
		if (!strcmp(conference_s->name, conference->name) && !strcmp(conference_s->schedule_id, schedule_id)) {
			ao2_ref(conference_s, +1);
			break;
		}
	}

	if (!conference_s) {
		if ((conference_s = ao2_alloc(sizeof(*conference_s), schedule_conference_destroy))) {
			ast_log(LOG_DEBUG, "schedule_conference_wait_add_info Add a new schedule conference '%s'.\n", schedule_id);
			ast_copy_string(conference_s->name, conference->name, sizeof(conference_s->name));
			ast_copy_string(conference_s->schedule_id, schedule_id, sizeof(conference_s->schedule_id));
			ast_copy_string(conference_s->utc_start_time, utc_start_time, sizeof(conference_s->utc_start_time));
			ast_copy_string(conference_s->conf_subject, conf_subject, sizeof(conference_s->conf_subject));
			ast_copy_string(conference_s->host_exten, host_exten, sizeof(conference_s->host_exten));
			ast_copy_string(conference_s->host_email, host_email, sizeof(conference_s->host_email));
			//ast_copy_string(conference_s->start_time, start_time, sizeof(conference_s->start_time));
			conference_s->type = conf_type;
			AST_LIST_INSERT_TAIL(&schedule_conference_list, conference_s, entry);
			ao2_ref(conference_s, +1);
		}
	}
	AST_LIST_UNLOCK(&schedule_conference_list);

	ao2_lock(conference);
	if (conference->schedule_subject) {
		ast_free(conference->schedule_subject);
		conference->schedule_subject = NULL;
	}
	if (!ast_strlen_zero(conf_subject)) {
		conference->schedule_subject = ast_strdup(conf_subject);
	}

	if (conference->schedule_id_processing) {
		ast_free(conference->schedule_id_processing);
		conference->schedule_id_processing = NULL;
	}
	if (!ast_strlen_zero(schedule_id)) {
		conference->schedule_id_processing = ast_strdup(schedule_id);
	}

	if (conference->schedule_host_exten) {
		ast_free(conference->schedule_host_exten);
		conference->schedule_host_exten = NULL;
	}
	conference->schedule_host_exten = ast_strdup(host_exten);

	if (conference->schedule_host_email) {
		ast_free(conference->schedule_host_email);
		conference->schedule_host_email = NULL;
	}
	conference->schedule_host_email = ast_strdup(host_email);

	if (conference->schedule_utc_start_time ) {
		ast_free(conference->schedule_utc_start_time );
		conference->schedule_utc_start_time  = NULL;
	}
	conference->schedule_utc_start_time  = ast_strdup(utc_start_time);

	/*Set start timeval*/
	if (ast_strlen_zero(start_time)) {
		conference->schedule_start_timeval = ast_tvnow();
	} else {
		struct tm start_tm = {0};
		struct timeval start_tv = {0};
		strptime(start_time, "%Y-%m-%d %H:%M", &start_tm);
		start_tv.tv_sec = mktime(&start_tm);
		conference->schedule_start_timeval = start_tv;
		ast_log(LOG_DEBUG, "schedule_conference_wait_add_info time diff %ld\n", ast_tvdiff_sec(ast_tvnow(), start_tv));
	}

	ao2_unlock(conference);

	return 0;
}

/*
* For UCM dial out of trunk,
* if the terminal joins the meeting through assistant,
* other terminals will stop ringing
*/
static void cancel_remote_calling(struct confbridge_user*user)
{
	struct ast_channel *remote_chan = NULL;

	if (!user || !user->chan || !user->conference) {
		return;
	}
	if (!ast_strlen_zero(user->related_ext) &&
		ast_channel_name(user->chan) &&
		strcasestr(ast_channel_name(user->chan), "PJSIP/conf")) {
			char channel_name[AST_CHANNEL_NAME] = {0};
			snprintf(channel_name, sizeof(channel_name), "Local/%s", user->related_ext);
			remote_chan = ast_channel_get_by_name_prefix(channel_name, strlen(channel_name));
			if (remote_chan) {
				const char *conf_room = pbx_builtin_getvar_helper(remote_chan, "CONF_ROOM");
				if (!ast_strlen_zero(conf_room) &&
					!strcasecmp(user->conference->name, conf_room)) {
					ast_log(LOG_DEBUG,"softhangup channel name %s", ast_channel_name(remote_chan));
					ast_softhangup(remote_chan, AST_SOFTHANGUP_DEV);
				}
				ast_channel_unref(remote_chan);
			}
		}
}
static int conf_status_astdb_del(const char *family_name, const char *value, const char *key_name)
{
	int res;
	char family[128]={0};
	char key[50]={0};

	if (ast_strlen_zero(family_name)) {
		ast_log(LOG_ERROR,"No key_name specified");
		return 0;
	}
	if (ast_strlen_zero(value)) {
		ast_log(LOG_ERROR,"No value specified");
		return 0;
	}

	if (ast_strlen_zero(key_name))
	{
		snprintf(family, sizeof(family),"%s/%s", family_name, value);
	} else {
		snprintf(family, sizeof(family),"%s/%s", family_name, value);
		snprintf(key,sizeof(key),"%s", key_name);
	}
	ast_log(LOG_DEBUG,"family %s key %s", family, key);

	res = ast_db_deltree(family,key);
	if (res) {
		ast_log(LOG_ERROR,"Failed to update entry");
	} else {
		ast_log(LOG_DEBUG,"Success to update entry");
	}
	return 0;
}

static void im_conf_list_remove(const char *conf_number)
{
	struct im_conference *conference = NULL;

	if (ast_strlen_zero(conf_number)) {
		return;
	}

	ast_log(LOG_DEBUG, "Remove the im conference '%s'.\n", conf_number);
	AST_LIST_LOCK(&im_conference_list);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&im_conference_list, conference, entry)
	{
		if (!strcasecmp(conf_number, "all") || !strcmp(conference->name, conf_number)) {
			AST_LIST_REMOVE_CURRENT(entry);
			ao2_cleanup(conference);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&im_conference_list);
}

static void im_conf_list_remove_error(void)
{
	struct im_conference *conference = NULL;

	AST_LIST_LOCK(&im_conference_list);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&im_conference_list, conference, entry)
	{
		if (conference->status != CONF_IM_INUSE) {
			/*if im group conference not entry in 60s we clean it.*/
			if ((time(NULL) - conference->im_conference_start) >= 60) {
				ast_log(LOG_DEBUG, "Remove the timeout im conference '%s'.\n", conference->name);
				conf_status_astdb_del("CONF_STATUS", conference->name, NULL);
				conf_status_astdb_del("MCB", conference->name, NULL);
				conf_status_astdb_del("IM_CHAT_GROUP", conference->im_group_id, NULL);
				AST_LIST_REMOVE_CURRENT(entry);
				ao2_cleanup(conference);
				break;
			}
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	AST_LIST_UNLOCK(&im_conference_list);
}

static char *get_link_result(const char *link_path,const char *local_path, const char *nas_path, char *result)
{
	char full_path[AST_MAX_PATH_LENGTH] = {0};
	char mount_path[AST_MAX_PATH_LENGTH] = {0};
	char usb1_path[AST_MAX_PATH_LENGTH] = {0};
	char usb2_path[AST_MAX_PATH_LENGTH] = {0};
	char remote_direct[AST_MAX_PATH_LENGTH] = {0};

	if (link_path == NULL) {
		return NULL;
	}

	ast_safe_nvram_get("Device_USB1_path", usb1_path, sizeof(usb1_path) - 1);
	ast_safe_nvram_get("Device_USB2_path", usb2_path, sizeof(usb2_path) - 1);

	memset(full_path, 0, sizeof(full_path));
	if (readlink(link_path, full_path, sizeof(full_path) - 1) != -1) {
		ast_log(LOG_DEBUG, "get_link_result, read link success\n");
	}

	ast_log(LOG_DEBUG, "get_link_result full_path = %s, usb1_path=%s, usb2_path=%s\n", full_path, usb1_path, usb2_path);

	ast_safe_nvram_get("netdisk_remote_direct", remote_direct, sizeof(remote_direct) - 1);
	if (remote_direct[0] != '\0') {
		snprintf(mount_path, sizeof(mount_path), "/media/%s", remote_direct);
	}

	ast_log(LOG_DEBUG, "get_link_result mount_path = %s", mount_path);

	if (local_path != NULL && strncmp(full_path, local_path, strlen(local_path)) == 0) {
		strcpy(result, "local");
	} else if(strncmp(full_path, "/media/mmcblk", strlen("/media/mmcblk")) == 0) {
		strcpy(result, "SD");
	} else if(strncmp(full_path, "/media/sd", strlen("/media/sd")) == 0) {
		if (usb1_path != NULL && usb1_path[0] != '\0' && strstr(full_path, usb1_path) != NULL) {
			strcpy(result, "USB1");
		} else if (usb2_path != NULL && usb2_path[0] != '\0' && strstr(full_path, usb2_path) != NULL) {
			strcpy(result, "USB2");
		} else {
			strcpy(result, "wrong");
		}
	} else if(nas_path != NULL && strncmp(full_path, nas_path, strlen(nas_path)) == 0) {
		strcpy(result, "netdisk");
	} else {
		strcpy(result, "wrong");
	}

	if (access(full_path, F_OK) != 0) {
		strcpy(result, "refresh");
	}

	ast_log(LOG_DEBUG, "get_link_result, result=%s\n", result);

	return result;
}

static void set_conference_record_link(struct confbridge_conference *conference, enum conf_record_type type)
{
    char record_result[32] = {0};
	char conf_recording[256] = { 0 };
	char conf_recording_local[256] = { 0 };
	char conf_recording_nas[256] = { 0 };

	if (!conference) {
		return;
	}

	switch (type) {
		case CONF_RECORD_AUDIO:
			snprintf(conf_recording, sizeof(conf_recording), "%s/meetme", ast_config_AST_SPOOL_DIR);
			snprintf(conf_recording_local, sizeof(conf_recording_local), "%s/meetme_local", ast_config_AST_SPOOL_DIR);
			snprintf(conf_recording_nas, sizeof(conf_recording_nas), "%s/meetme_nas", ast_config_AST_SPOOL_DIR);
			if (get_link_result(conf_recording, conf_recording_local, conf_recording_nas, record_result) == NULL) {
				snprintf(record_result, sizeof(record_result), "wrong");
			}
			break;
		case CONF_RECORD_VIDEO:
		case CONF_RECORD_SLIDES:
		case CONF_RECORD_FOUCS:
			snprintf(conf_recording, sizeof(conf_recording), "%s/video", ast_config_AST_SPOOL_DIR);
			snprintf(conf_recording_nas, sizeof(conf_recording_nas), "%s/video_nas", ast_config_AST_SPOOL_DIR);
#ifdef BUILD_ON_CLOUD_PBX
			snprintf(conf_recording_local, sizeof(conf_recording_local), "%s/video_local", ast_config_AST_SPOOL_DIR);
			if (get_link_result(conf_recording, conf_recording_local, conf_recording_nas, record_result) == NULL) {
				snprintf(record_result, sizeof(record_result), "wrong");
			}
#else
			if (get_link_result(conf_recording, NULL, conf_recording_nas, record_result) == NULL) {
				snprintf(record_result, sizeof(record_result), "wrong");
			}
#endif
			break;
		default:
			break;
	}

	strncpy(conference->record_link, record_result, sizeof(conference->record_link) - 1);

	return;
}
#endif

/*! \brief Called when module is being unloaded */
static int unload_module(void)
{
	ast_unregister_application(app);
#ifdef GRANDSTREAM_NETWORKS
	ast_unregister_application(app_sfu);
	ast_unregister_application(app_confearlymedia);
	ast_unregister_application(app_confbridge_command);
	ast_unregister_application(app_conffloorcontrl);
	ast_unregister_application(app_pageconf_change_caller);
	ast_unregister_application(app_privateintercom_callflip);
	ast_unregister_application(app_confrecordcontrl);
#endif

	ast_custom_function_unregister(&confbridge_function);
	ast_custom_function_unregister(&confbridge_info_function);
#ifdef GRANDSTREAM_NETWORKS
	ast_custom_function_unregister(&confbridge_sfu_info_function);
	ast_custom_function_unregister(&confbridge_im_conf_function);
	ast_custom_function_unregister(&confbridge_user_terminal_type_function);
	ast_custom_function_unregister(&confbridge_early_join_function);
	ast_custom_function_unregister(&confbridge_schedule_check_role_function);
	ast_custom_function_unregister(&confbridge_user_related_info_function);
#endif

	ast_cli_unregister_multiple(cli_confbridge, ARRAY_LEN(cli_confbridge));

	ast_manager_unregister("ConfbridgeList");
	ast_manager_unregister("ConfbridgeListRooms");
	ast_manager_unregister("ConfbridgeMute");
	ast_manager_unregister("ConfbridgeUnmute");
	ast_manager_unregister("ConfbridgeKick");
	ast_manager_unregister("ConfbridgeUnlock");
	ast_manager_unregister("ConfbridgeLock");
	ast_manager_unregister("ConfbridgeStartRecord");
	ast_manager_unregister("ConfbridgeStopRecord");
	ast_manager_unregister("ConfbridgeSetSingleVideoSrc");
#ifdef GRANDSTREAM_NETWORKS
	ast_manager_unregister("ConfbridgeSfuAuthorizeUser");
	ast_manager_unregister("ConfbridgeSfuTransferModerator");
	ast_manager_unregister("ConfbridgeSfuSwitchVideoQuality");
	ast_manager_unregister("ConfbridgeAuthorizeUser");
	ast_manager_unregister("ConfbridgeTransferModerator");
	ast_manager_unregister("ConfbridgeRenameUser");
	ast_manager_unregister("ConfbridgeUpdateAdminPin");
	ast_manager_unregister("ConfbridgeReportStatisAndRemoveSchedule");
	ast_manager_unregister("ConfbridgeImAuthorize");
	ast_manager_unregister("ConfbridgeScheduleStart");
	ast_manager_unregister("ConfbridgeScheduleEnd");
	ast_manager_unregister("ConfbridgeScheduleJoinEarly");
	ast_manager_unregister("ConfbridgeScheduleWaitInfoModify");
	ast_manager_unregister("ConfbridgeDumpPCMData");
	ast_manager_unregister("ConfbridgeHaSyncRecordCount");
	ast_manager_unregister("ConfbridgeRemoveModeratorCache");//added by zyxia for callflip
	ast_manager_unregister("ConfbridgeSfuToggleRecord");
	ast_manager_unregister("ConfbridgeToggleAutoRecord");
	ast_unregister_operation_log_interface("conference");
	ast_manager_unregister("ConfbridgeStartVideoRecord");
	ast_manager_unregister("ConfbridgeStopVideoRecord");
	ast_manager_unregister("ConfbridgeEnableMCU");
	ast_manager_unregister("ConfbridgeDisableMCU");
	ast_manager_unregister("ConfbridgeVideoRecordAlarmOn");
	ast_manager_unregister("ConfbridgeVideoRecordAlarmOff");
	ast_manager_unregister("VideoRecordAlarmOn");
	ast_manager_unregister("VideoRecordAlarmOff");
	ast_manager_unregister("ConfbridgeStartSlidesRecord");
	ast_manager_unregister("ConfbridgeStopSlidesRecord");
	ast_manager_unregister("ConfbridgeAllowExtendTime");
	ast_manager_unregister("ConfbridgeRejectExtendTime");
	ast_manager_unregister("ConfbridgeSetLayout");
	ast_manager_unregister("ConfbridgeCancelLayout");
	ast_manager_unregister("ConfbridgeSetUserPermission");
	ast_manager_unregister("ConfbridgeCancelUserPermission");
	ast_manager_unregister("DeviceMonitor");
	ast_manager_unregister("ConfbridgeRemoveWaitUsers");
#endif

	/* Unsubscribe from stasis confbridge message type and clean it up. */
	manager_confbridge_shutdown();
#ifdef GRANDSTREAM_NETWORKS
#if 0
	if (mcb_release_thread != AST_PTHREADT_NULL) {
		stop_mcb_release_thread();
	}
#endif
	manager_cei_shutdown();

	if (ccm_command_router) {
		stasis_message_router_unsubscribe(ccm_command_router);
		ccm_command_router = NULL;
	}

	if (cei_toggle_enable_router) {
		stasis_message_router_unsubscribe(cei_toggle_enable_router);
		cei_toggle_enable_router = NULL;
	}

	if (operation_log_router) {
		stasis_message_router_unsubscribe(operation_log_router);
		operation_log_router = NULL;
	}

	destroy_conf_contact_groups();
	cache_conf_remove("all");
	schedule_conference_remove("all","");
	im_conf_list_remove("all");
	schedule_conference_wait_list_remove("all");

	unload_conf_resource();
#ifdef HAVE_UNIXODBC
	ast_unregister_application(app_cbl);
#endif
	if (conf_sched) {
		ast_sched_context_destroy(conf_sched);
		conf_sched = NULL;
	}
	if (user_sched) {
		ast_sched_context_destroy(user_sched);
		user_sched = NULL;
	}
	if (recycle_sched) {
		ast_sched_context_destroy(recycle_sched);
		recycle_sched = NULL;
	}
#endif

	/* Get rid of the conference bridges container. Since we only allow dynamic ones none will be active. */
	ao2_cleanup(conference_bridges);
	conference_bridges = NULL;

	conf_destroy_config();

	unregister_channel_tech(conf_announce_get_tech());
	unregister_channel_tech(conf_record_get_tech());

	return 0;
}

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	int res = 0;
#ifdef GRANDSTREAM_NETWORKS
	char sqlprefix[AST_DB_NAME_MAX_LEN] = {0};
	struct stasis_topic *ccm_command_topic = NULL;
	struct stasis_topic *cei_toggle_enable_topic = NULL;
	struct stasis_topic *operation_log_topic = NULL;
	globalEnabledCEI = ast_cei_get_enable_configuration();

	if (ast_safe_nvram_get("sqlprefix", sqlprefix, sizeof(sqlprefix)) != NULL) {
		snprintf(s_dbAmiName, sizeof(s_dbAmiName), "%s%s", sqlprefix, "ami_status");
		snprintf(s_dbConfigName, sizeof(s_dbConfigName), "%s%s", sqlprefix, "ucm_config");
		snprintf(s_dbOperationLog, sizeof(s_dbOperationLog), "%s%s", sqlprefix, "operation_log");
	}
#endif

	if (conf_load_config()) {
		ast_log(LOG_ERROR, "Unable to load config. Not loading module.\n");
		return AST_MODULE_LOAD_DECLINE;
	}

	if (register_channel_tech(conf_record_get_tech())
		|| register_channel_tech(conf_announce_get_tech())) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	/* Create a container to hold the conference bridges */
	conference_bridges = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
		CONFERENCE_BRIDGE_BUCKETS,
		conference_bridge_hash_cb, NULL, conference_bridge_cmp_cb);
	if (!conference_bridges) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

		/* Handle statsis topics and types of CCM commands*/
#ifdef GRANDSTREAM_NETWORKS
	ccm_command_topic = ast_ccm_command_topic_all();
	if (!ccm_command_topic) {
		ast_log(LOG_WARNING, "Failed to get ast_ccm_command_topic_all\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	ccm_command_router = stasis_message_router_create(ccm_command_topic);
	if (ccm_command_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create ccm_command_router\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(ccm_command_router,
								  ast_ccm_command_kick_type(),
								  ccm_command_kick_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ccm_command_kick_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(ccm_command_router,
								  ast_ccm_command_mute_unmute_type(),
								  ccm_command_mute_unmute_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ccm_command_mute_unmute_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(ccm_command_router,
								  ast_ccm_command_lock_type(),
								  ccm_command_lock_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ccm_command_lock_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(ccm_command_router,
								  ast_ccm_command_unlock_type(),
								  ccm_command_unlock_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ccm_command_unlock_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	cei_toggle_enable_topic = ast_cei_toggle_enable_topic_all();
	if (!cei_toggle_enable_topic) {
		ast_log(LOG_WARNING, "[WL] Failed to get cei_toggle_enable_topic\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	cei_toggle_enable_router = stasis_message_router_create(cei_toggle_enable_topic);
	if (cei_toggle_enable_router == NULL) {
		ast_log(LOG_WARNING, "[WL] Failed to create cei_toggle_enable_router\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(cei_toggle_enable_router,
									ast_cei_toggle_enable_type(),
									cei_toggle_enable_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_toggle_enable_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	operation_log_topic = ast_operation_log_topic_all();
	if (!operation_log_topic) {
		ast_log(LOG_WARNING, "[WL] Failed to get operation_log_topic\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	operation_log_router = stasis_message_router_create(operation_log_topic);
	if (operation_log_router == NULL) {
		ast_log(LOG_WARNING, "[WL] Failed to create operation_log_router\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	if (stasis_message_router_add(operation_log_router,
									ast_conf_operation_log_type(),
									conf_operation_log_stasis_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_toggle_enable_type\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}
#if 0
	start_mcb_release_thread();
#endif
	conference_contact_groups = ao2_container_alloc_list(AO2_ALLOC_OPT_LOCK_MUTEX, 0, NULL, conf_contact_groups_cmp_cb);
	if (!conference_contact_groups) {
		ast_log(LOG_ERROR, "Unable to create a contact groups container.\n");
		unload_module();
		return AST_MODULE_LOAD_FAILURE;
	}
	if (load_conf_contact_groups_conf(0)) {
		ast_log(LOG_ERROR, "Unable to load contact groups config. Not loading module.\n");
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}
	load_conf_resource();

#ifdef HAVE_UNIXODBC
	count = conf_report_get_call_count();
#endif
#endif

	/* Setup manager stasis subscriptions */
	res |= manager_confbridge_init();

#ifdef GRANDSTREAM_NETWORKS
	res |= manager_cei_init();

#ifdef HAVE_UNIXODBC
	res |= ast_register_application_xml(app_cbl, confbridge_log_exec);
#endif
#endif

	res |= ast_register_application_xml(app, confbridge_exec);
#ifdef GRANDSTREAM_NETWORKS
	res |= ast_register_application_xml(app_sfu, confbridge_sfu_helper);
	res |= ast_register_application_xml(app_confearlymedia, confbridge_early_media);
	res |= ast_register_application_xml(app_confbridge_command, confbridge_command);
	res |= ast_register_application_xml(app_conffloorcontrl, confbridge_floor_contrl);
	res |= ast_register_application_xml(app_pageconf_change_caller, pageconf_change_caller);
	res |= ast_register_application_xml(app_privateintercom_callflip, privateintercom_callflip);
	res |= ast_register_application_xml(app_confrecordcontrl, confbridge_record_contrl);
#endif
	res |= ast_custom_function_register_escalating(&confbridge_function, AST_CFE_WRITE);
	res |= ast_custom_function_register(&confbridge_info_function);
#ifdef GRANDSTREAM_NETWORKS
	res |= ast_custom_function_register(&confbridge_sfu_info_function);
	res |= ast_custom_function_register(&confbridge_im_conf_function);
	res |= ast_custom_function_register(&confbridge_user_terminal_type_function);
	res |= ast_custom_function_register(&confbridge_early_join_function);
	res |= ast_custom_function_register(&confbridge_schedule_check_role_function);
	res |= ast_custom_function_register(&confbridge_user_related_info_function);
#endif

	res |= ast_cli_register_multiple(cli_confbridge, ARRAY_LEN(cli_confbridge));

	res |= ast_manager_register_xml("ConfbridgeList", EVENT_FLAG_REPORTING, action_confbridgelist);
	res |= ast_manager_register_xml("ConfbridgeListRooms", EVENT_FLAG_REPORTING, action_confbridgelistrooms);
	res |= ast_manager_register_xml("ConfbridgeMute", EVENT_FLAG_CALL, action_confbridgemute);
	res |= ast_manager_register_xml("ConfbridgeUnmute", EVENT_FLAG_CALL, action_confbridgeunmute);
	res |= ast_manager_register_xml("ConfbridgeKick", EVENT_FLAG_CALL, action_confbridgekick);
	res |= ast_manager_register_xml("ConfbridgeUnlock", EVENT_FLAG_CALL, action_confbridgeunlock);
	res |= ast_manager_register_xml("ConfbridgeLock", EVENT_FLAG_CALL, action_confbridgelock);
	res |= ast_manager_register_xml("ConfbridgeStartRecord", EVENT_FLAG_SYSTEM, action_confbridgestartrecord);
	res |= ast_manager_register_xml("ConfbridgeStopRecord", EVENT_FLAG_SYSTEM, action_confbridgestoprecord);
	res |= ast_manager_register_xml("ConfbridgeSetSingleVideoSrc", EVENT_FLAG_CALL, action_confbridgesetsinglevideosrc);
#ifdef GRANDSTREAM_NETWORKS
	res |= ast_manager_register_xml("ConfbridgeSfuAuthorizeUser", EVENT_FLAG_CALL, action_confbridge_sfu_authorize_user);
	res |= ast_manager_register_xml("ConfbridgeSfuTransferModerator", EVENT_FLAG_CALL, action_confbridge_sfu_transfer_moderator);
	res |= ast_manager_register_xml("ConfbridgeSfuSwitchVideoQuality", EVENT_FLAG_CALL, action_confbridge_sfu_switch_video_quality);
	res |= ast_manager_register_xml("ConfbridgeSfuToggleRecord", EVENT_FLAG_CALL, action_confbridge_sfu_toggle_record);
	res |= ast_manager_register_xml("ConfbridgeToggleAutoRecord", EVENT_FLAG_CALL, action_confbridge_toggle_auto_record);
	res |= ast_manager_register_xml("ConfbridgeAuthorizeUser", EVENT_FLAG_CALL, action_confbridge_authorize_user);
	res |= ast_manager_register_xml("ConfbridgeTransferModerator", EVENT_FLAG_CALL, action_confbridge_transfer_moderator);
	res |= ast_manager_register_xml("ConfbridgeRenameUser", EVENT_FLAG_CALL, action_confbridge_rename_user);
	res |= ast_manager_register_xml("ConfbridgeUpdateAdminPin", EVENT_FLAG_CALL, action_confbridge_update_admin_pincode);
	res |= ast_manager_register_xml("ConfbridgeReportStatisAndRemoveSchedule", EVENT_FLAG_CALL, action_confbridge_report_statistics_remove_schedule);
	res |= ast_manager_register_xml("ConfbridgeImAuthorize", EVENT_FLAG_CALL, action_confbridge_im_authorize);
	res |= ast_manager_register_xml("ConfbridgeScheduleStart", EVENT_FLAG_CALL, action_confbridge_schedule_start);
	res |= ast_manager_register_xml("ConfbridgeScheduleEnd", EVENT_FLAG_CALL, action_confbridge_schedule_end);
	res |= ast_manager_register_xml("ConfbridgeScheduleJoinEarly", EVENT_FLAG_CALL, schedule_conference_join_early);
	res |= ast_manager_register_xml("ConfbridgeScheduleWaitInfoModify", EVENT_FLAG_CALL, schedule_conference_wait_info_modify);
	res |= ast_manager_register_xml("ConfbridgeDumpPCMData", EVENT_FLAG_SYSTEM, action_confbridge_dump_pcm_data);
	res |= ast_manager_register_xml("ConfbridgeHaSyncRecordCount", EVENT_FLAG_SYSTEM, action_confbridge_ha_sync_record_count);
	res |= ast_manager_register_xml("ConfbridgeRemoveModeratorCache", EVENT_FLAG_CALL, action_confbridge_remove_moderator_cache);//added by zyxia for  call flip,2021.8.16
	res |= ast_manager_register_xml("ConfbridgeStartVideoRecord", EVENT_FLAG_SYSTEM, action_confbridgestartvideorecord);
	res |= ast_manager_register_xml("ConfbridgeStopVideoRecord", EVENT_FLAG_SYSTEM, action_confbridgestopvideorecord);
	res |= ast_manager_register_xml("ConfbridgeEnableMCU", EVENT_FLAG_SYSTEM, action_confbridgeenablemcu);
	res |= ast_manager_register_xml("ConfbridgeDisableMCU", EVENT_FLAG_SYSTEM, action_confbridgedisablemcu);
	res |= ast_manager_register_xml("ConfbridgeVideoRecordAlarmOn", EVENT_FLAG_SYSTEM, action_videorecordalarmon);
	res |= ast_manager_register_xml("ConfbridgeVideoRecordAlarmOff", EVENT_FLAG_SYSTEM, action_videorecordalarmoff);
	res |= ast_manager_register_xml("ConfbridgeStartSlidesRecord", EVENT_FLAG_SYSTEM, action_confbridgestartslidesrecord);
	res |= ast_manager_register_xml("ConfbridgeStopSlidesRecord", EVENT_FLAG_SYSTEM, action_confbridgestopslidesrecord);
	res |= ast_manager_register_xml("ConfbridgeAllowExtendTime", EVENT_FLAG_SYSTEM, action_confbridge_allow_extend_time);
	res |= ast_manager_register_xml("ConfbridgeRejectExtendTime", EVENT_FLAG_SYSTEM, action_confbridge_reject_extend_time);
	res |= ast_manager_register_xml("ConfbridgeSetLayout", EVENT_FLAG_SYSTEM, action_confbridge_set_layout);
	res |= ast_manager_register_xml("ConfbridgeCancelLayout", EVENT_FLAG_SYSTEM, action_confbridge_cancel_layout);
	res |= ast_manager_register_xml("ConfbridgeSetUserPermission", EVENT_FLAG_SYSTEM, action_confbridge_set_user_permission);
	res |= ast_manager_register_xml("ConfbridgeCancelUserPermission", EVENT_FLAG_SYSTEM, action_confbridge_cancel_user_permission);
	res |= ast_manager_register_xml("DeviceMonitor", EVENT_FLAG_CALL, action_device_monitor);
	res |= ast_manager_register_xml("ConfbridgeRemoveWaitUsers", EVENT_FLAG_SYSTEM, action_confbridge_remove_wait_users);

	res |= ast_register_operation_log_interface("conference", conf_operation_log_interface);

	conf_sched = ast_sched_context_create();
	if (conf_sched) {
		if (ast_sched_start_thread(conf_sched)) {
			ast_log(LOG_ERROR, "Could not start conf scheduler thread\n");
			unload_module();
			return AST_MODULE_LOAD_FAILURE;
		}
	}
	user_sched = ast_sched_context_create();
	if (user_sched) {
		if (ast_sched_start_thread(user_sched)) {
			ast_log(LOG_ERROR, "Could not start user scheduler thread\n");
			unload_module();
			return AST_MODULE_LOAD_FAILURE;
		}
	}
	recycle_sched = ast_sched_context_create();
	if (recycle_sched) {
		if (ast_sched_start_thread(recycle_sched)) {
			ast_log(LOG_ERROR, "Could not start mcb user scheduler thread\n");
			unload_module();
			return AST_MODULE_LOAD_FAILURE;
		}
	}
#endif

	if (res) {
		unload_module();
		return AST_MODULE_LOAD_DECLINE;
	}

	return AST_MODULE_LOAD_SUCCESS;
}

static int reload(void)
{
#ifdef  GRANDSTREAM_NETWORKS
	globalEnabledCEI = ast_cei_get_enable_configuration();
	load_conf_contact_groups_conf(1);
	if (reload_conf_resource() < 0) {
		return AST_MODULE_LOAD_FAILURE;
	}
#endif

	return conf_reload_config();
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Conference Bridge Application",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload,
	.load_pri = AST_MODPRI_DEVSTATE_PROVIDER,
	.optional_modules = "codec_speex,func_jitterbuffer",
);
