/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2012, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief dial() & retrydial() - Trivial application to dial a channel and send an URL on answer
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \ingroup applications
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/


#include "asterisk.h"

#include <sys/time.h>
#include <signal.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "asterisk/paths.h" /* use ast_config_AST_DATA_DIR */
#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/translate.h"
#include "asterisk/say.h"
#include "asterisk/config.h"
#include "asterisk/features.h"
#include "asterisk/musiconhold.h"
#include "asterisk/callerid.h"
#include "asterisk/utils.h"
#include "asterisk/app.h"
#include "asterisk/causes.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/manager.h"
#include "asterisk/privacy.h"
#include "asterisk/stringfields.h"
#include "asterisk/dsp.h"
#include "asterisk/aoc.h"
#include "asterisk/ccss.h"
#include "asterisk/indications.h"
#include "asterisk/framehook.h"
#include "asterisk/dial.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/bridge_after.h"
#include "asterisk/features_config.h"
#include "asterisk/max_forwards.h"
#include "asterisk/stream.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/astdb.h"
#include "asterisk/avs_controller.h"
#include "asterisk/core_local.h"
#include "asterisk/security_events.h"
#include "asterisk/ucm_dbi.h"
#include "asterisk/app.h"

#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#include "asterisk/avs_controller.h"
#endif
AST_MODULE_LOG("dial");
#endif

/*** DOCUMENTATION
	<application name="Dial" language="en_US">
		<synopsis>
			Attempt to connect to another device or endpoint and bridge the call.
		</synopsis>
		<syntax>
			<parameter name="Technology/Resource" required="true" argsep="&amp;">
				<argument name="Technology/Resource" required="true">
					<para>Specification of the device(s) to dial.  These must be in the format of
					<literal>Technology/Resource</literal>, where <replaceable>Technology</replaceable>
					represents a particular channel driver, and <replaceable>Resource</replaceable>
					represents a resource available to that particular channel driver.</para>
				</argument>
				<argument name="Technology2/Resource2" required="false" multiple="true">
					<para>Optional extra devices to dial in parallel</para>
					<para>If you need more than one enter them as
					Technology2/Resource2&amp;Technology3/Resource3&amp;.....</para>
				</argument>
			</parameter>
			<parameter name="timeout" required="false">
				<para>Specifies the number of seconds we attempt to dial the specified devices.</para>
				<para>If not specified, this defaults to 136 years.</para>
			</parameter>
			<parameter name="options" required="false">
				<optionlist>
				<option name="A">
					<argument name="x" required="true">
						<para>The file to play to the called party</para>
					</argument>
					<para>Play an announcement to the called party, where <replaceable>x</replaceable> is the prompt to be played</para>
				</option>
				<option name="a">
					<para>Immediately answer the calling channel when the called channel answers in
					all cases. Normally, the calling channel is answered when the called channel
					answers, but when options such as <literal>A()</literal> and
					<literal>M()</literal> are used, the calling channel is
					not answered until all actions on the called channel (such as playing an
					announcement) are completed.  This option can be used to answer the calling
					channel before doing anything on the called channel. You will rarely need to use
					this option, the default behavior is adequate in most cases.</para>
				</option>
				<option name="b" argsep="^">
					<para>Before initiating an outgoing call, <literal>Gosub</literal> to the specified
					location using the newly created channel.  The <literal>Gosub</literal> will be
					executed for each destination channel.</para>
					<argument name="context" required="false" />
					<argument name="exten" required="false" />
					<argument name="priority" required="true" hasparams="optional" argsep="^">
						<argument name="arg1" multiple="true" required="true" />
						<argument name="argN" />
					</argument>
				</option>
				<option name="B" argsep="^">
					<para>Before initiating the outgoing call(s), <literal>Gosub</literal> to the
					specified location using the current channel.</para>
					<argument name="context" required="false" />
					<argument name="exten" required="false" />
					<argument name="priority" required="true" hasparams="optional" argsep="^">
						<argument name="arg1" multiple="true" required="true" />
						<argument name="argN" />
					</argument>
				</option>
				<option name="C">
					<para>Reset the call detail record (CDR) for this call.</para>
				</option>
				<option name="c">
					<para>If the Dial() application cancels this call, always set
					<variable>HANGUPCAUSE</variable> to 'answered elsewhere'</para>
				</option>
				<option name="d">
					<para>Allow the calling user to dial a 1 digit extension while waiting for
					a call to be answered. Exit to that extension if it exists in the
					current context, or the context defined in the <variable>EXITCONTEXT</variable> variable,
					if it exists.</para>
					<para>NOTE: Many SIP and ISDN phones cannot send DTMF digits until the call is
					connected.  If you wish to use this option with these phones, you
					can use the <literal>Answer</literal> application before dialing.</para>
				</option>
				<option name="D" argsep=":">
					<argument name="called" />
					<argument name="calling" />
					<argument name="progress" />
					<para>Send the specified DTMF strings <emphasis>after</emphasis> the called
					party has answered, but before the call gets bridged.  The
					<replaceable>called</replaceable> DTMF string is sent to the called party, and the
					<replaceable>calling</replaceable> DTMF string is sent to the calling party.  Both arguments
					can be used alone.  If <replaceable>progress</replaceable> is specified, its DTMF is sent
					to the called party immediately after receiving a <literal>PROGRESS</literal> message.</para>
					<para>See <literal>SendDTMF</literal> for valid digits.</para>
				</option>
				<option name="e">
					<para>Execute the <literal>h</literal> extension for peer after the call ends</para>
				</option>
				<option name="f">
					<argument name="x" required="false" />
					<para>If <replaceable>x</replaceable> is not provided, force the CallerID sent on a call-forward or
					deflection to the dialplan extension of this <literal>Dial()</literal> using a dialplan <literal>hint</literal>.
					For example, some PSTNs do not allow CallerID to be set to anything
					other than the numbers assigned to you.
					If <replaceable>x</replaceable> is provided, force the CallerID sent to <replaceable>x</replaceable>.</para>
				</option>
				<option name="F" argsep="^">
					<argument name="context" required="false" />
					<argument name="exten" required="false" />
					<argument name="priority" required="true" />
					<para>When the caller hangs up, transfer the <emphasis>called</emphasis> party
					to the specified destination and <emphasis>start</emphasis> execution at that location.</para>
					<para>NOTE: Any channel variables you want the called channel to inherit from the caller channel must be
					prefixed with one or two underbars ('_').</para>
				</option>
				<option name="F">
					<para>When the caller hangs up, transfer the <emphasis>called</emphasis> party to the next priority of the current extension
					and <emphasis>start</emphasis> execution at that location.</para>
					<para>NOTE: Any channel variables you want the called channel to inherit from the caller channel must be
					prefixed with one or two underbars ('_').</para>
					<para>NOTE: Using this option from a Macro() or GoSub() might not make sense as there would be no return points.</para>
				</option>
				<option name="g">
					<para>Proceed with dialplan execution at the next priority in the current extension if the
					destination channel hangs up.</para>
				</option>
				<option name="G" argsep="^">
					<argument name="context" required="false" />
					<argument name="exten" required="false" />
					<argument name="priority" required="true" />
					<para>If the call is answered, transfer the calling party to
					the specified <replaceable>priority</replaceable> and the called party to the specified
					<replaceable>priority</replaceable> plus one.</para>
					<para>NOTE: You cannot use any additional action post answer options in conjunction with this option.</para>
				</option>
				<option name="h">
					<para>Allow the called party to hang up by sending the DTMF sequence
					defined for disconnect in <filename>features.conf</filename>.</para>
				</option>
				<option name="H">
					<para>Allow the calling party to hang up by sending the DTMF sequence
					defined for disconnect in <filename>features.conf</filename>.</para>
					<para>NOTE: Many SIP and ISDN phones cannot send DTMF digits until the call is
					connected.  If you wish to allow DTMF disconnect before the dialed
					party answers with these phones, you can use the <literal>Answer</literal>
					application before dialing.</para>
				</option>
				<option name="i">
					<para>Asterisk will ignore any forwarding requests it may receive on this dial attempt.</para>
				</option>
				<option name="I">
					<para>Asterisk will ignore any connected line update requests or any redirecting party
					update requests it may receive on this dial attempt.</para>
				</option>
				<option name="k">
					<para>Allow the called party to enable parking of the call by sending
					the DTMF sequence defined for call parking in <filename>features.conf</filename>.</para>
				</option>
				<option name="K">
					<para>Allow the calling party to enable parking of the call by sending
					the DTMF sequence defined for call parking in <filename>features.conf</filename>.</para>
				</option>
				<option name="L" argsep=":">
					<argument name="x" required="true">
						<para>Maximum call time, in milliseconds</para>
					</argument>
					<argument name="y">
						<para>Warning time, in milliseconds</para>
					</argument>
					<argument name="z">
						<para>Repeat time, in milliseconds</para>
					</argument>
					<para>Limit the call to <replaceable>x</replaceable> milliseconds. Play a warning when <replaceable>y</replaceable> milliseconds are
					left. Repeat the warning every <replaceable>z</replaceable> milliseconds until time expires.</para>
					<para>This option is affected by the following variables:</para>
					<variablelist>
						<variable name="LIMIT_PLAYAUDIO_CALLER">
							<value name="yes" default="true" />
							<value name="no" />
							<para>If set, this variable causes Asterisk to play the prompts to the caller.</para>
						</variable>
						<variable name="LIMIT_PLAYAUDIO_CALLEE">
							<value name="yes" />
							<value name="no" default="true"/>
							<para>If set, this variable causes Asterisk to play the prompts to the callee.</para>
						</variable>
						<variable name="LIMIT_TIMEOUT_FILE">
							<value name="filename"/>
							<para>If specified, <replaceable>filename</replaceable> specifies the sound prompt to play when the timeout is reached.
							If not set, the time remaining will be announced.</para>
						</variable>
						<variable name="LIMIT_CONNECT_FILE">
							<value name="filename"/>
							<para>If specified, <replaceable>filename</replaceable> specifies the sound prompt to play when the call begins.
							If not set, the time remaining will be announced.</para>
						</variable>
						<variable name="LIMIT_WARNING_FILE">
							<value name="filename"/>
							<para>If specified, <replaceable>filename</replaceable> specifies the sound prompt to play as
							a warning when time <replaceable>x</replaceable> is reached. If not set, the time remaining will be announced.</para>
						</variable>
					</variablelist>
				</option>
				<option name="m">
					<argument name="class" required="false"/>
					<para>Provide hold music to the calling party until a requested
					channel answers. A specific music on hold <replaceable>class</replaceable>
					(as defined in <filename>musiconhold.conf</filename>) can be specified.</para>
				</option>
				<option name="M" argsep="^">
					<argument name="macro" required="true">
						<para>Name of the macro that should be executed.</para>
					</argument>
					<argument name="arg" multiple="true">
						<para>Macro arguments</para>
					</argument>
					<para>Execute the specified <replaceable>macro</replaceable> for the <emphasis>called</emphasis> channel
					before connecting to the calling channel. Arguments can be specified to the Macro
					using <literal>^</literal> as a delimiter. The macro can set the variable
					<variable>MACRO_RESULT</variable> to specify the following actions after the macro is
					finished executing:</para>
					<variablelist>
						<variable name="MACRO_RESULT">
							<para>If set, this action will be taken after the macro finished executing.</para>
							<value name="ABORT">
								Hangup both legs of the call
							</value>
							<value name="CONGESTION">
								Behave as if line congestion was encountered
							</value>
							<value name="BUSY">
								Behave as if a busy signal was encountered
							</value>
							<value name="CONTINUE">
								Hangup the called party and allow the calling party to continue dialplan execution at the next priority
							</value>
							<value name="GOTO:[[&lt;context&gt;^]&lt;exten&gt;^]&lt;priority&gt;">
								Transfer the call to the specified destination.
							</value>
						</variable>
					</variablelist>
					<para>NOTE: You cannot use any additional action post answer options in conjunction
					with this option. Also, pbx services are run on the peer (called) channel,
					so you will not be able to set timeouts via the <literal>TIMEOUT()</literal> function in this macro.</para>
					<para>WARNING: Be aware of the limitations that macros have, specifically with regards to use of
					the <literal>WaitExten</literal> application. For more information, see the documentation for
					<literal>Macro()</literal>.</para>
					<para>NOTE: Macros are deprecated, GoSub should be used instead,
					see the <literal>U</literal> option.</para>
				</option>
				<option name="n">
					<argument name="delete">
						<para>With <replaceable>delete</replaceable> either not specified or set to <literal>0</literal>,
						the recorded introduction will not be deleted if the caller hangs up while the remote party has not
						yet answered.</para>
						<para>With <replaceable>delete</replaceable> set to <literal>1</literal>, the introduction will
						always be deleted.</para>
					</argument>
					<para>This option is a modifier for the call screening/privacy mode. (See the
					<literal>p</literal> and <literal>P</literal> options.) It specifies
					that no introductions are to be saved in the <directory>priv-callerintros</directory>
					directory.</para>
				</option>
				<option name="N">
					<para>This option is a modifier for the call screening/privacy mode. It specifies
					that if CallerID is present, do not screen the call.</para>
				</option>
				<option name="o">
					<argument name="x" required="false" />
					<para>If <replaceable>x</replaceable> is not provided, specify that the CallerID that was present on the
					<emphasis>calling</emphasis> channel be stored as the CallerID on the <emphasis>called</emphasis> channel.
					This was the behavior of Asterisk 1.0 and earlier.
					If <replaceable>x</replaceable> is provided, specify the CallerID stored on the <emphasis>called</emphasis> channel.
					Note that <literal>o(${CALLERID(all)})</literal> is similar to option <literal>o</literal> without the parameter.</para>
				</option>
				<option name="O">
					<argument name="mode">
						<para>With <replaceable>mode</replaceable> either not specified or set to <literal>1</literal>,
						the originator hanging up will cause the phone to ring back immediately.</para>
						<para>With <replaceable>mode</replaceable> set to <literal>2</literal>, when the operator
						flashes the trunk, it will ring their phone back.</para>
					</argument>
					<para>Enables <emphasis>operator services</emphasis> mode.  This option only
					works when bridging a DAHDI channel to another DAHDI channel
					only. if specified on non-DAHDI interfaces, it will be ignored.
					When the destination answers (presumably an operator services
					station), the originator no longer has control of their line.
					They may hang up, but the switch will not release their line
					until the destination party (the operator) hangs up.</para>
				</option>
				<option name="p">
					<para>This option enables screening mode. This is basically Privacy mode
					without memory.</para>
				</option>
				<option name="P">
					<argument name="x" />
					<para>Enable privacy mode. Use <replaceable>x</replaceable> as the family/key in the AstDB database if
					it is provided. The current extension is used if a database family/key is not specified.</para>
				</option>
				<option name="Q">
					<argument name="cause" required="true"/>
					<para>Specify the Q.850/Q.931 <replaceable>cause</replaceable> to send on
					unanswered channels when another channel answers the call.
					As with <literal>Hangup()</literal>, <replaceable>cause</replaceable>
					can be a numeric cause code or a name such as
						<literal>NO_ANSWER</literal>,
						<literal>USER_BUSY</literal>,
						<literal>CALL_REJECTED</literal> or
						<literal>ANSWERED_ELSEWHERE</literal> (the default if Q isn't specified).
						You can also specify <literal>0</literal> or <literal>NONE</literal>
						to send no cause.  See the <filename>causes.h</filename> file for the
						full list of valid causes and names.
						</para>
					<para>NOTE: chan_sip does not support setting the cause on a CANCEL to anything
					other than ANSWERED_ELSEWHERE.</para>
				</option>
				<option name="r">
					<para>Default: Indicate ringing to the calling party, even if the called party isn't actually ringing. Pass no audio to the calling
					party until the called channel has answered.</para>
					<argument name="tone" required="false">
						<para>Indicate progress to calling party. Send audio 'tone' from the <filename>indications.conf</filename> tonezone currently in use.</para>
					</argument>
				</option>
				<option name="R">
					<para>Default: Indicate ringing to the calling party, even if the called party isn't actually ringing.
					Allow interruption of the ringback if early media is received on the channel.</para>
				</option>
				<option name="S">
					<argument name="x" required="true" />
					<para>Hang up the call <replaceable>x</replaceable> seconds <emphasis>after</emphasis> the called party has
					answered the call.</para>
				</option>
				<option name="s">
					<argument name="x" required="true" />
					<para>Force the outgoing CallerID tag parameter to be set to the string <replaceable>x</replaceable>.</para>
					<para>Works with the <literal>f</literal> option.</para>
				</option>
				<option name="t">
					<para>Allow the called party to transfer the calling party by sending the
					DTMF sequence defined in <filename>features.conf</filename>. This setting does not perform policy enforcement on
					transfers initiated by other methods.</para>
				</option>
				<option name="T">
					<para>Allow the calling party to transfer the called party by sending the
					DTMF sequence defined in <filename>features.conf</filename>. This setting does not perform policy enforcement on
					transfers initiated by other methods.</para>
				</option>
				<option name="U" argsep="^">
					<argument name="x" required="true">
						<para>Name of the subroutine context to execute via <literal>Gosub</literal>.
						The subroutine execution starts in the named context at the s exten and priority 1.</para>
					</argument>
					<argument name="arg" multiple="true" required="false">
						<para>Arguments for the <literal>Gosub</literal> routine</para>
					</argument>
					<para>Execute via <literal>Gosub</literal> the routine <replaceable>x</replaceable> for the <emphasis>called</emphasis> channel before connecting
					to the calling channel. Arguments can be specified to the <literal>Gosub</literal>
					using <literal>^</literal> as a delimiter. The <literal>Gosub</literal> routine can set the variable
					<variable>GOSUB_RESULT</variable> to specify the following actions after the <literal>Gosub</literal> returns.</para>
					<variablelist>
						<variable name="GOSUB_RESULT">
							<value name="ABORT">
								Hangup both legs of the call.
							</value>
							<value name="CONGESTION">
								Behave as if line congestion was encountered.
							</value>
							<value name="BUSY">
								Behave as if a busy signal was encountered.
							</value>
							<value name="CONTINUE">
								Hangup the called party and allow the calling party
								to continue dialplan execution at the next priority.
							</value>
							<value name="GOTO:[[&lt;context&gt;^]&lt;exten&gt;^]&lt;priority&gt;">
								Transfer the call to the specified destination.
							</value>
						</variable>
					</variablelist>
					<para>NOTE: You cannot use any additional action post answer options in conjunction
					with this option. Also, pbx services are run on the <emphasis>called</emphasis> channel,
					so you will not be able to set timeouts via the <literal>TIMEOUT()</literal> function in this routine.</para>
				</option>
				<option name="u">
					<argument name = "x" required="true">
						<para>Force the outgoing callerid presentation indicator parameter to be set
						to one of the values passed in <replaceable>x</replaceable>:
						<literal>allowed_not_screened</literal>
						<literal>allowed_passed_screen</literal>
						<literal>allowed_failed_screen</literal>
						<literal>allowed</literal>
						<literal>prohib_not_screened</literal>
						<literal>prohib_passed_screen</literal>
						<literal>prohib_failed_screen</literal>
						<literal>prohib</literal>
						<literal>unavailable</literal></para>
					</argument>
					<para>Works with the <literal>f</literal> option.</para>
				</option>
				<option name="w">
					<para>Allow the called party to enable recording of the call by sending
					the DTMF sequence defined for one-touch recording in <filename>features.conf</filename>.</para>
				</option>
				<option name="W">
					<para>Allow the calling party to enable recording of the call by sending
					the DTMF sequence defined for one-touch recording in <filename>features.conf</filename>.</para>
				</option>
				<option name="x">
					<para>Allow the called party to enable recording of the call by sending
					the DTMF sequence defined for one-touch automixmonitor in <filename>features.conf</filename>.</para>
				</option>
				<option name="X">
					<para>Allow the calling party to enable recording of the call by sending
					the DTMF sequence defined for one-touch automixmonitor in <filename>features.conf</filename>.</para>
				</option>
				<option name="z">
					<para>On a call forward, cancel any dial timeout which has been set for this call.</para>
				</option>
				</optionlist>
			</parameter>
			<parameter name="URL">
				<para>The optional URL will be sent to the called party if the channel driver supports it.</para>
			</parameter>
		</syntax>
		<description>
			<para>This application will place calls to one or more specified channels. As soon
			as one of the requested channels answers, the originating channel will be
			answered, if it has not already been answered. These two channels will then
			be active in a bridged call. All other channels that were requested will then
			be hung up.</para>

			<para>Unless there is a timeout specified, the Dial application will wait
			indefinitely until one of the called channels answers, the user hangs up, or
			if all of the called channels are busy or unavailable. Dialplan execution will
			continue if no requested channels can be called, or if the timeout expires.
			This application will report normal termination if the originating channel
			hangs up, or if the call is bridged and either of the parties in the bridge
			ends the call.</para>
			<para>If the <variable>OUTBOUND_GROUP</variable> variable is set, all peer channels created by this
			application will be put into that group (as in <literal>Set(GROUP()=...</literal>).
			If the <variable>OUTBOUND_GROUP_ONCE</variable> variable is set, all peer channels created by this
			application will be put into that group (as in <literal>Set(GROUP()=...</literal>). Unlike <variable>OUTBOUND_GROUP</variable>,
			however, the variable will be unset after use.</para>

			<example title="Dial with 30 second timeout">
			 same => n,Dial(PJSIP/alice,30)
			</example>
			<example title="Parallel dial with 45 second timeout">
			 same => n,Dial(PJSIP/alice&amp;PJIP/bob,45)
			</example>
			<example title="Dial with 'g' continuation option">
			 same => n,Dial(PJSIP/alice,,g)
			 same => n,Log(NOTICE, Alice call result: ${DIALSTATUS})
			</example>
			<example title="Dial with transfer/recording features for calling party">
			 same => n,Dial(PJSIP/alice,,TX)
			</example>
			<example title="Dial with call length limit">
			 same => n,Dial(PJSIP/alice,,L(60000:30000:10000))
			</example>
			<example title="Dial alice and bob and send NO_ANSWER to bob instead of ANSWERED_ELSEWHERE when alice answers">
			 same => n,Dial(PJSIP/alice&amp;PJSIP/bob,,Q(NO_ANSWER))
			</example>
			<example title="Dial with pre-dial subroutines">
			[default]

			exten => callee_channel,1,NoOp(ARG1=${ARG1} ARG2=${ARG2})
			 same => n,Log(NOTICE, I'm called on channel ${CHANNEL} prior to it starting the dial attempt)
			 same => n,Return()

			exten => called_channel,1,NoOp(ARG1=${ARG1} ARG2=${ARG2})
			 same => n,Log(NOTICE, I'm called on outbound channel ${CHANNEL} prior to it being used to dial someone)
			 same => n,Return()

			exten => _X.,1,NoOp()
			 same => n,Dial(PJSIP/alice,,b(default^called_channel^1(my_gosub_arg1^my_gosub_arg2))B(default^callee_channel^1(my_gosub_arg1^my_gosub_arg2)))
			 same => n,Hangup()
			</example>
			<example title="Dial with post-answer subroutine executed on outbound channel">
			[my_gosub_routine]

			exten => s,1,NoOp(ARG1=${ARG1} ARG2=${ARG2})
			 same => n,Playback(hello)
			 same => n,Return()

			[default]

			exten => _X.,1,NoOp()
			 same => n,Dial(PJSIP/alice,,U(my_gosub_routine^my_gosub_arg1^my_gosub_arg2))
			 same => n,Hangup()
			</example>
			<example title="Dial into ConfBridge using 'G' option">
			 same => n,Dial(PJSIP/alice,,G(jump_to_here))
			 same => n(jump_to_here),Goto(confbridge)
			 same => n,Goto(confbridge)
			 same => n(confbridge),ConfBridge(${EXTEN})
			</example>
			<para>This application sets the following channel variables:</para>
			<variablelist>
				<variable name="DIALEDTIME">
					<para>This is the time from dialing a channel until when it is disconnected.</para>
				</variable>
				<variable name="ANSWEREDTIME">
					<para>This is the amount of time for actual call.</para>
				</variable>
				<variable name="DIALEDPEERNAME">
					<para>The name of the outbound channel that answered the call.</para>
				</variable>
				<variable name="DIALEDPEERNUMBER">
					<para>The number that was dialed for the answered outbound channel.</para>
				</variable>
				<variable name="FORWARDERNAME">
					<para>If a call forward occurred, the name of the forwarded channel.</para>
				</variable>
				<variable name="DIALSTATUS">
					<para>This is the status of the call</para>
					<value name="CHANUNAVAIL" />
					<value name="CONGESTION" />
					<value name="NOANSWER" />
					<value name="BUSY" />
					<value name="ANSWER" />
					<value name="CANCEL" />
					<value name="DONTCALL">
						For the Privacy and Screening Modes.
						Will be set if the called party chooses to send the calling party to the 'Go Away' script.
					</value>
					<value name="TORTURE">
						For the Privacy and Screening Modes.
						Will be set if the called party chooses to send the calling party to the 'torture' script.
					</value>
					<value name="INVALIDARGS" />
				</variable>
			</variablelist>
		</description>
		<see-also>
			<ref type="application">RetryDial</ref>
			<ref type="application">SendDTMF</ref>
			<ref type="application">Gosub</ref>
			<ref type="application">Macro</ref>
		</see-also>
	</application>
	<application name="RetryDial" language="en_US">
		<synopsis>
			Place a call, retrying on failure allowing an optional exit extension.
		</synopsis>
		<syntax>
			<parameter name="announce" required="true">
				<para>Filename of sound that will be played when no channel can be reached</para>
			</parameter>
			<parameter name="sleep" required="true">
				<para>Number of seconds to wait after a dial attempt failed before a new attempt is made</para>
			</parameter>
			<parameter name="retries" required="true">
				<para>Number of retries</para>
				<para>When this is reached flow will continue at the next priority in the dialplan</para>
			</parameter>
			<parameter name="dialargs" required="true">
				<para>Same format as arguments provided to the Dial application</para>
			</parameter>
		</syntax>
		<description>
			<para>This application will attempt to place a call using the normal Dial application.
			If no channel can be reached, the <replaceable>announce</replaceable> file will be played.
			Then, it will wait <replaceable>sleep</replaceable> number of seconds before retrying the call.
			After <replaceable>retries</replaceable> number of attempts, the calling channel will continue at the next priority in the dialplan.
			If the <replaceable>retries</replaceable> setting is set to 0, this application will retry endlessly.
			While waiting to retry a call, a 1 digit extension may be dialed. If that
			extension exists in either the context defined in <variable>EXITCONTEXT</variable> or the current
			one, The call will jump to that extension immediately.
			The <replaceable>dialargs</replaceable> are specified in the same format that arguments are provided
			to the Dial application.</para>
		</description>
		<see-also>
			<ref type="application">Dial</ref>
		</see-also>
	</application>
 ***/

static const char app[] = "Dial";
static const char rapp[] = "RetryDial";

enum {
	OPT_ANNOUNCE =          (1 << 0),
	OPT_RESETCDR =          (1 << 1),
	OPT_DTMF_EXIT =         (1 << 2),
	OPT_SENDDTMF =          (1 << 3),
	OPT_FORCECLID =         (1 << 4),
	OPT_GO_ON =             (1 << 5),
	OPT_CALLEE_HANGUP =     (1 << 6),
	OPT_CALLER_HANGUP =     (1 << 7),
	OPT_ORIGINAL_CLID =     (1 << 8),
	OPT_DURATION_LIMIT =    (1 << 9),
	OPT_MUSICBACK =         (1 << 10),
	OPT_CALLEE_MACRO =      (1 << 11),
	OPT_SCREEN_NOINTRO =    (1 << 12),
	OPT_SCREEN_NOCALLERID = (1 << 13),
	OPT_IGNORE_CONNECTEDLINE = (1 << 14),
	OPT_SCREENING =         (1 << 15),
	OPT_PRIVACY =           (1 << 16),
	OPT_RINGBACK =          (1 << 17),
	OPT_DURATION_STOP =     (1 << 18),
	OPT_CALLEE_TRANSFER =   (1 << 19),
	OPT_CALLER_TRANSFER =   (1 << 20),
	OPT_CALLEE_MONITOR =    (1 << 21),
	OPT_CALLER_MONITOR =    (1 << 22),
	OPT_GOTO =              (1 << 23),
	OPT_OPERMODE =          (1 << 24),
	OPT_CALLEE_PARK =       (1 << 25),
	OPT_CALLER_PARK =       (1 << 26),
	OPT_IGNORE_FORWARDING = (1 << 27),
	OPT_CALLEE_GOSUB =      (1 << 28),
	OPT_CALLEE_MIXMONITOR = (1 << 29),
	OPT_CALLER_MIXMONITOR = (1 << 30),
};

/* flags are now 64 bits, so keep it up! */
#define DIAL_STILLGOING      (1LLU << 31)
#define DIAL_NOFORWARDHTML   (1LLU << 32)
#define DIAL_CALLERID_ABSENT (1LLU << 33) /* TRUE if caller id is not available for connected line. */
#define OPT_CANCEL_ELSEWHERE (1LLU << 34)
#define OPT_PEER_H           (1LLU << 35)
#define OPT_CALLEE_GO_ON     (1LLU << 36)
#define OPT_CANCEL_TIMEOUT   (1LLU << 37)
#define OPT_FORCE_CID_TAG    (1LLU << 38)
#define OPT_FORCE_CID_PRES   (1LLU << 39)
#define OPT_CALLER_ANSWER    (1LLU << 40)
#define OPT_PREDIAL_CALLEE   (1LLU << 41)
#define OPT_PREDIAL_CALLER   (1LLU << 42)
#define OPT_RING_WITH_EARLY_MEDIA (1LLU << 43)
#define OPT_HANGUPCAUSE      (1LLU << 44)
#ifdef GRANDSTREAM_NETWORKS
#define DIAL_NOANSWER_TIMEOUT_ATTENDED_TRANSFER	15020	/*!< ms, added by yfyang */
#endif

enum {
	OPT_ARG_ANNOUNCE = 0,
	OPT_ARG_SENDDTMF,
	OPT_ARG_GOTO,
	OPT_ARG_DURATION_LIMIT,
	OPT_ARG_MUSICBACK,
	OPT_ARG_CALLEE_MACRO,
	OPT_ARG_RINGBACK,
	OPT_ARG_CALLEE_GOSUB,
	OPT_ARG_CALLEE_GO_ON,
	OPT_ARG_PRIVACY,
	OPT_ARG_DURATION_STOP,
	OPT_ARG_OPERMODE,
	OPT_ARG_SCREEN_NOINTRO,
	OPT_ARG_ORIGINAL_CLID,
	OPT_ARG_FORCECLID,
	OPT_ARG_FORCE_CID_TAG,
	OPT_ARG_FORCE_CID_PRES,
	OPT_ARG_PREDIAL_CALLEE,
	OPT_ARG_PREDIAL_CALLER,
	OPT_ARG_HANGUPCAUSE,
	/* note: this entry _MUST_ be the last one in the enum */
	OPT_ARG_ARRAY_SIZE
};

AST_APP_OPTIONS(dial_exec_options, BEGIN_OPTIONS
	AST_APP_OPTION_ARG('A', OPT_ANNOUNCE, OPT_ARG_ANNOUNCE),
	AST_APP_OPTION('a', OPT_CALLER_ANSWER),
	AST_APP_OPTION_ARG('b', OPT_PREDIAL_CALLEE, OPT_ARG_PREDIAL_CALLEE),
	AST_APP_OPTION_ARG('B', OPT_PREDIAL_CALLER, OPT_ARG_PREDIAL_CALLER),
	AST_APP_OPTION('C', OPT_RESETCDR),
	AST_APP_OPTION('c', OPT_CANCEL_ELSEWHERE),
	AST_APP_OPTION('d', OPT_DTMF_EXIT),
	AST_APP_OPTION_ARG('D', OPT_SENDDTMF, OPT_ARG_SENDDTMF),
	AST_APP_OPTION('e', OPT_PEER_H),
	AST_APP_OPTION_ARG('f', OPT_FORCECLID, OPT_ARG_FORCECLID),
	AST_APP_OPTION_ARG('F', OPT_CALLEE_GO_ON, OPT_ARG_CALLEE_GO_ON),
	AST_APP_OPTION('g', OPT_GO_ON),
	AST_APP_OPTION_ARG('G', OPT_GOTO, OPT_ARG_GOTO),
	AST_APP_OPTION('h', OPT_CALLEE_HANGUP),
	AST_APP_OPTION('H', OPT_CALLER_HANGUP),
	AST_APP_OPTION('i', OPT_IGNORE_FORWARDING),
	AST_APP_OPTION('I', OPT_IGNORE_CONNECTEDLINE),
	AST_APP_OPTION('k', OPT_CALLEE_PARK),
	AST_APP_OPTION('K', OPT_CALLER_PARK),
	AST_APP_OPTION_ARG('L', OPT_DURATION_LIMIT, OPT_ARG_DURATION_LIMIT),
	AST_APP_OPTION_ARG('m', OPT_MUSICBACK, OPT_ARG_MUSICBACK),
	AST_APP_OPTION_ARG('M', OPT_CALLEE_MACRO, OPT_ARG_CALLEE_MACRO),
	AST_APP_OPTION_ARG('n', OPT_SCREEN_NOINTRO, OPT_ARG_SCREEN_NOINTRO),
	AST_APP_OPTION('N', OPT_SCREEN_NOCALLERID),
	AST_APP_OPTION_ARG('o', OPT_ORIGINAL_CLID, OPT_ARG_ORIGINAL_CLID),
	AST_APP_OPTION_ARG('O', OPT_OPERMODE, OPT_ARG_OPERMODE),
	AST_APP_OPTION('p', OPT_SCREENING),
	AST_APP_OPTION_ARG('P', OPT_PRIVACY, OPT_ARG_PRIVACY),
	AST_APP_OPTION_ARG('Q', OPT_HANGUPCAUSE, OPT_ARG_HANGUPCAUSE),
	AST_APP_OPTION_ARG('r', OPT_RINGBACK, OPT_ARG_RINGBACK),
	AST_APP_OPTION('R', OPT_RING_WITH_EARLY_MEDIA),
	AST_APP_OPTION_ARG('S', OPT_DURATION_STOP, OPT_ARG_DURATION_STOP),
	AST_APP_OPTION_ARG('s', OPT_FORCE_CID_TAG, OPT_ARG_FORCE_CID_TAG),
	AST_APP_OPTION('t', OPT_CALLEE_TRANSFER),
	AST_APP_OPTION('T', OPT_CALLER_TRANSFER),
	AST_APP_OPTION_ARG('u', OPT_FORCE_CID_PRES, OPT_ARG_FORCE_CID_PRES),
	AST_APP_OPTION_ARG('U', OPT_CALLEE_GOSUB, OPT_ARG_CALLEE_GOSUB),
	AST_APP_OPTION('w', OPT_CALLEE_MONITOR),
	AST_APP_OPTION('W', OPT_CALLER_MONITOR),
	AST_APP_OPTION('x', OPT_CALLEE_MIXMONITOR),
	AST_APP_OPTION('X', OPT_CALLER_MIXMONITOR),
	AST_APP_OPTION('z', OPT_CANCEL_TIMEOUT),
END_OPTIONS );

#define CAN_EARLY_BRIDGE(flags,chan,peer) (!ast_test_flag64(flags, OPT_CALLEE_HANGUP | \
	OPT_CALLER_HANGUP | OPT_CALLEE_TRANSFER | OPT_CALLER_TRANSFER | \
	OPT_CALLEE_MONITOR | OPT_CALLER_MONITOR | OPT_CALLEE_PARK |  \
	OPT_CALLER_PARK | OPT_ANNOUNCE | OPT_CALLEE_MACRO | OPT_CALLEE_GOSUB) && \
	!ast_channel_audiohooks(chan) && !ast_channel_audiohooks(peer) && \
	ast_framehook_list_is_empty(ast_channel_framehooks(chan)) && ast_framehook_list_is_empty(ast_channel_framehooks(peer)))

/*
 * The list of active channels
 */
struct chanlist {
	AST_LIST_ENTRY(chanlist) node;
	struct ast_channel *chan;
	/*! Channel interface dialing string (is tech/number).  (Stored in stuff[]) */
	const char *interface;
	/*! Channel technology name.  (Stored in stuff[]) */
	const char *tech;
	/*! Channel device addressing.  (Stored in stuff[]) */
	const char *number;
	/*! Original channel name.  Must be freed.  Could be NULL if allocation failed. */
	char *orig_chan_name;
	uint64_t flags;
	/*! Saved connected party info from an AST_CONTROL_CONNECTED_LINE. */
	struct ast_party_connected_line connected;
	/*! TRUE if an AST_CONTROL_CONNECTED_LINE update was saved to the connected element. */
	unsigned int pending_connected_update:1;
	struct ast_aoc_decoded *aoc_s_rate_list;
	/*! The interface, tech, and number strings are stuffed here. */
	char stuff[0];
};

AST_LIST_HEAD_NOLOCK(dial_head, chanlist);

static int detect_disconnect(struct ast_channel *chan, char code, struct ast_str **featurecode);

static void chanlist_free(struct chanlist *outgoing)
{
	ast_party_connected_line_free(&outgoing->connected);
	ast_aoc_destroy_decoded(outgoing->aoc_s_rate_list);
	ast_free(outgoing->orig_chan_name);
	ast_free(outgoing);
}

static void hanguptree(struct dial_head *out_chans, struct ast_channel *exception, int hangupcause)
{
	/* Hang up a tree of stuff */
	struct chanlist *outgoing;

	while ((outgoing = AST_LIST_REMOVE_HEAD(out_chans, node))) {
		/* Hangup any existing lines we have open */
		if (outgoing->chan && (outgoing->chan != exception)) {
			if (hangupcause >= 0) {
				/* This is for the channel drivers */
				ast_channel_hangupcause_set(outgoing->chan, hangupcause);
			}
			ast_hangup(outgoing->chan);
		}
		chanlist_free(outgoing);
	}
}

#define AST_MAX_WATCHERS 256

/*
 * argument to handle_cause() and other functions.
 */
struct cause_args {
	struct ast_channel *chan;
	int busy;
	int congestion;
	int nochan;
};

static void handle_cause(int cause, struct cause_args *num)
{
	switch(cause) {
	case AST_CAUSE_BUSY:
		num->busy++;
		break;
	case AST_CAUSE_CONGESTION:
		num->congestion++;
		break;
	case AST_CAUSE_NO_ROUTE_DESTINATION:
	case AST_CAUSE_UNREGISTERED:
		num->nochan++;
		break;
	case AST_CAUSE_NO_ANSWER:
#ifdef GRANDSTREAM_NETWORKS
	case AST_CAUSE_NO_USER_RESPONSE:
#endif
	case AST_CAUSE_NORMAL_CLEARING:
		break;
	default:
		num->nochan++;
		break;
	}
}

static int onedigit_goto(struct ast_channel *chan, const char *context, char exten, int pri)
{
	char rexten[2] = { exten, '\0' };

	if (context) {
		if (!ast_goto_if_exists(chan, context, rexten, pri))
			return 1;
	} else {
		if (!ast_goto_if_exists(chan, ast_channel_context(chan), rexten, pri))
			return 1;
		else if (!ast_strlen_zero(ast_channel_macrocontext(chan))) {
			if (!ast_goto_if_exists(chan, ast_channel_macrocontext(chan), rexten, pri))
				return 1;
		}
	}
	return 0;
}

/* do not call with chan lock held */
static const char *get_cid_name(char *name, int namelen, struct ast_channel *chan)
{
	const char *context;
	const char *exten;

	ast_channel_lock(chan);
	context = ast_strdupa(S_OR(ast_channel_macrocontext(chan), ast_channel_context(chan)));
	exten = ast_strdupa(S_OR(ast_channel_macroexten(chan), ast_channel_exten(chan)));
	ast_channel_unlock(chan);

	return ast_get_hint(NULL, 0, name, namelen, chan, context, exten) ? name : "";
}

/*!
 * helper function for wait_for_answer()
 *
 * \param o Outgoing call channel list.
 * \param num Incoming call channel cause accumulation
 * \param peerflags Dial option flags
 * \param single TRUE if there is only one outgoing call.
 * \param caller_entertained TRUE if the caller is being entertained by MOH or ringback.
 * \param to Remaining call timeout time.
 * \param forced_clid OPT_FORCECLID caller id to send
 * \param stored_clid Caller id representing the called party if needed
 *
 * XXX this code is highly suspicious, as it essentially overwrites
 * the outgoing channel without properly deleting it.
 *
 * \todo eventually this function should be intergrated into and replaced by ast_call_forward()
 */
static void do_forward(struct chanlist *o, struct cause_args *num,
	struct ast_flags64 *peerflags, int single, int caller_entertained, int *to,
	struct ast_party_id *forced_clid, struct ast_party_id *stored_clid)
{
	char tmpchan[256];
	char forwarder[AST_CHANNEL_NAME];
	struct ast_channel *original = o->chan;
	struct ast_channel *c = o->chan; /* the winner */
	struct ast_channel *in = num->chan; /* the input channel */
	char *stuff;
	char *tech;
	int cause;
	struct ast_party_caller caller;

	ast_copy_string(forwarder, ast_channel_name(c), sizeof(forwarder));
	ast_copy_string(tmpchan, ast_channel_call_forward(c), sizeof(tmpchan));
	if ((stuff = strchr(tmpchan, '/'))) {
		*stuff++ = '\0';
		tech = tmpchan;
	} else {
		const char *forward_context;
		ast_channel_lock(c);
		forward_context = pbx_builtin_getvar_helper(c, "FORWARD_CONTEXT");
		if (ast_strlen_zero(forward_context)) {
			forward_context = NULL;
		}
		snprintf(tmpchan, sizeof(tmpchan), "%s@%s", ast_channel_call_forward(c), forward_context ? forward_context : ast_channel_context(c));
		ast_channel_unlock(c);
		stuff = tmpchan;
		tech = "Local";
	}
	if (!strcasecmp(tech, "Local")) {
		/*
		 * Drop the connected line update block for local channels since
		 * this is going to run dialplan and the user can change his
		 * mind about what connected line information he wants to send.
		 */
		ast_clear_flag64(o, OPT_IGNORE_CONNECTEDLINE);
	}

	/* Before processing channel, go ahead and check for forwarding */
	ast_verb(3, "Now forwarding %s to '%s/%s' (thanks to %s)\n", ast_channel_name(in), tech, stuff, ast_channel_name(c));
	/* If we have been told to ignore forwards, just set this channel to null and continue processing extensions normally */
	if (ast_test_flag64(peerflags, OPT_IGNORE_FORWARDING)) {
		ast_verb(3, "Forwarding %s to '%s/%s' prevented.\n", ast_channel_name(in), tech, stuff);
		ast_channel_publish_dial_forward(in, original, NULL, NULL, "CANCEL",
			ast_channel_call_forward(original));
		c = o->chan = NULL;
		cause = AST_CAUSE_BUSY;
	} else {
		struct ast_stream_topology *topology;

#ifdef GRANDSTREAM_NETWORKS
		/*
		 *  added by zgmo for NBF merge from 20.x
		 *  the original NBF info:
		 *  [NBF] modified for the permission for phone do forward
		 *  by jklou
		 */
		struct ast_channel* from_chan = c;
#endif
		ast_channel_lock(in);
		topology = ast_stream_topology_clone(ast_channel_get_stream_topology(in));
		ast_channel_unlock(in);

		/* Setup parameters */
		c = o->chan = ast_request_with_stream_topology(tech, topology, NULL, in, stuff, &cause);
#ifdef GRANDSTREAM_NETWORKS
		const char *in_chan_name = ast_channel_name(in);
		if (NULL != strcasestr(in_chan_name, "from-internal-xfer") && NULL != strcasestr(in_chan_name, "Local") && NULL != strstr(in_chan_name, ";2")){
			pbx_builtin_setvar_helper(o->chan, "IS_ATX", "yes");
		}
#endif

		ast_stream_topology_free(topology);

		if (c) {
			if (single && !caller_entertained) {
				ast_channel_make_compatible(in, o->chan);
			}
			ast_channel_lock_both(in, o->chan);
			ast_channel_inherit_variables(in, o->chan);
			ast_channel_datastore_inherit(in, o->chan);
			pbx_builtin_setvar_helper(o->chan, "FORWARDERNAME", forwarder);
			ast_max_forwards_decrement(o->chan);
			ast_channel_unlock(in);
			ast_channel_unlock(o->chan);
			/* When a call is forwarded, we don't want to track new interfaces
			 * dialed for CC purposes. Setting the done flag will ensure that
			 * any Dial operations that happen later won't record CC interfaces.
			 */
			ast_ignore_cc(o->chan);
			ast_log(LOG_NOTICE, "Not accepting call completion offers from call-forward recipient %s\n", ast_channel_name(o->chan));
#ifdef GRANDSTREAM_NETWORKS
			pbx_builtin_setvar_helper(in, "custom", ast_channel_name(o->chan));  /* Added by yfyang, for Bug 222972 */

			/* Added by yfyang, for Bug185340 */
			const char *en_reason = pbx_builtin_getvar_helper(c, "CF_REASON");
			if (!ast_strlen_zero(en_reason)) {
				if (atoi(en_reason) == AST_REDIRECTING_REASON_BLIND_TRANSFER || atoi(en_reason) == AST_REDIRECTING_REASON_DEFLECTION) {
					ast_log(LOG_NOTICE, "Clear the cf variables for chan[%s], reason[%s]\n", ast_channel_name(c), en_reason);
					pbx_builtin_setvar_helper(c, "CF_REASON", NULL);
					pbx_builtin_setvar_helper(c, "CF_ALL", NULL);
				}
			}
			/*
			 *  added by zgmo for NBF merge from 20.x
			 *  the original NBF info:
			 *  [NBF] modified for the permission for phone do forward
			 *  by jklou
			 */
			ast_log(LOG_NOTICE, "from_chann %s\n", ast_channel_name(from_chan));
			pbx_builtin_setvar_helper(o->chan, "DO_FORWARD", ast_channel_name(from_chan));
			pbx_builtin_setvar_helper(in, "AF_FORWORD", ast_channel_name(o->chan));
			pbx_builtin_setvar_helper(o->chan, "__FORWARD_THX", ast_channel_name(from_chan));

			pbx_builtin_setvar_helper(c, "HAVE_PROMPTED", pbx_builtin_getvar_helper(in, "HAVE_PROMPTED")); /* added by zgmo for Bug144632 */
			pbx_builtin_setvar_helper(c, "REALCALLERIDNUM", pbx_builtin_getvar_helper(in, "REALCALLERIDNUM")); /* Bug190648 */
#endif
		} else
			ast_log(LOG_NOTICE,
				"Forwarding failed to create channel to dial '%s/%s' (cause = %d)\n",
				tech, stuff, cause);
	}
	if (!c) {
		ast_channel_publish_dial(in, original, stuff, "BUSY");
		ast_clear_flag64(o, DIAL_STILLGOING);
		handle_cause(cause, num);
		ast_hangup(original);
	} else {
		ast_channel_lock_both(c, original);
		ast_party_redirecting_copy(ast_channel_redirecting(c),
			ast_channel_redirecting(original));
		ast_channel_unlock(c);
		ast_channel_unlock(original);

		ast_channel_lock_both(c, in);

		if (single && !caller_entertained && CAN_EARLY_BRIDGE(peerflags, c, in)) {
			ast_rtp_instance_early_bridge_make_compatible(c, in);
		}

		if (!ast_channel_redirecting(c)->from.number.valid
			|| ast_strlen_zero(ast_channel_redirecting(c)->from.number.str)) {
			/*
			 * The call was not previously redirected so it is
			 * now redirected from this number.
			 */
			ast_party_number_free(&ast_channel_redirecting(c)->from.number);
			ast_party_number_init(&ast_channel_redirecting(c)->from.number);
			ast_channel_redirecting(c)->from.number.valid = 1;
			ast_channel_redirecting(c)->from.number.str =
				ast_strdup(S_OR(ast_channel_macroexten(in), ast_channel_exten(in)));
		}

		ast_channel_dialed(c)->transit_network_select = ast_channel_dialed(in)->transit_network_select;

		/* Determine CallerID to store in outgoing channel. */
		ast_party_caller_set_init(&caller, ast_channel_caller(c));
		if (ast_test_flag64(peerflags, OPT_ORIGINAL_CLID)) {
			caller.id = *stored_clid;
			ast_channel_set_caller_event(c, &caller, NULL);
			ast_set_flag64(o, DIAL_CALLERID_ABSENT);
		} else if (ast_strlen_zero(S_COR(ast_channel_caller(c)->id.number.valid,
			ast_channel_caller(c)->id.number.str, NULL))) {
			/*
			 * The new channel has no preset CallerID number by the channel
			 * driver.  Use the dialplan extension and hint name.
			 */
			caller.id = *stored_clid;
			ast_channel_set_caller_event(c, &caller, NULL);
			ast_set_flag64(o, DIAL_CALLERID_ABSENT);
		} else {
			ast_clear_flag64(o, DIAL_CALLERID_ABSENT);
		}

		/* Determine CallerID for outgoing channel to send. */
		if (ast_test_flag64(o, OPT_FORCECLID)) {
			struct ast_party_connected_line connected;

			ast_party_connected_line_init(&connected);
			connected.id = *forced_clid;
			ast_party_connected_line_copy(ast_channel_connected(c), &connected);
		} else {
			ast_connected_line_copy_from_caller(ast_channel_connected(c), ast_channel_caller(in));
		}

		ast_channel_req_accountcodes(c, in, AST_CHANNEL_REQUESTOR_BRIDGE_PEER);

		ast_channel_appl_set(c, "AppDial");
		ast_channel_data_set(c, "(Outgoing Line)");
		ast_channel_publish_snapshot(c);

		ast_channel_unlock(in);
		if (single && !ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
			struct ast_party_redirecting redirecting;

			/*
			 * Redirecting updates to the caller make sense only on single
			 * calls.
			 *
			 * We must unlock c before calling
			 * ast_channel_redirecting_macro, because we put c into
			 * autoservice there.  That is pretty much a guaranteed
			 * deadlock.  This is why the handling of c's lock may seem a
			 * bit unusual here.
			 */
			ast_party_redirecting_init(&redirecting);
			ast_party_redirecting_copy(&redirecting, ast_channel_redirecting(c));
			ast_channel_unlock(c);
			if (ast_channel_redirecting_sub(c, in, &redirecting, 0) &&
				ast_channel_redirecting_macro(c, in, &redirecting, 1, 0)) {
				ast_channel_update_redirecting(in, &redirecting, NULL);
			}
			ast_party_redirecting_free(&redirecting);
		} else {
			ast_channel_unlock(c);
		}

		if (ast_test_flag64(peerflags, OPT_CANCEL_TIMEOUT)) {
			*to = -1;
		}

		if (ast_call(c, stuff, 0)) {
			ast_log(LOG_NOTICE, "Forwarding failed to dial '%s/%s'\n",
				tech, stuff);
			ast_channel_publish_dial(in, original, stuff, "CONGESTION");
			ast_clear_flag64(o, DIAL_STILLGOING);
			ast_hangup(original);
			ast_hangup(c);
			c = o->chan = NULL;
			num->nochan++;
		} else {
#ifdef GRANDSTREAM_NETWORKS
			if (strcasestr(ast_channel_name(in), "Local/REC") && strcasestr(ast_channel_name(in), ";2")){
				char tmp_callee_name[64] = {0};
				char family[64] = {0};
				char key[64] = {0};
				char tmp_language[32] = {0};
				char local_chan1[128] = {0};
				sscanf(ast_channel_name(c), "%*[^/]/%63[^@]", tmp_callee_name);
				snprintf(family, sizeof(family)-1, "AAUSER/%s", tmp_callee_name);
				snprintf(key, sizeof(key)-1, "language");
				ast_ucm_db_get(family, key, tmp_language, sizeof(tmp_language) - 1);
				snprintf(local_chan1, sizeof(local_chan1)-1, "%s", ast_channel_name(in));
				local_chan1[strlen(local_chan1)-1] = '1';
				ast_log(LOG_DEBUG, "local chan:%s", local_chan1);
				struct ast_channel *Local_chan = NULL;
				Local_chan = ast_channel_get_by_name(local_chan1);
				if (Local_chan){
					ast_channel_lock(Local_chan);
					if (0 != strlen(tmp_language)){
						ast_channel_language_set(Local_chan, tmp_language);
						pbx_builtin_setvar_helper(Local_chan, "REC_CALLFORWARD", "YES");
					}
					else{
						pbx_builtin_setvar_helper(Local_chan, "REC_CALLFORWARD", "NO");
					}
					ast_channel_unlock(Local_chan);
					ast_channel_unref(Local_chan);
				}
			}
#endif
			ast_channel_publish_dial_forward(in, original, c, NULL, "CANCEL",
				ast_channel_call_forward(original));

			ast_channel_publish_dial(in, c, stuff, NULL);
#ifdef GRANDSTREAM_NETWORKS
			if (NULL != strcasestr(stuff, "call-forward") && NULL != strcasestr(tech, "Local")) {
				ast_cdr_setvar(ast_channel_name(in),"record_files","");

				/*Added by zgmo for BUG144632 */
				ast_cdr_setvar(ast_channel_name(in), "action_type", "CALLFORWARD");
			}
#endif

			/* Hangup the original channel now, in case we needed it */
			ast_hangup(original);
		}
		if (single && !caller_entertained) {
			ast_indicate(in, -1);
		}
	}
}

/* argument used for some functions. */
struct privacy_args {
	int sentringing;
	int privdb_val;
	char privcid[256];
	char privintro[1024];
	char status[256];
};

static void publish_dial_end_event(struct ast_channel *in, struct dial_head *out_chans, struct ast_channel *exception, const char *status)
{
	struct chanlist *outgoing;
	AST_LIST_TRAVERSE(out_chans, outgoing, node) {
		if (!outgoing->chan || outgoing->chan == exception) {
			continue;
		}
		ast_channel_publish_dial(in, outgoing->chan, NULL, status);
	}
}


/*End added by zyxia for push ansered else for wave app,2022.7.5*/

/*!
 * \internal
 * \brief Update connected line on chan from peer.
 * \since 13.6.0
 *
 * \param chan Channel to get connected line updated.
 * \param peer Channel providing connected line information.
 * \param is_caller Non-zero if chan is the calling channel.
 *
 * \return Nothing
 */
static void update_connected_line_from_peer(struct ast_channel *chan, struct ast_channel *peer, int is_caller)
{
	struct ast_party_connected_line connected_caller;

	ast_party_connected_line_init(&connected_caller);

	ast_channel_lock(peer);
	ast_connected_line_copy_from_caller(&connected_caller, ast_channel_caller(peer));
	ast_channel_unlock(peer);
	connected_caller.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;
	if (ast_channel_connected_line_sub(peer, chan, &connected_caller, 0)
		&& ast_channel_connected_line_macro(peer, chan, &connected_caller, is_caller, 0)) {
		ast_channel_update_connected_line(chan, &connected_caller, NULL);
	}
	ast_party_connected_line_free(&connected_caller);
}

#ifdef GRANDSTREAM_NETWORKS
struct ast_app;

static int sfu_handle_early_media(struct ast_channel *c, int join)
{
	struct ast_app *app = NULL;
	char appdata[64] = {0};
	const char *conf_num = pbx_builtin_getvar_helper(c, "CONF_ROOM");

	snprintf(appdata, sizeof(appdata), "%s,%s", conf_num, join ? "join" : "leave");
	app = pbx_findapp("ConfBridgeEarlyMedia");
	if (app) {
		ast_log(LOG_VERBOSE, "Launching %s(%s) on %s\n", "ConfBridgeEarlyMedia", S_OR(appdata, ""),
			ast_channel_name(c));
		pbx_exec(c, app, S_OR(appdata, ""));
	}

	return 0;
}

/*
 * 2021-11/auto-1638183467-4001-6400  ===>  2021-11/auto-1638183467RG4005-4001-6400
 *                                                              (callee_num)
 *         from                                          to
 */
static int pending_ringgroup_callee_info(char *to, const char *from, char *callee_num)
{
	int len = 0;
	char *insert_point = NULL;

	ast_log(LOG_DEBUG, "Enter pending_ringgroup_callee_info()");
	if (!to || ast_strlen_zero(from)
		|| ast_strlen_zero(callee_num)) {
		ast_log(LOG_ERROR, "Invalid Param, to:%p, from:%s, callee_num:%s", to,
			ast_strlen_zero(from)? from:"invalid",
			ast_strlen_zero(callee_num)? callee_num:"invalid");
		return -1;
	}

	insert_point = strchr(from, '-');
	if (!insert_point) {
		ast_log(LOG_ERROR, "Invalid from string:%s", from);
		return -1;
	}

	insert_point = strchr(insert_point + 1, '-');
	if (!insert_point) {
		ast_log(LOG_ERROR, "Invalid from string:%s", from);
		return -1;
	}

	insert_point = strchr(insert_point + 1, '-');
	if (!insert_point) {
		ast_log(LOG_ERROR, "Invalid from string:%s", from);
		return -1;
	}

	len = insert_point - from;
	strncpy(to, from, len);
	to[len] = '\0';

	sprintf(to + len, "RG%s", callee_num);
	strcat(to, insert_point);

	ast_log(LOG_DEBUG, "Exit pending_ringgroup_callee_info() from:%s  ====> to:%s", from, to);

	return 0;
}

static int process_outside_stop_record_for_dial(struct ast_channel *chan,
																	struct ast_bridge_config *config)
{
	int need_free = 0;
	int need_mark_prompt = 0;
	const char *from_disa = pbx_builtin_getvar_helper(chan, "FROM_DISA");
	struct ast_channel *chan_check = chan;

	if (!ast_strlen_zero(from_disa) && !strcmp(from_disa, "yes")
		&& strstr(ast_channel_name(chan), "Local/") && strstr(ast_channel_name(chan), ";2")) {
		const char *fst_caller_chan_name = pbx_builtin_getvar_helper(chan,  "FIRST_CALLER_CHANNEL_NAME");
		if (!ast_strlen_zero(fst_caller_chan_name)
			&& (strstr(fst_caller_chan_name, "PJSIP/trunk_")
				|| strstr(fst_caller_chan_name, "DAHDI/o"))) {
			chan_check = ast_channel_get_by_name(fst_caller_chan_name);
			if (chan_check) {
				const char *fm_enable = pbx_builtin_getvar_helper(chan, "FM_ENABLE");
				if (ast_strlen_zero(fm_enable) || strcmp(fm_enable, "yes")) {
					pbx_builtin_setvar_helper(chan_check, "HAVE_PROMPTED",
							pbx_builtin_getvar_helper(chan, "HAVE_PROMPTED"));
					pbx_builtin_setvar_helper(chan_check, "OUTSIDE_EXT_STOP_RECORD_PROMPTED",
							pbx_builtin_getvar_helper(chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED"));
				} else {
					need_mark_prompt = 1;
				}

				pbx_builtin_setvar_helper(chan, "LOCAL2_IS_TRUNK", "yes");
				need_free = 1;
			}
		}
	}

	if (ast_dial_check_outside_can_stop_record(chan_check)) {
		const char *have_prompted = pbx_builtin_getvar_helper(chan_check, "HAVE_PROMPTED");
		if (ast_strlen_zero(have_prompted) || strcmp(have_prompted, "yes")) {
			ast_answer(chan_check);
		}

		ast_dial_prompt_outside_stop_record(chan_check);
		ast_set_flag(&(config->features_caller), AST_FEATURE_AUTOMIXMON);
		ast_log(LOG_WARNING, "Set caller %s automixmon flag", ast_channel_name(chan));

		if (need_mark_prompt) {
			pbx_builtin_setvar_helper(chan, "HAVE_PROMPTED",
							pbx_builtin_getvar_helper(chan_check, "HAVE_PROMPTED"));
			pbx_builtin_setvar_helper(chan, "OUTSIDE_EXT_STOP_RECORD_PROMPTED",
							pbx_builtin_getvar_helper(chan_check, "OUTSIDE_EXT_STOP_RECORD_PROMPTED"));
		}
	}

	if (need_free) {
		ast_channel_unref(chan_check);
	}

	return 0;
}

/* Modify by bfdu for Bug 296720, 2024.5.14 */
static void set_timeout_for_blind_transfer_callback(struct ast_channel *chan, int *to)
{
	unsigned int blind_transfer_timeout = 0;
	const char *blind_transfer_callback_num = NULL;
	const char *cfnext = NULL;
	const char *cfuext = NULL;
	const char *cfbext = NULL;
	const char *cfn_flag = NULL;
	const char *cfu_flag = NULL;
	const char *cfb_flag = NULL;

	blind_transfer_callback_num = pbx_builtin_getvar_helper(chan, "BLIND_TRANSFER_CALLBACK_NUM");
	if (ast_strlen_zero(blind_transfer_callback_num)) {
		ast_log(LOG_DEBUG, "Can't get callback number, don't need change, chan:%s\n", ast_channel_name((chan)));
		return;
	}

	cfnext = pbx_builtin_getvar_helper(chan, "CFNEXT");
	cfuext = pbx_builtin_getvar_helper(chan, "CFUEXT");
	cfbext = pbx_builtin_getvar_helper(chan, "CFBEXT");
	cfn_flag = pbx_builtin_getvar_helper(chan, "CFN_FLAG");
	cfu_flag = pbx_builtin_getvar_helper(chan, "CFU_FLAG");
	cfb_flag = pbx_builtin_getvar_helper(chan, "CFB_FLAG");
	if (ast_strlen_zero(cfnext) && ast_strlen_zero(cfuext) && ast_strlen_zero(cfbext) &&
		ast_strlen_zero(cfn_flag) && ast_strlen_zero(cfu_flag) && ast_strlen_zero(cfb_flag)) {
		blind_transfer_timeout = ast_channel_blind_transfer_timeout(chan);
		*to = blind_transfer_timeout*1000;
		ast_log(LOG_DEBUG, "Get blind_transfer_callback_num:%s, update timeout value:%d, chan:%s\n",
			blind_transfer_callback_num, blind_transfer_timeout, ast_channel_name((chan)));
	} else {
		ast_log(LOG_DEBUG, "Get blind_transfer_callback_num:%s, Chan:%s set call forward, cfn:%s, cfu:%s, cfb:%s, cfn_flag:%s, cfu_flag:%s, cfb_flag:%s, don't update timeout\n",
			blind_transfer_callback_num, ast_channel_name(chan), cfnext, cfuext, cfbext, cfn_flag, cfu_flag, cfb_flag);
	}

	return;
}

static void publish_answered_else_event2im(struct ast_channel *caller,
            struct dial_head *out_chans, struct ast_channel *answered_chan,
            const char* terminal_ext, const char* reason)
{
	struct chanlist *outgoing;
	const char* dst_num = NULL;
	const char* pre_dst_num = NULL;
	const char* caller_num = NULL;
	const char* caller_name = NULL;
	const char *var_ext = NULL;
	const char* instance = NULL;

	if (caller) {
		caller_num = S_COR(ast_channel_caller(caller)->id.number.valid,
				ast_channel_caller(caller)->id.number.str, "");

		caller_name =S_COR(ast_channel_caller(caller)->id.name.valid,
			ast_channel_caller(caller)->id.name.str, "");
	}

	if(answered_chan)
		instance = pbx_builtin_getvar_helper(answered_chan,"app_instance");//added by zyxia for Bug 308693

	AST_LIST_TRAVERSE(out_chans, outgoing, node) {

		/*if is not app channel and pjsip channel, is answed channel ,dont handle it*/
		if (!outgoing->chan  || outgoing->chan == answered_chan
			|| NULL == strstr(ast_channel_name(outgoing->chan), "PJSIP/")
			|| strstr(ast_channel_name(outgoing->chan), "PJSIP/trunk")) {
			continue;
		}

		if (!ast_strlen_zero(reason)) {
			if (!strcmp(reason,"refused")) {
				var_ext = pbx_builtin_getvar_helper(outgoing->chan, "CANCEL_OTHERS_IF_ONE_REJECT");
				if (!ast_strlen_zero(var_ext) && terminal_ext && !strncmp(terminal_ext, var_ext, sizeof(terminal_ext)-1)) {
					ast_log(LOG_DEBUG, "var_ext:%s\n",var_ext);
				} else {
					continue;
				}
			} else if (!strcmp(reason,"answer")) {
				ast_log(LOG_DEBUG, "Ringing Chan:%s, Call is answered by others",
				ast_channel_name(outgoing->chan));
				char chan_reason[256] = {0};
				snprintf(chan_reason, sizeof(chan_reason), "answer-elsewhere%s", ast_channel_name(outgoing->chan));
				ast_cdr_newest_setvar(ast_channel_name(caller), "reason", chan_reason);
			} else {
				continue;
			}
		}


		/*added by zyxia for Bug236334,2022.9.9*/
		if (!ast_channel_get_is_wave_app(outgoing->chan)) {
			ast_log(LOG_DEBUG, "Chan:%s is not wave app channel,dont need send AnsweredByElse",ast_channel_name(outgoing->chan));
			continue;
		}
		/*End by zyxia for Bug236334,2022.9.9*/

		dst_num = S_COR(ast_channel_caller(outgoing->chan)->id.number.valid,
					ast_channel_caller(outgoing->chan)->id.number.str, NULL);

		/*the fist dst numer need send ,also compare every channel number with previous channel,2022,7.4*/
		if ((pre_dst_num == NULL && dst_num) || (pre_dst_num && dst_num && strcmp(pre_dst_num,dst_num))){
			pre_dst_num = S_COR(ast_channel_caller(outgoing->chan)->id.number.valid,
						ast_channel_caller(outgoing->chan)->id.number.str, NULL);

			/* send event to cti pbxmid */
			manager_event(EVENT_FLAG_CALL, "AnsweredByElse",
											"Caller_Num: %s\r\n"
											"Caller_Name: %s\r\n"
											"Callee_Num: %s\r\n"
											"Reason: %s\r\n"
											"AnsweredInstance: %s\r\n",
							caller_num,caller_name,dst_num,reason,instance?instance:"");

			ast_log(LOG_DEBUG, "Send App AnsweredRefuesdByElse Event  Caller_Num:%s Callee_Num:%s Reason:%s Instance:%s\n",caller_num,dst_num,reason,instance ? instance:"");
		}
	}
	return ;
}

#endif

static struct ast_channel *wait_for_answer(struct ast_channel *in,
	struct dial_head *out_chans, int *to, struct ast_flags64 *peerflags,
	char *opt_args[],
	struct privacy_args *pa,
	const struct cause_args *num_in, int *result, char *dtmf_progress,
	const int ignore_cc,
	struct ast_party_id *forced_clid, struct ast_party_id *stored_clid)
{
	struct cause_args num = *num_in;
	int prestart = num.busy + num.congestion + num.nochan;
	int orig = *to;
	struct ast_channel *peer = NULL;
	struct chanlist *outgoing = AST_LIST_FIRST(out_chans);
	/* single is set if only one destination is enabled */
	int single = outgoing && !AST_LIST_NEXT(outgoing, node);
	int caller_entertained = outgoing
		&& ast_test_flag64(outgoing, OPT_MUSICBACK | OPT_RINGBACK);
	struct ast_str *featurecode = ast_str_alloca(AST_FEATURE_MAX_LEN + 1);
	int cc_recall_core_id;
	int is_cc_recall;
	int cc_frame_received = 0;
	int num_ringing = 0;
	struct timeval start = ast_tvnow();
#ifdef GRANDSTREAM_NETWORKS
	int progress_flag = 0;
	int remain_num = 0;
	int progress_num = 0;

	int cc_play_rbt = 0;
	int xfer_ringing = 0;
	struct timeval  time_ringing;
	int is_pri = 0;
	struct ast_channel *spyee_chan = NULL;
	if(NULL != strcasestr(ast_channel_name(in), "Local") && NULL != strstr(ast_channel_name(in), ";2")){
		const char* spy_exten = pbx_builtin_getvar_helper(in, "SPYEXTEN");
		ast_log(LOG_NOTICE, "spy_exten:%s", spy_exten);
		if (spy_exten && strlen(spy_exten)>0){
			if (!(spyee_chan = ast_channel_get_by_name(spy_exten))) {
				return NULL;
			}
			const char *inbound_index = pbx_builtin_getvar_helper(spyee_chan, "INBOUND_TRUNK_INDEX");
			if (inbound_index){
				ast_cdr_setvar(ast_channel_name(in),"inbound_trunk_index", inbound_index);
			}
			if (spyee_chan) {
				spyee_chan = ast_channel_unref(spyee_chan);
			}
		}
	}
#endif

	if (single) {
		/* Turn off hold music, etc */
		if (!caller_entertained) {
			ast_deactivate_generator(in);
			/* If we are calling a single channel, and not providing ringback or music, */
			/* then, make them compatible for in-band tone purpose */
			if (ast_channel_make_compatible(in, outgoing->chan) < 0) {
				/* If these channels can not be made compatible,
				 * there is no point in continuing.  The bridge
				 * will just fail if it gets that far.
				 */
				*to = -1;
				strcpy(pa->status, "CONGESTION");
				ast_channel_publish_dial(in, outgoing->chan, NULL, pa->status);
				return NULL;
			}
		}

		if (!ast_test_flag64(outgoing, OPT_IGNORE_CONNECTEDLINE)
			&& !ast_test_flag64(outgoing, DIAL_CALLERID_ABSENT)) {
			update_connected_line_from_peer(in, outgoing->chan, 1);
		}
	}

	is_cc_recall = ast_cc_is_recall(in, &cc_recall_core_id, NULL);

	while ((*to = ast_remaining_ms(start, orig)) && !peer) {
		struct chanlist *o;
		int pos = 0; /* how many channels do we handle */
		int numlines = prestart;
		struct ast_channel *winner;
		struct ast_channel *watchers[AST_MAX_WATCHERS];

		watchers[pos++] = in;
		AST_LIST_TRAVERSE(out_chans, o, node) {
			/* Keep track of important channels */
			if (ast_test_flag64(o, DIAL_STILLGOING) && o->chan)
				watchers[pos++] = o->chan;
			numlines++;
		}

#ifdef GRANDSTREAM_NETWORKS
		remain_num = numlines;

		if (xfer_ringing) {
			if (ast_tvdiff_ms(ast_tvnow(), time_ringing) >= DIAL_NOANSWER_TIMEOUT_ATTENDED_TRANSFER){
				ast_log(LOG_NOTICE, "wait_for_answer, xfer ringing timeout!\n");
				*to = 0;

				if (cc_play_rbt) {
					ast_moh_stop(in);
					cc_play_rbt = 0;
				}
				return NULL;
			}
		}
#endif
		if (pos == 1) { /* only the input channel is available */
			if (numlines == (num.busy + num.congestion + num.nochan)) {
				ast_verb(2, "Everyone is busy/congested at this time (%d:%d/%d/%d)\n", numlines, num.busy, num.congestion, num.nochan);
#ifdef GRANDSTREAM_NETWORKS
				if (1 == is_pri && num.nochan)
				{
					pbx_builtin_setvar_helper(num.chan, "PRI_NONUMBER", "YES");
				}
#endif
				if (num.busy)
					strcpy(pa->status, "BUSY");
				else if (num.congestion)
					strcpy(pa->status, "CONGESTION");
				else if (num.nochan)
					strcpy(pa->status, "CHANUNAVAIL");
			} else {
				ast_verb(3, "No one is available to answer at this time (%d:%d/%d/%d)\n", numlines, num.busy, num.congestion, num.nochan);
			}
			*to = 0;
			if (is_cc_recall) {
				ast_cc_failed(cc_recall_core_id, "Everyone is busy/congested for the recall. How sad");
			}

#ifdef GRANDSTREAM_NETWORKS
			if (cc_play_rbt) {
				ast_moh_stop(in);
				cc_play_rbt = 0;
			}
#endif
			return NULL;
		}
		winner = ast_waitfor_n(watchers, pos, to);
		AST_LIST_TRAVERSE(out_chans, o, node) {
			struct ast_frame *f;
			struct ast_channel *c = o->chan;

			if (c == NULL)
				continue;
			if (ast_test_flag64(o, DIAL_STILLGOING) && ast_channel_state(c) == AST_STATE_UP) {
				if (!peer) {
					ast_verb(3, "%s answered %s\n", ast_channel_name(c), ast_channel_name(in));
					if (o->orig_chan_name
						&& strcmp(o->orig_chan_name, ast_channel_name(c))) {
						/*
						 * The channel name changed so we must generate COLP update.
						 * Likely because a call pickup channel masqueraded in.
						 */
						update_connected_line_from_peer(in, c, 1);
					} else if (!single && !ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
						if (o->pending_connected_update) {
							if (ast_channel_connected_line_sub(c, in, &o->connected, 0) &&
								ast_channel_connected_line_macro(c, in, &o->connected, 1, 0)) {
								ast_channel_update_connected_line(in, &o->connected, NULL);
							}
						} else if (!ast_test_flag64(o, DIAL_CALLERID_ABSENT)) {
							update_connected_line_from_peer(in, c, 1);
						}
					}
					if (o->aoc_s_rate_list) {
						size_t encoded_size;
						struct ast_aoc_encoded *encoded;
						if ((encoded = ast_aoc_encode(o->aoc_s_rate_list, &encoded_size, o->chan))) {
							ast_indicate_data(in, AST_CONTROL_AOC, encoded, encoded_size);
							ast_aoc_destroy_encoded(encoded);
						}
					}
					peer = c;
					publish_dial_end_event(in, out_chans, peer, "CANCEL");
					ast_copy_flags64(peerflags, o,
						OPT_CALLEE_TRANSFER | OPT_CALLER_TRANSFER |
						OPT_CALLEE_HANGUP | OPT_CALLER_HANGUP |
						OPT_CALLEE_MONITOR | OPT_CALLER_MONITOR |
						OPT_CALLEE_PARK | OPT_CALLER_PARK |
						OPT_CALLEE_MIXMONITOR | OPT_CALLER_MIXMONITOR |
						DIAL_NOFORWARDHTML);
					ast_channel_dialcontext_set(c, "");
					ast_channel_exten_set(c, "");
				}
				continue;
			}
			if (c != winner)
				continue;
			/* here, o->chan == c == winner */
			if (!ast_strlen_zero(ast_channel_call_forward(c))) {
				pa->sentringing = 0;
				if (!ignore_cc && (f = ast_read(c))) {
					if (f->frametype == AST_FRAME_CONTROL && f->subclass.integer == AST_CONTROL_CC) {
						/* This channel is forwarding the call, and is capable of CC, so
						 * be sure to add the new device interface to the list
						 */
						ast_handle_cc_control_frame(in, c, f->data.ptr);
					}
					ast_frfree(f);
				}

				if (o->pending_connected_update) {
					/*
					 * Re-seed the chanlist's connected line information with
					 * previously acquired connected line info from the incoming
					 * channel.  The previously acquired connected line info could
					 * have been set through the CONNECTED_LINE dialplan function.
					 */
					o->pending_connected_update = 0;
					ast_channel_lock(in);
					ast_party_connected_line_copy(&o->connected, ast_channel_connected(in));
					ast_channel_unlock(in);
				}

				do_forward(o, &num, peerflags, single, caller_entertained, &orig,
					forced_clid, stored_clid);

				if (o->chan) {
					ast_free(o->orig_chan_name);
					o->orig_chan_name = ast_strdup(ast_channel_name(o->chan));
					if (single
						&& !ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)
						&& !ast_test_flag64(o, DIAL_CALLERID_ABSENT)) {
						update_connected_line_from_peer(in, o->chan, 1);
					}
				}
				continue;
			}
			f = ast_read(winner);
			if (!f) {
				ast_channel_hangupcause_set(in, ast_channel_hangupcause(c));
#ifdef GRANDSTREAM_NETWORKS
				ast_channel_hangupsipcode_set(in, ast_channel_hangupsipcode(c));
				if (ast_channel_sfu_early_media(c)) {
					sfu_handle_early_media(c, 0);
				}

				/*Added by pyzhong for Bug165993*/
				int is_exit = 0;
				int sipcode = ast_channel_hangupsipcode(c);
				if (!strcasecmp(ast_channel_tech(c)->type, "Local")) {
					const char *is_sfu = pbx_builtin_getvar_helper(c, "IS_SFU");
					if (!strcasecmp(S_OR(is_sfu, ""), "yes")) {
						const char *local_status = pbx_builtin_getvar_helper(c, "CHANLOCALSTATUS");
						if (!strcasecmp(S_OR(local_status, ""), "ANSWER")) {
							strcpy(pa->status, "ANSWER");
							is_exit = 1;
						}
					}
				}

				char terminal_ext[128] = {0};
				const char *cancel_ext = pbx_builtin_getvar_helper(c, "CANCEL_OTHERS_IF_ONE_REJECT");
				if (!ast_strlen_zero(cancel_ext)) {
					snprintf(terminal_ext, sizeof(terminal_ext)-1, "%s", cancel_ext);
					ast_log(LOG_NOTICE, "terminal_ext is [%s]", terminal_ext);
				}

				/* Added by bfdu for Bug 311695, 2024.8.16 */
				if (progress_num > 0 && progress_flag) {
					const char *indicate_progress = pbx_builtin_getvar_helper(c, "INDICATE_PROGRESS");
					if (!ast_strlen_zero(indicate_progress)) {
						progress_num--;
						if (progress_num == 0) {
							ast_log(LOG_NOTICE, "Has no indicate progress channel, indicate RINGING to [%s]", ast_channel_name(in));
							pbx_builtin_setvar_helper(in, "RING_BY_END", "yes");
							ast_indicate(in, AST_CONTROL_RINGING);
							progress_flag = 0;
						}
					}
				}
#endif
				ast_channel_publish_dial(in, c, NULL, ast_hangup_cause_to_dial_status(ast_channel_hangupcause(c)));
				ast_hangup(c);
				c = o->chan = NULL;
				ast_clear_flag64(o, DIAL_STILLGOING);
				handle_cause(ast_channel_hangupcause(in), &num);
#ifdef GRANDSTREAM_NETWORKS
				ast_log(LOG_NOTICE, "ast_channel_hangupcause %d(%s)",
					ast_channel_hangupcause(in), ast_channel_name(in));
				//if (AST_CAUSE_ANSWERED_ELSEWHERE == ast_channel_hangupcause(in)) {
				//	ast_cdr_newest_setvar(ast_channel_name(in), "reason", "answer-elsewhere");
				//}

				/* Added by yfyang, for Bug 199859 */
				remain_num--;
				if ((remain_num == 1) && progress_flag) {
					ast_log(LOG_NOTICE, "Only one callee remains, indicate RINGING to [%s]", ast_channel_name(in));
					ast_indicate(in, AST_CONTROL_RINGING);
					progress_flag = 0;
				}

				/* Added by yfyang, for the feature "cancel other terminals if one rejected" */
				if ((486 == sipcode) && (strlen(terminal_ext) > 0)) {
					struct chanlist *outgoing;
					const char *var_ext = NULL;
					int  leave_num = 0;
					publish_answered_else_event2im(in, out_chans, NULL, terminal_ext, "refused");

					AST_LIST_TRAVERSE_SAFE_BEGIN(out_chans, outgoing, node) {
						if (!outgoing->chan || outgoing->chan == c) {
							continue;
						}

						var_ext = pbx_builtin_getvar_helper(outgoing->chan, "CANCEL_OTHERS_IF_ONE_REJECT");
						if (!ast_strlen_zero(var_ext) && !strncmp(terminal_ext, var_ext, sizeof(terminal_ext)-1)) {
							ast_log(LOG_NOTICE, "cancel chan[%s]", ast_channel_name(outgoing->chan));
							ast_hangup(outgoing->chan);
							outgoing->chan = NULL;
							AST_LIST_REMOVE_CURRENT(node);
							chanlist_free(outgoing);
						} else {
							leave_num++;
						}
					}
					AST_LIST_TRAVERSE_SAFE_END;

					if (leave_num == 0) {
						ast_channel_hangupcause_set(in, AST_CAUSE_USER_BUSY);
						strcpy(pa->status, "BUSY");
						is_exit = 1;
						*to = 0;
					}
				}

				if (is_exit) {
					return NULL;
				}
#endif
				continue;
			}
			switch (f->frametype) {
			case AST_FRAME_CONTROL:
				switch (f->subclass.integer) {
				case AST_CONTROL_ANSWER:
					/* This is our guy if someone answered. */
					if (!peer) {
						ast_verb(3, "%s answered %s\n", ast_channel_name(c), ast_channel_name(in));

					/*Start added by zyxia for Bug 204898 call ipc device,200 ok need add*/
#ifdef GRANDSTREAM_NETWORKS
						/*for special sca*/
						if (0 == strncmp(ast_channel_name(in), "Local/", strlen("Local/")) && NULL != strstr(ast_channel_name(in), ";2")){
							char special_act[128] = {0};
							ast_cdr_getvar(ast_channel_name(in), "special_action_type", special_act, sizeof(special_act));
							if (0 == strcmp(special_act, "SPECIAL_SCA")){
								ast_cdr_setvar(ast_channel_name(in), "reason", "answer-elsewhere");
							}

							/* Added by yfyang, for Bug 264860 */
							if (pa->sentringing) {
								char *from_ivr = (char *)pbx_builtin_getvar_helper(in, "FROM_IVR");
								if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")) {
									ast_moh_stop(in);
									ast_log(LOG_NOTICE, "stop ringback moh, in[%s]", ast_channel_name(in));
								}
							}
						}
						if (!strncasecmp(ast_channel_name(c),"RTSP/",5)) {
							pbx_builtin_setvar_helper(in, "CALL_TO_IPC", "YES");
						}

						char  family[80] = {0};
						char  data[80] = {0};
						int high_freq_count = 0;
						int q_index, xtransfer = 0;
						struct timeval now;
						const char *aauser,*out_rt_pwd,*pinsetid;
						aauser = pbx_builtin_getvar_helper(in, "AAUSER");
						out_rt_pwd = pbx_builtin_getvar_helper(in, "outrt_passwd");
						pinsetid = pbx_builtin_getvar_helper(in, "PINSETID");
						/* specify transfer */
						if( strstr(ast_channel_name(in),"Local/") && strstr(ast_channel_name(in),"from-internal-xfer")
							&& strstr(ast_channel_name(in),";2"))
						{
							xtransfer = 1;
						}
						/* outbound route set password */
 						if( (strstr(ast_channel_name(c),"DAHDI/o") || strstr(ast_channel_name(c),"PJSIP/trunk"))
							&& (NULL == strstr(ast_channel_name(in),"PJSIP/trunk"))
							&& aauser && ((out_rt_pwd && strlen(out_rt_pwd)) || xtransfer || (pinsetid && strlen(pinsetid))))
						{
							snprintf(family, sizeof(family), "AAUSER/%s", aauser);
							ast_ucm_db_get(family, "HIGHCALL_THRESHOLD", data, sizeof(data) -1);
						}

						if(data[0] != 0 && atoi(data) != 0)
						{
							high_freq_count = atoi(data);
							q_index = ast_exten_map2queue_index((char *)aauser);
							gettimeofday(&now, NULL);

							if(-1 != q_index )
							{
								ast_log(LOG_DEBUG,"chan:%s insert a observe node, time=%ld, thres=%d", ast_channel_name(in),now.tv_sec,high_freq_count);
								ast_high_frequence_enqueue(ast_get_high_frequence_queue(q_index), (char *)aauser,high_freq_count, &now);
							}
						}
#endif
					/*End added by zyxia for Bug 204898 call ipc device,200 ok need add*/

						if (o->orig_chan_name
							&& strcmp(o->orig_chan_name, ast_channel_name(c))) {
							/*
							 * The channel name changed so we must generate COLP update.
							 * Likely because a call pickup channel masqueraded in.
							 */
							update_connected_line_from_peer(in, c, 1);
						} else if (!single && !ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
							if (o->pending_connected_update) {
								if (ast_channel_connected_line_sub(c, in, &o->connected, 0) &&
									ast_channel_connected_line_macro(c, in, &o->connected, 1, 0)) {
									ast_channel_update_connected_line(in, &o->connected, NULL);
								}
							} else if (!ast_test_flag64(o, DIAL_CALLERID_ABSENT)) {
								update_connected_line_from_peer(in, c, 1);
							}
						}
						if (o->aoc_s_rate_list) {
							size_t encoded_size;
							struct ast_aoc_encoded *encoded;
							if ((encoded = ast_aoc_encode(o->aoc_s_rate_list, &encoded_size, o->chan))) {
								ast_indicate_data(in, AST_CONTROL_AOC, encoded, encoded_size);
								ast_aoc_destroy_encoded(encoded);
							}
						}
						peer = c;

#ifdef GRANDSTREAM_NETWORKS
						publish_answered_else_event2im(in, out_chans, peer, NULL, "answer");
						/*Start added by zyxia for Bug 263733 ,callqueue time to destination and enable recorded*/
						if(ast_channel_get_is_queue_call(in) && strstr(ast_channel_name(in),"PJSIP")
							&& !strcasestr(ast_channel_name(in),"PJSIP/trunk") && ast_channel_tech(in)->indicate ){
							ast_channel_tech(in)->indicate(in, AST_CONTROL_REAL_ANSWERED, NULL, 0);
						}
						/*End added by zyxia for Bug 263733 ,callqueue time to destination and enable recorded*/
#endif
						/* Inform everyone else that they've been canceled.
						 * The dial end event for the peer will be sent out after
						 * other Dial options have been handled.
						 */
						publish_dial_end_event(in, out_chans, peer, "CANCEL");
						ast_copy_flags64(peerflags, o,
							OPT_CALLEE_TRANSFER | OPT_CALLER_TRANSFER |
							OPT_CALLEE_HANGUP | OPT_CALLER_HANGUP |
							OPT_CALLEE_MONITOR | OPT_CALLER_MONITOR |
							OPT_CALLEE_PARK | OPT_CALLER_PARK |
							OPT_CALLEE_MIXMONITOR | OPT_CALLER_MIXMONITOR |
							DIAL_NOFORWARDHTML);
						ast_channel_dialcontext_set(c, "");
						ast_channel_exten_set(c, "");
						if (CAN_EARLY_BRIDGE(peerflags, in, peer)) {
							/* Setup early bridge if appropriate */
							ast_channel_early_bridge(in, peer);
						}
#ifdef GRANDSTREAM_NETWORKS
						if (ast_channel_sfu_early_media(peer)) {
							sfu_handle_early_media(c, 0);
						}

						if (ast_channel_tech(in)->set_multicast_format) {
							ast_channel_tech(in)->set_multicast_format(in, ast_channel_readformat(peer));
						}
#endif
					}
					/* If call has been answered, then the eventual hangup is likely to be normal hangup */
					ast_channel_hangupcause_set(in, AST_CAUSE_NORMAL_CLEARING);
					ast_channel_hangupcause_set(c, AST_CAUSE_NORMAL_CLEARING);
#ifdef GRANDSTREAM_NETWORKS
					xfer_ringing = 0;
#endif
					break;
				case AST_CONTROL_BUSY:
					ast_verb(3, "%s is busy\n", ast_channel_name(c));
					ast_channel_hangupcause_set(in, ast_channel_hangupcause(c));
					ast_channel_publish_dial(in, c, NULL, "BUSY");
#ifdef GRANDSTREAM_NETWORKS
					if (ast_channel_sfu_early_media(c)) {
						sfu_handle_early_media(c, 0);
					}
#endif
					ast_hangup(c);
					c = o->chan = NULL;
					ast_clear_flag64(o, DIAL_STILLGOING);
					handle_cause(AST_CAUSE_BUSY, &num);
					break;
#ifdef GRANDSTREAM_NETWORKS
				case AST_CONTROL_UPDATE_VIDEO_PORT_ZERO:
				{
					RAII_VAR(struct ast_channel *, mychan, ao2_bump(ast_channel_get_mychan(in)), ast_channel_cleanup);
					if (mychan) {
						ast_log(LOG_NOTICE, "AppDial for update sdp channel:%s\n", ast_channel_name(mychan));
						ast_channel_lock(mychan);
						pbx_builtin_setvar_helper(mychan, "UPDATE_VIDEO_PORT_ZERO", "YES");
						ast_channel_unlock(mychan);
					} else {
						ast_indicate(in, AST_CONTROL_CLOSE_CAMERA);
						ast_log(LOG_DEBUG, "The channel(%s) second channel(;2) is null.\n", ast_channel_name(in));
					}
					break;
				}
#endif
				case AST_CONTROL_CONGESTION:
					ast_verb(3, "%s is circuit-busy\n", ast_channel_name(c));
					ast_channel_hangupcause_set(in, ast_channel_hangupcause(c));
					ast_channel_publish_dial(in, c, NULL, "CONGESTION");
#ifdef GRANDSTREAM_NETWORKS
					if (ast_channel_sfu_early_media(c)) {
						sfu_handle_early_media(c, 0);
					}
#endif
					ast_hangup(c);
					c = o->chan = NULL;
					ast_clear_flag64(o, DIAL_STILLGOING);
					handle_cause(AST_CAUSE_CONGESTION, &num);
					break;
				case AST_CONTROL_RINGING:
					/* This is a tricky area to get right when using a native
					 * CC agent. The reason is that we do the best we can to send only a
					 * single ringing notification to the caller.
					 *
					 * Call completion complicates the logic used here. CCNR is typically
					 * offered during a ringing message. Let's say that party A calls
					 * parties B, C, and D. B and C do not support CC requests, but D
					 * does. If we were to receive a ringing notification from B before
					 * the others, then we would end up sending a ringing message to
					 * A with no CCNR offer present.
					 *
					 * The approach that we have taken is that if we receive a ringing
					 * response from a party and no CCNR offer is present, we need to
					 * wait. Specifically, we need to wait until either a) a called party
					 * offers CCNR in its ringing response or b) all called parties have
					 * responded in some way to our call and none offers CCNR.
					 *
					 * The drawback to this is that if one of the parties has a delayed
					 * response or, god forbid, one just plain doesn't respond to our
					 * outgoing call, then this will result in a significant delay between
					 * when the caller places the call and hears ringback.
					 *
					 * Note also that if CC is disabled for this call, then it is perfectly
					 * fine for ringing frames to get sent through.
					 */
					++num_ringing;
#ifndef GRANDSTREAM_NETWORKS
					if (ignore_cc || cc_frame_received || num_ringing == numlines)
#else
					if (ignore_cc || cc_frame_received || num_ringing <= numlines)
#endif
					{
						ast_verb(3, "%s is ringing\n", ast_channel_name(c));
						/* Setup early media if appropriate */
						if (single && !caller_entertained
							&& CAN_EARLY_BRIDGE(peerflags, in, c)) {
							ast_channel_early_bridge(in, c);
						}
#ifndef GRANDSTREAM_NETWORKS
						if (!(pa->sentringing) && !ast_test_flag64(outgoing, OPT_MUSICBACK) && ast_strlen_zero(opt_args[OPT_ARG_RINGBACK])) {
							ast_indicate(in, AST_CONTROL_RINGING);
							pa->sentringing++;
						}
#else
						if ((0 == cc_play_rbt) && (NULL != pbx_builtin_getvar_helper(in, "CC_INTERFACES"))) {
							/* Modified by yfyang, for Bug303208 */
							if (ast_channel_state(in) <= AST_STATE_RING) {
								ast_indicate(in, AST_CONTROL_RINGING);
							} else {
								ast_moh_start(in, "rbt", NULL);
								cc_play_rbt = 1;
							}
						} else if (!(pa->sentringing) && !ast_test_flag64(outgoing, OPT_MUSICBACK) && ast_strlen_zero(opt_args[OPT_ARG_RINGBACK])) {
							ast_indicate(in, AST_CONTROL_RINGING);
							pa->sentringing++;
							progress_flag = 0;

							/* Added by yfyang, for Bug 200018 */
							if (strstr(ast_channel_name(in), "Local/")) {
								char *from_ivr = (char *)pbx_builtin_getvar_helper(in, "FROM_IVR");
								if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")) {
									char *special_tone = (char*)pbx_builtin_getvar_helper(in, "SPECIAL_TONE_NAME");
									if (!ast_strlen_zero(special_tone) && strcmp(special_tone, "nil")) {
										ast_moh_start(in, special_tone, NULL);
									} else {
										ast_moh_start(in, "rbt", NULL);
									}
								}
							}
						}
#endif
					}
					ast_channel_publish_dial(in, c, NULL, "RINGING");
					break;
				case AST_CONTROL_PROGRESS:
					ast_verb(3, "%s is making progress passing it to %s\n", ast_channel_name(c), ast_channel_name(in));
					/* Setup early media if appropriate */
					if (single && !caller_entertained
						&& CAN_EARLY_BRIDGE(peerflags, in, c)) {
						ast_channel_early_bridge(in, c);
					}
					if (!ast_test_flag64(outgoing, OPT_RINGBACK)) {
#ifdef GRANDSTREAM_NETWORKS
						int use_final_sdp = ast_channel_get_use_final_sdp(c);
						if (single || (!single && !pa->sentringing) || use_final_sdp) {
							ast_indicate_data2(in, AST_CONTROL_PROGRESS, c, ast_channel_size_get());
							/* Add by bfdu for Bug 311695, 2024.8.16 */
							pbx_builtin_setvar_helper(c, "INDICATE_PROGRESS", "yes");
							progress_num++;
							/* Add by bfdu for Bug 194445, 2021.5.11 */
							if (!strcasecmp(ast_channel_tech(in)->type, "Local")) {
								ast_log(LOG_DEBUG, "Chan:%s is Local, make early bridge\n", ast_channel_name(c));
								ast_avs_join_early_bridge(in, c);
							}
						}
#else
						if (single || (!single && !pa->sentringing)) {
							ast_indicate(in, AST_CONTROL_PROGRESS);
						}
#endif
					}
					if (!ast_strlen_zero(dtmf_progress)) {
						ast_verb(3,
							"Sending DTMF '%s' to the called party as result of receiving a PROGRESS message.\n",
							dtmf_progress);
						ast_dtmf_stream(c, in, dtmf_progress, 250, 0);
					}
#ifdef GRANDSTREAM_NETWORKS
					if (!strcasecmp(ast_channel_tech(c)->type, "PJSIP")) {
						const char *is_sfu = pbx_builtin_getvar_helper(c, "IS_SFU");
						if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
							sfu_handle_early_media(c, 1);
						}
					}
					/*for bug 153395*/
					if (ast_channel_generatordata(in)) {
						/* added by zgmo for Bug164169 and Bug164307 */
						const char *ringgroup_moh = pbx_builtin_getvar_helper(in, "RINGGROUP_MOH");
						const char *do_forward = pbx_builtin_getvar_helper(c, "DO_FORWARD");
						if (ast_strlen_zero(ringgroup_moh) && ast_strlen_zero(do_forward)) {
							ast_deactivate_generator(in);
						}
					}
					progress_flag = 1;
#endif
					ast_channel_publish_dial(in, c, NULL, "PROGRESS");
					break;
#ifdef GRANDSTREAM_NETWORKS
				case AST_CONTROL_PAI_INFO:
					ast_indicate_data(in, AST_CONTROL_PAI_INFO, f->data.ptr, f->datalen);
					break;
#endif
				case AST_CONTROL_VIDUPDATE:
				case AST_CONTROL_SRCUPDATE:
				case AST_CONTROL_SRCCHANGE:
					if (!single || caller_entertained) {
						break;
					}
					ast_verb(3, "%s requested media update control %d, passing it to %s\n",
						ast_channel_name(c), f->subclass.integer, ast_channel_name(in));
					ast_indicate(in, f->subclass.integer);
					break;
				case AST_CONTROL_CONNECTED_LINE:
					if (ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
						ast_verb(3, "Connected line update to %s prevented.\n", ast_channel_name(in));
						break;
					}
					if (!single) {
						struct ast_party_connected_line connected;

						ast_verb(3, "%s connected line has changed. Saving it until answer for %s\n",
							ast_channel_name(c), ast_channel_name(in));
						ast_party_connected_line_set_init(&connected, &o->connected);
						ast_connected_line_parse_data(f->data.ptr, f->datalen, &connected);
						ast_party_connected_line_set(&o->connected, &connected, NULL);
						ast_party_connected_line_free(&connected);
						o->pending_connected_update = 1;
						break;
					}
					if (ast_channel_connected_line_sub(c, in, f, 1) &&
						ast_channel_connected_line_macro(c, in, f, 1, 1)) {
						ast_indicate_data(in, AST_CONTROL_CONNECTED_LINE, f->data.ptr, f->datalen);
					}
					break;
				case AST_CONTROL_AOC:
					{
						struct ast_aoc_decoded *decoded = ast_aoc_decode(f->data.ptr, f->datalen, o->chan);
						if (decoded && (ast_aoc_get_msg_type(decoded) == AST_AOC_S)) {
							ast_aoc_destroy_decoded(o->aoc_s_rate_list);
							o->aoc_s_rate_list = decoded;
						} else {
							ast_aoc_destroy_decoded(decoded);
						}
					}
					break;
				case AST_CONTROL_REDIRECTING:
					if (!single) {
						/*
						 * Redirecting updates to the caller make sense only on single
						 * calls.
						 */
						break;
					}
					if (ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
						ast_verb(3, "Redirecting update to %s prevented.\n", ast_channel_name(in));
						break;
					}
					ast_verb(3, "%s redirecting info has changed, passing it to %s\n",
						ast_channel_name(c), ast_channel_name(in));
					if (ast_channel_redirecting_sub(c, in, f, 1) &&
						ast_channel_redirecting_macro(c, in, f, 1, 1)) {
						ast_indicate_data(in, AST_CONTROL_REDIRECTING, f->data.ptr, f->datalen);
					}
					pa->sentringing = 0;
					break;
				case AST_CONTROL_PROCEEDING:
					ast_verb(3, "%s is proceeding passing it to %s\n", ast_channel_name(c), ast_channel_name(in));
#ifdef GRANDSTREAM_NETWORKS
					if (0 == strncmp(ast_channel_name(c), "DAHDI/i", strlen("DAHDI/i"))){
						is_pri = 1;
					}
					ast_log(LOG_DEBUG, "is_pri:%s", is_pri?"yes":"no");
#endif
					if (single && !caller_entertained
						&& CAN_EARLY_BRIDGE(peerflags, in, c)) {
						ast_channel_early_bridge(in, c);
					}
					if (!ast_test_flag64(outgoing, OPT_RINGBACK))
						ast_indicate(in, AST_CONTROL_PROCEEDING);
					ast_channel_publish_dial(in, c, NULL, "PROCEEDING");
					break;
				case AST_CONTROL_HOLD:
					/* XXX this should be saved like AST_CONTROL_CONNECTED_LINE for !single || caller_entertained */
					ast_verb(3, "Call on %s placed on hold\n", ast_channel_name(c));
					ast_indicate_data(in, AST_CONTROL_HOLD, f->data.ptr, f->datalen);
					break;
				case AST_CONTROL_UNHOLD:
					/* XXX this should be saved like AST_CONTROL_CONNECTED_LINE for !single || caller_entertained */
					ast_verb(3, "Call on %s left from hold\n", ast_channel_name(c));
					ast_indicate(in, AST_CONTROL_UNHOLD);
					break;
				case AST_CONTROL_OFFHOOK:
				case AST_CONTROL_FLASH:
					/* Ignore going off hook and flash */
					break;
				case AST_CONTROL_CC:
					if (!ignore_cc) {
						ast_handle_cc_control_frame(in, c, f->data.ptr);
						cc_frame_received = 1;
					}
					break;
				case AST_CONTROL_PVT_CAUSE_CODE:
					ast_indicate_data(in, AST_CONTROL_PVT_CAUSE_CODE, f->data.ptr, f->datalen);
					break;
#ifdef GRANDSTREAM_NETWORKS
				case AST_CONTROL_TRANSFER_COMPLETE:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_TRANSFER_COMPLETE\n", ast_channel_name(in));
					ast_indicate(in, AST_CONTROL_TRANSFER_COMPLETE);
					break;
				case AST_CONTROL_CHANNEL_IS_NOT_SIP:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_CHANNEL_IS_NOT_SIP\n", ast_channel_name(in));
					ast_indicate(in, AST_CONTROL_CHANNEL_IS_NOT_SIP);
					break;
				case AST_CONTROL_GET_PEER_SESSION:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_GET_PEER_SESSION\n", ast_channel_name(in));
					ast_indicate(in, AST_CONTROL_GET_PEER_SESSION);
					break;
				case AST_CONTROL_PEER_SESSION:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_GET_PEER_SESSION\n", ast_channel_name(in));
					ast_indicate_data(in, AST_CONTROL_PEER_SESSION, f->data.ptr, f->datalen);
					break;
				case AST_CONTROL_TRANSFERER_HANGUP:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_TRANSFERER_HANGUP\n", ast_channel_name(in));
					ast_channel_set_transferer_hanguped(in, 1);
					break;
				case AST_CONTROL_REFER_TRANSFERER_HANGUP:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_REFER_TRANSFERER_HANGUP, out:%s\n", ast_channel_name(in), ast_channel_name(c));
					ast_channel_set_send_slides_video(c, ast_channel_get_send_slides_video(in));
					ast_channel_set_recv_slides_video(c, ast_channel_get_recv_slides_video(in));
					ast_channel_set_transferer_hanguped(o->chan, 1);
					break;
				case AST_CONTROL_NOTIFY_TOPOLOY:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_NOTIFY_TOPOLOY, out:%s\n", ast_channel_name(in), ast_channel_name(c));
					ast_indicate_data(in, AST_CONTROL_NOTIFY_TOPOLOY, f->data.ptr, f->datalen);
					break;
				case AST_CONTROL_PEER_IS_NAT_CALL:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_PEER_IS_NAT_CALL\n", ast_channel_name(in));
					ast_indicate(in, AST_CONTROL_PEER_IS_NAT_CALL);
					break;
				case AST_CONTROL_NOTIFY_TRUNK_INFO:
					ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_NOTIFY_TRUNK_INFO\n", ast_channel_name(in));
					ast_indicate_data(in, AST_CONTROL_NOTIFY_TRUNK_INFO, f->data.ptr, f->datalen);
					break;
#endif
				case -1:
					if (single && !caller_entertained) {
						ast_verb(3, "%s stopped sounds\n", ast_channel_name(c));
						ast_indicate(in, -1);
						pa->sentringing = 0;
					}
					break;
				default:
					ast_debug(1, "Dunno what to do with control type %d\n", f->subclass.integer);
					break;
				}
				break;
			case AST_FRAME_VIDEO:
			case AST_FRAME_VOICE:
			case AST_FRAME_IMAGE:
				if (caller_entertained) {
					break;
				}
				/* Fall through */
			case AST_FRAME_TEXT:
#ifdef GRANDSTREAM_NETWORKS
			case AST_FRAME_DTMF_BEGIN:
			case AST_FRAME_DTMF_END:
			case AST_FRAME_TEXT_DATA:
				// When receive multi 183 response, only the last channel in early bridge.
				if (ast_write(in, f)) {
					ast_log(LOG_WARNING, "Unable to write frametype: %u\n",
						f->frametype);
				}
#else
				if (single && ast_write(in, f)) {
					ast_log(LOG_WARNING, "Unable to write frametype: %u\n",
						f->frametype);
				}
#endif
				break;
			case AST_FRAME_HTML:
				if (single && !ast_test_flag64(outgoing, DIAL_NOFORWARDHTML)
					&& ast_channel_sendhtml(in, f->subclass.integer, f->data.ptr, f->datalen) == -1) {
					ast_log(LOG_WARNING, "Unable to send URL\n");
				}
				break;
			default:
				break;
			}
			ast_frfree(f);
		} /* end for */
		if (winner == in) {
			struct ast_frame *f = ast_read(in);
#if 0
			if (f && (f->frametype != AST_FRAME_VOICE))
				printf("Frame type: %d, %d\n", f->frametype, f->subclass);
			else if (!f || (f->frametype != AST_FRAME_VOICE))
				printf("Hangup received on %s\n", in->name);
#endif
			if (!f || ((f->frametype == AST_FRAME_CONTROL) && (f->subclass.integer == AST_CONTROL_HANGUP))) {
#ifdef GRANDSTREAM_NETWORKS
				const char *in_group = pbx_builtin_getvar_helper(in, "IN_RINGGROUP");
				if (!ast_strlen_zero(in_group) && !strncmp(in_group, "yes", strlen("yes"))) {
					*to = 0;
				} else {
					*to = -1;
				}
#else
				/* Got hung up */
				*to = -1;
#endif
				strcpy(pa->status, "CANCEL");
				publish_dial_end_event(in, out_chans, NULL, pa->status);
				if (f) {
					if (f->data.uint32) {
						ast_channel_hangupcause_set(in, f->data.uint32);
					}
					ast_frfree(f);
				}
				if (is_cc_recall) {
					ast_cc_completed(in, "CC completed, although the caller hung up (cancelled)");
#ifdef GRANDSTREAM_NETWORKS
					if (cc_play_rbt) {
						ast_moh_stop(in);
						cc_play_rbt = 0;
					}
#endif
				}
				return NULL;
			}

			/* now f is guaranteed non-NULL */
			if (f->frametype == AST_FRAME_DTMF) {
				if (ast_test_flag64(peerflags, OPT_DTMF_EXIT)) {
					const char *context;
					ast_channel_lock(in);
					context = pbx_builtin_getvar_helper(in, "EXITCONTEXT");
					if (onedigit_goto(in, context, (char) f->subclass.integer, 1)) {
						ast_verb(3, "User hit %c to disconnect call.\n", f->subclass.integer);
						*to = 0;
						*result = f->subclass.integer;
						strcpy(pa->status, "CANCEL");
						publish_dial_end_event(in, out_chans, NULL, pa->status);
						ast_frfree(f);
						ast_channel_unlock(in);
						if (is_cc_recall) {
							ast_cc_completed(in, "CC completed, but the caller used DTMF to exit");
						}

#ifdef GRANDSTREAM_NETWORKS
						if (cc_play_rbt) {
							ast_moh_stop(in);
							cc_play_rbt = 0;
						}
#endif
						return NULL;
					}
					ast_channel_unlock(in);
				}

				if (ast_test_flag64(peerflags, OPT_CALLER_HANGUP) &&
					detect_disconnect(in, f->subclass.integer, &featurecode)) {
					ast_verb(3, "User requested call disconnect.\n");
					*to = 0;
					strcpy(pa->status, "CANCEL");
					publish_dial_end_event(in, out_chans, NULL, pa->status);
					ast_frfree(f);
					if (is_cc_recall) {
						ast_cc_completed(in, "CC completed, but the caller hung up with DTMF");
					}

#ifdef GRANDSTREAM_NETWORKS
					if (cc_play_rbt) {
						ast_moh_stop(in);
						cc_play_rbt = 0;
					}
#endif
					return NULL;
				}
			}

			/* Send the frame from the in channel to all outgoing channels. */
			AST_LIST_TRAVERSE(out_chans, o, node) {
				if (!o->chan || !ast_test_flag64(o, DIAL_STILLGOING)) {
					/* This outgoing channel has died so don't send the frame to it. */
					continue;
				}
				switch (f->frametype) {
				case AST_FRAME_HTML:
					/* Forward HTML stuff */
					if (!ast_test_flag64(o, DIAL_NOFORWARDHTML)
						&& ast_channel_sendhtml(o->chan, f->subclass.integer, f->data.ptr, f->datalen) == -1) {
						ast_log(LOG_WARNING, "Unable to send URL\n");
					}
					break;
				case AST_FRAME_VIDEO:
				case AST_FRAME_VOICE:
				case AST_FRAME_IMAGE:
					if (!single || caller_entertained) {
						/*
						 * We are calling multiple parties or caller is being
						 * entertained and has thus not been made compatible.
						 * No need to check any other called parties.
						 */
						goto skip_frame;
					}
					/* Fall through */
				case AST_FRAME_TEXT:
				case AST_FRAME_DTMF_BEGIN:
				case AST_FRAME_DTMF_END:
#ifdef GRANDSTREAM_NETWORKS
				case AST_FRAME_TEXT_DATA:
#endif
					if (ast_write(o->chan, f)) {
						ast_log(LOG_WARNING, "Unable to forward frametype: %u\n",
							f->frametype);
					}
					break;
				case AST_FRAME_CONTROL:
					switch (f->subclass.integer) {
					case AST_CONTROL_HOLD:
#ifdef GRANDSTREAM_NETWORKS
						/* Added by yfyang, for Bug293297 */
						if (ast_channel_state(o->chan) < AST_STATE_UP) {
							ast_verb(3, "chan[%s] in state[%s], ignore HOLD indicate.\n", ast_channel_name(o->chan), ast_state2str(ast_channel_state(o->chan)));
							break;
						}
#endif
						ast_verb(3, "Call on %s placed on hold\n", ast_channel_name(o->chan));
						ast_indicate_data(o->chan, AST_CONTROL_HOLD, f->data.ptr, f->datalen);
						break;
					case AST_CONTROL_UNHOLD:
						ast_verb(3, "Call on %s left from hold\n", ast_channel_name(o->chan));
						ast_indicate(o->chan, AST_CONTROL_UNHOLD);
						break;
					case AST_CONTROL_VIDUPDATE:
					case AST_CONTROL_SRCUPDATE:
					case AST_CONTROL_SRCCHANGE:
						if (!single || caller_entertained) {
							/*
							 * We are calling multiple parties or caller is being
							 * entertained and has thus not been made compatible.
							 * No need to check any other called parties.
							 */
							goto skip_frame;
						}
						ast_verb(3, "%s requested media update control %d, passing it to %s\n",
							ast_channel_name(in), f->subclass.integer, ast_channel_name(o->chan));
						ast_indicate(o->chan, f->subclass.integer);
						break;
					case AST_CONTROL_CONNECTED_LINE:
						if (ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
							ast_verb(3, "Connected line update to %s prevented.\n", ast_channel_name(o->chan));
							break;
						}
						if (ast_channel_connected_line_sub(in, o->chan, f, 1) &&
							ast_channel_connected_line_macro(in, o->chan, f, 0, 1)) {
							ast_indicate_data(o->chan, f->subclass.integer, f->data.ptr, f->datalen);
						}
						break;
					case AST_CONTROL_REDIRECTING:
						if (ast_test_flag64(o, OPT_IGNORE_CONNECTEDLINE)) {
							ast_verb(3, "Redirecting update to %s prevented.\n", ast_channel_name(o->chan));
							break;
						}
						if (ast_channel_redirecting_sub(in, o->chan, f, 1) &&
							ast_channel_redirecting_macro(in, o->chan, f, 0, 1)) {
							ast_indicate_data(o->chan, f->subclass.integer, f->data.ptr, f->datalen);
						}
						break;
#ifdef GRANDSTREAM_NETWORKS
					case AST_CONTROL_EARLY_MEDIA:
						ast_log(LOG_DEBUG, "Recv AST_CONTROL_EARLY_MEDIA\n");
						if (ast_channel_is_avs_engine(in) && ast_channel_recv_early_media_rsp_get(o->chan)) {
							ast_avs_join_early_bridge(in, o->chan);
							ast_channel_recv_early_media_rsp_set(o->chan, 0);
						}
						break;
					case AST_CONTROL_TRANSFER_COMPLETE:
						ast_log(LOG_DEBUG, "Out channel:%s recieve AST_CONTROL_TRANSFER_COMPLETE\n", ast_channel_name(o->chan));
						ast_indicate(o->chan, AST_CONTROL_TRANSFER_COMPLETE);
						break;
					case AST_CONTROL_CHANNEL_IS_NOT_SIP:
						ast_log(LOG_DEBUG, "Out channel:%s recieve AST_CONTROL_CHANNEL_IS_NOT_SIP\n", ast_channel_name(o->chan));
						ast_indicate(o->chan, AST_CONTROL_CHANNEL_IS_NOT_SIP);
						break;
					case AST_CONTROL_GET_PEER_SESSION:
						ast_log(LOG_DEBUG, "Out channel:%s recieve AST_CONTROL_GET_PEER_SESSION\n", ast_channel_name(o->chan));
						ast_indicate(o->chan, AST_CONTROL_GET_PEER_SESSION);
						break;
					case AST_CONTROL_PEER_SESSION:
						ast_log(LOG_DEBUG, "Out channel:%s recieve AST_CONTROL_PEER_SESSION\n", ast_channel_name(o->chan));
						ast_indicate_data(o->chan, AST_CONTROL_PEER_SESSION, f->data.ptr, f->datalen);
						break;
					case AST_CONTROL_TRANSFERER_HANGUP:
						ast_log(LOG_DEBUG, "Out channel:%s recieve AST_CONTROL_TRANSFERER_HANGUP\n", ast_channel_name(o->chan));
						ast_channel_set_transferer_hanguped(o->chan, 1);
						break;
					case AST_CONTROL_REFER_TRANSFERER_HANGUP:
						ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_REFER_TRANSFERER_HANGUP, out:%s\n", ast_channel_name(in), ast_channel_name(o->chan));
						ast_channel_set_send_slides_video(o->chan, ast_channel_get_send_slides_video(in));
						ast_channel_set_recv_slides_video(o->chan, ast_channel_get_recv_slides_video(in));
						ast_channel_set_transferer_hanguped(o->chan, 1);
						if (strstr(ast_channel_name(o->chan), "trunk_")){
							ast_channel_reg_callernum_set(o->chan, "");
							ast_channel_reg_callername_set(o->chan, "");
							ast_channel_reg_calleenum_set(o->chan, "");
							pbx_builtin_setvar_helper(o->chan,"REG_CALLERNUM","");
							pbx_builtin_setvar_helper(o->chan,"REG_CALELRNAME","");
							pbx_builtin_setvar_helper(o->chan,"REG_CALLEENUM","");
						}
						break;
					case AST_CONTROL_NOTIFY_TOPOLOY:
						ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_NOTIFY_TOPOLOY, out:%s\n", ast_channel_name(in), ast_channel_name(o->chan));
						ast_indicate_data(o->chan, AST_CONTROL_NOTIFY_TOPOLOY, f->data.ptr, f->datalen);
						break;
					case AST_CONTROL_PEER_IS_NAT_CALL:
						ast_log(LOG_DEBUG, "Out channel:%s recieve AST_CONTROL_PEER_IS_NAT_CALL\n", ast_channel_name(o->chan));
						ast_indicate(o->chan, AST_CONTROL_PEER_IS_NAT_CALL);
						break;
					case AST_CONTROL_NOTIFY_TRUNK_INFO:
						ast_log(LOG_DEBUG, "Channel:%s recieve AST_CONTROL_NOTIFY_TRUNK_INFO, out:%s\n", ast_channel_name(in), ast_channel_name(o->chan));
						ast_indicate_data(o->chan, AST_CONTROL_NOTIFY_TRUNK_INFO, f->data.ptr, f->datalen);
						break;
#endif
					default:
						/* We are not going to do anything with this frame. */
						goto skip_frame;
					}
					break;
				default:
					/* We are not going to do anything with this frame. */
					goto skip_frame;
				}
			}
skip_frame:;
			ast_frfree(f);
		}
	}

	if (!*to || ast_check_hangup(in)) {
		ast_verb(3, "Nobody picked up in %d ms\n", orig);
		publish_dial_end_event(in, out_chans, NULL, "NOANSWER");
	}

	if (is_cc_recall) {
		ast_cc_completed(in, "Recall completed!");
	}

#ifdef GRANDSTREAM_NETWORKS
	if (cc_play_rbt) {
		ast_moh_stop(in);
		cc_play_rbt = 0;
	}
#endif
	return peer;
}

static int detect_disconnect(struct ast_channel *chan, char code, struct ast_str **featurecode)
{
	char disconnect_code[AST_FEATURE_MAX_LEN];
	int res;

	ast_str_append(featurecode, 1, "%c", code);

	res = ast_get_builtin_feature(chan, "disconnect", disconnect_code, sizeof(disconnect_code));
	if (res) {
		ast_str_reset(*featurecode);
		return 0;
	}

	if (strlen(disconnect_code) > ast_str_strlen(*featurecode)) {
		/* Could be a partial match, anyway */
		if (strncmp(disconnect_code, ast_str_buffer(*featurecode), ast_str_strlen(*featurecode))) {
			ast_str_reset(*featurecode);
		}
		return 0;
	}

	if (strcmp(disconnect_code, ast_str_buffer(*featurecode))) {
		ast_str_reset(*featurecode);
		return 0;
	}

	return 1;
}

/* returns true if there is a valid privacy reply */
static int valid_priv_reply(struct ast_flags64 *opts, int res)
{
	if (res < '1')
		return 0;
	if (ast_test_flag64(opts, OPT_PRIVACY) && res <= '5')
		return 1;
	if (ast_test_flag64(opts, OPT_SCREENING) && res <= '4')
		return 1;
	return 0;
}

static int do_privacy(struct ast_channel *chan, struct ast_channel *peer,
	struct ast_flags64 *opts, char **opt_args, struct privacy_args *pa)
{

	int res2;
	int loopcount = 0;

	/* Get the user's intro, store it in priv-callerintros/$CID,
	   unless it is already there-- this should be done before the
	   call is actually dialed  */

	/* all ring indications and moh for the caller has been halted as soon as the
	   target extension was picked up. We are going to have to kill some
	   time and make the caller believe the peer hasn't picked up yet */

	if (ast_test_flag64(opts, OPT_MUSICBACK) && !ast_strlen_zero(opt_args[OPT_ARG_MUSICBACK])) {
		char *original_moh = ast_strdupa(ast_channel_musicclass(chan));
		ast_indicate(chan, -1);
		ast_channel_musicclass_set(chan, opt_args[OPT_ARG_MUSICBACK]);
		ast_moh_start(chan, opt_args[OPT_ARG_MUSICBACK], NULL);
		ast_channel_musicclass_set(chan, original_moh);
	} else if (ast_test_flag64(opts, OPT_RINGBACK) || ast_test_flag64(opts, OPT_RING_WITH_EARLY_MEDIA)) {
		ast_indicate(chan, AST_CONTROL_RINGING);
		pa->sentringing++;
	}

	/* Start autoservice on the other chan ?? */
	res2 = ast_autoservice_start(chan);
	/* Now Stream the File */
	for (loopcount = 0; loopcount < 3; loopcount++) {
		if (res2 && loopcount == 0) /* error in ast_autoservice_start() */
			break;
		if (!res2) /* on timeout, play the message again */
			res2 = ast_play_and_wait(peer, "priv-callpending");
		if (!valid_priv_reply(opts, res2))
			res2 = 0;
		/* priv-callpending script:
		   "I have a caller waiting, who introduces themselves as:"
		*/
		if (!res2)
			res2 = ast_play_and_wait(peer, pa->privintro);
		if (!valid_priv_reply(opts, res2))
			res2 = 0;
		/* now get input from the called party, as to their choice */
		if (!res2) {
			/* XXX can we have both, or they are mutually exclusive ? */
			if (ast_test_flag64(opts, OPT_PRIVACY))
				res2 = ast_play_and_wait(peer, "priv-callee-options");
			if (ast_test_flag64(opts, OPT_SCREENING))
				res2 = ast_play_and_wait(peer, "screen-callee-options");
		}

		/*! \page DialPrivacy Dial Privacy scripts
		 * \par priv-callee-options script:
		 * \li Dial 1 if you wish this caller to reach you directly in the future,
		 * 	and immediately connect to their incoming call.
		 * \li Dial 2 if you wish to send this caller to voicemail now and forevermore.
		 * \li Dial 3 to send this caller to the torture menus, now and forevermore.
		 * \li Dial 4 to send this caller to a simple "go away" menu, now and forevermore.
		 * \li Dial 5 to allow this caller to come straight thru to you in the future,
		 * 	but right now, just this once, send them to voicemail.
		 *
		 * \par screen-callee-options script:
		 * \li Dial 1 if you wish to immediately connect to the incoming call
		 * \li Dial 2 if you wish to send this caller to voicemail.
		 * \li Dial 3 to send this caller to the torture menus.
		 * \li Dial 4 to send this caller to a simple "go away" menu.
		 */
		if (valid_priv_reply(opts, res2))
			break;
		/* invalid option */
		res2 = ast_play_and_wait(peer, "vm-sorry");
	}

	if (ast_test_flag64(opts, OPT_MUSICBACK)) {
		ast_moh_stop(chan);
	} else if (ast_test_flag64(opts, OPT_RINGBACK) || ast_test_flag64(opts, OPT_RING_WITH_EARLY_MEDIA)) {
		ast_indicate(chan, -1);
		pa->sentringing = 0;
	}
	ast_autoservice_stop(chan);
	if (ast_test_flag64(opts, OPT_PRIVACY) && (res2 >= '1' && res2 <= '5')) {
		/* map keypresses to various things, the index is res2 - '1' */
		static const char * const _val[] = { "ALLOW", "DENY", "TORTURE", "KILL", "ALLOW" };
		static const int _flag[] = { AST_PRIVACY_ALLOW, AST_PRIVACY_DENY, AST_PRIVACY_TORTURE, AST_PRIVACY_KILL, AST_PRIVACY_ALLOW};
		int i = res2 - '1';
		ast_verb(3, "--Set privacy database entry %s/%s to %s\n",
			opt_args[OPT_ARG_PRIVACY], pa->privcid, _val[i]);
		ast_privacy_set(opt_args[OPT_ARG_PRIVACY], pa->privcid, _flag[i]);
	}
	switch (res2) {
	case '1':
		break;
	case '2':
		ast_copy_string(pa->status, "NOANSWER", sizeof(pa->status));
		break;
	case '3':
		ast_copy_string(pa->status, "TORTURE", sizeof(pa->status));
		break;
	case '4':
		ast_copy_string(pa->status, "DONTCALL", sizeof(pa->status));
		break;
	case '5':
		if (ast_test_flag64(opts, OPT_PRIVACY)) {
			ast_copy_string(pa->status, "NOANSWER", sizeof(pa->status));
			break;
		}
		/* if not privacy, then 5 is the same as "default" case */
	default: /* bad input or -1 if failure to start autoservice */
		/* well, if the user messes up, ... he had his chance... What Is The Best Thing To Do?  */
		/* well, there seems basically two choices. Just patch the caller thru immediately,
			  or,... put 'em thru to voicemail. */
		/* since the callee may have hung up, let's do the voicemail thing, no database decision */
		ast_log(LOG_NOTICE, "privacy: no valid response from the callee. Sending the caller to voicemail, the callee isn't responding\n");
		/* XXX should we set status to DENY ? */
		/* XXX what about the privacy flags ? */
		break;
	}

	if (res2 == '1') { /* the only case where we actually connect */
		/* if the intro is NOCALLERID, then there's no reason to leave it on disk, it'll
		   just clog things up, and it's not useful information, not being tied to a CID */
		if (strncmp(pa->privcid, "NOCALLERID", strlen("NOCALLERID")) == 0 || ast_test_flag64(opts, OPT_SCREEN_NOINTRO)) {
			ast_filedelete(pa->privintro, NULL);
			if (ast_fileexists(pa->privintro, NULL, NULL) > 0)
				ast_log(LOG_NOTICE, "privacy: ast_filedelete didn't do its job on %s\n", pa->privintro);
			else
				ast_verb(3, "Successfully deleted %s intro file\n", pa->privintro);
		}
		return 0; /* the good exit path */
	} else {
		return -1;
	}
}

/*! \brief returns 1 if successful, 0 or <0 if the caller should 'goto out' */
static int setup_privacy_args(struct privacy_args *pa,
	struct ast_flags64 *opts, char *opt_args[], struct ast_channel *chan)
{
	char callerid[60];
	int res;
	char *l;

	if (ast_channel_caller(chan)->id.number.valid
		&& !ast_strlen_zero(ast_channel_caller(chan)->id.number.str)) {
		l = ast_strdupa(ast_channel_caller(chan)->id.number.str);
		ast_shrink_phone_number(l);
		if (ast_test_flag64(opts, OPT_PRIVACY) ) {
			ast_verb(3, "Privacy DB is '%s', clid is '%s'\n", opt_args[OPT_ARG_PRIVACY], l);
			pa->privdb_val = ast_privacy_check(opt_args[OPT_ARG_PRIVACY], l);
		} else {
			ast_verb(3, "Privacy Screening, clid is '%s'\n", l);
			pa->privdb_val = AST_PRIVACY_UNKNOWN;
		}
	} else {
		char *tnam, *tn2;

		tnam = ast_strdupa(ast_channel_name(chan));
		/* clean the channel name so slashes don't try to end up in disk file name */
		for (tn2 = tnam; *tn2; tn2++) {
			if (*tn2 == '/')  /* any other chars to be afraid of? */
				*tn2 = '=';
		}
		ast_verb(3, "Privacy-- callerid is empty\n");

		snprintf(callerid, sizeof(callerid), "NOCALLERID_%s%s", ast_channel_exten(chan), tnam);
		l = callerid;
		pa->privdb_val = AST_PRIVACY_UNKNOWN;
	}

	ast_copy_string(pa->privcid, l, sizeof(pa->privcid));

	if (strncmp(pa->privcid, "NOCALLERID", strlen("NOCALLERID")) != 0 && ast_test_flag64(opts, OPT_SCREEN_NOCALLERID)) {
		/* if callerid is set and OPT_SCREEN_NOCALLERID is set also */
		ast_verb(3, "CallerID set (%s); N option set; Screening should be off\n", pa->privcid);
		pa->privdb_val = AST_PRIVACY_ALLOW;
	} else if (ast_test_flag64(opts, OPT_SCREEN_NOCALLERID) && strncmp(pa->privcid, "NOCALLERID", strlen("NOCALLERID")) == 0) {
		ast_verb(3, "CallerID blank; N option set; Screening should happen; dbval is %d\n", pa->privdb_val);
	}

	if (pa->privdb_val == AST_PRIVACY_DENY) {
		ast_verb(3, "Privacy DB reports PRIVACY_DENY for this callerid. Dial reports unavailable\n");
		ast_copy_string(pa->status, "NOANSWER", sizeof(pa->status));
		return 0;
	} else if (pa->privdb_val == AST_PRIVACY_KILL) {
		ast_copy_string(pa->status, "DONTCALL", sizeof(pa->status));
		return 0; /* Is this right? */
	} else if (pa->privdb_val == AST_PRIVACY_TORTURE) {
		ast_copy_string(pa->status, "TORTURE", sizeof(pa->status));
		return 0; /* is this right??? */
	} else if (pa->privdb_val == AST_PRIVACY_UNKNOWN) {
		/* Get the user's intro, store it in priv-callerintros/$CID,
		   unless it is already there-- this should be done before the
		   call is actually dialed  */

		/* make sure the priv-callerintros dir actually exists */
		snprintf(pa->privintro, sizeof(pa->privintro), "%s/sounds/priv-callerintros", ast_config_AST_DATA_DIR);
		if ((res = ast_mkdir(pa->privintro, 0755))) {
			ast_log(LOG_WARNING, "privacy: can't create directory priv-callerintros: %s\n", strerror(res));
			return -1;
		}

		snprintf(pa->privintro, sizeof(pa->privintro), "priv-callerintros/%s", pa->privcid);
		if (ast_fileexists(pa->privintro, NULL, NULL ) > 0 && strncmp(pa->privcid, "NOCALLERID", strlen("NOCALLERID")) != 0) {
			/* the DELUX version of this code would allow this caller the
			   option to hear and retape their previously recorded intro.
			*/
		} else {
			int duration; /* for feedback from play_and_wait */
			/* the file doesn't exist yet. Let the caller submit his
			   vocal intro for posterity */
			/* priv-recordintro script:

			   "At the tone, please say your name:"

			*/
			int silencethreshold = ast_dsp_get_threshold_from_settings(THRESHOLD_SILENCE);
			ast_answer(chan);
			res = ast_play_and_record(chan, "priv-recordintro", pa->privintro, 4, "sln", &duration, NULL, silencethreshold, 2000, 0);  /* NOTE: I've reduced the total time to 4 sec */
									/* don't think we'll need a lock removed, we took care of
									   conflicts by naming the pa.privintro file */
			if (res == -1) {
				/* Delete the file regardless since they hung up during recording */
				ast_filedelete(pa->privintro, NULL);
				if (ast_fileexists(pa->privintro, NULL, NULL) > 0)
					ast_log(LOG_NOTICE, "privacy: ast_filedelete didn't do its job on %s\n", pa->privintro);
				else
					ast_verb(3, "Successfully deleted %s intro file\n", pa->privintro);
				return -1;
			}
			if (!ast_streamfile(chan, "vm-dialout", ast_channel_language(chan)) )
				ast_waitstream(chan, "");
		}
	}
	return 1; /* success */
}

static void end_bridge_callback(void *data)
{
	char buf[80];
	time_t end;
	struct ast_channel *chan = data;

	time(&end);

	ast_channel_lock(chan);
	ast_channel_stage_snapshot(chan);
	snprintf(buf, sizeof(buf), "%d", ast_channel_get_up_time(chan));
	pbx_builtin_setvar_helper(chan, "ANSWEREDTIME", buf);
	snprintf(buf, sizeof(buf), "%d", ast_channel_get_duration(chan));
	pbx_builtin_setvar_helper(chan, "DIALEDTIME", buf);
	ast_channel_stage_snapshot_done(chan);
	ast_channel_unlock(chan);
}

static void end_bridge_callback_data_fixup(struct ast_bridge_config *bconfig, struct ast_channel *originator, struct ast_channel *terminator) {
	bconfig->end_bridge_callback_data = originator;
}

static int dial_handle_playtones(struct ast_channel *chan, const char *data)
{
	struct ast_tone_zone_sound *ts = NULL;
	int res;
	const char *str = data;

	if (ast_strlen_zero(str)) {
		ast_debug(1,"Nothing to play\n");
		return -1;
	}

	ts = ast_get_indication_tone(ast_channel_zone(chan), str);

	if (ts && ts->data[0]) {
		res = ast_playtones_start(chan, 0, ts->data, 0);
	} else {
		res = -1;
	}

	if (ts) {
		ts = ast_tone_zone_sound_unref(ts);
	}

	if (res) {
		ast_log(LOG_WARNING, "Unable to start playtone \'%s\'\n", str);
	}

	return res;
}


/*!
 * \internal
 * \brief Setup the after bridge goto location on the peer.
 * \since 12.0.0
 *
 * \param chan Calling channel for bridge.
 * \param peer Peer channel for bridge.
 * \param opts Dialing option flags.
 * \param opt_args Dialing option argument strings.
 *
 * \return Nothing
 */
static void setup_peer_after_bridge_goto(struct ast_channel *chan, struct ast_channel *peer, struct ast_flags64 *opts, char *opt_args[])
{
	const char *context;
	const char *extension;
	int priority;

	if (ast_test_flag64(opts, OPT_PEER_H)) {
		ast_channel_lock(chan);
		context = ast_strdupa(ast_channel_context(chan));
		ast_channel_unlock(chan);
		ast_bridge_set_after_h(peer, context);
	} else if (ast_test_flag64(opts, OPT_CALLEE_GO_ON)) {
		ast_channel_lock(chan);
		context = ast_strdupa(ast_channel_context(chan));
		extension = ast_strdupa(ast_channel_exten(chan));
		priority = ast_channel_priority(chan);
		ast_channel_unlock(chan);
		ast_bridge_set_after_go_on(peer, context, extension, priority,
			opt_args[OPT_ARG_CALLEE_GO_ON]);
	}
}

static int dial_exec_full(struct ast_channel *chan, const char *data, struct ast_flags64 *peerflags, int *continue_exec)
{
	int res = -1; /* default: error */
	char *rest, *cur; /* scan the list of destinations */
	struct dial_head out_chans = AST_LIST_HEAD_NOLOCK_INIT_VALUE; /* list of destinations */
	struct chanlist *outgoing;
	struct chanlist *tmp;
	struct ast_channel *peer;
	int to; /* timeout */
	struct cause_args num = { chan, 0, 0, 0 };
	int cause;

	struct ast_bridge_config config = { { 0, } };
	struct timeval calldurationlimit = { 0, };
	char *dtmfcalled = NULL, *dtmfcalling = NULL, *dtmf_progress=NULL;
	struct privacy_args pa = {
		.sentringing = 0,
		.privdb_val = 0,
		.status = "INVALIDARGS",
	};
	int sentringing = 0, moh = 0;
	const char *outbound_group = NULL;
	int result = 0;
	char *parse;
	int opermode = 0;
	int delprivintro = 0;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(peers);
		AST_APP_ARG(timeout);
		AST_APP_ARG(options);
		AST_APP_ARG(url);
	);
	struct ast_flags64 opts = { 0, };
	char *opt_args[OPT_ARG_ARRAY_SIZE];
	int fulldial = 0, num_dialed = 0;
	int ignore_cc = 0;
	char device_name[AST_CHANNEL_NAME];
#ifdef GRANDSTREAM_NETWORKS
	char forced_clid_name[AST_MAX_CID_NAME];
	char stored_clid_name[AST_MAX_CID_NAME];
	int record_check_forbiden = 0;
#else
	char forced_clid_name[AST_MAX_EXTENSION];
	char stored_clid_name[AST_MAX_EXTENSION];
#endif
	int force_forwards_only;	/*!< TRUE if force CallerID on call forward only. Legacy behaviour.*/
	/*!
	 * \brief Forced CallerID party information to send.
	 * \note This will not have any malloced strings so do not free it.
	 */
	struct ast_party_id forced_clid;
	/*!
	 * \brief Stored CallerID information if needed.
	 *
	 * \note If OPT_ORIGINAL_CLID set then this is the o option
	 * CallerID.  Otherwise it is the dialplan extension and hint
	 * name.
	 *
	 * \note This will not have any malloced strings so do not free it.
	 */
	struct ast_party_id stored_clid;
#ifdef GRANDSTREAM_NETWORKS
	/*!
	 * \brief Outbound CallerID party information to send when the way to dial is ringboth
	 * \note This will not have any malloced strings so do not free it.
	 */
	struct ast_party_id ringboth_clid;
#endif
	/*!
	 * \brief CallerID party information to store.
	 * \note This will not have any malloced strings so do not free it.
	 */
	struct ast_party_caller caller;
	int max_forwards;
#ifdef GRANDSTREAM_NETWORKS
	int  cc_core_id;    /*Added by yfyang, Grandstream, for ccss feature */
	struct timeval now;
	int dura, flag = 0;
	char  family[80] = {0};
	char  buff[80] = {0};

	const char *outbound_group_ext = NULL;  /*Added by yfyang, for Bug 173103 */
	const char *inbound_group = NULL;       /*Added by yfyang, for Bug 252812 */

	/*! Begin: Fax processing flow optimization*/
	if (chan) {
		ast_channel_set_caller_extension_flag(chan, 1);
		/*Start added by zyxia for Bug 257462 ,2023.4.19*/
		const char *auto_answer = pbx_builtin_getvar_helper(chan, "auto_answer");
		if( ast_channel_name(chan) && strstr(ast_channel_name(chan),"trunk")
			&& !ast_strlen_zero(auto_answer) && !strcasecmp(auto_answer,"yes")){
				pbx_builtin_setvar_helper(chan, "STOP_UNSYNC_PLAY","yes");
				const char *fax_wait_prompt = pbx_builtin_getvar_helper(chan, "FAX_WAIT_PROMPT_PLAY");
				if(!ast_strlen_zero(fax_wait_prompt) && !strcasecmp(fax_wait_prompt,"yes")) {
					ast_stopstream(chan);
					ast_log(LOG_DEBUG, "Stop the prompt  in playbackUnSync_exec\n");
				}
		}
		/*End added by zyxia for Bug 257462 ,2023.4.19*/
	}
	/*! End: Fax processing flow optimization*/
	pbx_builtin_setvar_helper(chan, "_DIAL_EXEC_FULL_CALLER", ast_channel_name(chan));
#endif
	/* Reset all DIAL variables back to blank, to prevent confusion (in case we don't reset all of them). */
	ast_channel_lock(chan);
	ast_channel_stage_snapshot(chan);
	pbx_builtin_setvar_helper(chan, "DIALSTATUS", "");
	pbx_builtin_setvar_helper(chan, "DIALEDPEERNUMBER", "");
	pbx_builtin_setvar_helper(chan, "DIALEDPEERNAME", "");
	pbx_builtin_setvar_helper(chan, "ANSWEREDTIME", "");
	pbx_builtin_setvar_helper(chan, "DIALEDTIME", "");
	ast_channel_stage_snapshot_done(chan);
	max_forwards = ast_max_forwards_get(chan);
	ast_channel_unlock(chan);

	if (max_forwards <= 0) {
		ast_log(LOG_WARNING, "Cannot place outbound call from channel '%s'. Max forwards exceeded\n",
				ast_channel_name(chan));
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", "BUSY");
		return -1;
	}

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "Dial requires an argument (technology/resource)\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", pa.status);
		return -1;
	}

	if (ast_check_hangup_locked(chan)) {
		/*
		 * Caller hung up before we could dial.  If dial is executed
		 * within an AGI then the AGI has likely eaten all queued
		 * frames before executing the dial in DeadAGI mode.  With
		 * the caller hung up and no pending frames from the caller's
		 * read queue, dial would not know that the call has hung up
		 * until a called channel answers.  It is rather annoying to
		 * whoever just answered the non-existent call.
		 *
		 * Dial should not continue execution in DeadAGI mode, hangup
		 * handlers, or the h exten.
		 */
		ast_verb(3, "Caller hung up before dial.\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", "CANCEL");
		return -1;
	}

	char high_frequence_ext[64] = {0};
	const char *aauser   = pbx_builtin_getvar_helper(chan, "AAUSER");
	const char *istrunk  = pbx_builtin_getvar_helper(chan, "FROM_TRUNK");
	const char *urglevel = pbx_builtin_getvar_helper(chan, "URGENCYLEVEL");
	const char *mode     = pbx_builtin_getvar_helper(chan, "STRATEGY");
	const char *conf_invite = pbx_builtin_getvar_helper(chan, "CONF_INVITE");
	const char *invite_user = pbx_builtin_getvar_helper(chan, "INVITE_USER");

	/*caller need inside user, and callee need be trunk */
	if(!urglevel && aauser && !(istrunk && 0 == strcmp(istrunk, "yes"))
		&& (strstr(data,"@trunk_") || strstr(data,"DAHDI/g"))
		&& !(mode && 0 == strcasecmp(mode,"SIMULTA")))
	{
		/* Added by lywei, Bug 216203 */
		if (!ast_strlen_zero(conf_invite) && !strcmp(conf_invite, "yes"))
		{
			snprintf(high_frequence_ext, sizeof(high_frequence_ext)-1, "%s", (char *)invite_user);
		}
		else
		{
			snprintf(high_frequence_ext, sizeof(high_frequence_ext)-1, "%s", (char *)aauser);
		}
		/* End */

		if(1 == ast_get_high_frequence_lockflag(high_frequence_ext))
		{
			ast_log(LOG_WARNING,"user:%s reach the high frequence call,banned!",high_frequence_ext);
			pbx_builtin_setvar_helper(chan, "DIALSTATUS", "CANCEL");
			return -1;
		}

		snprintf(family, sizeof(family), "AAUSER/%s", high_frequence_ext);
		ast_ucm_db_get(family, "HIGHCALL_THRESHOLD", buff, sizeof(buff) -1);
		if(buff[0] != 0 && atoi(buff) != 0)
		{
			int call_threshold = atoi(buff);
			memset(buff,0, sizeof(buff));
			ast_ucm_db_get(family, "HIGHCALL_DURATION", buff, sizeof(buff) -1);
			dura = atoi(buff);
			gettimeofday(&now, NULL);
			ast_high_frequence_queue_search(high_frequence_ext, &now, dura*60, &flag);
			if(1 == flag || 2 == flag)
			{
				if(1 == flag){
					ast_append_high_frequence_ban(high_frequence_ext);/* first reach */
					ast_report_high_freq_call_alarm(high_frequence_ext, call_threshold, dura);

					manager_event(EVENT_FLAG_CALL, "HighFreqLimit", "Extension: %s\r\nFlag: %d\r\n\r\n",
								high_frequence_ext,1);
				}

				ast_log(LOG_WARNING,"user:%s reach the high frequence call,banned!",high_frequence_ext);
				pbx_builtin_setvar_helper(chan, "DIALSTATUS", "CANCEL");
				return -1;
			}
		}
	}

	parse = ast_strdupa(data);

	AST_STANDARD_APP_ARGS(args, parse);

	if (!ast_strlen_zero(args.options) &&
		ast_app_parse_options64(dial_exec_options, &opts, opt_args, args.options)) {
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", pa.status);
		goto done;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (strstr(ast_channel_name(chan), "Local/")) { /* added by zgmo for bug147917 */
		const char *from_callback = pbx_builtin_getvar_helper(chan, "FROM_CALLBACK");
		const char *in_followme = pbx_builtin_getvar_helper(chan, "IN_FOLLOWME");
		const char *fst_caller_name = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");

		if (from_callback && !strcmp(from_callback, "yes")) {
			ast_clear_flag64(&opts, OPT_CALLER_MIXMONITOR);
			ast_log(LOG_DEBUG,
				"chan:%s in callback and is a local channel, clear OPT_CALLER_MIXMONITOR flag",
				ast_channel_name(chan));
		}

		ast_log(LOG_DEBUG, "in_followme:%s, fst_caller:%s",
			ast_strlen_zero(in_followme)? "":in_followme,
			ast_strlen_zero(fst_caller_name)? "":fst_caller_name);

		if (!ast_strlen_zero(in_followme) && !strcmp(in_followme, "yes")) {
			const char *fm_caller_delay = pbx_builtin_getvar_helper(chan, "FM_CALLER_DELAY");

			if (!ast_strlen_zero(fst_caller_name)) {
				struct ast_channel *fst_caller = ast_channel_get_by_name(fst_caller_name);
				if (fst_caller) {
					pbx_builtin_setvar_helper(fst_caller, "FM_DIALED_STR", data);
					ast_channel_unref(fst_caller);
				}
			}

			if (!ast_strlen_zero(fm_caller_delay) && !strcmp(fm_caller_delay, "yes")) {
				int wait_seconds = 3;
				ast_log(LOG_DEBUG,
					"Detect FM outbound through FXO, need wait %ds to ensure FXS-Chan totally hangup!",
					wait_seconds);
				sleep(wait_seconds);
				ast_log(LOG_DEBUG, "Wait done, continue to dial");
			}
		}
	}
#if 0
	else if (strstr(ast_channel_name(chan), "PJSIP/")) { /* added by zgmo for bug183396 */
		const char *user_agent = pbx_builtin_getvar_helper(chan, "USER_AGENT");
		if (!ast_strlen_zero(user_agent) && strstr(user_agent, "Grandstream UCM6")) {
			ast_clear_flag64(&opts, OPT_CALLER_MIXMONITOR);
			ast_log(LOG_DEBUG,
				"chan:%s MAYBE a register trunk, clear OPT_CALLER_MIXMONITOR flag",
				ast_channel_name(chan));
		}
	}
#endif
	/* added by zgmo for sdp record attr */
	{
		const char *cf_reason = pbx_builtin_getvar_helper(chan, "CF_REASON");
		const char *rec_enable = pbx_builtin_getvar_helper(chan,"REC_ENABLE");

		if (!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "13")) {
			ast_indicate(chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
		}

		if (strstr(ast_channel_name(chan), "Local/") && strstr(ast_channel_name(chan), ";2")
			&& strstr(ast_channel_name(chan), "call-forward-")) {
			const char *initial_caller = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");

			if (!ast_strlen_zero(initial_caller) && strcmp(ast_channel_name(chan), initial_caller)) {
			struct ast_channel *initial_chan = ast_channel_get_by_name(initial_caller);
				if (initial_chan) {
					pbx_builtin_setvar_helper(initial_chan, "REC_ENABLE",
						pbx_builtin_getvar_helper(chan, "REC_ENABLE"));
					ast_indicate(initial_chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
					ast_channel_unref(initial_chan);
				}
			}
		}

		if (!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")) {
			process_outside_stop_record_for_dial(chan, &config);
		}
	}
	/* end */
#endif

	if (ast_strlen_zero(args.peers)) {
		ast_log(LOG_WARNING, "Dial requires an argument (technology/resource)\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", pa.status);
		goto done;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_cc_clear_ignore(chan);  /* Added by yfyang, for ccss feature, clear the ignore flag */
#endif
	if (ast_cc_call_init(chan, &ignore_cc)) {
		goto done;
	}

	if (ast_test_flag64(&opts, OPT_SCREEN_NOINTRO) && !ast_strlen_zero(opt_args[OPT_ARG_SCREEN_NOINTRO])) {
		delprivintro = atoi(opt_args[OPT_ARG_SCREEN_NOINTRO]);

		if (delprivintro < 0 || delprivintro > 1) {
			ast_log(LOG_WARNING, "Unknown argument %d specified to n option, ignoring\n", delprivintro);
			delprivintro = 0;
		}
	}

	if (!ast_test_flag64(&opts, OPT_RINGBACK)) {
		opt_args[OPT_ARG_RINGBACK] = NULL;
	}

	if (ast_test_flag64(&opts, OPT_OPERMODE)) {
		opermode = ast_strlen_zero(opt_args[OPT_ARG_OPERMODE]) ? 1 : atoi(opt_args[OPT_ARG_OPERMODE]);
		ast_verb(3, "Setting operator services mode to %d.\n", opermode);
	}

	if (ast_test_flag64(&opts, OPT_DURATION_STOP) && !ast_strlen_zero(opt_args[OPT_ARG_DURATION_STOP])) {
		calldurationlimit.tv_sec = atoi(opt_args[OPT_ARG_DURATION_STOP]);
		if (!calldurationlimit.tv_sec) {
			ast_log(LOG_WARNING, "Dial does not accept S(%s), hanging up.\n", opt_args[OPT_ARG_DURATION_STOP]);
			pbx_builtin_setvar_helper(chan, "DIALSTATUS", pa.status);
			goto done;
		}
		ast_verb(3, "Setting call duration limit to %.3lf seconds.\n", calldurationlimit.tv_sec + calldurationlimit.tv_usec / 1000000.0);
	}

	if (ast_test_flag64(&opts, OPT_SENDDTMF) && !ast_strlen_zero(opt_args[OPT_ARG_SENDDTMF])) {
		dtmf_progress = opt_args[OPT_ARG_SENDDTMF];
		dtmfcalled = strsep(&dtmf_progress, ":");
		dtmfcalling = strsep(&dtmf_progress, ":");
	}

	if (ast_test_flag64(&opts, OPT_DURATION_LIMIT) && !ast_strlen_zero(opt_args[OPT_ARG_DURATION_LIMIT])) {
		if (ast_bridge_timelimit(chan, &config, opt_args[OPT_ARG_DURATION_LIMIT], &calldurationlimit))
			goto done;
	}

	/* Setup the forced CallerID information to send if used. */
	ast_party_id_init(&forced_clid);
	force_forwards_only = 0;
	if (ast_test_flag64(&opts, OPT_FORCECLID)) {
		if (ast_strlen_zero(opt_args[OPT_ARG_FORCECLID])) {
			ast_channel_lock(chan);
			forced_clid.number.str = ast_strdupa(S_OR(ast_channel_macroexten(chan), ast_channel_exten(chan)));
			ast_channel_unlock(chan);
			forced_clid_name[0] = '\0';
			forced_clid.name.str = (char *) get_cid_name(forced_clid_name,
				sizeof(forced_clid_name), chan);
			force_forwards_only = 1;
		} else {
			/* Note: The opt_args[OPT_ARG_FORCECLID] string value is altered here. */
			ast_callerid_parse(opt_args[OPT_ARG_FORCECLID], &forced_clid.name.str,
				&forced_clid.number.str);
		}
		if (!ast_strlen_zero(forced_clid.name.str)) {
			forced_clid.name.valid = 1;
		}
		if (!ast_strlen_zero(forced_clid.number.str)) {
			forced_clid.number.valid = 1;
		}
	}
	if (ast_test_flag64(&opts, OPT_FORCE_CID_TAG)
		&& !ast_strlen_zero(opt_args[OPT_ARG_FORCE_CID_TAG])) {
		forced_clid.tag = opt_args[OPT_ARG_FORCE_CID_TAG];
	}
	forced_clid.number.presentation = AST_PRES_ALLOWED_USER_NUMBER_PASSED_SCREEN;
	if (ast_test_flag64(&opts, OPT_FORCE_CID_PRES)
		&& !ast_strlen_zero(opt_args[OPT_ARG_FORCE_CID_PRES])) {
		int pres;

		pres = ast_parse_caller_presentation(opt_args[OPT_ARG_FORCE_CID_PRES]);
		if (0 <= pres) {
			forced_clid.number.presentation = pres;
		}
	}

	/* Setup the stored CallerID information if needed. */
	ast_party_id_init(&stored_clid);
	if (ast_test_flag64(&opts, OPT_ORIGINAL_CLID)) {
		if (ast_strlen_zero(opt_args[OPT_ARG_ORIGINAL_CLID])) {
			ast_channel_lock(chan);
			ast_party_id_set_init(&stored_clid, &ast_channel_caller(chan)->id);
			if (!ast_strlen_zero(ast_channel_caller(chan)->id.name.str)) {
				stored_clid.name.str = ast_strdupa(ast_channel_caller(chan)->id.name.str);
			}
			if (!ast_strlen_zero(ast_channel_caller(chan)->id.number.str)) {
				stored_clid.number.str = ast_strdupa(ast_channel_caller(chan)->id.number.str);
			}
			if (!ast_strlen_zero(ast_channel_caller(chan)->id.subaddress.str)) {
				stored_clid.subaddress.str = ast_strdupa(ast_channel_caller(chan)->id.subaddress.str);
			}
			if (!ast_strlen_zero(ast_channel_caller(chan)->id.tag)) {
				stored_clid.tag = ast_strdupa(ast_channel_caller(chan)->id.tag);
			}
			ast_channel_unlock(chan);
		} else {
			/* Note: The opt_args[OPT_ARG_ORIGINAL_CLID] string value is altered here. */
			ast_callerid_parse(opt_args[OPT_ARG_ORIGINAL_CLID], &stored_clid.name.str,
				&stored_clid.number.str);
			if (!ast_strlen_zero(stored_clid.name.str)) {
				stored_clid.name.valid = 1;
			}
			if (!ast_strlen_zero(stored_clid.number.str)) {
				stored_clid.number.valid = 1;
			}
		}
	} else {
		/*
		 * In case the new channel has no preset CallerID number by the
		 * channel driver, setup the dialplan extension and hint name.
		 */
		stored_clid_name[0] = '\0';
		stored_clid.name.str = (char *) get_cid_name(stored_clid_name,
			sizeof(stored_clid_name), chan);
		if (ast_strlen_zero(stored_clid.name.str)) {
			stored_clid.name.str = NULL;
		} else {
			stored_clid.name.valid = 1;
		}
		ast_channel_lock(chan);
		stored_clid.number.str = ast_strdupa(S_OR(ast_channel_macroexten(chan), ast_channel_exten(chan)));
		stored_clid.number.valid = 1;

#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, Bug 234390 */
		if (!strcmp(stored_clid.number.str, "s")) {
			char *out_callee = (char *)pbx_builtin_getvar_helper(chan, "THE_REAL_OUT_CALLEE");
			if (!ast_strlen_zero(out_callee)) {
				stored_clid.number.str = ast_strdupa(out_callee);
			}
		}
#endif
		ast_channel_unlock(chan);
	}

	if (ast_test_flag64(&opts, OPT_RESETCDR)) {
		ast_cdr_reset(ast_channel_name(chan), 0);
	}
	if (ast_test_flag64(&opts, OPT_PRIVACY) && ast_strlen_zero(opt_args[OPT_ARG_PRIVACY]))
		opt_args[OPT_ARG_PRIVACY] = ast_strdupa(ast_channel_exten(chan));

	if (ast_test_flag64(&opts, OPT_PRIVACY) || ast_test_flag64(&opts, OPT_SCREENING)) {
		res = setup_privacy_args(&pa, &opts, opt_args, chan);
		if (res <= 0)
			goto out;
		res = -1; /* reset default */
	}

	if (continue_exec)
		*continue_exec = 0;

	/* If a channel group has been specified, get it for use when we create peer channels */

	ast_channel_lock(chan);
	if ((outbound_group = pbx_builtin_getvar_helper(chan, "OUTBOUND_GROUP_ONCE"))) {
		outbound_group = ast_strdupa(outbound_group);
		pbx_builtin_setvar_helper(chan, "OUTBOUND_GROUP_ONCE", NULL);
	} else if ((outbound_group = pbx_builtin_getvar_helper(chan, "OUTBOUND_GROUP"))) {
		outbound_group = ast_strdupa(outbound_group);
	}
#ifdef GRANDSTREAM_NETWORKS
	/*Added by yfyang, for Bug 173103 */
	if ((outbound_group_ext = pbx_builtin_getvar_helper(chan, "OUTBOUND_GROUP_EXT"))) {
		outbound_group_ext = ast_strdupa(outbound_group_ext);
	}

	/*Added by yfyang, for Bug 252812 */
	if ((inbound_group = pbx_builtin_getvar_helper(chan, "INBOUND_GROUP"))) {
		inbound_group = ast_strdupa(inbound_group);
	}

	ast_party_id_init(&ringboth_clid);
	const char* ringboth_callerid= pbx_builtin_getvar_helper(chan, "RINGBOTH_CALLERID");
	if(ringboth_callerid){
		char *str = ast_strdupa(ringboth_callerid);
		ast_callerid_parse(str, &ringboth_clid.name.str,&ringboth_clid.number.str);
		if (!ast_strlen_zero(ringboth_clid.name.str)) {
			ringboth_clid.name.valid = 1;
		}
		if (!ast_strlen_zero(ringboth_clid.number.str)) {
			ringboth_clid.number.valid = 1;
		} else if (ringboth_clid.name.valid == 1) {
			ringboth_clid.number.str = ringboth_clid.name.str;
			ast_log(LOG_NOTICE, "number is null, set name as number -- %s --", ringboth_clid.number.str);
			ringboth_clid.number.valid = 1;
		}
	}
#endif
	ast_channel_unlock(chan);

	/* Set per dial instance flags.  These flags are also passed back to RetryDial. */
	ast_copy_flags64(peerflags, &opts, OPT_DTMF_EXIT | OPT_GO_ON | OPT_ORIGINAL_CLID
		| OPT_CALLER_HANGUP | OPT_IGNORE_FORWARDING | OPT_CANCEL_TIMEOUT
		| OPT_ANNOUNCE | OPT_CALLEE_MACRO | OPT_CALLEE_GOSUB | OPT_FORCECLID);

	/* PREDIAL: Run gosub on the caller's channel */
	if (ast_test_flag64(&opts, OPT_PREDIAL_CALLER)
		&& !ast_strlen_zero(opt_args[OPT_ARG_PREDIAL_CALLER])) {
		ast_replace_subargument_delimiter(opt_args[OPT_ARG_PREDIAL_CALLER]);
		ast_app_exec_sub(NULL, chan, opt_args[OPT_ARG_PREDIAL_CALLER], 0);
	}

	/* loop through the list of dial destinations */
	rest = args.peers;
/*Start added by zyxia for Bug 183156,2021.3.3*/
#ifdef GRANDSTREAM_NETWORKS
	if(strstr(rest,"&"))
		pbx_builtin_setvar_helper(chan,"HAS_SEVERAL_DIAL_DESTINATION","yes");
	else
		pbx_builtin_setvar_helper(chan,"HAS_SEVERAL_DIAL_DESTINATION","no");
#endif
/*End added by zyxia for Bug 183156,2021.3.3*/

	while ((cur = strsep(&rest, "&")) ) {
		struct ast_channel *tc; /* channel for this destination */
		/* Get a technology/resource pair */
		char *number = cur;
		char *tech = strsep(&number, "/");
		size_t tech_len;
		size_t number_len;
		struct ast_stream_topology *topology;

		num_dialed++;
		if (ast_strlen_zero(number)) {
			ast_log(LOG_WARNING, "Dial argument takes format (technology/resource)\n");
			goto out;
		}

		tech_len = strlen(tech) + 1;
		number_len = strlen(number) + 1;
		tmp = ast_calloc(1, sizeof(*tmp) + (2 * tech_len) + number_len);
		if (!tmp) {
			goto out;
		}

		/* Save tech, number, and interface. */
		cur = tmp->stuff;
		strcpy(cur, tech);
		tmp->tech = cur;
		cur += tech_len;
		strcpy(cur, tech);
		cur[tech_len - 1] = '/';
		tmp->interface = cur;
		cur += tech_len;
		strcpy(cur, number);
		tmp->number = cur;

		if (opts.flags) {
			/* Set per outgoing call leg options. */
			ast_copy_flags64(tmp, &opts,
				OPT_CANCEL_ELSEWHERE |
				OPT_CALLEE_TRANSFER | OPT_CALLER_TRANSFER |
				OPT_CALLEE_HANGUP | OPT_CALLER_HANGUP |
				OPT_CALLEE_MONITOR | OPT_CALLER_MONITOR |
				OPT_CALLEE_PARK | OPT_CALLER_PARK |
				OPT_CALLEE_MIXMONITOR | OPT_CALLER_MIXMONITOR |
				OPT_RINGBACK | OPT_MUSICBACK | OPT_FORCECLID | OPT_IGNORE_CONNECTEDLINE |
				OPT_RING_WITH_EARLY_MEDIA);
			ast_set2_flag64(tmp, args.url, DIAL_NOFORWARDHTML);
		}

		/* Request the peer */

		ast_channel_lock(chan);
		/*
		 * Seed the chanlist's connected line information with previously
		 * acquired connected line info from the incoming channel.  The
		 * previously acquired connected line info could have been set
		 * through the CONNECTED_LINE dialplan function.
		 */
		ast_party_connected_line_copy(&tmp->connected, ast_channel_connected(chan));

		topology = ast_stream_topology_clone(ast_channel_get_stream_topology(chan));

		ast_channel_unlock(chan);

#ifdef GRANDSTREAM_NETWORKS
		if (!strcasecmp(tmp->tech, "MulticastRTP") && !strcasecmp(ast_channel_tech(chan)->type, "Local")) {
			RAII_VAR(struct ast_format_cap *, request_caps, ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT), ao2_cleanup);
			RAII_VAR(struct ast_format *, multicast_format, NULL, ao2_cleanup);

			if (ast_channel_tech(chan)->get_multicast_format) {
				multicast_format = ast_channel_tech(chan)->get_multicast_format(chan);
			}

			if (multicast_format) {
				ast_format_cap_append(request_caps, multicast_format, 0);
			} else {
				ast_format_cap_append(request_caps, ast_format_slin, 0);
			}
			tc = ast_request(tmp->tech, request_caps, NULL, chan, tmp->number, &cause);
		} else {
			tc = ast_request_with_stream_topology(tmp->tech, topology, NULL, chan, tmp->number, &cause);
		}
#else
		tc = ast_request_with_stream_topology(tmp->tech, topology, NULL, chan, tmp->number, &cause);
#endif

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 84597, 2017.04.28 */
		if (tc && ast_channel_allow_blind_transfer_callback(chan)) {
			ast_log(LOG_DEBUG, "Peer chan allow blindtransfer callback, set allow too\n");
			ast_channel_allow_blind_transfer_callback_set(tc, 1);
			ast_channel_blind_transfer_timeout_set(tc, ast_channel_blind_transfer_timeout(chan));
		}
		if (tc && ast_channel_allow_blind_transfer_callback(tc) &&
			!ast_channel_allow_blind_transfer_callback(chan)) {
			ast_log(LOG_DEBUG, "Local chan allow blindtransfer callback, set peer allow too\n");
			ast_channel_allow_blind_transfer_callback_set(chan, 1);
			ast_channel_blind_transfer_timeout_set(chan, ast_channel_blind_transfer_timeout(tc));
		}
#endif

		ast_stream_topology_free(topology);

		if (!tc) {
			/* If we can't, just go on to the next call */
			ast_log(LOG_WARNING, "Unable to create channel of type '%s' (cause %d - %s)\n",
				tmp->tech, cause, ast_cause2str(cause));
			handle_cause(cause, &num);
			if (!rest) {
				/* we are on the last destination */
				ast_channel_hangupcause_set(chan, cause);
			}
			if (!ignore_cc && (cause == AST_CAUSE_BUSY || cause == AST_CAUSE_CONGESTION)) {
				if (!ast_cc_callback(chan, tmp->tech, tmp->number, ast_cc_busy_interface)) {
					ast_cc_extension_monitor_add_dialstring(chan, tmp->interface, "");
				}
			}
			chanlist_free(tmp);
#ifdef GRANDSTREAM_NETWORKS
			/*added by wxu@grangstream.cn for followme extension unregister*/
			const char *fm_enable = pbx_builtin_getvar_helper(chan,"FM_ENABLE");
			char action_type[32] = {0};
			ast_cdr_getvar(ast_channel_name(chan), "action_type", action_type,sizeof(action_type));
			if(fm_enable && !strcasecmp(fm_enable,"yes") && (0 != strncmp(action_type, "CONFERENCE", strlen("CONFERENCE")) || 0 != strncmp(action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING")))){
				struct ast_flags flags_op = {
					.flags = 0x70,
				};
				ast_cdr_fork(ast_channel_name(chan),&flags_op);
			}
			/*end*/

			/* added by zgmo for Bug118650 */
			pbx_builtin_setvar_helper(chan, "DIAL_FAIL_CAUSE", ast_cause2str(cause));
			/* end */
#endif
			continue;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (!strncmp(ast_channel_name(chan), "Local", strlen("Local"))) {
			ast_log(LOG_VERBOSE, "Local value !\n");
			ast_channel_value_to_originate_call(tc, 1);
			const char* org_session = pbx_builtin_getvar_helper(chan, "ORG_SESSION");
			char local_caller[128] = {0};
			sscanf(ast_channel_name(chan), "%*[^/]/%127[^@]", local_caller);
			if (org_session && strlen(local_caller)>0){
				char time_stamp[32] = {0};
				sscanf(org_session, "%[^-]-%*s", time_stamp);
				if(strlen(time_stamp)>0){
					char now_session[256] = {0};
					snprintf(now_session, sizeof(now_session), "%s-%s", time_stamp, local_caller);
					ast_log(LOG_NOTICE, "now_session:%s", now_session);
					ast_channel_session_set_value(chan, now_session);
				}
			}
		}
#endif

		ast_channel_get_device_name(tc, device_name, sizeof(device_name));
		if (!ignore_cc) {
			ast_cc_extension_monitor_add_dialstring(chan, tmp->interface, device_name);
		}

		ast_channel_lock_both(tc, chan);
		ast_channel_stage_snapshot(tc);

		pbx_builtin_setvar_helper(tc, "DIALEDPEERNUMBER", tmp->number);
		char dialed_peer_num[256] = {0};
		if (!ast_strlen_zero(tmp->number) && NULL != strstr(tmp->number, "trunk_")){
			char format[32] = {0};

#ifdef GRANDSTREAM_NETWORKS
			/* Added by yfyang, for Bug 238200 */
			if (!strcmp(tmp->tech, "IAX2") && !strncmp(tmp->number, "trunk_", 6)) {
				char tk_id[16] = {0};
				sscanf(tmp->number, "%15[^/]/%31[^;]", tk_id, dialed_peer_num);
			} else {
				snprintf(format, sizeof(format), "%%%lu[^@]@%%*s", sizeof(dialed_peer_num)-1);
				sscanf(tmp->number, format, dialed_peer_num);
			}
#else
			snprintf(format, sizeof(format), "%%%lu[^@]@%%*s", sizeof(dialed_peer_num)-1);
			sscanf(tmp->number, format, dialed_peer_num);
#endif
			const char *real_caller = pbx_builtin_getvar_helper(chan, "REALCALLERIDNUM");
			const char *callerid_name = pbx_builtin_getvar_helper(chan, "AAUSERCIDNAME");
			if (!ast_strlen_zero(dialed_peer_num)){
				ast_channel_reg_calleenum_set(tc, dialed_peer_num);
			}
			if (!ast_strlen_zero(real_caller)){
				ast_channel_reg_callernum_set(tc, real_caller);
			}
			if (!ast_strlen_zero(callerid_name)){
				ast_channel_reg_callername_set(tc, callerid_name);
			}
		}

		/* Setup outgoing SDP to match incoming one */
		if (!AST_LIST_FIRST(&out_chans) && !rest && CAN_EARLY_BRIDGE(peerflags, chan, tc)) {
			/* We are on the only destination. */
			ast_rtp_instance_early_bridge_make_compatible(tc, chan);
		}

		/* Inherit specially named variables from parent channel */
		ast_channel_inherit_variables(chan, tc);
#ifdef GRANDSTREAM_NETWORKS
		const char* tmp_chan_name = ast_channel_name(chan);
		const char* tmp_tc_name = ast_channel_name(tc);
		const char* the_real_callee = pbx_builtin_getvar_helper(chan, "THE_REAL_CALLEE");
		if (!ast_strlen_zero(the_real_callee) && 0 != strcmp(the_real_callee, "s") && 0 != strcmp(the_real_callee, "play_file") && strcasestr(tmp_chan_name, "Local") && strstr(tmp_chan_name, ";2")){
			ast_channel_context_set(tc, "originate-from-internal");
			ast_channel_connect_num_set(tc, the_real_callee);
		}
		const char *chan_out = pbx_builtin_getvar_helper(chan, "OUTBOUND_TRUNK_INDEX");
		if (!ast_strlen_zero(chan_out)
			&& (strcasestr(tmp_chan_name, "call-forward") && strcasestr(tmp_chan_name, "Local") && strstr(tmp_chan_name, ";2"))
			&& (strcasestr(tmp_tc_name, "DAHDI/o") || strcasestr(tmp_tc_name, "trunk_"))){
			char local_1[256] = {0};
			snprintf(local_1, sizeof(local_1), "%s", tmp_chan_name);
			local_1[strlen(tmp_chan_name)-1] = '1';
			ast_log(LOG_NOTICE, "the other local chan is %s", local_1);
			ast_channel_unlock(chan);  /*Add by weidu, for asterisk deadlock*/
			ast_channel_unlock(tc);
			struct ast_channel *Local_chan = ast_channel_get_by_name(local_1);
			ast_channel_lock_both(tc, chan);
			if(Local_chan){
				ast_channel_set_local_outbound_index(Local_chan,chan_out?atoi(chan_out):0);
				ast_channel_unref(Local_chan);
			}
		}
		const char *same_session = pbx_builtin_getvar_helper(chan, "SAME_SESSION");
		if (!ast_strlen_zero(same_session) && strcasestr(tmp_chan_name, "Local") && strstr(tmp_chan_name, ";2")){
			ast_channel_session_set_value(chan, same_session);
		}

		const char *is_atx = pbx_builtin_getvar_helper(chan, "IS_ATX");
		if (!ast_strlen_zero(is_atx) && strcasestr(tmp_chan_name, "Local") && strstr(tmp_chan_name, ";2")){
			char special_action_type[32] = {0};
			ast_cdr_getvar(tmp_chan_name, "special_action_type", special_action_type,sizeof(special_action_type));
			if (!ast_strlen_zero(special_action_type)){
				snprintf(special_action_type+strlen(special_action_type), sizeof(special_action_type)-strlen(special_action_type), "/IS_ATX");
			}else{
				snprintf(special_action_type, sizeof(special_action_type), "IS_ATX");
			}
			ast_cdr_setvar(tmp_chan_name,"special_action_type",special_action_type);
		}

		/*inbound call dial by DISA*/
		char disa_type[32] = {0};
		ast_cdr_getvar(tmp_chan_name, "action_type", disa_type,sizeof(disa_type));
		if (!ast_strlen_zero(disa_type) && !strncmp(disa_type, "DISA",strlen("DISA")) && strstr(tmp_tc_name, "Local") && strstr(tmp_tc_name, ";1")){
			char local_2[256] = {0};
			snprintf(local_2, sizeof(local_2), "%s", tmp_tc_name);
			local_2[strlen(local_2)-1] = '2';
			ast_channel_unlock(chan);  /*Add by weidu, for asterisk deadlock*/
			ast_channel_unlock(tc);
			struct ast_channel *Local_chan = ast_channel_get_by_name(local_2);
			ast_channel_lock_both(tc, chan);
			if(Local_chan){
				ast_channel_session_set_value(Local_chan,ast_channel_session(chan));
				if ((strstr(tmp_chan_name, "DAHDI/o") || strstr(tmp_chan_name, "trunk_") || 0 == strcmp(ast_channel_context(chan), "disa-dial"))){
					ast_channel_set_is_nat_call(Local_chan, ast_channel_get_is_nat_call(chan));
				}
				ast_channel_unref(Local_chan);
			}
		}
		const char *in_trunk_index = pbx_builtin_getvar_helper(chan, "INBOUND_TRUNK_INDEX");
		if (in_trunk_index != NULL && strlen(in_trunk_index)>0
				&& (strstr(tmp_tc_name, "Local") && strstr(tmp_tc_name, ";1"))
				&& (strstr(tmp_chan_name, "DAHDI/o") || strstr(tmp_chan_name, "trunk_") || 0 == strcmp(ast_channel_context(chan), "disa-dial"))){
				char *local_2 = ast_strdup(tmp_tc_name);
				local_2[strlen(local_2)-1]='2';
				ast_cdr_setvar(local_2,"inbound_trunk_index", in_trunk_index);
				ast_free(local_2);
		}
		if (in_trunk_index != NULL && strlen(in_trunk_index)>0
			&& strcasestr(tmp_chan_name, "Local") && strcasestr(tmp_chan_name, "call-forward") && strstr(tmp_chan_name, ";2")){
			ast_cdr_setvar(tmp_chan_name,"inbound_trunk_index", in_trunk_index);
			const char *ori_session = pbx_builtin_getvar_helper(chan, "ORI_SESSION");
			if (ori_session){
				ast_channel_session_set_value(chan, ori_session);
			}
		}
		const char * conf_inv = pbx_builtin_getvar_helper(chan, "CONF_INVITE");
		const char * is_sfu = pbx_builtin_getvar_helper(chan, "IS_SFU");
		if ((NULL != conf_inv && 0 == strcmp(conf_inv, "yes")) && (NULL != is_sfu && 0 == strcmp(is_sfu, "yes"))){
			char tmp_action[256] = {0};
			const char *r_callername = pbx_builtin_getvar_helper(chan, "REG_CALELRNAME");
			snprintf(tmp_action, sizeof(tmp_action), "issfu|%s", S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, ""));
			if (0 == strcmp(tmp_action, "issfu|") && NULL != r_callername){
				snprintf(tmp_action, sizeof(tmp_action), "issfu|%s", r_callername);
			}
			ast_channel_session_set_value(tc,ast_channel_session(chan));
			ast_cdr_setvar(tmp_tc_name,"special_action_type",tmp_action);
		}

		char tmp_sp_action[256] = {0};
		ast_cdr_getvar(tmp_chan_name,"special_action_type",tmp_sp_action, sizeof(tmp_sp_action));
		if (0 == strncmp(tmp_sp_action, "COUNTRY_CODE/FOLLOWME", strlen("COUNTRY_CODE/FOLLOWME"))){
			const char *chan_s = pbx_builtin_getvar_helper(chan, "CHANNEL_SESSION");
			if (!ast_strlen_zero(chan_s)){
				ast_channel_session_set_value(chan, chan_s);
			}
		}
#endif
		ast_channel_datastore_inherit(chan, tc);
		ast_max_forwards_decrement(tc);

		ast_channel_appl_set(tc, "AppDial");
		ast_channel_data_set(tc, "(Outgoing Line)");

		memset(ast_channel_whentohangup(tc), 0, sizeof(*ast_channel_whentohangup(tc)));

		/* Determine CallerID to store in outgoing channel. */
		ast_party_caller_set_init(&caller, ast_channel_caller(tc));

#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug108605 */
		if (strcasestr(ast_channel_name(tc),"PJSIP/trunk") && strlen(ast_channel_connect_num(chan)) > 0) {
			caller.id = stored_clid;
			caller.id.number.valid = 1;
			char action_type[32] = {0};
			char number[128] = {0};
			ast_cdr_getvar(ast_channel_name(chan), "action_type", action_type,sizeof(action_type));
			const char* is_ivr = pbx_builtin_getvar_helper(chan, "IS_IVR");
			if (0 == strncmp(action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING")) && !ast_strlen_zero(is_ivr) && 0 == strcmp(is_ivr, "yes")){
				sscanf(action_type, "%*[^[][%127[^]]", number);
				if ('\0'!=number[0]){
					caller.id.number.str = ast_strdupa(number);
				}
			}else{
				caller.id.number.str = ast_strdupa(ast_channel_connect_num(chan));
			}
			ast_channel_set_caller_event(tc, &caller, NULL);
		}
#endif

		if (ast_test_flag64(peerflags, OPT_ORIGINAL_CLID)) {
			caller.id = stored_clid;
			ast_channel_set_caller_event(tc, &caller, NULL);
			ast_set_flag64(tmp, DIAL_CALLERID_ABSENT);
		} else if (ast_strlen_zero(S_COR(ast_channel_caller(tc)->id.number.valid,
			ast_channel_caller(tc)->id.number.str, NULL))) {
			/*
			 * The new channel has no preset CallerID number by the channel
			 * driver.  Use the dialplan extension and hint name.
			 */
			caller.id = stored_clid;
#ifdef GRANDSTREAM_NETWORKS
			/* Add by chdruan for Bug 45590 - activity call:if callee enable ring both,it will display error during ringing */
			if ((strcasestr(ast_channel_name(tc),"PJSIP/trunk") || strcasestr(ast_channel_name(tc),"DAHDI/")) && strlen(ast_channel_connect_num(chan)) > 0) {
				caller.id.number.valid = 1;
				caller.id.number.str = ast_strdupa(ast_channel_connect_num(chan));
			}
			/*Begin: Added by shfguan@grandstream.cn for bug98005, 2018-01-04*/
			const char *cf_exten = pbx_builtin_getvar_helper(chan, "CF_EXTEN");
			if (strcasestr(ast_channel_name(tc),"DAHDI/o") && cf_exten && strlen(cf_exten) > 0) {
				caller.id.number.valid = 1;
				caller.id.number.str = ast_strdupa(cf_exten);
			}
			/*End: Added by shfguan@grandstream.cn for bug98005, 2018-01-04*/
#endif
			if (!caller.id.name.valid
				&& !ast_strlen_zero(S_COR(ast_channel_connected(chan)->id.name.valid,
					ast_channel_connected(chan)->id.name.str, NULL))) {
				/*
				 * No hint name available.  We have a connected name supplied by
				 * the dialplan we can use instead.
				 */
				caller.id.name.valid = 1;
				caller.id.name = ast_channel_connected(chan)->id.name;
			}
			ast_channel_set_caller_event(tc, &caller, NULL);
			ast_set_flag64(tmp, DIAL_CALLERID_ABSENT);
		} else if (ast_strlen_zero(S_COR(ast_channel_caller(tc)->id.name.valid, ast_channel_caller(tc)->id.name.str,
			NULL))) {
#ifdef GRANDSTREAM_NETWORKS
			if (strcasestr(ast_channel_name(tc),"PJSIP/trunk") && strlen(ast_channel_connect_num(chan)) > 0) {
				caller.id.number.valid = 1;
				char action_type[32] = {0};
				char number[128] = {0};
				ast_cdr_getvar(ast_channel_name(chan), "action_type", action_type,sizeof(action_type));
				const char* is_ivr = pbx_builtin_getvar_helper(chan, "IS_IVR");
				if (0 == strncmp(action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING")) && !ast_strlen_zero(is_ivr) && 0 == strcmp(is_ivr, "yes")){
					sscanf(action_type, "%*[^[][%127[^]]", number);
					if ('\0'!=number[0]){
						caller.id.number.str = ast_strdupa(number);
					}
				}else{
					caller.id.number.str = ast_strdupa(ast_channel_connect_num(chan));
				}
				ast_channel_set_caller_event(tc, &caller, NULL);
			}
#endif
			/* The new channel has no preset CallerID name by the channel driver. */
			if (!ast_strlen_zero(S_COR(ast_channel_connected(chan)->id.name.valid,
				ast_channel_connected(chan)->id.name.str, NULL))) {
#ifdef GRANDSTREAM_NETWORKS
				/* Add by chdruan for Bug 46633 */
				if (!ast_strlen_zero(S_COR(ast_channel_caller(tc)->id.number.valid, ast_channel_caller(tc)->id.number.str, NULL)) &&
						!ast_strlen_zero(S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, NULL))) {
					if (!strcasecmp(ast_channel_caller(tc)->id.number.str,ast_channel_caller(chan)->id.number.str)) {
						/*
						 * We have a connected name supplied by the dialplan we can
						 * use instead.
						 */
						caller.id.name.valid = 1;
						caller.id.name = ast_channel_connected(chan)->id.name;
						ast_channel_set_caller_event(tc, &caller, NULL);
					}
				}
#else
				/*
				 * We have a connected name supplied by the dialplan we can
				 * use instead.
				 */
				caller.id.name.valid = 1;
				caller.id.name = ast_channel_connected(chan)->id.name;
				ast_channel_set_caller_event(tc, &caller, NULL);
#endif
			}
		}

#ifdef GRANDSTREAM_NETWORKS
		const char *r_both = pbx_builtin_getvar_helper(chan, "EN_BOTH");
		const char *in_ringgrp = pbx_builtin_getvar_helper(chan, "IN_RINGGROUP");
		const char *conf_user_inv = pbx_builtin_getvar_helper(chan, "CONF_USER_INVITE");
		const char *realcaller = pbx_builtin_getvar_helper(chan, "REALCALLERIDNUM");
		const char *cf_reason = pbx_builtin_getvar_helper(chan, "CF_REASON");

		if(ast_strlen_zero(ringboth_callerid)){
			const char* from_exten = pbx_builtin_getvar_helper(chan, "FROM-EXTEN");
			if ((strcasestr(ast_channel_name(tc),"PJSIP/trunk") || strcasestr(ast_channel_name(tc),"DAHDI/o"))
				&& (!ast_strlen_zero(r_both) && 0 == strcmp(r_both, "yes"))
				&& (!ast_strlen_zero(is_sfu) && 0 == strcmp(is_sfu, "yes"))
				&& (!ast_strlen_zero(conf_user_inv) && 0 != strcmp(conf_user_inv, "yes"))
				&& !ast_strlen_zero(from_exten)){
				char *str = ast_strdupa(from_exten);
				ast_callerid_parse(str, &ringboth_clid.name.str,&ringboth_clid.number.str);
				if (!ast_strlen_zero(ringboth_clid.name.str)) {
					ringboth_clid.name.valid = 1;
				}
				if (!ast_strlen_zero(ringboth_clid.number.str)) {
					ringboth_clid.number.valid = 1;
				} else if (ringboth_clid.name.valid == 1) {
					ringboth_clid.number.str = ringboth_clid.name.str;
					ast_log(LOG_NOTICE, "number is null, set name as number -- %s --", ringboth_clid.number.str);
					ringboth_clid.number.valid = 1;
				}
			}
		}
#endif
		/* Determine CallerID for outgoing channel to send. */
		if (ast_test_flag64(peerflags, OPT_FORCECLID) && !force_forwards_only) {
			struct ast_party_connected_line connected;

			ast_party_connected_line_set_init(&connected, ast_channel_connected(tc));
			connected.id = forced_clid;
			ast_channel_set_connected_line(tc, &connected, NULL);
#ifdef GRANDSTREAM_NETWORKS
		} else if((strcasestr(ast_channel_name(tc),"PJSIP/trunk") || strcasestr(ast_channel_name(tc),"DAHDI/o"))
			&& ((!ast_strlen_zero(r_both) && 0 == strcmp(r_both, "yes")) || (!ast_strlen_zero(in_ringgrp) && 0 == strcmp(in_ringgrp, "yes")))
			&& (1 == ringboth_clid.name.valid || 1 == ringboth_clid.number.valid)){
			const char *r_callernum = pbx_builtin_getvar_helper(chan, "REG_CALLERNUM");
			const char *r_callername = pbx_builtin_getvar_helper(chan, "REG_CALELRNAME");
			const char *r_calleenum = pbx_builtin_getvar_helper(chan, "REG_CALLEENUM");
			if (!ast_strlen_zero(r_callernum)){
				ast_channel_reg_callernum_set(tc, r_callernum);
			}
			if (!ast_strlen_zero(r_callername)){
				ast_channel_reg_callername_set(tc, r_callername);
			}
			if (!ast_strlen_zero(r_calleenum)){
				ast_channel_reg_calleenum_set(tc, r_calleenum);
			}
			if (1 == ringboth_clid.name.valid || 1 == ringboth_clid.number.valid){
				struct ast_party_connected_line connected;
				ast_party_connected_line_set_init(&connected, ast_channel_connected(tc));
				connected.id = ringboth_clid;
				ast_log(LOG_NOTICE,"ringboth? change the callerid of outbound channel!");
				ast_channel_set_connected_line(tc, &connected, NULL);
			}else {
				ast_connected_line_copy_from_caller(ast_channel_connected(tc), ast_channel_caller(chan));
			}
		} else if (strcasestr(ast_channel_name(tc),"PJSIP/trunk") && !ast_strlen_zero(realcaller) && !ast_strlen_zero(cf_reason) && !strncmp(cf_reason, "13", 2)
		           && strcasestr(ast_channel_name(chan),"PJSIP/trunk")) {
			struct ast_party_connected_line connected1;
			struct ast_party_id bt_clid;

			ast_party_connected_line_set_init(&connected1, ast_channel_connected(tc));
			ast_party_id_init(&bt_clid);
			bt_clid.number.str = ast_strdup(realcaller);
			bt_clid.number.valid = 1;
			connected1.id = bt_clid;

			ast_log(LOG_NOTICE,"blind transfer, set the callerid: %s !", realcaller);
			ast_channel_set_connected_line(tc, &connected1, NULL);
			ast_party_id_free(&bt_clid);
#endif
		} else {
			ast_connected_line_copy_from_caller(ast_channel_connected(tc), ast_channel_caller(chan));
		}

		ast_party_redirecting_copy(ast_channel_redirecting(tc), ast_channel_redirecting(chan));

		ast_channel_dialed(tc)->transit_network_select = ast_channel_dialed(chan)->transit_network_select;

		ast_channel_req_accountcodes(tc, chan, AST_CHANNEL_REQUESTOR_BRIDGE_PEER);
		if (ast_strlen_zero(ast_channel_musicclass(tc))) {
			ast_channel_musicclass_set(tc, ast_channel_musicclass(chan));
		}

		/* Pass ADSI CPE and transfer capability */
		ast_channel_adsicpe_set(tc, ast_channel_adsicpe(chan));
		ast_channel_transfercapability_set(tc, ast_channel_transfercapability(chan));

		/* If we have an outbound group, set this peer channel to it */
		if (outbound_group)
			ast_app_group_set_channel(tc, outbound_group);
		/* If the calling channel has the ANSWERED_ELSEWHERE flag set, inherit it. This is to support local channels */
		if (ast_channel_hangupcause(chan) == AST_CAUSE_ANSWERED_ELSEWHERE)
			ast_channel_hangupcause_set(tc, AST_CAUSE_ANSWERED_ELSEWHERE);

		/* Check if we're forced by configuration */
		if (ast_test_flag64(&opts, OPT_CANCEL_ELSEWHERE))
			 ast_channel_hangupcause_set(tc, AST_CAUSE_ANSWERED_ELSEWHERE);


		/* Inherit context and extension */
		ast_channel_dialcontext_set(tc, ast_strlen_zero(ast_channel_macrocontext(chan)) ? ast_channel_context(chan) : ast_channel_macrocontext(chan));
		if (!ast_strlen_zero(ast_channel_macroexten(chan)))
			ast_channel_exten_set(tc, ast_channel_macroexten(chan));
		else
			ast_channel_exten_set(tc, ast_channel_exten(chan));

		ast_channel_stage_snapshot_done(tc);
#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, Bug 173103 */
		if (outbound_group_ext) {
			ast_app_group_set_channel(chan, outbound_group_ext);
		}

		/*Added by yfyang, for Bug 252812 */
		if (inbound_group) {
			ast_app_group_set_channel(chan, inbound_group);
		}

		/*added by wxu@grandstream.cn for set callback session*/
		if (strncasecmp(ast_channel_name(tc), "Local",5)){
			const char *from_callback = pbx_builtin_getvar_helper(chan,"FROM_CALLBACK");

			ast_log(LOG_VERBOSE,"tc->name:%s",ast_channel_name(tc));
			if(from_callback){
				ast_channel_session_set_value(tc,ast_channel_session(chan));
			}
		}
		/*end*/
		else if (!strcasecmp(S_OR(pbx_builtin_getvar_helper(tc, "IS_SFU"), ""), "yes")) {
			char sfu_call_id[64] = {0};

			snprintf(sfu_call_id, sizeof(sfu_call_id), "%s-%ld", S_OR(ast_channel_caller_num(tc), "NULL"), ast_tvnow().tv_sec);
			pbx_builtin_setvar_helper(tc, "SFU_CALL_ID", sfu_call_id);
		}
		const char *redirect = pbx_builtin_getvar_helper(chan, "REDIRECT");
		if (redirect && 0 == strcmp(redirect, "yes")){
			ast_cdr_setvar(ast_channel_name(chan), "action_type", "TRANSFER");
			if (NULL == strcasestr(ast_channel_name(chan), "trunk_")){
				ast_channel_reg_callernum_set(chan, "");
				ast_channel_reg_callername_set(chan, "");
			}
			if (NULL == strcasestr(ast_channel_name(tc), "trunk_")){
				ast_channel_reg_calleenum_set(tc, "");
			}
		}
#endif

		/* Save the original channel name to detect call pickup masquerading in. */
		tmp->orig_chan_name = ast_strdup(ast_channel_name(tc));

		ast_channel_unlock(tc);
		ast_channel_unlock(chan);

#ifdef GRANDSTREAM_NETWORKS
		/*add by ychp to delete useless record file name for bug 109363/bug 109539*/
		const char *rec_enable = pbx_builtin_getvar_helper(chan,"REC_ENABLE");
		const char *mixautorec = pbx_builtin_getvar_helper(chan,"MIXAUTORECORDING");
		const char *queue_timeout = pbx_builtin_getvar_helper(chan,"QUEUE-TIMEOUT");
		const char *c_option = pbx_builtin_getvar_helper(chan,"CALL_OPTIONS");
		const char *ring_both = pbx_builtin_getvar_helper(chan,"EN_BOTH");
		if (!(
			(!ast_strlen_zero(rec_enable) && 0 == strcmp(rec_enable, "yes"))
			|| (!ast_strlen_zero(mixautorec) && 0 == strcmp(mixautorec, "ON") && !ast_strlen_zero(queue_timeout) && 0 == strcmp(queue_timeout, "yes"))
			|| (!ast_strlen_zero(mixautorec) && 0 == strcmp(mixautorec, "ON") && !ast_strlen_zero(ring_both) && 0 == strcmp(ring_both, "yes"))
			)
			&& (NULL != c_option && NULL == strstr(c_option, "dialog-being-recorded"))){
			char action_type[32] = {0};
			ast_cdr_getvar(ast_channel_name(chan), "action_type", action_type,sizeof(action_type));
			if (strcmp(action_type, "EMERGENCYCALL"))
				ast_cdr_setvar(ast_channel_name(chan),"record_files","");
		}

		pbx_builtin_setvar_helper(tc, "REC_ENABLE", pbx_builtin_getvar_helper(chan, "REC_ENABLE"));  /* added by zgmo for sdp record attr */
#endif

		/* Put channel in the list of outgoing thingies. */
		tmp->chan = tc;
		AST_LIST_INSERT_TAIL(&out_chans, tmp, node);
	}

#ifdef GRANDSTREAM_NETWORKS
	/*added by wxu@grandstream.cn for bug 43554*/
	const char *group_callee = pbx_builtin_getvar_helper(chan,"GROUP_CALLEE");
	const char *group_rec = pbx_builtin_getvar_helper(chan,"RINGGROUP_REC");
	const char *mixfilename = pbx_builtin_getvar_helper(chan,"MIXFILENAME");
	const char *mixmon_format = pbx_builtin_getvar_helper(chan,"MIXMON_FORMAT");
	char *p_group_callee = NULL;
	if(group_callee){
		p_group_callee = strstr(group_callee,"&");
	}
	/*end*/
#endif

	/*
	 * PREDIAL: Run gosub on all of the callee channels
	 *
	 * We run the callee predial before ast_call() in case the user
	 * wishes to do something on the newly created channels before
	 * the channel does anything important.
	 *
	 * Inside the target gosub we will be able to do something with
	 * the newly created channel name ie: now the calling channel
	 * can know what channel will be used to call the destination
	 * ex: now we will know that SIP/abc-123 is calling SIP/def-124
	 */
	if (ast_test_flag64(&opts, OPT_PREDIAL_CALLEE)
		&& !ast_strlen_zero(opt_args[OPT_ARG_PREDIAL_CALLEE])
		&& !AST_LIST_EMPTY(&out_chans)) {
		const char *predial_callee;

		ast_replace_subargument_delimiter(opt_args[OPT_ARG_PREDIAL_CALLEE]);
		predial_callee = ast_app_expand_sub_args(chan, opt_args[OPT_ARG_PREDIAL_CALLEE]);
		if (predial_callee) {
			ast_autoservice_start(chan);
			AST_LIST_TRAVERSE(&out_chans, tmp, node) {
#ifdef GRANDSTREAM_NETWORKS
				pbx_builtin_setvar_helper(tmp->chan, "ORIG_CID_NUMBER",
					S_COR(ast_channel_caller(tmp->chan)->ani.number.valid, ast_channel_caller(tmp->chan)->ani.number.str, ""));
#endif
				ast_pre_call(tmp->chan, predial_callee);
			}
			ast_autoservice_stop(chan);
			ast_free((char *) predial_callee);
		}
	}

	/* Start all outgoing calls */
	AST_LIST_TRAVERSE_SAFE_BEGIN(&out_chans, tmp, node) {
		res = ast_call(tmp->chan, tmp->number, 0); /* Place the call, but don't wait on the answer */
		ast_channel_lock(chan);

		/* check the results of ast_call */
		if (res) {
			/* Again, keep going even if there's an error */
			ast_debug(1, "ast call on peer returned %d\n", res);
			ast_verb(3, "Couldn't call %s\n", tmp->interface);
			if (ast_channel_hangupcause(tmp->chan)) {
				ast_channel_hangupcause_set(chan, ast_channel_hangupcause(tmp->chan));
			}
			ast_channel_unlock(chan);
			ast_cc_call_failed(chan, tmp->chan, tmp->interface);
			ast_hangup(tmp->chan);
			tmp->chan = NULL;
			AST_LIST_REMOVE_CURRENT(node);
			chanlist_free(tmp);
			continue;
		}

		ast_channel_publish_dial(chan, tmp->chan, tmp->number, NULL);
		ast_channel_unlock(chan);

#ifdef GRANDSTREAM_NETWORKS
		/*modified by yjwu for bug 54496,43554*/
		if(p_group_callee){
			char callee_num[64] = {0};
			sscanf(tmp->number,"%63[0-9*#]%*s",callee_num);

			if( !strncmp(tmp->interface,"DAHDI/",strlen("DAHDI/")) && strlen(callee_num) > 0 ){
				char varname_tmp[256] = {0};
				char family[64] = {0};
				char key[64] = {0};

				snprintf(family, sizeof(family), "DAHDI/%s", callee_num);
				snprintf(key, sizeof(key), "FXS_EXTENSION");
				ast_ucm_db_get(family, key, varname_tmp, sizeof(varname_tmp) - 1);
				snprintf(callee_num, sizeof(callee_num), "%s", varname_tmp);
			} else if ( !strncmp(tmp->interface,"DAHDI/",strlen("DAHDI/"))) {
				sscanf(tmp->number,"%*[^/]/%63[0-9*#]%*s",callee_num);
			}

			//ast_cdr_setvar(ast_channel_name(chan),"app_callee",callee_num);
			ast_cdr_setvar(ast_channel_name(tmp->chan),"__app_callee", callee_num);

			if(group_rec && !strcmp(group_rec,"yes") && mixfilename && mixmon_format){
				char record_file_name[512] = {0};
				snprintf(record_file_name,sizeof(record_file_name)-1,"%s.%s@",mixfilename,mixmon_format);
				ast_cdr_setvar(ast_channel_name(chan),"record_files",record_file_name);
			}
		}
		/*end*/
		/*for bug 118698 by ychpang*/
		const char *is_order = pbx_builtin_getvar_helper(chan,"STRATEGY");
		if (NULL !=is_order && 0 == strncmp(is_order, "ORDER", strlen("ORDER"))){
			const char *cur_member = pbx_builtin_getvar_helper(chan,"CUR_MEMBER");
			const char *members_ldap = pbx_builtin_getvar_helper(chan,"MEMBERS_LDAP");
			if (NULL != cur_member && NULL != members_ldap && strstr(members_ldap, cur_member)){
					ast_cdr_setvar(ast_channel_name(chan),"app_callee",cur_member);
			}
		}
		const char *tmp_tc_name = ast_channel_name(tmp->chan);
		const char *tmp_chan_name = ast_channel_name(chan);
		const char *schdule_sfu = pbx_builtin_getvar_helper(chan, "CONF_SCHEDULE_INVITE");
		if (tmp_tc_name && tmp_chan_name && !ast_strlen_zero(schdule_sfu) && !strcmp(schdule_sfu, "yes")){
			char tmp_action[256] = {0};
			const char* sfu_sub1 = pbx_builtin_getvar_helper(chan, "CONF_SCHEDULE_SUBJECT");
			if (!ast_strlen_zero(sfu_sub1)){
				snprintf(tmp_action, sizeof(tmp_action), "Scheduled Conference|%s", sfu_sub1);
				ast_cdr_setvar(tmp_tc_name,"special_action_type",tmp_action);
                        }
			ast_channel_session_set_value(tmp->chan,ast_channel_session(chan));
			char chan_action_type[32] = {0};
			ast_cdr_getvar(tmp_chan_name, "action_type", chan_action_type,sizeof(chan_action_type));
			ast_log(LOG_DEBUG, "action_type:%s", chan_action_type);
			if (!ast_strlen_zero(chan_action_type)){
				ast_cdr_setvar(tmp_tc_name, "action_type",chan_action_type);
			}


			char chan_app_caller[256] = {0};
			ast_cdr_getvar(tmp_chan_name, "app_caller", chan_app_caller,sizeof(chan_app_caller));
			ast_log(LOG_DEBUG, "app_callee:%s", chan_app_caller);
			if (!ast_strlen_zero(chan_app_caller)){
				ast_cdr_setvar(tmp_tc_name, "app_callee", chan_app_caller);
			}

			char chan_app_callee[256] = {0};
			ast_cdr_getvar(tmp_chan_name, "app_callee", chan_app_callee,sizeof(chan_app_callee));
			ast_log(LOG_DEBUG, "app_caller:%s", chan_app_callee);
			if (!ast_strlen_zero(chan_app_callee)){
				ast_cdr_setvar(tmp_tc_name, "app_caller", chan_app_callee);
			}

			char chan_outbound_trunk_index[32] = {0};
			ast_cdr_getvar(tmp_chan_name, "outbound_trunk_index", chan_outbound_trunk_index,sizeof(chan_outbound_trunk_index));
			ast_log(LOG_DEBUG, "outbound_trunk_index:%s", chan_outbound_trunk_index);
			if (!ast_strlen_zero(chan_outbound_trunk_index)){
				ast_cdr_setvar(tmp_tc_name, "outbound_trunk_index", chan_outbound_trunk_index);
			}
		}
		/*end*/
#endif

		ast_verb(3, "Called %s\n", tmp->interface);
		ast_set_flag64(tmp, DIAL_STILLGOING);

		/* If this line is up, don't try anybody else */
		if (ast_channel_state(tmp->chan) == AST_STATE_UP) {
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	if (ast_strlen_zero(args.timeout)) {
		to = -1;
	} else {
		to = atoi(args.timeout);
		if (to > 0)
			to *= 1000;
		else {
			ast_log(LOG_WARNING, "Invalid timeout specified: '%s'. Setting timeout to infinite\n", args.timeout);
			to = -1;
		}
	}

	outgoing = AST_LIST_FIRST(&out_chans);
	if (!outgoing) {
		strcpy(pa.status, "CHANUNAVAIL");
		if (fulldial == num_dialed) {
			res = -1;
			goto out;
		}
	} else {
		/* Our status will at least be NOANSWER */
		strcpy(pa.status, "NOANSWER");
		if (ast_test_flag64(outgoing, OPT_MUSICBACK)) {
			moh = 1;
#ifdef GRANDSTREAM_NETWORKS
			/* added by zgmo for Bug180799 */
			if (strstr(ast_channel_name(chan), "DAHDI/o")) {
				ast_answer(chan);
			}
			/* end */
#endif
			if (!ast_strlen_zero(opt_args[OPT_ARG_MUSICBACK])) {
				char *original_moh = ast_strdupa(ast_channel_musicclass(chan));
				ast_channel_musicclass_set(chan, opt_args[OPT_ARG_MUSICBACK]);
				ast_moh_start(chan, opt_args[OPT_ARG_MUSICBACK], NULL);
				ast_channel_musicclass_set(chan, original_moh);
			} else {
				ast_moh_start(chan, NULL, NULL);
			}
			ast_indicate(chan, AST_CONTROL_PROGRESS);
		} else if (ast_test_flag64(outgoing, OPT_RINGBACK) || ast_test_flag64(outgoing, OPT_RING_WITH_EARLY_MEDIA)) {
			if (!ast_strlen_zero(opt_args[OPT_ARG_RINGBACK])) {
				if (dial_handle_playtones(chan, opt_args[OPT_ARG_RINGBACK])){
					ast_indicate(chan, AST_CONTROL_RINGING);
					sentringing++;
				} else {
					ast_indicate(chan, AST_CONTROL_PROGRESS);
				}
			} else {
				ast_indicate(chan, AST_CONTROL_RINGING);
				sentringing++;
			}
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for blind transfer callback, 2017.03.16 */
	set_timeout_for_blind_transfer_callback(chan, &to);

	ast_channel_set_peer_is_nat_call(chan, 0);
#endif

	peer = wait_for_answer(chan, &out_chans, &to, peerflags, opt_args, &pa, &num, &result,
		dtmf_progress, ignore_cc, &forced_clid, &stored_clid);

#ifdef GRANDSTREAM_NETWORKS
	cc_core_id = ast_cc_get_current_core_id(chan);
	if (-1 != cc_core_id) {
		if (0 == strcmp(pa.status, "CANCEL")) {
			ast_cc_failed(cc_core_id, "The call of '%s' has been cancelled", ast_channel_name(chan));
		}
	}
#endif

	if (!peer) {
		if (result) {
			res = result;
		} else if (to) { /* Musta gotten hung up */
			res = -1;
		} else { /* Nobody answered, next please? */
			res = 0;
		}
	} else {
		const char *number;
		const char *name;
		int dial_end_raised = 0;
		int cause = -1;
#ifdef GRANDSTREAM_NETWORKS
		const char *pickup_caller = pbx_builtin_getvar_helper(peer, "PICKUP_CALLER");
#endif

		if (ast_test_flag64(&opts, OPT_CALLER_ANSWER)) {
			ast_answer(chan);
		}

#ifdef GRANDSTREAM_NETWORKS
		/*Start: chzhdeng  [Bug 37676 ] add */
		pbx_builtin_setvar_helper(chan, "CALLED_REALHOOK", "yes");
		/*End: chzhdeng  [Bug 37676 ] add */
		if (-1 != cc_core_id) {
			if (!strncmp(ast_channel_name(peer), "DAHDI/", strlen("DAHDI/"))) {
				ast_cc_failed(cc_core_id, "The call of '%s' has been answered", ast_channel_name(chan));
			}
		}

		pbx_builtin_setvar_helper(peer, "DIAL_EXEC_FULL_CALLER", ast_channel_name(chan));

		if (!ast_strlen_zero(pickup_caller) && !strcmp(pickup_caller, "yes")) {
			ast_pickup_check_record(chan, peer);
		}

		if (strstr(ast_channel_name(chan), "Local/") && strstr(ast_channel_name(chan), ";2")) {
			const char *do_forward = pbx_builtin_getvar_helper(chan, "DO_FORWARD");
			const char *ringboth = pbx_builtin_getvar_helper(chan, "RINGBOTH");
			const char *seamless_transfer = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER");

			if (ast_test_flag(ast_channel_flags(chan), AST_FLAG_MOH)) {
				const char *from_ivr = pbx_builtin_getvar_helper(chan, "FROM_IVR");
				const char *from_disa = pbx_builtin_getvar_helper(chan, "FROM_DISA");

				if (!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes")
					&& !ast_strlen_zero(from_disa) && !strcmp(from_disa, "yes")) {
					ast_moh_stop(chan);
					ast_log(LOG_DEBUG,
						"IVR->DISA| %s MOH is ongoing, now stop it!", ast_channel_name(chan));
				}
			}

			/* added by zgmo for Bug144632 */
			if (!ast_strlen_zero(do_forward)) {
				const char *fst_caller = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");
				if (!ast_strlen_zero(fst_caller)) {
					char cdr_record[512];
					ast_cdr_getvar(ast_channel_name(chan), "RECORD_FILES", cdr_record, sizeof(cdr_record));

					if (!ast_strlen_zero(cdr_record)) {
						ast_cdr_setvar(fst_caller, "RECORD_FILES", cdr_record);
						ast_log(LOG_DEBUG, "caller:%s cdr-record:%s set to ==> %s's cdr", ast_channel_name(chan), cdr_record, fst_caller);
					}
				}
			}
			/* end */

			/* added by zgmo for Bug178459 */
			if (!ast_strlen_zero(pickup_caller) && !strcmp(pickup_caller, "yes")) {
				ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES",
					pbx_builtin_getvar_helper(peer, "RECORDING_FILE_CDR"));
				ast_log(LOG_DEBUG, "Set Chan:%s CDR var:RECORD_FILES=%s", ast_channel_name(chan), pbx_builtin_getvar_helper(peer, "RECORDING_FILE_CDR"));
			}
			/* end */

			/* added by zgmo for Bug191175 */
			if (!ast_strlen_zero(seamless_transfer) && !strcmp(seamless_transfer, "yes")
				&& !ast_strlen_zero(ringboth) && !strcmp(ringboth, "yes")) {
				ast_clear_flag64(peerflags, OPT_CALLEE_MIXMONITOR);
				ast_log(LOG_DEBUG, "In this case disable %s mixmonitor feature",
					ast_channel_name(peer));
			}
			/* end */
		}

		/*added by zgmo for Bug171856 */
		if (!ast_strlen_zero(pbx_builtin_getvar_helper(peer, "DO_FORWARD"))) {
			if (strstr(ast_channel_name(peer), "Local/") && strstr(ast_channel_name(peer), ";1")) {
				ast_clear_flag64(peerflags, OPT_CALLEE_MIXMONITOR);
				ast_log(LOG_DEBUG, "In this case disable %s mixmonitor feature",
					ast_channel_name(peer));
			}
		}
		/* end */
#if 0
		/* added by zgmo for Bug149059 */
		if(strstr(ast_channel_name(peer), "trunk_") && ast_test_flag64(peerflags, OPT_CALLEE_MIXMONITOR)) {
			ast_clear_flag64(peerflags, OPT_CALLEE_MIXMONITOR);
			ast_log(LOG_DEBUG, "peer(%s) is a trunk chan, disable mixmonitor feature",
				ast_channel_name(peer));
		}
		/* end */

		/* added by zgmo for Bug183396 */
		{
			const char *user_agent = pbx_builtin_getvar_helper(peer, "USER_AGENT");
			if (!ast_strlen_zero(user_agent) && strstr(user_agent, "Grandstream UCM")) {
				ast_clear_flag64(peerflags, OPT_CALLEE_MIXMONITOR);
				ast_log(LOG_DEBUG,
					"peer(%s) as CALLEE maybe a register trunk, disable mixmonitor feature",
					ast_channel_name(peer));
			}
		}
		/* end */
#endif
		{
			const char *in_ringgroup = pbx_builtin_getvar_helper(chan, "IN_RINGGROUP");

			if (!ast_strlen_zero(in_ringgroup) && !strcmp(in_ringgroup, "yes")) {
				const char *call_filename = pbx_builtin_getvar_helper(chan, "CALLFILENAME");
				const char *pickupee = pbx_builtin_getvar_helper(peer, "PICKUPEE");

				char cdr_record_files[1024];

				cdr_record_files[0] = '\0';
				ast_cdr_getvar(ast_channel_name(chan), "RECORD_FILES", cdr_record_files, sizeof(cdr_record_files));

				if (!ast_strlen_zero(call_filename) && !ast_strlen_zero(cdr_record_files)
					&& ast_channel_caller(peer)->ani.number.valid && ast_strlen_zero(pickupee)) {
					char *peer_num = ast_channel_caller(peer)->ani.number.str;
					char *call_filename_new = NULL;
					char *cdr_record_files_new = NULL;

					call_filename_new = ast_malloc(strlen(call_filename) + strlen(peer_num) + 16);
					cdr_record_files_new = ast_malloc(strlen(cdr_record_files) + strlen(peer_num) + 16);

					if(call_filename_new && cdr_record_files_new) {
						pending_ringgroup_callee_info(call_filename_new, call_filename, peer_num);
						pending_ringgroup_callee_info(cdr_record_files_new, (const char*)cdr_record_files, peer_num);

						pbx_builtin_setvar_helper(chan, "CALLFILENAME", call_filename_new);
						pbx_builtin_setvar_helper(peer, "CALLFILENAME", call_filename_new);
						pbx_builtin_setvar_helper(chan, "_MIXFILENAME", call_filename_new);
						pbx_builtin_setvar_helper(peer, "_MIXFILENAME", call_filename_new);

						ast_cdr_setvar(ast_channel_name(chan), "special_action_type", "overwrt_record");
						ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES", cdr_record_files_new);
					}

					if (call_filename_new) {
						ast_free(call_filename_new);
					}

					if (cdr_record_files_new) {
						ast_free(cdr_record_files_new);
					}
				}
			}
		}
#endif

		/* Ah ha!  Someone answered within the desired timeframe.  Of course after this
		   we will always return with -1 so that it is hung up properly after the
		   conversation.  */

		if (ast_test_flag64(&opts, OPT_HANGUPCAUSE)
			&& !ast_strlen_zero(opt_args[OPT_ARG_HANGUPCAUSE])) {
			cause = ast_str2cause(opt_args[OPT_ARG_HANGUPCAUSE]);
			if (cause <= 0) {
				if (!strcasecmp(opt_args[OPT_ARG_HANGUPCAUSE], "NONE")) {
					cause = 0;
				} else if (sscanf(opt_args[OPT_ARG_HANGUPCAUSE], "%30d", &cause) != 1
					|| cause < 0) {
					ast_log(LOG_WARNING, "Invalid cause given to Dial(...Q(<cause>)): \"%s\"\n",
						opt_args[OPT_ARG_HANGUPCAUSE]);
					cause = -1;
				}
			}
		}
		hanguptree(&out_chans, peer, cause >= 0 ? cause : AST_CAUSE_ANSWERED_ELSEWHERE);

		/* If appropriate, log that we have a destination channel and set the answer time */

		ast_channel_lock(peer);
		name = ast_strdupa(ast_channel_name(peer));

		number = pbx_builtin_getvar_helper(peer, "DIALEDPEERNUMBER");
		if (ast_strlen_zero(number)) {
			number = NULL;
		} else {
			number = ast_strdupa(number);
		}
		ast_channel_unlock(peer);

		ast_channel_lock(chan);
		ast_channel_stage_snapshot(chan);

		strcpy(pa.status, "ANSWER");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", pa.status);

		pbx_builtin_setvar_helper(chan, "DIALEDPEERNAME", name);
		pbx_builtin_setvar_helper(chan, "DIALEDPEERNUMBER", number);

		ast_channel_stage_snapshot_done(chan);
		ast_channel_unlock(chan);

		if (!ast_strlen_zero(args.url) && ast_channel_supports_html(peer) ) {
			ast_debug(1, "app_dial: sendurl=%s.\n", args.url);
			ast_channel_sendurl( peer, args.url );
		}
		if ( (ast_test_flag64(&opts, OPT_PRIVACY) || ast_test_flag64(&opts, OPT_SCREENING)) && pa.privdb_val == AST_PRIVACY_UNKNOWN) {
			if (do_privacy(chan, peer, &opts, opt_args, &pa)) {
				ast_channel_publish_dial(chan, peer, NULL, pa.status);
				/* hang up on the callee -- he didn't want to talk anyway! */
				ast_autoservice_chan_hangup_peer(chan, peer);
				res = 0;
				goto out;
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		/* added by zgmo for Bug166199 */
		{
			const char *a_xfer_state = pbx_builtin_getvar_helper(chan, "A_XFER_STATE");
			if (!ast_strlen_zero(a_xfer_state)
				&& !strcmp(a_xfer_state, "blond")) {
				const char *transferee_chan_name;
				transferee_chan_name = pbx_builtin_getvar_helper(chan, "A_XFER_XFEREE_CHAN");
				if (!ast_strlen_zero(transferee_chan_name)) {
					struct ast_channel *transferee_chan = ast_channel_get_by_name(transferee_chan_name);
					if (transferee_chan) {
						char *transferee_cid_num = NULL;
						char *target_cid_num = NULL;
						const char *rec_enable;
						const char *leave_trunk;
						const char *callee_num;

						leave_trunk = pbx_builtin_getvar_helper(chan, "LEAVE_TRUNK");
						callee_num = pbx_builtin_getvar_helper(chan, "CALLEE_NUM");

						if (!ast_strlen_zero(leave_trunk) && !strcmp(leave_trunk, "yes")
							&& !ast_strlen_zero(callee_num)) {
							target_cid_num = (char *)callee_num;
						} else {
							if (!ast_strlen_zero(ast_channel_caller(peer)->ani.number.str)) {
								target_cid_num = ast_channel_caller(peer)->ani.number.str;
							}
						}

						if (!ast_strlen_zero(ast_channel_caller(transferee_chan)->ani.number.str)) {
							transferee_cid_num = ast_channel_caller(transferee_chan)->ani.number.str;
						}

						pbx_builtin_setvar_helper(peer, "__CF_REASON", "14");
						pbx_builtin_setvar_helper(peer, "__CF_TRANSFEREE", transferee_cid_num);
						pbx_builtin_setvar_helper(peer, "__CF_TRANSFER_TARGET", target_cid_num);
						pbx_builtin_setvar_helper(peer, "TRUNK_INDEX", pbx_builtin_getvar_helper(transferee_chan, "TRUNK_INDEX"));
						pbx_builtin_setvar_helper(peer, "IVR_REC", pbx_builtin_getvar_helper(transferee_chan, "IVR_REC"));

						if (strstr(ast_channel_name(peer), "/trunk")) {
							pbx_builtin_setvar_helper(peer, "OUTBOUND_RT_INDEX",
								pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_INDEX"));
						}

						if (ast_test_flag64(&opts, OPT_ANNOUNCE)
							&& !ast_strlen_zero(opt_args[OPT_ARG_ANNOUNCE])
							&& !strcmp(opt_args[OPT_ARG_ANNOUNCE], "dialog-being-recorded")) {
							ast_clear_flag64(&opts, OPT_ANNOUNCE);
						}

						if (ast_test_flag64(&opts, OPT_CALLEE_MACRO)
							&& !ast_strlen_zero(opt_args[OPT_ARG_CALLEE_MACRO])
							&& !strcmp(opt_args[OPT_ARG_CALLEE_MACRO], "record-check,StartMonitor")) {
							ast_clear_flag64(&opts, OPT_CALLEE_MACRO);
						}

						ast_dial_check_record(transferee_chan, peer);

						rec_enable = pbx_builtin_getvar_helper(peer, "REC_ENABLE");
						if (!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")) {
							char record_file_name[512] = {0};
							const char *mixfilename = pbx_builtin_getvar_helper(peer, "MIXFILENAME");
							const char *mixmon_format = pbx_builtin_getvar_helper(peer, "MIXMON_FORMAT");
							snprintf(record_file_name, sizeof(record_file_name)-1,
								"%s.%s@", mixfilename, mixmon_format);
							ast_cdr_setvar(ast_channel_name(chan), "record_files", record_file_name);
						} else {
							ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES", "");
						}

						/* added by zgmo for sdp record attr */
						pbx_builtin_setvar_helper(transferee_chan, "REC_ENABLE", rec_enable);

						ast_indicate(transferee_chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
						ast_indicate(peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
						/* end */

						ast_channel_unref(transferee_chan);
					}
				}
			}
		}
		/* add end */

		/* added by zgmo for bug133681 */
		if (!ast_strlen_zero(pbx_builtin_getvar_helper(chan, "CF_REASON"))
			&& !strcmp(pbx_builtin_getvar_helper(chan, "CF_REASON"), "15")) {
			/*
			 * check whether this dial is caused by attended transfer, if so, the caller channel
			 * is setted previously by adding a variable CF_REASON = 15 in refer_attended_task() when attended
			 * transferer hung up before transfer target pick up.
			 */
			const char *rec_enable;
			const char *recording_chan;
			const char *leave_trunk;
			const char *callee_num;
			const char *mixfilename;
			const char *a_xferer_hangup;
			const char *a_xfereree_name;

			char macro_args[128] = {0};
			struct ast_app *macro_app = pbx_findapp("Macro");

			ast_log(LOG_DEBUG, "Detect attented transfer by endpoint");

			pbx_builtin_setvar_helper(peer, "CF_REASON", "15");
			pbx_builtin_setvar_helper(peer, "TRUNK_INDEX", pbx_builtin_getvar_helper(chan, "TRUNK_INDEX"));
			pbx_builtin_setvar_helper(peer, "INBOUND_RT_INDEX", pbx_builtin_getvar_helper(chan, "INBOUND_RT_INDEX"));
			pbx_builtin_setvar_helper(peer, "OUTBOUND_RT_INDEX", pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_INDEX"));

			leave_trunk = pbx_builtin_getvar_helper(chan, "LEAVE_TRUNK");
			callee_num = pbx_builtin_getvar_helper(chan, "CALLEE_NUM");

			if (!ast_strlen_zero(leave_trunk) && !strcmp(leave_trunk, "yes")
				&& !ast_strlen_zero(callee_num)) {
				snprintf(macro_args, 128, "record-check,%s-%s",
					ast_channel_caller(chan)->ani.number.str, callee_num);
				ast_cdr_setvar(ast_channel_name(chan), "app_callee", callee_num);
			} else {
				snprintf(macro_args, 128, "record-check,%s-%s",
					ast_channel_caller(chan)->ani.number.str,
					ast_channel_caller(peer)->ani.number.str);
			}

			pbx_exec(peer, macro_app, macro_args);

			rec_enable = pbx_builtin_getvar_helper(peer, "REC_ENABLE");
			recording_chan = pbx_builtin_getvar_helper(peer, "RECORDING_CHAN");
			if (!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")
				&& (ast_strlen_zero(recording_chan) || strcmp(recording_chan, ast_channel_name(peer)))) {
				snprintf(macro_args, 128, "record-check,StartMonitor");
				pbx_exec(peer, macro_app, macro_args);

				ast_cdr_setvar(ast_channel_name(chan), "special_action_type", "atxee");
				ast_cdr_setvar(ast_channel_name(chan), "app_caller",
					ast_channel_caller(chan)->ani.number.str);
			}

			mixfilename = pbx_builtin_getvar_helper(peer, "MIXFILENAME");
			a_xferer_hangup = pbx_builtin_getvar_helper(chan, "A_XFER_STATE");
			a_xfereree_name = pbx_builtin_getvar_helper(chan, "A_XFEREE_NAME");
			if (!ast_strlen_zero(mixfilename) && !ast_strlen_zero(a_xferer_hangup)
					&& !ast_strlen_zero(a_xfereree_name)
					&& !strcmp(a_xferer_hangup, "transferer hangup")
					&& !strcmp(a_xfereree_name, ast_channel_name(chan))) {
				char *f;
				ast_asprintf(&f, "%s.wav@", mixfilename);
				if (f) {
					pbx_builtin_setvar_helper(chan, "A_XFER_HANGUP_RECORD", f);
					ast_free(f);
				}

				ast_clear_flag64(&opts, OPT_CALLEE_MACRO);

				if (ast_test_flag64(&opts, OPT_ANNOUNCE) && !ast_strlen_zero(opt_args[OPT_ARG_ANNOUNCE])
					&& !strcmp(opt_args[OPT_ARG_ANNOUNCE], "dialog-being-recorded")) {
					ast_clear_flag64(&opts, OPT_ANNOUNCE);
				}
			}

			/* added by zgmo for sdp record attr */
			pbx_builtin_setvar_helper(chan, "REC_ENABLE", rec_enable);
			ast_indicate(chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
			ast_indicate(peer, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);
			/* end */

			ast_dial_prompt_outside_stop_record(chan);
		}

		/* added by zgmo for Bug 171813, ring both process optimization  */
		{
			const char *rec_check_on_answer = pbx_builtin_getvar_helper(chan, "REC_CHECK_ON_ANSWER");
			const char *pickup_caller = pbx_builtin_getvar_helper(peer, "PICKUP_CALLER");  /* Bug178826 */

			if (!ast_strlen_zero(rec_check_on_answer) && !strcmp(rec_check_on_answer, "yes")
				&& ast_strlen_zero(pickup_caller)
				&& !strstr(ast_channel_name(peer), "call-forward")
				&& !(strstr(ast_channel_name(peer), "Local/") && strstr(ast_channel_name(peer), ";1"))) {
				const char *from_trunk = pbx_builtin_getvar_helper(chan, "FROM_TRUNK");
				const char *leave_trunk = pbx_builtin_getvar_helper(chan, "LEAVE_TRUNK");
				const char *cf_reason = pbx_builtin_getvar_helper(chan, "CF_REASON");

				pbx_builtin_setvar_helper(peer, "TRUNK_INDEX", "");

				/*
				 *  when call ast_dial_check_record(), the executor channel is peer,
				 *  so we should pass some variables help to judge recording.
				 */
				if (!ast_strlen_zero(from_trunk) && !strcmp(from_trunk, "yes")) {
					const char *inbound_trunk_index = pbx_builtin_getvar_helper(chan, "INBOUND_TRUNK_INDEX");
					const char *inbound_rt_index = pbx_builtin_getvar_helper(chan, "INBOUND_RT_INDEX");
					const char *trunk_index = pbx_builtin_getvar_helper(chan, "TRUNK_INDEX");

					pbx_builtin_setvar_helper(peer, "INBOUND_TRUNK_INDEX", inbound_trunk_index);
					pbx_builtin_setvar_helper(peer, "INBOUND_RT_INDEX", inbound_rt_index);
					if (ast_strlen_zero(inbound_trunk_index)) {
						pbx_builtin_setvar_helper(peer, "TRUNK_INDEX", trunk_index);
					}
				}

				if (strstr(ast_channel_name(peer), "trunk")
					|| strstr(ast_channel_name(peer), "DAHDI/")) {
					//pbx_builtin_setvar_helper(peer, "TRUNK_INDEX", pbx_builtin_getvar_helper(chan, "TRUNK_INDEX"));
					pbx_builtin_setvar_helper(peer, "OUTBOUND_TRUNK_INDEX", pbx_builtin_getvar_helper(chan, "OUTBOUND_TRUNK_INDEX"));
					pbx_builtin_setvar_helper(peer, "OUTBOUND_RT_INDEX", pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_INDEX"));
					pbx_builtin_setvar_helper(peer, "NEED_PROMPT_EXT_STOP_RECORD", "yes");
				} else {
					if (ast_strlen_zero(leave_trunk) || strcmp(leave_trunk, "yes")) {
						pbx_builtin_setvar_helper(peer, "OUTBOUND_TRUNK_INDEX", NULL);
						pbx_builtin_setvar_helper(peer, "OUTBOUND_RT_INDEX", NULL);
					}
				}

				/* added by zgmo for Bug189180 */
				if ((!ast_strlen_zero(leave_trunk) && !strcmp(leave_trunk, "yes"))
					&& (!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "13")))  { /* Bug190442 */
					pbx_builtin_setvar_helper(peer, "TRUNK_INDEX", pbx_builtin_getvar_helper(chan, "TRUNKINDEX_NUM"));
					pbx_builtin_setvar_helper(peer, "INBOUND_TRUNK_INDEX", pbx_builtin_getvar_helper(chan, "OUTBOUND_TRUNK_PRE_INDEX"));
					pbx_builtin_setvar_helper(peer, "INBOUND_RT_INDEX", pbx_builtin_getvar_helper(chan, "OUTBOUND_RT_PRE_INDEX"));
				}
				/* end */

				pbx_builtin_setvar_helper(peer, "REC_CHECK_ON_ANSWER", rec_check_on_answer);
				pbx_builtin_setvar_helper(peer, "REC_CHECK_ON_ANSWER_CALLER", ast_channel_name(chan));
				pbx_builtin_setvar_helper(peer, "RINGBOTH", pbx_builtin_getvar_helper(chan, "RINGBOTH"));
				pbx_builtin_setvar_helper(peer, "FM_ENABLE", pbx_builtin_getvar_helper(chan, "FM_ENABLE"));
				pbx_builtin_setvar_helper(peer, "REG_CALLEENUM", pbx_builtin_getvar_helper(chan, "REG_CALLEENUM"));
				pbx_builtin_setvar_helper(peer, "A_XFER_STATE", pbx_builtin_getvar_helper(chan, "A_XFER_STATE"));

				if (1 == ast_dial_check_record(chan, peer)) {
					const char *recording_file_cdr = pbx_builtin_getvar_helper(chan, "RECORDING_FILE_CDR");
					const char *recording_file_key_cdr = pbx_builtin_getvar_helper(chan, "RECORDING_FILE_KEY_CDR");
					const char *fm_caller = pbx_builtin_getvar_helper(chan, "FOLLOWME_CALLER");

					if (!ast_strlen_zero(recording_file_cdr)) {
						ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES", recording_file_cdr);

						if (!ast_strlen_zero(recording_file_key_cdr)) {
							ast_cdr_setvar(ast_channel_name(chan), "record_files_uuid", recording_file_key_cdr);
							ast_log(LOG_DEBUG, "[dial_exec_full] channel:%s file-key:%s", ast_channel_name(chan), recording_file_key_cdr);
						}
					}

					pbx_builtin_setvar_helper(chan, "REC_ENABLE", "yes");
					pbx_builtin_setvar_helper(peer, "RECORD_ATTR_SDP_TRIGGER", "yes");

					if (!ast_strlen_zero(fm_caller)) {
						struct ast_channel *fm_caller_chan = ast_channel_get_by_name(fm_caller);
						if (fm_caller_chan) {
							pbx_builtin_setvar_helper(fm_caller_chan, "REC_ENABLE", "yes");
							ast_channel_unref(fm_caller_chan);
						}
					}

					process_outside_stop_record_for_dial(chan, &config);
				}

				record_check_forbiden = 1;

				/* added by zgmo for Bug186429 */
				pbx_builtin_setvar_helper(chan, "REC_CHECK_ON_ANSWER", "");
				pbx_builtin_setvar_helper(peer, "REC_CHECK_ON_ANSWER", "");
				/* end */
			}
		}

		/* added by zgmo for Bug186429 */
		{
			const char *fm_enable = pbx_builtin_getvar_helper(chan, "FM_ENABLE");
			if (!ast_strlen_zero(fm_enable) && !strcmp(fm_enable, "yes")) {
				pbx_builtin_setvar_helper(chan, "FM_ENABLE", "");
				ast_log(LOG_DEBUG, "Chan:%s Clean ChanVar:FM_ENABLE", ast_channel_name(chan));
			}
		}
		/* end */

		{
			const char *from_park_dest = pbx_builtin_getvar_helper(chan, "FROM_PARK_DEST");
			const char *from_park_calling = pbx_builtin_getvar_helper(chan, "FROM_PARK_CALLING");
			const char *from_park = pbx_builtin_getvar_helper(chan, "FROM_PARK");

			ast_log(LOG_DEBUG, "from_park:%s, from_park_calling:%s, from_park_dest:%s",
				ast_strlen_zero(from_park) ? "":from_park,
				ast_strlen_zero(from_park_calling) ? "":from_park_calling,
				ast_strlen_zero(from_park_dest) ? "":from_park_dest);

			if (!ast_strlen_zero(from_park) && !strcmp(from_park, "yes")
				&& (ast_strlen_zero(from_park_dest) || strcmp(from_park_dest, "yes"))
				&& !record_check_forbiden) {
				int peer_is_callee = 1;
				const char *caller_do_park = pbx_builtin_getvar_helper(chan, "CALLER_DO_PARK");
				const char *callee_do_park = pbx_builtin_getvar_helper(chan, "CALLEE_DO_PARK");

				ast_log(LOG_DEBUG, "Detect callback for parking timeout");

				if (!ast_strlen_zero(caller_do_park) && !strcmp(caller_do_park, "yes")) {
					ast_log(LOG_DEBUG, "Parker:%s originally is caller", ast_channel_name(peer));
					peer_is_callee = 0;
				} else if (!ast_strlen_zero(callee_do_park) && !strcmp(callee_do_park, "yes")) {
					ast_log(LOG_DEBUG, "Parker:%s originally is callee", ast_channel_name(peer));
					pbx_builtin_setvar_helper(peer, "IS_CALLEE", "yes");
					peer_is_callee = 1;
				} else {
					const char *is_callee = pbx_builtin_getvar_helper(chan, "IS_CALLEE");
					if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
						peer_is_callee = 0;
					}
				}

				if (strstr(ast_channel_name(chan), "Local/")) {
					struct ast_channel *real_peer = ast_channel_real_peer(chan);
					if (real_peer) {
						int is_recording = ast_channel_avs_is_recording(real_peer);

						ast_log(LOG_DEBUG, "%s is_recording:%d, peer_is_callee:%d",
							ast_channel_name(real_peer), is_recording, peer_is_callee);
						if (!is_recording) {
							pbx_builtin_setvar_helper(peer, "IVR_REC",
								pbx_builtin_getvar_helper(real_peer, "IVR_REC"));
							pbx_builtin_setvar_helper(peer, "FROM_PARK",
								pbx_builtin_getvar_helper(chan, "FROM_PARK"));

							if (ast_dial_check_record(real_peer, peer)) {
								char cdr_record[512];
								ast_cdr_getvar(ast_channel_name(peer), "RECORD_FILES", cdr_record, sizeof(cdr_record));
								ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES", cdr_record);
							}
						}

						ast_channel_unref(real_peer);
					}
				} else {
					pbx_builtin_setvar_helper(peer, "IVR_REC",
						pbx_builtin_getvar_helper(chan, "IVR_REC"));
					pbx_builtin_setvar_helper(peer, "FROM_PARK",
						pbx_builtin_getvar_helper(chan, "FROM_PARK"));

					if (peer_is_callee) {
						ast_dial_check_record(chan, peer);
					} else {
						ast_dial_check_record(peer, chan);
					}
				}
			}

			if (!ast_strlen_zero(from_park_calling) && !strcmp(from_park_calling, "yes")) {
				pbx_builtin_setvar_helper(chan, "FROM_PARK_CALLING", NULL);
			}
		}

		/* Added by gangwang for Bug 259522 */
		{
			const char *from_privileged_call = pbx_builtin_getvar_helper(peer, "FROM_PRIVILEGED_CALL");
			if (ast_channel_name(peer) && strstr(ast_channel_name(peer), "Local/") &&
				!ast_strlen_zero(from_privileged_call) && !strcmp(from_privileged_call, "yes")) {
				char cdr_record[512] = {0};
				ast_cdr_getvar(ast_channel_name(peer), "RECORD_FILES", cdr_record, sizeof(cdr_record));
				ast_log(LOG_DEBUG, "%s get cdr record:%s", ast_channel_name(peer), cdr_record);
				if (!ast_strlen_zero(cdr_record)) {
						ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES", cdr_record);
						ast_cdr_setvar(ast_channel_name(chan), "special_action_type", "PRIVILEGED_CALL");
						ast_log(LOG_DEBUG, "set cdr record to %s", ast_channel_name(chan));
				}
		}

		}
		if (ast_test_flag64(&opts, OPT_ANNOUNCE) && !ast_strlen_zero(opt_args[OPT_ARG_ANNOUNCE])
			&& !strcmp(opt_args[OPT_ARG_ANNOUNCE], "dialog-being-recorded")) {
			const char *have_prompted = pbx_builtin_getvar_helper(peer, "HAVE_PROMPTED");
			const char *pickup_caller = pbx_builtin_getvar_helper(peer, "PICKUP_CALLER");
			const char *seamless_transfer_record = pbx_builtin_getvar_helper(chan, "SEAMLESS_TRANSFER_RECORD");

			/*
			 *  added by zgmo for bug 144632
			 */
			if (strstr(ast_channel_name(peer),"call-forward")) {
				ast_clear_flag64(&opts, OPT_ANNOUNCE);
			}

			if (!ast_strlen_zero(have_prompted) && !strcmp(have_prompted, "yes")) {
				ast_clear_flag64(&opts, OPT_ANNOUNCE);
			}

			/* Add by bfdu for bug 211171, 2021.12.23 */
			const char *peer_enter_vm = pbx_builtin_getvar_helper(peer, "PEER_ENTER_VM");
			if (!ast_strlen_zero(peer_enter_vm) && !strcmp(peer_enter_vm, "yes")) {
				ast_log(LOG_DEBUG, "Chan:%s receive 200OK, but peer enter voicemail, don't need play sound\n", ast_channel_name(peer));
				ast_clear_flag64(&opts, OPT_ANNOUNCE);
			}

			/*
			 * added by zgmo for Bug155326
			 */
			if (!ast_strlen_zero(pickup_caller) && !strcmp(pickup_caller, "yes")) {
				const char *have_prompted = pbx_builtin_getvar_helper(chan, "HAVE_PROMPTED");
				const char *rec_enable = pbx_builtin_getvar_helper(peer, "REC_ENABLE");

				if (!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "no")) {
					if (!ast_strlen_zero(have_prompted) && !strcmp(have_prompted, "yes")) {
						if (!ast_streamfile(chan, "dialog-stop-recording", ast_channel_language(chan))) {
							ast_waitstream(chan, "");
							ast_stopstream(chan);

							pbx_builtin_setvar_helper(chan, "HAVE_PROMPTED", "no");
						}
					}
				}

				ast_clear_flag64(&opts, OPT_ANNOUNCE);

				if (ast_test_flag64(&opts, OPT_CALLEE_MACRO)
						&& !ast_strlen_zero(opt_args[OPT_ARG_CALLEE_MACRO])
						&& !strcmp(opt_args[OPT_ARG_CALLEE_MACRO], "record-check,StartMonitor")) {
						ast_clear_flag64(&opts, OPT_CALLEE_MACRO);
				}
			}

			if (!ast_strlen_zero(seamless_transfer_record)
				&& strstr(ast_channel_name(chan), "Local/")) {
				struct ast_channel *local_peer = ast_local_get_peer(chan);

				//ast_clear_flag64(&opts, OPT_ANNOUNCE);
				if (ast_test_flag64(&opts, OPT_CALLEE_MACRO)
						&& !ast_strlen_zero(opt_args[OPT_ARG_CALLEE_MACRO])
						&& !strcmp(opt_args[OPT_ARG_CALLEE_MACRO], "record-check,StartMonitor")) {
						ast_clear_flag64(&opts, OPT_CALLEE_MACRO);
				}

				if (local_peer) {
					pbx_builtin_setvar_helper(local_peer, "SEAMLESS_TRANSFER_RECORD", seamless_transfer_record);
					ast_channel_unref(local_peer);
				}
			}
		}

		{
			const char *vq_cb_caller = pbx_builtin_getvar_helper(chan, "VQ_CALLBACK_CALLER");
			const char *vq_cb_rec_enable = pbx_builtin_getvar_helper(chan, "VQ_CALLBACK_REC_ENABLE");
			if (!ast_strlen_zero(vq_cb_caller) && !strcmp(vq_cb_caller, "yes")
				&& !ast_strlen_zero(vq_cb_rec_enable) && !strcmp(vq_cb_rec_enable, "yes")
				&& strstr(ast_channel_name(chan), "Local/")
				&& strstr(ast_channel_name(chan), ";2")) {
				if (ast_test_flag64(&opts, OPT_CALLEE_MACRO)
						&& !ast_strlen_zero(opt_args[OPT_ARG_CALLEE_MACRO])
						&& !strcmp(opt_args[OPT_ARG_CALLEE_MACRO], "record-check,StartMonitor")) {
						ast_clear_flag64(&opts, OPT_CALLEE_MACRO);
						ast_cdr_newest_setvar(ast_channel_name(chan), "record_files", "");
				}
			}
		}
#endif
		if (!ast_test_flag64(&opts, OPT_ANNOUNCE) || ast_strlen_zero(opt_args[OPT_ARG_ANNOUNCE])
			|| ast_test_flag(ast_channel_flags(peer), AST_FLAG_RECOVERYING)) {
			res = 0;
		} else {
			int digit = 0;
			struct ast_channel *chans[2];
			struct ast_channel *active_chan;

			chans[0] = chan;
			chans[1] = peer;

			/* we need to stream the announcement to the called party when the OPT_ARG_ANNOUNCE (-A) is setted */

			/* stream the file */
			res = ast_streamfile(peer, opt_args[OPT_ARG_ANNOUNCE], ast_channel_language(peer));
			if (res) {
				res = 0;
				ast_log(LOG_ERROR, "error streaming file '%s' to callee\n", opt_args[OPT_ARG_ANNOUNCE]);
			}

			ast_channel_set_flag(peer, AST_FLAG_END_DTMF_ONLY);
			while (ast_channel_stream(peer)) {
				int ms;

				ms = ast_sched_wait(ast_channel_sched(peer));

				if (ms < 0 && !ast_channel_timingfunc(peer)) {
					ast_stopstream(peer);
					break;
				}
				if (ms < 0)
					ms = 1000;

				active_chan = ast_waitfor_n(chans, 2, &ms);
				if (active_chan) {
					struct ast_channel *other_chan;
					struct ast_frame *fr = ast_read(active_chan);

					if (!fr) {
						ast_autoservice_chan_hangup_peer(chan, peer);
						res = -1;
						goto done;
					}
					switch (fr->frametype) {
					case AST_FRAME_DTMF_END:
						digit = fr->subclass.integer;
						if (active_chan == peer && strchr(AST_DIGIT_ANY, res)) {
							ast_stopstream(peer);
							res = ast_senddigit(chan, digit, 0);
						}
						break;
					case AST_FRAME_CONTROL:
						switch (fr->subclass.integer) {
						case AST_CONTROL_HANGUP:
							ast_frfree(fr);
							ast_autoservice_chan_hangup_peer(chan, peer);
							res = -1;
							goto done;
						case AST_CONTROL_CONNECTED_LINE:
							/* Pass COLP update to the other channel. */
							if (active_chan == chan) {
								other_chan = peer;
							} else {
								other_chan = chan;
							}
							if (ast_channel_connected_line_sub(active_chan, other_chan, fr, 1)
								&& ast_channel_connected_line_macro(active_chan,
									other_chan, fr, other_chan == chan, 1)) {
								ast_indicate_data(other_chan, fr->subclass.integer,
									fr->data.ptr, fr->datalen);
							}
							break;
						default:
#ifdef GRANDSTREAM_NETWORKS
							ast_log(LOG_ERROR, "Ignore frame control %d\n",fr->subclass.integer);
#endif
							break;
						}
						break;
#ifdef GRANDSTREAM_NETWORKS
					case AST_FRAME_VOICE:
						if (active_chan == peer) {
							ast_write(chan, fr);
						}
						break;
#endif
					default:
						/* Ignore all others */
						break;
					}
					ast_frfree(fr);
				}
				ast_sched_runq(ast_channel_sched(peer));
			}
			ast_channel_clear_flag(peer, AST_FLAG_END_DTMF_ONLY);
#ifdef GRANDSTREAM_NETWORKS
			/*begin: add by weidu ,for 110820, 20180621,paging caller Paging/ Intercom Announcement Warning */
			if (!ast_channel_reg_sendbeep_get(chan) && strcasestr(ast_channel_name(peer),"MulticastRTP")){
				res = ast_streamfile(chan, "beep", ast_channel_language(chan));
				if (res) {
					res = 0;
					ast_log(LOG_ERROR, "error streaming file to caller\n");
				}
				else{
					ast_channel_reg_sendbeep_set(chan, 1);
				}
			}
			/*end: add by weidu ,for 110820*/

			/* added by zgmo for Bug161683  */
			if (!res && !strcmp(opt_args[OPT_ARG_ANNOUNCE], "dialog-being-recorded")) {
				pbx_builtin_setvar_helper(peer, "HAVE_PROMPTED", "yes");
			}
			/* add end */
#endif
		}

		if (chan && peer && ast_test_flag64(&opts, OPT_GOTO) && !ast_strlen_zero(opt_args[OPT_ARG_GOTO])) {
#ifdef GRANDSTREAM_NETWORKS
			RAII_VAR(struct ast_channel *, tmppeerchan, ao2_bump(peer), ao2_cleanup); // Added by pxie@grandstrema.cn for bug[205015], 2022-07-09.
#endif
			/* chan and peer are going into the PBX; as such neither are considered
			 * outgoing channels any longer */
			ast_channel_clear_flag(chan, AST_FLAG_OUTGOING);

			ast_replace_subargument_delimiter(opt_args[OPT_ARG_GOTO]);
			ast_parseable_goto(chan, opt_args[OPT_ARG_GOTO]);
			/* peer goes to the same context and extension as chan, so just copy info from chan*/
			ast_channel_lock(peer);
			ast_channel_stage_snapshot(peer);
			ast_clear_flag(ast_channel_flags(peer), AST_FLAG_OUTGOING);
			ast_channel_context_set(peer, ast_channel_context(chan));
			ast_channel_exten_set(peer, ast_channel_exten(chan));
			ast_channel_priority_set(peer, ast_channel_priority(chan) + 2);
			ast_channel_stage_snapshot_done(peer);
#ifdef GRANDSTREAM_NETWORKS
			if (!strcasecmp(ast_channel_tech(peer)->type, "Local") && !ast_strlen_zero(pbx_builtin_getvar_helper(peer, "CONF_ROOM"))) {
				ast_channel_no_increase_call_set(peer, 1);
			}

			if (!strcasecmp(ast_channel_tech(chan)->type, "DAHDI")) {
				ast_indicate(chan, -1);
			}
#endif
			ast_channel_unlock(peer);
			if (ast_pbx_start(peer)) {
				ast_autoservice_chan_hangup_peer(chan, peer);
			}
			if (continue_exec)
				*continue_exec = 1;
			res = 0;
			ast_channel_publish_dial(chan, peer, NULL, "ANSWER");
			goto done;
		}

		if (ast_test_flag64(&opts, OPT_CALLEE_MACRO) && !ast_strlen_zero(opt_args[OPT_ARG_CALLEE_MACRO])) {
			const char *macro_result_peer;
			int macro_res;

			/* Set peer->exten and peer->context so that MACRO_EXTEN and MACRO_CONTEXT get set */
			ast_channel_lock_both(chan, peer);
			ast_channel_context_set(peer, ast_channel_context(chan));
			ast_channel_exten_set(peer, ast_channel_exten(chan));
			ast_channel_unlock(peer);
			ast_channel_unlock(chan);
			ast_replace_subargument_delimiter(opt_args[OPT_ARG_CALLEE_MACRO]);
			macro_res = ast_app_exec_macro(chan, peer, opt_args[OPT_ARG_CALLEE_MACRO]);

			ast_channel_lock(peer);

			if (!macro_res && (macro_result_peer = pbx_builtin_getvar_helper(peer, "MACRO_RESULT"))) {
				char *macro_result = ast_strdupa(macro_result_peer);
				char *macro_transfer_dest;

				ast_channel_unlock(peer);

				if (!strcasecmp(macro_result, "BUSY")) {
					ast_copy_string(pa.status, macro_result, sizeof(pa.status));
					ast_set_flag64(peerflags, OPT_GO_ON);
					macro_res = -1;
				} else if (!strcasecmp(macro_result, "CONGESTION") || !strcasecmp(macro_result, "CHANUNAVAIL")) {
					ast_copy_string(pa.status, macro_result, sizeof(pa.status));
					ast_set_flag64(peerflags, OPT_GO_ON);
					macro_res = -1;
				} else if (!strcasecmp(macro_result, "CONTINUE")) {
					/* hangup peer and keep chan alive assuming the macro has changed
					   the context / exten / priority or perhaps
					   the next priority in the current exten is desired.
					*/
					ast_set_flag64(peerflags, OPT_GO_ON);
					macro_res = -1;
				} else if (!strcasecmp(macro_result, "ABORT")) {
					/* Hangup both ends unless the caller has the g flag */
					macro_res = -1;
				} else if (!strncasecmp(macro_result, "GOTO:", 5)) {
					macro_transfer_dest = macro_result + 5;
					macro_res = -1;
					/* perform a transfer to a new extension */
					if (strchr(macro_transfer_dest, '^')) { /* context^exten^priority*/
						ast_replace_subargument_delimiter(macro_transfer_dest);
					}
					if (!ast_parseable_goto(chan, macro_transfer_dest)) {
						ast_set_flag64(peerflags, OPT_GO_ON);
					}
				}
				if (macro_res && !dial_end_raised) {
					ast_channel_publish_dial(chan, peer, NULL, macro_result);
					dial_end_raised = 1;
				}
			} else {
				ast_channel_unlock(peer);
			}
			res = macro_res;
		}

		if (ast_test_flag64(&opts, OPT_CALLEE_GOSUB) && !ast_strlen_zero(opt_args[OPT_ARG_CALLEE_GOSUB])) {
			const char *gosub_result_peer;
			char *gosub_argstart;
			char *gosub_args = NULL;
			int gosub_res = -1;

			ast_replace_subargument_delimiter(opt_args[OPT_ARG_CALLEE_GOSUB]);
			gosub_argstart = strchr(opt_args[OPT_ARG_CALLEE_GOSUB], ',');
			if (gosub_argstart) {
				const char *what_is_s = "s";
				*gosub_argstart = 0;
				if (!ast_exists_extension(peer, opt_args[OPT_ARG_CALLEE_GOSUB], "s", 1, S_COR(ast_channel_caller(peer)->id.number.valid, ast_channel_caller(peer)->id.number.str, NULL)) &&
					 ast_exists_extension(peer, opt_args[OPT_ARG_CALLEE_GOSUB], "~~s~~", 1, S_COR(ast_channel_caller(peer)->id.number.valid, ast_channel_caller(peer)->id.number.str, NULL))) {
					what_is_s = "~~s~~";
				}
				if (ast_asprintf(&gosub_args, "%s,%s,1(%s)", opt_args[OPT_ARG_CALLEE_GOSUB], what_is_s, gosub_argstart + 1) < 0) {
					gosub_args = NULL;
				}
				*gosub_argstart = ',';
			} else {
				const char *what_is_s = "s";
				if (!ast_exists_extension(peer, opt_args[OPT_ARG_CALLEE_GOSUB], "s", 1, S_COR(ast_channel_caller(peer)->id.number.valid, ast_channel_caller(peer)->id.number.str, NULL)) &&
					 ast_exists_extension(peer, opt_args[OPT_ARG_CALLEE_GOSUB], "~~s~~", 1, S_COR(ast_channel_caller(peer)->id.number.valid, ast_channel_caller(peer)->id.number.str, NULL))) {
					what_is_s = "~~s~~";
				}
				if (ast_asprintf(&gosub_args, "%s,%s,1", opt_args[OPT_ARG_CALLEE_GOSUB], what_is_s) < 0) {
					gosub_args = NULL;
				}
			}
			if (gosub_args) {
				gosub_res = ast_app_exec_sub(chan, peer, gosub_args, 0);
				ast_free(gosub_args);
			} else {
				ast_log(LOG_ERROR, "Could not Allocate string for Gosub arguments -- Gosub Call Aborted!\n");
			}

			ast_channel_lock_both(chan, peer);

			if (!gosub_res && (gosub_result_peer = pbx_builtin_getvar_helper(peer, "GOSUB_RESULT"))) {
				char *gosub_transfer_dest;
				char *gosub_result = ast_strdupa(gosub_result_peer);
				const char *gosub_retval = pbx_builtin_getvar_helper(peer, "GOSUB_RETVAL");

				/* Inherit return value from the peer, so it can be used in the master */
				if (gosub_retval) {
					pbx_builtin_setvar_helper(chan, "GOSUB_RETVAL", gosub_retval);
				}

				ast_channel_unlock(peer);
				ast_channel_unlock(chan);

				if (!strcasecmp(gosub_result, "BUSY")) {
					ast_copy_string(pa.status, gosub_result, sizeof(pa.status));
					ast_set_flag64(peerflags, OPT_GO_ON);
					gosub_res = -1;
				} else if (!strcasecmp(gosub_result, "CONGESTION") || !strcasecmp(gosub_result, "CHANUNAVAIL")) {
					ast_copy_string(pa.status, gosub_result, sizeof(pa.status));
					ast_set_flag64(peerflags, OPT_GO_ON);
					gosub_res = -1;
				} else if (!strcasecmp(gosub_result, "CONTINUE")) {
					/* Hangup peer and continue with the next extension priority. */
					ast_set_flag64(peerflags, OPT_GO_ON);
					gosub_res = -1;
				} else if (!strcasecmp(gosub_result, "ABORT")) {
					/* Hangup both ends unless the caller has the g flag */
					gosub_res = -1;
				} else if (!strncasecmp(gosub_result, "GOTO:", 5)) {
					gosub_transfer_dest = gosub_result + 5;
					gosub_res = -1;
					/* perform a transfer to a new extension */
					if (strchr(gosub_transfer_dest, '^')) { /* context^exten^priority*/
						ast_replace_subargument_delimiter(gosub_transfer_dest);
					}
					if (!ast_parseable_goto(chan, gosub_transfer_dest)) {
						ast_set_flag64(peerflags, OPT_GO_ON);
					}
				}
				if (gosub_res) {
					res = gosub_res;
					if (!dial_end_raised) {
						ast_channel_publish_dial(chan, peer, NULL, gosub_result);
						dial_end_raised = 1;
					}
				}
			} else {
				ast_channel_unlock(peer);
				ast_channel_unlock(chan);
			}
		}

		if (!res) {

			/* None of the Dial options changed our status; inform
			 * everyone that this channel answered
			 */
			if (!dial_end_raised) {
				ast_channel_publish_dial(chan, peer, NULL, "ANSWER");
				dial_end_raised = 1;
			}

			if (!ast_tvzero(calldurationlimit)) {
				struct timeval whentohangup = ast_tvadd(ast_tvnow(), calldurationlimit);
#ifdef GRANDSTREAM_NETWORKS
				/*Start added by zyxia for Bug 224512 ,2023.3.7,edit by zyxia for Bug 266881 */
				if(peer && ast_channel_name(peer) && (strstr(ast_channel_name(peer),"trunk") || strstr(ast_channel_name(peer),"DAHDI/o"))){
					const char *outbound_trunk_index = pbx_builtin_getvar_helper(peer, "OUTBOUND_TRUNK_INDEX");
					if(outbound_trunk_index){
						config.timelimit = calldurationlimit.tv_sec*1000;
					}
					else {
						/* Added by zgmo,
						 * BUG 143684 - basic call:Call duration restrictions affect transfer calls
						 */
						ast_channel_lock(chan);
						ast_channel_whentohangup_set(chan, &whentohangup);
						ast_channel_unlock(chan);
					}
				}
				/*End added by zyxia for Bug 224512 ,2023.3.7*/
				else {
					/* Added by zgmo,
					 * BUG 143684 - basic call:Call duration restrictions affect transfer calls
					 */
					ast_channel_lock(chan);
					ast_channel_whentohangup_set(chan, &whentohangup);
					ast_channel_unlock(chan);
				}
#else
				ast_channel_lock(peer);
				ast_channel_whentohangup_set(peer, &whentohangup);
				ast_channel_unlock(peer);
#endif
			}
			if (!ast_strlen_zero(dtmfcalled)) {
				ast_verb(3, "Sending DTMF '%s' to the called party.\n", dtmfcalled);
				res = ast_dtmf_stream(peer, chan, dtmfcalled, 250, 0);
			}
			if (!ast_strlen_zero(dtmfcalling)) {
				ast_verb(3, "Sending DTMF '%s' to the calling party.\n", dtmfcalling);
				res = ast_dtmf_stream(chan, peer, dtmfcalling, 250, 0);
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		int trans_prohibited = 0; /*use for SCA calls raised by the originate command.*/
		const char* callee_is_sca = pbx_builtin_getvar_helper(chan, "ORIGINATE_SCA");
		if (!ast_strlen_zero(callee_is_sca) && 0 == strcmp(callee_is_sca, "yes")) {
			ast_log(LOG_DEBUG,"The real callee is SCA extension, transfer is not allowed.");
			trans_prohibited = 1;
		}
#endif
		if (res) { /* some error */
			if (!ast_check_hangup(chan) && ast_check_hangup(peer)) {
				ast_channel_hangupcause_set(chan, ast_channel_hangupcause(peer));
			}
			setup_peer_after_bridge_goto(chan, peer, &opts, opt_args);
			if (ast_bridge_setup_after_goto(peer)
				|| ast_pbx_start(peer)) {
				ast_autoservice_chan_hangup_peer(chan, peer);
			}
			res = -1;
		} else {
#ifdef GRANDSTREAM_NETWORKS
			if (ast_test_flag64(peerflags, OPT_CALLEE_TRANSFER) && trans_prohibited == 0)
				ast_set_flag(&(config.features_callee), AST_FEATURE_REDIRECT);
			if (ast_test_flag64(peerflags, OPT_CALLER_TRANSFER) && trans_prohibited == 0)
				ast_set_flag(&(config.features_caller), AST_FEATURE_REDIRECT);
#else
			if (ast_test_flag64(peerflags, OPT_CALLEE_TRANSFER))
				ast_set_flag(&(config.features_callee), AST_FEATURE_REDIRECT);
			if (ast_test_flag64(peerflags, OPT_CALLER_TRANSFER))
				ast_set_flag(&(config.features_caller), AST_FEATURE_REDIRECT);
#endif
			if (ast_test_flag64(peerflags, OPT_CALLEE_HANGUP))
				ast_set_flag(&(config.features_callee), AST_FEATURE_DISCONNECT);
			if (ast_test_flag64(peerflags, OPT_CALLER_HANGUP))
				ast_set_flag(&(config.features_caller), AST_FEATURE_DISCONNECT);
			if (ast_test_flag64(peerflags, OPT_CALLEE_MONITOR))
				ast_set_flag(&(config.features_callee), AST_FEATURE_AUTOMON);
			if (ast_test_flag64(peerflags, OPT_CALLER_MONITOR))
				ast_set_flag(&(config.features_caller), AST_FEATURE_AUTOMON);
			if (ast_test_flag64(peerflags, OPT_CALLEE_PARK))
				ast_set_flag(&(config.features_callee), AST_FEATURE_PARKCALL);
			if (ast_test_flag64(peerflags, OPT_CALLER_PARK))
				ast_set_flag(&(config.features_caller), AST_FEATURE_PARKCALL);
			if (ast_test_flag64(peerflags, OPT_CALLEE_MIXMONITOR))
				ast_set_flag(&(config.features_callee), AST_FEATURE_AUTOMIXMON);
			if (ast_test_flag64(peerflags, OPT_CALLER_MIXMONITOR))
				ast_set_flag(&(config.features_caller), AST_FEATURE_AUTOMIXMON);

			config.end_bridge_callback = end_bridge_callback;
			config.end_bridge_callback_data = chan;
			config.end_bridge_callback_data_fixup = end_bridge_callback_data_fixup;

			if (moh) {
				moh = 0;
				ast_moh_stop(chan);
			} else if (sentringing) {
				sentringing = 0;
				ast_indicate(chan, -1);
			}
			/* Be sure no generators are left on it and reset the visible indication */
			ast_deactivate_generator(chan);
			ast_channel_visible_indication_set(chan, 0);
			/* Make sure channels are compatible */
			res = ast_channel_make_compatible(chan, peer);
			if (res < 0) {
				ast_log(LOG_WARNING, "Had to drop call because I couldn't make %s compatible with %s\n", ast_channel_name(chan), ast_channel_name(peer));
				ast_autoservice_chan_hangup_peer(chan, peer);
				res = -1;
				goto done;
			}
			if (opermode) {
				struct oprmode oprmode;

				oprmode.peer = peer;
				oprmode.mode = opermode;

				ast_channel_setoption(chan, AST_OPTION_OPRMODE, &oprmode, sizeof(oprmode), 0);
			}
			setup_peer_after_bridge_goto(chan, peer, &opts, opt_args);
#ifdef GRANDSTREAM_NETWORKS
			if (ast_channel_get_is_nat_call(peer)) {
				ast_log(LOG_DEBUG, "Peer(%s) is NAT", ast_channel_name(peer));
				if (strstr(ast_channel_name(chan), "Local/") && strstr(ast_channel_name(chan), ";2")) {
					struct ast_channel *local_peer = ast_local_get_peer(chan);
					if (local_peer) {
						ast_log(LOG_DEBUG, "%s is NAT, Mark %s NAT", ast_channel_name(peer), ast_channel_name(local_peer));
						ast_channel_set_turn_session_count(local_peer, ast_channel_get_turn_session_count(peer));
						ast_channel_unref(local_peer);
					}
				}
			}

			if (ast_strlen_zero(pbx_builtin_getvar_helper(peer, "HAVE_PROMPTED"))
				|| strcmp(pbx_builtin_getvar_helper(peer, "HAVE_PROMPTED"), "yes")) {
				if (strstr(ast_channel_name(chan), "Local/")) {
					/*
					 *  peer who pickup the call without prompted record, if caller is Local,
					 *  there maybe multiple bridges across apps,
					 *  e.g. A -- Queue -- Followme --- B.
					 *  Trunk -------- Local1 | Local2 --- Local1 | Local2 ----- 4003
					 *          QUEUE       (302)             (Followme)
					 */
					const char *followme_caller = pbx_builtin_getvar_helper(chan, "FOLLOWME_CALLER");
					const char *initial_caller = pbx_builtin_getvar_helper(chan, "FIRST_CALLER_CHANNEL_NAME");
					if (!ast_strlen_zero(initial_caller) && !ast_strlen_zero(followme_caller)) {
						struct ast_channel *initial_caller_chan = ast_channel_get_by_name(initial_caller);
						if (initial_caller_chan) {
							if (ast_channel_avs_is_recording(initial_caller_chan)) {
								ast_stream_and_wait(peer, "dialog-being-recorded", "");
							}
							ast_channel_unref(initial_caller_chan);
						}
					}
				}
			}

			{
				const char *from_ivr = pbx_builtin_getvar_helper(chan, "FROM_IVR");
				const char *from_dir = pbx_builtin_getvar_helper(chan, "FROM_DIRECTORY");
				const char *rec_enable = pbx_builtin_getvar_helper(chan, "REC_ENABLE");

				if ((!ast_strlen_zero(from_ivr) && !strcmp(from_ivr, "yes"))
					|| (!ast_strlen_zero(from_dir) && !strcmp(from_dir, "yes"))) {
					if(!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")) {
						pbx_builtin_setvar_helper(chan, "RECORD_ATTR_SDP_TRIGGER", "yes");
						ast_log(LOG_DEBUG, "Trigger record sdp update for Chan:%s",
							ast_channel_name(chan));
					}
				}

				if(!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")) {
					if (ast_dial_prompt_outside_stop_record(peer)) {
						ast_set_flag(&(config.features_callee), AST_FEATURE_AUTOMIXMON);
						ast_log(LOG_WARNING, "Set callee automixmon flag");
					}
				}
			}
#endif
			res = ast_bridge_call(chan, peer, &config);
#ifdef GRANDSTREAM_NETWORKS
			{
				const char *a_xfer_hangup_record = pbx_builtin_getvar_helper(chan, "A_XFER_HANGUP_RECORD");
				if (!ast_strlen_zero(a_xfer_hangup_record)) {
					ast_log(LOG_DEBUG, "Chan:%s Set cdr:record_files to '%s'", ast_channel_name(chan), a_xfer_hangup_record);
					ast_cdr_newest_setvar(ast_channel_name(chan), "record_files", a_xfer_hangup_record);
				}
			}
#endif
		}
	}
out:
	if (moh) {
		moh = 0;
		ast_moh_stop(chan);
	} else if (sentringing) {
		sentringing = 0;
		ast_indicate(chan, -1);
	}

	if (delprivintro && ast_fileexists(pa.privintro, NULL, NULL) > 0) {
		ast_filedelete(pa.privintro, NULL);
		if (ast_fileexists(pa.privintro, NULL, NULL) > 0) {
			ast_log(LOG_NOTICE, "privacy: ast_filedelete didn't do its job on %s\n", pa.privintro);
		} else {
			ast_verb(3, "Successfully deleted %s intro file\n", pa.privintro);
		}
	}

	ast_channel_early_bridge(chan, NULL);
	 /* forward 'answered elsewhere' if we received it */
	hanguptree(&out_chans, NULL,
		ast_channel_hangupcause(chan) == AST_CAUSE_ANSWERED_ELSEWHERE
		|| ast_test_flag64(&opts, OPT_CANCEL_ELSEWHERE)
		? AST_CAUSE_ANSWERED_ELSEWHERE : -1);
	pbx_builtin_setvar_helper(chan, "DIALSTATUS", pa.status);

#ifdef GRANDSTREAM_NETWORKS
	if (!strncmp(ast_channel_name(chan), "DAHDI/", strlen("DAHDI/"))) {
		/* Add by weidu for blind transfer callback BUG 155999*/
		const char *flag_clear_callback = pbx_builtin_getvar_helper((struct ast_channel *)chan, "FLAG_CLEAR_CALLBACK");
		if (flag_clear_callback && !strcasecmp(flag_clear_callback, "CLEAR")) {
			pbx_builtin_setvar_helper(chan, "BLIND_TRANSFER_CALLBACK_NUM", "");
			ast_log(LOG_DEBUG, "clear bland transter callback \n");
		}
	}

	//Start add by zyxia 2020.12.29
	if (strcmp((ast_channel_tech(chan)->type), "Local") == 0)
	{
		struct ast_channel *local_peer = ast_local_get_peer(chan);
		if (local_peer) {
			pbx_builtin_setvar_helper(local_peer, "DIALSTATUS", pa.status);
			ast_channel_unref(local_peer);
		}
	}
	//add by zyxia@grandstream.cn 2020.12.29
#endif

	ast_debug(1, "Exiting with DIALSTATUS=%s.\n", pa.status);

	if ((ast_test_flag64(peerflags, OPT_GO_ON)) && !ast_check_hangup(chan) && (res != AST_PBX_INCOMPLETE)) {
		if (!ast_tvzero(calldurationlimit))
			memset(ast_channel_whentohangup(chan), 0, sizeof(*ast_channel_whentohangup(chan)));
		res = 0;
	}

done:
	if (config.warning_sound) {
		ast_free((char *)config.warning_sound);
	}
	if (config.end_sound) {
		ast_free((char *)config.end_sound);
	}
	if (config.start_sound) {
		ast_free((char *)config.start_sound);
	}
	ast_ignore_cc(chan);
	return res;
}

static int dial_exec(struct ast_channel *chan, const char *data)
{
	struct ast_flags64 peerflags;

	memset(&peerflags, 0, sizeof(peerflags));

	return dial_exec_full(chan, data, &peerflags, NULL);
}

#ifdef GRANDSTREAM_NETWORKS
static int function_getlivechatserver(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
{
	int res = -1;
	RAII_VAR(char *, live_chat_server, NULL, ast_free);
	const char *caller_num = pbx_builtin_getvar_helper(chan, "CALLER_NUM");
	const char *dialed_exten = pbx_builtin_getvar_helper(chan, "DIALED_EXTEN");
	const char *from_exten = pbx_builtin_getvar_helper(chan, "FROM_EXTEN");
	const char *sipreferredbyhdr = pbx_builtin_getvar_helper(chan, "SIPREFERREDBYHDR");

	// transfer
	if (sipreferredbyhdr != NULL) {
		return res;
	}
	if (caller_num == NULL || dialed_exten == NULL || from_exten == NULL) {
		return res;
	}
	if (ast_channel_get_is_webrtc_trunk(chan) != 1) {
		return res;
	}
	if (!strcmp(from_exten, "service")) {
		return res;
	}
	buf[0] = '\0';
	struct ast_app *upload_live_chat_id_app = pbx_findapp("UPLOAD_LIVE_CHAT_ID");
	if (upload_live_chat_id_app) {
		pbx_exec(chan, upload_live_chat_id_app, S_OR(caller_num, ""));
	}
	for (int i = 0; i < 3; i++) {
		sleep(1);
		live_chat_server = ast_app_find_live_chat_server(caller_num);
		if (live_chat_server != NULL) {
			break;
		}
	}

	if (live_chat_server == NULL || strlen(live_chat_server) == 0) {
		ast_log(LOG_WARNING, "live chat server is not found.\n");
		ast_app_remove_live_chat_server(caller_num);
	}
	else {
		char read_buf[1024] = {0};
		char extension[256] = {0};

		ast_app_remove_live_chat_server(caller_num);
		ast_log(LOG_DEBUG, "live_chat_server:%s", live_chat_server);

		if (strcmp(live_chat_server, dialed_exten)) {
			char function[128] = {0};

			snprintf(function, sizeof(function), "PJSIP_DIAL_CONTACTS(%s)", live_chat_server);

			if (!ast_func_read(chan, function, read_buf, sizeof(read_buf))) {
				if (strlen(read_buf) > 0) {
					strncat(buf, read_buf, strlen(read_buf));
					res = 0;
				}
				else {
					snprintf(extension, sizeof(extension), "PJSIP/%s", live_chat_server);
					if (strlen(extension) < len - 2) {
						strncat(buf, extension, strlen(extension));
						res = 0;
					}
				}
			}
		}
	}

	return res;
}
#endif

static int retrydial_exec(struct ast_channel *chan, const char *data)
{
	char *parse;
	const char *context = NULL;
	int sleepms = 0, loops = 0, res = -1;
	struct ast_flags64 peerflags = { 0, };
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(announce);
		AST_APP_ARG(sleep);
		AST_APP_ARG(retries);
		AST_APP_ARG(dialdata);
	);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "RetryDial requires an argument!\n");
		return -1;
	}

	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);

	if (!ast_strlen_zero(args.sleep) && (sleepms = atoi(args.sleep)))
		sleepms *= 1000;

	if (!ast_strlen_zero(args.retries)) {
		loops = atoi(args.retries);
	}

	if (!args.dialdata) {
		ast_log(LOG_ERROR, "%s requires a 4th argument (dialdata)\n", rapp);
		goto done;
	}

	if (sleepms < 1000)
		sleepms = 10000;

	if (!loops)
		loops = -1; /* run forever */

	ast_channel_lock(chan);
	context = pbx_builtin_getvar_helper(chan, "EXITCONTEXT");
	context = !ast_strlen_zero(context) ? ast_strdupa(context) : NULL;
	ast_channel_unlock(chan);

	res = 0;
	while (loops) {
		int continue_exec;

		ast_channel_data_set(chan, "Retrying");
		if (ast_test_flag(ast_channel_flags(chan), AST_FLAG_MOH))
			ast_moh_stop(chan);

		res = dial_exec_full(chan, args.dialdata, &peerflags, &continue_exec);
		if (continue_exec)
			break;

		if (res == 0) {
			if (ast_test_flag64(&peerflags, OPT_DTMF_EXIT)) {
				if (!ast_strlen_zero(args.announce)) {
					if (ast_fileexists(args.announce, NULL, ast_channel_language(chan)) > 0) {
						if (!(res = ast_streamfile(chan, args.announce, ast_channel_language(chan))))
							ast_waitstream(chan, AST_DIGIT_ANY);
					} else
						ast_log(LOG_WARNING, "Announce file \"%s\" specified in Retrydial does not exist\n", args.announce);
				}
				if (!res && sleepms) {
					if (!ast_test_flag(ast_channel_flags(chan), AST_FLAG_MOH))
						ast_moh_start(chan, NULL, NULL);
					res = ast_waitfordigit(chan, sleepms);
				}
			} else {
				if (!ast_strlen_zero(args.announce)) {
					if (ast_fileexists(args.announce, NULL, ast_channel_language(chan)) > 0) {
						if (!(res = ast_streamfile(chan, args.announce, ast_channel_language(chan))))
							res = ast_waitstream(chan, "");
					} else
						ast_log(LOG_WARNING, "Announce file \"%s\" specified in Retrydial does not exist\n", args.announce);
				}
				if (sleepms) {
					if (!ast_test_flag(ast_channel_flags(chan), AST_FLAG_MOH))
						ast_moh_start(chan, NULL, NULL);
					if (!res)
						res = ast_waitfordigit(chan, sleepms);
				}
			}
		}

		if (res < 0 || res == AST_PBX_INCOMPLETE) {
			break;
		} else if (res > 0) { /* Trying to send the call elsewhere (1 digit ext) */
			if (onedigit_goto(chan, context, (char) res, 1)) {
				res = 0;
				break;
			}
		}
		loops--;
	}
	if (loops == 0)
		res = 0;
	else if (res == 1)
		res = 0;

	if (ast_test_flag(ast_channel_flags(chan), AST_FLAG_MOH))
		ast_moh_stop(chan);
 done:
	return res;
}

#ifdef GRANDSTREAM_NETWORKS
static struct ast_custom_function getlivechatserver_function = {
	.name = "GET_LIVE_CHAT_SERVER",
	.read = function_getlivechatserver,
};
#endif

static int unload_module(void)
{
	int res;

	res = ast_unregister_application(app);
	res |= ast_unregister_application(rapp);
#ifdef GRANDSTREAM_NETWORKS
	res |= ast_custom_function_unregister(&getlivechatserver_function);
#endif

	return res;
}

static int load_module(void)
{
	int res;

	res = ast_register_application_xml(app, dial_exec);
	res |= ast_register_application_xml(rapp, retrydial_exec);
#ifdef GRANDSTREAM_NETWORKS
	res |= ast_custom_function_register(&getlivechatserver_function);
#endif

	return res;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_DEFAULT, "Dialing Application",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.requires = "ccss",
);
