/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \author Joshua Colp <jcolp@digium.com>
 *
 * \brief PSJIP SIP Channel Driver
 *
 * \ingroup channel_drivers
 */

/*** MODULEINFO
    <depend>ucmsecret</depend>
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjlib.h>

#include "asterisk/lock.h"
#include "asterisk/channel.h"
#include "asterisk/module.h"
#include "asterisk/pbx.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/acl.h"
#include "asterisk/callerid.h"
#include "asterisk/file.h"
#include "asterisk/cli.h"
#include "asterisk/app.h"
#include "asterisk/musiconhold.h"
#include "asterisk/causes.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/dsp.h"
#include "asterisk/stasis_endpoints.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/indications.h"
#include "asterisk/format_cache.h"
#include "asterisk/translate.h"
#include "asterisk/threadstorage.h"
#include "asterisk/features_config.h"
#include "asterisk/pickup.h"
#include "asterisk/test.h"
#include "asterisk/message.h"

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/stream.h"

#include "pjsip/include/chan_pjsip.h"
#include "pjsip/include/dialplan_functions.h"
#include "pjsip/include/cli_functions.h"

#include "ucm_crypt.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/ccss.h"
#include "asterisk/app.h"
#include "asterisk/pjsip_cei.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/astdb.h"
#include "asterisk/security_events.h"
#include "asterisk/res_pjsip_bfcp_info.h"
#include "asterisk/options.h"
#include "asterisk/avs_controller.h"
#include "../apps/confbridge/include/conf_resource.h"
#include "../apps/confbridge/include/confbridge.h"
#include "asterisk/res_pjsip_camera_control.h"
#include "asterisk/switch_xml.h"
#include "asterisk/haserver.h"
#include "asterisk/stasis_bridges.h"
#include "asterisk/mixmonitor.h"
#include "asterisk/multi_media.h"
#include "asterisk/say.h"
#include "asterisk/ucm_dbi.h"
#ifdef HAVE_UNIXODBC
#include "asterisk/core_odbc.h"
#endif
#include <libgen.h>

AST_MODULE_LOG("pjsip");

/*** DOCUMENTATION
<application name="SIPSetCallType" language="en_US">
	<synopsis>
		Set Call Type Header.
	</synopsis>
	<syntax>
		<parameter name="type" required="true" />
	</syntax>
	<description>
		<para>Set Call Type header for SIP response</para>
	</description>
</application>
<application name="SIPSetMCBResp" language="en_US">
	<synopsis>
		Set MCB Header Response.
	</synopsis>
	<syntax>
		<parameter name="header" required="true" />
	</syntax>
	<description>
		<para>Set MCB header for SIP response</para>
	</description>
</application>
<application name="SIPSetConfFocus" language="en_US">
	<synopsis>
		Add isfocus in a contact parameter.
	</synopsis>
	<syntax>
		<parameter name="room_number" required="true" />
	</syntax>
	<description>
		<para>Add isfocus in a contact parameter during responding of an incoming INVITE</para>
	</description>
</application>
<application name="SIPSetContactUser" language="en_US">
	<synopsis>
		Set Contact URI Host User Part.
	</synopsis>
	<syntax>
		<parameter name="user" required="true" />
	</syntax>
	<description>
		<para>Set Contact URI Host User Part during responding of an incoming INVITE</para>
	</description>
</application>
<application name="SIPTerminatedNotify" language="en_US">
	<synopsis>
		Send SIP NOTIFY message with terminated state.
	</synopsis>
	<syntax>
		<parameter name="conference" required="true" />
	</syntax>
	<description>
		<para>Send SIP NOTIFY message with terminated state and clean up SIP subscription.</para>
	</description>
</application>
<application name="SIPCEIJoined" language="en_US">
	<synopsis>
		Given channel joined to a conference with valid authentication.
	</synopsis>
	<syntax>
		<parameter name="conference" required="true" />
	</syntax>
	<description>
		<para>Given channel joined to a conference with valid authentication.</para>
	</description>
</application>
<application name="SIPSubscribeExtMCB" language="en_US">
	<synopsis>
		Subscribe an external MCB.
	</synopsis>
	<syntax>
		<parameter name="local_conference" required="true" />
		<parameter name="local_conference_id" required="true" />
		<parameter name="remote_conference" required="true" />
		<parameter name="remote_conference_id" required="true" />
		<parameter name="remote_channel" required="true" />
		<parameter name="endpoint_name" required="true" />
	</syntax>
	<description>
		<para>Subscribe to and handle subscription process of an external MCB</para>
	</description>
</application>
<application name="SIPUnsubscribeExtMCB" language="en_US">
	<synopsis>
		Unsubscribe an external MCB.
	</synopsis>
	<syntax>
		<parameter name="local_conference_id" required="true" />
		<parameter name="remote_conference_id" required="true" />
		<parameter name="endpoint_name" required="true" />
	</syntax>
	<description>
		<para>Unsubscribe to and handle unsubscription process of an external MCB</para>
	</description>
</application>
<application name="SIPSubscribeMCB" language="en_US">
	<synopsis>
		Subscribe an internal MCB.
	</synopsis>
	<syntax>
		<parameter name="local_conference" required="true" />
		<parameter name="local_conference_id" required="true" />
		<parameter name="remote_conference" required="true" />
		<parameter name="remote_conference_id" required="true" />
	</syntax>
	<description>
		<para>Subscribe to and handle subscription process of an internal MCB</para>
	</description>
</application>
<application name="SIPUnsubscribeMCB" language="en_US">
	<synopsis>
		Unsubscribe an internal MCB.
	</synopsis>
	<syntax>
		<parameter name="local_conference_id" required="true" />
		<parameter name="remote_conference_id" required="true" />
	</syntax>
	<description>
		<para>Unsubscribe to and handle unsubscription process of an internal MCB</para>
	</description>
</application>
***/
#endif

AST_THREADSTORAGE(uniqueid_threadbuf);
#define UNIQUEID_BUFSIZE 256

static const char channel_type[] = "PJSIP";

#ifdef GRANDSTREAM_NETWORKS
static int globalEnabledCEI = 0;
static struct stasis_message_router *cei_toggle_enable_router = NULL;
static struct ast_taskprocessor *emergency_mail_tps;

#define NAT_CALL_THRESHOLD_PERCENTAGE 80

static char *app_sipsetcalltype = "SIPSetCallType";
static char *app_sipsetmcbresp = "SIPSetMCBResp";
static char *app_sipsetconffocus = "SIPSetConfFocus";
static char *app_sipsetcontactuser = "SIPSetContactUser";
static char *app_sipterminatednotify = "SIPTerminatedNotify";
static char *app_sipceijoined = "SIPCEIJoined";
// Internal MCB
 static char *app_sipsubscribemcb = "SIPSubscribeMCB";
 static char *app_sipunsubscribemcb = "SIPUnsubscribeMCB";
// External MCB
static char *app_sipsubscribeextmcb = "SIPSubscribeExtMCB";
static char *app_sipunsubscribeextmcb = "SIPUnsubscribeExtMCB";
static int sip_setcalltype(struct ast_channel *chan, const char *data);
static int sip_setmcbresp(struct ast_channel *chan, const char *data);
static int sip_set_conf_focus(struct ast_channel *chan, const char *data);
static int sip_set_contact_user(struct ast_channel *chan, const char *data);
static int sip_terminated_notify(struct ast_channel *chan, const char *data);
static int sip_cei_joined(struct ast_channel *chan, const char *data);
static int sip_external_mcb_sub(struct ast_channel *chan, const char *data);
static int sip_external_mcb_unsub(struct ast_channel *chan, const char *data);
static int sip_internal_mcb_sub(struct ast_channel *chan, const char *data);
static int sip_internal_mcb_unsub(struct ast_channel *chan, const char *data);
static unsigned char judge_chan_type_is_sip(const struct ast_channel *channel);
static void update_sdp_for_local(struct ast_channel* chan, struct ast_channel* peer);
static int check_channel_is_video_call(struct ast_channel *chan);
static void disable_main_video(struct ast_channel *chan,int ignore_video_caps);
static int set_pending_media_stream_by_local_session(struct ast_sip_session *local_session,struct ast_sip_session *remote_session);
static void update_auto_record_sdp(struct ast_channel* chan);
static void change_media_direction(pj_pool_t *pool, pjmedia_sdp_media *media, char *dir);
static void change_video_media_direction(struct ast_channel *chan, enum ast_media_sub_type sub_type,char* dir);
static int create_new_main_video_stream_append(struct ast_sip_session *session, struct ast_stream_topology *topology,enum ast_stream_media_direction media_direction);
static void change_session_media_direction(struct ast_sip_session *session,enum ast_media_sub_type sub_type ,enum ast_stream_media_direction media_direction);
static void get_peer_trunk_info(struct ast_sip_session *session, struct ast_sip_session *peer_session);
static int sip_exteral_trunk_conf_sub(struct ast_sip_session *session, const char *local_virtual_conf);
static int sip_exteral_trunk_conf_unsub(struct ast_sip_endpoint *endpoint ,struct cei_external_trunk_conf_peer* etc_peer);
static void check_if_outbound_to_trunk_conf(struct ast_sip_session *session,struct pjsip_rx_data *rdata);
static void send_trunk_conf_cei_subscribe(struct ast_sip_session *session);
static void add_isfocus_for_trunk_conf(struct ast_sip_session *session, pjsip_tx_data *tdata);
static int process_trunk_info_for_callflip(struct ast_channel *chan,struct ast_channel *new_chan);
//static void remove_imageattr_for_callflip(struct ast_sip_session *session);
static int chan_pjsip_cc_callback(struct ast_channel *inbound, const char *dest, ast_cc_callback_fn callback);
static int process_real_answered_control_frame(void *session);
//static void disable_slides_video(struct ast_channel *chan,int ignore_video_caps);
/*Start added by zyxia for Bug 256745,dont wait the prompt play finished*/
static int process_play_prompt_to_caller(void* data);
static int process_parkingstate_aor(void *obj, void *arg, int flags);
static int process_open_ipc_video(void* data);

struct main_video_data {
	struct ast_sip_session *session;
	const void             *data;
};
static void main_video_dtor(void *data)
{
	struct main_video_data *main_video = data;
	ao2_cleanup(main_video->session);
}
static int process_update_main_video_control_frame(void *data);
static int change_active_main_video_direction(struct ast_sip_session *session,enum ast_stream_media_direction stream_direction);
//static int change_slides_video_direction_by_peer_session(struct ast_sip_session *local_session,struct ast_sip_session *remote_session);
#endif

static unsigned int chan_idx;
char first_delay_foripv6_recovery = 0;
static void chan_pjsip_pvt_dtor(void *obj)
{
#ifdef GRANDSTREAM_NETWORKS
	struct chan_pjsip_pvt *pvt = obj;

    if (pvt->call_type != NULL) {
        ast_free(pvt->call_type);
        pvt->call_type = NULL;
    }

	if (pvt->mcb_request != NULL) {
		ast_free(pvt->mcb_request);
		pvt->mcb_request = NULL;
	}

	if (pvt->mcb_response != NULL) {
		ast_free(pvt->mcb_response);
		pvt->mcb_response = NULL;
	}

    if (pvt->contact_user != NULL) {
		ast_free(pvt->contact_user);
		pvt->contact_user = NULL;
	}
#endif
}

/* \brief Asterisk core interaction functions */
static struct ast_channel *chan_pjsip_request(const char *type, struct ast_format_cap *cap, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *data, int *cause);
static struct ast_channel *chan_pjsip_request_with_stream_topology(const char *type,
	struct ast_stream_topology *topology, const struct ast_assigned_ids *assignedids,
	const struct ast_channel *requestor, const char *data, int *cause);
static int chan_pjsip_sendtext_data(struct ast_channel *ast, struct ast_msg_data *msg);
static int chan_pjsip_sendtext(struct ast_channel *ast, const char *text);
static int chan_pjsip_digit_begin(struct ast_channel *ast, char digit);
static int chan_pjsip_digit_end(struct ast_channel *ast, char digit, unsigned int duration);
static int chan_pjsip_call(struct ast_channel *ast, const char *dest, int timeout);
static int chan_pjsip_hangup(struct ast_channel *ast);
static int chan_pjsip_answer(struct ast_channel *ast);
#ifdef GRANDSTREAM_NETWORKS

struct ast_gs_feature_map gs_feature_map[GS_FEATURE_MAP_COUNT] = {
    { "blindxfer", "#", GS_FEATURE_DISABLE },
    { "atxfer",    "",  GS_FEATURE_DISABLE },
    { "parkcall",  "",  GS_FEATURE_DISABLE },
    { "automixmon", "",  GS_FEATURE_DISABLE }
};

void add_feature_map_header(pjsip_tx_data *tdata, int outgoing_call, struct ast_sip_session *session);
static int get_feature_code(int reload);
// CEI
//static void clear_peer_conferences(struct ast_sip_endpoint *endpoint);
static int add_peer_conference(const char *channel, struct ast_sip_endpoint *endpoint, const char *conference,enum ast_conference_type conf_type);
static void publish_cei_unsub(const char *conference, const char *channel, const char *cid_number);
static void publish_cei_joined(const char *conference, const char *channel);
void ast_sip_report_emergency_outbound(const char *endpoint_id, const char *emergency_num); /* add by jlhu for trunk hunt */
int chan_pjsip_groupfull(struct ast_channel *requestor, const char *data);
#endif
static struct ast_frame *chan_pjsip_read_stream(struct ast_channel *ast);
static int chan_pjsip_write(struct ast_channel *ast, struct ast_frame *f);
static int chan_pjsip_write_stream(struct ast_channel *ast, int stream_num, struct ast_frame *f);
static int chan_pjsip_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen);
static int chan_pjsip_transfer(struct ast_channel *ast, const char *target);
static int chan_pjsip_fixup(struct ast_channel *oldchan, struct ast_channel *newchan);
static int chan_pjsip_devicestate(const char *data);
static int chan_pjsip_queryoption(struct ast_channel *ast, int option, void *data, int *datalen);
static const char *chan_pjsip_get_uniqueid(struct ast_channel *ast);
#ifdef GRANDSTREAM_NETWORKS
static int chan_pjsip_special_answer(struct ast_channel *chan, struct ast_channel *peer);
static int send_reinvite_request(void *data);
static int send_reinvite_response(void *data);
static int send_reinvite_failed_response(void *data);
static void send_reinvite_request_to_callee(struct ast_sip_session *session, const struct ast_sip_reinvite_data *reinvite_data);
static void send_reinvite_response_to_caller(struct ast_sip_session *session, const struct ast_sip_reinvite_data *reinvite_data);
static void chan_pjsip_modify_answer_sdp(struct ast_sip_session *session, struct ast_channel *chan, struct ast_channel *peer);
static unsigned char judge_if_need_send_reinvite(struct ast_sip_session *session, const struct ast_sip_reinvite_data *req_data);
static void get_peer_param_for_init_invite(struct ast_sip_session *session, const struct ast_channel *requestor);
static enum ast_terminal_type chan_pjsip_get_terminal_type(struct ast_channel *chan);
static void bfcp_create_params(struct ast_sip_session *session);
static void bfcp_create_params_for_sfu_invite(struct ast_sip_session *session);
static void update_msid_for_change_host(struct ast_sip_session *session, char *host_msid);
static void send_reinivte_for_change_ssrc(struct ast_sip_session *session);
static unsigned char judge_if_has_slides_video(const struct ast_sip_session *session);
static int chan_pjsip_get_if_negotiate_state_done(struct ast_channel *chan);
static unsigned int chan_pjsip_get_delayed_reinvite_count(struct ast_channel *chan);
static void chan_pjsip_inc_delayed_reinvite_count(struct ast_channel *chan);
static int chan_pjsip_get_direct_enable(struct ast_channel* chan);
static void chan_pjsip_release_video_resource(struct ast_channel *chan);
static void update_sfu_params_for_join_conference(struct ast_sip_session *session);
static void notify_released_bfcp_for_transfer(struct ast_sip_session *session);
static void notify_bfcp_for_refer_transfer(struct ast_sip_session *session);
static void disable_media_except_audio(struct ast_channel *chan);
static void set_pai_payload(struct ast_channel *channel, const void *data, size_t datalen);
static void process_topology_control_frame(struct ast_sip_session *session, struct ast_sip_session *peer_session);
static void send_digit_with_both_rfc4733_and_info(struct ast_sip_session *session,
	char digit, unsigned int duration, struct ast_sip_session_media *media);
static void process_reinvite_req_control_frame(struct ast_sip_session *session, const void *data);
static void process_reinvite_rsp_control_frame(struct ast_sip_session *session, const void *data);
static void process_bfcp_control_frame(struct ast_sip_session *session, const void *data);
static int process_update_turn_session_count_frame(struct ast_sip_session *session, const void *data, size_t datalen);
static void send_response_control_frame_to_peer(struct ast_sip_session *session);
//static void send_reinvite_for_change_host(struct ast_sip_session *session, char *host_msid);
static int process_camera_control_frame(struct ast_sip_session *session, const void *data, size_t datalen);
static void process_camera_change_frame(struct ast_sip_session *session, char *host_msid);
static unsigned int chan_pjsip_get_media_turn_relay(struct ast_channel *chan);
static void process_notify_cei_status_frame(struct ast_sip_session *session, const void *data, size_t datalen);
static void set_local_gds_call_from_peer(struct ast_sip_session* session, struct ast_sip_session* peer_session, int is_caller);
static void set_callee_gds_call(struct ast_sip_session* session, pjsip_rx_data* rdata);
static void set_local_gsc_call_from_peer(struct ast_sip_session* session, struct ast_sip_session* peer_session, int is_caller);
static void set_callee_gsc_call(struct ast_sip_session* session, pjsip_rx_data* rdata);
static void set_local_hikvision_call_from_peer(struct ast_sip_session* session, struct ast_sip_session* peer_session, int is_caller);
static int send_update_record_attr(void *data);
static int chan_pjsip_get_session_ua(struct ast_channel* chan, char* buf, size_t size);
static unsigned int chan_pjsip_get_gswave_bundle_enabled(struct ast_channel *chan);
static unsigned int chan_pjsip_get_mobile_bundle_enabled(struct ast_channel *chan);
static void process_join_sfu_control_frame(struct ast_sip_session *session);
static void notify_info_data_destroy(void* obj);
static int generate_notify_and_send(void* data);
static struct conf_notify_info_data* notify_info_data_alloc2(struct ast_sip_session* session, struct conf_notify_switch_data* data);
static void send_notify_to_checkout_conf(struct ast_sip_session* session, const void* data, int datalen);
static int chan_pjsip_get_participant_if_negotiate_state_done(struct ast_channel *chan);
static unsigned int chan_pjsip_get_delayed_notify_count(struct ast_channel *chan);
static void chan_pjsip_inc_delayed_notify_count(struct ast_channel *chan);
static void process_reinvite_failed_frame(struct ast_sip_session *session, const void *data);
static void process_wait_join_conference_frame(struct ast_sip_session *session);
static void process_join_nway_conf_frame(struct ast_sip_session *session);
static void send_avs_slides_video_control(struct ast_sip_session *session, enum avs_ctrl_opt opt);
static void hangup_some_associated_call(struct ast_sip_session* session);
static char *chan_pjsip_get_slides_video_local_msid(struct ast_channel *chan);
static void update_avs_slides_video_msid(struct ast_sip_session *session, const char *msid);
static int chan_pjsip_append_sip_header(struct ast_channel *chan, void *xml_cdr, void *msg, int *offset);
static void chan_pjsip_decrease_turn_session_count(struct ast_channel *chan,unsigned int decrease_count);
static int chan_pjsip_get_enable_mcu(struct ast_channel *chan);
static void process_conf_timeout_notify_frame(struct ast_sip_session* session, const void* data, int datalen);

static int chan_pjsip_recovery_throw_pbx(struct ast_channel *channel);
static char *cli_pjsip_call_recovery(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
static void process_close_camera_frame(struct ast_sip_session *session);
static void set_media_encryption_from_restore_xml(struct ast_channel_media_encryption *encryption, switch_xml_t restore_xml);
static void process_open_slides_video_frame(struct ast_sip_session *session);
static void process_stop_slides_video_frame(struct ast_sip_session *session);
static int process_gs_state_frame(struct ast_sip_session *session, const void *data, size_t datalen);
static int process_ami_hold_frame(struct ast_sip_session *session);
static int process_ami_unhold_frame(struct ast_sip_session *session);
//static void send_reinvite_control_slides(struct ast_sip_session *local_session,struct ast_sip_session *session);
static int send_reinvite_for_session_refresh(void *data);
static void send_reinvite_for_neg_slides_video(struct ast_sip_session* session);
static int chan_pjsip_get_limit_time_reached(struct ast_channel *chan);
static void process_open_focus_video_notify_frame(struct ast_sip_session *session);
static void process_close_focus_video_notify_frame(struct ast_sip_session *session);
static void process_disable_focus_video_notify_frame(struct ast_sip_session *session);
static int process_conf_video_record_start_frame(struct ast_sip_session *session);
static void process_stop_record_frame(struct ast_sip_session *session, const void *data, size_t datalen);
static void process_stop_slides_video_by_host_frame(struct ast_sip_session *session, const void *data);
static void send_reinvite_close_camera(struct ast_sip_session *session, enum ast_stream_media_direction direction);
static void process_reinvite_req_failed_frame(struct ast_sip_session *session);
static void process_reinvite_rsp_failed_frame(struct ast_sip_session *session);
static void process_reinvite_rsp_again_frame(struct ast_sip_session *session);
static void process_queue_peer_session_failed_frame(struct ast_sip_session *session);
static void process_wave_released_hd_resolution_frame(struct ast_sip_session *session);
static int process_notify_trunk_info_frame(struct ast_sip_session *session, const void *data, size_t datalen);
static void notify_trunk_info_for_transfer(struct ast_sip_session *session);
static void process_update_candidates_control_frame(struct ast_sip_session *session, const void *data);
static int process_change_to_trunk_conf_frame(struct ast_sip_session *session, char* trunk_conf_num);
static void send_reinvite_to_peer(struct ast_sip_session *session);
static int change_main_video_direction_by_local_session(struct ast_sip_session *local_session,struct ast_sip_session *remote_session);
static int chan_pjsip_get_sca_sharedline(struct ast_channel* chan, char* buf, size_t size);
static void process_conf_end_control_frame(struct ast_sip_session *session);
static int process_callflip_reinvite_frame(struct ast_sip_session *session, const void *data, size_t datalen);

struct pjmedia_sdp_neg
{
    pjmedia_sdp_neg_state state;	    /**< Negotiator state.	     */
    pj_bool_t		  prefer_remote_codec_order;
    pj_bool_t             answer_with_multiple_codecs;
    pj_bool_t		  has_remote_answer;
    pj_bool_t		  answer_was_remote;

    pjmedia_sdp_session	*initial_sdp,	    /**< Initial local SDP	     */
			*initial_sdp_tmp,   /**< Temporary initial local SDP */
			*active_local_sdp,  /**< Currently active local SDP. */
			*active_remote_sdp, /**< Currently active remote's.  */
			*neg_local_sdp,	    /**< Temporary local SDP.	     */
			*neg_remote_sdp;    /**< Temporary remote SDP.	     */
};

struct ast_format {
	/*! Name of the format */
	const char *name;
	/*! \brief Pointer to the codec in use for this format */
	struct ast_codec *codec;
	/*! \brief Attribute specific data, implementation specific */
	void *attribute_data;
	/*! \brief Pointer to the optional format interface */
	const struct ast_format_interface *interface;
};

struct format_cap_framed {
	/*! \brief A pointer to the format */
	struct ast_format *format;
	/*! \brief The format framing size */
	unsigned int framing;
	/*! \brief Linked list information */
	AST_LIST_ENTRY(format_cap_framed) entry;
};

struct ast_format_cap {
	/*! \brief Vector of formats, indexed using the codec identifier */
	AST_VECTOR(, struct format_cap_framed_list) formats;
	/*! \brief Vector of formats, added in preference order */
	AST_VECTOR(, struct format_cap_framed *) preference_order;
	/*! \brief Global framing size, applies to all formats if no framing present on format */
	unsigned int framing;
};

struct ast_sip_special_answer_data
{
	struct ast_channel *chan;
	struct ast_channel *peer;
};

struct camera_control_info_data {
	struct ast_sip_session *session;
	struct ast_camera_control_info *camera_control_info;
};

struct conf_notify_info_data {
	struct ast_sip_session *session;
	struct conf_notify_info *notify_info;
	char conf_id[MAX_CONF_NAME];
	int is_video_conf;
	int is_admin;
	int enabled_cei;
	int event_type;      /*0 means nway type, 1 means call flip type*/
};

struct general_info_data {
	struct ast_sip_session* session;
	char* message;
};

struct conf_schedule_wait_notify_info_data {
	struct ast_sip_session *session;
	struct conf_wait_notify_data * wait_notify_info;
};

struct trunk_info_data {
	struct ast_sip_session *session;
	struct ast_trunk_info trunk_info;
};
#endif

/*! \brief PBX interface structure for channel registration */
struct ast_channel_tech chan_pjsip_tech = {
	.type = channel_type,
	.description = "PJSIP Channel Driver",
	.requester = chan_pjsip_request,
	.requester_with_stream_topology = chan_pjsip_request_with_stream_topology,
	.send_text = chan_pjsip_sendtext,
	.send_text_data = chan_pjsip_sendtext_data,
	.send_digit_begin = chan_pjsip_digit_begin,
	.send_digit_end = chan_pjsip_digit_end,
	.call = chan_pjsip_call,
	.hangup = chan_pjsip_hangup,
	.answer = chan_pjsip_answer,
#ifdef GRANDSTREAM_NETWORKS
	.special_answer = chan_pjsip_special_answer,
	.get_terminal_type = chan_pjsip_get_terminal_type,
	.get_if_negotiate_state_done = chan_pjsip_get_if_negotiate_state_done,
	.get_delayed_reinvite_count = chan_pjsip_get_delayed_reinvite_count,
	.inc_delayed_reinvite_count = chan_pjsip_inc_delayed_reinvite_count,
	.get_direct_enable = chan_pjsip_get_direct_enable,
	.release_video_resource = chan_pjsip_release_video_resource,
	.get_media_turn_relay = chan_pjsip_get_media_turn_relay,
	.get_session_ua = chan_pjsip_get_session_ua,
	.get_gswave_bundle_enabled = chan_pjsip_get_gswave_bundle_enabled,
	.get_mobile_bundle_enabled = chan_pjsip_get_mobile_bundle_enabled,
	.get_participant_if_negotiate_state_done = chan_pjsip_get_participant_if_negotiate_state_done,
	.get_delayed_notify_count = chan_pjsip_get_delayed_notify_count,
	.inc_delayed_notify_count = chan_pjsip_inc_delayed_notify_count,
	.get_slides_video_local_msid = chan_pjsip_get_slides_video_local_msid,
	.append_sip_header = chan_pjsip_append_sip_header, /* for ha recovery */
	.decrease_turn_session_count  = chan_pjsip_decrease_turn_session_count, /*start added by zyxia for bug202204,2021.8.27*/
	.get_enable_mcu  = chan_pjsip_get_enable_mcu,
	.get_limit_time_reached  = chan_pjsip_get_limit_time_reached,
	.get_sca_sharedline  = chan_pjsip_get_sca_sharedline,
    .cc_callback = chan_pjsip_cc_callback,   /* Added by yfyang, for Bug 221341 */
#endif
	.read_stream = chan_pjsip_read_stream,
	.write = chan_pjsip_write,
	.write_stream = chan_pjsip_write_stream,
	.exception = chan_pjsip_read_stream,
	.indicate = chan_pjsip_indicate,
	.transfer = chan_pjsip_transfer,
	.fixup = chan_pjsip_fixup,
	.devicestate = chan_pjsip_devicestate,
	.queryoption = chan_pjsip_queryoption,
	.func_channel_read = pjsip_acf_channel_read,
	.get_pvt_uniqueid = chan_pjsip_get_uniqueid,
	.properties = AST_CHAN_TP_WANTSJITTER | AST_CHAN_TP_CREATESJITTER | AST_CHAN_TP_SEND_TEXT_DATA
};

/*! \brief SIP session interaction functions */
static void chan_pjsip_session_begin(struct ast_sip_session *session);
static void chan_pjsip_session_end(struct ast_sip_session *session);
static int chan_pjsip_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata);
static void chan_pjsip_incoming_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata);

/*! \brief SIP session supplement structure */
static struct ast_sip_session_supplement chan_pjsip_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL,
	.session_begin = chan_pjsip_session_begin,
	.session_end = chan_pjsip_session_end,
	.incoming_request = chan_pjsip_incoming_request,
	/* It is important that this supplement runs after media has been negotiated */
	.response_priority = AST_SIP_SESSION_AFTER_MEDIA,
};

#ifdef GRANDSTREAM_NETWORKS
static int chan_pjsip_incoming_request_fastprov(struct ast_sip_session *session, struct pjsip_rx_data *rdata);
static struct ast_sip_session_supplement chan_pjsip_supplement_fastprov = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL + 1,
	.incoming_request = chan_pjsip_incoming_request_fastprov,
	/* It is important that this supplement runs after media has been negotiated */
	.response_priority = AST_SIP_SESSION_AFTER_MEDIA,
};
#endif

/*! \brief SIP session supplement structure just for responses */
static struct ast_sip_session_supplement chan_pjsip_supplement_response = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL,
	.incoming_response = chan_pjsip_incoming_response,
#ifdef GRANDSTREAM_NETWORKS
	.response_priority = AST_SIP_SESSION_AFTER_MEDIA,
#else
	.response_priority = AST_SIP_SESSION_BEFORE_MEDIA | AST_SIP_SESSION_AFTER_MEDIA,
#endif
};

static int chan_pjsip_incoming_ack(struct ast_sip_session *session, struct pjsip_rx_data *rdata);

static struct ast_sip_session_supplement chan_pjsip_ack_supplement = {
	.method = "ACK",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_CHANNEL,
	.incoming_request = chan_pjsip_incoming_ack,
};

#ifdef GRANDSTREAM_NETWORKS

#ifdef HAVE_UNIXODBC
struct generic_prepare_struct {
	char *sql;
	int argc;
	char **argv;
};
#endif

static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct generic_prepare_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}

static int get_emergency_email(const char *match)
{
	char cmd[256] = {0};
	int ret = 0;

#ifdef USE_DB_MYSQL

#ifdef HAVE_UNIXODBC
	struct odbc_obj *obj = NULL;
	SQLHSTMT stmt = NULL;
	char rowdata[20];
	int res = -1;
	struct generic_prepare_struct gps = { .sql = cmd };

	obj = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);
	if (!obj) {
		ast_log(LOG_ERROR, "Failed to obtain database object for '%s'!\n", UCM_CONFIG_DB);
		return ret;
	}
#else
	dbi_inst pInst =NULL;
	dbi_conn connection;
	connection = ast_dbi_connect_db(UCM_CONFIG_DB, &pInst);
	if(connection == NULL )
	{
		ast_log(LOG_ERROR, "get_queue_member_count Connect ucm_config failed\n");
		return ret;
	}
#endif

#else
	sqlite3* ucm_config_db = NULL;
	char **db_result = NULL;
	int row = 0, col = 0;
	int err;

	err = sqlite3_open("/tmp/ucm_config.db",&ucm_config_db);
	if(err != SQLITE_OK){
		ast_log(LOG_WARNING, "sqlite3_open failed\n");
		return ret;
	}
	sqlite3_busy_timeout(ucm_config_db,3000);
#endif
	memset(cmd,0,sizeof(cmd));
	snprintf(cmd,sizeof(cmd)-1,"SELECT email FROM emergency_outbound WHERE `match` ='%s'", match);
	// ast_log(LOG_DEBUG, "cmd is %s\n", cmd);

#ifdef USE_DB_MYSQL

#ifdef HAVE_UNIXODBC
	stmt = ast_core_odbc_prepare_and_execute(obj, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_ERROR, "SQL Execute error!\n[%s]\n\n", cmd);
		goto bail;
	}
	res = SQLFetch(stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(LOG_ERROR, "SQL Fetch error!\n[%s]\n\n", cmd);
		goto bail_with_handle;
	}
	res = SQLGetData(stmt, 1, SQL_CHAR, rowdata, sizeof(rowdata), NULL);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(LOG_ERROR, "SQL Get Data error!\n[%s]\n\n", cmd);
		goto bail_with_handle;
	}
	if(strlen(rowdata) > 0 && rowdata[0] != ' ')
	{
		ret = 1;
	}

bail_with_handle:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);
bail:
	ast_core_odbc_release_obj(obj);

#else
	queue_member_count = ast_dbi_getOneIntField(connection, cmd, "cnt");
	if (queue_member_count < 0) {
		return -1;
	}
	ast_dbi_disconnect_db(connection, &pInst);
#endif

#else
	err = sqlite3_get_table(ucm_config_db,cmd,&db_result,&row,&col,NULL);
	if( err != SQLITE_OK ){
		ast_log(LOG_WARNING, "sqlite3_get_table failed\n");
		sqlite3_close(ucm_config_db);
		return -1;
	}

	if (db_result) {
		queue_member_count = atoi(*(db_result + 1));
		sqlite3_free_table(db_result);
		db_result = NULL;
	}

	sqlite3_close(ucm_config_db);
#endif
	return ret;
}
#endif

/*! \brief Function called by RTP engine to get local audio RTP peer */
static enum ast_rtp_glue_result chan_pjsip_get_rtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_endpoint *endpoint;
	struct ast_datastore *datastore;
	struct ast_sip_session_media *media;

	if (!channel || !channel->session) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	/* XXX Getting the first RTP instance for direct media related stuff seems just
	 * absolutely wrong. But the native RTP bridge knows no other method than single-stream
	 * for direct media. So this is the best we can do.
	 */
	media = channel->session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO];
	if (!media || !media->rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	datastore = ast_sip_session_get_datastore(channel->session, "t38");
	if (datastore) {
		ao2_ref(datastore, -1);
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = media->rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	if (endpoint->media.direct_media.enabled) {
		return AST_RTP_GLUE_RESULT_REMOTE;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}

/*! \brief Function called by RTP engine to get local video RTP peer */
static enum ast_rtp_glue_result chan_pjsip_get_vrtp_peer(struct ast_channel *chan, struct ast_rtp_instance **instance)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_endpoint *endpoint;
	struct ast_sip_session_media *media;

	if (!channel || !channel->session) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	media = channel->session->active_media_state->default_session[AST_MEDIA_TYPE_VIDEO];
	if (!media || !media->rtp) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	endpoint = channel->session->endpoint;

	*instance = media->rtp;
	ao2_ref(*instance, +1);

	ast_assert(endpoint != NULL);
	if (endpoint->media.rtp.encryption != AST_SIP_MEDIA_ENCRYPT_NONE) {
		return AST_RTP_GLUE_RESULT_FORBID;
	}

	return AST_RTP_GLUE_RESULT_LOCAL;
}

/*! \brief Function called by RTP engine to get peer capabilities */
static void chan_pjsip_get_codec(struct ast_channel *chan, struct ast_format_cap *result)
{
	ast_format_cap_append_from_cap(result, ast_channel_nativeformats(chan), AST_MEDIA_TYPE_UNKNOWN);
}

/*! \brief Destructor function for \ref transport_info_data */
static void transport_info_destroy(void *obj)
{
	struct transport_info_data *data = obj;
	ast_free(data);
}

/*! \brief Datastore used to store local/remote addresses for the
 * INVITE request that created the PJSIP channel */
static struct ast_datastore_info transport_info = {
	.type = "chan_pjsip_transport_info",
	.destroy = transport_info_destroy,
};

static struct ast_datastore_info direct_media_mitigation_info = { };

static int direct_media_mitigate_glare(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);

	if (session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_NONE) {
		return 0;
	}

	datastore = ast_sip_session_get_datastore(session, "direct_media_glare_mitigation");
	if (!datastore) {
		return 0;
	}

	/* Removing the datastore ensures we won't try to mitigate glare on subsequent reinvites */
	ast_sip_session_remove_datastore(session, "direct_media_glare_mitigation");

	if ((session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_OUTGOING &&
			session->inv_session->role == PJSIP_ROLE_UAC) ||
			(session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_INCOMING &&
			session->inv_session->role == PJSIP_ROLE_UAS)) {
		return 1;
	}

	return 0;
}

/*! \brief Helper function to find the position for RTCP */
static int rtp_find_rtcp_fd_position(struct ast_sip_session *session, struct ast_rtp_instance *rtp)
{
	int index;

	for (index = 0; index < AST_VECTOR_SIZE(&session->active_media_state->read_callbacks); ++index) {
		struct ast_sip_session_media_read_callback_state *callback_state =
			AST_VECTOR_GET_ADDR(&session->active_media_state->read_callbacks, index);

		if (callback_state->fd != ast_rtp_instance_fd(rtp, 1)) {
			continue;
		}

		return index;
	}

	return -1;
}

/*!
 * \pre chan is locked
 */
static int check_for_rtp_changes(struct ast_channel *chan, struct ast_rtp_instance *rtp,
		struct ast_sip_session_media *media, struct ast_sip_session *session)
{
	int changed = 0, position = -1;

	if (media->rtp) {
		position = rtp_find_rtcp_fd_position(session, media->rtp);
	}

	if (rtp) {
		changed = ast_rtp_instance_get_and_cmp_remote_address(rtp, &media->direct_media_addr);
		if (media->rtp) {
			if (position != -1) {
				ast_channel_set_fd(chan, position + AST_EXTENDED_FDS, -1);
			}
			ast_rtp_instance_set_prop(media->rtp, AST_RTP_PROPERTY_RTCP, 0);
		}
	} else if (!ast_sockaddr_isnull(&media->direct_media_addr)){
		ast_sockaddr_setnull(&media->direct_media_addr);
		changed = 1;
		if (media->rtp) {
			ast_rtp_instance_set_prop(media->rtp, AST_RTP_PROPERTY_RTCP, 1);
			if (position != -1) {
				ast_channel_set_fd(chan, position + AST_EXTENDED_FDS, ast_rtp_instance_fd(media->rtp, 1));
			}
		}
	}

	return changed;
}

struct rtp_direct_media_data {
	struct ast_channel *chan;
	struct ast_rtp_instance *rtp;
	struct ast_rtp_instance *vrtp;
	struct ast_format_cap *cap;
	struct ast_sip_session *session;
};

static void rtp_direct_media_data_destroy(void *data)
{
	struct rtp_direct_media_data *cdata = data;

	ao2_cleanup(cdata->session);
	ao2_cleanup(cdata->cap);
	ao2_cleanup(cdata->vrtp);
	ao2_cleanup(cdata->rtp);
	ao2_cleanup(cdata->chan);
}

static struct rtp_direct_media_data *rtp_direct_media_data_create(
	struct ast_channel *chan, struct ast_rtp_instance *rtp, struct ast_rtp_instance *vrtp,
	const struct ast_format_cap *cap, struct ast_sip_session *session)
{
	struct rtp_direct_media_data *cdata = ao2_alloc(sizeof(*cdata), rtp_direct_media_data_destroy);

	if (!cdata) {
		return NULL;
	}

	cdata->chan = ao2_bump(chan);
	cdata->rtp = ao2_bump(rtp);
	cdata->vrtp = ao2_bump(vrtp);
	cdata->cap = ao2_bump((struct ast_format_cap *)cap);
	cdata->session = ao2_bump(session);

	return cdata;
}

static int send_direct_media_request(void *data)
{
	struct rtp_direct_media_data *cdata = data;
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(cdata->chan);
	struct ast_sip_session *session;
	int changed = 0;
	int res = 0;

	/* XXX In an ideal world each media stream would be direct, but for now preserve behavior
	 * and connect only the default media sessions for audio and video.
	 */

	/* The channel needs to be locked when checking for RTP changes.
	 * Otherwise, we could end up destroying an underlying RTCP structure
	 * at the same time that the channel thread is attempting to read RTCP
	 */
	ast_channel_lock(cdata->chan);
	session = channel->session;
	if (session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO]) {
		changed |= check_for_rtp_changes(
			cdata->chan, cdata->rtp, session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO], session);
	}
	if (session->active_media_state->default_session[AST_MEDIA_TYPE_VIDEO]) {
		changed |= check_for_rtp_changes(
			cdata->chan, cdata->vrtp, session->active_media_state->default_session[AST_MEDIA_TYPE_VIDEO], session);
	}
	ast_channel_unlock(cdata->chan);

	if (direct_media_mitigate_glare(cdata->session)) {
		ast_debug(4, "Disregarding setting RTP on %s: mitigating re-INVITE glare\n", ast_channel_name(cdata->chan));
		ao2_ref(cdata, -1);
		return 0;
	}

	if (cdata->cap && ast_format_cap_count(cdata->cap) &&
		!ast_format_cap_identical(cdata->session->direct_media_cap, cdata->cap)) {
		ast_format_cap_remove_by_type(cdata->session->direct_media_cap, AST_MEDIA_TYPE_UNKNOWN);
		ast_format_cap_append_from_cap(cdata->session->direct_media_cap, cdata->cap, AST_MEDIA_TYPE_UNKNOWN);
		changed = 1;
	}

	if (changed) {
		ast_debug(4, "RTP changed on %s; initiating direct media update\n", ast_channel_name(cdata->chan));
		res = ast_sip_session_refresh(cdata->session, NULL, NULL, NULL,
			cdata->session->endpoint->media.direct_media.method, 1, NULL);
	}

	ao2_ref(cdata, -1);
	return res;
}

/*! \brief Function called by RTP engine to change where the remote party should send media */
static int chan_pjsip_set_rtp_peer(struct ast_channel *chan,
		struct ast_rtp_instance *rtp,
		struct ast_rtp_instance *vrtp,
		struct ast_rtp_instance *tpeer,
		const struct ast_format_cap *cap,
		int nat_active)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_session *session = channel->session;
	struct rtp_direct_media_data *cdata;

	/* Don't try to do any direct media shenanigans on early bridges */
	if ((rtp || vrtp || tpeer) && !ast_channel_is_bridged(chan)) {
		ast_debug(4, "Disregarding setting RTP on %s: channel is not bridged\n", ast_channel_name(chan));
		return 0;
	}

	if (nat_active && session->endpoint->media.direct_media.disable_on_nat) {
		ast_debug(4, "Disregarding setting RTP on %s: NAT is active\n", ast_channel_name(chan));
		return 0;
	}

	cdata = rtp_direct_media_data_create(chan, rtp, vrtp, cap, session);
	if (!cdata) {
		return 0;
	}

	if (ast_sip_push_task(session->serializer, send_direct_media_request, cdata)) {
		ast_log(LOG_ERROR, "Unable to send direct media request for channel %s\n", ast_channel_name(chan));
		ao2_ref(cdata, -1);
	}

	return 0;
}

/*! \brief Local glue for interacting with the RTP engine core */
static struct ast_rtp_glue chan_pjsip_rtp_glue = {
	.type = "PJSIP",
	.get_rtp_info = chan_pjsip_get_rtp_peer,
	.get_vrtp_info = chan_pjsip_get_vrtp_peer,
	.get_codec = chan_pjsip_get_codec,
	.update_peer = chan_pjsip_set_rtp_peer,
};

static void set_channel_on_rtp_instance(const struct ast_sip_session *session,
	const char *channel_id)
{
	int i;

	for (i = 0; i < AST_VECTOR_SIZE(&session->active_media_state->sessions); ++i) {
		struct ast_sip_session_media *session_media;

		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
		if (!session_media || !session_media->rtp) {
			continue;
		}

		ast_rtp_instance_set_channel_id(session_media->rtp, channel_id);
	}
}

/*!
 * \brief Determine if a topology is compatible with format capabilities
 *
 * This will return true if ANY formats in the topology are compatible with the format
 * capabilities.
 *
 * XXX When supporting true multistream, we will need to be sure to mark which streams from
 * top1 are compatible with which streams from top2. Then the ones that are not compatible
 * will need to be marked as "removed" so that they are negotiated as expected.
 *
 * \param top Topology
 * \param cap Format capabilities
 * \retval 1 The topology has at least one compatible format
 * \retval 0 The topology has no compatible formats or an error occurred.
 */
static int compatible_formats_exist(struct ast_stream_topology *top, struct ast_format_cap *cap)
{
	struct ast_format_cap *cap_from_top;
	int res;

	cap_from_top = ast_format_cap_from_stream_topology(top);

	if (!cap_from_top) {
		return 0;
	}

	res = ast_format_cap_iscompatible(cap_from_top, cap);
	ao2_ref(cap_from_top, -1);

	return res;
}

#ifdef GRANDSTREAM_NETWORKS
static void set_chan_avs_address(struct ast_sip_session *session, struct ast_channel *chan)
{
	char *avs_address = NULL;

	if (!session || !chan) {
		return;
	}

	if (!ast_get_avs_master_value()) {
		return;
	}

	/*added by zyxia for Bug 306514 */
	if(session->diff_multiavs_ip) {
		strncpy(session->avs_address, "127.0.0.1", sizeof(session->avs_address) - 1);
		ast_channel_set_avs_address(chan, "127.0.0.1");
	}
	else if (!ast_strlen_zero(session->avs_address)) {
		ast_channel_set_avs_address(chan, session->avs_address);
	}
	else {
		avs_address = ast_avs_get_avs_address();
		if(avs_address) {
			ast_channel_set_avs_address(chan, avs_address);
			strncpy(session->avs_address, avs_address, sizeof(session->avs_address) - 1);
			ast_free(avs_address);
		}
	}

	ast_log(LOG_DEBUG, "[session:%p] set_chan_avs_address:%s, endpoint:%s\n",
		session, ast_channel_get_avs_address(chan), ast_sorcery_object_get_id(session->endpoint));

	return;
}
#endif

/*! \brief Function called to create a new PJSIP Asterisk channel */
static struct ast_channel *chan_pjsip_new(struct ast_sip_session *session, int state, const char *exten, const char *title, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *cid_name)
{
	struct ast_channel *chan;
#ifdef GRANDSTREAM_NETWORKS
	struct ast_format_cap *caps = NULL;
#else
	struct ast_format_cap *caps;
#endif
	RAII_VAR(struct chan_pjsip_pvt *, pvt, NULL, ao2_cleanup);
	struct ast_sip_channel_pvt *channel;
	struct ast_variable *var;
#ifdef GRANDSTREAM_NETWORKS
	struct ast_stream_topology *topology = NULL;
#else
	struct ast_stream_topology *topology;
#endif
	if (!(pvt = ao2_alloc_options(sizeof(*pvt), chan_pjsip_pvt_dtor, AO2_ALLOC_OPT_LOCK_NOLOCK))) {
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	pvt->call_type = NULL;
	pvt->mcb_request = NULL;
	pvt->mcb_response = NULL;
	pvt->contact_user = NULL;
	pvt->isConference = FALSE;
#endif

	chan = ast_channel_alloc_with_endpoint(1, state,
		S_COR(session->id.number.valid, session->id.number.str, ""),
		S_COR(session->id.name.valid, session->id.name.str, ""),
		session->endpoint->accountcode,
		exten, session->endpoint->context,
		assignedids, requestor, 0,
		session->endpoint->persistent, "PJSIP/%s-%08x",
		ast_sorcery_object_get_id(session->endpoint),
		(unsigned) ast_atomic_fetchadd_int((int *) &chan_idx, +1));
	if (!chan) {
		return NULL;
	}

	ast_channel_set_port_group_id(chan, session->port_group_id);
	ast_channel_tech_set(chan, &chan_pjsip_tech);

	if (!(channel = ast_sip_channel_pvt_alloc(pvt, session))) {
		ast_channel_unlock(chan);
		ast_hangup(chan);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 218331 */
	if (!ast_strlen_zero(session->endpoint->anonymous) && !strncmp(session->endpoint->anonymous, "anonymous", strlen("anonymous"))) {
		pbx_builtin_setvar_helper(chan, "__ANONYMOUS", "anonymous");
		ast_string_field_set(session->endpoint, anonymous, "");
	}
	/* Added by gangwang for Bug 224258*/
	if (session->endpoint->click2call_enable) {
		ast_channel_set_is_webrtc_trunk(chan, 1);
	}
#endif

	ast_channel_tech_pvt_set(chan, channel);

	if (!ast_stream_topology_get_count(session->pending_media_state->topology) ||
		!compatible_formats_exist(session->pending_media_state->topology, session->endpoint->media.codecs)) {
		caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!caps) {
			ast_channel_unlock(chan);
			ast_hangup(chan);
			return NULL;
		}

		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);
		topology = ast_stream_topology_clone(session->endpoint->media.topology);

	} else {
		caps = ast_format_cap_from_stream_topology(session->pending_media_state->topology);
		topology = ast_stream_topology_clone(session->pending_media_state->topology);
	}

	if (!topology || !caps) {
		ao2_cleanup(caps);
		ast_stream_topology_free(topology);
		ast_channel_unlock(chan);
		ast_hangup(chan);
		return NULL;
	}

	ast_channel_stage_snapshot(chan);

	ast_channel_nativeformats_set(chan, caps);
	ast_channel_set_stream_topology(chan, topology);

	if (!ast_format_cap_empty(caps)) {
		struct ast_format *fmt;

		fmt = ast_format_cap_get_best_by_type(caps, AST_MEDIA_TYPE_AUDIO);
		if (!fmt) {
			/* Since our capabilities aren't empty, this will succeed */
			fmt = ast_format_cap_get_format(caps, 0);
		}
		ast_channel_set_writeformat(chan, fmt);
		ast_channel_set_rawwriteformat(chan, fmt);
		ast_channel_set_readformat(chan, fmt);
		ast_channel_set_rawreadformat(chan, fmt);
		ao2_ref(fmt, -1);
	}

	ao2_ref(caps, -1);

	if (state == AST_STATE_RING) {
		ast_channel_rings_set(chan, 1);
#ifdef GRANDSTREAM_NETWORKS
		ast_setstate(chan, AST_STATE_RING);
#endif
	}

	ast_channel_adsicpe_set(chan, AST_ADSI_UNAVAILABLE);

	ast_party_id_copy(&ast_channel_caller(chan)->id, &session->id);
	ast_party_id_copy(&ast_channel_caller(chan)->ani, &session->id);
	/*BEGIN: add by weidu ,for stir/shaken*/
#ifdef GRANDSTREAM_NETWORKS
	if (requestor){
		ast_channel_set_stir_shaken_origtn(chan, ast_channel_caller((struct ast_channel *)requestor)->id.number.str);
	}else{
		ast_channel_set_stir_shaken_origtn(chan, session->id.number.str);
	}
#endif
	/*END: add by weidu ,for stir/shaken*/
	if (!ast_strlen_zero(exten)) {
		/* Set provided DNID on the new channel. */
		ast_channel_dialed(chan)->number.str = ast_strdup(exten);
	}

	ast_channel_priority_set(chan, 1);

	ast_channel_callgroup_set(chan, session->endpoint->pickup.callgroup);
	ast_channel_pickupgroup_set(chan, session->endpoint->pickup.pickupgroup);

	ast_channel_named_callgroups_set(chan, session->endpoint->pickup.named_callgroups);
	ast_channel_named_pickupgroups_set(chan, session->endpoint->pickup.named_pickupgroups);
#ifdef GRANDSTREAM_NETWORKS
	ast_channel_pickupgroup_list_set(chan, &session->endpoint->pickup);
#endif

	if (!ast_strlen_zero(session->endpoint->language)) {
		ast_channel_language_set(chan, session->endpoint->language);
	}

	if (!ast_strlen_zero(session->endpoint->zone)) {
		struct ast_tone_zone *zone = ast_get_indication_zone(session->endpoint->zone);
		if (!zone) {
			ast_log(LOG_ERROR, "Unknown country code '%s' for tonezone. Check indications.conf for available country codes.\n", session->endpoint->zone);
		}
		ast_channel_zone_set(chan, zone);
	}

	for (var = session->endpoint->channel_vars; var; var = var->next) {
		char buf[512];
		pbx_builtin_setvar_helper(chan, var->name, ast_get_encoded_str(
						  var->value, buf, sizeof(buf)));
	}

	ast_channel_stage_snapshot_done(chan);
	ast_channel_unlock(chan);

	set_channel_on_rtp_instance(session, ast_channel_uniqueid(chan));

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for blind transfer callback when blind transfer is use ucm feature code, 2017.04.20 */
	char *allow_blind_transfer_callback = ast_sip_get_allow_blind_transfer_callback();
	if (strcasecmp(allow_blind_transfer_callback, "yes") == 0) {
		ast_channel_allow_blind_transfer_callback_set(chan, 1);
		unsigned int blind_transfer_timeout = ast_sip_get_blind_transfer_timeout();
		ast_channel_blind_transfer_timeout_set(chan, blind_transfer_timeout);
	}
	ast_free(allow_blind_transfer_callback);
	allow_blind_transfer_callback = NULL;

	/* Added by yfyang, for sca feature */
	ast_channel_set_line_index(chan, session->line_index);
	ast_channel_set_sharedline(chan, session->endpoint->sca_sharedline);

	ast_channel_cc_params_init(chan, session->endpoint->cc_params);

	if (session->endpoint->media.rtp.accept_multiple_sdp_answers) {
		ast_channel_set_use_final_sdp(chan, 1);
		if (requestor) {
			ast_channel_set_use_final_sdp((struct ast_channel*)requestor, 1);
		}
	}

	set_chan_avs_address(session, chan);

	if (ast_opt_avs) {
		ast_avs_create_channel(chan, avs_channel_type_rtp, 0, ast_channel_get_port_group_id(chan));
		ast_log(LOG_DEBUG, "[seesion:%p] chan_pjsip_new create chan:%s, chan_tick:%ld, port_group_id:%d, avs_chanid:%d, endpoint:%s\n",
			session, ast_channel_name(chan), ast_channel_get_avs_tick(chan), ast_channel_get_port_group_id(chan),
			ast_channel_get_avs_chanid(chan), ast_sorcery_object_get_id(session->endpoint));
	}
#endif

	return chan;
}

#ifdef GRANDSTREAM_NETWORKS

static int append_isfocus_contact(pjsip_tx_data *packet)
{
    pjsip_contact_hdr *contact = NULL;
    static const pj_str_t isfocus_str = { "isfocus", 7 };
    static const pj_str_t blank_str = { "", 0 };
    pjsip_param *isfocus = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL);
    if (contact == NULL) {
        ast_log(LOG_WARNING, "contact header is NULL\n");
        return -1;
    }

    isfocus = PJ_POOL_ALLOC_T(packet->pool, pjsip_param);
    if (isfocus == NULL) {
        ast_log(LOG_WARNING, "isfocus is NULL\n");
        return -1;
    }

    isfocus->name = isfocus_str;
    isfocus->value = blank_str;
    pj_list_push_back(&contact->other_param, isfocus);

    return 0;
}

static int set_contact_user(pjsip_tx_data *packet, char *user)
{
    pjsip_contact_hdr *contact = NULL;
    pjsip_sip_uri *contact_uri = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    if (user == NULL) {
        ast_log(LOG_WARNING, "user is NULL\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL);
    if (contact == NULL || contact->uri == NULL) {
        ast_log(LOG_WARNING, "contact URI header is NULL\n");
        return -1;
    }

    contact_uri = pjsip_uri_get_uri(contact->uri);
    if (contact_uri == NULL) {
        ast_log(LOG_WARNING, "Failed to get contact URI\n");
        return -1;
    }

    pj_strdup2(packet->pool, &contact_uri->user, user);

    return 0;
}


static int set_from_name(pjsip_tx_data *packet, const char *name)
{
	pjsip_name_addr *name_addr = NULL;

	// Sanity check
	if (packet == NULL) {
		ast_log(LOG_WARNING, "TX packet is NULL\n");
		return -1;
	}

	if (name == NULL) {
		ast_log(LOG_WARNING, "name is NULL\n");
		return -1;
	}

	name_addr = (pjsip_name_addr *)PJSIP_MSG_FROM_HDR(packet->msg)->uri;

	if (name_addr) {
		pj_strdup2(packet->pool, &name_addr->display, name);
	}

	return 0;
}

/* Added by yfyang, Grandstream networks, Bug 46889 */
static void add_call_info_to_response_tmp(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	char uuid_str[AST_UUID_STR_LEN] = {0};
	char *user = NULL;
	pjsip_fromto_hdr *hdr;
	pjsip_sip_uri    *uri;
	pjsip_name_addr  *name_addr;
	pjsip_param *param_p;
	pjsip_param *param_m;
	pjsip_contact_hdr  *contact_hdr = (pjsip_contact_hdr*) pjsip_msg_find_hdr(tdata->msg,PJSIP_H_CONTACT, NULL);
	pjsip_uri *base;
	char cc_uri[PJSIP_MAX_URL_SIZE] = {0};
	pj_str_t callinfo_name = { "Call-Info", 9 };

	if (!session || !tdata || !contact_hdr) {
		return;
	}

	base = contact_hdr->uri;
	hdr = pjsip_from_hdr_create(tdata->pool);
	hdr->type = PJSIP_H_OTHER;
	pj_strdup(tdata->pool, &hdr->name, &callinfo_name);
	hdr->sname = hdr->name;

	name_addr = pjsip_uri_clone(tdata->pool, base);
	uri = pjsip_uri_get_uri(name_addr->uri);
	user = ast_uuid_generate_str(uuid_str, sizeof(uuid_str));
	pj_strdup2(tdata->pool, &name_addr->display, "");
	pj_strdup2(tdata->pool, &uri->user, user);

	pjsip_uri_print(PJSIP_URI_IN_FROMTO_HDR, pjsip_uri_clone(tdata->pool, (pjsip_uri *) name_addr), cc_uri, sizeof(cc_uri));
	ast_log(LOG_NOTICE, "Get subscribe_uri[%s]\n", cc_uri);

	hdr->uri = (pjsip_uri *) name_addr;

	pj_strdup2(session->inv_session->pool, &session->cc_subs_user, user);
	session->cc_subs_uri = pjsip_uri_clone(session->inv_session->pool, (pjsip_uri *) name_addr);

	param_p = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param_p->name = pj_str("purpose");
	param_p->value = pj_str("call-completion");
	pj_list_insert_before(&hdr->other_param, param_p);

	param_m = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
	param_m->name = pj_str("m");
	param_m->value = pj_str("BS");
	pj_list_insert_before(&hdr->other_param, param_m);

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
}

static unsigned int judge_if_conference_enter_vm(struct ast_channel *chan, struct chan_pjsip_pvt *pvt)
{
	if (!chan || !pvt) {
		return 0;
	}

	if (!ast_channel_vm_get(chan)) {
		return 0;
	}

	if (pvt && pvt->call_type && (!strcasecmp(pvt->call_type, "Conference") || !strcasecmp(pvt->call_type, "VideoConference"))) {
		return 1;
	}

	return 0;
}

static int send_emergency_mail(void *datap)
{
	if (!datap)
	{
		ast_log(LOG_WARNING, "emergency mail data is null!!!");
		return -1;
	}

	char *cmd = datap;
	ast_safe_system2(cmd);
	if(cmd)
	{
		ast_free(cmd);
	}

	ast_log(LOG_DEBUG, "send emergency mail successful!!!");
	return 0;
}


static void add_sfu_header_for_trunk(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_supported_hdr *hdr = NULL;
	pj_str_t channel_id = { "Channel-ID", 10 };

	if (!session || !session->channel || !tdata) {
		ast_log(LOG_DEBUG, "add_sfu_header_for_trunk param is NULL\n");
		return;
	}

	if (!session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] add_sfu_header_for_trunk, is not SFU, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->ucm_trunk) {
		ast_log(LOG_DEBUG, "[session:%p] add_sfu_header_for_trunk, is not UCM trunk, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->conf_link) {
		ast_log(LOG_DEBUG, "[session:%p] add_sfu_header_for_trunk, already set conf_link, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	} else {
		if (ast_sip_session_set_conf_link_from_db(session)) {
			return;
		}
	}

	ast_log(LOG_DEBUG, "[session:%p] add_sfu_header_for_trunk, conf_link:%s, channel name:%s, endpoint:%s\n",
		session, session->conf_link, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));

	ast_sip_add_header(tdata, "X-GS-Conf-Link", session->conf_link);

	hdr = pjsip_msg_find_hdr_by_name(tdata->msg, &channel_id, NULL);
	if (!hdr) {
		ast_sip_add_header(tdata, "Channel-ID", ast_channel_name(session->channel));
	}

	return;
}

static void add_conf_link_for_wave(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	if (!session || !session->channel || !tdata) {
		ast_log(LOG_DEBUG, "add_conf_link_for_wave param is NULL\n");
		return;
	}

	if (ast_strlen_zero(session->conf_link)) {
		ast_log(LOG_DEBUG, "[session:%p] add_conf_link_for_wave, has no conf_link, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_strlen_zero(session->trunk_channel_id)) {
		ast_log(LOG_DEBUG, "[session:%p] add_conf_link_for_wave, has no trunk_channel_id, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if ((session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) && (session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		ast_log(LOG_DEBUG, "[session:%p] add_conf_link_for_wave, is not wave, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_conf_link_for_wave, conf_link:%s, channel name:%s, endpoint:%s\n",
		session, session->conf_link, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));

	ast_sip_add_header(tdata, "X-GS-Conf-Link", session->conf_link);

	ast_sip_add_header(tdata, "X-GS-Replace-Channel-ID", session->trunk_channel_id);

	return;
}


static void add_isfocus_for_trunk_conf(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	if (!session || !session->channel || !tdata ) {
		ast_log(LOG_DEBUG, "add_sfu_header_for_trunk param is NULL\n");
		return;
	}

	if (ast_strlen_zero(session->trunk_conf_num)) {
		ast_log(LOG_DEBUG, "[session:%p] add_isfocus_for_trunk_conf, has no trunk_conf_num, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if(ast_channel_name(session->channel) && strcasestr(ast_channel_name(session->channel),"PJSIP/trunk")) {
		ast_log(LOG_DEBUG, "[session:%p] add_isfocus_for_trunk_conf, is trunk channel, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		ast_log(LOG_DEBUG, "[session:%p] add_isfocus_for_trunk_conf, is wave terminal, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	append_isfocus_contact(tdata);
	ast_channel_set_is_join_trunk_conf(session->channel,1);

	//sip call to ucm trunk conf   or call from to trunk conf
	const char *trunk_index= pbx_builtin_getvar_helper(session->channel, "TRUNK_INDEX");
	if(!ast_strlen_zero(trunk_index)){
		char local_virtual_conf[512] = {0};
		char family[64] = {0};
		char trunkName[128] = {0};
		ast_log(LOG_DEBUG, "add_isfocus_for_trunk_conf:%s,%s\n",session->trunk_conf_num,ast_channel_name(session->channel));

		snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(trunk_index));
		ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
		snprintf(local_virtual_conf, sizeof(local_virtual_conf) - 1, "%s_%s", trunkName,session->trunk_conf_num);

		set_contact_user(tdata, local_virtual_conf);
		struct ast_sip_endpoint *endpoint = session->endpoint;
		const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
		if (endpoint != NULL && endpoint_id != NULL) {//remove
			if (add_peer_conference(ast_channel_name(session->channel), endpoint, local_virtual_conf,AST_CONF_TYPE_VIRTUAL) == 0){
				ast_channel_set_virtual_conf_num(session->channel, local_virtual_conf);
			}
		}

		ast_log(LOG_DEBUG, "[session:%p] add_isfocus_for_trunk_conf, contact_user:%s, channel name:%s, endpoint:%s\n",
			session, local_virtual_conf, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}
/*Start added by zyxia for trunk conf cei.2022.11.18*/

static void add_header_for_voicemail(struct ast_sip_session *session, pjsip_tx_data *packet)
{
	if (!session || !session->channel || !packet) {
		return;
	}

	if (ast_channel_vm_get(session->channel)) {
		ast_sip_add_header(packet, "GS-voicemail", "yes");
		return;
	}

	const char *is_vmmain = pbx_builtin_getvar_helper(session->channel, "IS_VMMAIN");
	if (!ast_strlen_zero(is_vmmain) && !strcasecmp(is_vmmain, "yes")) {
		ast_sip_add_header(packet, "GS-voicemail", "yes");
	}

	return;
}

static void add_client_info_for_wave(struct ast_sip_session *session, pjsip_tx_data *tdata, unsigned int is_caller)
{
	if (!session || !tdata) {
		return;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC && session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		return;
	}

	if (is_caller && ast_strlen_zero(session->caller_gs_client_info)) {
		return;
	}

	if (!is_caller && ast_strlen_zero(session->callee_gs_client_info)) {
		return;
	}

	if (is_caller) {
		ast_sip_add_header(tdata, "X-GS-Client-Info", session->caller_gs_client_info);
		ast_log(LOG_DEBUG, "[session:%p] add_client_info_for_wave caller info:%s, endpoint:%s\n",
			session, session->caller_gs_client_info, ast_sorcery_object_get_id(session->endpoint));
	} else {
		ast_sip_add_header(tdata, "X-GS-Client-Info", session->callee_gs_client_info);
		ast_log(LOG_DEBUG, "[session:%p] add_client_info_for_wave callee info:%s, endpoint:%s\n",
			session, session->callee_gs_client_info, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}
#endif

static int answer(void *data)
{
	pj_status_t status = PJ_SUCCESS;
	pjsip_tx_data *packet = NULL;
	struct ast_sip_session *session = data;
#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_channel *, peer, ast_channel_bridge_peer(session->channel), ast_channel_cleanup);
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(session->channel);
	struct chan_pjsip_pvt *pvt = NULL;
	int pjsip_channel = 0;
	int call_info_flag = 0;
#endif


	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			session->inv_session->cause,
			pjsip_get_status_text(session->inv_session->cause)->ptr);
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(session->inv_session);
#endif
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (peer) {
		if (0 != strncmp(ast_channel_name(peer), "PJSIP/trunk_", 12)) {
			session->cc_offer = 0;
		}
	}

	pvt = channel->pvt;
	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP") == 0 && pvt) {
		pjsip_channel = 1;

		ast_log(LOG_DEBUG, "[TK] channel: %s, mcb_response: %s, isConference: %d\n",
							ast_channel_name(session->channel),
							(pvt->mcb_response != NULL) ? pvt->mcb_response : "NULL",
							pvt->isConference);
	}

	const char *transparentsdp = pbx_builtin_getvar_helper((struct ast_channel *)session->channel, "TRANSPARENTSDP");
	if(transparentsdp && !strcasecmp(transparentsdp, "no")){
		ast_log(LOG_DEBUG, "channel transparentsdp can not be supported");
		session->is_direct_media = 0;
	}
#endif

	pjsip_dlg_inc_lock(session->inv_session->dlg);
	if (session->inv_session->invite_tsx) {
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for 183 neg, 2016.01.20 */
		/* Modify by bfdu for bug 76728, 2017.08.03 */
		session->send_final_rsp = 1;

		if (judge_if_conference_enter_vm(session->channel, pvt)) {
			ast_log(LOG_DEBUG, "[session:%p] Dialplan is enter voicemail, but endpoint:%s is join conference, don't response 200OK\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			/*BEGIN: Mod by weidu ,for 165972, if no answer, response 408*/
			if (AST_CAUSE_NO_USER_RESPONSE == ast_channel_hangupcause(session->channel) || AST_CAUSE_NO_ANSWER == ast_channel_hangupcause(session->channel)){
				status = pjsip_inv_answer(session->inv_session, 480, NULL, NULL, &packet);
			}else{
				status = pjsip_inv_answer(session->inv_session, 603, NULL, NULL, &packet);
			}
			/*END: Mod by weidu ,for 165972*/
		} else if (ast_channel_hangupcause(session->channel) == AST_CAUSE_BEARERCAPABILITY_NOTAVAIL) {
			status = pjsip_inv_answer(session->inv_session, 488, NULL, NULL, &packet);
		} else if (session->inv_session->updated_sdp_answer) {
			ast_log(LOG_DEBUG, "Has send 183 before, when send 200OK according callee answer\n");
			status = pjsip_inv_answer(session->inv_session, 200, NULL, session->inv_session->neg->neg_local_sdp, &packet);
		} else {
			status = pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet);
		}

		if (session->inv_session->neg && session->inv_session->neg->active_local_sdp) {
			char buf[1024] = {0};
			pjmedia_sdp_session* caller_answer_sdp = session->inv_session->neg->active_local_sdp;
			pjmedia_sdp_print(caller_answer_sdp, buf, sizeof(buf) - 1);
			//ast_log(LOG_CODEC_NEG, "[caller] A <<<=== UCM. send 200 ok sdp to the caller!\n%s", buf);
		}
#else
		status = pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet);
#endif
	} else {
		ast_log(LOG_ERROR,"Cannot answer '%s' because there is no associated SIP transaction\n",
			ast_channel_name(session->channel));
	}

#ifndef GRANDSTREAM_NETWORKS
	pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif

	if (status == PJ_SUCCESS && packet) {
#ifdef GRANDSTREAM_NETWORKS
		if (pjsip_channel) {
			if (pvt->mcb_response != NULL) {
				ast_sip_add_header(packet, "X-Grandstream-Mcb-Resp", pvt->mcb_response);
			}

			//not add isfocus if CEI notify disabled, modified by yfyang, for bug58890
			if ((pvt->isConference == TRUE) && (globalEnabledCEI || session->is_sfu)) {
				append_isfocus_contact(packet);
				ast_sip_add_header(packet, "Channel-ID", ast_channel_name(session->channel));
			} else {
				if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
					ast_sip_add_header(packet, "Channel-ID", ast_channel_name(session->channel));
				}
			}

			if (pvt->contact_user != NULL) {
				set_contact_user(packet, pvt->contact_user);
				if (pvt->isConference == TRUE) {
					pjsip_sip_uri *contact_uri = pjsip_uri_get_uri(session->inv_session->dlg->local.contact->uri);
					if (contact_uri != NULL) {
						pj_strdup2(session->inv_session->dlg->pool, &contact_uri->user, pvt->contact_user);
					}
				}
			} else if (session->contact_user != NULL) {
				set_contact_user(packet, session->contact_user);
			}

			/*Start added by zyxia for Bug 204898 call ipc device,200 ok need add*/
			const char *call_to_ipc = pbx_builtin_getvar_helper(session->channel, "CALL_TO_IPC");
			if (!ast_strlen_zero(call_to_ipc) && !strcasecmp(call_to_ipc, "yes")) {
				ast_log(LOG_NOTICE, "channel:%s  answer, add X-Grandstream-Call-Type header!",ast_channel_name(session->channel));
				ast_sip_add_header(packet, "X-Grandstream-Call-Type", "IPC");
			} else if(session->is_hikvision_call)
				ast_sip_add_header(packet, "X-Grandstream-Call-Type", "DoorStation/Hikvision");
			else if(session->is_gds_call)
				ast_sip_add_header(packet, "X-Grandstream-Call-Type", "DoorStation/Grandstream");//Bug 228701 ,call gds
			/*End added by zyxia for Bug 204898 call ipc,gds, device,200 ok need add*/
			else if (pvt->call_type != NULL) {
				ast_sip_add_header(packet, "X-Grandstream-Call-Type", pvt->call_type);
			}

			add_header_for_voicemail(session, packet);

			if (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
				ast_sip_add_timestamp_header(packet);
			}

			/* For register trunk, if we receive X-GS-Trunk header, add this header too */
			if (!strncasecmp(session->endpoint->aors, "trunk_", 6) || session->ucm_trunk) {
				ast_sip_add_header(packet, "X-GS-Trunk", "yes");
			}

			add_client_info_for_wave(session, packet, 0);
		}

		/* Modified by yfyang, for 73741, 91950, 108547 */

		if((ast_channel_get_terminate_type(session->channel) == AST_TERMINAL_TYPE_GSWAVE_MOBILE
			|| ast_channel_get_terminate_type(session->channel) == AST_TERMINAL_TYPE_WEBRTC)
			&& session->endpoint && !ast_sip_endpoint_is_have_h264_codec(session->endpoint)) {
				ast_sip_add_header(packet, "X-GS-Video-Codec", "no");
		}


		/*Start added by zyxia for Bug 235007,2023.3.17*/
		if(ast_channel_page_mode_get(session->channel) != AST_PAGE_MODE_NONE)
			ast_sip_add_header(packet, "X-Grandstream-Call-Type", "Paging");
		/*End added by zyxia for Bug 235007,2023.3.17*/

		const char *seamless_transfer = pbx_builtin_getvar_helper(session->channel, "SEAMLESS_TRANSFER");
		const char *sca_status = pbx_builtin_getvar_helper(session->channel, "SCA_DIAL_STATUS");
		const char *frompage = pbx_builtin_getvar_helper(session->channel, "PAGING_ALERT_INFO");
		const char *is_multipage = pbx_builtin_getvar_helper(session->channel, "IS_MULTIPAGE");

		/*Start added by xiazeyang  for bug114484,voicemail 200 ok,2018.7.3  */
		const char *is_vmmain = pbx_builtin_getvar_helper(session->channel, "IS_VMMAIN");
		/*End added by xiazeyang  for bug114484,voicemail 200 ok,2018.7.3  */


		if (session->endpoint && strncmp(session->endpoint->context, "sla-stations", 12)
			&& !session->is_direct_media
			&& (!seamless_transfer || strcasecmp(seamless_transfer, "yes"))
			&& (!pvt->call_type || (pvt->call_type && strncmp(pvt->call_type, "Conference", 10) && strcasecmp(pvt->call_type, "VideoConference")))
			&& (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12))
			&& (session->line_index == 0)
			&& (ast_strlen_zero(sca_status))
			&& (ast_strlen_zero(frompage))
			&& (!is_multipage || strcasecmp(is_multipage, "yes"))
			&& (!is_vmmain || strcasecmp(is_vmmain, "yes"))){ //added by xiazeyang  for bug114484,voicemail 200 ok,2018.7.3  */
			add_feature_map_header(packet, 0, session);
		}

		/* Added by yfyang, Grandstream, Bug 46889 */
		if (!peer && (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) &&
			(PJSIP_INV_STATE_INCOMING == session->inv_session->state) &&
			(AST_CC_AGENT_NATIVE == ast_get_cc_agent_policy(session->endpoint->cc_params))) {

			char *dialstatus = (char *)pbx_builtin_getvar_helper(session->channel, "DIALSTATUS");
			if (ast_strlen_zero(dialstatus) || (strcmp(dialstatus, "CHANUNAVAIL") && strcmp(dialstatus, "BUSY") && strcmp(dialstatus, "ANSWER"))) {
				ast_log(LOG_NOTICE, "answer, add callinfo header!");
				if (!ast_cc_check_max_agents(session->channel)) {
					add_call_info_to_response_tmp(session, packet);
					call_info_flag = 1;
				}
			}
		}

		if (!call_info_flag) {
			pj_str_t callinfo_name = { "Call-Info", 9 };
			pjsip_generic_string_hdr  *hdr_old;

			pj_lock_acquire(packet->lock);
			while ((hdr_old = pjsip_msg_find_hdr_by_name(packet->msg, &callinfo_name, NULL))) {
				pj_list_erase(hdr_old);
			}
			pj_lock_release(packet->lock);
		}

		/*Start added by zyxia for Bug210287,2021.12.24*/
		if(check_channel_is_video_call(session->channel))
			ast_channel_set_is_videocall(session->channel, TRUE);
		else
			ast_channel_set_is_videocall(session->channel, FALSE);
		/*End added by zyxia for Bug210287,2021.12.24*/

		/* Add by bfdu for whiteboard, 2022.1.7 */
		ast_sip_session_add_support_header_for_whiteboard(session, packet);

		add_sfu_header_for_trunk(session, packet);
		add_conf_link_for_wave(session, packet);

		/*send 200 ok to sip user,when call to trunk conf outbound*/
		add_isfocus_for_trunk_conf(session,packet);
		/*send 200 ok to remote trunk ucm,need subscribe remote conf*/
		send_trunk_conf_cei_subscribe(session);

		ast_sip_session_add_support_video_call_for_webrtc_trunk(session, packet);

		/* Add by bfdu for Bug 284438, clear IS_183_BEFORE flag when answer, 2024.1.9 */
		if (session->channel) {
			pbx_builtin_setvar_helper(session->channel, "_IS_183_BEFORE", "");
		}
#endif

		ast_sip_session_send_response(session, packet);
	}

#ifdef GRANDSTREAM_NETWORKS
	pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif

	if (status != PJ_SUCCESS) {
		char err[PJ_ERR_MSG_SIZE];

		pj_strerror(status, err, sizeof(err));
		ast_log(LOG_WARNING,"Cannot answer '%s': %s\n",
			ast_channel_name(session->channel), err);
		/*
		 * Return this value so we can distinguish between this
		 * failure and the threadpool synchronous push failing.
		 */
		return -2;
	}
	return 0;
}

/*! \brief Function called by core when we should answer a PJSIP session */
static int chan_pjsip_answer(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session *session;
	int res;

	if (ast_channel_state(ast) == AST_STATE_UP) {
		return 0;
	}

	ast_setstate(ast, AST_STATE_UP);
	session = ao2_bump(channel->session);

#ifdef HAVE_PJSIP_INV_SESSION_REF
	if (pjsip_inv_add_ref(session->inv_session) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
		ao2_ref(session, -1);
		return -1;
	}
#endif
#ifdef GRANDSTREAM_NETWORKS
	if (session->line_index) {
		enum sca_status status;

		/* Added by yfyang, for Bug 187604 */
		const char *FROM_SpeedDial = pbx_builtin_getvar_helper(session->channel, "FROM_SpeedDial");
		if (!ast_strlen_zero(FROM_SpeedDial) && !strncmp(FROM_SpeedDial, "yes", 3)) {
			char *dest = (char *)pbx_builtin_getvar_helper(session->channel, "SpeedDial_Dest");
			if (!ast_strlen_zero(dest)) {
				ast_sip_set_sca_user_info((char *)session->endpoint->sca_sharedline, session->line_index, NULL, dest);
			} else {
				if (!ast_strlen_zero(S_COR(ast_channel_connected(ast)->id.number.valid, ast_channel_connected(ast)->id.number.str, NULL))) {
					ast_sip_set_sca_user_info((char *)session->endpoint->sca_sharedline, session->line_index, NULL, ast_channel_connected(ast)->id.number.str);
				} else if (strlen(ast_channel_connect_num(session->channel))) {
					ast_sip_set_sca_user_info((char *)session->endpoint->sca_sharedline, session->line_index, NULL, (char *)ast_channel_connect_num(session->channel));
				}
			}
		}

		status = ast_sip_get_sca_line_status((char *)session->endpoint->sca_sharedline, session->line_index);
		ast_log(LOG_NOTICE, "chan_pjsip_answer, sharedline[%s], line[%d] in state[%d]",
				session->endpoint->sca_sharedline, session->line_index, status);

		if (status < SCA_Active) {
			ast_sip_set_sca_line_status((char *)session->endpoint->sca_sharedline, (int)session->line_index, SCA_Active);
		}
	}
#endif

	/* the answer task needs to be pushed synchronously otherwise a race condition
	   can occur between this thread and bridging (specifically when native bridging
	   attempts to do direct media) */
	ast_channel_unlock(ast);
	res = ast_sip_push_task_wait_serializer(session->serializer, answer, session);
	if (res) {
		if (res == -1) {
			ast_log(LOG_ERROR,"Cannot answer '%s': Unable to push answer task to the threadpool.\n",
				ast_channel_name(session->channel));
#ifdef HAVE_PJSIP_INV_SESSION_REF
			pjsip_inv_dec_ref(session->inv_session);
#endif
		}
		ao2_ref(session, -1);
		ast_channel_lock(ast);
		return -1;
	}
	ao2_ref(session, -1);
	ast_channel_lock(ast);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/*! \brief Internal helper function called when CNG tone is detected */
static void chan_pjsip_avs_cng_tone_detected(struct ast_sip_session *session)
{
	const char *target_context;
	int exists;

	ast_log(LOG_DEBUG, "Enter chan_pjsip_avs_cng_tone_detected channel_name:%s,extension:%s\n",
		ast_channel_name(session->channel),ast_channel_exten(session->channel));

	/* If already executing in the fax extension don't do anything */
	if (!strcmp(ast_channel_exten(session->channel), "fax")) {
		return ;
	}

	/*if sendfax  or the ucm is caller ,ignore the cng*/
	/*
	if (ast_channel_get_is_send_fax(session->channel) || (!ast_channel_get_caller_extension_flag(session->channel)
		&& !strncasecmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12))) {
		return ;
	}
	*/
	/*if sendfax  or the ucm is caller ,ignore the cng*/
	if (ast_channel_get_is_send_fax(session->channel) && !strncasecmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		return ;
	}


/*
	if (strncasecmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		return ;
	}
*/
	target_context = S_OR(ast_channel_macrocontext(session->channel), ast_channel_context(session->channel));

	/* We need to unlock the channel here because ast_exists_extension has the
	 * potential to start and stop an autoservice on the channel. Such action
	 * is prone to deadlock if the channel is locked.
	 */
	ast_channel_unlock(session->channel);
	exists = ast_exists_extension(session->channel, target_context, "fax", 1,
		S_COR(ast_channel_caller(session->channel)->id.number.valid,
			ast_channel_caller(session->channel)->id.number.str, NULL));
	ast_channel_lock(session->channel);

	if (exists && 0) {
		ast_verb(2, "Redirecting '%s' to fax extension due to CNG detection\n",
			ast_channel_name(session->channel));
		pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));

		/*Start added by zyxia for Bug 258665  ,2023.4.19*/
		const char *auto_answer = pbx_builtin_getvar_helper(session->channel, "auto_answer");
		if( ast_channel_name(session->channel) && strstr(ast_channel_name(session->channel),"trunk")
			&& !ast_strlen_zero(auto_answer) && !strcasecmp(auto_answer,"yes")){
			pbx_builtin_setvar_helper(session->channel, "FAX_WAIT_END", "yes");
			ast_cdr_setvar(ast_channel_name(session->channel),"special_action_type","AUTO_ANSWER_TO_FAX");
		}
		/*End added by zyxia for Bug 257462 ,2023.4.19*/

		/*! Added by hzhou@grandstream.cn for bug[113859], 2018-06-28*/
		const char *is_from_disa = pbx_builtin_getvar_helper(session->channel, "FROM_DISA");
		//const char *fax_detect = pbx_builtin_getvar_helper(session->channel, "enable_fax_detect");
		const char *fax_intelligent_route = pbx_builtin_getvar_helper(session->channel, "fax_intelligent_route");
		const char *fax_intelligent_route_destination = pbx_builtin_getvar_helper(session->channel, "fax_intelligent_route_destination");

		if (is_from_disa && fax_intelligent_route && fax_intelligent_route_destination
			&& !strcasecmp(is_from_disa, "yes") && !strcasecmp(fax_intelligent_route, "yes")
			&& strcasecmp(fax_intelligent_route_destination, "no")) {
			ast_channel_set_is_enable_fax_detect(session->channel, 0);
			pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE", "yes");
			pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE_DETINATION",
										fax_intelligent_route_destination);
			session->is_intelligent_route = 1;
		}
		session->fax_alreay_work = 1;//added by zyxia for Bug 186437,2021.1.19
		if (ast_async_goto(session->channel, target_context, "fax", 1)) {
			ast_log(LOG_ERROR, "Failed to async goto '%s' into fax extension in '%s'\n",
				ast_channel_name(session->channel), target_context);
		}
	} else {

		short intelligent_route_flag = 0;
		const char *callee_num = pbx_builtin_getvar_helper(session->channel, "EXTTOCALL");
		const char *from_trunk = pbx_builtin_getvar_helper(session->channel, "FROM_TRUNK");
		//const char *fax_detect = pbx_builtin_getvar_helper(session->channel, "enable_fax_detect");
		const char *fax_intelligent_route = pbx_builtin_getvar_helper(session->channel, "fax_intelligent_route");
		const char *fax_intelligent_route_destination = pbx_builtin_getvar_helper(session->channel, "fax_intelligent_route_destination");
		const char *fax_to_extension = pbx_builtin_getvar_helper(session->channel, "fax_to_extension");

		ast_log(LOG_VERBOSE, "callee_num:%s,from_trunk:%s,fax_intelligent_route:%s,fax_intelligent_route_destination:%s,fax_to_extension:%s,intelligent_route_flag:%d\n", callee_num,from_trunk,fax_intelligent_route,fax_intelligent_route_destination,fax_to_extension,intelligent_route_flag);

		if (callee_num && fax_intelligent_route_destination
			&& fax_intelligent_route && !strcasecmp(fax_intelligent_route, "yes")
			&& !strcasecmp(callee_num, fax_intelligent_route_destination)
			&& fax_to_extension && !strcasecmp(fax_to_extension, "extension")
			&& !ast_channel_get_vfax_flag(session->channel)) {
			intelligent_route_flag = 1;
			ast_log(LOG_VERBOSE, "The same purpose(destination[%s]) and called extension(callee[%s])",
				fax_intelligent_route_destination, callee_num);
		} else if (from_trunk && fax_intelligent_route_destination
			&& ast_channel_get_vfax_flag(session->channel)) {
			intelligent_route_flag = 1;
			ast_log(LOG_VERBOSE, "The same fax extension purpose(destination[%s])", fax_intelligent_route_destination);
		}
		/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
		if (!intelligent_route_flag && !strcasecmp(session->endpoint->faxcontext, "faxcontext")) {
			ast_verb(2, "Redirecting '%s' to fax extension due to CNG detection\n",
				ast_channel_name(session->channel));

			/*Start added by zyxia for Bug Bug 258665  ,2023.4.19*/
			const char *auto_answer = pbx_builtin_getvar_helper(session->channel, "auto_answer");
			if( ast_channel_name(session->channel) && strstr(ast_channel_name(session->channel),"trunk")
				&& !ast_strlen_zero(auto_answer) && !strcasecmp(auto_answer,"yes")){
				pbx_builtin_setvar_helper(session->channel, "FAX_WAIT_END", "yes");
				ast_cdr_setvar(ast_channel_name(session->channel),"special_action_type","AUTO_ANSWER_TO_FAX");//added by zyxia for Bug 257611,2023.4.28
			}
			/*End added by zyxia for Bug 258665  ,2023.4.19*/

			pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));
			if (fax_intelligent_route && !strcasecmp(fax_intelligent_route, "yes")
				&& fax_intelligent_route_destination && strcasecmp(fax_intelligent_route_destination, "no")
				&& from_trunk && !strcasecmp(from_trunk, "yes")) {
				ast_channel_set_is_enable_fax_detect(session->channel, 0);
				pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE", "yes");
				pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE_DETINATION",
					fax_intelligent_route_destination);
				session->is_intelligent_route = 1;
			}
			session->fax_alreay_work = 1;//added by zyxia for Bug 186437,2021.1.19
			char action_type[32] = {0};
			ast_cdr_getvar(ast_channel_name(session->channel), "action_type", action_type,sizeof(action_type));
			if (ast_async_goto(session->channel, session->endpoint->faxcontext, "fax", 1)) {
				ast_log(LOG_NOTICE, "Failed to async goto '%s' into fax of '%s'\n",
					ast_channel_name(session->channel), session->endpoint->faxcontext);
			}
			else if (0 == strncmp(action_type, "IVR", 3)){
				struct ast_flags options;
				options.flags = 0x70;	//set option "are"
				ast_cdr_fork(ast_channel_name(session->channel),&options);
			}

		} else
		/*! Added by pxie@grandstream.cn for bug[42104], 2015-04-14*/
		ast_log(LOG_NOTICE, "FAX CNG detected on '%s' but no fax extension in '%s'\n",
			ast_channel_name(session->channel), target_context);
	}

	return ;
}

#endif

/*! \brief Internal helper function called when CNG tone is detected */
static struct ast_frame *chan_pjsip_cng_tone_detected(struct ast_sip_session *session, struct ast_frame *f)
{
	const char *target_context;
	int exists;
	int dsp_features;

	dsp_features = ast_dsp_get_features(session->dsp);
	dsp_features &= ~DSP_FEATURE_FAX_DETECT;
	if (dsp_features) {
		ast_dsp_set_features(session->dsp, dsp_features);
	} else {
		ast_dsp_free(session->dsp);
		session->dsp = NULL;
	}

	/* If already executing in the fax extension don't do anything */
	if (!strcmp(ast_channel_exten(session->channel), "fax")) {
		return f;
	}

	target_context = S_OR(ast_channel_macrocontext(session->channel), ast_channel_context(session->channel));

	/*
	 * We need to unlock the channel here because ast_exists_extension has the
	 * potential to start and stop an autoservice on the channel. Such action
	 * is prone to deadlock if the channel is locked.
	 *
	 * ast_async_goto() has its own restriction on not holding the channel lock.
	 */
	ast_channel_unlock(session->channel);
	ast_frfree(f);
	f = &ast_null_frame;
	exists = ast_exists_extension(session->channel, target_context, "fax", 1,
		S_COR(ast_channel_caller(session->channel)->id.number.valid,
			ast_channel_caller(session->channel)->id.number.str, NULL));
	if (exists) {
		ast_verb(2, "Redirecting '%s' to fax extension due to CNG detection\n",
			ast_channel_name(session->channel));
		pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));
#ifdef GRANDSTREAM_NETWORKS
		char action_type[32] = {0};
		ast_cdr_getvar(ast_channel_name(session->channel), "action_type", action_type,sizeof(action_type));
#endif
		if (ast_async_goto(session->channel, target_context, "fax", 1)) {
			ast_log(LOG_ERROR, "Failed to async goto '%s' into fax extension in '%s'\n",
				ast_channel_name(session->channel), target_context);
		}
#ifdef GRANDSTREAM_NETWORKS
		else if (0 == strncmp(action_type, "IVR", 3)){
			struct ast_flags options;
			options.flags = 0x70;	//set option "are"
			ast_cdr_fork(ast_channel_name(session->channel),&options);
		}
#endif
	} else {
		ast_log(LOG_NOTICE, "FAX CNG detected on '%s' but no fax extension in '%s'\n",
			ast_channel_name(session->channel), target_context);
	}
	ast_channel_lock(session->channel);

	return f;
}

/*!
 * \brief Function called by core to read any waiting frames
 *
 * \note The channel is already locked.
 */
static struct ast_frame *chan_pjsip_read_stream(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session *session = channel->session;
	struct ast_sip_session_media_read_callback_state *callback_state;
	struct ast_frame *f;
	int fdno = ast_channel_fdno(ast) - AST_EXTENDED_FDS;

	if (fdno >= AST_VECTOR_SIZE(&session->active_media_state->read_callbacks)) {
		return &ast_null_frame;
	}

	callback_state = AST_VECTOR_GET_ADDR(&session->active_media_state->read_callbacks, fdno);

#ifdef GRANDSTREAM_NETWORKS
	if (callback_state->session->bfcp_udp) {
		ast_log(LOG_DEBUG, "read BFCP frame\n");
		f = ast_bfcp_read(callback_state->session->bfcp_udp);		/* UDP for bfcp */
		return f;
	} else {
		if (callback_state->read_callback) {
			f = callback_state->read_callback(session, callback_state->session);
		} else {
			return &ast_null_frame;
		}
	}
#else
	f = callback_state->read_callback(session, callback_state->session);
#endif

	if (!f) {
		return f;
	}

	if (f->frametype != AST_FRAME_VOICE ||
		callback_state->session != session->active_media_state->default_session[callback_state->session->type]) {
		return f;
	}

	session = channel->session;

	/*
	 * Asymmetric RTP only has one native format set at a time.
	 * Therefore we need to update the native format to the current
	 * raw read format BEFORE the native format check
	 */
	if (!session->endpoint->asymmetric_rtp_codec &&
		ast_format_cmp(ast_channel_rawwriteformat(ast), f->subclass.format) == AST_FORMAT_CMP_NOT_EQUAL) {
		struct ast_format_cap *caps;

		/* For maximum compatibility we ensure that the formats match that of the received media */
		ast_debug(1, "Oooh, got a frame with format of %s on channel '%s' when we're sending '%s', switching to match\n",
			ast_format_get_name(f->subclass.format), ast_channel_name(ast),
			ast_format_get_name(ast_channel_rawwriteformat(ast)));

		caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (caps) {
			ast_format_cap_append_from_cap(caps, ast_channel_nativeformats(ast), AST_MEDIA_TYPE_UNKNOWN);
			ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_AUDIO);
			ast_format_cap_append(caps, f->subclass.format, 0);
			ast_channel_nativeformats_set(ast, caps);
			ao2_ref(caps, -1);
		}

		ast_set_write_format_path(ast, ast_channel_writeformat(ast), f->subclass.format);
		ast_set_read_format_path(ast, ast_channel_readformat(ast), f->subclass.format);

		if (ast_channel_is_bridged(ast)) {
			ast_channel_set_unbridged_nolock(ast, 1);
		}
	}

	if (ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), f->subclass.format) == AST_FORMAT_CMP_NOT_EQUAL) {
		ast_debug(1, "Oooh, got a frame with format of %s on channel '%s' when it has not been negotiated\n",
			ast_format_get_name(f->subclass.format), ast_channel_name(ast));

		ast_frfree(f);
		return &ast_null_frame;
	}

	if (session->dsp) {
		int dsp_features;

		dsp_features = ast_dsp_get_features(session->dsp);
		if ((dsp_features & DSP_FEATURE_FAX_DETECT)
			&& session->endpoint->faxdetect_timeout
			&& session->endpoint->faxdetect_timeout <= ast_channel_get_up_time(ast)) {
			dsp_features &= ~DSP_FEATURE_FAX_DETECT;
			if (dsp_features) {
				ast_dsp_set_features(session->dsp, dsp_features);
			} else {
				ast_dsp_free(session->dsp);
				session->dsp = NULL;
			}
			ast_debug(3, "Channel driver fax CNG detection timeout on %s\n",
				ast_channel_name(ast));
		}
	}
	if (session->dsp) {
		f = ast_dsp_process(ast, session->dsp, f);
		if (f && (f->frametype == AST_FRAME_DTMF)) {
			if (f->subclass.integer == 'f') {
				ast_debug(3, "Channel driver fax CNG detected on %s\n",
					ast_channel_name(ast));
				f = chan_pjsip_cng_tone_detected(session, f);
			} else {
				ast_debug(3, "* Detected inband DTMF '%c' on '%s'\n", f->subclass.integer,
					ast_channel_name(ast));
			}
		}
	}

	return f;
}

static int chan_pjsip_write_stream(struct ast_channel *ast, int stream_num, struct ast_frame *frame)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session *session = channel->session;
	struct ast_sip_session_media *media = NULL;
	int res = 0;

	/* The core provides a guarantee that the stream will exist when we are called if stream_num is provided */
	if (stream_num >= 0) {
		/* What is not guaranteed is that a media session will exist */
		if (stream_num < AST_VECTOR_SIZE(&channel->session->active_media_state->sessions)) {
			media = AST_VECTOR_GET(&channel->session->active_media_state->sessions, stream_num);
		}
	}

	switch (frame->frametype) {
	case AST_FRAME_VOICE:
		if (!media) {
			return 0;
		} else if (media->type != AST_MEDIA_TYPE_AUDIO) {
			ast_debug(3, "Channel %s stream %d is of type '%s', not audio!\n",
				ast_channel_name(ast), stream_num, ast_codec_media_type2str(media->type));
			return 0;
		} else if (media == channel->session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO] &&
			ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), frame->subclass.format) == AST_FORMAT_CMP_NOT_EQUAL) {
			struct ast_str *cap_buf = ast_str_alloca(AST_FORMAT_CAP_NAMES_LEN);
			struct ast_str *write_transpath = ast_str_alloca(256);
			struct ast_str *read_transpath = ast_str_alloca(256);

			ast_log(LOG_WARNING,
				"Channel %s asked to send %s frame when native formats are %s (rd:%s->%s;%s wr:%s->%s;%s)\n",
				ast_channel_name(ast),
				ast_format_get_name(frame->subclass.format),
				ast_format_cap_get_names(ast_channel_nativeformats(ast), &cap_buf),
				ast_format_get_name(ast_channel_rawreadformat(ast)),
				ast_format_get_name(ast_channel_readformat(ast)),
				ast_translate_path_to_str(ast_channel_readtrans(ast), &read_transpath),
				ast_format_get_name(ast_channel_writeformat(ast)),
				ast_format_get_name(ast_channel_rawwriteformat(ast)),
				ast_translate_path_to_str(ast_channel_writetrans(ast), &write_transpath));
			return 0;
		} else if (media->write_callback) {
			res = media->write_callback(session, media, frame);

		}
		break;
	case AST_FRAME_VIDEO:
		if (!media) {
			return 0;
		} else if (media->type != AST_MEDIA_TYPE_VIDEO) {
			ast_debug(3, "Channel %s stream %d is of type '%s', not video!\n",
				ast_channel_name(ast), stream_num, ast_codec_media_type2str(media->type));
			return 0;
		} else if (media->write_callback) {
			res = media->write_callback(session, media, frame);
		}
		break;
	case AST_FRAME_MODEM:
		if (!media) {
			return 0;
		} else if (media->type != AST_MEDIA_TYPE_IMAGE) {
			ast_debug(3, "Channel %s stream %d is of type '%s', not image!\n",
				ast_channel_name(ast), stream_num, ast_codec_media_type2str(media->type));
			return 0;
		} else if (media->write_callback) {
			res = media->write_callback(session, media, frame);
		}
		break;
	case AST_FRAME_CNG:
		break;
	case AST_FRAME_RTCP:
		/* We only support writing out feedback */
		if (frame->subclass.integer != AST_RTP_RTCP_PSFB || !media) {
			return 0;
		} else if (media->type != AST_MEDIA_TYPE_VIDEO) {
			ast_debug(3, "Channel %s stream %d is of type '%s', not video! Unable to write RTCP feedback.\n",
				ast_channel_name(ast), stream_num, ast_codec_media_type2str(media->type));
			return 0;
		} else if (media->write_callback) {
			res = media->write_callback(session, media, frame);
		}
		break;
#ifdef GRANDSTREAM_NETWORKS
	case AST_FRAME_BFCP:
		ast_log(LOG_DEBUG, "write BFCP frame\n");
		if (media && media->bfcp_udp) {
			res = ast_bfcp_write(media->bfcp_udp, frame);
		}
		break;
#endif
	default:
		ast_log(LOG_WARNING, "Can't send %u type frames with PJSIP\n", frame->frametype);
		break;
	}

	return res;
}

static int chan_pjsip_write(struct ast_channel *ast, struct ast_frame *frame)
{
	return chan_pjsip_write_stream(ast, -1, frame);
}

/*! \brief Function called by core to change the underlying owner channel */
static int chan_pjsip_fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(newchan);

	if (channel->session->channel != oldchan) {
		return -1;
	}

	/*
	 * The masquerade has suspended the channel's session
	 * serializer so we can safely change it outside of
	 * the serializer thread.
	 */
	channel->session->channel = newchan;

	set_channel_on_rtp_instance(channel->session, ast_channel_uniqueid(newchan));

	return 0;
}

/*! AO2 hash function for on hold UIDs */
static int uid_hold_hash_fn(const void *obj, const int flags)
{
	const char *key = obj;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_KEY:
		break;
	case OBJ_SEARCH_OBJECT:
		break;
	default:
		/* Hash can only work on something with a full key. */
		ast_assert(0);
		return 0;
	}
	return ast_str_hash(key);
}

/*! AO2 sort function for on hold UIDs */
static int uid_hold_sort_fn(const void *obj_left, const void *obj_right, const int flags)
{
	const char *left = obj_left;
	const char *right = obj_right;
	int cmp;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_OBJECT:
	case OBJ_SEARCH_KEY:
		cmp = strcmp(left, right);
		break;
	case OBJ_SEARCH_PARTIAL_KEY:
		cmp = strncmp(left, right, strlen(right));
		break;
	default:
		/* Sort can only work on something with a full or partial key. */
		ast_assert(0);
		cmp = 0;
		break;
	}
	return cmp;
}

static struct ao2_container *pjsip_uids_onhold;

/*!
 * \brief Add a channel ID to the list of PJSIP channels on hold
 *
 * \param chan_uid - Unique ID of the channel being put into the hold list
 *
 * \retval 0 Channel has been added to or was already in the hold list
 * \retval -1 Failed to add channel to the hold list
 */
static int chan_pjsip_add_hold(const char *chan_uid)
{
	RAII_VAR(char *, hold_uid, NULL, ao2_cleanup);

	hold_uid = ao2_find(pjsip_uids_onhold, chan_uid, OBJ_SEARCH_KEY);
	if (hold_uid) {
		/* Device is already on hold. Nothing to do. */
		return 0;
	}

	/* Device wasn't in hold list already. Create a new one. */
	hold_uid = ao2_alloc_options(strlen(chan_uid) + 1, NULL,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!hold_uid) {
		return -1;
	}

	ast_copy_string(hold_uid, chan_uid, strlen(chan_uid) + 1);

	if (ao2_link(pjsip_uids_onhold, hold_uid) == 0) {
		return -1;
	}

	return 0;
}

/*!
 * \brief Remove a channel ID from the list of PJSIP channels on hold
 *
 * \param chan_uid - Unique ID of the channel being taken out of the hold list
 */
static void chan_pjsip_remove_hold(const char *chan_uid)
{
	ao2_find(pjsip_uids_onhold, chan_uid, OBJ_SEARCH_KEY | OBJ_UNLINK | OBJ_NODATA);
}

/*!
 * \brief Determine whether a channel ID is in the list of PJSIP channels on hold
 *
 * \param chan_uid - Channel being checked
 *
 * \retval 0 The channel is not in the hold list
 * \retval 1 The channel is in the hold list
 */
static int chan_pjsip_get_hold(const char *chan_uid)
{
	RAII_VAR(char *, hold_uid, NULL, ao2_cleanup);

	hold_uid = ao2_find(pjsip_uids_onhold, chan_uid, OBJ_SEARCH_KEY);
	if (!hold_uid) {
		return 0;
	}

	return 1;
}

/*! \brief Function called to get the device state of an endpoint */
static int chan_pjsip_devicestate(const char *data)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", data), ao2_cleanup);
	enum ast_device_state state = AST_DEVICE_UNKNOWN;
	RAII_VAR(struct ast_endpoint_snapshot *, endpoint_snapshot, NULL, ao2_cleanup);
	RAII_VAR(struct stasis_cache *, cache, NULL, ao2_cleanup);
	struct ast_devstate_aggregate aggregate;
	int num, inuse = 0;

	if (!endpoint) {
		return AST_DEVICE_INVALID;
	}

	endpoint_snapshot = ast_endpoint_latest_snapshot(ast_endpoint_get_tech(endpoint->persistent),
		ast_endpoint_get_resource(endpoint->persistent));

	if (!endpoint_snapshot) {
		return AST_DEVICE_INVALID;
	}

	if (endpoint_snapshot->state == AST_ENDPOINT_OFFLINE) {
		state = AST_DEVICE_UNAVAILABLE;
	} else if (endpoint_snapshot->state == AST_ENDPOINT_ONLINE) {
		state = AST_DEVICE_NOT_INUSE;
	}

	if (!endpoint_snapshot->num_channels || !(cache = ast_channel_cache())) {
		return state;
	}

	ast_devstate_aggregate_init(&aggregate);

	ao2_ref(cache, +1);

	for (num = 0; num < endpoint_snapshot->num_channels; num++) {
		RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
		struct ast_channel_snapshot *snapshot;

		msg = stasis_cache_get(cache, ast_channel_snapshot_type(),
			endpoint_snapshot->channel_ids[num]);

		if (!msg) {
			continue;
		}

		snapshot = stasis_message_data(msg);

		if (chan_pjsip_get_hold(snapshot->uniqueid)) {
			ast_devstate_aggregate_add(&aggregate, AST_DEVICE_ONHOLD);
		} else {
			ast_devstate_aggregate_add(&aggregate, ast_state_chan2dev(snapshot->state));
		}

		if ((snapshot->state == AST_STATE_UP) || (snapshot->state == AST_STATE_RING) ||
			(snapshot->state == AST_STATE_BUSY)) {
			inuse++;
		}
	}

	if (endpoint->devicestate_busy_at && (inuse == endpoint->devicestate_busy_at)) {
		state = AST_DEVICE_BUSY;
	} else if (ast_devstate_aggregate_result(&aggregate) != AST_DEVICE_INVALID) {
		state = ast_devstate_aggregate_result(&aggregate);
	}

	return state;
}

/*! \brief Function called to query options on a channel */
static int chan_pjsip_queryoption(struct ast_channel *ast, int option, void *data, int *datalen)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session *session = channel->session;
	int res = -1;
	enum ast_t38_state state = T38_STATE_UNAVAILABLE;

	switch (option) {
	case AST_OPTION_T38_STATE:
		if (session->endpoint->media.t38.enabled) {
			switch (session->t38state) {
			case T38_LOCAL_REINVITE:
			case T38_PEER_REINVITE:
				state = T38_STATE_NEGOTIATING;
				break;
			case T38_ENABLED:
				state = T38_STATE_NEGOTIATED;
				break;
			case T38_REJECTED:
				state = T38_STATE_REJECTED;
				break;
			default:
				state = T38_STATE_UNKNOWN;
				break;
			}
		}

		*((enum ast_t38_state *) data) = state;
		res = 0;

		break;
	default:
		break;
	}

	return res;
}

static const char *chan_pjsip_get_uniqueid(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	char *uniqueid = ast_threadstorage_get(&uniqueid_threadbuf, UNIQUEID_BUFSIZE);

	if (!uniqueid) {
		return "";
	}

	ast_copy_pj_str(uniqueid, &channel->session->inv_session->dlg->call_id->id, UNIQUEID_BUFSIZE);

	return uniqueid;
}

struct indicate_data {
	struct ast_sip_session *session;
	int condition;
	int response_code;
	void *frame_data;
	size_t datalen;
};

static void indicate_data_destroy(void *obj)
{
	struct indicate_data *ind_data = obj;

	ast_free(ind_data->frame_data);
	ao2_ref(ind_data->session, -1);
}

static struct indicate_data *indicate_data_alloc(struct ast_sip_session *session,
		int condition, int response_code, const void *frame_data, size_t datalen)
{
	struct indicate_data *ind_data = ao2_alloc(sizeof(*ind_data), indicate_data_destroy);

	if (!ind_data) {
		return NULL;
	}

	ind_data->frame_data = ast_malloc(datalen);
	if (!ind_data->frame_data) {
		ao2_ref(ind_data, -1);
		return NULL;
	}

	memcpy(ind_data->frame_data, frame_data, datalen);
	ind_data->datalen = datalen;
	ind_data->condition = condition;
	ind_data->response_code = response_code;
	ao2_ref(session, +1);
	ind_data->session = session;

	return ind_data;
}

static int indicate(void *data)
{
	pjsip_tx_data *packet = NULL;
	struct indicate_data *ind_data = data;
	struct ast_sip_session *session = ind_data->session;
	int response_code = ind_data->response_code;

#ifdef GRANDSTREAM_NETWORKS
	//pjsip_dlg_inc_lock(session->inv_session->dlg); // Added by pxie@grandstream.cn for bug[131634], 2019-03-06.
	/* Begin add by chdruan for Bug 42574 , 2015.7.9 */
	if (response_code == 183 && session->inv_session->options == PJSIP_INV_REQUIRE_100REL) {
		/* pjsip_inv_answer will clear sdp when 100rel is active, then "200 OK" without sdp*/
		session->inv_session->options &= ~PJSIP_INV_REQUIRE_100REL;
		if ((session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) &&
			(pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS)) {
			ast_sip_session_send_response(session, packet);
		}
		session->inv_session->options &= PJSIP_INV_REQUIRE_100REL;
	} else if (session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) {
		/* Modify by bfdu for bug 79193, 2017.02.22 */
		pj_status_t status = PJ_SUCCESS;
		if (session->inv_session->updated_sdp_answer) {
			ast_log(LOG_DEBUG, "Has send 183 before, when send 183 according callee's 183\n");
			/* for bug 263820 , add by yjwu */
			if (session->channel)
			{
				pbx_builtin_setvar_helper(session->channel, "_IS_183_BEFORE", "yes");
				if (ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_MOH))
					ast_moh_stop(session->channel);
			}
			status = pjsip_inv_answer(session->inv_session, response_code, NULL, session->inv_session->neg->neg_local_sdp, &packet);
		} else {
			ast_sip_notify_avs_to_make_early_bridge(session);
			status = pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet);
		}
		if (status == PJ_SUCCESS) {
			ast_sip_session_send_response(session, packet);
		}
	}
	/* End add by chdruan for Bug 42574 , 2015.7.9 */
	//pjsip_dlg_dec_lock(session->inv_session->dlg); // Added by pxie@grandstream.cn for bug[131634], 2019-03-06.
#else
	if ((session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) &&
		(pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS)) {
		ast_sip_session_send_response(session, packet);
	}
#endif

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif
	ao2_ref(ind_data, -1);

	return 0;
}

/*! \brief Send SIP INFO with video update request */
static int transmit_info_with_vidupdate(void *data)
{
	const char * xml =
		"<?xml version=\"1.0\" encoding=\"utf-8\" ?>\r\n"
		" <media_control>\r\n"
		"  <vc_primitive>\r\n"
		"   <to_encoder>\r\n"
		"    <picture_fast_update/>\r\n"
		"   </to_encoder>\r\n"
		"  </vc_primitive>\r\n"
		" </media_control>\r\n";

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "media_control+xml",
		.body_text = xml
	};

	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);
	struct pjsip_tx_data *tdata;

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			session->inv_session->cause,
			pjsip_get_status_text(session->inv_session->cause)->ptr);
		goto failure;
	}

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text video update INFO request\n");
		goto failure;
	}
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to text video update INFO request\n");
		goto failure;
	}
	ast_sip_session_send_request(session, tdata);

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif

	return 0;

failure:
#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif
	return -1;

}

/*!
 * \internal
 * \brief TRUE if a COLP update can be sent to the peer.
 * \since 13.3.0
 *
 * \param session The session to see if the COLP update is allowed.
 *
 * \retval 0 Update is not allowed.
 * \retval 1 Update is allowed.
 */
static int is_colp_update_allowed(struct ast_sip_session *session)
{
	struct ast_party_id connected_id;
	int update_allowed = 0;

	if (!session->endpoint->send_connected_line
		|| (!session->endpoint->id.send_pai && !session->endpoint->id.send_rpid)) {
		return 0;
	}

	/*
	 * Check if privacy allows the update.  Check while the channel
	 * is locked so we can work with the shallow connected_id copy.
	 */
	ast_channel_lock(session->channel);
	connected_id = ast_channel_connected_effective_id(session->channel);
	if (connected_id.number.valid
		&& (session->endpoint->id.trust_outbound
			|| (ast_party_id_presentation(&connected_id) & AST_PRES_RESTRICTION) == AST_PRES_ALLOWED)) {
		update_allowed = 1;
	} 
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 311270 */
	else if (!connected_id.number.valid) {
		char *cf_reason = (char *)pbx_builtin_getvar_helper(session->channel, "CF_REASON");
		if (!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "14")) {
			char *bridge_peer = (char *)pbx_builtin_getvar_helper(session->channel, "BRIDGEPEER");
			if (!ast_strlen_zero(bridge_peer) && !strncmp(bridge_peer, "Local/", 6)) {
				char *ptr1 = bridge_peer + strlen("Local/");
				char *ptr2 = strchr(ptr1, '@');

				if (ptr2 && ptr2 > ptr1) {
					char number[128] = {0};
					strncpy(number, ptr1, ((ptr2-ptr1)<128)?(ptr2-ptr1):127);

					ast_channel_connected(session->channel)->id.number.valid = 1;
					ast_channel_connected(session->channel)->id.number.str = ast_strdup(number);
					ast_log(LOG_NOTICE, "Set connected number[%s] to channel[%s]", number, ast_channel_name(session->channel));

					update_allowed = 1;
				}
			}
		}
	}
#endif

	ast_channel_unlock(session->channel);

	return update_allowed;
}

/*! \brief Update connected line information */
static int update_connected_line_information(void *data)
{
	struct ast_sip_session *session = data;

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			session->inv_session->cause,
			pjsip_get_status_text(session->inv_session->cause)->ptr);
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(session->inv_session);
#endif
		ao2_ref(session, -1);
		return -1;
	}

	if (ast_channel_state(session->channel) == AST_STATE_UP
		|| session->inv_session->role == PJSIP_ROLE_UAC) {
#ifdef GRANDSTREAM_NETWORKS
		/*Start added by zyxia for callflip update connect number,2021.11.19*/
		if (is_colp_update_allowed(session) || ast_channel_get_is_callflip(session->channel)) { //added by zyxia
		/*End added by zyxia for callflip update connect number,2021.11.19*/
#else
		if (is_colp_update_allowed(session)) {
#endif
			enum ast_sip_session_refresh_method method;
			int generate_new_sdp;

			method = session->endpoint->id.refresh_method;
			if (session->inv_session->options & PJSIP_INV_SUPPORT_UPDATE) {
				method = AST_SIP_SESSION_REFRESH_METHOD_UPDATE;
			}

			/* Only the INVITE method actually needs SDP, UPDATE can do without */
			generate_new_sdp = (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE);

		/*Start added by zyxia for Bug 215084,2022.2.26*/
#ifdef GRANDSTREAM_NETWORKS
			int need_refresh = 1;
			if (ast_channel_get_is_callflip_connected(session->channel) && ast_channel_name(session->channel) && strstr(ast_channel_name(session->channel), "PJSIP/trunk") && (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE)) {
				need_refresh = 0;
				ast_channel_set_is_callflip_connected(session->channel,0);
				ast_log(LOG_WARNING, "[session:%p] dont need send reinvite to sip trunk when receive connected line for callflip for channel:%s\n",
					session, ast_channel_name(session->channel));
			}

			/*told action callflip ,have send update,dont need update again,added by zyxia for callflip update callid */
			if(ast_channel_get_is_callflip(session->channel) && need_refresh && method == AST_SIP_SESSION_REFRESH_METHOD_UPDATE) {
				ast_channel_set_is_callflip_have_sendupdate(session->channel,1);
			}

			/* Added by yfyang, for Bug 226167 */
			if ((session->inv_session->role == PJSIP_ROLE_UAC) && (ast_channel_state(session->channel) == AST_STATE_DOWN)) {
				need_refresh = 0;
			}

			/* Added by yfyang, for Bug 264865 */
			if ((session->inv_session->role == PJSIP_ROLE_UAC) && (ast_channel_state(session->channel) == AST_STATE_UP)) {
				char *vq_cb_caller = (char *)pbx_builtin_getvar_helper(session->channel, "VQ_CALLBACK_CALLER");
				if (!ast_strlen_zero(vq_cb_caller) && !strncmp(vq_cb_caller, "yes", 3)) {
					ast_log(LOG_NOTICE, "It's vq callback caller, not update connected line");
					need_refresh = 0;
				}
			}

			if (need_refresh) {
				ast_sip_session_refresh(session, NULL, NULL, NULL, method, generate_new_sdp, NULL);
			}
#else
			ast_sip_session_refresh(session, NULL, NULL, NULL, method, generate_new_sdp, NULL);
#endif
		/*End added by zyxia for Bug 215084,2022.2.26*/
		}
	} else if (session->endpoint->id.rpid_immediate
		&& session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED
		&& is_colp_update_allowed(session)) {
		int response_code = 0;

		if (ast_channel_state(session->channel) == AST_STATE_RING) {
#ifdef GRANDSTREAM_NETWORKS
			response_code = !(session->endpoint->inband_progress && !session->is_direct_media) ? 180 : 183;
#else
			response_code = !session->endpoint->inband_progress ? 180 : 183;
#endif
		} else if (ast_channel_state(session->channel) == AST_STATE_RINGING) {
			response_code = 183;
		}

		if (response_code) {
			struct pjsip_tx_data *packet = NULL;

			if (pjsip_inv_answer(session->inv_session, response_code, NULL, NULL, &packet) == PJ_SUCCESS) {
				ast_sip_session_send_response(session, packet);
			}
		}
	}

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif

	ao2_ref(session, -1);
	return 0;
}

/*! \brief Callback which changes the value of locally held on the media stream */
static void local_hold_set_state(struct ast_sip_session_media *session_media, unsigned int held)
{
	if (session_media) {
		session_media->locally_held = held;
	}
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for AMI hold, 2021.12.02 */
static void local_hold_set_stream_state(struct ast_sip_session *session, unsigned int held)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream_topology *topology = NULL;

	if (!session || !session->active_media_state) {
		return;
	}

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
		if (!session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] local_hold_set_stream_state clone topology failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
	}

	topology = session->pending_media_state->topology;

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (held) {
			if (ast_stream_get_local_media_direction(stream) == SENDRECV) {
				ast_stream_set_local_media_direction(stream, SENDONLY);
			}
		} else {
			if (ast_stream_get_local_media_direction(stream) == SENDONLY) {
				ast_stream_set_local_media_direction(stream, SENDRECV);
			}
		}
	}

	return;
}
#endif

/*! \brief Update local hold state and send a re-INVITE with the new SDP */
static int remote_send_hold_refresh(struct ast_sip_session *session, unsigned int held)
{
	AST_VECTOR_CALLBACK_VOID(&session->active_media_state->sessions, local_hold_set_state, held);
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for AMI hold, 2021.12.02 */
	local_hold_set_stream_state(session, held);
#endif
	ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	ao2_ref(session, -1);

	return 0;
}

/*! \brief Update local hold state to be held */
static int remote_send_hold(void *data)
{
	return remote_send_hold_refresh(data, 1);
}

/*! \brief Update local hold state to be unheld */
static int remote_send_unhold(void *data)
{
	return remote_send_hold_refresh(data, 0);
}

struct topology_change_refresh_data {
	struct ast_sip_session *session;
	struct ast_sip_session_media_state *media_state;
};

static void topology_change_refresh_data_free(struct topology_change_refresh_data *refresh_data)
{
	ao2_cleanup(refresh_data->session);

	ast_sip_session_media_state_free(refresh_data->media_state);
	ast_free(refresh_data);
}

static struct topology_change_refresh_data *topology_change_refresh_data_alloc(
	struct ast_sip_session *session, const struct ast_stream_topology *topology)
{
	struct topology_change_refresh_data *refresh_data;

	refresh_data = ast_calloc(1, sizeof(*refresh_data));
	if (!refresh_data) {
		return NULL;
	}

	refresh_data->session = ao2_bump(session);
	refresh_data->media_state = ast_sip_session_media_state_alloc();
	if (!refresh_data->media_state) {
		topology_change_refresh_data_free(refresh_data);
		return NULL;
	}
	refresh_data->media_state->topology = ast_stream_topology_clone(topology);
	if (!refresh_data->media_state->topology) {
		topology_change_refresh_data_free(refresh_data);
		return NULL;
	}

	return refresh_data;
}

static int on_topology_change_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (PJSIP_IS_STATUS_IN_CLASS(rdata->msg_info.msg->line.status.code, 200)) {
		/* The topology was changed to something new so give notice to what requested
		 * it so it queries the channel and updates accordingly.
		 */
		if (session->channel) {
			ast_queue_control(session->channel, AST_CONTROL_STREAM_TOPOLOGY_CHANGED);
		}
	} else if (300 <= rdata->msg_info.msg->line.status.code) {
		/* The topology change failed, so drop the current pending media state */
		ast_sip_session_media_state_reset(session->pending_media_state);
	}

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int send_topology_change_refresh_according_type(struct topology_change_refresh_data *refresh_data)
{
	int ret = 0;

	ast_log(LOG_DEBUG, "[session:%p] send_topology_change_refresh for endpoint:%s, refresh_data->session->terminal_type:%d\n",
		refresh_data->session, ast_sorcery_object_get_id(refresh_data->session->endpoint), refresh_data->session->terminal_type);

	switch (refresh_data->session->terminal_type) {
		case AST_TERMINAL_TYPE_WEBRTC:
		case AST_TERMINAL_TYPE_GSWAVE_MOBILE:
			if (ast_sip_session_media_use_bundle(refresh_data->session)) {
				ret = ast_sip_session_refresh_participant_for_sfu(refresh_data->session, 1, ast_sip_session_participant_topology_change_response,
					refresh_data->media_state);
			} else {
				ret = ast_sip_session_refresh(refresh_data->session, NULL, NULL, on_topology_change_response,
					AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, refresh_data->media_state);
			}
			break;
		default:
			ret = ast_sip_session_refresh(refresh_data->session, NULL, NULL, on_topology_change_response,
				AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, refresh_data->media_state);
			break;
	}

	return ret;
}
#endif

static int send_topology_change_refresh(void *data)
{
	struct topology_change_refresh_data *refresh_data = data;
	int ret;

#ifdef GRANDSTREAM_NETWORKS
	ret = send_topology_change_refresh_according_type(refresh_data);
#else
	ret = ast_sip_session_refresh(refresh_data->session, NULL, NULL, on_topology_change_response,
		AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, refresh_data->media_state);
#endif
	refresh_data->media_state = NULL;

	topology_change_refresh_data_free(refresh_data);

	return ret;
}

static int handle_topology_request_change(struct ast_sip_session *session,
	const struct ast_stream_topology *proposed)
{
	struct topology_change_refresh_data *refresh_data;
	int res;

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_sip_session_support_multiple_video(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Ignore topology request change for terminal type '%d', endpoint:%s\n",
			session, session->terminal_type, ast_sorcery_object_get_id(session->endpoint));
		if (!session->open_slides_video) {
			return 0;
		} else {
			ast_log(LOG_DEBUG, "[session:%p] It's opening slides video, need send re-INVITE to change SSRC for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	session->is_outgoing = 1;
#endif

	refresh_data = topology_change_refresh_data_alloc(session, proposed);
	if (!refresh_data) {
		return -1;
	}

	res = ast_sip_push_task(session->serializer, send_topology_change_refresh, refresh_data);
	if (res) {
		topology_change_refresh_data_free(refresh_data);
	}
	return res;
}

#ifdef GRANDSTREAM_NETWORKS

static int send_general_sip_info(void* data)
{
	struct pjsip_tx_data* tdata = NULL;
	struct ast_sip_session* session = NULL;
	struct general_info_data* info_data = (struct general_info_data*)data;

	if (!info_data || !info_data->session) {
		ast_log(LOG_ERROR, "[send_general_sip_info] Invalid data for info.");
		if (info_data) {
			ast_free(info_data->message);
			ast_free(info_data);
		}

		return -1;
	}

	session = (struct ast_sip_session*)info_data->session;

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "[send_general_sip_info] Create request failed.");
		ao2_ref(session, -1);
		ast_free(info_data->message);
		ast_free(info_data);
		return -1;
	}

	ast_log(LOG_NOTICE, "[send_general_sip_info] The message body:\n%s", info_data->message);

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "general_sip_control+xml",
		.body_text = info_data->message
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "[send_general_sip_info] Add tdata body failed.");
		ao2_ref(session, -1);
		ast_free(info_data->message);
		ast_free(info_data);
		return -1;
	}

	ast_sip_session_send_request(session, tdata);

	ao2_ref(session, -1);
	ast_free(info_data->message);
	ast_free(info_data);

	return 0;
}

static void close_focus_record_for_leave(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	if (!session->focus_video) {
		return;
	}

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable video recording, don't send avs mix all video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->open_slides_video || session->sfu_recv_slides_video) {
		ast_log(LOG_DEBUG, "[session:%p] Conference is opening slides video, don't send avs mix all video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);

	return;
}
#endif

/*! \brief Function called by core to ask the channel to indicate some sort of condition */
static int chan_pjsip_indicate(struct ast_channel *ast, int condition, const void *data, size_t datalen)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session_media *media;
	int response_code = 0;
	int res = 0;
	char *device_buf;
	size_t device_buf_size;
	int i;
	const struct ast_stream_topology *topology;

#ifdef GRANDSTREAM_NETWORKS
	struct general_info_data* info_data = NULL;

	const char *ishunt;
	struct ast_sip_session_media *bfcp_media = NULL;

	unsigned int video_update_info_sended = 0;

	if (channel && channel->session) {
		RAII_VAR(struct ast_sip_session *, session, ao2_bump(channel->session), ao2_cleanup);
	}

	if (ast) {
		ast_log(LOG_DEBUG, "chan[%s] in state[%d], indicate[%d]\n", ast_channel_name(ast), ast_channel_state(ast), condition);
	}
#endif

	switch (condition) {
	case AST_CONTROL_RINGING:
#ifdef GRANDSTREAM_NETWORKS
		ishunt = pbx_builtin_getvar_helper(ast, "TRIGGERHUNT");
		if (ishunt) {
			pbx_builtin_setvar_helper(ast,"EMERGENCYALERT","yes");
		}
#endif
		if (ast_channel_state(ast) == AST_STATE_RING)
		{
#ifdef GRANDSTREAM_NETWORKS
			if (channel->session->endpoint->inband_progress && !channel->session->is_direct_media && !channel->session->joint_ipvt)
#else
			if (channel->session->endpoint->inband_progress)
#endif
			{
				response_code = 183;
#ifdef GRANDSTREAM_NETWORKS
				ast_channel_cHasSend183_set(ast,1);
#endif
				res = -1;
			} else {
				response_code = 180;
#ifdef GRANDSTREAM_NETWORKS
				if (ast_channel_cHasSend183(ast)) {
					res = -1;

					if (channel->session->endpoint->ignore_180_after_183) {
						const char *have_prompt = pbx_builtin_getvar_helper(ast, "HAVE_PROMPTED");
						const char *ring_by_end = pbx_builtin_getvar_helper(ast, "RING_BY_END");
						if (ast_strlen_zero(have_prompt) || strcasecmp(have_prompt, "yes")) {

							const char *prompt_tone = pbx_builtin_getvar_helper(ast, "RINGGROUP_MOH");
							const char *in_group = pbx_builtin_getvar_helper(ast, "IN_RINGGROUP");
							if ((ast_strlen_zero(in_group) || ast_strlen_zero(prompt_tone)) &&
								ast_strlen_zero(ring_by_end)) {
								response_code = 0;
								res = 0;
							}
						}

						if (ring_by_end) {
							ast_log(LOG_NOTICE, "Have RING_BY_END value, clear it for channel:%s\n", ast_channel_name(ast));
							pbx_builtin_setvar_helper(ast, "RING_BY_END", "");
						}
					}
				}
#endif
			}
		} else {
			res = -1;
		}
		ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ast_sorcery_object_get_id(channel->session->endpoint));
		break;
	case AST_CONTROL_BUSY:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 486;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_CONGESTION:
#ifdef GRANDSTREAM_NETWORKS
		/*Begin added by zhywang . See asterisk 1.8.23 sip_indicate */
		if(ast_channel_state(ast) == AST_STATE_RING){
			/* Modify by bfdu for bug 102947, 2018.02.23 */
			if (ast_channel_hangupsipcode(ast) == 488) {
				ast_log(LOG_NOTICE, "When hangupsipcode is 488, channel is RING, response 488\n");
				response_code = 488;
			} else if (ast_channel_hangupcause(ast) == AST_CAUSE_NO_ROUTE_DESTINATION) {
				ast_log(LOG_NOTICE, "When CONGESTION, channel is RING, response 404");
				response_code = 404;
			} else {
				ast_log(LOG_NOTICE, "When CONGESTION, channel is RING, response 480");
				response_code = 480;
			}
		}
		/*End added by zhywang */
		else if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 503;
		} else {
			res = -1;
		}
#else
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 503;
		} else {
			res = -1;
		}
#endif
		break;
	case AST_CONTROL_INCOMPLETE:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 484;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_PROCEEDING:
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 100;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_PROGRESS:
#ifdef GRANDSTREAM_NETWORKS
		ishunt = pbx_builtin_getvar_helper(ast, "TRIGGERHUNT");
		if (ishunt) {
			pbx_builtin_setvar_helper(ast,"EMERGENCYALERT","yes");
		}
#endif
		if (ast_channel_state(ast) != AST_STATE_UP) {
			response_code = 183;
#ifdef GRANDSTREAM_NETWORKS
			/* Begin add by bfdu for 183 neg, 2015.09.29 */
			struct ast_channel *peer = (struct ast_channel *)data;
			if (peer && (ast_channel_size_get() == datalen)) {
				if (ast_channel_recv_early_media_rsp_get(peer) && channel->session) {
					ast_log(LOG_DEBUG, "Peer recv early media, set need_notify_avs_early_media true, peer[%p]\n", peer);
					channel->session->need_notify_avs_early_media = 1;
				}
				chan_pjsip_modify_answer_sdp(channel->session, ast, peer);
			}
			/* End add by bfdu for 183 neg, 2015.09.29 */
			ast_channel_cHasSend183_set(ast,1);
#endif
		} else {
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for make early bridge for caller is answered before */
			struct ast_channel *peer = (struct ast_channel *)data;
			if (peer && (ast_channel_size_get() == datalen)) {
				ast_log(LOG_DEBUG, "Channel:%s state is already up, notify avs to make early bridge\n", ast_channel_name(ast));
				ast_queue_control(ast, AST_CONTROL_EARLY_MEDIA);
			}
#endif
			res = -1;
		}
#ifndef GRANDSTREAM_NETWORKS
		ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ast_sorcery_object_get_id(channel->session->endpoint));
#endif
		break;
	case AST_CONTROL_VIDUPDATE:
		for (i = 0; i < AST_VECTOR_SIZE(&channel->session->active_media_state->sessions); ++i) {
			media = AST_VECTOR_GET(&channel->session->active_media_state->sessions, i);
			if (!media || media->type != AST_MEDIA_TYPE_VIDEO) {
				continue;
			}
			if (media->rtp) {
				/* FIXME: Only use this for VP8. Additional work would have to be done to
				 * fully support other video codecs */

#ifdef GRANDSTREAM_NETWORKS
				/* Modify by bfdu, gswave can't use SIP INFO to request video update, just send RTCP FIR */
				if (ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_vp8) != AST_FORMAT_CMP_NOT_EQUAL ||
					ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_vp9) != AST_FORMAT_CMP_NOT_EQUAL ||
					(channel->session->endpoint->media.webrtc &&
					 ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_h264) != AST_FORMAT_CMP_NOT_EQUAL) ||
					 channel->session->is_webrtc_call)
#else
				if (ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_vp8) != AST_FORMAT_CMP_NOT_EQUAL ||
					ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_vp9) != AST_FORMAT_CMP_NOT_EQUAL ||
					(channel->session->endpoint->media.webrtc &&
					 ast_format_cap_iscompatible_format(ast_channel_nativeformats(ast), ast_format_h264) != AST_FORMAT_CMP_NOT_EQUAL))
#endif
				{
					/* FIXME Fake RTP write, this will be sent as an RTCP packet. Ideally the
					 * RTP engine would provide a way to externally write/schedule RTCP
					 * packets */
					struct ast_frame fr;
					fr.frametype = AST_FRAME_CONTROL;
					fr.subclass.integer = AST_CONTROL_VIDUPDATE;
					res = ast_rtp_instance_write(media->rtp, &fr);
				} else {
#ifdef GRANDSTREAM_NETWORKS
					if (video_update_info_sended) {
						continue;
					}
#endif
					ao2_ref(channel->session, +1);
#ifdef HAVE_PJSIP_INV_SESSION_REF
					if (pjsip_inv_add_ref(channel->session->inv_session) != PJ_SUCCESS) {
						ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
						ao2_cleanup(channel->session);
					} else {
#endif
						if (ast_sip_push_task(channel->session->serializer, transmit_info_with_vidupdate, channel->session)) {
							ao2_cleanup(channel->session);
						}
#ifdef GRANDSTREAM_NETWORKS
						else {
							video_update_info_sended = 1;
						}
#endif

#ifdef HAVE_PJSIP_INV_SESSION_REF
					}
#endif
				}
				ast_test_suite_event_notify("AST_CONTROL_VIDUPDATE", "Result: Success");
			} else {
				ast_test_suite_event_notify("AST_CONTROL_VIDUPDATE", "Result: Failure");
				res = -1;
			}
		}
		/* XXX If there were no video streams, then this should set
		 * res to -1
		 */
		break;
	case AST_CONTROL_CONNECTED_LINE:
		ao2_ref(channel->session, +1);
#ifdef GRANDSTREAM_NETWORKS
		if (channel->session->line_index > 0) {
			if (!ast_strlen_zero(S_COR(ast_channel_connected(ast)->id.number.valid, ast_channel_connected(ast)->id.number.str, NULL))) {

				ast_log(LOG_NOTICE, "chan[%s], connected number[%s]",ast_channel_name(ast), ast_channel_connected(ast)->id.number.str);
				ast_sip_set_sca_user_info((char *)channel->session->endpoint->sca_sharedline, channel->session->line_index, NULL, ast_channel_connected(ast)->id.number.str);
			}
		}

		/* Added by yfyang, for bug 223133*/
		if (data) {
			struct ast_party_connected_line connected;
			ast_party_connected_line_init(&connected);
			if (!ast_connected_line_parse_data(data, datalen, &connected) && (connected.source == AST_CONNECTED_LINE_UPDATE_SOURCE_RPID)) {
				ast_log(LOG_NOTICE, "chan[%s], set REG_CALLEENUM to %s\n", ast_channel_name(ast), connected.id.number.str);
				pbx_builtin_setvar_helper(ast, "REG_CALLEENUM", connected.id.number.str);
				if (connected.id.name.valid) {
					pbx_builtin_setvar_helper(ast, "REG_CALLEENAME", connected.id.name.str);
				}
			}

			ast_party_connected_line_free(&connected);
		}
#endif

#ifdef HAVE_PJSIP_INV_SESSION_REF
		if (pjsip_inv_add_ref(channel->session->inv_session) != PJ_SUCCESS) {
			ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
			ao2_cleanup(channel->session);
			return -1;
		}
#endif
		if (ast_sip_push_task(channel->session->serializer, update_connected_line_information, channel->session)) {
#ifdef HAVE_PJSIP_INV_SESSION_REF
			pjsip_inv_dec_ref(channel->session->inv_session);
#endif
			ao2_cleanup(channel->session);
		}
		break;
	case AST_CONTROL_UPDATE_RTP_PEER:
		break;
	case AST_CONTROL_PVT_CAUSE_CODE:
		res = -1;
		break;
	case AST_CONTROL_MASQUERADE_NOTIFY:
		ast_assert(datalen == sizeof(int));
		if (*(int *) data) {
			/*
			 * Masquerade is beginning:
			 * Wait for session serializer to get suspended.
			 */
			ast_channel_unlock(ast);
			ast_sip_session_suspend(channel->session);
			ast_channel_lock(ast);
		} else {
			/*
			 * Masquerade is complete:
			 * Unsuspend the session serializer.
			 */
			ast_sip_session_unsuspend(channel->session);
		}
		break;
	case AST_CONTROL_HOLD:
		chan_pjsip_add_hold(ast_channel_uniqueid(ast));
		device_buf_size = strlen(ast_channel_name(ast)) + 1;
		device_buf = alloca(device_buf_size);
		ast_channel_get_device_name(ast, device_buf, device_buf_size);
		ast_devstate_changed_literal(AST_DEVICE_ONHOLD, 1, device_buf);
		if (!channel->session->endpoint->moh_passthrough) {
#ifdef GRANDSTREAM_NETWORKS
			// Skip Music on hold in a conference
			struct chan_pjsip_pvt *pvt = channel->pvt;
			if (pvt && pvt->call_type && (!strcasecmp(pvt->call_type, "Conference") || !strcasecmp(pvt->call_type, "VideoConference"))) {
				ast_log(LOG_DEBUG, "[session:%p] Channel:%s, Call Type:%s, don't play music on hold in a conference\n",
					channel->session, ast_channel_name(ast), S_OR(pvt->call_type, "NULL"));
			} else {
				ast_moh_start(ast, data, NULL);
			}
#else
			ast_moh_start(ast, data, NULL);
#endif
		} else {
			if (ast_sip_push_task(channel->session->serializer, remote_send_hold, ao2_bump(channel->session))) {
				ast_log(LOG_WARNING, "Could not queue task to remotely put session '%s' on hold with endpoint '%s'\n",
					ast_sorcery_object_get_id(channel->session), ast_sorcery_object_get_id(channel->session->endpoint));
				ao2_ref(channel->session, -1);
			}
		}
		break;
	case AST_CONTROL_UNHOLD:
		chan_pjsip_remove_hold(ast_channel_uniqueid(ast));
		device_buf_size = strlen(ast_channel_name(ast)) + 1;
		device_buf = alloca(device_buf_size);
		ast_channel_get_device_name(ast, device_buf, device_buf_size);
		ast_devstate_changed_literal(AST_DEVICE_UNKNOWN, 1, device_buf);
		if (!channel->session->endpoint->moh_passthrough) {
			ast_moh_stop(ast);
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu, change ssrc for gswave when transfer */
			if (ast_channel_get_feature_transfer(ast)) {
				if (channel->session->is_webrtc_call) {
					ast_log(LOG_DEBUG, "It's webrtc and feature transfer, need send re-invite to change SSRC\n");
					send_reinivte_for_change_ssrc(channel->session);
					ast_queue_control(ast, AST_CONTROL_CHANGE_SSRC);
				} else {
					ast_queue_control(ast, AST_CONTROL_GET_PEER_SESSION);
					ast_channel_set_feature_transfer(ast, 0);
				}
			}
#endif
		} else {
			if (ast_sip_push_task(channel->session->serializer, remote_send_unhold, ao2_bump(channel->session))) {
				ast_log(LOG_WARNING, "Could not queue task to remotely take session '%s' off hold with endpoint '%s'\n",
					ast_sorcery_object_get_id(channel->session), ast_sorcery_object_get_id(channel->session->endpoint));
				ao2_ref(channel->session, -1);
			}
		}
		break;
	case AST_CONTROL_BLIND_REFER_UNHOLD:
		if (ast_sip_session_get_allow_forward_hold_request()) {
			ast_log(LOG_DEBUG, "Send unhold request for local channel blind refer.\n");
			ast_sip_session_init_stream_media_direction(channel->session);
			if (ast_sip_push_task(channel->session->serializer, remote_send_unhold, ao2_bump(channel->session))) {
				ast_log(LOG_WARNING, "Could not queue task to remotely take session '%s' off hold with endpoint '%s'\n",
				ast_sorcery_object_get_id(channel->session), ast_sorcery_object_get_id(channel->session->endpoint));
				ao2_ref(channel->session, -1);
			}
		}
		break;
	case AST_CONTROL_SRCUPDATE:
		break;
	case AST_CONTROL_SRCCHANGE:
		break;
	case AST_CONTROL_REDIRECTING:
		if (ast_channel_state(ast) != AST_STATE_UP) {
#ifdef GRANDSTREAM_NETWORKS
			if (!ast_channel_respond_18x_flag(ast)) {
				/* send 180 first */
				struct indicate_data *ind_data1 = indicate_data_alloc(channel->session, condition, 180, data, datalen);
				if (!ind_data1) {
					return -1;
				}
#ifdef HAVE_PJSIP_INV_SESSION_REF
				if (pjsip_inv_add_ref(ind_data1->session->inv_session) != PJ_SUCCESS) {
					ao2_cleanup(ind_data1);
					return -1;
				}
#endif

				ast_channel_unlock(ast);
				if (ast_sip_push_task_wait_serializer(channel->session->serializer, indicate, ind_data1)) {
#ifdef HAVE_PJSIP_INV_SESSION_REF
					pjsip_inv_dec_ref(ind_data1->session->inv_session);
#endif
					ao2_cleanup(ind_data1);
				}
				ast_channel_lock(ast);
				ast_channel_respond_18x_flag_set(ast, 1);
			}
#endif
			response_code = 181;
		} else {
			res = -1;
		}
		break;
	case AST_CONTROL_T38_PARAMETERS:
		res = 0;

		if (channel->session->t38state == T38_PEER_REINVITE) {
			const struct ast_control_t38_parameters *parameters = data;

			if (parameters->request_response == AST_T38_REQUEST_PARMS) {
				res = AST_T38_REQUEST_PARMS;
			}
		}

		break;
	case AST_CONTROL_STREAM_TOPOLOGY_REQUEST_CHANGE:
		topology = data;
		res = handle_topology_request_change(channel->session, topology);
		break;
	case AST_CONTROL_STREAM_TOPOLOGY_CHANGED:
		break;
	case AST_CONTROL_STREAM_TOPOLOGY_SOURCE_CHANGED:
		break;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for video reinivite */
	case AST_CONTROL_REINVITE_REQ:
		if (datalen != sizeof(struct ast_sip_reinvite_data)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_REINVITE_REQ. Expected %d, got %d\n", (int) sizeof(struct ast_sip_reinvite_data), (int) datalen);
		} else {
			process_reinvite_req_control_frame(channel->session, data);
		}
		break;
	case AST_CONTROL_REINVITE_RSP:
		if (datalen != sizeof(struct ast_sip_reinvite_data)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_REINVITE_RSP. Expected %d, got %d\n", (int) sizeof(struct ast_sip_reinvite_data), (int) datalen);
		} else {
			process_reinvite_rsp_control_frame(channel->session, data);
		}
		break;
	case AST_CONTROL_BFCP_TCP_DATA:
		bfcp_media = ast_sip_session_get_bfcp_active_media(channel->session);
		if (bfcp_media) {
			if (bfcp_media->bfcp_tcp) {
				ast_bfcp_tcp_write(bfcp_media->bfcp_tcp, data, datalen);
			}
		}
		break;
	case AST_CONTROL_REINVITE_BYE_CONFLICT:
		ao2_ref(channel->session, +1);
		if (ast_sip_push_task(channel->session->serializer, send_reinvite_response, channel->session)) {
			ao2_cleanup(channel->session);
		}
		break;
	case AST_CONTROL_REINVITE_FAILED:
		if (datalen != sizeof(int)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_REINVITE_FAILED. Expected %d, got %d\n", (int) sizeof(int), (int) datalen);
		} else {
			process_reinvite_failed_frame(channel->session, data);
		}
		break;
	case AST_CONTROL_BFCP_CONTROL:
		if (datalen != sizeof(struct ast_bfcp_info)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_BFCP_CONTROL. Expected %d, got %d\n", (int) sizeof(struct ast_bfcp_info), (int) datalen);
		} else {
			process_bfcp_control_frame(channel->session, data);
		}
		break;
	case AST_CONTROL_GERERAL_SIP_INFO:
		info_data = (struct general_info_data*)ast_malloc(sizeof(struct general_info_data));
		if (!info_data) {
			res = -1;
		} else {
			ast_log(LOG_NOTICE, "[AST_CONTROL_GERERAL_SIP_INFO] Recv control frame.\n");
			info_data->session = ao2_bump(channel->session);
			info_data->message = ast_strdup(data);
			if (ast_sip_push_task(channel->session->serializer, send_general_sip_info, info_data)) {
				ast_log(LOG_WARNING, "[session:%p] Send sip info push to task failed for endpoint:%s\n",
					channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
				ao2_ref(info_data->session, -1);
				ast_free(info_data->message);
				ast_free(info_data);
			}
		}
		break;
	case AST_CONTROL_CONF_HOST_CHANGED:
		ast_log(LOG_DEBUG, "[session:%p] Recv AST_CONTROL_CONF_HOST_CHANGED for endpoint:%s\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		update_msid_for_change_host(channel->session, (char *)data);
		break;
	case AST_CONTROL_CHANGE_SSRC:
		ast_log(LOG_DEBUG, "[session:%p] Recv AST_CONTROL_CHANGE_SSRC, endpoint:%s\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		send_reinivte_for_change_ssrc(channel->session);
		break;
	case AST_CONTROL_JOIN_SFU:
		ast_log(LOG_DEBUG, "[session:%p] Recv AST_CONTROL_JOIN_SFU, for endpoint:%s\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_join_sfu_control_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_N_WAY_CONF:
		send_reinvite_for_neg_slides_video(channel->session);
		send_notify_to_checkout_conf(channel->session, data, datalen);
		break;
	case AST_CONTROL_UPDATE_CANDIDATES:
		ast_log(LOG_DEBUG, "[session:%p] Recv AST_CONTROL_UPDATE_CANDIDATES, for endpoint:%s\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_update_candidates_control_frame(channel->session, data);
		break;
	case AST_CONTROL_TRANSFER_COMPLETE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_TRANSFER_COMPLETE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		notify_released_bfcp_for_transfer(channel->session);
		notify_trunk_info_for_transfer(channel->session);
		ast_queue_control(ast, AST_CONTROL_GET_PEER_SESSION);
		break;
	case AST_CONTROL_CHANNEL_IS_NOT_SIP:
		if (channel->session->ref_added) {
			if (TRUE == channel->session->is_media_change) {
				disable_media_except_audio(ast);
				ast_channel_unlock(ast);
				/* when receive re-INVITE, session ref is alreay add one, so don't add ref here */
				if (ast_sip_push_task_synchronous(channel->session->serializer, send_reinvite_response, channel->session)) {
					ast_log(LOG_WARNING, "Unable to push send_reinvite_response task to the threadpool.\n");
					ao2_ref(channel->session, -1);
				}
				ast_channel_lock(ast);
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Don't need to send video re-invite response, endpoint:%s\n",
					channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
				ao2_ref(channel->session, -1);
			}

			channel->session->ref_added = 0;
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Don't need to process, ref not added before, endpoint:%s\n",
				channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		}
		break;
	/* Added by yfyang, for Broadsoft Test */
	case AST_CONTROL_PAI_INFO:
		ast_log(LOG_NOTICE, "chan[%s], indicate PAI INFO", ast_channel_name(ast));
		set_pai_payload(ast, data, datalen);
		break;
	case AST_CONTROL_GET_PEER_SESSION:
		ast_sip_session_send_session_to_peer(channel->session);
		break;
	case AST_CONTROL_PEER_SESSION:
		if (datalen != sizeof(struct ast_sip_reinvite_data)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_PEER_SESSION. Expected %d, got %d for channel:%s\n",
				(int) sizeof(struct ast_sip_reinvite_data), (int) datalen, ast_channel_name(ast));
		} else {
			const struct ast_sip_reinvite_data *reinvite_data = data;
			if (reinvite_data->session->ref_added) {
				process_topology_control_frame(channel->session, reinvite_data->session);
				ao2_ref(reinvite_data->session, -1);
				reinvite_data->session->ref_added = 0;
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Seesion's ref is not added, don't need process,endpoint:%s\n",
					channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
			}
		}
		break;
	case AST_CONTROL_UPDATE_VIDEO_PORT_ZERO:
		break;
	case AST_CONTROL_REFER_COMPLETE:
		notify_bfcp_for_refer_transfer(channel->session);
		ast_sip_session_send_session_to_peer(channel->session);
		ast_queue_control(ast, AST_CONTROL_GET_PEER_SESSION);
		break;
	case AST_CONTROL_TOPOLOGY_CHANGE_FOR_LOCAL:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_TOPOLOGY_CHANGE_FOR_LOCAL frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		const struct ast_channel_topology_data *topology_data = data;
		res = handle_topology_request_change(channel->session, topology_data->topology);
		break;
	case AST_CONTROL_UPDATE_TURN_SESSION_COUNT:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_UPDATE_TURN_SESSION_COUNT frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		res = process_update_turn_session_count_frame(channel->session, data, datalen);
		break;
	case AST_CONTROL_TURN_SESSION_LIMITED:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_TURN_SESSION_LIMITED frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		ast_sip_decrease_used_turn_session_count(ast_channel_get_turn_session_count(ast));
		ast_channel_set_turn_session_count(ast, 0);
		ast_channel_hangupcause_set(ast, AST_CAUSE_TURN_SESSION_LIMITED);
		ast_softhangup(ast, AST_CAUSE_TURN_SESSION_LIMITED);
		break;
	case AST_CONTROL_PEER_IS_NAT_CALL:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_PEER_IS_NAT_CALL frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		ast_channel_set_peer_is_nat_call(ast, 1);
		ast_channel_set_external_call_event(ast);
		break;
	case AST_CONTROL_NOTIFY_GDS_CALL:
		ast_channel_set_external_call_event(ast);
		ast_queue_control(ast, AST_CONTROL_NOTIFY_PEER_GDS_CALL);
		break;
	case AST_CONTROL_NOTIFY_PEER_GDS_CALL:
		ast_channel_set_external_call_event(ast);
		break;
	case AST_CONTROL_NOTIFY_GSC_CALL:
		ast_channel_set_external_call_event(ast);
		ast_queue_control(ast, AST_CONTROL_NOTIFY_PEER_GSC_CALL);
		break;
	case AST_CONTROL_NOTIFY_PEER_GSC_CALL:
		ast_channel_set_external_call_event(ast);
		break;
	case AST_CONTROL_CAMERA_CONTROL:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_CAMERA_CONTROL frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		res = process_camera_control_frame(channel->session, data, datalen);
		break;
	case AST_CONTROL_HOST_CAMERA_CHANGE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_HOST_CAMERA_CHANGE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_camera_change_frame(channel->session, (char *)data);
		break;
	case AST_CONTROL_FAX_CNG_TONE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s  receive AST_CONTROL_FAX_CNG_TONE,need check fax_alreay_work:%d\n",channel->session, ast_sorcery_object_get_id(channel->session->endpoint),channel->session->fax_alreay_work);

		const char* enable_fax_dectct = pbx_builtin_getvar_helper(ast, "enable_fax_detect");
		int enable_fax = 0;
		if (channel->session->wFaxDetect || (!ast_strlen_zero(enable_fax_dectct) && !strcasecmp(enable_fax_dectct,"yes") && !strncasecmp(ast_channel_name(ast), "PJSIP/trunk_", 12))){
			enable_fax =  1;
		}

		if (enable_fax && !(ast_channel_get_is_send_fax(ast)) && !channel->session->fax_alreay_work) {
			chan_pjsip_avs_cng_tone_detected(channel->session);
		}
		break;
	case AST_CONTROL_NOTIFY_CEI_STATUS:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CEI_STATUS frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_notify_cei_status_frame(channel->session, data, datalen);
		break;
	/* added by zgmo for indicating sdp record attr has changed */
	case AST_CONTROL_UPDATE_RECORD_SDP_ATTR:
	{
		/* Added by yfyang, for Bug 197337 ,edit by zyxia for Bug267240*/
		if (!strncmp(ast_channel_name(ast), "PJSIP/trunk_", 12) && !channel->session->ucm_trunk) {
			break;
		}

		int trigger = 0;
		const char *rec_enable = pbx_builtin_getvar_helper(ast, "REC_ENABLE");
		const char *record_atr_sdp_trigger = pbx_builtin_getvar_helper(ast, "RECORD_ATTR_SDP_TRIGGER");

		ast_log(LOG_DEBUG, "[RecordSDP]chan:%s, rec_enable:%s, session->auto_record:%d", ast_channel_name(ast),
			ast_strlen_zero(rec_enable) ? "":rec_enable,
			channel->session->auto_record);

		if ((!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes"))) {
			trigger = channel->session->auto_record ? 0 : 1;
		}else {
			trigger = channel->session->auto_record ? 1 : 0;
		}

		if (!ast_strlen_zero(record_atr_sdp_trigger) && !strcmp(record_atr_sdp_trigger, "yes")) {
			trigger = 1;
			pbx_builtin_setvar_helper(ast, "RECORD_ATTR_SDP_TRIGGER", "");
		}

		ast_log(LOG_DEBUG, "trigger:%d", trigger);

		if (trigger) {
			ao2_ref(channel->session, +1);
#ifdef HAVE_PJSIP_INV_SESSION_REF
			if (pjsip_inv_add_ref(channel->session->inv_session) != PJ_SUCCESS) {
				ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
				ao2_cleanup(channel->session);
				return -1;
			}
#endif
			if (ast_sip_push_task(channel->session->serializer, send_update_record_attr, channel->session)) {
#ifdef HAVE_PJSIP_INV_SESSION_REF
				pjsip_inv_dec_ref(channel->session->inv_session);
#endif
				ao2_cleanup(channel->session);
			}
		}
	}

	break;
	/* end */
	case AST_CONTROL_WAIT_JOIN_CONFERENCE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_WAIT_JOIN_CONFERENCE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_wait_join_conference_frame(channel->session);
		break;
	case AST_CONTROL_JOIN_NWAY_CONF:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_JOIN_NWAY_CONF frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_join_nway_conf_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_RECV_BYE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_RECV_BYE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		hangup_some_associated_call(channel->session);
		break;
	case AST_CONTROL_NOTIFY_SLIDER_MSID:
		ast_log(LOG_DEBUG, "[session:%p] Recv AST_CONTROL_NOTIFY_SLIDER_MSID for endpoint:%s\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		update_avs_slides_video_msid(channel->session, (char *)data);
		channel->session->update_avs_slides_video = 1;
		channel->session->open_slides_video = 1;
		break;
	case AST_CONTROL_NOTIFY_CLOSE_CAMERA:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CLOSE_CAMERA frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_close_camera_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_START:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_START frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_conf_video_record_start_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_STOP:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CONF_VIDEO_RECORD_STOP frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		channel->session->conf_video_recording = 0;
		break;
	case AST_CONTROL_NOTIFY_OPEN_SLIDES_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_OPEN_SLIDES_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_open_slides_video_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_stop_slides_video_frame(channel->session);
	case AST_CONTROL_NOTIFY_GS_STATE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_GS_STATE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_gs_state_frame(channel->session, data, datalen);
		break;
	case AST_CONTROL_AMI_HOLD:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_AMI_HOLD frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_ami_hold_frame(channel->session);
		break;
	case AST_CONTROL_AMI_UNHOLD:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_AMI_UNHOLD frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_ami_unhold_frame(channel->session);
		break;
	case AST_CONTROL_TOTAL_TIME_LIMITED:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_TOTAL_TIME_LIMITED frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		ast_sip_decrease_used_turn_session_count(ast_channel_get_turn_session_count(ast));
		ast_channel_set_turn_session_count(ast, 0);
		ast_channel_hangupcause_set(ast, AST_CAUSE_TOTAL_TIME_LIMITED);
		ast_softhangup(ast, AST_CAUSE_TOTAL_TIME_LIMITED);
		break;
	case AST_CONTROL_NOTIFY_CONF_TIMEOUT:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CONF_TIMEOUT frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_conf_timeout_notify_frame(channel->session, data, datalen);
		break;
	case AST_CONTROL_NOTIFY_OPEN_FOCUS_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_OPEN_FOCUS_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_open_focus_video_notify_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_CLOSE_FOCUS_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CLOSE_FOCUS_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_close_focus_video_notify_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_STOP_RECORD:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_STOP_RECORD frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_stop_record_frame(channel->session, data, datalen);
		break;
	case AST_CONTROL_NOTIFY_CHANNEL_LEAVE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_CHANNEL_LEAVE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		close_focus_record_for_leave(channel->session);
		break;
	case AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO_BY_HOST:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO_BY_HOST frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		if (datalen != sizeof(struct ast_bfcp_info)) {
			res = -1;
			ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO_BY_HOST. Expected %d, got %d\n", (int) sizeof(struct ast_bfcp_info), (int) datalen);
		} else {
			process_stop_slides_video_by_host_frame(channel->session, data);
		}
		break;
	case AST_CONTROL_CLOSE_CAMERA:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_CLOSE_CAMERA frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		send_reinvite_close_camera(channel->session, INACTIVE);
		break;
	case AST_CONTROL_NOTIFY_DISABLE_FOCUS_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_DISABLE_FOCUS_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_disable_focus_video_notify_frame(channel->session);
		break;
	case AST_CONTROL_QUEUE_REINVITE_REQ_FAILED:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_QUEUE_REINVITE_REQ_FAILED frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_reinvite_req_failed_frame(channel->session);
		break;
	case AST_CONTROL_QUEUE_REINVITE_RSP_FAILED:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_QUEUE_REINVITE_RSP_FAILED frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_reinvite_rsp_failed_frame(channel->session);
		break;
	case AST_CONTROL_QUEUE_REINVITE_RSP_AGAIN:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_QUEUE_REINVITE_RSP_AGAIN frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_reinvite_rsp_again_frame(channel->session);
		break;
	case AST_CONTROL_QUEUE_PEER_SESSION_FAILED:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_QUEUE_REINVITE_RSP_FAILED frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_queue_peer_session_failed_frame(channel->session);
		break;
	case AST_CONTROL_UPDATE_PAGE_MSID:
		ast_log(LOG_DEBUG, "[session:%p] Recv AST_CONTROL_UPDATE_PAGE_MSID for endpoint:%s\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		update_msid_for_change_host(channel->session, (char *)data);
		break;
	case AST_CONTROL_WAVE_RELEASED_HD_RESOLUTION:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_WAVE_RELEASED_HD_RESOLUTION frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_wave_released_hd_resolution_frame(channel->session);
		break;
	case AST_CONTROL_NOTIFY_TRUNK_INFO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_TRUNK_INFO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		res = process_notify_trunk_info_frame(channel->session, data, datalen);
		break;
	case AST_CONTROL_PAGE_OPEN_FOCUS_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_PAGE_OPEN_FOCUS_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		ast_sip_session_send_avs_mcu_control(channel->session, AVS_MCU_MIX_TYPE_FOCUS_VIDEO);
		break;

	case AST_CONTROL_TRUNK_CALL_TO_CONF:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_TRUNK_CALL_TO_CONF frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		res = process_change_to_trunk_conf_frame(channel->session, (char *)data);
		break;
	case AST_CONTROL_NOTIFY_HANGUP_SWITCH_TO_CONF:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_HANGUP_SWITCH_TO_CONF frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		ast_channel_hangupcause_set(ast, AST_CAUSE_SWITCH_TO_CONF_MODE);
		break;
	case AST_CONTROL_END_CONF:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_END_CONF frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		process_conf_end_control_frame(channel->session);
		break;
	case AST_CONTROL_REAL_ANSWERED:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_REAL_ANSWERED frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		if(!channel->session->had_send_answered) { //added  by zyxia for Bug 262501
			if (ast_sip_push_task(channel->session->serializer, process_real_answered_control_frame, ao2_bump(channel->session))) {
				ao2_cleanup(channel->session);
			}
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s had handled AST_CONTROL_REAL_ANSWERED frame before,ignore it\n",
				channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		}
		break;

	case AST_CONTROL_UPDATE_MAIN_VIDEO:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_UPDATE_MAIN_VIDEO frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		struct main_video_data *main_video = ao2_alloc(sizeof(*main_video), main_video_dtor);
		if (main_video) {
			main_video->session = ao2_bump(channel->session);
			main_video->data = data;
			if (ast_sip_push_task(channel->session->serializer, process_update_main_video_control_frame, main_video)) {
				ao2_cleanup(main_video);
			}
		}
		break;

	case AST_CONTROL_PLAY_PAGING_VIDEO_RECORD:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_PLAY_PAGING_VIDEO_RECORD frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		char* file = (char*)data;
		if (!ast_strlen_zero(file))
			ast_streamfile(channel->session->channel, file, ast_channel_language(channel->session->channel));

		break;

	case AST_CONTROL_CALLFLIP_REINVITE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_CALLFLIP_REINVITE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		res = process_callflip_reinvite_frame(channel->session, data, datalen);
		break;

	case AST_CONTROL_PLAY_PROMPT:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_PLAY_PROMPT frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
		//res = process_play_prompt_to_caller(channel->session->channel);
		ao2_ref(channel->session, +1);
		if (ast_sip_push_task(channel->session->serializer, process_play_prompt_to_caller, channel->session)) {
			ao2_cleanup(channel->session);
		}

		break;
	/*Start added by zyxia for Bug 291797,2024.3.14*/
	case AST_CONTROL_NOTIFY_PARKING_STATE:
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive AST_CONTROL_NOTIFY_PARKING_STATE frame\n",
			channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
			if(channel->session->endpoint)
				ast_sip_for_each_aor(channel->session->endpoint->aors, process_parkingstate_aor, (void *)data);
		break;
	/*End added by zyxia for Bug 291797,2024.3.14*/
	case AST_CONTROL_OPEN_IPC_VIDEO:
		ao2_ref(channel->session, +1);
		if (ast_sip_push_task(channel->session->serializer, process_open_ipc_video, channel->session)) {
			ao2_cleanup(channel->session);
		}

		break;
#endif
	case -1:
		res = -1;
		break;
	default:
		ast_log(LOG_WARNING, "Don't know how to indicate condition %d\n", condition);
		res = -1;
		break;
	}

	if (response_code) {
		struct indicate_data *ind_data = indicate_data_alloc(channel->session, condition, response_code, data, datalen);

		if (!ind_data) {
			return -1;
		}
#ifdef HAVE_PJSIP_INV_SESSION_REF
		if (pjsip_inv_add_ref(ind_data->session->inv_session) != PJ_SUCCESS) {
			ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
			ao2_cleanup(ind_data);
			return -1;
		}
#endif

#ifdef GRANDSTREAM_NETWORKS
		if ((response_code == 183) || (response_code == 180)) {
			ast_channel_respond_18x_flag_set(ast, 1);
		}

		if (response_code == 183) {
			ast_channel_unlock(ast);
			if (ast_sip_push_task_wait_serializer(channel->session->serializer, indicate, ind_data)) {
				ast_log(LOG_NOTICE, "Cannot send response code %d to endpoint %s. Could not queue task properly\n",
					response_code, ast_sorcery_object_get_id(channel->session->endpoint));
#ifdef HAVE_PJSIP_INV_SESSION_REF
				pjsip_inv_dec_ref(ind_data->session->inv_session);
#endif
				ao2_cleanup(ind_data);
				res = -1;
			}
			ast_channel_lock(ast);
		} else {
			if (ast_sip_push_task(channel->session->serializer, indicate, ind_data)) {
				ast_log(LOG_NOTICE, "Cannot send response code %d to endpoint %s. Could not queue task properly\n",
						response_code, ast_sorcery_object_get_id(channel->session->endpoint));
#ifdef HAVE_PJSIP_INV_SESSION_REF
				pjsip_inv_dec_ref(ind_data->session->inv_session);
#endif
				ao2_cleanup(ind_data);
				res = -1;
			}
		}
#else
		if (ast_sip_push_task(channel->session->serializer, indicate, ind_data)) {
			ast_log(LOG_NOTICE, "Cannot send response code %d to endpoint %s. Could not queue task properly\n",
					response_code, ast_sorcery_object_get_id(channel->session->endpoint));
#ifdef HAVE_PJSIP_INV_SESSION_REF
			pjsip_inv_dec_ref(ind_data->session->inv_session);
#endif
			ao2_cleanup(ind_data);
			res = -1;
		}
#endif
	}

	return res;
}

struct transfer_data {
	struct ast_sip_session *session;
	char *target;
};

static void transfer_data_destroy(void *obj)
{
	struct transfer_data *trnf_data = obj;

	ast_free(trnf_data->target);
	ao2_cleanup(trnf_data->session);
}

static struct transfer_data *transfer_data_alloc(struct ast_sip_session *session, const char *target)
{
	struct transfer_data *trnf_data = ao2_alloc(sizeof(*trnf_data), transfer_data_destroy);

	if (!trnf_data) {
		return NULL;
	}

	if (!(trnf_data->target = ast_strdup(target))) {
		ao2_ref(trnf_data, -1);
		return NULL;
	}

	ao2_ref(session, +1);
	trnf_data->session = session;

	return trnf_data;
}

static void transfer_redirect(struct ast_sip_session *session, const char *target)
{
	pjsip_tx_data *packet;
	enum ast_control_transfer message = AST_TRANSFER_SUCCESS;
	pjsip_contact_hdr *contact;
	pj_str_t tmp;

	if (pjsip_inv_end_session(session->inv_session, 302, NULL, &packet) != PJ_SUCCESS
		|| !packet) {
		ast_log(LOG_WARNING, "Failed to redirect PJSIP session for channel %s\n",
			ast_channel_name(session->channel));
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));

		return;
	}

	if (!(contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL))) {
		contact = pjsip_contact_hdr_create(packet->pool);
	}

	pj_strdup2_with_null(packet->pool, &tmp, target);
	if (!(contact->uri = pjsip_parse_uri(packet->pool, tmp.ptr, tmp.slen, PJSIP_PARSE_URI_AS_NAMEADDR))) {
		ast_log(LOG_WARNING, "Failed to parse destination URI '%s' for channel %s\n",
			target, ast_channel_name(session->channel));
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
		pjsip_tx_data_dec_ref(packet);

		return;
	}
	pjsip_msg_add_hdr(packet->msg, (pjsip_hdr *) contact);

	ast_sip_session_send_response(session, packet);
	ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
}

static void transfer_refer(struct ast_sip_session *session, const char *target)
{
	pjsip_evsub *sub;
	enum ast_control_transfer message = AST_TRANSFER_SUCCESS;
	pj_str_t tmp;
	pjsip_tx_data *packet;
	const char *ref_by_val;
	char local_info[pj_strlen(&session->inv_session->dlg->local.info_str) + 1];

	if (pjsip_xfer_create_uac(session->inv_session->dlg, NULL, &sub) != PJ_SUCCESS) {
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));

		return;
	}

	if (pjsip_xfer_initiate(sub, pj_cstr(&tmp, target), &packet) != PJ_SUCCESS) {
		message = AST_TRANSFER_FAILED;
		ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
		pjsip_evsub_terminate(sub, PJ_FALSE);

		return;
	}

	ref_by_val = pbx_builtin_getvar_helper(session->channel, "SIPREFERREDBYHDR");
	if (!ast_strlen_zero(ref_by_val)) {
		ast_sip_add_header(packet, "Referred-By", ref_by_val);
	} else {
		ast_copy_pj_str(local_info, &session->inv_session->dlg->local.info_str, sizeof(local_info));
		ast_sip_add_header(packet, "Referred-By", local_info);
	}

	pjsip_xfer_send_request(sub, packet);
	ast_queue_control_data(session->channel, AST_CONTROL_TRANSFER, &message, sizeof(message));
}

static int transfer(void *data)
{
	struct transfer_data *trnf_data = data;
	struct ast_sip_endpoint *endpoint = NULL;
	struct ast_sip_contact *contact = NULL;
	const char *target = trnf_data->target;

	if (trnf_data->session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			trnf_data->session->inv_session->cause,
			pjsip_get_status_text(trnf_data->session->inv_session->cause)->ptr);
	} else {
		/* See if we have an endpoint; if so, use its contact */
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", target);
		if (endpoint) {
			contact = ast_sip_location_retrieve_contact_from_aor_list(endpoint->aors);
			if (contact && !ast_strlen_zero(contact->uri)) {
				target = contact->uri;
			}
		}

		if (ast_channel_state(trnf_data->session->channel) == AST_STATE_RING) {
			transfer_redirect(trnf_data->session, target);
		} else {
			transfer_refer(trnf_data->session, target);
		}
	}

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(trnf_data->session->inv_session);
#endif

	ao2_ref(trnf_data, -1);
	ao2_cleanup(endpoint);
	ao2_cleanup(contact);
	return 0;
}

/*! \brief Function called by core for Asterisk initiated transfer */
static int chan_pjsip_transfer(struct ast_channel *chan, const char *target)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct transfer_data *trnf_data = transfer_data_alloc(channel->session, target);

	if (!trnf_data) {
		return -1;
	}

#ifdef HAVE_PJSIP_INV_SESSION_REF
	if (pjsip_inv_add_ref(trnf_data->session->inv_session) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
		ao2_cleanup(trnf_data);
		return -1;
	}
#endif

	if (ast_sip_push_task(channel->session->serializer, transfer, trnf_data)) {
		ast_log(LOG_WARNING, "Error requesting transfer\n");
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(trnf_data->session->inv_session);
#endif
		ao2_cleanup(trnf_data);
		return -1;
	}

	return 0;
}

/*! \brief Function called by core to start a DTMF digit */
static int chan_pjsip_digit_begin(struct ast_channel *chan, char digit)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_session_media *media;
	int res = 0;

	media = channel->session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO];

	switch (channel->session->dtmf) {
	case AST_SIP_DTMF_RFC_4733:
		if (!media || !media->rtp) {
			return -1;
		}

		ast_rtp_instance_dtmf_begin(media->rtp, digit);
		break;
	case AST_SIP_DTMF_AUTO:
		if (!media || !media->rtp || (ast_rtp_instance_dtmf_mode_get(media->rtp) == AST_RTP_DTMF_MODE_INBAND)) {
			return -1;
		}

		ast_rtp_instance_dtmf_begin(media->rtp, digit);
		break;
	case AST_SIP_DTMF_AUTO_INFO:
		if (!media || !media->rtp || (ast_rtp_instance_dtmf_mode_get(media->rtp) == AST_RTP_DTMF_MODE_NONE)) {
			return -1;
		}
		ast_rtp_instance_dtmf_begin(media->rtp, digit);
		break;
	case AST_SIP_DTMF_NONE:
		break;
	case AST_SIP_DTMF_INBAND:
		res = -1;
		break;
	default:
		break;
	}

	return res;
}

struct info_dtmf_data {
	struct ast_sip_session *session;
	char digit;
	unsigned int duration;
};

static void info_dtmf_data_destroy(void *obj)
{
	struct info_dtmf_data *dtmf_data = obj;
	ao2_ref(dtmf_data->session, -1);
}

static struct info_dtmf_data *info_dtmf_data_alloc(struct ast_sip_session *session, char digit, unsigned int duration)
{
	struct info_dtmf_data *dtmf_data = ao2_alloc(sizeof(*dtmf_data), info_dtmf_data_destroy);
	if (!dtmf_data) {
		return NULL;
	}
	ao2_ref(session, +1);
	dtmf_data->session = session;
	dtmf_data->digit = digit;
	dtmf_data->duration = duration;
	return dtmf_data;
}

static int transmit_info_dtmf(void *data)
{
	RAII_VAR(struct info_dtmf_data *, dtmf_data, data, ao2_cleanup);

	struct ast_sip_session *session = dtmf_data->session;
	struct pjsip_tx_data *tdata;

	RAII_VAR(struct ast_str *, body_text, NULL, ast_free_ptr);

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "dtmf-relay",
	};

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			session->inv_session->cause,
			pjsip_get_status_text(session->inv_session->cause)->ptr);
		goto failure;
	}

	if (!(body_text = ast_str_create(32))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for INFO DTMF.\n");
		goto failure;
	}
	ast_str_set(&body_text, 0, "Signal=%c\r\nDuration=%u\r\n", dtmf_data->digit, dtmf_data->duration);

	body.body_text = ast_str_buffer(body_text);

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create DTMF INFO request\n");
		goto failure;
	}
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to DTMF INFO request\n");
		pjsip_tx_data_dec_ref(tdata);
		goto failure;
	}
	ast_sip_session_send_request(session, tdata);

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif

	return 0;

failure:
#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif
	return -1;

}

/*! \brief Function called by core to stop a DTMF digit */
static int chan_pjsip_digit_end(struct ast_channel *ast, char digit, unsigned int duration)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session_media *media;
	int res = 0;

	media = channel->session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO];

	switch (channel->session->dtmf) {
	case AST_SIP_DTMF_AUTO_INFO:
	{
		if (!media || !media->rtp) {
			return -1;
		}
		if (ast_rtp_instance_dtmf_mode_get(media->rtp) != AST_RTP_DTMF_MODE_NONE) {
			ast_debug(3, "Told to send end of digit on Auto-Info channel %s RFC4733 negotiated so using it.\n", ast_channel_name(ast));
#ifdef GRANDSTREAM_NETWORKS
			if (ast_opt_avs) {
				struct ast_dtmf_send_param dtmfreq = {0};
				dtmfreq.digit = digit;
				dtmfreq.opt = 1;
				res = ast_avs_control(ast, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
				if (!res){
					usleep(100000);
					dtmfreq.opt = 0;
					ast_avs_control(ast, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
				}
			} else {
#endif
				ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
#ifdef GRANDSTREAM_NETWORKS
			}
#endif
			break;
		}
		/* If RFC_4733 was not negotiated, fail through to the DTMF_INFO processing */
		ast_debug(3, "Told to send end of digit on Auto-Info channel %s RFC4733 NOT negotiated using INFO instead.\n", ast_channel_name(ast));
	}

	case AST_SIP_DTMF_INFO:
	{
		struct info_dtmf_data *dtmf_data = info_dtmf_data_alloc(channel->session, digit, duration);

		if (!dtmf_data) {
			return -1;
		}

#ifdef HAVE_PJSIP_INV_SESSION_REF
		if (pjsip_inv_add_ref(dtmf_data->session->inv_session) != PJ_SUCCESS) {
			ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
			ao2_cleanup(dtmf_data);
			return -1;
		}
#endif

		if (ast_sip_push_task(channel->session->serializer, transmit_info_dtmf, dtmf_data)) {
			ast_log(LOG_WARNING, "Error sending DTMF via INFO.\n");
#ifdef HAVE_PJSIP_INV_SESSION_REF
			pjsip_inv_dec_ref(dtmf_data->session->inv_session);
#endif
			ao2_cleanup(dtmf_data);
			return -1;
		}
		break;
	}
	case AST_SIP_DTMF_RFC_4733:
		if (!media || !media->rtp) {
			return -1;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (ast_rtp_instance_dtmf_mode_get(media->rtp) != AST_RTP_DTMF_MODE_NONE) {
			if (ast_opt_avs) {
				struct ast_dtmf_send_param dtmfreq = {0};
				dtmfreq.digit = digit;
				dtmfreq.opt = 1;
				res = ast_avs_control(ast, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
				if (!res) {
					usleep(100000);
					dtmfreq.opt = 0;
					ast_avs_control(ast, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
				}
			} else {
				ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
			}
		}

		break;
#else
		ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
		break;
#endif
	case AST_SIP_DTMF_AUTO:
		if (!media || !media->rtp || (ast_rtp_instance_dtmf_mode_get(media->rtp) == AST_RTP_DTMF_MODE_INBAND)) {
			 return -1;
		}

#ifdef GRANDSTREAM_NETWORKS
			if (ast_opt_avs) {
				struct ast_dtmf_send_param dtmfreq = {0};
				dtmfreq.digit = digit;
				dtmfreq.opt = 1;
				res = ast_avs_control(ast, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
				if (!res){
					usleep(100000);
					dtmfreq.opt = 0;
					ast_avs_control(ast, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
				}
			} else {
#endif
				ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
#ifdef GRANDSTREAM_NETWORKS
			}
#endif
		break;


	case AST_SIP_DTMF_NONE:
		break;
	case AST_SIP_DTMF_INBAND:
		res = -1;
		break;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for bug 158037 */
	case AST_SIP_DTMF_BOTH_RFC4733_AND_INFO:
		send_digit_with_both_rfc4733_and_info(channel->session, digit, duration, media);
		break;
#endif
	}

	return res;
}

#ifdef GRANDSTREAM_NETWORKS

static char* ast_terminal_type_to_string(enum ast_terminal_type terminal_type)
{
	switch (terminal_type) {
	case AST_TERMINAL_TYPE_SIP:
		return "GENERAL_SIP";
	case AST_TERMINAL_TYPE_GVC:
		return "GVC";
	case AST_TERMINAL_TYPE_WEBRTC:
		return "WEBRTC";
	case AST_TERMINAL_TYPE_GSWAVE_MOBILE:
		return "APP";
	case AST_TERMINAL_TYPE_ICE_TRUNK:
		return "ICE_TRUNK";
	default:
		return "unknow";
	}

	return "unknow";
}


#endif

static void update_initial_connected_line(struct ast_sip_session *session)
{
	struct ast_party_connected_line connected;

	/*
	 * Use the channel CALLERID() as the initial connected line data.
	 * The core or a predial handler may have supplied missing values
	 * from the session->endpoint->id.self about who we are calling.
	 */
	ast_channel_lock(session->channel);
	ast_party_id_copy(&session->id, &ast_channel_caller(session->channel)->id);
	ast_channel_unlock(session->channel);

	/* Supply initial connected line information if available. */
	if (!session->id.number.valid && !session->id.name.valid) {
		return;
	}

	ast_party_connected_line_init(&connected);
	connected.id = session->id;
	connected.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;

	ast_channel_queue_connected_line_update(session->channel, &connected, NULL);
}

static int call(void *data)
{
	struct ast_sip_channel_pvt *channel = data;
	struct ast_sip_session *session = channel->session;
	pjsip_tx_data *tdata;
#ifdef GRANDSTREAM_NETWORKS
	const char *forward_thx = NULL;
	const char *inviter_ext = NULL;
	struct chan_pjsip_pvt *pvt = channel->pvt;
	const char* local_channel_name = NULL;
	struct ast_channel * local_channel = NULL;
	if (session && session->channel) {
		const char* endpoint = ast_sorcery_object_get_id(session->endpoint);
		const char* conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
		if (((conf_room && !ast_strlen_zero(conf_room)) || session->reject_call) &&
			endpoint && strstr(endpoint, "trunk") == NULL &&
			(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC )) {
			ast_log(LOG_DEBUG, "Inviting members(%s/%s)[%s] is not supported in meeting.",
					endpoint, ast_terminal_type_to_string(session->terminal_type),
					ast_channel_name(session->channel));


			forward_thx = pbx_builtin_getvar_helper(session->channel, "FORWARD_THX");
			inviter_ext = pbx_builtin_getvar_helper(session->channel, "PHONECONFINVITE");
			if (!ast_strlen_zero(forward_thx)) {
				if (ast_strlen_zero(inviter_ext)) {
					inviter_ext = conf_room;
				}
				ast_log(LOG_DEBUG, "inviter_ext %s!", inviter_ext);
				ast_log(LOG_DEBUG, "forward_thx %s!", forward_thx);
				char *cmd_array = NULL;
				ast_asprintf(&cmd_array, "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/pushInviteNotifyToIM.py --invite_users '%s' --inviter '%s' --conf_room '%s'", endpoint, inviter_ext, conf_room);
				if (cmd_array) {
					ast_safe_system2(cmd_array);
					ast_log(LOG_DEBUG, "call forward! try %s!",cmd_array);
				}
				ast_free(cmd_array);
			}
			local_channel_name = pbx_builtin_getvar_helper(session->channel, "ORI_CALLER_CHANNEL_NAME");
			if (!ast_strlen_zero(local_channel_name)) {
				local_channel = ast_channel_get_by_name(local_channel_name);
				if (local_channel) {
					pbx_builtin_setvar_helper(local_channel, "WAVE_CONF_REJECT", "yes");
					ast_channel_unref(local_channel);
					if (ast_channel_get_mychan(local_channel)) {
						pbx_builtin_setvar_helper(ast_channel_get_mychan(local_channel), "WAVE_CONF_REJECT", "yes");
					}
				}
			}
			const char* ccm_channel_name = pbx_builtin_getvar_helper(session->channel, "CCM_CHANNEL");
			if (!ast_strlen_zero(ccm_channel_name)) {
				struct ast_channel *ccm_channel = ast_channel_get_by_name(ccm_channel_name);
				if (ccm_channel) {
					ast_log(LOG_DEBUG, "ccm invite set WAVE_CONF_REJECT!");
					pbx_builtin_setvar_helper(ccm_channel, "WAVE_CONF_REJECT", "yes");
					ast_channel_unref(ccm_channel);
				}
			}

			//ast_channel_hangupcause_set(session->channel, AST_CAUSE_USER_BUSY);
			ast_channel_hangupcause_set(session->channel, AST_CAUSE_CONF_SKIP_CALL_WAVE);
			ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
			ast_queue_hangup(session->channel);
			ao2_ref(channel, -1);
			return 0;
		}
	}

	/*Start added by zyxia for Bug 234969,2022.8.30*/
	const char *service = pbx_builtin_getvar_helper(session->channel, "SPECIAL");
	if (!ast_strlen_zero(service) && !strcasecmp(service, "schedule_paging")) {
		const char*page_video_support = pbx_builtin_getvar_helper(session->channel,"PAGE_VIDEO_SUPPORT");
		if (page_video_support && !strcasecmp(page_video_support,"yes")) {
			ast_log(LOG_DEBUG, "Set channel:%s page video\n", ast_channel_name(session->channel));
			ast_channel_set_is_page_video_support(session->channel,1);
			session->is_video_page_call = 1;
			change_session_media_direction(session, AST_MEDIA_SUB_VIDEO_TYPE_MAIN,RECVONLY);
		}
	}
	/*End added by zyxia for Bug 234969,2022.8.30*/
	/*Start added by zyxia for Bug 235752,play video prompt file on web,2022.9.13*/
	else {
		const char *prompt_video_call = pbx_builtin_getvar_helper(session->channel, "PROMPT_VIDEO_CALL");
		if (!ast_strlen_zero(prompt_video_call) && !strcasecmp(prompt_video_call, "yes")) {
			ast_log(LOG_DEBUG, "Set channel:%s  video call,direction is sendonly\n", ast_channel_name(session->channel));
			change_session_media_direction(session, AST_MEDIA_SUB_VIDEO_TYPE_MAIN,SENDONLY);
		}
	}
	/*End added by zyxia for Bug 235752,play video prompt file on web,2022.9.13*/
#endif


	int res = ast_sip_session_create_invite(session, &tdata);

	if (res) {
		ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
		ast_queue_hangup(session->channel);
	} else {
		set_channel_on_rtp_instance(session, ast_channel_uniqueid(session->channel));
		update_initial_connected_line(session);
#ifdef GRANDSTREAM_NETWORKS
		/* Modified by yfyang, for 73741 */
		const char *seamless_transfer = pbx_builtin_getvar_helper(session->channel, "SEAMLESS_TRANSFER");
		const char *frompage = pbx_builtin_getvar_helper(session->channel, "PAGING_ALERT_INFO");
		int  add_feature_map = 1;

		/* Added by yfyang, for Bug 109923*/
		if (ast_strlen_zero(frompage)) {
			const char *service = pbx_builtin_getvar_helper(session->channel, "SPECIAL");
			if (!ast_strlen_zero(service) && !strcasecmp(service, "schedule_paging")) {
				add_feature_map = 0;
				ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "Paging");
			}
		} else {
			add_feature_map = 0;
			ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "Paging");
		}

		/* for bug 121470 */
		const char *vq_callback = pbx_builtin_getvar_helper(session->channel, "VQ_CALLBACK_CALLEE");
		if (!ast_strlen_zero(vq_callback) && !strcmp(vq_callback,"yes"))
		{
			add_feature_map = 0;
		}

		if (session->endpoint && strncmp(session->endpoint->context, "sla-stations", 12)
			&& !session->is_direct_media
			&& (!seamless_transfer || strcasecmp(seamless_transfer, "yes"))
			&& (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12))
			&& (session->line_index <= 0)
			&& (add_feature_map == 1)) {
			add_feature_map_header(tdata, 1, session);
		}

		/*add by chdruan, Bug 42500 - Diversion:the ucm carry 4 diversion header when going through register trunk*/
		if (session->channel != NULL && pbx_builtin_getvar_helper(session->channel, "CF_REASON") != NULL
				&& !ast_channel_redirecting(session->channel)->count) {
			ast_channel_redirecting(session->channel)->count++;
		}

		if((ast_channel_get_terminate_type(session->channel) == AST_TERMINAL_TYPE_GSWAVE_MOBILE
			|| ast_channel_get_terminate_type(session->channel) == AST_TERMINAL_TYPE_WEBRTC)
			&& session->endpoint && !ast_sip_endpoint_is_have_h264_codec(session->endpoint)) {
				ast_sip_add_header(tdata, "X-GS-Video-Codec", "no");
		}

		add_client_info_for_wave(session, tdata, 1);
		//==========================================
		// Support CEI for user invitation
		//==========================================
		const char *cei_support = pbx_builtin_getvar_helper(session->channel, "CEI_SUPPORT");
		const char *conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
		const char *mcb_contact = pbx_builtin_getvar_helper(session->channel, "MCB_CONTACT");
		const char *mcb_conf_header = pbx_builtin_getvar_helper(session->channel, "MCB_CONF_HEADER");
		const char *cid_name = pbx_builtin_getvar_helper(session->channel, "CID_NAME");
		const char *transport_type = pbx_builtin_getvar_helper(session->channel, "TRANSPORT_TYPE");
		struct ast_sip_endpoint *endpoint = session->endpoint;
		const char *endpoint_id = ast_sorcery_object_get_id(endpoint);

#ifdef GRANDSTREAM_NETWORKS
		if (cei_support != NULL && strcmp(cei_support, "yes") == 0 &&
			!ast_strlen_zero(conf_room)) {
#else
		if (cei_support != NULL && strcmp(cei_support, "yes") == 0 &&
			!ast_strlen_zero(conf_room) && endpoint != NULL &&
			endpoint_id != NULL && strstr(endpoint_id, "trunk") == NULL) {//remove trunk limit by zyxia for trunk conf
#endif
			// Set isfocus
			pvt->isConference = TRUE;

			if (globalEnabledCEI || session->is_sfu) {
				add_peer_conference(ast_channel_name(session->channel), endpoint, conf_room,AST_CONF_TYPE_NORMAL);
			}

			// Set contact user
			if (pvt->contact_user != NULL) {
				ast_free(pvt->contact_user);
			}
			pvt->contact_user = ast_strdup(conf_room);

			// Update tdata
			if (globalEnabledCEI || session->is_sfu) {  //not add isfocus if CEI notify disabled, modified by yfyang, for bug58890
				append_isfocus_contact(tdata);
				ast_sip_add_header(tdata, "Channel-ID", ast_channel_name(session->channel));
			} else {
				if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
					ast_sip_add_header(tdata, "Channel-ID", ast_channel_name(session->channel));
				}
			}
			set_contact_user(tdata, pvt->contact_user);
			pjsip_sip_uri *contact_uri = pjsip_uri_get_uri(session->inv_session->dlg->local.contact->uri);
			if (contact_uri != NULL) {
				pj_strdup2(session->inv_session->dlg->pool, &contact_uri->user, pvt->contact_user);
			}
		} else {
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
				ast_sip_add_header(tdata, "Channel-ID", ast_channel_name(session->channel));
			}
		}

/*Start added by zyxia for door station support.2021.11.19*/
#ifdef GRANDSTREAM_NETWORKS
		if(session->is_hikvision_call)
			ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "DoorStation/Hikvision");
		else if(session->is_gds_call)
			ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "DoorStation/Grandstream");
		else if (session->is_rtsp_call)
			ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "IPC");//added by zyxia for Bug 208397
#endif
/*En added by zyxia for door station support.2021.11.19*/

		if (!ast_strlen_zero(conf_room)) {
			ast_log(LOG_DEBUG, "[session:%p] It's conference invite, add Call-Type for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->conf_invite_call = 1;//added by zyxia for Bug 271949,2023.9.15
			if (session->is_sfu) {
				ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "VideoConference");
			} else {
				ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "Conference");
			}

			if (endpoint_id != NULL && strstr(endpoint_id, "trunk")) {
				char SIP_X_UCM_IM_VALUE[256] = {0};
				char cloud_im_enabled[8] = {0};
				char cloud_im_addr[129] = {0};
				char cloud_im_service_id[64] = {0};
				ast_safe_nvram_get("cloud_im_enabled", cloud_im_enabled, sizeof(cloud_im_enabled));
				if (strncmp(cloud_im_enabled, "on", 2) == 0) {
					ast_safe_nvram_get("cloud_im_address", cloud_im_addr, sizeof(cloud_im_addr));
					snprintf(SIP_X_UCM_IM_VALUE, sizeof(SIP_X_UCM_IM_VALUE) - 1, "cloud;%s", cloud_im_addr);

					ast_safe_nvram_get("cloud_im_service_id", cloud_im_service_id, sizeof(cloud_im_service_id));
					if (strlen(cloud_im_service_id)) {
						ast_sip_add_header(tdata, "X-GS-IM-SERVICE-ID", cloud_im_service_id);
					}
				} else {
					snprintf(SIP_X_UCM_IM_VALUE, sizeof(SIP_X_UCM_IM_VALUE) - 1, "local");
				}

				ast_sip_add_header(tdata, "X-UCM-IM", SIP_X_UCM_IM_VALUE);
			}
		}

		//==========================================
		// Update SIP headers
		//==========================================
		// Assign From header information
		if (cid_name != NULL) {
			/* Modified by yfyang, for Bug 159397 */
			char *cf_user = (char *)pbx_builtin_getvar_helper(session->channel, "CF_USER");;
			if (ast_strlen_zero(cf_user)) {
				set_from_name(tdata, cid_name ? cid_name : "");
			}
		}

		// Update contact information
		if (mcb_contact != NULL && strlen(mcb_contact) != 0) {
			set_contact_user(tdata, (char *)mcb_contact);
		}

		// Add SIP custom header to indicate transport type
		/*X-Grandstream-Transport is to distinguish internal mcb or external mcb.*/
		if (transport_type != NULL && !ast_strlen_zero(mcb_conf_header)) {
			ast_sip_add_header(tdata, "X-Grandstream-Transport",
								S_OR(transport_type, ""));
		}

		if (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
			ast_sip_add_timestamp_header(tdata);
		}

		/* Add by bfdu for whiteboard, 2022.1.7 */
		ast_sip_session_add_support_header_for_whiteboard(session, tdata);
		/* Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
		ast_sip_session_add_support_header_for_gs_state(session,tdata);

		if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
			ast_sip_add_header(tdata, "X-GS-Trunk", "yes");
		}

		add_sfu_header_for_trunk(session, tdata);
		add_conf_link_for_wave(session, tdata);

		add_isfocus_for_trunk_conf(session,tdata);
		ast_channel_terminal_type_set(session->channel, session->terminal_type);
#endif
		ast_sip_session_send_request(session, tdata);
	}
	ao2_ref(channel, -1);
	return res;
}

/*! \brief Function called by core to actually start calling a remote party */
static int chan_pjsip_call(struct ast_channel *ast, const char *dest, int timeout)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);

	ao2_ref(channel, +1);
	if (ast_sip_push_task(channel->session->serializer, call, channel)) {
		ast_log(LOG_WARNING, "Error attempting to place outbound call to '%s'\n", dest);
		ao2_cleanup(channel);
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (strncmp(ast_channel_name(ast), "PJSIP/trunk_", 12)) {
		char callee_dial_string[256] = {0};
		snprintf(callee_dial_string, sizeof(callee_dial_string), "PJSIP/%s", dest);
		pbx_builtin_setvar_helper(ast, "CALLEE_DIAL_STRING", callee_dial_string);
	}
#endif


	return 0;
}

/*! \brief Internal function which translates from Asterisk cause codes to SIP response codes */
static int hangup_cause2sip(int cause)
{
	switch (cause) {
	case AST_CAUSE_UNALLOCATED:             /* 1 */
	case AST_CAUSE_NO_ROUTE_DESTINATION:    /* 3 IAX2: Can't find extension in context */
	case AST_CAUSE_NO_ROUTE_TRANSIT_NET:    /* 2 */
		return 404;
	case AST_CAUSE_CONGESTION:              /* 34 */
	case AST_CAUSE_SWITCH_CONGESTION:       /* 42 */
		return 503;
	case AST_CAUSE_NO_USER_RESPONSE:        /* 18 */
		return 408;
	case AST_CAUSE_NO_ANSWER:               /* 19 */
	case AST_CAUSE_UNREGISTERED:        /* 20 */
		return 480;
	case AST_CAUSE_CALL_REJECTED:           /* 21 */
		return 403;
	case AST_CAUSE_NUMBER_CHANGED:          /* 22 */
		return 410;
	case AST_CAUSE_NORMAL_UNSPECIFIED:      /* 31 */
		return 480;
	case AST_CAUSE_INVALID_NUMBER_FORMAT:
		return 484;
	case AST_CAUSE_USER_BUSY:
#ifdef GRANDSTREAM_NETWORKS
	case AST_CAUSE_CONF_SKIP_CALL_WAVE:
#endif
		return 486;
	case AST_CAUSE_FAILURE:
		return 500;
	case AST_CAUSE_FACILITY_REJECTED:       /* 29 */
		return 501;
	case AST_CAUSE_CHAN_NOT_IMPLEMENTED:
		return 503;
	case AST_CAUSE_DESTINATION_OUT_OF_ORDER:
		return 502;
	case AST_CAUSE_BEARERCAPABILITY_NOTAVAIL:       /* Can't find codec to connect to host */
		return 488;
	case AST_CAUSE_INTERWORKING:    /* Unspecified Interworking issues */
		return 500;
	case AST_CAUSE_NOTDEFINED:
	default:
		ast_debug(1, "AST hangup cause %d (no match found in PJSIP)\n", cause);
		return 0;
	}

	/* Never reached */
	return 0;
}

struct hangup_data {
	int cause;
	struct ast_channel *chan;
};

static void hangup_data_destroy(void *obj)
{
	struct hangup_data *h_data = obj;

	h_data->chan = ast_channel_unref(h_data->chan);
}

static struct hangup_data *hangup_data_alloc(int cause, struct ast_channel *chan)
{
	struct hangup_data *h_data = ao2_alloc(sizeof(*h_data), hangup_data_destroy);

	if (!h_data) {
		return NULL;
	}

	h_data->cause = cause;
	h_data->chan = ast_channel_ref(chan);

	return h_data;
}

/*! \brief Clear a channel from a session along with its PVT */
static void clear_session_and_channel(struct ast_sip_session *session, struct ast_channel *ast)
{
	session->channel = NULL;
	set_channel_on_rtp_instance(session, "");
	ast_channel_tech_pvt_set(ast, NULL);
}

static int hangup(void *data)
{
	struct hangup_data *h_data = data;
	struct ast_channel *ast = h_data->chan;
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct ast_sip_session *session = channel->session;
#ifdef GRANDSTREAM_NETWORKS
	struct ast_sip_endpoint *endpoint = (session ? session->endpoint : NULL);
#endif
	int cause = h_data->cause;

#ifdef GRANDSTREAM_NETWORKS
	if (session && session->endpoint) {
		ast_log(LOG_NOTICE, "Hangup clear endpoint global var: %s\n", session->endpoint->aors);
		session->endpoint->ignore_endpt_direct_media = 0;
	}

	if (session->line_index && endpoint) {
		char *user = NULL;

		if (session->aor) {
			user = (char *)ast_sorcery_object_get_id(session->aor);
		} else {
			user = (char *)endpoint->aors;
		}
		ast_log(LOG_NOTICE, "hangup, sharedline[%s], index[%d], user[%s]", endpoint->sca_sharedline, session->line_index, user);
		ast_sip_sca_line_remove_user(session->channel, (char *)endpoint->sca_sharedline, session->line_index, user);

		session->line_index = 0;
	}

	if (session->channel) {
		session->conf_hangupcause = ast_channel_conf_hangupcause(session->channel);
	}

	// Clear current channel's subscription
	if (endpoint) {
		ast_sip_clear_peer_conference(endpoint, ast_channel_name(ast));
	}

	session->ignore_avs_response = (ast_channel_ignore_avs_response(ast) ? 1 : 0);
	/*Start added by zyxia for 256617 ,2023.4.10*/
	if(ast_channel_get_is_wave_app(ast) && session && session->inv_session
		&& PJSIP_INV_STATE_CALLING == session->inv_session->state) {
		ast_sip_session_defer_termination_cancel(session);
		ast_log(LOG_NOTICE, "session:%p session_defer_termination_cancel  channel:%s", session,ast_channel_name(ast));
	}
	/*End added by zyxia for 256617 ,2023.4.10*/
#endif

	/*
	 * It's possible that session_terminate might cause the session to be destroyed
	 * immediately so we need to keep a reference to it so we can NULL session->channel
	 * afterwards.
	 */
	ast_sip_session_terminate(ao2_bump(session), cause);
	clear_session_and_channel(session, ast);
	ao2_cleanup(session);
	ao2_cleanup(channel);
	ao2_cleanup(h_data);
	return 0;
}

/*start added by zyxia for Bug 316141 */
#ifdef GRANDSTREAM_NETWORKS
static  char* ast_get_record_uuid_from_cdr(struct ast_channel *ast,const char* file)
{
	char cdr_record_uuid[1024] = {0};
	char cdr_record_files[1024] = {0};
	char *sep = "@";
	char *record_files_token = NULL;
	char *record_files_uuid_token = NULL;
	char *record_files_sub = NULL;
	char *record_files_uuid_sub = NULL;

	if(!ast || ast_strlen_zero(file)){
		return NULL;
	}

	ast_cdr_getvar(ast_channel_name(ast), "record_files_uuid", cdr_record_uuid, sizeof(cdr_record_uuid));
	ast_cdr_getvar(ast_channel_name(ast), "record_files", cdr_record_files, sizeof(cdr_record_files));
	if(!ast_strlen_zero(cdr_record_files) && !ast_strlen_zero(cdr_record_uuid)){

		record_files_token = strtok_r(cdr_record_files, sep, &record_files_sub);
		record_files_uuid_token = strtok_r(cdr_record_uuid, sep, &record_files_uuid_sub);
		while (record_files_token && record_files_uuid_token) {
			if(strstr(file,record_files_token)){
				ast_log(LOG_DEBUG, "[sendMail] decrypted key-from-cdr:%s",record_files_uuid_token);
				return ast_strdup(record_files_uuid_token);
			}
			record_files_token = strtok_r(NULL, sep, &record_files_sub);
			record_files_uuid_token = strtok_r(NULL, sep, &record_files_uuid_sub);
		}
	}

	return NULL;
}


#endif
/*End added by zyxia for Bug 316141 */

/*! \brief Function called by core to hang up a PJSIP session */
static int chan_pjsip_hangup(struct ast_channel *ast)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	int cause;
	struct hangup_data *h_data;
	char  family[80] = {0};
	char  data[80] = {0};
	int high_freq_count = 0;
	int q_index;
	struct timeval answer_val;

	if (!channel || !channel->session) {
		return -1;
	}

	/* when one channel answered , other channel hangup, can't delete recovery db with linkedid */
	if(AST_CAUSE_ANSWERED_ELSEWHERE == ast_channel_hangupcause(ast)
		&& !ast_channel_pbx(ast))
	{
		ast_channel_avoid_deltrack_flag_set(ast,1);
	}

	cause = hangup_cause2sip(ast_channel_hangupcause(channel->session->channel));
	h_data = hangup_data_alloc(cause, ast);

/*Add by yuyang, for Bug 127827*/
#ifdef GRANDSTREAM_NETWORKS
	const char *ishunt = pbx_builtin_getvar_helper(ast, "TRIGGERHUNT");
	const char *emergency_service_num = pbx_builtin_getvar_helper(ast, "EMERGENCY_SERVICE_NUM");
	const char *emergency_dir = pbx_builtin_getvar_helper(ast, "EMERGENCY_DIR");
	const char *emergency_filename = pbx_builtin_getvar_helper(ast, "EMERGENCYFILENAME");

	ast_log(LOG_DEBUG, "Channel:%s, hangupCause:%d(%d)",
		ast_channel_name(channel->session->channel), cause,
		ast_channel_hangupcause(channel->session->channel));

	if( !strstr(ast_channel_name(ast),"PJSIP/trunk"))
	{
		const char *isivr = pbx_builtin_getvar_helper(ast, "IS_IVR");
		const char *trunkidx = pbx_builtin_getvar_helper(ast, "OUTBOUND_TRUNK_INDEX");
		const char *dialstate = pbx_builtin_getvar_helper(ast, "DIALSTATUS");
		/* outbound via ivr */
		if(isivr && strstr(isivr,"YES") && trunkidx && (0 != strlen(trunkidx))
			&& channel->session->endpoint
			&& dialstate && !strcasecmp(dialstate,"ANSWER"))
		{
			snprintf(family, sizeof(family), "AAUSER/%s", channel->session->endpoint->sca_sharedline);
			ast_ucm_db_get(family, "HIGHCALL_THRESHOLD", data, sizeof(data));
		}
	}

	/*start added by zyxia for trunk conf,if trunk active user is 0,need unscbcribe*/
	if(channel->session->had_subscribe_trunk_cei  && strstr(ast_channel_name(channel->session->channel),"PJSIP/trunk")){
		struct cei_external_trunk_conf_peer* etc_peer = NULL;
		etc_peer = ast_cei_find_external_trunk_conf_peer_by_virtual_conf(channel->session->endpoint,ast_channel_get_virtual_conf_num(ast));

		ast_log(LOG_DEBUG, "Channel:%s, hangupCause:%d(%d),trunk_conf:%s",
			ast_channel_name(ast), cause,
			ast_channel_hangupcause(ast),ast_channel_get_virtual_conf_num(ast));

		if (etc_peer && etc_peer->local_active_user) {

			ast_log(LOG_DEBUG, "Channel:%s, hangupCause:%d(%d),etc_peer->local_active_user:%d",
			ast_channel_name(ast), cause,
			ast_channel_hangupcause(ast),etc_peer->local_active_user);

			etc_peer->local_active_user = etc_peer->local_active_user - 1;
			if (etc_peer->local_active_user == 0) {
				sip_exteral_trunk_conf_unsub(channel->session->endpoint ,etc_peer);
			}
		}
	}

	if(!strstr(ast_channel_name(ast),"PJSIP/trunk")
		&& ast_channel_get_is_join_trunk_conf(ast) && ast_channel_get_virtual_conf_num(ast)) {
		ast_log(LOG_DEBUG, "Channel:%s, ast_channel_get_is_join_trunk_conf:%d",
		ast_channel_name(ast), ast_channel_get_is_join_trunk_conf(ast));

		publish_cei_unsub(ast_channel_get_virtual_conf_num(ast), ast_channel_name(ast), S_COR(ast_channel_caller(ast)->id.number.valid, ast_channel_caller(ast)->id.number.str, ""));
	}
	/*End added by zyxia for trunk conf,if trunk active user is 0,need unscbcribe*/

	if(data[0] != 0 && atoi(data) != 0)
	{
		high_freq_count = atoi(data);
		q_index = ast_exten_map2queue_index((char *)channel->session->endpoint->sca_sharedline);

		if(-1 != q_index )
		{
			answer_val = ast_channel_answertime(ast);
			ast_log(LOG_DEBUG,"chan:%s insert a observe node, time=%ld, thres=%d", ast_channel_name(ast),answer_val.tv_sec,high_freq_count);
			ast_high_frequence_enqueue(ast_get_high_frequence_queue(q_index), (char *)channel->session->endpoint->sca_sharedline,
						high_freq_count, &answer_val);
		}
	}

	if(ishunt)
	{
		char *sendMail_cmd = NULL;
        const char *notifyonce = pbx_builtin_getvar_helper(ast, "NOTIFYONCE");

        if (NULL != notifyonce && 0 == strcmp(notifyonce, "no")) {
            ast_sip_report_emergency_outbound(ast_channel_caller_num(ast),emergency_service_num);
            pbx_builtin_setvar_helper(ast, "_NOTIFYONCE", "yes");
        }
		
		int flag = 0;
		char emergency_filename_decrypt[1024] = "";
		if (get_emergency_email(emergency_service_num)) {
			int is_file_encrypted = 0;
			int encrypted_file_header_offset = 0;
			char file_key_from_file[256] = "";
			char *file_key_from_file_decrypt = NULL;
			char emergency_path_encrypt[1024] = "";
			char emergency_path_decrypt[1024] = "";
			FILE *fp_encrypt = NULL;
			FILE *fp_decrypt = NULL;
			char *in_buffer = NULL;
    		char *out_buffer = NULL;

			snprintf(emergency_path_encrypt, sizeof(emergency_path_encrypt), "%s/%s%s", emergency_dir, emergency_filename, ".wav");
			ast_log(LOG_DEBUG, "[sendMail] encrypt full-file-path:%s", emergency_path_encrypt);

			is_file_encrypted = file_crypt_is_file_encrypted(emergency_path_encrypt);
			if (UCM_CRYPT_E_OK == is_file_encrypted) {
				file_crypt_file_key(emergency_path_encrypt, file_key_from_file, sizeof(file_key_from_file));
				ast_log(LOG_DEBUG, "[sendMail] key-from-file:%s", file_key_from_file);

				ucm_secret_decrypt(file_key_from_file, &file_key_from_file_decrypt);
				ast_log(LOG_DEBUG, "[sendMail] decrypted key-from-file:%s", file_key_from_file_decrypt);

				/*added by zyxia for 316141,get key from cdr*/
				if(ast_strlen_zero(file_key_from_file_decrypt)){
					AST_SAFE_FREE(file_key_from_file_decrypt);
					file_key_from_file_decrypt = ast_get_record_uuid_from_cdr(ast,emergency_path_encrypt);
				}

				if (file_key_from_file_decrypt && strlen(file_key_from_file_decrypt)) {
					char *ptr = strstr(emergency_filename, "/");
					if (ptr && ptr++) {
						snprintf(emergency_filename_decrypt, sizeof(emergency_filename_decrypt), "%s", ptr);
					} else {
						snprintf(emergency_filename_decrypt, sizeof(emergency_filename_decrypt), "%s", emergency_filename);
					}
					ast_log(LOG_DEBUG, "[sendMail] decrypt full-file-name:%s", emergency_filename_decrypt);

					snprintf(emergency_path_decrypt, sizeof(emergency_path_decrypt), "%s/%s%s", "/tmp", emergency_filename_decrypt, ".wav");
					ast_log(LOG_DEBUG, "[sendMail] decrypt full-file-path:%s", emergency_path_decrypt);

					fp_encrypt = fopen(emergency_path_encrypt, "r");
                	fp_decrypt = fopen(emergency_path_decrypt, "w+");

					if (fp_encrypt && fp_decrypt) {
						encrypted_file_header_offset = file_crypt_file_header_length(emergency_path_encrypt);
                    	fseek(fp_encrypt, encrypted_file_header_offset, SEEK_SET);

						in_buffer = (char *)malloc(sizeof(char) * (1024 + 64));
                    	out_buffer = (char *)malloc(sizeof(char) * (1024 + 64) * 2);

						if (in_buffer && out_buffer) {
							int length_read_from_file = 0;
							while ((length_read_from_file = fread(in_buffer, sizeof(char), 1024, fp_encrypt)) > 0) {
								int length = file_crypt_decrypt(in_buffer, length_read_from_file, file_key_from_file_decrypt, strlen(file_key_from_file_decrypt), "v1", out_buffer, (1024 + 64) * 2);
								fwrite(out_buffer, sizeof(char), length, fp_decrypt);
							}
							flag = 1;
						}
					}
				}
			}
			if (fp_encrypt) {
        		fclose(fp_encrypt);
    		}
    		if (fp_decrypt) {
        		fclose(fp_decrypt);
    		}
    		AST_SAFE_FREE(file_key_from_file_decrypt);
    		AST_SAFE_FREE(in_buffer);
    		AST_SAFE_FREE(out_buffer);
		}

		if(get_emergency_email(emergency_service_num) && 
		   asprintf(&sendMail_cmd, "LD_PRELOAD=/usr/lib/libjemalloc.so python /cfg/var/lib/asterisk/scripts/sendMail.py emergency '' '%s' '%s' '%s' '%s'", ast_channel_caller_num(ast), emergency_service_num, flag ? "/tmp" : emergency_dir, flag ? emergency_filename_decrypt : emergency_filename) >= 0)
		{
			ast_log(LOG_DEBUG, "sendMail_cmd: %s", sendMail_cmd);
#ifdef HAVE_GCC601X
            ast_safe_nvram_set("emergency_nas_falg","1");
#endif

			if (ast_taskprocessor_push(emergency_mail_tps, send_emergency_mail, sendMail_cmd) < 0) {
				ast_log(LOG_WARNING, "push emergency mail task failed!!!");
				ast_free(sendMail_cmd);
			}
		}
	}
#ifdef HAVE_GCC601X
    else
    {
        ast_safe_nvram_set("emergency_nas_falg","0");
    }
#endif

	/* Add by bfdu for turn license, 2020.5.19 */

	if (ast_channel_get_turn_session_count(ast)) {
		ast_sip_decrease_used_turn_session_count(ast_channel_get_turn_session_count(ast));
		ast_channel_set_turn_session_count(ast, 0);
	}
#endif
/*End by yuyang, for Bug 127827*/

	if (!h_data) {
		goto failure;
	}

	if (ast_sip_push_task(channel->session->serializer, hangup, h_data)) {
		ast_log(LOG_WARNING, "Unable to push hangup task to the threadpool. Expect bad things\n");
		goto failure;
	}

	return 0;

failure:
	/* Go ahead and do our cleanup of the session and channel even if we're not going
	 * to be able to send our SIP request/response
	 */
	clear_session_and_channel(channel->session, ast);
	ao2_cleanup(channel);
	ao2_cleanup(h_data);
	return -1;
}

struct request_data {
	struct ast_sip_session *session;
	struct ast_stream_topology *topology;
#ifdef GRANDSTREAM_NETWORKS
	struct ast_channel *requestor;
	struct ast_channel *old_channel;
	int app_wakeup;
#endif
	const char *dest;
	int cause;
};

#ifdef GRANDSTREAM_NETWORKS
struct app_prompt_data{
	struct ast_channel *requestor;
	const char* file_name;
};
#endif


/*Start added by zyxia for wakeup ios gswave,2018.7.27*/
#ifdef GRANDSTREAM_NETWORKS
static unsigned char judge_chan_type_is_sip(const struct ast_channel *channel)
{
	if (NULL == channel){
		ast_log(LOG_WARNING, "chan is NULL\n");
		return FALSE;
	}

	if (ast_channel_tech(channel) &&
		(0 == strcmp(ast_channel_tech(channel)->type, "PJSIP"))) {
		return TRUE;
	} else {
		return FALSE;
	}
}

/*! \brief remove sip: sips @ and port from uri
*/
char* ast_sip_get_ip_addr_from_contact_uri(const char *contact_uri)
{
	const char *origin_uri = NULL;
	char *uri = NULL;
	char *pUri = NULL;

	if( NULL == contact_uri){
		return NULL;
	}

	// Valid URI starts with sip: or sips:
	if (strncmp(contact_uri, "sip:", 4) == 0) {
		origin_uri = contact_uri + 4;
	}
	else if (strncmp(contact_uri, "sips:", 5) == 0) {
		origin_uri = contact_uri + 5;
	}
	else {
		return NULL;
	}

	// IP address is next to username@
	pUri = strchr(origin_uri, '@');
	if (pUri == NULL) {
		uri = ast_strdup(origin_uri);
	}
	else {
		uri = ast_strdup(pUri + 1);
	}

	// Remove port, modified for ipv6, bug Bug 67059
	if ((uri[0] == '[') && (pUri = strchr(uri, ']'))) {
		*pUri = '\0';
		pUri = ast_strdup(uri + 1);
		ast_free(uri);
		uri = pUri;
	} else {
		pUri = strchr(uri, ':');
		if (pUri != NULL) {
		*pUri = '\0';
		}
	}

	// Remove transport
	pUri = strchr(uri, ';');
	if (pUri != NULL) {
		*pUri = '\0';
	}

	return uri;
}

/*Start added by zyxia for Bug 281532, check the uri if exist in wave app contact*/
/*0: not exist,offline 1:exist,online*/
static int check_uri_in_contact(const char* dst_addr, struct ast_sip_endpoint *endpoint)
{
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
  	struct ast_sip_contact *contact = NULL;
	struct ao2_iterator i;
	int  contact_matched = 0;


	if (!dst_addr || !endpoint) {
		return 0;
	}

	aor = ast_sip_location_retrieve_aor(endpoint->aors); 
	if (!aor) {
		return 0;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts ) {
		return 0;
	}

	if (contacts) {
		i = ao2_iterator_init(contacts, 0);
		while ((contact = ao2_iterator_next(&i))) {
			if ( contact->gswave_mobile && !strcmp(contact->uri, dst_addr)) {
				contact_matched = 1;
				ao2_ref(contact, -1);
				break;
			}
			ao2_ref(contact, -1);
		}
		ao2_iterator_destroy(&i);

		if (contact_matched) {
			ast_log(LOG_NOTICE, "check_uri_in_contact endpoint:%s,contact_uri:%s contact_matched!",ast_sorcery_object_get_id(endpoint), contact->uri);
			return 1;
		}
	}
	return 0;
}
/*End added by zyxia for Bug 281532, check the uri if exist in wave app contact*/

static int check_registr_max_contacts(struct ast_sip_endpoint *endpoint)
{
#if 0
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	struct ast_sip_contact *contact = NULL;
	struct ao2_iterator i;
	int  exist_app = 0;

	if (!endpoint) {
		return 0;
	}

	aor = ast_sip_location_retrieve_aor(endpoint->aors);
	if (!aor) {
		return 0;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts ) {
		return 0;
	}

	//if contacs is full,need check the contacst if exist app mobile
	if(ao2_container_count(contacts) >= aor->max_contacts) {
		i = ao2_iterator_init(contacts, 0);
		while ((contact = ao2_iterator_next(&i))) {
			if ( contact->gswave_mobile) {
				exist_app = 1;
				ao2_ref(contact, -1);
				break;
			}
			ao2_ref(contact, -1);
		}
		ao2_iterator_destroy(&i);

		if (exist_app) {
			return 0;
		}
		else{
			ast_log(LOG_WARNING, "check_registr_max_contacts %s had full register full and no app exist\n",ast_sorcery_object_get_id(endpoint));
			return 1;
		}
	}
#endif
	return 0;

}


static unsigned int ast_check_if_wakeup_needed(struct ast_sip_endpoint *endpoint, const char *endpoint_name, const char *args_aor,	struct ast_channel *requestor )
{
	char register_key[64]={0};
	char register_time[256]={0},gs_uri[256]={0};
	char   family[80]={0};
	char  db_data[80]={0};
	struct sip_waveapp_info* app_info = NULL;

	/*Start added by zyxia for Bug 198869 ,2021.7.2*/
	if ( ast_channel_get_reject_call(requestor) || check_registr_max_contacts(endpoint)) {//remove it conf invite call,need call wave app
	//if ( ast_channel_get_reject_call(requestor) || (conf_room && !ast_strlen_zero(conf_room) && endpoint_from && strstr(endpoint_from, "trunk") == NULL)) {
		ast_log(LOG_DEBUG, "Call  dont need wake up endpoint_name:%s\n",endpoint_name);
		return 0;
	}
	/*End added by zyxia for Bug 198869 ,2021.7.2*/

	/*get gs_appid from astdb*/

	app_info = ast_sip_find_waveapp_info_by_uri(endpoint,args_aor);

	if(app_info != NULL){


		snprintf(family, sizeof(family), "AAUSER/%s", endpoint_name);
		ast_ucm_db_get(family, "IMDND", db_data, sizeof(db_data));

		ast_log(LOG_DEBUG, "ast_sip_handle_gswave_mobile_wake aors:%s endpoint_name:%s,gs_uri:%s ,im_dnd:%s,conf_invite:%d\n",args_aor,endpoint_name,app_info->gs_uri,strlen(db_data)? db_data:"",app_info->conf_invite);

		if(strlen(db_data) && !strncasecmp(db_data, "no",2)){

			const char* endpoint_from = ast_sorcery_object_get_id(endpoint);
			const char* conf_room = pbx_builtin_getvar_helper(requestor, "CONF_ROOM");
			if(conf_room && !ast_strlen_zero(conf_room) && endpoint_from && strstr(endpoint_from, "trunk") == NULL &&
				!app_info->conf_invite){
				ast_log(LOG_DEBUG, "Conf Invite app,but the app dont support sip conf invite :%s\n",endpoint_name);
				ao2_ref(app_info, -1);
				return 0;
			}

			/*get register_time from astdb*/
			/*if reload ,endpoint need reset,set register_time again from astdb*/
			snprintf(register_key,sizeof(register_key),"%s/%s/%s",endpoint_name,app_info->client_uniqueId,"register_time");
			if((endpoint->register_time.tv_sec == 0) && !ast_db_get(GSWAVE_DB_FAMILY, register_key, register_time, sizeof(register_time)))
				ast_get_timeval(register_time, &app_info->register_time, ast_tv(0, 0), NULL);

			/*! args.aor is null ,when from callqueue */
			if(((!ast_strlen_zero(app_info->gs_uri)) && (NULL != args_aor) && strstr(args_aor,gs_uri))||(NULL == args_aor)) {

				if (ast_tvdiff_ms(ast_tvnow(), app_info->register_time) < REGISGTER_VALID_TIME*1000) {
					return 1;
				} else {
					/*! gswave didn't use long time,del it. */
					ast_sip_remove_waveinfo_from_database(endpoint_name,app_info);
					ast_sip_clear_waveapp_info(endpoint,app_info->client_uniqueId);
					ast_log(LOG_ERROR, "Gswave endpoint[%s] is Rigister timeout!,Remove Data from database!\n", endpoint_name);
					return 0;
				}
			}
			//gswave register timeout and just exist a siphone register，need remove gswave data in database
			else if (ast_tvdiff_ms(ast_tvnow(), app_info->register_time) > REGISGTER_VALID_TIME*1000) {
				/*! gswave didn't use long time,del it. */
				ast_sip_remove_waveinfo_from_database(endpoint_name,app_info);
				ast_sip_clear_waveapp_info(endpoint,app_info->client_uniqueId);
				ast_log(LOG_ERROR, "Gswave endpoint[%s] is Rigister timeout,Remove Data from database!\n", endpoint_name);
				return 0;
			}
		}
	}
	return 0;
}

static unsigned int ast_sip_handle_gswave_mobile_wake(struct ast_sip_endpoint *endpoint, const char *endpoint_name,
	struct ast_channel *requestor, const char *args_aor, char **dst,int need_prompt,struct ast_channel *callee_chan)
{
	char gs_uri[256]={0};
	struct timeval register_time_old;
	struct timeval send_msg_time;
	const char *dst_num = endpoint_name;
	const char *caller_num = dst_num;
	const char *caller_name = NULL;
	char family[64] = {0};
	char ring_time[64] = {0};
	int ring_timeout = 0;
	int play_prompt = 0;

	struct sip_waveapp_info* app_info = NULL;

	app_info = ast_sip_find_waveapp_info_by_uri(endpoint,args_aor);

	if(app_info == NULL){
		ast_log(LOG_DEBUG, "ast_sip_handle_gswave_mobile_wake can not find appinfo by uri:%s,endpoinrt:%s\n",args_aor,endpoint_name);
		return 0;
	}

	ast_log(LOG_DEBUG, "ast_sip_handle_gswave_mobile_wake aors:%s endpoint_name:%s,gs_uri:%s\n",args_aor,endpoint_name,app_info->gs_uri);

	/*some time the requestor is null from cc bug for 120770*/
	if (NULL != requestor) {
		/*Start added by zyxia for Bug 230147,2022.7.23*/
		caller_num = (char *)pbx_builtin_getvar_helper(requestor, "ANONYMOUS");
		if (caller_num && !strcasecmp(caller_num, "anonymous")) {
			ast_log(LOG_DEBUG, "Channel:%s  wake up  caller_num is anonymous\n",ast_channel_name(requestor));
			caller_name  = "Anonymous";
		}
		/*End added by zyxia for Bug 230147,2022.7.23*/
		else {
			/*get callerid number first for Bug 229143 ,2022.7.6*/
			caller_num = S_COR(ast_channel_caller(requestor)->id.number.valid,
						ast_channel_caller(requestor)->id.number.str, NULL);
			if (caller_num == NULL)
				caller_num = ast_channel_caller_num(requestor);

			caller_name = S_COR(ast_channel_caller(requestor)->id.name.valid,
						ast_channel_caller(requestor)->id.name.str, NULL);
			if (caller_name == NULL) {
				caller_name = pbx_builtin_getvar_helper(requestor, "Monitor_Name");
				if (ast_strlen_zero(caller_name))
					caller_name = pbx_builtin_getvar_helper(requestor, "AAUSERCIDNAME");
			}
		}
	}
	/*Start added by zyxia for Bug 195558,2021.5.25*/
	if (caller_num == NULL || caller_num[0] =='\0') {
		const char * conf_inv = pbx_builtin_getvar_helper(requestor, "CONF_INVITE");
		if ((NULL != conf_inv && 0 == strcmp(conf_inv, "yes"))){
			caller_num = pbx_builtin_getvar_helper(requestor,"CONF_ROOM");
			ast_log(LOG_DEBUG, "ast_sip_handle_gswave_mobile_wake aors:%s endpoint_name:%s,gs_uri:%s  channel:%s conf_room:%s\n",args_aor,endpoint_name,gs_uri,ast_channel_name(requestor),caller_num);
		}
	}

	register_time_old= app_info->register_time;

	/*if reload ,endpoint need reset,set gs_appid again from astdb*/
	//if(!ast_db_get(GSWAVE_DB_FAMILY, gs_appid_key, gs_appid, sizeof(gs_appid))&& !ast_strlen_zero(gs_appid)){
	//	ast_string_field_set(endpoint, gs_appid, gs_appid);
	//}
//	ast_string_field_set(endpoint, gs_uri, gs_uri);

	send_msg_time = ast_tvnow();
	/* send event to cti pbxmid */
	manager_event(EVENT_FLAG_COMMAND, "WakeUpGsWave",
	"Token: %s\r\n"
	"Caller_Num: %s\r\n"
	"Dst_Num: %s\r\n"
	"Caller_Name: %s\r\n",
	app_info->client_uniqueId,caller_num,dst_num,(caller_name == NULL ||  caller_name[0] == '\0')? caller_num:caller_name);
	ast_log(LOG_WARNING, "GsWave endpoint[%s] Send WakeUpGsWave Event  uniqueId:%s Caller_Num:%s Dst_Num:%s state:%d,Caller_name:%s\n",endpoint_name,app_info->client_uniqueId,caller_num,dst_num,ast_endpoint_get_state(endpoint->persistent),caller_name);


	/*if the gswave is offline and the caller destianion has only gswave app,need play prompt*/
	const char *has_several_destianion = pbx_builtin_getvar_helper(requestor, "HAS_SEVERAL_DIAL_DESTINATION");
	if (ast_endpoint_get_state(endpoint->persistent) == AST_ENDPOINT_OFFLINE_FOR_MOBILE
		&& !ast_strlen_zero(has_several_destianion) && !strcmp(has_several_destianion,"no")){
		/*! send 183 to caller and play prompt */
		if (judge_chan_type_is_sip(requestor) && !ast_channel_get_is_gds_call(requestor)) {//edit by Bug 261667
			ast_indicate(requestor,AST_CONTROL_PROGRESS);
			usleep(500000);//bug for 122423, some times can nott play the prompt
		}

		/*Start added by zyxia for Bug 256745,dont wait the prompt play finished*/
		if(requestor != NULL && !ast_channel_get_is_gds_call(requestor)) {//edit by Bug 261667
			ast_indicate_data(requestor, AST_CONTROL_PLAY_PROMPT, NULL, 0);
			play_prompt = 1;
		}
		/*End added by zyxia for Bug 256745,dont wait the prompt play finished*/

	}

	/*wait */
	/*Start added by zyxia for Bug 230192,2022.7.21*/
	snprintf(family, sizeof(family), "AAUSER/%s", endpoint_name);
	if (ast_ucm_db_get(family, "RT", ring_time, sizeof(ring_time))) {
		ast_log(LOG_WARNING, "DB: %s/RT not found in database.\n", family);
		ring_timeout = 60*1000;
	} else {
		if (ring_time[0] == 0)
			ring_timeout = 60*1000;
		else
			ring_timeout = atoi(ring_time)*1000;
	}
	ast_log(LOG_DEBUG, "Wave App Wait time %d ms\n", ring_timeout);
	/*End added by zyxia for Bug 230192,2022.7.21*/

	/*in ring time  ,if refresh register, get contact uri from endpoint */
	while( ast_tvdiff_ms(ast_tvnow(), send_msg_time) <= ring_timeout) {
		if (requestor != NULL && ast_channel_softhangup_internal_flag(requestor)) {
			ast_log(LOG_DEBUG, "The Caller '%s' is hang up,dont call the gswave \n",ast_channel_name(requestor));
			return 0;
		}
		/*if the callee channel is ringing or answerd,also dont wait*/
		else if(callee_chan != NULL && (ast_channel_state(callee_chan) == AST_STATE_UP
			|| ast_channel_state(callee_chan) == AST_STATE_RINGING )){
			ast_log(LOG_DEBUG, "The callee app '%s' is answered ,dont need wait the new contact \n",ast_channel_name(callee_chan));
			return 0;
		}

		/*! if gswave refresh register after wake up, get contact from aors */
		if(ast_tvdiff_ms(app_info->register_time, register_time_old) != 0 ) {
			if(!ast_strlen_zero(app_info->gs_new_contact_uri)) {
				/*Start added by zyxia for Bug 256745,dont wait the prompt play finished*/
				if(play_prompt && requestor) {
					pbx_builtin_setvar_helper(requestor, "STOP_PLAY_APP_PROMPT", "yes");//tell play thread ,dont need play
					ast_stopstream(requestor);
				}
				/*End added by zyxia for Bug 256745,dont wait the prompt play finished*/
				*dst = ast_strdup(app_info->gs_new_contact_uri);
				ast_endpoint_set_state(endpoint->persistent,AST_ENDPOINT_ONLINE);
				//ast_string_field_set(endpoint, gs_new_contact_uri,"");//delete by zyxia for bug263894,2023.7.4
				ast_log(LOG_DEBUG, "Gswave endpoint[%s] New Contact Uri is '%s'\n", endpoint_name,*dst);
				break;
			}
		}

		usleep(5000);
	}

	return 1;
}

/*End added by zyxia for wakeup ios gswave,2018.7.27*/

static int session_end_gswave_wakeup_defer(void *vsession)
{
	RAII_VAR(struct ast_sip_session *, session, vsession, ao2_cleanup);

	ast_sip_session_end_if_deferred(session);
	ast_sip_session_defer_termination_cancel(session);
	return 0;
}

static void* ast_call_gswave_mobile_app(void *data)
{
	struct request_data *req_data = data;
	char *tmp = ast_strdupa(req_data->dest), *endpoint_name = NULL, *request_user = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	unsigned char wake_ios_gswave = 0;
	int res=-1;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(endpoint);
		AST_APP_ARG(aor);
	);
	ast_log(LOG_DEBUG, "Enter ast_call_gswave_mobile_app dest:%s\n",tmp);

	if (ast_strlen_zero(tmp)) {
		ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty destination\n");
		goto out ;
	}

	AST_NONSTANDARD_APP_ARGS(args, tmp, '/');

	if (ast_sip_get_disable_multi_domain()) {
		/* If a request user has been specified extract it from the endpoint name portion */
		if ((endpoint_name = strchr(args.endpoint, '@'))) {
			request_user = args.endpoint;
			*endpoint_name++ = '\0';
		} else {
			endpoint_name = args.endpoint;
		}

		if (ast_strlen_zero(endpoint_name)) {
			if (request_user) {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name: %s@<endpoint-name>\n",
					request_user);
			} else {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name\n");
			}
			goto  out;
		}
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			endpoint_name);
		if (!endpoint) {
			ast_log(LOG_ERROR, "Unable to create PJSIP channel - endpoint '%s' was not found\n", endpoint_name);
			goto out;
		}
	} else {
		/* First try to find an exact endpoint match, for single (user) or multi-domain (user@domain) */
		endpoint_name = args.endpoint;
		if (ast_strlen_zero(endpoint_name)) {
			ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name\n");
			goto out;
		}

		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			endpoint_name);
		if (!endpoint) {
			/* It seems it's not a multi-domain endpoint or single endpoint exact match,
			 * it's possible that it's a SIP trunk with a specified user (user@trunkname),
			 * so extract the user before @ sign.
			 */
			endpoint_name = strchr(args.endpoint, '@');
			if (!endpoint_name) {
				/*
				 * Couldn't find an '@' so it had to be an endpoint
				 * name that doesn't exist.
				 */
				ast_log(LOG_ERROR, "Unable to create PJSIP channel - endpoint '%s' was not found\n",
					args.endpoint);
				goto out;
			}
			request_user = args.endpoint;
			*endpoint_name++ = '\0';

			if (ast_strlen_zero(endpoint_name)) {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name: %s@<endpoint-name>\n",
					request_user);
				goto out;
			}

			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
				endpoint_name);
			if (!endpoint) {
				goto out;
			}
		}
	}

	struct sip_waveapp_info* app_info = ast_sip_find_waveapp_info_by_uri(endpoint,args.aor);
	if(app_info){
		pbx_builtin_setvar_helper(req_data->old_channel, "app_instance",app_info->client_uniqueId);
	}

	if (ast_check_if_wakeup_needed(endpoint, endpoint_name,args.aor,req_data->requestor)) {
		req_data->session->is_gswave_app_call = 1;

		/*Start added by zyxia for Bug 281532,2024.3.21*/
		if(!ast_strlen_zero(args.aor) &&!check_uri_in_contact(args.aor,endpoint)){
			ast_log(LOG_DEBUG, "The Wave App addr is disconnect,Dont need send initial invite to %s endpoint_name:%s\n",args.aor,endpoint_name);
			req_data->session->inv_session->need_ignore = 1;//tellp pjsip stack,dont need send initial invite 
		}
		/*End added by zyxia for Bug 281532,2024.3.21*/

		if (ast_sip_session_defer_termination(req_data->session)) {
			ast_log(LOG_ERROR, "Received Gswave Wake up request on channel '%s' from endpoint '%s'	but could not defer termination, rejecting\n",
				ast_channel_name(req_data->session->channel), ast_sorcery_object_get_id(req_data->session->endpoint));
			goto out;
		}
	} else {
			req_data->session->is_gswave_app_call = 0;
			ast_log(LOG_DEBUG, "Dont need wake up,endpoint '%s',dst:%s\n", ast_sorcery_object_get_id(req_data->session->endpoint),args.aor);
			goto out;
	}

	RAII_VAR(char *, dst, NULL, ast_free);
	ast_channel_set_is_wave_app(req_data->old_channel,1);
	wake_ios_gswave = ast_sip_handle_gswave_mobile_wake(endpoint, endpoint_name,req_data->requestor, args.aor, &dst,0,req_data->old_channel);
	/*modify by zyxia for Bug 208588 ,the app is ringing ,dont call*/
	ast_log(LOG_DEBUG, "endpoint '%s',wake_ios_gswave:%d,gswave_app_session_active:%d,has_add_gs_call_count:%d,has_dec_gs_call_count:%d\n", ast_sorcery_object_get_id(req_data->session->endpoint),req_data->session->gswave_app_session_active,req_data->session->gswave_app_session_active,req_data->session->has_add_gs_call_count,req_data->session->has_dec_gs_call_count);



	/*modify by zyxia for Bug 208588 ,the app is ringing ,dont call*/
	if(!wake_ios_gswave || req_data->session->gswave_app_session_active || (req_data->session->has_add_gs_call_count && (req_data->session->has_dec_gs_call_count == 0))) {
		if (ast_sip_push_task(req_data->session->serializer, session_end_gswave_wakeup_defer, ao2_bump(req_data->session))) {
			/* Do it anyway even though this is not the right thread. */
			ast_sip_session_end_if_deferred(req_data->session);
			ast_sip_session_defer_termination_cancel(req_data->session);
			ao2_cleanup(req_data->session);
		}
		goto out;
	}

	ast_log(LOG_DEBUG, "call dst is %s\n",dst);

	if ( dst != NULL && req_data->requestor != NULL ) {
		struct ast_channel *new_channel; /* channel for this destination */
		int cause;
		char addr[256] = {0};
		snprintf(addr, sizeof(addr), "%s/%s", endpoint_name,dst);
		struct ast_assigned_ids assignedids = {
			.app_wakeup = 1,
		};

		new_channel = ast_request_with_stream_topology("PJSIP", req_data->topology, &assignedids, req_data->requestor, addr, &cause);
		if (new_channel) {

			struct ast_party_caller caller;
			ast_party_caller_set_init(&caller, ast_channel_caller(new_channel));
			ast_channel_set_caller_event(new_channel, ast_channel_caller(new_channel), NULL);
			struct ast_party_connected_line connected;
			ast_party_connected_line_set_init(&connected, ast_channel_connected(new_channel));
			ast_channel_set_connected_line(new_channel, ast_channel_connected(req_data->old_channel), NULL);
			ast_channel_set_is_wave_app(new_channel,1);
			/*set is readdy to calling app,tel monitor_dial for page call Bug252231*/
			ast_channel_set_is_in_calling_wave_app(new_channel,1);
			ast_channel_set_is_in_calling_wave_app(req_data->old_channel,1);
			if(app_info){
				pbx_builtin_setvar_helper(new_channel, "app_instance",app_info->client_uniqueId);
			}

			/*Start added by zyxia for Bug 254962 ,2023.3.24*/
			ast_channel_lock_both(req_data->old_channel, new_channel);
			ast_party_redirecting_copy(ast_channel_redirecting(new_channel), ast_channel_redirecting(req_data->old_channel));
			ast_channel_unlock(req_data->old_channel);
			ast_channel_unlock(new_channel);
			/*End added by zyxia for Bug 254962 ,2023.3.24*/


			if (ast_channel_move(req_data->old_channel, new_channel)) {
				ast_log(LOG_ERROR, "Unable to complete move channel pickup \n");
				ast_sip_session_end_if_deferred(req_data->session);
				ast_sip_session_defer_termination_cancel(req_data->session);
				ast_channel_set_is_in_calling_wave_app(new_channel,0);
				ast_channel_set_is_in_calling_wave_app(req_data->old_channel,0);
				ast_hangup(new_channel);
				goto out;
			}

			ast_sip_session_end_if_deferred(req_data->session);
			ast_sip_session_defer_termination_cancel(req_data->session);
			ast_log(LOG_DEBUG, "Channel name:%s \n",ast_channel_name(new_channel));
			ast_cdr_newest_setvar(ast_channel_name(new_channel), "special_action_type", "WAKEUP_APP");//added by zyxia for Bug 269407,dont need this cdr,2023.8.22
			ast_hangup(new_channel);
			/*Bug 281996*/
			if (!ast_strlen_zero(ast_channel_name(req_data->requestor)) && strstr(ast_channel_name(req_data->requestor),"Local/") && strstr(ast_channel_name(req_data->requestor),";2")) {
				const char* is_sfu = pbx_builtin_getvar_helper(req_data->requestor, "IS_SFU");
				if (ast_true(is_sfu)) {
					ast_cdr_oldest_setvar(ast_channel_name(req_data->requestor), "special_action_type", "CONF_WAKEUP");
				}
			}

			res = ast_call(req_data->old_channel, addr, 0); /* Place the call, but don't wait on the answer */
			ast_channel_set_is_in_calling_wave_app(req_data->old_channel,0);

			/* check the results of ast_call */
			if (res) {
				/* Again, keep going even if there's an error */
				ast_log(LOG_ERROR, "Unable to Call the %s \n",addr);
				goto  out;
			}
		}
	}
out:
	ao2_cleanup(req_data->topology);
	ao2_cleanup(req_data->requestor);
	ao2_cleanup(req_data->old_channel);
	ao2_cleanup(req_data->session);
	ast_free((char *)req_data->dest);
	ast_free(req_data);
	ast_log(LOG_DEBUG, "Leave  ast_call_gswave_mobile_app\n");
	pthread_exit(NULL);

	return NULL;
}

#if 0
static int call_number_is_only_gswave(struct ast_sip_endpoint* endpoint)
{
	struct ao2_container* contacts = NULL;
	struct ast_sip_contact* contact = NULL;
	struct ast_sip_aor* aor = NULL;
	struct ao2_iterator i;
	int is_only_gswave = 1;

	aor = ast_sip_location_retrieve_aor(endpoint->aors);
	if (!aor) {
		return 0;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	ao2_ref(aor, -1);
	if (!contacts) {
		return 0;
	}

	i = ao2_iterator_init(contacts, 0);

	while ((contact = ao2_iterator_next(&i))) {
		if (!contact->gswave_mobile && !strstr(contact->uri, "WS")) {
			is_only_gswave = 0;
		}

		ao2_ref(contact, -1);

		if (!is_only_gswave) {
			break;
		}
	}

	ao2_iterator_destroy(&i);
	ao2_ref(contacts, -1);

	ast_log(LOG_DEBUG, "The endpoint:(%s) %s only gswave register.",
			ast_sorcery_object_get_id(endpoint), is_only_gswave ? "is" : "isn't");

	return is_only_gswave;
}
#endif
#endif

static int request(void *obj)
{
	struct request_data *req_data = obj;
	struct ast_sip_session *session = NULL;
	char *tmp = ast_strdupa(req_data->dest), *endpoint_name = NULL, *request_user = NULL;
	struct ast_sip_endpoint *endpoint;
#ifdef GRANDSTREAM_NETWORKS
	/*start added by zyxia for wakeup ios gswave*/
	enum ast_endpoint_state state;
	unsigned char wake_ios_gswave = 0;
	/*end added by zyxia for wakeup ios gswave*/
#endif

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(endpoint);
		AST_APP_ARG(aor);
		AST_APP_ARG(local_tag);
		AST_APP_ARG(remote_tag);
		AST_APP_ARG(callid);
	);

	if (ast_strlen_zero(tmp)) {
		ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty destination\n");
		req_data->cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
		return -1;
	}

	AST_NONSTANDARD_APP_ARGS(args, tmp, '/');

	if (ast_sip_get_disable_multi_domain()) {
		/* If a request user has been specified extract it from the endpoint name portion */
		if ((endpoint_name = strchr(args.endpoint, '@'))) {
			request_user = args.endpoint;
			*endpoint_name++ = '\0';
		} else {
			endpoint_name = args.endpoint;
		}

		if (ast_strlen_zero(endpoint_name)) {
			if (request_user) {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name: %s@<endpoint-name>\n",
					request_user);
			} else {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name\n");
			}
			req_data->cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
			return -1;
		}
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			endpoint_name);
		if (!endpoint) {
			ast_log(LOG_ERROR, "Unable to create PJSIP channel - endpoint '%s' was not found\n", endpoint_name);
			req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
			return -1;
		}
	} else {
		/* First try to find an exact endpoint match, for single (user) or multi-domain (user@domain) */
		endpoint_name = args.endpoint;
		if (ast_strlen_zero(endpoint_name)) {
			ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name\n");
			req_data->cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
			return -1;
		}
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			endpoint_name);
		if (!endpoint) {
			/* It seems it's not a multi-domain endpoint or single endpoint exact match,
			 * it's possible that it's a SIP trunk with a specified user (user@trunkname),
			 * so extract the user before @ sign.
			 */
			endpoint_name = strchr(args.endpoint, '@');
			if (!endpoint_name) {
				/*
				 * Couldn't find an '@' so it had to be an endpoint
				 * name that doesn't exist.
				 */
				ast_log(LOG_ERROR, "Unable to create PJSIP channel - endpoint '%s' was not found\n",
					args.endpoint);
				req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
				return -1;
			}
			request_user = args.endpoint;
			*endpoint_name++ = '\0';

			if (ast_strlen_zero(endpoint_name)) {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel with empty endpoint name: %s@<endpoint-name>\n",
					request_user);
				req_data->cause = AST_CAUSE_CHANNEL_UNACCEPTABLE;
				return -1;
			}

			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
				endpoint_name);
			if (!endpoint) {
				ast_log(LOG_ERROR, "Unable to create PJSIP channel - endpoint '%s' was not found\n", endpoint_name);
				req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
				return -1;
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Bug 193469 */
		if (!strncmp(endpoint_name, "trunk_", 6) && !request_user) {
			request_user = args.aor;
			if (!strchr(args.aor, '@')) {
				args.aor = NULL;
			}
		}
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
		RAII_VAR(char *, dst, NULL, ast_free);
		state = ast_endpoint_get_state(endpoint->persistent);
		if (state == AST_ENDPOINT_OFFLINE) {
			/* Added by yfyang, for Bug 185341*/
			if (!strncmp(endpoint_name, "trunk_", 6)) {
				ast_log(LOG_ERROR, "endpoint[%s] is OFFLINE!\n", endpoint_name);
				req_data->cause = AST_CAUSE_UNREACHABLE;

				/* Added by yfyang, for Bug312578 */
				if (req_data->requestor) {
					ast_channel_hangupsipcode_set(req_data->requestor, 0);
				}

				ao2_ref(endpoint, -1); 
				return -1;
			}

			/*start if the register data is not exsitd, need check the endpoint */
			if ( ast_check_if_wakeup_needed(endpoint, endpoint_name,args.aor,req_data->requestor) ) {
				wake_ios_gswave = ast_sip_handle_gswave_mobile_wake(endpoint, endpoint_name, req_data->requestor, args.aor, &dst,1,NULL);
				if(!wake_ios_gswave) {
					ast_log(LOG_ERROR, "endpoint[%s] is OFFLINE!\n", endpoint_name);
					if (!strncmp(endpoint_name, "trunk_", 6)) {
						req_data->cause = AST_CAUSE_UNREACHABLE;	/* for bug 86650 */
					} else {
						req_data->cause = ast_sip_get_unavailable_cause();	  /* for bug 178724, 265759 */
					}
					ao2_ref(endpoint, -1);
					return -1;
				}
				/* end added by zyxia for  gswave wakeup,2018.8.16*/
			} else {
				ast_log(LOG_ERROR, "endpoint[%s] is OFFLINE!\n", endpoint_name);

				req_data->cause = ast_sip_get_unavailable_cause();	  /* for bug 178724, 265759 */
				ao2_ref(endpoint, -1);
				return -1;
			}
		}

	/* Added by yfyang, for Bug 221341 */
	/* check callwaiting & extension state: /AAUSER/9004/CALLWAITING : no  */
	char db_family[80] = {0};
	char db_value[80] = {0};

	snprintf(db_family, sizeof(db_family), "AAUSER/%s", endpoint_name);
	ast_ucm_db_get(db_family, "CALLWAITING", db_value, sizeof(db_value));

	if (!ast_strlen_zero(db_value) && !strcmp(db_value, "no")) {
		enum ast_extension_states ext_state = ast_extension_state(NULL, "ext-local", endpoint_name);

		/*Start added by zyxia for Bug 262678,2023.6.14*/
		if(ast_sip_exist_waveapp_info(endpoint) && (req_data->app_wakeup || ext_state == AST_EXTENSION_UNAVAILABLE)){//edit by zyxia for Bug 292255
			ast_log(LOG_NOTICE, "endpoint[%s] disabled callwaiting,but is wave app and the state is [%s] or is wakeup call %d, need call it!\n", endpoint_name, ast_extension_state2str(ext_state),req_data->app_wakeup);
		}
		/*End added by zyxia for Bug 262678,2023.6.14*/
		else if (ext_state != AST_EXTENSION_NOT_INUSE) {
			ast_log(LOG_NOTICE, "endpoint[%s] disabled callwaiting, and the state is [%s]!\n", endpoint_name, ast_extension_state2str(ext_state));
			req_data->cause = AST_CAUSE_BUSY;
			ao2_ref(endpoint, -1);
			return -1;
		}
	}

	pj_str_t local_tag = {0}, remote_tag = {0}, cid = {0};
	struct recovery_db_entry* entry = NULL;
	struct recovery_db_entry* inbound_mode_sub_entry = NULL;
	switch_xml_t vars,restore_xml = NULL;
	switch_xml_t from,to,call_id;
	char need_recovery = 0;

	if(args.argc == 5)
	{
		/* for rebuild caller dialog */
		local_tag.ptr = args.local_tag;
		local_tag.slen = strlen(args.local_tag);
		remote_tag.ptr = args.remote_tag;
		remote_tag.slen = strlen(args.remote_tag);
		cid.ptr = args.callid;
		cid.slen = strlen(args.callid);
		need_recovery=1;
	}
	else if(req_data->requestor && ast_test_flag(ast_channel_flags(req_data->requestor), AST_FLAG_RECOVERYING))
	{
		inbound_mode_sub_entry = ast_recovery_db_get_all();

		for (entry = inbound_mode_sub_entry; entry; entry = entry->next) {
			if(0 == strcmp(entry->runtime_uuid, ast_channel_recovery_linkedid( req_data->requestor)))
			{
				if(strcmp(entry->uuid, ast_channel_recovery_uniqueid( req_data->requestor)))
				{
					/* for rebuild callee dialog */
					restore_xml = switch_xml_parse_str_dynamic(entry->data, SWITCH_TRUE);
					vars = switch_xml_child(restore_xml, "variables");
					from = switch_xml_child(vars, "from_tag");
					to = switch_xml_child(vars, "to_tag");
					call_id = switch_xml_child(vars, "call_id");

					local_tag.ptr = from->txt;
					local_tag.slen = strlen(from->txt);
					remote_tag.ptr = to->txt;
					remote_tag.slen = strlen(to->txt);
					cid.ptr = call_id->txt;
					cid.slen = strlen(call_id->txt);
					need_recovery=1;
					break;

				}
			}
		}


	}
	if(need_recovery)
	{
		session = ast_sip_session_create_outgoing(endpoint, NULL, dst?dst:args.aor, request_user,
			req_data->topology,req_data->requestor, &local_tag, &remote_tag,&cid);
	}
	else
	{
#ifdef BUILD_ON_CLOUD_PBX
		char *sbc_param = NULL;
		if (args.aor) {
			sbc_param = strstr(args.aor, ";sbc-dest=");
			if (sbc_param) {
				*sbc_param = '\0';
				sbc_param += strlen(";sbc-dest=");
			}
		}
#endif
        session = ast_sip_session_create_outgoing(endpoint, NULL, dst ? dst : args.aor, request_user,
			req_data->topology, req_data->requestor,NULL, NULL, NULL);

#ifdef BUILD_ON_CLOUD_PBX
		if (sbc_param) {
			snprintf(session->sbc_endpoint_addr, sizeof(session->sbc_endpoint_addr), "%s", sbc_param);
		}
#endif
	}
#else
	session = ast_sip_session_create_outgoing(endpoint, NULL, args.aor, request_user,
		req_data->topology);
#endif
	if(restore_xml)
	{
		switch_xml_free(restore_xml);
	}

	if (inbound_mode_sub_entry) {
		ast_recovery_freetree(inbound_mode_sub_entry);
	}
	ao2_ref(endpoint, -1);
	if (!session) {
		ast_log(LOG_ERROR, "Failed to create outgoing session to endpoint '%s'\n", endpoint_name);
#ifndef GRANDSTREAM_NETWORKS
		req_data->cause = AST_CAUSE_NO_ROUTE_DESTINATION;
#else
		req_data->cause = AST_CAUSE_SUBSCRIBER_ABSENT;	  /* for bug 209460 */
#endif
		return -1;
	}

#if 0
	int reject_call = ast_channel_get_reject_call(req_data->requestor);
	if (reject_call && call_number_is_only_gswave(session->endpoint)) {
		if (reject_call & (0x01 << 1)) {
			ast_log(LOG_DEBUG, "Need reject this call, response cause 'AST_CAUSE_CONF_SKIP_CALL_WAVE'.");
			req_data->cause = AST_CAUSE_CONF_SKIP_CALL_WAVE;
		} else {
			ast_log(LOG_DEBUG, "Need reject this call, response 603 Decline.");
			req_data->cause = 603;
		}

		ast_queue_hangup_with_cause(req_data->requestor, req_data->cause);
		ast_queue_hangup(req_data->requestor);
		return -1;
	}
#endif
	req_data->session = session;

	return 0;
}

/* add by jlhu for trunk hunt begin */
#ifdef GRANDSTREAM_NETWORKS
int chan_pjsip_groupfull(struct ast_channel *requestor, const char *data)
{
	struct ast_group_info *gi = NULL;
        int loop = 0;
        char *s = NULL;
	char *t = NULL;
	const char *regstr;
	struct ast_channel *dest = NULL;
        unsigned char reqlevel;
	unsigned char minlevel;

        if (NULL == data || NULL == requestor)
	{
	    return 0;
	}

	if (requestor)
	{
	        regstr = pbx_builtin_getvar_helper(requestor, "URGENCYLEVEL");
		if (!regstr)
		{
		    ast_log(LOG_WARNING, " not a urgent call \n ");
		    return 0;
		}
		reqlevel = atoi(regstr);
		minlevel = reqlevel;
	}

        /* e.g. data: 1020@trunk_3 */
        if((s = strchr(data, '_')) != NULL)
	{
	   s++;
	}

	ast_app_group_list_rdlock();
	gi = ast_app_group_list_head();
	while (gi)
	{

		if (gi->chan)
		{
			/* e.g. gi-group: OUT_3, 3 is trunk number  */
			if((t = strchr(gi->group, '_')) != NULL)
			{
			        t++;

				/* belong to the same trunk number */
				if (0 == strcmp(s,t)
					&& (AST_STATE_UP == ast_channel_state(gi->chan)
					|| AST_STATE_DIALING == ast_channel_state(gi->chan)
					|| AST_STATE_RING == ast_channel_state(gi->chan)
					|| AST_STATE_RINGING == ast_channel_state(gi->chan)))
				{
					if (ast_channel_trunk_priority_get(gi->chan) < reqlevel)
					{
						if (ast_channel_trunk_priority_get(gi->chan) < minlevel)
						{
							minlevel =	ast_channel_trunk_priority_get(gi->chan);
							dest = gi->chan;
						}

					}
				}
			}
		}

		gi = AST_LIST_NEXT(gi, group_list);
		loop++;
		t = NULL;
	}

	ast_app_group_list_unlock();

	if (NULL != dest)
	{
		ast_log(LOG_NOTICE, " force release siptk channel!!! chan = %p,levl=%d\n",dest, minlevel);
		ast_queue_hangup_with_cause(dest, AST_CAUSE_NORMAL_CLEARING);
		return 1;
	}
	else
	{
		ast_log(LOG_WARNING, "  no higher hunt priority ,reqlevl = %d \n",reqlevel);
		return 0;
	}

}

static enum ast_stream_video_resolution get_configured_video_resolution(const struct ast_channel *chan)
{
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;
	const char *video_resolution_str = NULL;

	if (!chan) {
		return resolution;
	}

	video_resolution_str = pbx_builtin_getvar_helper((struct ast_channel *)chan, "VIDEO_RESOLUTION");
	if (ast_strlen_zero(video_resolution_str)) {
		return resolution;
	}

	if (!strcasecmp(video_resolution_str, "1080P")) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_1080P;
	} else if (!strcasecmp(video_resolution_str, "720P")) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_720P;
	} else if (!strcasecmp(video_resolution_str, "360P")) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_360P;
	}

	return resolution;
}

static void set_session_sfu_params_according_chan(struct ast_sip_session *session, const struct ast_channel *chan)
{
	const char *is_sfu = NULL;
	const char *conf_room = NULL;

	if (!session || !chan) {
		return;
	}

	if (session->is_sfu) {
		return;
	}

	/*Start added by zyxia for Bug266995,2023.7.28*/
	if(ast_channel_tech(chan)->type && !strncasecmp(ast_channel_tech(chan)->type, "PJSIP", 5)){
		struct ast_sip_channel_pvt *channel = NULL;
		channel = ast_channel_tech_pvt(chan);
		if (channel && channel->session) {
			ast_log(LOG_DEBUG, "[session:%p] Set is_trunk_conf_invite true for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->is_trunk_conf_invite = channel->session->is_trunk_conf_invite;
			if(session->is_trunk_conf_invite)
				session->reject_call = channel->session->reject_call; //added by zyxia from Bug 267608
		}
	}
	/*End added by zyxia for Bug266995,2023.7.28*/


	is_sfu = pbx_builtin_getvar_helper((struct ast_channel *)chan, "IS_SFU");
	if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
		ast_log(LOG_DEBUG, "[session:%p] Set is_sfu true for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->is_sfu = 1;

		if (ast_sip_session_judge_if_type_is_local(chan)) {
			session->sfu_invite_by_local_chan = 1;
		}

		session->configured_video_resolution = get_configured_video_resolution(chan);
	}

	conf_room = pbx_builtin_getvar_helper((struct ast_channel *)chan, "CONF_ROOM");
	if (!ast_strlen_zero(conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] Set conf_room:%s for endpoint:%s\n",
			session, conf_room, ast_sorcery_object_get_id(session->endpoint));
		strncpy(session->conf_room, conf_room, sizeof(session->conf_room) - 1);
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_UNKNOW) {
		return;
	}

	/* For SFU invite call, we just distinguish webrtc and SIP, set terminal_type again when recieve answer */
	if (session->is_webrtc_call) {
		session->terminal_type = AST_TERMINAL_TYPE_WEBRTC;
	} else {
		session->terminal_type = AST_TERMINAL_TYPE_SIP;
	}

	return;
}

static unsigned int judge_if_topology_has_slides_video(struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	unsigned int res = 0;
	struct ast_stream *stream = NULL;

	if (!topology) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			continue;
		}

		if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
			res = 0;
		} else {
			res = 1;
		}

		ast_log(LOG_DEBUG, "Find slides video stream for topology:%p\n", topology);

		break;
	}

	return res;
}

static void set_opus_codec_inbandfec(struct ast_sip_session *session, struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !topology) {
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
			stream = ast_stream_topology_get_stream(topology, i);
		if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO) {
			continue;
		}
		ast_sip_session_process_special_audio_codec(session, stream, ast_stream_get_formats(stream));
		break;
	}

	return;
}

static void append_video_codec_to_caps(struct ast_sip_session *session,
		struct ast_format_cap *caps, struct ast_format *video_format)
{
	/* Parms checks before, don't need do again */

	if (!ast_format_cap_has_format(session->endpoint->media.codecs, video_format)) {
		return;
	}

	ast_format_cap_append(caps, video_format, 0);

	/* append RTX codec if endpoint configure  */
	if (ast_format_cap_has_format(session->endpoint->media.codecs, ast_format_rtx)) {
		ast_format_cap_append(caps, ast_format_rtx, 0);
	}

	/* append video FEC codec if endpoint configure  */
	if (ast_format_cap_has_format(session->endpoint->media.codecs, ast_format_fec)) {
		ast_format_cap_append(caps, ast_format_fec, 0);
	}

	return;
}

static void append_codecs_to_caps(struct ast_sip_session *session, struct ast_format_cap *caps)
{
	if (!session || !caps) {
		return;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			append_video_codec_to_caps(session, caps, ast_format_h264);
			break;
		case AST_SFU_VIDEO_H265:
			append_video_codec_to_caps(session, caps, ast_format_h265);
			break;
		case AST_SFU_VIDEO_VP9:
			append_video_codec_to_caps(session, caps, ast_format_vp9);
			break;
		default:
			append_video_codec_to_caps(session, caps, ast_format_h264);
			break;
	}

	return;
}

static int chan_pjsip_handle_no_pending_media(struct ast_sip_session *session)
{
	unsigned int create_bfcp = 0;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	enum ast_sfu_invite_type invite_type = AST_SFU_INVITE_TYPE_VIDEO;

	ast_log(LOG_DEBUG, "chan_pjsip_handle_no_pending_media begin\n");

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate capabilities for handle no pending media\n");
		return -1;
	}

	/* SFU we need append video caps, otherwise we just append audio caps */
	if (session->is_sfu) {
		append_codecs_to_caps(session, caps);
		ast_sip_session_change_gs_fec_version(caps);

		if (ast_sip_session_get_only_support_audio()) {
			invite_type = AST_SFU_INVITE_TYPE_AUDIO;
		}

		if (session->is_webrtc_call) {
			/* We use video call for default */
			ast_log(LOG_DEBUG, "Create pending media state for webrtc call\n");
			session->pending_media_state->topology = ast_stream_topology_webrtc_create_from_format_cap_for_sfu(caps, invite_type);
		} else {
			create_bfcp = ast_sip_session_need_create_bfcp_media(session);
			session->pending_media_state->topology = ast_stream_topology_sip_create_from_format_cap_for_sfu(caps, invite_type, create_bfcp);
			if (create_bfcp && session->pending_media_state->topology &&
				judge_if_topology_has_slides_video(session->pending_media_state->topology)) {
				bfcp_create_params_for_sfu_invite(session);
			}
		}
	} else {
		ast_log(LOG_DEBUG, "Create pending media state for normal call, just add audio codec\n");
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
		if (session->is_video_page_call)
			ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);
		ast_sip_session_change_gs_fec_version(caps);
		session->pending_media_state->topology = ast_stream_topology_create_from_format_cap(caps);
		set_opus_codec_inbandfec(session, session->pending_media_state->topology);

		if (session->is_webrtc_call && ast_strlen_zero(session->conf_room)) {
			ast_sip_session_add_missed_line_for_gswave(session, NULL);
		}
	}

	if (!session->pending_media_state->topology) {
		ast_log(LOG_ERROR, "Can't alloc pending media\n");
		pjsip_inv_terminate(session->inv_session, 500, PJ_FALSE);
		ao2_ref(session, -1);
		return -1;
	}

	return 0;
}

static int chan_pjsip_modify_sfu_pending_media(struct ast_sip_session *session)
{
	unsigned int create_bfcp = 0;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (!session->is_sfu) {
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] chan_pjsip_modify_sfu_pending_media for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "[session:%p] Failed to allocate capabilities for handle no pending media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	/* Clear pending_media_state topology first, then use endpoint's codec to create topology */
	ast_stream_topology_free(session->pending_media_state->topology);
	session->pending_media_state->topology = NULL;

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);
	ast_sip_session_change_gs_fec_version(caps);
	if (session->is_webrtc_call) {
		session->pending_media_state->topology = ast_stream_topology_webrtc_create_from_format_cap_for_sfu(caps, AST_SFU_INVITE_TYPE_VIDEO);
	} else {
		create_bfcp = ast_sip_session_need_create_bfcp_media(session);
		session->pending_media_state->topology = ast_stream_topology_sip_create_from_format_cap_for_sfu(caps, AST_SFU_INVITE_TYPE_VIDEO, create_bfcp);
		if (create_bfcp && session->pending_media_state->topology &&
			judge_if_topology_has_slides_video(session->pending_media_state->topology)) {
			bfcp_create_params_for_sfu_invite(session);
		}
	}

	if (!session->pending_media_state->topology) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc pending media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		pjsip_inv_terminate(session->inv_session, 500, PJ_FALSE);
		ao2_ref(session, -1);
		return -1;
	}

	return 0;
}

static void set_stream_cap_use_exist_codec(struct ast_sip_session *session,
	struct ast_stream *stream, char *conf_codec)
{
	struct ast_format_cap *caps = NULL;

	caps = ast_stream_get_formats(stream);
	if (!caps) {
		return;
	}

	ast_format_cap_remove_codecs_except_conf_video_codec(caps, conf_codec);

	return;
}

static void set_stream_cap_as_default(struct ast_sip_session *session,
	struct ast_stream *stream)
{
	struct ast_format_cap *caps = NULL;

	caps = ast_stream_get_formats(stream);
	if (!caps) {
		return;
	}

	ast_format_cap_remove_codecs_except_conf_video_codec(caps, "H264");

	return;
}

static void set_conf_video_codec_for_outgoing(struct ast_sip_session *session,
	struct ast_stream *stream)
{
	char *conf_codec = NULL;

	if (!stream || !session) {
		return;
	}

	conf_codec = ast_conf_get_video_codec(session->channel, session->conf_room);

	/* If conf already has video codec, don't need set again */
	if (!ast_strlen_zero(conf_codec)) {
		set_stream_cap_use_exist_codec(session, stream, conf_codec);
		ast_conf_set_video_codec(session->channel, session->conf_room, conf_codec);
		ast_free(conf_codec);
		conf_codec = NULL;
	} else {
		/* If we can't get conf codec, use H264 as default */
		set_stream_cap_as_default(session, stream);
		ast_conf_set_video_codec(session->channel, session->conf_room, "H264");
	}

	return;
}

static enum ast_stream_video_resolution get_sip_terminal_resolution_for_no_resource(struct ast_sip_session *session,
	struct ast_stream *stream)
{
	RAII_VAR(struct ast_channel *, host_channel, NULL, ao2_cleanup);
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_stream *host_stream = NULL;
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session || !stream) {
		return AST_STREAM_VIDEO_RESOLUTION_NONE;
	}

	ast_stream_set_local_media_direction(stream, SENDONLY);

	host_channel = ast_conf_get_sfu_moderator(NULL, session->conf_room);
	if (!host_channel) {
		/* If has no host, use 360P as default */
		ast_log(LOG_DEBUG, "[session:%p] get_sip_terminal_resolution_for_no_resource can't get host channel, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return AST_STREAM_VIDEO_RESOLUTION_360P;
	}

	chan_pvt = ast_channel_tech_pvt(host_channel);
	if (!chan_pvt || !chan_pvt->session) {
		return AST_STREAM_VIDEO_RESOLUTION_360P;
	}

	host_stream = ast_sip_session_get_active_main_video_stream(chan_pvt->session);
	if (!host_stream) {
		return AST_STREAM_VIDEO_RESOLUTION_360P;
	}

	resolution = ast_stream_get_video_resolution(host_stream);
	if (resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get host resolution, use 360P as default, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		resolution = AST_STREAM_VIDEO_RESOLUTION_360P;
	}

	ast_log(LOG_DEBUG, "[session:%p] get_sip_terminal_resolution_for_no_resource host's resolution:%d, endpoint:%s\n",
		session, resolution, ast_sorcery_object_get_id(session->endpoint));

	return resolution;
}

static void chan_pjsip_modify_sfu_stream_params_for_audio(struct ast_sip_session *session, struct ast_stream *stream)
{
	struct ast_format_cap *caps = NULL;

	caps = ast_stream_get_formats(stream);

	if (ast_format_cap_has_opus(caps)) {
		/* set useinbandfec true as default */
		ast_sip_session_set_format_useinbandfec(session, caps, stream, 1);
		ast_sip_session_set_format_dtx(session, caps, stream);
	}

	return;
}

static unsigned int is_video_caps_valid(struct ast_sip_session *session, struct ast_stream *stream)
{
	struct ast_format_cap *caps = NULL;

	caps = ast_stream_get_formats(stream);
	if (ast_format_cap_count(caps) == 0 ) {
		ast_log(LOG_DEBUG, "[session:%p] Get no video codec, remove this video stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (ast_format_cap_has_rtx(caps) && ast_format_cap_count(caps) == 1) {
		ast_log(LOG_DEBUG, "[session:%p] Only has RTX codec, remove this video stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (ast_format_cap_has_video_fec(caps) && ast_format_cap_count(caps) == 1) {
		ast_log(LOG_DEBUG, "[session:%p] Only has video FEC, remove this video stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (ast_format_cap_has_rtx(caps) && ast_format_cap_has_video_fec(caps) && ast_format_cap_count(caps) == 2) {
		ast_log(LOG_DEBUG, "[session:%p] Only has RTX and video FEC, remove this video stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	return 1;
}

static enum ast_stream_video_resolution modify_main_video_for_no_resource(struct ast_sip_session *session,
	struct ast_stream *stream, struct ast_stream_topology *topology, int postion)
{
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!ast_sip_session_support_multiple_video(session)) {
		/* For SIP terminal, we get host's video resolution */
		resolution = get_sip_terminal_resolution_for_no_resource(session, stream);
	} else {
		/* For gswave we need to reserve main video m line, for SIP terminal we don't reserve, just delete */
		if (session->is_webrtc_call) {
			ast_log(LOG_DEBUG, "[session:%p] Has no video resolution, need to change stream state to removed, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			if (ast_sip_session_media_use_bundle(session)) {
				ast_sip_session_set_video_format(session, stream, VIDEO_RESOLUTION_360P);
				ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
				ast_stream_set_local_media_direction(stream, INACTIVE);
				ast_stream_set_remote_media_direction(stream, INACTIVE);
			} else {
				ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
			}
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Has no video resolution, delete main video stream for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_topology_del_stream(topology, postion);
		}
	}

	return resolution;
}

static void chan_pjsip_modify_sfu_stream_params_for_main_video(struct ast_sip_session *session,
	struct ast_stream *stream, struct ast_stream_topology *topology, int postion)
{
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session || !stream) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] chan_pjsip_modify_sfu_stream_params_for_main_video stream:%p, endpoint:%s\n",
		session, stream, ast_sorcery_object_get_id(session->endpoint));

	if (!ast_sip_session_support_multiple_video(session) && ast_format_cap_count(ast_stream_get_formats(stream)) == 0) {
		ast_log(LOG_DEBUG, "[session:%p] SIP terminal has no video codec, delete main video stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_topology_del_stream(topology, postion);
		return;
	}

	/* For outgoing, if has configured value, use configured, otherwise invite as audio call */
	if (session->configured_video_resolution) {
		resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, session->configured_video_resolution);
	}
	if ((resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) || (resolution == -1)) {
		resolution = modify_main_video_for_no_resource(session, stream, topology, postion);
		if (resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) {
			return;
		}
	}

	set_conf_video_codec_for_outgoing(session, stream);

	if (!is_video_caps_valid(session, stream)) {
		ast_log(LOG_DEBUG, "[session:%p] Get no video codec or only has RTX codec after set conf codec for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
		ast_conf_put_main_video_resource(session->conf_room, session->channel);
		return;
	}

	ast_sip_session_set_video_format(session, stream, resolution);
	ast_stream_set_video_resolution(stream, resolution);
	ast_sip_session_set_as_tias_for_sfu(session, stream);
	ast_sip_session_set_imageattr_for_stream(session, stream, resolution);

	return;
}

static void modify_stream_state(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!session || !stream) {
		return;
	}

	if (!is_video_caps_valid(session, stream)) {
		ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
		return;
	}

	return;
}

static void chan_pjsip_modify_sfu_stream_params_for_slides_video(struct ast_sip_session *session,
	struct ast_stream *stream)
{
	if (!session || !stream) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] chan_pjsip_modify_sfu_stream_params_for_slides_video stream:%p endpoint:%s\n",
		session, stream, ast_sorcery_object_get_id(session->endpoint));

	modify_stream_state(session, stream);

	/* For outgoing request, we use 1080P as default */
	ast_sip_session_set_video_format(session, stream, AST_STREAM_VIDEO_RESOLUTION_1080P);

	ast_sip_session_set_as_tias_for_sfu(session, stream);

	/* For outgoing request, we use 1080P as default */
	ast_sip_session_set_imageattr_for_stream(session, stream, AST_STREAM_VIDEO_RESOLUTION_1080P);

	return;
}

static void chan_pjsip_modify_sfu_stream_params(struct ast_sip_session *session)
{
	int i = 0;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!session || !session->pending_media_state) {
		return;
	}

	topology = session->pending_media_state->topology;
	if (!topology) {
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		sub_type = ast_stream_get_sub_type(stream);
		switch (sub_type) {
			case AST_MEDIA_SUB_AUDIO_TYPE:
				chan_pjsip_modify_sfu_stream_params_for_audio(session, stream);
				break;
			case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
				chan_pjsip_modify_sfu_stream_params_for_main_video(session, stream, topology, i);
				break;
			case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
				chan_pjsip_modify_sfu_stream_params_for_slides_video(session, stream);
				break;
			default:
				break;
		}
	}

	return;
}

static int set_peer_src_addr(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "Get peer src addr error\n");
		return -1;
	}

	if (0 != strcmp(ast_channel_tech(requestor)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "requestor type is not pjsip");
		return -1;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_log(LOG_WARNING, "Get peer src addr, requestor has no session\n");
		return -1;
	}

	ast_channel_set_elin_ip_ver(session->channel, channel->session->src_addr_ver);
	if (AF_INET == channel->session->src_addr_ver) {
		ast_channel_set_elin_ip(session->channel, &channel->session->src_addr);
	} else if (AF_INET6 == channel->session->src_addr_ver) {
		ast_channel_set_elin_ip6(session->channel, &channel->session->src_addr6);
	}

	return 0;
}

static void set_channel_media_encryption_for_recovery(struct ast_sip_session *session, switch_xml_t restore_xml)
{
	struct ast_channel_media_encryption encryption;

	if (!session || !session->channel) {
		return;
	}

	memset(&encryption, 0, sizeof(struct ast_channel_media_encryption));
	set_media_encryption_from_restore_xml(&encryption, restore_xml);
	ast_channel_media_encryption_set(session->channel, encryption);

	return;
}

static int check_meeting_invitation(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	char *user_exten = NULL;
	int requestor_hangup = 0;
	if (requestor
		&& session->is_sfu
		&& session->channel
		&& ast_channel_name(requestor)
		&& strstr(ast_channel_name(requestor), "Local/")
		&& strstr(ast_channel_name(requestor), ";2")) {
			if (session->endpoint) {
				user_exten = (char *)session->endpoint->aors;
				ast_log(LOG_DEBUG, "user_exten aors %s", user_exten);
				if (!ast_strlen_zero(user_exten) && strstr(ast_channel_name(requestor), user_exten)) {
					requestor_hangup = 1;
					ast_log(LOG_DEBUG, "Local channel call the Pjsip from meeting %d", requestor_hangup);
				}
			}
	}
	return requestor_hangup;
}

static void set_session_avs_address_according_requestor(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (!session || !requestor) {
		return;
	}

	if (!ast_get_avs_master_value()) {
		return;
	}

	if(session->diff_multiavs_ip)
		strcpy(session->avs_address, "127.0.0.1");
	else
		strcpy(session->avs_address, ast_channel_get_avs_address(requestor));

	ast_log(LOG_DEBUG, "[session:%p] Set avs address:%s according requestor endpoint:%s\n",
		session, session->avs_address, ast_sorcery_object_get_id(session->endpoint));

	return;
}
#endif
/* add by jlhu for trunk hunt end */

/*! \brief Function called by core to create a new outgoing PJSIP session */
static struct ast_channel *chan_pjsip_request_with_stream_topology(const char *type, struct ast_stream_topology *topology, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *data, int *cause)
{
	struct request_data req_data;
#ifdef GRANDSTREAM_NETWORKS
	const char *ucgency_levl;
	const char *ishunt;
	const char *frompage;
	const char *is_reg;
	const char *reg_calleenum;
	const char *reg_callernum;
	const char *reg_callername;
	const char *has_elin;
	int reg_flag = 0;
	struct request_data *gswave_req_data;
#endif
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);

	req_data.topology = topology;
	req_data.dest = data;
	/* Default failure value in case ast_sip_push_task_wait_servant() itself fails. */
	req_data.cause = AST_CAUSE_FAILURE;
#ifdef GRANDSTREAM_NETWORKS
	req_data.requestor = (struct ast_channel *)requestor;
	/*Start added nu zyxia for Bug 292255,2024.5.29 */
	if(assignedids)
		req_data.app_wakeup = assignedids->app_wakeup;
	else
		req_data.app_wakeup = 0;
	/*End added nu zyxia for Bug 292255 */
#endif
	if(requestor){
		const char *mode = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "STRATEGY");
		const  char *aauser = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "AAUSER");
		struct timeval now;
		int dura, flag = 0;
		char  family[80] = {0};
		char  buff[80] = {0};
		/* in ringgroup ,  trunk member need check if can go out */
		if(aauser && mode && !strcasecmp(mode,"SIMULTA") && strstr(data,"@trunk_"))
		{
			if(1 == ast_get_high_frequence_lockflag((char *)aauser))
			{
				ast_log(LOG_WARNING,"user:%s reach the high frequence call,banned!",aauser);
				return NULL;
			}

			snprintf(family, sizeof(family), "AAUSER/%s", aauser);
			ast_ucm_db_get(family, "HIGHCALL_THRESHOLD", buff, sizeof(buff));
			if(buff[0] != 0 && atoi(buff) != 0)
			{
				int call_threshold = atoi(buff);
				memset(buff,0, sizeof(buff));
				ast_ucm_db_get(family, "HIGHCALL_DURATION", buff, sizeof(buff));
				dura = atoi(buff);
				gettimeofday(&now, NULL);
				ast_high_frequence_queue_search((char *)aauser, &now, dura*60, &flag);
				if(1 == flag)
				{
					ast_append_high_frequence_ban((char *)aauser);/* first reach */
					manager_event(EVENT_FLAG_CALL, "HighFreqLimit", "Extension: %s\r\nFlag: %d\r\n\r\n",
								aauser,1);
					ast_report_high_freq_call_alarm((char *)aauser, call_threshold, dura);
					ast_log(LOG_WARNING,"user:%s reach the high frequence call,banned!",aauser);
					return NULL;
				}
			}

		}
	}


	if (ast_sip_push_task_wait_servant(NULL, request, &req_data)) {
		*cause = req_data.cause;
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* add by jlhu for trunk hunt begin */
	ishunt = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "TRIGGERHUNT");
	if (ishunt && (0 == strcmp(ishunt, "yes"))) {
		if (!chan_pjsip_groupfull((struct ast_channel *)requestor, data)) {
			*cause = AST_CAUSE_CONGESTION;
			return NULL;
		}
	}
	/* add by jlhu for trunk hunt end */
#endif

	session = req_data.session;

#ifdef GRANDSTREAM_NETWORKS
	/*Start added by zyxia for sigle announcement ,the var is be set in exe_ucm_announicement_page.py*/
	const char* page_video_support = pbx_builtin_getvar_helper((struct ast_channel *)requestor,"PAGE_VIDEO_SUPPORT");
	if (page_video_support && !strcasecmp(page_video_support,"yes")) {
		ast_log(LOG_DEBUG, "Set channel:%s page video\n", ast_channel_name(requestor));
		session->is_video_page_call = 1;
		ast_channel_set_is_page_video_support((struct ast_channel *)requestor, 1);
	}
	/*End added by zyxia for sigle announcement ,the var is be set in exe_ucm_announicement_page.py*/

	/* Add by bfdu for has no joint caps */
	set_session_sfu_params_according_chan(session, requestor);
	if (!session->pending_media_state->topology) {
		if (chan_pjsip_handle_no_pending_media(session)) {
			return NULL;
		}
	} else {
		if (chan_pjsip_modify_sfu_pending_media(session)) {
			return NULL;
		}
	}


	/*Start added by zyxia for Bug 266995 ,2023.7.28*/
	if(session->is_trunk_conf_invite &&  session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ){
		if(session->reject_call & (0x01 << 1)) {//reject_call set at  set_session_sfu_params_according_chan  by zyxia for Bug 267608
			ast_log(LOG_DEBUG, "[session:%p] requset:%s chan_pjsip_request_with_stream_topology is conf invite from trunk  to webrtc wave and the same cloud im,cancel it, endpoint:%s\n",
				session, ast_channel_name(requestor),ast_sorcery_object_get_id(session->endpoint));
			*cause = AST_CAUSE_CONF_SKIP_CALL_WAVE; //added by zyxia for Bug 267696
			ast_channel_hangupcause_set((struct ast_channel *)requestor, AST_CAUSE_CONF_SKIP_CALL_WAVE);
			return NULL;
		}
	}
	/*End added by zyxia for Bug 266995 ,2023.7.28*/

	if (ast_sip_session_cancel_invite_call(session, requestor,data)) {
		return NULL;
	}

	set_session_avs_address_according_requestor(session, requestor);
#endif

	if (!(session->channel = chan_pjsip_new(session, AST_STATE_DOWN, NULL, NULL, assignedids, requestor, NULL))) {
		/* Session needs to be terminated prematurely */
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->is_video_page_call) {
		ast_log(LOG_DEBUG, "set channel:%s page video\n", ast_channel_name(session->channel));
		ast_channel_set_is_page_video_support(session->channel,1);
	}

	if (requestor) {
		const char *one_way_page = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "ONE_WAY_PAGING");
		if (!ast_strlen_zero(one_way_page) && !strcasecmp(one_way_page, "yes")) {
			ast_log(LOG_DEBUG, "[session:%p] Get peer one way paging flag, set channel:%s flag too, endpoint:%s\n",
				session, ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));
			pbx_builtin_setvar_helper(session->channel, "ONE_WAY_PAGING", "yes");
		}
	}

	/* Added by yfyang, for ssh proxy process */
	if (session->ssh_proxy_flag) {
		pbx_builtin_setvar_helper(session->channel, "SSH_PROXY_FLAG", "yes");
	}

	/* Add by bfdu for SFU INVITE */
	if (session->is_sfu) {
		if (requestor) {
			const char *sfu_call_id = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "SFU_CALL_ID");
			if (!ast_strlen_zero(sfu_call_id)) {
				pbx_builtin_setvar_helper(session->channel, "SFU_CALL_ID", sfu_call_id);
			}
		}

		chan_pjsip_modify_sfu_stream_params(session);
	}

	/* Added by yfyang, for sca feature */
	if (session->endpoint->sca_enable && (session->line_index <= 0)) {
		if (check_meeting_invitation(session, requestor)) {
			session->line_index = ast_sip_sca_line_alloc((char *)session->endpoint->sca_sharedline, (char *)ast_sorcery_object_get_id(session->endpoint), NULL);
		} else {
			session->line_index = ast_sip_sca_line_alloc((char *)session->endpoint->sca_sharedline, (char *)ast_sorcery_object_get_id(session->endpoint), (struct ast_channel *)requestor);
		}
	}

	/* Added by yfyang, for the feature "Cancel others if one terminal reject" */
	if (session->endpoint->cancel_others_if_one_reject) {
		char endpoint_id[128] = {0};
		snprintf(endpoint_id, sizeof(endpoint_id)-1, "%s", ast_sorcery_object_get_id(session->endpoint));
		pbx_builtin_setvar_helper(session->channel, "CANCEL_OTHERS_IF_ONE_REJECT", endpoint_id);
	}

	if (requestor) {
		/*Start  added by zyxia for Bug186503,2021.1.20*/
		const char *send_fax_file_path = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "filepath");
		if (send_fax_file_path) {
			session->is_fax_sending_request = 1;
		} else {
			session->is_fax_sending_request = 0;
		}
		/*End  added by zyxia for Bug186503,2021.1.20*/

		/* Add by bfdu for blind transfer callback,  */
		const char *requestor_num = ast_channel_caller_num(session->channel);
		const char *blind_transfer_callback_num = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "BLIND_TRANSFER_CALLBACK_NUM");
		if (requestor_num && blind_transfer_callback_num && !strcmp(requestor_num, blind_transfer_callback_num)) {
			ast_log(LOG_DEBUG, "Set BLIND_TRANSFER_CALLBACK_NUM as NULL, only need callback for one time\n");
			pbx_builtin_setvar_helper((struct ast_channel *)requestor, "BLIND_TRANSFER_CALLBACK_NUM", NULL);
			/* There is no need to add Referred-By header, clear SIPREFERREDBYHDR value */
			pbx_builtin_setvar_helper((struct ast_channel *)requestor, "SIPREFERREDBYHDR", NULL);
		}
	}

	 /* add by jlhu for trunk hunt begin */
	ucgency_levl = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "URGENCYLEVEL");
	if (ucgency_levl != NULL) {
		ast_channel_trunk_priority_set(session->channel, atoi(ucgency_levl));
	}

	/*if (ishunt) {
		notifyonce = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "NOTIFYONCE");
		if (NULL != notifyonce && 0 == strcmp(notifyonce, "no")) {
			const char *emergency_service_num = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "EMERGENCY_SERVICE_NUM");
			ast_sip_report_emergency_outbound(ast_channel_caller_num(requestor),emergency_service_num);
			pbx_builtin_setvar_helper((struct ast_channel *)requestor, "_NOTIFYONCE", "yes");
		}
	}*/
	/* add by jlhu for trunk hunt end */

	is_reg = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "EMERGENCY_TRUNK_TYPE");
	/*only support peer trunk by now*/
	if (NULL != is_reg && 0 == strcmp(is_reg, "register")){
		reg_flag = 1;
	}
	has_elin = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "HASELIN");
	if (requestor && (NULL != has_elin) && (0 == strcmp(has_elin, "yes")) && (0 == strcmp(ast_channel_tech(requestor)->type, "PJSIP") || 0 == strcmp(ast_channel_tech(requestor)->type, "IAX2"))) {
		ast_channel_set_elin_flag(session->channel, 1);
		if(0 == strcmp(ast_channel_tech(requestor)->type, "IAX2"))
		{
			struct in_addr addr;
			struct in6_addr addr6;
			char *start,*end;
			char addr6_str[256] = {0};
			const char *iax_addr = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "IAX_TERMINAL_ADDR");
			if(iax_addr && !ast_strlen_zero(iax_addr))
			{
				if(strchr(iax_addr,'['))
				{
					ast_channel_set_elin_ip_ver(session->channel,AF_INET6);
					start = (char *)iax_addr+1;
					end = strchr(start, ']');
					strncpy(addr6_str, start, end-start);
					inet_pton(AF_INET6, addr6_str, (void *)(&addr6));
					ast_channel_set_elin_ip6(session->channel, &addr6);
				}
				else
				{
					ast_channel_set_elin_ip_ver(session->channel,AF_INET);
					inet_pton(AF_INET, iax_addr, (void *)(&addr));
					ast_channel_set_elin_ip(session->channel, &addr);
				}
			}
		}
		else
		{
			set_peer_src_addr(session, requestor);
		}
		ast_channel_set_elin_dialer(session->channel, ast_channel_caller_num(requestor));
		if(reg_flag)
		{
			pbx_builtin_setvar_helper((struct ast_channel *)session->channel, "ELIN_ON_REGTRUNK", "yes");
		}

	} else {
		ast_channel_set_elin_flag(session->channel, 0);
	}

	frompage = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "PAGING_ALERT_INFO");
	if (frompage != NULL) {
		ast_channel_set_pagecall_flag(session->channel, 1);
	}

	/*edit by zyxia for bug113812 peer sip trunk ,active call,2018.6.27*/
	if (NULL != is_reg && (0 == strcmp(is_reg, "register") || 0 == strcmp(is_reg,"peer"))){
		reg_calleenum = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "REG_CALLEENUM");
		if (NULL != reg_calleenum){
			ast_channel_reg_calleenum_set(session->channel, reg_calleenum);
		}

		reg_callernum = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "REG_CALLERNUM");
		if (NULL != reg_callernum){
			ast_channel_reg_callernum_set(session->channel, reg_callernum);
		}

		reg_callername = pbx_builtin_getvar_helper((struct ast_channel *)requestor, "REG_CALELRNAME");
		if (NULL != reg_callername){
			ast_channel_reg_callername_set(session->channel, reg_callername);
		}
	}

	/*BEGIN: Added by yfyang, for Bug 155444 */
	if (requestor && !strncmp(ast_channel_name(requestor), "PJSIP/trunk_", 12) &&
		!ast_strlen_zero(pbx_builtin_getvar_helper((struct ast_channel *)requestor, "CF_ALL"))) {

		reg_calleenum = ast_channel_reg_calleenum(requestor);
		if (!ast_strlen_zero(reg_calleenum)) {
			ast_log(LOG_NOTICE, "set reg_calleenum[%s] to chan[%s], requestor[%s]", reg_calleenum, ast_channel_name(session->channel), ast_channel_name(requestor));
			ast_channel_reg_calleenum_set(session->channel, reg_calleenum);
		}
	}
	/*END: Added by yfyang, for Bug 155444 */

	ast_channel_set_parking(session->channel, ast_channel_get_parking(requestor));
#if 0   /* Removed by yfyang, for Bug 168224 */
	ast_channel_set_parking((struct ast_channel*)requestor, 0);
#endif
	const char* tmp_alert_info = pbx_builtin_getvar_helper((struct ast_channel*)requestor, "PARKING_LOT_ALERT_INFO");
	if (!ast_strlen_zero(tmp_alert_info)) {
		pbx_builtin_setvar_helper(session->channel, "PARKING_LOT_ALERT_INFO", tmp_alert_info);
		ast_log(LOG_DEBUG, "[chan_pjsip_request] channel: %s, set parking lot timeout ring back call Alert-Info header.\n", ast_channel_name(session->channel));
	}

	get_peer_param_for_init_invite(session, requestor);

	ast_channel_set_avs_address(session->channel, session->avs_address);

	/*Start added by zyxia for Bug 208397,2021.12.1,dial ipc rtsp, and transfer the call to sip,the caller will be rtsp channel*/
	if (ast_sip_session_judge_if_type_is_rtsp(requestor)) {
		ast_log(LOG_WARNING, "[session %p] caller %s is  a rtsp channel \n" ,session,ast_channel_name(requestor));
		session->is_rtsp_call = 1;
	}
	/*End added by zyxia for Bug 208397,2021.12.1 */

	/*Start added by zyxia for gswave app push*/
	if (session->terminal_type ==  AST_TERMINAL_TYPE_GSWAVE_MOBILE
		&& !ast_channel_get_reject_call((struct ast_channel *)requestor)
		&& session->channel  &&  (assignedids == NULL || (assignedids && !assignedids->app_wakeup))){
		gswave_req_data=(struct request_data*)ast_calloc(1,sizeof(struct request_data));
		gswave_req_data->topology =  ao2_bump(topology);
		gswave_req_data->dest = ast_strdup(data);
		gswave_req_data->session = ao2_bump((struct ast_sip_session *)session);
		gswave_req_data->requestor =  ao2_bump((struct ast_channel *)requestor);
		gswave_req_data->old_channel = ao2_bump((struct ast_channel *)session->channel);
		pthread_t threadid;
		if (ast_pthread_create_detached(&threadid, NULL, ast_call_gswave_mobile_app, gswave_req_data)) {
			ao2_cleanup(gswave_req_data->topology);
			ao2_cleanup(gswave_req_data->requestor);
			ao2_cleanup(gswave_req_data->old_channel);
			ao2_cleanup(gswave_req_data->session);
			ast_free((char*)gswave_req_data->dest);
			ast_free(gswave_req_data);
			ast_log(LOG_ERROR, "Error call to the gswave new address\n");
			return session->channel;
		}
	}
	/*End added by zyxia for gswave app push*/
	/* restore original callee channel */
	if(requestor && ast_test_flag(ast_channel_flags((struct ast_channel *)requestor), AST_FLAG_RECOVERYING))
	{
		ast_channel_set_flag(session->channel, AST_FLAG_RECOVERYING);
		struct recovery_db_entry* entry = NULL;
		struct recovery_db_entry* inbound_mode_sub_entry = NULL;

		inbound_mode_sub_entry = ast_recovery_db_get_all();

		for (entry = inbound_mode_sub_entry; entry; entry = entry->next) {
			if(0 == strcmp(entry->runtime_uuid, ast_channel_recovery_linkedid( requestor)))
			{
				if(strcmp(entry->uuid, ast_channel_recovery_uniqueid( requestor)))
				{
					/* it is original callee side */
					char *linkid = ast_malloc(AST_MAX_UNIQUEID);
					char *uniqid = ast_malloc(AST_MAX_UNIQUEID);
					strncpy(linkid, entry->runtime_uuid, AST_MAX_UNIQUEID);
					strncpy(uniqid, entry->uuid, AST_MAX_UNIQUEID);
					ast_channel_set_recovery_linkedid(session->channel, linkid);
					ast_channel_set_recovery_uniqueid(session->channel, uniqid);
					switch_xml_t restore_xml,times, sec ,cseq,vars,connect_line;
					struct timeval restore_time;
					int last_cseq;
					struct ast_party_connected_line connected;
					restore_xml = switch_xml_parse_str_dynamic(entry->data, SWITCH_TRUE);
					times = switch_xml_child(restore_xml, "times");

					sec = switch_xml_child(times, "create_sec");
					restore_time.tv_sec = atoi(sec->txt);
					sec = switch_xml_child(times, "create_usec");
					restore_time.tv_usec = atoi(sec->txt);
					ast_channel_set_recovery_createtime(session->channel, &restore_time);

					sec = switch_xml_child(times, "answer_sec");
					restore_time.tv_sec = atoi(sec->txt);
					sec = switch_xml_child(times, "answer_usec");
					restore_time.tv_usec = atoi(sec->txt);
					ast_channel_set_recovery_answertime(session->channel, &restore_time);

					vars = switch_xml_child(restore_xml, "variables");
					cseq = switch_xml_child(vars, "cseq");
					last_cseq = atoi(cseq->txt);
					connect_line = switch_xml_child(vars, "connect_line_id");

					if(connect_line && !ast_strlen_zero(connect_line->txt)){
						ast_party_connected_line_set_init(&connected, ast_channel_connected(session->channel));
						connected.id.number.valid = TRUE;
						connected.id.number.str = ast_strdup(connect_line->txt);
						ast_party_connected_line_copy(ast_channel_connected(session->channel), &connected);
						ast_free(connected.id.number.str);
					}

					set_channel_media_encryption_for_recovery(session, restore_xml);

					pjsip_dlg_inc_lock(session->inv_session->dlg);
					session->inv_session->dlg->local.first_cseq = last_cseq+10;
					session->inv_session->dlg->local.cseq = last_cseq+10;
					pjsip_dlg_dec_lock(session->inv_session->dlg);

					if(restore_xml)
					{
						switch_xml_free(restore_xml);
					}
					break;

				}
			}
		}

		if (inbound_mode_sub_entry) {
			ast_recovery_freetree(inbound_mode_sub_entry);
		}
	}
#endif

	return session->channel;
}

static struct ast_channel *chan_pjsip_request(const char *type, struct ast_format_cap *cap, const struct ast_assigned_ids *assignedids, const struct ast_channel *requestor, const char *data, int *cause)
{
	struct ast_stream_topology *topology;
	struct ast_channel *chan;

	topology = ast_stream_topology_create_from_format_cap(cap);
	if (!topology) {
		return NULL;
	}

	chan = chan_pjsip_request_with_stream_topology(type, topology, assignedids, requestor, data, cause);

	ast_stream_topology_free(topology);

	return chan;
}

struct sendtext_data {
	struct ast_sip_session *session;
	struct ast_msg_data *msg;
};

static void sendtext_data_destroy(void *obj)
{
	struct sendtext_data *data = obj;
	ao2_cleanup(data->session);
	ast_free(data->msg);
}

static struct sendtext_data* sendtext_data_create(struct ast_channel *chan,
	struct ast_msg_data *msg)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct sendtext_data *data = ao2_alloc(sizeof(*data), sendtext_data_destroy);

	if (!data) {
		return NULL;
	}

	data->msg = ast_msg_data_dup(msg);
	if (!data->msg) {
		ao2_cleanup(data);
		return NULL;
	}
	data->session = channel->session;
	ao2_ref(data->session, +1);

	return data;
}

static int sendtext(void *obj)
{
	struct sendtext_data *data = obj;
	pjsip_tx_data *tdata;
	const char *body_text = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_BODY);
	const char *content_type = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_CONTENT_TYPE);
	char *sep;
	struct ast_sip_body body = {
		.type = "text",
		.subtype = "plain",
		.body_text = body_text,
	};

	if (!ast_strlen_zero(content_type)) {
		sep = strchr(content_type, '/');
		if (sep) {
			*sep = '\0';
			body.type = content_type;
			body.subtype = ++sep;
		}
	}

	if (data->session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			data->session->inv_session->cause,
			pjsip_get_status_text(data->session->inv_session->cause)->ptr);
	} else {
		pjsip_from_hdr *hdr;
		pjsip_name_addr *name_addr;
		const char *from = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_FROM);
		const char *to = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_TO);
		int invalidate_tdata = 0;

#ifdef GRANDSTREAM_NETWORKS
		if (!strncasecmp(body.body_text, "Event: x-gs-schedule-conference", strlen("Event: x-gs-schedule-conference"))) {
			if (!strncasecmp(ast_sorcery_object_get_id(data->session->endpoint), "trunk_", strlen("trunk_"))) {
				goto out;
			}

			pjsip_event_hdr *event = NULL;

			ast_sip_create_request("NOTIFY", data->session->inv_session->dlg, data->session->endpoint, NULL, NULL, &tdata);
			event = pjsip_event_hdr_create(tdata->pool);
			pj_cstr(&event->event_type, "x-gs-schedule-conference");
			pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *) event);
			body.body_text += strlen("Event: x-gs-schedule-conference\r\n");
		} else {
			ast_sip_create_request("MESSAGE", data->session->inv_session->dlg, data->session->endpoint, NULL, NULL, &tdata);
		}
#else
		ast_sip_create_request("MESSAGE", data->session->inv_session->dlg, data->session->endpoint, NULL, NULL, &tdata);
#endif
		ast_sip_add_body(tdata, &body);

		/*
		 * If we have a 'from' in the msg, set the display name in the From
		 * header to it.
		 */
		if (!ast_strlen_zero(from)) {
			hdr = PJSIP_MSG_FROM_HDR(tdata->msg);
			name_addr = (pjsip_name_addr *) hdr->uri;
			pj_strdup2(tdata->pool, &name_addr->display, from);
			invalidate_tdata = 1;
		}

		/*
		 * If we have a 'to' in the msg, set the display name in the To
		 * header to it.
		 */
		if (!ast_strlen_zero(to)) {
			hdr = PJSIP_MSG_TO_HDR(tdata->msg);
			name_addr = (pjsip_name_addr *) hdr->uri;
			pj_strdup2(tdata->pool, &name_addr->display, to);
			invalidate_tdata = 1;
		}

		if (invalidate_tdata) {
			pjsip_tx_data_invalidate_msg(tdata);
		}
#ifdef BUILD_ON_CLOUD_PBX
		ast_sip_add_sbc_endpoint_address_header(data->session->sbc_endpoint_addr, tdata);
#endif
		ast_sip_send_request(tdata, data->session->inv_session->dlg, data->session->endpoint, NULL, NULL);
	}

#ifdef GRANDSTREAM_NETWORKS
out:
#endif

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(data->session->inv_session);
#endif

	ao2_cleanup(data);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int send_callcontrol(void *obj)
{
	RAII_VAR(struct sendtext_data *, data, obj, ao2_cleanup);
	pjsip_tx_data   *tdata;
	const char *body_text = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_BODY);

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-grandstream-slides-video-control+xml",
		.body_text = body_text + strlen("INFO gs-callcontrol:")
	};

	ast_log(LOG_NOTICE, "Sending in dialog SIP callcontrol INFO\n");

	ast_sip_create_request("INFO", data->session->inv_session->dlg, data->session->endpoint, NULL, NULL, &tdata);

	ast_sip_add_body(tdata, &body);

#ifdef BUILD_ON_CLOUD_PBX
	ast_sip_add_sbc_endpoint_address_header(data->session->sbc_endpoint_addr, tdata);
#endif
	ast_sip_send_request(tdata, data->session->inv_session->dlg, data->session->endpoint, NULL, NULL);

	return 0;
}

static int send_notify_for_mobile_app(void *obj)
{
	RAII_VAR(struct sendtext_data *, data, obj, ao2_cleanup);
	pjsip_tx_data *tdata = NULL;
	const char *body_text = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_BODY);

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "conference-information",
		.body_text = body_text + strlen("NOTIFY mobile-app-conf-info:")
	};

	ast_log(LOG_DEBUG, "[session:%p] Sending in dialog SIP NOTIFY for mobile app, endpoint:%s\n",
		data->session, ast_sorcery_object_get_id(data->session->endpoint));

	ast_sip_create_request("NOTIFY", data->session->inv_session->dlg, data->session->endpoint, NULL, NULL, &tdata);

	ast_sip_add_body(tdata, &body);
#ifdef BUILD_ON_CLOUD_PBX
	ast_sip_add_sbc_endpoint_address_header(data->session->sbc_endpoint_addr, tdata);
#endif
	ast_sip_send_request(tdata, data->session->inv_session->dlg, data->session->endpoint, NULL, NULL);

	return 0;
}

static unsigned int judge_if_need_send_notify(struct ast_sip_session *session)
{
	if (!session) {
		return FALSE;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE &&
		strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_DEBUG, "[session:%p] judge_if_need_send_notify endpoint:%s is not gswave mobile or trunk\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	return TRUE;
}
#endif

/*! \brief Function called by core to send text on PJSIP session */
static int chan_pjsip_sendtext_data(struct ast_channel *ast, struct ast_msg_data *msg)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(ast);
	struct sendtext_data *data = sendtext_data_create(ast, msg);

#ifdef GRANDSTREAM_NETWORKS
	const char *body_text = ast_msg_data_get_attribute(data->msg, AST_MSG_DATA_ATTR_BODY);
	if (!strncmp(body_text, "INFO gs-callcontrol:", strlen("INFO gs-callcontrol:"))) {
		ast_log(LOG_DEBUG, "It's INFO gs-callcontrol\n");
		if (!data || ast_sip_push_task(channel->session->serializer, send_callcontrol, data)) {
			ao2_ref(data, -1);
			return -1;
		}
		return 0;
	}

	if (!strncmp(body_text, "NOTIFY mobile-app-conf-info:", strlen("NOTIFY mobile-app-conf-info:"))) {
		if (judge_if_need_send_notify(channel->session)) {
			if (!data || ast_sip_push_task(channel->session->serializer, send_notify_for_mobile_app, data)) {
				ao2_ref(data, -1);
				return -1;
			}
		} else {
			ao2_ref(data, -1);
		}
		return 0;
	}
#endif

	ast_debug(1, "Sending MESSAGE from '%s' to '%s:%s': %s\n",
		ast_msg_data_get_attribute(msg, AST_MSG_DATA_ATTR_FROM),
		ast_msg_data_get_attribute(msg, AST_MSG_DATA_ATTR_TO),
		ast_channel_name(ast),
		ast_msg_data_get_attribute(msg, AST_MSG_DATA_ATTR_BODY));

	if (!data) {
		return -1;
	}

#ifdef HAVE_PJSIP_INV_SESSION_REF
	if (pjsip_inv_add_ref(data->session->inv_session) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
		ao2_ref(data, -1);
		return -1;
	}
#endif

	if (ast_sip_push_task(channel->session->serializer, sendtext, data)) {
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(data->session->inv_session);
#endif
		ao2_ref(data, -1);
		return -1;
	}
	return 0;
}

static int chan_pjsip_sendtext(struct ast_channel *ast, const char *text)
{
	struct ast_msg_data *msg;
	int rc;
	struct ast_msg_data_attribute attrs[] =
	{
		{
			.type = AST_MSG_DATA_ATTR_BODY,
			.value = (char *)text,
		}
	};

	msg = ast_msg_data_alloc(AST_MSG_DATA_SOURCE_TYPE_UNKNOWN, attrs, ARRAY_LEN(attrs));
	if (!msg) {
		return -1;
	}
	rc = chan_pjsip_sendtext_data(ast, msg);
	ast_free(msg);

	return rc;
}

/*! \brief Convert SIP hangup causes to Asterisk hangup causes */
static int hangup_sip2cause(int cause)
{
	/* Possible values taken from causes.h */

	switch(cause) {
	case 401:       /* Unauthorized */
		return AST_CAUSE_CALL_REJECTED;
	case 403:       /* Not found */
		return AST_CAUSE_CALL_REJECTED;
	case 404:       /* Not found */
		return AST_CAUSE_UNALLOCATED;
	case 405:       /* Method not allowed */
		return AST_CAUSE_INTERWORKING;
	case 407:       /* Proxy authentication required */
		return AST_CAUSE_CALL_REJECTED;
	case 408:       /* No reaction */
		return AST_CAUSE_NO_USER_RESPONSE;
	case 409:       /* Conflict */
		return AST_CAUSE_NORMAL_TEMPORARY_FAILURE;
	case 410:       /* Gone */
		return AST_CAUSE_NUMBER_CHANGED;
	case 411:       /* Length required */
		return AST_CAUSE_INTERWORKING;
	case 413:       /* Request entity too large */
		return AST_CAUSE_INTERWORKING;
	case 414:       /* Request URI too large */
		return AST_CAUSE_INTERWORKING;
	case 415:       /* Unsupported media type */
		return AST_CAUSE_INTERWORKING;
	case 420:       /* Bad extension */
		return AST_CAUSE_NO_ROUTE_DESTINATION;
	case 480:       /* No answer */
		return AST_CAUSE_NO_ANSWER;
	case 481:       /* No answer */
		return AST_CAUSE_INTERWORKING;
	case 482:       /* Loop detected */
		return AST_CAUSE_INTERWORKING;
	case 483:       /* Too many hops */
		return AST_CAUSE_NO_ANSWER;
	case 484:       /* Address incomplete */
		return AST_CAUSE_INVALID_NUMBER_FORMAT;
	case 485:       /* Ambiguous */
		return AST_CAUSE_UNALLOCATED;
	case 486:       /* Busy everywhere */
		return AST_CAUSE_BUSY;
	case 487:       /* Request terminated */
		return AST_CAUSE_INTERWORKING;
	case 488:       /* No codecs approved */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	case 491:       /* Request pending */
		return AST_CAUSE_INTERWORKING;
	case 493:       /* Undecipherable */
		return AST_CAUSE_INTERWORKING;
	case 500:       /* Server internal failure */
		return AST_CAUSE_FAILURE;
	case 501:       /* Call rejected */
		return AST_CAUSE_FACILITY_REJECTED;
	case 502:
		return AST_CAUSE_DESTINATION_OUT_OF_ORDER;
	case 503:       /* Service unavailable */
		return AST_CAUSE_CONGESTION;
	case 504:       /* Gateway timeout */
		return AST_CAUSE_RECOVERY_ON_TIMER_EXPIRE;
	case 505:       /* SIP version not supported */
		return AST_CAUSE_INTERWORKING;
	case 600:       /* Busy everywhere */
		return AST_CAUSE_USER_BUSY;
	case 603:       /* Decline */
		return AST_CAUSE_CALL_REJECTED;
	case 604:       /* Does not exist anywhere */
		return AST_CAUSE_UNALLOCATED;
	case 606:       /* Not acceptable */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	default:
		if (cause < 500 && cause >= 400) {
			/* 4xx class error that is unknown - someting wrong with our request */
			return AST_CAUSE_INTERWORKING;
		} else if (cause < 600 && cause >= 500) {
			/* 5xx class error - problem in the remote end */
			return AST_CAUSE_CONGESTION;
		} else if (cause < 700 && cause >= 600) {
			/* 6xx - global errors in the 4xx class */
			return AST_CAUSE_INTERWORKING;
		}
		return AST_CAUSE_NORMAL;
	}
	/* Never reached */
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS

/* Added for yfyang, for Broadsoft Test */
static int get_pai_payload(struct ast_channel *channel, struct pjsip_rx_data *rdata, struct ast_sip_pai_payload *payload)
{
	static const pj_str_t pai_str = { "P-Asserted-Identity", 19 };
	static const pj_str_t privacy_str = { "Privacy", 7 };
	static const pj_str_t from = { "From", 4 };
	struct ast_party_id  pai_id;
	pjsip_generic_string_hdr *privacy = NULL;
	pjsip_generic_string_hdr *ident = NULL;
	pjsip_fromto_hdr *pai_hdr = NULL;
	pj_str_t pai_content;
	pjsip_name_addr *id_name_addr = NULL;
	pjsip_sip_uri   *uri;
	int parsed_len;

	if (!channel || !rdata || !payload) {
		return -1;
	}

	/* get PAI header string */
	ident = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &pai_str, NULL);
	if (!ident) {
		return -1;
	}

	/* transform pai string to uri */
	pj_strdup_with_null(rdata->tp_info.pool, &pai_content, &ident->hvalue);
	pai_hdr = pjsip_parse_hdr(rdata->tp_info.pool, &from, pai_content.ptr, pj_strlen(&pai_content), &parsed_len);
	if (!pai_hdr) {
		return -1;
	}

	/* get id number and name */
	ast_party_id_init(&pai_id);
	id_name_addr = (pjsip_name_addr *) pai_hdr->uri;
	uri = pjsip_uri_get_uri(id_name_addr);

	if (id_name_addr->display.slen) {
		pai_id.name.str = ast_strndup(id_name_addr->display.ptr, id_name_addr->display.slen);
		if (!ast_strlen_zero(pai_id.name.str)) {
			pai_id.name.valid = 1;
		}
	}

	pai_id.number.str = ast_strndup(uri->user.ptr, uri->user.slen);
	if (!ast_strlen_zero(pai_id.number.str)) {
		pai_id.number.valid = 1;
	}

	if (!pai_id.number.valid) {
		ast_party_id_free(&pai_id);
		return -1;
	}

	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) {
		payload->is_response = 0;
		payload->code = 0;

		if (rdata->msg_info.msg->line.req.method.id == PJSIP_INVITE_METHOD) {
			strncpy(payload->method, "INVITE", sizeof(payload->method)-1);
		} else if (rdata->msg_info.msg->line.req.method.id == PJSIP_OTHER_METHOD) {
			/* Only process PAI in UPDATE and INVITE here */
			if (strncmp(rdata->msg_info.msg->line.req.method.name.ptr, "UPDATE", 6)) {
				ast_party_id_free(&pai_id);
				return -1;
			}

			strncpy(payload->method, "UPDATE", sizeof(payload->method)-1);
		}
	} else {
		payload->is_response = 1;
		payload->code = rdata->msg_info.msg->line.status.code;
		strncpy(payload->method, "INVITE", sizeof(payload->method)-1);  /* only process response for INVITE */
	}

	privacy = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &privacy_str, NULL);
	if (privacy && pj_stricmp2(&privacy->hvalue, "none")) {
		payload->is_blocked = 1;
	} else {
		payload->is_blocked = 0;
	}

	snprintf(payload->pai_string, sizeof(payload->pai_string), "%s<%s>", pai_id.name.valid ? pai_id.name.str : "", pai_id.number.str);

	ast_party_id_free(&pai_id);
	return 0;
}


static void set_pai_payload(struct ast_channel *channel, const void *data, size_t datalen)
{
	struct ast_sip_pai_payload *payload = (struct ast_sip_pai_payload *)data;

	char buffer[512] = {0};

	if (!channel || !data || !datalen) {
		return;
	}

	if (sizeof(struct ast_sip_pai_payload) != datalen) {
		return;
	}

	snprintf(buffer, sizeof(buffer)-1, "%d,%d,%d,%s,%s", payload->is_blocked, payload->is_response, payload->code,
	                                                     payload->method, payload->pai_string);

	ast_log(LOG_NOTICE, "PEER_PAI_INFO: %s\n", buffer);
	pbx_builtin_setvar_helper(channel, "PEER_PAI_INFO", buffer);
}


/* Added for Bug75836 */
static void set_caller_pai_string(struct ast_channel *channel, struct pjsip_rx_data *rdata)
{
	static const pj_str_t pai_str = { "P-Asserted-Identity", 19 };
	static const pj_str_t privacy_str = { "Privacy", 7 };
	static const pj_str_t from = { "From", 4 };
	struct ast_party_id  pai_id;
	pjsip_generic_string_hdr *privacy = NULL;
	pjsip_generic_string_hdr *ident = NULL;
	pjsip_fromto_hdr *pai_hdr = NULL;
	pj_str_t pai_content;
	pjsip_name_addr *id_name_addr = NULL;
	pjsip_sip_uri   *uri;
	int parsed_len;

	if (!channel || !rdata) {
		return;
	}

	/* get PAI header string */
	ident = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &pai_str, NULL);
	if (!ident) {
		return;
	}

	/* transform pai string to uri */
	pj_strdup_with_null(rdata->tp_info.pool, &pai_content, &ident->hvalue);
	pai_hdr = pjsip_parse_hdr(rdata->tp_info.pool, &from, pai_content.ptr, pj_strlen(&pai_content), &parsed_len);
	if (!pai_hdr) {
		return;
	}

	/* get id number and name */
	ast_party_id_init(&pai_id);
	id_name_addr = (pjsip_name_addr *) pai_hdr->uri;
	uri = pjsip_uri_get_uri(id_name_addr);

	if (id_name_addr->display.slen) {
		pai_id.name.str = ast_strndup(id_name_addr->display.ptr, id_name_addr->display.slen);
		if (!ast_strlen_zero(pai_id.name.str)) {
			pai_id.name.valid = 1;
		}
	}

	pai_id.number.str = ast_strndup(uri->user.ptr, uri->user.slen);
	if (!ast_strlen_zero(pai_id.number.str)) {
		pai_id.number.valid = 1;
	}

	if (!pai_id.number.valid) {
		ast_party_id_free(&pai_id);
		return;
	}

	privacy = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &privacy_str, NULL);
	if (privacy && pj_stricmp2(&privacy->hvalue, "none")) {
		pai_id.number.presentation = AST_PRES_PROHIB_USER_NUMBER_NOT_SCREENED;
		pai_id.name.presentation = AST_PRES_PROHIB_USER_NUMBER_NOT_SCREENED;
	} else {
		pai_id.number.presentation = AST_PRES_ALLOWED_USER_NUMBER_NOT_SCREENED;
		pai_id.name.presentation = AST_PRES_ALLOWED_USER_NUMBER_NOT_SCREENED;
	}

	if (pai_id.number.presentation == AST_PRES_ALLOWED_USER_NUMBER_NOT_SCREENED) {
		char pai_caller[128] = {0};

		snprintf(pai_caller, sizeof(pai_caller), "%s<%s>", pai_id.name.valid ? pai_id.name.str : "", pai_id.number.str);
		pbx_builtin_setvar_helper(channel, "__CALLER_PAI", pai_caller);
	}

	ast_party_id_free(&pai_id);
}

static void set_caller_dial_string(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_contact_hdr *contact_hdr = NULL;
	pjsip_sip_uri *sip_contact;
	char contact_uri[PJSIP_MAX_URL_SIZE];
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	char caller_dial_string[256] = {0};

	if (!session || !rdata) {
		return;
	}

	contact_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);

	if (contact_hdr != NULL && contact_hdr->uri != NULL) {
		sip_contact = pjsip_uri_get_uri(contact_hdr->uri);
		pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, sip_contact, contact_uri, sizeof(contact_uri));
	}

	if (!session->endpoint) {
		return;
	}

	aor = ast_sip_location_retrieve_aor(session->endpoint->aors);
	if (!aor) {
		return;
	}

	//if (aor->max_contacts > 1) {
#ifdef BUILD_ON_CLOUD_PBX
		if (!ast_strlen_zero(session->sbc_endpoint_addr)) {
			char tmp_uri[128] = {0};
			snprintf(tmp_uri, sizeof(tmp_uri), "sip:%s@%s:%d", session->endpoint->aors, rdata->pkt_info.src_name, rdata->pkt_info.src_port);
			snprintf(caller_dial_string, sizeof(caller_dial_string), "PJSIP/%s/%s;sbc-dest=%s", session->endpoint->aors, tmp_uri, session->sbc_endpoint_addr);
		} else {
			snprintf(caller_dial_string, sizeof(caller_dial_string), "PJSIP/%s/%s", session->endpoint->aors, contact_uri);
		}
#else
		snprintf(caller_dial_string, sizeof(caller_dial_string), "PJSIP/%s/%s", session->endpoint->aors, contact_uri);
#endif
		pbx_builtin_setvar_helper(session->channel, "__CALLER_DIAL_STRING", caller_dial_string);
		ast_log(LOG_DEBUG, "set_caller_dial_string %s\n", caller_dial_string);
	//}
}

static void set_call_type(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr   *hdr;
	pj_str_t calltype = {"X-Grandstream-Call-Type", 23};
	char calltype_str[64] = {0};

	if (!session || !session->channel || !rdata) {
		return;
	}

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &calltype, NULL))) {
		return;
	}

	if (0 == hdr->hvalue.slen) {
		return;
	}

	snprintf(calltype_str, sizeof(calltype_str)-1, "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);
	pbx_builtin_setvar_helper(session->channel, "__CALL_TYPE", calltype_str);

	sip_setcalltype(session->channel, calltype_str);

	ast_log(LOG_NOTICE, "[session:%p] set_call_type: %s for endpoint:%s\n",
		session, calltype_str, ast_sorcery_object_get_id(session->endpoint));
}


/* Added by yfyang, for IPVT Server application */
static void set_ipvt_client_info_string(struct ast_channel *channel, struct pjsip_rx_data *rdata, int is_response)
{
	static const pj_str_t clientinfo_str = { "X-IPVT-Client-Info", 18};
	pjsip_generic_string_hdr   *hdr;
	char clientinfo_rec[256] = {0};

	if (!channel || !rdata) {
		return;
	}

	/* get X-IPVT-Client-Info header string */
	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &clientinfo_str, NULL);
	if (!hdr) {
		return;
	}

	snprintf(clientinfo_rec, sizeof(clientinfo_rec)-1, "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);
	pbx_builtin_setvar_helper(channel, "__IPVT_CLIENT_INFO", clientinfo_rec);

	if (is_response) {
		pbx_builtin_setvar_helper(channel, "__IPVT_CLIENT_INFO_MSG", "response");
	} else {
		pbx_builtin_setvar_helper(channel, "__IPVT_CLIENT_INFO_MSG", "request");
	}
}

/* Added by yfyang, for Bug 232651, support P-Called-Party-ID */
static void set_pcpid_string(struct ast_channel *channel, struct pjsip_rx_data *rdata)
{
	pjsip_uri      *ruri;
	char pcpid[256]    = {0};
	char uri_info[256] = {0};

	if (!channel || !rdata) {
		return;
	}

	/* get request-url string */
	ruri = rdata->msg_info.msg->line.req.uri;

	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri) && !PJSIP_URI_SCHEME_IS_TEL(ruri)) {
		return;
	}

	pjsip_uri_print(PJSIP_URI_IN_REQ_URI, ruri, uri_info, sizeof(uri_info));

	snprintf(pcpid, sizeof(pcpid)-1, "<%s>", uri_info);
	pbx_builtin_setvar_helper(channel, "__PEER_PCPID_INFO", pcpid);

	ast_log(LOG_DEBUG, "pcpid string: %s", pcpid);
	return;
}

/* Added by bfdu for Bug 211171, 2021.12.23 */
static void set_peer_enter_voicemail(struct ast_channel *channel, struct pjsip_rx_data *rdata)
{
	static const pj_str_t vm_str = { "GS-voicemail", 12};
	pjsip_generic_string_hdr *hdr;

	if (!channel || !rdata) {
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &vm_str, NULL);
	if (!hdr) {
		return;
	}

	ast_log(LOG_DEBUG, "set_peer_enter_voicemail for channel:%s\n", ast_channel_name(channel));

	pbx_builtin_setvar_helper(channel, "__PEER_ENTER_VM", "yes");

	return;
}

/* Add by bfdu for Bug 78568, 2017.02.08 */
static int transmit_info_with_crm_update(void *data)
{
	char *text = NULL;
	int text_len = 0;

	ast_log(LOG_DEBUG, "transmit_info_with_crm_update begin\n");

	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);
	struct pjsip_tx_data *tdata;

	text_len = sizeof(session->crm_info) + 48;

	text = ast_malloc(text_len);
	if (!text) {
		ast_log(LOG_ERROR, "Could not malloc CRM info text\n");
		return -1;
	}

	memset(text, 0, text_len);

	if (strlen(session->crm_info.caller_name)) {
		strcat(text, "caller-name=");
		strcat(text, session->crm_info.caller_name);
		strcat(text, "\r\n");
	}

	if (strlen(session->crm_info.caller_number)) {
		strcat(text, "caller-number=");
		strcat(text, session->crm_info.caller_number);
		strcat(text, "\r\n");
	}

	if (strlen(session->crm_info.company)) {
		strcat(text, "company=");
		strcat(text, session->crm_info.company);
		strcat(text, "\r\n");
	}

	ast_log(LOG_DEBUG, "transmit_info_with_crm_update text:%s\n", text);

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "crm-info",
		.body_text = text
	};

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text CRM update INFO request\n");
		ast_free(text);
		return -1;
	}
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to text CRM update INFO request\n");
		ast_free(text);
		return -1;
	}
	ast_sip_session_send_request(session, tdata);

	ast_free(text);

	return 0;
}

static int ami_push_crm_info(struct mansession *s, const struct message *m)
{
	struct ast_channel *c = NULL;
	const char *channel_name = astman_get_header(m, "Channel");
	const char *caller_name = astman_get_header(m, "CallerName");
	const char *caller_number = astman_get_header(m, "CallerNumber");
	const char *company = astman_get_header(m, "Company");

	ast_log(LOG_DEBUG, "Get CRM info: channel_name:%s, caller_name:%s, caller_number:%s, company:%s\n", channel_name, caller_name, caller_number, company);

	if (!(c = ast_channel_get_by_name(channel_name))) {
		ast_log(LOG_WARNING, "ami_push_crm_info can't get channel by channel_name\n");
		astman_send_error(s, m, "No such channel");
		return AMI_SUCCESS;
	}

	if (0 != strcmp(ast_channel_tech(c)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "ami_push_crm_info channel type is not pjsip\n");
		c = ast_channel_unref(c);
		astman_send_error(s, m, "Can't send CRM info as not a PJSIP channel");
		return AMI_SUCCESS;
	}

	struct ast_sip_channel_pvt *chan_pvt;
	chan_pvt = ast_channel_tech_pvt(c);
	if (!chan_pvt) {
		ast_log(LOG_WARNING, "ami_push_crm_info can't get chan_pvt\n");
		c = ast_channel_unref(c);
		astman_send_error(s, m, "Can't send CRM info as not have a chan_pvt");
		return AMI_SUCCESS;
	}

	if (!chan_pvt->session) {
		ast_log(LOG_WARNING, "ami_push_crm_info can't get session\n");
		c = ast_channel_unref(c);
		astman_send_error(s, m, "Can't send CRM info as not have a session");
		return AMI_SUCCESS;
	}

	if (0 < strlen(caller_name) && strlen(caller_name) < sizeof(chan_pvt->session->crm_info.caller_name)) {
		strcpy(chan_pvt->session->crm_info.caller_name, caller_name);
	} else {
		ast_log(LOG_DEBUG, "AMI don't send CRM caller name\n");
	}

	if (0 < strlen(caller_number) && strlen(caller_number) < sizeof(chan_pvt->session->crm_info.caller_number)) {
		strcpy(chan_pvt->session->crm_info.caller_number, caller_number);
	} else {
		ast_log(LOG_DEBUG, "AMI don't send CRM caller_number\n");
	}

	if (0 < strlen(company) && strlen(company) < sizeof(chan_pvt->session->crm_info.company)) {
		strcpy(chan_pvt->session->crm_info.company, company);
	} else {
		ast_log(LOG_DEBUG, "AMI don't send CRM company\n");
	}

	ao2_ref(chan_pvt->session, +1);
	if (ast_sip_push_task(chan_pvt->session->serializer, transmit_info_with_crm_update, chan_pvt->session)) {
		ao2_cleanup(chan_pvt->session);
	}

	c = ast_channel_unref(c);

	return AMI_SUCCESS;
}

static int ami_query_nat_status(struct mansession *s, const struct message *m)
{
	int res = 0;
	struct avs_query_nat_status_rsp resp;
	char ami_rsp[MAX_MESSAGE_REPONSE*2 + 32] = { 0 };

	memset(&resp, 0, sizeof(struct avs_query_nat_status_rsp));

	res = ast_avs_control(NULL, AVS_CONTROL_QUERY_NAT_STATUS, &resp, sizeof(struct avs_query_nat_status_rsp));
	if (res == R_SUCCESS) {
		ast_log(LOG_DEBUG, "Query nat status success, stun_code:%d, stun_reason:%s, stun_desc:%s, turn_code:%d, turn_reason:%s, turn_desc:%s, turn_protocol:%s\n",
			resp.stun.code, resp.stun.reason, resp.stun.desc, resp.turn.code, resp.turn.reason, resp.turn.desc, resp.turn.protocol);
		if (ast_strlen_zero(resp.turn.protocol)) {
			snprintf(ami_rsp, sizeof(ami_rsp), "Success\r\n"
				"stun_code: %d\r\n"
				"stun_reason: %s\r\n"
				"stun_desc: %s\r\n"
				"turn_code: %d\r\n"
				"turn_reason: %s\r\n"
				"turn_desc: %s",
				resp.stun.code, resp.stun.reason, resp.stun.desc, resp.turn.code, resp.turn.reason, resp.turn.desc);
		} else {
			snprintf(ami_rsp, sizeof(ami_rsp), "Success\r\n"
				"stun_code: %d\r\n"
				"stun_reason: %s\r\n"
				"stun_desc: %s\r\n"
				"turn_code: %d\r\n"
				"turn_reason: %s\r\n"
				"turn_desc: %s\r\n"
				"turn_protocol: %s",
				resp.stun.code, resp.stun.reason, resp.stun.desc, resp.turn.code, resp.turn.reason, resp.turn.desc, resp.turn.protocol);
		}
		astman_send_response(s, m, ami_rsp, NULL);
	} else {
		ast_log(LOG_ERROR, "Query nat status failed\n");
		snprintf(ami_rsp, sizeof(ami_rsp), "Success\r\n"
				"stun_code: %d\r\n"
				"stun_reason: %s\r\n"
				"stun_desc: %s\r\n"
				"turn_code: %d\r\n"
				"turn_reason: %s\r\n"
				"turn_desc: %s",
				5002, "Internal Error", "Can't send avs to check nat status", 5002, "Internal Error", "Can't send avs to check nat status");
		astman_send_response(s, m, ami_rsp, NULL);
	}

	return AMI_SUCCESS;
}

/*For bug 181104, jzha, 2020-11-16*/
static int get_calltype_params_of_from(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	pj_str_t calltype_str = { "x-gs-call-type", 14};
	pjsip_from_hdr *from_hdr = NULL;
	char str_call_type[16] = {0};

	if (!session || !session->channel || !rdata) {
		ast_log(LOG_WARNING, "RX packet is NULL\n");
		return -1;
	}

	from_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_FROM, NULL);
	if (!from_hdr) {
		return -1;
	}

	pjsip_param *call_type = pjsip_param_find(&from_hdr->other_param, &calltype_str);
	if (call_type) {
		ast_copy_pj_str(str_call_type, &call_type->value, 14);
		if(str_call_type[0] != '\0' && !strcasecmp(str_call_type, "RG")) {
			pbx_builtin_setvar_helper(session->channel, "__IN_RINGGROUP", "yes");
		}
	}

	return 0;
}

/* added by zgmo for bug183396 */
static int get_useragent_from_hdr(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	const pj_str_t user_agent_name = { "User-Agent", 10 };
	pjsip_generic_string_hdr *useragent_hdr = NULL;

	if (!session || !session->channel || !rdata) {
		ast_log(LOG_WARNING, "RX packet is NULL\n");
		return -1;
	}

	useragent_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &user_agent_name, NULL);
	if (useragent_hdr) {
		pj_str_t value;
		pj_strdup_with_null(rdata->tp_info.pool, &value, &useragent_hdr->hvalue);

		ast_log(LOG_DEBUG, "User-Agent: %s, len: %ld", pj_strbuf(&value), pj_strlen(&value));
		if (!ast_strlen_zero(pj_strbuf(&value))) {
			pbx_builtin_setvar_helper(session->channel, "USER_AGENT", pj_strbuf(&value));
			ast_channel_device_info_set(session->channel, pj_strbuf(&value));
		}
	}

	return 0;
}

static int get_server_from_hdr(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	const pj_str_t server_name = { "Server", 6 };
	pjsip_generic_string_hdr *server_hdr = NULL;

	if (!session || !session->channel || !rdata) {
		ast_log(LOG_WARNING, "RX packet is NULL\n");
		return -1;
	}

	server_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &server_name, NULL);
	if (server_hdr) {
		pj_str_t value;
		pj_strdup_with_null(rdata->tp_info.pool, &value, &server_hdr->hvalue);

		ast_log(LOG_DEBUG, "Server: %s, len: %ld", pj_strbuf(&value), pj_strlen(&value));
		if (!ast_strlen_zero(pj_strbuf(&value))) {
			pbx_builtin_setvar_helper(session->channel, "USER_AGENT", pj_strbuf(&value));
			ast_channel_device_info_set(session->channel, pj_strbuf(&value));
		}
	}

	return 0;
}
/* end */

/*Start added by zyxia for Bug 291797,2024.3.14*/
static int process_contact_container_for_parkingstate(void *obj, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = obj;

	if ( ast_sip_check_if_need_send_parkingstate(wrapper->contact,arg)) {
		ast_sip_send_notify_to_update_gs_state(wrapper->contact,GS_STATE_TYPE_PARKING,arg);
	}
	return 0;
}

static int process_parkingstate_aor(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;
	ast_sip_for_each_contact(aor, process_contact_container_for_parkingstate, arg);
	return 0;
}
/*End added by zyxia for Bug 291797,2024.3.14*/

/*Start added by zyxia for callflip,2021.11.24 */

static int process_contact_container_for_callflip(void *obj, void *arg, int flags)
{
	struct ast_sip_contact_wrapper *wrapper = obj;
	/*modify by zyxia for Bug 209316,dont need send callflip to callee*/
	if (ast_sip_get_callflip_enable() && ast_sip_check_if_need_sendcallflip(wrapper->contact,arg)) {
		ast_sip_send_notify_to_update_gs_state(wrapper->contact,GS_STATE_TYPE_CALLFLIP,arg);
	}
	return 0;
}

static int process_callflip_aor(void *obj, void *arg, int flags)
{
	struct ast_sip_aor *aor = obj;
	ast_sip_for_each_contact(aor, process_contact_container_for_callflip, arg);
	return 0;
}


/*
"Action: SendCallFlipStatus\r\n"
"ActionID: %d\r\n"
"Extension: %s\r\n"
"Status:%s\r\n"
"RemoteID: %s\r\n"
"RemoteName: %s\r\n"
"CallType: %s\r\n"
*/
static int action_send_callflip_status(struct mansession *s, const struct message *m)
{
	char buf[256];
	struct ast_channel *chan= NULL;
	const char *extension = astman_get_header(m, "Extension");
	const char *callflip_status = astman_get_header(m, "Status");
	const char *remote_id = astman_get_header(m, "RemoteID");
	const char *remote_name = astman_get_header(m, "RemoteName");
	const char *calltype = astman_get_header(m, "CallType");
	const char *channelname = astman_get_header(m, "Channel");
	struct ast_sip_callflip_state callflip_info;

	if (ast_strlen_zero(extension)) {
		astman_send_error(s, m, "Extension not specified");
		return 0;
	}

	if (ast_strlen_zero(callflip_status)) {
		astman_send_error(s, m, "Status not specified");
		return 0;
	}

	/*enable call flip,calltype must given*/
	if ( !strcasecmp(callflip_status,"yes")) {
		if (ast_strlen_zero(calltype)) {
			astman_send_error(s, m, "CallType not specified");
			return 0;
		}

		if (ast_strlen_zero(remote_id)) {
			astman_send_error(s, m, "RemoteID not specified");
			return 0;
		}

		if (ast_strlen_zero(channelname)) {
			astman_send_error(s, m, "Channel not specified");
			return 0;
		}

		memset(&callflip_info, 0, sizeof(struct ast_sip_callflip_state));

		snprintf(callflip_info.enable, sizeof(callflip_info.enable), "%s", callflip_status);
		snprintf(callflip_info.remote_id, sizeof(callflip_info.remote_id), "%s", remote_id);
		snprintf(callflip_info.remote_name, sizeof(callflip_info.remote_name), "%s", remote_name);
		snprintf(callflip_info.call_type, sizeof(callflip_info.call_type), "%s", calltype);
		snprintf(callflip_info.channel_id, sizeof(callflip_info.channel_id), "%s", channelname);


		/*if conference,need get subject from database */
		if (!strcasecmp(calltype,"Conference")) {
			char family[64] = {0};
			char schedule_id[32] = {0};
			char conf_subject[512] = {0};
			/*added by zyxia for Bug 214247 -2022.1.25*/
			snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", "audio");

			snprintf(family, sizeof(family), "SCHEDULE/%s", remote_id);
			if (!ast_db_get(family, "ScheduleID", schedule_id, sizeof(schedule_id)) && !ast_strlen_zero(schedule_id)) {
				/*Conf Subject*/
				snprintf(family, sizeof(family), "SCHEDULE/%s", schedule_id);
				if (ast_db_get(family, "ConfSubject", conf_subject, sizeof(conf_subject)) || ast_strlen_zero(conf_subject)) {
					ast_log(LOG_DEBUG, "Did not get the ConfSubject!");
				}
				else{
					snprintf(callflip_info.remote_name, sizeof(callflip_info.remote_name), "%s", conf_subject);
				}
			}
		}else{

			chan = ast_channel_get_by_name(channelname);
			if (!chan) {
				snprintf(buf, sizeof(buf), "Channel does not exist: %s", channelname);
				astman_send_error(s, m, buf);
				return 0;
			}

			struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
			//added by zyxia for Bug 248369 ,2023.1.12
			if (channel && channel->session && channel->session->conf_link && channel->session->trunk_channel_id) {
				snprintf(callflip_info.conf_link, sizeof(callflip_info.conf_link), "%s", channel->session->conf_link);
				snprintf(callflip_info.replace_channel_id, sizeof(callflip_info.replace_channel_id), "%s", channel->session->trunk_channel_id);
			}

			if (ast_channel_get_is_videocall(chan)){
				snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", "video");
			}else{
				snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", "audio");
			}

			ast_channel_unref(chan);
		}
	}
	else
	{
		memset(&callflip_info, 0, sizeof(struct ast_sip_callflip_state));
		snprintf(callflip_info.enable, sizeof(callflip_info.enable), "%s", callflip_status);
	}


	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", extension);
	if (!endpoint) {
		snprintf(buf, sizeof(buf), "endpoint does not exist: %s", extension);
		ast_log(LOG_WARNING, "endpoint does not exist:%s \n",extension);
		astman_send_error(s, m, buf);
		return 0;
	}

	//save the callflip state to the ao2_container
	ast_sip_callflip_build_user(&callflip_info,extension);

	/*send notify to all contact*/
	ast_sip_for_each_aor(endpoint->aors, process_callflip_aor, &callflip_info);

	return 0;
}
/*End added by zyxia for 2021.11.24*/


/*Start added by zyxia for 2021.8.13*/

static int ast_get_extension_from_channel_name(const char *channel_name, char *buf, int size)
{
	const char *start = NULL;
	const char *end =  NULL;

	memset(buf, 0, size);

	if (strncasecmp(channel_name, "PJSIP/", 6)) {
		return -1;
	}
	start = channel_name + 6;
	end = strrchr(start, '-');
	if (!end) {
		return -1;
	}

	strncpy(buf, start, size > (end - start) ? (end - start) : size);
	return 0;
}

static int gen_p2p_recordfile_name_raw(struct ast_channel* chan, struct ast_channel *peer_chan, char **filename_out)
{
	int len =0;
	int use_peer_exten = 0;

	char *touch_filename = NULL;
	char *caller_chan_id = NULL;
	char *peer_chan_id = NULL;

	const char *from_sca = pbx_builtin_getvar_helper(chan, "FROM_SCA");
	const char *is_callee = pbx_builtin_getvar_helper(chan, "IS_CALLEE");
	const char *cf_reason = pbx_builtin_getvar_helper(chan, "CF_REASON");
	const char *caller_num = pbx_builtin_getvar_helper(chan, "CALLER_NUM");

	char exten[AST_MAX_EXTENSION] = {0};
	char peer_chan_exten[AST_MAX_EXTENSION] = {0};

	if (!filename_out) {
		return -1;
	}

	if(strlen(ast_channel_caller_num(chan)) > 0) {
		const char *pickup_caller = pbx_builtin_getvar_helper(chan, "PICKUP_CALLER");

		ast_log(LOG_DEBUG, "%s caller_num:%s", ast_channel_name(chan), ast_channel_caller_num(chan));
		if (!ast_strlen_zero(pickup_caller) && !strcmp(pickup_caller, "yes")) {
			if (strstr(ast_channel_name(chan), "DAHDI/")) {
				struct ast_party_caller *party_caller = ast_channel_caller(chan);

				if (party_caller->ani.number.valid && !ast_strlen_zero(party_caller->ani.number.str)) {
					caller_chan_id = ast_strdupa(party_caller->ani.number.str);
				} else {
					ast_log(LOG_WARNING,
						"%s is a pickup-caller and should be callee, but cannot retrive ANI NUMBER",
						ast_channel_name(chan));
					caller_chan_id = "unknown";
				}
			} else {
				if (sscanf(ast_channel_name(chan), "%*[^/]/%[^-]", exten)) {
					caller_chan_id = exten;
				}
			}
		} else {
			caller_chan_id = ast_strdupa(ast_channel_caller_num(chan));
		}
	} else {
		if(!strcmp(ast_channel_tech(chan)->type,"IAX2")){
			if(sscanf(ast_channel_name(chan),"IAX2/%[^-]-%*s",exten) == 1){
				caller_chan_id = ast_strdupa(exten);
			}
		} else if ((!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "14"))
			&& strstr(ast_channel_name(chan), "Local/")
			&& strstr(ast_channel_name(chan), ";2")
			&& !ast_strlen_zero(caller_num)) {
			caller_chan_id = (char *)caller_num;
		} else {
			if(sscanf(ast_channel_name(chan), "%*[^/]/%[^@]", exten)) {
				caller_chan_id = exten;
			}
		}
	}

	if (strstr(ast_channel_name(peer_chan), "Local/")) {
		const char *from_cb = pbx_builtin_getvar_helper(chan, "FROM_CALLBACK");
		const char *do_forward = pbx_builtin_getvar_helper(peer_chan, "DO_FORWARD");
		const char *aa_user = pbx_builtin_getvar_helper(peer_chan, "AAUSER");
		const char *from_park = pbx_builtin_getvar_helper(peer_chan, "FROM_PARK");
		const char *caller_do_park = pbx_builtin_getvar_helper(peer_chan, "CALLER_DO_PARK");
		const char *orig_callee = pbx_builtin_getvar_helper(peer_chan, "ORIG_CALLEE");

		if ((!ast_strlen_zero(cf_reason) && !strcmp(cf_reason, "14"))
			|| (!ast_strlen_zero(from_cb) && !strcmp(from_cb, "yes"))) {
			if (!ast_strlen_zero(caller_num)) {
				strcpy(peer_chan_exten, caller_num);
			} else if (strstr(ast_channel_name(peer_chan), ";2")
				&& !ast_strlen_zero(from_park) && !strcmp(from_park, "yes")
				&& !ast_strlen_zero(caller_do_park) && !strcmp(caller_do_park, "yes")
				&& !ast_strlen_zero(orig_callee)) {
				strcpy(peer_chan_exten, orig_callee);
			} else {
				sscanf(ast_channel_name(peer_chan), "%*[^/]/%[^@]", peer_chan_exten);
			}
		} else if (!ast_strlen_zero(do_forward) && !ast_strlen_zero(aa_user)
			&& strstr(ast_channel_name(peer_chan), ";2")) {
			strcpy(peer_chan_exten, aa_user);
		} else {
			sscanf(ast_channel_name(peer_chan), "%*[^/]/%[^@]", peer_chan_exten);
		}

		use_peer_exten = 1;
		ast_log(LOG_DEBUG, "using peer_chan_exten: %s", peer_chan_exten);
	} else {
		const char *is_callee = pbx_builtin_getvar_helper(peer_chan, "IS_CALLEE");
		const char *callee_num = pbx_builtin_getvar_helper(peer_chan, "CALLEE_NUM");

		if (strstr(ast_channel_name(peer_chan), "/trunk_")) {
			const char *ringboth = pbx_builtin_getvar_helper(chan, "RINGBOTH");
			const char *org_is_callee = pbx_builtin_getvar_helper(peer_chan, "ORG_IS_CALLEE");
			const char *orig_callee = pbx_builtin_getvar_helper(peer_chan, "ORIG_CALLEE");

			if (!ast_strlen_zero(ringboth) && !strcmp(ringboth, "yes")) {
				const char *external_number = pbx_builtin_getvar_helper(peer_chan, "EXTERNAL_NUMBER");

				if (!ast_strlen_zero(orig_callee)
					&& !ast_strlen_zero(org_is_callee) && !strcmp(org_is_callee, "yes")) {
					strcpy(peer_chan_exten, orig_callee);
					use_peer_exten = 1;
					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				} else if (!ast_strlen_zero(external_number)) {
					strcpy(peer_chan_exten, external_number);
					use_peer_exten = 1;
					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", external_number);
				}
			} else {
				if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")
							&& !ast_strlen_zero(callee_num)) {
					strcpy(peer_chan_exten, callee_num);
					use_peer_exten = 1;

					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				} else if (!ast_strlen_zero(org_is_callee) && !strcmp(org_is_callee, "yes")
							&& !ast_strlen_zero(orig_callee)) {
					strcpy(peer_chan_exten, orig_callee);
					use_peer_exten = 1;
					ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
				}
			}
		} else if (strstr(ast_channel_name(peer_chan), "DAHDI/o")) {
			if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")
						&& !ast_strlen_zero(callee_num)) {
				strcpy(peer_chan_exten, callee_num);
				use_peer_exten = 1;

				ast_log(LOG_DEBUG, "In this case, using peer_exten:%s", peer_chan_exten);
			}
		} else {
			sscanf(ast_channel_name(peer_chan), "%*[^/]/%[^-]", peer_chan_exten);
		}
	}

	if(strstr(ast_channel_caller(peer_chan)->id.number.str, "Callback")) {
		use_peer_exten = 1;
		ast_log(LOG_DEBUG, "In this case, mark use_peer_exten");
	}

	if (use_peer_exten) {
		peer_chan_id = peer_chan_exten;
	} else {
		peer_chan_id = ast_strdupa(S_COR(ast_channel_caller(peer_chan)->ani.number.valid,
			ast_channel_caller(peer_chan)->ani.number.str,
			ast_channel_caller(peer_chan)->id.number.str));

		ast_log(LOG_DEBUG, "use peer_exten from party-caller :%s", peer_chan_id ? : "none");
	}

	len = strlen(caller_chan_id) + strlen(peer_chan_id) + 60;
	touch_filename = ast_malloc(len);

	time_t timep;
	struct tm *p;
	time(&timep);
	p = localtime(&timep);

	if (!ast_strlen_zero(from_sca) && !strcmp(from_sca, "yes")) {
		if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
			snprintf(touch_filename, len, "%d-%02d/%s-%s-%s-%ld.%s",
				1900 + p->tm_year, 1 + p->tm_mon,
				"rec",
				peer_chan_id,
				caller_chan_id,
				(long) time(NULL),
				"wav");
		} else {
			snprintf(touch_filename, len, "%d-%02d/%s-%s-%s-%ld.%s",
				1900 + p->tm_year, 1 + p->tm_mon,
				"rec",
				caller_chan_id,
				peer_chan_id,
				(long) time(NULL),
				"wav");
		}
	} else {
		/* modified by zgmo for bug153684 */
		if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
			snprintf(touch_filename, len, "%d-%02d/%s-%ld-%s-%s.%s",
				1900 + p->tm_year, 1 + p->tm_mon,
				"auto",
				(long) time(NULL),
				peer_chan_id,
				caller_chan_id,
				"wav");
		} else {
			snprintf(touch_filename, len, "%d-%02d/%s-%ld-%s-%s.%s",
				1900 + p->tm_year, 1 + p->tm_mon,
				"auto",
				(long) time(NULL),
				caller_chan_id,
				peer_chan_id,
				"wav");
		}
	}

	*filename_out = touch_filename;
	return 0;
}

static char* ast_set_record_filename_and_cdr(struct ast_channel* chan, struct ast_channel *peer_chan)
{
	int ret = 0;
	int len = 0;
	int queue_record = 0;
	char *touch_filename = NULL;

	const char *from_sca = NULL;
	const char *from_queue = NULL;
	const char *mixmonitor_filename = NULL;
	const char *mix_filename = NULL;
	const char *in_ringgroup = NULL;
	//const char *merge_audio_recording_files = NULL;

	char val[2*AST_MAX_USER_FIELD] = {0};
	char tmp[2*AST_MAX_USER_FIELD] = {0};
	char record_files[2*AST_MAX_USER_FIELD] = {0};

	if (peer_chan == NULL || chan == NULL) {
		return NULL;
	}

	ast_log(LOG_DEBUG, "ast_set_record_filename_and_cdr|chan:%s, peer_chan: %s",
		ast_channel_name(chan), ast_channel_name(peer_chan));

	from_sca = pbx_builtin_getvar_helper(chan, "FROM_SCA");
	from_queue = pbx_builtin_getvar_helper(peer_chan, "FROM-QUEUE");
	mixmonitor_filename = pbx_builtin_getvar_helper(peer_chan, "MIXMONITOR_FILENAME");
	mix_filename = pbx_builtin_getvar_helper(peer_chan, "MIXFILENAME");
	in_ringgroup = pbx_builtin_getvar_helper(peer_chan, "IN_RINGGROUP");

	ast_log(LOG_DEBUG, "from_sca:%s, from_queue:%s, in_ringgroup:%s",
		ast_strlen_zero(from_sca)? "":from_sca,
		ast_strlen_zero(from_queue)? "":from_queue,
		ast_strlen_zero(in_ringgroup)? "":in_ringgroup);

	ast_log(LOG_DEBUG, "mixmonitor_filename:%s, mix_filename:%s",
		ast_strlen_zero(mixmonitor_filename)? "":mixmonitor_filename,
		ast_strlen_zero(mix_filename)? "":mix_filename);

	if (!ast_strlen_zero(from_queue) && !strcmp(from_queue, "yes")
		&& !ast_strlen_zero(mixmonitor_filename) && strstr(mixmonitor_filename, "asterisk/queue/")) {
		char *filename_dump = NULL;
		char *base_name = NULL, *dir_name = NULL;
		char *head = NULL, *tail = NULL, *sep = NULL;

		/*
		 *  .e.g:
		 *  /app/asterisk/var/spool/asterisk/queue/2022-10/q6500-4002-20221019-170540-1666170337.167-4001.wav
		 */
		ast_log(LOG_DEBUG, "Detect Callflip in CallQueue, Curent recordingFile:%s", mixmonitor_filename);

		ast_asprintf(&filename_dump, "%s", mixmonitor_filename);
		if (!filename_dump) {
			ast_log(LOG_DEBUG, "Insufficient memory");
			return NULL;
		}

		base_name = basename(filename_dump);
		dir_name = dirname(filename_dump);
		dir_name = strrchr(dir_name, '/');
		if (!base_name || !dir_name) {
			ast_free(filename_dump);
			ast_log(LOG_ERROR, "Aboard in this case");

			return NULL;
		}

		dir_name++;

		ast_log(LOG_DEBUG, "dir_name:%s", dir_name);
		ast_log(LOG_DEBUG, "base_name:%s", base_name);

		sep = strrchr(base_name, '-');
		if (!sep) {
			ast_free(filename_dump);
			ast_log(LOG_ERROR, "Aboard in this case");

			return NULL;
		}

		*sep = '\0';
		tail = sep + 1;

		sep = strrchr(base_name, '-');
		if (!sep) {
			ast_free(filename_dump);
			ast_log(LOG_ERROR, "Aboard in this case");

			return NULL;
		}

		*sep = '\0';
		head = base_name;

		ast_asprintf(&touch_filename, "%s/%s-%s-%s", dir_name, head,
			ast_channel_uniqueid(chan), tail);

		ast_log(LOG_DEBUG, "New touch_filename:%s",
			ast_strlen_zero(touch_filename) ? "":touch_filename);

		ast_free(filename_dump);
		queue_record = 1;
	} else if (!ast_strlen_zero(in_ringgroup) && !strcmp(in_ringgroup, "yes")
		&& !ast_strlen_zero(mix_filename)) {
		/*
		 *  MIXFILENAME=2022-10/auto-1666226379#4001-4002-6400
		 *  Now we need to change the TimeStamp
		 */
		char *filename_dump = NULL;
		char *base_name = NULL, *dir_name = NULL;
		char *head = NULL, *tail = NULL, *sep = NULL;
		struct timeval cur_time = {0};

		ast_log(LOG_DEBUG, "Detect Callflip in RingGroup, Curent recordingFile:%s", mix_filename);

		ast_asprintf(&filename_dump, "%s", mix_filename);
		if (!filename_dump) {
			ast_log(LOG_DEBUG, "Insufficient memory");
			return NULL;
		}

		base_name = basename(filename_dump);
		dir_name = dirname(filename_dump);
		if (!base_name || !dir_name) {
			ast_free(filename_dump);
			ast_log(LOG_ERROR, "Aboard in this case");

			return NULL;
		}

		ast_log(LOG_DEBUG, "dir_name:%s", dir_name);
		ast_log(LOG_DEBUG, "base_name:%s", base_name);

		sep = strchr(base_name, '-');
		if (!sep) {
			ast_free(filename_dump);
			ast_log(LOG_ERROR, "Aboard in this case");

			return NULL;
		}

		*sep = '\0';  /* auto'\0'1666226379#4001-4002-6400 */
		sep++;
		head = base_name;

		sep = strrchr(sep, '#');
		if (!sep) {
			ast_free(filename_dump);
			ast_log(LOG_ERROR, "Aboard in this case");

			return NULL;
		}

		head = base_name;
		tail = sep;  /* #4001-4002-6400 */

		gettimeofday(&cur_time, NULL);
		ast_asprintf(&touch_filename, "%s/%s-%ld%s.wav", dir_name, head, cur_time.tv_sec, tail);

		ast_log(LOG_DEBUG, "New touch_filename:%s",
			ast_strlen_zero(touch_filename) ? "":touch_filename);

		ast_free(filename_dump);
	} else {
		/*merge_audio_recording_files = pbx_builtin_getvar_helper(NULL, "MERGE_AUDIO_RECORDING_FILES");
		if ( (!ast_strlen_zero(merge_audio_recording_files) && !strcmp(merge_audio_recording_files, "yes")) &&
			 !ast_strlen_zero(mix_filename) ) {
			ast_asprintf(&touch_filename, "%s.wav", mix_filename);
		} else {*/
			ret = gen_p2p_recordfile_name_raw(chan, peer_chan, &touch_filename);
			if (ret || !touch_filename) {
				ast_log(LOG_ERROR, "Failed to generate recording file name");
				return NULL;
			}
		//}
	}

	ast_log(LOG_DEBUG, "[ast_set_record_filename_and_cdr] ast_set_record_filename:%s \n", touch_filename);

	len = sizeof(tmp)-1;
	strncat(tmp, touch_filename, len);
	len -= strlen(touch_filename);
	strncat(tmp, "@", len);

	/*get cdr var  from old chan and set it to new_channel*/
	ast_cdr_getvar(ast_channel_name(chan), "lastapp", val, sizeof(val));
	ast_log(LOG_VERBOSE, "lastapp = %s, new record:%s", val, tmp);

	if (strcmp(val, "ParkedCall") != 0) {
		val[0] = '\0';
		ast_cdr_getvar(ast_channel_name(chan), "RECORD_FILES", val, sizeof(val));

		ast_log(LOG_VERBOSE, "Chan:%s, old-cdr-record:%s",
			ast_channel_name(chan), val);

		if (!strstr(val, tmp)) {
			snprintf(record_files,sizeof(record_files)-1,"%s%s",val,tmp);
			ast_cdr_setvar(ast_channel_name(chan), "RECORD_FILES", record_files);

			ast_log(LOG_VERBOSE, "Set %s's cdr RECORD_FILES:%s",
				ast_channel_name(chan), record_files);
		}
	}

	for (int x = 8; x < strlen(touch_filename); x++) {
		if (touch_filename[x] == '/') {
			touch_filename[x] = '-';
		}
	}

	pbx_builtin_setvar_helper(peer_chan, "MIXFILENAME", touch_filename);

	if (!ast_strlen_zero(from_sca) && !strcmp(from_sca, "yes")) {
		char *touch_filename_sca = NULL;

		ast_asprintf(&touch_filename_sca, "/app/asterisk/var/spool/asterisk/sca/%s", touch_filename);
		if (touch_filename_sca) {
			ast_log(LOG_DEBUG, "Using absolute path for SCA:%s", touch_filename_sca);
			ast_free(touch_filename);
			touch_filename = touch_filename_sca;
		}
	} else if (queue_record) {
		char *touch_filename_queue = NULL;

		ast_asprintf(&touch_filename_queue, "/app/asterisk/var/spool/asterisk/queue/%s", touch_filename);
		if (touch_filename_queue) {
			ast_log(LOG_DEBUG, "Using absolute path for QUEUE:%s", touch_filename_queue);
			ast_free(touch_filename);
			touch_filename = touch_filename_queue;
		}
	}

	return touch_filename;
}

static void set_bfcp_account_id(struct ast_channel *peer, struct ast_bfcp_info *bfcp_info)
{
	const char *slider_extension = NULL;
	char *caller_name = NULL;
	char extension[AST_MAX_EXTENSION] = { 0 };

	if (!peer || !bfcp_info) {
		return;
	}

	caller_name = S_COR(ast_channel_caller(peer)->id.name.valid, ast_channel_caller(peer)->id.name.str, "");

	if (strstr(ast_channel_name(peer), "PJSIP/trunk")) {
		slider_extension = pbx_builtin_getvar_helper(peer, "FLOOR_OWNER_ID");
		if (!ast_strlen_zero(slider_extension)) {
			strncpy(bfcp_info->account_id, slider_extension, sizeof(bfcp_info->account_id) - 1);
			ast_log(LOG_DEBUG, "Trunk use FLOOR_OWNER_ID set_bfcp_account_id:%s, peer:%s\n", bfcp_info->account_id, ast_channel_name(peer));
		} else {
			if (!ast_strlen_zero(caller_name)) {
				strncpy(bfcp_info->account_id, caller_name, sizeof(bfcp_info->account_id) - 1);
				ast_log(LOG_DEBUG, "Trunk use caller_name set_bfcp_account_id:%s, peer:%s\n", bfcp_info->account_id, ast_channel_name(peer));
			} else {
				slider_extension = pbx_builtin_getvar_helper(peer, "AAUSER");
				if (!ast_strlen_zero(slider_extension)) {
					strncpy(bfcp_info->account_id, slider_extension, sizeof(bfcp_info->account_id) - 1);
					ast_log(LOG_DEBUG, "Trunk use AAUSER set_bfcp_account_id:%s, peer:%s\n", bfcp_info->account_id, ast_channel_name(peer));
				}
			}
		}
	} else {
		if (!ast_strlen_zero(caller_name)) {
			strncpy(bfcp_info->account_id, caller_name, sizeof(bfcp_info->account_id) - 1);
			ast_log(LOG_DEBUG, "Use caller_name set_bfcp_account_id:%s, peer:%s\n", bfcp_info->account_id, ast_channel_name(peer));
		} else {
			ast_get_extension_from_channel_name(ast_channel_name(peer), extension, sizeof(extension));
			strncpy(bfcp_info->account_id, extension, sizeof(bfcp_info->account_id) - 1);
			ast_log(LOG_DEBUG, "Use extension set_bfcp_account_id:%s, peer:%s\n", bfcp_info->account_id, ast_channel_name(peer));
		}
	}

	ast_log(LOG_DEBUG, "set_bfcp_account_id:%s, peer:%s\n", bfcp_info->account_id, ast_channel_name(peer));

	return;
}

/*Start added by zyxia for Bug 235579, call flip the page caller,need update the video track msid*/
static void page_change_caller(struct ast_channel *old_caller,struct ast_channel *new_caller)
{
	RAII_VAR(char *, msid, NULL, ast_free);
	struct ast_app *app = NULL;
	char appdata[64] = {0};

	if( !old_caller || !new_caller)
		return ;

	msid = ast_channel_get_msid(new_caller, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	if (!msid) {
		msid = ast_strdup("");
	}

	snprintf(appdata, sizeof(appdata), "%s,%s", msid,ast_channel_name(new_caller));//edit by zyxia for Bug 250509,2023.2.14
	app = pbx_findapp("PageConfChangeCaller");
	ast_channel_set_msid(old_caller,AST_MEDIA_SUB_VIDEO_TYPE_MAIN,msid);//added by zyxia for Bug 250509,2023.2.14

	if (app) {
		ast_log(LOG_DEBUG, "CallFlip run  PageConfChangeCaller APP, chan:%s,msid:%s\n",ast_channel_name(old_caller),appdata);
		pbx_exec(old_caller, app, S_OR(appdata, ""));
	} else {
		ast_log(LOG_ERROR, "CallFlip can't PageConfChangeCaller APP, chan:%s\n",ast_channel_name(old_caller));
	}
	return;
}
/*End added by zyxia for Bug 235579, call flip the page caller,need update the video track msid*/

/*Start added by zyxia for Bug251489,2023.2.23*/
static void page_private_intercom_callflip(struct ast_channel *old_chan,struct ast_channel *new_chan)
{
	struct ast_app *app = NULL;
	char appdata[64] = {0};

	if( !old_chan || !new_chan)
		return ;

	snprintf(appdata, sizeof(appdata), "%s",ast_channel_name(new_chan));
	app = pbx_findapp("PrivateintercomCallflip");
	if (app) {
		ast_log(LOG_DEBUG, "CallFlip run  PrivateintercomCallflip APP, chan:%s,new_channame:%s\n",ast_channel_name(old_chan),appdata);
		pbx_exec(old_chan, app, S_OR(appdata, ""));
	} else {
		ast_log(LOG_ERROR, "CallFlip can't PrivateintercomCallflip APP, chan:%s\n",ast_channel_name(old_chan));
	}
	return;
}
/*End added by zyxia for Bug251489,2023.2.23*/

/*!
 * \internal
 * \brief Update connected line on chan from peer.
 * \since 13.6.0
 *
 * \param chan Channel to get connected line updated.
 * \param peer Channel providing connected line information.
 * \param is_caller Non-zero if chan is the calling channel.
 *
 * \return Nothing
 */
static void update_connected_line_from_peer(struct ast_channel *chan, struct ast_channel *peer, int is_caller)
{
	struct ast_party_connected_line connected_caller;

	ast_party_connected_line_init(&connected_caller);

	ast_channel_lock(peer);
	ast_connected_line_copy_from_caller(&connected_caller, ast_channel_caller(peer));
	ast_channel_unlock(peer);
	connected_caller.source = AST_CONNECTED_LINE_UPDATE_SOURCE_ANSWER;
	if (ast_channel_connected_line_sub(peer, chan, &connected_caller, 0)
		&& ast_channel_connected_line_macro(peer, chan, &connected_caller, is_caller, 0)) {
		ast_channel_update_connected_line(chan, &connected_caller, NULL);
	}
	ast_party_connected_line_free(&connected_caller);
}

/*Start added by zyxia for Bug 251671,2023.2.23*/
static void process_update_connected_line_for_callflip(struct ast_channel *chan,struct ast_channel *peer,const char *bridge_peer)
{
	RAII_VAR(struct ast_channel *, peer_chan, NULL, ao2_cleanup);

	if(chan == NULL ){
		ast_log(LOG_ERROR, "process_update_connected_line_for_callflip chan is null\n");
		return ;
	}

	const char *page_intercom = pbx_builtin_getvar_helper(chan, "PAGE_INTERCOM");
	if(page_intercom && !strcasecmp(page_intercom,"yes")) {
		if(ast_channel_get_is_page_caller(chan)) {
			ast_indicate(chan,AST_CONTROL_CONNECTED_LINE);
			ast_channel_set_is_callflip_have_sendupdate(chan,1);
		}
		else {
			RAII_VAR(struct ast_channel *, caller_chan, NULL, ao2_cleanup);
			const char* caller_name = pbx_builtin_getvar_helper(chan, "PAGING_INTERCOM_CALLER_CHAN_NAME");
			caller_chan = ast_channel_get_by_name(caller_name);
			if (caller_chan) {
				update_connected_line_from_peer(chan,caller_chan,1);
				ast_channel_set_is_callflip_have_sendupdate(chan,1);
			}
		}
	}

	/*if when the bridge reconfig,dont send the sip update, need send at here to make sure the remote party-id is true,*/
	if (!ast_channel_get_is_callflip_have_sendupdate(chan) && !strstr(ast_channel_name(chan), "PJSIP/trunk")){
		if (!peer && !ast_strlen_zero(bridge_peer) && strstr(bridge_peer, ",") == NULL) {
			peer_chan = ast_channel_get_by_name(bridge_peer);
		}

		if (peer_chan) {
			update_connected_line_from_peer(chan,peer_chan,1);
			ast_channel_set_is_callflip_have_sendupdate(chan,1);
		} else if(peer) {
			update_connected_line_from_peer(chan,peer,1);
			ast_channel_set_is_callflip_have_sendupdate(chan,1);
		}
	}
	return;
}
/*End added by zyxia for Bug 251671,2023.2.23*/


static int process_trunk_info_for_callflip(struct ast_channel *chan,struct ast_channel *new_chan){

	if(chan == NULL || new_chan == NULL){
		ast_log(LOG_ERROR, "process_trunk_info_for_callflip chan is null\n");
		return 0;
	}

	struct ast_sip_channel_pvt *old_channel = ast_channel_tech_pvt(chan);
	struct ast_sip_channel_pvt *new_channel = ast_channel_tech_pvt(new_chan);
	if(old_channel && old_channel->session && new_channel && new_channel->session) {
		get_peer_trunk_info(new_channel->session,old_channel->session);//for Bug 238123
	}

	const char *trunk_index= pbx_builtin_getvar_helper(chan, "TRUNK_INDEX");
	if(trunk_index) {
		pbx_builtin_setvar_helper(new_chan, "TRUNK_INDEX", pbx_builtin_getvar_helper(chan, "TRUNK_INDEX"));

		/*if old channel is webrtc or wavea pp trunk conf,need set the od*/
		if( !ast_strlen_zero(new_channel->session->trunk_conf_num)
			&& new_channel->session->terminal_type  != AST_TERMINAL_TYPE_WEBRTC
			&& new_channel->session->terminal_type  != AST_TERMINAL_TYPE_GSWAVE_MOBILE
			&& (old_channel->session->terminal_type  == AST_TERMINAL_TYPE_WEBRTC || old_channel->session->terminal_type  == AST_TERMINAL_TYPE_GSWAVE_MOBILE)){

			char local_virtual_conf[512] = {0};
			char family[64] = {0};
			char trunkName[128] = {0};

			snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(trunk_index));
			ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
			snprintf(local_virtual_conf, sizeof(local_virtual_conf) - 1, "%s_%s", trunkName,new_channel->session->trunk_conf_num);
			ast_channel_set_is_join_trunk_conf(chan,1);
			ast_channel_set_virtual_conf_num(chan, local_virtual_conf); //set the old channel ,make sure the new channel have the value after channel move
			ast_log(LOG_DEBUG, "process_trunk_info_for_callflip chan %s set conf:%s\n",ast_channel_name(chan),local_virtual_conf);

		}
	}
	return 1;
}

static void reset_bfcp_channel(struct ast_channel *chan)
{
	int i = 0;
	struct ast_sip_channel_pvt *pvt = NULL;
	struct ast_sip_session *session = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!chan) {
		ast_log(LOG_DEBUG, "reset_bfcp_channel can't get chan\n");
		return;
	}

	pvt = ast_channel_tech_pvt(chan);
	if (!pvt) {
		ast_log(LOG_DEBUG, "reset_bfcp_channel can't get pvt\n");
		return;
	}

	session = pvt->session;
	if (!session) {
		ast_log(LOG_DEBUG, "reset_bfcp_channel can't get session\n");
		return;
	}

	if (!session->active_media_state) {
		ast_log(LOG_DEBUG, "[session:%p] reset_bfcp_channel can't get active_media_state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&session->active_media_state->sessions); ++i) {
		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
		if (!session_media) {
			continue;
		}

		if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP) {
			if (session_media->bfcp_udp) {
				ast_log(LOG_DEBUG, "[session:%p] Reset bfcp udp channel from %p to %p, endpoint:%s\n",
					session, session_media->bfcp_udp->channel, chan, ast_sorcery_object_get_id(session->endpoint));
				session_media->bfcp_udp->channel = chan;
			}
			break;
		}

		if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
			if (session_media->bfcp_tcp) {
				ast_log(LOG_DEBUG, "[session:%p] Reset bfcp tcp channel from %p to %p, endpoint:%s\n",
					session, session_media->bfcp_tcp->channel, chan, ast_sorcery_object_get_id(session->endpoint));
				if (session_media->bfcp_tcp->channel) {
					ast_channel_unref(session_media->bfcp_tcp->channel);
					session_media->bfcp_tcp->channel = chan;
					ast_channel_ref(chan);
				} else {
					session_media->bfcp_tcp->channel = chan;
					ast_channel_ref(chan);
				}
			}
			break;
		}
	}

	return;
}

static int action_callflip(struct mansession *s, const struct message *m)
{
	int is_slides_ok=0; //slides sdp negation success
	int is_auto_record = 0;
	unsigned int peer_is_callee = 0;
	int need_reinvite= 0;
	char buf[256] = {0};
	char *bridgeid = NULL;
	char* filename= NULL;
	const char *name = astman_get_header(m, "OldChannel");
	const char *new_name = astman_get_header(m, "NewChannel");
	const char *conf_rec_chan_name = NULL;
	int page_role = 0;//1 caller, 2 callee
	RAII_VAR(struct ast_bridge *, bridge, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, chan, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, peer_chan, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, peer, NULL, ao2_cleanup);
	struct ast_sip_channel_pvt *channel = NULL;
	const char *outbound_group_ext = NULL;
	char val[2*AST_MAX_USER_FIELD] = {0};
	char record_files[2*AST_MAX_USER_FIELD] = {0};

	if (ast_strlen_zero(name)) {
		astman_send_error(s, m, "Channel not specified");
		return 0;
	}

	chan = ast_channel_get_by_name(name);
	if (!chan) {
		snprintf(buf, sizeof(buf), "Channel does not exist: %s", name);
		astman_send_error(s, m, buf);
		return 0;
	}

	bridge = ast_channel_get_bridge(chan);
	if (!bridge) {
		snprintf(buf, sizeof(buf), "Bridge does not exist for channel %s", name);
		astman_send_error(s, m, buf);
		return 0;
	}

	if (ast_strlen_zero(new_name)) {
		goto FAILED;
	}

	ast_log(LOG_DEBUG, "Enter action_callflip, OldChannel:%s, NewChannel:%s\n", name, new_name);

	RAII_VAR(struct ast_channel *, new_chan, ast_channel_get_by_name(new_name), ao2_cleanup);

	if (!new_chan) {
		goto FAILED;
	}

	/*Start added by zyxia for Bug 261524 ,2023.7.20*/
	if ((outbound_group_ext = pbx_builtin_getvar_helper(new_chan, "OUTBOUND_GROUP_EXT"))) {
		outbound_group_ext = ast_strdupa(outbound_group_ext);
	}

	if(outbound_group_ext)
		ast_app_group_set_channel(new_chan, outbound_group_ext);
	/*End added by zyxia for Bug 261524 ,2023.7.20*/

	peer = ast_channel_real_peer(chan);
	if (peer && NULL != strstr(ast_channel_name(peer), "trunk_")) {
		const char *r_calleenum = pbx_builtin_getvar_helper(chan, "REG_CALLEENUM");
		if (NULL != r_calleenum){
			ast_channel_connect_num_set(peer, r_calleenum);
		}
	}
	/*start added by zyxia for Bug 210743 ,2021.12.16*/
	const char *call_to_ipc = pbx_builtin_getvar_helper(chan, "CALL_TO_IPC");
	const char *bridge_peer = pbx_builtin_getvar_helper(chan, "BRIDGEPEER");
	if (!peer) {
		const char *from_sca = pbx_builtin_getvar_helper(chan, "FROM_SCA");

		ast_log(LOG_DEBUG, "BRIDGEPEER:%s", ast_strlen_zero(bridge_peer) ? "":bridge_peer);
		ast_log(LOG_DEBUG, "FROM_SCA:%s", ast_strlen_zero(from_sca) ? "":from_sca);
		if (!ast_strlen_zero(from_sca) && !strcmp(from_sca, "yes")
				&& !ast_strlen_zero(bridge_peer) && strstr(bridge_peer, "CBRec/SCA_")) {
			char *sep = strchr(bridge_peer, ',');
			if (sep) {
				sep=  sep + 1;
				ast_log(LOG_DEBUG, "Detect SCA with recording, Try to retrieve %s", sep);
				peer = ast_channel_get_by_name(sep);
			}
		}
		else if( !ast_strlen_zero(bridge_peer) && strstr(bridge_peer, "CBAnn/")) {
			char *sep = strchr(bridge_peer, ',');
			if (sep) {
				sep=  sep + 1;
				ast_log(LOG_DEBUG, "Detect Announcement , Try to retrieve %s", sep);
				peer = ast_channel_get_by_name(sep);
			}
		}
	}

	if (!strncmp(ast_channel_name(chan), "PJSIP/", 6)
		&& strncmp(ast_channel_name(chan), "PJSIP/trunk_", 12)) {
		const char *rec_enable = NULL;
		const char *rec_enable_peer = NULL;

		rec_enable = pbx_builtin_getvar_helper(chan, "REC_ENABLE");
		rec_enable_peer = pbx_builtin_getvar_helper(peer, "REC_ENABLE");

		if ((!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes"))
			|| (!ast_strlen_zero(rec_enable_peer) && !strcmp(rec_enable_peer, "yes"))) {
			is_auto_record =1;
		}

		if (is_auto_record) {
			pbx_builtin_setvar_helper(new_chan, "REC_ENABLE","yes");
			update_auto_record_sdp(new_chan);
		}
	}

	ast_log(LOG_DEBUG, "is_auto_record:%d, peer(%p):%s",
				is_auto_record, peer, peer ? ast_channel_name(peer):"");

	/*Start added by zyxia for Bug201571,2021.9.1*/
	const char *is_callee = pbx_builtin_getvar_helper(chan, "IS_CALLEE");
	if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
		peer_is_callee = 1;
		pbx_builtin_setvar_helper(new_chan, "IS_CALLEE","yes");//added by zyxia for Bug 259859
	}
	/*End added by zyxia for Bug201571,2021.9.1*/

	/*added by zyxia for Bug 235579,need set the vedeo direction */
	if(ast_channel_get_is_page_video_support(chan) ){
		if(check_channel_is_video_call(chan)){
			ast_log(LOG_DEBUG, " channel_name %s,is_callee:%s", ast_channel_name(chan),is_callee ? is_callee:"");
			ast_channel_set_is_page_video_support(new_chan,1);//edit by zyxia for Bug 258912
			if (peer_is_callee) {
				change_video_media_direction(new_chan,AST_MEDIA_SUB_VIDEO_TYPE_MAIN, "sendonly");
				change_video_media_direction(new_chan,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES, "inactive");
				page_role = 2;
			}
			else {// caller
				change_video_media_direction(new_chan,AST_MEDIA_SUB_VIDEO_TYPE_MAIN, "recvonly");
				change_video_media_direction(new_chan,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES, "inactive");

				page_role = 1;
			}
		}
		/*added by zyxia for Bug 246497 ,if the paging is video,disable it*/
		else{
			disable_main_video(new_chan,TRUE);
			ast_channel_set_is_videocall(new_chan,FALSE);
		}
	}
	else if (!ast_strlen_zero(call_to_ipc) && !strcasecmp(call_to_ipc, "yes")) {
		pbx_builtin_setvar_helper(new_chan, "CALL_TO_IPC","yes");

		struct ast_sip_channel_pvt* chan_pvt = ast_channel_tech_pvt(new_chan);
		if (chan_pvt && chan_pvt->session) {
			chan_pvt->session->is_rtsp_call = 1;
		}
		//added by zyxia for Bug208161,peer is rtsp,need update local sdp,2021.12.23
		if(peer)
			update_sdp_for_local(new_chan, peer);
	}
	//if the peer is not video call,200 ok need diable main video
	else if (!peer || (peer && !strcasecmp(ast_channel_tech(peer)->type, "PJSIP") && !check_channel_is_video_call(peer))
				|| ( peer && !strcasecmp(ast_channel_tech(peer)->type, "DAHDI"))) {
		disable_main_video(new_chan,TRUE);
		ast_channel_set_is_videocall(new_chan,FALSE);
	}
	/*Start added by zyxia for Bug 228808,2022.8.24*/
	else {
		const char *chan_calltype = pbx_builtin_getvar_helper(chan, "CALL_TYPE");
		if (NULL != chan_calltype && !strcmp(chan_calltype,"VideoConference")
			&& peer && ast_channel_name(peer) && strstr(ast_channel_name(peer), "PJSIP/trunk")
			&& check_channel_is_video_call(peer)){
			change_video_media_direction(new_chan,AST_MEDIA_SUB_VIDEO_TYPE_MAIN, "sendonly");
		}
	}
	/*End added by zyxia for Bug 228808,2022.8.24*/


	/*Start added by zyxia for Bug 235007,2023.3.17*/
	if(ast_channel_page_mode_get(chan) != AST_PAGE_MODE_NONE )
		ast_channel_page_mode_set(new_chan, ast_channel_page_mode_get(chan));
	/*End added by zyxia for Bug 235007,2023.3.17*/

	if(ast_media_role_get() == MEDIA_ROLE_MASTER && ast_channel_get_syncflag(chan) && !ast_strlen_zero(ast_channel_get_avs_address(chan))
		&& strcmp("127.0.0.1",ast_channel_get_avs_address(chan))){
		ast_channel_set_syncflag(new_chan, 1);
		ast_channel_set_recordfile_slave_media(new_chan,ast_channel_get_recordfile_slave_media(chan));
		ast_channel_set_recordfile_master_media(new_chan,ast_channel_get_recordfile_master_media(chan));
		ast_channel_set_record_in_media(new_chan,ast_channel_get_record_in_media(chan));//added it for total records cnt 
	}

	process_trunk_info_for_callflip(chan,new_chan);//added by zyxia for Bug 246351

	int ret = ast_answer(new_chan);
	/*start added by zyxia for Bug 28773.2024.2.24*/
	if (ret < 0) {
		ast_log(LOG_ERROR, " channel_name %s callflip answer failed", ast_channel_name(new_chan));
		astman_send_ack(s, m, "CALL FLIP failed");
		return 0;
	}
	/*End added by zyxia for Bug 28773.2024.2.24*/

	/*End added by zyxia for Bug 210743 ,2021.12.16*/

	/*added for Bug 235579,if the callflip the page callee video call,need  update track to the */
	if (page_role == 2) {
		const char* page_caller_name = pbx_builtin_getvar_helper(chan, "PAGING_INTERCOM_CALLER_CHAN_NAME");
		if (!ast_strlen_zero(page_caller_name)) {
			struct ast_channel *page_caller_chan = NULL;
			page_caller_chan = ast_channel_get_by_name(page_caller_name);
			if (page_caller_chan) {
				RAII_VAR(char *, msid, NULL, ast_free);
				msid = ast_channel_get_msid(page_caller_chan, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
				if (!msid) {
					msid = ast_strdup("");
				}
				ast_log(LOG_NOTICE, "channel:%s Msid:%s\n",ast_channel_name(chan),msid);
				ast_indicate_data(new_chan, AST_CONTROL_UPDATE_PAGE_MSID, msid, strlen(msid) + 1);
				ast_channel_unref(page_caller_chan);
			}
		}
	}
	else if (page_role == 1) {//if is video caller callflip,need update video track msid of members
		page_change_caller(chan, new_chan);
		ast_channel_set_is_page_caller(new_chan, 1);
		ast_channel_set_is_page_focus_record(new_chan, ast_channel_get_is_page_focus_record(chan));
		/*Start added by zyxia for Bug 275931*/
		val[0] = '\0';
		ast_cdr_getvar(ast_channel_name(chan), "RECORD_FILES", val, sizeof(val));
		ast_log(LOG_VERBOSE, "Chan:%s, old-cdr-record:%s",
			ast_channel_name(chan), val);
		snprintf(record_files,sizeof(record_files)-1,"%s",val);
		ast_cdr_setvar(ast_channel_name(new_chan), "RECORD_FILES", record_files);
		/*End added by zyxia for Bug 275931*/
	}

	ast_log(LOG_DEBUG, "old_chan:%s, new_chan:%s \n",ast_channel_name(chan),ast_channel_name(new_chan));
	/*added by zyxia for Bug 275047 ,2023.12.6*/
	if (bridge && ast_get_avs_master_value() && bridge->avs_bridge_type == AVS_BRIDGE_TYPE_P2P
		&& (!peer || (peer && strstr(ast_channel_name(peer), "DAHDI") && strstr(ast_channel_name(peer), "IAX"))  || (peer && (strstr(ast_channel_name(peer), "DAHDI")|| strstr(ast_channel_name(peer), "IAX")) && !strcmp(ast_channel_get_avs_address(new_chan),"127.0.0.1")) )){//edit by zyxia for Bug 291750,Bug 297274 
		ast_avs_destroy_multicast(bridge);

		/*new channel callflip to local avs, need set bridge avs address to local,added by zyxia for Bug 291750,Bug 297274  */
		if(peer && (strstr(ast_channel_name(peer), "DAHDI") || strstr(ast_channel_name(peer), "IAX"))&& !strcmp(ast_channel_get_avs_address(new_chan),"127.0.0.1")){
			strcpy(bridge->avs_address, "127.0.0.1");
		}
	}

	/*let the old channel leave the bridege and new channel join the bridge*/
	ast_avs_leave_bridge(chan, bridge, bridge->avs_bridge_id, bridge->avs_bridge_tick);
	ast_avs_join_bridge(new_chan, bridge, bridge->avs_bridge_id, bridge->avs_bridge_tick);

	/*Start added by zyxia for Bug251489,2023.2.23*/
	if(ast_channel_callerintercom_get(chan)){
		page_private_intercom_callflip(chan,new_chan);
	}
	/*End added by zyxia for Bug251489,2023.2.23*/

	/*235945,page announecment callee need mute*/
	if(ast_channel_page_mode_get(chan) ==AST_PAGE_MODE_ANNOUNCE && peer_is_callee) {
		ast_avs_control(new_chan, AVS_CONTROL_MUTE, NULL, 0);
	}


	ast_channel_lock(chan);
	ast_channel_decrease_turn_session_count(chan,ast_channel_get_turn_session_count(chan));//added by zyxia for bug202204,2021.8.27
	ast_channel_unlock(chan);

	bridgeid = ast_channel_get_sca_bridge_uniqueid(chan);
	/*old channel leave*/
	manager_event(EVENT_FLAG_CALL, "BridgeLeave", "BridgeUniqueid: %s\r\n"
								   "BridgeNumChannels: %d\r\n"
								   "Channel: %s\r\n"
								   "Uniqueid: %s\r\n",
				  ast_strlen_zero(bridgeid)?bridge->uniqueid:bridgeid, bridge->num_channels,
				  ast_channel_name(chan),
				  ast_channel_uniqueid(chan)
				  );

	/*Start added by zyxia for Bug202192 ,2021.10.22*/
	//if the callflip is the callee ,need update the callee_dial_string var
	const char *is_callee_chan = pbx_builtin_getvar_helper(chan, "IS_CALLEE");
	if (peer) {
		const char *caller_dial_string = pbx_builtin_getvar_helper(new_chan, "CALLER_DIAL_STRING");
		if (!ast_strlen_zero(caller_dial_string) ) {
			if (!ast_strlen_zero(is_callee_chan) && !strcasecmp(is_callee_chan, "yes"))//is callee,need update peer callee
				pbx_builtin_setvar_helper(peer, "CALLEE_DIAL_STRING",caller_dial_string);
			else
				pbx_builtin_setvar_helper(peer, "__CALLER_DIAL_STRING",caller_dial_string);
		}
	}
	/*End added by zyxia for Bug202192 ,2021.10.22*/

	/*Start added by zyxia for Bug 249764 ,park call 2023.2.7*/
	if (!ast_strlen_zero(is_callee_chan) && !strcasecmp(is_callee_chan, "yes")) {//is callee,need update peer callee
		pbx_builtin_setvar_helper(new_chan, "CALLER_SIDE", NULL);
		const char *chan_caller_dial_string = pbx_builtin_getvar_helper(chan, "CALLER_DIAL_STRING");
		if(!ast_strlen_zero(chan_caller_dial_string))
			pbx_builtin_setvar_helper(new_chan, "__CALLER_DIAL_STRING",chan_caller_dial_string);
	}
	/*End added by zyxia for Bug 249764 ,park call 2023.2.7*/

	/*send update message*/
	ast_channel_set_is_callflip(chan,1); //added by zyxia for Bug 208315,update remote id from callflip


	/*Start added by zyxia for callflip cdr bridge enter, Bug 209297,2021.12.7*/
	ast_bridge_publish_enter_for_callflip(bridge, new_chan, chan);
	/*End added by zyxia for callflip cdr bridge enter,Bug 209297,2021.12.7*/

	//added by zyxia for bug210287 set videocall from callflip channel,2021.12.24
	int is_video_call = ast_channel_get_is_videocall(new_chan);
	int is_hold_state = ast_channel_hold_state(chan);
	//move channel ,will reconfiger bridge,have connected,dont need set,only first connected need set,added by zyxia for 214641,2022.2.9
	ast_channel_set_is_callflip_connected(chan,1);

	/*start added by zyxia for Bug 215084,2022.2.26*/
	if(peer && ast_channel_name(peer) && strstr(ast_channel_name(peer), "PJSIP/trunk")) {
		ast_channel_set_is_callflip_connected(peer,1);
	}
	/*end by zyxia for Bug 215084,2022.2.26*/

	//for bug 211471, Bug 217083.2022.3.10 need set it befor move channel,cdr will use it.
	ast_cdr_setvar(ast_channel_name(chan), "special_action_type", "TRIGGER_CALL_FLIP_TRANSFER");

	ast_log(LOG_DEBUG, "Before swap: chan:%s(%p) new_chan:%s(%p)",
		ast_channel_name(chan), chan,
		ast_channel_name(new_chan), new_chan);

	/*for Bug 234665,calltime limit*/
	struct timeval old_chan_hangup_time;
	struct timeval new_chan_hangup_time;

	old_chan_hangup_time = *ast_channel_whentohangup(chan);
	new_chan_hangup_time = *ast_channel_whentohangup(new_chan);

	/*new channel timelimit is larger than old channel,use old channel,the new channel may is nat turn call*/
	if (!ast_tvzero(old_chan_hangup_time) && !ast_tvzero(new_chan_hangup_time)
		&& ast_tvcmp(new_chan_hangup_time, old_chan_hangup_time) > 0) {
		ast_channel_whentohangup_set(new_chan, ast_channel_whentohangup(chan));
	}
	/*new channel time is null ,need copy from old channel*/
	else if (!ast_tvzero(old_chan_hangup_time) && ast_tvzero(new_chan_hangup_time)){
		ast_channel_whentohangup_set(new_chan, ast_channel_whentohangup(chan));
	}
	/*for Bug 234665,calltime limit*/

	if(ast_channel_get_is_callflip_have_sendupdate(chan))
		ast_channel_set_is_callflip_have_sendupdate(chan,0); //reset it,this will move to new channel

	/*Bug 291708,set bridge id and bridge tick again,otherwise ,the new channel  bridge tick will be set to 0 after move */
	if(bridge) {
		ast_channel_set_avs_bridgeid(chan, bridge->avs_bridge_id);
		ast_channel_set_avs_bridgetick(chan, bridge->avs_bridge_tick);
	}

	if (ast_channel_move(chan, new_chan)) {
		ast_log(LOG_ERROR, "Unable to complete move channel pickup \n");
		astman_send_ack(s, m, "CALL FLIP failed");
		return 0;
	}

	/*Bug 291708, reset bridge id and bridge tick to 0 after do move*/
	ast_channel_set_avs_bridgeid(new_chan, 0);
	ast_channel_set_avs_bridgetick(new_chan, 0);

	reset_bfcp_channel(chan);

	ast_log(LOG_DEBUG, "After swap: chan:%s(%p) new_chan:%s(%p)",
		ast_channel_name(chan), chan,
		ast_channel_name(new_chan), new_chan);

	conf_rec_chan_name = pbx_builtin_getvar_helper(chan, "CONF_RECORDING_CHAN");
	if (!ast_strlen_zero(conf_rec_chan_name) && strstr(conf_rec_chan_name, "CBRec/")) {
		struct ast_channel *conf_rec_chan = ast_channel_get_by_name(conf_rec_chan_name);
		if (conf_rec_chan) {
			const char *parent_chan_name = pbx_builtin_getvar_helper(conf_rec_chan, "PARENT_CHAN");

			ast_log(LOG_DEBUG, "%s's PARENT_CHAN:%s", conf_rec_chan_name, parent_chan_name);
			if (!ast_strlen_zero(parent_chan_name) && !strcmp(parent_chan_name, ast_channel_name(new_chan))) {
				char old_chan_rec[256];
				char new_chan_rec[256];
				ast_cdr_newest_getvar(ast_channel_name(chan), "record_files", old_chan_rec, sizeof(old_chan_rec));
				ast_cdr_newest_getvar(ast_channel_name(new_chan), "record_files", new_chan_rec, sizeof(new_chan_rec));

				ast_log(LOG_DEBUG, "old_chan_rec: %s", ast_strlen_zero(old_chan_rec)?"":old_chan_rec);
				ast_log(LOG_DEBUG, "new_chan_rec: %s", ast_strlen_zero(new_chan_rec)?"":new_chan_rec);

				pbx_builtin_setvar_helper(conf_rec_chan, "PARENT_CHAN", ast_channel_name(chan));
				ast_log(LOG_DEBUG, "UPDATE %s's PARENT_CHAN => %s", conf_rec_chan_name, ast_channel_name(chan));

				if (ast_strlen_zero(old_chan_rec) && !ast_strlen_zero(new_chan_rec)) {
					ast_cdr_newest_setvar(ast_channel_name(chan), "record_files", new_chan_rec);
					ast_log(LOG_DEBUG, "Inherit CDR record_files");
				}
			}

			ast_channel_unref(conf_rec_chan);
		}
	}

	pbx_builtin_setvar_helper(chan, "NEW_FLIP_CHAN", "yes");
	pbx_builtin_setvar_helper(chan, "OLD_FLIP_CHAN", name);

	if(ast_channel_get_is_page_focus_record(chan) && ast_channel_get_is_page_caller(chan) ) {
		if (ast_channel_tech(chan)->indicate) {
			ast_channel_tech(chan)->indicate(chan, AST_CONTROL_PAGE_OPEN_FOCUS_VIDEO, NULL, 0);
		}
	}

	process_update_connected_line_for_callflip(chan,peer,bridge_peer);//added  by zyxia for  Bug 251671

	//added by zyxia for bug210287 set videocall from callflip channel,2021.12.24
	ast_channel_set_is_videocall(chan,is_video_call);
	//for bug 211471, this cdr  dont need push to wave

	if (peer) {
		/*Start added by zyxia for Bug 209511,after callflip,need closed slides first*/
		if (ast_channel_get_send_slides_video(new_chan)) {
			ast_log(LOG_DEBUG, "CallFlip Channel have opend slides,need closed first\n");
			struct ast_bfcp_info bfcp_info;
			memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
			bfcp_info.primitive = BFCP_FLOOR_STATUS;
			bfcp_info.status = BFCP_RELEASED;
			char account_id[AST_MAX_EXTENSION] = { 0 };
			ast_get_extension_from_channel_name(ast_channel_name(new_chan), account_id, sizeof(account_id));
			strcpy(bfcp_info.account_id, account_id);
			if (ast_channel_tech(peer)->indicate) {
				ast_channel_tech(peer)->indicate(peer, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));
			}
			is_slides_ok = 1;
		/*if the callflip is recv slides,send floor granted status to callflip channel to force tell it recv slides*/
		}else if (ast_channel_get_recv_slides_video(new_chan)){
			struct ast_bfcp_info bfcp_info;
			memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
			set_bfcp_account_id(peer, &bfcp_info);
			bfcp_info.primitive = BFCP_FLOOR_STATUS;
			bfcp_info.status = BFCP_GRANTED;

			if (ast_channel_tech(chan)->indicate) {
				ast_channel_tech(chan)->indicate(chan, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));
			}
			is_slides_ok =1;
		}

		struct ast_sip_channel_pvt *old_channel = ast_channel_tech_pvt(chan);
		/*Start added by zyxia for Bug 211311,send hold/unhold,2021.12.27*/
		if(ast_sip_session_get_allow_forward_hold_request()) {
			if(is_hold_state == AST_CONTROL_HOLD && strstr(ast_channel_name(chan), "PJSIP/")) { //remote peer is hold, need send hold to callflip channel
				ast_log(LOG_WARNING, "CallFlip need send hold to channel '%s' \n",ast_channel_name(chan));
				if (old_channel && old_channel->session) {
					ast_channel_hold_state_set(chan,AST_CONTROL_HOLD);
					ast_channel_hold_state_set(peer,AST_CONTROL_UNHOLD);
					if (ast_sip_push_task(old_channel->session->serializer, remote_send_hold, ao2_bump(old_channel->session))) {
						ast_log(LOG_WARNING, "Could not queue task to remotely take session '%s' off hold with endpoint '%s'\n",
						ast_sorcery_object_get_id(old_channel->session), ast_sorcery_object_get_id(old_channel->session->endpoint));
						ast_channel_hold_state_set(chan,AST_CONTROL_UNHOLD);
						ao2_ref(old_channel->session, -1);
					}
				}
			}
			else if(ast_channel_hold_state(peer) ==  AST_CONTROL_HOLD) //callflip channel is hold,need send unhold to remote
			{
				/*Start callflip channel is hold,need send unhold to  stpo music hold.Bug 218527,2022.3.21*/
				ast_queue_unhold(chan);
				ast_queue_frame(chan, &ast_null_frame);
				/*End callflip channel is hold,need send unhold to  stpo music hold.Bug 218527,2022.3.21*/
				ast_channel_hold_state_set(peer,AST_CONTROL_UNHOLD);
				ast_channel_hold_state_set(chan,AST_CONTROL_UNHOLD);
				if (strstr(ast_channel_name(peer), "PJSIP/")) {
					struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(peer);
					if (channel && channel->session) {
						if (!is_slides_ok  && ((ast_channel_get_terminate_type(peer) == AST_TERMINAL_TYPE_WEBRTC)
							|| (ast_channel_get_terminate_type(peer) == AST_TERMINAL_TYPE_GSWAVE_MOBILE) || strstr(ast_channel_name(peer), "PJSIP/trunk"))) {
								ast_log(LOG_DEBUG, "CallFlip Remote is WebRtc ,need update reinvite to remote to open slides\n");
								struct ast_sip_channel_pvt *local_channel = ast_channel_tech_pvt(chan);
								set_pending_media_stream_by_local_session(local_channel->session,channel->session);
								is_slides_ok = 1;
						}
						ast_log(LOG_WARNING, "CallFlip need send unhold to channel '%s' \n",ast_channel_name(peer));
						if (ast_sip_push_task(channel->session->serializer, remote_send_unhold, ao2_bump(channel->session))) {
							ast_log(LOG_WARNING, "Could not queue task to remotely take session '%s' off hold with endpoint '%s'\n",
							ast_sorcery_object_get_id(channel->session), ast_sorcery_object_get_id(channel->session->endpoint));
							ast_channel_hold_state_set(peer,AST_CONTROL_HOLD);
							ao2_ref(channel->session, -1);
						}
					}
				}
			}
		}
		/*End added by zyxia for Bug 211311,send hold/unhold,2021.12.27*/
		/*Start callflip channel is hold,need send unhold to  stpo music hold.Bug 218527,2022.3.21*/
		else if(ast_channel_hold_state(peer) ==  AST_CONTROL_HOLD) {
			ast_queue_unhold(chan);
			ast_queue_frame(chan, &ast_null_frame);
		}
		/*End callflip channel is hold,need send unhold to  stpo music hold.Bug 218527,2022.3.21*/


		/*Start added by zyxia for callflip, slides,send open slides to peer chan.2021.12.10*/
		if (strstr(ast_channel_name(peer), "PJSIP/"))//added by zyxia for Bug 250865,not pjsip channel ,
			channel = ast_channel_tech_pvt(peer);
		if (!is_slides_ok  && strstr(ast_channel_name(peer), "PJSIP/") && ((ast_channel_get_terminate_type(peer) == AST_TERMINAL_TYPE_WEBRTC)
			|| (ast_channel_get_terminate_type(peer) == AST_TERMINAL_TYPE_GSWAVE_MOBILE) || strstr(ast_channel_name(peer), "PJSIP/trunk"))) {
				if (channel && channel->session && old_channel && old_channel->session)
				{
					if(set_pending_media_stream_by_local_session( old_channel->session, channel->session) == 0){
						ast_log(LOG_DEBUG, "CallFlip Remote is WebRtc ,need update reinvite to remote to open slides\n");
						need_reinvite = 1;
					}
				}
		}
		/*End added by zyxia for callflip, slides,send open slides to peer chan.2021.12.10*/

		/*Start added by zyxia for Bug 242037 ,2022.12.21*/
		if (strstr(ast_channel_name(peer), "PJSIP/")&& !ast_channel_get_is_page_caller(peer)
				&& channel && channel->session && old_channel && old_channel->session
				&& change_main_video_direction_by_local_session(old_channel->session ,channel->session) == 1){
					need_reinvite = 1;
					ast_log(LOG_DEBUG, "CallFlip need update reinvite to remote to change main direction\n");
		}


		/*Start added by zyxia for Bug 248301, pending_media need reset  2023.2.7*/
		if( !need_reinvite && channel && channel->session
			&& channel->session->pending_media_state
			&& channel->session->pending_media_state->topology
			&& channel->session->inv_session
			&& channel->session->inv_session->neg
			&& pjmedia_sdp_neg_get_state(channel->session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_DONE) {
			ast_stream_topology_free(channel->session->pending_media_state->topology);
			channel->session->pending_media_state->topology = NULL;
		}
		/*End added by zyxia for Bug 248301, pending_media need reset  2023.2,7*/

		/*start added by zyxia for Bug 201640*/
		if(channel && channel->session && channel->session->is_direct_media && !chan_pjsip_get_direct_enable(chan)) {
			need_reinvite = 1;
			channel->session->is_direct_media = 0;
		}
		/*End added by zyxia for Bug 201640*/

		if(need_reinvite && channel && channel->session) {
			send_reinvite_to_peer(channel->session);
			if (peer && strstr(ast_channel_name(peer), "PJSIP/trunk") && old_channel && old_channel->session) {
				ast_log(LOG_DEBUG, "[session:%p] Need update reinvite to callflip chan by trunk session sdp negation, endpoint:%s\n",
					channel->session, ast_sorcery_object_get_id(channel->session->endpoint));
				channel->session->send_peer_reinvite_for_callflip = 1;
			}
		}
		/*End added by zyxia for Bug 242037 ,2022.12.21*/

		/*Start added by zyxia for Bug 262467,2023.6.13 ,send reinvite to call chan to update slides direction*/
		/*if (need_reinvite  && peer && strstr(ast_channel_name(peer), "PJSIP/trunk")
			&& channel && channel->session && old_channel && old_channel->session)
		{
			usleep(400000);
			if(change_slides_video_direction_by_peer_session( channel->session, old_channel->session)){
				ast_log(LOG_DEBUG, "Trunk call ,need update reinvite to callflip chan by trunk session sdp negation\n");
				send_reinvite_to_peer(old_channel->session);
			}
		}*/
		/*End added by zyxia for Bug 262467,2023.6.13 ,send reinvite to call chan to update slides direction*/

		/*Restart  record the channel ,the record file create again ,also be named by old filename with the new timestamp*/
		if (ast_channel_page_mode_get(chan) != AST_PAGE_MODE_ANNOUNCE  && is_auto_record) {
			//ast_indicate(chan, AST_CONTROL_UPDATE_RECORD_SDP_ATTR);//dont need sent the update sdp
			ast_channel_lock(chan); //edit by zyxia for Bug 203068 ,2021.9.3
			filename = ast_set_record_filename_and_cdr(chan, peer);
			ast_channel_unlock(chan);

			if (filename) {
				ast_log(LOG_ERROR, "bridge chan%s  chan:%s filename:%s\n",ast_channel_name(chan),ast_channel_name(peer),filename);
				
				size_t uuid_len = 0;
				char *uuid = NULL;
				
				uuid_len = 64;
				uuid = ast_alloca(uuid_len);
				
				memset(uuid, 0, uuid_len);
				file_crypt_new_uuid(uuid, uuid_len);
				ast_log(LOG_DEBUG, "[call-flip action] new uuid:%s\n", uuid);
				ast_cdr_newest_setvar(ast_channel_name(chan), "record_files_uuid", uuid);//added by zyxia for Bug 316450

				if (ast_start_mixmonitor(chan, filename, uuid, "b")) {
					ast_verb(4, "Call File restart record chann:%s faild.\n",
					ast_channel_name(chan));
				}
				ast_free(filename);
			}
		}
	}
	/*End added by zyxia for Bug 209511,after callflip,need closed slides first*/

	char *caller_name;
	char *connected_name;
	caller_name = ast_escape_c_alloc(S_COR(ast_channel_caller(chan)->id.name.valid, ast_channel_caller(chan)->id.name.str, ""));
	connected_name = ast_escape_c_alloc(S_COR(ast_channel_connected(new_chan)->id.name.valid, ast_channel_connected(new_chan)->id.name.str, ""));
	manager_event(EVENT_FLAG_CALL, "BridgeEnter", "BridgeUniqueid: %s\r\n"
								   "BridgeTechnology: simple_bridge\r\n"
								   "BridgeCreator: <unknown>\r\n"
								   "BridgeNumChannels: %d\r\n"
								   "Channel: %s\r\n"
								   "ChannelStateDesc: Up\r\n"
								   "CallerIDNum: %s\r\n"
								   "CallerIDName: %s\r\n"
								   "ConnectedLineNum: %s\r\n"
								   "ConnectedLineName: %s\r\n"
								   "Context: from-internal\r\n"
								   "Uniqueid: %s\r\n"
								   "Callee:%s\r\n"
								   "Application: CallFlip\r\n"
								   "Old_Context: %s\r\n",

				  ast_strlen_zero(bridgeid)?bridge->uniqueid:bridgeid, bridge->num_channels,//modify Bug 210035 sca bridge id update */
				  ast_channel_name(chan),
				  S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, ""),
				  caller_name,
				  S_COR(ast_channel_connected(new_chan)->id.number.valid, ast_channel_connected(new_chan)->id.number.str, ""),//connect number use old chan number
				  connected_name,//connect name use old chan's name
				  ast_channel_uniqueid(chan),
				  peer_is_callee? "yes":"no",
				  ast_channel_context(chan)
				  );
	ast_free(caller_name);
	ast_free(connected_name);

	astman_send_ack(s, m, "CALL FLIP succeed");
	return 0;

FAILED:
	astman_send_ack(s, m, "CALL FLIP failed");
	return 0;
}
/*End added by zyxia for 2021.8.13*/

#endif

static void chan_pjsip_session_begin(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);

	if (session->endpoint->media.direct_media.glare_mitigation ==
			AST_SIP_DIRECT_MEDIA_GLARE_MITIGATION_NONE) {
		return;
	}

	datastore = ast_sip_session_alloc_datastore(&direct_media_mitigation_info,
			"direct_media_glare_mitigation");

	if (!datastore) {
		return;
	}

	ast_sip_session_add_datastore(session, datastore);
}

/*! \brief Function called when the session ends */
static void chan_pjsip_session_end(struct ast_sip_session *session)
{
	if (!session->channel) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->line_index) {
		char *user = NULL;

		if (session->aor) {
			user = (char *)ast_sorcery_object_get_id(session->aor);
		} else {
			user = (char *)session->endpoint->aors;
		}
		ast_log(LOG_NOTICE, "chan_pjsip_session_end, sharedline[%s], index[%d], user[%s]", session->endpoint->sca_sharedline, session->line_index, user);
		ast_sip_sca_line_remove_user(session->channel, (char *)session->endpoint->sca_sharedline, session->line_index, user);
		session->line_index = 0;
	}

	/* Add by bfdu for BYE re-INVITE conflict, 2021.1.20 */
	if (session->is_media_change && session->inv_session && session->inv_session->invite_tsx) {
		session->is_media_change = 0;
		ast_queue_control(session->channel, AST_CONTROL_REINVITE_BYE_CONFLICT);
		ast_log(LOG_DEBUG, "[session:%p] Recv bye, but has a reinvite pending, just decline this reinvite request, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	if ((ast_channel_state(session->channel) == AST_STATE_DOWN) && !ast_channel_rings(session->channel) && !strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_NO_ROUTE_DESTINATION);
		ast_log(LOG_NOTICE, "Set NO_ROUTE_DESTINATION to channel[%s]", ast_channel_name(session->channel));
	}
#endif

	chan_pjsip_remove_hold(ast_channel_uniqueid(session->channel));

	/* send reinvite to caller for 30 seconds ,but no response from phone */
	if(ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)
		&& ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ORIGINATED)
		&& ast_channel_state(session->channel) == AST_STATE_DOWN)
	{
		ast_hangup(session->channel);
		return;
	}

	ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
	if (!ast_channel_hangupcause(session->channel) && session->inv_session) {
		int cause = hangup_sip2cause(session->inv_session->cause);

		ast_queue_hangup_with_cause(session->channel, cause);
	} else {
		ast_queue_hangup(session->channel);
	}
}

#ifdef GRANDSTREAM_NETWORKS
static unsigned int reject_call_for_reach_rc_session_count_limit(struct ast_sip_session *session)
{
	unsigned int all_used_count = 0;
	unsigned int total_count = 0;

	total_count = ast_sip_get_total_turn_session_count();
	if (total_count == 0) {
		ast_log(LOG_DEBUG, "[session:%p] Has no turn session count, don't need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	all_used_count = ast_sip_get_used_turn_session_count();
	ast_log(LOG_DEBUG, "[session:%p] total_count:%d all_used_count:%d, endpoint:%s\n",
		session, total_count, all_used_count, ast_sorcery_object_get_id(session->endpoint));

	if (all_used_count >= total_count) {
		ast_log(LOG_DEBUG, "[session:%p] Used RC count large than total count, need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	return 0;
}

static unsigned int reject_call_for_reach_rc_time_limit(struct ast_sip_session *session)
{
	unsigned int nat_limit_time = 0;
	unsigned int total_limit_time = 0;
	unsigned int used_limit_time = 0;
	unsigned int total_month_limit_time = 0;
	unsigned int used_month_limit_time = 0;

	nat_limit_time = ast_sip_get_turn_limit_time();
	if (!nat_limit_time) {
		ast_log(LOG_DEBUG, "[session:%p] Has no limit time, don't need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	total_limit_time = ast_sip_get_turn_day_total_limit_time();
	used_limit_time = ast_cdr_get_nat_used_time();
	if (total_limit_time && (used_limit_time >= total_limit_time * 60)) {
		ast_log(LOG_DEBUG, "[session:%p] used_limit_time is larger than total_limit_time, need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	total_month_limit_time = ast_sip_get_turn_month_total_limit_time();
	used_month_limit_time = ast_cdr_get_nat_month_used_time();
	if (total_month_limit_time && (used_month_limit_time >= total_month_limit_time * 60)) {
		ast_log(LOG_DEBUG, "[session:%p] used_month_limit_time is larger than total_month_limit_time, need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	ast_log(LOG_DEBUG, "[session:%p] total_limit_time:%d used_limit_time:%d total_month_limit_time:%d used_month_limit_time:%d, endpoint:%s\n",
		session, total_limit_time, used_limit_time, total_month_limit_time, used_month_limit_time, ast_sorcery_object_get_id(session->endpoint));

	return 0;
}

static unsigned int is_emergemcy_call(struct ast_sip_session *session)
{
	const char *emergemcy_serive_num = NULL;
	const char *emergemcy_dialer = NULL;

	if (!session || !session->channel) {
		return 0;
	}

	if (session->emergemcy_call) {
		return 1;
	}

	emergemcy_serive_num = pbx_builtin_getvar_helper(session->channel, "EMERGENCY_SERVICE_NUM");
	if (ast_strlen_zero(emergemcy_serive_num)) {
		return 0;
	}

	emergemcy_dialer = pbx_builtin_getvar_helper(session->channel, "EMERGENCY_DIALER");
	if (ast_strlen_zero(emergemcy_dialer)) {
		return 0;
	}

	return 1;
}

struct reject_nat_call_data {
	unsigned int reach_session_count_limit;
	unsigned int reach_time_limit;
	struct ast_sip_session *session;
};

static int play_reject_nat_call_prompt_task(void *data)
{
	RAII_VAR(struct reject_nat_call_data *, task_data, data, ao2_cleanup);
	struct ast_sip_session *session = NULL;

	if (!task_data) {
		return 0;
	}

	session = task_data->session;
	if (!session || !session->channel) {
		return 0;
	}

	if (task_data->reach_session_count_limit) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_TURN_SESSION_LIMITED);
		if (!ast_channel_stream(session->channel)) {
			ast_stream_and_wait(session->channel, AST_NAT_CALL_SESSION_LIMIT_PROMPT, "");
		} else {
			ast_log(LOG_WARNING, "[session:%p] Channel already play other stream file, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	} else if (task_data->reach_time_limit) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_TOTAL_TIME_LIMITED);
		if (!ast_channel_stream(session->channel)) {
			ast_stream_and_wait(session->channel, AST_NAT_CALL_TOTAL_TIME_LIMIT_PROMPT, "");
		} else {
			ast_log(LOG_WARNING, "[session:%p] Channel already play other stream file, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	ast_hangup(session->channel);

	return 0;
}

static void reject_nat_call_data_destroy(void *obj)
{
	struct reject_nat_call_data *data = obj;

	if (!data) {
		return;
	}

	ao2_ref(data->session, -1);

	return;
}

static struct reject_nat_call_data *reject_nat_call_data_alloc(struct ast_sip_session *session,
	unsigned int reach_session_count_limit, unsigned int reach_time_limit)
{
	struct reject_nat_call_data *data = ao2_alloc(sizeof(struct reject_nat_call_data), reject_nat_call_data_destroy);

	if (!data) {
		return NULL;
	}

	data->reach_session_count_limit = reach_session_count_limit;
	data->reach_time_limit = reach_time_limit;

	ao2_ref(session, +1);
	data->session = session;

	return data;
}

static void schedule_task_for_reject_nat_call(struct ast_sip_session *session,
	unsigned int timeout, unsigned int reach_session_count_limit, unsigned int reach_time_limit)
{
	struct reject_nat_call_data *data = NULL;

	if (!session || !session->channel) {
		return;
	}

	if (session->reject_nat_call_task) {
		ast_log(LOG_WARNING, "[session:%p] Has play prompt task before, release first, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_sched_task_cancel(session->reject_nat_call_task);
		ao2_ref(session->reject_nat_call_task, -1);
		session->reject_nat_call_task = NULL;
	}

	data = reject_nat_call_data_alloc(session, reach_session_count_limit, reach_time_limit);
	if (!data) {
		ast_log(LOG_WARNING, "[session:%p] reject_nat_call_data_alloc failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		/* Alloc data faile, hangup call immediately */
		ast_hangup(session->channel);
		return;
	}

	session->reject_nat_call_task = ast_sip_schedule_task(session->serializer, timeout*1000,
		play_reject_nat_call_prompt_task, ast_taskprocessor_name(session->serializer),
		data, AST_SIP_SCHED_TASK_VARIABLE);
	if (!session->reject_nat_call_task) {
		ast_log(LOG_WARNING, "[session:%p] schedule play prompt task failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(data, -1);
		/* schedule task faile, hangup call immediately */
		ast_hangup(session->channel);
	}

	return;
}

static unsigned int update_candidate_complete(struct ast_sip_session *session)
{
	unsigned int i = 0;
	unsigned int res = 0;
	struct ast_sip_session_media_state *media_state = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->inv_session) {
		ast_log(LOG_WARNING, "[session:%p] update_candidate_complete session or inv_session is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	if (!session->pending_media_state || !session->pending_media_state->topology) {
		media_state = session->active_media_state;
	} else {
		media_state = session->pending_media_state;
	}

	if (!media_state || !media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] There's no topology for create sdp for trickle ice, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	for (i = 0; i < ast_stream_topology_get_count(media_state->topology); ++i) {
		session_media = AST_VECTOR_GET(&media_state->sessions, i);
		if (!session_media || !session_media->rtp) {
			ast_log(LOG_WARNING, "[session:%p] Can't get session_media for create trickle ice sdp, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		if (ast_sip_session_media_use_bundle(session)) {
			if (session_media->end_of_candidates) {
				res = 1;
				ast_log(LOG_DEBUG, "[session:%p] session_media end_of_candidates flag is marked, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				break;
			}
		} else {
			res = 1;
			if (!session_media->end_of_candidates) {
				res = 0;
				break;
			}
		}
	}

	return res;
}

static void wait_for_update_candidate(struct ast_sip_session *session)
{
	int to = 0;
	int i =0;
	int timeout = 3*1000; /* Just wait 3s */
	struct ast_frame *f = NULL;
	struct timeval start = ast_tvnow();

	if (!session) {
		return;
	}

	if (!session->support_trickle_ice) {
		ast_log(LOG_DEBUG, "[session:%p] Can't support trickle ICE, don't need wait, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	while ((to = ast_remaining_ms(start, timeout))) {
		i = ast_waitfor(session->channel, 100);
		if (i < 0) {
			ast_log(LOG_NOTICE, "waitfor failed\n");
			break;
		}

		f = ast_read(session->channel);
		if (!f) {
			ast_log(LOG_NOTICE, "ast_read failed\n");
			break;
		}

		ast_frfree(f);

		if (update_candidate_complete(session)) {
			ast_log(LOG_DEBUG, "[session:%p] update_candidate_complete, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			to = 0;
		}
	}

	return;
}

static unsigned int reject_call_for_rc_limit(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	unsigned int reach_session_count_limit = 0;
	unsigned int reach_time_limit = 0;
	pjsip_tx_data *packet = NULL;

	if (!session) {
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] reject_call_for_rc_limit begin, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->ssh_proxy_flag) {
		ast_log(LOG_DEBUG, "[session:%p] Call has no ssh_proxy flag, don't need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (is_emergemcy_call(session)) {
		ast_log(LOG_DEBUG, "[session:%p] It's emergemcy call, don't need reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	reach_session_count_limit = reject_call_for_reach_rc_session_count_limit(session);
	reach_time_limit = reject_call_for_reach_rc_time_limit(session);

	if (!reach_session_count_limit && !reach_time_limit) {
		ast_log(LOG_DEBUG, "[session:%p] Call didn't reach any RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	session->rc_limit_already_processed = 1;

	if ((session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) &&
		(pjsip_inv_answer(session->inv_session, 183, NULL, NULL, &packet) == PJ_SUCCESS)) {
		ast_sip_session_send_response(session, packet);
	} else {
		ast_log(LOG_WARNING, "[session:%p] Can't send 183 for reject call by RC limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->rc_limit_already_processed = 0;
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] reject_call_for_rc_limit reach_session_count_limit:%d, reach_time_limit:%d, endpoint:%s\n",
		session, reach_session_count_limit, reach_time_limit, ast_sorcery_object_get_id(session->endpoint));

	if (!session->channel) {
		ast_log(LOG_WARNING, "[session:%p] Channel should be exsit, it must be something wrong, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->rc_limit_already_processed = 0;
		return 0;
	}

	if (reach_session_count_limit) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_TURN_SESSION_LIMITED);
	} else if (reach_time_limit) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_TOTAL_TIME_LIMITED);
	}

	wait_for_update_candidate(session);

	/* Delay 3s play sound to let the trickle ICE complete */
	schedule_task_for_reject_nat_call(session, 3, reach_session_count_limit, reach_time_limit);

	return 1;
}

/* Added by yfyang, for Bug 307551 */
static int set_sca_barge_call_no_increase(struct ast_sip_session *session)
{
	struct ast_sip_endpoint *endpoint = NULL;

	if (!session || !session->endpoint || !session->channel) {
		return -1;
	}

	endpoint = session->endpoint;
	if (SCA_Active == ast_sip_get_sca_line_status((char *)endpoint->sca_sharedline, session->line_index)) {
		ast_channel_no_increase_call_set(session->channel, 1);
		ast_log(LOG_NOTICE, "set_sca_barge_call_no_increase, chan[%s]", ast_channel_name(session->channel));
		return 0;
	}

	return -1;
}
#endif

/*! \brief Function called when a request is received on the session */
static int chan_pjsip_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);
	struct transport_info_data *transport_data;
	pjsip_tx_data *packet = NULL;

	if (session->channel) {
#ifdef GRANDSTREAM_NETWORKS
		/* Added by yfyang, for Broadsoft Test */
		struct ast_sip_pai_payload *pai_payload = NULL;

		pai_payload = ast_alloca(sizeof(struct ast_sip_pai_payload));
		if (!get_pai_payload(session->channel, rdata, pai_payload)) {
			ast_queue_control_data(session->channel, AST_CONTROL_PAI_INFO, pai_payload, sizeof(struct ast_sip_pai_payload));
		}
#endif
		return 0;
	}

	/* Check for a to-tag to determine if this is a reinvite */
	if (rdata->msg_info.to->tag.slen) {
		/* Weird case. We've received a reinvite but we don't have a channel. The most
		 * typical case for this happening is that a blind transfer fails, and so the
		 * transferer attempts to reinvite himself back into the call. We already got
		 * rid of that channel, and the other side of the call is unrecoverable.
		 *
		 * We treat this as a failure, so our best bet is to just hang this call
		 * up and not create a new channel. Clearing defer_terminate here ensures that
		 * calling ast_sip_session_terminate() can result in a BYE being sent ASAP.
		 */
		session->defer_terminate = 0;
		ast_sip_session_terminate(session, 400);
		return -1;
	}

	datastore = ast_sip_session_alloc_datastore(&transport_info, "transport_info");
	if (!datastore) {
		return -1;
	}

	transport_data = ast_calloc(1, sizeof(*transport_data));
	if (!transport_data) {
		return -1;
	}
	pj_sockaddr_cp(&transport_data->local_addr, &rdata->tp_info.transport->local_addr);
	pj_sockaddr_cp(&transport_data->remote_addr, &rdata->pkt_info.src_addr);
	datastore->data = transport_data;
	ast_sip_session_add_datastore(session, datastore);

	if (!(session->channel = chan_pjsip_new(session, AST_STATE_RING, session->exten, NULL, NULL, NULL, NULL))) {
		if (pjsip_inv_end_session(session->inv_session, 503, NULL, &packet) == PJ_SUCCESS
			&& packet) {
			ast_sip_session_send_response(session, packet);
		}

		ast_log(LOG_ERROR, "Failed to allocate new PJSIP channel on incoming SIP INVITE\n");
		return -1;
	}
	/* channel gets created on incoming request, but we wait to call start
		so other supplements have a chance to run */

#ifdef GRANDSTREAM_NETWORKS
	/* Added by gangwang for Bug 288942*/
	if (!ast_strlen_zero(session->replace_channel_id)) {
		pbx_builtin_setvar_helper(session->channel, "GS_REPLACE", "yes");
	}
	/* Added by yfyang, for ssh proxy process */
	if (session->ssh_proxy_flag) {
		pbx_builtin_setvar_helper(session->channel, "SSH_PROXY_FLAG", "yes");
	}

	/*Added by yfyang, for Bug 179774 */
	if (session->need_orig_callee) {
		pbx_builtin_setvar_helper(session->channel, "__ORIG_CALLEE", session->exten);
	}

	/* Added for Bug75836 */
	set_caller_pai_string(session->channel, rdata);
	ast_sip_set_call_info_string(session->channel, rdata);

	/* Added by yfyang, for IPVT Server application */
	set_ipvt_client_info_string(session->channel, rdata, 0);

	/* Added by yfyang, for Bug 232651, support P-Called-Party-ID */
	set_pcpid_string(session->channel, rdata);

	if (session->core_id >= 0) {
		ast_setup_cc_recall_datastore(session->channel, session->core_id);
		ast_cc_agent_set_interfaces_chanvar(session->channel);
	}

	/* for Bug 102894 */
	if (session->endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_PRIVATE) {
		pbx_builtin_setvar_helper(session->channel, "SCA_PRIV_CALLER", "yes");
	}

	/* for Bug 103746 */
	pbx_builtin_setvar_helper(session->channel, "CALLER_SIDE", "yes");

	/* add by pengwang */
	if (session->blocking_collect_calls == 1) {
		pbx_builtin_setvar_helper(session->channel, "ENABLE_BLOCKING_COLLECT_CALLS", "yes");
	} else {
		pbx_builtin_setvar_helper(session->channel, "ENABLE_BLOCKING_COLLECT_CALLS", "no");
	}

	ast_channel_set_reject_call(session->channel, session->reject_call);

	/* Added by xxcui for Bug 87187 */
	set_caller_dial_string(session, rdata);

	/* Added by yfyang for Bug 93558 */
	if (0 != strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		char  family[80] = {0};
		char  data[80]   = {0};
		char  group[80]  = {0};

		snprintf(family, sizeof(family), "AAUSER/%s", ast_sorcery_object_get_id(session->endpoint));
		ast_ucm_db_get(family, "LIMITIME", data, sizeof(data));

		if (strlen(data)) {
			pbx_builtin_setvar_helper(session->channel, "CALLER_DUR_LIMIT", data);
			ast_log(LOG_NOTICE, "user[%s], limitime[%s]!\n", family, data);
		}

		/* Added by yfyang, for Bug 173103 */
		snprintf(group, sizeof(group), "OUT_EXT_%s", ast_sorcery_object_get_id(session->endpoint));
		pbx_builtin_setvar_helper(session->channel, "OUTBOUND_GROUP_EXT", group);
	}

	/* Modify by bfdu for Bug 152952 */
	set_call_type(session, rdata); /* call from trunk, record call type*/

	/*For bug 181104, jzha, 2020-11-16*/
	get_calltype_params_of_from(session, rdata);

	/* added by zgmo for bug183396 */
	get_useragent_from_hdr(session, rdata);

	/*call from trunk conf, check is conference inviteby zyxia 2022.12.1*/
	ast_sip_session_trunk_conf_need_subscribe(session, rdata);
	if(session->is_trunk_conf_invite){
		char tmp_action[256] = {0};
		snprintf(tmp_action, sizeof(tmp_action), "MULTIMEDIAMEETING[%s]", ast_channel_caller_num(session->channel));
		ast_cdr_setvar(ast_channel_name(session->channel), "action_type", tmp_action);
	}

	ast_sip_session_call_to_emergency_call(session,rdata);//added  by zyxia for Bug 265588,2023.7.19

	if (reject_call_for_rc_limit(session, rdata)) {
		return -1;
	}

	/* Added by yfyang, for Bug 307551 */
	if (session->line_index > 0) {
		set_sca_barge_call_no_increase(session);
	}
#endif
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
struct sip_fastprov_context {
	int status;
	int addition;
	int port;

	struct ast_channel *chan;
	char *addr;	/* ip */

	/*
	 *  DO NOT Free
	 */
	const char *profcode;
	const char *user_agent;
};

/*
 *	error_code:
 *	1001 - Invalid profile-code, 'additional' will show number of try left
 *	1002 - The requested Device is locked for too much fail attempt
 *		   'additional' will show how many seconds before it's unlocked
 *	1003 - Profile-Code is valid but out-of-date
 *	1004 - Profile-Code is valid but exceed it's use-limits
 */
#define FASTPROV_SUCCESS								"0"
#define FASTPROV_ERR_CODE_INVALID						"1001"
#define FASTPROV_ERR_DEV_LOCK							"1002"
#define FASTPROV_ERR_CODE_OUT_OF_DATE					"1003"
#define FASTPROV_ERR_CODE_EXCEED_USELIMITS				"1004"
#define FASTPROV_ERR_CODE_DEV_NOT_FOUND					"1010"

enum ast_sip_fastprov_stat {
	AST_SIP_FASTPROV_INIT = 0,
	AST_SIP_FASTPROV_REQUESTED,
	AST_SIP_FASTPROV_RESPONSED_SUCCESS,
	AST_SIP_FASTPROV_RESPONSED_ERR_CODE_INVALID,
	AST_SIP_FASTPROV_RESPONSED_ERR_DEV_LOCK,
	AST_SIP_FASTPROV_RESPONSED_ERR_CODE_OUT_OF_DATE,
	AST_SIP_FASTPROV_RESPONSED_ERR_CODE_EXCEED_USELIMITS,

	AST_SIP_FASTPROV_FAILED
};

static void stasis_fastprov_resp_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	const char *msg_type, *msg_data, *msg_addr, *msg_ua;
	const char *msg_result, *msg_addition;

	struct ast_json *json_payload;
	struct ast_json_payload *msg_payload = NULL;
	struct sip_fastprov_context *fastprov_ctx = data;

	if (stasis_message_type(message) != ast_channel_fastprov_type()) {
		return;
	}

	msg_payload = stasis_message_data(message);
	if (!fastprov_ctx || !msg_payload || !msg_payload->json) {
		ast_log(LOG_ERROR, "Fatal error");
		return;
	}

	json_payload = msg_payload->json;

	msg_type = ast_json_string_get(ast_json_object_get(json_payload, "Type"));
	msg_data = ast_json_string_get(ast_json_object_get(json_payload, "Data"));
	msg_addr = ast_json_string_get(ast_json_object_get(json_payload, "Addr"));
	msg_ua = ast_json_string_get(ast_json_object_get(json_payload, "UserAgent"));

	msg_result = ast_json_string_get(ast_json_object_get(json_payload, "Result"));
	msg_addition = ast_json_string_get(ast_json_object_get(json_payload, "Addition"));

#if 0
	ast_log(LOG_DEBUG, "type:%s, data:%s, addr:%s, ua:%s, result:%s, addition:%s",
		S_OR(msg_type, "null"), S_OR(msg_data, "null"), S_OR(msg_addr, "null"),
		S_OR(msg_ua, "null"), S_OR(msg_result, "null"), S_OR(msg_addition, "null"));
	ast_log(LOG_DEBUG, "json:{%s}", S_OR(ast_json_string_get(json_payload), ""));
#endif
	if (ast_strlen_zero(msg_type) || strcmp(msg_type, "Profcode")
		|| ast_strlen_zero(msg_data) || ast_strlen_zero(msg_addr)
		|| ast_strlen_zero(msg_ua)) {
		ast_log(LOG_WARNING, "Invalid message");
		return;
	}

	if (strcmp(msg_data, fastprov_ctx->profcode)
		|| strcmp(msg_addr, fastprov_ctx->addr)
		|| strcmp(msg_ua, fastprov_ctx->user_agent)) {
		ast_log(LOG_DEBUG, "Not my message");
		goto END;
	}

	if (ast_strlen_zero(msg_result))
	{
		fastprov_ctx->status = AST_SIP_FASTPROV_FAILED;
		goto END;
	}

	fastprov_ctx->addition = 0;
	if (!strcmp(msg_result, FASTPROV_SUCCESS)) {
		fastprov_ctx->status = AST_SIP_FASTPROV_RESPONSED_SUCCESS;
	} else if (!strcmp(msg_result, FASTPROV_ERR_CODE_INVALID)
		&& !ast_strlen_zero(msg_addition)) {
		fastprov_ctx->status = AST_SIP_FASTPROV_RESPONSED_ERR_CODE_INVALID;
		fastprov_ctx->addition = atoi(msg_addition);
	} else if (!strcmp(msg_result, FASTPROV_ERR_DEV_LOCK)
		&& !ast_strlen_zero(msg_addition)) {
		fastprov_ctx->status = AST_SIP_FASTPROV_RESPONSED_ERR_DEV_LOCK;
		fastprov_ctx->addition = atoi(msg_addition);
	} else if (!strcmp(msg_result, FASTPROV_ERR_CODE_OUT_OF_DATE)) {
		fastprov_ctx->status = AST_SIP_FASTPROV_RESPONSED_ERR_CODE_OUT_OF_DATE;
	} else if (!strcmp(msg_result, FASTPROV_ERR_CODE_EXCEED_USELIMITS)) {
		fastprov_ctx->status = AST_SIP_FASTPROV_RESPONSED_ERR_CODE_EXCEED_USELIMITS;
	} else {
		fastprov_ctx->status = AST_SIP_FASTPROV_FAILED;
	}

	ast_log(LOG_DEBUG, "stasis_fastprov_resp_cb | status:%d, addition:%d",
									fastprov_ctx->status, fastprov_ctx->addition);
END:
	return;
}

static int get_addr_from_rdata(struct pjsip_rx_data *in_rdata, char **out_addr, int *out_port)
{
	int alloc_size;
	pjsip_via_hdr *via_hdr;

	via_hdr = (pjsip_via_hdr*) pjsip_msg_find_hdr(in_rdata->msg_info.msg, PJSIP_H_VIA, NULL);
	if (!via_hdr) {
		ast_log(LOG_DEBUG, "Fatal error");
		return -1;
	}

	*out_port=via_hdr->sent_by.port;

	alloc_size = pj_strlen(&via_hdr->sent_by.host) + 1;
	*out_addr = ast_malloc(alloc_size);
	if (!*out_addr) {
		ast_log(LOG_DEBUG, "Fatal error");
		return -1;
	}

	ast_copy_pj_str(*out_addr, &via_hdr->sent_by.host, alloc_size);
	return 0;
}

#define VOICE_PROMPT_PROFCODE_CORRECT				"profcode-correct"
#define VOICE_PROMPT_PROFCODE_INVALID				"profcode-invalid"
#define VOICE_PROMPT_PROFCODE_CHANCES_LEFT			"profcode-chances-left"
#define VOICE_PROMPT_PROFCODE_DEVLOCK_PRE			"profcode-devlock-pre"
#define VOICE_PROMPT_PROFCODE_DEVLOCK_POST			"profcode-devlock-post"
#define VOICE_PROMPT_PROFCODE_EXPIRED				"profcode-expired"
#define VOICE_PROMPT_PROFCODE_EXCEED_USELIMITS		"profcode-exceed-uselimits"
#define VOICE_PROMPT_PROFCODE_DEPLOY_FAIL			"profcode-deploy-failed"

static int prompt_for_profilecode(struct sip_fastprov_context *fastprov_ctx)
{
	int ret = 0;
	ast_log(LOG_DEBUG, "Enter prompt_for_profilecode(%s)", ast_channel_name(fastprov_ctx->chan));

	if (AST_SIP_FASTPROV_RESPONSED_SUCCESS == fastprov_ctx->status) {
		ast_log(LOG_DEBUG, "ProfCode CORRECT");
		ret  = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_CORRECT, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		}

		goto END;
	}

	if (AST_SIP_FASTPROV_RESPONSED_ERR_CODE_INVALID == fastprov_ctx->status && fastprov_ctx->addition) {
		ast_log(LOG_DEBUG, "ProfCode Invalid");

		ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_INVALID, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		} else {
			goto END;
		}

		ast_say_digits(fastprov_ctx->chan, fastprov_ctx->addition, "", ast_channel_language(fastprov_ctx->chan));
		ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_CHANCES_LEFT, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		}

		goto END;
	}

	if (AST_SIP_FASTPROV_RESPONSED_ERR_DEV_LOCK == fastprov_ctx->status && fastprov_ctx->addition) {
		ast_log(LOG_DEBUG, "Device locked");
		ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_DEVLOCK_PRE, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		} else {
			goto END;
		}

		ast_say_digits(fastprov_ctx->chan, fastprov_ctx->addition/60 + 1, "", ast_channel_language(fastprov_ctx->chan));
		ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_DEVLOCK_POST, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		}

		goto END;
	}

	if (AST_SIP_FASTPROV_RESPONSED_ERR_CODE_OUT_OF_DATE == fastprov_ctx->status) {
		ast_log(LOG_DEBUG, "ProfCode expired");
		ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_EXPIRED, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		}

		goto END;
	}

	if (AST_SIP_FASTPROV_RESPONSED_ERR_CODE_EXCEED_USELIMITS == fastprov_ctx->status) {
		ast_log(LOG_DEBUG, "ProfCode exceed use limits");
		ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_EXCEED_USELIMITS, ast_channel_language(fastprov_ctx->chan));
		if (!ret) {
			ast_waitstream(fastprov_ctx->chan, "");
		}

		goto END;
	}

	/*
	 *  Common failure
	 */
	ast_log(LOG_DEBUG, "Endpoint Device Not Found");
	ret = ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_PROFCODE_DEPLOY_FAIL, ast_channel_language(fastprov_ctx->chan));
	if (!ret) {
		ast_waitstream(fastprov_ctx->chan, "");
	}

END:
	ast_stopstream(fastprov_ctx->chan);

	ast_log(LOG_DEBUG, "Exit prompt_for_profilecode");
	return 0;
}

static void *process_fastprov_request(void *data)
{
	int ret;
	int timeout_loops;
	struct sip_fastprov_context *fastprov_ctx = data;
	struct stasis_subscription *cur_subscription = NULL;

	ast_log(LOG_DEBUG, "Enter process_fastprov_request|chan:%p, profilecode:%s, peer:[%s:%d]",
		fastprov_ctx->chan, fastprov_ctx->profcode, fastprov_ctx->addr, fastprov_ctx->port);

#if 0
	ast_streamfile(fastprov_ctx->chan, VOICE_PROMPT_IDCODE_CORRECT, ast_channel_language(fastprov_ctx->chan));
	ast_waitstream(fastprov_ctx->chan, "");

	goto END;
#endif
	cur_subscription = stasis_subscribe(ast_manager_get_topic(), stasis_fastprov_resp_cb, fastprov_ctx);
	if (!cur_subscription) {
		ast_log(LOG_ERROR, "Fatal error");
		goto END;
	}

	/*
	 *  Push message to CGI through AMI and get the response from AMI
	 */
	ret = manager_event(EVENT_FLAG_CALL, "SipFastProv",
		"Method: Request\r\n"
		"Type: Profcode\r\n"
		"Data: %s\r\n"		/* Profile-code */
		"Addr: %s\r\n"		/* IP */
		"UserAgent: %s\r\n", fastprov_ctx->profcode, fastprov_ctx->addr, fastprov_ctx->user_agent);
	if (ret) {
		ast_log(LOG_ERROR, "Publis manager-event failed");
		goto END;
	}

	fastprov_ctx->status = AST_SIP_FASTPROV_REQUESTED;
	timeout_loops = 50; /* Wait Max 5 secs */

	while(timeout_loops--) {
		if (fastprov_ctx->status <= AST_SIP_FASTPROV_REQUESTED)
		{
			/*
			 *  Keep waiting
			 */
			usleep(100*1000);	/* sleep 100ms */
			continue;
		}

		ast_log(LOG_DEBUG, "Receive response for SipFastProvReq");
		prompt_for_profilecode(fastprov_ctx);
		break;
	}

	if (fastprov_ctx->status <= AST_SIP_FASTPROV_REQUESTED) {
		ast_log(LOG_WARNING, "SipFastProvReq TIMEOUT");
	}

END:
	if (cur_subscription) cur_subscription = stasis_unsubscribe(cur_subscription);
	ast_hangup(fastprov_ctx->chan);
	ast_free(fastprov_ctx->addr);
	ast_free(fastprov_ctx);

	return NULL;
}

static int chan_pjsip_incoming_request_fastprov(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	int ret;
	int peer_port;
	char *peer_addr = NULL;
	pthread_t t;

	pjsip_tx_data *packet = NULL;
	struct sip_fastprov_context *fastprov_ctx = NULL;

	if (!session->is_fastprov) {
		//ast_log(LOG_DEBUG, "Not fastprov session, skip");
		return 0;
	}

	ret = get_addr_from_rdata(rdata, &peer_addr, &peer_port);
	if (ret || !peer_addr) {
		ast_log(LOG_ERROR, "Fatal error");
		goto FAILED;
	}

	ast_log(LOG_DEBUG, "Fastprov profile-code(%s) registration coming from [%s|%d]. User-agent:%s",
		session->exten, peer_addr, peer_port, session->user_agent);
	if (PJ_SUCCESS == pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet)
		&& packet) {
		ast_sip_session_send_response(session, packet);
	} else {
		ast_log(LOG_ERROR, "Fatal error");
		goto FAILED;
	}

	/*
	 *  Pre-setup environment
	 */
	fastprov_ctx = ast_malloc(sizeof(struct sip_fastprov_context));
	if (!fastprov_ctx) {
		ast_log(LOG_ERROR, "Insufficient memory");
		goto FAILED;
	}

	fastprov_ctx->chan = session->channel;
	fastprov_ctx->addr = peer_addr;
	fastprov_ctx->port = peer_port;
	fastprov_ctx->user_agent = session->user_agent;
	fastprov_ctx->profcode = session->exten;

	fastprov_ctx->status = AST_SIP_FASTPROV_INIT;

	if (ast_pthread_create_detached(&t, NULL, process_fastprov_request, fastprov_ctx)) {
		ast_log(LOG_ERROR, "Fatal error");
		goto FAILED;
	}

	return 1;

FAILED:
	ast_hangup(session->channel);
	if (peer_addr) ast_free(peer_addr);
	if (fastprov_ctx) ast_free(fastprov_ctx);

	return 1;
}
#endif

static int call_pickup_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	struct ast_features_pickup_config *pickup_cfg;
	struct ast_channel *chan;

	/* Check for a to-tag to determine if this is a reinvite */
	if (rdata->msg_info.to->tag.slen) {
		/* We don't care about reinvites */
		return 0;
	}

	pickup_cfg = ast_get_chan_features_pickup_config(session->channel);
	if (!pickup_cfg) {
		ast_log(LOG_ERROR, "Unable to retrieve pickup configuration options. Unable to detect call pickup extension.\n");
		return 0;
	}

	if (strcmp(session->exten, pickup_cfg->pickupexten)) {
		ao2_ref(pickup_cfg, -1);
		return 0;
	}
	ao2_ref(pickup_cfg, -1);

	/* We can't directly use session->channel because the pickup operation will cause a masquerade to occur,
	 * changing the channel pointer in session to a different channel. To ensure we work on the right channel
	 * we store a pointer locally before we begin and keep a reference so it remains valid no matter what.
	 */
	chan = ast_channel_ref(session->channel);
	if (ast_pickup_call(chan)) {
		ast_channel_hangupcause_set(chan, AST_CAUSE_CALL_REJECTED);
	} else {
#ifdef GRANDSTREAM_NETWORKS
		/*jklou Bug 41546 - Pickup Group:It will have a missed call record when pickup a call*/
		if (ast_channel_hangupcause(chan) != AST_CAUSE_ANSWERED_ELSEWHERE){
			ast_channel_hangupcause_set(chan, AST_CAUSE_NORMAL_CLEARING);
		}
#else
		ast_channel_hangupcause_set(chan, AST_CAUSE_NORMAL_CLEARING);
#endif
	}
	/* A hangup always occurs because the pickup operation will have either failed resulting in the call
	 * needing to be hung up OR the pickup operation was a success and the channel we now have is actually
	 * the channel that was replaced, which should be hung up since it is literally in limbo not connected
	 * to anything at all.
	 */
	ast_hangup(chan);
	ast_channel_unref(chan);

	return 1;
}

static struct ast_sip_session_supplement call_pickup_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_LAST - 1,
	.incoming_request = call_pickup_incoming_request,
};

static int pbx_start_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	int res;

	/* Check for a to-tag to determine if this is a reinvite */
	if (rdata->msg_info.to->tag.slen) {
		/* We don't care about reinvites */
		return 0;
	}

	res = ast_pbx_start(session->channel);

	switch (res) {
	case AST_PBX_FAILED:
		ast_log(LOG_WARNING, "Failed to start PBX ;(\n");
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_SWITCH_CONGESTION);
		ast_hangup(session->channel);
		break;
	case AST_PBX_CALL_LIMIT:
		ast_log(LOG_WARNING, "Failed to start PBX (call limit reached) \n");
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_SWITCH_CONGESTION);
		ast_hangup(session->channel);
		break;
	case AST_PBX_SUCCESS:
	default:
		break;
	}

	ast_debug(3, "Started PBX on new PJSIP channel %s\n", ast_channel_name(session->channel));

	return (res == AST_PBX_SUCCESS) ? 0 : -1;
}

/*BEIGN: Add by weidu ,for 196889*/
static int ast_incoming_without_sdp(struct pjsip_rx_data *rdata)
{
	pjsip_rdata_sdp_info *sdp_info = pjsip_rdata_get_sdp_info(rdata);
	if (!(sdp_info && sdp_info->sdp)){
		return 1;
	}
	return 0;
}
/*END: Add by weidu*/

static struct ast_sip_session_supplement pbx_start_supplement = {
	.method = "INVITE",
	.priority = AST_SIP_SUPPLEMENT_PRIORITY_LAST,
	.incoming_request = pbx_start_incoming_request,
};

#ifdef GRANDSTREAM_NETWORKS
static void queue_camera_control_frame_for_failed(struct ast_sip_session *session)
{
	struct ast_camera_control_info camera_control_info;

	memset(&camera_control_info, 0, sizeof(struct ast_camera_control_info));

	switch (session->camera_control_action) {
		case AST_SIP_CAMERA_CONTROL_ACTION_OPEN:
			camera_control_info.action = CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA_STATUS;
			break;
		case AST_SIP_CAMERA_CONTROL_ACTION_CLOSE:
			camera_control_info.action = CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA_STATUS;
			break;
		default:
			break;
	}

	camera_control_info.status = CAMERA_CONTROL_STATUS_FAILED;
	strcpy(camera_control_info.failed_reason, "Peer send non-2XX response, decline request");

	ast_queue_control_data(session->channel, AST_CONTROL_CAMERA_CONTROL, &camera_control_info, sizeof(camera_control_info));

	return;
}

/*start added by zyxia for trunk info cei ,2022.11.17*/
static void send_trunk_conf_cei_subscribe(struct ast_sip_session * session)
{
	if (!session )
		return;

	ast_log(LOG_DEBUG, "send_trunk_conf_cei_subscribe\n");

	if(ast_strlen_zero(session->trunk_conf_num)) {
		ast_log(LOG_WARNING, "send_trunk_conf_cei_subscribe trunk conf num is null\n");
		return;
	}

	//if is trunk conf, need subscribe
	if(session->channel && ast_channel_name(session->channel) && strcasestr(ast_channel_name(session->channel),"PJSIP/trunk")) {
		const char *trunk_index= pbx_builtin_getvar_helper(session->channel, "TRUNK_INDEX");
		if(!ast_strlen_zero(trunk_index)){
			char local_virtual_conf[512] = {0};
			char family[64] = {0};
			char trunkName[128] = {0};
			ast_log(LOG_DEBUG, "send_trunk_conf_cei_subscribe:%s,%s\n",session->trunk_conf_num,ast_channel_name(session->channel));
			snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(trunk_index));
			ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
			snprintf(local_virtual_conf, sizeof(local_virtual_conf) - 1, "%s_%s", trunkName,session->trunk_conf_num);
			sip_exteral_trunk_conf_sub(session,local_virtual_conf);//send subscribe to trunk
			session->had_subscribe_trunk_cei = 1;
			ast_channel_set_virtual_conf_num(session->channel,local_virtual_conf);//added by zyxia for Bug 246318
		}
	}

	return;
}

static void check_if_outbound_to_trunk_conf(struct ast_sip_session *session,struct pjsip_rx_data *rdata)
{
	if (!session  || !rdata)
		return;

	if(ast_sip_session_trunk_conf_need_subscribe(session, rdata)) {
		send_trunk_conf_cei_subscribe(session);
	}
}

/*End added by zyxia for trunk info cei ,2022.11.17*/

/* Add by bfdu for Bug 260277, 2023.6.1 */
static int rdata_is_100_rel(struct pjsip_rx_data *rdata)
{
	const pj_str_t RSeq_hdr = { "RSeq", 4 };
	pjsip_generic_string_hdr *rseq_hdr = NULL;
	pjsip_msg *msg = NULL;

	if (!rdata) {
		return 0;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		return 0;
	}

	/* Get the RSeq header */
	rseq_hdr = (pjsip_generic_string_hdr*)pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &RSeq_hdr, NULL);
	if (rseq_hdr == NULL) {
		return 0;
	}

	return 1;
}

/*added by zyxia for report queue answer state,2023.6.2*/
static int send_real_answered_status(void *data){
	struct ast_sip_session *session = data;

	if (!session) {
		return 0;
	}

	if (!session->channel) {
		ast_log(LOG_DEBUG, "[session:%p] send_real_answered_status, channel is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
		return 0;
	}

	ast_log(LOG_DEBUG, "send_real_answered_status\n");
	ast_queue_control(session->channel, AST_CONTROL_REAL_ANSWERED);
	ao2_cleanup(session);
	return 0;
}

static void schedule_task_for_real_answered(struct ast_sip_session *session, unsigned int timeout)
{
	if (!session || !session->inv_session) {
		ast_log(LOG_ERROR, "schedule_task_for_real_answered param is NULL\n");
		return;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		ast_log(LOG_DEBUG, "[session:%p] schedule_task_for_real_answered inv state is CONFIRMED, don't need schedule, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->send_real_answered) {
		ast_log(LOG_WARNING, "[session:%p] schedule_task_for_real_answered already schedule, don't schedule again, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ao2_ref(session, +1);

	session->send_real_answered = ast_sip_schedule_task(session->serializer, timeout,
		send_real_answered_status, ast_taskprocessor_name(session->serializer),
		session, AST_SIP_SCHED_TASK_VARIABLE);
	if (!session->send_real_answered) {
		ast_log(LOG_WARNING, "[session:%p] schedule_task_for_real_answered task failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}
/*End added by zyxia for report queue answer state,2023.6.2*/

#endif

/*! \brief Function called when a response is received on the session */
static void chan_pjsip_incoming_response(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	struct pjsip_status_line status = rdata->msg_info.msg->line.status;
	struct ast_control_pvt_cause_code *cause_code;
	int data_size = sizeof(*cause_code);
#ifdef GRANDSTREAM_NETWORKS
	const pj_str_t mcb_resp_name = { "X-Grandstream-Mcb-Resp", 22 };
	const pj_str_t call_type_name = { "X-Grandstream-Call-Type", 23 };
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t value;
	const char *cei_support = NULL;
	const char *conf_room = NULL;
	struct ast_sip_pai_payload *pai_payload = NULL;
#endif

	if (!session->channel) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (status.code >= 200) {
		pbx_builtin_setvar_helper(session->channel, "__CALL_INFO", NULL);
		pbx_builtin_setvar_helper(session->channel, "__CALL_INFO_MSG", NULL);

		/* Added by yfyang, for IPVT Server application */
		pbx_builtin_setvar_helper(session->channel, "__IPVT_CLIENT_INFO", NULL);
		pbx_builtin_setvar_helper(session->channel, "__IPVT_CLIENT_INFO_MSG", NULL);
	}
#endif

	/* Build and send the tech-specific cause information */
	/* size of the string making up the cause code is "SIP " number + " " + reason length */
	data_size += 4 + 4 + pj_strlen(&status.reason);
	cause_code = ast_alloca(data_size);
	memset(cause_code, 0, data_size);

	ast_copy_string(cause_code->chan_name, ast_channel_name(session->channel), AST_CHANNEL_NAME);

	snprintf(cause_code->code, data_size - sizeof(*cause_code) + 1, "SIP %d %.*s", status.code,
	(int) pj_strlen(&status.reason), pj_strbuf(&status.reason));

	cause_code->ast_cause = hangup_sip2cause(status.code);
	ast_queue_control_data(session->channel, AST_CONTROL_PVT_CAUSE_CODE, cause_code, data_size);
	ast_channel_hangupcause_hash_set(session->channel, cause_code, data_size);

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_session_end_if_deferred(session);
	ast_sip_session_defer_termination_cancel(session);
	ast_sip_session_set_ucm_trunk(session, rdata);
	ast_sip_session_set_trunk_info(session, rdata);
#endif

	switch (status.code) {
	case 180:
		if (ast_incoming_without_sdp(rdata)){  /*Mod by weidu ,for 196889*/
#ifdef GRANDSTREAM_NETWORKS
			/*start add by zyxia for wake up gswave,if receive 180 from gswave,*/
			if (!session->has_add_gs_call_count) {
				ast_sip_set_used_gs_call_count(session,rdata,1);
			}
			/*end add by zyxia for wake up gswave,if receive 180 from gswave */

			/* Added by yfyang, for Broadsoft Test */
			pai_payload = ast_alloca(sizeof(struct ast_sip_pai_payload));
			if (!get_pai_payload(session->channel, rdata, pai_payload)) {
				ast_queue_control_data(session->channel, AST_CONTROL_PAI_INFO, pai_payload, sizeof(struct ast_sip_pai_payload));
			}

			/* Added by yfyang, for Bug 311191 */
			if (ast_channel_respond_18x_flag(session->channel) && session->endpoint->ignore_180_after_183) {
				ast_log(LOG_NOTICE, "ignore 180 after 183.");
				break;
			}
#endif

			ast_queue_control(session->channel, AST_CONTROL_RINGING);
			ast_channel_lock(session->channel);
			if (ast_channel_state(session->channel) != AST_STATE_UP) {
				ast_setstate(session->channel, AST_STATE_RINGING);
			}
			ast_channel_unlock(session->channel);
			break;
		}
	case 183:

#ifdef GRANDSTREAM_NETWORKS
		/*start add by zyxia for wake up gswave,if receive 183 from gswave,*/
		if (!session->has_dec_gs_call_count) {
			ast_sip_set_used_gs_call_count(session,rdata,1);
		}
		/*end add by zyxia for wake up gswave,if receive 183 from gswave,*/

		/* Added by yfyang, for Broadsoft Test */
		pai_payload = ast_alloca(sizeof(struct ast_sip_pai_payload));
		if (!get_pai_payload(session->channel, rdata, pai_payload)) {
			ast_queue_control_data(session->channel, AST_CONTROL_PAI_INFO, pai_payload, sizeof(struct ast_sip_pai_payload));
		}

		if (ast_incoming_without_sdp(rdata) && !session->endpoint->ignore_180_after_183 && !rdata_is_100_rel(rdata)) {
			ast_queue_control(session->channel, AST_CONTROL_RINGING);
		} else {
			ast_channel_recv_early_media_rsp_set(session->channel, 1);
			ast_queue_control(session->channel, AST_CONTROL_PROGRESS);
		}

		ast_channel_respond_18x_flag_set(session->channel, 1);
#else
		ast_queue_control(session->channel, AST_CONTROL_PROGRESS);
#endif

#ifdef GRANDSTREAM_NETWORKS
		/* Add by jzha for Bug 96183, 2017.11.28
		 * when enable Record Prompt, make peer trunk calls
		 * the Active Calls of the caller's UCM should display ringing
		 */
		ast_channel_lock(session->channel);
		if (ast_channel_state(session->channel) != AST_STATE_UP) {
			ast_setstate(session->channel, AST_STATE_RINGING);
		}
		ast_channel_unlock(session->channel);

#endif
		break;
	case 200:
		/* original caller side goto pbx_thread */
		if(ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)
			&& ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ORIGINATED)
			&& !ast_channel_already_recovery_get(session->channel))
		//	&& !strcmp(ast_channel_recovery_linkedid(session->channel), ast_channel_recovery_uniqueid(session->channel)))
		{
			ast_channel_already_recovery_set(session->channel, 1);
			chan_pjsip_recovery_throw_pbx(session->channel);
		}
#ifdef GRANDSTREAM_NETWORKS
		hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &mcb_resp_name, NULL);
		if (hdr) {
			pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);
			ast_log(LOG_DEBUG, "[TK] X-Grandstream-Mcb-Resp: %s, len: %ld",
								 pj_strbuf(&value), pj_strlen(&value));
			if (pj_strlen(&value) > 0) {
				sip_setmcbresp(session->channel, pj_strbuf(&value));
			}
		}

		hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &call_type_name, NULL);
		if (hdr) {
			pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);

			if (pj_strlen(&value) > 0) {
				sip_setcalltype(session->channel, pj_strbuf(&value));
			}
		}

		get_server_from_hdr(session, rdata);

		set_callee_gds_call(session, rdata);
		set_callee_gsc_call(session, rdata);

		ast_sip_session_set_device_ua(session, rdata, PJSIP_OTHER_METHOD);

		ast_sip_set_call_info_string(session->channel, rdata);  /* to record the Call-Info header */

		/* Added by yfyang, for IPVT Server application */
		set_ipvt_client_info_string(session->channel, rdata, 1);

		//added by zyxia for call to trunk conf ,receive 200 ok from remote  200 ok
		check_if_outbound_to_trunk_conf(session,rdata);
		/* Added by yfyang, for Broadsoft Test */
		pai_payload = ast_alloca(sizeof(struct ast_sip_pai_payload));
		if (!get_pai_payload(session->channel, rdata, pai_payload)) {
			ast_queue_control_data(session->channel, AST_CONTROL_PAI_INFO, pai_payload, sizeof(struct ast_sip_pai_payload));
		}

		/* Add by bfdu for Bug 211171, 2021.12.23 */
		set_peer_enter_voicemail(session->channel, rdata);

		ast_sip_session_set_gs_client_info(rdata, session, 0);

		if (0 == session->is_media_change) {
			ast_queue_control(session->channel, AST_CONTROL_ANSWER);
		} else {
			send_response_control_frame_to_peer(session);
		}

		//added by zyxia for report queue answer state, ediy for Bug262507 2023.6.2
		if(session->channel && ast_channel_name(session->channel) && !strcasestr(ast_channel_name(session->channel),"PJSIP/trunk"))
			schedule_task_for_real_answered(session,200);
#else
		ast_queue_control(session->channel, AST_CONTROL_ANSWER);
#endif
		break;
#ifdef GRANDSTREAM_NETWORKS
	case 486:
		/*start add by zyxia for wake up gswave,if receive 486 from gswave,*/
		if (!session->has_dec_gs_call_count) {
			ast_sip_set_used_gs_call_count(session,rdata,0);
		}
		/*end add by zyxia for wake up gswave,receive 486 from gswave,*/

		cei_support = pbx_builtin_getvar_helper(session->channel, "CEI_SUPPORT");
		conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");

		// Invited user rejected a call, so we need to terminate its subscription
		if (!ast_strlen_zero(cei_support) && !ast_strlen_zero(conf_room)) {
			publish_cei_unsub(conf_room, ast_channel_name(session->channel), S_COR(ast_channel_caller(session->channel)->id.number.valid, ast_channel_caller(session->channel)->id.number.str, ""));
		}
		break;
	/*start added by zyxia for Bug 121281  receive reponse 408 and other failed invite from gswave */
	case 408:
		if (!session->has_dec_gs_call_count) {
			ast_sip_set_used_gs_call_count(session,rdata,0);
		}
		break;
	/*end added by zyxia for Bug 121281  receive reponse 408 and other failed invite from gswave */

#endif
	default:
#ifdef GRANDSTREAM_NETWORKS
	/*start added by zyxia for Bug 121281  receive reponse 408 and other failed invite from gswave */
	if (status.code >= 300 && !session->has_dec_gs_call_count) {
		ast_sip_set_used_gs_call_count(session,rdata,0);
	}
	/*end added by zyxia for Bug 121281  receive reponse 408 and other failed invite from gswave */
#endif
		break;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for re-INVITE for receive 2XX response, 2019.10.11 */
	if (status.code > 200 && status.code < 300 && session->is_media_change) {
		send_response_control_frame_to_peer(session);
	}

	/* Add by bfdu for re-INVITE failed, like 491 respond, 2018.05.28 */
	if ((status.code >= 300) && (status.code != 401) && (status.code != 407) && session->is_media_change) {
		session->is_media_change = 0;
		session->recv_reinvite_no_sdp = 0;
		ast_log(LOG_DEBUG, "chan_pjsip_incoming_response failed for re-INVITE respond is code:%d\n", status.code);
		ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_FAILED, &status.code, sizeof(status.code));
	}

	if (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED &&
		status.code >= 300 && (status.code != 401) && (status.code != 407)) {
		ast_log(LOG_DEBUG, "chan_pjsip_incoming_response respond is code:%d, reset pending media\n", status.code);
		ast_sip_session_media_state_reset(session->pending_media_state);
	}

	if ((status.code >= 300) && (status.code != 401) && (status.code != 407) && session->camera_control_action) {
		ast_log(LOG_DEBUG, "chan_pjsip_incoming_response failed for re-INVITE respond is code:%d\n", status.code);
		queue_camera_control_frame_for_failed(session);
		session->camera_control_action = 0;
	}
#endif
}

static int chan_pjsip_incoming_ack(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	if (rdata->msg_info.msg->line.req.method.id == PJSIP_ACK_METHOD) {
		if (session->endpoint->media.direct_media.enabled && session->channel) {
			ast_queue_control(session->channel, AST_CONTROL_SRCCHANGE);
		}
	}
	return 0;
}

static int update_devstate(void *obj, void *arg, int flags)
{
	ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE,
			     "PJSIP/%s", ast_sorcery_object_get_id(obj));
	return 0;
}

static struct ast_custom_function chan_pjsip_dial_contacts_function = {
	.name = "PJSIP_DIAL_CONTACTS",
	.read = pjsip_acf_dial_contacts_read,
};

static struct ast_custom_function chan_pjsip_parse_uri_function = {
	.name = "PJSIP_PARSE_URI",
	.read = pjsip_acf_parse_uri_read,
};

static struct ast_custom_function media_offer_function = {
	.name = "PJSIP_MEDIA_OFFER",
	.read = pjsip_acf_media_offer_read,
	.write = pjsip_acf_media_offer_write
};

static struct ast_custom_function dtmf_mode_function = {
	.name = "PJSIP_DTMF_MODE",
	.read = pjsip_acf_dtmf_mode_read,
	.write = pjsip_acf_dtmf_mode_write
};

static struct ast_custom_function session_refresh_function = {
	.name = "PJSIP_SEND_SESSION_REFRESH",
	.write = pjsip_acf_session_refresh_write,
};

#ifdef GRANDSTREAM_NETWORKS

/* Added by yfyang, for missed call feature */
static struct ast_custom_function chan_pjsip_missed_call_contacts_function = {
	.name = "PJSIP_MISSED_CALL_CONTACTS",
	.read = pjsip_acf_missed_call_contacts_read,
};


static struct ast_custom_function chan_pjsip_gs_message_contacts_function = {
	.name = "PJSIP_GS_MESSAGE_CONTACTS",
	.read = pjsip_acf_gs_message_contacts_read,
};


static struct ast_custom_function chan_pjsip_plain_message_contacts_function = {
	.name = "PJSIP_PLAIN_MESSAGE_CONTACTS",
	.read = pjsip_acf_plain_message_contacts_read,
};

/* Added by yfyang, for Bug 221341 */
static int chan_pjsip_cc_callback(struct ast_channel *inbound, const char *dest, ast_cc_callback_fn callback)
{
	struct ast_cc_config_params *cc_params = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	enum ast_cc_monitor_policies monitor_type = AST_CC_MONITOR_NEVER;
	char device_name[80] = {0};
	char *num   = NULL;
	char *slash = NULL;

	if (!inbound || !dest) {
		return -1;
	}

	ast_log(LOG_NOTICE, "chan_pjsip_cc_callback, inbound[%s], dest[%s]", ast_channel_name(inbound), dest);

	num = ast_strdupa(dest);
	slash = strchr(num, '/');
	if (slash) {
		*slash = '\0';
	}

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", num);
	if (!endpoint) {
		return -1;
	}

	cc_params = endpoint->cc_params;
	if (!cc_params) {
		return -1;
	}

	monitor_type = ast_get_cc_monitor_policy(cc_params);
	if ((AST_CC_MONITOR_ALWAYS != monitor_type) && (AST_CC_MONITOR_GENERIC != monitor_type)) {
		return -1;
	}

	snprintf(device_name, sizeof(device_name), "PJSIP/%s", num);
	if (ast_cc_monitor_count(device_name, "generic") >= ast_get_cc_max_monitors(cc_params)) {
		return -1;
	}

	callback(inbound, cc_params, "generic", device_name, device_name, NULL);

	ast_log(LOG_NOTICE, "chan_pjsip_cc_callback, finished.");
	return 0;
}

static int is_pjsip_channel(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *channel = NULL;

	if (chan == NULL) {
		return -1;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP") != 0) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return -1;
	}

	channel = ast_channel_tech_pvt(chan);
	if (channel == NULL || channel->pvt == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
		ast_channel_name(chan));
		return -1;
	}

	return 0;
}

/*! Destroy conference subscriptions */
static void destroy_conference(void *data)
{
	struct sip_conference *conference = data;

	// Sanity check
	if (conference == NULL) {
		return;
	}

	ast_debug(LOG_DEBUG, "[TK] Destroying Channel %s, Conference %s\n", conference->channelName, conference->conference);

	if (conference->channelName != NULL) {
		ast_free(conference->channelName);
		conference->channelName = NULL;
	}

	if (conference->conference != NULL) {
		ast_free(conference->conference);
		conference->conference = NULL;
	}
}

static int add_peer_conference(const char *channel, struct ast_sip_endpoint *endpoint, const char *conference,enum ast_conference_type conf_type)
{
	struct sip_conference* conf = NULL;

	// Sanity check
	if (channel == NULL || endpoint == NULL || conference == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return -1;
	}

	conf = ast_sip_cei_find_peer_conference(endpoint, conference, channel);
	if (conf != NULL) {
		ast_log(LOG_WARNING, "[TK] Channel %s already subscribed to Conference %s\n",
		channel, conference);
		ao2_ref(conf, -1);
		return -1;
	}

	AST_LIST_LOCK(&endpoint->conferences);

	conf = ao2_alloc(sizeof(*conf), destroy_conference);
	if (conf == NULL) {
		ast_log(LOG_WARNING, "Failed to allocate new peer conference\n");
		AST_LIST_UNLOCK(&endpoint->conferences);
		return -1;
	}

	conf->conference = ast_strdup(conference);
	conf->channelName = ast_strdup(channel);
	conf->state = SIP_CONFERENCE_STATE_WAITSUB;
	conf->update_number = 1;
	conf->ceiJoined = 0;
	conf->ceiRole = CEI_ROLE_UNKNOWN; /* Added by achan, Grandstream for CCM */

	/*Start added by zyxia for trunk conf,set is virtual conf,2022.11.22*/
	conf->is_virtual_conf = 0;
	if( conf_type == AST_CONF_TYPE_VIRTUAL){
		conf->is_virtual_conf = 1;
	}
	/*End added by zyxia for trunk conf,set is virtual conf,2022.11.22*/

	AST_LIST_INSERT_TAIL(&endpoint->conferences, conf, entry);

	AST_LIST_UNLOCK(&endpoint->conferences);

	ast_log(LOG_DEBUG, "[TK] Channel %s subscribed to Conference %s",
	conf->channelName, conf->conference);

	return 0;
}

/*! \brief Retrive contact URI from AORs moved sip:.
    This returned URI MUST be freed.
*/
static char* get_contact_uri_from_aors(const char *aors)
{
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
	const char *origin_uri = NULL;
	char *uri = NULL;
	char *pUri = NULL;

	// Sanity check
	if (aors == NULL) {
		return NULL;
	}

	contact = ast_sip_location_retrieve_contact_from_aor_list(aors);
	if (contact == NULL || ast_strlen_zero(contact->uri)) {
		return NULL;
	}

	// Valid URI starts with sip: or sips:
	if (strncmp(contact->uri, "sip:", 4) == 0) {
		origin_uri = contact->uri + 4;
	} else if (strncmp(contact->uri, "sips:", 5) == 0) {
		origin_uri = contact->uri + 5;
	} else {
		return NULL;
	}

	// IP address is next to username@
	pUri = strchr(origin_uri, '@');
	if (pUri == NULL) {
		uri = ast_strdup(origin_uri);
	} else {
		uri = ast_strdup(pUri + 1);
	}

	// Remove port, modified for ipv6, bug Bug 67059
	if ((uri[0] == '[') && (pUri = strchr(uri, ']'))) {
		*pUri = '\0';
		pUri = ast_strdup(uri + 1);
		ast_free(uri);
		uri = pUri;
	} else {
		pUri = strchr(uri, ':');
		if (pUri != NULL) {
		*pUri = '\0';
		}
	}

	// Remove transport
	pUri = strchr(uri, ';');
	if (pUri != NULL) {
		*pUri = '\0';
	}

	return uri;
}

/*!
 * \brief Prepare and publish an \ref ast_cei_ext_mcb_sub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static void publish_cei_ext_mcb_sub(const char *peer_name,
                                    struct cei_external_mcb_peer *mcb_peer)
{
	struct stasis_topic *cei_mcb_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (peer_name == NULL || mcb_peer == NULL) {
		return;
	}

	cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s, s: s}",
								"peer_name", peer_name,
								"mcb_peer_id", mcb_peer->ref_id);

	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return;
	}

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

	msg = stasis_message_create(ast_cei_ext_mcb_sub_type(), payload);
	if (msg == NULL) {
		ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_mcb_sub_type");
		return;
	}

	stasis_publish(cei_mcb_sub_topic, msg);
}

/*!
 * \brief Prepare and publish an \ref ast_cei_ext_mcb_unsub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static int publish_cei_ext_mcb_unsub(const char *peer_name,
                                      struct cei_external_mcb_peer *mcb_peer)
{
	struct stasis_topic *cei_mcb_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (peer_name == NULL || mcb_peer == NULL) {
		return -1;
	}

	cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
		return -1;
	}

	json_object = ast_json_pack("{s: s, s: s, s: s, s: s}",
								"peer_name", peer_name,
								"mcb_peer_id", mcb_peer->ref_id,
								"remote_conf_id", mcb_peer->remote_conf_id,
								"local_conf_id", mcb_peer->local_conf_id);

	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return -1;
	}

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return -1;
	}

	msg = stasis_message_create(ast_cei_ext_mcb_unsub_type(), payload);
	if (msg == NULL) {
		ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_mcb_unsub_type");
		return -1;
	}

	stasis_publish(cei_mcb_sub_topic, msg);

	return 0;
}

/*!
 * \brief Prepare and publish an \ref ast_cei_int_mcb_sub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static void publish_cei_int_mcb_sub(struct cei_internal_mcb_peer *mcb_peer)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	struct stasis_topic *cei_mcb_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (mcb_peer == NULL) {
		return;
	}

	cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s}",
								"mcb_peer_id", mcb_peer->ref_id);

	ast_log(LOG_DEBUG, "[WL] publish_cei_int_mcb_sub prepare to send data: %s,%s,%s,%s,%s\n",
	mcb_peer->local_conf,
	mcb_peer->local_conf_id,
	mcb_peer->remote_conf,
	mcb_peer->remote_conf_id,
	mcb_peer->channelName);

	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return;
	}

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

	msg = stasis_message_create(ast_cei_int_mcb_sub_type(), payload);
	if (msg == NULL) {
		ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_mcb_sub_type");
		return;
	}

	stasis_publish(cei_mcb_sub_topic, msg);
}

/*!
 * \brief Prepare and publish an \ref ast_cei_int_mcb_unsub_type message to the
 * \ref ast_cei_mcb_sub_topic_all.
 */
static void publish_cei_int_mcb_unsub(struct cei_internal_mcb_peer *mcb_peer)
{
	struct stasis_topic *cei_mcb_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (mcb_peer == NULL) {
		return;
	}

	cei_mcb_sub_topic = ast_cei_mcb_sub_topic_all();
	if (cei_mcb_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_mcb_sub_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s}",
								"mcb_peer_id", mcb_peer->ref_id);

	ast_log(LOG_DEBUG, "[WL] publish_cei_int_mcb_unsub prepare to send data: %s,%s,%s,%s,%s\n",
	mcb_peer->local_conf,
	mcb_peer->local_conf_id,
	mcb_peer->remote_conf,
	mcb_peer->remote_conf_id,
	mcb_peer->channelName);

	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return;
	}

	payload = ast_json_payload_create(json_object);
	if(NULL == payload) {
		return;
	}

	msg = stasis_message_create(ast_cei_int_mcb_unsub_type(), payload);
	if (msg == NULL) {
		ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_int_mcb_unsub_type");
		return;
	}

	stasis_publish(cei_mcb_sub_topic, msg);
}

/*! \brief Set call type in SIP response */
static int sip_setcalltype(struct ast_channel *chan, const char *data)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt;
    const char *inbuf = data;
    char *subbuf = NULL;
    size_t len = 0;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: Type\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

    ast_channel_lock(chan);

	pvt = channel->pvt;
	if (!pvt) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
                             ast_channel_name(chan));
        ast_channel_unlock(chan);
		return -1;
	}

    len = strlen(inbuf);
	subbuf = (char*)ast_malloc(len + 1);
	ast_get_encoded_str(inbuf, subbuf, len + 1);

	if (pvt->call_type != NULL) {
		ast_log(LOG_DEBUG, "Channel: %s, call_type: %s\n",
							ast_channel_name(chan), pvt->call_type);

		ast_free(pvt->call_type);
		pvt->call_type = NULL;
	}
	pvt->call_type = subbuf;

    ast_log(LOG_DEBUG, "[TK] Channel: %s sets call type response hash to: %s\n",
						ast_channel_name(chan), pvt->call_type);

	if(pvt->call_type && !strcmp(pvt->call_type,"IPC")){
		ast_channel_set_is_ipc_call(chan, 1);
	}

	ast_channel_unlock(chan);

	return 0;
}

static struct ast_custom_function pjsip_mcb_resp_function = {
	.name = "MCB_RESPONSE",
	.read = func_mcb_resp_get,
};

/*! \brief Add a SIP header to an outbound INVITE */
static int sip_setmcbresp(struct ast_channel *chan, const char *data)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt;
	const char *inbuf = data;
	char *subbuf = NULL;
	size_t len = 0;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: Header\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

	ast_channel_lock(chan);

	pvt = channel->pvt;
	if (!pvt) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return -1;
	}

	len = strlen(inbuf);
	subbuf = (char*)ast_malloc(len + 1);
	ast_get_encoded_str(inbuf, subbuf, len + 1);

	if (pvt->mcb_response != NULL) {
		ast_log(LOG_DEBUG, "Channel: %s, mcb_response: %s\n", ast_channel_name(chan), pvt->mcb_response);
		ast_free(pvt->mcb_response);
		pvt->mcb_response = NULL;
	}
	pvt->mcb_response = subbuf;

	ast_log(LOG_DEBUG, "[TK] Channel: %s sets MCB response hash to: %s\n",
	ast_channel_name(chan), pvt->mcb_response);

	ast_channel_unlock(chan);

	return 0;
}

static int sip_set_conf_focus(struct ast_channel *chan, const char *data)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = NULL;
	struct ast_sip_session *session = NULL;
	struct ast_sip_endpoint *endpoint = NULL;
	const char *inbuf = data;
	const char *id = NULL;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: Room number\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

	ast_channel_lock(chan);

	pvt = channel->pvt;
	if (pvt == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return 0;
	}

	session = channel->session;
	if (session == NULL || session->endpoint == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no end point!\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return 0;
	}

	endpoint = session->endpoint;
	id = ast_sorcery_object_get_id(endpoint);
	if (id == NULL) {
		ast_log(LOG_WARNING, "Channel %s has invalid end point (%s)!\n", ast_channel_name(chan), S_COR(id, id, "NULL"));
		ast_channel_unlock(chan);
		return 0;
	}

	pvt->isConference = TRUE;

	if (!session->is_sfu) {
		set_session_sfu_params_according_chan(session, chan);
	}

	if (globalEnabledCEI || session->is_sfu) {
		add_peer_conference(ast_channel_name(chan), endpoint, inbuf,AST_CONF_TYPE_NORMAL);
	}

	ast_channel_unlock(chan);

	return 0;
}

/*! \brief Set Contact URI Host User Part during responding of an incoming INVITE */
static int sip_set_contact_user(struct ast_channel *chan, const char *data)
{
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt;
	const char *inbuf = data;
	char *subbuf = NULL;
	size_t len = 0;

	if (ast_strlen_zero(inbuf)) {
		ast_log(LOG_WARNING, "This application requires the argument: User\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP channel\n");
		return 0;
	}

	ast_channel_lock(chan);

	pvt = channel->pvt;
	if (!pvt) {
		ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return -1;
	}

	len = strlen(inbuf);
	subbuf = (char*)ast_malloc(len + 1);
	ast_get_encoded_str(inbuf, subbuf, len + 1);

	if (pvt->contact_user != NULL) {
		ast_log(LOG_DEBUG, "Channel: %s, contact_user: %s\n", ast_channel_name(chan), pvt->contact_user);
		ast_free(pvt->contact_user);
		pvt->contact_user = NULL;
	}
	pvt->contact_user = subbuf;

	ast_log(LOG_DEBUG, "[TK] Channel: %s sets Contact URI Host User Part to: %s\n",
	ast_channel_name(chan), pvt->contact_user);

	ast_channel_unlock(chan);

	return 0;
}

/*!
 * \brief Prepare and publish an \ref ast_cei_unsub_type message to the
 * \ref ast_cei_sub_topic_all.
 */
static void publish_cei_unsub(const char *conference, const char *channel, const char *cid_number)
{
	struct stasis_topic *cei_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (conference == NULL || channel == NULL) {
		return;
	}

	cei_sub_topic = ast_cei_sub_topic_all();
	if (cei_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_sub_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s, s: s, s: s}",
								"conference", conference,
								"channel", channel,
								"cid_number", cid_number?cid_number:"Anonymous");

	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return;
	}

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

	msg = stasis_message_create(ast_cei_unsub_type(), payload);
	if (msg == NULL) {
		ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_unsub_type");
		return;
	}

	stasis_publish(cei_sub_topic, msg);
}

/*!
 * \brief Prepare and publish an \ref ast_cei_joined_type message to the
 * \ref ast_cei_sub_topic_all.
 */
static void publish_cei_joined(const char *conference, const char *channel)
{
	struct stasis_topic *cei_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (conference == NULL || channel == NULL) {
		return;
	}

	cei_sub_topic = ast_cei_sub_topic_all();
	if (cei_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_sub_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s, s: s}",
								"conference", conference,
								"channel", channel);

	if (json_object == NULL) {
		ast_log(LOG_WARNING, "Failed to create json_object");
		return;
	}

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		return;
	}

	msg = stasis_message_create(ast_cei_joined_type(), payload);
	if (msg == NULL) {
		ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_joined_type");
		return;
	}

	stasis_publish(cei_sub_topic, msg);
}

/*! \brief Send stasis message to generate SIP NOTIFY message with terminated state */
static int sip_terminated_notify(struct ast_channel *chan, const char *data)
{
    // Sanity check
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires the argument: Conference\n");
		return 0;
	}

    ast_channel_lock(chan);

    publish_cei_unsub(data, ast_channel_name(chan), S_COR(ast_channel_caller(chan)->id.number.valid, ast_channel_caller(chan)->id.number.str, ""));

	ast_channel_unlock(chan);

	return 0;
}

/*! \brief Given channel joined to a conference with valid authentication */
static int sip_cei_joined(struct ast_channel *chan, const char *data)
{
	struct ast_sip_channel_pvt *channel = NULL;
	struct ast_sip_endpoint *endpoint = NULL;
	RAII_VAR(struct sip_conference *, conf, NULL, ao2_cleanup);

	// Sanity check
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires the argument: Conference\n");
		return 0;
	}

	if (is_pjsip_channel(chan) == -1) {
		return 0;
	}

	ast_channel_lock(chan);

	// Update ceiJoined status
	channel = ast_channel_tech_pvt(chan);
	if (channel == NULL || channel->session == NULL ||
		channel->session->endpoint == NULL) {
		ast_log(LOG_WARNING, "Failed to get endpoint of channel %s\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return 0;
	}

	endpoint = channel->session->endpoint;
	conf = ast_sip_cei_find_peer_conference(endpoint, data, ast_channel_name(chan));
	if (conf == NULL) {
		ast_log(LOG_WARNING, "Failed to get peer conference: chan: %s, conf: %s\n",
				ast_channel_name(chan), data);
		ast_channel_unlock(chan);
		return 0;
	}
	conf->ceiJoined = 1;

	ast_log(LOG_DEBUG, "[TK] Channel %s in conference %s has valid authentication\n",
			ast_channel_name(chan), data);

	// Notify ceiJoined status to CEI module
	publish_cei_joined(data, ast_channel_name(chan));

	ast_channel_unlock(chan);

	return 0;
}

/*! \brief find the caller id of incoming call bridged by register sip trunk */
/*! \return need to be free*/
static char* get_register_trunk_caller_id(const char * extension_num)
{
	char family[80] = {0};
	char caller_id[64] = {0};
	char outbound_cid[128] = {0};
	char *ret;

	// Sanity check
	if (extension_num == NULL || ast_strlen_zero(extension_num)) {
		return NULL;
	}

	snprintf(family, sizeof(family), "AAUSER/%s", extension_num);
	ast_ucm_db_get(family, "OUTBOUNDCID", outbound_cid, sizeof(outbound_cid));
	if (ast_strlen_zero(outbound_cid)) {
		ast_log(LOG_WARNING, "[WL] No customized caller id of %s\n", family);
		return NULL;
	}

	ret = strrchr(outbound_cid, '<');
	if (ret == NULL || ast_strlen_zero(ret + 2)) {
		return NULL;
	}

	strncpy(caller_id, ret + 1, strlen(ret) -2);
	return ast_strdup(caller_id);
}

static int sip_external_mcb_sub(struct ast_channel *chan, const char *data)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	if (!globalEnabledCEI) {
		return 0;
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = NULL;
	struct ast_sip_session *session = NULL;
	struct ast_sip_endpoint *endpoint = NULL;
	char *inbuf = NULL;
	char *local_conf = NULL;
	char *local_conf_id = NULL;
	char *remote_conf = NULL;
	char *remote_conf_id = NULL;
	char *remote_channel = NULL;
	char *endpoint_name = NULL;
	char *call_type = NULL;
	char *saveptr = NULL;
	char local_conf_private_num[128];
	char remote_conf_public_num[128];
	char *remote_conf_private_num = NULL;
	char *ip_addr = NULL;
	char outbound_prepend[MAX_PREPEND_LEN] = {0};
	char inbound_prepend[MAX_PREPEND_LEN]  = {0};
	int pjsip_channel = 0;
	int outbound_strip = 0;
	int inbound_strip = 0;
	struct cei_external_mcb_peer *mcb_peer = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint0, NULL, ao2_cleanup);
	const char *conf_room = NULL;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires the arguments: "
							"local conference, local conference ID, "
							"remote conference, remote conference ID, "
							"remote channel, endpoint name, call type\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP") &&
		strcmp(ast_channel_tech(chan)->type, "Local")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP/Local channel\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP") == 0) {
		pjsip_channel = 1;
	}

	ast_channel_lock(chan);

	//=================================================
	// Parse arguments
	//=================================================
	inbuf = ast_strdupa(data);

	// Get local conference
	local_conf = strtok_r(inbuf, ",", &saveptr);
	if (ast_strlen_zero(local_conf)) {
		ast_log(LOG_WARNING, "Failed to parse local conference\n");
		ast_channel_unlock(chan);
		return 0;
	}

	local_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!local_conf_id || strlen(local_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_conf = strtok_r(NULL, ",", &saveptr);
	if (ast_strlen_zero(remote_conf)) {
		ast_log(LOG_WARNING, "Failed to parse remote conference\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!remote_conf_id || strlen(remote_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_channel = strtok_r(NULL, ",", &saveptr);
	if (ast_strlen_zero(remote_channel)) {
		ast_log(LOG_WARNING, "Failed to parse remote channel\n");
		ast_channel_unlock(chan);
		return 0;
	}

	endpoint_name = strtok_r(NULL, ",", &saveptr);
	if (ast_strlen_zero(endpoint_name)) {
		ast_log(LOG_WARNING, "Failed to parse endpoint name\n");
		ast_channel_unlock(chan);
		return 0;
	}

	call_type  = strtok_r(NULL, "", &saveptr);
	if (ast_strlen_zero(call_type)) {
		ast_log(LOG_WARNING, "Failed to parse call type\n");
		ast_channel_unlock(chan);
		return 0;
	}

	//=================================================
	// Get session and endpoint
	//=================================================
	// PJSIP channel
	if (pjsip_channel) {
		pvt = channel->pvt;
		if (pvt == NULL) {
			ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
			ast_channel_name(chan));
			ast_channel_unlock(chan);
			return 0;
		}

		// Get session
		session = channel->session;

		// Get endpoint
		if (session != NULL) {
			endpoint = session->endpoint;
		}
	} else { // Local channel
		// Get session
		session = channel->session;

		// Get endpoint
		endpoint = endpoint0 = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", endpoint_name);
	}

	if (session == NULL || endpoint == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no end point!\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return 0;
	}

	//=================================================
	// Save new subscription in endpoint
	//=================================================
	if (ast_strlen_zero(endpoint->aors)) {
		ast_log(LOG_WARNING, "Failed to get endpoint aors\n");
		ast_channel_unlock(chan);
		return 0;
	}

	memset(remote_conf_public_num, 0, sizeof(remote_conf_public_num));
	if (strcmp(call_type, CEI_MCB_DIR_OUTGOING) == 0) {
		if (ast_cei_get_trunk_info_by_ext(endpoint->aors, CEI_MCB_DIR_OUTGOING, remote_conf, outbound_prepend, &outbound_strip)) {
			ast_log(LOG_WARNING, "Failed to get outgoing informations for trunk: %s\n", endpoint->aors);
			ast_channel_unlock(chan);
			return 0;
		}
		if (!ast_strlen_zero(outbound_prepend)) {
			strncpy(remote_conf_public_num, outbound_prepend, strlen(outbound_prepend));
		}
	}

	strcat(remote_conf_public_num, remote_conf + outbound_strip);

	// Get remote address
	ip_addr = get_contact_uri_from_aors(endpoint->aors);
	if (ast_strlen_zero(ip_addr)) {
		ast_channel_unlock(chan);
		return 0;
	}

	conf_room = pbx_builtin_getvar_helper(chan, "CONF_ROOM");
	if (ast_strlen_zero(conf_room)) {
		conf_room = local_conf;
	}

	mcb_peer = ast_cei_add_external_mcb_peer(endpoint, local_conf, local_conf_id,
				remote_conf_public_num, remote_conf_id, ip_addr,
				(pjsip_channel) ? ast_channel_name(chan) : remote_channel,
				conf_room);

	AST_LIST_LOCK(&endpoint->external_mcb_peers);

	if(mcb_peer == NULL) {
		ast_channel_unlock(chan);
		ast_free(ip_addr);
		AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

		return 0;
	}

	//Retrieve correct provate num of local conf and remote conf
	if (strcmp(call_type, CEI_MCB_DIR_INCOMING) == 0) {
		if (strstr(endpoint_name, "trunk_") == NULL) {
			ast_log(LOG_DEBUG, "[WL] Incoming mcb via Register SIP Trunk \n");
			remote_conf_private_num = get_register_trunk_caller_id(endpoint_name);

			if (remote_conf_private_num != NULL && !ast_strlen_zero(remote_conf_private_num)) {
				ast_string_field_set(mcb_peer, conf_private_num, remote_conf_private_num);
				ast_free(remote_conf_private_num);
			}
		} else {
			ast_log(LOG_DEBUG, "[WL] Incoming mcb via Peer SIP Trunk \n");

			memset(local_conf_private_num, 0, sizeof(local_conf_private_num));
			if (ast_cei_get_trunk_info_by_ext(endpoint->aors, CEI_MCB_DIR_INCOMING, mcb_peer->local_conf, inbound_prepend, &inbound_strip)) {
				ast_log(LOG_WARNING, "Failed to get incoming informations for trunk: %s\n", mcb_peer->local_conf);
				ast_channel_unlock(chan);
				return 0;
			}

			if (!ast_strlen_zero(inbound_prepend)) {
				strncpy(local_conf_private_num, inbound_prepend, strlen(inbound_prepend));
			}

			strcat(local_conf_private_num, mcb_peer->local_conf + inbound_strip);
		}
	} else if (strcmp(call_type, CEI_MCB_DIR_OUTGOING) == 0) {
		ast_log(LOG_DEBUG, "[WL] Outgoing mcb \n");
		ast_string_field_set(mcb_peer, conf_private_num, remote_conf);
	}

	// Notify res_pjsip_cei to generate SIP SUBSCRIBE
	publish_cei_ext_mcb_sub(ast_sorcery_object_get_id(endpoint), mcb_peer);

	// Clean up
	ast_channel_unlock(chan);
	ast_free(ip_addr);
	AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

	return 0;
}

static int sip_external_mcb_unsub(struct ast_channel *chan, const char *data)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	if (!globalEnabledCEI) {
		return 0;
	}
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct chan_pjsip_pvt *pvt = NULL;
	struct ast_sip_session *session = NULL;
	struct ast_sip_endpoint *endpoint = NULL;
	char *inbuf = NULL;
	char *local_conf_id = NULL;
	char *remote_conf_id = NULL;
	char *endpoint_name = NULL;
	char *saveptr = NULL;
	int pjsip_channel = 0;
	struct cei_external_mcb_peer *mcb_peer = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint0, NULL, ao2_cleanup);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "This application requires the arguments: "
				"local conference ID, remote conference ID, "
				"endpoint name\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP") &&
		strcmp(ast_channel_tech(chan)->type, "Local")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP/Local channel\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP") == 0) {
		pjsip_channel = 1;
	}

	ast_channel_lock(chan);

	//=================================================
	// Parse arguments
	//=================================================
	inbuf = ast_strdupa(data);

	// Get local conference
	local_conf_id = strtok_r(inbuf, ",", &saveptr);
	if (!local_conf_id || strlen(local_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!remote_conf_id || strlen(remote_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	endpoint_name = strtok_r(NULL, "", &saveptr);
	if (ast_strlen_zero(endpoint_name)) {
		ast_log(LOG_WARNING, "Failed to parse endpoint name\n");
		ast_channel_unlock(chan);
		return 0;
	}

	//=================================================
	// Get session and endpoint
	//=================================================
	// PJSIP channel
	if (pjsip_channel) {
		pvt = channel->pvt;
		if (pvt == NULL) {
			ast_log(LOG_WARNING, "Channel %s has no chan_pjsip pvt!\n",
			ast_channel_name(chan));
			ast_channel_unlock(chan);
			return 0;
		}

		// Get session
		session = channel->session;

		// Get endpoint
		if (session != NULL) {
			endpoint = session->endpoint;
		}
	} else { // Local channel
		// Get session
		session = channel->session;

		// Get endpoint
		endpoint = endpoint0 = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
		endpoint_name);
	}

	if (session == NULL || endpoint == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no end point!\n", ast_channel_name(chan));
		ast_channel_unlock(chan);
		return 0;
	}

	//=================================================
	// Unsubscription
	//=================================================
	// Get MCB peer
	mcb_peer = ast_cei_find_external_mcb_peer_by_id(endpoint, local_conf_id, remote_conf_id);
	if (mcb_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to get mcb peer of local id: %s, remote id: %s\n",
				local_conf_id, remote_conf_id);
		ast_channel_unlock(chan);
		return 0;
	}

	// Notify res_pjsip_pubsub to generate SIP Un-SUBSCRIBE
	if (publish_cei_ext_mcb_unsub(ast_sorcery_object_get_id(endpoint), mcb_peer) == -1) {
		// Remove given subscription in endpoint if the subscription is failed.
		ast_cei_remove_external_mcb_peer(endpoint, local_conf_id, remote_conf_id);
	}

	// Clean up
	ast_channel_unlock(chan);

	return 0;
}

static int sip_internal_mcb_sub_rev(const char *local_conf, const char *local_conf_id,
                                    const char *remote_conf, const char *remote_conf_id,
                                    const char *channelName)
{
	// Sanity check
	if (local_conf == NULL || local_conf_id == NULL ||
		remote_conf == NULL || remote_conf_id == NULL || channelName == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return 0;
	}

	struct cei_internal_mcb_peer *mcb_peer = NULL;

	mcb_peer = ast_cei_add_int_mcb_peer(local_conf, local_conf_id,
	remote_conf, remote_conf_id,
	channelName);
	if(mcb_peer == NULL) {
		return 0;
	}

	// Notify res_pjsip_cei to generate SIP SUBSCRIBE
	publish_cei_int_mcb_sub(mcb_peer);

	return 0;
}

static int sip_internal_mcb_sub(struct ast_channel *chan, const char *data)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	if (!globalEnabledCEI) {
		return 0;
	}
	char *inbuf = NULL;
	char *local_conf = NULL;
	char *local_conf_id = NULL;
	char *remote_conf = NULL;
	char *remote_conf_id = NULL;
	char *saveptr = NULL;
	struct cei_internal_mcb_peer *mcb_peer = NULL;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "[WL] Invalid data\n");
		return 0;
	}

	if (strcmp(ast_channel_tech(chan)->type, "Local")) {
		ast_log(LOG_WARNING, "This application supports only Local channel\n");
		return 0;
	}

	ast_channel_lock(chan);

	//=================================================
	// Parse arguments
	//=================================================
	inbuf = ast_strdupa(data);

	// Get local conference
	local_conf = strtok_r(inbuf, ",", &saveptr);
	if (ast_strlen_zero(local_conf)) {
		ast_log(LOG_WARNING, "Failed to parse local conference\n");
		ast_channel_unlock(chan);
		return 0;
	}

	local_conf_id = strtok_r(NULL, ",", &saveptr);
	if (!local_conf_id || strlen(local_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_conf = strtok_r(NULL, ",", &saveptr);
	if (ast_strlen_zero(remote_conf)) {
		ast_log(LOG_WARNING, "Failed to parse remote conference\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_conf_id = strtok_r(NULL, "", &saveptr);
	if (!remote_conf_id || strlen(remote_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	mcb_peer = ast_cei_add_int_mcb_peer(local_conf, local_conf_id,
	remote_conf, remote_conf_id,
	ast_channel_name(chan));
	if(mcb_peer == NULL) {
		ast_channel_unlock(chan);
		return 0;
	}

	// Notify res_pjsip_cei to generate SIP SUBSCRIBE
	publish_cei_int_mcb_sub(mcb_peer);

	sip_internal_mcb_sub_rev(remote_conf, remote_conf_id,
	local_conf, local_conf_id,
	ast_channel_name(chan));
	// Clean up
	ast_channel_unlock(chan);

	return 0;
}

static int sip_internal_mcb_unsub_rev(const char *local_conf_id,
                                      const char *remote_conf_id)
{
	// Sanity check
	if (local_conf_id == NULL || remote_conf_id == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return 0;
	}

	struct cei_internal_mcb_peer *mcb_peer = NULL;

	mcb_peer = ast_cei_find_int_mcb_peer(local_conf_id, remote_conf_id);
	if(mcb_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to get mcb peer of local id: %s, remote id: %s\n",
		local_conf_id, remote_conf_id);
		return 0;
	}

	// Notify res_pjsip_cei to generate SIP SUBSCRIBE
	publish_cei_int_mcb_unsub(mcb_peer);

	return 0;
}

static int sip_internal_mcb_unsub(struct ast_channel *chan, const char *data)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	if (!globalEnabledCEI) {
		return 0;
	}
	char *inbuf = NULL;
	char *local_conf_id = NULL;
	char *remote_conf_id = NULL;
	char *saveptr = NULL;
	struct cei_internal_mcb_peer *mcb_peer = NULL;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "[WL] Invalid Data\n");
		return 0;
	}
	if (strcmp(ast_channel_tech(chan)->type, "Local")) {
		ast_log(LOG_WARNING, "This application supports only Local channel\n");
		return 0;
	}

	ast_channel_lock(chan);

	//=================================================
	// Parse arguments
	//=================================================
	inbuf = ast_strdupa(data);

	// Get local conference
	local_conf_id = strtok_r(inbuf, ",", &saveptr);
	if (!local_conf_id || strlen(local_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse local conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}

	remote_conf_id = strtok_r(NULL, "", &saveptr);
	if (!remote_conf_id || strlen(remote_conf_id) != 40) {
		ast_log(LOG_WARNING, "Failed to parse remote conference ID in SHA1 hash\n");
		ast_channel_unlock(chan);
		return 0;
	}
	//=================================================
	// Save new subscription in endpoint
	//=================================================


	mcb_peer = ast_cei_find_int_mcb_peer(local_conf_id, remote_conf_id);
	if(mcb_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to get mcb peer of local id: %s, remote id: %s\n",
				local_conf_id, remote_conf_id);
		ast_channel_unlock(chan);
		return 0;
	}

	// [TK] TODO: Notify CEI manager about new subscription

	// Notify res_pjsip_cei to generate SIP SUBSCRIBE
	publish_cei_int_mcb_unsub(mcb_peer);

	sip_internal_mcb_unsub_rev(remote_conf_id, local_conf_id);
	// Clean up
	ast_channel_unlock(chan);

	return 0;
}

static void cei_toggle_enable_cb(void *data, struct stasis_subscription *sub,
								  struct stasis_message *message)
{
	struct ast_json			*json_object = NULL;
	struct ast_json_payload *payload	 = NULL;
	const char *enableCEI = NULL;

	// Sanity check
	if (message == NULL) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (NULL == payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (json_object == NULL) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	enableCEI = ast_json_string_get(ast_json_object_get(json_object, "enableCEI"));
	if (enableCEI == NULL) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}

	if (strcmp(enableCEI, "true") == 0) {
		globalEnabledCEI = 1;
	} else if (strcmp(enableCEI, "false") == 0) {
		globalEnabledCEI = 0;
	}
}

/*!
* \brief Prepare and publish an \ref ast_cei_ext_mcb_sub_type message to the
* \ref ast_cei_mcb_sub_topic_all.
*/
static void publish_cei_ext_trunk_conf_sub(const char *peer_name,struct cei_external_trunk_conf_peer *etc_peer)
{
	struct stasis_topic *cei_etc_sub_topic = NULL;
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);

	// Sanity check
	if (peer_name == NULL || etc_peer == NULL) {
		return;
	}

	cei_etc_sub_topic = ast_cei_etc_sub_topic_all();
	if (cei_etc_sub_topic == NULL) {
		ast_log(LOG_WARNING, "Failed to get ast_cei_etc_sub_topic_all");
		return;
	}

	json_object = ast_json_pack("{s: s, s: s,s: s}",
							  "peer_name", peer_name,
							  "etc_peer_id", etc_peer->ref_id,
							  "remote_channel_id",etc_peer->channelName);//addded by zyxia for Bug 280759 ,trunk conf ,2023.12.20 */

	  if (json_object == NULL) {
		  ast_log(LOG_WARNING, "Failed to create json_object");
		  return;
	  }

	  payload = ast_json_payload_create(json_object);
	  if (NULL == payload) {
		  return;
	  }

	  msg = stasis_message_create(ast_cei_ext_trunk_conf_sub_type(), payload);
	  if (msg == NULL) {
		  ast_log(LOG_WARNING, "Failed to create stasis message for ast_cei_ext_trunk_conf_sub_type");
		  return;
	  }

	  stasis_publish(cei_etc_sub_topic, msg);
 }

static int sip_exteral_trunk_conf_sub(struct ast_sip_session *session, const char *local_virtual_conf)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	if (!globalEnabledCEI) {
		return 0;
	}

	struct ast_sip_endpoint *endpoint = NULL;
	char *ip_addr = NULL;
	struct cei_external_trunk_conf_peer *etc_peer = NULL;
	RAII_VAR(struct ast_sip_endpoint *, endpoint0, NULL, ao2_cleanup);

	if (session == NULL || session->trunk_conf_num == NULL || local_virtual_conf == NULL) {
		ast_log(LOG_WARNING, "This application requires the arguments: "
						  "local virtual conf , remote conf\n");
		return 0;
	}

	if (session->channel && ast_channel_name(session->channel)  && !strcasestr(ast_channel_name(session->channel),"PJSIP/trunk")) {
		ast_log(LOG_WARNING, "This application supports only PJSIP/trunk channel\n");
		return 0;
	}

	ast_channel_lock(session->channel);

	//=================================================
	  endpoint = session->endpoint;

	if (endpoint == NULL) {
		ast_log(LOG_WARNING, "Channel %s has no end point!\n", ast_channel_name(session->channel));
		ast_channel_unlock(session->channel);
		return 0;
	}

	//=================================================
	// Save new subscription in endpoint
	//=================================================
	if (ast_strlen_zero(endpoint->aors)) {
		ast_log(LOG_WARNING, "Failed to get endpoint aors\n");
		ast_channel_unlock(session->channel);
		return 0;
	}

	// Get remote address
	ip_addr = get_contact_uri_from_aors(endpoint->aors);
	if (ast_strlen_zero(ip_addr)) {
		ast_channel_unlock(session->channel);
		return 0;
	}

	etc_peer = ast_cei_add_external_trunk_conf_peer(endpoint, local_virtual_conf,session->trunk_conf_num,ip_addr,ast_channel_name(session->channel));

	AST_LIST_LOCK(&endpoint->external_tc_peers);

	if(etc_peer == NULL) {
		ast_channel_unlock(session->channel);
		ast_free(ip_addr);
		AST_LIST_UNLOCK(&endpoint->external_tc_peers);
		return 0;
	}

	ast_string_field_set(etc_peer, remote_conf_num, session->trunk_conf_num);
	ast_string_field_set(etc_peer,channelName,session->remote_trunk_channel_id);// addded by zyxia for Bug 280759 ,trunk conf ,2023.12.20 */

	// Notify res_pjsip_cei to generate SIP SUBSCRIBE
	publish_cei_ext_trunk_conf_sub(ast_sorcery_object_get_id(endpoint), etc_peer);

	// Clean up
	ast_channel_unlock(session->channel);
	ast_free(ip_addr);
	AST_LIST_UNLOCK(&endpoint->external_tc_peers);

	return 0;
}

static int sip_exteral_trunk_conf_unsub(struct ast_sip_endpoint *endpoint ,struct cei_external_trunk_conf_peer* etc_peer)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
	if (!globalEnabledCEI || !endpoint || !etc_peer) {
		return 0;
	}

	ast_cei_remove_external_trunk_conf_peer(endpoint, etc_peer->local_virtual_conf_num);
	return 0;
}

static int ast_cei_get_enable_configuration(void)
{
	char buf[64] = { 0 };
	char buffer[256] = { 0 };
	int enableCEI = 0;

	FILE *fp = NULL;
	fp = popen("nvram get enableCEI", "r");

	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
			&& (fgets(buf, sizeof(buf) - 1, fp) != NULL)) {
			strcat(buffer, buf);
		}
		pclose(fp);
		fp = NULL;

		if (buffer[0] != '\0') {
			/*parsing result*/
			enableCEI = atoi( buffer );
		}
	}

	return enableCEI;
}

/*! dial_options convert to feature right */
static void set_feature_map(const char* dial_options)
{
	if (NULL == dial_options) {
		return;
	}

	if (strchr(dial_options, 't')) {
		gs_feature_map[GS_FEATURE_BLINDXFER].f_right |= GS_FEATURE_ALLOW_CALLEE;
		gs_feature_map[GS_FEATURE_ATXFER].f_right    |= GS_FEATURE_ALLOW_CALLEE;
	}

	if (strchr(dial_options, 'T')) {
		gs_feature_map[GS_FEATURE_BLINDXFER].f_right |= GS_FEATURE_ALLOW_CALLER;
		gs_feature_map[GS_FEATURE_ATXFER].f_right    |= GS_FEATURE_ALLOW_CALLER;
	}

	if (strchr(dial_options, 'x')) {
		gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right |= GS_FEATURE_ALLOW_CALLEE;
	}

	if (strchr(dial_options, 'X')) {
		gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right |= GS_FEATURE_ALLOW_CALLER;
	}

	if (strchr(dial_options, 'k')) {
		gs_feature_map[GS_FEATURE_PARKCALL].f_right |= GS_FEATURE_ALLOW_CALLEE;
	}

	if (strchr(dial_options, 'K')) {
		gs_feature_map[GS_FEATURE_PARKCALL].f_right |= GS_FEATURE_ALLOW_CALLER;
	}

	return;
}

/*! \brief get feature code map from features.conf, and get feature right from extensions.conf (DIAL_OPTIONS) */
int get_feature_code(int reload)
{
	struct ast_flags config_flags = {reload ? CONFIG_FLAG_FILEUNCHANGED:0};
	struct ast_config    *pstCfg  = NULL;
	struct ast_config    *pstECfg = NULL;
	struct ast_variable  *pstVarGen = NULL;
	int    i;

	/* load features.conf file*/
	pstCfg = ast_config_load("features.conf", config_flags);

	if (pstCfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstCfg) {
		ast_log(LOG_NOTICE, "features.conf unchanged\n");
		goto feature_map;
	}

	if (pstCfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of features.conf are invalid and cannot be parsed\n");
		goto feature_map;
	}

	pstVarGen = ast_variable_browse(pstCfg, "featuremap");

	for (; (NULL != pstVarGen); pstVarGen = pstVarGen->next) {
		for (i = 0; i < GS_FEATURE_MAP_COUNT; i++) {
			if (0 == strcasecmp(gs_feature_map[i].f_name, pstVarGen->name)) {
				ast_copy_string(gs_feature_map[i].f_code, pstVarGen->value,  sizeof(gs_feature_map[i].f_code));
				break;
			}
		}
	}

	ast_config_destroy(pstCfg);

feature_map:
	/* load extensions.conf file*/
	pstECfg = ast_config_load("extensions.conf", config_flags);

	if (pstECfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstECfg) {
		ast_log(LOG_NOTICE, "extensions.conf unchanged\n");
		return -1;
	}

	if (pstECfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of extensions.conf are invalid and cannot be parsed\n");
		return -1;
	}

	pstVarGen = ast_variable_browse(pstECfg, "globals");

	/* clear feature right */
	for (i = 0; i < GS_FEATURE_MAP_COUNT; i++) {
		gs_feature_map[i].f_right = GS_FEATURE_DISABLE;
	}

	for (; (NULL != pstVarGen); pstVarGen = pstVarGen->next) {
		if (0 == strcasecmp(pstVarGen->name, "DIAL_OPTIONS")) {
			if (NULL == pstVarGen->value) {
				break;
			}

			set_feature_map(pstVarGen->value);
			break;
		}
	}

	ast_config_destroy(pstECfg);
	return 0;
}

/*! \brief Add "X-UCM-AudioRecord" header and "X-UCM-CallPark" header to outgoing message
 *
 * if the right is allow callee or allow both, add the header in Invite message;
   if the right is allow caller or allow both, add the header in 200 for Invite message
 *
 * \param req The request/response to which we will add the header
 */
void add_feature_map_header(pjsip_tx_data *tdata, int outgoing_call, struct ast_sip_session *session)
{
	char  audio_record_code[12] = {0};
	char  call_park_code[12]    = {0};

	if (NULL == tdata) {
		return;
	}
#ifdef GRANDSTREAM_NETWORKS
	const char *feature_enable = pbx_builtin_getvar_helper(session->channel, "FEATURE_ENABLE");
	const char *in_queue = pbx_builtin_getvar_helper(session->channel, "IN_QUEUE");
#endif
	/* check right */
	if (outgoing_call == 0) {
		/* copy code */
		if (0 != (gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right & GS_FEATURE_ALLOW_CALLER)) {
			strncpy(audio_record_code, gs_feature_map[GS_FEATURE_AUTOMIXMON].f_code, sizeof(audio_record_code));
		}

		if (0 != (gs_feature_map[GS_FEATURE_PARKCALL].f_right & GS_FEATURE_ALLOW_CALLER)) {
			strncpy(call_park_code, gs_feature_map[GS_FEATURE_PARKCALL].f_code, sizeof(call_park_code));
		}
	}
	else {
		if (0 != (gs_feature_map[GS_FEATURE_AUTOMIXMON].f_right & GS_FEATURE_ALLOW_CALLEE)) {
			strncpy(audio_record_code, gs_feature_map[GS_FEATURE_AUTOMIXMON].f_code, sizeof(audio_record_code));
		}

		if (0 != (gs_feature_map[GS_FEATURE_PARKCALL].f_right & GS_FEATURE_ALLOW_CALLEE)) {
			strncpy(call_park_code, gs_feature_map[GS_FEATURE_PARKCALL].f_code, sizeof(call_park_code));
		}
	}

	/* add header */
	if (strlen(audio_record_code) > 0) {
#ifdef GRANDSTREAM_NETWORKS
#ifdef BUILD_ON_CLOUD_PBX
		char result[16] = {0};
		int gdms_record_support = 0;

		ast_safe_nvram_get("gdms_record_support", result, sizeof(result) - 1);
		if (strlen(result) > 0) {
			gdms_record_support = atoi(result);
		}
		
		if ( gdms_record_support && (in_queue == NULL || !strcmp(in_queue, "no") || (feature_enable && !strcmp(feature_enable, "yes") && !strcmp(in_queue, "yes")))) {
#else
		if (in_queue == NULL || !strcmp(in_queue, "no") || (feature_enable && !strcmp(feature_enable, "yes") && !strcmp(in_queue, "yes"))) {
#endif // BUILD_ON_CLOUD_PBX
#endif // GRANDSTREAM_NETWORKS
			if (session->id.number.valid && !ast_strlen_zero(session->id.number.str)) {
				if (ast_channel_check_record_feature_whitelist(session->channel, session->id.number.str)) {
					ast_sip_add_header(tdata, "X-UCM-AudioRecord", audio_record_code);
				}
			} else {
				ast_sip_add_header(tdata, "X-UCM-AudioRecord", audio_record_code);
			}
#ifdef GRANDSTREAM_NETWORKS
		}
#endif
	}

	if (strlen(call_park_code) > 0) {
		ast_sip_add_header(tdata, "X-UCM-CallPark", call_park_code);
	}

	return;
}

/* add by jlhu for trunk hunt  begin */
#ifdef GRANDSTREAM_NETWORKS
void ast_sip_report_emergency_outbound(const char *endpoint_id, const char *emergency_num)
{
    char num[256] = {0};

    struct ast_security_event_emergency_outbound ast_emergency_outbound = {
		.common.event_type = AST_SECURITY_EVENT_EMERGENCY_OUTBOUND,
		.common.version = AST_SECURITY_EVENT_EMERGENCY_OUTBOUND_VERSION,
		.common.service = "PJSIP",
		.common.account_id = endpoint_id,
		.emergency_num = num,

    };

    ast_copy_string(num, emergency_num, sizeof(num));
    ast_log(LOG_DEBUG, "Emergency call occur, user_name:%s, emergency_number:%s \n", endpoint_id, num);
    ast_security_event_report(AST_SEC_EVT(&ast_emergency_outbound));
}
#endif
/* add by jlhu for trunk hunt end */

/* Add by bfdu, 2019.03.28 */
static void bfcp_create_params(struct ast_sip_session *session)
{
	if (!session->bfcp_para.has_udp_bfcp && !session->bfcp_para.has_tcp_bfcp) {
		session->bfcp_para.has_udp_bfcp = 1;
	}
	session->bfcp_para.confid = 1;
	session->bfcp_para.userid = 1;
	snprintf(session->bfcp_para.floorid, sizeof(session->bfcp_para.floorid), "1 mstrm:%d", DEFALUT_SLIDES_VIDEO_LABEL);
	strcpy(session->bfcp_para.local_floorctrl, "c-s");
	strcpy(session->bfcp_para.local_setup, "actpass");
	strcpy(session->bfcp_para.connection, "new");

	return;
}

static void bfcp_create_params_for_sfu_invite(struct ast_sip_session *session)
{
	session->bfcp_para.has_udp_bfcp = 1;
	session->bfcp_para.confid = 1;
	session->bfcp_para.userid = 1;
	snprintf(session->bfcp_para.floorid, sizeof(session->bfcp_para.floorid), "1 mstrm:%d", DEFALUT_SLIDES_VIDEO_LABEL);
	strcpy(session->bfcp_para.local_floorctrl, "s-only");
	strcpy(session->bfcp_para.local_setup, "passive");
	strcpy(session->bfcp_para.connection, "new");

	return;
}

static void bfcp_create_stream(struct ast_sip_session *session)
{
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "bfcp create stream, has no pending media\n");
		return;
	}

	bfcp_create_params(session);

	stream = ast_stream_alloc(ast_codec_media_type2str(AST_MEDIA_TYPE_APPLICATION), AST_MEDIA_TYPE_APPLICATION);
	if (!stream) {
		return;
	}

	ast_stream_set_sub_type(stream, AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP);

	if (ast_stream_topology_append_stream(session->pending_media_state->topology, stream) < 0) {
		ast_stream_free(stream);
	}

	return;
}

static void bfcp_set_info(struct ast_sip_session *session, struct ast_sip_session_media *bfcp_media, struct ast_bfcp_info *bfcp_info)
{
	struct ast_bfcp_info rsp_bfcp_info;
	memset(&rsp_bfcp_info, 0, sizeof(struct ast_bfcp_info));

	if (!session || !bfcp_info) {
		ast_log(LOG_WARNING, "bfcp_set_info param is NULL\n");
		return;
	}

	if (!bfcp_media ||
		(bfcp_media && bfcp_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP && !bfcp_media->bfcp_udp) ||
		(bfcp_media && bfcp_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP && !bfcp_media->bfcp_tcp)) {
		ast_log(LOG_DEBUG, "[session:%p] bfcp_media is already destoryed, just denied or released for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		switch (bfcp_info->primitive) {
			case BFCP_FLOOR_REQUEST:
				rsp_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
				rsp_bfcp_info.status = BFCP_DENIED;
				bfcp_info->status = BFCP_DENIED;
				strcpy(rsp_bfcp_info.account_id, bfcp_info->account_id);
				ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &rsp_bfcp_info, sizeof(rsp_bfcp_info));
				break;
			case BFCP_FLOOR_RELEASE:
				rsp_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
				rsp_bfcp_info.status = BFCP_RELEASED;
				ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &rsp_bfcp_info, sizeof(rsp_bfcp_info));
				break;
			default:
				break;
		}

		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] bfcp_set_info local floorctrl:%s, primitive:%d, status:%d for endpoint:%s\n",
		session, session->bfcp_para.local_floorctrl, bfcp_info->primitive, bfcp_info->status, ast_sorcery_object_get_id(session->endpoint));

	if (!strcasecmp(session->bfcp_para.local_floorctrl, "s-only")) {
		if (bfcp_media->bfcp_udp) {
			switch (bfcp_info->primitive) {
				case BFCP_FLOOR_REQUEST:
					bfcp_media->bfcp_udp->bfcp_info.primitive = BFCP_FLOOR_STATUS;
					bfcp_media->bfcp_udp->bfcp_info.status = BFCP_GRANTED;
					rsp_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
					rsp_bfcp_info.status = bfcp_media->bfcp_udp->bfcp_info.status;
					strcpy(rsp_bfcp_info.account_id, bfcp_info->account_id);
					if (strlen(bfcp_info->floor_owner_id)) {
						strcpy(bfcp_media->bfcp_udp->bfcp_info.floor_owner_id, bfcp_info->floor_owner_id);
					}
					ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &rsp_bfcp_info, sizeof(rsp_bfcp_info));
					break;
				case BFCP_FLOOR_RELEASE:
					bfcp_media->bfcp_udp->bfcp_info.primitive = BFCP_FLOOR_STATUS;
					bfcp_media->bfcp_udp->bfcp_info.status = BFCP_RELEASED;
					rsp_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
					rsp_bfcp_info.status = bfcp_media->bfcp_udp->bfcp_info.status;
					ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &rsp_bfcp_info, sizeof(rsp_bfcp_info));
					break;
				default:
					bfcp_media->bfcp_udp->bfcp_info.primitive = bfcp_info->primitive;
					bfcp_media->bfcp_udp->bfcp_info.status = bfcp_info->status;
					break;
			}
		} else if(bfcp_media->bfcp_tcp) {
			switch (bfcp_info->primitive) {
				case BFCP_FLOOR_REQUEST:
					bfcp_media->bfcp_tcp->bfcp_info.primitive = BFCP_FLOOR_STATUS;
					bfcp_media->bfcp_tcp->bfcp_info.status = BFCP_GRANTED;
					rsp_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
					rsp_bfcp_info.status = bfcp_media->bfcp_tcp->bfcp_info.status;
					if (strlen(bfcp_info->floor_owner_id)) {
						strcpy(bfcp_media->bfcp_tcp->bfcp_info.floor_owner_id, bfcp_info->floor_owner_id);
					}
					ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &rsp_bfcp_info, sizeof(rsp_bfcp_info));
					break;
				case BFCP_FLOOR_RELEASE:
					bfcp_media->bfcp_tcp->bfcp_info.primitive = BFCP_FLOOR_STATUS;
					bfcp_media->bfcp_tcp->bfcp_info.status = BFCP_RELEASED;
					rsp_bfcp_info.primitive = BFCP_FLOOR_REQUEST_STATUS;
					rsp_bfcp_info.status = bfcp_media->bfcp_tcp->bfcp_info.status;
					ast_queue_control_data(session->channel, AST_CONTROL_BFCP_CONTROL, &rsp_bfcp_info, sizeof(rsp_bfcp_info));
					break;
				default:
					bfcp_media->bfcp_tcp->bfcp_info.primitive = bfcp_info->primitive;
					bfcp_media->bfcp_tcp->bfcp_info.status = bfcp_info->status;
					break;
			}
		}
	} else {
		if (bfcp_media->bfcp_udp) {
			bfcp_media->bfcp_udp->bfcp_info.primitive = bfcp_info->primitive;
			bfcp_media->bfcp_udp->bfcp_info.status = bfcp_info->status;
			if (strlen(bfcp_info->floor_owner_id)) {
				strcpy(bfcp_media->bfcp_udp->bfcp_info.floor_owner_id, bfcp_info->floor_owner_id);
			}
		} else if(bfcp_media->bfcp_tcp) {
			bfcp_media->bfcp_tcp->bfcp_info.primitive = bfcp_info->primitive;
			bfcp_media->bfcp_tcp->bfcp_info.status = bfcp_info->status;
			if (strlen(bfcp_info->floor_owner_id)) {
				strcpy(bfcp_media->bfcp_tcp->bfcp_info.floor_owner_id, bfcp_info->floor_owner_id);
			}
		}
	}

	return;
}

static void notify_released_bfcp_for_transfer(struct ast_sip_session *session)
{
	struct ast_bfcp_info bfcp_info;

	if (!session) {
		return;
	}

	memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
	bfcp_info.primitive = BFCP_FLOOR_STATUS;

	if (ast_channel_get_recv_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Release slides video for transfer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_set_recv_slides_video(session->channel, 0);
		bfcp_info.status = BFCP_RELEASED;
		if (session->endpoint && session->endpoint->aors) {
			strncpy(bfcp_info.account_id, session->endpoint->aors, sizeof(bfcp_info.account_id) - 1);
		}
		ast_indicate_data(session->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));
	}

	return;
}

static void notify_trunk_info_for_transfer(struct ast_sip_session *session)
{
	struct ast_trunk_info trunk_info;

	if (!session || !session->channel) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_transfer param is NULL\n");
		return;
	}

	if (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		ast_log(LOG_DEBUG, "notify_trunk_info_for_transfer it's not trunk\n");
		return;
	}

	if (ast_strlen_zero(session->conf_link)) {
		ast_log(LOG_DEBUG, "[session:%p] notify_trunk_info_for_transfer can't get conf_link, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_strlen_zero(session->trunk_channel_id)) {
		ast_log(LOG_DEBUG, "[session:%p] notify_trunk_info_for_transfer can't get trunk_channel_id, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	memset(&trunk_info, 0, sizeof(struct ast_trunk_info));
	strncpy(trunk_info.conf_link, session->conf_link, sizeof(trunk_info.conf_link) - 1);
	strncpy(trunk_info.channel_id, session->trunk_channel_id, sizeof(trunk_info.channel_id) - 1);

	ast_log(LOG_DEBUG, "[session:%p] notify_trunk_info_for_transfer conf_link:%s, trunk_channel_id:%s, chan:%s, endpoint:%s\n",
		session, trunk_info.conf_link, trunk_info.channel_id,
		ast_channel_name(session->channel), ast_sorcery_object_get_id(session->endpoint));

	if (ast_queue_control_data(session->channel, AST_CONTROL_NOTIFY_TRUNK_INFO, &trunk_info, sizeof(trunk_info))) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_NOTIFY_TRUNK_INFO frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void notify_bfcp_for_refer_transfer(struct ast_sip_session *session)
{
	struct ast_bfcp_info bfcp_info;

	if (!session) {
		return;
	}

	memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
	bfcp_info.primitive = BFCP_FLOOR_STATUS;

	if (ast_channel_get_recv_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Notify to recv slides video for refer transfer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		bfcp_info.status = BFCP_GRANTED;
		ast_indicate_data(session->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));
	}

	if (ast_channel_get_send_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Release slides video for refer transfer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_set_recv_slides_video(session->channel, 0);
		bfcp_info.status = BFCP_RELEASED;
		ast_indicate_data(session->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));
	}

	return;
}

/* Begin add by bfdu for Bug 43278, 2015.05.20 */
static int judge_if_transport_compatible(const struct ast_sip_session *session, const pjmedia_sdp_media *stream)
{
	unsigned int use_received_transport = 0;
	enum ast_sip_application_type app_type = AST_SIP_APP_INVALID;
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;

	if (NULL == session || NULL == stream) {
		ast_log(LOG_WARNING, "judge_if_transport_compatible para is null\n");
		return 0;
	}
	/* FEC stream, should not use our config, just use received encryption type */
	if (ast_sip_session_judge_if_fec_stream(stream)) {
		return 1;
	}

	if (session->is_webrtc_call) {
		use_received_transport = session->endpoint->media.rtp.use_received_transport;
	} else {
		use_received_transport = session->endpoint->media.rtp.configure_use_received_transport;
	}

	app_type = ast_sip_session_get_application_type(stream->desc.transport);
	if (!use_received_transport &&
		AST_SIP_APP_UDP_BFCP != app_type && AST_SIP_APP_TCP_BFCP != app_type &&
		AST_SIP_APP_RTP != app_type) {
		encryption = ast_sip_session_check_endpoint_media_transport(session->endpoint, stream, session->is_webrtc_call);

		if (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID) {
			ast_log(LOG_WARNING, "[session:%p] Transport is not compatible for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 0;
		}
	}

	return 1;
}

static unsigned char judge_if_need_send_reinvite(struct ast_sip_session *session, const struct ast_sip_reinvite_data *req_data)
{
	unsigned int i = 0;
	const pjmedia_sdp_media *m = NULL;
	pjmedia_sdp_session *active_local_sdp = NULL;

	if (!session || !req_data || !req_data->session) {
		return 0;
	}

	if (!req_data->session->recv_reinvite_no_sdp) {
		ast_log(LOG_DEBUG, "[session:%p] It's not a reinvite with no sdp, just send to peer for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		return 0;
	}

	active_local_sdp = session->inv_session->neg->active_local_sdp;

	if (!active_local_sdp) {
		return 0;
	}

	/* For re-INVITE with no SDP, we just send re-INVITE to peer if preivous meida direction is not sendrecv */
	for (i = 0; i < active_local_sdp->media_count; i++) {
		m = active_local_sdp->media[i];
		if (m->desc.port != 0 &&
			(pjmedia_sdp_media_find_attr2(m, "sendonly", NULL) ||
			pjmedia_sdp_media_find_attr2(m, "recvonly", NULL) ||
			pjmedia_sdp_media_find_attr2(m, "inactive", NULL)))  {
			ast_log(LOG_DEBUG, "[session:%p] Can't find sendrecv, need send reinvite to peer when recv no sdp for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		}
	}

	return 0;
}

static unsigned char judge_if_is_audio_rtcp_fb_attr(char *attr_name)
{
	if (!attr_name) {
		return 0;
	}

	/* Now for audio media, we just support rtcp transparent, if need other, support later, 2017.07.13 */
	if (!strcmp(attr_name, "rtcp-fb")) {
		return 1;
	}

	return 0;
}

static unsigned char judge_if_need_update_rtcp_address(struct ast_stream *caller_stream, struct ast_stream *callee_stream)
{
	if (ast_stream_get_has_rtcp_address(caller_stream) && !ast_stream_get_has_rtcp_address(callee_stream)) {
		return TRUE;
	}

	return FALSE;
}

static unsigned char judge_if_need_update_nack_param(struct ast_stream *caller_stream, struct ast_stream *callee_stream)
{
	if (ast_stream_get_has_nack(caller_stream) && !ast_stream_get_has_nack(callee_stream)) {
		return TRUE;
	}

	if (ast_stream_get_has_nack_pli(caller_stream) && !ast_stream_get_has_nack_pli(callee_stream)) {
		return TRUE;
	}

	if (ast_stream_get_has_ccm_fir(caller_stream) && !ast_stream_get_has_ccm_fir(callee_stream)) {
		return TRUE;
	}

	return FALSE;
}

static void set_session_sfu_params(struct ast_sip_session *session, const char *conf_room)
{
	if (!session || !conf_room) {
		return;
	}

	if (session->inv_session && session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) {
		session->join_sfu = 1;
	}

	if (session->is_sfu) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Set is_sfu to yes, and conf_room to %s for endpoint:%s\n",
		session, conf_room, ast_sorcery_object_get_id(session->endpoint));

	session->is_sfu = 1;
	strncpy(session->conf_room, conf_room, sizeof(session->conf_room) - 1);

	return;
}

static void update_rtcp_attr(struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	int i = 0;
	pjmedia_sdp_attr *attr_old = NULL;
	char attr_name[16]   = {0};
	char attr_value[128] = {0};
	unsigned char update_rtcp_address = 0;
	unsigned char update_nack_param = 0;

	if (!(*p_result) || !caller_stream || !callee_stream) {
		ast_log(LOG_DEBUG, "Param is NULL, don't need to change rtcp attr\n");
		return;
	}

	ast_log(LOG_DEBUG, "update_rtcp_attr begin\n");

	update_rtcp_address = judge_if_need_update_rtcp_address(caller_stream, callee_stream);
	update_nack_param = judge_if_need_update_nack_param(caller_stream, callee_stream);

	if (update_rtcp_address) {
		attr_old = pjmedia_sdp_media_find_attr2(*p_result, "rtcp", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(*p_result, attr_old);
		}
	}

	if (!update_nack_param) {
		ast_log(LOG_DEBUG, "Don't need update nack param\n");
		return;
	}

	for (i = 0; i < (*p_result)->attr_count; i++) {
		memset(attr_name, 0, sizeof(attr_name));
		memset(attr_value, 0, sizeof(attr_value));
		ast_copy_pj_str(attr_name, &(*p_result)->attr[i]->name, sizeof(attr_name));
		ast_copy_pj_str(attr_value, &(*p_result)->attr[i]->value, sizeof(attr_value));

		if (judge_if_is_audio_rtcp_fb_attr(attr_name)) {
			if ((!ast_stream_get_has_nack(caller_stream) && strstr(attr_value, "nack")) ||
				(!ast_stream_get_has_nack_pli(caller_stream) && strstr(attr_value, "nack pli")) ||
				(!ast_stream_get_has_ccm_fir(caller_stream) && strstr(attr_value, "ccm fir"))) {
				pjmedia_sdp_media_remove_attr(*p_result, (*p_result)->attr[i]);
			} else {
				ast_log(LOG_DEBUG, "Can't support other rtcp-fb value:%s\n", attr_value);
			}
		}
	}

	return;
}

static void update_framerate_param(pj_pool_t *pool,
					struct ast_sip_session *local_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	const char *caller_framerate = ast_stream_get_metadata(caller_stream, "FRAMERATE");
	const char *callee_framerate = ast_stream_get_metadata(callee_stream, "FRAMERATE");

	if (!caller_framerate || !callee_framerate) {
		return;
	}

	if (!strcmp(caller_framerate, callee_framerate)) {
		return;
	}

	attr_old = pjmedia_sdp_media_find_attr2(*p_result, "framerate", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(*p_result, attr_old);
	}

	snprintf(tmp, sizeof(tmp), "%s", callee_framerate);
	attr_new = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
	pjmedia_sdp_media_add_attr(*p_result, attr_new);

	ast_log(LOG_DEBUG, "[session:%p] Update %s imageattr param:%s for endpoint:%s\n",
		local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
		tmp, ast_sorcery_object_get_id(local_session->endpoint));

	return;
}

static void add_as_to_media_for_basic_call(pj_pool_t *pool, struct ast_sip_session *local_session,
	struct ast_stream *callee_stream, pjmedia_sdp_media *media)
{
	int as_value = 0;
	const pj_str_t STR_BANDW_MODIFIER = { "AS", 2 };
	pjmedia_sdp_bandw *bandw = NULL;

	const char *callee_as = ast_stream_get_metadata(callee_stream, "AS");
	if (!callee_as) {
		return;
	}

	as_value = atoi(callee_as);
	if (!as_value) {
		return;
	}

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Add %s AS value:%d for endpoint:%s\n",
		local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(callee_stream)),
		as_value, ast_sorcery_object_get_id(local_session->endpoint));

	bandw->modifier = STR_BANDW_MODIFIER;
	bandw->value = as_value;
	media->bandw[media->bandw_count++] = bandw;

	return;
}

static void add_tias_to_media_for_basic_call(pj_pool_t *pool,  struct ast_sip_session *local_session,
	struct ast_stream *callee_stream, pjmedia_sdp_media *media)
{
	int tias_value = 0;
	const pj_str_t STR_BANDW_MODIFIER = { "TIAS", 4 };
	pjmedia_sdp_bandw *bandw = NULL;

	const char *callee_tias = ast_stream_get_metadata(callee_stream, "TIAS");
	if (!callee_tias) {
		return;
	}

	tias_value = atoi(callee_tias);
	if (!tias_value) {
		return;
	}

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Add %s TIAS value:%d for endpoint:%s\n",
		local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(callee_stream)),
		tias_value, ast_sorcery_object_get_id(local_session->endpoint));

	bandw->modifier = STR_BANDW_MODIFIER;
	bandw->value = tias_value;
	media->bandw[media->bandw_count++] = bandw;

	return;
}

static void update_as_tias_param(pj_pool_t *pool,
					struct ast_sip_session *local_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	if (!pool || !caller_stream || !callee_stream || !*p_result) {
		return;
	}

	(*p_result)->bandw_count = 0;

	add_as_to_media_for_basic_call(pool, local_session, callee_stream, *p_result);
	add_tias_to_media_for_basic_call(pool, local_session, callee_stream, *p_result);

	return;
}

static void update_imageattr_param(pj_pool_t *pool,
					struct ast_sip_session *local_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	unsigned int i = 0;
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *attr_old = NULL;
	const char *imageattr = NULL;

	attr_old = pjmedia_sdp_media_find_attr2(*p_result, "imageattr", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(*p_result, attr_old);
	}

	imageattr = ast_stream_get_metadata(callee_stream, "IMAGEATTR");
	if (ast_strlen_zero(imageattr)) {
		return;
	}

	for (i = 0; i < (*p_result)->desc.fmt_count; ++i) {
	    attr = pjmedia_sdp_media_find_attr2(*p_result, "rtpmap", &((*p_result)->desc.fmt[i]));
	    if (!attr) {
			continue;
		}

		pjmedia_sdp_rtpmap lr;
		pjmedia_sdp_attr_get_rtpmap(attr, &lr);
		if (pj_stricmp2(&lr.enc_name, "h264")) {
		    continue;
		}

		snprintf(tmp, sizeof(tmp), "%ld %s", pj_strtoul(&((*p_result)->desc.fmt[i])), imageattr);
		attr = pjmedia_sdp_attr_create(pool, "imageattr", pj_cstr(&stmp, tmp));
		(*p_result)->attr[(*p_result)->attr_count++] = attr;

		ast_log(LOG_DEBUG, "[session:%p] Update %s imageattr param:%s for endpoint:%s\n",
			local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
			tmp, ast_sorcery_object_get_id(local_session->endpoint));

		break;
	}

	return;
}

static void update_video_param(pj_pool_t *pool, struct ast_sip_session *local_session,
	struct ast_stream *caller_stream, struct ast_stream *callee_stream, pjmedia_sdp_media **p_result)
{
	update_framerate_param(pool, local_session, caller_stream, callee_stream, p_result);

	update_as_tias_param(pool, local_session, caller_stream, callee_stream, p_result);

	update_imageattr_param(pool, local_session, caller_stream, callee_stream, p_result);

	return;
}

static void update_bfcp_info_conf_id(struct ast_sip_session *session)
{
	struct ast_sip_session_media *bfcp_media = NULL;

	if (!session) {
		return;
	}

	const char *conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
	if (ast_strlen_zero(conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] update_bfcp_info_conf_id, endpoint:%s can't get conf room number\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	bfcp_media = ast_sip_session_get_bfcp_active_media(session);
	if (!bfcp_media) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no bfcp media\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (bfcp_media->bfcp_udp) {
		strncpy(bfcp_media->bfcp_udp->bfcp_info.info_conf_id, conf_room, sizeof(bfcp_media->bfcp_udp->bfcp_info.info_conf_id) - 1);
	} else if(bfcp_media->bfcp_tcp) {
		strncpy(bfcp_media->bfcp_tcp->bfcp_info.info_conf_id, conf_room, sizeof(bfcp_media->bfcp_tcp->bfcp_info.info_conf_id) - 1);
	}

	return;
}

static void update_sfu_params_for_join_conference(struct ast_sip_session *session)
{
	const char *is_sfu = NULL;
	const char *conf_room = NULL;
	const char *from_disa = NULL;
	const char *from_callback = NULL;

	if (!session || !session->channel) {
		return;
	}

	from_disa = pbx_builtin_getvar_helper(session->channel, "FROM_DISA");
	from_callback = pbx_builtin_getvar_helper(session->channel, "FROM_CALLBACK");

	is_sfu = pbx_builtin_getvar_helper(session->channel, "IS_SFU");

	if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
		ast_log(LOG_DEBUG, "[session:%p] Join SFU, set is_sfu true for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->is_sfu = 1;
		conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
		if (!ast_strlen_zero(conf_room)) {
			ast_log(LOG_DEBUG, "[session:%p] Join SFU, set conf_room:%s for endpoint:%s\n",
				session, conf_room, ast_sorcery_object_get_id(session->endpoint));
			strncpy(session->conf_room, conf_room, sizeof(session->conf_room) - 1);
		}
	}
	/*Start added by zyxia for Bug 303622 ,2024.8.7*/
	else if (strstr(ast_channel_name(session->channel), "PJSIP/trunk") && !ast_strlen_zero(from_disa) && !strcasecmp(from_disa, "yes") ) {
			RAII_VAR(struct ast_channel*, local_chan, ast_channel_bridge_peer(session->channel), ast_channel_cleanup);
			if(local_chan && strstr(ast_channel_name(local_chan), "Local/")){
				RAII_VAR(struct ast_channel*, local_remote_peer, ast_local_get_peer(local_chan), ast_channel_cleanup);
				if (local_remote_peer) {
					is_sfu = pbx_builtin_getvar_helper(local_remote_peer, "IS_SFU");
					conf_room = pbx_builtin_getvar_helper(local_remote_peer, "CONF_ROOM");
					if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes") && !ast_strlen_zero(conf_room)) {
						ast_log(LOG_DEBUG, "[session:%p] Join SFU, set conf_room:%s for endpoint:%s\n",
							session, conf_room, ast_sorcery_object_get_id(session->endpoint));
						pbx_builtin_setvar_helper(session->channel,"IS_SFU","yes");
						session->is_sfu = 1;
						pbx_builtin_setvar_helper(session->channel,"CONF_ROOM",conf_room);
						strncpy(session->conf_room, conf_room, sizeof(session->conf_room) - 1);
					}
				}
			}
	}
	/* inbound to callback and disa to speeddial  to conf,trunk--local/853700;2--local/853700;1--local0/2*/
	else if (strstr(ast_channel_name(session->channel), "PJSIP/trunk") && !ast_strlen_zero(from_callback) && !strcasecmp(from_callback, "yes") ) {
			RAII_VAR(struct ast_channel*, local_chan, ast_channel_bridge_peer(session->channel), ast_channel_cleanup);
			if(local_chan && strstr(ast_channel_name(local_chan), "Local/")){
				RAII_VAR(struct ast_channel*, local_remote_peer, ast_local_get_peer(local_chan), ast_channel_cleanup);
				if (local_remote_peer) {//Local/853007@from-internal-00000006;1

					from_disa = pbx_builtin_getvar_helper(local_remote_peer, "FROM_DISA");
					if(!ast_strlen_zero(from_disa) && !strcasecmp(from_disa, "yes")){
						RAII_VAR(struct ast_channel*, local_chan1, ast_channel_bridge_peer(local_remote_peer), ast_channel_cleanup);
						if(local_chan1){
							RAII_VAR(struct ast_channel*, local_chan2, ast_local_get_peer(local_chan1), ast_channel_cleanup);
							if (local_chan2) {//Local/0@from-internal-00000006;2
								is_sfu = pbx_builtin_getvar_helper(local_chan2, "IS_SFU");
								conf_room = pbx_builtin_getvar_helper(local_chan2, "CONF_ROOM");
								if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes") && !ast_strlen_zero(conf_room)) {
									ast_log(LOG_DEBUG, "[session:%p] Join SFU, set conf_room:%s for endpoint:%s\n",
										session, conf_room, ast_sorcery_object_get_id(session->endpoint));
									pbx_builtin_setvar_helper(session->channel,"IS_SFU","yes");
									session->is_sfu = 1;
									pbx_builtin_setvar_helper(session->channel,"CONF_ROOM",conf_room);
									strncpy(session->conf_room, conf_room, sizeof(session->conf_room) - 1);
								}
							}
						}
					}
				}
			}
	}
	/*End added by zyxia for Bug 303622 ,2024.8.7*/

	update_bfcp_info_conf_id(session);

	return;
}

static void ast_remove_all_media_directions(pjmedia_sdp_media *m)
{
	if (!m) {
		ast_log(LOG_WARNING, "Remove all media directions, media is null\n");
		return;
	}

	pjmedia_sdp_media_remove_all_attr(m, "inactive");
	pjmedia_sdp_media_remove_all_attr(m, "sendrecv");
	pjmedia_sdp_media_remove_all_attr(m, "sendonly");
	pjmedia_sdp_media_remove_all_attr(m, "recvonly");
}

static void ast_update_media_direction(pj_pool_t *pool, const struct ast_sip_session *local_session,
	pjmedia_sdp_media *local, const pjmedia_sdp_media *callee_answer)
{
	char dir[16] = { 0 };
	char media_type[32] = { 0 };

	if (!pool || !local || !callee_answer) {
		ast_log(LOG_WARNING, "Update media direction, param is null\n");
		return;
	}

	if (0 == callee_answer->desc.port) {
		ast_log(LOG_DEBUG, "Update media direction, callee media port is zero\n");
		return;
	}

	pjmedia_sdp_attr *a = NULL;

	if (pjmedia_sdp_media_find_attr2(callee_answer, "sendrecv", NULL) != NULL) {
		a = pjmedia_sdp_attr_create(pool, "sendrecv", NULL);
	} else if(pjmedia_sdp_media_find_attr2(callee_answer, "sendonly", NULL) != NULL) {
		a = pjmedia_sdp_attr_create(pool, "sendonly", NULL);
	} else if(pjmedia_sdp_media_find_attr2(callee_answer, "inactive", NULL) != NULL) {
		a = pjmedia_sdp_attr_create(pool, "inactive", NULL);
	} else if(pjmedia_sdp_media_find_attr2(callee_answer, "recvonly", NULL) != NULL) {
		a = pjmedia_sdp_attr_create(pool, "recvonly", NULL);
	} else {
		ast_log(LOG_WARNING, "Can't find media direction, just use sendrecv as default\n");
		a = pjmedia_sdp_attr_create(pool, "sendrecv", NULL);
	}

	if (a) {
		ast_remove_all_media_directions(local);
		pjmedia_sdp_media_add_attr(local, a);

		ast_copy_pj_str(dir, &a->name, sizeof(dir));
		ast_copy_pj_str(media_type, &local->desc.media, sizeof(media_type));

		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s update media direction to %s for media:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint), dir, media_type);
	}

	return;
}

static int remove_sdp_image_cb(struct ast_sip_session *session, pjmedia_sdp_session *sdp)
{
	int stream;
	int media_count = 0;

	for (stream = 0; stream < sdp->media_count; ++stream) {
		if (pj_strcmp2(&sdp->media[stream]->desc.media, "image")) {
			sdp->media[media_count++] = sdp->media[stream];
		}
	}

	if (media_count > 0) {
		sdp->media_count = media_count;
	}

	return 0;
}

static void reset_media_encrytion(struct ast_sip_session *session)
{
	unsigned int index = 0;
	unsigned int use_received_transport = 0;
	struct ast_sip_session_media *session_media = NULL;

	if (!session) {
		ast_log(LOG_WARNING, "reset_media_encrytion, param is invalid\n");
		return;
	}

	if (session->is_webrtc_call) {
		use_received_transport = session->endpoint->media.rtp.use_received_transport;
	} else {
		use_received_transport = session->endpoint->media.rtp.configure_use_received_transport;
	}

	if (use_received_transport) {
		ast_log(LOG_DEBUG, "[session:%p] Use received transport, don't reset media encryption for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->active_media_state) {
		ast_log(LOG_WARNING, "[session:%p] Reset media encrytion, active_media_state is NULL for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (index = 0; index < AST_VECTOR_SIZE(&session->active_media_state->sessions); ++index) {
		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, index);
		if (!session_media || session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_FEC) {
			continue;
		}

		if (session->is_webrtc_call) {
			session_media->encryption = session->endpoint->media.rtp.encryption;
		} else if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			session_media->encryption = AST_SIP_MEDIA_ENCRYPT_SDES;
		} else {
			session_media->encryption = session->endpoint->media.rtp.configure_encryption;
		}
	}

	return;
}

/* Add the following interfaces to make further codec negotiation between the caller and the callee leg. */
static void str_swap(pj_str_t *str1, pj_str_t *str2)
{
	pj_str_t tmp = *str1;
	*str1 = *str2;
	*str2 = tmp;
}

static void print_sdp_media(const pjmedia_sdp_media* media)
{
	unsigned j = 0;

	for (j = 0; j < media->desc.fmt_count; ++j) {
		ast_log(LOG_DEBUG, "media->desc.fmt[%d]: %ld\n", j, pj_strtoul(&media->desc.fmt[j]));
	}
}

static int get_pjmedia_sdp_attr_index (unsigned count,
						pjmedia_sdp_attr *const attr_array[],
						const pj_str_t *name,
						const pj_str_t *c_fmt)
{
	unsigned int i = 0;
	unsigned c_pt = 0xFFFF;

	if (c_fmt) {
		c_pt = pj_strtoul(c_fmt);
	}

	for (i = 0; i < count; ++i) {
		if (pj_strcmp(&attr_array[i]->name, name) == 0) {
			const pjmedia_sdp_attr *a = attr_array[i];
			if (c_fmt) {
				unsigned pt = (unsigned) pj_strtoul2(&a->value, NULL, 10);
				if (pt == c_pt) {
					return i;
				}
			} else {
				return i;
			}
		}
	}

	return -1;
}

static int index_of_pjmedia_sdp_attr(unsigned count,
						pjmedia_sdp_attr *const attr_array[],
						const char *c_name,
						const pj_str_t *c_fmt)
{
	pj_str_t name;

	name.ptr = (char*)c_name;
	name.slen = strlen(c_name);

	return get_pjmedia_sdp_attr_index(count, attr_array, &name, c_fmt);
}

static void arrange_pjmedia_sdp_attr_by_m_line(pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *attr = NULL;
	unsigned int j = 0;
	unsigned int attr_index = 0;
	unsigned int index = 0;

	for (j = 0; j < media->desc.fmt_count; j++) {
		index = index_of_pjmedia_sdp_attr(media->attr_count, media->attr, "rtpmap", &media->desc.fmt[j]);
		if (index > -1) {
			if (index != attr_index) {
				attr = media->attr[attr_index];
				media->attr[attr_index] = media->attr[index];
				media->attr[index] = attr;
			}
			attr_index++;
		}

		index = index_of_pjmedia_sdp_attr(media->attr_count, media->attr, "fmtp", &media->desc.fmt[j]);
		if (index > -1) {
			if (index != attr_index) {
				attr = media->attr[attr_index];
				media->attr[attr_index] = media->attr[index];
				media->attr[index] = attr;
			}
			attr_index++;
		}
	}
}

static void update_bfcp_conf_id(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	if (local_session->bfcp_para.confid == peer_session->bfcp_para.confid) {
		/* confid is already the same, don't need update */
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, confid is not the same, need change for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	local_session->bfcp_para.confid = peer_session->bfcp_para.confid;

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "confid", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (peer_session->bfcp_para.confid) {
		snprintf(tmp, sizeof(tmp), "%d", peer_session->bfcp_para.confid);
		attr_new = pjmedia_sdp_attr_create(pool, "confid", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp confid is zero, don't add confid attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static void update_bfcp_user_id(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	if (local_session->bfcp_para.userid == peer_session->bfcp_para.userid) {
		/* userid is already the same, don't need update */
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, userid is not the same, need change for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	local_session->bfcp_para.userid = peer_session->bfcp_para.userid;

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "userid", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (peer_session->bfcp_para.userid) {
		snprintf(tmp, sizeof(tmp), "%d", peer_session->bfcp_para.userid);
		attr_new = pjmedia_sdp_attr_create(pool, "userid", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp userid is zero, don't add userid attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static void update_bfcp_floor_id(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	if (local_session->bfcp_para.floorid == peer_session->bfcp_para.floorid) {
		/* floorid is already the same, don't need update */
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, floorid is not the same, need change for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	strcpy(local_session->bfcp_para.floorid, peer_session->bfcp_para.floorid);

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "floorid", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (strlen(peer_session->bfcp_para.floorid)) {
		snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.floorid);
		attr_new = pjmedia_sdp_attr_create(pool, "floorid", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp floorid is zero, don't add floorid attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static void update_bfcp_floor_ctrl(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, update floorctrl for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	strcpy(local_session->bfcp_para.local_floorctrl, peer_session->bfcp_para.remote_floorctrl);

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "floorctrl", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (strlen(peer_session->bfcp_para.remote_floorctrl)) {
		snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.remote_floorctrl);
		attr_new = pjmedia_sdp_attr_create(pool, "floorctrl", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp floorctrl is zero, don't add floorctrl attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static void update_bfcp_setup(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, update setup for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	strcpy(local_session->bfcp_para.local_setup, peer_session->bfcp_para.remote_setup);

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "setup", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (strlen(peer_session->bfcp_para.remote_setup)) {
		snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.remote_setup);
		attr_new = pjmedia_sdp_attr_create(pool, "setup", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp setup is zero, don't add setup attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static void update_bfcp_connection(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, update connection for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	strcpy(local_session->bfcp_para.connection, peer_session->bfcp_para.connection);

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "connection", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (strlen(peer_session->bfcp_para.connection)) {
		snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.connection);
		attr_new = pjmedia_sdp_attr_create(pool, "connection", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp connection is zero, don't add connection attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static void update_bfcp_fingerprint(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session)
{
	char tmp[512] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, update fingerprint for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	strcpy(local_session->bfcp_para.fingerprint, peer_session->bfcp_para.fingerprint);

	attr_old = pjmedia_sdp_media_find_attr2(caller_local, "fingerprint", NULL);
	if (attr_old) {
		pjmedia_sdp_media_remove_attr(caller_local, attr_old);
	}

	if (strlen(peer_session->bfcp_para.fingerprint)) {
		snprintf(tmp, sizeof(tmp), "%s", peer_session->bfcp_para.fingerprint);
		attr_new = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&stmp, tmp));
		pjmedia_sdp_media_add_attr(caller_local, attr_new);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] bfcp fingerprint is zero, don't add fingerprint attr for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
	}

	return;
}

static int match_callee_answer_app_media_bfcp(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					const pjmedia_sdp_media *callee_answer,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	if (0 == peer_session->bfcp_para.has_udp_bfcp && 0 == peer_session->bfcp_para.has_tcp_bfcp) {
		ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, peer bfcp is decline, local should decline too for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, caller_local);
		return 0;
	}

	if (AST_STREAM_STATE_REMOVED == ast_stream_get_state(callee_stream)) {
		ast_log(LOG_DEBUG, "[session:%p] match callee bfcp answer, peer bfcp is removed, local should removed too for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, caller_local);
		return 0;
	}

	update_bfcp_conf_id(pool, caller_local, local_session, peer_session);
	update_bfcp_user_id(pool, caller_local, local_session, peer_session);
	update_bfcp_floor_id(pool, caller_local, local_session, peer_session);
	update_bfcp_floor_ctrl(pool, caller_local, local_session, peer_session);
	update_bfcp_setup(pool, caller_local, local_session, peer_session);
	update_bfcp_connection(pool, caller_local, local_session, peer_session);
	update_bfcp_fingerprint(pool, caller_local, local_session, peer_session);

	*p_result = pjmedia_sdp_media_clone(pool, caller_local);
	return 0;
}

static int match_callee_answer_app_media_fecc(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					const pjmedia_sdp_media *callee_answer,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	/* now we only need to process peer disable fecc, if change fmtp param, support later */
	ast_log(LOG_DEBUG, "match_callee_answer_app_media_fecc begin\n");

	if (0 == callee_answer->desc.port) {
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, caller_local);
		return 0;
	}

	if (0 == caller_local->desc.port) {
		*p_result = pjmedia_sdp_media_clone(pool, caller_local);
		return 0;
	}

	update_rtcp_attr(caller_stream, callee_stream, p_result);

	*p_result = pjmedia_sdp_media_clone(pool, caller_local);

	return 0;
}

static int match_callee_answer_app_media(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					const pjmedia_sdp_media *callee_answer,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
					pjmedia_sdp_media **p_result)
{
	int res = -1;
	enum ast_media_sub_type sub_type = ast_stream_get_sub_type(caller_stream);
	switch(sub_type) {
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP:
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP:
			res = match_callee_answer_app_media_bfcp(pool, caller_local, callee_answer, local_session, peer_session, caller_stream, callee_stream, p_result);
			return res;
		case AST_MEDIA_SUB_APPLICATION_TYPE_FECC:
			res = match_callee_answer_app_media_fecc(pool, caller_local, callee_answer, local_session, peer_session, caller_stream, callee_stream, p_result);
			return res;
		default:
			break;
	}

	return res;
}

/* Now RTX and GS-FEC codec, we can negotiate independently */
static unsigned int is_codec_can_negotiate_independent(char *rtpmap_name)
{
	if (!rtpmap_name) {
		return 0;
	}

	if (!strncasecmp(rtpmap_name, "rtx", 3) ||
		!strncasecmp(rtpmap_name, "GS-FEC", 6) ||
		!strncasecmp(rtpmap_name, "RED", 3)) {
		return 1;
	}

	return 0;
}

static unsigned char is_pt_can_negotiate_independent(pjmedia_sdp_media *media, pj_str_t *pt)
{
	pjmedia_sdp_attr *a = NULL;
	pjmedia_sdp_rtpmap lr;
	char rtpmap_name[64] = { 0 };

	a = pjmedia_sdp_media_find_attr2(media, "rtpmap", pt);
	if (!a) {
		return 0;;
	}

	pjmedia_sdp_attr_get_rtpmap(a, &lr);
	ast_copy_pj_str(rtpmap_name, &lr.enc_name, sizeof(rtpmap_name));

	return is_codec_can_negotiate_independent(rtpmap_name);
}

static unsigned char is_rtpmap_can_negotiate_independent(pjmedia_sdp_rtpmap *rtpmap)
{
	char rtpmap_name[64] = { 0 };

	if (!rtpmap) {
		return 0;
	}

	ast_copy_pj_str(rtpmap_name, &(rtpmap->enc_name), sizeof(rtpmap_name));

	return is_codec_can_negotiate_independent(rtpmap_name);
}

static unsigned int media_has_multi_same_codec(pjmedia_sdp_media *media, pj_str_t enc_name)
{
	unsigned i = 0;
	pjmedia_sdp_attr *a = NULL;
	pjmedia_sdp_rtpmap ar;
	unsigned count = 0;

	/* We just need to compare video codec now */
	if (strncmp(media->desc.media.ptr, "video", 5)) {
		return 0;
	}

	if (pj_stricmp2(&enc_name, "H264") && pj_stricmp2(&enc_name, "RTX")) {
		return 0;
	}

	for (i = 0; i < media->desc.fmt_count; ++i) {
		a = pjmedia_sdp_media_find_attr2(media, "rtpmap", &media->desc.fmt[i]);
		if (!a) {
			continue;
		}

		pjmedia_sdp_attr_get_rtpmap(a, &ar);
		if (!pj_stricmp(&ar.enc_name, &enc_name)) {
			count++;
		}
	}

	if (count > 1) {
		return 1;
	}

	return 0;
}

static void change_fmtp_params_according_callee(pj_pool_t *pool, struct ast_sip_session *local_session,
	pjmedia_sdp_media *result, const pjmedia_sdp_media *callee_answer, unsigned int local_fmt_count)
{
	unsigned int i = 0;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *a_o_fmtp = NULL;
	pjmedia_sdp_attr *a_a_fmtp = NULL;
	pjmedia_sdp_attr *new_a_o_fmtp= NULL;
	pjmedia_sdp_rtpmap caller_rtpmap;
	pjmedia_sdp_rtpmap callee_rtpmap;
	char rtpmap_name[64] = { 0 };

	attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[local_fmt_count]);
	if (!attr) {
		ast_log(LOG_DEBUG, "[session:%p] It's has no rtpmap for caller, don't consider a dynamic payload, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	pjmedia_sdp_attr_get_rtpmap(attr, &caller_rtpmap);
	ast_copy_pj_str(rtpmap_name, &caller_rtpmap.enc_name, sizeof(rtpmap_name));
	if (is_codec_can_negotiate_independent(rtpmap_name) && !media_has_multi_same_codec(result, caller_rtpmap.enc_name)) {
		ast_log(LOG_DEBUG, "[session:%p] It's codec:%s can negotiate independently, don't need change fmtp, endpoint:%s\n",
			local_session, rtpmap_name, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	for (i = 0; i < callee_answer->desc.fmt_count; ++i) {
		attr = pjmedia_sdp_media_find_attr2(callee_answer, "rtpmap", &callee_answer->desc.fmt[i]);
		if (!attr) {
			ast_log(LOG_DEBUG, "[session:%p] It's has no rtpmap for caller, don't consider a video payload, endpoint:%s\n",
				local_session, ast_sorcery_object_get_id(local_session->endpoint));
			continue;
		}

		pjmedia_sdp_attr_get_rtpmap(attr, &callee_rtpmap);
		if (!pj_stricmp(&caller_rtpmap.enc_name, &callee_rtpmap.enc_name)) {
			break;
		}
	}

	if (i == callee_answer->desc.fmt_count) {
		ast_log(LOG_DEBUG, "[session:%p] Can't find matched codec:%s, endpoint:%s\n",
			local_session, rtpmap_name, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	/* need to update fmtp attributes to pass through some video attributes. */
	a_o_fmtp = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
	a_a_fmtp = pjmedia_sdp_media_find_attr2(callee_answer, "fmtp", &callee_answer->desc.fmt[i]);

	if (a_o_fmtp && a_a_fmtp) {
		pjmedia_sdp_media_remove_attr(result, a_o_fmtp);

		new_a_o_fmtp = pjmedia_sdp_attr_change_fmtp_payload(pool, a_a_fmtp, &callee_answer->desc.fmt[i], &result->desc.fmt[local_fmt_count]);

		if (new_a_o_fmtp) {
			pjmedia_sdp_media_add_attr(result, new_a_o_fmtp);
		}
	}

	return;
}

static void remove_fmtp_param_according_callee(pjmedia_sdp_media *result,
	struct ast_sip_session *local_session, unsigned int local_fmt_count)
{
	pjmedia_sdp_attr *a_rtpmap = NULL;
	pjmedia_sdp_attr *a_fmtp = NULL;
	pjmedia_sdp_rtpmap rtpmap;
	char rtpmap_name[64] = { 0 };

	ast_log(LOG_DEBUG, "[session:%p] Video media can't find local fmt:%ld in callee answer, just remove this attr for endpoint:%s\n",
		local_session, pj_strtoul(&result->desc.fmt[local_fmt_count]), ast_sorcery_object_get_id(local_session->endpoint));

	a_rtpmap = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[local_fmt_count]);
	if (!a_rtpmap) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get rtpmap, just remove fmtp endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		a_fmtp = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
		if (NULL != a_fmtp) {
			pjmedia_sdp_media_remove_attr(result, a_fmtp);
		}
		return;
	}

	pjmedia_sdp_attr_get_rtpmap(a_rtpmap, &rtpmap);
	ast_copy_pj_str(rtpmap_name, &rtpmap.enc_name, sizeof(rtpmap_name));
	if (is_codec_can_negotiate_independent(rtpmap_name) && !media_has_multi_same_codec(result, rtpmap.enc_name)) {
		ast_log(LOG_DEBUG, "[session:%p] It's codec:%s can negotiate independently, don't need remove fmtp, endpoint:%s\n",
			local_session, rtpmap_name, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	if (NULL != a_rtpmap) {
		pjmedia_sdp_media_remove_attr(result, a_rtpmap);
	}

	a_fmtp = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
	if (NULL != a_fmtp) {
		pjmedia_sdp_media_remove_attr(result, a_fmtp);
	}

	return;
}

/* If caller local SDP has no fmtp value, just use callee's SDP fmtp */
static void add_fmtp_param_according_callee(pj_pool_t *pool, struct ast_sip_session *local_session,
	struct ast_stream *caller_stream, pjmedia_sdp_media *result, const pjmedia_sdp_media *callee_answer)
{
	int local_fmt_count = 0;
	int remote_fmt_count = 0;
	int local_has_no_fmtp = 0;

	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *local_fmtp = NULL;
	pjmedia_sdp_attr *remote_fmtp = NULL;
	pjmedia_sdp_attr *new_fmtp = NULL;

	pjmedia_sdp_rtpmap local_rtpmap;
	pjmedia_sdp_rtpmap remote_rtpmap;

	for (remote_fmt_count = 0; remote_fmt_count < callee_answer->desc.fmt_count; ++remote_fmt_count) {
		attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &callee_answer->desc.fmt[remote_fmt_count]);
		if (!attr) {
			ast_log(LOG_DEBUG, "[session:%p] Media is %s, can't get callee rtpmap, endpoint:%s\n",
				local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
				ast_sorcery_object_get_id(local_session->endpoint));
			continue;
		}

		pjmedia_sdp_attr_get_rtpmap(attr, &remote_rtpmap);

		remote_fmtp = pjmedia_sdp_media_find_attr2(callee_answer, "fmtp", &callee_answer->desc.fmt[remote_fmt_count]);
		if (!remote_fmtp) {
			ast_log(LOG_DEBUG, "[session:%p] Media is %s, can't get callee fmtp, endpoint:%s\n",
				local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
				ast_sorcery_object_get_id(local_session->endpoint));
			continue;
		}

		for (local_fmt_count = 0; local_fmt_count < result->desc.fmt_count; ++local_fmt_count) {
			attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[local_fmt_count]);
			if (!attr) {
				ast_log(LOG_DEBUG, "[session:%p] Media is %s, can't get caller rtpmap, endpoint:%s\n",
					local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
					ast_sorcery_object_get_id(local_session->endpoint));
				continue;
			}

			pjmedia_sdp_attr_get_rtpmap(attr, &local_rtpmap);

			if (pj_stricmp(&local_rtpmap.enc_name, &remote_rtpmap.enc_name)) {
				continue;
			}

			if (pj_stricmp2(&local_rtpmap.enc_name, "H264")) {
				continue;
			}

			local_fmtp = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
			if (!local_fmtp) {
				local_has_no_fmtp = 1;
				ast_log(LOG_DEBUG, "[session:%p] Media is %s, local has no H264 fmtp, just use callee, endpoint:%s\n",
					local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
					ast_sorcery_object_get_id(local_session->endpoint));
			}

			break;
		}

		if (local_has_no_fmtp) {
			new_fmtp = pjmedia_sdp_attr_change_fmtp_payload(pool, remote_fmtp, &callee_answer->desc.fmt[remote_fmt_count], &result->desc.fmt[local_fmt_count]);
			if (new_fmtp) {
				pjmedia_sdp_media_add_attr(result, new_fmtp);
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Media is %s, create new fmtp failed, endpoint:%s\n",
					local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
					ast_sorcery_object_get_id(local_session->endpoint));
			}

			break;
		}
	}

	return;
}

/* video payload and attrs need to modify result according to callee answer */
static void match_video_payload_callee_answer(pj_pool_t *pool, struct ast_sip_session *local_session,
	struct ast_stream *caller_stream, pjmedia_sdp_media *result, const pjmedia_sdp_media *callee_answer,
	int pt_answer_count, pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT])
{
	int res_fmt_count = 0;
	int local_fmt_count = 0;
	int fmt_found_in_callee = FALSE;

	ast_log(LOG_DEBUG, "[session:%p] Media is %s, just use callee answer for endpoint:%s\n",
		local_session, ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)),
		ast_sorcery_object_get_id(local_session->endpoint));

	/* to remove rtpmap/fmtp for payload can't find in callee answer */
	for (local_fmt_count = 0; local_fmt_count < result->desc.fmt_count; ++local_fmt_count) {
		for(res_fmt_count = 0; res_fmt_count < pt_answer_count; ++res_fmt_count) {
			if (pj_strtoul(&pt_local[res_fmt_count]) == pj_strtoul(&result->desc.fmt[local_fmt_count])) {
				fmt_found_in_callee = TRUE;
				break;
			}
		}

		if (fmt_found_in_callee) {
			change_fmtp_params_according_callee(pool, local_session, result, callee_answer, local_fmt_count);
			fmt_found_in_callee = FALSE;
		} else {
			remove_fmtp_param_according_callee(result, local_session, local_fmt_count);
		}
	}

	add_fmtp_param_according_callee(pool, local_session, caller_stream, result, callee_answer);

	/* to remove payload in m line */
	result->desc.fmt_count = pt_answer_count;
	for (res_fmt_count = 0; res_fmt_count < pt_answer_count; ++res_fmt_count) {
		result->desc.fmt[res_fmt_count] = pt_local[res_fmt_count];
	}

	return;
}

static void add_audio_payload_for_can_transcode(struct ast_sip_session *local_session,
	enum ast_media_type media_type, pjmedia_sdp_media *caller_local,
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT], unsigned int *pt_answer_count)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int find_codec = 0;
	pjmedia_sdp_attr *a = NULL;
	pjmedia_sdp_rtpmap lr;
	char rtpmap_name[64] = { 0 };

	if (media_type != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	for (i = 0; i < caller_local->desc.fmt_count; ++i) {
		a = pjmedia_sdp_media_find_attr2(caller_local, "rtpmap", &caller_local->desc.fmt[i]);
		if (!a) {
			continue;
		}

		pjmedia_sdp_attr_get_rtpmap(a, &lr);
		ast_copy_pj_str(rtpmap_name, &lr.enc_name, sizeof(rtpmap_name));

		find_codec = 0;
		for (j = 0; j < *pt_answer_count; ++j) {
			if (pj_strtoul(&pt_local[j]) == pj_strtoul(&caller_local->desc.fmt[i])) {
				find_codec = 1;
				break;
			}
		}

		if (!find_codec) {
			ast_log(LOG_DEBUG, "[session:%p] It's codec:%s can transcode, and can't match before, need to add to sdp, endpoint:%s\n",
				local_session, rtpmap_name, ast_sorcery_object_get_id(local_session->endpoint));
			pt_local[(*pt_answer_count)++] = caller_local->desc.fmt[i];
		}
	}

	return;
}

static void append_rtx_pt_for_matched_apt_value(struct ast_sip_session *local_session,
	pjmedia_sdp_media *caller_local, pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT],
	unsigned int *pt_answer_count, pjmedia_sdp_rtpmap lr, unsigned int position)
{
	int i = 0;
	int find = 0;
	char fmt_param[256] = { 0 };
	char apt_str[8] = { 0 };
	char *attribs = NULL;
	char *attrib = NULL;
	pjmedia_sdp_attr *a = NULL;
	pjmedia_sdp_fmtp fmtp;

	if (!(a = pjmedia_sdp_media_find_attr2(caller_local, "fmtp", &lr.pt))) {
		ast_log(LOG_WARNING, "[session:%p] Can't find according rtx fmtp attr, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	if ((pjmedia_sdp_attr_get_fmtp(a, &fmtp)) != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "[session:%p] Can't get rtx fmtp, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));
	ast_log(LOG_DEBUG, "[session:%p] Get rtx fmtp param:%s, endpoint:%s\n",
			local_session, fmt_param, ast_sorcery_object_get_id(local_session->endpoint));

	attribs = ast_strdupa(fmt_param);
	while ((attrib = strsep(&attribs, ";"))) {
		attrib = ast_strip(attrib);
		if (sscanf(attrib, "apt=%s", apt_str) == 1) {
			break;
		}
	}

	for (i = 0; i < *pt_answer_count; i++) {
		if (!pj_strnicmp2(&pt_local[i], apt_str, sizeof(apt_str))) {
			find = 1;
			break;
		}
	}

	if (find) {
		ast_log(LOG_DEBUG, "[session:%p] Find matched video PT, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		pt_local[(*pt_answer_count)++] = caller_local->desc.fmt[position];
	}

	return;
}

/* For RTX and audio/video FEC we can negotiate, don't remove if callee is not support */
static void add_payload_for_negotiate_independent(struct ast_sip_session *local_session,
	pjmedia_sdp_media *caller_local, pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT], unsigned int *pt_answer_count)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int find_neg_independent_codec = 0;
	pjmedia_sdp_attr *a = NULL;
	pjmedia_sdp_rtpmap lr;
	char rtpmap_name[64] = { 0 };

	for (i = 0; i < caller_local->desc.fmt_count; ++i) {
		a = pjmedia_sdp_media_find_attr2(caller_local, "rtpmap", &caller_local->desc.fmt[i]);
		if (!a) {
			continue;
		}

		pjmedia_sdp_attr_get_rtpmap(a, &lr);
		ast_copy_pj_str(rtpmap_name, &lr.enc_name, sizeof(rtpmap_name));
		if (!is_codec_can_negotiate_independent(rtpmap_name)) {
			continue;
		}

		find_neg_independent_codec = 0;
		for (j = 0; j < *pt_answer_count; ++j) {
			if (pj_strtoul(&pt_local[j]) == pj_strtoul(&caller_local->desc.fmt[i])) {
				find_neg_independent_codec = 1;
				break;
			}
		}

		if (!find_neg_independent_codec) {
			ast_log(LOG_DEBUG, "[session:%p] It's codec:%s can negotiate independently, and can't match before, need to add to sdp, endpoint:%s\n",
				local_session, rtpmap_name, ast_sorcery_object_get_id(local_session->endpoint));
			if (!strcasecmp(rtpmap_name, "RTX") && media_has_multi_same_codec(caller_local, lr.enc_name)) {
				append_rtx_pt_for_matched_apt_value(local_session, caller_local, pt_local, pt_answer_count, lr, i);
			} else {
				pt_local[(*pt_answer_count)++] = caller_local->desc.fmt[i];
			}
		}
	}

	return;
}

static void get_matched_local_static_payload(pjmedia_sdp_media *caller_local,
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT], unsigned int *pt_answer_count, unsigned int callee_pt)
{
	/* For static payload type, it's enough to compare just
	 * the payload number.
	 */
	unsigned int i = 0;
	unsigned int caller_pt = 0;

	/* Find matching codec in local descriptor. */
	for (i = 0; i < caller_local->desc.fmt_count; ++i) {
		caller_pt = pj_strtoul(&caller_local->desc.fmt[i]);
		if (caller_pt == callee_pt && pj_isdigit(*caller_local->desc.fmt[i].ptr)) {
			pt_local[(*pt_answer_count)++] = caller_local->desc.fmt[i];
			break;
		}
	}

	return;
}

static unsigned int compare_is_the_same_rtpmap(pjmedia_sdp_media *offer,
	const pjmedia_sdp_media *answer, pj_str_t enc_name, unsigned i, unsigned j)
{
	if (!media_has_multi_same_codec(offer, enc_name)) {
		return 1;
	}

	if (pj_strtoul(&offer->desc.fmt[i]) == pj_strtoul(&answer->desc.fmt[j])) {
		return 1;
	}

	return 0;
}

static void get_matched_local_dynamic_payload(struct ast_sip_session *local_session,
	pjmedia_sdp_media *caller_local, const pjmedia_sdp_media *callee_answer,
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT], unsigned int *pt_answer_count,
	unsigned int callee_pt, unsigned int callee_postion)
{
	/* This is dynamic payload type.
	 * For dynamic payload type, we must look the rtpmap and compare the encoding name.
	 */

	unsigned int i = 0;
	const pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap callee_rtpmap;
	pjmedia_sdp_rtpmap caller_rtpmap;

	/* Get the rtpmap for the payload type. */
	attr = pjmedia_sdp_media_find_attr2(callee_answer, "rtpmap", &callee_answer->desc.fmt[callee_postion]);
	if (!attr) {
		ast_log(LOG_WARNING, "[session:%p] It's dynamic payload:%d, but has no rtpmap, endpoint:%s\n",
			local_session, callee_pt, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	pjmedia_sdp_attr_get_rtpmap(attr, &callee_rtpmap);

	/* Find payload in our initial SDP with matching encoding name and clock rate. */
	for (i = 0; i < caller_local->desc.fmt_count; ++i) {
	    attr = pjmedia_sdp_media_find_attr2(caller_local, "rtpmap", &caller_local->desc.fmt[i]);
	    if (!attr) {
	    	ast_log(LOG_DEBUG, "[session:%p] It's has no rtpmap for caller, don't consider a dynamic payload, endpoint:%s\n",
	    		local_session, ast_sorcery_object_get_id(local_session->endpoint));
			continue;
	    }
		pjmedia_sdp_attr_get_rtpmap(attr, &caller_rtpmap);

		if (is_rtpmap_can_negotiate_independent(&caller_rtpmap)) {
			break;
		}

		/* See if encoding name, clock rate, and channel count match */
		if (!pj_stricmp(&callee_rtpmap.enc_name, &caller_rtpmap.enc_name) &&
			callee_rtpmap.clock_rate == caller_rtpmap.clock_rate &&
			(pj_stricmp(&callee_rtpmap.param, &caller_rtpmap.param) == 0 ||
			(caller_rtpmap.param.slen == 0 && callee_rtpmap.param.slen == 1 && *callee_rtpmap.param.ptr == '1') ||
			(callee_rtpmap.param.slen == 0 && caller_rtpmap.param.slen == 1 && *caller_rtpmap.param.ptr == '1')) &&
			compare_is_the_same_rtpmap(caller_local, callee_answer, caller_rtpmap.enc_name, i, callee_postion)) {
		    /* Match! */
			pt_local[(*pt_answer_count)++] = caller_local->desc.fmt[i];
			break;
		}
	}
}

static void get_matched_local_payload(pj_pool_t *pool,
					struct ast_sip_session *local_session,
					pjmedia_sdp_media *caller_local,
					const pjmedia_sdp_media *callee_answer,
					pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT],
					unsigned int *pt_answer_count)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int callee_pt = 0;

	for (i = 0; i < callee_answer->desc.fmt_count; ++i) {
		if (pj_isdigit(*callee_answer->desc.fmt[i].ptr)) {
			/* This is normal/standard payload type, where it's identified
			 * by payload number.
			 */
			callee_pt = pj_strtoul(&callee_answer->desc.fmt[i]);
			if (callee_pt < 96) {
				get_matched_local_static_payload(caller_local, pt_local, pt_answer_count, callee_pt);
			} else {
				get_matched_local_dynamic_payload(local_session, caller_local, callee_answer, pt_local, pt_answer_count, callee_pt, i);
			}
		} else {
			/* This is a non-standard, brain damaged SDP where the payload
			 * type is non-numeric. It exists e.g. in Microsoft RTC based
			 * UA, to indicate instant messaging capability.
			 * Example:
			 *	- m=x-ms-message 5060 sip null
			 */
			for (j = 0; j < caller_local->desc.fmt_count; ++j) {
				if (!pj_strcmp(&callee_answer->desc.fmt[i], &caller_local->desc.fmt[j])) {
					/* Match */
					pt_local[(*pt_answer_count)++] = caller_local->desc.fmt[j];
					break;
				}
			}
		}
	}

	return;
}

static void change_dtmf_payload_position_to_last(pjmedia_sdp_media *result)
{
	unsigned int i = 0;
	unsigned int got_dtmf = 0;
	const pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap rtpmap;

	for (i = 0; i < result->desc.fmt_count; ++i) {
		if (i == (result->desc.fmt_count - 1)) {
			/* reach the end, don't need process */
			break;
		}

		if (got_dtmf) {
			str_swap(&result->desc.fmt[i], &result->desc.fmt[i + 1]);
			continue;
		}

		attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[i]);
		if (!attr) {
	    	ast_log(LOG_DEBUG, "It's has no rtpmap for caller, don't consider a dynamic payload\n");
			continue;
		}

		pjmedia_sdp_attr_get_rtpmap(attr, &rtpmap);

		if (!pj_stricmp2(&rtpmap.enc_name, "telephone-event")) {
			str_swap(&result->desc.fmt[i], &result->desc.fmt[i + 1]);
			got_dtmf = 1;
		}
	}

	return;
}

static void change_caller_payload_order_according_callee(struct ast_sip_session *local_session,
	struct ast_stream *caller_stream, pjmedia_sdp_media *result,
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT], unsigned int pt_answer_count)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int find = 0;

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, media:%s before change payload order according callee, result is:\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint),
		ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)));
	print_sdp_media(result);

	for (i = 0; i < pt_answer_count; ++i) {
		for (j = i; j < result->desc.fmt_count; ++j) {
			if (!pj_strcmp(&result->desc.fmt[j], &pt_local[i])) {
				break;
			}
			if (is_pt_can_negotiate_independent(result, &result->desc.fmt[j])) {
				find = 1;
				break;
			}
		}
		if (find == 1) {
			ast_log(LOG_DEBUG, "[session:%p] It's special code like FEC/RTX for endpoint:%s, don't need change\n",
				local_session, ast_sorcery_object_get_id(local_session->endpoint));
			find = 0;
			continue;
		}
		if (j == result->desc.fmt_count) {
			ast_log(LOG_ERROR, "[session:%p] Can't find matched fmt in pt_local for endpoint:%s\n",
				local_session, ast_sorcery_object_get_id(local_session->endpoint));
			continue;
		}

		ast_log(LOG_DEBUG, "pt_local[%d]: %ld, swap format position: (result->desc.fmt[%d])%ld <==> (result->desc.fmt[%d])%ld\n",
			i, pj_strtoul(&pt_local[i]), i, pj_strtoul(&result->desc.fmt[i]), j, pj_strtoul(&result->desc.fmt[j]));
		str_swap(&result->desc.fmt[i], &result->desc.fmt[j]);
	}

	/* In normal case, DTMF payload is alway in the last postion in the m line */
	change_dtmf_payload_position_to_last(result);

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, media:%s after change payload order according callee, result is:\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint),
		ast_codec_media_sub_type2str(ast_stream_get_sub_type(caller_stream)));
	print_sdp_media(result);

	return;
}

static void change_video_params_according_callee(pj_pool_t *pool,
					const pjmedia_sdp_media *callee_answer,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
				  	enum ast_media_type media_type,
					pjmedia_sdp_media *result,
					pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT],
					unsigned int pt_answer_count)
{
	if (media_type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	match_video_payload_callee_answer(pool, local_session, caller_stream, result, callee_answer, pt_answer_count, pt_local);

	update_video_param(pool, local_session, caller_stream, callee_stream, &result);

	ast_sip_session_modify_outgoing_yealink_sdp(local_session, pool, result);

	return;
}

static unsigned char judge_if_support_rtx_video_codec(char *name)
{
	if (!name) {
		return 0;
	}

	if (!strcasecmp(name, "h264") ||
		!strcasecmp(name, "h265") ||
		!strcasecmp(name, "vp8") ||
		!strcasecmp(name, "vp9")) {
		return 1;
	}

	return 0;
}

static unsigned int get_first_video_pt(pj_pool_t *pool, pjmedia_sdp_media *result)
{
	unsigned int i = 0;
	char name[16] = { 0 };
	int first_video_pt = 0;
	char first_video_pt_str[8] = { 0 };
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;

	for (i = 0; i < result->desc.fmt_count; i++) {
		if (!(attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[i]))) {
			continue;
		}

		if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
		if (judge_if_support_rtx_video_codec(name)) {
			ast_copy_pj_str(first_video_pt_str, &result->desc.fmt[i], sizeof(first_video_pt_str));
			first_video_pt = atoi(first_video_pt_str);
			break;
		}
	}

	return first_video_pt;
}

static void get_old_rtx_params(pj_pool_t *pool, struct ast_sip_session *local_session,
	pjmedia_sdp_media *result, int *rtx_pt, int *apt_val, int *rtx_time)
{
	unsigned int i = 0;
	char name[16] = { 0 };
	char rtx_pt_str[8] = { 0 };
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	char *attrib = NULL;
	char *attribs = NULL;
	unsigned int val = 0;
	char fmtp_params[64] = { 0 };
	pjmedia_sdp_fmtp fmtp;

	for (i = 0; i < result->desc.fmt_count; i++) {
		if (!(attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[i]))) {
			continue;
		}

		if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
		if (strcmp(name, "rtx")) {
			continue;
		}

		ast_copy_pj_str(rtx_pt_str, &rtpmap->pt, sizeof(rtx_pt_str));
		*rtx_pt = atoi(rtx_pt_str);

		if (!(attr = pjmedia_sdp_media_find_attr2(result, "fmtp", &rtpmap->pt))) {
			continue;
		}

		if ((pjmedia_sdp_attr_get_fmtp(attr, &fmtp)) == PJ_SUCCESS) {
			ast_copy_pj_str(fmtp_params, &fmtp.fmt_param, sizeof(fmtp_params));
			attribs = ast_strdupa(fmtp_params);
			while ((attrib = strsep(&attribs, ";"))) {
				attrib = ast_strip(attrib);
				if (sscanf(attrib, "apt=%30u", &val) == 1) {
					*apt_val = val;
				} else if (sscanf(attrib, "rtx-time=%30u", &val) == 1) {
					*rtx_time = val;
				}
			}
		}

		break;
	}

	return;
}

static int get_new_rtx_pt(pj_pool_t *pool, struct ast_sip_session *local_session,
	struct ast_stream *caller_stream, int first_video_pt, pjmedia_sdp_media *result)
{
	char first_video_rtpmap[8] = { 0 };
	pj_str_t first_video_str;
	char name[16] = { 0 };
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	int according_rtx_pt = 0;

	snprintf(first_video_rtpmap, sizeof(first_video_rtpmap), "%d", first_video_pt);
	first_video_str = pj_str(first_video_rtpmap);
	if (!(attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &first_video_str))) {
		return 0;
	}

	if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
		return 0;
	}

	ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));

	according_rtx_pt = ast_stream_get_video_rtx_pt(caller_stream, name);

	ast_log(LOG_DEBUG, "[session:%p] get_new_rtx_pt:%d, endpoint:%s\n",
		local_session, according_rtx_pt, ast_sorcery_object_get_id(local_session->endpoint));

	return according_rtx_pt;
}

static void modify_rtx_rtpmap_params(pj_pool_t *pool,
	struct ast_sip_session *local_session, pjmedia_sdp_media *result, int old_rtx_pt, int new_rtx_pt)
{
	unsigned int i = 0;
	char rtx_str[8] = { 0 };
	pj_str_t rtx_pj_str;
	pjmedia_sdp_rtpmap rtpmap;
	pjmedia_sdp_attr *attr = NULL;

	if (new_rtx_pt == old_rtx_pt) {
		ast_log(LOG_DEBUG, "[session:%p] modify_rtx_rtpmap_params RTX PT already matched, don't need change, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	for (i = 0; i < result->desc.fmt_count; i++) {
		if (pj_strtoul(&result->desc.fmt[i]) == old_rtx_pt) {
			snprintf(rtx_str, sizeof(rtx_str), "%d", new_rtx_pt);
			pj_strdup2(pool, &result->desc.fmt[i], rtx_str);
			break;
		}
	}

	if (i == result->desc.fmt_count) {
		ast_log(LOG_DEBUG, "[session:%p] change_rtx_apt_value_according_callee can't matched old rtx pt, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	memset(rtx_str, 0, sizeof(rtx_str));
	snprintf(rtx_str, sizeof(rtx_str), "%d", old_rtx_pt);
	rtx_pj_str = pj_str(rtx_str);
	if ((attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &rtx_pj_str))) {
		pjmedia_sdp_attr_remove(&result->attr_count, result->attr, attr);
	}

	rtpmap.pt = result->desc.fmt[i];
	rtpmap.clock_rate = ast_rtp_lookup_sample_rate2(1, ast_format_rtx, 0);
	pj_strdup2(pool, &rtpmap.enc_name, "RTX");

	pjmedia_sdp_rtpmap_to_attr(pool, &rtpmap, &attr);
	pjmedia_sdp_media_add_attr(result, attr);

	return;
}

static void modify_rtx_fmtp_params(pj_pool_t *pool,
	struct ast_sip_session *local_session, pjmedia_sdp_media *result,
	int rtx_time, int old_rtx_pt, int new_rtx_pt, int first_video_pt)
{
	char new_rtx_fmtp[64] = { 0 };
	char rtx_fmtp[8] = { 0 };
	pj_str_t rtx_pj_str;
	pj_str_t fmtp_pj_str;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *new_fmtp_attr = NULL;

	if (rtx_time == 0) {
		snprintf(new_rtx_fmtp, sizeof(new_rtx_fmtp), "%d apt=%d", new_rtx_pt, first_video_pt);
	} else {
		snprintf(new_rtx_fmtp, sizeof(new_rtx_fmtp), "%d apt=%d rtx-time=%d", new_rtx_pt, first_video_pt, rtx_time);
	}

	ast_log(LOG_DEBUG, "[session:%p] modify_rtx_fmtp_params new_rtx_fmtp:%s, endpoint:%s\n",
		local_session, new_rtx_fmtp, ast_sorcery_object_get_id(local_session->endpoint));

	snprintf(rtx_fmtp, sizeof(rtx_fmtp), "%d", old_rtx_pt);
	rtx_pj_str = pj_str(rtx_fmtp);
	if ((attr = pjmedia_sdp_media_find_attr2(result, "fmtp", &rtx_pj_str))) {
		pjmedia_sdp_attr_remove(&result->attr_count, result->attr, attr);
	}
	new_fmtp_attr = pjmedia_sdp_attr_create(pool, "fmtp", pj_cstr(&fmtp_pj_str, new_rtx_fmtp));
	if (new_fmtp_attr) {
		pjmedia_sdp_media_add_attr(result, new_fmtp_attr);
	}

	return;
}

static void modify_rtx_params_for_pt_change(pj_pool_t *pool,
	struct ast_sip_session *local_session, pjmedia_sdp_media *result,
	int old_rtx_pt, int new_rtx_pt, int rtx_time, int first_video_pt)
{
	modify_rtx_rtpmap_params(pool, local_session, result, old_rtx_pt, new_rtx_pt);
	modify_rtx_fmtp_params(pool, local_session, result, rtx_time, old_rtx_pt, new_rtx_pt, first_video_pt);

	return;
}

static void change_rtx_apt_value_according_callee(pj_pool_t *pool,
		struct ast_sip_session *local_session,
		struct ast_stream *caller_stream,
		enum ast_media_type media_type,
		pjmedia_sdp_media *result)
{
	int first_video_pt = 0;
	int new_rtx_pt = 0;
	int old_rtx_pt = 0;
	int apt_val = 0;
	int rtx_time = 0;

	if (media_type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	first_video_pt = get_first_video_pt(pool, result);
	if (!first_video_pt) {
		ast_log(LOG_DEBUG, "[session:%p] change_rtx_apt_value_according_callee can't first_video_pt, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] change_rtx_apt_value_according_callee first_video_pt:%d, endpoint:%s\n",
		local_session, first_video_pt, ast_sorcery_object_get_id(local_session->endpoint));

	get_old_rtx_params(pool, local_session, result, &old_rtx_pt, &apt_val, &rtx_time);
	if (!old_rtx_pt) {
		ast_log(LOG_DEBUG, "[session:%p] change_rtx_apt_value_according_callee can't get old rtx PT, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	new_rtx_pt = get_new_rtx_pt(pool, local_session, caller_stream, first_video_pt, result);
	if (!new_rtx_pt) {
		ast_log(LOG_DEBUG, "[session:%p] change_rtx_apt_value_according_callee can't get according rtx PT, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	if (apt_val == first_video_pt) {
		ast_log(LOG_DEBUG, "[session:%p] change_rtx_apt_value_according_callee apt value is aleady the same, endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		return;
	}

	modify_rtx_params_for_pt_change(pool, local_session, result, old_rtx_pt, new_rtx_pt, rtx_time, first_video_pt);

	return;
}

/* added by zgmo for recording indication in sdp */
static int change_audio_record_attr(struct ast_sip_session *session,
	pjmedia_sdp_media *sdp_media, pj_pool_t *pool)
{
	const char *rec_enable;
	pj_str_t stmp;
	pjmedia_sdp_attr *record_attr;
	pjmedia_sdp_attr *record_time_attr;
	char media_type_str[16] = {0};

	if (!session || !sdp_media || !pool) {
		return -1;
	}

	/*BEGIN; Add by weidu , for 290147, heap-buffer-overflow */
	ast_copy_pj_str(media_type_str, &sdp_media->desc.media, sizeof(media_type_str));
	if (!strstr(media_type_str, "audio")) {
		return 0;
	}
	/*END; Add by weidu , for 290147*/
	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p] change_audio_record_attr, support_special_attrs configure is disable, don't need add record attr, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	rec_enable = pbx_builtin_getvar_helper(session->channel, "REC_ENABLE");
	if (!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")) {
		session->auto_record = 1;
	} else {
		session->auto_record = 0;
	}

	record_attr = pjmedia_sdp_attr_find2(sdp_media->attr_count, sdp_media->attr, "record", NULL);
	if (record_attr) {
		/*
		ast_log(LOG_DEBUG, "Found record_attr:%p,name:%.*s, value:%.*s, session->auto_record:%d",
			record_attr, (int)record_attr->name.slen, record_attr->name.ptr, (int)record_attr->value.slen, record_attr->value.ptr,
			session->auto_record);
		*/
		pjmedia_sdp_attr_remove(&sdp_media->attr_count, sdp_media->attr, record_attr);
	}

	record_time_attr = pjmedia_sdp_attr_find2(sdp_media->attr_count, sdp_media->attr, "record-time", NULL);
	if (record_time_attr) {
		pjmedia_sdp_attr_remove(&sdp_media->attr_count, sdp_media->attr, record_time_attr);
		record_time_attr = NULL;
	}

	if (session->auto_record) {
		char start_time[16] = "";
		time_t start_time_value = ast_channel_get_recordstart_utc(session->channel);
		snprintf(start_time, sizeof(start_time)-1, "%ld", start_time_value ?: ast_now_utc_timestamp());

		record_time_attr = pjmedia_sdp_attr_create(pool, "record-time", pj_cstr(&stmp, start_time));
		pjmedia_sdp_attr_add(&sdp_media->attr_count, sdp_media->attr, record_time_attr);

		record_attr = pjmedia_sdp_attr_create(pool, "record", pj_cstr(&stmp, "on"));
	} else {
		record_attr = pjmedia_sdp_attr_create(pool, "record", pj_cstr(&stmp, "off"));
	}

	pjmedia_sdp_attr_add(&sdp_media->attr_count, sdp_media->attr, record_attr);

	return 0;
}
/* end */

static void disable_gswave_media(pj_pool_t *pool, const struct ast_sip_session *session,
	pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *a = NULL;
	pj_str_t stmp;

	if (!pool || !media) {
		ast_log(LOG_WARNING, "Update media direction, param is null\n");
		return;
	}

	a = pjmedia_sdp_attr_create(pool, "inactive", NULL);

	if (a) {
		ast_remove_all_media_directions(media);
		pjmedia_sdp_media_add_attr(media, a);

		ast_log(LOG_DEBUG, "[session:%p] Peer's media port is zero, but local is gswave, need to use inactive to disable media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	if (session->is_sfu && ast_sip_session_get_only_support_audio()) {
		a = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "version limit"));
		media->attr[media->attr_count++] = a;
	}

	return;
}

static void add_decline_reason_according_callee(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, struct ast_stream *caller_stream, struct ast_stream *callee_stream)
{
	enum ast_video_decline_reason callee_decline_reason = 0;

	if (ast_stream_get_type(caller_stream) != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	callee_decline_reason = ast_stream_get_video_decline_reason(callee_stream);

	if (!callee_decline_reason) {
		ast_log(LOG_DEBUG, "[session:%p] add_decline_reason_according_callee has no reason, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (pjmedia_sdp_attr_find2(media->attr_count, media->attr, "declined-reason", NULL)) {
		ast_log(LOG_DEBUG, "[session:%p] add_decline_reason_according_callee already has reason, don't need add, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_decline_reason_according_callee add decline reason, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_stream_set_video_decline_reason(caller_stream, callee_decline_reason);

	ast_sip_session_add_decline_video_reason(pool, session, media, caller_stream);

	/* clear callee decline reason */
	ast_stream_set_video_decline_reason(callee_stream, 0);

	return;
}

static int match_callee_answer(pj_pool_t *pool,
					pjmedia_sdp_media *caller_local,
					const pjmedia_sdp_media *callee_answer,
					struct ast_sip_session *local_session,
				  	const struct ast_sip_session *peer_session,
				  	struct ast_stream *caller_stream,
				  	struct ast_stream *callee_stream,
				  	enum ast_media_type media_type,
					pjmedia_sdp_media **p_result)
{
	unsigned pt_answer_count = 0;
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT];
	pjmedia_sdp_media *result = NULL;

	if (0 == callee_answer->desc.port) {
		if (local_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC &&
			peer_session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) {
			*p_result = pjmedia_sdp_media_clone(pool, caller_local);
			disable_gswave_media(pool, local_session, *p_result);
			add_decline_reason_according_callee(pool, local_session, *p_result, caller_stream, callee_stream);
		} else {
			*p_result = pjmedia_sdp_media_clone_deactivate(pool, caller_local);
			ast_sip_session_remove_attrs_for_disable_media(*p_result);
			add_decline_reason_according_callee(pool, local_session, *p_result, caller_stream, callee_stream);
		}
		return 0;
	}

	if (peer_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC &&
		local_session->terminal_type != AST_TERMINAL_TYPE_WEBRTC &&
		ast_stream_get_state(callee_stream) == AST_STREAM_STATE_INACTIVE &&
		(media_type != AST_MEDIA_TYPE_AUDIO)) {
		*p_result = pjmedia_sdp_media_clone_deactivate(pool, caller_local);
		ast_sip_session_remove_attrs_for_disable_media(*p_result);
		add_decline_reason_according_callee(pool, local_session, *p_result, caller_stream, callee_stream);
		return 0;
	}

	if (peer_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC &&
		local_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC &&
		(media_type == AST_MEDIA_TYPE_VIDEO) &&
		ast_stream_get_state(callee_stream) == AST_STREAM_STATE_INACTIVE) {
		*p_result = pjmedia_sdp_media_clone(pool, caller_local);
		disable_gswave_media(pool, local_session, *p_result);
		add_decline_reason_according_callee(pool, local_session, *p_result, caller_stream, callee_stream);
		return 0;
	}

	if (0 == caller_local->desc.port) {
		*p_result = pjmedia_sdp_media_clone(pool, caller_local);
		return 0;
	}

	if (ast_sip_session_get_local_negotiate_prefer() && media_type == AST_MEDIA_TYPE_AUDIO) {
		ast_log(LOG_DEBUG, "[session:%p] Create outgoing, enable local negotiate prefer, use local codec prefer for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		*p_result = pjmedia_sdp_media_clone(pool, caller_local);
		return 0;
	}

	/* We find caller local SDP's payload which can find in callee answer SDP */
	get_matched_local_payload(pool, local_session, caller_local, callee_answer, pt_local, &pt_answer_count);

	/* append the payload which can't find in callee answer, audio codec can transcode */
	add_audio_payload_for_can_transcode(local_session, media_type, caller_local, pt_local, &pt_answer_count);

	/* append the payload which can't find in callee answer, but can negotiate independently, like GS-FEC/RTX codec */
	add_payload_for_negotiate_independent(local_session, caller_local, pt_local, &pt_answer_count);

	/* rearrange the payload to suit the offer. */
	result = pjmedia_sdp_media_clone(pool, caller_local);
	change_caller_payload_order_according_callee(local_session, caller_stream, result, pt_local, pt_answer_count);

	/* For video, we don't encode/decode, just use callee answer's fmtp params, and other video params.
	 * And remove the local SDP which payload can't find in callee answer except codec can negotiate independently, like GS-FEC/RTX codec.
	 */
	change_video_params_according_callee(pool, callee_answer, local_session, peer_session,
		caller_stream, callee_stream, media_type, result, pt_local, pt_answer_count);

	/* If callee video first codec change, need to change RTX apt value */
	change_rtx_apt_value_according_callee(pool, local_session, caller_stream, media_type, result);

	arrange_pjmedia_sdp_attr_by_m_line(result);

	change_audio_record_attr(local_session, result, pool);

	*p_result = result;

	return 0;
}

static void modify_local_sdp_media_direction(pj_pool_t *pool,
				  pjmedia_sdp_session *caller_local,
				  const pjmedia_sdp_session *callee_remote,
				  const struct ast_sip_session *local_session,
				  const struct ast_sip_session *peer_session)
{
	unsigned i, j;
	pjmedia_sdp_media *im = NULL;
	pjmedia_sdp_media *om = NULL;
	char media_used[PJMEDIA_MAX_SDP_MEDIA];
	pj_bzero(media_used, sizeof(media_used));

	if (!pool || !caller_local || !callee_remote || !local_session || !peer_session) {
		ast_log(LOG_WARNING, "Modify media direction, param is null\n");
		return;
	}

	for (im = NULL, j = 0; j < caller_local->media_count; ++j) {
		im = caller_local->media[j];
		for (i = 0; i < callee_remote->media_count; ++i) {
			om = callee_remote->media[i];
			if (media_used[i] == 0 && im && om &&
				pj_strcmp(&om->desc.media, &im->desc.media) == 0 &&
				judge_if_transport_compatible(local_session, im) &&
				judge_if_transport_compatible(peer_session, om)) {
				if (strncmp(om->desc.media.ptr, "audio", 5) == 0 &&
					ast_sip_session_get_allow_forward_hold_request()) {
					ast_update_media_direction(pool, local_session, im, om);
					media_used[i] = 1;
					break;
				} else if (strncmp(om->desc.media.ptr, "video", 5) == 0) {
					if (FALSE == ast_sip_session_judge_if_slides_video(im) &&
						FALSE == ast_sip_session_judge_if_slides_video(om)) {
						ast_update_media_direction(pool, local_session, im, om);
						media_used[i] = 1;
						break;
					}
					if (TRUE == ast_sip_session_judge_if_slides_video(im) &&
						TRUE == ast_sip_session_judge_if_slides_video(om)) {
						ast_update_media_direction(pool, local_session, im, om);
						media_used[i] = 1;
						break;
					}
				} else if (strncmp(om->desc.media.ptr, "application", 11) == 0) {
					if (ast_sip_session_judge_if_fec_stream(om) &&
						ast_sip_session_judge_if_fec_stream(im)) {
						ast_update_media_direction(pool, local_session, im, om);
						media_used[i] = 1;
						break;
					}
					if (ast_sip_session_judge_if_fecc_stream(om) &&
						ast_sip_session_judge_if_fecc_stream(im)) {
						ast_update_media_direction(pool, local_session, im, om);
						media_used[i] = 1;
						break;
					}
				} else {
					ast_log(LOG_WARNING, "Can't process media type except audio/video/application\n");
					break;
				}
			}
		}
	}

	return;
}
/* End add by bfdu for Bug 43278, 2015.05.20 */

/* For FEC we can negotiated independent for caller and callee.
 * For gswave and GVC call, gswave don't have BFCP, we need to judge if have slides video to add BFCP.
 */
static unsigned int judge_if_need_clone_for_special_media(struct ast_stream *caller_stream,
	const struct ast_sip_session *peer_session, struct ast_sip_session *local_session)
{
	if (!caller_stream || !peer_session || !local_session) {
		return 0;
	}

	if (local_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC &&
		peer_session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) {
		return 1;
	}

	if (ast_stream_get_sub_type(caller_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_FEC) {
		return 1;
	}

	if ((ast_stream_get_sub_type(caller_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		ast_stream_get_sub_type(caller_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) &&
		!ast_sip_session_support_bfcp(peer_session))  {
		return 1;
	}

	return 0;
}

static struct ast_stream_topology *get_local_topology(const struct ast_sip_session *local_session,
	const struct ast_sip_session *peer_session)
{
	struct ast_stream_topology *local_topology = NULL;

	if (!peer_session->active_media_state || !peer_session->active_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] Peer endpoint:%s has no active media, don't need modify local sdp for endpoint:%s\n",
			local_session, ast_sorcery_object_get_id(peer_session->endpoint), ast_sorcery_object_get_id(local_session->endpoint));
		return NULL;
	}

	if (!local_session->pending_media_state || !local_session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no pending media, use active media\n",
			local_session, ast_sorcery_object_get_id(local_session->endpoint));
		if (local_session->active_media_state && local_session->active_media_state->topology) {
			local_topology = local_session->active_media_state->topology;
		} else{
			ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no active media too, don't modify local sdp\n",
				local_session, ast_sorcery_object_get_id(local_session->endpoint));
			return NULL;
		}
	} else {
		local_topology = local_session->pending_media_state->topology;
	}

	return local_topology;
}

static pjmedia_sdp_session* modify_local_sdp_with_callee_answer(pj_pool_t *pool,
				  pjmedia_sdp_session *caller_local,
				  const pjmedia_sdp_session *callee_answer,
				  struct ast_sip_session *local_session,
				  const struct ast_sip_session *peer_session)
{
	unsigned i = 0;
	unsigned j = 0;
	unsigned found = 0;
	char media_used[PJMEDIA_MAX_SDP_MEDIA] = {0};
	const pjmedia_sdp_media *callee_media;
	pjmedia_sdp_media *caller_media = NULL;
	pjmedia_sdp_media *result_media = NULL;
	pjmedia_sdp_session *result_session = NULL;
	struct ast_stream *caller_stream = NULL;
	struct ast_stream *callee_stream = NULL;
	struct ast_stream_topology *local_topology = NULL;
	unsigned int caller_max_media_count = 0;
	unsigned int callee_max_media_count = 0;

	char media_type_str[16] = {0};
	enum ast_media_type media_type = AST_MEDIA_TYPE_UNKNOWN;

	if (!pool || !caller_local || !callee_answer || !local_session || !peer_session) {
		ast_log(LOG_WARNING, "modify_local_sdp_with_callee_answer param is NULL\n");
		return NULL;
	}

	local_topology = get_local_topology(local_session, peer_session);
	if (!local_topology) {
		return caller_local;
	}

	result_session = pjmedia_sdp_session_clone(pool, caller_local);
	result_session->media_count = 0;
	memset(media_used, 0, sizeof(media_used));

	if (caller_local->media_count > ast_stream_topology_get_count(local_topology)) {
		caller_max_media_count = ast_stream_topology_get_count(local_topology);
	} else {
		caller_max_media_count = caller_local->media_count;
	}

	if (callee_answer->media_count > ast_stream_topology_get_count(peer_session->active_media_state->topology)) {
		callee_max_media_count = ast_stream_topology_get_count(peer_session->active_media_state->topology);
	} else {
		callee_max_media_count = callee_answer->media_count;
	}

	for (caller_media = NULL, j = 0; j < caller_max_media_count; ++j) {
		caller_media = caller_local->media[j];
		if (!caller_media) {
			continue;
		}
		caller_stream = ast_stream_topology_get_stream(local_topology, j);
		if (!caller_stream) {
			continue;
		}
		found = 0;
		for (i = 0; i < callee_max_media_count; ++i) {
			callee_media = callee_answer->media[i];
			if (!callee_media) {
				continue;
			}
			callee_stream = ast_stream_topology_get_stream(peer_session->active_media_state->topology, i);
			if (!callee_stream) {
				continue;
			}
			if (pj_strcmp(&callee_media->desc.media, &caller_media->desc.media) == 0 &&
				judge_if_transport_compatible(local_session, caller_media) &&
				judge_if_transport_compatible(peer_session, callee_media) &&
				(ast_stream_get_sub_type(caller_stream) == ast_stream_get_sub_type(callee_stream)) &&
				/* Modify by bfdu, we can only support audio has two m line, RTP and SRTP media, 2018.10.29 */
				(media_used[i] == 0 || !strncmp(callee_media->desc.media.ptr, "audio", 5))) {
				pj_status_t status = -1;
				ast_copy_pj_str(media_type_str, &callee_media->desc.media, sizeof(media_type_str));
				media_type = ast_media_type_from_str(media_type_str);
				switch (media_type) {
					case AST_MEDIA_TYPE_AUDIO:
					case AST_MEDIA_TYPE_VIDEO:
						status = match_callee_answer(pool, caller_media, callee_media, local_session, peer_session, caller_stream, callee_stream, media_type, &result_media);
						break;
					case AST_MEDIA_TYPE_APPLICATION:
						status = match_callee_answer_app_media(pool, caller_media, callee_media, local_session, peer_session, caller_stream, callee_stream, &result_media);
						break;
					default:
						break;
				}

				/* End modify by bfdu for BFCP, 2015.07.13 */
				if (status == 0) {
					media_used[i] = 1;
					found = 1;
					break;
				}
			}
		}

		if (found) {
			result_session->media[result_session->media_count++] = result_media;
			if (ast_sip_session_judge_if_disable_media(result_media)) {
				ast_sip_session_remove_media_from_bundle_group(local_session, pool, result_session, result_media);
			}
		} else {
			if (judge_if_need_clone_for_special_media(caller_stream, peer_session, local_session)) {
				if (local_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
					disable_gswave_media(pool, local_session, caller_media);
				}
				result_session->media[result_session->media_count++] = pjmedia_sdp_media_clone(pool, caller_media);
			} else {
				result_media = pjmedia_sdp_media_clone_deactivate(pool, caller_media);
				result_session->media[result_session->media_count++] = result_media;
				ast_sip_session_remove_attrs_for_disable_media(result_media);
				ast_sip_session_remove_media_from_bundle_group(local_session, pool, result_session, result_media);
			}
		}
	}

	return result_session;
}

static int update_local_sdp_for_passthrough(pj_pool_t *pool,
						pjmedia_sdp_media *local,
						const pjmedia_sdp_media *callee_answer,
						const struct ast_sip_session *local_session,
						const struct ast_sip_session *peer_session,
						pjmedia_sdp_media **p_result)
{
	unsigned int i = 0;
	pjmedia_sdp_media *tmp_local = NULL;

	tmp_local = pjmedia_sdp_media_clone(pool, local);
	tmp_local->attr_count = 0;

	for (i = 0; i < local->attr_count; i++) {
		if (!ast_sip_session_judge_if_passthrough_attr(local->attr[i]->name.ptr)) {
			tmp_local->attr[tmp_local->attr_count++] = local->attr[i];
		}
	}

	for (i = 0; i < callee_answer->attr_count; i++) {
		if (ast_sip_session_judge_if_passthrough_attr(callee_answer->attr[i]->name.ptr)) {
			tmp_local->attr[tmp_local->attr_count++] = callee_answer->attr[i];
		}
	}

	if (local_session->endpoint->media.support_special_attrs) {
		tmp_local->bandw_count = 0;
		for (i = 0; i < callee_answer->bandw_count; i++) {
			tmp_local->bandw[tmp_local->bandw_count++] = callee_answer->bandw[i];
		}
	}

	*p_result = pjmedia_sdp_media_clone(pool, tmp_local);

	return 0;
}

static int modify_application_rtp_local_sdp_for_passthrough(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	int res = 0;

	if (ast_sip_session_judge_if_fec_stream(local) &&
		ast_sip_session_judge_if_fec_stream(callee_answer)) {
		res = update_local_sdp_for_passthrough(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
    } else if (ast_sip_session_judge_if_fecc_stream(local) &&
		ast_sip_session_judge_if_fecc_stream(callee_answer)) {
		res = update_local_sdp_for_passthrough(pool, local, callee_answer, local_session, peer_session, p_result);
		return res;
    } else {
		ast_log(LOG_DEBUG, "Don't have rtpmap, just deactivate media\n");
		return -1;
    }

	return -1;
}


static int modify_video_local_sdp_for_passthrough(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	int res = 0;

	res = update_local_sdp_for_passthrough(pool, local, callee_answer, local_session, peer_session, p_result);

	return res;
}

static int modify_application_local_sdp_for_passthrough(pj_pool_t *pool,
					pjmedia_sdp_media *local,
					const pjmedia_sdp_media *callee_answer,
					const struct ast_sip_session *local_session,
					const struct ast_sip_session *peer_session,
					pjmedia_sdp_media **p_result)
{
	int status = -1;
	enum ast_sip_application_type app_type = ast_sip_session_get_application_type(callee_answer->desc.transport);
	enum ast_sip_application_type i_app_type = ast_sip_session_get_application_type(local->desc.transport);
	if (app_type == i_app_type) {
		switch(app_type) {
			case AST_SIP_APP_UDP_BFCP:
			case AST_SIP_APP_TCP_BFCP:
				status = update_local_sdp_for_passthrough(pool, local, callee_answer, local_session, peer_session, p_result);
				break;
			case AST_SIP_APP_RTP:
				status = modify_application_rtp_local_sdp_for_passthrough(pool, local, callee_answer, local_session, peer_session, p_result);
				break;
			default:
				ast_log(LOG_DEBUG, "Can't match other application media, except bfcp/fec/fecc\n");
		}
	}

	return status;
}

/*added by jklou for fec/fecc/bfcp attribure passthrough*/
static pjmedia_sdp_session* modify_local_sdp_for_passthrough( pj_pool_t *pool,
				  pjmedia_sdp_session *caller_local,
				  const pjmedia_sdp_session *callee_answer,
				  const struct ast_sip_session *local_session,
				  const struct ast_sip_session *peer_session
				  )
{
	unsigned i = 0;
	unsigned j = 0;
	unsigned found = 0;
	pjmedia_sdp_media *im = NULL;
	pjmedia_sdp_session *result = NULL;
	const pjmedia_sdp_media *om;
	pjmedia_sdp_media *am = NULL;
	char media_used[PJMEDIA_MAX_SDP_MEDIA] = {0};
	char media_type_str[16] = {0};
	enum ast_media_type media_type = AST_MEDIA_TYPE_UNKNOWN;
	struct ast_stream *caller_stream = NULL;
	struct ast_stream *callee_stream = NULL;
	struct ast_stream_topology *local_topology = NULL;
	unsigned int caller_max_media_count = 0;
	unsigned int callee_max_media_count = 0;

	if (!caller_local) {
		return NULL;
	}

	if (!ast_sip_session_judge_if_need_passthrough()) {
		return caller_local;
	}

	local_topology = get_local_topology(local_session, peer_session);
	if (!local_topology) {
		return caller_local;
	}

	result = pjmedia_sdp_session_clone(pool, caller_local);
	result->media_count = 0;

	if (caller_local->media_count > ast_stream_topology_get_count(local_topology)) {
		caller_max_media_count = ast_stream_topology_get_count(local_topology);
	} else {
		caller_max_media_count = caller_local->media_count;
	}

	if (callee_answer->media_count > ast_stream_topology_get_count(peer_session->active_media_state->topology)) {
		callee_max_media_count = ast_stream_topology_get_count(peer_session->active_media_state->topology);
	} else {
		callee_max_media_count = callee_answer->media_count;
	}

	pj_bzero(media_used, sizeof(media_used));
	for (im = NULL, j = 0; j < caller_max_media_count; ++j) {
		im = caller_local->media[j];
		if (!im) {
			continue;
		}
		found = 0;

		caller_stream = ast_stream_topology_get_stream(local_topology, j);
		if (!caller_stream) {
			continue;
		}

		for (i = 0; i < callee_max_media_count; ++i) {
			om = callee_answer->media[i];
			if (!om) {
				continue;
			}
			callee_stream = ast_stream_topology_get_stream(peer_session->active_media_state->topology, i);
			if (!callee_stream) {
				continue;
			}
			if (pj_strcmp(&om->desc.media, &im->desc.media) == 0 &&
				judge_if_transport_compatible(local_session, im) &&
				judge_if_transport_compatible(peer_session, om) &&
				(ast_stream_get_sub_type(caller_stream) == ast_stream_get_sub_type(callee_stream)) &&
				media_used[i] == 0) {
				int status = -1;

				ast_copy_pj_str(media_type_str, &om->desc.media, sizeof(media_type_str));
				media_type = ast_media_type_from_str(media_type_str);
				switch (media_type) {
					case AST_MEDIA_TYPE_VIDEO:
						status = modify_video_local_sdp_for_passthrough(pool, im, om, local_session, peer_session, &am);
						break;
					case AST_MEDIA_TYPE_APPLICATION:
						status = modify_application_local_sdp_for_passthrough(pool, im, om, local_session, peer_session, &am);
						break;
					default:
						break;
				}

				if (status == 0){
					found = 1;
					break;
				}
			}
		}
		if (found) {
			result->media[result->media_count++] = am;
		} else {
			result->media[result->media_count++] = im;
		}
	}

	return result;
}
/*ended by jklou*/

static struct ast_sip_session *get_sip_session_from_channel(struct ast_channel *chan)
{
	struct ast_sip_session *session = NULL;
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		ast_log(LOG_DEBUG, "get_sip_session_from_channel, chan is invalid\n");
		return NULL;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "get_sip_session_from_channel, channel:%s type is PJSIP\n", ast_channel_name(chan));
		return NULL;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt) {
		ast_log(LOG_WARNING, "get_sip_session_from_channel, chan_pvt is invalid\n");
		return NULL;
	}

	session = chan_pvt->session;
	if (!session || !session->inv_session || !session->inv_session->neg || !session->inv_session->pool_prov) {
		ast_log(LOG_WARNING, "get_sip_session_from_channel, session or inv_session is invalid\n");
		return NULL;
	}

	return session;
}

static pjmedia_sdp_session *get_local_sdp_from_session(pj_pool_t *pool, struct ast_sip_session *session)
{
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;

	chan_neg_state = session->inv_session->neg->state;
	switch (chan_neg_state) {
		case PJMEDIA_SDP_NEG_STATE_WAIT_NEGO:
			caller_local_sdp = session->inv_session->neg->neg_local_sdp;
			break;
		case PJMEDIA_SDP_NEG_STATE_DONE:
			if (session->endpoint->media.rtp.accept_multiple_sdp_answers) {
				session->inv_session->updated_sdp_answer = 1;
				session->inv_session->neg->neg_local_sdp = pjmedia_sdp_session_clone(pool, session->inv_session->neg->active_local_sdp);
				caller_local_sdp = session->inv_session->neg->neg_local_sdp;
				pjmedia_sdp_neg_set_remote_offer(pool, session->inv_session->neg, session->inv_session->neg->active_remote_sdp);
				break;
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Can't get local sdp for done state, endpoint:%s not enable accept_multiple_sdp_answers\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				return NULL;
			}
		default:
			ast_log(LOG_ERROR, "[session:%p] Can't get local sdp for endpoint:%s, neg state(%d) is not WAIT or DONE\n",
				session, ast_sorcery_object_get_id(session->endpoint), chan_neg_state);
			return NULL;
	}

	return caller_local_sdp;
}

/* Add by bfdu for disable video media when use voicemail, 2017.04.26 */
static void disable_media_except_audio(struct ast_channel *chan)
{
	unsigned i = 0;
	struct ast_sip_session *session = NULL;
	pjmedia_sdp_media *local_media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_media *disable_res = NULL;
	pj_pool_t *pool = NULL;

	session = get_sip_session_from_channel(chan);
	if (!session) {
		ast_log(LOG_WARNING, "Disable media chan_pvt session is invalid\n");
		return;
	}

	pool = session->inv_session->pool_prov;

	ast_log(LOG_DEBUG, "[session:%p] disable_media_except_audio for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	caller_local_sdp= get_local_sdp_from_session(pool, session);
	if (!caller_local_sdp) {
		ast_log(LOG_WARNING, "[session:%p] Disable media neg_local_sdp is invalid for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (local_media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		local_media = caller_local_sdp->media[i];
		disable_res = NULL;
		if (!local_media) {
			ast_log(LOG_WARNING, "[session:%p] Disable media local_media is invalid for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		if (strncmp(local_media->desc.media.ptr, "video", 5) == 0) {
			ast_log(LOG_DEBUG, "[session:%p] Disable video media for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ) {
				disable_res = pjmedia_sdp_media_clone(pool, local_media);
				disable_gswave_media(pool, session, disable_res);
			} else {
				disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);
				ast_sip_session_remove_attrs_for_disable_media(disable_res);
				ast_sip_session_remove_media_from_bundle_group(session, pool, caller_local_sdp, disable_res);
			}
			caller_local_sdp->media[i] = disable_res;
		} else if (strncmp(local_media->desc.media.ptr, "application", 11) == 0) {
			ast_log(LOG_DEBUG, "[session:%p] Disable application media for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);
			caller_local_sdp->media[i] = disable_res;
		}
	}

	return;
}

static void disable_main_video(struct ast_channel *chan,int ignore_video_caps)
{
	unsigned i = 0;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;
	pjmedia_sdp_media *local_media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_media *disable_res = NULL;
	pj_pool_t *pool = NULL;

	if (!chan) {
		ast_log(LOG_DEBUG, "disable media chan_pvt chan is invalid\n");
		return;
	}

	if (!ignore_video_caps){

		const char *has_no_video_caps = pbx_builtin_getvar_helper(chan, "HAS_NO_VIDEO_CAPS");
		if (ast_strlen_zero(has_no_video_caps)) {
			ast_log(LOG_DEBUG, "disable_main_video, HAS_NO_VIDEO_CAPS not set, don't need process\n");
			return;
		}
		if (strcasecmp(has_no_video_caps, "yes")) {
			ast_log(LOG_DEBUG, "disable_main_video, HAS_NO_VIDEO_CAPS is not yes, don't need process\n");
			return;
		}
	}

	ast_log(LOG_DEBUG, "disable_main_video begin\n");

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "Disable main video just suppot channel type is PJSIP\n");
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		ast_log(LOG_WARNING, "disable main video chan_pvt session is invalid\n");
		return;
	}
	chan_neg_state = chan_pvt->session->inv_session->neg->state;
	if (chan_neg_state != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_DEBUG, "Can't disable main video, neg state(%d) is not WAIT_NEGO", chan_neg_state);
		return;
	}

	pool = chan_pvt->session->inv_session->pool_prov;
	caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;

	if (!pool || !caller_local_sdp) {
		ast_log(LOG_WARNING, "disable main video pool or neg_local_sdp is invalid\n");
		return;
	}

	for (local_media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		local_media = caller_local_sdp->media[i];
		disable_res = NULL;
		if (!local_media) {
			ast_log(LOG_WARNING, "disable main video local_media is invalid\n");
			return;
		}

		if (strncmp(local_media->desc.media.ptr, "video", 5) == 0 &&
			!ast_sip_session_judge_if_slides_video(local_media)) {
			ast_log(LOG_DEBUG, "disable video media\n");

			if (chan_pvt->session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ) {
				disable_res = pjmedia_sdp_media_clone(pool, local_media);
				disable_gswave_media(pool, chan_pvt->session, disable_res);
			} else {
				disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);
				ast_sip_session_remove_attrs_for_disable_media(disable_res);
				ast_sip_session_remove_media_from_bundle_group(chan_pvt->session, pool, caller_local_sdp, disable_res);
			}
			caller_local_sdp->media[i] = disable_res;
			break;
		}
	}

	return;
}
/*Start added by zyxia for video paging,2022.8.5*/

#if 0
/*Start  added by zyxia for Bug 261644,2023.6.7*/
static void disable_slides_video(struct ast_channel *chan,int ignore_video_caps)
{
	unsigned i = 0;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;
	pjmedia_sdp_media *local_media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_media *disable_res = NULL;
	pj_pool_t *pool = NULL;

	if (!chan) {
		ast_log(LOG_DEBUG, "disable media chan_pvt chan is invalid\n");
		return;
	}

	if (!ignore_video_caps){

		const char *has_no_video_caps = pbx_builtin_getvar_helper(chan, "HAS_NO_VIDEO_CAPS");
		if (ast_strlen_zero(has_no_video_caps)) {
			ast_log(LOG_DEBUG, "disable_slides_video, HAS_NO_VIDEO_CAPS not set, don't need process\n");
			return;
		}
		if (strcasecmp(has_no_video_caps, "yes")) {
			ast_log(LOG_DEBUG, "disable_slides_video, HAS_NO_VIDEO_CAPS is not yes, don't need process\n");
			return;
		}
	}

	ast_log(LOG_DEBUG, "disable_slides_video begin\n");

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "Disable slides video just suppot channel type is PJSIP\n");
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		ast_log(LOG_WARNING, "disable slides video chan_pvt session is invalid\n");
		return;
	}
	chan_neg_state = chan_pvt->session->inv_session->neg->state;
	if (chan_neg_state != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_DEBUG, "Can't disable slides video, neg state(%d) is not WAIT_NEGO", chan_neg_state);
		return;
	}

	pool = chan_pvt->session->inv_session->pool_prov;
	caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;

	if (!pool || !caller_local_sdp) {
		ast_log(LOG_WARNING, "disable slides video pool or neg_local_sdp is invalid\n");
		return;
	}

	for (local_media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		local_media = caller_local_sdp->media[i];
		disable_res = NULL;
		if (!local_media) {
			ast_log(LOG_WARNING, "disable slides video local_media is invalid\n");
			return;
		}

		if (strncmp(local_media->desc.media.ptr, "video", 5) == 0 &&
			ast_sip_session_judge_if_slides_video(local_media)) {
			ast_log(LOG_DEBUG, "disable slides video media\n");

			if (chan_pvt->session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ) {
				disable_res = pjmedia_sdp_media_clone(pool, local_media);
				disable_gswave_media(pool, chan_pvt->session, disable_res);
			} else {
				disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);
				ast_sip_session_remove_attrs_for_disable_media(disable_res);
				ast_sip_session_remove_media_from_bundle_group(chan_pvt->session, pool, caller_local_sdp, disable_res);
			}
			caller_local_sdp->media[i] = disable_res;
			break;
		}
	}

	return;
}
/*Start  added by zyxia for Bug 261644,2023.6.7*/
#endif
static void change_video_media_direction(struct ast_channel *chan, enum ast_media_sub_type sub_type,char* dir)
{
	unsigned i = 0;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;
	pjmedia_sdp_media *local_media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pj_pool_t *pool = NULL;

	if (!chan) {
		ast_log(LOG_DEBUG, "change_video_media_direction media chan_pvt chan is invalid\n");
		return;
	}

	ast_log(LOG_DEBUG, "change_video_media_direction begin\n");

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "change_video_media_direction just suppot channel type is PJSIP\n");
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		ast_log(LOG_WARNING, "change_video_media_direction chan_pvt session is invalid\n");
		return;
	}
	chan_neg_state = chan_pvt->session->inv_session->neg->state;
	if (chan_neg_state != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_DEBUG, "Can't change_video_media_direction_for_page, neg state(%d) is not WAIT_NEGO", chan_neg_state);
		return;
	}

	pool = chan_pvt->session->inv_session->pool_prov;
	caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;

	if (!pool || !caller_local_sdp) {
		ast_log(LOG_WARNING, "change_video_media_direction pool or neg_local_sdp is invalid\n");
		return;
	}

	for (local_media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		local_media = caller_local_sdp->media[i];
		if (!local_media) {
			ast_log(LOG_WARNING, "change_video_media_direction local_media is invalid\n");
			return;
		}

		if (strncmp(local_media->desc.media.ptr, "video", 5) == 0) {

			if(sub_type == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES && ast_sip_session_judge_if_slides_video(local_media)){
				change_media_direction(pool, local_media, dir);
				ast_log(LOG_DEBUG, "change_video_direction to %s,type:slides\n",dir);
				break;
			}
			else if(sub_type == AST_MEDIA_SUB_VIDEO_TYPE_MAIN && ast_sip_session_judge_if_main_video(local_media)){
				change_media_direction(pool, local_media, dir);
				ast_log(LOG_DEBUG, "change_video_direction to %s,type:main\n",dir);
				break;
			}
		}
	}

	return;
}

static void change_session_media_direction(struct ast_sip_session *session,enum ast_media_sub_type sub_type ,enum ast_stream_media_direction media_direction)
{
	unsigned int i = 0;
	unsigned int find = 0;
	struct ast_stream *stream = NULL;


	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) != sub_type) {
			continue;
		}

		find = 1;
		ast_stream_set_local_media_direction(stream, media_direction);
		switch (media_direction) {
			case SENDRECV:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
				break;
			case RECVONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_RECVONLY);
				break;
			case SENDONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDONLY);
				break;
			case INACTIVE:
				ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
				break;
			default:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
				break;
			}

		break;
	}
	if(!find)
		create_new_main_video_stream_append(session,session->pending_media_state->topology,media_direction);

}
/*End added by zyxia for video paging,2022.8.5*/

static void disable_multiplexed_fec(pjmedia_sdp_media *local_media)
{
	unsigned int i = 0;
	unsigned int j = 0;
	pjmedia_sdp_attr *fec_rtpmap_attr = NULL;
	pjmedia_sdp_attr *fec_fmtp_attr = NULL;
	pjmedia_sdp_rtpmap fec_rtpmap;

	for (i = 0; i < local_media->desc.fmt_count; ++i) {
		/* Get the rtpmap for the payload type. */
		fec_rtpmap_attr = pjmedia_sdp_media_find_attr2(local_media, "rtpmap", &local_media->desc.fmt[i]);
		if (!fec_rtpmap_attr) {
			pj_assert(!"Bug! Offer should have been validated");
			return;
		}
		pjmedia_sdp_attr_get_rtpmap(fec_rtpmap_attr, &fec_rtpmap);

		if (!pj_stricmp2(&fec_rtpmap.enc_name, "GS-FEC")) {
			ast_log(LOG_DEBUG, "remove GS-FEC rtpmap attr\n");
			pjmedia_sdp_media_remove_attr(local_media, fec_rtpmap_attr);
			fec_fmtp_attr = pjmedia_sdp_media_find_attr2(local_media, "fmtp", &local_media->desc.fmt[i]);
			if (fec_fmtp_attr) {
				ast_log(LOG_DEBUG, "remove GS-FEC fmtp attr\n");
				pjmedia_sdp_media_remove_attr(local_media, fec_fmtp_attr);
			}
			fec_fmtp_attr = pjmedia_sdp_media_find_attr2(local_media, "gs-fec-version", NULL);
			if (fec_fmtp_attr) {
				ast_log(LOG_DEBUG, "remove gs-fec-version fmtp attr\n");
				pjmedia_sdp_media_remove_attr(local_media, fec_fmtp_attr);
			}

			for (j = i; j < local_media->desc.fmt_count - 1; j++) {
				local_media->desc.fmt[j] = local_media->desc.fmt[j + 1];
			}
			local_media->desc.fmt_count = local_media->desc.fmt_count - 1;
			break;
		}
	}

	return;
}

/*Start  added by zyxia for Bug 227857 - 2022.6.28*/
static void disable_application_bfcp(struct ast_channel *chan)
{
	unsigned i = 0;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;
	pjmedia_sdp_media *local_media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_media *disable_res = NULL;
	pj_pool_t *pool = NULL;

	if (!chan) {
		ast_log(LOG_DEBUG, "disable media chan_pvt chan is invalid\n");
		return;
	}


	ast_log(LOG_DEBUG, "disable_application begin\n");

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "Disable application just suppot channel type is PJSIP\n");
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		ast_log(LOG_WARNING, "disable application  chan_pvt session is invalid\n");
		return;
	}
	chan_neg_state = chan_pvt->session->inv_session->neg->state;
	if (chan_neg_state != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_DEBUG, "Can't disable application , neg state(%d) is not WAIT_NEGO", chan_neg_state);
		return;
	}

	pool = chan_pvt->session->inv_session->pool_prov;
	caller_local_sdp = chan_pvt->session->inv_session->neg->neg_local_sdp;

	if (!pool || !caller_local_sdp) {
		ast_log(LOG_WARNING, "disable main video pool or neg_local_sdp is invalid\n");
		return;
	}

	for (local_media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		local_media = caller_local_sdp->media[i];
		disable_res = NULL;
		if (!local_media) {
			ast_log(LOG_WARNING, "disable application  is invalid\n");
			return;
		}

		if (strncmp(local_media->desc.media.ptr, "application", 11) == 0) {
			disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);

			caller_local_sdp->media[i] = disable_res;
			break;
		}
	}

	return;
}
/*End  added by zyxia for Bug 227857 - 2022.6.28*/

/* Add by bfdu for we only reserve audio/main video, 2018.10.10 */
static void disable_external_media(struct ast_channel *chan)
{
	unsigned int i = 0;
	struct ast_sip_session *session = NULL;
	pjmedia_sdp_media *local_media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_media *disable_res = NULL;
	pj_pool_t *pool = NULL;

	session = get_sip_session_from_channel(chan);
	if (!session) {
		ast_log(LOG_WARNING, "Disable media chan_pvt session is invalid\n");
		return;
	}

	pool = session->inv_session->pool_prov;

	ast_log(LOG_DEBUG, "[session:%p] disable_external_media for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	caller_local_sdp= get_local_sdp_from_session(pool, session);
	if (!caller_local_sdp) {
		ast_log(LOG_WARNING, "[session:%p] Disable external media neg_local_sdp is invalid for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (local_media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		local_media = caller_local_sdp->media[i];
		disable_res = NULL;
		if (!local_media) {
			ast_log(LOG_WARNING, "[session:%p] Disable media local_media is invalid for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		if (strncmp(local_media->desc.media.ptr, "video", 5) == 0) {
			if (ast_sip_session_judge_if_fec_stream(local_media)) {
				disable_multiplexed_fec(local_media);
			}

			if (ast_sip_session_judge_if_slides_video(local_media)) {
				ast_log(LOG_DEBUG, "[session:%p] Disable slides video media for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ) {
					disable_res = pjmedia_sdp_media_clone(pool, local_media);
					disable_gswave_media(pool, session, disable_res);
				} else {
					disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);
					ast_sip_session_remove_attrs_for_disable_media(disable_res);
					ast_sip_session_remove_media_from_bundle_group(session, pool, caller_local_sdp, disable_res);
				}
				caller_local_sdp->media[i] = disable_res;
			}
		} else if (strncmp(local_media->desc.media.ptr, "application", 11) == 0) {
			ast_log(LOG_DEBUG, "[session:%p] Disable application media for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			disable_res = pjmedia_sdp_media_clone_deactivate(pool, local_media);
			caller_local_sdp->media[i] = disable_res;
		}
	}

	return;
}

static void disable_direct_media(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (NULL == chan){
		ast_log(LOG_WARNING, "disable_direct_media, chan is NULL");
		return;
	}

	ast_log(LOG_DEBUG, "Has no peer, disable_direct_media begin\n");

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		ast_log(LOG_WARNING, "disable_direct_media, chan_pvt or session is NULL");
		return;
	}

	chan_pvt->session->is_direct_media = 0;
	pbx_builtin_setvar_helper(chan, "TRANSPARENTSDP", "no");

	return;
}

static int send_direct_media_reinvite_request(void *data)
{
	struct ast_sip_session *session = data;
	int res;

	res = ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	ao2_ref(session, -1);
	return res;
}

static int check_special_answer_param(struct ast_sip_channel_pvt *chan_pvt, struct ast_sip_channel_pvt *peer_pvt)
{
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		ast_log(LOG_WARNING, "special answer chan_pvt session is invalid\n");
		return -1;
	}

	if (!peer_pvt || !peer_pvt->session || !peer_pvt->session->inv_session || !peer_pvt->session->inv_session->neg) {
		ast_log(LOG_WARNING, "special answer peer_pvt session is invalid\n");
		return -1;
	}

	if (!chan_pvt->session->inv_session->pool_prov) {
		ast_log(LOG_WARNING, "special answer chan_pvt pool is invalid\n");
		return -1;
	}

	if (!peer_pvt->session->inv_session->neg->active_local_sdp ||
		!peer_pvt->session->inv_session->neg->active_remote_sdp) {
		ast_log(LOG_WARNING, "special answer peer_pvt active local/remote sdp is invalid\n");
		return -1;
	}

	if (peer_pvt->session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_WARNING, "special answer peer_pvt inv_session state is DISCONNECTED\n");
		return -1;
	}

	if (peer_pvt->session->inv_session->neg->state != PJMEDIA_SDP_NEG_STATE_DONE) {
		ast_log(LOG_WARNING, "special answer peer_pvt neg state(%d) is not PJMEDIA_SDP_NEG_STATE_DONE\n", peer_pvt->session->inv_session->neg->state);
		return -1;
	}

	return 0;
}

static void chan_pjsip_use_final_response_sdp(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	pj_pool_t *pool = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_session *callee_local_sdp = NULL;
	pjmedia_sdp_session *callee_answer_sdp = NULL;
	char *sdp_body = NULL;

	if (ast_channel_state(session->channel) == AST_STATE_UP) {
		ast_log(LOG_WARNING, "chan_pjsip_use_final_response_sdp channel is already up\n");
		return;
	}

	pool = session->inv_session->pool_prov;
	caller_local_sdp = session->inv_session->neg->initial_sdp;
	callee_local_sdp = peer_session->inv_session->neg->active_local_sdp;
	callee_answer_sdp = peer_session->inv_session->neg->active_remote_sdp;

	if (!caller_local_sdp) {
		ast_log(LOG_WARNING, "chan_pjsip_use_final_response_sdp has no active_local_sdp\n");
		return;
	}

	sdp_body = ast_malloc(MAX_SDP_BODY_LEN);
	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(caller_local_sdp, sdp_body, MAX_SDP_BODY_LEN - 1); // Must use MAX_SDP_BODY_LEN - 1, otherwise there is a potential overflow
	ast_log(LOG_DEBUG, "[caller] caller local sdp is:\n%s", sdp_body);

	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(callee_answer_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
	ast_log(LOG_DEBUG, "[callee] UCM <<<=== B . received 200 ok from the callee\n%s", sdp_body);

	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(callee_local_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
	ast_log(LOG_DEBUG, "[callee] local sdp of the callee is:\n%s", sdp_body);

	ast_log(LOG_DEBUG, "[caller] modify caller local sdp according to callee answer sdp!\n");

	session->inv_session->neg->neg_local_sdp = modify_local_sdp_with_callee_answer(pool, caller_local_sdp, callee_local_sdp, session, peer_session);

	session->inv_session->neg->neg_local_sdp = modify_local_sdp_for_passthrough(pool, session->inv_session->neg->neg_local_sdp, callee_answer_sdp, session, peer_session);

	modify_local_sdp_media_direction(pool, session->inv_session->neg->neg_local_sdp, callee_answer_sdp, session, peer_session);

	if (!session->inv_session->neg->neg_local_sdp) {
		ast_log(LOG_DEBUG, "neg_local_sdp is NULL\n");
		ast_free(sdp_body);
		sdp_body = NULL;
		return;
	}

	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(session->inv_session->neg->neg_local_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
	ast_log(LOG_DEBUG, "[caller] caller local sdp is:\n%s", sdp_body);

	ast_channel_set_use_final_sdp(session->channel, 1);
	ast_channel_set_use_final_sdp(peer_session->channel, 1);

	pjmedia_sdp_neg_set_remote_offer(pool, session->inv_session->neg, session->inv_session->neg->active_remote_sdp);

	ast_free(sdp_body);
	sdp_body = NULL;

	return;
}

static void modify_caller_sdp_for_wait_state(pj_pool_t *pool, struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	const char *transparentsdp = NULL;
	pjmedia_sdp_session *callee_answer_sdp = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pjmedia_sdp_session *callee_local_sdp = NULL;
	char *sdp_body = NULL;

	caller_local_sdp = session->inv_session->neg->neg_local_sdp;
	callee_local_sdp = peer_session->inv_session->neg->active_local_sdp;
	callee_answer_sdp = peer_session->inv_session->neg->active_remote_sdp;

	sdp_body = ast_malloc(MAX_SDP_BODY_LEN);
	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(caller_local_sdp, sdp_body, MAX_SDP_BODY_LEN - 1); // Must use MAX_SDP_BODY_LEN - 1, otherwise there is a potential overflow
	ast_log(LOG_DEBUG, "[session:%p] [caller] caller local sdp is:\n%s", session, sdp_body);

	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(callee_answer_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
	ast_log(LOG_DEBUG, "[session:%p] [callee] UCM <<<=== B . received 200 ok from the callee\n%s", session, sdp_body);

	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(callee_local_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
	ast_log(LOG_DEBUG, "[session:%p] [callee] local sdp of the callee is:\n%s", session, sdp_body);

	transparentsdp = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "TRANSPARENTSDP");
	if (transparentsdp && !strcasecmp(transparentsdp, "no")) {
		ast_log(LOG_DEBUG, "peer transparentsdp can not be supported");
		session->is_direct_media = 0;
	}

	if (!peer_session->is_direct_media) {
		session->is_direct_media = 0;
	}

	if (session->joint_ipvt || peer_session->joint_ipvt) {
		session->is_direct_media = 1;
		session->inv_session->neg->neg_local_sdp = pjmedia_sdp_session_clone(pool, callee_answer_sdp);
		{
			memset(sdp_body, 0, MAX_SDP_BODY_LEN);
			pjmedia_sdp_print(callee_answer_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
			ast_log(LOG_NOTICE, "[IPVT] modify_caller_sdp_for_wait_state ipvt sdp is callee sdp:\n%s", sdp_body);
		}
	} else if (session->is_direct_media) {
		session->inv_session->neg->neg_local_sdp = pjmedia_sdp_session_clone(session->inv_session->pool, callee_answer_sdp);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Modify caller local sdp according to callee answer sdp for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->inv_session->neg->neg_local_sdp = modify_local_sdp_with_callee_answer(pool, caller_local_sdp, callee_local_sdp, session, peer_session);

		session->inv_session->neg->neg_local_sdp = modify_local_sdp_for_passthrough(pool, session->inv_session->neg->neg_local_sdp, callee_answer_sdp, session, peer_session);

		modify_local_sdp_media_direction(pool, session->inv_session->neg->neg_local_sdp, callee_answer_sdp, session, peer_session);
	}

	memset(sdp_body, 0, MAX_SDP_BODY_LEN);
	pjmedia_sdp_print(session->inv_session->neg->neg_local_sdp, sdp_body, MAX_SDP_BODY_LEN - 1);
	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s caller local sdp is:\n%s",
		session, ast_sorcery_object_get_id(session->endpoint), sdp_body);

	ast_free(sdp_body);
	sdp_body = NULL;

	return;
}

static void modify_caller_sdp_for_done_state(pj_pool_t *pool, struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	if (session->is_direct_media) {
		ast_log(LOG_DEBUG, "Enable direct media, don't modify caller sdp for done state\n");
		return;
	}

	if (!peer_session->is_direct_media) {
		if (!session->inv_session->invite_tsx) {
			ast_log(LOG_WARNING, "Has no invite tsx, don't need process when both neg is done\n");
			return;
		}

		if (session->endpoint->media.rtp.accept_multiple_sdp_answers) {
			session->inv_session->updated_sdp_answer = 1;
			chan_pjsip_use_final_response_sdp(session, peer_session);
		}
	} else {
		/* modified by xxcui for Bug 76639, attended transfer under direct media,
		 * one channel is bridged and the other isn't. Refresh the peer target
		 */
		pjsip_dlg_dec_lock(peer_session->inv_session->dlg);
		peer_session->is_direct_media = 0;
		ao2_ref(peer_session, +1);
		if (ast_sip_push_task(peer_session->serializer, send_direct_media_reinvite_request, peer_session)) {
			ao2_cleanup(peer_session);
		}
		pjsip_dlg_inc_lock(peer_session->inv_session->dlg);
		return;
	}

	return;
}

static int check_params_for_sfu(struct ast_sip_session *session)
{
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;

	if (!session) {
		ast_log(LOG_DEBUG, "check_params_for_sfu session is invalid\n");
		return -1;
	}

	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "check_params_for_sfu, channel type is not PJSIP\n");
		return -1;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "check_params_for_sfu session is invalid\n");
		return -1;
	}

	chan_neg_state = session->inv_session->neg->state;
	if (chan_neg_state != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_DEBUG, "check_params_for_sfu, neg state(%d) is not WAIT_NEGO", chan_neg_state);
		return -1;
	}

	return 0;
}

#if 0
static unsigned int judge_if_has_host_video(struct ast_sip_session *session)
{
	unsigned int res = FALSE;

	/* To avoid channel and conference deadlock, unlock channel first */
	ast_channel_unlock(session->channel);

	if (ast_sip_session_get_host_main_video_media(session->channel, session->conf_room)) {
		res = TRUE;
	}

	ast_channel_lock(session->channel);

	return res;
}

static void modify_media_direction_for_sfu(struct ast_sip_session *session)
{
	unsigned i = 0;
	pjmedia_sdp_media *media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;
	pj_pool_t *pool = NULL;
	pjmedia_sdp_attr *attr_old = NULL;
	pjmedia_sdp_attr *attr_new = NULL;

	ast_log(LOG_DEBUG, "[session:%p] modify_media_direction_for_sfu for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!ast_sip_session_support_multiple_video(session) && judge_if_has_host_video(session)) {
		ast_log(LOG_WARNING, "[session:%p] modify_media_direction_for_sfu terminal_type is not webrtc or GVC, don't need modify for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	pool = session->inv_session->pool_prov;
	caller_local_sdp = session->inv_session->neg->neg_local_sdp;

	if (!pool || !caller_local_sdp) {
		ast_log(LOG_WARNING, "[session:%p] modify_media_direction_for_sfu pool or neg_local_sdp is invalid for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		media = caller_local_sdp->media[i];
		if (!media) {
			ast_log(LOG_WARNING, "[session:%p] modify_media_direction_for_sfu local_media is invalid for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}

		if (!ast_sip_session_judge_if_main_video(media)) {
			continue;
		}

		attr_old = pjmedia_sdp_media_find_attr2(media, "sendrecv", NULL);
		if (attr_old) {
			pjmedia_sdp_media_remove_attr(media, attr_old);
			attr_new = pjmedia_sdp_attr_create(pool, "recvonly", NULL);
			pjmedia_sdp_media_add_attr(media, attr_new);
			ast_log(LOG_DEBUG, "[session:%p] modify_media_direction_for_sfu change sendrecv to recvonly for main video, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
		}
	}

	return;
}
#endif

static struct ast_sip_session_media *get_main_video_session_media(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
		return session_media;
	}

	return NULL;
}

static void update_msid_for_change_host(struct ast_sip_session *session, char *host_msid)
{
	struct ast_sip_session_media *session_media = NULL;
	const char *local_msid = NULL;
	char msid[MAX_MSID_LEN] = {0};

	if (!session || !session->channel || !host_msid) {
		ast_log(LOG_WARNING, "update_msid_for_change_host params is NULL\n");
		return;
	}

	session_media = get_main_video_session_media(session);
	if (!session_media || !session_media->rtp) {
		ast_log(LOG_WARNING, "[session:%p] update_msid_for_change_host can' get main video session media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] update_msid_for_change_host update local msid to %s for endpoint:%s\n",
		session, host_msid, ast_sorcery_object_get_id(session->endpoint));

	local_msid = ast_rtp_instance_get_local_msid(session_media->rtp);
	if (!strcmp(local_msid, host_msid)) {
		ast_log(LOG_DEBUG, "[session:%p] local_msid is already the same as host_msid, don't need update for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	//send_reinvite_for_change_host(session, host_msid);

	if (!ast_strlen_zero(host_msid)) {
		snprintf(msid, sizeof(msid), "%s", host_msid);
	} else {
		snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
	}
	ast_rtp_instance_set_local_msid(session_media->rtp, msid);
	ast_avs_after_nego_notify(session->channel, session_media->rtp);

	return;
}

static void add_as_to_media(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	int as_value = 0;
	const pj_str_t STR_BANDW_MODIFIER = { "AS", 2 };
	pjmedia_sdp_bandw *bandw = NULL;

	as_value = ast_sip_session_get_as_according_type(session, resolution);
	if (!as_value) {
		return;
	}

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_as_to_media, AS:%d for endpoint:%s\n",
		session, as_value, ast_sorcery_object_get_id(session->endpoint));

	bandw->modifier = STR_BANDW_MODIFIER;
	bandw->value = as_value;
	media->bandw[media->bandw_count++] = bandw;

	return;
}

static void add_tias_to_media(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	int tias_value = 0;
	const pj_str_t STR_BANDW_MODIFIER = { "TIAS", 4 };
	pjmedia_sdp_bandw *bandw = NULL;

	tias_value = ast_sip_session_get_tias_according_type(session, resolution);
	if (!tias_value) {
		return;
	}

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_tias_to_media, TIAS:%d for endpoint:%s\n",
		session, tias_value, ast_sorcery_object_get_id(session->endpoint));

	bandw->modifier = STR_BANDW_MODIFIER;
	bandw->value = tias_value;
	media->bandw[media->bandw_count++] = bandw;

	return;
}

static void add_as_to_slides_video_media(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media)
{
	int as_value = AST_SLIDES_VIDEO_AS;
	const pj_str_t STR_BANDW_MODIFIER = { "AS", 2 };
	pjmedia_sdp_bandw *bandw = NULL;

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_as_to_slides_video_media, AS:%d for endpoint:%s\n",
		session, as_value, ast_sorcery_object_get_id(session->endpoint));

	bandw->modifier = STR_BANDW_MODIFIER;
	bandw->value = as_value;
	media->bandw[media->bandw_count++] = bandw;

	return;
}

static void add_tias_to_slides_video_media(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media)
{
	int tias_value = AST_SLIDES_VIDEO_TIAS;
	const pj_str_t STR_BANDW_MODIFIER = { "TIAS", 4 };
	pjmedia_sdp_bandw *bandw = NULL;

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_tias_to_slides_video_media, TIAS:%d for endpoint:%s\n",
		session, tias_value, ast_sorcery_object_get_id(session->endpoint));

	bandw->modifier = STR_BANDW_MODIFIER;
	bandw->value = tias_value;
	media->bandw[media->bandw_count++] = bandw;

	return;
}

static unsigned int get_slides_video_level_asymmetry_allowed(struct ast_sip_session *session)
{
	const unsigned int *level_asymmetry_allowed = NULL;
	struct ast_stream *stream = NULL;
	RAII_VAR(struct ast_format *, format, NULL, ao2_cleanup);

	stream = ast_sip_session_get_pending_slides_video_stream(session);
	if (!stream) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get slides video stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	format = ast_sip_session_get_video_format(stream, "h264");
	if (!format) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get slides video format, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	level_asymmetry_allowed = ast_format_attribute_get(format, "level-asymmetry-allowed");
	if (!level_asymmetry_allowed) {
		ast_log(LOG_WARNING, "[session:%p] Can't get profile level for H264 format, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (*level_asymmetry_allowed == 0) {
		return 0;
	}

	return 1;
}

/*Start added by zyxia for Bug 247293 ,2023.2.8*/
static void modify_video_fmtp_params(struct ast_sip_session *session,pjmedia_sdp_media *media,enum ast_media_sub_type sub_type)
{
	RAII_VAR(struct ast_format *, format, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format *, new_format, NULL, ao2_cleanup);
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *image_attr = NULL;
	struct ast_stream *stream = NULL;
	unsigned int payload =0;
	char fmtp_string[512]={0};
	char image_string[512]={0};

	if ( !session || !media ) {
		return;
	}

	/* Get fmtp for first codec */
	attr = pjmedia_sdp_media_find_attr2(media, "fmtp", &media->desc.fmt[0]);
	if (!attr) {
		return;
	}

	stream = ast_sip_session_get_pending_stream_by_subtype(session,sub_type);
	if(!stream){
		return;
	}

	image_attr = pjmedia_sdp_media_find_attr2(media, "imageattr", &media->desc.fmt[0]);
	if (image_attr && sscanf(image_attr->value.ptr, "%30u %255[^\t\n]", &payload, image_string) == 2) {
		ast_stream_set_metadata(stream, "IMAGEATTR", image_string);
		ast_log(LOG_DEBUG, "[session:%p] modify_video_fmtp_params imageattr %s endpoint:%s\n",
			session,image_string, ast_sorcery_object_get_id(session->endpoint));
	}
	else
		ast_stream_set_metadata(stream, "IMAGEATTR", NULL);

	format = ast_sip_session_get_video_format(stream, "h264");
	if (!format) {
		ast_log(LOG_ERROR, "[session:%p] Can't get video format, don't get H264 video level, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return ;
	}

	if(sscanf(attr->value.ptr, "%30u %255[^\t\n]", &payload, fmtp_string) != 2){
		return;
	}

	ast_format_attribute_reset(format);

	char *attribs = ast_strdupa(fmtp_string), *attrib;
	while ((attrib = strsep(&attribs, ";"))) {
		char val[256] = "";
		attrib = ast_strip(attrib);
		if (sscanf(attrib, "profile-level-id=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "profile-level-id", attrib);
		}
		else if (sscanf(attrib, "sprop-parameter-sets=%s",val) == 1) {
			ast_format_attribute_set_directly(format, "sprop-parameter-sets", attrib);
		} else if (sscanf(attrib, "packetization-mode=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "packetization-mode", val);
		}else if (sscanf(attrib, "level-asymmetry-allowed=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "level-asymmetry-allowed", val);
		} else if (sscanf(attrib, "max-fs=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-fs", val);
		}
		#if 0
		else if (sscanf(attrib, "max-mbps=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-mbps", val);
		else if (sscanf(attrib, "max-cpb=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-cpb", val);
		} else if (sscanf(attrib, "max-dpb=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-dpb", val);
		} else if (sscanf(attrib, "max-br=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-br", val);
		} else if (sscanf(attrib, "max-smbps=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-smbps", val);
		} else if (sscanf(attrib, "max-fps=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-fps", val);
		}

		else if (sscanf(attrib, "redundant-pic-cap=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "redundant-pic-cap", val);
		} else if (sscanf(attrib, "parameter-add=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "parameter-add", val);
		} else if (sscanf(attrib, "sprop-interleaving-depth=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "sprop-interleaving-depth", val);
		} else if (sscanf(attrib, "sprop-deint-buf-req=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "sprop-deint-buf-req", val);
		} else if (sscanf(attrib, "deint-buf-cap=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "deint-buf-cap", val);
		} else if (sscanf(attrib, "sprop-init-buf-time=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "sprop-init-buf-time", val);
		} else if (sscanf(attrib, "sprop-max-don-diff=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "sprop-max-don-diff", val);
		} else if (sscanf(attrib, "max-rcmd-nalu-size=%s", val) == 1) {
			ast_format_attribute_set_directly(format, "max-rcmd-nalu-size", val);
		}
		#endif
	}


		ast_log(LOG_DEBUG, "[session:%p] modify   endpoint:%s,format:%p\n",
			session,ast_sorcery_object_get_id(session->endpoint),format);

	return;
}
static void modify_video_fmtp_params_for_callflip(struct ast_sip_session *remote_session,struct ast_sip_session *local_session)
{
	unsigned int i = 0;
	pjmedia_sdp_media *media = NULL;
	pjmedia_sdp_session *local_sdp = NULL;

	if (!remote_session || !local_session) {
		return;
	}

	if (!local_session->inv_session || !local_session->inv_session->neg) {
		return;
	}

	local_sdp = local_session->inv_session->neg->active_local_sdp;
	if (!local_sdp) {
		ast_log(LOG_WARNING, "[session:%p] modify_video_fmtp_params_for_callflip ,local active sdp is null,endpoint:%s\n",
			remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
		return;
	}

	for (media = NULL, i = 0; i < local_sdp->media_count; ++i) {
		media = local_sdp->media[i];
		if (ast_sip_session_judge_if_main_video(media)) {
			modify_video_fmtp_params(remote_session,media,AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		}
		else if (ast_sip_session_judge_if_slides_video(media)){
			modify_video_fmtp_params(remote_session,media,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
		}
	}

	return;
}


#if 0
static void remove_imageattr_for_callflip(struct ast_sip_session *session)
{
	if (!session ) {
		ast_log(LOG_DEBUG, "[session:%p] remove_imageattr_for_callflipendpoint:%s\n",
			session,  ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	struct ast_stream *main_stream = NULL;
	struct ast_stream *slides_stream = NULL;

	main_stream = ast_sip_session_get_pending_stream_by_subtype(session,AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	if(!main_stream){
		return;
	}
	ast_stream_set_metadata(main_stream, "IMAGEATTR", NULL);

	slides_stream = ast_sip_session_get_pending_stream_by_subtype(session,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if(!slides_stream){
		return;
	}
	ast_stream_set_metadata(slides_stream, "IMAGEATTR", NULL);

	return;
}
#endif
static void modify_profile_level_id_for_sfu(struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	unsigned int i = 0;
	pjmedia_sdp_attr *attr = NULL;
	char *profile_level_id_str = NULL;
	char *video_resolution = NULL;

	if (!session || !media) {
		return;
	}

	if (!media->desc.port) {
		return;
	}

	/* Get fmtp for first codec */
	attr = pjmedia_sdp_media_find_attr2(media, "fmtp", &media->desc.fmt[i]);
	if (!attr) {
		return;
	}
	profile_level_id_str = strstr(attr->value.ptr, "profile-level-id=");
	if (!profile_level_id_str) {
		return;
	}

	video_resolution = ast_sip_session_get_video_resolution_according_type(session, resolution);
	if (video_resolution) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s replace profile-level-id to %s\n",
			session, ast_sorcery_object_get_id(session->endpoint), video_resolution);
		strncpy(profile_level_id_str, video_resolution, strlen(video_resolution));
	} else {
		if (ast_sip_session_judge_if_main_video(media)) {
			ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s can't get video resource, just decline this media\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			media->desc.port = 0;
			ast_conf_put_main_video_resource(session->conf_room, session->channel);
		}
	}

	return;
}

static void modify_resolution_sdp_for_sfu(struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	if (!session) {
		return;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			modify_profile_level_id_for_sfu(session, media, resolution);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			break;
		default:
			break;
	}

	return;
}

static void modify_main_video_as_and_tias_for_sfu(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	if (!pool || !session || !media) {
		return;
	}

	if (resolution == VIDEO_RESOLUTION_NONE) {
		return;
	}

	if (!media->desc.port) {
		return;
	}

	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p]  support_special_attrs is disable, don't need modify main video AS/TIAS for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	media->bandw_count = 0;

	add_as_to_media(pool, session, media, resolution);
	add_tias_to_media(pool, session, media, resolution);

	return;
}

static void modify_slides_video_as_and_tias_for_sfu(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media)
{
	if (!pool || !session || !media) {
		return;
	}

	if (!media->desc.port) {
		return;
	}

	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p]  support_special_attrs is disable, don't need modify slides video AS/TIAS for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	media->bandw_count = 0;

	add_as_to_slides_video_media(pool, session, media);
	add_tias_to_slides_video_media(pool, session, media);

	return;
}

static void modify_slides_video_framerate_for_sfu(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr = NULL;

	if (!pool || !session || !media) {
		return;
	}

	if (!media->desc.port) {
		return;
	}

	attr = pjmedia_sdp_media_find_attr2(media, "framerate", NULL);
	if (attr) {
		pjmedia_sdp_media_remove_attr(media, attr);
	}

	snprintf(tmp, sizeof(tmp), "%d", AST_SLIDES_VIDEO_FRAMERATE);
	attr = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
	media->attr[media->attr_count++] = attr;

	return;
}

static void modify_imageattr_for_sfu(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *new_attr = NULL;
	char *imageattr_str = NULL;
	char new_attr_value[128] = {0};
	pj_str_t new_attr_str;

	if (!session || !media) {
		return;
	}

	if (!media->desc.port) {
		return;
	}

	/* Get fmtp for first codec */
	attr = pjmedia_sdp_media_find_attr2(media, "imageattr", &media->desc.fmt[0]);
	if (!attr) {
		return;
	}

	if ((session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) ||
		(session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		imageattr_str = ast_sip_session_get_wave_imageattr_according_type(resolution);
	} else {
		imageattr_str = ast_sip_session_get_imageattr_according_type(resolution);
	}

	if (imageattr_str) {
		snprintf(new_attr_value, sizeof(new_attr_value), "%ld %s", pj_strtoul(&media->desc.fmt[0]), imageattr_str);
		ast_log(LOG_DEBUG, "Replace imageattr to %s\n", new_attr_value);
		pjmedia_sdp_media_remove_attr(media, attr);
		new_attr = pjmedia_sdp_attr_create(pool, "imageattr", pj_cstr(&new_attr_str, new_attr_value));
		pjmedia_sdp_media_add_attr(media, new_attr);
	}

	return;
}

static void set_resolution_for_stream(struct ast_stream *stream, enum video_resolution resolution)
{
	if (!stream) {
		return;
	}

	switch (resolution) {
		case VIDEO_RESOLUTION_NONE:
			ast_stream_set_video_resolution(stream, AST_STREAM_VIDEO_RESOLUTION_NONE);
			break;
		case VIDEO_RESOLUTION_360P:
			ast_stream_set_video_resolution(stream, AST_STREAM_VIDEO_RESOLUTION_360P);
			break;
		case VIDEO_RESOLUTION_720P:
			ast_stream_set_video_resolution(stream, AST_STREAM_VIDEO_RESOLUTION_720P);
			break;
		case VIDEO_RESOLUTION_1080P:
			ast_stream_set_video_resolution(stream, AST_STREAM_VIDEO_RESOLUTION_1080P);
			break;
		default :
			break;
	}

	return;
}

static void change_media_direction(pj_pool_t *pool, pjmedia_sdp_media *media, char *dir)
{
	pjmedia_sdp_attr *a = NULL;

	if (!pool || !media) {
		return;
	}

	ast_log(LOG_DEBUG, "Change media direction to %s\n", dir);

	a = pjmedia_sdp_attr_create(pool, dir, NULL);
	if (a) {
		ast_remove_all_media_directions(media);
		pjmedia_sdp_media_add_attr(media, a);
	}

	return;
}

static unsigned int need_to_get_video_resource(pjmedia_sdp_media *media)
{
	if (!media) {
		return 0;
	}

	if (pjmedia_sdp_attr_find2(media->attr_count, media->attr, "sendonly", NULL) ||
		pjmedia_sdp_attr_find2(media->attr_count, media->attr, "inactive", NULL)) {
		return 0;
	}

	if (ast_sip_session_judge_if_disable_media(media)) {
		return 0;
	}

	return 1;
}

static void modify_main_video_for_no_resolution(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, int resolution)
{
	pjmedia_sdp_attr *attr = NULL;
	pj_str_t stmp;

	if (!pool || !session || !media) {
		return;
	}

	/* That's GVC3210 BUG, GCV3210 even support multiple video media, but need to reserve main video, otherwise will abnormal */
	if (!ast_sip_session_support_multiple_video(session) ||
		session->terminal_type == AST_TERMINAL_TYPE_GVC) {
		change_media_direction(pool, media, "sendonly");
		ast_sip_session_modify_video_codec_order_for_sfu(pool, session, media, resolution);
	} else {
		if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && ast_sip_session_media_use_bundle(session)) {
			change_media_direction(pool, media, "inactive");
		} else {
			media->desc.port = 0;
			if (ast_sip_session_media_use_bundle(session)) {
				ast_sip_session_remove_attrs_for_disable_media(media);
				ast_sip_session_remove_media_from_bundle_group(session, pool, session->inv_session->neg->neg_local_sdp, media);
			}
		}
	}

	/* Add declined-reason attr for no video resource or cpu limit */
	if (resolution == -2) {
		attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "CPU limit"));
	} else {
		attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "no video resource"));
	}

	media->attr[media->attr_count++] = attr;

	return;
}

static enum video_resolution get_resolution_from_h264_level(struct ast_sip_session *session, unsigned int is_nway)
{
	unsigned int h264_level = 0;
	enum video_resolution result = VIDEO_RESOLUTION_NONE;

	h264_level = ast_sip_session_get_h264_level(session, is_nway);
	if (!h264_level) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get H264 level, can't modify video resolution, disable video for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return result;
	}

	/* For mobile APP, use 360P as default */
	if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		h264_level = VIDEO_RESOLUTION_360P;
	}

	result = ast_sip_session_get_main_video_resource(session, h264_level);

	return result;
}

static enum video_resolution get_resolution_from_imageattr(struct ast_sip_session *session,
	pjmedia_sdp_attr *attr, unsigned int is_nway)
{
	int rtp_code = 0;
	int send_x = 0;
	int send_y = 0;
	char imageattr_all[MAX_IMAGERTTR_LEN + 4] = {0};
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;

	ast_copy_pj_str(imageattr_all, &attr->value, sizeof(imageattr_all));
	if (!strlen(imageattr_all)) {
		ast_log(LOG_DEBUG, "[session:%p] get_video_resolution_from_imageattr can't get imageattr vaule, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		resolution = get_resolution_from_h264_level(session, is_nway);
		return resolution;
	}

	if (sscanf(imageattr_all, "%d send [x=%d,y=%d]", &rtp_code, &send_x, &send_y) != 3) {
		ast_log(LOG_DEBUG, "[session:%p] get_video_resolution_from_imageattr, can't process this imageattr attr:%s, endpoint:%s\n",
			session, imageattr_all, ast_sorcery_object_get_id(session->endpoint));
		resolution = get_resolution_from_h264_level(session, is_nway);
		return resolution;
	}

	ast_log(LOG_DEBUG, "[session:%p] get_video_resolution_from_imageattr x=%d, y=%d, endpoint:%s\n",
		session, send_x, send_y, ast_sorcery_object_get_id(session->endpoint));

	if (send_y < 720) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_360P;
	} else if (send_y < 1080) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_720P;
	} else {
		resolution = AST_STREAM_VIDEO_RESOLUTION_1080P;
	}

	resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, resolution);

	return resolution;
}

static enum video_resolution get_resolution_for_h264_codec(struct ast_sip_session *session, int is_nway)
{
	unsigned int i = 0;
	pj_pool_t *pool = NULL;
	pjmedia_sdp_media *media = NULL;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_session *caller_remote_sdp = NULL;
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;

	if (!session || !session->inv_session) {
		return VIDEO_RESOLUTION_NONE;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED && !is_nway) {
		ast_log(LOG_DEBUG, "[session:%p] We don't need modify video resolution for re-INVITE, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return VIDEO_RESOLUTION_NONE;
	}

	pool = session->inv_session->pool_prov;
	if (!is_nway) {
		caller_remote_sdp = session->inv_session->neg->neg_remote_sdp;
	} else {
		ast_log(LOG_DEBUG, "[session:%p] get_resolution_for_h264_codec is nway process, use active sdp, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		caller_remote_sdp = session->inv_session->neg->active_remote_sdp;
	}

	if (!pool || !caller_remote_sdp) {
		ast_log(LOG_WARNING, "[session:%p] get_resolution_for_h264_codec pool or neg_local_sdp is invalid for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return VIDEO_RESOLUTION_NONE;
	}

	ast_log(LOG_DEBUG, "[session:%p] get_resolution_for_h264_codec for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	for (media = NULL, i = 0; i < caller_remote_sdp->media_count; ++i) {
		media = caller_remote_sdp->media[i];
		if (!ast_sip_session_judge_if_main_video(media)) {
			continue;
		}

		if (!(attr = pjmedia_sdp_media_find_attr2(media, "imageattr", NULL))) {
			resolution = get_resolution_from_h264_level(session, is_nway);
			break;
		}

		resolution = get_resolution_from_imageattr(session, attr, is_nway);

		break;
	}

	return resolution;
}

static enum video_resolution get_resolution_for_main_video(struct ast_sip_session *session)
{
	enum video_resolution result = VIDEO_RESOLUTION_NONE;

	if (!session) {
		return result;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			result = get_resolution_for_h264_codec(session, 0);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			result = VIDEO_RESOLUTION_360P;
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			result = VIDEO_RESOLUTION_360P;
			break;
		default:
			break;
	}

	return result;
}

static int media_has_conf_codec(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	unsigned int i = 0;
	int has_conf_codec = 0;
	char conf_codec[8] = { 0 };
	char codec_name[32] = {0};
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;

	if (!pool || !media || !session) {
		return 0;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			strcpy(conf_codec, "H264");
			break;
		case AST_SFU_VIDEO_H265:
			strcpy(conf_codec, "H265");
			break;
		case AST_SFU_VIDEO_VP9:
			strcpy(conf_codec, "VP9");
			break;
		default:
			strcpy(conf_codec, "H264");
			break;
	}

	for (i = 0; i < media->desc.fmt_count; ++i) {
		if (!(attr = pjmedia_sdp_media_find_attr2(media, "rtpmap", &media->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(codec_name, &rtpmap->enc_name, 32);

		if (!strcasecmp(codec_name, conf_codec)) {
			has_conf_codec = 1;
			break;
		}
	}

	return has_conf_codec;
}

static void disable_video_as_no_video_codec(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *attr = NULL;
	pj_str_t stmp;

	ast_log(LOG_DEBUG, "[session:%p] disable_video_as_no_video_codec, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && ast_sip_session_media_use_bundle(session)) {
		change_media_direction(pool, media, "inactive");
	} else {
		media->desc.port = 0;
		if (ast_sip_session_media_use_bundle(session)) {
			ast_sip_session_remove_attrs_for_disable_media(media);
			ast_sip_session_remove_media_from_bundle_group(session, pool, session->inv_session->neg->neg_local_sdp, media);
		}
	}

	/* Add declined-reason attr for no video codec */
	attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "no video codec"));
	if (attr) {
		media->attr[media->attr_count++] = attr;
	}

	return;
}

static void modify_main_video_params_for_sfu(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	struct ast_stream *stream = NULL;
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;

	/* Don't need to check params, we already check before */

	if (!need_to_get_video_resource(media)) {
		ast_log(LOG_DEBUG, "[session:%p] Media direction is sendonly or inactive, don't need get video resource for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!media_has_conf_codec(pool, session, media)) {
		disable_video_as_no_video_codec(pool, session, media);
		return;
	}

	resolution = get_resolution_for_main_video(session);
	if ((resolution <= VIDEO_RESOLUTION_NONE) || (resolution >= VIDEO_RESOLUTION_MAX)) {
		modify_main_video_for_no_resolution(pool, session, media, resolution);
		return;
	}

	ast_sip_session_set_conf_video_codec(pool, session, media, resolution);
	if (ast_sip_session_modify_video_codec_order_for_sfu(pool, session, media, resolution)) {
		ast_log(LOG_DEBUG, "[session:%p] There is something wrong for chose video codec, don't modify other param for sfu, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] modify_main_video_params_for_sfu enpoint:%s, resolution:%d\n",
		session, ast_sorcery_object_get_id(session->endpoint), resolution);

	modify_resolution_sdp_for_sfu(session, media, resolution);
	modify_main_video_as_and_tias_for_sfu(pool, session, media, resolution);
	modify_imageattr_for_sfu(pool, session, media, resolution);

	stream = ast_sip_session_get_pending_main_video_stream(session);
	set_resolution_for_stream(stream, resolution);

	return;
}

static void modify_slides_video_to_h264_codec(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	unsigned int i = 0;
	unsigned int final_fmtp_count = 0;
	unsigned int find_h264_codec = 0;
	pjmedia_sdp_attr *a_rtpmap = NULL;
	pjmedia_sdp_attr *a_fmtp = NULL;
	const pj_str_t str_h264 = { "H264", 4 };
	const pj_str_t str_rtx = { "rtx", 3 };
	const pj_str_t str_gs_fec = { "GS-FEC", 6 };
	pj_str_t pt_local[PJMEDIA_MAX_SDP_FMT];

	for (i = 0; i < media->desc.fmt_count; i++) {
		a_rtpmap = pjmedia_sdp_media_find_attr2(media, "rtpmap", &media->desc.fmt[i]);
		if (!a_rtpmap) {
			continue;
		}

		if (pj_strstr(&a_rtpmap->value, &str_h264) != NULL) {
			find_h264_codec = 1;
			pt_local[final_fmtp_count] = media->desc.fmt[i];
			final_fmtp_count++;
			continue;
		}

		if (pj_strstr(&a_rtpmap->value, &str_rtx) != NULL ||
			pj_strstr(&a_rtpmap->value, &str_gs_fec) != NULL) {
			pt_local[final_fmtp_count] = media->desc.fmt[i];
			final_fmtp_count++;
			continue;
		}

		pjmedia_sdp_media_remove_attr(media, a_rtpmap);

		a_fmtp = pjmedia_sdp_media_find_attr2(media, "fmtp", &media->desc.fmt[i]);
		if (NULL != a_fmtp) {
			pjmedia_sdp_media_remove_attr(media, a_fmtp);
		}
	}

	if (!find_h264_codec) {
		ast_log(LOG_DEBUG, "We can't find H264 codec, disable slides video for endpoint:%s\n", ast_sorcery_object_get_id(session->endpoint));
		media->desc.port = 0;
		return;
	}

	/* We only to retain H264 codec for slides video */
	media->desc.fmt_count = final_fmtp_count;
	for (i = 0; i < final_fmtp_count; i++) {
		media->desc.fmt[i] = pt_local[i];
	}

	ast_sip_session_change_rtx_params(pool, session, media, "H264", media->desc.fmt[0]);

	return;
}

static void modify_slides_video_codec_for_sfu(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	if (!session) {
		return;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			modify_slides_video_to_h264_codec(pool, session, media);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			break;
		default:
			break;
	}

	return;
}

static void modify_slides_video_resolution_for_h264_codec(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	if (get_slides_video_level_asymmetry_allowed(session)) {
		modify_profile_level_id_for_sfu(session, media, VIDEO_RESOLUTION_1080P);
		modify_imageattr_for_sfu(pool, session, media, VIDEO_RESOLUTION_1080P);
	}

	return;
}

static void modify_slides_video_resolution_for_sfu(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	if (!session) {
		return;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			modify_slides_video_resolution_for_h264_codec(pool, session, media);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			break;
		default:
			break;
	}

	return;
}

static void modify_slides_video_params_for_sfu(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	modify_slides_video_codec_for_sfu(pool, session, media);
	modify_slides_video_resolution_for_sfu(pool, session, media);
	modify_slides_video_as_and_tias_for_sfu(pool, session, media);
	modify_slides_video_framerate_for_sfu(pool, session, media);

	return;
}

static void modify_video_params_for_sfu(struct ast_sip_session *session)
{
	unsigned int i = 0;
	pj_pool_t *pool = NULL;
	pjmedia_sdp_media *media = NULL;
	pjmedia_sdp_session *caller_local_sdp = NULL;

	if (!session || !session->inv_session) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		ast_log(LOG_DEBUG, "[session:%p] We don't need modify video resolution for re-INVITE, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	pool = session->inv_session->pool_prov;
	caller_local_sdp = session->inv_session->neg->neg_local_sdp;

	if (!pool || !caller_local_sdp) {
		ast_log(LOG_WARNING, "[session:%p] modify_video_params_for_sfu pool or neg_local_sdp is invalid for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] modify_video_params_for_sfu for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	for (media = NULL, i = 0; i < caller_local_sdp->media_count; ++i) {
		media = caller_local_sdp->media[i];

		if (ast_sip_session_judge_if_main_video(media)) {
			modify_main_video_params_for_sfu(pool, session, media);
		} else if (ast_sip_session_judge_if_slides_video(media)) {
			modify_slides_video_params_for_sfu(pool, session, media);
		}
	}

	return;
}

static void modify_sdp_for_sfu(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	if (check_params_for_sfu(session)) {
		return;
	}

	if (ast_sip_session_get_only_support_audio()) {
		disable_media_except_audio(session->channel);
		return;
	}

	//modify_media_direction_for_sfu(session);

	modify_video_params_for_sfu(session);

	return;
}

static void chan_pjsip_special_answer_for_no_peer(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	const char *conf_room = NULL;
	const char *is_sfu = NULL;

	if (!chan) {
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return;
	}

	conf_room = pbx_builtin_getvar_helper(chan, "CONF_ROOM");
	is_sfu = pbx_builtin_getvar_helper(chan, "IS_SFU");

#if 0
	/*Start added by zyxia for trunk conf info,2022.11.17*/
	struct ast_sip_endpoint *endpoint = chan_pvt->session->endpoint;
	const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
	if (!ast_strlen_zero(conf_room) && endpoint != NULL &&
		endpoint_id != NULL && strcasestr(endpoint_id, "trunk") ) {//remove
		if (globalEnabledCEI || chan_pvt->session->is_sfu) {
			ast_log(LOG_DEBUG, "[session:%p] Set is_sfu to yes, and conf_room to %s for endpoint:%s\n",
				chan_pvt->session, conf_room, ast_sorcery_object_get_id(chan_pvt->session->endpoint));
			add_peer_conference(ast_channel_name(chan), endpoint, conf_room);
		}
	}
	/*End added by zyxia for trunk conf info,2022.11.17*/
#endif

	if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
		set_session_sfu_params(chan_pvt->session, conf_room);
		modify_sdp_for_sfu(chan_pvt->session);

		/*Start Added by zyxia for Bug213762,2022.1.22, callflip conf, default is audio*/
		if (ast_channel_get_is_callflip(chan))
			disable_main_video(chan,TRUE);
		/*End by zyxia for Bug213762,2022.1.22*/
	} else if (!ast_strlen_zero(conf_room) || (ast_channel_page_mode_get(chan) && !ast_channel_get_is_page_video_support(chan)) || ast_channel_vm_get(chan)) {
		ast_log(LOG_DEBUG, "[session:%p] Enter audio conference room/page/VM, disable media except audio, endpoint:%s\n",
			chan_pvt->session, ast_sorcery_object_get_id(chan_pvt->session->endpoint));
			disable_media_except_audio(chan);
	}

	/*Start added by zyxia for video paging,2022.8.5*/
	if (ast_channel_page_mode_get(chan) && ast_channel_get_is_page_video_support(chan) && !ast_channel_get_is_callflip(chan)){//edit by zyxia for Bug 258912
		change_video_media_direction(chan,AST_MEDIA_SUB_VIDEO_TYPE_MAIN,"recvonly");
		change_video_media_direction(chan,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES,"inactive");//added by zyxia for Bug 267391 ,not support slides
	}
	else
	/*End added by zyxia for video paging,2022.8.5*/
		/* Add by bfdu for bug 130876, 2019.01.28 */
		disable_main_video(chan,FALSE);

	/* Disable direct media for answered with no peer channel, 2019.04.08 */
	disable_direct_media(chan);

	return;
}

/*start added by zyxia for  Bug 241677,2022.11.3 */
static pjmedia_sdp_attr* pjmedia_sdp_attr_change_fmtp_param_by_name(pj_pool_t *pool,
	pjmedia_sdp_attr *attr_fmtp,const char* param_str,const char *param_value)
{
	pj_str_t attr_name = pj_str("fmtp");
	char find_param_str[256] = { 0 };
	char* fmtp_profile = NULL;
	char* fmtp_buf = NULL;
	pj_str_t fmtp_value;
	int len_fmtp_value = 0;

	if (!pool || !attr_fmtp  ||!param_str || !param_value ) {
		ast_log(LOG_DEBUG, "Param is null, don't change");
		return NULL;
	}

	snprintf(find_param_str, sizeof(find_param_str), "%s", param_str);
	pj_str_t find_str = pj_str(find_param_str);

	if (pj_strcmp(&attr_fmtp->name, &attr_name)) {
		ast_log(LOG_DEBUG, "It's not a fmtp, don't change");
		return NULL;
	}

	len_fmtp_value = attr_fmtp->value.slen;

	fmtp_buf = pj_pool_alloc(pool, len_fmtp_value);
	pj_strset(&fmtp_value, fmtp_buf, 0);
	pj_strcpy(&fmtp_value, &attr_fmtp->value);
	fmtp_profile = pj_strstr(&fmtp_value, &find_str);
	if (!fmtp_profile) {
		ast_log(LOG_DEBUG, "Can't find %s string, don't need change\n",param_str);
		return pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_value);
	}

	/* don't copy '\0' to fmtp_profile, 23 is the len of "profile-level-id=428014" */
	strncpy(fmtp_profile, param_value,strlen(param_value));
	ast_log(LOG_DEBUG, "pjmedia_sdp_attr_change_fmtp_param_by_name fmtp_profile :%s old:%s",fmtp_profile,pj_strbuf(&fmtp_value));
	return pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_value);
}
	/*start added by zyxia for	Bug 241677,2022.11.3 */

/*Start added by zyxia for Bug 205228,2021.10.9 */
/* video payload and attrs need to modify result according to rtsp sdp info */
static void update_sdp_for_callee_answer_rtsp(pj_pool_t *pool,struct ast_sip_session* local_session, pjmedia_sdp_media *result,struct ast_channel *callee_chan)
{
	int local_fmt_count = 0;
	pjmedia_sdp_attr *a_o_fmtp = NULL;
	pjmedia_sdp_attr *a_o_rtpmap = NULL;
	pjmedia_sdp_rtpmap caller_rtpmap;
	int final_fmt_count=0;
	char rtpmap_name[64] = { 0 };
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *fmtp_attr = NULL;
	char local_video_format_name[64]={0};

	const char* video_format_name = pbx_builtin_getvar_helper(callee_chan, "RTSP_VIDEO_FORMAT");
	const char* video_rtsp_level_id = pbx_builtin_getvar_helper(callee_chan, "RTSP_PROFILE_LEVEL_ID");


	/* to remove rtpmap/fmtp for payload can't find in callee answer */
	for (local_fmt_count = 0; local_fmt_count < result->desc.fmt_count; ++local_fmt_count) {
		attr = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[local_fmt_count]);
		if (!attr) {
			ast_log(LOG_DEBUG, "[session:%p] It's has no rtpmap for caller, don't consider a dynamic payload, endpoint:%s\n",
				local_session, ast_sorcery_object_get_id(local_session->endpoint));
			return;
		}

		pjmedia_sdp_attr_get_rtpmap(attr, &caller_rtpmap);
		ast_copy_pj_str(rtpmap_name, &caller_rtpmap.enc_name, sizeof(rtpmap_name));
		snprintf(local_video_format_name,(int)caller_rtpmap.enc_name.slen+1,"%s",caller_rtpmap.enc_name.ptr);
		ast_log(LOG_NOTICE, "[update_sdp_for_local]local video formats:%s\n",local_video_format_name);

		/*start added by zyxia for 206254 ,2021.10.18*/
		if ((!ast_strlen_zero(video_format_name) && !strcasecmp(video_format_name, local_video_format_name))
			|| (!strcasecmp("RTX", local_video_format_name)) || (!strcasecmp("GS-FEC", local_video_format_name)) || (!strcasecmp("H264", local_video_format_name))) { //added by zyxia for Bug207787  fec
			/* to remove payload in m line */
			result->desc.fmt[final_fmt_count] = result->desc.fmt[local_fmt_count];
			final_fmt_count++;

			/*set the rtsp profile-level-id to 200 ok,for  Bug 241677,2022.11.3 */
			if(!strcasecmp("H264", local_video_format_name) ){
				fmtp_attr = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
				if (!fmtp_attr) {
					ast_log(LOG_DEBUG, "[session:%p] It's has no fmtp for caller, don't consider a dynamic payload, endpoint:%s\n",
						local_session, ast_sorcery_object_get_id(local_session->endpoint));
				}
				else {
					pjmedia_sdp_attr *new_fmtp_attr = NULL;
					new_fmtp_attr = pjmedia_sdp_attr_change_fmtp_param_by_name(pool,fmtp_attr,"profile-level-id",video_rtsp_level_id);
					if(new_fmtp_attr) {
						pjmedia_sdp_media_remove_attr(result, fmtp_attr);
						pjmedia_sdp_media_add_attr(result, new_fmtp_attr);
					}
				}
			}
			continue;
		}
		/*End added by zyxia for 206254 ,2021.10.18*/

		/* need to update fmtp attributes to pass through some video attributes. */
		a_o_fmtp = pjmedia_sdp_media_find_attr2(result, "fmtp", &result->desc.fmt[local_fmt_count]);
		a_o_rtpmap = pjmedia_sdp_media_find_attr2(result, "rtpmap", &result->desc.fmt[local_fmt_count]);

		//remove fmtp
		if (a_o_fmtp )
			pjmedia_sdp_media_remove_attr(result, a_o_fmtp);
		//remove rtpmtp
		if(a_o_rtpmap)
			pjmedia_sdp_media_remove_attr(result, a_o_rtpmap);
	}

	/*start added by zyxia for 206254 ,2021.10.18*/
	if(final_fmt_count)
		result->desc.fmt_count = final_fmt_count;
	else
	/*End added by zyxia for 206254 ,2021.10.18*/
		result->desc.port = 0; //added by zyxia for Bug 207797 ,2021.11.17

	ast_log(LOG_NOTICE, "[update_sdp_for_local]final_fmt_count:%d\n",final_fmt_count);

	return;
}
/*End added by zyxia for Bug 205228,2021.10.9 */

static void update_sdp_for_local(struct ast_channel* chan, struct ast_channel* peer) {
	if (!peer || !chan || strcasecmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_NOTICE, "[update_sdp_for_local] Invalid args.\n");
		return;
	}

	int caller_is_gswave = 0;

	if (strncasecmp(ast_channel_tech(peer)->type, "RTSP", 4)){ //not RTSP channel

		const char* video_port_zero = pbx_builtin_getvar_helper(peer, "UPDATE_VIDEO_PORT_ZERO");
		if (!video_port_zero) {
			ast_log(LOG_DEBUG, "[update_sdp_for_local] The channel(%s) no UPDATE_VIDEO_PORT_ZERO channel var.\n", ast_channel_name(chan));
			return;
		}

		if (strcasecmp(video_port_zero, "YES")) {
			ast_log(LOG_NOTICE, "[update_sdp_for_local] The channel var(video_port_zero): %s.\n", video_port_zero);
			return;
		}
	}


	struct ast_sip_session* caller_session = get_sip_session_from_channel(chan);
	if (!caller_session) {
		ast_log(LOG_NOTICE, "[update_sdp_for_local] The caller_session is nullptr.\n");
		return;
	}

	if (strncasecmp(ast_channel_tech(peer)->type, "RTSP", 4)){ //not RTSP channel
		if (caller_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
			caller_is_gswave = 1;
		}
	}

	struct pjmedia_sdp_session* caller_local_sdp_sess = NULL;
	caller_local_sdp_sess = caller_session->inv_session->neg->neg_local_sdp;
	if (!caller_local_sdp_sess) {
		ast_log(LOG_NOTICE, "[update_sdp_for_local] The caller_local_sdp_sess is nullptr.\n");
		return;
	}

	struct ast_stream_topology *topology = NULL;
	if (caller_session->pending_media_state && caller_session->pending_media_state->topology) {
		topology = caller_session->pending_media_state->topology;
	} else if (caller_session->active_media_state && caller_session->active_media_state->topology) {
		topology = caller_session->active_media_state->topology;
	}
	if (!topology) {
		ast_log(LOG_NOTICE, "[update_sdp_for_local] The caller's media topology is nullptr.\n");
		return;
	}

	struct pjmedia_sdp_media* sdp_media = NULL;

	for (int i = 0; i < caller_local_sdp_sess->media_count; ++ i) {
		sdp_media = caller_local_sdp_sess->media[i];
		struct ast_stream* stream = ast_stream_topology_get_stream(topology, i);
		if (!sdp_media || !stream) {
			continue;
		}

		//ipc media need set to sendonly
		if (strncasecmp(sdp_media->desc.media.ptr, "audio", 5) && !strncasecmp(ast_channel_tech(peer)->type, "RTSP", 4)) {
			change_media_direction(caller_session->inv_session->pool_prov,sdp_media,"sendonly");
		}
		if (!strncasecmp(sdp_media->desc.media.ptr, "video", 5) && ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
			/*Start added by zyxia for Bug 205228,2021.10.9 */
			if (!strncasecmp(ast_channel_tech(peer)->type, "RTSP", 4)){
				ast_log(LOG_NOTICE, "[update_sdp_for_local] Set caller local_sdp main video for RTSP.\n");
				update_sdp_for_callee_answer_rtsp(caller_session->inv_session->pool_prov,caller_session,sdp_media,peer);
			}
			/*End added by zyxia for Bug 205228,2021.10.9 */
			else {
				if (caller_is_gswave) {
					disable_gswave_media(caller_session->inv_session->pool_prov, caller_session, sdp_media);
				} else {
					sdp_media->desc.port = 0;
				}
				ast_log(LOG_NOTICE, "[update_sdp_for_local] Set caller local_sdp main video port 0.\n");
			}
			break;
		}
	}
}
/*Start added by zyxia for Bug 223328 - Call Flip_Edge browser flip call was hung up,2022.5.12*/
static void update_auto_record_sdp(struct ast_channel* chan)
{
	if (!chan || strcasecmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_NOTICE, "[update_auto_record_sdp] Invalid args.\n");
		return;
	}

	struct ast_sip_session* caller_session = get_sip_session_from_channel(chan);
	if (!caller_session) {
		ast_log(LOG_NOTICE, "[update_auto_record_sdp] The caller_session is nullptr.\n");
		return;
	}

	struct pjmedia_sdp_session* caller_local_sdp_sess = NULL;
	caller_local_sdp_sess = caller_session->inv_session->neg->neg_local_sdp;
	if (!caller_local_sdp_sess) {
		ast_log(LOG_NOTICE, "[update_auto_record_sdp] The update_auto_record_sdp is nullptr.\n");
		return;
	}

	struct ast_stream_topology *topology = NULL;
	if (caller_session->pending_media_state && caller_session->pending_media_state->topology) {
		topology = caller_session->pending_media_state->topology;
	} else if (caller_session->active_media_state && caller_session->active_media_state->topology) {
		topology = caller_session->active_media_state->topology;
	}
	if (!topology) {
		ast_log(LOG_NOTICE, "[update_auto_record_sdp] The caller's media topology is nullptr.\n");
		return;
	}

	struct pjmedia_sdp_media* sdp_media = NULL;

	for (int i = 0; i < caller_local_sdp_sess->media_count; ++ i) {
		sdp_media = caller_local_sdp_sess->media[i];
		struct ast_stream* stream = ast_stream_topology_get_stream(topology, i);
		if (!sdp_media || !stream) {
			continue;
		}

		if (!strncasecmp(sdp_media->desc.media.ptr, "audio", 5) && ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_AUDIO_TYPE) {
			change_audio_record_attr(caller_session, sdp_media, caller_session->inv_session->pool_prov);
			break;
		}
	}
}
/*End added by zyxia for Bug 223328 - Call Flip_Edge browser flip call was hung up,2022.5.12*/

static void chan_pjsip_special_answer_for_peer_not_pjsip(struct ast_channel *chan, struct ast_channel *peer)
{
	if (strcasestr(ast_channel_tech(peer)->type, "DAHDI")) {
		ast_log(LOG_DEBUG, "It's peer channel is DAHDI type,disable media except audio\n");
		disable_media_except_audio(chan);
	} else {

		if (!strncasecmp(ast_channel_tech(peer)->type, "IAX2", 4) &&
			!strncasecmp(ast_channel_tech(chan)->type, "PJSIP", 5)  ) {
			ast_log(LOG_NOTICE, "peer channel name:%s, chan channel name:%s\n", ast_channel_name(peer), ast_channel_name(chan));

			if (ast_channel_tech(peer) &&
				ast_channel_tech(peer)->support_video &&
				!ast_channel_tech(peer)->support_video(peer)) {
				pbx_builtin_setvar_helper(chan, "HAS_NO_VIDEO_CAPS", "yes");
				disable_main_video(chan,FALSE);
			}
		}

		/*Start added  by zyxia for Bug 207787 ,2021.11.16*/
		if (!strncasecmp(ast_channel_tech(peer)->type, "RTSP", 4)){ //is RTSP channel
			struct ast_sip_channel_pvt* chan_pvt = ast_channel_tech_pvt(chan);
			if (chan_pvt && chan_pvt->session) {
				chan_pvt->session->is_rtsp_call = 1;
				/*Start added by zyxia for Bug 241677,2022.11.3 */
				const char* video_rtsp_level_id = pbx_builtin_getvar_helper(peer, "RTSP_PROFILE_LEVEL_ID");
				if (video_rtsp_level_id ) {
					memset(chan_pvt->session->rtsp_profile_level_id,0,sizeof(chan_pvt->session->rtsp_profile_level_id));
					strncpy(chan_pvt->session->rtsp_profile_level_id,video_rtsp_level_id, sizeof(chan_pvt->session->rtsp_profile_level_id)-1);
				}
				/*End added by zyxia for Bug 241677,2022.11.3 */

				ast_log(LOG_WARNING, "[session %p] is  a rtsp call.the caller name is %s\n" ,chan_pvt->session,ast_channel_name(chan));
			}
			disable_application_bfcp(chan);//added by zyxia for Bug 227857
		} else if (strncasecmp(ast_channel_tech(peer)->type, "Local", 5)) {
			disable_external_media(chan);
		}

		/* In this, the peer channel is a local channel, We need modify (chan SDP) according to (peer channel var). */
		update_sdp_for_local(chan, peer);
	}

	return;
}

static void ast_session_ignore_direct_media_attr(struct ast_channel* chan)
{
	if (!chan) {
		ast_log(LOG_WARNING, "[ast_session_ignore_direct_media_attr] the channel is nullptr.\n");
		return;
	}

	if (strcasecmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "[ast_session_ignore_direct_media_attr] channel:%s type isn't PJSIP\n", ast_channel_name(chan));
		return;
	}

	struct ast_sip_channel_pvt* chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt) {
		ast_log(LOG_WARNING, "[ast_session_ignore_direct_media_attr] the channel pvt is nullptr.\n");
		return;
	}

	if (chan_pvt->session) {
		chan_pvt->session->is_direct_media = 0;
		return;
	}

	ast_log(LOG_WARNING, "[ast_session_ignore_direct_media_attr] the channel session is nullptr.\n");
}

static int check_channel_is_video_call(struct ast_channel *chan)
{
	int i;
	int active_stream_count = 0;
	int pend_stream_count = 0;
	struct ast_sip_channel_pvt *chan_pvt = ast_channel_tech_pvt(chan);
	struct ast_stream *active_stream;
	struct ast_stream *pend_stream;
	struct ast_sip_session *session;

	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	session = chan_pvt->session;
	ast_channel_lock(session->channel);
	if (session->active_media_state && session->active_media_state->topology) {
		active_stream_count = ast_stream_topology_get_count(session->active_media_state->topology);
		for(i=0; i<active_stream_count; i++) {
			active_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)active_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(chan), i, ast_stream_get_type(active_stream),
				ast_stream_get_sub_type(active_stream), ast_stream_get_state(active_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(active_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(active_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(active_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(active_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}

	if (session->pending_media_state && session->pending_media_state->topology) {
		pend_stream_count = ast_stream_topology_get_count(session->pending_media_state->topology);
		for(i=0; i<pend_stream_count; i++) {
			pend_stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)pend_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(chan), i, ast_stream_get_type(pend_stream),
				ast_stream_get_sub_type(pend_stream), ast_stream_get_state(pend_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(pend_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(pend_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(pend_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(pend_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}
	ast_channel_unlock(session->channel);
	return 0;
}

static int check_is_video_call(struct ast_channel *chan, struct ast_channel *peer)
{
	int i;
	int active_stream_count = 0;
	int pend_stream_count = 0;
	struct ast_sip_channel_pvt *chan_pvt = ast_channel_tech_pvt(chan);
	struct ast_sip_channel_pvt *peer_pvt = ast_channel_tech_pvt(peer);
	struct ast_stream *active_stream;
	struct ast_stream *pend_stream;
	struct ast_sip_session *session;

	if (!chan_pvt || !chan_pvt->session || !peer_pvt || !peer_pvt->session) {
		return 0;
	}

	session = chan_pvt->session;
	ast_channel_lock(session->channel);
	if (session->active_media_state && session->active_media_state->topology) {
		active_stream_count = ast_stream_topology_get_count(session->active_media_state->topology);
		for(i=0; i<active_stream_count; i++) {
			active_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)active_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(chan), i, ast_stream_get_type(active_stream),
				ast_stream_get_sub_type(active_stream), ast_stream_get_state(active_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(active_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(active_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(active_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(active_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}

	if (session->pending_media_state && session->pending_media_state->topology) {
		pend_stream_count = ast_stream_topology_get_count(session->pending_media_state->topology);
		for(i=0; i<pend_stream_count; i++) {
			pend_stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)pend_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(chan), i, ast_stream_get_type(pend_stream),
				ast_stream_get_sub_type(pend_stream), ast_stream_get_state(pend_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(pend_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(pend_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(pend_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(pend_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}
	ast_channel_unlock(session->channel);

	session = peer_pvt->session;
	ast_channel_lock(session->channel);
	if (session->active_media_state && session->active_media_state->topology) {
		active_stream_count = ast_stream_topology_get_count(session->active_media_state->topology);
		for(i=0; i<active_stream_count; i++) {
			active_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)active_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(peer), i, ast_stream_get_type(active_stream),
				ast_stream_get_sub_type(active_stream), ast_stream_get_state(active_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(active_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(active_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(active_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(active_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}

	if (session->pending_media_state && session->pending_media_state->topology) {
		pend_stream_count = ast_stream_topology_get_count(session->pending_media_state->topology);
		for(i=0; i<pend_stream_count; i++) {
			pend_stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)pend_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(peer), i, ast_stream_get_type(pend_stream),
				ast_stream_get_sub_type(pend_stream), ast_stream_get_state(pend_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(pend_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(pend_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(pend_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(pend_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}
	ast_channel_unlock(session->channel);

	return 0;
}

static void get_peer_trunk_info(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	if (!session || !peer_session) {
		return;
	}

	if (!session->conf_link && peer_session->conf_link) {
		session->conf_link = ast_calloc(1, strlen(peer_session->conf_link) + 1);
		if (session->conf_link) {
			strcpy(session->conf_link, peer_session->conf_link);
		}
	}

	if (!session->trunk_channel_id && peer_session->trunk_channel_id) {
		session->trunk_channel_id = ast_calloc(1, strlen(peer_session->trunk_channel_id) + 1);
		if (session->trunk_channel_id) {
			strcpy(session->trunk_channel_id, peer_session->trunk_channel_id);
		}
	}

	if (!session->trunk_conf_num && peer_session->trunk_conf_num) {
		session->trunk_conf_num = ast_calloc(1, strlen(peer_session->trunk_conf_num) + 1);
		if (session->trunk_conf_num) {
			strcpy(session->trunk_conf_num, peer_session->trunk_conf_num);
		}
	}

	return;
}

static int special_answer(void *data)
{
	struct ast_channel *chan = NULL;
	struct ast_channel *peer = NULL;
	struct ast_sip_special_answer_data *answer_data = data;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_channel_pvt *peer_pvt = NULL;
	pjmedia_sdp_neg_state chan_neg_state = PJMEDIA_SDP_NEG_STATE_NULL;
	pj_pool_t *pool = NULL;
	int release_chan = 0;
	int i = 0;

	chan = answer_data->chan;
	peer = answer_data->peer;

	ast_channel_lock(chan);
	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(chan), AST_FLAG_ZOMBIE) || ast_check_hangup(chan)) {
		ast_channel_unlock(chan);
		return -1;
	}

	const char* from_call_flip = pbx_builtin_getvar_helper(chan, "FROM_CALL_FLIP");
	if (!ast_strlen_zero(from_call_flip) && strcasecmp(from_call_flip, "yes") == 0) {
		ast_channel_set_is_callflip(chan, 1);
	}

	if (ast_channel_get_peer_is_sca(chan) && !peer) {
		char peer_sca_chan_name[80];
		memset(peer_sca_chan_name, 0, sizeof(peer_sca_chan_name));
		ast_channel_get_peer_sca_name(chan, peer_sca_chan_name);

		ast_channel_unlock(chan);
		peer = ast_channel_get_by_name(peer_sca_chan_name);
		if (peer) {
			release_chan = 1;
		}
		ast_channel_lock(chan);
	}

	/* Add by bfdu for Bug 303077, 2024.6.19.
	 * Another thread do ast_channel_move will change chan to local type, just return if not pjsip type.
	 */
	if (strcasecmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_ERROR, "special_answer chan type is not PJSIP\n");
		ast_channel_unlock(chan);
		return -1;
	}

	if (!peer) {
		chan_pjsip_special_answer_for_no_peer(chan);
		ast_channel_unlock(chan);
		return 1;
	}

	if (0 != strcmp(ast_channel_tech(peer)->type, "PJSIP")) {
		/* DAHDI call not support direct media option. */
		ast_session_ignore_direct_media_attr(chan);
		//ast_session_ignore_direct_media_attr(peer);

		chan_pjsip_special_answer_for_peer_not_pjsip(chan, peer);

		ast_channel_unlock(chan);
		if (release_chan) {
			ast_channel_unref(peer);
		}

		return 1;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	peer_pvt = ast_channel_tech_pvt(peer);

	if (!peer_pvt || !peer_pvt->session || !peer_pvt->session->inv_session) {
		ast_log(LOG_WARNING, "special answer peer_pvt session is invalid\n");
		ast_channel_unlock(chan);
		if (release_chan) {
			ast_channel_unref(peer);
		}

		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] [peer_session:%p] special_answer begin for caller endpoint:%s callee endpoint:%s\n",
		chan_pvt->session, peer_pvt->session, ast_sorcery_object_get_id(chan_pvt->session->endpoint), ast_sorcery_object_get_id(peer_pvt->session->endpoint));

	/* Added by yfyang, for Bug 107211 */
	if (strncmp(ast_channel_name(peer), "PJSIP/trunk_", 12)) {
		const char *callee_dial_str = pbx_builtin_getvar_helper(peer, "CALLEE_DIAL_STRING");
		if (!ast_strlen_zero(callee_dial_str)) {
			pbx_builtin_setvar_helper(chan, "CALLEE_DIAL_STRING", callee_dial_str);
		}
	}

	/* Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
	if (!strncmp(ast_channel_name(peer), "PJSIP/trunk_", 12) && peer_pvt) {
		struct chan_pjsip_pvt *pvt = peer_pvt->pvt;
		if(pvt && pvt->call_type && !strcasecmp(pvt->call_type,"CQ")) {//edit by zyxia for Bug263552,2023.6.27
			sip_setcalltype(chan, pvt->call_type);
			ast_channel_set_is_queue_call(chan, 1);
		}
	}
	/* End by zyxia for x-gs-state Bug 262507 , 2023.6.14 */

	/* Added by yfyang, for IPVT Server application */
	const char *client_info_msg = pbx_builtin_getvar_helper(peer, "IPVT_CLIENT_INFO_MSG");
	if (!ast_strlen_zero(client_info_msg)) {
		const char *client_info = pbx_builtin_getvar_helper(peer, "IPVT_CLIENT_INFO");
		if (!ast_strlen_zero(client_info)) {
			pbx_builtin_setvar_helper(chan, "IPVT_CLIENT_INFO_MSG", client_info_msg);
			pbx_builtin_setvar_helper(chan, "IPVT_CLIENT_INFO", client_info);
		}
	}

	/* Added by yfyang, for Bug 160811 */
	const char *callinfo_msg = pbx_builtin_getvar_helper(peer, "CALL_INFO_MSG");
	if (!ast_strlen_zero(callinfo_msg) && !strncmp(callinfo_msg, "response", 8)) {
		const char *call_info = pbx_builtin_getvar_helper(peer, "CALL_INFO");
		if (!ast_strlen_zero(call_info)) {
			pbx_builtin_setvar_helper(chan, "CALL_INFO_MSG", callinfo_msg);
			pbx_builtin_setvar_helper(chan, "CALL_INFO", call_info);
		}
	}

	while (PJ_SUCCESS != pjsip_dlg_try_inc_lock(peer_pvt->session->inv_session->dlg)) {
		if (i < 3) {
			usleep(20000);
		} else {
			chan_pjsip_special_answer_for_no_peer(chan);
			ast_channel_unlock(chan);
			if (release_chan) {
				ast_channel_unref(peer);
			}
			ast_log(LOG_DEBUG, "[session:%p] special_answer dlg try lock failed %d times, don't use peer to answer, endpoint:%s\n",
				chan_pvt->session, i, ast_sorcery_object_get_id(chan_pvt->session->endpoint));
			return 1;
		}
		i++;
	}

	if (check_special_answer_param(chan_pvt, peer_pvt)) {
		ast_log(LOG_WARNING, "[session:%p] check_special_answer_param failed for endpoint:%s\n",
			chan_pvt->session, ast_sorcery_object_get_id(chan_pvt->session->endpoint));
		pjsip_dlg_dec_lock(peer_pvt->session->inv_session->dlg);
		ast_channel_unlock(chan);
		if (release_chan) {
			ast_channel_unref(peer);
		}

		return -1;
	}

	pool = chan_pvt->session->inv_session->pool_prov;

	chan_neg_state = chan_pvt->session->inv_session->neg->state;
	switch (chan_neg_state) {
		case PJMEDIA_SDP_NEG_STATE_WAIT_NEGO:
		case PJMEDIA_SDP_NEG_STATE_LOCAL_OFFER:
			modify_caller_sdp_for_wait_state(pool, chan_pvt->session, peer_pvt->session);
			break;
		case PJMEDIA_SDP_NEG_STATE_DONE:
			modify_caller_sdp_for_done_state(pool, chan_pvt->session, peer_pvt->session);
			break;
		default:
			ast_log(LOG_WARNING, "[session:%p] Can't process neg state:%d for endpoint:%s\n",
				chan_pvt->session, chan_neg_state, ast_sorcery_object_get_id(chan_pvt->session->endpoint));
			break;
	}

	get_peer_trunk_info(chan_pvt->session, peer_pvt->session);

	strcpy(chan_pvt->session->callee_gs_client_info, peer_pvt->session->callee_gs_client_info);

	pjsip_dlg_dec_lock(peer_pvt->session->inv_session->dlg);

#ifdef GRANDSTREAM_NETWORKS
	ast_channel_unlock(chan);
	int retval = check_is_video_call(chan, peer);
	ast_channel_lock(chan);
	if (retval) {
		ast_channel_set_is_videocall(chan, TRUE);
		ast_channel_set_is_videocall(peer, TRUE);
		/*start added by zyxia for Bug 212610 ,2022.1.11*/
		ast_sip_callflip_update_user_media_type(chan_pvt->session->endpoint->aors,"video");
		ast_sip_callflip_update_user_media_type(peer_pvt->session->endpoint->aors,"video");
		/*End added by zyxia for Bug 212610 ,2022.1.11*/
	} else {
		ast_channel_set_is_videocall(chan, FALSE);
		ast_channel_set_is_videocall(peer, FALSE);
		/*start added by zyxia for Bug 212610 ,2022.1.11*/
		ast_sip_callflip_update_user_media_type(chan_pvt->session->endpoint->aors,"audio");
		ast_sip_callflip_update_user_media_type(peer_pvt->session->endpoint->aors,"audio");
		/*end added by zyxia for Bug 212610 ,2022.1.11*/
	}

	/*start added by zyxia for Bug 212610 ,2022.1.11*/
	if (ast_sip_get_callflip_enable()) {
		struct ast_sip_callflip_state callflip_info;
		memset(&callflip_info, 0, sizeof(struct ast_sip_callflip_state));
		struct callflip_user *callflip_state,*peer_callflip_state;

		if ((callflip_state = ast_sip_callflip_find_user(chan_pvt->session->endpoint->aors))) {
			if (!strcasecmp(callflip_state->enable,"yes")) {
				ast_log(LOG_DEBUG, "Chan:%s, set ast_sip_callflip_find_user:%d",ast_channel_name(chan), retval);

				snprintf(callflip_info.enable, sizeof(callflip_info.enable), "%s", callflip_state->enable);
				snprintf(callflip_info.remote_id, sizeof(callflip_info.remote_id), "%s", callflip_state->remote_id);
				snprintf(callflip_info.remote_name, sizeof(callflip_info.remote_name), "%s",callflip_state->remote_name);
				snprintf(callflip_info.call_type, sizeof(callflip_info.call_type), "%s", callflip_state->call_type);
				snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", callflip_state->media_type);
				snprintf(callflip_info.channel_id, sizeof(callflip_info.channel_id), "%s", callflip_state->channel_id);
				snprintf(callflip_info.replace_channel_id, sizeof(callflip_info.replace_channel_id), "%s", callflip_state->replace_channel_id);
				snprintf(callflip_info.conf_link, sizeof(callflip_info.conf_link), "%s", callflip_state->conf_link);

				ast_channel_unlock(chan);
				ast_sip_for_each_aor(chan_pvt->session->endpoint->aors, process_callflip_aor, &callflip_info);
				ast_channel_lock(chan);
			}
		}

		if ((peer_callflip_state = ast_sip_callflip_find_user(peer_pvt->session->endpoint->aors))) {
			if (!strcasecmp(peer_callflip_state->enable,"yes")) {
				memset(&callflip_info, 0, sizeof(struct ast_sip_callflip_state));
				//ast_log(LOG_DEBUG, "Chan:%s, set ast_sip_callflip_find_user:%d",ast_channel_name(chan), retval);
				ast_log(LOG_DEBUG, "peerChan:%s, set ast_sip_callflip_find_user:%d",ast_channel_name(peer), retval);
				snprintf(callflip_info.enable, sizeof(callflip_info.enable), "%s", peer_callflip_state->enable);
				snprintf(callflip_info.remote_id, sizeof(callflip_info.remote_id), "%s", peer_callflip_state->remote_id);
				snprintf(callflip_info.remote_name, sizeof(callflip_info.remote_name), "%s",peer_callflip_state->remote_name);
				snprintf(callflip_info.call_type, sizeof(callflip_info.call_type), "%s", peer_callflip_state->call_type);
				snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", peer_callflip_state->media_type);
				snprintf(callflip_info.channel_id, sizeof(callflip_info.channel_id), "%s", peer_callflip_state->channel_id);
				snprintf(callflip_info.replace_channel_id, sizeof(callflip_info.replace_channel_id), "%s", peer_callflip_state->replace_channel_id);
				snprintf(callflip_info.conf_link, sizeof(callflip_info.conf_link), "%s", peer_callflip_state->conf_link);
				ast_channel_unlock(chan);
				ast_sip_for_each_aor(peer_pvt->session->endpoint->aors, process_callflip_aor, &callflip_info);
				ast_channel_lock(chan);
			}
		}
	}

	/*End added by zyxia for Bug 212610 ,2022.1.11*/

	if (chan_pvt && peer_pvt) {
		set_local_gds_call_from_peer(peer_pvt->session, chan_pvt->session, 0);

		ast_log(LOG_NOTICE, "channel1 name:%s gds_call:%d, channel2 name:%s gds_call:%d\n",
			ast_channel_name(peer_pvt->session->channel), peer_pvt->session->is_gds_call,
			ast_channel_name(chan_pvt->session->channel), chan_pvt->session->is_gds_call);

		set_local_gsc_call_from_peer(peer_pvt->session, chan_pvt->session, 0);

		ast_log(LOG_NOTICE, "channel1 name:%s gsc_call:%d, channel2 name:%s gsc_call:%d\n",
			ast_channel_name(peer_pvt->session->channel), peer_pvt->session->is_gsc_call,
			ast_channel_name(chan_pvt->session->channel), chan_pvt->session->is_gsc_call);
	}

	if (chan && !strcmp(ast_channel_tech(chan)->type, "PJSIP") && chan_pvt && peer_pvt) {
		if (chan_pvt->session->is_gds_call || peer_pvt->session->is_gds_call) {
			ast_channel_tech(chan)->indicate(chan, AST_CONTROL_NOTIFY_GDS_CALL, NULL, 0);
		}

		if (chan_pvt->session->is_gsc_call || peer_pvt->session->is_gsc_call) {
			ast_channel_tech(chan)->indicate(chan, AST_CONTROL_NOTIFY_GSC_CALL, NULL, 0);
		}
	}
#endif

	ast_channel_unlock(chan);

	if (release_chan) {
		ast_channel_unref(peer);
	}

	return 1;
}

static int chan_pjsip_get_direct_enable(struct ast_channel* chan)
{
	if (!chan) {
		ast_log(LOG_ERROR, "[chan_pjsip_get_direct_enable] Invalid channel\n");
		return 0;
	}

	if(!strstr(ast_channel_name(chan), "PJSIP/")){
		ast_log(LOG_ERROR, "[chan_pjsip_get_direct_enable] Invalid channel\n");
		return 0;
	}

	struct ast_sip_channel_pvt* chan_pvt = ast_channel_tech_pvt(chan);
	if (chan_pvt && chan_pvt->session) {
		return chan_pvt->session->is_direct_media;
	}

	return 0;
}

static int chan_pjsip_special_answer(struct ast_channel *chan, struct ast_channel *peer)
{
	int ret = 0;
	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	struct ast_sip_session *session = NULL;
	struct ast_sip_special_answer_data answer_data;
	char  family[80] = {0};
	char  data[80]	 = {0};
	int high_freq_count = 0;
	int q_index;
	struct timeval now;

	answer_data.chan = chan;
	answer_data.peer = peer;

	session = ao2_bump(channel->session);

	ast_log(LOG_DEBUG, "[session:%p] chan_pjsip_special_answer begin for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_unlock(chan);
	if ((ret = ast_sip_push_task_synchronous(session->serializer, special_answer, &answer_data)) <= 0) {
		ao2_ref(session, -1);
		ast_channel_lock(chan);
		return ret;
	}
	ao2_ref(session, -1);
	ast_channel_lock(chan);
	ast_log(LOG_DEBUG, "[session:%p] chan_pjsip_special_answer end for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if( peer && (strstr(ast_channel_name(peer),"DAHDI/o") || strstr(ast_channel_name(peer),"PJSIP/trunk"))
		&& NULL == strstr(ast_channel_name(chan),"PJSIP/trunk"))
	{
		snprintf(family, sizeof(family), "AAUSER/%s", session->endpoint->sca_sharedline);
		ast_ucm_db_get(family, "HIGHCALL_THRESHOLD", data, sizeof(data));
	}
	else if(peer && strstr(ast_channel_name(peer),"Local/") )
	{
		/*  cfu forward on phone side */
		char local_2[64] = {0};
		struct ast_channel *local2;
		const char *trunk_index;
		strcpy(local_2, ast_channel_name(peer));
		local_2[strlen(ast_channel_name(peer)) - 1] = '2';

		ast_channel_unlock(chan);
		local2 = ast_channel_get_by_name(local_2);
		ast_channel_lock(chan);
		if(local2)
		{
			trunk_index = pbx_builtin_getvar_helper(local2, "OUTBOUND_TRUNK_INDEX");
			if(trunk_index && 0 != strlen(trunk_index))
			{
				snprintf(family, sizeof(family), "AAUSER/%s", session->endpoint->sca_sharedline);
				ast_ucm_db_get(family, "HIGHCALL_THRESHOLD", data, sizeof(data));
			}
			ast_channel_unref(local2);
		}
	}

	if(data[0] != 0 && atoi(data) != 0)
	{
		high_freq_count = atoi(data);
		q_index = ast_exten_map2queue_index((char *)session->endpoint->sca_sharedline);

		gettimeofday(&now, NULL);

		if(-1 != q_index )
		{
			ast_log(LOG_DEBUG,"chan:%s insert a observe node, time=%ld, thres=%d", ast_channel_name(chan),now.tv_sec,high_freq_count);
			ast_high_frequence_enqueue(ast_get_high_frequence_queue(q_index), (char *)session->endpoint->sca_sharedline,high_freq_count, &now);
		}
	}
	return chan_pjsip_answer(chan);
}

static unsigned char judge_if_removed_stream_in_active_media(struct ast_sip_session *session, struct ast_stream *stream)
{
	unsigned int i = 0;
	struct ast_stream_topology *active_topology = NULL;
	struct ast_stream *active_stream = NULL;

	if (!session || !stream) {
		return 0;
	}

	if (!session->active_media_state) {
		return 0;
	}
	active_topology = session->active_media_state->topology;
	if (!active_topology) {
		return 0;
	}

	if (ast_stream_topology_get_count(active_topology) > 0) {
		for (i = 0; i < ast_stream_topology_get_count(active_topology); ++i) {
			active_stream = ast_stream_topology_get_stream(active_topology, i);
			if (active_stream && ast_stream_get_sub_type(active_stream) == ast_stream_get_sub_type(stream)) {
				ast_log(LOG_DEBUG, "Find the remove stream type(%d) is active media topology\n", ast_stream_get_sub_type(stream));
				return 1;
			}
		}
	}

	return 0;
}

static unsigned char judge_if_has_slides_video(const struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state) {
		return 0;
	}

	topology = session->pending_media_state->topology;
	if (!topology) {
		/* Need to find active media state, maybe we already have early media */
		if (!session->active_media_state) {
			return 0;
		}
		topology = session->active_media_state->topology;
		if (!topology) {
			ast_log(LOG_WARNING, "We can't get any pending or active media state\n");
			return 0;
		}
	}

	if (ast_stream_topology_get_count(topology) == 0) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (stream && ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			ast_log(LOG_DEBUG, "Find slides video stream type in topology\n");
			return 1;
		}
	}

	return 0;
}

static unsigned char judge_if_pending_topology_has_bfcp_stream(const struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state) {
		return 0;
	}

	topology = session->pending_media_state->topology;
	if (!topology) {
		return 0;
	}

	if (ast_stream_topology_get_count(topology) == 0) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (stream &&
			(ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
			ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP)) {
			ast_log(LOG_DEBUG, "Find bfcp stream type in topology\n");
			return 1;
		}
	}

	return 0;
}

static unsigned char judge_if_has_bfcp_stream(const struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state) {
		return 0;
	}

	topology = session->pending_media_state->topology;
	if (!topology) {
		if (!session->active_media_state) {
			return 0;
		}
		topology = session->active_media_state->topology;
		if (!topology) {
			ast_log(LOG_WARNING, "We can't get any pending or active media state\n");
			return 0;
		}
	}

	if (ast_stream_topology_get_count(topology) == 0) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (stream &&
			(ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
			ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP)) {
			ast_log(LOG_DEBUG, "Find bfcp stream type in topology\n");
			return 1;
		}
	}

	return 0;
}

static int find_topology_audio_stream(struct ast_stream_topology* topology)
{
	int i = 0;
	int ret = 0;

	for (i = 0; i < ast_stream_topology_get_count(topology); ++ i) {
		struct ast_stream* tmp = ast_stream_topology_get_stream(topology, i);

		if (ast_stream_get_type(tmp) == AST_MEDIA_TYPE_AUDIO) {
			ret = 1;
			break;
		}
	}

	return ret;
}

static struct ast_stream *get_audio_stream(struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	struct ast_stream* stream= NULL;

	if (!topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_AUDIO) {
			break;
		}
	}

	return stream;
}

static void copy_audio_stream_direction_from_peer(struct ast_stream* stream, struct ast_stream_topology *peer_topology)
{
	struct ast_stream *peer_stream = NULL;

	if (!stream || !peer_topology) {
		return;
	}

	peer_stream = get_audio_stream(peer_topology);
	if (!peer_stream) {
		return;
	}

	/* copy media direction value directly, will change in funtion:get_peer_media_direction */
	ast_stream_set_local_media_direction(stream, ast_stream_get_local_media_direction(peer_stream));
	ast_stream_set_remote_media_direction(stream, ast_stream_get_remote_media_direction(peer_stream));

	return;
}

static void add_reinvite_default_audio_stream(struct ast_stream_topology* pending_topology,
	struct ast_stream_topology* endpoint_topology, struct ast_stream_topology *peer_topology)
{
	if (pending_topology == NULL || endpoint_topology == NULL) {
		ast_log(LOG_DEBUG, "add_default_audio_stream met error: invalid args\n");
		return;
	}

	int ret = find_topology_audio_stream(pending_topology);
	if (ret == 1) {
		ast_log(LOG_NOTICE, "pending_topology is already had audio stream\n");
		return;
	}

	int i = 0;

	struct ast_stream* tmp_audio_stream = NULL;
	for (i = 0; i < ast_stream_topology_get_count(endpoint_topology); ++ i) {
		struct ast_stream* tmp = ast_stream_topology_get_stream(endpoint_topology, i);
		if (ast_stream_get_type(tmp) == AST_MEDIA_TYPE_AUDIO) {
			tmp_audio_stream = ast_stream_clone(tmp, NULL);
			ast_stream_set_type(tmp_audio_stream, AST_MEDIA_TYPE_AUDIO);
			ast_stream_set_sub_type(tmp_audio_stream, AST_MEDIA_SUB_AUDIO_TYPE);
			copy_audio_stream_direction_from_peer(tmp_audio_stream, peer_topology);
			if (ast_format_cap_has_audio_fec(ast_stream_get_formats(tmp_audio_stream))) {
				ast_sip_session_change_gs_fec_version(ast_stream_get_formats(tmp_audio_stream));
			}
			break;
		}
	}

	int old_length = ast_stream_topology_get_count(pending_topology);
	struct ast_stream* stream_array[10];
	ast_stream_topology_append_stream(pending_topology, tmp_audio_stream);

	for (i = 0; i < old_length; ++ i) {
		struct ast_stream* tmp = ast_stream_clone(ast_stream_topology_get_stream(pending_topology, 0), NULL);
		stream_array[i] = tmp;
		ast_stream_topology_del_stream(pending_topology, 0);
	}

	for (i = 0; i < old_length; ++ i) {
		ast_stream_topology_append_stream(pending_topology, stream_array[i]);
	}
}

static void create_default_topology(struct ast_sip_session *session, struct ast_stream_topology *peer_topology)
{
	unsigned int i = 0;
	struct ast_stream* stream= NULL;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (session->pending_media_state->topology) {
		return;
	}

	ast_log(LOG_DEBUG, "create_default_topology begin\n");

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate capabilities for handle no pending media\n");
		return;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
	if (ast_format_cap_has_audio_fec(caps)) {
		ast_sip_session_change_gs_fec_version(caps);
	}
	session->pending_media_state->topology = ast_stream_topology_create_from_format_cap(caps);

	if (!session->pending_media_state->topology) {
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); ++ i) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_AUDIO) {
			ast_stream_set_sub_type(stream, AST_MEDIA_SUB_AUDIO_TYPE);
			ast_sip_session_process_special_audio_codec(session, stream, ast_stream_get_formats(stream));
			copy_audio_stream_direction_from_peer(stream, peer_topology);
			break;
		}
	}

	return;
}

static struct ast_stream *get_active_stream_according_peer_stream(struct ast_sip_session *session, struct ast_stream *stream)
{
	unsigned int i = 0;
	struct ast_stream_topology *active_topology = NULL;
	struct ast_stream *active_stream = NULL;

	if (!session || !stream) {
		return 0;
	}

	if (!session->active_media_state) {
		return 0;
	}
	active_topology = session->active_media_state->topology;
	if (!active_topology) {
		return 0;
	}

	if (ast_stream_topology_get_count(active_topology) > 0) {
		for (i = 0; i < ast_stream_topology_get_count(active_topology); ++i) {
			active_stream = ast_stream_topology_get_stream(active_topology, i);
			if (active_stream && ast_stream_get_sub_type(active_stream) == ast_stream_get_sub_type(stream)) {
				ast_log(LOG_DEBUG, "Find according stream type(%d) is active media topology\n", ast_stream_get_sub_type(stream));
				return active_stream;
			}
		}
	}

	return 0;
}

static void append_disable_stream(struct ast_sip_session *session, struct ast_stream *stream)
{
	struct ast_stream *clone_stream = NULL;
	struct ast_stream *active_stream = NULL;

	if (session->terminal_type ==  AST_TERMINAL_TYPE_WEBRTC &&
		ast_sip_session_media_use_bundle(session)) {
		active_stream = get_active_stream_according_peer_stream(session, stream);
		clone_stream = ast_stream_clone(active_stream, NULL);
		if (!clone_stream) {
			ast_log(LOG_WARNING, "[session:%p] Can't clone the stream for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		ast_stream_set_state(clone_stream, AST_STREAM_STATE_INACTIVE);
		ast_stream_set_local_media_direction(clone_stream, INACTIVE);
		ast_stream_set_remote_media_direction(clone_stream, INACTIVE);
	} else {
		clone_stream = ast_stream_clone(stream, NULL);
		if (!clone_stream) {
			ast_log(LOG_WARNING, "[session:%p] Can't clone the stream for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		ast_stream_set_state(clone_stream, AST_STREAM_STATE_REMOVED);
	}

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_alloc();
		if (!session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] Can't alloc topology for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
	}

	if (ast_stream_topology_append_stream(session->pending_media_state->topology, clone_stream) < 0) {
		ast_log(LOG_WARNING, "[session:%p] Append stream to topology failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_free(clone_stream);
		return;
	}

	return;
}

static int process_disabled_media(struct ast_sip_session *session,
	struct ast_sip_session *peer_session, struct ast_stream *req_stream)
{
	if (ast_stream_get_state(req_stream) == AST_STREAM_STATE_REMOVED &&
		judge_if_removed_stream_in_active_media(session, req_stream)) {
		append_disable_stream(session, req_stream);
		return 1;
	}

	if (ast_stream_get_remote_media_direction(req_stream) == INACTIVE &&
		(ast_stream_get_type(req_stream) != AST_MEDIA_TYPE_AUDIO) &&
		(peer_session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) &&
		(session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) &&
		judge_if_removed_stream_in_active_media(session, req_stream)) {
		append_disable_stream(session, req_stream);
		return 1;
	}

	return 0;
}

static int process_common_stream(struct ast_sip_session *session, struct ast_stream *req_stream)
{
	struct ast_format_cap *req_cap = NULL;
	RAII_VAR(struct ast_format_cap *, joint_cap, NULL, ao2_cleanup);
	struct ast_stream *clone_stream = NULL;
	struct ast_str *format_buf = ast_str_alloca(128);

	req_cap = ast_stream_get_formats(req_stream);
	if (!req_cap) {
		ast_log(LOG_WARNING, "req stream has no cap\n");
		return -1;
	}

	joint_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!joint_cap) {
		return -1;
	}

	if (ast_sip_session_get_joint_caps(session, req_stream, joint_cap, req_cap) &&
		(ast_stream_get_type(req_stream) != AST_MEDIA_TYPE_AUDIO)) {
		return -1;
	}

	if (ast_sip_session_process_special_video_codec(session, joint_cap)) {
		return -1;
	}

	clone_stream = ast_stream_clone(req_stream, NULL);
	if (!clone_stream) {
		return -1;
	}

	if (ast_stream_get_type(req_stream) == AST_MEDIA_TYPE_AUDIO) {
		ast_format_cap_append_from_cap(joint_cap, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
		ast_sip_session_process_special_audio_codec(session, req_stream, joint_cap);
	}

	if (ast_stream_get_type(req_stream) == AST_MEDIA_TYPE_VIDEO) {
		ast_sip_session_change_video_independent_codec(session, joint_cap, clone_stream);
	}

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, outgoing peer: %s\n",
		session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(req_cap, &format_buf));
	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, outgoing caps: %s\n",
		session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(session->endpoint->media.codecs, &format_buf));
	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, outgoing joint_cap: %s\n",
		session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(joint_cap, &format_buf));

	ast_stream_set_formats(clone_stream, joint_cap);

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_alloc();
		if (!session->pending_media_state->topology) {
			return -1;
		}
	}

	ast_log(LOG_DEBUG, "process_common_stream, append stream type:%d\n", ast_stream_get_type(clone_stream));

	if (ast_stream_topology_append_stream(session->pending_media_state->topology, clone_stream) < 0) {
		ast_stream_free(clone_stream);
		return -1;
	}

	return 0;
}

static unsigned int judge_if_pending_topology_has_valid_slides_video(struct ast_sip_session *session)
{
	unsigned int i = 0;
	unsigned int res = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_WARNING, "judge_if_has_topology_valid_slides_video params error\n");
		return 0;
	}

	if (ast_sip_session_get_only_support_audio()) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			continue;
		}

		if (ast_stream_get_local_media_direction(stream) == UNKNOW ||
			ast_stream_get_local_media_direction(stream) == INACTIVE) {
			res = 0;
		} else {
			res = 1;
		}

		ast_log(LOG_DEBUG, "[session:%p] Find slides video stream, and state is %d, endpoint:%s\n",
			session, ast_stream_get_state(stream), ast_sorcery_object_get_id(session->endpoint));

		break;
	}

	return res;
}

static void process_not_find_in_peer_media(struct ast_sip_session *session,
	struct ast_sip_session *peer_session, struct ast_stream *local_stream)
{
	if ((ast_stream_get_sub_type(local_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		ast_stream_get_sub_type(local_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) &&
		!ast_sip_session_support_bfcp(peer_session) &&
		ast_sip_session_support_bfcp(session) &&
		judge_if_pending_topology_has_valid_slides_video(peer_session)) {
		ast_stream_set_state(local_stream, AST_STREAM_STATE_SENDRECV);
		ast_sip_session_process_special_stream(session, local_stream);
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Can't find local stream in peer, just disable this stream:%d for endpoint:%s\n",
		session, ast_stream_get_sub_type(local_stream), ast_sorcery_object_get_id(session->endpoint));
	append_disable_stream(session, local_stream);

	return;
}

static void add_pending_stream_in_active_topology(struct ast_sip_session *session,
	struct ast_sip_session *peer_session, struct ast_stream_topology *peer_topology)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int find_in_peer = 0;
	struct ast_stream *local_stream = NULL;
	struct ast_stream *req_stream = NULL;
	struct ast_stream_topology *active_topology = NULL;

	active_topology = session->active_media_state->topology;
	if (!active_topology) {
		ast_log(LOG_ERROR, "[session:%p] Send Re-INVITE, but active_topology is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_pending_stream_in_active_topology for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	for (i = 0; i < ast_stream_topology_get_count(active_topology); i++) {
		local_stream = ast_stream_topology_get_stream(active_topology, i);
		if (!local_stream) {
			continue;
		}

		find_in_peer = 0;
		for (j = 0; j < ast_stream_topology_get_count(peer_topology); j++) {
			req_stream = ast_stream_topology_get_stream(peer_topology, j);
			if (!req_stream) {
				continue;
			}

			if (ast_stream_get_sub_type(local_stream) != ast_stream_get_sub_type(req_stream)) {
				continue;
			}

			find_in_peer = 1;

			if (process_disabled_media(session, peer_session, req_stream)) {
				break;
			}

			if ((ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
				ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) &&
				!ast_sip_session_need_create_bfcp_media(session)) {
				append_disable_stream(session, local_stream);
				break;
			}

			if (!ast_sip_session_process_special_stream(session, req_stream)) {
				break;
			}

			if (process_common_stream(session, req_stream)) {
				find_in_peer = 0;
			}

			break;
		}

		if (!find_in_peer) {
			process_not_find_in_peer_media(session, peer_session, local_stream);
		}
	}
}

static void add_pending_stream_increased_by_peer(struct ast_sip_session *session,
	struct ast_sip_session *peer_session, struct ast_stream_topology *peer_topology)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int find_in_local = 0;
	struct ast_stream *local_stream = NULL;
	struct ast_stream *req_stream = NULL;
	struct ast_stream_topology *active_topology = NULL;

	active_topology = session->active_media_state->topology;
	if (!active_topology) {
		ast_log(LOG_ERROR, "[session:%p] Send Re-INVITE, but active_topology is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_pending_stream_increased_by_peer for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	for (i = 0; i < ast_stream_topology_get_count(peer_topology); i++) {
		req_stream = ast_stream_topology_get_stream(peer_topology, i);
		if (!req_stream) {
			continue;
		}

		find_in_local = 0;
		for (j = 0; j < ast_stream_topology_get_count(active_topology); j++) {
			local_stream = ast_stream_topology_get_stream(active_topology, j);
			if (!local_stream) {
				continue;
			}

			if (ast_stream_get_sub_type(local_stream) == ast_stream_get_sub_type(req_stream)) {
				find_in_local = 1;
				break;
			}
		}

		/* If we find in local, we must process this before */
		if (!find_in_local) {
			if (process_disabled_media(session, peer_session, req_stream)) {
				continue;
			}

			if (ast_stream_get_state(req_stream) == AST_STREAM_STATE_REMOVED) {
				ast_log(LOG_DEBUG, "[session:%p] Peer stream state is removed, and local seesion can't get according stream, don't add, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				continue;
			}

			if (ast_sip_session_translate_gswave_to_sip_behavior(session, req_stream, peer_session->channel)) {
				continue;
			}

			if (!ast_sip_session_process_special_stream(session, req_stream)) {
				continue;
			}

			if (!ast_sip_session_jugde_if_media_support(session, ast_stream_get_sub_type(req_stream))) {
				ast_log(LOG_DEBUG, "[session:%p] Can't support BFCP/slides video/FECC for A series, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				continue;
			}

			process_common_stream(session, req_stream);
		}
	}

	return;
}

/* Add by bfdu for video reinivite */
static void get_peer_reinvite_topology(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	struct ast_stream_topology *peer_topology = NULL;

	if (!session || !session->pending_media_state ||
		!peer_session || !peer_session->pending_media_state) {
		ast_log(LOG_DEBUG, "copy_reinvite_caps param is NULL\n");
		return;
	}

	peer_topology = peer_session->pending_media_state->topology;
	if (!peer_topology) {
		if (peer_session->active_media_state && peer_session->active_media_state->topology) {
			peer_topology = peer_session->active_media_state->topology;
		} else {
			ast_log(LOG_DEBUG, "get_peer_reinvite_topology param is NULL\n");
			return;
		}
	}

	add_pending_stream_in_active_topology(session, peer_session, peer_topology);
	add_pending_stream_increased_by_peer(session, peer_session, peer_topology);

	create_default_topology(session, peer_topology);

	// session(callee), peer_session(caller)
	// We make sure any request sdp carry the audio stream, if not this, add default.
	add_reinvite_default_audio_stream(session->pending_media_state->topology, session->endpoint->media.topology, peer_topology);

	return;
}

static void get_peer_reinvite_remote_sdp(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	ast_log(LOG_DEBUG, "copy peer remote sdp begin");

	if (!session || !session->inv_session || !session->endpoint) {
		ast_log(LOG_ERROR, "copy peer remote sdp, session or endpoint is not exist");
		return;
	}

	if (!peer_session->inv_session || !peer_session->endpoint) {
		ast_log(LOG_ERROR, "copy peer remote sdp, reinvite_data session or endpoint is not exist");
		return;
	}

	if (!session->is_direct_media && !session->joint_ipvt) {
		ast_log(LOG_DEBUG, "copy peer remote sdp, direct media is not enabled");
		return;
	}

	pjmedia_sdp_neg_state chan_neg_state = peer_session->inv_session->neg->state;
	if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO || chan_neg_state == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER)
	{
		pjmedia_sdp_session *caller_remote_sdp = peer_session->inv_session->neg->neg_remote_sdp;

		pjsip_inv_set_local_sdp(session->inv_session, caller_remote_sdp);
		{
			char buf[1024] = {0};
			pjmedia_sdp_print(session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
		}
	}
	return;
}

static int get_endpoint_ip_address_version(struct ast_sip_endpoint *endpoint)
{
	struct ao2_container *contacts;
	struct ao2_iterator i;
	struct ast_sip_contact *contact;
	struct ast_sip_aor *aor;
	char aor_id[80]      = {0};
	char contact_uri[80] = {0};
	struct in_addr addr;
	struct in6_addr addr6;
	int ver = 0;
	int ver_tmp = 0;

	aor = ast_sip_location_retrieve_aor(endpoint->aors);
	if (!aor) {
		return 0;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	ao2_ref(aor, -1);
	if (!contacts) {
		return 0;
	}

	i = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_iterator_next(&i))) {
		memset(contact_uri, 0, sizeof(contact_uri));
		sscanf(contact->uri, "%79[^@]@%79[^;]", aor_id, contact_uri);
		ast_log(LOG_DEBUG, "get_endpoint_ip_address_version contact_uri %s\n", contact_uri);
		char *port = NULL;
		char *tmp_ts = NULL;

		/* for ipv6 */
		if (contact_uri[0] != '[') {
			port = strrchr(contact_uri, ':');
		} else {
			tmp_ts = strrchr(contact_uri, ']');
			if (tmp_ts) {
				port = strrchr(tmp_ts, ':');
			}
		}

		if (port) {
			*port = '\0';
		}

		int i = 0, j = 0;
		char host[MAXHOSTNAMELEN] = {0};

		if (contact_uri[0] == '[') {
			for (++i; contact_uri[i] && contact_uri[i] != ']';) {
				host[j++] = contact_uri[i++];
			}
			if (inet_pton(AF_INET6, host, &addr6) > 0) {
				ver = 6;
			}
		}

		if (inet_pton(AF_INET, contact_uri, &addr) > 0) {
			ver = 4;
		}

		if (inet_pton(AF_INET6, contact_uri, &addr6) > 0) {
			ver = 6;
		}

		ao2_ref(contact, -1);
		if (!ver_tmp) {
			ver_tmp = ver;
		}
		if (ver != ver_tmp) {
			ver = 0;
			break;
		}
	}

	ao2_iterator_destroy(&i);
	ao2_ref(contacts, -1);

	ast_log(LOG_DEBUG, "get endpoint %s ip address version %d\n", ast_sorcery_object_get_id(endpoint), ver);
	return ver;
}

static void ast_ignore_session_direct_media_attr(struct ast_sip_session* session)
{
	if (!session) {
		ast_log(LOG_ERROR, "[ast_ignore_session_direct_media_attr] met error: invalid args.\n");
		return;
	}

	// The both must be set.
	session->is_direct_media = 0;
	session->inv_session->is_direct_media = 0;
	if (session->endpoint->media.direct_media.enabled) {
		session->endpoint->ignore_endpt_direct_media = 1;
	} else {
		session->endpoint->ignore_endpt_direct_media = 0;
	}
}

static int chan_pjsip_get_session_ua(struct ast_channel* chan, char* buf, size_t size)
{
	int ret = -1;
	struct ast_sip_channel_pvt* channel_pvt = NULL;

	if (!chan || !buf || size <= 0) {
		return ret;
	}

	channel_pvt = ast_channel_tech_pvt(chan);
	if (channel_pvt && channel_pvt->session) {
		ret = ast_sip_session_get_device_ua(channel_pvt->session, buf, size);
	}

	ast_log(LOG_DEBUG, "[chan_pjsip_get_session_ua] The result: %d\n", ret);

	return ret;
}

static void set_callee_gds_call(struct ast_sip_session* session, pjsip_rx_data* rdata) {
	if (!session || !rdata) {
		ast_log(LOG_ERROR, "[set_callee_gds_call] The session or rdata is nil\n");
		return;
	}

	char useragent[128] = {0};
	const pj_str_t USER_AGENT = { "User-Agent", 10 };
	const char* GDS_USER_AGENT = "Grandstream GDS";
	pjsip_user_agent_hdr* pjsip_ua_hdr = NULL;
	char* is_gds_ua = NULL;

	pjsip_ua_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (!pjsip_ua_hdr) {
		ast_log(LOG_NOTICE, "[set_callee_gds_call] The invite request not had User-Agent header\n");
		return;
	}

	ast_copy_string(useragent, pjsip_ua_hdr->hvalue.ptr,
		(pjsip_ua_hdr->hvalue.slen > sizeof(useragent)) ? sizeof(useragent) - 1 : pjsip_ua_hdr->hvalue.slen);
	is_gds_ua = strstr(useragent, GDS_USER_AGENT);
	if (is_gds_ua) {
		session->is_gds_call = 1;
		ast_channel_set_is_gds_call(session->channel, session->is_gds_call);
	}

	ast_log(LOG_NOTICE, "[set_callee_gds_call] The session: %s is gds call: %d\n",
		ast_sorcery_object_get_id(session->endpoint), session->is_gds_call);
}

static void set_local_gds_call_from_peer(struct ast_sip_session* session, struct ast_sip_session* peer_session, int is_caller) {
	if (!session || !peer_session) {
		return;
	}

	if (is_caller) {
		session->is_gds_call = peer_session->is_gds_call;
		ast_channel_set_is_gds_call(peer_session->channel, peer_session->is_gds_call);
		ast_channel_set_is_gds_call(session->channel, peer_session->is_gds_call);
	} else {
		if (!peer_session->is_gds_call){
			peer_session->is_gds_call = session->is_gds_call;
		}

		ast_channel_set_is_gds_call(session->channel, session->is_gds_call);
		ast_channel_set_is_gds_call(peer_session->channel, session->is_gds_call);
	}
}

static void set_local_hikvision_call_from_peer(struct ast_sip_session* session, struct ast_sip_session* peer_session, int is_caller) {
	if (!session || !peer_session) {
		return;
	}

	if (is_caller) {
		session->is_hikvision_call = peer_session->is_hikvision_call;
		ast_channel_set_is_hikvision_call(peer_session->channel, peer_session->is_hikvision_call);
		ast_channel_set_is_hikvision_call(session->channel, peer_session->is_hikvision_call);
	} else {
		if (!peer_session->is_hikvision_call){
			peer_session->is_hikvision_call = session->is_hikvision_call;
		}

		ast_channel_set_is_hikvision_call(session->channel, session->is_hikvision_call);
		ast_channel_set_is_hikvision_call(peer_session->channel, session->is_hikvision_call);
	}
}

static void set_callee_gsc_call(struct ast_sip_session* session, pjsip_rx_data* rdata) {
	if (!session || !rdata) {
		ast_log(LOG_ERROR, "[set_callee_gsc_call] The session or rdata is nil\n");
		return;
	}

	char useragent[128] = {0};
	const pj_str_t USER_AGENT = { "User-Agent", 10 };
	const char* GDS_USER_AGENT = "Grandstream GSC";
	pjsip_user_agent_hdr* pjsip_ua_hdr = NULL;
	char* is_gsc_ua = NULL;

	pjsip_ua_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (!pjsip_ua_hdr) {
		ast_log(LOG_NOTICE, "[set_callee_gsc_call] The invite request not had User-Agent header\n");
		return;
	}

	ast_copy_string(useragent, pjsip_ua_hdr->hvalue.ptr,
		(pjsip_ua_hdr->hvalue.slen > sizeof(useragent)) ? sizeof(useragent) - 1 : pjsip_ua_hdr->hvalue.slen);
	is_gsc_ua = strstr(useragent, GDS_USER_AGENT);
	if (is_gsc_ua) {
		session->is_gsc_call = 1;
		ast_channel_set_is_gsc_call(session->channel, session->is_gsc_call);
	}

	ast_log(LOG_NOTICE, "[set_callee_gsc_call] The session: %s is gsc call: %d\n",
		ast_sorcery_object_get_id(session->endpoint), session->is_gsc_call);
}

static void set_local_gsc_call_from_peer(struct ast_sip_session* session, struct ast_sip_session* peer_session, int is_caller) {
	if (!session || !peer_session) {
		return;
	}

	if (is_caller) {
		session->is_gsc_call = peer_session->is_gsc_call;
		ast_channel_set_is_gsc_call(peer_session->channel, peer_session->is_gsc_call);
		ast_channel_set_is_gsc_call(session->channel, peer_session->is_gsc_call);
	} else {
		if (!peer_session->is_gsc_call){
			peer_session->is_gsc_call = session->is_gsc_call;
		}

		ast_channel_set_is_gsc_call(session->channel, session->is_gsc_call);
		ast_channel_set_is_gsc_call(peer_session->channel, session->is_gsc_call);
	}
}

static void get_peer_sdp_for_ipvt(struct ast_sip_session* session, struct ast_sip_session* peer_session)
{
	ast_log(LOG_NOTICE, "[IPVT] get_peer_sdp_for_ipvt begin!\n");

	ast_log(LOG_NOTICE, "session channel name: %s,  peer_session channel name: %s\n",
		ast_channel_name(session->channel), ast_channel_name(peer_session->channel));

	ast_log(LOG_NOTICE, "session endpoint joint ipvt: %d,  peer_session endpoint joint ipvt: %d\n",
		session->endpoint->joint_ipvt, peer_session->endpoint->joint_ipvt);

	if (!session->endpoint->joint_ipvt && !peer_session->endpoint->joint_ipvt) {
		ast_log(LOG_NOTICE, "[IPVT] It's a ordinary call, not joint ipvt\n");
		return;
    }

	if ((session->channel && ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)) ||
		(peer_session->channel && ast_test_flag(ast_channel_flags(peer_session->channel), AST_FLAG_RECOVERYING))){
		ast_log(LOG_DEBUG, "It's RECOVERYING for HA, don't use direct media\n");
		return;
	}

	pjmedia_sdp_session* callee_local_sdp = NULL;
	pjmedia_sdp_neg_state neg_state = peer_session->inv_session->neg->state;

	if (neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		callee_local_sdp = pjmedia_sdp_session_clone(session->inv_session->pool_prov, peer_session->inv_session->neg->neg_remote_sdp);
	} else if (neg_state == PJMEDIA_SDP_NEG_STATE_DONE) {
		callee_local_sdp = pjmedia_sdp_session_clone(session->inv_session->pool_prov, peer_session->inv_session->neg->active_remote_sdp);
	}

	if (callee_local_sdp && PJ_SUCCESS == pjsip_inv_set_local_sdp(session->inv_session, callee_local_sdp)) {
		session->joint_ipvt = 1;
		peer_session->joint_ipvt = 1;
		session->inv_session->is_direct_media = 1;
		peer_session->inv_session->is_direct_media = 1;

		char sdp_info[2048];
		memset(sdp_info, 0, sizeof(sdp_info));
		pjmedia_sdp_print(session->inv_session->neg->neg_local_sdp, sdp_info, sizeof(sdp_info) - 1);
		ast_log(LOG_NOTICE, "[IPVT] callee local sdp info:\n%s", sdp_info);
    }

	ast_log(LOG_NOTICE, "[IPVT] get_peer_sdp_for_ipvt end!\n");
}

static void get_peer_init_invite_remote_sdp(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	ast_log(LOG_DEBUG, "get peer remote sdp begin\n");

	int caller_direct_media_ignore = 0;
	int callee_direct_media_ignore = 0;

	ast_get_direct_media_ignore(session->channel, &caller_direct_media_ignore);
	ast_get_direct_media_ignore(peer_session->channel, &callee_direct_media_ignore);

	if (caller_direct_media_ignore || callee_direct_media_ignore) {
		ast_log(LOG_DEBUG, "[get_peer_init_invite_remote_sdp] caller channel or callee channel set direct meida ignore.\n");
		session->is_direct_media = 0;
		peer_session->is_direct_media = 0;
		ast_ignore_session_direct_media_attr(session);
		ast_ignore_session_direct_media_attr(peer_session);
		return;
	}

    if (!session->endpoint->media.direct_media.enabled || !peer_session->endpoint->media.direct_media.enabled) {
		ast_ignore_session_direct_media_attr(session);
		ast_ignore_session_direct_media_attr(peer_session);
		ast_log(LOG_DEBUG, "get peer remote sdp, direct media is not enabled\n");
		return;
	}

	if (session->ssh_proxy_flag || peer_session->ssh_proxy_flag) {
		session->is_direct_media = 0;
		ast_log(LOG_DEBUG, "get_peer_init_invite_remote_sdp, the session had ssh_proxy_flag.");
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE || peer_session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		session->is_direct_media = 0;
		ast_log(LOG_DEBUG, "get_peer_init_invite_remote_sdp, the session is gswave mobile.");
		return;
	}

	if ((session->channel && ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)) ||
		(peer_session->channel && ast_test_flag(ast_channel_flags(peer_session->channel), AST_FLAG_RECOVERYING))){
		ast_log(LOG_DEBUG, "[session:%p] It's RECOVERYING for HA, don't use direct media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	int cfn_flag = 0;
	const char* cfn = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "CFN_FLAG");
	if (cfn && !strncasecmp(cfn, "yes", 3)) {
		ast_log(LOG_DEBUG, "[get_peer_init_invite_remote_sdp] the caller enabled CFN feature, caller and callee media is direct mode.\n");
		cfn_flag = 1;
	}

	int cfb_flag = 0;
	const char* cfb = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "CFB_FLAG");
	if (cfb && !strncasecmp(cfb, "yes", 3)) {
		ast_log(LOG_DEBUG, "[get_peer_init_invite_remote_sdp] the caller enabled CFB feature, caller and callee media is direct mode.\n");
		cfb_flag = 1;
	}

	if ((session->endpoint->ignore_endpt_direct_media > 0 || peer_session->endpoint->ignore_endpt_direct_media > 0) && !cfb_flag && !cfn_flag) {
		ast_ignore_session_direct_media_attr(session);
		ast_ignore_session_direct_media_attr(peer_session);
		session->endpoint->ignore_endpt_direct_media = 0;
		peer_session->endpoint->ignore_endpt_direct_media = 0;
		ast_log(LOG_DEBUG, "get peer remote sdp, ignore endpoint direct media configure option\n");
		return;
	}

	const char *transparentsdp = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "TRANSPARENTSDP");
	if(transparentsdp && !strcasecmp(transparentsdp, "no")){
		ast_log(LOG_DEBUG, "get peer remote sdp, transparentsdp can not be supported\n");
		return;
	}

	const char *from_ivr = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "FROM_IVR");
	if(from_ivr && !strcasecmp(from_ivr, "yes")){
		ast_log(LOG_DEBUG, "get peer remote sdp, requestor is from ivr\n");
		return;
	}

	const char *from_directory = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "FROM_DIRECTORY");
	if(from_directory && !strcasecmp(from_directory, "yes")){
		ast_log(LOG_DEBUG, "get peer remote sdp, requestor is from directory\n");
		return;
	}

	const char *cc_interfaces = pbx_builtin_getvar_helper((struct ast_channel *)peer_session->channel, "CC_INTERFACES");
	if (cc_interfaces) {
		ast_log(LOG_DEBUG, "get peer remote sdp, requestor is recalling\n");
		return;
	}

	if (!get_endpoint_ip_address_version(session->endpoint) ||
		(peer_session->is_sdp_ipv6 && (4 == get_endpoint_ip_address_version(session->endpoint))) ||
		(!peer_session->is_sdp_ipv6 && (6 == get_endpoint_ip_address_version(session->endpoint)))) {
		ast_log(LOG_WARNING, "get peer remote sdp, caller and callee ip address version are not the same\n");
		return;
	}

	if (session->is_webrtc_call ^ peer_session->is_webrtc_call) {
		ast_log(LOG_DEBUG, "get peer remote sdp, caller and callee type are not the same, one is gswave and another is sip phone\n");
		return;
	}

	session->is_direct_media = 1;
	peer_session->is_direct_media = 1;
	session->inv_session->is_direct_media = 1;
	peer_session->inv_session->is_direct_media = 1;

	pjmedia_sdp_neg_state chan_neg_state = peer_session->inv_session->neg->state;
	pjmedia_sdp_session *callee_local_sdp = NULL;
	pj_pool_t *pool = session->inv_session->pool_prov;
	if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		callee_local_sdp = pjmedia_sdp_session_clone(pool, peer_session->inv_session->neg->neg_remote_sdp);
	} else if (chan_neg_state == PJMEDIA_SDP_NEG_STATE_DONE) {
		callee_local_sdp = pjmedia_sdp_session_clone(pool, peer_session->inv_session->neg->active_remote_sdp);
	}

	if (callee_local_sdp) {
		pjsip_inv_set_local_sdp(session->inv_session, callee_local_sdp);
		{
			char buf[1024] = {0};
			pjmedia_sdp_print(session->inv_session->neg->neg_local_sdp, buf, sizeof(buf) - 1); // Must use sizeof(buf) - 1, otherwise there is a potential overflow
			//ast_log(LOG_CODEC_NEG, "[callee] callee local sdp is:\n%s", buf);
		}
	}

	return;
}

static void get_peer_audio_param(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	memcpy(&session->audio_ecan_param, &peer_session->audio_ecan_param, sizeof(session->audio_ecan_param));
	session->audio_has_modem = peer_session->audio_has_modem;

	return;
}

static unsigned int need_create_bfcp_stream_by_self(struct ast_sip_session *peer_session)
{
	if (!judge_if_has_slides_video(peer_session)) {
		return 0;
	}

	if (!ast_sip_session_support_bfcp(peer_session)) {
		ast_log(LOG_DEBUG, "[session:%p] Peer seesion can't support bfcp, add bfcp para by asterisk self for endpoint:%s\n",
			peer_session, ast_sorcery_object_get_id(peer_session->endpoint));
		return 1;
	}

	if (!strncasecmp(peer_session->endpoint->aors, "trunk_", 6) && !judge_if_has_bfcp_stream(peer_session)) {
		ast_log(LOG_DEBUG, "[session:%p] Peer seesion is trunk and didn't contain bfcp, add bfcp para by asterisk self for endpoint:%s\n",
			peer_session, ast_sorcery_object_get_id(peer_session->endpoint));
		return 1;
	}

	return 0;
}

static void get_peer_application_param(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	if (!ast_sip_session_support_bfcp(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't support BFCP, don't need get peer application param for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/* Modify by bfdu for Bug 268012, 2023.8.8 */
	if (need_create_bfcp_stream_by_self(peer_session)) {
		if (!judge_if_pending_topology_has_bfcp_stream(session)) {
			ast_log(LOG_DEBUG, "[session:%p] Session has no bfcp stream, create bfcp stream for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			bfcp_create_stream(session);
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Session already has bfcp stream, don't create bfcp stream for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}

		return;
	}

	memcpy(&session->bfcp_para, &peer_session->bfcp_para, sizeof(session->bfcp_para));
	strcpy(session->bfcp_para.local_floorctrl, peer_session->bfcp_para.remote_floorctrl);
	strcpy(session->bfcp_para.local_setup, peer_session->bfcp_para.remote_setup);

	return;
}

static void get_peer_common_param(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	session->is_yealink_msg = peer_session->is_yealink_msg;
	session->recv_reinvite_no_sdp = peer_session->recv_reinvite_no_sdp;
	session->is_media_count_change = peer_session->is_media_count_change;
	session->reject_call = peer_session->reject_call;
	session->emergemcy_call = peer_session->emergemcy_call;

	get_peer_trunk_info(session, peer_session);

	strcpy(session->caller_gs_client_info, peer_session->caller_gs_client_info);

	return;
}

/* Peer media direction is already clone in local stream, need change local/remote direction value */
static void get_peer_media_direction(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream_topology *topology = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "get_peer_media_direction param is NULL\n");
		return;
	}

	topology = session->pending_media_state->topology;

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		/* Local and remote direction is already clone peer's, need to change.
		 * We need to forward media direction, use peer's remote media direction as local.
		 * We set remote direction sendrecv as default, and later use remote answer to set remote direction.
		 */
		if (session->recv_reinvite_no_sdp) {
			ast_stream_set_local_media_direction(stream, SENDRECV);
		} else {
			ast_stream_set_local_media_direction(stream, ast_stream_get_remote_media_direction(stream));
		}

		ast_stream_set_remote_media_direction(stream, SENDRECV);
	}

	return;
}

static void clear_media_direction(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream_topology *topology = NULL;

	if (ast_channel_state(requestor) < AST_STATE_UP) {
		return;
	}

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "clear_media_direction param is NULL\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] requestor channel state:%d, clear media direction for endpoint:%s\n",
		session, ast_channel_state(requestor), ast_sorcery_object_get_id(session->endpoint));

	topology = session->pending_media_state->topology;

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (ast_sip_session_media_use_bundle(session) && (ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO)) {
			ast_log(LOG_DEBUG, "[session:%p] Media bundle is used and not audio, don't need clear media direction, for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		ast_stream_set_local_media_direction(stream, SENDRECV);
	}

	return;
}

static void get_peer_param_for_local_type(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	unsigned int i = 0;
	unsigned int has_slides_video = 0;
	unsigned int has_bfcp = 0;
	unsigned int create_bfcp = 0;
	struct ast_stream *stream = NULL;

	if (!ast_sip_session_judge_if_type_is_sip(session->channel) ||
		!ast_sip_session_judge_if_type_is_local(requestor)) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] get_peer_param_for_local_type for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->pending_media_state) {
		ast_log(LOG_WARNING, "[session:%p] Has no pending media for peer is local type, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	struct ast_stream_topology *topology = session->pending_media_state->topology;
	if (!topology) {
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);

		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			if (session->is_sfu && session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
				ast_log(LOG_DEBUG, "[session:%p] It's webrtc don't change direction:%d in sfu, for endpoint:%s\n",
					session, ast_stream_get_local_media_direction(stream), ast_sorcery_object_get_id(session->endpoint));
			} else {
				ast_stream_set_local_media_direction(stream, SENDRECV);
			}
			has_slides_video = 1;
		}
		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP) {
			session->bfcp_para.has_udp_bfcp = 1;
			has_bfcp = 1;
		}
		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
			session->bfcp_para.has_tcp_bfcp = 1;
			has_bfcp = 1;
		}
	}

	create_bfcp = ast_sip_session_need_create_bfcp_media(session);
	if (has_slides_video && !has_bfcp && create_bfcp) {
		ast_log(LOG_DEBUG, "[session:%p] Add bfcp media as peer channel is local for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (!judge_if_pending_topology_has_bfcp_stream(session)) {
			bfcp_create_stream(session);
		}
	}

	return;
}

static void get_peer_slides_status(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	if (ast_sip_session_judge_if_type_is_sip(requestor)) {
		ast_channel_set_send_slides_video(session->channel, ast_channel_get_recv_slides_video(requestor));
		ast_channel_set_recv_slides_video(session->channel, ast_channel_get_send_slides_video(requestor));
	} else if (ast_sip_session_judge_if_type_is_local(requestor)) {
		ast_channel_set_send_slides_video(session->channel, ast_channel_get_send_slides_video(requestor));
		ast_channel_set_recv_slides_video(session->channel, ast_channel_get_recv_slides_video(requestor));
	}

	return;
}

static void get_peer_param_for_init_invite(struct ast_sip_session *session, const struct ast_channel *requestor)
{
	int count = 0;
	struct ast_sip_channel_pvt *channel = NULL;

	if (NULL == session || NULL == requestor) {
		ast_log(LOG_ERROR, "get_peer_param_for_init_invite param is NULL\n");
		return;
	}

	if (NULL == session->channel || NULL == session->inv_session) {
		ast_log(LOG_ERROR, "[session:%p] get_peer_param_for_init_invite param is NULL for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!ast_sip_session_judge_if_type_is_sip(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] get_peer_param_for_init_invite, channel type is not PJSIP for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] get_peer_param_for_init_invite begin for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	/*Deadlock avoidance*/
	while(ast_channel_trylock((struct ast_channel *)requestor)) {
		/*cannot goto done since the channel is not locked*/
		if(count++ > 10) {
			ast_debug(1, "[session:%p] Deadlock avoided for channel '%s'\n", session, ast_channel_name(requestor));
			return;
		}
		usleep(1);
	}
	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags((struct ast_channel *)requestor), AST_FLAG_ZOMBIE) || ast_check_hangup((struct ast_channel *)requestor)) {
		ast_channel_unlock((struct ast_channel *)requestor);
		ast_log(LOG_DEBUG, "[session:%p] Peer has been hung up\n", session);
		return;
	}

	get_peer_param_for_local_type(session, requestor);
	get_peer_slides_status(session, requestor);

	if (!ast_sip_session_judge_if_type_is_sip((struct ast_channel *)requestor)) {
		ast_channel_unlock((struct ast_channel *)requestor);
		ast_log(LOG_DEBUG, "[session:%p] get_peer_param_for_init_invite, peer type is not PJSIP for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	channel = ast_channel_tech_pvt(requestor);
	if (!channel || !channel->session) {
		ast_channel_unlock((struct ast_channel *)requestor);
		ast_log(LOG_WARNING, "[session:%p] Requestor has no session for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	get_peer_audio_param(session, channel->session);
	get_peer_application_param(session, channel->session);
	get_peer_common_param(session, channel->session);
	get_peer_init_invite_remote_sdp(session, channel->session);
	get_peer_media_direction(session);

	get_peer_sdp_for_ipvt(session, channel->session);

	set_local_gds_call_from_peer(session, channel->session, 1);
	set_local_gsc_call_from_peer(session, channel->session, 1);
	set_local_hikvision_call_from_peer(session, channel->session, 1);

	if(!session->diff_multiavs_ip)
		strcpy(session->avs_address, channel->session->avs_address);//only copy the avs addess when the init ininvite ,edit for Bug 305270

	/* Clear media direction for transfer, or media direction may sendonly, and has voice issue */
	clear_media_direction(session, requestor);

	ast_channel_unlock((struct ast_channel *)requestor);

	return;
}

static void get_peer_param_for_reinvite(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	int count = 0;

	if (NULL == session || NULL == peer_session) {
		ast_log(LOG_ERROR, "get_peer_session_param_for_reinvite param is NULL\n");
		return;
	}

	if (NULL == session->channel || NULL == peer_session->channel || NULL == session->inv_session) {
		ast_log(LOG_ERROR, "get_peer_session_param_for_reinvite para error\n");
		return;
	}

	if (!ast_sip_session_judge_if_type_is_sip(session->channel) || !ast_sip_session_judge_if_type_is_sip(peer_session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] [peer_session:%p] get_peer_session_param_for_reinvite, channel type is not PJSIP, endpoint:%s\n",
			session, peer_session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] [peer_session:%p] get_peer_session_param_for_reinvite for endpoint:%s\n",
		session, peer_session, ast_sorcery_object_get_id(session->endpoint));

	/*Deadlock avoidance*/
	while(ast_channel_trylock(peer_session->channel)) {
		/*cannot goto done since the channel is not locked*/
		if(count++ > 10) {
			ast_debug(1, "Deadlock avoided for channel '%s'\n", ast_channel_name(peer_session->channel));
			return;
		}
		usleep(1);
	}
	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(peer_session->channel), AST_FLAG_ZOMBIE) || ast_check_hangup(peer_session->channel)) {
		ast_channel_unlock(peer_session->channel);
		ast_log(LOG_DEBUG, "[session:%p] Peer has been hung up, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	get_peer_audio_param(session, peer_session);
	get_peer_common_param(session, peer_session);
	get_peer_reinvite_topology(session, peer_session);
	get_peer_reinvite_remote_sdp(session, peer_session);
	get_peer_media_direction(session);
	get_peer_application_param(session, peer_session);

	ast_channel_unlock(peer_session->channel);

	return;
}

static void chan_pjsip_modify_answer_sdp(struct ast_sip_session *session, struct ast_channel *chan, struct ast_channel *peer)
{
	struct ast_sip_special_answer_data answer_data;
	answer_data.chan = chan;
	answer_data.peer = peer;

	ast_log(LOG_DEBUG, "chan_pjsip_modify_answer_sdp begin, peer[%p]\n", peer);

	ao2_ref(session, +1);
	ast_channel_unlock(chan);

	ast_sip_push_task_wait_serializer(session->serializer, special_answer, &answer_data);

	ao2_ref(session, -1);
	ast_channel_lock(chan);

	return;
}

static void send_reinvite_request_to_callee(struct ast_sip_session *session, const struct ast_sip_reinvite_data *reinvite_data)
{
	if (!session || !session->channel || !ast_channel_tech(session->channel) || !reinvite_data || !reinvite_data->session) {
		return;
	}

	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_request_to_callee for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	get_peer_param_for_reinvite(session, reinvite_data->session);

	reset_media_encrytion(session);

	ao2_ref(session, +1);
	session->is_outgoing = 1;

	if (ast_sip_push_task(session->serializer, send_reinvite_request, session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_reinvite_request_to_callee push task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}

static void send_reinvite_response_to_caller(struct ast_sip_session *session, const struct ast_sip_reinvite_data *reinvite_data)
{
	if (!session || !session->channel || !ast_channel_tech(session->channel) || !reinvite_data || !reinvite_data->session) {
		ast_log(LOG_WARNING, "param is null, can not send video re-invite rsp to peer\n");
		return;
	}

	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "channel type is not pjsip, do not send video re-invite rsp to peer\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_response_to_caller for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	chan_pjsip_modify_answer_sdp(session, session->channel, reinvite_data->session->channel);

	ao2_ref(session, +1);
	ast_channel_unlock(session->channel);
	if (ast_sip_push_task_synchronous(session->serializer, send_reinvite_response, session)) {
		ast_log(LOG_WARNING, "[session:%p] Unable to push send_reinvite_response task to the threadpool for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
		ast_channel_lock(session->channel);
		return;
	}
	ast_channel_lock(session->channel);

	return;
}

static int send_reinvite_request(void *data)
{
	struct ast_sip_session *session = data;
	session->is_media_change = 1;

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_request set is_media_change true for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->is_direct_media && !session->joint_ipvt) {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	} else {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 0, NULL);
	}

	ao2_ref(session, -1);

	return 0;
}

static int send_reinvite_response(void *data)
{
	struct ast_sip_session *session = data;

	ast_log(LOG_DEBUG, "[session:%p] send_video_reinvite_response set is_media_change False for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_send_reinvite_response(session);
	session->is_media_change = 0;
	ao2_ref(session, -1);

	return 0;
}

static int send_reinvite_failed_response(void *data)
{
	pjsip_tx_data *packet = NULL;
	struct ast_sip_session *session = data;
	if (!session) {
		ast_log(LOG_DEBUG, "send_video_reinvite_conflict_response session is NULL\n");
		return 0;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_DEBUG, "[session:%p] send_video_reinvite_conflict_response param is NULL for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
		return 0;
	}

	if (ast_sip_session_delay_request(session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_video_reinvite_conflict_response has delayed request, don't reset is_media_change, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	} else {
		session->is_media_change = 0;
	}

	if (session->channel && ast_sip_session_judge_if_type_is_sip(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] send_video_reinvite_conflict_response answer re-INVITE status code:%d for endpoint:%s\n",
			session, session->reinvite_failed_code, ast_sorcery_object_get_id(session->endpoint));
		if ((pjsip_inv_answer(session->inv_session, session->reinvite_failed_code, NULL, NULL, &packet) == PJ_SUCCESS)) {
			pjmedia_sdp_neg_cancel_wait_nego(session->inv_session->neg);
			ast_sip_session_send_response(session, packet);
		}
	}

	ao2_ref(session, -1);

	return 0;
}

static int send_ssrc_change_reinvite_request(void *data)
{
	struct ast_sip_session *session = data;

	if (!session->is_direct_media) {
		ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	} else {
		ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 0, NULL);
	}

	ao2_ref(session, -1);

	return 0;
}

static void send_reinivte_for_change_ssrc(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinivte_for_change_ssrc for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_set_all_stream_change_ssrc(session);

	ao2_ref(session, +1);
	session->is_outgoing = 1;

	if (ast_sip_push_task(session->serializer, send_ssrc_change_reinvite_request, session)) {
		ao2_cleanup(session);
	}

	return;
}

static enum ast_terminal_type chan_pjsip_get_terminal_type(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return AST_TERMINAL_TYPE_UNKNOW;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return AST_TERMINAL_TYPE_UNKNOW;
	}

	return chan_pvt->session->terminal_type;
}

static int chan_pjsip_get_if_negotiate_state_done(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		return 0;
	}

	if (chan_pvt->session->inv_session->neg->state == PJMEDIA_SDP_NEG_STATE_DONE &&
		chan_pvt->session->pending_media_state &&
		chan_pvt->session->pending_media_state->topology == NULL) {
		return 1;
	}

	ast_log(LOG_DEBUG, "chan_pjsip_get_negotiate_state is not done state for channel:%s\n", ast_channel_name(chan));

	return 0;
}

static unsigned int chan_pjsip_get_delayed_reinvite_count(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	return chan_pvt->session->delayed_reinvite_count;
}

static void chan_pjsip_inc_delayed_reinvite_count(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *session = NULL;

	if (!chan) {
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt) {
		return;
	}

	session = chan_pvt->session;
	if (!session) {
		return;
	}

	session->delayed_reinvite_count++;

	ast_log(LOG_DEBUG, "[session:%p] After increase endpoint:%s has delayed_reinvite_count:%d\n",
		session, ast_sorcery_object_get_id(session->endpoint), session->delayed_reinvite_count);

	return;
}

static int chan_pjsip_get_participant_if_negotiate_state_done(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session || !chan_pvt->session->inv_session || !chan_pvt->session->inv_session->neg) {
		return 0;
	}

	if (chan_pvt->session->pending_participant_media_state &&
		chan_pvt->session->pending_participant_media_state->topology == NULL) {
		return 1;
	}

	ast_log(LOG_DEBUG, "chan_pjsip_get_participant_if_negotiate_state_done is not done state for channel:%s\n", ast_channel_name(chan));

	return 0;
}

static unsigned int chan_pjsip_get_delayed_notify_count(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	return chan_pvt->session->sending_notify_count;
}

static void chan_pjsip_inc_delayed_notify_count(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *session = NULL;

	if (!chan) {
		return;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt) {
		return;
	}

	session = chan_pvt->session;
	if (!session) {
		return;
	}

	session->sending_notify_count++;

	ast_log(LOG_DEBUG, "[session:%p] After increase endpoint:%s has delayed notify count:%d\n",
		session, ast_sorcery_object_get_id(session->endpoint), session->sending_notify_count);

	return;
}

/*Start added by zyxia for Bug202204,2021.8.27*/
static void chan_pjsip_decrease_turn_session_count(struct ast_channel *chan,unsigned int decrease_count)
{
	unsigned int count = 0;
	if (!chan) {
		return;
	}

	ast_sip_decrease_used_turn_session_count(decrease_count);
	count = ast_channel_get_turn_session_count(chan) - decrease_count;
	if (count >= 0)
		ast_channel_set_turn_session_count(chan,count);
	else
		ast_channel_set_turn_session_count(chan,0);

	ast_log(LOG_DEBUG, "[channel:%s] decrease %d nat calls,have %d nat calls\n",
		ast_channel_name(chan), decrease_count,ast_channel_get_turn_session_count(chan));

	return;
}
/*End added by zyxia for Bug202204,2021.8.27*/

static int chan_pjsip_get_enable_mcu(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	/* Now we just enable MCU when enable video recording, add other case later */
	if (chan_pvt->session->conf_video_recording) {
		return 1;
	}

	return 0;
}

static int chan_pjsip_get_limit_time_reached(struct ast_channel *chan)
{
	unsigned int nat_limit_time = 0;
	unsigned int total_limit_time = 0;
	unsigned int used_limit_time = 0;

	nat_limit_time = ast_sip_get_turn_limit_time();
	if (!nat_limit_time) {
		return 0;
	}

	total_limit_time = ast_sip_get_turn_day_total_limit_time();
	used_limit_time = ast_cdr_get_nat_used_time();
	if (used_limit_time >= total_limit_time * 60) {
		ast_log(LOG_DEBUG, "Used limit time is lager than total limit time\n");
		return 1;
	}

	if (total_limit_time * 60 - used_limit_time <= PROMPT_BEFORE_HANGUP_TIME) {
		ast_log(LOG_DEBUG, "Used limit time is less than before hangup time\n");
		return 1;
	}

	return 0;
}

static int chan_pjsip_get_sca_sharedline(struct ast_channel* chan, char* buf, size_t size)
{
	int ret = -1;
	struct ast_sip_channel_pvt* chan_pvt = NULL;
	struct ast_sip_session *session = NULL;

	if (!chan || !buf || size <= 0) {
		return ret;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt) {
		return ret;
	}

	session = chan_pvt->session;
	if (!session) {
		return ret;
	}

	if (session->endpoint && session->endpoint->sca_enable &&
		(session->endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_PRIVATE) &&
		!ast_strlen_zero(session->endpoint->sca_sharedline)) {
			memset(buf, 0, size);
			memcpy(buf, session->endpoint->sca_sharedline, size);
			ret = 0;
		}

	ast_log(LOG_DEBUG, "[chan_pjsip_get_sca_sharedline] The result: %d\n", ret);

	return ret;
}

static char *chan_pjsip_get_slides_video_local_msid(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *session = NULL;
	struct ast_sip_session_media *session_media = NULL;
	char msid[128] = { 0 };

	if (!chan) {
		return NULL;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt) {
		return NULL;
	}

	session = chan_pvt->session;
	if (!session) {
		return NULL;
	}

	session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if (!session_media) {
		ast_log(LOG_ERROR, "[session:%p] Can't get active slides video media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);

	return ast_strdup(msid);
}

static unsigned int chan_pjsip_get_media_turn_relay(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	return chan_pvt->session->media_turn_relay;
}

static unsigned int chan_pjsip_get_gswave_bundle_enabled(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	return chan_pvt->session->endpoint->media.gswave_bundle_enabled;
}

static unsigned int chan_pjsip_get_mobile_bundle_enabled(struct ast_channel *chan)
{
	struct ast_sip_channel_pvt *chan_pvt = NULL;

	if (!chan) {
		return 0;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	return ast_sip_session_media_use_bundle(chan_pvt->session);
}

static void release_slides_video_resource(const char *conf_room, struct ast_channel *chan)
{
	struct ast_bfcp_info bfcp_info;
	char slider_extension[AST_MAX_EXTENSION] = { 0 };
	RAII_VAR(char *, slider_channel_name, NULL, ast_free);

	slider_channel_name = ast_conf_get_slider(conf_room);
	if (!slider_channel_name) {
		ast_log(LOG_DEBUG, "Channel:%s can't get slider channel, don't need to release\n",
			ast_channel_name(chan));
		return;
	}

	if (strcmp(slider_channel_name, ast_channel_name(chan))) {
		ast_log(LOG_DEBUG, "Channel:%s is not the same with slider channel:%s, don't need to release\n",
			ast_channel_name(chan), slider_channel_name);
		return;
	}

	ast_log(LOG_DEBUG, "Get slider channel:%s, and is the owner of this channel, release slides video\n", slider_channel_name);

	ast_conf_put_slides_video_resource(conf_room, chan);

	ast_channel_lock(chan);
	if (ast_channel_tech(chan)->indicate) {
		ast_channel_tech(chan)->indicate(chan, AST_CONTROL_NOTIFY_STOP_SLIDES_VIDEO, NULL, 0);
	}
	ast_channel_unlock(chan);

	ast_sip_get_extension_from_channel_name(ast_channel_name(chan), slider_extension, sizeof(slider_extension));

	/* Need to send bfcp info to notify other users */
	memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
	bfcp_info.primitive = BFCP_FLOOR_STATUS;
	bfcp_info.status = BFCP_RELEASED;
	strcpy(bfcp_info.floor_owner_id, slider_extension);
	strcpy(bfcp_info.info_conf_id, conf_room);
	ast_queue_control_data(chan, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));

	return;
}

static void chan_pjsip_release_video_resource(struct ast_channel *chan)
{
	const char *is_sfu = NULL;
	const char *conf_room = NULL;

	if (!chan) {
		return;
	}

	is_sfu = pbx_builtin_getvar_helper(chan, "IS_SFU");

	if (ast_strlen_zero(is_sfu)) {
		return;
	}

	if (strcasecmp(is_sfu, "yes")) {
		return;
	}

	conf_room = pbx_builtin_getvar_helper(chan, "CONF_ROOM");
	if (ast_strlen_zero(conf_room)) {
		return;
	}

	ast_log(LOG_DEBUG, "chan_pjsip_release_video_resource for channel:%s\n", ast_channel_name(chan));

	ast_conf_put_main_video_resource(conf_room, chan);
	release_slides_video_resource(conf_room, chan);

	return;
}

static int check_params_for_create_pending_media(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	if (!session || !peer_session) {
		return -1;
	}

	if (session->pending_media_state && session->pending_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] It's already has a pending media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->get_peer_session = 1;
		return -1;
	}

	if (!session->active_media_state || !session->active_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] It's has no active media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (!peer_session->active_media_state || !peer_session->active_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] Peer has no active media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	return 0;
}

static unsigned int judge_if_need_create_pending_media(struct ast_sip_session *session,
	struct ast_sip_session *peer_session, struct ast_stream *local_stream, struct ast_stream *peer_stream)
{
	unsigned int need_disable = 0;
	unsigned int need_change = 0;

	if (ast_stream_get_state(local_stream) == ast_stream_get_state(peer_stream)) {
		return 0;
	}

	switch (peer_session->terminal_type) {
		case AST_TERMINAL_TYPE_WEBRTC:
			if (ast_stream_get_state(peer_stream) == AST_STREAM_STATE_REMOVED ||
				ast_stream_get_state(peer_stream) == AST_STREAM_STATE_INACTIVE) {
				need_disable = 1;
			}
			if (ast_stream_get_remote_media_direction(peer_stream) == RECVONLY) {
				ast_stream_set_local_media_direction(local_stream, RECVONLY);
				need_change = 1;
			} else if (ast_stream_get_remote_media_direction(peer_stream) == SENDONLY) {
				ast_stream_set_local_media_direction(local_stream, SENDONLY);
				need_change = 1;
			}
			break;
		default :
			if (ast_stream_get_state(peer_stream) == AST_STREAM_STATE_REMOVED) {
				need_disable = 1;
			}
			if (ast_stream_get_remote_media_direction(peer_stream) == RECVONLY) {
				ast_stream_set_local_media_direction(local_stream, RECVONLY);
				need_change = 1;
			} else if (ast_stream_get_remote_media_direction(peer_stream) == SENDONLY) {
				ast_stream_set_local_media_direction(local_stream, SENDONLY);
				need_change = 1;
			}
			break;
	}

	if (need_disable) {
		ast_sip_session_disable_stream(session, local_stream);
	}

	return (need_disable | need_change);
}

static unsigned int create_pending_media_for_transfer(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int find = 0;
	unsigned int need_create = 0;
	unsigned int count = 0;
	struct ast_stream *local_stream = NULL;
	struct ast_stream *peer_stream = NULL;

	if (check_params_for_create_pending_media(session, peer_session)) {
		return 0;
	}

	if (!peer_session->channel) {
		ast_log(LOG_WARNING, "[session:%p] create_pending_media_for_transfer peer session channel is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/*Deadlock avoidance*/
	while(ast_channel_trylock(peer_session->channel)) {
		/*cannot goto done since the channel is not locked*/
		if(count++ > 10) {
			ast_debug(1, "Deadlock avoided for channel '%s'\n", ast_channel_name(peer_session->channel));
			return 0;
		}
		usleep(1);
	}
	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(peer_session->channel), AST_FLAG_ZOMBIE) || ast_check_hangup(peer_session->channel)) {
		ast_channel_unlock(peer_session->channel);
		ast_log(LOG_DEBUG, "[session:%p] Peer channel:%s has been hung up, endpoint:%s\n",
			session, ast_channel_name(peer_session->channel), ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); ++i) {
		find = 0;
		local_stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!local_stream) {
			continue;
		}
		for (j = 0; j < ast_stream_topology_get_count(peer_session->active_media_state->topology); ++j) {
			peer_stream = ast_stream_topology_get_stream(peer_session->active_media_state->topology, j);
			if (!peer_stream) {
				continue;
			}
			if (!session->is_webrtc_call && peer_session->is_webrtc_call &&
				(ast_stream_get_sub_type(local_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
				ast_stream_get_sub_type(local_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP)) {
				if (ast_stream_get_sub_type(peer_stream) != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
					continue;
				}
			} else if (ast_stream_get_sub_type(local_stream) != ast_stream_get_sub_type(peer_stream)) {
				continue;
			}
			find = 1;
			/* Now we just send re-INVITE for peer media is disable, but local still active */
			if (judge_if_need_create_pending_media(session, peer_session, local_stream, peer_stream)) {
				need_create = 1;
			}
		}
		if (find == 0) {
			need_create = 1;
			ast_sip_session_disable_stream(session, local_stream);
		}
	}

	ast_channel_unlock(peer_session->channel);

	if (need_create == 0) {
		ast_sip_session_media_state_reset(session->pending_media_state);
		return 0;
	}

	return 1;
}

struct topology_control_data {
	struct ast_sip_session *session;
	struct ast_sip_session *peer_session;
};

static void topology_control_data_destroy(void *obj)
{
	struct topology_control_data *data = obj;

	if (!data) {
		return;
	}

	ao2_ref(data->session, -1);
	ao2_ref(data->peer_session, -1);

	return;
}

static struct topology_control_data *topology_control_data_alloc(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	struct topology_control_data *data = ao2_alloc(sizeof(*data), topology_control_data_destroy);

	if (!data) {
		return NULL;
	}

	data->session = session;
	data->peer_session = peer_session;

	ao2_ref(session, +1);
	ao2_ref(peer_session, +1);

	return data;
}

static int send_reinvite_request_for_transfer(void *data)
{
	struct ast_sip_session *session = NULL;
	struct ast_sip_session *peer_session = NULL;
	struct topology_control_data *topology_control = data;

	if (!topology_control) {
		return -1;
	}

	session = topology_control->session;
	peer_session = topology_control->peer_session;
	if (!session || !peer_session) {
		ast_log(LOG_ERROR, "send_reinvite_request_for_transfer session is NULL\n");
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_request_for_transfer endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!create_pending_media_for_transfer(session, peer_session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_reinvite_request_for_transfer don't need send session refresh:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(topology_control, -1);
		return 0;
	}

	if (!session->is_direct_media && !session->joint_ipvt) {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	} else {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 0, NULL);
	}

	ao2_ref(topology_control, -1);

	return 0;
}

static void process_reinvite_req_control_frame(struct ast_sip_session *session, const void *data)
{
	int status_code = 0;
	const struct ast_sip_reinvite_data *req_data = NULL;

	if (!session || !session->channel || !data) {
		return;
	}

	req_data = data;

	ast_log(LOG_DEBUG, "[session:%p] process_reinvite_req_control_frame for endpoint:%s",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session && session->inv_session && session->inv_session->invite_tsx) {
		/* Both channel has receieve reinvite.
		 * re-INVITE conflict, respond 491 to re-INVITE.
		 */
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s already has invite tsx, can't send reinvite\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		status_code = 491;
		ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_FAILED, &status_code, sizeof(status_code));
	} else if (1 == session->is_media_change) {
		ast_log(LOG_DEBUG, "[session:%p] It's dealing with reinvite now, reject this request for endpoint:%s",
			session, ast_sorcery_object_get_id(session->endpoint));
		status_code = 491;
		ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_FAILED, &status_code, sizeof(status_code));
	} else {
		if (judge_if_need_send_reinvite(session, req_data)) {
			send_reinvite_request_to_callee(session, req_data);
		}
	}

	if (req_data && req_data->session && req_data->session->ref_added) {
		req_data->session->ref_added = 0;
		ao2_ref(req_data->session, -1);
	}

	return;
}

static void process_reinvite_rsp_control_frame(struct ast_sip_session *session, const void *data)
{
	const struct ast_sip_reinvite_data *rsp_data = NULL;

	if (!session || !session->channel || !data) {
		return;
	}

	rsp_data = data;

	session->recv_reinvite_no_sdp = 0;

	if (TRUE == session->is_media_change) {
		send_reinvite_response_to_caller(session, rsp_data);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Media change flag is not set, don't need to send re-invite rsp for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	if (rsp_data && rsp_data->session) {
		ao2_ref(rsp_data->session, -1);
	}

	return;
}

/*Start added by zyxia for Bug 260870,2023.6.7*/
static int process_update_main_video_control_frame(void *data)
{
	struct main_video_data *main_video = data;
	const struct ast_channel *callee_chan = NULL;
	struct ast_sip_channel_pvt *callee_channel_pvt = NULL;

	if (!data || !main_video->session || !main_video->session->channel) {
		return -1;
	}

	callee_chan = main_video->data;
	if (callee_chan && !strcmp(ast_channel_tech(callee_chan)->type, "PJSIP") )
	{
		callee_channel_pvt = ast_channel_tech_pvt(callee_chan);
		if(callee_channel_pvt && callee_channel_pvt->session  && change_main_video_direction_by_local_session(callee_channel_pvt->session ,main_video->session)){
			send_reinvite_to_peer(main_video->session);
		}
	}
	/*callee is not pjsip, and caller is video call, need update main video media*/
	else if (callee_chan && strcmp(ast_channel_tech(callee_chan)->type, "PJSIP")
		&& ast_channel_get_is_videocall(main_video->session->channel) && change_active_main_video_direction(main_video->session,INACTIVE)) {
		send_reinvite_to_peer(main_video->session);
	}
	ao2_cleanup(main_video);

	return 0;
}
/*End added by zyxia for Bug 260870,2023.6.7*/

static void process_bfcp_control_frame_for_no_bfcp(struct ast_sip_session *session, const struct ast_bfcp_info *bfcp_info)
{
	struct bfcp_info_data *bfcp_data = NULL;

	if (!session || !bfcp_info) {
		return;
	}

	bfcp_data = ast_sip_session_bfcp_data_alloc(session, bfcp_info);
	if (!bfcp_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc bfcp data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) {
		memset(bfcp_data->bfcp_info->info_conf_id, 0, sizeof(bfcp_data->bfcp_info->info_conf_id));
	}

	if (strlen(bfcp_data->bfcp_info->info_conf_id)) {
		if (ast_sip_push_task(session->serializer, ast_sip_session_bfcp_send_conference_sip_info, bfcp_data)) {
			ast_log(LOG_WARNING, "[session:%p] Send conference sip info push to task failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_cleanup(bfcp_data);
		}
	} else {
		if (ast_sip_push_task(session->serializer, ast_sip_session_bfcp_send_sip_info, bfcp_data)) {
			ast_log(LOG_WARNING, "[session:%p] Send sip info push to task failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_cleanup(bfcp_data);
		}
	}

	return;
}

static void process_bfcp_control_frame_for_has_bfcp(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	struct ast_sip_session_media *bfcp_media = NULL;

	if (!session || !bfcp_info) {
		return;
	}

	bfcp_media = ast_sip_session_get_bfcp_active_media(session);
	if (!bfcp_media) {
		bfcp_media = ast_sip_session_get_bfcp_pending_media(session);
		session->pending_bfcp_message = 1;
	}
	if (session->inv_session->neg->state != PJMEDIA_SDP_NEG_STATE_DONE) {
		ast_log(LOG_DEBUG, "[session:%p] SDP NEG is not done, send bfcp message later, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->pending_bfcp_message = 1;
	}
	bfcp_set_info(session, bfcp_media, bfcp_info);

	if (session->pending_bfcp_message) {
		ast_log(LOG_DEBUG, "[session:%p] Just has bfcp pending media, send bfcp message later, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!bfcp_media) {
		ast_log(LOG_DEBUG, "[session:%p] Has no bfcp active media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (bfcp_media->bfcp_udp) {
		ast_send_bfcp_msg(bfcp_media->bfcp_udp);
	} else if(bfcp_media->bfcp_tcp) {
		ast_send_bfcp_tcp_msg(bfcp_media->bfcp_tcp);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Has no bfcp udp or tcp, can't send bfcp message for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void update_avs_slides_video_msid(struct ast_sip_session *session, const char *msid)
{
	struct ast_sip_session_media *session_media = NULL;
	RAII_VAR(struct ast_channel *, slider_channel, NULL, ao2_cleanup);
	RAII_VAR(char *, slider_msid, NULL, ast_free);

	if (!session) {
		return;
	}

	session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if (!session_media || !session_media->rtp) {
		ast_log(LOG_WARNING, "[session:%p] Can't get active slides video media for update msid, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!ast_strlen_zero(msid)) {
		ast_log(LOG_DEBUG, "[session:%p] update_avs_slides_video_msid msid:%s, endpoint:%s\n",
			session, msid, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_local_msid(session_media->rtp, msid);
		ast_avs_after_nego_notify(session->channel, session_media->rtp);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Can't get msid, get default slider msid, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		slider_channel = ast_conf_get_slider_channel(session->conf_room);
		if (slider_channel) {
			slider_msid = ast_channel_get_slides_video_local_msid(slider_channel);
			if (!ast_strlen_zero(slider_msid)) {
				ast_rtp_instance_set_local_msid(session_media->rtp, slider_msid);
				ast_avs_after_nego_notify(session->channel, session_media->rtp);
			}
		}
	}

	return;
}

static void disable_avs_slides_video_msid(struct ast_sip_session *session)
{
	struct ast_sip_session_media *session_media = NULL;
	char msid[128] = { 0 };

	if (!session || !session->channel) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] disable_avs_slides_video_msid for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if (!session_media || !session_media->rtp) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get slides video media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);

	ast_rtp_instance_set_local_msid(session_media->rtp, msid);
	ast_avs_after_nego_notify(session->channel, session_media->rtp);

	return;

}

static int judge_if_need_update_slides_msid(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	if (!session->is_sfu) {
		return 0;
	}

	if (!session->support_whiteboard) {
		ast_log(LOG_DEBUG, "[session:%p] Can't support whiteboard, need update msid, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	if (strcasecmp(bfcp_info->share_type, "whiteboard") && strcasecmp(bfcp_info->share_type, "document")) {
		ast_log(LOG_DEBUG, "[session:%p] Share type is not whiteboard or document, need update msid, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	return 0;
}

static void change_main_video_stream_resolution(struct ast_sip_session *session,
	struct ast_stream_topology *topology, enum video_resolution resolution)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}

		ast_sip_session_set_video_format(session, stream, resolution);
		ast_stream_set_previous_video_resolution(stream, ast_stream_get_video_resolution(stream));
		ast_stream_set_video_resolution(stream, resolution);
		ast_sip_session_set_as_tias_for_sfu(session, stream);
		ast_sip_session_set_imageattr_for_stream(session, stream, resolution);

		break;
	}

	return;
}

static void send_reinvite_for_change_resolution(struct ast_sip_session *session, enum video_resolution resolution)
{
	if (!session || !session->pending_media_state || !session->active_media_state) {
		ast_log(LOG_ERROR, "send_reinvite_for_change_resolution param error\n");
		return;
	}

	if (session->pending_media_state->topology) {
		ast_log(LOG_ERROR, "[session:%p] There already pending topology, can't send reinvite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->active_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
	} else {
		ast_log(LOG_ERROR, "[session:%p] There's no active topology, can't send reinvite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->pending_media_state->topology) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc pending topology, can't send reinvite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	change_main_video_stream_resolution(session, session->pending_media_state->topology, resolution);

	ao2_ref(session, +1);
	session->is_outgoing = 1;
	session->change_resolution_reinvite = 1;

	if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_camera_control_reinvite push task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_media_state);
		session->is_outgoing = 0;
		session->change_resolution_reinvite = 0;
		ao2_cleanup(session);
	}

	return;
}

static void change_video_resolution(struct ast_sip_session *session, enum video_resolution resolution)
{
	/* 1. Change to 360P, if it's wave, release video resource by wave self,
	 *    If Focus is not wave, wait wave to release then send re-INVITE to change video resolution.
	 * 2. Change to 1080P, if it's wave, reserve 1080P video resource.
	 */
	if (resolution == VIDEO_RESOLUTION_360P &&
		(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		ast_log(LOG_ERROR, "[session:%p] It's wave, change video resolution by themself, and relese resource by themself, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->release_hd_by_wave = 1;
		return;
	}

	ast_conf_put_main_video_resource(session->conf_room, session->channel);

	if (ast_conf_get_main_video_resource(session->conf_room, session->channel, resolution) != resolution) {
		ast_log(LOG_ERROR, "[session:%p] Can't get resolution:%d video resource, endpoint:%s\n",
			session, resolution, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		ast_log(LOG_ERROR, "[session:%p] It's wave, change video resolution by themself, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	send_reinvite_for_change_resolution(session, resolution);

	return;
}

static void process_hd_chan_name_is_focus_chan(struct ast_sip_session *session)
{
	if (ast_conf_get_hd_resolution(session->conf_room) == VIDEO_RESOLUTION_1080P) {
		ast_log(LOG_DEBUG, "[session:%p] Focus video is alreay 1080P, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	change_video_resolution(session, VIDEO_RESOLUTION_1080P);

	return;
}

static void change_hd_user_video_resolution(char *hd_chan_name)
{
	RAII_VAR(struct ast_channel *, chan, NULL, ao2_cleanup);
	struct ast_sip_session *session = NULL;

	if (!hd_chan_name) {
		return;
	}

	chan = ast_channel_get_by_name(hd_chan_name);
	if (!chan) {
		return;
	}

	session = get_sip_session_from_channel(chan);
	if (!session) {
		return;
	}

	change_video_resolution(session, VIDEO_RESOLUTION_360P);

	return;
}

static void process_hd_chan_name_is_not_focus_chan(struct ast_sip_session *session, char *hd_chan_name)
{
	/* Change the hd user to 360P */
	change_hd_user_video_resolution(hd_chan_name);

	/* Change the focus user to 1080P */
	change_video_resolution(session, VIDEO_RESOLUTION_1080P);

	return;
}

static void change_resolution_for_focus_has_video(struct ast_sip_session *session)
{
	RAII_VAR(char *, hd_chan_name, NULL, ast_free);

	ast_log(LOG_DEBUG, "[session:%p] change_resolution_for_focus_has_video, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	hd_chan_name = ast_conf_get_hd_chan_name(session->conf_room);

	if (hd_chan_name && !strcasecmp(hd_chan_name, ast_channel_name(session->channel))) {
		process_hd_chan_name_is_focus_chan(session);
	} else {
		process_hd_chan_name_is_not_focus_chan(session, hd_chan_name);
	}

	return;
}

static void change_resolution_for_focus_no_video(struct ast_sip_session *session)
{
	RAII_VAR(char *, hd_chan_name, NULL, ast_free);

	ast_log(LOG_DEBUG, "[session:%p] change_resolution_for_focus_no_video, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	hd_chan_name = ast_conf_get_hd_chan_name(session->conf_room);

	change_hd_user_video_resolution(hd_chan_name);

	/* reserve video resource */
	if (ast_conf_get_main_video_resource(session->conf_room, session->channel, VIDEO_RESOLUTION_1080P) != VIDEO_RESOLUTION_1080P) {
		ast_log(LOG_ERROR, "[session:%p] Can't reserve 1080P resource, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	return;
}

static void change_video_resolution_for_focus(struct ast_sip_session *session)
{
	if (!session || !session->channel) {
		return;
	}

	if (ast_channel_has_enabled_main_video_stream(session->channel)) {
		change_resolution_for_focus_has_video(session);
	} else {
		change_resolution_for_focus_no_video(session);
	}

	return;
}

static int process_open_focus_video_to_task(void *data)
{
	struct ast_stream *stream = NULL;
	RAII_VAR(char *, slider, NULL, ast_free);
	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);

	if (!session) {
		ast_log(LOG_DEBUG, "[session:%p] process_open_focus_video_to_task session is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	change_video_resolution_for_focus(session);

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable video recording, don't need process this frame, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	slider = ast_conf_get_slider(session->conf_room);
	if (slider) {
		ast_log(LOG_DEBUG, "[session:%p] Conference is opening slides video, don't send avs focus video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	stream = ast_sip_session_get_active_main_video_stream(session);
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) != AST_STREAM_STATE_SENDRECV && ast_stream_get_state(stream) != AST_STREAM_STATE_RECVONLY) {
		ast_log(LOG_WARNING, "[session:%p] process_open_focus_video_notify_frame has active video stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_FOCUS_VIDEO);

	return 0;
}

static void process_wave_released_hd_resolution_frame(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_ERROR, "process_wave_released_hd_resolution_frame session is NULL\n");
		return;
	}

	if (!session->focus_video) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_wave_released_hd_resolution_frame, for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	/* Change the focus user to 1080P */
	change_video_resolution(session, VIDEO_RESOLUTION_1080P);

	return;
}

static int process_callflip_reinvite_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	enum ast_stream_media_direction peer_dir = 0;
	enum ast_stream_media_direction local_dir = 0;
	struct ast_stream *active_stream = NULL;
	struct ast_stream *pending_stream = NULL;

	if (!session) {
		ast_log(LOG_ERROR, "process_callflip_reinvite_frame session is NULL\n");
		return -1;
	}

	if (datalen != sizeof(enum ast_stream_media_direction)) {
		ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_CALLFLIP_REINVITE. Expected %d, got %d\n", (int) sizeof(struct ast_trunk_info), (int) datalen);
		return -1;
	}

	if (session->pending_media_state && session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Already have pending re-INVITE, don't send now, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	peer_dir = *(enum ast_stream_media_direction *)data;

	active_stream = ast_sip_session_get_active_slides_video_stream(session);
	if (!active_stream) {
		ast_log(LOG_DEBUG, "[session:%p] Have no slides video stream, don't send now, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	local_dir = ast_stream_get_local_media_direction(active_stream);
	if (peer_dir == local_dir) {
		ast_log(LOG_DEBUG, "[session:%p] Already has the same direction:%d, don't need send, endpoint:%s\n",
			session, peer_dir, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
	if (!session->pending_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] process_callflip_reinvite_frame clone topology failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	pending_stream = ast_sip_session_get_pending_slides_video_stream(session);
	if (!pending_stream) {
		ast_log(LOG_ERROR, "[session:%p] Have no pending slides video stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_topology_free(session->pending_media_state->topology);
		session->pending_media_state->topology = NULL;
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_callflip_reinvite_frame get peer direction:%d, send re-INVTE to re-neg, endpoint:%s\n",
		session, peer_dir, ast_sorcery_object_get_id(session->endpoint));

	switch (peer_dir) {
		case SENDRECV:
			ast_stream_set_state(pending_stream, AST_STREAM_STATE_SENDRECV);
			ast_stream_set_local_media_direction(pending_stream, SENDRECV);
			ast_stream_set_remote_media_direction(pending_stream, SENDRECV);
			break;
		case SENDONLY:
			ast_stream_set_state(pending_stream, AST_STREAM_STATE_SENDONLY);
			ast_stream_set_local_media_direction(pending_stream, SENDONLY);
			ast_stream_set_remote_media_direction(pending_stream, SENDONLY);
			break;
		case RECVONLY:
			ast_stream_set_state(pending_stream, AST_STREAM_STATE_RECVONLY);
			ast_stream_set_local_media_direction(pending_stream, RECVONLY);
			ast_stream_set_remote_media_direction(pending_stream, RECVONLY);
			break;
		default:
			ast_stream_set_state(pending_stream, AST_STREAM_STATE_INACTIVE);
			ast_stream_set_local_media_direction(pending_stream, INACTIVE);
			ast_stream_set_remote_media_direction(pending_stream, INACTIVE);
			break;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
		ast_log(LOG_ERROR, "[session:%p] Can't push task to send reinvite for add slides video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_topology_free(session->pending_media_state->topology);
		session->pending_media_state->topology = NULL;
		ao2_cleanup(session);
	}

	return 0;
}

static void send_callflip_notify_to_update_trunk_info(struct ast_sip_session *session, struct ast_trunk_info *trunk_info)
{
	struct ast_sip_callflip_state callflip_info;
	struct callflip_user *callflip_state = NULL;

	if (!ast_sip_get_callflip_enable()) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable callflip, don't need send NOTIFY to update, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	memset(&callflip_info, 0, sizeof(struct ast_sip_callflip_state));

	if (!(callflip_state = ast_sip_callflip_find_user(session->endpoint->aors))) {
		ast_log(LOG_WARNING, "[session:%p] Can't find callflip state, don't need send NOTIFY to update, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (strcasecmp(callflip_state->enable, "yes")) {
		ast_log(LOG_WARNING, "[session:%p] Callflip state is not enabled, don't need send NOTIFY to update, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(callflip_info.enable, sizeof(callflip_info.enable), "%s", callflip_state->enable);
	snprintf(callflip_info.remote_id, sizeof(callflip_info.remote_id), "%s", callflip_state->remote_id);
	snprintf(callflip_info.remote_name, sizeof(callflip_info.remote_name), "%s",callflip_state->remote_name);
	snprintf(callflip_info.call_type, sizeof(callflip_info.call_type), "%s", callflip_state->call_type);
	snprintf(callflip_info.media_type, sizeof(callflip_info.media_type), "%s", callflip_state->media_type);
	snprintf(callflip_info.channel_id, sizeof(callflip_info.channel_id), "%s", callflip_state->channel_id);
	snprintf(callflip_info.conf_link, sizeof(callflip_info.conf_link), "%s", trunk_info->conf_link);
	snprintf(callflip_info.replace_channel_id, sizeof(callflip_info.replace_channel_id), "%s", trunk_info->channel_id);

	ast_sip_callflip_build_user(&callflip_info,session->endpoint->aors);//save the callflip state to contain

	ast_sip_for_each_aor(session->endpoint->aors, process_callflip_aor, &callflip_info);

	return;
}

/*Start added by zyxia for Bug 256745,277884,284874 dont wait the prompt play finished*/
static int process_play_prompt_to_caller(void* data){

	struct ast_sip_session *session = data;

	if (!session || !session->channel) {
		ast_log(LOG_ERROR, "send_reinvite_for_session_refresh invalid session\n");
		ao2_cleanup(session);
		return -1;
	}

	ast_log(LOG_DEBUG, "play_prompt_to_caller enter\n");
	/*play prompt*/

	const char* play_moh = pbx_builtin_getvar_helper(session->channel, "STOP_PLAY_APP_PROMPT");
	if(ast_strlen_zero(play_moh)) {//if has get new register ,dont play it
		ast_stream_and_wait(session->channel, "pls-wait-connect-call", "");
	}

	ao2_ref(session, -1);
	ast_log(LOG_DEBUG, "play_prompt_to_caller exit\n");
	return 0;
}

/*End added by zyxia for Bug Bug 256745,277884,284874,dont wait the prompt play finished*/

static int process_open_ipc_video(void* data){
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream_topology *topology = NULL;
	struct ast_sip_session *session = data;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->active_media_state) {
		return -1;
	}

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
		if (!session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] process_open_ipc_video clone topology failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}
	}
	session_media = ast_sip_session_get_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	if (!session_media) {
		ast_log(LOG_WARNING, "[session:%p] Can't get main video session_media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	topology = session->pending_media_state->topology;

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO && ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN && ast_stream_get_state(stream) != AST_STREAM_STATE_RECVONLY) {
			ast_stream_set_state(stream, AST_STREAM_STATE_RECVONLY);
			ast_stream_set_local_media_direction(stream, RECVONLY);
			//ast_stream_set_remote_media_direction(stream, RECVONLY);
			session_media->open_camera = 1;
			session_media->changed = 1;
			ast_sip_session_set_main_video_stream_uptime(session);
			ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
		}
	}

	ao2_ref(session, -1);
	ast_log(LOG_DEBUG, "process_open_ipc_video exit\n");
	return 0;
}

static void trunk_info_data_destroy(void *obj)
{
	struct trunk_info_data *data = obj;

	if (!data) {
		return;
	}

	if (data->session) {
		ao2_ref(data->session, -1);
	}

	return;
}

static struct trunk_info_data *trunk_info_data_alloc(struct ast_sip_session *session,
	const struct ast_trunk_info *trunk_info)
{
	struct trunk_info_data *data = NULL;

	if (!session || !trunk_info) {
		return NULL;
	}

	data = ao2_alloc(sizeof(*data), trunk_info_data_destroy);
	if (!data) {
		return NULL;
	}

	strcpy(data->trunk_info.conf_link, trunk_info->conf_link);
	strcpy(data->trunk_info.channel_id, trunk_info->channel_id);

	ao2_ref(session, +1);
	data->session = session;

	return data;
}

static int process_notify_trunk_info_frame_to_task(void *data)
{
	struct trunk_info_data *info_data = data;
	struct ast_sip_session *session = NULL;

	if (!info_data) {
		return -1;
	}

	session = info_data->session;

	send_callflip_notify_to_update_trunk_info(session, &info_data->trunk_info);

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		ast_sip_session_send_notify_to_update_gs_state(session, AST_GS_STATE_TYPE_TRUNK_INFO, &info_data->trunk_info);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] It's not wave, don't need send NOTIFY to update, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	ao2_ref(info_data, -1);

	return 0;
}

static int process_notify_trunk_info_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	struct ast_trunk_info *trunk_info = NULL;
	struct trunk_info_data *info_data = NULL;

	if (!session) {
		ast_log(LOG_ERROR, "process_notify_trunk_info_frame session is NULL\n");
		return -1;
	}

	if (datalen != sizeof(struct ast_trunk_info)) {
		ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_NOTIFY_TRUNK_INFO. Expected %d, got %d\n", (int) sizeof(struct ast_trunk_info), (int) datalen);
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_notify_trunk_info_frame, for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	trunk_info = (struct ast_trunk_info *)data;

	/*Start added by zyxia for Bug 236617,2022.9.26*/
	if (!session->conf_link && trunk_info->conf_link) {
		session->conf_link = ast_calloc(1, strlen(trunk_info->conf_link) + 1);
		if (session->conf_link) {
			strcpy(session->conf_link, trunk_info->conf_link);
		}
	}
	if (!session->trunk_channel_id && trunk_info->channel_id) {
		session->trunk_channel_id = ast_calloc(1, strlen(trunk_info->channel_id) + 1);
		if (session->trunk_channel_id) {
			strcpy(session->trunk_channel_id, trunk_info->channel_id);
		}
	}
	/*End added by zyxia for Bug 236617,2022.9.26*/

	info_data = trunk_info_data_alloc(session, trunk_info);
	if (!info_data) {
		ast_log(LOG_WARNING, "[session:%p] process_notify_trunk_info_frame alloc data failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (ast_sip_push_task(session->serializer, process_notify_trunk_info_frame_to_task, info_data)) {
		ast_log(LOG_WARNING, "[session:%p] process_notify_trunk_info_frame to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(info_data, -1);
	}

	return 0;
}


/*Start added by zyxia for Bug 252996 ,2023.3.9*/
static int send_update_for_trunk_conf(void *data)
{
	struct ast_sip_session *session = data;

	ast_log(LOG_DEBUG, "[session:%p] send_update_for_trunk_conf endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_UPDATE, 0, NULL);

	ao2_cleanup(session);
	return 0;
}
/*End added by zyxia for Bug 252996 ,2023.3.9*/

static int process_change_to_trunk_conf_frame(struct ast_sip_session *session, char* trunk_conf_num)
{
	if (!session || !trunk_conf_num) {
		ast_log(LOG_ERROR, "process_change_to_trunk_conf_frame session or trunk_conf_num is NULL\n");
		return -1;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		ast_log(LOG_DEBUG, "[session:%p] process_change_to_trunk_conf_frame, is wave terminal, don't need support trunk cei, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	/*trunk channel ,send */
	if (session->channel && !strncmp(ast_channel_name(session->channel),"PJSIP/trunk_",12)){
		send_trunk_conf_cei_subscribe(session);
		ast_log(LOG_ERROR, "process_change_to_trunk_conf_frame for subscribe %s through trunk channel %s\n",trunk_conf_num,ast_channel_name(session->channel));
		return 1;
	}

	if (!session->trunk_conf_num ) {
		session->trunk_conf_num = ast_calloc(1, strlen(trunk_conf_num) + 1);
		if (session->trunk_conf_num) {
			strcpy(session->trunk_conf_num, trunk_conf_num);
		}
	}

	ast_channel_set_is_join_trunk_conf(session->channel,1);
	ast_log(LOG_DEBUG, "process_change_to_trunk_conf_frame session or trunk_conf_num is %s\n",trunk_conf_num);

	//sip call to ucm trunk conf   or call from to trunk conf
	const char *trunk_index= pbx_builtin_getvar_helper(session->channel, "TRUNK_INDEX");
	if(!ast_strlen_zero(trunk_index)){
		char local_virtual_conf[512] = {0};
		char family[64] = {0};
		char trunkName[128] = {0};
		ast_log(LOG_DEBUG, "process_change_to_trunk_conf_frame:%s,%s\n",session->trunk_conf_num,ast_channel_name(session->channel));

		snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(trunk_index));
		ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
		snprintf(local_virtual_conf, sizeof(local_virtual_conf) - 1, "%s_%s", trunkName,session->trunk_conf_num);

		struct ast_sip_endpoint *endpoint = session->endpoint;
		const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
		if (endpoint != NULL && endpoint_id != NULL) {//remove
			if (add_peer_conference(ast_channel_name(session->channel), endpoint, local_virtual_conf,AST_CONF_TYPE_VIRTUAL) == 0){
				ast_channel_set_virtual_conf_num(session->channel, local_virtual_conf);
			}
		}
	}

	/*Start added by zyxia for Bug 252996,2023.3.9*/
	/*send update to sip terminal*/
	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_update_for_trunk_conf, session)) {
		ao2_cleanup(session);
	}
	/*End added by zyxia for Bug 252996,2023.3.9*/

	return 0;
}

/* when host enable focus video record or close screen share, we need to judge if need send avs cmd to set focus video */
static int open_focus_video_after_operate(struct ast_sip_session *session)
{
	if (!session) {
		return -1;
	}

	if (!session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] open_focus_video_after_operate is not SFU, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (!session->focus_video) {
		ast_log(LOG_DEBUG, "[session:%p] open_focus_video_after_operate is not focus video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, process_open_focus_video_to_task, session)) {
		ast_log(LOG_WARNING, "[session:%p] open_focus_video_after_operate to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return 0;
}

/* when host enable focus video record or close screen share, we need to judge if need send avs cmd to set focus video */
static int open_slides_video_after_operate(struct ast_sip_session *session)
{
	if (!session) {
		return -1;
	}

	if (!session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] open_slides_video_after_operate is not SFU, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (!session->sfu_recv_slides_video && !ast_channel_get_opening_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] open_slides_video_after_operate is not open slides video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_SLIDES_VIDEO);

	return 0;
}

static int process_conf_video_record_start_frame(struct ast_sip_session *session)
{
	if (!session) {
		return -1;
	}

	session->conf_video_recording = 1;

	open_focus_video_after_operate(session);

	open_slides_video_after_operate(session);

	return 0;
}

static int judge_if_send_bfcp_by_info(struct ast_sip_session *session, struct ast_bfcp_info *bfcp_info)
{
	if (!session || !bfcp_info) {
		return 0;
	}

	if (!ast_sip_session_support_bfcp(session) || session->ucm_trunk) {
		return 1;
	}

	if (session->support_whiteboard &&
		(!strcasecmp(bfcp_info->share_type, "whiteboard") || !strcasecmp(bfcp_info->share_type, "document"))) {
		ast_log(LOG_DEBUG, "[session:%p] Support whiteboard and share type is whiteboard, use SIP INFO to send BFCP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->send_bfcp_request_by_info = 1;
		return 1;
	}

	if (bfcp_info->status == BFCP_RELEASED && session->send_bfcp_request_by_info) {
		ast_log(LOG_DEBUG, "[session:%p] Send BFCP by SIP INFO before, when release need use SIP INFO too, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->send_bfcp_request_by_info = 0;
		return 1;
	}

	return 0;
}

static int process_bfcp_control_frame_to_task(void *data)
{
	struct ast_sip_session *session = NULL;
	struct ast_bfcp_info *bfcp_info = NULL;

	struct bfcp_info_data *bfcp_data = data;

	if (!bfcp_data) {
		return -1;
	}

	session = bfcp_data->session;
	bfcp_info = bfcp_data->bfcp_info;

	/*Start added by zyxia for Bug 218922*/
	if (!ast_sip_session_if_need_send_bfcp_info(session, bfcp_info->share_type)){
		ao2_ref(bfcp_data, -1);
		return -1;
	}
	/*End added by zyxia for Bug 218922*/

	/* Don't need check sesssion and bfcp_info again */

	if (judge_if_send_bfcp_by_info(session, bfcp_info)) {
		process_bfcp_control_frame_for_no_bfcp(session, bfcp_info);
	} else {
		process_bfcp_control_frame_for_has_bfcp(session, bfcp_info);
	}

	if (bfcp_info->primitive == BFCP_FLOOR_REQUEST && bfcp_info->status == BFCP_DENIED) {
		ast_log(LOG_DEBUG, "[session:%p] Denied the floor request, don't need do other process, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(bfcp_data, -1);
		return 0;
	}

	if ((bfcp_info->primitive == BFCP_FLOOR_REQUEST || bfcp_info->primitive == BFCP_FLOOR_STATUS) &&
		bfcp_info->status == BFCP_GRANTED) {
		ast_log(LOG_DEBUG, "It's a open screen action, set session open slides video true\n");
		/* For wave switch share source will not release first, so we need to judge if need to update avs slides video */
		if (session->open_slides_video) {
			if (session->update_avs_slides_video) {
				if (session->support_whiteboard &&
					(!strcasecmp(bfcp_info->share_type, "whiteboard") || !strcasecmp(bfcp_info->share_type, "document"))) {
					send_avs_slides_video_control(session, AVS_CTRL_OPT_STOP);
					disable_avs_slides_video_msid(session);
					session->update_avs_slides_video = 0;
				}
			} else {
				if (judge_if_need_update_slides_msid(session, bfcp_info)) {
					send_avs_slides_video_control(session, AVS_CTRL_OPT_START);
					update_avs_slides_video_msid(session, bfcp_info->msid);
					session->update_avs_slides_video = 1;
				}
			}
		} else {
			session->open_slides_video = 1;
			if (judge_if_need_update_slides_msid(session, bfcp_info)) {
				send_avs_slides_video_control(session, AVS_CTRL_OPT_START);
				update_avs_slides_video_msid(session, bfcp_info->msid);
				session->update_avs_slides_video = 1;
			}
		}
	}

	if (!session->is_sfu && (bfcp_info->primitive == BFCP_FLOOR_REQUEST ||
		(bfcp_info->primitive == BFCP_FLOOR_STATUS && bfcp_info->status == BFCP_GRANTED))) {
		ast_log(LOG_DEBUG, "It's a open screen action for P2P call, set session open slides video true\n");
		session->open_slides_video = 1;
	}

	if (bfcp_info->primitive == BFCP_FLOOR_STATUS && bfcp_info->status == BFCP_RELEASED) {
		ast_log(LOG_DEBUG, "[session:%p] It's a close screen action, set session open slides video false, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->open_slides_video = 0;
		if (judge_if_need_update_slides_msid(session, bfcp_info) && session->update_avs_slides_video) {
			send_avs_slides_video_control(session, AVS_CTRL_OPT_STOP);
			disable_avs_slides_video_msid(session);
			session->update_avs_slides_video = 0;
		}
		open_focus_video_after_operate(session);
	}

	ast_bfcp_set_channel_slides_status(session->channel, bfcp_info, 0);

	ao2_ref(bfcp_data, -1);

	return 0;
}

static void process_bfcp_control_frame(struct ast_sip_session *session, const void *data)
{
	struct bfcp_info_data *bfcp_data = NULL;
	const struct ast_bfcp_info *bfcp_info = data;

	if (!session || !bfcp_info) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_bfcp_control_frame for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	bfcp_data = ast_sip_session_bfcp_data_alloc(session, bfcp_info);
	if (!bfcp_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc bfcp data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_push_task(session->serializer, process_bfcp_control_frame_to_task, bfcp_data)) {
		ast_log(LOG_WARNING, "[session:%p] process_bfcp_control_frame to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(bfcp_data);
	}

	return;
}

static void process_stop_slides_video_by_host_frame(struct ast_sip_session *session, const void *data)
{
	struct ast_bfcp_info *bfcp_info = (struct ast_bfcp_info *)data;

	if (!session || !bfcp_info) {
		return;
	}

	if (bfcp_info->primitive == BFCP_FLOOR_STATUS && bfcp_info->status == BFCP_RELEASED) {
		ast_log(LOG_DEBUG, "[session:%p] It's a close screen action, set session open slides video false, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->open_slides_video = 0;
		if (judge_if_need_update_slides_msid(session, bfcp_info) && session->update_avs_slides_video) {
			send_avs_slides_video_control(session, AVS_CTRL_OPT_STOP);
			disable_avs_slides_video_msid(session);
			session->update_avs_slides_video = 0;
		}
		open_focus_video_after_operate(session);
	}

	return;
}

static void process_topology_control_frame(struct ast_sip_session *session, struct ast_sip_session *peer_session)
{
	struct topology_control_data *data = NULL;

	if (!session || !peer_session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_topology_control_frame, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	data = topology_control_data_alloc(session, peer_session);
	if (!data) {
		ast_log(LOG_ERROR, "[session:%p] process_topology_control_frame can't alloc, for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->is_outgoing = 1;

	if (ast_sip_push_task(session->serializer, send_reinvite_request_for_transfer, data)) {
		ao2_cleanup(data);
	}

	return;
}

static void process_reinvite_failed_frame(struct ast_sip_session *session, const void *data)
{
	if (!session || !session->inv_session || !session->inv_session->neg || !data) {
		ast_log(LOG_WARNING, "Recv process_reinvite_failed_frame, invalid param\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Recv process_reinvite_failed_frame, for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->inv_session->neg->state != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_WARNING, "[session:%p] Recv process_reinvite_failed_frame, neg state is not wait nego, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->recv_reinvite_no_sdp = 0;
	ast_sip_session_media_state_reset(session->pending_media_state);
	if (TRUE == session->is_media_change) {
		ao2_ref(session, +1);
		session->reinvite_failed_code = *(int *)data;
		if (ast_sip_push_task(session->serializer, send_reinvite_failed_response, session)) {
			ao2_cleanup(session);
		}
	}

	return;
}

static void process_reinvite_req_failed_frame(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_ERROR, "process_reinvite_req_failed_frame session is NULL\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Recv process_reinvite_req_failed_frame, for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->ref_added) {
		ast_log(LOG_DEBUG, "[session:%p] Recv process_reinvite_req_failed_frame, ref didn't add, maybe something wrong for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->ref_added = 0;

	if (session->recv_reinvite_no_sdp) {
		session->recv_reinvite_no_sdp = 0;
		ao2_cleanup(session);
		return;
	}

	ast_sip_session_media_state_reset(session->pending_media_state);

	session->reinvite_failed_code = 488;

	/* For previous we already add ref for queue AST_CONTROL_REINVITE_REQ, don't need add ref again */
	if (ast_sip_push_task(session->serializer, send_reinvite_failed_response, session)) {
		ao2_cleanup(session);
	}

	return;
}

static void process_reinvite_rsp_failed_frame(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_ERROR, "process_reinvite_rsp_failed_frame session is NULL\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_reinvite_rsp_failed_frame, for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ao2_ref(session, -1);

	return;
}

static void process_reinvite_rsp_again_frame(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_ERROR, "process_reinvite_rsp_again_frame session is NULL\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_reinvite_rsp_again_frame for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	send_response_control_frame_to_peer(session);

	return;
}

static void process_queue_peer_session_failed_frame(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_ERROR, "process_queue_peer_session_failed_frame session is NULL\n");
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_queue_peer_session_failed_frame, for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session->ref_added = 0;

	ao2_ref(session, -1);

	return;
}

static void free_candidate_list(struct avs_update_candidate_info *info)
{
	struct candidate *cand_tmp = NULL;

	if (!info) {
		return;
	}

	AST_LIST_TRAVERSE_SAFE_BEGIN(&info->candidate_list, cand_tmp, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_free(cand_tmp);
		cand_tmp = NULL;
	}
	AST_LIST_TRAVERSE_SAFE_END;

	return;
}

static void process_update_candidates_control_frame(struct ast_sip_session *session, const void *data)
{
	unsigned int i = 0;
	struct ast_sip_session_media_state *media_state = NULL;
	struct ast_sip_session_media *session_media = NULL;
	struct avs_update_candidate_info *info = NULL;
	char port_id[MAX_PORTID_LEN] = { 0 };

	if (!session || !session->inv_session) {
		return;
	}

	info = (struct avs_update_candidate_info *)data;
	if (!info) {
		return;
	}

	if (!session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no pending media, use active media\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		media_state = session->active_media_state;
	} else {
		media_state = session->pending_media_state;
	}

	if (!media_state || !media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] There's no topology for create sdp for trickle ice, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		free_candidate_list(info);
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(media_state->topology); ++i) {
		session_media = AST_VECTOR_GET(&media_state->sessions, i);
		if (!session_media || !session_media->rtp) {
			ast_log(LOG_WARNING, "[session:%p] Can't get session_media for create trickle ice sdp, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		snprintf(port_id, sizeof(port_id) - 1, "%ld", ast_rtp_instance_get_avs_portid(session_media->rtp));

		if (strcasecmp(port_id, info->port_id)) {
			ast_log(LOG_DEBUG, "[session:%p] session_media port_id:%s is not the same as info port_id:%s, endpoint:%s\n",
				session, port_id, info->port_id, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		ast_rtp_instance_clear_local_candidates(session_media->rtp);

		ast_rtp_instance_set_ice_flag(session_media->rtp, 1);
		ast_rtp_instance_set_local_ice_ufrag(session_media->rtp, info->ice_ufrag);
		ast_rtp_instance_set_local_ice_pwd(session_media->rtp, info->ice_pwd);
		ast_rtp_instance_set_local_candidates(session_media->rtp, &info->candidate_list);
		session_media->end_of_candidates = info->end_of_candidate;

		break;
	}

	free_candidate_list(info);

	if (session->inv_session->state < PJSIP_INV_STATE_EARLY) {
		ast_log(LOG_DEBUG, "[session:%p] Dialog is not established really, don't send sip info for trickle ice, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->trickle_info_delayed = 1;
		return ;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, ast_sip_session_send_info_for_add_candidates, session)) {
		ao2_cleanup(session);
	}

	return;
}

static unsigned int if_has_valid_video_stream(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_INACTIVE ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_SENDONLY) {
		return 0;
	}

	return 1;
}

static void process_nway_join_confenerce(struct ast_sip_session *session)
{
	struct ast_stream *main_video = NULL;
	int resolution = -1;
	enum video_resolution used_resolution = VIDEO_RESOLUTION_NONE;

	ast_log(LOG_DEBUG, "[session:%p] process_nway_join_confenerce for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	update_sfu_params_for_join_conference(session);

	main_video = ast_sip_session_get_active_main_video_stream(session);
	if (if_has_valid_video_stream(main_video)) {
		used_resolution = get_resolution_for_h264_codec(session, 1);
		if ((used_resolution <= VIDEO_RESOLUTION_NONE) || (used_resolution >= VIDEO_RESOLUTION_MAX)) {
			ast_log(LOG_ERROR, "[session:%p] process_nway_join_confenerce can't get video resolution, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}

		resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, used_resolution);
		if ((resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) || (resolution == -1)) {
			/* Todo, It seems can't failed, if happen need to close camera */
			ast_log(LOG_ERROR, "[session:%p] process_nway_join_confenerce can't get video resource, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		} else {
			ast_log(LOG_DEBUG, "[session:%p] process_nway_join_confenerce get resolution:%d for endpoint:%s\n",
				session, resolution, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_set_video_resolution(main_video, resolution);
			ast_conf_cei_video_resolution_update(session->channel, session->conf_room, resolution);
		}
	}

	return;
}

static int join_nway_conf_frame_task(void *data)
{
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	struct ast_stream *slides_video = NULL;

	session = data;
	if (!session || !session->channel) {
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] join_nway_conf_frame_task for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	slides_video = ast_sip_session_get_active_slides_video_stream(session);
	if (if_has_valid_video_stream(slides_video) && ast_channel_get_send_slides_video(session->channel)) {
		if (VIDEO_RESOLUTION_1080P == ast_conf_get_slides_video_resource(session->conf_room, session->channel, session->notify_share_audio)) {
			ast_log(LOG_DEBUG, "[session:%p] Get slides video resource sucess for conf:%s, channel:%s\n",
				session, session->conf_room, ast_channel_name(session->channel));
			ast_channel_set_opening_slides_video(session->channel, 1);
		} else {
			/* Todo, It seems can't failed, if happen need to close slides video */
			ast_log(LOG_ERROR, "[session:%p] join_nway_conf_frame_task can't get video resource, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	} else if (!ast_channel_get_recv_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] join_nway_conf_frame_task send BFCP or INFO if need, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_notify_sfu_bfcp_for_join_conference(session);
	}

	return 0;
}

static void process_join_nway_conf_frame(struct ast_sip_session *session)
{
	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, join_nway_conf_frame_task, session)) {
		ast_log(LOG_DEBUG, "[session:%p] Push join_sfu_control_frame_task to task failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}

static void hangup_some_associated_call(struct ast_sip_session* session)
{
	if (!session || !session->channel) {
		return;
	}

	char associated_name[80] = {0};
	struct ast_channel* associated_channel = NULL;

	if (!ast_channel_spy_with_ringgroup(session->channel)) {
		return;
	}

	ast_channel_get_ringgroup_name(session->channel, associated_name, sizeof(associated_name));
	if (strlen(associated_name) <= 0) {
		return;
	}

	associated_channel = ast_channel_get_by_name(associated_name);
	if (associated_channel) {
		ast_queue_hangup(associated_channel);
		ast_channel_unref(associated_channel);
	}
}

static enum avs_device_type get_avs_device_type(struct ast_sip_session *session)
{
	if (!session) {
		return AVS_DEVICE_SIP;
	}

	switch (session->terminal_type) {
		case AST_TERMINAL_TYPE_WEBRTC:
			return AVS_DEVICE_WEBRTC;
		default:
			return AVS_DEVICE_SIP;
	}

	return AVS_DEVICE_SIP;
}

static void send_avs_slides_video_control(struct ast_sip_session *session, enum avs_ctrl_opt opt)
{
	struct ast_sip_session_media *slides_video = NULL;
	enum avs_device_type device_type = 0;

	if (!session || !session->channel) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_avs_slides_video_control for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	slides_video = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if (!slides_video || !slides_video->rtp) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get slides video media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	device_type = get_avs_device_type(session);

	ast_avs_content_control(session->channel, slides_video->rtp, opt, device_type);

	return;
}

static void get_video_resource_for_ivr(struct ast_sip_session *session)
{
	int resolution = -1;
	struct ast_stream *main_video = NULL;

	main_video = ast_sip_session_get_active_main_video_stream(session);
	if (!if_has_valid_video_stream(main_video)) {
		return;
	}

	resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, AST_STREAM_VIDEO_RESOLUTION_360P);
	if ((resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) || (resolution == -1)) {
		ast_log(LOG_DEBUG, "[session:%p] get_video_resource_for_ivr can't get video resource, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		/* AST_CONTROL_WAIT_JOIN_CONFERENCE this frame is used to close camera for wait join conference, just multiplex for this case */
		session->close_camera_forcely = 1;
		/* wave init conference will cost some time, just delay 2s to send this info, or will ignore this info,
		 * but this can't complete resovle this issue, need wave to support later.
		 */
		sleep(2);
		ast_indicate(session->channel, AST_CONTROL_WAIT_JOIN_CONFERENCE);
	}

	return;
}

static void change_to_conference_for_ivr(struct ast_sip_session *session)
{
	struct conf_notify_switch_data data;

	if (!session || !session->channel || ast_strlen_zero(session->conf_room)) {
		return;
	}

	memset(&data, 0, sizeof(data));
	snprintf(data.conf_id, sizeof(data.conf_id), "%s", session->conf_room);
	data.is_video_conf = 1;
	data.enabled_cei = 1;
	data.is_admin = 0;

	/* Just use nway process */
	ast_indicate_data(session->channel, AST_CONTROL_NOTIFY_N_WAY_CONF, &data, sizeof(data));

	return;
}

static void process_join_conference_from_special_feature(struct ast_sip_session *session)
{
	const char *from_ivr = NULL;
	const char *from_disa = NULL;
	const char *from_announcement = NULL;

	if (!session || !session->channel) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	from_ivr = pbx_builtin_getvar_helper(session->channel, "FROM_IVR");
	from_disa = pbx_builtin_getvar_helper(session->channel, "FROM_DISA");
	from_announcement = pbx_builtin_getvar_helper(session->channel, "ANNOUNCEMENT_FLAG");

	if (ast_strlen_zero(from_ivr) && ast_strlen_zero(from_disa) && ast_strlen_zero(from_announcement)) {
		return;
	}

	if ((from_ivr && !strcasecmp(from_ivr, "yes")) ||
		(from_disa && !strcasecmp(from_disa, "yes")) ||
		(from_announcement && !strcasecmp(from_announcement, "yes"))) {
		ast_log(LOG_DEBUG, "[session:%p] process_join_conference_from_special_feature for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));

		change_to_conference_for_ivr(session);
		get_video_resource_for_ivr(session);
		ast_sip_session_notfiy_trunk_info(session);//added by zyxia for Bug 267028 ,2023.7.31
	}
	return;
}

static void notify_open_share_audio(struct ast_sip_session *session)
{
	struct ast_app *app = NULL;
	char appdata[64] = {0};

	if (!session || !session->channel) {
		return;
	}

	if (!session->notify_share_audio) {
		ast_log(LOG_DEBUG, "[session:%p] notify_open_share_audio, notify_share_audio is disable, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(appdata, sizeof(appdata), "%s,%s", session->conf_room, "true");
	app = pbx_findapp("ConfBridgeFloorContrl");
	if (app) {
		ast_log(LOG_VERBOSE, "[session:%p] notify_open_share_audio launching %s(%s) on %s\n",
			session, "ConfBridgeFloorContrl", S_OR(appdata, ""), ast_sorcery_object_get_id(session->endpoint));
		pbx_exec(session->channel, app, S_OR(appdata, ""));
		session->notify_share_audio = 1;
	} else {
		ast_log(LOG_ERROR, "[session:%p] notify_open_share_audio, can't ConfBridgeFloorContrl APP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static int hangup_replace_channel(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_channel *, chan, NULL, ao2_cleanup);

	if (!session) {
		return -1;
	}

	if (ast_strlen_zero(session->replace_channel_id)) {
		ast_log(LOG_DEBUG, "[session:%p] hangup_replace_channel, can't get replace_channel_id, don't need hangup endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	chan = ast_channel_get_by_name(session->replace_channel_id);
	if (!chan) {
		ast_log(LOG_ERROR, "[session:%p] hangup_replace_channel, can't get chan from channel name:%s, endpoint:%s\n",
			session, session->replace_channel_id, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ast_channel_lock(chan);
	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(chan), AST_FLAG_ZOMBIE) || ast_check_hangup(chan)) {
		ast_channel_unlock(chan);
		return -1;
	}

	/*Modify for auto transfer moderator when switch to conf mode*/
	if (session->channel && chan) {
		pbx_builtin_setvar_helper(chan, "SWITCH_HOST", ast_channel_name(session->channel));
	}

	ast_log(LOG_DEBUG, "[session:%p] hangup_replace_channel, channel name:%s, endpoint:%s\n",
		session, session->replace_channel_id, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_hangupcause_set(chan, AST_CAUSE_SWITCH_TO_CONF_MODE);

	ast_softhangup(chan, AST_CAUSE_SWITCH_TO_CONF_MODE);

	ast_channel_unlock(chan);

	return 0;
}

static int join_sfu_control_frame_task(void *data)
{
	const char *is_nway_conf = NULL;
	const char *is_sfu = NULL;
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	char ipc_autopoen[8] = {0};
	int need_open_ipc=0;
	session = data;
	if (!session) {
		return -1;
	}

	if (!session->channel) {
		return -1;
	}


	is_nway_conf = pbx_builtin_getvar_helper(session->channel, "NWAY_CONF_NUMBER");
	is_sfu = pbx_builtin_getvar_helper(session->channel, "IS_SFU");

	if(session->channel &&  strstr(ast_channel_name(session->channel),"PJSIP/trunk") && ast_channel_get_is_ipc_call(session->channel)){

		ast_safe_nvram_get(":ipc_auto_open", ipc_autopoen, sizeof(ipc_autopoen));
		if ( ipc_autopoen != NULL && ipc_autopoen[0] != '\0')
			need_open_ipc = atoi(ipc_autopoen);

		if(need_open_ipc)
			ast_indicate(session->channel, AST_CONTROL_OPEN_IPC_VIDEO);
	}

	if (!ast_strlen_zero(is_nway_conf) && !ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
		process_nway_join_confenerce(session);
	} else {
		update_sfu_params_for_join_conference(session);
		process_join_conference_from_special_feature(session);
		ast_sip_session_notify_sfu_bfcp_for_join_conference(session);
	}

	if (session->inv_session && session->inv_session->role == PJSIP_ROLE_UAS && session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) {
		ast_sip_session_cancel_pending_call(session);
	}

	ast_sip_session_set_main_video_stream_uptime(session);

	notify_open_share_audio(session);

	ast_sip_session_notfiy_trunk_info(session);

	hangup_replace_channel(session);

	return 0;
}

static void process_join_sfu_control_frame(struct ast_sip_session *session)
{
	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, join_sfu_control_frame_task, session)) {
		ast_log(LOG_DEBUG, "[session:%p] Push join_sfu_control_frame_task to task failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}

static int conf_end_control_frame_task(void *data)
{
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);

	session = data;
	if (!session) {
		return -1;
	}

	if (!session->channel) {
		return -1;
	}

	if (session->inv_session
		&& session->inv_session->role == PJSIP_ROLE_UAS
		&& session->inv_session->state != PJSIP_INV_STATE_DISCONNECTED) {
		ast_sip_session_cancel_conf_invitee(session);
	}

	return 0;
}

static void process_conf_end_control_frame(struct ast_sip_session *session)
{
	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, conf_end_control_frame_task, session)) {
		ast_log(LOG_DEBUG, "[session:%p] Push process_conf_end_control_frame to task failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}

static int process_real_answered_control_frame(void *session)
{
	ast_sip_session_notfiy_answered_status(session);
	ao2_cleanup(session);
	return 0;
}

static int generate_notify_and_send(void* data)
{
	pjsip_tx_data* tdata = NULL;
	pj_str_t hdr_name = {"Allow-Events", 12};
	pj_str_t hdr_value = {"conference, presence, dialog, gs-nway-info, call-completion, refer", 66};
	pjsip_event_hdr* hdr_event = NULL;
	pjsip_generic_string_hdr* hdr_allowevent = NULL;

	char buf[512] = {0};
	int buf_cursor = 0;
	struct ast_sip_session* session = NULL;
	RAII_VAR(struct conf_notify_info_data*, notify_info_data, data, ao2_cleanup);

	if (!notify_info_data) {
		return -1;
	}

	session = notify_info_data->session;

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			session->inv_session->cause,
			pjsip_get_status_text(session->inv_session->cause)->ptr);
		return -1;
	}

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n");

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor, "<NwayControl>\r\n");

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor,
						"  <ConfID>%s</ConfID>\r\n", notify_info_data->conf_id);

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor,
						"  <ConfType>%s</ConfType>\r\n", notify_info_data->is_video_conf == 1 ? "video" : "audio");

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor,
						"  <ChannelID>%s</ChannelID>\r\n", ast_channel_name(session->channel));

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor,
						"  <Moderator>%d</Moderator>\r\n", notify_info_data->is_admin);

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor,
						"  <EnabledCEI>%d</EnabledCEI>\r\n", notify_info_data->enabled_cei);

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor,
						"  <EventType>%s</EventType>\r\n", notify_info_data->event_type == 1 ? "callflip" : "nway");

	buf_cursor += snprintf(buf + buf_cursor, sizeof(buf) - buf_cursor, "</NwayControl>");


	ast_log(LOG_NOTICE, "Generate notify body: %s\n", buf);

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-grandstream-nway-info+xml",
		.body_text = buf
	};

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create notify request\n");
		return -1;
	}

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body for notify request\n");
		return -1;
	}

	hdr_event = pjsip_event_hdr_create(tdata->pool);
	if (hdr_event) {
		pj_cstr(&hdr_event->event_type, "gs-nway-info");
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr*)hdr_event);
	}

	hdr_allowevent = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_allowevent);

	ast_sip_session_send_request(session, tdata);

	return 0;
}

static struct conf_notify_info_data* notify_info_data_alloc2(struct ast_sip_session* session, struct conf_notify_switch_data* data)
{
	struct conf_notify_info_data* notify_info_data = ao2_alloc(sizeof(*notify_info_data), notify_info_data_destroy);

	if (!notify_info_data || !data) {
		return NULL;
	}

	ao2_ref(session, +1);
	notify_info_data->session = session;
	notify_info_data->notify_info = NULL;
	notify_info_data->is_video_conf = data->is_video_conf;
	notify_info_data->is_admin = data->is_admin;
	notify_info_data->enabled_cei = data->enabled_cei;
	notify_info_data->event_type = data->event_type;
	memcpy(notify_info_data->conf_id, data->conf_id, sizeof(notify_info_data->conf_id));

	return notify_info_data;
}

static void send_notify_to_checkout_conf(struct ast_sip_session* session, const void* data, int datalen)
{
	struct conf_notify_switch_data* s_data = NULL;
	struct conf_notify_info_data* notify_info_data = NULL;

	if (!session || !session->channel || !data || datalen <= 0) {
		return;
	}

	if (datalen != sizeof(struct conf_notify_switch_data)) {
		return;
	}

	if (!strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		return;
	}

	if (session->join_sfu || session->conf_invite_call) {//edit  by zyxia for Bug 271949,2023.9.15
		ast_log(LOG_NOTICE, "[session:%p] Already joined SFU or Join from conf invite call, don't need notify again, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	s_data = (struct conf_notify_switch_data*)data;

	const char* is_n_way_chan = pbx_builtin_getvar_helper(session->channel, "NWAY_CONF_NUMBER");
	const char *from_ivr = pbx_builtin_getvar_helper(session->channel, "FROM_IVR");
	const char *from_disa = pbx_builtin_getvar_helper(session->channel, "FROM_DISA");
	/*Start added by zyxia for Bug 202171,call flip in conference need send notify*/
	const char* from_call_flip = pbx_builtin_getvar_helper(session->channel, "FROM_CALL_FLIP");
	const char *from_announcement = pbx_builtin_getvar_helper(session->channel, "ANNOUNCEMENT_FLAG");
	if (!ast_strlen_zero(from_call_flip) && strcasecmp(from_call_flip, "yes") == 0) {
		ast_log(LOG_NOTICE, "send_notify_to_checkout_conf the channel is call flip channel%s.",ast_channel_name(session->channel));
	}
	/*End added by zyxia for Bug 202171,call flip in conference need send notify*/
	else if (!ast_strlen_zero(from_disa)) {
		ast_log(LOG_NOTICE, "send_notify_to_checkout_conf the call from disa.");
	} else if (!ast_strlen_zero(from_announcement) && !strcasecmp(from_announcement, "yes")) {
		ast_log(LOG_NOTICE, "send_notify_to_checkout_conf the call from announcement.");
	} else if (ast_strlen_zero(is_n_way_chan) && ast_strlen_zero(from_ivr)) {
		ast_log(LOG_NOTICE, "send_notify_to_checkout_conf the channel is not n way channel.");
		return;
	}

	notify_info_data = notify_info_data_alloc2(session, s_data);
	if (!notify_info_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc conf notify for no CEI data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_push_task(session->serializer, generate_notify_and_send, notify_info_data)) {
		ao2_cleanup(notify_info_data);
	}
}

static char *get_conf_timeout_notify_status(int status)
{
	switch (status) {
	case EXTEND_PENDING:
		return "pending";
	case EXTEND_CONFIRMED:
		return "confirmed";
	case EXTEND_REJECTED:
		return "rejected";
	}
	return "";
}

static void conf_timeout_state_data_destroy(void *obj)
{
	struct ast_conf_timeout_state *conf_timeout_state = obj;

	if (!conf_timeout_state) {
		return;
	}
}

static void process_conf_timeout_notify_frame(struct ast_sip_session* session, const void* data, int datalen)
{
	struct conf_timeout_notify_data* s_data = NULL;
	struct ast_conf_timeout_state* conf_timeout_state = NULL;

	if (!session || !session->channel || !data || datalen <= 0) {
		ast_log(LOG_WARNING, "process_conf_timeout_notify_frame params is NULL\n");
		return;
	}

	if (datalen != sizeof(struct conf_timeout_notify_data)) {
		ast_log(LOG_WARNING, "[session:%p] process_conf_timeout_notify_frame data len is wrong, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	s_data = (struct conf_timeout_notify_data*)data;
	if (s_data) {
		conf_timeout_state = ao2_alloc(sizeof(*conf_timeout_state), conf_timeout_state_data_destroy);
		if (!conf_timeout_state) {
			ast_log(LOG_WARNING, "process_conf_timeout_notify_frame ao2_alloc error\n");
			return;
		}
		ast_copy_string(conf_timeout_state->conf, s_data->conf_id, sizeof(conf_timeout_state->conf));
		ast_copy_string(conf_timeout_state->type, s_data->notify_type ? "conf" : "user", sizeof(conf_timeout_state->type));
		ast_copy_string(conf_timeout_state->status, get_conf_timeout_notify_status(s_data->status), sizeof(conf_timeout_state->status));
		snprintf(conf_timeout_state->time, sizeof(conf_timeout_state->time), "%d", s_data->time);
		ast_log(LOG_DEBUG, "conf_timeout_state %s %s %s %s\n",conf_timeout_state->conf, conf_timeout_state->type, conf_timeout_state->status, conf_timeout_state->time);
	}
	ast_sip_session_send_notify_to_update_gs_state(session, AST_GS_STATE_TYPE_CONF_TIMEOUT, conf_timeout_state);

	if (conf_timeout_state) {
		ao2_cleanup(conf_timeout_state);
	}
}

static void process_open_focus_video_notify_frame(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	session->focus_video = 1;

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, process_open_focus_video_to_task, session)) {
		ast_log(LOG_WARNING, "[session:%p] process_open_focus_video_notify_frame to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}

static int process_close_focus_video_to_task(void *data)
{
	RAII_VAR(char *, slider, NULL, ast_free);
	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);

	if (!session) {
		ast_log(LOG_DEBUG, "[session:%p] process_open_focus_video_to_task session is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	slider = ast_conf_get_slider(session->conf_room);
	if (slider) {
		ast_log(LOG_DEBUG, "[session:%p] Conference is opening slides video, don't send avs focus video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);

	return 0;
}

static unsigned int judge_if_main_video_stream_use_360p_resolution(struct ast_sip_session *session)
{
	unsigned int i = 0;
	unsigned int res = 0;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream *stream = NULL;

	ast_channel_lock(session->channel);

	topology = ast_channel_get_stream_topology(session->channel);
	if (!topology) {
		ast_channel_unlock(session->channel);
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (ast_stream_is_enabled_main_video(stream)) {
			if (ast_stream_get_video_resolution(stream) == AST_STREAM_VIDEO_RESOLUTION_360P) {
				res = 1;
			}
			break;
		}
	}

	ast_channel_unlock(session->channel);

	return res;
}

static void change_video_resolution_for_close_focus(struct ast_sip_session *session)
{
	if (!ast_channel_has_enabled_main_video_stream(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Disable focus video, and has no enabled or hd video, release video resource, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_conf_put_main_video_resource(session->conf_room, session->channel);
		return;
	}

	/* If focus actually use 360P, need release 1080P and alloc 360P instead */
	if (judge_if_main_video_stream_use_360p_resolution(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Disable focus video, video stream actually use 360P, change to 360P, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_conf_put_main_video_resource(session->conf_room, session->channel);
		if (ast_conf_get_main_video_resource(session->conf_room, session->channel, AST_STREAM_VIDEO_RESOLUTION_360P) != AST_STREAM_VIDEO_RESOLUTION_360P) {
			ast_log(LOG_ERROR, "[session:%p] Can't get 360P video resolution resource, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	return;
}

static void process_close_focus_video_notify_frame(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	session->focus_video = 0;

	change_video_resolution_for_close_focus(session);

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable video recording, don't need process this frame, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, process_close_focus_video_to_task, session)) {
		ast_log(LOG_WARNING, "[session:%p] process_close_focus_video_notify_frame to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}

static void process_disable_focus_video_notify_frame(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	session->focus_video = 0;

	change_video_resolution_for_close_focus(session);

	return;
}

static int create_sdp_for_in_dialog_gs_state_notify(struct ast_sip_session *session, char *content, int len)
{
	char buf[256] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<state version=\"1\">\r\n"
	"  <item name=\"state\">\r\n"
	"    <record>\r\n");

	switch (session->record_stop_type) {
		case AST_RECORD_STOP_TYPE_DEVICE_OUT:
			snprintf(buf, sizeof(buf) - 1, "      <state>stop</state>\r\n");
			strncat(content, buf, len-strlen(content));

			snprintf(buf, sizeof(buf) - 1, "      <errorCode>10000</errorCode>\r\n");
			strncat(content, buf, len-strlen(content));
			break;
		default:
			break;
	}

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    </record>\r\n"
	"  </item>\r\n"
	"</state>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs state body[%s]\n", content);

	return 0;
}

static int send_in_dialog_gs_state_notify(void *data)
{
	struct pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;
	pj_str_t hdr_name = {"Allow-Events", 12};
	pj_str_t hdr_value = {"conference, presence, dialog, message-summary, call-completion, refer, x-gs-state", 81};
	pjsip_generic_string_hdr *hdr_allowevent = NULL;
	char sdp[4096] = {0};
	struct ast_sip_session *session = data;

	if (!session) {
		return -1;
	}

	if (session && session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_DEBUG, "Dialog is not established really, don't send in dialog gs state notify\n");
		ao2_ref(session, -1);
		return 0;
	}

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create in dialog gs state notify\n");
		ao2_ref(session, -1);
		return -1;
	}

	create_sdp_for_in_dialog_gs_state_notify(session, sdp, sizeof(sdp));

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-state+xml",
		.body_text = sdp
	};

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "x-gs-state");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	hdr_allowevent = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_allowevent);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to trickle ICE INFO request\n");
		ao2_ref(session, -1);
		return -1;
	}

	ast_sip_session_send_request(session, tdata);

	ao2_ref(session, -1);

	return 0;
}


static void process_stop_record_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	enum ast_record_stop_type type;

	if (!session) {
		return;
	}

	if (datalen != sizeof(enum ast_record_stop_type)) {
		ast_log(LOG_DEBUG, "process_stop_record_frame datalen is invalid\n");
		return;
	}

	type = *(enum ast_record_stop_type *)data;
	session->record_stop_type = type;

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_in_dialog_gs_state_notify, session)) {
		ao2_cleanup(session);
	}

	return;
}

static int create_slides_video_stream(struct ast_sip_session *session)
{
	struct ast_stream *stream = NULL;
	struct ast_format_cap *new_cap = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "create_slides_video_stream, has no pending media\n");
		return -1;
	}

	new_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!new_cap) {
		return -1;
	}

	if (ast_format_cap_append_from_cap(new_cap, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO)) {
		ao2_cleanup(new_cap);
		return -1;
	}

	stream = ast_stream_alloc(ast_codec_media_type2str(AST_MEDIA_TYPE_VIDEO), AST_MEDIA_TYPE_VIDEO);
	if (!stream) {
		ao2_cleanup(new_cap);
		return -1;
	}

	ast_stream_set_formats(stream, new_cap);
	ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
	ast_stream_set_local_media_direction(stream, SENDRECV);
	ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	ast_stream_set_has_nack_pli(stream, 1);
	ast_stream_set_has_ccm_fir(stream, 1);

	if (session->endpoint->enable_nack) {
		ast_stream_set_has_nack(stream, 1);
	}

	if (ast_stream_topology_append_stream(session->pending_media_state->topology, stream) < 0) {
		ao2_cleanup(new_cap);
		ast_stream_free(stream);
		return -1;
	}

	return 0;
}

static void send_reinvite_for_neg_slides_video(struct ast_sip_session* session)
{
	unsigned int i = 0;
	unsigned int has_bfcp = 0;
	unsigned int has_slides = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->channel) {
		return;
	}

	/* just re-negotiate WAVE slides video for nway conference */
	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC && session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		ast_log(LOG_DEBUG, "[session:%p] It's not wave, don't need re-negotiate slides video endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/*check the endpoint if have h264 codec,added by zyxia for bug 263761,2023.6.26*/
	if(session->endpoint && !ast_sip_endpoint_is_have_h264_codec(session->endpoint)){
		ast_log(LOG_DEBUG, "[session:%p] Dont have h264 codec, don't need re-negotiate slides video endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_session_judge_if_has_valid_slides_video(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Already has valid slides video stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->active_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Has no active media state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Already has pending media state, can't send re-invite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
	if (!session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Can't alloc pending media state topology, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); ++i) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP) {
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			session->bfcp_para.has_udp_bfcp = 1;
			has_bfcp = 1;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			session->bfcp_para.has_tcp_bfcp = 1;
			has_bfcp = 1;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			ast_stream_set_local_media_direction(stream, SENDRECV);
			has_slides = 1;
		}
	}

	if (!has_bfcp && ast_sip_session_support_bfcp(session)) {
		bfcp_create_stream(session);
	}

	if (!has_slides) {
		if (create_slides_video_stream(session)) {
			ast_log(LOG_DEBUG, "[session:%p] Can't create slides video stream, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
		ast_log(LOG_ERROR, "[session:%p] Can't push task to send reinvite for add slides video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}

static unsigned char judge_if_the_same_max_used_date(void)
{
	unsigned char res = FALSE;
	time_t now_time;
	struct tm *now_time_info = NULL;
	char now_date[32] = { 0 };
	char *max_used_date = NULL;

	time(&now_time);
	now_time_info = localtime(&now_time);

	snprintf(now_date, sizeof(now_date) -1, "%d%d%d", now_time_info->tm_year, now_time_info->tm_mon, now_time_info->tm_mday);

	max_used_date = ast_sip_get_max_used_date();
	if (!max_used_date) {
		ast_sip_set_max_used_date(now_date);
		return FALSE;
	}

	ast_log(LOG_DEBUG, "Now date:%s, max_used_date:%s\n", now_date, max_used_date);

	if (!strcmp(now_date, max_used_date)) {
		res = TRUE;
	} else {
		ast_sip_set_max_used_date(now_date);
		res = FALSE;
	}

	ast_free(max_used_date);
	max_used_date = NULL;

	return res;
}

static void update_max_used_turn_session_count(unsigned int all_used_count)
{
	if (ast_sip_get_day_max_used_turn_session_count() >= all_used_count &&
		judge_if_the_same_max_used_date()) {
		return;
	}

	ast_sip_set_day_max_used_turn_session_count(all_used_count);

	return;
}

#if 0
static unsigned char judge_if_the_same_alert_exceed_threshold_date(void)
{
	unsigned char res = FALSE;
	time_t now_time;
	struct tm *now_time_info = NULL;
	char now_date[32] = { 0 };
	char *exceed_threshold_date = NULL;

	time(&now_time);
	now_time_info = localtime(&now_time);

	snprintf(now_date, sizeof(now_date) -1, "%d%d%d", now_time_info->tm_year, now_time_info->tm_mon, now_time_info->tm_mday);

	exceed_threshold_date = ast_sip_get_alert_exceed_threshold_date();
	if (!exceed_threshold_date) {
		ast_sip_set_alert_exceed_threshold_date(now_date);
		return FALSE;
	}

	ast_log(LOG_DEBUG, "Now date:%s, alert_exceed_threshold_date:%s\n", now_date, exceed_threshold_date);

	if (!strcmp(now_date, exceed_threshold_date)) {
		res = TRUE;
	} else {
		ast_sip_set_alert_exceed_threshold_date(now_date);
		res = FALSE;
	}

	ast_free(exceed_threshold_date);
	exceed_threshold_date = NULL;

	return res;
}

static unsigned char judge_if_the_same_alert_exceed_total_date(void)
{
	unsigned char res = FALSE;
	time_t now_time;
	struct tm *now_time_info = NULL;
	char now_date[32] = { 0 };
	char *exceed_total_date = NULL;

	time(&now_time);
	now_time_info = localtime(&now_time);

	snprintf(now_date, sizeof(now_date) -1, "%d%d%d", now_time_info->tm_year, now_time_info->tm_mon, now_time_info->tm_mday);

	exceed_total_date = ast_sip_get_alert_exceed_total_date();
	if (!exceed_total_date) {
		ast_sip_set_alert_exceed_total_date(now_date);
		return FALSE;
	}

	ast_log(LOG_DEBUG, "Now date:%s, exceed_total_date:%s\n", now_date, exceed_total_date);

	if (!strcmp(now_date, exceed_total_date)) {
		res = TRUE;
	} else {
		ast_sip_set_alert_exceed_total_date(now_date);
		res = FALSE;
	}

	ast_free(exceed_total_date);
	exceed_total_date = NULL;

	return res;

}
#endif

static void ast_sip_report_exceed_threshold(void)
{
	struct ast_security_event_nat_call ast_nat_call = {
		.common.event_type = AST_SECURITY_EVENT_NAT_CALL,
		.common.version = AST_SECURITY_EVENT_EXCEED_NAT_CALL_THRESHOLD_VERSION,
		.common.service = "PJSIP",
		.type = AST_SECURITY_NAT_CALL_EVENT_EXCEED_THRESHOLD,
	};

	ast_log(LOG_DEBUG, "Report exceed threshold\n");

	ast_security_event_report(AST_SEC_EVT(&ast_nat_call));

	return;
}

static void ast_sip_report_exceed_total(void)
{
	struct ast_security_event_nat_call ast_nat_call = {
		.common.event_type = AST_SECURITY_EVENT_NAT_CALL,
		.common.version = AST_SECURITY_EVENT_EXCEED_NAT_CALL_THRESHOLD_VERSION,
		.common.service = "PJSIP",
		.type = AST_SECURITY_NAT_CALL_EVENT_EXCEED_TOTAL,
	};

	ast_log(LOG_DEBUG, "Report exceed total\n");

	ast_security_event_report(AST_SEC_EVT(&ast_nat_call));

	return;
}

static void alert_exceed_threshold(unsigned int all_used_count, unsigned int total_count)
{
	int threshold = 0;

	threshold = (all_used_count*1.0/total_count*1.0)*100;

	if (threshold < NAT_CALL_THRESHOLD_PERCENTAGE) {
		return;
	}

#if 0
	if (judge_if_the_same_alert_exceed_threshold_date()) {
		return;
	}
#endif

	ast_log(LOG_DEBUG, "alert_exceed_threshold threshold:%d, %d, %d\n", threshold, all_used_count, total_count);

	ast_sip_report_exceed_threshold();

	return;
}

static void alert_exceed_total(unsigned int all_used_count, unsigned int total_count)
{
	if (all_used_count < total_count) {
		return;
	}

#if 0
	if (judge_if_the_same_alert_exceed_total_date()) {
		ast_log(LOG_DEBUG, "Date is the same, don't need alert again\n");
		return;
	}
#endif

	ast_sip_report_exceed_total();

	return;
}

static void alert_for_external_call(unsigned int all_used_count, unsigned int total_count)
{
	alert_exceed_threshold(all_used_count, total_count);
	alert_exceed_total(all_used_count, total_count);

	return;
}

static int hangup_call_for_exceed_total_count(void *data)
{
	struct ast_sip_session *session = data;

	ast_log(LOG_DEBUG, "[session:%p] The call used turn session is reached the limited of license, hangup the call, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_lock(session->channel);

	ast_queue_control(session->channel, AST_CONTROL_TURN_SESSION_LIMITED);
	ast_sip_decrease_used_turn_session_count(ast_channel_get_turn_session_count(session->channel));
	ast_channel_set_turn_session_count(session->channel, 0);
	if (!ast_channel_stream(session->channel)) {
		ast_stream_and_wait(session->channel, AST_NAT_CALL_SESSION_LIMIT_PROMPT, "");
	}
	ast_channel_hangupcause_set(session->channel, AST_CAUSE_TURN_SESSION_LIMITED);
	ast_softhangup(session->channel, AST_CAUSE_TURN_SESSION_LIMITED);

	ast_channel_unlock(session->channel);

	ao2_cleanup(session);

	return 0;
}

static int hangup_call_for_exceed_total_limit_time(void *data)
{
	struct ast_sip_session *session = data;

	ast_log(LOG_DEBUG, "[session:%p] The call used turn session is reached the total limit time, hangup the call, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_lock(session->channel);

	if (!ast_channel_stream(session->channel)) {
		ast_stream_and_wait(session->channel, AST_NAT_CALL_TOTAL_TIME_LIMIT_PROMPT, "");
	} else {
		ast_log(LOG_WARNING, "[session:%p] hangup_call_for_exceed_total_limit_time, playing other prompt now, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_queue_control(session->channel, AST_CONTROL_TOTAL_TIME_LIMITED);
	ast_sip_decrease_used_turn_session_count(ast_channel_get_turn_session_count(session->channel));
	ast_channel_set_turn_session_count(session->channel, 0);

	ast_channel_hangupcause_set(session->channel, AST_CAUSE_TOTAL_TIME_LIMITED);
	ast_softhangup(session->channel, AST_CAUSE_TOTAL_TIME_LIMITED);

	ast_channel_unlock(session->channel);

	ao2_cleanup(session);

	return 0;
}

static void set_hangup_time(struct ast_sip_session *session, unsigned int left_time)
{
	struct timeval nat_call_time_limit;
	struct timeval nat_call_when_to_hangup;

	if (!session || !session->channel) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] set_hangup_time, left_time:%d, endpoint:%s\n",
		session, left_time, ast_sorcery_object_get_id(session->endpoint));

	nat_call_time_limit.tv_sec = left_time;
	nat_call_time_limit.tv_usec = 0;
	nat_call_when_to_hangup = ast_tvadd(ast_tvnow(), nat_call_time_limit);

	ast_channel_lock(session->channel);

	ast_channel_whentohangup_set(session->channel, &nat_call_when_to_hangup);

	ast_channel_unlock(session->channel);

	return;
}

static int play_prompt_for_call_time_limit(void *data)
{
	struct ast_sip_session *session = data;

	if (!session ) {
		return -1;
	}

	if (!session->channel) {
		ast_log(LOG_DEBUG, "[session:%p] play_prompt_for_call_time_limit, has no channel, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] play_prompt_for_call_time_limit, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_lock(session->channel);

	if (!ast_channel_stream(session->channel)) {
		ast_stream_and_wait(session->channel, AST_NAT_CALL_TIME_LIMIT_PROMPT, "");
	} else {
		ast_log(LOG_WARNING, "[session:%p] play_prompt_for_call_time_limit, is playing other prompt now, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_inc_rc_call_exceed_num();
	set_hangup_time(session, session->nat_call_left_time);

	ast_channel_unlock(session->channel);

	ao2_cleanup(session);

	return 0;
}

static int play_prompt_task(void *data)
{
	struct ast_sip_session *session = data;

	if (!session) {
		return 0;
	}

	if (!session->channel) {
		ast_log(LOG_DEBUG, "[session:%p] play_prompt_task, channel is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] play_prompt_task, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_lock(session->channel);

	if (!ast_channel_stream(session->channel)) {
		ast_stream_and_wait(session->channel, AST_NAT_CALL_TIME_LIMIT_PROMPT, "");
	} else {
		ast_log(LOG_WARNING, "[session:%p] play_prompt_task, is playing other prompt now, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_inc_rc_call_exceed_num();
	ast_channel_unlock(session->channel);

	ao2_cleanup(session);

	return 0;
}

static void schedule_task_for_play_prompt(struct ast_sip_session *session, unsigned int timeout)
{
	if (!session) {
		return;
	}

	if (session->play_sound_task) {
		ast_log(LOG_WARNING, "[session:%p] Has play prompt task before, don't need schedule again, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ao2_ref(session, +1);
	session->play_sound_task = ast_sip_schedule_task(session->serializer, timeout*1000,
		play_prompt_task, ast_taskprocessor_name(session->serializer),
		session, AST_SIP_SCHED_TASK_VARIABLE);
	if (!session->play_sound_task) {
		ast_log(LOG_WARNING, "[session:%p] schedule play prompt task failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}

/* If both UCM's trunk use remote RC, and play sound at the same time, AVS will not forward media when playsound
 * so if trunk's role is UAC, play sound early to avoid conflict.
 */
static void set_call_limit_prompt_for_trunk(struct ast_sip_session *session, unsigned int timeout)
{
	ast_log(LOG_DEBUG, "[session:%p] set_call_limit_prompt_for_trunk, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->inv_session->role == PJSIP_ROLE_UAC) {
		ast_log(LOG_DEBUG, "[session:%p] Trunk's invite session role is UAC, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (timeout > PROMPT_BEFORE_HANGUP_TIME) {
			ast_log(LOG_DEBUG, "[session:%p] Set trunk play sound timeout:%d, endpoint:%s\n",
				session, timeout - PROMPT_BEFORE_HANGUP_TIME, ast_sorcery_object_get_id(session->endpoint));
			ast_channel_nat_call_play_sound_timeout_set(session->channel, timeout - PROMPT_BEFORE_HANGUP_TIME);
		} else {
			ast_log(LOG_DEBUG, "[session:%p] It's trunk, and no much left time, play prompt tone now, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(session, +1);
			if (ast_sip_push_task(session->serializer, play_prompt_for_call_time_limit, session)) {
				ao2_cleanup(session);
			}
		}
	} else {
		ast_channel_nat_call_play_sound_timeout_set(session->channel, timeout);
	}

	return;
}

static void set_call_limit_prompt_without_bridge(struct ast_sip_session *session, unsigned int timeout)
{
	if (ast_channel_vm_get(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] It's VM, will not build bridge, schedule a task to play sound, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		schedule_task_for_play_prompt(session, timeout);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Channel has not in bridge, timeout:%d, set hook later, endpoint:%s\n",
			session, timeout, ast_sorcery_object_get_id(session->endpoint));
		if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
			set_call_limit_prompt_for_trunk(session, timeout);
		} else {
			ast_channel_nat_call_play_sound_timeout_set(session->channel, timeout);
		}
	}

	return;
}

static void set_call_limit_prompt(struct ast_sip_session *session, unsigned int current_limit_time)
{
	unsigned int timeout = 0;

	/* If left call time is large than 5s, set channel when to play limit sound flag, other play limit sound now */
	if (current_limit_time > PROMPT_BEFORE_HANGUP_TIME) {
		set_hangup_time(session, current_limit_time);
		timeout = current_limit_time - PROMPT_BEFORE_HANGUP_TIME;

		if (ast_channel_internal_bridge(session->channel)) {
			ast_log(LOG_DEBUG, "[session:%p] Channel has already in bridge, set hook here for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_bridge_nat_call_limit_interval_hook(session->channel, timeout);
		} else {
			set_call_limit_prompt_without_bridge(session, timeout);
		}
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Call limit time is less than 5s, play prompt tone, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->nat_call_left_time = current_limit_time;
		ao2_ref(session, +1);
		if (ast_sip_push_task(session->serializer, play_prompt_for_call_time_limit, session)) {
			ao2_cleanup(session);
		}
	}

	return;
}

static void set_channel_when_to_hangup(struct ast_sip_session *session)
{
	unsigned int nat_limit_time = 0;
	struct timeval nat_call_time_limit;
	struct timeval nat_call_when_to_hangup;
	struct timeval channel_hangup_time;
	unsigned int total_limit_time = 0;
	unsigned int used_limit_time = 0;
	unsigned int left_limit_time = 0;
	unsigned int current_limit_time = 0;
	unsigned int total_month_limit_time = 0;
	unsigned int used_month_limit_time = 0;

	if (!session || !session->channel) {
		return;
	}

	if (is_emergemcy_call(session)) {
		ast_log(LOG_DEBUG, "[session:%p] It's emergemcy call, don't set hangup time, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	nat_limit_time = ast_sip_get_turn_limit_time();
	if (!nat_limit_time) {
		ast_log(LOG_DEBUG, "[session:%p] Has no limit time for NAT call, don't need set for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	total_limit_time = ast_sip_get_turn_day_total_limit_time();
	used_limit_time = ast_cdr_get_nat_used_time();
	if (used_limit_time >= total_limit_time * 60) {
		ast_log(LOG_DEBUG, "[session:%p] used_limit_time is larger than total_limit_time, hangup the call, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, +1);
		if (ast_sip_push_task(session->serializer, hangup_call_for_exceed_total_limit_time, session)) {
			ao2_cleanup(session);
		}
		return;
	}

	total_month_limit_time = ast_sip_get_turn_month_total_limit_time();
	used_month_limit_time = ast_cdr_get_nat_month_used_time();
	if (total_month_limit_time && (used_month_limit_time >= total_month_limit_time * 60)) {
		ast_log(LOG_DEBUG, "[session:%p] used_month_limit_time is larger than total_month_limit_time, hangup the call, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, +1);
		if (ast_sip_push_task(session->serializer, hangup_call_for_exceed_total_limit_time, session)) {
			ao2_cleanup(session);
		}
		return;
	}

	ast_channel_lock(session->channel);

	left_limit_time = total_limit_time * 60 - used_limit_time;

	if (left_limit_time > nat_limit_time * 60) {
		current_limit_time = nat_limit_time * 60;
	} else {
		current_limit_time = left_limit_time;
	}

	if (total_month_limit_time && (current_limit_time > (total_month_limit_time * 60 - used_month_limit_time))) {
		ast_log(LOG_DEBUG, "[session:%p] Current month left time is less than day limit time, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		current_limit_time = total_month_limit_time * 60 - used_month_limit_time;
	}

	nat_call_time_limit.tv_sec = current_limit_time;
	nat_call_time_limit.tv_usec = 0;
	nat_call_when_to_hangup = ast_tvadd(ast_tvnow(), nat_call_time_limit);

	channel_hangup_time = *ast_channel_whentohangup(session->channel);

	if (ast_tvzero(channel_hangup_time) ||
		ast_tvcmp(channel_hangup_time, nat_call_when_to_hangup) > 0) {
		ast_log(LOG_DEBUG, "[session:%p] Channel has not set whentohangup or lager than nat call limit time, hangup call after:%ld seconds for endpoint:%s\n",
			session, nat_call_time_limit.tv_sec, ast_sorcery_object_get_id(session->endpoint));
		set_call_limit_prompt(session, current_limit_time);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Channel has already set hangup time, and is less than nat call limit, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_channel_unlock(session->channel);

	return;
}

static int process_update_turn_session_count_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	unsigned int session_count = 0;
	unsigned int all_used_count = 0;
	unsigned int total_count = 0;
	const struct avs_turn_session_count_string *count_string = NULL;

	if (!session || !data || !datalen) {
		return -1;
	}

	if (datalen != sizeof(struct avs_turn_session_count_string)) {
		ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_UPDATE_TURN_SESSION_COUNT. Expected %d, got %d for channel:%s\n",
			(int) sizeof(struct avs_turn_session_count_string), (int) datalen, ast_channel_name(session->channel));
		return -1;
	}

	if (session->rc_limit_already_processed) {
		ast_log(LOG_DEBUG, "[session:%p] RC limit already processed, don't need process again, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (is_emergemcy_call(session)) {
		ast_log(LOG_DEBUG, "[session:%p] It's emergemcy call, don't need limit by RC call, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	total_count = ast_sip_get_total_turn_session_count();
	if (total_count == 0) {
		ast_log(LOG_DEBUG, "[session:%p] Has no turn session count for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	count_string = data;

	if (strlen(count_string->turn_session_count) == 0) {
		ast_log(LOG_ERROR, "[session:%p] Can't get turn session count for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	session_count = atoi(count_string->turn_session_count);

	if (session_count > 0 && ast_channel_get_turn_session_count(session->channel) > 0) {
		ast_log(LOG_DEBUG, "[session:%p] Already get turn session count for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!count_string->ssh_proxy_flag && !ast_sip_get_max_media_nat_call_count()) {
		ast_log(LOG_DEBUG, "[session:%p] It's not from ssh proxy, and can't get max media nat call count for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (session_count == 0 && session->ssh_proxy_flag) {
		ast_log(LOG_DEBUG, "[session:%p] Receive AVS turn count is 0, but session is use ssh proxy, don't decrease, for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (ast_channel_get_turn_session_count(session->channel) > session_count) {
		all_used_count = ast_sip_decrease_used_turn_session_count(ast_channel_get_turn_session_count(session->channel) - session_count);
	} else {
		all_used_count = ast_sip_increase_used_turn_session_count(session_count - ast_channel_get_turn_session_count(session->channel));
	}

	ast_log(LOG_DEBUG, "[session:%p] Get turn session count:%d, total_count:%d, all_used_count:%d for endpoint:%s\n",
			session, session_count, total_count, all_used_count, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_set_turn_session_count(session->channel, session_count);

	set_channel_when_to_hangup(session);

	if (session_count > 0 && all_used_count <= total_count) {
		ast_channel_set_external_call_event(session->channel);
		ast_queue_control(session->channel, AST_CONTROL_PEER_IS_NAT_CALL);
	}

	if (all_used_count <= total_count) {
		update_max_used_turn_session_count(all_used_count);
	}

	alert_for_external_call(all_used_count, total_count);

	if (all_used_count > total_count) {
		ao2_ref(session, +1);
		if (ast_sip_push_task(session->serializer, hangup_call_for_exceed_total_count, session)) {
			ao2_cleanup(session);
		}
	}

	return 0;
}

static const char *camera_control_info_action_2str(enum ast_camera_control_action action)
{
	switch (action) {
		case CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_OPEN_OWN_CAMERA][0];
		case CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA][0];
		case CAMERA_CONTROL_ACTION_OPEN_OWN_CAMERA:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_OPEN_OWN_CAMERA][0];
		case CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA][0];
		case CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA_STATUS:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA_STATUS][0];
		case CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA_STATUS:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA_STATUS][0];
		case CAMERA_CONTROL_ACTION_OPEN_OWN_CAMERA_STATUS:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA_STATUS][0];
		case CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA_STATUS:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA_STATUS][0];
		case CAMERA_CONTROL_ACTION_REPORT_CAMERA_CAPABILITY:
			return camera_control_action_strmap[CAMERA_CONTROL_ACTION_REPORT_CAMERA_CAPABILITY][0];
		default:
			break;
	}

	ast_log(LOG_WARNING, "Can't process this action:%d\n", action);

	return NULL;
}

static const char *camera_control_info_status_2str(enum ast_camera_control_status status)
{
	switch (status) {
		case CAMERA_CONTROL_STATUS_SUCCEED:
			return camera_control_status_strmap[CAMERA_CONTROL_STATUS_SUCCEED][0];
		case CAMERA_CONTROL_STATUS_FAILED:
			return camera_control_status_strmap[CAMERA_CONTROL_STATUS_FAILED][0];
		default:
			break;
	}

	ast_log(LOG_WARNING, "Can't process this action:%d\n", status);

	return NULL;
}

static int camera_control_info_generate_body_content(const struct ast_camera_control_info *camera_control_info,char *content, int len)
{
	char buf[256] = {0};

	if (!camera_control_info || !content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	                   "<ConfControl>\r\n");

	snprintf(buf, sizeof(buf), "  <Action name=\"%s\">\r\n", camera_control_info_action_2str(camera_control_info->action));
	strncat(content, buf, len-strlen(content));

	if (camera_control_info->status > CAMERA_CONTROL_STATUS_UNKNOW &&
		camera_control_info->status < CAMERA_CONTROL_STATUS_MAX) {
		snprintf(buf, sizeof(buf), "    <Status>%s</Status>\r\n", camera_control_info_status_2str(camera_control_info->status));
		strncat(content, buf, len-strlen(content));
		if (strlen(camera_control_info->failed_reason)) {
			snprintf(buf, sizeof(buf), "    <FailedReason>%s</FailedReason>\r\n", camera_control_info->failed_reason);
			strncat(content, buf, len-strlen(content));
		}
	}

	strncat(content, "  </Action>\r\n", len-strlen(content));

	strncat(content, "</ConfControl>", len-strlen(content));

	ast_log(LOG_DEBUG, "Generate camera control info body[%s]\n", content);

	return 0;
}

static int camera_control_send_sip_info(void *data)
{
	struct pjsip_tx_data *tdata = NULL;
	char content[1024] = {0};
	struct ast_sip_session *session = NULL;
	const struct ast_camera_control_info *camera_control_info = NULL;

	struct camera_control_info_data *camera_control_data = data;

	if (!camera_control_data) {
		return -1;
	}

	session = camera_control_data->session;
	camera_control_info = camera_control_data->camera_control_info;

	if (!session || !camera_control_info) {
		ao2_ref(camera_control_data, -1);
		return -1;
	}

	if (session && session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_DEBUG, "Dialog is not established really, don't send sip info\n");
		ao2_ref(camera_control_data, -1);
		return 0;
	}

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text camera control INFO request\n");
		ao2_ref(camera_control_data, -1);
		return -1;
	}

	camera_control_info_generate_body_content(camera_control_info, content, sizeof(content));

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-grandstream-camera-confcontrol+xml",
		.body_text = content
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to camera control INFO request\n");
		ao2_ref(camera_control_data, -1);
		return -1;
	}
	ast_sip_session_send_request(session, tdata);

	ao2_ref(camera_control_data, -1);

	return 0;
}

static void camera_control_data_destroy(void *obj)
{
	struct camera_control_info_data *camera_control_data = obj;

	if (!camera_control_data) {
		return;
	}

	if (camera_control_data->camera_control_info) {
		ast_free(camera_control_data->camera_control_info);
		camera_control_data->camera_control_info = NULL;
	}

	if (camera_control_data->session) {
		ao2_ref(camera_control_data->session, -1);
	}

	return;
}

static struct camera_control_info_data *camera_control_data_alloc(struct ast_sip_session *session,
	const struct ast_camera_control_info *camera_control_info)
{
	struct camera_control_info_data *camera_control_data = ao2_alloc(sizeof(*camera_control_data), camera_control_data_destroy);

	if (!camera_control_data) {
		return NULL;
	}

	camera_control_data->camera_control_info = (struct ast_camera_control_info *)ast_malloc(sizeof(struct ast_camera_control_info));
	if (!camera_control_data->camera_control_info) {
		ao2_ref(camera_control_data, -1);
		return NULL;
	}
	memcpy(camera_control_data->camera_control_info, camera_control_info, sizeof(struct ast_camera_control_info));

	ao2_ref(session, +1);
	camera_control_data->session = session;

	return camera_control_data;
}

static void send_camera_control_info(struct ast_sip_session *session, const struct ast_camera_control_info *camera_control_info)
{
	struct camera_control_info_data *camera_control_data = NULL;

	if (!session || !camera_control_info) {
		return;
	}

	camera_control_data = camera_control_data_alloc(session, camera_control_info);
	if (!camera_control_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc camera control data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_push_task(session->serializer, camera_control_send_sip_info, camera_control_data)) {
		ao2_cleanup(camera_control_data);
	}

	return;
}

static void append_conf_video_codecs_to_caps(struct ast_sip_session *session, struct ast_format_cap *caps)
{
	if (!session || !caps) {
		return;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			append_video_codec_to_caps(session, caps, ast_format_h264);
			break;
		case AST_SFU_VIDEO_H265:
			append_video_codec_to_caps(session, caps, ast_format_h265);
			break;
		case AST_SFU_VIDEO_VP9:
			append_video_codec_to_caps(session, caps, ast_format_vp9);
			break;
		default:
			append_video_codec_to_caps(session, caps, ast_format_h264);
			break;
	}

	return;
}


static int create_new_main_video_stream_append(struct ast_sip_session *session, struct ast_stream_topology *topology,enum ast_stream_media_direction media_direction)
{
	struct ast_stream *stream = NULL;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate capabilities\n");
		return -1;
	}

	append_conf_video_codecs_to_caps(session, caps);

	stream = ast_stream_alloc(ast_codec_media_type2str(AST_MEDIA_TYPE_VIDEO), AST_MEDIA_TYPE_VIDEO);
	if (!stream) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ast_stream_set_formats(stream, caps);

	if (session->endpoint->enable_nack) {
		ast_stream_set_has_nack(stream, 1);
	}

	ast_stream_set_has_nack_pli(stream, 1);
	ast_stream_set_has_ccm_fir(stream, 1);
	ast_stream_set_metadata(stream, "CONTENT", "main");
	ast_stream_set_sub_type(stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	ast_stream_set_local_media_direction(stream, media_direction);
	switch (media_direction) {
		case SENDRECV:
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			break;
		case RECVONLY:
			ast_stream_set_state(stream, AST_STREAM_STATE_RECVONLY);
			break;
		case SENDONLY:
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDONLY);
			break;
		case INACTIVE:
			ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
			break;
		default:
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			break;
		}

	if (ast_stream_topology_append_stream(topology, stream) < 0) {
		ast_stream_free(stream);
		ast_log(LOG_ERROR, "[session:%p] Can't append stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	return 0;
}

static int active_main_video_stream(struct ast_sip_session *session, struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	unsigned int find = 0;
	struct ast_stream *stream = NULL;

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
			continue;
		}

		find = 1;
		ast_stream_set_local_media_direction(stream, SENDRECV);
		ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
		break;
	}

	if (!find) {
		if (create_new_main_video_stream_append(session, topology,SENDRECV)) {
			ast_log(LOG_ERROR, "[session:%p] Can't create new main video stream, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}
	}

	return 0;
}

static int disable_main_video_stream(struct ast_sip_session *session, struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
			continue;
		}

		ast_stream_set_local_media_direction(stream, SENDONLY);
		break;
	}

	return 0;
}

static void send_camera_control_reinvite(struct ast_sip_session *session,
	const struct ast_camera_control_info *camera_control_info)
{
	if (!session || !camera_control_info || !session->pending_media_state) {
		ast_log(LOG_ERROR, "send_camera_control_reinvite param error\n");
		return;
	}

	if (session->pending_media_state->topology) {
		ast_log(LOG_ERROR, "[session:%p] There already pending topology, can't send reinvite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->active_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
	} else {
		ast_log(LOG_ERROR, "[session:%p] There's no active topology, can't send reinvite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->pending_media_state->topology) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc pending topology, can't send reinvite, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	switch (camera_control_info->action) {
		case CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA:
			session->camera_control_action = AST_SIP_CAMERA_CONTROL_ACTION_OPEN;
			if (active_main_video_stream(session, session->pending_media_state->topology)) {
				ast_log(LOG_ERROR, "[session:%p] Can't active main video stream, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				ast_sip_session_media_state_reset(session->pending_media_state);
				return;
			}
			break;
		case CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA:
			session->camera_control_action = AST_SIP_CAMERA_CONTROL_ACTION_CLOSE;
			disable_main_video_stream(session, session->pending_media_state->topology);
			break;
		default:
			ast_log(LOG_ERROR, "[session:%p] Can't process this camera action:%d now, endpoint:%s\n",
				session, camera_control_info->action, ast_sorcery_object_get_id(session->endpoint));
			ast_sip_session_media_state_reset(session->pending_media_state);
			return;
	}

	ao2_ref(session, +1);
	session->is_outgoing = 1;

	if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_camera_control_reinvite push task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_media_state);
		ao2_cleanup(session);
	}

	return;
}

static unsigned int need_send_camera_control_info(struct ast_sip_session *session,
	const struct ast_camera_control_info *camera_control_info)
{
	ast_log(LOG_DEBUG, "[session:%p] Camera control action is :%d, endpoint:%s\n",
		session, camera_control_info->action, ast_sorcery_object_get_id(session->endpoint));

	if (!ast_strlen_zero(camera_control_info->participant_channel_id) &&
		strcmp(ast_channel_name(session->channel), camera_control_info->participant_channel_id)) {
		ast_log(LOG_DEBUG, "[session:%p] Channel ID is not the same as participant_channel_id:%s, endpoint:%s\n",
			session, camera_control_info->participant_channel_id, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (camera_control_info->action == CAMERA_CONTROL_ACTION_OPEN_OWN_CAMERA_STATUS ||
		camera_control_info->action == CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA_STATUS ||
		camera_control_info->action == CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA_STATUS ||
		camera_control_info->action == CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA_STATUS) {
		if (session->camera_control_operater) {
			session->camera_control_operater = 0;
			return 1;
		} else {
			return 0;
		}
	}

	return 1;
}

static int process_camera_control_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	const struct ast_camera_control_info *camera_control_info = NULL;

	if (!session || !data || !datalen) {
		return -1;
	}

	if (datalen != sizeof(struct ast_camera_control_info)) {
		ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_CAMERA_CONTROL. Expected %d, got %d for channel:%s\n",
			(int) sizeof(struct ast_camera_control_info), (int) datalen, ast_channel_name(session->channel));
		return -1;
	}

	camera_control_info = data;

	if (!need_send_camera_control_info(session, camera_control_info)) {
		return 0;
	}

	switch (session->terminal_type) {
		case AST_TERMINAL_TYPE_WEBRTC:
		case AST_TERMINAL_TYPE_GSWAVE_MOBILE:
			send_camera_control_info(session, camera_control_info);
			break;
		default:
			send_camera_control_reinvite(session, camera_control_info);
			break;
	}

	return 0;
}

static void process_camera_change_frame(struct ast_sip_session *session, char *host_msid)
{
	if (!session || !session->channel || !host_msid) {
		ast_log(LOG_WARNING, "update_msid_for_change_host params is NULL\n");
		return;
	}

	if (ast_sip_session_support_multiple_video(session)) {
		return;
	}

	update_msid_for_change_host(session, host_msid);

	return;
}

static const char *conf_role_info_2str(enum conf_user_role role)
{
	switch (role) {
		case CONF_USER_ROLE_ADMIN:
			return "admin";
		case CONF_USER_ROLE_PARTICIPANT:
			return "participant";
		default:
			/* Use participant as default */
			return "participant";
	}

	return NULL;
}

static int generate_update_cei_status_body_content(struct ast_sip_session *session,
	struct conf_notify_info *notify_info, char *content, int len)
{
	char buf[64] = {0};

	if (!session || !notify_info || !content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "CEI:no\r\n");

	snprintf(buf, sizeof(buf), "Role:%s\r\n", conf_role_info_2str(notify_info->role));
	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify body for conference has no CEI[%s]\n", content);

	return 0;
}

static int send_notify_to_update_cei_status(void *data)
{
	pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;
	pj_str_t hdr_name = {"Allow-Events", 12};
	pj_str_t hdr_value = {"conference, presence, dialog, message-summary, call-completion, refer, cei-status", 81};
	pjsip_generic_string_hdr *hdr_allowevent = NULL;
	char content[128] = { 0 };
	RAII_VAR(struct conf_notify_info_data *, notify_info_data, data, ao2_cleanup);
	struct ast_sip_session *session = NULL;
	struct conf_notify_info *notify_info = NULL;

	if (!notify_info_data) {
		return -1;
	}

	session = notify_info_data->session;
	notify_info = notify_info_data->notify_info;

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			session->inv_session->cause,
			pjsip_get_status_text(session->inv_session->cause)->ptr);
		return -1;
	}

	generate_update_cei_status_body_content(session, notify_info, content, sizeof(content));

	const struct ast_sip_body body = {
		.type = "application",
		.subtype = "conference-information",
		.body_text = content
	};

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text video update INFO request\n");
		return -1;
	}

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to text video update INFO request\n");
		return -1;
	}

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "cei-status");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	hdr_allowevent = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_allowevent);

	ast_sip_session_send_request(session, tdata);

	return 0;
}

static void notify_info_data_destroy(void *obj)
{
	struct conf_notify_info_data *notify_info_data = obj;

	if (!notify_info_data) {
		return;
	}

	if (notify_info_data->notify_info) {
		ast_free(notify_info_data->notify_info);
		notify_info_data->notify_info = NULL;
	}

	if (notify_info_data->session) {
		ao2_ref(notify_info_data->session, -1);
	}

	return;
}

static struct conf_notify_info_data *notify_info_data_alloc(struct ast_sip_session *session,
	const struct conf_notify_info *notify_info)
{
	struct conf_notify_info_data *notify_info_data = ao2_alloc(sizeof(*notify_info_data), notify_info_data_destroy);

	if (!notify_info_data) {
		return NULL;
	}

	notify_info_data->notify_info = (struct conf_notify_info *)ast_malloc(sizeof(struct conf_notify_info));
	if (!notify_info_data->notify_info) {
		ao2_ref(notify_info_data, -1);
		return NULL;
	}
	memcpy(notify_info_data->notify_info, notify_info, sizeof(struct conf_notify_info));

	ao2_ref(session, +1);
	notify_info_data->session = session;

	return notify_info_data;
}

static void send_conf_notify(struct ast_sip_session *session, const struct conf_notify_info *notify_info)
{
	struct conf_notify_info_data *notify_info_data = NULL;

	if (!session || !notify_info) {
		return;
	}

	notify_info_data = notify_info_data_alloc(session, notify_info);
	if (!notify_info_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc conf notify for no CEI data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_push_task(session->serializer, send_notify_to_update_cei_status, notify_info_data)) {
		ao2_cleanup(notify_info_data);
	}

	return;
}


static void process_notify_cei_status_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	if (!session || !session->channel) {
		ast_log(LOG_WARNING, "process_notify_cei_status_frame params is NULL\n");
		return;
	}

	if (datalen != sizeof(struct conf_notify_info)) {
		ast_log(LOG_WARNING, "[session:%p] process_notify_cei_status_frame data len is wrong, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE &&
		session->terminal_type != AST_TERMINAL_TYPE_WEBRTC &&
		strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_DEBUG, "[session:%p] process_notify_cei_status_frame endpoint:%s is not gswave mobile or trunk\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	struct conf_notify_info *notify_info = (struct conf_notify_info *)data;
	send_conf_notify(session, notify_info);

	return;
}

static unsigned int judge_if_has_video_resource(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_SENDRECV ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_RECVONLY) {
		return 1;
	}

	return 0;
}

static void send_info_indicate_close_camera(struct ast_sip_session *session)
{
	struct ast_camera_control_info camera_control_info;

	if (!session) {
		return;
	}

	memset(&camera_control_info, 0, sizeof(struct ast_camera_control_info));
	camera_control_info.action = CAMERA_CONTROL_ACTION_CLOSE_OWN_CAMERA;

	send_camera_control_info(session, &camera_control_info);

	return;
}

static int send_reinvite_for_session_refresh(void *data)
{
	struct ast_sip_session *session = data;

	if (!session) {
		ast_log(LOG_ERROR, "send_reinvite_for_session_refresh invalid session\n");
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_for_session_refresh endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->is_direct_media && !session->joint_ipvt) {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	} else {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 0, NULL);
	}

	ao2_ref(session, -1);

	return 0;
}

static void send_reinvite_close_camera(struct ast_sip_session *session, enum ast_stream_media_direction direction)
{
	struct ast_stream *stream = NULL;

	if (!session || !session->channel) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_close_camera for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) {
		ast_log(LOG_WARNING, "[session:%p] send_reinvite_close_camera inv state is not confirmed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->active_media_state|| !session->active_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] send_reinvite_close_camera for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] send_reinvite_close_camera already have pending topology, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
	if (!session->pending_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] send_reinvite_close_camera clone topology failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	stream = ast_sip_session_get_pending_main_video_stream(session);
	if (!judge_if_has_video_resource(stream)) {
		ast_log(LOG_DEBUG, "[session:%p] send_reinvite_close_camera has no video resource for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_media_state);
		return;
	}

	switch (direction) {
		case INACTIVE:
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
				ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
				ast_stream_set_local_media_direction(stream, INACTIVE);
				ast_stream_set_remote_media_direction(stream, INACTIVE);
			} else {
				ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
			}
			break;
		default:
			ast_stream_set_local_media_direction(stream, SENDONLY);
			break;
	}

	ao2_ref(session, +1);
	session->is_outgoing = 1;

	if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_reinvite_close_camera push task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_media_state);
		ao2_cleanup(session);
	}

	return;
}


static void close_camera_by_media_direction(struct ast_sip_session *session, enum ast_stream_media_direction direction)
{
	struct ast_stream *stream = NULL;

	if (!session || !session->channel) {
		ast_log(LOG_WARNING, "process_wait_join_conference_frame params is NULL\n");
		return;
	}

	stream = ast_sip_session_get_active_main_video_stream(session);
	if (!judge_if_has_video_resource(stream)) {
		return;
	}

	switch (session->terminal_type) {
		case AST_TERMINAL_TYPE_WEBRTC:
			send_info_indicate_close_camera(session);
			break;
		case AST_TERMINAL_TYPE_GSWAVE_MOBILE:
			if (ast_sip_session_media_use_bundle(session)) {
				send_info_indicate_close_camera(session);
			} else {
				send_reinvite_close_camera(session, direction);
			}
			break;
		default:
			send_reinvite_close_camera(session, direction);
			break;
	}

	return;
}

static void process_wait_join_conference_frame(struct ast_sip_session *session)
{
	close_camera_by_media_direction(session, INACTIVE);

	return;
}

static void process_close_camera_frame(struct ast_sip_session *session)
{
	if (ast_sip_session_get_only_support_audio()) {
		send_reinvite_close_camera(session, INACTIVE);
	} else {
		send_reinvite_close_camera(session, SENDONLY);
	}

	return;
}

static void process_open_slides_video_frame(struct ast_sip_session *session)
{
	if (!session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] It's not SFU, don't need process this frame, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->sfu_recv_slides_video = 1;

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not open video recording, don't need process this frame, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_SLIDES_VIDEO);

	return;
}

static void notify_close_share_audio(struct ast_sip_session *session)
{
	struct ast_app *app = NULL;
	char appdata[64] = {0};

	if (!session || !session->channel) {
		return;
	}

	if (!session->notify_share_audio) {
		return;
	}

	session->notify_share_audio = 0;

	if (ast_strlen_zero(session->conf_room)) {
		ast_log(LOG_ERROR, "[session:%p] notify_close_share_audio, can't get conf_room endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(appdata, sizeof(appdata), "%s,%s", session->conf_room, "false");
	app = pbx_findapp("ConfBridgeFloorContrl");
	if (app) {
		ast_log(LOG_VERBOSE, "[session:%p] notify_close_share_audio launching ConfBridgeFloorContrl(%s) on %s\n",
			session, S_OR(appdata, ""), ast_channel_name(session->channel));
		pbx_exec(session->channel, app, S_OR(appdata, ""));
	} else {
		ast_log(LOG_ERROR, "[session:%p] notify_close_share_audio, can't ConfBridgeFloorContrl APP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static unsigned int has_enabled_main_video_stream(struct ast_sip_session *session)
{
	struct ast_stream *stream = NULL;

	stream = ast_sip_session_get_active_main_video_stream(session);
	if (!stream) {
		ast_log(LOG_DEBUG, "[session:%p] has_enabled_main_video_stream has no active video stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (ast_stream_get_state(stream) != AST_STREAM_STATE_SENDRECV && ast_stream_get_state(stream) != AST_STREAM_STATE_RECVONLY) {
		ast_log(LOG_DEBUG, "[session:%p] has_enabled_main_video_stream not enable video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	return 1;
}

static int stop_slides_video_task(void *data)
{
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, focus_channel, NULL, ao2_cleanup);
	unsigned int focus_has_video = 0;

	session = data;
	if (!session) {
		return -1;
	}

	session->sfu_recv_slides_video = 0;

	ast_channel_set_opening_slides_video(session->channel, 0);

	notify_close_share_audio(session);

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not open video recording, don't need process this frame, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	focus_channel = ast_conf_get_sfu_focus(session->channel, session->conf_room);
	if (focus_channel) {
		focus_has_video = ast_channel_has_enabled_main_video_stream(focus_channel);
		ast_log(LOG_DEBUG, "[session:%p] Focus channel:%s, video enabled:%d, endpoint:%s\n",
			session, ast_channel_name(focus_channel), focus_has_video, ast_sorcery_object_get_id(session->endpoint));
	}

	if (!has_enabled_main_video_stream(session) && !focus_has_video) {
		ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);
		return 0;
	}

	if (open_focus_video_after_operate(session) && !focus_has_video) {
		ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);
	}

	return 0;
}

static void process_stop_slides_video_frame(struct ast_sip_session *session)
{
	if (!session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] It's not SFU, don't need process this frame, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, stop_slides_video_task, session)) {
		ast_log(LOG_ERROR, "[session:%p] Can't push task to stop slides video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}

static void sip_gs_state_info_data_destroy(void *obj)
{
	struct gs_state_info_data *gs_info_data = obj;

	if (!gs_info_data) {
		return;
	}
	if (gs_info_data->gs_state_data) {
		ast_free(gs_info_data->gs_state_data);
		gs_info_data->gs_state_data =NULL;
	}
	ao2_ref(gs_info_data->session, -1);

	return;
}

static struct gs_state_info_data *sip_gs_state_info_data_alloc(struct ast_sip_session *session, const struct ast_sip_gs_state_data *gs_state)
{
	struct gs_state_info_data *gs_info_data = ao2_alloc(sizeof(*gs_info_data), sip_gs_state_info_data_destroy);

	if (!gs_info_data) {
		return NULL;
	}

	gs_info_data->gs_state_data = (struct ast_sip_gs_state_data *)ast_malloc(sizeof(struct ast_sip_gs_state_data));
	if (!gs_info_data->gs_state_data) {
		ao2_ref(gs_info_data, -1);
		return NULL;
	}

	memcpy(gs_info_data->gs_state_data, gs_state, sizeof(struct ast_sip_gs_state_data));
	ao2_ref(session, +1);
	gs_info_data->session = session;

	return gs_info_data;
}

static int process_gs_state_frame_to_task(void *data)
{
	struct ast_sip_session *session = NULL;
	struct ast_sip_gs_state_data *gs_state = NULL;

	struct gs_state_info_data *gs_info_data = data;

	if (!gs_info_data) {
		return -1;
	}

	session = gs_info_data->session;
	gs_state = gs_info_data->gs_state_data;

	if (!session || !gs_state || ast_strlen_zero(session->conf_room) || !session->channel) {
		ao2_ref(gs_info_data, -1);
		return -1;
	}

	ast_log(LOG_DEBUG, "session->conf_room %s", session->conf_room);

	switch (gs_state->type) {
		case AST_GS_STATE_ITEM_TYPE_TERMINAL:
			ast_conf_mic_mute_unmute_helper(gs_state->muted, session->conf_room,  ast_channel_name(session->channel));
			break;
		case AST_GS_STATE_ITEM_TYPE_GESTURE:
			ast_conf_user_hand_up_helper(gs_state->hand_up, session->conf_room, gs_state->channel_name, session->channel);
			break;
		default:
			break;
	}

	ao2_ref(gs_info_data, -1);

	return 0;
}

static int process_gs_state_frame(struct ast_sip_session *session, const void *data, size_t datalen)
{
	struct gs_state_info_data *gs_info_data = NULL;
	const struct ast_sip_gs_state_data *gs_state = data;

	if (!session || !gs_state || !datalen) {
		return -1;
	}

	ast_log(LOG_DEBUG, "process_gs_state_frame");

	if (datalen != sizeof(struct ast_sip_gs_state_data)) {
		ast_log(LOG_ERROR, "Invalid datalen for AST_CONTROL_NOTIFY_GS_STATE. Expected %d, got %d for channel:%s\n",
			(int) sizeof(struct ast_sip_gs_state_data), (int) datalen, session->channel ? ast_channel_name(session->channel) : "");
		return -1;
	}

	gs_info_data = sip_gs_state_info_data_alloc(session, gs_state);
	if (!gs_info_data) {
		ast_log(LOG_ERROR, "[session:%p] Can't alloc gs state data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (ast_sip_push_task(session->serializer, process_gs_state_frame_to_task, gs_info_data)) {
		ast_log(LOG_WARNING, "[session:%p] process_bfcp_control_frame to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(gs_info_data);
	}

	return 0;
}

static int process_ami_hold_frame(struct ast_sip_session *session)
{
	if (!session) {
		return -1;
	}

	if (ast_sip_push_task(session->serializer, remote_send_hold, ao2_bump(session))) {
		ast_log(LOG_WARNING, "Could not queue task to remotely put session '%s' on hold with endpoint '%s'\n",
			ast_sorcery_object_get_id(session), ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return 0;
}

static int process_ami_unhold_frame(struct ast_sip_session *session)
{
	if (!session) {
		return -1;
	}

	if (ast_sip_push_task(session->serializer, remote_send_unhold, ao2_bump(session))) {
		ast_log(LOG_WARNING, "Could not queue task to remotely put session '%s' on hold with endpoint '%s'\n",
			ast_sorcery_object_get_id(session), ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return 0;
}
/*Start added by zyxia for Bug 215084,2022.2.26*/
#if 0
/*Start added by zyxia for callflip, slides,send open slides to peer chan.2021.12.10*/
static void send_reinvite_control_slides(struct ast_sip_session *local_session,struct ast_sip_session *session)
{
	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_control_slides for endpoint:%s,channel_name:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint),ast_channel_name(session->channel));

	if (!session || !local_session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_control_slides for endpoint:%s,channel_name:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint),ast_channel_name(session->channel));

	if (set_pending_media_stream_by_local_session(local_session,session) == 0){
		ao2_ref(session, +1);
		session->is_outgoing = 1;
		if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
			ast_log(LOG_DEBUG, "[session:%p] send_reinvite_close_camera push task failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_sip_session_media_state_reset(session->pending_media_state);
			ao2_cleanup(session);
		}
	}
	return;
}
#endif

static void send_reinvite_to_peer(struct ast_sip_session *session)
{
	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_to_peer for endpoint:%s,channel_name:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint),ast_channel_name(session->channel));

	if (!session ) {
		return;
	}

	ao2_ref(session, +1);
	session->is_outgoing = 1;
	if (ast_sip_push_task(session->serializer, send_reinvite_for_session_refresh, session)) {
		ast_log(LOG_DEBUG, "[session:%p] send_reinvite_to_peer push task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_media_state);
		ao2_cleanup(session);
	}

	return;
}

/*local_session, callflip channel session*/
/*remote_session, remote channel session*/
static int set_pending_media_stream_by_local_session(struct ast_sip_session *local_session,struct ast_sip_session *remote_session)
{
	struct ast_stream *stream = NULL;

	if (!local_session || !remote_session) {
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] set_pending_media_stream_direction  for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	if (!remote_session->pending_media_state->topology) {

		if (!remote_session->active_media_state|| !remote_session->active_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] set_pending_media_stream_direction for endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			return -1;
		}

		remote_session->pending_media_state->topology = ast_stream_topology_clone(remote_session->active_media_state->topology);
		if (!remote_session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] set_pending_media_stream_direction clone topology failed for endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			return -1;
		}
	}

	/*local have main ,remote dont have ,need append*/
	if (ast_sip_session_get_active_main_video_stream(local_session) ) {
		stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		if(!stream && remote_session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE){ //edit by zyxia for Bug 265085 ,app wave dont need append main video
			ast_log(LOG_WARNING, "[session:%p] ast_sip_session_get_pending_stream_by_subtype main video topology failed for endpoint:%s,create it\n",
			remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			ast_sip_session_append_stream_to_topology(remote_session, remote_session->pending_media_state->topology,NULL, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		}
	}

	stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if(!stream){
		ast_log(LOG_WARNING, "[session:%p] ast_sip_session_get_pending_stream_by_subtype slides video topology failed for endpoint:%s,create it\n",
		remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
		ast_sip_session_append_stream_to_topology(remote_session, remote_session->pending_media_state->topology,NULL, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	}

	if (ast_sip_session_get_active_slides_video_stream(local_session)) {
		stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
		if(stream){
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			ast_stream_set_local_media_direction(stream, SENDRECV);
			ast_stream_set_remote_media_direction(stream, SENDRECV);
		}
	}
	/*Start added by zyxia for Bug 248286,2023.1.17*/
	else {
		stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
		if(stream){
			ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
			ast_stream_set_local_media_direction(stream, INACTIVE);
			ast_stream_set_remote_media_direction(stream, INACTIVE);
		}
	}
	/*End added by zyxia for Bug 248286,2023.1.17*/

	/*Start added by zyxia for Bug 215280,2022.2.24*/
	if (!remote_session->support_ice && strstr(ast_channel_name(remote_session->channel), "PJSIP/trunk")) {
		stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP);
		if(!stream){
			ast_log(LOG_DEBUG, "[session:%p] get_pending_media_stream_direction  bfcp topology failed for endpoint:%s,need create stream\n",
			remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			bfcp_create_stream(remote_session);
		}
		else
		{
			ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
			ast_stream_set_local_media_direction(stream, SENDRECV);
			ast_stream_set_remote_media_direction(stream, SENDRECV);
			bfcp_create_params(remote_session);
			ast_log(LOG_DEBUG, "[session:%p] bfcp_create_params for endpoint:%s,channel_name:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint),ast_channel_name(remote_session->channel));
		}
	}
	/*End added by zyxia for Bug 215280,2022.2.24*/

	return 0;
}

/*Start added by zyxia for Bug 242037 ,2022.12.21*/
static int change_main_video_direction_by_local_session(struct ast_sip_session *local_session,struct ast_sip_session *remote_session){

	struct ast_stream *stream = NULL;
	struct ast_stream *local_stream = NULL;
	enum ast_stream_media_direction locale_stream_direction= INACTIVE;
	enum ast_stream_media_direction remote_stream_direction= INACTIVE;
	int ret = 0;
	int create_pending = 0;
	if (!local_session || !remote_session) {
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] change_main_video_direction_by_local_session  for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	if (!remote_session->pending_media_state->topology) {

		if (!remote_session->active_media_state|| !remote_session->active_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] set_pending_media_stream_direction for endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			return 0;
		}

		remote_session->pending_media_state->topology = ast_stream_topology_clone(remote_session->active_media_state->topology);
		if (!remote_session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] set_pending_media_stream_direction clone topology failed for endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			return 0;
		}
		create_pending = 1;
	}

	stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	local_stream = ast_sip_session_get_active_main_video_stream(local_session);
	/*local have main ,remote also have have,check the direction */
	if (local_stream && stream) {
		locale_stream_direction = ast_stream_get_local_media_direction(local_stream);
		remote_stream_direction = ast_stream_get_local_media_direction(stream);
		ast_log(LOG_WARNING, "[session:%p] change_main_video_direction_by_local_session local_dir:%d,remote_dir:%d\n",
		remote_session, locale_stream_direction,remote_stream_direction);

		/*change video fmtp param*/
		modify_video_fmtp_params_for_callflip(remote_session,local_session);

		if(locale_stream_direction == remote_stream_direction) {
			ast_log(LOG_WARNING, "[session:%p] change_main_video_direction_by_local_session remote main dir is same with local ,dont need update endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));

			ret = 0;
			goto back;
		}

		switch (locale_stream_direction) {
			case SENDRECV:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
				ast_stream_set_local_media_direction(stream, SENDRECV);
				ast_stream_set_remote_media_direction(stream, SENDRECV);
				break;
			case RECVONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDONLY);
				ast_stream_set_local_media_direction(stream, SENDONLY);
				ast_stream_set_remote_media_direction(stream, SENDONLY);
				break;
			case SENDONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_RECVONLY);
				ast_stream_set_local_media_direction(stream, RECVONLY);
				ast_stream_set_remote_media_direction(stream, RECVONLY);
				break;
			default:
				ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
				ast_stream_set_local_media_direction(stream, INACTIVE);
				ast_stream_set_remote_media_direction(stream, INACTIVE);
				break;
		}
		ret = 1;
	}
	//remove have ,but local dont have
	else if(stream && !local_stream) {
		ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
		ast_stream_set_local_media_direction(stream, INACTIVE);
		ast_stream_set_remote_media_direction(stream, INACTIVE);
		ret = 1;
	}

back:

	/* Modify by bfdu for Bug 311634, 2024.8.12 */
	if (0 == ret) {
		/*Start added by zyxia for Bug 264767 ,2023.7.10*/
		if (create_pending && remote_session->pending_media_state->topology) {
			ast_stream_topology_free(remote_session->pending_media_state->topology);
			remote_session->pending_media_state->topology = NULL;
		}
		/*End added by zyxia for Bug 264767 ,2023.7.10*/
	}

	return ret;
}

/*End added by zyxia for Bug 242037 ,2022.12.21*/

#if 0
/*Start added by zyxia for Bug 262467,2023.6.13 ,send reinvite to call chan to update slides direction*/
static int change_slides_video_direction_by_peer_session(struct ast_sip_session *local_session,struct ast_sip_session *remote_session){

	struct ast_stream *stream = NULL;
	struct ast_stream *local_stream = NULL;
	enum ast_stream_media_direction locale_stream_direction= INACTIVE;
	enum ast_stream_media_direction remote_stream_direction= INACTIVE;
	int ret = 0;
	int create_pending=0;
	if (!local_session || !remote_session) {
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] change_slides_video_direction_by_peer_session  for endpoint:%s\n",
		local_session, ast_sorcery_object_get_id(local_session->endpoint));

	if (!remote_session->pending_media_state->topology) {

		if (!remote_session->active_media_state|| !remote_session->active_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] change_slides_video_direction_by_peer_session for endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			return 0;
		}

		remote_session->pending_media_state->topology = ast_stream_topology_clone(remote_session->active_media_state->topology);
		if (!remote_session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] change_slides_video_direction_by_peer_session clone topology failed for endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			return 0;
		}
		create_pending =1;
	}

	stream = ast_sip_session_get_pending_stream_by_subtype(remote_session,AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	local_stream = ast_sip_session_get_active_slides_video_stream(local_session);
	/*local have main ,remote also have have,check the direction */
	if (local_stream && stream) {
		locale_stream_direction = ast_stream_get_local_media_direction(local_stream);
		remote_stream_direction = ast_stream_get_local_media_direction(stream);
		ast_log(LOG_WARNING, "[session:%p] change_slides_video_direction_by_peer_session local_dir:%d,remote_dir:%d\n",
		remote_session, locale_stream_direction,remote_stream_direction);

		/*change video fmtp param*/
		//modify_video_fmtp_params_for_callflip(remote_session,local_session);

		if (ast_stream_get_state(local_stream) == AST_STREAM_STATE_REMOVED) {
			ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
			ast_stream_set_local_media_direction(stream, INACTIVE);
			ast_stream_set_remote_media_direction(stream, INACTIVE);
			ast_log(LOG_WARNING, "[session:%p] change_slides_video_direction_by_peer_session remote slides is removed, inactive it endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));
			ret = 1;
			goto back;
		}

		if(locale_stream_direction == remote_stream_direction) {
			ast_log(LOG_WARNING, "[session:%p] change_slides_video_direction_by_peer_session remote main dir is same with local ,dont need update endpoint:%s\n",
				remote_session, ast_sorcery_object_get_id(remote_session->endpoint));

			/*Start added by zyxia for Bug 264767 ,2023.7.10*/
			if (create_pending && remote_session->pending_media_state->topology) {
				ast_stream_topology_free(remote_session->pending_media_state->topology);
				remote_session->pending_media_state->topology = NULL;
			}
			/*End added by zyxia for Bug 264767 ,2023.7.10*/

			ret = 0;
			goto back;
		}

		switch (locale_stream_direction) {
			case SENDRECV:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
				ast_stream_set_local_media_direction(stream, SENDRECV);
				ast_stream_set_remote_media_direction(stream, SENDRECV);
				break;
			case SENDONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDONLY);
				ast_stream_set_local_media_direction(stream, SENDONLY);
				ast_stream_set_remote_media_direction(stream, SENDONLY);
				break;
			case RECVONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_RECVONLY);
				ast_stream_set_local_media_direction(stream, RECVONLY);
				ast_stream_set_remote_media_direction(stream, RECVONLY);
				break;
			default:
				ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
				ast_stream_set_local_media_direction(stream, INACTIVE);
				ast_stream_set_remote_media_direction(stream, INACTIVE);
				break;
		}
		ret = 1;
	}
	//remove have ,but local dont have
	else if(stream && !local_stream) {
		ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
		ast_stream_set_local_media_direction(stream, INACTIVE);
		ast_stream_set_remote_media_direction(stream, INACTIVE);
		ret = 1;
	}

back:

	return ret;
}
/*End added by zyxia for Bug 262467,2023.6.13 ,send reinvite to call chan to update slides direction*/
#endif

/*Start added by zyxia for Bug 260870,2023.6.7*/
static int change_active_main_video_direction(struct ast_sip_session *session,enum ast_stream_media_direction stream_direction){

	struct ast_stream *stream = NULL;
	int ret = 0;
	if (!session ) {
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] change_active_main_video_direction  for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->pending_media_state->topology) {

		if (!session->active_media_state|| !session->active_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] change_active_main_video_direction for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 0;
		}

		session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
		if (!session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] change_active_main_video_direction clone topology failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 0;
		}
	}

	stream = ast_sip_session_get_pending_stream_by_subtype(session,AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	if (stream) {
		switch (stream_direction) {
			case SENDRECV:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDRECV);
				ast_stream_set_local_media_direction(stream, SENDRECV);
				ast_stream_set_remote_media_direction(stream, SENDRECV);
				break;
			case SENDONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_SENDONLY);
				ast_stream_set_local_media_direction(stream, SENDONLY);
				ast_stream_set_remote_media_direction(stream, SENDONLY);
				break;
			case RECVONLY:
				ast_stream_set_state(stream, AST_STREAM_STATE_RECVONLY);
				ast_stream_set_local_media_direction(stream, RECVONLY);
				ast_stream_set_remote_media_direction(stream, RECVONLY);
				break;
			default:
				ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
				ast_stream_set_local_media_direction(stream, INACTIVE);
				ast_stream_set_remote_media_direction(stream, INACTIVE);
				break;
		}
		ret = 1;
	}

	return ret;
}
/*End added by zyxia for Bug 260870,2023.6.7*/


#if 0
/* Now media_direction param is only support two options
 * media_direction: RECVONLY is indicate moderator leave, SENDRECV is indicate moderator join.
 */
static void set_final_local_media_direction(struct ast_stream *stream , enum ast_stream_media_direction media_direction)
{
	enum ast_stream_media_direction local_direction = UNKNOW;
	enum ast_stream_media_direction final_direction = SENDRECV;

	local_direction = ast_stream_get_local_media_direction(stream);

	switch (local_direction) {
		case SENDRECV:
			if (media_direction == RECVONLY) {
				final_direction = RECVONLY;
			} else if (media_direction == SENDRECV) {
				final_direction = SENDRECV;
			}
			break;
		case SENDONLY:
			if (media_direction == RECVONLY) {
				final_direction = INACTIVE;
			} else if (media_direction == SENDRECV) {
				final_direction = SENDONLY;
			}
			break;
		case INACTIVE:
			if (media_direction == RECVONLY) {
				final_direction = INACTIVE;
			} else if (media_direction == SENDRECV) {
				final_direction = SENDONLY;
			}
			break;
		case RECVONLY:
			if (media_direction == RECVONLY) {
				final_direction = RECVONLY;
			} else if (media_direction == SENDRECV) {
				final_direction = SENDRECV;
			}
			break;
		default:
			break;
	}

	ast_stream_set_local_media_direction(stream, final_direction);

	return;
}

static void set_main_video_stream_to_direction(struct ast_sip_session *session, enum ast_stream_media_direction media_direction)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		ast_log(LOG_WARNING, "set_all_stream_change_ssrc params error\n");
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}

		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}

		set_final_local_media_direction(stream, media_direction);

		ast_log(LOG_DEBUG, "[session:%p] Final set_main_video_stream_to_direction to %d for endpoint:%s\n",
			session, ast_stream_get_local_media_direction(stream), ast_sorcery_object_get_id(session->endpoint));
		break;
	}

	return;
}

static int send_reinvite_for_update_meida_direction(void *data)
{
	struct ast_sip_session *session = data;

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_for_update_meida_direction for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->is_direct_media && !session->joint_ipvt) {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, NULL);
	} else {
		ast_sip_session_refresh(session, NULL, remove_sdp_image_cb, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 0, NULL);
	}

	ao2_ref(session, -1);

	return 0;
}

static enum ast_stream_media_direction get_main_video_stream_local_direction(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	enum ast_stream_media_direction direction = UNKNOW;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return UNKNOW;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}

		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}

		direction = ast_stream_get_local_media_direction(stream);

		ast_log(LOG_DEBUG, "[session:%p] Get main video stream local direction %d for endpoint:%s\n",
			session, ast_stream_get_local_media_direction(stream), ast_sorcery_object_get_id(session->endpoint));
		break;
	}

	return direction;
}

static void send_reinvite_for_change_host(struct ast_sip_session *session, char *host_msid)
{
	enum ast_stream_media_direction main_video_local_direction = UNKNOW;

	if (!session || !host_msid) {
		return;
	}

	if (ast_sip_session_support_multiple_video(session)) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_for_change_host for endpoint:%s as host msid changed to %s\n",
		session, ast_sorcery_object_get_id(session->endpoint), host_msid);

	main_video_local_direction = get_main_video_stream_local_direction(session);

	if (ast_strlen_zero(host_msid)) {
		if (main_video_local_direction == RECVONLY || main_video_local_direction == INACTIVE) {
			ast_log(LOG_DEBUG, "[session:%p] Don't need send reinvite for change host, local direction:%d already can't send media, endpoint:%s\n",
				session, main_video_local_direction, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		set_main_video_stream_to_direction(session, RECVONLY);
	} else {
		if (main_video_local_direction == SENDRECV || main_video_local_direction == SENDONLY) {
			ast_log(LOG_DEBUG, "[session:%p] Don't need send reinvite for change host, local direction:%d already can send media, endpoint:%s\n",
				session, main_video_local_direction, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		set_main_video_stream_to_direction(session, SENDRECV);
	}

	ao2_ref(session, +1);
	session->is_outgoing = 1;

	if (ast_sip_push_task(session->serializer, send_reinvite_for_update_meida_direction, session)) {
		ast_log(LOG_DEBUG, "[session:%p] Push task failed for send reinvit as change host, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
	}

	return;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
/* added by zgmo for sdp media record attribute */
static int send_update_record_attr(void *data)
{
	struct ast_sip_session *session = data;
	enum ast_sip_session_refresh_method method;

	if (!session) {
		return -1;
	}

	if (!session->inv_session) {
		ao2_ref(session, -1);
		return -1;
	}

	if ((session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) || (ast_channel_state(session->channel) != AST_STATE_UP)) {
		ast_log(LOG_NOTICE, "Session state is[%d], ignore\n", session->inv_session->state);
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(session->inv_session);
#endif
		ao2_ref(session, -1);
		return -1;
	}

	method = session->endpoint->id.refresh_method;
	if (session->inv_session->options & PJSIP_INV_SUPPORT_UPDATE) {
		method = AST_SIP_SESSION_REFRESH_METHOD_UPDATE;
	}

	ast_sip_session_refresh(session, NULL, NULL, NULL, method, 1, NULL);

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(session->inv_session);
#endif

	ao2_ref(session, -1);
	return 0;

#if 0
	int i;
	int ret = -1;
	struct ast_sip_session *session = data;
	struct pjsip_tx_data *tdata = NULL;
	pjmedia_sdp_session *sdp_session;
	pjsip_inv_session *inv_session;
	pj_pool_t *pool;

	if (!session) {
		goto END;
	}

	if (session && session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_DEBUG, "Dialog is not established really, don't send sip update for record attr\n");
		ret = 0;
		goto END;
	}

	inv_session = session->inv_session;
	if (!inv_session || !inv_session->neg || !inv_session->neg->active_local_sdp) {
		return -1;
	}

	pool = inv_session->pool;
	if (!pool) {
		return -1;
	}

	sdp_session = inv_session->neg->active_local_sdp;
	if (!sdp_session) {
		return -1;
	}

	for (i = 0; i < sdp_session->media_count; ++i) {
		if (!strstr(sdp_session->media[i]->desc.media.ptr, "audio")) {
			continue;
		}

		change_audio_record_attr(session, sdp_session->media[i], pool);
	}

	ret = pjsip_inv_update(session->inv_session, NULL, session->inv_session->neg->active_local_sdp, &tdata);
	if (PJ_SUCCESS  == ret && tdata) {
		ast_sip_session_send_request(session, tdata);
		ret = 0;
	} else {
		ast_log(LOG_WARNING, "pjsip_inv_update fail, ret:%d, tdata:%p", ret, tdata);
	}
	ret = 0;

END:
	ao2_ref(session, -1);
	ast_channel_unref(session->channel);

	return ret;
#endif
}
/* end */
#endif

static void send_digit_with_rfc4733(struct ast_sip_session *session,
	char digit, unsigned int duration, struct ast_sip_session_media *media)
{
	int res = 0;
	struct ast_dtmf_send_param dtmfreq = {0};

	if (!session || !session->channel || !media || !media->rtp) {
		ast_log(LOG_WARNING, "send_digit_with_rfc4733 param is NULL\n");
		return;
	}

	if (ast_rtp_instance_dtmf_mode_get(media->rtp) == AST_RTP_DTMF_MODE_NONE) {
		ast_log(LOG_DEBUG, "[session:%p] DTMF mode is none, don't send digit:%d, duration:%d with rfc4733 for endpoint:%s\n",
			session, digit, duration, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_digit_with_rfc4733 digit:%d, duration:%d for endpoint:%s\n",
		session, digit, duration, ast_sorcery_object_get_id(session->endpoint));

	if (ast_opt_avs) {
		dtmfreq.digit = digit;
		dtmfreq.opt = 1;
		res = ast_avs_control(session->channel, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
		if (!res) {
			usleep(100000);
			dtmfreq.opt = 0;
			ast_avs_control(session->channel, AVS_CONTROL_DTMF, &dtmfreq, sizeof(dtmfreq));
		}
	} else {
		ast_rtp_instance_dtmf_end_with_duration(media->rtp, digit, duration);
	}

	return;
}

static int send_digit_with_sip_info(struct ast_sip_session *session,
	char digit, unsigned int duration)
{
	struct info_dtmf_data *dtmf_data = NULL;

	if (!session) {
		return -1;
	}

	dtmf_data = info_dtmf_data_alloc(session, digit, duration);
	if (!dtmf_data) {
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_digit_with_sip_info digit:%d, duration:%d for endpoint:%s\n",
		session, digit, duration, ast_sorcery_object_get_id(session->endpoint));

#ifdef HAVE_PJSIP_INV_SESSION_REF
	if (pjsip_inv_add_ref(dtmf_data->session->inv_session) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "[session:%p] Can't increase the session reference counter for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(dtmf_data);
		return -1;
	}
#endif

	if (ast_sip_push_task(session->serializer, transmit_info_dtmf, dtmf_data)) {
		ast_log(LOG_WARNING, "[session:%p] Error sending DTMF via INFO for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(dtmf_data->session->inv_session);
#endif
		ao2_cleanup(dtmf_data);
		return -1;
	}

	return 0;
}

static void send_digit_with_both_rfc4733_and_info(struct ast_sip_session *session,
	char digit, unsigned int duration, struct ast_sip_session_media *media)
{
	send_digit_with_rfc4733(session, digit, duration, media);

	send_digit_with_sip_info(session, digit, duration);

	return;
}

static void send_response_control_frame_to_peer(struct ast_sip_session *session)
{
	struct ast_sip_reinvite_data reinvite_data;

	ast_log(LOG_DEBUG, "[session:%p] send_response_control_frame_to_peer for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session) {
		return;
	}

	session->is_media_change = 0;
	session->recv_reinvite_no_sdp = 0;

	reinvite_data.session = session;
	ao2_ref(session, +1);
	if (ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_RSP, &reinvite_data, sizeof(reinvite_data))) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_REINVITE_RSP frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}

static int reload_module(void)
{
	globalEnabledCEI = ast_cei_get_enable_configuration();
	get_feature_code(1);

	ast_high_frequence_queue_reload();
	return 0;
}

static int chan_pjsip_recovery_throw_pbx(struct ast_channel *channel)
{
	int res;
	char buf[3072] = {0};
	char *passdata;
	const char *conf_room;

	if (!channel) {
		return -1;
	}

	if(0 == ast_recovery_db_get(ast_channel_recovery_linkedid(channel), ast_channel_recovery_uniqueid(channel), buf, sizeof(buf)-1))
	{
		ast_channel_lock(channel);
		/* restore database to xml struct, need free */
		switch_xml_t restore_xml = switch_xml_parse_str_dynamic(buf, SWITCH_TRUE);

		switch_xml_t channel_data = switch_xml_child(restore_xml, "variables");
		switch_xml_t data;
		/* restore app data for dialplan */
		data = switch_xml_child(channel_data, "app_data");
		passdata = ast_malloc(256);  /* equal with EXT_DATA_SIZE*/
		strncpy(passdata, data->txt,256);
		if(ast_channel_recovery_data(channel))
		{
			free((char *)ast_channel_recovery_data(channel));
		}
		ast_channel_recovery_data_set(channel, passdata);

		data = switch_xml_child(channel_data, "context");
		ast_channel_context_set(channel, data->txt);

		data = switch_xml_child(channel_data, "exten");
		ast_channel_exten_set(channel, data->txt);

		data = switch_xml_child(channel_data, "priority");
		ast_channel_priority_set(channel, atoi(data->txt));

		if (restore_xml) {
			switch_xml_free(restore_xml);
		}

		conf_room = pbx_builtin_getvar_helper((struct ast_channel *)channel, "CONF_ROOM");
		if(conf_room)
		{
			/* for cdr info correct, need in up state when sending snapshot to cdr */
			ast_setstate(channel, AST_STATE_UP);
		}

		ast_set_flag(ast_channel_flags(channel), AST_FLAG_RECOVERYING);
		ast_channel_unlock(channel);
	}
	else
	{
		ast_log(LOG_WARNING, "query recovery data failed chan:%s\n", ast_channel_name(channel));
		return -1;
	}

	res = ast_pbx_start(channel);

	switch (res) {
		case AST_PBX_FAILED:
			ast_log(LOG_WARNING, "Failed to start PBX ;(\n");
			ast_channel_hangupcause_set(channel, AST_CAUSE_SWITCH_CONGESTION);
			ast_hangup(channel);
			break;
		case AST_PBX_CALL_LIMIT:
			ast_log(LOG_WARNING, "Failed to start PBX (call limit reached) \n");
			ast_channel_hangupcause_set(channel, AST_CAUSE_SWITCH_CONGESTION);
			ast_hangup(channel);
			break;
		case AST_PBX_SUCCESS:
		default:
			break;
	}

	ast_debug(3, "Started PBX on recovery PJSIP channel %s\n", ast_channel_name(channel));

	return (res == AST_PBX_SUCCESS) ? 0 : -1;

}

/* note: need free entry pointer member when exit*/
static int chan_pjsip_start_recovery(void *data)
{
	int cause;
	struct ast_stream_topology *top = NULL;
	char dest[256] = {0};
	char addr[256] = {0};
	struct ast_channel *tc = NULL;
	int ret = 0;
	struct ast_party_connected_line connected;
	struct timeval time_val;
	struct ast_sip_channel_pvt *channel;
	struct ast_sip_session *session;
	struct ast_sip_recovery_entry *entry = data;
	char *sql = NULL;
	struct ast_party_caller caller;
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	struct ao2_iterator it_contacts;
	struct ast_sip_endpoint *endpoint;
	RAII_VAR(struct ast_sip_aor *, found_aor, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
	char isfound = 0;

	if(!data)
	{
		return -1;
	}

	/* for sip trunk */
	if(entry->peertrunk_aor[0] != '\0')
	{
		strcat(dest, entry->exten);
		strcat(dest, "@");
		strcat(dest, entry->peertrunk_aor);
		strcpy(addr, dest);
		strcat(dest, "/");
		strcat(dest, entry->peertrunk_aor);
		strcat(dest, "/");
		strcat(dest, entry->from_tag);
		strcat(dest, "/");
		strcat(dest, entry->to_tag);
		strcat(dest, "/");
		strcat(dest, entry->call_id);

	}
	else
	{
		if(entry->remote_ip)
		{
			aor = ast_sip_location_retrieve_aor(entry->exten);
			if (!aor) {
				ret= -1;
				goto error;
			}

			if (!(contacts = ast_sip_location_retrieve_aor_contacts_filtered(aor, AST_SIP_CONTACT_FILTER_REACHABLE))) {
				ret= -1;
				goto error;
			}

			if (!ao2_container_count(contacts)) {
				ret= -1;
				goto error;
			}
			ast_log(LOG_DEBUG,"ao2_container_count(contacts)=%d", ao2_container_count(contacts));
			it_contacts = ao2_iterator_init(contacts, 0);
			for (; (contact = ao2_iterator_next(&it_contacts)); ao2_ref(contact, -1)) {
				if(strstr(contact->uri, entry->remote_ip))
				{
					isfound = 1;
					ast_log(LOG_NOTICE, "Found contact[%p], aor[%s], uri[%s]", contact, contact->aor, contact->uri);
					break;
				}
			}
			ao2_iterator_destroy(&it_contacts);


		}
		else
		{
			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",entry->exten);
			if(endpoint)
			{
				contact = NULL;
				ast_sip_location_retrieve_contact_and_aor_from_list_filtered(endpoint->aors, AST_SIP_CONTACT_FILTER_REACHABLE,
					&found_aor, &contact);
				if(contact)
				{
					isfound = 1;
				}
			}
		}

		if(isfound)
		{
			strcat(dest, contact->aor);
			strcat(dest, "/");
			strcat(dest, contact->uri);
			strcpy(addr, dest);
			strcat(dest, "/");
			strcat(dest, entry->from_tag);
			strcat(dest, "/");
			strcat(dest, entry->to_tag);
			strcat(dest, "/");
			strcat(dest, entry->call_id);
			if(entry->delay_foripv6 == 1  && 0 == first_delay_foripv6_recovery )
			{
				usleep(500000); /* sleep  for wait ipv6 address becaming valid */
				first_delay_foripv6_recovery = 1;
			}
		}
		else
		{
			ast_log(LOG_ERROR," aor:%s find contact failed!!!",entry->exten);
			ret= -1;
			goto error;
		}

	}

	top =ast_stream_topology_alloc();/* need free top*/
	tc = chan_pjsip_request_with_stream_topology("PJSIP",top,NULL,NULL,
		dest,&cause);

	if(!tc)
	{
		ast_log(LOG_ERROR," original caller %s request channel failed!!!", entry->exten);
		ret = -1;
		goto error;
	}

	channel = ast_channel_tech_pvt(tc);
	session = channel->session;

	pjsip_dlg_inc_lock(session->inv_session->dlg);
	session->inv_session->dlg->local.first_cseq = entry->cseq+10;
	session->inv_session->dlg->local.cseq = entry->cseq+10;
	pjsip_dlg_dec_lock(session->inv_session->dlg);

	if (entry->session_conf_link)
	{
		session->conf_link = ast_calloc(1, strlen(entry->session_conf_link) + 1);
		if (session->conf_link) {
			strcpy(session->conf_link, entry->session_conf_link);
		}
	}

	if (entry->session_trunk_channel_id)
	{
		session->trunk_channel_id = ast_calloc(1, strlen(entry->session_trunk_channel_id) + 1);
		if (session->trunk_channel_id) {
			strcpy(session->trunk_channel_id, entry->session_trunk_channel_id);
		}
	}

	ast_set_flag(ast_channel_flags(tc), AST_FLAG_RECOVERYING);
	ast_set_flag(ast_channel_flags(tc), AST_FLAG_ORIGINATED);

	if(entry->conf_room[0] != '\0')
	{
		pbx_builtin_setvar_helper(tc, "__CONF_ROOM", entry->conf_room);
		session->is_sfu = 1;
		strncpy(session->conf_room, entry->conf_room, sizeof(session->conf_room) - 1);
	}
	if(entry->dahdi_channel != 0)
	{
		char dahdi_chan[16] = {0};
		sprintf(dahdi_chan, "%d", entry->dahdi_channel);
		pbx_builtin_setvar_helper(tc, "__RECOVERY_DAHDI_CHANNEL", dahdi_chan);
	}
	if(entry->orig_options != '\0')
	{
		pbx_builtin_setvar_helper(tc, "__ORI_OPTIONS", entry->orig_options);
	}

	if(entry->sca_answer_user)
	{
		pbx_builtin_setvar_helper(tc, "__SCA_ANSWERED_USER", entry->sca_answer_user);
	}

	ast_party_connected_line_set_init(&connected, ast_channel_connected(tc));
	connected.id.number.valid = TRUE;
	connected.id.number.str = ast_strdup(entry->connected_line_id);
	ast_party_connected_line_copy(ast_channel_connected(tc), &connected);
	ast_free(connected.id.number.str);
	ast_channel_connect_num_set(tc, entry->connected_line_id);
	ast_cdr_setvar(ast_channel_name(tc), "__app_callee", entry->connected_line_id);
	if (!ast_strlen_zero(entry->inbound_trunk_index)){
		ast_cdr_setvar(ast_channel_name(tc), "inbound_trunk_index", entry->inbound_trunk_index);
	}
	if(strstr(ast_channel_name(tc), "trunk_")){
		ast_cdr_setvar(ast_channel_name(tc), "__app_caller", entry->exten);
	}

	time_val.tv_sec = entry->create_sec;
	time_val.tv_usec = entry->create_usec;
	ast_channel_set_recovery_createtime(tc, &time_val);

	time_val.tv_sec = entry->answer_sec;
	time_val.tv_usec = entry->answer_usec;
	ast_channel_set_recovery_answertime(tc, &time_val);
	ast_channel_session_set(tc,entry->exten);

	char *linkid = ast_malloc(AST_MAX_UNIQUEID);
	char *uniqid = ast_malloc(AST_MAX_UNIQUEID);
	strncpy(linkid, entry->linkid, AST_MAX_UNIQUEID);
	strncpy(uniqid, entry->uuid, AST_MAX_UNIQUEID);

	ast_channel_set_recovery_linkedid(tc, linkid);
	ast_channel_set_recovery_uniqueid(tc, uniqid);
	if(entry->exten[0] != '\0')
	{
		ast_party_caller_init(&caller);
		caller.id.number.valid = 1;
		caller.id.number.str = ast_strdup(entry->exten);
		if(entry->caller_name[0] != '\0')
		{
			caller.id.name.valid = 1;
			caller.id.name.str = ast_strdup(entry->caller_name);
		}
		ast_party_caller_set(ast_channel_caller(tc), &caller, NULL);
		ast_free(caller.id.number.str);
		ast_free(caller.id.name.str);
	}

	ast_channel_media_encryption_set(tc, entry->encryption);
	/*we need init bridge profile*/
	if(entry->conf_room[0] != '\0')
	{
		ast_func_write(tc, "CONFBRIDGE(bridge,video_mode)", "sfu");
	}
	ret =ast_call(tc,addr, 0 );

error:
	if(top){
		ast_stream_topology_free(top);
	}

	if(ret == -1)
	{
		sql = switch_xml_mprintf("delete from recovery where runtime_uuid='%q'", entry->linkid);
		if (sql) {
			ast_log(LOG_NOTICE, "cmd is %s\n", sql);
			ast_recovery_db_put(sql, 1, 1); /*Mod by weidu ,for ha recovery_astdb sync*/
			ast_free(sql);
		}
	}

	ast_free(entry->from_tag);
	ast_free(entry->to_tag);
	ast_free(entry->call_id);
	ast_free(entry->remote_ip);
	if(entry->sca_answer_user)
	{
		ast_free(entry->sca_answer_user);
	}

	if (entry->session_conf_link)
	{
		ast_free(entry->session_conf_link);
	}

	if (entry->session_trunk_channel_id)
	{
		ast_free(entry->session_trunk_channel_id);
	}

	ast_free(entry);

	return ret;
}


static void set_media_encryption_from_restore_xml(struct ast_channel_media_encryption *encryption, switch_xml_t restore_xml)
{
	switch_xml_t media_encryption, media_encryption_child;

	media_encryption = switch_xml_child(restore_xml, "media_encryption");
	if (!media_encryption) {
		ast_log(LOG_DEBUG, "Can't get media_encryption xml\n");
		return;
	}

	media_encryption_child = switch_xml_child(media_encryption, "audio");
	if (media_encryption_child && !ast_strlen_zero(media_encryption_child->txt)) {
		encryption->audio_encryption = atoi(media_encryption_child->txt);
	} else {
		ast_log(LOG_DEBUG, "Can't get audio_media_encryption\n");
	}

	media_encryption_child = switch_xml_child(media_encryption, "main_video");
	if (media_encryption_child && !ast_strlen_zero(media_encryption_child->txt)) {
		encryption->main_video_encryption = atoi(media_encryption_child->txt);
	} else {
		ast_log(LOG_DEBUG, "Can't get main_video_media_encryption\n");
	}

	media_encryption_child = switch_xml_child(media_encryption, "slides_video");
	if (media_encryption_child && !ast_strlen_zero(media_encryption_child->txt)) {
		encryption->slides_video_encryption = atoi(media_encryption_child->txt);
	} else {
		ast_log(LOG_DEBUG, "Can't get slides_video_media_encryption\n");
	}

	return;
}

static void pjsip_start_call_recovery(void)
{
	struct recovery_db_entry* entry = NULL;
	struct recovery_db_entry* inbound_mode_sub_entry = NULL;
	switch_xml_t restore_xml,channel_data,di,vars, callid,from, to,callerid,time, sec, peerid,cseq,confroom, d_chan, session_conf_link, session_trunk_channel_id;
	switch_xml_t peer_trunk_aor = NULL;
	struct ast_taskprocessor *serializer = NULL;
	pj_str_t remote_tag, cid;
	int loop = 0;
	struct ast_sip_recovery_entry *one_entry;
	switch_xml_t chan_name, orig_options, caller_name, inbound_trunk_index,remote_ip,sca_answer_user,app_data;

	inbound_mode_sub_entry = ast_recovery_db_get_all();

	for (entry = inbound_mode_sub_entry; entry; entry = entry->next) {
		restore_xml = switch_xml_parse_str_dynamic(entry->data, SWITCH_TRUE);
		channel_data = switch_xml_child(restore_xml, "channel_data");
		di = switch_xml_child(channel_data, "direction");
		vars = switch_xml_child(restore_xml, "variables");
		chan_name = switch_xml_child(vars, "channel_name");
		if(strstr(chan_name->txt,"DAHDI"))
		{
			/* restore call on chan_dahdi module */
			switch_xml_free(restore_xml);
			continue;
		}
		callid = switch_xml_child(vars, "call_id");
		from = switch_xml_child(vars, "from_tag");
		to = switch_xml_child(vars, "to_tag");
		callerid = switch_xml_child(vars, "caller_id");
		peerid = switch_xml_child(vars, "connect_line_id");
		cseq = switch_xml_child(vars, "cseq");
		time = switch_xml_child(restore_xml, "times");
		peer_trunk_aor = switch_xml_child(vars, "peertrunk_aor");
		confroom = switch_xml_child(vars, "conf_room");
		d_chan = switch_xml_child(vars, "peer_dahdi_channel");
		orig_options = switch_xml_child(vars, "orig_options");
		caller_name = switch_xml_child(vars, "caller_id_name");
		inbound_trunk_index = switch_xml_child(vars, "inbound_trunk_index");
		remote_ip = switch_xml_child(vars, "contact_host");
		sca_answer_user = switch_xml_child(vars, "sca_answer_user");
		app_data = switch_xml_child(vars, "app_data");
		session_conf_link = switch_xml_child(vars, "session_conf_link");
		session_trunk_channel_id = switch_xml_child(vars, "session_trunk_channel_id");

		if (NULL == callid || NULL == from || NULL == to || NULL == callerid || NULL == peerid || NULL == cseq) {
			goto ended;
		}

		if (0 == strcmp(di->txt, "inbound") ||
			(0 == strcmp(di->txt, "outbound") && confroom))
		{
			remote_tag.ptr = from->txt;
			remote_tag.slen = strlen(from->txt);
			cid.ptr = callid->txt;
			cid.slen = strlen(callid->txt);

			serializer = ast_sip_distributor_serializer_for_recovery(&remote_tag, &cid);
			one_entry = ast_malloc(sizeof(struct ast_sip_recovery_entry));
			if(!one_entry)
			{
				switch_xml_free(restore_xml);
				ast_log(LOG_ERROR," alloc mem for recovery entry failed!");
				continue;
			}
			memset(one_entry,0, sizeof(struct ast_sip_recovery_entry));
			strcpy(one_entry->linkid, entry->runtime_uuid);
			strcpy(one_entry->uuid, entry->uuid);
			strcpy(one_entry->exten, callerid->txt);
			strcpy(one_entry->connected_line_id, peerid->txt);
			if(confroom)
			{
				strcpy(one_entry->conf_room, confroom->txt);
			}

			if(d_chan)
			{
				one_entry->dahdi_channel = atoi(d_chan->txt);
			}

			if(orig_options)
			{
				strcpy(one_entry->orig_options, orig_options->txt);
			}

			if(peer_trunk_aor)
			{
				strcpy(one_entry->peertrunk_aor, peer_trunk_aor->txt);
			}

			if(caller_name)
			{
				strcpy(one_entry->caller_name, caller_name->txt);
			}
			if(inbound_trunk_index && !ast_strlen_zero(inbound_trunk_index->txt))
			{
				strcpy(one_entry->inbound_trunk_index, inbound_trunk_index->txt);
			}

			if(remote_ip)
			{
				one_entry->remote_ip = ast_strdup(remote_ip->txt);
				if(strchr(one_entry->remote_ip,':'))
				{
					one_entry->delay_foripv6 = 1; /* wait ipv6 address being valid */
				}
			}

			if(app_data)
			{
				if(strstr(app_data->txt,"@["))
				{
					one_entry->delay_foripv6 = 1;/* wait ipv6 address being valid */
				}
			}

			if(sca_answer_user)
			{
				one_entry->sca_answer_user = ast_strdup(sca_answer_user->txt);
			}

			if(0 == strcmp(di->txt, "outbound"))/* conference user , press 1 for be invited */
			{
				one_entry->from_tag = ast_strdup(from->txt);
				one_entry->to_tag = ast_strdup(to->txt);
			}
			else
			{
				one_entry->from_tag = ast_strdup(to->txt);
				one_entry->to_tag = ast_strdup(from->txt);
			}
			one_entry->call_id = ast_strdup(callid->txt);
			one_entry->cseq = atoi(cseq->txt);

			sec = switch_xml_child(time, "create_sec");
			one_entry->create_sec = atoi(sec->txt);
			sec = switch_xml_child(time, "create_usec");
			one_entry->create_usec = atoi(sec->txt);
			sec = switch_xml_child(time, "answer_sec");
			one_entry->answer_sec = atoi(sec->txt);
			sec = switch_xml_child(time, "answer_usec");
			one_entry->answer_usec = atoi(sec->txt);

			if (session_conf_link)
			{
				one_entry->session_conf_link = ast_strdup(session_conf_link->txt);
			}

			if (session_trunk_channel_id)
			{
				one_entry->session_trunk_channel_id = ast_strdup(session_trunk_channel_id->txt);
			}

			set_media_encryption_from_restore_xml(&one_entry->encryption, restore_xml);

			ast_log(LOG_ERROR," hey hey serializer=%p entry=%p",serializer,one_entry);
			if (ast_sip_push_task(serializer, chan_pjsip_start_recovery, one_entry)) {
				switch_xml_free(restore_xml);
				ast_free(one_entry);
				ast_log(LOG_ERROR," push recovery task failed!");
				continue;
			}

			ast_taskprocessor_unreference(serializer);
			loop++;

		}

ended:
		if(restore_xml)
		{
			switch_xml_free(restore_xml);
		}
	}
	if (inbound_mode_sub_entry) {
		ast_recovery_freetree(inbound_mode_sub_entry);
	}

	return;
}



static char *cli_pjsip_call_recovery(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct recovery_db_entry* entry = NULL;
	struct recovery_db_entry* inbound_mode_sub_entry = NULL;
	switch_xml_t restore_xml,channel_data,di,vars, callid,from, to,callerid,time, sec, peerid,cseq,confroom;
	switch_xml_t peer_trunk_aor;
	struct ast_taskprocessor *serializer = NULL;
	pj_str_t remote_tag, cid;
	int loop = 0;
	struct ast_sip_recovery_entry *one_entry;


	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip call recovery";
		e->usage = "Usage: pjsip call recovery\n";
		return NULL;
	case CLI_GENERATE:
		break;
	}

	inbound_mode_sub_entry = ast_recovery_db_get_all();

	for (entry = inbound_mode_sub_entry; entry; entry = entry->next) {
		restore_xml = switch_xml_parse_str_dynamic(entry->data, SWITCH_TRUE);
		channel_data = switch_xml_child(restore_xml, "channel_data");
		di = switch_xml_child(channel_data, "direction");
		vars = switch_xml_child(restore_xml, "variables");
		callid = switch_xml_child(vars, "call_id");
		from = switch_xml_child(vars, "from_tag");
		to = switch_xml_child(vars, "to_tag");
		callerid = switch_xml_child(vars, "caller_id");
		peerid = switch_xml_child(vars, "connect_line_id");
		cseq = switch_xml_child(vars, "cseq");
		time = switch_xml_child(restore_xml, "times");
		peer_trunk_aor = switch_xml_child(vars, "peertrunk_aor");
		confroom = switch_xml_child(vars, "conf_room");
		if (0 == strcmp(di->txt, "inbound") ||
			(0 == strcmp(di->txt, "outbound") && confroom))
		{
			remote_tag.ptr = from->txt;
			remote_tag.slen = strlen(from->txt);
			cid.ptr = callid->txt;
			cid.slen = strlen(callid->txt);

			serializer = ast_sip_distributor_serializer_for_recovery(&remote_tag, &cid);
			one_entry = ast_malloc(sizeof(struct ast_sip_recovery_entry));
			if(!one_entry)
			{
				switch_xml_free(restore_xml);
				ast_log(LOG_ERROR," alloc mem for recovery entry failed!");
				continue;
			}
			memset(one_entry,0, sizeof(struct ast_sip_recovery_entry));
			strcpy(one_entry->linkid, entry->runtime_uuid);
			strcpy(one_entry->uuid, entry->uuid);
			strcpy(one_entry->exten, callerid->txt);
			strcpy(one_entry->connected_line_id, peerid->txt);
			if(confroom)
			{
				strcpy(one_entry->conf_room, confroom->txt);
			}

			if(peer_trunk_aor)
			{
				strcpy(one_entry->peertrunk_aor, peer_trunk_aor->txt);
			}

			if(0 == strcmp(di->txt, "outbound"))/* conference user , press 1 for be invited */
			{
				one_entry->from_tag = ast_strdup(from->txt);
				one_entry->to_tag = ast_strdup(to->txt);
			}
			else
			{
				one_entry->from_tag = ast_strdup(to->txt);
				one_entry->to_tag = ast_strdup(from->txt);
			}
			one_entry->call_id = ast_strdup(callid->txt);
			one_entry->cseq = atoi(cseq->txt);

			sec = switch_xml_child(time, "create_sec");
			one_entry->create_sec = atoi(sec->txt);
			sec = switch_xml_child(time, "create_usec");
			one_entry->create_usec = atoi(sec->txt);
			sec = switch_xml_child(time, "answer_sec");
			one_entry->answer_sec = atoi(sec->txt);
			sec = switch_xml_child(time, "answer_usec");
			one_entry->answer_usec = atoi(sec->txt);

			ast_log(LOG_ERROR," hey hey serializer=%p entry=%p",serializer,one_entry);
			if (ast_sip_push_task(serializer, chan_pjsip_start_recovery, one_entry)) {
				switch_xml_free(restore_xml);
				ast_log(LOG_ERROR," push recovery task failed!");
				continue;
			}

			ast_taskprocessor_unreference(serializer);
			loop++;

		}

		if(restore_xml)
		{
			switch_xml_free(restore_xml);
		}
	}
	if (inbound_mode_sub_entry) {
		ast_recovery_freetree(inbound_mode_sub_entry);
	}

	return CLI_SUCCESS;
}




static struct ast_cli_entry cli_commands[] = {
	AST_CLI_DEFINE(cli_pjsip_call_recovery, "ha call recovery")
};

static int chan_pjsip_append_sip_header(struct ast_channel *chan, void *xml_cdr, void *msg, int *offset)
{
	pjsip_msg *sip_msg;
	pjsip_from_hdr *from_hdr = NULL;

	pjsip_from_hdr *to_hdr = NULL;
	pjsip_cid_hdr *cid = NULL;
	pjsip_cseq_hdr *cseq = NULL;
	char cseq_str[16] = {0};
	switch_xml_t vars,x_field;
	char tmp[512] = {0};


	if(!chan || !msg || !xml_cdr || !offset)
	{
		return -1;
	}
	vars = (switch_xml_t)xml_cdr;
	sip_msg = (pjsip_msg *)msg;

	from_hdr = pjsip_msg_find_hdr(sip_msg, PJSIP_H_FROM, NULL);
	if (from_hdr){
		x_field = switch_xml_add_child_d(vars, "from_tag",(*offset)++);
		ast_copy_pj_str(tmp, &from_hdr->tag, pj_strlen(&from_hdr->tag)+1);
		switch_xml_set_txt_d(x_field, tmp);
	}


	to_hdr = pjsip_msg_find_hdr(sip_msg, PJSIP_H_TO, NULL);
	if(to_hdr){
		x_field = switch_xml_add_child_d(vars, "to_tag", (*offset)++);
		ast_copy_pj_str(tmp, &to_hdr->tag, pj_strlen(&to_hdr->tag)+1);
		switch_xml_set_txt_d(x_field, tmp);

	}

	cid = pjsip_msg_find_hdr(sip_msg, PJSIP_H_CALL_ID, NULL);
	if(cid){
		x_field = switch_xml_add_child_d(vars, "call_id",(*offset)++);
		ast_copy_pj_str(tmp, &cid->id, pj_strlen(&cid->id)+1);
		switch_xml_set_txt_d(x_field, tmp);
	}

	cseq = pjsip_msg_find_hdr(sip_msg, PJSIP_H_CSEQ, NULL);
	if(cseq){
		x_field = switch_xml_add_child_d(vars, "cseq",(*offset)++);
		sprintf(cseq_str,"%d",cseq->cseq);
		switch_xml_set_txt_d(x_field,cseq_str);
	}

	struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
	if(!channel || !channel->session)
	{
		return -1;
	}
	if(channel->session->active_contact_addr)
	{
		x_field = switch_xml_add_child_d(vars, "contact_host",(*offset)++);
		switch_xml_set_txt_d(x_field,channel->session->active_contact_addr);
	}

	if(channel->session->conf_link)
	{
		x_field = switch_xml_add_child_d(vars, "session_conf_link", (*offset)++);
		switch_xml_set_txt_d(x_field, channel->session->conf_link);
	}

	if(channel->session->trunk_channel_id)
	{
		x_field = switch_xml_add_child_d(vars, "session_trunk_channel_id", (*offset)++);
		switch_xml_set_txt_d(x_field, channel->session->trunk_channel_id);
	}

	return 0;
}


static struct stasis_message_router *chan_pjsip_ha_data_router = NULL;

static void chan_pjsip_ha_role_switch_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	static int num=0;
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	int    role = 0;
	ast_log(LOG_NOTICE,"----->chan_pjsip_ha_role_switch_cb seq=%d", num);

	if(!ast_get_redundance_flag() && num >= 1)
		return;
	/* Only do sync when switch to active */
	if (HA_ROLE_ACTIVE != ast_ha_get_role()) {
		return;
	}

	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (!payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	role = ast_json_integer_get(ast_json_object_get(json_object, "ha_role"));
	if (!role) {
		ast_log(LOG_WARNING, "Incomplete JSON object data");
		return;
	}

	ast_log(LOG_NOTICE, "chan_pjsip swtiched to Active.\n");

	ast_sip_reload_waveappinfo_from_astdb(0);//added by zyxia for Bug 315367,reload appinfo form astdb
	while(!ast_sip_smooth_notify_avs_flag())
	{
		usleep(1);
	}

	usleep(1500000);/* wait ipv6 address valid exactly,for avs bind socket */
	ast_log(LOG_NOTICE, " Start recovery calls on new master.\n");
	pjsip_start_call_recovery();
	num++;

	ast_high_frequence_queue_initload();
	return;
}


static int chan_pjsip_ha_stasis_route_init(void)
{
	struct stasis_topic * pjsip_ha_data_topic = ast_ha_sync_data_topic_all();
	if (!pjsip_ha_data_topic) {
		ast_log(LOG_WARNING, "Failed to get chan_pjsip_ha_data_topic\n");
		return -1;
	}

	chan_pjsip_ha_data_router = stasis_message_router_create(pjsip_ha_data_topic);
	if (chan_pjsip_ha_data_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create chan_pjsip_ha_data_router\n");
		return -1;
	}

	if (stasis_message_router_add(chan_pjsip_ha_data_router, ast_ha_role_switch_type(), chan_pjsip_ha_role_switch_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_role_switch_type\n");
		return -1;
	}
	ast_log(LOG_ERROR,"chan_pjsip_ha_stasis_route_init succeed!!!");
	return 0;
}


#endif

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	struct ao2_container *endpoints;

	if (!(chan_pjsip_tech.capabilities = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		return AST_MODULE_LOAD_DECLINE;
	}

	ast_format_cap_append_by_type(chan_pjsip_tech.capabilities, AST_MEDIA_TYPE_AUDIO);

	ast_rtp_glue_register(&chan_pjsip_rtp_glue);

	if (ast_channel_register(&chan_pjsip_tech)) {
		ast_log(LOG_ERROR, "Unable to register channel class %s\n", channel_type);
		goto end;
	}

	if (ast_custom_function_register(&chan_pjsip_dial_contacts_function)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP_DIAL_CONTACTS dialplan function\n");
		goto end;
	}

	if (ast_custom_function_register(&chan_pjsip_parse_uri_function)) {
		ast_log(LOG_ERROR, "Unable to register PJSIP_PARSE_URI dialplan function\n");
		goto end;
	}

	if (ast_custom_function_register(&media_offer_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_MEDIA_OFFER dialplan function\n");
		goto end;
	}

	if (ast_custom_function_register(&dtmf_mode_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_DTMF_MODE dialplan function\n");
		goto end;
	}

	if (ast_custom_function_register(&session_refresh_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_SEND_SESSION_REFRESH dialplan function\n");
		goto end;
	}

	ast_sip_session_register_supplement(&chan_pjsip_supplement);
	ast_sip_session_register_supplement(&chan_pjsip_supplement_response);

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_session_register_supplement(&chan_pjsip_supplement_fastprov);
#endif

	if (!(pjsip_uids_onhold = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_RWLOCK,
			AO2_CONTAINER_ALLOC_OPT_DUPS_REJECT, 37, uid_hold_hash_fn,
			uid_hold_sort_fn, NULL))) {
		ast_log(LOG_ERROR, "Unable to create held channels container\n");
		goto end;
	}

	ast_sip_session_register_supplement(&call_pickup_supplement);
	ast_sip_session_register_supplement(&pbx_start_supplement);
	ast_sip_session_register_supplement(&chan_pjsip_ack_supplement);

#ifdef GRANDSTREAM_NETWORKS
	if (ast_custom_function_register(&chan_pjsip_missed_call_contacts_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_MISSED_CALL_CONTACTS dialplan function!\n");
		goto end;
	}

	if (ast_custom_function_register(&chan_pjsip_gs_message_contacts_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_GS_MESSAGE_CONTACTS dialplan function!\n");
		goto end;
	}

	if (ast_custom_function_register(&chan_pjsip_plain_message_contacts_function)) {
		ast_log(LOG_WARNING, "Unable to register PJSIP_PLAIN_MESSAGE_CONTACTS dialplan function!\n");
		goto end;
	}

	globalEnabledCEI = ast_cei_get_enable_configuration();
	get_feature_code(0);

	struct stasis_topic *cei_toggle_enable_topic = NULL;

	cei_toggle_enable_topic = ast_cei_toggle_enable_topic_all();
	if (!cei_toggle_enable_topic) {
		ast_log(LOG_WARNING, "[WL] Failed to get cei_toggle_enable_topic\n");
		goto end;
	}

	cei_toggle_enable_router = stasis_message_router_create(cei_toggle_enable_topic);
	if (cei_toggle_enable_router == NULL) {
		ast_log(LOG_WARNING, "[WL] Failed to create cei_toggle_enable_router\n");
		goto end;
	}

	if (stasis_message_router_add(cei_toggle_enable_router, ast_cei_toggle_enable_type(),
			cei_toggle_enable_cb, NULL))
	{
		ast_log(LOG_WARNING, "Failed to subscribe to ast_cei_toggle_enable_type\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsetcalltype, sip_setcalltype)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetCallType dialplan function\n");
		goto end;
	}

	if (ast_custom_function_register(&pjsip_mcb_resp_function)) {
		ast_log(LOG_ERROR, "Unable to register MCB_RESPONSE dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsetmcbresp, sip_setmcbresp)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetMCBResp dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsetconffocus, sip_set_conf_focus)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetConfFocus dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsetcontactuser, sip_set_contact_user)) {
		ast_log(LOG_ERROR, "Unable to register SIPSetContactUser dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipterminatednotify, sip_terminated_notify)) {
		ast_log(LOG_ERROR, "Unable to register SIPTerminatedNotify dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipceijoined, sip_cei_joined)) {
		ast_log(LOG_ERROR, "Unable to register SIPCEIJoined dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsubscribeextmcb, sip_external_mcb_sub)) {
		ast_log(LOG_ERROR, "Unable to register SIPSubscribeExtMCB dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipunsubscribeextmcb, sip_external_mcb_unsub)) {
		ast_log(LOG_ERROR, "Unable to register SIPUnsubscribeExtMCB dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipsubscribemcb, sip_internal_mcb_sub)) {
		ast_log(LOG_ERROR, "Unable to register SIPSubscribeMCB dialplan function\n");
		goto end;
	}

	if (ast_register_application_xml(app_sipunsubscribemcb, sip_internal_mcb_unsub)) {
		ast_log(LOG_ERROR, "Unable to register SIPUnsubscribeMCB dialplan function\n");
		goto end;
	}

	/* Add by bfdu for Bug 78568, 2017.02.08 */
	ast_manager_register_xml("CRMInfo", EVENT_FLAG_SYSTEM, ami_push_crm_info);
	ast_manager_register_xml("QueryNatStatus", EVENT_FLAG_SYSTEM, ami_query_nat_status);
	/*Start added by zyxia for callflip.2021.11.24*/
	ast_manager_register_xml("SendCallFlipStatus", EVENT_FLAG_SYSTEM, action_send_callflip_status);
	ast_manager_register_xml("CallFlip", EVENT_FLAG_SYSTEM, action_callflip);

	/*End added by zyxia for callflip.2021.11.24*/

	ast_cli_register_multiple(cli_commands, ARRAY_LEN(cli_commands));

	if (!(emergency_mail_tps = ast_taskprocessor_get("emergency_mail", TPS_REF_DEFAULT))) {
		ast_log(AST_LOG_WARNING, "failed to reference emergency mail taskprocessor. Emergency mail will not work\n");
	}
#endif

	if (pjsip_channel_cli_register()) {
		ast_log(LOG_ERROR, "Unable to register PJSIP Channel CLI\n");
		goto end;
	}

	/* since endpoints are loaded before the channel driver their device
	   states get set to 'invalid', so they need to be updated */
	if ((endpoints = ast_sip_get_endpoints())) {
		ao2_callback(endpoints, OBJ_NODATA, update_devstate, NULL);
		ao2_ref(endpoints, -1);
	}

	/*if (HA_ROLE_STANDBY == ast_ha_get_role())*/
	if (1)
	{
		chan_pjsip_ha_stasis_route_init();
	}

	ast_high_frequence_queue_initload();

	return 0;

end:
	ao2_cleanup(pjsip_uids_onhold);
	pjsip_uids_onhold = NULL;
	ast_sip_session_unregister_supplement(&chan_pjsip_ack_supplement);
	ast_sip_session_unregister_supplement(&pbx_start_supplement);
	ast_sip_session_unregister_supplement(&chan_pjsip_supplement_response);
	ast_sip_session_unregister_supplement(&chan_pjsip_supplement);
	ast_sip_session_unregister_supplement(&call_pickup_supplement);
	ast_custom_function_unregister(&dtmf_mode_function);
	ast_custom_function_unregister(&media_offer_function);
	ast_custom_function_unregister(&chan_pjsip_dial_contacts_function);
	ast_custom_function_unregister(&chan_pjsip_parse_uri_function);
	ast_custom_function_unregister(&session_refresh_function);
	ast_channel_unregister(&chan_pjsip_tech);
	ast_rtp_glue_unregister(&chan_pjsip_rtp_glue);
#ifdef GRANDSTREAM_NETWORKS
	ast_custom_function_unregister(&chan_pjsip_missed_call_contacts_function);

	ast_unregister_application(app_sipsetcalltype);
	ast_custom_function_unregister(&pjsip_mcb_resp_function);
	ast_unregister_application(app_sipsetmcbresp);
	ast_unregister_application(app_sipsetconffocus);
	ast_unregister_application(app_sipsetcontactuser);
	ast_unregister_application(app_sipterminatednotify);
	ast_unregister_application(app_sipceijoined);
	ast_unregister_application(app_sipsubscribeextmcb);
	ast_unregister_application(app_sipunsubscribeextmcb);
	ast_unregister_application(app_sipsubscribemcb);
	ast_unregister_application(app_sipunsubscribemcb);

	if (cei_toggle_enable_router) {
		stasis_message_router_unsubscribe(cei_toggle_enable_router);
		cei_toggle_enable_router = NULL;
	}

	/* Add by bfdu for Bug 78568, 2017.02.08 */
	ast_manager_unregister("CRMInfo");

	/*Start added by zyxia for callflip,2021.11.23*/
	ast_manager_unregister("SendCallFlipStatus");
	ast_manager_unregister("CallFlip");
	/*End added by zyxia for callflip,2021.11.23*/
#endif

	return AST_MODULE_LOAD_DECLINE;
}

/*! \brief Unload the PJSIP channel from Asterisk */
static int unload_module(void)
{
	ao2_cleanup(pjsip_uids_onhold);
	pjsip_uids_onhold = NULL;

	ast_cli_unregister_multiple(cli_commands, ARRAY_LEN(cli_commands));

	pjsip_channel_cli_unregister();

	ast_sip_session_unregister_supplement(&chan_pjsip_supplement_response);
	ast_sip_session_unregister_supplement(&chan_pjsip_supplement);
	ast_sip_session_unregister_supplement(&pbx_start_supplement);
	ast_sip_session_unregister_supplement(&chan_pjsip_ack_supplement);
	ast_sip_session_unregister_supplement(&call_pickup_supplement);

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_session_unregister_supplement(&chan_pjsip_supplement_fastprov);
#endif

	ast_custom_function_unregister(&dtmf_mode_function);
	ast_custom_function_unregister(&media_offer_function);
	ast_custom_function_unregister(&chan_pjsip_dial_contacts_function);
	ast_custom_function_unregister(&chan_pjsip_parse_uri_function);
	ast_custom_function_unregister(&session_refresh_function);

	ast_channel_unregister(&chan_pjsip_tech);
	ao2_ref(chan_pjsip_tech.capabilities, -1);
	ast_rtp_glue_unregister(&chan_pjsip_rtp_glue);

#ifdef GRANDSTREAM_NETWORKS
	ast_unregister_application(app_sipsetcalltype);
	ast_custom_function_unregister(&pjsip_mcb_resp_function);
	ast_unregister_application(app_sipsetmcbresp);
	ast_unregister_application(app_sipsetconffocus);
	ast_unregister_application(app_sipsetcontactuser);
	ast_unregister_application(app_sipterminatednotify);
	ast_unregister_application(app_sipceijoined);
	ast_unregister_application(app_sipsubscribeextmcb);
	ast_unregister_application(app_sipunsubscribeextmcb);
	ast_unregister_application(app_sipsubscribemcb);
	ast_unregister_application(app_sipunsubscribemcb);

	ast_custom_function_unregister(&chan_pjsip_gs_message_contacts_function);
	ast_custom_function_unregister(&chan_pjsip_plain_message_contacts_function);

	if (cei_toggle_enable_router) {
		stasis_message_router_unsubscribe(cei_toggle_enable_router);
		cei_toggle_enable_router = NULL;
	}

	if (chan_pjsip_ha_data_router) {
		stasis_message_router_unsubscribe(chan_pjsip_ha_data_router);
		chan_pjsip_ha_data_router = NULL;
	}

	/* Add by bfdu for Bug 78568, 2017.02.08 */
	ast_manager_unregister("CRMInfo");
	ast_manager_unregister("SendCallFlipStatus");//added by zyxia for callflip notify
	ast_manager_unregister("CallFlip");

	emergency_mail_tps = ast_taskprocessor_unreference(emergency_mail_tps);
#endif

	return 0;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP Channel Driver",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
#ifdef GRANDSTREAM_NETWORKS
	.reload = reload_module,
#endif
	.load_pri = AST_MODPRI_CHANNEL_DRIVER,
	.requires = "res_pjsip,res_pjsip_session",
);

