/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2006, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Call Detail Record API
 *
 * \author Mark Spencer <markster@digium.com>
 *
 * \note Includes code and algorithms from the Zapata library.
 *
 * \note We do a lot of checking here in the CDR code to try to be sure we don't ever let a CDR slip
 * through our fingers somehow.  If someone allocates a CDR, it must be completely handled normally
 * or a WARNING shall be logged, so that we can best keep track of any escape condition where the CDR
 * isn't properly generated and posted.
 */

/*! \li \ref cdr.c uses the configuration file \ref cdr.conf
 * \addtogroup configuration_file Configuration Files
 */

/*!
 * \page cdr.conf cdr.conf
 * \verbinclude cdr.conf.sample
 */

/*** MODULEINFO
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <signal.h>
#include <inttypes.h>

#include "asterisk/lock.h"
#include "asterisk/channel.h"
#include "asterisk/cdr.h"
#include "asterisk/callerid.h"
#include "asterisk/manager.h"
#include "asterisk/module.h"
#include "asterisk/causes.h"
#include "asterisk/linkedlists.h"
#include "asterisk/utils.h"
#include "asterisk/sched.h"
#include "asterisk/config.h"
#include "asterisk/cli.h"
#include "asterisk/stringfields.h"
#include "asterisk/config_options.h"
#include "asterisk/json.h"
#include "asterisk/parking.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_channels.h"
#include "asterisk/stasis_bridges.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/astobj2.h"
#include "asterisk/taskprocessor.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/astdb.h"
#include "asterisk/mixmonitor.h"
AST_MODULE_LOG("cdr");
#endif

/*** DOCUMENTATION
	<configInfo name="cdr" language="en_US">
		<synopsis>Call Detail Record configuration</synopsis>
		<description>
			<para>CDR is Call Detail Record, which provides logging services via a variety of
			pluggable backend modules. Detailed call information can be recorded to
			databases, files, etc. Useful for billing, fraud prevention, compliance with
			Sarbanes-Oxley aka The Enron Act, QOS evaluations, and more.</para>
		</description>
		<configFile name="cdr.conf">
			<configObject name="general">
				<synopsis>Global settings applied to the CDR engine.</synopsis>
				<configOption name="debug">
					<synopsis>Enable/disable verbose CDR debugging.</synopsis>
					<description><para>When set to <literal>True</literal>, verbose updates
					of changes in CDR information will be logged. Note that this is only
					of use when debugging CDR behavior.</para>
					</description>
				</configOption>
				<configOption name="enable">
					<synopsis>Enable/disable CDR logging.</synopsis>
					<description><para>Define whether or not to use CDR logging. Setting this to "no" will override
					any loading of backend CDR modules.  Default is "yes".</para>
					</description>
				</configOption>
				<configOption name="unanswered">
					<synopsis>Log calls that are never answered and don't set an outgoing party.</synopsis>
					<description><para>
					Define whether or not to log unanswered calls that don't involve an outgoing party. Setting
					this to "yes" will make calls to extensions that don't answer and don't set a side B channel
					(such as by using the Dial application) receive CDR log entries. If this option is set to
					"no", then those log entries will not be created. Unanswered calls which get offered to an
					outgoing line will always receive log entries regardless of this option, and that is the
					intended behavior.
					</para>
					</description>
				</configOption>
				<configOption name="congestion">
					<synopsis>Log congested calls.</synopsis>
					<description><para>Define whether or not to log congested calls. Setting this to "yes" will
					report each call that fails to complete due to congestion conditions.</para>
					</description>
				</configOption>
				<configOption name="endbeforehexten">
					<synopsis>Don't produce CDRs while executing hangup logic</synopsis>
					<description>
						<para>As each CDR for a channel is finished, its end time is updated
						and the CDR is finalized. When a channel is hung up and hangup
						logic is present (in the form of a hangup handler or the
						<literal>h</literal> extension), a new CDR is generated for the
						channel. Any statistics are gathered from this new CDR. By enabling
						this option, no new CDR is created for the dialplan logic that is
						executed in <literal>h</literal> extensions or attached hangup handler
						subroutines. The default value is <literal>yes</literal>, indicating
						that a CDR will be generated during hangup logic.</para>
					</description>
				</configOption>
				<configOption name="initiatedseconds">
					<synopsis>Count microseconds for billsec purposes</synopsis>
					<description><para>Normally, the <literal>billsec</literal> field logged to the CDR backends
					is simply the end time (hangup time) minus the answer time in seconds. Internally,
					asterisk stores the time in terms of microseconds and seconds. By setting
					initiatedseconds to <literal>yes</literal>, you can force asterisk to report any seconds
					that were initiated (a sort of round up method). Technically, this is
					when the microsecond part of the end time is greater than the microsecond
					part of the answer time, then the billsec time is incremented one second.</para>
					</description>
				</configOption>
				<configOption name="batch">
					<synopsis>Submit CDRs to the backends for processing in batches</synopsis>
					<description><para>Define the CDR batch mode, where instead of posting the CDR at the end of
					every call, the data will be stored in a buffer to help alleviate load on the
					asterisk server.</para>
					<warning><para>Use of batch mode may result in data loss after unsafe asterisk termination,
					i.e., software crash, power failure, kill -9, etc.</para>
					</warning>
					</description>
				</configOption>
				<configOption name="size">
					<synopsis>The maximum number of CDRs to accumulate before triggering a batch</synopsis>
					<description><para>Define the maximum number of CDRs to accumulate in the buffer before posting
					them to the backend engines. batch must be set to <literal>yes</literal>.</para>
					</description>
				</configOption>
				<configOption name="time">
					<synopsis>The maximum time to accumulate CDRs before triggering a batch</synopsis>
					<description><para>Define the maximum time to accumulate CDRs before posting them in a batch to the
					backend engines. If this time limit is reached, then it will post the records, regardless of the value
					defined for size. batch must be set to <literal>yes</literal>.</para>
					<note><para>Time is expressed in seconds.</para></note>
					</description>
				</configOption>
				<configOption name="scheduleronly">
					<synopsis>Post batched CDRs on their own thread instead of the scheduler</synopsis>
					<description><para>The CDR engine uses the internal asterisk scheduler to determine when to post
					records.  Posting can either occur inside the scheduler thread, or a new
					thread can be spawned for the submission of every batch.  For small batches,
					it might be acceptable to just use the scheduler thread, so set this to <literal>yes</literal>.
					For large batches, say anything over size=10, a new thread is recommended, so
					set this to <literal>no</literal>.</para>
					</description>
				</configOption>
				<configOption name="safeshutdown">
					<synopsis>Block shutdown of Asterisk until CDRs are submitted</synopsis>
					<description><para>When shutting down asterisk, you can block until the CDRs are submitted.  If
					you don't, then data will likely be lost.  You can always check the size of
					the CDR batch buffer with the CLI <astcli>cdr status</astcli> command.  To enable blocking on
					submission of CDR data during asterisk shutdown, set this to <literal>yes</literal>.</para>
					</description>
				</configOption>
			</configObject>
		</configFile>
	</configInfo>
 ***/


/* The prime here should be similar in size to the channel container. */
#ifdef LOW_MEMORY
#define NUM_CDR_BUCKETS 61
#else
#define NUM_CDR_BUCKETS 769
#endif

#define DEFAULT_ENABLED "1"
#define DEFAULT_BATCHMODE "0"
#define DEFAULT_UNANSWERED "0"
#define DEFAULT_CONGESTION "0"
#define DEFAULT_END_BEFORE_H_EXTEN "1"
#define DEFAULT_INITIATED_SECONDS "0"

#define DEFAULT_BATCH_SIZE "100"
#define MAX_BATCH_SIZE 1000
#define DEFAULT_BATCH_TIME "300"
#define MAX_BATCH_TIME 86400
#define DEFAULT_BATCH_SCHEDULER_ONLY "0"
#define DEFAULT_BATCH_SAFE_SHUTDOWN "1"

#ifdef GRANDSTREAM_NETWORKS
static unsigned int limit_time_used = 0; /* Use for day time used */
static unsigned int limit_month_time_used = 0;  /* Use for month time used */
static unsigned int has_limit_time = 0;
static unsigned int has_month_limit_time = 0;
static ast_rwlock_t time_lock;
static pthread_t clear_limit_time_thread = AST_PTHREADT_NULL;
static int clear_limit_time_run_flag = 1;
static unsigned int day_total_limit_time = 0;

#define TURN_INFO_LIMITTIMEUSED_KEY "LimitTimeUsed"
#define TURN_INFO_LIMITMONTHTIMEUSED_KEY "LimitMonthTimeUsed"
#define TURN_INFO_CLRARDAYTIME_KEY "ClearDayTime"
#define TURN_INFO_CLRARMONTHTIME_KEY "ClearMonthTime"
#define TURN_INFO_LASTUSEDTIME_KEY "LastUsedTime"

//"answered elsewhere" is the same meaning to "answer-elsewhere", but should consider the compatibility
#define ANSWERED_ELSEWHERE "answer-elsewhere"
#define ANSWERED_ELSEWHERE_NEW "answered elsewhere"
#define NOT_SEND_TO_WAVE "not_send"
#endif

#define cdr_set_debug_mode(mod_cfg) \
	do { \
		cdr_debug_enabled = ast_test_flag(&(mod_cfg)->general->settings, CDR_DEBUG); \
	} while (0)

static int cdr_debug_enabled;

#define CDR_DEBUG(fmt, ...) \
	do { \
		if (cdr_debug_enabled) { \
			ast_verbose((fmt), ##__VA_ARGS__); \
		} \
	} while (0)

static void cdr_detach(struct ast_cdr *cdr);
static void cdr_submit_batch(int shutdown);
static int cdr_toggle_runtime_options(void);

/*! \brief The configuration settings for this module */
struct module_config {
	struct ast_cdr_config *general;		/*!< CDR global settings */
};

/*! \brief The container for the module configuration */
static AO2_GLOBAL_OBJ_STATIC(module_configs);

/*! \brief The type definition for general options */
static struct aco_type general_option = {
	.type = ACO_GLOBAL,
	.name = "general",
	.item_offset = offsetof(struct module_config, general),
	.category = "general",
	.category_match = ACO_WHITELIST_EXACT,
};

/*! Config sections used by existing modules. Do not add to this list. */
static const char *ignore_categories[] = {
	"csv",
	"custom",
	"manager",
	"odbc",
	"pgsql",
	"radius",
	"sqlite",
	"tds",
	"mysql",
	NULL,
};

static struct aco_type ignore_option = {
	.type = ACO_IGNORE,
	.name = "modules",
	.category = (const char*)ignore_categories,
	.category_match = ACO_WHITELIST_ARRAY,
};

#ifdef GRANDSTREAM_NETWORKS
static int cdr_write_database_disable = 0;
static char device_name[16] = {0};
static void get_device_name(void){
	FILE *name_fp = NULL;
	name_fp = popen("nvram get syslog_id", "r");
	if (name_fp != NULL) {
		if (name_fp != NULL && (!feof(name_fp)) && (fgets(device_name, sizeof(device_name) - 1, name_fp) != NULL)) {
			ast_log(LOG_DEBUG, "the device name: %s", device_name);
		}
	}
	pclose(name_fp);
}
#endif

static void *module_config_alloc(void);
static void module_config_destructor(void *obj);
static void module_config_post_apply(void);

/*! \brief The file definition */
static struct aco_file module_file_conf = {
	.filename = "cdr.conf",
	.types = ACO_TYPES(&general_option, &ignore_option),
};

CONFIG_INFO_CORE("cdr", cfg_info, module_configs, module_config_alloc,
	.files = ACO_FILES(&module_file_conf),
	.post_apply_config = module_config_post_apply,
);

static struct aco_type *general_options[] = ACO_TYPES(&general_option);

static void module_config_post_apply(void)
{
	struct module_config *mod_cfg;

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return;
	}
	cdr_set_debug_mode(mod_cfg);
	ao2_cleanup(mod_cfg);
}

/*! \brief Dispose of a module config object */
static void module_config_destructor(void *obj)
{
	struct module_config *cfg = obj;

	if (!cfg) {
		return;
	}
	ao2_ref(cfg->general, -1);
}

/*! \brief Create a new module config object */
static void *module_config_alloc(void)
{
	struct module_config *mod_cfg;
	struct ast_cdr_config *cdr_config;

	mod_cfg = ao2_alloc(sizeof(*mod_cfg), module_config_destructor);
	if (!mod_cfg) {
		return NULL;
	}

	cdr_config = ao2_alloc(sizeof(*cdr_config), NULL);
	if (!cdr_config) {
		ao2_ref(cdr_config, -1);
		return NULL;
	}
	mod_cfg->general = cdr_config;

	return mod_cfg;
}

/*! \brief Registration object for CDR backends */
struct cdr_beitem {
	char name[20];
	char desc[80];
	ast_cdrbe be;
	AST_RWLIST_ENTRY(cdr_beitem) list;
	int suspended:1;
};

struct cdr_sql_beitem{
	ast_cdr_sql_backend sql_be;
};

static struct cdr_sql_beitem cdr_sql_be = {.sql_be = NULL,};
/*! \brief List of registered backends */
static AST_RWLIST_HEAD_STATIC(be_list, cdr_beitem);

/*! \brief List of registered modifiers */
static AST_RWLIST_HEAD_STATIC(mo_list, cdr_beitem);

/*! \brief Queued CDR waiting to be batched */
struct cdr_batch_item {
	struct ast_cdr *cdr;
	struct cdr_batch_item *next;
};

/*! \brief The actual batch queue */
static struct cdr_batch {
	int size;
	struct cdr_batch_item *head;
	struct cdr_batch_item *tail;
} *batch = NULL;

/*! \brief The global sequence counter used for CDRs */
static int global_cdr_sequence =  0;

/*! \brief Scheduler items */
static struct ast_sched_context *sched;
static int cdr_sched = -1;
AST_MUTEX_DEFINE_STATIC(cdr_sched_lock);
static pthread_t cdr_thread = AST_PTHREADT_NULL;

#ifdef GRANDSTREAM_NETWORKS
static int cdr_batch_disable = 0;
#endif

/*! \brief Lock protecting modifications to the batch queue */
AST_MUTEX_DEFINE_STATIC(cdr_batch_lock);

/*! \brief These are used to wake up the CDR thread when there's work to do */
AST_MUTEX_DEFINE_STATIC(cdr_pending_lock);
static ast_cond_t cdr_pending_cond;

/*! \brief A container of the active master CDRs indexed by Party A channel uniqueid */
static struct ao2_container *active_cdrs_master;

/*! \brief A container of all active CDRs with a Party B indexed by Party B channel name */
static struct ao2_container *active_cdrs_all;

/*! \brief Message router for stasis messages regarding channel state */
static struct stasis_message_router *stasis_router;

/*! \brief Our subscription for bridges */
static struct stasis_forward *bridge_subscription;

/*! \brief Our subscription for channels */
static struct stasis_forward *channel_subscription;

/*! \brief Our subscription for parking */
static struct stasis_forward *parking_subscription;

#ifdef GRANDSTREAM_NETWORKS
static struct stasis_forward *mixmonitor_subscription;
#endif

/*! \brief The parent topic for all topics we want to aggregate for CDRs */
static struct stasis_topic *cdr_topic;

/*! \brief A message type used to synchronize with the CDR topic */
STASIS_MESSAGE_TYPE_DEFN_LOCAL(cdr_sync_message_type);

struct cdr_object;

/*! \brief Return types for \ref process_bridge_enter functions */
enum process_bridge_enter_results {
	/*!
	 * The CDR was the only party in the bridge.
	 */
	BRIDGE_ENTER_ONLY_PARTY,
	/*!
	 * The CDR was able to obtain a Party B from some other party already in the bridge
	 */
	BRIDGE_ENTER_OBTAINED_PARTY_B,
	/*!
	 * The CDR was not able to obtain a Party B
	 */
	BRIDGE_ENTER_NO_PARTY_B,
	/*!
	 * This CDR can't handle a bridge enter message and a new CDR needs to be created
	 */
	BRIDGE_ENTER_NEED_CDR,
};

/*!
 * \brief A virtual table used for \ref cdr_object.
 *
 * Note that all functions are optional - if a subclass does not need an
 * implementation, it is safe to leave it NULL.
 */
struct cdr_object_fn_table {
	/*! \brief Name of the subclass */
	const char *name;

	/*!
	 * \brief An initialization function. This will be called automatically
	 * when a \ref cdr_object is switched to this type in
	 * \ref cdr_object_transition_state
	 *
	 * \param cdr The \ref cdr_object that was just transitioned
	 */
	void (* const init_function)(struct cdr_object *cdr);

	/*!
	 * \brief Process a Party A update for the \ref cdr_object
	 *
	 * \param cdr The \ref cdr_object to process the update
	 * \param snapshot The snapshot for the CDR's Party A
	 * \retval 0 the CDR handled the update or ignored it
	 * \retval 1 the CDR is finalized and a new one should be made to handle it
	 */
	int (* const process_party_a)(struct cdr_object *cdr,
			struct ast_channel_snapshot *snapshot);

	/*!
	 * \brief Process a Party B update for the \ref cdr_object
	 *
	 * \param cdr The \ref cdr_object to process the update
	 * \param snapshot The snapshot for the CDR's Party B
	 */
	void (* const process_party_b)(struct cdr_object *cdr,
			struct ast_channel_snapshot *snapshot);

	/*!
	 * \brief Process the beginning of a dial. A dial message implies one of two
	 * things:
	 * The \ref cdr_object's Party A has been originated
	 * The \ref cdr_object's Party A is dialing its Party B
	 *
	 * \param cdr The \ref cdr_object
	 * \param caller The originator of the dial attempt
	 * \param peer The destination of the dial attempt
	 *
	 * \retval 0 if the parties in the dial were handled by this CDR
	 * \retval 1 if the parties could not be handled by this CDR
	 */
	int (* const process_dial_begin)(struct cdr_object *cdr,
			struct ast_channel_snapshot *caller,
			struct ast_channel_snapshot *peer);

	/*!
	 * \brief Process the end of a dial. At the end of a dial, a CDR can be
	 * transitioned into one of two states - DialedPending
	 * (\ref dialed_pending_state_fn_table) or Finalized
	 * (\ref finalized_state_fn_table).
	 *
	 * \param cdr The \ref cdr_object
	 * \param caller The originator of the dial attempt
	 * \param peer the Destination of the dial attempt
	 * \param dial_status What happened
	 *
	 * \retval 0 if the parties in the dial were handled by this CDR
	 * \retval 1 if the parties could not be handled by this CDR
	 */
	int (* const process_dial_end)(struct cdr_object *cdr,
			struct ast_channel_snapshot *caller,
			struct ast_channel_snapshot *peer,
			const char *dial_status);

	/*!
	 * \brief Process the entering of a bridge by this CDR. The purpose of this
	 * callback is to have the CDR prepare itself for the bridge and attempt to
	 * find a valid Party B. The act of creating new CDRs based on the entering
	 * of this channel into the bridge is handled by the higher level message
	 * handler.
	 *
	 * Note that this handler is for when a channel enters into a "normal"
	 * bridge, where people actually talk to each other. Parking is its own
	 * thing.
	 *
	 * \param cdr The \ref cdr_object
	 * \param bridge The bridge that the Party A just entered into
	 * \param channel The \ref ast_channel_snapshot for this CDR's Party A
	 *
	 * \retval process_bridge_enter_results Defines whether or not this CDR was able
	 * to fully handle the bridge enter message.
	 */
	enum process_bridge_enter_results (* const process_bridge_enter)(
			struct cdr_object *cdr,
			struct ast_bridge_snapshot *bridge,
			struct ast_channel_snapshot *channel);

	/*!
	 * \brief Process entering into a parking bridge.
	 *
	 * \param cdr The \ref cdr_object
	 * \param bridge The parking bridge that Party A just entered into
	 * \param channel The \ref ast_channel_snapshot for this CDR's Party A
	 *
	 * \retval 0 This CDR successfully transitioned itself into the parked state
	 * \retval 1 This CDR couldn't handle the parking transition and we need a
	 *  new CDR.
	 */
	int (* const process_parking_bridge_enter)(struct cdr_object *cdr,
			struct ast_bridge_snapshot *bridge,
			struct ast_channel_snapshot *channel);

	/*!
	 * \brief Process the leaving of a bridge by this CDR.
	 *
	 * \param cdr The \ref cdr_object
	 * \param bridge The bridge that the Party A just left
	 * \param channel The \ref ast_channel_snapshot for this CDR's Party A
	 *
	 * \retval 0 This CDR left successfully
	 * \retval 1 Error
	 */
	int (* const process_bridge_leave)(struct cdr_object *cdr,
			struct ast_bridge_snapshot *bridge,
			struct ast_channel_snapshot *channel);

	/*!
	 * \brief Process an update informing us that the channel got itself parked
	 *
	 * \param cdr The \ref cdr_object
	 * \param channel The parking information for this CDR's party A
	 *
	 * \retval 0 This CDR successfully parked itself
	 * \retval 1 This CDR couldn't handle the park
	 */
	int (* const process_parked_channel)(struct cdr_object *cdr,
			struct ast_parked_call_payload *parking_info);
};

static int base_process_party_a(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot);
static enum process_bridge_enter_results base_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);
static int base_process_bridge_leave(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);
static int base_process_dial_end(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer, const char *dial_status);
static int base_process_parked_channel(struct cdr_object *cdr, struct ast_parked_call_payload *parking_info);

static void single_state_init_function(struct cdr_object *cdr);
static void single_state_process_party_b(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot);
static int single_state_process_dial_begin(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer);
static enum process_bridge_enter_results single_state_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);
static int single_state_process_parking_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);

/*!
 * \brief The virtual table for the Single state.
 *
 * A \ref cdr_object starts off in this state. This represents a channel that
 * has no Party B information itself.
 *
 * A \ref cdr_object from this state can go into any of the following states:
 * * \ref dial_state_fn_table
 * * \ref bridge_state_fn_table
 * * \ref finalized_state_fn_table
 */
struct cdr_object_fn_table single_state_fn_table = {
	.name = "Single",
	.init_function = single_state_init_function,
	.process_party_a = base_process_party_a,
	.process_party_b = single_state_process_party_b,
	.process_dial_begin = single_state_process_dial_begin,
	.process_dial_end = base_process_dial_end,
	.process_bridge_enter = single_state_process_bridge_enter,
	.process_parking_bridge_enter = single_state_process_parking_bridge_enter,
	.process_bridge_leave = base_process_bridge_leave,
	.process_parked_channel = base_process_parked_channel,
};

static void dial_state_process_party_b(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot);
static int dial_state_process_dial_begin(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer);
static int dial_state_process_dial_end(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer, const char *dial_status);
static enum process_bridge_enter_results dial_state_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);

/*!
 * \brief The virtual table for the Dial state.
 *
 * A \ref cdr_object that has begun a dial operation. This state is entered when
 * the Party A for a CDR is determined to be dialing out to a Party B or when
 * a CDR is for an originated channel (in which case the Party A information is
 * the originated channel, and there is no Party B).
 *
 * A \ref cdr_object from this state can go in any of the following states:
 * * \ref dialed_pending_state_fn_table
 * * \ref bridge_state_fn_table
 * * \ref finalized_state_fn_table
 */
struct cdr_object_fn_table dial_state_fn_table = {
	.name = "Dial",
	.process_party_a = base_process_party_a,
	.process_party_b = dial_state_process_party_b,
	.process_dial_begin = dial_state_process_dial_begin,
	.process_dial_end = dial_state_process_dial_end,
	.process_bridge_enter = dial_state_process_bridge_enter,
	.process_bridge_leave = base_process_bridge_leave,
};

static int dialed_pending_state_process_party_a(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot);
static int dialed_pending_state_process_dial_begin(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer);
static enum process_bridge_enter_results dialed_pending_state_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);
static int dialed_pending_state_process_parking_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);

/*!
 * \brief The virtual table for the Dialed Pending state.
 *
 * A \ref cdr_object that has successfully finished a dial operation, but we
 * don't know what they're going to do yet. It's theoretically possible to dial
 * a party and then have that party not be bridged with the caller; likewise,
 * an origination can complete and the channel go off and execute dialplan. The
 * pending state acts as a bridge between either:
 * * Entering a bridge
 * * Getting a new CDR for new dialplan execution
 * * Switching from being originated to executing dialplan
 *
 * A \ref cdr_object from this state can go in any of the following states:
 * * \ref single_state_fn_table
 * * \ref dialed_pending_state_fn_table
 * * \ref bridge_state_fn_table
 * * \ref finalized_state_fn_table
 */
struct cdr_object_fn_table dialed_pending_state_fn_table = {
	.name = "DialedPending",
	.process_party_a = dialed_pending_state_process_party_a,
	.process_dial_begin = dialed_pending_state_process_dial_begin,
	.process_bridge_enter = dialed_pending_state_process_bridge_enter,
	.process_parking_bridge_enter = dialed_pending_state_process_parking_bridge_enter,
	.process_bridge_leave = base_process_bridge_leave,
	.process_parked_channel = base_process_parked_channel,
};

static void bridge_state_process_party_b(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot);
static int bridge_state_process_bridge_leave(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);

/*!
 * \brief The virtual table for the Bridged state
 *
 * A \ref cdr_object enters this state when it receives notification that the
 * channel has entered a bridge.
 *
 * A \ref cdr_object from this state can go to:
 * * \ref finalized_state_fn_table
 */
struct cdr_object_fn_table bridge_state_fn_table = {
	.name = "Bridged",
	.process_party_a = base_process_party_a,
	.process_party_b = bridge_state_process_party_b,
	.process_bridge_leave = bridge_state_process_bridge_leave,
	.process_parked_channel = base_process_parked_channel,
};

static int parked_state_process_bridge_leave(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel);

/*!
 * \brief The virtual table for the Parked state
 *
 * Parking is weird. Unlike typical bridges, it has to be treated somewhat
 * uniquely - a channel in a parking bridge (which is a subclass of a holding
 * bridge) has to be handled as if the channel went into an application.
 * However, when the channel comes out, we need a new CDR - unlike the Single
 * state.
 */
struct cdr_object_fn_table parked_state_fn_table = {
	.name = "Parked",
	.process_party_a = base_process_party_a,
	.process_bridge_leave = parked_state_process_bridge_leave,
	.process_parked_channel = base_process_parked_channel,
};

static void finalized_state_init_function(struct cdr_object *cdr);
static int finalized_state_process_party_a(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot);
#ifdef GRANDSTREAM_NETWORKS
static enum ast_cdr_disposition dial_status_to_disposition(const char *dial_status);
#endif

/*!
 * \brief The virtual table for the finalized state.
 *
 * Once in the finalized state, the CDR is done. No modifications can be made
 * to the CDR.
 */
struct cdr_object_fn_table finalized_state_fn_table = {
	.name = "Finalized",
	.init_function = finalized_state_init_function,
	.process_party_a = finalized_state_process_party_a,
	.process_bridge_enter = base_process_bridge_enter,
};

/*! \brief A wrapper object around a snapshot.
 * Fields that are mutable by the CDR engine are replicated here.
 */
struct cdr_object_snapshot {
	struct ast_channel_snapshot *snapshot;  /*!< The channel snapshot */
	char userfield[AST_MAX_USER_FIELD];     /*!< Userfield for the channel */
	unsigned int flags;                     /*!< Specific flags for this party */
	struct varshead variables;              /*!< CDR variables for the channel */
};

/*! \brief An in-memory representation of an active CDR */
struct cdr_object {
	struct cdr_object_snapshot party_a;     /*!< The Party A information */
	struct cdr_object_snapshot party_b;     /*!< The Party B information */
	struct cdr_object_fn_table *fn_table;   /*!< The current virtual table */

	enum ast_cdr_disposition disposition;   /*!< The disposition of the CDR */
	struct timeval start;                   /*!< When this CDR was created */
	struct timeval answer;                  /*!< Either when the channel was answered, or when the path between channels was established */
	struct timeval end;                     /*!< When this CDR was finalized */
	unsigned int sequence;                  /*!< A monotonically increasing number for each CDR */
	struct ast_flags flags;                 /*!< Flags on the CDR */
	AST_DECLARE_STRING_FIELDS(
		AST_STRING_FIELD(linkedid);         /*!< Linked ID. Cached here as it may change out from party A, which must be immutable */
		AST_STRING_FIELD(uniqueid);         /*!< Unique id of party A. Cached here as it is the master CDR container key */
		AST_STRING_FIELD(name);             /*!< Channel name of party A. Cached here as the party A address may change */
		AST_STRING_FIELD(bridge);           /*!< The bridge the party A happens to be in. */
		AST_STRING_FIELD(appl);             /*!< The last accepted application party A was in */
		AST_STRING_FIELD(data);             /*!< The data for the last accepted application party A was in */
		AST_STRING_FIELD(context);          /*!< The accepted context for Party A */
		AST_STRING_FIELD(exten);            /*!< The accepted extension for Party A */
		AST_STRING_FIELD(party_b_name);     /*!< Party B channel name. Cached here as it is the all CDRs container key */
#ifdef GRANDSTREAM_NETWORKS
		AST_STRING_FIELD(session);
		/*add by wxu, add two words for cdr table*/
		AST_STRING_FIELD(action_owner);
		AST_STRING_FIELD(action_type);
		AST_STRING_FIELD(special_action_type);
		AST_STRING_FIELD(special_dst);
		AST_STRING_FIELD(inbound_trunk_index);
		AST_STRING_FIELD(outbound_trunk_index);
		AST_STRING_FIELD(status);
		AST_STRING_FIELD(device_info);      /*!<for caller device info*/
		AST_STRING_FIELD(device_info_peer); /*!<for callee device info*/
		/*add end*/
#endif
	);
#ifdef GRANDSTREAM_NETWORKS
	/*add by wxu@grandstream.cn for callee and caller*/
	char app_caller[AST_MAX_ACCOUNT_CODE];
	char app_callee[AST_MAX_ACCOUNT_CODE];
	char feature_transfer[32];
	char record_files[1024];
	char record_files_desc[4096];
	char record_files_uuid[4096];
	char reason[32];
	unsigned int dst_need;
	unsigned int src_need;
	/*end*/
	ast_mutex_t party_lock;

	int is_nat_call;
	int peer_is_nat_call;
	int is_video_call;
	int is_gds_call;
	int is_gsc_call;
	struct timeval recovery_start;                   /*!< When this CDR was originally created */
	struct timeval recovery_answer;                  /*!<  when the channel was originally answered */
#endif
	struct cdr_object *next;                /*!< The next CDR object in the chain */
	struct cdr_object *last;                /*!< The last CDR object in the chain */
	int is_root;                            /*!< True if this is the first CDR in the chain */
};

/*!
 * \brief Copy variables from one list to another
 * \param to_list destination
 * \param from_list source
 * \retval The number of copied variables
 */
static int copy_variables(struct varshead *to_list, struct varshead *from_list)
{
	struct ast_var_t *variables;
	struct ast_var_t *newvariable;
	const char *var;
	const char *val;
	int x = 0;

	AST_LIST_TRAVERSE(from_list, variables, entries) {
		var = ast_var_name(variables);
		if (ast_strlen_zero(var)) {
			continue;
		}
		val = ast_var_value(variables);
		if (ast_strlen_zero(val)) {
			continue;
		}
		newvariable = ast_var_assign(var, val);
		if (newvariable) {
			AST_LIST_INSERT_HEAD(to_list, newvariable, entries);
			++x;
		}
	}

	return x;
}

/*!
 * \brief Delete all variables from a variable list
 * \param headp The head pointer to the variable list to delete
 */
static void free_variables(struct varshead *headp)
{
	struct ast_var_t *vardata;

	while ((vardata = AST_LIST_REMOVE_HEAD(headp, entries))) {
		ast_var_delete(vardata);
	}
}

/*!
 * \brief Copy a snapshot and its details
 * \param dst The destination
 * \param src The source
 */
#ifdef GRANDSTREAM_NETWORKS
static void cdr_object_snapshot_copy(struct cdr_object *cdr, struct cdr_object_snapshot *dst, struct cdr_object_snapshot *src)
#else
static void cdr_object_snapshot_copy(struct cdr_object_snapshot *dst, struct cdr_object_snapshot *src)
#endif
{
#ifdef GRANDSTREAM_NETWORKS
	SCOPED_MUTEX(lock, &cdr->party_lock);
#endif
	ao2_t_replace(dst->snapshot, src->snapshot, "CDR snapshot copy");
	strcpy(dst->userfield, src->userfield);
	dst->flags = src->flags;
	copy_variables(&dst->variables, &src->variables);
}

/*!
 * \brief Transition a \ref cdr_object to a new state
 * \param cdr The \ref cdr_object to transition
 * \param fn_table The \ref cdr_object_fn_table state to go to
 */
static void cdr_object_transition_state(struct cdr_object *cdr, struct cdr_object_fn_table *fn_table)
{
	ast_log(LOG_DEBUG, "cdr:%p - Transitioning CDR for %s from state %s to %s\n",
		cdr, cdr->party_a.snapshot->name,
		cdr->fn_table ? cdr->fn_table->name : "NONE", fn_table->name);
#ifdef GRANDSTREAM_NETWORKS
	if (cdr->action_type && !strncmp(cdr->action_type, "RINGGROUP", strlen("RINGGROUP"))
		&& cdr->fn_table && cdr->fn_table->name && !strcmp(cdr->fn_table->name, "Single")
		&& fn_table && fn_table->name && !strcmp(fn_table->name, "Dial")){
		cdr->start = ast_tvnow();
		ast_log(LOG_NOTICE, "reset the ringgroup member start time");
	}
#endif
	cdr->fn_table = fn_table;
#ifdef GRANDSTREAM_NETWORKS
	if (cdr->action_type && !strncmp(cdr->action_type, "QUEUE", strlen("QUEUE")) && !strcmp(fn_table->name, "Bridged")){
		ast_log(LOG_NOTICE, "reset queue member answer time");
		cdr->answer = ast_tvnow();
	}
	if (!strcmp(fn_table->name, "Bridged")
		&& (cdr->party_b.snapshot && cdr->party_b.snapshot->name && !strncmp(cdr->party_b.snapshot->name, "MulticastRTP", strlen("MulticastRTP")))){
		ast_log(LOG_NOTICE, "reset multicast paging answer time");
		cdr->answer = ast_tvnow();
	}
#endif
	if (cdr->fn_table->init_function) {
		cdr->fn_table->init_function(cdr);
	}
}

/*!
 * \internal
 * \brief Hash function for master CDR container indexed by Party A uniqueid.
 */
static int cdr_master_hash_fn(const void *obj, const int flags)
{
	const struct cdr_object *cdr;
	const char *key;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_KEY:
		key = obj;
		break;
	case OBJ_SEARCH_OBJECT:
		cdr = obj;
		key = cdr->uniqueid;
		break;
	default:
		ast_assert(0);
		return 0;
	}
	return ast_str_case_hash(key);
}

/*!
 * \internal
 * \brief Comparison function for master CDR container indexed by Party A uniqueid.
 */
static int cdr_master_cmp_fn(void *obj, void *arg, int flags)
{
    struct cdr_object *left = obj;
    struct cdr_object *right = arg;
    const char *right_key = arg;
    int cmp;

    switch (flags & OBJ_SEARCH_MASK) {
    case OBJ_SEARCH_OBJECT:
        right_key = right->uniqueid;
        /* Fall through */
    case OBJ_SEARCH_KEY:
        cmp = strcmp(left->uniqueid, right_key);
        break;
    case OBJ_SEARCH_PARTIAL_KEY:
        /*
         * We could also use a partial key struct containing a length
         * so strlen() does not get called for every comparison instead.
         */
        cmp = strncmp(left->uniqueid, right_key, strlen(right_key));
        break;
    default:
        /* Sort can only work on something with a full or partial key. */
        ast_assert(0);
        cmp = 0;
        break;
    }
    return cmp ? 0 : CMP_MATCH;
}

/*!
 * \internal
 * \brief Hash function for all CDR container indexed by Party B channel name.
 */
static int cdr_all_hash_fn(const void *obj, const int flags)
{
	const struct cdr_object *cdr;
	const char *key;

	switch (flags & OBJ_SEARCH_MASK) {
	case OBJ_SEARCH_KEY:
		key = obj;
		break;
	case OBJ_SEARCH_OBJECT:
		cdr = obj;
		key = cdr->party_b_name;
		break;
	default:
		ast_assert(0);
		return 0;
	}
	return ast_str_case_hash(key);
}

/*!
 * \internal
 * \brief Comparison function for all CDR container indexed by Party B channel name.
 */
static int cdr_all_cmp_fn(void *obj, void *arg, int flags)
{
    struct cdr_object *left = obj;
    struct cdr_object *right = arg;
    const char *right_key = arg;
    int cmp;

    switch (flags & OBJ_SEARCH_MASK) {
    case OBJ_SEARCH_OBJECT:
        right_key = right->party_b_name;
        /* Fall through */
    case OBJ_SEARCH_KEY:
        cmp = strcasecmp(left->party_b_name, right_key);
        break;
    case OBJ_SEARCH_PARTIAL_KEY:
        /*
         * We could also use a partial key struct containing a length
         * so strlen() does not get called for every comparison instead.
         */
        cmp = strncasecmp(left->party_b_name, right_key, strlen(right_key));
        break;
    default:
        /* Sort can only work on something with a full or partial key. */
        ast_assert(0);
        cmp = 0;
        break;
    }
    return cmp ? 0 : CMP_MATCH;
}

/*!
 * \internal
 * \brief Relink the CDR because Party B's snapshot changed.
 * \since 13.19.0
 *
 * \return Nothing
 */
static void cdr_all_relink(struct cdr_object *cdr)
{
	ao2_lock(active_cdrs_all);

#ifdef GRANDSTREAM_NETWORKS
	ao2_ref(cdr, +1); // Added by pxie@grandstream.cn for bug[144956], 2019-08-26.
#endif

	if (cdr->party_b.snapshot) {
		if (strcasecmp(cdr->party_b_name, cdr->party_b.snapshot->name)) {
			ao2_unlink_flags(active_cdrs_all, cdr, OBJ_NOLOCK);
			ast_string_field_set(cdr, party_b_name, cdr->party_b.snapshot->name);
			ao2_link_flags(active_cdrs_all, cdr, OBJ_NOLOCK);
		}
	} else {
		ao2_unlink_flags(active_cdrs_all, cdr, OBJ_NOLOCK);
		ast_string_field_set(cdr, party_b_name, "");
	}

#ifdef GRANDSTREAM_NETWORKS
	ao2_ref(cdr, -1); // Added by pxie@grandstream.cn for bug[144956], 2019-08-26.
#endif

	ao2_unlock(active_cdrs_all);
}

/*!
 * \internal
 * \brief Unlink the master CDR and chained records from the active_cdrs_all container.
 * \since 13.19.0
 *
 * \return Nothing
 */
static void cdr_all_unlink(struct cdr_object *cdr)
{
	struct cdr_object *cur;
	struct cdr_object *next;

	ast_assert(cdr->is_root);

	/* Hold a ref to the root CDR to ensure the list members don't go away on us. */
	ao2_ref(cdr, +1);
	ao2_lock(active_cdrs_all);
	for (cur = cdr; cur; cur = next) {
		next = cur->next;
		ao2_unlink_flags(active_cdrs_all, cur, OBJ_NOLOCK);
		/*
		 * It is safe to still use cur after unlinking because the
		 * root CDR holds a ref to all the CDRs in the list and we
		 * have a ref to the root CDR.
		 */
		ast_string_field_set(cur, party_b_name, "");
	}
	ao2_unlock(active_cdrs_all);
	ao2_ref(cdr, -1);
}

/*!
 * \brief \ref cdr_object Destructor
 */
static void cdr_object_dtor(void *obj)
{
	struct cdr_object *cdr = obj;
	struct ast_var_t *it_var;

	ao2_cleanup(cdr->party_a.snapshot);
	ao2_cleanup(cdr->party_b.snapshot);
	while ((it_var = AST_LIST_REMOVE_HEAD(&cdr->party_a.variables, entries))) {
		ast_var_delete(it_var);
	}
	while ((it_var = AST_LIST_REMOVE_HEAD(&cdr->party_b.variables, entries))) {
		ast_var_delete(it_var);
	}
	ast_string_field_free_memory(cdr);

	/* CDR destruction used to work by calling ao2_cleanup(next) and
	 * allowing the chain to destroy itself neatly. Unfortunately, for
	 * really long chains, this can result in a stack overflow. So now
	 * when the root CDR is destroyed, it is responsible for unreffing
	 * all CDRs in the chain
	 */
	if (cdr->is_root) {
		struct cdr_object *curr = cdr->next;
		struct cdr_object *next;

		while (curr) {
			next = curr->next;
			ao2_cleanup(curr);
			curr = next;
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_mutex_destroy(&cdr->party_lock);
#endif
}

/*!
 * \brief \ref cdr_object constructor
 * \param chan The \ref ast_channel_snapshot that is the CDR's Party A
 *
 * This implicitly sets the state of the newly created CDR to the Single state
 * (\ref single_state_fn_table)
 */
static struct cdr_object *cdr_object_alloc(struct ast_channel_snapshot *chan)
{
	struct cdr_object *cdr;

	ast_assert(chan != NULL);

	cdr = ao2_alloc(sizeof(*cdr), cdr_object_dtor);
	if (!cdr) {
		return NULL;
	}
	cdr->last = cdr;
	if (ast_string_field_init(cdr, 64)) {
		ao2_cleanup(cdr);
		return NULL;
	}
	ast_string_field_set(cdr, uniqueid, chan->uniqueid);
	ast_string_field_set(cdr, name, chan->name);
	ast_string_field_set(cdr, linkedid, chan->linkedid);
	cdr->disposition = AST_CDR_NULL;
	cdr->sequence = ast_atomic_fetchadd_int(&global_cdr_sequence, +1);

	cdr->party_a.snapshot = chan;
	ao2_t_ref(cdr->party_a.snapshot, +1, "bump snapshot during CDR creation");

	ast_log(LOG_DEBUG, "%p - Created CDR for channel %s\n", cdr, chan->name);

	cdr_object_transition_state(cdr, &single_state_fn_table);

	return cdr;
}

/*!
 * \brief Create a new \ref cdr_object and append it to an existing chain
 * \param cdr The \ref cdr_object to append to
 */
static struct cdr_object *cdr_object_create_and_append(struct cdr_object *cdr)
{
	struct cdr_object *new_cdr;
	struct cdr_object *it_cdr;
	struct cdr_object *cdr_last;

	cdr_last = cdr->last;
	new_cdr = cdr_object_alloc(cdr_last->party_a.snapshot);
	if (!new_cdr) {
		return NULL;
	}
	new_cdr->disposition = AST_CDR_NULL;

	/* Copy over the linkedid, as it may have changed */
	ast_string_field_set(new_cdr, linkedid, cdr_last->linkedid);
	ast_string_field_set(new_cdr, appl, cdr_last->appl);
	ast_string_field_set(new_cdr, data, cdr_last->data);
	ast_string_field_set(new_cdr, context, cdr_last->context);
#ifndef GRANDSTREAM_NETWORKS
	ast_string_field_set(new_cdr, exten, cdr_last->exten);
#endif

#ifdef GRANDSTREAM_NETWORKS
	/*add by wxu*/
	ast_string_field_set(new_cdr, session, cdr_last->session);
	ast_string_field_set(new_cdr, action_type, cdr_last->action_type);
	ast_string_field_set(new_cdr, special_action_type, cdr_last->special_action_type);
	ast_string_field_set(new_cdr, special_dst, cdr_last->special_dst);
	ast_string_field_set(new_cdr, inbound_trunk_index, cdr_last->inbound_trunk_index);
	ast_string_field_set(new_cdr, outbound_trunk_index, cdr_last->outbound_trunk_index);
	ast_string_field_set(new_cdr, status, cdr_last->status);
	/*add end*/

	if( !strcasecmp(cdr_last->feature_transfer,"yes")){
		ast_log(LOG_VERBOSE,"cdr disable");
		ast_set_flag(&new_cdr->flags,AST_CDR_FLAG_DISABLE);
	}
	/*the sca only don't need private call record*/
	if (0 == strcmp(new_cdr->action_type, "SCA") || ((!strcmp(new_cdr->action_type, "TRANSFER") || !strncmp(new_cdr->action_type, "IVR", strlen("IVR"))) && !strcmp(new_cdr->special_action_type, "SCA")))
	{
		ast_log(LOG_VERBOSE,"cdr disable");
		ast_set_flag(&new_cdr->flags,AST_CDR_FLAG_DISABLE);
	}

	new_cdr->dst_need = 0;
	new_cdr->src_need = 0;

	if( cdr_last->dst_need ){
		if( strlen(cdr_last->app_callee) > 0 ){
			snprintf(new_cdr->app_callee,sizeof(new_cdr->app_callee),"%s",cdr_last->app_callee);
		}
	}
	/*modify by zyxia for Bug 113157, announcements center to sca ,cdr error,2018.8.4 */
	if( cdr_last->src_need || !strcasecmp(cdr_last->special_action_type,"SPECIAL_SCA")
		|| !strcasecmp(cdr_last->special_action_type, "atxee")){
		if( strlen(cdr_last->app_caller) > 0 ){
			snprintf(new_cdr->app_caller,sizeof(new_cdr->app_caller),"%s",cdr_last->app_caller);
		}
	}

	if(strlen(cdr_last->record_files) > 0){
		ast_log(LOG_DEBUG, "In cdr_object_create_and_append:last_cdr(%p) record_file:%s",
			cdr_last, cdr_last->record_files);

		if (cdr_last->party_b.snapshot && 0 == strncmp(new_cdr->action_type, "DISA", strlen("DISA"))){
			char tmp_buff[512] = {0};
			ast_cdr_getvar(cdr_last->party_b.snapshot->name, "record_files", tmp_buff,sizeof(tmp_buff));
			ast_log(LOG_NOTICE, "the party b record file name:%s", tmp_buff);
			snprintf(new_cdr->record_files,sizeof(new_cdr->record_files),"%s",tmp_buff);
		} else{
			snprintf(new_cdr->record_files,sizeof(new_cdr->record_files),"%s",cdr_last->record_files);
		}
	}

	if (!ast_strlen_zero(cdr_last->special_action_type)
		&& !strcmp(cdr_last->special_action_type, "WAVE_SKIP_CALL")) {
		ast_log(LOG_DEBUG, "DO NOT inherit special_action_type of 'WAVE_SKIP_CALL' to the new one");
		ast_string_field_set(new_cdr, special_action_type, "");
	}
#endif

	/*
	 * If the current CDR says to disable all future ones,
	 * keep the disable chain going
	 */
	if (ast_test_flag(&cdr_last->flags, AST_CDR_FLAG_DISABLE_ALL)) {
		ast_set_flag(&new_cdr->flags, AST_CDR_FLAG_DISABLE_ALL);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_WARNING, "new cdr(%p|%s) disable all flags", new_cdr, new_cdr->name);
#endif
	}

	/* Copy over other Party A information */
#ifdef GRANDSTREAM_NETWORKS
	cdr_object_snapshot_copy(new_cdr, &new_cdr->party_a, &cdr_last->party_a);
#else
	cdr_object_snapshot_copy(&new_cdr->party_a, &cdr_last->party_a);
#endif

	/* Append the CDR to the end of the list */
	for (it_cdr = cdr; it_cdr->next; it_cdr = it_cdr->next) {
		it_cdr->last = new_cdr;
	}
	it_cdr->last = new_cdr;
	it_cdr->next = new_cdr;

	ast_log(LOG_DEBUG, "[%s]Created new cdr(%p) and appended to cdr(%p)",
		it_cdr->name, new_cdr, it_cdr);

	return new_cdr;
}

/*!
 * \internal
 * \brief Determine if CDR flag is configured.
 *
 * \param cdr_flag The configured CDR flag to check.
 *
 * \retval 0 if the CDR flag is not configured.
 * \retval non-zero if the CDR flag is configured.
 *
 * \return Nothing
 */
static int is_cdr_flag_set(unsigned int cdr_flag)
{
	struct module_config *mod_cfg;
	int flag_set;

	mod_cfg = ao2_global_obj_ref(module_configs);
	flag_set = mod_cfg && ast_test_flag(&mod_cfg->general->settings, cdr_flag);
	ao2_cleanup(mod_cfg);
	return flag_set;
}

/*!
 * \brief Return whether or not a channel has changed its state in the dialplan, subject
 * to endbeforehexten logic
 *
 * \param old_snapshot The previous state
 * \param new_snapshot The new state
 *
 * \retval 0 if the state has not changed
 * \retval 1 if the state changed
 */
static int snapshot_cep_changed(struct ast_channel_snapshot *old_snapshot,
	struct ast_channel_snapshot *new_snapshot)
{
	/* If we ignore hangup logic, don't indicate that we're executing anything new */
	if (ast_test_flag(&new_snapshot->softhangup_flags, AST_SOFTHANGUP_HANGUP_EXEC)
		&& is_cdr_flag_set(CDR_END_BEFORE_H_EXTEN)) {
		return 0;
	}

	/* When Party A is originated to an application and the application exits, the stack
	 * will attempt to clear the application and restore the dummy originate application
	 * of "AppDialX". Ignore application changes to AppDialX as a result.
	 */
	if (strcmp(new_snapshot->appl, old_snapshot->appl)
		&& strncasecmp(new_snapshot->appl, "appdial", 7)
		&& (strcmp(new_snapshot->context, old_snapshot->context)
			|| strcmp(new_snapshot->exten, old_snapshot->exten)
			|| new_snapshot->priority != old_snapshot->priority)) {
		return 1;
	}

	return 0;
}

/*!
 * \brief Return whether or not a \ref ast_channel_snapshot is for a channel
 * that was created as the result of a dial operation
 *
 * \retval 0 the channel was not created as the result of a dial
 * \retval 1 the channel was created as the result of a dial
 */
static int snapshot_is_dialed(struct ast_channel_snapshot *snapshot)
{
	return (ast_test_flag(&snapshot->flags, AST_FLAG_OUTGOING)
			&& !(ast_test_flag(&snapshot->flags, AST_FLAG_ORIGINATED)));
}

/*!
 * \brief Given two CDR snapshots, figure out who should be Party A for the
 * resulting CDR
 * \param left One of the snapshots
 * \param right The other snapshot
 * \retval The snapshot that won
 */
static struct cdr_object_snapshot *cdr_object_pick_party_a(struct cdr_object_snapshot *left, struct cdr_object_snapshot *right)
{
	/* Check whether or not the party is dialed. A dialed party is never the
	 * Party A with a party that was not dialed.
	 */
	if (!snapshot_is_dialed(left->snapshot) && snapshot_is_dialed(right->snapshot)) {
		return left;
	} else if (snapshot_is_dialed(left->snapshot) && !snapshot_is_dialed(right->snapshot)) {
		return right;
	}

	/* Try the Party A flag */
	if (ast_test_flag(left, AST_CDR_FLAG_PARTY_A) && !ast_test_flag(right, AST_CDR_FLAG_PARTY_A)) {
		return left;
	} else if (!ast_test_flag(right, AST_CDR_FLAG_PARTY_A) && ast_test_flag(right, AST_CDR_FLAG_PARTY_A)) {
		return right;
	}

	/* Neither party is dialed and neither has the Party A flag - defer to
	 * creation time */
	if (left->snapshot->creationtime.tv_sec < right->snapshot->creationtime.tv_sec) {
		return left;
	} else if (left->snapshot->creationtime.tv_sec > right->snapshot->creationtime.tv_sec) {
		return right;
	} else if (left->snapshot->creationtime.tv_usec > right->snapshot->creationtime.tv_usec) {
		return right;
	} else {
		/* Okay, fine, take the left one */
		return left;
	}
}

/*!
 * Compute the duration for a \ref cdr_object
 */
static long cdr_object_get_duration(struct cdr_object *cdr)
{
	return (long)(ast_tvdiff_ms(ast_tvzero(cdr->end) ? ast_tvnow() : cdr->end, cdr->start) / 1000);
}

/*!
 * \brief Compute the billsec for a \ref cdr_object
 */
static long cdr_object_get_billsec(struct cdr_object *cdr)
{
	long int ms;

	if (ast_tvzero(cdr->answer)) {
		return 0;
	}

	ms = ast_tvdiff_ms(ast_tvzero(cdr->end) ? ast_tvnow() : cdr->end, cdr->answer);
	if (ms % 1000 >= 500
		&& is_cdr_flag_set(CDR_INITIATED_SECONDS)) {
		ms = (ms / 1000) + 1;
	} else {
		ms = ms / 1000;
	}

	return ms;
}

/*!
 * \internal
 * \brief Set a variable on a CDR object
 *
 * \param headp The header pointer to the variable to set
 * \param name The name of the variable
 * \param value The value of the variable
 */
static void set_variable(struct varshead *headp, const char *name, const char *value)
{
	struct ast_var_t *newvariable;

	AST_LIST_TRAVERSE_SAFE_BEGIN(headp, newvariable, entries) {
		if (!strcasecmp(ast_var_name(newvariable), name)) {
			AST_LIST_REMOVE_CURRENT(entries);
			ast_var_delete(newvariable);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	if (value && (newvariable = ast_var_assign(name, value))) {
		AST_LIST_INSERT_HEAD(headp, newvariable, entries);
	}
}

#ifdef GRANDSTREAM_NETWORKS
/*
 *  This Function needs to be optimized.
 *  But currently we are about to official release, so we won't optimize. (20200604)
 */
static int cdr_check_record_file(struct ast_cdr *cdr)
{
	int pass = 0;
	int malloc_len = 256;  /* should be enough */
	char *cid_num_start;
	char *cid_num_stop;
	char *call_list = (char *)ast_malloc(malloc_len);

	if (!call_list) {
		ast_log(LOG_WARNING, "Insufficient memory");
		goto END;
	}

	/* case 1, normal one */
	snprintf(call_list, malloc_len, "%s-%s", cdr->src, cdr->dst);
	ast_log(LOG_DEBUG, "[cdr src/dst] call_list:%s", call_list);

	pass = strstr(cdr->record_files, call_list) ? 1 : 0;
	if (pass) {
		goto END;
	}

	if (strchr(call_list, '*') && strchr(cdr->record_files, '_')) {
		char *p = call_list;
		while ((p = strchr((const char *)p, '*')) != NULL) {
			*p = '_';
			p++;
		}

		ast_log(LOG_DEBUG, "['*' -> '_'] call_list:%s", call_list);
		pass = strstr(cdr->record_files, call_list) ? 1 : 0;
		if (pass) {
			goto END;
		}

		snprintf(call_list, malloc_len, "%s-%s", cdr->dst, cdr->src);
		p = call_list;
		while ((p = strchr((const char *)p, '*')) != NULL) {
			*p = '_';
			p++;
		}

		ast_log(LOG_DEBUG, "['*' -> '_'] [swap] call_list:%s", call_list);
		pass = strstr(cdr->record_files, call_list) ? 1 : 0;
		if (pass) {
			goto END;
		}
	}

	/* case 2 */
	cid_num_start = strchr(cdr->clid, '<');
	cid_num_stop = strchr(cdr->clid, '>');
	if (cid_num_start && cid_num_stop && cid_num_stop > cid_num_start) {
		int cid_num_len;
		char cid_num[AST_MAX_EXTENSION];

		cid_num_start++;
		cid_num_len = cid_num_stop-cid_num_start < AST_MAX_EXTENSION ? cid_num_stop-cid_num_start : AST_MAX_EXTENSION - 1;

		strncpy(cid_num, cid_num_start, cid_num_len);
		cid_num[cid_num_len] = '\0';

		ast_log(LOG_DEBUG, "cdr->clid_num: <%s>", cid_num);
		
		if (!strcmp(cid_num, "Unavailable") && strstr(cdr->channel, "DAHDI/o")) {
			ast_log(LOG_DEBUG, "DAHDI channel and cid_name is Unavailable, abandon src cid");
			cid_num[0] = '\0';
		}

		snprintf(call_list, malloc_len, "%s-%s", cid_num, cdr->dst);

		ast_log(LOG_DEBUG, "[case 2 - cid] call_list:%s", call_list);
		pass = strstr(cdr->record_files, call_list) ? 1 : 0;
		if (pass) {
			goto END;
		}

		snprintf(call_list, malloc_len, "%s-%s", cdr->dst, cid_num);
		ast_log(LOG_DEBUG, "[case 2 - cid] call_list(reverse):%s", call_list);
		pass = strstr(cdr->record_files, call_list) ? 1 : 0;
		if (pass) {
			goto END;
		}
	}

	/* case 3 */
	if (strstr(cdr->dstchannel, "/trunk_") || strstr(cdr->dstchannel, "DAHDI/o")) {
		cid_num_start = strrchr(cdr->record_files, '-');
		cid_num_stop = strchr(cdr->record_files, '.');

		if (cid_num_start && cid_num_stop && cid_num_stop > cid_num_start) {
			int len = cid_num_stop - cid_num_start -1;
			strncpy(call_list, cid_num_start+1, len);
			call_list[len] = '\0';

			ast_log(LOG_DEBUG, "[case 3] call_list:%s, cdr->dst:%s", call_list, cdr->dst);
			pass = strstr(cdr->dst, call_list) ? 1 : 0;
			if (pass) {
				goto END;
			}
		}
	}

	/* case 4 for Bug 164359*/
	if (strstr(cdr->channel, "/trunk_")) {
		cid_num_start = strrchr(cdr->dstchannel, '/');
		cid_num_stop = strchr(cdr->dstchannel, '-');

		if (cid_num_start && cid_num_stop && cid_num_stop > cid_num_start) {
			int len = cid_num_stop - cid_num_start -1;
			strncpy(call_list, cdr->src, malloc_len);
			strncat(call_list, "-", malloc_len - strlen(call_list) - 1);
			strncat(call_list, cid_num_start + 1, malloc_len - strlen(call_list) - 1);

			call_list[strlen(cdr->src) + 1 + len] = '\0';
			ast_log(LOG_DEBUG, "[case 4] call_list:%s", call_list);
			pass = strstr(cdr->record_files, call_list) ? 1 : 0;
			if (pass) {
				goto END;
			}
		}
	}

	if (!pass) {
		if ( (!ast_strlen_zero(cdr->special_action_type) && 
		     (!strcmp(cdr->special_action_type, "IVR_TIMEOUT") || 
			  !strcmp(cdr->special_action_type, "IVR_INVALID") || 
			  !strcmp(cdr->special_action_type, "PRIVILEGED_CALL") || 
			  !strcmp(cdr->special_action_type, "COUNTRY_CODE/ringboth") || 
			  !strcmp(cdr->special_action_type, "atx_refer") || 
			  !strcmp(cdr->special_action_type, "overwrt_record")))
		  || ((!ast_strlen_zero(cdr->src) && (strstr(cdr->src, "livechat"))) ||
		  	  (!ast_strlen_zero(cdr->dst) && (strstr(cdr->dst, "livechat")))) ) {
			ast_log(LOG_DEBUG, "Detect SPECIAL_ACTION_TYPE:%s, DO NOT CLEAR record icon or src/dst is livechat", cdr->special_action_type);

			if (!strcmp(cdr->special_action_type, "atx_refer")) {
				ast_log(LOG_DEBUG, ", But need do re-check for SPECIAL_ACTION_TYPE:%s", cdr->special_action_type);
				if (!ast_strlen_zero(cdr->dstchannel) && strstr(cdr->dstchannel, "PJSIP/trunk_") && !strstr(cdr->record_files, cdr->dst)) {
					cdr->record_files[0] = '\0';
				}
			}
		} else {
			ast_log(LOG_DEBUG, "clear original var:record_file[%s],special_action_type:%s", cdr->record_files,cdr->special_action_type);
			cdr->record_files[0] = '\0';
		}
	}

END:
	if (call_list) {
		ast_free(call_list);
	}

	return pass;
}

static unsigned int cdr_is_create_next_day(struct ast_cdr *cdr)
{
	struct timeval now;
	struct ast_tm now_tm;
	struct ast_tm cdr_start_tm;

	if (!cdr) {
		return 0;
	}

	now = ast_tvnow();
	ast_localtime(&now, &now_tm, NULL);
	ast_localtime(&cdr->start, &cdr_start_tm, NULL);

	if (cdr_start_tm.tm_hour > now_tm.tm_hour) {
		return 1;
	}

	return 0;
}

static unsigned int today_remain_time(void)
{
	unsigned int remain_time = 0;
	struct timeval now;
	struct ast_tm tm;

	now = ast_tvnow();
	ast_localtime(&now, &tm, NULL);

	remain_time = (23 - tm.tm_hour)*3600 + (59 - tm.tm_min)*60 + (60 - tm.tm_sec);

	ast_log(LOG_DEBUG, "Now time is %d:%d:%d, get remain time:%d\n", tm.tm_hour, tm.tm_min, tm.tm_sec, remain_time);

	return remain_time;
}

enum ast_clear_time_type {
	AST_CLEAR_TIME_UNKNOW,
	AST_CLEAR_TIME_DAY,
	AST_CLEAR_TIME_MONTH,
};

static void set_clear_time(enum ast_clear_time_type type)
{
	struct timeval now;
	struct ast_tm tm;
	char time[16] = { 0 };

	now = ast_tvnow();
	ast_localtime(&now, &tm, NULL);

	snprintf(time, sizeof(time), "%d-%d-%d", tm.tm_year, tm.tm_mon, tm.tm_mday);

	ast_log(LOG_DEBUG, "set_clear_day_time time:%s, type:%d\n", time, type);

	switch (type) {
		case AST_CLEAR_TIME_DAY:
			ast_db_put("TurnInfo", TURN_INFO_CLRARDAYTIME_KEY, time);
			break;
		case AST_CLEAR_TIME_MONTH:
			ast_db_put("TurnInfo", TURN_INFO_CLRARMONTHTIME_KEY, time);
			break;
		default:
			break;
	}

	return;
}

static void clear_limit_time_used(void)
{
	ast_rwlock_wrlock(&time_lock);

	limit_time_used = 0;

	ast_rwlock_unlock(&time_lock);

	ast_db_del("TurnInfo", TURN_INFO_LIMITTIMEUSED_KEY);

	set_clear_time(AST_CLEAR_TIME_DAY);

	ast_log(LOG_DEBUG, "clear_limit_time_used\n");

	return;
}

static unsigned int is_month_first_day(void)
{
	struct timeval now;
	struct ast_tm tm;

	now = ast_tvnow();
	ast_localtime(&now, &tm, NULL);

	if (tm.tm_mday == 1) {
		ast_log(LOG_DEBUG, "Now is the first day of month\n");
		return 1;
	} else {
		return 0;
	}
}

static void clear_month_limit_time_used(void)
{
	ast_rwlock_wrlock(&time_lock);

	limit_month_time_used = 0;

	ast_rwlock_unlock(&time_lock);

	ast_db_del("TurnInfo", TURN_INFO_LIMITMONTHTIMEUSED_KEY);

	set_clear_time(AST_CLEAR_TIME_MONTH);

	ast_log(LOG_DEBUG, "clear_month_limit_time_used\n");

	return;
}

static unsigned int need_clear_day_time_for_reboot(void)
{
	struct timeval now;
	struct ast_tm tm;
	char day_time[16] = { 0 };
	int year = 0;
	int month = 0;
	int day = 0;

	ast_db_get("TurnInfo", TURN_INFO_CLRARDAYTIME_KEY, day_time, sizeof(day_time) - 1);
	if (!strlen(day_time)) {
		ast_db_get("TurnInfo", TURN_INFO_LASTUSEDTIME_KEY, day_time, sizeof(day_time) - 1);
		if (!strlen(day_time)) {
			ast_log(LOG_DEBUG, "Not set clear day time and last used time before, don't need clear\n");
			return 0;
		}
	}

	sscanf(day_time, "%d-%d-%d", &year, &month, &day);

	now = ast_tvnow();
	ast_localtime(&now, &tm, NULL);

	ast_log(LOG_DEBUG, "Previous clear day time:%s, now time:%d-%d-%d\n", day_time, tm.tm_year, tm.tm_mon, tm.tm_mday);

	if (tm.tm_year > year) {
		return 1;
	}

	if (tm.tm_mon != month) {
		return 1;
	}

	if (tm.tm_mday > day) {
		return 1;
	}

	return 0;
}

static unsigned int need_clear_month_time_for_reboot(void)
{
	struct timeval now;
	struct ast_tm tm;
	char month_time[16] = { 0 };
	int year = 0;
	int month = 0;
	int day = 0;

	ast_db_get("TurnInfo", TURN_INFO_CLRARMONTHTIME_KEY, month_time, sizeof(month_time) - 1);
	if (!strlen(month_time)) {
		ast_db_get("TurnInfo", TURN_INFO_LASTUSEDTIME_KEY, month_time, sizeof(month_time) - 1);
		if (!strlen(month_time)) {
			ast_log(LOG_DEBUG, "Not set clear month time and last used time before, don't need clear\n");
			return 0;
		}
	}

	sscanf(month_time, "%d-%d-%d", &year, &month, &day);

	now = ast_tvnow();
	ast_localtime(&now, &tm, NULL);

	ast_log(LOG_DEBUG, "previous clear month time:%s, now time:%d-%d-%d\n", month_time, tm.tm_year, tm.tm_mon, tm.tm_mday);

	if (tm.tm_year > year) {
		return 1;
	}

	if (tm.tm_mon != month) {
		return 1;
	}

	return 0;
}

static void clear_time_for_reboot(void)
{
	if (need_clear_day_time_for_reboot()) {
		clear_limit_time_used();
	}

	if (need_clear_month_time_for_reboot()) {
		clear_month_limit_time_used();
	}

	return;
}

static void *clear_limit_time_run(void *data)
{
	struct timeval timeout;

	ast_log(LOG_DEBUG, "clear_limit_time_run\n");

	while (clear_limit_time_run_flag) {
		timeout.tv_sec = today_remain_time();
		timeout.tv_usec = 0;
		ast_select(0, NULL, NULL, NULL, &timeout);
		clear_limit_time_used();
		if (is_month_first_day()) {
			clear_month_limit_time_used();
		}
	}

	return NULL;
}

static int start_clear_limit_time_thread(void)
{
	if (clear_limit_time_thread != AST_PTHREADT_NULL) {
		ast_log(LOG_DEBUG, "clear_limit_time_thread already start\n");
		return 0;
	}

	if (ast_pthread_create_background(&clear_limit_time_thread, NULL, clear_limit_time_run, NULL)) {
		ast_log(LOG_ERROR, "Failed to create scheduler thread\n");
		return -1;
	}

	return 0;
}

static unsigned int need_set_limit_used_time(struct ast_cdr *cdr)
{
	if (!cdr) {
		return 0;
	}

	if (!has_limit_time) {
		ast_log(LOG_DEBUG, "need_set_limit_used_time no has_limit_time\n");
		return 0;
	}

	if (!cdr->billsec) {
		ast_log(LOG_DEBUG, "need_set_limit_used_time no billsec\n");
		return 0;
	}

#ifndef BUILD_ON_CLOUD_PBX
	if (cdr->is_nat_call != 1 && cdr->peer_is_nat_call != 1) {
		ast_log(LOG_DEBUG, "need_set_limit_used_time no nat call\n");
		return 0;
	}
#endif

	if (cdr_is_create_next_day(cdr)) {
		ast_log(LOG_DEBUG, "need_set_limit_used_time is creat by next day\n");
		return 0;
	}

	return 1;
}

static void set_last_used_time(void)
{
	struct timeval now;
	struct ast_tm tm;
	char time[16] = { 0 };

	now = ast_tvnow();
	ast_localtime(&now, &tm, NULL);

	snprintf(time, sizeof(time), "%d-%d-%d", tm.tm_year, tm.tm_mon, tm.tm_mday);

	ast_log(LOG_DEBUG, "set_last_used_time time:%s\n", time);

	ast_db_put("TurnInfo", TURN_INFO_LASTUSEDTIME_KEY, time);

	return;
}

static void cdr_set_limit_used_time(struct ast_cdr *cdr)
{
	char used[16] = { 0 };
	char month_time_used[16] = { 0 };

	if (!need_set_limit_used_time(cdr)) {
		return;
	}

	ast_rwlock_wrlock(&time_lock);

	if (limit_time_used >= day_total_limit_time*60) {
		ast_log(LOG_DEBUG, "limit time used is already larger than total, don't need set again\n");
		ast_rwlock_unlock(&time_lock);
		return;
	}


#ifndef BUILD_ON_CLOUD_PBX
	if (cdr->is_nat_call == 1) {
		limit_time_used += cdr->billsec;
		if (has_month_limit_time) {
			limit_month_time_used += cdr->billsec;
		}
	}

	if (cdr->peer_is_nat_call == 1) {
		limit_time_used += cdr->billsec;
		if (has_month_limit_time) {
			limit_month_time_used += cdr->billsec;
		}
	}
#else
	limit_time_used += cdr->billsec;
	if (has_month_limit_time) {
		limit_month_time_used += cdr->billsec;
	}
#endif

	ast_log(LOG_DEBUG, "CDR billsec:%ld, limit_time_used change to %d, is_nat_call:%d, peer_is_nat_call:%d\n",
		cdr->billsec, limit_time_used, cdr->is_nat_call, cdr->peer_is_nat_call);

	ast_rwlock_unlock(&time_lock);

	snprintf(used, sizeof(used), "%d", limit_time_used);
	ast_db_put("TurnInfo", TURN_INFO_LIMITTIMEUSED_KEY, used);

	set_last_used_time();

	if (has_month_limit_time) {
		snprintf(month_time_used, sizeof(month_time_used), "%d", limit_month_time_used);
		ast_db_put("TurnInfo", TURN_INFO_LIMITMONTHTIMEUSED_KEY, month_time_used);
		ast_log(LOG_DEBUG, "cdr_set_limit_used_time:%s\n", month_time_used);
	}

	return;
}

unsigned int ast_cdr_get_nat_used_time(void)
{
	unsigned int time = 0;

	ast_rwlock_wrlock(&time_lock);
	time = limit_time_used;
	ast_rwlock_unlock(&time_lock);

	ast_log(LOG_DEBUG, "ast_cdr_get_nat_used_time:%d", time);

	return time;
}

unsigned int ast_cdr_get_nat_month_used_time(void)
{
	unsigned int time = 0;

	ast_rwlock_wrlock(&time_lock);
	time = limit_month_time_used;
	ast_rwlock_unlock(&time_lock);

	ast_log(LOG_DEBUG, "ast_cdr_get_nat_month_used_time:%d", time);

	return time;
}

void ast_cdr_set_has_limit_time(unsigned int value)
{
	ast_log(LOG_DEBUG, "ast_cdr_set_has_limit_time value:%d", value);

	has_limit_time = value;

	if (value) {
		start_clear_limit_time_thread();
	}

	return;
}

void ast_cdr_set_has_month_limit_time(unsigned int value)
{
	ast_log(LOG_DEBUG, "ast_cdr_set_has_month_limit_time value:%d", value);

	has_month_limit_time = value;

	return;
}

void ast_cdr_set_day_total_limit_time(unsigned int value)
{
	ast_log(LOG_DEBUG, "ast_cdr_set_day_total_limit_time value:%d", value);

	day_total_limit_time = value;

	return;
}

static int mark_answered_elsewhere_for_queue(struct cdr_object *cdr)
{
	int skip_mark = 1;
	char tmp_action[256] = {0};
	char cmp_action[256] = {0};
	struct cdr_object *it_cdr;

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		snprintf(cmp_action, sizeof(cmp_action), "%s[%s]", it_cdr->appl, it_cdr->party_a.snapshot ? it_cdr->party_a.snapshot->connected_number : "NULL");
		if ((!ast_strlen_zero(it_cdr->appl) && !strncmp(it_cdr->appl, "Queue", strlen("Queue"))
			&& ((!ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "QUEUE["))
				|| (!ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "TRANSFER") && '\0' != tmp_action[0] && 0 == strcasecmp(cmp_action, tmp_action)))
			&& AST_CDR_ANSWERED == it_cdr->disposition)) {
			skip_mark = 0;
			ast_log(LOG_DEBUG, "Detect Queue member(cdr_obj:%p) answered the Call", it_cdr);
		}
		if (!ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "QUEUE[")){
			snprintf(tmp_action, sizeof(tmp_action), "%s", it_cdr->action_type);
		}
	}

	if (skip_mark) {
		ast_log(LOG_DEBUG, "No Queue member answered the call, skip mark ANSWERED_ELSEWHERE");
		return 0;
	}

	memset(tmp_action, 0, sizeof(tmp_action));
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		snprintf(cmp_action, sizeof(cmp_action), "%s[%s]", it_cdr->appl, it_cdr->party_a.snapshot ? it_cdr->party_a.snapshot->connected_number : "NULL");
		if ((!ast_strlen_zero(it_cdr->appl) && !strncmp(it_cdr->appl, "Queue", strlen("Queue"))
			&& ((!ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "QUEUE["))
				|| (!ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "TRANSFER") && '\0' != tmp_action[0] && 0 == strcasecmp(cmp_action, tmp_action)))
			&& AST_CDR_NOANSWER == it_cdr->disposition)) {
			ast_log(LOG_DEBUG,
				"Mark cdr_obj(%p) disposition 'ANSWERED ELSEWHERE' for Queue Call",
				it_cdr);
			snprintf(it_cdr->reason, sizeof(it_cdr->reason), ANSWERED_ELSEWHERE_NEW);
		}
		if (!ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "QUEUE[")){
			snprintf(tmp_action, sizeof(tmp_action), "%s", it_cdr->action_type);
		}
	}

	return 0;
}

/*
 *  Please be noticed that if this function is called,
 *  It is sure that Calling ringgroup is answered.
 */
static int mark_answered_elsewhere_for_ringgroup(struct cdr_object *cdr)
{
	struct cdr_object *it_cdr;
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (!ast_strlen_zero(it_cdr->appl) && strncmp(it_cdr->appl, "ForkCDR", AST_MAX_EXTENSION)
			&& !ast_strlen_zero(it_cdr->action_type) && strstr(it_cdr->action_type, "RINGGROUP[")
			&& AST_CDR_NOANSWER == it_cdr->disposition) {
			ast_log(LOG_DEBUG,
				"Mark cdr_obj(%p) disposition 'ANSWERED ELSEWHERE' for Ringgroup Call",
				it_cdr);
			snprintf(it_cdr->reason, sizeof(it_cdr->reason), ANSWERED_ELSEWHERE_NEW);
		}
	}

	return 0;
}

#endif

/*!
 * \brief Create a chain of \ref ast_cdr objects from a chain of \ref cdr_object
 * suitable for consumption by the registered CDR backends
 * \param cdr The \ref cdr_object to convert to a public record
 * \retval A chain of \ref ast_cdr objects on success
 * \retval NULL on failure
 */
static struct ast_cdr *cdr_object_create_public_records(struct cdr_object *cdr)
{
	struct ast_cdr *pub_cdr = NULL, *cdr_prev = NULL;
	struct cdr_object *it_cdr;
	struct ast_var_t *it_var, *it_copy_var;
	struct ast_channel_snapshot *party_a;
	struct ast_channel_snapshot *party_b;
#ifdef GRANDSTREAM_NETWORKS
	char tmp_ext[128] = {0};
	char tmp_page_ext[128] = {0};
	char tmp_format[32] = {0};
	int transfer_rp = 0;
	char tmp_caller[128] = {0};
	char callee_cfu[64] = {0};
	char action_owner[AST_MAX_EXTENSION] = {0};
	char transfer_callernum[AST_MAX_EXTENSION] = {0};
	char transfer_callername[512] = {0};
	char atx_inbound_index[AST_MAX_EXTENSION] = {0};
	char sla_chan[256] = {0};
	int clear_atx_rec = 0;
	int callee_followme = 0;
	int announce_disable = 0;
	int reset_paging_start = 0;
	struct cdr_object *pre_it_cdr = NULL;
	int atx_keep_nat_flag = 0;
	int is_app_page=0;
	long int tmp_main_duration = 0;
	long int tmp_main_billsec = 0;
	int first_enable_cdr = 0;
	struct timeval tmp_start =  { 0, };
	struct timeval tmp_answer =  { 0, };
	struct timeval tmp_end =  { 0, };
	struct ast_cdr *tmp_cdr_copy = NULL;

	ast_log(LOG_DEBUG, "Enter cdr_object_create_public_records(%p)", cdr);
#endif

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		struct ast_cdr *cdr_copy;

#ifdef GRANDSTREAM_NETWORKS
		ast_debug(1, "cdr:%p, record:%s, spec_action_type:%s, appl:%s, disposition:%d",
				it_cdr, it_cdr->record_files, 
				ast_strlen_zero(it_cdr->special_action_type)? "":it_cdr->special_action_type,
				ast_strlen_zero(it_cdr->appl) ? "":it_cdr->appl,it_cdr->disposition);
#endif
		/* Don't create records for CDRs where the party A was a dialed channel */
#ifdef GRANDSTREAM_NETWORKS
		if (snapshot_is_dialed(it_cdr->party_a.snapshot) && !it_cdr->party_b.snapshot
		&& !(strstr(it_cdr->party_a.snapshot->name,";1") && !strncmp(it_cdr->action_type, "PAGE", strlen("PAGE")) && strcasestr(it_cdr->session, "SchedulePaging/Intercom") && !strcmp(it_cdr->appl,"ForkCDR")))
#else
		if (snapshot_is_dialed(it_cdr->party_a.snapshot) && !it_cdr->party_b.snapshot)
#endif
		{
			ast_debug(1, "CDR for %s is dialed and has no Party B; discarding\n",
				it_cdr->party_a.snapshot->name);
			continue;
		}

		cdr_copy = ast_calloc(1, sizeof(*cdr_copy));
		if (!cdr_copy) {
			ast_free(pub_cdr);
			return NULL;
		}

		party_a = it_cdr->party_a.snapshot;
		party_b = it_cdr->party_b.snapshot;

		/* Party A */
		ast_assert(party_a != NULL);
/*Start added by zyxia for Bug 206949,2021.10.29*/
#ifdef GRANDSTREAM_NETWORKS
		if( strncmp(party_a->name,"RTSP/",5)  ){
			ast_copy_string(cdr_copy->accountcode, party_a->accountcode, sizeof(cdr_copy->accountcode));
		}
#else
		/*End added by zyxia for Bug 206949,2021.10.29*/
		ast_copy_string(cdr_copy->accountcode, party_a->accountcode, sizeof(cdr_copy->accountcode));
#endif
		cdr_copy->amaflags = party_a->amaflags;
		ast_copy_string(cdr_copy->channel, party_a->name, sizeof(cdr_copy->channel));
		ast_callerid_merge(cdr_copy->clid, sizeof(cdr_copy->clid), party_a->caller_name, party_a->caller_number, "");
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "Initially USE party_a->caller_num:%s AS src", party_a->caller_num);

		ast_copy_string(cdr_copy->src, party_a->caller_num, sizeof(cdr_copy->src));
		ast_copy_string(cdr_copy->new_src, party_a->inbound_caller, sizeof(cdr_copy->new_src));
		ast_copy_string(cdr_copy->inbound_trunk_index, it_cdr->inbound_trunk_index, sizeof(cdr_copy->inbound_trunk_index));
		ast_copy_string(cdr_copy->outbound_trunk_index, it_cdr->outbound_trunk_index, sizeof(cdr_copy->outbound_trunk_index));
		if (ast_strlen_zero(cdr_copy->outbound_trunk_index) && party_b && 0 != party_b->local_outbound_index){
			snprintf(cdr_copy->outbound_trunk_index, sizeof(cdr_copy->outbound_trunk_index), "%d", party_b->local_outbound_index);
		}
		ast_copy_string(cdr_copy->status, it_cdr->status, sizeof(cdr_copy->status));
		cdr_copy->main_disposition = -1;
#else
		ast_copy_string(cdr_copy->src, party_a->caller_number, sizeof(cdr_copy->src));
#endif
		ast_copy_string(cdr_copy->uniqueid, party_a->uniqueid, sizeof(cdr_copy->uniqueid));
		ast_copy_string(cdr_copy->lastapp, it_cdr->appl, sizeof(cdr_copy->lastapp));
		ast_copy_string(cdr_copy->lastdata, it_cdr->data, sizeof(cdr_copy->lastdata));
		ast_copy_string(cdr_copy->dst, it_cdr->exten, sizeof(cdr_copy->dst));
		ast_copy_string(cdr_copy->dcontext, it_cdr->context, sizeof(cdr_copy->dcontext));
		ast_copy_string(cdr_copy->reason, it_cdr->reason, sizeof(cdr_copy->reason));

		/* Party B */
		if (party_b) {
#ifdef GRANDSTREAM_NETWORKS
			ast_copy_string(cdr_copy->special_action_type, it_cdr->special_action_type, sizeof(cdr_copy->special_action_type));
			ast_copy_string(cdr_copy->special_dst, it_cdr->special_dst, sizeof(cdr_copy->special_dst));
			/*add by wxu@grandstream.cn for set callee when callee is null or callee is error*/
			char callee[64] = {0};	// modify for long callee num by wxu@grandstream.cn

			ast_log(LOG_DEBUG,"party_b->name:%s",(char *)(party_b->name));
			if( strstr((char *)(party_b->name),"/") ){
				sscanf((char *)(party_b->name),"%*[^/]/%63[\\+0-9*#]%*s",callee);
			}

			if( !strncmp(party_b->name,"DAHDI/",strlen("DAHDI/")) && strlen(callee) > 0 ){
				char varname_tmp[256] = {0};
				char family[64] = {0};
				char key[64] = {0};

				snprintf(family, sizeof(family), "DAHDI/%s", callee);
				snprintf(key, sizeof(key), "FXS_EXTENSION");
				ast_ucm_db_get(family, key, varname_tmp, sizeof(varname_tmp) - 1);
				snprintf(callee, sizeof(callee), "%s", varname_tmp);
			}

			if (ast_strlen_zero(cdr_copy->dst) && !strncmp(party_b->name, "MulticastRTP", strlen("MulticastRTP")) && !ast_strlen_zero(party_a->connected_number)){
				ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
				snprintf(callee, sizeof(callee), "%s", cdr_copy->dst);
			}

			if ((!strcasestr(cdr_copy->special_action_type, "SCA")
				|| strcasestr(cdr_copy->special_action_type, "AUTO_ANSWER_TO_FAX") //added by zyxia for Bug 257611,2023.4.28
				|| ast_strlen_zero(cdr_copy->dst)
				|| (NULL != strcasestr(it_cdr->action_type, "CALLFORWARD") && NULL != strcasestr(cdr_copy->special_action_type, "SCA")))){
				memset(cdr_copy->dst,0,sizeof(cdr_copy->dst));
				snprintf(cdr_copy->dst,sizeof(cdr_copy->dst)-1,"%s",callee);
							ast_log(LOG_DEBUG,"cdr_copy->dst:%s",cdr_copy->dst);
			}
			ast_log(LOG_DEBUG,"callee:%s",callee);

			if (NULL != strstr(party_b->name, "call-forward") && !ast_strlen_zero(callee)){
				snprintf(callee_cfu,sizeof(callee_cfu)-1,"%s",callee);
			}

			if (!ast_strlen_zero(callee_cfu) && NULL == strstr(party_b->name, "call-forward")){
				ast_copy_string(cdr_copy->dst, callee_cfu, sizeof(cdr_copy->dst));
				memset(callee_cfu,0,sizeof(callee_cfu));
			}

			/*for call forward setted by phone, using local chan to dial with err owner*/
			if (!strcmp(it_cdr->action_type, "CALLFORWARD")
				&& !strstr(cdr_copy->channel, "call-forward")
				&& !strstr(party_b->name, "call-forward")
				&& !ast_strlen_zero(callee)){
				snprintf(action_owner,sizeof(action_owner)-1,"%s",callee);
			}
			/*for followme use local chan to dial with err dst*/
			if (!strncmp(it_cdr->action_type, "FOLLOWME", strlen("FOLLOWME"))
				&& strstr(party_b->name, "Local")
				&& strstr(party_b->name, ";1")
				&& !ast_strlen_zero(callee) && strcmp(cdr_copy->src, callee)){
				ast_log(LOG_NOTICE, "the callee of followme should be retention");
				callee_followme = 1;
			}
			if (0 == strcmp(it_cdr->action_type, "TRANSFER")
				&& !ast_strlen_zero(it_cdr->action_owner)
				&& it_cdr->next && it_cdr->next->party_b.snapshot)
			{
				const char *next_src_chan=it_cdr->next->party_a.snapshot->name;
				const char *next_dst_chan = it_cdr->next->party_b.snapshot->name;
				if (!ast_strlen_zero(next_src_chan) && 0 == strcmp(cdr_copy->channel, next_src_chan)
					&&(!ast_strlen_zero(next_dst_chan) && !ast_strlen_zero(party_b->name)))
				{
					int len_now=strlen(party_b->name);
					int len_next=strlen(next_dst_chan);
					if (len_now==len_next && len_next > strlen("00000000") && 0 == strncmp(party_b->name, next_dst_chan, len_next - strlen("00000000"))){
						snprintf(action_owner,sizeof(action_owner)-1,"%s",it_cdr->action_owner);
					}
				}
			}
			/*end*/
#endif
			ast_copy_string(cdr_copy->dstchannel, party_b->name, sizeof(cdr_copy->dstchannel));
			ast_copy_string(cdr_copy->peeraccount, party_b->accountcode, sizeof(cdr_copy->peeraccount));
			if (!ast_strlen_zero(it_cdr->party_b.userfield)) {
				snprintf(cdr_copy->userfield, sizeof(cdr_copy->userfield), "%s;%s", it_cdr->party_a.userfield, it_cdr->party_b.userfield);
			}
		}
		if (ast_strlen_zero(cdr_copy->userfield) && !ast_strlen_zero(it_cdr->party_a.userfield)) {
			ast_copy_string(cdr_copy->userfield, it_cdr->party_a.userfield, sizeof(cdr_copy->userfield));
		}

		/* Timestamps/durations */
		cdr_copy->start = it_cdr->start;
#ifdef GRANDSTREAM_NETWORKS
		if (!reset_paging_start && !strncmp(it_cdr->action_type,"PAGE",strlen("PAGE")) && party_b){
			struct cdr_object *tmp_it_cdr = it_cdr;
			struct timeval tv_start = it_cdr->start;
			for(tmp_it_cdr=tmp_it_cdr->next;tmp_it_cdr;tmp_it_cdr=tmp_it_cdr->next){
				if (!strncmp(it_cdr->action_type,"PAGE",strlen("PAGE"))){
					tmp_it_cdr->start = tv_start;
				}
			}
			reset_paging_start = 1;
		}
#endif
		cdr_copy->answer = it_cdr->answer;
		cdr_copy->end = it_cdr->end;
		cdr_copy->billsec = cdr_object_get_billsec(it_cdr);
		cdr_copy->duration = cdr_object_get_duration(it_cdr);

		/* Flags and IDs */
		ast_copy_flags(cdr_copy, &it_cdr->flags, AST_FLAGS_ALL);
		ast_copy_string(cdr_copy->linkedid, it_cdr->linkedid, sizeof(cdr_copy->linkedid));
		cdr_copy->disposition = it_cdr->disposition;
		cdr_copy->sequence = it_cdr->sequence;

#ifdef GRANDSTREAM_NETWORKS
		if (ast_strlen_zero(cdr_copy->dst) && !strcmp(it_cdr->action_type, "VM") && !ast_strlen_zero(party_a->connected_number)){
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}
		if (!ast_strlen_zero(it_cdr->special_action_type) && (
			(0 == strncmp(it_cdr->special_action_type, "DO_DND", strlen("DO_DND")) && ast_strlen_zero(cdr_copy->dstchannel))
			|| (!strcmp(it_cdr->special_action_type, "callpark"))
			)){
			ast_copy_string(cdr_copy->special_action_type, it_cdr->special_action_type, sizeof(cdr_copy->special_action_type));
		}
		/*add by wxu@grandstream.cn for set callee adnd caller*/
		snprintf(cdr_copy->record_files,sizeof(cdr_copy->record_files)-1,"%s",it_cdr->record_files);
		if (ast_strlen_zero(cdr_copy->dst) && (NULL != party_a->connected_number && !ast_strlen_zero(party_a->connected_number)) && strcasestr(cdr_copy->dstchannel, "trunk_")){
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}

		if( !ast_strlen_zero(it_cdr->app_caller) ){
			memset(cdr_copy->src,0,sizeof(cdr_copy->src));
			snprintf(cdr_copy->src,sizeof(cdr_copy->src)-1,"%s",it_cdr->app_caller);

			ast_log(LOG_DEBUG,"Use app_caller:%s as cdr'src", it_cdr->app_caller);
		}

		if( !ast_strlen_zero(it_cdr->app_callee) && !strcasestr(cdr_copy->special_action_type, "AUTO_ANSWER_TO_FAX")//added by zyxia for Bug 257611,2023.4.28
			&& (!callee_followme || (!strcmp(it_cdr->action_type, "VM") && !strcmp(it_cdr->special_action_type, "FOLLOWME")))){
			char *tmp_sched = NULL;

			memset(cdr_copy->dst,0,sizeof(cdr_copy->dst));
			snprintf(cdr_copy->dst,sizeof(cdr_copy->dst)-1,"%s",it_cdr->app_callee);

			tmp_sched = strstr(cdr_copy->dst, "-schedule-");
			if (tmp_sched) {
				ast_log(LOG_DEBUG,"Detect redundant schedule-conference-number, strip it!");
				*tmp_sched = '\0';
			}

			ast_log(LOG_DEBUG,"Use app_callee:%s as cdr'dst", it_cdr->app_callee);
		}

		callee_followme = 0;

		if (!ast_strlen_zero(cdr_copy->dstchannel) && cdr == it_cdr && NULL == it_cdr->next && 0 == strcmp(cdr_copy->lastapp, "SLATrunk")){
			memset(cdr_copy->dst,0,sizeof(cdr_copy->dst));
			char tmp_num[128] = {0};
			sscanf(cdr_copy->dstchannel, "%*[^/]/%127[^-]-%*s", tmp_num);
			if (!ast_strlen_zero(tmp_num)){
				ast_copy_string(cdr_copy->dst, tmp_num, sizeof(cdr_copy->dst));
			}
		}

		if (2!= transfer_rp && 0 == strcmp(it_cdr->action_type, "TRANSFER")
			&& (NULL != strstr(cdr_copy->channel, "Local") && NULL != strstr(cdr_copy->channel, "from-internal-xfer") && NULL != strstr(cdr_copy->channel, ";2"))){
			transfer_rp = 1;
		}

		if (ast_strlen_zero(cdr_copy->src) && !ast_strlen_zero(party_a->caller_number)
			&& strstr(cdr_copy->channel, "asterisk_guitools")
			&& cdr_copy->dstchannel && strstr(cdr_copy->dstchannel, "call-forward")){
			ast_copy_string(cdr_copy->src, party_a->caller_number, sizeof(cdr_copy->src));
		}

		if (0 == strcmp(it_cdr->special_action_type, "COUNTRY_CODE_LIMIT")){
			cdr_copy->disposition = AST_CDR_FAILED;
			ast_copy_string(cdr_copy->special_action_type, it_cdr->special_action_type, sizeof(cdr_copy->special_action_type));
		}

		if( ast_strlen_zero(cdr_copy->src)
			&& strcasestr(it_cdr->session, "SchedulePaging/Intercom")
			&& strcasestr(it_cdr->name, "Local") && strcasestr(it_cdr->name, ";1")){
			char src[64] = {0};
			const char *format = "%*[^/]/%[^@]";
			sscanf(it_cdr->name,format,src);
			if ('\0' != src[0]){
				ast_copy_string(cdr_copy->src, src, sizeof(cdr_copy->src));
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"SchedulePaging/Intercom\" <%s>", cdr_copy->src);
				ast_copy_string(cdr_copy->action_owner, cdr_copy->src, sizeof(cdr_copy->action_owner));
			}
		}
		if (ast_strlen_zero(cdr_copy->src) && !ast_strlen_zero(party_a->caller_number) && 0 == strcmp(it_cdr->action_type, "PARKEDCALL")){
			ast_copy_string(cdr_copy->src, party_a->caller_number, sizeof(cdr_copy->src));
		}

		if (!ast_strlen_zero(it_cdr->special_action_type) && 0 == strcmp(it_cdr->special_action_type, "N_WAY_CONF") && ast_strlen_zero(cdr_copy->src) && NULL != strstr(cdr_copy->channel, "trunk_") && !ast_strlen_zero(party_a->caller_number)){
			ast_copy_string(cdr_copy->src, party_a->caller_number, sizeof(cdr_copy->src));
		}
		/*if sfu invites the extension(not registered in wave), should swap the caller and callee*/
		if (NULL != strstr(cdr_copy->lastdata, "sfu_user_menu") && NULL != strstr(cdr_copy->dcontext, "conf-inv-menu") && !ast_strlen_zero(it_cdr->special_action_type) && 0 == strcmp(it_cdr->special_action_type, "N_WAY_CONF")){
			char tmp_save_dst[128] = {0};
			snprintf(tmp_save_dst, sizeof(tmp_save_dst),"%s", cdr_copy->dst);
			ast_copy_string(cdr_copy->dst, cdr_copy->src, sizeof(cdr_copy->dst));
			ast_copy_string(cdr_copy->src, tmp_save_dst, sizeof(cdr_copy->src));
			ast_copy_string(cdr_copy->action_owner, tmp_save_dst, sizeof(cdr_copy->action_owner));
			snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"\" <%s>", cdr_copy->src);
		}
		if (NULL != strstr(cdr_copy->src, "webrtc_trunk") && 0 == strcmp(cdr_copy->src, party_a ? party_a->caller_name : "NULL")) {
			snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"\" <%s>", cdr_copy->src);
		}
		if( ast_strlen_zero(cdr_copy->src) ){
			char src[64] = {0};
			sscanf(it_cdr->session,"%*[^-]-%63[\\+0-9a-zA-Z_*#-]%*s",src);
			snprintf(cdr_copy->src,sizeof(cdr_copy->src)-1,"%s", src);

			ast_log(LOG_DEBUG, "cdr_copy->src is NULL string, So scan it from session:%s",
				it_cdr->session);
			if (ast_strlen_zero(cdr_copy->src) && ast_strlen_zero(it_cdr->session) && !ast_strlen_zero(cdr_copy->clid))
			{
				ast_log(LOG_DEBUG, "session is null, see if we can get src from clid:%s", cdr_copy->clid);
				sscanf(cdr_copy->clid, "%*[^<]<%63[\\+0-9a-zA-Z_*#-]%*s", src);
				snprintf(cdr_copy->src,sizeof(cdr_copy->src)-1,"%s", src);
			}
			if(NULL != strstr(cdr_copy->channel, "Local") && NULL != strstr(cdr_copy->channel, ";2")
				&& 0 == strcmp(it_cdr->action_type, "ANNOUNCEMENT PAGE") && 0 == strcmp(cdr_copy->src, "Page")
				&& party_a->connected_number && strcasestr(party_a->connected_number, "announcement_paging")){
				ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
			}
		}

		if( (!strcmp(cdr_copy->dst,"1") && !strstr(it_cdr->action_type, "EMERGENCYCALL") && !strstr(it_cdr->action_type, "PARKEDCALL"))  
		 || (!strcmp(cdr_copy->dst,"o")))	{
			memset(cdr_copy->dst,0,sizeof(cdr_copy->dst));
		}
		/*end*/

		if (strcasestr(it_cdr->special_action_type, "DISA_CALLEE_BLT") && 0 == strncmp(it_cdr->action_type, "TRANSFER", strlen("TRANSFER"))){
			/*to clear disable flag*/
			ast_clear_flag(cdr_copy, AST_CDR_FLAG_DISABLE|AST_CDR_FLAG_DISABLE_ALL);
			ast_log(LOG_VERBOSE, "clear cdr disable flag");
		}

		if(!strncmp(cdr_copy->dcontext, "outbound-route-", strlen("outbound-route-"))
			&& !ast_strlen_zero(cdr_copy->outbound_trunk_index)
			&& (strstr(cdr_copy->lastdata, "DAHDI/g") || strstr(cdr_copy->lastdata, "trunk_"))){
			if (ast_strlen_zero(cdr_copy->dst) && !ast_strlen_zero(party_a->connected_number)){
				ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
			}
			ast_clear_flag(cdr_copy, AST_CDR_FLAG_DISABLE|AST_CDR_FLAG_DISABLE_ALL);
			ast_log(LOG_VERBOSE, "clear cdr disable flag");
		}

		/*add by wxu@grandstream.cn for disable Local channel ;2 CDR in Dial*/
		if(!strncmp(cdr_copy->dstchannel,"Local/",strlen("Local/")) && strstr(cdr_copy->dstchannel,";2")
			&&(!strncmp(cdr_copy->channel,"PJSIP/",strlen("PJSIP/")) || (!strncmp(cdr_copy->channel,"DAHDI/",strlen("DAHDI/"))))//modify by wxu@grandstream.cn for DAHDI trunk
			&& !strcmp(cdr_copy->lastapp,"Dial")){	/*Local in Dial function*/
			ast_log(LOG_VERBOSE,"cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		/*end*/
		/*added by wxu@grandstream.cn for disable Local channel ;2 to PJSIP dstchannel for same extension*/
		if( !strncmp(cdr_copy->channel,"Local/",strlen("Local/")) && strstr(cdr_copy->channel,";2")
			&& !strcmp(cdr_copy->lastapp,"Dial")){	//Local in Dial function
			/*channel and dstchannel extension is same*/
			char p_channel_num[64] = {0};
			char p_dstchannel_num[64] = {0};

			sscanf(cdr_copy->channel,"%*[^/]/%63[\\+0-9]%*s",p_channel_num);

			if( strncmp(cdr_copy->dstchannel,"DAHDI/",strlen("DAHDI/")) ) {
				sscanf(cdr_copy->dstchannel,"%*[^/]/%63[0-9]%*s",p_dstchannel_num);
			}
			if( (strlen(p_channel_num) > 0
				&& (!strcmp(p_channel_num,p_dstchannel_num) || !strcmp(p_channel_num,cdr_copy->dst))
				&& !strcasestr(it_cdr->action_type,"CONFERENCE")
				&& !strcasestr(it_cdr->action_type,"MEETING")
				&& !strcasestr(cdr_copy->clid,"Call Bargin")
				&& !strcasestr(cdr_copy->clid,"Call Whisper")
				&& !strcasestr(cdr_copy->clid,"Call Listen")
				&& !strcasestr(it_cdr->action_type,"FAX")
				&& !strcasestr(it_cdr->action_type,"WAKEUP")
				&& !strcasestr(it_cdr->action_type,"SCHEDULE CALL")
				&& !strcasestr(it_cdr->action_type,"Callback")
				&& !strcasestr(it_cdr->action_type,"TRANSFER")
				&& !strcasestr(it_cdr->special_action_type, "VQ_CALLBACK")
				&& !strcasestr(it_cdr->special_action_type, "SEAMLESS_TRANSFER")
				&& !strcasestr(it_cdr->action_type,"EMERGENCYNOTIFY")  /* add by jlhu for trunk hunt */
				&& !strcasestr(it_cdr->action_type,"ANNOUNCEMENT PAGE")
				&& !strcasestr(it_cdr->special_action_type,"local_disa")
				&& !strcasestr(it_cdr->special_action_type,"announcement_disa")
				&& !strcasestr(cdr_copy->dstchannel, "MulticastRTP")
				&& !(strcasestr(it_cdr->special_action_type,"ringboth") && !(strstr(cdr_copy->dstchannel, "Local/") && strstr(cdr_copy->dstchannel, ";1")))
				&& !strcasestr(it_cdr->special_action_type,"DISA_CALLEE_BLT"))
			|| strcasestr(cdr_copy->clid,"Record Prompt")
			|| (strcasestr(cdr_copy->clid,"Multi-conference Bridge") && ast_strlen_zero(cdr_copy->dst))){
				if (strstr(cdr_copy->channel, "Local/") && strstr(cdr_copy->channel, ";2")
					&& (strstr(cdr_copy->channel, "from-internal") || (strstr(cdr_copy->channel, "call-forward") && strstr(it_cdr->special_action_type, "IS_ATX")))
					&& strstr(cdr_copy->dstchannel, "trunk") && !strcmp(p_channel_num,cdr_copy->dst)
					&& cdr_copy->disposition == AST_CDR_ANSWERED && 0 != strcmp(it_cdr->special_action_type, "COUNTRY_CODE/FOLLOWME")
					&& !ast_strlen_zero(cdr_copy->record_files)) {
					ast_log(LOG_VERBOSE, "In this case dont disable CDR");
					ast_clear_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
				} else {
					if (cdr_copy->disposition == AST_CDR_ANSWERED
						&& !strcmp(it_cdr->special_action_type, "TRIGGER_CALL_FLIP_TRANSFER")) {
						ast_log(LOG_VERBOSE, "In this case dont disable CDR");
						ast_clear_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
					} else {
						ast_log(LOG_VERBOSE,"------ set cdr disable ------");
						ast_log(LOG_VERBOSE,"p_channel_num:%s, p_dstchannel_num:%s", p_channel_num, p_dstchannel_num);
						ast_log(LOG_VERBOSE,"clid:%s", cdr_copy->clid);
						ast_log(LOG_VERBOSE,"dst:%s", cdr_copy->dst);
						ast_log(LOG_VERBOSE,"action_type:%s", it_cdr->action_type);
						ast_log(LOG_VERBOSE,"special_action_type:%s", it_cdr->special_action_type);
						ast_log(LOG_VERBOSE,"dstchannel:%s", cdr_copy->dstchannel);
						ast_log(LOG_VERBOSE,"-----------------------------");
						ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
					}
				}
			}
		}
		/*end*/

		/*added by wxu@grandstream.c for callee #72 Park caller*/
		/*if(!strcmp(cdr_copy->lastapp,"Park") && strlen(cdr_copy->dstchannel) == 0){
			ast_log(LOG_VERBOSE,"cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}*/
		/*end*/

		if( !strcmp(cdr_copy->src,"Call Bargin") && strlen(cdr_copy->dstchannel) == 0
			&& !strncmp(cdr_copy->channel,"Local/",strlen("Local/")) && strstr(cdr_copy->channel,";1")){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/* bug 105270 */
		if((strstr(cdr_copy->channel,"mcb-create") && !strncmp(cdr_copy->dstchannel,"Local/",strlen("Local/")) && !strstr(cdr_copy->channel,";2"))
		|| (!strncmp(cdr_copy->channel,"Local/",strlen("Local/")) && !strstr(cdr_copy->channel, ";2") && !strncmp(cdr_copy->dstchannel,"Local/",strlen("Local/")) && strcasestr(it_cdr->action_type,"CONFERENCE"))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if( !strncmp(cdr_copy->channel,"Local",strlen("Local")) && strlen(cdr_copy->dstchannel) == 0 && 0 != strcmp(cdr_copy->dst, "send_fax_hangup")){
			/* bug 59309, 49292, 70380, 81392*/
			//if(!(strcasestr(cdr_copy->lastapp,"ConfBridge") && strcasestr(cdr_copy->channel,"from-internal"))){
			if(!((strcasestr(it_cdr->special_action_type, "SEAMLESS_TRANSFER") || strcasestr(it_cdr->special_action_type, "INBOUND_FAX")) && strcasestr(cdr_copy->channel,"from-internal") && strstr(cdr_copy->channel,";1"))
				&& !(strcasestr(it_cdr->special_action_type, "SCA") && strstr(cdr_copy->channel,";2"))
				&& !(strcasestr(it_cdr->action_type, "IVR") && strcasestr(cdr_copy->src,"Callback"))
				&& !(strcasestr(it_cdr->special_action_type, "VQ_CALLBACK") && strstr(cdr_copy->channel,";2"))
				&& !(0 == strncmp(it_cdr->action_type, "ANNOUNCEMENT PAGE", strlen("ANNOUNCEMENT PAGE")) && 0 == strncmp(it_cdr->appl, "Dial", strlen("Dial")))
				&& !(strcasestr(it_cdr->action_type, "VM") && strstr(cdr_copy->channel,";2"))
				&& !(strstr(cdr_copy->channel, ";2") && strcasestr(it_cdr->action_type, "PAGE"))
				&& !(strstr(cdr_copy->channel,";1") && !strncmp(it_cdr->action_type, "PAGE", strlen("PAGE")) && strcasestr(it_cdr->session, "SchedulePaging/Intercom") && !strcmp(it_cdr->appl,"ForkCDR"))
				&& !(strcasestr(cdr_copy->lastapp,"ConfBridge") && strcasestr(cdr_copy->channel,"from-internal") && strcasestr(cdr_copy->channel,";2"))
				&& !strcasestr(it_cdr->special_action_type, "DISA_CALLEE_BLT")
				&& !(0 == strcmp(it_cdr->action_type, "SCHEDULE CALL") && 0 == strcmp(it_cdr->action_type, "WAKEUP") && 0 == strcmp(it_cdr->special_action_type, "FOLLOWME") && strstr(cdr_copy->channel,";2"))) {
				ast_log(LOG_VERBOSE,"------ set cdr disable ------");
				ast_log(LOG_VERBOSE, "special_action_type:%s", it_cdr->special_action_type);
				ast_log(LOG_VERBOSE, "channel:%s", cdr_copy->channel);
				ast_log(LOG_VERBOSE, "action_type:%s", it_cdr->action_type);
				ast_log(LOG_VERBOSE, "lastapp:%s", cdr_copy->lastapp);
				ast_log(LOG_VERBOSE,"-----------------------------");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}
		if(0 == strcmp(cdr_copy->src, "api") && !ast_strlen_zero(party_a->connected_number)
			&& strstr(cdr_copy->channel,";1") && strstr(cdr_copy->channel,"Local")
			&& !strncmp(cdr_copy->lastapp, "ForkCDR", strlen("ForkCDR"))){
			ast_log(LOG_VERBOSE,"clean the disable flag.");
			ast_clear_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}

		if (AST_CDR_NULL == cdr_copy->disposition && 0 == strncmp(it_cdr->action_type, "ANNOUNCEMENT PAGE", strlen("ANNOUNCEMENT PAGE"))){
			cdr_copy->disposition = AST_CDR_NOANSWER;
		}

		/*for bug 115292 by ychpang*/
		if (0 == strcmp(cdr_copy->dst, "send_fax_hangup") && !strncmp(cdr_copy->channel,"Local",strlen("Local")) && strstr(cdr_copy->channel, ";2")){
			snprintf(tmp_format, sizeof(tmp_format), "%%*[^/]/%%%lu[^@]", sizeof(tmp_ext)-1);
			sscanf(cdr_copy->channel, tmp_format, tmp_ext);
			if (0 != strlen(tmp_ext)){
				ast_copy_string(cdr_copy->dst, tmp_ext, sizeof(cdr_copy->dst));
			}
			memset(tmp_ext, 0, sizeof(tmp_ext));
		}

		if( strlen(cdr_copy->dst) == 0 && strlen(cdr_copy->special_dst) != 0 && strncmp(cdr_copy->dstchannel,"Local",strlen("Local")) ){
			snprintf(cdr_copy->dst,sizeof(cdr_copy->dst)-1,"%s",cdr_copy->special_dst);
		}

		if (0 == strncmp(it_cdr->special_action_type, "ATX", strlen("ATX")) && ast_strlen_zero(cdr_copy->dst) && !ast_strlen_zero(party_a->connected_number))
		{
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}

		if( strlen(cdr_copy->dst) == 0 && (strlen(cdr_copy->dstchannel) == 0 || strncmp(cdr_copy->dstchannel, "Local/h@from-internal", strlen("Local/h@from-internal")) == 0) ){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		//add for cdr display error,callee is i, s or t; bug 78610
		if( (strstr(it_cdr->action_type, "IVR") || strcasestr(it_cdr->action_type, "VFAX")) && (strcmp(cdr_copy->dst,"s") == 0 || strcmp(cdr_copy->dst,"i") == 0 || strcmp(cdr_copy->dst,"t") == 0 || strcmp(cdr_copy->dst,"h") == 0)){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*bug 165234*/
		if (!strcmp(cdr_copy->dst,"s") && !strcmp(cdr_copy->lastapp, "Park")){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if (NULL != strstr(it_cdr->special_action_type, "ATX") && 0 == strcmp(cdr_copy->dst,"h")){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if ( strstr(it_cdr->action_type, "EMERGENCYCALL") && ( 0 == strcmp(cdr_copy->dst,"nopriority") || 0 == strcmp(cdr_copy->dst,"normalrel")))
		{
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*bug 97692  bug 116623*/
		if( !strncmp(cdr_copy->lastapp, "AppQueue", AST_MAX_EXTENSION)
			&& (strlen(cdr_copy->dstchannel) == 0 || (!strncmp(cdr_copy->dstchannel,"Local/",strlen("Local/")) && strstr(cdr_copy->dstchannel, ";1")))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		ast_copy_string(cdr_copy->session, it_cdr->session, sizeof(cdr_copy->session));
		if( it_cdr->action_owner == NULL || it_cdr->action_owner[0] == '\0' ){
			ast_copy_string(cdr_copy->action_owner, cdr_copy->src, sizeof(cdr_copy->action_owner));
		} else{
			ast_copy_string(cdr_copy->action_owner, it_cdr->action_owner, sizeof(cdr_copy->action_owner));
		}

		if(!ast_strlen_zero(action_owner) && party_b && strstr(party_b->name, "call-forward")){
			ast_copy_string(cdr_copy->action_owner, action_owner, sizeof(cdr_copy->action_owner));
			memset(action_owner,0,sizeof(action_owner));
		}

		if (!ast_strlen_zero(action_owner) && !ast_strlen_zero(cdr_copy->dstchannel) && 0 == strcmp(it_cdr->action_type, "TRANSFER"))
		{
			ast_copy_string(cdr_copy->action_owner, action_owner, sizeof(cdr_copy->action_owner));
		}

		if( it_cdr->action_type == NULL || it_cdr->action_type[0] == '\0'  ){
			ast_copy_string(cdr_copy->action_type,"DIAL", sizeof(cdr_copy->action_type));
		} else{
			ast_copy_string(cdr_copy->action_type,it_cdr->action_type, sizeof(cdr_copy->action_type));
		}

		if (0 == strcmp(cdr_copy->action_type, "PARKEDCALL") && 0 == strcmp(cdr_copy->special_action_type, "SCA")){
			ast_clear_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			ast_log(LOG_VERBOSE, "clear cdr disable flag");
		}

		if (0 == strncmp(it_cdr->special_action_type, "ATX[", strlen("ATX[")) && ast_strlen_zero(cdr_copy->dstchannel)){
			sscanf(it_cdr->special_action_type, "%*[^[][%[^]]", cdr_copy->action_owner);
		}
		if (0 == strncmp(it_cdr->special_action_type, "ATX", strlen("ATX")) && 0 == strcmp(cdr_copy->lastapp, "VoiceMail"))
		{
			ast_copy_string(cdr_copy->action_type, "VM", sizeof(cdr_copy->action_type));
		}

		if (0 == strncmp(it_cdr->special_action_type, "COUNTRY_CODE/FOLLOWME", strlen("COUNTRY_CODE/FOLLOWME"))){
			ast_copy_string(cdr_copy->action_type, it_cdr->special_action_type+strlen("COUNTRY_CODE/"), sizeof(cdr_copy->action_type));
			if (strstr(cdr_copy->channel, "Local/") && strstr(cdr_copy->channel, ";2")
				&& (strstr(cdr_copy->dstchannel, "trunk") || strstr(cdr_copy->dstchannel, "DAHDI"))
				&& !ast_strlen_zero(cdr_copy->outbound_trunk_index)
				&& 0 != strcmp(it_cdr->special_action_type, "COUNTRY_CODE/FOLLOWME")) {
				ast_log(LOG_VERBOSE, "In this case dont disable CDR");
				ast_clear_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		if (0 == strncmp(it_cdr->special_action_type, "FOLLOWME/external", strlen("FOLLOWME/external"))
			&& strcasestr(cdr_copy->dstchannel, "Local") && strcasestr(cdr_copy->dstchannel, ";1")
			&& !ast_strlen_zero(cdr_copy->outbound_trunk_index)){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if (NULL != strcasestr(cdr_copy->dstchannel, "call-forward") && 0 == strncmp(it_cdr->action_type, "RINGGROUP", strlen("RINGGROUP")))
		{
			ast_copy_string(cdr_copy->action_type,"CALLFORWARD", sizeof(cdr_copy->action_type));
			if(strlen(cdr_copy->record_files) > 0) {
				ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
				memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
			}
		}

		if (strstr(cdr_copy->action_type, "FAILEDIVR")){
			snprintf(tmp_ext, sizeof(tmp_ext)-1, "%s", cdr_copy->action_type+6);
			memset(cdr_copy->action_type,0,sizeof(cdr_copy->action_type));
			ast_copy_string(cdr_copy->action_type, tmp_ext, sizeof(cdr_copy->action_type));
			cdr_copy->disposition = AST_CDR_FAILED;
			cdr_copy->billsec = 0;
			memset(tmp_ext, 0, sizeof(tmp_ext));
		}

		if( !strcmp(cdr_copy->src,"Callback")) {
			char src[64] = {0};
			sscanf(it_cdr->session,"%*[^-]-%63[\\+0-9][a-z][A-Z]%*s",src);
			/*modify for bug 109482, callback have prepend cause cdr disable*/
			int diff = 0;
			if (strlen(cdr_copy->dst) > strlen(src))
			{
				diff = strlen(cdr_copy->dst) - strlen(src);
			}
			if ((!strcasecmp(cdr_copy->action_type,"Dial")&& !strncmp(cdr_copy->dstchannel,"Local/",strlen("Local/")) && strstr(cdr_copy->dstchannel,";1"))
			|| (strlen(cdr_copy->dstchannel) == 0 && !strcasecmp(cdr_copy->lastapp, "Hangup"))
			|| (!strncmp(cdr_copy->channel,"Local/",strlen("Local/")) && strstr(cdr_copy->channel,";2") && strcmp(cdr_copy->dst + diff, src))){

				ast_log(LOG_VERBOSE,"set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		/*bug 117293*/
		if (strlen(cdr_copy->dstchannel) == 0 && !strcasecmp(cdr_copy->lastapp, "Hangup") && !strcmp(cdr_copy->action_type, "CALLBACK")){
				ast_log(LOG_VERBOSE,"set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if( strstr(cdr_copy->dstchannel, "call-forward") && strstr(cdr_copy->action_type, "DIAL")){
			ast_copy_string(cdr_copy->action_type,"CALLFORWARD", sizeof(cdr_copy->action_type));
			ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
			memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
			cdr_copy->billsec = 0;
		}

		if(0 != strcmp(it_cdr->status, "0") && strlen(it_cdr->status) > 0 && strcasecmp(cdr_copy->action_type, "VM") && !strstr(cdr_copy->action_type, "ANNOUNCEMENTS")){
			cdr_copy->disposition = dial_status_to_disposition(it_cdr->status);
		}

		snprintf(tmp_format, sizeof(tmp_format), "IVR[%%%lu[0-9]]", sizeof(tmp_ext)-1);
		sscanf(cdr_copy->action_type, tmp_format, tmp_ext);
		//if( strstr(cdr_copy->action_type, "IVR") && strstr(cdr_copy->action_type, cdr_copy->dst)) {
		if( strstr(cdr_copy->action_type, "IVR") && strlen(tmp_ext) != 0 && strcmp(tmp_ext, cdr_copy->dst) == 0) {
			cdr_copy->disposition = AST_CDR_ANSWERED;
		}

		/* add for bug 60257 - 60259, 62614, add by pchwu */
		if (cdr_copy->disposition == AST_CDR_NOANSWER){
			cdr_copy->answer = cdr_copy->end;
			cdr_copy->billsec = 0;
			/*sometimes, action type will by like CALLFORWARD/QUEUE*/
			if (NULL != strstr(cdr_copy->action_type,"QUEUE")
				&& !strncmp(cdr_copy->lastapp, "ForkCDR", AST_MAX_EXTENSION)) {
				ast_log(LOG_DEBUG, "Mark disposition 'ANSWERED' for Queue Call");
				cdr_copy->disposition = AST_CDR_ANSWERED;
				mark_answered_elsewhere_for_queue(cdr);
			}
		}

		if ((cdr_copy->disposition == AST_CDR_FAILED && strncmp(it_cdr->special_action_type, "EMPTYQUEUE", strlen("EMPTYQUEUE")))
			&&(!strncmp(cdr_copy->action_type,"QUEUE",strlen("QUEUE")) && !strncmp(cdr_copy->lastapp, "ForkCDR", AST_MAX_EXTENSION))){
			cdr_copy->disposition = AST_CDR_ANSWERED;
		}

		/*add by ychpang to discriminate whether ringgroup member is answered*/
		if (NULL != strstr(cdr_copy->action_type,"RINGGROUP")
			&& !strncmp(cdr_copy->lastapp, "ForkCDR", strlen("ForkCDR"))) {
			struct cdr_object *tmp_cdr = NULL;
			for(tmp_cdr = cdr; tmp_cdr; tmp_cdr=tmp_cdr->next){
				if (AST_CDR_ANSWERED == tmp_cdr->disposition && NULL != strstr(tmp_cdr->action_type,"RINGGROUP")){
					cdr_copy->disposition = AST_CDR_ANSWERED;
					mark_answered_elsewhere_for_ringgroup(cdr);
					break;
				}
			}
			tmp_cdr = NULL;
		}
		/* end */

		if (2 == announce_disable && ((party_b && ast_strlen_zero(party_b->name)) || !party_b)){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			announce_disable = 0;
		}
		if (strstr(cdr_copy->action_type, "ANNOUNCE[")){
			struct cdr_object *tmp_cdr = NULL;
			for(tmp_cdr = it_cdr->next; tmp_cdr; tmp_cdr=tmp_cdr->next){
				if (1 == announce_disable && !strcmp(tmp_cdr->action_type, "VM")){
					announce_disable = 2;
					break;
				}
				if (0 == announce_disable
					&& !strcmp(tmp_cdr->action_type, "DIAL")
					&& !strcmp(tmp_cdr->appl, "Dial")
					&& ((tmp_cdr->party_b.snapshot && ast_strlen_zero(tmp_cdr->party_b.snapshot->name)) || !tmp_cdr->party_b.snapshot)){
					announce_disable = 1;
				}
			}
			tmp_cdr = NULL;
		}

		if( strstr(cdr_copy->action_type,"DISA") || 0 == strncmp(cdr_copy->action_type,"PAGE",strlen("PAGE")) || strstr(cdr_copy->lastapp, "Directory")){
			cdr_copy->disposition = AST_CDR_ANSWERED;
		}

		/* add by yjwu for BUG 92132 */
		if(strstr(cdr_copy->lastapp, "Authenticate")){
			cdr_copy->disposition = AST_CDR_FAILED;
		}

		/*add by wxu@grandstream.cn for remove record_files when call is not ANSWERED*/
		/* modify by yjwu, for remove the record_files when callee is not a extension like: DISA) */
		if( cdr_copy->disposition != AST_CDR_ANSWERED || !strcasecmp(cdr_copy->action_type,"VM") || strcasestr(cdr_copy->dst,"DISA")){
			ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
			memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
		}
		/*end*/

		if (strcasestr(cdr_copy->action_type, "IVR")) {
			/*modify by ychpang when callee is IVR number and caller is callback, to remove record_files*/
			if (strcasestr(cdr_copy->action_type, cdr_copy->dst) && strcasestr(cdr_copy->src, "Callback")){
				ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
				memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
			}
			/* end */

			/* added by zgmo for Bug181980 */
			if (ast_strlen_zero(cdr_copy->dstchannel)) {
				ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
				memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
			}
			/* end */
		}

		if( !strncmp(cdr_copy->dstchannel,"Local",strlen("Local")) && strstr(cdr_copy->dstchannel,";1") && !strcmp(it_cdr->action_type, "TRANSFER") && strcmp(it_cdr->special_action_type, "SEAMLESS_TRANSFER"))
		{
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if( strstr(cdr_copy->action_type, "FOLLOWME") && (strlen(cdr_copy->dst) == 0 || !strncmp(cdr_copy->dstchannel,"IAX2/",strlen("Local"))) ){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		/*added by ychpang@grandstream.cn for bug 71581*/
		if( !strncmp(cdr_copy->channel,"Local",strlen("Local")) && strstr(cdr_copy->dst,cdr_copy->src) && strcasestr(cdr_copy->lastapp,"ConfBridge") && strstr(it_cdr->action_type, "CONFERENCE")){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if (AST_CDR_NULL == cdr_copy->disposition && ast_strlen_zero(cdr_copy->dstchannel) 
			&& !strcmp(it_cdr->special_action_type, "WAVE_SKIP_CALL")
			&& !strcmp(cdr_copy->action_type, "DIAL")) {
			ast_log(LOG_VERBOSE, "set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if (AST_CDR_NOANSWER == cdr_copy->disposition && !ast_strlen_zero(cdr_copy->dstchannel) 
			&& !strcmp(cdr_copy->action_type, "DIAL")
			&& (strstr(cdr_copy->channel, "trunk_") || strstr(cdr_copy->channel, "DAHDI"))) {
			if (!strcmp(it_cdr->special_action_type, "WAVE_SKIP_CALL")) {
				ast_log(LOG_VERBOSE, "set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}

			if (!strcmp(it_cdr->reason, ANSWERED_ELSEWHERE)) {
				struct cdr_object *it_cdr_prev_or_next = it_cdr->next;
				if (!it_cdr_prev_or_next) {
					it_cdr_prev_or_next = pre_it_cdr;
					ast_log(LOG_VERBOSE, "it_cdr_next is NULL, try previous one:%p", it_cdr_prev_or_next);
				}

				if (it_cdr_prev_or_next && ast_strlen_zero(it_cdr_prev_or_next->reason)
					&& !strcmp(it_cdr_prev_or_next->special_action_type, "WAVE_SKIP_CALL")) {
					ast_log(LOG_VERBOSE, "set cdr disable");
					ast_set_flag(cdr_copy, AST_CDR_FLAG_DISABLE);
				}
			}
		}

		/*add for bug 205663 by gangwang*/
		if (!strncmp(cdr_copy->action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING")) 
			&& AST_CDR_ANSWERED != cdr_copy->disposition) {
			if (!ast_strlen_zero(it_cdr->special_action_type) 
					&& !strncmp(it_cdr->special_action_type, "WAVE_SKIP_CALL", strlen("WAVE_SKIP_CALL"))) {
				int disable = 1;
				if (party_b) {
					ast_log(LOG_VERBOSE,"party_b->device_info %s", party_b->device_info);
					ast_log(LOG_VERBOSE,"party_b->device_info_peer %s", party_b->device_info_peer);
					if (!ast_strlen_zero(party_b->device_info) && !strcasestr(party_b->device_info, "Wave")) {
						disable = 0;
						cdr_copy->disposition = AST_CDR_ANSWERED;
						snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), ANSWERED_ELSEWHERE_NEW);
					}
				}

				if (disable) {
					ast_set_flag(cdr_copy, AST_CDR_FLAG_DISABLE);
					ast_log(LOG_DEBUG,"wave_skip_call set cdr disable");
				}
			}

			if (ast_strlen_zero(it_cdr->special_action_type)
				&& (AST_CDR_BUSY == cdr_copy->disposition || AST_CDR_NOANSWER == cdr_copy->disposition)) {
				struct cdr_object *it_cdr_next = it_cdr->next;

				if (!strcmp(cdr_copy->reason, ANSWERED_ELSEWHERE)
					|| (party_b && party_b->conf_hangupcause == AST_CAUSE_ANSWERED_ELSEWHERE)) {
					if (party_b && party_b->terminal_type == AST_TERMINAL_TYPE_SIP){
						ast_log(LOG_DEBUG,"should not disable not wave call cdr");
					}else{
						ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
						ast_log(LOG_DEBUG,"set cdr disable because wave join conference");
					}
				}

				if (AST_CDR_NOANSWER == cdr_copy->disposition && it_cdr_next
					&& !ast_strlen_zero(it_cdr_next->party_a.snapshot->name)
					&& !ast_strlen_zero(it_cdr_next->special_action_type)
					&& !strcmp(cdr_copy->channel, it_cdr_next->party_a.snapshot->name)
					&& !strcmp(it_cdr_next->reason, ANSWERED_ELSEWHERE)
					&& !strcmp(it_cdr_next->special_action_type, "WAVE_SKIP_CALL")) {
					ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
					ast_log(LOG_DEBUG,"set cdr disable because wave join conference");
				}

				if (0 == strncmp(cdr_copy->reason, ANSWERED_ELSEWHERE, strlen(ANSWERED_ELSEWHERE))
					&& ((strlen(cdr_copy->reason) > strlen(ANSWERED_ELSEWHERE)) || (party_b && party_b->terminal_type == AST_TERMINAL_TYPE_SIP))){
					snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), NOT_SEND_TO_WAVE );
				}
			}
		}

		if (party_b){
			if(strcasestr(party_b->device_info, "Wave") && cdr_copy->disposition != AST_CDR_ANSWERED
				&& 0 == strcmp(cdr_copy->action_type, "TRANSFER") && 0 == strcmp(cdr_copy->special_action_type, "ATTENDED_TRANSFER")){
				snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), ANSWERED_ELSEWHERE_NEW);
			}
		}

		/*emergency notify lack of src and owner*/
		if( !strlen(cdr_copy->src) && !strlen(cdr_copy->action_owner) && !strcmp(cdr_copy->action_type, "EMERGENCYNOTIFY")){
			snprintf(tmp_format, sizeof(tmp_format)-1, "%%*d-%%%lus", sizeof(tmp_ext)-1);
			sscanf(cdr_copy->session, tmp_format, tmp_ext);
			ast_copy_string(cdr_copy->src, tmp_ext, sizeof(cdr_copy->src));
			ast_copy_string(cdr_copy->action_owner, tmp_ext, sizeof(cdr_copy->action_owner));
		}

		/*voicemailmain dial out failed*/
		if(!strcmp(cdr_copy->dcontext, "bad-number") && !strcmp(cdr_copy->lastapp, "Congestion")){
			cdr_copy->disposition = AST_CDR_FAILED;
			cdr_copy->billsec = 0;
		}

		/*add for bug 109549 by ychpang*/
		if(0 == strcmp(cdr_copy->special_action_type, "VoiceMailMainCallback") &&(AST_CDR_BUSY == cdr_copy->disposition || AST_CDR_NOANSWER == cdr_copy->disposition)){
			cdr_copy->billsec = 0;
		}

		/*bug 115693*/
		if(!strcmp(cdr_copy->dst ,"s") && strcasestr(cdr_copy->clid, "Schedule Paging/Intercom") && !strlen(cdr_copy->dstchannel) && !strcmp(cdr_copy->lastapp, "AppDial2")){
			snprintf(tmp_format, sizeof(tmp_format)-1, "%%*d-%%%lus", sizeof(tmp_ext)-1);
			sscanf(cdr_copy->session, tmp_format, tmp_ext);
			ast_copy_string(cdr_copy->dst, tmp_ext, sizeof(cdr_copy->dst));
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type), "PAGE[%s]", tmp_ext);
		}

		/*bug144130*/
		if (!strcmp(cdr_copy->dst ,"s") && strcasestr(cdr_copy->session, "seamless") && AST_CDR_ANSWERED!= cdr_copy->disposition){
			ast_copy_string(cdr_copy->dst, strstr(cdr_copy->session, "_")+1, sizeof(cdr_copy->dst));
			ast_copy_string(cdr_copy->action_type,"TRANSFER", sizeof(cdr_copy->action_type));
			if (AST_CDR_FAILED == cdr_copy->disposition){
				cdr_copy->disposition = AST_CDR_NOANSWER;
			}
		}


		/*pincheck failed hang up*/
		if(!strcmp(cdr_copy->special_action_type, "CONF_INV_FAIL") && AST_CDR_NOANSWER == cdr_copy->disposition){
			   ast_log(LOG_VERBOSE,"set cdr disable");
			   ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}


		/*bug 118475*/
		if (0 == strncmp(it_cdr->special_action_type, "NOPM_TRANSFER", strlen("Local")) && 0 == strlen(cdr_copy->dstchannel)){
			ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
			memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
			cdr_copy->disposition = AST_CDR_FAILED;
			cdr_copy->billsec = 0;
		}
		/*Bug 144379 The CDR of dial the empty queue is incorrect. by gangwang*/
		long int disposition_res = cdr_copy->disposition;
		if ( 0 == strncmp(it_cdr->special_action_type, "EMPTYQUEUE", strlen("EMPTYQUEUE"))
			 && !strncmp(cdr_copy->action_type, "QUEUE", strlen("QUEUE"))){
			cdr_copy->disposition = AST_CDR_FAILED;
		}
		if ( 0 == strncmp(it_cdr->special_action_type, "EMPTYQUEUE_DEST", strlen("EMPTYQUEUE_DEST"))){
			if (disposition_res != AST_CDR_FAILED){
				cdr_copy->disposition = disposition_res;
			}
		}
		if (0 == strncmp(cdr_copy->action_type, "QUEUE", strlen("QUEUE")) && AST_CDR_ANSWERED == cdr_copy->disposition
			&& (it_cdr->next && 0 == strncmp(it_cdr->next->action_type, "QUEUE", strlen("QUEUE")) && 0 == strcmp(it_cdr->next->special_action_type, "HAVE_MEMBER_BUT_NULL"))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*the queue has agents but the agents all can't access*/
		if ((!strcmp(cdr_copy->lastapp, "ForkCDR") && !strncmp(cdr_copy->action_type, "QUEUE", strlen("QUEUE")))
			&& (it_cdr->next && !strncmp(it_cdr->next->action_type, "QUEUE", strlen("QUEUE")) && !strcmp(it_cdr->next->special_action_type, "EMPTYQUEUE"))) {
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);

			/* added by zgmo for Bug175523 */
			if (!ast_strlen_zero(cdr_copy->record_files)) {
				const char *mixmon_dir = pbx_builtin_getvar_helper(NULL, "MIXMON_DIR");
				if (!ast_strlen_zero(mixmon_dir)) {
					int path_len = strlen(mixmon_dir) + strlen(cdr_copy->record_files) + 2;
					char *path = (char *)ast_malloc(path_len);
					if (path) {
						char *p;
						sprintf(path, "%s/", mixmon_dir);
						strcat(path, cdr_copy->record_files);

						p = strchr(path, '@');
						if (p) {
							*p = '\0';
						}
						ast_log(LOG_DEBUG, "remove record file:%s", path);
						remove(path);
						ast_free(path);
					}
				}
			}
			/* end */
		}

		/*bug 118481*/
		if (0 == strncmp(it_cdr->special_action_type, "OPERATOR_TRANSFER_FAILED", strlen("OPERATOR_TRANSFER_FAILED"))) {
			cdr_copy->disposition = AST_CDR_FAILED;
			cdr_copy->billsec = 0;
		}

		if(!strncmp(cdr_copy->action_type, "DIRECTORY", strlen("DIRECTORY"))
			|| (!strncmp(cdr_copy->action_type, "PRESENCE_STATUS", strlen("PRESENCE_STATUS")) && (!strncmp(cdr_copy->channel,"IAX2/",strlen("IAX2/")) || !strncmp(cdr_copy->channel,"DAHDI/",strlen("DAHDI/"))))){
				ast_copy_string(cdr_copy->action_type,"DIAL", sizeof(cdr_copy->action_type));
		}

		/*bug 119752*/
		if (0 == strncmp(cdr_copy->dst,"normalrel",strlen("normalrel")) && 0 == strncmp(cdr_copy->action_type,"CALLFORWARD",strlen("CALLFORWARD"))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		if (strcasestr(it_cdr->special_action_type, "SEAMLESS_TRANSFER") && party_b && !ast_strlen_zero(party_b->caller_number) && strcasestr(cdr_copy->dstchannel, "trunk_")){
			ast_copy_string(cdr_copy->dst,party_b->caller_number, sizeof(cdr_copy->dst));
		}

		/*bug 155257*/
		if (!ast_strlen_zero(cdr_copy->channel) && !ast_strlen_zero(cdr_copy->dstchannel) && strncmp(cdr_copy->dstchannel,"Local",strlen("Local")) && strncmp(cdr_copy->channel,"Local",strlen("Local"))
			&& !strcmp(cdr_copy->lastapp, "ConfBridge")
			&& ((!strcmp(cdr_copy->action_type, "TRANSFER") && strstr(it_cdr->special_action_type, "SEAMLESS_TRANSFER")) || strstr(it_cdr->special_action_type, "SEAMLESS_TRANSFER_FXS"))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*bug 119044*/
		if (0 == strncmp(it_cdr->special_action_type, "SEAMLESS_TRANSFER_HANGUP", strlen("SEAMLESS_TRANSFER_HANGUP"))
			|| (ast_strlen_zero(cdr_copy->dst) && (strcasestr(cdr_copy->dstchannel, "DAHDI") || strcasestr(cdr_copy->dstchannel, "trunk_")) && 0 == strncmp(cdr_copy->lastapp, "ConfBridge", strlen("ConfBridge")))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*bug 120648*/
		if (NULL != strcasestr(it_cdr->action_type, "CONFERENCE") || NULL != strcasestr(it_cdr->action_type, "MULTIMEDIAMEETING")){
			ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
			memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
		}

		/*bug 121093*/
		if (0 == strncmp(cdr_copy->action_type, "CALLFORWARD", strlen("CALLFORWARD")) && ast_strlen_zero(cdr_copy->dstchannel) && 0 == strncmp(cdr_copy->lastapp, "Park", strlen("Park"))){
			ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
			memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
		}

		/*bug 122622*/
		if (0 == strncmp(it_cdr->special_action_type, "COUNTRY_CODE", strlen("COUNTRY_CODE")) && 0 == strncmp(cdr_copy->lastapp, "Hangup", strlen("Hangup"))){
			cdr_copy->disposition = AST_CDR_FAILED;
		}
		/*bug 123834*/
		if (strcasestr(it_cdr->special_action_type, "SCA") && 0 == strncmp(cdr_copy->dcontext, "from-did-direct", strlen("from-did-direct"))){
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type),"SCA");
		}

		/*bug 143565*/
		if (strcasestr(cdr_copy->action_type, "SCA") && !strcasecmp(cdr_copy->lastapp, "Sca")
			&& (!ast_strlen_zero(cdr_copy->lastdata) && !ast_strlen_zero(cdr_copy->dst) && strcmp(cdr_copy->dst, cdr_copy->lastdata))){
			snprintf(cdr_copy->dst, sizeof(cdr_copy->action_type),cdr_copy->lastdata);
		}

		if(strcasestr(it_cdr->special_action_type,"local_disa")
			&& (!ast_strlen_zero(cdr_copy->dstchannel) && strcasestr(cdr_copy->dstchannel, "trunk_"))){
			ast_clear_flag(cdr_copy, AST_CDR_FLAG_DISABLE|AST_CDR_FLAG_DISABLE_ALL);
			ast_log(LOG_VERBOSE, "clear cdr disable flag");
		}


		/*wakeup app,old channel hang up,Bug 269407 added by zyxia,2023.8.22*/
		if(!strcmp(it_cdr->special_action_type, "WAKEUP_APP") && AST_CDR_FAILED == it_cdr->disposition){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*Bug 281996*/
		if (strcasestr(cdr_copy->action_type,"MULTIMEDIAMEETING") && strcasestr(cdr_copy->channel, "Local") && strcasestr(cdr_copy->channel, ";2")) {
			if (AST_CDR_ANSWERED != cdr_copy->disposition && !strcmp(cdr_copy->special_action_type, "CONF_WAKEUP")) {
				ast_log(LOG_VERBOSE,"CONF_WAKEUP set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		if (strcasestr(cdr_copy->action_type,"CALLFORWARD") && strcasestr(cdr_copy->dstchannel, "Local") && strcasestr(cdr_copy->dstchannel, "call-forward")) {
			if (AST_CDR_ANSWERED != cdr_copy->disposition && !strcmp(cdr_copy->special_action_type, "WAVE_SKIP_CALL")) {
				ast_log(LOG_VERBOSE,"CALLFORWARD WAVE_SKIP_CALL set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		if (strcasestr(cdr_copy->channel, "Local") && strcasestr(cdr_copy->channel, ";2") && strcasestr(cdr_copy->channel, "call-forward")
			&& !ast_strlen_zero(cdr_copy->dst) && strcasestr(cdr_copy->channel, cdr_copy->dst)) {
			if (AST_CDR_ANSWERED != cdr_copy->disposition && !strcmp(cdr_copy->special_action_type, "WAVE_SKIP_CALL")) {
				ast_log(LOG_VERBOSE,"CALLFORWARD WAVE_SKIP_CALL set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		if ( strcasestr(it_cdr->special_action_type,"local_disa") && strcasestr(cdr_copy->channel, "Local") && strcasestr(cdr_copy->channel, ";2")){
			if (!ast_strlen_zero(cdr_copy->dstchannel) && strcasestr(cdr_copy->dstchannel, "Local") && strcasestr(cdr_copy->dstchannel, ";1")){
				ast_log(LOG_VERBOSE,"set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		/*bug 120554*/
		if (0 != strncmp(cdr_copy->dst, "DISA", strlen("DISA")) && 0 == strncmp(cdr_copy->action_type, "DISA", strlen("DISA"))
			&& (strcasestr(cdr_copy->dstchannel, "Local") && strcasestr(cdr_copy->dstchannel, ";1"))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		if (0 == cdr_copy->billsec && 0 == strncmp(cdr_copy->action_type, "DISA", strlen("DISA"))){
			cdr_copy->disposition = AST_CDR_NOANSWER;
		}

		/*bug 124932*/
		if (0 != strncmp(cdr_copy->dst, "DISA", strlen("DISA")) && 0 == strncmp(cdr_copy->action_type, "DISA", strlen("DISA")) && !ast_strlen_zero(it_cdr->special_action_type)){
			ast_copy_string(cdr_copy->clid, it_cdr->special_action_type, sizeof(cdr_copy->clid));
		}

		/*for forkcdr with null dst chan, bug 214177*/
		if (!party_b && 0 == strcmp(cdr_copy->lastapp, "ForkCDR")
			&& 0 == strncmp(cdr_copy->action_type, "RINGGROUP", strlen("RINGGROUP"))
			&& strstr(cdr_copy->action_type, (party_a ? party_a->connected_number : "NULL"))){
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}

		/*do special process for callee transfer by *2*/
		if ('\0' != transfer_callernum[0]){
			ast_copy_string(cdr_copy->src, transfer_callernum, sizeof(cdr_copy->src));
			ast_copy_string(cdr_copy->inbound_trunk_index, atx_inbound_index, sizeof(cdr_copy->inbound_trunk_index));
			ast_log(LOG_DEBUG, "Update cdr_copy->src -> '%s' due to transfer_callernum", transfer_callernum);

			if ('\0' != transfer_callername[0]){
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", transfer_callername, cdr_copy->src);
			} else{
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"\" <%s>", cdr_copy->src);
			}
			snprintf(cdr_copy->action_owner, sizeof(cdr_copy->action_owner)-1, "%s", transfer_callernum);
			memset(transfer_callernum, 0, sizeof(transfer_callernum));
			memset(transfer_callername, 0, sizeof(transfer_callername));
			memset(atx_inbound_index, 0, sizeof(atx_inbound_index));
			//pre_it_cdr->party_a.snapshot->name
			if (pre_it_cdr
				&& strcasestr(cdr_copy->dstchannel, "Local") && strcasestr(cdr_copy->dstchannel, ";1") && strcasestr(cdr_copy->dstchannel, "xfer")){
				atx_keep_nat_flag = 1;
				cdr_copy->is_nat_call = pre_it_cdr->peer_is_nat_call;
			}
		}
		struct ast_channel_snapshot *next_party_b = NULL;
		if (it_cdr->next && it_cdr->next->party_b.snapshot){
			next_party_b = it_cdr->next->party_b.snapshot;
		}

		ast_log(LOG_DEBUG, "party_a:%p | caller_number:%s name:%s", party_a, party_a ? party_a->caller_number : "NULL", party_a ? party_a->caller_name : "NULL");
		ast_log(LOG_DEBUG, "party_a:%p | connected_number:%s name:%s", party_a, party_a ? party_a->connected_number : "NULL", party_a ? party_a->connected_name : "NULL");
		ast_log(LOG_DEBUG, "party_b:%p | caller_number:%s name:%s", party_b, party_b ? party_b->caller_number : "NULL", party_b ? party_b->caller_name : "NULL");
		ast_log(LOG_DEBUG, "party_b:%p | connected_number:%s name:%s", party_b, party_b ? party_b->connected_number : "NULL", party_b ? party_b->connected_name : "NULL");
		ast_log(LOG_DEBUG, "next_party_b:%p | name:%s", next_party_b, next_party_b ? next_party_b->name : "NULL");

		if (!ast_strlen_zero(cdr_copy->dstchannel) && NULL != strcasestr(cdr_copy->dstchannel, "trunk_")
			&& 0 == strcmp(cdr_copy->action_type, "CALL FLIP")
			&& 0 == strcmp(cdr_copy->special_action_type, "CALL_FLIP_TRANSFER")
			&& party_b && party_b->connected_number){
			ast_copy_string(cdr_copy->dst, party_b->connected_number, sizeof(cdr_copy->dst));
		}

		if (ast_strlen_zero(cdr_copy->dstchannel)
			&&(it_cdr->next && 0 == strcmp(it_cdr->next->appl, "SLATrunk"))
			&& (next_party_b && !ast_strlen_zero(next_party_b->name) && NULL == strstr(next_party_b->name, cdr_copy->dst))){
			char tmp_num[128] = {0};
			sscanf(next_party_b->name, "%*[^/]/%127[^-]-%*s", tmp_num);
			if (!ast_strlen_zero(tmp_num)){
				ast_copy_string(cdr_copy->dst, tmp_num, sizeof(cdr_copy->dst));
			}
		}

		if (party_a->connected_number && party_b && party_b->caller_number
			&& next_party_b && next_party_b->name
			&& strstr(next_party_b->name, "xfer") && strstr(next_party_b->name, ";1")){
			int lets_do_it = 0;
			if (strcmp(party_a->connected_number, party_b->caller_number) || NULL != strstr(cdr_copy->channel, "trunk")) {
				lets_do_it = 1;
			} else {
				if (!ast_strlen_zero(cdr->next->record_files)
					&& strstr(cdr->next->record_files, party_b->caller_number)) {
					lets_do_it = 1;
				}
			}

			if (lets_do_it) {
				/*cp the party b's caller number and name*/
				snprintf(transfer_callernum, sizeof(transfer_callernum)-1, "%s", party_b->caller_number);
				ast_log(LOG_DEBUG, "UPDATE transfer_callernum to '%s'", transfer_callernum);
				if (NULL != strstr(cdr_copy->channel, "trunk_")){
					snprintf(atx_inbound_index, sizeof(atx_inbound_index)-1, "%s", cdr_copy->outbound_trunk_index);
				}

				if (party_b->caller_name){
					snprintf(transfer_callername, sizeof(transfer_callername)-1, "%s", party_b->caller_name);
				}
			}
		}

		if (party_a->connected_number && party_b && party_b->caller_number && !next_party_b
			&& 0 == strcmp(party_b->context, "followme-from-internal-xfer")){
			memset(cdr_copy->dst, 0, sizeof(cdr_copy->dst));
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}

		if (party_a && party_a->connected_number && !strcmp(it_cdr->special_action_type, "EMPTYQUEUE") 
			&& !strcmp(cdr_copy->lastapp, "Hangup") && !strncmp(cdr_copy->action_type, "QUEUE", strlen("QUEUE")) 
			&& !strcmp(cdr_copy->dst, "t")) {
			memset(cdr_copy->dst, 0, sizeof(cdr_copy->dst));
			ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
		}

		if (party_b && 0 == strcmp(party_b->call_flip, "yes"))
		{
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type),"CALL FLIP");
		}

		/*bug 119237*/
		if (1 == transfer_rp && NULL != strstr(cdr_copy->src, "<")){
			/*need some data to set src and clid*/
			char tmp_name[128] = {0};
			ast_copy_string(cdr_copy->clid, cdr_copy->src, sizeof(cdr_copy->clid));
			if (strstr(cdr_copy->src, "\"\"")){
				sscanf(cdr_copy->src, "%*[^<]<%127[\\+0-9a-zA-Z]%*s", tmp_caller);
				snprintf(tmp_name, sizeof(tmp_name), "%s", tmp_caller);
			}
			else{
				sscanf(cdr_copy->src, "\"%[^\"]\"<%127[\\+0-9a-zA-Z]%*s", tmp_name, tmp_caller);
			}
			if (0 < strlen(tmp_caller)){
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", tmp_name, tmp_caller);
				ast_copy_string(cdr_copy->src, tmp_caller, sizeof(cdr_copy->src));
			}

			ast_log(LOG_DEBUG, "cdr_copy->src is UPDATED to '%s'", cdr_copy->src);
		}

		if (!strncmp(cdr_copy->channel,"Local/",strlen("Local/")) && strstr(cdr_copy->channel,";2") && strstr(cdr_copy->channel,"xfer")){
			if (party_a->caller_number && strcmp(cdr_copy->src, party_a->caller_number)){
				ast_copy_string(cdr_copy->src, party_a->caller_number, sizeof(cdr_copy->src));
				ast_log(LOG_DEBUG, "cdr_copy->src is UPDATED to '%s'", cdr_copy->src);
			}
			if (party_a->caller_name && '\0' != party_a->caller_name[0]){
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", party_a->caller_name, cdr_copy->src);
			}else{
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"\" <%s>", cdr_copy->src);
			}
		}

		if (2 == transfer_rp
			&& (NULL != strstr(cdr_copy->channel, "Local") && NULL != strstr(cdr_copy->channel, "from-internal-xfer") && NULL != strstr(cdr_copy->channel, ";2"))
			&& !ast_strlen_zero(tmp_caller)){
			ast_copy_string(cdr_copy->src, tmp_caller, sizeof(cdr_copy->src));
			ast_log(LOG_DEBUG, "only for extension have more than one contacts");
		}

		if (1 == transfer_rp && it_cdr->next && next_party_b){
			const char* next_src_chan=it_cdr->next->party_a.snapshot->name;
			const char* next_dst_chan=next_party_b->name;
			if (it_cdr->next->disposition != AST_CDR_ANSWERED
				&& !ast_strlen_zero(next_src_chan) && 0 == strcmp(next_src_chan, cdr_copy->channel)
				&& (NULL != strstr(next_src_chan, "Local") && NULL != strstr(next_src_chan, "from-internal-xfer") && NULL != strstr(next_src_chan, ";2"))
				&& !ast_strlen_zero(next_dst_chan) && !ast_strlen_zero(cdr_copy->dstchannel)){
				int len_now=strlen(cdr_copy->dstchannel);
				int len_next=strlen(next_dst_chan);
				if (len_now==len_next && len_next > strlen("00000000") && 0 == strncmp(cdr_copy->dstchannel, next_dst_chan, len_next - strlen("00000000"))){
					transfer_rp = 2;
					ast_log(LOG_DEBUG, "only for extension have more than one contacts");
				}
			}
		}

		if (!strncmp(cdr_copy->channel,"Local/",strlen("Local/")) && strstr(cdr_copy->channel,";1") && strstr(cdr_copy->channel,"xfer") && !strcmp(cdr_copy->action_type, "DIAL")){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		/*bug 128404*/
		if (ast_strlen_zero(cdr_copy->action_owner) && !ast_strlen_zero(cdr_copy->clid)){
			char tmp_src[128] = {0};
			sscanf(cdr_copy->clid, "%*[^<]<%127[^>]", tmp_src);
			if (!ast_strlen_zero(tmp_src))
			{
				snprintf(cdr_copy->action_owner, sizeof(cdr_copy->action_owner), tmp_src);
			}
			else
			{
				snprintf(cdr_copy->action_owner, sizeof(cdr_copy->action_owner), "anonymous");
			}
		}

		/*bug 133655*/
		if (0 == strcmp(cdr_copy->dst, "fax") && 0 == strncmp(cdr_copy->channel, "PJSIP/trunk_", strlen("PJSIP/trunk_")) && ast_strlen_zero(cdr_copy->dstchannel)){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*136075*/
		if ( 0 == strcmp(cdr_copy->src, "VFAX-"))
		{
			ast_copy_string(cdr_copy->src, "VFAX-admin", sizeof(cdr_copy->src));
		}
		if ( 0 == strcmp(cdr_copy->action_owner, "VFAX-"))
		{
			ast_copy_string(cdr_copy->action_owner, "VFAX-admin", sizeof(cdr_copy->action_owner));
		}
		#if 0
		/*bug 146434 / It is different with 65xx*/
		if (NULL != strstr(cdr_copy->dstchannel, ";1") && 0 == strncmp(cdr_copy->dstchannel, "Local", 5) && NULL != strstr(cdr_copy->dstchannel, "call-forward")
			&& AST_CDR_ANSWERED == cdr_copy->disposition && !ast_tvzero(cdr_copy->answer) && !it_cdr->next && '\0' == cdr_copy->record_files[0]){
			cdr_copy->end = cdr_copy->answer;
			cdr_copy->duration = (long)(ast_tvdiff_ms(ast_tvzero(cdr_copy->end) ? ast_tvnow() : cdr_copy->end, cdr_copy->start) / 1000);
			cdr_copy->disposition = AST_CDR_NOANSWER;
		}
		#endif
		/*bug 139428*/
		if (strstr(cdr_copy->action_type, "ANNOUNCE[") && AST_CDR_ANSWERED != cdr_copy->disposition){
			cdr_copy->duration = (long)(ast_tvdiff_ms(ast_tvzero(cdr_copy->end) ? ast_tvnow() : cdr_copy->end, cdr_copy->start) / 1000);
			cdr_copy->disposition = AST_CDR_ANSWERED;
			cdr_copy->billsec = cdr_copy->duration;
			cdr_copy->answer = cdr_copy->start;
		}
		if (AST_CDR_ANSWERED != cdr_copy->disposition && !ast_tvzero(cdr_copy->answer)){
			cdr_copy->billsec = 0;
			cdr_copy->end = cdr_copy->answer;
		}

		/*bug 147875*/
		if (0 == strcmp(cdr_copy->src, "Callback") && 0 == strcmp(cdr_copy->action_type, "CALLBACK")){
			snprintf(cdr_copy->clid, sizeof(cdr_copy->clid)-1, "\"Callback\" <Callback>");
			cdr_copy->start.tv_sec = cdr_copy->start.tv_sec-1;
		}
		/*bug 152184*/
		if (0 == strncmp(cdr_copy->action_type, "CALLFORWARD", strlen("CALLFORWARD"))
			&& (!strcasestr(cdr_copy->channel, "call-forward") && !strcasestr(cdr_copy->dstchannel, "call-forward"))
			&& (it_cdr->next
				&& !strncmp(it_cdr->next->action_type, "CALLFORWARD", strlen("CALLFORWARD"))
				&& (strcmp(it_cdr->next->appl, "Busy") && strcmp(it_cdr->next->appl, "Hangup"))
				&& (it_cdr->next->party_b.snapshot && it_cdr->next->party_b.snapshot->name && strcasestr(it_cdr->next->party_b.snapshot->name, "call-forward")))){
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type),"DIAL");
		}

		/*182792*/
		if(1 == is_app_page && 0 < strlen(tmp_page_ext)){
			snprintf(cdr_copy->dst,sizeof(cdr_copy->dst),"%s",tmp_page_ext);
		}
		if(-1 == is_app_page && 0 < strlen(tmp_page_ext)){
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type), "CALLFORWARD");
		}
		if (0 == strcmp(cdr_copy->lastapp, "ForkCDR") && 0 == strcmp(it_cdr->special_action_type, "APP_PAGING")
			&& (it_cdr->next && !it_cdr->next->party_b.snapshot)){
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		if (0 == strcmp(cdr_copy->lastapp, "ForkCDR") && 0 == strcmp(it_cdr->special_action_type, "APP_PAGING")
			&& it_cdr->next && it_cdr->next->session && it_cdr->next->party_b.snapshot && it_cdr->next->party_b.snapshot->name){
			sscanf(cdr_copy->session, "%*d-%s", tmp_page_ext);
			char next_callee[128] = {0};
			sscanf((char *)(it_cdr->next->party_b.snapshot->name),"%*[^/]/%63[\\+0-9*#]%*s", next_callee);
			if(0 != strcmp(next_callee, tmp_page_ext)){
				is_app_page = -1;
			}else{
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
				is_app_page = 1;
				snprintf(tmp_page_ext,sizeof(tmp_page_ext),"%s",it_cdr->exten);
			}
		}

		/*fxs caller dial dnd callee, if the callee disable vm and have set dnd, the disposition should be busy*/
		if (!strncmp(cdr_copy->channel, "DAHDI/", strlen("DAHDI/"))
			&& !strcmp(cdr_copy->special_action_type, "DO_DND")
			&& AST_CDR_ANSWERED == cdr_copy->disposition && strcmp(it_cdr->action_type, "VM")){
			cdr_copy->disposition = AST_CDR_BUSY;
		}

		/*add reason for wave call history, be the same as sip's warning head*/
		if (AST_CDR_BUSY == cdr_copy->disposition ){
			if (0 == strcmp(it_cdr->special_action_type, "DO_DND")){
				snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), "DND");
			}else if (0 == strcmp(it_cdr->special_action_type, "CALLWAITING")){
				snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), "INUSE");
			}
		}

		/*only for seamless transfer by trunk group using local channel, extremly special modify*/
		if (!strcasecmp(cdr_copy->src,cdr_copy->dst) && strcasestr(it_cdr->special_action_type, "SEAMLESS_TRANSFER")
			&& strcasestr(cdr_copy->channel, "Local") && strstr(cdr_copy->channel, ";1") && ast_strlen_zero(cdr_copy->dstchannel)
			&& !ast_strlen_zero(cdr_copy->session) && !ast_strlen_zero(cdr_copy->outbound_trunk_index)){
			char family[64] = {0};
			char trunkName[128] = {0};
			memset(tmp_format, 0, sizeof(tmp_format));
			memset(tmp_ext, 0, sizeof(tmp_ext));
			snprintf(tmp_format, sizeof(tmp_format)-1, "%%*d-%%%lus", sizeof(tmp_ext)-1);
			sscanf(cdr_copy->session, tmp_format, tmp_ext);
			ast_copy_string(cdr_copy->dst, tmp_ext, sizeof(cdr_copy->dst));
			ast_copy_string(cdr_copy->inbound_trunk_index, cdr_copy->outbound_trunk_index, sizeof(cdr_copy->inbound_trunk_index));
			memset(cdr_copy->outbound_trunk_index, 0, sizeof(cdr_copy->outbound_trunk_index));
			snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(cdr_copy->inbound_trunk_index));
			ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
			ast_copy_string(cdr_copy->src_trunk_name, trunkName, sizeof(cdr_copy->src_trunk_name));
		}
		if (strcasestr(it_cdr->special_action_type, "SEAMLESS_TRANSFER") && strcasestr(cdr_copy->dstchannel, "Local") && strstr(cdr_copy->dstchannel, ";1")){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		/*end*/
		/*callee seamless transfer trunk err*/
		if (strcasestr(cdr_copy->session, "calleeseamless") && ast_strlen_zero(cdr_copy->dstchannel)
			&& (!ast_strlen_zero(cdr_copy->outbound_trunk_index) || !ast_strlen_zero(cdr_copy->inbound_trunk_index))){
			if (!ast_strlen_zero(cdr_copy->outbound_trunk_index) && !ast_strlen_zero(cdr_copy->inbound_trunk_index)){
				char *tmp_trunkidx = NULL;
				tmp_trunkidx = strdup(cdr_copy->outbound_trunk_index);
				memset(cdr_copy->outbound_trunk_index, 0, sizeof(cdr_copy->outbound_trunk_index));
				ast_copy_string(cdr_copy->outbound_trunk_index, cdr_copy->inbound_trunk_index, sizeof(cdr_copy->outbound_trunk_index));
				memset(cdr_copy->inbound_trunk_index, 0, sizeof(cdr_copy->inbound_trunk_index));
				ast_copy_string(cdr_copy->inbound_trunk_index, tmp_trunkidx, sizeof(cdr_copy->inbound_trunk_index));
				if (tmp_trunkidx){
					ast_free(tmp_trunkidx);
					tmp_trunkidx = NULL;
				}
			}else if (!ast_strlen_zero(cdr_copy->outbound_trunk_index)){
				memset(cdr_copy->inbound_trunk_index, 0, sizeof(cdr_copy->inbound_trunk_index));
				ast_copy_string(cdr_copy->inbound_trunk_index, cdr_copy->outbound_trunk_index, sizeof(cdr_copy->inbound_trunk_index));
				memset(cdr_copy->outbound_trunk_index, 0, sizeof(cdr_copy->outbound_trunk_index));
			}else if (!ast_strlen_zero(cdr_copy->inbound_trunk_index)){
				memset(cdr_copy->outbound_trunk_index, 0, sizeof(cdr_copy->outbound_trunk_index));
				ast_copy_string(cdr_copy->outbound_trunk_index, cdr_copy->inbound_trunk_index, sizeof(cdr_copy->outbound_trunk_index));
				memset(cdr_copy->inbound_trunk_index, 0, sizeof(cdr_copy->inbound_trunk_index));
			}
		}
		/*end*/

		/*bug 233580*/
		if (ast_strlen_zero(cdr_copy->dstchannel)
			&& (!ast_strlen_zero(it_cdr->special_action_type) && strstr(it_cdr->special_action_type, "SEAMLESS_TRANSFER"))
			&& 0 == strcmp(cdr_copy->dst, "s"))
		{
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*fxs call forward busy*/
		if (AST_CDR_NOANSWER == cdr_copy->disposition && it_cdr->special_action_type && !strcmp(it_cdr->special_action_type, "FXS_CFB")){
			cdr_copy->disposition = AST_CDR_BUSY;
		}
		/*end*/

		/*for attend transfer callee not answer then dial goback*/
		if (clear_atx_rec && !ast_strlen_zero(cdr_copy->record_files)){
			ast_log(LOG_DEBUG, "should clear the record file");
			memset(cdr_copy->record_files,0,sizeof(cdr_copy->record_files));
			if (ast_strlen_zero(cdr_copy->dst) && !ast_strlen_zero(party_a->connected_number)){
				ast_copy_string(cdr_copy->dst, party_a->connected_number, sizeof(cdr_copy->dst));
			}
		}
		if (strcasestr(cdr_copy->dstchannel, "Local") && strcasestr(cdr_copy->dstchannel, ";1") && strcasestr(cdr_copy->dstchannel, "xfer")
			&& (party_b && (NULL == party_b->bridgeid || '\0'==party_b->bridgeid[0])) && ast_strlen_zero(party_b->hangupsource)) {
				ast_log(LOG_VERBOSE,"cdr disable");
				clear_atx_rec = 1;
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}
		/*end*/

		/*use phone do attend transfer to outbound number through reg trunk, callee err*/
		if (!ast_strlen_zero(party_a->reg_calleenum) && !ast_strlen_zero(cdr_copy->dstchannel) && strcasestr(cdr_copy->dstchannel, "trunk_") && 0 == strcmp(it_cdr->special_action_type, "atx_refer")){
			ast_copy_string(cdr_copy->dst, party_a->reg_calleenum, sizeof(cdr_copy->dst));
		}

		/* added by zgmo for Bug182653 */
		if (strcasestr(cdr_copy->action_type, "PAGE")) {
			if (ast_strlen_zero(cdr_copy->dstchannel) && AST_CDR_ANSWERED == cdr_copy->disposition && ast_strlen_zero(it_cdr->record_files_desc)) {
				ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
				cdr_copy->record_files[0] = '\0';
			}
		}
		/* end */

		/* Bug 166209 */
		if (AST_CDR_FAILED == cdr_copy->disposition
			&& !ast_strlen_zero(cdr_copy->record_files)) {
			ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]", it_cdr, cdr_copy->record_files);
			cdr_copy->record_files[0] = '\0';
		}

		/*bug 171834, schedule paging/intercom with not registed caller*/
		if (!strcmp(cdr_copy->channel, "OutgoingSpoolFailed") && !strcmp(cdr_copy->dst, "failed") && strcasestr(cdr_copy->clid, "<SchedulePaging/Intercom>")){
			char dst[64] = {0};
			sscanf(it_cdr->session,"%*[^-]-%63[\\+0-9a-zA-Z]%*s",dst);
			ast_copy_string(cdr_copy->dst, dst, sizeof(cdr_copy->dst));
			ast_copy_string(cdr_copy->action_owner, cdr_copy->src, sizeof(cdr_copy->action_owner));
			snprintf(cdr_copy->action_type, AST_MAX_EXTENSION, "PAGE[%s]", cdr_copy->dst);
		}
		int set_clid = 0;

		/*phone press key do attend transfer, caller is err, modify it*/
		if (0 == strcmp(it_cdr->special_action_type, "atx_refer") && !ast_strlen_zero(cdr_copy->inbound_trunk_index) && !ast_strlen_zero(cdr_copy->clid)){
			char tmp_src[256] = {0};
			sscanf(cdr_copy->clid, "%*[^<]<%255[^>]", tmp_src);
			ast_copy_string(cdr_copy->src, tmp_src, sizeof(cdr_copy->src));
		}

		/*bug 200252*/
		if (NULL == strstr(cdr_copy->session, "seamless") && 0 == strcmp(cdr_copy->action_type, "TRANSFER") && 0 == strcmp(cdr_copy->special_action_type, "SEAMLESS_TRANSFER")){
			memset(cdr_copy->action_type, 0, sizeof(cdr_copy->action_type));
			ast_copy_string(cdr_copy->action_type,"DIAL", sizeof(cdr_copy->action_type));
		}

		/*bug 158787, local;2 hangup too early cause record err*/
		if (strstr(cdr_copy->dstchannel, "DAHDI")
			&& strstr(cdr_copy->action_type, "MULTIMEDIAMEETING") && strstr(it_cdr->special_action_type, "Scheduled Conference")){
			if (cdr_copy->disposition == AST_CDR_ANSWERED
				&& (strstr(cdr_copy->channel, "Local")&& strstr(cdr_copy->channel, ";2"))){
				ast_log(LOG_VERBOSE,"set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}else{
			if (cdr_copy->session && it_cdr->special_action_type && (strstr(it_cdr->special_action_type, "Scheduled Conference")||strstr(it_cdr->special_action_type, "issfu"))){
				if (strstr(it_cdr->special_action_type, "issfu")){
					if (!ast_strlen_zero(party_a->caller_num)){
						ast_copy_string(tmp_ext, party_a->caller_num, sizeof(tmp_ext));
					}
					else{
						ast_copy_string(tmp_ext, party_a->caller_number, sizeof(tmp_ext));
					}
					ast_copy_string(cdr_copy->dstchannel, cdr_copy->channel, sizeof(cdr_copy->dstchannel));
					snprintf(cdr_copy->channel, sizeof(cdr_copy->channel), "Local/%s@from-internal-%s;2", tmp_ext, cdr_copy->session);
					memset(cdr_copy->inbound_trunk_index, 0, sizeof(cdr_copy->inbound_trunk_index));
				}else{
					ast_copy_string(tmp_ext, cdr_copy->src, sizeof(tmp_ext));
				}
				if (0 != strcmp(cdr_copy->dst, "s")){
					ast_copy_string(cdr_copy->src, cdr_copy->dst, sizeof(cdr_copy->src));
				}else{
					ast_copy_string(cdr_copy->src, strstr(cdr_copy->session, "-")+1, sizeof(cdr_copy->src));
					snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type), "MULTIMEDIAMEETING[%s]", cdr_copy->src);
				}
				ast_copy_string(cdr_copy->action_owner, cdr_copy->src, sizeof(cdr_copy->action_owner));
				ast_copy_string(cdr_copy->dst, tmp_ext, sizeof(cdr_copy->dst));
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", strstr(it_cdr->special_action_type, "|")+1, cdr_copy->src);
				memset(tmp_ext, 0, sizeof(tmp_ext));
				set_clid = 1;

				ast_log(LOG_DEBUG, "cdr_copy->src is UPDATED to '%s'", cdr_copy->src);
			}
			if (strstr(cdr_copy->action_type, "MULTIMEDIAMEETING")
				&& (!strstr(it_cdr->special_action_type, "Scheduled Conference")&&!strstr(it_cdr->special_action_type, "issfu"))
				&& AST_CDR_ANSWERED == cdr_copy->disposition
				&& !ast_strlen_zero(cdr_copy->dstchannel) && 0 == cdr_copy->billsec && 0 == set_clid){
				ast_log(LOG_VERBOSE,"cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		if (0 == strncmp(cdr_copy->channel, "MulticastRTP", strlen("MulticastRTP")) && !ast_strlen_zero(cdr_copy->clid)){
			char tmp_src[128] = {0};
			sscanf(cdr_copy->clid, "%*[^<]<%127[^>]", tmp_src);
			if (!ast_strlen_zero(tmp_src))
			{
				snprintf(cdr_copy->src, sizeof(cdr_copy->src), tmp_src);
			}
		}

		if (0 == strcmp(it_cdr->special_action_type, "UNREACHABLE") && !party_b){
			cdr_copy->disposition=AST_CDR_FAILED;
			cdr_copy->billsec = 0;
		}

		/*for park lot callback cause callee err, modify it.*/
		if ((!strcmp(cdr_copy->action_type, "PARKEDCALL") || !strcmp(cdr_copy->dcontext, "park-dial")) && strstr(cdr_copy->dst, "_")){
			char tmp_dst[128] = {0};
			if (NULL != strcasestr(cdr_copy->dst, "DAHDI")){
				int i;
				char family[128] = {0};
				char key[32] = {0};
				ast_copy_string(key, "FXS_EXTENSION", sizeof(key));
				ast_copy_string(family, cdr_copy->dst, sizeof(family));
				for (i = 0; family[i]; i++) {
					if (family[i] == '_') {
						family[i] = '/';
					}
				}
				ast_ucm_db_get(family, key, tmp_dst, sizeof(tmp_dst) - 1);
			} else{
				char tmp_format[32] = {0};
				snprintf(tmp_format, sizeof(tmp_format)-1, "%%*[^_]_%%%lu[\\\\+0-9a-zA-Z]", sizeof(tmp_format)-1);
				sscanf(cdr_copy->dst, tmp_format, tmp_dst);
			}
			if (0 != strlen(tmp_dst) && '\0' != tmp_dst[0]){
				ast_copy_string(cdr_copy->dst, tmp_dst, sizeof(cdr_copy->dst));
				if (strcmp(cdr_copy->action_type, "PARKEDCALL") && !strcmp(cdr_copy->dcontext, "park-dial") && 0 == strcmp(it_cdr->appl, "VoiceMail")){
					snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type),"VM");
				}else if (strcmp(cdr_copy->action_type, "PARKEDCALL") && !strcmp(cdr_copy->dcontext, "park-dial")){
					ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
				}
			}
		}

		/* bug 150309 */
		if (strlen(cdr_copy->record_files) > 0) {
			if (0 == strcmp(cdr_copy->special_action_type, "ringboth")
				|| (0 == strcmp(cdr_copy->action_type, "DIAL") /* bug 152731 */
						&& !strstr(cdr_copy->special_action_type, "FOLLOWME")    /* bug 153497 */
						&& !strstr(cdr_copy->special_action_type, "local_disa")  /* bug 160028 */
						&& !strstr(cdr_copy->special_action_type, "callpark")   /* bug 149053 */
						&& !strstr(cdr_copy->special_action_type, "COUNTRY_CODE/VQ_CALLBACK")  /* bug 182487 */
						&& !strstr(cdr_copy->special_action_type, "APP_PAGING")
						&& !strstr(cdr_copy->special_action_type, "a_xfer_complete")
						&& !strstr(it_cdr->special_action_type, "CALL_FLIP_TRANSFER")//added by zyxia for callflip
						&& !(strstr(cdr_copy->channel, "/trunk") && strstr(cdr_copy->dstchannel, "/trunk")))  /* bug 168792 */
				|| (strstr(cdr_copy->action_type, "RINGGROUP") && !strcmp(cdr_copy->lastapp, "ForkCDR"))
				|| (strstr(cdr_copy->action_type, "RINGGROUP") && strstr(cdr_copy->dstchannel, "Local/")
					&& strstr(cdr_copy->dstchannel, "from-internal-xfer") && strstr(cdr_copy->dstchannel, ";1"))
				|| (!strcmp(cdr_copy->action_type, "PARKEDCALL") && ast_strlen_zero(cdr_copy->dstchannel))
				|| (strstr(cdr_copy->action_type, "IVR["))
				|| (0 == strcmp(cdr_copy->action_type, "SCA") && 0 == strcmp(cdr_copy->special_action_type, "SCA"))) { /* Bug 183024 */
				cdr_check_record_file(cdr_copy);
			}
		}

		/* bug 182126*/
		if (strcasestr(cdr_copy->channel, "Local") && strcasestr(cdr_copy->channel, ";2") && 0 != strcmp(it_cdr->special_action_type, "CCM_INV")
			&& 0 == strncmp(cdr_copy->action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING"))
			&& 0 == strcmp(cdr_copy->dcontext, "conf-tools")
			&& (0 == strcmp(cdr_copy->dst, "exit-gate") || (0 == strcmp(cdr_copy->dst, cdr_copy->src) && strstr(cdr_copy->action_type, cdr_copy->dst)))){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*ccm inv answer and press 1 or 2*/
		if (NULL == strcasestr(cdr_copy->channel, "Local") && 0 == strncmp(it_cdr->special_action_type, "CCM_INV", strlen("CCM_INV"))){
			if (ast_strlen_zero(cdr_copy->session) && !ast_strlen_zero(cdr_copy->uniqueid)){
				ast_copy_string(cdr_copy->dst, cdr_copy->src, sizeof(cdr_copy->dst));
				sscanf(it_cdr->special_action_type, "%*[^[][%127[^]]", tmp_ext);
				ast_copy_string(cdr_copy->src, tmp_ext, sizeof(cdr_copy->src));
				memset(tmp_ext, 0, sizeof(tmp_ext));
				strncpy(tmp_ext,cdr_copy->uniqueid,10);
				snprintf(cdr_copy->session, sizeof(cdr_copy->session), "%s00000-%s", tmp_ext, cdr_copy->src);
			}else{
				ast_copy_string(tmp_ext, cdr_copy->src, sizeof(tmp_ext));
				ast_copy_string(cdr_copy->src, cdr_copy->dst, sizeof(cdr_copy->src));
				ast_copy_string(cdr_copy->dst, tmp_ext, sizeof(cdr_copy->dst));
			}
			memset(tmp_ext, 0, sizeof(tmp_ext));
			ast_copy_string(cdr_copy->action_owner, cdr_copy->src, sizeof(cdr_copy->action_owner));
			int sp_len = strlen(cdr_copy->src)+strlen("CCM_INV[]");
			if (strlen(it_cdr->special_action_type)>sp_len){
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", it_cdr->special_action_type+sp_len, cdr_copy->src);
			}
			else if (!ast_strlen_zero(party_a->caller_name)){
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", party_a->caller_name, cdr_copy->src);
			}else{
				snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"Conference invitation\" <%s>", cdr_copy->src);
			}
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type), "MULTIMEDIAMEETING[%s]", cdr_copy->src);
		}

		/*modify for all invalid pbx run steps*/
		if(strcasecmp(cdr_copy->dst,"T") == 0 || strcasecmp(cdr_copy->dst,"normalrel") == 0){
			ast_log(LOG_VERBOSE,"set cdr disable");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/*bug189937 added by zyxia 2021.3.18*/
		if (!strcmp(cdr_copy->src ,"s") && strcasestr(it_cdr->special_action_type, "Conference invitation") && AST_CDR_ANSWERED == cdr_copy->disposition){
			snprintf(tmp_format, sizeof(tmp_format)-1, "%%*d-%%%lus", sizeof(tmp_ext)-1);
			sscanf(cdr_copy->session, tmp_format, tmp_ext);
			ast_copy_string(cdr_copy->src, tmp_ext, sizeof(cdr_copy->src));
			ast_copy_string(cdr_copy->action_owner, cdr_copy->src, sizeof(cdr_copy->action_owner));
			ast_log(LOG_VERBOSE, "Rewrite cdr src: %s-->%s for Conference invitation!","s",cdr_copy->src);
		}

		if (ast_strlen_zero(cdr_copy->new_src)){
			ast_copy_string(cdr_copy->new_src, cdr_copy->src, sizeof(cdr_copy->new_src));
		}

		/*added by zyxia for Bug 210837 ,2021.12.20*/
		if (!strcmp(cdr_copy->src ,cdr_copy->dst) && cdr_copy->duration == 0 && cdr_copy->billsec == 0){
			ast_log(LOG_VERBOSE, "Cdr dst is src, and duration is 0,set cdr disable !");
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		/* modified by zgmo for Bug179879 */
		/* for cloud upload, nat data should be same as channel data*/
		if (!ast_strlen_zero(it_cdr->special_action_type) &&
			!strcmp(it_cdr->special_action_type, "\"SEAMLESS_TRANSFER\"")) {
			cdr_copy->is_nat_call = it_cdr->peer_is_nat_call;
			cdr_copy->peer_is_nat_call = it_cdr->is_nat_call;
			if (ast_strlen_zero(cdr_copy->dstchannel)){
				ast_copy_string(cdr_copy->dstchannel, cdr_copy->channel, sizeof(cdr_copy->dstchannel));
				snprintf(cdr_copy->channel+strlen(cdr_copy->channel), sizeof(cdr_copy->channel)-strlen(cdr_copy->channel), "-seam");
				if ((party_a->connected_number && 0 == strcmp(cdr_copy->src, party_a->connected_number))
					|| (party_a->caller_number && 0 == strcmp(cdr_copy->dst, party_a->caller_number)))
				{
					/*this channel should swap the caller and callee data!*/
					snprintf(cdr_copy->clid, sizeof(cdr_copy->clid), "\"%s\" <%s>", party_a->connected_name, cdr_copy->src);
				}
			}
			ast_log(LOG_VERBOSE, "Swap NAT flag!");
		} else if (!ast_strlen_zero(it_cdr->special_action_type)
			&& !strcmp(it_cdr->special_action_type, "issfu|Conference invitation")
			&& ast_strlen_zero(cdr_copy->special_action_type)) {
			cdr_copy->is_nat_call = it_cdr->peer_is_nat_call;
			cdr_copy->peer_is_nat_call = it_cdr->is_nat_call;
			if (ast_strlen_zero(cdr_copy->dstchannel)){
				ast_copy_string(cdr_copy->dstchannel, cdr_copy->channel, sizeof(cdr_copy->dstchannel));
				snprintf(cdr_copy->channel+strlen(cdr_copy->channel), sizeof(cdr_copy->channel)-strlen(cdr_copy->channel), "-sfuinv");
			}
			ast_log(LOG_VERBOSE, "Swap NAT flag!");
		} else if (!ast_strlen_zero(it_cdr->special_action_type)
			&& strstr(it_cdr->special_action_type, "Scheduled Conference|")
			&& ast_strlen_zero(cdr_copy->special_action_type)) {
			cdr_copy->is_nat_call = it_cdr->peer_is_nat_call;
			cdr_copy->peer_is_nat_call = it_cdr->is_nat_call;
			if (ast_strlen_zero(cdr_copy->dstchannel)){
				ast_copy_string(cdr_copy->dstchannel, cdr_copy->channel, sizeof(cdr_copy->dstchannel));
				snprintf(cdr_copy->channel+strlen(cdr_copy->channel), sizeof(cdr_copy->channel)-strlen(cdr_copy->channel), "-sfuinv");
			}
			ast_log(LOG_VERBOSE, "Swap NAT flag!");
		} else {
			if (!atx_keep_nat_flag){
				cdr_copy->is_nat_call = it_cdr->is_nat_call;
			}
			cdr_copy->peer_is_nat_call = it_cdr->peer_is_nat_call;
			if (!cdr_copy->peer_is_nat_call && party_a->peer_is_nat_call
				&& strstr(cdr_copy->dstchannel, "Local") && strstr(cdr_copy->dstchannel, ";1")){
				cdr_copy->peer_is_nat_call = party_a->peer_is_nat_call;
			}
		}
		if (NULL != strstr(cdr_copy->channel, "Local")
			&& (NULL == strstr(it_cdr->special_action_type, "Conference") && NULL == strstr(it_cdr->special_action_type, "SEAMLESS_TRANSFER"))){
			if ((NULL != strstr(it_cdr->special_action_type, "overwrt_record") && NULL != strstr(cdr_copy->action_type,"TRANSFER"))
				|| (!ast_strlen_zero(cdr_copy->inbound_trunk_index) && NULL != strstr(it_cdr->special_action_type, "local_disa"))) {
				/*Keep NAT flag in cdr*/
			}else{
				cdr_copy->is_nat_call = -1;
			}
		}
		if (NULL != strstr(cdr_copy->dstchannel, "Local")
			&& (NULL == strstr(it_cdr->special_action_type, "Conference")
				&& NULL == strstr(it_cdr->special_action_type, "SEAMLESS_TRANSFER")
				&& NULL == strstr(it_cdr->special_action_type, "FOLLOWME"))){
			cdr_copy->peer_is_nat_call = -1;
		}

		cdr_copy->is_video_call = it_cdr->is_video_call;
		cdr_copy->is_gds_call = it_cdr->is_gds_call;
		cdr_copy->is_gsc_call = it_cdr->is_gsc_call;
		cdr_copy->is_anonymous_call = it_cdr->party_a.snapshot->is_anonymous;

		/*for 63xxA, this condition will delete in v1.0.10.x*/
		if (!ast_strlen_zero(device_name) && strlen(device_name) >= 7 && device_name[7]=='A' && strstr(cdr_copy->action_type, "CONFERENCE")){
			snprintf(cdr_copy->action_type, sizeof(cdr_copy->action_type), "MEETING%s", strstr(cdr_copy->action_type, "CONFERENCE")+strlen("CONFERENCE"));
		}

		if (!ast_strlen_zero(sla_chan) && 0 ==strcmp(cdr_copy->dst, "s") && ast_strlen_zero(cdr_copy->dstchannel)){
			char tmp_num[128] = {0};
			sscanf(sla_chan, "%*[^/]/%127[^-]-%*s", tmp_num);
			if (!ast_strlen_zero(tmp_num)){
				ast_copy_string(cdr_copy->dst, tmp_num, sizeof(cdr_copy->dst));
			}
		}

		/*bug 196077, first cdr should remain for callee do not answer*/
		if ((it_cdr != cdr || NULL != it_cdr->next) && NULL != strstr(cdr_copy->dcontext, "sla-did") && 0 == strcmp(cdr_copy->lastapp, "SLATrunk") && !ast_strlen_zero(cdr_copy->dstchannel)){
			ast_log(LOG_VERBOSE,"set cdr disable");
			snprintf(sla_chan, sizeof(sla_chan), "%s", cdr_copy->dstchannel);
			ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
		}

		if (!ast_strlen_zero(it_cdr->special_action_type) && ast_strlen_zero(cdr_copy->dstchannel)
			&& 0 == strcmp(cdr_copy->dst, "i") && 0 == strcmp(cdr_copy->action_type, "PARKEDCALL")){
			ast_copy_string(cdr_copy->dst, it_cdr->special_action_type, sizeof(cdr_copy->dst));
			cdr_copy->disposition = AST_CDR_FAILED;
		}

		if(0 != strcmp(cdr_copy->action_type, "DIAL")){
			char action_num[256] = {0};
			snprintf(tmp_format, sizeof(tmp_format), "%%*[^[]\[%%%lu[^]]", sizeof(action_num)-1);
			sscanf(cdr_copy->action_type, tmp_format, action_num);
			ast_log(LOG_NOTICE, "Need reset the data? see action:%s callee:%s caller:%s", cdr_copy->action_type, cdr_copy->dst, cdr_copy->src);
			if ((!ast_strlen_zero(action_num) && 0 == strcmp(cdr_copy->dst, action_num))
				|| 0 == strcmp(cdr_copy->dst, cdr_copy->action_type)){
				cdr_copy->peer_is_nat_call = -1;
			}
			if ((!ast_strlen_zero(action_num) && 0 == strcmp(cdr_copy->src, action_num))
				|| 0 == strcmp(cdr_copy->action_type, "EMERGENCYNOTIFY")){
				cdr_copy->is_nat_call = -1;
			}
		}

		/*bug 203208*/
		if (ast_strlen_zero(cdr_copy->dstchannel) && 0 == strcasecmp(cdr_copy->lastapp, "Park")){
			ast_copy_string(cdr_copy->action_type, "PARKEDCALL", sizeof(cdr_copy->action_type));
		}
		/*bug 203213*/
		if((NULL != strcasestr(cdr_copy->clid,"Call Bargin") || NULL != strcasestr(cdr_copy->clid,"Call Whisper") || NULL != strcasestr(cdr_copy->clid,"Call Listen"))
			&& (NULL != strcasestr(cdr_copy->channel, "Local/") && NULL != strcasestr(cdr_copy->channel, ";2"))){
			cdr_copy->is_nat_call = -1;
		}

		if(0 == strcasecmp(it_cdr->action_type, "ANNOUNCEMENT PAGE")
			&& 0 == strcasecmp(cdr_copy->src, "Page")
			&& (NULL != strcasestr(cdr_copy->channel, "Local/") && NULL != strcasestr(cdr_copy->channel, ";2"))){
			cdr_copy->is_nat_call = -1;
		}

		if (NULL != strcasestr(cdr_copy->lastdata, "conf-tools,conf-inv")
			&& 0 == strncmp(cdr_copy->action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING"))
			&& !ast_strlen_zero(cdr_copy->dstchannel)){
			char inv_conf[256] = {0};
			sscanf(cdr_copy->action_type, "%*[^[]\[%[^]]", inv_conf);
			if (0 != strcmp(cdr_copy->src, inv_conf)){
				ast_log(LOG_VERBOSE,"set cdr disable");
				ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
			}
		}

		/*a probabilistic issue, the dst channel is come from the same out trunk*/
		if (cdr_copy->channel && NULL != strcasestr(cdr_copy->channel, "trunk_")
			&& cdr_copy->dstchannel && NULL != strcasestr(cdr_copy->dstchannel, "trunk_")
			&& 0 == strncmp(cdr_copy->action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING"))){
			int chan_cmp_len = strlen(cdr_copy->channel) - strlen("00000000");
			if (ast_strlen_zero(cdr_copy->outbound_trunk_index) && 0 == strncmp(cdr_copy->channel, cdr_copy->dstchannel, chan_cmp_len)){
				ast_log(LOG_DEBUG, "this is a faker dst channel, we should clean it!");
				memset(cdr_copy->dstchannel, 0, sizeof(cdr_copy->dstchannel));
			}
		}

		/*only for special sca*/
		if (0 == strcmp(cdr_copy->special_action_type, "SPECIAL_SCA") && 0 == strcmp(cdr_copy->lastapp, "Dial")){
			sscanf(cdr_copy->lastdata, "%*[^/]/%127[^/]%*s", tmp_ext);
			if (!ast_strlen_zero(tmp_ext)){
				if (0 != strcmp(tmp_ext, cdr_copy->dst)){
					/*private account cdr process*/
					if (0 == strcmp(cdr_copy->reason, ANSWERED_ELSEWHERE)){
						ast_copy_string(cdr_copy->dst, tmp_ext, sizeof(cdr_copy->dst));
					}else{
						/*disable the not answered private account*/
						ast_log(LOG_VERBOSE,"set cdr disable");
						ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
					}
				}else if (0 != strcmp(cdr_copy->reason, ANSWERED_ELSEWHERE)){
					/*if no one anwered, should show the sca number selves cdr*/
					ast_log(LOG_VERBOSE, "clear cdr disable flag");
					ast_clear_flag(cdr_copy, AST_CDR_FLAG_DISABLE|AST_CDR_FLAG_DISABLE_ALL);
				}
			}

			/*enable the right duration cdr*/
			if (0 == strcmp(cdr_copy->reason, ANSWERED_ELSEWHERE)){
				if (cdr_copy->disposition == AST_CDR_ANSWERED){
					ast_log(LOG_VERBOSE, "clear cdr disable flag");
					ast_clear_flag(cdr_copy, AST_CDR_FLAG_DISABLE|AST_CDR_FLAG_DISABLE_ALL);
				}
				else{
					ast_log(LOG_VERBOSE,"set cdr disable");
					ast_set_flag(cdr_copy,AST_CDR_FLAG_DISABLE);
				}
			}
			ast_copy_string(cdr_copy->action_type, "SCA", sizeof(cdr_copy->dst));
			memset(cdr_copy->reason, 0, sizeof(cdr_copy->reason));
		}

		/*add for reason with channel*/
		if (NULL != strcasestr(cdr_copy->reason, ANSWERED_ELSEWHERE)){
			if (NULL != strstr(cdr_copy->action_type, "MULTIMEDIAMEETING")){
				snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), NOT_SEND_TO_WAVE);
			}else{
				snprintf(cdr_copy->reason, sizeof(cdr_copy->reason), ANSWERED_ELSEWHERE_NEW);
			}
		}

		/*add device info for CDR*/
		if (ast_strlen_zero(cdr_copy->device_info) && !ast_strlen_zero(party_a->device_info)){
			ast_copy_string(cdr_copy->device_info, party_a->device_info, sizeof(cdr_copy->device_info));
		}
		if (ast_strlen_zero(cdr_copy->device_info_peer) && party_b && !ast_strlen_zero(party_b->device_info)){
			ast_copy_string(cdr_copy->device_info_peer, party_b->device_info, sizeof(cdr_copy->device_info_peer));
		}
		if (strcasestr(cdr_copy->channel, "IAX")){
			ast_copy_string(cdr_copy->device_info, "IAX", sizeof(cdr_copy->device_info));
		}
		if (strcasestr(cdr_copy->dstchannel, "IAX")){
			ast_copy_string(cdr_copy->device_info_peer, "IAX", sizeof(cdr_copy->device_info_peer));
		}
		if (strcasestr(cdr_copy->channel, "DAHDI")){
			sscanf(cdr_copy->channel, "%79[^-]", cdr_copy->device_info);
		}
		if (strcasestr(cdr_copy->dstchannel, "DAHDI")){
			sscanf(cdr_copy->dstchannel, "%79[^-]", cdr_copy->device_info_peer);
		}

		if (ast_strlen_zero(cdr_copy->session)
			&& !strcmp(it_cdr->special_action_type, "a_xfer_resume")) {
			struct timeval now = ast_tvnow();
			snprintf(cdr_copy->session, sizeof(cdr_copy->session),
				"%d%d-%s", (int)now.tv_sec, (int)now.tv_usec, cdr_copy->src);
			ast_log(LOG_DEBUG, "In this case, generate session");
		}

		if (!ast_strlen_zero(cdr_copy->record_files)
			&& !ast_strlen_zero(it_cdr->record_files_desc)) {
			snprintf(cdr_copy->record_files_desc, sizeof(cdr_copy->record_files_desc)-1,
				"%s", it_cdr->record_files_desc);
		}
		
		
		ast_log(LOG_DEBUG, "[cdr_object_create_public_records] cdr_copy->record_files:%s cdr_copy->record_files_uuid:%s", 
										cdr_copy->record_files, cdr_copy->record_files_uuid);
		ast_log(LOG_DEBUG, "[cdr_object_create_public_records] it_cdr->record_files:%s it_cdr->record_files_uuid:%s", 
										it_cdr->record_files, it_cdr->record_files_uuid);
		if (!ast_strlen_zero(cdr_copy->record_files) /*&&
			!ast_strlen_zero(it_cdr->record_files_uuid)*/) {
				snprintf(cdr_copy->record_files_uuid, sizeof(cdr_copy->record_files_uuid),
					"%s", it_cdr->record_files_uuid);               
		}

		if (ast_strlen_zero(cdr_copy->record_files_desc)
			&& !ast_strlen_zero(cdr_copy->record_files)) {
			ast_log(LOG_VERBOSE, "Detect record_files_desc empty having record_files");

			if (strstr(cdr_copy->action_type, "IVR") && AST_CDR_ANSWERED == cdr_copy->disposition && !strstr(cdr_copy->src, "Callback")) {
				cdr_copy->record_files[0] = '\0';
				ast_log(LOG_DEBUG, "cdr[%p] clear original var:record_file[%s]",
									it_cdr, cdr_copy->record_files);
			} else {
				snprintf(cdr_copy->record_files_desc, sizeof(cdr_copy->record_files_desc)-1,
					"duration=0,size=0B,start-time=0@");
			}
		}
		/*Added by gangwang for Bug275685*/
		if (!ast_strlen_zero(cdr_copy->dst) && 0 == strcasecmp(cdr_copy->dst, "s") && !ast_strlen_zero(it_cdr->appl) && 0 == strcasecmp(it_cdr->appl, "ParkedCall")) {
			if (it_cdr->special_dst && !ast_strlen_zero(it_cdr->special_dst)) {
				ast_copy_string(cdr_copy->dst, it_cdr->special_dst, sizeof(cdr_copy->dst));
				ast_log(LOG_DEBUG, "ParkedCall copy %s to cdr_copy->dst", it_cdr->special_dst);
			}
		}
		/*End gangwang*/

		//set the main cdr's data to last ast_cdr
		if( !ast_test_flag(cdr_copy, AST_CDR_FLAG_DISABLE)){
			tmp_main_duration += cdr_copy->duration;
			tmp_main_billsec += cdr_copy->billsec;
			if (0 == first_enable_cdr){
				cdr_copy->main_disposition = cdr_copy->disposition;
				tmp_start = cdr_copy->start;
				tmp_answer = cdr_copy->answer;
				tmp_end = cdr_copy->end;
			}
			if (AST_CDR_ANSWERED == cdr_copy->disposition){
				cdr_copy->main_disposition = AST_CDR_ANSWERED;
			}
			//get the min start
			if (tmp_start.tv_sec > cdr_copy->start.tv_sec
				|| (tmp_start.tv_sec == cdr_copy->start.tv_sec && tmp_start.tv_usec > cdr_copy->start.tv_usec)){
				tmp_start = cdr_copy->start;
			}
			//get the min answer
			if (tmp_answer.tv_sec > cdr_copy->answer.tv_sec
				|| (tmp_answer.tv_sec == cdr_copy->answer.tv_sec && tmp_answer.tv_usec > cdr_copy->answer.tv_usec)){
				tmp_answer = cdr_copy->answer;
			}
			//get the max end
			if (tmp_end.tv_sec < cdr_copy->end.tv_sec
				|| (tmp_end.tv_sec == cdr_copy->end.tv_sec && tmp_end.tv_usec < cdr_copy->end.tv_usec)){
				tmp_end = cdr_copy->end;
			}
			//calculate the main duration and billsec
			cdr_copy->main_duration = (long)ast_tvdiff_ms(ast_tvzero(tmp_end) ? ast_tvnow() : tmp_end, tmp_start) / 1000;
			cdr_copy->main_billsec =  (long)ast_tvdiff_ms(ast_tvzero(tmp_end) ? ast_tvnow() : tmp_end, tmp_answer);
			if (cdr_copy->main_billsec % 1000 >= 500
				&& is_cdr_flag_set(CDR_INITIATED_SECONDS)) {
				cdr_copy->main_billsec = (cdr_copy->main_billsec / 1000) + 1;
			} else {
				cdr_copy->main_billsec = cdr_copy->main_billsec / 1000;
			}

			if (tmp_main_billsec < cdr_copy->main_billsec){
				cdr_copy->main_billsec = tmp_main_billsec;
			}
			if (tmp_main_duration < cdr_copy->main_duration){
				cdr_copy->main_duration = tmp_main_duration;
			}
			first_enable_cdr++;
		}

		ast_log(LOG_VERBOSE, 
			"cdrhdl:%p, src:%s, dst:%s, disp:%ld, "
			"action_type:%s, special_action_type:%s",
			it_cdr, cdr_copy->src, cdr_copy->dst, cdr_copy->disposition,
			cdr_copy->action_type, it_cdr->special_action_type);

		ast_log(LOG_VERBOSE, "channel:%s, dstchannel:%s, session:%s, lastapp:%s,owner:%s",
			cdr_copy->channel, cdr_copy->dstchannel, cdr_copy->session,
			cdr_copy->lastapp, cdr_copy->action_owner);

		ast_log(LOG_VERBOSE, "record_files:%s", cdr_copy->record_files);
		ast_log(LOG_VERBOSE, "record_files_desc:%s", cdr_copy->record_files_desc);
		ast_log(LOG_VERBOSE, "[cdr_object_create_public_records] record_files_uuid:%s", cdr_copy->record_files_uuid);

#ifdef BUILD_ON_CLOUD_PBX
		cdr_copy->peer_is_nat_call = 0;
		cdr_copy->is_nat_call = 0;
#else
		ast_log(LOG_VERBOSE, "is_nat_call:%s, peer_is_nat_call:%s",
			1 == cdr_copy->is_nat_call ? "yes" : "no", 1 == cdr_copy->peer_is_nat_call ? "yes" : "no");
#endif
		ast_log(LOG_VERBOSE, "is_video_call:%s", cdr_copy->is_video_call ? "yes" : "no");
		ast_log(LOG_VERBOSE, "status:%s, disable:%d, reason:%s", cdr_copy->status,
			ast_test_flag(cdr_copy, AST_CDR_FLAG_DISABLE) ? 1 : 0, cdr_copy->reason);
		ast_log(LOG_NOTICE, "is_gds_call: %s", cdr_copy->is_gds_call ? "yes" : "no");
		ast_log(LOG_NOTICE, "is_gsc_call: %s", cdr_copy->is_gsc_call ? "yes" : "no");
		ast_log(LOG_NOTICE, "start:%ld, answer:%ld, end:%ld",
			cdr_copy->start.tv_sec, cdr_copy->answer.tv_sec, cdr_copy->end.tv_sec);
		ast_log(LOG_VERBOSE, "=======================================================");

		/* Add by bfdu for nat call limit, 2022.3.8 */
		cdr_set_limit_used_time(cdr_copy);
#endif

		/* Variables */
		copy_variables(&cdr_copy->varshead, &it_cdr->party_a.variables);
		AST_LIST_TRAVERSE(&it_cdr->party_b.variables, it_var, entries) {
			int found = 0;
			struct ast_var_t *newvariable;
			AST_LIST_TRAVERSE(&cdr_copy->varshead, it_copy_var, entries) {
				if (!strcasecmp(ast_var_name(it_var), ast_var_name(it_copy_var))) {
					found = 1;
					break;
				}
			}
			if (!found && (newvariable = ast_var_assign(ast_var_name(it_var), ast_var_value(it_var)))) {
				AST_LIST_INSERT_TAIL(&cdr_copy->varshead, newvariable, entries);
			}
		}

#ifdef GRANDSTREAM_NETWORKS
		/*special process for ringboth to external extension*/
		if (1 == first_enable_cdr && NULL == tmp_cdr_copy && it_cdr->next
			&& AST_CDR_ANSWERED != cdr_copy->disposition && !ast_strlen_zero(cdr_copy->outbound_trunk_index)
			&& NULL != it_cdr->special_action_type && (NULL != strstr(it_cdr->special_action_type, "ringboth") && NULL != strstr(it_cdr->special_action_type, "COUNTRY_CODE"))
			&& NULL != cdr_copy->dstchannel && (NULL != strstr(cdr_copy->dstchannel, "/trunk_") || NULL != strstr(cdr_copy->dstchannel, "DAHDI/o"))){
			tmp_cdr_copy = cdr_copy;
			continue;
		}
#endif

		if (!pub_cdr) {
			pub_cdr = cdr_copy;
			cdr_prev = pub_cdr;
		} else {
			cdr_prev->next = cdr_copy;
			cdr_prev = cdr_copy;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (tmp_cdr_copy){
			if (!pub_cdr) {
				pub_cdr = tmp_cdr_copy;
				cdr_prev = pub_cdr;
			} else {
				cdr_prev->next = tmp_cdr_copy;
				cdr_prev = tmp_cdr_copy;
			}
			tmp_cdr_copy = NULL;
		}
		pre_it_cdr = it_cdr;
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "Exit cdr_object_create_public_records()");
#endif
	return pub_cdr;
}

/*!
 * \brief Dispatch a CDR.
 * \param cdr The \ref cdr_object to dispatch
 *
 * This will create a \ref ast_cdr object and publish it to the various backends
 */
static void cdr_object_dispatch(struct cdr_object *cdr)
{
	struct ast_cdr *pub_cdr;

	CDR_DEBUG("%p - Dispatching CDR for Party A %s, Party B %s\n", cdr,
		cdr->party_a.snapshot->name,
		cdr->party_b.snapshot ? cdr->party_b.snapshot->name : "<none>");
#ifdef GRANDSTREAM_NETWORKS
	if (!ast_cdr_get_write_database_enable()) {
		ast_log(LOG_WARNING, "DiskSpace Protect, Forbidden cdr record write to Database!");
		return;
	}
#endif

	pub_cdr = cdr_object_create_public_records(cdr);
	cdr_detach(pub_cdr);
}

/*!
 * \brief Set the disposition on a \ref cdr_object based on a hangupcause code
 * \param cdr The \ref cdr_object
 * \param hangupcause The Asterisk hangup cause code
 */
static void cdr_object_set_disposition(struct cdr_object *cdr, int hangupcause)
{
	/* Change the disposition based on the hang up cause */
	switch (hangupcause) {
	case AST_CAUSE_BUSY:
#ifdef GRANDSTREAM_NETWORKS
	case AST_CAUSE_CALL_REJECTED:
#endif
		cdr->disposition = AST_CDR_BUSY;
		break;
	case AST_CAUSE_CONGESTION:
		if (!is_cdr_flag_set(CDR_CONGESTION)) {
			cdr->disposition = AST_CDR_FAILED;
		} else {
			cdr->disposition = AST_CDR_CONGESTION;
		}
		break;
	case AST_CAUSE_NO_ROUTE_DESTINATION:
	case AST_CAUSE_UNREGISTERED:
		cdr->disposition = AST_CDR_FAILED;
		break;
	case AST_CAUSE_NORMAL_CLEARING:
	case AST_CAUSE_NO_ANSWER:
		cdr->disposition = AST_CDR_NOANSWER;
		break;
	default:
		break;
	}
}

/*!
 * \brief Finalize a CDR.
 *
 * This function is safe to call multiple times. Note that you can call this
 * explicitly before going to the finalized state if there's a chance the CDR
 * will be re-activated, in which case the \ref cdr_object's end time should be
 * cleared. This function is implicitly called when a CDR transitions to the
 * finalized state and right before it is dispatched
 *
 * \param cdr_object The CDR to finalize
 */
#ifdef GRANDSTREAM_NETWORKS
static void cdr_object_finalize(struct cdr_object *cdr)
{
	if (!ast_tvzero(cdr->end)) {
		return;
	}
	cdr->end = ast_tvnow();

	if (cdr->disposition == AST_CDR_NULL) {
		/* add for bug 60257 - 60259, add by pchwu /110127*/
		int64_t ms;
		int64_t sec_dif = (int64_t)(cdr->answer.tv_sec - cdr->start.tv_sec)*1000000;
		int64_t usec_dif = cdr->answer.tv_usec - cdr->start.tv_usec;
		ms = sec_dif + usec_dif;
		if ((!ast_tvzero(cdr->answer) && strncmp(cdr->action_type, "QUEUE", strlen("QUEUE")) && strncmp(cdr->action_type, "RINGGROUP", strlen("RINGGROUP")))
			|| ms > 0) {
			cdr->disposition = AST_CDR_ANSWERED;
			ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
		} else if (cdr->party_a.snapshot->hangupcause) {
			/*pick up the ringgroup member bug 154848*/
			if (NULL != cdr->action_type && !strncmp(cdr->action_type, "RINGGROUP", strlen("RINGGROUP")) && !ast_tvzero(cdr->answer)
				&& cdr->party_b.snapshot && cdr->party_b.snapshot->hangupcause && AST_CAUSE_NORMAL_CLEARING == cdr->party_b.snapshot->hangupcause){
				cdr->disposition = AST_CDR_ANSWERED;
				ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
			}
			else{
				cdr_object_set_disposition(cdr, cdr->party_a.snapshot->hangupcause);
				ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
			}
		} else if (cdr->party_b.snapshot && cdr->party_b.snapshot->hangupcause) {
			if (NULL != cdr->action_type && !strncmp(cdr->action_type, "RINGGROUP", strlen("RINGGROUP"))
				&& !ast_tvzero(cdr->answer) && AST_CAUSE_NORMAL_CLEARING == cdr->party_b.snapshot->hangupcause){
				cdr->disposition = AST_CDR_ANSWERED;
				ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
			}
			else{
				cdr_object_set_disposition(cdr, cdr->party_b.snapshot->hangupcause);
				ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
			}
		} else {
			cdr->disposition = AST_CDR_FAILED;
			ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
		}
	}

	if (cdr->disposition == AST_CDR_NOANSWER && cdr->party_b.snapshot && cdr->party_b.snapshot->hangupcause){
		cdr_object_set_disposition(cdr, cdr->party_b.snapshot->hangupcause);
		ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);

		if (AST_CAUSE_CONF_SKIP_CALL_WAVE == cdr->party_b.snapshot->hangupcause) {
			ast_string_field_set(cdr, special_action_type, "WAVE_SKIP_CALL");
			ast_log(LOG_DEBUG, "cdr(%p) special_action_type ==> 'WAVE_SKIP_CALL'", cdr);
		}
	}

	//edit by zyxia for Bug 267696,AST_CDR_FAILED
	if ((AST_CDR_NULL == cdr->disposition || AST_CDR_FAILED == cdr->disposition ) && !cdr->party_b.snapshot
		&& AST_CAUSE_CONF_SKIP_CALL_WAVE == cdr->party_a.snapshot->hangupcause) {
		ast_string_field_set(cdr, special_action_type, "WAVE_SKIP_CALL");
		ast_log(LOG_DEBUG, "cdr(%p) special_action_type ==> 'WAVE_SKIP_CALL'", cdr);
	}

	if (cdr->party_b.snapshot
		&& cdr->party_b.snapshot->hangupcause
		&& AST_CAUSE_ANSWERED_ELSEWHERE == cdr->party_b.snapshot->hangupcause){
		cdr->disposition = AST_CDR_NOANSWER;
		ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
	}

	if (0 == strncmp(cdr->action_type, "MULTIMEDIAMEETING", strlen("MULTIMEDIAMEETING"))
		&& AST_CDR_NOANSWER == cdr->disposition
		&& AST_CAUSE_CALL_REJECTED == cdr->party_a.snapshot->hangupcause){
			cdr->disposition = AST_CDR_BUSY;
			ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
	}

	if (cdr->party_a.snapshot && AST_CAUSE_ANSWERED_ELSEWHERE == cdr->party_a.snapshot->hangupcause
		&& ast_strlen_zero(cdr->reason)) {
		snprintf(cdr->reason, sizeof(cdr->reason), ANSWERED_ELSEWHERE);
		ast_log(LOG_DEBUG, "mark reason for 'answer-elsewhere'");
	}

	if (cdr->party_a.snapshot && !ast_strlen_zero(cdr->party_a.snapshot->reject_reason)
		&& ast_strlen_zero(cdr->reason)) {
		snprintf(cdr->reason, sizeof(cdr->reason), cdr->party_a.snapshot->reject_reason);
		ast_log(LOG_DEBUG, "party_a mark reason for '%s'",cdr->reason);
	} else if (cdr->party_b.snapshot && !ast_strlen_zero(cdr->party_b.snapshot->reject_reason)
		&& ast_strlen_zero(cdr->reason)) {
		snprintf(cdr->reason, sizeof(cdr->reason), cdr->party_b.snapshot->reject_reason);
		ast_log(LOG_DEBUG, "party_b mark reason for '%s'",cdr->reason);
	}

	/* tv_usec is suseconds_t, which could be int or long */
	ast_debug(1, "Finalized CDR(%p) for %s - session[%s] start %ld.%06ld answer %ld.%06ld end %ld.%06ld dispo %s\n",
			cdr,
			cdr->party_a.snapshot->name,
			cdr->session,
			(long)cdr->start.tv_sec,
			(long)cdr->start.tv_usec,
			(long)cdr->answer.tv_sec,
			(long)cdr->answer.tv_usec,
			(long)cdr->end.tv_sec,
			(long)cdr->end.tv_usec,
			ast_cdr_disp2str(cdr->disposition));
}
#else
static void cdr_object_finalize(struct cdr_object *cdr)
{
	if (!ast_tvzero(cdr->end)) {
		return;
	}
	cdr->end = ast_tvnow();

	if (cdr->disposition == AST_CDR_NULL) {
		if (!ast_tvzero(cdr->answer)) {
			cdr->disposition = AST_CDR_ANSWERED;
		} else if (cdr->party_a.snapshot->hangupcause) {
			cdr_object_set_disposition(cdr, cdr->party_a.snapshot->hangupcause);
		} else if (cdr->party_b.snapshot && cdr->party_b.snapshot->hangupcause) {
			cdr_object_set_disposition(cdr, cdr->party_b.snapshot->hangupcause);
		} else {
			cdr->disposition = AST_CDR_FAILED;
		}
	}

	/* tv_usec is suseconds_t, which could be int or long */
	ast_debug(1, "Finalized CDR for %s - start %ld.%06ld answer %ld.%06ld end %ld.%06ld dispo %s\n",
			cdr->party_a.snapshot->name,
			(long)cdr->start.tv_sec,
			(long)cdr->start.tv_usec,
			(long)cdr->answer.tv_sec,
			(long)cdr->answer.tv_usec,
			(long)cdr->end.tv_sec,
			(long)cdr->end.tv_usec,
			ast_cdr_disp2str(cdr->disposition));
}
#endif

/*!
 * \brief Check to see if a CDR needs to move to the finalized state because
 * its Party A hungup.
 */
static void cdr_object_check_party_a_hangup(struct cdr_object *cdr)
{
	if (ast_test_flag(&cdr->party_a.snapshot->softhangup_flags, AST_SOFTHANGUP_HANGUP_EXEC)
		&& is_cdr_flag_set(CDR_END_BEFORE_H_EXTEN)) {
		cdr_object_finalize(cdr);
	}

	if (ast_test_flag(&cdr->party_a.snapshot->flags, AST_FLAG_DEAD)
		&& cdr->fn_table != &finalized_state_fn_table) {
		cdr_object_transition_state(cdr, &finalized_state_fn_table);
	}
}

/*!
 * \brief Check to see if a CDR needs to be answered based on its Party A.
 * Note that this is safe to call as much as you want - we won't answer twice
 */
static void cdr_object_check_party_a_answer(struct cdr_object *cdr)
{
	if (cdr->party_a.snapshot->state == AST_STATE_UP && ast_tvzero(cdr->answer)) {
		if(!ast_tvzero(cdr->recovery_answer))
		{
			cdr->answer = cdr->recovery_answer;
			cdr->start = cdr->recovery_start;
		}
		else
		{
			cdr->answer = ast_tvnow();
		}
		/* tv_usec is suseconds_t, which could be int or long */
		CDR_DEBUG("%p - Set answered time to %ld.%06ld\n", cdr,
			(long)cdr->answer.tv_sec,
			(long)cdr->answer.tv_usec);
	}
#ifdef GRANDSTREAM_NETWORKS
	if (cdr->party_b.snapshot && cdr->party_b.snapshot->state == AST_STATE_UP
		&& !ast_tvzero(cdr->answer)
		&& !ast_strlen_zero(cdr->special_action_type)
		&& strstr(cdr->special_action_type, "authenticate")){
		cdr->answer = ast_tvnow();
		if (0 == strncmp(cdr->special_action_type, "authenticate", strlen("authenticate"))){
			ast_cdr_setvar(cdr->party_a.snapshot->name,"special_action_type","");
		}else{
			ast_cdr_setvar(cdr->party_a.snapshot->name,"special_action_type","COUNTRY_CODE");
		}
	}
#endif
}

/* \brief Set Caller ID information on a CDR */
static void cdr_object_update_cid(struct cdr_object_snapshot *old_snapshot, struct ast_channel_snapshot *new_snapshot)
{
	if (!old_snapshot->snapshot) {
		set_variable(&old_snapshot->variables, "dnid", new_snapshot->caller_dnid);
		set_variable(&old_snapshot->variables, "callingsubaddr", new_snapshot->caller_subaddr);
		set_variable(&old_snapshot->variables, "calledsubaddr", new_snapshot->dialed_subaddr);
		return;
	}
	if (strcmp(old_snapshot->snapshot->caller_dnid, new_snapshot->caller_dnid)) {
		set_variable(&old_snapshot->variables, "dnid", new_snapshot->caller_dnid);
	}
	if (strcmp(old_snapshot->snapshot->caller_subaddr, new_snapshot->caller_subaddr)) {
		set_variable(&old_snapshot->variables, "callingsubaddr", new_snapshot->caller_subaddr);
	}
	if (strcmp(old_snapshot->snapshot->dialed_subaddr, new_snapshot->dialed_subaddr)) {
		set_variable(&old_snapshot->variables, "calledsubaddr", new_snapshot->dialed_subaddr);
	}
}

/*!
 * \brief Swap an old \ref cdr_object_snapshot's \ref ast_channel_snapshot for
 * a new \ref ast_channel_snapshot
 * \param old_snapshot The old \ref cdr_object_snapshot
 * \param new_snapshot The new \ref ast_channel_snapshot for old_snapshot
 */
#ifdef GRANDSTREAM_NETWORKS
static void cdr_object_swap_snapshot(struct cdr_object *cdr, struct cdr_object_snapshot *old_snapshot,
		struct ast_channel_snapshot *new_snapshot)
#else
static void cdr_object_swap_snapshot(struct cdr_object_snapshot *old_snapshot,
		struct ast_channel_snapshot *new_snapshot)
#endif
{
#ifdef GRANDSTREAM_NETWORKS
	SCOPED_MUTEX(lock, &cdr->party_lock);
#endif
	cdr_object_update_cid(old_snapshot, new_snapshot);
	ao2_t_replace(old_snapshot->snapshot, new_snapshot, "Swap CDR shapshot");
}

/* BASE METHOD IMPLEMENTATIONS */

static int base_process_party_a(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot)
{
	ast_assert(strcasecmp(snapshot->name, cdr->party_a.snapshot->name) == 0);

	/* Finalize the CDR if we're in hangup logic and we're set to do so */
	if (ast_test_flag(&snapshot->softhangup_flags, AST_SOFTHANGUP_HANGUP_EXEC)
		&& is_cdr_flag_set(CDR_END_BEFORE_H_EXTEN)) {
		cdr_object_finalize(cdr);
		return 0;
	}

	/*
	 * Only record the context and extension if we aren't in a subroutine, or if
	 * we are executing hangup logic.
	 */
	if (!ast_test_flag(&snapshot->flags, AST_FLAG_SUBROUTINE_EXEC)
		|| ast_test_flag(&snapshot->softhangup_flags, AST_SOFTHANGUP_HANGUP_EXEC)) {
		if (strcmp(cdr->context, snapshot->context)) {
			ast_string_field_set(cdr, context, snapshot->context);
		}
		if (strcmp(cdr->exten, snapshot->exten)) {
			ast_string_field_set(cdr, exten, snapshot->exten);
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	/*add by wxu*/
	ast_string_field_set(cdr, session, snapshot->session);
	if (ast_tvzero(cdr->recovery_answer))
	{
		if(!ast_tvzero(snapshot->recovery_answertime))
		{
			cdr->recovery_answer = snapshot->recovery_answertime;
			cdr->recovery_start = snapshot->recovery_creationtime;
		}
	}
	/*add end*/
	cdr_object_swap_snapshot(cdr, &cdr->party_a, snapshot);

	/* added by zgmo for new feature of marking nat call */
	if (!cdr->is_nat_call && snapshot->is_nat_call) {
		cdr->is_nat_call = snapshot->is_nat_call;
		ast_log(LOG_DEBUG, "cdr(%p|%s) Detect NAT", cdr, cdr->name);
	} else if (cdr->is_nat_call && !snapshot->is_nat_call) {
		cdr->is_nat_call = snapshot->is_nat_call;
		ast_log(LOG_DEBUG, "cdr(%p|%s) Detect clear NAT", cdr, cdr->name);
	}
	/* end */

	cdr->is_video_call = snapshot->is_video_call;
	/* add end */

	cdr->is_gds_call = snapshot->is_gds_call;
	cdr->is_gsc_call = snapshot->is_gsc_call;
#else
	cdr_object_swap_snapshot(&cdr->party_a, snapshot);
#endif

	/* When Party A is originated to an application and the application exits, the stack
	 * will attempt to clear the application and restore the dummy originate application
	 * of "AppDialX". Prevent that, and any other application changes we might not want
	 * here.
	 */
	if (!ast_test_flag(&cdr->flags, AST_CDR_LOCK_APP)
		&& !ast_strlen_zero(snapshot->appl)
		&& (strncasecmp(snapshot->appl, "appdial", 7) || ast_strlen_zero(cdr->appl))) {
		if (strcmp(cdr->appl, snapshot->appl)) {
			ast_string_field_set(cdr, appl, snapshot->appl);
		}
		if (strcmp(cdr->data, snapshot->data)) {
			ast_string_field_set(cdr, data, snapshot->data);
		}

		/* Dial (app_dial) is a special case. Because pre-dial handlers, which
		 * execute before the dial begins, will alter the application/data to
		 * something people typically don't want to see, if we see a channel enter
		 * into Dial here, we set the appl/data accordingly and lock it.
		 */
		if (!strcmp(snapshot->appl, "Dial")) {
			ast_set_flag(&cdr->flags, AST_CDR_LOCK_APP);
		}
	}

	if (strcmp(cdr->linkedid, snapshot->linkedid)) {
		ast_string_field_set(cdr, linkedid, snapshot->linkedid);
	}
	cdr_object_check_party_a_answer(cdr);
	cdr_object_check_party_a_hangup(cdr);

	return 0;
}

static int base_process_bridge_leave(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	return 0;
}

static int base_process_dial_end(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer, const char *dial_status)
{
	return 0;
}

static enum process_bridge_enter_results base_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	/* Base process bridge enter simply indicates that we can't handle it */
	return BRIDGE_ENTER_NEED_CDR;
}

static int base_process_parked_channel(struct cdr_object *cdr, struct ast_parked_call_payload *parking_info)
{
	char park_info[128];

	ast_assert(!strcasecmp(parking_info->parkee->name, cdr->party_a.snapshot->name));

	/* Update Party A information regardless */
	cdr->fn_table->process_party_a(cdr, parking_info->parkee);

	/* Fake out where we're parked */
	ast_string_field_set(cdr, appl, "Park");
	snprintf(park_info, sizeof(park_info), "%s:%u", parking_info->parkinglot, parking_info->parkingspace);
	ast_string_field_set(cdr, data, park_info);
#ifdef GRANDSTREAM_NETWORKS
	if (cdr->special_action_type && 0 == strcmp(cdr->special_action_type, "BLD2PARK")){
		if (0 == strcmp(parking_info->parkinglot, "default") && !ast_strlen_zero(cdr->session))
		{
			char callee_ext[256] = {0};
			sscanf(cdr->session, "%*[^-]-%255s", callee_ext);
			if (!ast_strlen_zero(callee_ext)){
				snprintf(cdr->app_callee,sizeof(cdr->app_callee)-1,"%s",callee_ext);
			}
		}else{
			snprintf(cdr->app_callee,sizeof(cdr->app_callee)-1,"%s",parking_info->parkinglot);
		}
	}
#endif

	/* Prevent any further changes to the App/Data fields for this record */
	ast_set_flag(&cdr->flags, AST_CDR_LOCK_APP);

	return 0;
}

/* SINGLE STATE */

static void single_state_init_function(struct cdr_object *cdr)
{
#ifdef GRANDSTREAM_NETWORKS
	if (ast_tvzero(cdr->start)){
		cdr->start = ast_tvnow();
	}
#else
	cdr->start = ast_tvnow();
#endif
	cdr_object_check_party_a_answer(cdr);
}

static void single_state_process_party_b(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot)
{
	/* This should never happen! */
	ast_assert(cdr->party_b.snapshot == NULL);
	ast_assert(0);
	return;
}

static int single_state_process_dial_begin(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer)
{
	if (caller && !strcasecmp(cdr->party_a.snapshot->name, caller->name)) {
		base_process_party_a(cdr, caller);
		CDR_DEBUG("%p - Updated Party A %s snapshot\n", cdr,
			cdr->party_a.snapshot->name);
#ifdef GRANDSTREAM_NETWORKS
		cdr_object_swap_snapshot(cdr, &cdr->party_b, peer);
#else
		cdr_object_swap_snapshot(&cdr->party_b, peer);
#endif
		cdr_all_relink(cdr);
		CDR_DEBUG("%p - Updated Party B %s snapshot\n", cdr,
			cdr->party_b.snapshot->name);

		/* If we have two parties, lock the application that caused the
		 * two parties to be associated. This prevents mid-call event
		 * macros/gosubs from perturbing the CDR application/data
		 */
		ast_set_flag(&cdr->flags, AST_CDR_LOCK_APP);
	} else if (!strcasecmp(cdr->party_a.snapshot->name, peer->name)) {
		/* We're the entity being dialed, i.e., outbound origination */
		base_process_party_a(cdr, peer);
		CDR_DEBUG("%p - Updated Party A %s snapshot\n", cdr,
			cdr->party_a.snapshot->name);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (NULL != cdr->action_type && 0 == strncmp(cdr->action_type, "QUEUE", strlen("QUEUE")))
	{
		cdr->start = ast_tvnow();
	}
#endif

	cdr_object_transition_state(cdr, &dial_state_fn_table);
	return 0;
}

/*!
 * \brief Handle a comparison between our \ref cdr_object and a \ref cdr_object
 * already in the bridge while in the Single state. The goal of this is to find
 * a Party B for our CDR.
 *
 * \param cdr Our \ref cdr_object in the Single state
 * \param cand_cdr The \ref cdr_object already in the Bridge state
 *
 * \retval 0 The cand_cdr had a Party A or Party B that we could use as our
 * Party B
 * \retval 1 No party in the cand_cdr could be used as our Party B
 */
#ifdef GRANDSTREAM_NETWORKS
static int single_state_bridge_enter_comparison(struct cdr_object *cdr,
		struct cdr_object *cand_cdr)
{
	struct cdr_object_snapshot *party_a;

	/* Don't match on ourselves */
	if (!strcasecmp(cdr->party_a.snapshot->name, cand_cdr->party_a.snapshot->name)) {
		return 1;
	}

	/* Try the candidate CDR's Party A first */
	party_a = cdr_object_pick_party_a(&cdr->party_a, &cand_cdr->party_a);

	/* yjwu, BUG 82931:The call time for conference displays incorrectly when other user enter/leave conference */
	if ( !strcasestr(cand_cdr->action_type, "Conference") && !strcasecmp(party_a->snapshot->name, cdr->party_a.snapshot->name)) {
		CDR_DEBUG("%p - Party A %s has new Party B %s\n",
			cdr, cdr->party_a.snapshot->name, cand_cdr->party_a.snapshot->name);
		cdr_object_snapshot_copy(cdr, &cdr->party_b, &cand_cdr->party_a);
		cdr_all_relink(cdr);
		if (!cand_cdr->party_b.snapshot && 0 != strcmp(cand_cdr->special_action_type, "\"SEAMLESS_TRANSFER\""))
		{
			/* We just stole them - finalize their CDR. Note that this won't
			 * transition their state, it just sets the end time and the
			 * disposition - if we need to re-activate them later, we can.
			 */
			cdr_object_finalize(cand_cdr);
		}
		return 0;
	}

	/* Try their Party B, unless it's us */
	if (!cand_cdr->party_b.snapshot
		|| !strcasecmp(cdr->party_a.snapshot->name, cand_cdr->party_b.snapshot->name)) {
		return 1;
	}
	party_a = cdr_object_pick_party_a(&cdr->party_a, &cand_cdr->party_b);
	if (!strcasecmp(party_a->snapshot->name, cdr->party_a.snapshot->name)) {
		CDR_DEBUG("%p - Party A %s has new Party B %s\n",
			cdr, cdr->party_a.snapshot->name, cand_cdr->party_b.snapshot->name);
		cdr_object_snapshot_copy(cdr, &cdr->party_b, &cand_cdr->party_b);
		cdr_all_relink(cdr);
		return 0;
	}

	return 1;
}
#else
static int single_state_bridge_enter_comparison(struct cdr_object *cdr,
		struct cdr_object *cand_cdr)
{
	struct cdr_object_snapshot *party_a;

	/* Don't match on ourselves */
	if (!strcasecmp(cdr->party_a.snapshot->name, cand_cdr->party_a.snapshot->name)) {
		return 1;
	}

	/* Try the candidate CDR's Party A first */
	party_a = cdr_object_pick_party_a(&cdr->party_a, &cand_cdr->party_a);
	if (!strcasecmp(party_a->snapshot->name, cdr->party_a.snapshot->name)) {
		CDR_DEBUG("%p - Party A %s has new Party B %s\n",
			cdr, cdr->party_a.snapshot->name, cand_cdr->party_a.snapshot->name);
		cdr_object_snapshot_copy(&cdr->party_b, &cand_cdr->party_a);
		cdr_all_relink(cdr);
		if (!cand_cdr->party_b.snapshot)
		{
			/* We just stole them - finalize their CDR. Note that this won't
			 * transition their state, it just sets the end time and the
			 * disposition - if we need to re-activate them later, we can.
			 */
			cdr_object_finalize(cand_cdr);
		}
		return 0;
	}

	/* Try their Party B, unless it's us */
	if (!cand_cdr->party_b.snapshot
		|| !strcasecmp(cdr->party_a.snapshot->name, cand_cdr->party_b.snapshot->name)) {
		return 1;
	}
	party_a = cdr_object_pick_party_a(&cdr->party_a, &cand_cdr->party_b);
	if (!strcasecmp(party_a->snapshot->name, cdr->party_a.snapshot->name)) {
		CDR_DEBUG("%p - Party A %s has new Party B %s\n",
			cdr, cdr->party_a.snapshot->name, cand_cdr->party_b.snapshot->name);
		cdr_object_snapshot_copy(&cdr->party_b, &cand_cdr->party_b);
		cdr_all_relink(cdr);
		return 0;
	}

	return 1;
}
#endif

static enum process_bridge_enter_results single_state_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	struct ao2_iterator it_cdrs;
	char *channel_id;
	int success = 0;

	ast_string_field_set(cdr, bridge, bridge->uniqueid);

	if (ao2_container_count(bridge->channels) == 1) {
		/* No one in the bridge yet but us! */
		cdr_object_transition_state(cdr, &bridge_state_fn_table);
		return BRIDGE_ENTER_ONLY_PARTY;
	}
#ifdef GRANDSTREAM_NETWORKS
	if (NULL != bridge->technology && !strcasecmp(bridge->technology, "softmix")
		&& NULL != cdr->name && strcasestr(cdr->name, "DAHDI/o")
		&& NULL != cdr->context && strcasestr(cdr->context, "sla-did")){
		return BRIDGE_ENTER_ONLY_PARTY;
	}
#endif

	for (it_cdrs = ao2_iterator_init(bridge->channels, 0);
		!success && (channel_id = ao2_iterator_next(&it_cdrs));
		ao2_ref(channel_id, -1)) {
		struct cdr_object *cand_cdr_master;
		struct cdr_object *cand_cdr;

		cand_cdr_master = ao2_find(active_cdrs_master, channel_id, OBJ_SEARCH_KEY);
		if (!cand_cdr_master) {
			continue;
		}

		ao2_lock(cand_cdr_master);
		for (cand_cdr = cand_cdr_master; cand_cdr; cand_cdr = cand_cdr->next) {
			/* Skip any records that are not in a bridge or in this bridge.
			 * I'm not sure how that would happen, but it pays to be careful. */
			if (cand_cdr->fn_table != &bridge_state_fn_table ||
					strcmp(cdr->bridge, cand_cdr->bridge)) {
				continue;
			}

			if (single_state_bridge_enter_comparison(cdr, cand_cdr)) {
				continue;
			}
			/* We successfully got a party B - break out */
			success = 1;
			break;
		}
		ao2_unlock(cand_cdr_master);
		ao2_cleanup(cand_cdr_master);
	}
	ao2_iterator_destroy(&it_cdrs);

	/* We always transition state, even if we didn't get a peer */
	cdr_object_transition_state(cdr, &bridge_state_fn_table);

	/* Success implies that we have a Party B */
	if (success) {
		return BRIDGE_ENTER_OBTAINED_PARTY_B;
	}

	return BRIDGE_ENTER_NO_PARTY_B;
}

static int single_state_process_parking_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	cdr_object_transition_state(cdr, &parked_state_fn_table);
	return 0;
}


/* DIAL STATE */

static void dial_state_process_party_b(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot)
{
	ast_assert(snapshot != NULL);
	ast_assert(cdr->party_b.snapshot
		&& !strcasecmp(cdr->party_b.snapshot->name, snapshot->name));

#ifdef GRANDSTREAM_NETWORKS
	cdr_object_swap_snapshot(cdr, &cdr->party_b, snapshot);
#else
	cdr_object_swap_snapshot(&cdr->party_b, snapshot);
#endif

	/* If party B hangs up, finalize this CDR */
	if (ast_test_flag(&cdr->party_b.snapshot->flags, AST_FLAG_DEAD)) {
		cdr_object_transition_state(cdr, &finalized_state_fn_table);
	}
}

static int dial_state_process_dial_begin(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer)
{
	/* Don't process a begin dial here. A party A already in the dial state will
	 * who receives a dial begin for something else will be handled by the
	 * message router callback and will add a new CDR for the party A */
	return 1;
}

/*!
 * \internal
 * \brief Convert a dial status to a CDR disposition
 */
static enum ast_cdr_disposition dial_status_to_disposition(const char *dial_status)
{
	if (!strcmp(dial_status, "ANSWER")) {
		return AST_CDR_ANSWERED;
	} else if (!strcmp(dial_status, "BUSY")) {
		return AST_CDR_BUSY;
	} else if (!strcmp(dial_status, "CANCEL") || !strcmp(dial_status, "NOANSWER")) {
		return AST_CDR_NOANSWER;
	} else if (!strcmp(dial_status, "CONGESTION")) {
		if (!is_cdr_flag_set(CDR_CONGESTION)) {
			return AST_CDR_FAILED;
		} else {
			return AST_CDR_CONGESTION;
		}
	} else if (!strcmp(dial_status, "FAILED")) {
		return AST_CDR_FAILED;
	}
	return AST_CDR_FAILED;
}

static int dial_state_process_dial_end(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer, const char *dial_status)
{
	struct ast_channel_snapshot *party_a;

	if (caller) {
		party_a = caller;
	} else {
		party_a = peer;
	}
	ast_assert(!strcasecmp(cdr->party_a.snapshot->name, party_a->name));
#ifdef GRANDSTREAM_NETWORKS
	cdr_object_swap_snapshot(cdr, &cdr->party_a, party_a);
#else
	cdr_object_swap_snapshot(&cdr->party_a, party_a);
#endif

	if (cdr->party_b.snapshot) {
		if (strcasecmp(cdr->party_b.snapshot->name, peer->name)) {
			/* Not the status for this CDR - defer back to the message router */
			return 1;
		}
#ifdef GRANDSTREAM_NETWORKS
		cdr_object_swap_snapshot(cdr, &cdr->party_b, peer);

		/* added by zgmo for BUG155038 */
		if (!strcmp(dial_status, "ANSWER") && !strcmp(cdr->context, "from-internal-xfer")) {
			cdr->answer = ast_tvnow();
		}
#else
		cdr_object_swap_snapshot(&cdr->party_b, peer);
#endif
	}

	/* Set the disposition based on the dial string. */
	cdr->disposition = dial_status_to_disposition(dial_status);
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_NOTICE, "cdr(%p) set disposition to %d", cdr, cdr->disposition);
#endif
	if (cdr->disposition == AST_CDR_ANSWERED) {
		/* Switch to dial pending to wait and see what the caller does */
		cdr_object_transition_state(cdr, &dialed_pending_state_fn_table);
	} else {
		cdr_object_transition_state(cdr, &finalized_state_fn_table);
#ifdef GRANDSTREAM_NETWORKS
		if (!ast_strlen_zero(cdr->record_files)) {
			cdr->record_files[0] = '\0';
		}
#endif
	}

	return 0;
}

static enum process_bridge_enter_results dial_state_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	int success = 0;

	ast_string_field_set(cdr, bridge, bridge->uniqueid);

	/* Get parties in the bridge */
	if (ao2_container_count(bridge->channels) == 1) {
		/* No one in the bridge yet but us! */
		cdr_object_transition_state(cdr, &bridge_state_fn_table);
		return BRIDGE_ENTER_ONLY_PARTY;
	}

	/* If we don't have a Party B (originated channel), skip it */
	if (cdr->party_b.snapshot) {
		struct ao2_iterator it_cdrs;
		char *channel_id;

		for (it_cdrs = ao2_iterator_init(bridge->channels, 0);
			!success && (channel_id = ao2_iterator_next(&it_cdrs));
			ao2_ref(channel_id, -1)) {
			struct cdr_object *cand_cdr_master;
			struct cdr_object *cand_cdr;

			cand_cdr_master = ao2_find(active_cdrs_master, channel_id, OBJ_SEARCH_KEY);
			if (!cand_cdr_master) {
				continue;
			}

			ao2_lock(cand_cdr_master);
			for (cand_cdr = cand_cdr_master; cand_cdr; cand_cdr = cand_cdr->next) {
				/* Skip any records that are not in a bridge or in this bridge.
				 * I'm not sure how that would happen, but it pays to be careful. */
				if (cand_cdr->fn_table != &bridge_state_fn_table
					|| strcmp(cdr->bridge, cand_cdr->bridge)) {
					continue;
				}

				/* Skip any records that aren't our Party B */
				if (strcasecmp(cdr->party_b.snapshot->name, cand_cdr->party_a.snapshot->name)) {
					continue;
				}
#ifdef GRANDSTREAM_NETWORKS
				cdr_object_snapshot_copy(cdr, &cdr->party_b, &cand_cdr->party_a);
#else
				cdr_object_snapshot_copy(&cdr->party_b, &cand_cdr->party_a);
#endif
				/* If they have a Party B, they joined up with someone else as their
				 * Party A. Don't finalize them as they're active. Otherwise, we
				 * have stolen them so they need to be finalized.
				 */
				if (!cand_cdr->party_b.snapshot) {
					cdr_object_finalize(cand_cdr);
				}
				success = 1;
				break;
			}
			ao2_unlock(cand_cdr_master);
			ao2_cleanup(cand_cdr_master);
		}
		ao2_iterator_destroy(&it_cdrs);
	}

	/* We always transition state, even if we didn't get a peer */
	cdr_object_transition_state(cdr, &bridge_state_fn_table);

	/* Success implies that we have a Party B */
	if (success) {
		return BRIDGE_ENTER_OBTAINED_PARTY_B;
	}
	return BRIDGE_ENTER_NO_PARTY_B;
}

/* DIALED PENDING STATE */

static int dialed_pending_state_process_party_a(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot)
{
	/* If we get a CEP change, we're executing dialplan. If we have a Party B
	 * that means we need a new CDR; otherwise, switch us over to single.
	 */
	if (snapshot_cep_changed(cdr->party_a.snapshot, snapshot)) {
		if (cdr->party_b.snapshot) {
			cdr_object_transition_state(cdr, &finalized_state_fn_table);
			cdr->fn_table->process_party_a(cdr, snapshot);
			return 1;
		} else {
			cdr_object_transition_state(cdr, &single_state_fn_table);
			cdr->fn_table->process_party_a(cdr, snapshot);
			return 0;
		}
	}
	base_process_party_a(cdr, snapshot);
	return 0;
}

static enum process_bridge_enter_results dialed_pending_state_process_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	cdr_object_transition_state(cdr, &dial_state_fn_table);
	return cdr->fn_table->process_bridge_enter(cdr, bridge, channel);
}

static int dialed_pending_state_process_parking_bridge_enter(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	if (cdr->party_b.snapshot) {
		/* We can't handle this as we have a Party B - ask for a new one */
		return 1;
	}
	cdr_object_transition_state(cdr, &parked_state_fn_table);
	return 0;
}

static int dialed_pending_state_process_dial_begin(struct cdr_object *cdr, struct ast_channel_snapshot *caller, struct ast_channel_snapshot *peer)
{
	cdr_object_transition_state(cdr, &finalized_state_fn_table);

	/* Ask for a new CDR */
	return 1;
}

/* BRIDGE STATE */

static void bridge_state_process_party_b(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot)
{
	ast_assert(cdr->party_b.snapshot
		&& !strcasecmp(cdr->party_b.snapshot->name, snapshot->name));

#ifdef GRANDSTREAM_NETWORKS
	cdr_object_swap_snapshot(cdr, &cdr->party_b, snapshot);
#else
	cdr_object_swap_snapshot(&cdr->party_b, snapshot);
#endif

	/* If party B hangs up, finalize this CDR */
	if (ast_test_flag(&cdr->party_b.snapshot->flags, AST_FLAG_DEAD)) {
		cdr_object_transition_state(cdr, &finalized_state_fn_table);
	}
}

static int bridge_state_process_bridge_leave(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	if (strcmp(cdr->bridge, bridge->uniqueid)) {
		return 1;
	}
	if (strcasecmp(cdr->party_a.snapshot->name, channel->name)
		&& cdr->party_b.snapshot
		&& strcasecmp(cdr->party_b.snapshot->name, channel->name)) {
		return 1;
	}
	cdr_object_transition_state(cdr, &finalized_state_fn_table);

	return 0;
}

/* PARKED STATE */

static int parked_state_process_bridge_leave(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge, struct ast_channel_snapshot *channel)
{
	if (strcasecmp(cdr->party_a.snapshot->name, channel->name)) {
		return 1;
	}
	cdr_object_transition_state(cdr, &finalized_state_fn_table);

	return 0;
}

/* FINALIZED STATE */

static void finalized_state_init_function(struct cdr_object *cdr)
{
#ifdef GRANDSTREAM_NETWORKS
	if (!ast_tvzero(cdr->end) && strcasecmp(cdr->action_type, "CONFERENCE")
		&& strcasecmp(cdr->special_action_type, "TRIGGER_CALL_FLIP_TRANSFER")){
		cdr->end.tv_sec = 0;
		cdr->end.tv_usec = 0;
		ast_log(LOG_DEBUG, "Reset cdr(%p) end", cdr);
	}
#endif
	cdr_object_finalize(cdr);
}

static int finalized_state_process_party_a(struct cdr_object *cdr, struct ast_channel_snapshot *snapshot)
{
	if (ast_test_flag(&snapshot->softhangup_flags, AST_SOFTHANGUP_HANGUP_EXEC)
		&& is_cdr_flag_set(CDR_END_BEFORE_H_EXTEN)) {
		return 0;
	}

	/* added by zgmo for new feature of marking nat call */
	if (!cdr->is_nat_call && snapshot->is_nat_call) {
		cdr->is_nat_call = snapshot->is_nat_call;
		ast_log(LOG_DEBUG, "cdr:%p(uid:%s), Chan:%s, Detect NAT", cdr, cdr->uniqueid, cdr->name);
	}
	/* end */

	/* Indicate that, if possible, we should get a new CDR */
	return 1;
}

/*!
 * \internal
 * \brief Filter channel snapshots by technology
 */
static int filter_channel_snapshot(struct ast_channel_snapshot *snapshot)
{
	return snapshot->tech_properties & AST_CHAN_TP_INTERNAL;
}

/*!
 * \internal
 * \brief Filter a channel cache update
 */
static int filter_channel_cache_message(struct ast_channel_snapshot *old_snapshot,
		struct ast_channel_snapshot *new_snapshot)
{
	int ret = 0;

	/* Drop cache updates from certain channel technologies */
	if (old_snapshot) {
		ret |= filter_channel_snapshot(old_snapshot);
	}
	if (new_snapshot) {
		ret |= filter_channel_snapshot(new_snapshot);
	}

	return ret;
}

static int dial_status_end(const char *dialstatus)
{
	return (strcmp(dialstatus, "RINGING") &&
			strcmp(dialstatus, "PROCEEDING") &&
			strcmp(dialstatus, "PROGRESS"));
}

#ifdef GRANDSTREAM_NETWORKS
/*add by wxu@grandstream.cn for set cdr_obj_vars*/
static const char * const cdr_obj_vars[] ={
	"app_caller",
	"app_callee",
	"app_callee_cdr", //ONLY set cdr, not change connected num in channel
	"action_type",
	"feature_transfer",
	"record_files",
	"record_files_desc",
	"record_files_uuid",
	"session",
	"dst_need",
	"action_owner",
	"src_need",
	"special_action_type",
	"special_dst",
	"inbound_trunk_index",
	"outbound_trunk_index",
	"status",
	"reason"
};

enum cdr_vars {
	APP_CALLER = 0,
	APP_CALLEE,
	APP_CALLEE_CDR,
	ACTION_TYPE,
	FEATURE_TRANSFER,
	RECORD_FILES,
	RECORD_FILES_DESC,
	RECORD_FILES_UUID,
	SESSION,
	DST_NEED,
	ACTION_OWNER,
	SRC_NEED,
	SPECIAL_ACTION_TYPE,
	SPECIAL_DST,
	INBOUND_TRUNK_INDEX,
	OUTBOUND_TRUNK_INDEX,
	STATUS,
	REASON,
	VAR_TOTAL
};

static int ast_set_cdr_obj_vars(struct cdr_object *gcdr, const char *channel_name,struct cdr_object *cdr, const char *var, const char *value)
{
	int x;

	ast_mutex_lock(&cdr->party_lock);	/*Add by weidu, for Bug 263475*/
	if (cdr->party_a.snapshot && 0 != strcasecmp(channel_name, cdr->party_a.snapshot->name)
		&& (var && NULL != strstr(var, "action_type"))
		&& (value && 0 == strcmp(value, "TRANSFER"))){
		ast_log(LOG_DEBUG,"channel_name:%s is diff with party a:%s",channel_name, cdr->party_a.snapshot->name);
		ast_mutex_unlock(&cdr->party_lock); /*Add by weidu, for Bug 263475*/
		return -1;
	}
	if (cdr->party_b.snapshot && !strcasecmp(channel_name, cdr->party_b.snapshot->name)){
		ast_log(LOG_DEBUG,"channel_name:%s is party_b | gcdr:%p, cdr:%p",
			channel_name, gcdr, cdr);
		ast_mutex_unlock(&cdr->party_lock); /*Add by weidu, for Bug 263475*/
		return -1;
	}
	if (strstr(cdr->special_action_type, "FOLLOWME/external") && 0 == strcmp(var, "outbound_trunk_index")){
		if (cdr->party_b.snapshot && NULL == strstr(cdr->special_action_type, cdr->party_b.snapshot->name)){
			ast_mutex_unlock(&cdr->party_lock); /*Add by weidu, for Bug 263475*/
			return -1;
		}
	}
	ast_mutex_unlock(&cdr->party_lock);	/*Add by weidu, for Bug 263475*/
	
	if ((!strcmp(var, "RECORD_FILES") || !strcmp(var ,"record_files_desc"))
		&& !ast_strlen_zero(value) && AST_CDR_NOANSWER ==cdr->disposition
		&& !strcasestr(cdr->special_action_type, "overwrt_record")) {
		ast_log(LOG_DEBUG,"cdr:%p disposition is no anser, forbit to set RECORD_FILES/record_files_desc", cdr);

		if (!strcmp(var ,"record_files_desc")
			&& !ast_strlen_zero(cdr->app_callee) && !ast_strlen_zero(cdr->action_type)
			&& strstr(cdr->action_type, "QUEUE[")
			&& strstr(cdr->action_type, cdr->app_callee)
			&& ast_strlen_zero(cdr->special_action_type)) {
			ast_log(LOG_DEBUG, "but in this case approve!");
		} else if (!strcmp(var ,"record_files_desc")
			&& !ast_strlen_zero(cdr->app_callee) && !ast_strlen_zero(cdr->action_type)
			&& strstr(cdr->action_type, "IVR[")
			&& strstr(cdr->action_type, cdr->app_callee)){
			ast_log(LOG_DEBUG, "but in this case approve!");
		} else {
			return -1;
		}
	}

	for(x = 0; x < VAR_TOTAL; x++){
		/* strip the '__'*/
		const char *tmp_var = var;
		if (tmp_var != NULL && !strncmp(tmp_var, "__", strlen("__"))) {
			tmp_var = tmp_var + 2;
		}
		if ( tmp_var != NULL && !strcasecmp(tmp_var, cdr_obj_vars[x])) {
			break;
		}
	}

	/*added by wxu@grandstream.cn for set cdr for Local channel ;1 if channel is Loal channel ;2 */
	if( channel_name != NULL && !strncmp(channel_name,"Local/",strlen("Local/")) && strstr(channel_name,";2") != NULL) {
		char local_channel[64] = {0};

		snprintf(local_channel,sizeof(local_channel)-1,"%s",channel_name);
		if(strlen(local_channel) > 0){
			char webcall[32] = {0};
			struct ast_channel *Local_chan = NULL;

			local_channel[strlen(local_channel)-1] = '1';
			ast_log(LOG_DEBUG,"local_channel:%s",local_channel);
			switch(x){
				case APP_CALLEE:
				case APP_CALLEE_CDR:
					ao2_unlock(gcdr);
					Local_chan = ast_channel_get_by_name(channel_name);
					if(Local_chan){
						const char *from_callback = pbx_builtin_getvar_helper(Local_chan, "FROM_CALLBACK");
						if( from_callback && !strcasecmp(from_callback,"yes")){
							ast_channel_unref(Local_chan);
							break;
						}
						ast_channel_unref(Local_chan);
					}
					ao2_lock(gcdr);
					ast_cdr_setvar(local_channel,var,value);
					break;
				case APP_CALLER:
				case ACTION_TYPE:
					ast_cdr_getvar(channel_name,"webcall",webcall,sizeof(webcall));
					if(!strcasecmp(webcall,"yes")){
						ast_cdr_setvar(local_channel,var,value);
					}
					break;
				case RECORD_FILES:
				case SPECIAL_ACTION_TYPE:
				case SPECIAL_DST:
				case INBOUND_TRUNK_INDEX:
				case OUTBOUND_TRUNK_INDEX:
				case STATUS:
				case SESSION:
				case REASON:
					ast_cdr_setvar(local_channel,var,value);
					break;
				default:
					break;
			}
		}
	}
	/*end*/

	ast_log(LOG_DEBUG,"[set cdr]channel_name:%s, var:%s, value:%s",
			S_OR(channel_name,"null"), var, value);
	ast_log(LOG_DEBUG,"gcdr(%p)[%s], cdr(%p)[%s]",
		gcdr, S_OR(gcdr->name,"null"),
		cdr, S_OR(cdr->name,"null"));

	switch(x){
		case APP_CALLER:
			snprintf(cdr->app_caller,sizeof(cdr->app_caller)-1,"%s",value);
			return 0;
			break;
		case APP_CALLEE:
		case APP_CALLEE_CDR:
			/* for bug 59297*/
			if(strlen(cdr->app_callee) == 0 || !strcasecmp(value,"null") || !strncmp(var, "__", strlen("__")) )
			{
				if(strcasecmp(value,"null")){
					snprintf(cdr->app_callee,sizeof(cdr->app_callee)-1,"%s",value);
				}else{
					memset(cdr->app_callee,0,sizeof(cdr->app_callee));
				}
			}
			return 0;
			break;
		case ACTION_TYPE:
			ast_string_field_set(cdr, action_type, value);
			/*bug 121093*/
			if (NULL != channel_name){
				char rec[512] = {0};
				ast_cdr_getvar(channel_name, "action_type", rec, sizeof(rec)-1);
				if ( 0 == strncmp(rec, "CALLFORWARD/IVR", strlen("CALLFORWARD/IVR"))){
					ast_cdr_setvar(channel_name, "record_files", "");
				}
			}
			return 0;
			break;
		case FEATURE_TRANSFER:
			memset(cdr->feature_transfer,0,sizeof(cdr->feature_transfer));
			snprintf(cdr->feature_transfer,sizeof(cdr->feature_transfer)-1,"%s",value);
			return 0;
			break;
		case RECORD_FILES:
			if (ast_strlen_zero(value) || !strcasecmp(value, "NO_FILE")) {
				cdr->record_files[0] = '\0';
				cdr->record_files_desc[0] = '\0';
			} else {
				if (AST_CDR_NULL == cdr->disposition || !strlen(cdr->record_files)
					|| (strlen(cdr->record_files) && strstr(value, cdr->record_files))
					|| !strcmp(cdr->special_action_type, "atxee")
					|| strcasestr(cdr->special_action_type, "overwrt_record")) {
					snprintf(cdr->record_files,sizeof(cdr->record_files)-1,"%s",value);
				} else {
					ast_log(LOG_WARNING, 
						"Can't set cdr record due to disposition isn't 'AST_CDR_NULL'(%s) "
						"and having cdr record:%s",
						ast_cdr_disp2str(cdr->disposition), cdr->record_files);
				}
			}
			return 0;
			break;
		case RECORD_FILES_DESC:
			if (ast_strlen_zero(cdr->record_files)) {
				ast_log(LOG_WARNING, "Try to set record-desc with NULL recordfile string");
				
				if (!ast_strlen_zero(cdr->record_files_desc)) {
					snprintf(cdr->record_files_desc, sizeof(cdr->record_files_desc)-1, "%s", "");
				}
				return -1;
			} else {
				snprintf(cdr->record_files_desc, sizeof(cdr->record_files_desc)-1, "%s", value);
			}
			return 0;
			break;
		case RECORD_FILES_UUID:
			if (ast_strlen_zero(cdr->record_files)) {
				ast_log(LOG_WARNING, "Try to set record-uuid with NULL recordfile string");
				return -1;
			} else {
				snprintf(cdr->record_files_uuid, sizeof(cdr->record_files_uuid)-1, "%s", value);
			}
			return 0;
			break;
		case SESSION:
			ast_string_field_set(cdr, session, value);
			return 0;
			break;
		case DST_NEED:
			if(value){
				if(!strcasecmp(value,"true")){
					cdr->dst_need = 1;
				}else if(!strcasecmp(value,"flase")){
					cdr->dst_need = 0;
				}else {
					ast_log(LOG_ERROR,"dst_need value is error,it must be true or flase or (null)!!!");
				}
			}else {
				cdr->dst_need = 0;
			}
			return 0;
			break;
		case ACTION_OWNER:
			ast_string_field_set(cdr, action_owner, value);
			return 0;
			break;
		case SRC_NEED:
			if(value){
				if(!strcasecmp(value,"true")){
					cdr->src_need= 1;
				}else if(!strcasecmp(value,"flase")){
					cdr->src_need = 0;
				}else {
					ast_log(LOG_ERROR,"src_need value is error,it must be true or flase or (null)!!!");
				}
			}else {
				cdr->src_need = 0;
			}
			return 0;
			break;
		case SPECIAL_ACTION_TYPE:
			ast_string_field_set(cdr, special_action_type, value);
			return 0;
			break;
		case SPECIAL_DST:
			ast_string_field_set(cdr, special_dst, value);
			return 0;
			break;
		case INBOUND_TRUNK_INDEX:
			ast_string_field_set(cdr, inbound_trunk_index, value);
			return 0;
			break;
		case OUTBOUND_TRUNK_INDEX:
			ast_string_field_set(cdr, outbound_trunk_index, value);
			return 0;
			break;
		case STATUS:
			ast_string_field_set(cdr, status, value);
			return 0;
			break;
		case REASON:
			snprintf(cdr->reason, sizeof(cdr->reason)-1, "%s", value);
			return 0;
			break;
		default:
			ast_log(LOG_DEBUG,"can't find var[%s] in cdr_obj_vars",var);
			break;
	}

	return -1;
}
/*end*/
#endif

/* TOPIC ROUTER CALLBACKS */

/*!
 * \brief Handler for Stasis-Core dial messages
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message The message
 */
static void handle_dial_message(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	struct cdr_object *cdr;
	struct ast_multi_channel_blob *payload = stasis_message_data(message);
	struct ast_channel_snapshot *caller;
	struct ast_channel_snapshot *peer;
	struct cdr_object *it_cdr;
	struct ast_json *dial_status_blob;
	const char *dial_status = NULL;
	int res = 1;

	caller = ast_multi_channel_blob_get_channel(payload, "caller");
	peer = ast_multi_channel_blob_get_channel(payload, "peer");
	if (!peer && !caller) {
		return;
	}

	if (peer && filter_channel_snapshot(peer)) {
		return;
	}

	if (caller && filter_channel_snapshot(caller)) {
		return;
	}

	dial_status_blob = ast_json_object_get(ast_multi_channel_blob_get_json(payload), "dialstatus");
	if (dial_status_blob) {
		dial_status = ast_json_string_get(dial_status_blob);
	}

	CDR_DEBUG("Dial %s message for %s, %s: %u.%08u\n",
		ast_strlen_zero(dial_status) ? "Begin" : "End",
		caller ? caller->name : "(none)",
		peer ? peer->name : "(none)",
		(unsigned int)stasis_message_timestamp(message)->tv_sec,
		(unsigned int)stasis_message_timestamp(message)->tv_usec);

#ifdef GRANDSTREAM_NETWORKS
	if(peer && caller){
		char tmp_buff[512] = {0};
		ast_cdr_getvar(caller->name, "special_action_type", tmp_buff,sizeof(tmp_buff));
		if(!strcmp(tmp_buff, "CONF_INV_FAIL")){
			ast_cdr_setvar(caller->name, "special_action_type", "");
		}
		memset(tmp_buff, 0, sizeof(tmp_buff));
		ast_cdr_getvar(peer->name, "special_action_type", tmp_buff,sizeof(tmp_buff));
		if(!strcmp(tmp_buff, "CONF_INV_FAIL")){
			ast_cdr_setvar(caller->name, "special_action_type", tmp_buff);
		}
	}
#endif
	/* Figure out who is running this show */
	if (caller) {
		cdr = ao2_find(active_cdrs_master, caller->uniqueid, OBJ_SEARCH_KEY);
	} else {
		cdr = ao2_find(active_cdrs_master, peer->uniqueid, OBJ_SEARCH_KEY);
	}
	if (!cdr) {
		ast_log(AST_LOG_WARNING, "No CDR for channel %s\n", caller ? caller->name : peer->name);
		ast_assert(0);
		return;
	}

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (ast_strlen_zero(dial_status)) {
			if (!it_cdr->fn_table->process_dial_begin) {
				continue;
			}
			CDR_DEBUG("%p - Processing Dial Begin message for channel %s, peer %s\n",
				it_cdr,
				caller ? caller->name : "(none)",
				peer ? peer->name : "(none)");
			res &= it_cdr->fn_table->process_dial_begin(it_cdr,
					caller,
					peer);
		} else if (dial_status_end(dial_status)) {
			if (!it_cdr->fn_table->process_dial_end) {
				continue;
			}
			CDR_DEBUG("%p - Processing Dial End message for channel %s, peer %s\n",
				it_cdr,
				caller ? caller->name : "(none)",
				peer ? peer->name : "(none)");
#ifdef GRANDSTREAM_NETWORKS
			/* add by pchwu for bug 60256 */
			if ( peer && (!strncmp(it_cdr->action_type, "QUEUE", strlen("QUEUE")) || !strncmp(it_cdr->action_type, "RINGGROUP", strlen("RINGGROUP"))) && it_cdr->context && it_cdr->context[0] == '\0' )
			{
				char dcontext[64] = { 0 };
				ast_cdr_getvar(peer->name, "dcontext", dcontext,sizeof(dcontext));
				if (strlen(dcontext) > 0)
				{
					ast_string_field_set(it_cdr, context, dcontext);
				}
			}
			/* end */

			/*addend by wxu@grandstream.cn for Local call need dst*/
			if( caller && peer && !strncmp(peer->name,"Local",strlen("Local")) ){
				if(it_cdr->dst_need){
					char peer_app_callee[64] = {0};
					ast_cdr_getvar(peer->name,"app_callee",peer_app_callee,sizeof(peer_app_callee));
					if( strlen(peer_app_callee) > 0 ){
						snprintf(it_cdr->app_callee,sizeof(it_cdr->app_callee),"%s",peer_app_callee);
					}
				}
			}
			if (peer && caller && (NULL != strstr(peer->name, "trunk") || NULL != strstr(peer->name, "DAHDI/o"))
				&& it_cdr->party_b.snapshot && 0 == strcmp(peer->name, it_cdr->party_b.snapshot->name)){
				char peer_ee[128] = { 0 };
				ast_cdr_getvar(peer->name, "app_callee", peer_ee,sizeof(peer_ee));
				if ('\0' != peer_ee[0] && caller->caller_number && 0 != strcmp(caller->caller_number, peer_ee)){
					ast_set_cdr_obj_vars(cdr, caller->name, it_cdr, "__app_callee",peer_ee);
				}
			}
			/*end*/
#endif
			it_cdr->fn_table->process_dial_end(it_cdr,
					caller,
					peer,
					dial_status);
		}
	}

	/* If no CDR handled a dial begin message, make a new one */
	if (res && ast_strlen_zero(dial_status)) {
		struct cdr_object *new_cdr;

		new_cdr = cdr_object_create_and_append(cdr);
		if (new_cdr) {
			new_cdr->fn_table->process_dial_begin(new_cdr, caller, peer);
		}
	}
	ao2_unlock(cdr);
	ao2_cleanup(cdr);
}

static int cdr_object_finalize_party_b(void *obj, void *arg, void *data, int flags)
{
	struct cdr_object *cdr = obj;
#ifdef GRANDSTREAM_NETWORKS
	ast_debug(1, "cdr_object_finalize_party_b");
#endif

	if (!strcasecmp(cdr->party_b_name, arg)) {
#ifdef AST_DEVMODE
		struct ast_channel_snapshot *party_b = data;

		/*
		 * For sanity's sake we also assert the party_b snapshot
		 * is consistent with the key.
		 */
		ast_assert(cdr->party_b.snapshot
			&& !strcasecmp(cdr->party_b.snapshot->name, party_b->name));
#endif

		/* Don't transition to the finalized state - let the Party A do
		 * that when its ready
		 */
		cdr_object_finalize(cdr);
	}
	return 0;
}

static int cdr_object_update_party_b(void *obj, void *arg, void *data, int flags)
{
	struct cdr_object *cdr = obj;

	if (cdr->fn_table->process_party_b
		&& !strcasecmp(cdr->party_b_name, arg)) {
		struct ast_channel_snapshot *party_b = data;

		/*
		 * For sanity's sake we also check the party_b snapshot
		 * for consistency with the key.  The callback needs and
		 * asserts the snapshot to be this way.
		 */
		if (!cdr->party_b.snapshot
			|| strcasecmp(cdr->party_b.snapshot->name, party_b->name)) {
			ast_log(LOG_NOTICE,
				"CDR for Party A %s(%s) has inconsistent Party B %s name.  Message can be ignored but this shouldn't happen.\n",
				cdr->linkedid,
				cdr->party_a.snapshot->name,
				cdr->party_b_name);
			return 0;
		}

		cdr->fn_table->process_party_b(cdr, party_b);

		{
			struct cdr_object *cdr_b = ao2_find(active_cdrs_master, party_b->uniqueid, OBJ_SEARCH_KEY);
			if (cdr_b) {
				int is_nat_call = 0;
				if (!strcmp(cdr_b->special_action_type, "masquerade")) {
					is_nat_call = atoi(cdr_b->status);
					ast_log(LOG_DEBUG, "cdr(%p|%s)'s Peer(%p|%s) is masqueraded, peer-status:%s",
						cdr, cdr->name, cdr_b, cdr_b->name, cdr_b->status);
				} else {
					is_nat_call = cdr_b->is_nat_call;
				}

				if (!cdr->peer_is_nat_call && is_nat_call) {
					cdr->peer_is_nat_call = 1;
					ast_log(LOG_DEBUG, "cdr(%p|%s), Detect PEER_NAT|Peer(%p|%s)",
						cdr, cdr->name, cdr_b, cdr_b->name);
				} else if (cdr->peer_is_nat_call && !is_nat_call) {
					cdr->peer_is_nat_call = 0;
					ast_log(LOG_DEBUG, "cdr(%p|%s), Detect clear PEER_NAT|Peer(%p|%s)",
						cdr, cdr->name, cdr_b, cdr_b->name);
				}

				ao2_cleanup(cdr_b);
			}
		}
	}
	return 0;
}

/*! \brief Determine if we need to add a new CDR based on snapshots */
static int check_new_cdr_needed(struct ast_channel_snapshot *old_snapshot,
		struct ast_channel_snapshot *new_snapshot)
{
	/* If we're dead, we don't need a new CDR */
	if (!new_snapshot
		|| (ast_test_flag(&new_snapshot->softhangup_flags, AST_SOFTHANGUP_HANGUP_EXEC)
			&& is_cdr_flag_set(CDR_END_BEFORE_H_EXTEN))) {
		return 0;
	}

	/* Auto-fall through will increment the priority but have no application */
	if (ast_strlen_zero(new_snapshot->appl)) {
		return 0;
	}

	if (old_snapshot && !snapshot_cep_changed(old_snapshot, new_snapshot)) {
		return 0;
	}

	return 1;
}

/*!
 * \brief Handler for Stasis-Core channel cache update messages
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message The message
 */
static void handle_channel_cache_message(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	struct cdr_object *cdr;
	struct stasis_cache_update *update = stasis_message_data(message);
	struct ast_channel_snapshot *old_snapshot;
	struct ast_channel_snapshot *new_snapshot;
	struct cdr_object *it_cdr;

	ast_assert(update != NULL);
	ast_assert(ast_channel_snapshot_type() == update->type);

	old_snapshot = stasis_message_data(update->old_snapshot);
	new_snapshot = stasis_message_data(update->new_snapshot);

	if (filter_channel_cache_message(old_snapshot, new_snapshot)) {
		return;
	}

	if (new_snapshot && !old_snapshot) {
		cdr = cdr_object_alloc(new_snapshot);
		if (!cdr) {
			return;
		}
		cdr->is_root = 1;
		ao2_link(active_cdrs_master, cdr);
	} else {
		const char *uniqueid;

		uniqueid = new_snapshot ? new_snapshot->uniqueid : old_snapshot->uniqueid;
		cdr = ao2_find(active_cdrs_master, uniqueid, OBJ_SEARCH_KEY);
	}

	/* Handle Party A */
	if (!cdr) {
		const char *name;

		name = new_snapshot ? new_snapshot->name : old_snapshot->name;
		ast_log(AST_LOG_WARNING, "No CDR for channel %s\n", name);
		ast_assert(0);
	} else if (new_snapshot) {
		int all_reject = 1;

		ao2_lock(cdr);
		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			if (!it_cdr->fn_table->process_party_a) {
				continue;
			}
			all_reject &= it_cdr->fn_table->process_party_a(it_cdr, new_snapshot);
		}
		if (all_reject && check_new_cdr_needed(old_snapshot, new_snapshot)) {
			/* We're not hung up and we have a new snapshot - we need a new CDR */
			struct cdr_object *new_cdr;

			new_cdr = cdr_object_create_and_append(cdr);
			if (new_cdr) {
				new_cdr->fn_table->process_party_a(new_cdr, new_snapshot);
			}
		}
		ao2_unlock(cdr);
	} else {
		ao2_lock(cdr);
		CDR_DEBUG("%p - Beginning finalize/dispatch for %s\n", cdr, old_snapshot->name);
		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			cdr_object_finalize(it_cdr);
		}
		cdr_object_dispatch(cdr);
		ao2_unlock(cdr);

		cdr_all_unlink(cdr);
		ao2_unlink(active_cdrs_master, cdr);
	}

	/* Handle Party B */
	if (new_snapshot) {
		ao2_callback_data(active_cdrs_all, OBJ_NODATA | OBJ_MULTIPLE | OBJ_SEARCH_KEY,
			cdr_object_update_party_b, (char *) new_snapshot->name, new_snapshot);
	} else {
		ao2_callback_data(active_cdrs_all, OBJ_NODATA | OBJ_MULTIPLE | OBJ_SEARCH_KEY,
			cdr_object_finalize_party_b, (char *) old_snapshot->name, old_snapshot);
	}

	ao2_cleanup(cdr);
}

struct bridge_leave_data {
	struct ast_bridge_snapshot *bridge;
	struct ast_channel_snapshot *channel;
};

/*! \brief Callback used to notify CDRs of a Party B leaving the bridge */
static int cdr_object_party_b_left_bridge_cb(void *obj, void *arg, void *data, int flags)
{
	struct cdr_object *cdr = obj;
	struct bridge_leave_data *leave_data = data;

	if (cdr->fn_table == &bridge_state_fn_table
		&& !strcmp(cdr->bridge, leave_data->bridge->uniqueid)
		&& !strcasecmp(cdr->party_b_name, arg)) {
		/*
		 * For sanity's sake we also assert the party_b snapshot
		 * is consistent with the key.
		 */
		ast_assert(cdr->party_b.snapshot
			&& !strcasecmp(cdr->party_b.snapshot->name, leave_data->channel->name));

		/* It is our Party B, in our bridge. Set the end time and let the handler
		 * transition our CDR appropriately when we leave the bridge.
		 */
		cdr_object_finalize(cdr);
	}
	return 0;
}

/*! \brief Filter bridge messages based on bridge technology */
static int filter_bridge_messages(struct ast_bridge_snapshot *bridge)
{
	/* Ignore holding bridge technology messages. We treat this simply as an application
	 * that a channel enters into.
	 */
	if (!strcmp(bridge->technology, "holding_bridge") && strcmp(bridge->subclass, "parking")) {
		return 1;
	}
	return 0;
}

/*!
 * \brief Handler for when a channel leaves a bridge
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message The message - hopefully a bridge one!
 */
static void handle_bridge_leave_message(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	struct ast_bridge_blob *update = stasis_message_data(message);
	struct ast_bridge_snapshot *bridge = update->bridge;
	struct ast_channel_snapshot *channel = update->channel;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct bridge_leave_data leave_data = {
		.bridge = bridge,
		.channel = channel,
	};
	int left_bridge = 0;

	if (filter_bridge_messages(bridge)) {
		return;
	}

	if (filter_channel_snapshot(channel)) {
		return;
	}

	CDR_DEBUG("Bridge Leave message for %s: %u.%08u\n",
		channel->name,
		(unsigned int)stasis_message_timestamp(message)->tv_sec,
		(unsigned int)stasis_message_timestamp(message)->tv_usec);

	cdr = ao2_find(active_cdrs_master, channel->uniqueid, OBJ_SEARCH_KEY);
	if (!cdr) {
		ast_log(AST_LOG_WARNING, "No CDR for channel %s\n", channel->name);
		ast_assert(0);
		return;
	}

	/* Party A */
	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (!it_cdr->fn_table->process_bridge_leave) {
			continue;
		}
		CDR_DEBUG("%p - Processing Bridge Leave for %s\n",
			it_cdr, channel->name);
		if (!it_cdr->fn_table->process_bridge_leave(it_cdr, bridge, channel)) {
			ast_string_field_set(it_cdr, bridge, "");
			left_bridge = 1;
		}

#ifdef GRANDSTREAM_NETWORKS
		/*added by wxu@grandstream.cn for attended transfer*/
		if(bridge->is_attended_transfer == 1){		//bridge is attended transfer
			if(it_cdr->last->party_a.snapshot){
				ast_log(LOG_VERBOSE,"cdr->party_a:%s",S_OR(it_cdr->last->party_a.snapshot->name,"null"));
			}

			if(it_cdr->last->party_b.snapshot){
				ast_log(LOG_VERBOSE,"cdr->party_b:%s",S_OR(it_cdr->last->party_b.snapshot->name,"null"));
			}

			if(it_cdr->party_a.snapshot){
				if(!strcmp(S_OR(it_cdr->party_a.snapshot->name,""),channel->name)){
					ast_log(LOG_VERBOSE,"set FEATURE_TRANSFER chan:%s",channel->name);
					ast_cdr_setvar(channel->name,"FEATURE_TRANSFER","yes");
				}
			}
		} else if( bridge->is_attended_transfer == 2 ) {
			if (it_cdr->last->party_b.snapshot && it_cdr->last->party_b.snapshot->name) {
				RAII_VAR(char *, cdr_partyb_sname, ast_strdup(it_cdr->last->party_b.snapshot->name), ast_free);
				RAII_VAR(char *, cdr_session, ast_strdup(it_cdr->session), ast_free);
				ao2_unlock(cdr);
				struct ast_channel *peer_chan = ast_channel_get_by_name(cdr_partyb_sname);
				ast_log(LOG_VERBOSE,"cdr->party_b:%s", S_OR(cdr_partyb_sname,"null"));
				if ( peer_chan ) {
					if (!strncmp(cdr_partyb_sname, "Local", 5)
						&&strlen(ast_channel_session(peer_chan)) == 0 && cdr_session) {
						ast_channel_lock(peer_chan);
						ast_channel_session_set_value(peer_chan, cdr_session);
						ast_channel_unlock(peer_chan);
					}
					ast_channel_unref(peer_chan);
				}
				ao2_lock(cdr);
			}
		}
		/*end*/
#endif
	}
	ao2_unlock(cdr);

	/* Party B */
	if (left_bridge
		&& strcmp(bridge->subclass, "parking")) {
		ao2_callback_data(active_cdrs_all, OBJ_NODATA | OBJ_MULTIPLE | OBJ_SEARCH_KEY,
			cdr_object_party_b_left_bridge_cb, (char *) leave_data.channel->name,
			&leave_data);
	}

	ao2_cleanup(cdr);
}

/*!
 * \internal
 * \brief Create a new CDR, append it to an existing CDR, and update its snapshots
 *
 * \note The new CDR will be automatically transitioned to the bridge state
 */
static void bridge_candidate_add_to_cdr(struct cdr_object *cdr,
		struct cdr_object_snapshot *party_b)
{
	struct cdr_object *new_cdr;

	new_cdr = cdr_object_create_and_append(cdr);
	if (!new_cdr) {
		return;
	}
#ifdef GRANDSTREAM_NETWORKS
	cdr_object_snapshot_copy(new_cdr, &new_cdr->party_b, party_b);
#else
	cdr_object_snapshot_copy(&new_cdr->party_b, party_b);
#endif
	cdr_all_relink(new_cdr);
	cdr_object_check_party_a_answer(new_cdr);
	ast_string_field_set(new_cdr, bridge, cdr->bridge);
#ifdef GRANDSTREAM_NETWORKS
	/*add by wxu@grandstream.cn for set action_type*/
	ast_string_field_set(new_cdr, action_type, cdr->action_type);
	/*end*/
	ast_string_field_set(new_cdr, special_action_type, cdr->special_action_type);
	ast_string_field_set(new_cdr, special_dst, cdr->special_dst);
	ast_string_field_set(new_cdr, inbound_trunk_index, cdr->inbound_trunk_index);
	ast_string_field_set(new_cdr, outbound_trunk_index, cdr->outbound_trunk_index);
	ast_string_field_set(new_cdr, status, cdr->status);
	if( party_b->snapshot->attend_transfered && !strncmp(party_b->snapshot->name,"Local",strlen("Local")) ){
		ast_log(LOG_VERBOSE,"cdr disable");
		ast_set_flag(&new_cdr->flags,AST_CDR_FLAG_DISABLE);
	}
#endif
	cdr_object_transition_state(new_cdr, &bridge_state_fn_table);
	CDR_DEBUG("%p - Party A %s has new Party B %s\n",
		new_cdr, new_cdr->party_a.snapshot->name,
		party_b->snapshot->name);
}

/*!
 * \brief Process a single \ref bridge_candidate
 *
 * When a CDR enters a bridge, it needs to make pairings with everyone else
 * that it is not currently paired with. This function determines, for the
 * CDR for the channel that entered the bridge and the CDR for every other
 * channel currently in the bridge, who is Party A and makes new CDRs.
 *
 * \param cdr The \ref cdr_obj being processed
 * \param cand_cdr The \ref cdr_object that is a candidate
 *
 */
static void bridge_candidate_process(struct cdr_object *cdr, struct cdr_object *base_cand_cdr)
{
	struct cdr_object_snapshot *party_a;
	struct cdr_object *cand_cdr;

	ao2_lock(base_cand_cdr);

	for (cand_cdr = base_cand_cdr; cand_cdr; cand_cdr = cand_cdr->next) {
		/* Skip any records that are not in this bridge */
		if (strcmp(cand_cdr->bridge, cdr->bridge)) {
			continue;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (strstr(cand_cdr->special_action_type, "COUNTRY_CODE") && !strstr(cand_cdr->action_type, "RINGGROUP")){
			if (NULL != strstr(cand_cdr->special_action_type, "COUNTRY_CODE/ringboth")){
				ast_log(LOG_NOTICE, "do need not reset data");
			}
			else{
				if (cand_cdr->answer.tv_sec <= cdr->answer.tv_sec){
					cand_cdr->answer.tv_sec = cdr->answer.tv_sec;
					cand_cdr->answer.tv_usec = cdr->answer.tv_usec;
				}
				if (cand_cdr->start.tv_sec <= cdr->start.tv_sec){
					cand_cdr->start.tv_sec = cdr->start.tv_sec;
					cand_cdr->start.tv_usec = cdr->start.tv_usec;
				}
				if (cand_cdr->end.tv_sec < cand_cdr->start.tv_sec || cand_cdr->end.tv_sec < cand_cdr->answer.tv_sec){
					cand_cdr->end.tv_sec = 0;
					cand_cdr->end.tv_usec = 0;
				}
			}
		}
		if (0 != strcmp(cdr->party_a.snapshot->name, cand_cdr->party_a.snapshot->name) && 0 == strcmp(cand_cdr->special_action_type, "\"SEAMLESS_TRANSFER\"")){
			cand_cdr->answer.tv_sec = cdr->answer.tv_sec;
			cand_cdr->answer.tv_usec = cdr->answer.tv_usec;
		}
		/*add for bug 109549 by ychpang*/
		if (0 == strcmp(cand_cdr->special_action_type, "VoiceMailMainCallback")){
			cand_cdr->answer.tv_sec = cdr->answer.tv_sec;
			cand_cdr->answer.tv_usec = cdr->answer.tv_usec;
		}
		if (strstr(cand_cdr->party_a.snapshot->name, "DAHDI") && 0 == strcmp(cand_cdr->special_action_type, "NOTANSWER") && ast_strlen_zero(cand_cdr->inbound_trunk_index)
			&& cand_cdr->answer.tv_sec != cdr->answer.tv_sec && cand_cdr->answer.tv_usec != cdr->answer.tv_usec){
			cand_cdr->answer.tv_sec = cdr->answer.tv_sec;
			cand_cdr->answer.tv_usec = cdr->answer.tv_usec;
			ast_cdr_setvar(cand_cdr->party_a.snapshot->name, "special_action_type", "");
		}

		/* Start added by zyxia for bug 113392 emptyqueue  */
		if ( 0 == strcmp(cand_cdr->special_action_type, "EMPTYQUEUE")
				&& cand_cdr->answer.tv_sec != cdr->answer.tv_sec && cand_cdr->answer.tv_usec != cdr->answer.tv_usec){
				cand_cdr->answer.tv_sec = cdr->answer.tv_sec;
				cand_cdr->answer.tv_usec = cdr->answer.tv_usec;
				ast_cdr_setvar(cand_cdr->party_a.snapshot->name, "special_action_type", "");
		}
		/* End added by zyxia for bug 113392 emptyqueue */

		/* add for bug 60257 - 60259, add by pchwu /110127*/
		long int ms;
		ms = ast_tvdiff_ms(cdr->answer, cand_cdr->answer);
		if (ms > 0
			&& ((!strncmp(cand_cdr->action_type, "QUEUE", strlen("QUEUE")) && NULL == strstr(cand_cdr->special_action_type, "atx_refer") && NULL == strstr(cdr->party_a.snapshot->name, "from-internal-xfer"))
				|| !strncmp(cand_cdr->action_type, "DIRECTORY", strlen("DIRECTORY"))
				|| (!strncmp(cand_cdr->action_type, "IVR", strlen("IVR")) && !strncmp(cdr->action_type, "IVR", strlen("IVR"))))){
			cand_cdr->answer.tv_sec = cdr->answer.tv_sec;
			cand_cdr->answer.tv_usec = cdr->answer.tv_usec;
		}
		if (ms < 0 && !strncmp(cdr->action_type, "IVR", strlen("IVR")) && ast_strlen_zero(cand_cdr->action_type)){
			ast_log(LOG_DEBUG, "later cdr's answer time must be later, set the answer time");
			cdr->answer.tv_sec = cand_cdr->answer.tv_sec;
			cdr->answer.tv_usec = cand_cdr->answer.tv_usec;
		}
#endif

		/* If the candidate is us or someone we've taken on, pass on by */
		if (!strcasecmp(cdr->party_a.snapshot->name, cand_cdr->party_a.snapshot->name)
			|| (cdr->party_b.snapshot
				&& !strcasecmp(cdr->party_b.snapshot->name, cand_cdr->party_a.snapshot->name))) {
			break;
		}

		party_a = cdr_object_pick_party_a(&cdr->party_a, &cand_cdr->party_a);
		/* We're party A - make a new CDR, append it to us, and set the candidate as
		 * Party B */
		if (!strcasecmp(party_a->snapshot->name, cdr->party_a.snapshot->name)) {
			bridge_candidate_add_to_cdr(cdr, &cand_cdr->party_a);
			break;
		}

		/* We're Party B. Check if we can add ourselves immediately or if we need
		 * a new CDR for them (they already have a Party B) */
		if (cand_cdr->party_b.snapshot
			&& strcasecmp(cand_cdr->party_b.snapshot->name, cdr->party_a.snapshot->name)) {
			bridge_candidate_add_to_cdr(cand_cdr, &cdr->party_a);
		} else {
			CDR_DEBUG("%p - Party A %s has new Party B %s\n",
				cand_cdr, cand_cdr->party_a.snapshot->name,
				cdr->party_a.snapshot->name);
#ifdef GRANDSTREAM_NETWORKS
			if( !strncmp(cdr->party_a.snapshot->name,"Local",strlen("Local"))
				&& strlen(cdr->record_files) > 0 && strlen(cand_cdr->record_files) == 0 ){
				CDR_DEBUG("cdr->record_files:%s",cdr->record_files);
				snprintf(cand_cdr->record_files,sizeof(cand_cdr->record_files),"%s",cdr->record_files);
			}
			cdr_object_snapshot_copy(cand_cdr, &cand_cdr->party_b, &cdr->party_a);
#else
			cdr_object_snapshot_copy(&cand_cdr->party_b, &cdr->party_a);
#endif
			cdr_all_relink(cand_cdr);
			/* It's possible that this joined at one point and was never chosen
			 * as party A. Clear their end time, as it would be set in such a
			 * case.
			 */
			memset(&cand_cdr->end, 0, sizeof(cand_cdr->end));
		}

		break;
	}

	ao2_unlock(base_cand_cdr);
}

/*!
 * \brief Handle creating bridge pairings for the \ref cdr_object that just
 * entered a bridge
 * \param cdr The \ref cdr_object that just entered the bridge
 * \param bridge The \ref ast_bridge_snapshot representing the bridge it just entered
 */
static void handle_bridge_pairings(struct cdr_object *cdr, struct ast_bridge_snapshot *bridge)
{
	struct ao2_iterator it_channels;
	char *channel_id;

	it_channels = ao2_iterator_init(bridge->channels, 0);
	while ((channel_id = ao2_iterator_next(&it_channels))) {
		struct cdr_object *cand_cdr;

		cand_cdr = ao2_find(active_cdrs_master, channel_id, OBJ_SEARCH_KEY);
		if (cand_cdr) {
			bridge_candidate_process(cdr, cand_cdr);
			ao2_ref(cand_cdr, -1);
		}

		ao2_ref(channel_id, -1);
	}
	ao2_iterator_destroy(&it_channels);
}

/*! \brief Handle entering into a parking bridge
 * \param cdr The CDR to operate on
 * \param bridge The bridge the channel just entered
 * \param channel The channel snapshot
 */
static void handle_parking_bridge_enter_message(struct cdr_object *cdr,
		struct ast_bridge_snapshot *bridge,
		struct ast_channel_snapshot *channel)
{
	int res = 1;
	struct cdr_object *it_cdr;
	struct cdr_object *new_cdr;

	ao2_lock(cdr);

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->fn_table->process_parking_bridge_enter) {
			res &= it_cdr->fn_table->process_parking_bridge_enter(it_cdr, bridge, channel);
		}
		if (it_cdr->fn_table->process_party_a) {
			CDR_DEBUG("%p - Updating Party A %s snapshot\n", it_cdr,
				channel->name);
			it_cdr->fn_table->process_party_a(it_cdr, channel);
		}
	}

	if (res) {
		/* No one handled it - we need a new one! */
		new_cdr = cdr_object_create_and_append(cdr);
		if (new_cdr) {
			/* Let the single state transition us to Parked */
			cdr_object_transition_state(new_cdr, &single_state_fn_table);
			new_cdr->fn_table->process_parking_bridge_enter(new_cdr, bridge, channel);
		}
	}
	ao2_unlock(cdr);
}

/*! \brief Handle a bridge enter message for a 'normal' bridge
 * \param cdr The CDR to operate on
 * \param bridge The bridge the channel just entered
 * \param channel The channel snapshot
 */
static void handle_standard_bridge_enter_message(struct cdr_object *cdr,
		struct ast_bridge_snapshot *bridge,
		struct ast_channel_snapshot *channel)
{
	enum process_bridge_enter_results result;
	struct cdr_object *it_cdr;
	struct cdr_object *new_cdr;
	struct cdr_object *handled_cdr = NULL;

	ao2_lock(cdr);

try_again:
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->fn_table->process_party_a) {
			CDR_DEBUG("%p - Updating Party A %s snapshot\n", it_cdr,
				channel->name);
			it_cdr->fn_table->process_party_a(it_cdr, channel);
		}

		/* Notify all states that they have entered a bridge */
		if (it_cdr->fn_table->process_bridge_enter) {
			CDR_DEBUG("%p - Processing bridge enter for %s\n", it_cdr,
				channel->name);
			result = it_cdr->fn_table->process_bridge_enter(it_cdr, bridge, channel);
			switch (result) {
			case BRIDGE_ENTER_ONLY_PARTY:
				/* Fall through */
			case BRIDGE_ENTER_OBTAINED_PARTY_B:
				if (!handled_cdr) {
					handled_cdr = it_cdr;
				}
				break;
			case BRIDGE_ENTER_NEED_CDR:
				/* Pass */
				break;
			case BRIDGE_ENTER_NO_PARTY_B:
				/* We didn't win on any - end this CDR. If someone else comes in later
				 * that is Party B to this CDR, it can re-activate this CDR.
				 */
				if (!handled_cdr) {
					handled_cdr = it_cdr;
				}
				cdr_object_finalize(cdr);
				break;
			}
		}
	}

	/* Create the new matchings, but only for either:
	 *  * The first CDR in the chain that handled it. This avoids issues with
	 *    forked CDRs.
	 *  * If no one handled it, the last CDR in the chain. This would occur if
	 *    a CDR joined a bridge and it wasn't Party A for anyone. We still need
	 *    to make pairings with everyone in the bridge.
	 */
	if (handled_cdr) {
#ifdef GRANDSTREAM_NETWORKS //add by zhrliu for bug 44535
		if(NULL==strcasestr(handled_cdr->action_type,"CONFERENCE") && NULL==strcasestr(handled_cdr->action_type,"MEETING")){
			handle_bridge_pairings(handled_cdr, bridge);
		}
#else
		handle_bridge_pairings(handled_cdr, bridge);
#endif
	} else {
		/* Nothing handled it - we need a new one! */
		new_cdr = cdr_object_create_and_append(cdr);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_VERBOSE,"cdr->name:%s",S_OR(cdr->name,"null"));
#endif
		if (new_cdr) {
			/* This is guaranteed to succeed: the new CDR is created in the single state
			 * and will be able to handle the bridge enter message
			 */
			/*need to check, does this really need after update to v.16*/
#ifdef GRANDSTREAM_NETWORKS
			if(cdr->last->party_a.snapshot){
				ast_log(LOG_VERBOSE,"cdr->party_a:%s",S_OR(cdr->last->party_a.snapshot->name,"null"));
				if( cdr->last->party_a.snapshot->attend_transfered
					&& !strncmp(S_OR(cdr->last->party_a.snapshot->name,"null"),"Local",strlen("Local")) ){
					ast_log(LOG_VERBOSE,"cdr disable");
					ast_set_flag(&new_cdr->flags,AST_CDR_FLAG_DISABLE);
				}
			}

			if(cdr->last->party_b.snapshot){
				ast_log(LOG_VERBOSE,"cdr->party_b:%s",S_OR(cdr->last->party_b.snapshot->name,"null"));
			}

			/*added by wxu@grandstream.cn for attended transfer callee *2 and C hangup*/
			if( bridge->is_attended_transfer ){
				if(bridge->is_attended_transfer == 1){
					ast_log(LOG_VERBOSE,"set FEATURE_TRANSFER chan:%s",S_OR(cdr->last->party_a.snapshot->name,"null"));
					memset(new_cdr->feature_transfer,0,sizeof(new_cdr->feature_transfer));
					snprintf(new_cdr->feature_transfer,sizeof(new_cdr->feature_transfer)-1,"%s","yes");
				}
				ast_string_field_set(new_cdr, action_type,"TRANSFER");
			}
			/*end*/
#endif

			goto try_again;
		}
	}
	ao2_unlock(cdr);
}

/*!
 * \internal
 * \brief Handler for Stasis-Core bridge enter messages
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message The message - hopefully a bridge one!
 */
static void handle_bridge_enter_message(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	struct ast_bridge_blob *update = stasis_message_data(message);
	struct ast_bridge_snapshot *bridge = update->bridge;
	struct ast_channel_snapshot *channel = update->channel;
	struct cdr_object *cdr;

	if (filter_bridge_messages(bridge)) {
		return;
	}

	if (filter_channel_snapshot(channel)) {
		return;
	}

	CDR_DEBUG("Bridge Enter message for channel %s: %u.%08u\n",
		channel->name,
		(unsigned int)stasis_message_timestamp(message)->tv_sec,
		(unsigned int)stasis_message_timestamp(message)->tv_usec);

	cdr = ao2_find(active_cdrs_master, channel->uniqueid, OBJ_SEARCH_KEY);
	if (!cdr) {
		ast_log(AST_LOG_WARNING, "No CDR for channel %s\n", channel->name);
		ast_assert(0);
		return;
	}

	if (!strcmp(bridge->subclass, "parking")) {
		handle_parking_bridge_enter_message(cdr, bridge, channel);
	} else {
		handle_standard_bridge_enter_message(cdr, bridge, channel);
	}
	ao2_cleanup(cdr);
}

/*!
 * \brief Handler for when a channel is parked
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message The message about who got parked
 * */
static void handle_parked_call_message(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	struct ast_parked_call_payload *payload = stasis_message_data(message);
	struct ast_channel_snapshot *channel = payload->parkee;
	struct cdr_object *cdr;
	int unhandled = 1;
	struct cdr_object *it_cdr;

	/* Anything other than getting parked will be handled by other updates */
	if (payload->event_type != PARKED_CALL) {
		return;
	}

	/* No one got parked? */
	if (!channel) {
		return;
	}

	if (filter_channel_snapshot(channel)) {
		return;
	}

	CDR_DEBUG("Parked Call message for channel %s: %u.%08u\n",
		channel->name,
		(unsigned int)stasis_message_timestamp(message)->tv_sec,
		(unsigned int)stasis_message_timestamp(message)->tv_usec);

	cdr = ao2_find(active_cdrs_master, channel->uniqueid, OBJ_SEARCH_KEY);
	if (!cdr) {
		ast_log(AST_LOG_WARNING, "No CDR for channel %s\n", channel->name);
		ast_assert(0);
		return;
	}

	ao2_lock(cdr);

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->fn_table->process_parked_channel) {
			unhandled &= it_cdr->fn_table->process_parked_channel(it_cdr, payload);
		}
	}

	if (unhandled) {
		/* Nothing handled the messgae - we need a new one! */
		struct cdr_object *new_cdr;

		new_cdr = cdr_object_create_and_append(cdr);
		if (new_cdr) {
			/* As the new CDR is created in the single state, it is guaranteed
			 * to have a function for the parked call message and will handle
			 * the message */
			new_cdr->fn_table->process_parked_channel(new_cdr, payload);
		}
	}

	ao2_unlock(cdr);

	ao2_cleanup(cdr);
}

#ifdef GRANDSTREAM_NETWORKS
static void handle_mixmonitor_message(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	char *record;
	struct cdr_object *cdr;
	struct ao2_iterator it_cdr;
	struct ast_mixmonitor_event_payload *payload = stasis_message_data(message);

	if (ast_strlen_zero(payload->record)) {
		return;
	}

	record = strstr(payload->record, "auto-");
	if (!record) {
		return;
	}

	ast_log(LOG_DEBUG, "Enter handle_mixmonitor_message, type:%d, record:%s",
		payload->event_type, record);

	it_cdr = ao2_iterator_init(active_cdrs_master, 0);
	while ((cdr = ao2_iterator_next(&it_cdr))) {
		ao2_lock(cdr);
		ast_log(LOG_DEBUG, "cdr(%p|%s), record:%s", cdr, cdr->name, cdr->record_files);
		if (strstr(cdr->record_files, record)) {
			ast_log(LOG_DEBUG, "Find match cdr, unlink from active_cdrs_master");
			cdr_all_unlink(cdr);
			ao2_unlink(active_cdrs_master, cdr);
		}

		ao2_unlock(cdr);
		ao2_cleanup(cdr);
	}
	ao2_iterator_destroy(&it_cdr);
}
#endif

/*!
 * \brief Handler for a synchronization message
 * \param data Passed on
 * \param sub The stasis subscription for this message callback
 * \param topic The topic this message was published for
 * \param message A blank ao2 object
 * */
static void handle_cdr_sync_message(void *data, struct stasis_subscription *sub,
		struct stasis_message *message)
{
	return;
}

struct ast_cdr_config *ast_cdr_get_config(void)
{
	struct ast_cdr_config *general;
	struct module_config *mod_cfg;

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return NULL;
	}
	general = ao2_bump(mod_cfg->general);
	ao2_cleanup(mod_cfg);
	return general;
}

void ast_cdr_set_config(struct ast_cdr_config *config)
{
	struct module_config *mod_cfg;

	if (!config) {
		return;
	}

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return;
	}

	ao2_replace(mod_cfg->general, config);

	cdr_set_debug_mode(mod_cfg);
	cdr_toggle_runtime_options();

	ao2_cleanup(mod_cfg);
}

int ast_cdr_is_enabled(void)
{
	return is_cdr_flag_set(CDR_ENABLED);
}

int ast_cdr_backend_suspend(const char *name)
{
	int success = -1;
	struct cdr_beitem *i = NULL;

	AST_RWLIST_WRLOCK(&be_list);
	AST_RWLIST_TRAVERSE(&be_list, i, list) {
		if (!strcasecmp(name, i->name)) {
			ast_debug(3, "Suspending CDR backend %s\n", i->name);
			i->suspended = 1;
			success = 0;
		}
	}
	AST_RWLIST_UNLOCK(&be_list);

	return success;
}

int ast_cdr_backend_unsuspend(const char *name)
{
	int success = -1;
	struct cdr_beitem *i = NULL;

	AST_RWLIST_WRLOCK(&be_list);
	AST_RWLIST_TRAVERSE(&be_list, i, list) {
		if (!strcasecmp(name, i->name)) {
			ast_debug(3, "Unsuspending CDR backend %s\n", i->name);
			i->suspended = 0;
			success = 0;
		}
	}
	AST_RWLIST_UNLOCK(&be_list);

	return success;
}

static int cdr_generic_register(struct be_list *generic_list, const char *name, const char *desc, ast_cdrbe be)
{
	struct cdr_beitem *i;
	struct cdr_beitem *cur;

	if (!name) {
		return -1;
	}

	if (!be) {
		ast_log(LOG_WARNING, "CDR engine '%s' lacks backend\n", name);

		return -1;
	}

	i = ast_calloc(1, sizeof(*i));
	if (!i) {
		return -1;
	}

	i->be = be;
	ast_copy_string(i->name, name, sizeof(i->name));
	ast_copy_string(i->desc, desc, sizeof(i->desc));

	AST_RWLIST_WRLOCK(generic_list);
	AST_RWLIST_TRAVERSE(generic_list, cur, list) {
		if (!strcasecmp(name, cur->name)) {
			ast_log(LOG_WARNING, "Already have a CDR backend called '%s'\n", name);
			AST_RWLIST_UNLOCK(generic_list);
			ast_free(i);

			return -1;
		}
	}

	AST_RWLIST_INSERT_HEAD(generic_list, i, list);
	AST_RWLIST_UNLOCK(generic_list);

	return 0;
}

int ast_cdr_register(const char *name, const char *desc, ast_cdrbe be)
{
	return cdr_generic_register(&be_list, name, desc, be);
}

int ast_cdr_modifier_register(const char *name, const char *desc, ast_cdrbe be)
{
	return cdr_generic_register((struct be_list *)&mo_list, name, desc, be);
}

int ast_cdr_sql_register(ast_cdr_sql_backend sqlbe)
{
    cdr_sql_be.sql_be = sqlbe;
	return 0;
}

int ast_cdr_sql_unregister(void)
{
    cdr_sql_be.sql_be = NULL;
	return 0;
}

void ast_cdr_sql_insert(char *sql)
{
	if(cdr_sql_be.sql_be)
	{
		cdr_sql_be.sql_be(sql);
	}
	
	return;
}


static int ast_cdr_generic_unregister(struct be_list *generic_list, const char *name)
{
	struct cdr_beitem *match = NULL;
	int active_count;

	AST_RWLIST_WRLOCK(generic_list);
	AST_RWLIST_TRAVERSE(generic_list, match, list) {
		if (!strcasecmp(name, match->name)) {
			break;
		}
	}

	if (!match) {
		AST_RWLIST_UNLOCK(generic_list);
		return 0;
	}

	active_count = ao2_container_count(active_cdrs_master);

	if (!match->suspended && active_count != 0) {
		AST_RWLIST_UNLOCK(generic_list);
		ast_log(AST_LOG_WARNING, "Unable to unregister CDR backend %s; %d CDRs are still active\n",
			name, active_count);
		return -1;
	}

	AST_RWLIST_REMOVE(generic_list, match, list);
	AST_RWLIST_UNLOCK(generic_list);

	ast_verb(2, "Unregistered '%s' CDR backend\n", name);
	ast_free(match);

	return 0;
}

int ast_cdr_unregister(const char *name)
{
	return ast_cdr_generic_unregister(&be_list, name);
}

int ast_cdr_modifier_unregister(const char *name)
{
	return ast_cdr_generic_unregister((struct be_list *)&mo_list, name);
}

struct ast_cdr *ast_cdr_dup(struct ast_cdr *cdr)
{
	struct ast_cdr *newcdr;

	if (!cdr) {
		return NULL;
	}
	newcdr = ast_cdr_alloc();
	if (!newcdr) {
		return NULL;
	}

	*newcdr = *cdr;
	AST_LIST_HEAD_INIT_NOLOCK(&newcdr->varshead);
	copy_variables(&newcdr->varshead, &cdr->varshead);
	newcdr->next = NULL;

	return newcdr;
}

static const char *cdr_format_var_internal(struct ast_cdr *cdr, const char *name)
{
	struct ast_var_t *variables;

	if (ast_strlen_zero(name)) {
		return NULL;
	}

	AST_LIST_TRAVERSE(&cdr->varshead, variables, entries) {
		if (!strcasecmp(name, ast_var_name(variables))) {
			return ast_var_value(variables);
		}
	}

	return NULL;
}

static void cdr_get_tv(struct timeval when, const char *fmt, char *buf, int bufsize)
{
	if (fmt == NULL) {	/* raw mode */
		snprintf(buf, bufsize, "%ld.%06ld", (long)when.tv_sec, (long)when.tv_usec);
	} else {
		buf[0] = '\0';/* Ensure the buffer is initialized. */
		if (when.tv_sec) {
			struct ast_tm tm;

			ast_localtime(&when, &tm, NULL);
			ast_strftime(buf, bufsize, fmt, &tm);
		}
	}
}

void ast_cdr_format_var(struct ast_cdr *cdr, const char *name, char **ret, char *workspace, int workspacelen, int raw)
{
	const char *fmt = "%Y-%m-%d %T";
	const char *varbuf;

	if (!cdr) {
		return;
	}

	*ret = NULL;

	if (!strcasecmp(name, "clid")) {
		ast_copy_string(workspace, cdr->clid, workspacelen);
	} else if (!strcasecmp(name, "src")) {
		ast_copy_string(workspace, cdr->src, workspacelen);
	} else if (!strcasecmp(name, "dst")) {
		ast_copy_string(workspace, cdr->dst, workspacelen);
	} else if (!strcasecmp(name, "dcontext")) {
		ast_copy_string(workspace, cdr->dcontext, workspacelen);
	} else if (!strcasecmp(name, "channel")) {
		ast_copy_string(workspace, cdr->channel, workspacelen);
	} else if (!strcasecmp(name, "dstchannel")) {
		ast_copy_string(workspace, cdr->dstchannel, workspacelen);
	} else if (!strcasecmp(name, "lastapp")) {
		ast_copy_string(workspace, cdr->lastapp, workspacelen);
	} else if (!strcasecmp(name, "lastdata")) {
		ast_copy_string(workspace, cdr->lastdata, workspacelen);
	} else if (!strcasecmp(name, "start")) {
		cdr_get_tv(cdr->start, raw ? NULL : fmt, workspace, workspacelen);
	} else if (!strcasecmp(name, "answer")) {
		cdr_get_tv(cdr->answer, raw ? NULL : fmt, workspace, workspacelen);
	} else if (!strcasecmp(name, "end")) {
		cdr_get_tv(cdr->end, raw ? NULL : fmt, workspace, workspacelen);
	} else if (!strcasecmp(name, "duration")) {
		snprintf(workspace, workspacelen, "%ld", cdr->end.tv_sec != 0 ? cdr->duration : (long)ast_tvdiff_ms(ast_tvnow(), cdr->start) / 1000);
	} else if (!strcasecmp(name, "billsec")) {
		snprintf(workspace, workspacelen, "%ld", (cdr->billsec || !ast_tvzero(cdr->end) || ast_tvzero(cdr->answer)) ? cdr->billsec : (long)ast_tvdiff_ms(ast_tvnow(), cdr->answer) / 1000);
	} else if (!strcasecmp(name, "disposition")) {
		if (raw) {
			snprintf(workspace, workspacelen, "%ld", cdr->disposition);
		} else {
			ast_copy_string(workspace, ast_cdr_disp2str(cdr->disposition), workspacelen);
		}
	} else if (!strcasecmp(name, "amaflags")) {
		if (raw) {
			snprintf(workspace, workspacelen, "%ld", cdr->amaflags);
		} else {
			ast_copy_string(workspace, ast_channel_amaflags2string(cdr->amaflags), workspacelen);
		}
	} else if (!strcasecmp(name, "accountcode")) {
		ast_copy_string(workspace, cdr->accountcode, workspacelen);
	} else if (!strcasecmp(name, "peeraccount")) {
		ast_copy_string(workspace, cdr->peeraccount, workspacelen);
	} else if (!strcasecmp(name, "uniqueid")) {
		ast_copy_string(workspace, cdr->uniqueid, workspacelen);
	} else if (!strcasecmp(name, "linkedid")) {
		ast_copy_string(workspace, cdr->linkedid, workspacelen);
	} else if (!strcasecmp(name, "userfield")) {
		ast_copy_string(workspace, cdr->userfield, workspacelen);
#ifdef GRANDSTREAM_NETWORKS
	/*add by wxu@grandstream.cn*/
	} else if (!strcasecmp(name, "session")) {
		ast_copy_string(workspace, cdr->session, workspacelen);
		/*add by wxu,  add two words for cdr table*/
	} else if (!strcasecmp(name, "action_owner")) {
		ast_copy_string(workspace, cdr->action_owner, workspacelen);
	} else if (!strcasecmp(name, "action_type")) {
		ast_copy_string(workspace, cdr->action_type, workspacelen);
		/*add end*/
	} else if (!strcasecmp(name, "src_trunk_name")) {
		ast_copy_string(workspace, cdr->src_trunk_name, workspacelen);
	} else if (!strcasecmp(name, "dst_trunk_name")) {
		ast_copy_string(workspace, cdr->dst_trunk_name, workspacelen);
	} else if (!strcasecmp(name, "recordfiles")) {
		ast_copy_string(workspace, cdr->record_files, workspacelen);
	} else if (!strcasecmp(name, "recordfiles_desc")) {
		ast_copy_string(workspace, cdr->record_files_desc, workspacelen);
	} else if (!strcasecmp(name, "special_action_type")) {
		ast_copy_string(workspace, cdr->special_action_type, workspacelen);
	} else if (!strcasecmp(name, "special_dst")) {
		ast_copy_string(workspace, cdr->special_dst, workspacelen);
	} else if (!strcasecmp(name, "inbound_trunk_index")) {
		ast_copy_string(workspace, cdr->inbound_trunk_index, workspacelen);
	} else if (!strcasecmp(name, "outbound_trunk_index")) {
		ast_copy_string(workspace, cdr->outbound_trunk_index, workspacelen);
	} else if (!strcasecmp(name, "device_info")) {
		ast_copy_string(workspace, cdr->device_info, workspacelen);
	} else if (!strcasecmp(name, "device_info_peer")) {
		ast_copy_string(workspace, cdr->device_info_peer, workspacelen);
	} else if (!strcasecmp(name, "status")) {
		ast_copy_string(workspace, cdr->status, workspacelen);
	/*add end*/
	/* added by zgmo for nat call */
	} else if (!strcasecmp(name, "nat_call")) {
		if (-1 == cdr->is_nat_call){
			memset(workspace, 0, workspacelen);
		}else{
			ast_copy_string(workspace, cdr->is_nat_call ? "yes" : "no", workspacelen);
		}
	} else if (!strcasecmp(name, "nat_call_peer")) {
		if (-1 == cdr->peer_is_nat_call){
			memset(workspace, 0, workspacelen);
		}else{
			ast_copy_string(workspace, cdr->peer_is_nat_call ? "yes" : "no", workspacelen);
		}
	} else if (!strcasecmp(name, "video_call")) {
		ast_copy_string(workspace, cdr->is_video_call ? "yes" : "no", workspacelen);
	/* add end */
	} else if (!strcasecmp(name, "gds_call")) {
		ast_copy_string(workspace, cdr->is_gds_call ? "yes" : "no", workspacelen);
	} else if (!strcasecmp(name, "gsc_call")) {
		ast_copy_string(workspace, cdr->is_gsc_call ? "yes" : "no", workspacelen);
	} else if (!strcasecmp(name, "anonymous_call")) {
		ast_copy_string(workspace, cdr->is_anonymous_call ? "yes" : "no", workspacelen);
	} else if (!strcasecmp(name, "reason")) {
		ast_copy_string(workspace, cdr->reason, workspacelen);
	} else if (!strcasecmp(name, "new_src")) {
		ast_copy_string(workspace, cdr->new_src, workspacelen);
#endif
	} else if (!strcasecmp(name, "sequence")) {
		snprintf(workspace, workspacelen, "%d", cdr->sequence);
	} else if ((varbuf = cdr_format_var_internal(cdr, name))) {
		ast_copy_string(workspace, varbuf, workspacelen);
	} else {
		workspace[0] = '\0';
	}

	if (!ast_strlen_zero(workspace)) {
		*ret = workspace;
	}
}

/*!
 * \internal
 * \brief Callback that finds all CDRs that reference a particular channel by name
 */
static int cdr_object_select_all_by_name_cb(void *obj, void *arg, int flags)
{
	struct cdr_object *cdr = obj;
	const char *name = arg;

#ifdef GRANDSTREAM_NETWORKS
	/* will case the call delay, remove this change */
	SCOPED_MUTEX(lock, &cdr->party_lock);
#endif

	if (!strcasecmp(cdr->party_a.snapshot->name, name) ||
			(cdr->party_b.snapshot && !strcasecmp(cdr->party_b.snapshot->name, name))) {
		return CMP_MATCH;
	}

	return 0;
}

/*!
 * \internal
 * \brief Callback that finds a CDR by channel name
 */
static int cdr_object_get_by_name_cb(void *obj, void *arg, int flags)
{
	struct cdr_object *cdr = obj;
	const char *name = arg;

#ifdef GRANDSTREAM_NETWORKS
	/* will case the call delay, remove this change */
	SCOPED_MUTEX(lock, &cdr->party_lock);
#endif

	if (!strcasecmp(cdr->party_a.snapshot->name, name)) {
		return CMP_MATCH;
	}
	return 0;
}

/* Read Only CDR variables */
static const char * const cdr_readonly_vars[] = {
	"clid",
	"src",
	"dst",
	"dcontext",
	"channel",
	"dstchannel",
	"lastapp",
	"lastdata",
	"start",
	"answer",
	"end",
	"duration",
	"billsec",
	"disposition",
	"amaflags",
	"accountcode",
	"uniqueid",
	"linkedid",
	"userfield",
	"sequence",
	NULL
};

int ast_cdr_setvar(const char *channel_name, const char *name, const char *value)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);
	int x;

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
			ast_log(LOG_ERROR, "Attempt to set the '%s' read-only variable!\n", name);
			return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);
		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;

			if (it_cdr->fn_table == &finalized_state_fn_table && it_cdr->next != NULL) {

/*Start added by zyxia for Bug 244700 ,2022.12.8*/
#ifdef GRANDSTREAM_NETWORKS
				if(strstr(cdr->action_type, "PAGE[") && !strcasecmp(name,"record_files_desc")) {
					ast_log(LOG_DEBUG,"gcdr(%p)[%s], cdr(%p)[%s] need set record_files_desc",
						cdr, S_OR(cdr->name,"null"),
						it_cdr, S_OR(it_cdr->name,"null"));
				}
				else
					continue;
#else
					continue;
#endif
/*End added by zyxia for Bug 244700 ,2022.12.8*/
			}
			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
#ifdef GRANDSTREAM_NETWORKS
			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			if(ast_set_cdr_obj_vars(cdr, channel_name, it_cdr, name, value)){
				if (headp) {
					set_variable(headp, name, value);
				}

			}
			/*end*/
#else
			if (headp) {
				set_variable(headp, name, value);
			}
#endif
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
int ast_cdr_oldest_setvar(const char *channel_name, const char *name, const char *value)
{
	int x;
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);
	

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
			ast_log(LOG_ERROR, "Attempt to set the '%s' read-only variable!\n", name);
			return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);
		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;

			ast_mutex_lock(&it_cdr->party_lock);   /*Add by weidu, for Bug 263475*/
			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
			ast_mutex_unlock(&it_cdr->party_lock);  /*Add by weidu, for Bug 263475*/
			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			ret = ast_set_cdr_obj_vars(cdr, channel_name,it_cdr,name,value);
			if(ret){
				if (headp) {
					set_variable(headp, name, value);
				}
			}
			/*end*/

			break;
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return ret;
}

int ast_cdr_newest_setvar(const char *channel_name, const char *name, const char *value)
{
	int x;
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);
	int right_reason = 0;
	int answlen = strlen(ANSWERED_ELSEWHERE);

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
			ast_log(LOG_ERROR, "Attempt to set the '%s' read-only variable!\n", name);
			return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name))
		{
			ast_log(AST_LOG_DEBUG, "Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;

			ast_mutex_lock(&it_cdr->party_lock);   /*Add by weidu, for Bug 263475*/
			if (0 == strcmp(name, "reason") 
				&& (strlen(value) > answlen && 0 == strncmp(value, ANSWERED_ELSEWHERE, answlen))
				&& (it_cdr->party_b.snapshot && 0 == strcmp(it_cdr->party_b.snapshot->name, strstr(value, ANSWERED_ELSEWHERE) + answlen))){
					ast_log(LOG_DEBUG, "Contact ext has more than one CDR, cause the newest cdr get err probabilistically");
					right_reason = 1;
			} else{
				right_reason = 0;
			}

			if (it_cdr->next != NULL && !right_reason) {
				ast_mutex_unlock(&it_cdr->party_lock);	/*Add by weidu, for Bug 263475*/
				continue;
			}

			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
			ast_mutex_unlock(&it_cdr->party_lock);  /*Add by weidu, for Bug 263475*/
			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			ret = ast_set_cdr_obj_vars(cdr, channel_name,it_cdr,name,value);
			if(ret) {
				if (headp) {
					set_variable(headp, name, value);
				}
			}
			/*end*/
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return ret;
}

int ast_cdr_last_newest_setvar(const char *channel_name, const char *name, const char *value)
{
	int x;
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
			ast_log(LOG_ERROR, "Attempt to set the '%s' read-only variable!\n", name);
			return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name))
		{
			ast_log(AST_LOG_DEBUG, "Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;

			if (it_cdr->next != NULL && it_cdr->next->next != NULL) {
				continue;
			}
			
			ast_mutex_lock(&it_cdr->party_lock);   /*Add by weidu, for Bug 263475*/
			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
			ast_mutex_unlock(&it_cdr->party_lock);  /*Add by weidu, for Bug 263475*/
			
			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			ret = ast_set_cdr_obj_vars(cdr, channel_name,it_cdr,name,value);
			if (ret) {
				if (headp) {
					set_variable(headp, name, value);
				}
			}
			/*end*/
			break;
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return ret;
}

int ast_cdr_all_setvar(const char *channel_name, const char *name, const char *value)
{
	int x;
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
			ast_log(LOG_ERROR, "Attempt to set the '%s' read-only variable!\n", name);
			return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name))
		{
			ast_log(AST_LOG_DEBUG, "Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;

			ast_mutex_lock(&it_cdr->party_lock); /*Added by pyzhong for Bug212881*/
			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
			ast_mutex_unlock(&it_cdr->party_lock);/*Added by pyzhong for Bug212881*/

			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			ret = ast_set_cdr_obj_vars(cdr, channel_name,it_cdr,name,value);
			if(ret) {
				if (headp) {
					set_variable(headp, name, value);
				}
			}
			/*end*/
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return ret;
}

int ast_cdr_all_except_newest_setvar(const char *channel_name, const char *name, const char *value)
{
	int x;
	int ret = 0;
	char *arg = ast_strdupa(channel_name);
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
			ast_log(LOG_ERROR, "Attempt to set the '%s' read-only variable!\n", name);
			return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name))
		{
			ast_log(AST_LOG_DEBUG, "Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;

			if (!it_cdr->next) {
				break;
			}

			if (name && 0 == strcasecmp(name, "record_files_desc") && !ast_strlen_zero(value)
				&& !ast_strlen_zero(it_cdr->record_files_desc)){
				char tmp_format[32] = {0};
				char start_time[32] = {0};
				snprintf(tmp_format, sizeof(tmp_format), "%%*[^-]-time=%%%lu[0-9]@", sizeof(start_time));
				sscanf(value, tmp_format, start_time);
				/*this mean the desc data is not the right data for rd file*/
				if (NULL == strstr(it_cdr->record_files, start_time)){
					continue;
				}
			}

			ast_mutex_lock(&it_cdr->party_lock); /*Added by pyzhong for Bug212881*/
			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
			ast_mutex_unlock(&it_cdr->party_lock);/*Added by pyzhong for Bug212881*/

			ret = ast_set_cdr_obj_vars(cdr, channel_name,it_cdr,name,value);
			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			if(ret) {
				if (headp) {
					set_variable(headp, name, value);
				}
			}
			/*end*/
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return ret;
}


int ast_cdr_set_key_by_channel_and_file(const char *channel_name, const char *file, const char *name, const char *value) {
	int x;
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);
	char file_to_compare[256] = "";

	if ((!channel_name || !strlen(channel_name)) || (!file || !strlen(file)) || !name || !value) {
	    ast_log(LOG_ERROR, "[ast_cdr_set_key_by_channel_and_file] Invalid params");
		return -1;
	}
	snprintf(file_to_compare, sizeof(file_to_compare), "%s@", file);

	for (x = 0; cdr_readonly_vars[x]; x++) {
		if (!strcasecmp(name, cdr_readonly_vars[x])) {
		    ast_log(LOG_ERROR, "[ast_cdr_set_key_by_channel_and_file] Attempt to set the '%s' read-only variable!\n", name);
		    return -1;
		}
	}

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
	    ast_log(AST_LOG_ERROR, "[ast_cdr_set_key_by_channel_and_file] Unable to find CDR for channel %s\n", channel_name);
	    return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name)) {
			ast_log(AST_LOG_DEBUG, "[ast_cdr_set_key_by_channel_and_file] Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			ast_log(AST_LOG_DEBUG, "[ast_cdr_set_key_by_channel_and_file] file:%s cdr-record-file:%s", file, it_cdr->record_files);
			if (!strlen(it_cdr->record_files) || !strstr(file_to_compare, it_cdr->record_files)) {
				ast_log(AST_LOG_DEBUG, "[ast_cdr_set_key_by_channel_and_file] Ignore mismatch record-file cdr");
				continue;
			}
			
			struct varshead *headp = NULL;

			ast_mutex_lock(&it_cdr->party_lock); 
			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot
				&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}
			ast_mutex_unlock(&it_cdr->party_lock);

			ret = ast_set_cdr_obj_vars(cdr, channel_name, it_cdr, name, value);
			if(ret) {
				if (headp) {
					set_variable(headp, name, value);
				}
			}
		}
	}
	ao2_iterator_destroy(it_cdrs);

	return ret;
}

void ast_cdr_set_write_database_enable(int enable)
{
	cdr_write_database_disable = enable ? 0 : 1;
}

int ast_cdr_get_write_database_enable()
{
	return cdr_write_database_disable ? 0 : 1;
}
#endif

/*!
 * \brief Format a variable on a \ref cdr_object
 */
static void cdr_object_format_var_internal(struct cdr_object *cdr, const char *name, char *value, size_t length)
{
	struct ast_var_t *variable;

	AST_LIST_TRAVERSE(&cdr->party_a.variables, variable, entries) {
		if (!strcasecmp(name, ast_var_name(variable))) {
			ast_copy_string(value, ast_var_value(variable), length);
			return;
		}
	}

	*value = '\0';
}

/*!
 * \brief Format one of the standard properties on a \ref cdr_object
 */
static int cdr_object_format_property(struct cdr_object *cdr_obj, const char *name, char *value, size_t length)
{
	struct ast_channel_snapshot *party_a = cdr_obj->party_a.snapshot;
	struct ast_channel_snapshot *party_b = cdr_obj->party_b.snapshot;

	if (!strcasecmp(name, "clid")) {
		ast_callerid_merge(value, length, party_a->caller_name, party_a->caller_number, "");
	} else if (!strcasecmp(name, "src")) {
		ast_copy_string(value, party_a->caller_number, length);
	} else if (!strcasecmp(name, "dst")) {
		ast_copy_string(value, party_a->exten, length);
	} else if (!strcasecmp(name, "dcontext")) {
		ast_copy_string(value, party_a->context, length);
	} else if (!strcasecmp(name, "channel")) {
		ast_copy_string(value, party_a->name, length);
	} else if (!strcasecmp(name, "dstchannel")) {
		if (party_b) {
			ast_copy_string(value, party_b->name, length);
		} else {
			ast_copy_string(value, "", length);
		}
	} else if (!strcasecmp(name, "lastapp")) {
		ast_copy_string(value, party_a->appl, length);
	} else if (!strcasecmp(name, "lastdata")) {
		ast_copy_string(value, party_a->data, length);
	} else if (!strcasecmp(name, "start")) {
		cdr_get_tv(cdr_obj->start, NULL, value, length);
	} else if (!strcasecmp(name, "answer")) {
		cdr_get_tv(cdr_obj->answer, NULL, value, length);
	} else if (!strcasecmp(name, "end")) {
		cdr_get_tv(cdr_obj->end, NULL, value, length);
	} else if (!strcasecmp(name, "duration")) {
		snprintf(value, length, "%ld", cdr_object_get_duration(cdr_obj));
	} else if (!strcasecmp(name, "billsec")) {
		snprintf(value, length, "%ld", cdr_object_get_billsec(cdr_obj));
	} else if (!strcasecmp(name, "disposition")) {
		snprintf(value, length, "%u", cdr_obj->disposition);
	} else if (!strcasecmp(name, "amaflags")) {
		snprintf(value, length, "%d", party_a->amaflags);
	} else if (!strcasecmp(name, "accountcode")) {
		ast_copy_string(value, party_a->accountcode, length);
#ifdef GRANDSTREAM_NETWORKS
	} else if (!strcasecmp(name, "session")) {
		ast_copy_string(value, party_a->session, length);
	/*add by wxu@grandstream.cn for get cdr_obj_vars*/
	} else if (!strcasecmp(name, "app_caller")) {
		ast_copy_string(value, cdr_obj->app_caller, length);
	} else if (!strcasecmp(name, "app_callee")) {
		ast_copy_string(value, cdr_obj->app_callee, length);
	} else if (!strcasecmp(name, "action_type")) {
		ast_copy_string(value, cdr_obj->action_type, length);
	} else if (!strcasecmp(name, "record_files")) {
		ast_copy_string(value, cdr_obj->record_files, length);
	} else if (!strcasecmp(name, "record_files_desc")) {
		ast_copy_string(value, cdr_obj->record_files_desc, length);
	} else if (!strcasecmp(name, "record_files_uuid")) {
		ast_copy_string(value, cdr_obj->record_files_uuid, length);
	} else if (!strcasecmp(name, "special_action_type")) {
		ast_copy_string(value, cdr_obj->special_action_type, length);
	} else if (!strcasecmp(name, "special_dst")) {
		ast_copy_string(value, cdr_obj->special_dst, length);
	} else if (!strcasecmp(name, "inbound_trunk_index")) {
		ast_copy_string(value, cdr_obj->inbound_trunk_index, length);
	} else if (!strcasecmp(name, "outbound_trunk_index")) {
		ast_copy_string(value, cdr_obj->outbound_trunk_index, length);
	} else if (!strcasecmp(name, "status")) {
		ast_copy_string(value, cdr_obj->status, length);
	/*end*/
#endif
	} else if (!strcasecmp(name, "peeraccount")) {
		if (party_b) {
			ast_copy_string(value, party_b->accountcode, length);
		} else {
			ast_copy_string(value, "", length);
		}
	} else if (!strcasecmp(name, "uniqueid")) {
		ast_copy_string(value, party_a->uniqueid, length);
	} else if (!strcasecmp(name, "linkedid")) {
		ast_copy_string(value, cdr_obj->linkedid, length);
	} else if (!strcasecmp(name, "userfield")) {
		ast_copy_string(value, cdr_obj->party_a.userfield, length);
	} else if (!strcasecmp(name, "sequence")) {
		snprintf(value, length, "%u", cdr_obj->sequence);
	} else {
		return 1;
	}

	return 0;
}

/*! \internal
 * \brief Look up and retrieve a CDR object by channel name
 * \param name The name of the channel
 * \retval NULL on error
 * \retval The \ref cdr_object for the channel on success, with the reference
 *	count bumped by one.
 */
static struct cdr_object *cdr_object_get_by_name(const char *name)
{
	char *param;

	if (ast_strlen_zero(name)) {
		return NULL;
	}

	param = ast_strdupa(name);
	return ao2_callback(active_cdrs_master, 0, cdr_object_get_by_name_cb, param);
}

int ast_cdr_getvar(const char *channel_name, const char *name, char *value, size_t length)
{
	struct cdr_object *cdr;
	struct cdr_object *cdr_obj;

	if (ast_strlen_zero(name)) {
		return 1;
	}

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return 1;
	}

	ao2_lock(cdr);

#ifdef GRANDSTREAM_NETWORKS
	/*added by jklou for bug 59632*/
	if (!strcasecmp(name, "record_files") || !strcasecmp(name, "record_files_desc") || !strcasecmp(name, "record_files_uuid")){
		for (cdr_obj = cdr; cdr_obj; cdr_obj = cdr_obj->next) {
			if (AST_CDR_NOANSWER == cdr_obj->disposition) {
				continue;
			}

			if (cdr_object_format_property(cdr_obj, name, value, length)) {
				/* Property failed; attempt variable */
				cdr_object_format_var_internal(cdr_obj, name, value, length);
			}
			if (!strncasecmp(cdr_obj->special_action_type, "ringboth", 8)){
				break;
			}
			/* Bug 74601 - CDR:only display one recording file when extension enable Ring Simultaneously and external extension answer */
		}
	} else {
		cdr_obj = cdr->last;
		if (cdr_object_format_property(cdr_obj, name, value, length)) {
			/* Property failed; attempt variable */
			 cdr_object_format_var_internal(cdr_obj, name, value, length);
		}
	}
#else
	cdr_obj = cdr->last;
	if (cdr_object_format_property(cdr_obj, name, value, length)) {
		/* Property failed; attempt variable */
		cdr_object_format_var_internal(cdr_obj, name, value, length);
	}
#endif

	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
int ast_cdr_oldest_getvar(const char *channel_name, const char *name, char *value, size_t length)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;

	if (ast_strlen_zero(name)) {
		return 1;
	}

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return 1;
	}

	ao2_lock(cdr);

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (cdr_object_format_property(it_cdr, name, value, length)) {
			/* Property failed; attempt variable */
			cdr_object_format_var_internal(it_cdr, name, value, length);
		}
		break;
	}

	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

int ast_cdr_last_newest_getvar(const char *channel_name, const char *name, char *value, size_t length)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;

	if (ast_strlen_zero(name)) {
		return 1;
	}

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return 1;
	}

	ao2_lock(cdr);

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->next != NULL && it_cdr->next->next != NULL) {
			continue;
		}

		if (cdr_object_format_property(it_cdr, name, value, length)) {
			/* Property failed; attempt variable */
			cdr_object_format_var_internal(it_cdr, name, value, length);
		}
		break;
	}

	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

int ast_cdr_newest_getvar(const char *channel_name, const char *name, char *value, size_t length)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;

	if (ast_strlen_zero(name)) {
		return 1;
	}

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return 1;
	}

	ao2_lock(cdr);

	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->next != NULL) {
			continue;
		}

		if (cdr_object_format_property(it_cdr, name, value, length)) {
			/* Property failed; attempt variable */
			cdr_object_format_var_internal(it_cdr, name, value, length);
		}
		break;
	}

	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

/*
 *  This is a special function to detect the cdr-object with recordfile but without recordfile-desc and
 *  fill the @desc if the stored recordfile-name match with @recordfile
 */
int ast_cdr_refill_recorddesc(const char *channel_name, const char *file, const char *desc)
{
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);

	char recordfiles[256];
	char recordfiles_desc[4096];

	ast_log(AST_LOG_DEBUG, "Enter ast_cdr_refill_recorddesc(chan:%s, recordfile:%s, desc:%s)",
		ast_strlen_zero(channel_name) ? "":channel_name,
		ast_strlen_zero(file) ? "":file, ast_strlen_zero(desc) ? "":desc);

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name))
		{
			ast_log(AST_LOG_DEBUG, "Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;
			recordfiles[0] = '\0';
			recordfiles_desc[0] = '\0';

			if (cdr_object_format_property(it_cdr, "record_files", recordfiles, sizeof(recordfiles))) {
				cdr_object_format_var_internal(it_cdr, "record_files", recordfiles, sizeof(recordfiles));
			}

			if (cdr_object_format_property(it_cdr, "record_files_desc", recordfiles_desc, sizeof(recordfiles_desc))) {
				cdr_object_format_var_internal(it_cdr, "record_files_desc", recordfiles_desc, sizeof(recordfiles_desc));
			}

			ast_log(AST_LOG_DEBUG, "cdr_obj:%p | file:%s, desc:%s", it_cdr, recordfiles, recordfiles_desc);
			if (!ast_strlen_zero(recordfiles) 
				&& ast_strlen_zero(recordfiles_desc) && strstr(recordfiles, file)) {
				ast_log(AST_LOG_DEBUG, "Need refill");
			} else {
				continue;
			}

			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot 
						&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}

			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			ret = ast_set_cdr_obj_vars(cdr, channel_name, it_cdr, "record_files_desc", desc);
			if(ret) {
				if (headp) {
					set_variable(headp, "record_files_desc", desc);
				}
			}
			/*end*/
		}
	}
	ao2_iterator_destroy(it_cdrs);

	ast_log(AST_LOG_DEBUG, "Exit ast_cdr_refill_recorddesc");
	return ret;
}

/*
 *  Be caution that this is a danger function and DO NOT call it if not necessary
 *  because this may accidently clear audio record in CDR
 */
int ast_cdr_clear_invalid_recordfile(const char *channel_name)
{
	int ret = 0;
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ao2_iterator *it_cdrs;
	char *arg = ast_strdupa(channel_name);

	char recordfiles[256];
	char recordfiles_desc[4096];

	ast_log(AST_LOG_DEBUG, "Enter ast_cdr_clear_invalid_recordfile(chan:%s)",
		ast_strlen_zero(channel_name) ? "":channel_name);

	it_cdrs = ao2_callback(active_cdrs_master, OBJ_MULTIPLE, cdr_object_select_all_by_name_cb, arg);
	if (!it_cdrs) {
		ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		return -1;
	}

	for (; (cdr = ao2_iterator_next(it_cdrs)); ao2_unlock(cdr), ao2_cleanup(cdr)) {
		ao2_lock(cdr);

		if (strcmp(cdr->name, channel_name))
		{
			ast_log(AST_LOG_DEBUG, "Ignore mismatch cdr");
			continue;
		}

		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			struct varshead *headp = NULL;
			recordfiles[0] = '\0';
			recordfiles_desc[0] = '\0';

			if (cdr_object_format_property(it_cdr, "record_files", recordfiles, sizeof(recordfiles))) {
				cdr_object_format_var_internal(it_cdr, "record_files", recordfiles, sizeof(recordfiles));
			}

			if (cdr_object_format_property(it_cdr, "record_files_desc", recordfiles_desc, sizeof(recordfiles_desc))) {
				cdr_object_format_var_internal(it_cdr, "record_files_desc", recordfiles_desc, sizeof(recordfiles_desc));
			}

			ast_log(AST_LOG_DEBUG, "cdr_obj:%p | file:%s, desc:%s", it_cdr, recordfiles, recordfiles_desc);
			if (!ast_strlen_zero(recordfiles) && ast_strlen_zero(recordfiles_desc)) {
				ast_log(AST_LOG_DEBUG, "Need clear");
			} else {
				continue;
			}

			if (!strcasecmp(channel_name, it_cdr->party_a.snapshot->name)) {
				headp = &it_cdr->party_a.variables;
			} else if (it_cdr->party_b.snapshot 
						&& !strcasecmp(channel_name, it_cdr->party_b.snapshot->name)) {
				headp = &it_cdr->party_b.variables;
			}

			/*add by wxu@grandstream.cn for set cdr_obj_vars*/
			ret = ast_set_cdr_obj_vars(cdr, channel_name, it_cdr, "record_files", NULL);
			if(ret) {
				if (headp) {
					set_variable(headp, "record_files", NULL);
				}
			}
			/*end*/
		}
	}
	ao2_iterator_destroy(it_cdrs);

	ast_log(AST_LOG_DEBUG, "Exit ast_cdr_clear_invalid_recordfile");
	return ret;
}
#endif

int ast_cdr_serialize_variables(const char *channel_name, struct ast_str **buf, char delim, char sep)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;
	struct ast_var_t *variable;
	const char *var;
	char workspace[256];
	int total = 0, x = 0, i;

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		if (is_cdr_flag_set(CDR_ENABLED)) {
			ast_log(AST_LOG_ERROR, "Unable to find CDR for channel %s\n", channel_name);
		}
		return 0;
	}

	ast_str_reset(*buf);

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (++x > 1) {
			ast_str_append(buf, 0, "\n");
		}

#ifdef GRANDSTREAM_NETWORKS
		ast_str_append(buf, 0, "---------- %p ----------\n", it_cdr);
#endif

		AST_LIST_TRAVERSE(&it_cdr->party_a.variables, variable, entries) {
			if (!(var = ast_var_name(variable))) {
				continue;
			}

			if (ast_str_append(buf, 0, "level %d: %s%c%s%c", x, var, delim, S_OR(ast_var_value(variable), ""), sep) < 0) {
				ast_log(LOG_ERROR, "Data Buffer Size Exceeded!\n");
				break;
			}

			total++;
		}

		for (i = 0; cdr_readonly_vars[i]; i++) {
			if (cdr_object_format_property(it_cdr, cdr_readonly_vars[i], workspace, sizeof(workspace))) {
				/* Unhandled read-only CDR variable. */
				ast_assert(0);
				continue;
			}

			if (!ast_strlen_zero(workspace)
				&& ast_str_append(buf, 0, "level %d: %s%c%s%c", x, cdr_readonly_vars[i], delim, workspace, sep) < 0) {
				ast_log(LOG_ERROR, "Data Buffer Size Exceeded!\n");
				break;
			}
			total++;
		}
	}
	ao2_unlock(cdr);
	ao2_cleanup(cdr);
	return total;
}

void ast_cdr_free(struct ast_cdr *cdr)
{
	while (cdr) {
		struct ast_cdr *next = cdr->next;

		free_variables(&cdr->varshead);
		ast_free(cdr);
		cdr = next;
	}
}

struct ast_cdr *ast_cdr_alloc(void)
{
	struct ast_cdr *x;

	x = ast_calloc(1, sizeof(*x));
	return x;
}

const char *ast_cdr_disp2str(int disposition)
{
	switch (disposition) {
	case AST_CDR_NULL:
		return "NO ANSWER"; /* by default, for backward compatibility */
	case AST_CDR_NOANSWER:
		return "NO ANSWER";
	case AST_CDR_FAILED:
		return "FAILED";
	case AST_CDR_BUSY:
		return "BUSY";
	case AST_CDR_ANSWERED:
		return "ANSWERED";
	case AST_CDR_CONGESTION:
		return "CONGESTION";
	}
	return "UNKNOWN";
}

struct party_b_userfield_update {
	const char *channel_name;
	const char *userfield;
};

/*! \brief Callback used to update the userfield on Party B on all CDRs */
static int cdr_object_update_party_b_userfield_cb(void *obj, void *arg, void *data, int flags)
{
	struct cdr_object *cdr = obj;

	if ((cdr->fn_table != &finalized_state_fn_table || !cdr->next)
		&& !strcasecmp(cdr->party_b_name, arg)) {
		struct party_b_userfield_update *info = data;

		/*
		 * For sanity's sake we also assert the party_b snapshot
		 * is consistent with the key.
		 */
		ast_assert(cdr->party_b.snapshot
			&& !strcasecmp(cdr->party_b.snapshot->name, info->channel_name));

		ast_copy_string(cdr->party_b.userfield, info->userfield,
			sizeof(cdr->party_b.userfield));
	}

	return 0;
}

void ast_cdr_setuserfield(const char *channel_name, const char *userfield)
{
	struct cdr_object *cdr;
	struct party_b_userfield_update party_b_info = {
		.channel_name = channel_name,
		.userfield = userfield,
	};
	struct cdr_object *it_cdr;

	/* Handle Party A */
	cdr = cdr_object_get_by_name(channel_name);
	if (cdr) {
		ao2_lock(cdr);
		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			if (it_cdr->fn_table == &finalized_state_fn_table && it_cdr->next != NULL) {
				continue;
			}
			ast_copy_string(it_cdr->party_a.userfield, userfield,
				sizeof(it_cdr->party_a.userfield));
		}
		ao2_unlock(cdr);
	}

	/* Handle Party B */
	ao2_callback_data(active_cdrs_all, OBJ_NODATA | OBJ_MULTIPLE | OBJ_SEARCH_KEY,
		cdr_object_update_party_b_userfield_cb, (char *) party_b_info.channel_name,
		&party_b_info);

	ao2_cleanup(cdr);
}

static void post_cdr(struct ast_cdr *cdr)
{
	struct module_config *mod_cfg;
	struct cdr_beitem *i;
#ifdef GRANDSTREAM_NETWORKS
	int concurrent = 1;
	int cdr_count = 0;
	int disable = 1;

	const int session_array_max = 6;
	int session_array_len = 0;
	char *session_array[6] = {NULL};
	//char *record_files_uuid_array[6] = {NULL};
#endif

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return;
	}

	for (; cdr ; cdr = cdr->next) {
		/* For people, who don't want to see unanswered single-channel events */
		if (!ast_test_flag(&mod_cfg->general->settings, CDR_UNANSWERED) &&
				cdr->disposition < AST_CDR_ANSWERED &&
				(ast_strlen_zero(cdr->channel) || ast_strlen_zero(cdr->dstchannel))) {
			ast_debug(1, "Skipping CDR for %s since we weren't answered\n", cdr->channel);
			continue;
		}

#ifdef GRANDSTREAM_NETWORKS
		/*add by wxu, if session is null set cdr disable*/
		if( (cdr->session[0] == '\0' || (atoi(cdr->session) == 0))
			&& !strstr(cdr->special_action_type, "a_xfer_resume") ){
			ast_log(LOG_VERBOSE,"cdr disable session is null");
			ast_set_flag(cdr,AST_CDR_FLAG_DISABLE);
		}
		/*add end*/

#if 0
		/*this is modify after all cf do not send missed call*/
		if (cdr->next && AST_CDR_ANSWERED != cdr->disposition && !strcmp(cdr->dst, cdr->next->action_owner)){
			concurrent--;
		}
#endif
		cdr->concurrent = concurrent;
		if (!cdr->next && !cdr_count && !strcasecmp(cdr->action_type, "VM") && strstr(cdr->lastdata, ",u")){
			ast_log(LOG_VERBOSE,"cdr set -1");
			cdr->concurrent = -1;
		}
		cdr_count ++;
#endif

		/* Modify CDR's */
		AST_RWLIST_RDLOCK(&mo_list);
		AST_RWLIST_TRAVERSE(&mo_list, i, list) {
			i->be(cdr);
		}
		AST_RWLIST_UNLOCK(&mo_list);

		if (ast_test_flag(cdr, AST_CDR_FLAG_DISABLE)) {
			continue;
		}
		AST_RWLIST_RDLOCK(&be_list);
		AST_RWLIST_TRAVERSE(&be_list, i, list) {
			if (!i->suspended) {
				i->be(cdr);
			}
		}
		AST_RWLIST_UNLOCK(&be_list);
#ifdef GRANDSTREAM_NETWORKS
		if (cdr->next && AST_CDR_ANSWERED != cdr->disposition
			&&((0 == strncmp(cdr->action_type, "IVR", strlen("IVR")) && 0 == strncmp(cdr->next->action_type, "IVR", strlen("IVR"))) || (0 == strcmp(cdr->action_type, "DIAL") && 0 == strcmp(cdr->next->action_type, "DIAL")))
			&&(!ast_strlen_zero(cdr->dst) && !ast_strlen_zero(cdr->next->dst) && 0 == strcmp(cdr->dst, cdr->next->dst))){
			concurrent++;
		}
		if (cdr->next && AST_CDR_ANSWERED != cdr->disposition
			&& !strncmp(cdr->action_type, "FOLLOWME", strlen("FOLLOWME")) && !strcasecmp(cdr->next->action_type, "VM")){
			concurrent = -1;
		}

		disable = 0;

		/*
		 * Gather session array, we found it may have multi-session string in one cdr list
		 */
		if (session_array_len < session_array_max) {
			if (ast_strlen_zero(session_array[session_array_len])
				&& 0 == session_array_len) {
				session_array[session_array_len] = cdr->session;
				//record_files_uuid_array[session_array_len] = cdr->record_files_uuid;
				session_array_len++;
			} else if (session_array_len) {
				if (strcmp(session_array[session_array_len-1], cdr->session)) {
					ast_log(LOG_WARNING, "Detect multi session within cdr list");
					session_array[session_array_len] = cdr->session;
					//record_files_uuid_array[session_array_len] = cdr->record_files_uuid;
					session_array_len++;
				}
			}
		}
#endif
	}
	ao2_cleanup(mod_cfg);

	if (!disable) {
		int j=0;
		ast_log(LOG_DEBUG, "session_array_len:%d", session_array_len);

		for (j=0; j<session_array_len; j++) {
			ast_log(LOG_DEBUG, "AMI Publish Event: NewCDR with Session:%s", session_array[j]);
			manager_event(EVENT_FLAG_CALL, "NewCDR", "Session: %s\r\n", session_array[j]);
		}
	}
}

int ast_cdr_set_property(const char *channel_name, enum ast_cdr_options option)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		return -1;
	}

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->fn_table == &finalized_state_fn_table) {
			continue;
		}
		/* Note: in general, set the flags on both the CDR record as well as the
		 * Party A. Sometimes all we have is the Party A to look at.
		 */
		ast_set_flag(&it_cdr->flags, option);
		ast_set_flag(&it_cdr->party_a, option);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "cdr(%p|%s) set flag - %d", it_cdr, it_cdr->name, option);
#endif
	}
	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

int ast_cdr_clear_property(const char *channel_name, enum ast_cdr_options option)
{
	struct cdr_object *cdr;
	struct cdr_object *it_cdr;

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		return -1;
	}

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		if (it_cdr->fn_table == &finalized_state_fn_table) {
			continue;
		}
		ast_clear_flag(&it_cdr->flags, option);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_NOTICE, "cdr(%p|%s) clear flag - %d", it_cdr, it_cdr->name, option);
#endif

	}
	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

int ast_cdr_reset(const char *channel_name, int keep_variables)
{
	struct cdr_object *cdr;
	struct ast_var_t *vardata;
	struct cdr_object *it_cdr;

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		return -1;
	}
#ifdef GRANDSTREAM_NETWORKS
	int first_disa = 0;
	int is_disa = 0;
	char pre_channel[256] = {0};
	struct cdr_object *tmp_cdr;
	ao2_lock(cdr);
	for (tmp_cdr = cdr; tmp_cdr; tmp_cdr = tmp_cdr->next) {
		if (NULL != strcasestr(tmp_cdr->context, "disa")){
			is_disa = 1;
			break;
		}
	}
	ao2_unlock(cdr);
#endif

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		/* clear variables */
		if (!keep_variables) {
			while ((vardata = AST_LIST_REMOVE_HEAD(&it_cdr->party_a.variables, entries))) {
				ast_var_delete(vardata);
			}
			if (cdr->party_b.snapshot) {
				while ((vardata = AST_LIST_REMOVE_HEAD(&it_cdr->party_b.variables, entries))) {
					ast_var_delete(vardata);
				}
			}
		}

		/* Reset to initial state */
#ifdef GRANDSTREAM_NETWORKS
		if (0 == first_disa && 0 == strncmp(it_cdr->action_type, "DISA", strlen("DISA"))){
			first_disa = 1;
		}
		if (1 == first_disa){
			first_disa = -1;
			ast_log(LOG_NOTICE, "not do reset");
			snprintf(pre_channel, sizeof(pre_channel), "%s", it_cdr->name);
			continue;
		}
		if (is_disa && cdr == it_cdr){
			ast_log(LOG_NOTICE, "not do reset");
			snprintf(pre_channel, sizeof(pre_channel), "%s", it_cdr->name);
			continue;
		}
		if (is_disa && 0 == strcmp(it_cdr->name, pre_channel)){
			snprintf(pre_channel, sizeof(pre_channel), "%s", it_cdr->name);
			ast_log(LOG_NOTICE, "not do reset");
			continue;
		}
		snprintf(pre_channel, sizeof(pre_channel), "%s", it_cdr->name);
		ast_log(LOG_NOTICE, "Do cdr start, answer and end time reset");
#endif
		memset(&it_cdr->start, 0, sizeof(it_cdr->start));
		memset(&it_cdr->end, 0, sizeof(it_cdr->end));
		memset(&it_cdr->answer, 0, sizeof(it_cdr->answer));
		it_cdr->start = ast_tvnow();
		cdr_object_check_party_a_answer(it_cdr);
	}
	ao2_unlock(cdr);

	ao2_cleanup(cdr);
	return 0;
}

int ast_cdr_fork(const char *channel_name, struct ast_flags *options)
{
	RAII_VAR(struct cdr_object *, cdr, cdr_object_get_by_name(channel_name), ao2_cleanup);
	struct cdr_object *new_cdr;
	struct cdr_object *it_cdr;
	struct cdr_object *cdr_obj;

	if (!cdr) {
		return -1;
	}

	{
		SCOPED_AO2LOCK(lock, cdr);

		cdr_obj = cdr->last;
		if (cdr_obj->fn_table == &finalized_state_fn_table) {
			/* If the last CDR in the chain is finalized, don't allow a fork -
			 * things are already dying at this point
			 */
			return -1;
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Start added by zyxia for bug 113392 emptyqueue  */
		if ( 0 == strcmp(cdr->special_action_type, "EMPTYQUEUE"))
			cdr_obj->disposition = AST_CDR_FAILED;
		else
		/* End added by zyxia for bug 113392 emptyqueue  */
			cdr_obj->disposition = AST_CDR_NOANSWER; //modify by wxu@grandstream.cn for forkCDR
		/*add for bug 109549 by ychpang*/
		if (0 == strcmp(cdr_obj->appl, "VoiceMailMain"))
		{
			cdr_obj->disposition = AST_CDR_ANSWERED;
		}
#endif

		/* Copy over the basic CDR information. The Party A information is
		 * copied over automatically as part of the append
		 */
		ast_debug(1, "Forking CDR for channel %s\n", cdr->party_a.snapshot->name);
		new_cdr = cdr_object_create_and_append(cdr);
		if (!new_cdr) {
			return -1;
		}
		new_cdr->fn_table = cdr_obj->fn_table;
		ast_string_field_set(new_cdr, bridge, cdr->bridge);
		ast_string_field_set(new_cdr, appl, cdr->appl);
		ast_string_field_set(new_cdr, data, cdr->data);
		ast_string_field_set(new_cdr, context, cdr->context);
		ast_string_field_set(new_cdr, exten, cdr->exten);
		new_cdr->flags = cdr->flags;
		/* Explicitly clear the AST_CDR_LOCK_APP flag - we want
		 * the application to be changed on the new CDR if the
		 * dialplan demands it
		 */
		ast_clear_flag(&new_cdr->flags, AST_CDR_LOCK_APP);

		/* If there's a Party B, copy it over as well */
		if (cdr_obj->party_b.snapshot) {
			new_cdr->party_b.snapshot = cdr_obj->party_b.snapshot;
			ao2_ref(new_cdr->party_b.snapshot, +1);
			cdr_all_relink(new_cdr);
			strcpy(new_cdr->party_b.userfield, cdr_obj->party_b.userfield);
			new_cdr->party_b.flags = cdr_obj->party_b.flags;
			if (ast_test_flag(options, AST_CDR_FLAG_KEEP_VARS)) {
				copy_variables(&new_cdr->party_b.variables, &cdr_obj->party_b.variables);
			}
		}
		new_cdr->start = cdr_obj->start;
		new_cdr->answer = cdr_obj->answer;

		/* Modify the times based on the flags passed in */
		if (ast_test_flag(options, AST_CDR_FLAG_SET_ANSWER)
				&& new_cdr->party_a.snapshot->state == AST_STATE_UP) {
			new_cdr->answer = ast_tvnow();
		}
		if (ast_test_flag(options, AST_CDR_FLAG_RESET)) {
#ifndef GRANDSTREAM_NETWORKS  //modify by yjwu for bug Bug 54416 - ringgroup:when ringgroup strategy is ring in order, CDR display error
			new_cdr->answer = ast_tvnow();
#endif
			new_cdr->start = ast_tvnow();
		}

		/* Create and append, by default, copies over the variables */
		if (!ast_test_flag(options, AST_CDR_FLAG_KEEP_VARS)) {
			free_variables(&new_cdr->party_a.variables);
		}

		/* Finalize any current CDRs */
		if (ast_test_flag(options, AST_CDR_FLAG_FINALIZE)) {
			for (it_cdr = cdr; it_cdr != new_cdr; it_cdr = it_cdr->next) {
				if (it_cdr->fn_table == &finalized_state_fn_table) {
					continue;
				}
				/* Force finalization on the CDR. This will bypass any checks for
				 * end before 'h' extension.
				 */
				cdr_object_finalize(it_cdr);
				cdr_object_transition_state(it_cdr, &finalized_state_fn_table);
			}
		}
	}

	return 0;
}

/*! \note Don't call without cdr_batch_lock */
static void reset_batch(void)
{
	batch->size = 0;
	batch->head = NULL;
	batch->tail = NULL;
}

/*! \note Don't call without cdr_batch_lock */
static int init_batch(void)
{
	/* This is the single meta-batch used to keep track of all CDRs during the entire life of the program */
	if (!(batch = ast_malloc(sizeof(*batch))))
		return -1;

	reset_batch();

	return 0;
}

static void *do_batch_backend_process(void *data)
{
	struct cdr_batch_item *processeditem;
	struct cdr_batch_item *batchitem = data;

	/* Push each CDR into storage mechanism(s) and free all the memory */
	while (batchitem) {
		post_cdr(batchitem->cdr);
		ast_cdr_free(batchitem->cdr);
		processeditem = batchitem;
		batchitem = batchitem->next;
		ast_free(processeditem);
	}

	return NULL;
}

static void cdr_submit_batch(int do_shutdown)
{
	struct module_config *mod_cfg;
	struct cdr_batch_item *oldbatchitems = NULL;
	pthread_t batch_post_thread = AST_PTHREADT_NULL;

	/* if there's no batch, or no CDRs in the batch, then there's nothing to do */
	if (!batch || !batch->head) {
		return;
	}

	/* move the old CDRs aside, and prepare a new CDR batch */
	ast_mutex_lock(&cdr_batch_lock);
	oldbatchitems = batch->head;
	reset_batch();
	ast_mutex_unlock(&cdr_batch_lock);

	mod_cfg = ao2_global_obj_ref(module_configs);

	/* if configured, spawn a new thread to post these CDRs,
	   also try to save as much as possible if we are shutting down safely */
	if (!mod_cfg
		|| ast_test_flag(&mod_cfg->general->batch_settings.settings, BATCH_MODE_SCHEDULER_ONLY)
		|| do_shutdown) {
		ast_debug(1, "CDR single-threaded batch processing begins now\n");
		do_batch_backend_process(oldbatchitems);
	} else {
		if (ast_pthread_create_detached_background(&batch_post_thread, NULL, do_batch_backend_process, oldbatchitems)) {
			ast_log(LOG_WARNING, "CDR processing thread could not detach, now trying in this thread\n");
			do_batch_backend_process(oldbatchitems);
		} else {
			ast_debug(1, "CDR multi-threaded batch processing begins now\n");
		}
	}

	ao2_cleanup(mod_cfg);
}

static int submit_scheduled_batch(const void *data)
{
	struct module_config *mod_cfg;
	int nextms;

	cdr_submit_batch(0);

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return 0;
	}

	/* Calculate the next scheduled interval */
	nextms = mod_cfg->general->batch_settings.time * 1000;

	ao2_cleanup(mod_cfg);

	return nextms;
}

/*! Do not hold the batch lock while calling this function */
static void start_batch_mode(void)
{
	/* Prevent two deletes from happening at the same time */
	ast_mutex_lock(&cdr_sched_lock);
	/* this is okay since we are not being called from within the scheduler */
	AST_SCHED_DEL(sched, cdr_sched);
	/* schedule the submission to occur ASAP (1 ms) */
	cdr_sched = ast_sched_add_variable(sched, 1, submit_scheduled_batch, NULL, 1);
	ast_mutex_unlock(&cdr_sched_lock);

	/* signal the do_cdr thread to wakeup early and do some work (that lazy thread ;) */
	ast_mutex_lock(&cdr_pending_lock);
	ast_cond_signal(&cdr_pending_cond);
	ast_mutex_unlock(&cdr_pending_lock);
}

static void cdr_detach(struct ast_cdr *cdr)
{
	struct cdr_batch_item *newtail;
	int curr;
	RAII_VAR(struct module_config *, mod_cfg, ao2_global_obj_ref(module_configs), ao2_cleanup);
	int submit_batch = 0;

	if (!cdr) {
		return;
	}

	/* maybe they disabled CDR stuff completely, so just drop it */
	if (!mod_cfg || !ast_test_flag(&mod_cfg->general->settings, CDR_ENABLED)) {
		ast_debug(1, "Dropping CDR !\n");
		ast_cdr_free(cdr);
		return;
	}

	/* post stuff immediately if we are not in batch mode, this is legacy behaviour */
	if (!ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE)) {
		post_cdr(cdr);
		ast_cdr_free(cdr);
		return;
	}

	/* otherwise, each CDR gets put into a batch list (at the end) */
	ast_debug(1, "CDR detaching from this thread\n");

	/* we'll need a new tail for every CDR */
	if (!(newtail = ast_calloc(1, sizeof(*newtail)))) {
		post_cdr(cdr);
		ast_cdr_free(cdr);
		return;
	}

	/* don't traverse a whole list (just keep track of the tail) */
	ast_mutex_lock(&cdr_batch_lock);
	if (!batch)
		init_batch();
	if (!batch->head) {
		/* new batch is empty, so point the head at the new tail */
		batch->head = newtail;
	} else {
		/* already got a batch with something in it, so just append a new tail */
		batch->tail->next = newtail;
	}
	newtail->cdr = cdr;
	batch->tail = newtail;
	curr = batch->size++;

	/* if we have enough stuff to post, then do it */
	if (curr >= (mod_cfg->general->batch_settings.size - 1)) {
		submit_batch = 1;
	}
	ast_mutex_unlock(&cdr_batch_lock);

	/* Don't submit a batch with cdr_batch_lock held */
	if (submit_batch) {
		start_batch_mode();
	}
}

#ifdef GRANDSTREAM_NETWORKS
static void cdr_batch_cleanup(void *arg)
{
	ast_mutex_t *mutex = arg;

	if (mutex) {
		ast_mutex_unlock(mutex);
	}
}
#endif

static void *do_cdr(void *data)
{
	struct timespec timeout;
	int schedms;
	int numevents = 0;

	pthread_cleanup_push(cdr_batch_cleanup, &cdr_pending_lock);

	for (;;) {
#ifdef GRANDSTREAM_NETWORKS
		if (cdr_batch_disable) {
			break;
		}
#endif
		struct timeval now;
		schedms = ast_sched_wait(sched);
		/* this shouldn't happen, but provide a 1 second default just in case */
		if (schedms < 0)
			schedms = 1000;
		now = ast_tvadd(ast_tvnow(), ast_samp2tv(schedms, 1000));
		timeout.tv_sec = now.tv_sec;
		timeout.tv_nsec = now.tv_usec * 1000;
		/* prevent stuff from clobbering cdr_pending_cond, then wait on signals sent to it until the timeout expires */
		ast_mutex_lock(&cdr_pending_lock);
		ast_cond_timedwait(&cdr_pending_cond, &cdr_pending_lock, &timeout);
		numevents = ast_sched_runq(sched);
		ast_mutex_unlock(&cdr_pending_lock);
		ast_debug(2, "Processed %d CDR batches from the run queue\n", numevents);
	}

	pthread_cleanup_pop(0);

	return NULL;
}

static char *handle_cli_debug(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct module_config *mod_cfg;

	switch (cmd) {
	case CLI_INIT:
		e->command = "cdr set debug [on|off]";
		e->usage = "Enable or disable extra debugging in the CDR Engine. Note\n"
				"that this will dump debug information to the VERBOSE setting\n"
				"and should only be used when debugging information from the\n"
				"CDR engine is needed.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 4) {
		return CLI_SHOWUSAGE;
	}

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		ast_cli(a->fd, "Could not set CDR debugging mode\n");
		return CLI_SUCCESS;
	}
	if (!strcasecmp(a->argv[3], "on")
		&& !ast_test_flag(&mod_cfg->general->settings, CDR_DEBUG)) {
		ast_set_flag(&mod_cfg->general->settings, CDR_DEBUG);
		ast_cli(a->fd, "CDR debugging enabled\n");
	} else if (!strcasecmp(a->argv[3], "off")
		&& ast_test_flag(&mod_cfg->general->settings, CDR_DEBUG)) {
		ast_clear_flag(&mod_cfg->general->settings, CDR_DEBUG);
		ast_cli(a->fd, "CDR debugging disabled\n");
	}
	cdr_set_debug_mode(mod_cfg);
	ao2_cleanup(mod_cfg);

	return CLI_SUCCESS;
}

/*! \brief Complete user input for 'cdr show' */
static char *cli_complete_show(struct ast_cli_args *a)
{
	int wordlen = strlen(a->word);
	struct ao2_iterator it_cdrs;
	struct cdr_object *cdr;

	it_cdrs = ao2_iterator_init(active_cdrs_master, 0);
	while ((cdr = ao2_iterator_next(&it_cdrs))) {
		if (!strncasecmp(a->word, cdr->party_a.snapshot->name, wordlen)) {
			if (ast_cli_completion_add(ast_strdup(cdr->party_a.snapshot->name))) {
				ao2_ref(cdr, -1);
				break;
			}
		}
		ao2_ref(cdr, -1);
	}
	ao2_iterator_destroy(&it_cdrs);

	return NULL;
}

static void cli_show_channels(struct ast_cli_args *a)
{
	struct ao2_iterator it_cdrs;
	struct cdr_object *cdr;
	char start_time_buffer[64];
	char answer_time_buffer[64];
	char end_time_buffer[64];

#define TITLE_STRING "%-25.25s %-25.25s %-15.15s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s\n"
#define FORMAT_STRING "%-25.25s %-25.25s %-15.15s %-8.8s %-8.8s %-8.8s %-8.8ld %-8.8ld\n"

	ast_cli(a->fd, "\n");
	ast_cli(a->fd, "Channels with Call Detail Record (CDR) Information\n");
	ast_cli(a->fd, "--------------------------------------------------\n");
	ast_cli(a->fd, TITLE_STRING, "Channel", "Dst. Channel", "LastApp", "Start", "Answer", "End", "Billsec", "Duration");

	it_cdrs = ao2_iterator_init(active_cdrs_master, 0);
	for (; (cdr = ao2_iterator_next(&it_cdrs)); ao2_cleanup(cdr)) {
		struct cdr_object *it_cdr;
		struct timeval start_time = { 0, };
		struct timeval answer_time = { 0, };
		struct timeval end_time = { 0, };

		SCOPED_AO2LOCK(lock, cdr);

		/* Calculate the start, end, answer, billsec, and duration over the
		 * life of all of the CDR entries
		 */
		for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
			if (snapshot_is_dialed(it_cdr->party_a.snapshot)) {
				continue;
			}
			if (ast_tvzero(start_time)) {
				start_time = it_cdr->start;
			}
			if (!ast_tvzero(it_cdr->answer) && ast_tvzero(answer_time)) {
				answer_time = it_cdr->answer;
			}
		}

		/* If there was no start time, then all CDRs were for a dialed channel; skip */
		if (ast_tvzero(start_time)) {
			continue;
		}
		it_cdr = cdr->last;

		end_time = ast_tvzero(it_cdr->end) ? ast_tvnow() : it_cdr->end;
		cdr_get_tv(start_time, "%T", start_time_buffer, sizeof(start_time_buffer));
		cdr_get_tv(answer_time, "%T", answer_time_buffer, sizeof(answer_time_buffer));
		cdr_get_tv(end_time, "%T", end_time_buffer, sizeof(end_time_buffer));
		ast_cli(a->fd, FORMAT_STRING, it_cdr->party_a.snapshot->name,
				it_cdr->party_b.snapshot ? it_cdr->party_b.snapshot->name : "<none>",
				it_cdr->appl,
				start_time_buffer,
				answer_time_buffer,
				end_time_buffer,
				ast_tvzero(answer_time) ? 0 : (long)ast_tvdiff_ms(end_time, answer_time) / 1000,
				(long)ast_tvdiff_ms(end_time, start_time) / 1000);
	}
	ao2_iterator_destroy(&it_cdrs);
#undef FORMAT_STRING
#undef TITLE_STRING
}

static void cli_show_channel(struct ast_cli_args *a)
{
	struct cdr_object *it_cdr;
	char clid[64];
	char start_time_buffer[64];
	char answer_time_buffer[64];
	char end_time_buffer[64];
	const char *channel_name = a->argv[3];
	struct cdr_object *cdr;

#define TITLE_STRING "%-10.10s %-20.20s %-25.25s %-15.15s %-15.15s %-8.8s %-8.8s %-8.8s %-8.8s %-8.8s\n"
#define FORMAT_STRING "%-10.10s %-20.20s %-25.25s %-15.15s %-15.15s %-8.8s %-8.8s %-8.8s %-8.8ld %-8.8ld\n"

	cdr = cdr_object_get_by_name(channel_name);
	if (!cdr) {
		ast_cli(a->fd, "Unknown channel: %s\n", channel_name);
		return;
	}

	ast_cli(a->fd, "\n");
	ast_cli(a->fd, "Call Detail Record (CDR) Information for %s\n", channel_name);
	ast_cli(a->fd, "--------------------------------------------------\n");
	ast_cli(a->fd, TITLE_STRING, "AccountCode", "CallerID", "Dst. Channel", "LastApp", "Data", "Start", "Answer", "End", "Billsec", "Duration");

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		struct timeval end;

		if (snapshot_is_dialed(it_cdr->party_a.snapshot)) {
			continue;
		}
		ast_callerid_merge(clid, sizeof(clid), it_cdr->party_a.snapshot->caller_name, it_cdr->party_a.snapshot->caller_number, "");
		if (ast_tvzero(it_cdr->end)) {
			end = ast_tvnow();
		} else {
			end = it_cdr->end;
		}
		cdr_get_tv(it_cdr->start, "%T", start_time_buffer, sizeof(start_time_buffer));
		cdr_get_tv(it_cdr->answer, "%T", answer_time_buffer, sizeof(answer_time_buffer));
		cdr_get_tv(end, "%T", end_time_buffer, sizeof(end_time_buffer));
		ast_cli(a->fd, FORMAT_STRING,
				it_cdr->party_a.snapshot->accountcode,
				clid,
				it_cdr->party_b.snapshot ? it_cdr->party_b.snapshot->name : "<none>",
				it_cdr->appl,
				it_cdr->data,
				start_time_buffer,
				answer_time_buffer,
				end_time_buffer,
				(long)ast_tvdiff_ms(end, it_cdr->answer) / 1000,
				(long)ast_tvdiff_ms(end, it_cdr->start) / 1000);
	}
	ao2_unlock(cdr);

	ao2_cleanup(cdr);

#undef FORMAT_STRING
#undef TITLE_STRING
}

static char *handle_cli_show(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
			e->command = "cdr show active";
			e->usage =
				"Usage: cdr show active [channel]\n"
				"	Displays a summary of all Call Detail Records when [channel]\n"
				"	is omitted; displays all of the Call Detail Records\n"
				"	currently in flight for a given [channel] when [channel] is\n"
				"	specified.\n\n"
				"	Note that this will not display Call Detail Records that\n"
				"	have already been dispatched to a backend storage, nor for\n"
				"	channels that are no longer active.\n";
			return NULL;
	case CLI_GENERATE:
		return cli_complete_show(a);
	}

	if (a->argc > 4) {
		return CLI_SHOWUSAGE;
	} else if (a->argc < 4) {
		cli_show_channels(a);
	} else {
		cli_show_channel(a);
	}

	return CLI_SUCCESS;
}

static char *handle_cli_status(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct cdr_beitem *beitem = NULL;
	struct module_config *mod_cfg;
	int cnt = 0;
	long nextbatchtime = 0;

	switch (cmd) {
	case CLI_INIT:
		e->command = "cdr show status";
		e->usage =
			"Usage: cdr show status\n"
			"	Displays the Call Detail Record engine system status.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc > 3) {
		return CLI_SHOWUSAGE;
	}

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return CLI_FAILURE;
	}

	ast_cli(a->fd, "\n");
	ast_cli(a->fd, "Call Detail Record (CDR) settings\n");
	ast_cli(a->fd, "----------------------------------\n");
	ast_cli(a->fd, "  Logging:                    %s\n", ast_test_flag(&mod_cfg->general->settings, CDR_ENABLED) ? "Enabled" : "Disabled");
	ast_cli(a->fd, "  Mode:                       %s\n", ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE) ? "Batch" : "Simple");
	if (ast_test_flag(&mod_cfg->general->settings, CDR_ENABLED)) {
		ast_cli(a->fd, "  Log unanswered calls:       %s\n", ast_test_flag(&mod_cfg->general->settings, CDR_UNANSWERED) ? "Yes" : "No");
		ast_cli(a->fd, "  Log congestion:             %s\n\n", ast_test_flag(&mod_cfg->general->settings, CDR_CONGESTION) ? "Yes" : "No");
		if (ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE)) {
			ast_cli(a->fd, "* Batch Mode Settings\n");
			ast_cli(a->fd, "  -------------------\n");
			if (batch)
				cnt = batch->size;
			if (cdr_sched > -1)
				nextbatchtime = ast_sched_when(sched, cdr_sched);
			ast_cli(a->fd, "  Safe shutdown:              %s\n", ast_test_flag(&mod_cfg->general->batch_settings.settings, BATCH_MODE_SAFE_SHUTDOWN) ? "Enabled" : "Disabled");
			ast_cli(a->fd, "  Threading model:            %s\n", ast_test_flag(&mod_cfg->general->batch_settings.settings, BATCH_MODE_SCHEDULER_ONLY) ? "Scheduler only" : "Scheduler plus separate threads");
			ast_cli(a->fd, "  Current batch size:         %d record%s\n", cnt, ESS(cnt));
			ast_cli(a->fd, "  Maximum batch size:         %u record%s\n", mod_cfg->general->batch_settings.size, ESS(mod_cfg->general->batch_settings.size));
			ast_cli(a->fd, "  Maximum batch time:         %u second%s\n", mod_cfg->general->batch_settings.time, ESS(mod_cfg->general->batch_settings.time));
			ast_cli(a->fd, "  Next batch processing time: %ld second%s\n\n", nextbatchtime, ESS(nextbatchtime));
		}
		ast_cli(a->fd, "* Registered Backends\n");
		ast_cli(a->fd, "  -------------------\n");
		AST_RWLIST_RDLOCK(&be_list);
		if (AST_RWLIST_EMPTY(&be_list)) {
			ast_cli(a->fd, "    (none)\n");
		} else {
			AST_RWLIST_TRAVERSE(&be_list, beitem, list) {
				ast_cli(a->fd, "    %s%s\n", beitem->name, beitem->suspended ? " (suspended) " : "");
			}
		}
		AST_RWLIST_UNLOCK(&be_list);
		ast_cli(a->fd, "\n");
	}

	ao2_cleanup(mod_cfg);
	return CLI_SUCCESS;
}

static char *handle_cli_submit(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct module_config *mod_cfg;

	switch (cmd) {
	case CLI_INIT:
		e->command = "cdr submit";
		e->usage =
			"Usage: cdr submit\n"
			"Posts all pending batched CDR data to the configured CDR\n"
			"backend engine modules.\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (a->argc > 2) {
		return CLI_SHOWUSAGE;
	}

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg) {
		return CLI_FAILURE;
	}

	if (!ast_test_flag(&mod_cfg->general->settings, CDR_ENABLED)) {
		ast_cli(a->fd, "Cannot submit CDR batch: CDR engine disabled.\n");
	} else if (!ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE)) {
		ast_cli(a->fd, "Cannot submit CDR batch: batch mode not enabled.\n");
	} else {
		start_batch_mode();
		ast_cli(a->fd, "Submitted CDRs to backend engines for processing.  This may take a while.\n");
	}
	ao2_cleanup(mod_cfg);

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_commands[] = {
	AST_CLI_DEFINE(handle_cli_submit, "Posts all pending batched CDR data"),
	AST_CLI_DEFINE(handle_cli_status, "Display the CDR status"),
	AST_CLI_DEFINE(handle_cli_show, "Display active CDRs for channels"),
	AST_CLI_DEFINE(handle_cli_debug, "Enable debugging in the CDR engine"),
};

/*!
 * \brief This dispatches *all* \ref cdr_objects. It should only be used during
 * shutdown, so that we get billing records for everything that we can.
 */
static int cdr_object_dispatch_all_cb(void *obj, void *arg, int flags)
{
	struct cdr_object *cdr = obj;
	struct cdr_object *it_cdr;

	ao2_lock(cdr);
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		cdr_object_transition_state(it_cdr, &finalized_state_fn_table);
	}
	cdr_object_dispatch(cdr);
	ao2_unlock(cdr);

	cdr_all_unlink(cdr);

	return CMP_MATCH;
}

static void finalize_batch_mode(void)
{
	if (cdr_thread == AST_PTHREADT_NULL) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	cdr_batch_disable = 1;
	ast_mutex_lock(&cdr_pending_lock);
	ast_cond_signal(&cdr_pending_cond);
	ast_mutex_unlock(&cdr_pending_lock);
#endif

	/* wake up the thread so it will exit */
	pthread_cancel(cdr_thread);
	pthread_kill(cdr_thread, SIGURG);
	pthread_join(cdr_thread, NULL);
	cdr_thread = AST_PTHREADT_NULL;
	ast_cond_destroy(&cdr_pending_cond);
	ast_cdr_engine_term();
}

struct stasis_message_router *ast_cdr_message_router(void)
{
	if (!stasis_router) {
		return NULL;
	}

	ao2_bump(stasis_router);
	return stasis_router;
}

/*!
 * \brief Destroy the active Stasis subscriptions
 */
static void destroy_subscriptions(void)
{
	channel_subscription = stasis_forward_cancel(channel_subscription);
	bridge_subscription = stasis_forward_cancel(bridge_subscription);
	parking_subscription = stasis_forward_cancel(parking_subscription);
#ifdef GRANDSTREAM_NETWORKS
	mixmonitor_subscription = stasis_forward_cancel(mixmonitor_subscription);
#endif
}

/*!
 * \brief Create the Stasis subcriptions for CDRs
 */
static int create_subscriptions(void)
{
	if (!cdr_topic) {
		return -1;
	}

	if (channel_subscription || bridge_subscription || parking_subscription) {
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (mixmonitor_subscription) {
		return 0;
	}
#endif

	channel_subscription = stasis_forward_all(ast_channel_topic_all_cached(), cdr_topic);
	if (!channel_subscription) {
		return -1;
	}
	bridge_subscription = stasis_forward_all(ast_bridge_topic_all_cached(), cdr_topic);
	if (!bridge_subscription) {
		return -1;
	}
	parking_subscription = stasis_forward_all(ast_parking_topic(), cdr_topic);
	if (!parking_subscription) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	mixmonitor_subscription = stasis_forward_all(ast_mixmonitor_topic(), cdr_topic);
	if (!mixmonitor_subscription) {
		return -1;
	}
#endif

	return 0;
}

static int process_config(int reload)
{
	if (!reload) {
		if (aco_info_init(&cfg_info)) {
			return 1;
		}

		aco_option_register(&cfg_info, "enable", ACO_EXACT, general_options, DEFAULT_ENABLED, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_ENABLED);
		aco_option_register(&cfg_info, "debug", ACO_EXACT, general_options, 0, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_DEBUG);
		aco_option_register(&cfg_info, "unanswered", ACO_EXACT, general_options, DEFAULT_UNANSWERED, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_UNANSWERED);
		aco_option_register(&cfg_info, "congestion", ACO_EXACT, general_options, 0, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_CONGESTION);
		aco_option_register(&cfg_info, "batch", ACO_EXACT, general_options, DEFAULT_BATCHMODE, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_BATCHMODE);
		aco_option_register(&cfg_info, "endbeforehexten", ACO_EXACT, general_options, DEFAULT_END_BEFORE_H_EXTEN, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_END_BEFORE_H_EXTEN);
		aco_option_register(&cfg_info, "initiatedseconds", ACO_EXACT, general_options, DEFAULT_INITIATED_SECONDS, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, settings), CDR_INITIATED_SECONDS);
		aco_option_register(&cfg_info, "scheduleronly", ACO_EXACT, general_options, DEFAULT_BATCH_SCHEDULER_ONLY, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, batch_settings.settings), BATCH_MODE_SCHEDULER_ONLY);
		aco_option_register(&cfg_info, "safeshutdown", ACO_EXACT, general_options, DEFAULT_BATCH_SAFE_SHUTDOWN, OPT_BOOLFLAG_T, 1, FLDSET(struct ast_cdr_config, batch_settings.settings), BATCH_MODE_SAFE_SHUTDOWN);
		aco_option_register(&cfg_info, "size", ACO_EXACT, general_options, DEFAULT_BATCH_SIZE, OPT_UINT_T, PARSE_IN_RANGE, FLDSET(struct ast_cdr_config, batch_settings.size), 0, MAX_BATCH_SIZE);
		aco_option_register(&cfg_info, "time", ACO_EXACT, general_options, DEFAULT_BATCH_TIME, OPT_UINT_T, PARSE_IN_RANGE, FLDSET(struct ast_cdr_config, batch_settings.time), 1, MAX_BATCH_TIME);
	}

	if (aco_process_config(&cfg_info, reload) == ACO_PROCESS_ERROR) {
		struct module_config *mod_cfg;

		if (reload) {
			return 1;
		}

		/* If we couldn't process the configuration and this wasn't a reload,
		 * create a default config
		 */
		mod_cfg = module_config_alloc();
		if (!mod_cfg
			|| aco_set_defaults(&general_option, "general", mod_cfg->general)) {
			ao2_cleanup(mod_cfg);
			return 1;
		}
		ast_log(LOG_NOTICE, "Failed to process CDR configuration; using defaults\n");
		ao2_global_obj_replace_unref(module_configs, mod_cfg);
		cdr_set_debug_mode(mod_cfg);
		ao2_cleanup(mod_cfg);
	}

	return 0;
}

static void cdr_engine_shutdown(void)
{
	stasis_message_router_unsubscribe_and_join(stasis_router);
	stasis_router = NULL;

	ao2_cleanup(cdr_topic);
	cdr_topic = NULL;

	STASIS_MESSAGE_TYPE_CLEANUP(cdr_sync_message_type);

	ao2_callback(active_cdrs_master, OBJ_NODATA | OBJ_MULTIPLE | OBJ_UNLINK,
		cdr_object_dispatch_all_cb, NULL);
	finalize_batch_mode();
	ast_cli_unregister_multiple(cli_commands, ARRAY_LEN(cli_commands));
	ast_sched_context_destroy(sched);
	sched = NULL;
	ast_free(batch);
	batch = NULL;

	aco_info_destroy(&cfg_info);
	ao2_global_obj_release(module_configs);

	ao2_container_unregister("cdrs_master");
	ao2_cleanup(active_cdrs_master);
	active_cdrs_master = NULL;

	ao2_container_unregister("cdrs_all");
	ao2_cleanup(active_cdrs_all);
	active_cdrs_all = NULL;
}

static void cdr_enable_batch_mode(struct ast_cdr_config *config)
{
	/* Only create the thread level portions once */
	if (cdr_thread == AST_PTHREADT_NULL) {
		ast_cond_init(&cdr_pending_cond, NULL);
#ifdef GRANDSTREAM_NETWORKS
		cdr_batch_disable = 0;
#endif
		if (ast_pthread_create_background(&cdr_thread, NULL, do_cdr, NULL) < 0) {
			ast_log(LOG_ERROR, "Unable to start CDR thread.\n");
			return;
		}
	}

	/* Start the batching process */
	start_batch_mode();

	ast_log(LOG_NOTICE, "CDR batch mode logging enabled, first of either size %u or time %u seconds.\n",
			config->batch_settings.size, config->batch_settings.time);
}

/*!
 * \internal
 * \brief Print master CDR container object.
 * \since 12.0.0
 *
 * \param v_obj A pointer to the object we want printed.
 * \param where User data needed by prnt to determine where to put output.
 * \param prnt Print output callback function to use.
 *
 * \return Nothing
 */
static void cdr_master_print_fn(void *v_obj, void *where, ao2_prnt_fn *prnt)
{
	struct cdr_object *cdr = v_obj;
	struct cdr_object *it_cdr;

	if (!cdr) {
		return;
	}
	for (it_cdr = cdr; it_cdr; it_cdr = it_cdr->next) {
		prnt(where, "Party A: %s; Party B: %s; Bridge %s\n",
			it_cdr->party_a.snapshot->name,
			it_cdr->party_b.snapshot ? it_cdr->party_b.snapshot->name : "<unknown>",
			it_cdr->bridge);
	}
}

/*!
 * \internal
 * \brief Print all CDR container object.
 * \since 13.19.0
 *
 * \param v_obj A pointer to the object we want printed.
 * \param where User data needed by prnt to determine where to put output.
 * \param prnt Print output callback function to use.
 *
 * \return Nothing
 */
static void cdr_all_print_fn(void *v_obj, void *where, ao2_prnt_fn *prnt)
{
	struct cdr_object *cdr = v_obj;

	if (!cdr) {
		return;
	}
	prnt(where, "Party A: %s; Party B: %s; Bridge %s",
		cdr->party_a.snapshot->name,
		cdr->party_b.snapshot ? cdr->party_b.snapshot->name : "<unknown>",
		cdr->bridge);
}

/*!
 * \brief Checks if CDRs are enabled and enables/disables the necessary options
 */
static int cdr_toggle_runtime_options(void)
{
	struct module_config *mod_cfg;

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (mod_cfg
		&& ast_test_flag(&mod_cfg->general->settings, CDR_ENABLED)) {
		if (create_subscriptions()) {
			destroy_subscriptions();
			ast_log(AST_LOG_ERROR, "Failed to create Stasis subscriptions\n");
			ao2_cleanup(mod_cfg);
			return -1;
		}
		if (ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE)) {
			cdr_enable_batch_mode(mod_cfg->general);
		} else {
			ast_log(LOG_NOTICE, "CDR simple logging enabled.\n");
		}
	} else {
		destroy_subscriptions();
		ast_log(LOG_NOTICE, "CDR logging disabled.\n");
	}
	ao2_cleanup(mod_cfg);

	return mod_cfg ? 0 : -1;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for limit basic UCM RC nat call time */
static void destroy_remote_connect_param(void)
{
	ast_rwlock_destroy(&time_lock);

	if (clear_limit_time_thread != AST_PTHREADT_NULL) {
		pthread_cancel(clear_limit_time_thread);
		pthread_kill(clear_limit_time_thread, SIGURG);
		pthread_join(clear_limit_time_thread, NULL);
		clear_limit_time_thread = AST_PTHREADT_NULL;
	}

	return;
}

static void init_remote_connect_param(void)
{
	char limit_time[8] = { 0 };
	char month_time[16] = { 0 };
	char time_used[16] = { 0 };
	char month_time_used[16] = { 0 };

	ast_rwlock_init(&time_lock);

	ast_db_get("TurnInfo", "DayTotalLimitTime", limit_time, sizeof(limit_time) - 1);
	if (strlen(limit_time) && atoi(limit_time) != 0) {
		has_limit_time = 1;
		day_total_limit_time = atoi(limit_time);
		start_clear_limit_time_thread();
	}

	ast_db_get("TurnInfo", "MonthTotalLimitTime", month_time, sizeof(month_time) - 1);
	if (strlen(month_time) && atoi(month_time) != 0) {
		has_month_limit_time = 1;
	}

	ast_db_get("TurnInfo", TURN_INFO_LIMITTIMEUSED_KEY, time_used, sizeof(time_used) - 1);
	if (strlen(time_used)) {
		limit_time_used = atoi(time_used);
	} else {
		limit_time_used = 0;
	}

	ast_db_get("TurnInfo", TURN_INFO_LIMITMONTHTIMEUSED_KEY, month_time_used, sizeof(month_time_used) - 1);
	if (strlen(month_time_used)) {
		limit_month_time_used = atoi(month_time_used);
	} else {
		limit_month_time_used = 0;
	}

	clear_time_for_reboot();

	return;
}
#endif

static int unload_module(void)
{
	destroy_subscriptions();

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for limit basic UCM RC nat call time */
	destroy_remote_connect_param();
#endif

	return 0;
}

static int load_module(void)
{
	if (process_config(0)) {
		return AST_MODULE_LOAD_FAILURE;
	}

	cdr_topic = stasis_topic_create("cdr_engine");
	if (!cdr_topic) {
		return AST_MODULE_LOAD_FAILURE;
	}
	ast_log(LOG_ERROR, "load module cdr");

#ifdef GRANDSTREAM_NETWORKS
	stasis_router = stasis_message_router_create_pool(cdr_topic);
	get_device_name();
#else
	stasis_router = stasis_message_router_create(cdr_topic);
#endif
	if (!stasis_router) {
		return AST_MODULE_LOAD_FAILURE;
	}
	stasis_message_router_set_congestion_limits(stasis_router, -1,
		10 * AST_TASKPROCESSOR_HIGH_WATER_LEVEL);

	if (STASIS_MESSAGE_TYPE_INIT(cdr_sync_message_type)) {
		return AST_MODULE_LOAD_FAILURE;
	}
	stasis_message_router_add_cache_update(stasis_router, ast_channel_snapshot_type(), handle_channel_cache_message, NULL);
	stasis_message_router_add(stasis_router, ast_channel_dial_type(), handle_dial_message, NULL);
	stasis_message_router_add(stasis_router, ast_channel_entered_bridge_type(), handle_bridge_enter_message, NULL);
	stasis_message_router_add(stasis_router, ast_channel_left_bridge_type(), handle_bridge_leave_message, NULL);
	stasis_message_router_add(stasis_router, ast_parked_call_type(), handle_parked_call_message, NULL);
	stasis_message_router_add(stasis_router, cdr_sync_message_type(), handle_cdr_sync_message, NULL);
#ifdef GRANDSTREAM_NETWORKS
	stasis_message_router_add(stasis_router, ast_mixmonitor_type(), handle_mixmonitor_message, NULL);
	/*Start added by zyxia for callflip cdr bridge enter,2021.12.7*/
	stasis_message_router_add(stasis_router, ast_channel_entered_bridge_type_callflip(), handle_bridge_enter_message, NULL);
	/*End added by zyxia for callflip cdr bridge enter,2021.12.7*/
#endif

	active_cdrs_master = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
		NUM_CDR_BUCKETS, cdr_master_hash_fn, NULL, cdr_master_cmp_fn);
	if (!active_cdrs_master) {
		return AST_MODULE_LOAD_FAILURE;
	}
	ao2_container_register("cdrs_master", active_cdrs_master, cdr_master_print_fn);

	active_cdrs_all = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
		NUM_CDR_BUCKETS, cdr_all_hash_fn, NULL, cdr_all_cmp_fn);
	if (!active_cdrs_all) {
		return AST_MODULE_LOAD_FAILURE;
	}
	ao2_container_register("cdrs_all", active_cdrs_all, cdr_all_print_fn);

	sched = ast_sched_context_create();
	if (!sched) {
		ast_log(LOG_ERROR, "Unable to create schedule context.\n");
		return AST_MODULE_LOAD_FAILURE;
	}

#ifdef GRANDSTREAM_NETWORKS
	init_remote_connect_param();
#endif

	ast_cli_register_multiple(cli_commands, ARRAY_LEN(cli_commands));
	ast_register_atexit(cdr_engine_shutdown);

	return cdr_toggle_runtime_options() ? AST_MODULE_LOAD_FAILURE : AST_MODULE_LOAD_SUCCESS;
}

void ast_cdr_engine_term(void)
{
	RAII_VAR(struct module_config *, mod_cfg, ao2_global_obj_ref(module_configs), ao2_cleanup);

	/* Since this is called explicitly during process shutdown, we might not have ever
	 * been initialized. If so, the config object will be NULL.
	 */
	if (!mod_cfg) {
		return;
	}

	if (cdr_sync_message_type()) {
		void *payload;
		struct stasis_message *message;

		if (!stasis_router) {
			return;
		}

		/* Make sure we have the needed items */
		payload = ao2_alloc(sizeof(*payload), NULL);
		if (!payload) {
			return;
		}

		ast_debug(1, "CDR Engine termination request received; waiting on messages...\n");

		message = stasis_message_create(cdr_sync_message_type(), payload);
		if (message) {
			stasis_message_router_publish_sync(stasis_router, message);
		}
		ao2_cleanup(message);
		ao2_cleanup(payload);
	}

	if (ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE)) {
		cdr_submit_batch(ast_test_flag(&mod_cfg->general->batch_settings.settings, BATCH_MODE_SAFE_SHUTDOWN));
	}
}

static int reload_module(void)
{
	struct module_config *old_mod_cfg;
	struct module_config *mod_cfg;

	old_mod_cfg = ao2_global_obj_ref(module_configs);

	if (!old_mod_cfg || process_config(1)) {
		ao2_cleanup(old_mod_cfg);
		return -1;
	}

	mod_cfg = ao2_global_obj_ref(module_configs);
	if (!mod_cfg
		|| !ast_test_flag(&mod_cfg->general->settings, CDR_ENABLED)
		|| !ast_test_flag(&mod_cfg->general->settings, CDR_BATCHMODE)) {
		/* If batch mode used to be enabled, finalize the batch */
		if (ast_test_flag(&old_mod_cfg->general->settings, CDR_BATCHMODE)) {
			finalize_batch_mode();
		}
	}
	ao2_cleanup(mod_cfg);

	ao2_cleanup(old_mod_cfg);
	return cdr_toggle_runtime_options();
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "CDR Engine",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
	.load_pri = AST_MODPRI_CORE,
	.requires = "extconfig",
);
