/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Mark Michelson <mmichelson@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

#include "asterisk.h"

#include <pjsip.h>
/* Needed for SUBSCRIBE, NOTIFY, and PUBLISH method definitions */
#include <pjsip_simple.h>
#include <pjsip/sip_transaction.h>
#include <pj/timer.h>
#include <pjlib.h>
#include <pjmedia/errno.h>

#include "asterisk/res_pjsip.h"
#include "res_pjsip/include/res_pjsip_private.h"
#include "asterisk/linkedlists.h"
#include "asterisk/logger.h"
#include "asterisk/lock.h"
#include "asterisk/utils.h"
#include "asterisk/astobj2.h"
#include "asterisk/module.h"
#include "asterisk/threadpool.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/uuid.h"
#include "asterisk/sorcery.h"
#include "asterisk/file.h"
#include "asterisk/cli.h"
#include "asterisk/res_pjsip_cli.h"
#include "asterisk/test.h"
#include "asterisk/res_pjsip_presence_xml.h"
#include "asterisk/res_pjproject.h"
#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/astdb.h"
#include "asterisk/pbx.h"
#include "asterisk/res_pjsip_pubsub.h"
#include "asterisk/app.h"
#include "asterisk/stasis.h"
#include "asterisk/stasis_message_router.h"
#include "asterisk/acl.h"
#include "asterisk/callerid.h"
#include "asterisk/avs_controller.h"
#include "asterisk/crypto.h"
#include "asterisk/http_websocket.h"
#include "asterisk/haserver.h"
#include "../apps/confbridge/include/conf_resource.h"
#include "ucm_crypt.h"
#include "asterisk/ucm_dbi.h"
#include "asterisk/core_odbc.h"
AST_MODULE_LOG("pjsip");
#endif

/*** MODULEINFO
	<depend>ucmsecret</depend>
	<depend>pjproject</depend>
	<depend>res_pjproject</depend>
	<depend>res_sorcery_config</depend>
	<depend>res_sorcery_memory</depend>
	<depend>res_sorcery_astdb</depend>
	<use type="module">res_statsd</use>
	<support_level>core</support_level>
 ***/

/*** DOCUMENTATION
	<configInfo name="res_pjsip" language="en_US">
		<synopsis>SIP Resource using PJProject</synopsis>
		<configFile name="pjsip.conf">
			<configObject name="endpoint">
				<synopsis>Endpoint</synopsis>
				<description><para>
					The <emphasis>Endpoint</emphasis> is the primary configuration object.
					It contains the core SIP related options only, endpoints are <emphasis>NOT</emphasis>
					dialable entries of their own. Communication with another SIP device is
					accomplished via Addresses of Record (AoRs) which have one or more
					contacts associated with them. Endpoints <emphasis>NOT</emphasis> configured to
					use a <literal>transport</literal> will default to first transport found
					in <filename>pjsip.conf</filename> that matches its type.
					</para>
					<para>Example: An Endpoint has been configured with no transport.
					When it comes time to call an AoR, PJSIP will find the
					first transport that matches the type. A SIP URI of <literal>sip:5000@[11::33]</literal>
					will use the first IPv6 transport and try to send the request.
					</para>
					<para>If the anonymous endpoint identifier is in use an endpoint with the name
					"anonymous@domain" will be searched for as a last resort. If this is not found
					it will fall back to searching for "anonymous". If neither endpoints are found
					the anonymous endpoint identifier will not return an endpoint and anonymous
					calling will not be possible.
					</para>
				</description>
				<configOption name="100rel" default="yes">
					<synopsis>Allow support for RFC3262 provisional ACK tags</synopsis>
					<description>
						<enumlist>
							<enum name="no" />
							<enum name="required" />
							<enum name="yes" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="aggregate_mwi" default="yes">
					<synopsis>Condense MWI notifications into a single NOTIFY.</synopsis>
					<description><para>When enabled, <replaceable>aggregate_mwi</replaceable> condenses message
					waiting notifications from multiple mailboxes into a single NOTIFY. If it is disabled,
					individual NOTIFYs are sent for each mailbox.</para></description>
				</configOption>
				<configOption name="allow">
					<synopsis>Media Codec(s) to allow</synopsis>
				</configOption>
				<configOption name="allow_overlap" default="yes">
					<synopsis>Enable RFC3578 overlap dialing support.</synopsis>
				</configOption>
				<configOption name="aors">
					<synopsis>AoR(s) to be used with the endpoint</synopsis>
					<description><para>
						List of comma separated AoRs that the endpoint should be associated with.
					</para></description>
				</configOption>
				<configOption name="auth">
					<synopsis>Authentication Object(s) associated with the endpoint</synopsis>
					<description><para>
						This is a comma-delimited list of <replaceable>auth</replaceable> sections defined
						in <filename>pjsip.conf</filename> to be used to verify inbound connection attempts.
						</para><para>
						Endpoints without an authentication object
						configured will allow connections without verification.</para>
						<note><para>
						Using the same auth section for inbound and outbound
						authentication is not recommended.  There is a difference in
						meaning for an empty realm setting between inbound and outbound
						authentication uses.  See the auth realm description for details.
						</para></note>
					</description>
				</configOption>
				<configOption name="callerid">
					<synopsis>CallerID information for the endpoint</synopsis>
					<description><para>
						Must be in the format <literal>Name &lt;Number&gt;</literal>,
						or only <literal>&lt;Number&gt;</literal>.
					</para></description>
				</configOption>
				<configOption name="callerid_privacy">
					<synopsis>Default privacy level</synopsis>
					<description>
						<enumlist>
							<enum name="allowed_not_screened" />
							<enum name="allowed_passed_screen" />
							<enum name="allowed_failed_screen" />
							<enum name="allowed" />
							<enum name="prohib_not_screened" />
							<enum name="prohib_passed_screen" />
							<enum name="prohib_failed_screen" />
							<enum name="prohib" />
							<enum name="unavailable" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="callerid_tag">
					<synopsis>Internal id_tag for the endpoint</synopsis>
				</configOption>
				<configOption name="context">
					<synopsis>Dialplan context for inbound sessions</synopsis>
				</configOption>
				<configOption name="direct_media_glare_mitigation" default="none">
					<synopsis>Mitigation of direct media (re)INVITE glare</synopsis>
					<description>
						<para>
						This setting attempts to avoid creating INVITE glare scenarios
						by disabling direct media reINVITEs in one direction thereby allowing
						designated servers (according to this option) to initiate direct
						media reINVITEs without contention and significantly reducing call
						setup time.
						</para>
						<para>
						A more detailed description of how this option functions can be found on
						the Asterisk wiki https://wiki.asterisk.org/wiki/display/AST/SIP+Direct+Media+Reinvite+Glare+Avoidance
						</para>
						<enumlist>
							<enum name="none" />
							<enum name="outgoing" />
							<enum name="incoming" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="direct_media_method" default="invite">
					<synopsis>Direct Media method type</synopsis>
					<description>
						<para>Method for setting up Direct Media between endpoints.</para>
						<enumlist>
							<enum name="invite" />
							<enum name="reinvite">
								<para>Alias for the <literal>invite</literal> value.</para>
							</enum>
							<enum name="update" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="trust_connected_line">
					<synopsis>Accept Connected Line updates from this endpoint</synopsis>
				</configOption>
				<configOption name="send_connected_line">
					<synopsis>Send Connected Line updates to this endpoint</synopsis>
				</configOption>
				<configOption name="connected_line_method" default="invite">
					<synopsis>Connected line method type</synopsis>
					<description>
						<para>Method used when updating connected line information.</para>
						<enumlist>
							<enum name="invite">
							<para>When set to <literal>invite</literal>, check the remote's Allow header and
							if UPDATE is allowed, send UPDATE instead of INVITE to avoid SDP
							renegotiation.  If UPDATE is not Allowed, send INVITE.</para>
							</enum>
							<enum name="reinvite">
								<para>Alias for the <literal>invite</literal> value.</para>
							</enum>
							<enum name="update">
							<para>If set to <literal>update</literal>, send UPDATE regardless of what the remote
							Allows. </para>
							</enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="direct_media" default="yes">
					<synopsis>Determines whether media may flow directly between endpoints.</synopsis>
				</configOption>
				<configOption name="disable_direct_media_on_nat" default="no">
					<synopsis>Disable direct media session refreshes when NAT obstructs the media session</synopsis>
				</configOption>
				<configOption name="disallow">
					<synopsis>Media Codec(s) to disallow</synopsis>
				</configOption>
				<configOption name="dtmf_mode" default="rfc4733">
					<synopsis>DTMF mode</synopsis>
					<description>
						<para>This setting allows to choose the DTMF mode for endpoint communication.</para>
						<enumlist>
							<enum name="rfc4733">
								<para>DTMF is sent out of band of the main audio stream.  This
								supercedes the older <emphasis>RFC-2833</emphasis> used within
								the older <literal>chan_sip</literal>.</para>
							</enum>
							<enum name="inband">
								<para>DTMF is sent as part of audio stream.</para>
							</enum>
							<enum name="info">
								<para>DTMF is sent as SIP INFO packets.</para>
							</enum>
							<enum name="auto">
								<para>DTMF is sent as RFC 4733 if the other side supports it or as INBAND if not.</para>
							</enum>
							<enum name="auto_info">
								<para>DTMF is sent as RFC 4733 if the other side supports it or as SIP INFO if not.</para>
							</enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="media_address">
					<synopsis>IP address used in SDP for media handling</synopsis>
					<description><para>
						At the time of SDP creation, the IP address defined here will be used as
						the media address for individual streams in the SDP.
					</para>
					<note><para>
						Be aware that the <literal>external_media_address</literal> option, set in Transport
						configuration, can also affect the final media address used in the SDP.
					</para></note>
					</description>
				</configOption>
				<configOption name="bind_rtp_to_media_address">
					<synopsis>Bind the RTP instance to the media_address</synopsis>
					<description><para>
						If media_address is specified, this option causes the RTP instance to be bound to the
						specified ip address which causes the packets to be sent from that address.
					</para>
					</description>
				</configOption>
				<configOption name="force_rport" default="yes">
					<synopsis>Force use of return port</synopsis>
				</configOption>
				<configOption name="ice_support" default="no">
					<synopsis>Enable the ICE mechanism to help traverse NAT</synopsis>
				</configOption>
				<configOption name="identify_by">
					<synopsis>Way(s) for the endpoint to be identified</synopsis>
					<description>
						<para>Endpoints and AORs can be identified in multiple ways.  This
						option is a comma separated list of methods the endpoint can be
						identified.
						</para>
						<note><para>
						This option controls both how an endpoint is matched for incoming
						traffic and also how an AOR is determined if a registration
						occurs.  You must list at least one method that also matches for
						AORs or the registration will fail.
						</para></note>
						<enumlist>
							<enum name="username">
								<para>Matches the endpoint or AOR ID based on the username
								and domain in the From header (or To header for AORs).  If
								an exact match on both username and domain/realm fails, the
								match is retried with just the username.
								</para>
							</enum>
							<enum name="auth_username">
								<para>Matches the endpoint or AOR ID based on the username
								and realm in the Authentication header.  If an exact match
								on both username and domain/realm fails, the match is
								retried with just the username.
								</para>
								<note><para>This method of identification has some security
								considerations because an Authentication header is not
								present on the first message of a dialog when digest
								authentication is used.  The client can't generate it until
								the server sends the challenge in a 401 response.  Since
								Asterisk normally sends a security event when an incoming
								request can't be matched to an endpoint, using this method
								requires that the security event be deferred until a request
								is received with the Authentication header and only
								generated if the username doesn't result in a match.  This
								may result in a delay before an attack is recognized.  You
								can control how many unmatched requests are received from
								a single ip address before a security event is generated
								using the <literal>unidentified_request</literal>
								parameters in the "global" configuration object.
								</para></note>
							</enum>
							<enum name="ip">
								<para>Matches the endpoint based on the source IP address.
								</para>
								<para>This method of identification is not configured here
								but simply allowed by this configuration option.  See the
								documentation for the <literal>identify</literal>
								configuration section for more details on this method of
								endpoint identification.
								</para>
							</enum>
							<enum name="header">
								<para>Matches the endpoint based on a configured SIP header
								value.
								</para>
								<para>This method of identification is not configured here
								but simply allowed by this configuration option.  See the
								documentation for the <literal>identify</literal>
								configuration section for more details on this method of
								endpoint identification.
								</para>
							</enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="redirect_method">
					<synopsis>How redirects received from an endpoint are handled</synopsis>
					<description><para>
						When a redirect is received from an endpoint there are multiple ways it can be handled.
						If this option is set to <literal>user</literal> the user portion of the redirect target
						is treated as an extension within the dialplan and dialed using a Local channel. If this option
						is set to <literal>uri_core</literal> the target URI is returned to the dialing application
						which dials it using the PJSIP channel driver and endpoint originally used. If this option is
						set to <literal>uri_pjsip</literal> the redirect occurs within chan_pjsip itself and is not exposed
						to the core at all. The <literal>uri_pjsip</literal> option has the benefit of being more efficient
						and also supporting multiple potential redirect targets. The con is that since redirection occurs
						within chan_pjsip redirecting information is not forwarded and redirection can not be
						prevented.
						</para>
						<enumlist>
							<enum name="user" />
							<enum name="uri_core" />
							<enum name="uri_pjsip" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="mailboxes">
					<synopsis>NOTIFY the endpoint when state changes for any of the specified mailboxes</synopsis>
					<description><para>
						Asterisk will send unsolicited MWI NOTIFY messages to the endpoint when state
						changes happen for any of the specified mailboxes. More than one mailbox can be
						specified with a comma-delimited string. app_voicemail mailboxes must be specified
						as mailbox@context; for example: mailboxes=6001@default. For mailboxes provided by
						external sources, such as through the res_external_mwi module, you must specify
						strings supported by the external system.
					</para><para>
						For endpoints that SUBSCRIBE for MWI, use the <literal>mailboxes</literal> option in your AOR
						configuration.
					</para></description>
				</configOption>
				<configOption name="mwi_subscribe_replaces_unsolicited">
					<synopsis>An MWI subscribe will replace sending unsolicited NOTIFYs</synopsis>
				</configOption>
				<configOption name="voicemail_extension">
					<synopsis>The voicemail extension to send in the NOTIFY Message-Account header</synopsis>
				</configOption>
				<configOption name="moh_suggest" default="default">
					<synopsis>Default Music On Hold class</synopsis>
				</configOption>
				<configOption name="didmode" default="request-line">
					<synopsis>Configure the didmode</synopsis>
				</configOption>
				<configOption name="faxcontext" default="">
					<synopsis>fax</synopsis>
				</configOption>
				<configOption name="fax_intelligent_route" default="">
					<synopsis>fax intelligent routing</synopsis>
				</configOption>
				<configOption name="fax_intelligent_route_destination" default="">
					<synopsis>fax intelligent routing destination</synopsis>
				</configOption>
				<configOption name="intranet_ip_filter" default="yes">
					<synopsis>Filter intranet IP address</synopsis>
				</configOption>
				<configOption name="outbound_auth">
					<synopsis>Authentication object(s) used for outbound requests</synopsis>
					<description><para>
						This is a comma-delimited list of <replaceable>auth</replaceable>
						sections defined in <filename>pjsip.conf</filename> used to respond
						to outbound connection authentication challenges.</para>
						<note><para>
						Using the same auth section for inbound and outbound
						authentication is not recommended.  There is a difference in
						meaning for an empty realm setting between inbound and outbound
						authentication uses.  See the auth realm description for details.
						</para></note>
					</description>
				</configOption>
				<configOption name="outbound_proxy">
					<synopsis>Full SIP URI of the outbound proxy used to send requests</synopsis>
				</configOption>
				<configOption name="rmv_obp_from_route" default="no">
					<synopsis>If yes, remove the outbound proxy from Route header of the outgoing message</synopsis>
				</configOption>
				<configOption name="custom_autoanswer" default="no">
					<synopsis>If yes, add the Call-Info header to INVITE message to indicate the endpoint to auto answer</synopsis>
                </configOption>
				<configOption name="allow_set_hb_state" default="yes">
					<synopsis>If yes, set extension state according to options </synopsis>
				</configOption>
				<configOption name="rewrite_contact">
					<synopsis>Allow Contact header to be rewritten with the source IP address-port</synopsis>
					<description><para>
						On inbound SIP messages from this endpoint, the Contact header or an
						appropriate Record-Route header will be changed to have the source IP
						address and port.  This option does not affect outbound messages sent to
						this endpoint.  This option helps servers communicate with endpoints
						that are behind NATs.  This option also helps reuse reliable transport
						connections such as TCP and TLS.
					</para></description>
				</configOption>
				<configOption name="rtp_ipv6" default="no">
					<synopsis>Allow use of IPv6 for RTP traffic</synopsis>
				</configOption>
				<configOption name="rtp_symmetric" default="no">
					<synopsis>Enforce that RTP must be symmetric</synopsis>
				</configOption>
				<configOption name="send_diversion" default="yes">
					<synopsis>Send the Diversion header, conveying the diversion
					information to the called user agent</synopsis>
				</configOption>
				<configOption name="send_pai" default="no">
					<synopsis>Send the P-Asserted-Identity header</synopsis>
				</configOption>
				<configOption name="passthrough_pai" default="yes">
					<synopsis>Passthrough the P-Asserted-Identity header from one side to the other side.</synopsis>
				</configOption>
				<configOption name="pai_number" default="">
					<synopsis>The user string specified in PAI header</synopsis>
				</configOption>
				<configOption name="send_ppi" default="no">
					<synopsis>Send the P-Preferred-Identity header</synopsis>
				</configOption>				
				<configOption name="send_pani" default="no">
					<synopsis>Send the P-Access-Network-Info header</synopsis>
				</configOption>
				<configOption name="access_info" default="">
					<synopsis>The access info string specified in P-Access-Network-Info header</synopsis>
				</configOption>
				<configOption name="send_anonymous" default="no">
					<synopsis>Call out with caller id blocked</synopsis>
				</configOption>
				<configOption name="send_pcpid" default="no">
					<synopsis>Send the P-Called-Party-ID header</synopsis>
				</configOption>
				<configOption name="use_for_sfu_conf" default="no">
					<synopsis>trunk enable configure, will only use for sfu conference</synopsis>
				</configOption>
				<configOption name="trickle_ice" default="yes">
					<synopsis>Endpoint enable configure, will only use trickle ice for gswave</synopsis>
				</configOption>
				<configOption name="block_backward_collectcall" default="no">
					<synopsis>To indicate whether to block the call with backward collect call or not</synopsis>
				</configOption>
				<configOption name="nack_enable" default="no">
					<synopsis>enable nack</synopsis>
				</configOption>
				<configOption name="use_dod_in_ppi" default="no">
					<synopsis>Use DOD number in PPI header if DOD exists</synopsis>
				</configOption>
				<configOption name="use_origcid_in_ppi" default="no">
					<synopsis>Use Original CallerID in PPI header if the call if from trunk</synopsis>
				</configOption>
				<configOption name="use_dod_as_from_name" default="no">
					<synopsis>Use DOD number in From header as display info if DOD exists and From user is not null </synopsis>
				</configOption>
				<configOption name="use_dod_as_contact_user" default="no">
					<synopsis>Use DOD number in Contact header as user if DOD exists </synopsis>
				</configOption>
				<configOption name="send_rpid" default="no">
					<synopsis>Send the Remote-Party-ID header</synopsis>
                </configOption>
				<configOption name="sca_enable" default="no">
					<synopsis>sca enable configure</synopsis>
				</configOption>
				<configOption name="sca_role" default="none">
					<synopsis>sca role none/shared/private</synopsis>
				</configOption>
				<configOption name="sca_sharedline" default="">
					<synopsis>sca sharedline for sca private role</synopsis>
				</configOption>
				<configOption name="cancel_others_if_one_reject" default="no">
					<synopsis>If one terminal reject the call, cancel the calls to the other terminals while the option enabled, </synopsis>
				</configOption>
				<configOption name="ignore_180_after_183" default="no">
					<synopsis>If 183 has received, ignore the following 180 response</synopsis>
				</configOption>
				<configOption name="strategy_ipacl" default="0">
					<synopsis>The strategy of ip acl, 0 for all, 1 for ip whitelist, 3 for whitelist and remote connect address</synopsis>
				</configOption>
				<configOption name="support_video_call" default="no">
					<synopsis>If enable, webrtc trunk can support video call</synopsis>
				</configOption>
				<configOption name="remote_mailbox" default="">
					<synopsis>enable remote mail box</synopsis>
				</configOption>
				<configOption name="remote_mailbox" default="">
					<synopsis>enable remote mail box</synopsis>
				</configOption>
				<configOption name="jitter_buffer_type" default="disable">
					<synopsis>jitter buffer type disable/normal/neteq</synopsis>
				</configOption>
				<configOption name="jitter_buffer_normal_type" default="">
					<synopsis>if select normal mode fixed/adaptive</synopsis>
				</configOption>
				<configOption name="jitter_buffer_normal_size" default="100">
					<synopsis>if select normal fixed mode, the jitter buffer size</synopsis>
				</configOption>
				<configOption name="jitter_buffer_normal_max_size" default="100">
					<synopsis>if select normal adaptive mode, the jitter buffer max size</synopsis>
				</configOption>
				<configOption name="rpid_immediate" default="no">
					<synopsis>Immediately send connected line updates on unanswered incoming calls.</synopsis>
					<description>
						<para>When enabled, immediately send <emphasis>180 Ringing</emphasis>
						or <emphasis>183 Progress</emphasis> response messages to the
						caller if the connected line information is updated before
						the call is answered.  This can send a <emphasis>180 Ringing</emphasis>
						response before the call has even reached the far end.  The
						caller can start hearing ringback before the far end even gets
						the call.  Many phones tend to grab the first connected line
						information and refuse to update the display if it changes.  The
						first information is not likely to be correct if the call
						goes to an endpoint not under the control of this Asterisk
						box.</para>
						<para>When disabled, a connected line update must wait for
						another reason to send a message with the connected line
						information to the caller before the call is answered.  You can
						trigger the sending of the information by using an appropriate
						dialplan application such as <emphasis>Ringing</emphasis>.</para>
					</description>
				</configOption>
				<configOption name="timers_min_se" default="90">
					<synopsis>Minimum session timers expiration period</synopsis>
					<description><para>
						Minimum session timer expiration period. Time in seconds.
					</para></description>
				</configOption>
				<configOption name="timers" default="yes">
					<synopsis>Session timers for SIP packets</synopsis>
					<description>
						<enumlist>
							<enum name="no" />
							<enum name="yes" />
							<enum name="required" />
							<enum name="always" />
							<enum name="forced"><para>Alias of always</para></enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="timers_sess_expires" default="1800">
					<synopsis>Maximum session timer expiration period</synopsis>
					<description><para>
						Maximum session timer expiration period. Time in seconds.
					</para></description>
				</configOption>
				<configOption name="transport">
					<synopsis>Explicit transport configuration to use</synopsis>
					<description>
						<para>This will <emphasis>force</emphasis> the endpoint to use the
						specified transport configuration to send SIP messages.  You need
						to already know what kind of transport (UDP/TCP/IPv4/etc) the
						endpoint device will use.
						</para>
						<note><para>Not specifying a transport will select the first
						configured transport in <filename>pjsip.conf</filename> which is
						compatible with the URI we are trying to contact.
						</para></note>
						<warning><para>Transport configuration is not affected by reloads. In order to
						change transports, a full Asterisk restart is required</para></warning>
					</description>
				</configOption>
				<configOption name="trust_id_inbound" default="no">
					<synopsis>Accept identification information received from this endpoint</synopsis>
					<description><para>This option determines whether Asterisk will accept
					identification from the endpoint from headers such as P-Asserted-Identity
					or Remote-Party-ID header. This option applies both to calls originating from the
					endpoint and calls originating from Asterisk. If <literal>no</literal>, the
					configured Caller-ID from pjsip.conf will always be used as the identity for
					the endpoint.</para></description>
				</configOption>
				<configOption name="trust_id_outbound" default="no">
					<synopsis>Send private identification details to the endpoint.</synopsis>
					<description><para>This option determines whether res_pjsip will send private
					identification information to the endpoint. If <literal>no</literal>,
					private Caller-ID information will not be forwarded to the endpoint.
					"Private" in this case refers to any method of restricting identification.
					Example: setting <replaceable>callerid_privacy</replaceable> to any
					<literal>prohib</literal> variation.
					Example: If <replaceable>trust_id_inbound</replaceable> is set to
					<literal>yes</literal>, the presence of a <literal>Privacy: id</literal>
					header in a SIP request or response would indicate the identification
					provided in the request is private.</para></description>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'endpoint'.</synopsis>
				</configOption>
				<configOption name="use_ptime" default="yes">
					<synopsis>Use Endpoint's requested packetization interval</synopsis>
				</configOption>
				<configOption name="use_avpf" default="no">
					<synopsis>Determines whether res_pjsip will use and enforce usage of AVPF for this
					endpoint.</synopsis>
					<description><para>
						If set to <literal>yes</literal>, res_pjsip will use the AVPF or SAVPF RTP
						profile for all media offers on outbound calls and media updates and will
						decline media offers not using the AVPF or SAVPF profile.
					</para><para>
						If set to <literal>no</literal>, res_pjsip will use the AVP or SAVP RTP
						profile for all media offers on outbound calls and media updates, and will
						decline media offers not using the AVP or SAVP profile.
					</para></description>
				</configOption>
				<configOption name="configure_use_avpf" default="no">
					<synopsis>Determines whether res_pjsip will use and enforce usage of AVPF for this
					endpoint when not enable webrtc.</synopsis>
					<description><para>
						If set to <literal>yes</literal>, res_pjsip will use the AVPF or SAVPF RTP
						profile for all media offers on outbound calls and media updates and will
						decline media offers not using the AVPF or SAVPF profile.
					</para><para>
						If set to <literal>no</literal>, res_pjsip will use the AVP or SAVP RTP
						profile for all media offers on outbound calls and media updates, and will
						decline media offers not using the AVP or SAVP profile.
					</para></description>
				</configOption>
				<configOption name="force_avp" default="no">
					<synopsis>Determines whether res_pjsip will use and enforce usage of AVP,
					regardless of the RTP profile in use for this endpoint.</synopsis>
					<description><para>
						If set to <literal>yes</literal>, res_pjsip will use the AVP, AVPF, SAVP, or
						SAVPF RTP profile for all media offers on outbound calls and media updates including
						those for DTLS-SRTP streams.
					</para><para>
						If set to <literal>no</literal>, res_pjsip will use the respective RTP profile
						depending on configuration.
					</para></description>
				</configOption>
				<configOption name="media_use_received_transport" default="no">
					<synopsis>Determines whether res_pjsip will use the media transport received in the
					offer SDP in the corresponding answer SDP.</synopsis>
					<description><para>
						If set to <literal>yes</literal>, res_pjsip will use the received media transport.
					</para><para>
						If set to <literal>no</literal>, res_pjsip will use the respective RTP profile
						depending on configuration.
					</para></description>
				</configOption>
				<configOption name="configure_media_use_received_transport" default="no">
					<synopsis>To record configure media_use_received_transport, webrtc will change this
					to yes default, if sip phone use the same extension with webrtc, need to change as
					configure value.</synopsis>
					<description><para>
						If set to <literal>yes</literal>, res_pjsip will use the received media transport.
					</para><para>
						If set to <literal>no</literal>, res_pjsip will use the respective RTP profile
						depending on configuration.
					</para></description>
				</configOption>
				<configOption name="media_encryption" default="no">
					<synopsis>Determines whether res_pjsip will use and enforce usage of media encryption
					for this endpoint.</synopsis>
					<description>
						<enumlist>
							<enum name="no"><para>
								res_pjsip will offer no encryption and allow no encryption to be setup.
							</para></enum>
							<enum name="sdes"><para>
								res_pjsip will offer standard SRTP setup via in-SDP keys. Encrypted SIP
								transport should be used in conjunction with this option to prevent
								exposure of media encryption keys.
							</para></enum>
							<enum name="dtls"><para>
								res_pjsip will offer DTLS-SRTP setup.
							</para></enum>
							<enum name="auot none"><para>
								res_pjsip will offer no encryption for media.
							</para></enum>
							<enum name="auot srtp"><para>
								res_pjsip will offer srtp encryption for media.
							</para></enum>
							<enum name="auot dtls"><para>
								res_pjsip will offer dtls encryption for media.
							</para></enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="configure_media_encryption" default="no">
					<synopsis>To record configure media_encryption, webrtc will change this to auto_dtls
					as default, if sip phone use the same extension with webrtc, need to change as
					configure value.</synopsis>
					<description>
						<enumlist>
							<enum name="no"><para>
								res_pjsip will offer no encryption and allow no encryption to be setup.
							</para></enum>
							<enum name="sdes"><para>
								res_pjsip will offer standard SRTP setup via in-SDP keys. Encrypted SIP
								transport should be used in conjunction with this option to prevent
								exposure of media encryption keys.
							</para></enum>
							<enum name="dtls"><para>
								res_pjsip will offer DTLS-SRTP setup.
							</para></enum>
							<enum name="auot none"><para>
								res_pjsip will offer no encryption for media.
							</para></enum>
							<enum name="auot srtp"><para>
								res_pjsip will offer srtp encryption for media.
							</para></enum>
							<enum name="auot dtls"><para>
								res_pjsip will offer dtls encryption for media.
							</para></enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="media_encryption_optimistic" default="no">
					<synopsis>Determines whether encryption should be used if possible but does not terminate the
					session if not achieved.</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>sdes</literal> or <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="g726_non_standard" default="no">
					<synopsis>Force g.726 to use AAL2 packing order when negotiating g.726 audio</synopsis>
					<description><para>
						When set to "yes" and an endpoint negotiates g.726 audio then use g.726 for AAL2
						packing order instead of what is recommended by RFC3551. Since this essentially
						replaces the underlying 'g726' codec with 'g726aal2' then 'g726aal2' needs to be
						specified in the endpoint's allowed codec list.
					</para></description>
				</configOption>
				<configOption name="inband_progress" default="no">
					<synopsis>Determines whether chan_pjsip will indicate ringing using inband
						progress.</synopsis>
					<description><para>
						If set to <literal>yes</literal>, chan_pjsip will send a 183 Session Progress
						when told to indicate ringing and will immediately start sending ringing
						as audio.
					</para><para>
						If set to <literal>no</literal>, chan_pjsip will send a 180 Ringing when told
						to indicate ringing and will NOT send it as audio.
					</para></description>
				</configOption>
				<configOption name="call_group">
					<synopsis>The numeric pickup groups for a channel.</synopsis>
					<description><para>
						Can be set to a comma separated list of numbers or ranges between the values
						of 0-63 (maximum of 64 groups).
					</para></description>
				</configOption>
				<configOption name="pickup_group">
					<synopsis>The numeric pickup groups that a channel can pickup.</synopsis>
					<description><para>
						Can be set to a comma separated list of numbers or ranges between the values
						of 0-63 (maximum of 64 groups).
					</para></description>
				</configOption>
				<configOption name="named_call_group">
					<synopsis>The named pickup groups for a channel.</synopsis>
					<description><para>
						Can be set to a comma separated list of case sensitive strings limited by
						supported line length.
					</para></description>
				</configOption>
				<configOption name="named_pickup_group">
					<synopsis>The named pickup groups that a channel can pickup.</synopsis>
					<description><para>
						Can be set to a comma separated list of case sensitive strings limited by
						supported line length.
					</para></description>
				</configOption>
				<configOption name="device_state_busy_at" default="0">
					<synopsis>The number of in-use channels which will cause busy to be returned as device state</synopsis>
					<description><para>
						When the number of in-use channels for the endpoint matches the devicestate_busy_at setting the
						PJSIP channel driver will return busy as the device state instead of in use.
					</para></description>
				</configOption>
				<configOption name="t38_udptl" default="no">
					<synopsis>Whether T.38 UDPTL support is enabled or not</synopsis>
					<description><para>
						If set to yes T.38 UDPTL support will be enabled, and T.38 negotiation requests will be accepted
						and relayed.
					</para></description>
				</configOption>
				<configOption name="t38_udptl_ec" default="none">
					<synopsis>T.38 UDPTL error correction method</synopsis>
					<description>
						<enumlist>
							<enum name="none"><para>
								No error correction should be used.
							</para></enum>
							<enum name="fec"><para>
								Forward error correction should be used.
							</para></enum>
							<enum name="redundancy"><para>
								Redundancy error correction should be used.
							</para></enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="t38_udptl_maxdatagram" default="0">
					<synopsis>T.38 UDPTL maximum datagram size</synopsis>
					<description><para>
						This option can be set to override the maximum datagram of a remote endpoint for broken
						endpoints.
					</para></description>
				</configOption>
				<configOption name="fax_detect" default="no">
					<synopsis>Whether CNG tone detection is enabled</synopsis>
					<description><para>
						This option can be set to send the session to the fax extension when a CNG tone is
						detected.
					</para></description>
				</configOption>
				<configOption name="fax_detect_timeout">
					<synopsis>How long into a call before fax_detect is disabled for the call</synopsis>
					<description><para>
						The option determines how many seconds into a call before the
						fax_detect option is disabled for the call.  Setting the value
						to zero disables the timeout.
					</para></description>
				</configOption>
				<configOption name="t38_udptl_nat" default="no">
					<synopsis>Whether NAT support is enabled on UDPTL sessions</synopsis>
					<description><para>
						When enabled the UDPTL stack will send UDPTL packets to the source address of
						received packets.
					</para></description>
				</configOption>
				<configOption name="t38_udptl_ipv6" default="no">
					<synopsis>Whether IPv6 is used for UDPTL Sessions</synopsis>
					<description><para>
						When enabled the UDPTL stack will use IPv6.
					</para></description>
				</configOption>
				<configOption name="tone_zone">
					<synopsis>Set which country's indications to use for channels created for this endpoint.</synopsis>
				</configOption>
				<configOption name="language">
					<synopsis>Set the default language to use for channels created for this endpoint.</synopsis>
				</configOption>
				<configOption name="one_touch_recording" default="no">
					<synopsis>Determines whether one-touch recording is allowed for this endpoint.</synopsis>
					<see-also>
						<ref type="configOption">record_on_feature</ref>
						<ref type="configOption">record_off_feature</ref>
					</see-also>
				</configOption>
				<configOption name="record_on_feature" default="automixmon">
					<synopsis>The feature to enact when one-touch recording is turned on.</synopsis>
					<description>
						<para>When an INFO request for one-touch recording arrives with a Record header set to "on", this
						feature will be enabled for the channel. The feature designated here can be any built-in
						or dynamic feature defined in features.conf.</para>
						<note><para>This setting has no effect if the endpoint's one_touch_recording option is disabled</para></note>
					</description>
					<see-also>
						<ref type="configOption">one_touch_recording</ref>
						<ref type="configOption">record_off_feature</ref>
					</see-also>
				</configOption>
				<configOption name="record_off_feature" default="automixmon">
					<synopsis>The feature to enact when one-touch recording is turned off.</synopsis>
					<description>
						<para>When an INFO request for one-touch recording arrives with a Record header set to "off", this
						feature will be enabled for the channel. The feature designated here can be any built-in
						or dynamic feature defined in features.conf.</para>
						<note><para>This setting has no effect if the endpoint's one_touch_recording option is disabled</para></note>
					</description>
					<see-also>
						<ref type="configOption">one_touch_recording</ref>
						<ref type="configOption">record_on_feature</ref>
					</see-also>
				</configOption>
				<configOption name="rtp_engine" default="asterisk">
					<synopsis>Name of the RTP engine to use for channels created for this endpoint</synopsis>
				</configOption>
				<configOption name="allow_transfer" default="yes">
					<synopsis>Determines whether SIP REFER transfers are allowed for this endpoint</synopsis>
				</configOption>
				<configOption name="user_eq_phone" default="no">
					<synopsis>Determines whether a user=phone parameter is placed into the request URI if the user is determined to be a phone number</synopsis>
				</configOption>
				<configOption name="moh_passthrough" default="no">
					<synopsis>Determines whether hold and unhold will be passed through using re-INVITEs with recvonly and sendrecv to the remote side</synopsis>
				</configOption>
				<configOption name="sdp_owner" default="-">
					<synopsis>String placed as the username portion of an SDP origin (o=) line.</synopsis>
				</configOption>
				<configOption name="sdp_session" default="Asterisk">
					<synopsis>String used for the SDP session (s=) line.</synopsis>
				</configOption>
				<configOption name="tos_audio">
					<synopsis>DSCP TOS bits for audio streams</synopsis>
					<description><para>
						See https://wiki.asterisk.org/wiki/display/AST/IP+Quality+of+Service for more information about QoS settings
					</para></description>
				</configOption>
				<configOption name="tos_video">
					<synopsis>DSCP TOS bits for video streams</synopsis>
					<description><para>
						See https://wiki.asterisk.org/wiki/display/AST/IP+Quality+of+Service for more information about QoS settings
					</para></description>
				</configOption>
				<configOption name="cos_audio">
					<synopsis>Priority for audio streams</synopsis>
					<description><para>
						See https://wiki.asterisk.org/wiki/display/AST/IP+Quality+of+Service for more information about QoS settings
					</para></description>
				</configOption>
				<configOption name="cos_video">
					<synopsis>Priority for video streams</synopsis>
					<description><para>
						See https://wiki.asterisk.org/wiki/display/AST/IP+Quality+of+Service for more information about QoS settings
					</para></description>
				</configOption>
				<configOption name="allow_subscribe" default="yes">
					<synopsis>Determines if endpoint is allowed to initiate subscriptions with Asterisk.</synopsis>
				</configOption>
				<configOption name="sub_min_expiry" default="60">
					<synopsis>The minimum allowed expiry time for subscriptions initiated by the endpoint.</synopsis>
				</configOption>
				<configOption name="sub_max_expiry" default="3600">
					<synopsis>The maximum allowed expiry time for subscriptions initiated by the endpoint.</synopsis>
                </configOption>
				<configOption name="from_user">
					<synopsis>Username to use in From header for requests to this endpoint.</synopsis>
				</configOption>
				<configOption name="mwi_from_user">
					<synopsis>Username to use in From header for unsolicited MWI NOTIFYs to this endpoint.</synopsis>
				</configOption>
				<configOption name="from_domain">
					<synopsis>Domain to user in From header for requests to this endpoint.</synopsis>
				</configOption>
				<configOption name="dtls_verify">
					<synopsis>Verify that the provided peer certificate is valid</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="dtls_rekey">
					<synopsis>Interval at which to renegotiate the TLS session and rekey the SRTP session</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para><para>
						If this is not set or the value provided is 0 rekeying will be disabled.
					</para></description>
				</configOption>
				<configOption name="dtls_auto_generate_cert" default="no">
					<synopsis>Whether or not to automatically generate an ephemeral X.509 certificate</synopsis>
					<description>
						<para>
							If enabled, Asterisk will generate an X.509 certificate for each DTLS session.
							This option only applies if <replaceable>media_encryption</replaceable> is set
							to <literal>dtls</literal>. This option will be automatically enabled if
							<literal>webrtc</literal> is enabled and <literal>dtls_cert_file</literal> is
							not specified.
						</para>
					</description>
				</configOption>
				<configOption name="dtls_cert_file">
					<synopsis>Path to certificate file to present to peer</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="dtls_private_key">
					<synopsis>Path to private key for certificate file</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="dtls_cipher">
					<synopsis>Cipher to use for DTLS negotiation</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para>
					<para>Many options for acceptable ciphers. See link for more:</para>
					<para>http://www.openssl.org/docs/apps/ciphers.html#CIPHER_STRINGS
					</para></description>
				</configOption>
				<configOption name="dtls_ca_file">
					<synopsis>Path to certificate authority certificate</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="dtls_ca_path">
					<synopsis>Path to a directory containing certificate authority certificates</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="dtls_setup">
					<synopsis>Whether we are willing to accept connections, connect to the other party, or both.</synopsis>
					<description>
						<para>
							This option only applies if <replaceable>media_encryption</replaceable> is
							set to <literal>dtls</literal>.
						</para>
						<enumlist>
							<enum name="active"><para>
								res_pjsip will make a connection to the peer.
							</para></enum>
							<enum name="passive"><para>
								res_pjsip will accept connections from the peer.
							</para></enum>
							<enum name="actpass"><para>
								res_pjsip will offer and accept connections from the peer.
							</para></enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="dtls_fingerprint">
					<synopsis>Type of hash to use for the DTLS fingerprint in the SDP.</synopsis>
					<description>
						<para>
							This option only applies if <replaceable>media_encryption</replaceable> is
							set to <literal>dtls</literal>.
						</para>
						<enumlist>
							<enum name="SHA-256"></enum>
							<enum name="SHA-1"></enum>
						</enumlist>
					</description>
				</configOption>
				<configOption name="srtp_tag_32">
					<synopsis>Determines whether 32 byte tags should be used instead of 80 byte tags.</synopsis>
					<description><para>
						This option only applies if <replaceable>media_encryption</replaceable> is
						set to <literal>sdes</literal> or <literal>dtls</literal>.
					</para></description>
				</configOption>
				<configOption name="set_var">
					<synopsis>Variable set on a channel involving the endpoint.</synopsis>
					<description><para>
						When a new channel is created using the endpoint set the specified
						variable(s) on that channel. For multiple channel variables specify
						multiple 'set_var'(s).
					</para></description>
				</configOption>
				<configOption name="message_context">
					<synopsis>Context to route incoming MESSAGE requests to.</synopsis>
					<description><para>
						If specified, incoming MESSAGE requests will be routed to the indicated
						dialplan context. If no <replaceable>message_context</replaceable> is
						specified, then the <replaceable>context</replaceable> setting is used.
					</para></description>
				</configOption>
				<configOption name="accountcode">
					<synopsis>An accountcode to set automatically on any channels created for this endpoint.</synopsis>
					<description><para>
						If specified, any channel created for this endpoint will automatically
						have this accountcode set on it.
					</para></description>
				</configOption>
				<configOption name="cc_agent_policy">
					<synopsis>the policy of ccss agent</synopsis>
					<description>
						<enumlist>
							<enum name="never" />
							<enum name="native" />
							<enum name="generic" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="cc_monitor_policy">
					<synopsis>the policy of ccss monitor</synopsis>
					<description>
						<enumlist>
							<enum name="never" />
							<enum name="native" />
							<enum name="generic" />
							<enum name="always" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="ccbs_available_timer" default="4800">
					<synopsis>the available timer of ccbs</synopsis>
				</configOption>
				<configOption name="ccnr_available_timer" default="7200">
					<synopsis>the available timer of ccnr</synopsis>
				</configOption>
				<configOption name="cc_offer_timer" default="120">
					<synopsis>the offered timer of ccss</synopsis>
				</configOption>
				<configOption name="cc_max_agents" default="1">
					<synopsis>the max number of cc agents</synopsis>
				</configOption>
				<configOption name="cc_max_monitors" default="2">
					<synopsis>the max number of cc monitors</synopsis>
				</configOption>
				<configOption name="cc_recall_timer" default="60">
					<synopsis>the recalled timer of ccss</synopsis>
				</configOption>
				<configOption name="username">
					<synopsis>the username of the register trunk</synopsis>
				</configOption>
				<configOption name="tel_uri" default="disabled">
    				<synopsis>the tel uri mode</synopsis>
					<description>
						<enumlist>
							<enum name="disabled" />
							<enum name="enabled" />
							<enum name="user_phone" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="preferred_codec_only" default="no">
					<synopsis>Respond to a SIP invite with the single most preferred codec rather than advertising all joint codec capabilities. This limits the other side's codec choice to exactly what we prefer.</synopsis>
				</configOption>
				<configOption name="rtp_keepalive">
					<synopsis>Number of seconds between RTP comfort noise keepalive packets.</synopsis>
					<description><para>
						At the specified interval, Asterisk will send an RTP comfort noise frame. This may
						be useful for situations where Asterisk is behind a NAT or firewall and must keep
						a hole open in order to allow for media to arrive at Asterisk.
					</para></description>
				</configOption>
				<configOption name="rtp_timeout" default="0">
					<synopsis>Maximum number of seconds without receiving RTP (while off hold) before terminating call.</synopsis>
					<description><para>
						This option configures the number of seconds without RTP (while off hold) before
						considering a channel as dead. When the number of seconds is reached the underlying
						channel is hung up. By default this option is set to 0, which means do not check.
					</para></description>
				</configOption>
				<configOption name="rtp_timeout_hold" default="0">
					<synopsis>Maximum number of seconds without receiving RTP (while on hold) before terminating call.</synopsis>
					<description><para>
						This option configures the number of seconds without RTP (while on hold) before
						considering a channel as dead. When the number of seconds is reached the underlying
						channel is hung up. By default this option is set to 0, which means do not check.
					</para></description>
				</configOption>
				<configOption name="acl">
					<synopsis>List of IP ACL section names in acl.conf</synopsis>
					<description><para>
						This matches sections configured in <literal>acl.conf</literal>. The value is
						defined as a list of comma-delimited section names.
					</para></description>
				</configOption>
				<configOption name="deny">
					<synopsis>List of IP addresses to deny access from</synopsis>
					<description><para>
						The value is a comma-delimited list of IP addresses. IP addresses may
						have a subnet mask appended. The subnet mask may be written in either
						CIDR or dotted-decimal notation. Separate the IP address and subnet
						mask with a slash ('/')
					</para></description>
				</configOption>
				<configOption name="permit">
					<synopsis>List of IP addresses to permit access from</synopsis>
					<description><para>
						The value is a comma-delimited list of IP addresses. IP addresses may
						have a subnet mask appended. The subnet mask may be written in either
						CIDR or dotted-decimal notation. Separate the IP address and subnet
						mask with a slash ('/')
					</para></description>
				</configOption>
				<configOption name="contact_acl">
					<synopsis>List of Contact ACL section names in acl.conf</synopsis>
					<description><para>
						This matches sections configured in <literal>acl.conf</literal>. The value is
						defined as a list of comma-delimited section names.
					</para></description>
				</configOption>
				<configOption name="contact_deny">
					<synopsis>List of Contact header addresses to deny</synopsis>
					<description><para>
						The value is a comma-delimited list of IP addresses. IP addresses may
						have a subnet mask appended. The subnet mask may be written in either
						CIDR or dotted-decimal notation. Separate the IP address and subnet
						mask with a slash ('/')
					</para></description>
				</configOption>
				<configOption name="contact_permit">
					<synopsis>List of Contact header addresses to permit</synopsis>
					<description><para>
						The value is a comma-delimited list of IP addresses. IP addresses may
						have a subnet mask appended. The subnet mask may be written in either
						CIDR or dotted-decimal notation. Separate the IP address and subnet
						mask with a slash ('/')
					</para></description>
				</configOption>
				<configOption name="subscribe_context">
					<synopsis>Context for incoming MESSAGE requests.</synopsis>
					<description><para>
						If specified, incoming SUBSCRIBE requests will be searched for the matching
						extension in the indicated context.
						If no <replaceable>subscribe_context</replaceable> is specified,
						then the <replaceable>context</replaceable> setting is used.
					</para></description>
				</configOption>
				<configOption name="contact_user" default="">
					<synopsis>Force the user on the outgoing Contact header to this value.</synopsis>
					<description><para>
						On outbound requests, force the user portion of the Contact header to this value.
					</para></description>
				</configOption>
				<configOption name="asymmetric_rtp_codec" default="no">
					<synopsis>Allow the sending and receiving RTP codec to differ</synopsis>
					<description><para>
						When set to "yes" the codec in use for sending will be allowed to differ from
						that of the received one. PJSIP will not automatically switch the sending one
						to the receiving one.
					</para></description>
				</configOption>
				<configOption name="rtcp_mux" default="no">
					<synopsis>Enable RFC 5761 RTCP multiplexing on the RTP port</synopsis>
					<description><para>
						With this option enabled, Asterisk will attempt to negotiate the use of the "rtcp-mux"
						attribute on all media streams. This will result in RTP and RTCP being sent and received
						on the same port. This shifts the demultiplexing logic to the application rather than
						the transport layer. This option is useful when interoperating with WebRTC endpoints
						since they mandate this option's use.
					</para></description>
				</configOption>
				<configOption name="refer_blind_progress" default="yes">
					<synopsis>Whether to notifies all the progress details on blind transfer</synopsis>
					<description><para>
						Some SIP phones (Mitel/Aastra, Snom) expect a sip/frag "200 OK"
						after REFER has been accepted. If set to <literal>no</literal> then asterisk
						will not send the progress details, but immediately will send "200 OK".
					</para></description>
				</configOption>
				<configOption name="notify_early_inuse_ringing" default="no">
					<synopsis>Whether to notifies dialog-info 'early' on InUse&amp;Ringing state</synopsis>
					<description><para>
						Control whether dialog-info subscriptions get 'early' state
						on Ringing when already INUSE.
					</para></description>
				</configOption>
				<configOption name="max_audio_streams" default="1">
					<synopsis>The maximum number of allowed audio streams for the endpoint</synopsis>
					<description><para>
						This option enforces a limit on the maximum simultaneous negotiated audio
						streams allowed for the endpoint.
					</para></description>
				</configOption>
				<configOption name="max_video_streams" default="1">
					<synopsis>The maximum number of allowed video streams for the endpoint</synopsis>
					<description><para>
						This option enforces a limit on the maximum simultaneous negotiated video
						streams allowed for the endpoint.
					</para></description>
				</configOption>
				<configOption name="bundle" default="no">
					<synopsis>Enable RTP bundling</synopsis>
					<description><para>
						With this option enabled, Asterisk will attempt to negotiate the use of bundle.
						If negotiated this will result in multiple RTP streams being carried over the same
						underlying transport. Note that enabling bundle will also enable the rtcp_mux option.
					</para></description>
				</configOption>
				<configOption name="gswave_bundle_enabled" default="yes">
					<synopsis>Enable gswave RTP bundling</synopsis>
					<description><para>
						With this option enabled, Asterisk will attempt to negotiate the use of bundle.
						If negotiated this will result in multiple RTP streams being carried over the same
						underlying transport. Note that enabling bundle will also enable the rtcp_mux option.
					</para></description>
				</configOption>
				<configOption name="bfcp_support" default="yes">
					<synopsis>Enable BFCP support</synopsis>
					<description><para>
						With this option enabled, Asterisk can accept BFCP media.
						otherwise will reject BFCP media.
					</para></description>
				</configOption>
				<configOption name="click2call_enable" default="no">
					<synopsis>Enable Click2Call</synopsis>
					<description><para>
						The endpoint is used for click2call.
					</para></description>
				</configOption>
				<configOption name="mobile_bundle_enabled" default="no">
					<synopsis>Enable mobile RTP bundling</synopsis>
					<description><para>
						With this option enabled, Asterisk will attempt to negotiate the use of bundle.
						If negotiated this will result in multiple RTP streams being carried over the same
						underlying transport. Note that enabling bundle will also enable the rtcp_mux option.
					</para></description>
				</configOption>
				<configOption name="dtx" default="no">
					<synopsis>Enable opus dtx</synopsis>
					<description><para>
						With this option enabled, AVS can decode opus as encode dtx mode.
					</para></description>
				</configOption>
				<configOption name="turn" default="no">
					<synopsis>Enable turn relay for trunk</synopsis>
					<description><para>
						With this option enabled, media will relayed by turn server.
					</para></description>
				</configOption>
				<configOption name="support_special_attrs" default="yes">
					<synopsis>Support ssrc/msid/mid/CT/AS/TIAS/record attrs in SDP</synopsis>
					<description><para>
						With this option enabled, SDP will add those attrs.
					</para></description>
				</configOption>
				<configOption name="support_multi_pt" default="yes">
					<synopsis>Support H264 and RTX codec has multi pt</synopsis>
					<description><para>
						With this option enabled, UCM will forward multi pt to callee
					</para></description>
				</configOption>
				<configOption name="comfort_noise" default="yes">
					<synopsis>Support comfort noise</synopsis>
					<description><para>
						With this option enabled, UCM will send comfort_noise RTP
					</para></description>
				</configOption>
				<configOption name="mpa" default="no">
					<synopsis>Support PMA codec for PM3 play</synopsis>
					<description><para>
						With this option enabled, UCM can negotiate PMA codec
					</para></description>
				</configOption>
				<configOption name="stir_shaken" default="no">
					<synopsis>Enable stir/shaken function</synopsis>
					<description><para>
						With this option enabled, when incoming call use the voip trunk, asterisk will 
						check the indentity header in sip INVITE message. when outgoing call, asterisk 
						will add indentity header in sip INVITE message.
					</para></description>
				</configOption>
				<configOption name="option_failed_count" default="3">
					<synopsis>Send option failed count</synopsis>
					<description><para>
						When try to send option failed because of has pending option, we will not
						send option, but when try count reach the configure, we will send again.
					</para></description>
				</configOption>
				<configOption name="webrtc" default="no">
					<synopsis>Defaults and enables some options that are relevant to WebRTC</synopsis>
					<description><para>
						When set to "yes" this also enables the following values that are needed in
						order for basic WebRTC support to work: rtcp_mux, use_avpf, ice_support, and
						use_received_transport. The following configuration settings also get defaulted
						as follows:</para>
						<para>media_encryption=dtls</para>
						<para>dtls_auto_generate_cert=yes (if dtls_cert_file is not set)</para>
						<para>dtls_verify=fingerprint</para>
						<para>dtls_setup=actpass</para>
					</description>
				</configOption>
				<configOption name="incoming_mwi_mailbox">
					<synopsis>Mailbox name to use when incoming MWI NOTIFYs are received</synopsis>
					<description><para>
						If an MWI NOTIFY is received <emphasis>from</emphasis> this endpoint,
						this mailbox will be used when notifying other modules of MWI status
						changes.  If not set, incoming MWI NOTIFYs are ignored.
					</para></description>
				</configOption>
				<configOption name="follow_early_media_fork">
					<synopsis>Follow SDP forked media when To tag is different</synopsis>
					<description><para>
						On outgoing calls, if the UAS responds with different SDP attributes
						on subsequent 18X or 2XX responses (such as a port update) AND the
						To tag on the subsequent response is different than that on the previous
						one, follow it. This usually happens when the INVITE is forked to multiple
						UASs and more than one sends an SDP answer.
						</para>
						<note><para>
							This option must also be enabled in the <literal>system</literal>
							section for it to take effect here.
						</para></note>
					</description>
				</configOption>
				<configOption name="accept_multiple_sdp_answers" default="no">
					<synopsis>Accept multiple SDP answers on non-100rel responses</synopsis>
					<description><para>
						On outgoing calls, if the UAS responds with different SDP attributes
						on non-100rel 18X or 2XX responses (such as a port update) AND the
						To tag on the subsequent response is the same as that on the previous one,
						process the updated SDP.  This can happen when the UAS needs to change ports
						for some reason such as using a separate port for custom ringback.
						</para>
						<note><para>
							This option must also be enabled in the <literal>system</literal>
							section for it to take effect here.
						</para></note>
					</description>
				</configOption>
				<configOption name="suppress_q850_reason_headers" default="no">
					<synopsis>Suppress Q.850 Reason headers for this endpoint</synopsis>
					<description><para>
						Some devices can't accept multiple Reason headers and get confused
						when both 'SIP' and 'Q.850' Reason headers are received.  This
						option allows the 'Q.850' Reason header to be suppressed.</para>
					</description>
				</configOption>
				<configOption name="sbc_rand_addr" default="">
					<synopsis>The sbc_rand_addr for this udp trunk endpoint</synopsis>
					<description><para>
						The CloudUCM will contain the header SBC-Rand-Addr with this value.</para>
					</description>
				</configOption>
				<configOption name="trunk_host" default="">
					<synopsis>The trunk host for this endpoint</synopsis>
					<description><para>
						The CloudUCM will contain the header Trunk-Host with this value.</para>
					</description>
				</configOption>
				<configOption name="local_vlan_address">
					<synopsis>Support use this local address to send SIP message</synopsis>
					<description><para>
						With this option configure, SIP trunk will use this address to send SIP message.</para>
					</description>
				</configOption>
				<configOption name="trunk_type">
					<synopsis>The type of the trunk</synopsis>
					<description><para>
						The value can be peer, register or account. If the endpoint is an extension, the value is null.</para>
					</description>
    			</configOption>
			</configObject>
			<configObject name="auth">
				<synopsis>Authentication type</synopsis>
				<description><para>
					Authentication objects hold the authentication information for use
					by other objects such as <literal>endpoints</literal> or <literal>registrations</literal>.
					This also allows for multiple objects to use a single auth object. See
					the <literal>auth_type</literal> config option for password style choices.
				</para></description>
				<configOption name="auth_type" default="userpass">
					<synopsis>Authentication type</synopsis>
					<description><para>
						This option specifies which of the password style config options should be read
						when trying to authenticate an endpoint inbound request. If set to <literal>userpass</literal>
						then we'll read from the 'password' option. For <literal>md5</literal> we'll read
						from 'md5_cred'.
						</para>
						<enumlist>
							<enum name="md5"/>
							<enum name="userpass"/>
						</enumlist>
					</description>
				</configOption>
				<configOption name="nonce_lifetime" default="60">
					<synopsis>Lifetime of a nonce associated with this authentication config.</synopsis>
				</configOption>
				<configOption name="md5_cred">
					<synopsis>MD5 Hash used for authentication.</synopsis>
					<description><para>Only used when auth_type is <literal>md5</literal>.</para></description>
				</configOption>
				<configOption name="password">
					<synopsis>Plain text password used for authentication.</synopsis>
					<description><para>Only used when auth_type is <literal>userpass</literal>.</para></description>
				</configOption>
				<configOption name="realm">
					<synopsis>SIP realm for endpoint</synopsis>
					<description><para>
						The treatment of this value depends upon how the authentication
						object is used.
						</para><para>
						When used as an inbound authentication object, the realm is sent
						as part of the challenge so the peer can know which key to use
						when responding.  An empty value will use the
						<replaceable>global</replaceable> section's
						<literal>default_realm</literal> value when issuing a challenge.
						</para><para>
						When used as an outbound authentication object, the realm is
						matched with the received challenge realm to determine which
						authentication object to use when responding to the challenge.  An
						empty value matches any challenging realm when determining
						which authentication object matches a received challenge.
						</para>
						<note><para>
						Using the same auth section for inbound and outbound
						authentication is not recommended.  There is a difference in
						meaning for an empty realm setting between inbound and outbound
						authentication uses.</para></note>
					</description>
				</configOption>
				<configOption name="type">
					<synopsis>Must be 'auth'</synopsis>
				</configOption>
				<configOption name="username">
					<synopsis>Username to use for account</synopsis>
				</configOption>
			</configObject>
			<configObject name="domain_alias">
				<synopsis>Domain Alias</synopsis>
				<description><para>
					Signifies that a domain is an alias. If the domain on a session is
					not found to match an AoR then this object is used to see if we have
					an alias for the AoR to which the endpoint is binding. This objects
					name as defined in configuration should be the domain alias and a
					config option is provided to specify the domain to be aliased.
				</para></description>
				<configOption name="type">
					<synopsis>Must be of type 'domain_alias'.</synopsis>
				</configOption>
				<configOption name="domain">
					<synopsis>Domain to be aliased</synopsis>
				</configOption>
			</configObject>
			<configObject name="transport">
				<synopsis>SIP Transport</synopsis>
				<description><para>
					<emphasis>Transports</emphasis>
					</para>
					<para>There are different transports and protocol derivatives
						supported by <literal>res_pjsip</literal>. They are in order of
						preference: UDP, TCP, and WebSocket (WS).</para>
					<note><para>Changes to transport configuration in pjsip.conf will only be
						effected on a complete restart of Asterisk. A module reload
						will not suffice.</para></note>
				</description>
				<configOption name="async_operations" default="1">
					<synopsis>Number of simultaneous Asynchronous Operations</synopsis>
				</configOption>
				<configOption name="bind">
					<synopsis>IP Address and optional port to bind to for this transport</synopsis>
				</configOption>
				<configOption name="ca_list_file">
					<synopsis>File containing a list of certificates to read (TLS ONLY, not WSS)</synopsis>
				</configOption>
				<configOption name="ca_list_path">
					<synopsis>Path to directory containing a list of certificates to read (TLS ONLY, not WSS)</synopsis>
				</configOption>
				<configOption name="cert_file">
					<synopsis>Certificate file for endpoint (TLS ONLY, not WSS)</synopsis>
					<description><para>
						A path to a .crt or .pem file can be provided.  However, only
						the certificate is read from the file, not the private key.
						The <literal>priv_key_file</literal> option must supply a
						matching key file.
					</para></description>
				</configOption>
				<configOption name="cipher">
					<synopsis>Preferred cryptography cipher names (TLS ONLY, not WSS)</synopsis>
					<description>
					<para>Comma separated list of cipher names or numeric equivalents.
						Numeric equivalents can be either decimal or hexadecimal (0xX).
					</para>
					<para>There are many cipher names.  Use the CLI command
						<literal>pjsip list ciphers</literal> to see a list of cipher
						names available for your installation.  See link for more:</para>
					<para>http://www.openssl.org/docs/apps/ciphers.html#CIPHER_SUITE_NAMES
					</para>
					</description>
				</configOption>
				<configOption name="domain">
					<synopsis>Domain the transport comes from</synopsis>
				</configOption>
				<configOption name="external_media_address">
					<synopsis>External IP address to use in RTP handling</synopsis>
					<description><para>
						When a request or response is sent out, if the destination of the
						message is outside the IP network defined in the option <literal>localnet</literal>,
						and the media address in the SDP is within the localnet network, then the
						media address in the SDP will be rewritten to the value defined for
						<literal>external_media_address</literal>.
					</para></description>
				</configOption>
				<configOption name="ip_in_sdp_connection" default="1">
					<synopsis>Require to use ip address in SDP Connection</synopsis>
				</configOption>
				<configOption name="get_extaddr_from_stun" default="0">
					<synopsis>Require to get external addr from STUN server</synopsis>
				</configOption>
				<configOption name="external_signaling_address">
					<synopsis>External address for SIP signalling</synopsis>
				</configOption>
				<configOption name="external_signaling_port" default="0">
					<synopsis>External port for SIP signalling</synopsis>
				</configOption>
				<configOption name="method">
					<synopsis>Method of SSL transport (TLS ONLY, not WSS)</synopsis>
					<description>
						<enumlist>
							<enum name="default">
								<para>The default as defined by PJSIP. This is currently TLSv1, but may change with future releases.</para>
							</enum>
							<enum name="unspecified">
								<para>This option is equivalent to setting 'default'</para>
							</enum>
							<enum name="tlsv1" />
							<enum name="tlsv1_1" />
							<enum name="tlsv1_2" />
							<enum name="sslv2" />
							<enum name="sslv3" />
							<enum name="sslv23" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="local_net">
					<synopsis>Network to consider local (used for NAT purposes).</synopsis>
					<description><para>This must be in CIDR or dotted decimal format with the IP
					and mask separated with a slash ('/').</para></description>
				</configOption>
				<configOption name="password">
					<synopsis>Password required for transport</synopsis>
				</configOption>
				<configOption name="priv_key_file">
					<synopsis>Private key file (TLS ONLY, not WSS)</synopsis>
				</configOption>
				<configOption name="protocol" default="udp">
					<synopsis>Protocol to use for SIP traffic</synopsis>
					<description>
						<enumlist>
							<enum name="udp" />
							<enum name="tcp" />
							<enum name="tls" />
							<enum name="ws" />
							<enum name="wss" />
						</enumlist>
					</description>
				</configOption>
				<configOption name="require_client_cert" default="false">
					<synopsis>Require client certificate (TLS ONLY, not WSS)</synopsis>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'transport'.</synopsis>
				</configOption>
				<configOption name="verify_client" default="false">
					<synopsis>Require verification of client certificate (TLS ONLY, not WSS)</synopsis>
				</configOption>
				<configOption name="verify_server" default="false">
					<synopsis>Require verification of server certificate (TLS ONLY, not WSS)</synopsis>
				</configOption>
				<configOption name="tos" default="false">
					<synopsis>Enable TOS for the signalling sent over this transport</synopsis>
					<description>
					<para>See <literal>https://wiki.asterisk.org/wiki/display/AST/IP+Quality+of+Service</literal>
					for more information on this parameter.</para>
					<note><para>This option does not apply to the <replaceable>ws</replaceable>
					or the <replaceable>wss</replaceable> protocols.</para></note>
					</description>
				</configOption>
				<configOption name="cos" default="false">
					<synopsis>Enable COS for the signalling sent over this transport</synopsis>
					<description>
					<para>See <literal>https://wiki.asterisk.org/wiki/display/AST/IP+Quality+of+Service</literal>
					for more information on this parameter.</para>
					<note><para>This option does not apply to the <replaceable>ws</replaceable>
					or the <replaceable>wss</replaceable> protocols.</para></note>
					</description>
				</configOption>
				<configOption name="websocket_write_timeout">
					<synopsis>The timeout (in milliseconds) to set on WebSocket connections.</synopsis>
					<description>
						<para>If a websocket connection accepts input slowly, the timeout
						for writes to it can be increased to keep it from being disconnected.
						Value is in milliseconds; default is 100 ms.</para>
					</description>
				</configOption>
				<configOption name="allow_reload" default="no">
					<synopsis>Allow this transport to be reloaded.</synopsis>
					<description>
						<para>Allow this transport to be reloaded when res_pjsip is reloaded.
						This option defaults to "no" because reloading a transport may disrupt
						in-progress calls.</para>
					</description>
				</configOption>
				<configOption name="symmetric_transport" default="no">
					<synopsis>Use the same transport for outgoing requests as incoming ones.</synopsis>
					<description>
						<para>When a request from a dynamic contact
							comes in on a transport with this option set to 'yes',
							the transport name will be saved and used for subsequent
							outgoing requests like OPTIONS, NOTIFY and INVITE.  It's
							saved as a contact uri parameter named 'x-ast-txp' and will
							display with the contact uri in CLI, AMI, and ARI output.
							On the outgoing request, if a transport wasn't explicitly
							set on the endpoint AND the request URI is not a hostname,
							the saved transport will be used and the 'x-ast-txp'
							parameter stripped from the outgoing packet.
						</para>
					</description>
				</configOption>
			</configObject>
			<configObject name="contact">
				<synopsis>A way of creating an aliased name to a SIP URI</synopsis>
				<description><para>
					Contacts are a way to hide SIP URIs from the dialplan directly.
					They are also used to make a group of contactable parties when
					in use with <literal>AoR</literal> lists.
				</para></description>
				<configOption name="type">
					<synopsis>Must be of type 'contact'.</synopsis>
				</configOption>
				<configOption name="uri">
					<synopsis>SIP URI to contact peer</synopsis>
				</configOption>
				<configOption name="expiration_time">
					<synopsis>Time to keep alive a contact</synopsis>
					<description><para>
						Time to keep alive a contact. String style specification.
					</para></description>
				</configOption>
				<configOption name="qualify_frequency" default="0">
					<synopsis>Interval at which to qualify a contact</synopsis>
					<description><para>
						Interval between attempts to qualify the contact for reachability.
						If <literal>0</literal> never qualify. Time in seconds.
					</para></description>
				</configOption>
				<configOption name="qualify_timeout" default="3.0">
					<synopsis>Timeout for qualify</synopsis>
					<description><para>
						If the contact doesn't respond to the OPTIONS request before the timeout,
						the contact is marked unavailable.
						If <literal>0</literal> no timeout. Time in fractional seconds.
					</para></description>
				</configOption>
				<configOption name="authenticate_qualify">
					<synopsis>Authenticates a qualify challenge response if needed</synopsis>
					<description>
						<para>If true and a qualify request receives a challenge response then
						authentication is attempted before declaring the contact available.
						</para>
						<note><para>This option does nothing as we will always complete
						the challenge response authentication if the qualify request is
						challenged.
						</para></note>
					</description>
				</configOption>
				<configOption name="outbound_proxy">
					<synopsis>Outbound proxy used when sending OPTIONS request</synopsis>
					<description><para>
						If set the provided URI will be used as the outbound proxy when an
						OPTIONS request is sent to a contact for qualify purposes.
					</para></description>
				</configOption>
				<configOption name="rmv_obp_from_route" default="no">
					<synopsis>If yes, remove the outbound proxy from Route header of the outgoing message</synopsis>
				</configOption>
				<configOption name="support_missed_call" default="no">
					<synopsis>If yes, the contact support the missed call process.</synopsis>
				</configOption>
				<configOption name="support_gs_message" default="no">
					<synopsis>If yes, the contact support gs_message process.</synopsis>
				</configOption>
				<configOption name="support_ice" default="no">
					<synopsis>If yes, the contact support ice process.</synopsis>
				</configOption>
				<configOption name="support_trickle_ice" default="no">
					<synopsis>If yes, the contact support trickle ice process.</synopsis>
				</configOption>
				<configOption name="gswave_mobile" default="no">
					<synopsis>If yes, is registered by gswave mobile.</synopsis>
				</configOption>
				<configOption name="ssh_proxy_flag" default="no">
					<synopsis>If yes, the contact is through ssh proxy.</synopsis>
				</configOption>
				<configOption name="support_ldap_settings" default="no">
					<synopsis>If yes, the contact support ldap settings.</synopsis>
				</configOption>
				<configOption name="support_gs_config" default="no">
					<synopsis>If yes, the contact support push config settings.</synopsis>
				</configOption>
				<configOption name="support_gs_state" default="no">
					<synopsis>If yes, the contact support push state.</synopsis>
				</configOption>
				<configOption name="support_gs_fec" default="no">
					<synopsis>If yes, the contact support GS-FEC.</synopsis>
				</configOption>
				<configOption name="support_gs_control" default="no">
					<synopsis>If yes, the contact support gs control.</synopsis>
				</configOption>
				<configOption name="require_feature_code_settings" default="no">
					<synopsis>If yes, need to push feature code setting when settings update.</synopsis>
				</configOption>
				<configOption name="require_turn_info_settings" default="no">
					<synopsis>If yes, need to push turn info when update.</synopsis>
				</configOption>
				<configOption name="require_basic_settings" default="no">
					<synopsis>If yes, need to push basic settings when update.</synopsis>
				</configOption>
				<configOption name="require_plugin" default="no">
					<synopsis>If yes, the contact support push UCMRC plugin state.</synopsis>
				</configOption>
				<configOption name="require_phonebook" default="no">
					<synopsis>If yes, the contact support push phonebook configure.</synopsis>
				</configOption>
				<configOption name="path">
					<synopsis>Stored Path vector for use in Route headers on outgoing requests.</synopsis>
				</configOption>
				<configOption name="user_agent">
					<synopsis>User-Agent header from registration.</synopsis>
					<description><para>
						The User-Agent is automatically stored based on data present in incoming SIP
						REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="endpoint">
					<synopsis>Endpoint name</synopsis>
					<description><para>
						The name of the endpoint this contact belongs to
					</para></description>
				</configOption>
				<configOption name="reg_server">
					<synopsis>Asterisk Server name</synopsis>
					<description><para>
						Asterisk Server name on which SIP endpoint registered.
					</para></description>
				</configOption>
				<configOption name="via_addr">
					<synopsis>IP-address of the last Via header from registration.</synopsis>
					<description><para>
						The last Via header should contain the address of UA which sent the request.
						The IP-address of the last Via header is automatically stored based on data present
						in incoming SIP REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="via_port">
					<synopsis>IP-port of the last Via header from registration.</synopsis>
					<description><para>
						The IP-port of the last Via header is automatically stored based on data present
						in incoming SIP REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="call_id">
					<synopsis>Call-ID header from registration.</synopsis>
					<description><para>
						The Call-ID header is automatically stored based on data present
						in incoming SIP REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="prune_on_boot">
					<synopsis>A contact that cannot survive a restart/boot.</synopsis>
					<description><para>
						The option is set if the incoming SIP REGISTER contact is rewritten
						on a reliable transport and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="original_address">
					<synopsis>contact-original-address header record in registration.</synopsis>
					<description><para>
						The contact-original-address is automatically stored based on data present in incoming SIP
						REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="to_host">
					<synopsis>To header host record in registration.</synopsis>
					<description><para>
						The to_host is automatically stored based on data present in incoming SIP
						REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
				<configOption name="sbc_endpoint_addr">
					<synopsis>SBC-Endpoint-Address header record in registration.</synopsis>
					<description><para>
						The SBC-Endpoint-Address is automatically stored based on data present in incoming SIP
						REGISTER requests and is not intended to be configured manually.
					</para></description>
				</configOption>
			</configObject>
			<configObject name="aor">
				<synopsis>The configuration for a location of an endpoint</synopsis>
				<description><para>
					An AoR is what allows Asterisk to contact an endpoint via res_pjsip. If no
					AoRs are specified, an endpoint will not be reachable by Asterisk.
					Beyond that, an AoR has other uses within Asterisk, such as inbound
					registration.
					</para><para>
					An <literal>AoR</literal> is a way to allow dialing a group
					of <literal>Contacts</literal> that all use the same
					<literal>endpoint</literal> for calls.
					</para><para>
					This can be used as another way of grouping a list of contacts to dial
					rather than specifying them each directly when dialing via the dialplan.
					This must be used in conjunction with the <literal>PJSIP_DIAL_CONTACTS</literal>.
					</para><para>
					Registrations: For Asterisk to match an inbound registration to an endpoint,
					the AoR object name must match the user portion of the SIP URI in the "To:"
					header of the inbound SIP registration. That will usually be equivalent
					to the "user name" set in your hard or soft phones configuration.
				</para></description>
				<configOption name="contact">
					<synopsis>Permanent contacts assigned to AoR</synopsis>
					<description><para>
						Contacts specified will be called whenever referenced
						by <literal>chan_pjsip</literal>.
						</para><para>
						Use a separate "contact=" entry for each contact required. Contacts
						are specified using a SIP URI.
					</para></description>
				</configOption>
				<configOption name="default_expiration" default="3600">
					<synopsis>Default expiration time in seconds for contacts that are dynamically bound to an AoR.</synopsis>
				</configOption>
				<configOption name="mailboxes">
					<synopsis>Allow subscriptions for the specified mailbox(es)</synopsis>
					<description><para>This option applies when an external entity subscribes to an AoR
						for Message Waiting Indications. The mailboxes specified will be subscribed to.
						More than one mailbox can be specified with a comma-delimited string.
						app_voicemail mailboxes must be specified as mailbox@context;
						for example: mailboxes=6001@default. For mailboxes provided by external sources,
						such as through the res_external_mwi module, you must specify strings supported by
						the external system.
					</para><para>
						For endpoints that cannot SUBSCRIBE for MWI, you can set the <literal>mailboxes</literal> option in your
						endpoint configuration section to enable unsolicited MWI NOTIFYs to the endpoint.
					</para></description>
				</configOption>
				<configOption name="voicemail_extension">
					<synopsis>The voicemail extension to send in the NOTIFY Message-Account header</synopsis>
				</configOption>
				<configOption name="maximum_expiration" default="7200">
					<synopsis>Maximum time to keep an AoR</synopsis>
					<description><para>
						Maximum time to keep a peer with explicit expiration. Time in seconds.
					</para></description>
				</configOption>
				<configOption name="max_contacts" default="0">
					<synopsis>Maximum number of contacts that can bind to an AoR</synopsis>
					<description><para>
						Maximum number of contacts that can associate with this AoR. This value does
						not affect the number of contacts that can be added with the "contact" option.
						It only limits contacts added through external interaction, such as
						registration.
						</para>
						<note><para>The <replaceable>rewrite_contact</replaceable> option
						registers the source address as the contact address to help with
						NAT and reusing connection oriented transports such as TCP and
						TLS.  Unfortunately, refreshing a registration may register a
						different contact address and exceed
						<replaceable>max_contacts</replaceable>.  The
						<replaceable>remove_existing</replaceable> option can help by
						removing the soonest to expire contact(s) over
						<replaceable>max_contacts</replaceable> which is likely the
						old <replaceable>rewrite_contact</replaceable> contact source
						address being refreshed.
						</para></note>
						<note><para>This should be set to <literal>1</literal> and
						<replaceable>remove_existing</replaceable> set to <literal>yes</literal> if you
						wish to stick with the older <literal>chan_sip</literal> behaviour.
						</para></note>
					</description>
				</configOption>
				<configOption name="minimum_expiration" default="60">
					<synopsis>Minimum keep alive time for an AoR</synopsis>
					<description><para>
						Minimum time to keep a peer with an explicit expiration. Time in seconds.
					</para></description>
				</configOption>
				<configOption name="remove_existing" default="no">
					<synopsis>Determines whether new contacts replace existing ones.</synopsis>
					<description><para>
						On receiving a new registration to the AoR should it remove enough
						existing contacts not added or updated by the registration to
						satisfy <replaceable>max_contacts</replaceable>?  Any removed
						contacts will expire the soonest.
						</para>
						<note><para>The <replaceable>rewrite_contact</replaceable> option
						registers the source address as the contact address to help with
						NAT and reusing connection oriented transports such as TCP and
						TLS.  Unfortunately, refreshing a registration may register a
						different contact address and exceed
						<replaceable>max_contacts</replaceable>.  The
						<replaceable>remove_existing</replaceable> option can help by
						removing the soonest to expire contact(s) over
						<replaceable>max_contacts</replaceable> which is likely the
						old <replaceable>rewrite_contact</replaceable> contact source
						address being refreshed.
						</para></note>
						<note><para>This should be set to <literal>yes</literal> and
						<replaceable>max_contacts</replaceable> set to <literal>1</literal> if you
						wish to stick with the older <literal>chan_sip</literal> behaviour.
						</para></note>
					</description>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'aor'.</synopsis>
				</configOption>
				<configOption name="qualify_frequency" default="0">
					<synopsis>Interval at which to qualify an AoR</synopsis>
					<description><para>
						Interval between attempts to qualify the AoR for reachability.
						If <literal>0</literal> never qualify. Time in seconds.
					</para></description>
				</configOption>
				<configOption name="qualify_timeout" default="3.0">
					<synopsis>Timeout for qualify</synopsis>
					<description><para>
						If the contact doesn't respond to the OPTIONS request before the timeout,
						the contact is marked unavailable.
						If <literal>0</literal> no timeout. Time in fractional seconds.
					</para></description>
				</configOption>
				<configOption name="authenticate_qualify">
					<synopsis>Authenticates a qualify challenge response if needed</synopsis>
					<description>
						<para>If true and a qualify request receives a challenge response then
						authentication is attempted before declaring the contact available.
						</para>
						<note><para>This option does nothing as we will always complete
						the challenge response authentication if the qualify request is
						challenged.
						</para></note>
					</description>
				</configOption>
				<configOption name="outbound_proxy">
					<synopsis>Outbound proxy used when sending OPTIONS request</synopsis>
					<description><para>
						If set the provided URI will be used as the outbound proxy when an
						OPTIONS request is sent to a contact for qualify purposes.
					</para></description>
				</configOption>
				<configOption name="rmv_obp_from_route" default="no">
					<synopsis>If yes, remove the outbound proxy from Route header of the outgoing message</synopsis>
				</configOption>
				<configOption name="support_path">
					<synopsis>Enables Path support for REGISTER requests and Route support for other requests.</synopsis>
					<description><para>
						When this option is enabled, the Path headers in register requests will be saved
						and its contents will be used in Route headers for outbound out-of-dialog requests
						and in Path headers for outbound 200 responses. Path support will also be indicated
						in the Supported header.
					</para></description>
				</configOption>
    			<configOption name="account_trunk" default="no">
					<synopsis>If yes, it's an aor of account trunk</synopsis>
    			</configOption>
			</configObject>
			<configObject name="system">
				<synopsis>Options that apply to the SIP stack as well as other system-wide settings</synopsis>
				<description><para>
					The settings in this section are global. In addition to being global, the values will
					not be re-evaluated when a reload is performed. This is because the values must be set
					before the SIP stack is initialized. The only way to reset these values is to either
					restart Asterisk, or unload res_pjsip.so and then load it again.
				</para></description>
				<configOption name="timer_t1" default="500">
					<synopsis>Set transaction timer T1 value (milliseconds).</synopsis>
					<description><para>
						Timer T1 is the base for determining how long to wait before retransmitting
						requests that receive no response when using an unreliable transport (e.g. UDP).
						For more information on this timer, see RFC 3261, Section 17.1.1.1.
					</para></description>
				</configOption>
				<configOption name="timer_b" default="32000">
					<synopsis>Set transaction timer B value (milliseconds).</synopsis>
					<description><para>
						Timer B determines the maximum amount of time to wait after sending an INVITE
						request before terminating the transaction. It is recommended that this be set
						to 64 * Timer T1, but it may be set higher if desired. For more information on
						this timer, see RFC 3261, Section 17.1.1.1.
					</para></description>
				</configOption>
				<configOption name="compact_headers" default="no">
					<synopsis>Use the short forms of common SIP header names.</synopsis>
				</configOption>
				<configOption name="threadpool_initial_size" default="0">
					<synopsis>Initial number of threads in the res_pjsip threadpool.</synopsis>
				</configOption>
				<configOption name="threadpool_auto_increment" default="5">
					<synopsis>The amount by which the number of threads is incremented when necessary.</synopsis>
				</configOption>
				<configOption name="threadpool_idle_timeout" default="60">
					<synopsis>Number of seconds before an idle thread should be disposed of.</synopsis>
				</configOption>
				<configOption name="threadpool_max_size" default="0">
					<synopsis>Maximum number of threads in the res_pjsip threadpool.
					A value of 0 indicates no maximum.</synopsis>
				</configOption>
				<configOption name="disable_tcp_switch" default="yes">
					<synopsis>Disable automatic switching from UDP to TCP transports.</synopsis>
					<description><para>
						Disable automatic switching from UDP to TCP transports if outgoing
						request is too large.  See RFC 3261 section 18.1.1.
					</para></description>
				</configOption>
				<configOption name="follow_early_media_fork">
					<synopsis>Follow SDP forked media when To tag is different</synopsis>
					<description><para>
						On outgoing calls, if the UAS responds with different SDP attributes
						on subsequent 18X or 2XX responses (such as a port update) AND the
						To tag on the subsequent response is different than that on the previous
						one, follow it.
						</para>
						<note><para>
							This option must also be enabled on endpoints that require
							this functionality.
						</para></note>
					</description>
				</configOption>
				<configOption name="accept_multiple_sdp_answers">
					<synopsis>Follow SDP forked media when To tag is the same</synopsis>
					<description><para>
						On outgoing calls, if the UAS responds with different SDP attributes
						on non-100rel 18X or 2XX responses (such as a port update) AND the
						To tag on the subsequent response is the same as that on the previous one,
						process the updated SDP.
						</para>
						<note><para>
							This option must also be enabled on endpoints that require
							this functionality.
						</para></note>
					</description>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'system' UNLESS the object name is 'system'.</synopsis>
				</configOption>
			</configObject>
			<configObject name="global">
				<synopsis>Options that apply globally to all SIP communications</synopsis>
				<description><para>
					The settings in this section are global. Unlike options in the <literal>system</literal>
					section, these options can be refreshed by performing a reload.
				</para></description>
				<configOption name="max_forwards" default="70">
					<synopsis>Value used in Max-Forwards header for SIP requests.</synopsis>
				</configOption>
				<configOption name="keep_alive_interval" default="90">
					<synopsis>The interval (in seconds) to send keepalives to active connection-oriented transports.</synopsis>
				</configOption>
				<configOption name="contact_expiration_check_interval" default="30">
					<synopsis>The interval (in seconds) to check for expired contacts.</synopsis>
				</configOption>
				<configOption name="disable_multi_domain" default="no">
					<synopsis>Disable Multi Domain support</synopsis>
					<description><para>
						If disabled it can improve realtime performance by reducing the number of database requests.
					</para></description>
				</configOption>
				<configOption name="max_initial_qualify_time" default="0">
					<synopsis>The maximum amount of time from startup that qualifies should be attempted on all contacts.
					If greater than the qualify_frequency for an aor, qualify_frequency will be used instead.</synopsis>
				</configOption>
				<configOption name="unidentified_request_period" default="5">
					<synopsis>The number of seconds over which to accumulate unidentified requests.</synopsis>
					<description><para>
					If <literal>unidentified_request_count</literal> unidentified requests are received
					during <literal>unidentified_request_period</literal>, a security event will be generated.
					</para></description>
				</configOption>
				<configOption name="unidentified_request_count" default="5">
					<synopsis>The number of unidentified requests from a single IP to allow.</synopsis>
					<description><para>
					If <literal>unidentified_request_count</literal> unidentified requests are received
					during <literal>unidentified_request_period</literal>, a security event will be generated.
					</para></description>
				</configOption>
				<configOption name="unidentified_request_prune_interval" default="30">
					<synopsis>The interval at which unidentified requests are older than
					twice the unidentified_request_period are pruned.</synopsis>
				</configOption>
				<configOption name="type">
					<synopsis>Must be of type 'global' UNLESS the object name is 'global'.</synopsis>
				</configOption>
				<configOption name="user_agent" default="Asterisk &lt;Asterisk Version&gt;">
					<synopsis>Value used in User-Agent header for SIP requests and Server header for SIP responses.</synopsis>
				</configOption>
				<configOption name="regcontext" default="">
					<synopsis>When set, Asterisk will dynamically create and destroy a NoOp priority 1 extension for a given
						peer who registers or unregisters with us.</synopsis>
				</configOption>
				<configOption name="default_outbound_endpoint" default="default_outbound_endpoint">
					<synopsis>Endpoint to use when sending an outbound request to a URI without a specified endpoint.</synopsis>
				</configOption>
				<configOption name="default_voicemail_extension">
					<synopsis>The voicemail extension to send in the NOTIFY Message-Account header if not specified on endpoint or aor</synopsis>
				</configOption>
				<configOption name="debug" default="no">
					<synopsis>Enable/Disable SIP debug logging.  Valid options include yes, no, or
						a host address</synopsis>
				</configOption>
				<configOption name="endpoint_identifier_order">
					<synopsis>The order by which endpoint identifiers are processed and checked.
						Identifier names are usually derived from and can be found in the endpoint
						identifier module itself (res_pjsip_endpoint_identifier_*).
						You can use the CLI command "pjsip show identifiers" to see the
						identifiers currently available.</synopsis>
					<description>
						<note><para>
						One of the identifiers is "auth_username" which matches on the username in
						an Authentication header.  This method has some security considerations because an
						Authentication header is not present on the first message of a dialog when
						digest authentication is used.  The client can't generate it until the server
						sends the challenge in a 401 response.  Since Asterisk normally sends a security
						event when an incoming request can't be matched to an endpoint, using auth_username
						requires that the security event be deferred until a request is received with
						the Authentication header and only generated if the username doesn't result in a
						match.  This may result in a delay before an attack is recognized.  You can control
						how many unmatched requests are received from a single ip address before a security
						event is generated using the unidentified_request parameters.
						</para></note>
					</description>
				</configOption>
				<configOption name="default_from_user" default="asterisk">
					<synopsis>When Asterisk generates an outgoing SIP request, the From header username will be
						set to this value if there is no better option (such as CallerID) to be
						used.</synopsis>
				</configOption>
				<configOption name="alwaysauthreject" default="no">
					<synopsis>Send 401 Unauthorized for all failing requests.</synopsis>
				</configOption>
				<configOption name="realm" default="grandstraem">
					<synopsis>the default value realm for authentication.</synopsis>
				</configOption>
				<configOption name="maxcallbitrate" default="384">
					<synopsis>the default value bit for stream.</synopsis>
				</configOption>
				<configOption name="attr_passthrough" default="yes">
					<synopsis>the default value bit attr_passthrough.</synopsis>
				</configOption>
				<configOption name="allow_blind_transfer_callback" default="no">
					<synopsis>the default value allow_blind_transfer_callback.</synopsis>
				</configOption>
				<configOption name="blind_transfer_timeout" default="60">
					<synopsis>the default value blind_transfer_timeout.</synopsis>
				</configOption>
				<configOption name="allow_forward_hold_request" default="no">
					<synopsis>the default value bit allow_forward_hold_request.</synopsis>
				</configOption>
				<configOption name="local_negotiate_prefer" default="no">
					<synopsis>the default value bit local_negotiate_prefer.</synopsis>
				</configOption>
				<configOption name="turn_session_count" default="0">
					<synopsis>the default total turn session count.</synopsis>
				</configOption>
				<configOption name="enable_p2p_video_control" default="no">
					<synopsis>the default value bit enable_p2p_video_control.</synopsis>
				</configOption>
				<configOption name="p2p_video_resolution" default="720P">
					<synopsis>the default video resolution for p2p call.</synopsis>
				</configOption>
				<configOption name="p2p_video_bitrate" default="512000">
					<synopsis>the default video max bitrate for p2p call.</synopsis>
				</configOption>
				<configOption name="only_support_audio" default="no">
					<synopsis>the default value bit only_support_audio.</synopsis>
				</configOption>
				<configOption name="default_realm" default="asterisk">
					<synopsis>When Asterisk generates a challenge, the digest realm will be
						set to this value if there is no better option (such as auth/realm) to be
						used.</synopsis>
				</configOption>
				<configOption name="mwi_tps_queue_high" default="500">
					<synopsis>MWI taskprocessor high water alert trigger level.</synopsis>
					<description>
						<para>On a heavily loaded system you may need to adjust the
						taskprocessor queue limits.  If any taskprocessor queue size
						reaches its high water level then pjsip will stop processing
						new requests until the alert is cleared.  The alert clears
						when all alerting taskprocessor queues have dropped to their
						low water clear level.
						</para>
					</description>
				</configOption>
				<configOption name="mwi_tps_queue_low" default="-1">
					<synopsis>MWI taskprocessor low water clear alert level.</synopsis>
					<description>
						<para>On a heavily loaded system you may need to adjust the
						taskprocessor queue limits.  If any taskprocessor queue size
						reaches its high water level then pjsip will stop processing
						new requests until the alert is cleared.  The alert clears
						when all alerting taskprocessor queues have dropped to their
						low water clear level.
						</para>
						<note><para>Set to -1 for the low water level to be 90% of
						the high water level.</para></note>
					</description>
				</configOption>
				<configOption name="mwi_disable_initial_unsolicited" default="no">
					<synopsis>Enable/Disable sending unsolicited MWI to all endpoints on startup.</synopsis>
					<description>
						<para>When the initial unsolicited MWI notification are
						enabled on startup then the initial notifications
						get sent at startup.  If you have a lot of endpoints
						(thousands) that use unsolicited MWI then you may
						want to consider disabling the initial startup
						notifications.
						</para>
						<para>When the initial unsolicited MWI notifications are
						disabled on startup then the notifications will start
						on the endpoint's next contact update.
						</para>
					</description>
				</configOption>
				<configOption name="ignore_uri_user_options">
					<synopsis>Enable/Disable ignoring SIP URI user field options.</synopsis>
					<description>
						<para>If you have this option enabled and there are semicolons
						in the user field of a SIP URI then the field is truncated
						at the first semicolon.  This effectively makes the semicolon
						a non-usable character for PJSIP endpoint names, extensions,
						and AORs.  This can be useful for improving compatibility with
						an ITSP that likes to use user options for whatever reason.
						</para>
						<example title="Sample SIP URI">
							sip:1235557890;phone-context=national@x.x.x.x;user=phone
						</example>
						<example title="Sample SIP URI user field">
							1235557890;phone-context=national
						</example>
						<example title="Sample SIP URI user field truncated">
							1235557890
						</example>
						<note><para>The caller-id and redirecting number strings
						obtained from incoming SIP URI user fields are always truncated
						at the first semicolon.</para></note>
					</description>
				</configOption>
				<configOption name="use_callerid_contact" default="no">
					<synopsis>Place caller-id information into Contact header</synopsis>
					<description><para>
						This option will cause Asterisk to place caller-id information into
						generated Contact headers.</para>
					</description>
				</configOption>
				<configOption name="send_contact_status_on_update_registration" default="no">
					<synopsis>Enable sending AMI ContactStatus event when a device refreshes its registration.</synopsis>
				</configOption>
				<configOption name="max_sub_expires" default="600">
					<synopsis>The maxinum value allowed in the subscription.</synopsis>
				</configOption>
 				<configOption name="reg_delay_interval" default="0">
					<synopsis>The delay interval of outbound registration.</synopsis>
 				</configOption>
 				<configOption name="send_deflection_div" default="yes">
					<synopsis>Enable/Disable send deflection Diversion header options.</synopsis>
 				</configOption>
 				<configOption name="unavailable_cause" default="20">
					<synopsis>The response cause when the called extension is unavailable.</synopsis>
 				</configOption>
			</configObject>
		</configFile>
	</configInfo>
	<manager name="PJSIPQualify" language="en_US">
		<synopsis>
			Qualify a chan_pjsip endpoint.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Endpoint" required="true">
				<para>The endpoint you want to qualify.</para>
			</parameter>
		</syntax>
		<description>
			<para>Qualify a chan_pjsip endpoint.</para>
		</description>
	</manager>
	<managerEvent language="en_US" name="IdentifyDetail">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an identify section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'identify'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="Endpoint">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip_endpoint_identifier_ip']/configFile[@name='pjsip.conf']/configObject[@name='identify']/configOption[@name='endpoint']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SrvLookups">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip_endpoint_identifier_ip']/configFile[@name='pjsip.conf']/configObject[@name='identify']/configOption[@name='srv_lookups']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Match">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip_endpoint_identifier_ip']/configFile[@name='pjsip.conf']/configObject[@name='identify']/configOption[@name='match']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MatchHeader">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip_endpoint_identifier_ip']/configFile[@name='pjsip.conf']/configObject[@name='identify']/configOption[@name='match_header']/synopsis/node())"/></para>
				</parameter>
				<parameter name="EndpointName">
					<para>The name of the endpoint associated with this information.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="AorDetail">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an Address of Record (AoR) section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'aor'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="MinimumExpiration">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='minimum_expiration']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MaximumExpiration">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='maximum_expiration']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DefaultExpiration">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='default_expiration']/synopsis/node())"/></para>
				</parameter>
				<parameter name="QualifyFrequency">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='qualify_frequency']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AuthenticateQualify">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='authenticate_qualify']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MaxContacts">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='max_contacts']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RemoveExisting">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='remove_existing']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Mailboxes">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='mailboxes']/synopsis/node())"/></para>
				</parameter>
				<parameter name="OutboundProxy">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='outbound_proxy']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SupportPath">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='support_path']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TotalContacts">
					<para>The total number of contacts associated with this AoR.</para>
				</parameter>
				<parameter name="ContactsRegistered">
					<para>The number of non-permanent contacts associated with this AoR.</para>
				</parameter>
				<parameter name="EndpointName">
					<para>The name of the endpoint associated with this information.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="AuthDetail">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an authentication section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'auth'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="Username">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='username']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Password">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='username']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Md5Cred">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='md5_cred']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Realm">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='realm']/synopsis/node())"/></para>
				</parameter>
				<parameter name="NonceLifetime">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='nonce_lifetime']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AuthType">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='auth_type']/synopsis/node())"/></para>
				</parameter>
				<parameter name="EndpointName">
					<para>The name of the endpoint associated with this information.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="TransportDetail">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an authentication section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'transport'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="Protocol">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='protocol']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Bind">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='bind']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AsycOperations">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='async_operations']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CaListFile">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='ca_list_file']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CaListPath">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='ca_list_path']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CertFile">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='cert_file']/synopsis/node())"/></para>
				</parameter>
				<parameter name="PrivKeyFile">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='priv_key_file']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Password">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='password']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ExternalSignalingAddress">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='external_signaling_address']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ExternalSignalingPort">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='external_signaling_port']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ExternalMediaAddress">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='external_media_address']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Domain">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='domain']/synopsis/node())"/></para>
				</parameter>
				<parameter name="VerifyServer">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='verify_server']/synopsis/node())"/></para>
				</parameter>
				<parameter name="VerifyClient">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='verify_client']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RequireClientCert">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='require_client_cert']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Method">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='method']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Cipher">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='cipher']/synopsis/node())"/></para>
				</parameter>
				<parameter name="LocalNet">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='local_net']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Tos">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='tos']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Cos">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='cos']/synopsis/node())"/></para>
				</parameter>
				<parameter name="WebsocketWriteTimeout">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='transport']/configOption[@name='websocket_write_timeout']/synopsis/node())"/></para>
				</parameter>
				<parameter name="EndpointName">
					<para>The name of the endpoint associated with this information.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="EndpointDetail">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an endpoint section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'endpoint'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="Context">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='context']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Disallow">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='disallow']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Allow">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='allow']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtmfMode">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtmf_mode']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RtpIpv6">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='rtp_ipv6']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RtpSymmetric">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='rtp_symmetric']/synopsis/node())"/></para>
				</parameter>
				<parameter name="IceSupport">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='ice_support']/synopsis/node())"/></para>
				</parameter>
				<parameter name="UsePtime">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='use_ptime']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ForceRport">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='force_rport']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RewriteContact">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='rewrite_contact']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Transport">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='transport']/synopsis/node())"/></para>
				</parameter>
				<parameter name="OutboundProxy">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='outbound_proxy']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MohSuggest">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='moh_suggest']/synopsis/node())"/></para>
				</parameter>
				<parameter name="100rel">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='100rel']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Timers">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='timers']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TimersMinSe">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='timers_min_se']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TimersSessExpires">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='timers_sess_expires']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Auth">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='auth']/synopsis/node())"/></para>
				</parameter>
				<parameter name="OutboundAuth">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='outbound_auth']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Aors">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='aors']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MediaAddress">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='media_address']/synopsis/node())"/></para>
				</parameter>
				<parameter name="IdentifyBy">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='identify_by']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DirectMedia">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='direct_media']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DirectMediaMethod">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='direct_media_method']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TrustConnectedLine">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='trust_connected_line']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SendConnectedLine">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='send_connected_line']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ConnectedLineMethod">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='connected_line_method']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DirectMediaGlareMitigation">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='direct_media_glare_mitigation']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DisableDirectMediaOnNat">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='disable_direct_media_on_nat']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Callerid">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='callerid']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CalleridPrivacy">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='callerid_privacy']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CalleridTag">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='callerid_tag']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TrustIdInbound">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='trust_id_inbound']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TrustIdOutbound">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='trust_id_outbound']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SendPai">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='send_pai']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SendRpid">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='send_rpid']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SendDiversion">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='send_diversion']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Mailboxes">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='mailboxes']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AggregateMwi">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='aggregate_mwi']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MediaEncryption">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='media_encryption']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MediaEncryptionOptimistic">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='media_encryption_optimistic']/synopsis/node())"/></para>
				</parameter>
				<parameter name="UseAvpf">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='use_avpf']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ForceAvp">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='force_avp']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MediaUseReceivedTransport">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='media_use_received_transport']/synopsis/node())"/></para>
				</parameter>
				<parameter name="OneTouchRecording">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='one_touch_recording']/synopsis/node())"/></para>
				</parameter>
				<parameter name="InbandProgress">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='inband_progress']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CallGroup">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='call_group']/synopsis/node())"/></para>
				</parameter>
				<parameter name="PickupGroup">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='pickup_group']/synopsis/node())"/></para>
				</parameter>
				<parameter name="NamedCallGroup">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='named_call_group']/synopsis/node())"/></para>
				</parameter>
				<parameter name="NamedPickupGroup">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='named_pickup_group']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DeviceStateBusyAt">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='device_state_busy_at']/synopsis/node())"/></para>
				</parameter>
				<parameter name="T38Udptl">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='t38_udptl']/synopsis/node())"/></para>
				</parameter>
				<parameter name="T38UdptlEc">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='t38_udptl_ec']/synopsis/node())"/></para>
				</parameter>
				<parameter name="T38UdptlMaxdatagram">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='t38_udptl_maxdatagram']/synopsis/node())"/></para>
				</parameter>
				<parameter name="FaxDetect">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='fax_detect']/synopsis/node())"/></para>
				</parameter>
				<parameter name="T38UdptlNat">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='t38_udptl_nat']/synopsis/node())"/></para>
				</parameter>
				<parameter name="T38UdptlIpv6">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='t38_udptl_ipv6']/synopsis/node())"/></para>
				</parameter>
				<parameter name="ToneZone">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='tone_zone']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Language">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='language']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RecordOnFeature">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='record_on_feature']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RecordOffFeature">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='record_off_feature']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AllowTransfer">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='allow_transfer']/synopsis/node())"/></para>
				</parameter>
				<parameter name="UserEqPhone">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='user_eq_phone']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MohPassthrough">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='moh_passthrough']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SdpOwner">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='sdp_owner']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SdpSession">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='sdp_session']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TosAudio">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='tos_audio']/synopsis/node())"/></para>
				</parameter>
				<parameter name="TosVideo">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='tos_video']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CosAudio">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='cos_audio']/synopsis/node())"/></para>
				</parameter>
				<parameter name="CosVideo">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='cos_video']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AllowSubscribe">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='allow_subscribe']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SubMinExpiry">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='sub_min_expiry']/synopsis/node())"/></para>
				</parameter>
				<parameter name="FromUser">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='from_user']/synopsis/node())"/></para>
				</parameter>
				<parameter name="FromDomain">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='from_domain']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MwiFromUser">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='mwi_from_user']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RtpEngine">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='rtp_engine']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsVerify">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_verify']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsRekey">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_rekey']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsCertFile">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_cert_file']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsPrivateKey">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_private_key']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsCipher">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_cipher']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsCaFile">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_ca_file']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsCaPath">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_ca_path']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DtlsSetup">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='dtls_setup']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SrtpTag32">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='srtp_tag_32']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RedirectMethod">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='redirect_method']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SetVar">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='set_var']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MessageContext">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='message_context']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Accountcode">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='accountcode']/synopsis/node())"/></para>
				</parameter>
				<parameter name="PreferredCodecOnly">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='preferred_codec_only']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DeviceState">
					<para>The aggregate device state for this endpoint.</para>
				</parameter>
				<parameter name="ActiveChannels">
					<para>The number of active channels associated with this endpoint.</para>
				</parameter>
				<parameter name="SubscribeContext">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='subscribe_context']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Allowoverlap">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='endpoint']/configOption[@name='allow_overlap']/synopsis/node())"/></para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="AorList">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an Address of Record (AoR) section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'aor'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="MinimumExpiration">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='minimum_expiration']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MaximumExpiration">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='maximum_expiration']/synopsis/node())"/></para>
				</parameter>
				<parameter name="DefaultExpiration">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='default_expiration']/synopsis/node())"/></para>
				</parameter>
				<parameter name="QualifyFrequency">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='qualify_frequency']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AuthenticateQualify">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='authenticate_qualify']/synopsis/node())"/></para>
				</parameter>
				<parameter name="MaxContacts">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='max_contacts']/synopsis/node())"/></para>
				</parameter>
				<parameter name="RemoveExisting">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='remove_existing']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Mailboxes">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='mailboxes']/synopsis/node())"/></para>
				</parameter>
				<parameter name="OutboundProxy">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='outbound_proxy']/synopsis/node())"/></para>
				</parameter>
				<parameter name="SupportPath">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='aor']/configOption[@name='support_path']/synopsis/node())"/></para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="AuthList">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about an Address of Record (Auth) section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'auth'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="Username">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='username']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Md5Cred">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='md5_cred']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Realm">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='realm']/synopsis/node())"/></para>
				</parameter>
				<parameter name="AuthType">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='auth_type']/synopsis/node())"/></para>
				</parameter>
				<parameter name="Password">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='password']/synopsis/node())"/></para>
				</parameter>
				<parameter name="NonceLifetime">
					<para><xi:include xpointer="xpointer(/docs/configInfo[@name='res_pjsip']/configFile[@name='pjsip.conf']/configObject[@name='auth']/configOption[@name='nonce_lifetime']/synopsis/node())"/></para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="ContactList">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about a contact section.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'contact'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="ViaAddr">
					<para>IP address of the last Via header in REGISTER request.
					Will only appear in the event if available.</para>
				</parameter>
				<parameter name="ViaPort">
					<para>Port number of the last Via header in REGISTER request.
					Will only appear in the event if available.</para>
				</parameter>
				<parameter name="QualifyTimeout">
					<para>The elapsed time in decimal seconds after which an OPTIONS
					message is sent before the contact is considered unavailable.</para>
				</parameter>
				<parameter name="CallId">
					<para>Content of the Call-ID header in REGISTER request.
					Will only appear in the event if available.</para>
				</parameter>
				<parameter name="RegServer">
					<para>Asterisk Server name.</para>
				</parameter>
				<parameter name="PruneOnBoot">
					<para>If true delete the contact on Asterisk restart/boot.</para>
				</parameter>
				<parameter name="Path">
					<para>The Path header received on the REGISTER.</para>
				</parameter>
				<parameter name="Endpoint">
					<para>The name of the endpoint associated with this information.</para>
				</parameter>
				<parameter name="AuthenticateQualify">
					<para>A boolean indicating whether a qualify should be authenticated.</para>
				</parameter>
				<parameter name="Uri">
					<para>This contact's URI.</para>
				</parameter>
				<parameter name="QualifyFrequency">
					<para>The interval in seconds at which the contact will be qualified.</para>
				</parameter>
				<parameter name="UserAgent">
					<para>Content of the User-Agent header in REGISTER request</para>
				</parameter>
				<parameter name="ExpirationTime">
					<para>Absolute time that this contact is no longer valid after</para>
				</parameter>
				<parameter name="OutboundProxy">
					<para>The contact's outbound proxy.</para>
				</parameter>
				<parameter name="Status">
					<para>This contact's status.</para>
					<enumlist>
						<enum name="Reachable"/>
						<enum name="Unreachable"/>
						<enum name="NonQualified"/>
						<enum name="Unknown"/>
					</enumlist>
				</parameter>
				<parameter name="RoundtripUsec">
					<para>The round trip time in microseconds.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="ContactStatusDetail">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about a contact's status.</synopsis>
			<syntax>
				<parameter name="AOR">
					<para>The AoR that owns this contact.</para>
				</parameter>
				<parameter name="URI">
					<para>This contact's URI.</para>
				</parameter>
				<parameter name="Status">
					<para>This contact's status.</para>
					<enumlist>
						<enum name="Reachable"/>
						<enum name="Unreachable"/>
						<enum name="NonQualified"/>
						<enum name="Unknown"/>
					</enumlist>
				</parameter>
				<parameter name="RoundtripUsec">
					<para>The round trip time in microseconds.</para>
				</parameter>
				<parameter name="EndpointName">
					<para>The name of the endpoint associated with this information.</para>
				</parameter>
				<parameter name="UserAgent">
					<para>Content of the User-Agent header in REGISTER request</para>
				</parameter>
				<parameter name="RegExpire">
					<para>Absolute time that this contact is no longer valid after</para>
				</parameter>
				<parameter name="ViaAddress">
					<para>IP address:port of the last Via header in REGISTER request.
					Will only appear in the event if available.</para>
				</parameter>
				<parameter name="CallID">
					<para>Content of the Call-ID header in REGISTER request.
					Will only appear in the event if available.</para>
				</parameter>
				<parameter name="ID">
					<para>The sorcery ID of the contact.</para>
				</parameter>
				<parameter name="AuthenticateQualify">
					<para>A boolean indicating whether a qualify should be authenticated.</para>
				</parameter>
				<parameter name="OutboundProxy">
					<para>The contact's outbound proxy.</para>
				</parameter>
				<parameter name="Path">
					<para>The Path header received on the REGISTER.</para>
				</parameter>
				<parameter name="QualifyFrequency">
					<para>The interval in seconds at which the contact will be qualified.</para>
				</parameter>
				<parameter name="QualifyTimeout">
					<para>The elapsed time in decimal seconds after which an OPTIONS
					message is sent before the contact is considered unavailable.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<managerEvent language="en_US" name="EndpointList">
		<managerEventInstance class="EVENT_FLAG_COMMAND">
			<synopsis>Provide details about a contact's status.</synopsis>
			<syntax>
				<parameter name="ObjectType">
					<para>The object's type. This will always be 'endpoint'.</para>
				</parameter>
				<parameter name="ObjectName">
					<para>The name of this object.</para>
				</parameter>
				<parameter name="Transport">
					<para>The transport configurations associated with this endpoint.</para>
				</parameter>
				<parameter name="Aor">
					<para>The aor configurations associated with this endpoint.</para>
				</parameter>
				<parameter name="Auths">
					<para>The inbound authentication configurations associated with this endpoint.</para>
				</parameter>
				<parameter name="OutboundAuths">
					<para>The outbound authentication configurations associated with this endpoint.</para>
				</parameter>
				<parameter name="DeviceState">
					<para>The aggregate device state for this endpoint.</para>
				</parameter>
				<parameter name="ActiveChannels">
					<para>The number of active channels associated with this endpoint.</para>
				</parameter>
			</syntax>
		</managerEventInstance>
	</managerEvent>
	<manager name="PJSIPShowEndpoints" language="en_US">
		<synopsis>
			Lists PJSIP endpoints.
		</synopsis>
		<syntax />
		<description>
			<para>
			Provides a listing of all endpoints.  For each endpoint an <literal>EndpointList</literal> event
			is raised that contains relevant attributes and status information.  Once all
			endpoints have been listed an <literal>EndpointListComplete</literal> event is issued.
			</para>
		</description>
		<responses>
			<list-elements>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='EndpointList'])" />
			</list-elements>
			<managerEvent language="en_US" name="EndpointListComplete">
				<managerEventInstance class="EVENT_FLAG_COMMAND">
					<synopsis>Provide final information about an endpoint list.</synopsis>
					<syntax>
						<parameter name="EventList"/>
						<parameter name="ListItems"/>
					</syntax>
				</managerEventInstance>
			</managerEvent>
		</responses>
	</manager>
	<manager name="PJSIPShowEndpoint" language="en_US">
		<synopsis>
			Detail listing of an endpoint and its objects.
		</synopsis>
		<syntax>
			<xi:include xpointer="xpointer(/docs/manager[@name='Login']/syntax/parameter[@name='ActionID'])" />
			<parameter name="Endpoint" required="true">
				<para>The endpoint to list.</para>
			</parameter>
		</syntax>
		<description>
			<para>
			Provides a detailed listing of options for a given endpoint.  Events are issued
			showing the configuration and status of the endpoint and associated objects.  These
			events include <literal>EndpointDetail</literal>, <literal>AorDetail</literal>,
			<literal>AuthDetail</literal>, <literal>TransportDetail</literal>, and
			<literal>IdentifyDetail</literal>.  Some events may be listed multiple times if multiple objects are
			associated (for instance AoRs).  Once all detail events have been raised a final
			<literal>EndpointDetailComplete</literal> event is issued.
			</para>
		</description>
		<responses>
			<list-elements>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='EndpointDetail'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='IdentifyDetail'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='ContactStatusDetail'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='AuthDetail'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='TransportDetail'])" />
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='AorDetail'])" />
			</list-elements>
			<managerEvent language="en_US" name="EndpointDetailComplete">
				<managerEventInstance class="EVENT_FLAG_COMMAND">
					<synopsis>Provide final information about endpoint details.</synopsis>
					<syntax>
						<parameter name="EventList"/>
						<parameter name="ListItems"/>
					</syntax>
				</managerEventInstance>
			</managerEvent>
		</responses>
	</manager>
	<manager name="PJSIPShowAors" language="en_US">
		<synopsis>
			Lists PJSIP AORs.
		</synopsis>
		<syntax />
		<description>
			<para>
			Provides a listing of all AORs. For each AOR an <literal>AorList</literal> event
			is raised that contains relevant attributes and status information.  Once all
			aors have been listed an <literal>AorListComplete</literal> event is issued.
			</para>
		</description>
		<responses>
			<list-elements>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='AorList'])" />
			</list-elements>
			<managerEvent language="en_US" name="AorListComplete">
				<managerEventInstance class="EVENT_FLAG_COMMAND">
					<synopsis>Provide final information about an aor list.</synopsis>
					<syntax>
						<parameter name="EventList"/>
						<parameter name="ListItems"/>
					</syntax>
				</managerEventInstance>
			</managerEvent>
		</responses>
	</manager>
	<manager name="PJSIPShowAuths" language="en_US">
		<synopsis>
			Lists PJSIP Auths.
		</synopsis>
		<syntax />
		<description>
			<para>Provides a listing of all Auths. For each Auth an <literal>AuthList</literal> event
			is raised that contains relevant attributes and status information.  Once all
			auths have been listed an <literal>AuthListComplete</literal> event is issued.
			</para>
		</description>
		<responses>
			<list-elements>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='AuthList'])" />
			</list-elements>
			<managerEvent language="en_US" name="AuthListComplete">
				<managerEventInstance class="EVENT_FLAG_COMMAND">
					<synopsis>Provide final information about an auth list.</synopsis>
					<syntax>
						<parameter name="EventList"/>
						<parameter name="ListItems"/>
					</syntax>
				</managerEventInstance>
			</managerEvent>
		</responses>
	</manager>
	<manager name="PJSIPShowContacts" language="en_US">
		<synopsis>
			Lists PJSIP Contacts.
		</synopsis>
		<syntax />
		<description>
			<para>Provides a listing of all Contacts. For each Contact a <literal>ContactList</literal>
			event is raised that contains relevant attributes and status information.
			Once all contacts have been listed a <literal>ContactListComplete</literal> event
			is issued.
			</para>
		</description>
		<responses>
			<list-elements>
				<xi:include xpointer="xpointer(/docs/managerEvent[@name='ContactList'])" />
			</list-elements>
			<managerEvent language="en_US" name="ContactListComplete">
				<managerEventInstance class="EVENT_FLAG_COMMAND">
					<synopsis>Provide final information about a contact list.</synopsis>
					<syntax>
						<parameter name="EventList"/>
						<parameter name="ListItems"/>
					</syntax>
				</managerEventInstance>
			</managerEvent>
		</responses>
	</manager>
	<application name="SipCheckIsConf" language="en_US">
			<synopsis>
				Check whether the number is a conference room number.
			</synopsis>
			<syntax>
				<parameter name="conf" required="true" />
			</syntax>
			<description>
				<para>Check whether the number is a conference room number,The result is stored in the channel variable IS_CONF_ROOM.</para>
				<para>Example: SipCheckIsConf(6300)</para>
			</description>
	</application>
	<application name="CheckIsConference" language="en_US">
			<synopsis>
				Check whether the number is a conference room number.
			</synopsis>
			<syntax>
				<parameter name="conf" required="true" />
			</syntax>
			<description>
				<para>Check whether the number is a conference room number,The result is as the returned value.</para>
				<para>Example: CheckIsConference(6300)</para>
			</description>
	</application>
 ***/

#define MOD_DATA_CONTACT "contact"

/*! Number of serializers in pool if one not supplied. */
#define SERIALIZER_POOL_SIZE		8

/*! Pool of serializers to use if not supplied. */
static struct ast_taskprocessor *serializer_pool[SERIALIZER_POOL_SIZE];

static pjsip_endpoint *ast_pjsip_endpoint;

static struct ast_threadpool *sip_threadpool;

/*! Local host address for IPv4 */
static pj_sockaddr host_ip_ipv4;

/*! Local host address for IPv4 (string form) */
static char host_ip_ipv4_string[PJ_INET6_ADDRSTRLEN];
#ifdef GRANDSTREAM_NETWORKS
/* Todo HA just support IPV4, just record IPV4 address */
static char local_ip_ipv4_string[PJ_INET6_ADDRSTRLEN];
static char local_vlan_ip_string[PJ_INET6_ADDRSTRLEN];

#define AES_ENCRYPT_BUFFER_LEN  2048
#define AES_ALL_ENCRYPT_BUFFER_LEN  2048*4
#define MAX_PHONEBOOK_URL_LEN 128
#define MAX_PHONEBOOK_XML_LEN 256

static char ldap_xml_buffer[AES_ENCRYPT_BUFFER_LEN];  /* Added by yfyang, for ldap setting sync */
static char ldap_xml_public_buffer[AES_ENCRYPT_BUFFER_LEN];
static char feature_code_xml_buffer[AES_ENCRYPT_BUFFER_LEN*2];
static char basic_xml_buffer[AES_ENCRYPT_BUFFER_LEN];
static char phonebook_xml_buffer[AES_ENCRYPT_BUFFER_LEN];
static char phonebook_xml_public_buffer[AES_ENCRYPT_BUFFER_LEN];
static char phonebook_public_url[MAX_PHONEBOOK_URL_LEN];
static ast_rwlock_t  config_xml_lock;
static struct ao2_container *callflip_users;//added by zyxia for callflip,2021.11.26
static int enable_callflip=0;//added by zyxia for callflip,2021.11.26
static const char app_sipcheckconf[] = "SipCheckIsConf";
static const char app_checkisconference[] = "CheckIsConference";
static const char app_checkifcanNway[] = "CheckIfCanNway";

#endif

/*! Local host address for IPv6 */
static pj_sockaddr host_ip_ipv6;

/*! Local host address for IPv6 (string form) */
static char host_ip_ipv6_string[PJ_INET6_ADDRSTRLEN];

#ifdef GRANDSTREAM_NETWORKS
#define MAX_OUTBOUND_PREFIX_LEN 64
#define DIRECT_MEDIA_OUTBOUND_PREFIX_CONF "direct_media_outbound_prefix.conf"
#define MAX_TURN_SERVER_LEN 128
#define MAX_TURN_USER_NAME_LEN 256
#define MAX_TURN_PASSWORD_LEN 256
#define MAX_TURN_PROTOCOL_TYPE_LEN 32
#define MAX_TURN_LIMIT_TIME_LEN 16
#define MAX_TURN_MAX_REGISTRAR_LEN 16
#define MAX_TURN_MAX_MEDIA_NAT_CALL_LEN 16
#define MAX_SUPPORT_WEBRTC_TRUNK_LEN 4
#define TURN_INFO_DB_FAMILY "TurnInfo"
#define TURN_INFO_ADDRESS_KEY "Address"
#define TURN_INFO_STUN_ADDRESS_KEY "StunAddress"
#define TURN_INFO_USERNAME_KEY "UserName"
#define TURN_INFO_PASSWORD_KEY "Password"
#define TURN_INFO_PROTOCOLTYPE_KEY "ProtocolType"
#define TURN_INFO_LIMITTIME_KEY "LimitTime"
#define TURN_INFO_DAY_TOTAL_LIMITTIME_KEY "DayTotalLimitTime"
#define TURN_INFO_MONTH_TOTAL_LIMITTIME_KEY "MonthTotalLimitTime"
#define TURN_INFO_MAXREGISTRAR_KEY "MaxRegistrar"
#define TURN_INFO_MAX_MEDIA_NAT_CALL_KEY "MaxMediaNatCall"
#define TURN_INFO_ENCRYPT_KEY "QSX%s#@j8dj+t*yH"
#define PLUGIN_MARKET_KEY "PluginMarket"
#define PLUGIN_THIRDPARTY_KEY "PluginThirdparty"
#define SUPPORT_WEBRTC_TRUNK "SupportWebRTCTrunk"


/* Add by bfdu for turn license control, 2020.4.30 */
static unsigned int total_turn_session_count;
static unsigned int turn_limit_time;
static unsigned int turn_day_total_limit_time;
static unsigned int turn_month_total_limit_time;
static unsigned int turn_max_registrar;
static unsigned int turn_max_media_nat_call;
static char *turn_server_address;
static char *stun_server_address;
static char *turn_user_name;
static char *turn_password;
static char *turn_protocol_type;
static ast_rwlock_t total_turn_session_lock;

static unsigned int used_turn_session_count;
static ast_rwlock_t used_turn_session_lock;

static unsigned int max_used_turn_session_count;
static ast_rwlock_t max_used_turn_session_lock;
static char max_used_date[32];
static char alert_exceed_threshold_date[32];
static char alert_exceed_total_date[32];

static unsigned int plugin_market = 0;
static unsigned int plugin_thirdparty = 0;
static unsigned int support_webrtc_trunk = 0;

int float_ip_to_avs = 0;

struct outbound_prefix_list {
	char outbound_prefix[MAX_OUTBOUND_PREFIX_LEN];
	AST_RWLIST_ENTRY(outbound_prefix_list) entry;
};

static AST_LIST_HEAD(, outbound_prefix_list) direct_media_outbound_prefix_list;

/* Add by bfdu for SFU, 2019.12.16 */
#define MCM_CONF "mcm_conference.conf"
#define MAX_MCM_CONF_LEN  (64)
struct mcm_room {
	char *number;
};
static struct ao2_container *mcm_rooms = NULL;

static AST_LIST_HEAD_STATIC(internal_mcb_peers, cei_internal_mcb_peer);
struct ast_gs_idle_feature_code  gs_idle_feature_code[GS_IDLE_FEATURE_COUNT] ={
	{ "dndActive", "*77"},
	{ "dndDeactive", "*78"},
	{ "fwdBusyActive", "*71"},
	{ "fwdBusyDeactive", "*72"},
	{ "fwdNoAnswerActive", "*73"},
	{ "fwdNoAnswerDeactive", "*74"},
	{ "fwdAllActive", "*75"},
	{ "fwdAllDeactive", "*76"},
	{ "agentPause", "*83"},
	{ "agentUnpause", "*84"},
	{ "pagingPrefix", "*81"},
	{ "intercomPrefix", "*80"},
	{ "callPickup", "**"},
	{ "blacklistAdd", "*40"},
	{ "blacklistRemove", "*41"},
	{ "voiceMailCode", "*98"},
	{ "myMailCode", "*97"}
};

static int prase_mcm_conference_config(int reload);

/* restore the Call-Info header to the channel */
void ast_sip_set_call_info_string(struct ast_channel *channel, struct pjsip_rx_data *rdata)
{
	static const pj_str_t callinfo_str = { "Call-Info", 9 };
	pjsip_generic_string_hdr   *hdr;
	pj_str_t callinfo_content;
	char callinfo_rec[1024] = {0};
	char *info_list[4] = {0};
	char *info;
	int i = 0;
	int len = 0;
	int j;

	if (!channel || !rdata) {
		return;
	}

	/* get PAI header string */
	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &callinfo_str, NULL);
	if (!hdr) {
		return;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &callinfo_content, &hdr->hvalue);
	do {
		info = ast_strsep(&callinfo_content.ptr, ',', AST_STRSEP_TRIM);
		if (!info) {
			break;
		}

		if (!strstr(info, "answer-after=") && !strstr(info, "call-completion")  && !strstr(info, "appearance-index=")) {
			info_list[i] = info;
			i++;
		}
	} while (i < 4);

	if (i == 0) {
		return;
	}

	snprintf(callinfo_rec, sizeof(callinfo_rec), "%s", info_list[0]);
	for (j=1; j<i; j++) {
		len = strlen(callinfo_rec);
		snprintf(callinfo_rec+len, sizeof(callinfo_rec)-len-1, ",%s", info_list[j]);
	}

	pbx_builtin_setvar_helper(channel, "__CALL_INFO", callinfo_rec);

	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) {
		pbx_builtin_setvar_helper(channel, "__CALL_INFO_MSG", "request");
	} else {
		pbx_builtin_setvar_helper(channel, "__CALL_INFO_MSG", "response");
	}
}

int ast_sip_judge_if_local_ip(const char *original_address)
{
	const char *local_ip = NULL;

	if (!original_address) {
		return 0;
	}

	/* Todo, HA just support ipv4 now, use this to compare */
	local_ip = ast_sip_get_local_ip_string(pj_AF_INET());
	if (!local_ip) {
		return 0;
	}

	if (!strcmp(original_address, local_ip)) {
		ast_log(LOG_DEBUG, "ast_sip_judge_if_local_ip original_address:%s, local_ip:%s\n", original_address, local_ip);
		return 1;
	}

	if (strlen(local_vlan_ip_string) && !strcmp(original_address, local_vlan_ip_string)) {
		ast_log(LOG_DEBUG, "ast_sip_judge_if_local_ip, local_vlan_ip:%s\n", local_vlan_ip_string);
		return 1;
	}

	return 0;
}

/* BEGIN: Added by yfyang, for sip ssh proxy feature */
int ast_sip_check_ssh_proxy_flag(pjsip_rx_data *rdata)
{
	char *trans_type = NULL;
	pjsip_uri      *ruri = NULL;
	pjsip_sip_uri  *sip_ruri = NULL;
	char  host_value[256] = {0};
	char  default_host_value[256] = {0};

	if (!rdata) {
		return 0;
	}

#ifdef BUILD_ON_CLOUD_PBX
	return 1;
#endif

	trans_type = rdata->tp_info.transport->type_name;
	ast_log(LOG_DEBUG, "check_ssh_proxy_flag, type[%s], src addr[%s]", trans_type, rdata->pkt_info.src_name);

	if (!strstr(rdata->pkt_info.src_name, "127.0.0.1") && !ast_sip_judge_if_local_ip(rdata->pkt_info.src_name)) {
		return 0;
	}

	if (!strcmp(trans_type, "TCP") || !strcmp(trans_type, "TLS")) {
		return 1;
	}

	if (strcmp(trans_type, "WS") && strcmp(trans_type, "WSS")) {
		return 0;
	}

	/* Get request-uri host */
	ruri = rdata->msg_info.msg->line.req.uri;
	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
		return 0;
	}

	/* cmp request-host and WsProxyHost */
	ast_sip_get_nvram_value(":WsProxyHost", host_value, sizeof(host_value));
	ast_sip_get_nvram_value(":DefaultWsProxyHost", default_host_value, sizeof(default_host_value));
	if (!strlen(host_value) && !strlen(default_host_value)) {
		return 0;
	}

	sip_ruri = pjsip_uri_get_uri(ruri);
	if (strlen(host_value) && !strncmp(host_value, sip_ruri->host.ptr, sip_ruri->host.slen)) {
		return 1;
	}

	if (strlen(default_host_value) && !strncmp(default_host_value, sip_ruri->host.ptr, sip_ruri->host.slen)) {
		return 1;
	}

	return 0;
}

int ast_sip_set_contact_param_for_sshproxy(pjsip_tx_data *tdata)
{
    pjsip_contact_hdr *contact = NULL;
    static const pj_str_t ssh_str = { "ssh-proxy", 9 };
    static const pj_str_t blank_str = { "", 0 };
    pjsip_param *ssh_param = NULL;

    if (!tdata) {
        ast_log(LOG_WARNING, "tdata is NULL.\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CONTACT, NULL);
    if (!contact) {
        ast_log(LOG_NOTICE, "contact header is NULL.\n");
        return -1;
    }

    ssh_param = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
    if (!ssh_param) {
        ast_log(LOG_WARNING, "alloc ssh_param failed.\n");
        return -1;
    }

    ssh_param->name  = ssh_str;
    ssh_param->value = blank_str;
    pj_list_push_back(&contact->other_param, ssh_param);

    return 0;
}

/* END: Added by yfyang, for sip ssh proxy feature */
#endif

void ast_sip_add_date_header(pjsip_tx_data *tdata)
{
	char date[256];
	struct tm tm;
	time_t t = time(NULL);

	gmtime_r(&t, &tm);
	strftime(date, sizeof(date), "%a, %d %b %Y %T GMT", &tm);

	ast_sip_add_header(tdata, "Date", date);
}

static int register_service(void *data)
{
	pjsip_module **module = data;
	if (!ast_pjsip_endpoint) {
		ast_log(LOG_ERROR, "There is no PJSIP endpoint. Unable to register services\n");
		return -1;
	}
	if (pjsip_endpt_register_module(ast_pjsip_endpoint, *module) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Unable to register module %.*s\n", (int) pj_strlen(&(*module)->name), pj_strbuf(&(*module)->name));
		return -1;
	}
	ast_debug(1, "Registered SIP service %.*s (%p)\n", (int) pj_strlen(&(*module)->name), pj_strbuf(&(*module)->name), *module);
	return 0;
}

int ast_sip_register_service(pjsip_module *module)
{
	return ast_sip_push_task_wait_servant(NULL, register_service, &module);
}

static int unregister_service(void *data)
{
	pjsip_module **module = data;
	if (!ast_pjsip_endpoint) {
		return -1;
	}
	pjsip_endpt_unregister_module(ast_pjsip_endpoint, *module);
	ast_debug(1, "Unregistered SIP service %.*s\n", (int) pj_strlen(&(*module)->name), pj_strbuf(&(*module)->name));
	return 0;
}

void ast_sip_unregister_service(pjsip_module *module)
{
	ast_sip_push_task_wait_servant(NULL, unregister_service, &module);
}

static struct ast_sip_authenticator *registered_authenticator;

int ast_sip_register_authenticator(struct ast_sip_authenticator *auth)
{
	if (registered_authenticator) {
		ast_log(LOG_WARNING, "Authenticator %p is already registered. Cannot register a new one\n", registered_authenticator);
		return -1;
	}
	registered_authenticator = auth;
	ast_debug(1, "Registered SIP authenticator module %p\n", auth);

	return 0;
}

void ast_sip_unregister_authenticator(struct ast_sip_authenticator *auth)
{
	if (registered_authenticator != auth) {
		ast_log(LOG_WARNING, "Trying to unregister authenticator %p but authenticator %p registered\n",
				auth, registered_authenticator);
		return;
	}
	registered_authenticator = NULL;
	ast_debug(1, "Unregistered SIP authenticator %p\n", auth);
}

int ast_sip_requires_authentication(struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata)
{
	if (!registered_authenticator) {
		ast_log(LOG_WARNING, "No SIP authenticator registered. Assuming authentication is not required\n");
		return 0;
	}

	return registered_authenticator->requires_authentication(endpoint, rdata);
}

enum ast_sip_check_auth_result ast_sip_check_authentication(struct ast_sip_endpoint *endpoint,
		pjsip_rx_data *rdata, pjsip_tx_data *tdata)
{
	if (!registered_authenticator) {
		ast_log(LOG_WARNING, "No SIP authenticator registered. Assuming authentication is successful\n");
		return AST_SIP_AUTHENTICATION_SUCCESS;
	}
	return registered_authenticator->check_authentication(endpoint, rdata, tdata);
}

static struct ast_sip_outbound_authenticator *registered_outbound_authenticator;

int ast_sip_register_outbound_authenticator(struct ast_sip_outbound_authenticator *auth)
{
	if (registered_outbound_authenticator) {
		ast_log(LOG_WARNING, "Outbound authenticator %p is already registered. Cannot register a new one\n", registered_outbound_authenticator);
		return -1;
	}
	registered_outbound_authenticator = auth;
	ast_debug(1, "Registered SIP outbound authenticator module %p\n", auth);

	return 0;
}

void ast_sip_unregister_outbound_authenticator(struct ast_sip_outbound_authenticator *auth)
{
	if (registered_outbound_authenticator != auth) {
		ast_log(LOG_WARNING, "Trying to unregister outbound authenticator %p but outbound authenticator %p registered\n",
				auth, registered_outbound_authenticator);
		return;
	}
	registered_outbound_authenticator = NULL;
	ast_debug(1, "Unregistered SIP outbound authenticator %p\n", auth);
}

int ast_sip_create_request_with_auth(const struct ast_sip_auth_vector *auths, pjsip_rx_data *challenge,
		pjsip_tx_data *old_request, pjsip_tx_data **new_request)
{
	if (!registered_outbound_authenticator) {
		ast_log(LOG_WARNING, "No SIP outbound authenticator registered. Cannot respond to authentication challenge\n");
		return -1;
	}
	return registered_outbound_authenticator->create_request_with_auth(auths, challenge, old_request, new_request);
}

struct endpoint_identifier_list {
	const char *name;
	unsigned int priority;
	struct ast_sip_endpoint_identifier *identifier;
	AST_RWLIST_ENTRY(endpoint_identifier_list) list;
};

static AST_RWLIST_HEAD_STATIC(endpoint_identifiers, endpoint_identifier_list);

int ast_sip_register_endpoint_identifier_with_name(struct ast_sip_endpoint_identifier *identifier,
						 const char *name)
{
	char *prev, *current, *identifier_order;
	struct endpoint_identifier_list *iter, *id_list_item;
	SCOPED_LOCK(lock, &endpoint_identifiers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	id_list_item = ast_calloc(1, sizeof(*id_list_item));
	if (!id_list_item) {
		ast_log(LOG_ERROR, "Unable to add endpoint identifier. Out of memory.\n");
		return -1;
	}
	id_list_item->identifier = identifier;
	id_list_item->name = name;

	ast_debug(1, "Register endpoint identifier %s(%p)\n", name ?: "", identifier);

	if (ast_strlen_zero(name)) {
		/* if an identifier has no name then place in front */
		AST_RWLIST_INSERT_HEAD(&endpoint_identifiers, id_list_item, list);
		return 0;
	}

	/* see if the name of the identifier is in the global endpoint_identifier_order list */
	identifier_order = prev = current = ast_sip_get_endpoint_identifier_order();

	if (ast_strlen_zero(identifier_order)) {
		id_list_item->priority = UINT_MAX;
		AST_RWLIST_INSERT_TAIL(&endpoint_identifiers, id_list_item, list);
		ast_free(identifier_order);
		return 0;
	}

	id_list_item->priority = 0;
	while ((current = strchr(current, ','))) {
		++id_list_item->priority;
		if (!strncmp(prev, name, current - prev)
			&& strlen(name) == current - prev) {
			break;
		}
		prev = ++current;
	}

	if (!current) {
		/* check to see if it is the only or last item */
		if (!strcmp(prev, name)) {
			++id_list_item->priority;
		} else {
			id_list_item->priority = UINT_MAX;
		}
	}

	if (id_list_item->priority == UINT_MAX || AST_RWLIST_EMPTY(&endpoint_identifiers)) {
		/* if not in the endpoint_identifier_order list then consider it less in
		   priority and add it to the end */
		AST_RWLIST_INSERT_TAIL(&endpoint_identifiers, id_list_item, list);
		ast_free(identifier_order);
		return 0;
	}

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&endpoint_identifiers, iter, list) {
		if (id_list_item->priority < iter->priority) {
			AST_RWLIST_INSERT_BEFORE_CURRENT(id_list_item, list);
			break;
		}

		if (!AST_RWLIST_NEXT(iter, list)) {
			AST_RWLIST_INSERT_AFTER(&endpoint_identifiers, iter, id_list_item, list);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

	ast_free(identifier_order);
	return 0;
}

int ast_sip_register_endpoint_identifier(struct ast_sip_endpoint_identifier *identifier)
{
	return ast_sip_register_endpoint_identifier_with_name(identifier, NULL);
}

void ast_sip_unregister_endpoint_identifier(struct ast_sip_endpoint_identifier *identifier)
{
	struct endpoint_identifier_list *iter;
	SCOPED_LOCK(lock, &endpoint_identifiers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&endpoint_identifiers, iter, list) {
		if (iter->identifier == identifier) {
			AST_RWLIST_REMOVE_CURRENT(list);
			ast_free(iter);
			ast_debug(1, "Unregistered endpoint identifier %p\n", identifier);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

struct ast_sip_endpoint *ast_sip_identify_endpoint(pjsip_rx_data *rdata)
{
	struct endpoint_identifier_list *iter;
	struct ast_sip_endpoint *endpoint = NULL;
	SCOPED_LOCK(lock, &endpoint_identifiers, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE(&endpoint_identifiers, iter, list) {
		ast_assert(iter->identifier->identify_endpoint != NULL);
		endpoint = iter->identifier->identify_endpoint(rdata);
		if (endpoint) {
			break;
		}
	}
	return endpoint;
}

#ifdef GRANDSTREAM_NETWORKS
/*Start added by zyxia for wakeup ios gswave,2018.7.27*/
/*! \brief remove sip: sips @ and port from uri
*/
char* ast_sip_get_ip_addr_from_contact_uri(const char *contact_uri)
{
	const char *origin_uri = NULL;
	char *uri = NULL;
	char *pUri = NULL;

	if( NULL == contact_uri){
		return NULL;
	}

	// Valid URI starts with sip: or sips:
	if (strncmp(contact_uri, "sip:", 4) == 0) {
		origin_uri = contact_uri + 4;
	}
	else if (strncmp(contact_uri, "sips:", 5) == 0) {
		origin_uri = contact_uri + 5;
	}
	else {
		return NULL;
	}

	// IP address is next to username@
	pUri = strchr(origin_uri, '@');
	if (pUri == NULL) {
		uri = ast_strdup(origin_uri);
	}
	else {
		uri = ast_strdup(pUri + 1);
	}

	// Remove port, modified for ipv6, bug Bug 67059
	/*
	if ((uri[0] == '[') && (pUri = strchr(uri, ']'))) {
		*pUri = '\0';
		pUri = ast_strdup(uri + 1);
		ast_free(uri);
		uri = pUri;
	} else {
		pUri = strchr(uri, ':');
		if (pUri != NULL) {
		*pUri = '\0';
		}
	}
	*/

	// Remove transport
	pUri = strchr(uri, ';');
	if (pUri != NULL) {
		*pUri = '\0';
	}

	return uri;
}
/*End added by zyxia for wakeup ios gswave,2018.7.27*/

char *ast_sip_identify_get_address_next(pjsip_rx_data *rdata, char *endpoint_id)
{
	struct endpoint_identifier_list *iter;

	SCOPED_LOCK(lock, &endpoint_identifiers, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE(&endpoint_identifiers, iter, list) {
		if (iter->identifier->identify_address_next) {
			return iter->identifier->identify_address_next(rdata, endpoint_id);
		}
	}
	return NULL;
}

/* Added by yfyang, for resolve in identifier */
struct ast_sip_resolve_identifier gs_res_identifier = {
	.get_dns_resolve_addr = NULL,
	.set_used_by_addr = NULL,
	.get_addrs_by_host = NULL,
};

int ast_sip_register_resolve_identifier(struct ast_sip_resolve_identifier *res_identifier)
{
	if (res_identifier) {
		gs_res_identifier.get_dns_resolve_addr = res_identifier->get_dns_resolve_addr;
		gs_res_identifier.set_used_by_addr = res_identifier->set_used_by_addr;
		gs_res_identifier.get_addrs_by_host = res_identifier->get_addrs_by_host;
	}

	return 0; 
}

void ast_sip_unregister_resolve_identifier(struct ast_sip_resolve_identifier *res_identifier)
{
	gs_res_identifier.get_dns_resolve_addr = NULL;
	gs_res_identifier.set_used_by_addr = NULL;
	gs_res_identifier.get_addrs_by_host = NULL;
}

int ast_sip_resolve_identifier_check(void) 
{
	if (NULL == gs_res_identifier.get_dns_resolve_addr) {
		return 0;
	}

	return 1;
}

int ast_sip_resolve_identifier_get(pj_str_t *host, pjsip_transport_type_e type, pjsip_server_addresses *addrs) 
{
	struct ast_ha  *ha   = NULL;
	struct ast_ha  *ha1  = NULL;
	struct ast_ha  *ha2  = NULL;
	char   ip_str[256]   = {0};
	int    ip_count = 0;
	pjsip_transport_type_e type4 = type;
	pjsip_transport_type_e type6 = type;

	if (!host || !addrs) {
		return 0;
	}

	if (!gs_res_identifier.get_dns_resolve_addr) {
		return 0;
	}

	if (type == PJSIP_TRANSPORT_UNSPECIFIED) {
		type4 = PJSIP_TRANSPORT_UDP;
		type6 = PJSIP_TRANSPORT_UDP6; 
	}

	if (type6 < PJSIP_TRANSPORT_IPV6) {
		type6 += PJSIP_TRANSPORT_IPV6;
	}

	ha = gs_res_identifier.get_dns_resolve_addr(host);
	ha1 = ha;

	/* adjust the ha order */
	ha2 = ast_dup_ha_list_by_used(ha);
	if (!ha2) {
		ast_log(LOG_NOTICE, "ast_sip_resolve_identifier_get, ha2 is null");
		return 0;
	}

	ha = ha2;
	for (; ha; ha = ha->next) {
		snprintf(ip_str, sizeof(ip_str), "%s", ast_sockaddr_stringify_addr(&ha->addr));

		if (ast_sockaddr_is_ipv4(&ha->addr)) {
			addrs->entry[ip_count].addr_len = sizeof(pj_sockaddr_in);
			pj_sockaddr_init(pj_AF_INET(), &addrs->entry[ip_count].addr, NULL, ha->port);
			inet_pton(AF_INET, ip_str, (struct in_addr*)&addrs->entry[ip_count].addr.ipv4.sin_addr);
			addrs->entry[ip_count].type = type4;
			pj_sockaddr_set_port(&addrs->entry[ip_count].addr, ha->port ? ha->port: pjsip_transport_get_default_port_for_type(type4));

			ast_log(LOG_NOTICE, "ast_sip_resolve_identifier_get, ip_str[%s], port[%d], ip_count[%d], type[%d]", ip_str, ha->port, ip_count, type4);
			ip_count++;
		} else if (ast_sockaddr_is_ipv6(&ha->addr)) {
			addrs->entry[ip_count].addr_len = sizeof(pj_sockaddr_in6);
			pj_sockaddr_init(pj_AF_INET6(), &addrs->entry[ip_count].addr, NULL, ha->port);
			inet_pton(pj_AF_INET6(), ip_str, (struct in_addr6*)&addrs->entry[ip_count].addr.ipv6.sin6_addr);
			addrs->entry[ip_count].type = type6;
			pj_sockaddr_set_port(&addrs->entry[ip_count].addr, ha->port ? ha->port: pjsip_transport_get_default_port_for_type(type6));

			ast_log(LOG_NOTICE, "ast_sip_resolve_identifier_get, ip_str[%s], port[%d], ip_count[%d], type[%d]", ip_str, ha->port, ip_count, type6);
			ip_count++;
		} else {
			continue;
		}

		if (ip_count >= PJSIP_MAX_RESOLVED_ADDRESSES) {
			break;
		}
	}

	ast_free_ha(ha1);
	ast_free_ha(ha2);

	addrs->count = ip_count;
	ast_log(LOG_NOTICE, "ast_sip_resolve_identifier_get, host[%.*s], ip_count[%d]", (int)host->slen, host->ptr, ip_count);

	return ip_count;
}

void ast_sip_resolve_identifier_set_inuse(char *endpoint_id, char *addr) 
{
	if (!endpoint_id || !addr) {
		return;
	}

	if (!gs_res_identifier.set_used_by_addr) {
		return;
	}

	gs_res_identifier.set_used_by_addr(endpoint_id, addr);
	return;
}

int ast_sip_resolve_identifier_get_destaddrs(char *dest, char *buf, int len) 
{
	char host[256] = {0};
	char *start = NULL;
	char *end   = NULL;

	if (!dest || !buf || !len) {
		return 0;
	}

	if (!gs_res_identifier.get_addrs_by_host) {
		return 0;
	}

	if (!strncmp(dest, "sip:", 4)) {
		start = dest + strlen("sip:");
	} else {
		start = dest;
	}

	end = strchr(start, ';');
	if (end) {
        strncpy(host, start, end-start);
	} else {
        strcpy(host, start);
    }

	gs_res_identifier.get_addrs_by_host(host, buf, len);

	return 0;
}

/* Added by yfyang, for secret process */
struct sql_cmd_struct {
	char *sql;
	int argc;
	char **argv;
};


static SQLHSTMT generic_prepare(struct odbc_obj *obj, void *data)
{
	struct sql_cmd_struct *gps = data;
	int res, i;
	SQLHSTMT stmt;

	res = SQLAllocHandle(SQL_HANDLE_STMT, obj->con, &stmt);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Alloc Handle failed!\n");
		return NULL;
	}
	res = SQLPrepare(stmt, (unsigned char *) gps->sql, SQL_NTS);
	if (!SQL_SUCCEEDED(res)) {
		ast_log(AST_LOG_WARNING, "SQL Prepare failed![%s]\n", gps->sql);
		SQLFreeHandle(SQL_HANDLE_STMT, stmt);
		return NULL;
	}
	for (i = 0; i < gps->argc; i++)
		SQLBindParameter(stmt, i + 1, SQL_PARAM_INPUT, SQL_C_CHAR, SQL_CHAR, strlen(gps->argv[i]), 0, gps->argv[i], 0, NULL);

	return stmt;
}

int ast_sip_get_plain_passwd_from_db(const char *account, char **passwd)
{
	int res = -1;
	char *sql_cmd;
	char passwd_encrypted[256];
	struct odbc_obj *odbc_hdl;
	SQLHSTMT stmt;

	struct sql_cmd_struct gps = {NULL, 0, NULL};

	if (!account || !passwd) {
		ast_log(LOG_WARNING, "Invalid param, account:%p, passwd:%p!", account, passwd);
		return res;
	}
	if (strcasestr(account, "trunk_")) {
		ast_asprintf(&sql_cmd, "SELECT secret FROM sip_trunks WHERE (trunk_type='register' OR trunk_type='account') AND trunk_index=REPLACE('%s','trunk_','')", account);
	} else {
		ast_asprintf(&sql_cmd, "SELECT secret FROM sip_accounts WHERE extension='%s'", account);
	}
	if (!sql_cmd) {
		ast_log(LOG_WARNING, "Failed to create sql cmd!");
		return res;
	}

	odbc_hdl = ast_core_odbc_request_obj(UCM_CONFIG_DB, 0);
	if (!odbc_hdl) {
		ast_log(LOG_WARNING, "Failed to create odbc obj!");
		return res;
	}

	gps.sql = sql_cmd;
	stmt = ast_core_odbc_prepare_and_execute(odbc_hdl, generic_prepare, &gps);
	if (!stmt) {
		ast_log(LOG_WARNING, "Failed to execute sql cmd!");
		goto END;
	}

	if (!SQL_SUCCEEDED(SQLFetch(stmt))) {
		//ast_log(LOG_WARNING, "Failed to fetch raw!");
		goto FREE_STMT_HDL;
	}

	if (!SQL_SUCCEEDED(SQLGetData(stmt, 1, SQL_CHAR, passwd_encrypted, sizeof(passwd_encrypted), NULL))) {
		ast_log(LOG_WARNING, "Failed to fetch data!");
		goto FREE_STMT_HDL;
	}

	if (ucm_secret_decrypt(passwd_encrypted, passwd)) {
		res = 0;
	}

FREE_STMT_HDL:
	SQLFreeHandle(SQL_HANDLE_STMT, stmt);

END:
	ast_core_odbc_release_obj(odbc_hdl);
	ast_free(sql_cmd);
	return res;
}

#endif

static int do_cli_dump_endpt(void *v_a)
{
	struct ast_cli_args *a = v_a;

	ast_pjproject_log_intercept_begin(a->fd);
	pjsip_endpt_dump(ast_sip_get_pjsip_endpoint(), a->argc == 4 ? PJ_TRUE : PJ_FALSE);
	ast_pjproject_log_intercept_end();

	return 0;
}

static char *cli_dump_endpt(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
#ifdef AST_DEVMODE
		e->command = "pjsip dump endpt [details]";
		e->usage =
			"Usage: pjsip dump endpt [details]\n"
			"       Dump the res_pjsip endpt internals.\n"
			"\n"
			"Warning: PJPROJECT documents that the function used by this\n"
			"CLI command may cause a crash when asking for details because\n"
			"it tries to access all active memory pools.\n";
#else
		/*
		 * In non-developer mode we will not document or make easily accessible
		 * the details option even though it is still available.  The user has
		 * to know it exists to use it.  Presumably they would also be aware of
		 * the potential crash warning.
		 */
		e->command = "pjsip dump endpt";
		e->usage =
			"Usage: pjsip dump endpt\n"
			"       Dump the res_pjsip endpt internals.\n";
#endif /* AST_DEVMODE */
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (4 < a->argc
		|| (a->argc == 4 && strcasecmp(a->argv[3], "details"))) {
		return CLI_SHOWUSAGE;
	}

	ast_sip_push_task_wait_servant(NULL, do_cli_dump_endpt, a);

	return CLI_SUCCESS;
}

static char *cli_show_endpoint_identifiers(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
#define ENDPOINT_IDENTIFIER_FORMAT "%-20.20s\n"
	struct endpoint_identifier_list *iter;

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show identifiers";
		e->usage = "Usage: pjsip show identifiers\n"
		            "      List all registered endpoint identifiers\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc != 3) {
                return CLI_SHOWUSAGE;
        }

	ast_cli(a->fd, ENDPOINT_IDENTIFIER_FORMAT, "Identifier Names:");
	{
		SCOPED_LOCK(lock, &endpoint_identifiers, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);
		AST_RWLIST_TRAVERSE(&endpoint_identifiers, iter, list) {
			ast_cli(a->fd, ENDPOINT_IDENTIFIER_FORMAT,
				iter->name ? iter->name : "name not specified");
		}
	}
	return CLI_SUCCESS;
#undef ENDPOINT_IDENTIFIER_FORMAT
}

static char *cli_show_settings(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ast_sip_cli_context context;

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show settings";
		e->usage = "Usage: pjsip show settings\n"
		            "      Show global and system configuration options\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	context.output_buffer = ast_str_create(256);
	if (!context.output_buffer) {
		ast_cli(a->fd, "Could not allocate output buffer.\n");
		return CLI_FAILURE;
	}

	if (sip_cli_print_global(&context) || sip_cli_print_system(&context)) {
		ast_free(context.output_buffer);
		ast_cli(a->fd, "Error retrieving settings.\n");
		return CLI_FAILURE;
	}

	ast_cli(a->fd, "%s", ast_str_buffer(context.output_buffer));
	ast_free(context.output_buffer);
	return CLI_SUCCESS;
}

#ifdef GRANDSTREAM_NETWORKS
static void cli_transaction(int fd, pjsip_transaction *tsx)
{
	ast_cli(fd, "%20s %20s %50s %20d %20s\n",
		tsx->obj_name, tsx->pool->obj_name,
		(tsx->last_tx ? pjsip_tx_data_get_info(tsx->last_tx): "none"),
		tsx->status_code,
		pjsip_tsx_state_str(tsx->state));
}

static char *cli_show_transaction(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show transaction";
		e->usage = "Usage: pjsip show transaction\n";
		return NULL;
	case CLI_GENERATE:
		break;
	}

	if (a->argc < 3) {
		return NULL;
	}

	ast_cli(a->fd, "%20s %80s %20s %20s\n", "tsx_name", "last_tx_tsx_info", "status_code", "state");
	pjsip_tsx_layer_dump2(a->fd, cli_transaction);
	ast_cli(a->fd, "\nTotal %d transaction!\n\n", pjsip_tsx_layer_get_tsx_count());

	return CLI_SUCCESS;
}

#include <pjsip-ua/sip_inv.h>
#include "asterisk/res_pjsip_session.h"

static void cli_dialog(int fd, pjsip_dialog *dlg, const char *fmt, ...)
{
	va_list ap;
	char str_dialog[1024] = {0};

	pjsip_inv_session * inv = NULL;
	RAII_VAR(struct ast_sip_session *, dlgsess, ast_sip_dialog_get_session(dlg), ao2_cleanup);

	va_start(ap, fmt);
	vsnprintf(str_dialog, sizeof(str_dialog) - 1, fmt, ap);
	va_end(ap);

	inv = pjsip_dlg_get_inv_session(dlg);
	if (inv) {
		ast_cli(fd, "%s %18s %18s %15s %5d %5d %18p[%2d]\n", str_dialog, inv->obj_name, inv->pool->obj_name,
			pjsip_inv_state_name(inv->state), dlg->sess_count, dlg->tsx_count, dlgsess, dlgsess ? ao2_ref(dlgsess, 0) : -1);
	} else {
		ast_cli(fd, "%s\n", str_dialog);
	}
}

static char *cli_show_dialog(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show dialog";
		e->usage = "Usage: pjsip show dialog\n";
		return NULL;
	case CLI_GENERATE:
		break;
	}

	if (a->argc < 3) {
		return NULL;
	}

	ast_cli(a->fd, "Dialog info:\n");
	ast_cli(a->fd, "%18s %18s %5s %5s %50s %18s %18s %15s %5s %5s %18s\n", "DlgName", "Cachpool", "Role",
		"State", "CallerID", "InvName", "Cachpool", "State", "sess", "tsx", "session");
	pjsip_ua_dump2(a->fd, cli_dialog);
	ast_cli(a->fd, "\nTotal %d dialog!\n", pjsip_ua_get_dlg_set_count());


	return CLI_SUCCESS;
}
static char *cli_show_callflip(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ao2_iterator i;
	struct callflip_user  *user;

	switch (cmd) {
	case CLI_INIT:
		e->command = "callflip show users";
		e->usage =
			"Usage: callflip show users\n"
			"       This will list all callflip status\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	ast_cli(a->fd, "\n" 
					"=============================================================\n"
					"--- List CallFlip Status ---------------------------------\n"
					"=============================================================\n"
					"===\n");

	i = ao2_iterator_init(callflip_users, 0);
	for (; (user = ao2_iterator_next(&i)); ao2_ref(user, -1)) {
		ao2_lock(user);

		ast_cli(a->fd, "=== ---------------------------------------------------------\n"
					"=== Extension:  %s\n"
					"=== Status:     %s\n"
					"=== RemoteId:  %s\n"
					"=== RemoteName: %s\n"
					"=== CallType: %s\n"
					"=== MediaType: %s\n",
					user->name, 
					user->enable,
					user->remote_id,
					user->remote_name,
					user->call_type,
					user->media_type);
		ast_cli(a->fd, "=== ---------------------------------------------------------\n===\n"); 

		ao2_unlock(user);
	}
	ao2_iterator_destroy(&i);
	ast_cli(a->fd, "=============================================================\n\n");

	return CLI_SUCCESS;

}

static char *cli_reload_conference_config(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "reload conference config";
		e->usage = "Usage: reload conference config\n";
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}

	if (a->argc < 3) {
		return NULL;
	}

	prase_mcm_conference_config(1);

	return CLI_SUCCESS;
}

static char *cli_show_cachpool(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a);
#endif

static struct ast_cli_entry cli_commands[] = {
	AST_CLI_DEFINE(cli_dump_endpt, "Dump the res_pjsip endpt internals"),
	AST_CLI_DEFINE(cli_show_settings, "Show global and system configuration options"),
	AST_CLI_DEFINE(cli_show_endpoint_identifiers, "List registered endpoint identifiers"),
#ifdef GRANDSTREAM_NETWORKS
	AST_CLI_DEFINE(cli_show_transaction, "List transaction"),
	AST_CLI_DEFINE(cli_show_dialog, "List dialog"),
	AST_CLI_DEFINE(cli_show_cachpool, "List cachpool"),
	AST_CLI_DEFINE(cli_show_callflip, "List callflip"),
	AST_CLI_DEFINE(cli_reload_conference_config, "reload conference config"),
#endif
};

AST_RWLIST_HEAD_STATIC(endpoint_formatters, ast_sip_endpoint_formatter);

void ast_sip_register_endpoint_formatter(struct ast_sip_endpoint_formatter *obj)
{
	SCOPED_LOCK(lock, &endpoint_formatters, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&endpoint_formatters, obj, next);
}

void ast_sip_unregister_endpoint_formatter(struct ast_sip_endpoint_formatter *obj)
{
	struct ast_sip_endpoint_formatter *i;
	SCOPED_LOCK(lock, &endpoint_formatters, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&endpoint_formatters, i, next) {
		if (i == obj) {
			AST_RWLIST_REMOVE_CURRENT(next);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

int ast_sip_format_endpoint_ami(struct ast_sip_endpoint *endpoint,
				struct ast_sip_ami *ami, int *count)
{
	int res = 0;
	struct ast_sip_endpoint_formatter *i;
	SCOPED_LOCK(lock, &endpoint_formatters, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);
	*count = 0;
	AST_RWLIST_TRAVERSE(&endpoint_formatters, i, next) {
		if (i->format_ami && ((res = i->format_ami(endpoint, ami)) < 0)) {
			return res;
		}

		if (!res) {
			(*count)++;
		}
	}
	return 0;
}

pjsip_endpoint *ast_sip_get_pjsip_endpoint(void)
{
	return ast_pjsip_endpoint;
}

int ast_sip_will_uri_survive_restart(pjsip_sip_uri *uri, struct ast_sip_endpoint *endpoint,
	pjsip_rx_data *rdata)
{
	pj_str_t host_name;
	int result = 1;

	/* Determine if the contact cannot survive a restart/boot. */
	if (uri->port == rdata->pkt_info.src_port
		&& !pj_strcmp(&uri->host,
			pj_cstr(&host_name, rdata->pkt_info.src_name))
		/* We have already checked if the URI scheme is sip: or sips: */
		&& PJSIP_TRANSPORT_IS_RELIABLE(rdata->tp_info.transport)) {
		pj_str_t type_name;

		/* Determine the transport parameter value */
		if (!strcasecmp("WSS", rdata->tp_info.transport->type_name)) {
			/* WSS is special, as it needs to be ws. */
			pj_cstr(&type_name, "ws");
		} else {
			pj_cstr(&type_name, rdata->tp_info.transport->type_name);
		}

		if (!pj_stricmp(&uri->transport_param, &type_name)
			&& (endpoint->nat.rewrite_contact
				/* Websockets are always rewritten */
				|| !pj_stricmp(&uri->transport_param,
					pj_cstr(&type_name, "ws")))) {
			/*
			 * The contact was rewritten to the reliable transport's
			 * source address.  Disconnecting the transport for any
			 * reason invalidates the contact.
			 */
			result = 0;
		}
	}

	return result;
}

int ast_sip_get_transport_name(const struct ast_sip_endpoint *endpoint,
	pjsip_sip_uri *sip_uri, char *buf, size_t buf_len)
{
	char *host = NULL;
	static const pj_str_t x_name = { AST_SIP_X_AST_TXP, AST_SIP_X_AST_TXP_LEN };
	pjsip_param *x_transport;
#if 0
#ifdef GRANDSTREAM_NETWORKS
	if (endpoint && !ast_strlen_zero(endpoint->transport))
#else
	if (!ast_strlen_zero(endpoint->transport))
#endif
	{
		ast_copy_string(buf, endpoint->transport, buf_len);
		return 0;
	}
#endif

	x_transport = pjsip_param_find(&sip_uri->other_param, &x_name);
	if (!x_transport) {
		return -1;
	}

	/* Only use x_transport if the uri host is an ip (4 or 6) address */
	host = ast_alloca(sip_uri->host.slen + 1);
	ast_copy_pj_str(host, &sip_uri->host, sip_uri->host.slen + 1);
	if (!ast_sockaddr_parse(NULL, host, PARSE_PORT_FORBID)) {
		return -1;
	}

	ast_copy_pj_str(buf, &x_transport->value, buf_len);

	return 0;
}

int ast_sip_dlg_set_transport(const struct ast_sip_endpoint *endpoint, pjsip_dialog *dlg,
	pjsip_tpselector *selector)
{
	pjsip_sip_uri *uri;
	pjsip_tpselector sel = { .type = PJSIP_TPSELECTOR_NONE, };

	uri = pjsip_uri_get_uri(dlg->target);
	if (!selector) {
		selector = &sel;
	}

	ast_sip_set_tpselector_from_ep_or_uri(endpoint, uri, selector);
	pjsip_dlg_set_transport(dlg, selector);

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for Bug 68712, 2016.08.28
 * check domain value, web have done some check, we check '[]' num.
 * return 1: sucess, return 0: false
 */
static unsigned int check_domain_value(const char *domain)
{
	unsigned int i = 0;
	unsigned int domain_len = 0;
	unsigned int left_bracket_num = 0;
	unsigned int right_bracket_num = 0;

	if (!domain) {
		ast_log(LOG_DEBUG, "Check domain name, domain is NULL\n");
	}

	domain_len = strlen(domain);

	for (i = 0; i < domain_len; i++) {
		if (domain[i] == '[') {
			left_bracket_num++;
			if (i != 0) {
				ast_log(LOG_WARNING, "domain name left bracket([) postion(%d) is not zero\n", i);
				return 0;
			}
		}
		if (domain[i] == ']') {
			right_bracket_num++;
		}
	}

	if (left_bracket_num != right_bracket_num) {
		ast_log(LOG_WARNING, "domain name left bracket num is not the same as right bracket num\n");
		return 0;
	}

	if (left_bracket_num > 1) {
		ast_log(LOG_WARNING, "domain name left bracket'[' num is %d\n", left_bracket_num);
		return 0;
	}

	if (right_bracket_num > 1) {
		ast_log(LOG_WARNING, "domain name right bracket']' num is %d\n", right_bracket_num);
		return 0;
	}

	return 1;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static int sip_dialog_create_from(pj_pool_t *pool, pj_str_t *from, const char *user,const char *domain,
	const pj_str_t *target, pjsip_tpselector *selector, pjsip_transport_type_e trans_type, const char *local_vlan_address)
#else
static int sip_dialog_create_from(pj_pool_t *pool, pj_str_t *from, const char *user,
	const char *domain, const pj_str_t *target, pjsip_tpselector *selector)
#endif
{
	pj_str_t tmp, local_addr;
	pjsip_uri *uri;
	pjsip_sip_uri *sip_uri = NULL;
	pjsip_transport_type_e type;
	int local_port;
	char default_user[PJSIP_MAX_URL_SIZE];
#ifdef GRANDSTREAM_NETWORKS
	/* Modify by bfdu for Bug 92041, 2017.09.11 */
	char *tmp_uri = NULL;
	pj_str_t tmp_local_uri;
	char uri_schema[8] = {0};
#endif

	if (ast_strlen_zero(user)) {
		ast_sip_get_default_from_user(default_user, sizeof(default_user));
		user = default_user;
	}

#ifdef GRANDSTREAM_NETWORKS
    if (PJSIP_TRANSPORT_UNSPECIFIED != trans_type) {
		type = trans_type;
    } else {
#endif
	/* Parse the provided target URI so we can determine what transport it will end up using */
	pj_strdup_with_null(pool, &tmp, target);

	if (!(uri = pjsip_parse_uri(pool, tmp.ptr, tmp.slen, 0)) ||
	    (!PJSIP_URI_SCHEME_IS_SIP(uri) && !PJSIP_URI_SCHEME_IS_SIPS(uri))) {
		return -1;
	}

	sip_uri = pjsip_uri_get_uri(uri);

	/* Determine the transport type to use */
	type = pjsip_transport_get_type_from_name(&sip_uri->transport_param);
	if (PJSIP_URI_SCHEME_IS_SIPS(sip_uri)) {
		if (type == PJSIP_TRANSPORT_UNSPECIFIED
			|| !(pjsip_transport_get_flag_from_type(type) & PJSIP_TRANSPORT_SECURE)) {
			type = PJSIP_TRANSPORT_TLS;
		}
	} else if (!sip_uri->transport_param.slen) {
		type = PJSIP_TRANSPORT_UDP;
	} else if (type == PJSIP_TRANSPORT_UNSPECIFIED) {
		return -1;
	}

	/* If the host is IPv6 turn the transport into an IPv6 version */
	if (pj_strchr(&sip_uri->host, ':')) {
		type |= PJSIP_TRANSPORT_IPV6;
	}
#ifdef GRANDSTREAM_NETWORKS
		if (PJSIP_URI_SCHEME_IS_SIPS(sip_uri)) {
			strncpy(uri_schema, "sips", sizeof(uri_schema)-1);
		}
	}

	if (!strlen(uri_schema) && from && from->slen) {
		pjsip_uri *uri_local;
		pjsip_sip_uri *sip_uri_local;

		if (!(uri_local = pjsip_parse_uri(pool, from->ptr, from->slen, 0)) ||
			 (!PJSIP_URI_SCHEME_IS_SIP(uri_local) && !PJSIP_URI_SCHEME_IS_SIPS(uri_local))) {
			return -1;
		}

		sip_uri_local = pjsip_uri_get_uri(uri_local);
		if (PJSIP_URI_SCHEME_IS_SIPS(sip_uri_local)) {
			strncpy(uri_schema, "sips", sizeof(uri_schema)-1);
		} 
	}

	if (!strlen(uri_schema)) {
		strncpy(uri_schema, "sip", sizeof(uri_schema)-1);
	}
#endif

	if (!ast_strlen_zero(domain)) {
		from->ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 68712, 2016.08.23 */
		/* web need check if both have '[]' or not. or may have some abnormal */
		unsigned int is_domain_right = check_domain_value(domain);
		if (!is_domain_right) {
			return -1;
		}
		unsigned char has_bracket = 0;
		if (*domain == '[') {
			has_bracket = 1;
		}
		from->slen = pj_ansi_snprintf(from->ptr, PJSIP_MAX_URL_SIZE,
				"<%s:%s@%s%s%s%s%s>",
				uri_schema,
				user,
				((type & PJSIP_TRANSPORT_IPV6) && !has_bracket) ? "[" : "",
				domain,
				((type & PJSIP_TRANSPORT_IPV6) && !has_bracket) ? "]" : "",
				(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? ";transport=" : "",
				(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? pjsip_transport_get_type_name(type) : "");

#else
		from->slen = pj_ansi_snprintf(from->ptr, PJSIP_MAX_URL_SIZE,
				"<sip:%s@%s%s%s>",
				user,
				domain,
				(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? ";transport=" : "",
				(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? pjsip_transport_get_type_name(type) : "");
#endif

#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu for Bug 92041, 2017.09.11 */
		pj_strdup_with_null(pool, &tmp_local_uri, from);
		tmp_uri = (char *)pjsip_parse_uri(pool, tmp_local_uri.ptr, tmp_local_uri.slen, 0);
		if (!tmp_uri) {
			ast_log(LOG_ERROR, "can't parse uri when set domain value is :%s, use the default process\n", domain);
		} else {
			return 0;
		}
#else
		return 0;
#endif
	}

	/* Get the local bound address for the transport that will be used when communicating with the provided URI */
	if (pjsip_tpmgr_find_local_addr(pjsip_endpt_get_tpmgr(ast_sip_get_pjsip_endpoint()), pool, type, selector,
							      &local_addr, &local_port) != PJ_SUCCESS) {

		/* If no local address can be retrieved using the transport manager use the host one */
		pj_strdup(pool, &local_addr, pj_gethostname());
		local_port = pjsip_transport_get_default_port_for_type(PJSIP_TRANSPORT_UDP);
	}

	/* If IPv6 was specified in the transport, set the proper type */
	if (pj_strchr(&local_addr, ':')) {
		type |= PJSIP_TRANSPORT_IPV6;

#ifdef GRANDSTREAM_NETWORKS
		if (!ast_strlen_zero(local_vlan_address)) {
			if (strstr(local_vlan_address, ":")) {
				ast_log(LOG_DEBUG, "Set local_addr use local_vlan_address '%s'\n", local_vlan_address);
				pj_strset2(&local_addr, (char *)local_vlan_address);
			} else {
				ast_log(LOG_ERROR, "local_vlan_address '%s' is not a IPV6 address\n", local_vlan_address);
			}
		}
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	else {
		const char *vlan_addr = ast_sip_get_vlan_addr();

		if (vlan_addr && sip_uri) {
			struct ast_sockaddr tmp_addr;
			char tmp_host[48] = {0};

			snprintf(tmp_host, sizeof(tmp_host), "%.*s", (int)pj_strlen(&sip_uri->host), pj_strbuf(&sip_uri->host));

			ast_sockaddr_parse(&tmp_addr, tmp_host, PARSE_ADDR);
			if (AST_SENSE_ALLOW != ast_apply_acl_with_name("lan", &tmp_addr)) {
				pj_strset2(&local_addr, (char *)vlan_addr);
			} else {
				pj_strset2(&local_addr, (char *)ast_sip_get_lan_addr());
			}
		}

		if (!ast_strlen_zero(local_vlan_address)) {
			if (!strstr(local_vlan_address, ":")) {
				ast_log(LOG_DEBUG, "Set local_addr use local_vlan_address '%s'\n", local_vlan_address);
				pj_strset2(&local_addr, (char *)local_vlan_address);
			} else {
				ast_log(LOG_ERROR, "local_vlan_address '%s' is not a IPV4 address\n", local_vlan_address);
			}
		}
	}
#endif
	from->ptr = pj_pool_alloc(pool, PJSIP_MAX_URL_SIZE);
	from->slen = pj_ansi_snprintf(from->ptr, PJSIP_MAX_URL_SIZE,
#ifdef GRANDSTREAM_NETWORKS
				      "<%s:%s@%s%.*s%s:%d%s%s>",
				      uri_schema,
#else
				      "<sip:%s@%s%.*s%s:%d%s%s>",
#endif
				      user,
				      (type & PJSIP_TRANSPORT_IPV6) ? "[" : "",
				      (int)local_addr.slen,
				      local_addr.ptr,
				      (type & PJSIP_TRANSPORT_IPV6) ? "]" : "",
				      local_port,
				      (type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? ";transport=" : "",
				      (type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? pjsip_transport_get_type_name(type) : "");

	return 0;
}

int ast_sip_set_tpselector_from_transport(const struct ast_sip_transport *transport, pjsip_tpselector *selector)
{
	RAII_VAR(struct ast_sip_transport_state *, transport_state, NULL, ao2_cleanup);

	transport_state = ast_sip_get_transport_state(ast_sorcery_object_get_id(transport));
	if (!transport_state) {
		ast_log(LOG_ERROR, "Unable to retrieve PJSIP transport state for '%s'\n",
			ast_sorcery_object_get_id(transport));
		return -1;
	}

	if (transport_state->transport) {
		selector->type = PJSIP_TPSELECTOR_TRANSPORT;
		selector->u.transport = transport_state->transport;
	} else if (transport_state->factory) {
		selector->type = PJSIP_TPSELECTOR_LISTENER;
		selector->u.listener = transport_state->factory;
	} else if (transport->type == AST_TRANSPORT_WS || transport->type == AST_TRANSPORT_WSS) {
		/* The WebSocket transport has no factory as it can not create outgoing connections, so
		 * even if an endpoint is locked to a WebSocket transport we let the PJSIP logic
		 * find the existing connection if available and use it.
		 */
		return 0;
	} else {
		return -1;
	}

	return 0;
}

int ast_sip_set_tpselector_from_transport_name(const char *transport_name, pjsip_tpselector *selector)
{
	RAII_VAR(struct ast_sip_transport *, transport, NULL, ao2_cleanup);

	if (ast_strlen_zero(transport_name)) {
		return 0;
	}

	transport = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "transport", transport_name);
	if (!transport) {
		ast_log(LOG_ERROR, "Unable to retrieve PJSIP transport '%s'\n",
			transport_name);
		return -1;
	}

	return ast_sip_set_tpselector_from_transport(transport, selector);
}

int ast_sip_set_tpselector_from_ep_or_uri(const struct ast_sip_endpoint *endpoint,
	pjsip_sip_uri *sip_uri, pjsip_tpselector *selector)
{
	char transport_name[128] = {0};

	if (ast_sip_get_transport_name(endpoint, sip_uri, transport_name, sizeof(transport_name))) {
		return 0;
	}

	return ast_sip_set_tpselector_from_transport_name(transport_name, selector);
}

void ast_sip_add_usereqphone(const struct ast_sip_endpoint *endpoint, pj_pool_t *pool, pjsip_uri *uri)
{
	pjsip_sip_uri *sip_uri;
	int i = 0;
	static const pj_str_t STR_PHONE = { "phone", 5 };

	if (!endpoint || !endpoint->usereqphone || (!PJSIP_URI_SCHEME_IS_SIP(uri) && !PJSIP_URI_SCHEME_IS_SIPS(uri))) {
		return;
	}

	sip_uri = pjsip_uri_get_uri(uri);

	if (!pj_strlen(&sip_uri->user)) {
		return;
	}

	if (pj_strbuf(&sip_uri->user)[0] == '+') {
		i = 1;
	}

	/* Test URI user against allowed characters in AST_DIGIT_ANY */
	for (; i < pj_strlen(&sip_uri->user); i++) {
		if (!strchr(AST_DIGIT_ANYNUM, pj_strbuf(&sip_uri->user)[i])) {
			break;
		}
	}

	if (i < pj_strlen(&sip_uri->user)) {
		return;
	}

	sip_uri->user_param = STR_PHONE;
}

pjsip_dialog *ast_sip_create_dialog_uac(const struct ast_sip_endpoint *endpoint,
#ifdef GRANDSTREAM_NETWORKS
	const char *uri, const char *request_user, pjsip_method_e method, pj_str_t *local, pj_str_t *remote, pj_str_t *callid
#else
	const char *uri, const char *request_user
#endif
)
{
	char enclosed_uri[PJSIP_MAX_URL_SIZE];
	pj_str_t local_uri = { "sip:temp@temp", 13 }, remote_uri, target_uri;
	pj_status_t res;
	pjsip_dialog *dlg = NULL;
	const char *outbound_proxy = endpoint->outbound_proxy;
	pjsip_tpselector selector = { .type = PJSIP_TPSELECTOR_NONE, };
	static const pj_str_t HCONTACT = { "Contact", 7 };

#ifdef GRANDSTREAM_NETWORKS
	char teluri[PJSIP_MAX_URL_SIZE];
	char orig_uri[PJSIP_MAX_URL_SIZE] = {0};
	const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
	unsigned int rmv_obp = endpoint->rmv_obp_from_route;
	const char *from_user = NULL;

	/* Added by yfyang, for HA service data sync, don't create uac when standby */
	if (HA_ROLE_STANDBY == ast_ha_get_role() && 0 ==ast_get_redundance_flag()) {
		return NULL;
	}

	pjsip_transport_type_e type = PJSIP_TRANSPORT_UNSPECIFIED;

	if ((AST_SIP_TEL_URI_ENABLED == endpoint->tel_uri) && (PJSIP_INVITE_METHOD == method)) {
		if (request_user) {
			snprintf(teluri, sizeof(teluri), "tel:%s", request_user);
		} else {
			snprintf(teluri, sizeof(teluri), "tel:%s", endpoint_id);
		}

		snprintf(orig_uri, sizeof(orig_uri), "%s;lr", uri);
		uri = teluri;

		if (!strncmp(orig_uri, "sips:", 5)) {
			pj_cstr(&local_uri, "sips:temp@temp");
		}
	}
#endif

	snprintf(enclosed_uri, sizeof(enclosed_uri), "<%s>", uri);
	pj_cstr(&remote_uri, enclosed_uri);

	pj_cstr(&target_uri, uri);

	if(local && remote && callid)
	{
		res = pjsip_dlg_create_uac_for_recovery(pjsip_ua_instance(), &local_uri, NULL, &remote_uri, &target_uri, local, remote, callid,&dlg);
	}
	else
	{
		res = pjsip_dlg_create_uac(pjsip_ua_instance(), &local_uri, NULL, &remote_uri, &target_uri, &dlg);
	}
	if (res != PJ_SUCCESS) {
		if (res == PJSIP_EINVALIDURI) {
			ast_log(LOG_ERROR,
				"Endpoint '%s': Could not create dialog to invalid URI '%s'.  Is endpoint registered and reachable?\n",
				ast_sorcery_object_get_id(endpoint), uri);
		}
		return NULL;
	}

	/* We have to temporarily bump up the sess_count here so the dialog is not prematurely destroyed */
	dlg->sess_count++;

#ifndef GRANDSTREAM_NETWORKS
	ast_sip_dlg_set_transport(endpoint, dlg, &selector);
#else
	if (AST_SIP_TEL_URI_ENABLED != endpoint->tel_uri) {
		ast_sip_dlg_set_transport(endpoint, dlg, &selector);
	} 
#endif
	
#ifdef GRANDSTREAM_NETWORKS
	if ((AST_SIP_TEL_URI_ENABLED == endpoint->tel_uri) && (PJSIP_INVITE_METHOD == method)) {
		pj_str_t tmp;
		pjsip_uri *uri;
		pjsip_sip_uri *sip_uri;

		pj_cstr(&tmp, orig_uri);
		if ((uri = pjsip_parse_uri(dlg->pool, tmp.ptr, tmp.slen, 0))) {
			sip_uri = pjsip_uri_get_uri(uri);

			if (PJSIP_URI_SCHEME_IS_SIPS(sip_uri)) {
				type = PJSIP_TRANSPORT_TLS;
			} else if (!sip_uri->transport_param.slen) {
				type = PJSIP_TRANSPORT_UDP;
			} else {
				type = pjsip_transport_get_type_from_name(&sip_uri->transport_param);
			}

			ast_sip_set_tpselector_from_ep_or_uri(endpoint, sip_uri, &selector);
			pjsip_dlg_set_transport(dlg, &selector);
		}
	}

	/* Added by yfyang, for Bug 97397 */
	if (PJSIP_INVITE_METHOD == method) {
		from_user = S_OR(endpoint->fromuser, endpoint->username);
	} else {
		/*Modify for Bug 187185*/
		from_user = S_OR(endpoint->cei_ext_fromuser, endpoint->username) ;
	}

	if (sip_dialog_create_from(dlg->pool, &local_uri, from_user, endpoint->fromdomain, &remote_uri, &selector, type, endpoint->local_vlan_address))
#else
	if (sip_dialog_create_from(dlg->pool, &local_uri, endpoint->fromuser, endpoint->fromdomain, &remote_uri, &selector)) 
#endif
	{
		dlg->sess_count--;
		pjsip_dlg_terminate(dlg);
		return NULL;
	}

	/* Update the dialog with the new local URI, we do it afterwards so we can use the dialog pool for construction */
	pj_strdup_with_null(dlg->pool, &dlg->local.info_str, &local_uri);
	dlg->local.info->uri = pjsip_parse_uri(dlg->pool, dlg->local.info_str.ptr, dlg->local.info_str.slen, 0);
	if (!dlg->local.info->uri) {
		ast_log(LOG_ERROR,
			"Could not parse URI '%s' for endpoint '%s'\n",
			dlg->local.info_str.ptr, ast_sorcery_object_get_id(endpoint));
		dlg->sess_count--;
		pjsip_dlg_terminate(dlg);
		return NULL;
	}

	dlg->local.contact = pjsip_parse_hdr(dlg->pool, &HCONTACT, local_uri.ptr, local_uri.slen, NULL);

	if (!ast_strlen_zero(endpoint->contact_user)) {
		pjsip_sip_uri *sip_uri;

		sip_uri = pjsip_uri_get_uri(dlg->local.contact->uri);
		pj_strdup2(dlg->pool, &sip_uri->user, endpoint->contact_user);
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for Bug 94413, 2017.11.22
	 * Don't use fromuser configure, update contact user based on username
	 */
	if (!ast_strlen_zero(endpoint->username) && dlg->local.contact) {
		pjsip_sip_uri *contact_hdr_uri = pjsip_uri_get_uri(dlg->local.contact->uri);
		if (contact_hdr_uri != NULL) {
			pj_strdup2(dlg->pool, &contact_hdr_uri->user, endpoint->username);
		}
	}

	if (!ast_strlen_zero(endpoint->cei_contactuser)) {
		pjsip_sip_uri *sip_uri;

		sip_uri = pjsip_uri_get_uri(dlg->local.contact->uri);
		pj_strdup2(dlg->pool, &sip_uri->user, endpoint->cei_contactuser);
	}
#endif

	/* If a request user has been specified and we are permitted to change it, do so */
	if (!ast_strlen_zero(request_user)) {
		pjsip_sip_uri *sip_uri;

		if (PJSIP_URI_SCHEME_IS_SIP(dlg->target) || PJSIP_URI_SCHEME_IS_SIPS(dlg->target)) {
			sip_uri = pjsip_uri_get_uri(dlg->target);
			pj_strdup2(dlg->pool, &sip_uri->user, request_user);
		}
		if (PJSIP_URI_SCHEME_IS_SIP(dlg->remote.info->uri) || PJSIP_URI_SCHEME_IS_SIPS(dlg->remote.info->uri)) {
			sip_uri = pjsip_uri_get_uri(dlg->remote.info->uri);
			pj_strdup2(dlg->pool, &sip_uri->user, request_user);
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	if (AST_SIP_URI_USER_PHONE == endpoint->tel_uri) {
		if (PJSIP_URI_SCHEME_IS_SIP(dlg->target) || PJSIP_URI_SCHEME_IS_SIPS(dlg->target)) {
			pjsip_sip_uri *sip_uri;
			sip_uri = pjsip_uri_get_uri(dlg->target);

			pj_strdup2(dlg->pool, &sip_uri->user_param, "phone");

			sip_uri = pjsip_uri_get_uri(dlg->remote.info->uri);
			pj_strdup2(dlg->pool, &sip_uri->user_param, "phone");
		}
	} else if(AST_SIP_TEL_URI_DISABLED == endpoint->tel_uri) {
		if (PJSIP_URI_SCHEME_IS_SIP(dlg->target) || PJSIP_URI_SCHEME_IS_SIPS(dlg->target)) {
			pjsip_sip_uri *sip_uri;
			sip_uri = pjsip_uri_get_uri(dlg->target);

			pj_strdup2(dlg->pool, &sip_uri->user_param, NULL);

			sip_uri = pjsip_uri_get_uri(dlg->remote.info->uri);
			pj_strdup2(dlg->pool, &sip_uri->user_param, NULL);
		}
	}
#endif

	/* Add the user=phone parameter if applicable */
	ast_sip_add_usereqphone(endpoint, dlg->pool, dlg->target);
	ast_sip_add_usereqphone(endpoint, dlg->pool, dlg->remote.info->uri);

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_strlen_zero(outbound_proxy) || ((AST_SIP_TEL_URI_ENABLED == endpoint->tel_uri) && (PJSIP_INVITE_METHOD == method)))
#else
	if (!ast_strlen_zero(outbound_proxy))
#endif
	{
		pjsip_route_hdr route_set, *route;
		static const pj_str_t ROUTE_HNAME = { "Route", 5 };
		pj_str_t tmp;

		pj_list_init(&route_set);

#ifdef GRANDSTREAM_NETWORKS
		if (ast_strlen_zero(outbound_proxy)) {
			pj_strdup2_with_null(dlg->pool, &tmp, orig_uri);
		} else {
			pj_strdup2_with_null(dlg->pool, &tmp, outbound_proxy);
		}
#else
		pj_strdup2_with_null(dlg->pool, &tmp, outbound_proxy);
#endif
		if (!(route = pjsip_parse_hdr(dlg->pool, &ROUTE_HNAME, tmp.ptr, tmp.slen, NULL))) {
			ast_log(LOG_ERROR, "Could not create dialog to endpoint '%s' as outbound proxy URI '%s' is not valid\n",
				ast_sorcery_object_get_id(endpoint), outbound_proxy);
			dlg->sess_count--;
			pjsip_dlg_terminate(dlg);
			return NULL;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (rmv_obp) {
			pjsip_param *p;

			p = PJ_POOL_ALLOC_T(dlg->pool, pjsip_param);
			p->name = pj_str("removed");
			p->value.slen = 0;
			pj_list_insert_before(&route->other_param, p);
		}
#endif
		pj_list_insert_nodes_before(&route_set, route);

		pjsip_dlg_set_route_set(dlg, &route_set);
	}

	dlg->sess_count--;

	return dlg;
}

/*!
 * \brief Determine if a SIPS Contact header is required.
 *
 * This uses the guideline provided in RFC 3261 Section 12.1.1 to
 * determine if the Contact header must be a sips: URI.
 *
 * \param rdata The incoming dialog-starting request
 * \retval 0 SIPS not required
 * \retval 1 SIPS required
 */
static int uas_use_sips_contact(pjsip_rx_data *rdata)
{
	pjsip_rr_hdr *record_route;

	if (PJSIP_URI_SCHEME_IS_SIPS(rdata->msg_info.msg->line.req.uri)) {
		return 1;
	}

	record_route = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_RECORD_ROUTE, NULL);
	if (record_route) {
		if (PJSIP_URI_SCHEME_IS_SIPS(&record_route->name_addr)) {
			return 1;
		}
	} else {
		pjsip_contact_hdr *contact;

		contact = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
		ast_assert(contact != NULL);
		if (PJSIP_URI_SCHEME_IS_SIPS(contact->uri)) {
			return 1;
		}
	}

	return 0;
}

pjsip_dialog *ast_sip_create_dialog_uas(const struct ast_sip_endpoint *endpoint, pjsip_rx_data *rdata, pj_status_t *status)
{
	pjsip_dialog *dlg;
	pj_str_t contact;
	pjsip_transport_type_e type = rdata->tp_info.transport->key.type;
	pjsip_tpselector selector = { .type = PJSIP_TPSELECTOR_NONE, };
	pjsip_transport *transport;
	pjsip_contact_hdr *contact_hdr;
#ifdef GRANDSTREAM_NETWORKS
	pjsip_contact_hdr *rdata_contact = NULL;
	pjsip_sip_uri *sip_rdata_contact = NULL;
	char rdata_contact_user[80] = {0};
	struct ast_sockaddr addr = { { 0, } };
	char *pos1 = NULL;
	char *pos2 = NULL;
	ast_sockaddr_parse(&addr, rdata->pkt_info.src_name, PARSE_PORT_FORBID);
#endif

	ast_assert(status != NULL);

	contact_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
	if (!contact_hdr || ast_sip_set_tpselector_from_ep_or_uri(endpoint, pjsip_uri_get_uri(contact_hdr->uri),
		&selector)) {
		return NULL;
	}

	transport = rdata->tp_info.transport;
	if (selector.type == PJSIP_TPSELECTOR_TRANSPORT) {
		transport = selector.u.transport;
	}
	type = transport->key.type;

#ifdef GRANDSTREAM_NETWORKS
	if (pj_stricmp2(&rdata->msg_info.cseq->method.name, "SUBSCRIBE")) {
		pjsip_fromto_hdr *to = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_TO, NULL);
		if (to) {
			pjsip_name_addr *to_name_addr = (pjsip_name_addr *)to->uri;
			if (to_name_addr) {
				sip_rdata_contact = pjsip_uri_get_uri(to_name_addr);
			}
		}
	} else {
		// Contact Header
		rdata_contact = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
		if ((rdata_contact && rdata_contact->uri)) {
			sip_rdata_contact = pjsip_uri_get_uri(rdata_contact->uri);
		}
	}

	if (sip_rdata_contact) {
		ast_copy_pj_str(rdata_contact_user, &sip_rdata_contact->user, sizeof(rdata_contact_user));
		if (!strcmp(rdata_contact_user, "\"\"")) {
			memset(rdata_contact_user, 0, sizeof(rdata_contact_user));
		/* BEGIN: Added by yfyang, for Bug247386 */
		} else if ((pos1 = strstr(rdata_contact_user, "@"))) {
			char rx_contact_user[80] = {0};
			char *unescaped = "%40";
			pos2 = pos1+1;
			*pos1 = '\0';
			snprintf(rx_contact_user, sizeof(rx_contact_user), "%s%s%s", rdata_contact_user, unescaped, pos2);
			strncpy(rdata_contact_user, rx_contact_user, sizeof(rdata_contact_user));
		}
		/* END: Added by yfyang, for Bug247386 */
	}
#endif

	contact.ptr = pj_pool_alloc(rdata->tp_info.pool, PJSIP_MAX_URL_SIZE);
#ifdef GRANDSTREAM_NETWORKS
	contact.slen = pj_ansi_snprintf(contact.ptr, PJSIP_MAX_URL_SIZE,
			"<%s:%s%s%s%.*s%s:%d%s%s>",
			uas_use_sips_contact(rdata) ? "sips" : "sip",
			rdata_contact_user,
			ast_strlen_zero(rdata_contact_user) ? "" : "@",
			((type & PJSIP_TRANSPORT_IPV6) || ast_sockaddr_is_ipv6(&addr)) ? "[" : "",
			(int)transport->local_name.host.slen,
			transport->local_name.host.ptr,
			((type & PJSIP_TRANSPORT_IPV6) || ast_sockaddr_is_ipv6(&addr)) ? "]" : "",
			transport->local_name.port,
			(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? ";transport=" : "",
			(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? pjsip_transport_get_type_name(type) : "");
#else
	contact.slen = pj_ansi_snprintf(contact.ptr, PJSIP_MAX_URL_SIZE,
			"<%s:%s%.*s%s:%d%s%s>",
			uas_use_sips_contact(rdata) ? "sips" : "sip",
			(type & PJSIP_TRANSPORT_IPV6) ? "[" : "",
			(int)transport->local_name.host.slen,
			transport->local_name.host.ptr,
			(type & PJSIP_TRANSPORT_IPV6) ? "]" : "",
			transport->local_name.port,
			(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? ";transport=" : "",
			(type != PJSIP_TRANSPORT_UDP && type != PJSIP_TRANSPORT_UDP6) ? pjsip_transport_get_type_name(type) : "");
#endif

#ifdef HAVE_PJSIP_DLG_CREATE_UAS_AND_INC_LOCK
	*status = pjsip_dlg_create_uas_and_inc_lock(pjsip_ua_instance(), rdata, &contact, &dlg);
#else
	*status = pjsip_dlg_create_uas(pjsip_ua_instance(), rdata, &contact, &dlg);
#endif
	if (*status != PJ_SUCCESS) {
		char err[PJ_ERR_MSG_SIZE];

		pj_strerror(*status, err, sizeof(err));
		ast_log(LOG_ERROR, "Could not create dialog with endpoint %s. %s\n",
				ast_sorcery_object_get_id(endpoint), err);
		return NULL;
	}

	dlg->sess_count++;
	pjsip_dlg_set_transport(dlg, &selector);
	dlg->sess_count--;
#ifdef HAVE_PJSIP_DLG_CREATE_UAS_AND_INC_LOCK
#ifndef GRANDSTREAM_NETWORKS
	pjsip_dlg_dec_lock(dlg);
#endif
#endif

	return dlg;
}

int ast_sip_create_rdata_with_contact(pjsip_rx_data *rdata, char *packet, const char *src_name, int src_port,
	char *transport_type, const char *local_name, int local_port, const char *contact)
{
	pj_str_t tmp;

	/*
	 * Initialize the error list in case there is a parse error
	 * in the given packet.
	 */
	pj_list_init(&rdata->msg_info.parse_err);

	rdata->tp_info.transport = PJ_POOL_ZALLOC_T(rdata->tp_info.pool, pjsip_transport);
	if (!rdata->tp_info.transport) {
		return -1;
	}

	ast_copy_string(rdata->pkt_info.packet, packet, sizeof(rdata->pkt_info.packet));
	ast_copy_string(rdata->pkt_info.src_name, src_name, sizeof(rdata->pkt_info.src_name));
	rdata->pkt_info.src_port = src_port;
	pj_sockaddr_parse(pj_AF_UNSPEC(), 0, pj_cstr(&tmp, src_name), &rdata->pkt_info.src_addr);
	pj_sockaddr_set_port(&rdata->pkt_info.src_addr, src_port);

	pjsip_parse_rdata(packet, strlen(packet), rdata);
	if (!rdata->msg_info.msg || !pj_list_empty(&rdata->msg_info.parse_err)) {
		return -1;
	}

	if (!ast_strlen_zero(contact)) {
		pjsip_contact_hdr *contact_hdr;

		contact_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
		if (contact_hdr) {
			contact_hdr->uri = pjsip_parse_uri(rdata->tp_info.pool, (char *)contact,
				strlen(contact), PJSIP_PARSE_URI_AS_NAMEADDR);
		}
	}

	pj_strdup2(rdata->tp_info.pool, &rdata->msg_info.via->recvd_param, rdata->pkt_info.src_name);
	rdata->msg_info.via->rport_param = -1;

	rdata->tp_info.transport->key.type = pjsip_transport_get_type_from_name(pj_cstr(&tmp, transport_type));
	rdata->tp_info.transport->type_name = transport_type;
	pj_strdup2(rdata->tp_info.pool, &rdata->tp_info.transport->local_name.host, local_name);
	rdata->tp_info.transport->local_name.port = local_port;

	return 0;
}

int ast_sip_create_rdata(pjsip_rx_data *rdata, char *packet, const char *src_name, int src_port,
	char *transport_type, const char *local_name, int local_port)
{
	return ast_sip_create_rdata_with_contact(rdata, packet, src_name, src_port, transport_type,
		local_name, local_port, NULL);
}

/* PJSIP doesn't know about the INFO method, so we have to define it ourselves */
static const pjsip_method info_method = {PJSIP_OTHER_METHOD, {"INFO", 4} };
static const pjsip_method message_method = {PJSIP_OTHER_METHOD, {"MESSAGE", 7} };
static const pjsip_method pjsip_update_method = {PJSIP_OTHER_METHOD, { "UPDATE", 6 } };

static struct {
	const char *method;
	const pjsip_method *pmethod;
} methods [] = {
	{ "INVITE", &pjsip_invite_method },
	{ "CANCEL", &pjsip_cancel_method },
	{ "ACK", &pjsip_ack_method },
	{ "BYE", &pjsip_bye_method },
	{ "REGISTER", &pjsip_register_method },
	{ "OPTIONS", &pjsip_options_method },
	{ "SUBSCRIBE", &pjsip_subscribe_method },
	{ "NOTIFY", &pjsip_notify_method },
	{ "PUBLISH", &pjsip_publish_method },
	{ "INFO", &info_method },
	{ "MESSAGE", &message_method },
	{ "UPDATE", &pjsip_update_method },
};

static const pjsip_method *get_pjsip_method(const char *method)
{
	int i;
	for (i = 0; i < ARRAY_LEN(methods); ++i) {
		if (!strcmp(method, methods[i].method)) {
			return methods[i].pmethod;
		}
	}
	return NULL;
}

static int create_in_dialog_request(const pjsip_method *method, struct pjsip_dialog *dlg, pjsip_tx_data **tdata)
{
	if (pjsip_dlg_create_request(dlg, method, -1, tdata) != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "Unable to create in-dialog request.\n");
		return -1;
	}

	return 0;
}

static pj_bool_t supplement_on_rx_request(pjsip_rx_data *rdata);
static pjsip_module supplement_module = {
	.name = { "Out of dialog supplement hook", 29 },
	.id = -1,
	.priority = PJSIP_MOD_PRIORITY_APPLICATION - 1,
	.on_rx_request = supplement_on_rx_request,
};

static int create_out_of_dialog_request(const pjsip_method *method, struct ast_sip_endpoint *endpoint,
		const char *uri, struct ast_sip_contact *provided_contact, pjsip_tx_data **tdata)
{
	RAII_VAR(struct ast_sip_contact *, contact, ao2_bump(provided_contact), ao2_cleanup);
	pj_str_t remote_uri;
#ifdef GRANDSTREAM_NETWORKS
	pj_str_t from = {NULL, 0};
#else
	pj_str_t from;
#endif
	pj_pool_t *pool;
	pjsip_tpselector selector = { .type = PJSIP_TPSELECTOR_NONE, };
	pjsip_uri *sip_uri;
	const char *fromuser;

	if (ast_strlen_zero(uri)) {
		if (!endpoint && (!contact || ast_strlen_zero(contact->uri))) {
			ast_log(LOG_ERROR, "An endpoint and/or uri must be specified\n");
			return -1;
		}

		if (!contact) {
			contact = ast_sip_location_retrieve_contact_from_aor_list(endpoint->aors);
		}
		if (!contact || ast_strlen_zero(contact->uri)) {
			ast_log(LOG_WARNING, "Unable to retrieve contact for endpoint %s\n",
					ast_sorcery_object_get_id(endpoint));
			return -1;
		}

		pj_cstr(&remote_uri, contact->uri);
	} else {
		pj_cstr(&remote_uri, uri);
	}

	pool = pjsip_endpt_create_pool(ast_sip_get_pjsip_endpoint(), "Outbound request", 256, 256);

	if (!pool) {
		ast_log(LOG_ERROR, "Unable to create PJLIB memory pool\n");
		return -1;
	}

	sip_uri = pjsip_parse_uri(pool, remote_uri.ptr, remote_uri.slen, 0);
	if (!sip_uri || (!PJSIP_URI_SCHEME_IS_SIP(sip_uri) && !PJSIP_URI_SCHEME_IS_SIPS(sip_uri))) {
		ast_log(LOG_ERROR, "Unable to create outbound %.*s request to endpoint %s as URI '%s' is not valid\n",
			(int) pj_strlen(&method->name), pj_strbuf(&method->name),
			endpoint ? ast_sorcery_object_get_id(endpoint) : "<none>",
			pj_strbuf(&remote_uri));
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), pool);
		return -1;
	}

	ast_sip_set_tpselector_from_ep_or_uri(endpoint, pjsip_uri_get_uri(sip_uri), &selector);

#ifndef GRANDSTREAM_NETWORKS
	fromuser = endpoint ? (!ast_strlen_zero(endpoint->fromuser) ? endpoint->fromuser : ast_sorcery_object_get_id(endpoint)) : NULL;
#else
	fromuser = NULL;
	char *endpoint_id = NULL;

	if (endpoint) {
		if (!ast_strlen_zero(endpoint->username)) {
			fromuser = endpoint->username;
		} else {
			endpoint_id = (char *)ast_sorcery_object_get_id(endpoint);
			if (strncmp(endpoint_id, "trunk_", 6)) {
				fromuser = endpoint_id;
			}
		}
	}
#endif

	if (sip_dialog_create_from(pool, &from, fromuser,
#ifdef GRANDSTREAM_NETWORKS
				endpoint ? endpoint->fromdomain : NULL, &remote_uri, &selector, PJSIP_TRANSPORT_UNSPECIFIED, endpoint ? endpoint->local_vlan_address : NULL))
#else
				endpoint ? endpoint->fromdomain : NULL, &remote_uri, &selector))
#endif
	{
		ast_log(LOG_ERROR, "Unable to create From header for %.*s request to endpoint %s\n",
				(int) pj_strlen(&method->name), pj_strbuf(&method->name),
				endpoint ? ast_sorcery_object_get_id(endpoint) : "<none>");
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), pool);
		return -1;
	}

	if (pjsip_endpt_create_request(ast_sip_get_pjsip_endpoint(), method, &remote_uri,
			&from, &remote_uri, &from, NULL, -1, NULL, tdata) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Unable to create outbound %.*s request to endpoint %s\n",
				(int) pj_strlen(&method->name), pj_strbuf(&method->name),
				endpoint ? ast_sorcery_object_get_id(endpoint) : "<none>");
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), pool);
		return -1;
	}

	pjsip_tx_data_set_transport(*tdata, &selector);

	if (endpoint && !ast_strlen_zero(endpoint->contact_user)){
		pjsip_contact_hdr *contact_hdr;
		pjsip_sip_uri *contact_uri;
		static const pj_str_t HCONTACT = { "Contact", 7 };
		static const pj_str_t HCONTACTSHORT = { "m", 1 };

		contact_hdr = pjsip_msg_find_hdr_by_names((*tdata)->msg, &HCONTACT, &HCONTACTSHORT, NULL);
		if (contact_hdr) {
			contact_uri = pjsip_uri_get_uri(contact_hdr->uri);
			pj_strdup2(pool, &contact_uri->user, endpoint->contact_user);
		}
	}

	/* Add the user=phone parameter if applicable */
	ast_sip_add_usereqphone(endpoint, (*tdata)->pool, (*tdata)->msg->line.req.uri);

#ifdef BUILD_ON_CLOUD_PBX
	if (contact && contact->sbc_endpoint_addr) {
		ast_sip_add_sbc_endpoint_address_header(contact->sbc_endpoint_addr, (*tdata));
	}
#endif

	/* If an outbound proxy is specified on the endpoint apply it to this request */
	if (endpoint && !ast_strlen_zero(endpoint->outbound_proxy) &&
#ifdef GRANDSTREAM_NETWORKS
		ast_sip_set_outbound_proxy((*tdata), endpoint->outbound_proxy, endpoint->rmv_obp_from_route))
#else
		ast_sip_set_outbound_proxy((*tdata), endpoint->outbound_proxy))
#endif
	{
		ast_log(LOG_ERROR, "Unable to apply outbound proxy on request %.*s to endpoint %s as outbound proxy URI '%s' is not valid\n",
			(int) pj_strlen(&method->name), pj_strbuf(&method->name), ast_sorcery_object_get_id(endpoint),
			endpoint->outbound_proxy);
		pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), pool);
		return -1;
	}

	ast_sip_mod_data_set((*tdata)->pool, (*tdata)->mod_data, supplement_module.id, MOD_DATA_CONTACT, ao2_bump(contact));

	/* We can release this pool since request creation copied all the necessary
	 * data into the outbound request's pool
	 */
	pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), pool);
	return 0;
}

int ast_sip_create_request(const char *method, struct pjsip_dialog *dlg,
		struct ast_sip_endpoint *endpoint, const char *uri,
		struct ast_sip_contact *contact, pjsip_tx_data **tdata)
{
	const pjsip_method *pmethod = get_pjsip_method(method);

	if (!pmethod) {
		ast_log(LOG_WARNING, "Unknown method '%s'. Cannot send request\n", method);
		return -1;
	}

	if (dlg) {
		return create_in_dialog_request(pmethod, dlg, tdata);
	} else {
		ast_assert(endpoint != NULL);
		return create_out_of_dialog_request(pmethod, endpoint, uri, contact, tdata);
	}
}

AST_RWLIST_HEAD_STATIC(supplements, ast_sip_supplement);

void ast_sip_register_supplement(struct ast_sip_supplement *supplement)
{
	struct ast_sip_supplement *iter;
	int inserted = 0;
	SCOPED_LOCK(lock, &supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&supplements, iter, next) {
		if (iter->priority > supplement->priority) {
			AST_RWLIST_INSERT_BEFORE_CURRENT(supplement, next);
			inserted = 1;
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;

	if (!inserted) {
		AST_RWLIST_INSERT_TAIL(&supplements, supplement, next);
	}
}

void ast_sip_unregister_supplement(struct ast_sip_supplement *supplement)
{
	struct ast_sip_supplement *iter;
	SCOPED_LOCK(lock, &supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&supplements, iter, next) {
		if (supplement == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

static int send_in_dialog_request(pjsip_tx_data *tdata, struct pjsip_dialog *dlg)
{
	if (pjsip_dlg_send_request(dlg, tdata, -1, NULL) != PJ_SUCCESS) {
		ast_log(LOG_WARNING, "Unable to send in-dialog request.\n");
		return -1;
	}
	return 0;
}

static pj_bool_t does_method_match(const pj_str_t *message_method, const char *supplement_method)
{
	pj_str_t method;

	if (ast_strlen_zero(supplement_method)) {
		return PJ_TRUE;
	}

	pj_cstr(&method, supplement_method);

	return pj_stristr(&method, message_method) ? PJ_TRUE : PJ_FALSE;
}

/*! Maximum number of challenges before assuming that we are in a loop */
#define MAX_RX_CHALLENGES	10
#define TIMER_INACTIVE		0
#define TIMEOUT_TIMER2		5

/*! \brief Structure to hold information about an outbound request */
struct send_request_data {
	/*! The endpoint associated with this request */
	struct ast_sip_endpoint *endpoint;
	/*! Information to be provided to the callback upon receipt of a response */
	void *token;
	/*! The callback to be called upon receipt of a response */
	void (*callback)(void *token, pjsip_event *e);
	/*! Number of challenges received. */
	unsigned int challenge_count;
};

static void send_request_data_destroy(void *obj)
{
	struct send_request_data *req_data = obj;

	ao2_cleanup(req_data->endpoint);
}

static struct send_request_data *send_request_data_alloc(struct ast_sip_endpoint *endpoint,
	void *token, void (*callback)(void *token, pjsip_event *e))
{
	struct send_request_data *req_data;

	req_data = ao2_alloc_options(sizeof(*req_data), send_request_data_destroy,
		AO2_ALLOC_OPT_LOCK_NOLOCK);
	if (!req_data) {
		return NULL;
	}

	req_data->endpoint = ao2_bump(endpoint);
	req_data->token = token;
	req_data->callback = callback;

	return req_data;
}

struct send_request_wrapper {
	/*! Information to be provided to the callback upon receipt of a response */
	void *token;
	/*! The callback to be called upon receipt of a response */
	void (*callback)(void *token, pjsip_event *e);
	/*! Non-zero when the callback is called. */
	unsigned int cb_called;
	/*! Non-zero if endpt_send_request_cb() was called. */
	unsigned int send_cb_called;
	/*! Timeout timer. */
	pj_timer_entry *timeout_timer;
	/*! Original timeout. */
	pj_int32_t timeout;
	/*! The transmit data. */
	pjsip_tx_data *tdata;
};

/*! \internal This function gets called by pjsip when the transaction ends,
 * even if it timed out.  The lock prevents a race condition if both the pjsip
 * transaction timer and our own timer expire simultaneously.
 */
static void endpt_send_request_cb(void *token, pjsip_event *e)
{
	struct send_request_wrapper *req_wrapper = token;
	unsigned int cb_called;

	/*
	 * Needed because we cannot otherwise tell if this callback was
	 * called when pjsip_endpt_send_request() returns error.
	 */
	req_wrapper->send_cb_called = 1;

	if (e->body.tsx_state.type == PJSIP_EVENT_TIMER) {
		ast_debug(2, "%p: PJSIP tsx timer expired\n", req_wrapper);

		if (req_wrapper->timeout_timer
			&& req_wrapper->timeout_timer->id != TIMEOUT_TIMER2) {
			ast_debug(3, "%p: Timeout already handled\n", req_wrapper);
			ao2_ref(req_wrapper, -1);
			return;
		}
	} else {
		ast_debug(2, "%p: PJSIP tsx response received\n", req_wrapper);
	}

	ao2_lock(req_wrapper);

	/* It's possible that our own timer was already processing while
	 * we were waiting on the lock so check the timer id.  If it's
	 * still TIMER2 then we still need to process.
	 */
	if (req_wrapper->timeout_timer
		&& req_wrapper->timeout_timer->id == TIMEOUT_TIMER2) {
		int timers_cancelled = 0;

		ast_debug(3, "%p: Cancelling timer\n", req_wrapper);

		timers_cancelled = pj_timer_heap_cancel_if_active(
			pjsip_endpt_get_timer_heap(ast_sip_get_pjsip_endpoint()),
			req_wrapper->timeout_timer, TIMER_INACTIVE);
		if (timers_cancelled > 0) {
			/* If the timer was cancelled the callback will never run so
			 * clean up its reference to the wrapper.
			 */
			ast_debug(3, "%p: Timer cancelled\n", req_wrapper);
			ao2_ref(req_wrapper, -1);
		} else {
			/*
			 * If it wasn't cancelled, it MAY be in the callback already
			 * waiting on the lock.  When we release the lock, it will
			 * now know not to proceed.
			 */
			ast_debug(3, "%p: Timer already expired\n", req_wrapper);
		}
	}

	cb_called = req_wrapper->cb_called;
	req_wrapper->cb_called = 1;
	ao2_unlock(req_wrapper);

	/* It's possible that our own timer expired and called the callbacks
	 * so no need to call them again.
	 */
	if (!cb_called && req_wrapper->callback) {
		req_wrapper->callback(req_wrapper->token, e);
		ast_debug(2, "%p: Callbacks executed\n", req_wrapper);
	}

	ao2_ref(req_wrapper, -1);
}

/*! \internal This function gets called by our own timer when it expires.
 * If the timer is cancelled however, the function does NOT get called.
 * The lock prevents a race condition if both the pjsip transaction timer
 * and our own timer expire simultaneously.
 */
static void send_request_timer_callback(pj_timer_heap_t *theap, pj_timer_entry *entry)
{
	struct send_request_wrapper *req_wrapper = entry->user_data;
	unsigned int cb_called;

	ast_debug(2, "%p: Internal tsx timer expired after %d msec\n",
		req_wrapper, req_wrapper->timeout);

	ao2_lock(req_wrapper);
	/*
	 * If the id is not TIMEOUT_TIMER2 then the timer was cancelled
	 * before we got the lock or it was already handled so just clean up.
	 */
	if (entry->id != TIMEOUT_TIMER2) {
		ao2_unlock(req_wrapper);
		ast_debug(3, "%p: Timeout already handled\n", req_wrapper);
		ao2_ref(req_wrapper, -1);
		return;
	}
	entry->id = TIMER_INACTIVE;

	ast_debug(3, "%p: Timer handled here\n", req_wrapper);

	cb_called = req_wrapper->cb_called;
	req_wrapper->cb_called = 1;
	ao2_unlock(req_wrapper);

	if (!cb_called && req_wrapper->callback) {
		pjsip_event event;

		PJSIP_EVENT_INIT_TX_MSG(event, req_wrapper->tdata);
		event.body.tsx_state.type = PJSIP_EVENT_TIMER;

		req_wrapper->callback(req_wrapper->token, &event);
		ast_debug(2, "%p: Callbacks executed\n", req_wrapper);
	}

	ao2_ref(req_wrapper, -1);
}

static void send_request_wrapper_destructor(void *obj)
{
	struct send_request_wrapper *req_wrapper = obj;

	pjsip_tx_data_dec_ref(req_wrapper->tdata);
	ast_debug(2, "%p: wrapper destroyed\n", req_wrapper);
}

static pj_status_t endpt_send_request(struct ast_sip_endpoint *endpoint,
	pjsip_tx_data *tdata, pj_int32_t timeout, void *token, pjsip_endpt_send_callback cb)
{
	struct send_request_wrapper *req_wrapper;
	pj_status_t ret_val;
	pjsip_endpoint *endpt = ast_sip_get_pjsip_endpoint();

	if (!cb && token) {
		/* Silly.  Without a callback we cannot do anything with token. */
		pjsip_tx_data_dec_ref(tdata);
		return PJ_EINVAL;
	}

	/* Create wrapper to detect if the callback was actually called on an error. */
	req_wrapper = ao2_alloc(sizeof(*req_wrapper), send_request_wrapper_destructor);
	if (!req_wrapper) {
		pjsip_tx_data_dec_ref(tdata);
		return PJ_ENOMEM;
	}

	ast_debug(2, "%p: Wrapper created\n", req_wrapper);

	req_wrapper->token = token;
	req_wrapper->callback = cb;
	req_wrapper->timeout = timeout;
	req_wrapper->timeout_timer = NULL;
	req_wrapper->tdata = tdata;
	/* Add a reference to tdata.  The wrapper destructor cleans it up. */
	pjsip_tx_data_add_ref(tdata);

	if (timeout > 0) {
		pj_time_val timeout_timer_val = { timeout / 1000, timeout % 1000 };

		req_wrapper->timeout_timer = PJ_POOL_ALLOC_T(tdata->pool, pj_timer_entry);

		ast_debug(2, "%p: Set timer to %d msec\n", req_wrapper, timeout);

		pj_timer_entry_init(req_wrapper->timeout_timer, TIMEOUT_TIMER2,
			req_wrapper, send_request_timer_callback);

		/* We need to insure that the wrapper and tdata are available if/when the
		 * timer callback is executed.
		 */
		ao2_ref(req_wrapper, +1);
		ret_val = pj_timer_heap_schedule(pjsip_endpt_get_timer_heap(endpt),
			req_wrapper->timeout_timer, &timeout_timer_val);
		if (ret_val != PJ_SUCCESS) {
			ast_log(LOG_ERROR,
				"Failed to set timer.  Not sending %.*s request to endpoint %s.\n",
				(int) pj_strlen(&tdata->msg->line.req.method.name),
				pj_strbuf(&tdata->msg->line.req.method.name),
				endpoint ? ast_sorcery_object_get_id(endpoint) : "<unknown>");
			ao2_t_ref(req_wrapper, -2, "Drop timer and routine ref");
			pjsip_tx_data_dec_ref(tdata);
			return ret_val;
		}
	}

	/* We need to insure that the wrapper and tdata are available when the
	 * transaction callback is executed.
	 */
	ao2_ref(req_wrapper, +1);
	ret_val = pjsip_endpt_send_request(endpt, tdata, -1, req_wrapper, endpt_send_request_cb);
	if (ret_val != PJ_SUCCESS) {
		char errmsg[PJ_ERR_MSG_SIZE];

		if (!req_wrapper->send_cb_called) {
			/* endpt_send_request_cb is not expected to ever be called now. */
			ao2_ref(req_wrapper, -1);
		}

		/* Complain of failure to send the request. */
		pj_strerror(ret_val, errmsg, sizeof(errmsg));
		ast_log(LOG_ERROR, "Error %d '%s' sending %.*s request to endpoint %s\n",
			(int) ret_val, errmsg, (int) pj_strlen(&tdata->msg->line.req.method.name),
			pj_strbuf(&tdata->msg->line.req.method.name),
			endpoint ? ast_sorcery_object_get_id(endpoint) : "<unknown>");

		if (timeout > 0) {
			int timers_cancelled;

			ao2_lock(req_wrapper);
			timers_cancelled = pj_timer_heap_cancel_if_active(
				pjsip_endpt_get_timer_heap(endpt),
				req_wrapper->timeout_timer, TIMER_INACTIVE);
			if (timers_cancelled > 0) {
				ao2_ref(req_wrapper, -1);
			}

			/* Was the callback called? */
			if (req_wrapper->cb_called) {
				/*
				 * Yes so we cannot report any error.  The callback
				 * has already freed any resources associated with
				 * token.
				 */
				ret_val = PJ_SUCCESS;
			} else {
				/*
				 * No so we claim it is called so our caller can free
				 * any resources associated with token because of
				 * failure.
				 */
				req_wrapper->cb_called = 1;
			}
			ao2_unlock(req_wrapper);
		} else if (req_wrapper->cb_called) {
			/*
			 * We cannot report any error.  The callback has
			 * already freed any resources associated with
			 * token.
			 */
			ret_val = PJ_SUCCESS;
		}
	}

	ao2_ref(req_wrapper, -1);
	return ret_val;
}

int ast_sip_failover_request(pjsip_tx_data *tdata)
{
	pjsip_via_hdr *via;

	if (!tdata || !tdata->dest_info.addr.count
		|| (tdata->dest_info.cur_addr >= tdata->dest_info.addr.count - 1)) {
		/* No more addresses to try */
		return 0;
	}

	/* Try next address */
	++tdata->dest_info.cur_addr;

	via = (pjsip_via_hdr*)pjsip_msg_find_hdr(tdata->msg, PJSIP_H_VIA, NULL);
	via->branch_param.slen = 0;

	pjsip_tx_data_invalidate_msg(tdata);

	return 1;
}

#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, for Bug157204 */
int ast_sip_failover_request_clone(pjsip_tx_data *tdata, pjsip_tx_data **new_tdata)
{
	pjsip_via_hdr *via;

	if (!tdata || !tdata->dest_info.addr.count
		|| (tdata->dest_info.cur_addr >= tdata->dest_info.addr.count - 1)) {
		/* No more addresses to try */
		return 0;
	}

	/* Try next address */
	++tdata->dest_info.cur_addr;

	via = (pjsip_via_hdr*)pjsip_msg_find_hdr(tdata->msg, PJSIP_H_VIA, NULL);
	via->branch_param.slen = 0;

	pjsip_tx_data_invalidate_msg(tdata);

	if (PJ_SUCCESS != pjsip_tx_data_clone_req(tdata, new_tdata)) {
		ast_log(LOG_NOTICE, "ast_sip_failover_request_clone, clone req failed.");
		return 0;
	}

	return 1;
}

#ifdef BUILD_ON_CLOUD_PBX
void ast_sip_add_trunk_host_header(struct ast_sip_endpoint *endpoint, pjsip_tx_data *tdata)
{
	pjsip_generic_string_hdr  *hdr_old;
	pj_str_t tkhost_name = { "Trunk-Host", 10};
	char  addrs[128] = {0};

	if (!tdata || !endpoint) {
		return;
	}

	if (ast_strlen_zero(endpoint->trunk_host)) {
		return;
	}

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &tkhost_name, NULL))) {
		return;
	}

	ast_sip_add_header(tdata, "Trunk-Host", endpoint->trunk_host);
	ast_sip_resolve_identifier_get_destaddrs((char *)endpoint->trunk_host, addrs, sizeof(addrs));

	if (strlen(addrs)) {
		ast_sip_add_header(tdata, "Trunk-IP", addrs);
	}

	return;
}

/* SBC-Endpoint-Address: ip:port */
void ast_sip_add_sbc_endpoint_address_header(const char *addr, pjsip_tx_data *tdata)
{
	pjsip_generic_string_hdr  *hdr_old;
	pj_str_t sbc_endpt_addr_name = { "SBC-Endpoint-Address", 20};

	if (!tdata || !addr) {
		return;
	}

	if (ast_strlen_zero(addr)) {
		return;
	}

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &sbc_endpt_addr_name, NULL))) {
		return;
	}

	ast_sip_add_header(tdata, "SBC-Endpoint-Address", addr);
    return;
}


void ast_sip_add_sbc_rand_addr_header(struct ast_sip_endpoint *endpoint, pjsip_tx_data *tdata)
{
	pjsip_generic_string_hdr  *hdr_old;
	pj_str_t sbc_rand_name = { "SBC-Rand-Addr", 13};
    char  addrs[128] = {0};

	if (!tdata || !endpoint) {
		return;
	}

	if (ast_strlen_zero(endpoint->sbc_rand_addr)) {
		return;
	}

	if ((hdr_old = pjsip_msg_find_hdr_by_name(tdata->msg, &sbc_rand_name, NULL))) {
		return;
	}

	ast_sip_add_header(tdata, "SBC-Rand-Addr", endpoint->sbc_rand_addr);

	return;
}

#endif
#endif

static void send_request_cb(void *token, pjsip_event *e);

static int check_request_status(struct send_request_data *req_data, pjsip_event *e)
{
	struct ast_sip_endpoint *endpoint;
	pjsip_transaction *tsx;
	pjsip_tx_data *tdata;
	pjsip_tx_data *new_tdata;
	int res = 0;

	if (!(endpoint = ao2_bump(req_data->endpoint))) {
		return 0;
	}

	tsx = e->body.tsx_state.tsx;

	switch (tsx->status_code) {
	case 401:
	case 407:
		/* Resend the request with a challenge response if we are challenged. */
		res = ++req_data->challenge_count < MAX_RX_CHALLENGES /* Not in a challenge loop */
			&& !ast_sip_create_request_with_auth(&endpoint->outbound_auths,
				e->body.tsx_state.src.rdata, tsx->last_tx, &tdata);
		break;
	case 408:
	case 503:
#ifdef GRANDSTREAM_NETWORKS
		/* Modified by yfyang, for Bug157204 */
		if ((res = ast_sip_failover_request_clone(tsx->last_tx, &new_tdata))) {
			ast_log(LOG_NOTICE, "check_request_status, call endpt_send_request(), new_tdata[%p], req_data[%p]", new_tdata, req_data);
			res = endpt_send_request(endpoint, new_tdata, -1, req_data, send_request_cb) == PJ_SUCCESS;
			ao2_ref(endpoint, -1);
			return res;
		}
#else
		if ((res = ast_sip_failover_request(tsx->last_tx))) {
			tdata = tsx->last_tx;

			/*
			 * Bump the ref since it will be on a new transaction and
			 * we don't want it to go away along with the old transaction.
			 */
			pjsip_tx_data_add_ref(tdata);
		}
#endif
		break;
	}

	if (res) {
		res = endpt_send_request(endpoint, tdata, -1,
					 req_data, send_request_cb) == PJ_SUCCESS;
	}

	ao2_ref(endpoint, -1);
	return res;
}

static void send_request_cb(void *token, pjsip_event *e)
{
	struct send_request_data *req_data = token;
	pjsip_rx_data *challenge;
	struct ast_sip_supplement *supplement;

	if (e->type == PJSIP_EVENT_TSX_STATE) {
		switch(e->body.tsx_state.type) {
		case PJSIP_EVENT_TRANSPORT_ERROR:
		case PJSIP_EVENT_TIMER:
			/*
			 * Check the request status on transport error or timeout. A transport
			 * error can occur when a TCP socket closes and that can be the result
			 * of a 503. Also we may need to failover on a timeout (408).
			 */
			if (check_request_status(req_data, e)) {
				return;
			}
			break;
		case PJSIP_EVENT_RX_MSG:
			challenge = e->body.tsx_state.src.rdata;

			/*
			 * Call any supplements that want to know about a response
			 * with any received data.
			 */
			AST_RWLIST_RDLOCK(&supplements);
			AST_LIST_TRAVERSE(&supplements, supplement, next) {
				if (supplement->incoming_response
					&& does_method_match(&challenge->msg_info.cseq->method.name,
						supplement->method)) {
					supplement->incoming_response(req_data->endpoint, challenge);
				}
			}
			AST_RWLIST_UNLOCK(&supplements);

			if (check_request_status(req_data, e)) {
				/*
				 * Request with challenge response or failover sent.
				 * Passed our req_data ref to the new request.
				 */
				return;
			}
			break;
		default:
			ast_log(LOG_ERROR, "Unexpected PJSIP event %u\n", e->body.tsx_state.type);
			break;
		}
	}

	if (req_data->callback) {
		req_data->callback(req_data->token, e);
	}
	ao2_ref(req_data, -1);
}

int ast_sip_send_out_of_dialog_request(pjsip_tx_data *tdata,
	struct ast_sip_endpoint *endpoint, int timeout, void *token,
	void (*callback)(void *token, pjsip_event *e))
{
	struct ast_sip_supplement *supplement;
	struct send_request_data *req_data;
	struct ast_sip_contact *contact;

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for HA service data sync, don't send messge out when standby */
	if (HA_ROLE_STANDBY == ast_ha_get_role()) {
		pjsip_tx_data_dec_ref(tdata);
		return -1;
	}
#endif

	req_data = send_request_data_alloc(endpoint, token, callback);
	if (!req_data) {
		pjsip_tx_data_dec_ref(tdata);
		return -1;
	}

	contact = ast_sip_mod_data_get(tdata->mod_data, supplement_module.id, MOD_DATA_CONTACT);

	AST_RWLIST_RDLOCK(&supplements);
	AST_LIST_TRAVERSE(&supplements, supplement, next) {
		if (supplement->outgoing_request
			&& does_method_match(&tdata->msg->line.req.method.name, supplement->method)) {
			supplement->outgoing_request(endpoint, contact, tdata);
		}
	}
	AST_RWLIST_UNLOCK(&supplements);

#ifdef BUILD_ON_CLOUD_PBX
	ast_sip_add_trunk_host_header(endpoint, tdata);
	ast_sip_add_sbc_rand_addr_header(endpoint, tdata);
#endif

	ast_sip_mod_data_set(tdata->pool, tdata->mod_data, supplement_module.id, MOD_DATA_CONTACT, NULL);
	ao2_cleanup(contact);

	if (endpt_send_request(endpoint, tdata, timeout, req_data, send_request_cb)
		!= PJ_SUCCESS) {
		ao2_cleanup(req_data);
		return -1;
	}

	return 0;
}

int ast_sip_send_request(pjsip_tx_data *tdata, struct pjsip_dialog *dlg,
	struct ast_sip_endpoint *endpoint, void *token,
	void (*callback)(void *token, pjsip_event *e))
{
	ast_assert(tdata->msg->type == PJSIP_REQUEST_MSG);
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for HA service data sync, don't send messge out when standby */
	if (HA_ROLE_STANDBY == ast_ha_get_role()) {
		pjsip_tx_data_dec_ref( tdata );
		return -1;
	}

#ifdef BUILD_ON_CLOUD_PBX
	ast_sip_add_trunk_host_header(endpoint, tdata);
	ast_sip_add_sbc_rand_addr_header(endpoint, tdata);
#endif
#endif

	if (dlg) {
		return send_in_dialog_request(tdata, dlg);
	} else {
		return ast_sip_send_out_of_dialog_request(tdata, endpoint, -1, token, callback);
	}
}

#ifdef GRANDSTREAM_NETWORKS
int ast_sip_set_outbound_proxy(pjsip_tx_data *tdata, const char *proxy, unsigned int rmv_flag)
#else
int ast_sip_set_outbound_proxy(pjsip_tx_data *tdata, const char *proxy)
#endif
{
	pjsip_route_hdr *route;
	static const pj_str_t ROUTE_HNAME = { "Route", 5 };
	pj_str_t tmp;

	pj_strdup2_with_null(tdata->pool, &tmp, proxy);
	if (!(route = pjsip_parse_hdr(tdata->pool, &ROUTE_HNAME, tmp.ptr, tmp.slen, NULL))) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (rmv_flag) {
		pjsip_param *p;

		p = PJ_POOL_ALLOC_T(tdata->pool, pjsip_param);
		p->name = pj_str("removed");
		p->value.slen = 0;
		pj_list_insert_before(&route->other_param, p);
	}
#endif
	pj_list_insert_nodes_before(&tdata->msg->hdr, (pjsip_hdr*)route);

	return 0;
}

int ast_sip_add_header(pjsip_tx_data *tdata, const char *name, const char *value)
{
	pj_str_t hdr_name;
	pj_str_t hdr_value;
	pjsip_generic_string_hdr *hdr;

	pj_cstr(&hdr_name, name);
	pj_cstr(&hdr_value, value);

	hdr = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);

	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *) hdr);
	return 0;
}

static pjsip_msg_body *ast_body_to_pjsip_body(pj_pool_t *pool, const struct ast_sip_body *body)
{
	pj_str_t type;
	pj_str_t subtype;
	pj_str_t body_text;

	pj_cstr(&type, body->type);
	pj_cstr(&subtype, body->subtype);
	pj_cstr(&body_text, body->body_text);

	return pjsip_msg_body_create(pool, &type, &subtype, &body_text);
}

int ast_sip_add_body(pjsip_tx_data *tdata, const struct ast_sip_body *body)
{
	pjsip_msg_body *pjsip_body = ast_body_to_pjsip_body(tdata->pool, body);
	tdata->msg->body = pjsip_body;
	return 0;
}

int ast_sip_add_body_multipart(pjsip_tx_data *tdata, const struct ast_sip_body *bodies[], int num_bodies)
{
	int i;
	/* NULL for type and subtype automatically creates "multipart/mixed" */
	pjsip_msg_body *body = pjsip_multipart_create(tdata->pool, NULL, NULL);

	for (i = 0; i < num_bodies; ++i) {
		pjsip_multipart_part *part = pjsip_multipart_create_part(tdata->pool);
		part->body = ast_body_to_pjsip_body(tdata->pool, bodies[i]);
		pjsip_multipart_add_part(tdata->pool, body, part);
	}

	tdata->msg->body = body;
	return 0;
}

int ast_sip_append_body(pjsip_tx_data *tdata, const char *body_text)
{
	size_t combined_size = strlen(body_text) + tdata->msg->body->len;
	struct ast_str *body_buffer = ast_str_alloca(combined_size);

	ast_str_set(&body_buffer, 0, "%.*s%s", (int) tdata->msg->body->len, (char *) tdata->msg->body->data, body_text);

	tdata->msg->body->data = pj_pool_alloc(tdata->pool, combined_size);
	pj_memcpy(tdata->msg->body->data, ast_str_buffer(body_buffer), combined_size);
	tdata->msg->body->len = combined_size;

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS

AST_RWLIST_HEAD_STATIC(sub_supplements, ast_sip_sub_supplement);

int ast_sip_sub_register_supplement(struct ast_sip_sub_supplement *supplement)
{
	if (!supplement->rx_response_notify || !supplement->rx_response_subscribe) {
		return -1;
	}

	SCOPED_LOCK(lock, &sub_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&sub_supplements, supplement, next);
	ast_module_ref(ast_module_info->self);

	return 0;
}

void ast_sip_sub_unregister_supplement(struct ast_sip_sub_supplement *supplement)
{
	struct ast_sip_sub_supplement *iter;
	SCOPED_LOCK(lock, &sub_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&sub_supplements, iter, next) {
		if (supplement == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

int ast_sip_process_for_supplement_sub(pjsip_dialog *dlg, int code, pjsip_rx_data *rdata)
{
	struct ast_sip_sub_supplement *iter;
	int  is_notify = 0;

	if (!pj_stricmp2(&rdata->msg_info.cseq->method.name, "NOTIFY")) {
		is_notify = 1;
	}

	SCOPED_LOCK(lock, &sub_supplements, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&sub_supplements, iter, next) {

		if (is_notify) {
			iter->rx_response_notify(rdata, dlg, code);
		} else {
			iter->rx_response_subscribe(rdata, code);
		}
	}
	return 0;
}

#endif

struct ast_taskprocessor *ast_sip_create_serializer_group(const char *name, struct ast_serializer_shutdown_group *shutdown_group)
{
	return ast_threadpool_serializer_group(name, sip_threadpool, shutdown_group);
}

struct ast_taskprocessor *ast_sip_create_serializer(const char *name)
{
	return ast_sip_create_serializer_group(name, NULL);
}

/*!
 * \internal
 * \brief Shutdown the serializers in the default pool.
 * \since 14.0.0
 *
 * \return Nothing
 */
static void serializer_pool_shutdown(void)
{
	int idx;

	for (idx = 0; idx < SERIALIZER_POOL_SIZE; ++idx) {
		ast_taskprocessor_unreference(serializer_pool[idx]);
		serializer_pool[idx] = NULL;
	}
}

/*!
 * \internal
 * \brief Setup the serializers in the default pool.
 * \since 14.0.0
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int serializer_pool_setup(void)
{
	char tps_name[AST_TASKPROCESSOR_MAX_NAME + 1];
	int idx;

	for (idx = 0; idx < SERIALIZER_POOL_SIZE; ++idx) {
		/* Create name with seq number appended. */
		ast_taskprocessor_build_name(tps_name, sizeof(tps_name), "pjsip/default");

		serializer_pool[idx] = ast_sip_create_serializer(tps_name);
		if (!serializer_pool[idx]) {
			serializer_pool_shutdown();
			return -1;
		}
	}
	return 0;
}

static struct ast_taskprocessor *serializer_pool_pick(void)
{
	int idx;
	int pos = 0;

	if (!serializer_pool[0]) {
		return NULL;
	}

	for (idx = 1; idx < SERIALIZER_POOL_SIZE; ++idx) {
		if (ast_taskprocessor_size(serializer_pool[idx]) < ast_taskprocessor_size(serializer_pool[pos])) {
			pos = idx;
		}
	}

	return serializer_pool[pos];
}

int ast_sip_push_task(struct ast_taskprocessor *serializer, int (*sip_task)(void *), void *task_data)
{
	if (!serializer) {
		serializer = serializer_pool_pick();
	}

	return ast_taskprocessor_push(serializer, sip_task, task_data);
}

struct sync_task_data {
	ast_mutex_t lock;
	ast_cond_t cond;
	int complete;
	int fail;
	int (*task)(void *);
	void *task_data;
};

static int sync_task(void *data)
{
	struct sync_task_data *std = data;
	int ret;

	std->fail = std->task(std->task_data);

	/*
	 * Once we unlock std->lock after signaling, we cannot access
	 * std again.  The thread waiting within ast_sip_push_task_wait()
	 * is free to continue and release its local variable (std).
	 */
	ast_mutex_lock(&std->lock);
	std->complete = 1;
	ast_cond_signal(&std->cond);
	ret = std->fail;
	ast_mutex_unlock(&std->lock);
	return ret;
}

static int ast_sip_push_task_wait(struct ast_taskprocessor *serializer, int (*sip_task)(void *), void *task_data)
{
	/* This method is an onion */
	struct sync_task_data std;

	memset(&std, 0, sizeof(std));
	ast_mutex_init(&std.lock);
	ast_cond_init(&std.cond, NULL);
	std.task = sip_task;
	std.task_data = task_data;

	if (ast_sip_push_task(serializer, sync_task, &std)) {
		ast_mutex_destroy(&std.lock);
		ast_cond_destroy(&std.cond);
		return -1;
	}

	ast_mutex_lock(&std.lock);
	while (!std.complete) {
		ast_cond_wait(&std.cond, &std.lock);
	}
	ast_mutex_unlock(&std.lock);

	ast_mutex_destroy(&std.lock);
	ast_cond_destroy(&std.cond);
	return std.fail;
}

int ast_sip_push_task_wait_servant(struct ast_taskprocessor *serializer, int (*sip_task)(void *), void *task_data)
{
	if (ast_sip_thread_is_servant()) {
		return sip_task(task_data);
	}

	return ast_sip_push_task_wait(serializer, sip_task, task_data);
}

int ast_sip_push_task_synchronous(struct ast_taskprocessor *serializer, int (*sip_task)(void *), void *task_data)
{
	return ast_sip_push_task_wait_servant(serializer, sip_task, task_data);
}

int ast_sip_push_task_wait_serializer(struct ast_taskprocessor *serializer, int (*sip_task)(void *), void *task_data)
{
	if (!serializer) {
		/* Caller doesn't care which PJSIP serializer the task executes under. */
		serializer = serializer_pool_pick();
		if (!serializer) {
			/* No serializer picked to execute the task */
			return -1;
		}
	}
	if (ast_taskprocessor_is_task(serializer)) {
		/*
		 * We are the requested serializer so we must execute
		 * the task now or deadlock waiting on ourself to
		 * execute it.
		 */
		return sip_task(task_data);
	}

	return ast_sip_push_task_wait(serializer, sip_task, task_data);
}

void ast_copy_pj_str(char *dest, const pj_str_t *src, size_t size)
{
	size_t chars_to_copy = MIN(size - 1, pj_strlen(src));
	memcpy(dest, pj_strbuf(src), chars_to_copy);
	dest[chars_to_copy] = '\0';
}

int ast_copy_pj_str2(char **dest, const pj_str_t *src)
{
	int res = ast_asprintf(dest, "%.*s", (int)pj_strlen(src), pj_strbuf(src));

	if (res < 0) {
		*dest = NULL;
	}

	return res;
}


int ast_sip_is_content_type(pjsip_media_type *content_type, char *type, char *subtype)
{
	pjsip_media_type compare;

	if (!content_type) {
		return 0;
	}

	pjsip_media_type_init2(&compare, type, subtype);

	return pjsip_media_type_cmp(content_type, &compare, 0) ? 0 : -1;
}

pj_caching_pool caching_pool;
pj_pool_t *memory_pool;
pj_thread_t *monitor_thread;
static int monitor_continue;
#ifdef GRANDSTREAM_NETWORKS
int timer_heap_support_ioqueue = 1;
#endif

static void *monitor_thread_exec(void *endpt)
{
#ifdef GRANDSTREAM_NETWORKS
	if (timer_heap_support_ioqueue) {
		pj_timer_heap_register_to_ioqueue(pjsip_endpt_get_timer_heap(ast_pjsip_endpoint), pjsip_endpt_get_ioqueue(ast_pjsip_endpoint));
		pj_timer_heap_set_servant_thread(pjsip_endpt_get_timer_heap(ast_pjsip_endpoint), monitor_thread);
	}
#endif

	while (monitor_continue) {
#ifndef GRANDSTREAM_NETWORKS
		const pj_time_val delay = {0, 10};
#else
		pj_time_val delay = {0, 10};
		if (timer_heap_support_ioqueue) {
			delay.sec = 10;
			delay.msec = 0;
		}
#endif
		pjsip_endpt_handle_events(ast_pjsip_endpoint, &delay);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (timer_heap_support_ioqueue) {
		pj_timer_heap_unregister_to_ioqueue(pjsip_endpt_get_timer_heap(ast_pjsip_endpoint));
		pj_timer_heap_set_servant_thread(pjsip_endpt_get_timer_heap(ast_pjsip_endpoint), NULL);
	}
#endif	
	return NULL;
}

static void stop_monitor_thread(void)
{
	monitor_continue = 0;
#ifdef GRANDSTREAM_NETWORKS
    if (timer_heap_support_ioqueue) {
        //pthread_kill(*(pthread_t *)pj_thread_get_os_handle(monitor_thread), SIGURG);
        pj_timer_heap_send_msg(pjsip_endpt_get_timer_heap(ast_pjsip_endpoint));
    }
#endif
	pj_thread_join(monitor_thread);
}

AST_THREADSTORAGE(pj_thread_storage);
AST_THREADSTORAGE(servant_id_storage);
#define SIP_SERVANT_ID 0x5E2F1D

static void sip_thread_start(void)
{
	pj_thread_desc *desc;
	pj_thread_t *thread;
	uint32_t *servant_id;

	servant_id = ast_threadstorage_get(&servant_id_storage, sizeof(*servant_id));
	if (!servant_id) {
		ast_log(LOG_ERROR, "Could not set SIP servant ID in thread-local storage.\n");
		return;
	}
	*servant_id = SIP_SERVANT_ID;

	desc = ast_threadstorage_get(&pj_thread_storage, sizeof(pj_thread_desc));
	if (!desc) {
		ast_log(LOG_ERROR, "Could not get thread desc from thread-local storage. Expect awful things to occur\n");
		return;
	}
	pj_bzero(*desc, sizeof(*desc));

	if (pj_thread_register("Asterisk Thread", *desc, &thread) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Couldn't register thread with PJLIB.\n");
	}
}

int ast_sip_thread_is_servant(void)
{
	uint32_t *servant_id;

	if (monitor_thread &&
			pthread_self() == *(pthread_t *)pj_thread_get_os_handle(monitor_thread)) {
		return 1;
	}

	servant_id = ast_threadstorage_get(&servant_id_storage, sizeof(*servant_id));
	if (!servant_id) {
		return 0;
	}

	return *servant_id == SIP_SERVANT_ID;
}

void *ast_sip_dict_get(void *ht, const char *key)
{
	unsigned int hval = 0;

	if (!ht) {
		return NULL;
	}

	return pj_hash_get(ht, key, PJ_HASH_KEY_STRING, &hval);
}

void *ast_sip_dict_set(pj_pool_t* pool, void *ht,
		       const char *key, void *val)
{
	if (!ht) {
#ifdef GRANDSTREAM_NETWORKS   /* Added by yfyang, for Bug 153949 */
		if (!val) {
			return NULL;
		}
#endif
		ht = pj_hash_create(pool, 11);
	}

	pj_hash_set(pool, ht, key, PJ_HASH_KEY_STRING, 0, val);

	return ht;
}

static pj_bool_t supplement_on_rx_request(pjsip_rx_data *rdata)
{
	struct ast_sip_supplement *supplement;

	if (pjsip_rdata_get_dlg(rdata)) {
		return PJ_FALSE;
	}

	AST_RWLIST_RDLOCK(&supplements);
	AST_LIST_TRAVERSE(&supplements, supplement, next) {
		if (supplement->incoming_request
			&& does_method_match(&rdata->msg_info.msg->line.req.method.name, supplement->method)) {
			struct ast_sip_endpoint *endpoint;

			endpoint = ast_pjsip_rdata_get_endpoint(rdata);
			supplement->incoming_request(endpoint, rdata);
			ao2_cleanup(endpoint);
		}
	}
	AST_RWLIST_UNLOCK(&supplements);

	return PJ_FALSE;
}

static void supplement_outgoing_response(pjsip_tx_data *tdata, struct ast_sip_endpoint *sip_endpoint)
{
	struct ast_sip_supplement *supplement;
	pjsip_cseq_hdr *cseq = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CSEQ, NULL);
	struct ast_sip_contact *contact = ast_sip_mod_data_get(tdata->mod_data, supplement_module.id, MOD_DATA_CONTACT);

	AST_RWLIST_RDLOCK(&supplements);
	AST_LIST_TRAVERSE(&supplements, supplement, next) {
		if (supplement->outgoing_response && does_method_match(&cseq->method.name, supplement->method)) {
			supplement->outgoing_response(sip_endpoint, contact, tdata);
		}
	}
	AST_RWLIST_UNLOCK(&supplements);

	ast_sip_mod_data_set(tdata->pool, tdata->mod_data, supplement_module.id, MOD_DATA_CONTACT, NULL);
	ao2_cleanup(contact);
}

int ast_sip_send_response(pjsip_response_addr *res_addr, pjsip_tx_data *tdata, struct ast_sip_endpoint *sip_endpoint)
{
	pj_status_t status;

	supplement_outgoing_response(tdata, sip_endpoint);
	status = pjsip_endpt_send_response(ast_sip_get_pjsip_endpoint(), res_addr, tdata, NULL, NULL);
	if (status != PJ_SUCCESS) {
		pjsip_tx_data_dec_ref(tdata);
	}

	return status == PJ_SUCCESS ? 0 : -1;
}

int ast_sip_send_stateful_response(pjsip_rx_data *rdata, pjsip_tx_data *tdata, struct ast_sip_endpoint *sip_endpoint)
{
	pjsip_transaction *tsx;

#ifdef GRANDSTREAM_NETWORKS
	if (pjsip_tsx_create_uas_and_lock(NULL, rdata, &tsx) != PJ_SUCCESS)
#else
	if (pjsip_tsx_create_uas(NULL, rdata, &tsx) != PJ_SUCCESS)
#endif
	{
		struct ast_sip_contact *contact;

		/* ast_sip_create_response bumps the refcount of the contact and adds it to the tdata.
		 * We'll leak that reference if we don't get rid of it here.
		 */
		contact = ast_sip_mod_data_get(tdata->mod_data, supplement_module.id, MOD_DATA_CONTACT);
		ao2_cleanup(contact);
		ast_sip_mod_data_set(tdata->pool, tdata->mod_data, supplement_module.id, MOD_DATA_CONTACT, NULL);
		pjsip_tx_data_dec_ref(tdata);
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	pj_grp_lock_add_ref(tsx->grp_lock); // Addec by pxie@grandstream.cn for bug[170997], 2020-07-06.
	/* Unlock transaction*/
	pj_grp_lock_release(tsx->grp_lock);
#endif

	pjsip_tsx_recv_msg(tsx, rdata);

	supplement_outgoing_response(tdata, sip_endpoint);

	if (pjsip_tsx_send_msg(tsx, tdata) != PJ_SUCCESS) {
#ifdef GRANDSTREAM_NETWORKS
		pj_grp_lock_dec_ref(tsx->grp_lock); // Addec by pxie@grandstream.cn for bug[170997], 2020-07-06.
#endif
		pjsip_tx_data_dec_ref(tdata);
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	pj_grp_lock_dec_ref(tsx->grp_lock); // Addec by pxie@grandstream.cn for bug[170997], 2020-07-06.
#endif

	return 0;
}

int ast_sip_create_response(const pjsip_rx_data *rdata, int st_code,
	struct ast_sip_contact *contact, pjsip_tx_data **tdata)
{
	int res = pjsip_endpt_create_response(ast_sip_get_pjsip_endpoint(), rdata, st_code, NULL, tdata);

	if (!res) {
		ast_sip_mod_data_set((*tdata)->pool, (*tdata)->mod_data, supplement_module.id, MOD_DATA_CONTACT, ao2_bump(contact));
	}

	return res;
}


#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, for Bug 156466 */
void ast_sip_inner_cid_to_outer_cid(char *inner_user, char *trunk_id, char *rt_id, char *num, int num_len, char *name, int name_len, int *skip_name)
{
	char  family[80]= {0};
	char  data[256] = {0};
	char  var[80]   = {0};
	char  outdod[256] = {0};
	const char* outcid = NULL;

	if (!inner_user || (!trunk_id && !rt_id) || !num || !num_len || !num || !num_len || !skip_name) {
		return;
	}

	num[0] = '\0';
	/* /TRUNK/SIP_5/OUTKEEPCID  : no */
	snprintf(family, sizeof(family), "TRUNK/SIP_%s", trunk_id);
	ast_ucm_db_get(family, "OUTKEEPCID", data, sizeof(data));
	if (!strcmp(data, "yes")) {
		/* OUT_CID_1 = tk7774<77774> */
		snprintf(var, sizeof(var), "OUT_CID_%s", trunk_id);
		outcid = pbx_builtin_getvar_helper(NULL, var);
		if (!ast_strlen_zero(outcid)) {
			ast_callerid_split(outcid, name, name_len, num, num_len);
			*skip_name = 0;
			ast_log(LOG_NOTICE, "inner_user[%s], trunk outcid data[%s], name[%s], num[%s]", inner_user, outcid, name, num);
			return;
		}

		return;   /*! not found or keep the trunk CID */
	}

	/* OUTDOD_1_9000 = rtrtrt1<r11111> */
	if (rt_id) {
		snprintf(outdod, sizeof(outdod), "OUTDOD_%s_%s", rt_id, inner_user);
		outcid = pbx_builtin_getvar_helper(NULL, outdod);
		if (!ast_strlen_zero(outcid)) {
			ast_callerid_split(outcid, name, name_len, num, num_len);
			*skip_name = 0;
			ast_log(LOG_NOTICE, "inner_user[%s], outrt dod data[%s], name[%s], num[%s]", inner_user, outcid, name, num);
			return;
		}
	}

	/*! get DOD of CF_USER: /TRUNK_5/DOD/6001 : DOD6001 */
	snprintf(family, sizeof(family), "TRUNK_%s/DOD", trunk_id);
	ast_ucm_db_get(family, inner_user, data, sizeof(data));
	if (strlen(data) != 0) {
		ast_callerid_split(data, name, name_len, num, num_len);
		*skip_name = 0;
		if (!strlen(num) && strlen(name)) {
			strncpy(num, name, num_len);
			memset(name, 0, name_len);
		}
		ast_log(LOG_NOTICE, "inner_user[%s], dod data[%s], name[%s], num[%s]", inner_user, data, name, num);
		return;
	}

	/*! get outbound cid: OBRT_CID_2 = 889977 */
	if (rt_id) {
		snprintf(outdod, sizeof(outdod), "OBRT_CID_%s", rt_id);
		outcid = pbx_builtin_getvar_helper(NULL, outdod);
		if (!ast_strlen_zero(outcid)) {
			ast_callerid_split(outcid, name, name_len, num, num_len);
			*skip_name = 0;
			ast_log(LOG_NOTICE, "rt_id[%s], outrt cid[%s], name[%s], num[%s]", rt_id, outcid, name, num);
			return;
		}
	}

	/*! get extension CID of the CF_USER:  /AAUSER/6001/OUTBOUNDCID  : John knight<user6001> */
	snprintf(family, sizeof(family), "AAUSER/%s", inner_user);
	ast_ucm_db_get(family, "OUTBOUNDCID", data, sizeof(data));
	if (strlen(data) != 0) {
		ast_callerid_split(data, name, name_len, num, num_len);
		*skip_name = 0;
		ast_log(LOG_NOTICE, "inner_user[%s], AAUSER/OUTBOUNDCID[%s], name[%s], num[%s]", inner_user, data, name, num);
		return;
	}

	/*! get trunk user name:  /TRUNK/SIP_5/USER_NAME  : 5005  */
	snprintf(family, sizeof(family), "TRUNK/SIP_%s", trunk_id);
	ast_ucm_db_get(family, "USER_NAME", data, sizeof(data));
	if (strlen(data) != 0) {
		memcpy(num, data, num_len);
		ast_log(LOG_NOTICE, "TRUNK/SIP_%s/USER_NAME=[%s], name[%s], num[%s]", trunk_id, data, name, num);
		return;
	}

	/*! get trunk user name:  /TRUNK/GLOBAL/OUTBOUNDCID: 9999  */
	ast_ucm_db_get("TRUNK/GLOBAL", "OUTBOUNDCID", data, sizeof(data));
	if (strlen(data) != 0) {
		ast_callerid_split(data, name, name_len, num, num_len);
		*skip_name = 0;
		ast_log(LOG_NOTICE, "GLOBAL OUTBOUNDCID=[%s], name[%s], num[%s]", data, name, num);
		return;
	}

	return;
}

int ast_sip_cei_same_channel(const char* channel, const char* mychannel, const char *cid_number)
{
    // Sanity check
    if (channel == NULL || mychannel == NULL) {
        return -1;
    }

    // Exact match
    if (strcmp(channel, mychannel) == 0) {
        return 0;
    }

    // Invited call: Match extension name
    // channel: Local/1002-446300@ccm-invite-00000001;2 (by CCM)
	//          Local/1002@from-internal-00000001;1 (by WebUI)
    // mychannel: PJSIP/1002-00000001
    if (strstr(channel, "Local") != NULL && strstr(mychannel, "PJSIP") != NULL) {
		char *extension = ast_strdupa(channel + 6);
        char *pchannel = NULL;
        if (extension == NULL) {
            return -1;
        }

		// CCM invite
		if (strstr(channel, "ccm-invite") != NULL) {
			pchannel = strchr(extension, '-');
		} else if (strstr(channel, "from-internal") != NULL) {
			pchannel = strchr(extension, '@');
		}

		if (pchannel == NULL) {
			return -1;
		}

		*pchannel = '\0';

		/* Modify by bfdu for Bug 311912, 2024.8.14 */
		char my_extension[AST_MAX_EXTENSION] = { 0 };
		ast_sip_get_extension_from_channel_name(mychannel, my_extension, sizeof(my_extension));
		if (!strcasecmp(my_extension, extension) || (cid_number && strstr(mychannel, cid_number) != NULL)) {
			return 0;
		}
	}

    return -1;
}

struct sip_conference* ast_sip_cei_find_peer_conference(struct ast_sip_endpoint *endpoint,
                                                		const char *conference, const char *channel)
{
    struct sip_conference* conf = NULL;

    // Sanity check
    if (endpoint == NULL || conference == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return NULL;
    }

    AST_LIST_LOCK(&endpoint->conferences);
    AST_LIST_TRAVERSE(&endpoint->conferences, conf, entry) {
        if (strcmp(conf->conference, conference) == 0 && ((channel == NULL && conf->state == SIP_CONFERENCE_STATE_WAITSUB) || (channel && strcmp(conf->channelName, channel) == 0))) {
            AST_LIST_UNLOCK(&endpoint->conferences);
            ao2_ref(conf, +1);
            return conf;
        }
    }
    AST_LIST_UNLOCK(&endpoint->conferences);

    return NULL;
}

/*! Destroy given peer-related conference subscription*/
void ast_sip_clear_peer_conference(struct ast_sip_endpoint *endpoint,
                                  const char *channel)
{
	struct sip_conference *conference;

    // Sanity check
    if (endpoint == NULL || channel == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return;
    }

    AST_LIST_LOCK(&endpoint->conferences);
    AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->conferences, conference, entry) {
        if (strcmp(conference->channelName, channel) == 0) {
            AST_LIST_REMOVE_CURRENT(entry);
            ao2_cleanup(conference);
        }
    }
    AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&endpoint->conferences);
}
/* END: Added by achan, Grandstream */

void ast_sip_peer_conference_set_subscribe_state(struct ast_sip_endpoint *endpoint, const char *channel, enum sip_conference_state state)
{
	struct sip_conference *conference = NULL;

	// Sanity check
	if (endpoint == NULL || channel == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return;
	}

	AST_LIST_LOCK(&endpoint->conferences);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->conferences, conference, entry) {
		if (!ast_strlen_zero(conference->channelName) && strcmp(conference->channelName, channel) == 0) {
			conference->state = state;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&endpoint->conferences);
}

struct cei_internal_mcb_peer* ast_cei_find_int_mcb_peer(const char *local_conf_id,
                                                        const char *remote_conf_id)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  struct cei_internal_mcb_peer* mcb_peer = NULL;

  // Sanity check
  if (local_conf_id == NULL || remote_conf_id == NULL) {
      ast_log(LOG_WARNING, "Invalid arguments\n");
      return NULL;
  }

  AST_LIST_LOCK(&internal_mcb_peers);
  AST_LIST_TRAVERSE(&internal_mcb_peers, mcb_peer, entry) {
	if (strcmp(mcb_peer->local_conf_id, local_conf_id) == 0 &&
      strcmp(mcb_peer->remote_conf_id, remote_conf_id) == 0) {
      AST_LIST_UNLOCK(&internal_mcb_peers);
      return mcb_peer;
    }
  }
  AST_LIST_UNLOCK(&internal_mcb_peers);

  return NULL;
}

struct cei_internal_mcb_peer* ast_cei_find_int_mcb_peer_by_id(const char *mcb_peer_id)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  struct cei_internal_mcb_peer* mcb_peer = NULL;

  // Sanity check
  if (mcb_peer_id == NULL) {
      ast_log(LOG_WARNING, "Invalid arguments\n");
      return NULL;
  }

  AST_LIST_LOCK(&internal_mcb_peers);
  AST_LIST_TRAVERSE(&internal_mcb_peers, mcb_peer, entry) {
	if (strcmp(mcb_peer->ref_id, mcb_peer_id) == 0) {
      AST_LIST_UNLOCK(&internal_mcb_peers);
      return mcb_peer;
    }
  }
  AST_LIST_UNLOCK(&internal_mcb_peers);

  return NULL;
}


struct cei_internal_mcb_peer* ast_cei_add_int_mcb_peer(const char *local_conf, const char *local_conf_id,
                                                       const char *remote_conf, const char *remote_conf_id,
                                                       const char *channelName)
{
  ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  struct cei_internal_mcb_peer* mcb_peer = NULL;
  size_t len = -1;
  char unix_time[16];
  char random_num[3];

  // Sanity check
  if (local_conf == NULL || local_conf_id == NULL ||
      remote_conf == NULL || remote_conf_id == NULL || channelName == NULL) {
    ast_log(LOG_WARNING, "Invalid arguments\n");
    return NULL;
  }

  mcb_peer = ast_cei_find_int_mcb_peer(local_conf_id, remote_conf_id);
  if (mcb_peer != NULL) {
    ast_log(LOG_WARNING, "%s already subscribed to Conference %s\n",
                         remote_conf, local_conf);
    return NULL;
  }

	AST_LIST_LOCK(&internal_mcb_peers);

  mcb_peer = ast_calloc(1, sizeof(*mcb_peer));
  len = strlen(local_conf) + strlen(remote_conf);

  if (mcb_peer == NULL) {
    ast_log(LOG_WARNING, "Failed to allocate new external MCB peer\n");
    AST_LIST_UNLOCK(&internal_mcb_peers);
    return NULL;
  }

  if (ast_string_field_init(mcb_peer, len)) {
    ast_log(LOG_WARNING, "Failed to allocate new string field\n");
    ast_free(mcb_peer);
    AST_LIST_UNLOCK(&internal_mcb_peers);
    return NULL;
  }

  ast_string_field_set(mcb_peer, local_conf, local_conf);
  ast_string_field_set(mcb_peer, remote_conf, remote_conf);
  ast_string_field_set(mcb_peer, channelName, channelName);
  strncpy(mcb_peer->local_conf_id, local_conf_id, sizeof(mcb_peer->local_conf_id));
  strncpy(mcb_peer->remote_conf_id, remote_conf_id, sizeof(mcb_peer->remote_conf_id));

  // Generate peer hash
	snprintf(unix_time, sizeof(unix_time), "%u", (int) time(NULL));
	snprintf(random_num, sizeof(random_num), "%d", (rand() % 90) + 10);
	char to_hash[strlen(remote_conf) + strlen(unix_time) + strlen(random_num) + 1];
	strcpy(to_hash, remote_conf);
	strcpy(to_hash + strlen(to_hash), unix_time);
	strcpy(to_hash + strlen(to_hash), random_num);
	ast_sha1_hash(mcb_peer->ref_id, to_hash);

  AST_LIST_INSERT_TAIL(&internal_mcb_peers, mcb_peer, entry);

  AST_LIST_UNLOCK(&internal_mcb_peers);

  ast_log(LOG_DEBUG, "%s subscribed to Conference %s\n",
                       remote_conf, local_conf);

  return mcb_peer;
}

int ast_cei_remove_int_mcb_peer(const char *local_conf_id, const char *remote_conf_id)
{
	ast_log(LOG_DEBUG, "[TK] Enter...%s\n", __FUNCTION__);
  struct cei_internal_mcb_peer* mcb_peer = NULL;

  // Sanity check
  if (local_conf_id == NULL || remote_conf_id == NULL) {
      ast_log(LOG_WARNING, "Invalid arguments\n");
      return -1;
  }

  // Remove MCB peer
  AST_LIST_LOCK(&internal_mcb_peers);
  AST_LIST_TRAVERSE_SAFE_BEGIN(&internal_mcb_peers, mcb_peer, entry) {
    if (strcmp(mcb_peer->local_conf_id, local_conf_id) == 0 &&
        strcmp(mcb_peer->remote_conf_id, remote_conf_id) == 0) {
			AST_LIST_REMOVE_CURRENT(entry);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
  AST_LIST_UNLOCK(&internal_mcb_peers);

  // Sanity check
  if (mcb_peer == NULL) {
    ast_log(LOG_WARNING, "Failed to get a subscription of local (%s) and remote (%s)\n",
                         local_conf_id, remote_conf_id);
    return -1;
  }

  ast_log(LOG_DEBUG, "%s unsubscribed from Conference %s\n",
                       mcb_peer->remote_conf, mcb_peer->local_conf);

  // Clean up
  ast_string_field_free_memory(mcb_peer);
  ast_free(mcb_peer);

  return 0;
}

struct cei_external_mcb_peer* ast_cei_find_external_mcb_peer(struct ast_sip_endpoint *endpoint,
							     const char *local_conf,
							     const char *remote_conf,
							     const char *aor,
							     const char *ip_addr)
{
    struct cei_external_mcb_peer* conf = NULL;

    // Sanity check
    if (endpoint == NULL || local_conf == NULL || remote_conf == NULL ||
		aor == NULL || ip_addr == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return NULL;
    }

    AST_LIST_LOCK(&endpoint->external_mcb_peers);
    AST_LIST_TRAVERSE(&endpoint->external_mcb_peers, conf, entry) {
		if (strcmp(conf->local_conf, local_conf) == 0 &&
            strcmp(conf->conf, remote_conf) == 0 &&
			strcmp(conf->endpoint, aor) == 0 &&
			strcmp(conf->ip_addr, ip_addr) == 0) {
            AST_LIST_UNLOCK(&endpoint->external_mcb_peers);
            return conf;
        }
	}
    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

    return NULL;
}

struct cei_external_mcb_peer* ast_cei_find_external_mcb_peer_by_id(struct ast_sip_endpoint *endpoint,
								   char *local_conf_id,
								   char *remote_conf_id)
{
    struct cei_external_mcb_peer* conf = NULL;

    // Sanity check
    if (endpoint == NULL || local_conf_id == NULL || remote_conf_id == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return NULL;
    }

    AST_LIST_LOCK(&endpoint->external_mcb_peers);
    AST_LIST_TRAVERSE(&endpoint->external_mcb_peers, conf, entry) {
		if (strcmp(conf->local_conf_id, local_conf_id) == 0 &&
            strcmp(conf->remote_conf_id, remote_conf_id) == 0) {
            AST_LIST_UNLOCK(&endpoint->external_mcb_peers);
            return conf;
        }
	}
    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

    return NULL;
}

struct cei_external_mcb_peer* ast_cei_find_external_mcb_peer_by_ref_id(struct ast_sip_endpoint *endpoint,
								       const char *mcb_peer_id)
{
    struct cei_external_mcb_peer* conf = NULL;

    // Sanity check
    if (endpoint == NULL || mcb_peer_id == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return NULL;
    }

    AST_LIST_LOCK(&endpoint->external_mcb_peers);
    AST_LIST_TRAVERSE(&endpoint->external_mcb_peers, conf, entry) {
		if (strcmp(conf->ref_id, mcb_peer_id) == 0) {
            AST_LIST_UNLOCK(&endpoint->external_mcb_peers);
            return conf;
        }
	}
    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

    return NULL;
}

struct cei_external_mcb_peer* ast_cei_add_external_mcb_peer(struct ast_sip_endpoint *endpoint,
                                                            char *local_conf, char *local_conf_id,
                                                            char *remote_conf, char *remote_conf_id,
                                                            char *ip_addr, const char *channelName, const char *actual_local_conf)
{
    struct cei_external_mcb_peer* mcb_peer = NULL;
    size_t len = -1;
    char unix_time[16];
	char random_num[3];

    // Sanity check
    if (endpoint == NULL || ast_strlen_zero(endpoint->aors) ||
        local_conf == NULL || local_conf_id == NULL ||
        remote_conf == NULL || remote_conf_id == NULL || ip_addr == NULL ||
        channelName == NULL || actual_local_conf == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return NULL;
    }

    mcb_peer = ast_cei_find_external_mcb_peer(endpoint, local_conf, remote_conf,
					      endpoint->aors, ip_addr);
    if (mcb_peer != NULL) {
	    AST_LIST_LOCK(&endpoint->external_mcb_peers);

        // Handle duplicate SIP subscription
		if (mcb_peer->isCompletion == 1) {
			ast_log(LOG_WARNING, "%s@%s via %s already subscribed to Conference %s\n",
	                             remote_conf, ip_addr, endpoint->aors, local_conf);

            // Update remote conference information
            if (strlen(remote_conf_id) > 0) {
                strncpy(mcb_peer->remote_conf_id, remote_conf_id, sizeof(mcb_peer->remote_conf_id));
            }

            // Destroy existing outgoing SIP subscription
            ao2_cleanup(mcb_peer->sip_sub);
            mcb_peer->sip_sub = NULL;

            // Update channel
            ast_log(LOG_WARNING, "Channel is changed from %s to %s\n",
                                 S_OR(mcb_peer->channelName, "NULL"), S_OR(channelName, "NULL"));
            // [TK] On case II of Bug 49550, we will need to destory previous channel by sending
            // stasis message to app_confbridge. However, this process will be done by Asterisk
            // after 30 minutes.
            ast_string_field_set(mcb_peer, channelName, channelName);

            AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

			return mcb_peer;
		}

		if (!ast_strlen_zero(local_conf_id) && !ast_strlen_zero(remote_conf_id) &&
			!ast_strlen_zero(channelName)) {
			strncpy(mcb_peer->local_conf_id, local_conf_id, sizeof(mcb_peer->local_conf_id));
			strncpy(mcb_peer->remote_conf_id, remote_conf_id, sizeof(mcb_peer->remote_conf_id));
			ast_string_field_set(mcb_peer, channelName, channelName);
			mcb_peer->isCompletion = 1;
			ast_log(LOG_WARNING, "%s@%s via %s in Conference %s has complete data\n",
	                             remote_conf, ip_addr, endpoint->aors, local_conf);
		}

        AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

        return mcb_peer;
    }

	AST_LIST_LOCK(&endpoint->external_mcb_peers);

    mcb_peer = ast_calloc(1, sizeof(*mcb_peer));
    len = strlen(local_conf) + strlen(remote_conf);
    len += strlen(endpoint->aors) + strlen(ip_addr);
    len += strlen(channelName);

    if (mcb_peer == NULL) {
        ast_log(LOG_WARNING, "Failed to allocate new external MCB peer\n");
        AST_LIST_UNLOCK(&endpoint->external_mcb_peers);
        return NULL;
    }

    if (ast_string_field_init(mcb_peer, len)) {
        ast_log(LOG_WARNING, "Failed to allocate new string field\n");
        ast_free(mcb_peer);
        AST_LIST_UNLOCK(&endpoint->external_mcb_peers);
        return NULL;
    }

    ast_string_field_set(mcb_peer, local_conf, local_conf);
    ast_string_field_set(mcb_peer, conf, remote_conf);
	ast_string_field_set(mcb_peer, local_conf_private_num, local_conf);
    ast_string_field_set(mcb_peer, conf_private_num, remote_conf);
    ast_string_field_set(mcb_peer, endpoint, endpoint->aors);
    ast_string_field_set(mcb_peer, ip_addr, ip_addr);
    ast_string_field_set(mcb_peer, channelName, channelName);
    ast_string_field_set(mcb_peer, actual_local_conf, actual_local_conf);
    strncpy(mcb_peer->local_conf_id, local_conf_id, sizeof(mcb_peer->local_conf_id));
    strncpy(mcb_peer->remote_conf_id, remote_conf_id, sizeof(mcb_peer->remote_conf_id));
    mcb_peer->state = SIP_CONFERENCE_STATE_WAITSUB;
    mcb_peer->sip_sub = NULL;
    mcb_peer->isCompletion = 1;

    if (strlen(mcb_peer->local_conf_id) == 0 ||
        strlen(mcb_peer->remote_conf_id) == 0) {
        mcb_peer->isCompletion = 0;
    }

    // Generate peer hash
    snprintf(unix_time, sizeof(unix_time), "%u", (int) time(NULL));
    snprintf(random_num, sizeof(random_num), "%d", (rand() % 90) + 10);
    char to_hash[strlen(remote_conf) + strlen(unix_time) + strlen(random_num) + 1];
    strcpy(to_hash, remote_conf);
    strcpy(to_hash + strlen(to_hash), unix_time);
    strcpy(to_hash + strlen(to_hash), random_num);
    ast_sha1_hash(mcb_peer->ref_id, to_hash);

    AST_LIST_INSERT_TAIL(&endpoint->external_mcb_peers, mcb_peer, entry);

    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

    ast_log(LOG_DEBUG, "%s@%s via %s subscribed to Conference %s\n",
                         remote_conf, ip_addr, endpoint->aors, local_conf);

    return mcb_peer;
}

int ast_cei_remove_external_mcb_peer(struct ast_sip_endpoint *endpoint,
                                     const char *local_conf_id,
									 const char *remote_conf_id)
{
    struct cei_external_mcb_peer* mcb_peer = NULL;

    // Sanity check
    if (endpoint == NULL || local_conf_id == NULL || remote_conf_id == NULL) {
        ast_log(LOG_WARNING, "Invalid arguments\n");
        return -1;
    }

    // Remove MCB peer
    AST_LIST_LOCK(&endpoint->external_mcb_peers);
    AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->external_mcb_peers, mcb_peer, entry) {
        if (strcmp(mcb_peer->local_conf_id, local_conf_id) == 0 &&
            strcmp(mcb_peer->remote_conf_id, remote_conf_id) == 0) {
			AST_LIST_REMOVE_CURRENT(entry);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
    AST_LIST_UNLOCK(&endpoint->external_mcb_peers);

    // Sanity check
    if (mcb_peer == NULL) {
        ast_log(LOG_WARNING, "Failed to get a subscription of local (%s) and remote (%s)\n",
                             local_conf_id, remote_conf_id);
        return -1;
    }

    ast_log(LOG_DEBUG, "%s@%s via %s unsubscribed from Conference %s\n",
                         mcb_peer->conf, mcb_peer->ip_addr, mcb_peer->endpoint,
                         mcb_peer->local_conf);

    // Clean up
	ast_sip_terminated_subscribe(mcb_peer->sip_sub, PJ_FALSE);
	/*Here decrease ref before increase ref in cei_ext_mcb_sub*/
	ast_sip_change_tree_ref(mcb_peer->sip_sub, -1);
    ast_string_field_free_memory(mcb_peer);
    ast_free(mcb_peer);

    return 0;
}



 struct cei_external_trunk_conf_peer* ast_cei_find_external_trunk_conf_peer(struct ast_sip_endpoint *endpoint,
																		const char  *local_virtual_conf,
																		const char *remote_conf,
																		const char *aor,
																		const char *ip_addr)
 {
	 struct cei_external_trunk_conf_peer* conf = NULL;
	 // Sanity check
	 if (endpoint == NULL || remote_conf == NULL ||
		 aor == NULL || ip_addr == NULL || local_virtual_conf == NULL) {
		 ast_log(LOG_WARNING, "Invalid arguments\n");
		 return NULL;
	 }
 
	 AST_LIST_LOCK(&endpoint->external_tc_peers);
	 AST_LIST_TRAVERSE(&endpoint->external_tc_peers, conf, entry) {
		 if (strcmp(conf->remote_conf_num, remote_conf) == 0 &&
			 strcmp(conf->endpoint, aor) == 0 &&
			 strcmp(conf->local_virtual_conf_num, local_virtual_conf) == 0 &&
			 strcmp(conf->ip_addr, ip_addr) == 0) {
			 AST_LIST_UNLOCK(&endpoint->external_tc_peers);
			 return conf;
		 }
	 }
	 AST_LIST_UNLOCK(&endpoint->external_tc_peers);
 
	 return NULL;
 }

struct cei_external_trunk_conf_peer* ast_cei_find_external_trunk_conf_peer_by_ref_id(struct ast_sip_endpoint *endpoint,
									 const char *etc_peer_id)
{
	struct cei_external_trunk_conf_peer* conf = NULL;

	// Sanity check
	if (endpoint == NULL || etc_peer_id == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return NULL;
	}

	AST_LIST_LOCK(&endpoint->external_tc_peers);
	AST_LIST_TRAVERSE(&endpoint->external_tc_peers, conf, entry) {
		if (strcmp(conf->ref_id, etc_peer_id) == 0) {
			AST_LIST_UNLOCK(&endpoint->external_tc_peers);
			return conf;
		}
	}
	AST_LIST_UNLOCK(&endpoint->external_tc_peers);

	return NULL;
}

 struct cei_external_trunk_conf_peer* ast_cei_find_external_trunk_conf_peer_by_virtual_conf(struct ast_sip_endpoint *endpoint,
									  const char *local_virtual_conf_num)
 {
	 struct cei_external_trunk_conf_peer* conf = NULL;
 
	 // Sanity check
	 if (endpoint == NULL || local_virtual_conf_num == NULL) {
		 ast_log(LOG_WARNING, "Invalid arguments\n");
		 return NULL;
	 }
 
	 AST_LIST_LOCK(&endpoint->external_tc_peers);
	 AST_LIST_TRAVERSE(&endpoint->external_tc_peers, conf, entry) {
		 if (strcmp(conf->local_virtual_conf_num, local_virtual_conf_num) == 0) {
			 AST_LIST_UNLOCK(&endpoint->external_tc_peers);
			 return conf;
		 }
	 }
	 AST_LIST_UNLOCK(&endpoint->external_tc_peers);
 
	 return NULL;
 }

 struct cei_external_trunk_conf_peer* ast_cei_add_external_trunk_conf_peer(struct ast_sip_endpoint *endpoint,
													const char  *local_virtual_conf,
													 const char *remote_conf,
													 const char *ip_addr, const char *channelName)
{
	struct cei_external_trunk_conf_peer* etc_peer = NULL;
	size_t len = -1;
	char unix_time[16];
	char random_num[3];

	// Sanity check
	if (endpoint == NULL || ast_strlen_zero(endpoint->aors) ||
		remote_conf == NULL || ip_addr == NULL || local_virtual_conf == NULL ||
		channelName == NULL ) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return NULL;
	}

	etc_peer = ast_cei_find_external_trunk_conf_peer(endpoint,local_virtual_conf,remote_conf,
						endpoint->aors, ip_addr);
	 if (etc_peer != NULL) {
		AST_LIST_LOCK(&endpoint->external_tc_peers);
		// Handle duplicate SIP subscription
		if (etc_peer->isCompletion == 1) {
			ast_log(LOG_WARNING, "%s@%s via %s already subscribed to Conference\n",
							  remote_conf, ip_addr, endpoint->aors);
			AST_LIST_UNLOCK(&endpoint->external_tc_peers);
			etc_peer->local_active_user = etc_peer->local_active_user +1;
			ast_log(LOG_DEBUG, "[WL] %d local users  join to the %s remote conf through %s \n",etc_peer->local_active_user,remote_conf,ast_sorcery_object_get_id(endpoint));

			return NULL;
		}

		if ( !ast_strlen_zero(channelName)) {
		ast_string_field_set(etc_peer, channelName, channelName);
		etc_peer->isCompletion = 1;
		ast_log(LOG_WARNING, "%s@%s via %s in Conference  has complete data\n",
						  remote_conf, ip_addr, endpoint->aors);
		}

		etc_peer->local_active_user = etc_peer->local_active_user +1;
		ast_log(LOG_DEBUG, "[WL] %d local users  join to the %s remote conf through %s \n",etc_peer->local_active_user,remote_conf,ast_sorcery_object_get_id(endpoint));

		AST_LIST_UNLOCK(&endpoint->external_tc_peers);
		return etc_peer;
	}

	 AST_LIST_LOCK(&endpoint->external_tc_peers);

	etc_peer = ast_calloc(1, sizeof(*etc_peer));
	len = strlen(remote_conf);
	len += strlen(endpoint->aors) + strlen(ip_addr);
	len += strlen(channelName);

	if (etc_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to allocate new external trunk conf peer\n");
		AST_LIST_UNLOCK(&endpoint->external_tc_peers);
		return NULL;
	}

	if (ast_string_field_init(etc_peer, len)) {
		ast_log(LOG_WARNING, "Failed to allocate new string field\n");
		ast_free(etc_peer);
		AST_LIST_UNLOCK(&endpoint->external_tc_peers);
		return NULL;
	}

	ast_string_field_set(etc_peer, remote_conf_num, remote_conf);
	ast_string_field_set(etc_peer, local_virtual_conf_num, local_virtual_conf);
	ast_string_field_set(etc_peer, endpoint, endpoint->aors);
	ast_string_field_set(etc_peer, ip_addr, ip_addr);
	ast_string_field_set(etc_peer, channelName, channelName);
	etc_peer->state = SIP_CONFERENCE_STATE_WAITSUB;
	etc_peer->sip_sub = NULL;
	etc_peer->isCompletion = 1;

	etc_peer->local_active_user = etc_peer->local_active_user +1;

	// Generate peer hash
	snprintf(unix_time, sizeof(unix_time), "%u", (int) time(NULL));
	snprintf(random_num, sizeof(random_num), "%d", (rand() % 90) + 10);
	char to_hash[strlen(remote_conf) + strlen(unix_time) + strlen(random_num) + 1];
	strcpy(to_hash, remote_conf);
	strcpy(to_hash + strlen(to_hash), unix_time);
	strcpy(to_hash + strlen(to_hash), random_num);
	ast_sha1_hash(etc_peer->ref_id, to_hash);

	AST_LIST_INSERT_TAIL(&endpoint->external_tc_peers, etc_peer, entry);

	AST_LIST_UNLOCK(&endpoint->external_tc_peers);
	ast_log(LOG_DEBUG, "[WL] %d local users  join to the %s remote conf through %s \n",etc_peer->local_active_user,remote_conf,ast_sorcery_object_get_id(endpoint));

	ast_log(LOG_DEBUG, "%s@%s via %s subscribed to Conference,%d local active users \n",
					  remote_conf, ip_addr, endpoint->aors,etc_peer->local_active_user);
	return etc_peer;
}

int ast_cei_remove_external_trunk_conf_peer(struct ast_sip_endpoint *endpoint,
								  const char *local_virtual_conf)
{
	struct cei_external_trunk_conf_peer* etc_peer = NULL;

	// Sanity check
	if (endpoint == NULL || local_virtual_conf == NULL ) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return -1;
	}

	// Remove etc peer
	AST_LIST_LOCK(&endpoint->external_tc_peers);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->external_tc_peers, etc_peer, entry) {
		if (strcmp(etc_peer->local_virtual_conf_num, local_virtual_conf) == 0 ) {
			AST_LIST_REMOVE_CURRENT(entry);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&endpoint->external_tc_peers);

	// Sanity check
	if (etc_peer == NULL) {
		ast_log(LOG_WARNING, "Failed to get a subscription of local (%s) \n",
						  local_virtual_conf);
		return -1;
	}

	ast_log(LOG_DEBUG, "%s@%s via %s unsubscribed from Conference\n",
			etc_peer->remote_conf_num, etc_peer->ip_addr, etc_peer->endpoint);

	
	// Clean up,use statis msg to clean, Bug 276133,added by zyxia 2023.11.9
	RAII_VAR(struct stasis_message *, msg, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json_payload *, payload, NULL, ao2_cleanup);
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);

	json_object = ast_json_pack("{s: s}",
		"local_virtual_conf_num", local_virtual_conf
	);

	payload = ast_json_payload_create(json_object);
	if (NULL == payload) {
		goto failed;
	}

	msg = stasis_message_create(ast_cei_etc_remove_type(), payload);
	if (msg == NULL) {
		goto failed;
	}
	stasis_publish(ast_cei_etc_process_topic_all(), msg);
failed:
	ast_sip_terminated_subscribe(etc_peer->sip_sub, PJ_FALSE);
	/*Here decrease ref before increase ref in cei_ext_mcb_sub*/
	ast_sip_change_tree_ref(etc_peer->sip_sub, -1);
	ast_string_field_free_memory(etc_peer);
	ast_free(etc_peer);

	return 0;
}

/* Get trunk informations by extension number.
 * @para trunk_name : IN
 * @para trunk_dir  : IN
 * @para ext        : IN, the extension number
 * @para prepend    : OUT, prepend string.
 * @para strip      : OUT, the number to be striped of digit.
 * @return          : 0-for success, -1 for false.
 */
int ast_cei_get_trunk_info_by_ext(const char *trunk_name,
							 const char *trunk_dir,
							 const char *ext,
							 char *prepend,
							 int  *strip)
{

	struct ast_exten       *pstExten = NULL;
	struct pbx_find_info    stFindInfo = { .stacklen = 0  }; /* the rest is reset in pbx_find_extension */
    char *trunk_number = NULL;
	char   family[80]  = {0};
	char   data[256]   = {0};
	char   *buf			= data;
	char   *arr[3] = {0};
	unsigned int  i = 0;

	if (NULL == trunk_name || NULL == trunk_dir || NULL == ext) {
		return -1;
	}

    trunk_number = strstr(trunk_name, "trunk_");
	if (trunk_number == NULL || ast_strlen_zero(trunk_number + 6)) {
		return -1;
    }
    trunk_number = trunk_number + 6;

	if (strcmp(trunk_dir, "outgoing") == 0) {
		pstExten = pbx_find_extension(NULL, NULL, &stFindInfo, "from-internal", ext, 1, NULL, NULL, E_MATCH);
		if (NULL == pstExten) {
			return -1;
		}

		/* get outbound route data from ucmdb */
		// snprintf(family, sizeof(family), "TRUNK/%s", ast_get_extension_name(pstExten));
		// ast_ucm_db_get(family, "EVENTLIST", data, sizeof(data));

		char function[128] = {0};
		char out_value[128] = {0};
		char *data_arr[4] = {0}; /* context,match,data, trunk_group */
		snprintf(function, sizeof(function), "MATCH_CALLERID(%s,)", ext);
		if (ast_func_read(NULL, function, out_value, sizeof(out_value))) {
			return -1;
		}

		if (out_value[0] == '\0') {
			return -1;
		}

		char *temp_buf = out_value;
		char *saveptr = NULL;
		while((i < 4) && ((data_arr[i] = strtok_r(temp_buf,",",&saveptr)) != NULL)) {
			i++;
			temp_buf = NULL;
		}
		i = 0;

		snprintf(data, sizeof(data), "%s", data_arr[2] ? : "");
		if (!strlen(data)) {
			return -1;
		}

		while((i < 3) && ((arr[i]=strtok_r(buf,"_",&saveptr)) != NULL)) {
			i++;
			buf = NULL;
		}

		if (strcmp(trunk_number, arr[0])) {
			return -1;
		}

		*strip = atoi(arr[1]);

		if (strcmp(arr[2], "a")) {
			snprintf(prepend, MAX_PREPEND_LEN, "%s", arr[2]);
		}

		return 0;

	} else if(strcmp(trunk_dir, "incoming") == 0) {
		snprintf(family, sizeof(family), "TRUNK/SIP_%s", trunk_number);
		ast_ucm_db_get(family, "INBOUND_PREPEND", data, sizeof(data));
		if (ast_strlen_zero(data)) {
			ast_log(LOG_WARNING, "No prepend: Failed to get prepend of %s\n", family);
		} else {
			snprintf(prepend, MAX_PREPEND_LEN, "%s", data);
		}

		/*get inbound strip*/
		memset(data, 0, 256);
		ast_ucm_db_get(family, "INBOUND_STRIP", data, sizeof(data));
		if (ast_strlen_zero(data)) {
			ast_log(LOG_WARNING, "No strip: Failed to get strip of %s\n", family);
			return 0;
		}

		*strip = atoi(data);
		return 0;
	} else {
		ast_log(LOG_WARNING, "Invalid trunk direction");
		return -1;
	}
}

struct ast_sip_endpoint *ast_sip_channel_get_endpoint(struct ast_channel *channel)
{
	struct ast_sip_endpoint *endpoint = NULL;
	struct ast_sip_channel_pvt *channel_pvt = NULL;

	if (!channel || strcasecmp(ast_channel_tech(channel)->type, "PJSIP")) {
		return NULL;
	}

	ast_channel_lock(channel);
	channel_pvt = ast_channel_tech_pvt(channel);
	if (channel_pvt && channel_pvt->session && channel_pvt->session->endpoint) {
		endpoint = ao2_bump(channel_pvt->session->endpoint);
	}
	ast_channel_unlock(channel);

	return endpoint;
}

/*Start added by zyxia for Bug 261870,2023.6.16*/
int ast_sip_endpoint_is_have_h264_codec(struct ast_sip_endpoint *endpoint)
{
	if(!endpoint)
		return 0;

	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return 0;
	}
	ast_log(LOG_DEBUG, "ast_sip_endpoint_is_have_video_codec endpoint:%s\n",
		ast_sorcery_object_get_id(endpoint));

	//check local endpoint if have video codec
	ast_format_cap_append_from_cap(caps, endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);
	if (ast_format_cap_count(caps) == 0 || !ast_format_cap_has_h264(caps)) {
		ast_log(LOG_WARNING, "Has no h264 video codec  endpoint:%s\n",
			 ast_sorcery_object_get_id(endpoint));
		return 0;
	}

	return  1;
}
/*End added by zyxia for Bug 261870,2023.6.16*/

int ast_sip_add_waveapp_info(struct ast_sip_endpoint *endpoint, const char *uniqueId,
const char* uri, int conf_invite,struct timeval register_time ,int update_new_uri,int del_db)
{
	struct sip_waveapp_info* app_info = NULL;
	size_t len = -1;


	if (endpoint == NULL || ast_strlen_zero(uniqueId) || ast_strlen_zero(uri))
	{
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return -1;
	}

	/*Bug 314611 ,if the enable_duplicate_logins is off,need force only one app login*/
	if(!endpoint->enable_duplicate_logins){
		ast_sip_clearall_waveapp_info(endpoint);
		if(del_db)
			ast_db_deltree(GSWAVE_DB_FAMILY, (char *)endpoint->aors);
	}

	app_info = ast_sip_find_waveapp_info(endpoint, uniqueId);
	if (app_info != NULL) {
		/*update it*/
		ast_string_field_set(app_info, gs_uri, uri);
		if(update_new_uri)
			ast_string_field_set(app_info, gs_new_contact_uri, uri);
		app_info->conf_invite = conf_invite;
		app_info->register_time = register_time;
		ast_log(LOG_DEBUG, "app clientId %s  already add to endpoint %s",
		uniqueId, ast_sorcery_object_get_id(endpoint));
		return -1;
	}

	AST_LIST_LOCK(&endpoint->waveapp_infos);
	app_info = ast_calloc(1, sizeof(*app_info));
	len = strlen(uniqueId) + strlen(uri)+strlen(uri);

	if (app_info == NULL) {
		ast_log(LOG_WARNING, "Failed to allocate new app_info\n");
		AST_LIST_UNLOCK(&endpoint->waveapp_infos);
		return -1;
	}
	if (ast_string_field_init(app_info, len)) {
		ast_log(LOG_WARNING, "Failed to allocate new string field\n");
		ast_free(app_info);
		AST_LIST_UNLOCK(&endpoint->waveapp_infos);
		return -1;
	}

	ast_string_field_set(app_info, client_uniqueId, uniqueId);
	ast_string_field_set(app_info, gs_uri, uri);
	if(update_new_uri)
		ast_string_field_set(app_info, gs_new_contact_uri, uri);

	app_info->conf_invite = conf_invite;
	app_info->register_time = register_time;

	AST_LIST_INSERT_TAIL(&endpoint->waveapp_infos, app_info, entry);

	AST_LIST_UNLOCK(&endpoint->waveapp_infos);

	ast_log(LOG_DEBUG, "Add app clientId %s to endpoint %s",
	uniqueId, ast_sorcery_object_get_id(endpoint));

	return 0;
}

int  ast_sip_reload_waveappinfo_from_astdb(int del_db) {
	struct ast_db_entry *tree, *entry;  
	ast_log(LOG_WARNING, "ast_sip_reload_waveappinfo_from_astdb ");
	char* extension = NULL;
	char *uniqueId = NULL;

	char* uri = NULL;
	char* register_time = NULL;
	char* conf_invite =NULL;

	char info_value[1024]={0};
	char info_key[1024]={0};
	char appid_key[32]={0},registime_key[32]={0},gs_uri_key[32]={0},conf_invite_key[32]={0};
	char appid_value[256]={0},registertime_value[64]={0},gs_uri_value[256]={0},conf_invite_value[16]={0};
	int conf_invite_s = 1;
	struct timeval timeval_struct; 
	const char *ptr = NULL;
	char pvalue[16] = {0};

	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	//get old data to new data
	ptr = ast_safe_nvram_get("o_appdata_save", pvalue, sizeof(pvalue));
	if (ptr == NULL) {
		if (!(tree = ast_db_gettree(GSWAVE_DB_FAMILY, "gs_appid"))) {
			ast_log(LOG_DEBUG, "Dont have  old app info\n");

			goto NEW_DATA;
		}

		//o_appdata_save have not set ,old app data need convert to new data to astdb
		for (entry = tree; entry; entry = entry->next) {  
			const char *key = entry->key;
			const char *delimiter = "/";
			char *saveptr = NULL;

			ast_safe_nvram_set("o_appdata_save","1");

			strtok_r((char *)key, delimiter, &saveptr);
			strtok_r(NULL, delimiter, &saveptr);
			extension = strtok_r(NULL, delimiter, &saveptr);

			snprintf(appid_key,sizeof(appid_key)-1,"%s/%s","gs_appid",extension);
			ast_db_get(GSWAVE_DB_FAMILY, appid_key, appid_value, sizeof(appid_value)-1);

			//exist old data
			if(!ast_strlen_zero(appid_value)){
				ast_log(LOG_DEBUG, "reload old app info key:%s\n",appid_key);
				ast_db_del(GSWAVE_DB_FAMILY, appid_key);

				memset(gs_uri_key,0,sizeof(gs_uri_key));
				memset(gs_uri_value,0,sizeof(gs_uri_value));
				snprintf(gs_uri_key,sizeof(gs_uri_key)-1,"%s/%s","gs_uri",extension);
				ast_db_get(GSWAVE_DB_FAMILY, gs_uri_key, gs_uri_value, sizeof(gs_uri_value)-1);
				ast_db_del(GSWAVE_DB_FAMILY, gs_uri_key);


				memset(registime_key,0,sizeof(registime_key));
				memset(registertime_value,0,sizeof(registertime_value));
				snprintf(registime_key,sizeof(registime_key)-1,"%s/%s","register_time",extension);
				ast_db_get(GSWAVE_DB_FAMILY, registime_key, registertime_value, sizeof(registertime_value)-1);
				ast_db_del(GSWAVE_DB_FAMILY, registime_key);


				memset(conf_invite_key,0,sizeof(conf_invite_key));
				memset(conf_invite_value,0,sizeof(conf_invite_value));
				snprintf(conf_invite_key,sizeof(conf_invite_key)-1,"%s/%s","conf_invite",extension);
				ast_db_get(GSWAVE_DB_FAMILY, conf_invite_key, conf_invite_value, sizeof(conf_invite_value)-1);
				ast_db_del(GSWAVE_DB_FAMILY, conf_invite_key);

				if(!ast_strlen_zero(conf_invite_value) && !strcmp(conf_invite_value,"no"))
					conf_invite_s = 0;

				/*added by zyxia for bug315083,2024.9.5*/
				if(ast_strlen_zero(gs_uri_value) || strstr(extension,"conf")){
					ast_log(LOG_DEBUG, "reload old app info key:%s gs_uri_value is NULL or the extension is null\n",appid_key);
					continue;
				}


				memset(info_key,0,sizeof(info_key));
				memset(info_value,0,sizeof(info_value));
				snprintf(info_key,sizeof(info_key),"%s/%s",extension,appid_value);
				snprintf(info_value, sizeof(info_value), "%s,%s,%d",gs_uri_value, registertime_value,conf_invite_s);
				ast_db_put(GSWAVE_DB_FAMILY, info_key, info_value);
			}

		}
		ast_db_freetree(tree);

	}

NEW_DATA:

	if (!(tree = ast_db_gettree(GSWAVE_DB_FAMILY, NULL))) {  
		return 0;
	}

	for (entry = tree; entry; entry = entry->next) {  
		 const char *key = entry->key;
		 const char *delimiter = "/";
		 ast_log(LOG_DEBUG, "ast_sip_reload_waveappinfo_from_astdb key:%s\n",key);

		 char *saveptr = NULL;
		 strtok_r((char *)key, delimiter, &saveptr);
		 extension = strtok_r(NULL, delimiter, &saveptr);
		 uniqueId =  strtok_r(NULL, delimiter, &saveptr);
		 ast_log(LOG_WARNING, "ast_sip_reload_waveappinfo_from_astdb :%s %s\n",extension,uniqueId);

		if(extension && uniqueId){

			endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", extension);
			if (!endpoint) {
				ast_log(LOG_WARNING, "endpoint does not exist:%s,delete the database  \n",extension);
				memset(info_key,0,sizeof(info_key));
				snprintf(info_key,sizeof(info_key)-1,"%s/%s",extension,uniqueId); //added by zyxia for 310163
				ast_db_del(GSWAVE_DB_FAMILY, info_key);
				continue;
			}

			memset(info_key,0,sizeof(info_key));
			snprintf(info_key,sizeof(info_key)-1,"%s/%s",GSWAVE_DB_FAMILY,extension);
			ast_db_get(info_key, uniqueId, info_value, sizeof(info_value)-1);
			if(!ast_strlen_zero(info_value)){

				saveptr = NULL;
				uri = strtok_r(info_value, ",", &saveptr);
				register_time = strtok_r(NULL, ",", &saveptr);
				conf_invite =  strtok_r(NULL, ",", &saveptr);

				if(register_time)
					timeval_struct.tv_sec = atoi(register_time);

				if(uri && register_time && conf_invite)
					ast_sip_add_waveapp_info(endpoint, uniqueId,  uri, atoi(conf_invite), timeval_struct, 0,del_db);
				ast_log(LOG_DEBUG, "ast_sip_reload_waveappinfo_from_astdb value: uri:%s time:%s,conf_invite:%s\n",uri,info_value,conf_invite);
			}

		}
	}

	ast_db_freetree(tree);
	return 1;
}


struct sip_waveapp_info* ast_sip_find_waveapp_info(struct ast_sip_endpoint *endpoint,const char *client_uniqueId)
{
	struct sip_waveapp_info* info = NULL;

	// Sanity check
	if (endpoint == NULL || ast_strlen_zero(client_uniqueId)) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return NULL;
	}

	AST_LIST_LOCK(&endpoint->waveapp_infos);
	AST_LIST_TRAVERSE(&endpoint->waveapp_infos, info, entry) {
		if (strcmp(info->client_uniqueId, client_uniqueId) == 0) {
			AST_LIST_UNLOCK(&endpoint->waveapp_infos);
			return info;
		}
	}
	AST_LIST_UNLOCK(&endpoint->waveapp_infos);

	return NULL;
}

/*added by zyxia for Bug 310016*/
int ast_sip_exist_waveapp_info(struct ast_sip_endpoint *endpoint)
{
	struct sip_waveapp_info* info = NULL;
	int exist = 0;
	// Sanity check
	if (endpoint == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return exist;
	}

	AST_LIST_LOCK(&endpoint->waveapp_infos);
	AST_LIST_TRAVERSE(&endpoint->waveapp_infos, info, entry) {
		exist = 1;
		break;
	}
	AST_LIST_UNLOCK(&endpoint->waveapp_infos);

	return exist;
}

struct sip_waveapp_info* ast_sip_find_waveapp_info_by_uri(struct ast_sip_endpoint *endpoint,const char *uri)
{
	struct sip_waveapp_info* info = NULL;

	// Sanity check
	if (endpoint == NULL || ast_strlen_zero(uri)) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return NULL;
	}

	AST_LIST_LOCK(&endpoint->waveapp_infos);
	AST_LIST_TRAVERSE(&endpoint->waveapp_infos, info, entry) {
		if (strstr(uri, info->gs_uri)) {
			AST_LIST_UNLOCK(&endpoint->waveapp_infos);
			return info;
		}
	}
	AST_LIST_UNLOCK(&endpoint->waveapp_infos);

	return NULL;
}

/*! Destroy given endport-related waveapp info*/
void ast_sip_clear_waveapp_info(struct ast_sip_endpoint *endpoint,const char *client_uniqueId)
{
	struct sip_waveapp_info* info = NULL;

	// Sanity check
	if (endpoint == NULL || client_uniqueId == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return;
	}

	AST_LIST_LOCK(&endpoint->waveapp_infos);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->waveapp_infos, info, entry) {
		if (strcmp(info->client_uniqueId, client_uniqueId) == 0) {
			AST_LIST_REMOVE_CURRENT(entry);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&endpoint->waveapp_infos);

	// Sanity check
	if (info == NULL) {
		ast_log(LOG_WARNING, "Failed to ast_sip_clear_waveapp_info client_uniqueId (%s)\n",
						client_uniqueId);
		return;
	}

	ast_string_field_free_memory(info);
	ast_free(info);
	return ;

}

/*! Destroy given endport-related waveapp info*/
void ast_sip_clearall_waveapp_info(struct ast_sip_endpoint *endpoint)
{
	struct sip_waveapp_info* info = NULL;

	// Sanity check
	if (endpoint == NULL) {
		ast_log(LOG_WARNING, "Invalid arguments\n");
		return;
	}

	AST_LIST_LOCK(&endpoint->waveapp_infos);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&endpoint->waveapp_infos, info, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_string_field_free_memory(info);
		ast_free(info);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&endpoint->waveapp_infos);
	return ;

}

void ast_sip_remove_waveinfo_from_database(const char *endpoint_name,struct sip_waveapp_info* app_info){

	char register_key[64]={0},gs_appid_key[64]={0},gs_uri_key[64]={0},gs_conf_invite_key[64]={0};
	char key[1024]={0};

	if(app_info == NULL || ast_strlen_zero(endpoint_name))
		return ;

	//rm new type wave data
	snprintf(key,sizeof(key),"%s/%s",endpoint_name,app_info->client_uniqueId);
	ast_db_del(GSWAVE_DB_FAMILY, key);

	//may exist old type wave data
	memset(gs_appid_key,0,sizeof(gs_appid_key));
	snprintf(gs_appid_key,sizeof(gs_appid_key),"%s/%s",endpoint_name,"gs_appid");
	
	memset(gs_uri_key,0,sizeof(gs_uri_key));
	snprintf(gs_uri_key,sizeof(gs_uri_key),"%s/%s",endpoint_name,"gs_uri");

	memset(register_key,0,sizeof(register_key));
	snprintf(register_key,sizeof(register_key),"%s/%s",endpoint_name,"register_time");

	memset(gs_conf_invite_key,0,sizeof(gs_conf_invite_key));
	snprintf(gs_conf_invite_key,sizeof(gs_conf_invite_key),"%s/%s",endpoint_name,"conf_invite");

	ast_db_del(GSWAVE_DB_FAMILY, gs_appid_key);
	ast_db_del(GSWAVE_DB_FAMILY, register_key);
	ast_db_del(GSWAVE_DB_FAMILY, gs_uri_key);
	ast_db_del(GSWAVE_DB_FAMILY, gs_conf_invite_key);

	return ;
}

#endif

int ast_sip_get_host_ip(int af, pj_sockaddr *addr)
{
	if (af == pj_AF_INET() && !ast_strlen_zero(host_ip_ipv4_string)) {
		pj_sockaddr_copy_addr(addr, &host_ip_ipv4);
		return 0;
	} else if (af == pj_AF_INET6() && !ast_strlen_zero(host_ip_ipv6_string)) {
		pj_sockaddr_copy_addr(addr, &host_ip_ipv6);
		return 0;
	}

	return -1;
}

const char *ast_sip_get_host_ip_string(int af)
{
	if (af == pj_AF_INET()) {
		return host_ip_ipv4_string;
	} else if (af == pj_AF_INET6()) {
		return host_ip_ipv6_string;
	}

	return NULL;
}

#ifdef GRANDSTREAM_NETWORKS
/* returns   1: network address was successfully converted
             0: is returned if src does not contain a character string representing a valid network address in the specified address family.
            -1: af does not contain a valid address family
*/
 int ast_sip_is_a_valid_IP_address(const char *str)
{
	struct in_addr addr;
	struct in6_addr addr6;
	char host[MAXHOSTNAMELEN] = {0};
	int i = 0;

	if (*str == '[') {
		++str;
		for (; *str && *str != ']';) {
			host[i++] = *str++;
		}
	}

	/* for ipv6 */
	if (inet_pton(AF_INET, str, &addr) < 1) {
		return inet_pton(AF_INET6, host, &addr6);
	}

	return 1;
}

 int ast_sip_sockaddr_from_pj_sockaddr(struct ast_sockaddr *addr, const pj_sockaddr *pjaddr)
 {
	 if (pjaddr->addr.sa_family == pj_AF_INET()) {
		 struct sockaddr_in *sin = (struct sockaddr_in *) &addr->ss;
		 sin->sin_family = AF_INET;
#ifdef HAVE_PJPROJECT_BUNDLED
		 sin->sin_addr = pjaddr->ipv4.sin_addr;
#else
		 sin->sin_addr.s_addr = pjaddr->ipv4.sin_addr.s_addr;
#endif
		 sin->sin_port	 = pjaddr->ipv4.sin_port;
		 addr->len = sizeof(struct sockaddr_in);
	 } else if (pjaddr->addr.sa_family == pj_AF_INET6()) {
		 struct sockaddr_in6 *sin = (struct sockaddr_in6 *) &addr->ss;
		 sin->sin6_family	= AF_INET6;
		 sin->sin6_port 	= pjaddr->ipv6.sin6_port;
		 sin->sin6_flowinfo = pjaddr->ipv6.sin6_flowinfo;
		 sin->sin6_scope_id = pjaddr->ipv6.sin6_scope_id;
		 memcpy(&sin->sin6_addr, &pjaddr->ipv6.sin6_addr, sizeof(sin->sin6_addr));
		 addr->len = sizeof(struct sockaddr_in6);
	 } else {
		 memset(addr, 0, sizeof(*addr));
		 return -1;
	 }
	 return 0;
 }

int ast_sip_check_host_ip(void)
{
    char host_str[32] = {0};
#ifdef BUILD_ON_CLOUD_PBX
    strcpy(host_str, "10.88.0.");
#else
#ifdef HAVE_GCC601X
    strcpy(host_str, "10.10.10.");
#else
    strcpy(host_str, "127.0.0.");
#endif
#endif
    if (!strncmp(local_ip_ipv4_string, host_str, strlen(host_str))) {
        if (!pj_gethostip(pj_AF_INET(), &host_ip_ipv4)) {
            pj_sockaddr_print(&host_ip_ipv4, local_ip_ipv4_string, sizeof(local_ip_ipv4_string), 0);
            if (HA_ROLE_ACTIVE == ast_ha_get_role() && (ast_get_redundance_flag() != 1)){
                host_ip_ipv4.ipv4.sin_addr.s_addr = get_ha_float_ip();
            }
            pj_sockaddr_print(&host_ip_ipv4, host_ip_ipv4_string, sizeof(host_ip_ipv4_string), 0);
            ast_verb(3, "Local IPv4 address determined to be: %s\n", host_ip_ipv4_string);
        }
        if (!pj_gethostip(pj_AF_INET6(), &host_ip_ipv6)) {
            pj_sockaddr_print(&host_ip_ipv6, host_ip_ipv6_string, sizeof(host_ip_ipv6_string), 0);
            ast_verb(3, "Local IPv6 address determined to be: %s\n", host_ip_ipv6_string);
        }
    }
    return 0;
}
const char *ast_sip_get_local_ip_string(int af)
{
	if (af == pj_AF_INET()) {
		return local_ip_ipv4_string;
	} else if (af == pj_AF_INET6()) {
		return host_ip_ipv6_string;
	}

	return NULL;
}


/*! \brief Helper function which adds a Date header to a response */
int ast_sip_add_timestamp_header(pjsip_tx_data *tdata)
{
	char ts[32];
	time_t t = time(NULL);

	snprintf(ts, sizeof(ts), "%d", (int)t);
	ast_sip_add_header(tdata, "Timestamp", ts);

	return 0;
}
#endif

#define _____HA_STANDBY_SWITCH_TO_ACTIVE_BEGIN_____

/* Added by yfyang, for HA service data sync */

static struct stasis_message_router *pjsip_ha_data_router = NULL;

static void pjsip_ha_role_switch_cb(void *data, struct stasis_subscription *sub, struct stasis_message *message)
{
	static int num=0;
	struct ast_json_payload *payload	 = NULL;
	struct ast_json			*json_object = NULL;
	int    role = 0;
	const unsigned int flags = 0; /* no port, no brackets */
	struct avs_set_float_ip_param para = {0};
	int floatip;
	struct in_addr ip;

	ast_log(LOG_NOTICE,"----->pjsip_ha_role_switch_cb seq=%d", num);

	if(!ast_get_redundance_flag() && num >= 1)
		return;
	/* Only do sync when switch to active */
	if (HA_ROLE_ACTIVE != ast_ha_get_role()) {
		return;
	}

	if (!message) {
		ast_log(LOG_WARNING, "Stasis message is NULL");
		return;
	}

	payload = stasis_message_data(message);
	if (!payload) {
		ast_log(LOG_WARNING, "Payload from stasis message is NULL");
		return;
	}

	json_object = payload->json;
	if (!json_object) {
		ast_log(LOG_WARNING, "JSON object from stasis message is NULL");
		return;
	}

	role = ast_json_integer_get(ast_json_object_get(json_object, "ha_role"));
	if (!role) {
		ast_log(LOG_WARNING, "Incomplete JSON object data"); 
		return;
	}

	ast_log(LOG_NOTICE,"pjsip_ha_role_switch_cb----->ast_sip_transport_update");

	if (ast_get_redundance_flag() != 1)
	{
		ast_sip_transport_update();
	}

	if (!pj_gethostip(pj_AF_INET(), &host_ip_ipv4)) {
		if (ast_get_redundance_flag() != 1)
		{
			host_ip_ipv4.ipv4.sin_addr.s_addr = get_ha_float_ip();
		}
		pj_sockaddr_print(&host_ip_ipv4, host_ip_ipv4_string, sizeof(host_ip_ipv4_string), flags);
		ast_log(LOG_NOTICE, "Local IPv4 address determined to be: %s\n", host_ip_ipv4_string);
	}
	

	if (!pj_gethostip(pj_AF_INET6(), &host_ip_ipv6)) {
		pj_sockaddr_print(&host_ip_ipv6, host_ip_ipv6_string, sizeof(host_ip_ipv6_string), flags);
		ast_log(LOG_NOTICE, "Local IPv6 address determined to be: %s\n", host_ip_ipv6_string);
	}

	floatip = get_ha_float_ip();
	if (-1 != floatip &&  (ast_get_redundance_flag() != 1))
	{
		ip.s_addr = floatip;
		char *ip_str = (char *)ast_inet_ntoa(ip);
		strcpy(para.ip_addr, ip_str);
		para.opt = 1;
		ast_avs_control(NULL, AVS_CONTROL_SET_HA_FIP, &para, sizeof(para));
		float_ip_to_avs = 1;
	}
	
	
	ast_log(LOG_NOTICE, "res_pjsip swtiched to Active.\n");
	num++;
	return;
}

static int pjsip_ha_stasis_route_init(void)
{
	struct stasis_topic * pjsip_ha_data_topic = ast_ha_sync_data_topic_all();
	if (!pjsip_ha_data_topic) {
		ast_log(LOG_WARNING, "Failed to get pjsip_ha_data_topic\n");
		return -1;
	}

	pjsip_ha_data_router = stasis_message_router_create(pjsip_ha_data_topic);
	if (pjsip_ha_data_router == NULL) {
		ast_log(LOG_WARNING, "Failed to create pjsip_ha_data_router\n");
		return -1;
	}

	if (stasis_message_router_add(pjsip_ha_data_router, ast_ha_role_switch_type(), pjsip_ha_role_switch_cb, NULL)) {
		ast_log(LOG_WARNING, "Failed to subscribe to ast_ha_role_switch_type\n");
		return -1;
	}

	return 0;
}

int get_ha_float_ip(void)
{
    FILE *ha_fp = NULL;
    char on_buf[64] = { 0 };
	int ip =-1;
	char *tmp=NULL;
        
    /* Get the ha_enable */
    ha_fp = popen("nvram get hafloatip", "r");
    if (ha_fp == NULL)
    {
        ast_log(LOG_ERROR, "hafloatip Read error");
        return -1;
    }
    while (ha_fp != NULL && (!feof(ha_fp)) && (fgets(on_buf, sizeof(on_buf) - 1, ha_fp) != NULL)) 
    {
    	/* delete control char */
		tmp = strchr(on_buf, 10);
		if(tmp)
		{
			*tmp = '\0';
		}

		inet_pton(AF_INET, on_buf, &ip);
		ast_log(LOG_DEBUG, "hafloatip Read out: [%s]. ip_addr=%d", on_buf, ip);
    }
	
    
    pclose(ha_fp);
    ha_fp = NULL;
    return ip;

}
int ast_sip_dtmf_to_str(const enum ast_sip_dtmf_mode dtmf,
		        char *buf, size_t buf_len)
{
	switch (dtmf) {
	case AST_SIP_DTMF_NONE:
		ast_copy_string(buf, "none", buf_len);
		break;
	case AST_SIP_DTMF_RFC_4733:
		ast_copy_string(buf, "rfc4733", buf_len);
		break;
	case AST_SIP_DTMF_INBAND:
		ast_copy_string(buf, "inband", buf_len);
		break;
	case AST_SIP_DTMF_INFO:
		ast_copy_string(buf, "info", buf_len);
		break;
	case AST_SIP_DTMF_AUTO:
		ast_copy_string(buf, "auto", buf_len);
		break;
	case AST_SIP_DTMF_AUTO_INFO:
		ast_copy_string(buf, "auto_info", buf_len);
		break;
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_DTMF_BOTH_RFC4733_AND_INFO:
		ast_copy_string(buf, "rfc4733_info", buf_len);
		break;
#endif
	default:
		buf[0] = '\0';
		return -1;
	}
	return 0;
}

int ast_sip_str_to_dtmf(const char * dtmf_mode)
{
	int result = -1;

	if (!strcasecmp(dtmf_mode, "info")) {
		result = AST_SIP_DTMF_INFO;
	} else if (!strcasecmp(dtmf_mode, "rfc4733")) {
		result = AST_SIP_DTMF_RFC_4733;
	} else if (!strcasecmp(dtmf_mode, "inband")) {
		result = AST_SIP_DTMF_INBAND;
	} else if (!strcasecmp(dtmf_mode, "none")) {
		result = AST_SIP_DTMF_NONE;
	} else if (!strcasecmp(dtmf_mode, "auto")) {
		result = AST_SIP_DTMF_AUTO;
	} else if (!strcasecmp(dtmf_mode, "auto_info")) {
		result = AST_SIP_DTMF_AUTO_INFO;
	}
#ifdef GRANDSTREAM_NETWORKS
	else if (!strcasecmp(dtmf_mode, "rfc4733_info")) {
		result = AST_SIP_DTMF_BOTH_RFC4733_AND_INFO;
	}
#endif

	return result;
}

/*!
 * \brief Set name and number information on an identity header.
 *
 * \param pool Memory pool to use for string duplication
 * \param id_hdr A From, P-Asserted-Identity, or Remote-Party-ID header to modify
 * \param id The identity information to apply to the header
 */
void ast_sip_modify_id_header(pj_pool_t *pool, pjsip_fromto_hdr *id_hdr, const struct ast_party_id *id)
{
	pjsip_name_addr *id_name_addr;
	pjsip_sip_uri *id_uri;

	id_name_addr = (pjsip_name_addr *) id_hdr->uri;
	id_uri = pjsip_uri_get_uri(id_name_addr->uri);

	if (id->name.valid) {
		if (!ast_strlen_zero(id->name.str)) {
			int name_buf_len = strlen(id->name.str) * 2 + 1;
			char *name_buf = ast_alloca(name_buf_len);

			ast_escape_quoted(id->name.str, name_buf, name_buf_len);
			pj_strdup2(pool, &id_name_addr->display, name_buf);
		} else {
			pj_strdup2(pool, &id_name_addr->display, NULL);
		}
	}

	if (id->number.valid) {
		pj_strdup2(pool, &id_uri->user, id->number.str);
	}
}


static void remove_request_headers(pjsip_endpoint *endpt)
{
	const pjsip_hdr *request_headers = pjsip_endpt_get_request_headers(endpt);
	pjsip_hdr *iter = request_headers->next;

	while (iter != request_headers) {
		pjsip_hdr *to_erase = iter;
		iter = iter->next;
		pj_list_erase(to_erase);
	}
}
#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, for Bug 295956 */
struct ao2_container *ast_sip_get_account_trunk_endpoints(void)
{
	struct ao2_container *endpoints;
	struct ast_variable fields = {
		.name = "trunk_type",
		.value = "account",
	};

	endpoints = ast_sorcery_retrieve_by_fields(ast_sip_get_sorcery(), "endpoint", AST_RETRIEVE_FLAG_MULTIPLE, &fields);
	return endpoints;
}


AST_MUTEX_DEFINE_STATIC(gs_feature_code_mutex);
/*! get feature code from  exten value in extensions.conf */
static void set_idle_feature_code(gs_idle_feature_enum type, const char* value)
{
	char  code[12] = {0};
	char* pTemp    = NULL;
	char *pStr     = (char *)value;
	int   iLen     = 0;

	if (type >= GS_IDLE_FEATURE_COUNT || NULL == value) {
		return;
	}

	pTemp = strchr(pStr, ',');
	if (NULL == pTemp) {
		return;
	}

	*pTemp = '\0';

	if ('_' == pStr[0]) {
		pStr++;
	}

	if ('s' == pStr[0]) {
		pStr++;
	}

	if (pStr >= pTemp) {
		return;
	}

	strncpy(code, pStr, sizeof(code)-1);
	iLen = strlen(code);

	if ('.' == code[iLen-1]) {
		code[iLen-1] = '\0';
		iLen--;
	}

	if (iLen <= 0) {
		return;
	}

	strncpy(gs_idle_feature_code[type].f_code, code, iLen);
	return;
}

/*! \brief get feature code map from features.conf, and get feature right from extensions.conf (DIAL_OPTIONS) */
static int get_feature_code(int reload)
{
	struct ast_flags config_flags = {reload ? CONFIG_FLAG_FILEUNCHANGED : 0};
	struct ast_config    *pstECfg = NULL;
	struct ast_variable  *pstVarGen = NULL;
	int    i;

	/* Added by yfyang, for Bug305432  */
	if (!reload) {
		struct ast_config *cfg = ast_config_load("pjsip_endpoints.conf", config_flags);  /* load pjsip_endpoints.conf while load module */
		if (cfg && (cfg != CONFIG_STATUS_FILEUNCHANGED) && (cfg != CONFIG_STATUS_FILEINVALID)) {
			ast_config_destroy(cfg);
		}
	} 

	/* load extensions.conf file*/
	pstECfg = ast_config_load("extensions.conf", config_flags);

	if (pstECfg == CONFIG_STATUS_FILEUNCHANGED || NULL == pstECfg) {
		return -1;
	}

	if (pstECfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_ERROR, "Contents of extensions.conf are invalid and cannot be parsed\n");
		return -1;
	}

	ast_mutex_lock(&gs_feature_code_mutex);
	for (i = 0; i < GS_IDLE_FEATURE_COUNT; i++) {
		memset(gs_idle_feature_code[i].f_code, 0, sizeof(gs_idle_feature_code[i].f_code));
	}

	/* get idle feature code */
	pstVarGen = ast_variable_browse(pstECfg, "app-dnd-on");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_DND_ON, pstVarGen->value);
	}

	pstVarGen = ast_variable_browse(pstECfg, "app-dnd-off");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_DND_OFF, pstVarGen->value);
	}

	/* FWD BUSY*/
	pstVarGen = ast_variable_browse(pstECfg, "app-cfb-on");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_FWD_BUSY_ON, pstVarGen->value);
	}

	pstVarGen = ast_variable_browse(pstECfg, "app-cfb-off");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_FWD_BUSY_OFF, pstVarGen->value);
	}

	/* FWD NO ANSWER */
	pstVarGen = ast_variable_browse(pstECfg, "app-cfn-on");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_FWD_NO_ANS_ON, pstVarGen->value);
	}

	pstVarGen = ast_variable_browse(pstECfg, "app-cfn-off");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_FWD_NO_ANS_OFF, pstVarGen->value);
	}

	/* FWD UNCONDITONAL */
	pstVarGen = ast_variable_browse(pstECfg, "app-cfu-on");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_FWD_UNCOND_ON, pstVarGen->value);
	}

	pstVarGen = ast_variable_browse(pstECfg, "app-cfu-off");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_FWD_UNCOND_OFF, pstVarGen->value);
	}

	/* AGENT PAUSE/UNPAUSE */
	pstVarGen = ast_variable_browse(pstECfg, "app-agent-pause");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_AGENT_PAUSE, pstVarGen->value);
	}

	pstVarGen = ast_variable_browse(pstECfg, "app-agent-unpause");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_AGENT_UNPAUSE, pstVarGen->value);
	}

	/* PAGEING PREFIX*/
	pstVarGen = ast_variable_browse(pstECfg, "app-paging");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_PAGING_PREFIX, pstVarGen->value);
	}

	/* INTERCOM PREFIX*/
	pstVarGen = ast_variable_browse(pstECfg, "app-intercom");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_INTERCOM_PREFIX, pstVarGen->value);
	}

	/* CALL PICKUP */
	pstVarGen = ast_variable_browse(pstECfg, "app-pickup");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_CALL_PICKUP, pstVarGen->value);
	}

	/* CALL BLACKLIST ADD/REMOVE */
	pstVarGen = ast_variable_browse(pstECfg, "app-blacklist-add-wrap");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_BLACKLIST_ADD, pstVarGen->value);
	}

	pstVarGen = ast_variable_browse(pstECfg, "app-blacklist-remove-wrap");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_BLACKLIST_REMOVE, pstVarGen->value);
	}

	/* VOICEMAIL_ACCESS_CODE */
	pstVarGen = ast_variable_browse(pstECfg, "app-dialvm");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_VOICEMAIL_ACCESS, pstVarGen->value);
	}

	/* MY_VOICEMAIL*/
	pstVarGen = ast_variable_browse(pstECfg, "app-vmmain");
	if (NULL != pstVarGen) {
		set_idle_feature_code(GS_FEATURE_MY_VOICEMAIL, pstVarGen->value);
	}

/*Start added by zyxia for callflip.2021.11.26*/
	/* CALL-FLIP*/
	pstVarGen = ast_variable_browse(pstECfg, "call-flip");
	if (NULL != pstVarGen) {
		enable_callflip = 1;
		//set_idle_feature_code(GS_FEATURE_CALLFLIP, pstVarGen->value);
	}
	else
		enable_callflip = 0;
/*End added by zyxia for callflip.2021.11.26*/

	ast_mutex_unlock(&gs_feature_code_mutex);
	ast_config_destroy(pstECfg);
	return 0;
}

static void mcm_room_destroy(void *obj)
{
	struct mcm_room *room = obj;

	ast_log(LOG_DEBUG, "Destroying the mcm room %s\n", room->number);
	ast_free(room->number);
}

static int mcm_room_hash_cb(const void *obj, const int flags)
{
	const struct mcm_room *room = obj;
	const char *number = obj;
	int hash;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
		default:
		case OBJ_POINTER:
			number = room->number;
			/* Fall through */
		case OBJ_KEY:
			hash = ast_str_case_hash(number);
			break;
		case OBJ_PARTIAL_KEY:
			/* Should never happen in hash callback. */
			ast_assert(0);
			hash = 0;
			break;
	}
	return hash;
}

static int mcm_room_cmp_cb(void *obj, void *arg, int flags)
{
	const struct mcm_room *left = obj;
	const struct mcm_room *right = arg;
	const char *right_number = arg;
	int cmp;

	switch (flags & (OBJ_POINTER | OBJ_KEY | OBJ_PARTIAL_KEY)) {
		default:
		case OBJ_POINTER:
			right_number = right->number;
			/* Fall through */
		case OBJ_KEY:
			cmp = strcasecmp(left->number, right_number);
			break;
		case OBJ_PARTIAL_KEY:
			cmp = strncasecmp(left->number, right_number, strlen(right_number));
			break;
	}
	return cmp ? 0 : CMP_MATCH;
}

static void create_mcm_conf_number(void)
{
	mcm_rooms = ao2_container_alloc(1, mcm_room_hash_cb, mcm_room_cmp_cb);
	if (!mcm_rooms) {
		ast_log(LOG_ERROR, "Fail to create mcm rooms.\n");
	}
	return;
}

static void destory_mcm_conf_number(void)
{
	if (mcm_rooms) {
		ao2_cleanup(mcm_rooms);
		mcm_rooms = NULL;
	}

	return;
}

static int prase_mcm_conference_config(int reload)
{
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct ast_config *cfg = NULL;
	char *cat = NULL;
	char conf_number[MAX_MCM_CONF_LEN] = {0};
	struct mcm_room *room = NULL;

	if (!mcm_rooms) {
		ast_log(LOG_WARNING, "prase_mcm_conference_config mcm_outbound_prefix is NULL\n");
		return 0;
	}

	cfg = ast_config_load2(MCM_CONF, "pjsip", config_flags);
	if (cfg == CONFIG_STATUS_FILEMISSING) {
		ast_log(LOG_DEBUG, MCM_CONF " file missing!");
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEUNCHANGED) {
		ast_log(LOG_DEBUG, MCM_CONF " file unchanged!");
		return 0;
	} else if (cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_DEBUG, MCM_CONF " file invalid!");
		return 0;
	}

	if (cfg) {
		ao2_lock(mcm_rooms);
		if (reload) {
			ao2_callback(mcm_rooms, OBJ_NODATA | OBJ_UNLINK | OBJ_MULTIPLE, NULL, NULL);
		}

		while ((cat = ast_category_browse(cfg, cat))) {
			memset(conf_number, 0, sizeof(conf_number));
			if (sscanf(cat, "conf-%63s", conf_number) == 1) {
				ast_log(LOG_DEBUG, "prase_mcm_conference_config get mcm_conf_number:%s\n", conf_number);
				room = ao2_alloc(sizeof(*room), mcm_room_destroy);
				if (room) {
					room->number = ast_strdup(conf_number);
					ao2_link(mcm_rooms, room);
					ao2_ref(room, -1);
				}
			}
		}
		ao2_unlock(mcm_rooms);
		ast_config_destroy(cfg);
	}

	return 0;
}

int ast_sip_check_is_mcm_conf_number(const char *exten)
{
	int res = 0;
	RAII_VAR(struct mcm_room*, room, NULL, ao2_cleanup);

	if (ast_strlen_zero(exten)) {
		return 0;
	}

	if (mcm_rooms) {
		room = ao2_find(mcm_rooms, exten, OBJ_KEY);
		if (room) {
			ast_log(LOG_DEBUG, "exten[%s], is a SFU video meeting\n", exten);
			res = 1;
		}
	}

	return res;
}

/* Add by bfdu for direct media trunk, 2018.10.24 */
static void out_bound_prefix_destructor(void *obj)
{
	return;
}

static int add_direct_media_outbound_prefix(char *outbound_prefix)
{
	struct outbound_prefix_list *prefix = NULL;

	if (!outbound_prefix) {
		return -1;
	}

	prefix = ao2_alloc(sizeof(*prefix), out_bound_prefix_destructor);
	if (prefix) {
		strncpy(prefix->outbound_prefix, outbound_prefix, MAX_OUTBOUND_PREFIX_LEN - 1);
		AST_LIST_LOCK(&direct_media_outbound_prefix_list);
		AST_LIST_INSERT_TAIL(&direct_media_outbound_prefix_list, prefix, entry);
		AST_LIST_UNLOCK(&direct_media_outbound_prefix_list);
	}

	return 0;
}

static int remove_all_direct_media_outbound_prefix(void)
{
	struct outbound_prefix_list *prefix = NULL;

	AST_LIST_LOCK(&direct_media_outbound_prefix_list);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&direct_media_outbound_prefix_list, prefix, entry) {
		AST_LIST_REMOVE_CURRENT(entry);
		ast_log(LOG_DEBUG, "remove_all_direct_media_outbound_prefix:%s\n", prefix->outbound_prefix);
		ao2_ref(prefix, -1);
	}
	AST_LIST_TRAVERSE_SAFE_END;
	AST_LIST_UNLOCK(&direct_media_outbound_prefix_list);

	return 0;
}

int ast_sip_match_direct_media_outbound_prefix(char *prefix_match)
{
	struct outbound_prefix_list *prefix = NULL;
	int res = 0;

	ast_log(LOG_DEBUG, "match_direct_media_outbound_prefix begin for prefix_match:%s\n", prefix_match);

	AST_LIST_LOCK(&direct_media_outbound_prefix_list);
	AST_LIST_TRAVERSE_SAFE_BEGIN(&direct_media_outbound_prefix_list, prefix, entry) {
		if (prefix && ast_extension_match(prefix->outbound_prefix, prefix_match)) {
			ast_log(LOG_DEBUG, "match_direct_media_outbound_prefix for prefix_match:%s\n", prefix_match);
			res = 1;
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END
	AST_LIST_UNLOCK(&direct_media_outbound_prefix_list);

	return res;
}

static int prase_direct_media_outbound_prefix_conf_to_list(int reload)
{
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	struct ast_config *cfg = NULL;
	const char *val = NULL;
	char *val_copy = NULL;
	char *outbound_prefix = NULL;

	ast_log(LOG_DEBUG, "prase_direct_media_outbound_prefix_conf_to_list begin\n");

	cfg = ast_config_load2(DIRECT_MEDIA_OUTBOUND_PREFIX_CONF, "pjsip", config_flags);
	if (cfg == CONFIG_STATUS_FILEMISSING || cfg == CONFIG_STATUS_FILEUNCHANGED || cfg == CONFIG_STATUS_FILEINVALID) {
		long reason = (long)cfg;
		ast_log(LOG_DEBUG, DIRECT_MEDIA_OUTBOUND_PREFIX_CONF " file %s!", reason == 0 ? "missing" : ( reason == -1 ? "unchanged" : (reason == -2 ? "invalid": "(unkown reason)")));
		return 0;
	}

	remove_all_direct_media_outbound_prefix();

	if (!cfg) {
		return 0;
	}

	if ((val = ast_variable_retrieve(cfg, "general", "outbound_prefix"))) {
		ast_log(LOG_DEBUG, "Get direct media outbound_prefix:%s", val);
		val_copy = ast_malloc(strlen(val)+1);
		if (val_copy) {
			char *tmp_val = val_copy;
			strcpy(val_copy, val);
			while ((outbound_prefix = strsep(&tmp_val, " ")) ) {
				ast_log(LOG_DEBUG, "Get direct media outbound_prefix:%s", outbound_prefix);
				add_direct_media_outbound_prefix(outbound_prefix);
			}
			ast_free(val_copy);
			val_copy = NULL;
		}
	}
	ast_config_destroy(cfg);

	return 0;
}

int add_feature_allow_body(pjsip_tx_data *tdata)
{
	int i = 0;

	if (tdata == NULL) {
		return -1;
	}

        RAII_VAR(struct ast_str *, body_text, NULL, ast_free_ptr);

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-feature-code+xml",
	};

	if (!(body_text = ast_str_create(1000))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for INFO DTMF.\n");
		return -1;
	}

	ast_str_append(&body_text, 0,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	ast_str_append(&body_text, 0, "<featureCodeIdle>\n");

	ast_mutex_lock(&gs_feature_code_mutex);
	for (i = 0; i < GS_IDLE_FEATURE_COUNT; i++) {
		ast_str_append(&body_text, 0, "  <%s>%s</%s>\n", gs_idle_feature_code[i].f_name,
				gs_idle_feature_code[i].f_code,
				gs_idle_feature_code[i].f_name);
	}
	ast_mutex_unlock(&gs_feature_code_mutex);

	ast_str_append(&body_text, 0, "</featureCodeIdle>\n");

	body.body_text = ast_str_buffer(body_text);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to text video update INFO request\n");
		return -1;
	}

	return 0;
}

int ast_sip_get_nvram_value(char *pvalue, char* buf, int len )
{
#if 1
	if (ast_safe_nvram_get(pvalue, buf, len) == NULL) {
		memset(buf, 0, len);
	}
#else
	char cmd[32] = {0};

	if (pvalue == NULL) {
		return -1;
	}

	FILE *fp = NULL;
	memset(buf, 0, len);
	snprintf(cmd, sizeof(cmd), "nvram get %s", pvalue);

	fp = popen(cmd, "r");
	if (fp != NULL) {
		while (fp != NULL && (!feof(fp))
			   && (fgets(buf, len - 1, fp) != NULL)) {
			int length = strlen(buf);
			if (buf[length-1] == '\n') {
				buf[length-1] = '\0';
			}
			ast_log(LOG_DEBUG, "cmd Read: %s", buf);
		}
		pclose(fp);
		fp = NULL;
	}
#endif

	return 0;
}


int add_http_server_allow_body(pjsip_tx_data *tdata, const char *gs_http_server_uri)
{
	if (tdata == NULL || gs_http_server_uri == NULL) {
		return -1;
	}
        RAII_VAR(struct ast_str *, body_text, NULL, ast_free_ptr);
	char value[64] = {0};
	int is_https = 0;
	char gs_http_type[6] = {0};
	int gs_http_server_port = 0;
	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-ucm-info+xml",
	};

	if (!(body_text = ast_str_create(1000))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for INFO DTMF.\n");
		return -1;
	}
	ast_sip_get_nvram_value("900", value, sizeof(value));
	is_https = atoi(value);
	snprintf(gs_http_type, sizeof(gs_http_type), "%s", is_https?"https" : "http");
	ast_sip_get_nvram_value("901", value, sizeof(value));
	gs_http_server_port = atoi(value);
	ast_log(LOG_DEBUG, "The http type is: %s, the port is %d \n", gs_http_type, gs_http_server_port);
	ast_str_append(&body_text, 0,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	ast_str_append(&body_text, 0, "<root>\n");
	ast_str_append(&body_text, 0, "  <cti>\n");
	ast_str_append(&body_text, 0, "    <server>%s://%s:%d/cticgi</server>\n", gs_http_type, gs_http_server_uri, gs_http_server_port);
	ast_str_append(&body_text, 0, "  </cti>\n");
	ast_str_append(&body_text, 0, "</root>\n");

	body.body_text = ast_str_buffer(body_text);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to text video update INFO request\n");
		return -1;
	}

	return 0;
}

int add_mailbox_allow_body(pjsip_rx_data *rdata, pjsip_tx_data *tdata)
{
	RAII_VAR(struct ast_str *, body_text, NULL, ast_free_ptr);
	pjsip_sip_uri *uri = NULL;
	char value[64] = {0};
	int is_https = 0;
	char gs_http_type[6] = {0};
	char gs_http_server_host[256] = {0};
	int gs_http_server_port = 0;
	char family[64] = {0};
	int skip_vmauth = 0;
	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-mailbox-info+xml",
	};

	if (rdata == NULL || tdata == NULL) {
		return -1;
	}

	if (!(body_text = ast_str_create(1000))) {
		ast_log(LOG_ERROR, "Could not allocate buffer for mailbox info\n");
		return -1;
	}
	ast_sip_get_nvram_value("900", value, sizeof(value));
	is_https = atoi(value);
	snprintf(gs_http_type, sizeof(gs_http_type), "%s", is_https?"https" : "http");
	ast_sip_get_nvram_value("901", value, sizeof(value));
	gs_http_server_port = atoi(value);
	uri = pjsip_uri_get_uri(rdata->msg_info.to->uri);
	if (!uri) {
		ast_log(LOG_ERROR, "Couldn't get uri in to header.\n");
		return -1;
	}
	ast_copy_pj_str(gs_http_server_host, &uri->host, sizeof(gs_http_server_host));

	/* get AAUSER SKIPVMAUTH data from ucmdb */
	snprintf(family, sizeof(family), "AAUSER/%.*s", (int)uri->user.slen, uri->user.ptr);
	if (ast_ucm_db_get(family, "SKIPVMAUTH", value, sizeof(value))) {
		ast_log(LOG_ERROR, "Couldn't find the %s/SKIPVMAUTH in ucmdb database.\n", family);
		return -1;
	}
	skip_vmauth = !strcasecmp(value, "yes") ? 1 : 0;

	ast_str_append(&body_text, 0,"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	ast_str_append(&body_text, 0, "<root>\n");
	ast_str_append(&body_text, 0, "  <mailbox>\n");
	ast_str_append(&body_text, 0, "    <name>%.*s</name>\n", (int)uri->user.slen, uri->user.ptr);
	ast_str_append(&body_text, 0, "    <auth>%s</auth>\n", !skip_vmauth ? "enable" : "disable");
	ast_str_append(&body_text, 0, "    <url>%s://%s:%d/vvmapi</url>\n", gs_http_type, gs_http_server_host, gs_http_server_port);
	ast_str_append(&body_text, 0, "  </mailbox>\n");
	ast_str_append(&body_text, 0, "</root>\n");

	body.body_text = ast_str_buffer(body_text);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to mailbox info\n");
		return -1;
	}

	return 0;
}

/*X-GS-Mailbox-Info*/
int ast_sip_add_mailbox_info_header(pjsip_tx_data *tdata, struct ast_sip_endpoint *endpoint)
{
	char value[64] = {0};
	int is_https = 0;
	char gs_http_type[6] = {0};
	char gs_http_server_host[256] = {0};
	int gs_http_server_port = 0;
	char family[64] = {0};
	int skip_vmauth = 0;
	int enable_vmbox = 0;
	char url_value[256] = {0};
	char header_value[256] = {0};
	int is_ipv6 = 0;

	/*api enable*/
	ast_sip_get_nvram_value("lighttpd_server", value, sizeof(value));
	if (strncmp(value, "yes", 3)) {
	ast_log(LOG_DEBUG, "API isn't enabled");
		return 0;
	}
	char dev_base[32];
	memset(dev_base, 0, sizeof(dev_base));
	ast_sip_get_nvram_value(":dev_base", dev_base, sizeof(dev_base));
	/*bug278543 modify by jmhuang, CloudUCM nvram 900=0, need set https*/

	/*schema*/
	ast_sip_get_nvram_value("900", value, sizeof(value));
	is_https = atoi(value);
	snprintf(gs_http_type, sizeof(gs_http_type), "%s", (is_https || NULL != strstr(dev_base,"CloudUCM")) ? "https" : "http");


	/*port*/
	ast_sip_get_nvram_value("901", value, sizeof(value));
	gs_http_server_port = atoi(value);

	/*host*/
	pjsip_fromto_hdr *from = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, NULL);
	if (!from) {
		ast_log(LOG_WARNING, "Couldn't find the from header in tdata message for '%s'\n", endpoint->aors);
		return -1;
	}
	pjsip_sip_uri *sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from->uri);
	ast_copy_pj_str(gs_http_server_host, &sip_from->host, sizeof(gs_http_server_host));
	if (pj_strchr(&sip_from->host, ':')) {
		is_ipv6 = 1;
	}

	/*get AAUSER SKIPVMAUTH data from ucmdb*/
	snprintf(family, sizeof(family), "AAUSER/%s", endpoint->aors);
	if (ast_ucm_db_get(family, "SKIPVMAUTH", value, sizeof(value))) {
		ast_log(LOG_ERROR, "Couldn't find the %s/SKIPVMAUTH in ucmdb database.\n", family);
		return -1;
	}
	skip_vmauth = !strcasecmp(value, "yes") ? 1 : 0;

	/*get AAUSER VMBOX data from ucmdb*/
	snprintf(family, sizeof(family), "AAUSER/%s", endpoint->aors);
	if (ast_ucm_db_get(family, "VMBOX", value, sizeof(value))) {
		ast_log(LOG_ERROR, "Couldn't find the %s/VMBOX in ucmdb database.\n", family);
		return -1;
	}
	enable_vmbox = !strcasecmp(value, "yes") ? 1 : 0;

	snprintf(url_value, sizeof(url_value), "%s://%s%s%s:%d", gs_http_type, is_ipv6 ? "[" : "", gs_http_server_host, is_ipv6 ? "]" : "", gs_http_server_port);
	memset(header_value, 0, sizeof(header_value));
	snprintf(header_value, sizeof(header_value), "auth=%s;url=%s", !skip_vmauth ? "enable" : "disable", enable_vmbox ? url_value: "");
	ast_sip_add_header(tdata, "X-GS-Mailbox-Info", header_value);

	return 0;
}

static int update_total_turn_session_count(unsigned int count)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	total_turn_session_count = count;
	ast_rwlock_unlock(&total_turn_session_lock);

	return total_turn_session_count;
}

static int update_max_media_nat_call_count(const char *count)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!count) {
		turn_max_media_nat_call = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_MAX_MEDIA_NAT_CALL_KEY);
	} else {
		turn_max_media_nat_call = atoi(count);
		ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_MAX_MEDIA_NAT_CALL_KEY, count);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return turn_max_media_nat_call;
}

static void update_turn_server_address(const char *address)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	strncpy(turn_server_address, address, MAX_TURN_SERVER_LEN - 1);
	ast_rwlock_unlock(&total_turn_session_lock);

	ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_ADDRESS_KEY, address);

	return;
}

static void update_stun_server_address(const char *address)
{
	int changed = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	if (address && stun_server_address && strncmp(stun_server_address, address, MAX_TURN_SERVER_LEN)) {
		changed = 1;
		strncpy(stun_server_address, address, MAX_TURN_SERVER_LEN - 1);
	}

	ast_rwlock_unlock(&total_turn_session_lock);

	if (changed) {
		ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_STUN_ADDRESS_KEY, address);
		ast_safe_system2("module reload res_stun_monitor.so");
	}

	return;
}

static void update_turn_user_name(const char *name)
{
	char *user_name_decrypted = NULL;

	ucm_secret_decrypt(name, &user_name_decrypted);
	if (!ast_strlen_zero(user_name_decrypted)) {
		ast_rwlock_wrlock(&total_turn_session_lock);
		strncpy(turn_user_name, user_name_decrypted, MAX_TURN_USER_NAME_LEN - 1);
		ast_rwlock_unlock(&total_turn_session_lock);

		ast_free(user_name_decrypted);
		user_name_decrypted = NULL;
	} else {
		ast_log(LOG_ERROR, "Turn user name decrypted failed\n");
	}

	if (ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_USERNAME_KEY, name)) {
		ast_log(LOG_ERROR, "update_turn_user_name put name to DB error\n");
	}

	return;
}

static void update_turn_password(const char *password)
{
	char *password_decrypted = NULL;

	ucm_secret_decrypt(password, &password_decrypted);
	if (!ast_strlen_zero(password_decrypted)) {
		ast_rwlock_wrlock(&total_turn_session_lock);
		strncpy(turn_password, password_decrypted, MAX_TURN_PASSWORD_LEN - 1);
		ast_rwlock_unlock(&total_turn_session_lock);

		ast_free(password_decrypted);
		password_decrypted = NULL;
	} else {
		ast_log(LOG_ERROR, "Turn password decrypted failed\n");
	}

	if (ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_PASSWORD_KEY, password)) {
		ast_log(LOG_ERROR, "update_turn_password put pw to DB error\n");
	}

	return;
}

static void update_turn_protocol_type(const char *protocol_type)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	strncpy(turn_protocol_type, protocol_type, MAX_TURN_PROTOCOL_TYPE_LEN - 1);
	ast_rwlock_unlock(&total_turn_session_lock);

	ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_PROTOCOLTYPE_KEY, protocol_type);

	return;
}

static unsigned int update_turn_limit_time(char *turn_limit_time_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!turn_limit_time_str) {
		turn_limit_time = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_LIMITTIME_KEY);
	} else {
		turn_limit_time = atoi(turn_limit_time_str);
		ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_LIMITTIME_KEY, turn_limit_time_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return turn_limit_time;
}

static unsigned int  update_turn_day_total_limit_time(char *turn_day_total_limit_time_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!turn_day_total_limit_time_str) {
		turn_day_total_limit_time = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_DAY_TOTAL_LIMITTIME_KEY);
	} else {
		turn_day_total_limit_time = atoi(turn_day_total_limit_time_str);
		ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_DAY_TOTAL_LIMITTIME_KEY, turn_day_total_limit_time_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return turn_day_total_limit_time;
}

static unsigned int  update_turn_month_total_limit_time(char *turn_month_total_limit_time_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!turn_month_total_limit_time_str) {
		turn_month_total_limit_time = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_MONTH_TOTAL_LIMITTIME_KEY);
	} else {
		turn_month_total_limit_time = atoi(turn_month_total_limit_time_str);
		ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_MONTH_TOTAL_LIMITTIME_KEY, turn_month_total_limit_time_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return turn_month_total_limit_time;
}

static unsigned int update_turn_max_registrar(char *turn_max_registrar_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!turn_max_registrar_str) {
		turn_max_registrar = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_MAXREGISTRAR_KEY);
	} else {
		turn_max_registrar = atoi(turn_max_registrar_str);
		ast_db_put(TURN_INFO_DB_FAMILY, TURN_INFO_MAXREGISTRAR_KEY, turn_max_registrar_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return turn_max_registrar;
}

static unsigned int update_plugin_market(const char *plugin_market_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!plugin_market_str) {
		plugin_market = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, PLUGIN_MARKET_KEY);
	} else {
		plugin_market = atoi(plugin_market_str);
		ast_db_put(TURN_INFO_DB_FAMILY, PLUGIN_MARKET_KEY, plugin_market_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return plugin_market;
}

static unsigned int update_plugin_thirdparty(const char *plugin_thirdparty_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!plugin_thirdparty_str) {
		plugin_thirdparty = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, PLUGIN_THIRDPARTY_KEY);
	} else {
		plugin_thirdparty = atoi(plugin_thirdparty_str);
		ast_db_put(TURN_INFO_DB_FAMILY, PLUGIN_THIRDPARTY_KEY, plugin_thirdparty_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return plugin_thirdparty;
}

static unsigned int update_support_webrtc_trunk(const char *support_webrtc_trunk_str)
{
	ast_rwlock_wrlock(&total_turn_session_lock);
	if (!support_webrtc_trunk_str) {
		support_webrtc_trunk = 0;
		ast_db_del(TURN_INFO_DB_FAMILY, SUPPORT_WEBRTC_TRUNK);
	} else {
		support_webrtc_trunk = atoi(support_webrtc_trunk_str);
		ast_db_put(TURN_INFO_DB_FAMILY, SUPPORT_WEBRTC_TRUNK, support_webrtc_trunk_str);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return support_webrtc_trunk;
}

unsigned int ast_sip_get_support_webrtc_trunk(void)
{
	unsigned int support = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	support = support_webrtc_trunk;
	ast_rwlock_unlock(&total_turn_session_lock);

	return support;
}

unsigned int ast_sip_get_plugin_market(void)
{
	unsigned int market = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	market = plugin_market;
	ast_rwlock_unlock(&total_turn_session_lock);

	return market;
}

unsigned int ast_sip_get_plugin_thirdparty(void)
{
	unsigned int thirdparty = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	thirdparty = plugin_thirdparty;
	ast_rwlock_unlock(&total_turn_session_lock);

	return thirdparty;
}

unsigned int ast_sip_get_turn_limit_time(void)
{
	unsigned int time = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	time = turn_limit_time;
	ast_rwlock_unlock(&total_turn_session_lock);

	return time;
}

unsigned int ast_sip_get_turn_day_total_limit_time(void)
{
	unsigned int time = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	time = turn_day_total_limit_time;
	ast_rwlock_unlock(&total_turn_session_lock);

	return time;
}

unsigned int ast_sip_get_turn_month_total_limit_time(void)
{
	unsigned int time = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	time = turn_month_total_limit_time;
	ast_rwlock_unlock(&total_turn_session_lock);

	return time;
}

unsigned int ast_sip_get_turn_max_registrar(void)
{
	unsigned int max_registrar = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	max_registrar = turn_max_registrar;
	ast_rwlock_unlock(&total_turn_session_lock);

	return max_registrar;
}

char *ast_sip_get_turn_server_address(void)
{
	char *address = NULL;

	ast_rwlock_wrlock(&total_turn_session_lock);
	address = ast_strdup(turn_server_address);
	ast_rwlock_unlock(&total_turn_session_lock);

	return address;
}

char *ast_sip_get_turn_user_name(void)
{
	char *user_name = NULL;

	ast_rwlock_wrlock(&total_turn_session_lock);
	user_name = ast_strdup(turn_user_name);
	ast_rwlock_unlock(&total_turn_session_lock);

	return user_name;
}

char *ast_sip_get_turn_password(void)
{
	char *password = NULL;

	ast_rwlock_wrlock(&total_turn_session_lock);
	password = ast_strdup(turn_password);
	ast_rwlock_unlock(&total_turn_session_lock);

	return password;
}

char *ast_sip_get_turn_protocol_type(void)
{
	char *protocol_type = NULL;

	ast_rwlock_wrlock(&total_turn_session_lock);
	protocol_type = ast_strdup(turn_protocol_type);
	ast_rwlock_unlock(&total_turn_session_lock);

	return protocol_type;
}

char *ast_sip_get_stun_address(void)
{
	char *stun_address = NULL;

	ast_rwlock_wrlock(&total_turn_session_lock);
	stun_address = ast_strdup(stun_server_address);
	ast_rwlock_unlock(&total_turn_session_lock);

	return stun_address;
}

static int turn_info_memcpy_encrypt(unsigned char *dst, const unsigned char *src, int len, ast_aes_encrypt_key *ecx)
{
	int dst_len = 0;
	int copy_len = 0;
	unsigned char curblock[16] = { 0 };
	int x = 0;
	while(len > 0) {
		if (len < 16) {
			copy_len = len;
		} else {
			copy_len = 16;
		}
		for (x=0; x < copy_len; x++) {
			curblock[x] ^= src[x];
		}
		ast_aes_encrypt(curblock, dst, ecx);
		memcpy(curblock, dst, sizeof(curblock));
		dst += 16;
		src += 16;
		len -= 16;
		dst_len += 16;
	}

	return dst_len;
}

static char *encrypt_turn_info_base64(const char *turn_info)
{
	ast_aes_encrypt_key ecx;
	char *encrypt_data = NULL;
	char *encrypt_data_base64 = NULL;
	int dst_len = 0;

	encrypt_data = ast_malloc(1024);
	if (!encrypt_data) {
		return NULL;
	}

	encrypt_data_base64 = ast_malloc(1024);
	if (!encrypt_data_base64) {
		ast_free(encrypt_data);
		encrypt_data = NULL;
		return NULL;
	}

	memset(encrypt_data, 0, 1024);
	memset(encrypt_data_base64, 0, 1024);

	ast_aes_set_encrypt_key((unsigned char *)TURN_INFO_ENCRYPT_KEY, &ecx);

	dst_len = turn_info_memcpy_encrypt((unsigned char *)encrypt_data, (const unsigned char *)turn_info, strlen(turn_info), &ecx);

	ast_base64encode(encrypt_data_base64, (const unsigned char *)encrypt_data, dst_len, 1024);

	ast_free(encrypt_data);
	encrypt_data = NULL;

	return encrypt_data_base64;
}

#if 0
/* For we don't need to decrypt, reserve now */
static void config_info_memcpy_decrypt(unsigned char *dst, const unsigned char *src, int len, ast_aes_decrypt_key *dcx)
{
	unsigned char lastblock[16] = { 0 };
	int x;
	while(len > 0) {
		ast_aes_decrypt(src, dst, dcx);
		for (x=0;x<16;x++)
			dst[x] ^= lastblock[x];
		memcpy(lastblock, src, sizeof(lastblock));
		dst += 16;
		src += 16;
		len -= 16;
	}
}

static char *decrypt_config_xml_base64(const char *encrypt_config_info, char *key)
{
	ast_aes_encrypt_key dcx;
	char *decrypt_data = NULL;
	char *decrypt_data_base64 = NULL;
	int len = 0;

	decrypt_data = ast_malloc(AES_ALL_ENCRYPT_BUFFER_LEN);
	if (!decrypt_data) {
		return NULL;
	}

	decrypt_data_base64 = ast_malloc(AES_ALL_ENCRYPT_BUFFER_LEN);
	if (!decrypt_data_base64) {
		ast_free(decrypt_data);
		decrypt_data = NULL;
		return NULL;
	}

	memset(decrypt_data, 0, AES_ALL_ENCRYPT_BUFFER_LEN);
	memset(decrypt_data_base64, 0, AES_ALL_ENCRYPT_BUFFER_LEN);

	ast_aes_set_decrypt_key((unsigned char *)key, &dcx);

	len = ast_base64decode((unsigned char *)decrypt_data_base64, encrypt_config_info, AES_ALL_ENCRYPT_BUFFER_LEN);

	config_info_memcpy_decrypt((unsigned char *)decrypt_data, (const unsigned char *)decrypt_data_base64, len, &dcx);

	ast_free(decrypt_data_base64);
	decrypt_data_base64 = NULL;

	return decrypt_data;
}
#endif

static char *encrypt_config_xml_base64(const char *xml_buf, char *key)
{
	ast_aes_encrypt_key ecx;
	char *encrypt_data = NULL;
	char *encrypt_data_base64 = NULL;
	int enc_len = 0;

	encrypt_data = ast_malloc(AES_ALL_ENCRYPT_BUFFER_LEN);
	if (!encrypt_data) {
		return NULL;
	}

	encrypt_data_base64 = ast_malloc(AES_ALL_ENCRYPT_BUFFER_LEN);
	if (!encrypt_data_base64) {
		ast_free(encrypt_data);
		encrypt_data = NULL;
		return NULL;
	}

	memset(encrypt_data, 0, AES_ALL_ENCRYPT_BUFFER_LEN);
	memset(encrypt_data_base64, 0, AES_ALL_ENCRYPT_BUFFER_LEN);

	ast_aes_set_encrypt_key((unsigned char *)key, &ecx);

	enc_len = turn_info_memcpy_encrypt((unsigned char *)encrypt_data, (const unsigned char *)xml_buf, strlen(xml_buf), &ecx);

	ast_log(LOG_NOTICE, "aes enc_len[%d]", enc_len);
	ast_base64encode(encrypt_data_base64, (const unsigned char *)encrypt_data, enc_len, AES_ALL_ENCRYPT_BUFFER_LEN);

	ast_free(encrypt_data);
	encrypt_data = NULL;

	return encrypt_data_base64;
}

static void add_ldap_config(struct ast_sip_config_info config_info, struct ast_str **config_string, int is_public)
{
	if (!config_info.require_ldap_settings) {
		return;
	}

	/* get xml text */
	if (!strlen(ldap_xml_buffer)) {
		ast_sip_init_ldap_xml_buffer();
	}

	if (!strlen(ldap_xml_buffer)) {
		ast_log(LOG_ERROR, "No ldap xml data.\n");
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	if (is_public && strlen(ldap_xml_public_buffer)) {
		ast_str_append(config_string, 0, "%s", ldap_xml_public_buffer);
	} else {
		ast_str_append(config_string, 0, "%s", ldap_xml_buffer);
	}
	ast_rwlock_unlock(&config_xml_lock);

	return;
}

static void add_feature_code_config(struct ast_sip_config_info config_info, struct ast_str **config_string)
{
	if (!config_info.require_feature_code_settings) {
		return;
	}

	/* get xml text */
	if (!strlen(feature_code_xml_buffer)) {
		ast_sip_init_feature_code_xml_buffer();
	}

	if (!strlen(feature_code_xml_buffer)) {
		ast_log(LOG_ERROR, "No feature code xml data.\n");
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	ast_str_append(config_string, 0, "%s", feature_code_xml_buffer);
	ast_rwlock_unlock(&config_xml_lock);

	return;
}

static void add_basic_config(struct ast_sip_config_info config_info, struct ast_str **config_string)
{
	if (!config_info.require_basic_settings) {
		return;
	}

	/* get xml text */
	if (!strlen(basic_xml_buffer)) {
		ast_sip_init_basic_xml_buffer();
	}

	if (!strlen(basic_xml_buffer)) {
		ast_log(LOG_ERROR, "No basic xml data.\n");
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	ast_str_append(config_string, 0, "%s", basic_xml_buffer);
	ast_rwlock_unlock(&config_xml_lock);

	return;
}

static void add_turn_info_config(struct ast_sip_config_info config_info, struct ast_str **config_string)
{
	unsigned int len = 0;
	char buf[512] = {0};
	char content[2048] = { 0 };
	RAII_VAR(char *, turn_address, NULL, ast_free);
	RAII_VAR(char *, name, NULL, ast_free);
	RAII_VAR(char *, password, NULL, ast_free);
	RAII_VAR(char *, protocol_type, NULL, ast_free);
	RAII_VAR(char *, stun_address, NULL, ast_free);

	if (!config_info.require_turn_info_settings) {
		return;
	}

	turn_address = ast_sip_get_turn_server_address();
	name = ast_sip_get_turn_user_name();
	password = ast_sip_get_turn_password();
	protocol_type = ast_sip_get_turn_protocol_type();
	stun_address = ast_sip_get_stun_address();

	len = sizeof(content);
	snprintf(content, len, "    <item name=\"turnInfo\">\r\n");

	if (turn_address) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <TurnServerAddress>%s</TurnServerAddress>\r\n", turn_address);
		strncat(content, buf, len - strlen(content));
	}

	if (name) {
		memset(buf, 0, sizeof(buf));
		char name_escape[MAX_TURN_USER_NAME_LEN + 8] = { 0 };
		ast_xml_escape(name, name_escape, sizeof(name_escape));
		snprintf(buf, sizeof(buf) - 1, "      <TurnServerUsername>%s</TurnServerUsername>\r\n", name_escape);
		strncat(content, buf, len - strlen(content));
	}

	if (password) {
		memset(buf, 0, sizeof(buf));
		char password_escape[MAX_TURN_PASSWORD_LEN + 8] = { 0 };
		ast_xml_escape(password, password_escape, sizeof(password_escape));
		snprintf(buf, sizeof(buf) - 1, "      <TurnServerPassword>%s</TurnServerPassword>\r\n", password_escape);
		strncat(content, buf, len - strlen(content));
	}

	if (protocol_type) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <TurnServerProtocolType>%s</TurnServerProtocolType>\r\n", protocol_type);
		strncat(content, buf, len - strlen(content));
	}

	if (stun_address) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <StunServerAddress>%s</StunServerAddress>\r\n", stun_address);
		strncat(content, buf, len - strlen(content));
	}

	strncat(content, "    </item>\r\n", len - strlen(content));

	ast_str_append(config_string, 0, "%s", content);

	return;
}

static void add_plugin_config(struct ast_sip_config_info config_info, struct ast_str **config_string)
{
	unsigned int len = 0;
	char buf[128] = {0};
	char content[256] = { 0 };
	char plugin_market[8] = { 0 };
	char plugin_thirdparty[8] = { 0 };
	unsigned int market = 0;
	unsigned int thirdparty = 0;

	if (!config_info.require_plugin) {
		return;
	}

	market = ast_sip_get_plugin_market();
	thirdparty = ast_sip_get_plugin_thirdparty();

	if (market) {
		strcpy(plugin_market, "yes");
	} else {
		strcpy(plugin_market, "no");
	}

	if (thirdparty) {
		strcpy(plugin_thirdparty, "yes");
	} else {
		strcpy(plugin_thirdparty, "no");
	}

	len = sizeof(content);
	snprintf(content, len, "    <item name=\"plugin\">\r\n");

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "      <market>%s</market>\r\n", plugin_market);
	strncat(content, buf, len - strlen(content));

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "      <thirdparty>%s</thirdparty>\r\n", plugin_thirdparty);
	strncat(content, buf, len - strlen(content));

	strncat(content, "    </item>\r\n", len - strlen(content));

	ast_str_append(config_string, 0, "%s", content);

	ast_log(LOG_DEBUG, "add_plugin_config content:\n%s\n", content);

	return;
}

static void add_phonebook_config(struct ast_sip_config_info config_info, struct ast_str **config_string, int is_public)
{
	if (!config_info.require_phonebook) {
		return;
	}

	/* get xml text */
	if (!strlen(phonebook_xml_buffer)) {
		ast_sip_init_phonebook_xml_buffer();
	}

	if (!strlen(phonebook_xml_buffer)) {
		ast_log(LOG_ERROR, "No ldap xml data.\n");
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	if (is_public && strlen(phonebook_xml_public_buffer)) {
		ast_str_append(config_string, 0, "%s", phonebook_xml_public_buffer);
	} else {
		ast_str_append(config_string, 0, "%s", phonebook_xml_buffer);
	}
	ast_rwlock_unlock(&config_xml_lock);

	return;
}

static void add_gs_config(struct ast_sip_config_info config_info, struct ast_str **config_string, int is_public)
{
	char config_header[128] = { 0 };

	if (!config_info.require_ldap_settings &&
		!config_info.require_feature_code_settings &&
		!config_info.require_turn_info_settings &&
		!config_info.require_basic_settings &&
		!config_info.require_plugin &&
		!config_info.require_phonebook) {
		ast_log(LOG_WARNING, "add_gs_config but all config no need to send\n");
		return;
	}

	snprintf(config_header, sizeof(config_header), "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n<gs_provision>\r\n  <config version=\"2\">\r\n");
	ast_str_append(config_string, 0, "%s", config_header);

	add_ldap_config(config_info, config_string, is_public);
	add_feature_code_config(config_info, config_string);
	add_turn_info_config(config_info, config_string);
	add_basic_config(config_info, config_string);
	add_plugin_config(config_info, config_string);
	add_phonebook_config(config_info, config_string, is_public);

	memset(config_header, 0, sizeof(config_header));
	snprintf(config_header, sizeof(config_header), "  </config>\r\n</gs_provision>\r\n");
	ast_str_append(config_string, 0, "%s", config_header);

	return;
}

int ast_sip_add_gs_config_body(struct ast_sip_endpoint *endpoint, pjsip_tx_data *tdata, int is_public,
	struct ast_sip_config_info config_info)
{
	RAII_VAR(struct ast_sip_auth *, auth, NULL, ao2_cleanup);
	char *encrypt_data = NULL;
	const char *endpoint_id = NULL;
	char key[128] = {0};
	struct ast_str *config_string = NULL;

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-config",
	};

	if (endpoint == NULL || tdata == NULL) {
		return -1;
	}

	endpoint_id = ast_sorcery_object_get_id(endpoint);
	if (!strncmp(endpoint_id, "trunk_", 6)) {
		return -1;
	}

	auth = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "auth", endpoint_id);
	if (!auth) {
		return -1;
	}

	/* generate key */
	snprintf(key, sizeof(key), "%s", auth->md5_creds);

	config_string = ast_str_create(AES_ALL_ENCRYPT_BUFFER_LEN);
	if (!config_string) {
		return -1;
	}

	add_gs_config(config_info, &config_string, is_public);

	encrypt_data = encrypt_config_xml_base64(ast_str_buffer(config_string), key);
	if (!encrypt_data) {
		ast_log(LOG_ERROR, "encrypt_config_xml_base64 failed.\n");
		ast_free(config_string);
		config_string = NULL;
		return -1;
	}

	body.body_text = (const char*)encrypt_data;
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add x-gs-config body\n");
		ast_free(encrypt_data);
		encrypt_data = NULL;
		ast_free(config_string);
		config_string = NULL;
		return -1;
	}

	ast_log(LOG_NOTICE, "Added x-gs-config body, is_public[%d]\n", is_public);

	ast_free(encrypt_data);
	encrypt_data = NULL;
	ast_free(config_string);
	config_string = NULL;

	return 0;
}

int ast_sip_add_webrtc_trunk_gs_config_body(struct ast_sip_endpoint *endpoint, pjsip_tx_data *tdata,
	struct ast_sip_config_info config_info)
{
	char *encrypt_data = NULL;
	char trunk_id[64] = {0};
	char key[128] = {0};
	struct ast_str *config_string = NULL;

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-config",
	};

	if (endpoint == NULL || tdata == NULL) {
		return -1;
	}

	/* generate key */
	snprintf(trunk_id, sizeof(trunk_id), "%s", ast_sorcery_object_get_id(endpoint));
	ast_md5_hash(key, trunk_id);

	config_string = ast_str_create(AES_ALL_ENCRYPT_BUFFER_LEN);
	if (!config_string) {
		return -1;
	}

	add_gs_config(config_info, &config_string, 1);

	encrypt_data = encrypt_config_xml_base64(ast_str_buffer(config_string), key);
	if (!encrypt_data) {
		ast_log(LOG_ERROR, "encrypt_config_xml_base64 failed.\n");
		ast_free(config_string);
		config_string = NULL;
		return -1;
	}

	body.body_text = (const char*)encrypt_data;
	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add x-gs-config body\n");
		ast_free(encrypt_data);
		encrypt_data = NULL;
		ast_free(config_string);
		config_string = NULL;
		return -1;
	}

	ast_log(LOG_NOTICE, "Added webrtc trunk x-gs-config body\n");

	ast_free(encrypt_data);
	encrypt_data = NULL;
	ast_free(config_string);
	config_string = NULL;

	return 0;
}

static int read_url_from_xml(char *xml, char *url)
{
	struct ast_xml_doc  *xml_doc          = NULL;
	struct ast_xml_node *phonebook_node   = NULL;
	struct ast_xml_node *child_nodes      = NULL;
	struct ast_xml_node *node_iterator    = NULL;
	struct ast_xml_node *node_item        = NULL;
	const char *node_name;
	const char *content;

	if (!xml) {
		ast_log(LOG_WARNING, "read_url_from_xml xml is NULL\n");
		return -1;
	}

	if (!(xml_doc = ast_xml_read_memory(xml, strlen(xml)))) {
		ast_log(LOG_WARNING, "read_url_from_xml xml can't read\n");
		return -1;
	}

	phonebook_node = ast_xml_get_root(xml_doc);
	if (!phonebook_node) {
		ast_log(LOG_WARNING, "read_url_from_xml xml can't get root node\n");
		ast_xml_close(xml_doc);
		return -1;
	}

	child_nodes = ast_xml_node_get_children(phonebook_node);
	if (child_nodes) {
		for (node_iterator = child_nodes; node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			if (!strcmp(node_name, "item") && !node_item) {
				node_item = node_iterator;
				break;
			}
		}
	}

	if (node_item) {
		for (node_iterator = ast_xml_node_get_children(node_item); node_iterator; node_iterator = ast_xml_node_get_next(node_iterator)) {
			node_name = ast_xml_node_get_name(node_iterator);
			content = ast_xml_get_text(node_iterator);
			if (!content) {
				continue;
			}

			if (!strcmp(node_name, "url")) {
				strncpy(url, content, MAX_PHONEBOOK_URL_LEN -1);
				ast_log(LOG_DEBUG, "read_url_from_xml, url[%s]\n", content);
				ast_xml_free_text(content);
				break;
			}

			ast_xml_free_text(content);
		}
	} else {
		ast_log(LOG_WARNING, "read_url_from_xml xml can't get item node\n");
	}

	ast_xml_close(xml_doc);

	return 0;
}

static void create_xml_for_phonebook(char *xml, int max_xml_len)
{
	if (!xml) {
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);

	strcpy(xml, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n<gs_provision>\r\n");
	strncat(xml, phonebook_xml_public_buffer, max_xml_len - strlen(xml));
	strncat(xml, "</gs_provision>\r\n", max_xml_len - strlen(xml));

	ast_log(LOG_DEBUG, "create_xml_for_phonebook xml:%s\r\n\n", xml);

	ast_rwlock_unlock(&config_xml_lock);

	return;
}

static void set_is_public(struct ast_sip_contact *contact, int *is_public)
{
	if (!contact || !is_public) {
		return;
	}

	if (contact->ssh_proxy_flag) {
		ast_log(LOG_DEBUG, "ssh_proxy_flag enable, set is_public to ture\n");
		*is_public = 1;
		return;
	}

	ast_log(LOG_DEBUG, "set_is_public url:%s, to_host:%s\n", phonebook_public_url, contact->to_host);

	if (strcasestr(phonebook_public_url, contact->to_host)) {
		ast_log(LOG_DEBUG, "Register To header match public host IP, set is_public to ture\n");
		*is_public = 1;
		return;
	}

	return;
}

int ast_sip_set_config_settings(pjsip_rx_data *rdata, int *is_public,
	struct ast_sip_config_info config_info)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_aor *, aor, NULL, ao2_cleanup);
	struct ao2_iterator i;
	struct ast_sip_contact *contact;
	char aor_id[80]      = {0};
	char contact_id[80]  = {0};
	char contact_uri[80] = {0};
	char src_addr[80]    = {0};
	char *endpoint_id    = NULL;

	if (!rdata) {
		return 0;
	}

	endpoint = ast_pjsip_rdata_get_endpoint(rdata);
	if (!endpoint) {
		ast_log(LOG_NOTICE, "can't get endpoint from rdata.");
		return -1;
	}

	endpoint_id = (char *)ast_sorcery_object_get_id(endpoint);
	if (!strncmp(endpoint_id, "trunk_", 6)) {
		return -1;
	}

	aor = ast_sip_location_retrieve_aor(endpoint->aors); 
	if (!aor) {
		return -1;
	}

	contacts = ast_sip_location_retrieve_aor_contacts(aor);
	if (!contacts || !ao2_container_count(contacts)) {
		return -1;
	}

	if (strchr(rdata->pkt_info.src_name, ':')) {
		snprintf(src_addr, sizeof(src_addr) - 1, "[%s]:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port); 
	} else {
		snprintf(src_addr, sizeof(src_addr) - 1, "%s:%d", rdata->pkt_info.src_name, rdata->pkt_info.src_port); 
	}

	i = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_iterator_next(&i))) {
		sscanf(contact->uri, "%79[^@]@%79[^;]", aor_id, contact_uri);
		if (!strcmp(contact_uri, src_addr)) {
			snprintf(contact_id, sizeof(contact_id), "%s",ast_sorcery_object_get_id(contact));
			ao2_ref(contact, -1); 
			break;
		}

		ao2_ref(contact, -1); 
	}
	ao2_iterator_destroy(&i);

	if (strlen(contact_id)) {
		contact = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "contact", contact_id);
		if (contact) {
			if (config_info.require_ldap_settings) {
				ast_log(LOG_NOTICE, "Set support_ldap_settings to contact[%s].", contact_id);
				contact->support_ldap_settings = 1;
				*is_public = contact->ssh_proxy_flag;
			}

			if (config_info.require_feature_code_settings) {
				ast_log(LOG_NOTICE, "Set require_feature_code_settings to contact[%s].", contact_id);
				contact->require_feature_code_settings = 1;
			}

			if (config_info.require_turn_info_settings) {
				ast_log(LOG_NOTICE, "Set require_turn_info_settings to contact[%s].", contact_id);
				contact->require_turn_info_settings = 1;
			}

			if (config_info.require_basic_settings) {
				ast_log(LOG_NOTICE, "Set require_basic_settings to contact[%s].", contact_id);
				contact->require_basic_settings = 1;
			}

			if (config_info.require_plugin) {
				ast_log(LOG_NOTICE, "Set require_plugin to contact[%s].", contact_id);
				contact->require_plugin = 1;
			}

			if (config_info.require_phonebook) {
				ast_log(LOG_NOTICE, "Set require_plugin to contact[%s].", contact_id);
				contact->require_phonebook = 1;
				set_is_public(contact, is_public);
			}

			ast_sorcery_update(ast_sip_get_sorcery(), contact);

			ao2_cleanup(contact);
		}
	}

	return 0;
}

struct gs_config_data {
	struct ast_sip_contact *contact;
	struct ast_sip_config_info config_info;
};

static void config_data_destroy(void *obj)
{
	struct gs_config_data *config_data = obj;

	if (!config_data) {
		return;
	}

	ao2_ref(config_data->contact, -1);

	return;
}

static struct gs_config_data *gs_config_data_alloc(struct ast_sip_contact *contact, struct ast_sip_config_info config_info)
{
	struct gs_config_data *config_data = ao2_alloc(sizeof(*config_data), config_data_destroy);

	if (!config_data) {
		return NULL;
	}

	config_data->contact = contact;
	memcpy(&config_data->config_info, &config_info, sizeof(struct ast_sip_config_info));

	return config_data;
}

#define APPEND_IF_NONZERO(field, str, name) do {		\
	if (field) {	\
		if (added) {	\
			ast_str_append(str, 0, ",");	\
		}	\
		ast_str_append(str, 0, "%s", name);	\
		added = 1;	\
	}	\
} while (0)

/* X-GS-Config: LDAP-Settings */
int ast_sip_add_gs_config_header(pjsip_rx_data *rdata, pjsip_tx_data *tdata,
	struct ast_sip_config_info config_info)
{
	const pj_str_t header_name = { "X-GS-Config", 11 };
	char gs_config[64] = {0};
	struct ast_str *gs_config_string = NULL;
	int added = 0;

	if (!tdata) {
		return 0;
	}

	if (rdata) {
		pjsip_generic_string_hdr *config_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &header_name, NULL);
		if (!config_hdr) {
			return -1;
		}

		snprintf(gs_config, sizeof(gs_config), "%.*s", (int)config_hdr->hvalue.slen, config_hdr->hvalue.ptr);

		ast_sip_add_header(tdata, "X-GS-Config", gs_config);
	} else {
		gs_config_string = ast_str_create(128);
		if (!gs_config_string) {
			return -1;
		}

		APPEND_IF_NONZERO(config_info.require_ldap_settings, &gs_config_string, "LDAP-Settings");
		APPEND_IF_NONZERO(config_info.require_feature_code_settings, &gs_config_string, "featureCode");
		APPEND_IF_NONZERO(config_info.require_turn_info_settings, &gs_config_string, "turnInfo");
		APPEND_IF_NONZERO(config_info.require_basic_settings, &gs_config_string, "basic");
		APPEND_IF_NONZERO(config_info.require_plugin, &gs_config_string, "plugin");
		APPEND_IF_NONZERO(config_info.require_phonebook, &gs_config_string, "phonebook");

		ast_sip_add_header(tdata, "X-GS-Config", ast_str_buffer(gs_config_string));

		ast_free(gs_config_string);
		gs_config_string = NULL;
	}

	return 0;
}

static int send_options_for_config(void *obj)
{
	RAII_VAR(struct gs_config_data *, config_data, obj, ao2_cleanup);
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	pjsip_tx_data *tdata = NULL;
	const char	*contact_id = NULL;
	const pjsip_hdr *hdr;
	char  *user_id	  = NULL;
	char  *ptr		  = NULL;
	struct ast_sip_config_info config_info;
	struct ast_sip_contact *contact = NULL;
	int is_public = 0;

	contact = config_data->contact;
	if (!contact) {
		return -1;
	}
	memcpy(&config_info, &config_data->config_info, sizeof(struct ast_sip_config_info));

	if (ast_sip_create_request("OPTIONS", NULL, contact->endpoint, NULL, contact, &tdata)) {
		ast_log(LOG_ERROR, "Unable to create ldap OPTIONS to contact %s \n", contact->uri);
		return 0;
	}

	if ((hdr = pjsip_endpt_get_capability(ast_sip_get_pjsip_endpoint(), PJSIP_H_ALLOW, NULL))) {
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)pjsip_hdr_clone(tdata->pool, hdr));
	}

	if ((hdr = pjsip_endpt_get_capability(ast_sip_get_pjsip_endpoint(), PJSIP_H_SUPPORTED, NULL))) {
		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)pjsip_hdr_clone(tdata->pool, hdr));
	}

	ast_sip_add_gs_config_header(NULL, tdata, config_info);

	contact_id = ast_sorcery_object_get_id(contact);
	user_id = ast_strdupa(contact_id);
	if ((ptr = strchr(user_id, ';'))) {
		*ptr = '\0';
	}

	set_is_public(contact, &is_public);

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", user_id);
	ast_sip_add_gs_config_body(endpoint, tdata, is_public, config_info);

	if (ast_sip_send_out_of_dialog_request(tdata, contact->endpoint, -1, NULL, NULL)) {
		ast_log(LOG_ERROR, "Unable to send ldap OPTIONS to contact %s \n", contact->uri);
	}

	return 0;
}

int ast_sip_send_config_to_contact(struct ast_sip_contact *contact, struct ast_sip_config_info config_info)
{
	struct gs_config_data *config_data = NULL;

	if (!contact) {
		return 0;
	}

	config_data = gs_config_data_alloc(contact, config_info);

	if (ast_sip_push_task(NULL, send_options_for_config, config_data)) {
		ast_log(LOG_NOTICE, "push task to send OPTIONS failed!");
		ao2_cleanup(config_data);
	}

	return 0;
}

#if 0
/* For we don't need to decrypt, reserve now */
static void turn_info_memcpy_decrypt(unsigned char *dst, const unsigned char *src, int len, ast_aes_decrypt_key *dcx)
{
	unsigned char lastblock[16] = { 0 };
	int x;
	while(len > 0) {
		ast_aes_decrypt(src, dst, dcx);
		for (x=0;x<16;x++)
			dst[x] ^= lastblock[x];
		memcpy(lastblock, src, sizeof(lastblock));
		dst += 16;
		src += 16;
		len -= 16;
	}
}

static char *decrypt_turn_info_base64(const char *encrypt_turn_info)
{
	ast_aes_encrypt_key dcx;
	char *decrypt_data = NULL;
	char *decrypt_data_base64 = NULL;
	int len = 0;

	decrypt_data = ast_malloc(1024);
	if (!decrypt_data) {
		return NULL;
	}

	decrypt_data_base64 = ast_malloc(1024);
	if (!decrypt_data_base64) {
		ast_free(decrypt_data);
		decrypt_data = NULL;
		return NULL;
	}

	memset(decrypt_data, 0, 1024);
	memset(decrypt_data_base64, 0, 1024);

	ast_aes_set_decrypt_key((unsigned char *)TURN_INFO_ENCRYPT_KEY, &dcx);

	len = ast_base64decode((unsigned char *)decrypt_data_base64, encrypt_turn_info, 1024);

	turn_info_memcpy_decrypt((unsigned char *)decrypt_data, (const unsigned char *)decrypt_data_base64, len, &dcx);

	ast_free(decrypt_data_base64);
	decrypt_data_base64 = NULL;

	return decrypt_data;
}
#endif

static int update_turn_notify_generate_body_content(char *content, int len,
	const char *address, const char *name, const char *password, const char *protocol_type)
{
	char buf[256] = {0};
	char *encrypt_address_base64 = NULL;
	char *encrypt_name_base64 = NULL;
	char *encrypt_password_base64 = NULL;
	char *encrypt_protocol_type_base64 = NULL;
	char *stun_address = NULL;
	char *stun_address_base64 = NULL;

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
		"<TurnInfo>\r\n");

	if (address) {
		memset(buf, 0, sizeof(buf));
		encrypt_address_base64 = encrypt_turn_info_base64(address);
		if (encrypt_address_base64) {
			snprintf(buf, sizeof(buf) - 1, "    <TurnServerAddress>%s</TurnServerAddress>\r\n", encrypt_address_base64);
			strncat(content, buf, len-strlen(content));
			ast_free(encrypt_address_base64);
			encrypt_address_base64 = NULL;
		}
	}

	if (name) {
		memset(buf, 0, sizeof(buf));
		encrypt_name_base64 = encrypt_turn_info_base64(name);
		if (encrypt_name_base64) {
			snprintf(buf, sizeof(buf) - 1, "    <TurnServerUsername>%s</TurnServerUsername>\r\n", encrypt_name_base64);
			strncat(content, buf, len-strlen(content));
			ast_free(encrypt_name_base64);
			encrypt_name_base64 = NULL;
		}
	}

	if (password) {
		memset(buf, 0, sizeof(buf));
		encrypt_password_base64 = encrypt_turn_info_base64(password);
		if (encrypt_password_base64) {
			snprintf(buf, sizeof(buf) - 1, "    <TurnServerPassword>%s</TurnServerPassword>\r\n", encrypt_password_base64);
			strncat(content, buf, len-strlen(content));
			ast_free(encrypt_password_base64);
			encrypt_password_base64 = NULL;
		}
	}

	if (protocol_type) {
		memset(buf, 0, sizeof(buf));
		encrypt_protocol_type_base64 = encrypt_turn_info_base64(protocol_type);
		if (encrypt_protocol_type_base64) {
			snprintf(buf, sizeof(buf) - 1, "    <TurnServerProtocolType>%s</TurnServerProtocolType>\r\n", encrypt_protocol_type_base64);
			strncat(content, buf, len-strlen(content));
			ast_free(encrypt_protocol_type_base64);
			encrypt_protocol_type_base64 = NULL;
		}
	}

	stun_address = ast_sip_get_stun_address();
	if (stun_address) {
		memset(buf, 0, sizeof(buf));
		stun_address_base64 = encrypt_turn_info_base64(stun_address);
		if (stun_address_base64) {
			snprintf(buf, sizeof(buf) - 1, "    <StunServerAddress>%s</StunServerAddress>\r\n", stun_address_base64);
			strncat(content, buf, len-strlen(content));
			ast_free(stun_address_base64);
			stun_address_base64 = NULL;
		}
		ast_free(stun_address);
		stun_address = NULL;
	}

	strncat(content, "</TurnInfo>", len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify turn info body[%s]\n", content);

	return 0;
}

int ast_sip_send_notify_to_update_turn_info(struct ast_sip_contact *contact,
	const char *address, const char *name, const char *password, const char *protocol_type)
{
	pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;
	pj_str_t hdr_name = {"Allow-Events", 12};
	pj_str_t hdr_value = {"conference, presence, dialog, message-summary, call-completion, refer, gs-turn-info", 83};
	pjsip_generic_string_hdr *hdr_allowevent = NULL;
	char content[1024] = { 0 };
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (ast_sip_create_request("NOTIFY", NULL, NULL, NULL, contact, &tdata)) {
		ast_log(LOG_ERROR, "Unable to create notify request to update turn info for contact %s\n", contact->uri);
		return -1;
	}

	update_turn_notify_generate_body_content(content, sizeof(content), address, name, password, protocol_type);

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-grandstream-turn-info+xml",
		.body_text = content
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to update turn info notify request\n");
		return -1;
	}

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "gs-turn-info");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	hdr_allowevent = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_allowevent);

	if (!ast_strlen_zero(contact->endpoint_name)) {
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			contact->endpoint_name);
		ast_log(LOG_DEBUG, "Get endpoint from contact's endpoint_name\n");
	}

#ifdef BUILD_ON_CLOUD_PBX
	/* Added by yfyang, for clouducm support UDP */
	if (contact && !ast_strlen_zero(contact->sbc_endpoint_addr)) {
		ast_sip_add_sbc_endpoint_address_header(contact->sbc_endpoint_addr, tdata);
	}
#endif

    if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "send_notify_to_update_turn_info - Could not send request\n");
		return -1;
	}

	return 0;
}

static void update_turn_info_for_gswave_app(void)
{
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	struct ast_sip_contact *contact = NULL;
	struct ao2_iterator i;
	RAII_VAR(char *, address, NULL, ast_free);
	RAII_VAR(char *, name, NULL, ast_free);
	RAII_VAR(char *, password, NULL, ast_free);
	RAII_VAR(char *, protocol_type, NULL, ast_free);

	ast_log(LOG_DEBUG, "update_turn_info_for_gswave_app begin\n");

	address = ast_sip_get_turn_server_address();
	name = ast_sip_get_turn_user_name();
	password = ast_sip_get_turn_password();
	protocol_type = ast_sip_get_turn_protocol_type();

	contacts = ast_sorcery_retrieve_by_regex(ast_sip_get_sorcery(), "contact", "");
	if (!contacts) {
		ast_log(LOG_WARNING, "update_turn_info_for_gswave_app, can't get any contacts\n");
		return;
	}

	i = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_iterator_next(&i))) {
		if (!contact->support_gs_config) {
			if (contact->gswave_mobile) {
				ast_sip_send_notify_to_update_turn_info(contact, address, name, password, protocol_type);
			}
			ao2_ref(contact, -1);
		} else {
			if (contact->require_turn_info_settings) {
				struct ast_sip_config_info config_info;
				memset(&config_info, 0, sizeof(struct ast_sip_config_info));
				config_info.require_turn_info_settings = 1;
				ast_sip_send_config_to_contact(contact, config_info);
			} else {
				ao2_ref(contact, -1);
			}
		}
	}
	ao2_iterator_destroy(&i);

	return;
}

static void send_notify_for_reload_configue(void)
{
	update_turn_info_for_gswave_app();

	return;
}

static int get_turn_info_from_config(int reload)
{
	struct ast_config *cfg;
	const char *s;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	char *user_name_decrypted = NULL;
	char *password_decrypted = NULL;

	cfg = ast_config_load2("rtp.conf", "rtp", config_flags);
	if (!cfg || cfg == CONFIG_STATUS_FILEUNCHANGED || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_DEBUG, "turn info config is not change\n");
		return -1;
	}

	ast_rwlock_wrlock(&total_turn_session_lock);
	if ((s = ast_variable_retrieve(cfg, "general", "stunaddr"))) {
		strncpy(stun_server_address, s, MAX_TURN_SERVER_LEN - 1);
	} else {
		memset(stun_server_address, 0, MAX_TURN_SERVER_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnaddr"))) {
		strncpy(turn_server_address, s, MAX_TURN_SERVER_LEN - 1);
	} else {
		memset(turn_server_address, 0, MAX_TURN_SERVER_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnusername"))) {
		ucm_secret_decrypt(s, &user_name_decrypted);
		if (!ast_strlen_zero(user_name_decrypted)) {
			strncpy(turn_user_name, user_name_decrypted, MAX_TURN_USER_NAME_LEN - 1);
			ast_free(user_name_decrypted);
			user_name_decrypted = NULL;
		} else {
			ast_log(LOG_ERROR, "Turn user name decrypted failed\n");
		}
	} else {
		memset(turn_user_name, 0, MAX_TURN_USER_NAME_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnpassword"))) {
		ucm_secret_decrypt(s, &password_decrypted);
		if (!ast_strlen_zero(password_decrypted)) {
			strncpy(turn_password, password_decrypted, MAX_TURN_PASSWORD_LEN - 1);
			ast_free(password_decrypted);
			password_decrypted = NULL;
		} else {
			ast_log(LOG_ERROR, "Turn password decrypted failed\n");
		}
	} else {
		memset(turn_password, 0, MAX_TURN_PASSWORD_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnprotocol"))) {
		strncpy(turn_protocol_type, s, MAX_TURN_PROTOCOL_TYPE_LEN - 1);
	} else {
		memset(turn_protocol_type, 0, MAX_TURN_PROTOCOL_TYPE_LEN);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	ast_config_destroy(cfg);

	ast_log(LOG_DEBUG, "get_turn_info_from_config reload:%d end\n", reload);

	return 0;
}

static int is_turn_info_changed(int reload)
{
	int res = 0;
	struct ast_config *cfg;
	const char *s;
	char *user_name_decrypted = NULL;
	char *password_decrypted = NULL;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };

	if (ast_sip_get_max_media_nat_call_count()) {
		ast_log(LOG_DEBUG, "It's enable RC media NAT traversal service, don't need update turn settings for apply change\n");
		return 0;
	}

	cfg = ast_config_load2("rtp.conf", "rtp", config_flags);
	if (!cfg || cfg == CONFIG_STATUS_FILEUNCHANGED || cfg == CONFIG_STATUS_FILEINVALID) {
		ast_log(LOG_DEBUG, "turn info config is not change\n");
		return 0;
	}

	ast_rwlock_wrlock(&total_turn_session_lock);
	if ((s = ast_variable_retrieve(cfg, "general", "stunaddr"))) {
		if (strcmp(stun_server_address, s)) {
			res |= 1;
		}
		strncpy(stun_server_address, s, MAX_TURN_SERVER_LEN - 1);
	} else {
		if (strlen(stun_server_address)) {
			res |= 1;
		}
		memset(stun_server_address, 0, MAX_TURN_SERVER_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnaddr"))) {
		if (strcmp(turn_server_address, s)) {
			res |= 1;
		}
		strncpy(turn_server_address, s, MAX_TURN_SERVER_LEN - 1);
	} else {
		if (strlen(turn_server_address)) {
			res |= 1;
		}
		memset(turn_server_address, 0, MAX_TURN_SERVER_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnusername"))) {
		ucm_secret_decrypt(s, &user_name_decrypted);
		if (!ast_strlen_zero(user_name_decrypted)) {
			if (strcmp(turn_user_name, user_name_decrypted)) {
				res |= 1;
			}
			strncpy(turn_user_name, user_name_decrypted, MAX_TURN_USER_NAME_LEN - 1);
			ast_free(user_name_decrypted);
			user_name_decrypted = NULL;
		} else {
			ast_log(LOG_ERROR, "Turn user name decrypted failed\n");
		}
	} else {
		if (strlen(turn_user_name)) {
			res |= 1;
		}
		memset(turn_user_name, 0, MAX_TURN_USER_NAME_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnpassword"))) {
		ucm_secret_decrypt(s, &password_decrypted);
		if (!ast_strlen_zero(password_decrypted)) {
			if (strcmp(turn_password, password_decrypted)) {
				res |= 1;
			}
			strncpy(turn_password, password_decrypted, MAX_TURN_PASSWORD_LEN - 1);
			ast_free(password_decrypted);
			password_decrypted = NULL;
		} else {
			ast_log(LOG_ERROR, "Turn password decrypted failed\n");
		}
	} else {
		if (strlen(turn_password)) {
			res |= 1;
		}
		memset(turn_password, 0, MAX_TURN_PASSWORD_LEN);
	}
	if ((s = ast_variable_retrieve(cfg, "general", "turnprotocol"))) {
		if (strcmp(turn_protocol_type, s)) {
			res |= 1;
		}
		strncpy(turn_protocol_type, s, MAX_TURN_PROTOCOL_TYPE_LEN - 1);
	} else {
		if (strlen(turn_protocol_type)) {
			res |= 1;
		}
		memset(turn_protocol_type, 0, MAX_TURN_PROTOCOL_TYPE_LEN);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	ast_config_destroy(cfg);

	ast_log(LOG_DEBUG, "is_turn_info_changed res:%d\n", res);

	return res;
}

static void update_turn_info_from_db(void)
{
	char server_address[MAX_TURN_SERVER_LEN] = { 0 };
	char stun_address[MAX_TURN_SERVER_LEN] = { 0 };
	char user_name[MAX_TURN_USER_NAME_LEN] = { 0 };
	char password[MAX_TURN_PASSWORD_LEN*3] = { 0 };
	char protocol_type[MAX_TURN_PROTOCOL_TYPE_LEN] = { 0 };
	char limit_time[MAX_TURN_LIMIT_TIME_LEN] = { 0 };
	char day_total_limit_time[MAX_TURN_LIMIT_TIME_LEN] = { 0 };
	char month_total_limit_time[MAX_TURN_LIMIT_TIME_LEN] = { 0 };
	char max_registrar[MAX_TURN_MAX_REGISTRAR_LEN] = { 0 };
	char max_media_nat_call[MAX_TURN_MAX_MEDIA_NAT_CALL_LEN] = { 0 };
	char webrtc_trunk[MAX_SUPPORT_WEBRTC_TRUNK_LEN] = { 0 };
	char *user_name_decrypted = NULL;
	char *password_decrypted = NULL;

	ast_rwlock_wrlock(&total_turn_session_lock);
	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_ADDRESS_KEY, server_address, MAX_TURN_SERVER_LEN - 1);
	if (strlen(server_address)) {
		strcpy(turn_server_address, server_address);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_STUN_ADDRESS_KEY, stun_address, MAX_TURN_SERVER_LEN - 1);
	if (strlen(stun_address)) {
		strcpy(stun_server_address, stun_address);
	}


	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_USERNAME_KEY, user_name, sizeof(user_name) - 1);
	ucm_secret_decrypt(user_name, &user_name_decrypted);
	if (!ast_strlen_zero(user_name_decrypted)) {
		strncpy(turn_user_name, user_name_decrypted, MAX_TURN_USER_NAME_LEN - 1);
		ast_free(user_name_decrypted);
		user_name_decrypted = NULL;
	} else {
		ast_log(LOG_ERROR, "Turn user name decrypted failed\n");
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_PASSWORD_KEY, password, sizeof(password) - 1);
	ucm_secret_decrypt(password, &password_decrypted);
	if (!ast_strlen_zero(password_decrypted)) {
		strncpy(turn_password, password_decrypted, MAX_TURN_PASSWORD_LEN - 1);
		ast_free(password_decrypted);
		password_decrypted = NULL;
	} else {
		ast_log(LOG_ERROR, "Turn password decrypted failed\n");
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_PROTOCOLTYPE_KEY, protocol_type, MAX_TURN_PROTOCOL_TYPE_LEN - 1);
	if (strlen(protocol_type)) {
		strcpy(turn_protocol_type, protocol_type);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_LIMITTIME_KEY, limit_time, MAX_TURN_LIMIT_TIME_LEN - 1);
	if (strlen(limit_time)) {
		turn_limit_time = atoi(limit_time);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_DAY_TOTAL_LIMITTIME_KEY, day_total_limit_time, MAX_TURN_LIMIT_TIME_LEN - 1);
	if (strlen(day_total_limit_time)) {
		turn_day_total_limit_time = atoi(day_total_limit_time);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_MONTH_TOTAL_LIMITTIME_KEY, month_total_limit_time, MAX_TURN_LIMIT_TIME_LEN - 1);
	if (strlen(month_total_limit_time)) {
		turn_month_total_limit_time = atoi(month_total_limit_time);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_MAXREGISTRAR_KEY, max_registrar, MAX_TURN_MAX_REGISTRAR_LEN - 1);
	if (strlen(max_registrar)) {
		turn_max_registrar = atoi(max_registrar);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, TURN_INFO_MAX_MEDIA_NAT_CALL_KEY, max_media_nat_call, MAX_TURN_MAX_MEDIA_NAT_CALL_LEN - 1);
	if (strlen(max_media_nat_call)) {
		turn_max_media_nat_call = atoi(max_media_nat_call);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, SUPPORT_WEBRTC_TRUNK, webrtc_trunk, MAX_SUPPORT_WEBRTC_TRUNK_LEN - 1);
	if (strlen(webrtc_trunk)) {
		support_webrtc_trunk = atoi(webrtc_trunk);
	}
	ast_rwlock_unlock(&total_turn_session_lock);

	return;
}

static void update_plugin_info(void)
{
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	struct ast_sip_contact *contact = NULL;
	struct ast_sip_config_info config_info;
	struct ao2_iterator i;

	ast_log(LOG_DEBUG, "update_plugin_info begin\n");

	contacts = ast_sorcery_retrieve_by_regex(ast_sip_get_sorcery(), "contact", "");
	if (!contacts) {
		ast_log(LOG_WARNING, "update_plugin_info, can't get any contacts\n");
		return;
	}

	memset(&config_info, 0, sizeof(struct ast_sip_config_info));
	config_info.require_plugin = 1;

	i = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_iterator_next(&i))) {
		if (contact->require_plugin) {
			ast_sip_send_config_to_contact(contact, config_info);
		} else {
			ao2_ref(contact, -1);
		}
	}
	ao2_iterator_destroy(&i);

	return;
}


static int ami_update_plugin_info(struct mansession *s, const struct message *m)
{
	const char *plugin_market = NULL;
	const char *plugin_thirdparty = NULL;

	plugin_market = astman_get_header(m, "support_wave_market");
	if (!ast_strlen_zero(plugin_market)) {
		update_plugin_market(plugin_market);
	} else {
		update_plugin_market(NULL);
	}

	plugin_thirdparty = astman_get_header(m, "support_wave_thirdparty");
	if (!ast_strlen_zero(plugin_thirdparty)) {
		update_plugin_thirdparty(plugin_thirdparty);
	} else {
		update_plugin_thirdparty(NULL);
	}

	ast_log(LOG_DEBUG, "ami_update_plugin_info plugin_market:%s, plugin_thirdparty:%s\n", plugin_market, plugin_thirdparty);

	astman_send_response(s, m, "Success", "Update plugin info success");

	update_plugin_info();

	return AMI_SUCCESS;
}

static int ami_update_turn_info(struct mansession *s, const struct message *m)
{
#ifndef BUILD_ON_CLOUD_PBX
	int res = 0;
#endif
	int turn_count = 0;
	const char *turn_server_address = NULL;
	const char *turn_user_name = NULL;
	const char *turn_password = NULL;
	const char *turn_protocol_type = NULL;
	const char *turn_count_str = NULL;
	const char *turn_limit_time_str = NULL;
	const char *turn_day_total_limit_time_str = NULL;
	const char *turn_month_total_limit_time_str = NULL;
	const char *turn_max_registrar_str = NULL;
	const char *webrtc_trunk_str = NULL;

	turn_count_str = astman_get_header(m, "TurnSessionCount");
	if (turn_count_str) {
		ast_log(LOG_DEBUG, "Get TURN info: turn_count:%s\n", turn_count_str);
		turn_count = atoi(turn_count_str);
		update_total_turn_session_count(turn_count);
	}
	update_max_media_nat_call_count(turn_count_str);

	turn_server_address = astman_get_header(m, "TurnServerAddress");
	if (!ast_strlen_zero(turn_server_address)) {
		update_turn_server_address(turn_server_address);
		update_stun_server_address(turn_server_address);
	} else {
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_ADDRESS_KEY);
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_STUN_ADDRESS_KEY);
	}

	turn_user_name = astman_get_header(m, "TurnServerUsername");
	if (!ast_strlen_zero(turn_user_name)) {
		update_turn_user_name(turn_user_name);
	} else {
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_USERNAME_KEY);
	}

	turn_password = astman_get_header(m, "TurnServerPassword");
	if (!ast_strlen_zero(turn_password)) {
		update_turn_password(turn_password);
	} else {
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_PASSWORD_KEY);
	}

	turn_protocol_type = astman_get_header(m, "TrunServerType");
	if (!ast_strlen_zero(turn_protocol_type)) {
		update_turn_protocol_type(turn_protocol_type);
	} else {
		ast_db_del(TURN_INFO_DB_FAMILY, TURN_INFO_PROTOCOLTYPE_KEY);
	}

	turn_limit_time_str = astman_get_header(m, "TurnLimitTime");
	if (!ast_strlen_zero(turn_limit_time_str)) {
		ast_log(LOG_DEBUG, "Get TURN limit time:%s\n", turn_limit_time_str);
		update_turn_limit_time((char *)turn_limit_time_str);
	} else {
		update_turn_limit_time(NULL);
	}

	turn_day_total_limit_time_str = astman_get_header(m, "PerDayTurnLimitTime");
	if (!ast_strlen_zero(turn_limit_time_str)) {
		ast_log(LOG_DEBUG, "Get every day TURN limit time:%s\n", turn_day_total_limit_time_str);
		update_turn_day_total_limit_time((char *)turn_day_total_limit_time_str);
		ast_cdr_set_has_limit_time(1);
		ast_cdr_set_day_total_limit_time(atoi(turn_day_total_limit_time_str));
	} else {
		update_turn_day_total_limit_time(NULL);
		ast_cdr_set_has_limit_time(0);
		ast_cdr_set_day_total_limit_time(0);
	}

	turn_month_total_limit_time_str = astman_get_header(m, "PerMonthTurnLimitTime");
	if (!ast_strlen_zero(turn_month_total_limit_time_str)) {
		ast_log(LOG_DEBUG, "Get every month TURN limit time:%s\n", turn_month_total_limit_time_str);
		update_turn_month_total_limit_time((char *)turn_month_total_limit_time_str);
		ast_cdr_set_has_month_limit_time(1);
	} else {
		update_turn_month_total_limit_time(NULL);
		ast_cdr_set_has_month_limit_time(0);
	}

	turn_max_registrar_str = astman_get_header(m, "TurnMaxRegistrar");
	if (!ast_strlen_zero(turn_max_registrar_str)) {
		ast_log(LOG_DEBUG, "Get TURN max registrar:%s\n", turn_max_registrar_str);
		update_turn_max_registrar((char *)turn_max_registrar_str);
	} else {
		update_turn_max_registrar(NULL);
	}

	webrtc_trunk_str = astman_get_header(m, "WebRTCTrunk");
	if (!ast_strlen_zero(webrtc_trunk_str)) {
		ast_log(LOG_DEBUG, "Get WebRTCTrunk:%s\n", webrtc_trunk_str);
		update_support_webrtc_trunk(webrtc_trunk_str);
	} else {
		update_support_webrtc_trunk(NULL);
	}

	if (ast_strlen_zero(turn_server_address) && ast_strlen_zero(turn_user_name) && 
		ast_strlen_zero(turn_password) && ast_strlen_zero(turn_protocol_type)) {
		ast_log(LOG_DEBUG, "AMI clean up the turn info, we use config instead\n");
		get_turn_info_from_config(0);
		update_max_media_nat_call_count(NULL);
	}

#ifndef BUILD_ON_CLOUD_PBX
	res = ast_avs_control(NULL, AVS_CONTROL_CONFIG, NULL, 0);
	if (res == R_SUCCESS) {
		astman_send_response(s, m, "Success", "Update turn info success");
	} else {
		ast_log(LOG_ERROR, "Update AVS turn info failed\n");
		astman_send_error(s, m, "Update turn info failed");
	}
#endif

	update_turn_info_for_gswave_app();

	return AMI_SUCCESS;
}

static void init_turn_info(void)
{
	ast_rwlock_init(&total_turn_session_lock);
	ast_rwlock_init(&used_turn_session_lock);
	ast_rwlock_init(&max_used_turn_session_lock);

	memset(max_used_date, 0, sizeof(max_used_date));
	memset(alert_exceed_threshold_date, 0, sizeof(alert_exceed_threshold_date));
	memset(alert_exceed_total_date, 0, sizeof(alert_exceed_total_date));

	turn_limit_time = 0;
	turn_max_registrar = 0;
	turn_max_media_nat_call = 0;

	turn_server_address = ast_malloc(MAX_TURN_SERVER_LEN);
	if (turn_server_address) {
		memset(turn_server_address, 0, MAX_TURN_SERVER_LEN);
	}
	stun_server_address = ast_malloc(MAX_TURN_SERVER_LEN);
	if (stun_server_address) {
		memset(stun_server_address, 0, MAX_TURN_SERVER_LEN);
	}
	turn_user_name = ast_malloc(MAX_TURN_USER_NAME_LEN);
	if (turn_user_name) {
		memset(turn_user_name, 0, MAX_TURN_USER_NAME_LEN);
	}
	turn_password = ast_malloc(MAX_TURN_PASSWORD_LEN);
	if (turn_password) {
		memset(turn_password, 0, MAX_TURN_PASSWORD_LEN);
	}
	turn_protocol_type = ast_malloc(MAX_TURN_PROTOCOL_TYPE_LEN);
	if (turn_protocol_type) {
		memset(turn_protocol_type, 0, MAX_TURN_PROTOCOL_TYPE_LEN);
	}

	get_turn_info_from_config(0);

	update_turn_info_from_db();

	update_total_turn_session_count(ast_sip_get_turn_session_count());

	ast_manager_register_xml("UpdateTurnInfo", EVENT_FLAG_SYSTEM, ami_update_turn_info);

	return;
}

static void init_plugin_info(void)
{
	char market[4] = { 0 };
	char thirdparty[4] = { 0 };

	ast_db_get(TURN_INFO_DB_FAMILY, PLUGIN_MARKET_KEY, market, MAX_TURN_SERVER_LEN - 1);
	if (strlen(market)) {
		plugin_market = atoi(market);
	}

	ast_db_get(TURN_INFO_DB_FAMILY, PLUGIN_THIRDPARTY_KEY, thirdparty, MAX_TURN_SERVER_LEN - 1);
	if (strlen(thirdparty)) {
		plugin_thirdparty = atoi(thirdparty);
	}

	return;
}

static void destory_turn_info(void)
{
	ast_rwlock_destroy(&total_turn_session_lock);
	ast_rwlock_destroy(&used_turn_session_lock);
	ast_rwlock_destroy(&max_used_turn_session_lock);

	if (turn_server_address) {
		ast_free(turn_server_address);
		turn_server_address = NULL;
	}
	if (stun_server_address) {
		ast_free(stun_server_address);
		stun_server_address = NULL;
	}
	if (turn_user_name) {
		ast_free(turn_user_name);
		turn_user_name = NULL;
	}
	if (turn_password) {
		ast_free(turn_password);
		turn_password = NULL;
	}
	if (turn_protocol_type) {
		ast_free(turn_protocol_type);
		turn_protocol_type = NULL;
	}

	ast_manager_unregister("UpdateTurnInfo");

	return;
}

unsigned int ast_sip_get_total_turn_session_count(void)
{
	unsigned int count = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	count = total_turn_session_count;
	ast_rwlock_unlock(&total_turn_session_lock);

	return count;
}

unsigned int ast_sip_get_used_turn_session_count(void)
{
	unsigned int count = 0;

	ast_rwlock_wrlock(&used_turn_session_lock);
	count = used_turn_session_count;
	ast_rwlock_unlock(&used_turn_session_lock);

	return count;
}

unsigned int ast_sip_get_max_media_nat_call_count(void)
{
	unsigned int count = 0;

	ast_rwlock_wrlock(&total_turn_session_lock);
	count = turn_max_media_nat_call;
	ast_rwlock_unlock(&total_turn_session_lock);

	return count;
}

unsigned int ast_sip_increase_used_turn_session_count(unsigned int increase_count)
{
	unsigned int count = 0;
	char buffer[128] = {0};

	ast_rwlock_wrlock(&used_turn_session_lock);
	used_turn_session_count = used_turn_session_count + increase_count;
	count = used_turn_session_count;

	if (count <= ast_sip_get_total_turn_session_count()) {
		manager_event(EVENT_FLAG_CALL, "ActiveCount", "ActiveNatCalls: %d\r\n", count);
	}

	ast_rwlock_unlock(&used_turn_session_lock);

	snprintf(buffer, sizeof(buffer), "nvram set :ActiveNatCalls=%d", count);
	ast_safe_system2(buffer);

	return count;
}

unsigned int ast_sip_decrease_used_turn_session_count(unsigned int decrease_count)
{
	unsigned int count = 0;
	char buffer[128] = {0};

	ast_rwlock_wrlock(&used_turn_session_lock);
	if (used_turn_session_count > decrease_count) {
		used_turn_session_count = used_turn_session_count - decrease_count;
	} else {
		used_turn_session_count = 0;
	}
	count = used_turn_session_count;

	if (count <= ast_sip_get_total_turn_session_count()) {
		manager_event(EVENT_FLAG_CALL, "ActiveCount", "ActiveNatCalls: %d\r\n", count);
	}

	ast_rwlock_unlock(&used_turn_session_lock);

	snprintf(buffer, sizeof(buffer), "nvram set :ActiveNatCalls=%d", count);
	ast_safe_system2(buffer);

	return count;
}

unsigned int ast_sip_set_day_max_used_turn_session_count(unsigned int max_count)
{
	ast_rwlock_wrlock(&max_used_turn_session_lock);
	max_used_turn_session_count = max_count;
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return max_count;
}

unsigned int ast_sip_get_day_max_used_turn_session_count(void)
{
	unsigned int count = 0;

	ast_rwlock_wrlock(&max_used_turn_session_lock);
	count = max_used_turn_session_count;
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return count;
}

char *ast_sip_set_max_used_date(char *date)
{
	ast_rwlock_wrlock(&max_used_turn_session_lock);
	strncpy(max_used_date, date, sizeof(max_used_date) - 1);
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return max_used_date;
}

char *ast_sip_get_max_used_date(void)
{
	char *date = NULL;

	ast_rwlock_wrlock(&max_used_turn_session_lock);
	date = ast_strdup(max_used_date);
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return date;
}

char *ast_sip_set_alert_exceed_threshold_date(char *date)
{
	ast_rwlock_wrlock(&max_used_turn_session_lock);
	strncpy(alert_exceed_threshold_date, date, sizeof(alert_exceed_threshold_date) - 1);
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return alert_exceed_threshold_date;
}

char *ast_sip_get_alert_exceed_threshold_date(void)
{
	char *date = NULL;

	ast_rwlock_wrlock(&max_used_turn_session_lock);
	date = ast_strdup(alert_exceed_threshold_date);
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return date;
}

char *ast_sip_set_alert_exceed_total_date(char *date)
{
	ast_rwlock_wrlock(&max_used_turn_session_lock);
	strncpy(alert_exceed_total_date, date, sizeof(alert_exceed_threshold_date) - 1);
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return alert_exceed_total_date;
}

char *ast_sip_get_alert_exceed_total_date(void)
{
	char *date = NULL;

	ast_rwlock_wrlock(&max_used_turn_session_lock);
	date = ast_strdup(alert_exceed_total_date);
	ast_rwlock_unlock(&max_used_turn_session_lock);

	return date;
}

/* Added by yfyang, get external ip from STUN */
int ast_sip_nat_get_external_ip(char *addr, int len)
{
	if (!addr || !len) {
		return -1;
	}

	ast_sip_get_nvram_value(":ext_ip", addr, len);
	if (!strlen(addr)) {
		ast_log(LOG_DEBUG, "No ext_ip in pvalue.");
		return -1;
	}

	ast_trim_blanks(addr);
	return 0;
}

int ast_sip_dialog_rewrite_remote_uri(pjsip_dialog *dlg, pj_sockaddr *rem_addr)
{
	int ret = 0;
	pjsip_sip_uri *uri = NULL;
	char remote_name[PJ_INET6_ADDRSTRLEN] = {0};
	int remote_port = 0;

	if (!dlg || !rem_addr) {
		return -1;
	}

	pjsip_dlg_inc_lock(dlg);
	if (!dlg->remote.contact) {
		pjsip_dlg_dec_lock(dlg);
		return -1;
	}

	if (!(PJSIP_URI_SCHEME_IS_SIP(dlg->remote.contact->uri) || PJSIP_URI_SCHEME_IS_SIPS(dlg->remote.contact->uri))) {
		pjsip_dlg_dec_lock(dlg);
		return -1;
	}

	uri = pjsip_uri_get_uri(dlg->remote.contact->uri);
	if (!uri) {
		pjsip_dlg_dec_lock(dlg);
		return -1;
	}

	pj_sockaddr_print(rem_addr, remote_name, sizeof(remote_name), 0);
	remote_port = pj_sockaddr_get_port(rem_addr);
	ast_log(LOG_DEBUG, "dlg->host: %.*s, dlg->port:%d, transport->host:%s, transport->port:%d",
						(int)uri->host.slen, uri->host.ptr, uri->port, 
						remote_name, remote_port);
	if (pj_strcmp2(&uri->host, remote_name) || uri->port != remote_port) {
		pj_strdup2(dlg->pool, &uri->host, remote_name);
		uri->port = remote_port;
		ret = 1;
	}
	pjsip_dlg_dec_lock(dlg);

	return ret;
}

/*Start added by zyxia for Bug 188323 - Wave_Android_1.0.6.2_Suggest to show only one line after disconnecting the network and then reconnecting during ringing*/
int ast_sip_dialog_rewrite_remote_uri_and_transport(struct ast_sip_endpoint *endpoint,pjsip_dialog *dlg, pj_sockaddr *rem_addr)
{
	int ret = 0;
	pjsip_sip_uri *uri = NULL;
	char remote_name[PJ_INET6_ADDRSTRLEN] = {0};
	int  remote_port = 0;
	pjsip_tpselector selector = { .type = PJSIP_TPSELECTOR_NONE, };

	if (!dlg || !rem_addr || !dlg->remote.contact) {
	return -1;
	}

	pjsip_dlg_inc_lock(dlg);
	if (!(PJSIP_URI_SCHEME_IS_SIP(dlg->remote.contact->uri) || PJSIP_URI_SCHEME_IS_SIPS(dlg->remote.contact->uri))) {
		pjsip_dlg_dec_lock(dlg);
		return -1;
	}

	uri = pjsip_uri_get_uri(dlg->remote.contact->uri);
	if (!uri) {
		pjsip_dlg_dec_lock(dlg);
		return -1;
	}

	pj_sockaddr_print(rem_addr, remote_name, sizeof(remote_name), 0);
	remote_port = pj_sockaddr_get_port(rem_addr);
	ast_log(LOG_DEBUG, "dlg->host: %.*s, dlg->port:%d, transport->host:%s, transport->port:%d,dlg->tp_select_type:%d",
			(int)uri->host.slen, uri->host.ptr, uri->port, 
			remote_name, remote_port,dlg->tp_sel.type);
	if (pj_strcmp2(&uri->host, remote_name) || uri->port != remote_port) {
		pj_strdup2(dlg->pool, &uri->host, remote_name);
		uri->port = remote_port;
		ret = 1;
	}
	ast_sip_dlg_set_transport(endpoint, dlg, &selector);
	pjsip_dlg_dec_lock(dlg);
	return ret;
}

int ast_sip_update_request_uri_and_dest_info(pjsip_transaction *tsx, pj_sockaddr *rem_addr)
{
	int ret = 0;
	pjsip_sip_uri *uri = NULL;
	char remote_name[PJ_INET6_ADDRSTRLEN] = {0};
	int  remote_port = 0;


	if (!tsx || !rem_addr || !tsx->last_tx || !tsx->last_tx->msg || tsx->last_tx->msg->type != PJSIP_REQUEST_MSG) {
		ast_log(LOG_ERROR, "ast_sip_update_request_uri_and_dest_info some tsx paras is NUll\n");
		return -1;
	}

	if (!(PJSIP_URI_SCHEME_IS_SIP(tsx->last_tx->msg->line.req.uri) || PJSIP_URI_SCHEME_IS_SIPS(tsx->last_tx->msg->line.req.uri))) {
		ast_log(LOG_ERROR, "ast_sip_update_request_uri_and_dest_info uri is not sip\n");
		return -1;
	}

	uri = pjsip_uri_get_uri(tsx->last_tx->msg->line.req.uri);
	if (!uri) {
		ast_log(LOG_ERROR, "ast_sip_update_request_uri_and_dest_info  can not get uri\n");
		return -1;
	}

	pj_sockaddr_print(rem_addr, remote_name, sizeof(remote_name), 0);
	remote_port = pj_sockaddr_get_port(rem_addr);

	if (pj_strcmp2(&uri->host, remote_name) || uri->port != remote_port) {
		pj_strdup2(tsx->last_tx->pool, &uri->host, remote_name);
		uri->port = remote_port;
		pj_strdup2(tsx->last_tx->pool, &tsx->last_tx->dest_info.name, remote_name);
		tsx->last_tx->dest_info.addr.entry[0].addr_len = sizeof(rem_addr);
		pj_memcpy(&tsx->last_tx->dest_info.addr.entry[0].addr, rem_addr,sizeof(rem_addr));
		ret =1 ;
	}

	return ret;
}
/*End added by zyxia for Bug 188323 - Wave_Android_1.0.6.2_Suggest to show only one line after disconnecting the network and then reconnecting during ringing*/

static const char *ucm_prot_event_string(ucm_prot_event event)
{
	switch(event) {
		case UCM_PROT_EVENT_CONF_TALK:
			return "conftalk";
		default:
			return "none";
	}
}

static struct ast_json *ucm_prot_build_notify_msg_data(ucm_prot_notify_msg_data *msg_data)
{
	struct ast_json *json_data = NULL;
	RAII_VAR(struct ast_json *, json_ctx, NULL, ast_json_unref);

	json_data = ast_json_object_create();
	if (!json_data) {
		return NULL;
	}

	ast_json_object_set(json_data, "event", ast_json_string_create(ucm_prot_event_string(msg_data->event)));
	ast_json_object_set(json_data, "seq", ast_json_integer_create(msg_data->seq));
	if (!ast_strlen_zero(msg_data->context)) {
		json_ctx = ast_json_load_string(msg_data->context, NULL);
		if (!json_ctx) {
			ast_json_unref(json_data);
			return NULL;
		}

		ast_json_object_update(json_data, json_ctx);
	}

	return json_data;
}

static struct ast_json *ucm_prot_build_msg_data(ucm_prot_msg *msg)
{
	struct ast_json *json_obj = NULL;

	if (!msg || !msg->data) {
		return NULL;
	}

	switch (msg->method) {
		case UCM_PROT_NOTIFY:
			json_obj = ucm_prot_build_notify_msg_data(msg->data);
			break;
		default:
			break;
	}

	return json_obj;
}

struct ast_json *ast_sip_ucm_prot_build_msg(ucm_prot_msg *msg)
{
	struct ast_json *json_msg = NULL;

	if (!msg) {
		return NULL;
	}

	json_msg = ast_json_object_create();
	if (!json_msg) {
		return NULL;
	}

	/*Header*/
	ast_json_object_set(json_msg, "typ", ast_json_integer_create(msg->type));
	ast_json_object_set(json_msg, "act", ast_json_integer_create(msg->method));
	ast_json_object_set(json_msg, "tid", ast_json_integer_create(msg->tid));
	/*Message*/
	if (msg->data) {
		ast_json_object_set(json_msg, "msg", ucm_prot_build_msg_data(msg));
	}

	return json_msg;
}

int ast_sip_ucm_prot_build_send_buf(ucm_prot_msg *msg, char **buf, unsigned int *buf_len)
{
	char *msg_buf = NULL;
	unsigned short payload_len = 0;
	unsigned int len = 0;
	RAII_VAR(struct ast_json *, json_object, NULL, ast_json_unref);
	RAII_VAR(char *, json_str, NULL, ast_json_free);

	if (!msg) {
		return -1;
	}

	json_object = ast_sip_ucm_prot_build_msg(msg);
	if (!json_object) {
		return -1;
	}
	json_str = ast_json_dump_string(json_object);
	if (!json_str) {
		return -1;
	}

	payload_len = strlen(json_str);
	/*Magic Code (3 Bytes) + Version (1 Bytes) + Length (2 Bytes) + Message*/
	len = UCM_PROT_MAGIC_CODE_SIZE + UCM_PROT_VERSION_SIZE + UCM_PROT_PAYLOAD_SIZE + payload_len;
	msg_buf = (char *)ast_calloc(1, len);
	if (!msg_buf) {
		return -1;
	}

	len = 0;
	/*Magic Code*/
	memcpy(msg_buf, UCM_PROT_MAGIC_CODE, UCM_PROT_MAGIC_CODE_SIZE);
	len += UCM_PROT_MAGIC_CODE_SIZE;
	/*Version*/
	*(msg_buf+len) = UCM_PROT_VERSION;
	len += UCM_PROT_VERSION_SIZE;
	/*Length*/
	*(unsigned short *)(msg_buf+len) = htons(payload_len);
	len += UCM_PROT_PAYLOAD_SIZE;
	/*Payload*/
	memcpy(msg_buf+len, json_str, payload_len);
	len += payload_len;

	ast_log(LOG_DEBUG, "send buf: %.*s", payload_len, json_str);

	*buf = msg_buf;
	*buf_len = len;
	return 0;
}

void ast_sip_ucm_prot_send_buf_print(const char *buf, unsigned int len)
{
	unsigned char version = 0;
	unsigned short payload_len = 0;

	if (!strncasecmp(buf, UCM_PROT_MAGIC_CODE, UCM_PROT_MAGIC_CODE_SIZE)) {
		version = *(buf + UCM_PROT_MAGIC_CODE_SIZE);
		payload_len = ntohs(*(unsigned short *)(buf + UCM_PROT_MAGIC_CODE_SIZE + UCM_PROT_VERSION_SIZE));

		ast_log(LOG_DEBUG, "ucm prot send buf: version:%d, payload length:%d, payload:%.*s",
				version, payload_len, payload_len, buf+UCM_PROT_PAYLOAD_OFFSET);
	}
}

int ast_sip_ucm_prot_check(const char *buf, unsigned int len)
{
	if (len < UCM_PROT_HEAD_SIZE || strncasecmp(buf, UCM_PROT_MAGIC_CODE, UCM_PROT_MAGIC_CODE_SIZE)) {
		return 0;
	}

	return 1;
}

void ast_sip_ucm_prot_reset_payload_size(char *buf, unsigned int len)
{
	memset(&buf[UCM_PROT_PAYLOAD_SIZE_OFFSET], 0, UCM_PROT_PAYLOAD_SIZE);
}

/* Added by yfyang, for websocket remote address */
static void init_websocket_remote(void)
{
	ast_db_deltree(WEBSOCKET_REMOTE, NULL);
}

void ast_sip_init_ldap_xml_buffer(void)
{
	const char *filename0 = "/data/gs_config/ldap_eth0.xml";
	const char *filename1 = "/data/gs_config/ldap_eth1.xml";
	const char *filenamep = "/data/gs_config/ldap_public.xml";
	const char *filename  = NULL;
	FILE *file = NULL;
	unsigned int len = 0;
	char  mode_value[8]   = {0};
	char  host_value[256] = {0};
	int   mode = 0;

	ast_sip_get_nvram_value(":WsProxyHost", host_value, sizeof(host_value));
	if (strlen(host_value)) {
		file = fopen(filenamep, "r");
		if (!file) {
			return;
		}

		ast_rwlock_wrlock(&config_xml_lock);
		memset(ldap_xml_public_buffer, 0, sizeof(ldap_xml_public_buffer));
		len = (unsigned int)fread(ldap_xml_public_buffer, 1, sizeof(ldap_xml_public_buffer), file);
		ast_rwlock_unlock(&config_xml_lock);

		if (!len) {
			ast_log(LOG_NOTICE, "Read file [%s] failed.", filenamep);
		}

		fclose(file);
	}

	ast_sip_get_nvram_value("231", mode_value, sizeof(mode_value));
	if (!strlen(mode_value)) {
		ast_log(LOG_NOTICE, "P231 not set.");
		return;
	}

	mode = atoi(mode_value);
	if (mode == 0) {
		/* Route mode */
		filename = filename1;
	} else if (mode == 1) {
		/* Switch mode */
		filename = filename0;
	} else if (mode == 2) {
		/* Dual mode */
		filename = filename1;
	} else {
		ast_log(LOG_NOTICE, "P231=[%s] is invalid.", mode_value);
		return;
	}

	file = fopen(filename, "r");
	if (!file) {
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	memset(ldap_xml_buffer, 0, sizeof(ldap_xml_buffer));
	len = (unsigned int)fread(ldap_xml_buffer, 1, sizeof(ldap_xml_buffer), file);
	ast_rwlock_unlock(&config_xml_lock);

	if (!len) {
		ast_log(LOG_NOTICE, "Read file [%s] failed.", filename);
	}

	fclose(file);
	return;
}

int ast_sip_check_ldap_xml_value(void)
{
	if (strlen(ldap_xml_buffer)) {
		return 1;
	}

	return 0;
}

void ast_sip_init_phonebook_xml_buffer(void)
{
	const char *filename0 = "/data/gs_config/phonebook_eth0.xml";
	const char *filename1 = "/data/gs_config/phonebook_eth1.xml";
	const char *filenamep = "/data/gs_config/phonebook_public.xml";
	const char *filename  = NULL;
	FILE *file = NULL;
	unsigned int len = 0;
	char  mode_value[8]   = {0};
	int   mode = 0;
	char phonebook_public_xml[MAX_PHONEBOOK_XML_LEN] = { 0 };

	file = fopen(filenamep, "r");
	if (file) {
		ast_rwlock_wrlock(&config_xml_lock);
		memset(phonebook_xml_public_buffer, 0, sizeof(phonebook_xml_public_buffer));
		len = (unsigned int)fread(phonebook_xml_public_buffer, 1, sizeof(phonebook_xml_public_buffer), file);
		ast_rwlock_unlock(&config_xml_lock);

		if (!len) {
			ast_log(LOG_NOTICE, "Read file [%s] failed.", filenamep);
		}

		create_xml_for_phonebook(phonebook_public_xml, sizeof(phonebook_public_xml) - 1);
		memset(phonebook_public_url, 0, sizeof(phonebook_public_url));
		read_url_from_xml(phonebook_public_xml, phonebook_public_url);

		fclose(file);
	}

	ast_sip_get_nvram_value("231", mode_value, sizeof(mode_value));
	if (!strlen(mode_value)) {
		ast_log(LOG_NOTICE, "P231 not set.");
		return;
	}

	mode = atoi(mode_value);
	if (mode == 0) {
		/* Route mode */
		filename = filename1;
	} else if (mode == 1) {
		/* Switch mode */
		filename = filename0;
	} else if (mode == 2) {
		/* Dual mode */
		filename = filename1;
	} else {
		ast_log(LOG_NOTICE, "P231=[%s] is invalid.", mode_value);
		return;
	}

	file = fopen(filename, "r");
	if (!file) {
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	memset(phonebook_xml_buffer, 0, sizeof(phonebook_xml_buffer));
	len = (unsigned int)fread(phonebook_xml_buffer, 1, sizeof(phonebook_xml_buffer), file);
	ast_rwlock_unlock(&config_xml_lock);

	if (!len) {
		ast_log(LOG_NOTICE, "Read file [%s] failed.", filename);
	}

	fclose(file);
	return;
}

int ast_sip_check_phonebook_xml_value(void)
{
	if (strlen(phonebook_xml_buffer)) {
		return 1;
	}

	return 0;
}

void ast_sip_init_feature_code_xml_buffer(void)
{
	const char *filename = "/data/gs_config/feature_code.xml";
	FILE *file = NULL;
	unsigned int len = 0;

	file = fopen(filename, "r");
	if (!file) {
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	memset(feature_code_xml_buffer, 0, sizeof(feature_code_xml_buffer));
	len = (unsigned int)fread(feature_code_xml_buffer, 1, sizeof(feature_code_xml_buffer), file);
	ast_rwlock_unlock(&config_xml_lock);

	if (!len) {
		ast_log(LOG_NOTICE, "Read file [%s] failed.", filename);
	}

	fclose(file);
	return;
}

int ast_sip_check_feature_code_xml_value(void)
{
	if (strlen(feature_code_xml_buffer)) {
		return 1;
	}

	return 0;
}

void ast_sip_init_basic_xml_buffer(void)
{
	const char *filename = "/data/gs_config/basic.xml";
	FILE *file = NULL;
	unsigned int len = 0;

	file = fopen(filename, "r");
	if (!file) {
		return;
	}

	ast_rwlock_wrlock(&config_xml_lock);
	memset(basic_xml_buffer, 0, sizeof(basic_xml_buffer));
	len = (unsigned int)fread(basic_xml_buffer, 1, sizeof(basic_xml_buffer), file);
	ast_rwlock_unlock(&config_xml_lock);

	if (!len) {
		ast_log(LOG_NOTICE, "Read file [%s] failed.", filename);
	}

	fclose(file);
	return;
}

int ast_sip_check_basic_xml_value(void)
{
	if (strlen(basic_xml_buffer)) {
		return 1;
	}

	return 0;
}

static int update_callflip_state_body(char *content, int len, struct ast_sip_callflip_state *callflip_state)
{
	char buf[512] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<state version=\"1\">\r\n"
	"  <item name=\"call-state\">\r\n"
	"    <callFlip>\r\n");

	if (!ast_strlen_zero(callflip_state->enable)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <enable>%s</enable>\r\n", callflip_state->enable);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->remote_id)) {
		char remote_id[MAX_REMOTE_NAME_LEN];
		ast_xml_escape(callflip_state->remote_id, remote_id, sizeof(remote_id));
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <remoteID>%s</remoteID>\r\n", remote_id);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->remote_name)) {
		//added by zyxia for Bug 212817,escape xml
		char remote_name[MAX_REMOTE_NAME_LEN];
		ast_xml_escape(callflip_state->remote_name, remote_name, sizeof(remote_name));
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <remoteName>%s</remoteName>\r\n", remote_name);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->call_type)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <callType>%s</callType>\r\n", callflip_state->call_type);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->media_type)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <mediaType>%s</mediaType>\r\n", callflip_state->media_type);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->conf_subject)) {
		//added by zyxia for Bug 212817,escape xml
		char conf_subject[MAX_CONF_SUBJECT_LEN];
		ast_xml_escape(callflip_state->conf_subject, conf_subject, sizeof(conf_subject));
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <confSubject>%s</confSubject>\r\n", conf_subject);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->replace_channel_id)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <replaceChannelID>%s</replaceChannelID>\r\n", callflip_state->replace_channel_id);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(callflip_state->conf_link)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <confLink>%s</confLink>\r\n", callflip_state->conf_link);
		strncat(content, buf, len-strlen(content));
	}

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    </callFlip>\r\n"
	"  </item>\r\n"
	"</state>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs state body[%s]\n", content);

	return 0;
}

static int update_parking_state_body(char *content, int len, struct ast_sip_parking_state *parking_state)
{
	char buf[512] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<state version=\"1\">\r\n"
	"  <item name=\"parking-state\">\r\n"
	"    <parking>\r\n");

	if (!ast_strlen_zero(parking_state->enable)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <enable>%s</enable>\r\n",parking_state->enable);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(parking_state->parking_num)) {
		char parknum[256];
		ast_xml_escape(parking_state->parking_num, parknum, sizeof(parknum));
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <parkingNum>%s</parkingNum>\r\n", parknum);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(parking_state->parkee_channel_id)) {
		char channelId[MAX_CHANNEL_ID];
		ast_xml_escape(parking_state->parkee_channel_id, channelId, sizeof(channelId));
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <channelId>%s</channelId>\r\n", channelId);
		strncat(content, buf, len-strlen(content));
	}

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    </parking>\r\n"
	"  </item>\r\n"
	"</state>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs state body[%s]\n", content);

	return 0;
}


/*return 0 dont send*/
/*return 1 need send,defalut*/
int ast_sip_check_if_need_sendcallflip(struct ast_sip_contact *contact,struct ast_sip_callflip_state *callflip_state)
{
	if (!ast_strlen_zero(callflip_state->enable) && !strcasecmp(callflip_state->enable,"yes")) {
		if (!ast_strlen_zero(callflip_state->channel_id)) {
			struct ast_channel *chan= NULL;
			chan = ast_channel_get_by_name(callflip_state->channel_id);
			if (!chan) {
				ast_log(LOG_ERROR, "Channel does not exist:%s \n",callflip_state->channel_id);
				return 0;
			}

			struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
			if(!channel){
				ast_log(LOG_ERROR, "Sip Channel pvt does not exist:%s \n",callflip_state->channel_id);
				ast_channel_unref(chan);
				return 0;
			}

			struct ast_sip_session *session = channel->session;
			ast_channel_unref(chan);

			if (session && session->inv_session && session->inv_session->dlg && session->inv_session->dlg->remote.contact){
				pjsip_sip_uri *sip_contact;
				char remote_uri[PJSIP_MAX_URL_SIZE]={0};
				sip_contact = pjsip_uri_get_uri(session->inv_session->dlg->remote.contact->uri);
				pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, sip_contact, remote_uri, sizeof(remote_uri));
				if (contact->uri && remote_uri[0] != '\0' && !strcasecmp(contact->uri, remote_uri)) {
					/*Start added by zyxia for Bug240451 ,after callflip,send the trunk info notify to the ternaml in calling*/
					if((session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)
						&& !ast_strlen_zero(callflip_state->conf_link) && !ast_strlen_zero(callflip_state->replace_channel_id)) {
						ast_log(LOG_DEBUG, "Need send notify trunk info to uri:%s\n",contact->uri);
						struct ast_trunk_info trunk_info;
						memset(&trunk_info, 0, sizeof(struct ast_trunk_info));
						strncpy(trunk_info.conf_link, callflip_state->conf_link, sizeof(trunk_info.conf_link) - 1);
						strncpy(trunk_info.channel_id, callflip_state->replace_channel_id, sizeof(trunk_info.channel_id) - 1);
						ast_sip_session_send_notify_to_update_gs_state(session, AST_GS_STATE_TYPE_TRUNK_INFO, &trunk_info);
					}
					/*End added by zyxia for Bug 240451 ,after callflip,send the trunk info notify to the ternaml in calling*/
					ast_log(LOG_DEBUG, "Dont need send callflip to uri:%s\n",contact->uri);
					return 0;
				}
			}
		}
	}
	return 1;
}


/*Start added by zyxia for Bug 291797,2024.3.14*/

/*return 0 dont send*/
/*return 1 need send,defalut*/
int ast_sip_check_if_need_send_parkingstate(struct ast_sip_contact *contact, struct ast_sip_parking_state *parking_state)
{
	if (contact && parking_state && parking_state->parkee_chan) {
		struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(parking_state->parkee_chan);
		struct ast_sip_session *session = channel->session;
		if (session && session->inv_session && session->inv_session->dlg && session->inv_session->dlg->remote.contact){
			pjsip_sip_uri *sip_contact;
			char remote_uri[PJSIP_MAX_URL_SIZE]={0};
			sip_contact = pjsip_uri_get_uri(session->inv_session->dlg->remote.contact->uri);
			pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, sip_contact, remote_uri, sizeof(remote_uri));
			if (contact->uri && remote_uri[0] != '\0' && !strcasecmp(contact->uri, remote_uri)) {
				if(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE){
					ast_log(LOG_DEBUG, "Need send parking State %s  to uri:%s\n",parking_state->enable,contact->uri);
					return 1;
				}
			}
		}
	}

	return 0;
}
/*End added by zyxia for Bug 291797,2024.3.14*/

int ast_sip_send_notify_to_update_gs_state(struct ast_sip_contact *contact,
	enum ast_sip_gs_state_type type, void *data)
{
	pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;
	pj_str_t hdr_name = {"Allow-Events", 12};
	pj_str_t hdr_value = {"conference, presence, dialog, message-summary, call-completion, refer, x-gs-state", 81};
	pjsip_generic_string_hdr *hdr_allowevent = NULL;
	char content[1024] = { 0 };
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (!contact->support_gs_state) {
		return -1;
	}

	switch (type) {
		case GS_STATE_TYPE_CALLFLIP:
			update_callflip_state_body(content, sizeof(content), data);
			break;
		case GS_STATE_TYPE_PARKING:
			update_parking_state_body(content, sizeof(content), data);
			break;

		default:
			ast_log(LOG_ERROR, "Can't support type:%d update gs state for contact %s\n", type, contact->uri);
			return -1;
	}

	if (ast_sip_create_request("NOTIFY", NULL, NULL, NULL, contact, &tdata)) {
		ast_log(LOG_ERROR, "Unable to create notify request to update gs state for contact %s\n", contact->uri);
		return -1;
	}

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-state+xml",
		.body_text = content
	};


	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "x-gs-state");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	hdr_allowevent = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_allowevent);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to update turn info notify request\n");
		return -1;
	}

	if (!ast_strlen_zero(contact->endpoint_name)) {
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			contact->endpoint_name);
		ast_log(LOG_DEBUG, "Get endpoint from contact's endpoint_name\n");
	}

#ifdef BUILD_ON_CLOUD_PBX
	/* Added by yfyang, for clouducm support UDP */
	if (contact && !ast_strlen_zero(contact->sbc_endpoint_addr)) {
		ast_sip_add_sbc_endpoint_address_header(contact->sbc_endpoint_addr, tdata);
	}
#endif
	if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "ast_sip_send_notify_to_update_gs_state - Could not send request\n");
		return -1;
	}

	return 0;
}

static int create_delete_call_log_body(char *content, int len, struct ast_sip_gs_control_data *gs_control_data)
{
	char buf[256] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<control version=\"1\">\r\n"
	"  <action name=\"DeleteCallLogs\">\r\n");

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    <target>%s</target>\r\n", gs_control_data->target);
	strncat(content, buf, len-strlen(content));

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "  </action>\r\n"
	"</control>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs control delete call log body[%s]\n", content);

	return 0;
}

static int create_delete_contacts_body(char *content, int len)
{
	char buf[256] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<control version=\"1\">\r\n"
	"  <action name=\"DeleteContacts\">\r\n");

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "  </action>\r\n"
	"</control>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs control delete contacts body[%s]\n", content);

	return 0;
}

int ast_sip_send_notify_for_gs_control(struct ast_sip_contact *contact, struct ast_sip_gs_control_data *gs_control_data)
{
	pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;
	char content[512] = { 0 };
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (!contact || !gs_control_data) {
		ast_log(LOG_ERROR, "ast_sip_send_notify_for_gs_control error params\n");
		return -1;
	}

	if (!contact->support_gs_control) {
		ast_log(LOG_DEBUG, "ast_sip_send_notify_for_gs_control can't support gs control, contact uri:%s\n", contact->uri);
		return 0;
	}

	ast_log(LOG_DEBUG, "ast_sip_send_notify_for_gs_control, type:%d, contact %s\n", gs_control_data->type, contact->uri);

	switch (gs_control_data->type) {
		case GS_CONTROL_TYPE_DELETE_CALL_LOGS:
			create_delete_call_log_body(content, sizeof(content), gs_control_data);
			break;
		case GS_CONTROL_TYPE_DELETE_CONTACTS:
			create_delete_contacts_body(content, sizeof(content));
			break;
		default:
			ast_log(LOG_ERROR, "Can't support action:%d for gs control, contact %s\n", gs_control_data->type, contact->uri);
			return -1;
	}

	if (ast_sip_create_request("NOTIFY", NULL, NULL, NULL, contact, &tdata)) {
		ast_log(LOG_ERROR, "Unable to create notify request for gs control, contact %s\n", contact->uri);
		return -1;
	}

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-control+xml",
		.body_text = content
	};

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "x-gs-control");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to update turn info notify request\n");
		return -1;
	}

	if (!ast_strlen_zero(contact->endpoint_name)) {
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
			contact->endpoint_name);
		ast_log(LOG_DEBUG, "Get endpoint from contact's endpoint_name\n");
	}

#ifdef BUILD_ON_CLOUD_PBX
	/* Added by yfyang, for clouducm support UDP */
	if (contact && !ast_strlen_zero(contact->sbc_endpoint_addr)) {
		ast_sip_add_sbc_endpoint_address_header(contact->sbc_endpoint_addr, tdata);
	}
#endif

	if (ast_sip_send_request(tdata, NULL, endpoint, NULL, NULL)) {
		ast_log(LOG_ERROR, "ast_sip_send_notify_to_update_gs_state - Could not send request\n");
		return -1;
	}

	return 0;
}

/*Start added by zyxia for Callflip,2021.11.25*/
/*!
 * \internal
 * \brief Find an callflip info by name
 */
 static void callflip_user_destructor(void *obj)
{
	struct callflip_user *user = obj;

	ast_log(LOG_NOTICE, "callflip_user destructor for [%s][%p]\n", user->name, user);
	ast_string_field_free_memory(user);
}

struct callflip_user *ast_sip_callflip_find_user(const char *name)
{
	struct callflip_user tmp_user = {
		.name = name,
	};

	return ao2_find(callflip_users, &tmp_user, OBJ_POINTER);
}

 static int callflip_user_hash(const void *obj, const int flags)
{
	const struct callflip_user *user = obj;

	return ast_str_case_hash(user->name);
}

 static int callflip_user_cmp(void *obj, void *arg, int flags)
{
	struct callflip_user *user = obj, *user2 = arg;
	return !strcasecmp(user->name, user2->name) ? CMP_MATCH | CMP_STOP : 0;
}

int ast_sip_callflip_build_user(struct ast_sip_callflip_state *callflip_state, const char *cat)
{
	RAII_VAR(struct callflip_user *, user, NULL, ao2_cleanup);
	int existing_user = 0;

	if ((user = ast_sip_callflip_find_user(cat))) {
		existing_user = 1;
	} else if ((user = ao2_alloc(sizeof(*user), callflip_user_destructor))) {
		if (ast_string_field_init(user, 32)) {
			return -1;
		}
		ast_string_field_set(user, name, cat);
	} else {
		return -1;
	}

	ao2_lock(user);
	ast_copy_string(user->enable,  callflip_state->enable, sizeof( callflip_state->enable));
	ast_copy_string(user->remote_id,  callflip_state->remote_id, sizeof( callflip_state->remote_id));
	ast_copy_string(user->remote_name,	callflip_state->remote_name, sizeof( callflip_state->remote_name));
	ast_copy_string(user->call_type,  callflip_state->call_type, sizeof( callflip_state->call_type));
	ast_copy_string(user->media_type,  callflip_state->media_type, sizeof( callflip_state->media_type));
	ast_copy_string(user->channel_id,  callflip_state->channel_id, sizeof( callflip_state->channel_id));
	ast_copy_string(user->conf_link,  callflip_state->conf_link, sizeof( callflip_state->conf_link));
	ast_copy_string(user->replace_channel_id,  callflip_state->replace_channel_id, sizeof( callflip_state->replace_channel_id));

	ao2_unlock(user);

	if (!existing_user) {
		ao2_link(callflip_users, user);
	}

	return 0;
}

int ast_sip_callflip_update_user_media_type(const char *cat, const char *media_type)
{
	RAII_VAR(struct callflip_user *, user, NULL, ao2_cleanup);

	if ((user = ast_sip_callflip_find_user(cat))) {
		ao2_lock(user);
		ast_copy_string(user->media_type,  media_type, sizeof( media_type));
		ao2_unlock(user);
	} else {
		return -1;
	}
	return 0;
}

/*End added by zyxia for Callflip,2021.11.25*/

int ast_sip_get_callflip_enable(void)
{
	return enable_callflip;
}

static int send_talk_event_notify(void *data)
{
	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);
	pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;

	if (!session) {
		return -1;
	}

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Could not create talk event notify request, inv session is disconnected\n");
		return -1;
	}

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create talk event notify request\n");
		return -1;
	}

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "talk");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	ast_sip_add_header(tdata, "Subscription-State", "active");

	ast_sip_session_send_request(session, tdata);

	return 0;
}

static int ami_talk_action(struct mansession *s, const struct message *m)
{
	const char *channel_name = NULL;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *session = NULL;
	RAII_VAR(struct ast_channel *, chan, NULL, ao2_cleanup);

	ast_log(LOG_DEBUG, "ami_talk_action begin\n");

	channel_name = astman_get_header(m, "Channel");
	if (ast_strlen_zero(channel_name)) {
		ast_log(LOG_WARNING, "ami_talk_action can't get channel name\n");
		astman_send_error(s, m, "Channel talk failed, can't get channel");
		return AMI_SUCCESS;
	}

	chan = ast_channel_get_by_name(channel_name);
	if (!chan) {
		ast_log(LOG_WARNING, "ami_talk_action can't get chan from channel name:%s\n", channel_name);
		astman_send_error(s, m, "Channel talk failed, can't get asterisk chan through channel name");
		return AMI_SUCCESS;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "ami_talk_action can't process channel type:%s\n", ast_channel_tech(chan)->type);
		astman_send_error(s, m, "Channel talk failed, can't process chan as not pjsip type");
		return AMI_SUCCESS;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		ast_log(LOG_WARNING, "ami_talk_action can't get sip session\n");
		astman_send_error(s, m, "Channel talk failed, can't get sip session");
		return AMI_SUCCESS;
	}

	session = chan_pvt->session;

	if (!session->allow_talk_event) {
		ast_log(LOG_WARNING, "[session:%p] Peer terminal can't support talk event, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		astman_send_error(s, m, "Channel talk failed, can't allow talk event");
		return AMI_SUCCESS;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_talk_event_notify, session)) {
		ast_log(LOG_ERROR, "[session:%p] Could not push talk event notify request to task, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(session);
		astman_send_error(s, m, "Channel talk failed, can't push talk event notify to task");
		return AMI_SUCCESS;
	}

	astman_send_response(s, m, "Success", "Channel talk success");

	return AMI_SUCCESS;
}

static int send_hold_event_notify(void *data)
{
	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);
	pjsip_tx_data *tdata = NULL;
	pjsip_event_hdr *sub_event = NULL;

	if (!session) {
		return -1;
	}

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Could not create hold event notify request, inv session is disconnected\n");
		return -1;
	}

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create hold event notify request\n");
		return -1;
	}

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "hold");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	ast_sip_add_header(tdata, "Subscription-State", "active");

	ast_sip_session_send_request(session, tdata);

	return 0;
}

static int ami_hold_action(struct mansession *s, const struct message *m)
{
	const char *channel_name = NULL;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *session = NULL;
	RAII_VAR(struct ast_channel *, chan, NULL, ao2_cleanup);

	channel_name = astman_get_header(m, "Channel");
	if (ast_strlen_zero(channel_name)) {
		ast_log(LOG_WARNING, "ami_hold_action can't get channel name\n");
		astman_send_error(s, m, "Channel hold failed, can't get channel");
		return AMI_SUCCESS;
	}

	chan = ast_channel_get_by_name(channel_name);
	if (!chan) {
		ast_log(LOG_WARNING, "ami_hold_action can't get chan from channel name:%s\n", channel_name);
		astman_send_error(s, m, "Channel hold failed, can't get asterisk chan through channel name");
		return AMI_SUCCESS;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "ami_hold_action can't process channel type:%s\n", ast_channel_tech(chan)->type);
		astman_send_error(s, m, "Channel hold failed, can't process chan as not pjsip type");
		return AMI_SUCCESS;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		ast_log(LOG_WARNING, "ami_hold_action can't get sip session\n");
		astman_send_error(s, m, "Channel hold failed, can't get sip session");
		return AMI_SUCCESS;
	}

	session = chan_pvt->session;

	if (!session->allow_hold_event) {
		ast_log(LOG_WARNING, "[session:%p] Peer terminal can't support hold event, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		astman_send_error(s, m, "Channel hold failed, can't allow hold event");
		return AMI_SUCCESS;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_hold_event_notify, session)) {
		ao2_cleanup(session);
	}

	astman_send_response(s, m, "Success", "Channel hold success");

	return AMI_SUCCESS;
}

static int ami_resume_action(struct mansession *s, const struct message *m)
{
	const char *channel_name = NULL;
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *session = NULL;
	RAII_VAR(struct ast_channel *, chan, NULL, ao2_cleanup);

	channel_name = astman_get_header(m, "Channel");
	if (ast_strlen_zero(channel_name)) {
		ast_log(LOG_WARNING, "ami_resume_action can't get channel name\n");
		astman_send_error(s, m, "Channel resume failed, can't get channel");
		return AMI_SUCCESS;
	}

	chan = ast_channel_get_by_name(channel_name);
	if (!chan) {
		ast_log(LOG_WARNING, "ami_resume_action can't get chan from channel name:%s\n", channel_name);
		astman_send_error(s, m, "Channel resume failed, can't get asterisk chan through channel name");
		return AMI_SUCCESS;
	}

	if (strcmp(ast_channel_tech(chan)->type, "PJSIP")) {
		ast_log(LOG_WARNING, "ami_resume_action can't process channel type:%s\n", ast_channel_tech(chan)->type);
		astman_send_error(s, m, "Channel resume failed, can't process chan as not pjsip type");
		return AMI_SUCCESS;
	}

	chan_pvt = ast_channel_tech_pvt(chan);
	if (!chan_pvt || !chan_pvt->session) {
		ast_log(LOG_WARNING, "ami_resume_action can't get sip session\n");
		astman_send_error(s, m, "Channel resume failed, can't get sip session");
		return AMI_SUCCESS;
	}

	session = chan_pvt->session;

	if (!session->allow_talk_event) {
		ast_log(LOG_WARNING, "[session:%p] Peer terminal can't support talk event, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		astman_send_error(s, m, "Channel talk failed, can't allow talk event");
		return AMI_SUCCESS;
	}

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_talk_event_notify, session)) {
		ao2_cleanup(session);
	}

	astman_send_response(s, m, "Success", "Channel resume success");

	return AMI_SUCCESS;
}

const char * const ast_gs_control_type_strmap[][1] = {
	{ /* 0 NULL          */      ""          },
	{ /* 1 GS_CONTROL_TYPE_DELETE_CALL_LOGS  */      "DeleteCallLogs"   },
	{ /* 2 GS_CONTROL_TYPE_DELETE_CONTACTS */        "DeleteContacts"   },
};

static enum ast_gs_control_type gs_control_str_2_type(const char *type)
{
	int i = 0;

	for (i = 0; i < GS_CONTROL_TYPE_MAX; i++) {
		if (!strcmp(type, ast_gs_control_type_strmap[i][0])) {
			return (enum ast_gs_control_type)i;
		}
	}

	return GS_CONTROL_TYPE_NULL;
}

static int ami_gs_control_notify(struct mansession *s, const struct message *m)
{
	const char *type = NULL;
	const char *extension = NULL;
	const char *target = NULL;
	RAII_VAR(struct ao2_container *, contacts, NULL, ao2_cleanup);
	struct ast_sip_contact *contact = NULL;
	struct ao2_iterator i;
	struct ast_sip_gs_control_data gs_control_data;

	type = astman_get_header(m, "Type");
	if (ast_strlen_zero(type)) {
		ast_log(LOG_WARNING, "ami_gs_control_notify can't get type\n");
		astman_send_error(s, m, "GS control notify failed, can't get type");
		return AMI_SUCCESS;
	}

	extension = astman_get_header(m, "Extension");
	if (ast_strlen_zero(extension)) {
		ast_log(LOG_WARNING, "ami_gs_control_notify can't get extension\n");
		astman_send_error(s, m, "GS control notify failed, can't get extension");
		return AMI_SUCCESS;
	}

	target = astman_get_header(m, "Target");
	if (ast_strlen_zero(extension)) {
		ast_log(LOG_WARNING, "ami_gs_control_notify can't get target\n");
		astman_send_error(s, m, "GS control notify failed, can't get target");
		return AMI_SUCCESS;
	}

	ast_log(LOG_DEBUG, "ami_gs_control_notify start\n");

	memset(&gs_control_data, 0, sizeof(struct ast_sip_gs_control_data));
	gs_control_data.type = gs_control_str_2_type(type);
	strncpy(gs_control_data.extension, extension, sizeof(gs_control_data.extension) - 1);
	strncpy(gs_control_data.target, target, sizeof(gs_control_data.target) - 1);

	contacts = ast_sorcery_retrieve_by_regex(ast_sip_get_sorcery(), "contact", "");
	if (!contacts) {
		ast_log(LOG_WARNING, "ami_gs_control_notify, can't get any contacts\n");
		astman_send_error(s, m, "GS control notify failed, can't get contacts");
		return AMI_SUCCESS;
	}

	i = ao2_iterator_init(contacts, 0);
	while ((contact = ao2_iterator_next(&i))) {
		if (!strcasecmp(contact->aor, extension)) {
			ast_sip_send_notify_for_gs_control(contact, &gs_control_data);
		}
		ao2_ref(contact, -1);
	}
	ao2_iterator_destroy(&i);

	astman_send_response(s, m, "Success", "Channel resume success");

	return AMI_SUCCESS;
}
/*Start aded by zyxia for 311299.311187,2024.8.8*/
static int action_wavelogout(struct mansession *s, const struct message *m)
{
	const char *extension = astman_get_header(m, "extension");
	const char *uniqueId = astman_get_header(m, "uniqueId");
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);
	struct sip_waveapp_info* app_info = NULL;

	if (ast_strlen_zero(extension)) {
		astman_send_error(s, m, "No field[extension] specified");
		return -1;
	}

	ast_log(LOG_DEBUG, "Enter action_wavelogout, extension=%s, uniqueId=%s\n", extension,uniqueId);

	endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint",
		extension);
	if (!endpoint) {
		astman_send_error(s, m, "Unable to retrieve endpoint");
		ast_log(LOG_ERROR, "Unable to retrieve endpoint %s\n", extension);
		return -1;

	}

	if(ast_strlen_zero(uniqueId)){
		ast_sip_clearall_waveapp_info(endpoint);
		ast_db_deltree(GSWAVE_DB_FAMILY, extension);
	}
	else{
		app_info = ast_sip_find_waveapp_info(endpoint,uniqueId);
		if(app_info != NULL){
			ast_sip_remove_waveinfo_from_database(endpoint->aors, app_info);
			ast_sip_clear_waveapp_info(endpoint,uniqueId);
		}
	}
	astman_send_response(s, m, "Success", "Wave dataremove success");

	return AMI_SUCCESS;
}
/*End aded by zyxia for 311299.311187,2024.8.8*/

#endif

long ast_sip_threadpool_queue_size(void)
{
	return ast_threadpool_queue_size(sip_threadpool);
}

#ifdef TEST_FRAMEWORK
AST_TEST_DEFINE(xml_sanitization_end_null)
{
	char sanitized[8];

	switch (cmd) {
	case TEST_INIT:
		info->name = "xml_sanitization_end_null";
		info->category = "/res/res_pjsip/";
		info->summary = "Ensure XML sanitization works as expected with a long string";
		info->description = "This test sanitizes a string which exceeds the output\n"
			"buffer size. Once done the string is confirmed to be NULL terminated.";
		return AST_TEST_NOT_RUN;
	case TEST_EXECUTE:
		break;
	}

	ast_sip_sanitize_xml("aaaaaaaaaaaa", sanitized, sizeof(sanitized));
	if (sanitized[7] != '\0') {
		ast_test_status_update(test, "Sanitized XML string is not null-terminated when it should be\n");
		return AST_TEST_FAIL;
	}

	return AST_TEST_PASS;
}

AST_TEST_DEFINE(xml_sanitization_exceeds_buffer)
{
	char sanitized[8];

	switch (cmd) {
	case TEST_INIT:
		info->name = "xml_sanitization_exceeds_buffer";
		info->category = "/res/res_pjsip/";
		info->summary = "Ensure XML sanitization does not exceed buffer when output won't fit";
		info->description = "This test sanitizes a string which before sanitization would\n"
			"fit within the output buffer. After sanitization, however, the string would\n"
			"exceed the buffer. Once done the string is confirmed to be NULL terminated.";
		return AST_TEST_NOT_RUN;
	case TEST_EXECUTE:
		break;
	}

	ast_sip_sanitize_xml("<><><>&", sanitized, sizeof(sanitized));
	if (sanitized[7] != '\0') {
		ast_test_status_update(test, "Sanitized XML string is not null-terminated when it should be\n");
		return AST_TEST_FAIL;
	}

	return AST_TEST_PASS;
}
#endif

/*!
 * \internal
 * \brief Reload configuration within a PJSIP thread
 */
static int reload_configuration_task(void *obj)
{
	ast_res_pjsip_reload_configuration();
	ast_res_pjsip_init_options_handling(1);
	ast_sip_initialize_dns();

#ifdef GRANDSTREAM_NETWORKS
	get_feature_code(1);
	prase_direct_media_outbound_prefix_conf_to_list(1);
	prase_mcm_conference_config(1);

	if (is_turn_info_changed(1)) {
		update_turn_info_from_db();
		send_notify_for_reload_configue();
	}

	ast_sip_acctk_unlink_marked();
#endif

	return 0;
}

static int unload_pjsip(void *data)
{
	/*
	 * These calls need the pjsip endpoint and serializer to clean up.
	 * If they're not set, then there's nothing to clean up anyway.
	 */
	if (ast_pjsip_endpoint && serializer_pool[0]) {
		ast_res_pjsip_cleanup_options_handling();
		ast_res_pjsip_cleanup_message_filter();
		ast_sip_destroy_distributor();
		ast_sip_destroy_transport_management();
		ast_res_pjsip_destroy_configuration();
		ast_sip_destroy_system();
		ast_sip_destroy_global_headers();
		ast_sip_unregister_service(&supplement_module);
		ast_sip_destroy_transport_events();
	}

	if (monitor_thread) {
		stop_monitor_thread();
		monitor_thread = NULL;
	}

	if (memory_pool) {
		/* This mimics the behavior of pj_pool_safe_release
		 * which was introduced in pjproject 2.6.
		 */
		pj_pool_t *temp_pool = memory_pool;

		memory_pool = NULL;
		pj_pool_release(temp_pool);
	}

	ast_pjsip_endpoint = NULL;

	if (caching_pool.lock) {
		ast_pjproject_caching_pool_destroy(&caching_pool);
	}

#ifdef GRANDSTREAM_NETWORKS
	destory_mcm_conf_number();
	destory_turn_info();

	ast_rwlock_destroy(&config_xml_lock);
#endif

	pj_shutdown();

	return 0;
}

static int load_pjsip(void)
{
	const unsigned int flags = 0; /* no port, no brackets */
	pj_status_t status;

#ifdef GRANDSTREAM_NETWORKS
#ifdef BUILD_ON_CLOUD_PBX
	/*Optimize the pj cache pool size*/
	pjsip_cfg()->tp.max_count = 1000;
	pjsip_cfg()->ua.max_count = ((16*1024)-1);
	pjsip_cfg()->tsx.max_count = ((16*1024)-1);
#endif
#endif

	/* The third parameter is just copied from
	 * example code from PJLIB. This can be adjusted
	 * if necessary.
	 */
	ast_pjproject_caching_pool_init(&caching_pool, NULL, 1024 * 1024);
	if (pjsip_endpt_create(&caching_pool.factory, "SIP", &ast_pjsip_endpoint) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Failed to create PJSIP endpoint structure. Aborting load\n");
		goto error;
	}

	/* PJSIP will automatically try to add a Max-Forwards header. Since we want to control that,
	 * we need to stop PJSIP from doing it automatically
	 */
	remove_request_headers(ast_pjsip_endpoint);

	memory_pool = pj_pool_create(&caching_pool.factory, "SIP", 1024, 1024, NULL);
	if (!memory_pool) {
		ast_log(LOG_ERROR, "Failed to create memory pool for SIP. Aborting load\n");
		goto error;
	}

	if (!pj_gethostip(pj_AF_INET(), &host_ip_ipv4)) {
#ifdef GRANDSTREAM_NETWORKS
		/* Record original local ipv4 address, this will change to float IP for HA */
		pj_sockaddr_print(&host_ip_ipv4, local_ip_ipv4_string, sizeof(local_ip_ipv4_string), flags);

		if (!ast_sockaddr_voip_vlan_addr_get(local_vlan_ip_string, sizeof(local_vlan_ip_string))) {
			memset(local_vlan_ip_string, 0, sizeof(local_vlan_ip_string));
		}
#endif
		if (HA_ROLE_ACTIVE == ast_ha_get_role() && (ast_get_redundance_flag() != 1)){
			host_ip_ipv4.ipv4.sin_addr.s_addr = get_ha_float_ip();
		}
		pj_sockaddr_print(&host_ip_ipv4, host_ip_ipv4_string, sizeof(host_ip_ipv4_string), flags);
		ast_verb(3, "Local IPv4 address determined to be: %s\n", host_ip_ipv4_string);
	}

	if (!pj_gethostip(pj_AF_INET6(), &host_ip_ipv6)) {
		pj_sockaddr_print(&host_ip_ipv6, host_ip_ipv6_string, sizeof(host_ip_ipv6_string), flags);
		ast_verb(3, "Local IPv6 address determined to be: %s\n", host_ip_ipv6_string);
	}

	pjsip_tsx_layer_init_module(ast_pjsip_endpoint);
	pjsip_ua_init_module(ast_pjsip_endpoint, NULL);

	monitor_continue = 1;
	status = pj_thread_create(memory_pool, "SIP", (pj_thread_proc *) &monitor_thread_exec,
			NULL, PJ_THREAD_DEFAULT_STACK_SIZE * 2, 0, &monitor_thread);
	if (status != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Failed to start SIP monitor thread. Aborting load\n");
		goto error;
	}

	return AST_MODULE_LOAD_SUCCESS;

error:
	return AST_MODULE_LOAD_DECLINE;
}

/*
 * This is a place holder function to ensure that pjmedia_strerr() is at
 * least directly referenced by this module to ensure that the loader
 * linker will link to the function.  If a module only indirectly
 * references a function from another module, such as a callback parameter
 * to a function, the loader linker has been known to miss the link.
 */
void never_called_res_pjsip(void);
void never_called_res_pjsip(void)
{
	pjmedia_strerror(0, NULL, 0);
}

#ifdef GRANDSTREAM_NETWORKS
static void cli_cachpool(int fd, const char *cachpool, const char *fmt, ...)
{
	va_list ap;
	char str_cachpool[1024] = {0};

	va_start(ap, fmt);
	vsnprintf(str_cachpool, sizeof(str_cachpool) - 1, fmt, ap);
	va_end(ap);

	ast_cli(fd, "%s\n", str_cachpool);
}

static char *cli_show_cachpool(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show cachpool";
		e->usage = "Usage: pjsip show cachpool\n";
		return NULL;
	case CLI_GENERATE:
		break;
	}

	if (a->argc < 3) {
		return NULL;
	}

	caching_pool.factory.dump2_status(a->fd, &caching_pool.factory, cli_cachpool);

	return CLI_SUCCESS;
}

static int sip_check_is_conf_help(struct ast_channel *chan, const char *data)
{
	char *parse = NULL;
	const char *family = "NWAY_CONF_STATUS";
	char key[64] = { 0 };
	char sub_value[64] = {0};

	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_sipcheckconf);
		return 0;
	}
	if (ast_strlen_zero(data)) {
		return 0;
	}

	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(conf);
	);
	parse = ast_strdupa(data);
	AST_STANDARD_APP_ARGS(args, parse);
	if (ast_strlen_zero(args.conf)) {
		return 0;
	}
	if (ast_sip_check_is_mcm_conf_number(args.conf)) {
		pbx_builtin_setvar_helper(chan, "IS_CONF_ROOM", "yes");
		return 0;
	}

	snprintf(key, sizeof(key), "%s/IS_N_WAY_CONF", args.conf);
	ast_db_get(family, key, sub_value, sizeof(sub_value) - 1);
	if (!ast_strlen_zero(sub_value)) {
		pbx_builtin_setvar_helper(chan, "IS_CONF_ROOM", "yes");
		return 0;
	}

	return 0;
}

static int check_is_conference(struct ast_channel *chan, const char *data)
{
	if (ast_strlen_zero(data)) {
		return 0;
	}

	if (ast_sip_check_is_mcm_conf_number(data)) {
		return 1;
	}

	if (ast_sip_session_is_nway_conference_number(data)) {
		return 1;
	}

	return 0;
}

/*Start added by zyxia for Bug 261870,2023.6.16*/
static int check_if_can_n_way(struct ast_channel *chan, const char *data)
{
	if (!chan) {
		ast_log(LOG_WARNING, "No channel was provided to %s application.\n", app_checkifcanNway);
		return 0;
	}

	if(ast_channel_get_terminate_type(chan) != AST_TERMINAL_TYPE_WEBRTC) {
		return 1;
	}


	RAII_VAR(struct ast_sip_endpoint *, endpoint, ast_sip_channel_get_endpoint(chan), ao2_cleanup);
	if(endpoint && !ast_sip_endpoint_is_have_h264_codec(endpoint)) {
		return 0;
	}

	return 1;
}
/*End added by zyxia for Bug 261870,2023.6.16*/

unsigned int ast_sip_judge_if_wave_terminal_type(pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_generic_string_hdr *hdr = NULL;
	const pj_str_t name = { "User-Agent", 10 };
	char user_agent[MAX_USER_AGENT_LEN] = { 0 };

	if (!rdata) {
		ast_log(LOG_ERROR, "ast_sip_judge_if_wave_terminal_type, para is invalid\n");
		return 0;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		ast_log(LOG_ERROR, "ast_sip_judge_if_wave_terminal_type, para is invalid\n");
		return 0;
	}

	hdr = pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	if (!hdr) {
		return 0;
	}

	snprintf(user_agent, sizeof(user_agent) - 1, "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);

	if (strcasestr(user_agent, "Grandstream Wave Web") || strcasestr(user_agent, "Grandstream Wave Desktop")) {
		return 1;
	}

	return 0;
}

#endif

static int load_module(void)
{
	struct ast_threadpool_options options;

	/* pjproject and config_system need to be initialized before all else */
	if (pj_init() != PJ_SUCCESS) {
		return AST_MODULE_LOAD_DECLINE;
	}

	if (pjlib_util_init() != PJ_SUCCESS) {
		goto error;
	}

	/* Register PJMEDIA error codes for SDP parsing errors */
	if (pj_register_strerror(PJMEDIA_ERRNO_START, PJ_ERRNO_SPACE_SIZE, pjmedia_strerror)
		!= PJ_SUCCESS) {
		ast_log(LOG_WARNING, "Failed to register pjmedia error codes.  Codes will not be decoded.\n");
	}

	if (ast_sip_initialize_system()) {
		ast_log(LOG_ERROR, "Failed to initialize SIP 'system' configuration section. Aborting load\n");
		goto error;
	}

	/* The serializer needs threadpool and threadpool needs pjproject to be initialized so it's next */
	sip_get_threadpool_options(&options);
	options.thread_start = sip_thread_start;
	sip_threadpool = ast_threadpool_create("SIP", NULL, &options);
	if (!sip_threadpool) {
		goto error;
	}

	if (serializer_pool_setup()) {
		ast_log(LOG_ERROR, "Failed to create SIP serializer pool. Aborting load\n");
		goto error;
	}

	if (ast_sip_initialize_scheduler()) {
		ast_log(LOG_ERROR, "Failed to start scheduler. Aborting load\n");
		goto error;
	}

	/* Now load all the pjproject infrastructure. */
	if (load_pjsip()) {
		goto error;
	}

	if (ast_sip_initialize_transport_events()) {
		ast_log(LOG_ERROR, "Failed to initialize SIP transport monitor. Aborting load\n");
		goto error;
	}

	ast_sip_initialize_dns();
	ast_sip_initialize_global_headers();

	if (ast_res_pjsip_preinit_options_handling()) {
		ast_log(LOG_ERROR, "Failed to pre-initialize OPTIONS handling. Aborting load\n");
		goto error;
	}

	if (ast_res_pjsip_initialize_configuration()) {
		ast_log(LOG_ERROR, "Failed to initialize SIP configuration. Aborting load\n");
		goto error;
	}

	ast_sip_initialize_resolver();
	ast_sip_initialize_dns();

	if (ast_sip_initialize_transport_management()) {
		ast_log(LOG_ERROR, "Failed to initialize SIP transport management. Aborting load\n");
		goto error;
	}

	if (ast_sip_initialize_distributor()) {
		ast_log(LOG_ERROR, "Failed to register distributor module. Aborting load\n");
		goto error;
	}

	if (ast_sip_register_service(&supplement_module)) {
		ast_log(LOG_ERROR, "Failed to initialize supplement hooks. Aborting load\n");
		goto error;
	}

	if (ast_res_pjsip_init_options_handling(0)) {
		ast_log(LOG_ERROR, "Failed to initialize OPTIONS handling. Aborting load\n");
		goto error;
	}
#ifdef GRANDSTREAM_NETWORKS
	get_feature_code(0);
	prase_direct_media_outbound_prefix_conf_to_list(0);
	create_mcm_conf_number();
	prase_mcm_conference_config(0);
	init_turn_info();
	init_plugin_info();
	ast_manager_register_xml("Hold", EVENT_FLAG_CALL, ami_hold_action);
	ast_manager_register_xml("Resume", EVENT_FLAG_CALL, ami_resume_action);
	ast_manager_register_xml("Talk", EVENT_FLAG_CALL, ami_talk_action);
	ast_manager_register_xml("UpdateWavePluginInfo", EVENT_FLAG_SYSTEM, ami_update_plugin_info);
	ast_manager_register_xml("GSControlNotify", EVENT_FLAG_USER | EVENT_FLAG_COMMAND, ami_gs_control_notify);
	ast_manager_register_xml("WaveLogOut",  EVENT_FLAG_COMMAND, action_wavelogout);

	ast_register_application_xml(app_sipcheckconf, sip_check_is_conf_help);

	ast_register_application_xml(app_checkisconference, check_is_conference);

	ast_register_application_xml(app_checkifcanNway, check_if_can_n_way);

	/* Added by yfyang, for websocket remote address */
	init_websocket_remote();

	ast_rwlock_init(&config_xml_lock);
	ast_sip_init_ldap_xml_buffer();
	ast_sip_init_phonebook_xml_buffer();
	callflip_users = ao2_container_alloc(1, callflip_user_hash, callflip_user_cmp);
	ast_sip_reload_waveappinfo_from_astdb(0);//Bug 316955 

#endif
	if (ast_res_pjsip_init_message_filter()) {
		ast_log(LOG_ERROR, "Failed to initialize message IP updating. Aborting load\n");
		goto error;
	}

	/* Added by yfyang, for HA service data sync, not do get hostip when standby */
	#if 0
	if (HA_ROLE_STANDBY != ast_ha_get_role()) {
		/* Add by bfdu for reduce the number of use pj_gethostip call performance, 2018.05.22 */
		const unsigned int flags = 0; /* no port, no brackets */
		if (!pj_gethostip(pj_AF_INET(), &host_ip_ipv4)) {
			pj_sockaddr_print(&host_ip_ipv4, host_ip_ipv4_string, sizeof(host_ip_ipv4_string), flags);
			ast_verb(3, "Local IPv4 address determined to be: %s\n", host_ip_ipv4_string);
		}

		if (!pj_gethostip(pj_AF_INET6(), &host_ip_ipv6)) {
			pj_sockaddr_print(&host_ip_ipv6, host_ip_ipv6_string, sizeof(host_ip_ipv6_string), flags);
			ast_verb(3, "Local IPv6 address determined to be: %s\n", host_ip_ipv6_string);
		}
	}
	#endif
	/* Init route for HA role switch sync */
	/*if ((HA_ROLE_STANDBY == ast_ha_get_role()) && pjsip_ha_stasis_route_init()) {*/
	if ( pjsip_ha_stasis_route_init()) {
		
		ast_log(LOG_ERROR, "Failed to initialize HA stasis route init. Aborting load\n");
		//internal_sip_unregister_service(&send_tsx_module);
		ast_sip_unregister_service(&supplement_module);
		ast_sip_destroy_distributor();
		ast_res_pjsip_destroy_configuration();
		ast_sip_destroy_global_headers();
		stop_monitor_thread();
		ast_sip_destroy_system();
		pj_pool_release(memory_pool);
		memory_pool = NULL;
		pjsip_endpt_destroy(ast_pjsip_endpoint);
		ast_pjsip_endpoint = NULL;
		pj_caching_pool_destroy(&caching_pool);
		return AST_MODULE_LOAD_DECLINE;
		
	}

	if (HA_ROLE_ACTIVE == ast_ha_get_role() && (ast_get_redundance_flag() != 1))
	{
		struct avs_set_float_ip_param para = {0};
		int floatip = get_ha_float_ip();
		struct in_addr ip;

		if (-1 != floatip)
		{
			ip.s_addr = floatip;
			char *ip_str = (char *)ast_inet_ntoa(ip);
			strcpy(para.ip_addr, ip_str);
			para.opt = 1;
			ast_avs_control(NULL, AVS_CONTROL_SET_HA_FIP, &para, sizeof(para));
		}
	}

	ast_cli_register_multiple(cli_commands, ARRAY_LEN(cli_commands));

	AST_TEST_REGISTER(xml_sanitization_end_null);
	AST_TEST_REGISTER(xml_sanitization_exceeds_buffer);



	return AST_MODULE_LOAD_SUCCESS;

error:
	unload_pjsip(NULL);

	/* These functions all check for NULLs and are safe to call at any time */
	ast_sip_destroy_scheduler();
	serializer_pool_shutdown();
	ast_threadpool_shutdown(sip_threadpool);

#ifdef GRANDSTREAM_NETWORKS
	/*Start add by zyxia  ,for callfllip*/
	ao2_ref(callflip_users, -1);
	callflip_users = NULL;
	/*End add by zyxia	,for callfllip*/

	ast_manager_unregister("Hold");
	ast_manager_unregister("Resume");
	ast_manager_unregister("Talk");
	ast_manager_unregister("UpdateWavePluginInfo");
	ast_manager_unregister("GSControlNotify");
	ast_manager_unregister("WaveLogOut");
	ast_unregister_application(app_sipcheckconf);
	ast_unregister_application(app_checkisconference);
	ast_unregister_application(app_checkifcanNway);
#endif

	return AST_MODULE_LOAD_DECLINE;
}

static int reload_module(void)
{
	/*
	 * We must wait for the reload to complete so multiple
	 * reloads cannot happen at the same time.
	 */
	if (ast_sip_push_task_wait_servant(NULL, reload_configuration_task, NULL)) {
		ast_log(LOG_WARNING, "Failed to reload PJSIP\n");
		return -1;
	}

	return 0;
}

static int unload_module(void)
{
	AST_TEST_UNREGISTER(xml_sanitization_end_null);
	AST_TEST_UNREGISTER(xml_sanitization_exceeds_buffer);
	ast_cli_unregister_multiple(cli_commands, ARRAY_LEN(cli_commands));

	/* The thread this is called from cannot call PJSIP/PJLIB functions,
	 * so we have to push the work to the threadpool to handle
	 */
	ast_sip_push_task_wait_servant(NULL, unload_pjsip, NULL);
	ast_sip_destroy_scheduler();
	serializer_pool_shutdown();
	ast_threadpool_shutdown(sip_threadpool);

#ifdef GRANDSTREAM_NETWORKS
	/*Start add by zyxia  ,for callfllip*/
	ao2_ref(callflip_users, -1);
	callflip_users = NULL;
	/*End add by zyxia  ,for callfllip*/

	ast_manager_unregister("Hold");
	ast_manager_unregister("Resume");
	ast_manager_unregister("Talk");
	ast_manager_unregister("UpdateWavePluginInfo");
	ast_manager_unregister("GSControlNotify");
	ast_manager_unregister("WaveLogOut");

	ast_unregister_application(app_sipcheckconf);
	ast_unregister_application(app_checkisconference);
	ast_unregister_application(app_checkifcanNway);
#endif

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS /*Add by weidu ,for stir_shaken*/
char *ast_sip_rdata_get_header_value(pjsip_rx_data *rdata, const pj_str_t str)
{
	pjsip_generic_string_hdr *hdr;
	pj_str_t hdr_val;

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &str, NULL);
	if (!hdr) {
		return NULL;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &hdr_val, &hdr->hvalue);

	return hdr_val.ptr;
}

int ast_sip_smooth_notify_avs_flag(void)
{
	return float_ip_to_avs;
}
#endif
AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "Basic SIP resource",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.reload = reload_module,
	.load_pri = AST_MODPRI_CHANNEL_DEPEND - 5,
	.requires = "dnsmgr,res_pjproject",
	.optional_modules = "res_statsd",
);
