/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 2013, Digium, Inc.
 *
 * Joshua Colp <jcolp@digium.com>
 * Kevin Harwell <kharwell@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \author Joshua Colp <jcolp@digium.com>
 *
 * \brief SIP SDP media stream handling
 */

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<depend>res_pjsip_session</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjmedia.h>
#include <pjlib.h>

#include "asterisk/utils.h"
#include "asterisk/module.h"
#include "asterisk/format.h"
#include "asterisk/format_cap.h"
#include "asterisk/rtp_engine.h"
#include "asterisk/netsock2.h"
#include "asterisk/channel.h"
#include "asterisk/causes.h"
#include "asterisk/sched.h"
#include "asterisk/acl.h"
#include "asterisk/sdp_srtp.h"
#include "asterisk/dsp.h"
#include "asterisk/linkedlists.h"       /* for AST_LIST_NEXT */
#include "asterisk/stream.h"
#include "asterisk/format_cache.h"

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"

#ifdef GRANDSTREAM_NETWORKS
#include "asterisk/avs_controller.h"
#include "../apps/confbridge/include/conf_resource.h"
AST_MODULE_LOG("pjsip");
#endif

/*! \brief Scheduler for RTCP purposes */
static struct ast_sched_context *sched;

/*! \brief Address for RTP */
static struct ast_sockaddr address_rtp;

static const char STR_AUDIO[] = "audio";
static const char STR_VIDEO[] = "video";

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
static const char STR_APPLICATION[] = "application";

/*! \send FIR RTCP internal */
#define SEND_FIR_RTCP_INTERNAL 5
#define SEND_BFCP_HELLO_MAX_COUNT 5
#define SEND_BFCP_KEEPALIVE_INTERNAL 20
#define SEND_FIRST_BFCP_HELLO 1

struct pjmedia_sdp_neg
{
    pjmedia_sdp_neg_state state;	    /**< Negotiator state.	     */
    pj_bool_t		  prefer_remote_codec_order;
    pj_bool_t             answer_with_multiple_codecs;
    pj_bool_t		  has_remote_answer;
    pj_bool_t		  answer_was_remote;

    pjmedia_sdp_session	*initial_sdp,	    /**< Initial local SDP	     */
			*initial_sdp_tmp,   /**< Temporary initial local SDP */
			*active_local_sdp,  /**< Currently active local SDP. */
			*active_remote_sdp, /**< Currently active remote's.  */
			*neg_local_sdp,	    /**< Temporary local SDP.	     */
			*neg_remote_sdp;    /**< Temporary remote SDP.	     */
};

static const char * const bfcp_sdp_floorctl_strmap[][1] = {
	{ /* 0 NULL                  */      ""        },
	{ /* 1 BFCP_FLOORCTL_C_ONLY  */      "c-only"  },
	{ /* 2 BFCP_FLOORCTL_S_ONLY  */      "s-only"  },
	{ /* 3 BFCP_FLOORCTL_C_S     */      "c-s"     },
};

static const char * const bfcp_sdp_setup_strmap[][1] = {
	{ /* 0 NULL                */      ""          },
	{ /* 1 BFCP_SETUP_ACTIVE   */      "active"    },
	{ /* 2 BFCP_SETUP_PASSIVE  */      "passive"   },
	{ /* 3 BFCP_SETUP_ACTPASS  */      "actpass"   },
};

enum ast_special_codec {
	AST_SPECIAL_CODEC_UNKNOW,
	AST_SPECIAL_CODEC_GS_FEC,
	AST_SPECIAL_CODEC_ILBC,
	AST_SPECIAL_CODEC_OPUS,
	AST_SPECIAL_CODEC_RTX,
	AST_SPECIAL_CODEC_G723,
	AST_SPECIAL_CODEC_G729,
	AST_SPECIAL_CODEC_MAX,
};

const char * const ast_special_codec_strmap[][1] = {
	{ /* 0 AST_SPECIAL_CODEC_UNKNOW        */      "unknow"   },
	{ /* 1 AST_SPECIAL_CODEC_GS_FEC        */      "GS-FEC"   },
	{ /* 2 AST_SPECIAL_CODEC_ILBC          */      "ilbc"     },
	{ /* 3 AST_SPECIAL_CODEC_OPUS          */      "opus"     },
	{ /* 4 AST_SPECIAL_CODEC_RTX           */      "RTX"      },
	{ /* 5 AST_SPECIAL_CODEC_G723          */      "g723"     },
	{ /* 6 AST_SPECIAL_CODEC_G729          */      "g729"     },
};

static void get_media_formats_by_order(struct ast_rtp_codecs *codecs, const struct pjmedia_sdp_media *stream,
	struct ast_format_cap *astformats, int *nonastformats);
static int check_if_direct_media_enabled(struct ast_sip_session *session);
static unsigned char judge_if_support_rtx_video_format(struct ast_format *format);
static void set_video_extra_attr_para(const struct pjmedia_sdp_media *stream,
	struct ast_stream *asterisk_stream, enum ast_media_type media_type);
static void set_gs_fec_attr_param(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream);
static int negotiate_incoming_application_bfcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream);
static int create_outgoing_application_bfcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_session *remote, struct ast_stream *stream);
static int apply_negotiated_application_bfcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
	const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream);
static void create_outgoing_disable_sdp(pj_pool_t *pool, struct ast_sip_session *session, 
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_session *remote, struct ast_stream *stream);
static void add_video_extra_attr_to_stream(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream);
static void add_rtcp_attrs_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream);
static void add_gs_fec_version_to_stream(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media,
	struct ast_sip_session_media *session_media, struct ast_stream *stream);
static int process_rtx_ssrc_group_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream, struct ast_stream *asterisk_stream);
static int process_fec_ssrc_group_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream, struct ast_stream *asterisk_stream);
static void add_rtx_ssrc_group_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream);
static void add_fec_ssrc_group_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream);
static void set_avs_params(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream);
static void remove_additional_codec_for_avs(struct ast_format_cap *joint_cap);
static void set_rtcp_attr_param(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream, int enable_nack);
static void change_rtcp_to_external_address(pj_pool_t *pool, struct pjmedia_sdp_media *stream, const char *new_address,
	struct ast_sip_transport_state *transport_state);
static void set_passthrough_param(const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream);
static void add_passthrough_param_to_stream(pj_pool_t *pool, struct pjmedia_sdp_media *media, struct ast_stream *stream);
static void determine_gs_fec_version(struct ast_sip_session_media *session_media, struct ast_stream *stream, char *gs_fec_version);
static void add_ssrc_cname_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, enum ast_ssrc_type ssrc_type);
static void add_ssrc_msid_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream, enum ast_ssrc_type ssrc_type);
static void add_stream_local_media_direction(struct ast_stream *asterisk_stream, pjmedia_sdp_attr *attr);
static void process_remote_stream_port(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, unsigned int remote_port);
static void add_imageattr_to_stream(pj_pool_t *pool, int rtp_code, pjmedia_sdp_media *media, struct ast_stream *stream);
static void set_stream_video_resolution(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream, pjmedia_sdp_media *stream);
static void set_as_tias_for_stream(struct ast_sip_session *session,
	const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream);
static void add_as_tias_to_steam(pj_pool_t *pool, struct pjmedia_sdp_media *media, struct ast_stream *stream);
static void process_gs_fec_codec(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	char *name, unsigned int *has_gs_fec, int fec_payload);
static void set_gvc_msid_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream);
static void set_stream_local_media_direction_according_remote(struct ast_stream *asterisk_stream,
	const struct pjmedia_sdp_media *remote_stream);
static void set_stream_local_media_direction_according_local(struct ast_stream *asterisk_stream,
	const struct pjmedia_sdp_media *local_stream);
static void set_stream_remote_media_direction(struct ast_stream *asterisk_stream,
	const struct pjmedia_sdp_media *remote_stream);
static void modify_joint_caps_for_sfu(struct ast_sip_session *session,
	struct ast_format_cap *joint, struct ast_stream *asterisk_stream);
static unsigned char judge_if_caps_has_invalid_codec(struct ast_sip_session *session, struct ast_format_cap *caps);
static void set_mid_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream);
static void set_trickle_ice_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream, const struct pjmedia_sdp_session *remote);
static void add_trickle_ice_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media);
static void set_session_negotiated_dtmf(struct ast_sip_session *session, struct ast_sip_session_media *session_media, int tel_event);
static int check_media_transport(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *stream);
static void reset_media_encryption(struct ast_sip_session *session, struct ast_sip_session_media *session_media);
static void change_rtx_apt_value(struct ast_sip_session *session, struct ast_format **format, int first_video_pt);
static int change_rtx_pt(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_format *format, char *first_video_name, int rtp_code);
static void process_rtx_codec(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream);
static void check_remote_ice_support(struct ast_sip_session *session, const struct pjmedia_sdp_media *remote_stream);
static void set_media_change_according_direction(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	enum ast_stream_media_direction pre_local_dir, struct ast_stream *asterisk_stream);
static void set_local_media_direction_for_page(struct ast_sip_session *session, struct ast_stream *stream);
static int add_record_indicate_attr(struct ast_sip_session * session,
	pj_pool_t * pool, pjmedia_sdp_media * sdp_media);
static void set_rtcp_rsize_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream);
static void process_remote_ssrc_attr(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_stream *asterisk_stream, pjmedia_sdp_media *stream);
static void set_avs_ice_restart(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *remote_stream, const pjmedia_sdp_session *remote);
static void set_avs_ice_flag(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *remote_stream);
static void set_media_port(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_media *media, int direct_media_enabled, struct ast_sockaddr *addr);
static const char *get_media_host_ip(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_media *media, int direct_media_enabled);
static void set_target_address_for_local_offer(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, unsigned int set_turn_address);
static void set_target_address_for_remote_offer(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_sockaddr *target_address, unsigned int set_turn_address);
static unsigned int judge_if_need_set_turn_address(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media);
static void create_bundled_media_line(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_session *sdp, pjmedia_sdp_media *media, struct ast_stream *stream, pj_pool_t *pool);
static enum ast_neg_pool_type select_pjsip_pool(struct ast_sip_session *session, struct ast_stream *stream);
static void add_msid_attr_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media);
static void add_content_attr_to_stream(struct ast_sip_session *session,
	struct ast_stream *stream, pj_pool_t *pool, pjmedia_sdp_media *media);
static void remove_connection_attr_for_gswave(struct ast_sip_session *session, pjmedia_sdp_media *media);
static void change_p2p_level_id_according_configure(struct ast_sip_session *session,
	struct ast_stream *stream, struct ast_format **format);
static unsigned int jugde_if_add_direction_as_local_value(struct ast_sip_session *session, struct ast_stream *stream);
static void add_bundle_only_to_stream(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media);
static void process_special_codec_params(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, char *name, struct ast_format *format_parsed, int position);
static void set_bundle_group_for_mobile(struct ast_sip_session *session, struct ast_sip_session_media *session_media);
static void set_stream_state_for_page(struct ast_sip_session *session, struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *remote_stream);
static void add_uptime_attr_to_stream(struct ast_sip_session *session,
	struct ast_stream *stream, pj_pool_t *pool, pjmedia_sdp_media *media);
static void set_target_address_for_trunk(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_sockaddr *addrs, int port);
static unsigned int judge_if_support_trickle_ice(struct ast_sip_session *session, struct ast_sip_session_media *session_media);
static unsigned int create_outgoing_sdp_multi_pt(struct ast_sip_session *session,
	struct ast_format *format, pj_pool_t *pool, struct ast_stream *stream, pjmedia_sdp_media *media);
static void set_stream_same_codec_info(struct ast_sip_session *session,
	const struct pjmedia_sdp_media *stream, char *name, int i, struct ast_stream *asterisk_stream, pjmedia_sdp_fmtp fmtp);
static void process_h264_format_no_profile_level_id(struct ast_sip_session *session, struct ast_format **format);
static void set_stream_final_video_resolution(struct ast_sip_session *session, struct ast_stream *asterisk_stream);
static void update_cei_for_change_video_resolution(struct ast_sip_session *session, struct ast_stream *asterisk_stream);
static void indicate_wave_release_hd_video_resource(struct ast_sip_session *session, struct ast_stream *asterisk_stream);
static int get_media_valid_pt(pjmedia_sdp_media *media, int pt);
static void add_comfort_noise_to_sdp(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media, struct ast_format_cap *caps,
	pj_pool_t *pool, const struct pjmedia_sdp_session *remote);
static void set_comfort_noise_pt(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int pt, char *rtpmap_name, pjmedia_sdp_rtpmap *rtpmap);
static void clear_comfort_noise_pt(struct ast_sip_session_media *session_media);
static void clear_video_codec_rtx_pt(struct ast_sip_session_media *session_media, struct ast_stream *stream);
static void modify_cn_according_joint_caps(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format_cap *caps);
static void set_media_net_type(struct ast_sip_session_media *session_media, struct ast_sockaddr *addrs);
static void set_media_net_type_for_local_offer(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, unsigned int set_turn_address);
static void set_session_media_dtls_flag(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_media *stream);
static void add_mpa_to_sdp(pj_pool_t *pool, struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media);
static void set_audio_media_mpa(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int pt);
static void clear_audio_media_mpa(struct ast_sip_session_media *session_media);
static void add_zrtphash_attr_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream);
#endif

static int send_keepalive(const void *data)
{
	struct ast_sip_session_media *session_media = (struct ast_sip_session_media *) data;
	struct ast_rtp_instance *rtp = session_media->rtp;
	int keepalive;
	time_t interval;
	int send_keepalive;

	if (!rtp) {
		return 0;
	}

	keepalive = ast_rtp_instance_get_keepalive(rtp);

	if (!ast_sockaddr_isnull(&session_media->direct_media_addr)) {
		ast_debug(3, "Not sending RTP keepalive on RTP instance %p since direct media is in use\n", rtp);
		return keepalive * 1000;
	}

	interval = time(NULL) - ast_rtp_instance_get_last_tx(rtp);
	send_keepalive = interval >= keepalive;

	ast_debug(3, "It has been %d seconds since RTP was last sent on instance %p. %sending keepalive\n",
			(int) interval, rtp, send_keepalive ? "S" : "Not s");

	if (send_keepalive) {
		ast_rtp_instance_sendcng(rtp, 0);
		return keepalive * 1000;
	}

	return (keepalive - interval) * 1000;
}

/*! \brief Check whether RTP is being received or not */
static int rtp_check_timeout(const void *data)
{
	struct ast_sip_session_media *session_media = (struct ast_sip_session_media *)data;
	struct ast_rtp_instance *rtp = session_media->rtp;
	int elapsed;
	struct ast_channel *chan;

	if (!rtp) {
		return 0;
	}

	elapsed = time(NULL) - ast_rtp_instance_get_last_rx(rtp);
	if (elapsed < ast_rtp_instance_get_timeout(rtp)) {
		return (ast_rtp_instance_get_timeout(rtp) - elapsed) * 1000;
	}

	chan = ast_channel_get_by_name(ast_rtp_instance_get_channel_id(rtp));
	if (!chan) {
		return 0;
	}

	ast_log(LOG_NOTICE, "Disconnecting channel '%s' for lack of RTP activity in %d seconds\n",
		ast_channel_name(chan), elapsed);

	ast_channel_lock(chan);
	ast_channel_hangupcause_set(chan, AST_CAUSE_REQUESTED_CHAN_UNAVAIL);
	ast_channel_unlock(chan);

	ast_softhangup(chan, AST_SOFTHANGUP_DEV);
	ast_channel_unref(chan);

	return 0;
}

/*!
 * \brief Enable RTCP on an RTP session.
 */
static void enable_rtcp(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_media)
{
	enum ast_rtp_instance_rtcp rtcp_type;

	if (session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux) {
		rtcp_type = AST_RTP_INSTANCE_RTCP_MUX;
	} else {
		rtcp_type = AST_RTP_INSTANCE_RTCP_STANDARD;
	}

	ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_RTCP, rtcp_type);
}

/*!
 * \brief Enable an RTP extension on an RTP session.
 */
static void enable_rtp_extension(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	enum ast_rtp_extension extension, enum ast_rtp_extension_direction direction,
	const pjmedia_sdp_session *sdp)
{
	int id = -1;

	/* For a bundle group the local unique identifier space is shared across all streams within
	 * it.
	 */
	if (session_media->bundle_group != -1) {
		int index;

		for (index = 0; index < sdp->media_count; ++index) {
			struct ast_sip_session_media *other_session_media;
			int other_id;

			if (index >= AST_VECTOR_SIZE(&session->pending_media_state->sessions)) {
				break;
			}

			other_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);
			if (!other_session_media->rtp || other_session_media->bundle_group != session_media->bundle_group) {
				continue;
			}

			other_id = ast_rtp_instance_extmap_get_id(other_session_media->rtp, extension);
			if (other_id == -1) {
				/* Worst case we have to fall back to the highest available free local unique identifier
				 * for the bundle group.
				 */
				other_id = ast_rtp_instance_extmap_count(other_session_media->rtp) + 1;
				if (id < other_id) {
					id = other_id;
				}
				continue;
			}

			id = other_id;
			break;
		}
	}

	ast_rtp_instance_extmap_enable(session_media->rtp, id, extension, direction);
}

/*! \brief Internal function which creates an RTP instance */
static int create_rtp(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const pjmedia_sdp_session *sdp)
{
	struct ast_rtp_engine_ice *ice;
	struct ast_sockaddr temp_media_address;
	struct ast_sockaddr *media_address =  &address_rtp;

	if (session->endpoint->media.bind_rtp_to_media_address && !ast_strlen_zero(session->endpoint->media.address)) {
		if (ast_sockaddr_parse(&temp_media_address, session->endpoint->media.address, 0)) {
			ast_debug(1, "Endpoint %s: Binding RTP media to %s\n",
				ast_sorcery_object_get_id(session->endpoint),
				session->endpoint->media.address);
			media_address = &temp_media_address;
		} else {
			ast_debug(1, "Endpoint %s: RTP media address invalid: %s\n",
				ast_sorcery_object_get_id(session->endpoint),
				session->endpoint->media.address);
		}
	} else {
		struct ast_sip_transport *transport;

		transport = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "transport",
			session->endpoint->transport);
		if (transport) {
			struct ast_sip_transport_state *trans_state;

			trans_state = ast_sip_get_transport_state(ast_sorcery_object_get_id(transport));
			if (trans_state) {
				char hoststr[PJ_INET6_ADDRSTRLEN];

				pj_sockaddr_print(&trans_state->host, hoststr, sizeof(hoststr), 0);
				if (ast_sockaddr_parse(&temp_media_address, hoststr, 0)) {
					ast_debug(1, "Transport %s bound to %s: Using it for RTP media.\n",
						session->endpoint->transport, hoststr);
					media_address = &temp_media_address;
				} else {
					ast_debug(1, "Transport %s bound to %s: Invalid for RTP media.\n",
						session->endpoint->transport, hoststr);
				}
				ao2_ref(trans_state, -1);
			}
			ao2_ref(transport, -1);
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	int endpt_ice_support = (session->is_webrtc_call || session->support_ice || (session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) || (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC));
	struct avs_port_param port_param = {0};
	if (session && session->endpoint && (session->is_webrtc_call || session->support_ice || (session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) || (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC)) && session->endpoint->intranet_ip_filter) {
		endpt_ice_support |= AST_RTP_ICE_IP_FILTER;
	}
	port_param.port_group_id = session->port_group_id;
	port_param.endpt_ice_support = endpt_ice_support;
	port_param.media_turn_relay = session->media_turn_relay;
	if (endpt_ice_support) {
		port_param.cand_host_only = session->cand_host_only;
	}
	port_param.trickle_ice = session->support_trickle_ice;
	port_param.if_type = session->if_type;

	port_param.parent_avs_fd = session_media->parent_avs_fd;
	port_param.net_type = session_media->net_type;
	port_param.enable_dtls = session_media->dtls;
	port_param.enable_zrtp = session_media->zrtp;

	if (!ast_strlen_zero(session->avs_address)) {
		strcpy(port_param.avs_address, session->avs_address);
	}
	if (!(session_media->rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, media_address, &port_param)))
#else
	if (!(session_media->rtp = ast_rtp_instance_new(session->endpoint->media.rtp.engine, sched, media_address, NULL)))
#endif
	{
		ast_log(LOG_ERROR, "Unable to create RTP instance using RTP engine '%s'\n", session->endpoint->media.rtp.engine);
		return -1;
	}

	ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_NAT, session->endpoint->media.rtp.symmetric);
	ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_ASYMMETRIC_CODEC, session->endpoint->asymmetric_rtp_codec);

	if (!session->endpoint->media.rtp.ice_support && (ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		ice->stop(session_media->rtp);
	}

	if (session->dtmf == AST_SIP_DTMF_RFC_4733 || session->dtmf == AST_SIP_DTMF_AUTO || session->dtmf == AST_SIP_DTMF_AUTO_INFO) {
		ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_RFC2833);
		ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_DTMF, 1);
	} else if (session->dtmf == AST_SIP_DTMF_INBAND) {
		ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_INBAND);
	}

	if (session_media->type == AST_MEDIA_TYPE_AUDIO &&
			(session->endpoint->media.tos_audio || session->endpoint->media.cos_audio)) {
		ast_rtp_instance_set_qos(session_media->rtp, session->endpoint->media.tos_audio,
				session->endpoint->media.cos_audio, "SIP RTP Audio");
	} else if (session_media->type == AST_MEDIA_TYPE_VIDEO) {
		ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_RETRANS_RECV, session->endpoint->media.webrtc);
		ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_RETRANS_SEND, session->endpoint->media.webrtc);
		ast_rtp_instance_set_prop(session_media->rtp, AST_RTP_PROPERTY_REMB, session->endpoint->media.webrtc);
		if (session->endpoint->media.webrtc) {
			enable_rtp_extension(session, session_media, AST_RTP_EXTENSION_ABS_SEND_TIME, AST_RTP_EXTENSION_DIRECTION_SENDRECV, sdp);
		}
		if (session->endpoint->media.tos_video || session->endpoint->media.cos_video) {
			ast_rtp_instance_set_qos(session_media->rtp, session->endpoint->media.tos_video,
					session->endpoint->media.cos_video, "SIP RTP Video");
		}
	}

	ast_rtp_instance_set_last_rx(session_media->rtp, time(NULL));

	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static void get_codecs(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream, struct ast_rtp_codecs *codecs,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream)
#else
static void get_codecs(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream, struct ast_rtp_codecs *codecs,
	struct ast_sip_session_media *session_media)
#endif
{
	pjmedia_sdp_attr *attr;
	pjmedia_sdp_rtpmap *rtpmap;
	pjmedia_sdp_fmtp fmtp;
	struct ast_format *format;
	int i, num = 0, tel_event = 0;
	char name[256];
	char media[20];
	char fmt_param[256];
	enum ast_rtp_options options = session->endpoint->media.g726_non_standard ?
		AST_RTP_OPT_G726_NONSTANDARD : 0;
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for avs dtmf param, 2019.05.24 */
	unsigned int has_red = 0;
	unsigned int has_gs_fec = 0;

	ast_stream_init_same_codec_info(asterisk_stream);
	clear_comfort_noise_pt(session_media);
	clear_audio_media_mpa(session_media);
#endif

	ast_rtp_codecs_payloads_initialize(codecs);

	/* Iterate through provided formats */
	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* The payload is kept as a string for things like t38 but for video it is always numerical */
		ast_rtp_codecs_payloads_set_m_type(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]));
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
#ifdef GRANDSTREAM_NETWORKS
			set_comfort_noise_pt(session, session_media, pj_strtoul(&stream->desc.fmt[i]), NULL, NULL);
			set_audio_media_mpa(session, session_media, pj_strtoul(&stream->desc.fmt[i]));
#endif
			continue;
		}

		/* Interpret the attribute as an rtpmap */
#ifdef GRANDSTREAM_NETWORKS
		if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, asterisk_stream)) {
			if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
				continue;
			}
		} else {
			if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_notify_prov, attr, &rtpmap)) != PJ_SUCCESS) {
				continue;
			}
		}
#else
		if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}
#endif

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
		if (strcmp(name, "telephone-event") == 0) {
			tel_event++;
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for avs dtmf param, 2019.05.13 */
			ast_rtp_instance_set_2833pt(session_media->rtp, pj_strtoul(&stream->desc.fmt[i]));
#endif
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for BFCP, 2015.11.17 */
		process_gs_fec_codec(session, session_media, name, &has_gs_fec, pj_strtoul(&stream->desc.fmt[i]));
		if (strcasecmp(name, "RED") == 0) {
			has_red = 1;
			ast_rtp_instance_set_red_pt(session_media->rtp, pj_strtoul(&stream->desc.fmt[i]));
		}
		if (strcasecmp(name, "H224") == 0) {
			ast_log(LOG_DEBUG, "[session:%p] It's FECC codec, set FECC PT:%ld to AVS for endpoint:%s\n",
				session, pj_strtoul(&stream->desc.fmt[i]), ast_sorcery_object_get_id(session->endpoint));
			ast_rtp_instance_set_fecc_pt(session_media->rtp, pj_strtoul(&stream->desc.fmt[i]));
		}

		set_comfort_noise_pt(session, session_media, pj_strtoul(&stream->desc.fmt[i]), name, rtpmap);
		set_audio_media_mpa(session, session_media, pj_strtoul(&stream->desc.fmt[i]));
#endif
		ast_copy_pj_str(media, (pj_str_t*)&stream->desc.media, sizeof(media));
		ast_rtp_codecs_payloads_set_rtpmap_type_rate(codecs, NULL,
			pj_strtoul(&stream->desc.fmt[i]), media, name, options, rtpmap->clock_rate);
		/* Look for an optional associated fmtp attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "fmtp", &rtpmap->pt))) {
			continue;
		}

		if ((pjmedia_sdp_attr_get_fmtp(attr, &fmtp)) == PJ_SUCCESS) {
			ast_copy_pj_str(fmt_param, &fmtp.fmt, sizeof(fmt_param));
			if (sscanf(fmt_param, "%30d", &num) != 1) {
				continue;
			}

#ifdef GRANDSTREAM_NETWORKS
			set_stream_same_codec_info(session, stream, name, i, asterisk_stream, fmtp);
#endif

			if ((format = ast_rtp_codecs_get_payload_format(codecs, num))) {
				struct ast_format *format_parsed;

				ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));

				format_parsed = ast_format_parse_sdp_fmtp(format, fmt_param);
				if (format_parsed) {
#ifdef GRANDSTREAM_NETWORKS
					process_special_codec_params(session, stream, session_media, name, format_parsed, i);
					/*start added by zyxia for  Bug 241677,2022.11.3 */
					if(session->is_rtsp_call && !ast_strlen_zero(session->rtsp_profile_level_id) && ast_stream_get_sub_type(asterisk_stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN && !strcasecmp(name, "H264")) {
						ast_format_attribute_set_directly(format_parsed, "profile-level-id", session->rtsp_profile_level_id); //720p
						ast_log(LOG_DEBUG, "[session:%p] Set the %s for rtsp\n",
							session, session->rtsp_profile_level_id);
					}
					/*End added by zyxia for  Bug 241677,2022.11.3 */
#endif
					ast_rtp_codecs_payload_replace_format(codecs, num, format_parsed);
					ao2_ref(format_parsed, -1);
				}

				ao2_ref(format, -1);
			}
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->dtmf == AST_SIP_DTMF_AUTO) {
		if (tel_event) {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_RFC2833);
		} else {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_INBAND);
		}
	}
#else
	if (!tel_event && (session->dtmf == AST_SIP_DTMF_AUTO)) {
		ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_INBAND);
	}
#endif

	if (session->dtmf == AST_SIP_DTMF_AUTO_INFO) {
		if  (tel_event) {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_RFC2833);
		} else {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_NONE);
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->dtmf == AST_SIP_DTMF_BOTH_RFC4733_AND_INFO) {
		if  (tel_event) {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_RFC2833);
		} else {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_NONE);
		}
	}

	if (session->dtmf == AST_SIP_DTMF_RFC_4733) {
		if (tel_event) {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_RFC2833);
		} else {
			ast_rtp_instance_dtmf_mode_set(session_media->rtp, AST_RTP_DTMF_MODE_NONE);
		}
	}

	if (!tel_event && session_media->type == AST_MEDIA_TYPE_AUDIO) {
		ast_log(LOG_DEBUG, "ast_rtp_instance_set_2833pt clear value\n");
		ast_rtp_instance_set_2833pt(session_media->rtp, 0);
	}
	if (!has_red) {
		ast_rtp_instance_set_red_pt(session_media->rtp, 0);
	}
	if (!has_gs_fec) {
		memset(session_media->gs_fec_version, 0, sizeof(session_media->gs_fec_version));
		ast_rtp_instance_set_gsfec_pt(session_media->rtp, 0);
	}
	set_session_negotiated_dtmf(session, session_media, tel_event);
#endif
	/* Get the packetization, if it exists */
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "ptime", NULL))) {
		unsigned long framing = pj_strtoul(pj_strltrim(&attr->value));
		if (framing && session->endpoint->media.rtp.use_ptime) {
			ast_rtp_codecs_set_framing(codecs, framing);
		}
	}
}

static int set_caps(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	struct ast_sip_session_media *session_media_transport,
	const struct pjmedia_sdp_media *stream,
	int is_offer, struct ast_stream *asterisk_stream)
{
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, peer, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, joint, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, endpoint_caps, NULL, ao2_cleanup);
#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_format_cap *, avs_caps, NULL, ao2_cleanup);
	int useinbandfec = 0;
#endif
	enum ast_media_type media_type = session_media->type;
	struct ast_rtp_codecs codecs = AST_RTP_CODECS_NULL_INIT;
	int fmts = 0;
	int direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);
	int dsp_features = 0;

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT)) ||
	    !(peer = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT)) ||
	    !(joint = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n",
			ast_codec_media_type2str(session_media->type));
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!(avs_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities for avs\n",
			ast_codec_media_type2str(session_media->type));
		return -1;
	}
#endif

	/* get the endpoint capabilities */
	if (direct_media_enabled) {
		ast_format_cap_get_compatible(session->endpoint->media.codecs, session->direct_media_cap, caps);
	} else {
#ifdef GRANDSTREAM_NETWORKS
		/*! Begin: Added by zyxia for Bug185737, 2021-1-15.*/
		if (session && session->wSwitchImageInvite488ToAudioInvite) {
			ast_format_cap_append(caps, ast_format_ulaw, 0);
			ast_format_cap_append(caps, ast_format_alaw, 0);
			ast_format_cap_sequence_append_from_cap(caps, session->endpoint->media.codecs, media_type);
		} else {
			ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
		}
		/*! End: Added by zyxia for Bug185737, 2021-1-15.*/
#else
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	/* get the capabilities on the peer */
	get_codecs(session, stream, &codecs,  session_media, asterisk_stream);
	get_media_formats_by_order(&codecs, stream, peer, &fmts);
	useinbandfec = ast_sip_session_get_format_useinbandfec(session, asterisk_stream, peer);
	ast_log(LOG_DEBUG, "[session:%p] aros:%s, fromdomain:%s\n", session, session->endpoint->aors, session->endpoint->fromdomain);
	if (ast_sip_session_get_local_negotiate_prefer() && media_type == AST_MEDIA_TYPE_AUDIO) {
		ast_log(LOG_DEBUG, "[session:%p] Enable local negotiate prefer, use local codec prefer for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_format_cap_get_compatible(caps, peer, joint);
	} else if (1 == session->is_outgoing && session->endpoint->aors && 0 == strncmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_DEBUG, "an inbound call from trunk!\n");
		ast_format_cap_get_compatible(caps, peer, joint);
	} else {
		ast_format_cap_get_compatible(peer, caps, joint);
	}

	ast_format_cap_set_framing(joint, ast_rtp_codecs_get_framing(&codecs));
#else
	/* get the capabilities on the peer */
	get_codecs(session, stream, &codecs,  session_media);

	ast_rtp_codecs_payload_formats(&codecs, peer, &fmts);

	/* get the joint capabilities between peer and endpoint */
	ast_format_cap_get_compatible(caps, peer, joint);
#endif

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_format_cap_count(joint) && !check_if_direct_media_enabled(session)) 
#else
	if (!ast_format_cap_count(joint))
#endif
	{
		struct ast_str *usbuf = ast_str_alloca(AST_FORMAT_CAP_NAMES_LEN);
		struct ast_str *thembuf = ast_str_alloca(AST_FORMAT_CAP_NAMES_LEN);

		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "No joint capabilities for '%s' media stream between our configuration(%s) and incoming SDP(%s)\n",
			ast_codec_media_type2str(session_media->type),
			ast_format_cap_get_names(caps, &usbuf),
			ast_format_cap_get_names(peer, &thembuf));
#ifdef GRANDSTREAM_NETWORKS
		if (media_type == AST_MEDIA_TYPE_VIDEO) {
			ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_CODEC);
		}
#endif
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (is_offer || (session_media->type == AST_MEDIA_TYPE_VIDEO) ||
		(ast_sip_session_media_use_bundle(session) && ast_stream_judge_if_stream_is_participant(asterisk_stream)))
#else
	if (is_offer)
#endif
	{
		/*
		 * Setup rx payload type mapping to prefer the mapping
		 * from the peer that the RFC says we SHOULD use.
		 */
		ast_rtp_codecs_payloads_xover(&codecs, &codecs, NULL);
	}
	ast_rtp_codecs_payloads_copy(&codecs, ast_rtp_instance_get_codecs(session_media->rtp),
		session_media->rtp);

#ifdef GRANDSTREAM_NETWORKS
	modify_joint_caps_for_sfu(session, joint, asterisk_stream);
	if (!judge_if_caps_has_invalid_codec(session, joint) && !check_if_direct_media_enabled(session)) {
		if (media_type == AST_MEDIA_TYPE_VIDEO) {
			ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_CODEC);
		}
		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "No joint capabilities after modify joint_caps, enpoint:%s\n", ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (media_type == AST_MEDIA_TYPE_VIDEO && ast_sip_session_process_special_video_codec(session, joint)) {
		ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_CODEC);
		ast_rtp_codecs_payloads_destroy(&codecs);
		ast_log(LOG_NOTICE, "No joint capabilities after check special video codec, enpoint:%s\n", ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (ast_format_cap_has_rtx(joint)) {
		/* Move rtx codec to the end of joint list */
		ast_format_cap_remove_rtx(joint);
		ast_format_cap_add_rtx(joint);
	}
	/* Add by bfdu for avs_engine, 2019.04.30 */
	ast_format_cap_append_from_cap(avs_caps, joint, AST_MEDIA_TYPE_UNKNOWN);
	remove_additional_codec_for_avs(avs_caps);
	ast_format_cap_set_framing(avs_caps, ast_rtp_codecs_get_framing(&codecs));
	ast_rtp_instance_set_caps(session_media->rtp, avs_caps);
	if (useinbandfec) {
		ast_sip_session_set_format_useinbandfec(session, joint, asterisk_stream, 1);
	} else {
		ast_sip_session_set_format_useinbandfec(session, joint, asterisk_stream, 0);
	}
	ast_sip_session_set_format_dtx(session, joint, asterisk_stream);

	modify_cn_according_joint_caps(session, session_media, joint);
#endif

	ast_stream_set_formats(asterisk_stream, joint);

#ifndef GRANDSTREAM_NETWORKS
	/* If this is a bundled stream then apply the payloads to RTP instance acting as transport to prevent conflicts */
	if (session_media_transport != session_media && session_media->bundled) {
		int index;

		for (index = 0; index < ast_format_cap_count(joint); ++index) {
			struct ast_format *format = ast_format_cap_get_format(joint, index);
			int rtp_code;

			/* Ensure this payload is in the bundle group transport codecs, this purposely doesn't check the return value for
			 * things as the format is guaranteed to have a payload already.
			 */
			rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media->rtp), 1, format, 0);
			ast_rtp_codecs_payload_set_rx(ast_rtp_instance_get_codecs(session_media_transport->rtp), rtp_code, format);

			ao2_ref(format, -1);
		}
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	{
		struct ast_str *format_buf = ast_str_alloca(128);
		ast_log(LOG_DEBUG, "[session:%p] endpoint:%s, get peer caps: %s\n",
			session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(peer, &format_buf));
		ast_log(LOG_DEBUG, "[session:%p] endpoint:%s, get local caps: %s\n",	
			session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(caps, &format_buf));
		ast_log(LOG_DEBUG, "[session:%p] endpoint:%s, set joint caps of caller: %s\n",
			session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(joint, &format_buf));
	}
#endif

	if (session->channel && ast_sip_session_is_pending_stream_default(session, asterisk_stream)) {
		ast_channel_lock(session->channel);
		ast_format_cap_remove_by_type(caps, AST_MEDIA_TYPE_UNKNOWN);
		ast_format_cap_append_from_cap(caps, ast_channel_nativeformats(session->channel),
			AST_MEDIA_TYPE_UNKNOWN);
		ast_format_cap_remove_by_type(caps, media_type);

		if (session->endpoint->preferred_codec_only){
			struct ast_format *preferred_fmt = ast_format_cap_get_format(joint, 0);
			ast_format_cap_append(caps, preferred_fmt, 0);
			ao2_ref(preferred_fmt, -1);
		} else if (!session->endpoint->asymmetric_rtp_codec) {
			struct ast_format *best;
			/*
			 * If we don't allow the sending codec to be changed on our side
			 * then get the best codec from the joint capabilities of the media
			 * type and use only that. This ensures the core won't start sending
			 * out a format that we aren't currently sending.
			 */

			best = ast_format_cap_get_best_by_type(joint, media_type);
			if (best) {
				ast_format_cap_append(caps, best, ast_format_cap_get_framing(joint));
				ao2_ref(best, -1);
			}
		} else {
			ast_format_cap_append_from_cap(caps, joint, media_type);
		}

		/*
		 * Apply the new formats to the channel, potentially changing
		 * raw read/write formats and translation path while doing so.
		 */
		ast_channel_nativeformats_set(session->channel, caps);
		if (media_type == AST_MEDIA_TYPE_AUDIO) {
			ast_set_read_format(session->channel, ast_channel_readformat(session->channel));
			ast_set_write_format(session->channel, ast_channel_writeformat(session->channel));
		}

		if ( ((session->dtmf == AST_SIP_DTMF_AUTO) || (session->dtmf == AST_SIP_DTMF_AUTO_INFO) )
		    && (ast_rtp_instance_dtmf_mode_get(session_media->rtp) == AST_RTP_DTMF_MODE_RFC2833)
		    && (session->dsp)) {
			dsp_features = ast_dsp_get_features(session->dsp);
			dsp_features &= ~DSP_FEATURE_DIGIT_DETECT;
			if (dsp_features) {
				ast_dsp_set_features(session->dsp, dsp_features);
			} else {
				ast_dsp_free(session->dsp);
				session->dsp = NULL;
			}
		}

		if (ast_channel_is_bridged(session->channel)) {
			ast_channel_set_unbridged_nolock(session->channel, 1);
		}

		ast_channel_unlock(session->channel);
	}

	ast_rtp_codecs_payloads_destroy(&codecs);
	return 0;
}

static pjmedia_sdp_attr* generate_rtpmap_attr(struct ast_sip_session *session, pjmedia_sdp_media *media, pj_pool_t *pool,
					      int rtp_code, int asterisk_format, struct ast_format *format, int code)
{
#ifndef HAVE_PJSIP_ENDPOINT_COMPACT_FORM
	extern pj_bool_t pjsip_use_compact_form;
#else
	pj_bool_t pjsip_use_compact_form = pjsip_cfg()->endpt.use_compact_form;
#endif
	pjmedia_sdp_rtpmap rtpmap;
	pjmedia_sdp_attr *attr = NULL;
	char tmp[64];
	enum ast_rtp_options options = session->endpoint->media.g726_non_standard ?
		AST_RTP_OPT_G726_NONSTANDARD : 0;

	snprintf(tmp, sizeof(tmp), "%d", rtp_code);
	pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], tmp);

	if (rtp_code <= AST_RTP_PT_LAST_STATIC && pjsip_use_compact_form) {
		return NULL;
	}

	rtpmap.pt = media->desc.fmt[media->desc.fmt_count - 1];
	rtpmap.clock_rate = ast_rtp_lookup_sample_rate2(asterisk_format, format, code);
	pj_strdup2(pool, &rtpmap.enc_name, ast_rtp_lookup_mime_subtype2(asterisk_format, format, code, options));
	if (!pj_stricmp2(&rtpmap.enc_name, "opus")) {
		pj_cstr(&rtpmap.param, "2");
	} else {
		pj_cstr(&rtpmap.param, NULL);
	}

	pjmedia_sdp_rtpmap_to_attr(pool, &rtpmap, &attr);

	return attr;
}

static pjmedia_sdp_attr* generate_fmtp_attr(pj_pool_t *pool, struct ast_format *format, int rtp_code)
{
	struct ast_str *fmtp0 = ast_str_alloca(256);
	pj_str_t fmtp1;
	pjmedia_sdp_attr *attr = NULL;
	char *tmp;

	ast_format_generate_sdp_fmtp(format, rtp_code, &fmtp0);
	if (ast_str_strlen(fmtp0)) {
		tmp = ast_str_buffer(fmtp0) + ast_str_strlen(fmtp0) - 1;
		/* remove any carriage return line feeds */
		while (*tmp == '\r' || *tmp == '\n') --tmp;
		*++tmp = '\0';
		/* ast...generate gives us everything, just need value */
		tmp = strchr(ast_str_buffer(fmtp0), ':');
		if (tmp && tmp[1] != '\0') {
			fmtp1 = pj_str(tmp + 1);
		} else {
			fmtp1 = pj_str(ast_str_buffer(fmtp0));
		}
		attr = pjmedia_sdp_attr_create(pool, "fmtp", &fmtp1);
	}
	return attr;
}

/*! \brief Function which adds ICE attributes to a media stream */
static void add_ice_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media,
	unsigned int include_candidates)
{
	struct ast_rtp_engine_ice *ice;
	struct ao2_container *candidates;
	const char *username, *password;
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	struct ao2_iterator it_candidates;
	struct ast_rtp_engine_ice_candidate *candidate;

	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		return;
	}

#ifndef GRANDSTREAM_NETWORKS
	if (!session_media->remote_ice) {
		return;
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	if (!session->is_webrtc_call && !session->support_ice && (session->terminal_type != AST_TERMINAL_TYPE_ICE_TRUNK)) {
		return;
	}
#endif

	if ((username = ice->get_ufrag(session_media->rtp))) {
		attr = pjmedia_sdp_attr_create(pool, "ice-ufrag", pj_cstr(&stmp, username));
		media->attr[media->attr_count++] = attr;
	}

	if ((password = ice->get_password(session_media->rtp))) {
		attr = pjmedia_sdp_attr_create(pool, "ice-pwd", pj_cstr(&stmp, password));
		media->attr[media->attr_count++] = attr;
	}

	if (!include_candidates) {
		return;
	}

	candidates = ice->get_local_candidates(session_media->rtp);
	if (!candidates) {
		return;
	}

	it_candidates = ao2_iterator_init(candidates, 0);
	for (; (candidate = ao2_iterator_next(&it_candidates)); ao2_ref(candidate, -1)) {
		struct ast_str *attr_candidate = ast_str_create(128);

		ast_str_set(&attr_candidate, -1, "%s %u %s %d %s ", candidate->foundation, candidate->id, candidate->transport,
					candidate->priority, ast_sockaddr_stringify_addr_remote(&candidate->address));
		ast_str_append(&attr_candidate, -1, "%s typ ", ast_sockaddr_stringify_port(&candidate->address));

		switch (candidate->type) {
			case AST_RTP_ICE_CANDIDATE_TYPE_HOST:
				ast_str_append(&attr_candidate, -1, "host");
				break;
			case AST_RTP_ICE_CANDIDATE_TYPE_SRFLX:
				ast_str_append(&attr_candidate, -1, "srflx");
				break;
			case AST_RTP_ICE_CANDIDATE_TYPE_RELAYED:
				ast_str_append(&attr_candidate, -1, "relay");
				break;
		}

		if (!ast_sockaddr_isnull(&candidate->relay_address)) {
			ast_str_append(&attr_candidate, -1, " raddr %s rport", ast_sockaddr_stringify_addr_remote(&candidate->relay_address));
			ast_str_append(&attr_candidate, -1, " %s", ast_sockaddr_stringify_port(&candidate->relay_address));
		}

		attr = pjmedia_sdp_attr_create(pool, "candidate", pj_cstr(&stmp, ast_str_buffer(attr_candidate)));
		media->attr[media->attr_count++] = attr;

		ast_free(attr_candidate);
	}

	ao2_iterator_destroy(&it_candidates);
	ao2_ref(candidates, -1);
}

/*! \brief Function which checks for ice attributes in an audio stream */
static void check_ice_support(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream)
{
	struct ast_rtp_engine_ice *ice;
	const pjmedia_sdp_attr *attr;
	unsigned int attr_i;

	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		session_media->remote_ice = 0;
		return;
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		attr = remote_stream->attr[attr_i];
		if (!pj_strcmp2(&attr->name, "candidate")) {
			session_media->remote_ice = 1;
#ifdef GRANDSTREAM_NETWORKS
			session->support_ice = 1;
#endif
			break;
		}
	}

	if (attr_i == remote_stream->attr_count) {
		session_media->remote_ice = 0;
	}
}

/*! \brief Function which processes ICE attributes in an audio stream */
static void process_ice_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	struct ast_rtp_engine_ice *ice;
	const pjmedia_sdp_attr *attr;
	char attr_value[256];
	unsigned int attr_i;

	/* If ICE support is not enabled or available exit early */
	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session_media->ice_restart) {
		ast_log(LOG_DEBUG, "[session:%p] ICE restart already for media type:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		session_media->ice_restart = 0;
		return;
	}
#endif

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-ufrag", NULL);
	if (!attr) {
		attr = pjmedia_sdp_attr_find2(remote->attr_count, remote->attr, "ice-ufrag", NULL);
	}
	if (attr) {
		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));
		ice->set_authentication(session_media->rtp, attr_value, NULL);
	} else {
		return;
	}

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-pwd", NULL);
	if (!attr) {
		attr = pjmedia_sdp_attr_find2(remote->attr_count, remote->attr, "ice-pwd", NULL);
	}
	if (attr) {
		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));
		ice->set_authentication(session_media->rtp, NULL, attr_value);
	} else {
		return;
	}

	if (pjmedia_sdp_media_find_attr2(remote_stream, "ice-lite", NULL)) {
		ice->ice_lite(session_media->rtp);
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		char foundation[33], transport[32], address[PJ_INET6_ADDRSTRLEN + 1], cand_type[6], relay_address[PJ_INET6_ADDRSTRLEN + 1] = "";
		unsigned int port, relay_port = 0;
		struct ast_rtp_engine_ice_candidate candidate = { 0, };

		attr = remote_stream->attr[attr_i];

		/* If this is not a candidate line skip it */
		if (pj_strcmp2(&attr->name, "candidate")) {
			continue;
		}

		ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));

		if (sscanf(attr_value, "%32s %30u %31s %30u %46s %30u typ %5s %*s %23s %*s %30u", foundation, &candidate.id, transport,
			(unsigned *)&candidate.priority, address, &port, cand_type, relay_address, &relay_port) < 7) {
			/* Candidate did not parse properly */
			continue;
		}

		if (session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux && candidate.id > 1) {
			/* Remote side may have offered RTP and RTCP candidates. However, if we're using RTCP MUX,
			 * then we should ignore RTCP candidates.
			 */
			continue;
		}

		candidate.foundation = foundation;
		candidate.transport = transport;

		ast_sockaddr_parse(&candidate.address, address, PARSE_PORT_FORBID);
		ast_sockaddr_set_port(&candidate.address, port);

		if (!strcasecmp(cand_type, "host")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_HOST;
		} else if (!strcasecmp(cand_type, "srflx")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_SRFLX;
		} else if (!strcasecmp(cand_type, "relay")) {
			candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_RELAYED;
		} else {
			continue;
		}

		if (!ast_strlen_zero(relay_address)) {
			ast_sockaddr_parse(&candidate.relay_address, relay_address, PARSE_PORT_FORBID);
		}

		if (relay_port) {
			ast_sockaddr_set_port(&candidate.relay_address, relay_port);
		}

		ice->add_remote_candidate(session_media->rtp, &candidate);
	}

	ice->set_role(session_media->rtp, pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_TRUE ?
		AST_RTP_ICE_ROLE_CONTROLLING : AST_RTP_ICE_ROLE_CONTROLLED);
	ice->start(session_media->rtp);
}

#ifndef GRANDSTREAM_NETWORKS
/* Modify by bfdu, those function will used by chan_pjsip, moved to res_pjsip_session.c */
/*! \brief figure out if media stream has crypto lines for sdes */
static int media_stream_has_crypto(const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		return 1;
	}

	return 0;
}

/*! \brief figure out media transport encryption type from the media transport string */
static enum ast_sip_session_media_encryption get_media_encryption_type(pj_str_t transport,
	const struct pjmedia_sdp_media *stream, unsigned int *optimistic)
{
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	*optimistic = 0;

	if (!transport_str) {
		return AST_SIP_MEDIA_TRANSPORT_INVALID;
	}
	if (strstr(transport_str, "UDP/TLS")) {
		return AST_SIP_MEDIA_ENCRYPT_DTLS;
	} else if (strstr(transport_str, "SAVP")) {
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else if (media_stream_has_crypto(stream)) {
		*optimistic = 1;
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}
}

/*!
 * \brief Checks whether the encryption offered in SDP is compatible with the endpoint's configuration
 * \internal
 *
 * \param endpoint_encryption Media encryption configured for the endpoint
 * \param stream pjmedia_sdp_media stream description
 *
 * \retval AST_SIP_MEDIA_TRANSPORT_INVALID on encryption mismatch
 * \retval The encryption requested in the SDP
 */
static enum ast_sip_session_media_encryption check_endpoint_media_transport(
	struct ast_sip_endpoint *endpoint,
	const struct pjmedia_sdp_media *stream)
{
	enum ast_sip_session_media_encryption incoming_encryption;
	char transport_end = stream->desc.transport.ptr[stream->desc.transport.slen - 1];
	unsigned int optimistic;

	if ((transport_end == 'F' && !endpoint->media.rtp.use_avpf)
		|| (transport_end != 'F' && endpoint->media.rtp.use_avpf)) {
		return AST_SIP_MEDIA_TRANSPORT_INVALID;
	}

	incoming_encryption = get_media_encryption_type(stream->desc.transport, stream, &optimistic);

	if (incoming_encryption == endpoint->media.rtp.encryption) {
		return incoming_encryption;
	}

	if (endpoint->media.rtp.force_avp ||
		endpoint->media.rtp.encryption_optimistic) {
		return incoming_encryption;
	}

	/* If an optimistic offer has been made but encryption is not enabled consider it as having
	 * no offer of crypto at all instead of invalid so the session proceeds.
	 */
	if (optimistic) {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}

	return AST_SIP_MEDIA_TRANSPORT_INVALID;
}
#endif

/*
static void parse_zrtp_hash(const char *sdp_attr, char *version, char *hash) {
	const char *version_end = strchr(sdp_attr, ' ');
	if (version_end == NULL) {
		ast_debug(1, "Invalid SDP attribute format.\n");
		return;
	}

	strncpy(version, sdp_attr, version_end - sdp_attr);
	version[version_end - sdp_attr] = '\0';

	strcpy(hash, version_end + 1); // +1 skip the spcae 
	return;
}
*/
static int setup_zrtp(struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;
		RAII_VAR(char *, zrtp_hash, NULL, ast_free);

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "zrtp-hash")) {
			continue;
		}

		zrtp_hash = ast_strndup(attr->value.ptr, attr->value.slen);
		if (!zrtp_hash) {
			return -1;
		}

		//parse_zrtp_hash(zrtp_hash,version,hash);
		ast_rtp_instance_set_remote_zrtphash(session_media->rtp,zrtp_hash);
		session_media->zrtp = 1;
		ast_debug(1, "Ignoring zrtp_hash offer with unsupported parameters:  hash:%s\n", zrtp_hash);
		return 0;

	}

	/* no usable zrtp_hash attributes found */
	return -1;
}

static int setup_srtp(struct ast_sip_session_media *session_media)
{
	if (!session_media->srtp) {
		session_media->srtp = ast_sdp_srtp_alloc();
		if (!session_media->srtp) {
			return -1;
		}
	}

	if (!session_media->srtp->crypto) {
		session_media->srtp->crypto = ast_sdp_crypto_alloc();
		if (!session_media->srtp->crypto) {
			return -1;
		}
	}

	return 0;
}

static int setup_dtls_srtp(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media)
{
	struct ast_rtp_engine_dtls *dtls;

	if (!session->endpoint->media.rtp.dtls_cfg.enabled || !session_media->rtp) {
		return -1;
	}

	dtls = ast_rtp_instance_get_dtls(session_media->rtp);
	if (!dtls) {
		return -1;
	}

	session->endpoint->media.rtp.dtls_cfg.suite = ((session->endpoint->media.rtp.srtp_tag_32) ? AST_AES_CM_128_HMAC_SHA1_32 : AST_AES_CM_128_HMAC_SHA1_80);
	if (dtls->set_configuration(session_media->rtp, &session->endpoint->media.rtp.dtls_cfg)) {
		ast_log(LOG_ERROR, "Attempted to set an invalid DTLS-SRTP configuration on RTP instance '%p'\n",
			session_media->rtp);
		return -1;
	}

	if (setup_srtp(session_media)) {
		return -1;
	}
	return 0;
}

static void apply_dtls_attrib(struct ast_sip_session_media *session_media,
	pjmedia_sdp_attr *attr)
{
	struct ast_rtp_engine_dtls *dtls = ast_rtp_instance_get_dtls(session_media->rtp);
	pj_str_t *value;

	if (!attr->value.ptr || !dtls) {
		return;
	}

	value = pj_strtrim(&attr->value);

	if (!pj_strcmp2(&attr->name, "setup")) {
		if (!pj_stricmp2(value, "active")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_ACTIVE);
		} else if (!pj_stricmp2(value, "passive")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_PASSIVE);
		} else if (!pj_stricmp2(value, "actpass")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_ACTPASS);
		} else if (!pj_stricmp2(value, "holdconn")) {
			dtls->set_setup(session_media->rtp, AST_RTP_DTLS_SETUP_HOLDCONN);
		} else {
			ast_log(LOG_WARNING, "Unsupported setup attribute value '%*s'\n", (int)value->slen, value->ptr);
		}
	} else if (!pj_strcmp2(&attr->name, "connection")) {
		if (!pj_stricmp2(value, "new")) {
			dtls->reset(session_media->rtp);
		} else if (!pj_stricmp2(value, "existing")) {
			/* Do nothing */
		} else {
			ast_log(LOG_WARNING, "Unsupported connection attribute value '%*s'\n", (int)value->slen, value->ptr);
		}
	} else if (!pj_strcmp2(&attr->name, "fingerprint")) {
		char hash_value[256], hash[32];
		char fingerprint_text[value->slen + 1];
		ast_copy_pj_str(fingerprint_text, value, sizeof(fingerprint_text));
			if (sscanf(fingerprint_text, "%31s %255s", hash, hash_value) == 2) {
			if (!strcasecmp(hash, "sha-1")) {
				dtls->set_fingerprint(session_media->rtp, AST_RTP_DTLS_HASH_SHA1, hash_value);
			} else if (!strcasecmp(hash, "sha-256")) {
				dtls->set_fingerprint(session_media->rtp, AST_RTP_DTLS_HASH_SHA256, hash_value);
			} else {
				ast_log(LOG_WARNING, "Unsupported fingerprint hash type '%s'\n",
				hash);
			}
		}
	}
}

static int parse_dtls_attrib(struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < sdp->attr_count; i++) {
		apply_dtls_attrib(session_media, sdp->attr[i]);
	}

	for (i = 0; i < stream->attr_count; i++) {
		apply_dtls_attrib(session_media, stream->attr[i]);
	}

	ast_set_flag(session_media->srtp, AST_SRTP_CRYPTO_OFFER_OK);

	return 0;
}

static int setup_sdes_srtp(struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;
		RAII_VAR(char *, crypto_str, NULL, ast_free);

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		crypto_str = ast_strndup(attr->value.ptr, attr->value.slen);
		if (!crypto_str) {
			return -1;
		}

		if (setup_srtp(session_media)) {
			return -1;
		}

		if (!ast_sdp_crypto_process(session_media->rtp, session_media->srtp, crypto_str)) {
			/* found a valid crypto attribute */
			return 0;
		}

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 153942, 2020.6.23 */
		if (session_media->srtp) {
			ast_clear_flag(session_media->srtp, AST_FLAGS_ALL);
		}
#endif

		ast_debug(1, "Ignoring crypto offer with unsupported parameters: %s\n", crypto_str);
	}

	/* no usable crypto attributes found */
	return -1;
}

static int setup_media_encryption(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
#ifdef GRANDSTREAM_NETWORKS
	unsigned int use_received_transport;
	unsigned int use_zrtp=0;
	if (session->is_webrtc_call) {
		use_received_transport = session->endpoint->media.rtp.use_received_transport;
	} else {
		use_received_transport = session->endpoint->media.rtp.configure_use_received_transport;
	}


	if (!session->is_webrtc_call) {
		use_zrtp = session->endpoint->media.rtp.configure_use_received_zrtp;
	}


	if(use_zrtp && ast_sip_session_media_stream_has_zrtphash(stream)) {
		if(ast_sip_session_media_stream_has_srtp(stream->desc.transport, stream))
			session_media->encryption = AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP;
		else
			session_media->encryption = AST_SIP_MEDIA_ENCRYPT_ZRTP;
	}

	else if (use_received_transport) {
		ast_log(LOG_DEBUG, "[session:%p] setup_media_encryption use received transport for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		unsigned int optimistic;
		session_media->encryption = ast_sip_session_get_media_encryption_type(stream->desc.transport, stream, &optimistic);
	}

	ast_log(LOG_DEBUG, "[session:%p] Setup media encryption:%s for endpoint:%s\n",
		session, ast_sip_session_media_encryption_type2str(session_media->encryption), ast_sorcery_object_get_id(session->endpoint));
#endif
	switch (session_media->encryption) {
	case AST_SIP_MEDIA_ENCRYPT_SDES:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
#endif
		if (setup_sdes_srtp(session_media, stream)) {
			return -1;
		}
		break;
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
#endif
		if (setup_dtls_srtp(session, session_media)) {
			return -1;
		}
		if (parse_dtls_attrib(session_media, sdp, stream)) {
			return -1;
		}
		break;
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_ENCRYPT_ZRTP:
	case AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP:
	if (setup_zrtp(session_media, stream)) {
		return -1;
	}
	break;
#endif
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
	case AST_SIP_MEDIA_ENCRYPT_NONE:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
	case AST_SIP_MEDIA_ENCRYPT_ZRTP_NONE:

#endif
		break;
	}

	return 0;
}

static void set_ice_components(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_rtp_engine_ice *ice;

	ast_assert(session_media->rtp != NULL);

	ice = ast_rtp_instance_get_ice(session_media->rtp);
	if (!session->endpoint->media.rtp.ice_support || !ice) {
		return;
	}

	if (session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux) {
		/* We both support RTCP mux. Only one ICE component necessary */
		ice->change_components(session_media->rtp, 1);
	} else {
		/* They either don't support RTCP mux or we don't know if they do yet. */
		ice->change_components(session_media->rtp, 2);
	}
}

#ifndef GRANDSTREAM_NETWORKS

/*! \brief Function which adds ssrc attributes to a media stream */
static void add_ssrc_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	char tmp[128];

	if (!session->endpoint->media.bundle || session_media->bundle_group == -1) {
		return;
	}

	snprintf(tmp, sizeof(tmp), "%u cname:%s", ast_rtp_instance_get_ssrc(session_media->rtp), ast_rtp_instance_get_cname(session_media->rtp));
	attr = pjmedia_sdp_attr_create(pool, "ssrc", pj_cstr(&stmp, tmp));
	media->attr[media->attr_count++] = attr;
}
#endif

/*! \brief Function which processes ssrc attributes in a stream */
#ifdef GRANDSTREAM_NETWORKS
static void process_ssrc_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream, struct ast_stream *asterisk_stream)
#else
static void process_ssrc_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream)
#endif
{
	int index;

#ifdef GRANDSTREAM_NETWORKS
	int get_remote_msid = 0;
	int get_remote_cname = 0;
	int get_remote_ssrc = 0;
	unsigned int previous_remote_ssrc = 0;

	/* Set reset flag first, if has remote SSRC, will clear this flag */
	previous_remote_ssrc = ast_rtp_instance_get_remote_ssrc(session_media->rtp);
	if (previous_remote_ssrc) {
		ast_rtp_instance_set_ssrc_reset_flag(session_media->rtp, 1);
		ast_rtp_instance_set_remote_ssrc(session_media->rtp, 0);
		ast_rtp_instance_set_remote_rtx_ssrc(session_media->rtp, 0);
	}
#else
	if (!session->endpoint->media.bundle) {
		return;
	}
#endif

	for (index = 0; index < remote_stream->attr_count; ++index) {
		pjmedia_sdp_attr *attr = remote_stream->attr[index];
		char attr_value[pj_strlen(&attr->value) + 1];
		char *ssrc_attribute_name, *ssrc_attribute_value = NULL;
		unsigned int ssrc;

		/* We only care about ssrc attributes */
		if (pj_strcmp2(&attr->name, "ssrc")) {
			continue;
		}

		ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

		if ((ssrc_attribute_name = strchr(attr_value, ' '))) {
			/* This has an actual attribute */
			*ssrc_attribute_name++ = '\0';
			ssrc_attribute_value = strchr(ssrc_attribute_name, ':');
			if (ssrc_attribute_value) {
				/* Values are actually optional according to the spec */
				*ssrc_attribute_value++ = '\0';
			}
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for prase msid, 2019.04.20 */
			if (!get_remote_msid && strstr(ssrc_attribute_name, "msid")) {
				ast_log(LOG_DEBUG, "[session:%p] Process ssrc msid attribute name:%s, value:%s, for endpoint:%s\n",
					session, ssrc_attribute_name, ssrc_attribute_value, ast_sorcery_object_get_id(session->endpoint));
				strncpy(session_media->remote_msid, ssrc_attribute_value, sizeof(session_media->remote_msid) - 1);
				get_remote_msid = 1;
			}

			if (!get_remote_cname && strstr(ssrc_attribute_name, "cname")) {
				ast_log(LOG_DEBUG, "[session:%p] Process ssrc cname attribute name:%s, value:%s for endpoint:%s\n",
					session, ssrc_attribute_name, ssrc_attribute_value, ast_sorcery_object_get_id(session->endpoint));
				strncpy(session_media->remote_cname, ssrc_attribute_value, sizeof(session_media->remote_cname) - 1);
				get_remote_cname = 1;
			}
#endif
		}

		if (sscanf(attr_value, "%30u", &ssrc) < 1) {
			continue;
		}

		/* If we are currently negotiating as a result of the remote side renegotiating then
		 * determine if the source for this stream has changed.
		 */
#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu for reduce SFU re-INVITE for receive init INVITE */
		if (pjmedia_sdp_neg_get_state(session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER &&
			(session->inv_session->state == PJSIP_INV_STATE_CONFIRMED))
#else
		if (pjmedia_sdp_neg_get_state(session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER &&
			session->active_media_state)
#endif
		{
#ifdef GRANDSTREAM_NETWORKS
			if (previous_remote_ssrc != ssrc && !get_remote_ssrc) {
				ast_log(LOG_DEBUG, "[session:%p] Media:%s SSRC change from %d to %d for endpoint:%s\n",
					session, ast_codec_media_sub_type2str(session_media->sub_type), previous_remote_ssrc, ssrc, ast_sorcery_object_get_id(session->endpoint));
				session_media->changed = 1;
			}
#else
			struct ast_rtp_instance_stats stats = { 0, };
			if (!ast_rtp_instance_get_stats(session_media->rtp, &stats, AST_RTP_INSTANCE_STAT_REMOTE_SSRC) &&
				stats.remote_ssrc != ssrc) {
				session_media->changed = 1;
			}
#endif
		}

#ifdef GRANDSTREAM_NETWORKS
		if (!get_remote_ssrc) {
			ast_rtp_instance_set_remote_ssrc(session_media->rtp, ssrc);
			get_remote_ssrc = 1;
			ast_rtp_instance_set_ssrc_reset_flag(session_media->rtp, 0);
		}
#else
		ast_rtp_instance_set_remote_ssrc(session_media->rtp, ssrc);
#endif
	}
}

#ifndef GRANDSTREAM_NETWORKS
static void add_msid_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media,
	struct ast_stream *stream)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	char msid[(AST_UUID_STR_LEN * 2) + 2];
	const char *stream_label = ast_stream_get_metadata(stream, "SDP:LABEL");

	if (!session->endpoint->media.webrtc) {
		return;
	}

	if (ast_strlen_zero(session_media->mslabel)) {
		/* If this stream is grouped with another then use its media stream label if possible */
		if (ast_stream_get_group(stream) != -1) {
			struct ast_sip_session_media *group_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, ast_stream_get_group(stream));

			ast_copy_string(session_media->mslabel, group_session_media->mslabel, sizeof(session_media->mslabel));
		}

		if (ast_strlen_zero(session_media->mslabel)) {
			ast_uuid_generate_str(session_media->mslabel, sizeof(session_media->mslabel));
		}
	}

	if (ast_strlen_zero(session_media->label)) {
		ast_uuid_generate_str(session_media->label, sizeof(session_media->label));
		/* add for stream identification to replace stream_name */
		ast_stream_set_metadata(stream, "MSID:LABEL", session_media->label);
	}

	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
	ast_debug(3, "Stream msid: %p %s %s\n", stream,
		ast_codec_media_type2str(ast_stream_get_type(stream)), msid);
	attr = pjmedia_sdp_attr_create(pool, "msid", pj_cstr(&stmp, msid));
	pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);

	/* 'label' must come after 'msid' */
	if (!ast_strlen_zero(stream_label)) {
		ast_debug(3, "Stream Label: %p %s %s\n", stream,
			ast_codec_media_type2str(ast_stream_get_type(stream)), stream_label);
		attr = pjmedia_sdp_attr_create(pool, "label", pj_cstr(&stmp, stream_label));
		pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
	}
}
#endif

static void add_rtcp_fb_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;

	if (!session->endpoint->media.webrtc || session_media->type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	/*
	 * For now just automatically add it the stream even though it hasn't
	 * necessarily been negotiated.
	 */
	attr = pjmedia_sdp_attr_create(pool, "rtcp-fb", pj_cstr(&stmp, "* ccm fir"));
	pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);

	attr = pjmedia_sdp_attr_create(pool, "rtcp-fb", pj_cstr(&stmp, "* goog-remb"));
	pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);

	attr = pjmedia_sdp_attr_create(pool, "rtcp-fb", pj_cstr(&stmp, "* nack"));
	pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
}

static void add_extmap_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	int idx;
	char extmap_value[256];

	if (!session->endpoint->media.webrtc || session_media->type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	/* RTP extension local unique identifiers start at '1' */
	for (idx = 1; idx <= ast_rtp_instance_extmap_count(session_media->rtp); ++idx) {
		enum ast_rtp_extension extension = ast_rtp_instance_extmap_get_extension(session_media->rtp, idx);
		const char *direction_str = "";
		pj_str_t stmp;
		pjmedia_sdp_attr *attr;

		/* If this is an unsupported RTP extension we can't place it into the SDP */
		if (extension == AST_RTP_EXTENSION_UNSUPPORTED) {
			continue;
		}

		switch (ast_rtp_instance_extmap_get_direction(session_media->rtp, idx)) {
		case AST_RTP_EXTENSION_DIRECTION_SENDRECV:
			/* Lack of a direction indicates sendrecv, so we leave it out */
			direction_str = "";
			break;
		case AST_RTP_EXTENSION_DIRECTION_SENDONLY:
			direction_str = "/sendonly";
			break;
		case AST_RTP_EXTENSION_DIRECTION_RECVONLY:
			direction_str = "/recvonly";
			break;
		case AST_RTP_EXTENSION_DIRECTION_NONE:
			/* It is impossible for a "none" direction extension to be negotiated but just in case
			 * we treat it as inactive.
			 */
		case AST_RTP_EXTENSION_DIRECTION_INACTIVE:
			direction_str = "/inactive";
			break;
		}

		snprintf(extmap_value, sizeof(extmap_value), "%d%s %s", idx, direction_str,
			ast_rtp_instance_extmap_get_uri(session_media->rtp, idx));
		attr = pjmedia_sdp_attr_create(pool, "extmap", pj_cstr(&stmp, extmap_value));
		pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
	}
}

/*! \brief Function which processes extmap attributes in a stream */
static void process_extmap_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream)
{
	int index;

#ifdef GRANDSTREAM_NETWORKS
	char transport_cc_str[128] = {0};
	unsigned int transport_cc_id = 0;
	if (!session->is_webrtc_call || session_media->type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}
#else
	if (!session->endpoint->media.webrtc || session_media->type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}
#endif

	ast_rtp_instance_extmap_clear(session_media->rtp);

	for (index = 0; index < remote_stream->attr_count; ++index) {
		pjmedia_sdp_attr *attr = remote_stream->attr[index];
		char attr_value[pj_strlen(&attr->value) + 1];
		char *uri;
		int id;
		char direction_str[10] = "";
		char *attributes;
		enum ast_rtp_extension_direction direction = AST_RTP_EXTENSION_DIRECTION_SENDRECV;

		/* We only care about extmap attributes */
		if (pj_strcmp2(&attr->name, "extmap")) {
			continue;
		}

		ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

#ifdef GRANDSTREAM_NETWORKS
		if ((sscanf(attr_value, "%d %s", &transport_cc_id, transport_cc_str) < 1) || (transport_cc_id < 1)) {
			/* We require at a minimum the unique identifier */
			continue;
		}

		if (!strcmp(transport_cc_str, USER_EXTMAP_TRANSPORT_CC_WHOLE)) {
			session_media->transport_cc_id = transport_cc_id;
			ast_log(LOG_DEBUG, "[session:%p] process_extmap_attributes get transport_cc_id:%d for endpoint:%s\n",
				session, transport_cc_id, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}
#endif

		/* Split the combined unique identifier and direction away from the URI and attributes for easier parsing */
		uri = strchr(attr_value, ' ');
		if (ast_strlen_zero(uri)) {
			continue;
		}
		*uri++ = '\0';

		if ((sscanf(attr_value, "%30d%9s", &id, direction_str) < 1) || (id < 1)) {
			/* We require at a minimum the unique identifier */
			continue;
		}

		/* Convert from the string to the internal representation */
		if (!strcasecmp(direction_str, "/sendonly")) {
			direction = AST_RTP_EXTENSION_DIRECTION_SENDONLY;
		} else if (!strcasecmp(direction_str, "/recvonly")) {
			direction = AST_RTP_EXTENSION_DIRECTION_RECVONLY;
		} else if (!strcasecmp(direction_str, "/inactive")) {
			direction = AST_RTP_EXTENSION_DIRECTION_INACTIVE;
		}

		attributes = strchr(uri, ' ');
		if (!ast_strlen_zero(attributes)) {
			*attributes++ = '\0';
		}

		ast_rtp_instance_extmap_negotiate(session_media->rtp, id, direction, uri, attributes);
	}
}

/*! \brief Function which negotiates an incoming media stream */
static int negotiate_incoming_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const pjmedia_sdp_session *sdp,
	int index, struct ast_stream *asterisk_stream)
{
	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	pjmedia_sdp_media *stream = sdp->media[index];
	struct ast_sip_session_media *session_media_transport;
	enum ast_media_type media_type = session_media->type;
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
	int res;

#ifdef GRANDSTREAM_NETWORKS
	enum ast_stream_media_direction pre_local_dir = UNKNOW;
	unsigned int set_turn_address = 0;

	/* Add by bfdu for BFCP/FEC/FECC */
	ast_sip_session_set_media_sub_type(session, stream, session_media, asterisk_stream);

	if (!ast_sip_session_jugde_if_media_support(session, session_media->sub_type)) {
		ast_log(LOG_DEBUG, "[session:%p] Media type is %s, UCM A series can't support this media, for endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
		res = negotiate_incoming_application_bfcp_sdp_stream(session, session_media, sdp, stream);
		return res;
	}
#endif

	/* If no type formats have been configured reject this stream */
	if (!ast_format_cap_has_type(session->endpoint->media.codecs, media_type)) {
		ast_debug(3, "Endpoint has no codecs for media type '%s', declining stream\n",
			ast_codec_media_type2str(session_media->type));
#ifdef GRANDSTREAM_NETWORKS
		if (media_type == AST_MEDIA_TYPE_VIDEO) {
			ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_CODEC);
		}
#endif
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (check_media_transport(session, session_media, stream)) {
		return -1;
	}
	ast_debug(3, " negotiate_incoming_sdp_stream media type '%s',zrtp:%d \n",
		ast_codec_media_type2str(session_media->type),session_media->zrtp);

	check_remote_ice_support(session, stream);

	set_session_media_dtls_flag(session, session_media, stream);
#else
	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport) {
		encryption = check_endpoint_media_transport(session->endpoint, stream);

		if (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID) {
			return -1;
		}
	}
#endif

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	set_turn_address = judge_if_need_set_turn_address(session, session_media);

	session_media_transport = ast_sip_session_media_get_transport(session, session_media);
	if (session_media_transport != session_media && session_media->bundled && session_media_transport->rtp) {
		session_media->parent_avs_fd = ast_rtp_instance_get_avs_fd(session_media_transport->rtp);
	}

	set_media_net_type(session_media, addrs);
#endif

	/* Using the connection information create an appropriate RTP instance */
	if (!session_media->rtp && create_rtp(session, session_media, sdp)) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_sockaddr_set_port(addrs, stream->desc.port);
	set_target_address_for_remote_offer(session, session_media, addrs, set_turn_address);
	if (session->media_turn_relay && ast_strlen_zero(session->turn_target_address)) {
		strncpy(session->turn_target_address, host, sizeof(session->turn_target_address) - 1);
	}
#endif

#ifndef GRANDSTREAM_NETWORKS
	/* Modify by bfdu, move after set_caps, we need to know if has rtx caps, 2019.04.20 */
	process_ssrc_attributes(session, session_media, stream);
#endif
	process_extmap_attributes(session, session_media, stream);
#ifndef GRANDSTREAM_NETWORKS
	/* Modify by bfdu for move this up */
	session_media_transport = ast_sip_session_media_get_transport(session, session_media);
#endif

	if (session_media_transport == session_media || !session_media->bundled) {
		/* If this media session is carrying actual traffic then set up those aspects */
		session_media->remote_rtcp_mux = (pjmedia_sdp_media_find_attr2(stream, "rtcp-mux", NULL) != NULL);
		set_ice_components(session, session_media);

		enable_rtcp(session, session_media, stream);

		res = setup_media_encryption(session, session_media, sdp, stream);
		if (res) {
#ifdef GRANDSTREAM_NETWORKS
		if (!check_if_direct_media_enabled(session) && 
			(!session->endpoint->media.rtp.encryption_optimistic ||
				!pj_strncmp2(&stream->desc.transport, "RTP/SAVP", 8)))
#else
		if (!session->endpoint->media.rtp.encryption_optimistic ||
				!pj_strncmp2(&stream->desc.transport, "RTP/SAVP", 8))
#endif
			{
				/* If optimistic encryption is disabled and crypto should have been enabled
				 * but was not this session must fail. This must also fail if crypto was
				 * required in the offer but could not be set up.
				 */
				return -1;
			}
			/* There is no encryption, sad. */
			session_media->encryption = AST_SIP_MEDIA_ENCRYPT_NONE;
		}

		/* If we've been explicitly configured to use the received transport OR if
		 * encryption is on and crypto is present use the received transport.
		 * This is done in case of optimistic because it may come in as RTP/AVP or RTP/SAVP depending
		 * on the configuration of the remote endpoint (optimistic themselves or mandatory).
		 */
		if ((session->endpoint->media.rtp.use_received_transport) ||
			((encryption == AST_SIP_MEDIA_ENCRYPT_SDES) && !res)) {
			pj_strdup(session->inv_session->pool, &session_media->transport, &stream->desc.transport);
		}
	} else {
		/* This is bundled with another session, so mark it as such */
		ast_rtp_instance_bundle(session_media->rtp, session_media_transport->rtp);

		enable_rtcp(session, session_media, stream);
	}

	/* If ICE support is enabled find all the needed attributes */
	check_ice_support(session, session_media, stream);

	if (set_caps(session, session_media, session_media_transport, stream, 1, asterisk_stream)) {
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	pre_local_dir = ast_stream_get_local_media_direction(asterisk_stream);
	set_video_extra_attr_para(stream, asterisk_stream, media_type);
	set_gs_fec_attr_param(session, stream, session_media, asterisk_stream);
	determine_gs_fec_version(session_media, asterisk_stream, session_media->gs_fec_version);
	set_stream_local_media_direction_according_remote(asterisk_stream, stream);
	set_stream_remote_media_direction(asterisk_stream, stream);
	set_rtcp_attr_param(session, stream, session_media, asterisk_stream, session->endpoint->enable_nack);
	set_passthrough_param(stream, asterisk_stream);
	set_stream_video_resolution(session, session_media, asterisk_stream, stream);
	set_as_tias_for_stream(session, stream, asterisk_stream);
	set_gvc_msid_for_stream(session, session_media, stream);
	set_mid_for_stream(session, session_media, stream);
	set_trickle_ice_for_stream(session, session_media, stream, sdp);
	set_media_change_according_direction(session, session_media, pre_local_dir, asterisk_stream);
	set_rtcp_rsize_for_stream(session, session_media, stream);
	set_avs_ice_restart(session, session_media, stream, sdp);
	process_rtx_codec(session, session_media, stream, asterisk_stream);
	process_remote_ssrc_attr(session, session_media, asterisk_stream, stream);
	set_stream_state_for_page(session, asterisk_stream,stream);
#endif

	return 1;
}

static int add_crypto_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	enum ast_rtp_dtls_hash hash;
	const char *crypto_attribute;
	struct ast_rtp_engine_dtls *dtls;
	struct ast_sdp_srtp *tmp;
	static const pj_str_t STR_NEW = { "new", 3 };
	static const pj_str_t STR_EXISTING = { "existing", 8 };
	static const pj_str_t STR_ACTIVE = { "active", 6 };
	static const pj_str_t STR_PASSIVE = { "passive", 7 };
	static const pj_str_t STR_ACTPASS = { "actpass", 7 };
	static const pj_str_t STR_HOLDCONN = { "holdconn", 8 };
	enum ast_rtp_dtls_setup setup;

	switch (session_media->encryption) {
	case AST_SIP_MEDIA_ENCRYPT_NONE:
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
	case AST_SIP_MEDIA_ENCRYPT_ZRTP:
	case AST_SIP_MEDIA_ENCRYPT_ZRTP_NONE:
	case AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP:
#endif
		break;
	case AST_SIP_MEDIA_ENCRYPT_SDES:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
#endif
		if (!session_media->srtp) {
			session_media->srtp = ast_sdp_srtp_alloc();
			if (!session_media->srtp) {
				return -1;
			}
		}

		tmp = session_media->srtp;

#ifdef GRANDSTREAM_NETWORKS
		ast_sdp_srtp_set_config_suite(tmp, session->endpoint->media.rtp.srtp_suite);
#endif

		do {
			crypto_attribute = ast_sdp_srtp_get_attrib(tmp,
				0 /* DTLS running? No */,
				session->endpoint->media.rtp.srtp_tag_32 /* 32 byte tag length? */);
			if (!crypto_attribute) {
				/* No crypto attribute to add, bad news */
				return -1;
			}

			attr = pjmedia_sdp_attr_create(pool, "crypto",
				pj_cstr(&stmp, crypto_attribute));
			media->attr[media->attr_count++] = attr;
		} while ((tmp = AST_LIST_NEXT(tmp, sdp_srtp_list)));

		break;
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
#ifdef GRANDSTREAM_NETWORKS
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
#endif
		if (setup_dtls_srtp(session, session_media)) {
			return -1;
		}

		dtls = ast_rtp_instance_get_dtls(session_media->rtp);
		if (!dtls) {
			return -1;
		}

		switch (dtls->get_connection(session_media->rtp)) {
		case AST_RTP_DTLS_CONNECTION_NEW:
			attr = pjmedia_sdp_attr_create(pool, "connection", &STR_NEW);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_CONNECTION_EXISTING:
			attr = pjmedia_sdp_attr_create(pool, "connection", &STR_EXISTING);
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
		}

		/* If this is an answer we need to use our current state, if it's an offer we need to use
		 * the configured value.
		 */
		if (session->inv_session->neg
			&& pjmedia_sdp_neg_get_state(session->inv_session->neg) != PJMEDIA_SDP_NEG_STATE_DONE) {
			setup = dtls->get_setup(session_media->rtp);
		} else {
			setup = session->endpoint->media.rtp.dtls_cfg.default_setup;
		}

		switch (setup) {
		case AST_RTP_DTLS_SETUP_ACTIVE:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTIVE);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_PASSIVE:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_PASSIVE);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_ACTPASS:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_ACTPASS);
			media->attr[media->attr_count++] = attr;
			break;
		case AST_RTP_DTLS_SETUP_HOLDCONN:
			attr = pjmedia_sdp_attr_create(pool, "setup", &STR_HOLDCONN);
			break;
		default:
			break;
		}

		hash = dtls->get_fingerprint_hash(session_media->rtp);
		crypto_attribute = dtls->get_fingerprint(session_media->rtp);
		if (crypto_attribute && (hash == AST_RTP_DTLS_HASH_SHA1 || hash == AST_RTP_DTLS_HASH_SHA256)) {
			RAII_VAR(struct ast_str *, fingerprint, ast_str_create(64), ast_free);
			if (!fingerprint) {
				return -1;
			}

			if (hash == AST_RTP_DTLS_HASH_SHA1) {
				ast_str_set(&fingerprint, 0, "SHA-1 %s", crypto_attribute);
			} else {
				ast_str_set(&fingerprint, 0, "SHA-256 %s", crypto_attribute);
			}

			attr = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&stmp, ast_str_buffer(fingerprint)));
			media->attr[media->attr_count++] = attr;
		}
		break;
	}

	return 0;
}

/*! \brief Function which creates an outgoing stream */
static int create_outgoing_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				      struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_session *remote, struct ast_stream *stream)
{
#ifdef GRANDSTREAM_NETWORKS
	pj_pool_t *pool = NULL;
#else
	pj_pool_t *pool = session->inv_session->pool_prov;
#endif
#ifndef GRANDSTREAM_NETWORKS
	static const pj_str_t STR_RTP_AVP = { "RTP/AVP", 7 };
#endif
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	static const pj_str_t STR_SENDONLY = { "sendonly", 8 };
	pjmedia_sdp_media *media;
	const char *hostip = NULL;
	struct ast_sockaddr addr;
	char tmp[512];
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	int index = 0;
#ifdef GRANDSTREAM_NETWORKS
	int noncodec = (session->dtmf == AST_SIP_DTMF_RFC_4733 || session->dtmf == AST_SIP_DTMF_AUTO || session->dtmf == AST_SIP_DTMF_AUTO_INFO || session->dtmf == AST_SIP_DTMF_BOTH_RFC4733_AND_INFO) ? AST_RTP_DTMF : 0;
#else
	int noncodec = (session->dtmf == AST_SIP_DTMF_RFC_4733 || session->dtmf == AST_SIP_DTMF_AUTO || session->dtmf == AST_SIP_DTMF_AUTO_INFO) ? AST_RTP_DTMF : 0;
#endif
	int min_packet_size = 0, max_packet_size = 0;
	int rtp_code;
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	enum ast_media_type media_type = session_media->type;
	struct ast_sip_session_media *session_media_transport;
	pj_sockaddr ip;

#ifdef GRANDSTREAM_NETWORKS
	int first_video_pt = 0;
	char first_video_name[16] = { 0 };
	unsigned int set_turn_address = 0;

	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
		return create_outgoing_application_bfcp_sdp_stream(session, session_media, sdp, remote, stream);
	}

	if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
		pool = session->inv_session->pool_prov;
	} else {
		pool = session->inv_session->pool_notify_prov;
	}

	ast_log(LOG_DEBUG, "[session:%p] create_outgoing_sdp_stream for session_media->sub_type:%s, endpoint:%s\n",
		session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
#endif

	int direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);

	media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media));
	if (!media) {
		return -1;
	}
	pj_strdup2(pool, &media->desc.media, ast_codec_media_type2str(session_media->type));

	/* If this is a removed (or declined) stream OR if no formats exist then construct a minimal stream in SDP */
	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED || !ast_stream_get_formats(stream) ||
		!ast_format_cap_count(ast_stream_get_formats(stream))) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "[session:%p] create_outgoing_sdp_stream for session_media->sub_type:%s, endpoint:%s,attr_count:%d\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint),sdp->attr_count);

		create_outgoing_disable_sdp(pool, session, session_media, sdp, remote, stream);
#else
		media->desc.port = 0;
		media->desc.port_count = 1;

		if (remote && remote->media[ast_stream_get_position(stream)]) {
			pjmedia_sdp_media *remote_media = remote->media[ast_stream_get_position(stream)];
			int index;

			media->desc.transport = remote_media->desc.transport;

			/* Preserve existing behavior by copying the formats provided from the offer */
			for (index = 0; index < remote_media->desc.fmt_count; ++index) {
				media->desc.fmt[index] = remote_media->desc.fmt[index];
			}
			media->desc.fmt_count = remote_media->desc.fmt_count;
		} else {
			/* This is actually an offer so put a dummy payload in that is ignored and sane transport */
			media->desc.transport = STR_RTP_AVP;
			pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], "32");
		}

		sdp->media[sdp->media_count++] = media;
#endif
		ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);

		return 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	set_turn_address = judge_if_need_set_turn_address(session, session_media);

	if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
		session_media_transport = ast_sip_session_media_get_transport(session, session_media);
	} else {
		session_media_transport = ast_sip_session_get_media_for_participant(session, session_media);
	}

	if (session_media_transport != session_media && session_media->bundled && session_media_transport->rtp) {
		session_media->parent_avs_fd = ast_rtp_instance_get_avs_fd(session_media_transport->rtp);
	}

	set_media_net_type_for_local_offer(session, session_media, set_turn_address);

	if( session->endpoint->media.rtp.configure_zrtp == AST_SIP_MEDIA_ENCRYPT_ZRTP)
		session_media->zrtp = 1;

#endif

	if (!session_media->rtp && create_rtp(session, session_media, sdp)) {
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	set_target_address_for_local_offer(session, session_media, set_turn_address);
	set_bundle_group_for_mobile(session, session_media);
#endif

	/* If this stream has not been bundled already it is new and we need to ensure there is no SSRC conflict */
	if (session_media->bundle_group != -1 && !session_media->bundled) {
		for (index = 0; index < sdp->media_count; ++index) {
			struct ast_sip_session_media *other_session_media;
#ifdef GRANDSTREAM_NETWORKS
			if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
				other_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);
			} else {
				other_session_media = session->participant_bundled_media;
			}
			if (!other_session_media) {
				ast_log(LOG_DEBUG, "[session:%p] Can't get according session_media sub_type:%s, endpoint:%s\n",
					session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
				continue;
			}
#else
			other_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);
#endif
			if (!other_session_media->rtp || other_session_media->bundle_group != session_media->bundle_group) {
				continue;
			}

			if (ast_rtp_instance_get_ssrc(session_media->rtp) == ast_rtp_instance_get_ssrc(other_session_media->rtp)) {
				ast_rtp_instance_change_source(session_media->rtp);
				/* Start the conflict check over again */
				index = -1;
				continue;
			}
		}
	}

#ifndef GRANDSTREAM_NETWORKS
	session_media_transport = ast_sip_session_media_get_transport(session, session_media);
#endif

	if (session_media_transport == session_media || !session_media->bundled) {
		set_ice_components(session, session_media);
		enable_rtcp(session, session_media, NULL);

		/* Crypto has to be added before setting the media transport so that SRTP is properly
		 * set up according to the configuration. This ends up changing the media transport.
		 */
#ifndef GRANDSTREAM_NETWORKS
		if (add_crypto_to_stream(session, session_media, pool, media)) {
			return -1;
		}
#endif

		if (pj_strlen(&session_media->transport)) {
			/* If a transport has already been specified use it */
			media->desc.transport = session_media->transport;
		} else {
#ifdef GRANDSTREAM_NETWORKS
			ast_sip_session_set_media_transport(session, session_media, media);
#else
			media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
				!session->endpoint->media.rtp.encryption_optimistic &&
					(session_media->encryption == AST_SIP_MEDIA_ENCRYPT_SDES),
				session_media->rtp, session->endpoint->media.rtp.use_avpf,
				session->endpoint->media.rtp.force_avp));
#endif
		}

		media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn));
		if (!media->conn) {
			return -1;
		}

		/* Add connection level details */
#ifdef GRANDSTREAM_NETWORKS
		hostip = get_media_host_ip(session, session_media, media, direct_media_enabled);
#else
		if (direct_media_enabled) {
			hostip = ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR);
		} else if (ast_strlen_zero(session->endpoint->media.address)) {
			hostip = ast_sip_get_host_ip_string(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET());
		} else {
			hostip = session->endpoint->media.address;
		}
#endif

		if (ast_strlen_zero(hostip)) {
			ast_log(LOG_ERROR, "No local host IP available for stream %s\n",
				ast_codec_media_type2str(session_media->type));
			return -1;
		}

		media->conn->net_type = STR_IN;
		/* Assume that the connection will use IPv4 until proven otherwise */
		media->conn->addr_type = STR_IP4;
		pj_strdup2(pool, &media->conn->addr, hostip);

		if ((pj_sockaddr_parse(pj_AF_UNSPEC(), 0, &media->conn->addr, &ip) == PJ_SUCCESS) &&
			(ip.addr.sa_family == pj_AF_INET6())) {
			media->conn->addr_type = STR_IP6;
		}

		/* Add ICE attributes and candidates */
#ifdef GRANDSTREAM_NETWORKS

		if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
			if (add_crypto_to_stream(session, session_media, pool, media)) {
				return -1;
			}
		}

		if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
			add_ice_to_stream(session, session_media, pool, media, 1);
		}
#else
		add_ice_to_stream(session, session_media, pool, media, 1);
#endif

		ast_rtp_instance_get_local_address(session_media->rtp, &addr);
#ifdef GRANDSTREAM_NETWORKS
		set_media_port(session, session_media, media, direct_media_enabled, &addr);
#else
		media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session_media->direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&addr);
#endif
		media->desc.port_count = 1;
	} else {
#ifdef GRANDSTREAM_NETWORKS
		create_bundled_media_line(session, session_media_transport, sdp, media, stream, pool);
		add_bundle_only_to_stream(session, pool, media);
#else
		pjmedia_sdp_media *bundle_group_stream = sdp->media[session_media_transport->stream_num];

		/* As this is in a bundle group it shares the same details as the group instance */
		media->desc.transport = bundle_group_stream->desc.transport;
		media->conn = bundle_group_stream->conn;
		media->desc.port = bundle_group_stream->desc.port;
#endif

#ifdef GRANDSTREAM_NETWORKS
		if (add_crypto_to_stream(session, session_media_transport, pool, media)) {
			ast_log(LOG_ERROR, "[session:%p] Add crypto to stream error, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}

		if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
			add_ice_to_stream(session, session_media_transport, pool, media, 0);
		}
#else
		if (add_crypto_to_stream(session, session_media_transport, pool, media)) {
			return -1;
		}

		add_ice_to_stream(session, session_media_transport, pool, media, 0);
#endif

		enable_rtcp(session, session_media, NULL);
	}

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n",
			ast_codec_media_type2str(session_media->type));
		return -1;
	}

	if (direct_media_enabled) {
		ast_format_cap_get_compatible(session->endpoint->media.codecs, session->direct_media_cap, caps);
	} else {
#ifdef GRANDSTREAM_NETWORKS
		/* If the endpoint is trunk, the joint cap order is not affected by the caller codec order */
		if (session->is_outgoing && !strncasecmp(session->endpoint->aors, "trunk_", 6) &&
			media_type == AST_MEDIA_TYPE_AUDIO) {
			ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, media_type);
			ast_sip_session_change_gs_fec_version(caps);
			if (ast_format_cap_has_opus(caps)) {
				/* set useinbandfec true as default */
				ast_sip_session_set_format_useinbandfec(session, caps, stream, 1);
				ast_sip_session_set_format_dtx(session, caps, stream);
			}
		}

		/*Start added by zyxia for reinvite after 488*/
		else if (session->wSwitchImageInvite488ToAudioInvite) {
			ast_format_cap_append(caps, ast_format_ulaw, 0);
			ast_format_cap_append(caps, ast_format_alaw, 0);
			if (session && session->channel) {
				ast_channel_nativeformats_set(session->channel, caps);
			}
		}
		/*End added by zyxia for reinvite after 488*/
		else {
			ast_format_cap_append_from_cap(caps, ast_stream_get_formats(stream), media_type);
		}
#else
		ast_format_cap_append_from_cap(caps, ast_stream_get_formats(stream), media_type);
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ast_stream_judge_if_main_video(stream)) {
		session->main_video_first_multi_h264_pt = 0;
		session->main_video_first_multi_rtx_pt = 0;
	}
#endif

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		struct ast_format *format = ast_format_cap_get_format(caps, index);

		if (ast_format_get_type(format) != media_type) {
			ao2_ref(format, -1);
			continue;
		}

#ifdef GRANDSTREAM_NETWORKS
		if (create_outgoing_sdp_multi_pt(session, format, pool, stream, media)) {
			ao2_ref(format, -1);
			continue;
		}

		process_h264_format_no_profile_level_id(session, &format);

		if (!strcasecmp(ast_format_get_name(format), "fec") && (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && !session->support_gs_fec)) {
			ast_log(LOG_DEBUG, "[session:%p] Codec is FEC, webrtc can't support, don't add it, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(format, -1);
			continue;
		}

		if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media->rtp), 1, format, 0)) == -1) {
			ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
			ao2_ref(format, -1);
			continue;
		}

		if (!strcasecmp(ast_format_get_name(format), "H264") && session->main_video_first_multi_h264_pt) {
			ast_log(LOG_DEBUG, "[session:%p] Main video H264 has multi codec, just use the first H264 codec PT as main video, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			rtp_code = session->main_video_first_multi_h264_pt;
		}
#else
		/* If this stream is not a transport we need to use the transport codecs structure for payload management to prevent
		 * conflicts.
		 */
		if (session_media_transport != session_media) {
			if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media_transport->rtp), 1, format, 0)) == -1) {
				ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
				ao2_ref(format, -1);
				continue;
			}
			/* Our instance has to match the payload number though */
			ast_rtp_codecs_payload_set_rx(ast_rtp_instance_get_codecs(session_media->rtp), rtp_code, format);
		} else {
			if ((rtp_code = ast_rtp_codecs_payload_code(ast_rtp_instance_get_codecs(session_media->rtp), 1, format, 0)) == -1) {
				ast_log(LOG_WARNING,"Unable to get rtp codec payload code for %s\n", ast_format_get_name(format));
				ao2_ref(format, -1);
				continue;
			}
		}
#endif

#ifdef GRANDSTREAM_NETWORKS
		if (!first_video_pt && judge_if_support_rtx_video_format(format)) {
			first_video_pt = rtp_code;
			strncpy(first_video_name, ast_format_get_name(format), sizeof(first_video_name) - 1);
		}
		rtp_code = change_rtx_pt(session, session_media, format, first_video_name, rtp_code);
		change_rtx_apt_value(session, &format, first_video_pt);
		change_p2p_level_id_according_configure(session, stream, &format);
#endif

		if ((attr = generate_rtpmap_attr(session, media, pool, rtp_code, 1, format, 0))) {
			media->attr[media->attr_count++] = attr;
		}

		if ((attr = generate_fmtp_attr(pool, format, rtp_code))) {
			media->attr[media->attr_count++] = attr;
		}
#ifdef GRANDSTREAM_NETWORKS
		else if (!strcasecmp(ast_format_get_name(format), "fec")) {
			ast_log(LOG_DEBUG, "It's GS-FEC, add gs-fec-version to stream\n");
			add_gs_fec_version_to_stream(session, pool, media, session_media, stream);
		}

		/* Add by bfdu, now we just support h264 codec for imageattr */
		if (!strcasecmp(ast_format_get_name(format), "h264")) {
			add_imageattr_to_stream(pool, rtp_code, media, stream);
		}
#endif

		if (ast_format_get_maximum_ms(format) &&
			((ast_format_get_maximum_ms(format) < max_packet_size) || !max_packet_size)) {
			max_packet_size = ast_format_get_maximum_ms(format);
		}
		ao2_ref(format, -1);

		if (media->desc.fmt_count == PJMEDIA_MAX_SDP_FMT) {
			break;
		}
	}

	/* Add non-codec formats */
#ifdef GRANDSTREAM_NETWORKS
	if (ast_sip_session_is_pending_stream_default(session, stream) && media_type != AST_MEDIA_TYPE_VIDEO
		&& media_type != AST_MEDIA_TYPE_APPLICATION && media->desc.fmt_count < PJMEDIA_MAX_SDP_FMT)
#else
	if (ast_sip_session_is_pending_stream_default(session, stream) && media_type != AST_MEDIA_TYPE_VIDEO
		&& media->desc.fmt_count < PJMEDIA_MAX_SDP_FMT)
#endif
	{
		for (index = 1LL; index <= AST_RTP_MAX; index <<= 1) {
			if (!(noncodec & index)) {
				continue;
			}
			rtp_code = ast_rtp_codecs_payload_code(
				ast_rtp_instance_get_codecs(session_media->rtp), 0, NULL, index);
			if (rtp_code == -1) {
				continue;
			}

			if ((attr = generate_rtpmap_attr(session, media, pool, rtp_code, 0, NULL, index))) {
				media->attr[media->attr_count++] = attr;
			}

			if (index == AST_RTP_DTMF) {
#ifdef GRANDSTREAM_NETWORKS
				snprintf(tmp, sizeof(tmp), "%d 0-15", rtp_code);  /* Modified by yfyang, RFC4733 decrepate 16 */
#else
				snprintf(tmp, sizeof(tmp), "%d 0-16", rtp_code);
#endif
				attr = pjmedia_sdp_attr_create(pool, "fmtp", pj_cstr(&stmp, tmp));
				media->attr[media->attr_count++] = attr;
			}

			if (media->desc.fmt_count == PJMEDIA_MAX_SDP_FMT) {
				break;
			}
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for CN codec */
	add_comfort_noise_to_sdp(session, session_media, media, caps, pool, remote);
	add_mpa_to_sdp(pool, session, session_media, media);
#endif

	/* If no formats were actually added to the media stream don't add it to the SDP */
	if (!media->desc.fmt_count) {
		return 1;
	}

	/* If ptime is set add it as an attribute */
	min_packet_size = ast_rtp_codecs_get_framing(ast_rtp_instance_get_codecs(session_media->rtp));
	if (!min_packet_size) {
		min_packet_size = ast_format_cap_get_framing(caps);
	}
	if (min_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", min_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "ptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

#ifndef GRANDSTREAM_NETWORKS
	if (max_packet_size) {
		snprintf(tmp, sizeof(tmp), "%d", max_packet_size);
		attr = pjmedia_sdp_attr_create(pool, "maxptime", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}
#endif

	/* Add the sendrecv attribute - we purposely don't keep track because pjmedia-sdp will automatically change our offer for us */
	attr = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_attr);
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for forward media direction, 2016.11.08 */
	set_local_media_direction_for_page(session, stream);
	if (jugde_if_add_direction_as_local_value(session, stream)) {
		add_stream_local_media_direction(stream, attr);
	} else {
		attr->name = !session_media->locally_held ? STR_SENDRECV : STR_SENDONLY;
	}
#else
	attr->name = !session_media->locally_held ? STR_SENDRECV : STR_SENDONLY;
#endif
	media->attr[media->attr_count++] = attr;

	/* If we've got rtcp-mux enabled, add it unless we received an offer without it */
	if (session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux) {
		attr = pjmedia_sdp_attr_create(pool, "rtcp-mux", NULL);
		pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (ast_format_cap_has_rtx(caps)) {
		add_rtx_ssrc_group_to_stream(session, session_media, pool, media, stream);
	} else {
		if (session->endpoint->media.support_special_attrs) {
			add_ssrc_cname_to_stream(session, session_media, pool, media, AST_SSRC_TYPE_MEDIA);
			add_ssrc_msid_to_stream(session, session_media, pool, media, stream, AST_SSRC_TYPE_MEDIA);
		}
	}

	if (ast_format_cap_has_video_fec(caps)) {
		add_fec_ssrc_group_to_stream(session, session_media, pool, media, stream);
	}
#else
	add_ssrc_to_stream(session, session_media, pool, media);
	add_msid_to_stream(session, session_media, pool, media, stream);
#endif

	add_rtcp_fb_to_stream(session, session_media, pool, media);
	add_extmap_to_stream(session, session_media, pool, media);

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.13 */
	if (AST_MEDIA_TYPE_VIDEO == media_type) {
		add_video_extra_attr_to_stream(pool, media, stream);
		ast_sip_session_modify_outgoing_yealink_sdp(session, pool, media);
	}

	add_passthrough_param_to_stream(pool, media, stream);

	add_rtcp_attrs_to_stream(session, session_media, pool, media, stream);

	if (session->endpoint->media.support_special_attrs) {
		add_as_tias_to_steam(pool, media, stream);
	}

	add_trickle_ice_to_stream(session, session_media, pool, media);

	ast_sip_session_add_decline_video_reason(pool, session, media, stream);

	add_record_indicate_attr(session, pool, media);

	add_msid_attr_to_stream(session, session_media, pool, media);

	add_content_attr_to_stream(session, stream, pool, media);

	remove_connection_attr_for_gswave(session, media);

	add_uptime_attr_to_stream(session, stream, pool, media);
	add_zrtphash_attr_to_stream(session, session_media, pool,media, stream);
#endif

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;

	return 1;
}

static struct ast_frame *media_session_rtp_read_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_frame *f;

	if (!session_media->rtp) {
		return &ast_null_frame;
	}

	f = ast_rtp_instance_read(session_media->rtp, 0);
	if (!f) {
		return NULL;
	}

	ast_rtp_instance_set_last_rx(session_media->rtp, time(NULL));

	return f;
}

static struct ast_frame *media_session_rtcp_read_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_frame *f;

	if (!session_media->rtp) {
		return &ast_null_frame;
	}

	f = ast_rtp_instance_read(session_media->rtp, 1);
	if (!f) {
		return NULL;
	}

	ast_rtp_instance_set_last_rx(session_media->rtp, time(NULL));

	return f;
}

static int media_session_rtp_write_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media, struct ast_frame *frame)
{
	if (!session_media->rtp) {
		return 0;
	}

	return ast_rtp_instance_write(session_media->rtp, frame);
}

static void set_port_update_flag(struct ast_sip_session *session,
	struct ast_rtp_instance *rtp, struct ast_sockaddr *new_addrs, int new_rtcp_port)
{
	struct ast_sockaddr old_addrs;
	int old_rtcp_port = 0;

	if (!session || !session->inv_session || !rtp) {
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ||
		session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE ||
		session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK ||
		session->support_ice) {
		return;
	}

	if (!new_rtcp_port) {
		ast_rtp_instance_set_port_update_flag(rtp, 0);
	}

	if (new_addrs) {
		ast_rtp_instance_get_remote_address(rtp, &old_addrs);
		if (ast_sockaddr_cmp(new_addrs, &old_addrs) != 0) {
			ast_rtp_instance_set_port_update_flag(rtp, 1);
		}
	}

	if (new_rtcp_port) {
		old_rtcp_port =  ast_rtp_instance_get_remote_rtcp(rtp);
		if (old_rtcp_port != new_rtcp_port) {
			ast_rtp_instance_set_port_update_flag(rtp, 1);
		}
	}

	return;
}

static int apply_negotiated_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_session *local,
	const struct pjmedia_sdp_session *remote, int index, struct ast_stream *asterisk_stream)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	struct pjmedia_sdp_media *remote_stream = remote->media[index];
	enum ast_media_type media_type = session_media->type;
	char host[NI_MAXHOST];
	int res;
	struct ast_sip_session_media *session_media_transport;
#ifdef GRANDSTREAM_NETWORKS
	struct pjmedia_sdp_media *final_stream = NULL;
#endif

	if (!session->channel) {
		return 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->is_direct_media) {
		ast_log(LOG_DEBUG, "Enable direct media, don't apply SDP\n");
		return 1;
	}

	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	struct pjmedia_sdp_media *local_stream = local->media[index];
	ast_sip_session_set_media_sub_type(session, local_stream, session_media, asterisk_stream);

	if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
		int app_res = 0;
		app_res = apply_negotiated_application_bfcp_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
		return app_res;
	}

	if (check_media_transport(session, session_media, remote_stream)) {
		return -1;
	}

	set_trickle_ice_for_stream(session, session_media, remote_stream, remote);

	set_avs_ice_flag(session, session_media, remote_stream);

	if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, asterisk_stream)) {
		session_media_transport = ast_sip_session_media_get_transport(session, session_media);
	} else {
		session_media_transport = ast_sip_session_get_media_for_participant(session, session_media);
	}

	if (session_media_transport != session_media && session_media->bundled && session_media_transport->rtp) {
		session_media->parent_avs_fd = ast_rtp_instance_get_avs_fd(session_media_transport->rtp);
	}
#else
	/* Ensure incoming transport is compatible with the endpoint's configuration */
	if (!session->endpoint->media.rtp.use_received_transport &&
		check_endpoint_media_transport(session->endpoint, remote_stream) == AST_SIP_MEDIA_TRANSPORT_INVALID) {
		return -1;
	}
#endif

	/* Create an RTP instance if need be */
	if (!session_media->rtp && create_rtp(session, session_media, local)) {
		return -1;
	}

#ifndef GRANDSTREAM_NETWORKS
	process_ssrc_attributes(session, session_media, remote_stream);
#endif
	process_extmap_attributes(session, session_media, remote_stream);

#ifndef GRANDSTREAM_NETWORKS
	session_media_transport = ast_sip_session_media_get_transport(session, session_media);
#endif

	if (session_media_transport == session_media || !session_media->bundled) {
		session_media->remote_rtcp_mux = (pjmedia_sdp_media_find_attr2(remote_stream, "rtcp-mux", NULL) != NULL);
		set_ice_components(session, session_media);

		enable_rtcp(session, session_media, remote_stream);

		res = setup_media_encryption(session, session_media, remote, remote_stream);
		if (!session->endpoint->media.rtp.encryption_optimistic && res) {
			/* If optimistic encryption is disabled and crypto should have been enabled but was not
			 * this session must fail.
			 */
			return -1;
		}

		if (!remote_stream->conn && !remote->conn) {
			return 1;
		}

		ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

		/* Ensure that the address provided is valid */
		if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
			/* The provided host was actually invalid so we error out this negotiation */
			return -1;
		}

		/* Apply connection information to the RTP instance */
#ifdef GRANDSTREAM_NETWORKS
		if (!session->media_turn_relay) {
			ast_sockaddr_set_port(addrs, remote_stream->desc.port);
			set_port_update_flag(session, session_media->rtp, addrs, 0);
		}
		set_target_address_for_trunk(session, session_media, addrs, remote_stream->desc.port);
#else
		ast_sockaddr_set_port(addrs, remote_stream->desc.port);
#endif
		ast_rtp_instance_set_remote_address(session_media->rtp, addrs);

		ast_sip_session_media_set_write_callback(session, session_media, media_session_rtp_write_callback);
		ast_sip_session_media_add_read_callback(session, session_media, ast_rtp_instance_fd(session_media->rtp, 0),
			media_session_rtp_read_callback);
		if (!session->endpoint->media.rtcp_mux || !session_media->remote_rtcp_mux) {
			ast_sip_session_media_add_read_callback(session, session_media, ast_rtp_instance_fd(session_media->rtp, 1),
				media_session_rtcp_read_callback);
		}

		/* If ICE support is enabled find all the needed attributes */
		process_ice_attributes(session, session_media, remote, remote_stream);
	} else {
		/* This is bundled with another session, so mark it as such */
		ast_rtp_instance_bundle(session_media->rtp, session_media_transport->rtp);
		ast_sip_session_media_set_write_callback(session, session_media, media_session_rtp_write_callback);
		enable_rtcp(session, session_media, remote_stream);
	}
#ifdef GRANDSTREAM_NETWORKS
	if (pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) ||
		(ast_sip_session_media_use_bundle(session) && ast_stream_judge_if_stream_is_participant(asterisk_stream))) {
		final_stream = remote_stream;
	} else {
		final_stream = local_stream;
	}
	if (set_caps(session, session_media, session_media_transport, final_stream, 0, asterisk_stream)) {
		return -1;
	}
#else
	if (set_caps(session, session_media, session_media_transport, remote_stream, 0, asterisk_stream)) {
		return 1;
	}
#endif

	/* Set the channel uniqueid on the RTP instance now that it is becoming active */
	ast_channel_lock(session->channel);
	ast_rtp_instance_set_channel_id(session_media->rtp, ast_channel_uniqueid(session->channel));
	ast_channel_unlock(session->channel);

	/* Ensure the RTP instance is active */
	ast_rtp_instance_set_stream_num(session_media->rtp, ast_stream_get_position(asterisk_stream));
	ast_rtp_instance_activate(session_media->rtp);

#ifdef GRANDSTREAM_NETWORKS
	process_remote_ssrc_attr(session, session_media, asterisk_stream, remote_stream);
	process_remote_stream_port(session, session_media, remote_stream->desc.port);
	set_gs_fec_attr_param(session, remote_stream, session_media, asterisk_stream);
	determine_gs_fec_version(session_media, asterisk_stream, session_media->gs_fec_version);
	if (pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg)) {
		set_stream_local_media_direction_according_remote(asterisk_stream, remote_stream);
	} else {
		set_stream_local_media_direction_according_local(asterisk_stream, local_stream);
	}
	set_stream_remote_media_direction(asterisk_stream, remote_stream);
	set_as_tias_for_stream(session, final_stream, asterisk_stream);

	set_rtcp_attr_param(session, remote_stream, session_media, asterisk_stream, session->endpoint->enable_nack);
	set_video_extra_attr_para(final_stream, asterisk_stream, media_type);
	set_avs_params(session, session_media, local_stream, asterisk_stream);
	ast_stream_set_media_encryption(asterisk_stream, session_media->encryption);
	set_stream_final_video_resolution(session, asterisk_stream);
	update_cei_for_change_video_resolution(session, asterisk_stream);
	indicate_wave_release_hd_video_resource(session, asterisk_stream);
	clear_video_codec_rtx_pt(session_media, asterisk_stream);
#endif

	/* audio stream handles music on hold */
	if (media_type != AST_MEDIA_TYPE_AUDIO) {
		if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
			&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
			ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
		}
		return 1;
	}

#ifdef GRANDSTREAM_NETWORKS
	if ((addrs && ast_sockaddr_isnull(addrs) && !judge_if_support_trickle_ice(session, session_media)) ||
		(addrs && ast_sockaddr_is_any(addrs) && !judge_if_support_trickle_ice(session, session_media)) ||
#else
	if (ast_sockaddr_isnull(addrs) ||
		ast_sockaddr_is_any(addrs) ||
#endif
		pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL) ||
		pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		if (!session_media->remotely_held) {
#ifdef GRANDSTREAM_NETWORKS
			/* Added by yfyang, for sca feature */
			if (session->line_index) {
				enum sca_status state = ast_sip_get_sca_line_status((char *)session->endpoint->sca_sharedline, session->line_index);
				ast_log(LOG_NOTICE, "Sharedline[%s], index[%d], the line state is [%s]",
						 session->endpoint->sca_sharedline, session->line_index, ast_sip_sca_state2str(state));

				if ((SCA_Active == state) && (session->recv_held_private)) {
					ast_sip_set_sca_line_status((char *)session->endpoint->sca_sharedline, session->line_index, SCA_Held_Private); 
				}

				if (SCA_Bridge_Held != state) {
					ast_queue_hold(session->channel, session->endpoint->mohsuggest);
				}
			} else {
				ast_queue_hold(session->channel, session->endpoint->mohsuggest);
			}

			if (ast_channel_is_avs_engine(session->channel)) {
				ast_avs_control(session->channel, AVS_CONTROL_HOLD, NULL, 0);
			}
#else
			/* The remote side has put us on hold */
			ast_queue_hold(session->channel, session->endpoint->mohsuggest);
#endif
			ast_rtp_instance_stop(session_media->rtp);
			ast_queue_frame(session->channel, &ast_null_frame);
			session_media->remotely_held = 1;
		}
	} else if (session_media->remotely_held) {
		/* The remote side has taken us off hold */
		ast_queue_unhold(session->channel);
		ast_queue_frame(session->channel, &ast_null_frame);
		session_media->remotely_held = 0;
#ifdef GRANDSTREAM_NETWORKS
		if (ast_channel_is_avs_engine(session->channel)) {
			ast_avs_control(session->channel, AVS_CONTROL_UNHOLD, NULL, 0);
		}
#endif
	} else if ((pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_FALSE)
		&& (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		ast_queue_control(session->channel, AST_CONTROL_UPDATE_RTP_PEER);
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Modify by bfdu for Bug 149572, 2019.10.29 */
	reset_media_encryption(session, session_media);
#else
	/* This purposely resets the encryption to the configured in case it gets added later */
	session_media->encryption = session->endpoint->media.rtp.encryption;
#endif

	if (session->endpoint->media.rtp.keepalive > 0 &&
			session_media->type == AST_MEDIA_TYPE_AUDIO) {
		ast_rtp_instance_set_keepalive(session_media->rtp, session->endpoint->media.rtp.keepalive);
		/* Schedule the initial keepalive early in case this is being used to punch holes through
		 * a NAT. This way there won't be an awkward delay before media starts flowing in some
		 * scenarios.
		 */
		AST_SCHED_DEL(sched, session_media->keepalive_sched_id);
#ifdef GRANDSTREAM_NETWORKS
		if (!ast_channel_is_avs_engine(session->channel)) {
			session_media->keepalive_sched_id = ast_sched_add_variable(sched, 500, send_keepalive,
				session_media, 1);
		}
#else
		session_media->keepalive_sched_id = ast_sched_add_variable(sched, 500, send_keepalive,
			session_media, 1);
#endif
	}

	/* As the channel lock is not held during this process the scheduled item won't block if
	 * it is hanging up the channel at the same point we are applying this negotiated SDP.
	 */
	AST_SCHED_DEL(sched, session_media->timeout_sched_id);

	/* Due to the fact that we only ever have one scheduled timeout item for when we are both
	 * off hold and on hold we don't need to store the two timeouts differently on the RTP
	 * instance itself.
	 */
	ast_rtp_instance_set_timeout(session_media->rtp, 0);
	if (session->endpoint->media.rtp.timeout && !session_media->remotely_held) {
		ast_rtp_instance_set_timeout(session_media->rtp, session->endpoint->media.rtp.timeout);
	} else if (session->endpoint->media.rtp.timeout_hold && session_media->remotely_held) {
		ast_rtp_instance_set_timeout(session_media->rtp, session->endpoint->media.rtp.timeout_hold);
	}

	if (ast_rtp_instance_get_timeout(session_media->rtp)) {
#ifdef GRANDSTREAM_NETWORKS
		if (!ast_channel_is_avs_engine(session->channel)) {
			session_media->timeout_sched_id = ast_sched_add_variable(sched,
				ast_rtp_instance_get_timeout(session_media->rtp) * 1000, rtp_check_timeout,
				session_media, 1);
		}
#else
		session_media->timeout_sched_id = ast_sched_add_variable(sched,
			ast_rtp_instance_get_timeout(session_media->rtp) * 1000, rtp_check_timeout,
			session_media, 1);
#endif
	}

	return 1;
}

/*! \brief Function which updates the media stream with external media address, if applicable */
static void change_outgoing_sdp_stream_media_address(pjsip_tx_data *tdata, struct pjmedia_sdp_media *stream, struct ast_sip_transport *transport)
{
	RAII_VAR(struct ast_sip_transport_state *, transport_state, ast_sip_get_transport_state(ast_sorcery_object_get_id(transport)), ao2_cleanup);
	char host[NI_MAXHOST];
	struct ast_sockaddr our_sdp_addr = { { 0, } };

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, get external ip from STUN */
	char ext_addr[256] = {0};

	if (!transport_state || !transport) {
		return;
	}

	if (transport->get_extaddr_from_stun) {
		ast_sip_nat_get_external_ip(ext_addr, sizeof(ext_addr)-1);
	}

	if (!strlen(ext_addr)) {
		if (!ast_sockaddr_isnull(&transport_state->external_media_address)) {
			strncpy(ext_addr, ast_sockaddr_stringify_host(&transport_state->external_media_address), sizeof(ext_addr)-1);
		} else if (!ast_strlen_zero(transport->external_media_address)) {
			strncpy(ext_addr, transport->external_media_address, sizeof(ext_addr)-1);
		}
	}

#ifdef BUILD_ON_CLOUD_PBX
	char *clouducm_host_ip = getenv("PUBLIC_IP");
	if (!ast_strlen_zero(clouducm_host_ip)) {
		strncpy(ext_addr, clouducm_host_ip, sizeof(ext_addr)-1);
		ast_log(LOG_DEBUG, "Get HOST_IP:%s for cloud UCM\n", clouducm_host_ip);
	}
#endif


	if (!strlen(ext_addr)) {
		ast_log(LOG_DEBUG, "No external addres\n");
		return;
	}

	/* maybe media has no c line, modify rtcp address here */
	change_rtcp_to_external_address(tdata->pool, stream, ext_addr, transport_state);
#endif

	/* If the stream has been rejected there will be no connection line */
	if (!stream->conn || !transport_state) {
		return;
	}

	ast_copy_pj_str(host, &stream->conn->addr, sizeof(host));
	ast_sockaddr_parse(&our_sdp_addr, host, PARSE_PORT_FORBID);

	/* Reversed check here. We don't check the remote endpoint being
	 * in our local net, but whether our outgoing session IP is
	 * local. If it is not, we won't do rewriting. No localnet
	 * configured? Always rewrite. */
#ifndef BUILD_ON_CLOUD_PBX
	if (ast_sip_transport_is_nonlocal(transport_state, &our_sdp_addr) && transport_state->localnet) {
		return;
	}
#endif
	ast_debug(5, "Setting media address to %s\n", ast_sockaddr_stringify_host(&transport_state->external_media_address));

#ifndef GRANDSTREAM_NETWORKS
	pj_strdup2(tdata->pool, &stream->conn->addr, ast_sockaddr_stringify_host(&transport_state->external_media_address));
#else
	if (transport->ip_in_sdp_connection) {
		pj_strdup2(tdata->pool, &stream->conn->addr, ext_addr); 
	} else {
		if (!ast_strlen_zero(transport->external_media_address)) {
#ifdef BUILD_ON_CLOUD_PBX
			pj_strdup2(tdata->pool, &stream->conn->addr, ext_addr); 
#else
			pj_strdup2(tdata->pool, &stream->conn->addr, transport->external_media_address); 
#endif
		}
	}
#endif
}

/*! \brief Function which stops the RTP instance */
static void stream_stop(struct ast_sip_session_media *session_media)
{
	if (!session_media->rtp) {
		return;
	}

	AST_SCHED_DEL(sched, session_media->keepalive_sched_id);
	AST_SCHED_DEL(sched, session_media->timeout_sched_id);
	ast_rtp_instance_stop(session_media->rtp);
}

/*! \brief Function which destroys the RTP instance when session ends */
static void stream_destroy(struct ast_sip_session_media *session_media)
{
	if (session_media->rtp) {
		stream_stop(session_media);
		ast_rtp_instance_destroy(session_media->rtp);
	}
	session_media->rtp = NULL;

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (session_media->bfcp_udp) {
		AST_SCHED_DEL(sched, session_media->bfcp_sched_id);
		AST_SCHED_DEL(sched, session_media->bfcp_keepalive_id);
		ast_bfcp_destroy(session_media->bfcp_udp);
	}
	session_media->bfcp_udp = NULL;

	if (session_media->bfcp_tcp) {
		AST_SCHED_DEL(sched, session_media->bfcp_sched_id);
		ast_bfcp_tcp_destroy(session_media->bfcp_tcp);
	}
	session_media->bfcp_tcp = NULL;
#endif
}

/*! \brief SDP handler for 'audio' media stream */
static struct ast_sip_session_sdp_handler audio_sdp_handler = {
	.id = STR_AUDIO,
	.negotiate_incoming_sdp_stream = negotiate_incoming_sdp_stream,
	.create_outgoing_sdp_stream = create_outgoing_sdp_stream,
	.apply_negotiated_sdp_stream = apply_negotiated_sdp_stream,
	.change_outgoing_sdp_stream_media_address = change_outgoing_sdp_stream_media_address,
	.stream_stop = stream_stop,
	.stream_destroy = stream_destroy,
};

/*! \brief SDP handler for 'video' media stream */
static struct ast_sip_session_sdp_handler video_sdp_handler = {
	.id = STR_VIDEO,
	.negotiate_incoming_sdp_stream = negotiate_incoming_sdp_stream,
	.create_outgoing_sdp_stream = create_outgoing_sdp_stream,
	.apply_negotiated_sdp_stream = apply_negotiated_sdp_stream,
	.change_outgoing_sdp_stream_media_address = change_outgoing_sdp_stream_media_address,
	.stream_stop = stream_stop,
	.stream_destroy = stream_destroy,
};

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
/*! \brief SDP handler for 'application' media stream */
static struct ast_sip_session_sdp_handler application_sdp_handler = {
	.id = STR_APPLICATION,
	.negotiate_incoming_sdp_stream = negotiate_incoming_sdp_stream,
	.create_outgoing_sdp_stream = create_outgoing_sdp_stream,
	.apply_negotiated_sdp_stream = apply_negotiated_sdp_stream,
	.change_outgoing_sdp_stream_media_address = change_outgoing_sdp_stream_media_address,
	.stream_stop = stream_stop,
	.stream_destroy = stream_destroy,
};
#endif

static int video_info_incoming_request(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	struct pjsip_transaction *tsx;
	pjsip_tx_data *tdata;

	if (!session->channel
		|| !ast_sip_is_content_type(&rdata->msg_info.msg->body->content_type,
			"application",
			"media_control+xml")) {
		return 0;
	}

	tsx = pjsip_rdata_get_tsx(rdata);

	ast_queue_control(session->channel, AST_CONTROL_VIDUPDATE);

	if (pjsip_dlg_create_response(session->inv_session->dlg, rdata, 200, NULL, &tdata) == PJ_SUCCESS) {
		pjsip_dlg_send_response(session->inv_session->dlg, tsx, tdata);
	}

	return 0;
}

static struct ast_sip_session_supplement video_info_supplement = {
	.method = "INFO",
	.incoming_request = video_info_incoming_request,
};

#ifdef GRANDSTREAM_NETWORKS
static unsigned int judge_if_need_set_turn_address(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media)
{
	if (!session || !session_media) {
		return FALSE;
	}

	if (!session->media_turn_relay) {
		return FALSE;
	}

	if (!session_media->rtp) {
		ast_log(LOG_DEBUG, "[session:%p] Has no rtp instance for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return TRUE;
	}

	return FALSE;
}

static unsigned int jugde_if_add_direction_as_local_value(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!session || !stream) {
		return 0;
	}

	if (!ast_stream_get_local_media_direction(stream)) {
		return 0;
	}

	if (!session->is_outgoing) {
		return 1;
	}

	if (session->inv_session && (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		return 1;
	}

	if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_AUDIO &&
		(ast_stream_get_local_media_direction(stream) == INACTIVE ||
		ast_stream_get_local_media_direction(stream) == SENDONLY) &&
		!ast_sip_session_get_allow_forward_hold_request()) {

		#if 0
		/*Start added by zyxia for Bug 242256*/
		if(session->channel) {
			enum ast_page_mode page_mode = AST_PAGE_MODE_NONE;
			page_mode = ast_channel_page_mode_get(session->channel);
			if (page_mode == AST_PAGE_MODE_ANNOUNCE) {
				return 1;
			}
		}
		/*End added by zyxia for Bug 242256*/
		#endif
		ast_log(LOG_DEBUG, "[session:%p] It's not enable forward hold request, don't add direction as caller's value, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	return 1;
}


static void get_media_formats_by_order(struct ast_rtp_codecs *codecs, const struct pjmedia_sdp_media *stream,
	struct ast_format_cap *astformats, int *nonastformats)
{
	unsigned j = 0;
	struct ast_rtp_payload_type *type = NULL;

	ast_rwlock_rdlock(&codecs->codecs_lock);
	for (j=0; j<stream->desc.fmt_count; ++j) {
		int payload = pj_strtoul(&stream->desc.fmt[j]);

		if (payload < 0 || payload >= AST_RTP_MAX_PT) {
			break;
		}

		if (payload < AST_VECTOR_SIZE(&codecs->payload_mapping_tx)) {
			type = AST_VECTOR_GET(&codecs->payload_mapping_tx, payload);
			if (type) {
				if ( type->asterisk_format) {
					ast_format_cap_append(astformats, type->format, 0);
				} else {
					*nonastformats |= type->rtp_code;
				}
			}
		}
	}

	if (codecs->framing) {
		ast_format_cap_set_framing(astformats, codecs->framing);
	}
	ast_rwlock_unlock(&codecs->codecs_lock);
}

static int check_if_direct_media_enabled(struct ast_sip_session *session)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint, NULL, ao2_cleanup);

	if (session->channel) {
		if (session->is_direct_media) {
			ast_log(LOG_DEBUG, "[session:%p] check direct media is enabled afer channel created, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		} else {
			return 0;
		}
	} else {
		endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", session->exten);
		if (endpoint && endpoint->media.direct_media.enabled && session->endpoint->media.direct_media.enabled) {
			ast_log(LOG_DEBUG, "[session:%p] check direct media is enabled, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		} else {
			return 0;
		}
	}
}

static unsigned char judge_if_support_rtx_video_format(struct ast_format *format)
{
	if (!format) {
		return 0;
	}
	
	if (!strcasecmp(ast_format_get_name(format), "h264") ||
		!strcasecmp(ast_format_get_name(format), "h265") ||
		!strcasecmp(ast_format_get_name(format), "vp8") ||
		!strcasecmp(ast_format_get_name(format), "vp9")) {
		return 1;
	}

	return 0;
}

static unsigned char judge_if_caps_has_invalid_codec(struct ast_sip_session *session, struct ast_format_cap *caps)
{
	if (!session || !caps) {
		return 0;
	}

	if (!ast_format_cap_count(caps)) {
		return 0;
	}

	/* If video codec only has RTX or FEC, it's invalid */
	if (ast_format_cap_count(caps) == 1 && 
		(ast_format_cap_has_rtx(caps) || ast_format_cap_has_video_fec(caps))) {
		ast_log(LOG_DEBUG, "[session:%p] The joint cap is only had RTX or video FEC, video media negotiated failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/* If audio codec only has FEC and RED, it's invalid */
	if (ast_format_cap_count(caps) == 2 && ast_format_cap_has_audio_fec(caps) && ast_format_cap_has_red(caps)) {
		ast_log(LOG_DEBUG, "[session:%p] The joint cap is only had audio_fec and audio_red, audio media negotiated failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	return 1;
}

static unsigned int judge_if_support_trickle_ice(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	if (!session || !session_media) {
		return 0;
	}

	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return 0;
	}

	if (session_media->trickle_ice) {
		return 1;
	}

	if (session->is_webrtc_call || session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
		return 1;
	}

	return 0;
}

static void swap_format(struct ast_format **p1, struct ast_format **p2)
{
    struct ast_format *tmp = *p1;
    *p1 = *p2;
    *p2 = tmp;

    return;
}

static enum ast_stream_video_resolution change_level_to_video_resolution(unsigned int level)
{
	enum ast_stream_video_resolution video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	/* We use profile-level-id last two bit to determine video resolution,
	 * 428028(1080P) -> 28(hex) = 40(Dec); 428016(360P) -> 16(hex) = 22(Dec)
	 * 42801F(720P) -> 1F(hex) = 31(Dec)
	 */
	if (level >= 40) {
		video_resolution = AST_STREAM_VIDEO_RESOLUTION_1080P;
	} else if (level == 31) {
		video_resolution = AST_STREAM_VIDEO_RESOLUTION_720P;
	} else {
		video_resolution = AST_STREAM_VIDEO_RESOLUTION_360P;
	}

	return video_resolution;
}

static void change_rtx_apt_value(struct ast_sip_session *session, struct ast_format **format, int first_video_pt)
{
	char apt_str[8] = {0};
	struct ast_format *rtx_format = NULL;

	if (!session || !*format) {
		return;
	}

	if (!first_video_pt) {
		return;
	}

	if (strcasecmp(ast_format_get_name(*format), "rtx")) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] It has RTX codec, set apt value:%d, endpoint:%s\n",
		session, first_video_pt, ast_sorcery_object_get_id(session->endpoint));

	snprintf(apt_str, sizeof(apt_str), "%d", first_video_pt);
	rtx_format = ast_format_attribute_set(*format, "apt", apt_str);
	if (rtx_format) {
		swap_format(&rtx_format, format);
		ao2_ref(rtx_format, -1);
	}

	return;
}

static unsigned int change_resolution_to_level_id(enum video_resolution resolution)
{
	unsigned int res = 0;

	switch (resolution) {
		case VIDEO_RESOLUTION_360P:
			res = 22;
			break;
		case VIDEO_RESOLUTION_720P:
			res = 31;
			break;
		case VIDEO_RESOLUTION_1080P:
			res = 40;
			break;
		default:
			res = 22;
			break;
	}

	return res;
}

static void change_imageattr_for_stream(struct ast_sip_session *session,
	struct ast_stream *stream, enum video_resolution resolution)
{
	const char *orig_imageattr = NULL;
	char *imageattr = NULL;

	if (!session || !stream) {
		return;
	}

	orig_imageattr = ast_stream_get_metadata(stream, "IMAGEATTR");
	if (ast_strlen_zero(orig_imageattr)) {
		return;
	}

	imageattr = ast_sip_session_get_imageattr_according_type(resolution);
	if (imageattr) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_imageattr_for_stream, imageattr:%s, endpoint:%s\n",
			session, imageattr, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_metadata(stream, "IMAGEATTR", imageattr);
	}

	return;
}

static void change_p2p_level_id_according_configure(struct ast_sip_session *session,
	struct ast_stream *stream, struct ast_format **format)
{
	unsigned int res = 0;
	char level_id_str[8] = {0};
	struct ast_format *video_format = NULL;
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;
	const unsigned int *h264_level = NULL;

	if (!session || !*format) {
		return;
	}

	if (session->is_sfu) {
		return;
	}

	if (strcasecmp(ast_format_get_name(*format), "H264")) {
		return;
	}

	if (!ast_sip_session_get_enable_p2p_video_control()) {
		return;
	}

	resolution = ast_sip_session_get_p2p_video_resolution();
	res = change_resolution_to_level_id(resolution);

	h264_level = ast_format_attribute_get(*format, "level");
	if (h264_level && *h264_level && (*h264_level < res)) {
		ast_log(LOG_DEBUG, "[session:%p] H264 level id:%d is less than configured:%d, don't need change, endpoint:%s\n",
			session, *h264_level, res, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Use configured H264 codec level id, set value:%d, endpoint:%s\n",
		session, res, ast_sorcery_object_get_id(session->endpoint));

	change_imageattr_for_stream(session, stream, resolution);

	snprintf(level_id_str, sizeof(level_id_str), "%d", res);
	video_format = ast_format_attribute_set(*format, "level-id", level_id_str);
	if (video_format) {
		swap_format(&video_format, format);
		ao2_ref(video_format, -1);
	}

	return;
}

/* Add by bfdu for bug 156722, 2020.4.24 */
static int change_rtx_pt(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_format *format, char *name, int rtp_code)
{
	int rtx_pt = 0;

	if (!session || !format || !name || !session_media || !session_media->rtp) {
		return rtp_code;
	}

	if (strcasecmp(ast_format_get_name(format), "rtx")) {
		return rtp_code;
	}

	ast_log(LOG_DEBUG, "[session:%p] It has RTX codec, first video name:%s, endpoint:%s\n",
		session, name, ast_sorcery_object_get_id(session->endpoint));

	if (!strcasecmp(name, "H264")) {
		rtx_pt = session_media->video_rtx_pt.h264_rtx_pt ? session_media->video_rtx_pt.h264_rtx_pt : rtp_code;
	} else if (!strcasecmp(name, "H265")) {
		rtx_pt = session_media->video_rtx_pt.h265_rtx_pt ? session_media->video_rtx_pt.h265_rtx_pt : rtp_code;
	} else if (!strcasecmp(name, "VP8")) {
		rtx_pt = session_media->video_rtx_pt.vp8_rtx_pt ? session_media->video_rtx_pt.vp8_rtx_pt : rtp_code;
	} else if (!strcasecmp(name, "VP9")) {
		rtx_pt = session_media->video_rtx_pt.vp9_rtx_pt ? session_media->video_rtx_pt.vp9_rtx_pt : rtp_code;
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Can't support video codec:%s for RTX, endpoint:%s\n",
			session, name, ast_sorcery_object_get_id(session->endpoint));
		rtx_pt = rtp_code;
	}

	if (session->main_video_first_multi_rtx_pt) {
		ast_log(LOG_DEBUG, "[session:%p] Main video RTX has multi codec, just use the first RTX codec PT as main video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		rtx_pt = session->main_video_first_multi_rtx_pt;
	}

	ast_rtp_instance_set_rtx_pt(session_media->rtp, rtx_pt);

	if (rtx_pt != rtp_code) {
		ast_log(LOG_DEBUG, "[session:%p] It has RTX codec, first video:%s RTX payload:%d is not the same store RTX payload:%d, endpoint:%s\n",
			session, name, rtx_pt, rtp_code, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_codecs_payload_set_rx(ast_rtp_instance_get_codecs(session_media->rtp), rtx_pt, format);
	}

	return rtx_pt;
}

static void set_video_according_rtx_payload(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *stream, struct ast_format *format_parsed, int rtx_payload, struct ast_stream *asterisk_stream)
{
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	pj_str_t apt_str;
	char name[64] = { 0 };
	const unsigned int *apt = NULL;
	char apt_s[8] = { 0 };

	apt = ast_format_attribute_get(format_parsed, "apt");
	if (!apt || (*apt == 0)) {
		return;
	}

	snprintf(apt_s, sizeof(apt_s) - 1, "%d", *apt);
	//pj_strdup2(session->inv_session->pool_prov, &apt_str, apt_s);
	apt_str = pj_str(apt_s);

	ast_log(LOG_DEBUG, "[session:%p] Get rtx_payload:%d, apt:%d in fmtp for endpoint:%s\n",
		session, rtx_payload, *apt, ast_sorcery_object_get_id(session->endpoint));

	if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &apt_str))) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get rtpmap for PT:%d, pj_strtoul:%ld for endpoint:%s\n",
			session, *apt, pj_strtoul(&apt_str), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
		ast_log(LOG_DEBUG, "[session:%p] Can't prase to rtpmap for PT:%d for endpoint:%s\n",
			session, *apt, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));

	if (!strcasecmp(name, "H264") && !session_media->video_rtx_pt.h264_rtx_pt) {
		session_media->video_rtx_pt.h264_rtx_pt = rtx_payload;
		ast_log(LOG_DEBUG, "[session:%p] Get apt:%d according H264 video codec for endpoint:%s\n",
			session, *apt, ast_sorcery_object_get_id(session->endpoint));
	} else if (!strcasecmp(name, "H265") && !session_media->video_rtx_pt.h265_rtx_pt) {
		session_media->video_rtx_pt.h265_rtx_pt = rtx_payload;
	} else if (!strcasecmp(name, "VP8") && !session_media->video_rtx_pt.vp8_rtx_pt) {
		session_media->video_rtx_pt.vp8_rtx_pt = rtx_payload;
	} else if (!strcasecmp(name, "VP9") && !session_media->video_rtx_pt.vp9_rtx_pt) {
		session_media->video_rtx_pt.vp9_rtx_pt = rtx_payload;
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Can't support video codec:%s for RTX,  endpoint:%s\n",
			session, name, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_stream_set_video_rtx_pt(asterisk_stream, name, rtx_payload);

	return;
}

static void init_video_codec_rtx_pt(struct ast_sip_session_media *session_media)
{
	if (!session_media) {
		return;
	}

	session_media->video_rtx_pt.h264_rtx_pt = 0;
	session_media->video_rtx_pt.h265_rtx_pt = 0;
	session_media->video_rtx_pt.vp8_rtx_pt = 0;
	session_media->video_rtx_pt.vp9_rtx_pt = 0;

	return;
}

static void process_rtx_codec(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	unsigned int i = 0;
	unsigned int num = 0;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	struct ast_format *format_parsed = NULL;
	pjmedia_sdp_fmtp fmtp;
	char name[64] = { 0 };
	char fmt_param[256] = { 0 };

	if (!session || !session_media || !stream) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	init_video_codec_rtx_pt(session_media);
	ast_stream_init_video_rtx_pt(asterisk_stream);

	/* Iterate through provided formats */
	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
		if (strcasecmp(name, "rtx")) {
			continue;
		}

		/* Look for an optional associated fmtp attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "fmtp", &rtpmap->pt))) {
			continue;
		}

		if ((pjmedia_sdp_attr_get_fmtp(attr, &fmtp)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(fmt_param, &fmtp.fmt, sizeof(fmt_param));
		if (sscanf(fmt_param, "%30d", &num) != 1) {
			continue;
		}

		ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));

		format_parsed = ast_format_parse_sdp_fmtp(ast_format_rtx, fmt_param);
		if (format_parsed) {
			set_video_according_rtx_payload(session, session_media, stream, format_parsed, pj_strtoul(&stream->desc.fmt[i]), asterisk_stream);
			ao2_ref(format_parsed, -1);
		}
	}

	return;
}


static void process_remote_stream_port(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, unsigned int remote_port)
{
	if (!session || !session->inv_session || !session_media) {
		return;
	}

	if ((pjmedia_sdp_neg_get_state(session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER) &&
		(session->inv_session->state == PJSIP_INV_STATE_CONFIRMED) &&
		((session_media->remote_port != remote_port) && (session_media->remote_port == 0 || remote_port == 0))) {
		ast_log(LOG_DEBUG, "[session:%p] It open/close media, set session media changed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session_media->changed = 1;
	}

	session_media->remote_port = remote_port;

	return;
}

static void process_gs_fec_codec(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	char *name, unsigned int *has_gs_fec, int fec_payload)
{
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (!session || !session_media || !name || !has_gs_fec) {
		return;
	}

	if (strcasecmp(name, "GS-FEC")) {
		return;
	}

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "Failed to allocate %s capabilities\n",
			ast_codec_media_type2str(session_media->type));
		return;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);

	if (!ast_format_cap_has_audio_fec(caps) && (session_media->type == AST_MEDIA_TYPE_AUDIO)) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint codec has no audio gs_fec, don't need process, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!ast_format_cap_has_video_fec(caps) &&
		(session_media->type == AST_MEDIA_TYPE_VIDEO || session_media->type == AST_MEDIA_TYPE_APPLICATION)) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint codec has no video gs_fec, don't need process, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	*has_gs_fec = 1;
	ast_rtp_instance_set_gsfec_pt(session_media->rtp, fec_payload);

	ast_log(LOG_DEBUG, "[session:%p] ast_rtp_instance_set_gsfec_pt fec_payload:%d, endpoint:%s\n",
		session, fec_payload, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void process_remote_ssrc_attr(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_stream *asterisk_stream, pjmedia_sdp_media *stream)
{
	struct ast_format_cap *caps = NULL;

	if (!session || !session_media || !asterisk_stream || !stream || !session_media->rtp) {
		return;
	}

	caps = ast_stream_get_formats(asterisk_stream);
	if (!caps) {
		ast_log(LOG_DEBUG, "[session:%p] Stream:%p has no formats, endpoint:%s\n",
			session, stream, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_format_cap_has_audio_fec(caps) ||
		ast_format_cap_has_video_fec(caps)) {
		process_fec_ssrc_group_attributes(session, session_media, stream, asterisk_stream);
	}

	if (ast_format_cap_has_rtx(caps)) {
		process_rtx_ssrc_group_attributes(session, session_media, stream, asterisk_stream);
	} else {
		process_ssrc_attributes(session, session_media, stream, asterisk_stream);
	}

	return;
}

static void process_h264_format_no_profile_level_id(struct ast_sip_session *session, struct ast_format **format)
{
	const unsigned int *h264_level = NULL;
	char *profile_level_id_str = NULL;
	struct ast_format *new_format = NULL;

	if (strcasecmp(ast_format_get_name(*format), "H264")) {
		return;
	}

	h264_level = ast_format_attribute_get(*format, "level");
	if (h264_level) {
		return;
	}

	/* If we has no profile level id, use 360P as default */
	profile_level_id_str = ast_sip_session_get_video_resolution_according_type(session, VIDEO_RESOLUTION_1080P);
	if (profile_level_id_str) {
		ast_log(LOG_DEBUG, "[session:%p] create SDP has no profile-level-id, set profile_level_id_str:%s, endpoint:%s\n",
			session, profile_level_id_str, ast_sorcery_object_get_id(session->endpoint));
		new_format = ast_format_attribute_set(*format, "profile-level-id", profile_level_id_str);
		if (new_format) {
			swap_format(&new_format, format);
			ao2_ref(new_format, -1);
		}
	}

	return;
}

static void set_stream_state_for_page(struct ast_sip_session *session, struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *remote_stream)
{
	int callee=0;
	enum ast_stream_media_direction local_media_direction = INACTIVE;
	enum ast_stream_state stream_state = AST_STREAM_STATE_INACTIVE;
	if (!session || !session->channel || !asterisk_stream || !remote_stream) {
		return;
	}

	if (ast_channel_page_mode_get(session->channel) == AST_PAGE_MODE_NONE) {
		return;
	}

	if(ast_channel_page_mode_get(session->channel) == AST_PAGE_MODE_DUPLEX ) {
		if (ast_stream_get_type(asterisk_stream) != AST_MEDIA_TYPE_AUDIO) {
			ast_stream_set_state(asterisk_stream, stream_state);
			if(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ) {
				ast_stream_set_local_media_direction(asterisk_stream, local_media_direction);
			}
		}
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] It's  paging announce mode, set stream direction for paging, channel:%s\n",
		session, ast_channel_name(session->channel));

	/*Start added  by zyxia for Bug 234952 ,2022.9.5*/
	const char *is_callee =pbx_builtin_getvar_helper(session->channel, "IS_CALLEE");
	if (!ast_strlen_zero(is_callee) && !strcmp(is_callee, "yes")) {
		callee = 1;
	}

	if (!callee) {//caller need recvonly
		if (pjmedia_sdp_media_find_attr2(remote_stream, "sendrecv", NULL)) {
			/*Start added for Bug 242583,gxv3370 send reinvite.zyxia*/
			if (ast_stream_get_type(asterisk_stream) == AST_MEDIA_TYPE_AUDIO){
				local_media_direction = SENDRECV;
				stream_state = AST_STREAM_STATE_SENDRECV;
			}
			else if (ast_stream_get_type(asterisk_stream) == AST_MEDIA_TYPE_VIDEO){
				local_media_direction = RECVONLY;
				stream_state = AST_STREAM_STATE_RECVONLY;
			}
			/*End add for Bug 242583,gxv3370 send reinvite.zyxia*/
		} else if (pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL)) {
			local_media_direction = RECVONLY;
			stream_state = AST_STREAM_STATE_RECVONLY;
		} else if (pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
			local_media_direction = INACTIVE;
			stream_state = AST_STREAM_STATE_INACTIVE;
		} else if (pjmedia_sdp_media_find_attr2(remote_stream, "recvonly", NULL)) {
			local_media_direction = SENDONLY;
			stream_state = AST_STREAM_STATE_SENDONLY;
		} else {
			ast_log(LOG_WARNING, "Can't find media direction attr, use inactive as default\n");
		}
	}
	/*start added by zyxia for Bug 237118,page callee hold unhold */
	else{
		if (pjmedia_sdp_media_find_attr2(remote_stream, "sendrecv", NULL)) {
			if (ast_stream_get_type(asterisk_stream) == AST_MEDIA_TYPE_AUDIO){
				local_media_direction = SENDRECV;
				stream_state = AST_STREAM_STATE_SENDRECV;
			}
			else if (ast_stream_get_type(asterisk_stream) == AST_MEDIA_TYPE_VIDEO){
				local_media_direction = SENDONLY;
				stream_state = AST_STREAM_STATE_SENDONLY;
			}
		} else if (pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL)) {
			local_media_direction = RECVONLY;
			stream_state = AST_STREAM_STATE_RECVONLY;
		} else if (pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
			local_media_direction = INACTIVE;
			stream_state = AST_STREAM_STATE_INACTIVE;
		} else if (pjmedia_sdp_media_find_attr2(remote_stream, "recvonly", NULL)) {
			local_media_direction = SENDONLY;
			stream_state = AST_STREAM_STATE_SENDONLY;
		} else {
			ast_log(LOG_WARNING, "Can't find media direction attr, use sendrecv as default\n");
		}
	}
	/*end added by zyxia for Bug 237118,page callee hold unhold */


	/*if not support video ,disable video sdp,added by zyxia for Bug 238192 */
	if(ast_channel_page_mode_get(session->channel) == AST_PAGE_MODE_ANNOUNCE
		&& ast_stream_get_type(asterisk_stream) == AST_MEDIA_TYPE_VIDEO
		&& ast_channel_get_is_page_video_support(session->channel) == 0) {
		local_media_direction = INACTIVE;
		stream_state = AST_STREAM_STATE_INACTIVE;
	}

	ast_stream_set_state(asterisk_stream, stream_state);
	ast_stream_set_local_media_direction(asterisk_stream, local_media_direction);

	/*End added  by zyxia for Bug 234952 ,2022.9.5*/

	return;
}

static void set_slides_video_media_change_according_direction(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	enum ast_stream_media_direction pre_local_dir,
	enum ast_stream_media_direction local_dir)
{
	if (pre_local_dir != UNKNOW && local_dir != pre_local_dir) {
		ast_log(LOG_DEBUG, "[session:%p] Slides video  media direction changed pre_local_dir:%d, now local_dir:%d, endpoint:%s\n",
			session, pre_local_dir, local_dir, ast_sorcery_object_get_id(session->endpoint));
		session_media->changed = 1;
	}

	if (session->channel && !ast_channel_get_send_slides_video(session->channel) &&
		session->inv_session->state == PJSIP_INV_STATE_CONFIRMED &&
		(pre_local_dir == UNKNOW || pre_local_dir == INACTIVE) &&
		local_dir == SENDRECV) {
		ast_log(LOG_DEBUG, "[session:%p] Set open_slides true, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session_media->open_slides = 1;
	}

	return;
}

static void set_main_video_media_change_according_direction(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	enum ast_stream_media_direction pre_local_dir,
	enum ast_stream_media_direction local_dir,
	struct ast_stream *asterisk_stream)
{
	ast_log(LOG_DEBUG, "[session:%p] set_main_video_media_change_according_direction pre_local_dir:%d, now local_dir:%d, endpoint:%s\n",
			session, pre_local_dir, local_dir, ast_sorcery_object_get_id(session->endpoint));

	if (local_dir == pre_local_dir) {
		return;
	}

	switch (pre_local_dir) {
		case SENDRECV:
		case RECVONLY:
			if (local_dir == SENDONLY || local_dir == INACTIVE) {
				ast_log(LOG_DEBUG, "[session:%p] It's close camera for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				session_media->changed = 1;
				if (session->is_sfu) {
					session_media->close_camera = 1;
				}
				ast_stream_set_video_resolution(asterisk_stream, AST_STREAM_VIDEO_RESOLUTION_NONE);
			}
			break;
		case SENDONLY:
		case INACTIVE:
			if ((local_dir == SENDRECV || local_dir == RECVONLY) &&
				ast_stream_get_video_resolution(asterisk_stream) != AST_STREAM_VIDEO_RESOLUTION_NONE) {
				ast_log(LOG_DEBUG, "[session:%p] It's open camera for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				session_media->changed = 1;
				if (session->is_sfu) {
					session_media->open_camera = 1;
				}
				ast_sip_session_set_main_video_stream_uptime(session);
			}
			break;
		default:
			break;
	}

	return;
}

static void set_media_change_according_direction(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	enum ast_stream_media_direction pre_local_dir, struct ast_stream *asterisk_stream)
{
	enum ast_stream_media_direction local_dir = UNKNOW;

	if (!session || !session_media || !asterisk_stream) {
		return;
	}

	local_dir = ast_stream_get_local_media_direction(asterisk_stream);

	if (ast_stream_get_sub_type(asterisk_stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
		set_slides_video_media_change_according_direction(session, session_media, pre_local_dir, local_dir);
	} else if (ast_stream_judge_if_main_video(asterisk_stream)) {
		set_main_video_media_change_according_direction(session, session_media, pre_local_dir, local_dir, asterisk_stream);
	}

	return;
}

static void set_stream_local_media_direction_according_remote(struct ast_stream *asterisk_stream,
	const struct pjmedia_sdp_media *remote_stream)
{
	enum ast_stream_media_direction local_media_direction = SENDRECV;

	if (!asterisk_stream || !remote_stream) {
		ast_log(LOG_WARNING, "Set local media direction, param is null\n");
		return;
	}

	if (pjmedia_sdp_media_find_attr2(remote_stream, "sendrecv", NULL)) {
		local_media_direction = SENDRECV;
	} else if (pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL)) {
		local_media_direction = RECVONLY;
	} else if (pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		local_media_direction = INACTIVE;
	} else if (pjmedia_sdp_media_find_attr2(remote_stream, "recvonly", NULL)) {
		local_media_direction = SENDONLY;
	} else {
		ast_log(LOG_WARNING, "Can't find media direction attr, use sendrecv as default\n");
	}

	ast_stream_set_local_media_direction(asterisk_stream, local_media_direction);

	return;
}

static void set_stream_local_media_direction_according_local(struct ast_stream *asterisk_stream,
	const struct pjmedia_sdp_media *local_stream)
{
	enum ast_stream_media_direction local_media_direction = SENDRECV;

	if (!asterisk_stream || !local_stream) {
		ast_log(LOG_WARNING, "Set local media direction, param is null\n");
		return;
	}

	if (pjmedia_sdp_media_find_attr2(local_stream, "sendrecv", NULL)) {
		local_media_direction = SENDRECV;
	} else if (pjmedia_sdp_media_find_attr2(local_stream, "sendonly", NULL)) {
		local_media_direction = SENDONLY;
	} else if (pjmedia_sdp_media_find_attr2(local_stream, "inactive", NULL)) {
		local_media_direction = INACTIVE;
	} else if (pjmedia_sdp_media_find_attr2(local_stream, "recvonly", NULL)) {
		local_media_direction = RECVONLY;
	} else {
		ast_log(LOG_WARNING, "Can't find media direction attr, use sendrecv as default\n");
	}

	ast_stream_set_local_media_direction(asterisk_stream, local_media_direction);

	return;
}

static void set_stream_remote_media_direction(struct ast_stream *asterisk_stream,
	const struct pjmedia_sdp_media *remote_stream)
{
	enum ast_stream_media_direction remote_media_direction = SENDRECV;

	if (!asterisk_stream || !remote_stream) {
		ast_log(LOG_WARNING, "Set media direction, param is null\n");
		return;
	}

	if (pjmedia_sdp_media_find_attr2(remote_stream, "sendrecv", NULL)) {
		remote_media_direction = SENDRECV;
	} else if (pjmedia_sdp_media_find_attr2(remote_stream, "sendonly", NULL)) {
		remote_media_direction = SENDONLY;
	} else if (pjmedia_sdp_media_find_attr2(remote_stream, "inactive", NULL)) {
		remote_media_direction = INACTIVE;
	} else if (pjmedia_sdp_media_find_attr2(remote_stream, "recvonly", NULL)) {
		remote_media_direction = RECVONLY;
	} else {
		ast_log(LOG_WARNING, "Can't find media direction attr, use sendrecv as default\n");
	}

	ast_stream_set_remote_media_direction(asterisk_stream, remote_media_direction);

	return;
}

static void set_local_media_direction_for_page(struct ast_sip_session *session, struct ast_stream *stream)
{
	enum ast_page_mode page_mode = AST_PAGE_MODE_NONE;

	if (!session || !session->channel) {
		return;
	}

	if (!ast_channel_get_is_page_video_support(session->channel) && ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		return;
	}

	/*start added by zyxia for video sigle announment page,2022.8.9*/
	/*support video,but is audio stream*/
	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_VIDEO  && session->is_video_page_call){
		ast_stream_set_local_media_direction(stream, SENDONLY);
		ast_log(LOG_DEBUG, "[session:%p] set_local_media_direction_for_page video sigle announment for endpoint:%s,%s\n",
			session, ast_sorcery_object_get_id(session->endpoint),ast_channel_name(session->channel));
		return;
	}

	const char *is_announcement_page = pbx_builtin_getvar_helper(session->channel, "IS_ANNOUNCEMENT_PAGING");
	 if (!ast_strlen_zero(is_announcement_page) && !strcmp(is_announcement_page,"1")  && ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO ){
		ast_stream_set_local_media_direction(stream, SENDONLY);
		ast_log(LOG_DEBUG, "[session:%p] set_local_media_direction_for_page video  announment paging for endpoint:%s,%s\n",
			session, ast_sorcery_object_get_id(session->endpoint),ast_channel_name(session->channel));
		return;
	}

	/*start added by zyxia for 235187 ,2022.8.30*/
	/*support video,but the caller is audio call, dont set,use the caller direction*/
	if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO && session->is_video_page_call && ast_stream_get_local_media_direction(stream) == INACTIVE){
		ast_log(LOG_DEBUG, "[session:%p] dont need set the video direction sigle announment for endpoint:%s,%s\n",
			session, ast_sorcery_object_get_id(session->endpoint),ast_channel_name(session->channel));
		return;
	}
	/*End added by zyxia for 235187 ,2022.8.30*/

	/*End added by zyxia for video sigle announment page,2022.8.9*/

	page_mode = ast_channel_page_mode_get(session->channel);

	if (page_mode == AST_PAGE_MODE_NONE) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] set_local_media_direction_for_page page_mode:%d, for endpoint:%s\n",
		session, page_mode, ast_sorcery_object_get_id(session->endpoint));

	switch (page_mode) {
		case AST_PAGE_MODE_DUPLEX:
			ast_stream_set_local_media_direction(stream, SENDRECV);
			break;
		case AST_PAGE_MODE_ANNOUNCE:
			ast_stream_set_local_media_direction(stream, SENDONLY);
			break;
		default:
			break;
	}



	return;
}

static void add_stream_local_media_direction(struct ast_stream *asterisk_stream, pjmedia_sdp_attr *attr)
{
	if (!asterisk_stream || !attr) {
		ast_log(LOG_WARNING, "Add media direction, param is null\n");
		return;
	}

	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	static const pj_str_t STR_SENDONLY = { "sendonly", 8 };
	static const pj_str_t STR_INACTIVE = { "inactive", 8 };
	static const pj_str_t STR_RECVONLY = { "recvonly", 8 };

	enum ast_stream_media_direction media_direction = SENDRECV;

	media_direction = ast_stream_get_local_media_direction(asterisk_stream);
	
	switch (media_direction) {
		case SENDRECV:
			attr->name = STR_SENDRECV;
			break;
		case SENDONLY:
			attr->name = STR_SENDONLY;
			break;
		case INACTIVE:
			attr->name = STR_INACTIVE;
			break;
		case RECVONLY:
			attr->name = STR_RECVONLY;
			break;
		default:
			ast_log(LOG_WARNING, "Can't process media direction:%d\n", media_direction);
			break;
	}

	return;
}

static void set_session_negotiated_dtmf(struct ast_sip_session *session, struct ast_sip_session_media *session_media, int tel_event)
{
	if (!session || !session_media) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	switch (session->dtmf) {
		case AST_SIP_DTMF_NONE:
			session->negotiated_dtmf = AST_SIP_DTMF_NONE;
			break;
		case AST_SIP_DTMF_RFC_4733:
			if (tel_event) {
				session->negotiated_dtmf = AST_SIP_DTMF_RFC_4733;
			}
			break;
		case AST_SIP_DTMF_INBAND:
			session->negotiated_dtmf = AST_SIP_DTMF_INBAND;
			break;
		case AST_SIP_DTMF_INFO:
			session->negotiated_dtmf = AST_SIP_DTMF_INFO;
			break;
		case AST_SIP_DTMF_AUTO:
			if (tel_event) {
				session->negotiated_dtmf = AST_SIP_DTMF_RFC_4733;
			} else {
				session->negotiated_dtmf = AST_SIP_DTMF_INBAND;
			}
			break;
		case AST_SIP_DTMF_AUTO_INFO:
		case AST_SIP_DTMF_BOTH_RFC4733_AND_INFO:
			if (tel_event) {
				session->negotiated_dtmf = AST_SIP_DTMF_RFC_4733;
			} else {
				session->negotiated_dtmf = AST_SIP_DTMF_INFO;
			}
			break;
		default:
			break;
	}

	return;
}

static void determine_gs_fec_version(struct ast_sip_session_media *session_media, struct ast_stream *stream, char *gs_fec_version)
{
	unsigned int version = 0;
	unsigned int max_version = 0;
	char *version_str = NULL;
	char max_version_str[MAX_GS_FEC_VERSION_LEN] = {0};

	if (!session_media || !stream || ast_strlen_zero(gs_fec_version)) {
		return;
	}

	ast_log(LOG_DEBUG, "determine_gs_fec_version init version:%s\n", gs_fec_version);

	if (strchr(gs_fec_version, ',') == NULL) {
		return;
	}

	while ((version_str = strsep(&gs_fec_version, ","))) {
		version_str = ast_strip(version_str);
		if (sscanf(version_str, "%d", &version) == 1) {
			if (max_version < version) {
				max_version = version;
			}
		}
	}

	snprintf(max_version_str, sizeof(max_version_str) - 1, "%d", max_version);
	strncpy(session_media->gs_fec_version, max_version_str, MAX_GS_FEC_VERSION_LEN - 1);
	ast_stream_set_gs_fec_version(stream, session_media->gs_fec_version);

	ast_log(LOG_DEBUG, "determine_gs_fec_version final version:%s\n", session_media->gs_fec_version);

	return;
}

static int check_set_stream_video_resolution_params_valid(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream)
{
	enum ast_stream_media_direction remote_dir = 0;

	if (!session || !session_media || !asterisk_stream) {
		return -1;
	}

	if (!session->channel || !session->inv_session || !session->inv_session->neg ) {
		return -1;
	}

	if (!session->is_sfu) {
		return -1;
	}

	if (pjmedia_sdp_neg_get_state(session->inv_session->neg) != PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER) {
		return -1;
	}

	if (session->inv_session->state != PJSIP_INV_STATE_CONFIRMED) {
		return -1;
	}

	if (ast_strlen_zero(session->conf_room)) {
		return -1;
	}

	if (!ast_stream_judge_if_main_video(asterisk_stream)) {
		return -1;
	}

	remote_dir = ast_stream_get_remote_media_direction(asterisk_stream);
	if (remote_dir == INACTIVE || remote_dir == RECVONLY) {
		ast_log(LOG_DEBUG, "[session:%p] Remote stream media direction is %d, don't need get video resolution, endpoint:%s\n",
			session, remote_dir, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	return 0;
}

static int check_media_transport(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *stream)
{
	unsigned int use_received_transport = 0;
	unsigned int optimistic = 0;
	unsigned int use_received_zrtp = 0;

	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;

	if (!session || !session_media || !stream) {
		return -1;
	}

	if (session->is_webrtc_call) {
		use_received_transport = session->endpoint->media.rtp.use_received_transport;
	} else {
		use_received_transport = session->endpoint->media.rtp.configure_use_received_transport;
	}

	if (!session->is_webrtc_call) {
		use_received_zrtp = session->endpoint->media.rtp.configure_use_received_zrtp;
	}

	if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_FEC) {
		ast_log(LOG_DEBUG, "[session:%p] It's application FEC stream, use received encryption type for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		use_received_transport = 1;
		session_media->encryption = ast_sip_session_get_media_encryption_type(stream->desc.transport, stream, &optimistic);
	}


	encryption = ast_sip_session_check_endpoint_media_transport(session->endpoint, stream, session->is_webrtc_call);
	if(encryption == AST_SIP_MEDIA_ENCRYPT_ZRTP  || encryption == AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP){
		session_media->zrtp = 1;
		if(use_received_zrtp)//not force
			return 0;
	}

	if (use_received_transport) {
		return 0;
	}

	if (!check_if_direct_media_enabled(session) && (encryption == AST_SIP_MEDIA_TRANSPORT_INVALID)) {
		ast_log(LOG_DEBUG, "[session:%p] check transport is invalid for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	return 0;
}

static void check_remote_ice_support(struct ast_sip_session *session, const struct pjmedia_sdp_media *remote_stream)
{
	const pjmedia_sdp_attr *attr = NULL;
	unsigned int attr_i = 0;

	if (!session || !remote_stream) {
		return;
	}

	if (session->support_ice) {
		return;
	}

	if (!session->endpoint->media.rtp.ice_support) {
		return;
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		attr = remote_stream->attr[attr_i];
		if (!pj_strcmp2(&attr->name, "candidate")) {
			session->support_ice = 1;
			break;
		}
	}

	if (!session->support_ice && (session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK)) {
		ast_log(LOG_DEBUG, "[session:%p] Remote can't support ICE, clear trunk ICE type for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->terminal_type = AST_TERMINAL_TYPE_SIP;
	}

	return;
}

static void set_session_media_dtls_flag(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_media *stream)
{
	unsigned int optimistic = 0;
	enum ast_sip_session_media_encryption encryption = AST_SIP_MEDIA_ENCRYPT_NONE;

	if (!session || !session_media || !stream) {
		return;
	}

	encryption = ast_sip_session_get_media_encryption_type(stream->desc.transport, stream, &optimistic);
	if (encryption == AST_SIP_MEDIA_ENCRYPT_DTLS) {
		session_media->dtls = 1;
	}

	return;
}

static void set_avs_ice_flag(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *remote_stream)
{
	const pjmedia_sdp_attr *attr = NULL;
	unsigned int attr_i = 0;
	unsigned int remote_support_ice = 0;

	if (!session || !session_media || !session_media->rtp || !remote_stream) {
		return;
	}

	if (!session->endpoint->media.rtp.ice_support) {
		return;
	}

	if (session->is_webrtc_call ||
		session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
		return;
	}

	if (judge_if_support_trickle_ice(session, session_media)) {
		ast_log(LOG_DEBUG, "[session:%p] Remote can support trickle ICE, don't need clear avs ice flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		attr = remote_stream->attr[attr_i];
		if (!pj_strcmp2(&attr->name, "candidate")) {
			remote_support_ice = 1;
			break;
		}
	}

	if (!remote_support_ice) {
		ast_log(LOG_DEBUG, "[session:%p] Remote can't support ICE, clear avs ice flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_ice_flag(session_media->rtp, 0);
	}

	return;
}

static void reset_media_encryption(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	unsigned int use_received_transport = 0;

	if (session->is_webrtc_call) {
		use_received_transport = session->endpoint->media.rtp.use_received_transport;
	} else {
		use_received_transport = session->endpoint->media.rtp.configure_use_received_transport;
	}

	if (use_received_transport) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s use received transport, don't reset media encryption for media:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint), ast_codec_media_sub_type2str(session_media->sub_type));
		return;
	}

	if (session->is_webrtc_call) {
		session_media->encryption = session->endpoint->media.rtp.encryption;
	} else {
		session_media->encryption = session->endpoint->media.rtp.configure_encryption;
	}

	return;
}

static void set_stream_params_according_video_resolution_for_h264_codec(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, enum ast_stream_video_resolution video_resolution)
{
	char *profile_level_id_str = NULL;

	profile_level_id_str = ast_sip_session_get_video_resolution_according_type(session, video_resolution);
	if (!profile_level_id_str) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] set_stream_params_according_video_resolution profile_level_id_str:%s, endpoint:%s\n",
		session, profile_level_id_str, ast_sorcery_object_get_id(session->endpoint));

	ast_format_attribute_set_directly(format, "profile-level-id", profile_level_id_str);

	ast_stream_set_video_resolution(asterisk_stream, video_resolution);
	ast_sip_session_set_imageattr_for_stream(session, asterisk_stream, video_resolution);

	return;
}

static void set_stream_params_according_video_resolution(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, enum ast_stream_video_resolution video_resolution)
{
	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			set_stream_params_according_video_resolution_for_h264_codec(session, format, asterisk_stream, video_resolution);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			break;
		default:
			break;
	}

	return;
}

static void set_h264_resolution_for_open_camera(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, struct ast_sip_session_media *session_media,
	pjmedia_sdp_media *media, enum ast_stream_video_resolution resolution)
{
	enum ast_stream_video_resolution obtain_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	obtain_video_resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, resolution);

	ast_log(LOG_DEBUG, "[session:%p] Open camera get resolution:%d for endpoint:%s\n",
		session, obtain_video_resolution, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_set_conf_video_codec(session->inv_session->pool_prov, session, media, obtain_video_resolution);
	if (ast_sip_session_modify_video_codec_order_for_sfu(session->inv_session->pool_prov, session, media, obtain_video_resolution)) {
		ast_log(LOG_DEBUG, "[session:%p] There is something wrong for chose video codec, don't modify other param for sfu, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if ((obtain_video_resolution <= AST_STREAM_VIDEO_RESOLUTION_NONE) || (obtain_video_resolution > AST_STREAM_VIDEO_RESOLUTION_1080P)) {
		if (!ast_sip_session_support_multiple_video(session) || session->terminal_type == AST_TERMINAL_TYPE_GVC) {
			if (ast_sip_session_get_only_support_audio()) {
				ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_REMOVED);
			} else {
				ast_stream_set_local_media_direction(asterisk_stream, SENDONLY);
			}
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Can't get video resource, and it's not SIP terminal, need to set stream state to removed, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && ast_sip_session_media_use_bundle(session)) {
				ast_stream_set_local_media_direction(asterisk_stream, INACTIVE);
			} else {
				ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_REMOVED);
				session_media->bundle_group = -1;
				session_media->bundled = 0;
			}
		}
		if (obtain_video_resolution == -2) {
			ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_CPU_LIMIT);
		} else if (obtain_video_resolution == -3) {
			ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_ROOM);
		} else {
			ast_stream_set_video_decline_reason(asterisk_stream, AST_VIDEO_DECLINE_REASON_NO_VIDEO_RESOURCE);
		}
	} else {
		session_media->changed = 1;
		set_stream_params_according_video_resolution(session, format, asterisk_stream, obtain_video_resolution);
		session_media->open_camera = 1;
		ast_sip_session_set_main_video_stream_uptime(session);
	}

	return;
}

static void set_stream_video_resolution_for_open_camera(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, struct ast_sip_session_media *session_media,
	pjmedia_sdp_media *media, enum ast_stream_video_resolution resolution)
{
	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			set_h264_resolution_for_open_camera(session, format, asterisk_stream, session_media, media, resolution);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			break;
		default:
			break;
	}

	return;
}

static void set_stream_video_resolution_for_close_camera(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, struct ast_sip_session_media *session_media, pjmedia_sdp_media *media)
{
	if (!session->is_sfu || !ast_stream_judge_if_main_video(asterisk_stream)) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] It's main video stream, and camera is closed, need to release video resource for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session_media->changed = 1;
	if (session->is_sfu) {
		session_media->close_camera = 1;
	}

	ast_stream_set_video_resolution(asterisk_stream, AST_STREAM_VIDEO_RESOLUTION_NONE);

	return;
}

static void set_stream_video_resolution_for_change_resolution(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, struct ast_sip_session_media *session_media,
	enum ast_stream_video_resolution recv_video_resolution, enum ast_stream_video_resolution prev_video_resolution)
{
	int res = 0;

	res = ast_conf_switch_main_video_resolution(session->channel, session->conf_room, recv_video_resolution);
	if (res != 0) {
		/* If we can't switch, we need to use previous video resolution */
		ast_log(LOG_DEBUG, "[session:%p] Change resolution failed, need to revert to previous resolution, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		set_stream_params_according_video_resolution(session, format, asterisk_stream, prev_video_resolution);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Change resolution success, change to new resolution:%d, endpoint:%s\n",
			session, recv_video_resolution, ast_sorcery_object_get_id(session->endpoint));
		set_stream_params_according_video_resolution(session, format, asterisk_stream, recv_video_resolution);
	}

	return;
}

static enum ast_stream_video_resolution get_received_video_resolution_for_h264_codec(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream)
{
	enum ast_stream_video_resolution result = AST_STREAM_VIDEO_RESOLUTION_NONE;
	const unsigned int *h264_level = NULL;

	if (ast_stream_get_remote_media_direction(asterisk_stream) == INACTIVE ||
		ast_stream_get_remote_media_direction(asterisk_stream) == RECVONLY) {
		ast_log(LOG_DEBUG, "[session:%p] Stream state is inactive, don't need get video resolution, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return result;
	}

	h264_level = ast_format_attribute_get(format, "level");
	if (!h264_level) {
		ast_log(LOG_WARNING, "[session:%p] Can't get profile level for H264 format, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return result;
	}

	ast_log(LOG_DEBUG, "[sesssion:%p] Get h264_level:%d, format:%p for endpoint:%s\n",
		session, *h264_level, format, ast_sorcery_object_get_id(session->endpoint));

	result = change_level_to_video_resolution(*h264_level);

	return result;
}

static enum ast_stream_video_resolution get_received_video_resolution(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream)
{
	enum ast_stream_video_resolution result = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session) {
		return result;
	}

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			result = get_received_video_resolution_for_h264_codec(session, format, asterisk_stream);
			break;
		case AST_SFU_VIDEO_H265:
			/* Todo */
			break;
		case AST_SFU_VIDEO_VP9:
			/* Todo */
			break;
		default:
			break;
	}

	return result;
}

static enum ast_stream_video_resolution get_video_resolution_from_imageattr(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, pjmedia_sdp_attr *attr)
{
	int rtp_code = 0;
	int send_x = 0;
	int send_y = 0;
	char imageattr_all[MAX_IMAGERTTR_LEN + 4] = {0};
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	ast_copy_pj_str(imageattr_all, &attr->value, sizeof(imageattr_all));
	if (!strlen(imageattr_all)) {
		ast_log(LOG_DEBUG, "[session:%p] get_video_resolution_from_imageattr can't get imageattr vaule, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		resolution = get_received_video_resolution(session, format, asterisk_stream);
		return resolution;
	}

	if (sscanf(imageattr_all, "%d send [x=%d,y=%d]", &rtp_code, &send_x, &send_y) != 3) {
		ast_log(LOG_DEBUG, "[session:%p] get_video_resolution_from_imageattr, can't process this imageattr attr:%s, endpoint:%s\n",
			session, imageattr_all, ast_sorcery_object_get_id(session->endpoint));
		resolution = get_received_video_resolution(session, format, asterisk_stream);
		return resolution;
	}

	ast_log(LOG_DEBUG, "[session:%p] get_video_resolution_from_imageattr x=%d, y=%d, endpoint:%s\n",
		session, send_x, send_y, ast_sorcery_object_get_id(session->endpoint));

	if (send_y < 720) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_360P;
	} else if (send_y < 1080) {
		resolution = AST_STREAM_VIDEO_RESOLUTION_720P;
	} else {
		resolution = AST_STREAM_VIDEO_RESOLUTION_1080P;
	}

	return resolution;
}

static enum ast_stream_video_resolution get_video_resolution(struct ast_sip_session *session,
	struct ast_format *format, struct ast_stream *asterisk_stream, pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!(attr = pjmedia_sdp_media_find_attr2(stream, "imageattr", NULL))) {
		resolution = get_received_video_resolution(session, format, asterisk_stream);
		return resolution;
	}

	resolution = get_video_resolution_from_imageattr(session, format, asterisk_stream, attr);

	return resolution;
}

static void set_stream_video_resolution(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream, pjmedia_sdp_media *stream)
{
	RAII_VAR(struct ast_format *, format, NULL, ao2_cleanup);
	enum ast_stream_video_resolution recv_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;
	enum ast_stream_video_resolution prev_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (check_set_stream_video_resolution_params_valid(session, session_media, asterisk_stream)) {
		return;
	}

	format = ast_sip_session_get_video_format(asterisk_stream, ast_sip_session_sfu_codec2str(session->sfu_video_codec));
	if (!format) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get video format, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	recv_video_resolution = get_video_resolution(session, format, asterisk_stream, stream);
	prev_video_resolution = ast_stream_get_video_resolution(asterisk_stream);

	/* If we have H264 level, but recv_video_resolution is NONE, we don't process this case now, todo later */
	if (recv_video_resolution == prev_video_resolution) {
		/* some profile-level-id we don't process, and decide a value by ourself, we set use prev value */
		set_stream_params_according_video_resolution(session, format, asterisk_stream, prev_video_resolution);
		ast_log(LOG_DEBUG, "[session:%p] Video resolution is the same as before, use prev video resolution, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	switch (prev_video_resolution) {
		case AST_STREAM_VIDEO_RESOLUTION_NONE:
			set_stream_video_resolution_for_open_camera(session, format, asterisk_stream, session_media, stream, recv_video_resolution);
			break;
		case AST_STREAM_VIDEO_RESOLUTION_360P:
		case AST_STREAM_VIDEO_RESOLUTION_720P:
		case AST_STREAM_VIDEO_RESOLUTION_1080P:
			if (recv_video_resolution == AST_STREAM_VIDEO_RESOLUTION_NONE) {
				set_stream_video_resolution_for_close_camera(session, format, asterisk_stream, session_media, stream);
			} else {
				set_stream_video_resolution_for_change_resolution(session, format, asterisk_stream, session_media, recv_video_resolution, prev_video_resolution);
			}
			break;
		default:
			break;
	}

	return;
}

static unsigned int need_set_local_msid(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	/*Start added by zyxia for Bug 250509,2023.2.14*/
	if(ast_channel_page_mode_get(session->channel) ==AST_PAGE_MODE_ANNOUNCE && session_media->sub_type == AST_MEDIA_SUB_VIDEO_TYPE_MAIN ){ 
		return FALSE;
	}
	/*End added by zyxia for Bug 250509,2023.2.14*/

	if (session_media->sub_type != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
		return TRUE;
	}

	if (session->is_sfu &&
		session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) {
		return TRUE;
	}

	return FALSE;
}

static void set_avs_sip_msid(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_sip_session_media *host_session_media = NULL;
	char msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };
	RAII_VAR(struct ast_channel *, focus_channel, NULL, ao2_cleanup);
	RAII_VAR(char *, focus_msid, NULL, ast_free);

	if (!session->channel) {
		return;
	}

	focus_channel = ast_conf_get_sfu_focus(session->channel, session->conf_room);
	if (focus_channel) {
		focus_msid = ast_channel_get_msid(focus_channel, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		if (!ast_strlen_zero(focus_msid)) {
			strncpy(msid, focus_msid, sizeof(msid) - 1);
			ast_log(LOG_DEBUG, "[session:%p] set_avs_sip_msid get focus msid:%s, endpoint:%s\n",
				session, msid, ast_sorcery_object_get_id(session->endpoint));
		} else {
			snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
			ast_log(LOG_DEBUG, "[session:%p] set_avs_sip_msid can't get focus msid, use local msid:%s, endpoint:%s\n",
				session, msid, ast_sorcery_object_get_id(session->endpoint));
		}
	} else {
		host_session_media = ast_sip_session_get_host_main_video_media(session->channel, session->conf_room);

		if ((session_media->sub_type == AST_MEDIA_SUB_VIDEO_TYPE_MAIN) && host_session_media) {
			snprintf(msid, sizeof(msid), "%s %s", host_session_media->mslabel, host_session_media->label);
		} else {
			snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
		}
	}

	if (need_set_local_msid(session, session_media) ){
		ast_rtp_instance_set_local_msid(session_media->rtp, msid);
		ast_log(LOG_DEBUG, "[session:%p] set_avs_sip_msid local msid:%s, endpoint:%s\n",
			session, msid, ast_sorcery_object_get_id(session->endpoint));
	}

	memset(msid, 0, sizeof(msid));
	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
	ast_rtp_instance_set_remote_msid(session_media->rtp, msid);
	ast_log(LOG_DEBUG, "[session:%p] set_avs_sip_msid remote msid:%s, endpoint:%s\n",
		session, msid, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_avs_gswave_gvc_msid(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	char msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

	/* Set local msid */
	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);

	if (need_set_local_msid(session, session_media)) {
		ast_rtp_instance_set_local_msid(session_media->rtp, msid);
	}

	/* Set remote msid, remote msid must the same as local msid, so we just use local to set remote */
	ast_rtp_instance_set_remote_msid(session_media->rtp, msid);

	ast_log(LOG_DEBUG, "[session:%p] set_avs_gswave_gvc_msid value:%s for endpoint:%s\n",
		session, msid, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_avs_gswave_mobile_msid(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_sip_session_media *host_session_media = NULL;
	char msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

	if (!session->channel) {
		return;
	}

	if (ast_sip_session_media_use_bundle(session)) {
		snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
		if (need_set_local_msid(session, session_media)) {
			ast_rtp_instance_set_local_msid(session_media->rtp, msid);
		}
		ast_log(LOG_DEBUG, "[session:%p] Enable bundle, set mobile local msid:%s, endpoint:%s\n",
			session, msid, ast_sorcery_object_get_id(session->endpoint));
	} else {
		host_session_media = ast_sip_session_get_host_main_video_media(session->channel, session->conf_room);

		if ((session_media->sub_type == AST_MEDIA_SUB_VIDEO_TYPE_MAIN) && host_session_media) {
			snprintf(msid, sizeof(msid), "%s %s", host_session_media->mslabel, host_session_media->label);
			ast_rtp_instance_set_local_msid(session_media->rtp, msid);
			ast_log(LOG_DEBUG, "[session:%p] set_avs_gswave_mobile_msid local msid:%s, endpoint:%s\n",
				session, msid, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	memset(msid, 0, sizeof(msid));
	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
	ast_rtp_instance_set_remote_msid(session_media->rtp, msid);
	ast_log(LOG_DEBUG, "[session:%p] set_avs_gswave_mobile_msid remote msid:%s, endpoint:%s\n",
		session, msid, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_avs_msid(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	enum ast_terminal_type terminal_type = AST_TERMINAL_TYPE_UNKNOW;
	
	if (!session || !session_media) {
		return;
	}

	terminal_type = session->terminal_type;

	switch (terminal_type) {
		case AST_TERMINAL_TYPE_SIP:
		case AST_TERMINAL_TYPE_ICE_TRUNK:
			set_avs_sip_msid(session, session_media);
			break;
		case AST_TERMINAL_TYPE_GVC:
		case AST_TERMINAL_TYPE_WEBRTC:
			set_avs_gswave_gvc_msid(session, session_media);
			break;
		case AST_TERMINAL_TYPE_GSWAVE_MOBILE:
			set_avs_gswave_mobile_msid(session, session_media);
			break;
		default:
			break;
	}

	return;
}

static void set_avs_crypto_update_flag(struct ast_sip_session *session,
	struct ast_rtp_instance *rtp, struct ast_sdp_srtp *srtp)
{
	const char *remote_crypto = NULL;
	char *new_remote_crypto = NULL;

	if (!session|| !rtp || !srtp) {
		return;
	}

	ast_rtp_instance_set_crypto_update_flag(rtp, 0);

	remote_crypto = ast_rtp_instance_get_remote_crypto(rtp);
	if (ast_strlen_zero(remote_crypto)) {
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params has no exist remote crypto, set update flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_crypto_update_flag(rtp, 1);
		return;
	}

	new_remote_crypto = ast_sdp_srtp_get_crypto_remote_key(srtp);
	if (ast_strlen_zero(new_remote_crypto)) {
		ast_log(LOG_WARNING, "[session:%p] set_avs_srtp_params has no new remote crypto, something very wrong for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (strcmp(remote_crypto, new_remote_crypto)) {
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params get new crypto is not the same, set update flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_crypto_update_flag(rtp, 1);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params get new crypto is the same, don't set update flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void set_avs_srtp_params(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_rtp_instance *rtp = NULL;
	struct ast_sdp_srtp *srtp = NULL;

	rtp = session_media->rtp;
	srtp = session_media->srtp;

	if (!rtp || !srtp) {
		return;
	}

	if (session_media->encryption == AST_SIP_MEDIA_TRANSPORT_INVALID ||
		session_media->encryption == AST_SIP_MEDIA_ENCRYPT_NONE ||
		session_media->encryption == AST_SIP_MEDIA_AUTO_NO_ENCRYPT ||
		session_media->encryption == AST_SIP_MEDIA_ENCRYPT_ZRTP) {
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params has no SRTP, clear SRTP params for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (!ast_strlen_zero(ast_rtp_instance_get_remote_crypto(rtp))) {
			ast_rtp_instance_set_crypto_update_flag(rtp, 1);
			ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params has no SRTP, but has SRTP preivous, need update for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
		ast_rtp_instance_set_local_crypto(rtp, "");
		ast_rtp_instance_set_remote_crypto(rtp, "");
		ast_rtp_instance_set_crypto_suite(rtp, 0);
	} else {
		set_avs_crypto_update_flag(session, rtp, srtp);

		ast_rtp_instance_set_local_crypto(rtp, ast_sdp_srtp_get_crypto_local_key(srtp));
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params local_key:%s for endpoint:%s\n",
			session, ast_sdp_srtp_get_crypto_local_key(srtp), ast_sorcery_object_get_id(session->endpoint));

		ast_rtp_instance_set_remote_crypto(rtp, ast_sdp_srtp_get_crypto_remote_key(srtp));
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params remote_key:%s for endpoint:%s\n",
			session, ast_sdp_srtp_get_crypto_remote_key(srtp), ast_sorcery_object_get_id(session->endpoint));

		ast_rtp_instance_set_crypto_suite(rtp, ast_sdp_srtp_get_suite(srtp));
		ast_log(LOG_DEBUG, "[session:%p] set_avs_srtp_params suite:%d for endpoint:%s\n",
			session, ast_sdp_srtp_get_suite(srtp), ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void set_avs_jitter_buffer_params(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	struct ast_rtp_instance *rtp = NULL;
	struct ast_rtp_jitterbuffer jitter_buffer = {0};

	rtp = session_media->rtp;
	if (!rtp) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no rtp, don't need set\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (AST_MEDIA_SUB_AUDIO_TYPE != session_media->sub_type) {
		ast_log(LOG_DEBUG, "[session:%p] media is not audio, don't need set for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!strcasecmp(session->endpoint->media.jitter_buffer.type, "normal")) {
		jitter_buffer.mode = ast_rtp_jitterbuffer_normal;
	} else if (!strcasecmp(session->endpoint->media.jitter_buffer.type, "neteq")) {
		jitter_buffer.mode = ast_rtp_jitterbuffer_neteq;
	} else {
		jitter_buffer.mode = ast_rtp_jitterbuffer_disable;
	}

	if (ast_rtp_jitterbuffer_normal == jitter_buffer.mode) {
		if (!strcasecmp(session->endpoint->media.jitter_buffer.normal_type, "adaptive")) {
			jitter_buffer.normal_mode = ast_rtp_jitterbuffer_normal_adaptive;
		} else {
			jitter_buffer.normal_mode = ast_rtp_jitterbuffer_normal_fixed;
		}
		jitter_buffer.normal_size = session->endpoint->media.jitter_buffer.normal_size;
		jitter_buffer.max_normal_size = session->endpoint->media.jitter_buffer.max_normal_size;
	}
	
	ast_rtp_instance_set_jitterbuffer(rtp, &jitter_buffer);
	ast_log(LOG_DEBUG, "[session:%p] set_avs_jitter_buffer_params mode[%d], normal_mode[%d], normal_size[%d], max_normal_size[%d] for endpoint:%s\n", 
			session, jitter_buffer.mode, jitter_buffer.normal_mode, jitter_buffer.normal_size,jitter_buffer.max_normal_size, ast_sorcery_object_get_id(session->endpoint));
	return;
}

static void set_avs_ice_restart(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *remote_stream, const pjmedia_sdp_session *remote)
{
	struct ast_rtp_engine_ice *ice = NULL;
	const char *username, *password;
	pjmedia_sdp_attr *attr = NULL;
	char attr_value[256] = { 0 };
	unsigned int need_ice_restart = 0;
	unsigned int attr_i = 0;

	if (!session || !session->channel || !session_media || !session_media->rtp || !remote_stream || !remote) {
		ast_log(LOG_DEBUG, "[session:%p] Param is NULL, don't need set ICE restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		ast_log(LOG_DEBUG, "[session:%p] rtp not ice support, don't need set ICE restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	username = ast_rtp_instance_get_remote_ice_ufrag(session_media->rtp);
	if (ast_strlen_zero(username)) {
		ast_log(LOG_DEBUG, "[session:%p] Has no remote ufrag, don't need set ICE restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] set_avs_ice_restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));

	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-ufrag", NULL);
	if (!attr) {
		attr = pjmedia_sdp_attr_find2(remote->attr_count, remote->attr, "ice-ufrag", NULL);
	}
	if (!attr) {
		return;
	}
	ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));
	if (strcmp(username, attr_value)) {
		ast_log(LOG_DEBUG, "[session:%p] Remote ICE ufrag is not the same, need set ICE restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ice->set_authentication(session_media->rtp, attr_value, NULL);
		need_ice_restart = 1;
	}

	if (!(password = ast_rtp_instance_get_remote_ice_pwd(session_media->rtp))) {
		ast_log(LOG_DEBUG, "[session:%p] Has no remote password, don't need set ICE restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-pwd", NULL);
	if (!attr) {
		attr = pjmedia_sdp_attr_find2(remote->attr_count, remote->attr, "ice-pwd", NULL);
	}
	if (!attr) {
		return;
	}
	ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));
	if (strcmp(password, attr_value)) {
		ast_log(LOG_DEBUG, "[session:%p] Remote ICE password is not the same, need set ICE restart for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ice->set_authentication(session_media->rtp, NULL, attr_value);
		need_ice_restart = 1;
	}

	if (need_ice_restart) {
		/* Clear exist remote candidates first */
		ast_rtp_instance_clear_remote_candidates(session_media->rtp);

		/* Find all of the candidates */
		for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
			char foundation[33], transport[32], address[PJ_INET6_ADDRSTRLEN + 1], cand_type[6], relay_address[PJ_INET6_ADDRSTRLEN + 1] = "";
			unsigned int port, relay_port = 0;
			struct ast_rtp_engine_ice_candidate candidate = { 0, };

			attr = remote_stream->attr[attr_i];

			/* If this is not a candidate line skip it */
			if (pj_strcmp2(&attr->name, "candidate")) {
				continue;
			}

			ast_copy_pj_str(attr_value, (pj_str_t*)&attr->value, sizeof(attr_value));

			if (sscanf(attr_value, "%32s %30u %31s %30u %46s %30u typ %5s %*s %23s %*s %30u", foundation, &candidate.id, transport,
				(unsigned *)&candidate.priority, address, &port, cand_type, relay_address, &relay_port) < 7) {
				/* Candidate did not parse properly */
				continue;
			}

			if (session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux && candidate.id > 1) {
				/* Remote side may have offered RTP and RTCP candidates. However, if we're using RTCP MUX,
				 * then we should ignore RTCP candidates.
				 */
				continue;
			}

			candidate.foundation = foundation;
			candidate.transport = transport;

			ast_sockaddr_parse(&candidate.address, address, PARSE_PORT_FORBID);
			ast_sockaddr_set_port(&candidate.address, port);

			if (!strcasecmp(cand_type, "host")) {
				candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_HOST;
			} else if (!strcasecmp(cand_type, "srflx")) {
				candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_SRFLX;
			} else if (!strcasecmp(cand_type, "relay")) {
				candidate.type = AST_RTP_ICE_CANDIDATE_TYPE_RELAYED;
			} else {
				continue;
			}

			if (!ast_strlen_zero(relay_address)) {
				ast_sockaddr_parse(&candidate.relay_address, relay_address, PARSE_PORT_FORBID);
			}

			if (relay_port) {
				ast_sockaddr_set_port(&candidate.relay_address, relay_port);
			}

			ice->add_remote_candidate(session_media->rtp, &candidate);
		}

		ice->set_role(session_media->rtp, pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) == PJ_TRUE ?
			AST_RTP_ICE_ROLE_CONTROLLING : AST_RTP_ICE_ROLE_CONTROLLED);
		ice->start(session_media->rtp);
		ast_avs_restart_ice(session->channel, session_media->rtp);
		session_media->ice_restart = 1;
	}

	return;
}

static void set_avs_params(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	ast_rtp_instance_set_media_type(session_media->rtp, ast_stream_get_sub_type(asterisk_stream));
	ast_rtp_instance_set_transport(session_media->rtp, session_media->transport_cc_id);
	ast_rtp_instance_set_teminal_type(session_media->rtp, session->terminal_type);
	if (strlen(session_media->gs_fec_version)) {
		ast_rtp_instance_set_gsfec_ver(session_media->rtp, session_media->gs_fec_version);
	}

	/* Set rtcp_mux */
	ast_rtp_instance_set_rtcp_mux(session_media->rtp, session->endpoint->media.rtcp_mux && session_media->remote_rtcp_mux);

	/* Set msid */
	set_avs_msid(session, session_media);

	/* Set remote cname */
	ast_rtp_instance_set_remote_cname(session_media->rtp, session_media->remote_cname);

	/* Set srtp params */
	set_avs_srtp_params(session, session_media);

	/* Set jitter buffer params */
	set_avs_jitter_buffer_params(session, session_media);

	/* Set rtcp-rsize */
	ast_rtp_instance_set_rsize_rtcp(session_media->rtp, session_media->rtcp_rsize);

	return;
}

static void set_target_address_for_local_offer(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, unsigned int set_turn_address)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);

	if (!session || !session_media || !session_media->rtp) {
		return;
	}

	if (!session->media_turn_relay) {
		return;
	}

	if (!set_turn_address) {
		ast_log(LOG_DEBUG, "[session:%p] Don't need set turn address for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_strlen_zero(session->turn_target_address)) {
		ast_log(LOG_DEBUG, "[session:%p] Has no turn target address for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sockaddr_resolve(&addrs, session->turn_target_address, PARSE_PORT_IGNORE, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_WARNING, "[session:%p] Can't resolve turn target address for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_sockaddr_set_port(addrs, 6666);

	ast_log(LOG_DEBUG, "[session:%p] Set turn target address:%s for media:%s, endpoint:%s\n",
		session, session->turn_target_address, ast_codec_media_sub_type2str(session_media->sub_type),
		ast_sorcery_object_get_id(session->endpoint));

	ast_rtp_instance_set_remote_address(session_media->rtp, addrs);

	ast_avs_set_targetaddr(session_media->rtp);

	session_media->setted_turn_target = 1;

	return;
}

static void set_target_address_for_remote_offer(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_sockaddr *target_address, unsigned int set_turn_address)
{
	if (!session || !session_media || !session_media->rtp || !target_address) {
		return;
	}

	if (!session->media_turn_relay) {
		return;
	}

	if (!set_turn_address) {
		ast_log(LOG_DEBUG, "[session:%p] Don't need set turn address for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Set turn target address:%s for media:%s, endpoint:%s\n",
		session, session->turn_target_address, ast_codec_media_sub_type2str(session_media->sub_type),
		ast_sorcery_object_get_id(session->endpoint));

	ast_rtp_instance_set_remote_address(session_media->rtp, target_address);

	ast_avs_set_targetaddr(session_media->rtp);

	session_media->setted_turn_target = 1;

	return;
}

static void set_target_address_for_trunk(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct ast_sockaddr *addrs, int port)
{
	struct ast_sockaddr target_address;

	if (!session || !session_media || !session_media->rtp || !addrs) {
		return;
	}

	if (!session->media_turn_relay) {
		return;
	}

	ast_rtp_instance_get_requested_target_address(session_media->rtp, &target_address);

	ast_sockaddr_set_port(addrs, port);
	ast_rtp_instance_set_remote_address(session_media->rtp, addrs);

	if (!session_media->setted_turn_target) {
		ast_avs_set_targetaddr(session_media->rtp);
		session_media->setted_turn_target = 1;
		return;
	}

	if (ast_sockaddr_cmp(addrs, &target_address) != 0) {
		if (!ast_rtp_instance_get_nego_flag(session_media->rtp)) {
			ast_avs_set_targetaddr(session_media->rtp);
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Update turn target address for media:%s, endpoint:%s\n",
				session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
			ast_avs_update_targetaddr(session->channel, session_media->rtp);
		}
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Media:%s target address is the same, don't need set again, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void set_media_net_type(struct ast_sip_session_media *session_media, struct ast_sockaddr *addrs)
{
	if (!session_media || !addrs) {
		return;
	}

	if (ast_sockaddr_is_ipv4(addrs)) {
		session_media->net_type = AST_MEDIA_NET_TYPE_IPV4;
	} else {
		session_media->net_type = AST_MEDIA_NET_TYPE_IPV6;
	}

	return;
}

static void set_media_net_type_for_local_offer(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, unsigned int set_turn_address)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);

	if (!session || !session_media) {
		return;
	}

	if (!session->media_turn_relay) {
		return;
	}

	if (!set_turn_address) {
		ast_log(LOG_DEBUG, "[session:%p] Don't need set net type for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_strlen_zero(session->turn_target_address)) {
		ast_log(LOG_DEBUG, "[session:%p] Has no turn target address for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sockaddr_resolve(&addrs, session->turn_target_address, PARSE_PORT_IGNORE, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_WARNING, "[session:%p] Can't resolve turn target address for media:%s, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	set_media_net_type(session_media, addrs);

	return;
}

static void remove_additional_codec_for_avs(struct ast_format_cap *joint_cap)
{
	/* AVS is consider RED/FEC as not common codec, need set in other way */

	if (ast_format_cap_has_red(joint_cap)) {
		ast_format_cap_remove_red(joint_cap);
	}
	if (ast_format_cap_has_audio_fec(joint_cap)) {
		ast_format_cap_remove_audio_fec(joint_cap);
	}
	if (ast_format_cap_has_video_fec(joint_cap)) {
		ast_format_cap_remove_video_fec(joint_cap);
	}

	return;
}

static void modify_video_joint_caps_for_sfu(struct ast_sip_session *session,
	struct ast_format_cap *joint)
{
	unsigned int index = 0;
	char conf_codec[8] = { 0 };
	struct ast_format *format = NULL;
	struct ast_str *format_buf = ast_str_alloca(64);

	switch (session->sfu_video_codec) {
		case AST_SFU_VIDEO_H264:
			strcpy(conf_codec, "H264");
			break;
		case AST_SFU_VIDEO_H265:
			strcpy(conf_codec, "H265");
			break;
		case AST_SFU_VIDEO_VP9:
			strcpy(conf_codec, "VP9");
			break;
		default:
			strcpy(conf_codec, "H264");
			break;
	}

	for (index = 0; index < ast_format_cap_count(joint); ++index) {
		format = ast_format_cap_get_format(joint, index);

		if (strcasecmp(ast_format_get_name(format), conf_codec) &&
			strcasecmp(ast_format_get_name(format), "rtx") &&
			strcasecmp(ast_format_get_name(format), "GS-FEC")&&
			strcasecmp(ast_format_get_name(format), "fec")) {
			ast_format_cap_remove(joint, format);
			ao2_ref(format, -1);
			continue;
		}

		ao2_ref(format, -1);
	}

	ast_log(LOG_DEBUG, "[session:%p] After modify joint video caps:%s for endpoint:%s\n",
		session, ast_format_cap_get_names(joint, &format_buf), ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void modify_joint_caps_for_sfu(struct ast_sip_session *session,
	struct ast_format_cap *joint, struct ast_stream *asterisk_stream)
{
	if (!session || !joint || !asterisk_stream) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!session->channel || ast_strlen_zero(session->conf_room)) {
		return;
	}

	if (ast_stream_judge_if_main_video(asterisk_stream) ||
		ast_stream_get_sub_type(asterisk_stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
		modify_video_joint_caps_for_sfu(session, joint);
	}

	return;
}

static void modify_cn_according_joint_caps(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format_cap *caps)
{
	unsigned int index = 0;
	unsigned int rate = 0;
	unsigned int cn_8k_format_find = 0;
	unsigned int cn_16k_format_find = 0;
	unsigned int cn_32k_format_find = 0;
	struct ast_format *format = NULL;

	if (!session || !session_media || !caps) {
		ast_log(LOG_ERROR, "[session:%p] modify_cn_according_joint_caps param is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		rate = ast_rtp_lookup_sample_rate2(1, format, 0);
		ao2_ref(format, -1);

		if (rate == 8000 && !cn_8k_format_find) {
			cn_8k_format_find = 1;
		}

		if (rate == 16000 && !cn_16k_format_find) {
			cn_16k_format_find = 1;
		}

		if (rate == 32000 && !cn_32k_format_find) {
			cn_32k_format_find = 1;
		}
	}

	if (!cn_8k_format_find && session_media->cn_8k_pt) {
		ast_log(LOG_DEBUG, "[session:%p] Can't find CN 8K according format, clear it, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session_media->cn_8k_pt = 0;
		ast_rtp_instance_set_cn_8k_pt(session_media->rtp, 0);
	}

	if (!cn_16k_format_find && session_media->cn_16k_pt) {
		ast_log(LOG_DEBUG, "[session:%p] Can't find CN 16K according format, clear it, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session_media->cn_16k_pt = 0;
		ast_rtp_instance_set_cn_16k_pt(session_media->rtp, 0);
	}

	if (!cn_32k_format_find && session_media->cn_32k_pt) {
		ast_log(LOG_DEBUG, "[session:%p] Can't find CN 32K according format, clear it, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session_media->cn_32k_pt = 0;
		ast_rtp_instance_set_cn_32k_pt(session_media->rtp, 0);
	}

	return;
}

/* Add by bfdu for audio other attr passthrough, 2017.07.04 */
static unsigned char is_audio_rtcp_attr(char *attr_name)
{
	if (!attr_name) {
		return 0;
	}

	/* Now for audio media, we just support rtcp transparent, if need other, support later, 2017.07.13 */
	if (!strcmp(attr_name, "rtcp")) {
		return 1;
	}

	return 0;
}

static unsigned char is_audio_rtcp_fb_attr(char *attr_name)
{
	if (!attr_name) {
		return 0;
	}

	/* Now for audio media, we just support rtcp transparent, if need other, support later, 2017.07.13 */
	if (!strcmp(attr_name, "rtcp-fb")) {
		return 1;
	}

	return 0;
}

static void set_rtcp_attr_param(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream, int enable_nack)
{
	int i = 0;
	char attr_name[16]   = {0};
	char attr_value[128] = {0};
	char rtcp_str[8] = { 0 };
	char net_type[8] = { 0 };
	char rtcp_address[128] = { 0 };
	unsigned int rtcp_port = 0;

	if (!session_media || !stream || !asterisk_stream) {
		ast_log(LOG_WARNING, "set_rtcp_attr_param para is NULL\n");
		return;
	}

	for (i = 0; i < stream->attr_count; i++) {
		memset(attr_name, 0, sizeof(attr_name));
		memset(attr_value, 0, sizeof(attr_value));
		ast_copy_pj_str(attr_name, &stream->attr[i]->name, sizeof(attr_name));
		ast_copy_pj_str(attr_value, &stream->attr[i]->value, sizeof(attr_value));

		if (is_audio_rtcp_attr(attr_name)) {
			if (sscanf(attr_value, "%30u %s %s %s", &rtcp_port, rtcp_str, net_type, rtcp_address) < 4) {
				continue;
			}
			session_media->remote_rtcp_port = rtcp_port;
			if (!session->media_turn_relay) {
				set_port_update_flag(session, session_media->rtp, NULL, rtcp_port);
				ast_rtp_instance_set_remote_rtcp(session_media->rtp, rtcp_port);
			}
			ast_log(LOG_DEBUG, "Media type:%s get rtcp value:%s\n", ast_codec_media_sub_type2str(session_media->sub_type), attr_value);
			session_media->has_rtcp_address = 1;
			ast_stream_set_has_rtcp_address(asterisk_stream, 1);
		}

		if (!is_audio_rtcp_fb_attr(attr_name)) {
			continue;
		}

		if (strstr(attr_value, "nack pli")) {
			session_media->has_nack_pli = 1;
			ast_stream_set_has_nack_pli(asterisk_stream, 1);
			ast_rtp_instance_set_pli(session_media->rtp, 1);
		} else if (strstr(attr_value, "nack") && enable_nack) {
			session_media->has_nack = 1;
			ast_stream_set_has_nack(asterisk_stream, 1);
			ast_rtp_instance_set_nack(session_media->rtp, 1);
		} else if (strstr(attr_value, "ccm fir")) {
			session_media->has_ccm_fir = 1;
			ast_stream_set_has_ccm_fir(asterisk_stream, 1);
			ast_rtp_instance_set_fir(session_media->rtp, 1);
		}
	}

	return;
}

static void set_stream_framerate(struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	char framerate[MAX_FRAMERATE_LEN] = {0};

	if ((attr = pjmedia_sdp_media_find_attr2(stream, "framerate", NULL))) {
		ast_copy_pj_str(framerate, &attr->value, sizeof(framerate));
		if (strlen(framerate)) {
			ast_log(LOG_DEBUG, "set_stream_framerate, framerate:%s\n", framerate);
			ast_stream_set_metadata(asterisk_stream, "FRAMERATE", framerate);
		}
	}

	return;
}

static void set_stream_content(struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	char content[MAX_CONTENT_LEN] = {0};

	if ((attr = pjmedia_sdp_media_find_attr2(stream, "content", NULL))) {
		ast_copy_pj_str(content, &attr->value, sizeof(content));
		if (strlen(content)) {
			ast_stream_set_metadata(asterisk_stream, "CONTENT", content);
		}
	}

	return;
}

static void set_stream_label(struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	char label[MAX_LABEL_LEN] = {0};

	if ((attr = pjmedia_sdp_media_find_attr2(stream, "label", NULL))) {
		ast_copy_pj_str(label, &attr->value, sizeof(label));
		if (strlen(label)) {
			ast_stream_set_metadata(asterisk_stream, "LABEL", label);
		}
	}

	return;
}

static void set_stream_imageattr(struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	char imageattr[MAX_IMAGERTTR_LEN] = {0};
	char imageattr_all[MAX_IMAGERTTR_LEN + 4] = {0};
	int rtp_code = 0;

	ast_stream_set_metadata(asterisk_stream, "IMAGEATTR", NULL);

	if (!(attr = pjmedia_sdp_media_find_attr2(stream, "imageattr", NULL))) {
		return;
	}

	ast_copy_pj_str(imageattr_all, &attr->value, sizeof(imageattr_all));
	if (!strlen(imageattr_all)) {
		return;
	}

	if (sscanf(imageattr_all, "%d %[^\n]", &rtp_code, imageattr) != 2) {
		ast_log(LOG_DEBUG, "set_stream_imageattr, can't process this imageattr attr:%s\n", imageattr_all);
		return;
	}

	ast_log(LOG_DEBUG, "set_stream_imageattr, imageattr:%s\n", imageattr);
	ast_stream_set_metadata(asterisk_stream, "IMAGEATTR", imageattr);

	return;
}

static void set_video_extra_attr_para(const struct pjmedia_sdp_media *stream,
	struct ast_stream *asterisk_stream, enum ast_media_type media_type)
{
	if (media_type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	set_stream_framerate(asterisk_stream, stream);
	set_stream_content(asterisk_stream, stream);
	set_stream_label(asterisk_stream, stream);
	set_stream_imageattr(asterisk_stream, stream);

	return;
}

static void set_passthrough_attr_param(const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	unsigned int index = 0;
	pjmedia_sdp_attr *attr = NULL;

	for (index = 0; index < stream->attr_count; ++index) {
		attr = stream->attr[index];
		if (!attr) {
			continue;
		}
		
		char attr_name[pj_strlen(&attr->name) + 1];
		char attr_value[pj_strlen(&attr->value) + 1];

		ast_copy_pj_str(attr_name, &attr->name, sizeof(attr_name));
		ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

		if (!ast_sip_session_judge_if_passthrough_attr(attr_name)) {
			continue;
		}

		ast_log(LOG_DEBUG, "set_passthrough_attr_param attr_name:%s, attr_value:%s\n", attr_name, attr_value);

		ast_stream_set_passthrough(asterisk_stream, attr_name, attr_value, AST_PASSTHROUGH_TYPE_ATTR);
	}

	return;
}

static void set_passthrough_bandw_param(const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	unsigned int index = 0;
	pjmedia_sdp_bandw *bandw = NULL;

	for (index = 0; index < stream->bandw_count; index++) {
		bandw = stream->bandw[index];
		if (!bandw ) {
			continue;
		}

		char bandw_name[pj_strlen(&bandw->modifier) + 1];
		char bandw_value[16] = {0};

		ast_copy_pj_str(bandw_name, &bandw->modifier, sizeof(bandw_name));
		if (!ast_sip_session_judge_if_passthrough_bandw(bandw_name)) {
			continue;
		}

		snprintf(bandw_value, sizeof(bandw_value), "%d", bandw->value);

		ast_log(LOG_DEBUG, "set_passthrough_bandw_param attr_name:%s, attr_value:%s\n", bandw_name, bandw_value);

		ast_stream_set_passthrough(asterisk_stream, bandw_name, bandw_value, AST_PASSTHROUGH_TYPE_BANDW);
	}

	return;
}

static void set_passthrough_param(const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	/* We don't process passthrough for audio type */
	if (ast_stream_get_type(asterisk_stream) == AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (!ast_sip_session_judge_if_need_passthrough()) {
		return;
	}

	ast_stream_clear_passthrough(asterisk_stream);

	set_passthrough_attr_param(stream, asterisk_stream);
	set_passthrough_bandw_param(stream, asterisk_stream);

	return;
}

static void set_media_port(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_media *media, int direct_media_enabled, struct ast_sockaddr *addr)
{
	if (!session || !session_media || !media) {
		return;
	}

	if (session->media_turn_relay) {
		media->desc.port = (pj_uint16_t) ast_rtp_instance_get_local_rtp(session_media->rtp);
	} else if (direct_media_enabled) {
		media->desc.port = ast_sockaddr_port(&session_media->direct_media_addr);
	} else {
		media->desc.port = (pj_uint16_t) ast_sockaddr_port(addr);
	}

	return;
}

static const char *get_turn_relay_address(struct ast_sip_session_media *session_media)
{
	const char *ipv4_addr = NULL;
	const char *ipv6_addr = NULL;

	ipv4_addr = ast_rtp_instance_get_local_ipaddr(session_media->rtp);
	if (!ast_strlen_zero(ipv4_addr)) {
		return ipv4_addr;
	}

	ipv6_addr = ast_rtp_instance_get_local_ipaddr_v6(session_media->rtp);

	return ipv6_addr;
}

static const char *get_media_host_ip(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_media *media, int direct_media_enabled)
{
	if (!session || !session_media || !media) {
		return NULL;
	}

	if (session->media_turn_relay) {
		return get_turn_relay_address(session_media);
	} else if (direct_media_enabled) {
		return ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR);
	} else if (!ast_strlen_zero(session->avs_address) && strcmp(session->avs_address,"127.0.0.1")) {
		return session->avs_address;
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		return ast_sip_get_host_ip_string(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET());
	} else {
		return session->endpoint->media.address;
	}

	return NULL;
}

static void add_passthrough_attr_param(pj_pool_t *pool, struct pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	struct ast_variable *attr_list = NULL;
	struct ast_variable *attr_var = NULL;

	attr_list = ast_stream_get_passthrough_list(stream, AST_PASSTHROUGH_TYPE_ATTR);

	for (attr_var = attr_list; attr_var; attr_var = attr_var->next) {
		attr = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_attr);
		pj_strdup2_with_null(pool, &attr->name, attr_var->name);
		pj_strdup2_with_null(pool, &attr->value, attr_var->value);
		media->attr[media->attr_count++] = attr;
		ast_log(LOG_DEBUG, "add_passthrough_attr_param name:%s, value:%s\n", attr_var->name, attr_var->value);
	}

	if (attr_list) {
		ast_variables_destroy(attr_list);
		attr_list = NULL;
	}

	return;
}

static void add_passthrough_bandw_param(pj_pool_t *pool, struct pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pjmedia_sdp_bandw *bandw = NULL;
	struct ast_variable *bandw_list = NULL;
	struct ast_variable *bandw_var = NULL;

	bandw_list = ast_stream_get_passthrough_list(stream, AST_PASSTHROUGH_TYPE_BANDW);

	for (bandw_var = bandw_list; bandw_var; bandw_var = bandw_var->next) {
		bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
		pj_strdup2_with_null(pool, &bandw->modifier, bandw_var->name);
		bandw->value = atoi(bandw_var->value);
		media->bandw[media->bandw_count++] = bandw;
		ast_log(LOG_DEBUG, "add_passthrough_bandw_param modifier:%s, value:%s\n", bandw_var->name, bandw_var->value);
	}

	if (bandw_list) {
		ast_variables_destroy(bandw_list);
		bandw_list = NULL;
	}

	return;
}

static void add_passthrough_param_to_stream(pj_pool_t *pool, struct pjmedia_sdp_media *media, struct ast_stream *stream)
{
	/* We don't process passthrough for audio type */
	if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (!ast_sip_session_judge_if_need_passthrough()) {
		return;
	}

	add_passthrough_attr_param(pool, media, stream);
	add_passthrough_bandw_param(pool, media, stream);

	return;
}

static void set_gs_fec_attr_param(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream)
{
	char attr_value[MAX_GS_FEC_VERSION_LEN] = {0};
	pjmedia_sdp_attr *fec_version_attr = NULL;

	if (!session || !stream || !session_media || !asterisk_stream) {
		ast_log(LOG_WARNING, "set_gs_fec_attr_param para is NULL\n");
		return;
	}

	fec_version_attr = pjmedia_sdp_media_find_attr2(stream, "gs-fec-version", NULL);
	if (fec_version_attr) {
		ast_copy_pj_str(attr_value, &fec_version_attr->value, sizeof(attr_value));
		strncpy(session_media->gs_fec_version, attr_value, sizeof(session_media->gs_fec_version) - 1);
		session_media->alone_fec_version = 1;
		ast_stream_set_alone_fec_version(asterisk_stream, 1);
		ast_log(LOG_DEBUG, "[session:%p] Find gs-fec-version attr:%s for endpoint:%s\n",
			session, attr_value, ast_sorcery_object_get_id(session->endpoint));
	}

	if (strlen(session_media->gs_fec_version)) {
		ast_stream_set_gs_fec_version(asterisk_stream, session_media->gs_fec_version);
	} else {
		ast_stream_set_gs_fec_version(asterisk_stream, NULL);
	}

	return;
}

static void set_as_tias_for_normal_call(const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	unsigned int index = 0;
	pjmedia_sdp_bandw *bandw = NULL;

	if (!stream || !asterisk_stream) {
		return;
	}

	for (index = 0; index < stream->bandw_count; index++) {
		bandw = stream->bandw[index];
		if (!bandw ) {
			continue;
		}

		char bandw_name[pj_strlen(&bandw->modifier) + 1];
		char bandw_value[16] = {0};

		ast_copy_pj_str(bandw_name, &bandw->modifier, sizeof(bandw_name));
		if (ast_sip_session_judge_if_passthrough_bandw(bandw_name)) {
			continue;
		}

		if (!strcasecmp(bandw_name, "AS")) {
			snprintf(bandw_value, sizeof(bandw_value), "%d", bandw->value);
			if (strlen(bandw_value)) {
				ast_log(LOG_DEBUG, "set_as_tias_for_normal_call, AS:%s\n", bandw_value);
				ast_stream_set_metadata(asterisk_stream, "AS", bandw_value);
			}
		}

		if (!strcasecmp(bandw_name, "TIAS")) {
			snprintf(bandw_value, sizeof(bandw_value), "%d", bandw->value);
			if (strlen(bandw_value)) {
				ast_log(LOG_DEBUG, "set_as_tias_for_normal_call, TIAS:%s\n", bandw_value);
				ast_stream_set_metadata(asterisk_stream, "TIAS", bandw_value);
			}
		}
	}

	return;
}

static void set_as_tias_for_stream(struct ast_sip_session *session,
	const struct pjmedia_sdp_media *stream, struct ast_stream *asterisk_stream)
{
	if (!session) {
		return;
	}

	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p]  support_special_attrs is disable, don't set AS/TIAS for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->is_sfu) {
		set_as_tias_for_normal_call(stream, asterisk_stream);
	} else {
		ast_sip_session_set_as_tias_for_sfu(session, asterisk_stream);
	}

	return;
}

static void add_bandw_to_media(pj_pool_t *pool, struct pjmedia_sdp_media *media,
	char *bandw_name, unsigned int bandw_value)
{
	pjmedia_sdp_bandw *bandw = NULL;

	bandw = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
	if (!bandw) {
		ast_log(LOG_ERROR, "Can't alloc bandw from pool\n");
		return;
	}

	pj_strdup2_with_null(pool, &bandw->modifier, bandw_name);
	bandw->value = bandw_value;
	media->bandw[media->bandw_count++] = bandw;

	ast_log(LOG_DEBUG, "add_bandw_to_media modifier, name:%s value:%d\n", bandw_name, bandw_value);

	return;
}

static void add_as_tias_to_steam(pj_pool_t *pool, struct pjmedia_sdp_media *media, struct ast_stream *stream)
{
	unsigned int enable_p2p_video_control = 0;
	unsigned int configured_bitrate = 0;

	if (!pool || !media || !stream) {
		return;
	}

	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	enable_p2p_video_control = ast_sip_session_get_enable_p2p_video_control();
	configured_bitrate = ast_sip_get_p2p_video_bitrate();

	const char *as = ast_stream_get_metadata(stream, "AS");
	const char *tias = ast_stream_get_metadata(stream, "TIAS");

	if (!ast_strlen_zero(as)) {
		/* Configured value is used for AS, need to plus 192 to compare */
		if (enable_p2p_video_control && ((configured_bitrate + 192) < atoi(as))) {
			add_bandw_to_media(pool, media, "AS", configured_bitrate + 192);
		} else {
			add_bandw_to_media(pool, media, "AS", atoi(as));
		}
	} else {
		if (enable_p2p_video_control && ast_strlen_zero(tias)) {
			add_bandw_to_media(pool, media, "AS", configured_bitrate + 192);
		}
	}

	if (!ast_strlen_zero(tias)) {
		if (enable_p2p_video_control && (configured_bitrate*1000 < atoi(tias))) {
			add_bandw_to_media(pool, media, "TIAS", configured_bitrate*1000);
		} else {
			add_bandw_to_media(pool, media, "TIAS", atoi(tias));
		}
	} else {
		if (enable_p2p_video_control && ast_strlen_zero(as)) {
			add_bandw_to_media(pool, media, "TIAS", configured_bitrate*1000);
		}
	}

	return;
}

static void add_trickle_ice_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !session_media || !pool || !media) {
		return;
	}

	if (!session_media->trickle_ice) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_trickle_ice_to_stream for media:%s endpoint:%s\n",
		session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));

	attr = pjmedia_sdp_attr_create(pool, "ice-options", pj_cstr(&stmp, "trickle"));
	media->attr[media->attr_count++] = attr;

	return;
}

static void add_rtcp_rsize_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !session_media || !pool || !media) {
		return;
	}

	if (!session_media->rtcp_rsize) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_rtcp_rsize_to_stream for media:%s endpoint:%s\n",
		session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));

	attr = pjmedia_sdp_attr_create(pool, "rtcp-rsize", NULL);
	media->attr[media->attr_count++] = attr;

	return;
}

static int add_record_indicate_attr(struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_media *sdp_media)
{
	if (!session || !pool || !sdp_media) {
		ast_log(LOG_ERROR, "Invalid param");
		return -1;
	}

	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p] support_special_attrs configure is disable, don't need add record attr, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (strstr(sdp_media->desc.media.ptr, "audio")) {
		const char *rec_enable;
		pj_str_t stmp;
		pjmedia_sdp_attr *attr;
		pjmedia_sdp_attr *record_time_attr = NULL;

		rec_enable = pbx_builtin_getvar_helper(session->channel, "REC_ENABLE");
		if (!ast_strlen_zero(rec_enable) && !strcmp(rec_enable, "yes")) {
			session->auto_record = 1;
		} else {
			session->auto_record = 0;
		}

		if (session->auto_record) {
			char start_time[16] = "";
			time_t start_time_value = ast_channel_get_recordstart_utc(session->channel);
			snprintf(start_time, sizeof(start_time)-1, "%ld", start_time_value ?: ast_now_utc_timestamp());

			record_time_attr = pjmedia_sdp_attr_create(pool, "record-time", pj_cstr(&stmp, start_time));
			pjmedia_sdp_attr_add(&sdp_media->attr_count, sdp_media->attr, record_time_attr);

			attr = pjmedia_sdp_attr_create(pool, "record", pj_cstr(&stmp, "on"));
		} else {
			attr = pjmedia_sdp_attr_create(pool, "record", pj_cstr(&stmp, "off"));
		}

		pjmedia_sdp_attr_add(&sdp_media->attr_count, sdp_media->attr, attr);
	}

	return 0;
}

static void add_msid_attr_to_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;
	char msid[256] = { 0 };

	if (!session || !session_media || !pool || !media) {
		return;
	}

	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p] support_special_attrs configure is disable, don't need add msid attr, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add_msid_to_stream for media:%s endpoint:%s\n",
		session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));

	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);

	attr = pjmedia_sdp_attr_create(pool, "msid", pj_cstr(&stmp, msid));
	media->attr[media->attr_count++] = attr;

	return;
}

static void add_content_attr_to_stream(struct ast_sip_session *session,
	struct ast_stream *stream, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!session || !stream || !pool || !media) {
		return;
	}

	attr = pjmedia_sdp_media_find_attr2(media, "content", NULL);
	if (attr) {
		return;
	}

	type = ast_stream_get_type(stream);
	sub_type = ast_stream_get_sub_type(stream);

	if (ast_stream_judge_if_stream_is_participant(stream) ||
		sub_type == AST_MEDIA_SUB_AUDIO_TYPE ||
		type == AST_MEDIA_TYPE_APPLICATION) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Add missed content attr to stream for media:%s, endpoint:%s\n",
		session, ast_codec_media_sub_type2str(sub_type), ast_sorcery_object_get_id(session->endpoint));

	switch (sub_type) {
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			attr = pjmedia_sdp_attr_create(pool, "content", pj_cstr(&stmp, "main"));
			media->attr[media->attr_count++] = attr;
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			attr = pjmedia_sdp_attr_create(pool, "content", pj_cstr(&stmp, "slides"));
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
	}

	return;
}

static void add_uptime_attr_to_stream(struct ast_sip_session *session,
	struct ast_stream *stream, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;
	char *uptime = NULL;

	if (!session || !stream || !pool || !media) {
		return;
	}

	attr = pjmedia_sdp_media_find_attr2(media, "uptime", NULL);
	if (attr) {
		pjmedia_sdp_media_remove_attr(media, attr);
	}

	uptime = ast_stream_get_uptime(stream);
	if (ast_strlen_zero(uptime)) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add uptime:%s to stream, endpoint:%s\n",
		session, uptime, ast_sorcery_object_get_id(session->endpoint));

	attr = pjmedia_sdp_attr_create(pool, "uptime", pj_cstr(&stmp, uptime));
	if (!attr) {
		return;
	}

	media->attr[media->attr_count++] = attr;

	return;
}

static void add_zrtphash_attr_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;
	const char *zrtphash = NULL;

	if (!session || !stream || !pool || !media) {
		return;
	}

	if (check_if_direct_media_enabled(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Enable direct media, don't need create zrtphash for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	attr = pjmedia_sdp_media_find_attr2(media, "zrtp-hash", NULL);
	if (attr) {
		pjmedia_sdp_media_remove_attr(media, attr);
	}

	zrtphash = ast_rtp_instance_get_local_zrtphash(session_media->rtp);
	if (ast_strlen_zero(zrtphash)) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] add zrtphash:%s to stream, endpoint:%s\n",
		session, zrtphash, ast_sorcery_object_get_id(session->endpoint));

	attr = pjmedia_sdp_attr_create(pool, "zrtp-hash", pj_cstr(&stmp, zrtphash));
	if (!attr) {
		return;
	}

	media->attr[media->attr_count++] = attr;

	return;
}

static pjmedia_sdp_attr* generate_cn_rtpmap_attr(struct ast_sip_session *session, pjmedia_sdp_media *media, pj_pool_t *pool,
					      int rtp_code, unsigned int rate)
{
#ifndef HAVE_PJSIP_ENDPOINT_COMPACT_FORM
	extern pj_bool_t pjsip_use_compact_form;
#else
	pj_bool_t pjsip_use_compact_form = pjsip_cfg()->endpt.use_compact_form;
#endif
	pjmedia_sdp_rtpmap rtpmap;
	pjmedia_sdp_attr *attr = NULL;
	char tmp[64];

	snprintf(tmp, sizeof(tmp), "%d", rtp_code);
	pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], tmp);

	if (rtp_code <= AST_RTP_PT_LAST_STATIC && pjsip_use_compact_form) {
		return NULL;
	}

	rtpmap.pt = media->desc.fmt[media->desc.fmt_count - 1];
	rtpmap.clock_rate = rate;
	pj_strdup2(pool, &rtpmap.enc_name, "CN");
	pj_cstr(&rtpmap.param, NULL);

	pjmedia_sdp_rtpmap_to_attr(pool, &rtpmap, &attr);

	return attr;
}

static void add_comfort_noise_to_sdp_use_remote(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media, pj_pool_t *pool)
{
	pjmedia_sdp_attr *attr = NULL;

	if (session_media->cn_8k_pt) {
		if ((attr = generate_cn_rtpmap_attr(session, media, pool, session_media->cn_8k_pt, 8000))) {
			media->attr[media->attr_count++] = attr;
		}
	}

	if (session_media->cn_16k_pt) {
		if ((attr = generate_cn_rtpmap_attr(session, media, pool, session_media->cn_16k_pt, 16000))) {
			media->attr[media->attr_count++] = attr;
		}
	}

	if (session_media->cn_32k_pt) {
		if ((attr = generate_cn_rtpmap_attr(session, media, pool, session_media->cn_32k_pt, 32000))) {
			media->attr[media->attr_count++] = attr;
		}
	}

	return;
}

static void add_comfort_noise_to_sdp_use_local(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format_cap *caps, pjmedia_sdp_media *media, pj_pool_t *pool)
{
	unsigned int index = 0;
	unsigned int rate = 0;
	unsigned int cn_8k_pt_added = 0;
	unsigned int cn_16k_pt_added = 0;
	unsigned int cn_32k_pt_added = 0;
	int valid_pt = 0;
	struct ast_format *format = NULL;
	pjmedia_sdp_attr *attr = NULL;

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		rate = ast_rtp_lookup_sample_rate2(1, format, 0);
		ao2_ref(format, -1);

		if (rate == 8000 && !cn_8k_pt_added) {
			if ((attr = generate_cn_rtpmap_attr(session, media, pool, 13, 8000))) {
				media->attr[media->attr_count++] = attr;
			}
			cn_8k_pt_added = 1;
		}

		if (rate == 16000 && !cn_16k_pt_added) {
			valid_pt = get_media_valid_pt(media, 104);
			if ((attr = generate_cn_rtpmap_attr(session, media, pool, valid_pt, 16000))) {
				media->attr[media->attr_count++] = attr;
			}
			cn_16k_pt_added = 1;
		}

		if (rate == 32000 && !cn_32k_pt_added) {
			valid_pt = get_media_valid_pt(media, 105);
			if ((attr = generate_cn_rtpmap_attr(session, media, pool, valid_pt, 32000))) {
				media->attr[media->attr_count++] = attr;
			}
			cn_32k_pt_added = 1;
		}

		if (cn_8k_pt_added && cn_16k_pt_added && cn_32k_pt_added) {
			break;
		}
	}

	return;
}

static void add_comfort_noise_to_sdp(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media, struct ast_format_cap *caps,
	pj_pool_t *pool, const struct pjmedia_sdp_session *remote)
{
	if (!session->endpoint->media.comfort_noise) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (remote) {
		add_comfort_noise_to_sdp_use_remote(session, session_media, media, pool);
	} else {
		add_comfort_noise_to_sdp_use_local(session, session_media, caps, media, pool);
	}

	return;
}

static void add_mpa_to_sdp(pj_pool_t *pool, struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media)
{
#ifndef HAVE_PJSIP_ENDPOINT_COMPACT_FORM
	extern pj_bool_t pjsip_use_compact_form;
#else
	pj_bool_t pjsip_use_compact_form = pjsip_cfg()->endpt.use_compact_form;
#endif

	pjmedia_sdp_attr *attr = NULL;
	enum ast_page_mode page_mode = AST_PAGE_MODE_NONE;
	const char *one_way_page = NULL;

	if (!session || !session_media || !media || !session->channel) {
		return;
	}

	if (!session->endpoint->media.mpa) {
		return;
	}

	page_mode = ast_channel_page_mode_get(session->channel);
	one_way_page = pbx_builtin_getvar_helper(session->channel, "ONE_WAY_PAGING");

	if (page_mode != AST_PAGE_MODE_ANNOUNCE && ast_strlen_zero(one_way_page)) {
		ast_log(LOG_DEBUG, "[session:%p] It's not one way page, don't need add MPA codec to SDP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/* MPA PT is less than AST_RTP_PT_LAST_STATIC, don't need add if use compact form */
	if (pjsip_use_compact_form) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (!session_media->mpa) {
		return;
	}

	if ((attr = generate_rtpmap_attr(session, media, pool, 14, 0, NULL, AST_RTP_MPA))) {
		media->attr[media->attr_count++] = attr;
		ast_log(LOG_DEBUG, "[session:%p] Add MPA codec to SDP, endpoint:%s\n", session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}


static void remove_connection_attr_for_gswave(struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	if (!session || !media) {
		return;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) {
		return;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return;
	}

	pjmedia_sdp_media_remove_all_attr(media, "connection");

	return;
}

/* Add by bfdu for bug 105429, 2017.3.22 */
static void change_rtcp_to_external_address(pj_pool_t *pool, struct pjmedia_sdp_media *stream, const char *new_address,
	struct ast_sip_transport_state *transport_state)
{
	int i = 0;
	unsigned int port = 0;
	char old_address[128] = {0};
	char str_address[64] = {0};
	char attr_name[16]   = {0};
	char attr_value[128] = {0};
	char addr_type[8] = {0};
	struct ast_sockaddr our_sdp_addr = { { 0, } };

	if (!pool || !stream || !new_address || !transport_state) {
		ast_log(LOG_DEBUG, "change_rtcp_to_external_address param is NULL\n");
		return;
	}

	for (i = 0; i < stream->attr_count; i++) {
		memset(attr_name, 0, sizeof(attr_name));
		memset(attr_value, 0, sizeof(attr_value));
		ast_copy_pj_str(attr_name, &stream->attr[i]->name, sizeof(attr_name));
		ast_copy_pj_str(attr_value, &stream->attr[i]->value, sizeof(attr_value));
		if (strcasecmp(attr_name, "rtcp")) {
			continue;
		}

		if (sscanf(attr_value, "%30u IN %7s %127s", &port, addr_type, old_address) != 3) {
			continue;
		}

		if (strcasecmp(addr_type, "IP4") && strcasecmp(addr_type, "IP6")) {
			ast_log(LOG_DEBUG, "Address type is not IP4 or IP6\n");
			continue;
		}

		ast_sockaddr_parse(&our_sdp_addr, old_address, PARSE_PORT_FORBID);

		if (ast_sip_transport_is_nonlocal(transport_state, &our_sdp_addr) && transport_state->localnet) {
			ast_log(LOG_DEBUG, "change_rtcp_to_external_address, it's already no local IP\n");
			break;
		}

		snprintf(str_address, sizeof(str_address) - 1 , "%d IN %s %s", port, addr_type, new_address);

		ast_log(LOG_DEBUG, "Get init rtcp port:%d, address:%s, and change to new address:%s\n", port, old_address, str_address);

		pj_strdup2(pool, &stream->attr[i]->value, str_address);

		break;
	}

	return;
}

/*! \brief Function which processes ssrc attributes in a stream */
static int process_rtx_ssrc_group_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream, struct ast_stream *asterisk_stream)
{
	int index = 0;
	pjmedia_sdp_attr *attr = NULL;
	unsigned int find = 0;
	unsigned int ssrc = 0;
	unsigned int rtx_ssrc = 0;
	unsigned int previous_remote_ssrc = 0;
	char ssrc_group[MAX_SSRC_ATTR_LEN] = {0};

	/* Set reset flag first, if has remote SSRC, will clear this flag */
	previous_remote_ssrc = ast_rtp_instance_get_remote_ssrc(session_media->rtp);
	if (previous_remote_ssrc) {
		ast_rtp_instance_set_ssrc_reset_flag(session_media->rtp, 1);
		ast_rtp_instance_set_remote_ssrc(session_media->rtp, 0);
		ast_rtp_instance_set_remote_rtx_ssrc(session_media->rtp, 0);
	}

	for (index = 0; index < remote_stream->attr_count; ++index) {
		attr = remote_stream->attr[index];
		if (!attr) {
			continue;
		}

		if (pj_strcmp2(&attr->name, "ssrc-group")) {
			continue;
		}

		memset(ssrc_group, 0, sizeof(ssrc_group));
		ast_copy_pj_str(ssrc_group, &attr->value, sizeof(ssrc_group));

		if (sscanf(ssrc_group, "FID %30u %30u", &ssrc, &rtx_ssrc) != 2) {
			continue;
		}

		find = 1;
		break;
	}

	if (!find) {
		ast_log(LOG_WARNING, "[session:%p] Has no ssrc-group, we consider only has ssrc, may has abnormal, RTX can't take effect, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		process_ssrc_attributes(session, session_media, remote_stream, asterisk_stream);
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_rtx_ssrc_group_attributes ssrc:%u, rtx_ssrc:%u, endpoint:%s\n",
		session, ssrc, rtx_ssrc, ast_sorcery_object_get_id(session->endpoint));

	if (pjmedia_sdp_neg_get_state(session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER &&
		(session->inv_session->state == PJSIP_INV_STATE_CONFIRMED)) {
		if (previous_remote_ssrc != ssrc) {
			ast_log(LOG_DEBUG, "[session:%p] Media:%s SSRC change from %u to %u for endpoint:%s\n",
				session, ast_codec_media_sub_type2str(session_media->sub_type), previous_remote_ssrc, ssrc, ast_sorcery_object_get_id(session->endpoint));
			session_media->changed = 1;
		}
	}

	ast_rtp_instance_set_remote_ssrc(session_media->rtp, ssrc);
	ast_rtp_instance_set_remote_rtx_ssrc(session_media->rtp, rtx_ssrc);

	ast_rtp_instance_set_ssrc_reset_flag(session_media->rtp, 0);

	for (index = 0; index < remote_stream->attr_count; ++index) {
		pjmedia_sdp_attr *attr = remote_stream->attr[index];
		char attr_value[pj_strlen(&attr->value) + 1];
		char *ssrc_attribute_name, *ssrc_attribute_value = NULL;

		/* We only care about ssrc attributes */
		if (pj_strcmp2(&attr->name, "ssrc")) {
			continue;
		}

		ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

		if ((ssrc_attribute_name = strchr(attr_value, ' '))) {
			/* This has an actual attribute */
			*ssrc_attribute_name++ = '\0';
			ssrc_attribute_value = strchr(ssrc_attribute_name, ':');
			if (ssrc_attribute_value) {
				/* Values are actually optional according to the spec */
				*ssrc_attribute_value++ = '\0';
			}

			if (strstr(ssrc_attribute_name, "msid")) {
				ast_log(LOG_DEBUG, "[session:%p] Process ssrc msid attribute name:%s, value:%s, endpoint:%s\n",
					session, ssrc_attribute_name, ssrc_attribute_value, ast_sorcery_object_get_id(session->endpoint));
				strncpy(session_media->remote_msid, ssrc_attribute_value, sizeof(session_media->remote_msid) - 1);
			}

			if (strstr(ssrc_attribute_name, "cname")) {
				ast_log(LOG_DEBUG, "[session:%p] Process ssrc cname attribute name:%s, value:%s, endpoint:%s\n",
					session, ssrc_attribute_name, ssrc_attribute_value, ast_sorcery_object_get_id(session->endpoint));
				strncpy(session_media->remote_cname, ssrc_attribute_value, sizeof(session_media->remote_cname) - 1);
			}
		}
	}

	return 0;
}

static int process_fec_ssrc_group_attributes(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				   const struct pjmedia_sdp_media *remote_stream, struct ast_stream *asterisk_stream)
{
	int index = 0;
	pjmedia_sdp_attr *attr = NULL;
	unsigned int find = 0;
	unsigned int ssrc = 0;
	unsigned int fec_ssrc = 0;
	char ssrc_group[MAX_SSRC_ATTR_LEN] = {0};

	for (index = 0; index < remote_stream->attr_count; ++index) {
		attr = remote_stream->attr[index];
		if (!attr) {
			continue;
		}

		if (pj_strcmp2(&attr->name, "ssrc-group")) {
			continue;
		}

		memset(ssrc_group, 0, sizeof(ssrc_group));
		ast_copy_pj_str(ssrc_group, &attr->value, sizeof(ssrc_group));

		if (sscanf(ssrc_group, "FEC-FR %30u %30u", &ssrc, &fec_ssrc) != 2) {
			continue;
		}

		find = 1;
		break;
	}

	if (!find) {
		ast_log(LOG_DEBUG, "[session:%p] Has no FEC ssrc-group, we will not set FEC SSRC to AVS, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] process_fec_ssrc_group_attributes ssrc:%u, fec_ssrc:%u, endpoint:%s\n",
		session, ssrc, fec_ssrc, ast_sorcery_object_get_id(session->endpoint));

	ast_rtp_instance_set_remote_fec_ssrc(session_media->rtp, fec_ssrc);

	return 0;
}

static void set_gvc_msid_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream)
{
	pjmedia_sdp_attr *attr = NULL;

	if (!(attr = pjmedia_sdp_media_find_attr2(remote_stream, "msid", NULL))) {
		return;
	}

	char attr_value[pj_strlen(&attr->value) + 1];
	ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));
	ast_log(LOG_DEBUG, "[session:%p] set_gvc_msid_for_stream get msid value:%s for endpoint:%s\n", 
		session, attr_value, ast_sorcery_object_get_id(session->endpoint));
	strncpy(session_media->remote_msid, attr_value, sizeof(session_media->remote_msid) - 1);

	return;
}

static void set_mid_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream)
{
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !session_media || !remote_stream) {
		return;
	}


	if (ast_sip_session_media_use_bundle(session)) {
		/* If endpoint enable bundle will add mid later */
		return;
	}

	if (!(attr = pjmedia_sdp_media_find_attr2(remote_stream, "mid", NULL))) {
		return;
	}

	if (!session_media->mid) {
		session_media->mid = ast_calloc(1, MAX_MID_LEN);
		if (!session_media->mid) {
			return;
		}
	}

	char attr_value[pj_strlen(&attr->value) + 1];
	ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));
	ast_log(LOG_DEBUG, "[session:%p] set_mid_for_stream get mid value:%s for endpoint:%s\n", 
		session, attr_value, ast_sorcery_object_get_id(session->endpoint));
	strncpy(session_media->mid, attr_value, MAX_MID_LEN - 1);

	return;
}

static unsigned int judge_if_has_session_level_attr(const pjmedia_sdp_session *sdp, const char *name, const char *val)
{
	unsigned int res = 0;
	unsigned int index = 0;
	pjmedia_sdp_attr *attr = NULL;

	if (!sdp || !name) {
		return 0;
	}

	for (index = 0; index < sdp->attr_count; ++index) {
		attr = sdp->attr[index];
		if (!attr) {
			continue;
		}

		if (pj_strcmp2(&attr->name, name)) {
			continue;
		}

		if (!val) {
			/* If don't need compare value, just return true */
			res = 1;
			break;
		}

		if (!pj_strcmp2(&attr->value, val)) {
			res = 1;
			break;
		}
	}

	return res;
}

static void set_trickle_ice_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream, const struct pjmedia_sdp_session *remote)
{
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !session_media || !remote_stream) {
		return;
	}

	if (!session->endpoint->trickle_ice) {
		return;
	}

	if (!(attr = pjmedia_sdp_media_find_attr2(remote_stream, "ice-options", NULL))) {
		if (judge_if_has_session_level_attr(remote, "ice-options", "trickle")) {
			session_media->trickle_ice = 1;
			return;
		}
		return;
	}

	char attr_value[pj_strlen(&attr->value) + 1];
	ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

	ast_log(LOG_DEBUG, "[session:%p] set_trickle_ice_for_stream get ice options value:%s for endpoint:%s\n",
		session, attr_value, ast_sorcery_object_get_id(session->endpoint));

	if (strstr(attr_value, "trickle")) {
		session_media->trickle_ice = 1;
	}

	return;
}

static void set_rtcp_rsize_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	const struct pjmedia_sdp_media *remote_stream)
{
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !session_media || !remote_stream) {
		return;
	}

	if (!(attr = pjmedia_sdp_media_find_attr2(remote_stream, "rtcp-rsize", NULL))) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Get rtcp-rsize attr for media:%s, endpoint:%s\n",
		session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));

	session_media->rtcp_rsize = 1;

	return;
}

static void add_ssrc_cname_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, enum ast_ssrc_type ssrc_type)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;
	char tmp[128] = {0};
	unsigned int ssrc = 0;

	switch (ssrc_type) {
		case AST_SSRC_TYPE_MEDIA:
			ssrc = ast_rtp_instance_get_ssrc(session_media->rtp);
			break;
		case AST_SSRC_TYPE_RTX:
			ssrc = ast_rtp_instance_get_local_rtx_ssrc(session_media->rtp);
			break;
		case AST_SSRC_TYPE_FEC:
			ssrc = ast_rtp_instance_get_local_fec_ssrc(session_media->rtp);
			break;
		default:
			ssrc = ast_rtp_instance_get_ssrc(session_media->rtp);
			break;
	}

	snprintf(tmp, sizeof(tmp), "%u cname:%s", ssrc, ast_rtp_instance_get_cname(session_media->rtp));
	attr = pjmedia_sdp_attr_create(pool, "ssrc", pj_cstr(&stmp, tmp));
	media->attr[media->attr_count++] = attr;

	return;
}

static void add_ssrc_msid_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream, enum ast_ssrc_type ssrc_type)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;
	char msid[(AST_UUID_STR_LEN * 2) + 24] = {0};
	unsigned int ssrc = 0;

	if (ast_strlen_zero(session_media->mslabel)) {
		/* If this stream is grouped with another then use its media stream label if possible */
		if (ast_stream_get_group(stream) != -1) {
			struct ast_sip_session_media *group_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, ast_stream_get_group(stream));

			ast_copy_string(session_media->mslabel, group_session_media->mslabel, sizeof(session_media->mslabel));
		}

		if (ast_strlen_zero(session_media->mslabel)) {
			ast_uuid_generate_str(session_media->mslabel, sizeof(session_media->mslabel));
		}
	}

	if (ast_strlen_zero(session_media->label)) {
		ast_uuid_generate_str(session_media->label, sizeof(session_media->label));
		/* add for stream identification to replace stream_name */
		ast_stream_set_metadata(stream, "MSID:LABEL", session_media->label);
	}

	switch (ssrc_type) {
		case AST_SSRC_TYPE_MEDIA:
			ssrc = ast_rtp_instance_get_ssrc(session_media->rtp);
			break;
		case AST_SSRC_TYPE_RTX:
			ssrc = ast_rtp_instance_get_local_rtx_ssrc(session_media->rtp);
			break;
		case AST_SSRC_TYPE_FEC:
			ssrc = ast_rtp_instance_get_local_fec_ssrc(session_media->rtp);
			break;
		default:
			ssrc = ast_rtp_instance_get_ssrc(session_media->rtp);
			break;
	}

	ast_stream_set_metadata(stream, "MSID:MSLABEL", session_media->mslabel);
	ast_stream_set_metadata(stream, "MSID:LABEL", session_media->label);

	snprintf(msid, sizeof(msid), "%u msid:%s %s", ssrc, session_media->mslabel, session_media->label);
	ast_log(LOG_DEBUG, "[session:%p] Add stream msid:%s %s for endpoint:%s\n", 
		 session, session_media->mslabel, session_media->label, ast_sorcery_object_get_id(session->endpoint));
	attr = pjmedia_sdp_attr_create(pool, "ssrc", pj_cstr(&stmp, msid));
	pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);

	return;
}

static void add_rtx_ssrc_group_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	char tmp[128];

	snprintf(tmp, sizeof(tmp), "FID %u %u", ast_rtp_instance_get_ssrc(session_media->rtp), ast_rtp_instance_get_local_rtx_ssrc(session_media->rtp));
	attr = pjmedia_sdp_attr_create(pool, "ssrc-group", pj_cstr(&stmp, tmp));
	media->attr[media->attr_count++] = attr;

	add_ssrc_cname_to_stream(session, session_media, pool, media, AST_SSRC_TYPE_MEDIA);
	add_ssrc_msid_to_stream(session, session_media, pool, media, stream, AST_SSRC_TYPE_MEDIA);

	add_ssrc_cname_to_stream(session, session_media, pool, media, AST_SSRC_TYPE_RTX);
	add_ssrc_msid_to_stream(session, session_media, pool, media, stream, AST_SSRC_TYPE_RTX);

	return;
}

static void add_fec_ssrc_group_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	char tmp[128];

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && !session->support_gs_fec) {
		ast_log(LOG_DEBUG, "[session:%p] Codec is GS-FEC, webrtc can't support, don't add FEC SSRC group, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(tmp, sizeof(tmp), "FEC-FR %u %u", ast_rtp_instance_get_ssrc(session_media->rtp), ast_rtp_instance_get_local_fec_ssrc(session_media->rtp));
	attr = pjmedia_sdp_attr_create(pool, "ssrc-group", pj_cstr(&stmp, tmp));
	media->attr[media->attr_count++] = attr;

	add_ssrc_cname_to_stream(session, session_media, pool, media, AST_SSRC_TYPE_FEC);
	add_ssrc_msid_to_stream(session, session_media, pool, media, stream, AST_SSRC_TYPE_FEC);

	return;
}

static int create_udp_bfcp(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	/* If we've already initialized bfcp, don't take any further action */
	if (session_media->bfcp_udp) {
		ast_log(LOG_WARNING, "[session:%p] BFCP has already initialized, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if ((session_media->bfcp_udp = ast_bfcp_udp_new_with_bindaddr("UDP", sched, NULL, &address_rtp, "asterisk", session_media->remote_address, session->media_turn_relay))) {
		/* todo: set bfcp qos use video now, need add a bfcp qos config */
		ast_bfcp_setqos(session_media->bfcp_udp, session->endpoint->media.tos_video,
				session->endpoint->media.cos_video);
		ast_log(LOG_DEBUG, "[session:%p] Setting NAT on BFCP to %s for endpoint:%s\n",
			session, session->endpoint->media.rtp.symmetric ? "On" : "Off", ast_sorcery_object_get_id(session->endpoint));
		ast_bfcp_setnat(session_media->bfcp_udp, session->endpoint->media.rtp.symmetric);
	} else {
		ast_log(LOG_WARNING, "[session:%p] BFCP creation failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	return 0;
}

static int create_tcp_bfcp(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	/* If we've already initialized bfcp, don't take any further action */
	if (session_media->bfcp_tcp) {
		ast_log(LOG_WARNING, "[session:%p] BFCP tcp has already initialized for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!(session_media->bfcp_tcp = ast_bfcp_tcp_new_with_bindaddr(&address_rtp))) {
		ast_log(LOG_WARNING, "[session:%p] bfcp tcp creation failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	return 0;
}

static const char *bfcp_sdp_floorctl_2_str(enum bfcp_sdp_floorctl floorctl)
{
	return bfcp_sdp_floorctl_strmap[floorctl][0];
}

static enum bfcp_sdp_floorctl bfcp_str_2_sdp_floorctl(const char *floorctl)
{
    int i = 0;

    for (i = 0; i < BFCP_FLOORCTL_MAX; i++)
    {
        if (!strcmp(floorctl, bfcp_sdp_floorctl_strmap[i][0]))
        {
            return (enum bfcp_sdp_floorctl)i;
        }
    }

    return 0;
}

static const char *bfcp_sdp_setup_2_str(enum bfcp_sdp_setup setup)
{
	return bfcp_sdp_setup_strmap[setup][0];
}

static enum bfcp_sdp_setup bfcp_str_2_sdp_setup(const char *floorctl)
{
    int i = 0;

    for (i = 0; i < BFCP_SETUP_MAX; i++)
    {
        if (!strcmp(floorctl, bfcp_sdp_setup_strmap[i][0]))
        {
            return (enum bfcp_sdp_setup)i;
        }
    }

    return 0;
}

static const char* bfcp_sdp_neg_floorctl_attr(const char *floorctl_str)
{
	enum bfcp_sdp_floorctl floorctl;
	enum bfcp_sdp_floorctl neg_res;

	if (!floorctl_str) {
		return NULL;
	}

	floorctl = bfcp_str_2_sdp_floorctl(floorctl_str);

	switch (floorctl) {
		case BFCP_FLOORCTL_C_ONLY:
			neg_res = BFCP_FLOORCTL_S_ONLY;
			break;
		case BFCP_FLOORCTL_S_ONLY:
			neg_res = BFCP_FLOORCTL_C_ONLY;
			break;
		case BFCP_FLOORCTL_C_S:
			neg_res = BFCP_FLOORCTL_S_ONLY;
			break;
		default:
			return NULL;
	}

	return bfcp_sdp_floorctl_2_str(neg_res);
}

static const char* bfcp_sdp_neg_setup_attr(const char *setup_str)
{
	enum bfcp_sdp_setup setup;
	enum bfcp_sdp_setup neg_res;

	if (!setup_str) {
		return NULL;
	}

	setup = bfcp_str_2_sdp_setup(setup_str);

	switch (setup) {
		case BFCP_SETUP_ACTIVE:
			neg_res = BFCP_SETUP_PASSIVE;
			break;
		case BFCP_SETUP_PASSIVE:
			neg_res = BFCP_SETUP_ACTIVE;
			break;
		case BFCP_SETUP_ACTPASS:
			neg_res = BFCP_SETUP_PASSIVE;
			break;
		default:
			return NULL;
	}

	return bfcp_sdp_setup_2_str(neg_res);
}

/* begin add by bfdu for bfcp, Grandstream Networks, 2015.01.19 */
static void process_sdp_a_app_bfcp(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr;
	int confid  = 0;
	int userid  = 0;

	if ((attr = pjmedia_sdp_media_find_attr2(stream, "confid", NULL))) {
		confid = pj_strtoul(pj_strltrim(&attr->value));
		if (confid) {
			session->bfcp_para.confid = confid;
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "userid", NULL))) {
		userid = pj_strtoul(pj_strltrim(&attr->value));
		if (userid) {
			session->bfcp_para.userid = userid;
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "floorctrl", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.remote_floorctrl, pj_strltrim(&attr->value), sizeof(session->bfcp_para.remote_floorctrl));
			ast_log(LOG_DEBUG, "[session:%p] Get BFCP remote floorctrl:%s for endpoint:%s\n",
				session, session->bfcp_para.remote_floorctrl, ast_sorcery_object_get_id(session->endpoint));
			const char *local_floorctrl = bfcp_sdp_neg_floorctl_attr(session->bfcp_para.remote_floorctrl);
			if (local_floorctrl) {
				strcpy(session->bfcp_para.local_floorctrl, local_floorctrl);
				ast_log(LOG_DEBUG, "[session:%p] After neg BFCP local floorctrl:%s for endpoint:%s\n",
					session, session->bfcp_para.local_floorctrl, ast_sorcery_object_get_id(session->endpoint));
			}
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "setup", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.remote_setup, pj_strltrim(&attr->value), sizeof(session->bfcp_para.remote_setup));
			ast_log(LOG_DEBUG, "[session:%p] Get BFCP remote setup %s for endpoint:%s\n",
				session, session->bfcp_para.remote_setup, ast_sorcery_object_get_id(session->endpoint));
			const char *local_setup = bfcp_sdp_neg_setup_attr(session->bfcp_para.remote_setup);
			if (local_setup) {
				strcpy(session->bfcp_para.local_setup, local_setup);
				ast_log(LOG_DEBUG, "[session:%p] After neg BFCP local setup:%s, endpoint:%s\n",
					session, session->bfcp_para.local_setup, ast_sorcery_object_get_id(session->endpoint));
			}
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "connection", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.connection, pj_strltrim(&attr->value), sizeof(session->bfcp_para.connection));
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "floorid", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.floorid, pj_strltrim(&attr->value), sizeof(session->bfcp_para.floorid));
		}
	}
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "fingerprint", NULL))) {
		if (pj_strltrim(&attr->value)) {
			ast_copy_pj_str(session->bfcp_para.fingerprint, pj_strltrim(&attr->value), sizeof(session->bfcp_para.fingerprint));
		}
	}

	return;
}

static void copy_para_to_bfcp_info(struct ast_sip_session *session, struct ast_bfcp_udp *bfcp)
{
	if (!session || !bfcp) {
		return;
	}

	bfcp->bfcp_info.conf_id = session->bfcp_para.confid;
	bfcp->bfcp_info.user_id = session->bfcp_para.userid;
	bfcp->bfcp_info.floor_id = atoi(session->bfcp_para.floorid);
	bfcp->bfcp_info.floor_requst_id = bfcp->bfcp_info.floor_id;
	if (session->endpoint && session->endpoint->aors) {
		strcpy(bfcp->bfcp_info.account_id, session->endpoint->aors);
	}

	return;
}

static void copy_para_to_bfcp_tcp_info(struct ast_sip_session *session, struct ast_bfcp_tcp *bfcp_tcp)
{
	if (!session || !bfcp_tcp) {
		return;
	}

	bfcp_tcp->bfcp_info.conf_id = session->bfcp_para.confid;
	bfcp_tcp->bfcp_info.user_id = session->bfcp_para.userid;
	bfcp_tcp->bfcp_info.floor_id = atoi(session->bfcp_para.floorid);
	bfcp_tcp->bfcp_info.floor_requst_id = bfcp_tcp->bfcp_info.floor_id;
	if (session->endpoint && session->endpoint->aors) {
		strcpy(bfcp_tcp->bfcp_info.account_id, session->endpoint->aors);
	}

	return;
}

static void set_bfcp_udp_conference_params(struct ast_sip_session *session, struct ast_bfcp_udp *bfcp)
{
	if (!session || !session->channel || !bfcp) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!ast_strlen_zero(session->conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] set_bfcp_udp_conference_params conference num:%s for endpoint:%s\n",
			session, session->conf_room, ast_sorcery_object_get_id(session->endpoint));
		strncpy(bfcp->bfcp_info.info_conf_id, session->conf_room, sizeof(bfcp->bfcp_info.info_conf_id) - 1);
	}

	return;
}

static void set_bfcp_tcp_conference_params(struct ast_sip_session *session, struct ast_bfcp_tcp *bfcp)
{
	if (!session || !session->channel || !bfcp) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!ast_strlen_zero(session->conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] set_bfcp_tcp_conference_params conference num:%s for endpoint:%s\n",
			session, session->conf_room, ast_sorcery_object_get_id(session->endpoint));
		strncpy(bfcp->bfcp_info.info_conf_id, session->conf_room, sizeof(bfcp->bfcp_info.info_conf_id) - 1);
	}

	return;
}

static void add_bfcp_param_to_sdp(pj_pool_t *pool, struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	char bfcp_para_str[512] = {0};
	pj_str_t bfcp_para_pj_str;
	pjmedia_sdp_attr *attr = NULL;

	if (!pool || !session || !media) {
		ast_log(LOG_WARNING, "add_bfcp_param_to_sdp param is NULL\n");
		return;
	}

	/* add bfcp attr */
	if (session->bfcp_para.confid) {
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%d", session->bfcp_para.confid);
		attr = pjmedia_sdp_attr_create(pool, "confid", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}
	if (session->bfcp_para.userid) {
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%d", session->bfcp_para.userid);
		attr = pjmedia_sdp_attr_create(pool, "userid", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}
	if (strlen(session->bfcp_para.floorid)) {
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%s", session->bfcp_para.floorid);
		attr = pjmedia_sdp_attr_create(pool, "floorid", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}
	if (strlen(session->bfcp_para.local_floorctrl)) {
		ast_log(LOG_DEBUG, "[session:%p] Creat bfcp floorctrl %s for endpoint:%s\n",
			session, session->bfcp_para.local_floorctrl, ast_sorcery_object_get_id(session->endpoint));
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%s", session->bfcp_para.local_floorctrl);
		attr = pjmedia_sdp_attr_create(pool, "floorctrl", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}
	if (strlen(session->bfcp_para.local_setup)) {
		ast_log(LOG_DEBUG, "[session:%p] Creat bfcp setup %s for endpoint:%s\n",
			session, session->bfcp_para.local_setup, ast_sorcery_object_get_id(session->endpoint));
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%s", session->bfcp_para.local_setup);
		attr = pjmedia_sdp_attr_create(pool, "setup", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}
	if (strlen(session->bfcp_para.connection)) {
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%s", session->bfcp_para.connection);
		attr = pjmedia_sdp_attr_create(pool, "connection", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}
	if (strlen(session->bfcp_para.fingerprint)) {
		snprintf(bfcp_para_str, sizeof(bfcp_para_str), "%s", session->bfcp_para.fingerprint);
		attr = pjmedia_sdp_attr_create(pool, "fingerprint", pj_cstr(&bfcp_para_pj_str, bfcp_para_str));
		media->attr[media->attr_count++] = attr;
		memset(bfcp_para_str, 0, sizeof(bfcp_para_str));
	}

	return;
}

static int negotiate_incoming_app_bfcp_udp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);

	ast_log(LOG_DEBUG, "[session:%p] Negotiate incoming bfcp udp sdp stream for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "Media stream bfcp is already declined\n");
		return 0;
	}

	if (!session->endpoint->media.bfcp_support) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable BFCP media support, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!ast_format_cap_has_type(session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO)) {
		ast_debug(3, "Endpoint has no codecs for media type video, declining udp bfcp stream\n");
		return 0;
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);

	if (session->media_turn_relay) {
		strncpy(session_media->remote_address, host, sizeof(session_media->remote_address) - 1);
	}

	/* Using the connection information create an appropriate RTP instance */
	if (!session_media->bfcp_udp && create_udp_bfcp(session, session_media)) {
		return -1;
	}

	/* for bfcp, just use recieved transport */
	pj_strdup(session->inv_session->pool, &session_media->transport, &stream->desc.transport);

	process_sdp_a_app_bfcp(session, stream);

	copy_para_to_bfcp_info(session, session_media->bfcp_udp);

	session->bfcp_para.has_udp_bfcp = 1;

	return 1;
}

static int negotiate_incoming_app_bfcp_tcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	char host[NI_MAXHOST];
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);

	ast_log(LOG_DEBUG, "[session:%p] Negotiate incoming bfcp tcp sdp stream for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	/* If port is 0, ignore this media stream */
	if (!stream->desc.port) {
		ast_debug(3, "[session:%p] Media stream bfcp is already declined, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!session->endpoint->media.bfcp_support) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable BFCP media support, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!ast_format_cap_has_type(session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO)) {
		ast_debug(3, "[session:%p] Endpoint has no codecs for media type video, declining tcp bfcp stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_copy_pj_str(host, stream->conn ? &stream->conn->addr : &sdp->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		return -1;
	}

	ast_sockaddr_set_port(addrs, stream->desc.port);

	/* Using the connection information create an appropriate RTP instance */
	if (!session_media->bfcp_tcp && create_tcp_bfcp(session, session_media)) {
		return -1;
	}

	/* for bfcp, just use recieved transport */
	pj_strdup(session->inv_session->pool, &session_media->transport, &stream->desc.transport);

	process_sdp_a_app_bfcp(session, stream);

	copy_para_to_bfcp_tcp_info(session, session_media->bfcp_tcp);

	session->bfcp_para.has_tcp_bfcp = 1;

	return 1;
}

static int negotiate_incoming_application_bfcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, const struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_media *stream)
{
	int res = 0;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (NULL == session || NULL == session_media || NULL == sdp || NULL == stream) {
		ast_log(LOG_ERROR, "Negotiate application sdp, param is invalid\n");
		return -1;
	}

	sub_type = session_media->sub_type;
	switch (sub_type) {
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP:
			res = negotiate_incoming_app_bfcp_udp_sdp_stream(session, session_media, sdp, stream);
			break;
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP:
			res = negotiate_incoming_app_bfcp_tcp_sdp_stream(session, session_media, sdp, stream);
			break;
		default:
			ast_log(LOG_ERROR, "[session:%p] Can't process other application media, ignore this for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
	}

	return res;
}

static void add_framerate_to_steam(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr = NULL;
	
	const char *framerate = ast_stream_get_metadata(stream, "FRAMERATE");
	if (!ast_strlen_zero(framerate)) {
		snprintf(tmp, sizeof(tmp), "%s", framerate);
		attr = pjmedia_sdp_attr_create(pool, "framerate", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	return;
}

static void add_content_to_steam(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr = NULL;
	
	const char *content = ast_stream_get_metadata(stream, "CONTENT");
	if (!ast_strlen_zero(content)) {
		snprintf(tmp, sizeof(tmp), "%s", content);
		attr = pjmedia_sdp_attr_create(pool, "content", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	return;
}

static void add_label_to_steam(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr = NULL;
	
	const char *label = ast_stream_get_metadata(stream, "LABEL");
	if (!ast_strlen_zero(label)) {
		snprintf(tmp, sizeof(tmp), "%s", label);
		attr = pjmedia_sdp_attr_create(pool, "label", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	return;
}

static void add_imageattr_to_stream(pj_pool_t *pool, int rtp_code, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	char tmp[512] = {0};
	pjmedia_sdp_attr *attr = NULL;

	const char *imageattr = ast_stream_get_metadata(stream, "IMAGEATTR");
	if (!ast_strlen_zero(imageattr)) {
		snprintf(tmp, sizeof(tmp), "%d %s", rtp_code, imageattr);
		attr = pjmedia_sdp_attr_create(pool, "imageattr", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	return;
}

static void add_video_extra_attr_to_stream(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	if (!pool || !media || !stream) {
		return;
	}

	add_framerate_to_steam(pool, media, stream);
	add_content_to_steam(pool, media, stream);
	add_label_to_steam(pool, media, stream);

	return;
}

static void add_rtcp_address_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	char tmp[256] = {0};
	pj_str_t stmp;
	uint16_t port = 0;
	const char *hostip = NULL;
	char net_type[8] = { 0 };
	pj_str_t	 pj_addr;
	pj_sockaddr ip;
	pjmedia_sdp_attr *attr = NULL;

	if (session_media->remote_rtcp_mux && session->endpoint->media.rtcp_mux) {
		return;
	}

	if (!ast_stream_get_has_rtcp_address(stream) && !session->media_turn_relay) {
		return;
	}

	hostip = get_media_host_ip(session, session_media, media, 0);
	if (ast_strlen_zero(hostip)) {
		ast_log(LOG_ERROR, "[sesssion:%p] No local host IP available for stream %s, don't add rtcp address for endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	strcpy(net_type, "IP4");
	pj_strdup2(pool, &pj_addr, hostip);
	if ((pj_sockaddr_parse(pj_AF_UNSPEC(), 0, &pj_addr, &ip) == PJ_SUCCESS) && (ip.addr.sa_family == pj_AF_INET6())) {
		strcpy(net_type, "IP6");
	}

	/* create rtcp attr */
	port = ast_rtp_instance_get_local_rtcp(session_media->rtp);
	if (!port) {
		ast_log(LOG_ERROR, "[sesssion:%p] Has no local rtcp port for stream %s, don't add rtcp address for endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	snprintf(tmp, sizeof(tmp), "%d IN %s %s", port, net_type, hostip);
	attr = pjmedia_sdp_attr_create(pool, "rtcp", pj_cstr(&stmp, tmp));
	media->attr[media->attr_count++] = attr;

	ast_log(LOG_DEBUG, "[session:%p] add_rtcp_address_to_stream host:%s port:%d for media type:%s, endpoint:%s\n",
		session, hostip, port, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void add_rtcp_nack_attrs_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	char tmp[256] = {0};
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;

	if (!session->endpoint->enable_nack) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has not enable nack, clear nack flag\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_has_nack(stream, 0);
	}

	if (ast_stream_get_has_nack(stream)) {
		memset(tmp, 0, sizeof(tmp));
		snprintf(tmp, sizeof(tmp), "%s", "* nack");
		attr = pjmedia_sdp_attr_create(pool, "rtcp-fb", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	if (ast_stream_get_has_nack_pli(stream)) {
		memset(tmp, 0, sizeof(tmp));
		snprintf(tmp, sizeof(tmp), "%s", "* nack pli");
		attr = pjmedia_sdp_attr_create(pool, "rtcp-fb", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	if (ast_stream_get_has_ccm_fir(stream)) {
		memset(tmp, 0, sizeof(tmp));
		snprintf(tmp, sizeof(tmp), "%s", "* ccm fir");
		attr = pjmedia_sdp_attr_create(pool, "rtcp-fb", pj_cstr(&stmp, tmp));
		media->attr[media->attr_count++] = attr;
	}

	return;
}

static void add_rtcp_attrs_to_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	if (check_if_direct_media_enabled(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Enable direct media, don't need create rtcp attr for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	add_rtcp_address_to_stream(session, session_media, pool, media, stream);

	add_rtcp_nack_attrs_to_stream(session, session_media, pool, media, stream);

	add_rtcp_rsize_to_stream(session, session_media, pool, media);

	return;
}

static void add_negotiate_gs_fec_version_to_stream(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;

	if (!pool || !media || !stream) {
		return;
	}

	if (strlen(ast_stream_get_gs_fec_version(stream)) == 0) {
		/* If we can't find fec version, add whatever we support */
		attr = pjmedia_sdp_attr_create(pool, "gs-fec-version", pj_cstr(&stmp, "2"));
		media->attr[media->attr_count++] = attr;

		attr = pjmedia_sdp_attr_create(pool, "gs-fec-version", pj_cstr(&stmp, "1"));
		media->attr[media->attr_count++] = attr;
	} else {
		attr = pjmedia_sdp_attr_create(pool, "gs-fec-version", pj_cstr(&stmp, ast_stream_get_gs_fec_version(stream)));
		media->attr[media->attr_count++] = attr;
	}

	return;
}

/* Now we only support gs-fec-version 2 and 1, we add both 2 and 1 for init outgoing call */
static void add_init_gs_fec_version_to_stream(pj_pool_t *pool, pjmedia_sdp_media *media, struct ast_stream *stream)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;

	if (!pool || !media || !stream) {
		return;
	}

	attr = pjmedia_sdp_attr_create(pool, "gs-fec-version", pj_cstr(&stmp, "2"));
	media->attr[media->attr_count++] = attr;

	attr = pjmedia_sdp_attr_create(pool, "gs-fec-version", pj_cstr(&stmp, "1"));
	media->attr[media->attr_count++] = attr;

	return;
}


static void add_gs_fec_version_to_stream(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media,
	struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	if (!session || !pool || !media || !session_media || !stream) {
		return;
	}

	if (!session_media->alone_fec_version && !ast_stream_get_alone_fec_version(stream)) {
		return;
	}

	if (session->is_outgoing) {
		add_init_gs_fec_version_to_stream(pool, media, stream);
	} else {
		add_negotiate_gs_fec_version_to_stream(pool, media, stream);
	}

	return;
}

static int check_msid_for_sdp_media(struct ast_sip_session_media *session_media, const pjmedia_sdp_media *media)
{
	unsigned int index = 0;
	char msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

	/* If we can't find msid attr, we just pass msid check */

	for (index = 0; index < media->attr_count; ++index) {
		pjmedia_sdp_attr *attr = media->attr[index];
		char attr_value[pj_strlen(&attr->value) + 1];
		char *ssrc_attribute_name, *ssrc_attribute_value = NULL;

		/* We only care about ssrc attributes */
		if (pj_strcmp2(&attr->name, "ssrc")) {
			continue;
		}

		ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

		if ((ssrc_attribute_name = strchr(attr_value, ' '))) {
			*ssrc_attribute_name++ = '\0';
			ssrc_attribute_value = strchr(ssrc_attribute_name, ':');
			if (ssrc_attribute_value) {
				*ssrc_attribute_value++ = '\0';
			}

			if (strstr(ssrc_attribute_name, "msid")) {
				snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
				ast_log(LOG_DEBUG, "check_msid_value value:%s, and session_media msid value:%s\n", ssrc_attribute_value, msid);
				if (strcmp(msid, ssrc_attribute_value)) {
					ast_log(LOG_DEBUG, "msid is not the same\n");
					return FALSE;
				}
			}
		}
	}

	return TRUE;
}

static pjmedia_sdp_media *add_disable_media_as_no_remote(struct ast_sip_session *session, pj_pool_t *pool, struct ast_stream *stream)
{
	pjmedia_sdp_media *media = NULL;
	static const pj_str_t STR_RTP_AVP = { "RTP/AVP", 7 };
	static const pj_str_t STR_DTLS_RTP = { "UDP/TLS/RTP/SAVPF", 17 };
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	const char *hostip = NULL;
	pjmedia_sdp_attr *attr = NULL;

	/* This is actually an offer so put a dummy payload in that is ignored and sane transport */
	media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media));
	if (!media) {
		return NULL;
	}
	if (session->is_webrtc_call) {
		media->desc.transport = STR_DTLS_RTP;
	} else {
		media->desc.transport = STR_RTP_AVP;
	}

	media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn));
	if (!media->conn) {
		return NULL;
	}

	/* Add connection level details */
	if (ast_strlen_zero(session->endpoint->media.address)) {
		hostip = ast_sip_get_host_ip_string(pj_AF_INET());
	} else {
		hostip = session->endpoint->media.address;
	}

	if (ast_strlen_zero(hostip)) {
		ast_log(LOG_ERROR, "No local host IP available for stream\n");
		return NULL;
	}

	media->conn->net_type = STR_IN;
	/* Assume that the connection will use IPv4 until proven otherwise */
	media->conn->addr_type = STR_IP4;
	pj_strdup2(pool, &media->conn->addr, hostip);

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
		media->desc.port = 9;
	} else {
		media->desc.port = 0;
	}

	media->desc.port_count = 1;

	if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO) {
		media->desc.media = pj_str("video");
		/* We use H264, and 99 payload for default */
		if ((attr = generate_rtpmap_attr(session, media, pool, 99, 1, ast_format_h264, 0))) {
			media->attr[media->attr_count++] = attr;
		}
		add_content_to_steam(pool, media, stream);
	} else {
		pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], "32");
	}

	return media;
}

static void remove_all_media_directions(pjmedia_sdp_media *m)
{
	if (!m) {
		ast_log(LOG_WARNING, "Remove all media directions, media is null\n");
		return;
	}

	pjmedia_sdp_media_remove_all_attr(m, "inactive");
	pjmedia_sdp_media_remove_all_attr(m, "sendrecv");
	pjmedia_sdp_media_remove_all_attr(m, "sendonly");
	pjmedia_sdp_media_remove_all_attr(m, "recvonly");

	return;
}

static void change_media_to_inactive(pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *a = NULL;

	if (!pool || !media) {
		ast_log(LOG_WARNING, "Change media direction, param is null\n");
		return;
	}

	a = pjmedia_sdp_attr_create(pool, "inactive", NULL);

	if (a) {
		remove_all_media_directions(media);
		pjmedia_sdp_media_add_attr(media, a);
	} else {
		ast_log(LOG_WARNING, "Create inactive media direction failed\n");
	}

	return;
}

static void create_outgoing_disable_sdp_stream_according_position(pj_pool_t *pool, struct ast_sip_session *session,
	struct pjmedia_sdp_session *sdp, const struct pjmedia_sdp_session *remote, struct ast_stream *stream)
{
	pjmedia_sdp_media *media = NULL;

	if (!session || !sdp || !stream) {
		ast_log(LOG_DEBUG, "create_outgoing_disable_sdp_stream_according_position param is NULL\n");
		return;
	}

	if (!session->inv_session) {
		ast_log(LOG_DEBUG, "create_outgoing_disable_sdp_stream_according_position inv_session is NULL\n");
		return;
	}

	if (!pool) {
		ast_log(LOG_DEBUG, "create_outgoing_disable_sdp_stream_according_position pool is NULL\n");
		return;
	}

	if (remote && remote->media[ast_stream_get_position(stream)]) {
		ast_log(LOG_DEBUG, "[session:%p] create_outgoing_disable_sdp_stream_according_position get stream position for media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		pjmedia_sdp_media *remote_media = remote->media[ast_stream_get_position(stream)];
		if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
			ast_log(LOG_DEBUG, "[session:%p] It's WAVE desktop/web, don't set port to zero, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			media = pjmedia_sdp_media_clone(pool, remote_media);
			change_media_to_inactive(pool, media);
		} else {
			media = pjmedia_sdp_media_clone_deactivate(pool, remote_media);
		}
		ast_sip_session_remove_attrs_for_disable_media(media);
		ast_sip_session_remove_media_from_bundle_group(session, pool, sdp, media);
	} else {
		/* This is actually an offer so put a dummy payload in that is ignored and sane transport */
		ast_log(LOG_DEBUG, "[session:%p] create_outgoing_disable_sdp_stream_according_position no remote stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		media = add_disable_media_as_no_remote(session, pool, stream);
		if (!media) {
			return;
		}
	}

	ast_sip_session_add_decline_video_reason(pool, session, media, stream);

	sdp->media[sdp->media_count++] = media;

	return;
}

static void create_outgoing_disable_sdp_according_msid(pj_pool_t *pool, struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp,
	struct ast_stream *stream, pjmedia_sdp_session *active_local_sdp)
{
	unsigned int i = 0;
	const pjmedia_sdp_media *im;
	pjmedia_sdp_media *res = NULL;
	enum ast_media_sub_type active_media_sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	sub_type = session_media->sub_type;

	ast_log(LOG_DEBUG, "[session:%p] create_outgoing_disable_sdp_according_msid, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	for (i = 0; i < active_local_sdp->media_count; ++i) {
		im = active_local_sdp->media[i];
		if (!im) {
			continue;
		}

		active_media_sub_type = ast_sip_session_get_media_sub_type(session, im);
		if (active_media_sub_type != sub_type) {
			continue;
		}

		if (session->is_sfu &&
			strncasecmp(session->endpoint->aors, "trunk_", 6) &&
			!check_msid_for_sdp_media(session_media, im)) {
			continue;
		}

		session_media->bundle_group = -1;
		session_media->bundled = 0;

		if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
			ast_log(LOG_DEBUG, "[session:%p] It's WAVE desktop/web, don't set port to zero, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			res = pjmedia_sdp_media_clone(pool, im);
			change_media_to_inactive(pool, res);
		} else {
			res = pjmedia_sdp_media_clone_deactivate(pool, im);
		}
		ast_sip_session_remove_attrs_for_disable_media(res);
		ast_sip_session_add_decline_video_reason(pool, session, res, stream);
		sdp->media[sdp->media_count++] = res;
		break;
	}

	return;
}

static void create_outgoing_bundled_mobile_disable_sdp(pj_pool_t *pool, struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp, struct ast_stream *stream)
{
	pjmedia_sdp_session *active_local_sdp = NULL;
	
	active_local_sdp = session->neg_notify_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_DEBUG, "Have no active_local_sdp, don't add disable media\n");
		return;
	}

	create_outgoing_disable_sdp_according_msid(pool, session, session_media, sdp, stream, active_local_sdp);

	return;
}

static int judge_if_create_disable_media_for_bundle_mobile(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!session || !stream) {
		return 0;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		return 0;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return 0;
	}

	if (!session->is_sfu) {
		return 0;
	}

	if (!ast_stream_judge_if_stream_is_participant(stream)) {
		return 0;
	}

	return 1;
}

static void create_outgoing_disable_sdp(pj_pool_t *pool, struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_session *remote, struct ast_stream *stream)
{
	pjmedia_sdp_session *active_local_sdp = NULL;
	const pjmedia_sdp_session *remote_sdp;

	if (!session || !session_media || !sdp) {
		return;
	}

	if (judge_if_create_disable_media_for_bundle_mobile(session, stream)) {
		create_outgoing_bundled_mobile_disable_sdp(pool, session, session_media, sdp, stream);
		return;
	}

	/* SFU we can't distinguish main video and participants video, and gswave will not disorder m line, just use stream position */
	//added neg_remote check by zyxia ,Bug 267515,need check if is outgoing sdp, not only check the is_outgoing ,also need check the neg state pjmedia_sdp_neg_get_neg_remote
	if ((!session->is_outgoing && pjmedia_sdp_neg_get_neg_remote(session->inv_session->neg, &remote_sdp)==PJ_SUCCESS ) ||
		(session->is_webrtc_call && session->is_sfu) ||
		ast_stream_judge_if_stream_is_participant(stream)) {
		create_outgoing_disable_sdp_stream_according_position(pool, session, sdp, remote, stream);
		return;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		if (session->is_webrtc_call) {
			create_outgoing_disable_sdp_stream_according_position(pool, session, sdp, remote, stream);
			return;
		}
		ast_log(LOG_DEBUG, "Have no inv session or neg, don't add disable media\n");
		return;
	}

	if (!pool) {
		ast_log(LOG_DEBUG, "Have no pool, don't add disable media\n");
		return;
	}
	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_DEBUG, "Have no active_local_sdp, don't add disable media\n");
		return;
	}

	create_outgoing_disable_sdp_according_msid(pool, session, session_media, sdp, stream, active_local_sdp);

	return;
}


static int create_outgoing_app_bfcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp)
{
	char fmt_str[8] = {0};
	int direct_media_enabled = 0;
	struct ast_sockaddr bfcpaddr = { {0,} };
	char hostip[PJ_INET6_ADDRSTRLEN + 2] = {0};
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	pj_str_t str_transport_bfcp = { "UDP/BFCP", 8 };
	pj_pool_t *pool = NULL;
	pjmedia_sdp_media *media = NULL;
	pj_sockaddr ip;

	if (!session || !session->inv_session || !session_media || !sdp) {
		ast_log(LOG_DEBUG, "create outgoing application bfcp sdp stream para is null\n");
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] create outgoing application bfcp sdp stream for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	pool = session->inv_session->pool_prov;
	if (!pool) {
		ast_log(LOG_DEBUG, "[session:%p] create outgoing application bfcp sdp stream pool is null for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (!(media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media))) ||
		!(media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn)))) {
		ast_log(LOG_ERROR, "[session:%p] Can't create outgoing bfcp media connection, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (session->media_turn_relay) {
		strncpy(session_media->remote_address, session->turn_target_address, sizeof(session_media->remote_address) - 1);
	}

	if (1 == session->bfcp_para.has_udp_bfcp) {
		pj_cstr(&str_transport_bfcp, "UDP/BFCP");
		if (!session_media->bfcp_udp && create_udp_bfcp(session, session_media)) {
			ast_log(LOG_ERROR, "[session:%p] Create outgoing bfcp sdp stream, create udp error for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}
		ast_bfcp_get_us(session_media->bfcp_udp, &bfcpaddr);
	} else if (1 == session->bfcp_para.has_tcp_bfcp) {
		pj_cstr(&str_transport_bfcp, "TCP/BFCP");
		if (!session_media->bfcp_tcp && create_tcp_bfcp(session, session_media)) {
			ast_log(LOG_ERROR, "[session:%p] Create outgoing bfcp sdp stream, create udp error for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}
		ast_bfcp_tcp_get_us(session_media->bfcp_tcp, &bfcpaddr);
	} else {
		ast_log(LOG_ERROR, "[session:%p] Can't create outgoing bfcp sdp for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	media->desc.media = pj_str("application");
	media->desc.transport = str_transport_bfcp;
	snprintf(fmt_str, sizeof(fmt_str), "%s", "*");
	pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], fmt_str);
	if (session->bfcp_para.has_udp_bfcp && session->media_turn_relay && !ast_strlen_zero(session_media->bfcp_udp->turn_relay_address)) {
		ast_log(LOG_DEBUG, "[session:%p] just use avs turn address, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		media->conn = NULL;
	} else {
		media->conn->net_type = STR_IN;
		/* Assume that the connection will use IPv4 until proven otherwise */
		media->conn->addr_type = STR_IP4;
	}

	direct_media_enabled = !ast_sockaddr_isnull(&session_media->direct_media_addr) &&
		ast_format_cap_count(session->direct_media_cap);
	if (session->bfcp_para.has_udp_bfcp && session->media_turn_relay && !ast_strlen_zero(session_media->bfcp_udp->turn_relay_address)) {
		strncpy(hostip, session_media->bfcp_udp->turn_relay_address, sizeof(hostip) - 1);
	} else if (direct_media_enabled) {
		ast_log(LOG_DEBUG, "[session:%p] Create outgoing bfcp sdp, enable direct media, get hostip:%s for endpoint:%s\n",
			session, ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR), ast_sorcery_object_get_id(session->endpoint));
		ast_copy_string(hostip, ast_sockaddr_stringify_fmt(&session_media->direct_media_addr, AST_SOCKADDR_STR_ADDR), sizeof(hostip));
	} else if (ast_strlen_zero(session->endpoint->media.address)) {
		ast_copy_string(hostip, ast_sip_get_host_ip_string(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET()), sizeof(hostip));
		ast_log(LOG_DEBUG, "[session:%p] Create outgoing bfcp sdp, hostip:%s for endpoint:%s\n",
			session, hostip, ast_sorcery_object_get_id(session->endpoint));
	} else {
		ast_copy_string(hostip, session->endpoint->media.address, sizeof(hostip));
	}

	if (session->bfcp_para.has_udp_bfcp && session->media_turn_relay && !ast_strlen_zero(session_media->bfcp_udp->turn_relay_address)) {
		ast_log(LOG_DEBUG, "[session:%p] just use avs turn address, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	} else {
		pj_strdup2(pool, &media->conn->addr, hostip);

		if ((pj_sockaddr_parse(pj_AF_UNSPEC(), 0, &media->conn->addr, &ip) == PJ_SUCCESS) &&
			(ip.addr.sa_family == pj_AF_INET6())) {
			media->conn->addr_type = STR_IP6;
		}
	}

	if (session->bfcp_para.has_udp_bfcp && session->media_turn_relay && !ast_strlen_zero(session_media->bfcp_udp->turn_relay_address)) {
		media->desc.port = session_media->bfcp_udp->turn_relay_port;
	} else {
		media->desc.port = direct_media_enabled ? ast_sockaddr_port(&session_media->direct_media_addr) : (pj_uint16_t) ast_sockaddr_port(&bfcpaddr);
	}
	media->desc.port_count = 1;

	add_bfcp_param_to_sdp(pool, session, media);

	/* Add the media stream to the SDP */
	sdp->media[sdp->media_count++] = media;

	return 1;
}

static int create_outgoing_application_bfcp_sdp_stream(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_session *remote, struct ast_stream *stream)
{
	int res = 0;

	if (!session || !session_media || !sdp || !stream) {
		ast_log(LOG_DEBUG, "create_outgoing_application_bfcp_sdp_stream, para is null\n");
		return -1;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
		create_outgoing_disable_sdp(session->inv_session->pool_prov, session, session_media, sdp, remote, stream);
		return 0;
	}

	if (!session->bfcp_para.has_udp_bfcp && !session->bfcp_para.has_tcp_bfcp) {
		ast_log(LOG_DEBUG, "[session:%p] create_outgoing_application_bfcp_sdp_stream, has no bfcp for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	res = create_outgoing_app_bfcp_sdp_stream(session, session_media, sdp);

	return res;
}

static void set_main_video_first_multi_pt(struct ast_sip_session *session,
	struct ast_stream *stream, int pt, enum ast_stream_same_codec_type type)
{
	if (!session || !stream || !pt) {
		return;
	}

	if (!ast_stream_judge_if_main_video(stream)) {
		return;
	}

	switch (type) {
		case AST_STREAM_SAME_CODEC_TYPE_H264:
			if (!session->main_video_first_multi_h264_pt) {
				session->main_video_first_multi_h264_pt = pt;
				ast_log(LOG_DEBUG, "[session:%p] set_main_video_first_multi_pt H264 first PT:%d for endpoint:%s\n",
					session, pt, ast_sorcery_object_get_id(session->endpoint));
			}
			break;
		case AST_STREAM_SAME_CODEC_TYPE_RTX:
			if (!session->main_video_first_multi_rtx_pt) {
				session->main_video_first_multi_rtx_pt = pt;
				ast_log(LOG_DEBUG, "[session:%p] set_main_video_first_multi_pt RTX first PT:%d for endpoint:%s\n",
					session, pt, ast_sorcery_object_get_id(session->endpoint));
			}
			break;
		default:
			break;
	}

	return;
}

static int create_outgoing_sdp_multi_pt_h264_stream(struct ast_sip_session *session,
	pj_pool_t *pool, struct ast_stream *stream, pjmedia_sdp_media *media)
{
	unsigned int i = 0;
	unsigned int size = 0;
	pjmedia_sdp_attr *attr = NULL;
	struct ast_same_codec_info *codec_info = NULL;
	char fmtp_str[256] = { 0 };
	pj_str_t fmtp_pj_str;

	ast_log(LOG_DEBUG, "[session:%p] create_outgoing_sdp_multi_pt_h264_stream for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	size = ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_H264);
	for (i = 0; i < size; i++) {
		codec_info = ast_stream_get_same_codec_info(stream, AST_STREAM_SAME_CODEC_TYPE_H264, i);

		set_main_video_first_multi_pt(session, stream, codec_info->payload, AST_STREAM_SAME_CODEC_TYPE_H264);

		if ((attr = generate_rtpmap_attr(session, media, pool, codec_info->payload, 1, ast_format_h264, 0))) {
			media->attr[media->attr_count++] = attr;
		} else {
			ast_log(LOG_ERROR, "[session:%p] create rtpmap attr failed for multi H264 PT, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}

		snprintf(fmtp_str, sizeof(fmtp_str), "%d %s", codec_info->payload, codec_info->fmtp_param);
		fmtp_pj_str = pj_str(fmtp_str);
		attr = pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_pj_str);
		if (attr) {
			media->attr[media->attr_count++] = attr;
		} else {
			ast_log(LOG_DEBUG, "[session:%p] create fmtp attr failed for multi H264 PT, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	if (media->desc.fmt_count == PJMEDIA_MAX_SDP_FMT) {
		return -1;
	}

	return 0;
}

static int create_outgoing_sdp_multi_pt_rtx_stream(struct ast_sip_session *session,
	pj_pool_t *pool, struct ast_stream *stream, pjmedia_sdp_media *media)
{
	unsigned int i = 0;
	unsigned int size = 0;
	pjmedia_sdp_attr *attr = NULL;
	struct ast_same_codec_info *codec_info = NULL;
	char fmtp_str[256] = { 0 };
	pj_str_t fmtp_pj_str;

	ast_log(LOG_DEBUG, "[session:%p] create_outgoing_sdp_multi_pt_h264_stream for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	size = ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_RTX);
	for (i = 0; i < size; i++) {
		codec_info = ast_stream_get_same_codec_info(stream, AST_STREAM_SAME_CODEC_TYPE_RTX, i);

		set_main_video_first_multi_pt(session, stream, codec_info->payload, AST_STREAM_SAME_CODEC_TYPE_RTX);

		if ((attr = generate_rtpmap_attr(session, media, pool, codec_info->payload, 1, ast_format_rtx, 0))) {
			media->attr[media->attr_count++] = attr;
		} else {
			ast_log(LOG_ERROR, "[session:%p] create rtpmap attr failed for multi RTX PT, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}

		snprintf(fmtp_str, sizeof(fmtp_str), "%d %s", codec_info->payload, codec_info->fmtp_param);
		fmtp_pj_str = pj_str(fmtp_str);
		attr = pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_pj_str);
		if (attr) {
			media->attr[media->attr_count++] = attr;
		} else {
			ast_log(LOG_ERROR, "[session:%p] create fmtp attr failed for multi RTX PT, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	if (media->desc.fmt_count == PJMEDIA_MAX_SDP_FMT) {
		return -1;
	}

	return 0;
}

static int get_media_valid_pt(pjmedia_sdp_media *media, int pt)
{
	unsigned int i = 0;
	pj_str_t pt_str;
	char pt_s[8] = { 0 };
	int find = 0;
	int retry_pt = 96;
	/* For now we only have five video codec, it seems try 10 times is enough, just expand 16 when add new video codec */
	int try_max_count = 16;

PT_LOOP:
	snprintf(pt_s, sizeof(pt_s), "%d", pt);
	pt_str = pj_str(pt_s);

	find = 0;
	for (i = 0; i < media->desc.fmt_count; ++i) {
		if (!pj_stricmp(&pt_str, &media->desc.fmt[i])) {
			find = 1;
			try_max_count--;
			pt = retry_pt;
			retry_pt++;
			break;
		}
	}

	if (!try_max_count) {
		ast_log(LOG_ERROR, "Can't get valid RTX PT\n");
		return 0;
	}

	if (find) {
		goto PT_LOOP;
	}

	return pt;
}

static int create_outgoing_sdp_multi_pt_rtx_stream_by_self(struct ast_sip_session *session,
	pj_pool_t *pool, struct ast_stream *stream, pjmedia_sdp_media *media)
{
	unsigned int i = 0;
	unsigned int size = 0;
	pjmedia_sdp_attr *attr = NULL;
	struct ast_same_codec_info *codec_info = NULL;
	char fmtp_str[256] = { 0 };
	pj_str_t fmtp_pj_str;
	int rtx_pt = 0;

	ast_log(LOG_DEBUG, "[session:%p] create_outgoing_sdp_multi_pt_rtx_stream_by_self for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	size = ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_H264);
	for (i = 0; i < size; i++) {
		codec_info = ast_stream_get_same_codec_info(stream, AST_STREAM_SAME_CODEC_TYPE_H264, i);

		rtx_pt = get_media_valid_pt(media, codec_info->payload + 1);
		if (rtx_pt == 0) {
			continue;
		}

		set_main_video_first_multi_pt(session, stream, rtx_pt, AST_STREAM_SAME_CODEC_TYPE_RTX);

		if ((attr = generate_rtpmap_attr(session, media, pool, rtx_pt, 1, ast_format_rtx, 0))) {
			media->attr[media->attr_count++] = attr;
		} else {
			ast_log(LOG_ERROR, "[session:%p] create rtpmap attr failed for multi RTX PT, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}

		snprintf(fmtp_str, sizeof(fmtp_str), "%d apt=%d; rtx-time=3000", rtx_pt, codec_info->payload);
		fmtp_pj_str = pj_str(fmtp_str);
		attr = pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_pj_str);
		if (attr) {
			media->attr[media->attr_count++] = attr;
		} else {
			ast_log(LOG_ERROR, "[session:%p] create fmtp attr failed for multi RTX PT, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	if (media->desc.fmt_count == PJMEDIA_MAX_SDP_FMT) {
		return -1;
	}

	return 0;
}

static unsigned int judge_if_rtx_apt_is_h264(struct ast_sip_session *session,
	const struct pjmedia_sdp_media *stream, pjmedia_sdp_fmtp fmtp)
{
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	pj_str_t apt_str;
	char name[64] = { 0 };
	const unsigned int *apt = NULL;
	char apt_s[8] = { 0 };
	char fmt_param[64] = { 0 };
	RAII_VAR(struct ast_format *, format_parsed, NULL, ao2_cleanup);

	ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));
	format_parsed = ast_format_parse_sdp_fmtp(ast_format_rtx, fmt_param);

	if (!format_parsed) {
		ast_log(LOG_DEBUG, "[session:%p] Can't parse sdp fmtp for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	apt = ast_format_attribute_get(format_parsed, "apt");
	if (!apt || (*apt == 0)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get apt value for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	snprintf(apt_s, sizeof(apt_s), "%d", *apt);
	apt_str = pj_str(apt_s);

	ast_log(LOG_DEBUG, "[session:%p] judge_if_rtx_apt_is_h264, apt:%d in fmtp for endpoint:%s\n",
		session, *apt, ast_sorcery_object_get_id(session->endpoint));

	if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &apt_str))) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get rtpmap for PT:%d, endpoint:%s\n",
			session, *apt, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
		ast_log(LOG_DEBUG, "[session:%p] Can't prase to rtpmap for PT:%d for endpoint:%s\n",
			session, *apt, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));

	ast_log(LOG_DEBUG, "[session:%p] RTX apt:%d is according %s codec for endpoint:%s\n",
		session, *apt, name, ast_sorcery_object_get_id(session->endpoint));

	if (!strcasecmp(name, "H264")) {
		return 1;
	}

	return 0;
}

static unsigned int create_outgoing_sdp_multi_pt(struct ast_sip_session *session,
	struct ast_format *format, pj_pool_t *pool, struct ast_stream *stream, pjmedia_sdp_media *media)
{
	if (session->is_sfu) {
		return 0;
	}

	if (!strcasecmp(ast_format_get_name(format), "H264") &&
		ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_H264) > 1) {
		create_outgoing_sdp_multi_pt_h264_stream(session, pool, stream, media);
		return 1;
	}

	if (!strcasecmp(ast_format_get_name(format), "RTX") &&
		ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_RTX) > 1) {
		create_outgoing_sdp_multi_pt_rtx_stream(session, pool, stream, media);
		return 1;
	}

	if (!strcasecmp(ast_format_get_name(format), "RTX") &&
		ast_stream_same_codec_info_get_count(stream, AST_STREAM_SAME_CODEC_TYPE_H264) > 1) {
		create_outgoing_sdp_multi_pt_rtx_stream_by_self(session, pool, stream, media);
		return 1;
	}

	return 0;
}

static void set_stream_same_codec_info(struct ast_sip_session *session,
	const struct pjmedia_sdp_media *stream, char *name, int i, struct ast_stream *asterisk_stream, pjmedia_sdp_fmtp fmtp)
{
	struct ast_same_codec_info codec_info;
	char fmt_param[256] = { 0 };

	if (!session->endpoint->media.support_multi_pt) {
		ast_log(LOG_DEBUG, "[session:%p] set_stream_same_codec_info, can't support multi pt, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->inv_session->neg->state != PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER) {
		ast_log(LOG_DEBUG, "[session:%p] set_stream_same_codec_info is not remote offer, don't need set, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->is_sfu) {
		ast_log(LOG_DEBUG, "[session:%p] set_stream_same_codec_info is SFU, don't need set, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));

	if (strcasecmp(name, "H264") == 0) {
		memset(&codec_info, 0, sizeof(codec_info));
		codec_info.payload = pj_strtoul(&stream->desc.fmt[i]);
		strncpy(codec_info.fmtp_param, fmt_param, sizeof(codec_info.fmtp_param) - 1);
		ast_log(LOG_DEBUG, "[session:%p] set H264 same codec info, pt:%d, fmtp param:%s, endpoint:%s\n",
			session, codec_info.payload, codec_info.fmtp_param, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_same_codec_info(asterisk_stream, AST_STREAM_SAME_CODEC_TYPE_H264, codec_info);
	}

	if (strcasecmp(name, "RTX") == 0 && judge_if_rtx_apt_is_h264(session, stream, fmtp)) {
		memset(&codec_info, 0, sizeof(codec_info));
		codec_info.payload = pj_strtoul(&stream->desc.fmt[i]);
		strncpy(codec_info.fmtp_param, fmt_param, sizeof(codec_info.fmtp_param) - 1);
		ast_log(LOG_DEBUG, "[session:%p] set RTX same codec info, pt:%d, fmtp param:%s, endpoint:%s\n",
			session, codec_info.payload, codec_info.fmtp_param, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_same_codec_info(asterisk_stream, AST_STREAM_SAME_CODEC_TYPE_RTX, codec_info);
	}

	return;
}

static int send_bfcp_keepalive(const void *data)
{
	int res = 0;
	struct ast_sip_session_media *media = NULL;
	char buf[4] = { "000" };

	if (!data) {
		ast_log(LOG_WARNING, "send_bfcp_keepalive data is null");
		return 0;
	}

	media = (struct ast_sip_session_media *)data;
	if (media && media->bfcp_udp) {
		if (media->bfcp_udp->media_turn_relay) {
			if (media->bfcp_udp->engine->turn_packet_send(media->bfcp_udp, buf, strlen(buf) + 1)) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&media->bfcp_udp->them), strerror(errno));
			}
		} else {
			if ((res = ast_sendto(media->bfcp_udp->fd, buf, strlen(buf) + 1, 0, &media->bfcp_udp->them)) < 0) {
				ast_log(LOG_NOTICE, "BFCP : Transmission error to %s: %s\n",
					ast_sockaddr_stringify(&media->bfcp_udp->them), strerror(errno));
			}
		}
	}

	return SEND_BFCP_KEEPALIVE_INTERNAL*1000;
}

static int send_bfcp_udp_hello(const void *data)
{
	struct ast_sip_session_media *media = NULL;

	if (!data) {
		ast_log(LOG_WARNING, "send_bfcp_udp_hello data is null");
		return 0;
	}

	media = (struct ast_sip_session_media *)data;
	if (media && media->bfcp_udp) {
		media->bfcp_udp->bfcp_info.primitive = BFCP_HELLO;
		ast_send_bfcp_msg(media->bfcp_udp);
	}

	return 0;
}

static int send_bfcp_tcp_hello(const void *data)
{
	struct ast_sip_session_media *media = NULL;

	if (!data) {
		ast_log(LOG_WARNING, "send_bfcp_tcp_hello data is null");
		return 0;
	}

	media = (struct ast_sip_session_media *)data;
	if (media && media->bfcp_tcp) {
		ast_log(LOG_DEBUG, "send_bfcp_tcp_hello for mslabel:%s\n", media->mslabel);
		media->bfcp_tcp->bfcp_info.primitive = BFCP_HELLO;
		ast_send_bfcp_tcp_msg(media->bfcp_tcp);
	}

	return 0;
}

static int apply_negotiated_app_bfcp_udp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		ast_log(LOG_DEBUG, "[session:%p] Apply UDP BFCP, peer BFCP is declined, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->bfcp_para.has_udp_bfcp = 0;
		return 1;
	}

	if (!session_media->bfcp_udp && create_udp_bfcp(session, session_media)) {
		ast_log(LOG_ERROR, "apply bfcp, create udp error\n");
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "[session:%p] Apply UDP BFCP, provided host was actually invalid, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	/* Apply connection information to the bfcp */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_bfcp_set_peer(session_media->bfcp_udp, addrs);
	ast_verbose("[session:%p] Peer UDP BFCP is at port %s, endpoint:%s\n",
		session, ast_sockaddr_stringify(addrs), ast_sorcery_object_get_id(session->endpoint));

	if (session->media_turn_relay && session_media->bfcp_udp->engine) {
		session_media->bfcp_udp->engine->set_peer_address(session_media->bfcp_udp, addrs);
	}

	if (session->channel) {
		session_media->bfcp_udp->channel = session->channel;
	}

	ast_sip_session_media_add_read_callback(session, session_media, ast_bfcp_fd(session_media->bfcp_udp), NULL);

	if (pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg)) {
		process_sdp_a_app_bfcp(session, remote_stream);
	}

	if (!strcasecmp(session->bfcp_para.local_floorctrl, "c-only") && !session->sent_bfcp_hello) {
		AST_SCHED_REPLACE_VARIABLE(session_media->bfcp_sched_id, sched, SEND_FIRST_BFCP_HELLO*1000, send_bfcp_udp_hello,
			session_media, 1);
		session->sent_bfcp_hello = 1;
	}

	if (session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) {
		AST_SCHED_REPLACE_VARIABLE(session_media->bfcp_keepalive_id, sched, SEND_BFCP_KEEPALIVE_INTERNAL*1000, send_bfcp_keepalive,
			session_media, 1);
	}

	copy_para_to_bfcp_info(session, session_media->bfcp_udp);

	set_bfcp_udp_conference_params(session, session_media->bfcp_udp);

	session->bfcp_para.has_udp_bfcp = 1;

	return 1;
}

static int apply_negotiated_app_bfcp_tcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	RAII_VAR(struct ast_sockaddr *, addrs, NULL, ast_free);
	char host[NI_MAXHOST];

	if (!local_stream->desc.port || !remote_stream->desc.port) {
		ast_log(LOG_DEBUG, "[session:%p] Apply TCP BFCP, peer BFCP is declined, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->bfcp_para.has_tcp_bfcp = 0;
		return 1;
	}

	if (!session_media->bfcp_tcp && create_tcp_bfcp(session, session_media)) {
		ast_log(LOG_ERROR, "[session:%p] Apply TCP BFCP, create tcp error, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (!remote_stream->conn && !remote->conn) {
		return 1;
	}

	ast_copy_pj_str(host, remote_stream->conn ? &remote_stream->conn->addr : &remote->conn->addr, sizeof(host));

	/* Ensure that the address provided is valid */
	if (ast_sockaddr_resolve(&addrs, host, PARSE_PORT_FORBID, AST_AF_UNSPEC) <= 0) {
		/* The provided host was actually invalid so we error out this negotiation */
		ast_log(LOG_ERROR, "[session:%p] Apply TCP BFCP, provided host was actually invalid, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	/* Apply connection information to the bfcp */
	ast_sockaddr_set_port(addrs, remote_stream->desc.port);
	ast_bfcp_tcp_set_peer(session_media->bfcp_tcp, addrs);
	ast_verbose("[session:%p] Peer TCP BFCP is at port %s, endpoint:%s\n",
		session, ast_sockaddr_stringify(addrs), ast_sorcery_object_get_id(session->endpoint));

	if (session && session->channel && !session_media->bfcp_tcp->channel) {
		session_media->bfcp_tcp->channel = session->channel;
		ast_channel_ref(session->channel);
	}

	if (pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg)) {
		process_sdp_a_app_bfcp(session, remote_stream);
	}

	if (!strcasecmp(session->bfcp_para.local_floorctrl, "c-only") && !session->sent_bfcp_hello) {
		AST_SCHED_REPLACE_VARIABLE(session_media->bfcp_sched_id, sched, SEND_FIRST_BFCP_HELLO*1000, send_bfcp_tcp_hello,
			session_media, 1);
		session->sent_bfcp_hello = 1;
	}

	copy_para_to_bfcp_tcp_info(session, session_media->bfcp_tcp);

	set_bfcp_tcp_conference_params(session, session_media->bfcp_tcp);

	session->bfcp_para.has_tcp_bfcp = 1;

	return 1;
}

static int apply_negotiated_application_bfcp_sdp_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
				       const struct pjmedia_sdp_session *local, const struct pjmedia_sdp_media *local_stream,
				       const struct pjmedia_sdp_session *remote, const struct pjmedia_sdp_media *remote_stream)
{
	int res = 0;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!session || !session_media || !local || !local_stream || !remote || !remote_stream) {
		ast_log(LOG_DEBUG, "Apply negotiated application bfcp sdp stream para is null\n");
		return -1;
	}

	sub_type = session_media->sub_type;
	switch (sub_type) {
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP:
			res = apply_negotiated_app_bfcp_udp_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
			break;
		case AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP:
			res = apply_negotiated_app_bfcp_tcp_sdp_stream(session, session_media, local, local_stream, remote, remote_stream);
			break;
		default:
			ast_log(LOG_DEBUG, "Can't apply other application bfcp sdp stream\n");
			break;
	}

	return res;
}

static void add_bundle_only_to_stream(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !pool || !media) {
		return;
	}

	attr = pjmedia_sdp_attr_create(pool, "bundle-only", NULL);
	pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);

	return;
}

static void create_bundled_media_line(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	struct pjmedia_sdp_session *sdp, pjmedia_sdp_media *media, struct ast_stream *stream, pj_pool_t *pool)
{
	pjmedia_sdp_media *bundle_group_stream = NULL;
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3};
	static const pj_str_t STR_IP6 = { "IP6", 3};
	const char *hostip = NULL;
	pj_sockaddr ip;

	if (AST_NEG_POOL_INV_PROV == select_pjsip_pool(session, stream)) {
		ast_log(LOG_DEBUG, "[session:%p] Find bundled group stream, use this to set media line params, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		bundle_group_stream = sdp->media[session_media->stream_num];
		if (!bundle_group_stream) {
			ast_log(LOG_WARNING, "[session:%p] Can't get according media, use first media as default, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			bundle_group_stream = sdp->media[0];
			if (!bundle_group_stream) {
				ast_log(LOG_ERROR, "[session:%p] Can't get any according media, there is something wrong, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				return;
			}
		}
		media->desc.transport = bundle_group_stream->desc.transport;
		media->conn = bundle_group_stream->conn;
		media->desc.port = 9;
	} else {
		ast_log(LOG_DEBUG, "[session:%p] It's gswave and participant stream, create bundled media line use default value, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (ast_strlen_zero(session->endpoint->media.address)) {
			hostip = ast_sip_get_host_ip_string(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET());
		} else {
			hostip = session->endpoint->media.address;
		}

		if (ast_strlen_zero(hostip)) {
			ast_log(LOG_ERROR, "[session:%p] No local host IP available for stream %s, endpoint:%s\n",
				session, ast_codec_media_type2str(session_media->type), ast_sorcery_object_get_id(session->endpoint));
			return;
		}

		media->conn = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_conn));
		if (!media->conn) {
			ast_log(LOG_ERROR, "[session:%p] Can't alloc conn for stream %s, endpoint:%s\n",
				session, ast_codec_media_type2str(session_media->type), ast_sorcery_object_get_id(session->endpoint));
			return;
		}

		/* Bundle is not care the port, we use port 9 as default */
		ast_sip_session_set_media_transport(session, session_media, media);
		media->desc.port = 9;
		media->conn->net_type = STR_IN;
		/* Assume that the connection will use IPv4 until proven otherwise */
		media->conn->addr_type = STR_IP4;
		pj_strdup2(pool, &media->conn->addr, hostip);

		if ((pj_sockaddr_parse(pj_AF_UNSPEC(), 0, &media->conn->addr, &ip) == PJ_SUCCESS) &&
			(ip.addr.sa_family == pj_AF_INET6())) {
			media->conn->addr_type = STR_IP6;
		}
	}

	return;
}

/* 0:pool_prov, 1:pool_notify_prov */
static enum ast_neg_pool_type select_pjsip_pool(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!ast_stream_judge_if_stream_is_participant(stream)) {
		return AST_NEG_POOL_INV_PROV;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC &&
		session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		return AST_NEG_POOL_INV_PROV;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && !session->endpoint->media.gswave_bundle_enabled) {
		return AST_NEG_POOL_INV_PROV;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE && !session->endpoint->media.mobile_bundle_enabled) {
		return AST_NEG_POOL_INV_PROV;
	}

	return AST_NEG_POOL_NOTIFY_PROV;
}

static enum ast_special_codec special_codec_str_2_enum(const char *str)
{
	int i = 0;

	for (i = 0; i < AST_SPECIAL_CODEC_MAX; i++) {
		if (!strcasecmp(str, ast_special_codec_strmap[i][0])) {
			return (enum ast_special_codec)i;
		}
	}

	return AST_SPECIAL_CODEC_UNKNOW;
}

static void process_special_codec_gs_fec(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format *format_parsed)
{
	const char *gs_fec_version = ast_format_attribute_get(format_parsed, "version");

	if (!ast_strlen_zero(gs_fec_version)) {
		ast_log(LOG_DEBUG, "Get gs-fec-version in fmtp:%s\n", gs_fec_version);
		strncpy(session_media->gs_fec_version, gs_fec_version, sizeof(session_media->gs_fec_version) - 1);
	}

	return;
}

static void process_special_codec_ilbc(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format *format_parsed)
{
	const unsigned int *ilbc_mode = ast_format_attribute_get(format_parsed, "mode");

	if (ilbc_mode) {
		ast_log(LOG_DEBUG, "[session:%p] Get ilbc mode:%d in fmtp for endpoint:%s\n",
			session, *ilbc_mode, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_ilbc_frame_mode(session_media->rtp, *ilbc_mode);
	}

	return;
}

static void process_special_codec_opus(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format *format_parsed)
{
	const int *useinbandfec = ast_format_attribute_get(format_parsed, "useinbandfec");

	if (useinbandfec && *useinbandfec) {
		ast_log(LOG_DEBUG, "[session:%p] Get useinbandfec:%d in fmtp for endpoint:%s\n",
			session, *useinbandfec, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_inbandfec(session_media->rtp, *useinbandfec);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Has no useinbandfec in fmtp, clear fec flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_rtp_instance_set_inbandfec(session_media->rtp, 0);
	}

	if (session->endpoint->media.dtx && session->endpoint->media.comfort_noise &&
		(pjmedia_sdp_neg_was_answer_remote(session->inv_session->neg) ||
		pjmedia_sdp_neg_get_state(session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER)) {
		const int *usedtx = ast_format_attribute_get(format_parsed, "usedtx");
		if (usedtx && *usedtx) {
			ast_log(LOG_DEBUG, "[session:%p] Get usedtx:%d in fmtp for endpoint:%s\n",
				session, *usedtx, ast_sorcery_object_get_id(session->endpoint));
			ast_rtp_instance_set_opus_dtx_flag(session_media->rtp, *usedtx);
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Has no usedtx in fmtp, clear dtx flag for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_rtp_instance_set_opus_dtx_flag(session_media->rtp, 0);
		}
	}

	return;
}

static void process_special_codec_rtx(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, int position)
{
	ast_log(LOG_DEBUG, "[session:%p] It's RTX codec, set RTX PT:%ld to AVS for endpoint:%s\n",
		session, pj_strtoul(&stream->desc.fmt[position]), ast_sorcery_object_get_id(session->endpoint));

	ast_rtp_instance_set_rtx_pt(session_media->rtp, pj_strtoul(&stream->desc.fmt[position]));

	return;
}

static void process_special_codec_g723(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format *format_parsed)
{
	//const char *annexa = NULL;

	/* Todo: AVS can't support now, disable annexa */
	ast_format_attribute_set_directly(format_parsed, "annexa", "no");
	return;

	if (!session->endpoint->media.comfort_noise) {
		ast_format_attribute_set_directly(format_parsed, "annexa", "no");
		return;
	}

	//annexa = ast_format_attribute_get(format_parsed, "annexa");

	/* Todo: Set avs annexa param */

	return;
}

static void process_special_codec_g729(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_format *format_parsed)
{
	//const char *annexb = NULL;

	/* Todo: AVS can't support now, disable annexb */
	ast_format_attribute_set_directly(format_parsed, "annexb", "no");
	return;

	if (!session->endpoint->media.comfort_noise) {
		ast_format_attribute_set_directly(format_parsed, "annexb", "no");
		return;
	}

	//annexb = ast_format_attribute_get(format_parsed, "annexb");

	/* Todo: Set avs annexb param */

	return;
}

static void process_special_codec_params(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, char *name, struct ast_format *format_parsed, int position)
{
	enum ast_special_codec codec = AST_SPECIAL_CODEC_UNKNOW;

	if (!session || !stream || !session_media || !name || !format_parsed) {
		return;
	}

	codec = special_codec_str_2_enum(name);

	switch (codec) {
		case AST_SPECIAL_CODEC_GS_FEC:
			process_special_codec_gs_fec(session, session_media, format_parsed);
			break;
		case AST_SPECIAL_CODEC_ILBC:
			process_special_codec_ilbc(session, session_media, format_parsed);
			break;
		case AST_SPECIAL_CODEC_OPUS:
			process_special_codec_opus(session, session_media, format_parsed);
			break;
		case AST_SPECIAL_CODEC_RTX:
			process_special_codec_rtx(session, stream, session_media, position);
			break;
		case AST_SPECIAL_CODEC_G723:
			process_special_codec_g723(session, session_media, format_parsed);
			break;
		case AST_SPECIAL_CODEC_G729:
			process_special_codec_g729(session, session_media, format_parsed);
			break;
		default:
			break;
	}

	return;
}

static void set_bundle_group_for_mobile(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	if (!session || !session_media) {
		return;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		return;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return;
	}

	/* For now, all media are bundle together to group 0, set 0 as default */
	session_media->bundle_group = 0;

	return;
}

static enum ast_stream_video_resolution get_resolution_from_stream(struct ast_sip_session *session,
	struct ast_stream *asterisk_stream)
{
	RAII_VAR(struct ast_format *, format, NULL, ao2_cleanup);
	enum ast_stream_video_resolution new_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	format = ast_sip_session_get_video_format(asterisk_stream, ast_sip_session_sfu_codec2str(session->sfu_video_codec));
	if (!format) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get video format, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return AST_STREAM_VIDEO_RESOLUTION_NONE;
	}

	new_video_resolution = get_received_video_resolution(session, format, asterisk_stream);

	return new_video_resolution;
}

static void set_stream_final_video_resolution(struct ast_sip_session *session, struct ast_stream *asterisk_stream)
{
	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session || !asterisk_stream) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!ast_stream_is_enabled_main_video(asterisk_stream)) {
		return;
	}

	resolution = get_resolution_from_stream(session, asterisk_stream);

	ast_log(LOG_DEBUG, "[session:%p] set_stream_final_video_resolution resolution:%d, endpoint:%s\n",
			session, resolution, ast_sorcery_object_get_id(session->endpoint));

	ast_stream_set_video_resolution(asterisk_stream, resolution);

	return;
}

static void update_cei_for_change_video_resolution(struct ast_sip_session *session, struct ast_stream *asterisk_stream)
{
	enum ast_stream_video_resolution previous_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;
	enum ast_stream_video_resolution new_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session || !session->channel || !asterisk_stream) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!ast_stream_judge_if_main_video(asterisk_stream)) {
		return;
	}

	if (!session->change_resolution_reinvite) {
		return;
	}

	session->change_resolution_reinvite = 0;

	previous_video_resolution = ast_stream_get_previous_video_resolution(asterisk_stream);
	new_video_resolution = ast_stream_get_video_resolution(asterisk_stream);

	ast_log(LOG_DEBUG, "[session:%p] update_cei_for_change_video_resolution previous resolution:%d, new resolution:%d, endpoint:%s\n",
			session, previous_video_resolution, new_video_resolution, ast_sorcery_object_get_id(session->endpoint));

	if (previous_video_resolution != new_video_resolution) {
		ast_conf_cei_video_resolution_update(session->channel, session->conf_room, new_video_resolution);
	}

	return;
}

static void indicate_wave_release_hd_video_resource(struct ast_sip_session *session, struct ast_stream *asterisk_stream)
{
	enum ast_stream_video_resolution new_video_resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session || !session->channel || !asterisk_stream) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!ast_stream_judge_if_main_video(asterisk_stream)) {
		return;
	}

	if (!session->release_hd_by_wave) {
		return;
	}

	session->release_hd_by_wave = 0;

	new_video_resolution = ast_stream_get_video_resolution(asterisk_stream);

	ast_log(LOG_DEBUG, "[session:%p] indicate_wave_release_hd_video_resource, new resolution:%d, endpoint:%s\n",
			session, new_video_resolution, ast_sorcery_object_get_id(session->endpoint));

	if (new_video_resolution == AST_STREAM_VIDEO_RESOLUTION_360P) {
		ast_queue_control(session->channel, AST_CONTROL_WAVE_RELEASED_HD_RESOLUTION);
	}

	return;
}

static void clear_video_codec_rtx_pt(struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	if (!session_media || !stream) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	init_video_codec_rtx_pt(session_media);
	ast_stream_init_video_rtx_pt(stream);

	return;
}

static void set_comfort_noise_pt(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int pt, char *rtpmap_name, pjmedia_sdp_rtpmap *rtpmap)
{
	if (!session || !session_media) {
		return;
	}

	if (!session->endpoint->media.comfort_noise) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (!rtpmap_name || !rtpmap) {
		if (pt == 13 || pt == 19) {
			session_media->cn_8k_pt = pt;
			ast_log(LOG_DEBUG, "[session:%p] set_comfort_noise_pt has no rtpmap, set static cn PT:%d endpoint:%s\n",
				session, pt, ast_sorcery_object_get_id(session->endpoint));
			ast_rtp_instance_set_cn_8k_pt(session_media->rtp, pt);
		}
		return;
	}

	if (strcasecmp(rtpmap_name, "CN")) {
		return;
	}

	if (rtpmap->clock_rate == 8000) {
		session_media->cn_8k_pt = pt;
		ast_rtp_instance_set_cn_8k_pt(session_media->rtp, pt);
		ast_log(LOG_DEBUG, "[session:%p] set_comfort_noise_pt set cn_8k_pt:%d endpoint:%s\n",
			session, pt, ast_sorcery_object_get_id(session->endpoint));
	} else if (rtpmap->clock_rate == 16000) {
		session_media->cn_16k_pt = pt;
		ast_rtp_instance_set_cn_16k_pt(session_media->rtp, pt);
		ast_log(LOG_DEBUG, "[session:%p] set_comfort_noise_pt set cn_16k_pt:%d endpoint:%s\n",
			session, pt, ast_sorcery_object_get_id(session->endpoint));
	} else if (rtpmap->clock_rate == 32000) {
		session_media->cn_32k_pt = pt;
		ast_rtp_instance_set_cn_32k_pt(session_media->rtp, pt);
		ast_log(LOG_DEBUG, "[session:%p] set_comfort_noise_pt set cn_32k_pt:%d endpoint:%s\n",
			session, pt, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

static void clear_comfort_noise_pt(struct ast_sip_session_media *session_media)
{
	if (!session_media) {
		return;
	}

	session_media->cn_8k_pt = 0;
	session_media->cn_16k_pt = 0;
	session_media->cn_32k_pt = 0;

	/* Clear CN PT first */
	ast_rtp_instance_set_cn_8k_pt(session_media->rtp, 0);
	ast_rtp_instance_set_cn_16k_pt(session_media->rtp, 0);
	ast_rtp_instance_set_cn_32k_pt(session_media->rtp, 0);

	return;
}

static void set_audio_media_mpa(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int pt)
{
	if (!session || !session_media) {
		return;
	}

	if (!session->endpoint->media.mpa) {
		return;
	}

	if (session_media->type != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	/* For MPA is a static codec, and PT value is 14, just compare PT value */
	if (pt != 14) {
		return;
	}

	session_media->mpa = 1;

	ast_rtp_instance_set_mpa(session_media->rtp, 1);

	return;
}

static void clear_audio_media_mpa(struct ast_sip_session_media *session_media)
{
	if (!session_media) {
		return;
	}

	session_media->mpa = 0;

	ast_rtp_instance_set_mpa(session_media->rtp, 0);

	return;
}

#endif

/*! \brief Unloads the sdp RTP/AVP module from Asterisk */
static int unload_module(void)
{
	ast_sip_session_unregister_supplement(&video_info_supplement);
	ast_sip_session_unregister_sdp_handler(&video_sdp_handler, STR_VIDEO);
	ast_sip_session_unregister_sdp_handler(&audio_sdp_handler, STR_AUDIO);

	if (sched) {
		ast_sched_context_destroy(sched);
	}

	return 0;
}

/*!
 * \brief Load the module
 *
 * Module loading including tests for configuration or dependencies.
 * This function can return AST_MODULE_LOAD_FAILURE, AST_MODULE_LOAD_DECLINE,
 * or AST_MODULE_LOAD_SUCCESS. If a dependency or environment variable fails
 * tests return AST_MODULE_LOAD_FAILURE. If the module can not load the
 * configuration file or other non-critical problem return
 * AST_MODULE_LOAD_DECLINE. On success return AST_MODULE_LOAD_SUCCESS.
 */
static int load_module(void)
{
	if (ast_check_ipv6()) {
		ast_sockaddr_parse(&address_rtp, "::", 0);
	} else {
		ast_sockaddr_parse(&address_rtp, "0.0.0.0", 0);
	}

	if (!(sched = ast_sched_context_create())) {
		ast_log(LOG_ERROR, "Unable to create scheduler context.\n");
		goto end;
	}

	if (ast_sched_start_thread(sched)) {
		ast_log(LOG_ERROR, "Unable to create scheduler context thread.\n");
		goto end;
	}

	if (ast_sip_session_register_sdp_handler(&audio_sdp_handler, STR_AUDIO)) {
		ast_log(LOG_ERROR, "Unable to register SDP handler for %s stream type\n", STR_AUDIO);
		goto end;
	}

	if (ast_sip_session_register_sdp_handler(&video_sdp_handler, STR_VIDEO)) {
		ast_log(LOG_ERROR, "Unable to register SDP handler for %s stream type\n", STR_VIDEO);
		goto end;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for BFCP/FEC/FECC, 2015.07.09 */
	if (ast_sip_session_register_sdp_handler(&application_sdp_handler, STR_APPLICATION)) {
		ast_log(LOG_ERROR, "Unable to register SDP handler for %s stream type\n", STR_APPLICATION);
		goto end;
	}
#endif

	ast_sip_session_register_supplement(&video_info_supplement);

	return AST_MODULE_LOAD_SUCCESS;
end:
	unload_module();

	return AST_MODULE_LOAD_DECLINE;
}

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_LOAD_ORDER, "PJSIP SDP RTP/AVP stream handler",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
	.load_pri = AST_MODPRI_CHANNEL_DRIVER,
	.requires = "res_pjsip,res_pjsip_session",
);
