/*
* Asterisk -- An open source telephony toolkit.
*
* Copyright (C) 2013, Digium, Inc.
*
* Mark Michelson <mmichelson@digium.com>
*
* See http://www.asterisk.org for more information about
* the Asterisk project. Please do not directly contact
* any of the maintainers of this project for assistance;
* the project provides a web site, mailing lists and IRC
* channels for your use.
*
* This program is free software, distributed under the terms of
* the GNU General Public License Version 2. See the LICENSE file
* at the top of the source tree.
*/

/*** MODULEINFO
	<depend>pjproject</depend>
	<depend>res_pjsip</depend>
	<support_level>core</support_level>
 ***/

#include "asterisk.h"

#include <pjsip.h>
#include <pjsip_ua.h>
#include <pjlib.h>
#include <ifaddrs.h>  
#include <netinet/in.h>  
#include <arpa/inet.h>  
#include <stdio.h> 

#include "asterisk/res_pjsip.h"
#include "asterisk/res_pjsip_session.h"
#include "asterisk/callerid.h"
#include "asterisk/datastore.h"
#include "asterisk/module.h"
#include "asterisk/logger.h"
#include "asterisk/res_pjsip.h"
#include "asterisk/astobj2.h"
#include "asterisk/lock.h"
#include "asterisk/uuid.h"
#include "asterisk/pbx.h"
#include "asterisk/taskprocessor.h"
#include "asterisk/causes.h"
#include "asterisk/sdp_srtp.h"
#include "asterisk/dsp.h"
#include "asterisk/acl.h"
#include "asterisk/features_config.h"
#include "asterisk/pickup.h"
#include "asterisk/test.h"
#include "asterisk/stream.h"

#ifdef GRANDSTREAM_NETWORKS
/*add by wxu@grandstream.cn*/
#include "asterisk/security_events.h"
/*end*/
#include "../channels/pjsip/include/chan_pjsip.h"
#include "asterisk/astdb.h"
#include "asterisk/avs_controller.h"
#include "asterisk/res_pjsip_bfcp_info.h"
#include "../apps/confbridge/include/conf_resource.h"
#include "asterisk/bridge.h"
#include "asterisk/cli.h"
#include "asterisk/app.h"
#include "asterisk/format_cache.h"
#include "asterisk/switch_xml.h"
#include "asterisk/haserver.h"
#include "asterisk/res_pjsip_camera_control.h"
#include "asterisk/bridge_internal.h"
#include "../apps/confbridge/include/confbridge.h"
#include "asterisk/multi_media.h"

AST_MODULE_LOG("pjsip");
#endif

#define SDP_HANDLER_BUCKETS 11

#define MOD_DATA_ON_RESPONSE "on_response"
#define MOD_DATA_NAT_HOOK "nat_hook"

/* Most common case is one audio and one video stream */
#define DEFAULT_NUM_SESSION_MEDIA 2

/* Some forward declarations */
static void handle_session_begin(struct ast_sip_session *session);
static void handle_session_end(struct ast_sip_session *session);
static void handle_session_destroy(struct ast_sip_session *session);
static void handle_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void handle_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata,
		enum ast_sip_session_response_priority response_priority);
static int handle_incoming(struct ast_sip_session *session, pjsip_rx_data *rdata,
		enum ast_sip_session_response_priority response_priority);
static void handle_outgoing_request(struct ast_sip_session *session, pjsip_tx_data *tdata);
static void handle_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata);

#ifdef GRANDSTREAM_NETWORKS
static void set_yealink_flag(pjsip_rx_data *rdata, struct ast_sip_session *session);
static void init_session_param(struct ast_sip_session *session);
static void set_webrtc_call_flag(struct ast_sip_session *session, const char *uri, struct ast_sip_contact *found_contact);
static void ast_sip_report_invite_failed(struct ast_sip_session *session,struct pjsip_rx_data *rdata,struct pjsip_tx_data *tdata,
	const char *sipmethod, int resp, const char* resp_text);
static pj_bool_t does_method_match(const pj_str_t *message_method, const char *supplement_method);
static void set_terminal_type(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_is_sdp_ipv6(struct ast_sip_session *session, pjmedia_sdp_session *sdp);
static void set_session_msid_count(struct ast_sip_session *session);
static void set_session_media_mslabel(struct ast_sip_session *session, struct ast_sip_session_media *session_media);
static void set_session_media_label(struct ast_sip_session_media *session_media, enum ast_media_sub_type sub_type, struct ast_sip_session *session);
static void set_stream_params_according_exist_stream(struct ast_stream *stream, struct ast_stream *existing_stream);
static void indicate_conference_control_camera(struct ast_sip_session *session);
static struct ast_sip_session_media_state *get_session_media_state(struct ast_sip_session *session, int is_noitfy);
static int append_isfocus_contact_for_reinvite(pjsip_tx_data *packet);
//static void add_bandw_ct_to_sdp(pjsip_inv_session *inv, struct ast_sip_session *session,
//	const pjmedia_sdp_session *offer, pjmedia_sdp_session *local);
static struct ast_sip_session_media *add_session_media_for_stream(struct ast_sip_session *session,
	const pjmedia_sdp_session *sdp, int position);
static void push_sfu_delayed_reinvite_to_task(struct ast_sip_session *session);
static void change_participant_stream_params_for_sfu(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream);
static void change_rtp_instance_local_ssrc(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream);
static void change_stream_msid_according_session_media(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream);
static void negotiated_for_transfer(struct ast_sip_session *session);
static unsigned char judge_if_bridge_exsit(struct ast_channel* chan);
static void notify_topogly_for_sfu_local_invite(struct ast_sip_session *session);
static void set_ice_flag(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_caller_webrtc_call_flag(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_ice_trunk_for_session(struct ast_sip_session *session);
static void notify_nat_call(struct ast_sip_session *session);
static void set_media_turn_relay(struct ast_sip_session *session);
static void clear_media_turn_relay_for_remote_support_ice(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_turn_target_address(struct ast_sip_session *session,
	struct ast_sip_contact *found_contact, const char *location);
static void set_caller_gds_call(struct ast_sip_session* session, pjsip_rx_data* rdata);
static void set_caller_gsc_call(struct ast_sip_session* session, pjsip_rx_data* rdata);
static void set_caller_hikvision_call(struct ast_sip_session* session, pjsip_rx_data* rdata);
static void set_stream_content(struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *stream, enum ast_media_type type);
static void move_dtls_and_ice_attrs_to_session_level(pj_pool_t *pool,
	pjmedia_sdp_session *local, struct ast_sip_session *session);
static unsigned int judge_if_invalid_caps(struct ast_format_cap *joint_cap);
static pj_bool_t session_search_audio_port_with_image_audio(struct ast_sip_session *session, int image_audio, int is_t38_enable, pjmedia_sdp_session *receive_sdp, const char *media_type_name);
static void notify_full_participant_sdp_for_reconnect(struct ast_sip_session *session);
void ast_sip_subscription_update_transport(pjsip_transport	*transport, const char *call_id);
static void set_session_remote_support_bundle(struct ast_sip_session *session, const pjmedia_sdp_session *remote);
static void send_bfcp_info_for_reconnect(struct ast_sip_session *session);
static void notify_remote_disable_media_for_transfer(struct ast_sip_session *session);
static void send_bfcp_for_open_slides(struct ast_sip_session *session);
static void send_delayed_bfcp_message(struct ast_sip_session *session);
static void change_bundle_state(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream);
static void set_support_whiteboard_flag(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void queue_camera_control_frame(struct ast_sip_session *session);
static void set_candidate_host_only_flag_for_caller(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_candidate_host_only_flag_for_callee(struct ast_sip_session *session,
	struct ast_sip_contact *found_contact, const char *location);
static void set_session_allow_events(struct ast_sip_session *session, struct pjsip_rx_data *rdata);
static void send_close_camera_for_webrtc(struct ast_sip_session *session, struct ast_stream *stream);
static void send_get_peer_session_frame(struct ast_sip_session *session);
static void set_session_emergency_call(struct ast_sip_session *session, pjsip_rx_data *rdata);
static unsigned int judge_if_need_destory_media(struct ast_sip_session *session);
static void send_reinvite_control_frame_for_no_sdp(struct ast_sip_session *session);
static void record_replace_channel(struct ast_sip_session *session, pjsip_rx_data *rdata);
static  int check_channel_is_video_call(struct ast_channel *chan);
static void set_decline_reason_for_basic_call(struct ast_sip_session *session,
	pjmedia_sdp_media *media, struct ast_stream *stream);
static void sip_session_defer_termination_stop_timer(struct ast_sip_session *session);
static void set_support_gs_state_flag(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void get_all_local_ip(void);
static void send_peer_reinvite_control_for_callflip(struct ast_sip_session *session);
static void set_session_diff_multiavs_ip_flag_for_caller(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_session_diff_multiavs_ip_flag_for_callee(struct ast_sip_session *session, struct ast_sip_contact *contact,const char *location,char *uri);
static void send_delayed_session_to_peer(struct ast_sip_session *session);

#ifdef BUILD_ON_CLOUD_PBX
static void set_sbc_endpoint_addr(struct ast_sip_session *session, pjsip_rx_data *rdata);
static void set_dest_addrs_to_inv_session(struct ast_sip_session *session);
#endif
/**
 * This structure describes SDP media negotiator.
 */
struct pjmedia_sdp_neg
{
    pjmedia_sdp_neg_state state;            /**< Negotiator state.           */
    pj_bool_t             prefer_remote_codec_order;
    pj_bool_t             answer_with_multiple_codecs;
    pj_bool_t             has_remote_answer;
    pj_bool_t             answer_was_remote;

    pjmedia_sdp_session *initial_sdp,       /**< Initial local SDP           */
                        *initial_sdp_tmp,   /**< Temporary initial local SDP */
                        *active_local_sdp,  /**< Currently active local SDP. */
                        *active_remote_sdp, /**< Currently active remote's.  */
                        *neg_local_sdp,     /**< Temporary local SDP.        */
                        *neg_remote_sdp;    /**< Temporary remote SDP.       */
};

static int hangup_sip2cause(int cause);
/* Added by yfyang, for sca feature */
static const pj_str_t callinfo_str = { "Call-Info", 9 };

/*! \brief Device state strings for printing */
static const struct {
	const char *string;
	enum sca_status state;

} sca_state2string[] = {
	{ "idle", SCA_Idle},
	{ "seized", SCA_Seized },
	{ "progressing", SCA_Progressing},
	{ "alerting", SCA_Alerting},
	{ "active", SCA_Active},
	{ "held", SCA_Held},
	{ "held-private", SCA_Held_Private},
	{ "bridge-active", SCA_Bridge_Active},
	{ "bridge-held", SCA_Bridge_Held},
};

struct ast_parklot_info {	
	char parkext[80];
	unsigned int pos_start;
	unsigned int pos_end;
	AST_LIST_ENTRY(ast_parklot_info) next;
};
AST_RWLIST_HEAD_STATIC(parklot_info_list, ast_parklot_info);

static int msid_count = 0;

#define SOFTBRIDGE_VIDEO_DEST_PREFIX "softbridge_dest"
#define SOFTBRIDGE_VIDEO_DEST_LEN strlen(SOFTBRIDGE_VIDEO_DEST_PREFIX)

/*! \brief All active session on the system */
static AST_LIST_HEAD_STATIC(sip_sessions, ast_sip_session);

static int im_support_whiteboard = 1;

static pj_sockaddr local_address[32];
static unsigned int local_address_count = 0;
static unsigned int get_local_ip_again = 0;

int g_vlan_wan_enable  = 0;
char sip_vlan_addr[48] = {0};
char sip_lan_addr[48]  = {0};
#endif

/*! \brief NAT hook for modifying outgoing messages with SDP */
static struct ast_sip_nat_hook *nat_hook;

/*!
 * \brief Registered SDP stream handlers
 *
 * This container is keyed on stream types. Each
 * object in the container is a linked list of
 * handlers for the stream type.
 */
static struct ao2_container *sdp_handlers;

/*!
 * These are the objects in the sdp_handlers container
 */
struct sdp_handler_list {
	/* The list of handlers to visit */
	AST_LIST_HEAD_NOLOCK(, ast_sip_session_sdp_handler) list;
	/* The handlers in this list handle streams of this type */
	char stream_type[1];
};

static struct pjmedia_sdp_session *create_local_sdp(pjsip_inv_session *inv, struct ast_sip_session *session, const pjmedia_sdp_session *offer);

static int sdp_handler_list_hash(const void *obj, int flags)
{
	const struct sdp_handler_list *handler_list = obj;
	const char *stream_type = flags & OBJ_KEY ? obj : handler_list->stream_type;

	return ast_str_hash(stream_type);
}

static int sdp_handler_list_cmp(void *obj, void *arg, int flags)
{
	struct sdp_handler_list *handler_list1 = obj;
	struct sdp_handler_list *handler_list2 = arg;
	const char *stream_type2 = flags & OBJ_KEY ? arg : handler_list2->stream_type;

	return strcmp(handler_list1->stream_type, stream_type2) ? 0 : CMP_MATCH | CMP_STOP;
}

int ast_sip_session_register_sdp_handler(struct ast_sip_session_sdp_handler *handler, const char *stream_type)
{
	RAII_VAR(struct sdp_handler_list *, handler_list,
			ao2_find(sdp_handlers, stream_type, OBJ_KEY), ao2_cleanup);
	SCOPED_AO2LOCK(lock, sdp_handlers);

	if (handler_list) {
		struct ast_sip_session_sdp_handler *iter;
		/* Check if this handler is already registered for this stream type */
		AST_LIST_TRAVERSE(&handler_list->list, iter, next) {
			if (!strcmp(iter->id, handler->id)) {
				ast_log(LOG_WARNING, "Handler '%s' already registered for stream type '%s'.\n", handler->id, stream_type);
				return -1;
			}
		}
		AST_LIST_INSERT_TAIL(&handler_list->list, handler, next);
		ast_debug(1, "Registered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);

		return 0;
	}

	/* No stream of this type has been registered yet, so we need to create a new list */
	handler_list = ao2_alloc(sizeof(*handler_list) + strlen(stream_type), NULL);
	if (!handler_list) {
		return -1;
	}
	/* Safe use of strcpy */
	strcpy(handler_list->stream_type, stream_type);
	AST_LIST_HEAD_INIT_NOLOCK(&handler_list->list);
	AST_LIST_INSERT_TAIL(&handler_list->list, handler, next);
	if (!ao2_link(sdp_handlers, handler_list)) {
		return -1;
	}
	ast_debug(1, "Registered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);

	return 0;
}

static int remove_handler(void *obj, void *arg, void *data, int flags)
{
	struct sdp_handler_list *handler_list = obj;
	struct ast_sip_session_sdp_handler *handler = data;
	struct ast_sip_session_sdp_handler *iter;
	const char *stream_type = arg;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&handler_list->list, iter, next) {
		if (!strcmp(iter->id, handler->id)) {
			AST_LIST_REMOVE_CURRENT(next);
			ast_debug(1, "Unregistered SDP stream handler '%s' for stream type '%s'\n", handler->id, stream_type);
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	if (AST_LIST_EMPTY(&handler_list->list)) {
		ast_debug(3, "No more handlers exist for stream type '%s'\n", stream_type);
		return CMP_MATCH;
	} else {
		return CMP_STOP;
	}
}

void ast_sip_session_unregister_sdp_handler(struct ast_sip_session_sdp_handler *handler, const char *stream_type)
{
	ao2_callback_data(sdp_handlers, OBJ_KEY | OBJ_UNLINK | OBJ_NODATA, remove_handler, (void *)stream_type, handler);
}


#ifdef GRANDSTREAM_NETWORKS

/*Start added by zyxia for Bug 188323 - Wave_Android_1.0.6.2_Suggest to show only one line after disconnecting the network and then reconnecting during ringing*/
void ast_sip_update_dlg_and_invite_tsx_transport_by_endponit(struct ast_sip_endpoint *endpoint,pj_sockaddr *rem_addr, pjsip_transport *tp)
{
	struct ast_sip_session *session = NULL;
	pjsip_sip_uri *sip_contact = NULL;
	char remote_uri[PJSIP_MAX_URL_SIZE]={0};

	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_TRAVERSE(&sip_sessions, session, next) {
		//check the session is call to gswave  app and the state is ringing 
		/*modify by zyxia for Bug 208588 ,the app is ringing ,refresh the dlg port*/
		if (session->inv_session &&  session->inv_session->dlg && &session->inv_session->dlg->remote.contact &&session->inv_session->state < PJSIP_INV_STATE_CONFIRMED && session->is_gswave_app_call  && session->endpoint && !strcmp(ast_sorcery_object_get_id(session->endpoint),ast_sorcery_object_get_id(endpoint)) && (session->has_add_gs_call_count && (session->has_dec_gs_call_count == 0))){
			sip_contact = NULL;
			memset(remote_uri,0,sizeof(remote_uri));
			sip_contact = pjsip_uri_get_uri(session->inv_session->dlg->remote.contact->uri);
			pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, sip_contact, remote_uri, sizeof(remote_uri));
			if(remote_uri[0] != '\0' && ast_sip_find_waveapp_info_by_uri(endpoint, remote_uri)) {
				ast_sip_dialog_rewrite_remote_uri_and_transport(endpoint,session->inv_session->dlg,rem_addr);
				ast_debug(3, "ast_sip_update_dlg_and_invite_tsx_transport_by_endponit inv_session '%p' tsx:%p,endpoint:%s\n", session->inv_session->dlg,session->inv_session->invite_tsx,ast_sorcery_object_get_id(endpoint));
				pjsip_tsx_update_transport(session->inv_session->invite_tsx,tp);
				ast_sip_update_request_uri_and_dest_info(session->inv_session->invite_tsx,rem_addr);
				break;
			}
		}
	}
	AST_LIST_UNLOCK(&sip_sessions);
}
/*End added by zyxia for Bug 188323 - Wave_Android_1.0.6.2_Suggest to show only one line after disconnecting the network and then reconnecting during ringing*/

AST_RWLIST_HEAD_STATIC(exten_supplements, ast_sip_exten_supplement);

int ast_sip_exten_register_supplement(struct ast_sip_exten_supplement *supplement)
{
	if (!supplement->find_exten) {
		return -1;
	}

	SCOPED_LOCK(lock, &exten_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&exten_supplements, supplement, next); 
	ast_module_ref(ast_module_info->self);

	return 0;
}

void ast_sip_exten_unregister_supplement(struct ast_sip_exten_supplement *supplement)
{
	struct ast_sip_exten_supplement *iter;
	SCOPED_LOCK(lock, &exten_supplements, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&exten_supplements, iter, next) {
		if (supplement == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}

static int find_sub_handler_for_supplement_exten(enum ast_sip_supplement_type type, char *exten, void *data)
{
	struct ast_sip_exten_supplement *iter;

	SCOPED_LOCK(lock, &exten_supplements, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&exten_supplements, iter, next) {
		if (iter->exten_type != type) {
			continue;
		}

		if (0 == iter->find_exten(exten, data)) {
			return 0;
		}
	}
	return -1;
}


int ast_sip_find_sub_handler_for_supplement_exten(enum ast_sip_supplement_type type, char *exten, void *data)
{
    return find_sub_handler_for_supplement_exten(type, exten, data);
}

/* Added by yfyang, for sca feature */
AST_RWLIST_HEAD_STATIC(sca_handlers, ast_sip_session_sca_handler);

int ast_sip_session_register_sca_handler(struct ast_sip_session_sca_handler *sca_handler)
{
	if (!sca_handler) {
		return -1;
	}

	SCOPED_LOCK(lock, &sca_handlers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_INSERT_TAIL(&sca_handlers, sca_handler, next); 
	ast_module_ref(ast_module_info->self);

	return 0;
}

void ast_sip_session_unregister_sca_handler(struct ast_sip_session_sca_handler *sca_handler)
{
	struct ast_sip_session_sca_handler *iter;

	SCOPED_LOCK(lock, &sca_handlers, AST_RWLIST_WRLOCK, AST_RWLIST_UNLOCK);
	AST_RWLIST_TRAVERSE_SAFE_BEGIN(&sca_handlers, iter, next) {
		if (sca_handler == iter) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_module_unref(ast_module_info->self);
			break;
		}
	}
	AST_RWLIST_TRAVERSE_SAFE_END;
}


static struct ast_sip_session_sca_handler *find_sca_handler(void)
{
	struct ast_sip_session_sca_handler *iter;

	SCOPED_LOCK(lock, &sca_handlers, AST_RWLIST_RDLOCK, AST_RWLIST_UNLOCK);

	AST_RWLIST_TRAVERSE(&sca_handlers, iter, next) {
		return iter;
	}

	return NULL;
}

void ast_sip_set_used_gs_call_count(struct ast_sip_session *session,pjsip_rx_data *rdata,int flag)
{
	char *gs_appid_tmp = NULL;
	pjsip_generic_string_hdr *gs_appid_hdr;
	static const pj_str_t GS_APPID = { "GS-APPID", 8 };
	if (!session->endpoint)
		return;

	gs_appid_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &GS_APPID, NULL);
	if (gs_appid_hdr) {
		size_t alloc_size = pj_strlen(&gs_appid_hdr->hvalue) + 1;
		gs_appid_tmp = ast_alloca(alloc_size);
		ast_copy_pj_str(gs_appid_tmp, &gs_appid_hdr->hvalue, alloc_size);
		if (!ast_strlen_zero(gs_appid_tmp)) {
			if (flag) {
				session->has_add_gs_call_count = 1;
				ast_log(LOG_WARNING, "Endpoint[%s] Recv Invite or 180,183,from gswave\n",session->endpoint->aors);
			}
			else if (session->has_add_gs_call_count) {
				session->has_dec_gs_call_count = 1;
				ast_log(LOG_WARNING, "Endpoint[%s] Recv Bye or Cancel or 486, 408 from gswave\n",session->endpoint->aors);
			}
		}
	}
	return;
}

void ast_sip_notify_avs_to_make_early_bridge(struct ast_sip_session *session)
{
	if (!session || !session->channel) {
		return;
	}

	if (!session->need_notify_avs_early_media) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] notify_avs_to_make_early_bridge begin for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_queue_control(session->channel, AST_CONTROL_EARLY_MEDIA);

	session->need_notify_avs_early_media = 0;

	return;
}

#endif

static struct ast_sip_session_media_state *internal_sip_session_media_state_alloc(
	size_t sessions, size_t read_callbacks)
{
	struct ast_sip_session_media_state *media_state;

	media_state = ast_calloc(1, sizeof(*media_state));
	if (!media_state) {
		return NULL;
	}

	if (AST_VECTOR_INIT(&media_state->sessions, sessions) < 0) {
		ast_free(media_state);
		return NULL;
	}

	if (AST_VECTOR_INIT(&media_state->read_callbacks, read_callbacks) < 0) {
		AST_VECTOR_FREE(&media_state->sessions);
		ast_free(media_state);
		return NULL;
	}

	return media_state;
}

struct ast_sip_session_media_state *ast_sip_session_media_state_alloc(void)
{
	return internal_sip_session_media_state_alloc(
		DEFAULT_NUM_SESSION_MEDIA, DEFAULT_NUM_SESSION_MEDIA);
}

void ast_sip_session_media_state_reset(struct ast_sip_session_media_state *media_state)
{
	int index;

	if (!media_state) {
		return;
	}

	AST_VECTOR_RESET(&media_state->sessions, ao2_cleanup);
	AST_VECTOR_RESET(&media_state->read_callbacks, AST_VECTOR_ELEM_CLEANUP_NOOP);

	for (index = 0; index < AST_MEDIA_TYPE_END; ++index) {
		media_state->default_session[index] = NULL;
	}

	ast_stream_topology_free(media_state->topology);
	media_state->topology = NULL;
}

struct ast_sip_session_media_state *ast_sip_session_media_state_clone(const struct ast_sip_session_media_state *media_state)
{
	struct ast_sip_session_media_state *cloned;
	int index;

	if (!media_state) {
		return NULL;
	}

	cloned = internal_sip_session_media_state_alloc(
		AST_VECTOR_SIZE(&media_state->sessions),
		AST_VECTOR_SIZE(&media_state->read_callbacks));
	if (!cloned) {
		return NULL;
	}

	if (media_state->topology) {
		cloned->topology = ast_stream_topology_clone(media_state->topology);
		if (!cloned->topology) {
			ast_sip_session_media_state_free(cloned);
			return NULL;
		}
	}

	for (index = 0; index < AST_VECTOR_SIZE(&media_state->sessions); ++index) {
		struct ast_sip_session_media *session_media = AST_VECTOR_GET(&media_state->sessions, index);
		enum ast_media_type type = ast_stream_get_type(ast_stream_topology_get_stream(cloned->topology, index));

		ao2_bump(session_media);
		if (AST_VECTOR_REPLACE(&cloned->sessions, index, session_media)) {
			ao2_cleanup(session_media);
		}
		if (ast_stream_get_state(ast_stream_topology_get_stream(cloned->topology, index)) != AST_STREAM_STATE_REMOVED &&
			!cloned->default_session[type]) {
			cloned->default_session[type] = session_media;
		}
	}

	for (index = 0; index < AST_VECTOR_SIZE(&media_state->read_callbacks); ++index) {
		struct ast_sip_session_media_read_callback_state *read_callback = AST_VECTOR_GET_ADDR(&media_state->read_callbacks, index);

		AST_VECTOR_REPLACE(&cloned->read_callbacks, index, *read_callback);
	}

	return cloned;
}

void ast_sip_session_media_state_free(struct ast_sip_session_media_state *media_state)
{
	if (!media_state) {
		return;
	}

	/* This will reset the internal state so we only have to free persistent things */
	ast_sip_session_media_state_reset(media_state);

	AST_VECTOR_FREE(&media_state->sessions);
	AST_VECTOR_FREE(&media_state->read_callbacks);

	ast_free(media_state);
}

int ast_sip_session_is_pending_stream_default(const struct ast_sip_session *session, const struct ast_stream *stream)
{
	int index;

	if (!session->pending_media_state->topology) {
		ast_log(LOG_WARNING, "Pending topology was NULL for channel '%s'\n",
			session->channel ? ast_channel_name(session->channel) : "unknown");
		return 0;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
		return 0;
	}

	for (index = 0; index < ast_stream_topology_get_count(session->pending_media_state->topology); ++index) {
		if (ast_stream_get_type(ast_stream_topology_get_stream(session->pending_media_state->topology, index)) !=
			ast_stream_get_type(stream)) {
			continue;
		}

		return ast_stream_topology_get_stream(session->pending_media_state->topology, index) == stream ? 1 : 0;
	}

	return 0;
}

int ast_sip_session_media_add_read_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	int fd, ast_sip_session_media_read_cb callback)
{
	struct ast_sip_session_media_read_callback_state callback_state = {
		.fd = fd,
		.read_callback = callback,
		.session = session_media,
	};

	/* The contents of the vector are whole structs and not pointers */
	return AST_VECTOR_APPEND(&session->pending_media_state->read_callbacks, callback_state);
}

int ast_sip_session_media_set_write_callback(struct ast_sip_session *session, struct ast_sip_session_media *session_media,
	ast_sip_session_media_write_cb callback)
{
	if (session_media->write_callback) {
		if (session_media->write_callback == callback) {
			return 0;
		}

		return -1;
	}

	session_media->write_callback = callback;

	return 0;
}

struct ast_sip_session_media *ast_sip_session_media_get_transport(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	int index;

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_sip_session_media_use_bundle(session) || ast_strlen_zero(session_media->mid))
#else
	if (!session->endpoint->media.bundle || ast_strlen_zero(session_media->mid))
#endif
	{
		return session_media;
	}

	for (index = 0; index < AST_VECTOR_SIZE(&session->pending_media_state->sessions); ++index) {
		struct ast_sip_session_media *bundle_group_session_media;

		bundle_group_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);

		/* The first session which is in the bundle group is considered the authoritative session for transport */
		if (bundle_group_session_media->bundle_group == session_media->bundle_group) {
			return bundle_group_session_media;
		}
	}

	return session_media;
}

/*!
 * \brief Set an SDP stream handler for a corresponding session media.
 *
 * \note Always use this function to set the SDP handler for a session media.
 *
 * This function will properly free resources on the SDP handler currently being
 * used by the session media, then set the session media to use the new SDP
 * handler.
 */
static void session_media_set_handler(struct ast_sip_session_media *session_media,
		struct ast_sip_session_sdp_handler *handler)
{
	ast_assert(session_media->handler != handler);

	if (session_media->handler) {
		session_media->handler->stream_destroy(session_media);
	}
	session_media->handler = handler;
}

#ifdef GRANDSTREAM_NETWORKS
/* to check if the exten is a park pos
   ret 0: not a park pos
   ret 1: is a park pos
 */
int ast_sip_park_pos_check(char* exten)
{
	int extennum = 0;
	struct ast_parklot_info *parklot = NULL;

	if (exten == NULL || exten[0] == '0') {
		return 0;
	}

	extennum = atoi(exten);
	AST_RWLIST_TRAVERSE(&parklot_info_list, parklot, next) {
		if (parklot && (extennum >= parklot->pos_start) && (extennum <= parklot->pos_end)) {
			return 1;
		}
	}
	return 0;
}

static void session_park_pos_get(int reload)
{
	struct ast_config *cfg;
	struct ast_flags config_flags = { reload ? CONFIG_FLAG_FILEUNCHANGED : 0 };
	const char *cat = NULL;
	const char *pos = NULL;
	int pos_start = 0;
	int pos_end = 0;
	int res = 0;
	struct ast_parklot_info *parklot = NULL;

	cfg = ast_config_load("res_parking.conf", config_flags);

	if (cfg == CONFIG_STATUS_FILEUNCHANGED || NULL == cfg) {
		return;
	} 

	/*remove old parklot */
	if (reload) {
		AST_RWLIST_TRAVERSE_SAFE_BEGIN(&parklot_info_list, parklot, next) {
			AST_RWLIST_REMOVE_CURRENT(next);
			ast_free(parklot);
		}
		AST_RWLIST_TRAVERSE_SAFE_END;
	}

	while ((cat = ast_category_browse(cfg, cat)) && !res) {
		if (!(pos = ast_variable_retrieve(cfg, cat, "parkpos"))) {
			continue;
		}

		pos_start = pos_end = 0;
		sscanf(pos, "%30d-%30d", &pos_start, &pos_end);

		if ((parklot = ast_calloc(1, sizeof(*parklot)))) {
			strncpy(parklot->parkext, cat, sizeof(parklot->parkext) -1);
			parklot->pos_start = pos_start;
			parklot->pos_end = pos_end;

			ast_log(LOG_NOTICE, "Parkext[%s], parkpos[%d]-[%d]", cat, pos_start, pos_end);
			AST_RWLIST_INSERT_TAIL(&parklot_info_list, parklot, next);
		}
	}

	ast_config_destroy(cfg);
}
#endif

static int stream_destroy(void *obj, void *arg, int flags)
{
	struct sdp_handler_list *handler_list = obj;
	struct ast_sip_session_media *session_media = arg;
	struct ast_sip_session_sdp_handler *handler;

	AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
		handler->stream_destroy(session_media);
	}

	return 0;
}

static void session_media_dtor(void *obj)
{
	struct ast_sip_session_media *session_media = obj;

	/* It is possible for multiple handlers to have allocated memory on the
	 * session media (usually through a stream changing types). Therefore, we
	 * traverse all the SDP handlers and let them all call stream_destroy on
	 * the session_media
	 */
	ao2_callback(sdp_handlers, 0, stream_destroy, session_media);

	if (session_media->srtp) {
		ast_sdp_srtp_destroy(session_media->srtp);
	}

	ast_free(session_media->mid);
	ast_free(session_media->remote_mslabel);
}

struct ast_sip_session_media *ast_sip_session_media_state_add(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state, enum ast_media_type type, int position)
{
	struct ast_sip_session_media *session_media = NULL;

	/* It is possible for this media state to already contain a session for the stream. If this
	 * is the case we simply return it.
	 */
	if (position < AST_VECTOR_SIZE(&media_state->sessions)) {
		return AST_VECTOR_GET(&media_state->sessions, position);
	}

	/* Determine if we can reuse the session media from the active media state if present */
	if (position < AST_VECTOR_SIZE(&session->active_media_state->sessions)) {
		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, position);
		/* A stream can never exist without an accompanying media session */
		if (session_media->type == type) {
			ao2_ref(session_media, +1);
		} else {
			session_media = NULL;
		}
	}

	if (!session_media) {
		/* No existing media session we can use so create a new one */
		session_media = ao2_alloc_options(sizeof(*session_media), session_media_dtor, AO2_ALLOC_OPT_LOCK_NOLOCK);
		if (!session_media) {
			return NULL;
		}

		session_media->encryption = session->endpoint->media.rtp.encryption;
		session_media->remote_ice = session->endpoint->media.rtp.ice_support;
#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu for rtcp_mux only used for webrtc. 2019.04.03
		 * Todo: If sip phone support rtcp_mux need to change again.
		 */
		session_media->remote_rtcp_mux = session->endpoint->media.rtcp_mux && (session->is_webrtc_call || session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK);
#else
		session_media->remote_rtcp_mux = session->endpoint->media.rtcp_mux;
#endif
		session_media->keepalive_sched_id = -1;
		session_media->timeout_sched_id = -1;
		session_media->type = type;
		session_media->stream_num = position;

#ifdef GRANDSTREAM_NETWORKS
		set_session_media_mslabel(session, session_media);
#endif

		if (session->endpoint->media.bundle) {
			/* This is a new stream so create a new mid based on media type and position, which makes it unique.
			 * If this is the result of an offer the mid will just end up getting replaced.
			 */
			if (ast_asprintf(&session_media->mid, "%s-%d", ast_codec_media_type2str(type), position) < 0) {
				ao2_ref(session_media, -1);
				return NULL;
			}
			session_media->bundle_group = 0;

			/* Some WebRTC clients can't handle an offer to bundle media streams. Instead they expect them to
			 * already be bundled. Every client handles this scenario though so if WebRTC is enabled just go
			 * ahead and treat the streams as having already been bundled.
			 */
			session_media->bundled = session->endpoint->media.webrtc;
		} else {
			session_media->bundle_group = -1;
		}
	}

	if (AST_VECTOR_REPLACE(&media_state->sessions, position, session_media)) {
		ao2_ref(session_media, -1);

		return NULL;
	}

	/* If this stream will be active in some way and it is the first of this type then consider this the default media session to match */
	if (!media_state->default_session[type] && ast_stream_get_state(ast_stream_topology_get_stream(media_state->topology, position)) != AST_STREAM_STATE_REMOVED) {
		media_state->default_session[type] = session_media;
	}

	return session_media;
}

static int is_stream_limitation_reached(enum ast_media_type type, const struct ast_sip_endpoint *endpoint, int *type_streams)
{
	switch (type) {
	case AST_MEDIA_TYPE_AUDIO:
		return !(type_streams[type] < endpoint->media.max_audio_streams);
	case AST_MEDIA_TYPE_VIDEO:
		return !(type_streams[type] < endpoint->media.max_video_streams);
	case AST_MEDIA_TYPE_IMAGE:
		/* We don't have an option for image (T.38) streams so cap it to one. */
		return (type_streams[type] > 0);
#ifdef GRANDSTREAM_NETWORKS
	case AST_MEDIA_TYPE_APPLICATION:
		/* Add by bfdu, we don't limit application media */
		return 0;
#endif
	case AST_MEDIA_TYPE_UNKNOWN:
	case AST_MEDIA_TYPE_TEXT:
	default:
		/* We don't want any unknown or "other" streams on our endpoint,
		 * so always just say we've reached the limit
		 */
		return 1;
	}
}

static int get_mid_bundle_group(const pjmedia_sdp_session *sdp, const char *mid)
{
	int bundle_group = 0;
	int index;

	for (index = 0; index < sdp->attr_count; ++index) {
		pjmedia_sdp_attr *attr = sdp->attr[index];
		char value[pj_strlen(&attr->value) + 1], *mids = value, *attr_mid;

		if (pj_strcmp2(&attr->name, "group") || pj_strncmp2(&attr->value, "BUNDLE", 6)) {
			continue;
		}

		ast_copy_pj_str(value, &attr->value, sizeof(value));

		/* Skip the BUNDLE at the front */
		mids += 7;

		while ((attr_mid = strsep(&mids, " "))) {
			if (!strcmp(attr_mid, mid)) {
				/* The ordering of attributes determines our internal identification of the bundle group based on number,
				 * with -1 being not in a bundle group. Since this is only exposed internally for response purposes it's
				 * actually even fine if things move around.
				 */
				return bundle_group;
			}
		}

		bundle_group++;
	}

	return -1;
}

static int set_mid_and_bundle_group(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	const pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr;

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_sip_session_media_use_bundle(session))
#else
	if (!session->endpoint->media.bundle)
#endif
	{
		return 0;
	}

	/* By default on an incoming negotiation we assume no mid and bundle group is present */
	ast_free(session_media->mid);
	session_media->mid = NULL;
	session_media->bundle_group = -1;
	session_media->bundled = 0;

	/* Grab the media identifier for the stream */
	attr = pjmedia_sdp_media_find_attr2(stream, "mid", NULL);
	if (!attr) {
		return 0;
	}

	session_media->mid = ast_calloc(1, attr->value.slen + 1);
	if (!session_media->mid) {
		return 0;
	}
	ast_copy_pj_str(session_media->mid, &attr->value, attr->value.slen + 1);

	/* Determine what bundle group this is part of */
	session_media->bundle_group = get_mid_bundle_group(sdp, session_media->mid);

	/* If this is actually part of a bundle group then the other side requested or accepted the bundle request */
	session_media->bundled = session_media->bundle_group != -1;

	return 0;
}

static void set_remote_mslabel_and_stream_group(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media,
	const pjmedia_sdp_session *sdp,
	const struct pjmedia_sdp_media *stream,
	struct ast_stream *asterisk_stream)
{
	int index;

	ast_free(session_media->remote_mslabel);
	session_media->remote_mslabel = NULL;

	for (index = 0; index < stream->attr_count; ++index) {
		pjmedia_sdp_attr *attr = stream->attr[index];
		char attr_value[pj_strlen(&attr->value) + 1];
		char *ssrc_attribute_name, *ssrc_attribute_value = NULL;
		char *msid, *tmp = attr_value;
		static const pj_str_t STR_msid = { "msid", 4 };
		static const pj_str_t STR_ssrc = { "ssrc", 4 };

		if (!pj_strcmp(&attr->name, &STR_msid)) {
			ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));
			msid = strsep(&tmp, " ");
			session_media->remote_mslabel = ast_strdup(msid);
			break;
		} else if (!pj_strcmp(&attr->name, &STR_ssrc)) {
			ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));

			if ((ssrc_attribute_name = strchr(attr_value, ' '))) {
				/* This has an actual attribute */
				*ssrc_attribute_name++ = '\0';
				ssrc_attribute_value = strchr(ssrc_attribute_name, ':');
				if (ssrc_attribute_value) {
					/* Values are actually optional according to the spec */
					*ssrc_attribute_value++ = '\0';
				}

				if (!strcasecmp(ssrc_attribute_name, "mslabel") && !ast_strlen_zero(ssrc_attribute_value)) {
					session_media->remote_mslabel = ast_strdup(ssrc_attribute_value);
					break;
				}
			}
		}
	}

	if (ast_strlen_zero(session_media->remote_mslabel)) {
		return;
	}

	/* Iterate through the existing streams looking for a match and if so then group this with it */
	for (index = 0; index < AST_VECTOR_SIZE(&session->pending_media_state->sessions); ++index) {
		struct ast_sip_session_media *group_session_media;

		group_session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);

		if (ast_strlen_zero(group_session_media->remote_mslabel) ||
			strcmp(group_session_media->remote_mslabel, session_media->remote_mslabel)) {
			continue;
		}

		ast_stream_set_group(asterisk_stream, index);
		break;
	}
}

static void remove_stream_from_bundle(struct ast_sip_session_media *session_media,
	struct ast_stream *stream)
{
	ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
#ifndef GRANDSTREAM_NETWORKS
	ast_free(session_media->mid);
	session_media->mid = NULL;
#endif
	session_media->bundle_group = -1;
	session_media->bundled = 0;
}

static int handle_incoming_sdp(struct ast_sip_session *session, const pjmedia_sdp_session *sdp)
{
	int i;
	int handled = 0;
	int type_streams[AST_MEDIA_TYPE_END] = {0};

	if (session->inv_session && session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Failed to handle incoming SDP. Session has been already disconnected\n");
		return -1;
	}

	/* It is possible for SDP deferral to have already created a pending topology */
	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_alloc();
		if (!session->pending_media_state->topology) {
			return -1;
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	set_session_remote_support_bundle(session, sdp);
#endif

	for (i = 0; i < sdp->media_count; ++i) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
		struct ast_sip_session_media *session_media = NULL;
		int res;
		enum ast_media_type type;
		struct ast_stream *stream = NULL;
		pjmedia_sdp_media *remote_stream = sdp->media[i];

		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &sdp->media[i]->desc.media, sizeof(media));
		type = ast_media_type_from_str(media);

#ifdef GRANDSTREAM_NETWORKS
		enum ast_media_sub_type sub_type = ast_sip_session_get_media_sub_type(session, remote_stream);
#endif

		/* See if we have an already existing stream, which can occur from SDP deferral checking */
		if (i < ast_stream_topology_get_count(session->pending_media_state->topology)) {
			stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		}
#ifdef GRANDSTREAM_NETWORKS
		struct ast_stream *existing_stream = NULL;

		if (session->active_media_state->topology &&
			(i < ast_stream_topology_get_count(session->active_media_state->topology))) {
			existing_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
			if (sub_type != ast_stream_get_sub_type(existing_stream)) {
				ast_log(LOG_DEBUG, "Sub type is not the same, can' get existing_stream\n");
				existing_stream = NULL;
			}
		}
#endif
		if (!stream) {
#ifndef GRANDSTREAM_NETWORKS
			struct ast_stream *existing_stream = NULL;

			if (session->active_media_state->topology &&
				(i < ast_stream_topology_get_count(session->active_media_state->topology))) {
				existing_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
			}
#endif

#ifdef GRANDSTREAM_NETWORKS
			/* Maybe we receive SDP is not according as previous order, if can't match sub_type, just use default name */
			stream = ast_stream_alloc((existing_stream && (ast_stream_get_sub_type(existing_stream) == sub_type)) ? 
				ast_stream_get_name(existing_stream) : ast_codec_media_type2str(type), type);
#else
			stream = ast_stream_alloc(existing_stream ? ast_stream_get_name(existing_stream) : ast_codec_media_type2str(type), type);
#endif
			if (!stream) {
				return -1;
			}
			if (ast_stream_topology_set_stream(session->pending_media_state->topology, i, stream)) {
				ast_stream_free(stream);
				return -1;
			}
		}

#ifdef GRANDSTREAM_NETWORKS
		ast_stream_set_sub_type(stream, sub_type);
		if (existing_stream) {
			set_stream_params_according_exist_stream(stream, existing_stream);
		}

		session_media = ast_sip_session_media_state_add_as_sub_type_and_msid(session, session->pending_media_state, stream, i);
#else
		session_media = ast_sip_session_media_state_add(session, session->pending_media_state, ast_media_type_from_str(media), i);
#endif
		if (!session_media) {
			return -1;
		}

		/* If this stream is already declined mark it as such, or mark it as such if we've reached the limit */
#ifdef GRANDSTREAM_NETWORKS
		if (ast_sip_session_judge_if_disable_media(remote_stream) || is_stream_limitation_reached(type, session->endpoint, type_streams))
#else
		if (!remote_stream->desc.port || is_stream_limitation_reached(type, session->endpoint, type_streams))
#endif
		{
			ast_debug(1, "Declining incoming SDP media stream '%s' at position '%d'\n",
				ast_codec_media_type2str(type), i);
			remove_stream_from_bundle(session_media, stream);
#ifdef GRANDSTREAM_NETWORKS
			ast_sip_session_set_media_sub_type(session, remote_stream, session_media, stream);
			set_stream_content(stream, sdp->media[i], session_media->type);
			ast_stream_set_local_media_direction(stream, UNKNOW);
			if ((pjmedia_sdp_neg_get_state(session->inv_session->neg) == PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER) &&
				!remote_stream->desc.port && (session->inv_session->state == PJSIP_INV_STATE_CONFIRMED) && session_media->remote_port) {
				ast_log(LOG_DEBUG, "Negotiating incoming SDP, remote media port:%d change to zero\n", session_media->remote_port);
				session_media->changed = 1;
				session_media->remote_port = 0;
				if (session->is_sfu) {
					session_media->close_camera = 1;
				}
			}
#endif
			continue;
		}

		set_mid_and_bundle_group(session, session_media, sdp, remote_stream);
		set_remote_mslabel_and_stream_group(session, session_media, sdp, remote_stream, stream);

		if (session_media->handler) {
			handler = session_media->handler;
			ast_debug(1, "Negotiating incoming SDP media stream '%s' using %s SDP handler\n",
				ast_codec_media_type2str(session_media->type),
				session_media->handler->id);
			res = handler->negotiate_incoming_sdp_stream(session, session_media, sdp, i, stream);
			if (res < 0) {
				/* Catastrophic failure. Abort! */
				return -1;
			} else if (res == 0) {
#ifdef GRANDSTREAM_NETWORKS
				if (AST_MEDIA_TYPE_AUDIO == type) {
					ast_log(LOG_NOTICE, "[session:%p] no audio codecs available for endpoint:%s!",
						session, ast_sorcery_object_get_id(session->endpoint));
					return -1;
				}
#endif
				ast_debug(1, "Declining incoming SDP media stream '%s' at position '%d'\n",
					ast_codec_media_type2str(type), i);
				remove_stream_from_bundle(session_media, stream);
				continue;
			} else if (res > 0) {
				ast_debug(1, "Media stream '%s' handled by %s\n",
					ast_codec_media_type2str(session_media->type),
					session_media->handler->id);
				/* Handled by this handler. Move to the next stream */
				handled = 1;
				++type_streams[type];
				continue;
			}
		}

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler == session_media->handler) {
				continue;
			}
			ast_debug(1, "Negotiating incoming SDP media stream '%s' using %s SDP handler\n",
				ast_codec_media_type2str(session_media->type),
				handler->id);
			res = handler->negotiate_incoming_sdp_stream(session, session_media, sdp, i, stream);
			if (res < 0) {
				/* Catastrophic failure. Abort! */
				return -1;
			} else if (res == 0) {
#ifdef GRANDSTREAM_NETWORKS
				if (AST_MEDIA_TYPE_AUDIO == type) {
					ast_log(LOG_NOTICE, "no audio codecs available!");
					return -1;
				}
#endif
				ast_debug(1, "Declining incoming SDP media stream '%s' at position '%d'\n",
					ast_codec_media_type2str(type), i);
				remove_stream_from_bundle(session_media, stream);
				continue;
			} else if (res > 0) {
				ast_debug(1, "Media stream '%s' handled by %s\n",
					ast_codec_media_type2str(session_media->type),
					handler->id);
				/* Handled by this handler. Move to the next stream */
				session_media_set_handler(session_media, handler);
				handled = 1;
				++type_streams[type];
				break;
			}
		}
	}
	if (!handled) {
		return -1;
	}
	return 0;
}

static int handle_negotiated_sdp_session_media(struct ast_sip_session_media *session_media,
		struct ast_sip_session *session, const pjmedia_sdp_session *local,
		const pjmedia_sdp_session *remote, int index, struct ast_stream *asterisk_stream)
{
	/* See if there are registered handlers for this media stream type */
	struct pjmedia_sdp_media *local_stream = local->media[index];
	char media[20];
	struct ast_sip_session_sdp_handler *handler;
	RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
	int res;

	/* For backwards compatibility we only reflect the stream state correctly on
	 * the non-default streams. This is because the stream state is also used for
	 * signaling that someone has placed us on hold. This situation is not handled
	 * currently and can result in the remote side being sort of placed on hold too.
	 */
#ifdef GRANDSTREAM_NETWORKS
	/* Determine the state of the stream based on our local SDP */
	if (pjmedia_sdp_media_find_attr2(local_stream, "sendonly", NULL)) {
		ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_SENDONLY);
	} else if (pjmedia_sdp_media_find_attr2(local_stream, "recvonly", NULL)) {
		ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_RECVONLY);
	} else if (pjmedia_sdp_media_find_attr2(local_stream, "inactive", NULL)) {
		ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_INACTIVE);
	} else {
		ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_SENDRECV);
	}
#else
	if (!ast_sip_session_is_pending_stream_default(session, asterisk_stream)) {
		/* Determine the state of the stream based on our local SDP */
		if (pjmedia_sdp_media_find_attr2(local_stream, "sendonly", NULL)) {
			ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_SENDONLY);
		} else if (pjmedia_sdp_media_find_attr2(local_stream, "recvonly", NULL)) {
			ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_RECVONLY);
		} else if (pjmedia_sdp_media_find_attr2(local_stream, "inactive", NULL)) {
			ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_INACTIVE);
		} else {
			ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_SENDRECV);
		}
	} else {
		ast_stream_set_state(asterisk_stream, AST_STREAM_STATE_SENDRECV);
	}
#endif

	/* We need a null-terminated version of the media string */
	ast_copy_pj_str(media, &local->media[index]->desc.media, sizeof(media));

	set_mid_and_bundle_group(session, session_media, remote, remote->media[index]);
	set_remote_mslabel_and_stream_group(session, session_media, remote, remote->media[index], asterisk_stream);

	handler = session_media->handler;
	if (handler) {
		ast_debug(1, "Applying negotiated SDP media stream '%s' using %s SDP handler\n",
			ast_codec_media_type2str(session_media->type),
			handler->id);
		res = handler->apply_negotiated_sdp_stream(session, session_media, local, remote, index, asterisk_stream);
		if (res >= 0) {
			ast_debug(1, "Applied negotiated SDP media stream '%s' using %s SDP handler\n",
				ast_codec_media_type2str(session_media->type),
				handler->id);
			return 0;
		}
		return -1;
	}

	handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
	if (!handler_list) {
		ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
		return -1;
	}
	AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
		if (handler == session_media->handler) {
			continue;
		}
		ast_debug(1, "Applying negotiated SDP media stream '%s' using %s SDP handler\n",
			ast_codec_media_type2str(session_media->type),
			handler->id);
		res = handler->apply_negotiated_sdp_stream(session, session_media, local, remote, index, asterisk_stream);
		if (res < 0) {
			/* Catastrophic failure. Abort! */
			return -1;
		}
		if (res > 0) {
			ast_debug(1, "Applied negotiated SDP media stream '%s' using %s SDP handler\n",
				ast_codec_media_type2str(session_media->type),
				handler->id);
			/* Handled by this handler. Move to the next stream */
			session_media_set_handler(session_media, handler);
			return 0;
		}
	}

	if (session_media->handler && session_media->handler->stream_stop) {
		ast_debug(1, "Stopping SDP media stream '%s' as it is not currently negotiated\n",
			ast_codec_media_type2str(session_media->type));
		session_media->handler->stream_stop(session_media);
	}

	return 0;
}

static int handle_negotiated_sdp(struct ast_sip_session *session, const pjmedia_sdp_session *local, const pjmedia_sdp_session *remote)
{
	int i;
	struct ast_stream_topology *topology;
	unsigned int changed = 0;

	if (!session->pending_media_state->topology) {
		if (session->active_media_state->topology) {
			/*
			 * This happens when we have negotiated media after receiving a 183,
			 * and we're now receiving a 200 with a new SDP.  In this case, there
			 * is active_media_state, but the pending_media_state has been reset.
			 */
			struct ast_sip_session_media_state *active_media_state_clone;

			active_media_state_clone =
				ast_sip_session_media_state_clone(session->active_media_state);
			if (!active_media_state_clone) {
				ast_log(LOG_WARNING, "Unable to clone active media state for channel '%s'\n",
					session->channel ? ast_channel_name(session->channel) : "unknown");
				return -1;
			}

			ast_sip_session_media_state_free(session->pending_media_state);
			session->pending_media_state = active_media_state_clone;
		} else {
			ast_log(LOG_WARNING, "No pending or active media state for channel '%s'\n",
				session->channel ? ast_channel_name(session->channel) : "unknown");
			return -1;
		}
	}

	/* If we're handling negotiated streams, then we should already have set
	 * up session media instances (and Asterisk streams) that correspond to
	 * the local SDP, and there should be the same number of session medias
	 * and streams as there are local SDP streams
	 */
	if (ast_stream_topology_get_count(session->pending_media_state->topology) != local->media_count
		|| AST_VECTOR_SIZE(&session->pending_media_state->sessions) != local->media_count) {
		ast_log(LOG_WARNING, "Local SDP for channel '%s' contains %d media streams while we expected it to contain %u\n",
			session->channel ? ast_channel_name(session->channel) : "unknown",
			ast_stream_topology_get_count(session->pending_media_state->topology), local->media_count);
		return -1;
	}

#ifdef GRANDSTREAM_NETWORKS
	// Bug 187062
	if (local->media_count == 1) {
		ast_log(LOG_NOTICE, "Send AST_CONTROL_UPDATE_VIDEO_PORT_ZERO frame to channel:%s\n", ast_channel_name(session->channel));
		ast_queue_control(session->channel, AST_CONTROL_UPDATE_VIDEO_PORT_ZERO);
	}

	set_session_remote_support_bundle(session, remote);
#endif

	for (i = 0; i < local->media_count; ++i) {
		struct ast_sip_session_media *session_media;
		struct ast_stream *stream;

		if (!remote->media[i]) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);


#ifdef GRANDSTREAM_NETWORKS
		set_decline_reason_for_basic_call(session, remote->media[i], stream);
#endif

		/* Make sure that this stream is in the correct state. If we need to change
		 * the state to REMOVED, then our work here is done, so go ahead and move on
		 * to the next stream.
		 */
#ifdef GRANDSTREAM_NETWORKS
		/* Modify by bfdu, for we may decline m line, 2019.07.11 */
		if (ast_sip_session_judge_if_disable_media(remote->media[i]) || ast_sip_session_judge_if_disable_media(local->media[i]))
#else
		if (!remote->media[i]->desc.port)
#endif
		{
			ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
#ifdef GRANDSTREAM_NETWORKS
			remove_stream_from_bundle(session_media, stream);
			ast_stream_set_local_media_direction(stream, UNKNOW);//added by zyxia for Bug 217208,2022.3.7

			if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
				ast_log(LOG_NOTICE, "Send AST_CONTROL_UPDATE_VIDEO_PORT_ZERO frame to channel:%s\n", ast_channel_name(session->channel));
				ast_queue_control(session->channel, AST_CONTROL_UPDATE_VIDEO_PORT_ZERO);
			}

			if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP) {
				ast_log(LOG_DEBUG, "[session:%p] BFCP UDP is disabled, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				session->bfcp_para.has_udp_bfcp = 0;
			}

			if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
				ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
				ast_log(LOG_DEBUG, "[session:%p] BFCP is disabled, clear send bfcp hello flag, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				session->sent_bfcp_hello = 0;
			}

			if (session->channel && session_media->rtp) {
				ast_log(LOG_DEBUG, "[session:%p] Negotiated SDP is complete, remote media port is zero, notify AVS, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				ast_rtp_instance_set_transmode(session_media->rtp, AST_STREAM_STATE_REMOVED);
				//edit by zyxia for Bug 186023 Bug 185971 ,2021.1.15
				if(session->wSwitchImageInvite488ToAudioInvite == 1 || ast_channel_get_is_send_fax(session->channel) || ast_channel_get_is_t30_vfax(session->channel)) {
					const char* faxmode =pbx_builtin_getvar_helper(session->channel,"FAX_MODE");
					ast_log(LOG_NOTICE, "handle_negotiated_sdp channel name:%s,fax mode:%s\n",ast_channel_name(session->channel),faxmode);
					if(!ast_strlen_zero(faxmode) && !strcmp(faxmode, "FAX_RECEIVE")) {
						const char* receive_filename =pbx_builtin_getvar_helper(session->channel,"FAX_RECEIVE_FILENAME");
						if (receive_filename != NULL) {
							ast_log(LOG_NOTICE, "handle_negotiated_sdp receive filename:%s\n",receive_filename);
							ast_rtp_instance_set_fax_filename(session_media->rtp, receive_filename);
							ast_rtp_instance_set_fax_mode(session_media->rtp, FAX_T30_RECEIVE);
							ast_rtp_instance_set_media_type(session_media->rtp, AST_MEDIA_SUB_IMAGE_TYPE);
						}
					} else if (!ast_strlen_zero(faxmode) && !strcmp(faxmode, "FAX_SENDING")) {
						const char* sending_filename =pbx_builtin_getvar_helper(session->channel,"FAX_SEND_FILENAME");
						if (sending_filename != NULL) {
							ast_log(LOG_NOTICE, "handle_negotiated_sdp sending filename:%s\n",sending_filename);
							if(ast_channel_get_is_send_fax(session->channel)) {
								const char* fax_local_station_id =pbx_builtin_getvar_helper(session->channel,"FAX_LOCAL_STATION_ID");
								if (fax_local_station_id != NULL) {
									ast_rtp_instance_set_fax_localstationid(session_media->rtp,fax_local_station_id);
								}
								const char* fax_header_info =pbx_builtin_getvar_helper(session->channel,"FAX_HEADER_INFO");
								if (fax_header_info != NULL) {
									ast_rtp_instance_set_fax_headerinfo(session_media->rtp,fax_header_info);
								}
							}

							ast_rtp_instance_set_fax_filename(session_media->rtp, sending_filename);
							ast_rtp_instance_set_fax_mode(session_media->rtp, FAX_T30_SENDING);
							ast_rtp_instance_set_media_type(session_media->rtp, AST_MEDIA_SUB_IMAGE_TYPE);
						}
					} else {
						int stream_framing = 0;
						stream_framing = ast_format_cap_get_framing(ast_stream_get_formats(stream));
						ast_rtp_instance_set_transmode(session_media->rtp, ast_stream_get_state(stream));

						if (ast_rtp_instance_get_caps(session_media->rtp)) {
							ast_format_cap_set_framing(ast_rtp_instance_get_caps(session_media->rtp), stream_framing);
						} else {
							ast_log(LOG_WARNING, "[session:%p] negotiated sdp met error: no joint codecs for endpoint:%s\n",
								session, ast_sorcery_object_get_id(session->endpoint));
							return -1;
						}

						ast_avs_after_nego_notify(session->channel, session_media->rtp);
						ast_log(LOG_NOTICE, "[session:%p] ast_avs_after_nego_notify framing is %d for endpoint:%s\n",
							session, stream_framing, ast_sorcery_object_get_id(session->endpoint));
					}

					if (session->wSwitchImageInvite488ToAudioInvite == 1)
						session->wSwitchImageInvite488ToAudioInvite = 0;
				}

				if (!judge_if_need_destory_media(session)) {
					ast_log(LOG_DEBUG, "[session:%p] It's not the final response, don't notify AVS disable media, endpoint:%s\n",
						session, ast_sorcery_object_get_id(session->endpoint));
				} else {
					ast_avs_after_nego_notify(session->channel, session_media->rtp);
				}
			}
			if (!ast_stream_judge_if_stream_is_participant(stream)) {
				changed |= session_media->changed;
			}
			session_media->changed = 0;
#endif
			continue;
		}

		/* If the stream state is REMOVED, nothing needs to be done, so move on to the
		 * next stream. This can occur if an internal thing has requested it to be
		 * removed, or if we remove it as a result of the stream limit being reached.
		 */
#ifndef GRANDSTREAM_NETWORKS
		/* Delete by bfdu, If stream state is removed, local or remote port must be zero, this case is processed before. */
		if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
			/*
			 * Defer removing the handler until we are ready to activate
			 * the new topology.  The channel's thread may still be using
			 * the stream and we could crash before we are ready.
			 */
			continue;
		}
#endif

		if (handle_negotiated_sdp_session_media(session_media, session, local, remote, i, stream)) {
#ifdef GRANDSTREAM_NETWORKS
			ast_log(LOG_WARNING, "[session:%p] Negotiated SDP is failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO) {
				ast_log(LOG_WARNING, "[session:%p] Negotiated SDP is failed, but media type:%d is not audio, just return 0 for endpoint:%s\n",
					session, ast_stream_get_type(stream), ast_sorcery_object_get_id(session->endpoint));
				continue;
			}
#endif
			return -1;
		}

#ifdef GRANDSTREAM_NETWORKS
		send_close_camera_for_webrtc(session, stream);

		if (session->channel && session_media->udptl) {
			{
				RAII_VAR(struct ast_channel *, peer, ast_channel_bridge_peer(session->channel), ast_channel_cleanup);
				if (peer && ast_channel_tech(peer) && !strcasecmp(ast_channel_tech(peer)->type, "DAHDI")) {
					ast_avs_set_udptl_mode(session_media->udptl, FAX_GATEWAY);
				} else if (peer && ast_channel_tech(peer) && !strcasecmp(ast_channel_tech(peer)->type, "PJSIP")) {
					ast_avs_set_udptl_mode(session_media->udptl, FAX_GATEWAY_PASSTHROUGH);
				}
			}
			const char* faxmode =pbx_builtin_getvar_helper(session->channel,"FAX_MODE");
			ast_log(LOG_NOTICE, "ast_avs_after_udptl_nego222 channel name:%s,fax mode:%s\n",ast_channel_name(session->channel),faxmode);
			if(!ast_strlen_zero(faxmode) && !strcmp(faxmode, "FAX_RECEIVE")) {
				const char* receive_filename =pbx_builtin_getvar_helper(session->channel,"FAX_RECEIVE_FILENAME");
				if (receive_filename != NULL)
				{
					ast_log(LOG_NOTICE, "ast_avs_after_udptl_nego receive filename:%s\n",receive_filename);
					ast_avs_set_udptl_mode(session_media->udptl, FAX_T38_RECEIVE);
					ast_avs_set_udptl_filename(session_media->udptl,receive_filename);
				}
			}
			else if(!ast_strlen_zero(faxmode) && !strcmp(faxmode, "FAX_SENDING")) {

				const char* send_filename =pbx_builtin_getvar_helper(session->channel,"FAX_SEND_FILENAME");
				if (send_filename != NULL)
				{
					if(ast_channel_get_is_send_fax(session->channel)) {
						const char* fax_local_station_id =pbx_builtin_getvar_helper(session->channel,"FAX_LOCAL_STATION_ID");
						if (fax_local_station_id != NULL) {
							ast_avs_set_udptl_localstationid(session_media->udptl,fax_local_station_id);
						}
						const char* fax_header_info =pbx_builtin_getvar_helper(session->channel,"FAX_HEADER_INFO");
						if (fax_header_info != NULL) {
							ast_avs_set_udptl_headerinfo(session_media->udptl,fax_header_info);
						}
					}
					ast_log(LOG_NOTICE, "ast_avs_after_udptl_nego send filename:%s\n",send_filename);
					ast_avs_set_udptl_mode(session_media->udptl, FAX_T38_SENDING);
					ast_avs_set_udptl_filename(session_media->udptl,send_filename);
				}
			}
			ast_avs_after_udptl_nego(session->channel, session_media->udptl);
			pbx_builtin_setvar_helper(session->channel, "FAX_SEND_FILENAME", "");
			pbx_builtin_setvar_helper(session->channel, "FAX_MODE","");

			ast_log(LOG_NOTICE, "ast_avs_after_udptl_nego framing\n");
		} else if (session->channel && session_media->rtp) {
			ast_log(LOG_DEBUG, "[session:%p] Negotiated SDP is complete, notify AVS for endpoint:%s,send_fax:%d\n",
				session, ast_sorcery_object_get_id(session->endpoint),ast_channel_get_is_send_fax(session->channel));
			//edit by zyxia for Bug 186023 Bug 185971,2021.1.15
			if(session->wSwitchImageInvite488ToAudioInvite == 1 || ast_channel_get_is_send_fax(session->channel) || ast_channel_get_is_t30_vfax(session->channel)) { 
				const char* faxmode =pbx_builtin_getvar_helper(session->channel,"FAX_MODE");
				ast_log(LOG_NOTICE, "ast_avs_after_nego_notify channel name:%s,fax mode:%s\n",ast_channel_name(session->channel),faxmode);
				if(!ast_strlen_zero(faxmode) && !strcmp(faxmode, "FAX_RECEIVE")) {
					const char* receive_filename =pbx_builtin_getvar_helper(session->channel,"FAX_RECEIVE_FILENAME");
					if (receive_filename != NULL) {
						ast_log(LOG_NOTICE, "ast_avs_after_udptl_nego receive filename:%s\n",receive_filename);
						ast_rtp_instance_set_media_type(session_media->rtp, AST_MEDIA_SUB_IMAGE_TYPE);
						int stream_framing = 0;
						stream_framing = ast_format_cap_get_framing(ast_stream_get_formats(stream));
						ast_rtp_instance_set_transmode(session_media->rtp, ast_stream_get_state(stream));
						
						if (ast_rtp_instance_get_caps(session_media->rtp)) {
							ast_format_cap_set_framing(ast_rtp_instance_get_caps(session_media->rtp), stream_framing);
						} else {
							ast_log(LOG_WARNING, "[session:%p] negotiated sdp met error: no joint codecs for endpoint:%s\n",
								session, ast_sorcery_object_get_id(session->endpoint));
							return -1;
						}
						ast_rtp_instance_set_fax_mode(session_media->rtp, FAX_T30_RECEIVE);
						ast_rtp_instance_set_fax_filename(session_media->rtp, receive_filename);
						ast_avs_after_nego_notify(session->channel, session_media->rtp);
						pbx_builtin_setvar_helper(session->channel, "FAX_SEND_FILENAME", "");
						pbx_builtin_setvar_helper(session->channel, "FAX_MODE","");
					}
				} else if(!ast_strlen_zero(faxmode) && !strcmp(faxmode, "FAX_SENDING")) { 
					const char* send_filename =pbx_builtin_getvar_helper(session->channel,"FAX_SEND_FILENAME");
					if (send_filename != NULL) {
						ast_log(LOG_NOTICE, "ast_avs_after_nego_notify send filename:%s\n",send_filename);
						ast_rtp_instance_set_media_type(session_media->rtp, AST_MEDIA_SUB_IMAGE_TYPE);
						int stream_framing = 0;
						stream_framing = ast_format_cap_get_framing(ast_stream_get_formats(stream));
						ast_rtp_instance_set_transmode(session_media->rtp, ast_stream_get_state(stream));
						
						if (ast_rtp_instance_get_caps(session_media->rtp)) {
							ast_format_cap_set_framing(ast_rtp_instance_get_caps(session_media->rtp), stream_framing);
						} else {
							ast_log(LOG_WARNING, "[session:%p] negotiated sdp met error: no joint codecs for endpoint:%s\n",
								session, ast_sorcery_object_get_id(session->endpoint));
							return -1;
						}
						if(ast_channel_get_is_send_fax(session->channel)) {
							const char* fax_local_station_id =pbx_builtin_getvar_helper(session->channel,"FAX_LOCAL_STATION_ID");
							if (fax_local_station_id != NULL) {
								ast_rtp_instance_set_fax_localstationid(session_media->rtp,fax_local_station_id);
							}
							const char* fax_header_info =pbx_builtin_getvar_helper(session->channel,"FAX_HEADER_INFO");
							if (fax_header_info != NULL) {
								ast_rtp_instance_set_fax_headerinfo(session_media->rtp,fax_header_info);
							}
						}
						ast_rtp_instance_set_fax_filename(session_media->rtp, send_filename);
						ast_rtp_instance_set_fax_mode(session_media->rtp, FAX_T30_SENDING);
						ast_avs_after_nego_notify(session->channel, session_media->rtp);
						pbx_builtin_setvar_helper(session->channel, "FAX_SEND_FILENAME", "");
						pbx_builtin_setvar_helper(session->channel, "FAX_MODE","");
					}
				} else {
					int stream_framing = 0;
					stream_framing = ast_format_cap_get_framing(ast_stream_get_formats(stream));
					ast_rtp_instance_set_transmode(session_media->rtp, ast_stream_get_state(stream));

					if (ast_rtp_instance_get_caps(session_media->rtp)) {
						ast_format_cap_set_framing(ast_rtp_instance_get_caps(session_media->rtp), stream_framing);
					} else {
						ast_log(LOG_WARNING, "[session:%p] negotiated sdp met error: no joint codecs for endpoint:%s\n",
							session, ast_sorcery_object_get_id(session->endpoint));
						return -1;
					}
					ast_avs_after_nego_notify(session->channel, session_media->rtp);
					ast_log(LOG_NOTICE, "[session:%p] ast_avs_after_nego_notify framing is %d for endpoint:%s\n",
						session, stream_framing, ast_sorcery_object_get_id(session->endpoint));
				}

				if (session->wSwitchImageInvite488ToAudioInvite == 1)
					session->wSwitchImageInvite488ToAudioInvite = 0;
			}
			else
			{
				int stream_framing = 0;
				stream_framing = ast_format_cap_get_framing(ast_stream_get_formats(stream));
				ast_rtp_instance_set_transmode(session_media->rtp, ast_stream_get_state(stream));

				if (ast_rtp_instance_get_caps(session_media->rtp)) {
					ast_format_cap_set_framing(ast_rtp_instance_get_caps(session_media->rtp), stream_framing);
				} else {
					ast_log(LOG_WARNING, "[session:%p] negotiated sdp met error: no joint codecs for endpoint:%s\n",
						session, ast_sorcery_object_get_id(session->endpoint));
					return -1;
				}

				/*start added by zyxia for Bug 237118,page callee hold unhold,need updat video msid*/
				if(ast_channel_page_mode_get(session->channel) == AST_PAGE_MODE_ANNOUNCE && ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN
					&&  AST_STREAM_STATE_REMOVED != ast_stream_get_state(stream)
					&&  AST_STREAM_STATE_RECVONLY != ast_stream_get_state(stream)
					&&  AST_STREAM_STATE_INACTIVE != ast_stream_get_state(stream))  {

					const char* page_caller_name = pbx_builtin_getvar_helper(session->channel, "PAGING_INTERCOM_CALLER_CHAN_NAME");
					if (!ast_strlen_zero(page_caller_name)) {
						struct ast_channel *page_caller_chan = NULL;
						page_caller_chan = ast_channel_get_by_name(page_caller_name);
						if (page_caller_chan) {
							RAII_VAR(char *, msid, NULL, ast_free);
							msid = ast_channel_get_msid(page_caller_chan, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
							if (!msid) {
								msid = ast_strdup("");
							}
							ast_log(LOG_NOTICE, "channel:%s Msid:%s\n",ast_channel_name(session->channel),msid);
							ast_rtp_instance_set_local_msid(session_media->rtp, msid);
							ast_channel_unref(page_caller_chan);
						}
					}
				}
				/*end added by zyxia for Bug 237118,page callee hold unhold */

				ast_avs_after_nego_notify(session->channel, session_media->rtp);
				ast_log(LOG_NOTICE, "[session:%p] ast_avs_after_nego_notify framing is %d for endpoint:%s\n",
					session, stream_framing, ast_sorcery_object_get_id(session->endpoint));

				/*Start added by zyxia for PEER FAX DETECTE,2020.9.9*/
				const char* enable_fax_dectct = pbx_builtin_getvar_helper(session->channel, "enable_fax_detect");
				int enable_fax = 0;
				if (session->wFaxDetect || (!ast_strlen_zero(enable_fax_dectct) && !strcasecmp(enable_fax_dectct,"yes") && !strncasecmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12))){
					enable_fax =  1;
				}

				ast_avs_control(session->channel, AVS_CONTROL_DETECT_FAX_TONE, &enable_fax, sizeof(enable_fax));
				/*End added by zyxia for PEER FAX DETECTE,2020.9.9*/
			}

		}
#endif

#ifdef GRANDSTREAM_NETWORKS
		if (!ast_stream_judge_if_stream_is_participant(stream)) {
			changed |= session_media->changed;
		}
#else
		changed |= session_media->changed;
#endif
		session_media->changed = 0;
	}

	/* Apply the pending media state to the channel and make it active */
	ast_channel_lock(session->channel);

	/* Now update the stream handler for any declined/removed streams */
	for (i = 0; i < local->media_count; ++i) {
		struct ast_sip_session_media *session_media;
		struct ast_stream *stream;

		if (!remote->media[i]) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);

		if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED
			&& session_media->handler) {
#ifdef GRANDSTREAM_NETWORKS
			if (session->is_sfu && ast_stream_judge_if_main_video(stream)) {
				ast_log(LOG_DEBUG, "[session:%p] It's main video stream, and stream removed, need to release video resource for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				ast_stream_set_video_resolution(stream, AST_STREAM_VIDEO_RESOLUTION_NONE);
			}
#endif
			/*
			 * This stream is no longer being used and the channel's thread
			 * is held off because we have the channel lock so release any
			 * resources the handler may have on it.
			 */
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for bug 233987, we retain media port for 183 we received port is zero, but 200OK port is not zero */
			if (!judge_if_need_destory_media(session)) {
				ast_log(LOG_DEBUG, "[session:%p] It's not the final response, reserve media port, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			} else {
				session_media_set_handler(session_media, NULL);
			}
#else
			session_media_set_handler(session_media, NULL);
#endif
		}
#ifdef GRANDSTREAM_NETWORKS
		if (session->is_sfu && ast_stream_judge_if_main_video(stream) &&
			ast_stream_get_state(stream) == AST_STREAM_STATE_SENDONLY) {
			ast_log(LOG_DEBUG, "[session:%p] It's main video stream, and stream mode is sendonly, need to release video resource for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_set_video_resolution(stream, AST_STREAM_VIDEO_RESOLUTION_NONE);
		}
#endif
	}

	/* Update the topology on the channel to match the accepted one */
	topology = ast_stream_topology_clone(session->pending_media_state->topology);
	if (topology) {
		ast_channel_set_stream_topology(session->channel, topology);
	}

	/* Remove all current file descriptors from the channel */
	for (i = 0; i < AST_VECTOR_SIZE(&session->active_media_state->read_callbacks); ++i) {
		ast_channel_internal_fd_clear(session->channel, i + AST_EXTENDED_FDS);
	}

	/* Add all the file descriptors from the pending media state */
	for (i = 0; i < AST_VECTOR_SIZE(&session->pending_media_state->read_callbacks); ++i) {
		struct ast_sip_session_media_read_callback_state *callback_state;

		callback_state = AST_VECTOR_GET_ADDR(&session->pending_media_state->read_callbacks, i);
#ifdef GRANDSTREAM_NETWORKS
		if (!ast_channel_is_avs_engine(session->channel)) {
			ast_channel_internal_fd_set(session->channel, i + AST_EXTENDED_FDS, callback_state->fd);
		} else {
			if (callback_state->session->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP && session->bfcp_para.has_udp_bfcp) {
				ast_channel_internal_fd_set(session->channel, i + AST_EXTENDED_FDS, callback_state->fd);
			}
		}
#else
		ast_channel_internal_fd_set(session->channel, i + AST_EXTENDED_FDS, callback_state->fd);
#endif
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_notify_avs_to_make_early_bridge(session);
#endif

	/* Active and pending flip flop as needed */
	SWAP(session->active_media_state, session->pending_media_state);
	ast_sip_session_media_state_reset(session->pending_media_state);

#ifdef GRANDSTREAM_NETWORKS
	if (session->is_media_count_change) {
		ast_log(LOG_DEBUG, "[session:%p] Media count changed, send control frame AST_CONTROL_STREAM_TOPOLOGY_CHANGED for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (session->channel) {
			ast_queue_control(session->channel, AST_CONTROL_STREAM_TOPOLOGY_CHANGED);
		}
	}

	if (session->delayed_reinvite_count) {
		push_sfu_delayed_reinvite_to_task(session);
	}

	negotiated_for_transfer(session);
	notify_topogly_for_sfu_local_invite(session);
	notify_nat_call(session);
	notify_full_participant_sdp_for_reconnect(session);
	notify_remote_disable_media_for_transfer(session);
	send_get_peer_session_frame(session);
	send_reinvite_control_frame_for_no_sdp(session);
	send_peer_reinvite_control_for_callflip(session);
#endif

	ast_channel_unlock(session->channel);

#ifdef GRANDSTREAM_NETWORKS
	indicate_conference_control_camera(session);
	send_bfcp_info_for_reconnect(session);
	send_bfcp_for_open_slides(session);
	send_delayed_bfcp_message(session);
	queue_camera_control_frame(session);
	/*Start added by zyxia for Bug 235770 ,2021.12.24*/
	if(check_channel_is_video_call(session->channel))
		ast_channel_set_is_videocall(session->channel, TRUE);
	/*End added by zyxia for Bug 235770 ,2021.12.24*/
	send_delayed_session_to_peer(session);
#endif

#ifdef GRANDSTREAM_NETWORKS
	if (changed || session->close_camera_forcely || session->camera_control_action)
#else
	if (changed)
#endif
	{
		struct ast_frame f = { AST_FRAME_CONTROL, .subclass.integer = AST_CONTROL_STREAM_TOPOLOGY_SOURCE_CHANGED };

		ast_queue_frame(session->channel, &f);
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_DEBUG, "[session:%p] Media is changed, send control frame AST_CONTROL_STREAM_TOPOLOGY_SOURCE_CHANGED for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->close_camera_forcely = 0;
		session->camera_control_action = 0;
#endif
	} else {
		ast_queue_frame(session->channel, &ast_null_frame);
	}

	return 0;
}

#define DATASTORE_BUCKETS 53
#define MEDIA_BUCKETS 7

static void session_datastore_destroy(void *obj)
{
	struct ast_datastore *datastore = obj;

	/* Using the destroy function (if present) destroy the data */
	if (datastore->info->destroy != NULL && datastore->data != NULL) {
		datastore->info->destroy(datastore->data);
		datastore->data = NULL;
	}

	ast_free((void *) datastore->uid);
	datastore->uid = NULL;
}

struct ast_datastore *ast_sip_session_alloc_datastore(const struct ast_datastore_info *info, const char *uid)
{
	RAII_VAR(struct ast_datastore *, datastore, NULL, ao2_cleanup);
	char uuid_buf[AST_UUID_STR_LEN];
	const char *uid_ptr = uid;

	if (!info) {
		return NULL;
	}

	datastore = ao2_alloc(sizeof(*datastore), session_datastore_destroy);
	if (!datastore) {
		return NULL;
	}

	datastore->info = info;
	if (ast_strlen_zero(uid)) {
		/* They didn't provide an ID so we'll provide one ourself */
		uid_ptr = ast_uuid_generate_str(uuid_buf, sizeof(uuid_buf));
	}

	datastore->uid = ast_strdup(uid_ptr);
	if (!datastore->uid) {
		return NULL;
	}

	ao2_ref(datastore, +1);
	return datastore;
}

int ast_sip_session_add_datastore(struct ast_sip_session *session, struct ast_datastore *datastore)
{
	ast_assert(datastore != NULL);
	ast_assert(datastore->info != NULL);
	ast_assert(ast_strlen_zero(datastore->uid) == 0);

	if (!ao2_link(session->datastores, datastore)) {
		return -1;
	}
	return 0;
}

struct ast_datastore *ast_sip_session_get_datastore(struct ast_sip_session *session, const char *name)
{
	return ao2_find(session->datastores, name, OBJ_KEY);
}

void ast_sip_session_remove_datastore(struct ast_sip_session *session, const char *name)
{
	ao2_callback(session->datastores, OBJ_KEY | OBJ_UNLINK | OBJ_NODATA, NULL, (void *) name);
}

enum delayed_method {
	DELAYED_METHOD_INVITE,
	DELAYED_METHOD_UPDATE,
	DELAYED_METHOD_BYE,
};

/*!
 * \internal
 * \brief Convert delayed method enum value to a string.
 * \since 13.3.0
 *
 * \param method Delayed method enum value to convert to a string.
 *
 * \return String value of delayed method.
 */
static const char *delayed_method2str(enum delayed_method method)
{
	const char *str = "<unknown>";

	switch (method) {
	case DELAYED_METHOD_INVITE:
		str = "INVITE";
		break;
	case DELAYED_METHOD_UPDATE:
		str = "UPDATE";
		break;
	case DELAYED_METHOD_BYE:
		str = "BYE";
		break;
	}

	return str;
}

/*!
 * \brief Structure used for sending delayed requests
 *
 * Requests are typically delayed because the current transaction
 * state of an INVITE. Once the pending INVITE transaction terminates,
 * the delayed request will be sent
 */
struct ast_sip_session_delayed_request {
	/*! Method of the request */
	enum delayed_method method;
	/*! Callback to call when the delayed request is created. */
	ast_sip_session_request_creation_cb on_request_creation;
	/*! Callback to call when the delayed request SDP is created */
	ast_sip_session_sdp_creation_cb on_sdp_creation;
	/*! Callback to call when the delayed request receives a response */
	ast_sip_session_response_cb on_response;
	/*! Whether to generate new SDP */
	int generate_new_sdp;
	/*! Requested media state for the SDP */
	struct ast_sip_session_media_state *media_state;
	AST_LIST_ENTRY(ast_sip_session_delayed_request) next;
};

static struct ast_sip_session_delayed_request *delayed_request_alloc(
	enum delayed_method method,
	ast_sip_session_request_creation_cb on_request_creation,
	ast_sip_session_sdp_creation_cb on_sdp_creation,
	ast_sip_session_response_cb on_response,
	int generate_new_sdp,
	struct ast_sip_session_media_state *media_state)
{
	struct ast_sip_session_delayed_request *delay = ast_calloc(1, sizeof(*delay));

	if (!delay) {
		return NULL;
	}
	delay->method = method;
	delay->on_request_creation = on_request_creation;
	delay->on_sdp_creation = on_sdp_creation;
	delay->on_response = on_response;
	delay->generate_new_sdp = generate_new_sdp;
	delay->media_state = media_state;
	return delay;
}

static void delayed_request_free(struct ast_sip_session_delayed_request *delay)
{
	ast_sip_session_media_state_free(delay->media_state);
	ast_free(delay);
}

static int send_delayed_request(struct ast_sip_session *session, struct ast_sip_session_delayed_request *delay)
{
	ast_debug(3, "Endpoint '%s(%s)' sending delayed %s request.\n",
		ast_sorcery_object_get_id(session->endpoint),
		session->channel ? ast_channel_name(session->channel) : "",
		delayed_method2str(delay->method));

#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for SFU re-INVITE conflict, 2019.08.09 */
	if (session->is_sfu) {
		if (delay->media_state) {
			ast_sip_session_media_state_free(delay->media_state);
		}
		delay->media_state = get_session_media_state(session, 0);
	}
#endif

	switch (delay->method) {
	case DELAYED_METHOD_INVITE:
		ast_sip_session_refresh(session, delay->on_request_creation,
			delay->on_sdp_creation, delay->on_response,
			AST_SIP_SESSION_REFRESH_METHOD_INVITE, delay->generate_new_sdp, delay->media_state);
		/* Ownership of media state transitions to ast_sip_session_refresh */
		delay->media_state = NULL;
		return 0;
	case DELAYED_METHOD_UPDATE:
		ast_sip_session_refresh(session, delay->on_request_creation,
			delay->on_sdp_creation, delay->on_response,
			AST_SIP_SESSION_REFRESH_METHOD_UPDATE, delay->generate_new_sdp, delay->media_state);
		/* Ownership of media state transitions to ast_sip_session_refresh */
		delay->media_state = NULL;
		return 0;
	case DELAYED_METHOD_BYE:
		ast_sip_session_terminate(session, 0);
		return 0;
	}
	ast_log(LOG_WARNING, "Don't know how to send delayed %s(%d) request.\n",
		delayed_method2str(delay->method), delay->method);
	return -1;
}

/*!
 * \internal
 * \brief The current INVITE transaction is in the PROCEEDING state.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int invite_proceeding(void *vsession)
{
	struct ast_sip_session *session = vsession;
	struct ast_sip_session_delayed_request *delay;
	int found = 0;
	int res = 0;

	AST_LIST_TRAVERSE_SAFE_BEGIN(&session->delayed_requests, delay, next) {
		switch (delay->method) {
		case DELAYED_METHOD_INVITE:
			break;
		case DELAYED_METHOD_UPDATE:
			AST_LIST_REMOVE_CURRENT(next);
			res = send_delayed_request(session, delay);
			delayed_request_free(delay);
			found = 1;
			break;
		case DELAYED_METHOD_BYE:
			/* A BYE is pending so don't bother anymore. */
			found = 1;
			break;
		}
		if (found) {
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	ao2_ref(session, -1);
	return res;
}

/*!
 * \internal
 * \brief The current INVITE transaction is in the TERMINATED state.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int invite_terminated(void *vsession)
{
	struct ast_sip_session *session = vsession;
	struct ast_sip_session_delayed_request *delay;
	int found = 0;
	int res = 0;
	int timer_running;

	/* re-INVITE collision timer running? */
	timer_running = pj_timer_entry_running(&session->rescheduled_reinvite);

	AST_LIST_TRAVERSE_SAFE_BEGIN(&session->delayed_requests, delay, next) {
		switch (delay->method) {
		case DELAYED_METHOD_INVITE:
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for SFU re-INVITE conflict, 2019.10.18 */
			if (session->inv_session && session->inv_session->invite_tsx) {
				ast_debug(3, "[session:%p] Don't send delay reinvite to %s because of outstanding transaction...\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				break;
			}
#endif
			if (!timer_running) {
				found = 1;
			}
			break;
		case DELAYED_METHOD_UPDATE:
		case DELAYED_METHOD_BYE:
			found = 1;
			break;
		}
		if (found) {
			AST_LIST_REMOVE_CURRENT(next);
			res = send_delayed_request(session, delay);
			delayed_request_free(delay);
			break;
		}
	}
	AST_LIST_TRAVERSE_SAFE_END;

	ao2_ref(session, -1);
	return res;
}

/*!
 * \internal
 * \brief INVITE collision timeout.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int invite_collision_timeout(void *vsession)
{
	struct ast_sip_session *session = vsession;
	int res;

	if (session->inv_session->invite_tsx) {
		/*
		 * INVITE transaction still active.  Let it send
		 * the collision re-INVITE when it terminates.
		 */
		ao2_ref(session, -1);
		res = 0;
	} else {
		res = invite_terminated(session);
	}

	return res;
}

/*!
 * \internal
 * \brief The current UPDATE transaction is in the COMPLETED state.
 * \since 13.3.0
 *
 * \param vsession Session object.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int update_completed(void *vsession)
{
	struct ast_sip_session *session = vsession;
	int res;

	if (session->inv_session->invite_tsx) {
		res = invite_proceeding(session);
	} else {
		res = invite_terminated(session);
	}

	return res;
}

static void check_delayed_requests(struct ast_sip_session *session,
	int (*cb)(void *vsession))
{
	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, cb, session)) {
		ao2_ref(session, -1);
	}
}

static int delay_request(struct ast_sip_session *session,
	ast_sip_session_request_creation_cb on_request,
	ast_sip_session_sdp_creation_cb on_sdp_creation,
	ast_sip_session_response_cb on_response,
	int generate_new_sdp,
	enum delayed_method method,
	struct ast_sip_session_media_state *media_state)
{
	struct ast_sip_session_delayed_request *delay = delayed_request_alloc(method,
			on_request, on_sdp_creation, on_response, generate_new_sdp, media_state);

	if (!delay) {
		ast_sip_session_media_state_free(media_state);
		return -1;
	}

	if (method == DELAYED_METHOD_BYE) {
		/* Send BYE as early as possible */
		AST_LIST_INSERT_HEAD(&session->delayed_requests, delay, next);
	} else {
		AST_LIST_INSERT_TAIL(&session->delayed_requests, delay, next);
	}
	return 0;
}

static pjmedia_sdp_session *generate_session_refresh_sdp(struct ast_sip_session *session)
{
	pjsip_inv_session *inv_session = session->inv_session;
	const pjmedia_sdp_session *previous_sdp = NULL;

	if (inv_session->neg) {
		if (pjmedia_sdp_neg_was_answer_remote(inv_session->neg)) {
			pjmedia_sdp_neg_get_active_remote(inv_session->neg, &previous_sdp);
		} else {
			pjmedia_sdp_neg_get_active_local(inv_session->neg, &previous_sdp);
		}
	}
	return create_local_sdp(inv_session, session, previous_sdp);
}

static void set_from_header(struct ast_sip_session *session)
{
	struct ast_party_id effective_id;
	struct ast_party_id connected_id;
	pj_pool_t *dlg_pool;
	pjsip_fromto_hdr *dlg_info;
	pjsip_contact_hdr *dlg_contact;
	pjsip_name_addr *dlg_info_name_addr;
	pjsip_sip_uri *dlg_info_uri;
	pjsip_sip_uri *dlg_contact_uri;
	int restricted;

	if (!session->channel || session->saved_from_hdr) {
		return;
	}

	/* We need to save off connected_id for RPID/PAI generation */
	ast_party_id_init(&connected_id);
	ast_channel_lock(session->channel);
	effective_id = ast_channel_connected_effective_id(session->channel);
	ast_party_id_copy(&connected_id, &effective_id);
	ast_channel_unlock(session->channel);

	restricted =
		((ast_party_id_presentation(&connected_id) & AST_PRES_RESTRICTION) != AST_PRES_ALLOWED);

	/* Now set up dlg->local.info so pjsip can correctly generate From */

	dlg_pool = session->inv_session->dlg->pool;
	dlg_info = session->inv_session->dlg->local.info;
	dlg_contact = session->inv_session->dlg->local.contact;
	dlg_info_name_addr = (pjsip_name_addr *) dlg_info->uri;
	dlg_info_uri = pjsip_uri_get_uri(dlg_info_name_addr);
	dlg_contact_uri = (pjsip_sip_uri*)pjsip_uri_get_uri(dlg_contact->uri);

	if (session->endpoint->id.trust_outbound || !restricted) {
		ast_sip_modify_id_header(dlg_pool, dlg_info, &connected_id);
		if (ast_sip_get_use_callerid_contact() && ast_strlen_zero(session->endpoint->contact_user)) {
			pj_strdup2(dlg_pool, &dlg_contact_uri->user, S_COR(connected_id.number.valid, connected_id.number.str, ""));
		}
	}

	ast_party_id_free(&connected_id);

	if (!ast_strlen_zero(session->endpoint->fromuser)) {
		dlg_info_name_addr->display.ptr = NULL;
		dlg_info_name_addr->display.slen = 0;
		pj_strdup2(dlg_pool, &dlg_info_uri->user, session->endpoint->fromuser);
	}

	if (!ast_strlen_zero(session->endpoint->fromdomain)) {
		pj_strdup2(dlg_pool, &dlg_info_uri->host, session->endpoint->fromdomain);
	}

	/* We need to save off the non-anonymized From for RPID/PAI generation (for domain) */
	session->saved_from_hdr = pjsip_hdr_clone(dlg_pool, dlg_info);
	ast_sip_add_usereqphone(session->endpoint, dlg_pool, session->saved_from_hdr->uri);

	/* In chan_sip, fromuser and fromdomain trump restricted so we only
	 * anonymize if they're not set.
	 */
	if (restricted) {
		/* fromuser doesn't provide a display name so we always set it */
		pj_strdup2(dlg_pool, &dlg_info_name_addr->display, "Anonymous");

		if (ast_strlen_zero(session->endpoint->fromuser)) {
			pj_strdup2(dlg_pool, &dlg_info_uri->user, "anonymous");
		}

		if (ast_sip_get_use_callerid_contact() && ast_strlen_zero(session->endpoint->contact_user)) {
			pj_strdup2(dlg_pool, &dlg_contact_uri->user, "anonymous");
		}

		if (ast_strlen_zero(session->endpoint->fromdomain)) {
			pj_strdup2(dlg_pool, &dlg_info_uri->host, "anonymous.invalid");
		}
	} else {
		ast_sip_add_usereqphone(session->endpoint, dlg_pool, dlg_info->uri);
    }
}

#ifdef GRANDSTREAM_NETWORKS
/*Start added by zyxia for confernce callflip,update contact is conf number,2021.12.20*/
int ast_sip_set_contact_user(pjsip_tx_data *packet, const char *user)
{
    pjsip_contact_hdr *contact = NULL;
    pjsip_sip_uri *contact_uri = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    if (user == NULL) {
        ast_log(LOG_WARNING, "user is NULL\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL);
    if (contact == NULL || contact->uri == NULL) {
        ast_log(LOG_WARNING, "contact URI header is NULL\n");
        return -1;
    }

    contact_uri = pjsip_uri_get_uri(contact->uri);
    if (contact_uri == NULL) {
        ast_log(LOG_WARNING, "Failed to get contact URI\n");
        return -1;
    }

    pj_strdup2(packet->pool, &contact_uri->user, user);

    return 0;
}
#endif
/*End added by zyxia for confernce callflip,update contact is conf number,2021.12.20*/

int ast_sip_session_refresh(struct ast_sip_session *session,
		ast_sip_session_request_creation_cb on_request_creation,
		ast_sip_session_sdp_creation_cb on_sdp_creation,
		ast_sip_session_response_cb on_response,
		enum ast_sip_session_refresh_method method, int generate_new_sdp,
		struct ast_sip_session_media_state *media_state)
{
	pjsip_inv_session *inv_session = session->inv_session;
	pjmedia_sdp_session *new_sdp = NULL;
	pjsip_tx_data *tdata;

	if (media_state && (!media_state->topology || !generate_new_sdp)) {
		ast_sip_session_media_state_free(media_state);
		return -1;
	}

	if (inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		/* Don't try to do anything with a hung-up call */
		ast_debug(3, "Not sending reinvite to %s because of disconnected state...\n",
				ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_free(media_state);
		return 0;
	}

	/* If the dialog has not yet been established we have to defer until it has */
	if (inv_session->dlg->state != PJSIP_DIALOG_STATE_ESTABLISHED) {
		ast_debug(3, "Delay sending request to %s because dialog has not been established...\n",
			ast_sorcery_object_get_id(session->endpoint));
		return delay_request(session, on_request_creation, on_sdp_creation, on_response,
			generate_new_sdp,
			method == AST_SIP_SESSION_REFRESH_METHOD_INVITE
				? DELAYED_METHOD_INVITE : DELAYED_METHOD_UPDATE,
			media_state);
	}

	if (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE) {
		if (inv_session->invite_tsx) {
			/* We can't send a reinvite yet, so delay it */
			ast_debug(3, "Delay sending reinvite to %s because of outstanding transaction...\n",
					ast_sorcery_object_get_id(session->endpoint));
			return delay_request(session, on_request_creation, on_sdp_creation,
				on_response, generate_new_sdp, DELAYED_METHOD_INVITE, media_state);
		} else if (inv_session->state != PJSIP_INV_STATE_CONFIRMED) {
			/* Initial INVITE transaction failed to progress us to a confirmed state
			 * which means re-invites are not possible
			 */
			ast_debug(3, "Not sending reinvite to %s because not in confirmed state...\n",
					ast_sorcery_object_get_id(session->endpoint));
			ast_sip_session_media_state_free(media_state);
			return 0;
		}
	}

	if (generate_new_sdp) {
		/* SDP can only be generated if current negotiation has already completed */
		if (inv_session->neg
			&& pjmedia_sdp_neg_get_state(inv_session->neg)
				!= PJMEDIA_SDP_NEG_STATE_DONE) {
			ast_debug(3, "Delay session refresh with new SDP to %s because SDP negotiation is not yet done...\n",
				ast_sorcery_object_get_id(session->endpoint));
			return delay_request(session, on_request_creation, on_sdp_creation,
				on_response, generate_new_sdp,
				method == AST_SIP_SESSION_REFRESH_METHOD_INVITE
					? DELAYED_METHOD_INVITE : DELAYED_METHOD_UPDATE, media_state);
		}

		/* If an explicitly requested media state has been provided use it instead of any pending one */
		if (media_state) {
			int index;
			int type_streams[AST_MEDIA_TYPE_END] = {0};
			struct ast_stream *stream;

			/* Prune the media state so the number of streams fit within the configured limits - we do it here
			 * so that the index of the resulting streams in the SDP match. If we simply left the streams out
			 * of the SDP when producing it we'd be in trouble. We also enforce formats here for media types that
			 * are configurable on the endpoint.
			 */
			for (index = 0; index < ast_stream_topology_get_count(media_state->topology); ++index) {
				struct ast_stream *existing_stream = NULL;

				stream = ast_stream_topology_get_stream(media_state->topology, index);

				if (session->active_media_state->topology &&
					index < ast_stream_topology_get_count(session->active_media_state->topology)) {
					existing_stream = ast_stream_topology_get_stream(session->active_media_state->topology, index);
				}

				if (is_stream_limitation_reached(ast_stream_get_type(stream), session->endpoint, type_streams)) {
					if (index < AST_VECTOR_SIZE(&media_state->sessions)) {
						struct ast_sip_session_media *session_media = AST_VECTOR_GET(&media_state->sessions, index);

						ao2_cleanup(session_media);
						AST_VECTOR_REMOVE(&media_state->sessions, index, 1);
					}

					ast_stream_topology_del_stream(media_state->topology, index);

					/* A stream has potentially moved into our spot so we need to jump back so we process it */
					index -= 1;
					continue;
				}

				/* No need to do anything with stream if it's media state is removed */
				if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
					/* If there is no existing stream we can just not have this stream in the topology at all. */
					if (!existing_stream) {
						ast_stream_topology_del_stream(media_state->topology, index);
						index -= 1;
					}
					continue;
				}

				/* Enforce the configured allowed codecs on audio and video streams */
				if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_AUDIO || ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO) {
					struct ast_format_cap *joint_cap;

					joint_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
					if (!joint_cap) {
						ast_sip_session_media_state_free(media_state);
						return 0;
					}
					ast_format_cap_get_compatible(ast_stream_get_formats(stream), session->endpoint->media.codecs, joint_cap);
#ifdef GRANDSTREAM_NETWORKS
					int useinbandfec = ast_sip_session_get_format_useinbandfec(session, stream, ast_stream_get_formats(stream));
					if (useinbandfec) {
						ast_sip_session_set_format_useinbandfec(session, joint_cap, stream, 1);
					} else {
						ast_sip_session_set_format_useinbandfec(session, joint_cap, stream, 0);
					}
					ast_sip_session_set_format_dtx(session, joint_cap, stream);

					/* Modify bf bfdu, if joint_cap only has rtx, don't add this media to stream */
					if (!ast_format_cap_count(joint_cap) ||
						(ast_format_cap_has_rtx(joint_cap) && ast_format_cap_count(joint_cap) == 1))
#else
					if (!ast_format_cap_count(joint_cap))
#endif
					{
						ao2_ref(joint_cap, -1);

						if (!existing_stream) {
							/* If there is no existing stream we can just not have this stream in the topology
							 * at all.
							 */
							ast_stream_topology_del_stream(media_state->topology, index);
							index -= 1;
							continue;
						} else if (ast_stream_get_state(stream) != ast_stream_get_state(existing_stream) ||
								strcmp(ast_stream_get_name(stream), ast_stream_get_name(existing_stream))) {
							/* If the underlying stream is a different type or different name then we have to
							 * mark it as removed, as it is replacing an existing stream. We do this so order
							 * is preserved.
							 */
#ifdef GRANDSTREAM_NETWORKS
							ast_log(LOG_DEBUG, "[session:%p] Can't get valid video codec in this stream, set stream:%p (index:%d) removed for endpoint:%s\n",
								session, stream, index, ast_sorcery_object_get_id(session->endpoint));
#endif
							ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
							continue;
						} else {
							/* However if the stream is otherwise remaining the same we can keep the formats
							 * that exist on it already which allows media to continue to flow.
							 */
							joint_cap = ao2_bump(ast_stream_get_formats(existing_stream));
						}
					}
#ifdef GRANDSTREAM_NETWORKS
					if (session->is_sfu && 
						(ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO) &&
						(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) &&
						!ast_format_cap_if_suppport_webrtc(joint_cap)) {
						ao2_ref(joint_cap, -1);
						if (!existing_stream) {
							ast_log(LOG_DEBUG, "[session:%p] Webrtc can't support this joint cap codec, delete this stream for endpoint:%s\n",
								session, ast_sorcery_object_get_id(session->endpoint));
							ast_stream_topology_del_stream(media_state->topology, index);
							index -= 1;
						} else {
							ast_log(LOG_DEBUG, "[session:%p] Webrtc can't support this joint cap codec, set this stream removed for endpoint:%s\n",
								session, ast_sorcery_object_get_id(session->endpoint));
							ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
						}
						continue;
					}
#endif
					ast_stream_set_formats(stream, joint_cap);
					ao2_cleanup(joint_cap);
				}

				++type_streams[ast_stream_get_type(stream)];
			}

			if (session->active_media_state->topology) {
				/* SDP is a fun thing. Take for example the fact that streams are never removed. They just become
				 * declined. To better handle this in the case where something requests a topology change for fewer
				 * streams than are currently present we fill in the topology to match the current number of streams
				 * that are active.
				 */
				for (index = ast_stream_topology_get_count(media_state->topology);
					index < ast_stream_topology_get_count(session->active_media_state->topology); ++index) {
					struct ast_stream *cloned;

					stream = ast_stream_topology_get_stream(session->active_media_state->topology, index);
					ast_assert(stream != NULL);

					cloned = ast_stream_clone(stream, NULL);
					if (!cloned) {
						ast_sip_session_media_state_free(media_state);
						return -1;
					}
#ifdef GRANDSTREAM_NETWORKS
					ast_log(LOG_DEBUG, "[session:%p] Active media stream can't find, set this stream:%p (index:%d) removed for endpoint:%s\n",
						session, cloned, index, ast_sorcery_object_get_id(session->endpoint));
#endif
					ast_stream_set_state(cloned, AST_STREAM_STATE_REMOVED);
					if (ast_stream_topology_append_stream(media_state->topology, cloned) < 0) {
						ast_stream_free(cloned);
						ast_sip_session_media_state_free(media_state);
						return -1;
					}
				}

				/* If the resulting media state matches the existing active state don't bother doing a session refresh */
#ifdef GRANDSTREAM_NETWORKS
				if (!session->wSwitchImageInvite488ToAudioInvite && ast_stream_topology_equal(session->active_media_state->topology, media_state->topology))
#else
				if (ast_stream_topology_equal(session->active_media_state->topology, media_state->topology))
#endif
				{
					ast_sip_session_media_state_free(media_state);
					return 0;
				}
			}

			ast_sip_session_media_state_free(session->pending_media_state);
			session->pending_media_state = media_state;
		}

		new_sdp = generate_session_refresh_sdp(session);
		if (!new_sdp) {
			ast_log(LOG_ERROR, "Failed to generate session refresh SDP. Not sending session refresh\n");
			ast_sip_session_media_state_reset(session->pending_media_state);
			return -1;
		}
		if (on_sdp_creation) {
			if (on_sdp_creation(session, new_sdp)) {
				ast_sip_session_media_state_reset(session->pending_media_state);
				return -1;
			}
		}
	}

	if (method == AST_SIP_SESSION_REFRESH_METHOD_INVITE) {
		if (pjsip_inv_reinvite(inv_session, NULL, new_sdp, &tdata)) {
			ast_log(LOG_WARNING, "Failed to create reinvite properly.\n");
			if (generate_new_sdp) {
				ast_sip_session_media_state_reset(session->pending_media_state);
			}
			return -1;
		}
	} else if (pjsip_inv_update(inv_session, NULL, new_sdp, &tdata)) {
		ast_log(LOG_WARNING, "Failed to create UPDATE properly.\n");
		if (generate_new_sdp) {
			ast_sip_session_media_state_reset(session->pending_media_state);
		}
		return -1;
	}
	if (on_request_creation) {
		if (on_request_creation(session, tdata)) {
			if (generate_new_sdp) {
				ast_sip_session_media_state_reset(session->pending_media_state);
			}
			return -1;
		}
	}
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for add isfocus for re-INVITE */
	if (session->is_sfu || session->joint_ipvt) {
		ast_log(LOG_DEBUG, "[session:%p] It a sfu call, add isfocus for re-INVITE, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));

		if(append_isfocus_contact_for_reinvite(tdata)== 0) {
			/*aaded by zyxia for Bug 264194 ,if add isfocus,need update cotnact user to conf num*/
			const char* conf_room  = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
			if (!ast_strlen_zero(conf_room)) 
				ast_sip_set_contact_user(tdata,conf_room);
		}
	#if 0
		/*Start added by zyxia for confernce callflip,update contact is conf number,2021.12.20*/
		const char* from_call_flip = pbx_builtin_getvar_helper(session->channel, "FROM_CALL_FLIP");
		if (!ast_strlen_zero(from_call_flip) && strcasecmp(from_call_flip, "yes") == 0) {
			ast_log(LOG_NOTICE, "[session:%p] It a sfu call from callflip conference, update contact  for updata,endpoint:%s",session, ast_sorcery_object_get_id(session->endpoint));
			const char* conf_room  = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
			if (!ast_strlen_zero(conf_room)) 
				ast_sip_set_contact_user(tdata,conf_room);
		}
		/*End added by zyxia for confernce callflip,update contact is conf number,2021.12.20*/
	#endif
	}

	if(session->channel && strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)
		&& ast_channel_get_virtual_conf_num(session->channel) 
		&& ast_channel_get_is_join_trunk_conf(session->channel) ) {
			append_isfocus_contact_for_reinvite(tdata);
			const char *trunk_index= pbx_builtin_getvar_helper(session->channel, "TRUNK_INDEX");
			if(!ast_strlen_zero(trunk_index)){
				char local_virtual_conf[512] = {0};
				char family[64] = {0};
				char trunkName[128] = {0};
				ast_log(LOG_DEBUG, "ast_sip_session_refresh:%s,%s\n",session->trunk_conf_num,ast_channel_name(session->channel));

				snprintf(family, sizeof(family) - 1, "TRUNK/SIP_%d", atoi(trunk_index));
				ast_ucm_db_get(family, "TRUNK_NAME", trunkName, sizeof(trunkName));
				snprintf(local_virtual_conf, sizeof(local_virtual_conf) - 1, "%s_%s", trunkName,session->trunk_conf_num);
				ast_sip_set_contact_user(tdata,local_virtual_conf);
			}
	}

	if (generate_new_sdp && session->channel && strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		ast_sip_add_timestamp_header(tdata);
	}

	if (session->channel) {
		ast_sip_add_header(tdata, "Channel-ID", ast_channel_name(session->channel));
	}

	/*Start added by zyxia for Bug 241705,2023.4.11*/
	const char *call_to_ipc = pbx_builtin_getvar_helper(session->channel, "CALL_TO_IPC");
	const char *call_to_gds = pbx_builtin_getvar_helper(session->channel, "CALL_TO_GDS");
	const char *call_to_hik = pbx_builtin_getvar_helper(session->channel, "CALL_TO_HIK");
	/*Start added by zyxia for Bug 261649 ,2023.6.9*/
	if (!ast_strlen_zero(call_to_gds) && !strcasecmp(call_to_gds, "yes")) {
		ast_log(LOG_NOTICE, "channel:%s   add X-Grandstream-Call-Type DoorStation/Grandstream header!",ast_channel_name(session->channel)); 
		ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "DoorStation/Grandstream");
	}
	else if (!ast_strlen_zero(call_to_hik) && !strcasecmp(call_to_hik, "yes")) {
		ast_log(LOG_NOTICE, "channel:%s   add X-Grandstream-Call-Type  DoorStation/Hikvision header!",ast_channel_name(session->channel)); 
		ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "DoorStation/Hikvision");
	}
		/*End added by zyxia for Bug 261649 ,2023.6.9*/
	else if (!ast_strlen_zero(call_to_ipc) && !strcasecmp(call_to_ipc, "yes")) {
		ast_log(LOG_NOTICE, "channel:%s   add X-Grandstream-Call-Type IPC header!",ast_channel_name(session->channel)); 
		ast_sip_add_header(tdata, "X-Grandstream-Call-Type", "IPC");
	}
	/*End added by zyxia for Bug 241705,2023.4.11*/

#endif
	ast_debug(3, "Sending session refresh SDP via %s to %s\n",
		method == AST_SIP_SESSION_REFRESH_METHOD_INVITE ? "re-INVITE" : "UPDATE",
		ast_sorcery_object_get_id(session->endpoint));
	ast_sip_session_send_request_with_cb(session, tdata, on_response);
	return 0;
}

int ast_sip_session_regenerate_answer(struct ast_sip_session *session,
		ast_sip_session_sdp_creation_cb on_sdp_creation)
{
	pjsip_inv_session *inv_session = session->inv_session;
	pjmedia_sdp_session *new_answer = NULL;
	const pjmedia_sdp_session *previous_offer = NULL;

	/* The SDP answer can only be regenerated if it is still pending to be sent */
	if (!inv_session->neg || (pjmedia_sdp_neg_get_state(inv_session->neg) != PJMEDIA_SDP_NEG_STATE_REMOTE_OFFER &&
		pjmedia_sdp_neg_get_state(inv_session->neg) != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO)) {
		ast_log(LOG_WARNING, "Requested to regenerate local SDP answer for channel '%s' but negotiation in state '%s'\n",
			ast_channel_name(session->channel), pjmedia_sdp_neg_state_str(pjmedia_sdp_neg_get_state(inv_session->neg)));
		return -1;
	}

	pjmedia_sdp_neg_get_neg_remote(inv_session->neg, &previous_offer);
	if (pjmedia_sdp_neg_get_state(inv_session->neg) == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		/* Transition the SDP negotiator back to when it received the remote offer */
		pjmedia_sdp_neg_negotiate(inv_session->pool, inv_session->neg, 0);
		pjmedia_sdp_neg_set_remote_offer(inv_session->pool, inv_session->neg, previous_offer);
	}

	new_answer = create_local_sdp(inv_session, session, previous_offer);
	if (!new_answer) {
		ast_log(LOG_WARNING, "Could not create a new local SDP answer for channel '%s'\n",
			ast_channel_name(session->channel));
		return -1;
	}

	if (on_sdp_creation) {
		if (on_sdp_creation(session, new_answer)) {
			return -1;
		}
	}

	pjsip_inv_set_sdp_answer(inv_session, new_answer);

	return 0;
}

void ast_sip_session_send_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	handle_outgoing_response(session, tdata);
	pjsip_inv_send_msg(session->inv_session, tdata);
	return;
}

static pj_bool_t session_on_rx_request(pjsip_rx_data *rdata);

static pjsip_module session_module = {
	.name = {"Session Module", 14},
	.priority = PJSIP_MOD_PRIORITY_APPLICATION,
	.on_rx_request = session_on_rx_request,
};

/*! \brief Determine whether the SDP provided requires deferral of negotiating or not
 *
 * \retval 1 re-invite should be deferred and resumed later
 * \retval 0 re-invite should not be deferred
 */
static int sdp_requires_deferral(struct ast_sip_session *session, const pjmedia_sdp_session *sdp)
{
	int i;

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_alloc();
		if (!session->pending_media_state->topology) {
			return -1;
		}
	}

	for (i = 0; i < sdp->media_count; ++i) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
		struct ast_sip_session_media *session_media = NULL;
		enum ast_sip_session_sdp_stream_defer res;

#ifdef GRANDSTREAM_NETWORKS
		session_media = add_session_media_for_stream(session, sdp, i);
#else
		struct ast_stream *existing_stream = NULL;
		struct ast_stream *stream;
		enum ast_media_type type;
		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &sdp->media[i]->desc.media, sizeof(media));

		if (session->active_media_state->topology &&
			(i < ast_stream_topology_get_count(session->active_media_state->topology))) {
			existing_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		}

		type = ast_media_type_from_str(media);
		stream = ast_stream_alloc(existing_stream ? ast_stream_get_name(existing_stream) : ast_codec_media_type2str(type), type);
		if (!stream) {
			return -1;
		}

		/* As this is only called on an incoming SDP offer before processing it is not possible
		 * for streams and their media sessions to exist.
		 */
		if (ast_stream_topology_set_stream(session->pending_media_state->topology, i, stream)) {
			ast_stream_free(stream);
			return -1;
		}

		session_media = ast_sip_session_media_state_add(session, session->pending_media_state, ast_media_type_from_str(media), i);
#endif
		if (!session_media) {
			return -1;
		}

		if (session_media->handler) {
			handler = session_media->handler;
			if (handler->defer_incoming_sdp_stream) {
				res = handler->defer_incoming_sdp_stream(session, session_media, sdp,
					sdp->media[i]);
				switch (res) {
				case AST_SIP_SESSION_SDP_DEFER_NOT_HANDLED:
					break;
				case AST_SIP_SESSION_SDP_DEFER_ERROR:
					return 0;
				case AST_SIP_SESSION_SDP_DEFER_NOT_NEEDED:
					break;
				case AST_SIP_SESSION_SDP_DEFER_NEEDED:
					return 1;
				}
			}
			/* Handled by this handler. Move to the next stream */
			continue;
		}

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler == session_media->handler) {
				continue;
			}
			if (!handler->defer_incoming_sdp_stream) {
				continue;
			}
			res = handler->defer_incoming_sdp_stream(session, session_media, sdp,
				sdp->media[i]);
			switch (res) {
			case AST_SIP_SESSION_SDP_DEFER_NOT_HANDLED:
				continue;
			case AST_SIP_SESSION_SDP_DEFER_ERROR:
				session_media_set_handler(session_media, handler);
				return 0;
			case AST_SIP_SESSION_SDP_DEFER_NOT_NEEDED:
				/* Handled by this handler. */
				session_media_set_handler(session_media, handler);
				break;
			case AST_SIP_SESSION_SDP_DEFER_NEEDED:
				/* Handled by this handler. */
				session_media_set_handler(session_media, handler);
				return 1;
			}
			/* Move to the next stream */
			break;
		}
	}
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static pj_bool_t session_check_sdp_media_type(const struct ast_sip_session *session, const pjmedia_sdp_session *sdp, const char *media_type_name)
{
	int i = 0;
	if (!sdp) {
		return PJ_FALSE;
	}

	for (i = 0; i < sdp->media_count; ++i) {
		if (!pj_strcmp2(&sdp->media[i]->desc.media, media_type_name)) {

			return PJ_TRUE;
		}
	}

	return PJ_FALSE;
}
/*Start added by zyxia for Bug185419,2021.1.22*/
static pj_bool_t session_search_audio_port_with_image_audio(struct ast_sip_session *session, int image_audio, int is_t38_enable, pjmedia_sdp_session *receive_sdp, const char *media_type_name)
{
	if (!image_audio || !is_t38_enable) {
		return PJ_FALSE;
	}

	int i = 0;

	for (; i < receive_sdp->media_count; ++i) {
		if (!pj_strcmp2(&receive_sdp->media[i]->desc.media, media_type_name)) {
			ast_log(LOG_VERBOSE, "receive_sdp->media[i]->desc.port[%d]", receive_sdp->media[i]->desc.port);
			receive_sdp->media[i]->desc.port = 0;
			break;
		}
	}
	return PJ_TRUE;
}
/*End added by zyxia for Bug185419,2021.1.22*/

static void session_handle_ucm_conf_header(pjsip_rx_data *rdata)
{
	pj_str_t hdr_str = {"X-UCM-Conf", 10};
	const char *hdr_value = NULL;
	char *parse = NULL;
	char *name = NULL;
	char *value = NULL;
	AST_DECLARE_APP_ARGS(args,
		AST_APP_ARG(subnames[20]);
	);
	int idx = 0;

	hdr_value = ast_sip_rdata_get_header_value(rdata, hdr_str);
	if (!hdr_value) {
		return;
	}

	parse = ast_strdupa(hdr_value);
	AST_NONSTANDARD_APP_ARGS(args, parse, ';');
	if (args.argc == 0 || ARRAY_LEN(args.subnames) <= args.argc) {
		/* Too few or too many subnames */
		return;
	}

	for (idx = 0; idx < args.argc; idx++) {
		name = args.argv[idx];
		value = strchr(name, '=');
		if (value) {
			*(value++) = '\0';
		}

		if (!strcasecmp(name, "update-subscription-call-id")) {
			if (!ast_strlen_zero(value)) {
				ast_sip_subscription_update_transport(rdata->tp_info.transport, value);
			}
		}
	}
}
#endif

static pj_bool_t session_reinvite_on_rx_request(pjsip_rx_data *rdata)
{
	pjsip_dialog *dlg;
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	pjsip_rdata_sdp_info *sdp_info;

#ifdef GRANDSTREAM_NETWORKS
	ast_debug(1, "session_reinvite_on_rx_request enter...\n");
#endif

	if (rdata->msg_info.msg->line.req.method.id != PJSIP_INVITE_METHOD ||
		!(dlg = pjsip_ua_find_dialog(&rdata->msg_info.cid->id, &rdata->msg_info.to->tag, &rdata->msg_info.from->tag, PJ_FALSE)) ||
		!(session = ast_sip_dialog_get_session(dlg)) ||
		!session->channel) {
		return PJ_FALSE;
	}

#ifdef GRANDSTREAM_NETWORKS
	session_handle_ucm_conf_header(rdata);

#if 0
/*Start added by zyxia for Bug 205312,RTSP dont support reinvite*/
	ast_log(LOG_NOTICE, "session :%p,is_rtsp_call:%d\n",session,session->is_rtsp_call);
	if (session->is_rtsp_call && !session->is_sfu) {//added by zyxia for Bug 210478,2021.12.15
		/*  this is a  re-invite, so reject it */
			pjsip_tx_data *tdata1;
		if (pjsip_dlg_create_response(dlg, rdata, 488, NULL, &tdata1) == PJ_SUCCESS) {
			if (pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata1, NULL, NULL) != PJ_SUCCESS) {
				pjsip_tx_data_dec_ref(tdata1);
			}
		}

		return PJ_TRUE;
	}
/*End added by zyxia for Bug 205312,RTSP dont support reinvite*/
#endif
#endif

	if (session->deferred_reinvite) {
		pj_str_t key, deferred_key;
		pjsip_tx_data *tdata;

		/* We use memory from the new request on purpose so the deferred reinvite pool does not grow uncontrollably */
		pjsip_tsx_create_key(rdata->tp_info.pool, &key, PJSIP_ROLE_UAS, &rdata->msg_info.cseq->method, rdata);
		pjsip_tsx_create_key(rdata->tp_info.pool, &deferred_key, PJSIP_ROLE_UAS, &session->deferred_reinvite->msg_info.cseq->method,
			session->deferred_reinvite);

		/* If this is a retransmission ignore it */
		if (!pj_strcmp(&key, &deferred_key)) {
			return PJ_TRUE;
		}

		/* Otherwise this is a new re-invite, so reject it */
		if (pjsip_dlg_create_response(dlg, rdata, 491, NULL, &tdata) == PJ_SUCCESS) {
			if (pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL) != PJ_SUCCESS) {
				pjsip_tx_data_dec_ref(tdata);
			}
		}

		return PJ_TRUE;
	}

	if (!(sdp_info = pjsip_rdata_get_sdp_info(rdata)) ||
		(sdp_info->sdp_err != PJ_SUCCESS)) {
		return PJ_FALSE;
	}

	if (!sdp_info->sdp) {
		const pjmedia_sdp_session *local = NULL;
		int i;

		ast_queue_unhold(session->channel);

#ifdef GRANDSTREAM_NETWORKS
		pjsip_dlg_inc_lock(session->inv_session->dlg);
#endif
		pjmedia_sdp_neg_get_active_local(session->inv_session->neg, &local);
		if (!local) {
#ifdef GRANDSTREAM_NETWORKS
			pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif
			return PJ_FALSE;
		}

		/*
		 * Some devices indicate hold with deferred SDP reinvites (i.e. no SDP in the reinvite).
		 * When hold is initially indicated, we
		 * - Receive an INVITE with no SDP
		 * - Send a 200 OK with SDP, indicating sendrecv in the media streams
		 * - Receive an ACK with SDP, indicating sendonly in the media streams
		 *
		 * At this point, the pjmedia negotiator saves the state of the media direction so that
		 * if we are to send any offers, we'll offer recvonly in the media streams. This is
		 * problematic if the device is attempting to unhold, though. If the device unholds
		 * by sending a reinvite with no SDP, then we will respond with a 200 OK with recvonly.
		 * According to RFC 3264, if an offerer offers recvonly, then the answerer MUST respond
		 * with sendonly or inactive. The result of this is that the stream is not off hold.
		 *
		 * Therefore, in this case, when we receive a reinvite while the stream is on hold, we
		 * need to be sure to offer sendrecv. This way, the answerer can respond with sendrecv
		 * in order to get the stream off hold. If this is actually a different purpose reinvite
		 * (like a session timer refresh), then the answerer can respond to our sendrecv with
		 * sendonly, keeping the stream on hold.
		 */
		for (i = 0; i < local->media_count; ++i) {
			pjmedia_sdp_media *m = local->media[i];
			pjmedia_sdp_attr *recvonly;
			pjmedia_sdp_attr *inactive;

			recvonly = pjmedia_sdp_attr_find2(m->attr_count, m->attr, "recvonly", NULL);
			inactive = pjmedia_sdp_attr_find2(m->attr_count, m->attr, "inactive", NULL);
			if (recvonly || inactive) {
				pjmedia_sdp_attr *to_remove = recvonly ?: inactive;
				pjmedia_sdp_attr *sendrecv;

				pjmedia_sdp_attr_remove(&m->attr_count, m->attr, to_remove);

				sendrecv = pjmedia_sdp_attr_create(session->inv_session->pool, "sendrecv", NULL);
				pjmedia_sdp_media_add_attr(m, sendrecv);
			}
		}
#ifdef GRANDSTREAM_NETWORKS
		pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif
		return PJ_FALSE;
	}

#ifdef GRANDSTREAM_NETWORKS
	int session_media_image = 0, session_media_audio = 0;
	session_media_image = session_check_sdp_media_type(session, sdp_info->sdp, "image");
	session_media_audio = session_check_sdp_media_type(session, sdp_info->sdp, "audio");
	if (session_media_image && session_media_audio) {
		session->is_image_audio_two_media = 1;
	} else {
		session->is_image_audio_two_media = 0;
	}
	/*Start added by zyxia for Bug 187744 for faxsending show error when receive audio reinvite too ealry,the spandsp did report the status,2020.12.7 */
	if (session->is_fax_sending_request && session_media_audio && !session_media_image ){
		ast_log(LOG_NOTICE, "session name:%s,need delay handler the reinvite wait for faxsending status\n",ast_channel_name(session->channel));
		usleep(300000);
	}
	/*End added by zyxia for Bug 187744 for faxsending show error when receive audio reinvite too ealry,the spandsp did report the status,2020.12.7 */

	ast_log(LOG_NOTICE, "name:%s,session_media_image:%d,session_media_audio:%d\n",ast_channel_name(session->channel),session_media_image,session_media_audio);

	const char *from_trunk = NULL;

	if (session->is_intelligent_route) {
		ast_channel_set_is_sdp_image_reinvite(session->channel, 1); //Added by pxie@grandstream.cn for bug[46127], 2015-07-20.
	}

	from_trunk = pbx_builtin_getvar_helper(session->channel, "FROM_TRUNK");
	const char *calleenum = pbx_builtin_getvar_helper(session->channel, "EXTTOCALL");
	const char *fax_detect = pbx_builtin_getvar_helper(session->channel, "enable_fax_detect");
	const char *fax_intelligent_route = pbx_builtin_getvar_helper(session->channel, "fax_intelligent_route");
	const char *fax_intelligent_route_destination = pbx_builtin_getvar_helper(session->channel, "fax_intelligent_route_destination");
	const char *fax_to_extension = pbx_builtin_getvar_helper(session->channel, "fax_to_extension");
	ast_log(LOG_NOTICE, "reinvite:channlename:%s, from_trunk:%s,calleenum:%s,fax_detect:%s,fax_intelligent_route:%s,fax_intelligent_route_destination:%s,fax_to_extension:%s,faxcontext:%s,session_media_image:%d,vfax_flag:%d\n", 
		ast_channel_name(session->channel),from_trunk,calleenum,fax_detect,fax_intelligent_route,fax_intelligent_route_destination,fax_to_extension,session->endpoint->faxcontext,session_media_image,ast_channel_get_vfax_flag(session->channel));

	if (!session->fax_alreay_work
		&& from_trunk && !strcasecmp(from_trunk, "yes")
		&& fax_detect && !strcasecmp(fax_detect, "yes")
		&& fax_intelligent_route_destination
		&& !strcasecmp(session->endpoint->faxcontext, "faxcontext")
		&& session_media_image
		&& calleenum && !strcasecmp(calleenum, fax_intelligent_route_destination)
		&& fax_to_extension && !strcasecmp(fax_to_extension, "extension")
		&& !ast_channel_get_vfax_flag(session->channel)) {
		session->fax_alreay_work = 1;
	}

	if (!session->fax_alreay_work
		&& from_trunk && !strcasecmp(from_trunk, "yes")
		&& fax_detect && !strcasecmp(fax_detect, "yes")
		&& fax_intelligent_route_destination
		&& !strcasecmp(session->endpoint->faxcontext, "faxcontext")
		&& session_media_image
		&& !ast_channel_get_vfax_flag(session->channel)) {
		ast_verb(2, "Redirecting '%s' to fax extension due to T.38 Re-Invite 'image'", ast_channel_name(session->channel));
		pbx_builtin_setvar_helper(session->channel, "FAXEXTEN", ast_channel_exten(session->channel));

		/*Start added by zyxia for Bug 257462 ,2023.4.19*/
		const char *auto_answer = pbx_builtin_getvar_helper(session->channel, "auto_answer");
		if( ast_channel_name(session->channel) && strstr(ast_channel_name(session->channel),"trunk")
			&& !ast_strlen_zero(auto_answer) && !strcasecmp(auto_answer,"yes")){
			pbx_builtin_setvar_helper(session->channel, "FAX_WAIT_END", "yes");
			ast_cdr_setvar(ast_channel_name(session->channel),"special_action_type","AUTO_ANSWER_TO_FAX");//added by zyxia for Bug 257611,2023.4.28
		}
		/*End added by zyxia for Bug 257462 ,2023.4.19*/

		if (fax_intelligent_route && !strcasecmp(fax_intelligent_route, "yes")) {
			pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE", "yes");
			pbx_builtin_setvar_helper(session->channel, "_FAX_INTELLIGENT_ROUTE_DETINATION", fax_intelligent_route_destination);
			session->is_intelligent_route = 1;
			ast_channel_set_is_sdp_image_reinvite(session->channel, 1); //Added by pxie@grandstream.cn for bug[46127], 2015-07-20.
		}

		/*Start Bug 298504,Start added by zyxia,2024.6.12 */
		if (!session->endpoint->media.t38.enabled) {
			pjsip_tx_data *tdata;
			if (pjsip_dlg_create_response(dlg, rdata, 488, NULL, &tdata) == PJ_SUCCESS) {
				if (pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL) != PJ_SUCCESS) {
					pjsip_tx_data_dec_ref(tdata);
				}
			}
		}
		else {
			pjsip_tx_data *tdata;
			if (pjsip_dlg_create_response(dlg, rdata, 200, NULL, &tdata) == PJ_SUCCESS) {
				if (pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL) != PJ_SUCCESS) {
					pjsip_tx_data_dec_ref(tdata);
				}
			}
		}
		/*End Bug 298504,Start added by zyxia,2024.6.12 */

		if (ast_async_goto(session->channel, session->endpoint->faxcontext, "fax", 1)) {
			ast_log(LOG_NOTICE, "Failed to async goto '%s' into fax of '%s'\n", ast_channel_name(session->channel), session->endpoint->faxcontext);
		} else {
			session->fax_alreay_work = 1;
		}

		return PJ_TRUE;
	}

	/*! Begin: Added by pxie@grandstream.cn for bug[46209, 46202], 2015-07-17.*/
	if (session && session->channel && !ast_channel_get_vfax_flag(session->channel)) {
		struct ast_bridge_channel *peer_bridge_channel = ast_channel_get_bridge_channel(session->channel);
		struct ast_channel *peer_chan = ast_bridge_channel_get_channel_from_bridge_channel(peer_bridge_channel);
		if (!peer_chan && session_media_image && session->is_intelligent_route) {
			if(peer_bridge_channel)
				ao2_ref(peer_bridge_channel, -1);
			return PJ_TRUE;
		}

		if (session_media_image && !peer_chan) {
			ast_log(LOG_VERBOSE, "Peer channel don't bridge success, wait bridge finished!\n");
			if(peer_bridge_channel)
				ao2_ref(peer_bridge_channel, -1);
			return PJ_TRUE;
		}

		if(peer_bridge_channel)
			ao2_ref(peer_bridge_channel, -1);

	}
	/*! End: Added by pxie@grandstream.cn for bug[46209, 46202], 2015-07-17.*/

	if (session->is_direct_media) {
		return PJ_FALSE;
	}
#endif


#ifdef GRANDSTREAM_NETWORKS
	//Bug 280728 ,reinvite sdp dont need handle in sdp_requires_deferral,just t38 reinvite need handle
	if (session_media_image && !sdp_requires_deferral(session, sdp_info->sdp)) {
		return PJ_FALSE;
	}
	else if(!session_media_image) {
		return PJ_FALSE;
	}

#else
	if (!sdp_requires_deferral(session, sdp_info->sdp)) {
		return PJ_FALSE;
	}
#endif
	pjsip_rx_data_clone(rdata, 0, &session->deferred_reinvite);

	return PJ_TRUE;
}

void ast_sip_session_resume_reinvite(struct ast_sip_session *session)
{
	if (!session->deferred_reinvite) {
		return;
	}

	if (session->channel) {
		pjsip_endpt_process_rx_data(ast_sip_get_pjsip_endpoint(),
			session->deferred_reinvite, NULL, NULL);
	}
	pjsip_rx_data_free_cloned(session->deferred_reinvite);
	session->deferred_reinvite = NULL;
}

static pjsip_module session_reinvite_module = {
	.name = { "Session Re-Invite Module", 24 },
	.priority = PJSIP_MOD_PRIORITY_UA_PROXY_LAYER - 1,
	.on_rx_request = session_reinvite_on_rx_request,
};


void ast_sip_session_send_request_with_cb(struct ast_sip_session *session, pjsip_tx_data *tdata,
		ast_sip_session_response_cb on_response)
{
#ifdef GRANDSTREAM_NETWORKS
	if (!session || !tdata){
		ast_log(LOG_ERROR, "ast_sip_session_send_request_with_cb session invalid params!\n");
		return;
	}
#endif
	pjsip_inv_session *inv_session = session->inv_session;

	/* For every request except BYE we disallow sending of the message when
	 * the session has been disconnected. A BYE request is special though
	 * because it can be sent again after the session is disconnected except
	 * with credentials.
	 */
	if (inv_session->state == PJSIP_INV_STATE_DISCONNECTED &&
		tdata->msg->line.req.method.id != PJSIP_BYE_METHOD) {
		return;
	}

	ast_sip_mod_data_set(tdata->pool, tdata->mod_data, session_module.id,
			     MOD_DATA_ON_RESPONSE, on_response);
#ifdef GRANDSTREAM_NETWORKS
	struct pjsip_request_line req = tdata->msg->line.req;
	/*add by wxu@grandstream.cn for add sip_invite_failed*/
	/*! begin: Added by pxie@grandstream.cn for bug[42228], 2015-04-13*/
	if(session && session->channel && does_method_match(&req.method.name, "BYE")) {
	/*! end:   Added by pxie@grandstream.cn for bug[42228], 2015-04-13*/
		/*add by zhrliu for bug[44336] 2015-6-10*/
		if(AST_CAUSE_NORMAL_CLEARING != ast_channel_hangupcause(session->channel)) {
			ast_sip_report_invite_failed(session,NULL,tdata,"BYE",0,ast_cause2str(ast_channel_hangupcause(session->channel)));
		}
	}
	/*end*/

	/* Add by bfdu, for wave we set timeout longer than normal, 2021.10.26 */
	/* Modify by bfdu for Bug 311455, 2024.8.13
	 * Remove SFU condition, for p2p call if transaction timeout is 32s, but tcp keepalive is 100s
	 * If transaction timeout happen first, the call wil be hangup, so change p2p timeout to 120 too.
	 * But it will not very match RFC request
	 */
	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		tdata->send_to_wave_in_sfu = 1;
	}

#ifdef BUILD_ON_CLOUD_PBX
	ast_sip_add_trunk_host_header(session->endpoint, tdata);
	ast_sip_add_sbc_endpoint_address_header(session->sbc_endpoint_addr, tdata);
	ast_sip_add_sbc_rand_addr_header(session->endpoint, tdata);
#endif
#endif

	handle_outgoing_request(session, tdata);
	pjsip_inv_send_msg(session->inv_session, tdata);

	return;
}

void ast_sip_session_send_request(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	ast_sip_session_send_request_with_cb(session, tdata, NULL);
}

int ast_sip_session_create_invite(struct ast_sip_session *session, pjsip_tx_data **tdata)
{
	pjmedia_sdp_session *offer;
#ifdef GRANDSTREAM_NETWORKS
	int  cc_core_id;
#endif

#ifdef GRANDSTREAM_NETWORKS
	if (!session->is_direct_media && !session->joint_ipvt) {
		if (!(offer = create_local_sdp(session->inv_session, session, NULL))) {
			pjsip_inv_terminate(session->inv_session, 500, PJ_FALSE);
			if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
				ast_log(LOG_ERROR, "[session:%p] Failed to create invite, session has been already disconnected", session);
			} else {
				ao2_ref(session, -1);
			}
			return -1;
		}
     
		pjsip_inv_set_local_sdp(session->inv_session, offer);
	}
#else
	if (!(offer = create_local_sdp(session->inv_session, session, NULL))) {
		pjsip_inv_terminate(session->inv_session, 500, PJ_FALSE);
		return -1;
	}

	pjsip_inv_set_local_sdp(session->inv_session, offer);
#endif
	pjmedia_sdp_neg_set_prefer_remote_codec_order(session->inv_session->neg, PJ_FALSE);
#ifdef PJMEDIA_SDP_NEG_ANSWER_MULTIPLE_CODECS
	if (!session->endpoint->preferred_codec_only) {
		pjmedia_sdp_neg_set_answer_multiple_codecs(session->inv_session->neg, PJ_TRUE);
	}
#endif
#ifdef GRANDSTREAM_NETWORKS   /* Added by yfyang, for ccss feature, get notify user to set as target user */
	if (ast_cc_is_recall(session->channel, &cc_core_id, "PJSIP")) {
		char   device_name[AST_CHANNEL_NAME] = {0};
		struct ast_cc_monitor *recall_monitor;

		ast_channel_get_device_name(session->channel, device_name, sizeof(device_name));
		if ((recall_monitor = ast_cc_get_monitor_by_recall_core_id(cc_core_id, device_name))) {
			struct ast_sip_cc_notify_user_data  notify_user_data = {
				.monitor = recall_monitor,
			};

			if (0 == find_sub_handler_for_supplement_exten(AST_SIP_CCSS_NOTIFY_USER, NULL, &notify_user_data)) {
				if (session->inv_session && session->inv_session->dlg) {
					pjsip_dialog   *dlg = session->inv_session->dlg;
					pjsip_sip_uri  *sip_uri;

                    ast_log(LOG_NOTICE, "ast_sip_session_create_invite, get notify user[%s].\n", notify_user_data.notify_user);
					if (PJSIP_URI_SCHEME_IS_SIP(dlg->target) || PJSIP_URI_SCHEME_IS_SIPS(dlg->target)) {
						sip_uri = pjsip_uri_get_uri(dlg->target);
						pj_strdup2(dlg->pool, &sip_uri->user, notify_user_data.notify_user);
					}

					if (PJSIP_URI_SCHEME_IS_SIP(dlg->remote.info->uri) || PJSIP_URI_SCHEME_IS_SIPS(dlg->remote.info->uri)) {
						sip_uri = pjsip_uri_get_uri(dlg->remote.info->uri);
						pj_strdup2(dlg->pool, &sip_uri->user, notify_user_data.notify_user);
					}
				}
			}

			ao2_t_ref(recall_monitor, -1, "Got the URI we need so unreffing monitor");
		}
	}
#endif


	/*
	 * We MUST call set_from_header() before pjsip_inv_invite.  If we don't, the
	 * From in the initial INVITE will be wrong but the rest of the messages will be OK.
	 */
	set_from_header(session);

	if (pjsip_inv_invite(session->inv_session, tdata) != PJ_SUCCESS) {
		return -1;
	}

	return 0;
}

static int datastore_hash(const void *obj, int flags)
{
	const struct ast_datastore *datastore = obj;
	const char *uid = flags & OBJ_KEY ? obj : datastore->uid;

	ast_assert(uid != NULL);

	return ast_str_hash(uid);
}

static int datastore_cmp(void *obj, void *arg, int flags)
{
	const struct ast_datastore *datastore1 = obj;
	const struct ast_datastore *datastore2 = arg;
	const char *uid2 = flags & OBJ_KEY ? arg : datastore2->uid;

	ast_assert(datastore1->uid != NULL);
	ast_assert(uid2 != NULL);

	return strcmp(datastore1->uid, uid2) ? 0 : CMP_MATCH | CMP_STOP;
}

#ifdef GRANDSTREAM_NETWORKS
static void session_media_state_set_ignore_avs_response(struct ast_sip_session_media_state *media_state, int ignore)
{
	int i = 0;
	struct ast_sip_session_media *media = NULL;

	if (!media_state) {
		return;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&media_state->sessions); ++i) {
		media = AST_VECTOR_GET(&media_state->sessions, i);
		if (!media) {
			continue;
		}

		if (media->rtp) {
			ast_rtp_instance_set_ignore_avs_response(media->rtp, (unsigned int)ignore);
		}
	}
}
#endif

static void session_destructor(void *obj)
{
	struct ast_sip_session *session = obj;
	struct ast_sip_session_delayed_request *delay;
	const char *endpoint_name = session->endpoint ?
		ast_sorcery_object_get_id(session->endpoint) : "<none>";

	ast_debug(3, "Destroying SIP session with endpoint %s, session[%p].\n", endpoint_name, session);

	ast_test_suite_event_notify("SESSION_DESTROYING",
		"Endpoint: %s\r\n"
		"AOR: %s\r\n"
		"Contact: %s"
		, endpoint_name
		, session->aor ? ast_sorcery_object_get_id(session->aor) : "<none>"
		, session->contact ? ast_sorcery_object_get_id(session->contact) : "<none>"
		);
#ifdef GRANDSTREAM_NETWORKS 
	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_REMOVE(&sip_sessions, session, next);
	AST_LIST_UNLOCK(&sip_sessions);
#endif

	/* fire session destroy handler */
	handle_session_destroy(session);

	/* remove all registered supplements */
	ast_sip_session_remove_supplements(session);
	AST_LIST_HEAD_DESTROY(&session->supplements);

	ast_taskprocessor_unreference(session->serializer);
	ao2_cleanup(session->datastores);
#ifdef GRANDSTREAM_NETWORKS
	if (session->ignore_avs_response) {
		session_media_state_set_ignore_avs_response(session->active_media_state, 1);
		session_media_state_set_ignore_avs_response(session->active_participant_media_state, 1);
	}
#endif
	ast_sip_session_media_state_free(session->active_media_state);
	ast_sip_session_media_state_free(session->pending_media_state);
#ifdef GRANDSTREAM_NETWORKS
	ast_sip_session_media_state_free(session->active_participant_media_state);
	ast_sip_session_media_state_free(session->pending_participant_media_state);
	if (session->participant_bundled_media) {
		ao2_ref(session->participant_bundled_media, -1);
		session->participant_bundled_media = NULL;
	}
	if (session->play_sound_task) {
		ast_sip_sched_task_cancel(session->play_sound_task);
		ao2_ref(session->play_sound_task, -1);
		session->play_sound_task = NULL;
	}

	if (session->reject_nat_call_task) {
		ast_sip_sched_task_cancel(session->reject_nat_call_task);
		ao2_ref(session->reject_nat_call_task, -1);
		session->reject_nat_call_task = NULL;
	}

	if (session->send_real_answered) {
		ast_sip_sched_task_cancel(session->send_real_answered);
		ao2_ref(session->send_real_answered, -1);
		session->send_real_answered = NULL;
	}

	if (session->conf_link) {
		ast_free(session->conf_link);
		session->conf_link = NULL;
	}
	if (session->trunk_channel_id) {
		ast_free(session->trunk_channel_id);
		session->trunk_channel_id = NULL;
	}
	if (session->replace_channel_id) {
		ast_free(session->replace_channel_id);
		session->replace_channel_id = NULL;
	}

	if (session->trunk_conf_num) {
		ast_free(session->trunk_conf_num);
		session->trunk_conf_num = NULL;
	}
	/*BEGIN: Add by weidu, for memory free*/
	if (session->remote_trunk_channel_id) {
		ast_free(session->remote_trunk_channel_id);
		session->remote_trunk_channel_id = NULL;
	}
	/*END: Add by weidu, for memory free*/

	if (session->gs_state_clone_data) {
		ao2_cleanup(session->gs_state_clone_data);
		session->gs_state_clone_data = NULL;
	}
#endif
	
	while ((delay = AST_LIST_REMOVE_HEAD(&session->delayed_requests, next))) {
		delayed_request_free(delay);
	}
	ast_party_id_free(&session->id);
	ao2_cleanup(session->endpoint);
	ao2_cleanup(session->aor);
	ao2_cleanup(session->contact);
	ao2_cleanup(session->direct_media_cap);

	ast_dsp_free(session->dsp);
	ast_free(session->active_contact_addr);
	ast_free(session->trans_type);
	
	if (session->inv_session) {
#ifdef GRANDSTREAM_NETWORKS
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_dlg_inc_lock(session->inv_session->dlg);
		pjsip_inv_dec_ref(session->inv_session);
		pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif
#endif
		pjsip_dlg_dec_session(session->inv_session->dlg, &session_module);
	}

	ast_test_suite_event_notify("SESSION_DESTROYED", "Endpoint: %s", endpoint_name);
}

/*! \brief Destructor for SIP channel */
static void sip_channel_destroy(void *obj)
{
	struct ast_sip_channel_pvt *channel = obj;

	ao2_cleanup(channel->pvt);
	ao2_cleanup(channel->session);
}

struct ast_sip_channel_pvt *ast_sip_channel_pvt_alloc(void *pvt, struct ast_sip_session *session)
{
	struct ast_sip_channel_pvt *channel = ao2_alloc(sizeof(*channel), sip_channel_destroy);

	if (!channel) {
		return NULL;
	}

	ao2_ref(pvt, +1);
	channel->pvt = pvt;
	ao2_ref(session, +1);
	channel->session = session;

	return channel;
}

struct ast_sip_session *ast_sip_session_alloc(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, pjsip_inv_session *inv_session, pjsip_rx_data *rdata)
{
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	struct ast_sip_session *ret_session;
	int dsp_features = 0;

	session = ao2_alloc(sizeof(*session), session_destructor);
	if (!session) {
		return NULL;
	}

	AST_LIST_HEAD_INIT(&session->supplements);
	AST_LIST_HEAD_INIT_NOLOCK(&session->delayed_requests);
	ast_party_id_init(&session->id);

#ifdef GRANDSTREAM_NETWORKS
	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_INSERT_TAIL(&sip_sessions, session, next);
	AST_LIST_UNLOCK(&sip_sessions);
#endif

	session->direct_media_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!session->direct_media_cap) {
		return NULL;
	}
	session->datastores = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
		DATASTORE_BUCKETS, datastore_hash, NULL, datastore_cmp);
	if (!session->datastores) {
		return NULL;
	}
	session->active_media_state = ast_sip_session_media_state_alloc();
	if (!session->active_media_state) {
		return NULL;
	}
	session->pending_media_state = ast_sip_session_media_state_alloc();
	if (!session->pending_media_state) {
		return NULL;
	}

	if (endpoint->dtmf == AST_SIP_DTMF_INBAND || endpoint->dtmf == AST_SIP_DTMF_AUTO) {
		dsp_features |= DSP_FEATURE_DIGIT_DETECT;
	}
	if (endpoint->faxdetect) {
		dsp_features |= DSP_FEATURE_FAX_DETECT;
	}
	if (dsp_features) {
		session->dsp = ast_dsp_new();
		if (!session->dsp) {
			return NULL;
		}

		ast_dsp_set_features(session->dsp, dsp_features);
	}

	session->endpoint = ao2_bump(endpoint);

	if (rdata) {
		/*
		 * We must continue using the serializer that the original
		 * INVITE came in on for the dialog.  There may be
		 * retransmissions already enqueued in the original
		 * serializer that can result in reentrancy and message
		 * sequencing problems.
		 */
		session->serializer = ast_sip_get_distributor_serializer(rdata);
	} else {
		char tps_name[AST_TASKPROCESSOR_MAX_NAME + 1];

		/* Create name with seq number appended. */
		ast_taskprocessor_build_name(tps_name, sizeof(tps_name), "pjsip/outsess/%s",
			ast_sorcery_object_get_id(endpoint));

		session->serializer = ast_sip_create_serializer(tps_name);
	}
	if (!session->serializer) {
		return NULL;
	}
	ast_sip_dialog_set_serializer(inv_session->dlg, session->serializer);
	ast_sip_dialog_set_endpoint(inv_session->dlg, endpoint);
	pjsip_dlg_inc_session(inv_session->dlg, &session_module);
	inv_session->mod_data[session_module.id] = ao2_bump(session);
	session->contact = ao2_bump(contact);
	session->inv_session = inv_session;
#ifdef GRANDSTREAM_NETWORKS
#ifdef HAVE_PJSIP_INV_SESSION_REF
	if (pjsip_inv_add_ref(inv_session) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Can't increase the inv_session reference counter\n");
		return NULL;
	}
#endif
#endif

	session->dtmf = endpoint->dtmf;
#ifdef GRANDSTREAM_NETWORKS
	session->core_id = -1;
	session->cc_offer = 0;
	session->cc_service = AST_CC_NONE;
	session->cc_subs_user.slen = 0;
	session->cc_subs_user.ptr = NULL;
	session->cc_subs_uri = NULL;
	session->is_outgoing = 0;
	session->initial_authed = 0;

	session->line_index = 0;   /* Added by yfyang, for sca feature */
	session->has_insert_sca_user = 0;
	session->port_group_id = ast_random();;
	set_session_msid_count(session);
	memset(session->user_agent, 0, sizeof(session->user_agent));
	memset(session->sbc_endpoint_addr, 0, sizeof(session->sbc_endpoint_addr));
	memset(session->sbc_addr, 0, sizeof(session->sbc_addr));

	session->sfu_video_codec = AST_SFU_VIDEO_H264;
	session->active_participant_media_state = ast_sip_session_media_state_alloc();
	if (!session->active_participant_media_state) {
		return NULL;
	}
	session->pending_participant_media_state = ast_sip_session_media_state_alloc();
	if (!session->pending_participant_media_state) {
		return NULL;
	}
	/*start added by zyxia for extension detect,2020.12.21*/
	session->wFaxDetect = endpoint->faxdetect;
	/*end added by zyxia for extension detect,2020.12.21*/

	session->active_contact_addr = NULL;
	session->trans_type = NULL;
#endif
	if (ast_sip_session_add_supplements(session)) {
		/* Release the ref held by session->inv_session */
#ifndef GRANDSTREAM_NETWORKS
		ao2_ref(session, -1);
#endif
		return NULL;
	}

	/* Fire seesion begin handlers */
	handle_session_begin(session);

#ifdef GRANDSTREAM_NETWORKS
	if (inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Invite Session %p already DISCONNECTED [reason=%d (%s)], session[%p] refcnt[%d]\n",
			inv_session,
			inv_session->cause,
			pjsip_get_status_text(inv_session->cause)->ptr, session, ao2_ref(session, 0));
		/* Release the ref held by session->inv_session */
		return NULL;
	}
#endif
	
	/* Avoid unnecessary ref manipulation to return a session */
	ret_session = session;
	session = NULL;

	return ret_session;
}

/*! \brief struct controlling the suspension of the session's serializer. */
struct ast_sip_session_suspender {
	ast_cond_t cond_suspended;
	ast_cond_t cond_complete;
	int suspended;
	int complete;
};

static void sip_session_suspender_dtor(void *vdoomed)
{
	struct ast_sip_session_suspender *doomed = vdoomed;

	ast_cond_destroy(&doomed->cond_suspended);
	ast_cond_destroy(&doomed->cond_complete);
}

/*!
 * \internal
 * \brief Block the session serializer thread task.
 *
 * \param data Pushed serializer task data for suspension.
 *
 * \retval 0
 */
static int sip_session_suspend_task(void *data)
{
	struct ast_sip_session_suspender *suspender = data;

	ao2_lock(suspender);

	/* Signal that the serializer task is now suspended. */
	suspender->suspended = 1;
	ast_cond_signal(&suspender->cond_suspended);

	/* Wait for the serializer suspension to be completed. */
	while (!suspender->complete) {
		ast_cond_wait(&suspender->cond_complete, ao2_object_get_lockaddr(suspender));
	}

	ao2_unlock(suspender);
	ao2_ref(suspender, -1);

	return 0;
}

void ast_sip_session_suspend(struct ast_sip_session *session)
{
	struct ast_sip_session_suspender *suspender;
	int res;

	ast_assert(session->suspended == NULL);

	if (ast_taskprocessor_is_task(session->serializer)) {
		/* I am the session's serializer thread so I cannot suspend. */
		return;
	}

	if (ast_taskprocessor_is_suspended(session->serializer)) {
		/* The serializer already suspended. */
		return;
	}

	suspender = ao2_alloc(sizeof(*suspender), sip_session_suspender_dtor);
	if (!suspender) {
		/* We will just have to hope that the system does not deadlock */
		return;
	}
	ast_cond_init(&suspender->cond_suspended, NULL);
	ast_cond_init(&suspender->cond_complete, NULL);

	ao2_ref(suspender, +1);
	res = ast_sip_push_task(session->serializer, sip_session_suspend_task, suspender);
	if (res) {
		/* We will just have to hope that the system does not deadlock */
		ao2_ref(suspender, -2);
		return;
	}

	session->suspended = suspender;

	/* Wait for the serializer to get suspended. */
	ao2_lock(suspender);
	while (!suspender->suspended) {
		ast_cond_wait(&suspender->cond_suspended, ao2_object_get_lockaddr(suspender));
	}
	ao2_unlock(suspender);

	ast_taskprocessor_suspend(session->serializer);
}

void ast_sip_session_unsuspend(struct ast_sip_session *session)
{
	struct ast_sip_session_suspender *suspender = session->suspended;

	if (!suspender) {
		/* Nothing to do */
		return;
	}
	session->suspended = NULL;

	/* Signal that the serializer task suspension is now complete. */
	ao2_lock(suspender);
	suspender->complete = 1;
	ast_cond_signal(&suspender->cond_complete);
	ao2_unlock(suspender);

	ao2_ref(suspender, -1);

	ast_taskprocessor_unsuspend(session->serializer);
}

/*!
 * \internal
 * \brief Handle initial INVITE challenge response message.
 * \since 13.5.0
 *
 * \param rdata PJSIP receive response message data.
 *
 * \retval PJ_FALSE Did not handle message.
 * \retval PJ_TRUE Handled message.
 */
static pj_bool_t outbound_invite_auth(pjsip_rx_data *rdata)
{
	pjsip_transaction *tsx;
	pjsip_dialog *dlg;
	pjsip_inv_session *inv;
	pjsip_tx_data *tdata;
	struct ast_sip_session *session;

	if (rdata->msg_info.msg->line.status.code != 401
		&& rdata->msg_info.msg->line.status.code != 407) {
		/* Doesn't pertain to us. Move on */
		return PJ_FALSE;
	}

	tsx = pjsip_rdata_get_tsx(rdata);
	dlg = pjsip_rdata_get_dlg(rdata);
	if (!dlg || !tsx) {
		return PJ_FALSE;
	}

	if (tsx->method.id != PJSIP_INVITE_METHOD) {
		/* Not an INVITE that needs authentication */
		return PJ_FALSE;
	}

	inv = pjsip_dlg_get_inv_session(dlg);
#ifdef GRANDSTREAM_NETWORKS
	if (!inv) {
		return PJ_FALSE;
	}
#endif

	if (PJSIP_INV_STATE_CONFIRMED <= inv->state) {
		/*
		 * We cannot handle reINVITE authentication at this
		 * time because the reINVITE transaction is still in
		 * progress.
		 */
		ast_debug(1, "A reINVITE is being challenged.\n");
		return PJ_FALSE;
	}
	ast_debug(1, "Initial INVITE is being challenged.\n");

	session = inv->mod_data[session_module.id];

#ifdef GRANDSTREAM_NETWORKS
	if (session && session->initial_authed) {
		int status = rdata->msg_info.msg->line.status.code;
		ast_debug(1, "Initial auth failed, report invite failed.\n");
		ast_sip_report_invite_failed(session, rdata, NULL, "INVITE", status, ast_cause2str(hangup_sip2cause(status)));

		return PJ_FALSE;
	}
#endif

	if (ast_sip_create_request_with_auth(&session->endpoint->outbound_auths, rdata,
		tsx->last_tx, &tdata)) {
#ifdef GRANDSTREAM_NETWORKS
		int code = rdata->msg_info.msg->line.status.code;

		ast_debug(1, "auth failed, report invite failed.\n");
		ast_sip_report_invite_failed(session, rdata, NULL, "INVITE", code, ast_cause2str(hangup_sip2cause(code)));
#endif
		return PJ_FALSE;
	}

	/*
	 * Restart the outgoing initial INVITE transaction to deal
	 * with authentication.
	 */
	pjsip_inv_uac_restart(inv, PJ_FALSE);

#ifdef GRANDSTREAM_NETWORKS
	/* check teluri and Route, Modified by yfyang, for Bug 289964 */
	if (AST_SIP_TEL_URI_ENABLED == session->endpoint->tel_uri)
	{
		static const pj_str_t ROUTE_HNAME = { "Route", 5 };
		pjsip_route_hdr *route = pjsip_msg_find_hdr_by_name(tdata->msg, &ROUTE_HNAME, NULL);
		if (!route) {
			char orig_uri[PJSIP_MAX_URL_SIZE] = {0};

			snprintf(orig_uri, sizeof(orig_uri), "sip:%s:%d;transport=%s;lr", rdata->pkt_info.src_name, rdata->pkt_info.src_port, rdata->tp_info.transport->type_name);
			route = pjsip_parse_hdr(tdata->pool, &ROUTE_HNAME, orig_uri, strlen(orig_uri), NULL);
			if (route) {
				pjsip_param *p;

				p = PJ_POOL_ALLOC_T(dlg->pool, pjsip_param);
				p->name = pj_str("removed");
				p->value.slen = 0;
				pj_list_insert_before(&route->other_param, p);

				ast_log(LOG_NOTICE, "outbound_invite_auth, route[%s]", orig_uri);
				pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)route);
			}
		}
	}

	session->initial_authed = 1;
#endif

	ast_sip_session_send_request(session, tdata);
	return PJ_TRUE;
}

static pjsip_module outbound_invite_auth_module = {
	.name = {"Outbound INVITE Auth", 20},
	.priority = PJSIP_MOD_PRIORITY_DIALOG_USAGE,
	.on_rx_response = outbound_invite_auth,
};

#ifdef GRANDSTREAM_NETWORKS

static int find_topology_audio_stream(struct ast_stream_topology* topology) 
{
	int i = 0;
	int ret = 0;
	
	for (i = 0; i < ast_stream_topology_get_count(topology); ++ i) {
		struct ast_stream* tmp = ast_stream_topology_get_stream(topology, i);

		if (ast_stream_get_type(tmp) == AST_MEDIA_TYPE_AUDIO) { 
			ret = 1;
			break;
		}
	}

	return ret;
}


static void add_default_audio_stream(struct ast_sip_session *session, struct ast_stream_topology* pending_topology, struct ast_stream_topology* endpoint_topology)
{
	if (pending_topology == NULL || endpoint_topology == NULL) {
		ast_log(LOG_DEBUG, "add_default_audio_stream met error: invalid args\n");
		return;
	}

	int ret = find_topology_audio_stream(pending_topology);
	if (ret == 1) {
		ast_log(LOG_NOTICE, "pending_topology is already had audio stream\n");
		return;
	}
	
	int i = 0;

	struct ast_stream* tmp_audio_stream = NULL;
	for (i = 0; i < ast_stream_topology_get_count(endpoint_topology); ++ i) {
		struct ast_stream* tmp = ast_stream_topology_get_stream(endpoint_topology, i);
		if (ast_stream_get_type(tmp) == AST_MEDIA_TYPE_AUDIO) {
			tmp_audio_stream = ast_stream_clone(tmp, NULL);
			ast_stream_set_type(tmp_audio_stream, AST_MEDIA_TYPE_AUDIO);
			ast_stream_set_sub_type(tmp_audio_stream, AST_MEDIA_SUB_AUDIO_TYPE);
			if (ast_format_cap_has_audio_fec(ast_stream_get_formats(tmp_audio_stream))) {
				ast_sip_session_change_gs_fec_version(ast_stream_get_formats(tmp_audio_stream));
			}
			ast_sip_session_process_special_audio_codec(session, tmp_audio_stream, ast_stream_get_formats(tmp_audio_stream));
			break;
		}
	}

	int old_length = ast_stream_topology_get_count(pending_topology);
	struct ast_stream* stream_array[10];
	ast_stream_topology_append_stream(pending_topology, tmp_audio_stream);

	for (i = 0; i < old_length; ++ i) {
		struct ast_stream* tmp = ast_stream_clone(ast_stream_topology_get_stream(pending_topology, 0), NULL);
		stream_array[i] = tmp;
		ast_stream_topology_del_stream(pending_topology, 0);
	}

	for (i = 0; i < old_length; ++ i) {
		ast_stream_topology_append_stream(pending_topology, stream_array[i]);	
	}
}

static void append_audio_to_topology(struct ast_sip_session *session,
	struct ast_stream_topology *topology, struct ast_stream *stream)
{
	unsigned int i = 0;
	struct ast_stream *endpoint_audio_stream = NULL;
	struct ast_stream *clone_stream = NULL;

	if (stream) {
		ast_stream_topology_append_stream(topology, stream);	
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->endpoint->media.topology); ++ i) {
		endpoint_audio_stream = ast_stream_topology_get_stream(session->endpoint->media.topology, i);
		if (ast_stream_get_type(endpoint_audio_stream) == AST_MEDIA_TYPE_AUDIO) {
			continue;
		}

		clone_stream = ast_stream_clone(endpoint_audio_stream, NULL);
		ast_stream_set_type(clone_stream, AST_MEDIA_TYPE_AUDIO);
		ast_stream_set_sub_type(clone_stream, AST_MEDIA_SUB_AUDIO_TYPE);
		ast_stream_topology_append_stream(topology, clone_stream);	
		break;
	}

	return;
}

/*Start added by zyxia for Bug 261644 ,2023.6.7*/
int ast_sip_session_is_have_video_codec(struct ast_sip_session *session)
{
	pjmedia_sdp_session *remote_sdp = NULL;
	pjmedia_sdp_media *remote_media = NULL;
	int  i =0;

	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return 0;
	}
	ast_log(LOG_WARNING, "[session:%p] ast_sip_session_is_have_video_codec endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	//check local endpoint if have video codec
	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);
	if (ast_format_cap_count(caps) == 0 ||
		ast_sip_session_process_special_video_codec(session, caps)) {
		ast_log(LOG_WARNING, "[session:%p] Has no video codec, can't append video stream, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}
	else {
		//check remote endpoint  if have video media from inv_session neg
		if(session->inv_session->neg->state == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO)
			remote_sdp = session->inv_session->neg->neg_remote_sdp;
		else if (session->inv_session->neg->state == PJMEDIA_SDP_NEG_STATE_DONE)
			remote_sdp = session->inv_session->neg->active_remote_sdp;

		if(remote_sdp) {
			for (remote_media = NULL, i = 0; i < remote_sdp->media_count; ++i) {
				remote_media = remote_sdp->media[i];
				if (!remote_media) {
					ast_log(LOG_WARNING, "[session:%p] Disable media local_media is invalid for endpoint:%s\n",
						session, ast_sorcery_object_get_id(session->endpoint));
					return 0;
				}
				if (strncmp(remote_media->desc.media.ptr, "video", 5) == 0) {
					ast_log(LOG_DEBUG, "[session:%p] Remote also have video media endpoint:%s\n",
						session, ast_sorcery_object_get_id(session->endpoint));
					return 1;
				} 
			}
			return 0;
		}
	}

	return  1;
}
/*End added by zyxia for Bug 261644 ,2023.6.7*/

static void append_video_to_topology(struct ast_sip_session *session,
	struct ast_stream_topology *topology, struct ast_stream *stream, enum ast_media_sub_type sub_type)
{
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
	struct ast_stream *clone_stream = NULL;

	if (stream) {
		ast_stream_topology_append_stream(topology, stream);	
		return;
	}

	caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!caps) {
		return;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);
	if (ast_format_cap_count(caps) == 0 ||
		ast_sip_session_process_special_video_codec(session, caps)) {
		if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
			ast_log(LOG_WARNING, "[session:%p] Has no video codec, but it's wave desktop/web, append H264 codec, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_format_cap_append(caps, ast_format_h264, 0);
		} else {
			ast_log(LOG_WARNING, "[session:%p] Has no video codec, can't append video stream, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
	}

	clone_stream = ast_stream_alloc(ast_codec_media_type2str(AST_MEDIA_TYPE_VIDEO), AST_MEDIA_TYPE_VIDEO);
	if (!clone_stream) {
		return;
	}

	ast_format_cap_remove_video_fec(caps);
	ast_stream_set_formats(clone_stream, caps);
	ast_sip_session_set_video_format(session, clone_stream, VIDEO_RESOLUTION_360P);

	ast_stream_set_local_media_direction(clone_stream, INACTIVE);
	ast_stream_set_remote_media_direction(clone_stream, INACTIVE);
	switch (sub_type) {
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			ast_stream_set_metadata(clone_stream, "CONTENT", "main");
			ast_stream_set_sub_type(clone_stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			ast_stream_set_metadata(clone_stream, "CONTENT", "slides");
			ast_stream_set_sub_type(clone_stream, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
			break;
		default:
			break;
	}

	ast_stream_topology_append_stream(topology, clone_stream);

	return;
}

void ast_sip_session_append_stream_to_topology(struct ast_sip_session *session,
	struct ast_stream_topology *topology, struct ast_stream *stream, enum ast_media_sub_type type)
{
	if (!session || !topology) {
		return;
	}

	switch (type) {
		case AST_MEDIA_SUB_AUDIO_TYPE:
			append_audio_to_topology(session, topology, stream);
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			append_video_to_topology(session, topology, stream, type);
			break;
		default:
			break;
	}

	return;
}

static unsigned int judge_if_need_modify_topology(struct ast_sip_session *session, struct ast_stream_topology *topology)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (ast_stream_topology_get_count(topology) != 3) {
		ast_log(LOG_DEBUG, "[session:%p] Pending topology stream count is not three, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}
	
	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			return 1;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_AUDIO_TYPE && (i != 0)) {
			ast_log(LOG_DEBUG, "[session:%p] Audio stream is mismatch, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN && (i != 1)) {
			ast_log(LOG_DEBUG, "[session:%p] Main video stream is mismatch, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES && (i != 2)) {
			ast_log(LOG_DEBUG, "[session:%p] Slides video stream is mismatch, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		}
	}

	return 0;
}

static unsigned int is_invite_by_audio_confenerce(struct ast_sip_session *session, struct ast_channel *requestor)
{
	const char *is_sfu = NULL;
	const char *conf_room = NULL;

	if (!session || !requestor) {
		return 0;
	}

	if (session->is_sfu) {
		return 0;
	}

	is_sfu = pbx_builtin_getvar_helper(requestor, "IS_SFU");

	if (!ast_strlen_zero(is_sfu) && !strcasecmp(is_sfu, "yes")) {
		return 0;
	}

	conf_room = pbx_builtin_getvar_helper(requestor, "CONF_ROOM");
	if (!ast_strlen_zero(conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] It's audio confenerce:%s, endpoint:%s\n",
			session, conf_room, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	return 0;
}

/* Gswave P2P call need has three media line forcely */
void ast_sip_session_add_missed_line_for_gswave(struct ast_sip_session *session, struct ast_channel *requestor)
{
	unsigned int i = 0;
	unsigned int audio_appended = 0;
	unsigned int main_appended = 0;
	unsigned int slides_appended = 0;
	struct ast_stream_topology *topology = NULL;
	struct ast_stream_topology *final_topology = NULL;
	struct ast_stream *stream = NULL;
	struct ast_stream *audio_clone_stream = NULL;
	struct ast_stream *main_video_clone_stream = NULL;
	struct ast_stream *slides_video_clone_stream = NULL;

	if (!session || !session->pending_media_state) {
		return;
	}

	if (!session->is_webrtc_call) {
		return;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return;
	}

	if (is_invite_by_audio_confenerce(session, requestor)) {
		return;
	}

	topology = session->pending_media_state->topology;
	if (!topology) {
		ast_log(LOG_DEBUG, "[session:%p] add_missed_line_for_gswave has no pending topology for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/* Now gswave has audio/main video/slides video */
	if (!judge_if_need_modify_topology(session, topology)) {
		ast_log(LOG_DEBUG, "[session:%p] Pending topology already matched, don't procees for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	final_topology = ast_stream_topology_alloc();
	if (!final_topology) {
		ast_log(LOG_DEBUG, "[session:%p] Can't alloc topology for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); ++i) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_AUDIO_TYPE) {
			audio_clone_stream = ast_stream_clone(stream, NULL);
			if (i == 0) {
				ast_stream_topology_append_stream(final_topology, audio_clone_stream);	
				audio_appended = 1;
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Audio is not first postion in topology for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
			main_video_clone_stream = ast_stream_clone(stream, NULL);
			if (i == 1) {
				ast_stream_topology_append_stream(final_topology, main_video_clone_stream);	
				main_appended = 1;
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Main video is not second postion in topology for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}

		if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			slides_video_clone_stream = ast_stream_clone(stream, NULL);
			if (i == 2) {
				ast_stream_topology_append_stream(final_topology, slides_video_clone_stream);	
				slides_appended = 1;
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Slides video is not third postion in topology for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}
	}

	if (!audio_appended) {
		ast_log(LOG_DEBUG, "[session:%p] It's sees something very wrong, add audio stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_append_stream_to_topology(session, final_topology, audio_clone_stream, AST_MEDIA_SUB_AUDIO_TYPE);
	}

	if (!main_appended) {
		ast_log(LOG_DEBUG, "[session:%p] Has no main video, add stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_append_stream_to_topology(session, final_topology, main_video_clone_stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	}

	if (!slides_appended) {
		ast_log(LOG_DEBUG, "[session:%p] Has no slides video, add stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_append_stream_to_topology(session, final_topology, slides_video_clone_stream, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	}

	ast_stream_topology_free(session->pending_media_state->topology);
	session->pending_media_state->topology = final_topology;

	return;
}


static void set_ice_for_session(struct ast_sip_session *session, 
	struct ast_sip_contact *found_contact, const char *location)
{
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (!session || !session->endpoint || !location) {
		ast_log(LOG_DEBUG, "[session:%p] set_ice_for_session can't get aor for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (found_contact) {
		session->support_ice = found_contact->support_ice;
		session->support_trickle_ice = found_contact->support_trickle_ice;
	} else {
		contact = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(session->endpoint), location);
		if (contact) {
			session->support_ice = contact->support_ice;
			session->support_trickle_ice = contact->support_trickle_ice;
		}
	}

	ast_log(LOG_DEBUG, "[session:%p] set_ice_for_session support_ice:%d, support_trickle_ice:%d for endpoint:%s\n",
		session, session->support_ice, session->support_trickle_ice, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_support_gs_fec_for_session(struct ast_sip_session *session, 
	struct ast_sip_contact *found_contact, const char *location)
{
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (!session || !session->endpoint) {
		ast_log(LOG_DEBUG, "[session:%p] set_support_gs_fec_for_session can't get aor for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (found_contact) {
		session->support_gs_fec = found_contact->support_gs_fec;
	} else {
		if (location) {
			contact = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(session->endpoint), location);
			if (contact) {
				session->support_gs_fec = contact->support_gs_fec;
			}
		}
	}

	ast_log(LOG_DEBUG, "[session:%p] set_support_gs_fec_for_session support_gs_fec:%d, for endpoint:%s\n",
		session, session->support_gs_fec, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_terminal_type_for_app_wakeup(struct ast_sip_session *session, const char *location)
{

	if (!session || !location) {
		return;
	}

	//edit by zyxia for Bug 277446,2023.11.17
	if (!strcasestr(location, "TLS") && !strcasestr(location, "TCP")) {
		return;
	}

	struct sip_waveapp_info *appinfo = ast_sip_find_waveapp_info_by_uri(session->endpoint,location);
	if(appinfo) {
		session->terminal_type = AST_TERMINAL_TYPE_GSWAVE_MOBILE;
		session->support_ice = 1;
	}

	return;
}

static int is_wave_terminal_type_according_contact(struct ast_sip_contact *contact)
{
	char user_agent[MAX_USER_AGENT_LEN] = { 0 };

	if (!contact || ast_strlen_zero(contact->user_agent)) {

		return 0;
	}

	strncpy(user_agent, contact->user_agent, sizeof(user_agent) - 1);

	if (strcasestr(user_agent, "Grandstream Wave Web") || strcasestr(user_agent, "Grandstream Wave Desktop")) {
		return 1;
	}

	return 0;
}

static void set_terminal_for_session(struct ast_sip_session* session,
    struct ast_sip_contact* found_contact, const char* location)
{
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (!session || !session->endpoint || !location) {
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_for_session can't get aor for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (found_contact && found_contact->gswave_mobile) {
		session->terminal_type = AST_TERMINAL_TYPE_GSWAVE_MOBILE;
	} else {
		contact = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(session->endpoint), location);
		if (contact && contact->gswave_mobile) {
			session->terminal_type = AST_TERMINAL_TYPE_GSWAVE_MOBILE;
		} else if (!contact) {
			set_terminal_type_for_app_wakeup(session, location);
		} else if (is_wave_terminal_type_according_contact(contact)) {
			ast_log(LOG_DEBUG, "[session:%p] Set WEBRTC terminal type for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->terminal_type = AST_TERMINAL_TYPE_WEBRTC;
			session->is_webrtc_call = 1;
		}
	}

	ast_log(LOG_DEBUG, "[session:%p] set_terminal_for_session terminal_type:%d, for endpoint:%s\n",
		session, session->terminal_type, ast_sorcery_object_get_id(session->endpoint));
}

static unsigned int ignore_video_media_for_page(struct ast_sip_session* session,
	struct ast_stream *stream, struct ast_channel *requestor)
{
	enum ast_page_mode page_mode = AST_PAGE_MODE_NONE;

	if (!session || !requestor || !stream) {
		return 0;
	}

	if (ast_stream_get_type(stream) == AST_MEDIA_TYPE_AUDIO) {
		return 0;
	}

	page_mode = ast_channel_page_mode_get(requestor);
	if (page_mode == AST_PAGE_MODE_NONE) {
		return 0;
	}

	/*Start added by zyxia for video page,2022.8.5*/
	if ( ast_channel_get_is_page_video_support(requestor)) {
		ast_log(LOG_DEBUG, "[session:%p] It's support video page, dont ignore video media for page, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}
	/*End added by zyxia for video page,2022.8.5*/

	ast_log(LOG_DEBUG, "[session:%p] It's not audio media, ignore this media for page, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	return 1;
}

static void set_if_type_for_session(struct ast_sip_session *session, struct ast_sip_contact *contact, char *uri)
{
    struct ast_sockaddr addr;
    char   tmp_str[128] = {0};

	if (!session) {
		return;
	}

	session->if_type = IF_TYPE_DEFAULT;
	if (!g_vlan_wan_enable) {
		return;
	}

	if (session->ssh_proxy_flag) {
		session->if_type = IF_TYPE_VLAN;
		return;
	}

	if (contact && (!ast_strlen_zero(contact->original_address))) {
        strncpy(tmp_str, contact->original_address, sizeof(tmp_str)-1);
	} 

	pjsip_inv_session *inv_session = session->inv_session;
	if (!strlen(tmp_str) && inv_session) {
		if (inv_session->dlg && inv_session->dlg->remote.contact && inv_session->dlg->remote.contact->uri) {
			pjsip_sip_uri *contact_uri = NULL;
			contact_uri = pjsip_uri_get_uri(inv_session->dlg->remote.contact->uri);
			if(contact_uri) {
				strncpy(tmp_str, contact_uri->host.ptr, contact_uri->host.slen);
			}
		}
	}

	if (!strlen(tmp_str) && uri) {
		//get host part
		char *start = uri;
		char *end = uri + strlen(uri);
		char *pos1 = strchr(uri, '@');

		if (pos1) {
			start = pos1+1;
		}

		char *pos2 = strchr(start, ';');
		if (pos2) {
			end  = pos2;
		}

		if (!strncmp(start, "sip:", 4)) {
			start = start+4;
		} else if (!strncmp(start, "sips:", 5)) {
			start = start+5;
		}

		strncpy(tmp_str, start, (end-start));
	}

	if (strlen(tmp_str)) {
		if (tmp_str[0] == '[') {
			return;  //ignore ipv6
		}

		ast_log(LOG_NOTICE, "remote addr: %s", tmp_str);
		ast_sockaddr_parse(&addr, tmp_str, PARSE_ADDR);
		if (AST_SENSE_ALLOW == ast_apply_acl_with_name("lan", &addr)) {
			session->if_type = IF_TYPE_LAN;
		} else {
			session->if_type = IF_TYPE_VLAN;
        }
	}

	return;
}

int ast_sip_get_vlan_flag(void)
{
	return g_vlan_wan_enable;
}

const char *ast_sip_get_vlan_addr(void)
{
	if (g_vlan_wan_enable) {
		return sip_vlan_addr;
	}

	return NULL;
}

const char *ast_sip_get_lan_addr(void)
{
	if (g_vlan_wan_enable) {
		return sip_lan_addr;
	}

	return NULL;
}
#endif

/*!
 * \internal
 * \brief Setup outbound initial INVITE authentication.
 * \since 13.5.0
 *
 * \param dlg PJSIP dialog to attach outbound authentication.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int setup_outbound_invite_auth(pjsip_dialog *dlg)
{
	pj_status_t status;

	++dlg->sess_count;
	status = pjsip_dlg_add_usage(dlg, &outbound_invite_auth_module, NULL);
	--dlg->sess_count;

	return status != PJ_SUCCESS ? -1 : 0;
}

#ifdef GRANDSTREAM_NETWORKS
struct ast_sip_session *ast_sip_session_create_outgoing(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, const char *location, const char *request_user,
	struct ast_stream_topology *req_topology, struct ast_channel *requestor,pj_str_t *local, pj_str_t *remote, pj_str_t *callid)
#else
struct ast_sip_session *ast_sip_session_create_outgoing(struct ast_sip_endpoint *endpoint,
	struct ast_sip_contact *contact, const char *location, const char *request_user,
	struct ast_stream_topology *req_topology)
#endif
{
	const char *uri = NULL;
	RAII_VAR(struct ast_sip_aor *, found_aor, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_contact *, found_contact, NULL, ao2_cleanup);
	pjsip_timer_setting timer;
	pjsip_dialog *dlg;
	struct pjsip_inv_session *inv_session;
	RAII_VAR(struct ast_sip_session *, session, NULL, ao2_cleanup);
	struct ast_sip_session *ret_session;
#ifdef GRANDSTREAM_NETWORKS
	char record_uri[PJSIP_MAX_URL_SIZE] = {0};
	int ssh_proxy_flag = 0;
	RAII_VAR(struct ast_sip_contact *, contact_tmp, NULL, ao2_cleanup);
#ifdef BUILD_ON_CLOUD_PBX
	char sbc_dest[80] = {0};
#endif
#endif

	/* If no location has been provided use the AOR list from the endpoint itself */
	if (location || !contact) {
		location = S_OR(location, endpoint->aors);

		ast_sip_location_retrieve_contact_and_aor_from_list_filtered(location, AST_SIP_CONTACT_FILTER_REACHABLE,
			&found_aor, &found_contact);
		if (!found_contact || ast_strlen_zero(found_contact->uri)) {
			uri = location;
		} else {
			uri = found_contact->uri;

#ifdef BUILD_ON_CLOUD_PBX
			if (!ast_strlen_zero(found_contact->sbc_endpoint_addr)) {
				strncpy(sbc_dest, found_contact->sbc_endpoint_addr, sizeof(sbc_dest));
			}
#endif
		}
	} else {
		uri = contact->uri;

#ifdef BUILD_ON_CLOUD_PBX
		if (!ast_strlen_zero(contact->sbc_endpoint_addr)) {
			strncpy(sbc_dest, contact->sbc_endpoint_addr, sizeof(sbc_dest));
		}
#endif
	}

	/* If we still have no URI to dial fail to create the session */
	if (ast_strlen_zero(uri)) {
		ast_log(LOG_ERROR, "Endpoint '%s': No URI available.  Is endpoint registered?\n",
			ast_sorcery_object_get_id(endpoint));
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (contact) {
		ssh_proxy_flag = contact->ssh_proxy_flag;
	} else if (found_contact) {
		ssh_proxy_flag = found_contact->ssh_proxy_flag;
		if (!ssh_proxy_flag && !strncmp(ast_sorcery_object_get_id(endpoint), "trunk_", 6)) {
			char host[256] = {0};
			
			ast_sip_get_nvram_value(":WsProxyHost", host, sizeof(host));
			if (strlen(host) && !ast_strlen_zero(endpoint->fromdomain)) {
				if (!strncmp(host, endpoint->fromdomain, sizeof(host))) {
					ssh_proxy_flag = 1;
				}
			}
		}
	} else {
		/* try to get contact object by location */
		contact_tmp = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(endpoint), location);
		if (contact_tmp) {
			ssh_proxy_flag = contact_tmp->ssh_proxy_flag;
		}
	}

	/* Add by bfdu for webrtc to record orig uri, 2017.07.12 */
	strncpy(record_uri, uri, PJSIP_MAX_URL_SIZE - 1);

	if (!(dlg = ast_sip_create_dialog_uac(endpoint, uri, request_user, PJSIP_INVITE_METHOD,local,remote,callid))) {
		return NULL;
	}
#else
	if (!(dlg = ast_sip_create_dialog_uac(endpoint, uri, request_user))) {
		return NULL;
	}
#endif

	if (setup_outbound_invite_auth(dlg)) {
		pjsip_dlg_terminate(dlg);
		return NULL;
	}

	if (pjsip_inv_create_uac(dlg, NULL, endpoint->extensions.flags, &inv_session) != PJ_SUCCESS) {
		pjsip_dlg_terminate(dlg);
		return NULL;
	}
#if defined(HAVE_PJSIP_REPLACE_MEDIA_STREAM) || defined(PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE)
	inv_session->sdp_neg_flags = PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE;
#endif

	pjsip_timer_setting_default(&timer);
	timer.min_se = endpoint->extensions.timer.min_se;
	timer.sess_expires = endpoint->extensions.timer.sess_expires;
	pjsip_timer_init_session(inv_session, &timer);

	session = ast_sip_session_alloc(endpoint, found_contact ? found_contact : contact,
		inv_session, NULL);
	if (!session) {
		pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	char call_id[256] = { 0 };
	ast_copy_pj_str(call_id, &dlg->call_id->id, sizeof(call_id));
	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_create_outgoing endpoint:%s, dialog:%p, call-id:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint), dlg, call_id);
	session->is_outgoing = 1;
	set_webrtc_call_flag(session, record_uri, found_contact);
	
	/* Added by yfyang, for sip ssh proxy feature */
	session->ssh_proxy_flag = ssh_proxy_flag;
	if (ssh_proxy_flag) {
		ast_log(LOG_NOTICE, "Set ssh_proxy_flag to dlg[%p]", dlg);
		pjsip_dlg_set_proxy_flag(dlg, 1);
	}
	set_ice_for_session(session, found_contact ? found_contact : contact, location);
	set_terminal_for_session(session, found_contact ? found_contact : contact, location);
	set_ice_trunk_for_session(session);
	set_candidate_host_only_flag_for_callee(session, found_contact ? found_contact : contact, location);
	set_media_turn_relay(session);
	set_turn_target_address(session, found_contact ? found_contact : contact, location);
	set_support_gs_fec_for_session(session, found_contact ? found_contact : contact, location);
	set_if_type_for_session(session, found_contact ? found_contact : contact, (char *)uri);

	if(ast_media_role_get() == MEDIA_ROLE_MASTER && ast_network_mode_get() != SWITCH_MODE 
		&& !ast_strlen_zero(ast_avs_get_avs_address()) && strcmp("127.0.0.1",ast_avs_get_avs_address())) {
		set_session_diff_multiavs_ip_flag_for_callee(session, found_contact ? found_contact : contact,location,(char*)uri);
	}
	ast_log(LOG_NOTICE, "[session:%p] outgoing, g_vlan_wan_enable[%d], if_type[%d],uri:%s", session, g_vlan_wan_enable, session->if_type,(char *)uri);

#ifdef BUILD_ON_CLOUD_PBX
	if (strlen(sbc_dest)) {
		snprintf(session->sbc_endpoint_addr, sizeof(session->sbc_endpoint_addr), "%s", sbc_dest);
	}

	/* Added by yfyang, for Bug292540 */
	snprintf(session->sbc_addr, sizeof(session->sbc_addr), "%s;lr", uri);
#endif
#endif
	session->aor = ao2_bump(found_aor);
	ast_party_id_copy(&session->id, &endpoint->id.self);

	if (ast_stream_topology_get_count(req_topology) > 0) {
		/* get joint caps between req_topology and endpoint topology */
		int i;

		for (i = 0; i < ast_stream_topology_get_count(req_topology); ++i) {
			struct ast_stream *req_stream;
			struct ast_format_cap *req_cap;
			struct ast_format_cap *joint_cap;
			struct ast_stream *clone_stream;

			req_stream = ast_stream_topology_get_stream(req_topology, i);

			if (ast_stream_get_state(req_stream) == AST_STREAM_STATE_REMOVED) {
				continue;
			}

#ifdef GRANDSTREAM_NETWORKS
			if (ast_sip_session_translate_gswave_to_sip_behavior(session, req_stream, requestor)) {
				continue;
			}

			if (!ast_sip_session_process_special_stream(session, req_stream)) {
				continue;
			}

			if (!ast_sip_session_jugde_if_media_support(session, ast_stream_get_sub_type(req_stream))) {
				ast_log(LOG_DEBUG, "[session:%p] Can't support BFCP/slides video/FECC for A series, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				continue;
			}

			if (ignore_video_media_for_page(session, req_stream, requestor)) {
				continue;
			}
#endif

			req_cap = ast_stream_get_formats(req_stream);

			joint_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
			if (!joint_cap) {
				continue;
			}
#ifdef GRANDSTREAM_NETWORKS
			if (ast_sip_session_get_joint_caps(session, req_stream, joint_cap, req_cap)) {
				ao2_ref(joint_cap, -1);
				continue;
			}

			if (ast_sip_session_process_special_video_codec(session, joint_cap)) {
				ao2_ref(joint_cap, -1);
				continue;
			}
#else
			ast_format_cap_get_compatible(req_cap, endpoint->media.codecs, joint_cap);

			if (!ast_format_cap_count(joint_cap)) {
				ao2_ref(joint_cap, -1);
				continue;
			}
#endif

			clone_stream = ast_stream_clone(req_stream, NULL);
			if (!clone_stream) {
				ao2_ref(joint_cap, -1);
				continue;
			}

			if (ast_stream_get_type(req_stream) == AST_MEDIA_TYPE_AUDIO) {
				/*
				 * By appending codecs from the endpoint after compatible ones this
				 * guarantees that priority is given to those while also allowing
				 * translation to occur for non-compatible.
				 */
				ast_format_cap_append_from_cap(joint_cap,
					endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
#ifdef GRANDSTREAM_NETWORKS
				ast_sip_session_process_special_audio_codec(session, req_stream, joint_cap);
#endif
			}

#ifdef GRANDSTREAM_NETWORKS
			if (ast_stream_get_type(req_stream) == AST_MEDIA_TYPE_VIDEO) {
				ast_sip_session_change_video_independent_codec(session, joint_cap, clone_stream);
			}

			{
				struct ast_str *format_buf = ast_str_alloca(128);
				ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s outgoing peer: %s\n",
					session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(req_cap, &format_buf));
				ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s outgoing caps: %s\n",
					session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(endpoint->media.codecs, &format_buf));
				ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s outgoing joint_cap: %s\n",
					session, ast_sorcery_object_get_id(session->endpoint), ast_format_cap_get_names(joint_cap, &format_buf));
			}
#endif

			ast_stream_set_formats(clone_stream, joint_cap);
			ao2_ref(joint_cap, -1);

			if (!session->pending_media_state->topology) {
				session->pending_media_state->topology = ast_stream_topology_alloc();
				if (!session->pending_media_state->topology) {
					pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
					ao2_ref(session, -1);
					return NULL;
				}
			}

			if (ast_stream_topology_append_stream(session->pending_media_state->topology, clone_stream) < 0) {
				ast_stream_free(clone_stream);
				continue;
			}
		}
	}

#ifndef GRANDSTREAM_NETWORKS
	if (!session->pending_media_state->topology) {
		/* Use the configured topology on the endpoint as the pending one */
		session->pending_media_state->topology = ast_stream_topology_clone(endpoint->media.topology);
		if (!session->pending_media_state->topology) {
			pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
			ao2_ref(session, -1);
			return NULL;
		}
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	add_default_audio_stream(session, session->pending_media_state->topology, endpoint->media.topology);
	ast_sip_session_add_missed_line_for_gswave(session, requestor);
#endif

	if (pjsip_dlg_add_usage(dlg, &session_module, NULL) != PJ_SUCCESS) {
		pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		/* Since we are not notifying ourselves that the INVITE session is being terminated
		 * we need to manually drop its reference to session
		 */
		ao2_ref(session, -1);
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for sca feature */
	if (endpoint->sca_enable) {
		if (!session->aor) {
			pj_str_t tmp, tmp1;
			pjsip_uri *v_uri;
			pjsip_sip_uri *sip_uri;
			char record_uri[PJSIP_MAX_URL_SIZE] = {0};

			strncpy(record_uri, uri, PJSIP_MAX_URL_SIZE - 1);
			pj_cstr(&tmp, record_uri);
			if ((v_uri = pjsip_parse_uri(dlg->pool, tmp.ptr, tmp.slen, 0))) {
				sip_uri = pjsip_uri_get_uri(v_uri);
				if (sip_uri->user.slen > 0) {
					pj_strdup_with_null(dlg->pool, &tmp1, &sip_uri->user);
					session->aor = ast_sip_location_retrieve_aor(tmp1.ptr);
					session->line_index = ast_sip_get_sca_line_index((char *)endpoint->sca_sharedline, SCA_Alerting, tmp1.ptr);

					ao2_ref(session->endpoint, -1); 
					session->endpoint = ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "endpoint", (tmp1.ptr));
				}
			}
		}

		if (session->line_index == 0) {
			session->line_index = ast_sip_get_sca_line_index((char *)endpoint->sca_sharedline, SCA_Alerting, NULL);
		}
	}
#endif

	/* Avoid unnecessary ref manipulation to return a session */
	ret_session = session;
	session = NULL;
	return ret_session;
}

#ifdef GRANDSTREAM_NETWORKS
/* Added by yfyang, for Bug 137192 */
static int ast_sip_session_set_credentials(pjsip_dialog *dlg, struct ast_sip_endpoint *endpoint)
{
	const char *endpoint_id = ast_sorcery_object_get_id(endpoint);
	RAII_VAR(struct ast_sip_auth *, auth, ast_sorcery_retrieve_by_id(ast_sip_get_sorcery(), "auth", endpoint_id), ao2_cleanup);
	pjsip_cred_info  cred;

	if (NULL == auth) {
		return 0;
	}

	pj_bzero(&cred, sizeof(cred));
	cred.realm = pj_str("*");
	cred.scheme = pj_str("digest");
	cred.username = pj_str((char *)auth->auth_user);
	cred.data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
	cred.data = pj_str((char *)auth->auth_pass);

	return pjsip_auth_clt_set_credentials(&dlg->auth_sess, 1, &cred);
}
#endif

static int session_end(void *vsession);
static int session_end_completion(void *vsession);

#ifdef GRANDSTREAM_NETWORKS
static void* faxsending_session_end(void *vsession); //added by zyxia for Bug185135,2021.1.22
/*Start added by zyxia for wakeup gswave,2018.9.19*/
#if 0
static void ast_sip_terminated_set_gswave_call_count(struct ast_sip_session *session,pjsip_tx_data *packet)
{
	char gs_appid_key[64];
	char gs_appid[256];

	/*get gs_appid from astdb*/
	snprintf(gs_appid_key,sizeof(gs_appid_key),"%s/%s","gs_appid",session->endpoint->aors);
	if((!ast_strlen_zero(session->endpoint->gs_appid)
		|| (!ast_db_get(GSWAVE_DB_FAMILY, gs_appid_key, gs_appid, sizeof(gs_appid)) && !ast_strlen_zero(gs_appid)))
		&& packet) {
		if (packet->msg->type == PJSIP_RESPONSE_MSG) {/* bug for 121062*/
			pjsip_via_hdr *via;
			via = (pjsip_via_hdr*) pjsip_msg_find_hdr(packet->msg, PJSIP_H_VIA, NULL);
			if (via && pj_strlen(&via->recvd_param) && strstr(pj_strbuf(&via->recvd_param) ,session->endpoint->gs_uri) && session->endpoint->gswave_call_count) {
				session->endpoint->gswave_call_count--;
				ast_log(LOG_WARNING, "Send Reponse 480 or 603 and other to Gswave [%s], gswave_call_count :%d,ip:%s\n",session->endpoint->aors,session->endpoint->gswave_call_count,pj_strbuf(&via->recvd_param));
			}
		} else {
			/*Get ip from Bye or Cancel requset "Requste line" header,check if the bye is sent to ios gswave ,for bug 120770,121085*/
			pjsip_uri *ruri = packet->msg->line.req.uri;
			if (PJSIP_URI_SCHEME_IS_SIP(ruri) || PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
				pjsip_sip_uri *sip_ruri;
				sip_ruri = pjsip_uri_get_uri(ruri);
				if (sip_ruri && strstr(pj_strbuf(&sip_ruri->host),session->endpoint->gs_uri) && session->endpoint->gswave_call_count) {
					session->endpoint->gswave_call_count--;
					ast_log(LOG_WARNING, "Send Request Cancel or Bye to Gswave [%s], gswave_call_count :%d,ip:%s\n",session->endpoint->aors,session->endpoint->gswave_call_count,pj_strbuf(&sip_ruri->host));
				}
			}
		}
	}
}
#endif
#endif
/*end added by zyxia for wakeup gswave,2018.9.19*/

void ast_sip_session_terminate(struct ast_sip_session *session, int response)
{
	pj_status_t status;
	pjsip_tx_data *packet = NULL;

#ifdef GRANDSTREAM_NETWORKS
	if (session == NULL || session->inv_session == NULL || session->inv_session->dlg == NULL) {
		ast_log(LOG_ERROR, "SIP session is invalid or PJSIP invite session/dialog is NULL!");
		return;
	}

	/* Added by yfyang, for Bug 137192, Bug 155688 */
	if ((PJSIP_ROLE_UAS == session->inv_session->role)
		&& (PJSIP_INV_STATE_DISCONNECTED != session->inv_session->state)
		&& session->channel && !strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
		ast_sip_session_set_credentials(session->inv_session->dlg, session->endpoint);
	}
#endif

	if (session->defer_terminate) {
		session->terminate_while_deferred = 1;
		return;
	}

	if (!response) {
		response = 603;
	}

	/* The media sessions need to exist for the lifetime of the underlying channel
	 * to ensure that anything (such as bridge_native_rtp) has access to them as
	 * appropriate. Since ast_sip_session_terminate is called by chan_pjsip and other
	 * places when the session is to be terminated we terminate any existing
	 * media sessions here.
	 */
#ifdef GRANDSTREAM_NETWORKS
	pjsip_dlg_inc_lock(session->inv_session->dlg);
#endif
	SWAP(session->active_media_state, session->pending_media_state);
	ast_sip_session_media_state_reset(session->pending_media_state);
#ifdef GRANDSTREAM_NETWORKS
	pjsip_dlg_dec_lock(session->inv_session->dlg);
#endif

	switch (session->inv_session->state) {
	case PJSIP_INV_STATE_NULL:
		if (!session->inv_session->invite_tsx) {
			/*
			 * Normally, it's pjproject's transaction cleanup that ultimately causes the
			 * final session reference to be released but if both STATE and invite_tsx are NULL,
			 * we never created a transaction in the first place.  In this case, we need to
			 * do the cleanup ourselves.
			 */
			/* Transfer the inv_session session reference to the session_end_task */
			session->inv_session->mod_data[session_module.id] = NULL;
			pjsip_inv_terminate(session->inv_session, response, PJ_TRUE);
			session_end(session);
			/*
			 * session_end_completion will cleanup the final session reference unless
			 * ast_sip_session_terminate's caller is holding one.
			 */
			session_end_completion(session);
		} else {
			pjsip_inv_terminate(session->inv_session, response, PJ_TRUE);
		}
		break;
	case PJSIP_INV_STATE_CONFIRMED:
#ifndef GRANDSTREAM_NETWORKS
		/* Remove by bfdu for bug 184080, 2022.11.28 */
		if (session->inv_session->invite_tsx) {
			ast_debug(3, "Delay sending BYE to %s because of outstanding transaction...\n",
					ast_sorcery_object_get_id(session->endpoint));
			/* If this is delayed the only thing that will happen is a BYE request so we don't
			 * actually need to store the response code for when it happens.
			 */
#ifdef GRANDSTREAM_NETWORKS
			/* Add by bfdu for copy conf_hangupcause because session->channel will set NULL later.
			 * We need get conference hangup reason when delay send BYE.
			 */
			if (session->channel) {
				session->conf_hangupcause = ast_channel_conf_hangupcause(session->channel);
			}
#endif
			delay_request(session, NULL, NULL, NULL, 0, DELAYED_METHOD_BYE, NULL);
			break;
		}
#endif

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 273977, 2023.10.17 */
		if (session->ref_added) {
			ast_log(LOG_WARNING, "[session:%p] ast_sip_session_terminate session ref added before, need decrease, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->ref_added = 0;
			ao2_ref(session, -1);
		}
#endif

		/* Fall through */
	default:
		status = pjsip_inv_end_session(session->inv_session, response, NULL, &packet);
		if (status == PJ_SUCCESS && packet) {
			struct ast_sip_session_delayed_request *delay;

			/* Flush any delayed requests so they cannot overlap this transaction. */
			while ((delay = AST_LIST_REMOVE_HEAD(&session->delayed_requests, next))) {
				delayed_request_free(delay);
			}

#ifdef GRANDSTREAM_NETWORKS
			if (session->channel) {
				session->conf_hangupcause = ast_channel_conf_hangupcause(session->channel);
			}
#endif

			if (packet->msg->type == PJSIP_RESPONSE_MSG) {
				ast_sip_session_send_response(session, packet);
			} else {
				ast_sip_session_send_request(session, packet);
			}

		}
		break;
	}
}

static int session_termination_task(void *data)
{
	struct ast_sip_session *session = data;

	if (session->defer_terminate) {
		session->defer_terminate = 0;
		if (session->inv_session) {
			ast_sip_session_terminate(session, 0);
		}
	}

	ao2_ref(session, -1);
	return 0;
}

static void session_termination_cb(pj_timer_heap_t *timer_heap, struct pj_timer_entry *entry)
{
	struct ast_sip_session *session = entry->user_data;

	if (ast_sip_push_task(session->serializer, session_termination_task, session)) {
		ao2_cleanup(session);
	}
}

int ast_sip_session_defer_termination(struct ast_sip_session *session)
{
	pj_time_val delay = { .sec = 60, };
	int res;

	/* The session should not have an active deferred termination request. */
	ast_assert(!session->defer_terminate);

	session->defer_terminate = 1;

	session->defer_end = 1;
	session->ended_while_deferred = 0;

#ifdef GRANDSTREAM_NETWORKS
	if (pj_timer_entry_running(&session->scheduled_termination)) {
		/* Stop the termination timer if it is still running. */
		sip_session_defer_termination_stop_timer(session);
	}
#endif

	ao2_ref(session, +1);
	pj_timer_entry_init(&session->scheduled_termination, 0, session, session_termination_cb);

	res = (pjsip_endpt_schedule_timer(ast_sip_get_pjsip_endpoint(),
		&session->scheduled_termination, &delay) != PJ_SUCCESS) ? -1 : 0;
	if (res) {
		session->defer_terminate = 0;
		ao2_ref(session, -1);
	}
	return res;
}

/*!
 * \internal
 * \brief Stop the defer termination timer if it is still running.
 * \since 13.5.0
 *
 * \param session Which session to stop the timer.
 *
 * \return Nothing
 */
static void sip_session_defer_termination_stop_timer(struct ast_sip_session *session)
{
	if (pj_timer_heap_cancel_if_active(pjsip_endpt_get_timer_heap(ast_sip_get_pjsip_endpoint()),
		&session->scheduled_termination, session->scheduled_termination.id)) {
		ao2_ref(session, -1);
	}
}

void ast_sip_session_defer_termination_cancel(struct ast_sip_session *session)
{
	if (!session->defer_terminate) {
		/* Already canceled or timer fired. */
		return;
	}

	session->defer_terminate = 0;

	if (session->terminate_while_deferred) {
		/* Complete the termination started by the upper layer. */
		ast_sip_session_terminate(session, 0);
	}

	/* Stop the termination timer if it is still running. */
	sip_session_defer_termination_stop_timer(session);
}

void ast_sip_session_end_if_deferred(struct ast_sip_session *session)
{
	if (!session->defer_end) {
		return;
	}

	session->defer_end = 0;

	if (session->ended_while_deferred) {
		/* Complete the session end started by the remote hangup. */
		ast_debug(3, "Ending session (%p) after being deferred\n", session);
		session->ended_while_deferred = 0;
		session_end(session);
	}
}

struct ast_sip_session *ast_sip_dialog_get_session(pjsip_dialog *dlg)
{
	pjsip_inv_session *inv_session = pjsip_dlg_get_inv_session(dlg);
	struct ast_sip_session *session;

	if (!inv_session ||
		!(session = inv_session->mod_data[session_module.id])) {
		return NULL;
	}

	ao2_ref(session, +1);

	return session;
}

enum sip_get_destination_result {
	/*! The extension was successfully found */
	SIP_GET_DEST_EXTEN_FOUND,
	/*! The extension specified in the RURI was not found */
	SIP_GET_DEST_EXTEN_NOT_FOUND,
	/*! The extension specified in the RURI was a partial match */
	SIP_GET_DEST_EXTEN_PARTIAL,
	/*! The RURI is of an unsupported scheme */
	SIP_GET_DEST_UNSUPPORTED_URI,
};

#ifdef GRANDSTREAM_NETWORKS
static int get_endpoint_details(pjsip_rx_data *rdata, char *endpoint, size_t endpoint_size, char *domain, size_t domain_size)
{
	pjsip_uri *from = rdata->msg_info.from->uri;
	pjsip_sip_uri *sip_from;

	if (PJSIP_URI_SCHEME_IS_SIP(from) || PJSIP_URI_SCHEME_IS_SIPS(from)) {
		sip_from = (pjsip_sip_uri *) pjsip_uri_get_uri(from);
		ast_copy_pj_str(endpoint, &sip_from->user, endpoint_size);
		ast_copy_pj_str(domain, &sip_from->host, domain_size);

		return 0;
	}

	if (PJSIP_URI_SCHEME_IS_TEL(from)) {
		pjsip_tel_uri *tel_uri = pjsip_uri_get_uri(from);
		ast_copy_pj_str(endpoint, &tel_uri->number, endpoint_size);

		return 0;
	}

	return -1;
}

int ast_sip_session_is_nway_conference_number(const char *exten)
{
	const char *family = "NWAY_CONF_STATUS";
	char key[64] = { 0 };
	char sub_value[64] = {0};

	if (!exten) {
		return 0;
	}

	snprintf(key, sizeof(key), "%s/IS_N_WAY_CONF", exten);

	ast_db_get(family, key, sub_value, sizeof(sub_value) - 1);
	if (ast_strlen_zero(sub_value)) {
		return 0;
	}

	ast_log(LOG_DEBUG, "%s is a nway conference number\n", exten);

	return 1;
}

int ast_sip_get_diversion_user(pjsip_rx_data *rdata, char *user, int len)
{
	pj_str_t from_name = { "From", 4 };
	pj_str_t div_name  = { "Diversion", 9 };
	pjsip_generic_string_hdr *hdr = NULL;
	pjsip_fromto_hdr *div_hdr = NULL;
	pjsip_sip_uri    *uri = NULL;
	pj_str_t          value;
	pj_str_t          uri_user;
	int   size;
	char *semi;

	if (!rdata || !user || (len == 0)) {
		return -1;
	}

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &div_name, NULL))) {
		return -1;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);

	/* parse as a fromto header */
	div_hdr = (pjsip_fromto_hdr *)pjsip_parse_hdr(rdata->tp_info.pool, &from_name, value.ptr, pj_strlen(&value), &size);
	if (!div_hdr) {
		return -1;
	}

	uri = pjsip_uri_get_uri(div_hdr->uri);
	if (!uri) {
		return -1;
	}

	if (!pj_strlen(&uri->user)) {
		return -1;
	}

	uri_user = uri->user;
	semi = pj_strchr(&uri_user, ';');
	if (semi) {
		pj_strset(&uri_user, (char *) pj_strbuf(&uri_user), semi - pj_strbuf(&uri_user));
	}

	if (pj_strlen(&uri_user)) {
		snprintf(user, len-1, "%.*s", (int)pj_strlen(&uri_user), pj_strbuf(&uri_user));
		ast_log(LOG_NOTICE, "Get Diversion user:%s", user);
		return 0;
	}

	return -1;
}

#endif

/*!
 * \brief Determine where in the dialplan a call should go
 *
 * This uses the username in the request URI to try to match
 * an extension in the endpoint's configured context in order
 * to route the call.
 *
 * \param session The inbound SIP session
 * \param rdata The SIP INVITE
 */
static enum sip_get_destination_result get_destination(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_uri *ruri = rdata->msg_info.msg->line.req.uri;
	pjsip_sip_uri *sip_ruri;
	struct ast_features_pickup_config *pickup_cfg;
	const char *pickupexten;
#ifdef GRANDSTREAM_NETWORKS
	char endpoint_name[64]={0}, domain_name[64]={0};
	char diversion_user[80] = {0};
	pjsip_tel_uri  *tel_ruri;
	struct ast_sip_cc_original_exten_data exten_data;
#endif

#ifndef GRANDSTREAM_NETWORKS
	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri)) 
#else
	if (!PJSIP_URI_SCHEME_IS_SIP(ruri) && !PJSIP_URI_SCHEME_IS_SIPS(ruri) && !PJSIP_URI_SCHEME_IS_TEL(ruri)) 
#endif
	{
		return SIP_GET_DEST_UNSUPPORTED_URI;
	}

#ifndef GRANDSTREAM_NETWORKS
	sip_ruri = pjsip_uri_get_uri(ruri);
	ast_copy_pj_str(session->exten, &sip_ruri->user, sizeof(session->exten));
#else

	pjsip_uri *to = rdata->msg_info.to->uri;
	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, didmode = %s\n",
		session, ast_sorcery_object_get_id(session->endpoint), session->endpoint->didmode);
	if (strcasecmp(session->endpoint->didmode, "to-header") == 0) {
		sip_ruri = pjsip_uri_get_uri(to);

		if (PJSIP_URI_SCHEME_IS_SIP(ruri) || PJSIP_URI_SCHEME_IS_SIPS(ruri)) {
			pjsip_sip_uri *sip_requri = pjsip_uri_get_uri(ruri);
			if (sip_requri->user.slen) {
				session->contact_user = ast_strndup(sip_requri->user.ptr, sip_requri->user.slen); 
			}
		} else if (PJSIP_URI_SCHEME_IS_TEL(ruri)) {
			pjsip_tel_uri *tel_requri = pjsip_uri_get_uri(ruri);
			if (tel_requri->number.slen) {
				session->contact_user = ast_strndup(tel_requri->number.ptr, tel_requri->number.slen); 
			}
		}
		ast_log(LOG_NOTICE, "contact_user = %s\n", session->contact_user);
	} else {
		sip_ruri = pjsip_uri_get_uri(ruri);
	}

    if (PJSIP_URI_SCHEME_IS_SIP(to) || PJSIP_URI_SCHEME_IS_SIPS(to)) {
		ast_copy_pj_str(session->exten, &sip_ruri->user, sizeof(session->exten));
    } else if (PJSIP_URI_SCHEME_IS_TEL(to)) {
		tel_ruri = (pjsip_tel_uri *)sip_ruri;
		ast_copy_pj_str(session->exten, &tel_ruri->number, sizeof(session->exten));
    }

	/* Added by yfyang, for Bug 124852 */
	char *sep = strchr(session->exten, ';');
	if (sep) {
		*sep = '\0';
	}

	/* Add by bfdu for security attacks */
	if (!strncmp(ast_sorcery_object_get_id(session->endpoint), "conf", 4) &&
		!ast_sip_check_is_mcm_conf_number(session->exten) &&
		!ast_sip_session_is_nway_conference_number(session->exten)) {
		ast_log(LOG_DEBUG, "Endpoint:%s is use for SFU, but destination is not SFU, reject\n", ast_sorcery_object_get_id(session->endpoint));
		return SIP_GET_DEST_EXTEN_NOT_FOUND;
	}
#endif

	/*
	 * We may want to match in the dialplan without any user
	 * options getting in the way.
	 */
	AST_SIP_USER_OPTIONS_TRUNCATE_CHECK(session->exten);

	pickup_cfg = ast_get_chan_features_pickup_config(session->channel);
	if (!pickup_cfg) {
		ast_log(LOG_ERROR, "Unable to retrieve pickup configuration options. Unable to detect call pickup extension\n");
		pickupexten = "";
	} else {
		pickupexten = ast_strdupa(pickup_cfg->pickupexten);
		ao2_ref(pickup_cfg, -1);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (get_endpoint_details(rdata, endpoint_name, sizeof(endpoint_name), domain_name, sizeof(domain_name))) {
		if (!strcmp(session->exten, pickupexten) ||
				 ast_exists_extension(NULL, session->endpoint->context, session->exten, 1, NULL)) {
			return SIP_GET_DEST_EXTEN_FOUND;
		}
	} else {
		/* Added by yfyang, Grandstream networks, for ccss feature */
		if (0 == find_sub_handler_for_supplement_exten(AST_SIP_CCSS_ORIGINAL_EXTEN, session->exten, &exten_data)) {
			strncpy(session->exten, exten_data.original_exten, sizeof(session->exten)-1);
			session->core_id = exten_data.core_id;
			ast_log(LOG_NOTICE, "get_destination, find cc recall, core_id[%d]", session->core_id);
			ast_cc_agent_recalling(exten_data.core_id, "SIP caller %s is attempting recall", exten_data.device_name);
			return SIP_GET_DEST_EXTEN_FOUND;
		}

		if (!strcmp(session->exten, pickupexten) && strcmp(session->endpoint->context, "sla-stations")) {
			return SIP_GET_DEST_EXTEN_FOUND;
		}

		/* get diversion userid */
		if (!ast_sip_get_diversion_user(rdata, diversion_user, sizeof(diversion_user))) {
			if (ast_exists_extension2(NULL, session->endpoint->context, session->exten, 1, diversion_user, 1)) {
				ast_log(LOG_NOTICE, "get_destination, context[%s], exten[%s], divid[%s], found!", session->endpoint->context, session->exten, diversion_user);
				session->need_orig_callee = 1;
				return SIP_GET_DEST_EXTEN_FOUND;
			}
		}

		if (ast_exists_extension(NULL, session->endpoint->context, session->exten, 1, endpoint_name)) {
			/* Added by yfyang, for Bug96013 */
			session->need_orig_callee = 0;
			if (!ast_exists_extension(NULL, "ext-park", session->exten, 1, NULL) &&
			    !ast_exists_extension(NULL, "voicemenus", session->exten, 1, NULL) &&
			    !ast_exists_extension(NULL, "app-pickup", session->exten, 1, NULL) &&
				!ast_sip_park_pos_check(session->exten)) {
				session->need_orig_callee = 1;
			}
			return SIP_GET_DEST_EXTEN_FOUND;
		}
	}
#else
	if (!strcmp(session->exten, pickupexten) ||
		ast_exists_extension(NULL, session->endpoint->context, session->exten, 1, NULL)) {
		size_t size = pj_strlen(&sip_ruri->host) + 1;
		char *domain = ast_alloca(size);

		ast_copy_pj_str(domain, &sip_ruri->host, size);
		pbx_builtin_setvar_helper(session->channel, "SIPDOMAIN", domain);

		/*
		 * Save off the INVITE Request-URI in case it is
		 * needed: CHANNEL(pjsip,request_uri)
		 */
		session->request_uri = pjsip_uri_clone(session->inv_session->pool, ruri);

		return SIP_GET_DEST_EXTEN_FOUND;
	}
#endif
	/*
	 * Check for partial match via overlap dialling (if enabled)
	 */

#ifndef GRANDSTREAM_NETWORKS
	if (session->endpoint->allow_overlap && (
		!strncmp(session->exten, pickupexten, strlen(session->exten)) ||
		ast_canmatch_extension(NULL, session->endpoint->context, session->exten, 1, NULL))) {
		/* Overlap partial match */
		return SIP_GET_DEST_EXTEN_PARTIAL;
	}
#else
	/* Modified by yfyang, for Bug 223703 */
	if (session->endpoint->allow_overlap) {
		if (!strncmp(session->exten, pickupexten, strlen(session->exten))) {
			return SIP_GET_DEST_EXTEN_PARTIAL;
		}

		if ((session->exten[0] != '*') && (ast_canmatch_extension2(NULL, session->endpoint->context, session->exten, 1, NULL, 0))) {
			return SIP_GET_DEST_EXTEN_PARTIAL;
		}
	}

	if (!strcmp(endpoint_name, "FastProvision")) {
		session->is_fastprov = 1;
		ast_log(LOG_NOTICE, "Detect Fastprovision SIP Public extension");
		return SIP_GET_DEST_EXTEN_FOUND;
	}
#endif

	return SIP_GET_DEST_EXTEN_NOT_FOUND;
}

/* Added by yfyang, for sca feature */
#ifdef GRANDSTREAM_NETWORKS

const char *ast_sip_sca_state2str(enum sca_status state)
{
	int i;
	for (i = 0; i < ARRAY_LEN(sca_state2string); i++) {
		if (state == sca_state2string[i].state) {
			return sca_state2string[i].string;
		}
	}
	return "";
}


static enum sca_status ast_sca_state_val(const char *val)
{
	int i;
	for (i = 0; i < ARRAY_LEN(sca_state2string); i++) {
		if (!strcasecmp(val, sca_state2string[i].string)) {
			return sca_state2string[i].state;
		}
	}

	return SCA_BUTT;
}

int ast_sip_get_sca_call_info_from_rdata(pjsip_rx_data *rdata, unsigned int *index, enum sca_status *status)
{
	pjsip_generic_string_hdr   *hdr;
	pj_str_t callinfo_content; 
	char  *info = NULL;
	char  *p_index  = NULL;
	char  *p_status = NULL;
	char  *p_status_end = NULL;

	if (!rdata || !index) {
		return -1;
	}

	/* get PAI header string */
	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &callinfo_str, NULL);
	if (!hdr) {
		return -1;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &callinfo_content, &hdr->hvalue);
	do {
		info = ast_strsep(&callinfo_content.ptr, ',', AST_STRSEP_TRIM);
		if (!info) {
			return -1;
		}

		if (strstr(info, "appearance-index=")) {
			break;
		}
	} while (1);

	p_index = strstr(info, "appearance-index=") + strlen("appearance-index=");
	*index = atoi(p_index);
	if (!status) {
		return 0;
	}

	if (status) {
		p_status = strstr(info, "appearance-state="); 
		if (p_status) {
			p_status = p_status + strlen("appearance-state=");
			p_status_end = strchr(p_status, ';');
			if (p_status_end) {
				*p_status_end = '\0';
			}
			*status = ast_sca_state_val(p_status); 
		}
	}

	return 0;
}

int ast_sip_get_sca_private_users(char *sharedline, char *user_buf, int len, int paging_flag)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline || !user_buf || !len) {
		return -1; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_get_private_users) {
		handler->sca_line_get_private_users(sharedline, user_buf, len, paging_flag);
	}

	return 0;
}

int ast_sip_set_sca_user_info(char *sharedline, int line_index, char *local_user, char *remote_user)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline || (line_index <= 0) || (!local_user && !remote_user)) {
		return -1; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_set_user_info) {
		handler->sca_line_set_user_info(sharedline, line_index, local_user, remote_user);
	}

	return 0;
}

int ast_sip_sca_line_remove_user(struct ast_channel *chan, char *sharedline, int line_index, char *user_name)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline || !user_name || (line_index <= 0)) {
		return -1; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_remove_user) {
		handler->sca_line_remove_user(chan, sharedline, line_index, user_name);
	}

	return 0;
}

int ast_sip_set_sca_line_status(char *sharedline, int line_index, enum sca_status status)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline || (status >= SCA_BUTT) || (line_index <= 0)) {
		return -1; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_status_set) {
		handler->sca_line_status_set(sharedline, line_index, status);
	}

	return 0;
}


enum sca_status  ast_sip_get_sca_line_status(char *sharedline, int line_index)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline || (line_index <= 0)) {
		return SCA_BUTT; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_status_get) {
		return handler->sca_line_status_get(sharedline, line_index, NULL);
	}

	return SCA_BUTT;
}

unsigned int ast_sip_get_sca_line_index(char *sharedline, enum sca_status status, char *local_info)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline) {
		return 0; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_index_get) {
		return handler->sca_line_index_get(sharedline, status, local_info);
	}

	return 0;
}

unsigned int ast_sip_sca_line_alloc(char *sharedline, char *local_info, struct ast_channel *remote_chan)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline) {
		return 0; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_alloc) {
		return handler->sca_line_alloc(sharedline, local_info, remote_chan);
	}

	return 0;
}

/* ret value means:
   -1: something wrong
   0 : idle  
   other value: inuse */
int ast_sip_check_sca_user_status(char *sharedline)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline) {
		return -1; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_user_check_status) {
		return handler->sca_user_check_status(sharedline);
	}

	return -1;
}


int ast_sip_sca_line_insert_user(char *sharedline, int line_index, char *local_user)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline) {
		return -1; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_insert_user) {
		return handler->sca_line_insert_user(sharedline, line_index, local_user);
	}

	return -1;
}

/* To check if the line is in a sca bridge, 
   return 0: not in bridge;  1: in bridge  */
int ast_sip_sca_line_check_in_bridge(char *sharedline, int line_index)
{
	struct ast_sip_session_sca_handler *handler = NULL;

	if (!sharedline) {
		return 0; 
	}

	handler = find_sca_handler(); 
	if (handler && handler->sca_line_check_in_bridge) {
		return handler->sca_line_check_in_bridge(sharedline, line_index);
	}

	return 0;
}

/*! \brief Retrive user from TO Header.  */
/* \note Returned string MUST be freed. */
static char* session_get_to_user_from_rdata(pjsip_rx_data *rdata)
{
	pjsip_to_hdr *to_hdr = NULL;
	size_t size;
	char *touser = NULL;

	// Sanity check
	if (rdata == NULL) {
		return NULL;
	}

	// Get TO header
	to_hdr = PJSIP_MSG_TO_HDR(rdata->msg_info.msg);
	if (to_hdr == NULL || to_hdr->uri == NULL) {
		return NULL;
	}

	// Get TO information
	if (PJSIP_URI_SCHEME_IS_SIP(to_hdr->uri) || PJSIP_URI_SCHEME_IS_SIPS(to_hdr->uri)) {
		pjsip_sip_uri *sip_uri = pjsip_uri_get_uri(to_hdr->uri);

		if (sip_uri && sip_uri->user.slen) {
			size = pj_strlen(&sip_uri->user) + 1;
			touser = ast_malloc(size);
            if (!touser) {
				return NULL;
            }

            ast_copy_pj_str(touser, &sip_uri->user, size); 
		}
	} else if (PJSIP_URI_SCHEME_IS_TEL(to_hdr->uri)) {
		pjsip_tel_uri *tel_uri = pjsip_uri_get_uri(to_hdr->uri);

		if (tel_uri && tel_uri->number.slen) {
			size = pj_strlen(&tel_uri->number) + 1;
			touser = ast_malloc(size);
            if (!touser) {
				return NULL;
            }

			ast_copy_pj_str(touser, &tel_uri->number, size);
		}
	}

	return touser;
}

static int check_sca_call_info(pjsip_rx_data *rdata, const struct ast_sip_endpoint *endpoint)
{
	struct ast_sip_session_sca_handler *handler = NULL;
	enum   sca_status  status = SCA_BUTT;
	unsigned int  index;
	char *user_name  = NULL;

	if (!endpoint || !rdata) {
		return -1;
	}

	/* get Call-Info header string */
	if (ast_sip_get_sca_call_info_from_rdata(rdata, &index, NULL)) {
		if (ast_sip_get_sca_line_index((char *)endpoint->sca_sharedline, SCA_Seized, (char *)ast_sorcery_object_get_id(endpoint))) {
			return 0;
		}
		return -1;
	}

	status = ast_sip_get_sca_line_status((char *)endpoint->sca_sharedline, index);
	if (status == SCA_Seized) {
		if (ast_sip_get_sca_line_index((char *)endpoint->sca_sharedline, SCA_Seized, (char *)ast_sorcery_object_get_id(endpoint))) {
			return 0; 
		}
		return -1;
	}

	if ((status == SCA_Held) || (status == SCA_Active)) {
		int retrv_remote = 0;
		int barge = 0;

		/* Added by yfyang, to check the to user, for Bug 109376 */
		user_name = session_get_to_user_from_rdata(rdata);
		if (!user_name) {
			return -1;
		}

		if (strcmp(user_name, ast_sorcery_object_get_id(endpoint))) {
			ast_log(LOG_NOTICE, "check_sca_call_info, sharedline[%s], line index[%d], status[%d], to user is [%s], can't process", 
						endpoint->sca_sharedline, index, status, user_name);
			ast_free(user_name); 
			return -1;
		}

		ast_free(user_name); 

		if (!(handler = find_sca_handler())) {
			return -1;
		}
         
		handler->sca_line_service_get((char *)endpoint->sca_sharedline, &retrv_remote, &barge);
		if (retrv_remote && (status == SCA_Held)) {
			ast_log(LOG_NOTICE, "check_sca_call_info, sharedline[%s], line index[%d], status is held, allow retrieve remote", 
						endpoint->sca_sharedline, index);
			return 0;
		}

		if (barge && (status == SCA_Active)) {
			ast_log(LOG_NOTICE, "check_sca_call_info, sharedline[%s], line index[%d], status is active, allow barge", 
						endpoint->sca_sharedline, index);
			return 0;
		}
	}

	return -1;
}

/* Added by yfyang, for Bug 130422 */
/* check the header: "P-Asserted-Service-Info: service-code=Backward Collect Call" */
static int get_backward_collect_call(pjsip_rx_data *rdata)
{
	static const pj_str_t hdr_name = { "P-Asserted-Service-Info", 23 };

	pjsip_generic_string_hdr *hdr;
	pj_str_t value;

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &hdr_name, NULL))) {
		return 0;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &value, &hdr->hvalue);

	if (!pj_strnicmp2(&value, "service-code=Backward Collect Call", 34) || !pj_strnicmp2(&value, "service-code=Collect Call", 25)) {
		return 1;
	}

	return 0;
}

static void conference_invite_need_reject(struct ast_sip_session* session, pjsip_rx_data* rdata)
{
	ast_log(LOG_DEBUG, "enter conference_invite_need_reject.");

	if (!session || !rdata) {
		return;
	}

	static const pj_str_t conf_call_type_field = { "X-Grandstream-Call-Type", 23 };
	static const pj_str_t conf_ucm_im_field = { "X-UCM-IM", 8 };
	static const pj_str_t conf_ucm_im_service_id_field = { "X-GS-IM-SERVICE-ID", 18 };

	pjsip_generic_string_hdr* conf_call_type_hdr = NULL;
	pjsip_generic_string_hdr* conf_ucm_im_hdr = NULL;
	pjsip_generic_string_hdr* conf_ucm_im_service_id_hdr = NULL;

	pj_str_t conf_call_type_value;
	pj_str_t conf_ucm_im_value;
	pj_str_t conf_ucm_im_service_id_value;

	char cloud_im_enabled[8] = {0};
	char cloud_im_addr[129] = {0};
	char cloud_im_value[256] = {0};
	char cloud_im_service_id[64] = {0};

	if (!(conf_call_type_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &conf_call_type_field, NULL))) {
		ast_log(LOG_DEBUG, "Not had X-Grandstream-Call-Type header.");
		return;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &conf_call_type_value, &conf_call_type_hdr->hvalue);

	if (pj_strnicmp2(&conf_call_type_value, "Conference", 10) &&
		pj_strnicmp2(&conf_call_type_value, "VideoConference", 15)) {
		ast_log(LOG_DEBUG, "The invite isn't conference invite.");
		return;
	}

	ast_log(LOG_DEBUG, "The invite is conference invite.");

	if (!(conf_ucm_im_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &conf_ucm_im_field, NULL))) {
		ast_log(LOG_DEBUG, "The invite message not carry im header.");
		return;
	}

	pj_strdup_with_null(rdata->tp_info.pool, &conf_ucm_im_value, &conf_ucm_im_hdr->hvalue);

	ast_safe_nvram_get("cloud_im_enabled", cloud_im_enabled, sizeof(cloud_im_enabled));

	conf_ucm_im_service_id_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &conf_ucm_im_service_id_field, NULL);
	if (conf_ucm_im_service_id_hdr) {
		pj_strdup_with_null(rdata->tp_info.pool, &conf_ucm_im_service_id_value, &conf_ucm_im_service_id_hdr->hvalue);
	}

	if (strncmp(cloud_im_enabled, "on", 2) == 0) {
		ast_safe_nvram_get("cloud_im_address", cloud_im_addr, sizeof(cloud_im_addr));
		snprintf(cloud_im_value, sizeof(cloud_im_value) - 1, "cloud;%s", cloud_im_addr);

		ast_safe_nvram_get("cloud_im_service_id", cloud_im_service_id, sizeof(cloud_im_service_id));

		ast_log(LOG_DEBUG, "[session:%p] Local is enabled cloud IM[%s] service_id[%s], endpoint:%s\n",
			session, cloud_im_value, cloud_im_service_id, ast_sorcery_object_get_id(session->endpoint));

		if (!pj_strnicmp2(&conf_ucm_im_value, cloud_im_value, strlen(cloud_im_value))) {
			if (conf_ucm_im_service_id_hdr && strlen(cloud_im_service_id)) {
				if (!pj_strnicmp2(&conf_ucm_im_service_id_value, cloud_im_service_id, strlen(cloud_im_service_id))) {
					session->reject_call = 1;
				} else {
					ast_log(LOG_DEBUG, "[session:%p] Receive IM service id hdr is not the same as cloud IM[%s] service_id[%s], endpoint:%s\n",
						session, cloud_im_value, cloud_im_service_id, ast_sorcery_object_get_id(session->endpoint));
				}
			} else {
				session->reject_call = 1;
			}
		}
	}

	if (session->reject_call) {
		session->reject_call |= (0x01 << 1);  /* set bit 1 to do special mark */
	}

	ast_log(LOG_NOTICE, "[session:%p] %s reject conference invite, endpoint:%s\n",
		session, session->reject_call ? "Do" : "Don't", ast_sorcery_object_get_id(session->endpoint));

	return;
}

#endif


static pjsip_inv_session *pre_session_setup(pjsip_rx_data *rdata, const struct ast_sip_endpoint *endpoint)
{
	pjsip_tx_data *tdata;
	pjsip_dialog *dlg;
	pjsip_inv_session *inv_session;
	unsigned int options = endpoint->extensions.flags;
	pj_status_t dlg_status;

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 130422 */
	if (endpoint->block_backward_collectcall) {
		if (get_backward_collect_call(rdata)) {
			ast_log(LOG_NOTICE, "Not support backward collect call!\n");
			pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 603, NULL, NULL, NULL);
			return NULL;
		}
	}

	/* Added by yfyang, Bug 124822 */
	static pj_str_t STR_100REL = {"100rel", 6};

	if (!(options & PJSIP_INV_SUPPORT_100REL || options & PJSIP_INV_REQUIRE_100REL)) {
		pjsip_endpt_remove_capability(ast_sip_get_pjsip_endpoint(), PJSIP_H_SUPPORTED, NULL, &STR_100REL);
	}
#endif

	if (pjsip_inv_verify_request(rdata, &options, NULL, NULL, ast_sip_get_pjsip_endpoint(), &tdata) != PJ_SUCCESS) {
		if (tdata) {
			if (pjsip_endpt_send_response2(ast_sip_get_pjsip_endpoint(), rdata, tdata, NULL, NULL) != PJ_SUCCESS) {
				pjsip_tx_data_dec_ref(tdata);
			}
		} else {
			pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 500, NULL, NULL, NULL);
		}
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* for sca feature */
	if (endpoint->sca_enable) {
		/* check the Call-Info header */
		if (check_sca_call_info(rdata, endpoint)) {
			ast_log(LOG_NOTICE, "Something wrong with the user or the line index, SCA user[%s]\n", ast_sorcery_object_get_id(endpoint));
			pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 403, NULL, NULL, NULL);

			return NULL;
		}
	}
#endif

	dlg = ast_sip_create_dialog_uas(endpoint, rdata, &dlg_status);
	if (!dlg) {
		if (dlg_status != PJ_EEXISTS) {
			pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 500, NULL, NULL, NULL);
		}
		return NULL;
	}
	if (pjsip_inv_create_uas(dlg, rdata, NULL, options, &inv_session) != PJ_SUCCESS) {
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata, 500, NULL, NULL, NULL);
		pjsip_dlg_terminate(dlg);
#ifdef GRANDSTREAM_NETWORKS
#ifdef HAVE_PJSIP_DLG_CREATE_UAS_AND_INC_LOCK
		pjsip_dlg_dec_lock(dlg);
#endif
#endif
		return NULL;
	}

#ifdef GRANDSTREAM_NETWORKS
#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_add_ref(inv_session);
#endif

#ifdef HAVE_PJSIP_DLG_CREATE_UAS_AND_INC_LOCK
	pjsip_dlg_dec_lock(dlg);
#endif
#endif

#if defined(HAVE_PJSIP_REPLACE_MEDIA_STREAM) || defined(PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE)
	inv_session->sdp_neg_flags = PJMEDIA_SDP_NEG_ALLOW_MEDIA_CHANGE;
#endif
	if (pjsip_dlg_add_usage(dlg, &session_module, NULL) != PJ_SUCCESS) {
		if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) != PJ_SUCCESS) {
			pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		}
		pjsip_inv_send_msg(inv_session, tdata);
#ifdef GRANDSTREAM_NETWORKS
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(inv_session);
#endif
#endif
		return NULL;
	}
	return inv_session;
}

struct new_invite {
	/*! \brief Session created for the new INVITE */
	struct ast_sip_session *session;

	/*! \brief INVITE request itself */
	pjsip_rx_data *rdata;
};

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for Bug 208318, 2021.12.02 */
static void report_call_failed(struct ast_sip_session *session, pjsip_tx_data *tdata, int failed_code)
{
	if (!session || !tdata) {
		return;
	}

	if (!strncmp(ast_sorcery_object_get_id(session->endpoint), "trunk_", 6)) {
		return;
	}

	ast_sip_report_invite_failed(session, NULL, tdata, "INVITE", failed_code, ast_cause2str(hangup_sip2cause(failed_code)));

	return;
}
#endif

static int new_invite(struct new_invite *invite)
{
	pjsip_tx_data *tdata = NULL;
	pjsip_timer_setting timer;
	pjsip_rdata_sdp_info *sdp_info;
	pjmedia_sdp_session *local = NULL;
	char buffer[AST_SOCKADDR_BUFLEN];
#ifdef GRANDSTREAM_NETWORKS
	int  from_trunk = 0;
	int  err_exit = 0;
	const char *ept_id;
	static const pj_str_t contact_name = { "Contact", 7 };
	//static const pj_str_t gs_addr_name = { "GS-Address", 10 };
	pjsip_contact_hdr *contact_hdr;
	
	ast_log(LOG_VERBOSE, "[session:%p] new invite starting for endpoint:%s\n",
		invite->session, ast_sorcery_object_get_id(invite->session->endpoint));

	ept_id = ast_sorcery_object_get_id(invite->session->endpoint);
	if (!ept_id || strstr(ept_id, "trunk_")) {
		from_trunk = 1;
	}

	/* Added by yfyang, grandstream networks, for Bug 47622*/
	if (!from_trunk) {
		ast_devstate_changed(AST_DEVICE_INUSE, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ept_id);
	}

	pjsip_dialog *dlg = invite->session->inv_session->dlg;
	if (dlg == NULL) {
		ast_log(LOG_WARNING, "PJSIP dialog is NULL!");
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(invite->session->inv_session);
#endif
		return -1;
	}
	pjsip_dlg_inc_lock(dlg);
#endif

	/* From this point on, any calls to pjsip_inv_terminate have the last argument as PJ_TRUE
	 * so that we will be notified so we can destroy the session properly
	 */

	if (invite->session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Session already DISCONNECTED [reason=%d (%s)]\n",
			invite->session->inv_session->cause,
			pjsip_get_status_text(invite->session->inv_session->cause)->ptr);
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(invite->session->inv_session);
#endif
#ifdef GRANDSTREAM_NETWORKS
		pjsip_dlg_dec_lock(dlg);
#endif
		return -1;
	}

	switch (get_destination(invite->session, invite->rdata)) {
	case SIP_GET_DEST_EXTEN_FOUND:
		/* Things worked. Keep going */
#ifdef GRANDSTREAM_NETWORKS
	{
		/* Added by yfyang, for sca feature, the private role need to change the callee number */
		struct ast_sip_session *session = invite->session;
		enum sca_status status;

		if (session && (session->line_index > 0)) {
			status = ast_sip_get_sca_line_status((char *)session->endpoint->sca_sharedline, session->line_index); 

			ast_log(LOG_NOTICE, "new_invite, sharedline[%s], line[%d] in state[%d]", session->endpoint->sca_sharedline, session->line_index, status);

			if (session->endpoint->sca_role == AST_SIP_SCA_ROLE_TYPE_PRIVATE) {
				if ((status == SCA_Held) || (status == SCA_Active)) {
					ast_copy_string(session->exten, (char *)session->endpoint->sca_sharedline, AST_MAX_EXTENSION); 
				}
			}

			if (status == SCA_Idle || status == SCA_Seized) {
				ast_sip_set_sca_user_info((char *)session->endpoint->sca_sharedline, session->line_index, NULL, session->exten);
				ast_sip_set_sca_line_status((char *)session->endpoint->sca_sharedline, (int)session->line_index, SCA_Progressing); 
			} 
		}
	}
#endif
		break;
	case SIP_GET_DEST_UNSUPPORTED_URI:
		if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 416, NULL, NULL, &tdata) == PJ_SUCCESS) {
			ast_sip_session_send_response(invite->session, tdata);
		} else  {
			pjsip_inv_terminate(invite->session->inv_session, 416, PJ_TRUE);
		}

#ifdef GRANDSTREAM_NETWORKS
		err_exit = 1;
#endif
		goto end;
	case SIP_GET_DEST_EXTEN_PARTIAL:
		ast_debug(1, "Call from '%s' (%s:%s) to extension '%s' - partial match\n",
			ast_sorcery_object_get_id(invite->session->endpoint),
			invite->rdata->tp_info.transport->type_name,
			pj_sockaddr_print(&invite->rdata->pkt_info.src_addr, buffer, sizeof(buffer), 3),
			invite->session->exten);

		if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 484, NULL, NULL, &tdata) == PJ_SUCCESS) {
			ast_sip_session_send_response(invite->session, tdata);
		} else  {
			pjsip_inv_terminate(invite->session->inv_session, 484, PJ_TRUE);
		}
#ifdef GRANDSTREAM_NETWORKS
		err_exit = 1;
#endif
		goto end;
	case SIP_GET_DEST_EXTEN_NOT_FOUND:
	default:
		ast_log(LOG_NOTICE, "Call from '%s' (%s:%s) to extension '%s' rejected because extension not found in context '%s'.\n",
			ast_sorcery_object_get_id(invite->session->endpoint),
			invite->rdata->tp_info.transport->type_name,
			pj_sockaddr_print(&invite->rdata->pkt_info.src_addr, buffer, sizeof(buffer), 3),
			invite->session->exten,
			invite->session->endpoint->context);

		if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 404, NULL, NULL, &tdata) == PJ_SUCCESS) {
			ast_sip_session_send_response(invite->session, tdata);
		} else  {
			pjsip_inv_terminate(invite->session->inv_session, 404, PJ_TRUE);
		}

#ifdef GRANDSTREAM_NETWORKS
		err_exit = 1;
#endif
		goto end;
	};

	pjsip_timer_setting_default(&timer);
	timer.min_se = invite->session->endpoint->extensions.timer.min_se;
	timer.sess_expires = invite->session->endpoint->extensions.timer.sess_expires;
	pjsip_timer_init_session(invite->session->inv_session, &timer);

	/*
	 * At this point, we've verified what we can that won't take awhile,
	 * so let's go ahead and send a 100 Trying out to stop any
	 * retransmissions.
	 */
#ifndef GRANDSTREAM_NETWORKS
	if (pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 100, NULL, NULL, &tdata) != PJ_SUCCESS) {
		pjsip_inv_terminate(invite->session->inv_session, 500, PJ_TRUE);
		goto end;
	}
#else
	/* If status is 422, we can set up this call*/
	pj_status_t status = pjsip_inv_initial_answer(invite->session->inv_session, invite->rdata, 100, NULL, NULL, &tdata);

	ast_log(LOG_NOTICE, "[new_invite] pjsip_inv_initial_answer result: %d\n", PJSIP_ERRNO_TO_SIP_STATUS(status));

	if (status != PJ_SUCCESS && PJSIP_ERRNO_TO_SIP_STATUS(status) != 422) {
		pjsip_inv_terminate(invite->session->inv_session, 500, PJ_TRUE);
		err_exit = 1;
		goto end;
	}
#endif

	ast_sip_session_send_response(invite->session, tdata);

#ifdef GRANDSTREAM_NETWORKS
	ast_sip_session_set_call_flip(invite->session,invite->rdata);
	/* Added by yfyang, for Bug 173103 */
	if (!from_trunk && !invite->session->is_callflip_call) {
		char  family[80] = {0};
		char  data[80]   = {0};
		char  group[80]  = {0};
		int   max_calls  = 0;
		int   cur_calls  = 0;

		snprintf(family, sizeof(family), "AAUSER/%s", ept_id);
		ast_ucm_db_get(family, "MAX_CHANS", data, sizeof(data));
		if (strlen(data) && (max_calls = atoi(data))) {
			snprintf(group, sizeof(group), "OUT_EXT_%s", ept_id);
			cur_calls = ast_app_group_get_count(group, NULL);
			if (cur_calls >= max_calls) {
				ast_log(LOG_WARNING, "[%s] has [%d] calls now, reached the max_calls[%d].\n", ept_id, cur_calls, max_calls);
				tdata = NULL;
				if ((pjsip_inv_end_session(invite->session->inv_session, 486, NULL, &tdata) == PJ_SUCCESS) && tdata) {
					ast_sip_session_send_response(invite->session, tdata);
				}
				err_exit = 1;
				goto end;
			}
		}
	}
#endif

	sdp_info = pjsip_rdata_get_sdp_info(invite->rdata);
	if (sdp_info && (sdp_info->sdp_err == PJ_SUCCESS) && sdp_info->sdp) {
#ifdef GRANDSTREAM_NETWORKS
		set_is_sdp_ipv6(invite->session, sdp_info->sdp);

		/*record the source ip*/
		 if (invite->session->is_webrtc_call) {

			if(invite->session->endpoint)
			{
				char *p;
				char webrtc_host[256] = {0};
				RAII_VAR(struct ast_sip_aor *, found_aor, NULL, ao2_cleanup);
				RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);
				ast_sip_location_retrieve_contact_and_aor_from_list_filtered(invite->session->endpoint->aors, AST_SIP_CONTACT_FILTER_REACHABLE,
					&found_aor, &contact);
				if(contact)
				{
					ast_log(LOG_DEBUG,"find webrtc original addr: %s",contact->original_address);
					if(NULL != (p = strchr(contact->original_address,':')))
					{
						strncpy(webrtc_host,contact->original_address, p-contact->original_address );
					}
					else
					{
						strcpy(webrtc_host,contact->original_address);
					}
					if(inet_pton(AF_INET, webrtc_host, (void *)(&invite->session->src_addr))){
							invite->session->src_addr_ver = AF_INET;
					} else if (inet_pton(AF_INET6, webrtc_host, (void *)(&invite->session->src_addr6))){
							invite->session->src_addr_ver = AF_INET6;
					}

					set_support_gs_fec_for_session(invite->session, contact, NULL);
				}
			}

			#if 0
			pjsip_generic_string_hdr *gs_addr_hdr;
			char webrtc_host[256];
			gs_addr_hdr = pjsip_msg_find_hdr_by_name(invite->rdata->msg_info.msg, &gs_addr_name, NULL);
			if (gs_addr_hdr) {
				ast_copy_pj_str(webrtc_host, &gs_addr_hdr->hvalue, sizeof(webrtc_host));
				ast_log(LOG_DEBUG, "host addr from webrtc is [%s]", webrtc_host);
				if(inet_pton(AF_INET, webrtc_host, (void *)(&invite->session->src_addr))){
					invite->session->src_addr_ver = AF_INET;
				} else if (inet_pton(AF_INET6, webrtc_host, (void *)(&invite->session->src_addr6))){
					invite->session->src_addr_ver = AF_INET6;
				}
			}
			#endif
		} else {
			contact_hdr = (pjsip_contact_hdr *)pjsip_msg_find_hdr_by_name(invite->rdata->msg_info.msg, &contact_name, NULL);
			if (contact_hdr) {
				pjsip_sip_uri *sip_uri = (pjsip_sip_uri *)pjsip_uri_get_uri(contact_hdr->uri);
				if (sip_uri) {
					char host[256];
					ast_copy_pj_str(host, &sip_uri->host, sizeof(host));
					ast_log(LOG_DEBUG, "host addr is [%s]", host);
					if(inet_pton(AF_INET, host, (void *)(&invite->session->src_addr))){
						invite->session->src_addr_ver = AF_INET;
					} else if (inet_pton(AF_INET6, host, (void *)(&invite->session->src_addr6))){
						invite->session->src_addr_ver = AF_INET6;
					}
				}
			}
		}
#endif
		if (handle_incoming_sdp(invite->session, sdp_info->sdp)) {
			tdata = NULL;
			if (pjsip_inv_end_session(invite->session->inv_session, 488, NULL, &tdata) == PJ_SUCCESS
				&& tdata) {
#ifdef GRANDSTREAM_NETWORKS
				/* Add by bfdu for Bug 208318, 2021.12.02 */
				report_call_failed(invite->session, tdata, 488);
#endif
				ast_sip_session_send_response(invite->session, tdata);
			}

#ifdef GRANDSTREAM_NETWORKS
			err_exit = 1;
#endif
			goto end;
		}
		/* We are creating a local SDP which is an answer to their offer */
		local = create_local_sdp(invite->session->inv_session, invite->session, sdp_info->sdp);
	} else {
		/* We are creating a local SDP which is an offer */
		local = create_local_sdp(invite->session->inv_session, invite->session, NULL);
	}

	/* If we were unable to create a local SDP terminate the session early, it won't go anywhere */
	if (!local) {
		tdata = NULL;
		if (pjsip_inv_end_session(invite->session->inv_session, 500, NULL, &tdata) == PJ_SUCCESS
			&& tdata) {
			ast_sip_session_send_response(invite->session, tdata);
		}

#ifdef GRANDSTREAM_NETWORKS
		err_exit = 1;
#endif
		goto end;
	}

	pjsip_inv_set_local_sdp(invite->session->inv_session, local);
	pjmedia_sdp_neg_set_prefer_remote_codec_order(invite->session->inv_session->neg, PJ_FALSE);
#ifdef PJMEDIA_SDP_NEG_ANSWER_MULTIPLE_CODECS
	if (!invite->session->endpoint->preferred_codec_only) {
		pjmedia_sdp_neg_set_answer_multiple_codecs(invite->session->inv_session->neg, PJ_TRUE);
	}
#endif

	handle_incoming_request(invite->session, invite->rdata);

end:
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for Bug 171882 */
	if (err_exit && invite->session && invite->session->line_index && invite->session->endpoint) {
		char *user = NULL;
		struct ast_sip_session *sess = invite->session;

		if (sess->aor) {
			user = (char *)ast_sorcery_object_get_id(sess->aor);
		} else {
			user = (char *)sess->endpoint->aors;
		}
		ast_log(LOG_NOTICE, "new_invite, sharedline[%s], index[%d], user[%s], remove user.", sess->endpoint->sca_sharedline, sess->line_index, user); 
		ast_sip_sca_line_remove_user(NULL, (char *)sess->endpoint->sca_sharedline, sess->line_index, user); 

		sess->line_index = 0;
	}

	/*Added by pyzhong, for Bug 177423*/
	if (err_exit && invite->session && (invite->session->terminal_type == AST_TERMINAL_TYPE_WEBRTC || invite->session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)
		&& (ast_exists_extension(NULL, "ext-conference", invite->session->exten, 1, NULL) || ast_exists_extension(NULL, "ext-video-conference", invite->session->exten, 1, NULL))) {
		struct ast_app *app = NULL;
		char appdata[128] = {0};
		const char *term_type = "other_type";

		if (invite->session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
			term_type = "webrtc";
		} else {
			term_type = "wave_app";
		}

		snprintf(appdata, sizeof(appdata), "DeleteAuthUser,%s,%s,%s", invite->session->exten, ast_sorcery_object_get_id(invite->session->endpoint), term_type);
		app = pbx_findapp("ConfBridgeCommand");
		if (app) {
			pbx_app_exec(app, appdata);
		}
	}
#endif

#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(invite->session->inv_session);
#endif
#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, grandstream networks, for Bug 47622*/
	if (!from_trunk) {
		ast_devstate_changed(AST_DEVICE_UNKNOWN, AST_DEVSTATE_CACHABLE, "PJSIP/%s", ept_id);
	}
	pjsip_dlg_dec_lock(dlg);
#endif

	return 0;
}

static void handle_new_invite_request(pjsip_rx_data *rdata)
{
	RAII_VAR(struct ast_sip_endpoint *, endpoint,
			ast_pjsip_rdata_get_endpoint(rdata), ao2_cleanup);
	static const pj_str_t identity_str = { "Identity", 8 };
	const pj_str_t use_identity_header_str = {
		AST_STIR_SHAKEN_RESPONSE_STR_USE_IDENTITY_HEADER,
		strlen(AST_STIR_SHAKEN_RESPONSE_STR_USE_IDENTITY_HEADER)
	};
			
	pjsip_tx_data *tdata = NULL;
	pjsip_inv_session *inv_session = NULL;
	struct ast_sip_session *session;
	struct new_invite invite;

	ast_assert(endpoint != NULL);

	if ((endpoint->stir_shaken & AST_SIP_STIR_SHAKEN_VERIFY) &&
		!ast_sip_rdata_get_header_value(rdata, identity_str)) {
		pjsip_endpt_respond_stateless(ast_sip_get_pjsip_endpoint(), rdata,
			AST_STIR_SHAKEN_RESPONSE_CODE_USE_IDENTITY_HEADER, &use_identity_header_str, NULL, NULL);
		ast_debug(3, "No Identity header when we require one\n");
		return;
	}

	inv_session = pre_session_setup(rdata, endpoint);
	if (!inv_session) {
		/* pre_session_setup() returns a response on failure */
		return;
	}

#ifdef HAVE_PJSIP_INV_SESSION_REF
	if (pjsip_inv_add_ref(inv_session) != PJ_SUCCESS) {
		ast_log(LOG_ERROR, "Can't increase the session reference counter\n");
		if (inv_session->state != PJSIP_INV_STATE_DISCONNECTED) {
			if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) == PJ_SUCCESS) {
				pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
			} else {
				pjsip_inv_send_msg(inv_session, tdata);
			}
		}
		return;
	}
#endif

#ifdef GRANDSTREAM_NETWORKS
	/*Increase ref in pre_session_setup, so decrease ref here*/
#ifdef HAVE_PJSIP_INV_SESSION_REF
	pjsip_inv_dec_ref(inv_session);
#endif
	if (endpoint->click2call_enable && !ast_sip_get_support_webrtc_trunk()) {
		ast_log(LOG_DEBUG, "handle_new_invite_request can't support webrtc trunk call\n");
		if (inv_session->state != PJSIP_INV_STATE_DISCONNECTED) {
			if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) == PJ_SUCCESS) {
				pjsip_inv_send_msg(inv_session, tdata);
			} else {
				pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
			}
		}
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(inv_session);
#endif
		return;
	}
#endif

	session = ast_sip_session_alloc(endpoint, NULL, inv_session, rdata);
	if (!session) {
#ifdef GRANDSTREAM_NETWORKS
		if (inv_session->state != PJSIP_INV_STATE_DISCONNECTED) {
#endif
		if (pjsip_inv_initial_answer(inv_session, rdata, 500, NULL, NULL, &tdata) == PJ_SUCCESS) {
			pjsip_inv_terminate(inv_session, 500, PJ_FALSE);
		} else {
			pjsip_inv_send_msg(inv_session, tdata);
		}
#ifdef GRANDSTREAM_NETWORKS
		}
#endif
#ifdef HAVE_PJSIP_INV_SESSION_REF
		pjsip_inv_dec_ref(inv_session);
#endif
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	char call_id[256] = { 0 };
	ast_copy_pj_str(call_id, &inv_session->dlg->call_id->id, sizeof(call_id));
	ast_log(LOG_DEBUG, "[session:%p] handle_new_invite_request endpoint:%s, dialog:%p, call-id:%s, session refcnt[%d]\n",
		session, ast_sorcery_object_get_id(session->endpoint), inv_session->dlg, call_id, ao2_ref(session, 0));

	/* Added by yfyang, for sca feature */
	if (endpoint->sca_enable) {
		if (ast_sip_get_sca_call_info_from_rdata(rdata, &session->line_index, NULL)) {
			session->line_index = ast_sip_get_sca_line_index((char *)endpoint->sca_sharedline, SCA_Seized, (char *)ast_sorcery_object_get_id(endpoint));
		}
	}
	set_yealink_flag(rdata, session);

	/*start add by zyxia for wake up gswave,if receive invite  from gswave,add the call count*/
	if (!session->has_add_gs_call_count) {
		ast_sip_session_end_if_deferred(session);
		ast_sip_session_defer_termination_cancel(session);
		ast_sip_set_used_gs_call_count(session,rdata,1);
	}
	/*end add by zyxia for wake up gswave,if receive invite from gswave,add the call count*/

	set_caller_webrtc_call_flag(session, rdata);

	set_ice_flag(session, rdata);

	set_terminal_type(session, rdata);

	set_support_whiteboard_flag(session, rdata);
	set_support_gs_state_flag(session, rdata);

	set_candidate_host_only_flag_for_caller(session, rdata);

	if(ast_media_role_get() == MEDIA_ROLE_MASTER && ast_network_mode_get() != SWITCH_MODE 
		&& !ast_strlen_zero(ast_avs_get_avs_address()) && strcmp("127.0.0.1",ast_avs_get_avs_address())){
		set_session_diff_multiavs_ip_flag_for_caller(session, rdata);
	}

	set_session_allow_events(session, rdata);

	ast_sip_session_set_ucm_trunk(session, rdata);

	set_session_emergency_call(session, rdata);

	ast_sip_session_set_trunk_info(session, rdata);

	record_replace_channel(session, rdata);

	ast_sip_session_set_avs_address(session);

	ast_sip_session_set_gs_client_info(rdata, session, 1);

	session->blocking_collect_calls = get_backward_collect_call(rdata);

	/* Added by yfyang, for sip ssh proxy feature */
	session->ssh_proxy_flag = ast_sip_check_ssh_proxy_flag(rdata);
	if (session->ssh_proxy_flag) {
		ast_log(LOG_NOTICE, "Set ssh_proxy_flag to dlg[%p]", inv_session->dlg);
		pjsip_dlg_set_proxy_flag(inv_session->dlg, 1);
	}


	session->trans_type = ast_strdup(rdata->tp_info.transport->type_name);
	if(HA_ROLE_ACTIVE == ast_ha_get_role()
		&& inv_session->dlg->remote.contact
		&& inv_session->dlg->remote.contact->uri)
	{
		pjsip_sip_uri *contact_uri = NULL;
		contact_uri = pjsip_uri_get_uri(inv_session->dlg->remote.contact->uri);
		if(contact_uri){
			char tmp_str[64] = {0};
			strncpy(tmp_str, contact_uri->host.ptr, contact_uri->host.slen);
			session->active_contact_addr = ast_strdup(tmp_str);
		}
	}

	set_if_type_for_session(session, NULL, NULL);
	ast_log(LOG_NOTICE, "session[%p] incoming, g_vlan_wan_enable[%d], if_type[%d]", session, g_vlan_wan_enable, session->if_type);

	set_media_turn_relay(session);

	clear_media_turn_relay_for_remote_support_ice(session, rdata);

	set_caller_gds_call(session, rdata);
	set_caller_gsc_call(session, rdata);
	set_caller_hikvision_call(session, rdata);

	ast_sip_session_set_device_ua(session, rdata, PJSIP_INVITE_METHOD);

	conference_invite_need_reject(session, rdata);//del by zyxia Bug 266995 ,2023.7.28

#ifdef BUILD_ON_CLOUD_PBX
	set_sbc_endpoint_addr(session, rdata);  /* Added by yfyang, for clouducm support UDP */
	set_dest_addrs_to_inv_session(session); /* Added by yfyang, for Bug 303715 */
#endif
#endif

	/*
	 * The current thread is supposed be the session serializer to prevent
	 * any initial INVITE retransmissions from trying to setup the same
	 * call again.
	 */
	ast_assert(ast_taskprocessor_is_task(session->serializer));

	invite.session = session;
	invite.rdata = rdata;
	new_invite(&invite);

	ao2_ref(session, -1);
}

static pj_bool_t does_method_match(const pj_str_t *message_method, const char *supplement_method)
{
	pj_str_t method;

	if (ast_strlen_zero(supplement_method)) {
		return PJ_TRUE;
	}

	pj_cstr(&method, supplement_method);

	return pj_stristr(&method, message_method) ? PJ_TRUE : PJ_FALSE;
}

static pj_bool_t has_supplement(const struct ast_sip_session *session, const pjsip_rx_data *rdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_method *method = &rdata->msg_info.msg->line.req.method;

	if (!session) {
		return PJ_FALSE;
	}

	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (does_method_match(&method->name, supplement->method)) {
			return PJ_TRUE;
		}
	}
	return PJ_FALSE;
}
/*!
 * \brief Called when a new SIP request comes into PJSIP
 *
 * This function is called under two circumstances
 * 1) An out-of-dialog request is received by PJSIP
 * 2) An in-dialog request that the inv_session layer does not
 *    handle is received (such as an in-dialog INFO)
 *
 * Except for INVITEs, there is very little we actually do in this function
 * 1) For requests we don't handle, we return PJ_FALSE
 * 2) For new INVITEs, handle them now to prevent retransmissions from
 *    trying to setup the same call again.
 * 3) For in-dialog requests we handle, we process them in the
 *    .on_state_changed = session_inv_on_state_changed or
 *    .on_tsx_state_changed = session_inv_on_tsx_state_changed
 *    callbacks instead.
 */
static pj_bool_t session_on_rx_request(pjsip_rx_data *rdata)
{
	pj_status_t handled = PJ_FALSE;
	pjsip_dialog *dlg = pjsip_rdata_get_dlg(rdata);
	pjsip_inv_session *inv_session;

	switch (rdata->msg_info.msg->line.req.method.id) {
	case PJSIP_INVITE_METHOD:
		if (dlg) {
			ast_log(LOG_WARNING, "on_rx_request called for INVITE in mid-dialog?\n");
			break;
		}
		handled = PJ_TRUE;
		handle_new_invite_request(rdata);
		break;
	default:
		/* Handle other in-dialog methods if their supplements have been registered */
		handled = dlg && (inv_session = pjsip_dlg_get_inv_session(dlg)) &&
			has_supplement(inv_session->mod_data[session_module.id], rdata);
		break;
	}

	return handled;
}

static void resend_reinvite(pj_timer_heap_t *timer, pj_timer_entry *entry)
{
	struct ast_sip_session *session = entry->user_data;

	ast_debug(3, "Endpoint '%s(%s)' re-INVITE collision timer expired.\n",
		ast_sorcery_object_get_id(session->endpoint),
		session->channel ? ast_channel_name(session->channel) : "");

	if (AST_LIST_EMPTY(&session->delayed_requests)) {
		/* No delayed request pending, so just return */
		ao2_ref(session, -1);
		return;
	}
	if (ast_sip_push_task(session->serializer, invite_collision_timeout, session)) {
		/*
		 * Uh oh.  We now have nothing in the foreseeable future
		 * to trigger sending the delayed requests.
		 */
		ao2_ref(session, -1);
	}
}

static void reschedule_reinvite(struct ast_sip_session *session, ast_sip_session_response_cb on_response)
{
	pjsip_inv_session *inv = session->inv_session;
	pj_time_val tv;

	ast_debug(3, "Endpoint '%s(%s)' re-INVITE collision.\n",
		ast_sorcery_object_get_id(session->endpoint),
		session->channel ? ast_channel_name(session->channel) : "");
	if (delay_request(session, NULL, NULL, on_response, 1, DELAYED_METHOD_INVITE, NULL)) {
		return;
	}
	if (pj_timer_entry_running(&session->rescheduled_reinvite)) {
		/* Timer already running.  Something weird is going on. */
		ast_debug(1, "Endpoint '%s(%s)' re-INVITE collision while timer running!!!\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "");
		return;
	}

	tv.sec = 0;
	if (inv->role == PJSIP_ROLE_UAC) {
		tv.msec = 2100 + ast_random() % 2000;
	} else {
		tv.msec = ast_random() % 2000;
	}
	pj_timer_entry_init(&session->rescheduled_reinvite, 0, session, resend_reinvite);

	ao2_ref(session, +1);
	if (pjsip_endpt_schedule_timer(ast_sip_get_pjsip_endpoint(),
		&session->rescheduled_reinvite, &tv) != PJ_SUCCESS) {
		ao2_ref(session, -1);
	}
}

static void __print_debug_details(const char *function, pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e)
{
	int id = session_module.id;
	struct ast_sip_session *session = NULL;

	if (!DEBUG_ATLEAST(5)) {
		/* Debug not spamy enough */
		return;
	}

	ast_log(LOG_DEBUG, "Function %s called on event %s\n",
		function, pjsip_event_str(e->type));
	if (!inv) {
		ast_log(LOG_DEBUG, "Transaction %p does not belong to an inv_session?\n", tsx);
		ast_log(LOG_DEBUG, "The transaction state is %s\n",
			pjsip_tsx_state_str(tsx->state));
		return;
	}
	if (id > -1) {
		session = inv->mod_data[session_module.id];
	}
	if (!session) {
		ast_log(LOG_DEBUG, "inv_session %p has no ast session\n", inv);
	} else {
		ast_log(LOG_DEBUG, "The state change pertains to the endpoint '%s(%s)'\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "");
	}
	if (inv->invite_tsx) {
		ast_log(LOG_DEBUG, "The inv session still has an invite_tsx (%p)\n",
			inv->invite_tsx);
	} else {
		ast_log(LOG_DEBUG, "The inv session does NOT have an invite_tsx\n");
	}
	if (tsx) {
		ast_log(LOG_DEBUG, "The %s %.*s transaction involved in this state change is %p\n",
			pjsip_role_name(tsx->role),
			(int) pj_strlen(&tsx->method.name), pj_strbuf(&tsx->method.name),
			tsx);
		ast_log(LOG_DEBUG, "The current transaction state is %s\n",
			pjsip_tsx_state_str(tsx->state));
		ast_log(LOG_DEBUG, "The transaction state change event is %s\n",
			pjsip_event_str(e->body.tsx_state.type));
	} else {
		ast_log(LOG_DEBUG, "There is no transaction involved in this state change\n");
	}
	ast_log(LOG_DEBUG, "The current inv state is %s\n", pjsip_inv_state_name(inv->state));
}

#define print_debug_details(inv, tsx, e) __print_debug_details(__PRETTY_FUNCTION__, (inv), (tsx), (e))

#ifdef GRANDSTREAM_NETWORKS
static void handle_incoming_bye_request(struct ast_sip_session *session, const pj_str_t *message_method, pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_generic_string_hdr *hdr = NULL;
	const pj_str_t name = { "Reason", 6 };
	char reason_str[64] = { 0 };

	if (!session || !message_method || !rdata) {
		return;
	}

	if (!does_method_match(message_method, "BYE")) {
		return;
	}

	if (!session->channel) {
		ast_log(LOG_DEBUG, "[session:%p] handle_incoming_bye_request channel is NULL, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_channel_recv_bye_set(session->channel, 1);
	ast_queue_control(session->channel, AST_CONTROL_NOTIFY_RECV_BYE);

	msg = rdata->msg_info.msg;
	if (!msg) {
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	if (!hdr) {
		return;
	}

	ast_copy_pj_str(reason_str, &hdr->hvalue, sizeof(reason_str));
	ast_log(LOG_DEBUG, "[session:%p] handle_incoming_bye_request get reason:%s, endpoint:%s\n",
		session, reason_str, ast_sorcery_object_get_id(session->endpoint));
	if (strcasestr(reason_str, "Switch to conference mode")) {
		ast_queue_control(session->channel, AST_CONTROL_NOTIFY_HANGUP_SWITCH_TO_CONF);
	}

	return;
}

/* Removed by yfyang, for Bug 241389 */
#if 0
static void set_cc_failed_recv_cancel(struct ast_sip_session *session)
{
	int cc_core_id = -1;

	if (!session || !session->channel) {
		return;
	}

	ast_cc_clear_ignore(session->channel);
	cc_core_id = ast_cc_get_current_core_id(session->channel);

	if (-1 != cc_core_id) {
		ast_cc_failed(cc_core_id, "The call of '%s' has been cancelled", ast_channel_name(session->channel));
	}

	return;
}
#endif

/* Add by bfdu for Bug 217885, 2022.9.30 */
static void send_notify_for_miss_call(struct ast_sip_session *session)
{
	struct ast_app *app = NULL;
	char app_data[1024] = { 0 };
	char *caller_name = NULL;
	char *caller_id = NULL;

	if (!session || !session->channel) {
		ast_log(LOG_ERROR, "send_notify_for_miss_call param is NULL\n");
		return;
	}

	const char *missed_contacts = pbx_builtin_getvar_helper(session->channel, "MISSED_CONTACTS");
	if (ast_strlen_zero(missed_contacts)) {
		return;
	}

	const char *in_group = pbx_builtin_getvar_helper(session->channel, "IN_RINGGROUP");
	if (!ast_strlen_zero(in_group) && strncmp(in_group, "yes", 3)) {
		ast_log(LOG_WARNING, "[session:%p] send_notify_for_miss_call is not ringgroup, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	const char *group_name = pbx_builtin_getvar_helper(session->channel, "GROUP_NAME");
	const char *group_number = pbx_builtin_getvar_helper(session->channel, "GROUP_NUMBER");
	if (ast_strlen_zero(group_name) || ast_strlen_zero(group_number)) {
		ast_log(LOG_ERROR, "[session:%p] send_notify_for_miss_call ringgroup has no name or number, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	caller_name = S_COR(ast_channel_caller(session->channel)->id.name.valid, ast_channel_caller(session->channel)->id.name.str, "Anonymous");
	caller_id = S_COR(ast_channel_caller(session->channel)->id.number.valid, ast_channel_caller(session->channel)->id.number.str, "Anonymous");

	snprintf(app_data, sizeof(app_data), "RG,%s,%s,%s <%s>,%s", group_name, group_number, caller_name, caller_id, missed_contacts);

	ast_log(LOG_DEBUG, "[session:%p] send_notify_for_miss_call app_data:%s, endpoint:%s\n",
		session, app_data, ast_sorcery_object_get_id(session->endpoint));

	app = pbx_findapp("MissedCall");
	if (app) {
		pbx_exec(session->channel, app, app_data);
	} else {
		ast_log(LOG_ERROR, "send_notify_for_miss_call can't get MissedCall APP\n");
	}

	return;
}

/*start added by zyxia for trunk info cei ,2022.12.9*/
int ast_sip_session_trunk_conf_need_subscribe(struct ast_sip_session *session,struct pjsip_rx_data *rdata)
{
	int ret = 0;
	static const pj_str_t conf_call_type_field = { "X-Grandstream-Call-Type", 23 };
	pjsip_generic_string_hdr* conf_call_type_hdr = NULL;
	pj_str_t conf_call_type_value;
	static const pj_str_t remote_channel_id = { "Channel-ID", 10 };
	pjsip_generic_string_hdr* channel_id_hdr = NULL;
	pj_str_t remote_channel_id_value;

	if (!session  || !rdata)
		return ret;

	ast_log(LOG_DEBUG, "ast_sip_session_trunk_conf_need_subscribe\n");

	if(!ast_strlen_zero(session->trunk_conf_num)) {
		ast_log(LOG_WARNING, "ast_sip_session_trunk_conf_need_subscribe  trunk conf num is not null,had already  subcirbe\n");
		return ret;
	}

	//if is trunk conf,
	if(session->channel && ast_channel_name(session->channel) && strcasestr(ast_channel_name(session->channel),"PJSIP/trunk")) {
		pjsip_contact_hdr *contact = NULL;
		pjsip_sip_uri *contact_uri = NULL;
		/*get ermote conf number from contact user*/
		contact = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
		if (contact == NULL || contact->uri == NULL) {
			ast_log(LOG_WARNING, "contact URI header is NULL\n");
			return ret;
		}

		pj_str_t isfocus_str = { "isfocus", 7};
		pjsip_param *isfocus = pjsip_param_find(&contact->other_param, &isfocus_str);
		if (!isfocus) {
			ast_log(LOG_DEBUG, "contact dont have isfocus,dont need subscribe\n");
			return ret;
		}

		/*Start added by zyxia for 261097,2023.6.15*/
		if (!(conf_call_type_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &conf_call_type_field, NULL))) {
			ast_log(LOG_DEBUG, "Not had X-Grandstream-Call-Type header. dont need subscribe");
			return ret;
		}

		pj_strdup_with_null(rdata->tp_info.pool, &conf_call_type_value, &conf_call_type_hdr->hvalue);
		if (pj_strnicmp2(&conf_call_type_value, "Conference", 10) &&
			pj_strnicmp2(&conf_call_type_value, "VideoConference", 15)) {
			ast_log(LOG_DEBUG, "The invite isn't conference invite. dont need subscribe");
			return ret;
		}
		/*End added by zyxia for 261097,2023.6.15*/


		/*Start addded by zyxia for Bug 280759 ,trunk conf ,2023.12.20 */
		if ((channel_id_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &remote_channel_id, NULL))) {

			pj_strdup_with_null(rdata->tp_info.pool, &remote_channel_id_value, &channel_id_hdr->hvalue);
			session->remote_trunk_channel_id = ast_calloc(1, remote_channel_id_value.slen + 1);
			if (session->remote_trunk_channel_id) {
				ast_copy_pj_str(session->remote_trunk_channel_id, &remote_channel_id_value, remote_channel_id_value.slen + 1);
			}
		}
		/*End addded by zyxia for Bug 280759 ,trunk conf ,2023.12.20 */

		session->is_trunk_conf_invite=1; //added by zyxia for 266995

		contact_uri = pjsip_uri_get_uri(contact->uri);
		if (contact_uri == NULL) {
			ast_log(LOG_WARNING, "Failed to get contact URI\n");
			return ret;
		}

		session->trunk_conf_num = ast_calloc(1, contact_uri->user.slen + 1);
		if (!session->trunk_conf_num) {
			ast_log(LOG_ERROR, "[session:%p] trunk_conf_num, can't malloc trunk_channel_id, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return ret;
		}
		ast_copy_pj_str(session->trunk_conf_num, &contact_uri->user, contact_uri->user.slen + 1);

		ast_log(LOG_DEBUG, "session:%p , ast_sip_session_trunk_conf_need_subscribe:%s,%s\n",session,session->trunk_conf_num,ast_channel_name(session->channel));
		ret = 1;
	}

	return  ret;

}

/*start added by zyxia for trunk record state,Bug 267640  ,2023.8.11*/
int ast_sip_session_trunk_record_need_forward(struct ast_sip_session *session,struct pjsip_rx_data *rdata)
{
	int ret = 0;
	pjsip_rdata_sdp_info *sdp_info;
	/* See if there are registered handlers for this media stream type */
	char media[20]={0};
	enum ast_media_type type;

	pjmedia_sdp_attr *record_attr = NULL;
	struct pjmedia_sdp_media* sdp_media = NULL;
	char attr_value[256];
	const char *pre_rec_enable = NULL;
	int last_rec_state=0,cur_rec_state = 0;

	if (!session  || !rdata)
		return ret;

	ast_log(LOG_DEBUG, "ast_sip_session_trunk_record_need_forward\n");

	if(!session->ucm_trunk) {
		ast_log(LOG_DEBUG, "ast_sip_session_trunk_record_need_forward is not from ucm trunk ,ignore it\n");
		return ret;
	}

	if(session->channel && ast_channel_name(session->channel) && strcasestr(ast_channel_name(session->channel),"PJSIP/trunk")) {

		pre_rec_enable = pbx_builtin_getvar_helper(session->channel, "REC_ENABLE");
		if (!ast_strlen_zero(pre_rec_enable) && !strcmp(pre_rec_enable, "yes")){
			last_rec_state = 1;
		}

		sdp_info = pjsip_rdata_get_sdp_info(rdata);
		if (sdp_info && (sdp_info->sdp_err == PJ_SUCCESS) && sdp_info->sdp) {

			for (int i = 0; i < sdp_info->sdp->media_count; ++i) {
				sdp_media = sdp_info->sdp->media[i];

				if(!sdp_media)
					continue;

				/* We need a null-terminated version of the media string */
				ast_copy_pj_str(media, &sdp_media->desc.media, sizeof(media));
				type = ast_media_type_from_str(media);
				if(type == AST_MEDIA_TYPE_AUDIO) {
					record_attr = pjmedia_sdp_attr_find2(sdp_media->attr_count, sdp_media->attr, "record", NULL);
					if (record_attr) {
						ast_copy_pj_str(attr_value, (pj_str_t*)&record_attr->value, sizeof(attr_value));
						if (!strcasecmp("on", attr_value)) {
							pbx_builtin_setvar_helper(session->channel, "REC_ENABLE", "yes");
							cur_rec_state = 1;
						}
						else
							pbx_builtin_setvar_helper(session->channel, "REC_ENABLE", "");
					}
					else
						pbx_builtin_setvar_helper(session->channel, "REC_ENABLE", "");
					break;
				}
			}

			//added for Bug 312846 , if the updata has no sdp,dont handle it
			if(last_rec_state != cur_rec_state) {
				struct ast_channel *real_peer = ast_channel_real_peer(session->channel);
				if (real_peer) {
					pbx_builtin_setvar_helper(real_peer, "RECORD_ATTR_SDP_TRIGGER", "yes");
					if(cur_rec_state)
						pbx_builtin_setvar_helper(real_peer, "REC_ENABLE", "yes");
					else
						pbx_builtin_setvar_helper(real_peer, "REC_ENABLE", "");
					ast_log(LOG_DEBUG, "ast_sip_session_trunk_record_need_forward  channel:%s,last rec state is %d,now rec_state is %d,need send update to peer chan:%s\n",ast_channel_name(session->channel),last_rec_state,cur_rec_state,ast_channel_name(real_peer));
					ast_channel_unref(real_peer);
				}
				ret = 1;
			}
		}
	}

	return  ret;

}
/*End added by zyxia for trunk record state,Bug 267640  ,2023.8.11*/

static const struct {
    const char *warning;
    const char *reason;
} warning2reason[] = {
    {"399 GS \"Do-not-disturb is on\"", "DND"},
    {"399 GS \"All lines are in use\"", "INUSE"},
    {"399 GS \"Dialog not found\"", "No dialog"},
    {"399 GS \"Anonymous call not allowed\"", "Anonymous not allowed"},
    {"399 GS \"Requested user ID is unknown\"", "Unknown user"},
    {"399 GS \"Missing Refer-To header\"", "No Refer-To"},
    {"399 GS \"Invalid Refer-To header\"", "Invalid Refer-To"},
    {"399 GS \"Request not supported\"", "Method not allowed"},
    {"399 GS \"Previous INVITE is not completed or terminated\"", "INVITE pending"},
};

static const char *ast_sip_warning2reason(const pj_str_t *warning)
{
    int i;
    for (i=0; i < ARRAY_LEN(warning2reason); i++) {
        if (!pj_stricmp2(warning, warning2reason[i].warning)) {
            return warning2reason[i].reason;
        }
    }

    return NULL;
}

static void ast_sip_warning_hdr_proc(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
    const pj_str_t warning_name  = {"Warning", 7};
    char *reason = NULL;

    if (!session || !session->channel || !rdata) {
        return;
    }

    pjsip_generic_string_hdr *hdr = (pjsip_generic_string_hdr *)&rdata->msg_info.msg->hdr;

    while ((hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &warning_name, hdr->next))) {
        /* check dnd warning */
        reason = (char *)ast_sip_warning2reason(&hdr->hvalue);
        if (!ast_strlen_zero(reason)) {
            ast_log(LOG_NOTICE, "chan[%s], get reason[%s]", session->channel ? ast_channel_name(session->channel) :"", reason);
            ast_channel_reject_reason_set(session->channel, reason);
            return;
        }
    }

    return;
}
#endif

static void handle_incoming_request(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_request_line req = rdata->msg_info.msg->line.req;

	ast_debug(3, "Method is %.*s\n", (int) pj_strlen(&req.method.name), pj_strbuf(&req.method.name));
	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (supplement->incoming_request && does_method_match(&req.method.name, supplement->method)) {
			if (supplement->incoming_request(session, rdata)) {
				break;
			}
		}
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, Bug 216100 */
	if (does_method_match(&rdata->msg_info.cseq->method.name, "CANCEL")) {
		/* Removed by yfyang, for Bug 241389 */
		#if 0
		set_cc_failed_recv_cancel(session); 
		#endif
		send_notify_for_miss_call(session);
		notify_nat_call(session);
	}
	else if (does_method_match(&rdata->msg_info.cseq->method.name, "UPDATE")) {
		if(ast_sip_session_trunk_conf_need_subscribe(session,rdata)){
			//tell the peer channel  sip terminal  this call have update to conf
			if (ast_queue_control_data(session->channel, AST_CONTROL_TRUNK_CALL_TO_CONF, session->trunk_conf_num, strlen(session->trunk_conf_num)+1)) {
				ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_TRUNK_CALL_TO_CONFAST_CONTROL_TRUNK_CALL_TO_CONFAST_CONTROL_TRUNK_CALL_TO_CONF frame failed for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}

			ast_indicate_data(session->channel, AST_CONTROL_TRUNK_CALL_TO_CONF,session->trunk_conf_num, strlen(session->trunk_conf_num)+1);//indicate the trunk channel
		}

		/*Start added by zyxia for 267640,2023.8.14*/
		if(ast_sip_session_trunk_record_need_forward(session,rdata)){
			if (ast_queue_control_data(session->channel, AST_CONTROL_UPDATE_RECORD_SDP_ATTR, NULL, 0)) {
				ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_UPDATE_RECORD_SDP_ATTR frame failed for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		}
		/*End added by zyxia for 267640,2023.8.14*/
	}

	if (HA_ROLE_ACTIVE == ast_ha_get_role()) {
		/* pickup service,two sides not write recovery db */
		if(does_method_match(&rdata->msg_info.cseq->method.name, "ACK"))
		{
			if( NULL != session->inv_session 
				&& PJSIP_INV_STATE_CONFIRMED == session->inv_session->state
				&& session->channel)
			{
				struct ast_channel *peer_user;
				struct ast_bridge *local_bridge = NULL;
				const char *tmp;
				tmp = pbx_builtin_getvar_helper((struct ast_channel *)session->channel, "PICKUP_CALLER");
				if(tmp && 0 == strcasecmp(tmp,"yes"))
				{
					ast_channel_del_track(session->channel);
				}
				else
				{
					local_bridge = ast_channel_internal_bridge(session->channel);
					if(local_bridge){
						peer_user = ast_bridge_peer(local_bridge, session->channel);
						if(peer_user)
						{
							tmp = pbx_builtin_getvar_helper((struct ast_channel *)peer_user, "PICKUP_CALLER");
							if(tmp && 0 == strcasecmp(tmp,"yes"))
							{
								ast_channel_del_track(session->channel);
							}
							ast_channel_unref(peer_user);
						}
					}
				}
			}
		}
	}

	/* Add by bfdu for terminate CEI when UCM send BYE, 2021.3.11 */
	handle_incoming_bye_request(session, &req.method.name, rdata);
#endif
}

static void handle_session_begin(struct ast_sip_session *session)
{
	struct ast_sip_session_supplement *iter;

	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
		if (iter->session_begin) {
			iter->session_begin(session);
		}
	}
}

static void handle_session_destroy(struct ast_sip_session *session)
{
	struct ast_sip_session_supplement *iter;

	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
		if (iter->session_destroy) {
			iter->session_destroy(session);
		}
	}
}

static void handle_session_end(struct ast_sip_session *session)
{
	struct ast_sip_session_supplement *iter;

	/* Session is dead.  Notify the supplements. */
	AST_LIST_TRAVERSE(&session->supplements, iter, next) {
		if (iter->session_end) {
			iter->session_end(session);
		}
	}
}

static void handle_incoming_response(struct ast_sip_session *session, pjsip_rx_data *rdata,
		enum ast_sip_session_response_priority response_priority)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_status_line status = rdata->msg_info.msg->line.status;

	ast_debug(3, "Response is %d %.*s\n", status.code, (int) pj_strlen(&status.reason),
			pj_strbuf(&status.reason));
#ifdef GRANDSTREAM_NETWORKS
	session->is_outgoing = 0;

	if(does_method_match(&rdata->msg_info.cseq->method.name, "INVITE")){

		if(status.code == 486){
			if(response_priority == AST_SIP_SESSION_BEFORE_MEDIA){
				/*start add by zyxia for wake up gswave,if receive 486 from gswave,Bug 150558 */
				if (!session->has_dec_gs_call_count) {
					ast_sip_session_end_if_deferred(session);
					ast_sip_session_defer_termination_cancel(session);
					ast_sip_set_used_gs_call_count(session,rdata,0);
				}
				/*end add by zyxia for wake up gswave,receive 486 from gswave,*/

				if (session->channel) {
					ast_channel_state_set(session->channel, AST_STATE_BUSY);//added by zyxia  for Bug208771,2021.11.25
				}
				ast_sip_report_invite_failed(session,rdata,NULL,"INVITE",status.code,ast_cause2str(hangup_sip2cause(status.code)));
			}

			/* check Warning header */
			ast_sip_warning_hdr_proc(session, rdata);

		} else if (status.code == 200) {
			if (session->is_gswave_app_call)
				session->gswave_app_session_active = 1; //added by zyxia for Bug 188207  check gswave  app if whehter receive invite  or not 
		
			if (session->line_index > 0) {
				ast_sip_set_sca_line_status((char *)session->endpoint->sca_sharedline, session->line_index, SCA_Active); 
				if (session->aor && !session->has_insert_sca_user) {
					ast_sip_sca_line_insert_user((char *)session->endpoint->sca_sharedline, session->line_index, (char *)ast_sorcery_object_get_id(session->aor)); 
					session->has_insert_sca_user = 1;

					/* update the caller number */
					if ((AST_SIP_SCA_ROLE_TYPE_PRIVATE == session->endpoint->sca_role) && 
						(session->channel && ast_channel_caller(session->channel)->id.number.valid)) {
						ast_free(ast_channel_caller(session->channel)->id.number.str);
						ast_channel_caller(session->channel)->id.number.str = ast_strdup((char *)ast_sorcery_object_get_id(session->aor));
					}
				}
			}
			if(session->channel && !ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)
				&& (NULL != rdata->tp_info.transport->type_name)
				&& !strcmp(rdata->tp_info.transport->type_name,"UDP"))
			{
				ast_channel_recovery_track(session->channel, rdata->msg_info.msg, NULL);
				//ast_sip_record_hdr_from_msg(session->channel, rdata->msg_info.msg, NULL);
			}
		} else if (status.code == 100) {
			if (session->line_index > 0) {
				if (session->aor && !session->has_insert_sca_user) {
					ast_sip_sca_line_insert_user((char *)session->endpoint->sca_sharedline, session->line_index, (char *)ast_sorcery_object_get_id(session->aor)); 
					session->has_insert_sca_user = 1;

					/* update the caller number */
					if ((AST_SIP_SCA_ROLE_TYPE_PRIVATE == session->endpoint->sca_role) && 
						(session->channel && ast_channel_caller(session->channel)->id.number.valid)) {
						ast_free(ast_channel_caller(session->channel)->id.number.str);
						ast_channel_caller(session->channel)->id.number.str = ast_strdup((char *)ast_sorcery_object_get_id(session->aor));
					}
				}
			}
		} else {
			if(session->channel != NULL 
				&& ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)
				&& ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ORIGINATED)
				&& status.code == 481)
			{
					/* release original caller */
					ast_log(LOG_ERROR,"481 received of reinvite, maybe user have hanguped  chan=%s", ast_channel_name(session->channel));
					ast_hangup(session->channel);
			}

			if ((status.code != 401) && (status.code != 407) && (status.code > 399)) {
				ast_sip_report_invite_failed(session,rdata,NULL,"INVITE",status.code,ast_cause2str(hangup_sip2cause(status.code)));
			}
		}

		/* Add by bfdu for Bug 268499, 2024.5.15 */
		notify_nat_call(session);
	}
	/*end*/

	if( (does_method_match(&rdata->msg_info.cseq->method.name, "UPDATE") || does_method_match(&rdata->msg_info.cseq->method.name, "INFO")) 
		&& status.code == 200 && NULL != session->inv_session 
		&& PJSIP_INV_STATE_CONFIRMED == session->inv_session->state
		&& NULL != session->channel
		&& ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_TRACKED)){
		/* update cseq for recovery record */
		ast_channel_recovery_track(session->channel, NULL, &(rdata->msg_info.cseq->cseq));
	}

	set_session_allow_events(session, rdata);
#endif

	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (!(supplement->response_priority & response_priority)) {
			continue;
		}
		if (supplement->incoming_response && does_method_match(&rdata->msg_info.cseq->method.name, supplement->method)) {
			supplement->incoming_response(session, rdata);
		}
	}
}

static int handle_incoming(struct ast_sip_session *session, pjsip_rx_data *rdata,
		enum ast_sip_session_response_priority response_priority)
{
	ast_debug(3, "Received %s\n", rdata->msg_info.msg->type == PJSIP_REQUEST_MSG ?
			"request" : "response");

#ifdef GRANDSTREAM_NETWORKS
	/*Start added by zyxia  for Bug 121182,Bug121199,after blind and attend transfer,the session channel is null, Bye Request will not
	 be deal ,so  deal with it at here*/
	if (NULL != session && ao2_ref(session,0) > 0) {
		if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) {
			struct pjsip_request_line req = rdata->msg_info.msg->line.req;
			if (!pjsip_method_cmp(&req.method, &pjsip_bye_method) || 
				!pjsip_method_cmp(&req.method, &pjsip_cancel_method)) {
				if(!session->has_dec_gs_call_count) {
					ast_sip_session_end_if_deferred(session);
					ast_sip_session_defer_termination_cancel(session);
					ast_sip_set_used_gs_call_count(session,rdata,0);
				}
			}
		}
	}
	/*End added by zyxia for Bug 121182,Bug121199,after blind and attend transfer,the session channel is null, Bye Request will not
		 be deal ,so  deal with it at here*/

	set_yealink_flag(rdata, session);
	set_terminal_type(session, rdata);
	set_support_whiteboard_flag(session, rdata);
#endif

	if (rdata->msg_info.msg->type == PJSIP_REQUEST_MSG) {
		handle_incoming_request(session, rdata);
	} else {
		handle_incoming_response(session, rdata, response_priority);
	}

	return 0;
}

static void handle_outgoing_request(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_request_line req = tdata->msg->line.req;

	ast_debug(3, "Method is %.*s\n", (int) pj_strlen(&req.method.name), pj_strbuf(&req.method.name));
	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (supplement->outgoing_request && does_method_match(&req.method.name, supplement->method)) {
			supplement->outgoing_request(session, tdata);
		}
	}
}

static void handle_outgoing_response(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	struct ast_sip_session_supplement *supplement;
	struct pjsip_status_line status = tdata->msg->line.status;
	pjsip_cseq_hdr *cseq = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_CSEQ, NULL);

	if (!cseq) {
		ast_log(LOG_ERROR, "Cannot send response due to missing sequence header");
		return;
	}

	ast_debug(3, "Method is %.*s, Response is %d %.*s\n", (int) pj_strlen(&cseq->method.name),
		pj_strbuf(&cseq->method.name), status.code, (int) pj_strlen(&status.reason),
		pj_strbuf(&status.reason));

#ifdef GRANDSTREAM_NETWORKS
		/* Add by bfdu for Bug 153236 , 2019.12.06 */
		if ((status.code == 200 || status.code == 183) && does_method_match(&cseq->method.name, "INVITE") &&
			session->endpoint->media.rtp.accept_multiple_sdp_answers) {
			ast_log(LOG_DEBUG, "[session:%p] When enable use final sdp, set nat_hook mod_data NULL to modify 200OK sdp addr for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_sip_mod_data_set(tdata->pool, tdata->mod_data, session_module.id, MOD_DATA_NAT_HOOK, NULL);
		}
#endif

	if(status.code == 200
		&& does_method_match(&cseq->method.name, "INVITE")
		&& !ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_RECOVERYING)
		&& tdata
		&& session->trans_type 
		&& !strcasecmp(session->trans_type, "UDP"))
	{
		ast_channel_recovery_track(session->channel,tdata->msg, NULL);
		//ast_sip_record_hdr_from_msg(session->channel, tdata->msg, NULL);
	}

	AST_LIST_TRAVERSE(&session->supplements, supplement, next) {
		if (supplement->outgoing_response && does_method_match(&cseq->method.name, supplement->method)) {
			supplement->outgoing_response(session, tdata);
		}
	}

}

static int session_end(void *vsession)
{
#ifdef GRANDSTREAM_NETWORKS
	RAII_VAR(struct ast_sip_session *, session, ao2_bump(vsession), ao2_cleanup);
#else
	struct ast_sip_session *session = vsession;
#endif


	/* Stop the scheduled termination */
	sip_session_defer_termination_stop_timer(session);

	/* Session is dead.  Notify the supplements. */
	handle_session_end(session);

#ifdef GRANDSTREAM_NETWORKS
	if (session->dump) {
		ao2_cleanup(session);
		session->dump = 0;
	}
#endif

	return 0;
}

/*Start added by zyxia for Bug185135  faxsending session end delay for wait faxsending result report,2021.1.22*/
#ifdef GRANDSTREAM_NETWORKS
static void* faxsending_session_end(void *vsession)
{
	RAII_VAR(struct ast_sip_session *, session, vsession, ao2_cleanup);
	ast_debug(1,"Session:%p Endpoint '%s':faxsending_session_end need delay wait for spandsp report\n",session,ast_sorcery_object_get_id(session->endpoint));
	usleep(900000);//add delay for Bug 248169.269031

	session->dump = 1;
	if (ast_sip_push_task(session->serializer, session_end, ao2_bump(session))) {
		/* Do it anyway even though this is not the right thread. */
		session_end(session);
		ao2_cleanup(session);
	}
	return NULL;
}
#endif

/*!
 * \internal
 * \brief Complete ending session activities.
 * \since 13.5.0
 *
 * \param vsession Which session to complete stopping.
 *
 * \retval 0 on success.
 * \retval -1 on error.
 */
static int session_end_completion(void *vsession)
{
	struct ast_sip_session *session = vsession;

	ast_sip_dialog_set_serializer(session->inv_session->dlg, NULL);
	ast_sip_dialog_set_endpoint(session->inv_session->dlg, NULL);

	/* Now we can release the ref that was held by session->inv_session */
	ao2_cleanup(session);
	return 0;
}

static int check_request_status(pjsip_inv_session *inv, pjsip_event *e)
{
	struct ast_sip_session *session = inv->mod_data[session_module.id];
	pjsip_transaction *tsx = e->body.tsx_state.tsx;

	if (tsx->status_code != 503 && tsx->status_code != 408) {
		return 0;
	}

	if (!ast_sip_failover_request(tsx->last_tx)) {
		return 0;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (tsx->method.id == PJSIP_INVITE_METHOD) {
		pjsip_inv_uac_restart(inv, PJ_FALSE);
	} else {
		ast_log(LOG_DEBUG, "Don't need restart uac inv, tsx id:%d\n", tsx->method.id);
	}
#else
	pjsip_inv_uac_restart(inv, PJ_FALSE);
#endif

#ifdef GRANDSTREAM_NETWORKS
	pjsip_tx_data *tdata = NULL;

	if (pjsip_tx_data_clone_req(tsx->last_tx, &tdata) != PJ_SUCCESS) {
		return 0;
	}

	ast_sip_session_send_request(session, tdata);
#else
	/*
	 * Bump the ref since it will be on a new transaction and
	 * we don't want it to go away along with the old transaction.
	 */
	pjsip_tx_data_add_ref(tsx->last_tx);
	ast_sip_session_send_request(session, tsx->last_tx);
#endif
	return 1;
}

static void handle_incoming_before_media(pjsip_inv_session *inv,
	struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_msg *msg;
#ifdef GRANDSTREAM_NETWORKS
	if(!rdata){
		return;
	}
#endif
	handle_incoming(session, rdata, AST_SIP_SESSION_BEFORE_MEDIA);
	msg = rdata->msg_info.msg;
	if (msg->type == PJSIP_REQUEST_MSG
		&& msg->line.req.method.id == PJSIP_ACK_METHOD
		&& pjmedia_sdp_neg_get_state(inv->neg) != PJMEDIA_SDP_NEG_STATE_DONE) {
		pjsip_tx_data *tdata;

		/*
		 * SDP negotiation failed on an incoming call that delayed
		 * negotiation and then gave us an invalid SDP answer.  We
		 * need to send a BYE to end the call because of the invalid
		 * SDP answer.
		 */
		ast_debug(1,
			"Endpoint '%s(%s)': Ending session due to incomplete SDP negotiation.  %s\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "",
			pjsip_rx_data_get_info(rdata));
		if (pjsip_inv_end_session(inv, 400, NULL, &tdata) == PJ_SUCCESS
			&& tdata) {
			ast_sip_session_send_request(session, tdata);
		}
	}
}

static void session_inv_on_state_changed(pjsip_inv_session *inv, pjsip_event *e)
{
	struct ast_sip_session *session;
	pjsip_event_id_e type;

	if (ast_shutdown_final()) {
		return;
	}

	if (e) {
		print_debug_details(inv, NULL, e);
		type = e->type;
	} else {
		type = PJSIP_EVENT_UNKNOWN;
	}

	session = inv->mod_data[session_module.id];
	if (!session) {
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "session_inv_on_state_changed session[%p], refcnt[%d], type[%d], inv->state[%d]", session, ao2_ref(session, 0), type, inv->state);
#endif

	switch(type) {
	case PJSIP_EVENT_TX_MSG:
		break;
	case PJSIP_EVENT_RX_MSG:
		handle_incoming_before_media(inv, session, e->body.rx_msg.rdata);
		break;
	case PJSIP_EVENT_TSX_STATE:
		ast_debug(3, "Source of transaction state change is %s\n", pjsip_event_str(e->body.tsx_state.type));
		/* Transaction state changes are prompted by some other underlying event. */
		switch(e->body.tsx_state.type) {
		case PJSIP_EVENT_TX_MSG:
			break;
		case PJSIP_EVENT_RX_MSG:
			if (!check_request_status(inv, e)) {
				handle_incoming_before_media(inv, session, e->body.tsx_state.src.rdata);
			}
			break;
		case PJSIP_EVENT_TRANSPORT_ERROR:
		case PJSIP_EVENT_TIMER:
			/*
			 * Check the request status on transport error or timeout. A transport
			 * error can occur when a TCP socket closes and that can be the result
			 * of a 503. Also we may need to failover on a timeout (408).
			 */
			check_request_status(inv, e);
			break;
		case PJSIP_EVENT_USER:
		case PJSIP_EVENT_UNKNOWN:
		case PJSIP_EVENT_TSX_STATE:
			/* Inception? */
			break;
		}
		break;
	case PJSIP_EVENT_TRANSPORT_ERROR:
	case PJSIP_EVENT_TIMER:
	case PJSIP_EVENT_UNKNOWN:
	case PJSIP_EVENT_USER:
	default:
		break;
	}

	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
		if (session->defer_end) {
			ast_debug(3, "Deferring session (%p) end\n", session);
			session->ended_while_deferred = 1;
			return;
		}

#ifdef GRANDSTREAM_NETWORKS
		/*Start added by zyxia for Bug185315  faxsending session end delay for wait faxsending result report,2021.1.22*/
		if (session->is_fax_sending_request) {
			pthread_t threadid;
			if (ast_pthread_create_detached(&threadid, NULL, faxsending_session_end, ao2_bump(session))) {
				session_end(session);
				ao2_cleanup(session);
			}
			return;
		}
#endif
		if (ast_sip_push_task(session->serializer, session_end, session)) {
			/* Do it anyway even though this is not the right thread. */
			session_end(session);
		}
	}
}

static void session_inv_on_new_session(pjsip_inv_session *inv, pjsip_event *e)
{
	/* XXX STUB */
}

static int session_end_if_disconnected(int id, pjsip_inv_session *inv)
{
	struct ast_sip_session *session;

	if (inv->state != PJSIP_INV_STATE_DISCONNECTED) {
		return 0;
	}

	/*
	 * We are locking because ast_sip_dialog_get_session() needs
	 * the dialog locked to get the session by other threads.
	 */
	pjsip_dlg_inc_lock(inv->dlg);
	session = inv->mod_data[id];
	inv->mod_data[id] = NULL;
	pjsip_dlg_dec_lock(inv->dlg);

	/*
	 * Pass the session ref held by session->inv_session to
	 * session_end_completion().
	 */
	if (session
		&& ast_sip_push_task(session->serializer, session_end_completion, session)) {
		/* Do it anyway even though this is not the right thread. */
		session_end_completion(session);
	}

	return 1;
}

#ifdef GRANDSTREAM_NETWORKS
/* Add by bfdu for send re-INVITE response when transport error, revert state, 2021.2.7 */
static void reset_reinvite_pending_state(struct ast_sip_session *session, pjsip_inv_session *inv, pjsip_transaction *tsx)
{
	int failed_code = 488;

	if (tsx->method.id != PJSIP_INVITE_METHOD) {
		return;
	}

	if (!inv->neg) {
		return;
	}

	if (pjmedia_sdp_neg_get_state(inv->neg) == PJMEDIA_SDP_NEG_STATE_WAIT_NEGO) {
		ast_log(LOG_DEBUG, "[session:%p] Send re-INVITE response meet transport error, reset pending state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		pjmedia_sdp_neg_cancel_wait_nego(inv->neg);
		ast_sip_session_media_state_reset(session->pending_media_state);
	}

	if (pjmedia_sdp_neg_get_state(inv->neg) == PJMEDIA_SDP_NEG_STATE_LOCAL_OFFER) {
		ast_log(LOG_DEBUG, "[session:%p] Send re-INVITE response meet transport error, neg state is local offer, reset pending state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		pjmedia_sdp_neg_cancel_offer(inv->neg);
		ast_sip_session_media_state_reset(session->pending_media_state);
		session->is_media_change = 0;
		if (ast_strlen_zero(session->conf_room)) {
			ast_log(LOG_DEBUG, "[session:%p] For P2P call other side may has pending re-INVITE, send failed code if need, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			if (session->channel) {
				ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_FAILED, &failed_code, sizeof(failed_code));
			}
		}
	}

	return;
}

static void reset_notify_pending_state(struct ast_sip_session *session, pjsip_inv_session *inv, pjsip_transaction *tsx)
{
	if (tsx->method.id != PJSIP_OTHER_METHOD) {
		return;
	}

	if (!tsx->last_tx || !tsx->last_tx->msg || !tsx->last_tx->msg->body) {
		return;
	}

	if (pj_strcmp2(&tsx->last_tx->msg->body->content_type.type, "application") ||
		pj_strcmp2(&tsx->last_tx->msg->body->content_type.subtype, "participant-sdp")) {
		ast_log(LOG_DEBUG, "[session:%p] reset_notify_pending_state this notify is not refresh participant notify, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->pending_participant_media_state && session->pending_participant_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Reset notify pending state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_participant_media_state);
		session->sending_notify_count = 0;
		session->notify_full_participant_sdp = 1;
		session->sending_participant_notify = 0;
	}

	return;
}

static void reset_pending_state(struct ast_sip_session *session, pjsip_inv_session *inv, pjsip_transaction *tsx)
{
	if (!session || !inv || !tsx) {
		return;
	}

	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
		return;
	}

	reset_reinvite_pending_state(session, inv, tsx);

	reset_notify_pending_state(session, inv, tsx);

	return;
}
#endif

#ifdef GRANDSTREAM_NETWORKS
static int get_useragent_for_device(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	const pj_str_t user_agent_name = { "User-Agent", 10 };
	pjsip_generic_string_hdr *useragent_hdr = NULL;

	if (!session || !session->channel || !rdata) {
		ast_log(LOG_WARNING, "RX packet is NULL\n");
		return -1;
	}

	useragent_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &user_agent_name, NULL);
	if (useragent_hdr) {
		pj_str_t value;
		pj_strdup_with_null(rdata->tp_info.pool, &value, &useragent_hdr->hvalue);

		ast_log(LOG_DEBUG, "User-Agent: %s, len: %ld", pj_strbuf(&value), pj_strlen(&value));
		if (!ast_strlen_zero(pj_strbuf(&value))) {
			ast_channel_device_info_set(session->channel, pj_strbuf(&value));
		}
	}

	return 0;
}

static int get_server_for_device(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	const pj_str_t server_name = { "Server", 6 };
	pjsip_generic_string_hdr *server_hdr = NULL;

	if (!session || !session->channel || !rdata) {
		ast_log(LOG_WARNING, "RX packet is NULL\n");
		return -1;
	}

	server_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &server_name, NULL);
	if (server_hdr) {
		pj_str_t value;
		pj_strdup_with_null(rdata->tp_info.pool, &value, &server_hdr->hvalue);

		ast_log(LOG_DEBUG, "Server: %s, len: %ld", pj_strbuf(&value), pj_strlen(&value));
		if (!ast_strlen_zero(pj_strbuf(&value))) {
			ast_channel_device_info_set(session->channel, pj_strbuf(&value));
		}
	}

	return 0;
}
/* end */

static void check_delayed_trickle_ice_info(struct ast_sip_session *session)
{
	if (!session || !session->inv_session) {
		return;
	}

	if (!session->trickle_info_delayed) {
		return;
	}

	if (session->inv_session->state < PJSIP_INV_STATE_EARLY) {
		ast_log(LOG_DEBUG, "[session:%p] Dialog is not established really, don't send sip info for trickle ice, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));	
		return;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_DEBUG, "[session:%p] State is disconnected, don't send sip info for trickle ice, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));	
		return ;
	}

	ast_log(LOG_DEBUG, "[session:%p] Send delayed sip info for trickle ice, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));	

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, ast_sip_session_send_info_for_add_candidates, session)) {
		ao2_cleanup(session);
	}

	session->trickle_info_delayed = 0;

	return;
}
#endif

static void session_inv_on_tsx_state_changed(pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_event *e)
{
	ast_sip_session_response_cb cb;
	int id = session_module.id;
	struct ast_sip_session *session;
	pjsip_tx_data *tdata;

	if (ast_shutdown_final()) {
		return;
	}

	session = inv->mod_data[id];

	print_debug_details(inv, tsx, e);
	if (!session) {
		/* The session has ended.  Ignore the transaction change. */
		return;
	}
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_DEBUG, "session_inv_on_tsx_state_changed session[%p], refcnt[%d]", session, ao2_ref(session, 0));
#endif

	/*
	 * If the session is disconnected really nothing else to do unless currently transacting
	 * a BYE. If a BYE then hold off destruction until the transaction timeout occurs. This
	 * has to be done for BYEs because sometimes the dialog can be in a disconnected
	 * state but the BYE request transaction has not yet completed.
	 */
	if (tsx->method.id != PJSIP_BYE_METHOD && session_end_if_disconnected(id, inv)) {
		return;
	}

	switch (e->body.tsx_state.type) {
	case PJSIP_EVENT_TX_MSG:
		/* When we create an outgoing request, we do not have access to the transaction that
		 * is created. Instead, We have to place transaction-specific data in the tdata. Here,
		 * we transfer the data into the transaction. This way, when we receive a response, we
		 * can dig this data out again
		 */
		tsx->mod_data[id] = e->body.tsx_state.src.tdata->mod_data[id];
#ifdef GRANDSTREAM_NETWORKS
		if (tsx->status_code == PJSIP_SC_NOT_ACCEPTABLE_HERE && tsx->method.id == PJSIP_INVITE_METHOD && inv->state == PJSIP_INV_STATE_DISCONNECTED) {
			char peer_name[64] = {0};
			ast_copy_string(peer_name, ast_sorcery_object_get_id(session->endpoint), sizeof(peer_name));
			if (0 != strncmp(peer_name, "trunk_", strlen("trunk_")))
			{
				if(e->body.tsx_state.src.tdata->msg->type == PJSIP_RESPONSE_MSG && e->body.tsx_state.src.tdata->msg->line.status.code == 488)	//add by yuyang for Bug 129275
				{
					//When the message type sent is a response and the response code is 488, the alert event is not reported
				}
				else
				{
					ast_sip_report_invite_failed(session,NULL,e->body.tsx_state.src.tdata,"INVITE",tsx->status_code,ast_cause2str(hangup_sip2cause(tsx->status_code)));
				}
			}
		}
#endif

		break;
	case PJSIP_EVENT_RX_MSG:
		cb = ast_sip_mod_data_get(tsx->mod_data, id, MOD_DATA_ON_RESPONSE);
		/* As the PJSIP invite session implementation responds with a 200 OK before we have a
		 * chance to be invoked session supplements for BYE requests actually end up executing
		 * in the invite session state callback as well. To prevent session supplements from
		 * running on the BYE request again we explicitly squash invocation of them here.
		 */
		if ((e->body.tsx_state.src.rdata->msg_info.msg->type != PJSIP_REQUEST_MSG) ||
			(tsx->method.id != PJSIP_BYE_METHOD)) {
			handle_incoming(session, e->body.tsx_state.src.rdata,
				AST_SIP_SESSION_AFTER_MEDIA);
		}
		if (tsx->method.id == PJSIP_INVITE_METHOD) {
			if (tsx->role == PJSIP_ROLE_UAC) {
				if (tsx->state == PJSIP_TSX_STATE_COMPLETED) {
					/* This means we got a non 2XX final response to our outgoing INVITE */
					if (tsx->status_code == PJSIP_SC_REQUEST_PENDING) {
						reschedule_reinvite(session, cb);
						return;
					}
					if (inv->state == PJSIP_INV_STATE_CONFIRMED) {
						ast_debug(1, "reINVITE received final response code %d\n",
							tsx->status_code);
						if ((tsx->status_code == 401 || tsx->status_code == 407)
							&& !ast_sip_create_request_with_auth(
								&session->endpoint->outbound_auths,
								e->body.tsx_state.src.rdata, tsx->last_tx, &tdata)) {
							/* Send authed reINVITE */
							ast_sip_session_send_request_with_cb(session, tdata, cb);
							return;
						}

						/*! Started added by zyxia for Bug 206995, t38 reinvte return 415,2021.10.28*/
#ifdef GRANDSTREAM_NETWORKS
						if (tsx->status_code != 488 && tsx->status_code != 500
							 && (session->wImageReInvite && 415 != tsx->status_code))
#else
						/*! End added by zyxia for Bug 206995, t38 reinvte return 415,2021.10.28*/

						if (tsx->status_code != 488 && tsx->status_code != 500)
#endif
						{
							/* Other reinvite failures (except 488 and 500) result in destroying the session. */
							if (pjsip_inv_end_session(inv, 500, NULL, &tdata) == PJ_SUCCESS
								&& tdata) {
								ast_sip_session_send_request(session, tdata);
							}
						}
					}
				} else if (tsx->state == PJSIP_TSX_STATE_TERMINATED) {
					if (inv->cancelling && tsx->status_code == PJSIP_SC_OK) {
						int sdp_negotiation_done =
							pjmedia_sdp_neg_get_state(inv->neg) == PJMEDIA_SDP_NEG_STATE_DONE;

						/*
						 * We can get here for the following reasons.
						 *
						 * 1) The race condition detailed in RFC5407 section 3.1.2.
						 * We sent a CANCEL at the same time that the UAS sent us a
						 * 200 OK with a valid SDP for the original INVITE.  As a
						 * result, we have now received a 200 OK for a cancelled
						 * call and the SDP negotiation is complete.  We need to
						 * immediately send a BYE to end the dialog.
						 *
						 * 2) We sent a CANCEL and hit the race condition but the
						 * UAS sent us an invalid SDP with the 200 OK.  In this case
						 * the SDP negotiation is incomplete and PJPROJECT has
						 * already sent the BYE for us because of the invalid SDP.
						 *
						 * 3) We didn't send a CANCEL but the UAS sent us an invalid
						 * SDP with the 200 OK.  In this case the SDP negotiation is
						 * incomplete and PJPROJECT has already sent the BYE for us
						 * because of the invalid SDP.
						 */
						ast_test_suite_event_notify("PJSIP_SESSION_CANCELED",
							"Endpoint: %s\r\n"
							"Channel: %s\r\n"
							"Message: %s\r\n"
							"SDP: %s",
							ast_sorcery_object_get_id(session->endpoint),
							session->channel ? ast_channel_name(session->channel) : "",
							pjsip_rx_data_get_info(e->body.tsx_state.src.rdata),
							sdp_negotiation_done ? "complete" : "incomplete");
						if (!sdp_negotiation_done) {
							ast_debug(1, "Endpoint '%s(%s)': Incomplete SDP negotiation cancelled session.  %s\n",
								ast_sorcery_object_get_id(session->endpoint),
								session->channel ? ast_channel_name(session->channel) : "",
								pjsip_rx_data_get_info(e->body.tsx_state.src.rdata));
						} else if (pjsip_inv_end_session(inv, 500, NULL, &tdata) == PJ_SUCCESS
							&& tdata) {
							ast_debug(1, "Endpoint '%s(%s)': Ending session due to RFC5407 race condition.  %s\n",
								ast_sorcery_object_get_id(session->endpoint),
								session->channel ? ast_channel_name(session->channel) : "",
								pjsip_rx_data_get_info(e->body.tsx_state.src.rdata));
							ast_sip_session_send_request(session, tdata);
						}
					}
				}
			}
		} else {
			/* All other methods */
			if (tsx->role == PJSIP_ROLE_UAC) {
				if (tsx->state == PJSIP_TSX_STATE_COMPLETED) {
					/* This means we got a final response to our outgoing method */
					ast_debug(1, "%.*s received final response code %d\n",
						(int) pj_strlen(&tsx->method.name), pj_strbuf(&tsx->method.name),
						tsx->status_code);
					if ((tsx->status_code == 401 || tsx->status_code == 407)
						&& !ast_sip_create_request_with_auth(
							&session->endpoint->outbound_auths,
							e->body.tsx_state.src.rdata, tsx->last_tx, &tdata)) {
						/* Send authed version of the method */
						ast_sip_session_send_request_with_cb(session, tdata, cb);
						return;
					}
				}
			}
		}
		if (cb) {
			cb(session, e->body.tsx_state.src.rdata);
		}
#ifdef GRANDSTREAM_NETWORKS
		/*! Begin: Added by pxie@grandstream.cn for bug[43844], 2015-06-19.*/
		ast_log(LOG_VERBOSE, "wImageReInvite:%d,wSwitchImageInvite488ToAudioInvite:%d!",session->wImageReInvite,session->wSwitchImageInvite488ToAudioInvite);
		if (session->wImageReInvite && ((tsx->status_code == 488) || (415 == tsx->status_code)) && (tsx->state == PJSIP_TSX_STATE_COMPLETED)
			&& (session->wSwitchImageInvite488ToAudioInvite) && session->channel) {
			ast_log(LOG_VERBOSE, "Send Re-Invite with audio sdp, when T.38 negotiate failed!");
			struct ast_sip_session_media_state *active_media_state_clone;

			active_media_state_clone = ast_sip_session_media_state_clone(session->active_media_state);
			if (!active_media_state_clone) {
				ast_log(LOG_WARNING, "Unable to clone active media state for channel '%s'\n",
					session->channel ? ast_channel_name(session->channel) : "unknown");
				return ;
			}
			ast_sip_session_refresh(session, NULL, NULL, NULL, AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1,active_media_state_clone);
			active_media_state_clone = NULL;
		}
		/*! End: Added by pxie@grandstream.cn	for bug[43844], 2015-06-19.*/
		/*set device data before answer*/
		if (session->channel && strcasestr(ast_channel_name(session->channel), "trunk_")){
			get_server_for_device(session, e->body.tsx_state.src.rdata);
		}else{
			get_useragent_for_device(session, e->body.tsx_state.src.rdata);
		}
#endif

		break;
	case PJSIP_EVENT_TRANSPORT_ERROR:
#ifdef GRANDSTREAM_NETWORKS
		reset_pending_state(session, inv, tsx);
#endif
	case PJSIP_EVENT_TIMER:
#ifdef GRANDSTREAM_NETWORKS
	/* Modify by bfdu, move PJSIP_EVENT_USER to here for SIPP response non-2XX's ACK branch ID is not the same */
	case PJSIP_EVENT_USER:
#endif
		/*
		 * The timer event is run by the pjsip monitor thread and not
		 * by the session serializer.
		 */
		if (session_end_if_disconnected(id, inv)) {
			return;
		}
		break;
#ifndef GRANDSTREAM_NETWORKS
	case PJSIP_EVENT_USER:
#endif
	case PJSIP_EVENT_UNKNOWN:
	case PJSIP_EVENT_TSX_STATE:
		/* Inception? */
		break;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (tsx->method.id == PJSIP_INVITE_METHOD) {
		check_delayed_trickle_ice_info(session);
	}
#endif

	if (AST_LIST_EMPTY(&session->delayed_requests)) {
		/* No delayed request pending, so just return */
		return;
	}

	if (tsx->method.id == PJSIP_INVITE_METHOD) {
		if (tsx->state == PJSIP_TSX_STATE_PROCEEDING) {
			ast_debug(3, "Endpoint '%s(%s)' INVITE delay check. tsx-state:%s\n",
				ast_sorcery_object_get_id(session->endpoint),
				session->channel ? ast_channel_name(session->channel) : "",
				pjsip_tsx_state_str(tsx->state));
			check_delayed_requests(session, invite_proceeding);
		} else if (tsx->state == PJSIP_TSX_STATE_TERMINATED) {
			/*
			 * Terminated INVITE transactions always should result in
			 * queuing delayed requests, no matter what event caused
			 * the transaction to terminate.
			 */
			ast_debug(3, "Endpoint '%s(%s)' INVITE delay check. tsx-state:%s\n",
				ast_sorcery_object_get_id(session->endpoint),
				session->channel ? ast_channel_name(session->channel) : "",
				pjsip_tsx_state_str(tsx->state));
			check_delayed_requests(session, invite_terminated);
		}
	} else if (tsx->role == PJSIP_ROLE_UAC
		&& tsx->state == PJSIP_TSX_STATE_COMPLETED
		&& !pj_strcmp2(&tsx->method.name, "UPDATE")) {
		ast_debug(3, "Endpoint '%s(%s)' UPDATE delay check. tsx-state:%s\n",
			ast_sorcery_object_get_id(session->endpoint),
			session->channel ? ast_channel_name(session->channel) : "",
			pjsip_tsx_state_str(tsx->state));
		check_delayed_requests(session, update_completed);
	}
}

static int add_sdp_streams(struct ast_sip_session_media *session_media,
	struct ast_sip_session *session, pjmedia_sdp_session *answer,
	const struct pjmedia_sdp_session *remote,
	struct ast_stream *stream)
{
	struct ast_sip_session_sdp_handler *handler = session_media->handler;
	RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);
	int res;
#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "add_sdp_streams(), handler[%p]\n,switchtoimage:%d,handler>id:%s", handler,session->wSwitchImageInvite488ToAudioInvite,handler ? handler->id:"");
#endif

	if (handler) {
		/* if an already assigned handler reports a catastrophic error, fail */
		res = handler->create_outgoing_sdp_stream(session, session_media, answer, remote, stream);
		if (res < 0) {
			return -1;
		}

		return 0;
	}

	handler_list = ao2_find(sdp_handlers, ast_codec_media_type2str(session_media->type), OBJ_KEY);
	if (!handler_list) {
		return 0;
	}

	/* no handler for this stream type and we have a list to search */
	AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
		if (handler == session_media->handler) {
			continue;
		}
		res = handler->create_outgoing_sdp_stream(session, session_media, answer, remote, stream);
		if (res < 0) {
			/* catastrophic error */
			return -1;
		}
		if (res > 0) {
			/* Handled by this handler. Move to the next stream */
			session_media_set_handler(session_media, handler);
			return 0;
		}
	}

	/* streams that weren't handled won't be included in generated outbound SDP */
	return 0;
}

/*! \brief Bundle group building structure */
struct sip_session_media_bundle_group {
	/*! \brief The media identifiers in this bundle group */
	char *mids[PJMEDIA_MAX_SDP_MEDIA];
	/*! \brief SDP attribute string */
	struct ast_str *attr_string;
};

static int add_bundle_groups(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_session *answer)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	struct sip_session_media_bundle_group bundle_groups[PJMEDIA_MAX_SDP_MEDIA];
	int index, mid_id;
	struct sip_session_media_bundle_group *bundle_group;

#ifdef GRANDSTREAM_NETWORKS
	if (session->endpoint->media.webrtc || ast_sip_session_media_use_bundle(session))
#else
	if (session->endpoint->media.webrtc)
#endif
	{
		attr = pjmedia_sdp_attr_create(pool, "msid-semantic", pj_cstr(&stmp, "WMS *"));
		pjmedia_sdp_attr_add(&answer->attr_count, answer->attr, attr);
	}

#ifdef GRANDSTREAM_NETWORKS
	if (!ast_sip_session_media_use_bundle(session))
#else
	if (!session->endpoint->media.bundle)
#endif
	{
		return 0;
	}

	memset(bundle_groups, 0, sizeof(bundle_groups));

	/* Build the bundle group layout so we can then add it to the SDP */
	for (index = 0; index < AST_VECTOR_SIZE(&session->pending_media_state->sessions); ++index) {
		struct ast_sip_session_media *session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);

		/* If this stream is not part of a bundle group we can't add it */
		if (session_media->bundle_group == -1) {
			continue;
		}

		bundle_group = &bundle_groups[session_media->bundle_group];

		/* If this is the first mid then we need to allocate the attribute string and place BUNDLE in front */
		if (!bundle_group->mids[0]) {
			bundle_group->mids[0] = session_media->mid;
			bundle_group->attr_string = ast_str_create(64);
			if (!bundle_group->attr_string) {
				continue;
			}

			ast_str_set(&bundle_group->attr_string, 0, "BUNDLE %s", session_media->mid);
			continue;
		}

		for (mid_id = 1; mid_id < PJMEDIA_MAX_SDP_MEDIA; ++mid_id) {
			if (!bundle_group->mids[mid_id]) {
				bundle_group->mids[mid_id] = session_media->mid;
				ast_str_append(&bundle_group->attr_string, 0, " %s", session_media->mid);
				break;
			} else if (!strcmp(bundle_group->mids[mid_id], session_media->mid)) {
				break;
			}
		}
	}

	/* Add all bundle groups that have mids to the SDP */
	for (index = 0; index < PJMEDIA_MAX_SDP_MEDIA; ++index) {
		bundle_group = &bundle_groups[index];

		if (!bundle_group->attr_string) {
			continue;
		}

		attr = pjmedia_sdp_attr_create(pool, "group", pj_cstr(&stmp, ast_str_buffer(bundle_group->attr_string)));
		pjmedia_sdp_attr_add(&answer->attr_count, answer->attr, attr);

		ast_free(bundle_group->attr_string);
	}

	return 0;
}

static struct pjmedia_sdp_session *create_local_sdp(pjsip_inv_session *inv, struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3 };
	static const pj_str_t STR_IP6 = { "IP6", 3 };
	pjmedia_sdp_session *local;
	int i;
	int stream;

#ifdef GRANDSTREAM_NETWORKS
	pjsip_dlg_inc_lock(inv->dlg);
#endif
	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "Failed to create session SDP. Session has been already disconnected\n");
#ifdef GRANDSTREAM_NETWORKS
		pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif
		return NULL;
	}

	if (!inv->pool_prov || !(local = PJ_POOL_ZALLOC_T(inv->pool_prov, pjmedia_sdp_session))) {
#ifdef GRANDSTREAM_NETWORKS
		pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif
		return NULL;
	}

	if (!offer) {
		local->origin.version = local->origin.id = (pj_uint32_t)(ast_random());
	} else {
		local->origin.version = offer->origin.version + 1;
		local->origin.id = offer->origin.id;
	}

	pj_strdup2(inv->pool_prov, &local->origin.user, session->endpoint->media.sdpowner);
	pj_strdup2(inv->pool_prov, &local->name, session->endpoint->media.sdpsession);

#ifdef GRANDSTREAM_NETWORKS
	/* Remove CT by bfdu, for SFU, we have AS/TIAS for bandw control */
	//add_bandw_ct_to_sdp(inv, session, offer, local);
#endif

	if (!session->pending_media_state->topology || !ast_stream_topology_get_count(session->pending_media_state->topology)) {
		/* We've encountered a situation where we have been told to create a local SDP but noone has given us any indication
		 * of what kind of stream topology they would like. We try to not alter the current state of the SDP negotiation
		 * by using what is currently negotiated. If this is unavailable we fall back to what is configured on the endpoint.
		 */
		ast_stream_topology_free(session->pending_media_state->topology);
		if (session->active_media_state->topology) {
			session->pending_media_state->topology = ast_stream_topology_clone(session->active_media_state->topology);
		} else {
			session->pending_media_state->topology = ast_stream_topology_clone(session->endpoint->media.topology);
		}
		if (!session->pending_media_state->topology) {
#ifdef GRANDSTREAM_NETWORKS
			pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif
			return NULL;
		}
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); ++i) {
		struct ast_sip_session_media *session_media;
		struct ast_stream *stream;
		unsigned int streams = local->media_count;

		/* This code does not enforce any maximum stream count limitations as that is done on either
		 * the handling of an incoming SDP offer or on the handling of a session refresh.
		 */

		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);

#ifdef GRANDSTREAM_NETWORKS
		session_media = ast_sip_session_media_state_add_as_sub_type_and_msid(session, session->pending_media_state, stream, i);
#else
		session_media = ast_sip_session_media_state_add(session, session->pending_media_state, ast_stream_get_type(stream), i);
#endif
		if (!session_media) {
#ifdef GRANDSTREAM_NETWORKS
			pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif
			return NULL;
		}

#ifdef GRANDSTREAM_NETWORKS
		change_participant_stream_params_for_sfu(session, session_media, stream);
		change_rtp_instance_local_ssrc(session, session_media, stream);
		change_stream_msid_according_session_media(session, session_media, stream);
		change_bundle_state(session, session_media, stream);
#endif

		if (add_sdp_streams(session_media, session, local, offer, stream)) {
#ifdef GRANDSTREAM_NETWORKS
			pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif
			return NULL;
		}

		/* If a stream was actually added then add any additional details */
		if (streams != local->media_count) {
			pjmedia_sdp_media *media = local->media[streams];
			pj_str_t stmp;
			pjmedia_sdp_attr *attr;

			/* Add the media identifier if present */
			if (!ast_strlen_zero(session_media->mid)) {
#ifdef GRANDSTREAM_NETWORKS
				if (session->endpoint->media.support_special_attrs) {
					attr = pjmedia_sdp_media_find_attr2(media, "mid", NULL);
					if (!attr) {
						attr = pjmedia_sdp_attr_create(inv->pool_prov, "mid", pj_cstr(&stmp, session_media->mid));
						pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
					}
				}
#else
				attr = pjmedia_sdp_attr_create(inv->pool_prov, "mid", pj_cstr(&stmp, session_media->mid));
				pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
#endif
			}
		}

		/* Ensure that we never exceed the maximum number of streams PJMEDIA will allow. */
		if (local->media_count == PJMEDIA_MAX_SDP_MEDIA) {
			break;
		}
	}

	/* Add any bundle groups that are present on the media state */
	if (add_bundle_groups(session, inv->pool_prov, local)) {
#ifdef GRANDSTREAM_NETWORKS
		pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif
		return NULL;
	}

	/* Use the connection details of an available media if possible for SDP level */
	for (stream = 0; stream < local->media_count; stream++) {
		if (!local->media[stream]->conn) {
			continue;
		}

		if (local->conn) {
			if (!pj_strcmp(&local->conn->net_type, &local->media[stream]->conn->net_type) &&
				!pj_strcmp(&local->conn->addr_type, &local->media[stream]->conn->addr_type) &&
				!pj_strcmp(&local->conn->addr, &local->media[stream]->conn->addr)) {
				local->media[stream]->conn = NULL;
			}
			continue;
		}

		/* This stream's connection info will serve as the connection details for SDP level */
		local->conn = local->media[stream]->conn;
		local->media[stream]->conn = NULL;

		continue;
	}

	/* If no SDP level connection details are present then create some */
	if (!local->conn) {
		local->conn = pj_pool_zalloc(inv->pool_prov, sizeof(struct pjmedia_sdp_conn));
		local->conn->net_type = STR_IN;
		local->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;

		if (!ast_strlen_zero(session->endpoint->media.address)) {
			pj_strdup2(inv->pool_prov, &local->conn->addr, session->endpoint->media.address);
		} else {
			pj_strdup2(inv->pool_prov, &local->conn->addr, ast_sip_get_host_ip_string(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET()));
		}
	}

	pj_strassign(&local->origin.net_type, &local->conn->net_type);
	pj_strassign(&local->origin.addr_type, &local->conn->addr_type);
	pj_strassign(&local->origin.addr, &local->conn->addr);

#ifdef GRANDSTREAM_NETWORKS
	move_dtls_and_ice_attrs_to_session_level(inv->pool_prov, local, session);
	pjsip_dlg_dec_lock(inv->dlg); // Added by pxie@grandstream.cn for bug[209790], 2021-12-13.
#endif

	return local;
}

static void session_inv_on_rx_offer(pjsip_inv_session *inv, const pjmedia_sdp_session *offer)
{
	struct ast_sip_session *session;
	pjmedia_sdp_session *answer;

	if (ast_shutdown_final()) {
		return;
	}

	session = inv->mod_data[session_module.id];

#ifdef GRANDSTREAM_NETWORKS
	ast_log(LOG_VERBOSE, "session_inv_on_rx_offer(),enter ...\n");
	init_session_param(session);
#endif

	if (session->joint_ipvt) {
		pjsip_inv_set_sdp_answer(inv, pjmedia_sdp_session_clone(session->inv_session->pool_prov, offer));
		return;
	}

	if (handle_incoming_sdp(session, offer)) {
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_VERBOSE, "[session:%p] handle_incoming_sdp failed, reset pending_media_state for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_media_state);
#endif
		return;
	}


	if ((answer = create_local_sdp(inv, session, offer))) {
	/*Start added by zyxia Bug185970,2021.1.22*/
#ifdef GRANDSTREAM_NETWORKS
		ast_log(LOG_VERBOSE, "'%s' two media(audo and image), because '%s'!\n",
		session->is_image_audio_two_media ? "Create" : "Not Create",
		session->two_media_is_support_image ? "support" : "not support");
		session_search_audio_port_with_image_audio(session, session->is_image_audio_two_media,
		session->two_media_is_support_image, answer, "audio");
#endif
	/*End added by zyxia Bug185970,2021.1.22*/
		pjsip_inv_set_sdp_answer(inv, answer);
	}
}

#if 0
static void session_inv_on_create_offer(pjsip_inv_session *inv, pjmedia_sdp_session **p_offer)
{
	/* XXX STUB */
}
#endif

static void session_inv_on_media_update(pjsip_inv_session *inv, pj_status_t status)
{
	struct ast_sip_session *session;
	const pjmedia_sdp_session *local, *remote;

	if (ast_shutdown_final()) {
		return;
	}

	session = inv->mod_data[session_module.id];
	if (!session || !session->channel) {
		/*
		 * If we don't have a session or channel then we really
		 * don't care about media updates.
		 * Just ignore
		 */
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	if (session->joint_ipvt) {
		ast_log(LOG_NOTICE, "[IPVT] session_inv_on_media_update it's a ipvt invite, don't need media update.\n");
		return;
	}

	if (session->is_direct_media) {
		ast_log(LOG_DEBUG, "[session:%p] SDP is negotiated as direct media for endpoint:%s, don't need media update\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}
#endif

	if (session->endpoint) {
		int bail = 0;

		/*
		 * If following_fork is set, then this is probably the result of a
		 * forked INVITE and SDP asnwers coming from the different fork UAS
		 * destinations.  In this case updated_sdp_answer will also be set.
		 *
		 * If only updated_sdp_answer is set, then this is the non-forking
		 * scenario where the same UAS just needs to change something like
		 * the media port.
		 */

		if (inv->following_fork) {
			if (session->endpoint->media.rtp.follow_early_media_fork) {
				ast_debug(3, "Following early media fork with different To tags\n");
			} else {
				ast_debug(3, "Not following early media fork with different To tags\n");
				bail = 1;
			}
		}
#ifdef HAVE_PJSIP_INV_ACCEPT_MULTIPLE_SDP_ANSWERS
		else if (inv->updated_sdp_answer) {
			if (session->endpoint->media.rtp.accept_multiple_sdp_answers) {
				ast_debug(3, "Accepting updated SDP with same To tag\n");
			} else {
				ast_debug(3, "Ignoring updated SDP answer with same To tag\n");
				bail = 1;
			}
		}
#endif
		if (bail) {
			return;
		}
	}

	if ((status != PJ_SUCCESS) || (pjmedia_sdp_neg_get_active_local(inv->neg, &local) != PJ_SUCCESS) ||
		(pjmedia_sdp_neg_get_active_remote(inv->neg, &remote) != PJ_SUCCESS)) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_BEARERCAPABILITY_NOTAVAIL);
		ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
		ast_queue_hangup(session->channel);
		return;
	}

#ifdef GRANDSTREAM_NETWORKS
	/*! Begin: Added by zyxia for Bug185737, 2021-1-15.*/
	if (session->wSwitchImageInvite488ToAudioInvite) {
		RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);
		if ((caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
			ast_format_cap_append(caps, ast_format_ulaw, 0);
			ast_format_cap_append(caps, ast_format_alaw, 0);
			ast_channel_nativeformats_set(session->channel, caps);
			ast_log(LOG_DEBUG, "Set channel '%s' nativeformats to 'ulaw and alaw'!\n", ast_channel_name(session->channel));
		}
	}
	/*! End: Added by zyxia for Bug185737, 2021-1-15.*/

	/* In this, if the negotiated sdp failed, send 488 to caller and BYE to callee */
	if (handle_negotiated_sdp(session, local, remote)) {
		ast_channel_hangupcause_set(session->channel, AST_CAUSE_BEARERCAPABILITY_NOTAVAIL);
		ast_set_hangupsource(session->channel, ast_channel_name(session->channel), 0);
		ast_queue_hangup(session->channel);
	}
#else
	handle_negotiated_sdp(session, local, remote);
#endif
}

static pjsip_redirect_op session_inv_on_redirected(pjsip_inv_session *inv, const pjsip_uri *target, const pjsip_event *e)
{
	struct ast_sip_session *session;
	const pjsip_sip_uri *uri;

	if (ast_shutdown_final()) {
		return PJSIP_REDIRECT_STOP;
	}

	session = inv->mod_data[session_module.id];
	if (!session || !session->channel) {
		return PJSIP_REDIRECT_STOP;
	}

#ifdef GRANDSTREAM_NETWORKS
	/* Added by yfyang, for bug106653 */
	if (session->line_index > 0) {
		if (ast_sip_sca_line_check_in_bridge((char *)session->endpoint->sca_sharedline, session->line_index)) {
			ast_log(LOG_NOTICE, "SIP SCA session on redirected, Stop redirect!");
			return PJSIP_REDIRECT_STOP;
		}
	}

#endif

	if (session->endpoint->redirect_method == AST_SIP_REDIRECT_URI_PJSIP) {
		return PJSIP_REDIRECT_ACCEPT;
	}

	if (!PJSIP_URI_SCHEME_IS_SIP(target) && !PJSIP_URI_SCHEME_IS_SIPS(target)) {
		return PJSIP_REDIRECT_STOP;
	}

	handle_incoming(session, e->body.rx_msg.rdata, AST_SIP_SESSION_BEFORE_REDIRECTING);

	uri = pjsip_uri_get_uri(target);

	if (session->endpoint->redirect_method == AST_SIP_REDIRECT_USER) {
		char exten[AST_MAX_EXTENSION];

		ast_copy_pj_str(exten, &uri->user, sizeof(exten));

		/*
		 * We may want to match in the dialplan without any user
		 * options getting in the way.
		 */
		AST_SIP_USER_OPTIONS_TRUNCATE_CHECK(exten);

		ast_channel_call_forward_set(session->channel, exten);
	} else if (session->endpoint->redirect_method == AST_SIP_REDIRECT_URI_CORE) {
		char target_uri[PJSIP_MAX_URL_SIZE];
		/* PJSIP/ + endpoint length + / + max URL size */
		char forward[8 + strlen(ast_sorcery_object_get_id(session->endpoint)) + PJSIP_MAX_URL_SIZE];

		pjsip_uri_print(PJSIP_URI_IN_REQ_URI, uri, target_uri, sizeof(target_uri));
		sprintf(forward, "PJSIP/%s/%s", ast_sorcery_object_get_id(session->endpoint), target_uri);
		ast_channel_call_forward_set(session->channel, forward);
	}

	return PJSIP_REDIRECT_STOP;
}

#ifdef GRANDSTREAM_NETWORKS
#ifdef BUILD_ON_CLOUD_PBX
static void set_sbc_endpoint_addr(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session || !rdata) {
		return;
	}

	static const pj_str_t sbc_endpt_addr_name = {"SBC-Endpoint-Address", 20};
	pjsip_generic_string_hdr *hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &sbc_endpt_addr_name, NULL);
	if (!hdr) {
		memset(session->sbc_endpoint_addr, 0, sizeof(session->sbc_endpoint_addr));
		return;
	}

	snprintf(session->sbc_endpoint_addr, sizeof(session->sbc_endpoint_addr), "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);
	return;
}

/* Added by yfyang, for Bug 303715 */
static void set_dest_addrs_to_inv_session(struct ast_sip_session *session)
{
	char *trunk_host = NULL;
	char *trunk_ip   = NULL;
	char *rand_addr  = NULL;
	char *endpt_addr = NULL;
	char  addrs[128] = {0};

	if (!session || !session->inv_session || !session->endpoint) {
		return;
	}

	if (!ast_strlen_zero(session->endpoint->trunk_host)) {
		trunk_host = (char *)session->endpoint->trunk_host;

		ast_sip_resolve_identifier_get_destaddrs(trunk_host, addrs, sizeof(addrs));
		if (strlen(addrs)) {
			trunk_ip = addrs;
		}
	}

	if (!ast_strlen_zero(session->endpoint->sbc_rand_addr)) {
		rand_addr = (char *)session->endpoint->sbc_rand_addr;
	}

	if (!ast_strlen_zero(session->sbc_endpoint_addr)) {
		endpt_addr = session->sbc_endpoint_addr;
	}

	ast_log(LOG_NOTICE, "set_dest_addrs_to_inv_session: trunk_host[%s], trunk_ip[%s], rand_addr[%s], endpt_addr[%s]", 
	        trunk_host?trunk_host:"", trunk_ip?trunk_ip:"", rand_addr?rand_addr:"", endpt_addr?endpt_addr:"");
	pjsip_inv_set_cloudtk_hdr_list(session->inv_session, trunk_host, trunk_ip, rand_addr, endpt_addr);
	return;
}
#endif

/* Add by bfdu for yealink flag, 2015.12.01 */
static void set_yealink_flag(pjsip_rx_data *rdata, struct ast_sip_session *session)
{
	pjsip_msg *msg = NULL;
	pjsip_user_agent_hdr *user_agent_hdr = NULL;
	const pj_str_t name = { "User-Agent", 10 };
	const pj_str_t str_yealink = { "Yealink", 7 };
	char *user_agent_str = NULL;
	int user_agent_size = 0;

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "Set yealink flag, para is invalid\n");
		return;
	}

	msg = rdata->msg_info.msg;
	if (msg) {
		user_agent_hdr = (pjsip_user_agent_hdr*) pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	}

	if (!user_agent_hdr) {
		return;
	}

	user_agent_size = pj_strlen(&user_agent_hdr->hvalue) + 1;
	user_agent_str = ast_malloc(user_agent_size);
	ast_copy_pj_str(user_agent_str, &user_agent_hdr->hvalue, user_agent_size);

	if (pj_strstr(&user_agent_hdr->hvalue, &str_yealink) != NULL) {
		ast_log(LOG_DEBUG, "[session:%p] Set yealink flag true for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->is_yealink_msg = TRUE;
	}

	ast_free(user_agent_str);
	user_agent_str = NULL;

	return;
}

/* To force change profile-level-id value when receive yealink message, 2015.12.02 */
static pjmedia_sdp_attr* pjmedia_sdp_attr_change_fmtp_param(pj_pool_t *pool,
	pjmedia_sdp_attr *attr_fmtp, const pj_str_t *payload)
{
	pj_str_t attr_name = pj_str("fmtp");
	pj_str_t find_str = pj_str("profile-level-id");
	char* fmtp_profile = NULL;
	char* change_param_str = "profile-level-id=428014";
	char* fmtp_buf = NULL;
	pj_str_t fmtp_value;
	int len_fmtp_value = 0;

	if (!pool || !attr_fmtp || !payload) {
		ast_log(LOG_DEBUG, "Param is null, don't change");
		return NULL;
	}

	if (pj_strcmp(&attr_fmtp->name, &attr_name)) {
		ast_log(LOG_DEBUG, "It's not a fmtp, don't change");
		return NULL;
	}

	len_fmtp_value = attr_fmtp->value.slen;

	fmtp_buf = pj_pool_alloc(pool, len_fmtp_value);
	pj_strset(&fmtp_value, fmtp_buf, 0);
	pj_strcpy(&fmtp_value, &attr_fmtp->value);
	fmtp_profile = pj_strstr(&fmtp_value, &find_str);
	if (!fmtp_profile) {
		ast_log(LOG_DEBUG, "Can't find profile-level-id string, don't need change\n");
		return pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_value);
	}

	/* don't copy '\0' to fmtp_profile, 23 is the len of "profile-level-id=428014" */
	strncpy(fmtp_profile, change_param_str, 23);

	return pjmedia_sdp_attr_create(pool, "fmtp", &fmtp_value);
}

void ast_sip_session_modify_outgoing_yealink_sdp(const struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_media *media)
{
	int i = 0;

	if (!session->is_yealink_msg) {
		return;
	}

	for (i = 0; i < media->desc.fmt_count; ++i) {
		pjmedia_sdp_attr *attr_fmtp = NULL;
		pjmedia_sdp_attr *new_attr_fmtp = NULL;

		attr_fmtp = pjmedia_sdp_media_find_attr2(media, "fmtp", &media->desc.fmt[i]);

		if (attr_fmtp) {
			ast_log(LOG_DEBUG, "[session:%p] Modify outgoing yealink sdp for endpoint:%s, find fmtp attr\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			pjmedia_sdp_media_remove_attr(media, attr_fmtp);

			new_attr_fmtp = pjmedia_sdp_attr_change_fmtp_param(pool, attr_fmtp, &media->desc.fmt[i]);
			if (new_attr_fmtp) {
				pjmedia_sdp_media_add_attr(media, new_attr_fmtp);
			}
		}
	}

	return;
}

/* Add by bfdu for webrtc, 2017.07.11 */
static void set_webrtc_call_flag(struct ast_sip_session *session, const char *uri, struct ast_sip_contact *found_contact)
{
	if (!session || !uri || !session->endpoint) {
		return;
	}

	if (!session->endpoint->media.rtp.ice_support) {
		ast_log(LOG_DEBUG, "[session:%p] ice_support is not enable for endpoint:%s, don't need set\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/* Now we just use transport is WS/WSS to judge is a webrtc call, maybe need add more condition later */
	if ((strcasestr(uri, "transport=WS") || strcasestr(uri, "transport=WSS"))) {
		ast_log(LOG_DEBUG, "[session:%p] It's a webrtc call for endpoint:%s, contact uri:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint), uri);
		session->is_webrtc_call = AST_RTP_ICE_ENABLE;
		session->terminal_type = AST_TERMINAL_TYPE_WEBRTC;
	} else {
		session->is_webrtc_call = 0;
	}

	if (is_wave_terminal_type_according_contact(found_contact)) {
		ast_log(LOG_DEBUG, "[session:%p] Set WEBRTC terminal type for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->terminal_type = AST_TERMINAL_TYPE_WEBRTC;
		session->is_webrtc_call = 1;
	}

	return;
}

static void set_ice_flag(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_contact_hdr *contact_hdr = NULL;
	pjsip_param *ice = NULL;
	static const pj_str_t ice_str = { "+sip.ice", 8 };
	pjsip_param *trickle_ice = NULL;
	static const pj_str_t trickle_ice_str = { "trickle-ice", 11 };

	contact_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT, NULL);
	if (!contact_hdr) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get contact header for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ice = pjsip_param_find(&contact_hdr->other_param, &ice_str);
	if (ice) {
		ast_log(LOG_DEBUG, "[session:%p] Contact contain sip.ice, support ice for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->support_ice = 1;
	}

	trickle_ice = pjsip_param_find(&contact_hdr->other_param, &trickle_ice_str);
	if (trickle_ice) {
		ast_log(LOG_DEBUG, "[session:%p] Contact contain trickle-ice, support trickle ice for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->support_trickle_ice = 1;
	}

	return;
}

static void set_support_whiteboard_flag(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	unsigned int i = 0;
	pjsip_msg *msg = NULL;
	pjsip_supported_hdr *supported_hdr = NULL;
	const pj_str_t whiteboard = { "whiteboard", 10 };

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "set_support_whiteboard_flag, para is invalid\n");
		return;
	}

	if (!im_support_whiteboard) {
		ast_log(LOG_DEBUG, "[session:%p] IM server can't support whiteboard, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!strncmp(ast_sorcery_object_get_id(session->endpoint), "trunk_", 6)) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s is a trunk, don't need support whiteboard\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		return;
	}

	supported_hdr = pjsip_msg_find_hdr(msg, PJSIP_H_SUPPORTED, NULL);
	if (!supported_hdr) {
		return;
	}

	for (i = 0; i < supported_hdr->count; i++) {
		if (!pj_stricmp(&supported_hdr->values[i], &whiteboard)) {
			ast_log(LOG_DEBUG, "[session:%p] Get Supported hdr has whiteboard vaule, set support_whiteboard to true, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->support_whiteboard = 1;
			break;
		}
	}

	return;
}


static void set_support_gs_state_flag(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	unsigned int i = 0;
	pjsip_msg *msg = NULL;
	pjsip_supported_hdr *supported_hdr = NULL;
	const pj_str_t gs_state = { "x-gs-state", 10 };

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "set_support_gs_state_flag, para is invalid\n");
		return;
	}


	msg = rdata->msg_info.msg;
	if (!msg) {
		return;
	}

	supported_hdr = pjsip_msg_find_hdr(msg, PJSIP_H_SUPPORTED, NULL);
	if (!supported_hdr) {
		return;
	}

	for (i = 0; i < supported_hdr->count; i++) {
		if (!pj_stricmp(&supported_hdr->values[i], &gs_state)) {
			ast_log(LOG_DEBUG, "[session:%p] Get Supported hdr has x-gs-state vaule, set support_gs_state to true, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->support_gs_state = 1;
			break;
		}
	}

	return;
}

static void set_candidate_host_only_flag(struct ast_sip_session *session, const char *to_host)
{
	unsigned int i = 0;
	unsigned int has_ipv4 = 0;
	const char *local_host_ip = NULL;
	char local_host[PJ_INET6_ADDRSTRLEN] = { 0 };

	if (!session || !to_host) {
		return;
	}

	local_host_ip = ast_sip_get_host_ip_string(pj_AF_INET());
	if (local_host_ip && !strcasecmp(to_host, local_host_ip)) {
		ast_log(LOG_DEBUG, "[session:%p] To hdr host matched host ipv4 address:%s, endpoint:%s\n",
			session, local_host_ip, ast_sorcery_object_get_id(session->endpoint));
		session->cand_host_only = 1;
		return;
	}

	local_host_ip = ast_sip_get_host_ip_string(pj_AF_INET6());
	if (local_host_ip && !strcasecmp(to_host, local_host_ip)) {
		ast_log(LOG_DEBUG, "[session:%p] To hdr host matched host ipv6 address:%s, endpoint:%s\n",
			session, local_host_ip, ast_sorcery_object_get_id(session->endpoint));
		session->cand_host_only = 1;
		return;
	}

	for (i = 0; i < local_address_count; i++) {
		memset(local_host, 0, sizeof(local_host));
		pj_sockaddr_print(&local_address[i], local_host, sizeof(local_host), 0);
		ast_log(LOG_DEBUG, "Get local host:%s, pos:%d\n", local_host, i);
		if (!strcasecmp(to_host, local_host)) {
			session->cand_host_only = 1;
			return;
		}
		if (local_address[i].addr.sa_family == PJ_AF_INET) {
			has_ipv4 = 1;
		}
	}

	/* If can't get ipv4 address, get local ip again and compare again */
	if (!has_ipv4 && !get_local_ip_again) {
		get_all_local_ip();
		get_local_ip_again = 1;
		for (i = 0; i < local_address_count; i++) {
			memset(local_host, 0, sizeof(local_host));
			pj_sockaddr_print(&local_address[i], local_host, sizeof(local_host), 0);
			ast_log(LOG_DEBUG, "Get local IP again, host:%s, pos:%d\n", local_host, i);
			if (!strcasecmp(to_host, local_host)) {
				session->cand_host_only = 1;
				return;
			}
		}
	}

	return;
}

static void set_candidate_host_only_flag_for_caller(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_from_hdr *to_hdr = NULL;
	pjsip_sip_uri *uri_to = NULL;
	char to_host[PJ_INET6_ADDRSTRLEN + 2] = { 0 };

	if (!session || !rdata) {
		return;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		ast_log(LOG_DEBUG, "[session:%p] No sip msg, can't set cand_host_only flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	to_hdr = pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
	if (!to_hdr) {
		ast_log(LOG_DEBUG, "[session:%p] No to header, can't set cand_host_only flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	uri_to = pjsip_uri_get_uri(to_hdr->uri);
	if (!uri_to) {
		ast_log(LOG_DEBUG, "[session:%p] No to header uri, can't set cand_host_only flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(to_host, &uri_to->host, sizeof(to_host));

	ast_log(LOG_DEBUG, "[session:%p] Receive INVITE To hdr host:%s, endpoint:%s\n",
		session, to_host, ast_sorcery_object_get_id(session->endpoint));

	set_candidate_host_only_flag(session, to_host);

	return;
}

static void set_candidate_host_only_flag_for_callee(struct ast_sip_session *session,
	struct ast_sip_contact *found_contact, const char *location)
{
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (!session || !session->endpoint || !location) {
		ast_log(LOG_DEBUG, "[session:%p] set_candidate_host_only_flag_for_callee can't get aor for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (found_contact) {
		ast_log(LOG_DEBUG, "[session:%p] set_candidate_host_only_flag_for_callee use found_contact endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		set_candidate_host_only_flag(session, found_contact->to_host);
	} else {
		contact = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(session->endpoint), location);
		if (contact) {
			ast_log(LOG_DEBUG, "[session:%p] set_candidate_host_only_flag_for_callee found contact by location, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			set_candidate_host_only_flag(session, contact->to_host);
		}
	}

	return;
}
/*Start added by zyxia for Bug 284585,cluster ip is wan, but reinvite from lan,2024.1.17 */
static int is_bound_any(pjsip_transport *transport)
{
	pj_uint32_t loop6[4] = {0, 0, 0, 0};

	if ((transport->local_addr.addr.sa_family == pj_AF_INET() &&
		transport->local_addr.ipv4.sin_addr.s_addr == PJ_INADDR_ANY) ||
		(transport->local_addr.addr.sa_family == pj_AF_INET6() &&
		!pj_memcmp(&transport->local_addr.ipv6.sin6_addr, loop6, sizeof(loop6)))) {
		return 1;
	}

	return 0;
}

static int getIfnameByIp(char * inbuf,char* ip){

	struct ifaddrs *ifaddr, *ifa;
	char host[NI_MAXHOST];
	int s;
	ifaddr = NULL;

	if(ast_strlen_zero(ip))
		return 0;

	getifaddrs(&ifaddr);
	for(ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next){
		if(!ifa->ifa_addr){
			continue;
		}
		if(ifa->ifa_addr->sa_family != AF_INET){
			continue;
		}
		s = getnameinfo(ifa->ifa_addr, sizeof(struct sockaddr_in), host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);
		if (s != 0) {
			continue;
		}
		if (strcmp(host, ip) == 0) {
			ast_log(LOG_DEBUG,"Interface name: %s\n", ifa->ifa_name);
			strcpy(inbuf, ifa->ifa_name);
			break;
		}
	}
	freeifaddrs(ifaddr);
	return 1;
}

static int getIpByIfname(char * inbuf,char* ifname){

	struct ifaddrs *ifaddr, *ifa;
	char host[NI_MAXHOST];
	int s;
	ifaddr = NULL;

	if(ast_strlen_zero(ifname))
		return 0;

	getifaddrs(&ifaddr);
	for(ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next){
		if(!ifa->ifa_addr){
			continue;
		}
		if(ifa->ifa_addr->sa_family != AF_INET){
			continue;
		}
		s = getnameinfo(ifa->ifa_addr, sizeof(struct sockaddr_in), host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);
		if (s != 0) {
			continue;
		}
		if (strcmp(ifa->ifa_name, ifname) == 0) {
			ast_log(LOG_DEBUG,"%s Interface Ip: %s\n", ifa->ifa_name,host);
			strcpy(inbuf, host);
			break;
		}
	}
	freeifaddrs(ifaddr);
	return 1;
}

static int getSessionLocalAddress(pjsip_rx_data *rdata,pj_pool_t *pool, char* remote_host, char* ipbuf, int buf_len)
{
	char local_ifname[64]={0};
	char floatip_ifname[64]={0};
	pjsip_tpmgr_fla2_param prm;
	char hafloatip[64] ={0};
	//char pvalue[512] = {0};
	char local_host[PJ_INET6_ADDRSTRLEN] = { 0 };

	pjsip_tpmgr_fla2_param_default(&prm);
	if(rdata) {
		prm.tp_type = rdata->tp_info.transport->key.type;
		pj_strset2(&prm.dst_host, rdata->pkt_info.src_name);
		prm.local_if = is_bound_any(rdata->tp_info.transport);
	}
	else {
		if (remote_host == NULL)
			return -1;
		prm.tp_type = PJSIP_TRANSPORT_TCP;
		pj_strset2(&prm.dst_host, remote_host);
		prm.local_if = TRUE;
	}


	if (pjsip_tpmgr_find_local_addr2(pjsip_endpt_get_tpmgr(ast_sip_get_pjsip_endpoint()), pool, &prm) == PJ_SUCCESS) {
		ast_copy_string(ipbuf, (&prm.ret_addr)->ptr, (prm.ret_addr.slen + 1) > buf_len ? buf_len : (prm.ret_addr.slen + 1));
	}

 	/*Start added by zyxia for Bug 286925 ,2024.1.22*/
	if (HA_ROLE_ACTIVE == ast_ha_get_role() && (ast_get_redundance_flag() != 1) && !ast_strlen_zero(ipbuf)) {

		/*start added by zyxia for Bug 307230*/
		if(remote_host) {
			for (int i = 0; i < local_address_count; i++) {
				memset(local_host, 0, sizeof(local_host));
				pj_sockaddr_print(&local_address[i], local_host, sizeof(local_host), 0);
				ast_log(LOG_DEBUG, "Get local host:%s, pos:%d\n", local_host, i);
				if ((local_address[i].addr.sa_family == PJ_AF_INET) && !strcasecmp(remote_host, local_host) ) {
					ast_copy_string(ipbuf, local_host, sizeof(local_host));
				}
			}
		}
		/*End added by zyxia for Bug 307230*/

		getIfnameByIp(local_ifname, ipbuf);
		ast_safe_nvram_get("ethx4floatip", floatip_ifname, sizeof(floatip_ifname));
		if(!ast_strlen_zero(local_ifname)&& !ast_strlen_zero(floatip_ifname) && !strcasecmp(local_ifname,floatip_ifname)){
			ast_safe_nvram_get("hafloatip", hafloatip, sizeof(hafloatip));
			ast_log(LOG_DEBUG, "local_ifname:%s, floatip_ifname:%s,float_ip:%s\n",local_ifname,floatip_ifname,hafloatip);
			if (!ast_strlen_zero(hafloatip)) {
				ast_copy_string(ipbuf, hafloatip, sizeof(hafloatip));
			}
		}
	}
	/*End added by zyxia for Bug 286925 ,2024.1.22*/

	return 0;
}

static void set_session_diff_multiavs_ip_flag_for_caller(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_from_hdr *to_hdr = NULL;
	pjsip_sip_uri *uri_to = NULL;
	char to_host[128] = { 0 };
	char *avs_network_ip = NULL;
	unsigned int i = 0;
	char local_host[128] = { 0 };
	
	char public_ip[256] = {0};
	char network_mode[8]={0};
	char ifname[8]={0};
	char ipaddr[256]={0};

	if (!session || !rdata) {
		return;
	}
	msg = rdata->msg_info.msg;
	if (!msg) {
		ast_log(LOG_DEBUG, "[session:%p] No sip msg, can't set_session_diff_multiavs_ip_flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	
	to_hdr = pjsip_msg_find_hdr(msg, PJSIP_H_TO, NULL);
	if (!to_hdr) {
		ast_log(LOG_DEBUG, "[session:%p] No to header, can't set_session_diff_multiavs_ip_flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	uri_to = pjsip_uri_get_uri(to_hdr->uri);
	if (!uri_to) {
		ast_log(LOG_DEBUG, "[session:%p] No to header uri, can't set_session_diff_multiavs_ip_flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(to_host, &uri_to->host, sizeof(to_host));

	if(!ast_sip_is_a_valid_IP_address(to_host)) {
		pjsip_server_addresses id_addresses = {
			.count = 0,
		};
		if (ast_sip_resolve_identifier_get((pj_str_t *)&uri_to->host, PJSIP_TRANSPORT_UNSPECIFIED, &id_addresses) > 0) {
			for (int idx = 0; idx < id_addresses.count; ++idx) {
				struct ast_sockaddr sock_addr = {{0,}};
				ast_sip_sockaddr_from_pj_sockaddr(&sock_addr, &id_addresses.entry[idx].addr);
				if (ast_sockaddr_is_ipv4(&sock_addr)) {
					snprintf(to_host, sizeof(to_host), "%s", ast_sockaddr_stringify_addr(&sock_addr));
				}
			}
		}
	}
	memset(local_host, 0, sizeof(local_host));
	//check to hdr from which network interface
	getSessionLocalAddress(rdata,rdata->tp_info.pool,to_host,local_host, sizeof(local_host));
	ast_log(LOG_DEBUG, "set_session_diff_multiavs_ip_flag_for_caller [session:%p] original_address:%s,%s endpoint:%s\n",
		session, to_host,local_host,ast_sorcery_object_get_id(session->endpoint));

	avs_network_ip = ast_get_avs_multicast_network_ip();
	if(avs_network_ip ){//to host is not equal to avs network ip
		if(!ast_strlen_zero(local_host) && strcasecmp(local_host,avs_network_ip)){
			session->diff_multiavs_ip = 1;
			ast_log(LOG_DEBUG, "[session:%p]  local host %s is different from the avs network ip %s, need set default avs addreess,endpoint:%s\n",
				session,local_host, avs_network_ip,ast_sorcery_object_get_id(session->endpoint));
		}
		else if(ast_strlen_zero(local_host)){

			for (i = 0; i < local_address_count; i++) {
				memset(local_host, 0, sizeof(local_host));
				pj_sockaddr_print(&local_address[i], local_host, sizeof(local_host), 0);
				ast_log(LOG_DEBUG, "Get local host:%s, pos:%d\n", local_host, i);
				if ((local_address[i].addr.sa_family == PJ_AF_INET) && !strcasecmp(to_host, local_host) && strcasecmp(to_host, avs_network_ip)) {
					session->diff_multiavs_ip = 1;
					ast_log(LOG_DEBUG, "[session:%p]  to uri %s is different from the avs network ip %s, need set default avs addreess,endpoint:%s\n",
						session,to_host, avs_network_ip,ast_sorcery_object_get_id(session->endpoint));
					ast_free(avs_network_ip);
					return;
				}
			}

			if (ast_safe_nvram_get("public_ip", public_ip, sizeof(public_ip)) != NULL) {
				//the to host is equal with  public up
				if(!ast_strlen_zero(to_host) && !strcasecmp(to_host,public_ip)){

					//route mode and dual mode default eth1
					const char *mode = ast_safe_nvram_get("231", network_mode, sizeof(network_mode));
					const char *default_ip = ast_safe_nvram_get("default_if", ifname, sizeof(ifname));
					if (!ast_strlen_zero(mode) && !ast_strlen_zero(default_ip) && (!strcmp(mode,"0") || (!strcmp(mode,"2") && !strcmp(default_ip,"LAN1") )
						)){
						getIpByIfname(ipaddr,"eth1");
					}
					//dual default eth0
					else if (!ast_strlen_zero(default_ip) &&  !strcmp(default_ip,"LAN2")){
						getIpByIfname(ipaddr,"eth0");
					}


					if(!ast_strlen_zero(ipaddr) && strcasecmp(ipaddr,avs_network_ip)){
						session->diff_multiavs_ip = 1;
						ast_log(LOG_DEBUG, "[session:%p]  local host %s is different from the avs network ip %s, need set default avs addreess,endpoint:%s\n",
							session,ipaddr, avs_network_ip,ast_sorcery_object_get_id(session->endpoint));
					}

				}
			}

		}
		ast_free(avs_network_ip);
	}
	return;
}
//check callee use which network interface
static int parse_ip_from_uri(const char* uri,char* output) {

	if(ast_strlen_zero(uri))
		return 0;

	const char* at_sign = strchr(uri, '@');
	if (at_sign == NULL)
		return 0;

	const char* colon = strchr(at_sign + 1, ':');

	if (colon == NULL)
		return 0;

	size_t ip_start = at_sign - uri + 1;
	size_t ip_length = colon - (at_sign + 1);

	strncpy(output, uri + ip_start, ip_length);
	output[ip_length] = '\0';

	return 1;
} 

static void set_session_diff_multiavs_ip_flag_for_callee(struct ast_sip_session *session, struct ast_sip_contact *found_contact,const char *location, char* uri)
{
	char *avs_network_ip = NULL;
	char local_host[PJ_INET6_ADDRSTRLEN] = { 0 };
	char *original_address= NULL;
	char* addr =NULL;
	char addrs[128] = { 0 };

	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (!session || !session->endpoint) {
		ast_log(LOG_DEBUG, "[session:%p] set_support_gs_fec_for_session can't get aor for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	pj_pool_t *pool;
	pool = pjsip_endpt_create_pool(ast_sip_get_pjsip_endpoint(), "Transport Availability", 256, 256);
	if (!pool) {
		return ;
	}

	if (found_contact) {
		original_address = ast_strdupa(found_contact->original_address);
	} else {
		if (location) {
			contact = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(session->endpoint), location);
			if(contact) {
				original_address = ast_strdupa(contact->original_address);
			}
		}
	}

	if(original_address && (!ast_strlen_zero(original_address))){
		const char* colon_pos = strstr(original_address, "]");//ipv6
		if (colon_pos) {//ipv6 [2408:8640:8fe:f7:c274:adff:fede:f0cd]:5666
			char ipv6[MAX_IPADDR_V6_LEN]={0};
			size_t ip_len = colon_pos - original_address-1;//remove [,]
			strncpy(ipv6, original_address+1, ip_len);
			ipv6[ip_len] = '\0';
			addr = ipv6;
		}
		else if(strstr(original_address, ":")){//ipv4
			addr = strtok(original_address, ":");
		}
		else
			addr = ast_strdupa(original_address);

	}
	//if is inbound call from trunk ,need get addr from uri
	/*start added by zyxia for Bug 299584*/
	else if (strstr(ast_sorcery_object_get_id(session->endpoint), "trunk_")&& !ast_strlen_zero(uri)) {
		const char* colon_pos = strstr(uri, ":");
		if (colon_pos) 
			addr = ast_strdupa(colon_pos+1);
		else
			addr = ast_strdupa(uri);

		//if is a  domain, need get dns resoult
		if(!ast_sip_is_a_valid_IP_address(addr)) {
			pj_str_t to_addr;
			pjsip_server_addresses id_addresses = {
				.count = 0,
			};
			pj_strset2(&to_addr, (char *)addr);
			if (ast_sip_resolve_identifier_get((pj_str_t *)&to_addr, PJSIP_TRANSPORT_UNSPECIFIED, &id_addresses) > 0) {
				for (int idx = 0; idx < id_addresses.count; ++idx) {
					struct ast_sockaddr sock_addr = {{0,}};
					ast_sip_sockaddr_from_pj_sockaddr(&sock_addr, &id_addresses.entry[idx].addr);
					if (ast_sockaddr_is_ipv4(&sock_addr)) {
						snprintf(addrs, sizeof(addrs), "%s", ast_sockaddr_stringify_addr(&sock_addr));
						addr = addrs;
					}
				}
			}
		}
	}
	else if (!ast_strlen_zero(uri)) {
		char ip[MAX_IPADDR_V6_LEN]={0};
		if(parse_ip_from_uri(uri,ip) && !ast_strlen_zero(ip))
			addr = ip;
	}

	memset(local_host, 0, sizeof(local_host));
	getSessionLocalAddress(NULL,pool,addr,local_host, sizeof(local_host));
	pjsip_endpt_release_pool(ast_sip_get_pjsip_endpoint(), pool);

	ast_log(LOG_DEBUG, "[session:%p] set_session_diff_multiavs_ip_flag_for_callee: addr:%s,local_host:%s endpoint:%s\n",
	session, addr ? addr :"",local_host,ast_sorcery_object_get_id(session->endpoint));
	
	avs_network_ip = ast_get_avs_multicast_network_ip();
	if(avs_network_ip ){//to host is not equal to avs network ip
		if(!ast_strlen_zero(local_host) && strcasecmp(local_host,avs_network_ip)){
			session->diff_multiavs_ip = 1;
			ast_log(LOG_DEBUG, "[session:%p]  local_host %s is different from the avs network ip %s, need set default avs addreess,endpoint:%s\n",
				session,local_host, avs_network_ip,ast_sorcery_object_get_id(session->endpoint));
		}
		ast_free(avs_network_ip);
	}
	return;
}

/*End added by zyxia for Bug 284585,cluster ip is wan, but invite from lan,2024.1.17  */

static void set_session_allow_events(struct ast_sip_session *session, struct pjsip_rx_data *rdata)
{
	const pj_str_t allow_events = { "Allow-Events", 12 };
	pjsip_generic_string_hdr *hdr = NULL;
	char hdr_buffer[128] = { 0 };
	char *hdr_str = NULL;
	char *allow_str = NULL;

	if (!session || !rdata) {
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &allow_events, NULL);
	if (!hdr) {
		return;
	}

	ast_copy_pj_str(hdr_buffer, &hdr->hvalue, 128);
	hdr_str = ast_strdupa(hdr_buffer);

	while((allow_str = ast_strsep(&hdr_str, ',', AST_STRSEP_TRIM))) {
		if (strcasestr(allow_str, "talk")) {
			ast_log(LOG_DEBUG, "[session:%p] Allow-Events has talk value, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->allow_talk_event = 1;
			continue;
		}

		if (strcasestr(allow_str, "hold")) {
			ast_log(LOG_DEBUG, "[session:%p] Allow-Events has hold value, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->allow_hold_event = 1;
			continue;
		}
	}

	return;
}

static void set_session_emergency_call(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	const char *emergency_numbers = NULL;
	char *emergency_number = NULL;
	char *tmp_emergency_numbers = NULL;
	char from_user[256] = {0};
	pjsip_sip_uri *uri_from = NULL;
	pjsip_from_hdr *from_hdr = NULL;

	if (!session || !rdata) {
		return;
	}

	if (strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return;
	}

	emergency_numbers = pbx_builtin_getvar_helper(NULL, "EMERGENCY_CALLS");
	if (ast_strlen_zero(emergency_numbers)) {
		return;
	}

	from_hdr = pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_FROM, NULL);
	if (!from_hdr) {
		ast_log(LOG_DEBUG, "[session:%p] set_session_emergency_call can't get from hdr from rdata, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	uri_from = pjsip_uri_get_uri(from_hdr->uri);
	if (!uri_from) {
		ast_log(LOG_DEBUG, "[session:%p] set_session_emergency_call can't get from uri from rdata, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(from_user, &uri_from->user, sizeof(from_user));

	ast_log(LOG_DEBUG, "[session:%p] set_session_emergency_call get from user:%s, emergency_numbers:%s, endpoint:%s\n",
		session, from_user, emergency_numbers, ast_sorcery_object_get_id(session->endpoint));

	tmp_emergency_numbers = ast_strdupa(emergency_numbers);
	while ((emergency_number = strsep(&tmp_emergency_numbers, "/")) ) {
		if (!strcasecmp(emergency_number, from_user)) {
			session->emergemcy_call = 1;
			ast_log(LOG_DEBUG, "[session:%p] set_session_emergency_call from user matched emergency_number, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
		}
	}

	return;
}
/*Start added  by zyxia for Bug 265588,2023.7.19*/
void ast_sip_session_call_to_emergency_call(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	const char *emergency_numbers = NULL;
	char *emergency_number = NULL;
	char *tmp_emergency_numbers = NULL;
	char request_user[256] = {0};
	const pjsip_msg *rx_msg = NULL;
	pjsip_sip_uri *uri_sip = NULL;

	if (!session || !rdata) {
		return;
	}

	//if from trunk ,ignore it
	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return;
	}

	rx_msg = rdata->msg_info.msg;
	if (rx_msg == NULL) {
		return ;
	}

	emergency_numbers = pbx_builtin_getvar_helper(NULL, "EMERGENCY_CALLS");
	if (ast_strlen_zero(emergency_numbers)) {
		return;
	}

	uri_sip = pjsip_uri_get_uri(rx_msg->line.req.uri);
	if (!uri_sip) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_call_to_emergency_call can't get request uri from rdata, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(request_user, &uri_sip->user, sizeof(request_user));

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_call_to_emergency_call get request user:%s, emergency_numbers:%s, endpoint:%s\n",
		session, request_user, emergency_numbers, ast_sorcery_object_get_id(session->endpoint));

	tmp_emergency_numbers = ast_strdupa(emergency_numbers);
	while ((emergency_number = strsep(&tmp_emergency_numbers, "/")) ) {
		if (!strcasecmp(emergency_number, request_user)) {
			session->emergemcy_call = 1;
			ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_call_to_emergency_call request user matched emergency_number, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
		}
	}

	return;
}
/*End added  by zyxia for Bug 265588,2023.7.19*/

/*Start added  by zyxia for Bug 261524 ,2023.7.19*/
void ast_sip_session_set_call_flip(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	const char *call_flip_num = NULL;

	char request_user[256] = {0};
	const pjsip_msg *rx_msg = NULL;
	pjsip_sip_uri *uri_sip = NULL;

	if (!session || !rdata) {
		return;
	}

	//if from trunk ,ignore it
	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return;
	}

	rx_msg = rdata->msg_info.msg;
	if (rx_msg == NULL) {
		return ;
	}

	call_flip_num = pbx_builtin_getvar_helper(NULL, "FEATURE_KEY_CALLFLIP");
	if (ast_strlen_zero(call_flip_num)) {
		return;
	}

	uri_sip = pjsip_uri_get_uri(rx_msg->line.req.uri);
	if (!uri_sip) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_call_flip can't get request uri from rdata, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(request_user, &uri_sip->user, sizeof(request_user));

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_call_flip get request user:%s, callflip number:%s, endpoint:%s\n",
		session, request_user, call_flip_num, ast_sorcery_object_get_id(session->endpoint));


	if (!strcasecmp(call_flip_num, request_user)) {
		session->is_callflip_call = 1;
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_call_flip request user matched callflip number, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}
/*End added  by zyxia for Bug 261524 ,2023.7.19*/

/* Add by bfdu for whiteboard, 2022.1.7 */
int ast_sip_session_add_support_header_for_whiteboard(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_supported_hdr *hdr = NULL;
	pj_str_t WHITEBOARD_SUPPORTED_NAME = { "whiteboard", 10 };

	if (!session || !tdata) {
		return 0;
	}

	if (!im_support_whiteboard) {
		ast_log(LOG_DEBUG, "[session:%p] IM server can't support whiteboard, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		ast_log(LOG_DEBUG, "[session:%p] It's trunk, don't need add whiteboard, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	hdr = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_SUPPORTED, NULL);
	if (!hdr) {
		/* insert a new Supported header */
		hdr = pjsip_supported_hdr_create(tdata->pool);
		if (!hdr) {
			return -1;
		}

		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
	}

	ast_log(LOG_DEBUG, "[session:%p] add_support_header_for_whiteboard, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	pj_strassign(&hdr->values[hdr->count++], &WHITEBOARD_SUPPORTED_NAME);

	return 0;
}

/* Start Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */
int ast_sip_session_add_support_header_for_gs_state(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_supported_hdr *hdr = NULL;
	pj_str_t GS_STATE_SUPPORTED_NAME = { "x-gs-state", 10 };

	if (!session || !tdata) {
		return 0;
	}


	hdr = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_SUPPORTED, NULL);
	if (!hdr) {
		/* insert a new Supported header */
		hdr = pjsip_supported_hdr_create(tdata->pool);
		if (!hdr) {
			return -1;
		}

		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
	}

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_add_support_header_for_gs_state, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	pj_strassign(&hdr->values[hdr->count++], &GS_STATE_SUPPORTED_NAME);

	return 0;
}
/* End Add by zyxia for x-gs-state Bug 262507 , 2023.6.14 */

/* We just use INVITE or 200OK to set terminal type */
static unsigned int judge_if_need_set_terminal_type(pjsip_rx_data *rdata)
{
	if (!rdata || !rdata->msg_info.msg || !rdata->msg_info.cseq) {
		return FALSE;
	}

	struct pjsip_request_line req = rdata->msg_info.msg->line.req;

	if (!pjsip_method_cmp(&req.method, &pjsip_invite_method)) {
		return TRUE;
	}

	if (!pj_stricmp2(&rdata->msg_info.cseq->method.name, "INVITE") && 
		(rdata->msg_info.msg->line.status.code)/100 == 2) {
		return TRUE;
	}

	return FALSE;
}

static unsigned int is_gvc_support_sfu_terminal(pjsip_rx_data *rdata)
{
	char s_line[256] = {0};
	pjsip_rdata_sdp_info *sdp_info = NULL;

	sdp_info = pjsip_rdata_get_sdp_info(rdata);
	if (sdp_info && sdp_info->sdp) {
		ast_copy_pj_str(s_line, &sdp_info->sdp->name, sizeof(s_line));
		if (strstr(s_line, TERMINAL_TYPE_GVC)) {
			return TRUE;
		}
	}

	return FALSE;
}

static unsigned int is_gswave_moblie_terminal(pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t hdr_type = {"Grandstream-type", 16};

	if (!rdata) {
		return FALSE;
	}

	if (!(hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &hdr_type, NULL))) {
		return FALSE;
	}

	if (pj_stricmp2(&hdr->hvalue, "GSWave mobile")) {
		return FALSE;
	}

	return TRUE;
}

static unsigned int is_trunk_support_ice(struct ast_sip_session *session)
{
	if (!session->endpoint->media.rtp.ice_support) {
		ast_log(LOG_DEBUG, "[session:%p] ice_support is not enable for endpoint:%s, don't need set\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return TRUE;
	}

	return FALSE;
}

static void set_terminal_type(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session || !rdata) {
		return;
	}

	if (!judge_if_need_set_terminal_type(rdata)) {
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_type is not INVITE method, don't need set for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->is_webrtc_call) {
		session->terminal_type = AST_TERMINAL_TYPE_WEBRTC;
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_type WEBRTC for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (is_gswave_moblie_terminal(rdata)) {
		session->terminal_type = AST_TERMINAL_TYPE_GSWAVE_MOBILE;
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_type gswave mobile for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_judge_if_wave_terminal_type(rdata)){
		session->terminal_type = AST_TERMINAL_TYPE_WEBRTC;
		session->is_webrtc_call = 1;
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_type WEBRTC according user agent for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (is_gvc_support_sfu_terminal(rdata)) {
		session->terminal_type = AST_TERMINAL_TYPE_GVC;
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_type GVC for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (is_trunk_support_ice(session) && !session->endpoint->media.turn) {
		if (session->terminal_type) {
			ast_log(LOG_DEBUG, "[session:%p] It's already set type before, don't set again for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		} else {
			session->terminal_type = AST_TERMINAL_TYPE_ICE_TRUNK;
			ast_log(LOG_DEBUG, "[session:%p] set_terminal_type ICE trunk for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
		return;
	}

	/* If terminal_type already set ICE_TRUNK,don't set again */
	if (session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) {
		ast_log(LOG_DEBUG, "[session:%p] Already set AST_TERMINAL_TYPE_ICE_TRUNK for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	} else {
		/* For other case, we conside as normal sip terminal */
		session->terminal_type = AST_TERMINAL_TYPE_SIP;
		ast_log(LOG_DEBUG, "[session:%p] set_terminal_type SIP for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}
	
	return;
}

static void set_caller_webrtc_call_flag(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session || !rdata || !session->endpoint) {
		return;
	}

	if (!session->endpoint->media.rtp.ice_support) {
		ast_log(LOG_DEBUG, "[session:%p] ice_support is not enable for endpoint:%s, don't need set\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	/* Now we just use transport is WS/WSS to judge is a webrtc call, maybe need add more condition later */
	if (!strcasecmp(rdata->tp_info.transport->type_name, "WS")
		|| !strcasecmp(rdata->tp_info.transport->type_name, "WSS")) {
		ast_log(LOG_NOTICE, "[session:%p] Set caller webrtc call flag for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->is_webrtc_call = AST_RTP_ICE_ENABLE;
	} else {
		session->is_webrtc_call = 0;
	}


	return;
}

static void set_ice_trunk_for_session(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	if (session->endpoint->media.turn) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint is enable TURN, don't set ICE flag, endpoint:%s",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (is_trunk_support_ice(session)) {
		session->terminal_type = AST_TERMINAL_TYPE_ICE_TRUNK;
		ast_log(LOG_DEBUG, "[session:%p] set_ice_trunk_for_session for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}


static void set_is_sdp_ipv6(struct ast_sip_session *session, pjmedia_sdp_session *sdp)
{
	if (!session || !sdp) {
		return;
	}

	if (sdp->conn) {
		session->is_sdp_ipv6 = (pj_strcmp2(&sdp->conn->addr_type, "IP6") ? 0 : 1);
	} else {
		/* todo: need to check both have IP4 and IP6 conn address */
		if (sdp->media[0] && sdp->media[0]->conn) {
			session->is_sdp_ipv6 = (pj_strcmp2(&sdp->media[0]->conn->addr_type, "IP6") ? 0 : 1);
		} else {
			ast_log(LOG_WARNING, "[session:%p] No audio connection line for endpoint:%s, can't set is_sdp_ipv6 value\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	return;
}

static void set_media_turn_relay(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

#ifdef BUILD_ON_CLOUD_PBX
	ast_log(LOG_DEBUG, "[session:%p] It's cloudUCM, don't need set media_turn_relay, endpoint:%s",
		session, ast_sorcery_object_get_id(session->endpoint));
	return;
#endif

	if (session->is_webrtc_call ||
		session->support_ice ||
		session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ||
		session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE ||
		session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) {
		return;
	}

	if (session->cand_host_only) {
		ast_log(LOG_DEBUG, "[session:%p] It's local call, don't need set media_turn_relay, endpoint:%s",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->endpoint->media.turn) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint is enable TURN, set media_turn_relay true, endpoint:%s",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->media_turn_relay = 1;
		return;
	}

	if (!session->ssh_proxy_flag) {
		return;
	}

	session->media_turn_relay = 1;

	return;
}

static void set_stream_content(struct ast_stream *asterisk_stream, const struct pjmedia_sdp_media *stream, enum ast_media_type type)
{
	pjmedia_sdp_attr *attr = NULL;
	char content[MAX_CONTENT_LEN] = {0};

	if (type != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	if ((attr = pjmedia_sdp_media_find_attr2(stream, "content", NULL))) {
		ast_copy_pj_str(content, &attr->value, sizeof(content));
		if (strlen(content)) {
			ast_stream_set_metadata(asterisk_stream, "CONTENT", content);
		}
	}

	return;
}

static void send_bfcp_for_open_slides(struct ast_sip_session *session)
{
	struct ast_sip_session_media *session_media = NULL;

	if (!session) {
		return;
	}

	session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);

	if (!session_media || !session_media->open_slides) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Send BFCP when SIP phone open camera with slides video media, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session_media->open_slides = 0;

	ast_sip_session_notify_sfu_bfcp_for_join_conference(session);

	return;
}

static void send_delayed_bfcp_message(struct ast_sip_session *session)
{
	struct ast_sip_session_media *bfcp_media = NULL;

	if (!session) {
		return;
	}

	if (!session->pending_bfcp_message) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_delayed_bfcp_message for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	bfcp_media = ast_sip_session_get_bfcp_active_media(session);
	if (!bfcp_media) {
		ast_log(LOG_DEBUG, "[session:%p] Has no bfcp active media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (bfcp_media->bfcp_udp) {
		ast_send_bfcp_msg(bfcp_media->bfcp_udp);
	} else if(bfcp_media->bfcp_tcp) {
		ast_send_bfcp_tcp_msg(bfcp_media->bfcp_tcp);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Has no bfcp udp or tcp, can't send bfcp message for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	session->pending_bfcp_message = 0;

	return;
}

static void send_delayed_session_to_peer(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_DEBUG, "send_delayed_session_to_peer param is NULL\n");
		return;
	}

	if (!session->delayed_send_session) {
		return;
	}

	session->delayed_send_session = 0;

	ast_sip_session_send_session_to_peer(session);

	return;
}

static unsigned int judge_if_has_valid_main_stream(struct ast_sip_session *session)
{
	struct ast_stream *stream = NULL;

	stream = ast_sip_session_get_active_main_video_stream(session);
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_SENDRECV) {
		return 1;
	}

	return 0;
}

static void queue_camera_control_frame(struct ast_sip_session *session)
{
	struct ast_camera_control_info camera_control_info;

	if (!session->camera_control_action) {
		return;
	}

	memset(&camera_control_info, 0, sizeof(struct ast_camera_control_info));

	switch (session->camera_control_action) {
		case AST_SIP_CAMERA_CONTROL_ACTION_OPEN:
			camera_control_info.action = CAMERA_CONTROL_ACTION_OPEN_PARTICIPANT_CAMERA_STATUS;
			if (judge_if_has_valid_main_stream(session)) {
				camera_control_info.status = CAMERA_CONTROL_STATUS_SUCCEED;
			} else {
				camera_control_info.status = CAMERA_CONTROL_STATUS_FAILED;
				strcpy(camera_control_info.failed_reason, "Peer decline open video camera request");
			}
			break;
		case AST_SIP_CAMERA_CONTROL_ACTION_CLOSE:
			camera_control_info.action = CAMERA_CONTROL_ACTION_CLOSE_PARTICIPANT_CAMERA_STATUS;
			camera_control_info.status = CAMERA_CONTROL_STATUS_SUCCEED;
			break;
		default:
			break;
	}

	ast_queue_control_data(session->channel, AST_CONTROL_CAMERA_CONTROL, &camera_control_info, sizeof(camera_control_info));

	return;
}

static void set_session_remote_support_bundle(struct ast_sip_session *session, const pjmedia_sdp_session *remote)
{
	unsigned int index = 0;
	pjmedia_sdp_attr *attr = NULL;

	if (!session || !remote) {
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE && !session->endpoint->media.mobile_bundle_enabled) {
		ast_log(LOG_DEBUG, "[session:%p] Mobile bundle is not enabled, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && !session->endpoint->media.gswave_bundle_enabled) {
		ast_log(LOG_DEBUG, "[session:%p] Gswave bundle is not enabled, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (session->remote_support_bundle) {
		/* Do not need set agadin */
		return;
	}

	for (index = 0; index < remote->attr_count; ++index) {
		attr = remote->attr[index];
		if (!attr) {
			continue;
		}

		if (pj_strcmp2(&attr->name, "group") || pj_strncmp2(&attr->value, "BUNDLE", 6)) {
			continue;
		}

		session->remote_support_bundle = 1;
		break;
	}

	return;
}

static void change_attr_to_session_level(pj_pool_t *pool,
	pjmedia_sdp_session *local, int stream, char *attr_name)
{
	pjmedia_sdp_attr *media_attr = NULL;
	pjmedia_sdp_attr *session_attr = NULL;
	pjmedia_sdp_attr *clone_attr = NULL;

	media_attr = pjmedia_sdp_media_find_attr2(local->media[stream], attr_name, NULL);
	if (!media_attr) {
		return;
	}

	session_attr = pjmedia_sdp_attr_find2(local->attr_count, local->attr, attr_name, NULL);
	if (!session_attr) {
		clone_attr = pjmedia_sdp_attr_clone(pool, media_attr);
		pjmedia_sdp_session_add_attr(local, clone_attr);
		pjmedia_sdp_media_remove_all_attr(local->media[stream], attr_name);
	} else {
		if (!pj_strcmp(&session_attr->value, &media_attr->value)) {
			pjmedia_sdp_media_remove_all_attr(local->media[stream], attr_name);
		} else {
			ast_log(LOG_DEBUG, "Media level attr is not the same session level attr\n");
		}
	}

	return;
}

static void move_dtls_and_ice_attrs_to_session_level(pj_pool_t *pool,
	pjmedia_sdp_session *local, struct ast_sip_session *session)
{
	int stream = 0;

	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) {
		return;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return;
	}

	for (stream = 0; stream < local->media_count; stream++) {
		if (!local->media[stream]) {
			continue;
		}

		change_attr_to_session_level(pool, local, stream, "ice-ufrag");
		change_attr_to_session_level(pool, local, stream, "ice-pwd");
		change_attr_to_session_level(pool, local, stream, "ice-options");
		change_attr_to_session_level(pool, local, stream, "fingerprint");
		change_attr_to_session_level(pool, local, stream, "setup");
	}

	return;
}

static unsigned int judge_if_remote_ice_support(struct ast_sip_session *session, const struct pjmedia_sdp_media *remote_stream)
{
	const pjmedia_sdp_attr *attr = NULL;
	unsigned int attr_i = 0;

	if (!session || !remote_stream) {
		return 0;
	}

	if (!session->endpoint->media.rtp.ice_support) {
		return 0;
	}

	/* Find all of the candidates */
	for (attr_i = 0; attr_i < remote_stream->attr_count; ++attr_i) {
		attr = remote_stream->attr[attr_i];
		if (!pj_strcmp2(&attr->name, "candidate")) {
			session->support_ice = 1;
			return 1;
		}
	}

	return 0;
}

static void clear_media_turn_relay_for_remote_support_ice(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	unsigned int remote_ice_support = 0;
	pjsip_rdata_sdp_info *sdp_info = NULL;

	if (!session || !rdata) {
		return;
	}

	if (!session->media_turn_relay) {
		return;
	}

	sdp_info = pjsip_rdata_get_sdp_info(rdata);

	if (sdp_info && (sdp_info->sdp_err == PJ_SUCCESS) && sdp_info->sdp) {
		remote_ice_support = judge_if_remote_ice_support(session, sdp_info->sdp->media[0]);
	}

	if (remote_ice_support) {
		ast_log(LOG_DEBUG, "[session:%p] Remote support ICE, clear media_turn_relay flag, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->media_turn_relay = 0;
		if (session->endpoint->aors && !strncmp(session->endpoint->aors, "trunk_", 6)) {
			session->terminal_type = AST_TERMINAL_TYPE_ICE_TRUNK;
			ast_log(LOG_DEBUG, "[session:%p] Remote support ICE, set_terminal_type ICE trunk for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	return;
}

static void set_turn_target_address(struct ast_sip_session *session,
	struct ast_sip_contact *found_contact, const char *location)
{
	RAII_VAR(struct ast_sip_contact *, contact, NULL, ao2_cleanup);

	if (!session) {
		return;
	}

	if (!session->media_turn_relay) {
		return;
	}

	if (found_contact) {
		strncpy(session->turn_target_address, found_contact->original_address, sizeof(session->turn_target_address) - 1);
		ast_log(LOG_WARNING, "[session:%p] Set turn target address:%s, endpoint:%s\n",
				session, session->turn_target_address,ast_sorcery_object_get_id(session->endpoint));
	} else {
		contact = ast_sip_location_retrieve_contact_from_aor_uri(ast_sorcery_object_get_id(session->endpoint), location);
		if (contact) {
			strncpy(session->turn_target_address, contact->original_address, sizeof(session->turn_target_address) - 1);
			ast_log(LOG_WARNING, "[session:%p] Set turn target address:%s, endpoint:%s\n",
				session, session->turn_target_address,ast_sorcery_object_get_id(session->endpoint));
		} else {
			ast_log(LOG_WARNING, "[session:%p] Can't get contact to set turn target address, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	return;
}

static void set_caller_gds_call(struct ast_sip_session* session, pjsip_rx_data* rdata) {
	if (!session || !rdata) {
		ast_log(LOG_ERROR, "[set_caller_gds_call] The session or rdata is nil\n");
		return;
	}

	char useragent[128] = {0};
	const pj_str_t USER_AGENT = { "User-Agent", 10 };
	const char* GDS_USER_AGENT = "Grandstream GDS";
	pjsip_user_agent_hdr* pjsip_ua_hdr = NULL;
	char* is_gds_ua = NULL;

	session->is_gds_call = 0;
	pjsip_ua_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (!pjsip_ua_hdr) {
		ast_log(LOG_NOTICE, "[set_caller_gds_call] The invite request not had User-Agent header\n");
		return;
	}

	ast_copy_string(useragent, pjsip_ua_hdr->hvalue.ptr,
		(pjsip_ua_hdr->hvalue.slen > sizeof(useragent)) ? sizeof(useragent) - 1 : pjsip_ua_hdr->hvalue.slen);
	is_gds_ua = strstr(useragent, GDS_USER_AGENT);
	if (is_gds_ua) {
		session->is_gds_call = 1;
	}

	ast_log(LOG_NOTICE, "[set_caller_gds_call] The session: %s is gds call: %d\n",
		ast_sorcery_object_get_id(session->endpoint), session->is_gds_call);
}

static void set_caller_hikvision_call(struct ast_sip_session* session, pjsip_rx_data* rdata) {
	if (!session || !rdata) {
		ast_log(LOG_ERROR, "[set_caller_gds_call] The session or rdata is nil\n");
		return;
	}

	char useragent[128] = {0};
	const pj_str_t USER_AGENT = { "User-Agent", 10 };
	const char* GDS_USER_AGENT = "YATE/";
	pjsip_user_agent_hdr* pjsip_ua_hdr = NULL;
	//char* is_hik_ua = NULL;

	session->is_hikvision_call = 0;
	pjsip_ua_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (!pjsip_ua_hdr) {
		ast_log(LOG_NOTICE, "[set_caller_gds_call] The invite request not had User-Agent header\n");
		return;
	}

	ast_copy_string(useragent, pjsip_ua_hdr->hvalue.ptr,
		(pjsip_ua_hdr->hvalue.slen > sizeof(useragent)) ? sizeof(useragent) - 1 : pjsip_ua_hdr->hvalue.slen);

	if (!strncasecmp(useragent, GDS_USER_AGENT,5)) {
		session->is_hikvision_call = 1;
	}

	ast_log(LOG_NOTICE, "[set_caller_hikvision_call] The session %p: %s is hikvision call: %d\n",
		session,ast_sorcery_object_get_id(session->endpoint), session->is_hikvision_call);
}

static void set_caller_gsc_call(struct ast_sip_session* session, pjsip_rx_data* rdata) {
	if (!session || !rdata) {
		ast_log(LOG_ERROR, "[set_caller_gsc_call] The session or rdata is nil\n");
		return;
	}

	char useragent[128] = {0};
	const pj_str_t USER_AGENT = { "User-Agent", 10 };
	const char* GDS_USER_AGENT = "Grandstream GSC";
	pjsip_user_agent_hdr* pjsip_ua_hdr = NULL;
	char* is_gsc_ua = NULL;

	session->is_gsc_call = 0;
	pjsip_ua_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (!pjsip_ua_hdr) {
		ast_log(LOG_NOTICE, "[set_caller_gsc_call] The invite request not had User-Agent header\n");
		return;
	}

	ast_copy_string(useragent, pjsip_ua_hdr->hvalue.ptr,
		(pjsip_ua_hdr->hvalue.slen > sizeof(useragent)) ? sizeof(useragent) - 1 : pjsip_ua_hdr->hvalue.slen);
	is_gsc_ua = strstr(useragent, GDS_USER_AGENT);
	if (is_gsc_ua) {
		session->is_gsc_call = 1;
	}

	ast_log(LOG_NOTICE, "[set_caller_gsc_call] The session: %s is gsc call: %d\n",
		ast_sorcery_object_get_id(session->endpoint), session->is_gsc_call);
}

int ast_sip_session_set_device_ua(struct ast_sip_session* session, pjsip_rx_data* rdata, pjsip_method_e method)
{
	size_t size = 0;
	const pj_str_t USER_AGENT = { "User-Agent", 10 };
	pjsip_user_agent_hdr* pjsip_ua_hdr = NULL;

	if (!session || !rdata) {
		ast_log(LOG_ERROR, "[ast_sip_session_set_device_ua] Invalid args.\n");
		return -1;
	}

	memset(session->user_agent, 0, sizeof(session->user_agent));

	pjsip_ua_hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &USER_AGENT, NULL);
	if (!pjsip_ua_hdr) {
		ast_log(LOG_NOTICE, "[ast_sip_session_set_device_ua] The %s method not carry User-Agent header.\n",
			method == PJSIP_INVITE_METHOD ? "INVITE" : "OTHER");
		return -1;
	}

	size = pj_strlen(&pjsip_ua_hdr->hvalue) < sizeof(session->user_agent) ? pj_strlen(&pjsip_ua_hdr->hvalue) : sizeof(session->user_agent);

	memcpy(session->user_agent, pj_strbuf(&pjsip_ua_hdr->hvalue), size);

	if (session->channel){
		ast_channel_device_info_set(session->channel, session->user_agent);
	}

	ast_log(LOG_DEBUG, "[ast_sip_session_set_device_ua] The %s method User-Agent: %s.\n",
		method == PJSIP_INVITE_METHOD ? "INVITE" : "OTHER", session->user_agent);

	return 0;
}

int ast_sip_session_get_device_ua(struct ast_sip_session* session, char* buf, uint64_t size)
{
	if (!session || !buf || size <= 0) {
		ast_log(LOG_ERROR, "[ast_sip_session_set_device_ua] Invalid args.\n");
		return -1;
	}

	memset(buf, 0, size);

	if (ast_strlen_zero(session->user_agent)) {
		ast_log(LOG_DEBUG, "[ast_sip_session_set_device_ua] The session user_agent is nil.\n");
		return -1;
	}

	memcpy(buf, session->user_agent, size);

	ast_log(LOG_DEBUG, "[ast_sip_session_get_device_ua] The get user_agent: %s.\n", buf);

	return 0;
}

static void str_swap(pj_str_t *str1, pj_str_t *str2)
{
	pj_str_t tmp = *str1;
	*str1 = *str2;
	*str2 = tmp;
}

static void set_session_msid_count(struct ast_sip_session *session)
{
	if (msid_count > 10000) {
		msid_count = 1;
	}
	msid_count++;

	session->msid_count = msid_count;

	return;
}

static void set_stream_params_according_exist_stream(struct ast_stream *stream, struct ast_stream *existing_stream)
{
	const char *mslabel = NULL;
	const char *label = NULL;

	if (!stream || !existing_stream) {
		return;
	}

	ast_stream_set_video_resolution(stream, ast_stream_get_video_resolution(existing_stream));

	ast_stream_set_held(stream, ast_stream_get_held(existing_stream));

	ast_stream_set_local_media_direction(stream, ast_stream_get_local_media_direction(existing_stream));

	ast_stream_set_uptime(stream, ast_stream_get_uptime(existing_stream));

	if (ast_stream_get_formats(existing_stream)) {
		ast_stream_set_formats(stream, ast_stream_get_formats(existing_stream));
	}

	mslabel = ast_stream_get_metadata(existing_stream, "MSID:MSLABEL");
	label = ast_stream_get_metadata(existing_stream, "MSID:LABEL");
	if (mslabel) {
		ast_stream_set_metadata(stream, "MSID:MSLABEL", mslabel);
	}
	if (label) {
		ast_stream_set_metadata(stream, "MSID:LABEL", label);
	}

	return;
}

static void set_decline_reason_for_basic_call(struct ast_sip_session *session,
	pjmedia_sdp_media *media, struct ast_stream *stream)
{
	if (!session || !media || !stream) {
		return;
	}

	if (session->is_sfu) {
		return;
	}

	if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
		return;
	}

	if (ast_stream_get_local_media_direction(stream) != SENDRECV) {
		return;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
		if (pjmedia_sdp_attr_find2(media->attr_count, media->attr, "recvonly", NULL) ||
			pjmedia_sdp_attr_find2(media->attr_count, media->attr, "inactive", NULL)) {
			ast_log(LOG_DEBUG, "[session:%p] Remote is wave, reject the video, set decline reason, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_set_video_decline_reason(stream, AST_VIDEO_DECLINE_REASON_PEER_REJECT);
		}
	} else {
		if (ast_sip_session_judge_if_disable_media(media)) {
			ast_log(LOG_DEBUG, "[session:%p] Remote reject the video, set decline reason, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_set_video_decline_reason(stream, AST_VIDEO_DECLINE_REASON_PEER_REJECT);
		}
	}

	return;
}

/* Add by bfdu for add isfocus for re-INVITE */
static int append_isfocus_contact_for_reinvite(pjsip_tx_data *packet)
{
    pjsip_contact_hdr *contact = NULL;
    static const pj_str_t isfocus_str = { "isfocus", 7 };
    static const pj_str_t blank_str = { "", 0 };
    pjsip_param *isfocus = NULL;

    // Sanity check
    if (packet == NULL) {
        ast_log(LOG_WARNING, "TX packet is NULL\n");
        return -1;
    }

    contact = pjsip_msg_find_hdr(packet->msg, PJSIP_H_CONTACT, NULL);
    if (contact == NULL) {
        ast_log(LOG_WARNING, "contact header is NULL\n");
        return -1;
    }

    isfocus = PJ_POOL_ALLOC_T(packet->pool, pjsip_param);
    if (isfocus == NULL) {
        ast_log(LOG_WARNING, "isfocus is NULL\n");
        return -1;
    }

    isfocus->name = isfocus_str;
    isfocus->value = blank_str;
    pj_list_push_back(&contact->other_param, isfocus);

    return 0;
}

static void add_video_fec_codec_to_stream_format(struct ast_sip_session *session, struct ast_stream *stream)
{
	struct ast_format_cap *formats = NULL;

	if (!session || !stream) {
		return;
	}

	if (session->is_webrtc_call && !session->support_gs_fec) {
		ast_log(LOG_DEBUG, "[session:%p] add_video_fec_codec_to_stream_format, wave can't support gs-fec, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	formats = ast_stream_get_formats(stream);
	if (!formats) {
		return;
	}

	if (ast_format_cap_has_video_fec(session->endpoint->media.codecs)) {
		ast_log(LOG_DEBUG, "[session:%p] add_video_fec_codec_to_stream_format endpoint:%s has video FEC codec, add to stream\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_format_cap_add_video_fec(formats);
		/* Now we use alone fec version for video as default */
		ast_stream_set_alone_fec_version(stream, 1);
	}

	return;
}

static void add_nack_to_stream(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!session || !stream) {
		return;
	}

	if (session->endpoint->enable_nack) {
		ast_log(LOG_DEBUG, "[session:%p] add_nack_to_stream endpoint:%s nack is enabled, add to stream\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_has_nack(stream, 1);
		ast_stream_set_has_nack_pli(stream, 1);
		ast_stream_set_has_ccm_fir(stream, 1);
	}

	return;
}

static void add_rtx_codec_to_stream_format(struct ast_sip_session *session, struct ast_stream *stream)
{
	struct ast_format_cap *formats = NULL;

	if (!session || !stream) {
		return;
	}

	formats = ast_stream_get_formats(stream);
	if (!formats) {
		return;
	}

	if (ast_format_cap_has_rtx(session->endpoint->media.codecs)) {
		ast_log(LOG_DEBUG, "[session:%p] add_rtx_codec_to_stream_format endpoint:%s has RTX codec, add to stream\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_format_cap_add_rtx(formats);
	}

	return;
}

/* Add by bfdu for BFCP, 2015.07.21 */
static void init_bfcp_para(struct ast_sip_session *session)
{
	memset(&session->bfcp_para, 0, sizeof(struct ast_bfcp_para));

	return;
}

static void init_yealink_para(struct ast_sip_session *session)
{
	session->is_yealink_msg = FALSE;

	return;
}

static void init_modem_and_ecan_para(struct ast_sip_session *session)
{
	session->audio_has_modem = 0;
	memset(&session->audio_ecan_param, 0, ECAN_PARAM_MAX_LEN);

	return;
}

static void init_session_param(struct ast_sip_session *session)
{
	if (!session) {
		ast_log(LOG_DEBUG, "init_session_param session is NULL\n");
		return;
	}

	session->is_outgoing = 0;
	session->recv_reinvite_no_sdp = 0;
	session->is_media_count_change = 0;
	init_bfcp_para(session);
	init_yealink_para(session);
	init_modem_and_ecan_para(session);

	return;
}

unsigned char judge_if_support_video_media(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	unsigned char res = FALSE;
	unsigned int i = 0;
	unsigned int has_video = 0;
	char media_type[16] = {0};
	const pjmedia_sdp_media *om = NULL;

	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (offer) {
		for (i = 0; i < offer->media_count; ++i) {
			om = offer->media[i];
			memset(media_type, 0, sizeof(media_type));
			ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));
			if (strncmp(media_type, "video", 5) == 0 && om->desc.port != 0) {
				has_video = 1;
				break;
			}
		}

		if (!has_video) {
			ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s receive offer has no video cap, don't need check endpoint video\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return TRUE;
		}
	}

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "[session:%p] Endpoint:%s failed to allocate capabilities\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		res = FALSE;
		return res;
	}

	ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);

	if (!ast_format_cap_count(caps)) {
		ast_log(LOG_WARNING, "[session:%p] Endpoint:%s has no video cap\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		res = FALSE;
	} else {
		res = TRUE;
	}

	return res;
}

#if 0
static void add_bandw_ct_to_sdp(pjsip_inv_session *inv, struct ast_sip_session *session,
	const pjmedia_sdp_session *offer, pjmedia_sdp_session *local)
{
	unsigned int maxbit= 0;
	const pj_str_t STR_BANDW_MODIFIER = { "CT", 2 };
	pjmedia_sdp_bandw *b = NULL;

	if (!inv || !session || !local) {
		return;
	}

	if (!session->endpoint->media.support_special_attrs) {
		ast_log(LOG_DEBUG, "[session:%p] endpoint:%s support_special_attrs configure is disabled, don't add CT value\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!judge_if_support_video_media(session, offer)) {
		ast_log(LOG_DEBUG, "[session:%p] Offer or endpoint:%s has no video, don't add CT value\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	maxbit = ast_sip_get_maxcallbitrate();
	if (!maxbit){
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s can't get maxbit, don't add CT value\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	b = PJ_POOL_ALLOC_T(inv->pool_prov, pjmedia_sdp_bandw);
	if (!b) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s can't alloc bandw, don't add CT value\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s add_bandw_ct_to_sdp maxbit:%d\n", 
		session, ast_sorcery_object_get_id(session->endpoint), maxbit);

	b->modifier = STR_BANDW_MODIFIER;
	b->value = maxbit;
	local->bandw[local->bandw_count++] = b;

	return;
}
#endif

static unsigned char judge_if_bridge_exsit(struct ast_channel* chan)
{
	if (NULL == chan){
		ast_log(LOG_WARNING, "Judge if both type is sip, chan is NULL\n");
		return FALSE;
	}
	RAII_VAR(struct ast_channel *, bridged, ast_channel_bridge_peer(chan), ast_channel_cleanup);
	if (NULL == bridged) {
		ast_log(LOG_DEBUG, "Judge if both type is sip, bridge is not exsit\n");
		return FALSE;
	}

	return TRUE;
}

unsigned char ast_sip_session_judge_if_type_is_sip(const struct ast_channel* chan)
{
	if (NULL == chan){
		ast_log(LOG_WARNING, "Judge if both type is sip, chan is NULL\n");
		return FALSE;
	}

	if (ast_channel_tech(chan) && (0 == strcasecmp(ast_channel_tech(chan)->type, "PJSIP"))) {
		return TRUE;
	} else {
		ast_log(LOG_DEBUG, "Judge if type is sip, type is not PJSIP\n");
		return FALSE;
	}
}

unsigned char ast_sip_session_judge_if_type_is_local(const struct ast_channel* chan)
{
	if (NULL == chan){
		ast_log(LOG_WARNING, "Judge if channel type is local, chan is NULL\n");
		return FALSE;
	}

	if (ast_channel_tech(chan) && (0 == strcasecmp(ast_channel_tech(chan)->type, "Local"))) {
		ast_log(LOG_DEBUG, "channel:%s type is local\n", ast_channel_name(chan));
		return TRUE;
	} else {
		return FALSE;
	}
}

unsigned char ast_sip_session_judge_if_type_is_rtsp(const struct ast_channel* chan)
{
	if (NULL == chan){
		ast_log(LOG_WARNING, "Judge if channel type is rtsp, chan is NULL\n");
		return FALSE;
	}

	if (ast_channel_tech(chan) && (0 == strcasecmp(ast_channel_tech(chan)->type, "RTSP"))) {
		ast_log(LOG_DEBUG, "channel:%s type is rtsp\n", ast_channel_name(chan));
		return TRUE;
	} else {
		return FALSE;
	}
}

unsigned char ast_sip_session_support_multiple_video(const struct ast_sip_session *session)
{
	if (!session) {
		return FALSE;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_GVC ||
		session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ||
		ast_sip_session_media_use_bundle(session) ||
		session->is_webrtc_call) {
		return TRUE;
	}

	return FALSE;
}

unsigned char ast_sip_session_support_bfcp(const struct ast_sip_session *session)
{
	RAII_VAR(struct ast_format_cap *, video_caps, NULL, ao2_cleanup);

	if (!session) {
		return FALSE;
	}

	if (!session->endpoint->media.bfcp_support) {
		ast_log(LOG_ERROR, "[session:%p] It's not enable BFCP media support, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	if (ast_sip_session_get_only_support_audio()) {
		return FALSE;
	}

	if (!(video_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "[session:%p] Failed to allocate video capabilities, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	ast_format_cap_append_from_cap(video_caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_VIDEO);

	if (judge_if_invalid_caps(video_caps)) {
		return FALSE;
	}

	if (session->support_ice ||
		session->is_webrtc_call ||
		session->terminal_type == AST_TERMINAL_TYPE_WEBRTC ||
		session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE ||
		session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) {
		return FALSE;
	}

	return TRUE;
}

static void get_codecs(struct ast_sip_session *session, const struct pjmedia_sdp_media *stream, struct ast_rtp_codecs *codecs)
{
	pjmedia_sdp_attr *attr;
	pjmedia_sdp_rtpmap *rtpmap;
	pjmedia_sdp_fmtp fmtp;
	struct ast_format *format;
	int i, num = 0;
	char name[256];
	char media[20];
	char fmt_param[256];

	ast_rtp_codecs_payloads_initialize(codecs);

	/* Iterate through provided formats */
	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* The payload is kept as a string for things like t38 but for video it is always numerical */
		ast_rtp_codecs_payloads_set_m_type(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]));
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(session->inv_session->pool_prov, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(name, &rtpmap->enc_name, sizeof(name));
		ast_copy_pj_str(media, (pj_str_t*)&stream->desc.media, sizeof(media));
		ast_rtp_codecs_payloads_set_rtpmap_type_rate(codecs, NULL, pj_strtoul(&stream->desc.fmt[i]),
							     media, name, 0, rtpmap->clock_rate);
		/* Look for an optional associated fmtp attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "fmtp", &rtpmap->pt))) {
			continue;
		}

		if ((pjmedia_sdp_attr_get_fmtp(attr, &fmtp)) == PJ_SUCCESS) {
			ast_copy_pj_str(fmt_param, &fmtp.fmt, sizeof(fmt_param));
			if (sscanf(fmt_param, "%30d", &num) != 1) {
				continue;
			}

			if ((format = ast_rtp_codecs_get_payload_format(codecs, num))) {
				struct ast_format *format_parsed;

				ast_copy_pj_str(fmt_param, &fmtp.fmt_param, sizeof(fmt_param));

				format_parsed = ast_format_parse_sdp_fmtp(format, fmt_param);
				if (format_parsed) {
					ast_rtp_codecs_payload_replace_format(codecs, num, format_parsed);
					ao2_ref(format_parsed, -1);
				}

				ao2_ref(format, -1);
			}
		}
	}

	/* Get the packetization, if it exists */
	if ((attr = pjmedia_sdp_media_find_attr2(stream, "ptime", NULL))) {
		unsigned long framing = pj_strtoul(pj_strltrim(&attr->value));
		if (framing && session->endpoint->media.rtp.use_ptime) {
			ast_rtp_codecs_set_framing(codecs, framing);
		}
	}
}

static enum ast_stream_media_direction get_sdp_media_direction(const pjmedia_sdp_media *stream)
{
    enum ast_stream_media_direction media_direction = SENDRECV;

	if (pjmedia_sdp_media_find_attr2(stream, "sendrecv", NULL)) {
		media_direction = SENDRECV;
	} else if (pjmedia_sdp_media_find_attr2(stream, "sendonly", NULL)) {
		media_direction = SENDONLY;
	} else if (pjmedia_sdp_media_find_attr2(stream, "inactive", NULL)) {
		media_direction = INACTIVE;
	} else if (pjmedia_sdp_media_find_attr2(stream, "recvonly", NULL)) {
		media_direction = RECVONLY;
	} else {
		ast_log(LOG_WARNING, "Can't find media direction attr, use default value(sendrecv)\n");
	}

	return media_direction;
}

/* compare framerate, if not the same, return 0, if is the same, return 1 */
static unsigned int compare_media_framerate(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer)
{
	unsigned int res = 0;
	unsigned int pre_framerate = 0;
	unsigned int offer_framerate = 0;
	pjmedia_sdp_attr *pre_attr = NULL;
	pjmedia_sdp_attr *offer_attr = NULL;

	if ((pre_attr = pjmedia_sdp_media_find_attr2(preanswer, "framerate", NULL))) {
		pre_framerate = pj_strtoul(pj_strltrim(&pre_attr->value));
	}

	if ((offer_attr = pjmedia_sdp_media_find_attr2(offer, "framerate", NULL))) {
		offer_framerate = pj_strtoul(pj_strltrim(&offer_attr->value));
	}

	if (pre_framerate != offer_framerate) {
		res = 0;
	} else {
		res = 1;
	}

	return res;
}

/* When we compare media direction with offer sdp and active local sdp, can't compare directly
 * local:sendonly is according offer:recvonly.
 * local:recvonly is according offer:sendonly.
 */
static unsigned int compare_media_direction(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer)
{
	unsigned int res = 0;
	enum ast_stream_media_direction preanswer_direction = get_sdp_media_direction(preanswer);
	enum ast_stream_media_direction offer_direction = get_sdp_media_direction(offer);

	switch (preanswer_direction) {
		case SENDRECV:
			if (offer_direction == SENDRECV) {
				res = 1;
			} else {
				res = 0;
			}
			break;
		case SENDONLY:
			if (offer_direction == RECVONLY) {
				res = 1;
			} else {
				res = 0;
			}
			break;
		case RECVONLY:
			if (offer_direction == SENDONLY) {
				res = 1;
			} else {
				res = 0;
			}
			break;
		case INACTIVE:
			if (offer_direction == INACTIVE) {
				res = 1;
			} else {
				res = 0;
			}
			break;
		default:
			break;
	}

	return res;
}

/* return TRUE is the same, FALSE IS NOT the same */
static unsigned char compare_static_payload_codec(const pjmedia_sdp_media *offer, unsigned preanswer_pt, unsigned position)
{
	unsigned offer_pt;

	if (position < offer->desc.fmt_count) {
		offer_pt = pj_strtoul(&offer->desc.fmt[position]);
		if (preanswer_pt == offer_pt) {
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		return FALSE;
	}
}

static unsigned char compare_codec_format(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer,
	struct ast_sip_session *session, unsigned position)
{
	unsigned preanswer_pt = 0;
	unsigned offer_pt = 0;
	struct ast_rtp_codecs pre_codecs = AST_RTP_CODECS_NULL_INIT;
	struct ast_rtp_codecs offer_codecs = AST_RTP_CODECS_NULL_INIT;
	struct ast_format *pre_format = NULL;
	struct ast_format *offer_format = NULL;

	get_codecs(session, preanswer, &pre_codecs);
	get_codecs(session, offer, &offer_codecs);

	preanswer_pt = pj_strtoul(&preanswer->desc.fmt[position]);
	offer_pt = pj_strtoul(&offer->desc.fmt[position]);

	pre_format = ast_rtp_codecs_get_payload_format(&pre_codecs, preanswer_pt);
	offer_format = ast_rtp_codecs_get_payload_format(&offer_codecs, offer_pt);
	if (pre_format && offer_format) {
		if (AST_FORMAT_CMP_NOT_EQUAL == ast_format_cmp(pre_format, offer_format)) {
			ast_log(LOG_DEBUG, "[session:%p] preanswer and offer format are not equal for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(pre_format, -1);
			ao2_ref(offer_format, -1);
			ast_rtp_codecs_payloads_destroy(&pre_codecs);
			ast_rtp_codecs_payloads_destroy(&offer_codecs);
			return FALSE;
		} else {
			ao2_ref(pre_format, -1);
			ao2_ref(offer_format, -1);
		}
	}

	ast_rtp_codecs_payloads_destroy(&pre_codecs);
	ast_rtp_codecs_payloads_destroy(&offer_codecs);

	return TRUE;
}

/* return TRUE is the same, FALSE IS NOT the same */
static unsigned char compare_dynamic_payload_codec(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer,
	struct ast_sip_session *session, unsigned position)
{
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap pre_rtpmap;
	pjmedia_sdp_rtpmap offer_rtpmap;

	/* Get the rtpmap for the payload type in the master. */
	attr = pjmedia_sdp_media_find_attr2(preanswer, "rtpmap", &preanswer->desc.fmt[position]);
	if (!attr) {
		pj_assert(!"Bug! Offer should have been validated");
		return TRUE;
	}

	pjmedia_sdp_attr_get_rtpmap(attr, &pre_rtpmap);

	if (position >= offer->desc.fmt_count) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s dynamic payload preanswer(%d) and offer(%d) is media fmt num is not the same\n",
			session, ast_sorcery_object_get_id(session->endpoint), preanswer->desc.fmt_count, offer->desc.fmt_count);
		return FALSE;
	}

	attr = pjmedia_sdp_media_find_attr2(offer, "rtpmap", &offer->desc.fmt[position]);
	if (!attr) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get rtpmap attr, maybe some wrong, don't send to peer for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return TRUE;
	}

	pjmedia_sdp_attr_get_rtpmap(attr, &offer_rtpmap);
	if (!(!pj_stricmp(&pre_rtpmap.enc_name, &offer_rtpmap.enc_name) &&
		pre_rtpmap.clock_rate == offer_rtpmap.clock_rate &&
		(pj_stricmp(&pre_rtpmap.param, &offer_rtpmap.param)==0 ||
		(offer_rtpmap.param.slen==0 && pre_rtpmap.param.slen==1 && *pre_rtpmap.param.ptr=='1') ||
		(pre_rtpmap.param.slen==0 && offer_rtpmap.param.slen==1 && *offer_rtpmap.param.ptr=='1')))) {
		/* Not Match! */
		ast_log(LOG_DEBUG, "[session:%p] rtpmap param not match for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	if (!compare_codec_format(preanswer, offer, session, position)) {
		return FALSE;
	}

	return TRUE;
}

/* return TRUE is the same, FALSE IS NOT the same */
static unsigned char compare_meida_codec(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer,
	struct ast_sip_session *session)
{
	unsigned i = 0;
	unsigned preanswer_pt = 0;

	for (i = 0; i < preanswer->desc.fmt_count; ++i) {
		if (pj_isdigit(*preanswer->desc.fmt[i].ptr)) {
			/* This is normal/standard payload type, where it's identified
			 * by payload number.
			 */
			preanswer_pt = pj_strtoul(&preanswer->desc.fmt[i]);

			/* if fmt's value\order\num is not same, send reinvite to peer */
			if (preanswer_pt < 96) {
				if (!compare_static_payload_codec(offer, preanswer_pt, i)) {
					ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s compare static payload is not the same for preanswer pt:%d, postion:%d\n",
						session, ast_sorcery_object_get_id(session->endpoint), preanswer_pt, i);
					return FALSE;
				}
			} else {
				if (!compare_dynamic_payload_codec(preanswer, offer, session, i)) {
					return FALSE;
				}
			}
		} else {
			ast_log(LOG_DEBUG, "[session:%p] payload is not digit, support later for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return TRUE;
		}
	}

	return TRUE;
}

static int compare_preanswer_offer_media_line(struct ast_sip_session *session,
	const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer)
{
	/* if audio or video turn on/off SRTP, send to peer */
	if (pj_strcmp(&preanswer->desc.transport, &offer->desc.transport) != 0) {
		ast_log(LOG_DEBUG, "[session:%p] Transport is not the same, need to send to peer for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	if (pj_strcmp(&offer->desc.media, &preanswer->desc.media)) {
		ast_log(LOG_DEBUG, "[session:%p] Media type is not the same, need to send to peer for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return FALSE;
	}

	if ((offer->desc.port == 0 && preanswer->desc.port != 0) || (offer->desc.port != 0 && preanswer->desc.port == 0)) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s, compare_preanswer_offer_sdp, offer port %d, preanswer port %d\n",
			session, ast_sorcery_object_get_id(session->endpoint), offer->desc.port, preanswer->desc.port);
		return FALSE;
	}

	return TRUE;
}

/* Now we just need to compare audio and video media */
static unsigned char media_type_need_compare(const pjmedia_sdp_media *preanswer)
{
	char media_type[16] = {0};

	memset(media_type, 0, sizeof(media_type));
	ast_copy_pj_str(media_type, &preanswer->desc.media, sizeof(media_type));
    if (!strncmp(media_type, "audio", 5) || !strncmp(media_type, "video", 5)) {
		return TRUE;
    }

    return FALSE;
}

static unsigned char compare_preanswer_offer_media_attrs(const pjmedia_sdp_media *preanswer, const pjmedia_sdp_media *offer,
				struct ast_sip_session *session, const char *media_type,
				const pjmedia_sdp_session *media_session)
{
	struct ast_sip_session_media *session_media = NULL;

	session_media = session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO];
	if (!session_media) {
		ast_log(LOG_WARNING, "[session:%p] compare_preanswer_offer_sdp, session_media is NULL for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return TRUE;
	}

	if (!media_type_need_compare(preanswer)) {
		return TRUE;
	}

	/* need to compare video framerate, if change, reinvite should send to peer */
	if (!compare_media_framerate(preanswer, offer)) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s compare media framerate is not the same for media type:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint), media_type);
		return FALSE;
	}

	if (!compare_media_direction(preanswer, offer)) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s compare media direction is not the same for media type:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint), media_type);
		return FALSE;
	}

	if (!compare_meida_codec(preanswer, offer, session)) {
		return FALSE;
	}

	return TRUE;
}

static unsigned char compare_preanswer_offer_sdp(struct ast_sip_session *session,
				  const pjmedia_sdp_session *offer,
				  const pjmedia_sdp_session *initial)
{
	unsigned char i = 0;
    char media_type[16] = {0};
    pjmedia_sdp_media *om = NULL;	/* offer */
	pjmedia_sdp_media *im = NULL;	/* initial media */

	if (initial->media_count != offer->media_count) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s initial and offer's media_count is not same\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->is_media_count_change = 1;
		return FALSE;
	}

	/* For each media line, create our answer based on our initial
     * capability.
     */
    for (i = 0; i < offer->media_count; ++i) {
		om = offer->media[i];
		if (!om) {
			ast_log(LOG_WARNING, "[session:%p] Offer media is NULL, something is very wrong, don't send to peer for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return TRUE;
		}

		if (i >= initial->media_count) {
			ast_log(LOG_DEBUG, "[session:%p] Offer's media_count is larger than initial for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return FALSE;
		}

		im = initial->media[i];
		if (!im) {
			ast_log(LOG_WARNING, "[session:%p] Initial media is NULL, something is very wrong, don't send to peer for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return TRUE;
		}

		if (FALSE == compare_preanswer_offer_media_line(session, im, om)) {
			return FALSE;
		}

		if (FALSE == compare_preanswer_offer_media_attrs(im, om, session, media_type, offer)) {
			return FALSE;
		}
    }

    return TRUE;
}

static int judge_if_hold_unhold_request(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	int i = 0;
	int res = 0;
	pjmedia_sdp_media *om = NULL;
	struct ast_sip_session_media *audio_media = NULL;

	if (!session || !offer) {
		ast_log(LOG_WARNING, "Change ssrc for hold, param is invalid\n");
		return 0;
	}

	audio_media = session->active_media_state->default_session[AST_MEDIA_TYPE_AUDIO];

	for (i = 0; i < offer->media_count; ++i) {
		om = offer->media[i];
		if (!om) {
			continue;
		}

		if (strncmp(om->desc.media.ptr, "audio", 5)) {
			ast_log(LOG_DEBUG, "[session:%p] It's not audio media, just continue for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		if (pjmedia_sdp_media_find_attr2(om, "sendonly", NULL) ||
			pjmedia_sdp_media_find_attr2(om, "inactive", NULL)) {
			ast_log(LOG_DEBUG, "[session:%p] It's a hold request for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			res = 1;
		}

		/* We just use audio media to judge hold/unhold request. */
		break;
	}

	if (audio_media && audio_media->remotely_held) {
		ast_log(LOG_DEBUG, "[session:%p] It's a unhold request for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		res = 1;
	}

	return res;
}

static int judge_if_conference(struct ast_sip_session *session)
{
	const char *conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");

	if (!ast_strlen_zero(conf_room)) {
		return TRUE;
	}

	return FALSE;
}

static unsigned int judge_if_ice_restart(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
		if (session_media && session_media->ice_restart) {
			/* If this reinvite is mix ice restart and other action, will have some abnormal */
			ast_log(LOG_DEBUG, "[session:%p] Find ice restart media, don't need forward to peer, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		}
	}

	return 0;
}

static int judge_if_need_send_reinvite_to_peer(struct ast_sip_session *session,
				  const pjmedia_sdp_session *offer,
				  const pjmedia_sdp_session *initial)
{
    int res = 0;
	unsigned int allow_forward_hold_request = 0;

	if (session->is_direct_media || session->joint_ipvt) {
		ast_log(LOG_DEBUG, "[session:%p] Enable direct media, need forward to peer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	/*Start added by zyxia for faxsending, and the peer disable t38, Bug 186503  2021.1.4*/
	if (session->is_fax_sending_request || (session->channel && ast_channel_get_is_send_fax(session->channel))) {
		ast_log(LOG_DEBUG, "[session:%p]It's faxsending channel,no forward to peer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}
	/*End added by zyxia for faxsending, and the peer disable t38,2021.1.4*/

	/*Start reinvited dont send to ipc ,205312 ,2022.6.7*/
	if (session->is_rtsp_call) {
		ast_log(LOG_DEBUG, "[session:%p]It's rtsp channel,no forward to peer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}
	/*End reinvited dont send to ipc ,205312 ,2022.6.7*/

	allow_forward_hold_request = ast_sip_session_get_allow_forward_hold_request();
	if (!allow_forward_hold_request && judge_if_hold_unhold_request(session, offer)) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable allow_forward_hold_request, and it's a hold/unhold request, not forward to peer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/* Added by yfyang, for Bug 181487 */
	const char *record_vm = pbx_builtin_getvar_helper(session->channel, "RECORD_VMENU");
	if (!ast_strlen_zero(record_vm) && !strncmp(record_vm, "ON", 2)) {
		ast_log(LOG_DEBUG, "[session:%p] It's a RECORD_VMENU call, not forward to peer\n", session);
		return 0;
	}

	if (FALSE == judge_if_support_video_media(session, offer)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't support video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (FALSE == judge_if_bridge_exsit(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get bridge, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (TRUE == judge_if_conference(session))  {
		ast_log(LOG_DEBUG, "[session:%p] It's in conference, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (judge_if_ice_restart(session)) {
		return 0;
	}

	if (ast_channel_get_opening_slides_video(session->channel) ||
		ast_channel_get_send_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] It's re-invite for open slides video, need forward to peer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	if (FALSE == compare_preanswer_offer_sdp(session, offer, initial)) {
		ast_log(LOG_DEBUG, "[session:%p] It's compare_preanswer_offer_sdp, need forward to peer, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

    return res;
}

static int judge_if_sdp_has_image(const pjmedia_sdp_session *offer)
{
    unsigned i = 0;
    const pjmedia_sdp_media *om;	/* offer */
    char media_type[16] = {0};

    for (i = 0; i < offer->media_count; ++i) {
		om = offer->media[i];
		if (!om) {
			continue;
		}
		memset(media_type, 0, sizeof(media_type));
		ast_copy_pj_str(media_type, &om->desc.media, sizeof(media_type));
	    if (strncmp(media_type, "image", 5) == 0) {
			ast_log(LOG_DEBUG, "offer sdp has image\n");
			return 1;
	    }
    }

    return 0;
}

static void notify_sfu_bfcp_for_sip(struct ast_sip_session *session)
{
	struct ast_sip_session_media *bfcp_media = NULL;

	bfcp_media = ast_sip_session_get_bfcp_active_media(session);
	if (!bfcp_media) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no bfcp media\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (bfcp_media->bfcp_udp) {
		bfcp_media->bfcp_udp->bfcp_info.primitive = BFCP_FLOOR_STATUS;
		bfcp_media->bfcp_udp->bfcp_info.status = BFCP_GRANTED;
	} else if(bfcp_media->bfcp_tcp) {
		bfcp_media->bfcp_tcp->bfcp_info.primitive = BFCP_FLOOR_STATUS;
		bfcp_media->bfcp_tcp->bfcp_info.status = BFCP_GRANTED;
	}

	if (bfcp_media->bfcp_udp) {
		ast_send_bfcp_msg(bfcp_media->bfcp_udp);
	} else if(bfcp_media->bfcp_tcp) {
		ast_send_bfcp_tcp_msg(bfcp_media->bfcp_tcp);
	}

	return;
}

static void set_bfcp_info_slides_type(struct ast_sip_session *session, char *share_type_str, int share_type)
{
	if (!session || !share_type_str) {
		return;
	}

	if (share_type == BFCP_INFO_SHARE_TYPE_UNKNOW) {
		strcpy(share_type_str, "screen");
	} else {
		strcpy(share_type_str, bfcp_info_share_type_strmap[share_type][0]);
	}

	ast_log(LOG_DEBUG, "[session:%p] set_bfcp_info_slides_type share_type:%d convert to str:%s, endpoint:%s\n",
		session, share_type, share_type_str, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void notify_sfu_bfcp_for_webrtc(struct ast_sip_session *session,
	const char *conf_room, const char *slider, int share_type)
{
	struct ast_bfcp_info bfcp_info;
	struct bfcp_info_data *bfcp_data = NULL;
	RAII_VAR(char *, slider_name, NULL, ast_free);

	memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));

	strncpy(bfcp_info.info_conf_id, conf_room, sizeof(bfcp_info.info_conf_id) - 1);
	strncpy(bfcp_info.floor_owner_id, slider, sizeof(bfcp_info.floor_owner_id) - 1);
	strncpy(bfcp_info.account_id, slider, sizeof(bfcp_info.account_id) - 1);
	slider_name = ast_conf_get_slider_name(conf_room);
	if (!ast_strlen_zero(slider_name)) {
		strncpy(bfcp_info.slider_name, slider_name, sizeof(bfcp_info.slider_name) - 1);
	}

	set_bfcp_info_slides_type(session, bfcp_info.share_type, share_type);

	bfcp_info.primitive = BFCP_FLOOR_STATUS;
	bfcp_info.status = BFCP_GRANTED;

	bfcp_data = ast_sip_session_bfcp_data_alloc(session, &bfcp_info);
	if (bfcp_data) {
		if (ast_sip_push_task(session->serializer, ast_sip_session_bfcp_send_conference_sip_info, bfcp_data)) {
			ao2_cleanup(bfcp_data);
		}
	}

	return;
}

static void notify_granted_bfcp_for_transfer(struct ast_sip_session *session)
{
	struct ast_bfcp_info bfcp_info;

	if (!session || !session->inv_session) {
		return;
	}

	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		ast_log(LOG_DEBUG, "[session:%p] notify_granted_bfcp_for_transfer is already in confirmed state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	memset(&bfcp_info, 0, sizeof(struct ast_bfcp_info));
	bfcp_info.primitive = BFCP_FLOOR_STATUS;

	if (ast_channel_get_recv_slides_video(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Notify to recv slides video for transfer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		bfcp_info.status = BFCP_GRANTED;
		ast_indicate_data(session->channel, AST_CONTROL_BFCP_CONTROL, &bfcp_info, sizeof(struct ast_bfcp_info));
	}

	return;
}

static void notify_topogly_for_sfu_local_invite(struct ast_sip_session *session)
{
	/* Todo: Reserve local channel process for SFU, need avs engine to support later */
	return;

	struct ast_stream_topology *topology = NULL;
	struct ast_channel_topology_data topology_data;

	if (!session || !session->channel) {
		return;
	}

	if (!session->sfu_invite_by_local_chan) {
		return;
	}

	topology = ast_stream_topology_clone(ast_channel_get_stream_topology(session->channel));
	if (topology) {
		topology_data.topology = topology;
		ast_log(LOG_DEBUG, "[session:%p] notify_topogly_for_sfu_local_invite for endpoint:%s, topology stream count:%d\n",
			session, ast_sorcery_object_get_id(session->endpoint), ast_stream_topology_get_count(topology));
		if (ast_queue_control_data(session->channel, AST_CONTROL_NOTIFY_TOPOLOY, &topology_data, sizeof(struct ast_channel_topology_data))) {
			ast_log(LOG_ERROR, "[session:%p] Can't queue AST_CONTROL_NOTIFY_TOPOLOY frame for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_topology_free(topology);
			topology = NULL;
		}
	}

	return;
}

static void notify_nat_call(struct ast_sip_session *session)
{
	struct avs_turn_session_count_string count_string;

	if (!session || !session->channel) {
		return;
	}

	if (!session->ssh_proxy_flag) {
		ast_log(LOG_DEBUG, "[session:%p] It has no ssh proxy flag, don't notify nat call for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_channel_get_turn_session_count(session->channel) > 0) {
		ast_log(LOG_DEBUG, "[session:%p] Already notify nat call for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Notify nat call for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	memset(&count_string, 0, sizeof(struct avs_turn_session_count_string));

	/* Now we only count as channel, don't need count as media, just set one count */
	snprintf(count_string.turn_session_count, sizeof(count_string.turn_session_count), "%d", 1);
	count_string.ssh_proxy_flag = 1;

	ast_indicate_data(session->channel, AST_CONTROL_UPDATE_TURN_SESSION_COUNT, &count_string, sizeof(struct avs_turn_session_count_string));

	return;
}

static void negotiated_for_transfer(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	/* This for transfer target send bfcp for receive slides video */
	notify_granted_bfcp_for_transfer(session);

	/* This for transferee, after unhold to get peer session */
	if (ast_channel_get_feature_transfer(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] It's feature transfer, get peer session for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_set_feature_transfer(session->channel, 0);
		ast_queue_control(session->channel, AST_CONTROL_GET_PEER_SESSION);
	}

	/* This for transferer, after answer to send session to peer */
	if (ast_channel_get_transferer_hanguped(session->channel)) {
		ast_log(LOG_DEBUG, "[session:%p] Transfer hanguped before transfer target answered, indicate transferee complete for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_set_transferer_hanguped(session->channel, 0);
		ast_queue_control(session->channel, AST_CONTROL_TRANSFER_COMPLETE);
	}

	return;
}

int ast_sip_get_extension_from_channel_name(const char *channel_name, char *buf, int size)
{
	const char *start = NULL;
	const char *end =  NULL;

	memset(buf, 0, size);

	if (strncasecmp(channel_name, "PJSIP/", 6)) {
		return -1;
	}
	start = channel_name + 6;
	end = strrchr(start, '-');
	if (!end) {
		return -1;
	}

	strncpy(buf, start, size > (end - start) ? (end - start) : size);
	return 0;
}

/*Start added by zyxia for Bug 218922,2022.3.24,if session dont have slides stream,dont need send bfcp info*/
int ast_sip_session_if_need_send_bfcp_info(struct ast_sip_session *session, char *share_type)
{
	struct ast_stream *slides_video = NULL;
	if (!session) {
		return 0;
	}

	if (session->support_whiteboard && share_type &&
		(!strcasecmp(share_type, "whiteboard") || !strcasecmp(share_type, "document"))) {
		ast_log(LOG_DEBUG, "[session:%p] Support whiteboard and share type is whiteboard, need send INFO, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	if (session->is_sfu) {
		slides_video = ast_sip_session_get_active_slides_video_stream(session);
		if (!slides_video) {
			ast_log(LOG_DEBUG, "[session:%p] done have slide video stream, don't need do process bfcp info endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 0;
		}

		if (ast_stream_get_state(slides_video) == AST_STREAM_STATE_REMOVED ||
			ast_stream_get_state(slides_video) == AST_STREAM_STATE_INACTIVE ) {
			ast_log(LOG_DEBUG, "[session:%p] done have slide video stream, don't need do process bfcp info endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 0;
		}
	}
	return 1;
}
/*End added by zyxia for Bug 218922,2022.3.24*/

static unsigned int send_by_bfcp(struct ast_sip_session *session, int slider_type)
{
	if (session->ucm_trunk) {
		return 0;
	}

	if (session->support_whiteboard &&
		((slider_type == BFCP_INFO_SHARE_TYPE_WHITEBOARD) || (slider_type == BFCP_INFO_SHARE_TYPE_DOCUMENT))) {
		return 0;
	}

	return 1;
}

void ast_sip_session_notify_sfu_bfcp_for_join_conference(struct ast_sip_session *session)
{
	char *slider = NULL;
	char slider_extension[AST_MAX_EXTENSION] = { 0 };
	RAII_VAR(struct ast_channel *, slider_channel, NULL, ao2_cleanup);
	RAII_VAR(char *, msid, NULL, ast_free);
	int slider_type = 0;
	char slider_type_str[16] = { 0 };

	if (!session || !session->channel) {
		ast_log(LOG_DEBUG, "Notify BFCP, has no session\n");
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	if (!session->active_media_state) {
		ast_log(LOG_DEBUG, "[session:%p] Notify BFCP, endpoint:%s has no active_media_state\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	const char *conf_room = pbx_builtin_getvar_helper(session->channel, "CONF_ROOM");
	if (ast_strlen_zero(conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] Notify BFCP, endpoint:%s can't get conf room number\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	slider_type = ast_conf_get_slides_type(conf_room);
	if (slider_type < 0){
		ast_log(LOG_DEBUG, "slider_type %d error\n", slider_type);
		return;
	}
	
	strcpy(slider_type_str, bfcp_info_share_type_strmap[slider_type][0]);
	if (!ast_sip_session_if_need_send_bfcp_info(session, slider_type_str)) {
		ast_log(LOG_DEBUG, "[session:%p] Notify BFCP, endpoint:%s has no active slide stream\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	slider = ast_conf_get_slider(conf_room);
	if (!slider) {
		ast_log(LOG_DEBUG, "[session:%p] Notify BFCP, endpoint:%s can't get slider, don't need send bfcp\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	if (!strlen(slider)) {
		ast_free(slider);
		return;
	}

	if (session->replace_channel_id && !strcasecmp(slider, session->replace_channel_id)) {
		ast_log(LOG_DEBUG, "[session:%p] Notify BFCP, endpoint:%s is the same as replace channel, don't need send bfcp\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_free(slider);
		return;
	}

	slider_channel = ast_conf_get_slider_channel(conf_room);
	if (slider_channel &&
		((slider_type != BFCP_INFO_SHARE_TYPE_WHITEBOARD && slider_type != BFCP_INFO_SHARE_TYPE_DOCUMENT) ||
		!session->support_whiteboard)) {
		msid = ast_channel_get_slides_video_local_msid(slider_channel);
		if (!ast_strlen_zero(msid)) {
			ast_indicate_data(session->channel, AST_CONTROL_NOTIFY_SLIDER_MSID, msid, strlen(msid) + 1);
		}
	}

	ast_sip_get_extension_from_channel_name(slider, slider_extension, sizeof(slider_extension));

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s get slider number:%s for room:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint), slider_extension, conf_room);

	switch (session->terminal_type) {
		case AST_TERMINAL_TYPE_SIP:
		case AST_TERMINAL_TYPE_GVC:
			if (send_by_bfcp(session, slider_type)) {
				notify_sfu_bfcp_for_sip(session);
				break;
			} else {
				ast_log(LOG_DEBUG, "[session:%p] Trunk is UCM, fall through to use SIP INFO, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
			}
		case AST_TERMINAL_TYPE_WEBRTC:
		case AST_TERMINAL_TYPE_ICE_TRUNK:
		case AST_TERMINAL_TYPE_GSWAVE_MOBILE:
			notify_sfu_bfcp_for_webrtc(session, conf_room, slider_extension, slider_type);
			break;
		default:
			break;
	}

	ast_free(slider);
	slider = NULL;

	return;
}

static struct ast_sip_session_media *get_active_main_video_media(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
		return session_media;
	}

	return NULL;
}

static struct ast_sip_session_media *get_pending_main_video_media(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
		return session_media;
	}

	return NULL;
}

static struct ast_sip_session_media_state *get_session_media_state(struct ast_sip_session *session, int is_notify)
{
	struct ast_sip_session_media_state *media_state= NULL;
	RAII_VAR(struct ast_bridge *, bridge, NULL, ao2_cleanup);
	struct ast_stream_topology *topology = NULL;

	if (!session || !session->channel) {
		ast_log(LOG_ERROR, "get_session_media_state param is NULL\n");
		return NULL;
	}

	ast_channel_lock(session->channel);
	bridge = ast_channel_get_bridge(session->channel);
	ast_channel_unlock(session->channel);
	if (!bridge) {
		ast_log(LOG_WARNING, "[session:%p] get_session_media_state endpoint:%s can't get bridge\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	topology = ast_bridge_get_topology(bridge, session->channel, is_notify);
	if (!topology) {
		ast_log(LOG_WARNING, "[session:%p] get_session_media_state endpoint:%s can't get topology\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	media_state = ast_sip_session_media_state_alloc();
	if (!media_state) {
		ast_log(LOG_WARNING, "[session:%p] get_session_media_state endpoint:%s can't alloc media_state\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_topology_free(topology);
		return NULL;
	}

	media_state->topology = topology;

	return media_state;
}

static void update_focus_msid_for_chan(struct ast_sip_session *session,
	struct ast_channel *focus_channel, int action)
{
	RAII_VAR(char *, msid, NULL, ast_free);

	if (action != CONF_CONTROL_OPEN_CAMERA) {
		return;
	}

	msid = ast_channel_get_msid(focus_channel, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);

	if (ast_strlen_zero(msid)) {
		ast_log(LOG_DEBUG, "[session:%p] update_focus_msid_for_chan can't get focus channel msid, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_indicate_data(session->channel, AST_CONTROL_CONF_HOST_CHANGED, msid, strlen(msid) + 1);

	return;
}

static unsigned int judge_if_need_indicate(struct ast_sip_session *session, int action)
{
	RAII_VAR(struct ast_channel *, host_channel, NULL, ao2_cleanup);
	RAII_VAR(struct ast_channel *, focus_channel, NULL, ao2_cleanup);

	focus_channel = ast_conf_get_sfu_focus(session->channel, session->conf_room);
	if (focus_channel) {
		if (focus_channel == session->channel) {
			ast_log(LOG_DEBUG, "[session:%p] Session channel is equal focus channel, need indicate for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Conference has focus channel, indicate focus msid for this chan, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			update_focus_msid_for_chan(session, focus_channel, action);
			return 0;
		}
	}

	host_channel = ast_conf_get_sfu_moderator(session->channel, session->conf_room);
	if (!host_channel) {
		ast_log(LOG_DEBUG, "[session:%p] Conference has no host, don't need indicate for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (host_channel == session->channel) {
		ast_log(LOG_DEBUG, "[session:%p] Session channel is equal host channel, need indicate for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return 1;
	}

	ast_log(LOG_DEBUG, "[session:%p] Conference has no focus chan, and chan is not equal host chan, don't need indicate for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	return 0;
}

static void indicate_participant_host_camera_change(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int action)
{
	char msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

	if (!session || !session_media|| !session->channel || ast_strlen_zero(session->conf_room)) {
		return;
	}

	if (!judge_if_need_indicate(session, action)) {
		return;
	}

	if (action == CONF_CONTROL_OPEN_CAMERA) {
		snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);
	}

	if (ast_queue_control_data(session->channel, AST_CONTROL_HOST_CAMERA_CHANGE, msid, strlen(msid) + 1)) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_HOST_CAMERA_CHANGE frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

void ast_sip_session_send_avs_mcu_control(struct ast_sip_session *session, int type)
{
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->channel) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_avs_mcu_control for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	switch (type) {
		case AVS_MCU_MIX_TYPE_SLIDES_VIDEO:
			session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
			break;
		case AVS_MCU_MIX_TYPE_FOCUS_VIDEO:
			session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
			break;
		default:
			ast_log(LOG_DEBUG, "[session:%p] Type:%d don't need get session_media, endpoint:%s\n",
				session, type, ast_sorcery_object_get_id(session->endpoint));
			ast_avs_mcu_control(session->channel, NULL, type);
			return;
	}

	if (!session_media || !session_media->rtp) {
		ast_log(LOG_WARNING, "[session:%p] Can't get slides video media for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_avs_mcu_control(session->channel, session_media->rtp, type);

	return;
}

static void indicate_video_record_for_open_camera(struct ast_sip_session *session)
{
	RAII_VAR(char *, slider, NULL, ast_free);

	if (!session) {
		return;
	}

	if (!session->focus_video) {
		return;
	}

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable video recording, don't send avs focus video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	slider = ast_conf_get_slider(session->conf_room);
	if (slider) {
		ast_log(LOG_DEBUG, "[session:%p] Conference is opening slides video, don't send avs focus video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_FOCUS_VIDEO);

	return;
}

static void indicate_video_record_for_close_camera(struct ast_sip_session *session)
{
	RAII_VAR(char *, slider, NULL, ast_free);

	if (!session) {
		return;
	}

	if (!session->focus_video) {
		return;
	}

	if (!session->conf_video_recording) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable video recording, don't send avs mix all video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	slider = ast_conf_get_slider(session->conf_room);
	if (slider) {
		ast_log(LOG_DEBUG, "[session:%p] Conference is opening slides video, don't send avs focus video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_sip_session_send_avs_mcu_control(session, AVS_MCU_MIX_TYPE_ALL_MAIN_VIDEO);

	return;
}

static int set_bridge_video_channel_for_camera_change(void *data)
{
	RAII_VAR(struct ast_bridge *, bridge, NULL, ao2_cleanup);
	RAII_VAR(struct ast_sip_session *, session, data, ao2_cleanup);
	struct ast_bridge_channel *bridge_channel = NULL;

	if (!session || !session->channel) {
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] set_bridge_video_channel_for_camera_change endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_channel_lock(session->channel);
	bridge = ast_channel_get_bridge(session->channel);
	ast_channel_unlock(session->channel);
	if (!bridge) {
		ast_log(LOG_WARNING, "[session:%p] set_bridge_video_channel_for_camera_change endpoint:%s can't get bridge\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ast_bridge_lock(bridge);

	if (!(bridge_channel = bridge_find_channel(bridge, session->channel))) {
		ast_log(LOG_ERROR, "[session:%p] set_bridge_video_channel_for_camera_change can't get bridge channel, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_bridge_unlock(bridge);
		return -1;
	}

	if (session->insert_video_channel) {
		ast_bridge_insert_video_channel_list(bridge, bridge_channel);
	} else {
		ast_bridge_remove_video_channel_list(bridge, bridge_channel);
	}

	ast_bridge_unlock(bridge);

	return 0;
}

static void push_operate_video_channels_list_to_task(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] push_operate_video_channels_list_to_task, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, set_bridge_video_channel_for_camera_change, session)) {
		ast_log(LOG_WARNING, "[session:%p] Endpoint:%s push push_operate_video_channels_list_to_task to task failed\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}

static void indicate_conference_control_camera(struct ast_sip_session *session)
{
	struct ast_sip_session_media *session_media = NULL;

	if (!session) {
		return;
	}

	if (!session->is_sfu) {
		return;
	}

	session_media = get_active_main_video_media(session);
	if (!session_media) {
		ast_log(LOG_DEBUG, "Can't get active main video media, don't need process\n");
		return;
	}

	if (session_media->open_camera || (session->camera_control_action == AST_SIP_CAMERA_CONTROL_ACTION_OPEN)) {
		ast_log(LOG_DEBUG, "[session:%p] indicate_conference_control_camera open camera for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_conf_indicate(session->channel, session->conf_room, CONF_CONTROL_OPEN_CAMERA, NULL);
		indicate_participant_host_camera_change(session, session_media, CONF_CONTROL_OPEN_CAMERA);
		indicate_video_record_for_open_camera(session);
		session_media->open_camera = 0;
		session->insert_video_channel = 1;
		push_operate_video_channels_list_to_task(session);
	} else if (session_media->close_camera || (session->camera_control_action == AST_SIP_CAMERA_CONTROL_ACTION_CLOSE)) {
		ast_log(LOG_DEBUG, "[session:%p] indicate_conference_control_camera close camera for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (!session->focus_video) {
			ast_conf_put_main_video_resource(session->conf_room, session->channel);
		} else {
			ast_log(LOG_DEBUG, "[session:%p] It's focus video, reserve the video resource, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
		ast_conf_indicate(session->channel, session->conf_room, CONF_CONTROL_CLOSE_CAMERA, NULL);
		indicate_participant_host_camera_change(session, session_media, CONF_CONTROL_CLOSE_CAMERA);
		indicate_video_record_for_close_camera(session);
		session_media->close_camera= 0;
		session->insert_video_channel = 0;
		push_operate_video_channels_list_to_task(session);
	}

	return;
}

static int delay_reinvite_topology_change_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (300 <= rdata->msg_info.msg->line.status.code) {
		/* The topology change failed, so drop the current pending media state */
		ast_sip_session_media_state_reset(session->pending_media_state);
	}

	return 0;
}

static int send_sfu_delayed_reinvite(void *data)
{
	int ret = 0;
	struct ast_sip_session *session = data;
	struct ast_sip_session_media_state *media_state = NULL;

	media_state = get_session_media_state(session, 0);
	if (!media_state) {
		ast_log(LOG_WARNING, "[session:%p] Endpoint:%s get can't media_state for send delayed re-INVITE\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->delayed_reinvite_count--;
		ao2_ref(session, -1);
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s send delayed re-INVITE...\n",
			session, ast_sorcery_object_get_id(session->endpoint));

	ret = ast_sip_session_refresh(session, NULL, NULL, delay_reinvite_topology_change_response,
		AST_SIP_SESSION_REFRESH_METHOD_INVITE, 1, media_state);

	if (session->pending_media_state && !session->pending_media_state->topology) {
		if (session->delayed_reinvite_count > 1) {
			ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s didn't send re-INVITE, and has no pending re-INVITE, try to send again later\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			push_sfu_delayed_reinvite_to_task(session);
			session->delayed_reinvite_count = 1;
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s didn't send re-INVITE, media is already the newest\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			session->delayed_reinvite_count = 0;
		}
	} else {
		session->delayed_reinvite_count--;
	}

	ao2_ref(session, -1);

	return ret;
}

static void push_sfu_delayed_reinvite_to_task(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has delayed re-INVITE, push to session task to send\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_sfu_delayed_reinvite, session)) {
		ast_log(LOG_WARNING, "[session:%p] Endpoint:%s push delayed re-INVITE to task failed\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->delayed_reinvite_count--;
		ao2_ref(session, -1);
	}

	return;
}

static void change_participant_stream_params_for_sfu(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	if (!session || !session_media || !stream) {
		return;
	}

	if (strncmp(ast_stream_get_name(stream), SOFTBRIDGE_VIDEO_DEST_PREFIX, SOFTBRIDGE_VIDEO_DEST_LEN)) {
		return;
	}

	if (ast_stream_get_metadata(stream, "MSID:MSLABEL")) {
		strncpy(session_media->mslabel, ast_stream_get_metadata(stream, "MSID:MSLABEL"), sizeof(session_media->mslabel) - 1);
	}
	if (ast_stream_get_metadata(stream, "MSID:LABEL")) {
		strncpy(session_media->label, ast_stream_get_metadata(stream, "MSID:LABEL"), sizeof(session_media->label) - 1);
	}

	ast_log(LOG_DEBUG, "[session:%p] Change other participant stream:%s params for endpoint:%s\n",
		session, session_media->mslabel, ast_sorcery_object_get_id(session->endpoint));

	/* It's participant media, we only support to send stream to SIP terminal, don't need to receive stream from SIP terminal 
	 * For gswave we use inactive to disable media, need to retain when inactive.
	 */
	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && ast_sip_session_media_use_bundle(session)) {
		if (ast_stream_get_local_media_direction(stream) != INACTIVE) {
			ast_stream_set_local_media_direction(stream, SENDONLY);
		}
	} else {
		ast_stream_set_local_media_direction(stream, SENDONLY);
	}

	add_video_fec_codec_to_stream_format(session, stream);
	add_nack_to_stream(session, stream);
	add_rtx_codec_to_stream_format(session, stream);

	return;
}

static void change_rtp_instance_local_ssrc(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	if (!session || !session_media || !stream || !session_media->rtp) {
		return;
	}

	/*remote it for Bug297853 */
	//if (ast_sip_session_media_use_bundle(session)) {
	//	return;
	//}

	if (ast_stream_get_change_ssrc(stream) || ast_stream_get_held(stream)) {
		ast_log(LOG_DEBUG, "[session:%p] It's session refresh for hold, change ssrc for session_media sub_type:%s, change_ssrc:%d, held:%d, endpoint:%s\n",
			session, ast_codec_media_sub_type2str(session_media->sub_type), ast_stream_get_change_ssrc(stream), ast_stream_get_held(stream), ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_change_ssrc(stream, 0);
		ast_stream_set_held(stream, 0);
		ast_rtp_instance_change_local_ssrc(session_media->rtp);
	}

	return;
}

static void change_stream_msid_according_session_media(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	if (!session || !session_media || !stream) {
		return;
	}

	ast_stream_set_metadata(stream, "MSID:MSLABEL", session_media->mslabel);
	ast_stream_set_metadata(stream, "MSID:LABEL", session_media->label);

	return;
}

static void change_bundle_state(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	if (!session || !session_media || !stream) {
		return;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
		return;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return;
	}

	session_media->bundled = 1;
	session_media->bundle_group = 0;

	return;
}

void ast_sip_session_bfcp_data_destroy(void *obj)
{
	struct bfcp_info_data *bfcp_data = obj;

	if (!bfcp_data) {
		return;
	}

	ast_free(bfcp_data->bfcp_info);
	ao2_ref(bfcp_data->session, -1);

	return;
}

struct bfcp_info_data *ast_sip_session_bfcp_data_alloc(struct ast_sip_session *session, const struct ast_bfcp_info *bfcp_info)
{
	struct bfcp_info_data *bfcp_data = ao2_alloc(sizeof(*bfcp_data), ast_sip_session_bfcp_data_destroy);

	if (!bfcp_data) {
		return NULL;
	}

	bfcp_data->bfcp_info = (struct ast_bfcp_info *)ast_malloc(sizeof(struct ast_bfcp_info));
	if (!bfcp_data->bfcp_info) {
		ao2_ref(bfcp_data, -1);
		return NULL;
	}

	memcpy(bfcp_data->bfcp_info, bfcp_info, sizeof(struct ast_bfcp_info));
	ao2_ref(session, +1);
	bfcp_data->session = session;

	return bfcp_data;
}

const char *ast_sip_session_bfcp_info_reqstatus_2str(enum bfcp_reqstatus status)
{
    if ((status > BFCP_REVOKED) || (status < BFCP_PENDING)) {
        return  "INVALID";
    }

    return bfcp_info_reqstatus_strmap[status][0];
}

const char *ast_sip_session_bfcp_info_prim_2str(enum bfcp_prim  prim)
{
	return bfcp_info_prim_strmap[prim][0];
}

static const char *bfcp_denied_reason_2str(enum bfcp_denied_reason denied_reason)
{
    if ((denied_reason >= BFCP_DENIED_REASON_MAX) || (denied_reason <= BFCP_DENIED_REASON_NONE)) {
        return  "INVALID";
    }

    return bfcp_denied_reason_strmap[denied_reason][0];
}

int ast_sip_session_bfcp_info_generate_body_content(const struct ast_bfcp_info *bfcp_info, char *content, int len)
{
	char   buf[256] = {0};

	if (!bfcp_info || !content || len <= 0) {
		return -1;
	}

	if (strlen(bfcp_info->info_conf_id)) {
		snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
		                   "<ConfControl>\r\n");
	} else {
		snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
		                   "<CallControl>\r\n");
	}

	snprintf(buf, sizeof(buf), "  <Action name=\"%s\">\r\n", ast_sip_session_bfcp_info_prim_2str(bfcp_info->primitive));
	strncat(content, buf, len-strlen(content));

	if (bfcp_info->status != BFCP_INFO_INVALID && bfcp_info->status <= BFCP_REVOKED && bfcp_info->status >= BFCP_PENDING) {
		snprintf(buf, sizeof(buf), "    <Status>%s</Status>\r\n", ast_sip_session_bfcp_info_reqstatus_2str(bfcp_info->status));
		strncat(content, buf, len-strlen(content));
		if (bfcp_info->status == BFCP_DENIED && bfcp_info->denied_reason) {
			snprintf(buf, sizeof(buf), "    <DeniedReason>%s</DeniedReason>\r\n", bfcp_denied_reason_2str(bfcp_info->denied_reason));
			strncat(content, buf, len-strlen(content));
		}
	}

	strncat(content, "  </Action>\r\n", len-strlen(content));

	if (strlen(bfcp_info->info_conf_id)) {
		snprintf(buf, sizeof(buf), "  <ConfID>%.*s</ConfID>\r\n", (int)strlen(bfcp_info->info_conf_id), bfcp_info->info_conf_id);
		strncat(content, buf, len-strlen(content));
	}

	if (strlen(bfcp_info->slider_name) && (bfcp_info->primitive != BFCP_FLOOR_REQUEST_STATUS)) {
		snprintf(buf, sizeof(buf), "  <UserID>%.*s</UserID>\r\n", (int)strlen(bfcp_info->slider_name), bfcp_info->slider_name);
		strncat(content, buf, len-strlen(content));
	} else {
		if (strlen(bfcp_info->account_id)) {
			snprintf(buf, sizeof(buf), "  <UserID>%.*s</UserID>\r\n", (int)strlen(bfcp_info->account_id), bfcp_info->account_id);
			strncat(content, buf, len-strlen(content));
		}
	}

	if (strlen(bfcp_info->floor_owner_id)) {
		snprintf(buf, sizeof(buf), "  <FloorOwnedUserID>%.*s</FloorOwnedUserID>\r\n", (int)strlen(bfcp_info->floor_owner_id), bfcp_info->floor_owner_id);
		strncat(content, buf, len-strlen(content));
	}

	if (strlen(bfcp_info->share_type)) {
		snprintf(buf, sizeof(buf), "  <ShareType>%.*s</ShareType>\r\n", (int)strlen(bfcp_info->share_type), bfcp_info->share_type);
		strncat(content, buf, len-strlen(content));
	}

	if (strlen(bfcp_info->info_conf_id)) {
		strncat(content, "</ConfControl>", len-strlen(content));
	} else {
		strncat(content, "</CallControl>", len-strlen(content));
	}

	ast_log(LOG_DEBUG, "Generate bfcp info body[%s]\n", content);

	return 0;
}

int ast_sip_session_bfcp_send_conference_sip_info(void *data)
{
	struct pjsip_tx_data *tdata = NULL;
	char content[1024] = {0};
	struct ast_sip_session *session = NULL;
	struct ast_bfcp_info *bfcp_info = NULL;

	struct bfcp_info_data *bfcp_data = data;

	if (!bfcp_data) {
		return -1;
	}

	session = bfcp_data->session;
	bfcp_info = bfcp_data->bfcp_info;

	if (!session || !bfcp_info) {
		return -1;
	}

	if (session && session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_DEBUG, "Dialog is not established really, don't send sip info\n");
		ao2_ref(bfcp_data, -1);
		return 0;
	}

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text bfcp control INFO request\n");
		ao2_ref(bfcp_data, -1);
		return -1;
	}

	ast_sip_session_bfcp_info_generate_body_content(bfcp_info, content, sizeof(content));
	
	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-grandstream-confcontrol+xml",
		.body_text = content
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to bfcp control INFO request\n");
		ao2_ref(bfcp_data, -1);
		return -1;
	}

	session->pending_bfcp_info = 1;
	memcpy(&session->last_bfcp_info, bfcp_info, sizeof(session->last_bfcp_info));

	ast_sip_session_send_request(session, tdata);

	ao2_ref(bfcp_data, -1);

	return 0;
}

int ast_sip_session_bfcp_send_sip_info(void *data)
{
	struct pjsip_tx_data *tdata = NULL;
	char content[1024] = {0};
	struct ast_sip_session *session = NULL;
	const struct ast_bfcp_info *bfcp_info = NULL;

	struct bfcp_info_data *bfcp_data = data;

	if (!bfcp_data) {
		return -1;
	}

	session = bfcp_data->session;
	bfcp_info = bfcp_data->bfcp_info;

	if (!session || !bfcp_info) {
		return -1;
	}

	if (session && session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_DEBUG, "Dialog is not established really, don't send sip info\n");
		ao2_ref(bfcp_data, -1);
		return 0;
	}

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create text bfcp control INFO request\n");
		ao2_ref(bfcp_data, -1);
		return -1;
	}

	ast_sip_session_bfcp_info_generate_body_content(bfcp_info, content, sizeof(content));

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-grandstream-slides-video-control+xml",
		.body_text = content
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to bfcp control INFO request\n");
		ao2_ref(bfcp_data, -1);
		return -1;
	}

	session->pending_bfcp_info = 1;
	memcpy(&session->last_bfcp_info, bfcp_info, sizeof(session->last_bfcp_info));

	ast_sip_session_send_request(session, tdata);

	ao2_ref(bfcp_data, -1);

	return 0;
}

static int create_conf_timeout_state_body(char *content, int len, struct ast_conf_timeout_state *conf_timeout_state)
{
	char buf[256] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<state version=\"1\">\r\n"
	"  <item name=\"conf-state\">\r\n"
	"    <ConfTimeout>\r\n");

	if (!ast_strlen_zero(conf_timeout_state->conf)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <Conf>%s</Conf>\r\n", conf_timeout_state->conf);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(conf_timeout_state->type)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <Type>%s</Type>\r\n", conf_timeout_state->type);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(conf_timeout_state->time)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <Time>%s</Time>\r\n", conf_timeout_state->time);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(conf_timeout_state->status)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <Status>%s</Status>\r\n", conf_timeout_state->status);
		strncat(content, buf, len-strlen(content));
	}

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    </ConfTimeout>\r\n"
	"  </item>\r\n"
	"</state>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs state body[%s]\n", content);

	return 0;
}

static int create_trunk_info_state_body(char *content, int len, struct ast_trunk_info *trunk_info)
{
	char buf[512] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<state version=\"1\">\r\n"
	"  <item name=\"call-state\">\r\n"
	"    <trunkInfo>\r\n");

	if (!ast_strlen_zero(trunk_info->conf_link)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <confLink>%s</confLink>\r\n", trunk_info->conf_link);
		strncat(content, buf, len-strlen(content));
	}

	if (!ast_strlen_zero(trunk_info->channel_id)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <replaceChannelID>%s</replaceChannelID>\r\n", trunk_info->channel_id);
		strncat(content, buf, len-strlen(content));
	}

	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    </trunkInfo>\r\n"
	"  </item>\r\n"
	"</state>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs state body[%s]\n", content);

	return 0;
}

static int create_answered_status_state_body(char *content, int len, struct ast_answered_status *queue_status)
{
	char buf[512] = {0};

	if (!content || len <= 0) {
		return -1;
	}

	snprintf(content, len, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\r\n"
	"<state version=\"1\">\r\n"
	"  <item name=\"call-state\">\r\n"
	"    <callStatus>\r\n");

	if (!ast_strlen_zero(queue_status->answered)) {
		memset(buf, 0, sizeof(buf));
		snprintf(buf, sizeof(buf) - 1, "      <answered>%s</answered>\r\n", queue_status->answered);
		strncat(content, buf, len-strlen(content));
	}


	memset(buf, 0, sizeof(buf));
	snprintf(buf, sizeof(buf) - 1, "    </callStatus>\r\n"
	"  </item>\r\n"
	"</state>\r\n");

	strncat(content, buf, len-strlen(content));

	ast_log(LOG_DEBUG, "Generate notify gs state body[%s]\n", content);

	return 0;
}

static void gs_state_clone_data_destroy(void *obj)
{
	struct ast_gs_state_clone_data *gs_state_data = obj;

	if (!gs_state_data) {
		return;
	}

	if (gs_state_data->data) {
		ast_free(gs_state_data->data);
		gs_state_data->data = NULL;
	}

	return;
}

static struct ast_gs_state_clone_data *gs_state_clone_data_alloc(struct ast_sip_session *session,
	enum ast_sip_session_gs_state_type type, void *data)
{
	struct ast_gs_state_clone_data *gs_state_data = ao2_alloc(sizeof(*gs_state_data), gs_state_clone_data_destroy);

	if (!gs_state_data) {
		return NULL;
	}

	gs_state_data->type = type;

	switch (type) {
		case AST_GS_STATE_TYPE_CONF_TIMEOUT:
			gs_state_data->data = (struct ast_conf_timeout_state *)ast_malloc(sizeof(struct ast_conf_timeout_state));
			if (!gs_state_data->data) {
				ao2_ref(gs_state_data, -1);
				return NULL;
			}
			memcpy(gs_state_data->data, data, sizeof(struct ast_conf_timeout_state));
			break;
		case AST_GS_STATE_TYPE_TRUNK_INFO:
			gs_state_data->data = (struct ast_trunk_info *)ast_malloc(sizeof(struct ast_trunk_info));
			if (!gs_state_data->data) {
				ao2_ref(gs_state_data, -1);
				return NULL;
			}
			memcpy(gs_state_data->data, data, sizeof(struct ast_trunk_info));
			break;
		case AST_GS_STATE_TYPE_ANSWERED_STATUS:
			gs_state_data->data = (struct ast_answered_status *)ast_malloc(sizeof(struct ast_answered_status));
			if (!gs_state_data->data) {
				ao2_ref(gs_state_data, -1);
				return NULL;
			}
			memcpy(gs_state_data->data, data, sizeof(struct ast_answered_status));  /*Mod by weidu,for Bug 264097*/
			break;

		default:
			ao2_ref(gs_state_data, -1);
			return NULL;
	}

	return gs_state_data;
}

static int gs_state_notify_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (!session || !rdata) {
		ast_log(LOG_ERROR, "gs_state_notify_response param is NULL\n");
		return -1;
	}

	if (!session->gs_state_clone_data) {
		ast_log(LOG_WARNING, "[session:%p] Has no gs_state_clone_data, don't need process, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	/* WAVE APP may response 500 when receive INVITE 200OK and NOTIFY at the same time, need send again */
	if (500 == rdata->msg_info.msg->line.status.code) {
		ast_log(LOG_WARNING, "[session:%p] recv gs state notify response failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_send_notify_to_update_gs_state(session, session->gs_state_clone_data->type, session->gs_state_clone_data->data);
	}

	/* Cleanup the data */
	ao2_cleanup(session->gs_state_clone_data);
	session->gs_state_clone_data = NULL;

	return 0;
}

static int send_notify_to_update_gs_state(void *data)
{
	struct pjsip_tx_data *tdata = NULL;
	char content[1024] = {0};
	pjsip_event_hdr *sub_event = NULL;
	pj_str_t hdr_name = {"Allow-Events", 12};
	pj_str_t hdr_value = {"conference, presence, dialog, message-summary, call-completion, refer, x-gs-state", 81};
	pjsip_generic_string_hdr *hdr_allowevent = NULL;
	struct ast_sip_session *session = NULL;
	enum ast_sip_session_gs_state_type type;
	RAII_VAR(struct ast_gs_state_data *, gs_state_data, data, ao2_cleanup);
	struct ast_gs_state_clone_data *gs_state_clone_data = NULL;

	if (!gs_state_data) {
		ast_log(LOG_ERROR, "send_notify_to_update_gs_state gs_state_data is NULL\n");
		return -1;
	}

	session = gs_state_data->session;

	if (session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_ERROR, "[session:%p] Dialog is not established really, don't send sip NOTIFY to update gs state, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] Send NOTIFY request to update gs state, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	type = gs_state_data->type;
	switch (type) {
		case AST_GS_STATE_TYPE_CONF_TIMEOUT:
			create_conf_timeout_state_body(content, sizeof(content), gs_state_data->data);
			break;
		case AST_GS_STATE_TYPE_TRUNK_INFO:
			create_trunk_info_state_body(content, sizeof(content), gs_state_data->data);
			break;
		case AST_GS_STATE_TYPE_ANSWERED_STATUS:
			session->had_send_answered=1;//added by zyxia for Bug 262501 -
			create_answered_status_state_body(content, sizeof(content), gs_state_data->data);
			break;

		default:
			ast_log(LOG_ERROR, "[session:%p] Can't support type:%d to update gs state, endpoint:%s\n",
				session, type, ast_sorcery_object_get_id(session->endpoint));
			return -1;
	}

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "[session:%p] Can't create NOTIFY request to update gs state, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "x-gs-state+xml",
		.body_text = content
	};

	sub_event = pjsip_event_hdr_create(tdata->pool);
	pj_cstr(&sub_event->event_type, "x-gs-state");
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)sub_event);

	hdr_allowevent = pjsip_generic_string_hdr_create(tdata->pool, &hdr_name, &hdr_value);
	pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr_allowevent);

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to update turn info notify request\n");
		return -1;
	}

	gs_state_clone_data = gs_state_clone_data_alloc(session, gs_state_data->type, gs_state_data->data);
	if (gs_state_clone_data) {
		if (session->gs_state_clone_data) {
			ast_log(LOG_WARNING, "[session:%p] Already has clone data, release first, use latest data, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_cleanup(session->gs_state_clone_data);
			session->gs_state_clone_data = NULL;
		}
		session->gs_state_clone_data = gs_state_clone_data;
	} else {
		ast_log(LOG_WARNING, "[session:%p] Clone gs_state_data failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}

	ast_sip_session_send_request_with_cb(session, tdata, gs_state_notify_response);

	return 0;
}

static void gs_state_data_destroy(void *obj)
{
	struct ast_gs_state_data *gs_state_data = obj;

	if (!gs_state_data) {
		return;
	}

	if (gs_state_data->data) {
		ast_free(gs_state_data->data);
		gs_state_data->data = NULL;
	}

	if (gs_state_data->session) {
		ao2_ref(gs_state_data->session, -1);
	}

	return;
}

static struct ast_gs_state_data *gs_state_data_alloc(struct ast_sip_session *session,
	enum ast_sip_session_gs_state_type type, void *data)
{
	struct ast_gs_state_data *gs_state_data = ao2_alloc(sizeof(*gs_state_data), gs_state_data_destroy);

	if (!gs_state_data) {
		return NULL;
	}

	gs_state_data->type = type;

	switch (type) {
		case AST_GS_STATE_TYPE_CONF_TIMEOUT:
			gs_state_data->data = (struct ast_conf_timeout_state *)ast_malloc(sizeof(struct ast_conf_timeout_state));
			if (!gs_state_data->data) {
				ao2_ref(gs_state_data, -1);
				return NULL;
			}
			memcpy(gs_state_data->data, data, sizeof(struct ast_conf_timeout_state));
			break;
		case AST_GS_STATE_TYPE_TRUNK_INFO:
			gs_state_data->data = (struct ast_trunk_info *)ast_malloc(sizeof(struct ast_trunk_info));
			if (!gs_state_data->data) {
				ao2_ref(gs_state_data, -1);
				return NULL;
			}
			memcpy(gs_state_data->data, data, sizeof(struct ast_trunk_info));
			break;
		case AST_GS_STATE_TYPE_ANSWERED_STATUS:
			gs_state_data->data = (struct ast_answered_status *)ast_malloc(sizeof(struct ast_answered_status));
			if (!gs_state_data->data) {
				ao2_ref(gs_state_data, -1);
				return NULL;
			}
			memcpy(gs_state_data->data, data, sizeof(struct ast_answered_status));  /*Mod by weidu,for Bug 264097*/
			break;

		default:
			ao2_ref(gs_state_data, -1);
			return NULL;
	}

	ao2_ref(session, +1);
	gs_state_data->session = session;

	return gs_state_data;
}

int ast_sip_session_send_notify_to_update_gs_state(struct ast_sip_session *session,
	enum ast_sip_session_gs_state_type type, void *data)
{
	struct ast_gs_state_data *gs_state_data = NULL;

	if (!session) {
		return -1;
	}

	gs_state_data = gs_state_data_alloc(session, type, data);
	if (!gs_state_data) {
		ast_log(LOG_WARNING, "[session:%p] Can't send NOTIFY to update gs_state because of alloc failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (ast_sip_push_task(session->serializer, send_notify_to_update_gs_state, gs_state_data)) {
		ast_log(LOG_WARNING, "[session:%p] Can't send NOTIFY to update gs_state because of push to task failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_cleanup(gs_state_data);
		return -1;
	}

	return 0;
}

int ast_sip_session_judge_if_main_video(const pjmedia_sdp_media *offer_media)
{
	pjmedia_sdp_attr *attr = NULL;
	char content[MAX_CONTENT_LEN];
	memset(content, 0, sizeof(content));
	char media_type[16] = {0};

	/* If the media type is not video, we just consider is not main video */
	ast_copy_pj_str(media_type, &offer_media->desc.media, sizeof(media_type));
	if (strncmp(media_type, "video", 5)) {
		return FALSE;
    }

	if ((attr = pjmedia_sdp_media_find_attr2(offer_media, "content", NULL))) {
		ast_copy_pj_str(content, pj_strltrim(&attr->value), sizeof(content));
		if (!strncmp(content, "main", 4)) {
			return TRUE;
		}
	} else {
		/* If we can't find content attr, we just consider as main video */
		return TRUE;
	}

	return FALSE;
}


int ast_sip_session_judge_if_slides_video(const pjmedia_sdp_media *offer_media)
{
	pjmedia_sdp_attr *attr = NULL;
	char content[MAX_CONTENT_LEN];
	memset(content, 0, sizeof(content));

	if ((attr = pjmedia_sdp_media_find_attr2(offer_media, "content", NULL))) {
		ast_copy_pj_str(content, pj_strltrim(&attr->value), sizeof(content));
		if (!strncmp(content, "slides", 6)) {
			return TRUE;
		}
	}

	return FALSE;
}

int ast_sip_session_judge_if_bfcp_stream(const pjmedia_sdp_media *stream)
{
	pj_str_t transport;

	if (!stream) {
		ast_log(LOG_WARNING, "ast_sip_session_judge_if_bfcp_stream, param is null\n");
		return FALSE;
	}

	transport = stream->desc.transport;

	if (NULL == transport.ptr) {
		ast_log(LOG_DEBUG, "ast_sip_session_judge_if_bfcp_stream, has no transport");
		return FALSE;
	}

	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	if (strstr(transport_str, "UDP/BFCP") || strstr(transport_str, "TCP/BFCP")) {
		ast_log(LOG_DEBUG, "ast_sip_session_judge_if_bfcp_stream transport is %s\n", transport_str);
		return TRUE;
	}

	return FALSE;
}

int ast_sip_session_judge_if_fec_stream(const pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	int i = 0;
	const pj_str_t str_gs_fec = { "GS-FEC", 6 };

	if (!stream) {
		ast_log(LOG_WARNING, "Judge fec stream, param is null\n");
		return FALSE;
	}

	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		if (pj_strstr(&attr->value, &str_gs_fec) != NULL) {
			return TRUE;
		}
	}
	return FALSE;
}


int ast_sip_session_judge_if_fecc_stream(const pjmedia_sdp_media *stream)
{
	pjmedia_sdp_attr *attr = NULL;
	int i = 0;
	const pj_str_t str_fecc = { "H224", 4 };

	if (!stream) {
		ast_log(LOG_WARNING, "Judge fecc stream, param is null\n");
		return FALSE;
	}

	for (i = 0; i < stream->desc.fmt_count; ++i) {
		/* Look for the optional rtpmap attribute */
		if (!(attr = pjmedia_sdp_media_find_attr2(stream, "rtpmap", &stream->desc.fmt[i]))) {
			continue;
		}

		if (pj_strstr(&attr->value, &str_fecc) != NULL) {
			return TRUE;
		}
	}
	return FALSE;
}

int ast_sip_session_judge_if_passthrough_attr(char *attr_value)
{
	if (!attr_value) {
		return FALSE;
	}

	if (strncasecmp(attr_value, "rtpmap", 6) &&
		strncasecmp(attr_value, "fmtp", 4) &&
		strncasecmp(attr_value, "sendrecv", 8) &&
		strncasecmp(attr_value, "sendonly", 8) &&
		strncasecmp(attr_value, "recvonly", 8) &&
		strncasecmp(attr_value, "inactive", 8) &&
		strncasecmp(attr_value, "framerate", 9) &&
		strncasecmp(attr_value, "label", 5) &&
		strncasecmp(attr_value, "content", 7) &&
		strncasecmp(attr_value, "crypto", 6) &&
		strncasecmp(attr_value, "ssrc", 4) &&
		strncasecmp(attr_value, "rtcp", 4) &&
		strncasecmp(attr_value, "gs-fec-version", 14) &&
		strncasecmp(attr_value, "candidate", 9) &&
		strncasecmp(attr_value, "ice-ufrag", 9) &&
		strncasecmp(attr_value, "ice-pwd", 7) &&
		strncasecmp(attr_value, "ice-options", 11) &&
		strncasecmp(attr_value, "fingerprint", 11) &&
		strncasecmp(attr_value, "setup", 5) &&
		strncasecmp(attr_value, "mid", 3) &&
		strncasecmp(attr_value, "extmap", 6) &&
		strncasecmp(attr_value, "confid", 6) &&
		strncasecmp(attr_value, "userid", 6) &&
		strncasecmp(attr_value, "floorctrl", 9) &&
		strncasecmp(attr_value, "setup", 5) &&
		strncasecmp(attr_value, "connection", 10) &&
		strncasecmp(attr_value, "floorid", 7) &&
		strncasecmp(attr_value, "fingerprint", 11) &&
		strncasecmp(attr_value, "imageattr", 9) &&
		strncasecmp(attr_value, "msid", 4) &&
		strncasecmp(attr_value, "bundle-only", 11) &&
		strncasecmp(attr_value, "declined-reason", 15) &&
		strncasecmp(attr_value, "zrtp-hash", 9)) {
		return TRUE;
	}

	return FALSE;
}

int ast_sip_session_judge_if_passthrough_bandw(char *bandw_name)
{
	if (!bandw_name) {
		return FALSE;
	}

	if (strncasecmp(bandw_name, "AS", 2) &&
		strncasecmp(bandw_name, "TIAS", 4)) {
		return TRUE;
	}

	return FALSE;
}

int ast_sip_session_judge_if_need_passthrough(void)
{
	int res = 0;
	char *passthrough = NULL;

	passthrough = ast_sip_get_attr_passthrough();
	if (!passthrough) {
		return 0;
	}
	if (strcasecmp(passthrough, "yes") == 0) {
		res = 1;
	}

	ast_free(passthrough);
	passthrough = NULL;

	return res;
}

void ast_sip_session_change_gs_fec_version(struct ast_format_cap *caps)
{
	unsigned int index = 0;
	struct ast_format *format = NULL;
	struct ast_format *fec_format = NULL;

	if (!caps) {
		return;
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}
		if (!strcasecmp(ast_format_get_name(format), "audio_fec")) {
			fec_format = ast_format_attribute_set(format, "version", AVS_SUPPORT_FEC_VERSION);
			ast_format_cap_remove(caps, format);
			ast_format_cap_append(caps, fec_format, 0);
			ao2_ref(fec_format, -1);
		}
		ao2_ref(format, -1);
	}

	return;
}

enum ast_sip_application_type ast_sip_session_get_application_type(pj_str_t transport)
{
	if (NULL == transport.ptr) {
		ast_log(LOG_DEBUG, "get application type has no transport");
		return AST_SIP_APP_INVALID;
	}
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	if (strstr(transport_str, "UDP/BFCP")) {
		return AST_SIP_APP_UDP_BFCP;
	} else if (strstr(transport_str, "TCP/BFCP")) {
		return AST_SIP_APP_TCP_BFCP;
	} else if (strstr(transport_str, "RTP")) {
		return AST_SIP_APP_RTP;
	}

	return AST_SIP_APP_INVALID;
}

static int ast_sip_session_media_stream_has_crypto(const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;

		/* check the stream for the required crypto attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "crypto")) {
			continue;
		}

		return 1;
	}

	return 0;
}
int ast_sip_session_media_stream_has_zrtphash(const struct pjmedia_sdp_media *stream)
{
	int i;

	for (i = 0; i < stream->attr_count; i++) {
		pjmedia_sdp_attr *attr;

		/* check the stream for the required zrtp-hash attribute */
		attr = stream->attr[i];
		if (pj_strcmp2(&attr->name, "zrtp-hash")) {
			continue;
		}

		return 1;
	}

	return 0;
}

int  ast_sip_session_media_stream_has_srtp(pj_str_t transport,
	const struct pjmedia_sdp_media *stream)
{
	
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	if (strstr(transport_str, "SAVP") && ast_sip_session_media_stream_has_crypto(stream)) {
		return 1;
	}
	return 0;
}

enum ast_sip_session_media_encryption ast_sip_session_get_media_encryption_type(pj_str_t transport,
	const struct pjmedia_sdp_media *stream, unsigned int *optimistic)
{
	RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);

	*optimistic = 0;

	if (strstr(transport_str, "UDP/TLS")) {
		return AST_SIP_MEDIA_ENCRYPT_DTLS;
	}
	/*remove it by zyxia for Bug 306320 
	else if (ast_sip_session_media_stream_has_zrtphash(stream)) {
		*optimistic = 1;
		return AST_SIP_MEDIA_ENCRYPT_ZRTP;
	}
	*/
	else if (strstr(transport_str, "SAVP")) {
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	}else if (ast_sip_session_media_stream_has_crypto(stream)) {
		*optimistic = 1;
		return AST_SIP_MEDIA_ENCRYPT_SDES;
	} else {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}
}


enum ast_sip_session_media_encryption ast_sip_session_check_endpoint_media_transport(
	struct ast_sip_endpoint *endpoint,
	const struct pjmedia_sdp_media *stream, unsigned int is_webrtc_call)
{
	enum ast_sip_session_media_encryption incoming_encryption;
	char transport_end = stream->desc.transport.ptr[stream->desc.transport.slen - 1];
	unsigned int optimistic;


	/*start added by zyxia for zrtp, check zrtp firsrt,2024.6.6*/
	if(endpoint->media.rtp.configure_use_received_zrtp && ast_sip_session_media_stream_has_zrtphash(stream)) {

		if(ast_sip_session_media_stream_has_srtp(stream->desc.transport, stream))
			return AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP;
		else
			return AST_SIP_MEDIA_ENCRYPT_ZRTP;
	}

	/*End added by zyxia for zrtp, check zrtp firsrt,2024.6.6*/

	if (is_webrtc_call) {
		if ((transport_end == 'F' && !endpoint->media.rtp.use_avpf)
			|| (transport_end != 'F' && endpoint->media.rtp.use_avpf)) {
			return AST_SIP_MEDIA_TRANSPORT_INVALID;
		}
	} else {
		/* endpoint->media.rtp.configure_use_avpf is no as default  */
		if ((transport_end == 'F' && !endpoint->media.rtp.configure_use_avpf)
			|| (transport_end != 'F' && endpoint->media.rtp.configure_use_avpf)) {
			return AST_SIP_MEDIA_TRANSPORT_INVALID;
		}
	}

	incoming_encryption = ast_sip_session_get_media_encryption_type(stream->desc.transport, stream, &optimistic);

	if (is_webrtc_call) {
		if (incoming_encryption == endpoint->media.rtp.encryption) {
			return incoming_encryption;
		}

		if ((endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_NO_ENCRYPT)
			|| (endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP)
			|| (endpoint->media.rtp.encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS)) {
			return incoming_encryption;
		}
	} else {

		if (incoming_encryption == endpoint->media.rtp.configure_encryption) {
			return incoming_encryption;
		}

		if ((endpoint->media.rtp.configure_encryption == AST_SIP_MEDIA_AUTO_NO_ENCRYPT)
			|| (endpoint->media.rtp.configure_encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP)
			|| (endpoint->media.rtp.configure_encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS)) {
			return incoming_encryption;
		}

	}

	if (endpoint->media.rtp.force_avp ||
		endpoint->media.rtp.encryption_optimistic) {
		return incoming_encryption;
	}

	/* If an optimistic offer has been made but encryption is not enabled consider it as having
	 * no offer of crypto at all instead of invalid so the session proceeds.
	 */
	if (optimistic) {
		return AST_SIP_MEDIA_ENCRYPT_NONE;
	}

	return AST_SIP_MEDIA_TRANSPORT_INVALID;
}

void ast_sip_session_set_media_sub_type(struct ast_sip_session *session, const pjmedia_sdp_media *stream,
	struct ast_sip_session_media *session_media, struct ast_stream *asterisk_stream)
{
	pj_str_t transport;

	if (!session || !stream || !session_media || !asterisk_stream) {
		return;
	}

	/* We only support one audio media, don't distinguish sub type */
	if (session_media->type == AST_MEDIA_TYPE_AUDIO) {
		session_media->sub_type = AST_MEDIA_SUB_AUDIO_TYPE;
		ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_AUDIO_TYPE);
		return;
	}

	transport = stream->desc.transport;
	if (transport.ptr) {
		RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);
		if (strstr(transport_str, "UDP/BFCP")) {
			session_media->sub_type =  AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP;
			ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP);
			return;
		} else if (strstr(transport_str, "TCP/BFCP")) {
			session_media->sub_type =  AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP;
			ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP);
			return;
		}
	}

	if (ast_sip_session_judge_if_slides_video(stream)) {
		session_media->sub_type = AST_MEDIA_SUB_VIDEO_TYPE_SLIDES;
		ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	} else if (ast_sip_session_judge_if_main_video(stream)) {
		session_media->sub_type = AST_MEDIA_SUB_VIDEO_TYPE_MAIN;
		ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
	} else if (ast_sip_session_judge_if_fec_stream(stream)) {
		session_media->sub_type = AST_MEDIA_SUB_APPLICATION_TYPE_FEC;
		ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_APPLICATION_TYPE_FEC);
	} else if (ast_sip_session_judge_if_fecc_stream(stream)) {
		session_media->sub_type = AST_MEDIA_SUB_APPLICATION_TYPE_FECC;
		ast_stream_set_sub_type(asterisk_stream, AST_MEDIA_SUB_APPLICATION_TYPE_FECC);
	} else {
		ast_log(LOG_DEBUG, "Can't set this stream sub type now\n");
	}

	return;
}

enum ast_media_sub_type ast_sip_session_get_media_sub_type(struct ast_sip_session *session, const pjmedia_sdp_media *stream)
{
	pj_str_t transport;

	if (!session || !stream) {
		return AST_MEDIA_SUB_TYPE_UNKNOWN;
	}

	if (!strncmp(stream->desc.media.ptr, "audio", 5)) {
		return AST_MEDIA_SUB_AUDIO_TYPE;
	}

	transport = stream->desc.transport;
	if (transport.ptr) {
		RAII_VAR(char *, transport_str, ast_strndup(transport.ptr, transport.slen), ast_free);
		if (strstr(transport_str, "UDP/BFCP")) {
			return AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP;
		} else if (strstr(transport_str, "TCP/BFCP")) {
			return AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP;
		}
	}

	if (ast_sip_session_judge_if_slides_video(stream)) {
		return AST_MEDIA_SUB_VIDEO_TYPE_SLIDES;
	} else if (ast_sip_session_judge_if_main_video(stream)) {
		return AST_MEDIA_SUB_VIDEO_TYPE_MAIN;
	} else if (ast_sip_session_judge_if_fec_stream(stream)) {
		return AST_MEDIA_SUB_APPLICATION_TYPE_FEC;
	} else if (ast_sip_session_judge_if_fecc_stream(stream)) {
		return AST_MEDIA_SUB_APPLICATION_TYPE_FECC;
	} else {
		ast_log(LOG_DEBUG, "Can't set this stream sub type now\n");
	}

	return AST_MEDIA_SUB_TYPE_UNKNOWN;
}

void ast_sip_session_init_stream_media_direction(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); ++i) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);

		/* gswave will contain video media even make audio call, so don't change */
		if ((session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) &&
			ast_sip_session_media_use_bundle(session) &&
			(ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO)) {
			continue;
		}

		ast_stream_set_local_media_direction(stream, SENDRECV);
	}

	return;
}

char *ast_sip_session_get_video_resolution_according_type(struct ast_sip_session *session, enum video_resolution resolution)
{
	if (!session) {
		return NULL;
	}

	switch (resolution) {
		case VIDEO_RESOLUTION_NONE:
			ast_log(LOG_DEBUG, "[session:%p] Can't get video resolution, just decline this video, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			break;
		case VIDEO_RESOLUTION_360P:
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
				return H264_PROFILE_LEVELID_360P;
			} else {
				return H264_GVC_PROFILE_LEVELID_360P;
			}
			break;
		case VIDEO_RESOLUTION_720P:
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
				return H264_PROFILE_LEVELID_720P;
			} else {
				return H264_GVC_PROFILE_LEVELID_720P;
			}
			break;
		case VIDEO_RESOLUTION_1080P:
			if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
				return H264_PROFILE_LEVELID_1080P;
			} else {
				return H264_GVC_PROFILE_LEVELID_1080P;
			}
			break;
		default:
			break;
	}

	return NULL;
}

char *ast_sip_session_get_imageattr_according_type(enum video_resolution resolution)
{
	switch (resolution) {
		case VIDEO_RESOLUTION_NONE:
			ast_log(LOG_DEBUG, "Can't get video resolution, just decline this video\n");
			break;
		case VIDEO_RESOLUTION_360P:
			return H264_IMAGEATTR_360P;
		case VIDEO_RESOLUTION_720P:
			return H264_IMAGEATTR_720P;
		case VIDEO_RESOLUTION_1080P:
			return H264_IMAGEATTR_1080P;
		default:
			break;
	}

	return NULL;
}

char *ast_sip_session_get_wave_imageattr_according_type(enum video_resolution resolution)
{
	switch (resolution) {
		case VIDEO_RESOLUTION_NONE:
			ast_log(LOG_DEBUG, "Can't get video resolution, just decline this video\n");
			break;
		case VIDEO_RESOLUTION_360P:
			return H264_WAVE_IMAGEATTR_360P;
		case VIDEO_RESOLUTION_720P:
			return H264_WAVE_IMAGEATTR_720P;
		case VIDEO_RESOLUTION_1080P:
			return H264_WAVE_IMAGEATTR_1080P;
		default:
			break;
	}

	return NULL;
}

int ast_sip_session_get_as_according_type(struct ast_sip_session *session, enum video_resolution resolution)
{
	int as_value = 0;

	switch (resolution) {
		case VIDEO_RESOLUTION_NONE:
			ast_log(LOG_DEBUG, "Can't get AS, because has no video resource\n");
			break;
		case VIDEO_RESOLUTION_360P:
			as_value = AST_360P_AS;
			break;
		case VIDEO_RESOLUTION_720P:
			as_value = AST_720P_AS;
			break;
		case VIDEO_RESOLUTION_1080P:
			as_value = AST_1080P_AS;
			break;
		default :
			break;
	}

	return as_value;
}

int ast_sip_session_get_tias_according_type(struct ast_sip_session *session, enum video_resolution resolution)
{
	int tias_value = 0;

	switch (resolution) {
		case VIDEO_RESOLUTION_NONE:
			ast_log(LOG_DEBUG, "Can't get TIAS, because has no video resource\n");
			break;
		case VIDEO_RESOLUTION_360P:
			tias_value = AST_360P_TIAS;
			break;
		case VIDEO_RESOLUTION_720P:
			tias_value = AST_720P_TIAS;
			break;
		case VIDEO_RESOLUTION_1080P:
			tias_value = AST_1080P_TIAS;
			break;
		default:
			break;
	}

	return tias_value;
}

void ast_sip_session_set_video_format(struct ast_sip_session *session, struct ast_stream *stream, int resolution)
{
	unsigned int index = 0;
	struct ast_format *format = NULL;
	struct ast_format_cap *caps = NULL;
	char *profile_level_id_str = NULL;
	RAII_VAR(struct ast_format *, video_format, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, video_caps, NULL, ao2_cleanup);

	if (!session || !stream) {
		return;
	}

	caps = ast_stream_get_formats(stream);
	if (!caps) {
		return;
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		if (strcasecmp(ast_format_get_name(format), "h264")) {
			ao2_ref(format, -1);
			continue;
		}

		profile_level_id_str = ast_sip_session_get_video_resolution_according_type(session, resolution);
		if (profile_level_id_str) {
			ast_log(LOG_DEBUG, "[session:%p] Modify video profile_level_id to:%s, endpoint:%s\n",
				session, profile_level_id_str, ast_sorcery_object_get_id(session->endpoint));
			video_format = ast_format_attribute_set(format, "profile-level-id", profile_level_id_str);
			/* Todo: for SFU INVITE, we set packetization-mode to 1 as default */
			ast_format_attribute_set_directly(video_format, "packetization-mode", "1");

			video_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
			if (!video_caps) {
				ao2_ref(format, -1);
				break;
			}

			ast_format_cap_append(video_caps, video_format, 0);
			ast_format_cap_replace_from_cap(caps, video_caps, AST_MEDIA_TYPE_VIDEO);
		}

		ao2_ref(format, -1);

	}

	return;
}

static void set_main_video_as_tias_for_sfu(struct ast_sip_session *session, struct ast_stream *stream)
{
	int as = 0;
	int tias = 0;
	char as_str[16] = {0};
	char tias_str[32] = {0};

	enum ast_stream_video_resolution resolution = AST_STREAM_VIDEO_RESOLUTION_NONE;

	if (!session || !stream) {
		return;
	}

	if (!ast_stream_judge_if_main_video(stream)) {
		return;
	}

	resolution = ast_stream_get_video_resolution(stream);

	as = ast_sip_session_get_as_according_type(session, resolution);
	if (as) {
		snprintf(as_str, sizeof(as_str), "%d", as);
		ast_stream_set_metadata(stream, "AS", as_str);
		ast_log(LOG_DEBUG, "[session:%p] set_main_video_as_tias_for_sfu, AS:%s, endpoint:%s\n",
			session, as_str, ast_sorcery_object_get_id(session->endpoint));
	}

	tias = ast_sip_session_get_tias_according_type(session, resolution);
	if (tias) {
		snprintf(tias_str, sizeof(tias_str), "%d", tias);
		ast_stream_set_metadata(stream, "TIAS", tias_str);
		ast_log(LOG_DEBUG, "[session:%p] set_main_video_as_tias_for_sfu, TIAS:%s, endpoint:%s\n",
			session, tias_str, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

/* For slides video AS/TIAS we use default value */
static void set_slides_video_as_tias_for_sfu(struct ast_sip_session *session, struct ast_stream *stream)
{
	char as_str[16] = {0};
	char tias_str[32] = {0};

	if (!stream) {
		return;
	}

	snprintf(as_str, sizeof(as_str), "%d", AST_SLIDES_VIDEO_AS);
	ast_stream_set_metadata(stream, "AS", as_str);
	ast_log(LOG_DEBUG, "[session:%p] set_slides_video_as_tias_for_sfu, AS:%s for endpoint:%s\n",
		session, as_str, ast_sorcery_object_get_id(session->endpoint));

	snprintf(tias_str, sizeof(tias_str), "%d", AST_SLIDES_VIDEO_TIAS);
	ast_stream_set_metadata(stream, "TIAS", tias_str);
	ast_log(LOG_DEBUG, "[session:%p] set_slides_video_as_tias_for_sfu, TIAS:%s for endpoint:%s\n",
		session, tias_str, ast_sorcery_object_get_id(session->endpoint));

	return;
}

void ast_sip_session_set_as_tias_for_sfu(struct ast_sip_session *session, struct ast_stream *stream)
{
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!stream) {
		return;
	}

	sub_type = ast_stream_get_sub_type(stream);

	switch (sub_type) {
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			set_main_video_as_tias_for_sfu(session, stream);
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			set_slides_video_as_tias_for_sfu(session, stream);
			break;
		default :
			break;
	}

	return;
}

void ast_sip_session_set_imageattr_for_stream(struct ast_sip_session *session,
	struct ast_stream *stream, enum video_resolution resolution)
{
	char *imageattr = NULL;

	if (!session || !stream) {
		return;
	}

	if ((session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) ||
		(session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		imageattr = ast_sip_session_get_wave_imageattr_according_type(resolution);
	} else {
		imageattr = ast_sip_session_get_imageattr_according_type(resolution);
	}

	if (imageattr) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_imageattr_for_stream, imageattr:%s, endpoint:%s\n",
			session, imageattr, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_set_metadata(stream, "IMAGEATTR", imageattr);
	}

	return;
}

/* Need to reduce refs when get format use this function */
struct ast_format *ast_sip_session_get_video_format(struct ast_stream *stream, const char *codec_name)
{
	unsigned int index = 0;
	struct ast_format *format = NULL;
	struct ast_format_cap *caps = NULL;

	if (!stream) {
		return NULL;
	}

	caps = ast_stream_get_formats(stream);
	if (!caps) {
		return NULL;
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		if (strcasecmp(ast_format_get_name(format), codec_name)) {
			ao2_ref(format, -1);
			continue;
		}

		return format;
	}

	return NULL;
}

void ast_sip_session_set_all_stream_change_ssrc(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		ast_log(LOG_WARNING, "set_all_stream_change_ssrc params error\n");
		return;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}

		ast_stream_set_change_ssrc(stream, 1);
	}

	return;
}

struct ast_stream *ast_sip_session_get_pending_main_video_stream(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}
		return stream;
	}

	return NULL;
}

struct ast_stream *ast_sip_session_get_active_main_video_stream(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}
		return stream;
	}

	return NULL;
}

struct ast_stream *ast_sip_session_get_pending_stream_by_subtype(struct ast_sip_session *session,enum ast_media_sub_type sub_type)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (ast_stream_get_sub_type(stream) != sub_type) {
			continue;
		}
		return stream;
	}

	return NULL;
}

struct ast_stream *ast_sip_session_get_pending_slides_video_stream(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			continue;
		}
		return stream;
	}

	return NULL;
}

struct ast_stream *ast_sip_session_get_active_slides_video_stream(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			continue;
		}
		return stream;
	}

	return NULL;
}

enum video_resolution ast_sip_session_get_main_video_resource(struct ast_sip_session *session, unsigned int h264_level)
{
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;

	if (h264_level >= AST_1080P_LEVEL_ID) {
		resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, AST_STREAM_VIDEO_RESOLUTION_1080P);
	} else if (h264_level >= AST_720P_LEVEL_ID && h264_level < AST_1080P_LEVEL_ID) {
		resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, AST_STREAM_VIDEO_RESOLUTION_720P);
	} else {
		resolution = ast_conf_get_main_video_resource(session->conf_room, session->channel, AST_STREAM_VIDEO_RESOLUTION_360P);
	}

	return resolution;
}

unsigned int ast_sip_session_get_h264_level(struct ast_sip_session *session, unsigned int is_nway)
{
	struct ast_stream *stream = NULL;
	RAII_VAR(struct ast_format *, format, NULL, ao2_cleanup);
	const unsigned int *h264_level = NULL;

	/* Don't need to check params, we already check before */
	if (!is_nway) {
		stream = ast_sip_session_get_pending_main_video_stream(session);
	} else {
		stream = ast_sip_session_get_active_main_video_stream(session);
	}
	if (!stream) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get main video stream, don't get H264 video level, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	format = ast_sip_session_get_video_format(stream, "h264");
	if (!format) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get video format, don't get H264 video level, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	h264_level = ast_format_attribute_get(format, "level");
	if (!h264_level) {
		ast_log(LOG_WARNING, "[session:%p] Can't get profile level for H264, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_get_h264_level get H264 level:%d, endpoint:%s\n",
		session, *h264_level, ast_sorcery_object_get_id(session->endpoint));

	return *h264_level;
}

const char *ast_sip_session_sfu_codec2str(enum ast_sfu_video_codec codec)
{
	switch (codec) {
		case AST_SFU_VIDEO_H264:
			return "h264";
		case AST_SFU_VIDEO_H265:
			return "h265";
		case AST_SFU_VIDEO_VP9:
			return "vp9";
		default:
			return "<unknown>";
	}
}

unsigned int ast_sip_session_judge_if_has_valid_slides_video(struct ast_sip_session *session)
{
	unsigned int i = 0;
	unsigned int res = 0;
	struct ast_stream *stream = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		ast_log(LOG_WARNING, "ast_sip_session_judge_if_has_valid_slides_video params error\n");
		return 0;
	}

	if (ast_sip_session_get_only_support_audio()) {
		return 0;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}

		if (ast_stream_get_sub_type(stream) != AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) {
			continue;
		}

		if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED || ast_stream_get_state(stream) == AST_STREAM_STATE_INACTIVE) {
			res = 0;
		} else {
			res = 1;
		}

		ast_log(LOG_DEBUG, "[session:%p] Find slides video stream, and state is %d, endpoint:%s\n",
			session, ast_stream_get_state(stream), ast_sorcery_object_get_id(session->endpoint));

		break;
	}

	return res;
}

struct ast_sip_session_media *ast_sip_session_get_bfcp_active_media(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->active_media_state) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&session->active_media_state->sessions); i++) {
		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
		if (!session_media) {
			continue;
		}
		
		if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
			session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
			ast_log(LOG_DEBUG, "[session:%p] Find bfcp active media for endpoint:%s, sub_type:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint), ast_codec_media_sub_type2str(session_media->sub_type));
			return session_media;
		}
	}

	return NULL;
}

struct ast_sip_session_media *ast_sip_session_get_bfcp_pending_media(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->pending_media_state) {
		return NULL;
	}

	for (i = 0; i < AST_VECTOR_SIZE(&session->pending_media_state->sessions); i++) {
		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
		if (!session_media) {
			continue;
		}

		if (session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
			session_media->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
			ast_log(LOG_DEBUG, "[session:%p] Find bfcp pending media for endpoint:%s, sub_type:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint), ast_codec_media_sub_type2str(session_media->sub_type));
			return session_media;
		}
	}

	return NULL;
}

unsigned int ast_sip_session_need_create_bfcp_media(struct ast_sip_session *session)
{
	if (!session) {
		return 0;
	}

	if (ast_sip_session_support_bfcp(session)) {
		return 1;
	}

	return 0;
}

void ast_sip_session_append_clone_stream(struct ast_sip_session *session, struct ast_stream *req_stream)
{
	struct ast_stream *clone_stream = NULL;

	clone_stream = ast_stream_clone(req_stream, NULL);
	if (!clone_stream) {
		ast_log(LOG_WARNING, "[session:%p] Can't clone the stream for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->pending_media_state->topology) {
		session->pending_media_state->topology = ast_stream_topology_alloc();
		if (!session->pending_media_state->topology) {
			ast_log(LOG_WARNING, "[session:%p] Can't alloc topology for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
	}

	if (ast_stream_topology_append_stream(session->pending_media_state->topology, clone_stream) < 0) {
		ast_log(LOG_WARNING, "[session:%p] Append stream to topology failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_stream_free(clone_stream);
		return;
	}

	return;
}

struct ast_sip_session_media *ast_sip_session_get_active_session_media(struct ast_sip_session *session,
	enum ast_media_sub_type sub_type)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->active_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (ast_stream_get_sub_type(stream) != sub_type) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
		return session_media;
	}

	return NULL;
}

struct ast_sip_session_media *ast_sip_session_get_pending_session_media(struct ast_sip_session *session,
	enum ast_media_sub_type sub_type)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->pending_media_state || !session->pending_media_state->topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_media_state->topology); i++) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
		if (!stream) {
			continue;
		}
		if (ast_stream_get_sub_type(stream) != sub_type) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
		return session_media;
	}

	return NULL;
}

struct ast_sip_session_media *ast_sip_session_get_session_media(struct ast_sip_session *session,
	enum ast_media_sub_type sub_type)
{
	struct ast_sip_session_media *session_media = NULL;

	session_media = ast_sip_session_get_active_session_media(session, sub_type);
	if (!session_media) {
		session_media = ast_sip_session_get_pending_session_media(session, sub_type);
		if (!session_media) {
			ast_log(LOG_WARNING, "[session:%p] Can't get media type:%s session_media, endpoint:%s\n",
				session, ast_codec_media_sub_type2str(sub_type), ast_sorcery_object_get_id(session->endpoint));
			return NULL;
		}
	}

	return session_media;
}

static unsigned int process_bfcp_stream(struct ast_sip_session *session, struct ast_stream *req_stream)
{
	if (!session || !req_stream) {
		return 0;
	}

	if (!ast_sip_session_need_create_bfcp_media(session)) {
		ast_log(LOG_DEBUG, "[session:%p] It's BFCP media, but register extension don't support BFCP, just discard for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] It's BFCP media, just clone the stream for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_append_clone_stream(session, req_stream);

	if (ast_stream_get_state(req_stream) != AST_STREAM_STATE_REMOVED) {
		if (ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP) {
			session->bfcp_para.has_udp_bfcp = 1;
		} else if (ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
			session->bfcp_para.has_tcp_bfcp = 1;
		}
	}

	return 0;
}

static unsigned int process_fecc_stream(struct ast_sip_session *session)
{
	if (!session) {
		return 0;
	}

	if (session->is_webrtc_call) {
		ast_log(LOG_DEBUG, "[session:%p] Req stream is FECC media, and local it's webrtc, just discard for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	return 1;
}

/* special media contain BFCP/FECC, return 0 don't need process anymore just continue, return 1 need to process later */
unsigned int ast_sip_session_process_special_stream(struct ast_sip_session *session, struct ast_stream *req_stream)
{
	if (ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {
		return process_bfcp_stream(session, req_stream);
	}

	if (ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_FECC) {
		return process_fecc_stream(session);
	}

	return 1;
}

static int check_removed_stream_in_active_topology(struct ast_sip_session* session, struct ast_stream* stream)
{
	int i = 0;
	int res = 0;
	int media_streams = 0;
	struct ast_stream* active_stream = NULL;
	struct ast_stream_topology* topology = NULL;

	if (!session || !session->active_media_state || !session->active_media_state->topology || !stream) {
		return 0;
	}

	topology = session->active_media_state->topology;
	media_streams = ast_stream_topology_get_count(topology);

	for (i = 0; i < media_streams; ++ i) {
		active_stream = ast_stream_topology_get_stream(topology, i);
		if (active_stream && ast_stream_get_sub_type(active_stream) == ast_stream_get_sub_type(stream)) {
			res = 1;
			break;
		}
	}

	ast_log(LOG_DEBUG, "[check_removed_stream_in_active_topology] The find result is: %d\n", res);

	return res;
}

int ast_sip_session_process_special_stream_for_reinvite(struct ast_sip_session* session, struct ast_stream* req_stream)
{
	if (ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP ||
		ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP) {

		if (ast_stream_get_state(req_stream) == AST_STREAM_STATE_REMOVED
			&& !check_removed_stream_in_active_topology(session, req_stream)) {
			return 0;
		}

		return process_bfcp_stream(session, req_stream);
	}

	if (ast_stream_get_sub_type(req_stream) == AST_MEDIA_SUB_APPLICATION_TYPE_FECC) {
		return process_fecc_stream(session);
	}

	return 1;
}

int ast_sip_session_get_joint_caps(struct ast_sip_session *session, struct ast_stream *req_stream,
	struct ast_format_cap *joint_cap, struct ast_format_cap *req_cap)
{
	if (!session || !session->endpoint || !req_stream || !joint_cap || !req_cap) {
		return -1;
	}

	if (ast_sip_session_get_local_negotiate_prefer() && ast_stream_get_type(req_stream) == AST_MEDIA_TYPE_AUDIO) {
		ast_log(LOG_DEBUG, "[session:%p] Create outgoing, enable local negotiate prefer, use local codec prefer for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_format_cap_get_compatible(session->endpoint->media.codecs, req_cap, joint_cap);
	} else {
		ast_format_cap_get_compatible(req_cap, session->endpoint->media.codecs, joint_cap);
	}

	if (!ast_format_cap_count(joint_cap)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get joint_cap for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	return 0;
}

void ast_sip_session_process_special_audio_codec(struct ast_sip_session *session,
	struct ast_stream *req_stream, struct ast_format_cap *joint_cap)
{
	if (!session || !req_stream || !joint_cap) {
		return;
	}

	if (session->is_webrtc_call && !session->support_gs_fec && ast_format_cap_has_red(joint_cap)) {
		ast_format_cap_remove_red(joint_cap);
	}

	if (session->is_webrtc_call && !session->support_gs_fec && ast_format_cap_has_audio_fec(joint_cap)) {
		ast_format_cap_remove_audio_fec(joint_cap);
	}

	/* Move red codec to last postion */
	if (ast_format_cap_has_red(joint_cap)) {
		ast_format_cap_remove_red(joint_cap);
		ast_format_cap_add_red(joint_cap);
	}

	ast_sip_session_change_gs_fec_version(joint_cap);

	if (ast_format_cap_has_opus(joint_cap)) {
		/* set useinbandfec true as default */
		ast_sip_session_set_format_useinbandfec(session, joint_cap, req_stream, 1);
		ast_sip_session_set_format_dtx(session, joint_cap, req_stream);
	}

	return;
}

/* If we only has RTX or video FEC codec for video, don't add video stream */
int ast_sip_session_process_special_video_codec(struct ast_sip_session *session, struct ast_format_cap *joint_cap)
{
	if (ast_format_cap_has_rtx(joint_cap) && ast_format_cap_count(joint_cap) == 1) {
		ast_log(LOG_DEBUG, "[session:%p] joint_cap only has RTX, don't add stream for this media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (ast_format_cap_has_video_fec(joint_cap) && ast_format_cap_count(joint_cap) == 1) {
		ast_log(LOG_DEBUG, "[session:%p] joint_cap only has video FEC, don't add stream for this media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (ast_format_cap_has_rtx(joint_cap) &&
		ast_format_cap_has_video_fec(joint_cap) &&
		ast_format_cap_count(joint_cap) == 2) {
		ast_log(LOG_DEBUG, "[session:%p] joint_cap only has RTX and video FEC, don't add stream for this media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	return 0;
}

struct ast_sip_session_media *ast_sip_session_get_host_main_video_media(struct ast_channel *channel, const char *conf_room)
{
	unsigned int i = 0;
	unsigned int find = 0;
	int active_media_count = 0;
	RAII_VAR(struct ast_channel *, host_channel, NULL, ao2_cleanup);
	struct ast_sip_channel_pvt *chan_pvt = NULL;
	struct ast_sip_session *host_session = NULL;
	struct ast_sip_session_media *host_session_media = NULL;
	struct ast_stream *stream = NULL;

	if (!channel || ast_strlen_zero(conf_room)) {
		return NULL;
	}

	host_channel = ast_conf_get_sfu_moderator(channel, conf_room);
	if (!host_channel) {
		return NULL;
	}

	chan_pvt = ast_channel_tech_pvt(host_channel);
	if (!chan_pvt) {
		ast_log(LOG_WARNING, "set_avs_sip_msid chan_pvt is NULL\n");
		return NULL;
	}

	host_session = chan_pvt->session;
	if (!host_session || !host_session->active_media_state || !host_session->active_media_state->topology) {
		ast_log(LOG_WARNING, "set_avs_sip_msid session or active_media_state is NULL\n");
		return NULL;
	}

	active_media_count = ast_stream_topology_get_count(host_session->active_media_state->topology);
	for (i = 0; i < active_media_count; i++) {
		host_session_media = AST_VECTOR_GET(&host_session->active_media_state->sessions, i);
		if (host_session_media->sub_type != AST_MEDIA_SUB_VIDEO_TYPE_MAIN) {
			continue;
		}

		stream = ast_stream_topology_get_stream(host_session->active_media_state->topology, i);
		if (ast_stream_has_video_resource(stream)) {
			find = 1;
			break;
		}
	}

	if (find == 1) {
		return host_session_media;
	} else {
		return NULL;
	}
}

void ast_sip_session_remove_attrs_for_disable_media(pjmedia_sdp_media *media)
{
	if (!media) {
		return;
	}

	/* For RFC5245, we should to remove all ice attributes when disable media */
	pjmedia_sdp_media_remove_all_attr(media, "candidate");
	pjmedia_sdp_media_remove_all_attr(media, "remote-candidates");
	pjmedia_sdp_media_remove_all_attr(media, "ice-lite");
	pjmedia_sdp_media_remove_all_attr(media, "ice-mismatch");
	pjmedia_sdp_media_remove_all_attr(media, "ice-ufrag");
	pjmedia_sdp_media_remove_all_attr(media, "ice-pwd");
	pjmedia_sdp_media_remove_all_attr(media, "ice-options");

	/* when disable media, can't carry bundle-only attr */
	pjmedia_sdp_media_remove_all_attr(media, "bundle-only");

	return;
}

static char *get_mid_from_media(pjmedia_sdp_media *media)
{
	pjmedia_sdp_attr *attr = NULL;
	char *mid = NULL;

	if (!media) {
		return NULL;
	}

	if (!(attr = pjmedia_sdp_media_find_attr2(media, "mid", NULL))) {
		return NULL;
	}

	mid = ast_calloc(1, MAX_MID_LEN);
	if (!mid) {
		return NULL;
	}

	char attr_value[pj_strlen(&attr->value) + 1];
	ast_copy_pj_str(attr_value, &attr->value, sizeof(attr_value));
	strncpy(mid, attr_value, MAX_MID_LEN - 1);

	return mid;
}

static void disable_session_media_bundle_group(struct ast_sip_session *session, pjmedia_sdp_media *media)
{
	char media_type[16] = {0};
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !media) {
		return;
	}

	memset(media_type, 0, sizeof(media_type));
	ast_copy_pj_str(media_type, &media->desc.media, sizeof(media_type));

	ast_log(LOG_DEBUG, "[session:%p] disable_session_media_bundle_group media type:%s, endpoint:%s\n",
		session, media_type, ast_sorcery_object_get_id(session->endpoint));

	if (ast_sip_session_judge_if_main_video(media)) {
		session_media = ast_sip_session_get_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
		if (!session_media) {
			ast_log(LOG_WARNING, "[session:%p] Can't get main video session_media, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		session_media->bundle_group = -1;
	} else if (ast_sip_session_judge_if_slides_video(media)) {
		session_media = ast_sip_session_get_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
		if (!session_media) {
			ast_log(LOG_WARNING, "[session:%p] Can't get slides video session_media, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}
		session_media->bundle_group = -1;
	} else {
		ast_log(LOG_WARNING, "[session:%p] Can't disable media type:%s except main/slides video, endpoint:%s\n",
			session, media_type, ast_sorcery_object_get_id(session->endpoint));
	}

	return;
}

void ast_sip_session_remove_media_from_bundle_group(struct ast_sip_session *session,
	pj_pool_t *pool, pjmedia_sdp_session *sdp, pjmedia_sdp_media *media)
{
	int index = 0;
	int find = 0;
	int bundle_added = 0;
	char *mid = NULL;
	pj_str_t stmp;
	struct ast_str *bundle_group_string = NULL;
	pjmedia_sdp_attr *old_bundle_group_attr = NULL;
	pjmedia_sdp_attr *new_bundle_group_attr = NULL;

	if (!session || !sdp || !media) {
		return;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
		return;
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return;
	}

	mid = get_mid_from_media(media);
	if (!mid) {
		ast_log(LOG_WARNING, "[session:%p] Can't get mid from media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	bundle_group_string = ast_str_create(64);
	if (!bundle_group_string) {
		ast_log(LOG_WARNING, "[session:%p] Can't alloc bundle group string, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_free(mid);
		mid = NULL;
		return;
	}

	for (index = 0; index < sdp->attr_count; ++index) {
		old_bundle_group_attr = sdp->attr[index];
		char value[pj_strlen(&old_bundle_group_attr->value) + 1], *mids = value, *attr_mid;

		if (pj_strcmp2(&old_bundle_group_attr->name, "group") || pj_strncmp2(&old_bundle_group_attr->value, "BUNDLE", 6)) {
			continue;
		}

		ast_copy_pj_str(value, &old_bundle_group_attr->value, sizeof(value));

		/* Skip the BUNDLE at the front */
		mids += 7;

		while ((attr_mid = strsep(&mids, " "))) {
			if (!strcmp(attr_mid, mid)) {
				ast_log(LOG_DEBUG, "[session:%p] Find the matched mid, remove from bundle group, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				find = 1;
				continue;
			}
			if (!bundle_added) {
				ast_str_set(&bundle_group_string, 0, "BUNDLE %s", attr_mid);
				bundle_added = 1;
				continue;
			}
			ast_str_append(&bundle_group_string, 0, " %s", attr_mid);
		}

		/* Todo:Now we just consider all media bundle together */
		break;
	}

	if (find) {
		pjmedia_sdp_attr_remove(&sdp->attr_count, sdp->attr, old_bundle_group_attr);
		new_bundle_group_attr = pjmedia_sdp_attr_create(pool, "group", pj_cstr(&stmp, ast_str_buffer(bundle_group_string)));
		pjmedia_sdp_attr_add(&sdp->attr_count, sdp->attr, new_bundle_group_attr);
		disable_session_media_bundle_group(session, media);
	}

	ast_free(mid);
	mid = NULL;

	ast_free(bundle_group_string);
	bundle_group_string = NULL;

	return;
}

unsigned int ast_sip_session_jugde_if_media_support(struct ast_sip_session *session, enum ast_media_sub_type sub_type)
{
	if (!session) {
		return 0;
	}

	if (!ast_sip_session_get_only_support_audio()) {
		return 1;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC && (sub_type == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES)) {
		return 1;
	}

	if ((sub_type == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES) ||
		(sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP) ||
		(sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_TCP)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't support BFCP/slides video for A series,  for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	return 1;
}

/* For gswave can't set port as zero, need to change direction to inactive instead */
void ast_sip_session_disable_stream(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!session || !stream) {
		return;
	}

	switch (session->terminal_type) {
		case AST_TERMINAL_TYPE_WEBRTC:
			ast_stream_set_state(stream, AST_STREAM_STATE_INACTIVE);
			ast_stream_set_local_media_direction(stream, INACTIVE);
			break;
		default :
			ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
			break;
	}

	return;
}


static void publish_answered_else_event2im(const char* caller_num,const char* dst_num, const char* reason)
{
	if (!caller_num || !dst_num || !reason)
		return ;

	/* send event to cti pbxmid */
	manager_event(EVENT_FLAG_CALL, "AnsweredByElse",
									"Caller_Num: %s\r\n"
									"Caller_Name: %s\r\n"
									"Callee_Num: %s\r\n"
									"Reason: %s\r\n"
									"AnsweredInstance: %s\r\n",
					caller_num,"",dst_num,reason,"");

	ast_log(LOG_DEBUG, "Send App AnsweredRefuesdByElse Event  Caller_Num:%s Callee_Num:%s Reason:%s\n",caller_num,dst_num,reason);

}

void ast_sip_session_cancel_pending_call(struct ast_sip_session *session)
{
	struct ast_sip_session *cur_session = NULL;
	const char *endpoint_id = NULL;
	const char *cur_endpoint_id = NULL;
	pjsip_sip_uri *sip_uri = NULL;
	char to_user[128] = { 0 };
	pjsip_sip_uri *cur_sip_uri = NULL;
	char cur_to_user[128] = { 0 };

	endpoint_id = ast_sorcery_object_get_id(session->endpoint);
	if (!endpoint_id) {
		return;
	}

	if (!strncasecmp(endpoint_id, "conf", 4)) {
		ast_log(LOG_DEBUG, "[session:%p] It's conference anonymous account, don't need cancel by this func, endpoint:%s\n",
			session, endpoint_id);
		return;
	}

	sip_uri = pjsip_uri_get_uri(session->inv_session->dlg->remote.info->uri);
	if (!sip_uri) {
		ast_log(LOG_WARNING, "[session:%p] can't get remote uri, endpoint:%s\n", session, endpoint_id);
		return;
	}
	ast_copy_pj_str(to_user, &sip_uri->user, sizeof(to_user));

	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_TRAVERSE(&sip_sessions, cur_session, next) {
		if (!cur_session->endpoint) {
			continue;
		}

		cur_endpoint_id = ast_sorcery_object_get_id(cur_session->endpoint);
		if (!cur_endpoint_id) {
			continue;
		}

		if (strcmp(endpoint_id, cur_endpoint_id)) {
			continue;
		}

		if (strcmp(session->conf_room, cur_session->conf_room)) {
			continue;
		}

		if (cur_session->inv_session &&
			cur_session->inv_session->role == PJSIP_ROLE_UAC &&
			(cur_session->inv_session->state < PJSIP_INV_STATE_CONNECTING ||(cur_session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED &&  cur_session->is_gswave_app_call)) ){
			ast_log(LOG_DEBUG, "[session:%p] Receive call, cancel pending session:%p call for the same endpoint:%s\n",
				session, cur_session, endpoint_id);

			/*Start added by zyxia for Bug 267037,2023.7.28*/
			if(cur_session->is_gswave_app_call)
				cur_session->gswave_app_session_active = 1;//mark it tell wakeup thread, dont send invite to app
			/*End added by zyxia for Bug 267037,2023.7.28*/

			/*Start added by zyxia for Bug 266962*/
			if(cur_session->channel && ast_channel_get_terminate_type(cur_session->channel) == AST_TERMINAL_TYPE_GSWAVE_MOBILE){
				publish_answered_else_event2im(session->conf_room,endpoint_id,"answer");
			}
			/*Start added by zyxia for Bug 266962*/

			cur_sip_uri = pjsip_uri_get_uri(cur_session->inv_session->dlg->remote.info->uri);
			if (!cur_sip_uri) {
				ast_log(LOG_WARNING, "[session:%p] ast_sip_session_cancel_invite_call cur_session:%p has no remote uri, endpoint:%s\n",
					session, cur_session, endpoint_id);
				continue;
			}
			ast_copy_pj_str(cur_to_user, &cur_sip_uri->user, sizeof(cur_to_user));

			/* If incoming from the same trunk, if ID can't match, we can't distinguish, don't need cancel */
			if (!strncasecmp(endpoint_id, "trunk_", 6) && strcmp(to_user, cur_to_user)) {
				ast_log(LOG_WARNING, "[session:%p] ast_sip_session_cancel_invite_call cur_to_user:%s is not equal as incoming user:%s, endpoint:%s\n",
					session, cur_to_user, to_user, endpoint_id);
				continue;
			}

			if (cur_session->channel) {
				ast_channel_conf_hangupcause_set(cur_session->channel, AST_CAUSE_ANSWERED_ELSEWHERE);
				ast_channel_hangupcause_set(cur_session->channel, AST_CAUSE_ANSWERED_ELSEWHERE);
			}
			ast_sip_session_terminate(cur_session, 603);
		}
	}
	AST_LIST_UNLOCK(&sip_sessions);

	return;
}

void ast_sip_session_cancel_pending_call_according_id(const char *conf_id, const char *id)
{
	struct ast_sip_session *cur_session = NULL;
	const char *exten = NULL;

	if (!conf_id || !id) {
		return;
	}

	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_TRAVERSE(&sip_sessions, cur_session, next) {
		if (!cur_session->channel) {
			continue;
		}

		exten = ast_channel_exten(cur_session->channel);
		if (ast_strlen_zero(exten)) {
			ast_log(LOG_DEBUG, "Can't get channel exten for session:%p\n", cur_session);
			continue;
		}

		if (ast_strlen_zero(cur_session->conf_room)) {
			continue;
		}

		if (strcmp(id, exten) ||
			strcmp(conf_id, cur_session->conf_room)) {
			continue;
		}

		ast_log(LOG_DEBUG, "Matched conf number and exten id, cancel pending session:%p call for the same exten:%s\n", cur_session, id);

		if (cur_session->inv_session &&
			cur_session->inv_session->role == PJSIP_ROLE_UAC &&
			cur_session->inv_session->state < PJSIP_INV_STATE_CONNECTING) {
			ast_log(LOG_DEBUG, "Receive call, cancel pending session:%p call for the same endpoint:%s\n", cur_session, id);
			if (cur_session->channel) {
				ast_channel_conf_hangupcause_set(cur_session->channel, AST_CAUSE_ANSWERED_ELSEWHERE);
				ast_channel_hangupcause_set(cur_session->channel, AST_CAUSE_ANSWERED_ELSEWHERE);
			}
			ast_sip_session_terminate(cur_session, 603);
		}
	}
	AST_LIST_UNLOCK(&sip_sessions);

	return;
}

/* For SFU, if there is terminal call to conference active, don't invite again */
int ast_sip_session_cancel_invite_call(struct ast_sip_session *session, const struct ast_channel *chan,const char* dest)
{
	int res = 0;
	const char *conf_invite = NULL;
	const char *conf_room = NULL;
	struct ast_sip_session *cur_session = NULL;
	const char *endpoint_id = NULL;
	const char *cur_endpoint_id = NULL;
	pjsip_sip_uri *sip_uri = NULL;
	char to_user[128] = { 0 };
	pjsip_sip_uri *from_sip_uri = NULL;
	char from_user[128] = { 0 };
	pjsip_sip_uri *invite_sip_uri = NULL;
	char invite_to_user[128] = { 0 };

	endpoint_id = ast_sorcery_object_get_id(session->endpoint);
	if (!endpoint_id) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call can't get endpoint_id\n", session);
		return 0;
	}


	if(session->is_sfu && session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE ){
		struct sip_waveapp_info* app_info = NULL;
		
		app_info = ast_sip_find_waveapp_info_by_uri(session->endpoint,dest);
		
		if(app_info == NULL){
			ast_log(LOG_DEBUG, "ast_sip_session_cancel_invite_call can not find appinfo by uri:%s,endpoinrt:%s\n", dest, endpoint_id);
			return 0;
		}

		if(app_info->conf_invite == 0)
		{
			ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call app dont support receive conf invte,cancel it, endpoint:%s\n",
				session, endpoint_id);
			return 1;
		}
	}

	conf_invite = pbx_builtin_getvar_helper((struct ast_channel *)chan, "CONF_SCHEDULE_INVITE");
	if (ast_strlen_zero(conf_invite) || strcasecmp(conf_invite, "yes")) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call it's not schedule invite, endpoint:%s\n",
			session, endpoint_id);
		return 0;
	}

	conf_room = pbx_builtin_getvar_helper((struct ast_channel *)chan, "CONF_ROOM");
	if (ast_strlen_zero(conf_room)) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call can't get conf_room, endpoint:%s\n",
			session, endpoint_id);
		return 0;
	}

	invite_sip_uri = pjsip_uri_get_uri(session->inv_session->dlg->remote.info->uri);
	if (!invite_sip_uri) {
		ast_log(LOG_WARNING, "[session:%p] can't get remote uri, endpoint:%s\n", session, endpoint_id);
		return 0;
	}
	ast_copy_pj_str(invite_to_user, &invite_sip_uri->user, sizeof(invite_to_user));
	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call invite_to_user:%s, endpoint:%s\n",
		session, invite_to_user, endpoint_id);

	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_TRAVERSE(&sip_sessions, cur_session, next) {
		if (!cur_session->endpoint) {
			ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call can't get endpoint, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		cur_endpoint_id = ast_sorcery_object_get_id(cur_session->endpoint);
		if (!cur_endpoint_id) {
			ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call can't get cur_endpoint_id, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		if (strcmp(endpoint_id, cur_endpoint_id)) {
			ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call cur_endpoint_id:%s is not the same, endpoint:%s\n",
				session, cur_endpoint_id, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		if (!cur_session->inv_session) {
			ast_log(LOG_WARNING, "[session:%p] ast_sip_session_cancel_invite_call cur_session:%p has no inv_session, endpoint:%s\n",
				session, cur_session, endpoint_id);
			continue;
		}

		if (cur_session->inv_session->role == PJSIP_ROLE_UAC) {
			continue;
		}

		if (cur_session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
			continue;
		}

		if (!cur_session->inv_session->dlg) {
			ast_log(LOG_WARNING, "[session:%p] ast_sip_session_cancel_invite_call cur_session:%p has no dlg, endpoint:%s\n",
				session, cur_session, endpoint_id);
			continue;
		}

		sip_uri = pjsip_uri_get_uri(cur_session->inv_session->dlg->local.info->uri);
		if (!sip_uri) {
			ast_log(LOG_WARNING, "[session:%p] ast_sip_session_cancel_invite_call cur_session:%p has no remote uri, endpoint:%s\n",
				session, cur_session, endpoint_id);
			continue;
		}

		ast_copy_pj_str(to_user, &sip_uri->user, sizeof(to_user));

		from_sip_uri = pjsip_uri_get_uri(cur_session->inv_session->dlg->remote.info->uri);
		if (!from_sip_uri) {
			ast_log(LOG_WARNING, "[session:%p] ast_sip_session_cancel_invite_call cur_session:%p has no local uri, endpoint:%s\n",
				session, cur_session, endpoint_id);
			continue;
		}

		ast_copy_pj_str(from_user, &from_sip_uri->user, sizeof(from_user));

		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_cancel_invite_call conf_room:%s, to_user:%s, from_user:%s, cur_endpoint:%s, endpoint:%s\n",
			session, conf_room, to_user, from_user, cur_endpoint_id, endpoint_id);

		if (!strcasecmp(to_user, conf_room)) {
			if (!strncasecmp(endpoint_id, "trunk_", 6)) {
				if (!strcasecmp(from_user, invite_to_user)) {
					ast_log(LOG_DEBUG, "[session:%p] It's trunk, and there is aleady incoming call, cancel call for the same endpoint:%s\n",
						session, endpoint_id);
					res = 1;
					break;
				}
			} else {
				ast_log(LOG_DEBUG, "[session:%p] There is aleady incoming call, cancel call for the same endpoint:%s\n",
					session, endpoint_id);
				res = 1;
				break;
			}
		}
	}
	AST_LIST_UNLOCK(&sip_sessions);

	return res;
}

void ast_sip_session_cancel_conf_invitee(struct ast_sip_session *session)
{
	struct ast_sip_session *cur_session = NULL;

	if (!session || ast_strlen_zero(session->conf_room)) {
		return;
	}

	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_TRAVERSE(&sip_sessions, cur_session, next) {
		if (ast_strlen_zero(cur_session->conf_room)) {
			continue;
		}

		if (strcmp(session->conf_room, cur_session->conf_room)) {
			continue;
		}

		if (cur_session->inv_session &&
			cur_session->inv_session->role == PJSIP_ROLE_UAC &&
			cur_session->inv_session->state < PJSIP_INV_STATE_CONNECTING) {
			ast_log(LOG_DEBUG, "[session:%p] Receive call, cancel pending session:%p call for the conf room:%s\n",
				session, cur_session, session->conf_room);

			if (cur_session->channel) {
				ast_log(LOG_DEBUG, "hangup chan :%s call for the conf room:%s\n", ast_channel_name(cur_session->channel), session->conf_room);

				ast_channel_conf_hangupcause_set(cur_session->channel, AST_CAUSE_CONF_END);
				ast_channel_hangupcause_set(cur_session->channel, AST_CAUSE_CONF_END);
			}
			ast_sip_session_terminate(cur_session, 603);
		}
	}
	AST_LIST_UNLOCK(&sip_sessions);

	return;
}

static void set_stream_uptime(struct ast_sip_session *session, struct ast_stream *stream)
{
	struct timespec tms;
	char uptime[64] = { 0 };

	if (clock_gettime(CLOCK_MONOTONIC, &tms) != 0) {
		ast_log(LOG_WARNING, "[session:%p] Can't get time, just use a random number for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		tms.tv_sec = ast_random();
		tms.tv_nsec = ast_random();
	}

	snprintf(uptime, sizeof(uptime) - 1, "%ld %ld", tms.tv_sec, tms.tv_nsec);

	ast_log(LOG_DEBUG, "[session:%p] set_stream_uptime:%s for endpoint:%s\n",
	session, uptime, ast_sorcery_object_get_id(session->endpoint));

	ast_stream_set_uptime(stream, uptime);

	return;
}

static unsigned int if_has_valid_main_video_stream(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_local_media_direction(stream) == SENDRECV ||
		ast_stream_get_local_media_direction(stream) == RECVONLY) {
		return 1;
	}

	if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_INACTIVE ||
		ast_stream_get_state(stream) == AST_STREAM_STATE_SENDONLY) {
		return 0;
	}

	return 1;
}

static struct ast_stream *ast_sip_session_get_channel_active_main_video_stream(struct ast_sip_session *session)
{
	unsigned int i = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream_topology *topology = NULL;

	if (!session || !session->channel) {
		return NULL;
	}

	topology = ast_channel_get_stream_topology(session->channel);
	if (!topology) {
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(topology); i++) {
		stream = ast_stream_topology_get_stream(topology, i);
		if (!stream) {
			continue;
		}
		if (!ast_stream_judge_if_main_video(stream)) {
			continue;
		}
		return stream;
	}

	return NULL;
}


void ast_sip_session_set_main_video_stream_uptime(struct ast_sip_session *session)
{
	struct ast_stream *main_video = NULL;

	if (!session || !session->channel) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_main_video_stream_uptime for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	main_video = ast_sip_session_get_pending_main_video_stream(session);
	if (main_video) {
		if (if_has_valid_main_video_stream(main_video)) {
			set_stream_uptime(session, main_video);
		}
	} else {
		main_video = ast_sip_session_get_channel_active_main_video_stream(session);
		if (if_has_valid_main_video_stream(main_video)) {
			set_stream_uptime(session, main_video);
		}
	}

	return;
}

static void get_prefer_video_codec_name(pj_pool_t *pool,
	pjmedia_sdp_media *media, char *name)
{
	unsigned int i = 0;
	unsigned int has_h264 = 0;
	unsigned int has_vp8 = 0;
	unsigned int has_h265 = 0;
	unsigned int has_vp9 = 0;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	char codec_name[32] = {0};

	if (!pool || !media || !name) {
		return;
	}

	for (i = 0; i < media->desc.fmt_count; ++i) {
		if (!(attr = pjmedia_sdp_media_find_attr2(media, "rtpmap", &media->desc.fmt[i]))) {
			return;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
			return;
		}

		ast_copy_pj_str(codec_name, &rtpmap->enc_name, 32);

		if (!strcasecmp(codec_name, "H264")) {
			has_h264 = 1;
		} else if (!strcasecmp(codec_name, "VP8")) {
			has_vp8 = 1;
		} else if (!strcasecmp(codec_name, "H265")) {
			has_h265 = 1;
		} else if (!strcasecmp(codec_name, "VP9")) {
			has_vp9 = 1;
		}
	}

	if (has_h264) {
		strcpy(name, "H264");
	} else if (has_vp8) {
		strcpy(name, "VP8");
	} else if (has_h265) {
		strcpy(name, "H265");
	} else if (has_vp9) {
		strcpy(name, "VP9");
	}

	return;
}

void ast_sip_session_set_conf_video_codec(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	char *conf_codec = NULL;
	char codec_name[64] = {0};

	if (!pool || !media || !session) {
		return;
	}

	if ((resolution <= VIDEO_RESOLUTION_NONE) || (resolution >= VIDEO_RESOLUTION_MAX)) {
		return;
	}

	conf_codec = ast_conf_get_video_codec(session->channel, session->conf_room);

	/* If conf already has video codec, don't need set again */
	if (!ast_strlen_zero(conf_codec)) {
		ast_log(LOG_DEBUG, "[session:%p] set_conf_video_codec already has conf codec:%s, endpoint:%s\n",
			session, conf_codec, ast_sorcery_object_get_id(session->endpoint));
		ast_conf_set_video_codec(session->channel, session->conf_room, conf_codec);
		ast_free(conf_codec);
		conf_codec = NULL;
		return;
	}

	get_prefer_video_codec_name(pool, media, codec_name);
	if (ast_strlen_zero(codec_name)) {
		ast_log(LOG_WARNING, "[session:%p] Can't get video codec name, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s get prefer video codec name:%s, use this codec to set\n",
		session, ast_sorcery_object_get_id(session->endpoint), codec_name);

	ast_conf_set_video_codec(session->channel, session->conf_room, codec_name);

	return;
}

static void relpace_rtpmap_codec_val(pj_pool_t *pool, pjmedia_sdp_attr *attr, int rtx_pt)
{
	int code_pt = 0;
	char codec_params[64] = { 0 };
	char rtpmap_attr_value[64] = { 0 };
	char new_rtpmap_attr_value[64] = { 0 };
	pj_str_t new_rtpmap_attr_value_str;

	ast_copy_pj_str(rtpmap_attr_value, &attr->value, sizeof(rtpmap_attr_value));

	sscanf(rtpmap_attr_value, "%d %s", &code_pt, codec_params);

	snprintf(new_rtpmap_attr_value, sizeof(new_rtpmap_attr_value), "%d %s", rtx_pt, codec_params);

	new_rtpmap_attr_value_str = pj_str(new_rtpmap_attr_value);

	pj_strdup(pool, &attr->value, &new_rtpmap_attr_value_str);

	return;
}

static void relpace_fmtp_codec_param(pj_pool_t *pool, pjmedia_sdp_attr *attr, int rtx_pt, char *video_pt_str)
{
	char new_fmtp_attr_value[64] = { 0 };
	pj_str_t new_fmtp_attr_value_str;

	snprintf(new_fmtp_attr_value, sizeof(new_fmtp_attr_value), "%d apt=%s", rtx_pt, video_pt_str);

	new_fmtp_attr_value_str = pj_str(new_fmtp_attr_value);

	pj_strdup(pool, &attr->value, &new_fmtp_attr_value_str);

	return;
}

static int get_according_video_rtx_pt(struct ast_sip_session *session, char *rtx_video_codec_name)
{
	int rtx_pt = 0;
	struct ast_sip_session_media *session_media = NULL;

	session_media = get_pending_main_video_media(session);
	if (!session_media) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get pending media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	if (!strcasecmp(rtx_video_codec_name, "H264")) {
		rtx_pt = session_media->video_rtx_pt.h264_rtx_pt;
	} else if (!strcasecmp(rtx_video_codec_name, "H265")) {
		rtx_pt = session_media->video_rtx_pt.h265_rtx_pt;
	} else if (!strcasecmp(rtx_video_codec_name, "VP8")) {
		rtx_pt = session_media->video_rtx_pt.vp8_rtx_pt;
	} else if (!strcasecmp(rtx_video_codec_name, "VP9")) {
		rtx_pt = session_media->video_rtx_pt.vp9_rtx_pt;
	} else {
		ast_log(LOG_DEBUG, "[session:%p] Can't support video codec:%s for RTX, endpoint:%s\n",
			session, rtx_video_codec_name, ast_sorcery_object_get_id(session->endpoint));
	}

	return rtx_pt;
}

void ast_sip_session_change_rtx_params(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, char *rtx_video_codec_name, pj_str_t video_pt_pj_str)
{
	unsigned int i = 0;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *fmtp_attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	char codec_name[32] = {0};
	int rtx_pt = 0;
	char rtx_pt_str[32] = { 0 };
	char video_pt_str[8] = { 0 };

	for (i = 0; i < media->desc.fmt_count; ++i) {
		if (!(attr = pjmedia_sdp_media_find_attr2(media, "rtpmap", &media->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		if (!(fmtp_attr = pjmedia_sdp_media_find_attr2(media, "fmtp", &media->desc.fmt[i]))) {
			continue;
		}

		ast_copy_pj_str(codec_name, &rtpmap->enc_name, sizeof(codec_name));

		if (strcasecmp(codec_name, "RTX")) {
			continue;
		}

		rtx_pt = get_according_video_rtx_pt(session, rtx_video_codec_name);
		if (!rtx_pt) {
			break;
		}

		snprintf(rtx_pt_str, sizeof(rtx_pt_str), "%d", rtx_pt);
		pj_strdup2(pool, &media->desc.fmt[i], rtx_pt_str);

		relpace_rtpmap_codec_val(pool, attr, rtx_pt);

		ast_copy_pj_str(video_pt_str, &video_pt_pj_str, sizeof(video_pt_str));
		relpace_fmtp_codec_param(pool, fmtp_attr, rtx_pt, video_pt_str);

		break;
	}

	return;
}

int ast_sip_session_modify_video_codec_order_for_sfu(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, enum video_resolution resolution)
{
	unsigned int i = 0;
	unsigned int find_matched_codec = 0;
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_rtpmap *rtpmap = NULL;
	char codec_name[32] = {0};
	const char *conf_codec = NULL;

	if (!media || !session) {
		return -1;
	}

	if (((resolution <= VIDEO_RESOLUTION_NONE) || (resolution >= VIDEO_RESOLUTION_MAX)) &&
		ast_sip_session_support_multiple_video(session)) {
		return 0;
	}

	conf_codec = ast_sip_session_sfu_codec2str(session->sfu_video_codec);

	/* If conf already has video codec, don't need set again */
	if (ast_strlen_zero(conf_codec)) {
		ast_log(LOG_WARNING, "[session:%p] Can't get conf video codec, don't modify video codec order, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	for (i = 0; i < media->desc.fmt_count; ++i) {
		if (!(attr = pjmedia_sdp_media_find_attr2(media, "rtpmap", &media->desc.fmt[i]))) {
			continue;
		}

		/* Interpret the attribute as an rtpmap */
		if ((pjmedia_sdp_attr_to_rtpmap(pool, attr, &rtpmap)) != PJ_SUCCESS) {
			continue;
		}

		ast_copy_pj_str(codec_name, &rtpmap->enc_name, 32);

		if (!strcasecmp(codec_name, conf_codec)) {
			ast_log(LOG_DEBUG, "[session:%p] Find conf codec in media stream(position:%d), use this codec as prefer, endpoint:%s\n",
				session, i, ast_sorcery_object_get_id(session->endpoint));
			if (i) {
				str_swap(&media->desc.fmt[i], &media->desc.fmt[0]);
				ast_sip_session_change_rtx_params(pool, session, media, codec_name, media->desc.fmt[0]);
			}
			find_matched_codec = 1;
			break;
		}
	}

	if (!find_matched_codec) {
		ast_log(LOG_DEBUG, "[session:%p] Can't find matched conf video codec, just decline this media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		media->desc.port = 0;
		ast_conf_put_main_video_resource(session->conf_room, session->channel);
		return -1;
	}

	return 0;
}

const char *ast_sip_session_media_encryption_type2str(enum ast_sip_session_media_encryption encryption)
{
	switch (encryption) {
	case AST_SIP_MEDIA_TRANSPORT_INVALID:
		return "invalid";
	case AST_SIP_MEDIA_ENCRYPT_NONE:
		return "none";
	case AST_SIP_MEDIA_ENCRYPT_SDES:
		return "sdes";
	case AST_SIP_MEDIA_ENCRYPT_DTLS:
		return "dtls";
	case AST_SIP_MEDIA_AUTO_NO_ENCRYPT:
		return "auto_no_encrypt";
	case AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP:
		return "auto_encrypt_srtp";
	case AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS:
		return "auto_encrypt_dtls";
	case AST_SIP_MEDIA_ENCRYPT_ZRTP:
		return "zrtp";
	case AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP:
		return "zrtp_srtp";
	default:
		return "<unknown>";
	}
}

int ast_sip_session_get_format_useinbandfec(struct ast_sip_session *session,
	struct ast_stream *stream, struct ast_format_cap *caps)
{
	int index = 0;
	struct ast_format *format = NULL;
	const unsigned int *useinbandfec = NULL;

	if (!session || !caps || !stream) {
		return 0;
	}

	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO) {
		return 0;
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		if (strcasecmp(ast_format_get_name(format), "opus")) {
			ao2_ref(format, -1);
			continue;
		}

		ast_log(LOG_DEBUG, "[session:%p] get_format_useinbandfec get opus codec in fmtp for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));

		useinbandfec = ast_format_attribute_get(format, "useinbandfec");
		if (useinbandfec && *useinbandfec) {
			ast_log(LOG_DEBUG, "[session:%p] get_format_useinbandfec get useinbandfec in fmtp for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(format, -1);
			return 1;
		}

		ao2_ref(format, -1);
		break;
	}

	return 0;
}

void ast_sip_session_set_format_useinbandfec(struct ast_sip_session *session,
	struct ast_format_cap *caps, struct ast_stream *stream, int useinbandfec)
{
	int index = 0;
	struct ast_format *format = NULL;
	RAII_VAR(struct ast_format *, opus_format, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, opus_caps, NULL, ao2_cleanup);

	char useinbandfec_str[4] = { 0 };

	if (!session || !caps || !stream) {
		return;
	}

	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		if (strcasecmp(ast_format_get_name(format), "opus")) {
			ao2_ref(format, -1);
			continue;
		}

		ast_log(LOG_DEBUG, "[session:%p] set_format_useinbandfec get opus codec in fmtp, set useinbandfec:%d for endpoint:%s\n",
				session, useinbandfec, ast_sorcery_object_get_id(session->endpoint));

		snprintf(useinbandfec_str, sizeof(useinbandfec_str) - 1, "%d", useinbandfec);

		opus_format = ast_format_attribute_set(format, "useinbandfec", useinbandfec_str);
		if (!opus_format) {
			ast_log(LOG_WARNING, "[session:%p] Can't set format useinbandfec attribute for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(format, -1);
			break;
		}

		opus_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!opus_caps) {
			ao2_ref(format, -1);
			break;
		}

		ast_format_cap_append(opus_caps, opus_format, 0);
		ast_format_cap_replace_from_cap(caps, opus_caps, AST_MEDIA_TYPE_AUDIO);

		ao2_ref(format, -1);

		break;
	}

	return;
}

void ast_sip_session_set_format_dtx(struct ast_sip_session *session,
	struct ast_format_cap *caps, struct ast_stream *stream)
{
	int index = 0;
	int dtx = 0;
	char dtx_str[4] = { 0 };
	struct ast_format *format = NULL;
	RAII_VAR(struct ast_format *, opus_format, NULL, ao2_cleanup);
	RAII_VAR(struct ast_format_cap *, opus_caps, NULL, ao2_cleanup);

	if (!session || !caps || !stream) {
		return;
	}

	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_AUDIO) {
		return;
	}

	if (!session->endpoint->media.dtx) {
		dtx = 0;
	} else {
		if (strcasecmp(session->endpoint->media.jitter_buffer.type, "neteq")) {
			dtx = 0;
		} else {
			dtx = 1;
		}
	}

	for (index = 0; index < ast_format_cap_count(caps); ++index) {
		format = ast_format_cap_get_format(caps, index);
		if (!format) {
			continue;
		}

		if (strcasecmp(ast_format_get_name(format), "opus")) {
			ao2_ref(format, -1);
			continue;
		}

		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_format_dtx get opus codec in fmtp, set dtx:%d for endpoint:%s\n",
				session, dtx, ast_sorcery_object_get_id(session->endpoint));

		snprintf(dtx_str, sizeof(dtx_str) - 1, "%d", dtx);

		opus_format = ast_format_attribute_set(format, "usedtx", dtx_str);
		if (!opus_format) {
			ast_log(LOG_WARNING, "[session:%p] Can't set format useinbandfec attribute for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(format, -1);
			break;
		}

		opus_caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
		if (!opus_caps) {
			ao2_ref(format, -1);
			break;
		}

		ast_format_cap_append(opus_caps, opus_format, 0);
		ast_format_cap_replace_from_cap(caps, opus_caps, AST_MEDIA_TYPE_AUDIO);

		ao2_ref(format, -1);

		break;
	}

	return;
}

static void change_caps_for_video_fec(struct ast_sip_session *session,
	struct ast_format_cap *joint_cap, struct ast_stream *clone_stream)
{
	if (!session || !session->endpoint || !joint_cap || !clone_stream) {
		return;
	}

	if (!ast_format_cap_has_video_fec(joint_cap)) {
		if (ast_format_cap_has_video_fec(session->endpoint->media.codecs) && 
			(!session->is_webrtc_call || (session->is_webrtc_call && session->support_gs_fec))) {
			ast_log(LOG_DEBUG, "Peer has no video FEC, but endpoint has video FEC, need add video FEC to caps\n");
			ast_format_cap_add_video_fec(joint_cap);
			/* Now we use alone fec version for video as default */
			ast_stream_set_alone_fec_version(clone_stream, 1);
		}
	} else {
		if (session->is_webrtc_call && !session->support_gs_fec) {
			ast_format_cap_remove_video_fec(joint_cap);
		}
	}

	return;
}

static void change_caps_for_rtx(struct ast_sip_session *session, struct ast_format_cap *joint_cap)
{
	if (!session || !session->endpoint || !joint_cap) {
		return;
	}

	if (!ast_format_cap_has_rtx(joint_cap) && ast_format_cap_has_rtx(session->endpoint->media.codecs)) {
		ast_log(LOG_DEBUG, "Peer has no RTX, but endpoint has RTX, need add RTX to caps\n");
		ast_format_cap_add_rtx(joint_cap);
	}

	return;
}

static void change_stream_for_nack(struct ast_sip_session *session, struct ast_stream *clone_stream)
{
	if (!session || !session->endpoint || !clone_stream) {
		return;
	}

	if (session->endpoint->enable_nack) {
		ast_log(LOG_DEBUG, "Endpoint has enable nack, set nack enable anyway\n");
		ast_stream_set_has_nack(clone_stream, 1);
		ast_stream_set_has_nack_pli(clone_stream, 1);
		ast_stream_set_has_ccm_fir(clone_stream, 1);
	}

	return;
}

void ast_sip_session_change_video_independent_codec(struct ast_sip_session *session,
	struct ast_format_cap *joint_cap, struct ast_stream *clone_stream)
{
	change_caps_for_video_fec(session, joint_cap, clone_stream);
	change_caps_for_rtx(session, joint_cap);
	change_stream_for_nack(session, clone_stream);

	return;
}

static unsigned int judge_if_mobile_use_bundle(const struct ast_sip_session *session)
{
	if (!session || !session->endpoint) {
		return FALSE;
	}

	if (!session->endpoint->media.mobile_bundle_enabled) {
		return FALSE;
	}

	if (!session->inv_session) {
		return TRUE;
	}

	if (session->inv_session->role == PJSIP_ROLE_UAC && session->inv_session->state < PJSIP_INV_STATE_CONFIRMED) {
		return TRUE;
	}

	if (session->remote_support_bundle) {
		return TRUE;
	}

	return FALSE;
}

unsigned int ast_sip_session_media_use_bundle(const struct ast_sip_session *session)
{
	if (!session) {
		return FALSE;
	}

	if (session->endpoint->media.bundle) {
		return TRUE;
	}

	if ((session->is_webrtc_call || session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) &&
		session->endpoint->media.gswave_bundle_enabled){
		return TRUE;
	}

	if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE){
		return judge_if_mobile_use_bundle(session);
	}

	return FALSE;
}

static unsigned char judge_if_need_send_session(struct ast_sip_session *session)
{
	unsigned int i = 0;
	const pjmedia_sdp_media *m = NULL;
	pjmedia_sdp_session *active_local_sdp = NULL;

	if (!session) {
		return 0;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		return 0;
	}

	if (session->inv_session->neg->state != PJMEDIA_SDP_NEG_STATE_DONE) {
		ast_log(LOG_DEBUG, "[session:%p] It has pending neg, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->delayed_send_session = 1;
		return 0;
	}

	active_local_sdp = session->inv_session->neg->active_local_sdp;

	if (!active_local_sdp) {
		return 0;
	}

	for (i = 0; i < active_local_sdp->media_count; i++) {
		m = active_local_sdp->media[i];
		/* We only send session to peer when has disable media */
		if (m->desc.port == 0 ||
			(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC &&
			ast_sip_session_judge_if_main_video(m) &&
			(pjmedia_sdp_media_find_attr2(m, "inactive", NULL) ||
			pjmedia_sdp_media_find_attr2(m, "sendonly", NULL) ||
			pjmedia_sdp_media_find_attr2(m, "recvonly", NULL))))  {
			ast_log(LOG_DEBUG, "[session:%p] Find special media for endpoint:%s, send session to peer\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return 1;
		}
	}

	/*Start added by zyxia for Bug 291590,2024.3.12*/
	if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE){

		for (i = 0; i < active_local_sdp->media_count; i++) {
			m = active_local_sdp->media[i];
			/* We only send session to peer when has disable media */
				if(ast_sip_session_judge_if_main_video(m) &&
				(pjmedia_sdp_media_find_attr2(m, "inactive", NULL) ||
				pjmedia_sdp_media_find_attr2(m, "sendonly", NULL) ||
				pjmedia_sdp_media_find_attr2(m, "recvonly", NULL)))  {
					ast_log(LOG_DEBUG, "[session:%p] Find special media for endpoint:%s, send session to peer\n",
						session, ast_sorcery_object_get_id(session->endpoint));
					return 1;
				}
				else if(ast_sip_session_judge_if_main_video(m) && pjmedia_sdp_media_find_attr2(m, "sendrecv", NULL)){
					ast_log(LOG_DEBUG, "[session:%p] Find special media for endpoint:%s, send session to peer\n",
						session, ast_sorcery_object_get_id(session->endpoint));
					return 0;
				}
		}
		/*End added by zyxia for Bug 291590,2024.3.12*/

		ast_log(LOG_DEBUG, "[session:%p] Find special media for endpoint:%s, send session to peer\n",
			session, ast_sorcery_object_get_id(session->endpoint));

		return 1;
	}

	return 0;
}

static unsigned int judge_if_need_destory_media(struct ast_sip_session *session)
{
	unsigned int res = 1;

	if (!session || !session->inv_session) {
		return 0;
	}

	switch (session->inv_session->role) {
		case PJSIP_ROLE_UAC:
			if (session->endpoint->media.rtp.accept_multiple_sdp_answers &&
				session->inv_session->state <= PJSIP_INV_STATE_EARLY) {
				res = 0;
			}
			break;
		case PJSIP_ROLE_UAS:
			if (session->endpoint->media.rtp.accept_multiple_sdp_answers &&
				session->inv_session->state <= PJSIP_INV_STATE_EARLY &&
				!session->send_final_rsp) {
				res = 0;
			}
			break;
		default:
			break;
	}

	return res;
}

void ast_sip_session_send_session_to_peer(struct ast_sip_session *session)
{
	struct ast_sip_reinvite_data reinvite_data;

	if (!session) {
		return;
	}

	if (!judge_if_need_send_session(session)) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Send session to peer to update for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	reinvite_data.session = session;
	ao2_ref(session, +1);
	session->ref_added = 1;
	if (ast_queue_control_data(session->channel, AST_CONTROL_PEER_SESSION, &reinvite_data, sizeof(reinvite_data))) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_PEER_SESSION frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
		session->ref_added = 0;
		return;
	}

	return;
}

static void notify_remote_disable_media_for_transfer(struct ast_sip_session *session)
{
	const char *ref_by_val = NULL;
	const char *rec_enable = NULL;

	if (!session || !session->channel || !session->inv_session) {
		return;
	}


	if (session->inv_session->state >= PJSIP_INV_STATE_CONFIRMED) {
		return;
	}

	ref_by_val = pbx_builtin_getvar_helper(session->channel, "SIPREFERREDBYHDR");
	if (ast_strlen_zero(ref_by_val)) {
		ast_log(LOG_DEBUG, "[session:%p] It's not transfer, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	rec_enable = pbx_builtin_getvar_helper(session->channel, "REC_ENABLE");
	if (ast_strlen_zero(rec_enable) || strcasecmp(rec_enable, "yes")) {
		ast_log(LOG_DEBUG, "[session:%p] It's not enable record, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Send AST_CONTROL_NOTIFY_DISABLE_MAIN_VIDEO to peer, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_send_session_to_peer(session);

	return;
}

static void send_get_peer_session_frame(struct ast_sip_session *session)
{
	if (!session || !session->channel) {
		ast_log(LOG_ERROR, "send_get_peer_session_frame param is NULL\n");
		return;
	}

	if (!session->get_peer_session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Has delayed AST_CONTROL_GET_PEER_SESSION for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_queue_control(session->channel, AST_CONTROL_GET_PEER_SESSION);
	session->get_peer_session = 0;

	return;
}

static void send_reinvite_control_frame_for_no_sdp(struct ast_sip_session *session)
{
	struct ast_sip_reinvite_data reinvite_data;
	if (!session ){
		ast_log(LOG_ERROR, "[session:%p] send_reinvite_control_frame_for_no_sdp\n", session);
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reinvite_control_frame_for_no_sdp for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->channel) {
		return;
	}

	if (!session->recv_reinvite_no_sdp) {
		return;
	}

	if (!session->disabled_video) {
		return;
	}

	session->disabled_video = 0;

	if (judge_if_has_valid_main_stream(session)) {
		session->recv_reinvite_no_sdp = 0;
	}

	reinvite_data.session = session;
	ao2_ref(session, +1);
	session->ref_added = 1;
	if (ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_REQ, &reinvite_data, sizeof(reinvite_data))) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_REINVITE_REQ frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
		session->ref_added = 0;
		return;
	}

	return;
}

static void send_peer_reinvite_control_for_callflip(struct ast_sip_session *session)
{
	struct ast_stream *stream = NULL;
	enum ast_stream_media_direction direction = 0;

	if (!session || !session->channel){
		ast_log(LOG_ERROR, "send_peer_reinvite_for_callflip param is NULL\n");
		return;
	}

	if (!session->send_peer_reinvite_for_callflip) {
		ast_log(LOG_DEBUG, "[session:%p] Not set the flag, don't need send re-INVITE for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->send_peer_reinvite_for_callflip = 0;

	ast_log(LOG_DEBUG, "[session:%p] send_peer_reinvite_for_callflip for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	stream = ast_sip_session_get_active_slides_video_stream(session);
	if (!stream) {
		direction = INACTIVE;
	} else {
		direction = ast_stream_get_local_media_direction(stream);
	}

	if (ast_queue_control_data(session->channel, AST_CONTROL_CALLFLIP_REINVITE, &direction, sizeof(enum ast_stream_media_direction))) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_REINVITE_REQ frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	return;
}


static struct ast_sip_session_media *add_session_media_for_stream(struct ast_sip_session *session,
	const pjmedia_sdp_session *sdp, int position)
{
	char media[20] = { 0 };
	struct ast_stream *existing_stream = NULL;
	struct ast_stream *stream = NULL;
	struct ast_sip_session_media *session_media = NULL;
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;
	pjmedia_sdp_media *remote_stream = NULL;

	if (!session || !sdp || !session->pending_media_state || !session->pending_media_state->topology) {
		return NULL;
	}
	
	/* We need a null-terminated version of the media string */
	ast_copy_pj_str(media, &sdp->media[position]->desc.media, sizeof(media));
	type = ast_media_type_from_str(media);

	remote_stream = sdp->media[position];
	sub_type = ast_sip_session_get_media_sub_type(session, remote_stream);

	/* See if we have an already existing stream, which can occur from SDP deferral checking */
	if (position < ast_stream_topology_get_count(session->pending_media_state->topology)) {
		stream = ast_stream_topology_get_stream(session->pending_media_state->topology, position);
	}

	if (session->active_media_state->topology &&
		(position < ast_stream_topology_get_count(session->active_media_state->topology))) {
		existing_stream = ast_stream_topology_get_stream(session->active_media_state->topology, position);
		if (sub_type != ast_stream_get_sub_type(existing_stream)) {
			ast_log(LOG_DEBUG, "Sub type is not the same, can' get existing_stream\n");
			existing_stream = NULL;
		}
	}

	if (!stream) {
		/* Maybe we receive SDP is not according as previous order, if can't match sub_type, just use default name */
		stream = ast_stream_alloc((existing_stream && (ast_stream_get_sub_type(existing_stream) == sub_type)) ? 
			ast_stream_get_name(existing_stream) : ast_codec_media_type2str(type), type);

		if (!stream) {
			return NULL;
		}
		if (ast_stream_topology_set_stream(session->pending_media_state->topology, position, stream)) {
			ast_stream_free(stream);
			return NULL;
		}
	}

	ast_stream_set_sub_type(stream, sub_type);

	if (existing_stream) {
		set_stream_params_according_exist_stream(stream, existing_stream);
	}

	session_media = ast_sip_session_media_state_add_as_sub_type_and_msid(session, session->pending_media_state, stream, position);

	return session_media;
}

static void set_session_media_mslabel(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	if (!session || !session_media || !session->endpoint) {
		return;
	}

	if (!ast_strlen_zero(session->endpoint->aors)) {
		snprintf(session_media->mslabel, sizeof(session_media->mslabel) - 1, "%s_%d-%s", session->endpoint->aors, session->msid_count, ast_codec_media_sub_type2str(session_media->sub_type));
	} else {
		snprintf(session_media->mslabel, sizeof(session_media->mslabel) - 1, "%s_%d-%s", ast_sorcery_object_get_id(session->endpoint), session->msid_count, ast_codec_media_sub_type2str(session_media->sub_type));
	}

	ast_log(LOG_DEBUG, "[session:%p] set_session_media_mslabel value:%s for endpoint:%s\n",
		session, session_media->mslabel, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_session_media_label(struct ast_sip_session_media *session_media, enum ast_media_sub_type sub_type, struct ast_sip_session *session)
{
	if (!session || !session_media) {
		return;
	}

	strncpy(session_media->label, ast_codec_media_sub_type2str(sub_type), sizeof(session_media->label) - 1);

	ast_log(LOG_DEBUG, "[session:%p] set_session_media_label value:%s for endpoint:%s\n",
		session, session_media->label, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static int check_msid_for_stream(struct ast_sip_session *session, struct ast_sip_session_media *session_media, struct ast_stream *stream)
{
	const char *mslabel = NULL;
	const char *label = NULL;
	char session_media_msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };
	char stream_media_msid[(AST_UUID_STR_LEN * 2) + 2] = { 0 };

	mslabel = ast_stream_get_metadata(stream, "MSID:MSLABEL");
	label = ast_stream_get_metadata(stream, "MSID:LABEL");

	if (!session || !session_media || !stream) {
		return FALSE;
	}

	if (!session->is_sfu) {
		return TRUE;
	}

	if (!strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		return TRUE;
	}

	/* If we can't find msid attr, we just pass msid check */
	if (!strlen(session_media->mslabel) || !strlen(session_media->label) || !mslabel || !label ) {
		ast_log(LOG_DEBUG, "[session:%p] check_msid_for_stream has no msid, just return true, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return TRUE;
	}

	snprintf(stream_media_msid, sizeof(stream_media_msid), "%s %s", mslabel, label);
	snprintf(session_media_msid, sizeof(session_media_msid), "%s %s", session_media->mslabel, session_media->label);

	ast_log(LOG_DEBUG, "[session:%p] check_msid_for_stream_and_session_media stream:%s, session_media:%s, endpoint:%s\n",
		session, stream_media_msid, session_media_msid, ast_sorcery_object_get_id(session->endpoint));

	if (strcmp(stream_media_msid, session_media_msid)) {
		return FALSE;
	}

	return TRUE;
}

static void set_session_media_mid(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int position)
{
	char mid[MAX_MID_LEN] = { 0 };

	if (!session || !session_media) {
		return;
	}

	if (ast_sip_session_media_use_bundle(session)) {
		/* If endpoint enable bundle will add mid later */
		return;
	}

	session_media->mid = ast_calloc(1, MAX_MID_LEN);
	if (!session_media->mid) {
		return;
	}

	if (snprintf(mid, sizeof(mid), "%d", position) < 0) {
		return;
	}

	strncpy(session_media->mid, mid, MAX_MID_LEN - 1);

	return;
}

static void set_session_media_trickle_ice(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	if (!session || !session_media) {
		return;
	}

	if (!session->endpoint->trickle_ice) {
		return;
	}

	if (!session->is_outgoing) {
		return;
	}

	/* when outgoing, set gswave support trickle ICE as default */
	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC && !session->support_trickle_ice) {
		return;
	}

	session_media->trickle_ice = 1;

	return;
}

static void set_session_media_encryption(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	enum ast_sip_session_media_encryption encryption = 0;

	if (!session) {
		return;
	}

	switch (session_media->sub_type) {
		case AST_MEDIA_SUB_AUDIO_TYPE:
			encryption = ast_channel_media_encryption_get(session->channel, AST_MEDIA_SUB_AUDIO_TYPE);
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_MAIN:
			encryption = ast_channel_media_encryption_get(session->channel, AST_MEDIA_SUB_VIDEO_TYPE_MAIN);
			break;
		case AST_MEDIA_SUB_VIDEO_TYPE_SLIDES:
			encryption = ast_channel_media_encryption_get(session->channel, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
			break;
		default:
			break;
	}

	if (encryption) {
		ast_log(LOG_DEBUG, "[session:%p] set_media_encryption from recovery encryption:%d, endpoint:%s\n",
			session, encryption, ast_sorcery_object_get_id(session->endpoint));
		session_media->encryption = encryption;
		ast_channel_media_encryption_set_by_type(session->channel, session_media->sub_type, 0);
	} else {
		if (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) {
			session_media->encryption = session->endpoint->media.rtp.encryption;
		} else if (session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE) {
			session_media->encryption = AST_SIP_MEDIA_ENCRYPT_SDES;
		} else {
			session_media->encryption = session->endpoint->media.rtp.configure_encryption;
		}
	}

	ast_log(LOG_DEBUG, "[session:%p] set_media_encryption encryption:%d, endpoint:%s\n",
		session, session_media->encryption, ast_sorcery_object_get_id(session->endpoint));

	return;
}

static void set_session_media_mpa(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	if (!session) {
		return;
	}

	if (!session->endpoint->media.mpa) {
		return;
	}

	/* If endpoint configure mpa, set session_media mpa as default, rewrite after negotiation */
	session_media->mpa = 1;

	return;
}

static unsigned int remote_support_rtcp_mux(struct ast_sip_session *session)
{
	if (!session) {
		return 0;
	}

	if (!session->endpoint->media.rtcp_mux) {
		return 0;
	}

	if (session->is_webrtc_call ||
		(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) ||
		(session->terminal_type == AST_TERMINAL_TYPE_ICE_TRUNK) ||
		(session->terminal_type == AST_TERMINAL_TYPE_GSWAVE_MOBILE)) {
		return 1;
	}

	return 0;
}

struct ast_sip_session_media *ast_sip_session_media_state_add_as_sub_type_and_msid(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state, struct ast_stream *stream, int position)
{
	unsigned int i = 0;
	struct ast_sip_session_media *session_media = NULL;
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!session || !media_state || !stream) {
		return NULL;
	}

	/* It is possible for this media state to already contain a session for the stream. If this
	 * is the case we simply return it.
	 */
	if (position < AST_VECTOR_SIZE(&media_state->sessions)) {
		return AST_VECTOR_GET(&media_state->sessions, position);
	}

	type = ast_stream_get_type(stream);
	sub_type = ast_stream_get_sub_type(stream);

	/* Determine if we can reuse the session media from the active media state if present */
	if (position < AST_VECTOR_SIZE(&session->active_media_state->sessions)) {
		session_media = AST_VECTOR_GET(&session->active_media_state->sessions, position);
		/* A stream can never exist without an accompanying media session */
		if (session_media->type == type && session_media->sub_type == sub_type &&
			check_msid_for_stream(session, session_media, stream)) {
			ao2_ref(session_media, +1);
		} else {
			session_media = NULL;
		}
	}

	/* Maybe we will receive SDP is not according preivous order, so need find it again */
	if (!session_media && (type != AST_MEDIA_TYPE_AUDIO)) {
		for (i = 0; i < AST_VECTOR_SIZE(&session->active_media_state->sessions); i++) {
			session_media = AST_VECTOR_GET(&session->active_media_state->sessions, i);
			if (session_media && session_media->type == type && session_media->sub_type == sub_type &&
				check_msid_for_stream(session, session_media, stream)) {
				ao2_ref(session_media, +1);
				break;
			} else {
				session_media = NULL;
			}
		}
	}

	if (!session_media) {
		/* No existing media session we can use so create a new one */
		session_media = ao2_alloc_options(sizeof(*session_media), session_media_dtor, AO2_ALLOC_OPT_LOCK_NOLOCK);
		if (!session_media) {
			return NULL;
		}

		session_media->remote_ice = session->endpoint->media.rtp.ice_support;
		/* Modify by bfdu for rtcp_mux only used for webrtc. 2019.04.03
		 * Todo: If sip phone support rtcp_mux need to change again.
		 */
		session_media->remote_rtcp_mux = remote_support_rtcp_mux(session);
		session_media->keepalive_sched_id = -1;
		session_media->timeout_sched_id = -1;
		session_media->bfcp_sched_id = -1;
		session_media->bfcp_keepalive_id = -1;
		session_media->type = type;
		session_media->stream_num = position;
		session_media->sub_type = sub_type;

		set_session_media_mslabel(session, session_media);
		set_session_media_label(session_media, sub_type, session);
		set_session_media_mid(session, session_media, position);
		set_session_media_trickle_ice(session, session_media);
		set_session_media_encryption(session, session_media);
		set_session_media_mpa(session, session_media);

		if (ast_sip_session_media_use_bundle(session)) {
			/* This is a new stream so create a new mid based on media type and position, which makes it unique.
			 * If this is the result of an offer the mid will just end up getting replaced.
			 */
			if (ast_asprintf(&session_media->mid, "%d", position) < 0) {
				ao2_ref(session_media, -1);
				return NULL;
			}

			if (ast_stream_get_state(stream) != AST_STREAM_STATE_REMOVED) {
				session_media->bundle_group = 0;

				/* Some WebRTC clients can't handle an offer to bundle media streams. Instead they expect them to
				 * already be bundled. Every client handles this scenario though so if WebRTC is enabled just go
				 * ahead and treat the streams as having already been bundled.
				 */
				session_media->bundled = session->endpoint->media.webrtc || ast_sip_session_media_use_bundle(session);
			} else {
				session_media->bundle_group = -1;
			}

			/* Todo: For SFU, gswave is bundle all media together, other media is bundled to the first media, we just record this */
			if (position == 0 && (session->participant_bundled_media == NULL)) {
				session->participant_bundled_media = session_media;
				ao2_ref(session_media, +1);
				ast_log(LOG_DEBUG, "set participant_bundled_media:%p\n", session_media);
			}
		} else {
			session_media->bundle_group = -1;
		}
	}

	if (AST_VECTOR_REPLACE(&media_state->sessions, position, session_media)) {
		ao2_ref(session_media, -1);

		return NULL;
	}

	/* If this stream will be active in some way and it is the first of this type then consider this the default media session to match */
	if (!media_state->default_session[type] && ast_stream_get_state(ast_stream_topology_get_stream(media_state->topology, position)) != AST_STREAM_STATE_REMOVED) {
		media_state->default_session[type] = session_media;
	}

	return session_media;
}

struct ast_sip_session_media *ast_sip_session_get_media_for_participant(struct ast_sip_session *session, struct ast_sip_session_media *session_media)
{
	if (!ast_sip_session_media_use_bundle(session) || ast_strlen_zero(session_media->mid)) {
		return session_media;
	}

	if (!session->participant_bundled_media) {
		return session_media;
	}

	ast_log(LOG_DEBUG, "ast_sip_session_get_media_for_participant %p\n", session->participant_bundled_media);

	return session->participant_bundled_media;
}

static int change_ssrc_for_hold(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	unsigned int i = 0;
	int is_changed = 0;
	struct ast_sip_session_media *session_media = NULL;
	struct ast_stream *stream = NULL;
	const pjmedia_sdp_media *om;

	if (!session || !offer || !session->pending_media_state) {
		ast_log(LOG_WARNING, "Change ssrc for hold, param is invalid\n");
		return 0;
	}

	if (session->is_sfu && session->is_webrtc_call) {
		ast_log(LOG_DEBUG, "[session:%p] It's SFU, and endpoint:%s is webrtc, don't need process for hold\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ao2_ref(session, +1);

	for (i = 0; i < offer->media_count; ++i) {
		om = offer->media[i];

		//only find audio
		if (strncmp(om->desc.media.ptr, "audio", 5)) {
			continue;
		}

		if (NULL == pjmedia_sdp_media_find_attr2(om, "sendonly", NULL) &&
			NULL == pjmedia_sdp_media_find_attr2(om, "inactive", NULL)) {
			break;
		}

		if (i < AST_VECTOR_SIZE(&session->pending_media_state->sessions)) {
			session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
			if (session_media && session_media->rtp) {
				ast_rtp_instance_change_source(session_media->rtp);
				ast_log(LOG_DEBUG, "[session:%p] Change media type:%s SSRC for hold, endpoint:%s\n",
					session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
				is_changed = 1;
			}

			if (session->pending_media_state->topology) {
				stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
				if (stream) {
					ast_stream_set_held(stream, 1);
				}
			}
		}
	}

	/*Start added by zyxia for 297853,2024.5.16*/
	//audio hold, other media also need hold
	if(is_changed) {
		for (i = 0; i < offer->media_count; ++i) {
			om = offer->media[i];

			if (!strncmp(om->desc.media.ptr, "audio", 5)) {
				continue;
			}

			if (i < AST_VECTOR_SIZE(&session->pending_media_state->sessions)) {
				session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
				if (session_media && session_media->rtp) {
					ast_rtp_instance_change_source(session_media->rtp);
					session_media->remotely_held = 1;
					ast_log(LOG_DEBUG, "[session:%p] Change media type:%s SSRC for hold, endpoint:%s\n",
						session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
					is_changed = 1;
				}

				if (session->pending_media_state->topology) {
					stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
					if (stream) {
						ast_stream_set_held(stream, 1);
					}
				}
			}
		}
		/*End added by zyxia for 297853,2024.5.16*/
	}
	ao2_ref(session, -1);

	return is_changed;
}

static int change_ssrc_for_unhold(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	unsigned int i = 0;
	int is_changed = 0;
	struct ast_sip_session_media *session_media = NULL;
	struct ast_stream *stream = NULL;

	if (!session || !offer || !session->pending_media_state) {
		ast_log(LOG_WARNING, "Change ssrc for unhold, param is invalid\n");
		return 0;
	}

	ao2_ref(session, +1);

	for (i = 0; i < offer->media_count; ++i) {
		const pjmedia_sdp_media *om;
		om = offer->media[i];
		if (NULL == pjmedia_sdp_media_find_attr2(om, "sendrecv", NULL)) {
			continue;
		}

		//remote it for 297853,2024.5.16*/
		/*
		if (strncmp(om->desc.media.ptr, "audio", 5)) {
			continue;
		}
		*/
		if (i < AST_VECTOR_SIZE(&session->pending_media_state->sessions)) {
			session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, i);
			if (session_media && session_media->rtp && session_media->remotely_held) {
				ast_rtp_instance_change_source(session_media->rtp);
				ast_log(LOG_DEBUG, "[session:%p] Change media type:%s SSRC for unhold, endpoint:%s\n",
					session, ast_codec_media_sub_type2str(session_media->sub_type), ast_sorcery_object_get_id(session->endpoint));
				is_changed = 1;
			}

			if (session->pending_media_state->topology) {
				stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
				if (stream && ast_stream_get_held(stream)) {
					ast_stream_set_change_ssrc(stream, 1);
				}
			}
		}
	}

	ao2_ref(session, -1);

	return is_changed;
}

static void change_ssrc_for_media_change(struct ast_sip_session *session)
{
	unsigned int index = 0;
	struct ast_sip_session_media *session_media = NULL;

	if (!session || !session->pending_media_state) {
		ast_log(LOG_WARNING, "Change ssrc for media change, param is invalid\n");
		return;
	}

	ao2_ref(session, +1);

	for (index = 0; index < AST_VECTOR_SIZE(&session->pending_media_state->sessions); ++index) {
		session_media = AST_VECTOR_GET(&session->pending_media_state->sessions, index);
		if (session_media && session_media->rtp) {
			ast_rtp_instance_change_source(session_media->rtp);
		}
	}

	ao2_ref(session, -1);

	return;
}

static unsigned int judge_if_sdp_has_video(pjmedia_sdp_session *sdp)
{
	unsigned int i = 0;
	pjmedia_sdp_media *media = NULL;

	if (!sdp) {
		return 0;
	}

	for (i = 0; i < sdp->media_count; ++i) {
		media = sdp->media[i];
		if (!media) {
			continue;
		}

		if (!strncmp(media->desc.media.ptr, "video", 5)) {
			return 1;
		}
	}

	return 0;
}

static void create_pending_media_for_reinvite_no_sdp(struct ast_sip_session *session, pjmedia_sdp_session *sdp)
{
	RAII_VAR(struct ast_format_cap *, caps, NULL, ao2_cleanup);

	if (!session || !sdp || !session->pending_media_state) {
		return;
	}

	if (!(caps = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT))) {
		ast_log(LOG_ERROR, "[session:%p] Failed to allocate capabilities for create pending media for no sdp, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] create_pending_media_for_reinvite_no_sdp for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!judge_if_sdp_has_video(sdp)) {
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_AUDIO);
	} else {
		ast_format_cap_append_from_cap(caps, session->endpoint->media.codecs, AST_MEDIA_TYPE_UNKNOWN);
	}
	ast_sip_session_change_gs_fec_version(caps);
	session->pending_media_state->topology = ast_stream_topology_create_from_format_cap(caps);

	return;
}

static unsigned int judge_if_sdp_has_disabled_video(pjmedia_sdp_session *sdp)
{
	unsigned int i = 0;
	pjmedia_sdp_media *media = NULL;

	if (!sdp) {
		return 0;
	}

	for (i = 0; i < sdp->media_count; ++i) {
		media = sdp->media[i];
		if (!media) {
			continue;
		}

		if (!ast_sip_session_judge_if_main_video(media) && media->desc.port == 0) {
			return 1;
		}
	}

	return 0;
}

static void modify_active_sdp_for_reinvite_no_sdp(struct ast_sip_session *session)
{
 	pjmedia_sdp_session *active_local_sdp = NULL;
	pjmedia_sdp_session *answer = NULL;
 
 	if (!session || !session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "modify_active_sdp_for_reinvite_no_sdp, param is null\n");
		return;
	}

	active_local_sdp = session->inv_session->neg->active_local_sdp;
	if (!active_local_sdp) {
		ast_log(LOG_WARNING, "[session:%p] modify_active_sdp_for_reinvite_no_sdp, has no active_local_sdp, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (judge_if_sdp_has_disabled_video(active_local_sdp)) {
		ast_log(LOG_DEBUG, "[session:%p] modify_active_sdp_for_reinvite_no_sdp, has disable video, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->disabled_video = 1;
	}

	create_pending_media_for_reinvite_no_sdp(session, active_local_sdp);

	session->is_outgoing = 1;

	if ((answer = create_local_sdp(session->inv_session, session, active_local_sdp))) {
		pjsip_inv_set_active_local_sdp(session->inv_session, answer);
	}

	return;
}

unsigned int ast_sip_session_get_allow_forward_hold_request(void)
{
	char *allow_forward_hold_request_str = NULL;
	unsigned int allow_forward_hold_request = 0;

	allow_forward_hold_request_str = ast_sip_get_allow_forward_hold_request();
	if (strcasecmp(allow_forward_hold_request_str, "yes") == 0) {
		allow_forward_hold_request = 1;
	} else {
		allow_forward_hold_request = 0;
	}

	ast_free(allow_forward_hold_request_str);
	allow_forward_hold_request_str = NULL;

	return allow_forward_hold_request;
}

unsigned int ast_sip_session_get_local_negotiate_prefer(void)
{
	char *local_negotiate_prefer_str = NULL;
	unsigned int local_negotiate_prefer = 0;

	local_negotiate_prefer_str = ast_sip_get_local_negotiate_prefer();
	if (strcasecmp(local_negotiate_prefer_str, "yes") == 0) {
		local_negotiate_prefer = 1;
	} else {
		local_negotiate_prefer = 0;
	}

	ast_free(local_negotiate_prefer_str);
	local_negotiate_prefer_str = NULL;

	return local_negotiate_prefer;
}

unsigned int ast_sip_session_get_enable_p2p_video_control(void)
{
	char *enable_p2p_video_control = NULL;
	unsigned int res = 0;

	enable_p2p_video_control = ast_sip_get_enable_p2p_video_control();
	if (!enable_p2p_video_control) {
		return 0;
	}

	if (!strcasecmp(enable_p2p_video_control, "yes")) {
		res = 1;
	}

	ast_free(enable_p2p_video_control);
	enable_p2p_video_control = NULL;

	return res;
}

unsigned int ast_sip_session_get_only_support_audio(void)
{
	char *only_support_audio_str = NULL;
	unsigned int only_support_audio = 0;

	only_support_audio_str = ast_sip_get_only_support_audio();
	if (strcasecmp(only_support_audio_str, "yes") == 0) {
		only_support_audio = 1;
	} else {
		only_support_audio = 0;
	}

	ast_free(only_support_audio_str);
	only_support_audio_str = NULL;

	return only_support_audio;
}

int ast_sip_session_get_p2p_video_resolution(void)
{
	char *video_resolution_str = NULL;
	enum video_resolution resolution = VIDEO_RESOLUTION_NONE;

	video_resolution_str = ast_sip_get_p2p_video_resolution();
	if (strcasecmp(video_resolution_str, "1080P") == 0) {
		resolution = VIDEO_RESOLUTION_1080P;
	} else if (strcasecmp(video_resolution_str, "720P") == 0) {
		resolution = VIDEO_RESOLUTION_720P;
	} else if (strcasecmp(video_resolution_str, "360P") == 0) {
		resolution = VIDEO_RESOLUTION_360P;
	}

	ast_free(video_resolution_str);
	video_resolution_str = NULL;

	return resolution;
}

unsigned char ast_sip_session_judge_if_disable_media(pjmedia_sdp_media *media)
{
	if (!media) {
		return FALSE;
	}

	if (!media->desc.port && !pjmedia_sdp_media_find_attr2(media, "bundle-only", NULL)) {
		return TRUE;
	}

	return FALSE;
}

void ast_sip_session_set_bfcp_info_msid(struct ast_sip_session *session,
	struct ast_bfcp_info *bfcp_info)
{
	struct ast_sip_session_media *session_media = NULL;
	char msid[128] = { 0 };

	if (!session || !bfcp_info) {
		return;
	}

	session_media = ast_sip_session_get_active_session_media(session, AST_MEDIA_SUB_VIDEO_TYPE_SLIDES);
	if (!session_media) {
		ast_log(LOG_ERROR, "[session:%p] Can't get active slides video media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	snprintf(msid, sizeof(msid), "%s %s", session_media->mslabel, session_media->label);

	strncpy(bfcp_info->msid, msid, sizeof(bfcp_info->msid) - 1);

	return;
}

static int send_notify_to_update_peer_register_trunk(struct ast_sip_session *session)
{
	struct pjsip_tx_data *tdata = NULL;

	if (session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_ERROR, "[session:%p] Dialog is not established really, don't send sip INFO to update trunk info, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] Send INFO request to update gs state, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "[session:%p] Can't create INFO request to update trunk info, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	ast_sip_add_header(tdata, "X-GS-Trunk", "yes");

	ast_sip_session_send_request(session, tdata);

	return 0;
}

void ast_sip_session_set_ucm_trunk(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_generic_string_hdr *hdr = NULL;
	const pj_str_t gs_trunk = { "X-GS-Trunk", 10 };

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "ast_sip_session_set_ucm_trunk, para is invalid\n");
		return;
	}

	if (session->ucm_trunk) {
		return;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get rdata msg, don't need set ucm_trunk, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(msg, &gs_trunk, NULL);
	if (!hdr) {
		ast_log(LOG_DEBUG, "[session:%p] Can't find hdr, don't need set ucm_trunk, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->ucm_trunk = 1;

	if (strncasecmp(session->endpoint->aors, "trunk_", 6)) {
		send_notify_to_update_peer_register_trunk(session);
	}

	return;
}

void ast_sip_session_set_trunk_info(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t conf_link = { "X-GS-Conf-Link", 14 };
	pj_str_t channel_id = { "Channel-ID", 10 };

	if (!session || !rdata) {
		return;
	}

	if (session->conf_link) {
		ast_log(LOG_DEBUG, "[session:%p] set_trunk_info, already set conf_link, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &conf_link, NULL);
	if (!hdr) {
		return;
	}

	if (0 == hdr->hvalue.slen) {
		ast_log(LOG_ERROR, "[session:%p] set_trunk_info, conf_link has no value, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->conf_link = ast_calloc(1, hdr->hvalue.slen + 1);
	if (!session->conf_link) {
		ast_log(LOG_ERROR, "[session:%p] set_trunk_info, can't malloc trunk_conf_link, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(session->conf_link, &hdr->hvalue, hdr->hvalue.slen + 1);
	ast_log(LOG_DEBUG, "[session:%p] set_trunk_info, trunk_conf_link:%s, endpoint:%s\n",
			session, session->conf_link, ast_sorcery_object_get_id(session->endpoint));

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &channel_id, NULL);
	if (!hdr) {
		ast_log(LOG_WARNING, "[session:%p] set_trunk_info, can't get Channel-ID header, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (0 == hdr->hvalue.slen) {
		ast_log(LOG_WARNING, "[session:%p] set_trunk_info, Channel-ID header has no value, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->trunk_channel_id = ast_calloc(1, hdr->hvalue.slen + 1);
	if (!session->trunk_channel_id) {
		ast_log(LOG_ERROR, "[session:%p] set_trunk_info, can't malloc trunk_channel_id, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	ast_copy_pj_str(session->trunk_channel_id, &hdr->hvalue, hdr->hvalue.slen + 1);
	ast_log(LOG_DEBUG, "[session:%p] set_trunk_info, trunk_channel_id:%s, endpoint:%s\n",
		session, session->trunk_channel_id, ast_sorcery_object_get_id(session->endpoint));

	return;
}

void ast_sip_session_set_avs_address(struct ast_sip_session *session)
{
	char *avs_address = NULL;

	if (!session) {
		return;
	}

	if (!ast_get_avs_master_value()) {
		return;
	}

	/*added by zyxia for Bug 284585,cluster ip is wan, but reinvite from lan,2024.1.17 */
	if(session->diff_multiavs_ip)
		strcpy(session->avs_address, "127.0.0.1");
	else {
		avs_address = ast_avs_get_avs_address();

		if(avs_address) {
			strncpy(session->avs_address, avs_address, sizeof(session->avs_address) - 1);
			ast_free(avs_address);
		}
	}
	return;
}

int ast_sip_session_set_conf_link_from_db(struct ast_sip_session* session)
{
	const char *family = "CONF_STATUS";
	char key[128] = { 0 };
	char sub_value[MAX_CONF_LINK_LEN] = {0};
	char *out_callee = NULL;
	char *decode_conf_link = NULL;
	char *encode_callee_number = NULL;
	char final_conf_link[MAX_CONF_LINK_LEN] = {0};

	if (!session) {
		return -1;
	}

	snprintf(key, sizeof(key), "%s/LINK_FULL_BASE64", session->conf_room);

	ast_db_get(family, key, sub_value, sizeof(sub_value) - 1);
	if (ast_strlen_zero(sub_value)) {
		ast_log(LOG_ERROR, "[session:%p] ast_sip_session_set_conf_link_from_db can't get conf_link from DB, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	/* Need add related_info when invite by conference */
	out_callee = (char *)pbx_builtin_getvar_helper(session->channel, "THE_REAL_OUT_CALLEE");
	if (!ast_strlen_zero(out_callee)) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_conf_link_from_db, THE_REAL_OUT_CALLEE[%s], endpoint:%s",
			session, out_callee, ast_sorcery_object_get_id(session->endpoint));

		decode_conf_link = ast_calloc(1, MAX_CONF_LINK_LEN);
		if (!decode_conf_link) {
			ast_log(LOG_ERROR, "[session:%p] Can't malloc memory for decode_conf_link, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}
		ast_base64decode((unsigned char *)decode_conf_link, sub_value, MAX_CONF_LINK_LEN);

		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_conf_link_from_db, decode_conf_link:%s, endpoint:%s",
			session, decode_conf_link, ast_sorcery_object_get_id(session->endpoint));

		encode_callee_number = ast_calloc(1, AST_MAX_EXTENSION*2);
		if (!encode_callee_number) {
			ast_log(LOG_ERROR, "[session:%p] Can't malloc memory for encode_callee_number, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}
		ast_base64encode(encode_callee_number, (const unsigned char *)out_callee, strlen(out_callee), AST_MAX_EXTENSION*2);

		snprintf(final_conf_link, sizeof(final_conf_link), "%s&related_info=%s", decode_conf_link, encode_callee_number);

		memset(sub_value, 0, sizeof(sub_value));
		ast_base64encode(sub_value, (const unsigned char *)final_conf_link, strlen(final_conf_link), MAX_CONF_LINK_LEN);

		ast_free(decode_conf_link);
		decode_conf_link = NULL;

		ast_free(encode_callee_number);
		encode_callee_number = NULL;
	}

	session->conf_link = ast_calloc(1, strlen(sub_value) + 1);
	if (!session->conf_link) {
		ast_log(LOG_ERROR, "[session:%p] Can't malloc memory for conf_link, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	strcpy(session->conf_link, sub_value);

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_set_conf_link_from_db, conf_link:%s, endpoint:%s\n",
		session, session->conf_link, ast_sorcery_object_get_id(session->endpoint));

	return 0;
}

void ast_sip_session_notfiy_trunk_info(struct ast_sip_session* session)
{
	struct ast_trunk_info trunk_info;

	if (!session || !session->channel) {
		ast_log(LOG_WARNING, "ast_sip_session_notfiy_trunk_info params is NULL\n");
		return;
	}

	if (!session->ucm_trunk) {
		ast_log(LOG_DEBUG, "ast_sip_session_notfiy_trunk_info params is not ucm_trunk\n");
		return;
	}

	if (session->conf_link) {
		ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_notfiy_trunk_info, conf_link:%s has already setted, endpoint:%s\n",
			session, session->conf_link, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (ast_sip_session_set_conf_link_from_db(session)) {
		return;
	}

	memset(&trunk_info, 0, sizeof(struct ast_trunk_info));

	strncpy(trunk_info.conf_link, session->conf_link, sizeof(trunk_info.conf_link) - 1);
	strncpy(trunk_info.channel_id, ast_channel_name(session->channel), sizeof(trunk_info.channel_id) - 1);

	ast_sip_session_send_notify_to_update_gs_state(session, AST_GS_STATE_TYPE_TRUNK_INFO, &trunk_info);

	return;
}


void ast_sip_session_notfiy_answered_status(struct ast_sip_session* session)
{
	struct ast_answered_status status;
	if (!session || !session->channel) {
		ast_log(LOG_WARNING, "ast_sip_session_notfiy_answered_status params is NULL\n");
		return;
	}

	if(!ast_channel_get_is_queue_call(session->channel) ){
		ast_log(LOG_WARNING, "[session:%p] ast_sip_session_notfiy_answered_status,  is not queue call,ignore it endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if( !session->support_gs_state) {
		ast_log(LOG_WARNING, "[session:%p] ast_sip_session_notfiy_answered_status,  dont support gs-state ,ignore it endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	memset(&status, 0, sizeof(struct ast_answered_status));
	snprintf(status.answered,sizeof(status.answered),"yes");

	ast_sip_session_send_notify_to_update_gs_state(session, AST_GS_STATE_TYPE_ANSWERED_STATUS, &status);

	return;
}

void ast_sip_session_set_media_transport(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, pjmedia_sdp_media *media)
{
	unsigned int use_avpf = 0;

	if (!session || !session_media || !media) {
		return;
	}

	if ((session_media->encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_DTLS) ||
		(session_media->encryption == AST_SIP_MEDIA_ENCRYPT_DTLS)) {
		if (session->is_webrtc_call) {
			use_avpf = session->endpoint->media.rtp.use_avpf;
		} else {
			use_avpf = session->endpoint->media.rtp.configure_use_avpf;
		}

		ast_log(LOG_NOTICE, "[session:%p] endpoint:%s, webrtc flag:%d, final use_avpf value is %d\n",
			session, ast_sorcery_object_get_id(session->endpoint), session->is_webrtc_call, use_avpf);
	}


	if (session->is_webrtc_call || (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC)) {
		media->desc.transport = pj_str("UDP/TLS/RTP/SAVPF");
	} else {
		media->desc.transport = pj_str(ast_sdp_get_rtp_profile(
			/* Optimistic encryption places crypto in the normal RTP/AVP profile */
			!session->endpoint->media.rtp.encryption_optimistic &&
				((session_media->encryption == AST_SIP_MEDIA_ENCRYPT_SDES)
				|| (session_media->encryption == AST_SIP_MEDIA_AUTO_ENCRYPT_SRTP) || (session_media->encryption == AST_SIP_MEDIA_ENCRYPT_ZRTP_SRTP)),
			session_media->rtp, use_avpf,
			session->endpoint->media.rtp.force_avp));
	}

	return;
}

void ast_sip_session_add_decline_video_reason(pj_pool_t *pool, struct ast_sip_session *session,
	pjmedia_sdp_media *media, struct ast_stream *stream)
{
	enum ast_video_decline_reason video_decline_reason = 0;
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;

	if (!pool || !session || !media || !stream) {
		return;
	}

	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_VIDEO) {
		return;
	}

	video_decline_reason = ast_stream_get_video_decline_reason(stream);

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_add_decline_video_reason:%d, for endpoint:%s\n",
		session, video_decline_reason, ast_sorcery_object_get_id(session->endpoint));

	switch (video_decline_reason) {
		case AST_VIDEO_DECLINE_REASON_NO_VIDEO_CODEC:
			attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "no video codec"));
			media->attr[media->attr_count++] = attr;
			break;
		case AST_VIDEO_DECLINE_REASON_NO_VIDEO_RESOURCE:
			attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "no video resource"));
			media->attr[media->attr_count++] = attr;
			break;
		case AST_VIDEO_DECLINE_REASON_CPU_LIMIT:
			attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "CPU limit"));
			media->attr[media->attr_count++] = attr;
			break;
		case AST_VIDEO_DECLINE_REASON_VERSION_LIMIT:
			attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "version limit"));
			media->attr[media->attr_count++] = attr;
			break;
		case AST_VIDEO_DECLINE_REASON_NO_VIDEO_ROOM:
			attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "no video room"));
			media->attr[media->attr_count++] = attr;
			break;
		case AST_VIDEO_DECLINE_REASON_PEER_REJECT:
			attr = pjmedia_sdp_attr_create(pool, "declined-reason", pj_cstr(&stmp, "peer reject"));
			media->attr[media->attr_count++] = attr;
			break;
		default:
			break;
	}

	ast_stream_set_video_decline_reason(stream, 0);

	return;
}


static void add_trickle_ice_attr_to_media(struct ast_sip_session *session,
	pj_pool_t *pool, struct ast_sip_session_media *session_media, pjmedia_sdp_media *media)
{
	struct ast_rtp_engine_ice *ice;
	struct ao2_container *candidates;
	const char *username, *password;
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	struct ao2_iterator it_candidates;
	struct ast_rtp_engine_ice_candidate *candidate;

	if (!session->endpoint->media.rtp.ice_support || !(ice = ast_rtp_instance_get_ice(session_media->rtp))) {
		return;
	}

	if ((username = ice->get_ufrag(session_media->rtp))) {
		attr = pjmedia_sdp_attr_create(pool, "ice-ufrag", pj_cstr(&stmp, username));
		media->attr[media->attr_count++] = attr;
	}

	if ((password = ice->get_password(session_media->rtp))) {
		attr = pjmedia_sdp_attr_create(pool, "ice-pwd", pj_cstr(&stmp, password));
		media->attr[media->attr_count++] = attr;
	}

	candidates = ice->get_local_candidates(session_media->rtp);
	if (!candidates) {
		return;
	}

	it_candidates = ao2_iterator_init(candidates, 0);
	for (; (candidate = ao2_iterator_next(&it_candidates)); ao2_ref(candidate, -1)) {
		struct ast_str *attr_candidate = ast_str_create(128);

		ast_str_set(&attr_candidate, -1, "%s %u %s %d %s ", candidate->foundation, candidate->id, candidate->transport,
					candidate->priority, ast_sockaddr_stringify_addr_remote(&candidate->address));
		ast_str_append(&attr_candidate, -1, "%s typ ", ast_sockaddr_stringify_port(&candidate->address));

		switch (candidate->type) {
			case AST_RTP_ICE_CANDIDATE_TYPE_HOST:
				ast_str_append(&attr_candidate, -1, "host");
				break;
			case AST_RTP_ICE_CANDIDATE_TYPE_SRFLX:
				ast_str_append(&attr_candidate, -1, "srflx");
				break;
			case AST_RTP_ICE_CANDIDATE_TYPE_RELAYED:
				ast_str_append(&attr_candidate, -1, "relay");
				break;
		}

		if (!ast_sockaddr_isnull(&candidate->relay_address)) {
			ast_str_append(&attr_candidate, -1, " raddr %s rport", ast_sockaddr_stringify_addr_remote(&candidate->relay_address));
			ast_str_append(&attr_candidate, -1, " %s", ast_sockaddr_stringify_port(&candidate->relay_address));
		}

		attr = pjmedia_sdp_attr_create(pool, "candidate", pj_cstr(&stmp, ast_str_buffer(attr_candidate)));
		media->attr[media->attr_count++] = attr;

		ast_free(attr_candidate);
	}

	if (session_media->end_of_candidates) {
		attr = pjmedia_sdp_attr_create(pool, "end-of-candidates", NULL);
		media->attr[media->attr_count++] = attr;
	}

	ao2_iterator_destroy(&it_candidates);
	ao2_ref(candidates, -1);

	return;
}

static pjmedia_sdp_media *create_media_descriptor_line(struct ast_sip_session *session,
	pj_pool_t *pool, struct ast_sip_session_media *session_media)
{
	pjmedia_sdp_media *media = NULL;
	struct ast_sockaddr addr;
	char tmp[8] = { 0 };

	if (!session || !session_media) {
		return NULL;
	}

	if (!session_media->rtp) {
		ast_log(LOG_DEBUG, "[session:%p] Has no rtp for session_media:%p, endpoint:%s\n",
			session, session_media, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	media = pj_pool_zalloc(pool, sizeof(struct pjmedia_sdp_media));
	if (!media) {
		return NULL;
	}

	ast_log(LOG_DEBUG, "[session:%p] create_media_descriptor_line for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	pj_strdup2(pool, &media->desc.media, ast_codec_media_type2str(session_media->type));

	ast_sip_session_set_media_transport(session, session_media, media);

	ast_rtp_instance_get_local_address(session_media->rtp, &addr);
	media->desc.port = (pj_uint16_t) ast_sockaddr_port(&addr);
	media->desc.port_count = 1;

	snprintf(tmp, sizeof(tmp), "%d", 0);
	pj_strdup2(pool, &media->desc.fmt[media->desc.fmt_count++], tmp);

	return media;
}

static void add_mid_to_media(struct ast_sip_session *session,
	pj_pool_t *pool, struct ast_sip_session_media *session_media, pjmedia_sdp_media *media)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr = NULL;

	/* Add the media identifier if present */
	if (!ast_strlen_zero(session_media->mid)) {
		attr = pjmedia_sdp_attr_create(pool, "mid", pj_cstr(&stmp, session_media->mid));
		pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
	}

	return;
}

static void create_sdp_for_trickle_ice(struct ast_sip_session *session, char *content, int len)
{
	unsigned int i = 0;
	pjmedia_sdp_session *info_sdp = NULL;
	struct ast_sip_session_media_state *media_state = NULL;
	struct ast_sip_session_media_state *media_state_clone = NULL;
	struct ast_sip_session_media *session_media = NULL;
	pjmedia_sdp_media *media = NULL;
	pj_pool_t *pool = NULL;

	if (!session || !session->inv_session || !session->channel) {
		return;
	}

	pool = session->inv_session->pool;
	if (!pool) {
		return;
	}

	if (!(info_sdp = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_session))) {
		return;
	}

	ast_channel_lock(session->channel);

	/* Stop if we're a zombie or need a soft hangup */
	if (ast_test_flag(ast_channel_flags(session->channel), AST_FLAG_ZOMBIE) || ast_check_hangup(session->channel)) {
		ast_channel_unlock(session->channel);
		ast_log(LOG_DEBUG, "[session:%p] Channel has been hung up, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->pending_media_state || !session->pending_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has no pending media, use active media\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		media_state = session->active_media_state;
	} else {
		media_state = session->pending_media_state;
	}

	if (!media_state || !media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] There's no topology for create sdp for trickle ice, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_unlock(session->channel);
		return;
	}

	media_state_clone = ast_sip_session_media_state_clone(media_state);
	if (!media_state_clone) {
		ast_log(LOG_WARNING, "[session:%p] Unable to clone active media state for channel, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_channel_unlock(session->channel);
		return;
	}

	ast_channel_unlock(session->channel);

	for (i = 0; i < ast_stream_topology_get_count(media_state_clone->topology); ++i) {
		session_media = AST_VECTOR_GET(&media_state_clone->sessions, i);
		if (!session_media) {
			ast_log(LOG_WARNING, "[session:%p] Can't get session_media for create trickle ice sdp, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		media = create_media_descriptor_line(session, pool, session_media);
		if (!media) {
			ast_log(LOG_ERROR, "[session:%p] Can't alloc media for trickle ice, position:%d, endpoint:%s\n",
				session, i, ast_sorcery_object_get_id(session->endpoint));
			continue;
		}

		add_mid_to_media(session, pool, session_media, media);

		add_trickle_ice_attr_to_media(session, pool, session_media, media);

		info_sdp->media[info_sdp->media_count++] = media;

		if (ast_sip_session_media_use_bundle(session)) {
			ast_log(LOG_DEBUG, "[session:%p] Use bundle just need add one media:%d, endpoint:%s\n",
				session, session_media->type, ast_sorcery_object_get_id(session->endpoint));
			break;
		}

		/* Ensure that we never exceed the maximum number of streams PJMEDIA will allow. */
		if (info_sdp->media_count == PJMEDIA_MAX_SDP_MEDIA) {
			break;
		}
	}

	ast_sip_session_print_sdp(info_sdp, content, len - 1);

	ast_sip_session_media_state_free(media_state_clone);
	media_state_clone = NULL;

	return;
}

static int add_candidates_info_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (400 <= rdata->msg_info.msg->line.status.code) {
		/* Failed, need send again */
		ast_log(LOG_WARNING, "[session:%p] recv add candidates info response failed, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, +1);
		if (ast_sip_push_task(session->serializer, ast_sip_session_send_info_for_add_candidates, session)) {
			ao2_cleanup(session);
		}
	}

	return 0;
}

int ast_sip_session_send_info_for_add_candidates(void *data)
{
	struct pjsip_tx_data *tdata = NULL;
	char sdp[4096] = {0};
	struct ast_sip_session *session = data;

	if (!session) {
		return -1;
	}

	if (session && session->inv_session && (session->inv_session->state < PJSIP_INV_STATE_EARLY)) {
		ast_log(LOG_DEBUG, "Dialog is not established really, don't send sip info for trickle ice\n");
		ao2_ref(session, -1);
		return 0;
	}

	if (ast_sip_create_request("INFO", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create INFO request for trickle ice\n");
		ao2_ref(session, -1);
		return -1;
	}

	create_sdp_for_trickle_ice(session, sdp, sizeof(sdp));

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "trickle-ice-sdpfrag",
		.body_text = sdp
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to trickle ICE INFO request\n");
		ao2_ref(session, -1);
		return -1;
	}

	ast_sip_add_header(tdata, "Info-Package", "trickle-ice");
	ast_sip_add_header(tdata, "Content-Disposition", "Info-Package");

	ast_sip_session_send_request_with_cb(session, tdata, add_candidates_info_response);

	ao2_ref(session, -1);

	return 0;
}

static void record_replace_channel(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	pjsip_generic_string_hdr *hdr = NULL;
	pj_str_t replace_channel_id = { "X-GS-Replace-Channel-ID", 23 };
	char channel_name[256] = { 0 };

	if (!session || !rdata) {
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &replace_channel_id, NULL);
	if (!hdr) {
		return;
	}

	ast_copy_pj_str(channel_name, &hdr->hvalue, sizeof(channel_name));
	if (ast_strlen_zero(channel_name)) {
		ast_log(LOG_ERROR, "[session:%p] record_replace_channel, can't get channel name, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	if (!session->replace_channel_id) {
		session->replace_channel_id = ast_calloc(1, hdr->hvalue.slen+ 1);
		if (session->replace_channel_id) {
			strcpy(session->replace_channel_id, channel_name);
			ast_log(LOG_DEBUG, "[session:%p] Record replace channel:%s, endpoint:%s\n",
				session, session->replace_channel_id, ast_sorcery_object_get_id(session->endpoint));
		} else {
			ast_log(LOG_ERROR, "[session:%p] Can't malloc for replace channel, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	}

	return;
}

void ast_sip_session_set_gs_client_info(pjsip_rx_data *rdata, struct ast_sip_session *session, unsigned int is_caller)
{
	pjsip_msg *msg = NULL;
	pjsip_generic_string_hdr *hdr = NULL;
	const pj_str_t name = { "X-GS-Client-Info", 16 };
	char client_info[MAX_USER_AGENT_LEN] = { 0 };

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "set_gs_client_info, para is invalid\n");
		return;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	if (!hdr) {
		return;
	}

	snprintf(client_info, sizeof(client_info) - 1, "%.*s", (int)hdr->hvalue.slen, hdr->hvalue.ptr);

	ast_log(LOG_DEBUG, "[session:%p] set_gs_client_info get info:%s, endpoint:%s\n",
		session, client_info, ast_sorcery_object_get_id(session->endpoint));

	if (is_caller) {
		strncpy(session->caller_gs_client_info, client_info, sizeof(session->caller_gs_client_info) - 1);
	} else {
		strncpy(session->callee_gs_client_info, client_info, sizeof(session->callee_gs_client_info) - 1);
	}

	return;
}

static unsigned int judge_if_is_participant_stream(struct ast_stream *stream)
{
	if (!stream) {
		return 0;
	}

	if (ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_AUDIO_TYPE ||
		ast_stream_get_sub_type(stream) == AST_MEDIA_SUB_VIDEO_TYPE_SLIDES ||
		ast_stream_judge_if_main_video(stream)) {
		return 0;
	}

	return 1;
}

static struct ast_sip_session_media *ast_sip_session_participant_media_state_add_as_sub_type_and_msid(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state, struct ast_stream *stream, int position)
{
	unsigned int i = 0;
	struct ast_sip_session_media *session_media = NULL;
	enum ast_media_type type = AST_MEDIA_TYPE_UNKNOWN;
	enum ast_media_sub_type sub_type = AST_MEDIA_SUB_TYPE_UNKNOWN;

	if (!session || !media_state || !stream) {
		return NULL;
	}

	/* It is possible for this media state to already contain a session for the stream. If this
	 * is the case we simply return it.
	 */
	if (position < AST_VECTOR_SIZE(&media_state->sessions)) {
		return AST_VECTOR_GET(&media_state->sessions, position);
	}

	type = ast_stream_get_type(stream);
	sub_type = ast_stream_get_sub_type(stream);

	/* Determine if we can reuse the session media from the active media state if present */
	if (position < AST_VECTOR_SIZE(&session->active_participant_media_state->sessions)) {
		session_media = AST_VECTOR_GET(&session->active_participant_media_state->sessions, position);
		/* A stream can never exist without an accompanying media session */
		if (session_media->type == type && session_media->sub_type == sub_type &&
			check_msid_for_stream(session, session_media, stream)) {
			ao2_ref(session_media, +1);
		} else {
			session_media = NULL;
		}
	}

	/* Maybe we will receive SDP is not according preivous order, so need find it again */
	if (!session_media) {
		for (i = 0; i < AST_VECTOR_SIZE(&session->active_participant_media_state->sessions); i++) {
			session_media = AST_VECTOR_GET(&session->active_participant_media_state->sessions, i);
			if (session_media && session_media->type == type && session_media->sub_type == sub_type &&
				check_msid_for_stream(session, session_media, stream)) {
				ao2_ref(session_media, +1);
				break;
			} else {
				session_media = NULL;
			}
		}
	}

	if (!session_media) {
		/* No existing media session we can use so create a new one */
		session_media = ao2_alloc_options(sizeof(*session_media), session_media_dtor, AO2_ALLOC_OPT_LOCK_NOLOCK);
		if (!session_media) {
			return NULL;
		}

		session_media->remote_ice = session->endpoint->media.rtp.ice_support;
		/* Modify by bfdu for rtcp_mux only used for webrtc. 2019.04.03
		 * Todo: If sip phone support rtcp_mux need to change again.
		 */
		session_media->remote_rtcp_mux = remote_support_rtcp_mux(session);
		session_media->keepalive_sched_id = -1;
		session_media->timeout_sched_id = -1;
		session_media->bfcp_sched_id = -1;
		session_media->bfcp_keepalive_id = -1;
		session_media->type = type;
		session_media->stream_num = position + 3;
		session_media->sub_type = sub_type;

		set_session_media_mslabel(session, session_media);
		set_session_media_label(session_media, sub_type, session);
		set_session_media_mid(session, session_media, position);
		set_session_media_encryption(session, session_media);

		if (ast_sip_session_media_use_bundle(session)) {
			/* For participant stream, we use three for begin, (audio/main video/slides video as before) */
			if (ast_asprintf(&session_media->mid, "%d", position + 3) < 0) {
				ao2_ref(session_media, -1);
				return NULL;
			}
			session_media->bundle_group = 0;

			/* Some WebRTC clients can't handle an offer to bundle media streams. Instead they expect them to
			 * already be bundled. Every client handles this scenario though so if WebRTC is enabled just go
			 * ahead and treat the streams as having already been bundled.
			 */
			session_media->bundled = session->endpoint->media.webrtc || ast_sip_session_media_use_bundle(session);
		} else {
			session_media->bundle_group = -1;
		}
	}

	if (AST_VECTOR_REPLACE(&media_state->sessions, position, session_media)) {
		ao2_ref(session_media, -1);
		return NULL;
	}

	/* If this stream will be active in some way and it is the first of this type then consider this the default media session to match */
	if (!media_state->default_session[type] && ast_stream_get_state(ast_stream_topology_get_stream(media_state->topology, position)) != AST_STREAM_STATE_REMOVED) {
		media_state->default_session[type] = session_media;
	}

	return session_media;
}

static int add_participant_bundle_groups(struct ast_sip_session *session, pj_pool_t *pool, pjmedia_sdp_session *answer)
{
	pj_str_t stmp;
	pjmedia_sdp_attr *attr;
	struct sip_session_media_bundle_group bundle_groups[PJMEDIA_MAX_SDP_MEDIA];
	int index, mid_id;
	struct sip_session_media_bundle_group *bundle_group;

	if (session->endpoint->media.webrtc) {
		attr = pjmedia_sdp_attr_create(pool, "msid-semantic", pj_cstr(&stmp, "WMS *"));
		pjmedia_sdp_attr_add(&answer->attr_count, answer->attr, attr);
	}

	if (!ast_sip_session_media_use_bundle(session)) {
		return 0;
	}

	memset(bundle_groups, 0, sizeof(bundle_groups));

	/* Build the bundle group layout so we can then add it to the SDP */
	for (index = 0; index < AST_VECTOR_SIZE(&session->active_media_state->sessions); ++index) {
		struct ast_sip_session_media *session_media = AST_VECTOR_GET(&session->active_media_state->sessions, index);

		/* If this stream is not part of a bundle group we can't add it */
		if (session_media->bundle_group == -1) {
			continue;
		}

		bundle_group = &bundle_groups[session_media->bundle_group];

		/* If this is the first mid then we need to allocate the attribute string and place BUNDLE in front */
		if (!bundle_group->mids[0]) {
			bundle_group->mids[0] = session_media->mid;
			bundle_group->attr_string = ast_str_create(64);
			if (!bundle_group->attr_string) {
				continue;
			}

			ast_str_set(&bundle_group->attr_string, 0, "BUNDLE %s", session_media->mid);
			continue;
		}

		for (mid_id = 1; mid_id < PJMEDIA_MAX_SDP_MEDIA; ++mid_id) {
			if (!bundle_group->mids[mid_id]) {
				bundle_group->mids[mid_id] = session_media->mid;
				ast_str_append(&bundle_group->attr_string, 0, " %s", session_media->mid);
				break;
			} else if (!strcmp(bundle_group->mids[mid_id], session_media->mid)) {
				break;
			}
		}
	}

	for (index = 0; index < AST_VECTOR_SIZE(&session->pending_participant_media_state->sessions); ++index) {
		struct ast_sip_session_media *session_media = AST_VECTOR_GET(&session->pending_participant_media_state->sessions, index);

		/* If this stream is not part of a bundle group we can't add it */
		if (session_media->bundle_group == -1) {
			continue;
		}

		bundle_group = &bundle_groups[session_media->bundle_group];

		/* If this is the first mid then we need to allocate the attribute string and place BUNDLE in front */
		if (!bundle_group->mids[0]) {
			bundle_group->mids[0] = session_media->mid;
			bundle_group->attr_string = ast_str_create(64);
			if (!bundle_group->attr_string) {
				continue;
			}

			ast_str_set(&bundle_group->attr_string, 0, "BUNDLE %s", session_media->mid);
			continue;
		}

		for (mid_id = 1; mid_id < PJMEDIA_MAX_SDP_MEDIA; ++mid_id) {
			if (!bundle_group->mids[mid_id]) {
				bundle_group->mids[mid_id] = session_media->mid;
				ast_str_append(&bundle_group->attr_string, 0, " %s", session_media->mid);
				break;
			} else if (!strcmp(bundle_group->mids[mid_id], session_media->mid)) {
				break;
			}
		}
	}

	/* Add all bundle groups that have mids to the SDP */
	for (index = 0; index < PJMEDIA_MAX_SDP_MEDIA; ++index) {
		bundle_group = &bundle_groups[index];

		if (!bundle_group->attr_string) {
			continue;
		}

		attr = pjmedia_sdp_attr_create(pool, "group", pj_cstr(&stmp, ast_str_buffer(bundle_group->attr_string)));
		pjmedia_sdp_attr_add(&answer->attr_count, answer->attr, attr);

		ast_free(bundle_group->attr_string);
	}

	return 0;
}

static int change_mid_for_participant_media_reuse(struct ast_sip_session *session,
	struct ast_sip_session_media *session_media, int position)
{
	char mid[MAX_MID_LEN] = { 0 };

	if (!session_media->mid) {
		ast_log(LOG_VERBOSE, "[session:%p] Has no mid value, just use position value for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		if (ast_asprintf(&session_media->mid, "%d", position + 3) < 0) {
			return -1;
		}
		return 0;
	}

	/* For participant stream, we use three for begin, (audio/main video/slides video as before) */
	snprintf(mid, sizeof(mid), "%d", position + 3);
	if (strcmp(session_media->mid, mid)) {
		ast_log(LOG_VERBOSE, "[session:%p] mid is not according to position, mid:%s need change to %s for endpoint:%s\n",
			session, session_media->mid, mid, ast_sorcery_object_get_id(session->endpoint));
		ast_free(session_media->mid);
		session_media->mid = NULL;
		if (ast_asprintf(&session_media->mid, "%d", position + 3) < 0) {
			return -1;
		}
	}

	return 0;
}

static struct pjmedia_sdp_session *create_participant_local_sdp(pjsip_inv_session *inv,
	struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	static const pj_str_t STR_IN = { "IN", 2 };
	static const pj_str_t STR_IP4 = { "IP4", 3 };
	static const pj_str_t STR_IP6 = { "IP6", 3 };
	pjmedia_sdp_session *local = NULL;
	int i = 0;
	int stream = 0;

	ast_log(LOG_VERBOSE, "[session:%p] create_participant_local_sdp for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (inv->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_ERROR, "[session:%p] Failed to create session SDP. Session has been already disconnected, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	if (!inv->pool_notify_prov || !(local = PJ_POOL_ZALLOC_T(inv->pool_notify_prov, pjmedia_sdp_session))) {
		ast_log(LOG_ERROR, "[session:%p] Failed to create session SDP as can't alloc local sdp, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	if (!session->pending_participant_media_state->topology ||
		!ast_stream_topology_get_count(session->pending_participant_media_state->topology)) {
		ast_log(LOG_ERROR, "[session:%p] Failed to create session SDP. has no pending_participant_media_state topology, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return NULL;
	}

	for (i = 0; i < ast_stream_topology_get_count(session->pending_participant_media_state->topology); ++i) {
		struct ast_sip_session_media *session_media = NULL;
		struct ast_stream *stream = NULL;
		unsigned int streams = local->media_count;

		/* This code does not enforce any maximum stream count limitations as that is done on either
		 * the handling of an incoming SDP offer or on the handling of a session refresh.
		 */

		stream = ast_stream_topology_get_stream(session->pending_participant_media_state->topology, i);

		session_media = ast_sip_session_participant_media_state_add_as_sub_type_and_msid(session, session->pending_participant_media_state, stream, i);

		if (!session_media) {
			return NULL;
		}

		if (change_mid_for_participant_media_reuse(session, session_media, i)) {
			ao2_ref(session_media, -1);
			return NULL;
		}
		change_participant_stream_params_for_sfu(session, session_media, stream);
		change_rtp_instance_local_ssrc(session, session_media, stream);
		change_stream_msid_according_session_media(session, session_media, stream);
		change_bundle_state(session, session_media, stream);

		if (add_sdp_streams(session_media, session, local, NULL, stream)) {
			return NULL;
		}

		/* If a stream was actually added then add any additional details */
		if (streams != local->media_count) {
			pjmedia_sdp_media *media = local->media[streams];
			pj_str_t stmp;
			pjmedia_sdp_attr *attr;

			/* Add the media identifier if present */
			if (!ast_strlen_zero(session_media->mid)) {
				attr = pjmedia_sdp_media_find_attr2(media, "mid", NULL);
				if (!attr) {
					attr = pjmedia_sdp_attr_create(inv->pool_notify_prov, "mid", pj_cstr(&stmp, session_media->mid));
					pjmedia_sdp_attr_add(&media->attr_count, media->attr, attr);
				}
			}
		}

		/* Ensure that we never exceed the maximum number of streams PJMEDIA will allow. */
		if (local->media_count == PJMEDIA_MAX_SDP_MEDIA) {
			break;
		}
	}

	/* Add any bundle groups that are present on the media state */
	if (add_participant_bundle_groups(session, inv->pool_notify_prov, local)) {
		return NULL;
	}

	/* Use the connection details of an available media if possible for SDP level */
	for (stream = 0; stream < local->media_count; stream++) {
		if (!local->media[stream]->conn) {
			continue;
		}

		if (local->conn) {
			if (!pj_strcmp(&local->conn->net_type, &local->media[stream]->conn->net_type) &&
				!pj_strcmp(&local->conn->addr_type, &local->media[stream]->conn->addr_type) &&
				!pj_strcmp(&local->conn->addr, &local->media[stream]->conn->addr)) {
				local->media[stream]->conn = NULL;
			}
			continue;
		}

		/* This stream's connection info will serve as the connection details for SDP level */
		local->conn = local->media[stream]->conn;
		local->media[stream]->conn = NULL;

		continue;
	}

	/* If no SDP level connection details are present then create some */
	if (!local->conn) {
		local->conn = pj_pool_zalloc(inv->pool_notify_prov, sizeof(struct pjmedia_sdp_conn));
		local->conn->net_type = STR_IN;
		local->conn->addr_type = session->endpoint->media.rtp.ipv6 ? STR_IP6 : STR_IP4;

		if (!ast_strlen_zero(session->endpoint->media.address)) {
			pj_strdup2(inv->pool_notify_prov, &local->conn->addr, session->endpoint->media.address);
		} else {
			pj_strdup2(inv->pool_notify_prov, &local->conn->addr, ast_sip_get_host_ip_string(session->endpoint->media.rtp.ipv6 ? pj_AF_INET6() : pj_AF_INET()));
		}
	}

	pj_strassign(&local->origin.net_type, &local->conn->net_type);
	pj_strassign(&local->origin.addr_type, &local->conn->addr_type);
	pj_strassign(&local->origin.addr, &local->conn->addr);

	return local;
}

static pjmedia_sdp_session *generate_session_participant_refresh_sdp(struct ast_sip_session *session)
{
	pjsip_inv_session *inv_session = session->inv_session;

	return create_participant_local_sdp(inv_session, session, NULL);
}

static int update_participant_media(struct ast_sip_session *session, const pjmedia_sdp_session *local, const pjmedia_sdp_session *remote)
{
	int i = 0;
	struct ast_stream_topology *topology = NULL;

	if (!session || !local || !remote) {
		return -1;
	}

	if (!session->pending_participant_media_state->topology) {
		ast_log(LOG_WARNING, "[session:%p] update_participant_media, no pending participant media state for channel '%s'\n",
			session, session->channel ? ast_channel_name(session->channel) : "unknown");
		return -1;
	}

	for (i = 0; i < local->media_count; ++i) {
		struct ast_sip_session_media *session_media;
		struct ast_stream *stream;

		if (!remote->media[i]) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->pending_participant_media_state->sessions, i);
		stream = ast_stream_topology_get_stream(session->pending_participant_media_state->topology, i);

		/* Make sure that this stream is in the correct state. If we need to change
		 * the state to REMOVED, then our work here is done, so go ahead and move on
		 * to the next stream.
		 */
		if (ast_sip_session_judge_if_disable_media(remote->media[i]) || ast_sip_session_judge_if_disable_media(local->media[i])) {
			ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);

			if (session->channel && session_media->rtp) {
				ast_log(LOG_DEBUG, "[session:%p] Negotiated SDP is complete, remote media port is zero, notify AVS, endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				ast_rtp_instance_set_transmode(session_media->rtp, AST_STREAM_STATE_REMOVED);
				ast_avs_after_nego_notify(session->channel, session_media->rtp);

			}
			session_media->changed = 0;
			continue;
		}

		if (handle_negotiated_sdp_session_media(session_media, session, local, remote, i, stream)) {
			ast_log(LOG_WARNING, "[session:%p] Negotiated SDP is failed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return -1;
		}

		if (session->channel && session_media->rtp) {
			ast_log(LOG_DEBUG, "[session:%p] Negotiated SDP is complete, notify AVS for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));

			int stream_framing = 0;
			stream_framing = ast_format_cap_get_framing(ast_stream_get_formats(stream));
			ast_rtp_instance_set_transmode(session_media->rtp, ast_stream_get_state(stream));

			if (ast_rtp_instance_get_caps(session_media->rtp)) {
				ast_format_cap_set_framing(ast_rtp_instance_get_caps(session_media->rtp), stream_framing);
			} else {
				ast_log(LOG_WARNING, "[session:%p] negotiated sdp met error: no joint codecs for endpoint:%s\n",
					session, ast_sorcery_object_get_id(session->endpoint));
				return -1;
			}

			ast_avs_after_nego_notify(session->channel, session_media->rtp);
			ast_log(LOG_NOTICE, "[session:%p] ast_avs_after_nego_notify framing is %d for endpoint:%s\n",
				session, stream_framing, ast_sorcery_object_get_id(session->endpoint));
		}

		session_media->changed = 0;
	}

	/* Apply the pending media state to the channel and make it active */
	ast_channel_lock(session->channel);

	/* Now update the stream handler for any declined/removed streams */
	for (i = 0; i < local->media_count; ++i) {
		struct ast_sip_session_media *session_media;
		struct ast_stream *stream;

		if (!remote->media[i]) {
			continue;
		}

		session_media = AST_VECTOR_GET(&session->pending_participant_media_state->sessions, i);
		stream = ast_stream_topology_get_stream(session->pending_participant_media_state->topology, i);

		if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED
			&& session_media->handler) {
			/*
			 * This stream is no longer being used and the channel's thread
			 * is held off because we have the channel lock so release any
			 * resources the handler may have on it.
			 */
			session_media_set_handler(session_media, NULL);
		}
	}

	/* Remove all current file descriptors from the channel */
	for (i = 0; i < AST_VECTOR_SIZE(&session->active_participant_media_state->read_callbacks); ++i) {
		ast_channel_internal_fd_clear(session->channel, i + AST_EXTENDED_FDS);
	}

	/* Add all the file descriptors from the pending media state */
	for (i = 0; i < AST_VECTOR_SIZE(&session->pending_participant_media_state->read_callbacks); ++i) {
		struct ast_sip_session_media_read_callback_state *callback_state;

		callback_state = AST_VECTOR_GET_ADDR(&session->pending_participant_media_state->read_callbacks, i);
		if (!ast_channel_is_avs_engine(session->channel)) {
			ast_channel_internal_fd_set(session->channel, i + AST_EXTENDED_FDS, callback_state->fd);
		} else {
			if (callback_state->session->sub_type == AST_MEDIA_SUB_APPLICATION_TYPE_BFCP_UDP && session->bfcp_para.has_udp_bfcp) {
				ast_channel_internal_fd_set(session->channel, i + AST_EXTENDED_FDS, callback_state->fd);
			}
		}
	}

	topology = ast_stream_topology_clone(session->pending_participant_media_state->topology);
	if (topology) {
		ast_channel_set_participant_stream_topology(session->channel, topology);
	}

	/* Active and pending flip flop as needed */
	SWAP(session->active_participant_media_state, session->pending_participant_media_state);
	ast_sip_session_media_state_reset(session->pending_participant_media_state);

	ast_channel_unlock(session->channel);

	return 0;
}

static void remove_media_direction(pjmedia_sdp_media *media)
{
	if (!media) {
		return;
	}

	pjmedia_sdp_media_remove_all_attr(media, "sendonly");
	pjmedia_sdp_media_remove_all_attr(media, "recvonly");
	pjmedia_sdp_media_remove_all_attr(media, "inactive");
	pjmedia_sdp_media_remove_all_attr(media, "sendrecv");

	return;
}

static void add_media_direction(struct ast_sip_session *session, pjmedia_sdp_media *media, enum ast_stream_media_direction dir)
{
	static const pj_str_t STR_SENDRECV = { "sendrecv", 8 };
	static const pj_str_t STR_SENDONLY = { "sendonly", 8 };
	static const pj_str_t STR_INACTIVE = { "inactive", 8 };
	static const pj_str_t STR_RECVONLY = { "recvonly", 8 };
	pjmedia_sdp_attr *attr = NULL;

	if (!media) {
		return;
	}

	attr = PJ_POOL_ZALLOC_T(session->inv_session->pool_notify_prov, pjmedia_sdp_attr);

	switch (dir) {
		case SENDRECV:
			attr->name = STR_SENDRECV;
			break;
		case SENDONLY:
			attr->name = STR_SENDONLY;
			break;
		case INACTIVE:
			attr->name = STR_INACTIVE;
			break;
		case RECVONLY:
			attr->name = STR_RECVONLY;
			break;
		default:
			attr->name = STR_SENDRECV;
			break;
	}

	media->attr[media->attr_count++] = attr;

	return;
}

static void change_remote_media_direction(struct ast_sip_session *session, pjmedia_sdp_session *remote)
{
	unsigned int i = 0;
	pjmedia_sdp_media *stream = NULL;
	enum ast_stream_media_direction changed_dir = UNKNOW;

	if (!remote) {
		return;
	}

	for (i = 0; i < remote->media_count; i++) {
		stream = remote->media[i];
		if (!stream) {
			continue;
		}

		if (pjmedia_sdp_media_find_attr2(stream, "sendonly", NULL)) {
			changed_dir = RECVONLY;
		} else if (pjmedia_sdp_media_find_attr2(stream, "recvonly", NULL)) {
			changed_dir = SENDONLY;
		} else if (pjmedia_sdp_media_find_attr2(stream, "inactive", NULL)) {
			changed_dir = INACTIVE;
		} else {
			changed_dir = SENDRECV;
		}

		remove_media_direction(stream);
		add_media_direction(session, stream, changed_dir);
	}
}

/* Util: swap pool */
static void swap_notify_pool(pj_pool_t **p1, pj_pool_t **p2)
{
    pj_pool_t *tmp = *p1;
    *p1 = *p2;
    *p2 = tmp;
}

static void change_ssrc_attr(struct ast_sip_session *session, pjmedia_sdp_media **orig_stream, pjmedia_sdp_media *recv_stream)
{
	unsigned int i = 0;
	pjmedia_sdp_attr *recv_attr = NULL;
	pjmedia_sdp_attr *clone_attr = NULL;

	pjmedia_sdp_media_remove_all_attr(*orig_stream, "ssrc");
	for (i = 0; i < recv_stream->attr_count; i++) {
		recv_attr = recv_stream->attr[i];
		if (pj_strcmp2(&recv_attr->name, "ssrc")) {
			continue;
		}
		clone_attr = pjmedia_sdp_attr_clone(session->inv_session->pool_notify_prov, recv_attr);
		pjmedia_sdp_media_add_attr(*orig_stream, clone_attr);
	}

	return;
}

static pjmedia_sdp_session *change_remote_ssrc_attrs(struct ast_sip_session *session,
	pjmedia_sdp_session *remote, pjsip_rx_data *rdata)
{
	unsigned int i = 0;
	pjmedia_sdp_media *orig_stream = NULL;
	pjmedia_sdp_media *recv_stream = NULL;
	pjsip_rdata_sdp_info *sdp_info = NULL;
	pjmedia_sdp_attr *orig_attr = NULL;
	pjmedia_sdp_attr *recv_attr = NULL;
	pjmedia_sdp_attr *clone_attr = NULL;

	sdp_info = pjsip_rdata_get_special_sdp_info(rdata, "participant-sdp");
	if (!sdp_info || !sdp_info->sdp) {
		ast_log(LOG_WARNING, "[session:%p] change_remote_ssrc can't get remote SDP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return remote;
	}

	for (i = 0; i < remote->media_count; i++) {
		orig_stream = remote->media[i];
		if (!orig_stream) {
			continue;
		}

		recv_stream = sdp_info->sdp->media[i];
		if (!recv_stream) {
			continue;
		}

		orig_attr = pjmedia_sdp_media_find_attr2(orig_stream, "ssrc-group", NULL);
		if (orig_attr) {
			recv_attr = pjmedia_sdp_media_find_attr2(recv_stream, "ssrc-group", NULL);
			if (recv_attr) {
				pjmedia_sdp_media_remove_all_attr(orig_stream, "ssrc-group");
				clone_attr = pjmedia_sdp_attr_clone(session->inv_session->pool_notify_prov, recv_attr);
				pjmedia_sdp_media_add_attr(orig_stream, clone_attr);
				change_ssrc_attr(session, &orig_stream, recv_stream);
			} else {
				ast_log(LOG_WARNING, "[session:%p] change_remote_ssrc can't get matched ssrc-group attr, remove anyway, may cause abnormal, postion:%d, endpoint:%s\n",
					session, i, ast_sorcery_object_get_id(session->endpoint));
				pjmedia_sdp_media_remove_all_attr(orig_stream, "ssrc-group");
				change_ssrc_attr(session, &orig_stream, recv_stream);
			}
		} else {
			change_ssrc_attr(session, &orig_stream, recv_stream);
		}
	}

	return remote;
}

static pjmedia_sdp_attr *sdp_session_find_attr(const pjmedia_sdp_session *sdp, const char *name)
{
	int index;
	pjmedia_sdp_attr *attr = NULL;

	for (index = 0; index < sdp->attr_count; ++index) {
		attr = sdp->attr[index];
		if (!attr) {
			continue;
		}
		if (pj_strcmp2(&attr->name, name)) {
			continue;
		}

		return attr;
	}

	return NULL;
}

static void add_group_for_participant(struct ast_sip_session *session,
	pjmedia_sdp_session *remote, pjmedia_sdp_session *local)
{
	pjmedia_sdp_attr *attr = NULL;
	pjmedia_sdp_attr *clone_attr = NULL;

	if (sdp_session_find_attr(remote, "group")) {
		return;
	}

	attr = sdp_session_find_attr(local, "group");
	if (!attr) {
		ast_log(LOG_WARNING, "[session:%p] Can't get group attr in local SDP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	clone_attr = pjmedia_sdp_attr_clone(session->inv_session->pool_notify_prov, attr);
	if (!clone_attr) {
		ast_log(LOG_WARNING, "[session:%p] Can't clone attr, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	pjmedia_sdp_session_add_attr(remote, clone_attr);

	return;
}

static pjmedia_sdp_session *make_remote_sdp_session(struct ast_sip_session *session,
	pjmedia_sdp_session *local, pjsip_rx_data *rdata)
{
	pjmedia_sdp_session *remote = NULL;

	pjsip_rdata_sdp_info *sdp_info = NULL;
	sdp_info = pjsip_rdata_get_special_sdp_info(rdata, "participant-sdp");
	if (sdp_info && sdp_info->sdp) {
		ast_log(LOG_DEBUG, "[session:%p] Get 200OK's SDP, just use remote SDP, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		remote = pjmedia_sdp_session_clone(session->inv_session->pool_notify_prov, sdp_info->sdp);
		add_group_for_participant(session, remote, local);
		return remote;
	}

	remote = pjmedia_sdp_session_clone(session->inv_session->pool_notify_prov, local);
	if (!remote) {
		return NULL;
	}

	change_remote_media_direction(session, remote);

	change_remote_ssrc_attrs(session, remote, rdata);

	return remote;
}

static int send_delayed_notify_refresh_participant(void *data)
{
	struct ast_sip_session_media_state *media_state = NULL;
	struct ast_sip_session *session = data;

	if (!session) {
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_delayed_notify_refresh_participant, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->sending_notify_count >= 2) {
		ast_log(LOG_DEBUG, "[session:%p] send_delayed_notify_refresh_participant, sending_notify_count:%d, retain one to get the newest topology, endpoint:%s\n",
			session, session->sending_notify_count, ast_sorcery_object_get_id(session->endpoint));
		session->sending_notify_count = 1;
	} else {
		session->sending_notify_count = 0;
	}
	media_state = get_session_media_state(session, 1);
	if (media_state && session->active_participant_media_state->topology &&
		ast_stream_topology_equal(session->active_participant_media_state->topology, media_state->topology)) {
		ast_log(LOG_DEBUG, "[session:%p] send_delayed_notify_refresh_participant, topology already equal, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_free(media_state);
		session->sending_notify_count = 0;
		ao2_ref(session, -1);
		return 0;
	}
	ast_sip_session_refresh_participant_for_sfu(session, 1, ast_sip_session_participant_topology_change_response, media_state);

	ao2_ref(session, -1);

	return 0;

}

static void push_sfu_delayed_notify_to_task(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] Endpoint:%s has delayed notify, push to session task to send\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, send_delayed_notify_refresh_participant, session)) {
		ast_log(LOG_WARNING, "[session:%p] Endpoint:%s push delayed notify to task failed\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->sending_notify_count = 0;
		ao2_ref(session, -1);
	}

	return;
}

static void notify_full_participant_sdp_for_reconnect(struct ast_sip_session *session)
{
	if (!session) {
		return;
	}

	if (!session->notify_full_participant_sdp) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] notify_full_participant_sdp_for_reconnect for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session->notify_full_participant_sdp = 0;

	push_sfu_delayed_notify_to_task(session);

	return;
}

static void send_bfcp_info_for_reconnect(struct ast_sip_session *session)
{
	struct bfcp_info_data *bfcp_data = NULL;

	if (!session) {
		return;
	}

	if (!session->pending_bfcp_info) {
		return;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_bfcp_info_for_reconnect for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	bfcp_data = ast_sip_session_bfcp_data_alloc(session, &session->last_bfcp_info);
	if (!bfcp_data) {
		ast_log(LOG_ERROR, "[session:%p] send_bfcp_info_for_reconnect can't alloc bfcp_data for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}

	session->pending_bfcp_info = 0;

	if (session->is_sfu) {
		if (ast_sip_push_task(session->serializer, ast_sip_session_bfcp_send_conference_sip_info, bfcp_data)) {
			ao2_cleanup(bfcp_data);
		}
	} else {
		if (ast_sip_push_task(session->serializer, ast_sip_session_bfcp_send_sip_info, bfcp_data)) {
			ao2_cleanup(bfcp_data);
		}
	}

	return;
}

static void handle_2XX_response_for_notify(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	int res = 0;
	pjmedia_sdp_session *remote_notify_sdp = NULL;

	if (!session || !session->channel) {
		return;
	}

	if (session->sending_notify_count > 0) {
		session->sending_notify_count--;	
	}

	remote_notify_sdp = make_remote_sdp_session(session, session->neg_notify_sdp, rdata);
	res = update_participant_media(session, session->neg_notify_sdp, remote_notify_sdp);

	if (res == 0) {
		swap_notify_pool(&session->inv_session->pool_notify_prov, &session->inv_session->pool_notify_active);
	} else {
		ast_log(LOG_DEBUG, "[session:%p] update_participant_media failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
	}
	pj_pool_reset(session->inv_session->pool_notify_prov);

	if (session->sending_notify_count) {
		push_sfu_delayed_notify_to_task(session);
	}

	session->sending_participant_notify = 0;

	return;
}

int ast_sip_session_participant_topology_change_response(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	if (PJSIP_IS_STATUS_IN_CLASS(rdata->msg_info.msg->line.status.code, 200)) {
		handle_2XX_response_for_notify(session, rdata);
	} else if (300 <= rdata->msg_info.msg->line.status.code) {
		/* The topology change failed, so drop the current pending media state */
		ast_sip_session_media_state_reset(session->pending_participant_media_state);
		session->sending_participant_notify = 0;
	}

	return 0;
}

static int ast_print_attr(const pjmedia_sdp_attr *attr,
			     char *buf, pj_size_t len)
{
	char *p = buf;

	if ((int)len < attr->name.slen + attr->value.slen + 10) {
		return -1;
	}

	*p++ = 'a';
	*p++ = '=';
	pj_memcpy(p, attr->name.ptr, attr->name.slen);
	p += attr->name.slen;

	if (attr->value.slen) {
		*p++ = ':';
		pj_memcpy(p, attr->value.ptr, attr->value.slen);
		p += attr->value.slen;
	}

	*p++ = '\r';
	*p++ = '\n';

	return p-buf;
}

static int ast_print_media_desc( pjmedia_sdp_media *m, char *buf, int len)
{
	char *p = buf;
	char *end = buf+len;
	unsigned i;
	int printed;

	/* check length for the "m=" line. */
	if (len < m->desc.media.slen+m->desc.transport.slen+12+24) {
		return -1;
	}
	*p++ = 'm';	    /* m= */
	*p++ = '=';
	pj_memcpy(p, m->desc.media.ptr, m->desc.media.slen);
	p += m->desc.media.slen;
	*p++ = ' ';
	printed = pj_utoa(m->desc.port, p);
	p += printed;
	if (m->desc.port_count > 1) {
		*p++ = '/';
		printed = pj_utoa(m->desc.port_count, p);
		p += printed;
	}
	*p++ = ' ';
	pj_memcpy(p, m->desc.transport.ptr, m->desc.transport.slen);
	p += m->desc.transport.slen;
	for (i=0; i<m->desc.fmt_count; ++i) {
		*p++ = ' ';
		pj_memcpy(p, m->desc.fmt[i].ptr, m->desc.fmt[i].slen);
		p += m->desc.fmt[i].slen;
	}
	*p++ = '\r';
	*p++ = '\n';

	/* print attributes. */
	for (i=0; i<m->attr_count; ++i) {
		printed = (int)ast_print_attr(m->attr[i], p, end-p);
		if (printed < 0) {
			return -1;
		}
		p += printed;
	}

	return (int)(p-buf);
}


int ast_sip_session_print_sdp(const pjmedia_sdp_session *ses,
			 char *buf, pj_ssize_t len)
{
    char *p = buf;
    char *end = buf+len;
    unsigned int i = 0;
    int printed = 0;

    /* Print media (m=) lines. */
    for (i = 0; i < ses->media_count; ++i) {
		printed = ast_print_media_desc(ses->media[i], p, (int)(end-p));
		if (printed < 0) {
		    return -1;
		}
		p += printed;
    }

    return (int)(p-buf);
}

static int send_notify_to_refresh_participant(struct ast_sip_session *session,
	ast_sip_session_response_cb on_response, pjmedia_sdp_session *new_sdp)
{
	char *sdp = NULL;
	int max_sdp_len = 12800;
	pjsip_tx_data *tdata = NULL;

	if (ast_sip_create_request("NOTIFY", session->inv_session->dlg, session->endpoint, NULL, NULL, &tdata)) {
		ast_log(LOG_ERROR, "Could not create NOTIFY request for refresh participant\n");
		return -1;
	}

	sdp = ast_malloc(max_sdp_len);
	if (!sdp) {
		ast_log(LOG_ERROR, "Could not malloc NOTIFY request SDP for refresh participant\n");
		return -1;
	}
	memset(sdp, 0, max_sdp_len);

	ast_sip_session_print_sdp(new_sdp, sdp, max_sdp_len - 1);

	struct ast_sip_body body = {
		.type = "application",
		.subtype = "participant-sdp",
		.body_text = sdp
	};

	if (ast_sip_add_body(tdata, &body)) {
		ast_log(LOG_ERROR, "Could not add body to refresh participant request\n");
		ao2_ref(session, -1);
		ast_free(sdp);
		sdp = NULL;
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] Sending session refresh SDP via NOTIFY to %s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session->neg_notify_sdp = new_sdp;

	ast_sip_session_send_request_with_cb(session, tdata, on_response);

	ast_free(sdp);
	sdp = NULL;

	return 0;
}

static int check_param_state_is_valid(struct ast_sip_session *session, struct ast_sip_session_media_state *media_state)
{
	pjsip_inv_session *inv_session = NULL;

	if (!session) {
		return -1;
	}

	inv_session = session->inv_session;
	if (!inv_session) {
		return -1;
	}

	if (!media_state) {
		ast_debug(3, "[session:%p] Not sending participant media refresh notify to %s because has no new media_state ...\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return -1;
	}

	if (media_state && !media_state->topology) {
		ast_debug(3, "[session:%p] Not sending participant media refresh notify to %s because new media_state has no topology...\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_free(media_state);
		return -1;
	}

	if (inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		/* Don't try to do anything with a hung-up call */
		ast_debug(3, "[session:%p] Not sending participant media refresh notify to %s because of disconnected state...\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_free(media_state);
		return 0;
	}

	/* If the dialog has not yet been established we have to defer until it has */
	if (inv_session->dlg->state != PJSIP_DIALOG_STATE_ESTABLISHED) {
		ast_debug(3, "[session:%p] Not sending participant media refresh notify to %s because dialog has not been established...\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_free(media_state);
		return 0;
	}

	return 1;
}

static unsigned int judge_if_invalid_caps(struct ast_format_cap *joint_cap)
{
	if (!joint_cap) {
		return TRUE;
	}

	if (!ast_format_cap_count(joint_cap)) {
		return TRUE;
	}

	if (ast_format_cap_has_rtx(joint_cap) && ast_format_cap_count(joint_cap) == 1) {
		return TRUE;
	}

	if (ast_format_cap_has_video_fec(joint_cap) && ast_format_cap_count(joint_cap) == 1) {
		return TRUE;
	}

	if (ast_format_cap_has_rtx(joint_cap) &&
		ast_format_cap_has_video_fec(joint_cap) &&
		ast_format_cap_count(joint_cap) == 2) {
		return TRUE;
	}

	return FALSE;
}

int ast_sip_session_translate_gswave_to_sip_behavior(struct ast_sip_session *session,
	struct ast_stream *req_stream, struct ast_channel *requestor)
{
	enum ast_terminal_type terminal_type = AST_TERMINAL_TYPE_UNKNOW;

	if (!session || !req_stream || !requestor) {
		return 0;
	}

	if (session->is_webrtc_call || (session->terminal_type == AST_TERMINAL_TYPE_WEBRTC)) {
		return 0;
	}

	if (ast_channel_tech(requestor)->get_terminal_type) {
		terminal_type = ast_channel_tech(requestor)->get_terminal_type(requestor);
		if (terminal_type != AST_TERMINAL_TYPE_WEBRTC && terminal_type != AST_TERMINAL_TYPE_UNKNOW) {
			return 0;
		}
	}

	if (ast_stream_get_type(req_stream) == AST_MEDIA_TYPE_AUDIO) {
		return 0;
	}

	if (ast_stream_get_remote_media_direction(req_stream) != INACTIVE) {
		return 0;
	}

	ast_log(LOG_DEBUG, "[session:%p] Peer terminal is WEBRTC, and local is SIP, don't append inactive media, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	return 1;
}

static int modify_stream_according_endpoint_configure(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state, struct ast_stream *stream, struct ast_stream *existing_stream, int *index)
{
	struct ast_format_cap *joint_cap = NULL;

	if (ast_stream_get_type(stream) != AST_MEDIA_TYPE_VIDEO) {
		return 0;
	}

	if (!ast_stream_get_formats(stream)) {
		ast_log(LOG_DEBUG, "[session:%p] Can't get stream:%p format, endpoint:%s\n",
			session, stream, ast_sorcery_object_get_id(session->endpoint));
		return 0;
	}
	
	joint_cap = ast_format_cap_alloc(AST_FORMAT_CAP_FLAG_DEFAULT);
	if (!joint_cap) {
		ast_sip_session_media_state_free(media_state);
		return -1;
	}

	ast_format_cap_get_compatible(ast_stream_get_formats(stream), session->endpoint->media.codecs, joint_cap);

	/* Modify bf bfdu, if joint_cap only has rtx, don't add this media to stream */
	if (judge_if_invalid_caps(joint_cap)) {
		ao2_ref(joint_cap, -1);

		if (!existing_stream) {
			/* If there is no existing stream we can just not have this stream in the topology
			 * at all.
			 */
			ast_stream_topology_del_stream(media_state->topology, *index);
			*index -= 1;
			return 0;
		} else if (ast_stream_get_state(stream) != ast_stream_get_state(existing_stream) ||
				strcmp(ast_stream_get_name(stream), ast_stream_get_name(existing_stream))) {
			/* If the underlying stream is a different type or different name then we have to
			 * mark it as removed, as it is replacing an existing stream. We do this so order
			 * is preserved.
			 */
			ast_log(LOG_DEBUG, "[session:%p] Can't get valid video codec in this stream, set stream:%p (index:%d) removed for endpoint:%s\n",
				session, stream, *index, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
			return 0;
		} else {
			/* However if the stream is otherwise remaining the same we can keep the formats
			 * that exist on it already which allows media to continue to flow.
			 */
			joint_cap = ao2_bump(ast_stream_get_formats(existing_stream));
		}
	}

	if (session->is_sfu && 
		(ast_stream_get_type(stream) == AST_MEDIA_TYPE_VIDEO) &&
		(session->terminal_type == AST_TERMINAL_TYPE_WEBRTC) &&
		!ast_format_cap_if_suppport_webrtc(joint_cap)) {
		ao2_ref(joint_cap, -1);
		if (!existing_stream) {
			ast_log(LOG_DEBUG, "[session:%p] Webrtc can't support this joint cap codec, delete this stream for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_topology_del_stream(media_state->topology, *index);
			*index -= 1;
		} else {
			ast_log(LOG_DEBUG, "[session:%p] Webrtc can't support this joint cap codec, set this stream removed for endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			ast_stream_set_state(stream, AST_STREAM_STATE_REMOVED);
		}
		return 0;
	}

	ast_stream_set_formats(stream, joint_cap);
	ao2_cleanup(joint_cap);

	return 0;
}

static int modify_pending_participant_media_state(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state)
{
	int res = 0;
	int index = 0;
	struct ast_stream *stream = NULL;
	struct ast_stream *existing_stream = NULL;

	for (index = 0; index < ast_stream_topology_get_count(media_state->topology); ++index) {
		stream = ast_stream_topology_get_stream(media_state->topology, index);
		if (!judge_if_is_participant_stream(stream)) {
			continue;
		}

		if (session->active_participant_media_state->topology &&
			index < ast_stream_topology_get_count(session->active_participant_media_state->topology)) {
			existing_stream = ast_stream_topology_get_stream(session->active_participant_media_state->topology, index);
		}

		/* No need to do anything with stream if it's media state is removed */
		if (ast_stream_get_state(stream) == AST_STREAM_STATE_REMOVED) {
			/* If there is no existing stream we can just not have this stream in the topology at all. */
			if (!existing_stream) {
				ast_stream_topology_del_stream(media_state->topology, index);
				index -= 1;
			}
			continue;
		}

		/* Enforce the configured allowed codecs on audio and video streams */
		res = modify_stream_according_endpoint_configure(session, media_state, stream, existing_stream, &index);
	}

	return res;
}

static struct ast_sip_session_media_state *clone_media_state_only_participant_part(struct ast_sip_session_media_state *media_state)
{
	struct ast_sip_session_media_state *media_state_new = NULL;

	media_state_new = ast_sip_session_media_state_alloc();
	if (!media_state_new) {
		return NULL;
	}

	media_state_new->topology = ast_stream_topology_clone_participant(media_state->topology);

	return media_state_new;
}

static int append_missed_stream_to_topology(struct ast_sip_session *session,
	struct ast_sip_session_media_state *media_state)
{
	int index = 0;
	struct ast_stream *stream = NULL;

	if (!session->active_participant_media_state->topology) {
		return 0;
	}

	/* SDP is a fun thing. Take for example the fact that streams are never removed. They just become
	 * declined. To better handle this in the case where something requests a topology change for fewer
	 * streams than are currently present we fill in the topology to match the current number of streams
	 * that are active.
	 */
	for (index = ast_stream_topology_get_count(media_state->topology);
		index < ast_stream_topology_get_count(session->active_participant_media_state->topology); ++index) {
		struct ast_stream *cloned;

		stream = ast_stream_topology_get_stream(session->active_participant_media_state->topology, index);
		ast_assert(stream != NULL);

		cloned = ast_stream_clone(stream, NULL);
		if (!cloned) {
			ast_sip_session_media_state_free(media_state);
			return -1;
		}

		ast_log(LOG_DEBUG, "[session:%p] Active media stream can't find, set this stream:%p (index:%d) removed for endpoint:%s\n",
			session, cloned, index, ast_sorcery_object_get_id(session->endpoint));

		ast_stream_set_state(cloned, AST_STREAM_STATE_REMOVED);
		if (ast_stream_topology_append_stream(media_state->topology, cloned) < 0) {
			ast_stream_free(cloned);
			ast_sip_session_media_state_free(media_state);
			return -1;
		}
	}

	/* If the resulting media state matches the existing active state don't bother doing a session refresh */
	if (ast_stream_topology_equal(session->active_participant_media_state->topology, media_state->topology)) {
		ast_sip_session_media_state_free(media_state);
		return 1;
	}

	return 0;
}

int ast_sip_session_refresh_participant_for_sfu(struct ast_sip_session *session,
	int generate_new_sdp, ast_sip_session_response_cb on_response, struct ast_sip_session_media_state *refresh_media_state)
{
	int res = 0;
	pjmedia_sdp_session *new_sdp = NULL;
	struct ast_sip_session_media_state *media_state = NULL;

	res = check_param_state_is_valid(session, refresh_media_state);
	if (res != 1) {
		ast_log(LOG_DEBUG, "[session:%p] Check param failed. don't send notify to refresh, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->sending_notify_count = 0;
		return res;
	}

	if (session->pending_participant_media_state && session->pending_participant_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] Alreay has pending notify. don't send notify to refresh, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_free(refresh_media_state);
		session->sending_notify_count++;
		return 0;
	}

	media_state = clone_media_state_only_participant_part(refresh_media_state);
	if (!media_state) {
		ast_log(LOG_DEBUG, "[session:%p] Can't clone participant media state. don't send notify to refresh, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->sending_notify_count = 0;
		ast_sip_session_media_state_free(refresh_media_state);
		return -1;
	}

	ast_sip_session_media_state_free(refresh_media_state);

	res = modify_pending_participant_media_state(session, media_state);
	if (res) {
		ast_log(LOG_DEBUG, "[session:%p] Modify pending media failed. don't send notify to refresh, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->sending_notify_count = 0;
		return res;
	}

	res = append_missed_stream_to_topology(session, media_state);
	if (res) {
		ast_log(LOG_DEBUG, "[session:%p] Append missed stream failed. don't send notify to refresh, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		session->sending_notify_count = 0;
		return res;
	}

	session->sending_notify_count++;

	ast_sip_session_media_state_free(session->pending_participant_media_state);
	session->pending_participant_media_state = media_state;

	new_sdp = generate_session_participant_refresh_sdp(session);
	if (!new_sdp) {
		ast_log(LOG_ERROR, "[session:%p] Failed to generate session refresh SDP. Not sending session refresh, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_participant_media_state);
		session->sending_notify_count = 0;
		return -1;
	}

	res = send_notify_to_refresh_participant(session, on_response, new_sdp);
	if (res) {
		ast_log(LOG_ERROR, "[session:%p] Failed to send NOTIFY to refresh participant SDP. endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_participant_media_state);
		session->sending_notify_count = 0;
		return res;
	}

	session->sending_participant_notify = 1;

	return 0;
}


static int send_reqeust_control_frame_to_peer(struct ast_sip_session *session)
{
	struct ast_sip_reinvite_data reinvite_data;

	if (!session) {
		ast_log(LOG_ERROR, "[session:%p] send_reqeust_control_frame_to_peer\n", session);
		return -1;
	}

	ast_log(LOG_DEBUG, "[session:%p] send_reqeust_control_frame_to_peer for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (!session->channel){
		return -1;
	}

	reinvite_data.session = session;
	ao2_ref(session, +1);
	session->ref_added = 1;
	if (ast_queue_control_data(session->channel, AST_CONTROL_REINVITE_REQ, &reinvite_data, sizeof(reinvite_data))) {
		ast_log(LOG_ERROR, "[session:%p] Queue AST_CONTROL_REINVITE_REQ frame failed for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
		session->ref_added = 0;
		return -1;
	}

	return 0;
}

static void send_close_camera_for_webrtc(struct ast_sip_session *session, struct ast_stream *stream)
{
	if (!session || !session->channel || !stream) {
		return;
	}

	if (session->terminal_type != AST_TERMINAL_TYPE_WEBRTC) {
		return;
	}

	if (!ast_stream_judge_if_main_video(stream)) {
		return;
	}

	if (ast_stream_get_state(stream) != AST_STREAM_STATE_INACTIVE) {
		return;
	}

	ast_log(LOG_NOTICE, "[session:%p] Send AST_CONTROL_UPDATE_VIDEO_PORT_ZERO for webrtc, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_queue_control(session->channel, AST_CONTROL_UPDATE_VIDEO_PORT_ZERO);

	return;
}

static void process_reinvite_with_no_sdp(struct ast_sip_session *session)
{
	ast_log(LOG_DEBUG, "[session:%p] Recv re-invite no sdp, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session->recv_reinvite_no_sdp = 1;
	modify_active_sdp_for_reinvite_no_sdp(session);

	if (!session->disabled_video) {
		send_reqeust_control_frame_to_peer(session);
	}

	return;
}

static pj_status_t check_reinvite_params(pjsip_inv_session *inv, const pjmedia_sdp_session *offer,
	pjsip_rx_data *rdata)
{
	pjmedia_sdp_session *caller_sdp_local = NULL;
	struct ast_sip_session *session = NULL;

	if (NULL == inv || NULL == rdata) {
		ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite, param is NULL\n");
		return PJ_TRUE;
	}

	session = inv->mod_data[session_module.id];
	if (NULL == session) {
		ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite, session is NULL\n");
		return PJ_TRUE;
	}

	if (NULL == offer) {
		process_reinvite_with_no_sdp(session);
		return PJ_TRUE;
	} else {
		session->recv_reinvite_no_sdp = 0;
	}

	if (judge_if_sdp_has_image(offer) && !session->is_direct_media && !session->joint_ipvt) {
		ast_log(LOG_DEBUG, "[session:%p] session_inv_on_rx_reinvite, sdp has image, just return, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_TRUE;
    }

	if (!session->is_direct_media && !session->joint_ipvt && (pjmedia_sdp_neg_get_state(inv->neg) != PJMEDIA_SDP_NEG_STATE_WAIT_NEGO)) {
		ast_log(LOG_DEBUG, "[session:%p] session_inv_on_rx_reinvite, neg state is not wait neg or remote offer endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_TRUE;
	}

	if (!session->inv_session || !session->inv_session->neg) {
		ast_log(LOG_WARNING, "[session:%p] session_inv_on_rx_reinvite, inv_session or neg is null, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_TRUE;
	}

	caller_sdp_local = session->inv_session->neg->active_local_sdp;
	if (!caller_sdp_local) {
		ast_log(LOG_WARNING, "[session:%p] session_inv_on_rx_reinvite, active_local_sdp is null. endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_TRUE;
	}

	return PJ_SUCCESS;
}

static int change_ssrc_for_reinvite(struct ast_sip_session *session, const pjmedia_sdp_session *offer)
{
	int is_ssrc_changed = 0;

	if (session->is_direct_media) {
		return 0;
	}

	is_ssrc_changed = change_ssrc_for_hold(session, offer);
	if (!is_ssrc_changed) {
		is_ssrc_changed = change_ssrc_for_unhold(session, offer);
	}

	return is_ssrc_changed;
}

static pj_status_t handle_for_reinvite_to_peer(struct ast_sip_session *session,
	pjsip_rx_data *rdata, int is_ssrc_changed)
{
	pjsip_tx_data *tdata = NULL;

	if (pjsip_inv_initial_answer(session->inv_session, rdata, 100, NULL, NULL, &tdata) != PJ_SUCCESS) {
		pjsip_inv_terminate(session->inv_session, 500, PJ_TRUE);
		return PJ_TRUE;
	}

	ast_sip_session_send_response(session, tdata);

	session->is_media_change = 1;

	if (!is_ssrc_changed) {
		change_ssrc_for_media_change(session);
	}

	if (send_reqeust_control_frame_to_peer(session)) {
		return PJ_TRUE;
	}

	return PJ_SUCCESS;
}

static int clear_pending_notify_state(void *data)
{
	struct ast_sip_session *session = data;

	if (!session) {
		return -1;
	}

	if (!session->sending_participant_notify) {
		ast_log(LOG_DEBUG, "[session:%p] clear_pending_notify_state, has no pending notify, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
		return 0;
	}

	if (session->pending_participant_media_state && session->pending_participant_media_state->topology) {
		ast_log(LOG_DEBUG, "[session:%p] clear_pending_notify_state get pending topology, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ast_sip_session_media_state_reset(session->pending_participant_media_state);
		session->sending_notify_count = 0;
		session->sending_participant_notify = 0;
	}

	ao2_ref(session, -1);

	return 0;
}

static void set_notify_participant_sdp_flag(pjsip_rx_data *rdata, struct ast_sip_session *session)
{
	pjsip_msg *msg = NULL;
	pjsip_generic_string_hdr *hdr = NULL;
	const pj_str_t name = { "Notify-participant-sdp", 22 };
	const pj_str_t new_name = { "X-GS-Notify-Participant-SDP", 27 };

	if (!rdata || !session) {
		ast_log(LOG_ERROR, "set_notify_participant_sdp_flag, para is invalid\n");
		return;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		return;
	}

	hdr = pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	if (!hdr) {
		hdr = pjsip_msg_find_hdr_by_name(msg, &new_name, NULL);
		if (!hdr) {
			return;
		}
	}

	/*start added  by zyxia for 268128, paging dont send notify sdp when reconnect*/
	if(session->channel && ast_channel_page_mode_get(session->channel)){
		ast_log(LOG_DEBUG, "[session:%p] Get Notify-participant-sdp hdr, is paging ,ignore it, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return;
	}
	/*End added  by zyxia for 268128, paging dont send notify sdp when reconnect*/

	ast_log(LOG_DEBUG, "[session:%p] Get Notify-participant-sdp hdr, set notify_full_participant_sdp to true, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	session->notify_full_participant_sdp = 1;

	ao2_ref(session, +1);
	if (ast_sip_push_task(session->serializer, clear_pending_notify_state, session)) {
		ast_log(LOG_WARNING, "[session:%p] Endpoint:%s push clear pending notify to task failed\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		ao2_ref(session, -1);
	}

	return;
}

unsigned int ast_sip_session_judge_if_need_add_rpid(const struct ast_sip_session *session)
{
	const char *is_nway_conf = NULL;

	if (!session || !session->channel) {
		return FALSE;
	}

	if (ast_channel_get_is_callflip(session->channel)) {
		return TRUE;
	}

	is_nway_conf = pbx_builtin_getvar_helper(session->channel, "NWAY_CONF_NUMBER");
	if (!ast_strlen_zero(is_nway_conf)) {
		return TRUE;
	}

	return FALSE;
}

static pj_status_t session_inv_on_session_timer_update(pjsip_inv_session *inv, pjsip_tx_data *tdata)
{
	struct ast_sip_session *session = NULL;

	if (!inv || !tdata) {
		return PJ_TRUE;
	}

	session = inv->mod_data[session_module.id];
	if (!session) {
		return PJ_TRUE;
	}

	if (!ast_sip_session_judge_if_need_add_rpid(session)) {
		ast_log(LOG_DEBUG, "[session:%p] Don't need update session timer request for endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_TRUE;
	}

	ast_log(LOG_DEBUG, "[session:%p] session_inv_on_session_timer_update for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->channel) {
		ast_sip_add_header(tdata, "Channel-ID", ast_channel_name(session->channel));
	}

	handle_outgoing_request(session, tdata);

	return PJ_SUCCESS;
}

/*Start added by zyxia for Bug 268897,2023.8.16*/
static int resend_nway_notify_frame(void *data)
{
	struct ast_sip_session *session = data;
	struct conf_notify_switch_data s_data;

	if (!session || !session->channel || ast_strlen_zero(session->conf_room)) {
		return -1;
	}

	memset(&s_data, 0, sizeof(s_data));
	snprintf(s_data.conf_id, sizeof(s_data.conf_id), "%s",session->conf_room);
	
	s_data.is_video_conf = 1;
	s_data.enabled_cei = 1;
	s_data.is_admin = ast_channel_get_is_admin(session->channel);
	if (ast_channel_get_is_callflip(session->channel)) {
		s_data.event_type = 1;
	}

	ast_channel_lock(session->channel);
	if (ast_channel_tech(session->channel)->indicate) {
		ast_channel_tech(session->channel)->indicate(session->channel, AST_CONTROL_NOTIFY_N_WAY_CONF, &s_data, sizeof(s_data));
	}
	ast_channel_unlock(session->channel);

	ao2_ref(session, -1);

	return 0;
}
/*End added by zyxia for Bug 268897,2023.8.16*/

static pj_status_t session_inv_on_rx_reinvite_conflict(pjsip_inv_session *inv, pjsip_transaction *tsx, pjsip_rx_data *rdata)
{
	pjsip_msg *msg = NULL;
	pjsip_generic_string_hdr *hdr = NULL;
	struct ast_sip_session *session = NULL;
	const pj_str_t name = { "Notify-participant-sdp", 22 };
	const pj_str_t new_name = { "X-GS-Notify-Participant-SDP", 27 };

	if (!inv || !tsx|| !rdata ) {
		ast_log(LOG_ERROR, "session_inv_on_rx_reinvite_conflict, para is invalid\n");
		return PJ_TRUE;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		return PJ_TRUE;
	}

	hdr = pjsip_msg_find_hdr_by_name(msg, &name, NULL);
	if (!hdr) {
		hdr = pjsip_msg_find_hdr_by_name(msg, &new_name, NULL);
		if (!hdr) {
			return PJ_TRUE;
		}
	}

	session = inv->mod_data[session_module.id];
	if (!session) {
		return PJ_TRUE;
	}

	ast_log(LOG_DEBUG, "[session:%p] It's a network reconnect reinvite, just reset previous reinvite, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	reset_reinvite_pending_state(session, inv, tsx);

	return PJ_SUCCESS;
}

static pj_status_t session_inv_on_cancel_offer(pjsip_inv_session *inv)
{
	struct ast_sip_session *session = NULL;

	if (!inv) {
		return PJ_TRUE;
	}

	session = inv->mod_data[session_module.id];
	if (!session) {
		return PJ_TRUE;
	}

	if (inv->invite_tsx) {
		ast_log(LOG_DEBUG, "[session:%p] Has pending invite, don't reset pending media, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_TRUE;
	}

	if (!session->pending_media_state) {
		return PJ_TRUE;
	}

	if (!session->pending_media_state->topology) {
		return PJ_TRUE;
	}

	ast_log(LOG_DEBUG, "[session:%p] Update's 200OK has no SDP, just reset pending media, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	ast_sip_session_media_state_reset(session->pending_media_state);

	return PJ_SUCCESS;
}

void ast_sip_session_add_support_video_call_for_webrtc_trunk(struct ast_sip_session *session, pjsip_tx_data *tdata)
{
	pjsip_supported_hdr *hdr = NULL;
	pj_str_t VIDEO_CALL_SUPPORTED_NAME = { "videoCall", 9 };

	if (!session || !session->endpoint || !tdata) {
		return;
	}

	if (strncasecmp(ast_sorcery_object_get_id(session->endpoint), "webrtc_trunk", strlen("webrtc_trunk"))) {
		return;
	}

	if (!session->endpoint->support_video_call) {
		return;
	}

	hdr = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_SUPPORTED, NULL);
	if (!hdr) {
		/* insert a new Supported header */
		hdr = pjsip_supported_hdr_create(tdata->pool);
		if (!hdr) {
			ast_log(LOG_WARNING, "[session:%p] Can't create supported hdr, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return;
		}

		pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr *)hdr);
	}

	ast_log(LOG_DEBUG, "[session:%p] ast_sip_session_add_support_video_call_for_webrtc_trunk, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	pj_strassign(&hdr->values[hdr->count++], &VIDEO_CALL_SUPPORTED_NAME);

	return;
}

void ast_sip_session_send_reinvite_response(struct ast_sip_session *session)
{
	pj_status_t status = PJ_SUCCESS;
	pjsip_tx_data *packet = NULL;

	if (session->inv_session->state == PJSIP_INV_STATE_DISCONNECTED) {
		ast_log(LOG_DEBUG, "Session is terminated\n");
		return;
	}

	if (strcmp(ast_channel_tech(session->channel)->type, "PJSIP")) {
		ast_log(LOG_DEBUG, "channel's type is not pjsip\n");
		return;
    }

	if (session->joint_ipvt) {
        char sdp_info[1024];
        memset(sdp_info, 0, sizeof(sdp_info));
        pjmedia_sdp_print(session->inv_session->neg->neg_local_sdp, sdp_info, sizeof(sdp_info) - 1);
        ast_log(LOG_NOTICE, "[IPVT] ast_sip_session_send_reinvite_response sdp:\n%s", sdp_info);
	}

	pjsip_dlg_inc_lock(session->inv_session->dlg);
	if (session->inv_session->invite_tsx) {
		status = pjsip_inv_answer(session->inv_session, 200, NULL, NULL, &packet);
		if (status == PJ_SUCCESS) {
			if (session->inv_session->neg && session->inv_session->neg->active_local_sdp) {
				char buf[1024] = {0};
				pjmedia_sdp_session* caller_answer_sdp = session->inv_session->neg->active_local_sdp;
				pjmedia_sdp_print(caller_answer_sdp, buf, sizeof(buf) - 1);// Must use sizeof(buf) - 1, otherwise there is a potential overflow
			}

			if (strncmp(ast_channel_name(session->channel), "PJSIP/trunk_", 12)) {
				ast_sip_add_timestamp_header(packet);
			}

			ast_sip_session_add_support_video_call_for_webrtc_trunk(session, packet);
		} else {
			ast_log(LOG_ERROR,"[session:%p] pjsip_inv_answer error, endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
		}
	} else {
		ast_log(LOG_ERROR,"Cannot answer reinvite '%s' because there is no associated SIP transaction\n",
			ast_channel_name(session->channel));
	}
	pjsip_dlg_dec_lock(session->inv_session->dlg);

	if (status == PJ_SUCCESS && packet) {
		ast_sip_session_send_response(session, packet);
	}

	return;
}

static int send_response_for_webrtc_trunk(struct ast_sip_session *session, pjsip_rx_data *rdata)
{
	int res = 0;
	pjsip_tx_data *tdata = NULL;

	if (!session || !session->channel) {
		return -1;
	}

	if (strncasecmp(ast_sorcery_object_get_id(session->endpoint), "webrtc_trunk", strlen("webrtc_trunk"))) {
		return -1;
	}

	if (pjsip_inv_initial_answer(session->inv_session, rdata, 100, NULL, NULL, &tdata) != PJ_SUCCESS) {
		pjsip_inv_terminate(session->inv_session, 500, PJ_TRUE);
		return -1;
	}

	ast_sip_session_send_reinvite_response(session);

	return res;
}

static pj_status_t session_inv_on_rx_reinvite(pjsip_inv_session *inv, const pjmedia_sdp_session *offer, pjsip_rx_data *rdata)
{
	pjmedia_sdp_session *caller_sdp_local = NULL;
	int send_reinvite = 0;
	struct ast_sip_session *session = NULL;
	int is_ssrc_changed = 0;
	const pj_str_t notify_nway_name = { "X-GS-Notify-Nway", 16 };
	pjsip_generic_string_hdr *notify_nway_hdr = NULL;
	pjsip_msg *msg = NULL;

	if (check_reinvite_params(inv, offer, rdata) == PJ_TRUE) {
		return PJ_TRUE;
	}

	msg = rdata->msg_info.msg;
	if (!msg) {
		return PJ_TRUE;
	}

	session = inv->mod_data[session_module.id];

	ast_log(LOG_DEBUG, "[session:%p] session_inv_on_rx_reinvite start for endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	if (session->channel) {
		pbx_builtin_setvar_helper(session->channel, "__CALL_INFO", NULL);
		pbx_builtin_setvar_helper(session->channel, "__CALL_INFO_MSG", NULL);

		/* Added by yfyang, for IPVT Server application */
		pbx_builtin_setvar_helper(session->channel, "__IPVT_CLIENT_INFO", NULL);
		pbx_builtin_setvar_helper(session->channel, "__IPVT_CLIENT_INFO_MSG", NULL);
	}

	/* to check sca held-private */
	if (session->line_index) {
		unsigned int index = 0;
		enum sca_status status = SCA_BUTT;

		ast_sip_get_sca_call_info_from_rdata(rdata, &index, &status);
		if (!index || (index != session->line_index)) {
			ast_log(LOG_NOTICE, "[session:%p] session_inv_on_rx_reinvite, the line index is different from the initial invite endpoint:%s\n",
				session, ast_sorcery_object_get_id(session->endpoint));
			return PJ_TRUE;
		}

		if (status == SCA_Held_Private) {
			session->recv_held_private = 1;
		} else {
			session->recv_held_private = 0;
		}
	}

	caller_sdp_local = session->inv_session->neg->active_local_sdp;

	set_yealink_flag(rdata, session);
	set_notify_participant_sdp_flag(rdata, session);

	/*Start added by zyxia for Bug 268897,2023.8.16*/
	notify_nway_hdr = pjsip_msg_find_hdr_by_name(msg, &notify_nway_name, NULL);
	if (notify_nway_hdr) {
		ast_log(LOG_DEBUG, "session_inv_on_rx_reinvite to channel '%s' need resend nway notify frame \n",ast_channel_name(session->channel));
		if (ast_sip_push_task(session->serializer, resend_nway_notify_frame, ao2_bump(session))) {
			ast_log(LOG_WARNING, "session_inv_on_rx_reinvite session '%s' resend nway notify failed with endpoint '%s'\n",
			ast_sorcery_object_get_id(session), ast_sorcery_object_get_id(session->endpoint));
			ao2_ref(session, -1);
		}
	}
	/*End added by zyxia for Bug 268897,2023.8.16*/

	/* ssrc only need to change for one time, hold maybe also a media change reinvite */
	is_ssrc_changed = change_ssrc_for_reinvite(session, offer);

	pj_mutex_unlock(inv->dlg->mutex_);
	send_reinvite = judge_if_need_send_reinvite_to_peer(session, offer, caller_sdp_local);
	pj_mutex_lock(inv->dlg->mutex_);
	if (send_reinvite) {
		return handle_for_reinvite_to_peer(session, rdata, is_ssrc_changed);
	}

	if (0 == send_response_for_webrtc_trunk(session, rdata)) {
		ast_log(LOG_DEBUG, "[session:%p] Already send reponse for webrtc trunk, endpoint:%s\n",
			session, ast_sorcery_object_get_id(session->endpoint));
		return PJ_SUCCESS;
	}

	ast_log(LOG_DEBUG, "[session:%p] Don't need to send reinvite to peer, endpoint:%s\n",
		session, ast_sorcery_object_get_id(session->endpoint));

	return PJ_TRUE;
}

static int check_channel_is_video_call(struct ast_channel *chan)
{
	int i;
	int active_stream_count = 0;
	int pend_stream_count = 0;
	struct ast_sip_channel_pvt *chan_pvt = ast_channel_tech_pvt(chan);
	struct ast_stream *active_stream;
	struct ast_stream *pend_stream;
	struct ast_sip_session *session;

	if (!chan_pvt || !chan_pvt->session) {
		return 0;
	}

	session = chan_pvt->session;
	ast_channel_lock(session->channel);
	if (session->active_media_state && session->active_media_state->topology) {
		active_stream_count = ast_stream_topology_get_count(session->active_media_state->topology);
		for(i=0; i<active_stream_count; i++) {
			active_stream = ast_stream_topology_get_stream(session->active_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)active_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(chan), i, ast_stream_get_type(active_stream),
				ast_stream_get_sub_type(active_stream), ast_stream_get_state(active_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(active_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(active_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(active_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(active_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}

	if (session->pending_media_state && session->pending_media_state->topology) {
		pend_stream_count = ast_stream_topology_get_count(session->pending_media_state->topology);
		for(i=0; i<pend_stream_count; i++) {
			pend_stream = ast_stream_topology_get_stream(session->pending_media_state->topology, i);
			ast_log(LOG_DEBUG, "(%s)pend_stream[%d] - type:%d, subtype:%d, status:%d",
				ast_channel_name(chan), i, ast_stream_get_type(pend_stream),
				ast_stream_get_sub_type(pend_stream), ast_stream_get_state(pend_stream));
			if (AST_MEDIA_TYPE_VIDEO == ast_stream_get_type(pend_stream)
				&& AST_MEDIA_SUB_VIDEO_TYPE_MAIN == ast_stream_get_sub_type(pend_stream)
				&& AST_STREAM_STATE_REMOVED != ast_stream_get_state(pend_stream)
				&& AST_STREAM_STATE_INACTIVE != ast_stream_get_state(pend_stream)) {
				ast_channel_unlock(session->channel);
				return 1;
			}
		}
	}
	ast_channel_unlock(session->channel);
	return 0;
}

#endif

static pjsip_inv_callback inv_callback = {
	.on_state_changed = session_inv_on_state_changed,
	.on_new_session = session_inv_on_new_session,
	.on_tsx_state_changed = session_inv_on_tsx_state_changed,
	.on_rx_offer = session_inv_on_rx_offer,
	.on_media_update = session_inv_on_media_update,
	.on_redirected = session_inv_on_redirected,
#ifdef GRANDSTREAM_NETWORKS
	/* Add by bfdu for video reinivite */
	.on_rx_reinvite = session_inv_on_rx_reinvite,
	.on_session_timer_update = session_inv_on_session_timer_update,
	.on_rx_reinvite_conflict = session_inv_on_rx_reinvite_conflict,
	.on_cancel_offer = session_inv_on_cancel_offer,
#endif
};

/*! \brief Hook for modifying outgoing messages with SDP to contain the proper address information */
static void session_outgoing_nat_hook(pjsip_tx_data *tdata, struct ast_sip_transport *transport)
{
	RAII_VAR(struct ast_sip_transport_state *, transport_state, ast_sip_get_transport_state(ast_sorcery_object_get_id(transport)), ao2_cleanup);
	struct ast_sip_nat_hook *hook = ast_sip_mod_data_get(
		tdata->mod_data, session_module.id, MOD_DATA_NAT_HOOK);
	struct pjmedia_sdp_session *sdp;
	int stream;

#ifndef BUILD_ON_CLOUD_PBX
	/* SDP produced by us directly will never be multipart */
	if (!transport_state || hook || !tdata->msg->body ||
		!ast_sip_is_content_type(&tdata->msg->body->content_type, "application", "sdp") ||
		ast_strlen_zero(transport->external_media_address)) {
		return;
	}
#else
	if (!transport_state || hook || !tdata->msg->body ||
		!ast_sip_is_content_type(&tdata->msg->body->content_type, "application", "sdp") ) {
		return;
	}
#endif
	sdp = tdata->msg->body->data;

	if (sdp->conn) {
		char host[NI_MAXHOST];
		struct ast_sockaddr our_sdp_addr = { { 0, } };

		ast_copy_pj_str(host, &sdp->conn->addr, sizeof(host));
		ast_sockaddr_parse(&our_sdp_addr, host, PARSE_PORT_FORBID);

		/* Reversed check here. We don't check the remote
		 * endpoint being in our local net, but whether our
		 * outgoing session IP is local. If it is, we'll do
		 * rewriting. No localnet configured? Always rewrite. */
#ifndef BUILD_ON_CLOUD_PBX
		if (ast_sip_transport_is_local(transport_state, &our_sdp_addr) || !transport_state->localnet) {
#endif
#ifndef GRANDSTREAM_NETWORKS
			ast_debug(5, "Setting external media address to %s\n", ast_sockaddr_stringify_host(&transport_state->external_media_address));
			pj_strdup2(tdata->pool, &sdp->conn->addr, ast_sockaddr_stringify_host(&transport_state->external_media_address));
#else
			/* Added by yfyang, get external ip from STUN */
			char ext_addr[256] = {0};

			if (transport->get_extaddr_from_stun) {
				ast_sip_nat_get_external_ip(ext_addr, sizeof(ext_addr)-1);
			}

			if (!strlen(ext_addr)) {
				if (ast_sockaddr_isnull(&transport_state->external_media_address)) {
					strncpy(ext_addr, transport->external_media_address, sizeof(ext_addr)-1);
				} else {
					strncpy(ext_addr, ast_sockaddr_stringify_host(&transport_state->external_media_address), sizeof(ext_addr)-1);
				}
			}

#ifdef BUILD_ON_CLOUD_PBX
			char *clouducm_host_ip = getenv("PUBLIC_IP");
			if (!ast_strlen_zero(clouducm_host_ip)) {
				strncpy(ext_addr, clouducm_host_ip, sizeof(ext_addr)-1);
				ast_log(LOG_DEBUG, "Get HOST_IP:%s for cloud UCM\n", clouducm_host_ip);
			}
#endif

			ast_log(LOG_DEBUG, "Setting external media address to %s\n", ext_addr);
			pj_strdup2(tdata->pool, &sdp->conn->addr, ext_addr);
#endif
			pj_strassign(&sdp->origin.addr, &sdp->conn->addr);
#ifndef BUILD_ON_CLOUD_PBX
		}
#endif
	}

	for (stream = 0; stream < sdp->media_count; ++stream) {
		/* See if there are registered handlers for this media stream type */
		char media[20];
		struct ast_sip_session_sdp_handler *handler;
		RAII_VAR(struct sdp_handler_list *, handler_list, NULL, ao2_cleanup);

		/* We need a null-terminated version of the media string */
		ast_copy_pj_str(media, &sdp->media[stream]->desc.media, sizeof(media));

		handler_list = ao2_find(sdp_handlers, media, OBJ_KEY);
		if (!handler_list) {
			ast_debug(1, "No registered SDP handlers for media type '%s'\n", media);
			continue;
		}
		AST_LIST_TRAVERSE(&handler_list->list, handler, next) {
			if (handler->change_outgoing_sdp_stream_media_address) {
				handler->change_outgoing_sdp_stream_media_address(tdata, sdp->media[stream], transport);
			}
		}
	}

	/* We purposely do this so that the hook will not be invoked multiple times, ie: if a retransmit occurs */
	ast_sip_mod_data_set(tdata->pool, tdata->mod_data, session_module.id, MOD_DATA_NAT_HOOK, nat_hook);
}

#ifdef GRANDSTREAM_NETWORKS
/*! \brief Convert SIP hangup causes to Asterisk hangup causes */
static int hangup_sip2cause(int cause)
{
	/* Possible values taken from causes.h */

	switch(cause) {
	case 401:       /* Unauthorized */
		return AST_CAUSE_CALL_REJECTED;
	case 403:       /* Not found */
		return AST_CAUSE_CALL_REJECTED;
	case 404:       /* Not found */
		return AST_CAUSE_UNALLOCATED;
	case 405:       /* Method not allowed */
		return AST_CAUSE_INTERWORKING;
	case 407:       /* Proxy authentication required */
		return AST_CAUSE_CALL_REJECTED;
	case 408:       /* No reaction */
		return AST_CAUSE_NO_USER_RESPONSE;
	case 409:       /* Conflict */
		return AST_CAUSE_NORMAL_TEMPORARY_FAILURE;
	case 410:       /* Gone */
		return AST_CAUSE_NUMBER_CHANGED;
	case 411:       /* Length required */
		return AST_CAUSE_INTERWORKING;
	case 413:       /* Request entity too large */
		return AST_CAUSE_INTERWORKING;
	case 414:       /* Request URI too large */
		return AST_CAUSE_INTERWORKING;
	case 415:       /* Unsupported media type */
		return AST_CAUSE_INTERWORKING;
	case 420:       /* Bad extension */
		return AST_CAUSE_NO_ROUTE_DESTINATION;
	case 480:       /* No answer */
		return AST_CAUSE_NO_ANSWER;
	case 481:       /* No answer */
		return AST_CAUSE_INTERWORKING;
	case 482:       /* Loop detected */
		return AST_CAUSE_INTERWORKING;
	case 483:       /* Too many hops */
		return AST_CAUSE_NO_ROUTE_DESTINATION;
	case 484:       /* Address incomplete */
		return AST_CAUSE_INVALID_NUMBER_FORMAT;
	case 485:       /* Ambiguous */
		return AST_CAUSE_UNALLOCATED;
	case 486:       /* Busy everywhere */
		return AST_CAUSE_BUSY;
	case 487:       /* Request terminated */
		return AST_CAUSE_INTERWORKING;
	case 488:       /* No codecs approved */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	case 491:       /* Request pending */
		return AST_CAUSE_INTERWORKING;
	case 493:       /* Undecipherable */
		return AST_CAUSE_INTERWORKING;
	case 500:       /* Server internal failure */
		return AST_CAUSE_FAILURE;
	case 501:       /* Call rejected */
		return AST_CAUSE_FACILITY_REJECTED;
	case 502:
		return AST_CAUSE_DESTINATION_OUT_OF_ORDER;
	case 503:       /* Service unavailable */
		return AST_CAUSE_CONGESTION;
	case 504:       /* Gateway timeout */
		return AST_CAUSE_RECOVERY_ON_TIMER_EXPIRE;
	case 505:       /* SIP version not supported */
		return AST_CAUSE_INTERWORKING;
	case 600:       /* Busy everywhere */
		return AST_CAUSE_USER_BUSY;
	case 603:       /* Decline */
		return AST_CAUSE_CALL_REJECTED;
	case 604:       /* Does not exist anywhere */
		return AST_CAUSE_UNALLOCATED;
	case 606:       /* Not acceptable */
		return AST_CAUSE_BEARERCAPABILITY_NOTAVAIL;
	default:
		if (cause < 500 && cause >= 400) {
			/* 4xx class error that is unknown - someting wrong with our request */
			return AST_CAUSE_INTERWORKING;
		} else if (cause < 600 && cause >= 500) {
			/* 5xx class error - problem in the remote end */
			return AST_CAUSE_CONGESTION;
		} else if (cause < 700 && cause >= 600) {
			/* 6xx - global errors in the 4xx class */
			return AST_CAUSE_INTERWORKING;
		}
		return AST_CAUSE_NORMAL;
	}
	/* Never reached */
	return 0;
}

static void ast_sip_report_invite_failed(struct ast_sip_session *session,struct pjsip_rx_data *rdata,struct pjsip_tx_data *tdata,
	const char *sipmethod, int resp, const char* resp_text)
{	
	if ((!strcmp(sipmethod,"INVITE")) || (!strcmp(sipmethod,"BYE") && 0 == resp) ){
		char reason_text[64] = {0};
		char peer_name[64] = {0};
		char from_user[256] = {0};
		char to_user[256] = {0};
		
		struct ast_security_event_sip_invite_failed sip_invite_failed = {
			.common.event_type = AST_SECURITY_EVENT_SIP_INVITE_FAILED,
			.common.version    = AST_SECURITY_EVENT_SIP_INVITE_FAILED_VERSION,
			.common.service    = "PJSIP",
			.is_outgoing = session->is_outgoing,
			.resp_code = resp,
			.reason_text = reason_text,
			.peer_name = peer_name,
			.from_user = from_user,
			.to_user = to_user,
		};



		pjsip_sip_uri *uri_to = NULL, *uri_from = NULL;

		if(!strcmp(sipmethod,"INVITE") && rdata != NULL){
			uri_to = pjsip_uri_get_uri(rdata->msg_info.to->uri);
			uri_from = pjsip_uri_get_uri(rdata->msg_info.from->uri);
		}else if( tdata != NULL ){
			pjsip_from_hdr *from_hd = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_FROM, NULL);
			pjsip_from_hdr *to_hd = pjsip_msg_find_hdr(tdata->msg, PJSIP_H_TO, NULL);
			if (strcmp(sipmethod,"INVITE") == 0)
			{
				if (session->inv_session->role == PJSIP_ROLE_UAC) {
					uri_from = pjsip_uri_get_uri(to_hd->uri);
					uri_to= pjsip_uri_get_uri(from_hd->uri);
				} else {
					uri_to = pjsip_uri_get_uri(to_hd->uri);
					uri_from = pjsip_uri_get_uri(from_hd->uri);
				}
			}else if(strcmp(sipmethod,"BYE") == 0){
				if (session->inv_session->role == PJSIP_ROLE_UAC) {
					uri_to = pjsip_uri_get_uri(to_hd->uri);
					uri_from = pjsip_uri_get_uri(from_hd->uri);
				} else {
					uri_from = pjsip_uri_get_uri(to_hd->uri);
					uri_to= pjsip_uri_get_uri(from_hd->uri);
				}
			}
		} else {
			ast_log(LOG_ERROR,"ast_sip_report_invite_failed error...");
		}

		if(uri_to != NULL && uri_from != NULL){
			ast_copy_pj_str(to_user, &uri_to->user, sizeof(to_user));
			ast_copy_pj_str(from_user, &uri_from->user, sizeof(from_user));
		}

		ast_log(LOG_DEBUG,"to:%s,from:%s",to_user,from_user);

		ast_copy_string(reason_text, resp_text, sizeof(reason_text));
		ast_copy_string(peer_name, ast_sorcery_object_get_id(session->endpoint), sizeof(peer_name));
		ast_log(LOG_VERBOSE, "endpoint: %s, peer_name: %s, from_user: %s, to_user: %s, resp code: %d %s", ast_sorcery_object_get_id(session->endpoint),peer_name, from_user, to_user, resp, reason_text);

		ast_security_event_report(AST_SEC_EVT(&sip_invite_failed));
	}
}
/*end*/

int ast_sip_session_delay_request(struct ast_sip_session *session)
{
	struct ast_sip_session_delayed_request *delay_request = NULL;
	int count = 0;

	AST_LIST_TRAVERSE(&session->delayed_requests, delay_request, next) {
		count++;
	}

	return count;
}

static char *cli_show_sip_sessions(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	struct ast_sip_session *session = NULL;
	char call_id[128] = {0};

#ifdef GRANDSTREAM_NETWORKS
	int count = 0;
#endif

	switch (cmd) {
	case CLI_INIT:
		e->command = "pjsip show sessions";
		e->usage = "Usage: pjsip show sessions\n";
		return NULL;
	case CLI_GENERATE:
		break;
	}

	if (a->argc < 3) {
		return NULL;
	}

	ast_cli(a->fd, "Session info:\n");
	ast_cli(a->fd, "%-12s %-18s %-18s %-64s %-18s %-8s\n", "Object", "Endpoint", "Session", "Dialog", "Channel", "Delay-Num");
	AST_LIST_LOCK(&sip_sessions);
	AST_LIST_TRAVERSE(&sip_sessions, session, next) {
		call_id[0] = '\0';
		if (session->inv_session) {
			snprintf(call_id, sizeof(call_id), "%.*s", (int)session->inv_session->dlg->call_id->id.slen, session->inv_session->dlg->call_id->id.ptr);
		}
		ast_cli(a->fd, "%p %-18s %-18s %-64s %-18s %d\n", session, session->endpoint ? ast_sorcery_object_get_id(session->endpoint) : "<none>", 
			session->inv_session ? session->inv_session->obj_name : "<none>",
			!ast_strlen_zero(call_id) ? call_id: "<none>",
			session->channel ? ast_channel_name(session->channel) : "<none>",
			ast_sip_session_delay_request(session));

#ifdef GRANDSTREAM_NETWORKS
		count++;
#endif
	}
	AST_LIST_UNLOCK(&sip_sessions);

#ifdef GRANDSTREAM_NETWORKS
	ast_cli(a->fd, "\nTotal %d session!\n", count);
#endif

	return CLI_SUCCESS;
}

static struct ast_cli_entry cli_commands[] = {
	AST_CLI_DEFINE(cli_show_sip_sessions, "List sip sessions")
};

static int action_whiteboard(struct mansession *s, const struct message *m)
{
	const char *supported = astman_get_header(m, "Supported");
	if (ast_strlen_zero(supported)) {
		astman_send_error(s, m, "No supported field");
		return AMI_SUCCESS;
	}

	ast_log(LOG_DEBUG, "action_whiteboard supported:%s\n", supported);

	if (!strcasecmp(supported, "yes")) {
		im_support_whiteboard = 1;
	} else {
		im_support_whiteboard = 0;
	}

	astman_send_ack(s, m, "push whiteboard status succeed");

	return AMI_SUCCESS;
}

static void get_all_local_ip(void)
{
	/* Max is the local_address array size */
	local_address_count = PJ_ARRAY_SIZE(local_address);

	pj_enum_ip_interface(pj_AF_UNSPEC(), &local_address_count, local_address);

	return;
}
#endif

static int load_module(void)
{
	pjsip_endpoint *endpt;

	if (!ast_sip_get_sorcery() || !ast_sip_get_pjsip_endpoint()) {
		return AST_MODULE_LOAD_DECLINE;
	}
	if (!(nat_hook = ast_sorcery_alloc(ast_sip_get_sorcery(), "nat_hook", NULL))) {
		return AST_MODULE_LOAD_DECLINE;
	}
	nat_hook->outgoing_external_message = session_outgoing_nat_hook;
	ast_sorcery_create(ast_sip_get_sorcery(), nat_hook);
	sdp_handlers = ao2_container_alloc_hash(AO2_ALLOC_OPT_LOCK_MUTEX, 0,
		SDP_HANDLER_BUCKETS, sdp_handler_list_hash, NULL, sdp_handler_list_cmp);
	if (!sdp_handlers) {
		return AST_MODULE_LOAD_DECLINE;
	}
	endpt = ast_sip_get_pjsip_endpoint();
	pjsip_inv_usage_init(endpt, &inv_callback);
	pjsip_100rel_init_module(endpt);
	pjsip_timer_init_module(endpt);
	if (ast_sip_register_service(&session_module)) {
		return AST_MODULE_LOAD_DECLINE;
	}
	ast_sip_register_service(&session_reinvite_module);
	ast_sip_register_service(&outbound_invite_auth_module);

#ifdef GRANDSTREAM_NETWORKS
	session_park_pos_get(0);
	ast_cli_register_multiple(cli_commands, ARRAY_LEN(cli_commands));

	/* Add by bfdu for whiteboard, 2022.1.7 */
	ast_manager_register_xml("Whiteboard", EVENT_FLAG_CONFIG, action_whiteboard);

	get_all_local_ip();

	g_vlan_wan_enable = ast_sockaddr_voip_vlan_addr_get(sip_vlan_addr, sizeof(sip_vlan_addr));
    if (g_vlan_wan_enable) {
		ast_sockaddr_route_lan_addr_get(sip_lan_addr, sizeof(sip_lan_addr));
    }
#endif
	ast_module_shutdown_ref(ast_module_info->self);

	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	ast_sip_unregister_service(&outbound_invite_auth_module);
	ast_sip_unregister_service(&session_reinvite_module);
	ast_sip_unregister_service(&session_module);
	ast_sorcery_delete(ast_sip_get_sorcery(), nat_hook);
	ao2_cleanup(nat_hook);
	ao2_cleanup(sdp_handlers);
#ifdef GRANDSTREAM_NETWORKS
	ast_cli_unregister_multiple(cli_commands, ARRAY_LEN(cli_commands));
#endif
	return 0;
}

#ifdef GRANDSTREAM_NETWORKS
static int reload_module(void)
{
	pjsip_endpoint *endpt;
	static pj_str_t STR_100REL = {"100rel", 6};
	static pj_str_t STR_TIMER = {"timer", 5};

	endpt = ast_sip_get_pjsip_endpoint();

	if (!pjsip_endpt_has_capability(endpt, PJSIP_H_SUPPORTED, NULL, &STR_TIMER)) {
		pjsip_endpt_add_capability(endpt, NULL, PJSIP_H_SUPPORTED, NULL, 1, &STR_TIMER);
	}

	if (!pjsip_endpt_has_capability(endpt, PJSIP_H_SUPPORTED, NULL, &STR_100REL)) {
		pjsip_endpt_add_capability(endpt, NULL, PJSIP_H_SUPPORTED, NULL, 1, &STR_100REL);
	}

	session_park_pos_get(1);   /* Added by yfyang, for Bug107609 */

	return AST_MODULE_LOAD_SUCCESS;
}
#endif

AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS | AST_MODFLAG_LOAD_ORDER, "PJSIP Session resource",
	.support_level = AST_MODULE_SUPPORT_CORE,
	.load = load_module,
	.unload = unload_module,
#ifdef GRANDSTREAM_NETWORKS
	.reload = reload_module,
#endif
	.load_pri = AST_MODPRI_APP_DEPEND,
	.requires = "res_pjsip",
);


#if 0
 int ast_sip_record_hdr_from_msg(struct ast_channel *chan, void *msg, void *xml_cdr)
{
	switch_xml_t cdr;
	char buf[3072] = {0};
	switch_xml_t vars,x_field;
	pjsip_msg *sip_msg;
	pjsip_from_hdr *from_hdr = NULL;
	
	pjsip_from_hdr *to_hdr = NULL;
	pjsip_cid_hdr *cid = NULL;
	pjsip_cseq_hdr *cseq = NULL;
	char cseq_str[16] = {0};
	const char *page;
	
	char *xml_cdr_text, *sql;

	char tmp[512] = {0};

	if(!chan || !msg)
	{
		return -1;
	}

	if(HA_ROLE_ACTIVE != ast_ha_get_role())
	{
		return 0;
	}

	if(!ast_test_flag(ast_channel_flags(chan), AST_FLAG_TRACKED))
	{
		return 0;
	}


	if(NULL != ast_channel_appl(chan))
	{
		if(0 == strcmp(ast_channel_appl(chan), "Park"))
		{
			return 0;
		}
	}

	page = pbx_builtin_getvar_helper((struct ast_channel *)chan, "PAGE_INTERCOM");
	if(page && 0 == strcasecmp(page,"yes"))
	{
		return 0;
	}

	if(0 == strcmp(ast_channel_context(chan),"macro-vm") && 0 == strcmp(ast_channel_exten(chan),"s-BUSY"))
	{
		return 0;
	}

	if(xml_cdr)
	{
		cdr = (switch_xml_t)xml_cdr;
	}
	else
	{
		if(0 != ast_recovery_db_get(ast_channel_linkedid( chan), ast_channel_uniqueid( chan), buf, sizeof(buf)-1))
		{
			return 0;
		}
		
		// restore database to xml struct, need free
		cdr = switch_xml_parse_str_dynamic(buf, SWITCH_TRUE);
		vars = switch_xml_child(cdr, "variables");
		sip_msg = (pjsip_msg *)msg;
		
		from_hdr = pjsip_msg_find_hdr(sip_msg, PJSIP_H_FROM, NULL);
		if (from_hdr){
			x_field = switch_xml_child(vars, "from_tag");
			ast_copy_pj_str(tmp, &from_hdr->tag, pj_strlen(&from_hdr->tag)+1);
			switch_xml_set_txt_d(x_field, tmp);
		}

		
		to_hdr = pjsip_msg_find_hdr(sip_msg, PJSIP_H_TO, NULL);
		if(to_hdr){
			x_field = switch_xml_child(vars, "to_tag");
			ast_copy_pj_str(tmp, &to_hdr->tag, pj_strlen(&to_hdr->tag)+1);
			switch_xml_set_txt_d(x_field, tmp);

		}

		cid = pjsip_msg_find_hdr(sip_msg, PJSIP_H_CALL_ID, NULL);
		if(cid){
			x_field = switch_xml_child(vars, "call_id");
			ast_copy_pj_str(tmp, &cid->id, pj_strlen(&cid->id)+1);
			switch_xml_set_txt_d(x_field, tmp);
		}

		cseq = pjsip_msg_find_hdr(sip_msg, PJSIP_H_CSEQ, NULL);
		if(cseq){
			x_field = switch_xml_child(vars, "cseq");
			sprintf(cseq_str,"%d",cseq->cseq);
			switch_xml_set_txt_d(x_field,cseq_str);
		}

		struct ast_sip_channel_pvt *channel = ast_channel_tech_pvt(chan);
		if(!channel || !channel->session)
		{
			return -1;
		}
		if(channel->session->active_contact_addr)
		{
			x_field = switch_xml_child(vars, "contact_host");
			switch_xml_set_txt_d(x_field,channel->session->active_contact_addr);
		}
		xml_cdr_text = switch_xml_toxml_nolock(cdr, SWITCH_FALSE);
		switch_xml_free(cdr);
		

		if(xml_cdr_text)
		{
			if (ast_test_flag(ast_channel_flags(chan), AST_FLAG_TRACKED)) {
				sql = switch_xml_mprintf("update recovery set metadata='%q' where uuid='%q'",  xml_cdr_text, ast_channel_uniqueid(chan));
			} else {
				sql = switch_xml_mprintf("insert into recovery (runtime_uuid, technology, profile_name, hostname, uuid, metadata) "
									 "values ('%q','%q','%q','%q','%q','%q')",
									 ast_channel_linkedid(chan), "pjsip",
									 "internal", "asterisk", ast_channel_uniqueid(chan), xml_cdr_text);
			}
			if (sql) {
				ast_log(LOG_NOTICE, "cmd is %s\n", sql);
				ast_recovery_db_put(sql, 1, 1);
				ast_free(sql);
			}

			ast_free(xml_cdr_text);
			ast_channel_lock(chan);
			ast_set_flag(ast_channel_flags(chan), AST_FLAG_TRACKED);
			ast_channel_unlock(chan);
		}
	}

	return 0;
}
#endif



